Package: rxode2
Version: 5.1.5
Title: Facilities for Simulating from ODE-Based Models
Authors@R: c(
   person("Matthew L.","Fidler", role = c("aut", "cre"), email = "matthew.fidler@gmail.com", comment=c(ORCID="0000-0001-8538-6691")),
   person("Wenping", "Wang", role = c("aut"), email = "wwang8198@gmail.com"),
   person("Aaron", "Collier", role="ctb", comment="SUNDIALS/CVODE"),
   person("Alan", "Hindmarsh",role="ctb"),
   person("Arun", "Srinivasan", role="ctb", email="arunkumar.sriniv@gmail.com"),
   person("Ashley", "Crawford", role="ctb", comment="SUNDIALS/CVODE"),
   person("Awad H.", "Al-Mohy",role="ctb"),
   person("Bill", "Denney", email="wdenney@humanpredictions.com", role="ctb", comment=c(ORCID="0000-0002-5759-428X")),
   person("Cleve", "Moler", role="ctb"),
   person("Cody J.", "Balos", role="ctb", comment="SUNDIALS/CVODE"),
   person("Dan", "Shumaker", role="ctb", comment="SUNDIALS/CVODE"),
   person("Daniel", "Kaschek", email="daniel.kaschek@gmail.com", role="ctb"),
   person("Daniel R.", "Reynolds", role="ctb", comment="SUNDIALS/CVODE"),
   person("David", "Cooley", role="ctb"),
   person("David J.", "Gardner", role="ctb", comment="SUNDIALS/CVODE"),
   person("Drew", "Schmidt", role="ctb"),
   person("Ernst", "Hairer", role="ctb"),
   person("Gabriel", "Staples", role="ctb"),
   person("Gerhard", "Wanner", role="ctb"),
   person("Gilbert", "Stewart", role="ctb"),
   person("Goro", "Fuji",  role="ctb", email="gfuji@cpan.org"),
   person("Hadley", "Wickham", role="ctb"),
   person("Hidde", "van de Beek", role="ctb", email="hiddevdbeek@icloud.com"),
   person("Igor", "Kushnir", email="igorkuo@gmail.com", role="ctb"),
   person("Imperial College of Science, Technology and Medicine", role="cph"),
   person("Jack", "Dongarra", role="ctb"),
   person("Jacob", "Williams", role="ctb"),
   person("Jim", "Bunch", role="ctb"),
   person("Karline", "Soetaert", role="ctb"),
   person("Kevin", "Ushey", role="ctb"),
   person("Lawrence Livermore National Security", role="cph", comment="SUNDIALS/CVODE"),
   person("Linda", "Petzold", role="ctb"),
   person("Mark M.", "Baum", role="ctb"),
   person("Martin", "Maechler", role="ctb"),
   person("Matt", "Dowle", role="ctb", email="mattjdowle@gmail.com"),
   person("Matteo", "Fasiolo", email = "matteo.fasiolo@gmail.com", role = "ctb"),
   person("Melissa", "Hallow", role = "aut", email = "hallowkm@uga.edu"),
   person("Michel", "Lang", , "michellang@gmail.com", role = "ctb"),
   person("Morwenn","", role="ctb"),
   person("Nicholas J.", "Higham",role="ctb"),
   person("Omar", "Elashkar", email="omar.i.elashkar@gmail.com", role="ctb"),
   person("Peter N.", "Brown", role="ctb"),
   person("Radu", "Serban", role="ctb", comment="SUNDIALS/CVODE"),
   person("Rich", "FitzJohn", role="ctb", email="rich.fitzjohn@gmail.com"),
   person("Richard", "Upton", role = "ctb"),
   person("Roger B.", "Sidje", role="ctb"),
   person("Scott D.", "Cohen", role="ctb", comment="SUNDIALS/CVODE"),
   person("Simon", "Frost", role="ctb"),
   person("Slaven", "Peles", role="ctb", comment="SUNDIALS/CVODE"),
   person("Southern Methodist University", role="cph", comment="SUNDIALS/CVODE"),
   person("Thomas", "Petzoldt", role="ctb"),
   person("University of Maryland Baltimore County", role="cph", comment="SUNDIALS/CVODE"),
   person("Wes", "Hinsley", role="ctb"),
   person("Woodrow",  "Setzer", role="ctb"),
   person("Yu", "Feng", role="ctb"),
   person("Zufar", "Mulyukov", role="ctb", email="zufar.mulyukov@novartis.com"))
Maintainer: Matthew L. Fidler <matthew.fidler@gmail.com>
Depends: R (>= 4.1.0)
Suggests: Matrix, DT, covr, crayon, curl, digest, dplyr (>= 0.8.0),
        ggrepel, gridExtra, htmltools, knitr, learnr, microbenchmark,
        nlme, remotes, rlang, rmarkdown, scales, shiny, stringi,
        symengine, testthat, tidyr, usethis, withr, xgxr, pillar,
        tibble, units (>= 0.6-0), rsconnect, devtools, patchwork,
        nlmixr2data, lifecycle, kableExtra, pmxTools, rootSolve, arrow,
        fst, duckdb, DBI
Imports: PreciseSums (>= 0.7), Rcpp (>= 0.12.3), RcppParallel,
        backports, cli (>= 2.0.0), checkmate, ggplot2 (>= 3.4.0),
        inline, lotri (>= 1.0.4), memoise, methods, mirai, rex, sys,
        tools, utils, dparser (>= 1.3.1-12), qs2, rxode2ll, data.table
        (>= 1.12.4), vctrs, stats
Description: Facilities for running simulations from ordinary
    differential equation ('ODE') models, such as pharmacometrics and other
    compartmental models.  A compilation manager translates the ODE model
    into C, compiles it, and dynamically loads the object code into R for
    improved computational efficiency.  An event table object facilitates
    the specification of complex dosing regimens (optional) and sampling
    schedules.  NB: The use of this package requires both C and
    Fortran compilers, for details on their use with R please see
    Section 6.3, Appendix A, and Appendix D in the "R Administration and
    Installation" manual. Also the code is mostly released under GPL.  The
    'VODE' and 'LSODA' are in the public domain.  The vendored 'SUNDIALS'
    'CVODE' sources and headers are released under the BSD-3-Clause license.
    The information is available in the inst/COPYRIGHTS.
BugReports: https://github.com/nlmixr2/rxode2/issues/
NeedsCompilation: yes
VignetteBuilder: knitr
License: GPL (>= 3)
URL: https://nlmixr2.github.io/rxode2/,
        https://github.com/nlmixr2/rxode2/
Biarch: true
LinkingTo: sitmo, lotri (>= 1.0.0), PreciseSums (>= 0.7), Rcpp,
        RcppArmadillo (>= 0.9.300.2.0), BH, RcppParallel, RcppEigen (>=
        0.3.3.9.2), dparser (>= 1.3.1-12), StanHeaders (>= 2.21.0.7)
Encoding: UTF-8
LazyData: true
Language: en-US
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
Packaged: 2026-07-28 02:55:20 UTC; matt-fidler
Author: Matthew L. Fidler [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-8538-6691>),
  Wenping Wang [aut],
  Aaron Collier [ctb] (SUNDIALS/CVODE),
  Alan Hindmarsh [ctb],
  Arun Srinivasan [ctb],
  Ashley Crawford [ctb] (SUNDIALS/CVODE),
  Awad H. Al-Mohy [ctb],
  Bill Denney [ctb] (ORCID: <https://orcid.org/0000-0002-5759-428X>),
  Cleve Moler [ctb],
  Cody J. Balos [ctb] (SUNDIALS/CVODE),
  Dan Shumaker [ctb] (SUNDIALS/CVODE),
  Daniel Kaschek [ctb],
  Daniel R. Reynolds [ctb] (SUNDIALS/CVODE),
  David Cooley [ctb],
  David J. Gardner [ctb] (SUNDIALS/CVODE),
  Drew Schmidt [ctb],
  Ernst Hairer [ctb],
  Gabriel Staples [ctb],
  Gerhard Wanner [ctb],
  Gilbert Stewart [ctb],
  Goro Fuji [ctb],
  Hadley Wickham [ctb],
  Hidde van de Beek [ctb],
  Igor Kushnir [ctb],
  Imperial College of Science, Technology and Medicine [cph],
  Jack Dongarra [ctb],
  Jacob Williams [ctb],
  Jim Bunch [ctb],
  Karline Soetaert [ctb],
  Kevin Ushey [ctb],
  Lawrence Livermore National Security [cph] (SUNDIALS/CVODE),
  Linda Petzold [ctb],
  Mark M. Baum [ctb],
  Martin Maechler [ctb],
  Matt Dowle [ctb],
  Matteo Fasiolo [ctb],
  Melissa Hallow [aut],
  Michel Lang [ctb],
  Morwenn [ctb],
  Nicholas J. Higham [ctb],
  Omar Elashkar [ctb],
  Peter N. Brown [ctb],
  Radu Serban [ctb] (SUNDIALS/CVODE),
  Rich FitzJohn [ctb],
  Richard Upton [ctb],
  Roger B. Sidje [ctb],
  Scott D. Cohen [ctb] (SUNDIALS/CVODE),
  Simon Frost [ctb],
  Slaven Peles [ctb] (SUNDIALS/CVODE),
  Southern Methodist University [cph] (SUNDIALS/CVODE),
  Thomas Petzoldt [ctb],
  University of Maryland Baltimore County [cph] (SUNDIALS/CVODE),
  Wes Hinsley [ctb],
  Woodrow Setzer [ctb],
  Yu Feng [ctb],
  Zufar Mulyukov [ctb]
Repository: CRAN
Date/Publication: 2026-07-28 10:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("$",rxEt)
S3method("$",rxSolve)
S3method("$",rxSolveCovs)
S3method("$",rxSolveOom)
S3method("$",rxSolveParams)
S3method("$",rxSolveSimType)
S3method("$",rxSymInvCholEnv)
S3method("$",rxUi)
S3method("$<-",rxEt)
S3method("$<-",rxEtPreview)
S3method("$<-",rxSolve)
S3method("$<-",rxSymInvCholEnv)
S3method("$<-",rxUi)
S3method("+",rxSolve)
S3method("+",solveRxDll)
S3method("[",rxEt)
S3method("[",rxEtPreview)
S3method("[",rxEvid)
S3method("[",rxRateDur)
S3method("[",rxSolve)
S3method("[<-",rxEtPreview)
S3method("[<-",rxEvid)
S3method("[<-",rxRateDur)
S3method("[<-",rxSolve)
S3method("[[",rxEvid)
S3method("[[",rxRateDur)
S3method("[[",rxSolve)
S3method("[[<-",rxEtPreview)
S3method("[[<-",rxSolve)
S3method("dimnames<-",rxSolve)
S3method("ini<-",default)
S3method("model<-",default)
S3method("rxode2<-","function")
S3method("rxode2<-",default)
S3method("units<-",rxEvid)
S3method(.DollarNames,rxEt)
S3method(.DollarNames,rxSolve)
S3method(.DollarNames,rxUi)
S3method(as.arrow,rxSolveOom)
S3method(as.character,rxEvid)
S3method(as.character,rxRateDur)
S3method(as.data.frame,rxEt)
S3method(as.data.frame,rxEtTran)
S3method(as.data.frame,rxEvid)
S3method(as.data.frame,rxRateDur)
S3method(as.data.frame,rxSolveOom)
S3method(as.data.table,rxEt)
S3method(as.data.table,rxSolveOom)
S3method(as.et,default)
S3method(as.function,character)
S3method(as.function,rxModelVars)
S3method(as.function,rxUi)
S3method(as.function,rxode2)
S3method(as.function,rxode2tos)
S3method(as.ini,call)
S3method(as.ini,character)
S3method(as.ini,data.frame)
S3method(as.ini,default)
S3method(as.ini,lotriFix)
S3method(as.ini,matrix)
S3method(as.model,call)
S3method(as.model,character)
S3method(as.model,default)
S3method(as.model,list)
S3method(as.rxUi,"function")
S3method(as.rxUi,default)
S3method(as.rxUi,rxModelVars)
S3method(as.rxUi,rxUi)
S3method(as.rxUi,rxode2)
S3method(as.rxUi,rxode2tos)
S3method(binomProbs,default)
S3method(c,rxEt)
S3method(c,rxEvid)
S3method(c,rxRateDur)
S3method(coef,"function")
S3method(coef,rxUi)
S3method(coef,rxode2)
S3method(confint,rxSolve)
S3method(dim,rxEt)
S3method(dim,rxSolveOom)
S3method(dimnames,rxSolve)
S3method(et,"function")
S3method(et,default)
S3method(et,rxParams)
S3method(et,rxSolve)
S3method(et,rxUi)
S3method(et,rxode2)
S3method(format,boundInits)
S3method(format,boundParams)
S3method(format,rxEvid)
S3method(format,rxRateDur)
S3method(format,rxSolveSimType)
S3method(getBaseIniSimModel,default)
S3method(getBaseSimModel,default)
S3method(getBaseSymengineModel,default)
S3method(head,rxSolveOom)
S3method(ini,default)
S3method(ini,rxUi)
S3method(meanProbs,default)
S3method(model,"function")
S3method(model,default)
S3method(model,rxModelVars)
S3method(model,rxUi)
S3method(model,rxode2)
S3method(modelExtract,"function")
S3method(modelExtract,default)
S3method(modelExtract,rxModelVars)
S3method(modelExtract,rxUi)
S3method(modelExtract,rxode2)
S3method(names,rxEt)
S3method(plot,rxSolve)
S3method(plot,rxSolveConfint1)
S3method(plot,rxSolveConfint2)
S3method(predict,"function")
S3method(predict,rxEt)
S3method(predict,rxParams)
S3method(predict,rxSolve)
S3method(predict,rxUi)
S3method(predict,rxode2)
S3method(print,boundInits)
S3method(print,boundParams)
S3method(print,rxC)
S3method(print,rxCoef)
S3method(print,rxDll)
S3method(print,rxEt)
S3method(print,rxEtPreview)
S3method(print,rxEtTran)
S3method(print,rxEvid)
S3method(print,rxHidden)
S3method(print,rxMemoryEstimate)
S3method(print,rxModelText)
S3method(print,rxModelVars)
S3method(print,rxRateDur)
S3method(print,rxSolve)
S3method(print,rxSolveCovs)
S3method(print,rxSolveInits)
S3method(print,rxSolveOom)
S3method(print,rxSolveParams)
S3method(print,rxSolveSimType)
S3method(print,rxSymInvCholEnv)
S3method(print,rxUi)
S3method(print,rxode2)
S3method(rbind,rxEt)
S3method(rbind,rxSolve)
S3method(rep,rxEt)
S3method(rxChain2,EventTable)
S3method(rxChain2,default)
S3method(rxCompile,character)
S3method(rxCompile,rxDll)
S3method(rxCompile,rxModelVars)
S3method(rxCompile,rxode2)
S3method(rxEtDispatchSolve,default)
S3method(rxEtDispatchSolve,rxode2et)
S3method(rxGetDistributionSimulationLines,cauchy)
S3method(rxGetDistributionSimulationLines,default)
S3method(rxGetDistributionSimulationLines,dnorm)
S3method(rxGetDistributionSimulationLines,norm)
S3method(rxGetDistributionSimulationLines,ordinal)
S3method(rxGetDistributionSimulationLines,rxUi)
S3method(rxGetDistributionSimulationLines,t)
S3method(rxHtml,rxSolve)
S3method(rxModelVarsS3,default)
S3method(rxModelVarsS3,rxUi)
S3method(rxParams,default)
S3method(rxParams,rxEt)
S3method(rxParams,rxSolve)
S3method(rxParams,rxode2)
S3method(rxRename,"function")
S3method(rxRename,default)
S3method(rxRename,rxUi)
S3method(rxShiny,default)
S3method(rxShiny,rxSolve)
S3method(rxSolve,"function")
S3method(rxSolve,default)
S3method(rxSolve,nlmixr2FitCore)
S3method(rxSolve,nlmixr2FitData)
S3method(rxSolve,rxSolve)
S3method(rxSolve,rxUi)
S3method(rxSolve,rxode2tos)
S3method(rxTrans,character)
S3method(rxTrans,default)
S3method(rxUdfUi,bolus)
S3method(rxUdfUi,default)
S3method(rxUdfUi,evid_)
S3method(rxUdfUi,expit)
S3method(rxUdfUi,infuse)
S3method(rxUdfUi,infuseDur)
S3method(rxUdfUi,linMod)
S3method(rxUdfUi,linMod0)
S3method(rxUdfUi,linModA)
S3method(rxUdfUi,linModA0)
S3method(rxUdfUi,linModB)
S3method(rxUdfUi,linModB0)
S3method(rxUdfUi,linModD)
S3method(rxUdfUi,linModD0)
S3method(rxUdfUi,linModM)
S3method(rxUdfUi,linModM0)
S3method(rxUdfUi,logit)
S3method(rxUdfUi,mix)
S3method(rxUdfUi,multiply)
S3method(rxUdfUi,obs)
S3method(rxUdfUi,phantom)
S3method(rxUdfUi,plogis)
S3method(rxUdfUi,replace)
S3method(rxUdfUi,ribeta)
S3method(rxUdfUi,ribinom)
S3method(rxUdfUi,ricauchy)
S3method(rxUdfUi,richisq)
S3method(rxUdfUi,riexp)
S3method(rxUdfUi,rif)
S3method(rxUdfUi,rigamma)
S3method(rxUdfUi,rigeom)
S3method(rxUdfUi,rinbinom)
S3method(rxUdfUi,rinorm)
S3method(rxUdfUi,ripois)
S3method(rxUdfUi,rit)
S3method(rxUdfUi,riunif)
S3method(rxUdfUi,riweibull)
S3method(rxUdfUi,rxbeta)
S3method(rxUdfUi,rxbinom)
S3method(rxUdfUi,rxcauchy)
S3method(rxUdfUi,rxchisq)
S3method(rxUdfUi,rxexp)
S3method(rxUdfUi,rxf)
S3method(rxUdfUi,rxgamma)
S3method(rxUdfUi,rxgeom)
S3method(rxUdfUi,rxnbinom)
S3method(rxUdfUi,rxnorm)
S3method(rxUdfUi,rxpois)
S3method(rxUdfUi,rxt)
S3method(rxUdfUi,rxunif)
S3method(rxUdfUi,rxweibull)
S3method(rxUiDeparse,default)
S3method(rxUiDeparse,lotriFix)
S3method(rxUiDeparse,rxControl)
S3method(rxUiGet,allCovs)
S3method(rxUiGet,cmtLines)
S3method(rxUiGet,covLhs)
S3method(rxUiGet,default)
S3method(rxUiGet,dvidLine)
S3method(rxUiGet,errParams)
S3method(rxUiGet,etaLhs)
S3method(rxUiGet,fun)
S3method(rxUiGet,funPartsDigest)
S3method(rxUiGet,funPrint)
S3method(rxUiGet,funTxt)
S3method(rxUiGet,ini)
S3method(rxUiGet,iniFun)
S3method(rxUiGet,interpLines)
S3method(rxUiGet,levelLhs)
S3method(rxUiGet,levels)
S3method(rxUiGet,lhsCov)
S3method(rxUiGet,lhsEta)
S3method(rxUiGet,lhsTheta)
S3method(rxUiGet,lhsVar)
S3method(rxUiGet,lstChr)
S3method(rxUiGet,md5)
S3method(rxUiGet,model)
S3method(rxUiGet,modelDesc)
S3method(rxUiGet,modelFun)
S3method(rxUiGet,muRefTable)
S3method(rxUiGet,multipleEndpoint)
S3method(rxUiGet,mvFromExpression)
S3method(rxUiGet,omega)
S3method(rxUiGet,paramsLine)
S3method(rxUiGet,props)
S3method(rxUiGet,sha1)
S3method(rxUiGet,simulationIniModel)
S3method(rxUiGet,simulationModel)
S3method(rxUiGet,simulationSigma)
S3method(rxUiGet,splitDose)
S3method(rxUiGet,splitDoseLines)
S3method(rxUiGet,state)
S3method(rxUiGet,stateDf)
S3method(rxUiGet,statePropDf)
S3method(rxUiGet,symengineModelNoPrune)
S3method(rxUiGet,symengineModelPrune)
S3method(rxUiGet,theta)
S3method(rxUiGet,thetaLhs)
S3method(rxUiGet,thetaLower)
S3method(rxUiGet,thetaUpper)
S3method(rxUiGet,varLhs)
S3method(seq,rxEt)
S3method(simulate,rxEt)
S3method(simulate,rxParams)
S3method(simulate,rxSolve)
S3method(simulate,rxode2)
S3method(solve,"function")
S3method(solve,rxEt)
S3method(solve,rxParams)
S3method(solve,rxSolve)
S3method(solve,rxUi)
S3method(solve,rxode2)
S3method(str,rxEt)
S3method(str,rxHidden)
S3method(str,rxSolve)
S3method(str,rxSymInvCholEnv)
S3method(str,rxUi)
S3method(str,rxode2)
S3method(summary,rxC)
S3method(summary,rxDll)
S3method(summary,rxSolve)
S3method(summary,rxode2)
S3method(t,rxSolve)
S3method(update,"function")
S3method(update,rxSolve)
S3method(update,rxUi)
S3method(vctrs::vec_cast,rxEt.tbl_df)
S3method(vctrs::vec_cast,tbl_df.rxEt)
S3method(vctrs::vec_ptype2,rxEt.tbl_df)
S3method(vctrs::vec_ptype2,tbl_df.rxEt)
S3method(vec_cast,data.frame.rxEt)
S3method(vec_cast,data.table.rxEt)
S3method(vec_cast,rxEt.data.frame)
S3method(vec_cast,rxEt.data.table)
S3method(vec_cast,rxEt.rxEt)
S3method(vec_cast,rxEt.tibble)
S3method(vec_cast,tibble.rxEt)
S3method(vec_proxy,rxEt)
S3method(vec_ptype2,data.frame.rxEt)
S3method(vec_ptype2,data.table.rxEt)
S3method(vec_ptype2,rxEt.data.frame)
S3method(vec_ptype2,rxEt.data.table)
S3method(vec_ptype2,rxEt.rxEt)
S3method(vec_restore,rxEt)
export("RxODE<-")
export("ini<-")
export("model<-")
export("rxode2<-")
export("rxode<-")
export(.assertRenameErrorModelLine)
export(.cbindOme)
export(.clearPipe)
export(.collectWarnings)
export(.convertId)
export(.copyUi)
export(.expandPars)
export(.extraC)
export(.extraCnow)
export(.getErrDist)
export(.getLastIdLvl)
export(.getNamedArgumentsToPredDf)
export(.getWh)
export(.handleSingleErrTypeNormOrTFoceiBase)
export(.iniGetAppendArg)
export(.iniHandleFixOrUnfix)
export(.iniHandleLine)
export(.malert)
export(.matchesLangTemplate)
export(.minfo)
export(.modelHandleModelLines)
export(.msuccess)
export(.mwarn)
export(.pipeDfObs)
export(.pipeDfSub)
export(.pipeEvents)
export(.pipeInits)
export(.pipeKeep)
export(.pipeNStud)
export(.pipeNSub)
export(.pipeOmega)
export(.pipeParams)
export(.pipeRx)
export(.pipeSigma)
export(.pipeThetaMat)
export(.quoteCallInfoLines)
export(.rstudioComplete)
export(.rxAdjoint)
export(.rxAdjointExpand)
export(.rxAdjointGrad)
export(.rxAdjointGradBuild)
export(.rxAdjointGradEval)
export(.rxAdjointGradEvalC)
export(.rxAdjointGradPop)
export(.rxAdjointMethodStiff)
export(.rxAdjointModel)
export(.rxAdjointSolve)
export(.rxAdjointSolveBuild)
export(.rxAdjointSolveEvalC)
export(.rxBlankIni)
export(.rxC)
export(.rxDelaySensJump)
export(.rxDelaySensJumpEvents)
export(.rxDelaySensJumpMap)
export(.rxDelaySensJumpMapCached)
export(.rxDiscreteAdjointBuild)
export(.rxDiscreteAdjointGrad)
export(.rxDiscreteForwardSens)
export(.rxDocTable)
export(.rxFromSE)
export(.rxGetHiBoundaryPred1AndIni)
export(.rxGetLambdaFromPred1AndIni)
export(.rxGetLowBoundaryPred1AndIni)
export(.rxGetPredictionDVTransform)
export(.rxGetPredictionF)
export(.rxGetPredictionFTransform)
export(.rxGetSeed)
export(.rxGetVarianceForErrorType)
export(.rxInjectMatExpOdes)
export(.rxIsLinCmt)
export(.rxJacobian)
export(.rxLastCompileSuccess)
export(.rxLinCmt)
export(.rxLinCmtGen)
export(.rxLinNcmt)
export(.rxMemoisedFns)
export(.rxPastBaseLinesFromEnv)
export(.rxPrune)
export(.rxRename)
export(.rxSEeqUsr)
export(.rxSens)
export(.rxSetSeed)
export(.rxToSE)
export(.rxTransInfo)
export(.rxTransform)
export(.rxWithOptions)
export(.rxWithSink)
export(.rxWithSinkBoth)
export(.rxWithWd)
export(.rxode2ptrs)
export(.s3register)
export(.solComp2)
export(.solComp3)
export(.symengineFs)
export(.toClassicEvid)
export(.udfCallFunArg)
export(.udfEnvReset)
export(.udfEnvSet)
export(.udfEnvSetUdf)
export(.udfExists)
export(.udfMd5Info)
export(.uiArg)
export(.useUtf)
export(.vecDf)
export(ELU)
export(GELU)
export(PReLU)
export(ReLU)
export(RxODE)
export(SELU)
export(Swish)
export(add.dosing)
export(add.sampling)
export(aes)
export(as.arrow)
export(as.et)
export(as.ini)
export(as.model)
export(as.rxEvid)
export(as.rxRateDur)
export(as.rxUi)
export(as_tibble.rxEt)
export(assertCompartmentExists)
export(assertCompartmentName)
export(assertCompartmentNew)
export(assertExists)
export(assertIniDf)
export(assertParameterValue)
export(assertRxLinCmt)
export(assertRxUi)
export(assertRxUiEstimatedResiduals)
export(assertRxUiIovNoCor)
export(assertRxUiMixedOnly)
export(assertRxUiMuRefOnly)
export(assertRxUiNoAutoregressive)
export(assertRxUiNoMix)
export(assertRxUiNormal)
export(assertRxUiPopulationOnly)
export(assertRxUiPrediction)
export(assertRxUiRandomOnIdOnly)
export(assertRxUiSingleEndpoint)
export(assertRxUiTransformNormal)
export(assertRxUnbounded)
export(assertVariableExists)
export(assertVariableName)
export(assertVariableNew)
export(binomProbs)
export(bolus)
export(boxCox)
export(boxCoxInv)
export(cvPost)
export(d2ELU)
export(d2ELUa)
export(d2GELU)
export(d2aELU)
export(d3GELU)
export(d4GELU)
export(dELU)
export(dELUa)
export(dGELU)
export(dPReLU)
export(dPReLUa)
export(dPReLUa1)
export(dReLU)
export(dSELU)
export(dSwish)
export(delay)
export(dfWishart)
export(dlReLU)
export(dmexpit)
export(dsoftplus)
export(erf)
export(et)
export(etExpand)
export(etRbind)
export(etRep)
export(etSeq)
export(etTrans)
export(eventTable)
export(evid_)
export(expand_limits)
export(expit)
export(facet_wrap)
export(findLhs)
export(forderForceBase)
export(gammap)
export(gammapDer)
export(gammapInv)
export(gammapInva)
export(gammaq)
export(gammaqInv)
export(gammaqInva)
export(genShinyApp.template)
export(geom_amt)
export(geom_cens)
export(geom_line)
export(getBaseIniSimModel)
export(getBaseSimModel)
export(getBaseSymengineModel)
export(getRxThreads)
export(ggplot)
export(guide_none)
export(indLin)
export(infuse)
export(infuseDur)
export(ini)
export(invWR1d)
export(is.rxEt)
export(is.rxSolve)
export(is.rxStackData)
export(lReLU)
export(label_both)
export(label_context)
export(label_value)
export(label_wrap_gen)
export(linMod)
export(linMod0)
export(linModA0)
export(linModB)
export(linModB0)
export(linModD)
export(linModD0)
export(linModM)
export(linModM0)
export(linToOde)
export(llikBeta)
export(llikBinom)
export(llikCauchy)
export(llikChisq)
export(llikExp)
export(llikF)
export(llikGamma)
export(llikGeom)
export(llikNbinom)
export(llikNbinomMu)
export(llikNorm)
export(llikPois)
export(llikT)
export(llikUnif)
export(llikWeibull)
export(logit)
export(logitNormInfo)
export(lotri)
export(lowergamma)
export(meanProbs)
export(mexpit)
export(mix)
export(mlogit)
export(model)
export(modelExtract)
export(obs)
export(odeMethodToInt)
export(odeToLin)
export(phantom)
export(phi)
export(probit)
export(probitInv)
export(probitNormInfo)
export(rLKJ1)
export(rinvchisq)
export(rxAllowUnload)
export(rxAppendModel)
export(rxAssignControlValue)
export(rxAssignPtr)
export(rxBlockZeros)
export(rxC)
export(rxCat)
export(rxCbindStudyIndividual)
export(rxChain)
export(rxChain2)
export(rxClean)
export(rxCombineErrorLines)
export(rxCompile)
export(rxCondition)
export(rxControl)
export(rxControlUpdateSens)
export(rxCores)
export(rxCreateCache)
export(rxD)
export(rxDelete)
export(rxDemoteAddErr)
export(rxDerived)
export(rxDeserialize)
export(rxDfdy)
export(rxDll)
export(rxDllLoaded)
export(rxDynLoad)
export(rxDynUnload)
export(rxErrTypeCombine)
export(rxEtDispatchSolve)
export(rxEventSensDeactivate)
export(rxEventSensLoadModel)
export(rxEventTableFile)
export(rxEvid)
export(rxExpandFEta_)
export(rxExpandGrid)
export(rxExpandGrid_)
export(rxExpandIfElse)
export(rxExpandSens2_)
export(rxExpandSens3_)
export(rxExpandSens_)
export(rxFixPop)
export(rxFixRes)
export(rxForget)
export(rxFromSE)
export(rxFun)
export(rxFunParse)
export(rxGetControl)
export(rxGetDefaultSerialize)
export(rxGetDistributionSimulationLines)
export(rxGetLin)
export(rxGetModel)
export(rxGetSeed)
export(rxGetSerialType_)
export(rxGetrxode2)
export(rxHasAr)
export(rxHtml)
export(rxIndLinState)
export(rxIndLinStrategy)
export(rxInit)
export(rxInits)
export(rxIntToBase)
export(rxIntToLetter)
export(rxInv)
export(rxIs)
export(rxIsAutoSwitch)
export(rxIsCurrent)
export(rxIsDense)
export(rxIsImplicit)
export(rxIsLoaded)
export(rxIsNonStiff)
export(rxIsStiff)
export(rxLastCompile)
export(rxLhs)
export(rxLoad)
export(rxLoadState)
export(rxLock)
export(rxMd5)
export(rxMemSummary)
export(rxMemoryEstimate)
export(rxModelVars)
export(rxModelVarsS3)
export(rxModels_)
export(rxNorm)
export(rxNumLoaded)
export(rxOdeToIndLin)
export(rxOldQsDes)
export(rxOmegaVarCovDeriv)
export(rxOmegaVarCovDeriv_)
export(rxOptExpr)
export(rxParam)
export(rxParams)
export(rxParseErr)
export(rxParsePk)
export(rxParsePred)
export(rxParseSetSilentErr)
export(rxParseSuppressMsg)
export(rxPkg)
export(rxPp)
export(rxPreferredDistributionName)
export(rxProgress)
export(rxProgressAbort)
export(rxProgressStop)
export(rxPrune)
export(rxRateDur)
export(rxRawToC)
export(rxReload)
export(rxRemoveControl)
export(rxRename)
export(rxRepR0_)
export(rxReq)
export(rxRmFun)
export(rxRmFunParse)
export(rxRmvn)
export(rxS)
export(rxSaveState)
export(rxSeedEng)
export(rxSensMatExp)
export(rxSerialize)
export(rxSetControl)
export(rxSetCovariateNamesForPiping)
export(rxSetIni0)
export(rxSetPipingAuto)
export(rxSetProd)
export(rxSetProgressBar)
export(rxSetSeed)
export(rxSetSilentErr)
export(rxSetSum)
export(rxSetupIni)
export(rxSetupScale)
export(rxShiny)
export(rxSimThetaOmega)
export(rxSolve)
export(rxSolveAdjoint)
export(rxSolveAdjointRk4)
export(rxSolveChunked)
export(rxSolveFree)
export(rxSolveSetup)
export(rxSplitPlusQ)
export(rxStack)
export(rxState)
export(rxStateOde)
export(rxSumProdModel)
export(rxSupportedFuns)
export(rxSuppressMsg)
export(rxSymInvChol)
export(rxSymInvCholCreate)
export(rxSymInvCholN)
export(rxSyncOptions)
export(rxTempDir)
export(rxTest)
export(rxTheme)
export(rxTick)
export(rxToIndLin)
export(rxToSE)
export(rxTrans)
export(rxUdfUi)
export(rxUdfUiControl)
export(rxUdfUiData)
export(rxUdfUiEst)
export(rxUdfUiExpr)
export(rxUdfUiFlag)
export(rxUdfUiIniDf)
export(rxUdfUiIniLhs)
export(rxUdfUiIsValue)
export(rxUdfUiMv)
export(rxUdfUiNum)
export(rxUdfUiParsing)
export(rxUdfUiReset)
export(rxUiCompress)
export(rxUiDecompress)
export(rxUiDeparse)
export(rxUiDevelop)
export(rxUiGet)
export(rxUnload)
export(rxUnloadAll)
export(rxUnlock)
export(rxUse)
export(rxValidate)
export(rxVersion)
export(rxWithPreserveSeed)
export(rxWithSeed)
export(rxbeta)
export(rxbinom)
export(rxcauchy)
export(rxchisq)
export(rxexp)
export(rxf)
export(rxgamma)
export(rxgeom)
export(rxnbinom)
export(rxnbinomMu)
export(rxnorm)
export(rxnormV)
export(rxode)
export(rxode2)
export(rxode2parse)
export(rxode2parseAssignPackagesToLoad)
export(rxode2parseAssignPointerTranslation)
export(rxode2parseAssignTranslation)
export(rxode2parseD)
export(rxode2parseGetPackagesToLoad)
export(rxode2parseGetPointerAssignment)
export(rxode2parseGetTranslation)
export(rxord)
export(rxpois)
export(rxt)
export(rxunif)
export(rxweibull)
export(scale_type)
export(scale_x_continuous)
export(scale_x_date)
export(scale_x_discrete)
export(scale_y_continuous)
export(scale_y_date)
export(scale_y_discrete)
export(setCvodeLinearSolver)
export(setRxThreads)
export(softplus)
export(splitBolus)
export(stat_amt)
export(stat_cens)
export(swapMatListWithCube)
export(testCompartmentExists)
export(testExists)
export(testIniDf)
export(testRxLinCmt)
export(testRxUnbounded)
export(testVariableExists)
export(toTrialDuration)
export(uppergamma)
export(waiver)
export(warnRxBounded)
export(write.template.server)
export(write.template.ui)
export(xlab)
export(yeoJohnson)
export(yeoJohnsonInv)
export(ylab)
export(zeroRe)
import(data.table)
import(tools)
importFrom(PreciseSums,fsum)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(RcppParallel,RcppParallelLibs)
importFrom(checkmate,qassert)
importFrom(dparser,dparse)
importFrom(ggplot2,"%+replace%")
importFrom(ggplot2,aes)
importFrom(ggplot2,expand_limits)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,label_both)
importFrom(ggplot2,label_context)
importFrom(ggplot2,label_value)
importFrom(ggplot2,label_wrap_gen)
importFrom(ggplot2,scale_type)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_x_date)
importFrom(ggplot2,scale_x_discrete)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,scale_y_date)
importFrom(ggplot2,scale_y_discrete)
importFrom(ggplot2,waiver)
importFrom(ggplot2,xlab)
importFrom(ggplot2,ylab)
importFrom(lotri,lotri)
importFrom(memoise,is.memoised)
importFrom(memoise,memoise)
importFrom(methods,is)
importFrom(methods,signature)
importFrom(stats,ave)
importFrom(stats,coef)
importFrom(stats,dnorm)
importFrom(stats,end)
importFrom(stats,integrate)
importFrom(stats,setNames)
importFrom(stats,simulate)
importFrom(stats,start)
importFrom(stats,update)
importFrom(sys,exec_internal)
importFrom(utils,.DollarNames)
importFrom(utils,assignInMyNamespace)
importFrom(utils,capture.output)
importFrom(utils,compareVersion)
importFrom(utils,download.file)
importFrom(utils,getFromNamespace)
importFrom(utils,head)
importFrom(utils,packageVersion)
importFrom(utils,removeSource)
importFrom(utils,sessionInfo)
importFrom(utils,str)
importFrom(utils,write.csv)
importFrom(vctrs,vec_cast)
importFrom(vctrs,vec_proxy)
importFrom(vctrs,vec_ptype2)
importFrom(vctrs,vec_restore)
useDynLib(rxode2, .registration=TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# rxode2

<!-- badges: start -->

![Cran updating
status](https://img.shields.io/badge/CRAN-Not%20Updating-green)
[![R-CMD-check](https://github.com/nlmixr2/rxode2/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/nlmixr2/rxode2/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/nlmixr2/rxode2/graph/badge.svg)](https://app.codecov.io/gh/nlmixr2/rxode2)
[![CRAN
version](http://www.r-pkg.org/badges/version/rxode2)](https://cran.r-project.org/package=rxode2)
[![CRAN total
downloads](https://cranlogs.r-pkg.org/badges/grand-total/rxode2)](https://cran.r-project.org/package=rxode2)
[![CRAN total
downloads](https://cranlogs.r-pkg.org/badges/rxode2)](https://cran.r-project.org/package=rxode2)
[![CodeFactor](https://www.codefactor.io/repository/github/nlmixr2/rxode2/badge)](https://www.codefactor.io/repository/github/nlmixr2/rxode2)
![r-universe](https://nlmixr2.r-universe.dev/badges/rxode2)
<!-- badges: end -->

## Overview

**rxode2** is an R package for solving and simulating from ode-based
models. These models are convert the rxode2 mini-language to C and
create a compiled dll for fast solving. ODE solving using rxode2 has a
few key parts:

- `rxode2()` which creates the C code for fast ODE solving based on a
  [simple
  syntax](https://nlmixr2.github.io/rxode2/articles/rxode2-syntax.html)
  related to Leibnitz notation.
- The event data, which can be:
  - a `NONMEM` or `deSolve` [compatible data
    frame](https://nlmixr2.github.io/rxode2/articles/rxode2-event-types.html),
    or
  - created with `et()` or `eventTable()` for [easy simulation of
    events](https://nlmixr2.github.io/rxode2/articles/rxode2-event-table.html)
  - The data frame can be augmented by adding
    [time-varying](https://nlmixr2.github.io/rxode2/articles/rxode2-covariates.html#time-varying-covariates)
    or adding [individual
    covariates](https://nlmixr2.github.io/rxode2/articles/rxode2-covariates.html#individual-covariates)
    (`iCov=` as needed)
- `rxSolve()` which solves the system of equations using initial
  conditions and parameters to make predictions
  - With multiple subject data, [this may be
    parallelized](https://nlmixr2.github.io/rxode2/articles/rxode2-speed.html).
  - With single subject the [output data frame is
    adaptive](https://nlmixr2.github.io/rxode2/articles/rxode2-data-frame.html)
  - Covariances and other metrics of uncertanty can be used to [simulate
    while
    solving](https://nlmixr2.github.io/rxode2/articles/rxode2-sim-var.html)

## Installation

You can install the released version of rxode2 from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("rxode2")
```

The fastest way to install the development version of `rxode2` is to use
the `r-universe` service. This service compiles binaries of the
development version for MacOS and for Windows so you don’t have to wait
for package compilation:

``` r
install.packages(c("dparser", "rxode2ll", "rxode2"),
                 repos=c(nlmixr2="https://nlmixr2.r-universe.dev",
                         CRAN="https://cloud.r-project.org"))
```

If this doesn’t work you install the development version of rxode2 with

``` r
devtools::install_github("nlmixr2/rxode2ll")
devtools::install_github("nlmixr2/rxode2")
```

To build models with rxode2, you need a working c compiler. To use
parallel threaded solving in rxode2, this c compiler needs to support
open-mp.

You can check to see if R has working c compiler you can check with:

``` r
## install.packages("pkgbuild")
pkgbuild::has_build_tools(debug = TRUE)
```

If you do not have the toolchain, you can set it up as described by the
platform information below:

### Windows

In windows you may simply use installr to install rtools:

``` r
install.packages("installr")
library(installr)
install.rtools()
```

Alternatively you can
[download](https://cran.r-project.org/bin/windows/Rtools/) and install
rtools directly.

### Mac OSX

To get the most speed you need OpenMP enabled and compile rxode2 with
that compiler. There are various options and the most up to date
discussion about this is likely the [data.table installation FAQ for
MacOS](https://github.com/Rdatatable/data.table/wiki/Installation#openmp-enabled-compiler-for-mac).
The last thing to keep in mind is that `rxode2` uses the code very
similar to the original `lsoda` which requires the `gfortran` compiler
to be setup as well as the `OpenMP` compilers.

If you are going to be using `rxode2` and `nlmixr` together and have an
older mac computer, I would suggest trying the following:

``` r
library(symengine)
```

If this crashes your R session then the binary does not work with your
Mac machine. To be able to run nlmixr, you will need to compile this
package manually. I will proceed assuming you have `homebrew` installed
on your system.

On your system terminal you will need to install the dependencies to
compile `symengine`:

``` sh
brew install cmake gmp mpfr libmpc
```

After installing the dependencies, you need to re-install `symengine`:

``` r
install.packages("symengine", type="source")
library(symengine)
```

### Linux

To install on linux make sure you install `gcc` (with openmp support)
and `gfortran` using your distribution’s package manager.

You will also have to install system dependencies like `udunits` and the
[symengine
dependencies](https://github.com/symengine/symengine.R?tab=readme-ov-file#installation)
for the complete installation to work in linux. You could also have this
done by system packages in your package manager if you add the
appropriate repositories. This is the point of the
[r2u](https://github.com/eddelbuettel/r2u) project.

### R versions 4.0 and 4.1

For installation on R versions 4.0.x and 4.1.x, please see the
instructions on how to install `symengine` in the `nlmixr2` installation
instructions:
<https://github.com/nlmixr2/nlmixr2#r-package-installation>

## Development version

Since the development version of rxode2 uses StanHeaders, you will need
to make sure your compiler is setup to support C++14, as described in
the [rstan setup
page](https://github.com/stan-dev/rstan/wiki/RStan-Getting-Started#configuration-of-the-c-toolchain).
For R 4.0, I do not believe this requires modifying the windows
toolchain any longer (so it is much easier to setup).

Once the C++ toolchain is setup appropriately, you can install the
development version from [GitHub](https://github.com/nlmixr2/rxode2)
with:

``` r
# install.packages("devtools")
devtools::install_github("nlmixr2/rxode2ll")
devtools::install_github("nlmixr2/rxode2")
```

# Illustrated Example

The model equations can be specified through a text string, a model file
or an R expression. Both differential and algebraic equations are
permitted. Differential equations are specified by `d/dt(var_name) =`.
Each equation can be separated by a semicolon.

To load `rxode2` package and compile the model:

``` r
library(rxode2)
#> rxode2 3.0.4.9000 using 8 threads (see ?getRxThreads)

mod1 <- function() {
  ini({
    # central
    KA=2.94E-01
    CL=1.86E+01
    V2=4.02E+01
    # peripheral
    Q=1.05E+01
    V3=2.97E+02
    # effects
    Kin=1
    Kout=1
    EC50=200
  })
  model({
    C2 <- centr/V2
    C3 <- peri/V3
    d/dt(depot) <- -KA*depot
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
    d/dt(peri)  <- Q*C2 - Q*C3
    eff(0) <- 1
    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
  })
}
```

Model parameters may be specified in the `ini({})` model block, initial
conditions can be specified within the model with the `cmt(0)= X`, like
in this model `eff(0) <- 1`.

You may also specify between subject variability initial conditions and
residual error components just like nlmixr2. This allows a single
interface for `nlmixr2`/`rxode2` models. Also note, the classic `rxode2`
interface still works just like it did in the past (so don’t worry about
breaking code at this time).

In fact, you can get the classic `rxode2` model `$simulationModel` in
the ui object:

``` r
mod1 <- mod1() # create the ui object (can also use `rxode2(mod1)`)
mod1

summary(mod1$simulationModel)
```

## Specify Dosing and sampling in rxode2

`rxode2` provides a simple and very flexible way to specify dosing and
sampling through functions that generate an event table. First, an empty
event table is generated through the “et()” function. This has an
interface that is similar to NONMEM event tables:

``` r
ev  <- et(amountUnits="mg", timeUnits="hours") |>
  et(amt=10000, addl=9,ii=12,cmt="depot") |>
  et(time=120, amt=2000, addl=4, ii=14, cmt="depot") |>
  et(0:240) # Add sampling
```

You can see from the above code, you can dose to the compartment named
in the rxode2 model. This slight deviation from NONMEM can reduce the
need for compartment renumbering.

These events can also be combined and expanded (to multi-subject events
and complex regimens) with `rbind`, `c`, `seq`, and `rep`. For more
information about creating complex dosing regimens using rxode2 see the
[rxode2 events
vignette](https://nlmixr2.github.io/rxode2/articles/rxode2-event-types.html).

## Solving ODEs

The ODE can now be solved using `rxSolve`:

``` r
x <- mod1 |> rxSolve(ev)
#> ℹ parameter labels from comments are typically ignored in non-interactive mode
#> ℹ Need to run with the source intact to parse comments
#> → creating rxode2 include directory
#> → getting R compile options
#> → precompiling headers
#> ✔ done
#> using C compiler: 'gcc.exe (GCC) 14.2.0'
x
#> ── Solved rxode2 object ──
#> ── Parameters (x$params): ──
#>      KA      CL      V2       Q      V3     Kin    Kout    EC50
#>   0.294  18.600  40.200  10.500 297.000   1.000   1.000 200.000
#> ── Initial Conditions (x$inits): ──
#> depot centr  peri   eff
#>     0     0     0     1
#> ── First part of data (object): ──
#> # A tibble: 241 × 7
#>   time    C2    C3  depot centr  peri   eff
#>    [h] <dbl> <dbl>  <dbl> <dbl> <dbl> <dbl>
#> 1    0   0   0     10000     0     0   1
#> 2    1  44.4 0.920  7453. 1784.  273.  1.08
#> 3    2  54.9 2.67   5554. 2206.  794.  1.18
#> 4    3  51.9 4.46   4140. 2087. 1324.  1.23
#> 5    4  44.5 5.98   3085. 1789. 1776.  1.23
#> 6    5  36.5 7.18   2299. 1467. 2132.  1.21
#> # ℹ 235 more rows
```

This returns a modified data frame. You can see the compartment values
in the plot below:

``` r
library(ggplot2)
plot(x,C2) + ylab("Central Concentration")
```

<img src="man/figures/README-intro-central-1.png" width="100%" />

Or,

``` r
plot(x,eff)  + ylab("Effect")
```

<img src="man/figures/README-intro-effect-1.png" width="100%" />

Note that the labels are automatically labeled with the units from the
initial event table. rxode2 extracts `units` to label the plot (if they
are present).

# Related R Packages

## ODE solving

This is a brief comparison of pharmacometric ODE solving R packages to
`rxode2`.

There are several [R packages for differential
equations](https://CRAN.R-project.org/view=DifferentialEquations). The
most popular is [deSolve](https://CRAN.R-project.org/package=deSolve).

However for pharmacometrics-specific ODE solving, there are only 2
packages other than [rxode2](https://cran.r-project.org/package=rxode2)
released on CRAN. Each uses compiled code to have faster ODE solving.

- [mrgsolve](https://cran.r-project.org/package=mrgsolve), which uses
  C++ lsoda solver to solve ODE systems. The user is required to write
  hybrid R/C++ code to create a mrgsolve model which is translated to
  C++ for solving.

  In contrast, `rxode2` has a R-like mini-language that is parsed into C
  code that solves the ODE system.

  Unlike `rxode2`, `mrgsolve` does not currently support symbolic
  manipulation of ODE systems, like automatic Jacobian calculation or
  forward sensitivity calculation (`rxode2` currently supports this and
  this is the basis of
  [nlmixr2](https://CRAN.R-project.org/package=nlmixr2)’s FOCEi
  algorithm)

- [dMod](https://CRAN.R-project.org/package=dMod), which uses a unique
  syntax to create “reactions”. These reactions create the underlying
  ODEs and then created c code for a compiled deSolve model.

  In contrast `rxode2` defines ODE systems at a lower level. `rxode2`’s
  parsing of the mini-language comes from C, whereas `dMod`’s parsing
  comes from R.

  Like `rxode2`, `dMod` supports symbolic manipulation of ODE systems
  and calculates forward sensitivities and adjoint sensitivities of
  systems.

  Unlike `rxode2`, `dMod` is not thread-safe since `deSolve` is not yet
  thread-safe.

- [PKPDsim](https://github.com/InsightRX/PKPDsim) which defines models
  in an R-like syntax and converts the system to compiled code.

  Like `mrgsolve`, `PKPDsim` does not currently support symbolic
  manipulation of ODE systems.

  `PKPDsim` is not thread-safe.

The open pharmacometrics open source community is fairly friendly, and
the rxode2 maintainers has had positive interactions with all of the
ODE-solving pharmacometric projects listed.

## PK Solved systems

`rxode2` supports 1-3 compartment models with gradients (using stan
math’s auto-differentiation). This currently uses the same equations as
`PKADVAN` to allow time-varying covariates.

`rxode2` can mix ODEs and solved systems.

### The following packages for solved PK systems are on CRAN

- [mrgsolve](https://cran.r-project.org/package=mrgsolve) currently has
  1-2 compartment (poly-exponential models) models built-in. The solved
  systems and ODEs cannot currently be mixed.

- [pmxTools](https://github.com/kestrel99/pmxTools) currently have 1-3
  compartment (super-positioning) models built-in. This is a R-only
  implementation.

- [PKPDsim](https://github.com/InsightRX/PKPDsim) uses 1-3 “ADVAN”
  solutions using non-superpositioning.

- [PKPDmodels](https://CRAN.R-project.org/package=PKPDmodels) has a
  one-compartment model with gradients.

### Non-CRAN libraries:

- [PKADVAN](https://github.com/abuhelwa/PKADVAN_Rpackage) Provides 1-3
  compartment models using non-superpositioning. This allows
  time-varying covariates.

## This is only for rxode2
for (f in c("inst/include/rxode2_RcppExports.h", "src/RcppExports.cpp")) {
  l <- readLines(f)
  w <- which(regexpr("^[#]include <RcppArmadillo.h>", l) != -1)
  if (length(w) > 0) {
    l <- l[-w]
  }
  w <- which(regexpr("^[#]define R_STRICT_HEADERS", l) != -1)
  if (length(w) > 0) {
    l <- l[-w]
  }
  l <- c("#define R_STRICT_HEADERS",
         "#include <RcppArmadillo.h>",
         l)
  file.out <- file(f, "wb")
  writeLines(l, file.out)
  close(file.out)
}

l <- readLines("R/RcppExports.R")
w <- which(regexpr("# Register entry points", l, fixed=TRUE) != -1)
if (length(w) >= 1) {
  w <- w[1]
  l <- l[seq(1, w-1)]
  RcppExports.R <- file("R/RcppExports.R", "wb")
  writeLines(l, RcppExports.R)
  close(RcppExports.R)
}

compilerPath <- tools::Rcmd("config CC", stdout=TRUE)

# To distinguish between them, check the version output
versionInfo <- try(system(paste(compilerPath, "--version"), intern = TRUE))
if (inherits(versionInfo, "try-error")) {
  .o2 <- "-O2 "
} else if (any(grepl("clang", versionInfo, ignore.case = TRUE))) {
  .o2 <- "-O3 -fno-math-errno -mtune=native "
} else if (any(grepl("gcc", versionInfo, ignore.case = TRUE))) {
  .o2 <- "-O3 -fno-math-errno -mtune=native "
} else {
  .o2 <- "-O2 "
}

.in <- suppressWarnings(readLines("src/Makevars.in"))
.in <- gsub("@ARMA@", file.path(find.package("RcppArmadillo"),"include"), .in)
.in <- gsub("@O2@", .o2, .in)
.in <- gsub("@BH@", file.path(find.package("BH"),"include"), .in)
.in <- gsub("@RCPP@", file.path(find.package("Rcpp"),"include"), .in)
.in <- gsub("@EG@", file.path(find.package("RcppEigen"),"include"), .in)


.in <- gsub("@SL@", paste(capture.output(StanHeaders:::LdFlags()), capture.output(RcppParallel:::RcppParallelLibs())), #nolint
            .in)


.badStan <- ""
.in <- gsub("@SH@", gsub("-I", "-@ISYSTEM@",
                         paste(capture.output(StanHeaders:::CxxFlags()), # nolint
                               capture.output(RcppParallel:::CxxFlags()), # nolint
                               paste0("-@ISYSTEM@'", system.file('include', 'src', package = 'StanHeaders', mustWork = TRUE), "'"),
                               .badStan)),
            .in)



if (.Platform$OS.type == "windows") {
  .makevars <- file("src/Makevars.win", "wb")
  .i <- "I"
} else {
  .makevars <- file("src/Makevars", "wb")
  if (any(grepl("Pop!_OS", utils::osVersion, fixed=TRUE)) ||
          any(grepl("Ubuntu", utils::osVersion, fixed=TRUE))) {
    .i <- "isystem"
  } else {
    .i <- "I"
  }
}

writeLines(gsub("@ISYSTEM@", .i, .in),
             .makevars)
close(.makevars)

if (file.exists("man/reexports.Rd")) {
  l <- readLines("man/reexports.Rd")
  if (!any(regexpr("[\\]value", l) != -1)) {
    l <- c(l, "\\value{ Inherited from parent routine }")
    file.out <- file("man/reexports.Rd", "wb")
    writeLines(l, file.out)
    close(file.out)
  }
}


unlink("R/rxode2_md5.R")

cpp <- list.files("src", pattern = ".(c|h|cpp|f)$")
include <- list.files("inst/include")
#Rfiles <- list.files("R/", pattern = ".R")

cmd <- file.path(R.home("bin"), "R")
args <- c("CMD", "config")

md5 <- digest::digest(c(lapply(c(paste0("src/", cpp),
                                 paste0("inst/include/", include)#,
                                 #paste0("R/", Rfiles)
                                 ), digest::digest, file = TRUE),
                        ## vapply(c("BLAS_LIBS", "CC",  "CFLAGS", "CPICFLAGS",
                        ##          "CXX", "CXXFLAGS", "CXXPICFLAGS",
                        ##          "CXX11", "CXX11STD", "CXX11FLAGS", "CXX11PICFLAGS",
                        ##          "CXX14", "CXX14STD", "CXX14FLAGS", "CXX14PICFLAGS",
                        ##          "CXX17", "CXX17STD", "CXX17FLAGS", "CXX17PICFLAGS",
                        ##          "CXX20", "CXX20STD", "CXX20FLAGS", "CXX20PICFLAGS",
                        ##          "FC", "FFLAGS", "FCFLAGS",  "FPICFLAGS"),
                        ##        function(cfg) {
                        ##          rawToChar(sys::exec_internal(cmd, c(args, cfg))$stdout)
                        ##        }, character(1)
                        ##       ),
                        ""
                        ))
unlink("R/rxode2_md5.R")
md5file <- file("R/rxode2_md5.R", "wb")
writeLines(sprintf("rxode2.md5 <- \"%s\"\n", md5), md5file)
close(md5file)

l <- readLines("DESCRIPTION")
w <- which(regexpr("Version[:] *(.*)$", l) != -1)
v <- gsub("Version[:] *(.*)$", "\\1", l[w])

unlink("inst/include/rxode2parseVer.h")
ode.h <- file("inst/include/rxode2parseVer.h", "wb")
writeLines(c(sprintf("#define __VER_md5__ \"%s\"", md5),
             "#define __VER_repo__ \"https://github.com/nlmixr2/rxode2\"",
             sprintf("#define __VER_ver__ \"%s\"", v)),
           ode.h)
close(ode.h)

unlink("src/sbuf.c")
l <- readLines("inst/include/sbuf.c")
sbuf.c <- file("src/sbuf.c", "wb")
writeLines(l, sbuf.c)
close(sbuf.c)


unlink("src/codegen2.h")
l <- readLines("inst/include/rxode2_model_shared.c")

l <- l[l != ""]
l <- gsub(" *= *NULL;", "=NULL;", l)

def <- l
w <- which(regexpr("double _prod", def) != -1) - 1
def <- def[1:w]
def <- gsub("=NULL", "", def)
def <- gsub("[^ ]* *[*]?([^;]*);", "\\1", def)

def <- unique(c(def, c("_sum", "_udf", "_sign", "_prod", "_max", "_min", "_transit4P", "_transit3P", "_assignFuns0", "_assignFuns", "_getRxSolve_", "_solveData", "_rxord", "__assignFuns2")))

w0 <- which(grepl("double +_prod", l))[1]
r <- 1:(w0 - 1)
l0 <- l[r]
l <- l[-r]


w1 <- which(regexpr("dynamic start", l) != -1)
l1 <- l[1:w1]



w2 <- which(regexpr("dynamic stop", l) != -1)
l2 <- l[seq(w2, length(l))]

w3 <- which(regexpr("assign start", l2) != -1)

l3 <- l2[seq(w3, length(l2))]
l2 <- l2[1:w3]

w4 <- which(regexpr("assign stop", l3) != -1)
l3 <- l3[seq(w4, length(l3))]

dfP <- l[seq(w1+1, w2-1)]

dfP <- dfP[regexpr("^ *$", dfP)==-1]
df <- setNames(do.call("rbind",lapply(seq_along(dfP),
                                      function(i) {
                                        .r <- sub("^ *", "", dfP[[i]])
                                        .r <- sub("^([^ ]*) *= *[(]", "\\1,", .r)
                                        .r <- sub("^([^ ]*) *[)] *R_GetCCallable *[(] *\"", "\\1,", .r, perl=TRUE)
                                        .r <- sub("^([^ ]*) *\" *, *\"", "\\1,", .r, perl=TRUE)
                                        .r <- sub("^([^ ]*)\" *[)] *; *", "\\1",.r, perl=TRUE)
                                        data.frame(t(strsplit(.r, ",")[[1]]),stringsAsFactors = FALSE)
                                      })), c("fun", "type", "package", "packageFun"))

df$rxFun <- df$fun
df$argMax <- df$argMin <- NA_integer_
df$threadSafe <- 1L
df <- df[,c("rxFun", "fun", "type", "package", "packageFun", "argMin", "argMax", "threadSafe")]
df$rxFun <- gsub("_llik", "llik", df$rxFun)

def <- def[!(def %in% df$rxFun)]
def <- def[!(def %in% df$fun)]

.parseEnv <- new.env(parent=emptyenv())
source("R/parseFuns.R")

df$argMin <- vapply(df$rxFun, function(f) {
  .n <- .parseEnv$.parseNum[f]
  if (is.na(.n)) return(NA_integer_)
  .n <-setNames(.n, NULL)
  as.integer(.n)
}, integer(1), USE.NAMES=TRUE)

df$argMax <- df$argMin

dfStr <- deparse(df)
dfStr[1] <- paste(".parseEnv$.rxode2parseDf <- ", dfStr[1])

dfIni.R <- file("R/dfIni.R", "wb")
writeLines(dfStr,
           dfIni.R)
close(dfIni.R)

## deparse1 came from R 4.0, use deparse2
deparse2 <- function(expr, collapse = " ", width.cutoff = 500L, ...) {
  paste(deparse(expr, width.cutoff, ...), collapse = collapse)
}

final <- c("#include <time.h>",
           "#include <stdlib.h>",
           "unsigned long int __timeId=0;",
           "char *genRandomChar(void);",
           "void writeHeader(const char *md5, const char *extra) {",
           paste0("sAppend(&sbOut, \"#define ", def, " _rx%s%s%ld_", def, "_%s\\n\", extra, md5, __timeId++, genRandomChar());"),
           "}",
           "void writeBody0(void) {",
           paste0("sAppendN(&sbOut, ", vapply(paste0(l0, "\n"), deparse2, character(1)), ", ", nchar(l0) + 1, ");"),
           "}",
           "void writeBody1(void) {",
           paste0("sAppendN(&sbOut, ", vapply(paste0(l1, "\n"), deparse2, character(1)), ", ", nchar(l1) + 1, ");"),
           "}",
           "void writeBody2(void) {",
           paste0("sAppendN(&sbOut, ", vapply(paste0(l2, "\n"), deparse2, character(1)), ", ", nchar(l2) + 1, ");"),
           "}",
           "void writeBody3(void) {",
           paste0("sAppendN(&sbOut, ", vapply(paste0(l3, "\n"), deparse2, character(1)), ", ", nchar(l3) + 1, ");"),
           "}",
           "void writeFooter(void) {",
           paste0("sAppendN(&sbOut, \"#undef ", def, "\\n\", ", nchar(def) + 8, ");"),
           "}"
           )

codegen2.h <- file("src/codegen2.h", "wb")
writeLines(final,
           codegen2.h)
close(codegen2.h)

ode <- "
   C2 = centr/V2;
   C3 = peri/V3;
   d/dt(depot) =-KA*depot;
   d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
   d/dt(peri)  =                    Q*C2 - Q*C3;
   d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
"

m1 <- RxODE(model = ode)

# create dosing and observation (sampling) events
# QD (once daily) dosing, 5 days

qd <- eventTable(amount.units = "ug", time.units = "hours")
qd$add.dosing(dose = 10000, nbr.doses = 5, dosing.interval = 24)

# hourly sampling during 1st day, every 4 hours afterwards
qd$add.sampling(0:24) # hourly
qd$add.sampling(seq(from = 28, to = 96, by = 4)) # every 4 hours

# BID dosing, 5 days

bid <- eventTable(amount.units = "ug", time.units = "days") # only dosing
bid$add.dosing(
  dose = 10000,
  nbr.doses = 2 * 5,
  dosing.interval = 12
)

bid$add.sampling(0:24)
bid$add.sampling(96 + 0:24)

# init values
theta <-
  c(
    KA = 2.94E-01, CL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
    Kin = 1, Kout = 1, EC50 = 200
  )

qd.cp <- m1$solve(theta, qd, inits = c(0, 0, 0, 1))
bid.cp <- m1$solve(theta, bid, inits = c(0, 0, 0, 1))

cp.plot <-
  function(cp, xlab = "Time (days)", ...) {
    xtime <- cp[, "time"]
    matplot(xtime, cp[, c("depot", "centr", "peri")],
      type = "l", ...,
      xlab = xlab, ylab = "Drug amount (ug)"
    )

    legend("topright",
      legend = c("Depot", "Central", "Peripheral"),
      title = "Compartment",
      col = 1:3, lty = 1:3, bty = "n"
    )

    plot(xtime, cp[, "eff"],
      type = "l", ...,
      xlab = xlab, ylab = "Effect compartment"
    )
  }

cp.plot(qd.cp, main = "QD dosing, 5 days")

cp.plot(bid.cp, main = "BID dosing, 5 days")
rxClean() # Remove dlls and extra files created by RxODE

#' Combine Model Lines
#'
#' @param model1 rxUi model1
#' @param model2 rxUi model2
#' @param ini Final ini for model
#' @return new model with both models appended together and ini from input
#' @noRd
#' @author Matthew L. Fidler
.combineModelLines <- function(model1, model2, ini) {
  # Add the meta information from model2 into the meta information of new model
  .ls <- ls(model2$meta, all.names=TRUE)
  for (.i in seq_along(.ls)) {
    if (.i != ".simModelBase") {
      assign(.ls[.i], model2$meta[[.ls[.i]]], envir=model1$meta)
    }
  }
  model1$iniDf <- ini
  model1$lstExpr <- c(model1$lstExpr, model2$lstExpr)
  model1$fun()
}
#' Append 2 models
#'
#'
#' @param model1 rxUi type of model
#' @param model2 rxUi type of model
#' @param common boolean; when `TRUE` warn if the models have no variables in common
#' @return rxUi combined model of model1 and model2
#' @noRd
#' @author Matthew L. Fidler
rxAppendModel_ <- function(model1, model2, common=TRUE) {
  model1 <- assertRxUi(model1)
  model1 <- .copyUi(model1) # so modifications do not affect first model
  model2 <- assertRxUi(model2)
  model2 <- .copyUi(model2)
  .ini1 <- model1$iniDf
  .ini2 <- model2$iniDf
  .bind <- intersect(c(model1$mv0$lhs, model1$mv0$state), model2$allCovs)
  if (common && length(.bind) == 0) {
    warning("the appended models have no variables in common (use `common=FALSE` to suppress this warning)",
            call.=FALSE)
  }
  if (is.null(.ini1) && is.null(.ini2)) {
    return(.combineModelLines(model1, model2, NULL))
  }
  if (!is.null(.ini1) && is.null(.ini2)) {
    return(.combineModelLines(model1, model2, .ini1))
  }
  if (is.null(.ini1) && !is.null(.ini2)) {
    return(.combineModelLines(model1, model2, .ini2))
  }
  # both exist
  .ini1theta <- .ini1[!is.na(.ini1$ntheta),, drop = FALSE]
  .ini2theta <- .ini2[!is.na(.ini2$ntheta),, drop = FALSE]
  .both <- intersect(.ini1theta$name, .ini2theta$name)
  if (length(.both) > 0) {
    .minfo("duplicated population parameters when combining 2 models")
    .minfo(paste0("keeping initialization from first model: '",
                  paste(.both, collapse="', '"), "'"))
    .ini2theta <- .ini2theta[!(.ini2theta$name %in% .both),, drop =FALSE]
  }
  .ini2theta$ntheta <- length(.ini1theta$ntheta) + seq_along(.ini2theta$ntheta)
  .iniT <- rbind(.ini1theta, .ini2theta)
  # now look at the etas
  .ini1eta <- .ini1[is.na(.ini1$ntheta),, drop = FALSE]
  .ini2eta <- .ini2[is.na(.ini2$ntheta),, drop = FALSE]
  .both <- intersect(.ini1eta$name, .ini2eta$name)
  if (length(.both) > 0) {
    # See if any of the items have covariances defined
    .complex1 <- which(vapply(.both, function(v) {
      .eta <- .ini1eta[.ini1eta$name == v, "neta1"]
      any((.ini1eta$neta1 == .eta & .ini1eta$neta2 != .eta) |
            (.ini1eta$neta2 == .eta & .ini1eta$neta1 != .eta))
    }, logical(1), USE.NAMES = FALSE))
    .complex2 <- which(vapply(.both, function(v) {
      .eta <- .ini2eta[.ini2eta$name == v, "neta1"]
      any((.ini2eta$neta1 == .eta & .ini2eta$neta2 != .eta) |
            (.ini2eta$neta2 == .eta & .ini2eta$neta1 != .eta) )
    }, logical(1), USE.NAMES = FALSE))
    .err <- unique(c(.both[.complex1], .both[.complex2]))
    if (length(.err) > 0) {
      stop("duplicated parameter has covariance, will not append models: '",
           paste0(.err, collapse="', '"), "'",
           call.=FALSE)
    } else {
      # drop in the second
      .minfo("duplicated eta parameters when combining 2 models")
      .minfo(paste0("keeping initialization from first model: '",
                    paste(.both, collapse="', '"), "'"))

      .ini2eta <- .ini2eta[!(.ini2eta$name %in% .both),, drop =FALSE]
    }
  }
  .maxEta <- suppressWarnings(max(.ini1eta$neta1))
  if (is.finite(.maxEta)) {
    .ini2eta$neta1 <- .ini2eta$neta1 + .maxEta
    .ini2eta$neta2 <- .ini2eta$neta2 + .maxEta
  }
  .iniE <- rbind(.ini1eta, .ini2eta)
  if (length(.iniE$name) > 0) {
    .iniE <- .iniE[order(.iniE$neta1, .iniE$neta2), ]
    .ini <- rbind(.iniT, .iniE)
  } else {
    .ini <- .iniT[order(.iniT$ntheta), ]
  }
  .combineModelLines(model1, model2, .ini)
}

#' Append two rxui models together
#'
#' @param ... models to append together
#' @param common boolean; when `TRUE` warn if the models have no variables in common
#' @return New model with both models appended together
#' @author Matthew L. Fidler
#' @export
#' @examples
#'
#' \donttest{
#'
#' ocmt <- function() {
#'   ini({
#'     tka <- exp(0.45) # Ka
#'     tcl <- exp(1) # Cl
#'     tv <- exp(3.45); # log V
#'     ## the label("Label name") works with all models
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- tka
#'     cl <- tcl
#'     v <- tv
#'     d/dt(depot) <- -ka * depot
#'     d/dt(center) <- ka * depot - cl / v * center
#'     cp <- center / v
#'     cp ~ add(add.sd)
#'   })
#' }
#'
#' idr <- function() {
#'   ini({
#'     tkin <- log(1)
#'     tkout <- log(1)
#'     tic50 <- log(10)
#'     gamma <- fix(1)
#'     idr.sd <- 1
#'   })
#'   model({
#'     kin <- exp(tkin)
#'     kout <- exp(tkout)
#'     ic50 <- exp(tic50)
#'     d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
#'     eff ~ add(idr.sd)
#'   })
#' }
#'
#' rxAppendModel(ocmt |>
#'   model(ceff=cp,append=TRUE), idr)
#'
#' }
#'
rxAppendModel <- function(..., common=TRUE) {
  .env <- new.env(parent=emptyenv())
  .env$ret <- NULL
  .lst <- list(...)
  lapply(seq_along(.lst), function(i) {
    .m <- .lst[[i]]
    if (is.null(.env$ret)) {
      .env$ret <- .m
    } else {
      .env$ret <- rxAppendModel_(.env$ret, .m, common=common)
    }
  })
  .env$ret
}

.forderEnv <- new.env()
.forderEnv$useBase <- FALSE


.forder3 <- function(c1,c2,c3, decreasing=FALSE) {
  data.table::data.table(c1=c1,
                         c2=c2,
                         c3=c3,
                         decreasing=decreasing,
                         na.last=TRUE)[order(c1, c2, c3), which=TRUE]
}

.border3 <- function(c1,c2,c3, decreasing=FALSE) {
  base::order(c1, c2, c3, decreasing=decreasing, na.last=NA, method="radix")
}

.forder1 <- function(c1, decreasing=FALSE) {
  data.table::data.table(c1=c1)[order(c1, decreasing=decreasing, na.last=TRUE), which=TRUE]
}

.border1 <- function(c1, decreasing=FALSE) {
  base::order(c1, na.last=NA, decreasing=decreasing, method="radix")
}

.order1 <- function(c1, decreasing=FALSE) {
  if (.forderEnv$useBase) {
    .border1(c1, decreasing=decreasing)
  } else {
    .forder1(c1, decreasing=decreasing)
  }
}

.order3 <- function(c1,c2,c3, decreasing=FALSE) {
  if (.forderEnv$useBase) {
    .border3(c1,c2,c3, decreasing=decreasing)
  } else {
    .forder3(c1,c2,c3, decreasing=decreasing)
  }
}
#' Force using base order for rxode2 radix sorting
#'
#' @param forceBase boolean indicating if rxode2 should use R's
#'   [order()] for radix sorting instead of
#'   `data.table`'s parallel radix sorting.
#'
#' @return value of `forceBase` (can change if `data.table` is not
#'   available)
#'
#' @examples
#' \donttest{
#' forderForceBase(TRUE) # Use base `order` for rxode2 sorts
#' forderForceBase(FALSE) # Use `data.table` for rxode2 sorts
#' }
#' @export
#' @keywords internal
forderForceBase <- function(forceBase = FALSE){
  if (forceBase) {
    .forderEnv$useBase <- forceBase
  } else if (requireNamespace("data.table", quietly = TRUE)) {
    .forderEnv$useBase <- forceBase
  } else {
    .forderEnv$useBase <- TRUE
  }
  invisible(.forderEnv$useBase)
}

.chin <- function(x, table) {
  x %in% table
}

rex::register_shortcuts("rxode2")
# Hack for Rcpp->R initial values problem
R_NegInf <- -Inf # nolint
R_PosInf <- Inf # nolint
NA_LOGICAL <- NA # nolint

## Must match what `rxode2()` stores for its `linCmtSens` default (the default
## choice vector collapsed to the value `match.arg()` picks): it is folded into
## the parsed md5, so a NULL here would make the first build of a session hash
## differently from every later one (c() drops NULL).
.linCmtSens <- "linCmtA"
.clearME <- function() {
  assignInMyNamespace(".rxMECode", "")
  assignInMyNamespace(".indLinInfo", list())
}
.rxFullPrint <- FALSE
#' Create an ODE-based model specification
#'
#' Create a dynamic ODE-based model object suitably for translation
#' into fast C code
#'
#' @param model This is the ODE model specification.  It can be:
#'
#'
#'  * a string containing the set of ordinary differential
#'     equations (ODE) and other expressions defining the changes in
#'     the dynamic system.
#'
#'  * a file name where the ODE system equation is contained
#'
#'   An ODE expression enclosed in `\{\}`
#'
#' (see also the `filename` argument). For
#'     details, see the sections \dQuote{Details} and
#'     `rxode2 Syntax` below.
#'
#' @param modName a string to be used as the model name. This string
#'     is used for naming various aspects of the computations,
#'     including generating C symbol names, dynamic libraries,
#'     etc. Therefore, it is necessary that `modName` consists of
#'     simple ASCII alphanumeric characters starting with a letter.
#'
#' @param wd character string with a working directory where to
#'     create a subdirectory according to `modName`. When
#'     specified, a subdirectory named after the
#'     \dQuote{`modName.d`} will be created and populated with a
#'     C file, a dynamic loading library, plus various other working
#'     files. If missing, the files are created (and removed) in the
#'     temporary directory, and the rxode2 DLL for the model is
#'     created in the current directory named `rx_????_platform`, for
#'     example `rx_129f8f97fb94a87ca49ca8dafe691e1e_i386.dll`
#'
#' @param filename A file name or connection object where the
#'     ODE-based model specification resides. Only one of `model`
#'     or `filename` may be specified.
#'
#' @param extraC  Extra c code to include in the model.  This can be
#'     useful to specify functions in the model.  These C functions
#'     should usually take `double` precision arguments, and
#'     return `double` precision values.
#'
#' @param debug is a boolean indicating if the executable should be
#'     compiled with verbose debugging information turned on.
#'
#' @param calcSens boolean indicating if rxode2 will calculate the
#'     sensitivities according to the specified ODEs.  May also be a
#'     character vector of the states/parameters whose first-order
#'     sensitivities (`rx__sens_<state>_BY_<param>__`) should be
#'     generated.
#'
#' @param calcSens2 character vector (or `NULL`) requesting
#'     second-order sensitivities in addition to the first-order ones
#'     from `calcSens`.  When supplied, rxode2 also generates the
#'     `rx__sens_<state>_BY_<p>_BY_<q>__` compartments (the Hessian
#'     path), where `p` ranges over `calcSens` and `q` over
#'     `calcSens2`.  Used, for example, for population (`THETA`)
#'     second-order event sensitivities.  `NULL` (the default) skips the
#'     second-order generation.
#'
#' @param calcSens3 character vector (or `NULL`) requesting third-order
#'     sensitivities in addition to the first- and second-order ones.
#'     Requires `calcSens2` to also be supplied (every `calcSens3`
#'     parameter needs its own already-built second-order sensitivity
#'     compartment, from the pairing of `calcSens2` with `calcSens3`, to
#'     reference -- so `calcSens3` should be a subset of `calcSens2`,
#'     which itself should be a subset of `calcSens`, mirroring how
#'     `calcSens2` is used everywhere else in rxode2 today).  When
#'     supplied, generates the `rx__sens_<state>_BY_<p>_BY_<q>_BY_<r>__`
#'     compartments, where `p` ranges over `calcSens`, `q` over
#'     `calcSens2`, and `r` over `calcSens3`, via `rxExpandSens3_()`.
#'     `NULL` (the default) skips the third-order generation.
#'
#' @param eventSens controls how dosing/event-parameter (alag, F, rate,
#'     dur, amt) sensitivities are computed when sensitivities are
#'     generated: `"jump"` injects the analytic event ("jump")
#'     sensitivities into the sensitivity states at each dosing event,
#'     `"fd"` keeps the legacy finite-difference behavior (the
#'     backward-compatible opt-out), and `"both"` computes both for
#'     cross-checking.  `NULL` (the default) uses
#'     `getOption("rxode2.eventSens", "fd")`.  When not `"fd"` and
#'     `calcSens` is supplied, `calcJac` is forced to `TRUE` so the
#'     Jacobian is available for the jump injection.
#'
#' @param calcJac boolean indicating if rxode2 will calculate the
#'     Jacobain according to the specified ODEs.
#'
#' @param collapseModel boolean indicating if rxode2 will remove all
#'     LHS variables when calculating sensitivities.
#'
#' @param package Package name for pre-compiled binaries.
#'
#' @param ... ignored arguments.
#'
#' @param linCmtSens The method to calculate the linCmt() solutions
#'
#' @param indLin Calculate inductive linearization matrices and
#'     compile with inductive linearization support.
#'
#' @param verbose When `TRUE` be verbose with the linear
#'   compartmental model
#'
#' @param fullPrint When using `printf` within the model, if `TRUE`
#'   print on every step (except ME/indLin), otherwise when `FALSE`
#'   print only when calculating the `d/dt`
#'
#' @inheritParams rxode2parse
#'
#' @details
#'
#' The `Rx` in the name `rxode2` is meant to suggest the
#' abbreviation *Rx* for a medical prescription, and thus to
#' suggest the package emphasis on pharmacometrics modeling, including
#' pharmacokinetics (PK), pharmacodynamics (PD), disease progression,
#' drug-disease modeling, etc.
#'
#'
#' @includeRmd man/rmdhunks/rxode2-create-models.Rmd
#'
#' @includeRmd man/rmdhunks/rxode2-syntax-hunk.Rmd
#'
#' @return An object (environment) of class `rxode2` (see Chambers and Temple Lang (2001))
#'      consisting of the following list of strings and functions:
#'
#'     * `model` a character string holding the source model specification.
#'     * `get.modelVars`a function that returns a list with 3 character
#'         vectors, `params`, `state`, and `lhs` of variable names used in the model
#'         specification. These will be output when the model is computed (i.e., the ODE solved by integration).
#'
#'       * `solve`{this function solves (integrates) the ODE. This
#'           is done by passing the code to [rxSolve()].
#'           This is as if you called `rxSolve(rxode2object, ...)`,
#'           but returns a matrix instead of a rxSolve object.
#'
#'           `params`: a numeric named vector with values for every parameter
#'           in the ODE system; the names must correspond to the parameter
#'           identifiers used in the ODE specification;
#'
#'           `events`: an `eventTable` object describing the
#'           input (e.g., doses) to the dynamic system and observation
#'           sampling time points (see  [eventTable()]);
#'
#'           `inits`: a vector of initial values of the state variables
#'           (e.g., amounts in each compartment), and the order in this vector
#'           must be the same as the state variables (e.g., PK/PD compartments);
#'
#'
#'           `stiff`: a logical (`TRUE` by default) indicating whether
#'           the ODE system is stiff or not.
#'
#'           For stiff ODE systems (`stiff = TRUE`), `rxode2` uses
#'           the LSODA (Livermore Solver for Ordinary Differential Equations)
#'           Fortran package, which implements an automatic method switching
#'           for stiff and non-stiff problems along the integration interval,
#'           authored by Hindmarsh and Petzold (2003).
#'
#'           For non-stiff systems (`stiff = FALSE`), `rxode2` uses `DOP853`,
#'           an explicit Runge-Kutta method of order 8(5, 3) of Dormand and Prince
#'           as implemented in C by Hairer and Wanner (1993).
#'
#'           `trans_abs`: a logical (`FALSE` by default) indicating
#'           whether to fit a transit absorption term
#'           (TODO: need further documentation and example);
#'
#'           `atol`: a numeric absolute tolerance (1e-08 by default);
#'
#'           `rtol`: a numeric relative tolerance (1e-06 by default).
#'
#'           The output of \dQuote{solve} is a matrix with as many rows as there
#'           are sampled time points and as many columns as system variables
#'           (as defined by the ODEs and additional assignments in the rxode2 model
#'               code).}
#'
#'       * `isValid` a function that (naively) checks for model validity,
#'           namely that the C object code reflects the latest model
#'           specification.
#'       * `version` a string with the version of the `rxode2`
#'           object (not the package).
#'       * `dynLoad` a function with one `force = FALSE` argument
#'           that dynamically loads the object code if needed.
#'       * `dynUnload` a function with no argument that unloads
#'           the model object code.
#'       * `delete` removes all created model files, including C and DLL files.
#'           The model object is no longer valid and should be removed, e.g.,
#'           `rm(m1)`.
#'       * `run` deprecated, use `solve`.
#'       * `get.index` deprecated.
#'       * `getObj` internal (not user callable) function.
#'
#' @references
#'
#' Chamber, J. M. and Temple Lang, D. (2001)
#' *Object Oriented Programming in R*.
#' R News, Vol. 1, No. 3, September 2001.
#' <https://cran.r-project.org/doc/Rnews/Rnews_2001-3.pdf>.
#'
#' Hindmarsh, A. C.
#' *ODEPACK, A Systematized Collection of ODE Solvers*.
#' Scientific Computing, R. S. Stepleman et al. (Eds.),
#' North-Holland, Amsterdam, 1983, pp. 55-64.
#'
#' Petzold, L. R.
#' *Automatic Selection of Methods for Solving Stiff and Nonstiff
#' Systems of Ordinary Differential Equations*.
#' Siam J. Sci. Stat. Comput. 4 (1983), pp. 136-148.
#'
#' Hairer, E., Norsett, S. P., and Wanner, G.
#' *Solving ordinary differential equations I, nonstiff problems*.
#' 2nd edition, Springer Series in Computational Mathematics,
#' Springer-Verlag (1993).
#'
#' Plevyak, J.
#' *`dparser`*, <https://dparser.sourceforge.net/>. Web. 12 Oct. 2015.
#'
#' @author Melissa Hallow, Wenping Wang and Matthew Fidler
#'
#' @seealso [eventTable()], [et()], [add.sampling()], [add.dosing()]
#'
#' @examples
#' \donttest{
#'
#' mod <- function() {
#'   ini({
#'     KA   <- .291
#'     CL   <- 18.6
#'     V2   <- 40.2
#'     Q    <- 10.5
#'     V3   <- 297.0
#'     Kin  <- 1.0
#'     Kout <- 1.0
#'     EC50 <- 200.0
#'   })
#'   model({
#'     # A 4-compartment model, 3 PK and a PD (effect) compartment
#'     # (notice state variable names 'depot', 'centr', 'peri', 'eff')
#'     C2 <- centr/V2
#'     C3 <- peri/V3
#'     d/dt(depot) <- -KA*depot;
#'     d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
#'     d/dt(peri)  <-                    Q*C2 - Q*C3;
#'     d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff;
#'     eff(0)      <- 1
#'   })
#' }
#'
#' m1 <- rxode2(mod)
#' print(m1)
#'
#' # Step 2 - Create the model input as an EventTable,
#' # including dosing and observation (sampling) events
#'
#' # QD (once daily) dosing for 5 days.
#'
#' qd <- et(amountUnits = "ug", timeUnits = "hours") |>
#'   et(amt = 10000, addl = 4, ii = 24)
#'
#' # Sample the system hourly during the first day, every 8 hours
#' # then after
#' qd <- qd |> et(0:24) |>
#'   et(from = 24 + 8, to = 5 * 24, by = 8)
#'
#' # Step 3 - solve the system
#'
#' qd.cp <- rxSolve(m1, qd)
#'
#' head(qd.cp)
#'
#' }
#'
#' @keywords models nonlinear
#' @concept Nonlinear regression
#' @concept ODE models
#' @concept Ordinary differential equations
#' @concept Pharmacokinetics (PK)
#' @concept Pharmacodynamics (PD)
#' @useDynLib rxode2, .registration=TRUE
#' @eval .rxodeBuildCode()
#' @importFrom PreciseSums fsum
#' @importFrom Rcpp evalCpp
#' @importFrom RcppParallel RcppParallelLibs
#' @importFrom checkmate qassert
#' @importFrom utils getFromNamespace assignInMyNamespace download.file head sessionInfo compareVersion packageVersion removeSource
#' @importFrom stats setNames update dnorm integrate
#' @importFrom methods signature is
#' @importFrom memoise memoise is.memoised
#' @importFrom utils capture.output
#' @import tools
#' @import data.table
#' @export
rxode2 <- # nolint
  function(model, modName = basename(wd),
           wd = getwd(),
           filename = NULL, extraC = NULL, debug = FALSE, calcJac = NULL, calcSens = NULL,
           calcSens2 = NULL,
           calcSens3 = NULL,
           collapseModel = FALSE, package = NULL, ...,
           linCmtSens = c("linCmtA", "linCmtB"),
           indLin = FALSE,
           eventSens = NULL,
           verbose = FALSE,
           fullPrint=getOption("rxode2.fullPrint", FALSE),
           envir=parent.frame()) {
    if (!missing(wd) && missing(modName)) {
      stop("working directory specified, but modName not declared, need to specify modName to create rxode2 c-files as a sub-directory of `wd`",
              call.=FALSE)
    }
    .udfEnvSet(envir)
    assignInMyNamespace(".rxFullPrint", fullPrint)
    rxSuppressMsg()
    rxParseSuppressMsg()
    .modelName <- try(as.character(substitute(model)), silent=TRUE)
    if (inherits(.modelName, "try-error")) .modelName <- NULL
    if (!missing(modName)) {
      if (!checkmate::testCharacter(modName, max.len = 1)) {
        stop("'modName' has to be a single length character", call. = FALSE)
      }
    }
    if (!missing(extraC)) {
      if (!checkmate::testAccess(extraC, "r")) {
        stop("'extraC' needs to point to a file that exists and is readable", call. = FALSE)
      }
    }
    if (!checkmate::checkLogical(collapseModel, max.len = 1, any.missing = FALSE)) {
      stop("'collapseModel' needs to be logical", call. = FALSE)
    }
    if (!checkmate::checkLogical(indLin, max.len = 1, any.missing = FALSE)) {
      stop("'indLin' needs to be logical", call. = FALSE)
    }
    if (!checkmate::checkLogical(debug, max.len = 1, any.missing = FALSE)) {
      stop("'debug' needs to be logical", call. = FALSE)
    }
    rxTempDir()
    if (!is.null(package)) {
      if (!checkmate::checkCharacter(package, max.len = 1, any.missing = FALSE)) {
        stop("'package' needs to a single character for the package name",
             call. = FALSE
             )
      }
      if (missing(modName)) {
        stop("with packages 'modName' is required",
             call. = FALSE
             )
      }
      modName <- paste0(package, "_", modName)
    }
    if (!missing(model) && !missing(filename)) {
      stop("must specify exactly one of 'model' or 'filename'",
           call. = FALSE
           )
    }
    if (missing(model) && !missing(filename)) {
      model <- filename
    }
    if (!missing(model) && missing(filename)) {
      if (is(substitute(model), "{")) {
        model <- deparse(substitute(model))
        if (model[1] == "{") {
          model <- model[-1]
          model <- model[-length(model)]
        }
        model <- paste(model, collapse = "\n")
      } else if (inherits(model, "rxUi")) {
        return(model)
      } else if (inherits(model, "function")) {
        .args <- as.list(match.call())[-1]
        if (length(.args) != 1L) {
          stop("model functions can only be called with one argument", call.=FALSE)
        }
        .tmp <- rxUiDecompress(.rxFunction2ui(model))
        assign("modelName", .modelName, envir=.tmp)
        return(rxUiCompress(.tmp))
      } else if (is(model, "rxode2")) {
        package <- get("package", model)
        if (!is.null(package)) {
          modName <- get("modName", model)
        }
        model <- model$.model
        class(model) <- NULL
      } else if ((is(model, "function") || is(model, "call"))) {
        model <- deparse(body(model))[-1]
        model <- paste(model[-length(model)], collapse = "\n")
      }
    }
    .env <- new.env(parent = baseenv())
    ## Event ("jump") sensitivities: when enabled, the runtime jump injection
    ## needs the full state Jacobian, which is otherwise generated only on
    ## demand -- force it on.  See ~/src/rxode2-event-sensitivities-plan.md.
    .eventSensMode <- .rxEventSensMode(eventSens)
    .eventSensActiveReq <- !missing(eventSens) && !identical(.eventSensMode, "fd")
    .eventSensNeedsJac <- .eventSensActiveReq && !is.null(calcSens)
    ## Fold the mode into the parsed md5/cache key for the duration of this build
    ## (reset after) so "fd" and "jump" of the same model do not collide in the
    ## compiled-DLL cache.  "fd" -> "" -> md5 unchanged.
    assignInMyNamespace(".rxEventSensCacheKey",
                        if (.eventSensActiveReq) .eventSensMode else "")
    on.exit(assignInMyNamespace(".rxEventSensCacheKey", ""), add = TRUE)
    ## Set BEFORE the parse: `.linCmtSens` is folded into the parsed md5, so
    ## assigning it afterwards hashed this build with the previous call's value.
    ## The default is the whole choice vector, which `match.arg()` resolves to
    ## its first element -- so collapse an in-effect default to that scalar and
    ## the default and an explicit "linCmtA" share one cache key instead of
    ## compiling the identical model twice.  Only the untouched default is
    ## collapsed; an explicit value is passed through so validation is unchanged.
    if (identical(linCmtSens, c("linCmtA", "linCmtB"))) {
      linCmtSens <- "linCmtA"
    }
    assignInMyNamespace(".linCmtSens", linCmtSens)
    .env$.mv <- rxGetModel(model, calcSens = calcSens, calcJac = calcJac, collapseModel = collapseModel, indLin = indLin, calcSens2 = calcSens2, calcSens3 = calcSens3)
    .isLinCmt <- .Call(`_rxode2_isLinCmt`) == 1L
    if (.eventSensNeedsJac && !.isLinCmt && !isTRUE(calcJac)) {
      calcJac <- TRUE
      .env$.mv <- rxGetModel(model, calcSens = calcSens, calcJac = calcJac,
                             collapseModel = collapseModel, indLin = indLin,
                             calcSens2 = calcSens2, calcSens3 = calcSens3)
      .isLinCmt <- .Call(`_rxode2_isLinCmt`) == 1L
    }
    if (.isLinCmt) {
      .env$.linCmtM <- rxNorm(.env$.mv)
      .vars <- c(.env$.mv$params, .env$.mv$lhs, .env$.mv$slhs)
      .env$.mv <- rxGetModel(.Call(
        `_rxode2_linCmtGen`,
        length(.env$.mv$state),
        .vars,
        setNames(
          c(
            "linCmtA" = 1L, "linCmtB" = 2L
          )[match.arg(linCmtSens)],
          NULL
        ), verbose
      ),
      calcSens = calcSens, calcJac = calcJac, collapseModel = collapseModel,
      indLin = indLin, calcSens2 = calcSens2, calcSens3 = calcSens3)
    }
    model <- rxNorm(.env$.mv)
    class(model) <- "rxModelText"
    .env$.model <- model
    .env$missing.modName <- missing(modName)
    wd <- .normalizePath(wd, "/", mustWork = FALSE)
    if (.env$missing.modName) {
      if (getOption("rxode2.tempfiles", TRUE)) {
        .env$mdir <- suppressMessages(.normalizePath(rxTempDir(), mustWork = FALSE))
      } else {
        .env$mdir <- suppressMessages(.normalizePath(wd, mustWork = FALSE))
      }
    } else {
      .env$mdir <- suppressMessages(.normalizePath(file.path(wd, sprintf("%s.d", modName)), mustWork = FALSE))
    }

    if (!file.exists(wd)) {
      dir.create(wd, recursive = TRUE, showWarnings = FALSE)
    }

    .env$modName <- modName
    .env$model <- model
    .env$extraC <- extraC
    .env$debug <- debug
    .env$calcJac <- calcJac
    .env$calcSens <- calcSens
    .eventSensEffectiveMode <- .rxEventSensEffectiveMode(.eventSensMode, .env$.mv)
    ## Warn when sensitivities are requested on a linCmt() model whose ODE
    ## compartment name collides with a linCmt reserved name: the ODE state loses
    ## its sensitivity expansion, so its sensitivities are silently incorrect.
    if (!is.null(calcSens)) {
      .linCollide <- .rxLinCmtNameCollision(.env$.mv)
      if (length(.linCollide) > 0L) {
        warning("ODE compartment(s) '", paste(.linCollide, collapse = "', '"),
                "' share a name with linCmt() reserved compartments; ",
                "sensitivities will be incorrect -- rename them", call. = FALSE)
      }
    }
    .indLinSens <- length(.env$.mv$indLin) > 0L &&
      length(.env$.mv$sens) > 0L
    if (.indLinSens) {
      .eventSensEffectiveMode <- "jump"
    }
    .eventSensActive <- (!missing(eventSens) && !identical(.eventSensEffectiveMode, "fd")) ||
      .indLinSens
    .eventSensOdeStates <- setdiff(.env$.mv$normal.state, .rxLinCmt(.env$.mv))
    .eventSensNeedsJac <- .eventSensActive && !is.null(calcSens) &&
      length(.eventSensOdeStates) > 0L
    if (.eventSensNeedsJac && !isTRUE(calcJac)) {
      calcJac <- TRUE
      .env$.mv <- rxGetModel(rxNorm(.env$.mv), calcSens = calcSens,
                             calcJac = calcJac, collapseModel = collapseModel,
                             indLin = indLin, calcSens2 = calcSens2, calcSens3 = calcSens3)
      .eventSensOdeStates <- setdiff(.env$.mv$normal.state, .rxLinCmt(.env$.mv))
    }
    .eventSensCacheKey <- if (.eventSensActive) .eventSensEffectiveMode else ""
    if (!identical(.eventSensCacheKey, .rxEventSensCacheKey)) {
      assignInMyNamespace(".rxEventSensCacheKey", .eventSensCacheKey)
      .env$.mv <- rxGetModel(rxNorm(.env$.mv), calcSens = calcSens,
                             calcJac = calcJac, collapseModel = collapseModel,
                             indLin = indLin, calcSens2 = calcSens2, calcSens3 = calcSens3)
      .eventSensOdeStates <- setdiff(.env$.mv$normal.state, .rxLinCmt(.env$.mv))
    }
    .env$eventSens <- .eventSensEffectiveMode
    ## Phase-A event-sensitivity metadata (index map + dosing-parameter total
    ## derivatives); NULL for mode "fd" or models without sensitivities.
    .env$eventSensInfo <- if (.eventSensActive) .rxEventSensInfo(.env$.mv, .eventSensEffectiveMode) else NULL
    .env$collapseModel <- collapseModel

    .env$wd <- wd
    .env$package <- package
    if (!is.null(.env$package)) {
      .env$mdir <- .rxPkgDir(.env)
    }
    .env$compile <- eval(bquote(function() {
      with(.(.env), {
        .rx <- base::loadNamespace("rxode2")
        if (!file.exists(wd)) {
          dir.create(wd, recursive = TRUE, showWarnings = FALSE)
        }
        on.exit({
          .rx$.clearME()
        })
        .rx$.rxWithWd(wd, {
          rxode2::.extraC(extraC)
          ## Event ("jump") sensitivities: dLag/dF body lines passed straight to
          ## codegen (empty for mode "fd" or models without sensitivities).
          .esCode <- .rx$.rxEventSensCodeStrings(eventSensInfo)
          if (missing.modName) {
            .rxDll <- .rx$rxCompile(.mv,
                                    debug = debug,
                                    package = .(.env$package),
                                    eventSensCode = .esCode
                                    )
          } else {
            .rxDll <- .rx$rxCompile(.mv,
                                    dir = mdir,
                                    debug = debug, modName = modName,
                                    package = .(.env$package),
                                    eventSensCode = .esCode
                                    )
          }
          .rxDll$linCmtM <- .(ifelse(exists(".linCmtM", .env),
                                     get(".linCmtM", .env), NA
                                     ))
          assign("rxDll", .rxDll, envir = .(.env))
          assign(".mv", .rxDll$modVars, envir = .(.env))
        })
      })
    }))
    rxode2::.extraC(extraC)
    .env$compile()
    .env$get.modelVars <- eval(bquote(function() {
      with(.(.env), {
        .ret <- .mv[c("params", "state", "lhs")]
        .p <- .ret["params"]
        .ini <- names(.mv$.ini)
        .init <- rxode2::rxInit(rxDll)
        .ret$params <- .ret$params[!(.ret$params %in%  names(.init))]
        class(.ret) <- "list"
        return(.ret)
      })
    }))
    .env$state <- .env$.mv$state
    if (.env$.mv$extraCmt == 1) {
      .extra <- c("central", .env$.mv$stateExtra)
    } else if (.env$.mv$extraCmt == 2) {
      .extra <- c("depot", "central", .env$.mv$stateExtra)
    } else {
      .extra <- .env$.mv$stateExtra
    }
    .env$stateExtra <- .extra
    .env$lhs <- .env$.mv$lhs
    .env$params <- .env$.mv$params
    .env$version <- .rxVersion["version"]
    .env$solve <- eval(bquote(function(..., returnType= "matrix", object = NULL) {
      rxode2::rxSolve(object = get("rxDll", envir = .(.env)), ..., returnType = "matrix")
    }))
    .env$dll <- new.env(parent = baseenv())
    .env$assignPtr <- eval(bquote(function() {
      rxode2::rxAssignPtr(get("rxDll", envir = .(.env)))
    }))
    .env$run <- .env$solve
    .env$modName <- modName
    .env$model <- model # actual model code
    ## cmpMgr = cmpMgr,
    .env$dynLoad <- eval(bquote(function(force = FALSE) {
      rx <- .(.env)
      rxode2::rxDynLoad(rx)
    }))
    .env$load <- .env$dynLoad
    .env$dynUnload <- eval(bquote(function() {
      rx <- .(.env)
      rxode2::rxDynUnload(rx)
    }))
    .env$unload <- .env$dynUnload
    .pkgStuff <- FALSE
    if (!is.null(.env$package)) {
      if (regexpr("_new", .env$modName) == -1) {
        .pkgStuff <- TRUE
        .env$isValid <- eval(bquote(function() {
          if (!all(is.null(getLoadedDLLs()[[.(.env$package)]]))) {
            if (loadNamespace("rxode2")$.pkgModelCurrent &&
                                        utils::packageVersion("rxode2") == .(utils::packageVersion("rxode2"))) {
              return(TRUE)
            } else {
              return(FALSE)
            }
          } else {
            return(file.exists(rxode2::rxDll(get("rxDll", envir = .(.env)))))
          }
        }))
        .env$isLoaded <- eval(bquote(function() {
          if ((!all(is.null(getLoadedDLLs()[[.(.env$package)]]))) &&
                loadNamespace("rxode2")$.pkgModelCurrent &&
                                      utils::packageVersion("rxode2") == .(utils::packageVersion("rxode2"))) {
            return(TRUE)
          } else {
            rx <- .(.env)
            rxode2::rxIsLoaded(rx)
          }
        }))
        .env$delete <- eval(bquote(function() {
          if ((!all(is.null(getLoadedDLLs()[[.(.env$package)]]))) &&
                loadNamespace("rxode2")$.pkgModelCurrent &&
                                      utils::packageVersion("rxode2") == .(utils::packageVersion("rxode2"))) {
            stop("cannot delete Dll in package", call. = FALSE)
          } else {
            rx <- .(.env)
            rxode2::rxDelete(rx)
          }
        }))
      }
    }
    if (!.pkgStuff) {
      .env$isValid <- eval(bquote(function() {
        return(file.exists(rxode2::rxDll(get("rxDll", envir = .(.env)))))
      }))
      .env$isLoaded <- eval(bquote(function() {
        rx <- .(.env)
        rxode2::rxIsLoaded(rx)
      }))
      .env$delete <- eval(bquote(function() {
        rx <- .(.env)
        rxode2::rxDelete(rx)
      }))
    }

    .env$parse <- with(.env, function() {
      stop("'$parse' is no longer supported", call. = FALSE)
    })
    .env$get.index <- eval(bquote(function(s) {
      return(rxState(get("rxDll", envir = .(.env)), s))
    }))
    .mv <- .env$.mv
    .env$lib.name <- .mv$trans["lib.name"] # nolint
    tmp <- list(
      dllfile = rxode2::rxDll(.env$rxDll),
      ode_solver = as.vector(.mv$trans["ode_solver"]),
      ode_solver_ptr = as.vector(.mv$trans["ode_solver_ptr"]),
      prefix = as.vector(.mv$trans["prefix"]),
      model = model,
      isValid = eval(bquote(function() {
        with(.(.env), isValid())
      })),
      parse = eval(bquote(function() {
        with(.(.env), parse())
      })),
      compile = eval(bquote(function() {
        with(.(.env), compile())
      })),
      dynLoad = eval(bquote(function() {
        with(.(.env), dynLoad())
      })),
      dynUnload = eval(bquote(function() {
        with(.(.env), dynUnload())
      })),
      modelDir = .env$mdir, # model directory
      get.modelVars = eval(bquote(function() {
        with(.(.env), get.modelVars())
      })),
      delete = eval(bquote(function() {
        with(.(.env), delete())
      })),
      get.index = eval(bquote(function(...) {
        with(.(.env), get.index(...))
      })),
      .rxDll = .env$rxDll,
      rxDll = eval(bquote(function() {
        with(.(.env), return(rxDll))
      }))
    )
    tmp <- list2env(tmp, parent = .env)
    class(tmp) <- "RxCompilationManager"
    .env$cmpMgr <- tmp
    ## Strip source refs from the generated helper closures.  Under
    ## devtools::load_all() (keep.source=TRUE) each eval(bquote(function(){...}))
    ## closure carries a srcref to the pkgload srcfile of rxode2.R, whose lazy
    ## `lines` promise captures the entire session -- serializing any one of them
    ## drags the whole session into the rds.  removeSource() drops the srcref
    ## (no-op for installed/source-less builds).
    for (.n in ls(.env, all.names = TRUE)) {
      if (is.function(.env[[.n]])) assign(.n, removeSource(.env[[.n]]), envir = .env)
    }
    for (.n in ls(.env$cmpMgr, all.names = TRUE)) {
      .f <- get(.n, envir = .env$cmpMgr)
      if (is.function(.f)) assign(.n, removeSource(.f), envir = .env$cmpMgr)
    }
    .env$calcJac <- (length(.mv$dfdy) > 0)
    .env$calcSens <- (length(.mv$sens) > 0)
    class(.env) <- "rxode2"
    rxode2::rxForget()
    if (!is.null(.env$package)) {
      .o <- rxDll(.env)
      .o <- paste0(substr(.o, 0, nchar(.o) - nchar(.Platform$dynlib.ext)), ".o")
      if (file.exists(.o)) {
        unlink(.o)
      }
      .make <- file.path(.env$mdir, "Makevars")
      if (file.exists(.make)) {
        unlink(.make)
      }
      if (.rxPkgLoaded(.env$package)) {
        .ns <- loadNamespace(.env$package)
        if (!exists(".rxUpdated", .ns)) {
          stop("cannot update package model", call. = FALSE)
        } else {
          .as <- .ns$.rxUpdated
          assign(.env$modName, .env)
        }
      }
    } else {
      rxode2::rxIsLoaded(.env) # Show this is loaded.
    }
    return(.env)
  }

#' @rdname rxode2
#' @export
RxODE <- rxode2

#' @rdname rxode2
#' @export
rxode <- rxode2

#' Get model properties without compiling it.
#'
#' @param mv rxode2 model variables
#'
#' Sensitivity-named states (`rx__sens_<state>_BY_<var>__`) are
#' stripped of their `d/dt()` when calcJac/calcSens regenerate the
#' sensitivity block.  A state that is read by `delay()` is a
#' load-bearing ODE (`delay(X, T)` requires `d/dt(X)`), so it must
#' never be stripped -- e.g. nlmixr2est's analytic-cov augmented model
#' supplies delayed sensitivity ODEs directly.  Returns the sens
#' states safe to strip (those not referenced by any delay() term).
#'
#' @return rxode2 trans list
#'
#' @author Matthew L. Fidler
.rxSensStrippable <- function(mv) {
  .sens <- mv$sens
  if (length(.sens) == 0) return(.sens)
  if (!isTRUE(mv$flags[["hasDelay"]] == 1L)) return(.sens)
  .delayed <- tryCatch(.rxDelayTerms(mv)$state, error = function(e) NULL)
  setdiff(.sens, .delayed)
}

## symengine rewrites a literal `THETA_1_`/`ETA_1_` parameter to the reserved
## indexed form `THETA[1]`/`ETA[1]` inside expressions, but the Jacobian dy()
## targets and the original param() declaration keep the literal form.  In a
## calcJac/calcSens reassembly the two forms then coexist and rxode2 treats them
## as two distinct parameters (so `params=` keyed by the literal name cannot fill
## the indexed one).  Restore the literal name for exactly the indices the
## original model declared as literals -- native `THETA[1]` models (whose params
## are the indexed form) declare no such literal and are left untouched.
.rxRestoreLiteralThetaEta <- function(txt, params) {
  .lit <- grep("^(THETA|ETA)_[0-9]+_$", params, value = TRUE)
  for (.p in .lit) {
    .m <- regmatches(.p, regexec("^(THETA|ETA)_([0-9]+)_$", .p))[[1]]
    txt <- gsub(paste0(.m[2], "[", .m[3], "]"), .p, txt, fixed = TRUE)
  }
  txt
}

#' Get the rxode2 model variables without compiling the model
#'
#' Normalizes the input to an `rxModelVars` object and optionally
#' rewrites the model text first: adding/stripping forward
#' sensitivity equations (`calcSens`/`calcSens2`/`calcSens3`),
#' adding/removing the `df()/dy()` Jacobian block (`calcJac`),
#' dropping LHS definitions (`collapseModel`), or converting to an
#' inductive linearization model (`indLin`).
#'
#' @param model The model to get the model variables from.  It can
#'   be: rxode2 model text (a character string, `rxModelText`, or a
#'   `{}` expression), a function or call whose body is the model, an
#'   `rxode2` object, an `rxModelVars` object, an `rxDll` object, or
#'   anything else `rxModelVars()` accepts.
#'
#' @inheritParams rxode2
#'
#' @return An `rxModelVars` object of the (possibly rewritten) model
#'
#' @author Matthew L. Fidler
#'
#' @examples
#'
#' rxGetModel("d/dt(depot) = -ka*depot; d/dt(central) = ka*depot - kel*central")
#'
#' @export
#' @keywords internal
rxGetModel <- function(model, calcSens = NULL, calcJac = NULL, collapseModel = NULL, indLin = FALSE,
                       calcSens2 = NULL, calcSens3 = NULL) {
  if (is(substitute(model), "call")) {
    model <- model
  }
  if (is(substitute(model), "{")) {
    model <- deparse(substitute(model))
    if (model[1] == "{") {
      model <- model[-1]
      model <- model[-length(model)]
    }
    model <- paste(model, collapse = "\n")
  } else if (inherits(model, "function") || inherits(model, "call")) {
    model <- deparse(body(model))
    if (model[1] == "{") {
      model <- model[-1]
      model <- model[-length(model)]
    }
    model <- paste(model, collapse = "\n")
  } else if (inherits(model, "name")) {
    model <- eval(model)
  } else if (inherits(model, "character") || inherits(model, "rxModelText")) {
    model <- as.vector(model)
  } else if (inherits(model, "rxode2")) {
    model <- rxModelVars(model)
    ## class(model) <- NULL;
  } else if (inherits(model, "rxModelVars")) {
  } else if (inherits(model, "rxDll")) {
    model <- model$args$model
  } else {
    model <- rxModelVars(model)
    if (!inherits(model, "rxModelVars")) {
      stop("cannot figure out how to handle the model argument", call. = FALSE)
    }
  }
  .oldEventSensKey <- .rxEventSensCacheKey
  on.exit(assignInMyNamespace(".rxEventSensCacheKey", .oldEventSensKey), add = TRUE)
  ## `.indLinInfo` is a session global folded into the parsed md5 by rxMd5().  It
  ## is an OUTPUT of parsing (this call's own indLin/mexp branches below re-set
  ## it before codegen), never an input, so clear any value left by a prior
  ## build BEFORE the parse below hashes the model -- otherwise a stale
  ## descriptor from an earlier matrix-exponential build (or from a rxSensMatExp
  ## that never fully compiled) makes an unrelated plain model hash differently
  ## depending on build order.
  assignInMyNamespace(".indLinInfo", list())
  .ret <- rxModelVars(model)
  if (!is.null(calcSens)) {
    .calcSens <- TRUE
    if (is(calcSens, "logical")) {
      if (!calcSens) {
        .calcSens <- FALSE
      }
    }
    if (.calcSens) {
      if (length(rxState(.ret)) == 0L) {
        stop("sensitivities do not make sense for models without ODEs", call. = FALSE)
      }
      .stateInfo <- .rxGenFunState(.ret)
      .s <- .rxLoadPrune(.ret, FALSE)
      .s$..stateInfo <- .stateInfo
      .rxJacobian(.s)
      ## base ODE states, captured BEFORE .rxSens splices in the rx__sens_*
      ## variational compartments -- used to build the full-system stiff Jacobian
      ## by reusing the F_X/F_p just computed (see .rxFwdSensJacBlock).
      .baseState <- rxStateOde(.s)
      if (!is(calcJac, "logical")) {
        calcJac <- FALSE
      }
      if (is.null(calcJac)) calcJac <- FALSE
      if (rxIs(calcSens, "logical")) {
        if (calcSens) {
          calcSens <- .rxParams(model, TRUE)
        }
      }
      ## .rxSens augments the sensitivity ODEs with the delayed (variational)
      ## terms for delay() models (no-op otherwise).
      .rxSens(.s, calcSens)
      ## Second-order sensitivities (for a Hessian / population THETA path): the
      ## 2nd-order ODEs reference the 1st-order sens compartments, so they are
      ## generated *after* the 1st-order ones and both are emitted into the model.
      ## `..sens2` stays NULL (dropped from the paste) when calcSens2 is absent.
      .sens2 <- NULL
      if (!is.null(calcSens2)) {
        .rxSens(.s, calcSens, calcSens2)
        .sens2 <- .s$..sens2
      }
      ## Third-order sensitivities (Phase H0): the 3rd-order ODEs reference
      ## the 2nd-order sens compartments (the pair calcSens2/calcSens3), so
      ## they are only generated when calcSens2 is *also* present; requires
      ## rxExpandSens3_ (PR #1092), which .rxSens() already dispatches to
      ## when given all three variable sets.
      .sens3 <- NULL
      if (!is.null(calcSens3)) {
        if (is.null(calcSens2)) {
          stop("'calcSens3' requires 'calcSens2' to be supplied", call. = FALSE)
        }
        .rxSens(.s, calcSens, calcSens2, calcSens3)
        .sens3 <- .s$..sens3
      }
      .tmp2 <- .s$..lhs
      if (collapseModel) .tmp2 <- ""
      ## Jacobian for stiff (Rosenbrock/implicit) solvers.  `.rxJacobian(.s)`
      ## above ran over the ORIGINAL states only (it feeds the variational
      ## equations), so `.s$..jacobian` is the base-system Jacobian -- INCOMPLETE
      ## for the expanded state, because each rx__sens_* variational compartment
      ## adds its own nonzero Jacobian rows/cols (the sens x sens block is F_X,
      ## sens x base couples through dF_X/dy).  A stiff solver integrates the
      ## FULL expanded system, so it needs that system's Jacobian.  Recompute
      ## calcJac AFTER the sensitivity expansion: assemble the expanded model
      ## once with no Jacobian, reload it (registers base + sens as ODE states),
      ## and take .rxJacobian over the full state, then splice that df()/dy()
      ## block back in.  Without it the stiff solvers get J == 0 and boost's
      ## rosenbrock4 cannot even find a step size (others fall back to a costly
      ## numeric Jacobian).
      ## The df()/dy() Jacobian block must come AFTER every d/dt definition it
      ## references -- the full-system Jacobian references the rx__sens_* sens
      ## compartments, so it is spliced in after .sens/.sens2/.sens3 (a base-only
      ## Jacobian could sit above them, the full one cannot).
      .assembleSens <- function(.jacBlock) paste(c(
        .s$..stateInfo["state"],
        .s$..lhs0,
        .s$..ddt,
        .s$..sens,
        .sens2,
        .sens3,
        .jacBlock,
        ## DDE non-constant pre-history: base past(state,tau)<-expr (dropped from
        ## ..ddt by the symengine interception) + per-sens-compartment histories
        ## past(rx__sens_state_BY_p__,tau)=d expr/d p.  Placed after every d/dt so
        ## the referenced (sensitivity) compartments are already defined.
        .s$..pastLines,
        ## DDE param-dependent-delay dose-jump: alag()/f() on the sens compartments
        ## (no-op unless rxSolve adds the mirroring sens-compartment doses).
        .s$..sensDelayAlagF,
        ## DDE param-dependent-delay 2nd-order breaking-point jump: alag()/f() on
        ## the 2nd-order sens compartments (no-op unless rxSolve adds the t0 doses).
        .s$..sens2DelayAlagF,
        .tmp2,
        .s$..stateInfo["statef"],
        .s$..stateInfo["dvid"],
        ""
      ), collapse = "\n")
      .jacBlock <- ""
      if (calcJac) {
        ## Ordinary ODE forward sensitivities: build the full-system Jacobian by
        ## REUSING the F_X/F_p already computed for the variational equations,
        ## computing only the sens x base block (== 0 for linear F).  Fall back to
        ## a fresh full-system .rxJacobian when 2nd/3rd-order sens compartments are
        ## present (their block structure is not covered by the reuse), or if the
        ## reuse errors for any reason.  (For ODE models the eventSens jump uses a
        ## central difference of dydt, not calc_jac, so this fuller Jacobian only
        ## ever feeds the stiff ODE steppers.  indLin / matExp models are a
        ## matrix-exponential method, out of scope for stiff ODE stepping.)
        .jacLines <- NULL
        if (is.null(calcSens2) && is.null(calcSens3)) {
          .jacLines <- tryCatch(.rxFwdSensJacBlock(.s, .baseState, calcSens),
                                 error = function(e) NULL)
        }
        if (is.null(.jacLines)) {
          .sj <- .rxLoadPrune(rxModelVars(.assembleSens("")), FALSE)
          .jacLines <- .rxJacobian(.sj, FALSE)
        }
        .jacBlock <- paste(.jacLines, collapse = "\n")
      }
      .new <- .assembleSens(.jacBlock)
      .ret <- rxModelVars(.new)
    } else {
      ## calcSens=FALSE removes the sensitivity equations.
      .stateInfo <- .rxGenFunState(.ret)
      .s <- .rxLoadPrune(.ret, FALSE)
      .s$..stateInfo <- .stateInfo
      .sensStrip <- .rxSensStrippable(.ret)
      if (length(.sensStrip) != 0) {
        .new <- setNames(gsub(
          rex::rex("d/dt(", or(.sensStrip), ")=", anything, "\n"), "",
          .ret$model["normModel"]
        ), NULL)
        .ret <- rxModelVars(.new)
      }
      .calcJac <- FALSE
      if (!is.null(calcJac)) {
        if (is(calcJac, "logical")) {
          if (calcJac) {
            .calcJac <- TRUE
          }
        }
      }
      if (.calcJac) {
        .rxJacobian(.s)
        ## calcJac=TRUE, calcSens=FALSE
      }
      .tmp1 <- .s$..jacobian
      if (!.calcJac) .tmp1 <- ""
      .tmp2 <- .s$..lhs
      if (collapseModel) .tmp2 <- ""
      .new <- paste(c(
        .s$..stateInfo["state"],
        .s$..lhs0,
        .s$..ddt,
        .tmp1,
        .tmp2,
        .s$..stateInfo["statef"],
        .s$..stateInfo["dvid"],
        ""
      ), collapse = "\n")
      .ret <- rxModelVars(.new)
    }
  } else if (!is.null(calcJac)) {
    .sensStrip <- .rxSensStrippable(.ret)
    if (length(.sensStrip) != 0) {
      .new <- setNames(gsub(
        rex::rex("d/dt(", or(.sensStrip), ")=", anything, "\n"), "",
        .ret$model["normModel"]
      ), NULL)
      .ret <- rxModelVars(.new)
    }
    .calcJac <- TRUE
    if (is(calcJac, "logical")) {
      if (!calcJac) {
        .calcJac <- FALSE
      }
    }
    .litParams <- .ret$params
    if (.calcJac) {
      if (length(rxState(.ret)) <= 0) {
        ## Jacobian capitalized because it should be spelled with a capital
        stop("Jacobians do not make sense for models without ODEs", call. = FALSE)
      }
      .stateInfo <- .rxGenFunState(.ret)
      .s <- .rxLoadPrune(.ret, FALSE)
      .s$..stateInfo <- .stateInfo
      .rxJacobian(.s)
      .tmp1 <- .s$..jacobian
      if (!.calcJac) .tmp1 <- ""
      .tmp2 <- .s$..lhs
      if (collapseModel) .tmp2 <- ""
      .new <- paste(c(
        .s$..stateInfo["state"],
        .s$..lhs0,
        .s$..ddt,
        .rxPastBaseLinesFromEnv(.s),
        .tmp1,
        .tmp2,
        .s$..stateInfo["statef"],
        .s$..stateInfo["dvid"],
        ""
      ), collapse = "\n")
      .new <- .rxRestoreLiteralThetaEta(.new, .litParams)
      .ret <- rxModelVars(.new)
    } else {
      ## remove Jacobian
      .stateInfo <- .rxGenFunState(.ret)
      .s <- .rxLoadPrune(.ret, FALSE)
      .s$..stateInfo <- .stateInfo
      .tmp2 <- .s$..lhs
      if (collapseModel) .tmp2 <- ""
      .new <- paste(c(
        .s$..stateInfo["state"],
        .s$..lhs0,
        .s$..ddt,
        .rxPastBaseLinesFromEnv(.s),
        .tmp2,
        .s$..stateInfo["statef"],
        .s$..stateInfo["dvid"],
        ""
      ), collapse = "\n")
      .new <- .rxRestoreLiteralThetaEta(.new, .litParams)
      .ret <- rxModelVars(.new)
    }
  }
  if (indLin) {
    ## Convert the ODE model to a pure matrix exponential model using
    ## rxToIndLin() so it uses doIndLin == 1/2 (not the old 3/4 paths).
    .mexpCode <- rxToIndLin(.ret, calcSens = calcSens)
    .ret <- rxModelVars(.mexpCode)
    assignInMyNamespace(".indLinInfo", .ret$indLin)
  } else if (length(.ret$indLin) == 4L) {
    ## NONMEM-like matrix exponential model: the parser already set up
    ## a 4-element indLin list via genModelVars.c (when tb.isMexp=1).
    ## Propagate it to .indLinInfo so codegen serializes it into the DLL.
    assignInMyNamespace(".indLinInfo", .ret$indLin)
  } else {
    ## This model has NO indLin structure -- `.indLinInfo` MUST be reset
    ## here rather than left as-is, or it silently carries over the LAST
    ## matExp/indLin model's 4-element descriptor into this (unrelated)
    ## plain-ODE model's compiled vars via rxCompile.rxModelVars()'s
    ## `.rxModelVarsLast[[17]] <- .indLinInfo` line. `.clearME()` resets
    ## this too, but only fires via `on.exit()` on the FULL compile
    ## closure (R/rxode2.R's `.env$compile`) -- code paths that call
    ## rxGetModel()/rxSensMatExp() WITHOUT ever fully compiling/solving a
    ## model (e.g. inspecting rxSensMatExp()'s generated text directly, as
    ## in several test-mexp-nonmem.R tests) never reach that on.exit, so
    ## `.indLinInfo` leaked forward and corrupted the NEXT unrelated
    ## model's `mv$indLin` (and hence its solve-time method dispatch:
    ## rxSolve.default force-selects method="indLin" whenever
    ## `length(rxModelVars(object)$indLin) > 0L`) -- confirmed by FD/CI as
    ## a real, reproducible "C stack overflow" / uncaught Rcpp::exception
    ## crash (`unsupported indLin code: 0`) when an unrelated population
    ## solve ran right after such a test. Resetting unconditionally here
    ## (whenever THIS model isn't indLin) closes the leak at its source.
    if (length(.indLinInfo) > 0L) {
      assignInMyNamespace(".indLinInfo", list())
    }
  }
  return(.ret)
}

#' Add item to solved system of equations
#'
#' @title rxChain  Chain or add item to solved system of equations
#'
#' @param obj1 Solved object.
#'
#' @param obj2 New object to be added/piped/chained to solved object.
#'
#' @return When `newObject` is an event table, return a new
#'     solved object with the new event table.
#'
#' @author Matthew L. Fidler
#'
#' @keywords internal
#'
#' @export
rxChain <- function(obj1, obj2) {
  .args <- rev(as.list(match.call())[-1])
  names(.args) <- c("obj", "solvedObject")
  return(do.call("rxChain2", .args, envir = parent.frame(1)))
}

#' @rdname rxChain
#' @export
"+.solveRxDll" <- function(obj1, obj2) {
  return(rxode2::rxChain(obj1, obj2))
}

#' Second command in chaining commands
#'
#' This is s3 method is called internally with `+` and `\%>\%` operators.
#'
#' @param obj the object being added/chained/piped to the solved object
#' @param solvedObject the solved object
#' @return chained operation
#' @keywords internal
#' @author Matthew L.Fidler
#' @export
rxChain2 <- function(obj, solvedObject) {
  UseMethod("rxChain2")
}

#' @rdname rxChain2
#' @export
rxChain2.default <- function(obj, solvedObject) {
  .args <- as.list(match.call())
  stop(sprintf(
    gettext("Do not know how to add %s to rxode2 solved object %s"),
    toString(.args[[2]]), toString(.args[[3]])
  ),
  call. = FALSE
  )
}

#' @rdname rxChain2
#' @export
rxChain2.EventTable <- function(obj, solvedObject) {
  .args <- rev(as.list(match.call())[-1])
  names(.args) <- c("object", "events")
  return(do.call("rxSolve", .args, envir = parent.frame(1)))
}

.isLatex <- function() {
  ## nocov start
  if (!("knitr" %in% loadedNamespaces())) {
    return(FALSE)
  }
  get("is_latex_output", asNamespace("knitr"))()
  ## nocov end
}
#' Internal function to figure out if this session supports Unicode
#'
#' @return boolean indicating if this session supports Unicode
#'
#' @keywords internal
#'
#' @export
.useUtf <- function() {
  ## nocov start
  opt <- getOption("cli.unicode", NULL)
  if (!is.null(opt)) {
    isTRUE(opt)
  } else {
    l10n_info()$`UTF-8` && !.isLatex()
  }
  ## nocov end
}
.getBoundRemember <- NULL
.getBound <- function(x, parent = parent.frame(2)) {
  ## nocov start
  if (!is.null(.getBoundRemember)) return(.getBoundRemember)
  .isRx <- try(rxIs(x, "rxode2"), silent = TRUE)
  if (inherits(.isRx, "try-error")) .isRx <- FALSE
  if (.isRx) {
    if (!is.null(x$package)) {
      return(substr(x$modName, nchar(x$package) + 2, nchar(x$modName)))
    }
  }
  bound <- do.call("c", lapply(ls(globalenv()), function(cur) {
    if (identical(parent[[cur]], x)) {
      return(cur)
    }
    return(NULL)
  }))
  if (length(bound) > 1) bound <- bound[1]
  if (length(bound) == 0) {
    bound <- do.call("c", lapply(ls(parent), function(cur) {
      if (identical(parent[[cur]], x)) {
        return(cur)
      }
      return(NULL)
    }))
    if (length(bound) > 1) bound <- bound[1]
    if (length(bound) == 0) {
      bound <- ""
    }
  }
  return(bound)
  ## nocov end
}
.getReal <- function(x) {
  ## Should always be in sync
  if (rxIs(x, "rxode2")) {
    if (!is.null(x$package)) {
      .ns <- loadNamespace(x$package)
      if (exists(".rxUpdated", .ns)) {
        .rxu <- get(".rxUpdated", .ns)
      }
      if (exists(x$modName, .rxu)) {
        return(get(x$modName, .rxu))
      }
    }
  }
  return(x)
}

#' Return the rxode2 coefficients
#'
#' This returns the parameters , state variables
#'
#' @param object is an rxode2 object
#' @param ... ignored arguments
#'
#' @return a rxCoef object with the following
#'
#' * `params`  is a list of strings for parameters for the rxode2 object
#' * `state` is a list of strings for the names of each state in
#'     the rxode2 object.
#' * `ini` is the model specified default values for the
#'     parameters.
#' * `rxode2` is the referring rxode2 object
#' @author Matthew L.Fidler
#' @keywords internal
#' @importFrom stats coef
#'
#' @export
coef.rxode2 <- function(object,
                       ...) {
  .ret <- rxode2::rxModelVars(object)[c("params", "state", "ini", "sens", "fn.ini")]
  .ret$rxode2 <- object
  class(.ret) <- "rxCoef"
  return(.ret)
}

.rxPre <- function(model,
                   modName = NULL) {
  if (!is.null(modName)) {
    if (is.null(.pkg)) {
      .modelPrefix <- paste0(gsub("\\W", "_", modName), "_", .Platform$r_arch, "_")
    } else {
      .modelPrefix <- paste0(gsub("\\W", "_", modName), "_")
    }
  } else {
    .mv <- rxModelVars(model)
    if (.Call(`_rxode2_codeLoaded`) == 0L) .rxModelVarsCharacter(setNames(rxNorm(.mv), NULL))
    .cache <- .rxModelVarsCCache
    .modelPrefix <- paste0("rx_", .mv$md5["parsed_md5"], "_", .Platform$r_arch, "_")
  }
  return(.modelPrefix)
}

.md5Rx <- NULL

#' Return the md5 of an rxode2 object or file
#'
#' This md5 is based on the model and possibly the extra c code
#' supplied for the model.  In addition the md5 is based on syntax
#' options, compiled rxode2 library md5, and the rxode2
#' version/repository.
#'
#' @inheritParams rxode2
#'
#'
#' @param ... ignored arguments
#'
#' @return If this is a rxode2 object, return a named list:
#'
#' * `file_md5` is the model's file's md5
#'
#' * `parsed_md5`  is the parsed model's file's md5.
#'
#' Otherwise return the md5 based on the arguments provided
#'
#' @keywords internal
#' @author Matthew L.Fidler
#' @export rxMd5
rxMd5 <- function(model, # Model File
                  ...) {
  ## rxMd5 returns MD5 of model file.
  ## digest(file = TRUE) includes file times, so it doesn't work for this needs.
  if (missing(model)) {
    return(rxode2.md5)
  } else if (is(model, "character")) {
    if (length(model) == 1) {
      if (file.exists(model)) {
        .ret <- readLines(model, warn = FALSE)
      } else {
        .ret <- model
      }
    } else {
      if (any(names(model) == "normModel")) {
        .ret <- setNames(model["normModel"], NULL)
        if (any(names(model) == "indLin")) {
          if (model["indLin"] != "") {
            .ret <- setNames(paste0(
              .ret, "\n",
              model["indLin"]
            ), NULL)
          }
        }
      } else {
        stop("unknown model", call. = FALSE)
      }
    }
    ## rxSyncOptions()
    .tmp <- c(
      getOption("rxode2.syntax.allow.ini", TRUE),
      getOption("rxode2.calculate.jacobian", FALSE),
      getOption("rxode2.calculate.sensitivity", FALSE))
    .ret <- c(
      .ret, .tmp, .rxIndLinStrategy, .rxIndLinState,
      .linCmtSens, .udfMd5Info(), .rxFullPrint
    )
    if (is.null(.md5Rx)) {
      .tmp <- getLoadedDLLs()$rxode2
      class(.tmp) <- "list"
      assignInMyNamespace(".md5Rx", digest::digest(.tmp$path, serialize = TRUE, file = TRUE, algo = "md5"))
    }
    ## new rxode2 DLLs gives different digests.
    .ret <- c(.ret, .md5Rx)
    ## Add version and github repository information
    .ret <- c(.ret, .rxVersion)
    return(list(
      text = model,
      digest = digest::digest(list(.ret, .indLinInfo), serialize = TRUE, algo = "md5")
    ))
  } else {
    rxode2::rxModelVars(model)$md5
  }
} # end function rxMd5

.rxLastModels <- NULL

.rxShouldUnload <- function(parseMd5) {
  if (is.null(.rxLastModels)) return(TRUE)
  return(!(parseMd5 %in% .rxLastModels))
}

.rxTimeId <- function(parseMd5) {
  if (exists(parseMd5, envir = .rxModels)) {
    .timeId <- get(parseMd5, envir = .rxModels)
  } else {
    .timeId <- as.integer(Sys.time())
    assign(parseMd5, .timeId, envir = .rxModels)
    .rxLastModels <- c(parseMd5, .rxLastModels)
    .nKeep <- getOption("rxode2.dontUnload", 10)
    .nKeep <- as.integer(.nKeep)
    if (.nKeep <= 0L) {
      .rxLastModels <- NULL
    } else if (length(.rxLastModels) < .nKeep) {
      .rxLastModels <- .rxLastModels[seq(1, .nKeep)]
    }
    assignInMyNamespace(".rxLastModels", .rxLastModels)
  }
  return(.timeId)
}

#' Translate the model to C code if needed
#'
#' This function translates the model to C code, if needed
#'
#'
#' @inheritParams rxode2
#'
#'
#' @param modelPrefix Prefix of the model functions that will be
#'     compiled to make sure that multiple rxode2 objects can coexist
#'     in the same R session.
#'
#' @param md5 Is the md5 of the model before parsing, and is used to
#'     embed the md5 into DLL, and then provide for functions like
#'     [rxModelVars()].
#'
#' @param eventSensKey event ("jump") sensitivity mode key folded into
#'     the parsed md5 cache key so the same model text translated in
#'     different modes (e.g. `"fd"` vs `"jump"`) compiles to distinct
#'     DLLs instead of colliding on a cached translation.  Defaults to
#'     the session global `.rxEventSensCacheKey` (set by [rxode2()]
#'     before parsing); an empty string leaves the md5 unchanged so the
#'     legacy `"fd"` path keeps existing caches valid.  It is a formal
#'     argument (rather than read from the global in the body) because
#'     `rxTrans.character` is memoised and `memoise` keys on
#'     default-valued formals, so the cache distinguishes the modes.
#'
#' @param ... Ignored parameters.
#'
#'
#' @param modVars returns the model variables instead of the named
#'     vector of translated properties.
#'
#'
#'
#' @return a named vector of translated model properties
#'       including what type of jacobian is specified, the `C` function prefixes,
#'       as well as the `C` functions names to be called through the compiled model.
#' @seealso [rxode2()], [rxCompile()].
#' @author Matthew L.Fidler
#' @export
rxTrans <- function(model,
                    modelPrefix = "", # Model Prefix
                    md5 = "", # Md5 of model
                    modName = NULL, # Model name for DLL
                    modVars = FALSE, # Return modVars
                    ...) {
  UseMethod("rxTrans")
} # end function rxTrans


#' @rdname rxTrans
#' @export
rxTrans.default <- function(model,
                            modelPrefix = "", # Model Prefix
                            md5 = "", # Md5 of model
                            modName = NULL, # Model name for DLL
                            modVars = FALSE, # Return modVars
                            ...) {
  .mv <- rxode2::rxModelVars(model)
  if (modVars) {
    return(.mv)
  } else {
    return(c(.mv$trans, .mv$md5))
  }
}

.rxMECode <- ""

## Event ("jump") sensitivities: current mode folded into the parsed md5 (cache
## key) so that the same model text built in different modes compiles to distinct
## DLLs.  Empty for the default "fd" path -> md5 unchanged (existing caches stay
## valid).  Set by rxode2() before parsing; see .rxEventSensMode().
.rxEventSensCacheKey <- ""

#' @rdname rxTrans
#' @export
rxTrans.character <- memoise::memoise(function(model,
                                               modelPrefix = "", # Model Prefix
                                               md5 = "", # Md5 of model
                                               modName = NULL, # Model name for DLL
                                               modVars = FALSE, # Return modVars
                                               eventSensKey = .rxEventSensCacheKey, # nolint
                                               ...) {
  ## `eventSensKey` MUST be a formal (defaulting to the session global) rather
  ## than read from `.rxEventSensCacheKey` inside the body: memoise keys on the
  ## arguments (including default-valued ones), so a body-read global would make
  ## "fd" and "jump" builds of the same model text share one memoise entry -- the
  ## first builder's mode-folded parsed_md5 would then be reused for the other
  ## mode and compiled with the wrong (empty/stale) dLag/dF.  As a formal it is
  ## part of the key, so each mode gets its own entry and its own parsed_md5.
  ## rxTrans returns a list of compiled properties
  if (file.exists(model)) {
    .isStr <- 0L
  } else {
    .isStr <- 1L
  }
  if (missing(md5)) {
    md5 <- rxMd5(model)$digest
  }
  .ret <- .Call(
    `_rxode2_trans`, model, modelPrefix, md5, .isStr,
    as.integer(crayon::has_color()),
    .rxMECode, .rxSupportedFuns(),
    .rxFullPrint
  )
  if (inherits(.ret, "try-error")) {
    message("model")
    if (.isStr == 0L) {
      message(suppressWarnings(readLines(model)))
    } else {
      message(model)
    }
    stop("cannot create rxode2 model", call. = FALSE)
  }
  .parsedMd5 <- rxMd5(c(
    .ret$model,
    .ret$ini,
    .ret$state,
    .ret$params,
    .ret$lhs
  ))$digest
  ## Fold the event-sensitivity mode into the parsed md5 so that the same
  ## normalized model built as "fd" vs "jump" compiles to distinct DLLs.  This is
  ## required for correctness independent of the codegen channel: the compiled
  ## DLL is keyed by parsed_md5, and a cache hit SKIPS codegen entirely -- so the
  ## mode-dependent dLag/dF bodies (which are NOT part of the normalized model
  ## text and so do not affect the base md5) must be reflected here, or a prior fd
  ## build (empty dLag/dF) is silently reused for a jump solve.  The fold must be
  ## a real re-digest: `rxMd5()` hashes only the `normModel`/`indLin` slots of its
  ## argument and ignores any extra vector elements, so appending the key to its
  ## input is a no-op.  "" (the "fd" default) leaves the md5 unchanged, preserving
  ## existing fd caches.
  if (nzchar(eventSensKey)) {
    .parsedMd5 <- digest::digest(c(.parsedMd5, paste0("eventSens=", eventSensKey)),
                                 serialize = TRUE, algo = "md5")
  }
  md5 <- c(file_md5 = md5, parsed_md5 = .parsedMd5)
  .ret$timeId <- .rxTimeId(md5["parsed_md5"])
  .ret$md5 <- md5
  if (.isStr == 1L) {
    ## Now update trans.
    .prefix <- paste0("rx_", md5["parsed_md5"], "_", .Platform$r_arch, "_")
    .libName <- substr(.prefix, 0, nchar(.prefix) - 1)
    .ret <- .Call(`_rxode2_rxUpdateTrans_`, .ret, .prefix, .libName)
  }
  ## dparser::dpReload();
  ## rxReload()
  if (modVars) {
    return(.ret)
  } else {
    return(c(.ret$trans, .ret$md5))
  }
})

#' @rdname rxIsLoaded
#' @export
rxDllLoaded <- rxIsLoaded
#' Compile a model if needed
#'
#' This is the compilation workhorse creating the rxode2 model DLL
#' files.
#'
#' @inheritParams rxode2
#'
#' @param dir This is the model directory where the C file will be
#'     stored for compiling.
#'
#'     If unspecified, the C code is stored in a temporary directory,
#'     then the model is compiled and moved to the current directory.
#'     Afterwards the C code is removed.
#'
#'     If specified, the C code is stored in the specified directory
#'     and then compiled in that directory.  The C code is not removed
#'     after the DLL is created in the same directory.  This can be
#'     useful to debug the c-code outputs.
#'
#' @param prefix is a string indicating the prefix to use in the C
#'     based functions.  If missing, it is calculated based on file
#'     name, or md5 of parsed model.
#'
#'
#' @param force is a boolean stating if the (re)compile should be
#'     forced if rxode2 detects that the models are the same as already
#'     generated.
#'
#' @param eventSensCode character vector of length 13 giving the C body
#'     lines for the event ("jump") sensitivity helper functions
#'     `dLag`/`dF`/`dRate`/`dDur`/`d2F`/`d2Lag`/`d2Rate`/`d2Dur`/`d3F`/`dFQ`/
#'     `dLagJac`/`dLagQ`/`dDurQ` (in that order).  These are the per-model
#'     dosing-parameter total-derivative assignments produced by the
#'     event-sensitivity code generator and are emitted verbatim into the
#'     corresponding generated functions.  The default of thirteen empty
#'     strings produces trivial (no-op) helpers, which is what
#'     non-sensitivity models and `eventSens = "fd"` models use.
#'
#' @param ... Other arguments sent to the [rxTrans()]
#'     function.
#'
#' @return An rxDll object that has the following components
#'
#' * `dll` DLL path
#' * `model` model specification
#' * `.c` A function to call C code in the correct context from the DLL
#'          using the [.C()] function.
#' * `.call` A function to call C code in the correct context from the DLL
#'          using the [.Call()] function.
#' * `args` A list of the arguments used to create the rxDll object.
#' @inheritParams rxode2
#' @seealso [rxode2()]
#' @author Matthew L.Fidler
#' @importFrom sys exec_internal
#' @export
rxCompile <- function(model, dir, prefix, force = FALSE, modName = NULL,
                      package = NULL,
                      ...) {
  UseMethod("rxCompile")
}

.getIncludeDir <- function() {
  .cache <- R_user_dir("rxode2", "cache")
  .parseInclude <- system.file("include", package = "rxode2")
  if (dir.exists(.cache)) {
    .include <- .normalizePath(file.path(.cache, "include"))
    if (!dir.exists(.include)) {
      .malert("creating rxode2 include directory")
      dir.create(.include, recursive = TRUE, showWarnings = FALSE)
      .sysInclude <- system.file("include", package = "rxode2")
      .files <- list.files(.parseInclude)
      lapply(.files, function(file) {
        file.copy(file.path(.parseInclude, file), file.path(.include, file))
      })
      .files <- list.files(.sysInclude)
      lapply(.files, function(file) {
        file.copy(file.path(.sysInclude, file), file.path(.include, file))
      })
      .malert("getting R compile options")
      .cc <- rawToChar(sys::exec_internal(file.path(R.home("bin"), "R"), c("CMD", "config", "CC"))$stdout)
      .cc <- gsub("\n", "", .cc)
      .cflags <- rawToChar(sys::exec_internal(file.path(R.home("bin"), "R"), c("CMD", "config", "CFLAGS"))$stdout)
      .cflags <- gsub("\n", "", .cflags)
      .cflags <- paste0(.cflags, " -O", getOption("rxode2.compile.O", "3"))
      .shlibCflags <- rawToChar(sys::exec_internal(file.path(R.home("bin"), "R"), c("CMD", "config", "SHLIB_CFLAGS"))$stdout)
      .shlibCflags <- gsub("\n", "", .shlibCflags)
      .cpicflags <- rawToChar(sys::exec_internal(file.path(R.home("bin"), "R"), c("CMD", "config", "CPICFLAGS"))$stdout)
      .cpicflags <- gsub("\n", "", .cpicflags)

      .malert("precompiling headers")
      .args <- paste0(
        .cc, " -I", gsub("[\\]", "/", .normalizePath(R.home("include"))), " ",
        " -I\"", .normalizePath(.parseInclude), "\" ",
        .cflags, " ", .shlibCflags, " ", .cpicflags, " -I", gsub("[\\]", "/", .normalizePath(.include)), " ",
        paste(gsub("[\\]", "/", .normalizePath(.include)), "rxode2_model_shared.h", sep = "/"),
        ""
      )
      system(.args)
      .msuccess("done")
      return(.include)
    }
  }
  return(.normalizePath(system.file("include", package = "rxode2")))
}

.rxCompileEnv <- new.env(parent = emptyenv())
.rxCompileEnv$success <- TRUE
.rxCompileEnv$lst <- list()
.rxCompileEnv$cc <- NA_character_
#' Get the last compiled model information as alist
#'
#' @return A list contains the following elements:
#'
#' * `msg` the message for a bad compilation, or NULL if successful.
#'
#' * `stdout` the standard output from the compilation
#'
#' * `stderr` the standard error from the compilation
#'
#' * `c` the code code that was used
#'
#' This list will be returned invisibly, but the function will
#' also message the contents to the console.
#'
#' @export
#' @author Matthew L. Fidler
#' @examples
#' \donttest{
#' rxode2({
#'   a <- b
#' })
#' rxLastCompile()
#' }
rxLastCompile <- function() {
  lapply(names(.rxCompileEnv$lst), function(nm) {
    cli::rule(left = nm)
    message(.rxCompileEnv$lst[[nm]])
  })
  return(invisible(.rxCompileEnv$lst))
}
#' Was the last compilation successful?
#'
#' @param set Logical flag. If missing, returns the current compilation
#'   status. If provided, sets the compilation status.
#'
#' @return this determines if the last compilation was successful.  This is useful for
#'   debugging and testing purposes.
#'
#' @export
#'
#' @author Matthew L. Fidler
#'
#' @keywords internal
#'
#' @examples
#'
#' .rxLastCompileSuccess()
#'
.rxLastCompileSuccess <- function(set) {
  if (missing(set)) {
    return(.rxCompileEnv$success)
  } else {
    .rxCompileEnv$success <- set
  }
  .rxCompileEnv$success
}
#' Get the number of loaded rxode2 DLLs
#'
#'
#' @return Number of loaded rxode2 DLLs
#' @export
#' @author Matthew L. Fidler
#' @examples
#' rxNumLoaded()
rxNumLoaded <- function() {
  .dlls <- getLoadedDLLs()
  length(grep(rex::rex(start, "rx_", n_times(any, 32), or("_x64", "_i386", "_", "")), names(.dlls)))
}
.pkg <- NULL
#' @rdname rxCompile
#' @export
rxCompile.rxModelVars <- function(model, # Model
                                  dir = NULL, # Directory
                                  prefix = NULL, # Prefix
                                  force = FALSE, # Force compile
                                  modName = NULL, # Model Name
                                  package = NULL,
                                  eventSensCode = rep("", 13L), # dLag/dF/dRate/dDur/d2F/d2Lag/d2Rate/d2Dur/d3F/dFQ/dLagJac/dLagQ/dDurQ body lines
                                  ...) {
  assignInMyNamespace(".pkg", package)
  ## rxCompile returns the DLL name that was created.
  model <- rxGetModel(model)

  if (is.null(prefix)) {
    prefix <- .rxPre(model, modName)
  }
  if (is.null(dir)) {
    if (getOption("rxode2.tempfiles", TRUE)) {
      .dir <- file.path(rxTempDir(), paste0(prefix, ".rxd"))
    } else {
      .dir <- getwd()
    }
  } else {
    .dir <- dir
    if (file.exists(.dir)) {
      if (!file.exists(file.path(.dir, paste0(rxode2.md5, ".md5")))) {
        .malert("remove old rxode2 dir {.file {.dir}}")
        unlink(.dir, recursive = TRUE, force = TRUE)
        .msuccess("done")
      }
    }
  }
  .dir <- suppressMessages(.normalizePath(.dir, mustWork = FALSE))
  if (!file.exists(.dir)) {
    dir.create(.dir, recursive = TRUE, showWarnings = FALSE)
    writeLines("rxode2", file.path(.dir, paste0(rxode2.md5, ".md5")))
  }

  .cFile <- file.path(.dir, sprintf("%s.c", substr(prefix, 0, nchar(prefix) - 1)))
  .cDllFile <- file.path(.dir, sprintf("%s%s", substr(prefix, 0, nchar(prefix) - 1), .Platform$dynlib.ext))
  .allModVars <- NULL
  .needCompile <- TRUE
  if (file.exists(.cDllFile)) {
    .modVars <- sprintf("%smodel_vars", prefix)
    if (!missing(prefix) && !missing(dir) &&
      regexpr(
        rex::rex(start, "rx_", n_times(any, 32), or("_x64", "_i386", "_", "")),
        prefix
      ) == -1 &&
        is.loaded(.modVars)) {
      try(dyn.unload(.cDllFile), silent=TRUE)
      unlink(.cFile)
      .tmp <- try(unlink(.cDllFile), silent=TRUE)
      if (inherits(.tmp, "try-error")) {
        if (file.exists(.cDllFile)) {
          stop("cannot seem to remove '", .cDllFile, "'",
               call.=FALSE)
        }
      }
    } else {
      if (!is.loaded(.modVars)) {
        try(dynLoad(.cDllFile), silent = TRUE)
      }
      if (is.loaded(.modVars)) {
        .allModVars <- eval(parse(text = sprintf(".Call(\"%s\")", .modVars)), envir = .GlobalEnv)
        .modVars <- .allModVars$md5
        if (!any(names(.modVars) == "file_md5")) {
          .needCompile <- FALSE
        } else {
          .needCompile <- FALSE
        }
      } else {
        .needCompile <- FALSE
      }
    }
  }
  if (force || .needCompile) {
    .lock <- paste0(.cFile, ".lock")
    if (file.exists(.lock)) {
      message("rxode2 already building model, waiting for lock file removal")
      message(sprintf("lock file: \"%s\"", .lock))
      while (file.exists(.lock)) {
        Sys.sleep(0.5)
        message(".", appendLF = FALSE)
      }
      message("")
      if (!(file.exists(.cDllFile))) {
        stop("error building model on another thread", call. = FALSE)
      }
    } else {
      sink(.lock)
      cat("\n")
      sink()
      on.exit(
        {
          unlink(.lock)
        },
        add = TRUE
      )
      .Makevars <- .normalizePath(file.path(.dir, "Makevars"))
      if (file.exists(.Makevars)) {
        .firstMake <- readLines(.Makevars, 1)
        if (length(.firstMake) == 0) {
          unlink(.Makevars)
        } else if ("#rxode2 Makevars" == .firstMake) {
          unlink(.Makevars)
        } else {
          file.rename(.Makevars, paste0(.Makevars, ".bakrx"))
        }
      }
      .trans <- model
      if (file.exists(.cDllFile)) {
        if (inherits(.modVars, "list")) {
          if (.modVars["parsed_md5"] == .trans["parsed_md5"]) {
            message("do not need to recompile, minimal change to model detected")
            .needCompile <- FALSE
          }
        }
      }
      if (force || .needCompile) {
        ## Setup Makevars
        ## Now create C file
        .mv <- model
        .j <- 0
        .i <- 0
        .trans <- c(.mv$trans, .mv$md5)
        ## Load model into memory if needed
        if (.Call(`_rxode2_codeLoaded`) == 0L) .rxModelVarsCharacter(setNames(.mv$model, NULL))
        .prefix2 <- .rxModelVarsCCache[[3]]
        ## SEXP pMd5, SEXP timeId, SEXP fixInis
        .newMod <- FALSE
        if (!is.null(modName)) {
          .newMod <- regexpr("_new", modName) != -1
        }
        .rxModelVarsLast[[17]] <- .indLinInfo
        .model <- .rxModelVarsLast$model
        .model["indLin"] <- .rxMECode
        .rxModelVarsLast$model <- .model
        if (!is.null(package) && !.newMod) {
          .libname <- c(package, gsub(.Platform$dynlib.ext, "", basename(.cDllFile)))
          .Call(
            `_rxode2_codegen`, .cFile, prefix, .libname,
            .trans["parsed_md5"], paste(.rxTimeId(.trans["parsed_md5"])),
            .rxModelVarsLast, .rxSupportedFuns(),
            eventSensCode[1], eventSensCode[2], eventSensCode[3], eventSensCode[4],
            eventSensCode[5], eventSensCode[6], eventSensCode[7], eventSensCode[8],
            eventSensCode[9], eventSensCode[10], eventSensCode[11], eventSensCode[12],
            eventSensCode[13]
          )
        } else {
          .libname <- gsub(.Platform$dynlib.ext, "", basename(.cDllFile))
          .libname <- c(.libname, .libname)
          .Call(
            `_rxode2_codegen`, .cFile, prefix, .libname,
            .trans["parsed_md5"], paste(.rxTimeId(.trans["parsed_md5"])),
            .rxModelVarsLast, .rxSupportedFuns(),
            eventSensCode[1], eventSensCode[2], eventSensCode[3], eventSensCode[4],
            eventSensCode[5], eventSensCode[6], eventSensCode[7], eventSensCode[8],
            eventSensCode[9], eventSensCode[10], eventSensCode[11], eventSensCode[12],
            eventSensCode[13]
          )
        }
        .defs <- ""

        if (is.na(.rxCompileEnv$cc)) {
          .compilerPath <- tools::Rcmd("config CC", stdout=TRUE)
          .versionInfo <- try(system(paste(.compilerPath, "--version"), intern = TRUE))
          if (inherits(.versionInfo, "try-error")) {
            .rxCompileEnv$cc <- "unknown"
          } else if (any(grepl("clang", .versionInfo, ignore.case = TRUE))) {
            .rxCompileEnv$cc <- "clang"
          } else if (any(grepl("gcc", .versionInfo, ignore.case = TRUE))) {
            .rxCompileEnv$cc <- "gcc"
          } else {
            .rxCompileEnv$cc <- "unknown"
          }
        }
        if (.rxCompileEnv$cc == "clang") {
          .extra <- "-fno-math-errno -mtune=native -fno-sanitize=address "
        } else if (.rxCompileEnv$cc == "gcc") {
          .extra <- "-fno-math-errno -mtune=native "
        } else {
          .extra <- " "
        }
        .ret <- sprintf(
          "#rxode2 Makevars\nPKG_CFLAGS=-O%s %s %s -I\"%s\" -I\"%s\"\nPKG_LIBS=$(BLAS_LIBS) $(LAPACK_LIBS) $(FLIBS)\n",
          getOption("rxode2.compile.O", "3"),
          .extra,
          .defs, .getIncludeDir(),
          system.file("include", package = "rxode2")
        )
        ## .ret <- paste(.ret, "-g")
        sink(.Makevars)
        cat(.ret)
        sink()
        sink(.normalizePath(file.path(.dir, "extraC.h")))
        cat(rxode2::.extraCnow())
        sink()
        try(dyn.unload(.cDllFile), silent = TRUE)
        try(unlink(.cDllFile))
        .cmd <- file.path(R.home("bin"), "R")
        .args <- c("CMD", "SHLIB", basename(.cFile))
        .rxBinpref <- Sys.getenv("rxBINPREF")
        if (.rxBinpref != "") {
          .oldBinpref <- Sys.getenv("BINPREF")
          Sys.setenv("BINPREF" = .rxBinpref)
          on.exit(Sys.setenv("BINPREF" = .oldBinpref), add = TRUE)
        }
        rxode2::rxReq("sys")
        .rxWithWd(.dir, {
          .out <- sys::exec_internal(cmd = .cmd, args = .args, error = FALSE)
        })
        .stderr <- rawToChar(.out$stderr)
        .rxCompileEnv$lst <- list()
        if (!(all(.stderr == "") && length(.stderr) == 1)) {
          .rxCompileEnv$lst[["stderr"]] <- paste(.stderr, sep = "\n")
        }
        .rxCompileEnv$success <- TRUE
        .badBuild <- function(msg, cSrc = TRUE) {
          .rxCompileEnv$lst[["msg"]] <- gettext(msg)
          .rxCompileEnv$lst[["stdout"]] <- rawToChar(.out$stdout)
          if (cSrc) {
            .rxCompileEnv$lst[["c"]] <- paste(readLines(.cFile), collapse = "\n")
          } else {
            dyn.load(.cDllFile)
          }
          message("Error building the model: see rxode2::rxLastCompile()")
          .rxCompileEnv$success <- FALSE
          if (.Platform$OS.type == "windows") {
            message("this could be because your Rtools is not set up correctly")
          } else {
            message("please make sure you have a working C compiler set up")
          }
          message("you may use nlmixr2::nlmixr2CheckInstall() to help diagnose installation issues")
          stop(msg, call. = FALSE)
        }
        if (!(.out$status == 0 && file.exists(.cDllFile))) {
          .badBuild("error building model")
        }
      }
    }
    .tmp <- try(dynLoad(.cDllFile), silent = FALSE)
    if (inherits(.tmp, "try-error")) {
      ## Try unloading rxode2 dlls now...
      rxUnloadAll()
      .tmp <- try(dynLoad(.cDllFile), silent = TRUE)
      if (inherits(.tmp, "try-error")) {
        .badBuild("Error loading model (though dll exists)", cSrc = FALSE)
      } else {
        warning("unloaded all rxode2 dlls before loading the current DLL", call. = FALSE)
      }
    }
    .modVars <- sprintf("%smodel_vars", prefix)
    if (is.loaded(.modVars)) {
      .allModVars <- eval(parse(text = sprintf(".Call(\"%s\")", .modVars)), envir = .GlobalEnv)
    } else {
      .badBuild("Error, model doesn't have correct model variables.")
    }
  }
  ## removeSource() so this generated closure does not carry a srcref back to
  ## the (pkgload) srcfile of rxode2.R -- under load_all that srcfile holds a
  ## lazy `lines` promise capturing the whole session, which would balloon any
  ## saveRDS() of a model/fit (see the fit-fixture size regression).
  .call <- removeSource(function(...) {
    return(.Call(...))
  })
  .args <- list(
    model = model, dir = .dir, prefix = prefix,
    force = force, modName = modName,
    ...
  )
  if (is.null(.allModVars)) {
    stop("something went wrong in compilation")
  }
  assign(.cDllFile, 0L, envir = .rxModels) ## Loaded model.
  ret <- suppressWarnings({
    list(
      dll = .cDllFile,
      c = .cFile,
      model = .allModVars$model["normModel"],
      modVars = .allModVars,
      .call = .call,
      args = .args
    )
  })
  class(ret) <- "rxDll"
  return(ret)
}

#' @rdname rxCompile
#' @export
rxCompile.character <- rxCompile.rxModelVars

#' @rdname rxCompile
#' @export
rxCompile.rxDll <- function(model, ...) {
  .args <- as.list(match.call(expand.dots = TRUE))
  .rxDllArgs <- model$args
  if (any(names(.rxDllArgs) == "dir")) {
    .args$dir <- .rxDllArgs$dir
  }
  if (any(names(.rxDllArgs) == "prefix")) {
    .args$prefix <- .rxDllArgs$prefix
  }
  if (any(names(.rxDllArgs) == "force")) {
    .args$force <- .rxDllArgs$force
  }
  if (any(names(.rxDllArgs) == "modName")) {
    .args$modName <- .rxDllArgs$modName
  }
  .args$model <- .rxDllArgs$model
  return(do.call(getFromNamespace("rxCompile", "rxode2"), .args, envir = parent.frame(1)))
}

#' @rdname rxCompile
#' @export
rxCompile.rxode2 <- function(model, ...) {
  model$compile()
}

#' @rdname rxDynLoad
#' @export
rxLoad <- rxDynLoad

#' @rdname rxDynUnload
#' @export
rxUnload <- rxDynUnload

.rxConditionLst <- list()
#' Current Condition for rxode2 object
#'
#' @param obj rxode2 object
#' @param condition If specified and is one of the conditions in the
#'     rxode2 object (as determined by [rxExpandIfElse()]),
#'     assign the rxode2 current condition to this parameter.  If the
#'     condition is not one of the known condition, the condition is
#'     set to `NULL`, implying no conditioning currently used.
#' @return Current condition for rxode2 object
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
rxCondition <- function(obj, condition = NULL) {
  .key <- digest::digest(rxode2::rxNorm(obj, FALSE), algo = "md5", serialize = TRUE)
  if (!missing(condition) && is.null(condition)) {
    condition <- FALSE
  }
  if (is.null(condition)) {
    return(getFromNamespace(".rxConditionLst", "rxode2")[[.key]])
  } else if (any(condition == rxNorm(obj, TRUE))) {
    .lst <- getFromNamespace(".rxConditionLst", "rxode2")
    .lst[[.key]] <- condition
    assignInMyNamespace(".rxConditionLst", .lst)
    return(getFromNamespace(".rxConditionLst", "rxode2")[[.key]])
  } else {
    .lst <- getFromNamespace(".rxConditionLst", "rxode2")
    .lst[[.key]] <- NULL
    assignInMyNamespace(".rxConditionLst", .lst)
    return(getFromNamespace(".rxConditionLst", "rxode2")[[.key]])
  }
}

#' Get the normalized model
#'
#'
#' This get the syntax preferred model for processing
#'
#' @inheritParams rxModelVars
#' @param condition Character string of a logical condition to use
#'   for subsetting the normalized model.  When missing, and a
#'   condition is not set via `rxCondition`, return the whole
#'   code with all the conditional settings intact.  When a condition
#'   is set with `rxCondition`, use that condition.
#' @param removeInis A boolean indicating if parameter initialization
#'   will be removed from the model
#' @param removeJac A boolean indicating if the Jacobians will be
#'   removed.
#' @param removeSens A boolean indicating if the sensitivities will
#'   be removed.
#' @return Normalized Normal syntax (no comments)
#' @author Matthew L. Fidler
#' @export
rxNorm <- function(obj, condition = NULL, removeInis, removeJac, removeSens) {
  if (!missing(removeInis) || !missing(removeJac) || !missing(removeSens)) {
    .ret <- strsplit(rxNorm(obj, condition), "\n")[[1]]
    if (missing(removeInis)) {
      removeInis <- FALSE
    }
    if (missing(removeJac)) {
      removeJac <- FALSE
    }
    if (missing(removeSens)) {
      removeSens <- FALSE
    }
    if (removeInis) {
      .ret <- .rxRmIni(.ret)
    }
    if (removeJac) {
      stop("'removeJac' is no longer supported")
      ## .ret <- .rxRmJac(.ret)
    }
    if (removeSens) {
      stop("'removeSens' is no longer supported")
      ## .ret <- .rxRmSens(.ret)
    }
    return(paste(.ret, collapse = "\n"))
  } else {
    if (is(condition, "logical")) {
      if (!condition) {
        condition <- NULL
      } else {
        .tmp <- rxode2::rxExpandIfElse(obj)
        return(names(.tmp))
      }
    } else if (is.null(condition)) {
      condition <- rxode2::rxCondition(obj)
    }
    if (is.null(condition)) {
      .tmp <- rxode2::rxModelVars(obj)$model["normModel"]
      names(.tmp) <- NULL
      return(.tmp)
    } else {
      if (is(condition, "character")) {
        .tmp <- rxode2::rxExpandIfElse(obj)[condition]
        names(.tmp) <- NULL
        return(.tmp)
      } else {
        return(rxNorm(obj, FALSE))
      }
    }
  }
}



.rxModelVarsCCache <- NULL
.rxModelVarsLast <- NULL
.rxModelVarsCharacter <- function(obj) {
  .oldEventSensKey <- .rxEventSensCacheKey
  on.exit(assignInMyNamespace(".rxEventSensCacheKey", .oldEventSensKey), add = TRUE)
  if (length(obj) == 1) {
    .parseModel <- tempfile("parseModel4")
    .prefix <- paste0(basename(.parseModel), "_", .Platform$r_arch, "_")
    .exists <- try(file.exists(obj), silent = TRUE)
    if (inherits(.exists, "try-error")) {
      .exists <- FALSE
    } else {
      .exists <- TRUE
    }
    if (.exists) {
      .parseModel <- obj
    } else {
      .parseModel <- paste(obj, collapse = "\n")
    }
    .ret <- rxTrans(.parseModel, modelPrefix = .prefix, modVars = TRUE)
    .cFile <- list(.exists, ifelse(.exists, obj, ""), .prefix)
    assignInMyNamespace(".rxModelVarsCCache", .cFile)
    assignInMyNamespace(".rxModelVarsLast", .ret)
    return(.ret)
  } else {
    .rxModelVarsCharacter(paste(obj, collapse = "\n"))
  }
}

#' @rdname rxInits
#' @export
rxInit <- rxInits
#' Reload rxode2 DLL
#'
#' Can be useful for debugging
#'
#' @author Matthew L. Fidler
#' @keywords internal
#' @return boolean of if the object is reloaded
#' @export
rxReload <- function() {
  .tmp <- getLoadedDLLs()$rxode2
  class(.tmp) <- "list"
  dyn.unload(.tmp$path)
  .ret <- is.null(getLoadedDLLs()$rxode2)
  dynLoad(.tmp$path)
  .ret <- .ret && !is.null(getLoadedDLLs()$rxode2)
  return(.ret)
}

.rxModels <- new.env(parent = emptyenv())
#' Get the rxModels  information
#' @param env boolean that returns the environment where models are stored (TRUE), or the currently assigned rxode2 model variables (FALSE).
#' @keywords internal
#' @return internal rxModels information environment
#' @export
rxModels_ <- # nolint
  function(env = TRUE) {
    if (env) {
      return(getFromNamespace(".rxModels", "rxode2"))
    } else {
      return(.Call(rxode2_get_mv, PACKAGE = "rxode2"))
    }
  }

#' All model variables for a rxode2 object
#'
#' Return all the known model variables for a specified rxode2 object
#'
#' These items are only calculated after compilation; they are
#' built-into the rxode2 compiled DLL.
#'
#' To allow extension, an s3 hook is added in the function `rxModelVarsS3`.
#'
#' @param obj rxode2 family of objects
#'
#' @return A list of rxode2 model properties including:
#'
#' * `params`  a character vector of names of the model parameters
#' * `lhs` a character vector of the names of the model calculated parameters
#' * `state` a character vector of the compartments in rxode2 object
#' * `trans` a named vector of translated model properties
#'       including what type of jacobian is specified, the `C` function prefixes,
#'       as well as the `C` functions names to be called through the compiled model.
#' * `md5` a named vector that gives the digest of the model (`file_md5`) and the parsed model
#'      (`parsed_md5`)
#' * `model`  a named vector giving the input model (`model`),
#'    normalized model (no comments and standard syntax for parsing, `normModel`),
#'    and interim code that is used to generate the final C file `parseModel`
#'
#' @keywords internal
#' @family Query model information
#' @author Matthew L. Fidler
#' @export
rxModelVars <- function(obj) {
  if (is(substitute(obj), "{")) {
    .obj <- deparse(substitute(obj))
    if (.obj[1] == "{") {
      .obj <- .obj[-1]
      .obj <- .obj[-length(.obj)]
    }
    .obj <- paste(.obj, collapse = "\n")
    return(rxModelVars_(.obj))
  }
  if ((is.list(obj) &&
         inherits(obj, "rxUi")) ||
        inherits(obj, "raw")) {
    obj <- rxUiDecompress(obj)
  }
  if (is(obj, "rxModelVars")) {
    return(obj)
  }
  .tmp <- try(obj, silent = TRUE)
  if (inherits(.tmp, "try-error")) {
    obj <- as.character(substitute(obj))
  }
  rxModelVars_(obj)
}

#' @rdname rxModelVars
#' @export
rxModelVarsS3 <- function(obj) {
  UseMethod("rxModelVarsS3")
}

#' @rdname rxModelVars
#' @export
rxModelVarsS3.rxUi <- function(obj) {
  .ret <- rxUiDecompress(obj)
  get("mv0", .ret)
}


#' @rdname rxModelVars
#' @export
rxModelVarsS3.default <- function(obj) {
  stop("need an rxode2-type object to extract model variables",
       call.=FALSE)
}

.rxGetParseModel <- function(type = c("normal", "dt"),
                             collapse = TRUE) {
  .type.idx <- c("normal" = 0L, "dt" = 1L)
  if (is(type, "character")) {
    type <- .type.idx[match.arg(type)]
  }
  .ret <- .Call(`_rxode2_parseModel`, type)
  if (collapse) {
    .ret <- paste(.ret, collapse = "")
  }
  return(.ret)
}

.rxGetModelInfoFromDll <- function(dll) {
  .base <- basename(dll)
  if (nchar(.base) >= 36) {
    if (substr(.base, 36, 36) == "_") {
      .md5 <- substring(.base, 4, 35)
      return(c(.md5, paste0("rx_", .md5, "_", .Platform$r_arch, "_")))
    }
  }
  .extra <- nchar(.Platform$r_arch) + 1 + nchar(.Platform$dynlib.ext)
  .mod <- substring(.base, 0, nchar(.base) - .extra)
  return(c(.mod, paste0(.mod, "_", .Platform$r_arch, "_")))
}

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    # Logical expressions test

    transTo <- function(model, syntax, match = TRUE) {
      mod <- rxModelVars(model)

      if (match) {
        test_that(sprintf("%s includes %s", model, syntax), {
          expect_true(regexpr(syntax, .rxGetParseModel(), fixed = TRUE) != -1)
        })
      } else {
        test_that(sprintf("%s dose not include %s", model, syntax), {
          expect_false(regexpr(syntax, .rxGetParseModel(), fixed = TRUE) != -1)
        })
      }
    }

    transTo("x=1;if (t != 0 & t != 1){x=0}", "&&")
    transTo("x=1;if ((t == 0) | (t == 1)){x=0}", "||")
    transTo("x=1;if ((t == 0) & !(t == 1)){x=0}", "&&")
    transTo("x=1;if ((t == 0) & !(t == 1)){x=0}", "!(")

    test_that("string comparisons track covariate string parameters", {
      p <- rxode2parse(paste(
        "x=1",
        "if (cov2 == \"str3\" || cov == \"str1\" || \"str4\" != cov2 || cov != \"str1\" || \"str2\" == cov || id == \"skip\" || \"skip2\" != ID){x=0}",
        sep = ";"
      ))

      expect_equal(p$strCmpParams,
                   list(cov2 = factor(c("str3", "str4"),
                                      levels = c("str3", "str4")),
                         cov = factor(c("str1", "str2"),
                                      levels = c("str1", "str2"))))
    })

    test_that("string comparisons use integer helpers for non-ID covariates", {
      rxode2parse(paste(
        "x=1",
        "if (cov == \"alpha\" || \"beta\" != cov || id == \"skip\" || \"skip2\" != ID){x=0}",
        sep = ";"
      ))
      parsed <- .rxGetParseModel()

      expect_true(grepl('_cmp2d(cov, "cov", 1, 1)', parsed, fixed = TRUE))
      expect_true(grepl('_cmp1d("beta", 0, cov, 2)', parsed, fixed = TRUE))
      expect_true(grepl('_cmp2((&_solveData->subjects[_cSub])->idReal, "ID", 1, "skip")',
                        parsed, fixed = TRUE))
      expect_true(grepl('_cmp1("skip2", 0, (&_solveData->subjects[_cSub])->idReal, "ID")',
                        parsed, fixed = TRUE))
    })
  }
})

rxTest({
  # Test that model specification can take string, file name or rxode expressions
  ode <- rxode2(model = "d/dt(y) = r * y * (1 - y/K);")

  ode2 <- rxode2({
    d / dt(y) <- r * y * (1 - y / K)
  })

  test_that("string and expression returns the same result", {
    expect_equal(rxModelVars(ode)$md5["parsed_md5"], rxModelVars(ode2)$md5["parsed_md5"])
  })

  tmp <- tempfile(fileext = ".rx")

  sink(tmp)
  cat("d/dt(y) = r * y * (1 - y/K);\n")
  sink()
  while (!file.exists(tmp)) {
    Sys.sleep(1)
  }
  ode3 <- rxode2(tmp)
  unlink(tmp)

  test_that("file and string returns the same result", {
    expect_equal(rxModelVars(ode)$md5["parsed_md5"], rxModelVars(ode3)$md5["parsed_md5"])
    expect_equal(rxNorm(ode), rxNorm(ode3))
  })


  ode <- rxode2("
         C2 = centr/V2;
         C3 = peri/V3;
         d/dt(depot) =-KA*depot;
         d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
         d/dt(peri)  =                    Q*C2 - Q*C3;
         d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;")

  ode2 <- rxode2({
    C2 <- centr / V2
    C3 <- peri / V3
    d / dt(depot) <- -KA * depot
    d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
    d / dt(peri) <- Q * C2 - Q * C3
    d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
  })

  tmp <- tempfile(fileext = ".rx")
  sink(tmp)
  cat("
         C2 = centr/V2;
         C3 = peri/V3;
         d/dt(depot) =-KA*depot;
         d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
         d/dt(peri)  =                    Q*C2 - Q*C3;
         d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;")
  sink()
  while (!file.exists(tmp)) {
    Sys.sleep(1)
  }
  ode3 <- rxode2(tmp)
  unlink(tmp)

  test_that("string and expression returns the same result", {
    expect_equal(rxModelVars(ode)$md5["parsed_md5"], rxModelVars(ode2)$md5["parsed_md5"])
  })
  test_that("file and string returns the same result", {
    expect_equal(rxModelVars(ode)$md5["parsed_md5"], rxModelVars(ode3)$md5["parsed_md5"])
  })

  .pmd5 <- function(m, ...) unname(rxModelVars(rxode2(m, ...))$md5["parsed_md5"])

  test_that("default linCmtSens shares a parsed_md5 with explicit linCmtA", {
    .m <- "cl <- 1\nv <- 20\ncp <- linCmt()"
    ## default is c(\"linCmtA\", \"linCmtB\"), which match.arg() resolves to
    ## \"linCmtA\" -- identical compiled code, so they must share a cache key
    expect_equal(.pmd5(.m), .pmd5(.m, linCmtSens = "linCmtA"))
    ## a genuinely different linCmtSens must NOT collide
    expect_false(.pmd5(.m, linCmtSens = "linCmtA") == .pmd5(.m, linCmtSens = "linCmtB"))
  })

  test_that("parsed_md5 does not depend on what was built before it", {
    ## a prior non-default linCmtSens build must not shift a later default hash
    .plain <- "d/dt(y) = r * y * (1 - y/K);"
    .h1 <- .pmd5(.plain)
    invisible(rxode2("cl <- 1\nv <- 20\ncp <- linCmt()", linCmtSens = "linCmtB"))
    expect_equal(.pmd5(.plain), .h1)
    ## a prior indLin build must not leak into a later plain model hash
    .p2 <- "d/dt(x) = -k * x"
    .h2 <- .pmd5(.p2)
    invisible(rxode2("d/dt(depot) = -ka*depot\nd/dt(central) = ka*depot - cl/v*central",
                     indLin = TRUE))
    expect_equal(.pmd5(.p2), .h2)
  })
})

rxTest({
  td <- tempdir()

  .rxWithWd(td, {
    test_that("Issue RxODE#178: deactivate active compiled model", {
      ## Define model
      skip_on_os("solaris")
      skip_on_os("mac")
      ode <- "
   C2 = centr/V2;
   C3 = peri/V3;
   d/dt(depot) = -KA*depot;
   d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
   d/dt(peri) = Q*C2 - Q*C3;
   d/dt(eff) = Kin*(1-C2/(EC50+C2)) - Kout*eff;
"
      
      mod1 <- rxode2(model = ode, modName = "mod1")
      
      ode <- "
   C2 = centr/V2;
   d/dt(depot) = -KA*depot;
   d/dt(centr) = KA*depot - CL*C2;
   d/dt(eff) = Kin*(1-C2/(EC50+C2)) - Kout*eff;
"
      
      ## Compile model
      mod1 <- rxode2(model = ode, modName = "mod1")
      
      expect_equal(rxModelVars(ode)$params, rxModelVars(mod1)$params)
    })
  })
})

rxTest({ 
  test_that("param() is not dropped in rxNorm model; Issue RxODE#431", {
    mod <- "param(parent_0,log_k_parent,log_k_m1,f_parent_qlogis,CMT);\ncmt(parent);\ncmt(m1);\nparent_0_model~parent_0;\nparent(0)=parent_0_model;\nk_parent~exp(log_k_parent);\nk_m1~exp(log_k_m1);\nf_parent_to_m1~expit(f_parent_qlogis);\nd/dt(parent)=-k_parent*parent;\nd/dt(m1)=+f_parent_to_m1*k_parent*parent-k_m1*m1;\nif (CMT==1){\nnlmixr_pred=parent;\n}\nif (CMT==2){\nnlmixr_pred=m1;\n}\n"
    expect_equal(rxNorm(rxode2(mod)), mod)
  })
})

rxTest({
  test_that("interpolation functions", {

    tmp <- rxModelVars("locf(a);\n ret=a+b")

    expect_equal(rxNorm(tmp), "locf(a);\nret=a+b;\n")

    expect_equal(as.character(tmp$interp["a"]), "locf")
    expect_equal(as.character(tmp$interp["b"]), "default")

    tmp <- rxModelVars("params(b, a);\nlocf(a);\n ret=a+b")

    expect_equal(as.character(tmp$interp["a"]), "locf")
    expect_equal(as.character(tmp$interp["b"]), "default")

    expect_error(rxModelVars("params(b, a);\nlocf(a);\nnocb(a);\n ret=a+b"))

  })

  test_that("line number increments correctly after interpolation error", {
    tmp <- normalizePath(tempfile(), mustWork = FALSE)
    on.exit(unlink(tmp))
    .rxWithSinkBoth(tmp, {
      expect_error(rxModelVars("params(b, a);\nlocf(a);\nnocb(a);\n ret=a+b"))
    })
    lines <- paste(readLines(tmp), collapse = "\n")
    expect_true(grepl(":004:", lines))
  })

  test_that("interpolation functions (continued)", {
    tmp <- rxModelVars("params(b, a);\nlinear(a);\n ret=a+b")

    expect_equal(as.character(tmp$interp["a"]), "linear")
    expect_equal(as.character(tmp$interp["b"]), "default")

    tmp <- rxModelVars("params(b, a);\nnocb(a);\n ret=a+b")
    expect_equal(as.character(tmp$interp["a"]), "nocb")
    expect_equal(as.character(tmp$interp["b"]), "default")

    tmp <- rxModelVars("params(b, a);\nmidpoint(a);\n ret=a+b")
    expect_equal(as.character(tmp$interp["a"]), "midpoint")
    expect_equal(as.character(tmp$interp["b"]), "default")


  })

  test_that("ui $interpLines", {

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        cl.wt <- 0
        v.wt <- 0
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        linear(WT)
        locf(b)
        nocb(c)
        midpoint(d)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)+ WT ^ 2* cl.wt
        v <- exp(tv + eta.v+ WT * v.wt + b + c + d)
        linCmt() ~ add(add.sd)
      })
    }

    ui <- rxode(f)

    expect_equal(ui$interpLines,
                 list(str2lang("linear(WT)"),
                      str2lang("locf(b)"),
                      str2lang("nocb(c)"),
                      str2lang("midpoint(d)")))

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        cl.wt <- 0
        v.wt <- 0
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        linear(WT)
        locf(b)
        midpoint(d)
        nocb(c)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)+ WT ^ 2* cl.wt
        v <- exp(tv + eta.v+ WT * v.wt + b + c + d)
        linCmt() ~ add(add.sd)
      })
    }

    ui <- rxode(f)

    expect_equal(ui$interpLines,
                 list(str2lang("linear(WT)"),
                      str2lang("locf(b)"),
                      str2lang("nocb(c)"),
                      str2lang("midpoint(d)")))

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        cl.wt <- 0
        v.wt <- 0
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        locf(WT, b, d, c)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)+ WT ^ 2* cl.wt
        v <- exp(tv + eta.v+ WT * v.wt + b + c + d)
        linCmt() ~ add(add.sd)
      })
    }

    ui <- rxode(f)

    expect_equal(ui$interpLines,
                 list(str2lang("locf(WT, b, d, c)")))

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        cl.wt <- 0
        v.wt <- 0
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)+ WT ^ 2* cl.wt
        v <- exp(tv + eta.v+ WT * v.wt + b + c + d)
        linCmt() ~ add(add.sd)
      })
    }

    ui <- rxode(f)

    expect_null(ui$interpLines)

  })


  test_that("interp $simulationModel", {

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        cl.wt <- 0
        v.wt <- 0
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        linear(WT)
        locf(b)
        nocb(c)
        midpoint(d)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)+ WT ^ 2* cl.wt
        v <- exp(tv + eta.v+ WT * v.wt + b + c + d)
        linCmt() ~ add(add.sd)
      })
    }

    ui <- rxode(f)

    expect_error(ui$simulationModel, NA)

    mod <- ui$simulationModel

    expect_true(rxModelVars(mod)$interp["WT"] == "linear")
    expect_true(rxModelVars(mod)$interp["b"] == "locf")
    expect_true(rxModelVars(mod)$interp["c"] == "nocb")
    expect_true(rxModelVars(mod)$interp["d"] == "midpoint")

    expect_error(ui$simulationIniModel, NA)

    mod <- ui$simulationIniModel

    expect_true(rxModelVars(mod)$interp["WT"] == "linear")
    expect_true(rxModelVars(mod)$interp["b"] == "locf")
    expect_true(rxModelVars(mod)$interp["c"] == "nocb")
    expect_true(rxModelVars(mod)$interp["d"] == "midpoint")

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        cl.wt <- 0
        v.wt <- 0
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)+ WT ^ 2* cl.wt
        v <- exp(tv + eta.v+ WT * v.wt + b + c + d)
        linCmt() ~ add(add.sd)
      })
    }

    ui <- rxode(f)


    expect_error(ui$simulationModel, NA)

    mod <- ui$simulationModel

    expect_true(rxModelVars(mod)$interp["WT"] == "default")
    expect_true(rxModelVars(mod)$interp["b"] == "default")
    expect_true(rxModelVars(mod)$interp["c"] == "default")
    expect_true(rxModelVars(mod)$interp["d"] == "default")

    expect_error(ui$simulationIniModel, NA)

    mod <- ui$simulationIniModel

    expect_true(rxModelVars(mod)$interp["WT"] == "default")
    expect_true(rxModelVars(mod)$interp["b"] == "default")
    expect_true(rxModelVars(mod)$interp["c"] == "default")
    expect_true(rxModelVars(mod)$interp["d"] == "default")


  })


  test_that("time varying character/factors should not be interpolated by linear solving", {

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        tviov.cl <- c(0, 0.1)
        iov.cl1 ~ fix(1)
        iov.cl2 ~ fix(1)
        add.sd <- 0.7
      })
      model({
        iov.cl <- sqrt(tviov.cl) * ((OCC=="first") * iov.cl1 +
                                      (OCC=="second") * iov.cl2)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }


    et <- et(amt=100) |>
      et(0:24) |>
      as.data.frame()

    et$OCC  <- "first"
    et$OCC[et$time > 12] <- "second"

    f <- suppressWarnings(f()$simulationIniModel)
    expect_warning(rxSolve(f, et, covsInterpolation="linear"))
    expect_warning(rxSolve(f, et, covsInterpolation="nocb"), NA)
    expect_warning(rxSolve(f, et, covsInterpolation="locf"), NA)
    expect_warning(rxSolve(f, et, covsInterpolation="midpoint"))

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        tviov.cl <- c(0, 0.1)
        iov.cl1 ~ fix(1)
        iov.cl2 ~ fix(1)
        add.sd <- 0.7
      })
      model({
        midpoint(OCC)
        iov.cl <- sqrt(tviov.cl) * ((OCC=="first") * iov.cl1 +
                                      (OCC=="second") * iov.cl2)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    f <- suppressWarnings(f()$simulationIniModel)

    expect_warning(rxSolve(f, et))

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        tviov.cl <- c(0, 0.1)
        iov.cl1 ~ fix(1)
        iov.cl2 ~ fix(1)
        add.sd <- 0.7
      })
      model({
        linear(OCC)
        iov.cl <- sqrt(tviov.cl) * ((OCC=="first") * iov.cl1 +
                                      (OCC=="second") * iov.cl2)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    f <- suppressWarnings(f()$simulationIniModel)

    expect_warning(rxSolve(f, et))

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        tviov.cl <- c(0, 0.1)
        iov.cl1 ~ fix(1)
        iov.cl2 ~ fix(1)
        add.sd <- 0.7
      })
      model({
        nocb(OCC)
        iov.cl <- sqrt(tviov.cl) * ((OCC=="first") * iov.cl1 +
                                      (OCC=="second") * iov.cl2)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    f <- suppressWarnings(f()$simulationIniModel)

    expect_warning(rxSolve(f, et), NA)

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        tviov.cl <- c(0, 0.1)
        iov.cl1 ~ fix(1)
        iov.cl2 ~ fix(1)
        add.sd <- 0.7
      })
      model({
        locf(OCC)
        iov.cl <- sqrt(tviov.cl) * ((OCC=="first") * iov.cl1 +
                                      (OCC=="second") * iov.cl2)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    f <- suppressWarnings(f()$simulationIniModel)

    expect_warning(rxSolve(f, et), NA)

  })
})

rxTest({
  test_that("param order", {
    mod <- rxode2({
      a <- 6
      b <- 0.6
      cmt(blood) # cmt = 1 now
      d / dt(intestine) <- -a * intestine
      d / dt(blood) <- a * intestine - b * blood
    })

    expect_equal(rxModelVars(mod)$param, c("a", "b"))
  })

  test_that("param order rev", {
    mod2 <- rxode2({
      param(b, a)
      a <- 6
      b <- 0.6
      cmt(blood) # cmt = 1 now
      d / dt(intestine) <- -a * intestine
      d / dt(blood) <- a * intestine - b * blood
    })

    expect_equal(rxModelVars(mod2)$param, c("b", "a"))
  })

  test_that("large params()", {
    expect_error(tmp <- rxode2("param(tktr,tka,tcl,tv,poplogit,tec50,tkout,te0)
cmt(depot)
cmt(gut)
cmt(center)
cmt(effect)
effect(0)=exp(te0)
rx_expr_3~exp(tktr)
d/dt(depot)=-rx_expr_3*depot
rx_expr_1~exp(tka)
d/dt(gut)=-rx_expr_1*gut+rx_expr_3*depot
d/dt(center)=rx_expr_1*gut-exp(tcl-tv)*center
rx_expr_2~exp(-tv)
rx_expr_4~rx_expr_2*center
d/dt(effect)=-exp(tkout)*effect+exp(te0+tkout)*(1-exp(poplogit-tv)*center/((1+exp(poplogit))*(rx_expr_4+exp(tec50))))
rx_expr_0~CMT==6
rx_pred_=effect*(rx_expr_0)+rx_expr_4*(CMT==5)*(1-(rx_expr_0))
cmt(cp)
cmt(pca)
dvid(5, 6)"), NA)

    expect_equal(rxModelVars(tmp)$param, c(
      "tktr", "tka", "tcl", "tv", "poplogit", "tec50", "tkout", "te0",
      "CMT"
    ))
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    transTo <- function(model, syntax, match = TRUE) {
      mv <- rxModelVars(model)
      if (match) {
        test_that(sprintf("%s includes %s", model, syntax), {
          expect_true(regexpr(syntax, .rxGetParseModel(), fixed = TRUE) != -1)
        })
      } else {
        test_that(sprintf("%s dose not include %s", model, syntax), {
          expect_false(regexpr(syntax, .rxGetParseModel(), fixed = TRUE) != -1)
        })
      }
    }

    transTo("d/dt(m)= factorial(1+fac)-cmt*ka", "factorial(1+fac)")
    transTo("d/dt(m)= lgamma(1+fac)-cmt*ka", "lgamma(1+fac)")
    transTo("d/dt(m)= gamma(1+fac)-cmt*ka", "lgammafn(1+fac)")
    transTo("d/dt(m)= lfactorial(1+fac)-cmt*ka", "lgamma1p(1+fac)")
  }
})

rxTest({
  .expectStrCmpParams <- function(obj, expected) {
    expect_equal(names(obj$strCmpParams), names(expected))
    expect_equal(obj$strCmpParams, expected)
    for (.nm in names(expected)) {
      expect_s3_class(obj$strCmpParams[[.nm]], "factor")
      expect_equal(levels(obj$strCmpParams[[.nm]]), levels(expected[[.nm]]))
      expect_equal(as.character(obj$strCmpParams[[.nm]]),
                   as.character(expected[[.nm]]))
    }
  }

  test_that("modelvars", {
    skip_on_cran()
    rxClean()
    rigid.txt <- "
y1(0)    = 1
y2(0)    = 0
y3(0)    = 0.9
a1       = -2
a2       = 1.25
a3       = -0.5
d/dt(y1) = a1*y2*y3
d/dt(y2) = a2*y1*y3
d/dt(y3) = a3*y1*y2
"
    rigid0 <- rxGetModel(rigid.txt)

    rigid <- rxode2(rigid.txt)

    et <- eventTable()
    et$add.sampling(seq(0, 20, by = 0.01))


    out <- solve(rigid, et)

    expect_equal(rxModelVars(rigid), rxModelVars(rigid$cmpMgr$rxDll()))
    expect_equal(rxModelVars(rigid), rxModelVars(out))
    # Drop the 'timeId' element because it is the compilation time, and it may
    # differ if the system is heavily loaded
    rigid0_compare <- rigid0[setdiff(names(rigid0), "timeId")]
    rigid_compare <- rxModelVars(rigid)
    rigid_compare <- rigid_compare[setdiff(names(rigid_compare), "timeId")]
    expect_equal(rigid0_compare, rigid_compare)
  })

  test_that("modelvars track exhaustive string comparison covariates", {
    skip_on_cran()
    rxClean()

    modTxt <- "
x = 1
if (covB == \"same\" ||
    covA == \"alpha\" ||
    \"beta\" == covA ||
    covA != \"alpha\" ||
    \"same\" != covB ||
    covC == \"shared\" ||
    \"shared\" == covA ||
    covB == \"same\" ||
    ID == \"drop\" ||
    \"drop2\" != id ||
    covC != \"shared\" ||
    covB == \"gamma\" ||
    \"alpha\" == covB ||
    covA == \"beta\" ||
    covC == \"omega\") {
  x = 0
}
"

    expected <- list(
      covB = factor(c("same", "gamma", "alpha"),
                    levels = c("same", "gamma", "alpha")),
      covA = factor(c("alpha", "beta", "shared"),
                    levels = c("alpha", "beta", "shared")),
      covC = factor(c("shared", "omega"),
                    levels = c("shared", "omega"))
    )

    parsed <- rxGetModel(modTxt)
    compiled <- rxode2(modTxt)

    .expectStrCmpParams(parsed, expected)
    .expectStrCmpParams(rxModelVars(compiled), expected)

    parsedCompare <- parsed[setdiff(names(parsed), "timeId")]
    compiledCompare <- rxModelVars(compiled)
    compiledCompare <- compiledCompare[setdiff(names(compiledCompare), "timeId")]
    expect_equal(parsedCompare, compiledCompare)
  })

  test_that("blank modelvars has same size", {
    blank <- rxModelVars("")
    full <-  rxModelVars("y = 1")
    expect_equal(length(rxModelVars(blank)), length(full))
    expect_length(blank$strCmpParams, 0)
    expect_equal(names(blank$strCmpParams), character(0))
    expect_equal(tail(names(full), 5),
                 c("lhsOrd", "splitBolus", "strCmpParams", "timeId", "md5"))
  })
})

rxTest({
  test_that("NONMEM-like matrix exponential syntax errors are caught", {
    # 1. matExp() with ODEs throws error
    expect_error(rxode2({
      matExp()
      d/dt(depot) = -k_depot_central * depot
      d/dt(central) = k_depot_central * depot - k_central_output * central
    }))

    # 2. indLin() without matExp() throws error
    expect_error(rxode2({
      cmt(depot)
      cmt(central)
      indLin(central) <- 1.5
      k_depot_central = 0.5
    }))

    # 3. self-transfer k_cmt1_cmt1 throws error
    expect_error(rxode2({
      matExp()
      cmt(depot)
      k_depot_depot = 0.1
    }))

    expect_error(rxode2({
      matExp()
      cmt(depot)
      k.depot.depot = 0.1
    }))
  })

  test_that("NONMEM-like matrix exponential compiles and solves correctly", {
    # Test a simple 2-compartment model
    mod1 <- rxode2({
      matExp()
      k_depot_central = 0.5
      k_central_output = 0.2
      # CP calculation using compartment amount
      cp = central / 10.0
    })

    # Verify that depot and central are state variables
    expect_true(any(mod1$state == "depot"))
    expect_true(any(mod1$state == "central"))

    # Solve it with some doses
    et1 <- eventTable() %>%
      add.dosing(dose = 100, nbr.doses = 1, start.time = 0) %>%
      add.sampling(seq(0, 10, by = 1))

    res1 <- rxSolve(mod1, et1, method = "indLin")

    # Solve the analytical/ode version to compare
    modOde <- rxode2({
      d/dt(depot) = -0.5 * depot
      d/dt(central) = 0.5 * depot - 0.2 * central
      cp = central / 10.0
    })
    resOde <- rxSolve(modOde, et1)

    expect_equal(res1$depot, resOde$depot, tolerance = 1e-4)
    expect_equal(res1$central, resOde$central, tolerance = 1e-4)
    expect_equal(res1$cp, resOde$cp, tolerance = 1e-4)
  })

  test_that("NONMEM-like matrix exponential works with dot notation and indLin prop", {
    mod2 <- rxode2({
      matExp()
      cmt(depot)
      cmt(central)
      k.depot.central = 0.5
      k.central.output = 0.2
      indLin(central) <- 1.5
    })

    et2 <- eventTable() %>%
      add.dosing(dose = 100, nbr.doses = 1, start.time = 0) %>%
      add.sampling(seq(0, 10, by = 1))

    res2 <- rxSolve(mod2, et2, method = "indLin")

    # Solve the ODE version to compare (since indLin(central) <- 1.5 adds a constant infusion/rate of 1.5 to central)
    modOde <- rxode2({
      d/dt(depot) = -0.5 * depot
      d/dt(central) = 0.5 * depot - 0.2 * central + 1.5
    })
    resOde <- rxSolve(modOde, et2)

    expect_equal(res2$depot, resOde$depot, tolerance = 1e-4)
    expect_equal(res2$central, resOde$central, tolerance = 1e-4)
  })

  test_that("NONMEM-like matrix exponential works with infusions", {
    mod3 <- rxode2({
      matExp()
      k_depot_central = 0.5
      k_central_output = 0.2
    })

    et3 <- eventTable() %>%
      add.dosing(dose = 100, rate = 10, start.time = 0) %>% # 10 hour infusion
      add.sampling(seq(0, 15, by = 1))

    res3 <- rxSolve(mod3, et3, method = "indLin")

    modOde <- rxode2({
      d/dt(depot) = -0.5 * depot
      d/dt(central) = 0.5 * depot - 0.2 * central
    })
    resOde <- rxSolve(modOde, et3)

    expect_equal(res3$depot, resOde$depot, tolerance = 1e-4)
    expect_equal(res3$central, resOde$central, tolerance = 1e-4)
  })

  test_that("indLin, rxOdeToIndLin, and rxToIndLin transition ODEs correctly", {
    # 1. Simple 2-compartment model
    ode_code <- "
      d/dt(depot) = -0.5 * depot
      d/dt(central) = 0.5 * depot - 0.2 * central
      cp = central / 10.0
    "
    mexp_code <- indLin(ode_code)
    
    # Verify that the generated code contains the correct elements
    expect_true(any(grepl("matExp\\(\\)", mexp_code)))
    expect_true(any(grepl("cmt\\(depot\\)", mexp_code)))
    expect_true(any(grepl("cmt\\(central\\)", mexp_code)))
    expect_true(any(grepl("k_depot_central\\s*=\\s*0.5", mexp_code)))
    expect_true(any(grepl("k_central_output\\s*=\\s*0.2", mexp_code)))
    expect_true(any(grepl("cp\\s*=\\s*central\\s*/\\s*10(\\.0)?", mexp_code)))
    expect_false(any(grepl("d/dt\\(", mexp_code)))
    
    # Verify compilation and solve
    mod_mexp <- rxode2(mexp_code)
    et <- eventTable() %>%
      add.dosing(dose = 100, nbr.doses = 1, start.time = 0) %>%
      add.sampling(seq(0, 10, by = 1))
    res_mexp <- rxSolve(mod_mexp, et, method = "indLin")
    
    mod_ode <- rxode2(ode_code)
    res_ode <- rxSolve(mod_ode, et)
    expect_equal(res_mexp$depot, res_ode$depot, tolerance = 1e-4)
    expect_equal(res_mexp$central, res_ode$central, tolerance = 1e-4)
    expect_equal(res_mexp$cp, res_ode$cp, tolerance = 1e-4)

    # 2. Parameter names matching micro-constants (self-assignment avoidance)
    ode_code_params <- "
      d/dt(depot) = -k_depot_central * depot
      d/dt(central) = k_depot_central * depot - k_central_output * central
    "
    mexp_code_params <- rxToIndLin(ode_code_params)
    expect_true(any(grepl("param\\(k_depot_central\\)", mexp_code_params)))
    expect_true(any(grepl("param\\(k_central_output\\)", mexp_code_params)))
    expect_false(any(grepl("k_depot_central\\s*=\\s*k_depot_central", mexp_code_params)))
    
    mod_mexp_params <- rxode2(mexp_code_params)
    res_mexp_params <- rxSolve(mod_mexp_params, et, method = "indLin",
                               params = c(k_depot_central = 0.5, k_central_output = 0.2))
    expect_equal(res_mexp_params$depot, res_ode$depot, tolerance = 1e-4)
    expect_equal(res_mexp_params$central, res_ode$central, tolerance = 1e-4)

    # 3. Forcing functions / constant inputs (indLin property)
    ode_code_forcing <- "
      d/dt(depot) = -0.5 * depot
      d/dt(central) = 0.5 * depot - 0.2 * central + 1.5
    "
    mexp_code_forcing <- rxOdeToIndLin(ode_code_forcing)
    expect_true(any(grepl("indLin\\(central\\)\\s*<-\\s*1\\.5", mexp_code_forcing)))
    
    mod_mexp_forcing <- rxode2(mexp_code_forcing)
    res_mexp_forcing <- rxSolve(mod_mexp_forcing, et, method = "indLin")
    
    mod_ode_forcing <- rxode2(ode_code_forcing)
    res_ode_forcing <- rxSolve(mod_ode_forcing, et)
    expect_equal(res_mexp_forcing$depot, res_ode_forcing$depot, tolerance = 1e-4)
    expect_equal(res_mexp_forcing$central, res_ode_forcing$central, tolerance = 1e-4)
  })

  test_that("NONMEM-like matrix exponential compartment naming restrictions", {
    # Invalid names containing '_' or '.' (without 'rx' prefix) must throw syntax error
    expect_error(rxode2({
      matExp()
      cmt(dep_ot)
      k_dep_ot_central = 0.5
    }))
    expect_error(rxode2({
      matExp()
      cmt(depot)
      cmt(cent.ral)
      k_depot_cent.ral = 0.5
    }))
    
    # Valid names (normal names or starting with 'rx' with '_' or '.') must succeed
    expect_error(rxode2({
      matExp()
      cmt(depot)
      cmt(central)
      cmt(rx__sens_depot_BY_ka__)
      k_depot_central = 0.5
      k_depot_rx__sens_depot_BY_ka___nd = -1.0
    }), NA)
  })

  test_that("NONMEM-like matrix exponential forward sensitivities solve correctly", {
    # 1. 2-compartment model
    ode_code <- "
      d/dt(depot) = -ka * depot
      d/dt(central) = ka * depot - cl/v * central
    "
    
    mexp_sens_code <- rxSensMatExp(ode_code, calcSens = c("ka", "cl"))
    expect_true(any(grepl("matExp\\(\\)", mexp_sens_code)))
    expect_true(any(grepl("cmt\\(rx__sens_depot_BY_ka__\\)", mexp_sens_code)))
    expect_true(any(grepl("k_depot_rx__sens_depot_BY_ka___nd\\s*=", mexp_sens_code)))
    
    mod_mexp <- rxode2(mexp_sens_code)
    mod_ode <- rxode2(ode_code, calcSens = c("ka", "cl"))
    
    et <- eventTable() |>
      add.dosing(dose = 100, nbr.doses = 1, start.time = 0) |>
      add.sampling(seq(0, 10, by = 1))
      
    res_mexp <- rxSolve(mod_mexp, et, method = "indLin", params = c(ka = 0.5, cl = 0.2, v = 10))
    res_ode <- rxSolve(mod_ode, et, params = c(ka = 0.5, cl = 0.2, v = 10))
    
    expect_equal(res_mexp$central, res_ode$central, tolerance = 1e-4)
    expect_equal(res_mexp$rx__sens_depot_BY_ka__, res_ode$rx__sens_depot_BY_ka__, tolerance = 1e-4)
    expect_equal(res_mexp$rx__sens_central_BY_ka__, res_ode$rx__sens_central_BY_ka__, tolerance = 1e-4)
    expect_equal(res_mexp$rx__sens_central_BY_cl__, res_ode$rx__sens_central_BY_cl__, tolerance = 1e-4)
    
    # 2. Non-linear Michaelis-Menten elimination model
    ode_code_mm <- "
      d/dt(depot) = -ka * depot
      d/dt(central) = ka * depot - Vm * central / (Km + central)
    "
    mexp_sens_mm <- rxSensMatExp(ode_code_mm, calcSens = c("ka", "Vm", "Km"))
    mod_mexp_mm <- rxode2(mexp_sens_mm)
    res_mexp_mm <- rxSolve(mod_mexp_mm, et, method = "indLin", params = c(ka = 0.5, Vm = 10, Km = 5))
    
    # Verify the sensitivity columns are present and contain numeric values
    expect_true(all(c("rx__sens_depot_BY_ka__", "rx__sens_central_BY_ka__",
                      "rx__sens_central_BY_Vm__", "rx__sens_central_BY_Km__") %in% colnames(res_mexp_mm)))
    expect_true(is.numeric(res_mexp_mm$rx__sens_central_BY_ka__))
    expect_true(is.numeric(res_mexp_mm$rx__sens_central_BY_Vm__))
    expect_true(is.numeric(res_mexp_mm$rx__sens_central_BY_Km__))
  })

  test_that("matExp symengine env can build jacobians and sensitivities", {
    mod <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        matExp()
        k_depot_central = exp(tka + eta.ka)
        k_central_output = exp(tcl + eta.cl) / exp(tv + eta.v)
        cp = central / exp(tv + eta.v)
        cp ~ add(add.sd)
      })
    }

    s <- rxS(rxode2(mod), doConst = TRUE, promoteLinSens = TRUE)
    expect_no_error(.rxJacobian(s, c("ETA_1_", "ETA_2_", "ETA_3_")))
    expect_no_error(.rxSens(s, c("ETA_1_", "ETA_2_", "ETA_3_")))
    expect_true(exists("..jacobian", envir = s, inherits = FALSE))
    expect_true(exists("..sens", envir = s, inherits = FALSE))
  })

  test_that("matrix-exp sensitivities obey reset and multiply event jumps", {
    ode_code <- "
      d/dt(depot) = -ka * depot
      d/dt(central) = ka * depot - cl / v * central
    "
    mod <- rxode2(rxSensMatExp(ode_code, calcSens = c("ka", "cl")))
    mod_ode <- rxode2(ode_code, calcSens = c("ka", "cl"), eventSens = "jump")
    expect_equal(mod$eventSens, "jump")
    expect_false(is.null(mod$eventSensInfo))

    pars <- c(ka = 0.5, cl = 0.2, v = 10)
    et_base <- et(amt = 100, cmt = "depot") |>
      et(seq(0, 10, 1))
    et_reset <- et(amt = 100, cmt = "depot") |>
      et(time = 5, evid = 3) |>
      et(seq(0, 10, 1))
    et_mult <- et(amt = 100, cmt = "depot") |>
      et(time = 5, amt = 0.5, cmt = "central", evid = 6) |>
      et(seq(0, 10, 1))

    res_base <- rxSolve(mod, et_base, params = pars, method = "indLin")
    res_reset <- rxSolve(mod, et_reset, params = pars, method = "indLin")
    res_mult <- rxSolve(mod, et_mult, params = pars, method = "indLin")
    res_ode_reset <- rxSolve(mod_ode, et_reset, params = pars)
    res_ode_mult <- rxSolve(mod_ode, et_mult, params = pars)

    expect_equal(res_reset$central, res_ode_reset$central, tolerance = 1e-5)
    expect_equal(res_reset$rx__sens_central_BY_ka__, res_ode_reset$rx__sens_central_BY_ka__,
                 tolerance = 1e-5)
    expect_equal(res_reset$rx__sens_central_BY_cl__, res_ode_reset$rx__sens_central_BY_cl__,
                 tolerance = 1e-5)
    expect_equal(res_mult$central, res_ode_mult$central, tolerance = 1e-5)
    expect_equal(res_mult$rx__sens_central_BY_ka__, res_ode_mult$rx__sens_central_BY_ka__,
                 tolerance = 1e-5)
    expect_equal(res_mult$rx__sens_central_BY_cl__, res_ode_mult$rx__sens_central_BY_cl__,
                 tolerance = 1e-5)
  })

  test_that("matExp additive-bolus F and lag jumps match the generic ODE path", {
    # Regression: rxSensMatExp()'s additive-bolus dtau (lag-time) jump row was
    # a silent no-op for matExp models -- matExp()/indLin() models have no
    # functional dydt() (matrix-exponential primal solve, not RHS evaluation),
    # so handle_evid's usual central-difference-of-dydt Jacobian column was
    # always exactly zero for them. Fixed by having rxSensMatExp() emit
    # explicit df(x)/dy(y) lines from its already-known Jacobian (legal in a
    # matExp() model even though d/dt() lines are not) and having handle_evid
    # read the column from calc_jac instead for these models
    # (`_rxEsUseCalcJac`, set from `.rxEventSensUseCalcJac()`/`mv$indLin`).
    ode_code <- "
      alag(depot) <- exp(tlag)
      f(depot)    <- expit(tf)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    mexp <- rxSensMatExp(ode_code, calcSens = c("tlag", "tf"))
    expect_true(any(grepl("df\\(depot\\)/dy\\(depot\\)", mexp)))
    expect_true(any(grepl("df\\(central\\)/dy\\(depot\\)", mexp)))
    expect_true(any(grepl("df\\(central\\)/dy\\(central\\)", mexp)))

    mod_mexp <- rxode2(mexp)
    mod_ode <- rxode2(ode_code, calcSens = c("tlag", "tf"), eventSens = "jump")

    pars <- c(ka = 0.5, cl = 0.2, v = 10, tlag = log(0.5), tf = qlogis(0.7))
    et1 <- et(amt = 100, cmt = "depot") |> et(seq(0, 10, by = 0.5))

    res_mexp <- rxSolve(mod_mexp, et1, params = pars, method = "indLin")
    res_ode <- rxSolve(mod_ode, et1, params = pars)

    for (cc in c(
      "central", "rx__sens_central_BY_tlag__", "rx__sens_central_BY_tf__"
    )) {
      expect_equal(res_mexp[[cc]], res_ode[[cc]], tolerance = 1e-4)
    }
  })

  test_that("matExp infusion jumps (modeled rate) match the generic ODE path", {
    ode_code <- "
      rate(depot) <- exp(trate)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    mexp <- rxSensMatExp(ode_code, calcSens = c("trate"))
    mod_mexp <- rxode2(mexp)
    mod_ode <- rxode2(ode_code, calcSens = c("trate"), eventSens = "jump")

    pars <- c(ka = 0.5, cl = 0.2, v = 10, trate = log(20))
    et1 <- et(amt = 100, cmt = "depot", rate = -1) |> et(seq(0, 10, by = 0.5))

    res_mexp <- rxSolve(mod_mexp, et1, params = pars, method = "indLin")
    res_ode <- rxSolve(mod_ode, et1, params = pars)

    expect_equal(res_mexp$central, res_ode$central, tolerance = 1e-4)
    expect_equal(res_mexp$rx__sens_central_BY_trate__,
      res_ode$rx__sens_central_BY_trate__,
      tolerance = 1e-4
    )
  })

  test_that("2nd-order (Hessian) jumps fire correctly for matExp compartments", {
    # Confirms the currently-implemented 2nd-order jump pieces (additive-bolus
    # F, replace, multiply -- see the event-sensitivities plan Phase F; the
    # dtau/infusion 2nd-order rows are not implemented for ANY model type
    # yet, so they are not exercised here) work identically for matExp's
    # standard-layout 2nd-order compartments as for the equivalent ODE model,
    # during an actual eventSens="jump" solve with parameter-dependent
    # dosing (not just plain ODE-sensitivity math with constant doses, which
    # is all the calcSens2/calcSens3 addition itself validated).
    ode_code_f <- "
      alag(depot) <- exp(tlag)
      f(depot)    <- expit(tf)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars_f <- c(ka = 0.5, cl = 0.2, v = 10, tlag = log(0.5), tf = qlogis(0.7))
    et_f <- et(amt = 100, cmt = "depot") |> et(seq(0, 10, by = 0.5))
    mexp_f <- rxode2(rxSensMatExp(ode_code_f, calcSens = c("tlag", "tf"), calcSens2 = c("tlag", "tf")))
    ode_f <- rxode2(ode_code_f, calcSens = c("tlag", "tf"), calcSens2 = c("tlag", "tf"), eventSens = "jump")
    res_mexp_f <- rxSolve(mexp_f, et_f, params = pars_f, method = "indLin")
    res_ode_f <- rxSolve(ode_f, et_f, params = pars_f)
    expect_equal(res_mexp_f$rx__sens_central_BY_tf_BY_tf__,
      res_ode_f$rx__sens_central_BY_tf_BY_tf__,
      tolerance = 1e-4
    )

    ode_code_rm <- "
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars_rm <- c(ka = 0.5, cl = 0.2, v = 10)
    mexp_rm <- rxode2(rxSensMatExp(ode_code_rm, calcSens = "ka", calcSens2 = "ka"))
    ode_rm <- rxode2(ode_code_rm, calcSens = "ka", calcSens2 = "ka", eventSens = "jump")
    for (.evid in c(5, 6)) {
      .amt <- if (.evid == 5) 50 else 0.5
      e <- et(amt = 100, cmt = "depot") |>
        et(time = 5, amt = .amt, cmt = "central", evid = .evid) |>
        et(seq(0.5, 12, 1))
      res_mexp <- rxSolve(mexp_rm, e, params = pars_rm, method = "indLin")
      res_ode <- rxSolve(ode_rm, e, params = pars_rm)
      expect_equal(res_mexp$rx__sens_central_BY_ka_BY_ka__,
        res_ode$rx__sens_central_BY_ka_BY_ka__,
        tolerance = 1e-4, info = paste0("evid: ", .evid)
      )
    }
  })

  test_that("2nd-order additive-bolus F jump uses matExp's OWN compartment layout for a proper-subset calcSens2", {
    # The 2nd-order ddelta (F) jump write used the ODE (rxExpandSens2_)
    # compartment-index formula UNCONDITIONALLY until 2026-07-01: harmless
    # when calcSens2==calcSens (d2F is symmetric in (p,q), so the ODE and
    # matExp layouts are just a transpose of each other and every value
    # lands on a slot expecting that same value by coincidence -- see the
    # "2nd-order (Hessian) jumps" test above, which only ever used
    # calcSens2==calcSens and so could never catch this), but a REAL bug
    # once calcSens2 is a PROPER SUBSET with fewer elements than calcSens
    # (some calcSens-only index no longer has a valid "transposed" slot in
    # the smaller calcSens2 range). Isolated to a pure-F (no alag) model to
    # avoid conflating with the separately-scoped alag/dtau-row gap for
    # parameters outside calcSens2 (see the plan/memory notes on that).
    ode_code_f <- "
      f(depot) <- expit(tf)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars_f <- c(ka = 0.5, cl = 0.2, v = 10, tf = qlogis(0.7))
    et_f <- et(amt = 100, cmt = "depot") |> et(seq(0, 10, by = 0.5))
    calcSens <- c("tf", "ka", "cl")
    calcSens2 <- c("tf")
    mexp_f <- rxode2(rxSensMatExp(ode_code_f, calcSens = calcSens, calcSens2 = calcSens2))
    ode_f <- rxode2(ode_code_f, calcSens = calcSens, calcSens2 = calcSens2, eventSens = "jump")
    res_mexp_f <- rxSolve(mexp_f, et_f, params = pars_f, method = "indLin")
    res_ode_f <- rxSolve(ode_f, et_f, params = pars_f)
    for (cc in c(
      "rx__sens_central_BY_tf_BY_tf__", "rx__sens_central_BY_ka_BY_tf__",
      "rx__sens_central_BY_cl_BY_tf__"
    )) {
      expect_equal(res_mexp_f[[cc]], res_ode_f[[cc]], tolerance = 1e-4)
    }
  })

  test_that("3rd-order additive-bolus F jump fires correctly for matExp compartments (Phase H1)", {
    # Same confirmation as the 2nd-order test above, one order deeper: the
    # H1 additive-bolus F-row 3rd-order jump (the only 3rd-order jump piece
    # implemented so far) works identically for matExp's standard-layout
    # 3rd-order compartments as for the equivalent ODE model.
    ode_code_f <- "
      f(depot)    <- expit(tf)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars_f <- c(ka = 0.5, cl = 0.2, v = 10, tf = qlogis(0.7))
    et_f <- et(amt = 100, cmt = "depot") |> et(seq(0, 10, by = 0.5))
    mexp3 <- rxode2(rxSensMatExp(ode_code_f,
      calcSens = c("tf", "ka"), calcSens2 = c("tf", "ka"), calcSens3 = "tf"
    ))
    ode3 <- rxode2(ode_code_f,
      calcSens = c("tf", "ka"), calcSens2 = c("tf", "ka"), calcSens3 = "tf",
      eventSens = "jump"
    )
    res_mexp <- rxSolve(mexp3, et_f, params = pars_f, method = "indLin")
    res_ode <- rxSolve(ode3, et_f, params = pars_f)
    expect_equal(res_mexp$rx__sens_central_BY_tf_BY_tf_BY_tf__,
      res_ode$rx__sens_central_BY_tf_BY_tf_BY_tf__,
      tolerance = 1e-4
    )
  })

  test_that("second-order dtau/lag row (Leibniz term) fires correctly for matExp compartments", {
    # Regression for a genuine compartment-ORDER mismatch found while
    # investigating why the dtau row's 2nd-order jump appeared to vanish for
    # matExp: rxSensMatExp() lists 2nd-order compartments p-outer/q-inner,
    # while the ODE path's rxExpandSens2_ layout is q-outer/p-inner. d2F/d3F
    # never exposed this because their (p,q) values are symmetric by
    # construction (differentiating F symbolically twice always commutes),
    # so writing into the "swapped" slot is invisible. The dtau row's
    # 3-term-only piece is NOT symmetric on its own; only after the Leibniz
    # term was added (making the full result symmetric again) did the
    # per-model-type compartment index actually matter for correctness. The
    # Leibniz term itself also needs a matExp-specific source for dS^p_k/dt
    # (dydt() is a no-op stub for matExp) -- computed from the constant
    # (time-invariant) physical Jacobian via calc_jac dotted with the
    # pre/post-jump sensitivity difference, exact for a linear system.
    ode_code <- "
      alag(depot) <- exp(tlag)
      f(depot)    <- expit(tf)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars <- c(ka = 0.5, cl = 0.2, v = 10, tlag = log(0.5), tf = qlogis(0.7))
    e <- et(amt = 100, cmt = "depot") |> et(seq(0.55, 10, by = 0.5))
    mexp2 <- rxode2(rxSensMatExp(ode_code, calcSens = c("tlag", "tf"), calcSens2 = c("tlag", "tf")))
    ode2 <- rxode2(ode_code, calcSens = c("tlag", "tf"), calcSens2 = c("tlag", "tf"),
                   eventSens = "jump")
    res_mexp <- rxSolve(mexp2, e, pars, method = "indLin", atol = 1e-12, rtol = 1e-12)
    res_ode <- rxSolve(ode2, e, pars, atol = 1e-12, rtol = 1e-12)
    expect_equal(res_mexp$rx__sens_central_BY_tlag_BY_tf__,
      res_ode$rx__sens_central_BY_tlag_BY_tf__,
      tolerance = 1e-8
    )
    expect_gt(max(abs(res_ode$rx__sens_central_BY_tlag_BY_tf__)), 1)
  })

  test_that("second-order infusion-boundary Leibniz term fires correctly for matExp compartments", {
    # Same compartment-order + matExp-specific-Jacobian-source fix as the
    # dtau row's own matExp regression above, applied to the infusion
    # moving-boundary-from-alag case: a fixed-rate infusion whose start/stop
    # time shifts with a modeled alag.
    ode_code <- "
      alag(central) <- exp(tlag)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars <- c(ka = 0.5, cl = 0.2, v = 10, tlag = log(0.5))
    e <- et(amt = 100, cmt = "central", rate = 20) |> et(seq(0.55, 15, by = 0.5))
    mexp2 <- rxode2(rxSensMatExp(ode_code, calcSens = "tlag", calcSens2 = "tlag"))
    ode2 <- rxode2(ode_code, calcSens = "tlag", calcSens2 = "tlag", eventSens = "jump")
    res_mexp <- rxSolve(mexp2, e, pars, method = "indLin", atol = 1e-12, rtol = 1e-12)
    res_ode <- rxSolve(ode2, e, pars, atol = 1e-12, rtol = 1e-12)
    expect_equal(res_mexp$rx__sens_central_BY_tlag_BY_tlag__,
      res_ode$rx__sens_central_BY_tlag_BY_tlag__,
      tolerance = 1e-8
    )
    expect_gt(max(abs(res_ode$rx__sens_central_BY_tlag_BY_tlag__)), 1)
  })

  test_that("second-order infusion-boundary Leibniz term propagates to a coupled matExp compartment", {
    # Same coupled-compartment regression as the dtau row above, for the
    # infusion-boundary Leibniz term: depot infused (alag/F/dur all
    # modeled), central observed.
    ode_code <- "
      alag(depot) <- tlag
      f(depot)    <- doseAmt
      dur(depot)  <- tinf
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars <- c(ka = 1, cl = 6, v = 60, tlag = 10, doseAmt = 200, tinf = 10)
    e <- et(amt = 1, cmt = "depot", rate = -2) |> et(seq(0.53, 60, by = 0.5))
    mexp2 <- rxode2(rxSensMatExp(ode_code,
      calcSens = c("tlag", "doseAmt", "tinf"), calcSens2 = c("tlag", "doseAmt", "tinf")
    ))
    ode2 <- rxode2(ode_code,
      calcSens = c("tlag", "doseAmt", "tinf"), calcSens2 = c("tlag", "doseAmt", "tinf"),
      eventSens = "jump"
    )
    res_mexp <- rxSolve(mexp2, e, pars, method = "indLin", atol = 1e-11, rtol = 1e-11)
    res_ode <- rxSolve(ode2, e, pars, atol = 1e-11, rtol = 1e-11)
    expect_equal(res_mexp$rx__sens_central_BY_tlag_BY_tlag__,
      res_ode$rx__sens_central_BY_tlag_BY_tlag__,
      tolerance = 1e-8
    )
    expect_gt(max(abs(res_ode$rx__sens_central_BY_tlag_BY_tlag__)), 1)
  })

  test_that("rxS() incorporates indLin() forcing (Michaelis-Menten) without error", {
    .mm <- paste("matExp()",
                 "cmt(depot)",
                 "cmt(central)",
                 "k_depot_central = exp(THETA[1])",
                 "indLin(central) <- -exp(THETA[2])*central/(exp(THETA[3])+central)",
                 "cp = central/exp(THETA[4])",
                 "rx_pred_~cp",
                 "rx_r_~1",
                 sep = "\n")
    # Regression: rxS() previously threw "attempt to use zero-length variable
    # name" on the indLin() left-hand side.
    .s <- rxode2::rxS(.mm, TRUE, promoteLinSens = FALSE)
    expect_true(is.environment(.s))
    # The materialized d/dt(central) Jacobian must include the Michaelis-Menten
    # term (forcing incorporated, not dropped) and a non-zero parameter
    # sensitivity for Vmax (THETA[2]).
    .jac <- rxode2::.rxJacobian(.s, c("depot", "central", "THETA_2_"))
    expect_true(any(grepl("df\\(central\\)/dy\\(central\\)", .jac) &
                      grepl("exp\\(THETA\\[3\\]\\)", .jac)))
    expect_true(any(grepl("df\\(central\\)/dy\\(THETA_2_\\)", .jac) &
                      grepl("central", .jac)))
  })

  test_that("rxSensMatExp() calcSens2/calcSens3 match the generic ODE path (linear)", {
    ode_code <- "
      d/dt(depot) = -ka * depot
      d/dt(central) = ka * depot - cl/v * central
    "
    et <- eventTable() |>
      add.dosing(dose = 100, nbr.doses = 1, start.time = 0) |>
      add.sampling(seq(0, 10, by = 1))
    pars <- c(ka = 0.5, cl = 0.2, v = 10)

    mexp2 <- rxSensMatExp(ode_code, calcSens = c("ka", "cl"), calcSens2 = c("ka", "cl"))
    expect_true(any(grepl("cmt\\(rx__sens_central_BY_ka_BY_cl__\\)", mexp2)))

    mod_mexp <- rxode2(mexp2)
    mod_ode <- rxode2(ode_code, calcSens = c("ka", "cl"), calcSens2 = c("ka", "cl"))
    res_mexp <- rxSolve(mod_mexp, et, method = "indLin", params = pars)
    res_ode <- rxSolve(mod_ode, et, params = pars)

    for (cc in c(
      "rx__sens_central_BY_ka_BY_ka__", "rx__sens_central_BY_ka_BY_cl__",
      "rx__sens_central_BY_cl_BY_ka__", "rx__sens_central_BY_cl_BY_cl__"
    )) {
      expect_equal(res_mexp[[cc]], res_ode[[cc]], tolerance = 1e-4)
    }

    # Third order: no generic-ODE calcSens3 path exists yet to compare against
    # (see the event-sensitivities plan), so validate against a finite
    # difference of the analytic *second*-order sensitivity (the same
    # cross-check strategy used for the second-order DDE/event-jump work).
    eps <- 1e-4
    p1 <- pars; p1["cl"] <- pars["cl"] + eps
    p2 <- pars; p2["cl"] <- pars["cl"] - eps
    r1 <- rxSolve(mod_mexp, et, method = "indLin", params = p1)
    r2 <- rxSolve(mod_mexp, et, method = "indLin", params = p2)
    fd3 <- (r1$rx__sens_central_BY_ka_BY_ka__ - r2$rx__sens_central_BY_ka_BY_ka__) / (2 * eps)

    mexp3 <- rxSensMatExp(ode_code,
      calcSens = c("ka", "cl"), calcSens2 = c("ka", "cl"),
      calcSens3 = c("ka", "cl")
    )
    expect_true(any(grepl("cmt\\(rx__sens_central_BY_ka_BY_ka_BY_cl__\\)", mexp3)))
    mod_mexp3 <- rxode2(mexp3)
    res_mexp3 <- rxSolve(mod_mexp3, et, method = "indLin", params = pars)
    expect_equal(fd3, res_mexp3$rx__sens_central_BY_ka_BY_ka_BY_cl__, tolerance = 1e-4)
  })

  test_that("rxSensMatExp() calcSens2/calcSens3 handle repeated parameters (Hessian diagonal)", {
    # calcSens2/calcSens3 reusing a calcSens parameter makes several distinct
    # cross-term families route to the same source compartment (e.g. the
    # "from S^p_k" and "from S^q_j" terms collapse when p == q); regression
    # test for the accumulator that sums (rather than duplicates/overwrites)
    # those contributions -- see .rxIndLinNdAccumulator().
    ode_code <- "
      d/dt(depot) = -ka * depot
      d/dt(central) = ka * depot - cl/v * central
    "
    mexp2 <- rxSensMatExp(ode_code, calcSens = c("ka", "cl"), calcSens2 = c("ka", "cl"))
    # exactly one definition per (from,to) pair -- no duplicate/conflicting
    # k_..._nd assignment for the ka-ka diagonal entry
    .lhs <- sub("\\s*=.*$", "", grep("^k[_.]", mexp2, value = TRUE))
    expect_equal(sum(duplicated(.lhs)), 0L)

    et <- eventTable() |>
      add.dosing(dose = 100, nbr.doses = 1, start.time = 0) |>
      add.sampling(seq(0, 10, by = 1))
    pars <- c(ka = 0.5, cl = 0.2, v = 10)
    mod_mexp <- rxode2(mexp2)
    mod_ode <- rxode2(ode_code, calcSens = c("ka", "cl"), calcSens2 = c("ka", "cl"))
    res_mexp <- rxSolve(mod_mexp, et, method = "indLin", params = pars)
    res_ode <- rxSolve(mod_ode, et, params = pars)
    expect_equal(res_mexp$rx__sens_central_BY_ka_BY_ka__,
      res_ode$rx__sens_central_BY_ka_BY_ka__,
      tolerance = 1e-4
    )

    # fully-degenerate third order (p == q == r)
    mexp3 <- rxSensMatExp(ode_code,
      calcSens = c("ka", "cl"), calcSens2 = c("ka", "cl"), calcSens3 = c("ka")
    )
    .lhs3 <- sub("\\s*=.*$", "", grep("^k[_.]", mexp3, value = TRUE))
    expect_equal(sum(duplicated(.lhs3)), 0L)
    mod_mexp3 <- rxode2(mexp3)
    res_mexp3 <- rxSolve(mod_mexp3, et, method = "indLin", params = pars)
    expect_true(is.numeric(res_mexp3$rx__sens_central_BY_ka_BY_ka_BY_ka__))
  })

  test_that("rxSensMatExp() calcSens2/calcSens3 validation errors", {
    ode_code <- "
      d/dt(depot) = -ka * depot
      d/dt(central) = ka * depot - cl/v * central
    "
    expect_error(rxSensMatExp(ode_code, calcSens = "ka", calcSens2 = "cl"))
    expect_error(rxSensMatExp(ode_code, calcSens = "ka", calcSens3 = "ka"))
    expect_error(rxSensMatExp(ode_code, calcSens = "ka", calcSens2 = "ka", calcSens3 = "cl"))
  })

  test_that("rxSensMatExp() excludes linCmt() states from calcSens2/calcSens3", {
    mod <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
      })
      model({
        matExp()
        k_depot_central <- exp(tka)
        k_central_output <- exp(tcl) / exp(tv)
        cp <- central / exp(tv)
        cp ~ add(0.1)
      })
    }
    .code <- rxSensMatExp(rxode2(mod), calcSens = "tka", calcSens2 = "tka")
    # a plain matExp() model has no linCmt() compartments, so nothing should
    # be excluded/changed -- this is the regression guard for the exclusion
    # logic itself (states are computed as setdiff(rxState(.env),
    # c("output", .rxLinCmt(.mv)))) not silently dropping ordinary states.
    expect_true(any(grepl("cmt\\(rx__sens_central_BY_tka_BY_tka__\\)", .code)))
    expect_true(any(grepl("cmt\\(rx__sens_depot_BY_tka_BY_tka__\\)", .code)))
  })

  test_that("inspecting rxSensMatExp() output without solving does not leak indLin state to the next model", {
    # Regression for a real crash (uncaught Rcpp::exception "unsupported
    # indLin code: 0", also observed as a C stack overflow depending on
    # platform/timing): `.indLinInfo` (R/rxode2.R) is a package-level
    # global used to carry matExp/indLin metadata into codegen. It was only
    # ever reset via `.clearME()`'s `on.exit()` on the FULL model-compile
    # closure -- a code path that calling `rxSensMatExp()` on a model and
    # merely inspecting the generated text (as above, never solving it)
    # does NOT reach. That left `.indLinInfo` set from this matExp model,
    # which then silently attached itself to the NEXT, completely
    # unrelated model built afterward -- corrupting its `mv$indLin` field
    # and making `rxSolve.default()` force-select `method="indLin"` on a
    # plain ODE model, which then crashed inside `indLin()` since the
    # model's own `doIndLin` was correctly 0. Fixed by resetting
    # `.indLinInfo` in `rxGetModel()` whenever the CURRENT model has no
    # indLin structure of its own.
    mod <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
      })
      model({
        matExp()
        k_depot_central <- exp(tka)
        k_central_output <- exp(tcl) / exp(tv)
        cp <- central / exp(tv)
        cp ~ add(0.1)
      })
    }
    .code <- rxSensMatExp(rxode2(mod), calcSens = "tka", calcSens2 = "tka")

    ode_code <- "
      d/dt(depot) = -ka * depot
      d/dt(central) = ka * depot - cl / v * central
    "
    plain <- rxode2(ode_code)
    expect_equal(length(rxModelVars(plain)$indLin), 0L)
    et <- eventTable() |> add.dosing(dose = 100, nbr.doses = 1, start.time = 0) |>
      add.sampling(seq(0, 10, by = 1))
    expect_no_error(
      rxSolve(plain, et, params = c(ka = 0.5, cl = 0.2, v = 10))
    )
  })
})

rxTest({
  ## test ODE parsing for syntax errors
  # Test Parsing of models

  badParse <- function(desc, code) {
    test_that(desc, {
      tmp <- normalizePath(tempfile(), mustWork = FALSE)
      on.exit(unlink(tmp))
      .rxWithSinkBoth(tmp, {
        .rxLastCompileSuccess(TRUE)
        expect_error(rxode2(code))
        # If the parser fails, it doesn't compile, which means
        # the last compiler success should be TRUE. When FALSE it
        # means the parser succeeded but the compiler failed, which
        # is not what we want here.
        expect_true(.rxLastCompileSuccess())
      })
    })
  }

  goodParse <- function(desc, code) {
    test_that(desc, {
      tmp <- normalizePath(tempfile(), mustWork = FALSE)
      on.exit(unlink(tmp))
      .rxWithSinkBoth(tmp, {
        rx <- rxode2(code)
        expect_s3_class(rx, "rxode2")
        rxDelete(rx)
      })
    })
  }

  equivSyntax <- function(desc, code1, code2) {
    test_that(desc, {
      tmp <- normalizePath(tempfile(), mustWork = FALSE)
      on.exit(unlink(tmp))
      .rxWithSinkBoth(tmp, {
        rx1 <- rxode2(code1)
        rx2 <- rxode2(code2)
        expect_equal(rxMd5(rx1)["parsed_md5"], rxMd5(rx2)["parsed_md5"])
        rxDelete(rx1)
        rxDelete(rx2)
      })
    })
  }

  badParse("incorrect d/dt operator", "d/dt(y = 1);")

  ## Statements don't require ; now.
  goodParse(
    "comments must be outside statements #1",
    "d/dt(y) = 1   # bad comment;")

  goodParse(
    'missing end of statement ";" dosen\'t cause errors',
    paste(
      sep = "\n",
      "d/dt(depot) = -ka * depot",
      "d/dt(centr) =  ka * depot - kout * centr;"))

  badParse(
    "arithmetic syntax error",
    paste(
      sep = "\n",
      "# comment, just to show error in line 3",
      "d/dt(y) = -ka;",
      "C1 = /y;"
    )
  )

  ## added ** operator
  goodParse("existing operator **",
            code = paste(
              sep = "\n",
              "d/dt(y) = -ka;",
              "C1 = ka *  y**2;"))
  badParse("unexistent operator %",
           code = paste(
             sep = "\n",
             "remainder = 4 % 3;",
             "d/dt(y) = -ka;",
             "C1 = ka * y;"))

  badParse(
    desc = 'incorrect "if" statement',
    code = paste(
      sep = "\n",
      "if(comed==0){",
      "   F = 1.0;",
      "else {", # missing "}"'
      "   F = 0.75;",
      "};",
      "d/dt(y) = F * y;"))

  badParse(
    desc = "illegal variable name (starting w. a digit)",
    code = paste(
      sep = "\n",
      "F = 0.75;",
      "12foo_bar = 1.0/2.0;",
      "d/dt(y) = F * y;"))

  goodParse(
    desc = "dot in variable name (ini0)",
    code = paste(
      sep = "\n",
      "F = 0.75;",
      "foo.bar = 1.0/2.0;",
      "d/dt(y) = F * y;"
    )
  )

  goodParse(
    desc = "dot in variable name in d/dt()",
    code = paste(
      sep = "\n",
      "d/dt(y_1) = F * y;", # okay
      "d/dt(y.1) = F * y;"
    ) # not okay
  )

  goodParse(
    desc = "leading dot in variable name",
    code = paste(
      sep = "\n",
      "F = 0.75;",
      ".foo.bar = 0.5;",
      "d/dt(y) = F * y;"
    )
  )

  goodParse(
    desc = "leading dot in variable name (ini0)",
    code = paste(
      sep = "\n",
      "F = 0.75;",
      ".foo.bar = 1.0/2.0;",
      "d/dt(y) = F * y;"
    )
  )

  goodParse(
    desc = "leading dot in variable name in d/dt()",
    code = paste(
      sep = "\n",
      "d/dt(y_1) = F * y;", # okay
      "d/dt(.y.1) = F * y;"
    ) # not okay
  )

  goodParse(
    desc = "leading dot in variable name",
    code = paste(
      sep = "\n",
      "F = 0.75;",
      ".foo.bar = 0.5;",
      "d/dt(y) = F * y;"))

  badParse(
    desc = "Assignment with <<- not supported",
    "d/dt(y_1) <<- F*y")

  goodParse(
    desc = "Assignment with <- supported #1",
    "d/dt(y_1) <- F*y")

  goodParse(
    desc = "Assignment with <- supported #2",
    "y_1(0) <- 1;d/dt(y_1) = F*y_1")

  goodParse(
    desc = "Assignment with <- supported #3",
    "y_2 <- 1;d/dt(y_1) = F*y"
  )

  goodParse(
    desc = "Assignment with <- supported #4",
    "y_2 <- 1+7;d/dt(y_1) = F*y"
  )

  goodParse(
    desc = "Assignment with <- supported #7",
    "d/dt(y_1) = F*y; df(y_1)/dy(y_1) <- 0"
  )

  badParse(
    desc = "Defining df(var1)/dy(var2) where var1 is not a state variable.",
    "
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
df(mu)/dy(y)=0;
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
"
)

  badParse(
    desc = "Defining df(var1)/dy(var2) where var1 is not a state variable.",
    "
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
df(mu)/dy(y)=0;
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
"
)

  goodParse(
    desc = "Defining df(var1)/dy(var2) where var1 is a state variable.",
    "
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
"
)

  goodParse(
    desc = "Defining df(var1)/dy(var2) where var2 is a variable.",
    "
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
df(dy)/dy(mu) = (1-y^2)*dy
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
"
)

  badParse(
    desc = "Defining df(var1)/dy(var2) where var2 is a calculated value.",
    "
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
df(dy)/dy(mu) = (1-y^2)*dy
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1+bad ## nonstiff; 10 moderately stiff; 1000 stiff
"
)

  goodParse(
    desc = "a*b/c^2",
    "d/dt(x)=a*b/c^2*x"
  )

  goodParse(
    desc = "a*b/c^2/d",
    "d/dt(x)=a*b/c^2/d*x"
  )

  goodParse(
    desc = "Transit as a compartment",
    "d/dt(transit) = -(1/mtt) * transit
        d/dt(depot)   =  (1/mtt) * transit - ka * depot
        d/dt(center)  = ka * depot - (cl/v1) * center - (q/v1) * center + (q/v2) * periph
        d/dt(periph)  = (q / v1) * center - (q / v2) * periph
        cp = center / v1"
  )

  goodParse(
    desc = "=+ parsing",
    "C2 = +centr/V2;
              C3 = peri/V3;
              d/dt(depot) =-KA*depot;
              d/dt(centr) = +KA*depot - CL*C2 - Q*C2 + Q*C3;
              d/dt(peri)  =                    Q*C2 - Q*C3;
              d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;"
  )

  for (v in c("ii", "evid")) {
    badParse(
      desc = sprintf("bad variables: %s", v),
      sprintf("var=%s", v)
    )
  }
  for (v in "tlast") {
    goodParse(
      desc = sprintf("good variables: %s", v),
      sprintf("var=%s", v)
    )
  }
  for (v in "abs") {
    goodParse(
      desc = sprintf("good functions: %s", v),
      sprintf("var=%s(x)", v)
    )
  }

  badParse(desc = "No duplicate dvid()", "a=b;dvid(1,2,3);dvid(3,4,5)")

  badParse(desc = "Bad dvid(0)", "a=b;dvid(0);")
  badParse(desc = "Bad dvid(0, 1)", "a=b;dvid(1,0);")

  goodParse(desc = "THETA/ETA parsing", "a=THETA[1]+ETA[2];")

  goodParse(desc = "Duplicate d/dt(x)", "d/dt(depot) = -depot*ka;\nd/dt(depot) = d/dt(depot)+0")

  goodParse(desc = "Duplicate d/dt(x)", "d/dt(depot) ~ -depot*ka;\nd/dt(depot) = d/dt(depot)+0")
  goodParse(desc = "Duplicate d/dt(x)", "d/dt(depot) = -depot*ka;\nd/dt(depot) ~ d/dt(depot)+0")

  goodParse(desc = "pi Parse", "a = pi+e2")

  for (v in c("f", "F", "alag", "lag", "rate", "dur")) {
    badParse(
      sprintf("%s can not depend on d/dt(state)", v),
      sprintf("d/dt(depot)=-depot*ka;\nd/dt(central)=ka*depot-kel*central\n%s(depot)=d/dt(central)+3", v)
    )
  }

  badParse(desc = "Using d/dt(x) before defined", "y=d/dt(x)+3")

  goodParse(
    desc = "Jacobain with theta and eta",
    "d/dt(x)=(THETA[1]+ETA[1])*x\ndf(x)/dy(THETA[1]) = 1\ndf(x)/dy(ETA[1]) = 1\n"
  )


  for (v in c("f", "F", "alag", "lag", "rate", "dur")) {
    badParse(
      sprintf("%s cannot depend on jacobain info", v),
      sprintf("d/dt(x)=(THETA[1]+ETA[1])*x\ndf(x)/dy(THETA[1]) = 1\ndf(x)/dy(ETA[1]) = 1\n%s(x)=df(x)/dt(ETA[1])+3", v)
    )
  }

  badParse(
    "mtime cannot depend on df(x)/dt(ETA[1])",
    "d/dt(x)=(THETA[1]+ETA[1])*x\ndf(x)/dy(THETA[1]) = 1\ndf(x)/dy(ETA[1]) = 1\nmtime(z)=df(x)/dt(ETA[1])+3"
  )
  badParse(
    "mtime cannot depend on d/dt(x)",
    "d/dt(x)=(THETA[1]+ETA[1])*x\ndf(x)/dy(THETA[1]) = 1\ndf(x)/dy(ETA[1]) = 1\nmtime(z)=d/dt(x)+3"
  )

  withr::with_options(list("rxode2.syntax.require.ode.first" = FALSE), {
    goodParse(
      "verbose output",
      "d/dt(x) = -k*x"
    )
  })
  goodParse(
    "functional initialization ok",
    "x(0) = y + 3\nd/dt(x) =-3*z"
  )

  ## 'rate' and 'dur' can be data items, so they cannot be variables
  ## in an rxode2 model
  for (var in c("alag", "f", "F")) {
    goodParse(
      sprintf("Parsing of %s as a variable and function work.", var),
      sprintf("d/dt(x) = -k*x;%s(x) = %s;", var, var)
    )
  }

  goodParse("x=ifelse(!matt,0,1)", "x=ifelse(!matt,0,1)")
  goodParse("x=ifelse(!(matt),0,1)", "x=ifelse(!(matt),0,1)")
  goodParse("x=ifelse((!matt),0,1)", "x=ifelse((!matt),0,1)")

  goodParse("mix lincmt with lags etc", "popCl <- 1
    popV <- 20
    popKa <- 1
    popVp <- 10
    popQ <- 2
    bsvCl <-0
    bsvV <- 0
    bsvKa <-0
    bsvVp <- 0
    bsvQ <-0
    popKeo <- 1.4
    bsvKeo <- 0
    popE0 <- 0
    popEmax <- 1
    popEC50 <- 5
    popGamma <- 1
    bsvE0 <- 0
    bsvEmax <- 0
    bsvEC50 <- 0
    ##
    cl ~ popCl * exp(bsvCl)
    v ~ popV * exp(bsvV)
    ka ~ popKa * exp(bsvKa)
    q ~ popQ * exp(bsvQ)
    vp ~ popVp * exp(bsvVp)
    keo ~ popKeo * exp(bsvKeo)
    popLagDepot <- 0
    popLagCentral <- 0
    popRateCentral <- 0
    popDurCentral <- 0
    bsvLagDepot <- 0
    bsvLagCentral <- 0
    bsvRateCentral <- 0
    bsvDurCentral <- 0
    alag(depot) <- popLagDepot * exp(bsvLagDepot)
    alag(central) <- popLagCentral * exp(bsvLagCentral)
    rate(central) <- popRateCentral * exp(bsvRateCentral)
    dur(central) <- popDurCentral * exp(bsvDurCentral)
    cp <- linCmt()
    d/dt(ce) = keo*(cp-ce)
    effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));")

  withr::with_options(list("rxode2.syntax.require.ode.first" = TRUE), {
    badParse(
      "Still cannot take undefined compartments",
      "popCl <- 1
    popV <- 20
    popKa <- 1
    popVp <- 10
    popQ <- 2
    bsvCl <-0
    bsvV <- 0
    bsvKa <-0
    bsvVp <- 0
    bsvQ <-0
    popKeo <- 1.4
    bsvKeo <- 0
    popE0 <- 0
    popEmax <- 1
    popEC50 <- 5
    popGamma <- 1
    bsvE0 <- 0
    bsvEmax <- 0
    bsvEC50 <- 0
    ##
    cl ~ popCl * exp(bsvCl)
    v ~ popV * exp(bsvV)
    ka ~ popKa * exp(bsvKa)
    q ~ popQ * exp(bsvQ)
    vp ~ popVp * exp(bsvVp)
    keo ~ popKeo * exp(bsvKeo)
    popLagDepot <- 0
    popLagCentral <- 0
    popRateCentral <- 0
    popDurCentral <- 0
    bsvLagDepot <- 0
    bsvLagCentral <- 0
    bsvRateCentral <- 0
    bsvDurCentral <- 0
    alag(depot) <- popLagDepot * exp(bsvLagDepot)
    alag(central) <- popLagCentral * exp(bsvLagCentral)
    rate(central) <- popRateCentral * exp(bsvRateCentral)
    dur(central) <- popDurCentral * exp(bsvDurCentral)
    dur(matt) <- 3
    cp <- linCmt()
    d/dt(ce) = keo*(cp-ce)
    effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));"
    )
  })

  badParse("cmt(depot) doesn't work with linCmt()", "popCl <- 1
    cmt(depot)
    popV <- 20
    popKa <- 1
    popVp <- 10
    popQ <- 2
    bsvCl <-0
    bsvV <- 0
    bsvKa <-0
    bsvVp <- 0
    bsvQ <-0
    popKeo <- 1.4
    bsvKeo <- 0
    popE0 <- 0
    popEmax <- 1
    popEC50 <- 5
    popGamma <- 1
    bsvE0 <- 0
    bsvEmax <- 0
    bsvEC50 <- 0
    ##
    cl ~ popCl * exp(bsvCl)
    v ~ popV * exp(bsvV)
    ka ~ popKa * exp(bsvKa)
    q ~ popQ * exp(bsvQ)
    vp ~ popVp * exp(bsvVp)
    keo ~ popKeo * exp(bsvKeo)
    popLagDepot <- 0
    popLagCentral <- 0
    popRateCentral <- 0
    popDurCentral <- 0
    bsvLagDepot <- 0
    bsvLagCentral <- 0
    bsvRateCentral <- 0
    bsvDurCentral <- 0
    alag(depot) <- popLagDepot * exp(bsvLagDepot)
    alag(central) <- popLagCentral * exp(bsvLagCentral)
    rate(central) <- popRateCentral * exp(bsvRateCentral)
    dur(central) <- popDurCentral * exp(bsvDurCentral)
    cp <- linCmt()
    d/dt(ce) = keo*(cp-ce)
    effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));")

  badParse("cmt(central) doesn't work with linCmt()", "popCl <- 1
    cmt(central)
    popV <- 20
    popKa <- 1
    popVp <- 10
    popQ <- 2
    bsvCl <-0
    bsvV <- 0
    bsvKa <-0
    bsvVp <- 0
    bsvQ <-0
    popKeo <- 1.4
    bsvKeo <- 0
    popE0 <- 0
    popEmax <- 1
    popEC50 <- 5
    popGamma <- 1
    bsvE0 <- 0
    bsvEmax <- 0
    bsvEC50 <- 0
    ##
    cl ~ popCl * exp(bsvCl)
    v ~ popV * exp(bsvV)
    ka ~ popKa * exp(bsvKa)
    q ~ popQ * exp(bsvQ)
    vp ~ popVp * exp(bsvVp)
    keo ~ popKeo * exp(bsvKeo)
    popLagDepot <- 0
    popLagCentral <- 0
    popRateCentral <- 0
    popDurCentral <- 0
    bsvLagDepot <- 0
    bsvLagCentral <- 0
    bsvRateCentral <- 0
    bsvDurCentral <- 0
    alag(depot) <- popLagDepot * exp(bsvLagDepot)
    alag(central) <- popLagCentral * exp(bsvLagCentral)
    rate(central) <- popRateCentral * exp(bsvRateCentral)
    dur(central) <- popDurCentral * exp(bsvDurCentral)
    cp <- linCmt()
    d/dt(ce) = keo*(cp-ce)
    effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));")

  badParse("theta0", "a = theta[0]")
  badParse("eta0", "a = eta[0]")
  goodParse("theta1", "a = theta[1]")
  goodParse("eta1", "a = eta[1]")
  badParse("matt1", "a = matt[1]")

  badParse("ifelse1", "ifelse=3")
  badParse("ifelse2", "a=ifelse+3")
  badParse("ifelse3", "d/dt(ifelse)=matt")

  badParse("if1", "if=3")
  badParse("if2", "a=if+3")
  badParse("if3", "d/dt(if)=matt")


  badParse("cmt1", "cmt=3")
  goodParse("cmt2", "a=cmt+3")
  badParse("cmt3", "d/dt(cmt)=matt")

  badParse("dvid1", "dvid=3")
  goodParse("dvid2", "a=dvid+3")
  badParse("dvid3", "d/dt(dvid)=matt")

  badParse("addl1", "addl=3")
  goodParse("addl2", "a=addl+3")
  badParse("addl3", "d/dt(addl)=matt")

  badParse("ss1", "ss=3")
  goodParse("ss2", "a=ss+3")
  badParse("ss3", "d/dt(ss)=matt")

  badParse("amt1", "amt=3")
  goodParse("amt2", "a=amt+3")
  badParse("amt3", "d/dt(amt)=matt")

  badParse("rate1", "rate=3")
  goodParse("rate2", "a=rate+3")
  badParse("rate3", "d/dt(rate)=matt")

  badParse("printf1", "printf=3")
  badParse("printf2", "a=printf+3")
  badParse("printf3", "d/dt(printf)=matt")

  badParse("Rprintf1", "Rprintf=3")
  badParse("Rprintf2", "a=Rprintf+3")
  badParse("Rprintf3", "d/dt(Rprintf)=matt")

  badParse("print1", "print=3")
  badParse("print2", "a=print+3")
  badParse("print3", "d/dt(print)=matt")

  goodParse("sum1", "a=sum(1,2,3,a,b,c)")
  goodParse("sum2", "a=lag(b, 1)")

  goodParse("transit1", "a=transit(n, mtt, bio)")
  goodParse("transit2", "a=transit(n, mtt)")
  badParse("transit3", "a=transit(n, mtt, bio,ack)")

  goodParse("fun1", "a=is.nan(x)")
  badParse("fun2", "a=is.nan(x,b)")
  badParse("fun3", "a=is.nan()")

  goodParse("fun4", "a=is.finite(x)")
  badParse("fun5", "a=is.finite(x,a)")
  badParse("fun6", "a=is.finite()")

  goodParse("fun7", "a=is.infinite(x)")
  badParse("fun8", "a=is.infinite(x,a)")
  badParse("fun9", "a=is.infinite()")

  badParse("fun10", "t=tinf")
  badParse("fun11", "time=tinf")

  badParse("while/else", "a=1;while(1){a=a+3} else { a=3}")

  goodParse("while", "a=1;while(1){a=a+3}")

  goodParse("while-break", "a=1;while(1){a=a+3; break;}")
  badParse("while-break-bad", "a=1;while(1){a=a+3;}; break;")

  goodParse("Dotted initial conditions",
            paste(c("d/dt(C.A) = - 1",
                    "C.A(0) = A"), collapse="\n"))


  # From rxode2parse
  badParse <- function(desc, code) {
    test_that(desc, {
      expect_error(suppressMessages(rxode2parse(code)))
    })
  }

  goodParse <- function(desc, code) {
    test_that(desc, {
      expect_s3_class(rxode2parse(code), "rxModelVars")
    })
  }

  badParse("incorrect d/dt operator", "d/dt(y = 1);")

  ## Statements don't require ; now.
  goodParse(
    "comments must be outside statements #1",
    "d/dt(y) = 1   # bad comment;")

  goodParse(
    'missing end of statement ";" dosen\'t cause errors',
    paste(
      sep = "\n",
      "d/dt(depot) = -ka * depot",
      "d/dt(centr) =  ka * depot - kout * centr;"))

  badParse(
    "arithmetic syntax error",
    paste(
      sep = "\n",
      "# comment, just to show error in line 3",
      "d/dt(y) = -ka;",
      "C1 = /y;"
    )
  )

  ## added ** operator
  goodParse("existing operator **",
            code = paste(
              sep = "\n",
              "d/dt(y) = -ka;",
              "C1 = ka *  y**2;"))
  badParse("unexistent operator %",
           code = paste(
             sep = "\n",
             "remainder = 4 % 3;",
             "d/dt(y) = -ka;",
             "C1 = ka * y;"))

  badParse(
    desc = 'incorrect "if" statement',
    code = paste(
      sep = "\n",
      "if(comed==0){",
      "   F = 1.0;",
      "else {", # missing "}"'
      "   F = 0.75;",
      "};",
      "d/dt(y) = F * y;"))

  badParse(
    desc = "illegal variable name (starting w. a digit)",
    code = paste(
      sep = "\n",
      "F = 0.75;",
      "12foo_bar = 1.0/2.0;",
      "d/dt(y) = F * y;"))

  goodParse(
    desc = "dot in variable name (ini0)",
    code = paste(
      sep = "\n",
      "F = 0.75;",
      "foo.bar = 1.0/2.0;",
      "d/dt(y) = F * y;"
    )
  )

  goodParse(
    desc = "dot in variable name in d/dt()",
    code = paste(
      sep = "\n",
      "d/dt(y_1) = F * y;", # okay
      "d/dt(y.1) = F * y;"
    ) # not okay
  )

  goodParse(
    desc = "leading dot in variable name",
    code = paste(
      sep = "\n",
      "F = 0.75;",
      ".foo.bar = 0.5;",
      "d/dt(y) = F * y;"
    )
  )

  goodParse(
    desc = "leading dot in variable name (ini0)",
    code = paste(
      sep = "\n",
      "F = 0.75;",
      ".foo.bar = 1.0/2.0;",
      "d/dt(y) = F * y;"
    )
  )

  goodParse(
    desc = "leading dot in variable name in d/dt()",
    code = paste(
      sep = "\n",
      "d/dt(y_1) = F * y;", # okay
      "d/dt(.y.1) = F * y;"
    ) # not okay
  )

  goodParse(
    desc = "leading dot in variable name",
    code = paste(
      sep = "\n",
      "F = 0.75;",
      ".foo.bar = 0.5;",
      "d/dt(y) = F * y;"))

  badParse(
    desc = "Assignment with <<- not supported",
    "d/dt(y_1) <<- F*y")

  goodParse(
    desc = "Assignment with <- supported #1",
    "d/dt(y_1) <- F*y")

  goodParse(
    desc = "Assignment with <- supported #2",
    "y_1(0) <- 1;d/dt(y_1) = F*y_1")

  goodParse(
    desc = "Assignment with <- supported #3",
    "y_2 <- 1;d/dt(y_1) = F*y"
  )

  goodParse(
    desc = "Assignment with <- supported #4",
    "y_2 <- 1+7;d/dt(y_1) = F*y"
  )

  goodParse(
    desc = "Assignment with <- supported #7",
    "d/dt(y_1) = F*y; df(y_1)/dy(y_1) <- 0"
  )

  badParse(
    desc = "Defining df(var1)/dy(var2) where var1 is not a state variable.",
    "
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
df(mu)/dy(y)=0;
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
"
)

  badParse(
    desc = "Defining df(var1)/dy(var2) where var1 is not a state variable.",
    "
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
df(mu)/dy(y)=0;
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
"
)

  goodParse(
    desc = "Defining df(var1)/dy(var2) where var1 is a state variable.",
    "
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
"
)

  goodParse(
    desc = "Defining df(var1)/dy(var2) where var2 is a variable.",
    "
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
df(dy)/dy(mu) = (1-y^2)*dy
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
"
)

  badParse(
    desc = "Defining df(var1)/dy(var2) where var2 is a calculated value.",
    "
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
df(dy)/dy(mu) = (1-y^2)*dy
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1+bad ## nonstiff; 10 moderately stiff; 1000 stiff
"
)

  goodParse(
    desc = "a*b/c^2",
    "d/dt(x)=a*b/c^2*x"
  )

  goodParse(
    desc = "a*b/c^2/d",
    "d/dt(x)=a*b/c^2/d*x"
  )

  goodParse(
    desc = "Transit as a compartment",
    "d/dt(transit) = -(1/mtt) * transit
        d/dt(depot)   =  (1/mtt) * transit - ka * depot
        d/dt(center)  = ka * depot - (cl/v1) * center - (q/v1) * center + (q/v2) * periph
        d/dt(periph)  = (q / v1) * center - (q / v2) * periph
        cp = center / v1"
  )

  goodParse(
    desc = "=+ parsing",
    "C2 = +centr/V2;
              C3 = peri/V3;
              d/dt(depot) =-KA*depot;
              d/dt(centr) = +KA*depot - CL*C2 - Q*C2 + Q*C3;
              d/dt(peri)  =                    Q*C2 - Q*C3;
              d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;"
  )

  for (v in c("ii", "evid")) {
    badParse(
      desc = sprintf("bad variables: %s", v),
      sprintf("var=%s", v)
    )
  }
  for (v in "tlast") {
    goodParse(
      desc = sprintf("good variables: %s", v),
      sprintf("var=%s", v)
    )
  }
  for (v in "abs") {
    goodParse(
      desc = sprintf("good functions: %s", v),
      sprintf("var=%s(x)", v)
    )
  }

  badParse(desc = "No duplicate dvid()", "a=b;dvid(1,2,3);dvid(3,4,5)")

  badParse(desc = "Bad dvid(0)", "a=b;dvid(0);")
  badParse(desc = "Bad dvid(0, 1)", "a=b;dvid(1,0);")

  goodParse(desc = "THETA/ETA parsing", "a=THETA[1]+ETA[2];")

  goodParse(desc = "Duplicate d/dt(x)", "d/dt(depot) = -depot*ka;\nd/dt(depot) = d/dt(depot)+0")

  goodParse(desc = "Duplicate d/dt(x)", "d/dt(depot) ~ -depot*ka;\nd/dt(depot) = d/dt(depot)+0")
  goodParse(desc = "Duplicate d/dt(x)", "d/dt(depot) = -depot*ka;\nd/dt(depot) ~ d/dt(depot)+0")

  goodParse(desc = "pi Parse", "a = pi+e2")

  for (v in c("f", "F", "alag", "lag", "rate", "dur")) {
    badParse(
      sprintf("%s can not depend on d/dt(state) (state-dependent not allowed)", v),
      sprintf("d/dt(depot)=-depot*ka;\nd/dt(central)=ka*depot-kel*central\n%s(depot)=d/dt(central)+3", v)
    )
  }

  badParse(desc = "Using d/dt(x) before defined", "y=d/dt(x)+3")

  goodParse(
    desc = "Jacobain with theta and eta",
    "d/dt(x)=(THETA[1]+ETA[1])*x\ndf(x)/dy(THETA[1]) = 1\ndf(x)/dy(ETA[1]) = 1\n"
  )


  for (v in c("f", "F", "alag", "lag", "rate", "dur")) {
    badParse(
      sprintf("%s cannot depend on jacobain info", v),
      sprintf("d/dt(x)=(THETA[1]+ETA[1])*x\ndf(x)/dy(THETA[1]) = 1\ndf(x)/dy(ETA[1]) = 1\n%s(x)=df(x)/dt(ETA[1])+3", v)
    )
  }

  badParse(
    "mtime cannot depend on df(x)/dt(ETA[1])",
    "d/dt(x)=(THETA[1]+ETA[1])*x\ndf(x)/dy(THETA[1]) = 1\ndf(x)/dy(ETA[1]) = 1\nmtime(z)=df(x)/dt(ETA[1])+3"
  )
  badParse(
    "mtime cannot depend on d/dt(x)",
    "d/dt(x)=(THETA[1]+ETA[1])*x\ndf(x)/dy(THETA[1]) = 1\ndf(x)/dy(ETA[1]) = 1\nmtime(z)=d/dt(x)+3"
  )

  goodParse(
    "functional initialization ok",
    "x(0) = y + 3\nd/dt(x) =-3*z"
  )

  ## 'rate' and 'dur' can be data items, so they cannot be variables
  ## in an rxode2 model
  for (var in c("alag", "f", "F")) {
    goodParse(
      sprintf("Parsing of %s as a variable and function work.", var),
      sprintf("d/dt(x) = -k*x;%s(x) = %s;", var, var)
    )
  }

  goodParse("x=ifelse(!matt,0,1)", "x=ifelse(!matt,0,1)")
  goodParse("x=ifelse(!(matt),0,1)", "x=ifelse(!(matt),0,1)")
  goodParse("x=ifelse((!matt),0,1)", "x=ifelse((!matt),0,1)")

  goodParse("mix lincmt with lags etc", "popCl <- 1
    popV <- 20
    popKa <- 1
    popVp <- 10
    popQ <- 2
    bsvCl <-0
    bsvV <- 0
    bsvKa <-0
    bsvVp <- 0
    bsvQ <-0
    popKeo <- 1.4
    bsvKeo <- 0
    popE0 <- 0
    popEmax <- 1
    popEC50 <- 5
    popGamma <- 1
    bsvE0 <- 0
    bsvEmax <- 0
    bsvEC50 <- 0
    ##
    cl ~ popCl * exp(bsvCl)
    v ~ popV * exp(bsvV)
    ka ~ popKa * exp(bsvKa)
    q ~ popQ * exp(bsvQ)
    vp ~ popVp * exp(bsvVp)
    keo ~ popKeo * exp(bsvKeo)
    popLagDepot <- 0
    popLagCentral <- 0
    popRateCentral <- 0
    popDurCentral <- 0
    bsvLagDepot <- 0
    bsvLagCentral <- 0
    bsvRateCentral <- 0
    bsvDurCentral <- 0
    alag(depot) <- popLagDepot * exp(bsvLagDepot)
    alag(central) <- popLagCentral * exp(bsvLagCentral)
    rate(central) <- popRateCentral * exp(bsvRateCentral)
    dur(central) <- popDurCentral * exp(bsvDurCentral)
    cp <- linCmt()
    d/dt(ce) = keo*(cp-ce)
    effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));")

  withr::with_options(list("rxode2.syntax.require.ode.first" = TRUE), {
  badParse(
    "Still cannot take undefined compartments",
    "popCl <- 1
    popV <- 20
    popKa <- 1
    popVp <- 10
    popQ <- 2
    bsvCl <-0
    bsvV <- 0
    bsvKa <-0
    bsvVp <- 0
    bsvQ <-0
    popKeo <- 1.4
    bsvKeo <- 0
    popE0 <- 0
    popEmax <- 1
    popEC50 <- 5
    popGamma <- 1
    bsvE0 <- 0
    bsvEmax <- 0
    bsvEC50 <- 0
    ##
    cl ~ popCl * exp(bsvCl)
    v ~ popV * exp(bsvV)
    ka ~ popKa * exp(bsvKa)
    q ~ popQ * exp(bsvQ)
    vp ~ popVp * exp(bsvVp)
    keo ~ popKeo * exp(bsvKeo)
    popLagDepot <- 0
    popLagCentral <- 0
    popRateCentral <- 0
    popDurCentral <- 0
    bsvLagDepot <- 0
    bsvLagCentral <- 0
    bsvRateCentral <- 0
    bsvDurCentral <- 0
    alag(depot) <- popLagDepot * exp(bsvLagDepot)
    alag(central) <- popLagCentral * exp(bsvLagCentral)
    rate(central) <- popRateCentral * exp(bsvRateCentral)
    dur(central) <- popDurCentral * exp(bsvDurCentral)
    dur(matta) <- 3 # undefined compartment
    cp <- linCmt()
    d/dt(ce) = keo*(cp-ce)
    effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));"
  )})

  badParse("cmt(depot) doesn't work with linCmt()", "popCl <- 1
    cmt(depot)
    popV <- 20
    popKa <- 1
    popVp <- 10
    popQ <- 2
    bsvCl <-0
    bsvV <- 0
    bsvKa <-0
    bsvVp <- 0
    bsvQ <-0
    popKeo <- 1.4
    bsvKeo <- 0
    popE0 <- 0
    popEmax <- 1
    popEC50 <- 5
    popGamma <- 1
    bsvE0 <- 0
    bsvEmax <- 0
    bsvEC50 <- 0
    ##
    cl ~ popCl * exp(bsvCl)
    v ~ popV * exp(bsvV)
    ka ~ popKa * exp(bsvKa)
    q ~ popQ * exp(bsvQ)
    vp ~ popVp * exp(bsvVp)
    keo ~ popKeo * exp(bsvKeo)
    popLagDepot <- 0
    popLagCentral <- 0
    popRateCentral <- 0
    popDurCentral <- 0
    bsvLagDepot <- 0
    bsvLagCentral <- 0
    bsvRateCentral <- 0
    bsvDurCentral <- 0
    alag(depot) <- popLagDepot * exp(bsvLagDepot)
    alag(central) <- popLagCentral * exp(bsvLagCentral)
    rate(central) <- popRateCentral * exp(bsvRateCentral)
    dur(central) <- popDurCentral * exp(bsvDurCentral)
    cp <- linCmt()
    d/dt(ce) = keo*(cp-ce)
    effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));")

  badParse("cmt(central) doesn't work with linCmt()", "popCl <- 1
    cmt(central)
    popV <- 20
    popKa <- 1
    popVp <- 10
    popQ <- 2
    bsvCl <-0
    bsvV <- 0
    bsvKa <-0
    bsvVp <- 0
    bsvQ <-0
    popKeo <- 1.4
    bsvKeo <- 0
    popE0 <- 0
    popEmax <- 1
    popEC50 <- 5
    popGamma <- 1
    bsvE0 <- 0
    bsvEmax <- 0
    bsvEC50 <- 0
    ##
    cl ~ popCl * exp(bsvCl)
    v ~ popV * exp(bsvV)
    ka ~ popKa * exp(bsvKa)
    q ~ popQ * exp(bsvQ)
    vp ~ popVp * exp(bsvVp)
    keo ~ popKeo * exp(bsvKeo)
    popLagDepot <- 0
    popLagCentral <- 0
    popRateCentral <- 0
    popDurCentral <- 0
    bsvLagDepot <- 0
    bsvLagCentral <- 0
    bsvRateCentral <- 0
    bsvDurCentral <- 0
    alag(depot) <- popLagDepot * exp(bsvLagDepot)
    alag(central) <- popLagCentral * exp(bsvLagCentral)
    rate(central) <- popRateCentral * exp(bsvRateCentral)
    dur(central) <- popDurCentral * exp(bsvDurCentral)
    cp <- linCmt()
    d/dt(ce) = keo*(cp-ce)
    effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));")

  badParse("theta0", "a = theta[0]")
  badParse("eta0", "a = eta[0]")
  goodParse("theta1", "a = theta[1]")
  goodParse("eta1", "a = eta[1]")
  badParse("matt1", "a = matt[1]")

  badParse("ifelse1", "ifelse=3")
  badParse("ifelse2", "a=ifelse+3")
  badParse("ifelse3", "d/dt(ifelse)=matt")

  badParse("if1", "if=3")
  badParse("if2", "a=if+3")
  badParse("if3", "d/dt(if)=matt")


  badParse("cmt1", "cmt=3")
  goodParse("cmt2", "a=cmt+3")
  badParse("cmt3", "d/dt(cmt)=matt")

  badParse("dvid1", "dvid=3")
  goodParse("dvid2", "a=dvid+3")
  badParse("dvid3", "d/dt(dvid)=matt")

  badParse("addl1", "addl=3")
  goodParse("addl2", "a=addl+3")
  badParse("addl3", "d/dt(addl)=matt")

  badParse("ss1", "ss=3")
  goodParse("ss2", "a=ss+3")
  badParse("ss3", "d/dt(ss)=matt")

  badParse("amt1", "amt=3")
  goodParse("amt2", "a=amt+3")
  badParse("amt3", "d/dt(amt)=matt")

  badParse("rate1", "rate=3")
  goodParse("rate2", "a=rate+3")
  badParse("rate3", "d/dt(rate)=matt")

  badParse("printf1", "printf=3")
  badParse("printf2", "a=printf+3")
  badParse("printf3", "d/dt(printf)=matt")

  badParse("Rprintf1", "Rprintf=3")
  badParse("Rprintf2", "a=Rprintf+3")
  badParse("Rprintf3", "d/dt(Rprintf)=matt")

  badParse("print1", "print=3")
  badParse("print2", "a=print+3")
  badParse("print3", "d/dt(print)=matt")

  goodParse("sum1", "a=sum(1,2,3,a,b,c)")
  goodParse("sum2", "a=lag(b, 1)")

  goodParse("transit1", "a=transit(n, mtt, bio)")
  goodParse("transit2", "a=transit(n, mtt)")
  badParse("transit3", "a=transit(n, mtt, bio,ack)")

  goodParse("fun1", "a=is.nan(x)")
  badParse("fun2", "a=is.nan(x,b)")
  badParse("fun3", "a=is.nan()")

  goodParse("fun4", "a=is.finite(x)")
  badParse("fun5", "a=is.finite(x,a)")
  badParse("fun6", "a=is.finite()")

  goodParse("fun7", "a=is.infinite(x)")
  badParse("fun8", "a=is.infinite(x,a)")
  badParse("fun9", "a=is.infinite()")

  badParse("fun10", "t=tinf")
  badParse("fun11", "time=tinf")

  badParse("while/else", "a=1;while(1){a=a+3} else { a=3}")

  goodParse("while", "a=1;while(1){a=a+3}")

  goodParse("while-break", "a=1;while(1){a=a+3; break;}")
  badParse("while-break-bad", "a=1;while(1){a=a+3;}; break;")

  goodParse("Dotted initial conditions",
            paste(c("d/dt(C.A) = - 1",
                    "C.A(0) = A"), collapse="\n"))

  goodParse("Less than expression",
            "a <- 3 < -1")

  badParse("Double assignment",
           "a <- 3 <- 1")


  badParse("Double assignment#2",
           "a = 3 = 1")

  badParse("Double assignment #3",
           "a <- 3 -> b")


  test_that("after isn't shown or garbled", {
    t <-try(rxode2parse("a+b<-fun+fun  +  fun"))
    expect_true(inherits(t, "try-error"))
    print(attr(t,"condition")$message)
    expect_true(regexpr("after", attr(t,"condition")$message)==-1)
  })

  test_that("throws parsing error with wrong number of arguments", {
    .trans <- rxode2parseGetTranslation()
    expect_error(rxode2parse("a= llikNorm(a, b, c, d, f)"))

    .trans2 <- .trans
    .w <- which(.trans2$rxFun=="llikNorm")
    .trans2$argMax[.w] <- 4L
    rxode2parseAssignTranslation(.trans2)

    expect_error(rxode2parse("a= llikNorm(a, b, c, d, f)"))

    .trans2$argMax[.w] <- 2L
    rxode2parseAssignTranslation(.trans2)

    expect_error(rxode2parse("a= llikNorm(a, b, c, d, f)"))

    rxode2parseAssignTranslation(.trans)
    .trans2 <- .trans
    expect_true(rxode2parse("a=llikNorm(a, b, c)")$flags["thread"] == 1L)

    # pretend that llikNorm is not thread safe
    .trans2$threadSafe[.w] <- 0L
    rxode2parseAssignTranslation(.trans2)

    expect_true(rxode2parse("a=llikNorm(a, b, c)")$flags["thread"] == 0L)
    rxode2parseAssignTranslation(.trans)

    expect_error(rxode2parse("a=cos(b, c, d, e, f)"))
  })


  test_that("linear compartmental error", {
    expect_error(rxode2parse('f(central) <- 1 + f_study1 * (STUDYID == "Study 1")\nka <- exp(tka + eta.ka)\ncl <- exp(tcl + eta.cl)\nv <- exp(tv + eta.v)\ncp <- linCmt()', linear=TRUE), NA)


  })


  test_that("TIME conundrums", {

    p <- rxode2parse("param(emax_fcfb,lec50,le0,let50_emax,propSd,etale0,TIME,PK);\ne0=exp(le0+etale0);\nemax=emax_fcfb;\nec50=exp(lec50);\net50_emax=exp(let50_emax);\nfoo=e0*(1+emax*(TIME/168)/(et50_emax+(TIME/168))*PK/(ec50+PK));\nrx_yj_~2;\nrx_lambda_~1;\nrx_low_~0;\nrx_hi_~1;\nrx_pred_f_~foo;\nrx_pred_~rx_pred_f_;\nrx_r_~(rx_pred_f_*propSd)^2;\nipredSim=rxTBSi(rx_pred_,rx_lambda_,rx_yj_,rx_low_,rx_hi_);\nsim=rxTBSi(rx_pred_+sqrt(rx_r_)*err.foo,rx_lambda_,rx_yj_,rx_low_,rx_hi_);\ncmt(foo);\ndvid(1);\n")

    expect_false(any(p$params == "TIME"))

  })


  test_that("pow problems", {
    expect_error(rxode2parse("pow=3+4"), NA)

    rxode2parse("pow=4\nif (CMT==5){pow = 3+1+pow(4, 3)}\npow2 = pow*2", code="rxode2parse_test_code.c")
    expect_true(file.exists("rxode2parse_test_code.c"))
    if (file.exists("rxode2parse_test_code.c")) {
      lines <- readLines("rxode2parse_test_code.c")
      unlink("rxode2parse_test_code.c")
      expect_false(file.exists("rxode2parse_test_code.c"))
      expect_true(any(regexpr("+pow(4,3)", lines, fixed=TRUE) != -1))
      expect_true(any(regexpr("_rxNotFun_pow=", lines, fixed=TRUE) != -1))
      expect_true(any(regexpr("=_rxNotFun_pow*2", lines, fixed=TRUE) != -1))
      expect_true(any(regexpr("(_CMT==5)", lines, fixed=TRUE)) != -1)
    }
  })

  test_that("cmt/locf interaction", {
    expect_equal(rxNorm(rxModelVars("cmt(a);\nlocf(b);d/dt(a)=b*kel")),
                 "cmt(a);\nlocf(b);\nd/dt(a)=b*kel;\n")
  })

})

rxTest({
  test_that("parsing for nlmixr; Issue RxODE#281", {
    .mv <- rxModelVars("Resp(0)=1;\nemaxD=exp(temaxD);\nec50=exp(tec50);\nemaxT=exp(temaxT);\net50=exp(tet50);\nslope=exp(tslope);\ndelay=exp(tdelay);\nkin=exp(tkin);\nkout=kin;\nGAM=exp(1)/2;\nC2=centr/V;\nCONC=(C2)^2;\nStim1=emaxT*(t)/(t+et50);\nStim2=emaxD*(CONC^GAM)/(CONC^GAM+ec50^GAM);\nStim=Stim1*(1+TRX*Stim2);\nDelta=1/(1+exp(-20*(t-delay)));\nd/dt(depot)=-KA*depot;\nd/dt(centr)=KA*depot-CL*C2;\nd/dt(Resp)=kin*(1+Delta*slope)-kout*(1+Stim)*Resp;\nnlmixr_pred=Resp;\nrx_yj_~2;\nrx_lambda_~1;\nrx_hi_~1;\nrx_low_~0;\nrx_pred_f_~nlmixr_pred;\nrx_pred_=nlmixr_pred;\nrx_r_=(nlmixrAdd)^2;\n")
    
    expect_error(rxS(.mv, TRUE, FALSE), NA)
    
    .mv <- rxModelVars("ktr=exp(tktr);\nka=exp(tka);\ncl=exp(tcl);\nv=exp(tv);\nemax=expit(temax);\nec50=exp(tec50);\nkout=exp(tkout);\ne0=exp(te0);\nDCP=center/v;\nPD=1-emax*DCP/(ec50+DCP);\neffect(0)=e0;\nkin=e0*kout;\nd/dt(depot)=-ktr*depot;\nd/dt(gut)=ktr*depot-ka*gut;\nd/dt(center)=ka*gut-cl/v*center;\nd/dt(effect)=kin*PD-kout*effect;\ncp=center/v;\nnlmixr_pred=(CMT==5)*(cp);\nnlmixr_pred=(CMT==6)*(effect)+(1-((CMT==6)))*(nlmixr_pred);\ncmt(cp);\ncmt(pca);\nrx_yj_~2;\nrx_lambda_~1;\nrx_hi_~1;\nrx_low_~0;\nrx_pred_f_~nlmixr_pred;\nrx_pred_=nlmixr_pred;\nrx_r_=(nlmixrAdd)^2;\n")
    
    expect_error(rxS(.mv, TRUE, FALSE), NA)
  })
})

rxTest({
  ## else if is actually already supported...
  test_that("Conditional statements: else if, RxODE#54", {
    m <- rxode2({
      if (cnd <= 1) {
        a <- 1.0
      } else if (cnd <= 2) {
        a <- 2.0
      } else if (cnd <= 3) {
        a <- 3.
      } else {
        a <- 100
      }
      tmp <- cnd
    })

    ## The prefered syntax is only if / else but it still works...
    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 100)
  })

  test_that("Conditional statements: ifelse", {
    m <- rxode2({
      a <- ifelse(cnd <= 1, 1.0, ifelse(cnd <= 2, 2, ifelse(cnd <= 3, 3, 100)))
      tmp <- cnd
    })

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 100)
  })

  test_that("Conditional statements: embedded logical expressions", {
    m <- rxode2({
      a <- (cnd == 1) * 1.0 + (cnd == 2) * 2 + (cnd == 3) * 3
      tmp <- cnd
    })

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 0)
  })

  test_that("Conditional statements: ifelse with assignments", {
    m <- rxode2({
      ifelse(cnd <= 1, a = 1.0, a = 2.0)
      tmp <- cnd
    })

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    m <- rxode2({
      ifelse(cnd <= 1, a <- 1.0, a <- 2.0)
      tmp <- cnd
    })

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)
  })

  test_that("prune checks", {
    tmp <- "C2=centr/V;\nC3=peri/V2;\nd/dt(depot)=-KA*depot;\nd/dt(centr)=KA*depot-CL*C2-Q*C2+Q*C3;\nd/dt(peri)=Q*C2-Q*C3;\nC4=CMT;\nif(CMT==1){\nprd=depot;\n}\nif(CMT==2){\nprd=centr;\n}\nif(CMT==3){\nprd=peri;\n}\n"
    expect_equal(rxPrune(tmp), "C2=centr/V\nC3=peri/V2\nd/dt(depot)=-KA*depot\nd/dt(centr)=KA*depot-CL*C2-Q*C2+Q*C3\nd/dt(peri)=Q*C2-Q*C3\nC4=CMT\nprd=(CMT==1)*(depot)\nprd=(CMT==2)*(centr)+(1-((CMT==2)))*(prd)\nprd=(CMT==3)*(peri)+(1-((CMT==3)))*(prd)")

    ## Advanced # context pruining:
    m <- rxode2({
      if (cnd <= 1) {
        a <- 1.0
      } else if (cnd <= 2) {
        a <- 2.0
      } else if (cnd <= 3) {
        a <- 3.
      } else {
        a <- 100
      }
      tmp <- cnd
    })

    m <- rxode2(rxPrune(m))

    ## The prefered syntax is only if / else but it still works...
    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 100)

    m <- suppressMessages(rxode2(rxOptExpr(rxNorm(m))))

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 100)

    m <- rxode2({
      a <- 100
      if (cnd <= 1) {
        a <- 1.0
      }
      if (cnd > 1 && cnd <= 2) {
        a <- 2.0
      }
      if (cnd > 2 && cnd <= 3) {
        a <- 3.
      }
      tmp <- cnd
    })

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 100)

    m <- rxode2(rxPrune(m))

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 100)

    m <- suppressMessages(rxode2(rxOptExpr(rxNorm(m))))

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 100)

    m <- rxode2({
      a <- ifelse(cnd <= 1, 1.0, ifelse(cnd <= 2, 2, ifelse(cnd <= 3, 3, 100)))
      tmp <- cnd
    })

    m <- rxode2(rxPrune(m))

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 100)

    m <- suppressMessages(rxode2(rxOptExpr(rxNorm(m))))

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 100)


    m <- rxode2({
      ifelse(cnd <= 1, a = 1.0, a = 2.0)
      tmp <- cnd
    })

    m <- rxode2(rxPrune(m))

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    m <- suppressMessages(rxode2(rxOptExpr(rxNorm(m))))

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    m <- rxode2({
      ifelse(cnd <= 1, a <- 1.0, a <- 2.0)
      tmp <- cnd
    })

    m <- rxode2(rxPrune(m))

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    m <- suppressMessages(rxode2(rxOptExpr(rxNorm(m))))

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    m <- rxode2({
      a <- (cnd == 1) * 1.0 + (cnd == 2) * 2 + (cnd == 3) * 3
      tmp <- cnd
    })

    m <- rxode2(rxPrune(m))

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 0)

    m <- suppressMessages(rxode2(rxOptExpr(rxNorm(m))))

    tmp <- rxSolve(m, c(cnd = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 0)

    m <- rxode2({
      if (cnd <= 1) {
        a <- theta[1]
      } else if (cnd <= 2) {
        a <- 2.0
      } else if (cnd <= 3) {
        a <- 3.
      } else {
        a <- 100
      }
      tmp <- cnd
    })

    m <- suppressMessages(rxode2(rxOptExpr(rxPrune(m))))

    tmp <- rxSolve(m, c(cnd = 1, `THETA[1]` = 1), et(0.1))
    expect_equal(tmp$tmp, 1)
    expect_equal(tmp$a, 1)

    tmp <- rxSolve(m, c(cnd = 2, `THETA[1]` = 1), et(0.1))
    expect_equal(tmp$tmp, 2)
    expect_equal(tmp$a, 2)

    tmp <- rxSolve(m, c(cnd = 3, `THETA[1]` = 1), et(0.1))
    expect_equal(tmp$tmp, 3)
    expect_equal(tmp$a, 3)

    tmp <- rxSolve(m, c(cnd = 4, `THETA[1]` = 1), et(0.1))
    expect_equal(tmp$tmp, 4)
    expect_equal(tmp$a, 100)
  })

  test_that("cimet pruning checks", {
    cimet.1 <- rxode2({
      dose <- 300
      eta.ka <- 0
      eta.cl <- 0
      eta.v <- 0
      eta.tgap <- 0
      eta.rkeb <- 0
      add.err <- 0
      tka <- log(0.5)
      tcl <- log(60)
      tv <- log(25)
      ttgap <- log(2)
      trkeb <- log(0.5)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      tgap <- exp(ttgap + eta.tgap)
      rkeb <- exp(trkeb + eta.rkeb)
      #
      bile <- 1
      if (t < tgap) {
        bile <- 0
      }
      #
      ha <- exp(-(cl / v) * tgap) / ((cl / v) - ka)
      hb <- exp(-ka * tgap) * (cl / v) / ka / ((cl / v) - ka)
      tote <- ka * dose * (1 / ka + ha - hb)
      #
      hc <- exp(-(cl / v) * t) - exp(-ka * t)
      timh <- bile * (t - tgap)
      hd <- exp(-(cl / v) * timh) - exp(-ka * timh)
      #
      cp <- dose / v * ka / (ka - (cl / v)) * hc + bile * rkeb * tote / v * ka / (ka - (cl / v)) * hd
      #
      cp <- cp + add.err # + prop(prop.err)
    })

    et <- et(seq(0, 24, length.out = 90))

    s1 <- rxSolve(cimet.1, et)

    cimet.2 <- rxode2(rxPrune(cimet.1))

    s2 <- rxSolve(cimet.2, et)

    expect_equal(s1$cp, s2$cp)

    cimet.3 <- suppressMessages(rxode2(rxOptExpr(rxPrune(cimet.1))))

    s3 <- rxSolve(cimet.3, et)

    expect_equal(s1$cp, s3$cp)
  })
})

## Similar to default behavior according to R extension manual
files <- Sys.glob(paste0("*", SHLIB_EXT))
dest <- file.path(R_PACKAGE_DIR, paste0('libs', R_ARCH))
dir.create(dest, recursive = TRUE, showWarnings = FALSE)
file.copy(files, dest, overwrite = TRUE)
## for (f in files)
##     file.copy(f,file.path(dest,paste0("lib",f)),overwrite = TRUE)
if(file.exists("symbols.rds")) {
    file.copy("symbols.rds", dest, overwrite = TRUE)
}

## Add headers
incl <- file.path(R_PACKAGE_DIR, "include")
headers <- c(Sys.glob("*.h"), Sys.glob("*.h.gch"))
if (any(file.exists(headers))) {
    dir.create(incl, recursive = TRUE, showWarnings = FALSE)
    file.copy(headers, incl, overwrite = TRUE)
}

#' Message about fixing or unfixing a parameter
#'
#' @param ini this is the iniDf data frame
#' @param w this indicates the row number of the item that is fixed or
#'   unfixed
#' @param fixedValue this is a boolean
#' @noRd
#' @author Matthew L. Fidler
.msgFix<- function(ini, w, fixedValue) {
  lapply(w, function(.w) {
    if (ini$fix[.w] != fixedValue) {
      if (fixedValue) {
        .minfo(paste0("fix {.code ", ini$name[.w], "} to {.code ", ini$est[.w], "}"))
      } else {
        .minfo(paste0("unfix {.code ", ini$name[.w], "} keeping initial estimate {.code ", ini$est[.w], "}"))
      }
    }
  })
}

#' This modifies the iniDf to fix (or unfix) parameters and related
#' values
#'
#' Note that the block of etas will be fixed/unfixed when a single
#' value is fixed/unfixed
#'
#' @param ini iniDf data.frame
#' @param w which item will be fixed
#' @param fixedValue should this be fixed `TRUE` or unfixed `FALSE`
#' @return nothing, called for side effects
#' @noRd
#' @author Matthew L. Fidler
.iniModifyFixedForThetaOrEtablock <- function(ini, w, fixedValue) {
  if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
    .msgFix(ini, w, fixedValue)
  }
  ini$fix[w] <- fixedValue
  .neta <- ini$neta1[w]
  if (!is.na(.neta)) {
    .etas <- .neta
    .fixedEtas <- NULL
    while (length(.etas) > 0) {
      .neta <- .etas[1]
      w <- which(ini$neta1 == .neta | ini$neta2 == .neta)
      if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
        .msgFix(ini, w, fixedValue)
      }
      ini$fix[w] <- fixedValue
      .etas <- unique(c(.etas, ini$neta1[w], ini$neta2[w]))
      .fixedEtas <- c(.neta, .fixedEtas)
      .etas <- .etas[!(.etas %in% .fixedEtas)]
    }
  }
  ini
}

#' Modify the population estimate in the internal `iniDf` data.frame
#'
#' @param ini This is the data frame for modifying
#' @param lhs This is the left hand expression as a character
#' @param rhs This is the right handed expression
#' @param doFix Fix the estimation variable
#' @param doUnfix Unfix the estimation variable
#' @param maxLen The maximum length is either 3 or 1
#' @return Modified ini variable
#' @author Matthew L. Fidler
#' @noRd
.iniModifyThetaOrSingleEtaDf <- function(ini, lhs, rhs, doFix, doUnfix, maxLen) {
  .w <- which(ini$name == lhs)
  if (length(.w) != 1) {
    stop("cannot find parameter '", lhs, "'", call.=FALSE)
  }
  .curFix <- ini$fix[.w]
  if (doFix) {
    if (.curFix) {
      warning("trying to fix '", lhs, "', but already fixed",
              call.=FALSE)
    } else {
      ini <- .iniModifyFixedForThetaOrEtablock(ini, .w, TRUE)
    }
  } else if (doUnfix) {
    if (.curFix) {
      ini <- .iniModifyFixedForThetaOrEtablock(ini, .w, FALSE)
    } else {
      warning("trying to unfix '", lhs, "', but already unfixed",
              call.=FALSE)
    }
  }

  if (is.null(rhs)) {
  } else if (length(rhs) == 1)  {
    ini$est[.w] <- rhs
    if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
      .minfo(paste0("change initial estimate of {.code ", ini$name[.w], "} to {.code ", ini$est[.w], "}"))
    }
    .lower <- ini$lower[.w]
    .upper <- ini$upper[.w]
    if (.lower >= rhs) {
      ini$lower[.w] <- -Inf
      if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
        .minfo(paste0("lower bound of  {.code ", ini$name[.w], "} reset to {.code -Inf}"))
      }
    }
    if (.upper <= rhs) {
      ini$upper[.w] <- Inf
      if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
        .minfo(paste0("upper bound of  {.code ", ini$name[.w], "} reset to {.code Inf}"))
      }
    }
  } else {
    if (maxLen == 1) {
      stop("piping for '", lhs, "' failed, the estimate should only be 1 value",
           call.=FALSE)
    } else if (length(rhs) == 2) {
      ini$lower[.w] <- rhs[1]
      ini$est[.w] <- rhs[2]
      if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
        .minfo(paste0("change initial estimate (", ini$est[.w], ") and lower bound (", ini$lower[.w], ") of {.code ", ini$name[.w], "}"))
      }
      # now check/change upper if needed
      .upper <- ini$upper[.w]
      if (.upper <= rhs[1] || .upper <= rhs[2]) {
        ini$upper[.w] <- Inf
        if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
          .minfo(paste0("upper bound for initial estimate (", ini$name[.w], ") reset to Inf"))
        }
      }
    } else if (length(rhs) == 3) {
      ini$lower[.w] <- rhs[1]
      ini$est[.w] <- rhs[2]
      ini$upper[.w] <- rhs[3]
      if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
        .minfo(paste0("change initial estimate (", ini$est[.w], ") and upper/lower bound (", ini$lower[.w], " to ", ini$upper[.w], ") of {.code ", ini$name[.w], "}"))
      }
    }
  }
  ini
}

#' Handle a fix or unfixed single expressionon for population or single eta
#'
#' This updates the `iniDf` within the rxode2 UI object
#'
#' @param expr Single assignment expression
#' @param rxui rxode2 UI object
#' @param envir Environment where the evaulation occurs
#' @param maxLen Maximum length of the argument
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @noRd
.iniHandleFixOrUnfixEqual <- function(expr, rxui, envir=parent.frame(), maxLen=3L) {
  .tilde <- .isLotriAssignment(expr)
  .covs <- rxui$allCovs
  .lhs <- as.character(expr[[2]])
  .rhs <- expr[[3]]
  .doFix <- .doUnfix <- FALSE
  if (is.name(.rhs)) {
    if (identical(.rhs, quote(`fix`))) { # variations on fix are handled upstream
      .doFix <- TRUE
      .rhs <- NULL
    } else if (identical(.rhs, quote(`unfix`))) { # variations on unfix are handled upstream
      .doUnfix <- TRUE
      .rhs <- NULL
    }
  } else if (identical(.rhs[[1]], quote(`fix`))) {
    .doFix <- TRUE
    .rhs[[1]] <- quote(`c`)
  } else if (identical(.rhs[[1]], quote(`unfix`))) {
    .doUnfix <- TRUE
    .rhs[[1]] <- quote(`c`)
  }

  if (!is.null(.rhs)) {
    .rhs <- eval(.rhs, envir=envir)
    checkmate::assertNumeric(.rhs, any.missing=FALSE, min.len=1, max.len=3, .var.name=.lhs)
    if (!all(sort(.rhs) == .rhs)) {
      stop("the '", .lhs, "' piping lower, estimate, and/or upper estimate is in the wrong order",
           call.=FALSE)
    }
  }
  if (.lhs %in% .covs) {
    .addVariableToIniDf(.lhs, rxui, toEta=.tilde, value=.rhs, promote=TRUE)
    # assign is called again to handle the fixing of the variable
  }
  .ini <- rxui$ini
  if (is.null(.ini)) .ini <- rxui$iniDf
  assign("iniDf", .iniModifyThetaOrSingleEtaDf(.ini, .lhs, .rhs, .doFix, .doUnfix, maxLen=maxLen),
         envir=rxui)
  invisible()
}

#'  Add a covariance term between two eta values
#'
#' @param ini Data frame of initial estimates
#' @param neta1 Name of the first eta term
#' @param neta2 Name of the second eta term
#' @param est Estimate of the covariance
#' @param doFix Should this term be fixed
#' @param rxui is the rxui value
#' @return A modified (unsorted) data frame with the new covariance term appended
#' @author Matthew L. Fidler
#' @noRd
.iniAddCovarianceBetweenTwoEtaValues <- function(ini, neta1, neta2, est, doFix, rxui) {
  .covs <- rxui$allCovs
  if (neta1 %in% .covs) {
    .addVariableToIniDf(neta1, rxui, toEta=TRUE, value=NA, promote=TRUE)
    ini <- rxui$iniDf
    .covs <- rxui$allCovs
  }
  if (neta2 %in% .covs) {
    .addVariableToIniDf(neta2, rxui, toEta=TRUE, value=NA, promote=TRUE)
    ini <- rxui$iniDf
    .covs <- rxui$allCovs
  }
  .w1 <- which(ini$name == neta1)
  .w2 <- which(ini$name == neta2)
  if (length(.w1) != 1) stop("cannot find parameter '", neta1, "'", call.=FALSE)
  if (length(.w2) != 1) stop("cannot find parameter '", neta2, "'", call.=FALSE)
  if (ini$neta1[.w1] < ini$neta1[.w2]) {
    .tmp <- .w1
    .w1 <- .w2
    .w2 <- .tmp

    .tmp <- neta1
    neta1 <- neta2
    neta2 <- .tmp
  }
  .fix <- FALSE
  if (doFix) .fix <- TRUE
  .ini2 <- data.frame(ntheta= NA_integer_, neta1=ini$neta1[.w1], neta2=ini$neta1[.w2],
                      name=paste0("(", neta2, ",", neta1, ")"), lower= -Inf, est=est, upper=Inf,
                      fix=.fix, label=NA_character_, backTransform=NA_character_, condition="id",
                      err=NA_character_)
  if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
    .minfo(paste0("add covariance between {.code ", ini$name[.w1], "} and {.code ", ini$name[.w2], "} with initial estimate {.code ", est, "}"))
  }
  rbind(ini,.ini2)
}

#' This function handles the lotri process and integrates into current UI
#'
#' This will update the matrix and integrate the initial estimates in the UI
#'
#' @param mat Lotri processed matrix from the piping ini function
#'
#' @param rxui rxode2 UI function
#'
#' @return Nothing, called for side effects
#'
#' @author Matthew L. Fidler
#'
#' @noRd
.iniHandleLotriMatrix <- function(mat, rxui) {
  .dn <- dimnames(mat)[[1]]
  .iniDf <- rxui$iniDf
  .drop <- FALSE
  .common <- rxui$iniDf$name[rxui$iniDf$name %in% .dn]
  if (all(is.na(rxui$iniDf$neta1))) {
    .maxEta <- 0
    .shift <- 0
  } else {
    .maxEta <- max(rxui$iniDf$neta1, na.rm=TRUE)
    .shift <- .maxEta - length(.common)
  }
  .ini2 <- NULL
  for (.i in seq_along(.dn)) {
    .n <- .dn[.i]
    .w <- which(.iniDf$name == .n)
    if (length(.w) == 1) {
      .oNum <- .iniDf$neta1[.w]
      .w2 <- which(.iniDf$neta1 == .oNum & .iniDf$neta2 != .oNum)
      .df1 <- .iniDf[.w, ]
      .df1$neta1 <- .i + .shift
      .df1$neta2 <- .i + .shift
      .ini2 <- c(.ini2, list(.df1))
      if (length(.w2) > 0) {
        .iniDf <- .iniDf[-.w2, ]
        .drop <- TRUE
      }
    }
  }
  if (isTRUE(getOption("rxode2.verbose.pipe", TRUE)) && .drop) {
    .minfo(paste0("some correlations may have been dropped for the variables: {.code ", paste(.dn, collapse="}, {.code "), "}"))
    .minfo("the piping should specify the needed covariances directly")
  }
  .dfTheta <- .iniDf[is.na(.iniDf$neta1), ]
  .dfEta <- .iniDf[!is.na(.iniDf$neta1), ]
  .dfEta <- .dfEta[!(.dfEta$name %in% .dn),, drop = FALSE]
  if (length(.dfEta$neta1) > 0) {
    .dfEta$neta1 <- factor(paste(.dfEta$neta1))
    .dfEta$neta2 <- factor(paste(.dfEta$neta2), levels=levels(.dfEta$neta1))
    .dfEta$neta1 <- as.integer(.dfEta$neta1)
    .dfEta$neta2 <- as.integer(.dfEta$neta2)
  }
  .iniDf <- do.call("rbind", c(list(.dfTheta), list(.dfEta), .ini2))
  assign("iniDf", .iniDf, envir=rxui)
  .covs <- rxui$allCovs
  .fixMatrix <- attr(mat, "lotriFix")
  .unfixMatrix <- attr(mat, "lotriUnfix")
  .n <- dimnames(mat)[[1]]
  .mat <- mat
  if (!inherits(.mat, "lotriFix"))
    class(.mat) <- c("lotriFix", class(.mat))
  .df <- as.data.frame(.mat)
  .lastFix <- FALSE
  for (i in seq_along(.df$neta1)) {
    if (!is.na(.df$neta1[i])) {
      .doFix <- FALSE
      .doUnfix <- FALSE
      if (!is.null(.fixMatrix) && .df$fix[i]) {
        .doFix <- TRUE
      }
      if (!is.null(.unfixMatrix) && !.df$fix[i]) {
        .doUnfix <- TRUE
      }
      if (.df$neta1[i] == .df$neta2[i]) {
        .var <- as.character(.df$name[i])
        if (.var %in% .covs) {
          .addVariableToIniDf(.var, rxui, toEta=TRUE, value=.df$est[i], promote=TRUE)
          .covs <- rxui$allCovs
        }
        assign("iniDf", .iniModifyThetaOrSingleEtaDf(rxui$iniDf, .var, .df$est[i], .doFix, .doUnfix, 1L),
               envir=rxui)
        .lastFix <- rxui$iniDf$fix[rxui$iniDf$name == .var]
      } else {
        .n1 <- paste0("(", .n[.df$neta1[i]], ",", .n[.df$neta2[i]], ")")
        assign("iniDf", .iniAddCovarianceBetweenTwoEtaValues(rxui$iniDf, .n[.df$neta1[i]], .n[.df$neta2[i]], .df$est[i],
                                                             .lastFix, rxui),
               envir=rxui)
        .covs <- rxui$allCovs
      }
    }
  }
}

# Determine if the input is an endpoint by being 3 long and the call part being
# a tilde
.isLotriAssignment <- function(expr) {
  .matchesLangTemplate(expr, str2lang(". ~ ."))
}

#' Modify the label in an iniDf
#'
#' @inheritParams .iniHandleLine
#' @return Nothing, called for side effects
#' @author Bill Denney & Matthew Fidler
#' @keywords internal
#' @noRd
.iniHandleLabel <- function(expr, rxui, envir) {
  .lhs <- as.character(expr[[2]])
  .newLabel <- expr[[3]][[2]]
  .ini <- rxui$ini
  .w <- which(.ini$name == .lhs)
  if (length(.w) != 1) {
    stop("cannot find parameter '", .lhs, "'", call.=FALSE)
  } else if (is.null(.newLabel)) {
    .newLabel <- NA_character_
  } else if (!is.character(.newLabel) || !(length(.newLabel) == 1)) {
    stop("the new label for '", .lhs, "' must be a character string",
         call.=FALSE)
  }
  .ini$label[.w] <- .newLabel
  assign("iniDf", .ini, envir=rxui)
  invisible()
}
#' This handles the backTransform() piping calls
#'
#' @param expr expression for backTransform() in `ini()` piping
#' @param rxui rxode2 ui function
#' @param envir evaluation environment
#' @return nothing, called for side effects
#' @noRd
#' @author Matthew L. Fidler
.iniHandleBackTransform <- function(expr, rxui, envir) {
  .lhs <- as.character(expr[[2]])
  .newExpr <- expr[[3]][[2]]
  .ini <- rxui$ini
  .w <- which(.ini$name == .lhs)
  .good <- TRUE
  if (length(.w) != 1) {
    stop("cannot find parameter '", .lhs, "'", call.=FALSE)
  } else if (is.null(.newExpr)) {
    .newExpr <- NA_character_
  } else if (checkmate::testCharacter(.newExpr, len=1, any.missing=FALSE,
                                      pattern="^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
                                      min.chars = 1)) {
  } else {
    .newExpr <- deparse1(.newExpr)
    if (!checkmate::testCharacter(.newExpr, len=1, any.missing=FALSE,
                                 pattern="^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
                                 min.chars = 1)) {
      .good <- FALSE
    }
  }
  if (!.good) {
    stop("backTransform specification malformed",
         call.=FALSE)
  }
  if (!is.na(.newExpr)) {
    if (!exists(.newExpr, envir=envir, mode="function")) {
      stop("tried use a backTransform(\"", .newExpr, "\") when the function does not exist",
           call.=FALSE)
    }
  }
  .ini$backTransform[.w] <- .newExpr
  assign("iniDf", .ini, envir=rxui)
  invisible()
}

#' Reorder rows in iniDf
#'
#' @inheritParams .iniHandleLine
#' @param append Reorder theta parameters.  \code{NULL} means no change to
#'   parameter order.  A parameter name (as a character string) means to put the
#'   new parameter after the named parameter.  A number less than or equal to
#'   zero means to put the parameter at the beginning of the list.  A number
#'   greater than the last parameter number means to put the parameter at the
#'   end of the list.
#' @return Nothing, called for side effects
#' @keywords internal
.iniHandleAppend <- function(expr, rxui, envir, append) {
  ini <- rxui$ini
  if (is.null(append)) {
    # Do nothing
    return()
  } else if (is.logical(append)) {
    checkmate::assertLogical(append, any.missing = FALSE, len = 1)
    if (isTRUE(append)) {
      appendClean <- Inf
    } else if (isFALSE(append)) {
      appendClean <- 0
    }
  } else if (is.numeric(append)) {
    checkmate::assertNumber(append, null.ok = FALSE, na.ok = FALSE)
    appendClean <- append
  } else if (is.character(append)) {
    checkmate::assertCharacter(append, any.missing = FALSE, len = 1, null.ok = FALSE)
    checkmate::assertChoice(append, choices = ini$name)
    appendClean <- which(ini$name == append)
  } else {
    stop("'append' must be NULL, logical, numeric, or character/expression of variable in model",
         call. = FALSE)
  }

  lhs <- as.character(expr[[2]])
  wLhs <- which(ini$name == lhs)
  if (length(wLhs) != 1) {
    stop("cannot find parameter '", lhs, "'", call.=FALSE)
  } else if (length(appendClean) != 1) {
    # This likely cannot be reached because all scenarios should be handled
    # above in the input checking.  The line remains in the code defensively.
    stop("Cannot find parameter '", append, "'", call.=FALSE) # nocov
  } else if (appendClean == wLhs) {
    warning("parameter '", lhs, "' set to be moved after itself, no change in order made",
            call. = FALSE)
    return()
  } else if (is.na(ini$ntheta[wLhs])) {
    stop("only theta parameters can be moved.  '", lhs, "' is not a theta parameter",
         call. = FALSE)
  }

  # Do the movement
  if (appendClean <= 0) {
    # put it at the top
    ret <- rbind(ini[wLhs, ], ini[-wLhs, ])
  } else if (appendClean >= nrow(ini)) {
    # put it at the bottom
    ret <- rbind(ini[-wLhs, ], ini[wLhs, ])
  } else {
    # put it in the middle
    rowsAbove <- setdiff(seq_len(appendClean), wLhs)
    rowsBelow <- setdiff(seq(appendClean + 1, nrow(ini)), wLhs)
    ret <- rbind(ini[rowsAbove, ], ini[wLhs, ], ini[rowsBelow, ])
  }
  # Ensure that ntheta stays in order
  ini$ntheta[!is.na(ini$ntheta)] <- seq_len(sum(!is.na(ini$ntheta)))
  assign("iniDf", ret, envir=rxui)
  invisible()
}

.iniHandleRecalc <- function(rxui) {
  .fun <- rxUiDecompress(rxui$fun())
  for (.i in ls(.fun, all.names=TRUE)) {
    if (.i != "meta") {
      assign(.i, get(.i, envir=.fun), envir=rxui)
    }
  }
  invisible()
}

#' Handle switching theta to eta and vice versa
#'
#' This is coded as model |> ini(~par)
#'
#' @param expr Expression, this would be the ~par expression
#' @param rxui rxui uncompressed environment
#' @param envir Environment for evaluation (if needed)
#' @return Nothing, called for side effects
#' @noRd
#' @author Matthew L. Fidler
.iniHandleSwitchType <- function(expr, rxui, envir=parent.frame()) {
  .var <- as.character(expr[[2]])
  .iniDf <- rxui$iniDf
  .w <- which(.iniDf$name == .var)
  if (length(.w) != 1L) stop("cannot switch parameter type for '", .var, "'", call.=FALSE)
  .theta <- .iniDf[!is.na(.iniDf$ntheta),, drop = FALSE]
  .eta <- .iniDf[is.na(.iniDf$ntheta),, drop = FALSE]
  if (is.na(.iniDf$ntheta[.w])) {
    # switch eta to theta
    .neta <- .iniDf$neta1[.w]
    .eta <- .eta[.eta$neta1 != .neta,, drop = FALSE]
    .eta <- .eta[.eta$neta2 != .neta,, drop = FALSE]
    .eta$neta1 <- .eta$neta1 - ifelse(.eta$neta1 < .neta, 0L, 1L)
    .eta$neta2 <- .eta$neta2 - ifelse(.eta$neta2 < .neta, 0L, 1L)
    .newTheta <- .iniDf[.w, ]
    .newTheta$neta1 <- NA_integer_
    .newTheta$neta2 <- NA_integer_
    if (length(.theta$ntheta) == 0L) {
      .newTheta$ntheta <- 1L
    } else {
      .newTheta$ntheta <- max(.theta$ntheta) + 1L
    }
    .minfo(paste0("convert '", .var, "' from between subject variability to population parameter"))
    .theta <- rbind(.theta, .newTheta)
  } else {
    # switch theta to eta
    if (!is.na(.iniDf$err[.w])) {
      stop("cannot switch error parameter '", .var,
           "' to a different type", call. = FALSE)
    }
    .ntheta <- .iniDf$ntheta[.w]
    .theta <- .theta[.theta$ntheta != .ntheta,, drop = FALSE]
    .theta$ntheta <- .theta$ntheta - ifelse(.theta$ntheta < .ntheta, 0L, 1L)
    .newEta <- .iniDf[.w, ]
    .newEta$ntheta <- NA_integer_
    if (length(.eta$neta1) == 0L) {
      .newEta$neta1 <- .newEta$neta2 <- 1L
    } else {
      .newEta$neta1 <- .newEta$neta2 <- max(.eta$neta1) + 1L
    }
    .minfo(paste0("convert '", .var, "' from population parameter to between subject variability"))
    if (.newEta$est == 0) {
      .minfo("old initial estimate is zero, changing to 1")
      .newEta$est <- 1
    } else if (.newEta$est < 0) {
      .minfo("old initial estimate was negative, changing to positive")
      .newEta$est <- -.newEta$est
    }
    .newEta$lower <- -Inf
    .newEta$upper <- Inf
    .newEta$condition <- "id"
    .eta <- rbind(.eta, .newEta)
  }
  .ini <- rbind(.theta, .eta)
  assign("iniDf", .ini, envir=rxui)
  .iniHandleRecalc(rxui)
  invisible()
}

#' Handle dropping parameter and treating as if it is a covariate
#'
#' This is coded as model |> ini(-par)
#'
#' @param expr Expression, this would be the ~par expression
#' @param rxui rxui uncompressed environment
#' @param envir Environment for evaluation (if needed)
#' @return Nothing, called for side effects
#' @noRd
#' @author Matthew L. Fidler
.iniHandleDropType <- function(expr, rxui, envir=parent.frame()) {
  .var <- as.character(expr[[2]])
  .iniDf <- rxui$iniDf
  .w <- which(.iniDf$name == .var)
  if (length(.w) != 1L) stop("no initial estimates for '", .var, "', cannot change to covariate", call.=FALSE)
  .theta <- .iniDf[!is.na(.iniDf$ntheta),, drop = FALSE]
  .eta <- .iniDf[is.na(.iniDf$ntheta),, drop = FALSE]
  if (is.na(.iniDf$ntheta[.w])) {
    .minfo(paste0("changing between subject variability parameter '", .var, "' to covariate parameter"))
    .neta <- .iniDf$neta1[.w]
    .eta <- .eta[.eta$neta1 != .neta,, drop = FALSE]
    .eta <- .eta[.eta$neta2 != .neta,, drop = FALSE]
    .eta$neta1 <- .eta$neta1 - ifelse(.eta$neta1 < .neta, 0L, 1L)
    .eta$neta2 <- .eta$neta2 - ifelse(.eta$neta2 < .neta, 0L, 1L)
  } else {
    if (!is.na(.iniDf$err[.w])) {
      stop("cannot switch error parameter '", .var,
           "' to a covariate", call. = FALSE)
    }
    .minfo(paste0("changing population parameter '", .var, "' to covariate parameter"))
    .ntheta <- .iniDf$ntheta[.w]
    .theta <- .theta[.theta$ntheta != .ntheta,, drop = FALSE]
    .theta$ntheta <- .theta$ntheta - ifelse(.theta$ntheta < .ntheta, 0L, 1L)
  }
  .ini <- rbind(.theta, .eta)
  assign("iniDf", .ini, envir=rxui)
  # This will change covariates, recalculate everything
  .iniHandleRecalc(rxui)
  invisible()
}

#' Update the iniDf of a model
#'
#' @param expr Expression for parsing
#' @param rxui User interface function
#' @param envir Environment for parsing
#' @inheritParams .iniHandleAppend
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
.iniHandleLine <- function(expr, rxui, envir=parent.frame(), append = NULL) {
  if (.matchesLangTemplate(expr, str2lang("~diag()"))) {
    .iniHandleDiag(expr=NULL, rxui=rxui)
    return(invisible())
  } else if (length(expr) == 2L &&
               identical(expr[[1]], quote(`~`)) &&
               is.call(expr[[2]]) && length(expr[[2]]) >= 2L &&
               identical(expr[[2]][[1]], quote(`diag`))) {
    # .matchesLangTemplate(expr, str2lang("~diag(.)")) doesn't work
    .iniHandleDiag(expr=expr, rxui=rxui)
    return(invisible())
  }
  # Convert all variations on fix, fixed, FIX, FIXED; unfix, unfixed, UNFIX,
  # UNFIXED to fix and unfix to simplify all downstream operations
  expr <- .iniSimplifyFixUnfix(expr)
  # Convert assignment equal ("=") to left arrows ("<-") to simplify all
  # downstream operations
  expr <- .iniSimplifyAssignArrow(expr)

  if (.matchesLangTemplate(expr, str2lang(".name <- NULL")) ||
        .matchesLangTemplate(expr, str2lang(".name ~ NULL")) ||
        .matchesLangTemplate(expr, str2lang("cov(.name, .name) <- NULL")) ||
        .matchesLangTemplate(expr, str2lang("cor(.name, .name) <- NULL")) ||
        .matchesLangTemplate(expr, str2lang("cov(.name, .name) ~ NULL")) ||
        .matchesLangTemplate(expr, str2lang("cor(.name, .name) ~ NULL"))) {
    expr <- as.call(list(quote(`-`), expr[[2]]))
  }

  # now handle dropping covariances
  if (.matchesLangTemplate(expr, str2lang("-cov(.name, .name)")) ||
        .matchesLangTemplate(expr, str2lang("-cor(.name, .name)"))) {
    .iniHandleRmCov(expr=expr, rxui=rxui)
    return(invisible())
  }

  # Convert fix(name) or unfix(name) to name <- fix or name <- unfix
  if (.matchesLangTemplate(expr, str2lang("fix(.name)"))) {
    expr <- as.call(list(quote(`<-`), expr[[2]], quote(`fix`)))
  } else if (.matchesLangTemplate(expr, str2lang("unfix(.name)"))) {
    expr <- as.call(list(quote(`<-`), expr[[2]], quote(`unfix`)))
  }
  if (.matchesLangTemplate(expr, str2lang(".name <- label(.)"))) {
    .iniHandleLabel(expr=expr, rxui=rxui, envir=envir)
  } else if (.matchesLangTemplate(expr, str2lang(".name <- backTransform(.)"))) {
    .iniHandleBackTransform(expr=expr, rxui=rxui, envir=envir)
  } else if (.isAssignment(expr) && is.character(expr[[3]])) {
    stop(
      sprintf(
        "to assign a new label, use '%s <- label(\"%s\")'",
        as.character(expr[[2]]), expr[[3]]
      ), call.=FALSE
    )
  } else if (.isAssignment(expr)) {
    .iniHandleFixOrUnfixEqual(expr=expr, rxui=rxui, envir=envir)
  } else if (.isLotriAssignment(expr)) {
    .rhs <- expr[[2]]
    if (length(.rhs) > 1) {
      if (identical(.rhs[[1]], quote(`+`))) {
        .iniHandleLotriMatrix(eval(as.call(list(quote(`lotri`), as.call(list(quote(`{`), expr)))),
                                   envir=envir),
                              rxui)
        return(invisible())
      }
    }
    expr[[3]] <- eval(as.call(list(quote(`lotri`), as.call(list(quote(`{`), expr)))),
                      envir=envir)[1, 1]
    .iniHandleFixOrUnfixEqual(expr=expr, rxui=rxui, envir=envir, maxLen=1L)
  } else if (.isTildeExpr(expr)) {
    .iniHandleSwitchType(expr=expr, rxui=rxui, envir=envir)
  } else if (.isIniDropExpression(expr)) {
    .iniHandleDropType(expr=expr, rxui=rxui, envir=envir)
  } else {
    # Can this error be improved to clarify what is the expression causing the
    # issue?  It needs a single character string representation of something
    # that is not a character string.
    stop("invalid expr for ini() modification", call.=FALSE)
  }

  # (Maybe) update parameter order; this must be at the end so that the
  # parameter exists in case it is promoted from a covariate
  .iniHandleAppend(expr = expr, rxui = rxui, envir = envir, append = append)

  # now take out ETAs that no longer exist
  .iniDf <- get("iniDf", envir=rxui)
  .w <- which(is.na(.iniDf$neta1) & !is.na(.iniDf$neta2))
  .reassign <- FALSE
  if (length(.w) > 0) {
    .iniDf <- .iniDf[-.w, ]
    .reassign <- TRUE
  }
  .iniDf <- get("iniDf", envir=rxui)
  .w <- which(!is.na(.iniDf$neta1) & is.na(.iniDf$neta2))
  if (length(.w) > 0) {
    .iniDf <- .iniDf[-.w, ]
    .reassign <- TRUE
  }
  if (.reassign) {
    assign("iniDf", .iniDf, envir=rxui)
  }
}

# TODO: while nlmixr2est is changed
#' @rdname dot-iniHandleLine
#' @export
.iniHandleFixOrUnfix <- .iniHandleLine

#' Simplify variants of fix and unfix to just those two
#'
#' @param expr An R call or similar object
#' @return \code{expr} where all variants of fix (fixed, FIX, FIXED) and unfix
#'   (unfixed, UNFIX, and UNFIXED) are converted to fix and unfix
#' @noRd
.iniSimplifyFixUnfix <- function(expr) {
  if (identical(expr, as.name("fixed")) ||
      identical(expr, as.name("FIX")) ||
      identical(expr, as.name("FIXED"))) {
    expr <- as.name("fix")
  } else if (identical(expr, as.name("unfixed")) ||
             identical(expr, as.name("UNFIX")) ||
             identical(expr, as.name("UNFIXED"))) {
    expr <- as.name("unfix")
  } else if (is.call(expr)) {
    for (idx in seq_along(expr)) {
      # Do not perform a NULL assignment so that NULL comes out of the result
      if (!is.null(expr[[idx]])) {
        expr[[idx]] <- .iniSimplifyFixUnfix(expr[[idx]])
      }
    }
  } else {
    # do nothing
  }
  expr
}

#' Simplify all assignments to be left arrows (and not equal signs)
#'
#' @param expr An R call or similar object
#' @return \code{expr} where all variants assignment equal signs are converted
#'   to \code{<-}
#' @noRd
.iniSimplifyAssignArrow <- function(expr) {
  if (is.call(expr) && length(expr) == 3) {
    if (expr[[1]] == as.name("=")) {
      expr[[1]] <- as.name("<-")
    }
  } else {
    # do nothing
  }
  expr
}
#' This gets the append arg for the ini({}) piping
#'
#' @param f this is the `try(force(append))` argument,
#' @param s this is the `as.character(substitute(append))` argument
#' @return corrected ini piping argument
#'
#' This is exported for creating new ini methods that have the same
#' requirements for piping
#'
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.iniGetAppendArg <- function(f, s) {
  if (inherits(f, "try-error") &&
        checkmate::testCharacter(s, len=1, any.missing=FALSE,
                                 pattern="^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
                                 min.chars = 1)) {
    return(s)
  }
  if (is.null(f)) {
    return(NULL)
  } else if (checkmate::testCharacter(f, len=1, any.missing=FALSE,
                                      pattern="^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
                                      min.chars = 1)) {
    return(f)
  } else if (is.infinite(f)) {
    return(f)
  } else if (checkmate::testIntegerish(f, len=1, any.missing=FALSE)) {
    if (f < 0) {
      stop("'append' cannot be a negative integer", call.=FALSE)
    }
    return(f)
  } else if (checkmate::testLogical(f, len=1)) {
    # NA for model piping prepends
    if (is.na(f)) return(FALSE)
    return(f)
  }
  stop("'append' must be NULL, logical, numeric, or character/expression of variable in model",
       call.=FALSE)
}

#' @export
#' @rdname ini
ini.rxUi <- function(x, ..., envir=parent.frame(), append = NULL) {
  .s  <- as.character(substitute(append))
  .f <- try(force(append), silent=TRUE)
  append <- .iniGetAppendArg(.f, .s)
  .ret <- rxUiDecompress(.copyUi(x)) # copy so (as expected) old UI isn't affected by the call
  .iniDf <- .ret$iniDf
  .iniLines <- .quoteCallInfoLines(match.call(expand.dots = TRUE)[-(1:2)], envir=envir, iniDf= .iniDf)
  if (length(.iniLines) == 0L) return(.ret$iniFun)
  lapply(.iniLines, function(line) {
    .iniHandleLine(expr = line, rxui = .ret, envir = envir, append=append)
  })
  if (inherits(x, "rxUi")) {
    .x <- rxUiDecompress(x)
    .ret <- .newModelAdjust(.ret, .x)
  }
  .ret <- rxUiCompress(.ret)
  if (inherits(x, "rxUi")) {
    .cls <- setdiff(class(x), class(.ret))
    if (length(.cls) > 0) {
      class(.ret) <- c(.cls, class(.ret))
    }
  }
  .ret
}

#' @rdname ini
#' @export
ini.default <- function(x, ..., envir=parent.frame(), append = NULL) {
  .s  <- as.character(substitute(append))
  .f <- try(force(append), silent=TRUE)
  append <- .iniGetAppendArg(.f, .s)
  .ret <- try(as.rxUi(x), silent = TRUE)
  .hasUi <- TRUE
  .err <- FALSE
  if (inherits(.ret, "try-error")) {
    .ret <- as.ini(x)
    if (is.call(.ret) && identical(.ret[[1]], quote(`ini`))) {
      lotri <- lotri::lotri
      .ret[[1]] <- quote(`lotri`)
      .lotri <- try(eval(.ret), silent = TRUE)
      if (inherits(.lotri, "try-error")) {
        .err <- TRUE
      } else if (!inherits(.lotri, "lotriFix")) {
        class(.lotri) <- c("lotriFix", class(.lotri))
      }
      if (!.err) {
        .iniDf <- as.data.frame(.lotri)
        .hasUi <- FALSE
      }
    } else {
      .err <- TRUE
    }
  } else {
    .ret <- rxUiDecompress(.ret)
    .iniDf <- .ret$iniDf
  }
  if (.err) {
    stop("cannot figure out what to do with the ini({}) function", call.=FALSE)
  }
  .iniLines <- .quoteCallInfoLines(match.call(expand.dots = TRUE)[-(1:2)], envir=envir, iniDf = .iniDf)
  if (length(.iniLines) == 0L) {
    if (.hasUi) {
      return(.ret$iniFun)
    } else {
      return(x)
    }
  }
  if (!.hasUi) {
    .ret <- new.env(parent=emptyenv())
    .ret$iniDf <- .iniDf

  }
  lapply(.iniLines, function(line) {
    .iniHandleLine(expr = line, rxui = .ret, envir=envir, append = append)
  })
  if (.hasUi) {
    rxUiCompress(.ret)
  } else {
    .ret <- as.ini(.ret$iniDf)
    if (inherits(x, "lotriFix") ||
          inherits(x, "matrix")) {
      .ret[[1]] <- quote(`lotri`)
      .ret <- try(eval(.ret), silent = TRUE)
    }
    .ret
  }
}

as.lotri.call <- function(x, ..., default = "") {
  lotri <- lotri::lotri
  .ret <- x
  if (identical(.ret[[1]], quote(`ini`))) {
    .ret[[1]] <- quote(`lotri`)
  }
  if (identical(.ret[[1]], quote(`lotri`))) {
    return(eval(.ret))
  } else {
    stop("unsupported object of class 'call' used with `as.lotri`")
  }
}

#' This tells if the line is modifying an estimate instead of a line of the model
#'
#' @param line Quoted line
#' @param rxui rxode2 UI object
#' @return Boolean indicating if the line defines an `ini` change.
#' @author Matthew L. Fidler
#' @noRd
.isQuotedLineRhsModifiesEstimates <- function(line, rxui) {
  if (length(line) != 3) return(FALSE)
  .rhs <- line[[2]]
  if (length(.rhs) > 1) {
    if (identical(.rhs[[1]], quote(`+`))) {
      return(TRUE)
    }
  }
  .c <- as.character(.rhs)
  if (any(rxui$iniDf$name == .c)) return(TRUE)
  return(FALSE)
}

#' Set random effects and residual error to zero
#'
#' @param object The model to modify
#' @param which The types of parameters to set to zero
#' @param fix Should the parameters be fixed to the zero value?
#' @return The `object` with some parameters set to zero
#' @family Initial conditions
#' @author Bill Denney
#' @examples
#' one.compartment <- function() {
#'   ini({
#'     tka <- log(1.57); label("Ka")
#'     tcl <- log(2.72); label("Cl")
#'     tv <- log(31.5); label("V")
#'     eta.ka ~ 0.6
#'     eta.cl ~ 0.3
#'     eta.v ~ 0.1
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl + eta.cl)
#'     v <- exp(tv + eta.v)
#'     d/dt(depot) = -ka * depot
#'     d/dt(center) = ka * depot - cl / v * center
#'     cp = center / v
#'     cp ~ add(add.sd)
#'   })
#' }
#' zeroRe(one.compartment)
#' @export
zeroRe <- function(object, which = c("omega", "sigma"), fix = TRUE) {
  object <- assertRxUi(object)
  which <- match.arg(which, several.ok = TRUE)
  .ret <- rxUiDecompress(.copyUi(object)) # copy so (as expected) old UI isn't affected by the call
  iniDf <- .ret$iniDf
  # In the code below there is no test for bounds since the bounds are typically (0, Inf).
  if ("omega" %in% which) {
    maskOmega <- !is.na(iniDf$neta1)
    if (sum(maskOmega) == 0) {
      cli::cli_warn("No omega parameters in the model")
    } else {
      iniDf$est[maskOmega] <- 0
      if (fix) {
        iniDf$fix[maskOmega] <- TRUE
      }
    }
  }
  if ("sigma" %in% which) {
    maskSigma <- !is.na(iniDf$err)
    if (sum(maskSigma) == 0) {
      cli::cli_warn("No sigma parameters in the model")
    } else {
      iniDf$est[maskSigma] <- 0
      maskLowerBound <- maskSigma & iniDf$lower > 0
      if (any(maskLowerBound)) {
        iniDf$lower[maskLowerBound] <- 0
      }
      if (fix) {
        iniDf$fix[maskSigma] <- TRUE
      }

    }
  }
  ini(.ret) <- iniDf
  .ret
}

#' This removes the off-diagonal BSV from a rxode2 iniDf
#'
#' @param ui rxode2 ui model
#'
#' @param diag character vector of diagonal values to remove
#'
#' @return iniDf with modified diagonal
#' @noRd
#' @author Matthew L. Fidler
.iniDfRmDiag <- function(iniDf, diag=character(0)) {
  .iniDf <- iniDf
  .theta <- .iniDf[!is.na(.iniDf$ntheta),,drop=FALSE]
  .eta <- .iniDf[is.na(.iniDf$ntheta),,drop=FALSE]
  if (length(diag) == 0) {
    .w <- which(.eta$neta1 == .eta$neta2)
    .rmNames <- .eta[-.w, "name"]
    .eta <- .eta[.w,, drop=FALSE]
    .iniDf <- rbind(.theta, .eta)
  } else {
    .rmNames <- character(0)
    for (.e in diag) {
      .w <- which(.eta$name == .e)
      if (length(.w) == 1L) {
        .n <- .eta$neta1[.w]
        .w <- vapply(seq_along(.eta$neta1),
                     function(i) {
                       if (.eta$neta1[i] == .eta$neta2[i]) {
                         TRUE
                       } else if (.eta$neta1[i] == .n && .eta$neta2[i] != .n) {
                         FALSE
                       } else if (.eta$neta2[i] == .n && .eta$neta1[i] != .n) {
                         FALSE
                       } else {
                         TRUE
                       }
                     }, logical(1), USE.NAMES = TRUE)
        .rmNames <- c(.rmNames, .eta$name[!.w])
        .eta <- .eta[.w,,drop=FALSE]
      } else {
        stop("cannot find parameter '", .e, "' for covariance removal", call.=FALSE)
      }
    }
    .mat <- lotri::as.lotri(.eta)
    .mat <- lotri::rcm(.mat)
    class(.mat) <- c("lotriFix", class(.mat))
    .eta <- as.data.frame(.mat)
    .eta$err <- NA_character_
    .iniDf <- rbind(.theta, .eta)
  }
  if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
    for (.v in .rmNames) {
      .minfo(paste0("remove covariance {.code ", .v, "}"))
    }
  }
  .iniDf
}

.iniHandleRmCov <- function(expr, rxui) {
  .iniDf <- rxui$iniDf
  .theta <- .iniDf[!is.na(.iniDf$ntheta),, drop = FALSE]
  .eta <- .iniDf[is.na(.iniDf$ntheta),, drop = FALSE]
  .mat <- lotri::as.lotri(.eta)
  .n1 <- as.character(expr[[2]][[2]])
  .v1 <- which(.n1==dimnames(.mat)[[1]])
  if (length(.v1) != 1) {
    stop("cannot find parameter '", .n1, "' for covariance removal", call.=FALSE)
  }
  .n2 <- as.character(expr[[2]][[3]])
  .v2 <- which(.n2==dimnames(.mat)[[1]])
  if (length(.v2) != 1) {
    stop("cannot find parameter '", .n2, "' for covariance removal", call.=FALSE)
  }
  if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
    .minfo(paste0("remove covariance {.code (", .n1, ", ", .n2, ")}"))
  }

  .mat[.v1, .v2] <- .mat[.v2, .v1] <- 0
  .mat <- lotri::rcm(.mat)
  class(.mat) <- c("lotriFix", class(.mat))
  .eta <- as.data.frame(.mat)
  .eta$err <- NA_character_
  .iniDf <- rbind(.theta, .eta)
  assign("iniDf", .iniDf, envir=rxui)
}

.iniHandleDiag <- function(expr, rxui){
  if (is.null(expr)) {
    assign("iniDf", .iniDfRmDiag(rxui$iniDf), envir=rxui)
  } else {
    # now get the variables in the diag expression
    .env <- new.env(parent=emptyenv())
    .env$names <- character(0)
    .f <- function(x) {
      if (is.name(x)) {
        .env$names <- c(.env$names, as.character(x))
      } else if (is.call(x)) {
        lapply(lapply(seq_along(x)[-1], function(i) {x[[i]]}), .f)
      }
    }
    expr <- expr[[2]]
    lapply(seq_along(expr)[-1],
           function(i) {
              .f(expr[[i]])
           })
    assign("iniDf", .iniDfRmDiag(rxui$iniDf, .env$names), envir=rxui)
  }
}

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("back transformation piping", {

      mod1 <- function() {
        ini({
          # central
          KA <- 2.94E-01
          backTransform("exp")
          CL <- 1.86E+01
          V2 <- 4.02E+01
          # peripheral
          Q <- 1.05E+01
          V3 <- 2.97E+02
          # effects
          Kin <- 1
          Kout <- 1
          EC50 <- 200
        })
        model({
          C2 <- centr/V2
          C3 <- peri/V3
          d/dt(depot) <- -KA*depot
          d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
          d/dt(peri)  <- Q*C2 - Q*C3
          eff(0) <- 1
          d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
        })
      }

      ui <- rxode(mod1)

      expect_equal(ui$iniDf$backTransform[ui$iniDf$name == "KA"], "exp")

      p1 <- ui |>
        ini(
          KA <- backTransform("log")
        )

      expect_equal(p1$iniDf$backTransform[ui$iniDf$name == "KA"], "log")

      p2 <-ui |>
        ini(
          KA <- backTransform(log)
        )

      expect_equal(p2$iniDf$backTransform[ui$iniDf$name == "KA"], "log")

      p3 <- ui |>
        ini(KA <- backTransform(NULL))

      expect_equal(p3$iniDf$backTransform[ui$iniDf$name == "KA"], NA_character_)

      expect_error(ui |>
                     ini(KA <- backTransform(matt)), "matt")

    })

    test_that("piping with ini can update labels (rxode2/issues#351)", {
      mod <- function() {
        ini({
          a <- 1
          label("foo")
          addSd <- 2
        })
        model({
          b <- a
          b ~ add(addSd)
        })
      }
      ui <- rxode2(mod)
      expect_equal(ui$iniDf$label[ui$iniDf$name == "a"], "foo")
      newLabelUi <- ini(ui, a = label("bar"))
      expect_equal(newLabelUi$iniDf$label[newLabelUi$iniDf$name == "a"], "bar")
    })

    test_that("piping with ini can remove labels (#627)", {

      mod <- function() {
        ini({
          a <- 1
          label("foo")
          addSd <- 2
        })
        model({
          b <- a
          b ~ add(addSd)
        })
      }
      ui <- rxode2(mod)
      expect_equal(ui$iniDf$label[ui$iniDf$name == "a"], "foo")
      newLabelUi <- ini(ui, a = label(NULL))
      expect_equal(newLabelUi$iniDf$label[ui$iniDf$name == "a"], NA_character_)
    })

    test_that("piping with ini gives an error pointing the user to use label for character rhs (rxode2/issues#351)", {
      mod <- function() {
        ini({
          a <- 1
          label("foo")
          addSd <- 2
        })
        model({
          b <- a
          b ~ add(addSd)
        })
      }
      ui <- rxode2(mod)
      expect_error(
        ini(ui, a = "bar"),
        regexp = "to assign a new label, use 'a <- label(\"bar\")'",
        fixed = TRUE
      )
    })

    test_that("piping with ini can update labels (rxode2/issues#351)", {
      mod <- function() {
        ini({
          a <- 1
          label("foo")
          addSd <- 2
        })
        model({
          b <- a
          b ~ add(addSd)
        })
      }
      ui <- rxode2(mod)
      expect_equal(ui$iniDf$label[ui$iniDf$name == "a"], "foo")
      newLabelUi <- ini(ui, a = label("bar"))
      expect_equal(newLabelUi$iniDf$label[newLabelUi$iniDf$name == "a"], "bar")
    })

    test_that("piping with ini gives an error pointing the user to use label for character rhs (rxode2/issues#351)", {
      mod <- function() {
        ini({
          a <- 1
          label("foo")
          addSd <- 2
        })
        model({
          b <- a
          b ~ add(addSd)
        })
      }
      ui <- rxode2(mod)
      expect_error(
        ini(ui, a = "bar"),
        regexp = "to assign a new label, use 'a <- label(\"bar\")'",
        fixed = TRUE
      )
    })


    test_that(".iniSimplifyFixUnfix", {
      expect_equal(
        .iniSimplifyFixUnfix(str2lang("fix")),
        as.name("fix")
      )
      expect_equal(
        .iniSimplifyFixUnfix(str2lang("fixed")),
        as.name("fix")
      )
      expect_equal(
        .iniSimplifyFixUnfix(str2lang("FIX")),
        as.name("fix")
      )
      expect_equal(
        .iniSimplifyFixUnfix(str2lang("FIXED")),
        as.name("fix")
      )

      expect_equal(
        .iniSimplifyFixUnfix(str2lang("unfix")),
        as.name("unfix")
      )
      expect_equal(
        .iniSimplifyFixUnfix(str2lang("unfixed")),
        as.name("unfix")
      )
      expect_equal(
        .iniSimplifyFixUnfix(str2lang("UNFIX")),
        as.name("unfix")
      )
      expect_equal(
        .iniSimplifyFixUnfix(str2lang("UNFIXED")),
        as.name("unfix")
      )

      expect_equal(
        .iniSimplifyFixUnfix(str2lang("FIXED(a)")),
        str2lang("fix(a)")
      )
      expect_equal(
        .iniSimplifyFixUnfix(str2lang("c <- FIXED(a+b)")),
        str2lang("c <- fix(a + b)")
      )
      expect_equal(
        .iniSimplifyFixUnfix(str2lang("c <- UNFIXED(a+b)")),
        str2lang("c <- unfix(a + b)")
      )
      expect_equal(
        .iniSimplifyFixUnfix(str2lang("c <- NULL")),
        str2lang("c <- NULL")
      )
    })

    test_that(".iniSimplifyAssignArrow", {
      expect_equal(
        .iniSimplifyAssignArrow(str2lang("a <- b")),
        str2lang("a <- b")
      )
      expect_equal(
        .iniSimplifyAssignArrow(str2lang("a = b")),
        str2lang("a <- b")
      )
      # non-assignment equal signs are not modified
      expect_equal(
        .iniSimplifyAssignArrow(str2lang("a = b(c=d)")),
        str2lang("a <- b(c=d)")
      )
    })

    test_that("piping with ini can update reorder parameters (rxode2/issues#352)", {

      mod <- function() {
        ini({
          a <- 1
          b <- 2
          c <- 3
          addSd <- 2
        })
        model({
          b <- a + b*log(c)
          b ~ add(addSd)
        })
      }

      ui <- rxode2(mod)

      # No modification
      expect_equal(ui$iniDf$name, c("a", "b", "c", "addSd"))
      # b to the top by number
      expect_equal(suppressMessages(ini(ui, b <- 1, append = 0))$iniDf$name, c("b", "a", "c", "addSd"))
      # b to the top by logical
      expect_equal(suppressMessages(ini(ui, b <- 1, append = FALSE))$iniDf$name, c("b", "a", "c", "addSd"))
      # b to the bottom by number
      expect_equal(suppressMessages(ini(ui, b <- 1, append = Inf))$iniDf$name, c("a", "c", "addSd", "b"))
      # b to the bottom by logical
      expect_equal(suppressMessages(ini(ui, b <- 1, append = TRUE))$iniDf$name, c("a", "c", "addSd", "b"))
      # b to the bottom by name
      expect_equal(suppressMessages(ini(ui, b <- 1, append = "addSd"))$iniDf$name, c("a", "c", "addSd", "b"))

      expect_equal(suppressMessages(ini(ui, b <- 1, append = addSd))$iniDf$name, c("a", "c", "addSd", "b"))

      # b after c
      expect_equal(suppressMessages(ini(ui, b <- 1, append = "c"))$iniDf$name, c("a", "c", "b", "addSd"))
      # a and b after c; counter-intuitive: the order of a and b are reversed
      expect_equal(suppressMessages(ini(ui, a <- 1, b <- 1, append = "c"))$iniDf$name, c("c", "b", "a", "addSd"))
      # b to b, warn and no change
      expect_warning(
        expect_equal(suppressMessages(ini(ui, b <- 1, append = "b"))$iniDf$name, c("a", "b", "c", "addSd")),
        regexp = "parameter 'b' set to be moved after itself, no change in order made"
      )

      expect_error(
        ini(ui, b <- 1, append = d/dt(fun)),
        "append")

      # Invalid parameter is correctly caught
      expect_error(
        ini(ui, b <- 1, append = "foo"),
        "append"
      )

    })

    test_that(".iniAddCovarianceBetweenTwoEtaValues", {
      # Promote a covariate to a correlated eta

      mod <- function() {
        ini({
          a <- 1
          b <- 2
          c <- 3
          d ~ 1
          h ~ 2
          addSd <- 2
        })
        model({
          b <- a + b*log(c)
          f <- a + d + e
          i <- j + h
          b ~ add(addSd)
        })
      }

      suppressMessages(
        expect_message(
          ini(mod, d + e ~ c(1, 0.5, 3)),
          regexp = "promote `e` to between subject variability"
        )
      )

      suppressMessages(
        expect_message(
          ini(mod, d ~ 1,  e ~ c(0.5, 3)),
          regexp = "promote `e` to between subject variability"
        )
      )

      suppressMessages(
        expect_message(
          ini(mod, {
            d ~ 1
            e ~ c(0.5, 3)})
        ))

      # Non-existent correlated eta
      suppressMessages(
        expect_error(
          ini(mod, d + g ~ c(1, 0.5, 3)),
          regexp = "cannot find parameter 'g'"
        )
      )

      suppressMessages(
        expect_error(
          ini(mod, d ~ 1, g ~ c(0.5, 3)),
          regexp = "cannot find parameter 'g'"
        )
      )


      # Update eta order
      suppressMessages(
        expect_equal(
          ini(mod, h + d ~ c(1, 0.5, 3))$iniDf$name,
          c("a", "b", "c", "addSd", "h", "d", "(h,d)")
        )
      )

      suppressMessages(
        expect_equal(
          ini(mod, h ~ 1,  d ~ c(0.5, 3))$iniDf$name,
          c("a", "b", "c", "addSd", "h", "d", "(h,d)")
        )
      )

    })

    test_that(".iniHandleLabel", {
      mod <- function() {
        ini({
          a <- 1
          b <- 2
          c <- 3
          d ~ 1
          h ~ 2
          addSd <- 2
        })
        model({
          b <- a + b*log(c)
          f <- a + d + e
          i <- j + h
          b ~ add(addSd)
        })
      }

      # non-existent parameter
      expect_error(
        ini(mod, q = label("foo")),
        regexp = "cannot find parameter 'q'"
      )
      # invalid label value
      expect_error(
        ini(mod, a = label(5)),
        regexp = "the new label for 'a' must be a character string"
      )
    })

    test_that(".iniHandleAppend", {
      mod <- function() {
        ini({
          a <- 1
          b <- 2
          c <- 3
          d ~ 1
          h ~ 2
          addSd <- 2
        })
        model({
          b <- a + b*log(c)
          f <- a + d + e
          i <- j + h
          b ~ add(addSd)
        })
      }

      expect_error(
        ini(mod, a <- 1, append=factor("A")),
        regexp = "'append' must be NULL, logical, numeric, or character/expression of variable in model"
      )
      expect_error(
        ini(mod, q <- 1, append=0),
        regexp = "cannot find parameter 'q'"
      )
      # Non-theta parameters cannot be moved
      expect_error(
        ini(mod, h ~ 1, append=0),
        regexp = "only theta parameters can be moved"
      )
    })

    test_that("ini tests for different types of expressions", {

      mod <- function() {
        ini({
          a <- 1
          b <- 2
          c <- 3
          d ~ 1
          h ~ 2
          addSd <- 2
        })
        model({
          b <- a + b*log(c)
          f <- a + d + e
          i <- j + h
          b ~ add(addSd)
        })
      }

      expect_error(mod |> ini("h~3"), NA)

      expect_error(mod |> ini("h~3;4*"))

      expect_error(mod |> ini(factor("A")))

    })

    test_that("zeroRe", {
      modOmegaSigma <- function() {
        ini({
          a <- 1; label("foo") #nolint
          iiva ~ 3
          addSd <- 2
        })
        model({
          b <- a + iiva
          b ~ add(addSd)
        })
      }
      modOmega <- function() {
        ini({
          a <- 1; label("foo") # nolint
          iiva ~ 3
        })
        model({
          b <- a + iiva
        })
      }
      modSigma <- function() {
        ini({
          a <- 1; label("foo") # nolint
          addSd <- 2
        })
        model({
          b <- a
          b ~ add(addSd)
        })
      }
      modSigmaBound <- function() {
        ini({
          a <- 1; label("foo") # nolint
          addSd <- c(1, 2)
        })
        model({
          b <- a
          b ~ add(addSd)
        })
      }
      modNone <- function() {
        ini({
          a <- 1; label("foo") # nolint
        })
        model({
          b <- a
        })
      }
      uiOmegaSigma <- rxode2(modOmegaSigma)
      uiOmega <- rxode2(modOmega)
      uiSigma <- rxode2(modSigma)
      uiSigmaBound <- rxode2(modSigmaBound)
      uiNone <- rxode2(modNone)

      expect_silent(
        suppressMessages(
          newMod <- zeroRe(modOmegaSigma, which = c("omega", "sigma"))
        )
      )
      expect_silent(
        suppressMessages(
          newUi <- zeroRe(uiOmegaSigma, which = c("omega", "sigma"))
        )
      )
      expect_equal(newMod$iniDf, newUi$iniDf)
      # detect change
      expect_equal(uiOmegaSigma$iniDf$est, c(1, 2, 3))
      expect_equal(newMod$iniDf$est, c(1, 0, 0))

      # Confirm that you can simulate from the model
      suppressMessages(
        expect_equal(
          rxSolve(newMod, events = data.frame(TIME = 0:2))$b,
          rep(1, 3)
        )
      )

      # Confirm that the `fix` flag is respected
      expect_silent(
        suppressMessages(
          newUiNoFix <- zeroRe(uiOmegaSigma, which = c("omega", "sigma"), fix = FALSE)
        )
      )
      # detect change
      expect_equal(uiOmegaSigma$iniDf$fix, rep(FALSE, 3))
      expect_equal(newUi$iniDf$fix, c(FALSE, TRUE, TRUE))
      expect_equal(newUiNoFix$iniDf$fix, rep(FALSE, 3))

      suppressMessages(
        expect_warning(
          newMod <- zeroRe(modOmega, which = c("omega", "sigma")),
          regexp = "No sigma parameters in the model"
        )
      )
      suppressMessages(
        expect_warning(
          newUi <- zeroRe(uiOmega, which = c("omega", "sigma")),
          regexp = "No sigma parameters in the model"
        )
      )
      expect_equal(newMod$iniDf, newUi$iniDf)
      # detect change
      expect_equal(uiOmega$iniDf$est, c(1, 3))
      expect_equal(newMod$iniDf$est, c(1, 0))

      suppressMessages(
        expect_warning(
          newMod <- zeroRe(modSigmaBound, which = c("omega", "sigma")),
          regexp = "No omega parameters in the model"
        )
      )
      suppressMessages(
        expect_warning(
          newUi <- zeroRe(uiSigmaBound, which = c("omega", "sigma")),
          regexp = "No omega parameters in the model"
        )
      )
      expect_equal(newMod$iniDf, newUi$iniDf)
      # detect change
      expect_equal(uiSigmaBound$iniDf$est, c(1, 2))
      expect_equal(newMod$iniDf$est, c(1, 0))
      # confirm lower bound change
      expect_equal(uiSigmaBound$iniDf$lower, c(-Inf, 1))
      expect_equal(newMod$iniDf$lower, c(-Inf, 0))

      suppressMessages(
        expect_warning(
          newMod <- zeroRe(modSigma, which = c("omega", "sigma")),
          regexp = "No omega parameters in the model"
        )
      )
      suppressMessages(
        expect_warning(
          newUi <- zeroRe(uiSigma, which = c("omega", "sigma")),
          regexp = "No omega parameters in the model"
        )
      )
      expect_equal(newMod$iniDf, newUi$iniDf)
      # detect change
      expect_equal(uiSigma$iniDf$est, c(1, 2))
      expect_equal(newMod$iniDf$est, c(1, 0))

      suppressMessages(
        expect_warning(expect_warning(
          newMod <- zeroRe(modNone, which = c("omega", "sigma")),
          regexp = "No omega parameters in the model"),
          regexp = "No sigma parameters in the model"
          )
      )
      suppressMessages(
        expect_warning(expect_warning(
          newUi <- zeroRe(uiNone, which = c("omega", "sigma")),
          regexp = "No omega parameters in the model"),
          regexp = "No sigma parameters in the model"
          )
      )
      expect_equal(newMod$iniDf, newUi$iniDf)
      # detect no change
      expect_equal(uiNone$iniDf$est, 1)
      expect_equal(newMod$iniDf$est, 1)

      # expected errors
      expect_error(zeroRe("A"), regexp = "'object' needs to be a rxUi model")
      expect_error(zeroRe(modOmegaSigma, which = "foo"), regexp = "should be one of")
    })

    test_that("zeroRe works with correlated etas (#480)", {
      mod <- function() {
        ini({
          lka <- 0.45
          lcl <- 1
          lvc <- 3.45
          propSd <- c(0, 0.5)
          etalka + etalcl + etalvc ~ c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        })
        model({
          ka <- exp(lka + etalka)
          cl <- exp(lcl + etalcl)
          vc <- exp(lvc + etalvc)
          cp <- linCmt()
          cp ~ prop(propSd)
        })
      }
      ui <- rxode2(mod)
      expect_equal(ui$iniDf$est[!is.na(ui$iniDf$neta1)], (1:6)/10)
      suppressMessages(zeroUi <- zeroRe(mod))
      expect_equal(zeroUi$iniDf$est[!is.na(zeroUi$iniDf$neta1)], c(0, 0, 0))
    })

    test_that("Piping outside the boundaries", {

      m1 <- function() {
        ini({
          x2 <- c(0, 1)
          x3 <- c(0, 1, 2)
        })
        model({
          f <- x2+x3*4
        })
      }

      suppressMessages({
        f2 <- m1 |> ini(x2=-1)
        expect_equal(f2$iniDf[f2$iniDf$name == "x2","lower"], -Inf)
      })
      suppressMessages({
        f2 <- m1 |> ini(x3=4)
        expect_equal(f2$iniDf[f2$iniDf$name == "x3","upper"], Inf)
      })
      suppressMessages({
        f2 <- m1 |> ini(x3=c(0,3))
        expect_equal(f2$iniDf[f2$iniDf$name == "x3","upper"], Inf)
      })
    })

    test_that("append allows promoting from covariate (#472)", {
      mod <- function() {
        ini({
          lka <- 0.45
          lcl <- 1
          lvc  <- 3.45
          propSd <- 0.5
        })
        model({
          ka <- exp(lka)
          cl <- exp(lcl)
          vc  <- exp(lvc)

          kel <- cl / vc

          d/dt(depot) <- -ka*depot
          d/dt(central) <- ka*depot-kel*central

          cp <- central / vc
          cp ~ prop(propSd)
        })
      }
      suppressMessages(
        newmod <-
          mod |>
          model(
            ka <- exp(lka + ka_dose*DOSE),
            auto = FALSE
          ) |>
          ini(
            ka_dose <- 1,
            append = "lka"
          )
      )
      expect_equal(newmod$iniDf$name, c("lka", "ka_dose", "lcl", "lvc", "propSd"))
    })

    test_that("change ini type with ~", {

      mod <- function() {
        ini({
          lka <- 0.45
          lcl <- 1
          lvc  <- 3.45
          propSd <- 0.5
        })
        model({
          ka <- exp(lka)
          cl <- exp(lcl)
          vc  <- exp(lvc)
          kel <- cl / vc
          d/dt(depot) <- -ka*depot
          d/dt(central) <- ka*depot-kel*central
          cp <- central / vc
          cp ~ prop(propSd)
        })
      }

      mod1 <- mod |> ini( ~ lka)
      expect_equal(mod1$omega, lotri(lka ~ 0.45))

      mod2 <- mod1 |> ini( ~ lka)
      expect_equal(mod2$omega, NULL)

      expect_error(mod1 |> ini( ~ propSd))

      expect_error(mod1 |> ini( ~ matt))

      ## all etas

      mod <- function() {
        ini({
          lka ~ 0.45
          lcl ~ 1
          lvc ~ 3.45
        })
        model({
          ka <- exp(lka)
          cl <- exp(lcl)
          vc  <- exp(lvc)
          kel <- cl / vc
          d/dt(depot) <- -ka*depot
          d/dt(central) <- ka*depot-kel*central
          cp <- central / vc
        })
      }

      mod2 <- mod |> ini( ~ lka)

      expect_equal(mod2$omega, lotri(lcl ~ 1, lvc ~ 3.45))

      # remove correlated eta

      mod <- function() {
        ini({
          lka + lcl + lvc ~
            c(0.45,
              0.01, 1,
              0.01, -0.01, 3.45)
        })
        model({
          ka <- exp(lka)
          cl <- exp(lcl)
          vc  <- exp(lvc)
          kel <- cl / vc
          d/dt(depot) <- -ka*depot
          d/dt(central) <- ka*depot-kel*central
          cp <- central / vc
        })
      }

      mod2 <- mod |> ini( ~ lka)

      expect_equal(mod2$omega, lotri(lcl + lvc ~ c(1,
                                                   -0.01, 3.45)))


      # negative and zero

      mod <- function() {
        ini({
          lka <- 0.45
          lcl <- -1
          lvc <- 0
        })
        model({
          ka <- exp(lka)
          cl <- exp(lcl)
          vc  <- exp(lvc)
          kel <- cl / vc
          d/dt(depot) <- -ka*depot
          d/dt(central) <- ka*depot-kel*central
          cp <- central / vc
        })
      }

      mod2 <- mod |> ini( ~ lcl)

      expect_equal(mod2$omega, lotri(lcl ~ 1))

      mod2 <- mod |> ini( ~ lvc)

      expect_equal(mod2$omega, lotri(lvc ~ 1))

      mod3 <- mod2 |> ini( ~ lvc)

      expect_equal(mod3$omega, NULL)

      mod4 <- mod3 |> ini( ~ lvc)

      expect_equal(mod4$omega, lotri(lvc ~ 1))

    })



    test_that("change ini variable to covariate with -", {

      mod <- function() {
        ini({
          lka + lcl + lvc ~
            c(0.45,
              0.01, 1,
              0.01, -0.01, 3.45)
        })
        model({
          ka <- exp(lka)
          cl <- exp(lcl)
          vc  <- exp(lvc)
          kel <- cl / vc
          d/dt(depot) <- -ka*depot
          d/dt(central) <- ka*depot-kel*central
          cp <- central / vc
        })
      }

      mod2 <- mod |> ini(-lka)

      expect_equal(mod2$allCovs, "lka")
      expect_equal(mod2$omega, lotri(lcl + lvc ~ c(1, -0.01, 3.45)))

      mod <- function() {
        ini({
          lka ~ 0.45
          lcl ~ 1
          lvc ~ 3.45
        })
        model({
          ka <- exp(lka)
          cl <- exp(lcl)
          vc  <- exp(lvc)
          kel <- cl / vc
          d/dt(depot) <- -ka*depot
          d/dt(central) <- ka*depot-kel*central
          cp <- central / vc
        })
      }

      mod2 <- mod |> ini(-lka)

      expect_equal(mod2$allCovs, "lka")


    })
  }

  test_that("empty arguments to rxRename() give a warning (#688)", {
    mod1 <- function() {
      ini({
        Kin=1
      })
      model({
        eff <- Kin
      })
    }

    expect_warning(
      rxRename(mod1, ),
      "empty argument ignored"
    )
    expect_warning(
      rxRename(mod1, foo = eff, ),
      "empty argument ignored"
    )
  })

  test_that("parameters can be promoted from covariate to parameter with bounds (#692)", {
    mod1 <- function() {
      model({
        eff <- Kin
      })
    }

    expect_message(
      mod1 |> ini(Kin = 2),
      "promote `Kin` to population parameter with initial estimate 2"
    )
    expect_message(
      expect_message(
        mod1 |> ini(Kin = c(1, 2)),
        "promote `Kin` to population parameter with initial estimate 2"
      ),
      regexp = "change initial estimate (2) and lower bound (1) of `Kin`",
      fixed = TRUE
    )
    expect_message(
      expect_message(
        mod1 |> ini(Kin = c(1, 2, 3)),
        "promote `Kin` to population parameter with initial estimate 2"
      ),
      regexp = "change initial estimate (2) and upper/lower bound (1 to 3) of `Kin`",
      fixed = TRUE
    )
  })

  test_that("ini(diag) and ini(-cov()) tests", {

    mod2 <- function() {
      ini({
        lka ~ 0.45
        lcl ~ c(0.01, 1)
        lvc ~ c(-0.01, 0.01, 3.45)
        lfun ~ c(-0.1, 0.1, 0.01, 4)
      })
      model({
        ka <- exp(lka)
        cl <- exp(lcl)
        vc  <- exp(lvc)
        kel <- cl / vc
        d/dt(depot) <- -ka*depot
        d/dt(central) <- ka*depot-kel*central
        cp <- central / vc + lfun
      })
    }

    expect_error(
      mod2 |> ini(diag(lcl, matt)),
      "matt"
    )

    expect_error(
      mod2 |> ini(diag(matt, lcl)),
      "matt"
    )

    tmp <- mod2 |> ini(-cov(lcl, lvc))
    expect_equal(tmp$omega,
                 lotri({
                   lvc ~ 3.45
                   lfun ~ c(0.01, 4)
                   lka ~ c(-0.01, -0.1, 0.45)
                   lcl ~ c(0, 0.1, 0.01, 1)
                 }))

    tmp <- mod2 |> ini(-cor(lcl, lvc))
    expect_equal(tmp$omega,
                 lotri({
                   lvc ~ 3.45
                   lfun ~ c(0.01, 4)
                   lka ~ c(-0.01, -0.1, 0.45)
                   lcl ~ c(0, 0.1, 0.01, 1)
                 }))

    tmp <- mod2 |> ini(cor(lcl, lvc) <- NULL)

    expect_equal(tmp$omega,
                 lotri({
                   lvc ~ 3.45
                   lfun ~ c(0.01, 4)
                   lka ~ c(-0.01, -0.1, 0.45)
                   lcl ~ c(0, 0.1, 0.01, 1)
                 }))

    tmp <- mod2 |> ini(cor(lcl, lvc) ~ NULL)
    expect_equal(tmp$omega,
                 lotri({
                   lvc ~ 3.45
                   lfun ~ c(0.01, 4)
                   lka ~ c(-0.01, -0.1, 0.45)
                   lcl ~ c(0, 0.1, 0.01, 1)
                 }))

    expect_error(mod2 |> ini(diag(matt)),
                 "matt")

    # Will reorder
    tmp <- mod2 |> ini(diag(lcl, lvc))
    expect_equal(tmp$omega,
                 lotri({
                   lfun ~ 4
                   lka ~ c(-0.1, 0.45)
                   lvc ~ 3.45
                   lcl ~ 1
                 }))

    tmp <- mod2 |> ini(diag)
    expect_equal(tmp$omega,
                 lotri({
                   lka ~ 0.45
                   lcl ~ 1
                   lvc ~ 3.45
                   lfun ~ 4
                 }))

    tmp <- mod2 |> ini(diag(lvc))

    expect_equal(tmp$omega,
                 lotri({
                   lfun ~ 4
                   lcl ~ c(0.1, 1)
                   lka ~ c(-0.1, 0.01, 0.45)
                   lvc ~ 3.45
                 }))

    mod <- function() {
      ini({
        lka ~ 0.45
        lcl ~ c(0.01, 1)
        lvc ~ c(-0.01, 0.01, 3.45)
      })
      model({
        ka <- exp(lka)
        cl <- exp(lcl)
        vc  <- exp(lvc)
        kel <- cl / vc
        d/dt(depot) <- -ka*depot
        d/dt(central) <- ka*depot-kel*central
        cp <- central / vc
      })
    }


    tmp <- mod |> ini(diag)

    expect_equal(tmp$omega,
                 lotri({
                   lka ~ 0.45
                   lcl ~ 1
                   lvc ~ 3.45
                 }))

    tmp <- mod |> ini(diag())

    expect_equal(tmp$omega,
                 lotri({
                   lka ~ 0.45
                   lcl ~ 1
                   lvc ~ 3.45
                 }))

  })
})

#' Extract fixed effects from an rxode2 model (rxUi)
#'
#' @param object rxode2 ui model or a function that evaluates to one
#' @param ... other parameters (currently ignored)
#' @return named numeric vector of the fixed-effect (theta) estimates
#' @author Matthew L. Fidler
#' @keywords internal
#' @examples
#' one.compartment <- function() {
#'   ini({
#'     tka <- 0.45
#'     tcl <- 1
#'     tv <- 3.45
#'     eta.ka ~ 0.6
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl)
#'     v <- exp(tv)
#'     d/dt(depot) <- -ka * depot
#'     d/dt(center) <- ka * depot - cl / v * center
#'     cp <- center / v
#'   })
#' }
#'
#' ui <- one.compartment()
#' nlme::fixef(ui)
#' @keywords internal
fixef.rxUi <- function(object, ...) {
  object$theta
}

#' @rdname fixef.rxUi
#' @keywords internal
fixef.function <- function(object, ...) {
  ui <- rxode2(object)
  ui$theta
}

#' Extract the initial coefficients from an rxode2 model (rxUi)
#'
#' For an rxode2 model specification (`rxUi`) there are no realized random
#' effects (etas), so `coef()` returns the fixed-effect (`theta`) estimates by
#' default.  The random-effect variability matrix (`omega`) can be requested
#' with `level="omega"`, or both can be returned together with `level="all"`.
#'
#' @param object rxode2 ui model or a function that evaluates to one
#' @param level which coefficients to return: `"theta"`/`"fixed"` (the default)
#'   returns the named fixed-effect estimates; `"omega"`/`"random"` returns the
#'   random-effect variability matrix (or `NULL` when the model has no random
#'   effects); `"all"`/`"both"` returns a list with `theta` and `omega`.
#' @param ... other parameters (currently ignored)
#' @return depends on `level`: a named numeric vector of fixed effects, the
#'   `omega` matrix (or `NULL`), or a list with both `theta` and `omega`.
#' @author Matthew L. Fidler
#' @importFrom stats coef
#' @examples
#' one.compartment <- function() {
#'   ini({
#'     tka <- 0.45
#'     tcl <- 1
#'     tv <- 3.45
#'     eta.ka ~ 0.6
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl)
#'     v <- exp(tv)
#'     d/dt(depot) <- -ka * depot
#'     d/dt(center) <- ka * depot - cl / v * center
#'     cp <- center / v
#'   })
#' }
#'
#' ui <- one.compartment()
#' coef(ui)
#' coef(ui, level="omega")
#' coef(ui, level="all")
#' @export
coef.rxUi <- function(object,
                      level = c("theta", "fixed", "omega", "random", "all", "both"),
                      ...) {
  level <- match.arg(level)
  if (level %in% c("theta", "fixed")) {
    return(object$theta)
  }
  if (level %in% c("omega", "random")) {
    return(object$omega)
  }
  list(theta = object$theta, omega = object$omega)
}

#' @rdname coef.rxUi
#' @export
coef.function <- function(object,
                          level = c("theta", "fixed", "omega", "random", "all", "both"),
                          ...) {
  coef.rxUi(rxode2(object), level = level, ...)
}

.rxD <- new.env(parent = emptyenv())
## This environment is a derivative table;
## For example:
## Derivative(f(a,b,c), a) = fa()
## Derivative(f(a,b,c), b) = fb()
## Derivative(f(a,b,c), c) = fc()
## Then
##
## .rxD$f <- list(fa(a,b,c), fb(a,b,c), fc(a,b,c))
##
##  fa translates the arguments to the derivative with respect to a
##  fb translates the arguments to the derivative with respect to b
##
## If any of the list is NULL then rxode2 won't know how to take a
## derivative with respect to the argument.
##
## If the list is shorter than the length of the arguments then the
## argument then the derivative of arguments that are not specified
## cannot be taken.

.rxD$atan2 <- list(
  function(y, x) {
    return(paste0("(", x, ")/((", x, ")^2+(", y, ")^2)"))
  },
  function(y, x) {
    return(paste0("-(", y, ")/((", x, ")^2+(", y, ")^2)"))
  }
)

.rxD$erfinv <- list(
  function(x) {
    ## http://specialfunctionswiki.org/index.php/Derivative_of_inverse_error_function
    return(paste0("sqrt(pi)/2*exp((erfinv(", x, "))^2)"))
  }
)

# FIXME should be able to be moved to rxode2ll
.rxD$llikNorm <- list(
  NULL,
  function(x, mean, sd) {
    paste0("llikNormDmean(",paste(c(x, mean, sd), collapse=", "), ")")
  },
  function(x, mean, sd) {
    paste0("llikNormDsd(",paste(c(x, mean, sd), collapse=", "), ")")
  }
)

.rxD$llikXNorm <- list(
  NULL,
  NULL,
  function(i, x, mean, sd) {
    paste0("llikXNormDmean(",paste(c(i, x, mean, sd), collapse=", "), ")")
  },
  function(i, x, mean, sd) {
    paste0("llikXNormDsd(",paste(c(i, x, mean, sd), collapse=", "), ")")
  }
)

.rxD$llikPois <- list(
  function(x, lambda) {
    "0"
  },
  function(x, lambda) {
    paste0("llikPoisDlambda(",paste(c(x, lambda), collapse=", "), ")")
  }
)

.rxD$llikXPois <- list(
  NULL,
  function(i, x, lambda) {
    "0"
  },
  function(i, x, lambda) {
    paste0("llikXPoisDlambda(",paste(c(i, x, lambda), collapse=", "), ")")
  }
)

.rxD$llikBinom <- list(
  function(x, size, prob) {
    "0"
  },
  function(x, size, prob) {
    "0"
  },
  function(x, size, prob) {
    paste0("llikBinomDprob(",paste(c(x, size, prob), collapse=", "), ")")
  }
)

.rxD$llikXBinom <- list(
  NULL,
  function(i, x, size, prob) {
    "0"
  },
  function(i, x, size, prob) {
    "0"
  },
  function(i, x, size, prob) {
    paste0("llikXBinomDprob(",paste(c(i, x, size, prob), collapse=", "), ")")
  }
)


.rxD$llikNbinom <- list(
  function(x, size, prob) {
    "0"
  },
  function(x, size, prob) {
    "0"
  },
  function(x, size, prob) {
    paste0("llikNbinomDprob(",paste(c(x, size, prob), collapse=", "), ")")
  }
)

.rxD$llikXNbinom <- list(
  NULL,
  function(i, x, size, prob) {
    "0"
  },
  function(i, x, size, prob) {
    "0"
  },
  function(i, x, size, prob) {
    paste0("llikXNbinomDprob(",paste(c(i, x, size, prob), collapse=", "), ")")
  }
)

.rxD$llikNbinomMu <- list(
  function(x, size, mu) {
    "0"
  },
  function(x, size, mu) {
    "0"
  },
  function(x, size, mu) {
    paste0("llikNbinomMuDmu(",paste(c(x, size, mu), collapse=", "), ")")
  }
)

.rxD$llikXNbinomMu <- list(
  NULL,
  function(i, x, size, mu) {
    "0"
  },
  function(i, x, size, mu) {
    "0"
  },
  function(i, x, size, mu) {
    paste0("llikXNbinomMuDmu(",paste(c(i, x, size, mu), collapse=", "), ")")
  }
)


.rxD$llikXBeta <- list(
  NULL,
  NULL,
  function(i, x, shape1, shape2) {
    paste0("llikXBetaDshape1(",paste(c(i, x, shape1, shape2), collapse=", "), ")")
  },
  function(i, x, shape1, shape2) {
    paste0("llikXBetaDshape2(",paste(c(i, x, shape1, shape2), collapse=", "), ")")
  }
)

.rxD$llikBeta <- list(
  NULL,
  function(x, shape1, shape2) {
    paste0("llikBetaDshape1(",paste(c(x, shape1, shape2), collapse=", "), ")")
  },
  function(x, shape1, shape2) {
    paste0("llikBetaDshape2(",paste(c(x, shape1, shape2), collapse=", "), ")")
  }
)

.rxD$llikT <- list(
  NULL,
  function(x, df, mean, sd) {
    paste0("llikTDdf(",paste(c(x, df, mean, sd), collapse=", "), ")")
  },
  function(x, df, mean, sd) {
    paste0("llikTDmean(",paste(c(x, df, mean, sd), collapse=", "), ")")
  },
  function(x, df, mean, sd) {
    paste0("llikTDsd(",paste(c(x, df, mean, sd), collapse=", "), ")")
  }
)

.rxD$llikXT <- list(
  NULL,
  NULL,
  function(i, x, df, mean, sd) {
    paste0("llikXTDdf(",paste(c(i, x, df, mean, sd), collapse=", "), ")")
  },
  function(i, x, df, mean, sd) {
    paste0("llikXTDmean(",paste(c(i, x, df, mean, sd), collapse=", "), ")")
  },
  function(i, x, df, mean, sd) {
    paste0("llikXTDsd(",paste(c(i, x, df, mean, sd), collapse=", "), ")")
  }
)

.rxD$llikChisq <- list(
  NULL,
  function(x, nu) {
    paste0("llikChisqDdf(",paste(c(x, nu), collapse=", "), ")")
  }
)

.rxD$llikXChisq <- list(
  NULL,
  NULL,
  function(i, x, nu) {
    paste0("llikXChisqDdf(",paste(c(i, x, nu), collapse=", "), ")")
  }
)

.rxD$llikExp <- list(
  NULL,
  function(x, rate) {
    paste0("llikExpDrate(",paste(c(x, rate), collapse=", "), ")")
  }
)

.rxD$llikXExp <- list(
  NULL,
  NULL,
  function(i, x, rate) {
    paste0("llikXExpDrate(",paste(c(i, x, rate), collapse=", "), ")")
  }
)

.rxD$llikF <- list(
  NULL,
  function(x, df1, df2) {
    paste0("llikFDdf1(",paste(c(x, df1, df2), collapse=", "), ")")
  },
  function(x, df1, df2) {
    paste0("llikFDdf2(",paste(c(x, df1, df2), collapse=", "), ")")
  }
)

.rxD$llikXF <- list(
  NULL,
  NULL,
  function(i, x, df1, df2) {
    paste0("llikXFDdf1(",paste(c(i, x, df1, df2), collapse=", "), ")")
  },
  function(i, x, df1, df2) {
    paste0("llikXFDdf2(",paste(c(i, x, df1, df2), collapse=", "), ")")
  }
)


.rxD$llikGeom <- list(
  NULL,
  function(x, p) {
    paste0("llikGeomDprob(",paste(c(x, p), collapse=", "), ")")
  }
)

.rxD$llikXGeom <- list(
  NULL,
  NULL,
  function(i, x, p) {
    paste0("llikXGeomDprob(",paste(c(i, x, p), collapse=", "), ")")
  }
)

.rxD$llikUnif <- list(
  NULL,
  function(x, alpha, beta) {
    paste0("llikUnifDalpha(",paste(c(x, alpha, beta), collapse=", "), ")")
  },
  function(x, alpha, beta) {
    paste0("llikUnifDbeta(",paste(c(x, alpha, beta), collapse=", "), ")")
  }
)

.rxD$llikXUnif <- list(
  NULL,
  NULL,
  function(i, x, alpha, beta) {
    paste0("llikXUnifDalpha(",paste(c(i, x, alpha, beta), collapse=", "), ")")
  },
  function(i, x, alpha, beta) {
    paste0("llikXUnifDbeta(",paste(c(i, x, alpha, beta), collapse=", "), ")")
  }
)

.rxD$llikWeibull <- list(
  NULL,
  function(x, shape, scale) {
    paste0("llikWeibullDshape(",paste(c(x, shape, scale), collapse=", "), ")")
  },
  function(x, shape, scale) {
    paste0("llikWeibullDscale(",paste(c(x, shape, scale), collapse=", "), ")")
  }
)

.rxD$llikXWeibull <- list(
  NULL,
  NULL,
  function(i, x, shape, scale) {
    paste0("llikXWeibullDshape(",paste(c(i, x, shape, scale), collapse=", "), ")")
  },
  function(i, x, shape, scale) {
    paste0("llikXWeibullDscale(",paste(c(i, x, shape, scale), collapse=", "), ")")
  }
)

.rxD$llikGamma <- list(
  NULL,
  function(x, shape, rate) {
    paste0("llikGammaDshape(",paste(c(x, shape, rate), collapse=", "), ")")
  },
  function(x, shape, rate) {
    paste0("llikGammaDrate(",paste(c(x, shape, rate), collapse=", "), ")")
  }
)

.rxD$llikXGamma <- list(
  NULL,
  NULL,
  function(i, x, shape, rate) {
    paste0("llikXGammaDshape(",paste(c(i, x, shape, rate), collapse=", "), ")")
  },
  function(i, x, shape, rate) {
    paste0("llikXGammaDrate(",paste(c(i, x, shape, rate), collapse=", "), ")")
  }
)

.rxD$llikCauchy <- list(
  NULL,
  function(x, location, scale) {
    paste0("llikCauchyDlocation(",paste(c(x, location, scale), collapse=", "), ")")
  },
  function(x, location, scale) {
    paste0("llikCauchyDscale(",paste(c(x, location, scale), collapse=", "), ")")
  }
)

.rxD$llikXCauchy <- list(
  NULL,
  NULL,
  function(i, x, location, scale) {
    paste0("llikXCauchyDlocation(",paste(c(i, x, location, scale), collapse=", "), ")")
  },
  function(i, x, location, scale) {
    paste0("llikXCauchyDscale(",paste(c(i, x, location, scale), collapse=", "), ")")
  }
)

.rxD$llikXGamma <- list(
  NULL,
  NULL,
  function(i, x, shape, rate) {
    paste0("llikXGammaDshape(",paste(c(i, x, shape, rate), collapse=", "), ")")
  },
  function(i, x, shape, rate) {
    paste0("llikXGammaDrate(",paste(c(i, x, shape, rate), collapse=", "), ")")
  }
)

# end likelihood piece

.rxD$abs0 <- list(function(x) {
  return(paste0("dabs(", x, ")"))
})

.rxD$abs <- list(function(x) {
  return(paste0("dabs(", x, ")"))
})


.rxD$abs1 <- list(function(x) {
  return(paste0("dabs1(", x, ")"))
})

.rxD$dabs1 <- list(function(x) {
  return("0")
})

.rxD$dabs <- list(function(x) {
  return(paste0("dabs2(", x, ")"))
})

.rxD$dabs2 <- list(function(x) {
  return("0")
})


.rxD$rxTBS <- list(function(a, lambda, yj, hi, low) {
  paste0("rxTBSd(", a, ",", lambda, ",", yj, ",", hi, ",", low, ")")
}, function(a, lambda, yj, hi, low) {
  paste0("rxTBSdL(", a, ",", lambda, ",", yj, ",", hi, ",", low, ")")
})

.rxD$rxTBSd <- list(function(a, lambda, yj, hi, low) {
  paste0("rxTBSd2(", a, ",", lambda, ",", yj, ",", hi, ",", low, ")")
}, function(a, lambda, yj, hi, low) {
  paste0("rxTBSdLx(", a, ",", lambda, ",", yj, ",", hi, ",", low, ")")
})

.rxD$rxTBSdL <- list(function(a, lambda, yj, hi, low) {
  paste0("rxTBSdLx(", a, ",", lambda, ",", yj, ",", hi, ",", low, ")")
}, function(a, lambda, yj, hi, low) {
  paste0("rxTBSdL2(", a, ",", lambda, ",", yj, ",", hi, ",", low, ")")
})

.rxD$..k <- 10

.rxD$rxMod <- list(
  # fmod(x, y) = x - y*trunc(x/y)
  # the trunc() doesn't exist at integers; this gives the derivative at non-integer values
  function(a, b) {
    paste0("1")
  },
  function(a, b) {
    paste0("0")
  }
)

## Approx a==b by
## (1-tanh(k*(a-b))^2) -- a bump centered at a==b
.rxD$rxEq <- list(
  function(a, b) {
    .ab <- paste0("(", a, "-", b, ")")
    return(paste0(
      "(", -2 * .rxD$..k, "*tanh(", .rxD$..k, "*", .ab, ")+",
      2 * .rxD$..k, "*tanh(", .rxD$..k, "*", .ab, ")^3)"
    ))
  }, function(a, b) {
    .ab <- paste0("(", a, "-", b, ")")
    return(paste0(
      "(", 2 * .rxD$..k, "*tanh(", .rxD$..k, "*", .ab, ")-",
      2 * .rxD$..k, "*tanh(", .rxD$..k, "*", .ab, ")^3)"
    ))
  }
)

## Derivative of the inequality operators (>=, <=, <, >): the smooth step
## 1/2 +/- 1/2*tanh(k*(a-b)) differentiates to the nascent-delta bump
## +/- (k/2 - (k/2)*tanh(k*(a-b))^2), which peaks at the boundary a==b and
## integrates to 1.  Do not shift the tanh (earlier versions added
## atanh(2*tol-1), moving the bump to a-b ~ +/-0.46): the forward pass emits
## the hard boolean, which jumps at a==b, so an off-center bump gives .rxSens
## consumers a spurious derivative where the value is constant.  The
## strict/non-strict distinction lives in the hard value, not the derivative.
.rxD$rxGeq <- list(
  function(a, b) {
    .ab <- paste0("(", a, "-", b, ")")
    return(paste0(
      "(", .rxD$..k / 2, "-", .rxD$..k / 2, "*tanh(", .rxD$..k, "*", .ab, ")^2)"
    ))
  }, function(a, b) {
    .ab <- paste0("(", a, "-", b, ")")
    return(paste0(
      "(", -.rxD$..k / 2, "+", .rxD$..k / 2, "*tanh(", .rxD$..k, "*", .ab, ")^2)"
    ))
  }
)

.rxD$rxLeq <- list(
  function(a, b) {
    .ab <- paste0("(", a, "-", b, ")")
    return(paste0(
      "(", -.rxD$..k / 2, "+", .rxD$..k / 2, "*tanh(", .rxD$..k, "*", .ab, ")^2)"
    ))
  }, function(a, b) {
    .ab <- paste0("(", a, "-", b, ")")
    return(paste0(
      "(", .rxD$..k / 2, "-", .rxD$..k / 2, "*tanh(", .rxD$..k, "*", .ab, ")^2)"
    ))
  }
)


.rxD$rxLt <- list(
  function(a, b) {
    .ab <- paste0("(", a, "-", b, ")")
    return(paste0(
      "(", -.rxD$..k / 2, "+", .rxD$..k / 2, "*tanh(", .rxD$..k, "*", .ab, ")^2)"
    ))
  },
  function(a, b) {
    .ab <- paste0("(", a, "-", b, ")")
    return(paste0(
      "(", .rxD$..k / 2, "-", .rxD$..k / 2, "*tanh(", .rxD$..k, "*", .ab, ")^2)"
    ))
  }
)


.rxD$rxGt <- list(
  function(a, b) {
    .ab <- paste0("(", a, "-", b, ")")
    return(paste0(
      "(", .rxD$..k / 2, "-", .rxD$..k / 2, "*tanh(", .rxD$..k, "*", .ab, ")^2)"
    ))
  },
  function(a, b) {
    .ab <- paste0("(", a, "-", b, ")")
    return(paste0(
      "(", -.rxD$..k / 2, "+", .rxD$..k / 2, "*tanh(", .rxD$..k, "*", .ab, ")^2)"
    ))
  }
)

.rxD$rxAnd <- list(
  function(a, b) {
    ## a*b
    return(b)
  }, function(a, b) {
    ## a*b
    return(a)
  }
)

.rxD$rxOr <- list(
  function(a, b) {
    ## Using DeMorgan's Theorem
    ## a+b = 1-(1-a)*(1-b)
    return(paste0("(1-(", b, "))"))
  }, function(a, b) {
    return(paste0("(1-(", a, "))"))
  }
)


.rxD$rxNot <- list(
  function(a) {
    ## 1 - a
    return("(-1)")
  }
)

.rxD$dose <- list(function(a) {
  return("0")
})

.rxD$podo <- list(function(a) {
  return("0")
})
.rxD$podo0 <- .rxD$podo
.rxD$dose0 <- .rxD$dose

.rxD$tlast <- list(function(a) {
  return("0")
})
.rxD$tfirst <- list(function(a) {
  return("0")
})

.rxD$tlast0 <- .rxD$tlast
.rxD$tfirst0 <- .rxD$tfirst

.rxD$first <- list(function(a) {
  return("0")
})
.rxD$last <- list(function(a) {
  return("0")
})
.rxD$diff <- list(function(a) {
  return("0")
})
.rxD$is.nan <- list(function(a) {
  return("0")
})
.rxD$is.na <- list(function(a) {
  return("0")
})
.rxD$is.finite <- list(function(a) {
  return("0")
})
.rxD$is.infinite <- list(function(a) {
  return("0")
})

.rxD$gammap <- list(
  NULL,
  function(a, z) {
    paste0("gammapDer(", a, ",", z, ")")
  }
)

.rxD$ReLU <- list(
  function(x) {
    paste0("dReLU(", x, ")")
  }
)

.rxD$dReLU <- list(
  function(x) {
    paste0("0")
  }
)

.rxD$GELU <- list(
  function(x) {
    paste0("dGELU(", x, ")")
  }
)

.rxD$dGELU <- list(
  function(x) {
    paste0("d2GELU(", x, ")")
  }
)

.rxD$d2GELU <- list(
  function(x) {
    paste0("d3GELU(", x, ")")
  }
)

.rxD$d3GELU <- list(
  function(x) {
    paste0("d4GELU(", x, ")")
  }
)

.rxD$ELU <- list(
  function(x, alpha) {
    paste0("dELU(", x, ", ", alpha, ")")
  },
  function(x, alpha) {
    paste0("dELUa(", x, ", ", alpha, ")")
  })

.rxD$dELU <- list(
  function(x, alpha) {
    paste0("d2ELU(", x, ", ", alpha, ")")
  },
  function(x, alpha) {
    paste0("d2aELU(", x, ", ", alpha, ")")
  })

.rxD$dELUa <- list(
  function(x, alpha) {
    paste0("d2ELUa(", x, ", ", alpha, ")")
  },
  function(x, alpha) {
    paste0("0")
  }
)
.rxD$d2ELUa <- list(
  function(x, alpha) {
    paste0("d2ELUa(", x, ", ", alpha, ")")
  },
  function(x, alpha) {
    paste0("0")
  }
)

.rxD$d2ELU <- list(
  function(x, alpha) {
    paste0("d2ELU(", x, ", ", alpha, ")")
  },
  function(x, alpha) {
    paste0("d2aELU(", x, ", ", alpha, ")")
  })

.rxD$d2aELU <- list(
  function(x, alpha) {
    paste0("d2aELU(", x, ", ", alpha, ")")
  },
  function(x, alpha) {
    paste0("0")
  })

.rxD$softplus <- list(
  function(x) {
    paste0("dsoftplus(", x, ")")
  })

.rxD$dsoftplus <- list(
  function(x) {
    paste0("d2softplus(", x, ")")
  })

.rxD$d2softplus <- list(
  function(x) {
    paste0("d3softplus(", x, ")")
  })

.rxD$d3softplus <- list(
  function(x) {
    paste0("d4softplus(", x, ")")
  })

.rxD$SELU <- list(
  function(x) {
    paste0("dSELU(", x, ")")
  })


.rxD$lReLU <- list(
  function(x) {
    paste0("dlReLU(", x, ")")
  }
)

.rxD$dlReLU <- list(
  function(x) {
    paste0("0")
  }
)

.rxD$PReLU <- list(
  function(x, alpha) {
    paste0("dPReLU(", x, ",", alpha, ")")
  },
  function(x, alpha) {
    paste0("dPReLUa(", x, ",", alpha, ")")
  })

.rxD$dPReLU <- list(
  function(x, alpha) {
    paste0("0")
  },
  function(x, alpha) {
    paste0("dPReLUa1(", x, ",", alpha, ")")
  })

.rxD$dPReLUa <- list(
  function(x, alpha) {
    paste0("dPReLUa1(", x, ",", alpha, ")")
  },
  function(x, alpha) {
    paste0("0")
  })

.rxD$dPReLUa1 <- list(
  function(x, alpha) {
    paste0("0")
  },
  function(x, alpha) {
    paste0("0")
  }
)

.rxD$Swish <- list(
  function(x) {
    paste0("dSwish(", x, ")")
  }
)
.linCmtBgen <- function(i) {
  # Ka isn't handled
  if (i == 9) {
    .which <- 0
  } else if (i == 10) {
    .which <- 1
  } else if (i == 11) {
    .which <- 2
  } else if (i == 12) {
    .which <- 3
  } else if (i == 13) {
    .which <- 4
  } else if (i == 14) {
    .which <- 5
  }
  .fun <- function(...) {}
  body(.fun) <- bquote({
    .args <- unlist(list(...))
    .ncmt <- .args[4] # ncmt
    .args5 <- .args[5] # oral0
    .oral0 <- as.numeric(.args5)
    .args6 <- .args[6] # which1
    .args7 <- .args[7] # which2
    .w <- .(paste(.which))
    .wn <- .(.which)
    .nc <- as.numeric(.ncmt)
    if (.wn + 1 > .nc * 2) {
      return("0")
    }
    if (.args6 == "-1" && .args7 == "-1") {
      ## This is the derivative of the linear compartment solution
      # Return the gradent with respect to the parameter
      .args[6] <- "-2"
      .args[7] <- .w
    } else if (.args7 == "-2") {
      ## This is the amount in each of the saved compartments
      ## and which1 represents the amount in the compartment (zero indexed)
      if (as.numeric(.args6) >= .oral0 + .nc) {
        return("0")
      }
      .args[7] <- .w
    } else {
      stop("bad 'linCmtB' derivative", call. = FALSE)
    }
    return(paste0("linCmtB(", paste(.args, collapse = ","), ")"))
  })
  return(.fun)
}

.rxD$linCmtB <- list(
  function(...) { # rx__PTR__
    stop("bad 'linCmtB' derivative", call. = FALSE)
  },
  function(...) { # t
    stop("bad 'linCmtB' derivative", call. = FALSE)
  },
  function(...) { # linCmt
    stop("bad 'linCmtB' derivative", call. = FALSE)
  },
  function(...) { # ncmt
    stop("bad 'linCmtB' derivative", call. = FALSE)
  },
  function(...) { # oral0
    stop("bad 'linCmtB' derivative", call. = FALSE)
  },
  function(...) { # which1
    stop("bad 'linCmtB' derivative", call. = FALSE)
  },
  function(...) { # which2
    stop("bad 'linCmtB' derivative", call. = FALSE)
  },
  function(...) { # trans
    stop("bad 'linCmtB' derivative", call. = FALSE)
  },
  .linCmtBgen(9),  # p1
  .linCmtBgen(10), # v1
  .linCmtBgen(11), # p2
  .linCmtBgen(12), # p3
  .linCmtBgen(13), # p4
  .linCmtBgen(14), # p5
  function(...) {  # ka
    .args <- unlist(list(...))
    .ncmt <- .args[4] # ncmt
    .nc <- as.numeric(.ncmt)
    .args5 <- .args[5] # oral0
    .oral0 <- as.numeric(.args5)
    if (.args5 != "1") return("0")
    .args6 <- .args[6]
    .args7 <- .args[7]
    .which <- "2"
    if (.ncmt == 3) {
      .which <- "6"
    } else if (.ncmt == 2) {
      .which <- "4"
    }
    if (.args6 == "-1" && .args7 == "-1") {
      ## This is the derivative of the linear compartment solution
      # Return the gradent with respect to the parameter
      .args[6] <- "-2"
      .args[7] <- .which
    } else if (.args7 == "-2") {
      ## This is the amount in each of the saved compartments
      ## and which1 represents the amount in the compartment (zero indexed)
      if (as.numeric(.args6) >= .oral0 + .nc) {
        return("0")
      }
      .args[7] <- .which
    } else {
      stop("bad 'linCmtB' derivative", call. = FALSE)
    }
    return(paste0("linCmtB(", paste(.args, collapse = ","), ")"))
  }
  # linCmtB(rx__PTR__, t, linCmt, ncmt, oral0, which1, which2 ,trans,
  #         p1, v1, p2, p3, p4, p5, ka)
)


# When which1 & which2 are -1 then the function is the linear compartment solution
#
# When which2 = -2, the function is the amount in each of the saved compartments and which1
# represents the amount in the compartment (zero indexed)
#
# When which1 = -2, the function returns the gradient of the linear compartment model
#
# Otherwise which1 & which2 are returns the Jacobain of the system

#' This gives the derivative table for rxode2
#'
#' This will help allow registration of functions in `rxode2`
#'
#' @return Derivative table environment for rxode2
#' @details
#'
#' This environment is a derivative table;
#'
#' For example:
#'
#' Derivative(f(a,b,c), a) = fa()
#' Derivative(f(a,b,c), b) = fb()
#' Derivative(f(a,b,c), c) = fc()
#'
#' Then the derivative table for `f` would be:
#'
#' assign("f", list(fa(a,b,c), fb(a,b,c), fc(a,b,c)), rxode2parseD())
#'
#'  fa translates the arguments to the derivative with respect to a
#'  fb translates the arguments to the derivative with respect to b
#'
#' If any of the list is NULL then rxode2 won't know how to take a
#' derivative with respect to the argument.
#'
#' If the list is shorter than the length of the arguments then the
#' argument then the derivative of arguments that are not specified
#' cannot be taken.
#' @author Matthew L. Fidler
#' @export
rxode2parseD <- function() {
  return(.rxD)
}

#' Simulate random normal variable from threefry generator
#'
#' @inheritParams stats::rnorm
#'
#' @param n number of observations
#'
#' @param ncores Number of cores for the simulation
#'
#' `rxnorm` simulates using the threefry sitmo generator.
#'
#'
#' `rxnormV` used to simulate with the vandercorput simulator, but
#' since it didn't satisfy the normal properties it was changed to simple be
#' an alias of `rxnorm`. It is no longer supported in `rxode2({})` blocks
#'
#' @return normal random number deviates
#'
#' @examples
#' \donttest{
#' ## Use threefry engine
#'
#' rxnorm(n = 10) # with rxnorm you have to explicitly state n
#' rxnorm(n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxnorm(2, 3) ## The first 2 arguments are the mean and standard deviation
#'
#'
#' ## This example uses `rxnorm` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxnorm()
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#' s <- rxSolve(rx, et)
#'
#' }
#' @export
rxnormV  <- function(mean = 0, sd = 1, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(mean, len = 1)
  checkmate::assertNumeric(sd, lower = 0, len = 1)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxnorm_`, mean, sd, n, ncores)
}
#' @rdname rxnormV
#' @export
rxnorm <- rxnormV



#' Simulate random Poisson variable from threefry generator
#'
#' @inheritParams stats::rpois
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#' @return poission random number deviates
#' @examples
#' \donttest{
#' ## Use threefry engine
#'
#' rxpois(lambda = 3, n = 10) # with rxpois you have to explicitly state n
#' rxpois(lambda = 3, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxpois(4) ## The first arguments are the lambda parameter
#'
#'
#' ## This example uses `rxpois` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxpois(3)
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#'
#' s <- rxSolve(rx, et)
#'
#' }
#' @export
rxpois <- function(lambda, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(lambda, len = 1)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxpois_`, lambda, n, ncores)
}

#' Simulate student t variable from threefry generator
#'
#' @inheritParams stats::rt
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#' @return t-distribution random numbers
#' @examples
#' \donttest{
#'
#' ## Use threefry engine
#' rxt(df = 3, n = 10) # with rxt you have to explicitly state n
#' rxt(df = 3, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxt(4) ## The first argument is the df parameter
#'
#'
#' ## This example uses `rxt` directly in the model
#'
#' rx <- function() {
#'    model({
#'     a <- rxt(3)
#'    })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#' s <- rxSolve(rx, et)
#' }
#' @export
rxt <- function(df, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(df, len = 1, lower = 0)
  if (df == 0) stop("'df' must be greater than 0", call. = FALSE)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxt__`, df, n, ncores)
}

#' Simulate uniform variable from threefry generator
#'
#' @inheritParams stats::runif
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#' @return uniform random numbers
#' @examples
#' \donttest{
#'
#' ## Use threefry engine
#'
#' rxunif(min = 0, max = 4, n = 10) # with rxunif you have to explicitly state n
#' rxunif(min = 0, max = 4, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxunif()
#'
#'
#' ## This example uses `rxunif` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxunif(0, 3)
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#'
#' s <- rxSolve(rx, et)
#'
#' }
#'
#' @export
rxunif <- function(min = 0, max = 1, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(min, len = 1)
  checkmate::assertNumeric(max, len = 1)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxunif_`, min, max, n, ncores)
}


#' Simulate Weibull variable from threefry generator
#'
#' @inheritParams stats::rweibull
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#' @return Weibull random deviates
#' @examples
#' \donttest{
#'
#' ## Use threefry engine
#'
#' # with rxweibull you have to explicitly state n
#' rxweibull(shape = 1, scale = 4, n = 10)
#'
#' # You can parallelize the simulation using openMP
#' rxweibull(shape = 1, scale = 4, n = 10, ncores = 2)
#'
#' rxweibull(3)
#'
#'
#' ## This example uses `rxweibull` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxweibull(1, 3)
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#'
#' s <- rxSolve(rx, et)
#'
#' }
#' @export
rxweibull <- function(shape, scale = 1, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(shape, len = 1)
  checkmate::assertNumeric(scale, len = 1)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxweibull_`, shape, scale, n, ncores)
}


#' Simulate geometric variable from threefry generator
#'
#' @inheritParams stats::rgeom
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#' @return geometric random deviates
#' @examples
#' \donttest{
#'
#' ## Use threefry engine
#'
#' rxgeom(0.5, n = 10) # with rxgeom you have to explicitly state n
#' rxgeom(0.25, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxgeom(0.75)
#'
#'
#' ## This example uses `rxgeom` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxgeom(0.24)
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#'
#' s <- rxSolve(rx, et)
#'
#' }
#'
#' @export
rxgeom <- function(prob, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(prob, len = 1, lower = 0, upper = 1)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxgeom_`, prob, n, ncores)
}


#' Simulate beta variable from threefry generator
#'
#' @inheritParams stats::rbeta
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#'
#' @return beta random deviates
#'
#' @examples
#' \donttest{
#'
#' ## Use threefry engine
#'
#' rxbeta(0.5, 0.5, n = 10) # with rxbeta you have to explicitly state n
#' rxbeta(5, 1, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxbeta(1, 3)
#'
#'
#' ## This example uses `rxbeta` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxbeta(2, 2)
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#'
#' s <- rxSolve(rx, et)
#'
#' }
#' @export
rxbeta <- function(shape1, shape2, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(shape1, len = 1, lower = 0)
  if (shape1 == 0) stop("'shape1' cannot be 0", call. = FALSE)
  checkmate::assertNumeric(shape2, len = 1, lower = 0)
  if (shape2 == 0) stop("'shape2' cannot be 0", call. = FALSE)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxbeta_`, shape1, shape2, n, ncores)
}

#' Simulate gamma variable from threefry generator
#'
#' @param shape The shape of the gamma random variable
#' @inheritParams stats::rgamma
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#'
#' @return gamma random deviates
#'
#' @examples
#'
#' \donttest{
#'
#' ## Use threefry engine
#'
#' rxgamma(0.5, n = 10) # with rxgamma you have to explicitly state n
#' rxgamma(5, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxgamma(1)
#'
#'
#' ## This example uses `rxbeta` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxgamma(2)
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#'
#' s <- rxSolve(rx, et)
#'
#' }
#'
#' @export
rxgamma <- function(shape, rate = 1, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(shape, len = 1, lower = 0)
  if (shape == 0) stop("'shape' cannot be 0", call. = FALSE)
  checkmate::assertNumeric(rate, len = 1, lower = 0)
  if (rate == 0) stop("'rate' cannot be 0", call. = FALSE)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxgamma_`, shape, rate, n, ncores)
}

#' Simulate F variable from threefry generator
#'
#' @inheritParams stats::rf
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#'
#' @return f random deviates
#'
#' @examples
#' \donttest{
#'
#' ## Use threefry engine
#'
#' rxf(0.5, 0.5, n = 10) # with rxf you have to explicitly state n
#' rxf(5, 1, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxf(1, 3)
#'
#'
#' ## This example uses `rxf` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxf(2, 2)
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#'
#' s <- rxSolve(rx, et)
#' }
#'
#' @export
rxf <- function(df1, df2, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(df1, len = 1, lower = 0)
  if (df1 == 0) stop("'df1' cannot be 0", call. = FALSE)
  checkmate::assertNumeric(df2, len = 1, lower = 0)
  if (df2 == 0) stop("'df2' cannot be 0", call. = FALSE)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxf_`, df1, df2, n, ncores)
}

#' Simulate exponential variable from threefry generator
#'
#' @inheritParams stats::rexp
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#'
#' @return exponential random deviates
#'
#' @examples
#' \donttest{
#'
#' ## Use threefry engine
#'
#' rxexp(0.5, n = 10) # with rxexp you have to explicitly state n
#' rxexp(5, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxexp(1)
#'
#'
#' ## This example uses `rxexp` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxexp(2)
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#'
#' s <- rxSolve(rx, et)
#'
#' }
#'
#' @export
rxexp <- function(rate, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(rate, len = 1, lower = 0)
  if (rate == 0) stop("'rate' cannot be 0", call. = FALSE)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxexp_`, rate, n, ncores)
}


#' Simulate chi-squared variable from threefry generator
#'
#' @inheritParams stats::rchisq
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#'
#' @return chi squared random deviates
#'
#' @examples
#' \donttest{
#'
#' ## Use threefry engine
#'
#' rxchisq(0.5, n = 10) # with rxchisq you have to explicitly state n
#' rxchisq(5, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxchisq(1)
#'
#'
#' ## This example uses `rxchisq` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxchisq(2)
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#'
#' s <- rxSolve(rx, et)
#'
#' }
#'
#' @export
rxchisq <- function(df, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(df, len = 1, lower = 0)
  if (df == 0) stop("'df' cannot be 0", call. = FALSE)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxchisq_`, df, n, ncores)
}

#' Simulate Cauchy variable from threefry generator
#'
#' @inheritParams stats::rcauchy
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#'
#' @return Cauchy random deviates
#'
#' @examples
#' \donttest{
#'
#' ## Use threefry engine
#'
#' rxcauchy(0, 1, n = 10) # with rxcauchy you have to explicitly state n
#' rxcauchy(0.5, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxcauchy(3)
#'
#'
#' ## This example uses `rxcauchy` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxcauchy(2)
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#' s <- rxSolve(rx, et)
#'
#' }
#'
#' @export
rxcauchy <- function(location = 0, scale = 1, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(location, len = 1)
  checkmate::assertNumeric(scale, len = 1, lower = 0)
  if (scale == 0) stop("'scale' cannot be 0", call. = FALSE)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxcauchy_`, location, scale, n, ncores)
}
#' Simulate ordinal value
#'
#' @param ... the probabilities to be simulated.  These should sum up to a number below one.
#'
#'
#' @return A number from 1 to the (number of probabilities specified + 1)
#'
#' @details
#'
#' The values entered into the 'rxord' simulation will simulate the
#' probability of falling each group. If it falls outside of the
#' specified probabilities, it will simulate the group (number of
#' probabilities specified + 1)
#'
#' @author Matthew L. Fidler
#' @examples
#'
#' # This will give values 1, and 2
#' rxord(0.5)
#' rxord(0.5)
#' rxord(0.5)
#' rxord(0.5)
#'
#' # This will give values 1, 2 and 3
#' rxord(0.3, 0.3)
#' rxord(0.3, 0.3)
#' rxord(0.3, 0.3)
#'
#' @export
rxord <- function(...) {
  .args <- c(...)
  if (any(.args < 0)) return(NA_real_)
  if (any(.args > 1)) return(NA_real_)
  .sum <- sum(.args)
  if (.sum >= 1) return(NA_real_)
  .v <- rxunif()
  .args <- cumsum(.args)
  .Call(`_rxode2_rxordSelect`, .v, .args)
}

#' Simulate Binomial variable from threefry generator
#'
#' @inheritParams stats::rbinom
#' @inheritParams rxnorm
#'
#' @template birthdayProblem
#'
#' @return binomial random deviates
#'
#' @examples
#' \donttest{
#' ## Use threefry engine
#'
#'
#' rxbinom(10, 0.9, n = 10) # with rxbinom you have to explicitly state n
#' rxbinom(3, 0.5, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxbinom(4, 0.7)
#'
#'
#' ## This example uses `rxbinom` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxbinom(1, 0.5)
#'   })
#' }
#'
#' et <- et(1, id = 1:2)
#'
#' s <- rxSolve(rx, et)
#' }
#'
#' @export
rxbinom <- function(size, prob, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(prob, len = 1, lower = 0, upper = 1)
  checkmate::assertCount(size)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxbinom_`, size, prob, n, ncores)
}

#' Simulate Binomial variable from threefry generator
#'
#' @inheritParams stats::rnbinom
#' @inheritParams rxbinom
#'
#' @template birthdayProblem
#'
#' @return negative binomial random deviates. Note that `rxbinom2`
#'   uses the `mu` parameterization an the `rxbinom` uses the `prob`
#'   parameterization (`mu=size/(prob+size)`)
#'
#' @examples
#' \donttest{
#' ## Use threefry engine
#'
#' rxnbinom(10, 0.9, n = 10) # with rxbinom you have to explicitly state n
#' rxnbinom(3, 0.5, n = 10, ncores = 2) # You can parallelize the simulation using openMP
#'
#' rxnbinom(4, 0.7)
#'
#' # use mu parameter
#' rxnbinomMu(40, 40, n=10)
#'
#' ## This example uses `rxbinom` directly in the model
#'
#' rx <- function() {
#'   model({
#'     a <- rxnbinom(10, 0.5)
#'   })
#' }
#'
#' et <- et(1, id = 1:100)
#'
#'
#' s <- rxSolve(rx, et)
#'
#'
#' rx <- function() {
#'   model({
#'     a <- rxnbinomMu(10, 40)
#'   })
#' }
#'
#'
#' s <- rxSolve(rx, et)
#'
#' }
#' @export
rxnbinom <- function(size, prob, mu, n = 1L, ncores = 1L) {

  if (!missing(mu)) {
    if (!missing(prob)) {
      stop("only one of 'mu' or 'prob' can be specified",
           call.=FALSE)
    } else {
      return(rxnbinomMu(size=size, mu=mu, n=n, ncores=ncores))
    }
  }
  checkmate::assertNumeric(prob, len = 1, lower = 0, upper = 1)
  checkmate::assertCount(size)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxnbinom_`, size, prob, n, ncores)
}

##' @rdname rxnbinom
##' @export
rxnbinomMu <- function(size, mu, n = 1L, ncores = 1L) {
  checkmate::assertNumeric(mu, len = 1, lower = 0)
  checkmate::assertCount(size)
  checkmate::assertCount(n)
  checkmate::assertCount(ncores)
  rxSeedEng(ncores)
  .Call(`_rxode2_rxnbinomMu_`, size, mu, n, ncores)
}

#' Simulate a from a Poisson process
#'
#' @param n Number of time points to simulate in the Poisson process
#'
#' @param lambda Rate of Poisson process
#'
#' @param gamma Asymmetry rate of Poisson process.  When gamma=1.0,
#'   this simulates a homogenous Poisson process.  When gamma<1.0,
#'   the Poisson process has more events early, when gamma > 1.0,
#'   the Poisson process has more events late in the process.
#'
#'   When gamma is non-zero, the tmax should not be infinite but indicate
#'   the end of the Poisson process to be simulated.  In most
#'   pharamcometric cases, this will be the end of the study.
#'   Internally this uses a rate of:
#'
#'   l(t) = lambda*gamma*(t/tmax)^(gamma-1)
#'
#'
#' @param prob When specified, this is a probability function with
#'   one argument, time, that gives the probability that a Poisson
#'   time t is accepted as a rejection time.
#'
#' @param t0 the starting time of the Poisson process
#'
#' @param tmax the maximum time of the Poisson process
#'
#' @param randomOrder when `TRUE` randomize the order of the Poisson
#'   events.  By default (`FALSE`) it returns the Poisson process is
#'   in order of how the events occurred.
#'
#' @return
#'
#' This returns a vector of the Poisson process times; If the dropout is >=
#' tmax, then all the rest of the times are = tmax to indicate the
#' dropout is equal to or after tmax.
#'
#' @author Matthew Fidler
#' @export
#' @examples
#'
#' ## Sample homogenous Poisson process of rate 1/10
#' rxPp(10, 1 / 10)
#'
#' ## Sample inhomogenous Poisson rate of 1/10
#'
#' rxPp(10, 1 / 10, gamma = 2, tmax = 100)
#'
#' ## Typically the Poisson process times are in a sequential order,
#' ## using randomOrder gives the Poisson process in random order
#'
#' rxPp(10, 1 / 10, gamma = 2, tmax = 10, randomOrder = TRUE)
#'
#' ## This uses an arbitrary function to sample a non-homogenous Poisson process
#'
#' rxPp(10, 1 / 10, prob = function(x) {
#'   1/(1+abs(x))
#' })
#'
rxPp <- function(n, lambda, gamma = 1.0, prob = NULL, t0 = 0.0, tmax = Inf, randomOrder = FALSE) {
  checkmate::assertNumeric(t0, len = 1, any.missing = FALSE)
  checkmate::assertNumeric(tmax, len = 1, any.missing = FALSE, lower = t0)
  checkmate::assertNumeric(gamma, len = 1, any.missing = FALSE, lower = .Machine$double.eps)
  checkmate::assertNumeric(lambda, len = 1, any.missing = FALSE, lower = .Machine$double.eps)
  checkmate::assertIntegerish(n, len = 1, any.missing = FALSE, lower = 1L)
  checkmate::assertLogical(randomOrder, len = 1, any.missing = FALSE)
  if (gamma != 1.0 && is.infinite(tmax)) {
    stop("when 'gamma' is not 1, 'tmax' cannot be infinite")
  }
  .Call(`_rxode2_rpp_`, n, lambda, gamma, prob, t0, tmax, randomOrder, PACKAGE = "rxode2")
}

#' cbind Ome
#'
#' @param et The theta data frame
#' @param mat The full matrix simulation from omegas
#' @param n number of subject simulated
#' @return data frame with et combined with simulated omega matrix values
#' @author Matthew Fidler
#' @export
.cbindOme <- function(et, mat, n) {
  .Call(`_cbindOme`, et, mat, as.integer(n)) # nolint
}


#' Cumulative distribution of standard normal
#'
#' @param q vector of quantiles
#'
#' @return cumulative distribution of standard normal distribution
#'
#' @author Matthew Fidler
#'
#' @examples
#'
#' # phi is equivalent to pnorm(x)
#' phi(3)
#'
#' # See
#' pnorm(3)
#'
#' # This is provided for NONMEM-like compatibility in rxode2 models
#' @export
phi <- function(q) {
  .Call(`_phi`, q)
}

#' Convert numeric vector to repeated data.frame
#'
#' @param vec Named input vector
#' @param n Number of columns
#' @return Data frame with repeated vec
#' @author Matthew Fidler
#' @export
.vecDf <- function(vec, n) {
  .Call(`_vecDF`, vec, as.integer(n)) # nolint
}



#' Expand parameters
#'
#'
#' @param object rxode2 model variables object
#' @param params parameters to expand
#' @param events event table to help with the expansion
#' @param control control structure to help with the parameter generation
#' @return Expanded parameters for simulation
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.expandPars <- function(object, params, events, control) {
  .Call(`_rxode2_expandPars_`, object, params, events, control)
}


#' Sample a covariance Matrix from the Posterior Inverse Wishart
#' distribution.
#'
#' Note this Inverse wishart rescaled to match the original scale of
#' the covariance matrix.
#'
#' If your covariance matrix is a 1x1 matrix, this uses an scaled
#' inverse chi-squared which is equivalent to the Inverse Wishart
#' distribution in the uni-directional case.
#'
#' @param nu Degrees of Freedom (Number of Observations) for
#'        covariance matrix simulation.
#'
#' @param omega Either the estimate of covariance matrix or the
#'     estimated standard deviations in matrix form each row forming
#'     the standard deviation simulated values
#'
#' @param n Number of Matrices to sample.  By default this is 1.
#'     This is only useful when `omega` is a matrix.  Otherwise
#'     it is determined by the number of rows in the input
#'     `omega` matrix of standard deviations
#'
#' @param omegaIsChol is an indicator of if the omega matrix is in
#'   the Cholesky decomposition. This is only used when \code{type="invWishart"}
#'
#' @param returnChol Return the Cholesky decomposition of the
#'   covariance matrix sample. This is only used when \code{type="invWishart"}
#'
#' @param diagXformType Diagonal transformation type.  These could be:
#'
#' * `log` The standard deviations are log transformed, so the
#'   actual standard deviations are exp(omega)
#'
#' * `identity` The standard deviations are not transformed. The
#' standard deviations are not transformed;  They should be positive.
#'
#' * `variance` The variances are specified in the `omega`
#' matrix; They are transformed into standard deviations.
#'
#' * `nlmixrSqrt` These standard deviations come from an nlmixr
#' omega matrix where diag(chol(inv(omega))) = x^2
#'
#' * `nlmixrLog` These standard deviations come from a nlmixr
#' omega matrix omega matrix where diag(chol(solve(omega))) = exp(x)
#'
#' * `nlmixrIdentity` These standard deviations come from a nlmixr
#' omega matrix omega matrix where diag(chol(solve(omega))) = x
#'
#'
#'  The nlmixr transformations only make sense when there is no
#'  off-diagonal correlations modeled.
#'
#' @param type The type of covariance posterior that is being
#'     simulated.  This can be:
#'
#'
#' * `invWishart` The posterior is an inverse wishart; This allows
#' for correlations between parameters to be modeled.  All the
#' uncertainty in the parameter is captured in the degrees of freedom
#' parameter.
#'
#' * `lkj` The posterior separates the standard deviation
#' estimates (modeled outside and provided in the `omega`
#' argument) and the correlation estimates. The correlation estimate
#' is simulated with the [rLKJ1()].  This simulation uses
#' the relationship `eta=(nu-1)/2`.  This is relationship based
#' on the proof of the relationship between the restricted
#' LKJ-distribution and inverse wishart distribution (XXXXXX).  Once
#' the correlation posterior is calculated, the estimated standard
#' deviations are then combined with the simulated correlation matrix
#' to create the covariance matrix.
#'
#' * `separation` Like the `lkj` option, this separates out
#' the estimation of the correlation and standard deviation.  Instead
#' of using the `LKJ` distribution to simulate the correlation,
#' it simulates the inverse wishart of the identity matrix and
#' converts the result to a correlation matrix.  This correlation
#' matrix is then used with the standard deviation to calculate the
#' simulated covariance matrix.
#'
#'
#' @return a matrix (n=1) or a list of matrices  (n > 1)
#'
#' @details
#'
#' In general, the separation strategy is preferred for diagonal
#' matrices.  If the dimension of the matrix is below 10, `lkj`
#' is numerically faster than `separation` method.  However, the
#' `lkj` method has densities too close to zero (XXXX) when the
#' dimension is above 10.  In that case, though computationally more
#' expensive `separation` method performs better.
#'
#' For matrices with modeled covariances, the easiest method to use
#' is the inverse Wishart which allows the simulation of correlation
#' matrices (XXXX).  This method is more well suited for well behaved
#' matrices, that is the variance components are not too low or too
#' high.  When modeling nonlinear mixed effects modeling matrices
#' with too high or low variances are considered sub-optimal in
#' describing a system.  With these rules in mind, it is reasonable
#' to use the inverse Wishart.
#'
#' @author Matthew L.Fidler & Wenping Wang
#'
#' @references
#'
#' Alvarez I, Niemi J and Simpson M. (2014) *Bayesian Inference for a
#' Covariance Matrix*. Conference on Applied Statistics in Agriculture.
#'
#' Wang1 Z, Wu Y, and Chu H. (2018) *On Equivalence of the LKJ
#' distribution and the restricted Wishart
#' distribution*. <doi:10.48550/arXiv.1809.047463
#'
#' @examples
#'
#' ## Sample a single covariance.
#' draw1 <- cvPost(3, matrix(c(1, .3, .3, 1), 2, 2))
#'
#' ## Sample 3 covariances
#' set.seed(42)
#' draw3 <- cvPost(3, matrix(c(1, .3, .3, 1), 2, 2), n = 3)
#'
#' ## Sample 3 covariances, but return the cholesky decomposition
#' set.seed(42)
#' draw3c <- cvPost(3, matrix(c(1, .3, .3, 1), 2, 2), n = 3, returnChol = TRUE)
#'
#' ## Sample 3 covariances with lognormal standard deviations via LKJ
#' ## correlation sample
#' cvPost(3, sapply(1:3, function(...) {
#'   rnorm(10)
#' }), type = "lkj")
#'
#' ## or return cholesky decomposition
#' cvPost(3, sapply(1:3, function(...) {
#'   rnorm(10)
#' }),
#' type = "lkj",
#' returnChol = TRUE
#' )
#'
#' ## Sample 3 covariances with lognormal standard deviations via separation
#' ## strategy using inverse Wishart correlation sample
#' cvPost(3, sapply(1:3, function(...) {
#'   rnorm(10)
#' }), type = "separation")
#'
#' ## or returning the cholesky decomposition
#' cvPost(3, sapply(1:3, function(...) {
#'   rnorm(10)
#' }),
#' type = "separation",
#' returnChol = TRUE
#' )
#' @export
cvPost <- function(nu, omega, n = 1L, omegaIsChol = FALSE, returnChol = FALSE,
                   type = c("invWishart", "lkj", "separation"),
                   diagXformType = c("log", "identity", "variance", "nlmixrSqrt", "nlmixrLog", "nlmixrIdentity")) {
  if (is.null(nu) && n == 1L) {
    return(omega)
  }
  if (inherits(type, "numeric") || inherits(type, "integer")) {
    .type <- as.integer(type)
  } else {
    .type <- as.vector(c(
      "invWishart" = 1L, "lkj" = 2L,
      "separation" = 3L
    )[match.arg(type)])
  }
  if (.type == 1L) {
    .xform <- 1L
  } else if (inherits(diagXformType, "numeric") || inherits(diagXformType, "integer")) {
    .xform <- as.integer(diagXformType)
  } else {
    .xform <- setNames(
      c(
        "variance" = 6L, "log" = 5L,
        "identity" = 4L, "nlmixrSqrt" = 1L,
        "nlmixrLog" = 2L,
        "nlmixrIdentity" = 3L
      )[match.arg(diagXformType)],
      NULL
    )
  }
  .ret <- .Call(`_rxode2_cvPost_`, nu, omega, n,
                omegaIsChol, returnChol, .type, .xform)
  return(.ret)
}

#' Get the rxode2 seed
#'
#' @return rxode2 seed state or -1 when the seed isn't set
#'
#' @export
#' @seealso rxSetSeed, rxWithSeed, rxWithPreserveSeed
#' @examples
#'
#' # without setting seed
#'
#' rxGetSeed()
#' # Now set the seed
#' rxSetSeed(42)
#'
#' rxGetSeed()
#'
#' rxnorm()
#'
#' rxGetSeed()
#'
#' # don't use the rxode2 seed again
#'
#' rxSetSeed(-1)
#'
#' rxGetSeed()
#'
#' rxnorm()
#'
#' rxGetSeed()
#'
rxGetSeed <- function() {
  .Call(`_rxode2_rxGetSeed`)
}

#' Set the parallel seed for rxode2 random number generation
#'
#' This sets the seed for the rxode2 parallel random number generation.
#' If set, then whenever a seed is set for the threefry or
#' vandercorput simulation engine, it will use this seed, increment
#' for the number of seeds and continue with the sequence the next
#' time the random number generator is called.
#'
#' In contrast, when this is not called, the time that the
#' vandercorput or threefry simulation engines are seeded it comes
#' from a uniform random number generated from the standard R random
#' seed.  This may cause a duplicate seed based on the R seed state.
#' This means that there could be correlations between simulations
#' that do not exist This will avoid the birthday problem picking
#' exactly the same seed using the seed state of the R random number
#' generator.  The more times the seed is called, the more likely this
#' becomes.
#'
#' @param seed An integer that represents the rxode2 parallel and
#'   internal random number generator seed.  When positive, use this
#'   seed for random number generation and increment and reseed any
#'   parallel or new engines that are being called. When negative,
#'   turn off the rxode2 seed and generate a seed from the R's uniform
#'   random number generator.  Best practice is to set this seed.
#'
#' @return Nothing, called for its side effects
#'
#' @author Matthew Fidler
#'
#' @examples
#'
#' rxSetSeed(42)
#'
#' # seed with generator 42
#' rxnorm()
#'
#' # Use R's random number generator
#' rnorm(1)
#'
#' rxSetSeed(42)
#'
#' # reproduces the same number
#' rxnorm()
#'
#' # But R's random number is not the same
#'
#' rnorm(1)
#'
#' # If we reset this to use the R's seed
#' # (internally rxode2 uses a uniform random number to span seeds)
#' # This can lead to duplicate sequences and seeds
#'
#' rxSetSeed(-1)
#'
#' # Now set seed works for both.
#'
#' # This is not recommended, but illustrates the different types of
#' # seeds that can be generated.
#'
#' set.seed(42)
#'
#' rxnorm()
#'
#' rnorm(1)
#'
#' set.seed(42)
#'
#' rxnorm()
#'
#' rnorm(1)
#'
#' @seealso rxGetSeed, rxWithSeed, rxWithPreserveSeed
#'
#' @references
#'
#' JD Cook. (2016). Random number generator seed mistakes.
#' \url{https://www.johndcook.com/blog/2016/01/29/random-number-generator-seed-mistakes/}
#'
#' @export
rxSetSeed <- function(seed) {
  .Call(`_rxSetSeed`, seed)
  invisible()
}

.rmRseed <- function() {
  if (!exists(".Random.seed", globalenv(), mode = "integer", inherits = FALSE)) {
    return(NULL)
  }
  set.seed(seed = NULL)
  rm(".Random.seed", envir = globalenv())
}

#' Get the random seed for the session
#'
#'
#' @return A list with the following components:
#'
#' * `seed` The random seed for the session.  This is the value of
#'  `.Random.seed` in the global environment.
#'
#' * `kind` The random number generator kind for the session.  This is
#'    the value of `RNGkind()` for the session.
#'
#' * `rxseed` The rxode2 random seed for the session.  This
#'    is the value of `rxGetSeed()` for the session.
#'
#'
#' @export
#' @keywords internal
#' @author Matthew L. Fidler
.rxGetSeed <- function() {
  if (!exists(".Random.seed", globalenv(), mode = "integer", inherits = FALSE)) {
    return(list(seed=NULL, kind=NULL, rxseed=rxGetSeed()))
  }
  list(seed = get(".Random.seed", globalenv(), mode = "integer",
                  inherits = FALSE), kind = RNGkind(), rxseed=rxGetSeed())
}
#' Set the random seed for the session
#'
#' This sets the random seed for the session.  This is used internally
#' to set the random seed for the session.
#'
#' @param seed Seed from `.rxGetSeed()`
#' @return Nothing, called for its side effects
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.rxSetSeed <- function(seed) {
  if (is.null(seed$seed)) {
    .rxGetSeed()
  } else {
    do.call(RNGkind, args = as.list(seed$kind))
    .Call(`_rxode2_setGlobalSeed`, seed$seed)
  }
  rxSetSeed(seed$rxseed)
  invisible(NULL)
}
#' Preserved seed and possibly set the seed
#'
#' @param seed R seed to use for the session
#'
#' @param code Is the code to evaluate
#'
#' @param rxseed is the rxode2 seed that is being preserved
#'
#' @return returns whatever the code is returning
#'
#' @inheritParams base::RNGkind
#'
#' @seealso rxGetSeed, rxSetSeed
#'
#' @examples
#'
#' rxGetSeed()
#' rxWithSeed(1, {
#'    print(rxGetSeed())
#'    rxnorm()
#'    print(rxGetSeed())
#'    rxnorm()
#' }, rxseed=3)
#'
#' @export
rxWithSeed <- function(seed, code, rxseed=rxGetSeed(), kind = "default", normal.kind = "default",
                       sample.kind = "default") {
  force(seed)
  force(rxseed)
  force(kind)
  force(normal.kind)
  force(sample.kind)
  .rxSeed <- .rxGetSeed()
  .origSeed <- .rxGetSeed()
  .newSeed <- .origSeed
  .newSeed$seed <- seed
  .newSeed$kind <- c(kind, normal.kind, sample.kind)
  .newSeed$rxseed <- rxseed
  on.exit(.rxSetSeed(.origSeed), add = TRUE)
  .rxSetSeed(.newSeed)
  force(code)
}

#' @rdname rxWithSeed
#' @export
rxWithPreserveSeed <- function(code) {
  .origSeed <- .rxGetSeed()
  on.exit(.rxSetSeed(.origSeed), add = TRUE)
  force(code)
}


#' Simulate from a (truncated) multivariate normal
#'
#' This is simulated with the fast, thread-safe threefry simulator
#' and can use multiple cores to generate the random deviates.
#'
#' @param n Number of random row vectors to be simulated OR the
#'     matrix to use for simulation (faster).
#'
#' @param mu mean vector
#'
#' @param sigma Covariance matrix for multivariate normal or a list
#'   of covariance matrices. If a list of covariance matrix, each
#'   matrix will simulate `n` matrices and combine them to a full
#'   matrix
#'
#' @param lower is a vector of the lower bound for the truncated
#'     multivariate norm
#'
#' @param upper is a vector of the upper bound for the truncated
#'     multivariate norm
#'
#' @param ncores Number of cores used in the simulation
#'
#' @param isChol A boolean indicating if `sigma` is a cholesky
#'     decomposition of the covariance matrix.
#'
#' @param keepNames Keep the names from either the mean or covariance
#'     matrix.
#'
#' @param a threshold for switching between methods; They can be
#'   tuned for maximum speed;  There are three cases that are considered:
#'
#'  case 1: a < l < u
#'
#'  case 2: l < u < -a
#'
#'  case 3: otherwise
#'
#' where l=lower and u = upper
#'
#' @param tol When case 3 is used from the above possibilities, the
#'   tol value controls the acceptance rejection and
#'   inverse-transformation;
#'
#' When abs(u-l)>tol, uses accept-reject from randn
#'
#' @param nlTol Tolerance for newton line-search
#'
#' @param nlMaxiter Maximum iterations for newton line-search
#'
#' @return
#'
#' If `n==integer` (default) the output is an (n x d) matrix
#' where the i-th row is the i-th simulated vector.
#'
#' If `is.matrix(n)` then the random vector are store in `n`,
#' which is provided by the user, and the function returns
#' `NULL` invisibly.
#'
#' @references John K. Salmon, Mark A. Moraes, Ron O. Dror, and David
#'     E. Shaw (2011). Parallel Random Numbers: As Easy as 1, 2, 3.
#'     D. E. Shaw Research, New York, NY 10036, USA.
#'
#' @examples
#'
#' ## From mvnfast
#' ## Unlike mvnfast, uses threefry simulation
#'
#' d <- 5
#' mu <- 1:d
#'
#' # Creating covariance matrix
#' tmp <- matrix(rnorm(d^2), d, d)
#' mcov <- tcrossprod(tmp, tmp)
#'
#'
#' set.seed(414)
#' rxRmvn(4, 1:d, mcov)
#'
#' set.seed(414)
#' rxRmvn(4, 1:d, mcov)
#'
#' set.seed(414)
#' rxRmvn(4, 1:d, mcov, ncores = 2) # r.v. generated on the second core are different
#'
#' ###### Here we create the matrix that will hold the simulated
#' #  random variables upfront.
#' A <- matrix(NA, 4, d)
#' class(A) <- "numeric" # This is important. We need the elements of A to be of class "numeric".
#'
#' set.seed(414)
#' rxRmvn(A, 1:d, mcov, ncores = 2) # This returns NULL ...
#' A # ... but the result is here
#'
#' ## You can also simulate from a truncated normal:
#'
#' rxRmvn(10, 1:d, mcov, lower = 1:d - 1, upper = 1:d + 1)
#'
#'
#' # You can also simulate from different matrices (if they match
#' # dimensions) by using a list of matrices.
#'
#' matL <- lapply(1:4, function(...) {
#'   tmp <- matrix(rnorm(d^2), d, d)
#'   tcrossprod(tmp, tmp)
#' })
#'
#'
#' rxRmvn(4, setNames(1:d, paste0("a", 1:d)), matL)
#' @author Matthew Fidler, Zdravko Botev and some from Matteo Fasiolo
#'
#' @references The thread safe multivariate normal was inspired from the `mvnfast` package by Matteo Fasiolo <https://CRAN.R-project.org/package=mvnfast>
#'
#' @references The concept of the truncated multivariate normal was
#'   taken from Zdravko Botev Botev (2017) \doi{10.1111/rssb.12162}
#'   and Botev and L'Ecuyer (2015) \doi{10.1109/WSC.2015.7408180} and
#'   converted to thread safe simulation;
#'
#' @export
rxRmvn <- function(n, mu = NULL, sigma, lower = -Inf, upper = Inf, ncores = 1, isChol = FALSE,
                   keepNames = TRUE, a = 0.4, tol = 2.05, nlTol = 1e-10, nlMaxiter = 100L) {
  .ret <- .Call(
    `_rxode2_rxRmvnSEXP`, n, mu, sigma, lower, upper, ncores,
    isChol, keepNames, a, tol, nlTol, nlMaxiter
  )
  if (is.matrix(n)) {
    return(invisible())
  }
  return(.ret)
}

#' @export
rxUdfUi.rxpois <- function(fun) {
  .fun <- fun
  .fun[[1]] <- str2lang(paste0(".", deparse1(fun[[1]])))
  eval(.fun)
}

#'@export
rxUdfUi.rxnbinom <- rxUdfUi.rxpois

#'  This processes the ui named functions to produce the correct
#'  underlying rxode2 code.
#'
#' In general this allows `rxnbinom()` and `ribinom()`to work more or less like
#' `rnbinom()` from R.
#'
#'
#' @inheritParams rbinom
#'
#' @param pre The prefix is to determine if this is a residual
#'   simulation (rx) or an individual simualtion (ri)
#'
#' @return A string for replacement in the ui code.
#'
#' @noRd
#' @author Matthew L. Fidler
.rxnbinom <- function(size, prob, mu, pre="rx") {
  .size <- as.character(substitute(size))
  .dp <- deparse1(substitute(size))
  .tmp <- suppressWarnings(try(force(size), silent = TRUE))
  .size <- .uiArg(.size, .tmp, .dp)

  if (!missing(mu)) {
    if (!missing(prob))
      stop("'prob' and 'mu' both specified")
    .mu <- as.character(substitute(mu))
    .dp <- deparse1(substitute(mu))
    .tmp <- suppressWarnings(try(force(mu), silent = TRUE))
    .mu <- .uiArg(.mu, .tmp, .dp)
    return(list(replace=paste0(pre, "nbinomMu(", .size, ", ", .mu, ")")))
  }
  .prob <- as.character(substitute(prob))
  .dp <- deparse1(substitute(prob))
  .tmp <- suppressWarnings(try(force(prob), silent = TRUE))
  .prob <- .uiArg(.prob, .tmp, .dp)
  list(replace = paste0(pre, "nbinom(", .size, ", ",  .prob, ")"))
}

#' @rdname rxUdfUi
#' @export
rxUdfUi.rinbinom <- function(fun) {
  .fun <- fun
  .fun[[1]] <- str2lang(".rxnbinom")
  .fun <- c(.fun, str2lang("prefix='ri'"))
  eval(.fun)
}

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("tests the internal df repetition routines", {
      expect_equal(
        .vecDf(c(a = 1, b = 1, c = 3), 3),
        structure(list(
          a = c(1, 1, 1),
          b = c(1, 1, 1),
          c = c(3, 3, 3)
        ),
        row.names = c(NA, -3L),
        class = "data.frame"
        )
      )
      expect_error(.vecDf(c(a = 1, b = 1, c = 3), 0))

      d <- 4

      rxWithSeed(
        42,
        matL <- lapply(1:4, function(...) {
          tmp <- matrix(rnorm(d^2), d, d)
          tcrossprod(tmp, tmp)
        })
      )

      theta <- data.frame(
        a = as.double(1:4), b = as.double(5:8),
        c = as.double(9:12)
      )

      omega <- rxRmvn(4, setNames(1:d, paste0("a", 1:d)), matL)

      expand <- .cbindOme(theta, omega, 4)

      expect_equal(expand$a, rep(theta$a, each = 4))
      expect_equal(expand$b, rep(theta$b, each = 4))
      expect_equal(expand$c, rep(theta$c, each = 4))

      expect_equal(expand$a1, omega[, "a1"])
      expect_equal(expand$a2, omega[, "a2"])
      expect_equal(expand$a3, omega[, "a3"])
      expect_equal(expand$a4, omega[, "a4"])

      expand2 <- .cbindOme(theta, NULL, 4)

      expect_equal(expand2$a, expand$a)
      expect_equal(expand2$b, expand$b)
      expect_equal(expand2$c, expand$c)

      expand3 <- .cbindOme(NULL, omega, 4)

      expect_equal(expand3$a1, expand$a1)
      expect_equal(expand3$b1, expand$b1)
      expect_equal(expand3$c1, expand$c1)
    })
  }
})

rxTest({

  warn1 <- function(code) {
    if (rxCores() == 1L) {
      x <- force(code)
    } else {
      expect_warning(x <- force(code))
    }
    x
  }

  test_that("rnorm", {

    rxWithSeed(1024, {

      rx <- rxode2({
        x1 <- rnorm()
        x2 <- rxnorm(a)
        x3 <- rnorm(b, c)
        d/dt(x0) <- 0
      })

      ev <- et(1, id = 1:70000)

      f <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 2))

      expect_equal(mean(f$x1), 0, tolerance = 1e-2)
      expect_equal(sd(f$x1), 1, tolerance = 1e-2)

      expect_equal(mean(f$x2), 3, tolerance = 1e-2)
      expect_equal(sd(f$x1), 1, tolerance = 1e-2)

      expect_equal(mean(f$x3), 5, tolerance = 1e-2)
      expect_equal(sd(f$x3), 2, tolerance = 1e-2)

      f2 <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 1))

      expect_equal(mean(f2$x1), 0, tolerance = 1e-2)
      expect_equal(sd(f2$x1), 1, tolerance = 1e-2)

      expect_equal(mean(f2$x2), 3, tolerance = 1e-2)
      expect_equal(sd(f2$x1), 1, tolerance = 1e-2)

      expect_equal(mean(f2$x3), 5, tolerance = 1e-2)
      expect_equal(sd(f2$x3), 2, tolerance = 1e-2)

      suppressMessages(expect_error(rxode2({
        x4 <- rnorm(a, b, c, d)
      })))

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

      x <- suppressMessages(rxnorm(n = 1e5))
      expect_equal(mean(x), 0, tolerance = 0.01)

      expect_equal(sd(x), 1, tolerance = 0.01)

    })
  })

  test_that("rbinom", {

    rx <- rxode2({
      x1 <- rbinom(4, 0.5)
      x2 <- rxbinom(10, 0.75)
    })

    ev <- et(1, id = 1:30000)

    rxWithSeed(1024, {

      f <- suppressMessages(rxSolve(rx, ev, cores = 2))

      expect_equal(max(f$x1), 4)
      expect_equal(min(f$x1), 0)
      expect_true(all(round(f$x1) == f$x1))
      expect_equal(mean(f$x1), 4 * 0.5, tolerance = 1e-2)
      expect_equal(sd(f$x1), sqrt(4 * 0.5 * 0.5), tolerance = 1e-2)

      expect_equal(max(f$x2), 10)
      expect_true(min(f$x2) > 0)
      expect_true(all(round(f$x2) == f$x2))

      expect_equal(mean(f$x2), 10 * 0.75, tolerance = 1e-2)
      expect_equal(sd(f$x2), sqrt(10 * 0.75 * 0.25), tolerance = 1e-2)

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

      suppressMessages(expect_error(rxode2({
        x1 <- rbinom()
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rbinom(a)
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rbinom(a, b, c)
      })))

    })

  })

  test_that("rxnbinom()", {

    rxWithSeed(1024, {

      n <- 1e5
      size <- 10
      prob <- 0.3
      # Call the function to generate negative binomial random deviates
      r <- rxnbinom(size=size, prob=prob, n=n)

      # Theoretical mean of the negative binomial distribution:
      mn <- round(size * (1 - prob) / prob, 1)
      mnr <- round(mean(r), 1)

      expect_equal(mn, mnr)

      mu <- 23.3

      r <- rxnbinomMu(size=size, mu=mu, n=n)

      # Theoretical mean of the negative binomial distribution:
      mnr <- round(mean(r), 1)

      expect_equal(mu, mnr)

      rx <- rxode2({
        x1 <- rxnbinom(10, 0.3)
        x2 <- rxnbinomMu(10, 23.3)
      })

      ev <- et(1:10000)

      f <- suppressMessages(rxSolve(rx, ev, cores = 2))

      expect_equal(round(mu), round(mean(f$x1)))
      expect_equal(round(mu), round(mean(f$x2)))

    })
  })

  test_that("rcauchy", {

    rxWithSeed(1024, {

      rx <- rxode2({
        x1 <- rcauchy()
        x2 <- rxcauchy(a)
        x3 <- rcauchy(b, c)
        d / dt(x0) <- 0
      })

      ev <- et(1, id = 1:100)

      f <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 2))
      ## Seed tests

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

      suppressMessages(expect_error(rxode2({
        x4 <- rcauchy(a, b, c, d)
      })))

    })

  })

  test_that("rchisq", {

    rx <- rxode2({
      x1 <- rchisq(15)
      x2 <- rxchisq(20)
    })

    ev <- et(1, id = 1:30000)

    rxWithSeed(1024, {

      f <- suppressMessages(rxSolve(rx, ev, cores = 2))

      expect_equal(mean(f$x1), 15, tolerance = 0.1)
      expect_equal(sd(f$x1), sqrt(2 * 15), tolerance = 0.1)

      expect_equal(mean(f$x2), 20, tolerance = 0.1)
      expect_equal(sd(f$x2), sqrt(2 * 20), tolerance = 0.1)

      ## Seed tests

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

      suppressMessages(expect_error(rxode2({
        x1 <- rchisq()
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rchisq(a, b)
      })))

    })
  })

  test_that("rexp tests", {

    rx <- rxode2({
      x1 <- rexp(0.5)
      x2 <- rxexp()
    })

    ev <- et(1, id = 1:30000)

    rxWithSeed(1024, {

      f <- suppressMessages(rxSolve(rx, ev, cores = 2))

      expect_equal(mean(f$x1), 2, tolerance = 0.1)
      expect_equal(sd(f$x1), sqrt(1 / (0.5 * 0.5)), tolerance = 0.1)

      expect_equal(mean(f$x2), 1, tolerance = 0.1)
      expect_equal(sd(f$x2), 1, tolerance = 0.1)

      ## Seed tests

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

      suppressMessages(expect_error(rxode2({
        x1 <- rexp(a, b)
      })))

    })

  })

  test_that("rf tests", {

    rx <- rxode2({
      x1 <- rf(10, 20)
      x2 <- rxf(30, 40)
    })

    ev <- et(1, id = 1:40000)

    rxWithSeed(1024, {

      f <- suppressMessages(rxSolve(rx, ev, cores = 2))

      sf <- function(d1, d2) {
        sqrt((2 * d2^2 * (d1 + d2 - 2)) / (d1 * (d2 - 2)^2 * (d2 - 4)))
      }

      mf <- function(d2) {
        d2 / (d2 - 2)
      }

      expect_equal(mean(f$x1), mf(20), tolerance = 0.01)
      expect_equal(sd(f$x1), sf(10, 20), tolerance = 0.02)

      expect_equal(mean(f$x2), mf(40), tolerance = 0.01)
      expect_equal(sd(f$x2), sf(30, 40), tolerance = 0.02)

      ## Seed tests

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

      suppressMessages(expect_error(rxode2({
        x1 <- rf(a, b, c)
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rf(a)
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rf()
      })))

    })
  })

  test_that("rgamma tests", {

    rx <- rxode2({
      x1 <- rgamma(9, 0.5)
      x2 <- rxgamma(7.5)
    })

    ev <- et(1, id = 1:30000)

    rxWithSeed(1024, {

      f <- suppressMessages(rxSolve(rx, ev, cores = 2))

      sgamma <- function(k, theta = 1) {
        sqrt(k / (theta^2))
      }

      expect_equal(sd(f$x1), sgamma(9, 0.5), tolerance = 0.01)

      expect_equal(sd(f$x2), sgamma(7.5), tolerance = 0.01)

      ## Seed tests

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

      suppressMessages(expect_error(rxode2({
        x1 <- rgamma(a, b, c)
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rgamma()
      })))

    })
  })

  test_that("rbeta tests", {

    rx <- rxode2({
      x1 <- rbeta(2, 5)
      x2 <- rxbeta(2, 2)
    })

    ev <- et(1, id = 1:30000)

    rxWithSeed(1024, {

      f <- suppressMessages(rxSolve(rx, ev, cores = 2))

      mbeta <- function(a, b) {
        a / (a + b)
      }
      sbeta <- function(a, b) {
        sqrt(a * b / ((a + b)^2 * (a + b + 1)))
      }

      expect_equal(mean(f$x1), mbeta(2, 5), tolerance = 0.01)
      # Using tolerance=0.02 for the Mac random number generator
      expect_equal(sd(f$x1), sbeta(2, 5), tolerance = 0.02)

      expect_equal(mean(f$x2), mbeta(2, 2), tolerance = 0.01)
      expect_equal(sd(f$x2), sbeta(2, 2), tolerance = 0.01)

      ## Seed tests

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

      suppressMessages(expect_error(rxode2({
        x1 <- rbeta(a, b, c)
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rbeta(a)
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rbeta()
      })))

    })

  })

  test_that("rgeom tests", {

    rx <- rxode2({
      # x1 <- rgeom(0.5)
      x2 <- rxgeom(0.1)
    })

    ev <- et(1, id = 1:30000)

    rxWithSeed(1024, {

      f <- suppressMessages(rxSolve(rx, ev, cores = 2))

      # expect_equal(median(f$x1), -ceiling(1 / log2(1 - 0.5)))
      expect_equal(median(f$x2), -ceiling(1 / log2(1 - 0.1)))

      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

      suppressMessages(expect_error(rxode2({
        x1 <- rgeom()
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rgeom(a, b)
      })))

    })
  })

  test_that("rpois", {

    rx <- rxode2({
      x1 <- rpois(1)
      x2 <- rxpois(2)
    })

    ev <- et(1, id = 1:30000)

    rxWithSeed(1024, {

      f <- suppressMessages(rxSolve(rx, ev, cores = 2))

      expect_equal(mean(f$x1), 1, tolerance = 0.02)
      expect_equal(sd(f$x1), 1, tolerance = 0.02)

      expect_equal(mean(f$x2), 2, tolerance = 0.02)
      expect_equal(sd(f$x2), sqrt(2), tolerance = 0.02)
      ## Seed tests

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

      suppressMessages(expect_error(rxode2({
        x1 <- rpois()
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rxpois(a, b)
      })))

    })
  })

  test_that("rt", {

    rx <- rxode2({
      x1 <- rt(15)
      x2 <- rxt(20)
    })

    ev <- et(1, id = 1:30000)

    rxWithSeed(1024, {

      f <- suppressMessages(rxSolve(rx, ev, cores = 2))

      expect_equal(mean(f$x1), 0, tolerance = 0.1)
      expect_equal(sd(f$x1), sqrt(15 / (15 - 2)), tolerance = 0.1)

      expect_equal(mean(f$x2), 0, tolerance = 0.1)
      expect_equal(sd(f$x2), sqrt(20 / (20 - 2)), tolerance = 0.1)

      ## Seed tests

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))


      suppressMessages(expect_error(rxode2({
        x1 <- rt()
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rt(a, b)
      })))

    })
  })

  test_that("runif", {

    rxWithSeed(1024, {

      rx <- rxode2({
        x1 <- runif()
        x2 <- rxunif(a)
        x3 <- runif(b, c)
        d / dt(x0) <- 0
      })

      ev <- et(1, id = 1:30000)

      f <- suppressMessages(rxSolve(rx, ev, c(a = 0.5, b = 0.25, c = 0.75), cores = 2))

      expect_equal(mean(f$x1), 0.5, tolerance = 1e-2)
      expect_equal(sd(f$x1), sqrt(1 / 12), tolerance = 1e-2)

      expect_equal(mean(f$x2), 0.5 * (0.5 + 1), tolerance = 1e-2)
      expect_equal(sd(f$x2), sqrt((1 - 0.5)^2 / 12), tolerance = 1e-2)

      expect_equal(mean(f$x3), 0.5 * (0.25 + 0.75), tolerance = 1e-2)
      expect_equal(sd(f$x3), sqrt((0.75 - 0.25)^2 / 12), tolerance = 1e-2)

      f2 <- suppressMessages(rxSolve(rx, ev, c(a = 0.5, b = 0.25, c = 0.75), cores = 1))

      expect_equal(mean(f2$x1), 0.5, tolerance = 1e-2)
      expect_equal(sd(f2$x1), sqrt(1 / 12), tolerance = 1e-2)

      expect_equal(mean(f2$x2), 0.5 * (0.5 + 1), tolerance = 1e-2)
      expect_equal(sd(f2$x2), sqrt((1 - 0.5)^2 / 12), tolerance = 1e-2)

      expect_equal(mean(f2$x3), 0.5 * (0.25 + 0.75), tolerance = 1e-2)
      expect_equal(sd(f2$x3), sqrt((0.75 - 0.25)^2 / 12), tolerance = 1e-2)

      suppressMessages(expect_error(rxode2({
        x4 <- runif(a, b, c, d)
      })))

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

    })
  })


  test_that("rweibull tests", {

    rx <- rxode2({
      x1 <- rweibull(9, 0.5)
      x2 <- rxweibull(7.5)
    })

    ev <- et(1, id = 1:30000)

    rxWithSeed(1024, {

      f <- suppressMessages(rxSolve(rx, ev, cores = 2))

      mweibull <- function(shape, scale = 1) {
        lambda <- scale
        k <- shape
        lambda * gamma(1 + 1 / k)
      }

      sweibull <- function(shape, scale = 1) {
        lambda <- scale
        k <- shape
        sqrt(lambda^2 * (gamma(1 + 2 / k)
          - (gamma(1 + 1 / k))^2))
      }

      expect_equal(mean(f$x1), mweibull(9, 0.5), tolerance = 0.01)
      expect_equal(sd(f$x1), sweibull(9, 0.5), tolerance = 0.01)

      expect_equal(mean(f$x2), mweibull(7.5), tolerance = 0.01)
      expect_equal(sd(f$x2), sweibull(7.5), tolerance = 0.01)

      ## Seed tests

      ## Make sure seeds are reproducible
      ev <- et(1, id = 1:10)

      set.seed(1)
      f <- suppressMessages(rxSolve(rx, ev, cores = 1))

      set.seed(1)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))
      expect_equal(as.data.frame(f), as.data.frame(f2))

      ## Make sure different seed value gives different result
      set.seed(2)
      f2 <- suppressMessages(rxSolve(rx, ev, cores = 1))

      expect_false(isTRUE(all.equal(as.data.frame(f), as.data.frame(f2))))

      suppressMessages(expect_error(rxode2({
        x1 <- rweibull(a, b, c)
      })))

      suppressMessages(expect_error(rxode2({
        x1 <- rweibull()
      })))

    })
  })

  test_that("individual random variable tests", {

    rx <- rxode2({
      x0 <- rxnorm()
      x1 <- rinorm(a)
      x2 <- rinorm(b, c)
      x3 <- rinorm()
      x7 <- ricauchy()
      x8 <- ricauchy(a)
      x9 <- ricauchy(b, c)
      x10 <- richisq(15)
      x11 <- riexp(0.5)
      x12 <- rif(10, 20)
      x13 <- rigamma(9, 0.5)
      x14 <- rigamma(7.5)
      x15 <- rit(20)
      x16 <- riunif()
      x17 <- riunif(a)
      x18 <- riunif(b, c)
      x19 <- riweibull(9, 0.5)
      x20 <- riweibull(7.5)
      ## int, likely to repeat
      x21 <- ripois(1)
      x22 <- ribeta(2, 5) ## ?
      x23 <- rigeom(0.5)
      x24 <- ribinom(10, 0.5)
      ##
      d / dt(xx) <- 0
    })

    rxWithSeed(10, {

      ev <- et(c(1, 2), id = 1:5)

      f <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 2))

      expect_equal(sum(duplicated(f$x0)), 0)

      for (i in 1:20) {
        expect_equal(sum(duplicated(paste0(f$id, f[[paste0("x", i)]]))), 5)
        .s <- sum(duplicated(f[[paste0("x", i)]]))
        expect_true(.s < 10)
      }

      rx <- rxode2({
        x0 <- rxnorm()
        x1 <- rinorm(a)
        x2 <- rinorm(b, c)
        x3 <- rinorm()
        x7 <- ricauchy()
        x8 <- ricauchy(a)
        x9 <- ricauchy(b, c)
        x10 <- richisq(15)
        x11 <- riexp(0.5)
        x12 <- rif(10, 20)
        x13 <- rigamma(9, 0.5)
        x14 <- rigamma(7.5)
        x15 <- rit(20)
        x16 <- riunif()
        x17 <- riunif(a)
        x18 <- riunif(b, c)
        x19 <- riweibull(9, 0.5)
        x20 <- riweibull(7.5)
        ## int, likely to repeat
        x21 <- ripois(1)
        x22 <- ribeta(2, 5) ## ?
        x23 <- rigeom(0.5)
        x24 <- ribinom(10, 0.5)
        ##
      })

      set.seed(10)

      ev <- et(c(1, 2), id = 1:5)

      f <- suppressMessages(rxSolve(rx, ev, c(a = 3, b = 5, c = 2), cores = 2))

      expect_equal(sum(duplicated(f$x0)), 0)

      for (i in 1:20) {
        expect_equal(sum(duplicated(paste0(f$id, f[[paste0("x", i)]]))), 5)
        .s <- sum(duplicated(f[[paste0("x", i)]]))
        expect_true(.s < 10)
      }

    })

  })

  test_that("simeps", {

    rx1 <- rxode2({
      c <- 0 + err
      i <- 0
    })

    e <- et(0, 10)

    rxWithSeed(10, {

      f1 <- suppressMessages(rxSolve(rx1, e, sigma = lotri(err ~ 1)))

      expect_true(f1$c[1] != 0)

      rx <- rxode2({
        c <- 0 + err
        i <- 0
        while (c < 0) {
          simeps()
          c <- 0 + err
          i <- i + 1
          if (i > 10) break
        }
      })

      set.seed(10)
      f2 <- suppressMessages(rxSolve(rx, e, sigma = lotri(err ~ 1)))

      expect_true(f2$c[1] != 0)

      expect_false(all(f1$c > 0))

      expect_true(all(f2$c > 0))

      f3 <- f2[f2$i == 0, c("time", "c")]

      f3 <- merge(f1, f3, by = "time")

      ## If the condition is already satisfied, it should keep the originally simulated values
      expect_equal(f3$c.x, f3$c.y)

      set.seed(10)
      f1 <- suppressMessages(rxSolve(rx, e, sigma = lotri(err ~ 1), nStud = 3))

      expect_true(all(f1$c > 0))

      expect_true(f1$c[1] != 0)

      set.seed(10)
      f2 <- suppressMessages(rxSolve(rx1, e, sigma = lotri(err ~ 1), nStud = 3))

      expect_false(all(f2$c > 0))

      expect_true(f2$c[1] != 0)

      f3 <- merge(f1, f2, by = c("sim.id", "time"))

      f3 <- f3[f3$i == 0, ]

      expect_equal(f3$c.x, f3$c.y)

      set.seed(10)
      f1 <- suppressMessages(rxSolve(rx, e, sigma = lotri(err ~ 1), nStud = 3, dfObs = 100))

      expect_true(all(f1$c > 0))

      expect_true(f1$c[1] != 0)


      set.seed(10)
      f2 <- suppressMessages(rxSolve(rx1, e, sigma = lotri(err ~ 1), nStud = 3, dfObs = 100))

      expect_false(all(f2$c > 0))

      expect_true(f2$c[1] != 0)

      f3b <- merge(f1, f2, by = c("sim.id", "time"))

      f3b <- f3b[f3b$i == 0, ]

      expect_equal(f3b$c.x, f3b$c.y)

      expect_false(identical(f3b$c.x, f3$c.x))
      expect_false(identical(f3b$c.y, f3$c.y))

      ## Check to make sure that this only accesses the
      f1 <- suppressMessages(rxSolve(rx, e, sigma = lotri(err ~ 1), nStud = 3))

      expect_true(all(f1$c > 0))

      expect_true(f1$c[1] != 0)
    })
  })

  test_that("simeta", {
    rx <- rxode2({
      wt <- 70 * exp(eta.wt)
      i <- 0
      while ((wt < 60) || (wt > 80)) {
        i <- i + 1
        if (i > 100) break
        simeta()
        wt <- 70 * exp(eta.wt)
      }
    })
    rxWithPreserveSeed({
      e <- et(1:2, id = 1:4)

      f <- suppressMessages(rxSolve(rx, e, omega = lotri(eta.wt ~ 0.1^2)))

      expect_true(all(f$wt > 60))
      expect_true(all(f$wt < 80))

      expect_equal(length(unique(f$wt)), 4)

      f <- suppressMessages(rxSolve(rx, e, omega = lotri(eta.wt ~ 0.5^2), nStud = 10))

      expect_true(all(f$wt > 60))
      expect_true(all(f$wt < 80))

      expect_equal(length(unique(f$wt)), 4 * 10)

      ## this one should work
      f <- suppressMessages(rxSolve(rx, e, omega = lotri(eta.wt ~ 0.5^2), nStud = 3, dfSub = 40))

      expect_true(all(f$wt > 60))
      expect_true(all(f$wt < 80))

      expect_equal(length(unique(f$wt)), 4 * 3)

    })
  })

  test_that("random variables work in R alone", {
    rxWithSeed(1024, {

      expect_true(is.numeric(rxcauchy()))

      p <- rxpois(2, n = 30000)
      expect_equal(mean(p), 2, tolerance = 0.01)
      expect_equal(sd(p), sqrt(2), tolerance = 0.01)

      r <- rxt(15, n = 30000)
      expect_equal(mean(r), 0, tolerance = 0.1)
      expect_equal(sd(r), sqrt(15 / (15 - 2)), tolerance = 0.1)

      r <- rxbinom(4, 0.5, n = 30000)
      expect_equal(max(r), 4)
      expect_equal(min(r), 0)
      expect_equal(mean(r), 4 * 0.5, tolerance = 1e-2)
      expect_equal(sd(r), sqrt(4 * 0.5 * 0.5), tolerance = 1e-2)

      chi <- rxchisq(15, n = 30000)
      expect_equal(mean(chi), 15, tolerance = 0.1)
      expect_equal(sd(chi), sqrt(2 * 15), tolerance = 0.1)

      xp <- rxexp(0.5, n = 30000)
      expect_equal(mean(xp), 2, tolerance = 0.1)
      expect_equal(sd(xp), sqrt(1 / (0.5 * 0.5)), tolerance = 0.1)

      f <- rxf(30, 40, n = 30000)

      sf <- function(d1, d2) {
        sqrt((2 * d2^2 * (d1 + d2 - 2)) / (d1 * (d2 - 2)^2 * (d2 - 4)))
      }

      mf <- function(d2) {
        d2 / (d2 - 2)
      }

      expect_equal(mean(f), mf(40), tolerance = 0.01)
      expect_equal(sd(f), sf(30, 40), tolerance = 0.1)

      x2 <- rxgamma(7.5, n = 30000)

      sgamma <- function(k, theta = 1) {
        sqrt(k / (theta^2))
      }

      ## expect_equal(sd(x2), sgamma(7.5), tolerance = 0.01)

      x2 <- rxbeta(2, 2, n = 30000)

      mbeta <- function(a, b) {
        a / (a + b)
      }

      sbeta <- function(a, b) {
        sqrt(a * b / ((a + b)^2 * (a + b + 1)))
      }

      expect_equal(mean(x2), mbeta(2, 2), tolerance = 0.01)
      expect_equal(sd(x2), sbeta(2, 2), tolerance = 0.01)

      x2 <- rxgeom(0.1, n = 30000)

      expect_equal(median(x2), -ceiling(1 / log2(1 - 0.1)))

      x2 <- rxpois(2, n = 30000)

      expect_equal(mean(x2), 2, tolerance = 0.01)
      expect_equal(sd(x2), sqrt(2), tolerance = 0.01)

      x2 <- rxunif(0.5, n = 30000)

      expect_equal(mean(x2), 0.5 * (0.5 + 1), tolerance = 1e-2)
      expect_equal(sd(x2), sqrt((1 - 0.5)^2 / 12), tolerance = 1e-2)

      x2 <- rxweibull(7.5, n = 30000)

      mweibull <- function(shape, scale = 1) {
        lambda <- scale
        k <- shape
        lambda * gamma(1 + 1 / k)
      }

      sweibull <- function(shape, scale = 1) {
        lambda <- scale
        k <- shape
        sqrt(lambda^2 * (gamma(1 + 2 / k)
          - (gamma(1 + 1 / k))^2))
      }

      expect_equal(mean(x2), mweibull(7.5), tolerance = 0.01)
      expect_equal(sd(x2), sweibull(7.5), tolerance = 0.01)
    })

  })


  test_that("rxord", {

    rxWithSeed(1024, {
      rx <- rxode2({
        tmp2 <- rxord(0.5)
        tmp3 <- rxord(0.33, 0.33)
        tmp4 <- rxord(0.25, 0.25, 0.25)
      })

      n <- 100000
      ev <- et(seq(1, n))

      f <- rxSolve(rx, ev)

      expect_equal(round(as.numeric(table(f$tmp2))/ n, 3), c(0.5, 0.5), tolerance=1e-2)
      expect_equal(round(as.numeric(table(f$tmp3))/n, 3), c(0.33, 0.33, 0.33), tolerance=1e-1)
      expect_equal(round(as.numeric(table(f$tmp4))/n, 3), c(0.25, 0.25, 0.25, 0.25), tolerance=1e-2)

      tmp2 <- vapply(seq(1, n), function(i){ rxord(0.5) }, numeric(1), USE.NAMES=TRUE)

      expect_equal(round(as.numeric(table(tmp2))/ n, 3), c(0.5, 0.5), tolerance=1e-2)

      tmp3 <- vapply(seq(1, n), function(i){ rxord(0.33, 0.33) }, numeric(1), USE.NAMES=TRUE)
      expect_equal(round(as.numeric(table(tmp3))/ n, 3), c(0.33, 0.33, 0.33), tolerance=1e-1)

      tmp4 <- vapply(seq(1, n), function(i){ rxord(0.25, 0.25, 0.25) }, numeric(1), USE.NAMES=TRUE)
      expect_equal(round(as.numeric(table(tmp4))/ n, 3), c(0.25, 0.25, 0.25, 0.25), tolerance=1e-2)

    })


  })

  test_that("rnorm", {
    rxWithSeed(1024, {
      x <- rxnorm(n = 1e5)
      expect_equal(mean(x), 0, tolerance = 0.01)
      expect_equal(sd(x), 1, tolerance = 0.01)
    })
  })

  test_that("random variables work in R alone", {
    rxWithSeed(1024, {
      expect_true(is.numeric(rxcauchy()))

      p <- rxpois(2, n = 30000)
      expect_equal(mean(p), 2, tolerance = 0.01)
      expect_equal(sd(p), sqrt(2), tolerance = 0.01)

      r <- rxt(15, n = 30000)
      expect_equal(mean(r), 0, tolerance = 0.1)
      expect_equal(sd(r), sqrt(15 / (15 - 2)), tolerance = 0.1)

      r <- rxbinom(4, 0.5, n = 30000)
      expect_equal(max(r), 4)
      expect_equal(min(r), 0)
      expect_equal(mean(r), 4 * 0.5, tolerance = 1e-2)
      expect_equal(sd(r), sqrt(4 * 0.5 * 0.5), tolerance = 1e-2)

      chi <- rxchisq(15, n = 30000)
      expect_equal(mean(chi), 15, tolerance = 0.1)
      expect_equal(sd(chi), sqrt(2 * 15), tolerance = 0.1)

      xp <- rxexp(0.5, n = 30000)
      expect_equal(mean(xp), 2, tolerance = 0.1)
      expect_equal(sd(xp), sqrt(1 / (0.5 * 0.5)), tolerance = 0.1)

      f <- rxf(30, 40, n = 30000)

      sf <- function(d1, d2) {
        sqrt((2 * d2^2 * (d1 + d2 - 2)) / (d1 * (d2 - 2)^2 * (d2 - 4)))
      }

      mf <- function(d2) {
        d2 / (d2 - 2)
      }

      expect_equal(mean(f), mf(40), tolerance = 0.01)
      expect_equal(sd(f), sf(30, 40), tolerance = 0.1)

      x2 <- rxgamma(7.5, n = 30000)

      sgamma <- function(k, theta = 1) {
        sqrt(k / (theta^2))
      }

      ## expect_equal(sd(x2), sgamma(7.5), tolerance = 0.01)

      x2 <- rxbeta(2, 2, n = 30000)

      mbeta <- function(a, b) {
        a / (a + b)
      }

      sbeta <- function(a, b) {
        sqrt(a * b / ((a + b)^2 * (a + b + 1)))
      }

      expect_equal(mean(x2), mbeta(2, 2), tolerance = 0.01)
      expect_equal(sd(x2), sbeta(2, 2), tolerance = 0.01)

      x2 <- rxgeom(0.1, n = 30000)

      expect_equal(median(x2), -ceiling(1 / log2(1 - 0.1)))

      x2 <- rxpois(2, n = 30000)

      expect_equal(mean(x2), 2, tolerance = 0.01)
      expect_equal(sd(x2), sqrt(2), tolerance = 0.01)

      x2 <- rxunif(0.5, n = 30000)

      expect_equal(mean(x2), 0.5 * (0.5 + 1), tolerance = 1e-2)
      expect_equal(sd(x2), sqrt((1 - 0.5)^2 / 12), tolerance = 1e-2)

      x2 <- rxweibull(7.5, n = 30000)

      mweibull <- function(shape, scale = 1) {
        lambda <- scale
        k <- shape
        lambda * gamma(1 + 1 / k)
      }

      sweibull <- function(shape, scale = 1) {
        lambda <- scale
        k <- shape
        sqrt(lambda^2 * (gamma(1 + 2 / k)
          - (gamma(1 + 1 / k))^2))
      }

      expect_equal(mean(x2), mweibull(7.5), tolerance = 0.01)
      expect_equal(sd(x2), sweibull(7.5), tolerance = 0.01)
    })

  })

  test_that("udf handling of rxbinomMu", {

    f <- function() {
      model({
        x <- rxnbinom(size=10, mu=0.5)
      })
    }

    f <- f()

    expect_equal(modelExtract(f, "x"),
                 "x <- rxnbinomMu(10, 0.5)")

    f <- function() {
      model({
        x <- rxnbinom(mu=0.5, size=10)
      })
    }

    f <- f()

    expect_equal(modelExtract(f, "x"),
                 "x <- rxnbinomMu(10, 0.5)")

    f <- function() {
      model({
        x <- rxnbinom(prob=0.5, size=10)
      })
    }

    f <- f()

    expect_equal(modelExtract(f, "x"),
                 "x <- rxnbinom(10, 0.5)")

    f <- function() {
      model({
        x <- rxnbinom(size=10, prob=0.5)
      })
    }

    f <- f()

    expect_equal(modelExtract(f, "x"),
                 "x <- rxnbinom(10, 0.5)")

    expect_error(rxnbinom(size=10, prob=.05, mu=30))

    expect_equal(rxWithSeed(1024, rxnbinom(size=10, mu=30)),
                 rxWithSeed(1024, rxnbinomMu(size=10, mu=30)))


  })

})

rxTest({
  test_that("Test Issue RxODE#163 -- Keep doesn't work with iCov", {

    rxWithSeed(
      42,
      {
        mod <- rxode2({
          C2 <- centr / V2
          C3 <- peri / V3
          KA <- TKA * exp(eta.ka + scale_rad + 0.5 * SEX)
          d / dt(depot) <- -KA * depot
          d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
          d / dt(peri) <- Q * C2 - Q * C3
          d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
        })

        theta <- c(
          TKA = 2.94E-01, CL = 1.86E+01, V2 = 4.02E+01, # central
          Q = 1.05E+01, V3 = 2.97E+02, # peripheral
          Kin = 1, Kout = 1, EC50 = 200
        ) # effects

        rdunif <- function(...) {
          round(runif(...))
        }

        et(amount.units = "mg", time.units = "hours") |>
          et(dose = 10000, addl = 9, ii = 12) |>
          et(dose = 20000, addl = 4, time = 120, ii = 24) |>
          et(id = 1:10) |>
          et(0:240) ->
          ev3_hum

        sim1 <- rxSolve(mod, theta,
                        ev = ev3_hum, omega = lotri(eta.ka ~ 0.01),
                        iCov = data.frame(id = 1:10, scale_rad = 3.5, SEX = rdunif(10, 0, 1)),
                        keep = c("SEX", "scale_rad")
                        )

        expect_true(all(sim1$rad == 3.5))

        sim1 <- rxSolve(mod, theta,
                        ev = ev3_hum, omega = lotri(eta.ka ~ 0.01),
                        iCov = data.frame(id = 1:10, scale_rad = 3.5, SEX = rdunif(10, 0, 1)),
                        keep = c("scale_rad", "SEX")
                        )

        expect_true(all(sim1$rad == 3.5))
      }
    )
  })
})

rxTest({
  test_that("test omegaCubeConvert", {

    matLst <- cvPost(10, lotri::lotri(a+b~c(1, 0.25, 1)), 3)
    expect_true(inherits(matLst, "list"))

    matCube <- swapMatListWithCube(matLst)
    expect_length(dim(matCube), 3L)

    matLst2 <- swapMatListWithCube(matCube)

    expect_identical(matLst2, matLst)
  })
})

rxTest({
  test_that("nesting works for single case #842", {

    library(dplyr)
    rxWithSeed(1234, {
      #Set model code
      model <-
        RxODE({
          #Parameters
          KA <- KA * exp(eta.KA)
          Fa <- Fa * exp(eta.Fa + iov.Fa) # IIV and IOV on bioavailability
          #Initial conditions
          A2(0) = 0
          #Model
          d/dt(DEPOT) = -KA * DEPOT
          f(DEPOT) = Fa
          d/dt(A2) = KA * DEPOT - (CL/V)*A2
          Cp = A2/V * 1000
        })

      #Define thetas
      theta <- c(KA = 0.925,   # Absorption rate constant (hr-1)
                 V  = 0.431,   # Central volume (mL)
                 CL = 0.137,  # Central clearance (mL/hr)
                 Fa = 0.134)

      #Define omegas
      omega <- lotri(lotri(eta.KA ~ 0.293^2,
                           eta.Fa ~ 0.0836^2) | id(nu=20),
                     lotri(iov.Fa ~ 0.767^2) | occ(nu=10))

      #Define event table
      ev <- et(amount.units = "mg", time.units = "hours") |>
        et(amt = 30*0.25 , cmt = "DEPOT", addl = 9, ii = 12) |> #Dose is 30mg/kg, multiply by average rat weight of 250 grams
        et(0,120) |>
        et(id = 1:20) |>
        mutate(occ = ceiling(as.numeric(time) / 12)) |> # 10 occasions in total
        mutate(occ = ifelse(occ == 0, 1, occ))

      expect_error(rxSolve(object=model, params=theta, events=ev, omega=omega, sigma=NULL), NA)

      results <- rxSolve(object=model, params=theta, events=ev, omega=omega, sigma=NULL)

      expect_equal(names(results$params),
                   c("id", "iov.Fa(occ==1)", "iov.Fa(occ==2)", "iov.Fa(occ==3)",
                     "iov.Fa(occ==4)", "iov.Fa(occ==5)", "iov.Fa(occ==6)", "iov.Fa(occ==7)",
                     "iov.Fa(occ==8)", "iov.Fa(occ==9)", "iov.Fa(occ==10)", "KA",
                     "eta.KA", "Fa", "eta.Fa", "CL", "V"))

    })


  })
})

rxTest({
  rxWithSeed(
    42,
    {

      m1 <- rxode2({
        CL ~ (1 - 0.2 * SEX) * (0.807 + 0.00514 * (CRCL - 91.2)) * exp(eta.cl)
        V1 ~ 4.8 * exp(eta.v1)
        Q ~ (3.46 + 0.0593 * (WT - 75.1)) * exp(eta.q)
        V2 ~ 1.93 * (3.13 + 0.0458 * (WT - 75.1)) * exp(eta.v2)
        cp <- max(linCmt() + err.sd, 0.01)
        if (cp == 0.01) cp <- NA
        mSEX <- SEX
        mWT <- WT
        mCRCL <- CRCL
      })

      # Make non-random covariates for testing
      AGE <- round(seq(18, 18 + 29))
      SEX <- c(rep(0, 15), rep(1, 15))
      WT <- seq(60, 60 + 29)
      CRCL <- seq(30, 30 + 29)
      ## id is in lower case to match the event table
      cov.df <- dplyr::tibble(id = seq_along(AGE), AGE = AGE, SEX = SEX, WT = WT, CRCL = CRCL)

      s <- c(0, 0.25, 0.5, 0.75, 1, 1.5, seq(2, 24, by = 1))
      ## Add 10% diff
      s <- lapply(s, function(x) {
        d <- x * 0.1
        c(x - d, x + d)
      })

      e <- et() |>
        ## Specify the id and weight based dosing from covariate data.frame
        ## This requires rxode2 XXX
        et(id = cov.df$id, amt = 6 * cov.df$WT, rate = 6 * cov.df$WT) |>
        ## Sampling is added for each ID
        et(s) |>
        as.data.frame() |>
        ## Merge the event table with the covarite information
        merge(cov.df, by = "id") |>
        dplyr::as_tibble()

      e2 <- et() |>
        ## Specify the id and weight based dosing from covariate data.frame
        ## This requires rxode2 XXX
        et(id = cov.df$id, amt = 6 * cov.df$WT, rate = 6 * cov.df$WT) |>
        ## Sampling is added for each ID
        et(s)

      e$WT <- e$WT + e$time / 30
      e$CRCL <- e$CRCL + e$time / 30

      test_that("test resampleID behavior", {
        for (nStud in c(1, 2)) {

          f1 <- rxSolve(m1, e,
                        ## Lotri uses lower-triangular matrix rep. for named matrix
                        omega = lotri(
                          eta.cl ~ .306,
                          eta.q ~ 0.0652,
                          eta.v1 ~ .567,
                          eta.v2 ~ .191
                        ),
                        sigma = lotri(err.sd ~ 0.5), addCov = TRUE,
                        addDosing = TRUE, nStud = nStud
                        )

          if (nStud == 1) {
            expect_equal(f1$WT, e$WT)
            expect_equal(f1$CRCL, e$CRCL)
            expect_equal(f1$SEX, e$SEX)
          }

          f2 <- rxSolve(m1, e,
                        ## Lotri uses lower-triangular matrix rep. for named matrix
                        omega = lotri(
                          eta.cl ~ .306,
                          eta.q ~ 0.0652,
                          eta.v1 ~ .567,
                          eta.v2 ~ .191
                        ),
                        sigma = lotri(err.sd ~ 0.5), addCov = TRUE,
                        resample = c("SEX", "WT", "CRCL"),
                        resampleID = TRUE,
                        addDosing = TRUE, nStud = nStud
                        )

          expect_equal(f2$mWT, f2$WT)
          expect_equal(f2$mCRCL, f2$CRCL)
          expect_equal(f2$mSEX, f2$SEX)

          r1 <- f1[!duplicated(f1$id), c("id", "SEX", "WT", "CRCL")]
          r2 <- f2[!duplicated(f2$id), c("id", "SEX", "WT", "CRCL")]

          expect_false(isTRUE(all.equal(r1, r2)))

          expect_true(isTRUE(all.equal(r1$WT - r1$CRCL, rep(30, nrow(r1)))))
          expect_true(isTRUE(all.equal(r2$WT - r2$CRCL, rep(30, nrow(r2)))))

          expect_true(all(r1$SEX[floor(r1$CRCL) <= 44] == 0))
          expect_true(all(r1$SEX[floor(r1$CRCL) > 44] == 1))

          expect_true(all(r2$SEX[floor(r2$CRCL) <= 44] == 0))
          expect_true(all(r2$SEX[floor(r2$CRCL) > 44] == 1))

          f3 <- rxSolve(m1, e,
                        ## Lotri uses lower-triangular matrix rep. for named matrix
                        omega = lotri(
                          eta.cl ~ .306,
                          eta.q ~ 0.0652,
                          eta.v1 ~ .567,
                          eta.v2 ~ .191
                        ),
                        sigma = lotri(err.sd ~ 0.5), addCov = TRUE,
                        resample = c("SEX", "WT", "CRCL"),
                        resampleID = FALSE,
                        addDosing = TRUE, nStud = nStud
                        )

          expect_equal(f3$mWT, f3$WT)
          expect_equal(f3$mCRCL, f3$CRCL)
          expect_equal(f3$mSEX, f3$SEX)

          r1 <- f1[!duplicated(f1$id), c("id", "SEX", "WT", "CRCL")]
          r3 <- f3[!duplicated(f3$id), c("id", "SEX", "WT", "CRCL")]

          expect_false(isTRUE(all.equal(r1, r3)))

          ## Now these should be false
          expect_false(all(r3$WT - r3$CRCL == 30))

          ## Use floor() so boundary subjects (base CRCL=44) are classified
          ## correctly regardless of time-varying shift.
          expect_false(all(r3$SEX[floor(r3$CRCL) <= 44] == 0))
          expect_false(all(r3$SEX[floor(r3$CRCL) > 44] == 1))


          f3 <- rxSolve(m1, e,
                        ## Lotri uses lower-triangular matrix rep. for named matrix
                        omega = lotri(
                          eta.cl ~ .306,
                          eta.q ~ 0.0652,
                          eta.v1 ~ .567,
                          eta.v2 ~ .191
                        ),
                        sigma = lotri(err.sd ~ 0.5), addCov = TRUE,
                        resample = TRUE,
                        resampleID = FALSE,
                        addDosing = TRUE, nStud = nStud
                        )

          expect_equal(f3$mWT, f3$WT)
          expect_equal(f3$mCRCL, f3$CRCL)
          expect_equal(f3$mSEX, f3$SEX)

          r1 <- f1[!duplicated(f1$id), c("id", "SEX", "WT", "CRCL")]
          r3 <- f3[!duplicated(f3$id), c("id", "SEX", "WT", "CRCL")]

          expect_false(isTRUE(all.equal(r1, r3)))

          ## Now these should be false
          expect_false(all(r3$WT - r3$CRCL == 30))

          ## Use floor() so boundary subjects (base CRCL=44) are classified
          ## correctly regardless of time-varying shift.
          expect_false(all(r3$SEX[floor(r3$CRCL) <= 44] == 0))
          expect_false(all(r3$SEX[floor(r3$CRCL) > 44] == 1))

          f2 <- rxSolve(m1, e,
                        ## Lotri uses lower-triangular matrix rep. for named matrix
                        omega = lotri(
                          eta.cl ~ .306,
                          eta.q ~ 0.0652,
                          eta.v1 ~ .567,
                          eta.v2 ~ .191
                        ),
                        sigma = lotri(err.sd ~ 0.5), addCov = TRUE,
                        resample = TRUE,
                        resampleID = TRUE,
                        addDosing = TRUE, nStud = nStud
                        )

          expect_equal(f2$mWT, f2$WT)
          expect_equal(f2$mCRCL, f2$CRCL)
          expect_equal(f2$mSEX, f2$SEX)

          r1 <- f1[!duplicated(f1$id), c("id", "SEX", "WT", "CRCL")]
          r2 <- f2[!duplicated(f2$id), c("id", "SEX", "WT", "CRCL")]

          expect_false(isTRUE(all.equal(r1, r2)))

          ## now test that the covariates are all shifted correctly
          ## Use all.equal tolerance: time-varying covariates go through
          ## rx_approxP interpolation which accumulates FP rounding errors,
          ## so exact == 30 is not guaranteed even when WT and CRCL come from
          ## the same subject.
          expect_true(isTRUE(all.equal(r1$WT - r1$CRCL, rep(30, nrow(r1)))))
          expect_true(isTRUE(all.equal(r2$WT - r2$CRCL, rep(30, nrow(r2)))))

          ## CRCL = base_CRCL + time/30; base values are integers and
          ## max time-shift is 26.4/30 < 1, so floor() recovers the base
          ## integer for correct SEX-CRCL co-resample categorisation.
          expect_true(all(r1$SEX[floor(r1$CRCL) <= 44] == 0))
          expect_true(all(r1$SEX[floor(r1$CRCL) > 44] == 1))

          expect_true(all(r2$SEX[floor(r2$CRCL) <= 44] == 0))
          expect_true(all(r2$SEX[floor(r2$CRCL) > 44] == 1))

          if (nStud == 1) {
            f1 <- rxSolve(m1, e,
                          ## Lotri uses lower-triangular matrix rep. for named matrix
                          omega = lotri(
                            eta.cl ~ .306,
                            eta.q ~ 0.0652,
                            eta.v1 ~ .567,
                            eta.v2 ~ .191
                          ),
                          sigma = lotri(err.sd ~ 0.5), addCov = TRUE,
                          resample = FALSE,
                          addDosing = TRUE, nStud = nStud
                          )

            expect_equal(f1$WT, e$WT)
            expect_equal(f1$CRCL, e$CRCL)
            expect_equal(f1$SEX, e$SEX)
          }
        }
      })


      # resample tests; time invariant

      nsub <- 30
      # Simulate Weight based on age and gender
      AGE <- round(runif(nsub, min = 18, max = 70))
      SEX <- round(runif(nsub, min = 0, max = 1))
      HTm <- round(rnorm(nsub, 176.3, 0.17 * sqrt(4482)), digits = 1)
      HTf <- round(rnorm(nsub, 162.2, 0.16 * sqrt(4857)), digits = 1)
      WTm <- round(exp(3.28 + 1.92 * log(HTm / 100)) * exp(rnorm(nsub, 0, 0.14)), digits = 1)
      WTf <- round(exp(3.49 + 1.45 * log(HTf / 100)) * exp(rnorm(nsub, 0, 0.17)), digits = 1)
      WT <- ifelse(SEX == 1, WTf, WTm)
      CRCL <- round(runif(nsub, 30, 140))
      ## id is in lower case to match the event table
      cov.df <- dplyr::tibble(id = seq_along(AGE), AGE = AGE, SEX = SEX, WT = WT, CRCL = CRCL)

      s <- c(0, 0.25, 0.5, 0.75, 1, 1.5, seq(2, 24, by = 1))
      ## Add 10% diff
      s <- lapply(s, function(x) {
        d <- x * 0.1
        c(x - d, x + d)
      })

      e <- et(time.units = "hr") |>
        ## Specify the id and weight based dosing from covariate data.frame
        ## This requires rxode2 XXX
        et(id = cov.df$id, amt = 6 * cov.df$WT, rate = 6 * cov.df$WT) |>
        ## Sampling is added for each ID
        et(s) |>
        as.data.frame() |>
        ## Merge the event table with the covarite information
        merge(cov.df, by = "id") |>
        dplyr::as_tibble()

      e2 <- et(time.units = "hr") |>
        ## Specify the id and weight based dosing from covariate data.frame
        ## This requires rxode2 XXX
        et(id = cov.df$id, amt = 6 * cov.df$WT, rate = 6 * cov.df$WT) |>
        ## Sampling is added for each ID
        et(s)


      test_that("resample tests: time invariant", {
        for (resampleID in c(TRUE, FALSE)) {
          f1 <- rxSolve(m1, e,
                        ## Lotri uses lower-triangular matrix rep. for named matrix
                        omega = lotri(
                          eta.cl ~ .306,
                          eta.q ~ 0.0652,
                          eta.v1 ~ .567,
                          eta.v2 ~ .191
                        ),
                        sigma = lotri(err.sd ~ 0.5), addCov = TRUE
                        )

          expect_equal(f1$mWT, f1$WT)
          expect_equal(f1$mCRCL, f1$CRCL)

          f2 <- rxSolve(m1, e,
                        ## Lotri uses lower-triangular matrix rep. for named matrix
                        omega = lotri(
                          eta.cl ~ .306,
                          eta.q ~ 0.0652,
                          eta.v1 ~ .567,
                          eta.v2 ~ .191
                        ),
                        sigma = lotri(err.sd ~ 0.5), addCov = TRUE,
                        resample = c("SEX", "WT", "CRCL"),
                        resampleID = resampleID
                        )

          expect_equal(f2$mWT, f2$WT)
          expect_equal(f2$mCRCL, f2$CRCL)

          f3 <- rxSolve(m1, e,
                        ## Lotri uses lower-triangular matrix rep. for named matrix
                        omega = lotri(
                          eta.cl ~ .306,
                          eta.q ~ 0.0652,
                          eta.v1 ~ .567,
                          eta.v2 ~ .191
                        ),
                        sigma = lotri(err.sd ~ 0.5), keep = c("SEX", "WT", "CRCL"),
                        resample = c("SEX", "WT", "CRCL"),
                        resampleID = resampleID
                        )

          expect_equal(f3$mWT, f3$WT)
          expect_equal(f3$mCRCL, f3$CRCL)

          r1 <- f1[!duplicated(f1$id), c("id", "SEX", "WT", "CRCL")]
          r2 <- f2[!duplicated(f2$id), c("id", "SEX", "WT", "CRCL")]

          expect_false(isTRUE(all.equal(r1, r2)))

          r3 <- f3[!duplicated(f3$id), c("id", "SEX", "WT", "CRCL")]

          expect_false(isTRUE(all.equal(r1, r3)))

          ## Now try icov option

          f1 <-
              rxSolve(m1, e2,
                      iCov = cov.df,
                      ## Lotri uses lower-triangular matrix rep. for named matrix
                      omega = lotri(
                        eta.cl ~ .306,
                        eta.q ~ 0.0652,
                        eta.v1 ~ .567,
                        eta.v2 ~ .191
                      ),
                      sigma = lotri(err.sd ~ 0.5), addCov = TRUE
                      )

          expect_equal(f1$mWT, f1$WT)
          expect_equal(f1$mCRCL, f1$CRCL)

          f2 <- rxSolve(m1, e2,
                    iCov = cov.df,
                    ## Lotri uses lower-triangular matrix rep. for named matrix
                    omega = lotri(
                      eta.cl ~ .306,
                      eta.q ~ 0.0652,
                      eta.v1 ~ .567,
                      eta.v2 ~ .191
                    ),
                    sigma = lotri(err.sd ~ 0.5), addCov = TRUE,
                    resample = c("SEX", "WT", "CRCL"),
                    resampleID = resampleID)

          expect_equal(f2$mWT, f2$WT)
          expect_equal(f2$mCRCL, f2$CRCL)

          f3 <-
              rxSolve(m1, e2,
                      iCov = cov.df,
                      ## Lotri uses lower-triangular matrix rep. for named matrix
                      omega = lotri(
                        eta.cl ~ .306,
                        eta.q ~ 0.0652,
                        eta.v1 ~ .567,
                        eta.v2 ~ .191
                      ),
                      sigma = lotri(err.sd ~ 0.5),
                      keep = c("SEX", "WT", "CRCL"),
                      resample = c("SEX", "WT", "CRCL"),
                      resampleID = resampleID
                      )

          expect_equal(f3$mWT, f3$WT)
          expect_equal(f3$mCRCL, f3$CRCL)
        }
      })

      # resample tests; time varying

      # Make these time-varying covariates

      e$WT <- e$WT + rnorm(length(e$WT), sd = 1)
      e$CRCL <- e$CRCL + rnorm(length(e$CRCL), sd = 1)

      test_that("resample tests: time varying", {
        for (resampleID in c(TRUE, FALSE)) {
          f1 <- rxSolve(m1, e,
                        ## Lotri uses lower-triangular matrix rep. for named matrix
                        omega = lotri(
                          eta.cl ~ .306,
                          eta.q ~ 0.0652,
                          eta.v1 ~ .567,
                          eta.v2 ~ .191
                        ),
                        sigma = lotri(err.sd ~ 0.5), addCov = TRUE
                        )

          expect_equal(f1$mWT, f1$WT)
          expect_equal(f1$mCRCL, f1$CRCL)

          f2 <- rxSolve(m1, e,
                        ## Lotri uses lower-triangular matrix rep. for named matrix
                        omega = lotri(
                          eta.cl ~ .306,
                          eta.q ~ 0.0652,
                          eta.v1 ~ .567,
                          eta.v2 ~ .191
                        ),
                        sigma = lotri(err.sd ~ 0.5), addCov = TRUE,
                        resample = c("SEX", "WT", "CRCL"),
                        resampleID = resampleID
                        )

          expect_equal(f2$mWT, f2$WT)
          expect_equal(f2$mCRCL, f2$CRCL)

          f3 <- rxSolve(m1, e,
                        omega = lotri(
                          eta.cl ~ .306,
                          eta.q ~ 0.0652,
                          eta.v1 ~ .567,
                          eta.v2 ~ .191
                        ),
                        sigma = lotri(err.sd ~ 0.5), keep = c("SEX", "WT", "CRCL"),
                        resample = c("SEX", "WT", "CRCL"),
                        resampleID = resampleID
                        )

          expect_equal(f3$mWT, f3$WT)
          expect_equal(f3$mCRCL, f3$CRCL)

          r1 <- f1[!duplicated(f1$id), c("id", "SEX", "WT", "CRCL")]
          r2 <- f2[!duplicated(f2$id), c("id", "SEX", "WT", "CRCL")]

          expect_false(isTRUE(all.equal(r1$WT, r2$WT)))

          ## Now test keep case

          r1 <- f1[!duplicated(f1$id), c("id", "SEX", "WT", "CRCL")]
          r3 <- f3[!duplicated(f3$id), c("id", "SEX", "WT", "CRCL")]

          expect_false(isTRUE(all.equal(r1$WT, r3$WT)))
        }
      })
    }
  )
})

if (FALSE) {
  # Setup the test data
  lognCv <- function(x){log((x/100)^2+1)}

  rxWithSeed(32, {
    nSub <- 3
    nStud <- 2

    theta <- c(lka=log(0.5), # log ka
               lCl=log(5), # log Cl
               lV=log(300) # log V
    )

    thetaMat <- lotri(lCl ~ lognCv(5),
                      lV  ~ lognCv(5),
                      lka ~ lognCv(5))

    nev <- nSub*nStud

    ev1 <- data.frame(COV1=rnorm(nev,50,30),COV2=rnorm(nev,75,10),
                      COV3=sample(c(1.0,2.0),nev,replace=TRUE))

    tmat <-rxRmvn(nStud, theta[dimnames(thetaMat)[[1]]], thetaMat)

    x1 <- rxCbindStudyIndividual(tmat, ev1)

    nev <- nStud

    ev2 <- data.frame(COV1=rnorm(nev,50,30),COV2=rnorm(nev,75,10),
                      COV3=sample(c(1.0,2.0),nev,replace=TRUE))


    x2 <- rxCbindStudyIndividual(tmat, ev2)
  })
  saveRDS(list(tmat=tmat, ev1=ev1, ev2=ev2, x1=x1, x2=x2), "data-cbind-study-individual.rds")
}

rxTest({
  test_that("cbind study and individual", {
    fileCbindStudyIndividual <- test_path("data-cbind-study-individual.rds")
    skip_if_not(file.exists(fileCbindStudyIndividual))

    l <- readRDS(fileCbindStudyIndividual)

    tmat <- l$tmat
    ev1 <- l$ev1
    ev2 <- l$ev2

    expect_equal(rxCbindStudyIndividual(tmat, ev1), l$x1)

    expect_equal(rxCbindStudyIndividual(as.data.frame(tmat), ev1), l$x1)

    expect_equal(rxCbindStudyIndividual(tmat, ev2), l$x2)

    expect_equal(rxCbindStudyIndividual(as.data.frame(tmat), ev2), l$x2)

    expect_error(rxCbindStudyIndividual(c(a = 1), ev1))

    expect_error(rxCbindStudyIndividual(tmat, c(a = 1)))

    tmat2 <- tmat

    dimnames(tmat2) <- list(c("c", "cq"), NULL)

    expect_error(rxCbindStudyIndividual(tmat2, ev1))
  })
})

rxTest({
  test_that("omega RSE function", {

    tmp <- cvPost(10.2, omega=lotri::lotri(a~0.1), 5)
    lst <- omegaListRse(tmp)

    expect_equal(mean(unlist(tmp)), as.vector(lst[["mean"]]))
    expect_equal(var(unlist(tmp)), as.vector(lst[["var"]]))
    expect_equal(sd(unlist(tmp)), as.vector(lst[["sd"]]))
    expect_equal(sd(unlist(tmp))/mean(unlist(tmp)), as.vector(lst[["rse"]]))

  })
})

rxTest({
  # Regression test for a heap-buffer-overflow in rxSolve_normalizeParms()
  # (src/rxData.cpp).  When subjects share a single event table, that table is
  # stored once (_globals.gall_times_n events) but rx->nall is the full per-sim
  # total (nsub * gall_times_n).  For an nsim > 1 solve that needs sorting (e.g.
  # a model with modeled rate()/dur() that expands events), the per-replicate
  # copy did `std::copy(gamt, gamt + rx->nall, ...)`, reading rx->nall doubles
  # from the gall_times_n-sized source -- a heap over-read (caught by ASAN /
  # valgrind).  The fix tiles the base table across each replicate's nall block.
  #
  # The over-read otherwise produced NA-amount errors / wrong replicate results,
  # so this test also pins numerical correctness against trusted single-sim
  # solves.
  test_that("nsim>1 shared-event solve with modeled rate/dur does not over-read the event table", {
    m1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      fdepot <- 1
      durDepot <- 8
      rateDepot <- 1250
      C2 <- centr / V2
      C3 <- peri / V3
      d / dt(depot) <- -KA * depot
      f(depot) <- fdepot
      dur(depot) <- durDepot       # modeled duration -> event expansion -> needSort
      rate(depot) <- rateDepot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- 1
    })

    ev <- et() |>
      et(amt = 10000, ii = 12, until = 24) |>
      et(seq(0, 24, length.out = 100)) |>
      et(id = 1:4)                # shared event table across 4 subjects

    pp <- rxWithSeed(42, data.frame(KA = 0.294 * exp(rnorm(8)),
                                    CL = 18.6 * exp(rnorm(8))))

    # 8 parameter sets over 4 subjects -> nsim = 2 (the over-read path)
    s8 <- suppressWarnings(rxSolve(m1, ev, params = pp, returnType = "data.frame"))

    expect_false(anyNA(s8$C2))
    expect_true(all(is.finite(s8$C2)))
    expect_equal(length(unique(s8$sim.id)), 2L)

    # trusted references: each replicate as its own single-sim (nsim == 1) solve
    refA <- suppressWarnings(rxSolve(m1, ev, params = pp[1:4, ], returnType = "data.frame"))
    refB <- suppressWarnings(rxSolve(m1, ev, params = pp[5:8, ], returnType = "data.frame"))

    expect_equal(sort(s8$C2), sort(c(refA$C2, refB$C2)), tolerance = 1e-8)
  })
})

rxTest({
  test_that("phi/pnorm/qnorm", {
    expect_equal(phi(1:3), pnorm(1:3))
    expect_equal(phi(as.double(1:3)), pnorm(as.double(1:3)))

    o <- rxode2({
      o <- phi(a)
    })

    expect_equal(
      rxSolve(o, data.frame(a = 1:3), et(0))$o,
      pnorm(as.double(1:3))
    )

    o <- rxode2({
      o <- pnorm(a)
    })

    expect_equal(
      rxSolve(o, data.frame(a = 1:3), et(0))$o,
      pnorm(as.double(1:3))
    )

    o <- rxode2({
      o <- pnorm(a, 0.5)
    })

    expect_equal(
      rxSolve(o, data.frame(a = 1:3), et(0))$o,
      pnorm(as.double(1:3), 0.5)
    )

    o <- rxode2({
      o <- pnorm(a, 0.5, 2)
    })

    expect_equal(
      rxSolve(o, data.frame(a = 1:3), et(0))$o,
      pnorm(as.double(1:3), 0.5, 2)
    )

    suppressMessages(expect_error(rxode2({
      o <- pnorm()
    })))

    suppressMessages(expect_error(rxode2({
      o <- pnorm(a, b, c, d)
    })))

    o <- rxode2({
      o <- qnorm(a)
    })

    expect_equal(
      rxSolve(o, data.frame(a = 1:3), et(0))$o,
      suppressWarnings(qnorm(as.double(1:3)))
    )

    o <- rxode2({
      o <- qnorm(a, 0.5)
    })

    expect_equal(
      rxSolve(o, data.frame(a = 1:3), et(0))$o,
      suppressWarnings(qnorm(as.double(1:3), 0.5))
    )

    o <- rxode2({
      o <- qnorm(a, 0.5, 2)
    })

    expect_equal(
      rxSolve(o, data.frame(a = 1:3), et(0))$o,
      suppressWarnings(qnorm(as.double(1:3), 0.5, 2))
    )

    suppressMessages(expect_error(rxode2({
      o <- qnorm()
    })))

    suppressMessages(expect_error(rxode2({
      o <- qnorm(a, b, c, d)
    })))

    m <- rxode2({
      o <- pnorm(a)
    })

    skip_if_not_installed("units")
    expect_error(rxS(m), NA)

    m <- rxode2({
      o <- pnorm(a, b)
    })

    expect_error(rxS(m), NA)

    m <- rxode2({
      o <- pnorm(a, b, c)
    })

    expect_error(rxS(m), NA)
  })
})

rxTest({
  test_that("nodup-keep", {

    #rxode2 issue #169
    mod3 <- rxode2({
      KA=2.94E-01
      ## Clearance with individuals
      CL=1.86E+01 * (WT / 70) ^ 0.75
      V2=4.02E+01
      Q=1.05E+01
      V3=2.97E+02
      Kin=1
      Kout=1
      EC50=200
      ## The linCmt() picks up the variables from above
      C2   = linCmt()
      Tz= 8
      amp=0.1
      eff(0) = 1  ## This specifies that the effect compartment starts at 1.
      d/dt(eff) =  Kin - Kout*(1-C2/(EC50+C2))*eff
    })

    ev <- readRDS(test_path("nodup-keep-ev.rds"))

    rxWithSeed(10, {

      r1 <- solve(mod3, ev,
                  # Create individual covariate data-frame
                  keep="WT", returnType="data.frame")

      expect_length(which(names(r1)=="WT"), 1)
    })


  })

})

rxTest({
  ## Tests adapted from
  ## https://github.com/cran/mvtnorm/blob/master/tests/rmvnorm.R
  test_that("mvtnormal simulations make sense", {
    rxWithSeed(1, {
      m <- 1:3

      s <- diag(1:3)
      s[2, 1] <- 1
      s[3, 1] <- 2
      s[3, 2] <- 3
      s <- s + t(s)

      set.seed(1)

      x <- rxRmvn(20000, m, s, ncores = 1)

      expect_equal(m, colMeans(x), tolerance = 0.01)
      expect_equal(s, var(x), tolerance = 0.1)

      set.seed(1)
      x <- rxRmvn(20000, m, s, ncores = 2)

      # Using tolerance=0.03 for the Mac random number generator
      expect_equal(m, colMeans(x), tolerance = 0.03)
      expect_equal(s, var(x), tolerance = 0.1)

    })
  })

  test_that("seeds behave as expected", {
    rxWithPreserveSeed({
      m <- 1:3

      s <- diag(1:3)
      s[2, 1] <- 1
      s[3, 1] <- 2
      s[3, 2] <- 3
      s <- s + t(s)

      set.seed(1)
      x1 <- rxRmvn(10, m, s, ncores = 1)

      set.seed(2)
      x2 <- rxRmvn(10, m, s, ncores = 1)
      expect_false(isTRUE(all.equal(x1, x2)))

      set.seed(1)
      x2 <- rxRmvn(10, m, s, ncores = 1)
      expect_equal(x1, x2)

      set.seed(1)
      x1 <- rxRmvn(10, m, s, ncores = 1)
      set.seed(2)
      x2 <- rxRmvn(10, m, s, ncores = 2)
      expect_false(isTRUE(all.equal(x1, x2)))

      set.seed(10)
      r1 <- rxRmvn(10, c(1, 1, 1), diag(3), c(-1, -1, -1), c(3, 3, 3), ncores = 1)

      set.seed(10)
      r2 <- rxRmvn(10, c(1, 1, 1), diag(3), c(-1, -1, -1), c(3, 3, 3), ncores = 1)

      expect_equal(r1, r2)


      set.seed(10)
      r1 <- rxRmvn(10, c(1, 1, 1), diag(3), c(-1, -1, -1), c(3, 3, 3), ncores = 1)

      set.seed(10)
      r2 <- rxRmvn(10, c(1, 1, 1), diag(3), c(-1, -1, -1), c(3, 3, 3), ncores = 2)
      expect_false(isTRUE(all.equal(x1, x2)))


      set.seed(10)
      r1 <- rxRmvn(10, c(1, 1, 1), diag(3), c(-1, -1, -1), c(3, 3, 3), ncores = 2)
      set.seed(10)
      r2 <- rxRmvn(10, c(1, 1, 1), diag(3), c(-1, -1, -1), c(3, 3, 3), ncores = 2)
      if (isTRUE(all.equal(r1, r2))) {
        expect_equal(r1, r2)
      } else {
        warning("r1 != r2 because of load/imbalance etc.")
      }

      rxRmvn(10, 1, diag(1) * 0.01, -1, 1.1)

      rxRmvn(10, 1, diag(1) * 0.01)
    })

  })

  ## test simulating with zero variance

  test_that("zero variance simulations", {
    m <- 1:3

    s <- diag(rep(0, 3))

    rxWithSeed(1, {
      x <- rxRmvn(20, m, s, ncores = 1)
      expect_true(all(x[, 1] == 1))
      expect_true(all(x[, 2] == 2))
      expect_true(all(x[, 3] == 3))

      x <- rxRmvn(20, sigma = s, ncores = 1)

      expect_true(all(x == 0))
    })
  })

  test_that("truncated normal simulation", {
    rxWithSeed(414, {
      d <- 5
      mu <- 1:d

      # Creating covariance matrix
      tmp <- matrix(rnorm(d^2), d, d)
      mcov <- tcrossprod(tmp, tmp)

      a <- rxRmvn(10, 1:d, mcov, lower = 1:d - 1, upper = 1:d + 1)

      for (i in 1:d) {
        expect_false(all(a[, i] == i))
        expect_false(any(a[, i] < i - 1))
        expect_false(any(a[, i] > i + 1))
      }

      set.seed(10)
      a1 <- rxRmvn(1, 10, matrix(2), 10, 11)
      a2 <- rxRmvn(1, 10, matrix(2), 10, 11)

      expect_false(all(a1 == a2))

      set.seed(10)
      a1 <- rxRmvn(1, 10, matrix(2), 10, 11)

      set.seed(10)
      a2 <- rxRmvn(1, 10, matrix(2), 10, 11)

      expect_equal(a1, a2)
    })

  })
})

*DECK DUMACH
      DOUBLE PRECISION FUNCTION DUMACH ()
C***BEGIN PROLOGUE  DUMACH
C***PURPOSE  Compute the unit roundoff of the machine.
C***CATEGORY  R1
C***TYPE      DOUBLE PRECISION (RUMACH-S, DUMACH-D)
C***KEYWORDS  MACHINE CONSTANTS
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***DESCRIPTION
C *Usage:
C        DOUBLE PRECISION  A, DUMACH
C        A = DUMACH()
C
C *Function Return Values:
C     A : the unit roundoff of the machine.
C
C *Description:
C     The unit roundoff is defined as the smallest positive machine
C     number u such that  1.0 + u .ne. 1.0.  This is computed by DUMACH
C     in a machine-independent manner.
C
C***REFERENCES  (NONE)
C***ROUTINES CALLED  DUMSUM
C***REVISION HISTORY  (YYYYMMDD)
C   19930216  DATE WRITTEN
C   19930818  Added SLATEC-format prologue.  (FNF)
C   20030707  Added DUMSUM to force normal storage of COMP.  (ACH)
C***END PROLOGUE  DUMACH
C
      DOUBLE PRECISION U, COMP
C***FIRST EXECUTABLE STATEMENT  DUMACH
      U = 1.0D0
 10   U = U*0.5D0
      CALL DUMSUM(1.0D0, U, COMP)
      IF (COMP .NE. 1.0D0) GO TO 10
      DUMACH = U*2.0D0
      RETURN
C----------------------- End of Function DUMACH ------------------------
      END
      SUBROUTINE DUMSUM(A,B,C)
C     Routine to force normal storing of A + B, for DUMACH.
      DOUBLE PRECISION A, B, C
      C = A + B
      RETURN
      END
*DECK DCFODE
      SUBROUTINE DCFODE (METH, ELCO, TESCO)
C***BEGIN PROLOGUE  DCFODE
C***SUBSIDIARY
C***PURPOSE  Set ODE integrator coefficients.
C***TYPE      DOUBLE PRECISION (SCFODE-S, DCFODE-D)
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***DESCRIPTION
C
C  DCFODE is called by the integrator routine to set coefficients
C  needed there.  The coefficients for the current method, as
C  given by the value of METH, are set for all orders and saved.
C  The maximum order assumed here is 12 if METH = 1 and 5 if METH = 2.
C  (A smaller value of the maximum order is also allowed.)
C  DCFODE is called once at the beginning of the problem,
C  and is not called again unless and until METH is changed.
C
C  The ELCO array contains the basic method coefficients.
C  The coefficients el(i), 1 .le. i .le. nq+1, for the method of
C  order nq are stored in ELCO(i,nq).  They are given by a genetrating
C  polynomial, i.e.,
C      l(x) = el(1) + el(2)*x + ... + el(nq+1)*x**nq.
C  For the implicit Adams methods, l(x) is given by
C      dl/dx = (x+1)*(x+2)*...*(x+nq-1)/factorial(nq-1),    l(-1) = 0.
C  For the BDF methods, l(x) is given by
C      l(x) = (x+1)*(x+2)* ... *(x+nq)/K,
C  where         K = factorial(nq)*(1 + 1/2 + ... + 1/nq).
C
C  The TESCO array contains test constants used for the
C  local error test and the selection of step size and/or order.
C  At order nq, TESCO(k,nq) is used for the selection of step
C  size at order nq - 1 if k = 1, at order nq if k = 2, and at order
C  nq + 1 if k = 3.
C
C***SEE ALSO  DLSODE
C***ROUTINES CALLED  (NONE)
C***REVISION HISTORY  (YYMMDD)
C   791129  DATE WRITTEN
C   890501  Modified prologue to SLATEC/LDOC format.  (FNF)
C   890503  Minor cosmetic changes.  (FNF)
C   930809  Renamed to allow single/double precision versions. (ACH)
C***END PROLOGUE  DCFODE
C**End
      INTEGER METH
      INTEGER I, IB, NQ, NQM1, NQP1
      DOUBLE PRECISION ELCO, TESCO
      DOUBLE PRECISION AGAMQ, FNQ, FNQM1, PC, PINT, RAGQ,
     1   RQFAC, RQ1FAC, TSIGN, XPIN
      DIMENSION ELCO(13,12), TESCO(3,12)
      DIMENSION PC(12)
C
C***FIRST EXECUTABLE STATEMENT  DCFODE
C     GO TO (100, 200), METH
      IF (METH .EQ. 1) GO TO 100
      IF (METH .EQ. 2) GO TO 200
C
 100  ELCO(1,1) = 1.0D0
      ELCO(2,1) = 1.0D0
      TESCO(1,1) = 0.0D0
      TESCO(2,1) = 2.0D0
      TESCO(1,2) = 1.0D0
      TESCO(3,12) = 0.0D0
      PC(1) = 1.0D0
      RQFAC = 1.0D0
      DO 140 NQ = 2,12
C-----------------------------------------------------------------------
C The PC array will contain the coefficients of the polynomial
C     p(x) = (x+1)*(x+2)*...*(x+nq-1).
C Initially, p(x) = 1.
C-----------------------------------------------------------------------
        RQ1FAC = RQFAC
        RQFAC = RQFAC/NQ
        NQM1 = NQ - 1
        FNQM1 = NQM1
        NQP1 = NQ + 1
C Form coefficients of p(x)*(x+nq-1). ----------------------------------
        PC(NQ) = 0.0D0
        DO 110 IB = 1,NQM1
          I = NQP1 - IB
C110      PC(I) = PC(I-1) + FNQM1*PC(I)
          PC(I) = PC(I-1) + FNQM1*PC(I)
 110      CONTINUE
        PC(1) = FNQM1*PC(1)
C Compute integral, -1 to 0, of p(x) and x*p(x). -----------------------
        PINT = PC(1)
        XPIN = PC(1)/2.0D0
        TSIGN = 1.0D0
        DO 120 I = 2,NQ
          TSIGN = -TSIGN
          PINT = PINT + TSIGN*PC(I)/I
C120      XPIN = XPIN + TSIGN*PC(I)/(I+1)
          XPIN = XPIN + TSIGN*PC(I)/(I+1)
 120      CONTINUE
C Store coefficients in ELCO and TESCO. --------------------------------
        ELCO(1,NQ) = PINT*RQ1FAC
        ELCO(2,NQ) = 1.0D0
        DO 130 I = 2,NQ
C130      ELCO(I+1,NQ) = RQ1FAC*PC(I)/I
          ELCO(I+1,NQ) = RQ1FAC*PC(I)/I
 130      CONTINUE
        AGAMQ = RQFAC*XPIN
        RAGQ = 1.0D0/AGAMQ
        TESCO(2,NQ) = RAGQ
        IF (NQ .LT. 12) TESCO(1,NQP1) = RAGQ*RQFAC/NQP1
        TESCO(3,NQM1) = RAGQ
 140    CONTINUE
      RETURN
C
 200  PC(1) = 1.0D0
      RQ1FAC = 1.0D0
      DO 230 NQ = 1,5
C-----------------------------------------------------------------------
C The PC array will contain the coefficients of the polynomial
C     p(x) = (x+1)*(x+2)*...*(x+nq).
C Initially, p(x) = 1.
C-----------------------------------------------------------------------
        FNQ = NQ
        NQP1 = NQ + 1
C Form coefficients of p(x)*(x+nq). ------------------------------------
        PC(NQP1) = 0.0D0
        DO 210 IB = 1,NQ
          I = NQ + 2 - IB
C210      PC(I) = PC(I-1) + FNQ*PC(I)
          PC(I) = PC(I-1) + FNQ*PC(I)
 210      CONTINUE
        PC(1) = FNQ*PC(1)
C Store coefficients in ELCO and TESCO. --------------------------------
        DO 220 I = 1,NQP1
C220      ELCO(I,NQ) = PC(I)/PC(2)
          ELCO(I,NQ) = PC(I)/PC(2)
 220      CONTINUE
        ELCO(2,NQ) = 1.0D0
        TESCO(1,NQ) = RQ1FAC
        TESCO(2,NQ) = NQP1/ELCO(1,NQ)
        TESCO(3,NQ) = (NQ+2)/ELCO(1,NQ)
        RQ1FAC = RQ1FAC/FNQ
 230    CONTINUE
      RETURN
C----------------------- END OF SUBROUTINE DCFODE ----------------------
      END
*DECK DINTDY
      SUBROUTINE DINTDY (T, K, YH, NYH, DKY, IFLAG)
C***BEGIN PROLOGUE  DINTDY
C***SUBSIDIARY
C***PURPOSE  Interpolate solution derivatives.
C***TYPE      DOUBLE PRECISION (SINTDY-S, DINTDY-D)
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***DESCRIPTION
C
C  DINTDY computes interpolated values of the K-th derivative of the
C  dependent variable vector y, and stores it in DKY.  This routine
C  is called within the package with K = 0 and T = TOUT, but may
C  also be called by the user for any K up to the current order.
C  (See detailed instructions in the usage documentation.)
C
C  The computed values in DKY are gotten by interpolation using the
C  Nordsieck history array YH.  This array corresponds uniquely to a
C  vector-valued polynomial of degree NQCUR or less, and DKY is set
C  to the K-th derivative of this polynomial at T.
C  The formula for DKY is:
C               q
C   DKY(i)  =  sum  c(j,K) * (T - tn)**(j-K) * h**(-j) * YH(i,j+1)
C              j=K
C  where  c(j,K) = j*(j-1)*...*(j-K+1), q = NQCUR, tn = TCUR, h = HCUR.
C  The quantities  nq = NQCUR, l = nq+1, N = NEQ, tn, and h are
C  communicated by COMMON.  The above sum is done in reverse order.
C  IFLAG is returned negative if either K or T is out of bounds.
C
C***SEE ALSO  DLSODE
C***ROUTINES CALLED  XERRWD
C***COMMON BLOCKS    DLS001
C***REVISION HISTORY  (YYMMDD)
C   791129  DATE WRITTEN
C   890501  Modified prologue to SLATEC/LDOC format.  (FNF)
C   890503  Minor cosmetic changes.  (FNF)
C   930809  Renamed to allow single/double precision versions. (ACH)
C   010418  Reduced size of Common block /DLS001/. (ACH)
C   031105  Restored 'own' variables to Common block /DLS001/, to
C           enable interrupt/restart feature. (ACH)
C   050427  Corrected roundoff decrement in TP. (ACH)
C***END PROLOGUE  DINTDY
C**End
      INTEGER K, NYH, IFLAG
      DOUBLE PRECISION T, YH, DKY
      DIMENSION YH(NYH,*), DKY(*)
      INTEGER INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH01,
     1   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     2   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     3   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     4   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      DOUBLE PRECISION CONIT, CRATE, EL, ELCO, HOLD, RMAX, TESCO,
     1   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND
      COMMON /DLS001/ CONIT, CRATE, EL(13), ELCO(13,12),
     1   HOLD, RMAX, TESCO(3,12),
     2   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND,
     3   INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH01,
     4   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     5   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     6   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     7   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      INTEGER I, IC, J, JB, JB2, JJ, JJ1, JP1
      DOUBLE PRECISION C, R, S, TP
      CHARACTER (LEN=80) MSG
C
C***FIRST EXECUTABLE STATEMENT  DINTDY
      IFLAG = 0
      IF (K .LT. 0 .OR. K .GT. NQ) GO TO 80
      TP = TN - HU -  100.0D0*UROUND*SIGN(ABS(TN) + ABS(HU), HU)
      IF ((T-TP)*(T-TN) .GT. 0.0D0) GO TO 90
C
      S = (T - TN)/H
      IC = 1
      IF (K .EQ. 0) GO TO 15
      JJ1 = L - K
      DO 10 JJ = JJ1,NQ
C10     IC = IC*JJ
        IC = IC*JJ
 10     CONTINUE
 15   C = IC
      DO 20 I = 1,N
C20     DKY(I) = C*YH(I,L)
        DKY(I) = C*YH(I,L)
 20     CONTINUE
      IF (K .EQ. NQ) GO TO 55
      JB2 = NQ - K
      DO 50 JB = 1,JB2
        J = NQ - JB
        JP1 = J + 1
        IC = 1
        IF (K .EQ. 0) GO TO 35
        JJ1 = JP1 - K
        DO 30 JJ = JJ1,J
C30       IC = IC*JJ
          IC = IC*JJ
 30       CONTINUE
 35     C = IC
        DO 40 I = 1,N
C40       DKY(I) = C*YH(I,JP1) + S*DKY(I)
          DKY(I) = C*YH(I,JP1) + S*DKY(I)
 40       CONTINUE
 50     CONTINUE
      IF (K .EQ. 0) RETURN
 55   R = H**(-K)
      DO 60 I = 1,N
C60     DKY(I) = R*DKY(I)
        DKY(I) = R*DKY(I)
 60     CONTINUE
      RETURN
C
 80   MSG = 'DINTDY-  K (=I1) illegal      '
c     CALL XERRWD (MSG, 30, 51, 0, 1, K, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (51, 0)
      IFLAG = -1
      RETURN
 90   MSG = 'DINTDY-  T (=R1) illegal      '
c     CALL XERRWD (MSG, 30, 52, 0, 0, 0, 0, 1, T, 0.0D0)
c     MSG='      T not in interval TCUR - HU (= R1) to TCUR (=R2)      '
c     CALL XERRWD (MSG, 60, 52, 0, 0, 0, 0, 2, TP, TN)
      CALL XERRWD (52, 0)
      IFLAG = -2
      RETURN
C----------------------- END OF SUBROUTINE DINTDY ----------------------
      END
*DECK DSOLSY
      SUBROUTINE DSOLSY (WM, IWM, X)
C     SUBROUTINE DSOLSY (WM, IWM, X, TEM)
C***BEGIN PROLOGUE  DSOLSY
C***SUBSIDIARY
C***PURPOSE  ODEPACK linear system solver.
C***TYPE      DOUBLE PRECISION (SSOLSY-S, DSOLSY-D)
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***DESCRIPTION
C
C  This routine manages the solution of the linear system arising from
C  a chord iteration.  It is called if MITER .ne. 0.
C  If MITER is 1 or 2, it calls DGESL to accomplish this.
C  If MITER = 3 it updates the coefficient h*EL0 in the diagonal
C  matrix, and then computes the solution.
C  If MITER is 4 or 5, it calls DGBSL.
C  Communication with DSOLSY uses the following variables:
C  WM    = real work space containing the inverse diagonal matrix if
C          MITER = 3 and the LU decomposition of the matrix otherwise.
C          Storage of matrix elements starts at WM(3).
C          WM also contains the following matrix-related data:
C          WM(1) = SQRT(UROUND) (not used here),
C          WM(2) = HL0, the previous value of h*EL0, used if MITER = 3.
C  IWM   = integer work space containing pivot information, starting at
C          IWM(21), if MITER is 1, 2, 4, or 5.  IWM also contains band
C          parameters ML = IWM(1) and MU = IWM(2) if MITER is 4 or 5.
C  X     = the right-hand side vector on input, and the solution vector
C          on output, of length N.
C  TEM   = vector of work space of length N, not used in this version.
C  IERSL = output flag (in COMMON).  IERSL = 0 if no trouble occurred.
C          IERSL = 1 if a singular matrix arose with MITER = 3.
C  This routine also uses the COMMON variables EL0, H, MITER, and N.
C
C***SEE ALSO  DLSODE
C***ROUTINES CALLED  DGBSL, DGESL
C***COMMON BLOCKS    DLS001
C***REVISION HISTORY  (YYMMDD)
C   791129  DATE WRITTEN
C   890501  Modified prologue to SLATEC/LDOC format.  (FNF)
C   890503  Minor cosmetic changes.  (FNF)
C   930809  Renamed to allow single/double precision versions. (ACH)
C   010418  Reduced size of Common block /DLS001/. (ACH)
C   031105  Restored 'own' variables to Common block /DLS001/, to
C           enable interrupt/restart feature. (ACH)
C***END PROLOGUE  DSOLSY
C**End
      INTEGER IWM
      DOUBLE PRECISION WM, X
      DIMENSION WM(*), IWM(*), X(*)
C     DOUBLE PRECISION WM, X, TEM
C     DIMENSION WM(*), IWM(*), X(*), TEM(*)
      INTEGER INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH,
     1   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     2   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     3   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     4   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      DOUBLE PRECISION CONIT, CRATE, EL, ELCO, HOLD, RMAX, TESCO,
     1   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND
      COMMON /DLS001/ CONIT, CRATE, EL(13), ELCO(13,12),
     1   HOLD, RMAX, TESCO(3,12),
     2   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND,
     3   INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH,
     4   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     5   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     6   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     7   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      INTEGER I, MEBAND, ML, MU
      DOUBLE PRECISION DI, HL0, PHL0, R
C
C***FIRST EXECUTABLE STATEMENT  DSOLSY
      IERSL = 0
C     GO TO (100, 100, 300, 400, 400), MITER
      IF (MITER .EQ. 1) GO TO 100
      IF (MITER .EQ. 2) GO TO 100
      IF (MITER .EQ. 3) GO TO 300
      IF (MITER .EQ. 4) GO TO 400
      IF (MITER .EQ. 5) GO TO 400
 100  CALL DGESL (WM(3), N, N, IWM(21), X, 0)
      RETURN
C
 300  PHL0 = WM(2)
      HL0 = H*EL0
      WM(2) = HL0
      IF (HL0 .EQ. PHL0) GO TO 330
      R = HL0/PHL0
      DO 320 I = 1,N
        DI = 1.0D0 - R*(1.0D0 - 1.0D0/WM(I+2))
        IF (ABS(DI) .EQ. 0.0D0) GO TO 390
C320    WM(I+2) = 1.0D0/DI
        WM(I+2) = 1.0D0/DI
 320    CONTINUE
 330  DO 340 I = 1,N
C340    X(I) = WM(I+2)*X(I)
        X(I) = WM(I+2)*X(I)
 340    CONTINUE
      RETURN
 390  IERSL = 1
      RETURN
C
 400  ML = IWM(1)
      MU = IWM(2)
      MEBAND = 2*ML + MU + 1
      CALL DGBSL (WM(3), MEBAND, N, ML, MU, IWM(21), X, 0)
      RETURN
C----------------------- END OF SUBROUTINE DSOLSY ----------------------
      END
*DECK DEWSET
      SUBROUTINE DEWSET (N, ITOL, RTOL, ATOL, YCUR, EWT)
C***BEGIN PROLOGUE  DEWSET
C***SUBSIDIARY
C***PURPOSE  Set error weight vector.
C***TYPE      DOUBLE PRECISION (SEWSET-S, DEWSET-D)
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***DESCRIPTION
C
C  This subroutine sets the error weight vector EWT according to
C      EWT(i) = RTOL(i)*ABS(YCUR(i)) + ATOL(i),  i = 1,...,N,
C  with the subscript on RTOL and/or ATOL possibly replaced by 1 above,
C  depending on the value of ITOL.
C
C***SEE ALSO  DLSODE
C***ROUTINES CALLED  (NONE)
C***REVISION HISTORY  (YYMMDD)
C   791129  DATE WRITTEN
C   890501  Modified prologue to SLATEC/LDOC format.  (FNF)
C   890503  Minor cosmetic changes.  (FNF)
C   930809  Renamed to allow single/double precision versions. (ACH)
C***END PROLOGUE  DEWSET
C**End
      INTEGER N, ITOL
      INTEGER I
      DOUBLE PRECISION RTOL, ATOL, YCUR, EWT
      DIMENSION RTOL(*), ATOL(*), YCUR(N), EWT(N)
C
C***FIRST EXECUTABLE STATEMENT  DEWSET
C     GO TO (10, 20, 30, 40), ITOL
      IF (ITOL .EQ. 1) GO TO 10
      IF (ITOL .EQ. 2) GO TO 20
      IF (ITOL .EQ. 3) GO TO 30
      IF (ITOL .EQ. 4) GO TO 40
 10   CONTINUE
      DO 15 I = 1,N
C15     EWT(I) = RTOL(1)*ABS(YCUR(I)) + ATOL(1)
        EWT(I) = RTOL(1)*ABS(YCUR(I)) + ATOL(1)
 15     CONTINUE
      RETURN
 20   CONTINUE
      DO 25 I = 1,N
C25     EWT(I) = RTOL(1)*ABS(YCUR(I)) + ATOL(I)
        EWT(I) = RTOL(1)*ABS(YCUR(I)) + ATOL(I)
 25     CONTINUE
      RETURN
 30   CONTINUE
      DO 35 I = 1,N
C35     EWT(I) = RTOL(I)*ABS(YCUR(I)) + ATOL(1)
        EWT(I) = RTOL(I)*ABS(YCUR(I)) + ATOL(1)
 35     CONTINUE
      RETURN
 40   CONTINUE
      DO 45 I = 1,N
C45     EWT(I) = RTOL(I)*ABS(YCUR(I)) + ATOL(I)
        EWT(I) = RTOL(I)*ABS(YCUR(I)) + ATOL(I)
 45     CONTINUE
      RETURN
C----------------------- END OF SUBROUTINE DEWSET ----------------------
      END
*DECK DSTODA
      SUBROUTINE DSTODA (NEQ, Y, YH, NYH, YH1, EWT, SAVF, ACOR,
     1   WM, IWM, F, JAC, PJAC, SLVS)
      EXTERNAL F, JAC, PJAC, SLVS
      INTEGER NEQ, NYH, IWM
      DOUBLE PRECISION Y, YH, YH1, EWT, SAVF, ACOR, WM
      DIMENSION NEQ(*), Y(*), YH(NYH,*), YH1(*), EWT(*), SAVF(*),
     1   ACOR(*), WM(*), IWM(*)
      INTEGER INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH01,
     1   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     2   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     3   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     4   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      INTEGER IOWND2, ICOUNT, IRFLAG, JTYP, MUSED, MXORDN, MXORDS
      DOUBLE PRECISION CONIT, CRATE, EL, ELCO, HOLD, RMAX, TESCO,
     2   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND
      DOUBLE PRECISION ROWND2, CM1, CM2, PDEST, PDLAST, RATIO,
     1   PDNORM
      COMMON /DLS001/ CONIT, CRATE, EL(13), ELCO(13,12),
     1   HOLD, RMAX, TESCO(3,12),
     2   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND,
     3   INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH01,
     4   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     5   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     6   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     7   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      COMMON /DLSA01/ ROWND2, CM1(12), CM2(5), PDEST, PDLAST, RATIO,
     1   PDNORM,
     2   IOWND2(3), ICOUNT, IRFLAG, JTYP, MUSED, MXORDN, MXORDS
      INTEGER I, I1, IREDO, IRET, J, JB, M, NCF, NEWQ
      INTEGER LM1, LM1P1, LM2, LM2P1, NQM1, NQM2
      DOUBLE PRECISION DCON, DDN, DEL, DELP, DSM, DUP, EXDN, EXSM, EXUP,
     1   R, RH, RHDN, RHSM, RHUP, TOLD, DMNORM
      DOUBLE PRECISION ALPHA, DM1,DM2, EXM1,EXM2,
     1   PDH, PNORM, RATE, RH1, RH1IT, RH2, RM, SM1(12)
      SAVE SM1
      DATA SM1/0.5D0, 0.575D0, 0.55D0, 0.45D0, 0.35D0, 0.25D0,
     1   0.20D0, 0.15D0, 0.10D0, 0.075D0, 0.050D0, 0.025D0/
C-----------------------------------------------------------------------
C DSTODA performs one step of the integration of an initial value
C problem for a system of ordinary differential equations.
C Note: DSTODA is independent of the value of the iteration method
C indicator MITER, when this is .ne. 0, and hence is independent
C of the type of chord method used, or the Jacobian structure.
C Communication with DSTODA is done with the following variables:
C
C Y      = an array of length .ge. N used as the Y argument in
C          all calls to F and JAC.
C NEQ    = integer array containing problem size in NEQ(1), and
C          passed as the NEQ argument in all calls to F and JAC.
C YH     = an NYH by LMAX array containing the dependent variables
C          and their approximate scaled derivatives, where
C          LMAX = MAXORD + 1.  YH(i,j+1) contains the approximate
C          j-th derivative of y(i), scaled by H**j/factorial(j)
C          (j = 0,1,...,NQ).  On entry for the first step, the first
C          two columns of YH must be set from the initial values.
C NYH    = a constant integer .ge. N, the first dimension of YH.
C YH1    = a one-dimensional array occupying the same space as YH.
C EWT    = an array of length N containing multiplicative weights
C          for local error measurements.  Local errors in y(i) are
C          compared to 1.0/EWT(i) in various error tests.
C SAVF   = an array of working storage, of length N.
C ACOR   = a work array of length N, used for the accumulated
C          corrections.  On a successful return, ACOR(i) contains
C          the estimated one-step local error in y(i).
C WM,IWM = real and integer work arrays associated with matrix
C          operations in chord iteration (MITER .ne. 0).
C PJAC   = name of routine to evaluate and preprocess Jacobian matrix
C          and P = I - H*EL0*Jac, if a chord method is being used.
C          It also returns an estimate of norm(Jac) in PDNORM.
C SLVS   = name of routine to solve linear system in chord iteration.
C CCMAX  = maximum relative change in H*EL0 before PJAC is called.
C H      = the step size to be attempted on the next step.
C          H is altered by the error control algorithm during the
C          problem.  H can be either positive or negative, but its
C          sign must remain constant throughout the problem.
C HMIN   = the minimum absolute value of the step size H to be used.
C HMXI   = inverse of the maximum absolute value of H to be used.
C          HMXI = 0.0 is allowed and corresponds to an infinite HMAX.
C          HMIN and HMXI may be changed at any time, but will not
C          take effect until the next change of H is considered.
C TN     = the independent variable. TN is updated on each step taken.
C JSTART = an integer used for input only, with the following
C          values and meanings:
C               0  perform the first step.
C           .gt.0  take a new step continuing from the last.
C              -1  take the next step with a new value of H,
C                    N, METH, MITER, and/or matrix parameters.
C              -2  take the next step with a new value of H,
C                    but with other inputs unchanged.
C          On return, JSTART is set to 1 to facilitate continuation.
C KFLAG  = a completion code with the following meanings:
C               0  the step was succesful.
C              -1  the requested error could not be achieved.
C              -2  corrector convergence could not be achieved.
C              -3  fatal error in PJAC or SLVS.
C          A return with KFLAG = -1 or -2 means either
C          ABS(H) = HMIN or 10 consecutive failures occurred.
C          On a return with KFLAG negative, the values of TN and
C          the YH array are as of the beginning of the last
C          step, and H is the last step size attempted.
C MAXORD = the maximum order of integration method to be allowed.
C MAXCOR = the maximum number of corrector iterations allowed.
C MSBP   = maximum number of steps between PJAC calls (MITER .gt. 0).
C MXNCF  = maximum number of convergence failures allowed.
C METH   = current method.
C          METH = 1 means Adams method (nonstiff)
C          METH = 2 means BDF method (stiff)
C          METH may be reset by DSTODA.
C MITER  = corrector iteration method.
C          MITER = 0 means functional iteration.
C          MITER = JT .gt. 0 means a chord iteration corresponding
C          to Jacobian type JT.  (The DLSODA/DLSODAR argument JT is
C          communicated here as JTYP, but is not used in DSTODA
C          except to load MITER following a method switch.)
C          MITER may be reset by DSTODA.
C N      = the number of first-order differential equations.
C-----------------------------------------------------------------------
      KFLAG = 0
      TOLD = TN
      NCF = 0
      IERPJ = 0
      IERSL = 0
      JCUR = 0
      ICF = 0
      DELP = 0.0D0
      IF (JSTART .GT. 0) GO TO 200
      IF (JSTART .EQ. -1) GO TO 100
      IF (JSTART .EQ. -2) GO TO 160
C-----------------------------------------------------------------------
C On the first call, the order is set to 1, and other variables are
C initialized.  RMAX is the maximum ratio by which H can be increased
C in a single step.  It is initially 1.E4 to compensate for the small
C initial H, but then is normally equal to 10.  If a failure
C occurs (in corrector convergence or error test), RMAX is set at 2
C for the next increase.
C DCFODE is called to get the needed coefficients for both methods.
C-----------------------------------------------------------------------
      LMAX = MAXORD + 1
      NQ = 1
      L = 2
      IALTH = 2
      RMAX = 10000.0D0
      RC = 0.0D0
      EL0 = 1.0D0
      CRATE = 0.7D0
      HOLD = H
      NSLP = 0
      IPUP = MITER
      IRET = 3
C Initialize switching parameters.  METH = 1 is assumed initially. -----
      ICOUNT = 20
      IRFLAG = 0
      PDEST = 0.0D0
      PDLAST = 0.0D0
      RATIO = 5.0D0
      CALL DCFODE (2, ELCO, TESCO)
      DO 10 I = 1,5
C10     CM2(I) = TESCO(2,I)*ELCO(I+1,I)
        CM2(I) = TESCO(2,I)*ELCO(I+1,I)
 10     CONTINUE
      CALL DCFODE (1, ELCO, TESCO)
      DO 20 I = 1,12
C20     CM1(I) = TESCO(2,I)*ELCO(I+1,I)
        CM1(I) = TESCO(2,I)*ELCO(I+1,I)
 20     CONTINUE
      GO TO 150
C-----------------------------------------------------------------------
C The following block handles preliminaries needed when JSTART = -1.
C IPUP is set to MITER to force a matrix update.
C If an order increase is about to be considered (IALTH = 1),
C IALTH is reset to 2 to postpone consideration one more step.
C If the caller has changed METH, DCFODE is called to reset
C the coefficients of the method.
C If H is to be changed, YH must be rescaled.
C If H or METH is being changed, IALTH is reset to L = NQ + 1
C to prevent further changes in H for that many steps.
C-----------------------------------------------------------------------
 100  IPUP = MITER
      LMAX = MAXORD + 1
      IF (IALTH .EQ. 1) IALTH = 2
      IF (METH .EQ. MUSED) GO TO 160
      CALL DCFODE (METH, ELCO, TESCO)
      IALTH = L
      IRET = 1
C-----------------------------------------------------------------------
C The el vector and related constants are reset
C whenever the order NQ is changed, or at the start of the problem.
C-----------------------------------------------------------------------
 150  DO 155 I = 1,L
C155    EL(I) = ELCO(I,NQ)
        EL(I) = ELCO(I,NQ)
 155    CONTINUE
      NQNYH = NQ*NYH
      RC = RC*EL(1)/EL0
      EL0 = EL(1)
      CONIT = 0.5D0/(NQ+2)
C     GO TO (160, 170, 200), IRET
      IF (IRET .EQ. 1) GO TO 160
      IF (IRET .EQ. 2) GO TO 170
      IF (IRET .EQ. 3) GO TO 200
C-----------------------------------------------------------------------
C If H is being changed, the H ratio RH is checked against
C RMAX, HMIN, and HMXI, and the YH array rescaled.  IALTH is set to
C L = NQ + 1 to prevent a change of H for that many steps, unless
C forced by a convergence or error test failure.
C-----------------------------------------------------------------------
 160  IF (H .EQ. HOLD) GO TO 200
      RH = H/HOLD
      H = HOLD
      IREDO = 3
      GO TO 175
 170  RH = MAX(RH,HMIN/ABS(H))
 175  RH = MIN(RH,RMAX)
      RH = RH/MAX(1.0D0,ABS(H)*HMXI*RH)
C-----------------------------------------------------------------------
C If METH = 1, also restrict the new step size by the stability region.
C If this reduces H, set IRFLAG to 1 so that if there are roundoff
C problems later, we can assume that is the cause of the trouble.
C-----------------------------------------------------------------------
      IF (METH .EQ. 2) GO TO 178
      IRFLAG = 0
      PDH = MAX(ABS(H)*PDLAST,0.000001D0)
      IF (RH*PDH*1.00001D0 .LT. SM1(NQ)) GO TO 178
      RH = SM1(NQ)/PDH
      IRFLAG = 1
 178  CONTINUE
      R = 1.0D0
      DO 181 J = 2,L
        R = R*RH
        DO 180 I = 1,N
C180      YH(I,J) = YH(I,J)*R
          YH(I,J) = YH(I,J)*R
 180      CONTINUE
 181      CONTINUE
      H = H*RH
      RC = RC*RH
      IALTH = L
      IF (IREDO .EQ. 0) GO TO 690
C-----------------------------------------------------------------------
C This section computes the predicted values by effectively
C multiplying the YH array by the Pascal triangle matrix.
C RC is the ratio of new to old values of the coefficient  H*EL(1).
C When RC differs from 1 by more than CCMAX, IPUP is set to MITER
C to force PJAC to be called, if a Jacobian is involved.
C In any case, PJAC is called at least every MSBP steps.
C-----------------------------------------------------------------------
 200  IF (ABS(RC-1.0D0) .GT. CCMAX) IPUP = MITER
      IF (NST .GE. NSLP+MSBP) IPUP = MITER
      TN = TN + H
      I1 = NQNYH + 1
      DO 215 JB = 1,NQ
        I1 = I1 - NYH
        DO 210 I = I1,NQNYH
C210      YH1(I) = YH1(I) + YH1(I+NYH)
          YH1(I) = YH1(I) + YH1(I+NYH)
 210      CONTINUE
 215    CONTINUE
      PNORM = DMNORM (N, YH1, EWT)
C-----------------------------------------------------------------------
C Up to MAXCOR corrector iterations are taken.  A convergence test is
C made on the RMS-norm of each correction, weighted by the error
C weight vector EWT.  The sum of the corrections is accumulated in the
C vector ACOR(i).  The YH array is not altered in the corrector loop.
C-----------------------------------------------------------------------
 220  M = 0
      RATE = 0.0D0
      DEL = 0.0D0
      DO 230 I = 1,N
C230    Y(I) = YH(I,1)
        Y(I) = YH(I,1)
 230    CONTINUE
      CALL F (NEQ, TN, Y, SAVF)
      NFE = NFE + 1
      IF (IPUP .LE. 0) GO TO 250
C-----------------------------------------------------------------------
C If indicated, the matrix P = I - H*EL(1)*J is reevaluated and
C preprocessed before starting the corrector iteration.  IPUP is set
C to 0 as an indicator that this has been done.
C-----------------------------------------------------------------------
      CALL PJAC (NEQ, Y, YH, NYH, EWT, ACOR, SAVF, WM, IWM, F, JAC)
      IPUP = 0
      RC = 1.0D0
      NSLP = NST
      CRATE = 0.7D0
      IF (IERPJ .NE. 0) GO TO 430
 250  DO 260 I = 1,N
C260    ACOR(I) = 0.0D0
        ACOR(I) = 0.0D0
 260    CONTINUE
 270  IF (MITER .NE. 0) GO TO 350
C-----------------------------------------------------------------------
C In the case of functional iteration, update Y directly from
C the result of the last function evaluation.
C-----------------------------------------------------------------------
      DO 290 I = 1,N
        SAVF(I) = H*SAVF(I) - YH(I,2)
C290    Y(I) = SAVF(I) - ACOR(I)
        Y(I) = SAVF(I) - ACOR(I)
 290    CONTINUE
      DEL = DMNORM (N, Y, EWT)
      DO 300 I = 1,N
        Y(I) = YH(I,1) + EL(1)*SAVF(I)
C300    ACOR(I) = SAVF(I)
        ACOR(I) = SAVF(I)
 300    CONTINUE
      GO TO 400
C-----------------------------------------------------------------------
C In the case of the chord method, compute the corrector error,
C and solve the linear system with that as right-hand side and
C P as coefficient matrix.
C-----------------------------------------------------------------------
 350  DO 360 I = 1,N
C360    Y(I) = H*SAVF(I) - (YH(I,2) + ACOR(I))
        Y(I) = H*SAVF(I) - (YH(I,2) + ACOR(I))
 360    CONTINUE
      CALL SLVS (WM, IWM, Y, SAVF)
      IF (IERSL .LT. 0) GO TO 430
      IF (IERSL .GT. 0) GO TO 410
      DEL = DMNORM (N, Y, EWT)
      DO 380 I = 1,N
        ACOR(I) = ACOR(I) + Y(I)
C380    Y(I) = YH(I,1) + EL(1)*ACOR(I)
        Y(I) = YH(I,1) + EL(1)*ACOR(I)
 380    CONTINUE
C-----------------------------------------------------------------------
C Test for convergence.  If M .gt. 0, an estimate of the convergence
C rate constant is stored in CRATE, and this is used in the test.
C
C We first check for a change of iterates that is the size of
C roundoff error.  If this occurs, the iteration has converged, and a
C new rate estimate is not formed.
C In all other cases, force at least two iterations to estimate a
C local Lipschitz constant estimate for Adams methods.
C On convergence, form PDEST = local maximum Lipschitz constant
C estimate.  PDLAST is the most recent nonzero estimate.
C-----------------------------------------------------------------------
 400  CONTINUE
      IF (DEL .LE. 100.0D0*PNORM*UROUND) GO TO 450
      IF (M .EQ. 0 .AND. METH .EQ. 1) GO TO 405
      IF (M .EQ. 0) GO TO 402
      RM = 1024.0D0
      IF (DEL .LE. 1024.0D0*DELP) RM = DEL/DELP
      RATE = MAX(RATE,RM)
      CRATE = MAX(0.2D0*CRATE,RM)
 402  DCON = DEL*MIN(1.0D0,1.5D0*CRATE)/(TESCO(2,NQ)*CONIT)
      IF (DCON .GT. 1.0D0) GO TO 405
      PDEST = MAX(PDEST,RATE/ABS(H*EL(1)))
      IF (PDEST .NE. 0.0D0) PDLAST = PDEST
      GO TO 450
 405  CONTINUE
      M = M + 1
      IF (M .EQ. MAXCOR) GO TO 410
      IF (M .GE. 2 .AND. DEL .GT. 2.0D0*DELP) GO TO 410
      DELP = DEL
      CALL F (NEQ, TN, Y, SAVF)
      NFE = NFE + 1
      GO TO 270
C-----------------------------------------------------------------------
C The corrector iteration failed to converge.
C If MITER .ne. 0 and the Jacobian is out of date, PJAC is called for
C the next try.  Otherwise the YH array is retracted to its values
C before prediction, and H is reduced, if possible.  If H cannot be
C reduced or MXNCF failures have occurred, exit with KFLAG = -2.
C-----------------------------------------------------------------------
 410  IF (MITER .EQ. 0 .OR. JCUR .EQ. 1) GO TO 430
      ICF = 1
      IPUP = MITER
      GO TO 220
 430  ICF = 2
      NCF = NCF + 1
      RMAX = 2.0D0
      TN = TOLD
      I1 = NQNYH + 1
      DO 445 JB = 1,NQ
        I1 = I1 - NYH
        DO 440 I = I1,NQNYH
C440      YH1(I) = YH1(I) - YH1(I+NYH)
          YH1(I) = YH1(I) - YH1(I+NYH)
 440      CONTINUE
 445    CONTINUE
      IF (IERPJ .LT. 0 .OR. IERSL .LT. 0) GO TO 680
      IF (ABS(H) .LE. HMIN*1.00001D0) GO TO 670
      IF (NCF .EQ. MXNCF) GO TO 670
      RH = 0.25D0
      IPUP = MITER
      IREDO = 1
      GO TO 170
C-----------------------------------------------------------------------
C The corrector has converged.  JCUR is set to 0
C to signal that the Jacobian involved may need updating later.
C The local error test is made and control passes to statement 500
C if it fails.
C-----------------------------------------------------------------------
 450  JCUR = 0
      IF (M .EQ. 0) DSM = DEL/TESCO(2,NQ)
      IF (M .GT. 0) DSM = DMNORM (N, ACOR, EWT)/TESCO(2,NQ)
      IF (DSM .GT. 1.0D0) GO TO 500
C-----------------------------------------------------------------------
C After a successful step, update the YH array.
C Decrease ICOUNT by 1, and if it is -1, consider switching methods.
C If a method switch is made, reset various parameters,
C rescale the YH array, and exit.  If there is no switch,
C consider changing H if IALTH = 1.  Otherwise decrease IALTH by 1.
C If IALTH is then 1 and NQ .lt. MAXORD, then ACOR is saved for
C use in a possible order increase on the next step.
C If a change in H is considered, an increase or decrease in order
C by one is considered also.  A change in H is made only if it is by a
C factor of at least 1.1.  If not, IALTH is set to 3 to prevent
C testing for that many steps.
C-----------------------------------------------------------------------
      KFLAG = 0
      IREDO = 0
      NST = NST + 1
      HU = H
      NQU = NQ
      MUSED = METH
      DO 461 J = 1,L
        DO 460 I = 1,N
C460      YH(I,J) = YH(I,J) + EL(J)*ACOR(I)
          YH(I,J) = YH(I,J) + EL(J)*ACOR(I)
 460      CONTINUE
 461      CONTINUE
      ICOUNT = ICOUNT - 1
      IF (ICOUNT .GE. 0) GO TO 488
      IF (METH .EQ. 2) GO TO 480
C-----------------------------------------------------------------------
C We are currently using an Adams method.  Consider switching to BDF.
C If the current order is greater than 5, assume the problem is
C not stiff, and skip this section.
C If the Lipschitz constant and error estimate are not polluted
C by roundoff, go to 470 and perform the usual test.
C Otherwise, switch to the BDF methods if the last step was
C restricted to insure stability (irflag = 1), and stay with Adams
C method if not.  When switching to BDF with polluted error estimates,
C in the absence of other information, double the step size.
C
C When the estimates are OK, we make the usual test by computing
C the step size we could have (ideally) used on this step,
C with the current (Adams) method, and also that for the BDF.
C If NQ .gt. MXORDS, we consider changing to order MXORDS on switching.
C Compare the two step sizes to decide whether to switch.
C The step size advantage must be at least RATIO = 5 to switch.
C-----------------------------------------------------------------------
      IF (NQ .GT. 5) GO TO 488
      IF (DSM .GT. 100.0D0*PNORM*UROUND .AND. PDEST .NE. 0.0D0)
     1   GO TO 470
      IF (IRFLAG .EQ. 0) GO TO 488
      RH2 = 2.0D0
      NQM2 = MIN(NQ,MXORDS)
      GO TO 478
 470  CONTINUE
      EXSM = 1.0D0/L
      RH1 = 1.0D0/(1.2D0*DSM**EXSM + 0.0000012D0)
      RH1IT = 2.0D0*RH1
      PDH = PDLAST*ABS(H)
      IF (PDH*RH1 .GT. 0.00001D0) RH1IT = SM1(NQ)/PDH
      RH1 = MIN(RH1,RH1IT)
      IF (NQ .LE. MXORDS) GO TO 474
         NQM2 = MXORDS
         LM2 = MXORDS + 1
         EXM2 = 1.0D0/LM2
         LM2P1 = LM2 + 1
         DM2 = DMNORM (N, YH(1,LM2P1), EWT)/CM2(MXORDS)
         RH2 = 1.0D0/(1.2D0*DM2**EXM2 + 0.0000012D0)
         GO TO 476
 474  DM2 = DSM*(CM1(NQ)/CM2(NQ))
      RH2 = 1.0D0/(1.2D0*DM2**EXSM + 0.0000012D0)
      NQM2 = NQ
 476  CONTINUE
      IF (RH2 .LT. RATIO*RH1) GO TO 488
C THE SWITCH TEST PASSED.  RESET RELEVANT QUANTITIES FOR BDF. ----------
 478  RH = RH2
      ICOUNT = 20
      METH = 2
      MITER = JTYP
      PDLAST = 0.0D0
      NQ = NQM2
      L = NQ + 1
      GO TO 170
C-----------------------------------------------------------------------
C We are currently using a BDF method.  Consider switching to Adams.
C Compute the step size we could have (ideally) used on this step,
C with the current (BDF) method, and also that for the Adams.
C If NQ .gt. MXORDN, we consider changing to order MXORDN on switching.
C Compare the two step sizes to decide whether to switch.
C The step size advantage must be at least 5/RATIO = 1 to switch.
C If the step size for Adams would be so small as to cause
C roundoff pollution, we stay with BDF.
C-----------------------------------------------------------------------
 480  CONTINUE
      EXSM = 1.0D0/L
      IF (MXORDN .GE. NQ) GO TO 484
         NQM1 = MXORDN
         LM1 = MXORDN + 1
         EXM1 = 1.0D0/LM1
         LM1P1 = LM1 + 1
         DM1 = DMNORM (N, YH(1,LM1P1), EWT)/CM1(MXORDN)
         RH1 = 1.0D0/(1.2D0*DM1**EXM1 + 0.0000012D0)
         GO TO 486
 484  DM1 = DSM*(CM2(NQ)/CM1(NQ))
      RH1 = 1.0D0/(1.2D0*DM1**EXSM + 0.0000012D0)
      NQM1 = NQ
      EXM1 = EXSM
 486  RH1IT = 2.0D0*RH1
      PDH = PDNORM*ABS(H)
      IF (PDH*RH1 .GT. 0.00001D0) RH1IT = SM1(NQM1)/PDH
      RH1 = MIN(RH1,RH1IT)
      RH2 = 1.0D0/(1.2D0*DSM**EXSM + 0.0000012D0)
      IF (RH1*RATIO .LT. 5.0D0*RH2) GO TO 488
      ALPHA = MAX(0.001D0,RH1)
      DM1 = (ALPHA**EXM1)*DM1
      IF (DM1 .LE. 1000.0D0*UROUND*PNORM) GO TO 488
C The switch test passed.  Reset relevant quantities for Adams. --------
      RH = RH1
      ICOUNT = 20
      METH = 1
      MITER = 0
      PDLAST = 0.0D0
      NQ = NQM1
      L = NQ + 1
      GO TO 170
C
C No method switch is being made.  Do the usual step/order selection. --
 488  CONTINUE
      IALTH = IALTH - 1
      IF (IALTH .EQ. 0) GO TO 520
      IF (IALTH .GT. 1) GO TO 700
      IF (L .EQ. LMAX) GO TO 700
      DO 490 I = 1,N
C490    YH(I,LMAX) = ACOR(I)
        YH(I,LMAX) = ACOR(I)
 490    CONTINUE
      GO TO 700
C-----------------------------------------------------------------------
C The error test failed.  KFLAG keeps track of multiple failures.
C Restore TN and the YH array to their previous values, and prepare
C to try the step again.  Compute the optimum step size for this or
C one lower order.  After 2 or more failures, H is forced to decrease
C by a factor of 0.2 or less.
C-----------------------------------------------------------------------
 500  KFLAG = KFLAG - 1
      TN = TOLD
      I1 = NQNYH + 1
      DO 515 JB = 1,NQ
        I1 = I1 - NYH
        DO 510 I = I1,NQNYH
C510      YH1(I) = YH1(I) - YH1(I+NYH)
          YH1(I) = YH1(I) - YH1(I+NYH)
 510      CONTINUE
 515    CONTINUE
      RMAX = 2.0D0
      IF (ABS(H) .LE. HMIN*1.00001D0) GO TO 660
      IF (KFLAG .LE. -3) GO TO 640
      IREDO = 2
      RHUP = 0.0D0
      GO TO 540
C-----------------------------------------------------------------------
C Regardless of the success or failure of the step, factors
C RHDN, RHSM, and RHUP are computed, by which H could be multiplied
C at order NQ - 1, order NQ, or order NQ + 1, respectively.
C In the case of failure, RHUP = 0.0 to avoid an order increase.
C The largest of these is determined and the new order chosen
C accordingly.  If the order is to be increased, we compute one
C additional scaled derivative.
C-----------------------------------------------------------------------
 520  RHUP = 0.0D0
      IF (L .EQ. LMAX) GO TO 540
      DO 530 I = 1,N
C530    SAVF(I) = ACOR(I) - YH(I,LMAX)
        SAVF(I) = ACOR(I) - YH(I,LMAX)
 530    CONTINUE
      DUP = DMNORM (N, SAVF, EWT)/TESCO(3,NQ)
      EXUP = 1.0D0/(L+1)
      RHUP = 1.0D0/(1.4D0*DUP**EXUP + 0.0000014D0)
 540  EXSM = 1.0D0/L
      RHSM = 1.0D0/(1.2D0*DSM**EXSM + 0.0000012D0)
      RHDN = 0.0D0
      IF (NQ .EQ. 1) GO TO 550
      DDN = DMNORM (N, YH(1,L), EWT)/TESCO(1,NQ)
      EXDN = 1.0D0/NQ
      RHDN = 1.0D0/(1.3D0*DDN**EXDN + 0.0000013D0)
C If METH = 1, limit RH according to the stability region also. --------
 550  IF (METH .EQ. 2) GO TO 560
      PDH = MAX(ABS(H)*PDLAST,0.000001D0)
      IF (L .LT. LMAX) RHUP = MIN(RHUP,SM1(L)/PDH)
      RHSM = MIN(RHSM,SM1(NQ)/PDH)
      IF (NQ .GT. 1) RHDN = MIN(RHDN,SM1(NQ-1)/PDH)
      PDEST = 0.0D0
 560  IF (RHSM .GE. RHUP) GO TO 570
      IF (RHUP .GT. RHDN) GO TO 590
      GO TO 580
 570  IF (RHSM .LT. RHDN) GO TO 580
      NEWQ = NQ
      RH = RHSM
      GO TO 620
 580  NEWQ = NQ - 1
      RH = RHDN
      IF (KFLAG .LT. 0 .AND. RH .GT. 1.0D0) RH = 1.0D0
      GO TO 620
 590  NEWQ = L
      RH = RHUP
      IF (RH .LT. 1.1D0) GO TO 610
      R = EL(L)/L
      DO 600 I = 1,N
C600    YH(I,NEWQ+1) = ACOR(I)*R
        YH(I,NEWQ+1) = ACOR(I)*R
 600    CONTINUE
      GO TO 630
 610  IALTH = 3
      GO TO 700
C If METH = 1 and H is restricted by stability, bypass 10 percent test.
 620  IF (METH .EQ. 2) GO TO 622
      IF (RH*PDH*1.00001D0 .GE. SM1(NEWQ)) GO TO 625
 622  IF (KFLAG .EQ. 0 .AND. RH .LT. 1.1D0) GO TO 610
 625  IF (KFLAG .LE. -2) RH = MIN(RH,0.2D0)
C-----------------------------------------------------------------------
C If there is a change of order, reset NQ, L, and the coefficients.
C In any case H is reset according to RH and the YH array is rescaled.
C Then exit from 690 if the step was OK, or redo the step otherwise.
C-----------------------------------------------------------------------
      IF (NEWQ .EQ. NQ) GO TO 170
 630  NQ = NEWQ
      L = NQ + 1
      IRET = 2
      GO TO 150
C-----------------------------------------------------------------------
C Control reaches this section if 3 or more failures have occured.
C If 10 failures have occurred, exit with KFLAG = -1.
C It is assumed that the derivatives that have accumulated in the
C YH array have errors of the wrong order.  Hence the first
C derivative is recomputed, and the order is set to 1.  Then
C H is reduced by a factor of 10, and the step is retried,
C until it succeeds or H reaches HMIN.
C-----------------------------------------------------------------------
 640  IF (KFLAG .EQ. -10) GO TO 660
      RH = 0.1D0
      RH = MAX(HMIN/ABS(H),RH)
      H = H*RH
      DO 645 I = 1,N
C645    Y(I) = YH(I,1)
        Y(I) = YH(I,1)
 645    CONTINUE
      CALL F (NEQ, TN, Y, SAVF)
      NFE = NFE + 1
      DO 650 I = 1,N
C650    YH(I,2) = H*SAVF(I)
        YH(I,2) = H*SAVF(I)
 650    CONTINUE
      IPUP = MITER
      IALTH = 5
      IF (NQ .EQ. 1) GO TO 200
      NQ = 1
      L = 2
      IRET = 3
      GO TO 150
C-----------------------------------------------------------------------
C All returns are made through this section.  H is saved in HOLD
C to allow the caller to change H on the next step.
C-----------------------------------------------------------------------
 660  KFLAG = -1
      GO TO 720
 670  KFLAG = -2
      GO TO 720
 680  KFLAG = -3
      GO TO 720
 690  RMAX = 10.0D0
 700  R = 1.0D0/TESCO(2,NQU)
      DO 710 I = 1,N
C710    ACOR(I) = ACOR(I)*R
        ACOR(I) = ACOR(I)*R
 710    CONTINUE
 720  HOLD = H
      JSTART = 1
      RETURN
C----------------------- End of Subroutine DSTODA ----------------------
      END
*DECK DPRJA
      SUBROUTINE DPRJA (NEQ, Y, YH, NYH, EWT, FTEM, SAVF, WM, IWM,
     1   F, JAC)
      EXTERNAL F, JAC
      INTEGER NEQ, NYH, IWM
      DOUBLE PRECISION Y, YH, EWT, FTEM, SAVF, WM
      DIMENSION NEQ(*), Y(*), YH(NYH,*), EWT(*), FTEM(*), SAVF(*),
     1   WM(*), IWM(*)
      INTEGER INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH01,
     1   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     2   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     3   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     4   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      INTEGER IOWND2, IOWNS2, JTYP, MUSED, MXORDN, MXORDS
      DOUBLE PRECISION CONIT, CRATE, EL, ELCO, HOLD, RMAX, TESCO,
     1   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND
      DOUBLE PRECISION ROWND2, ROWNS2, PDNORM
      COMMON /DLS001/ CONIT, CRATE, EL(13), ELCO(13,12),
     1   HOLD, RMAX, TESCO(3,12),
     2   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND,
     3   INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH01,
     4   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     5   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     6   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     7   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      COMMON /DLSA01/ ROWND2, ROWNS2(20), PDNORM,
     1   IOWND2(3), IOWNS2(2), JTYP, MUSED, MXORDN, MXORDS
      INTEGER I, I1, I2, IER, II, J, J1, JJ, LENP,
     1   MBA, MBAND, MEB1, MEBAND, ML, ML3, MU, NP1
      DOUBLE PRECISION CON, FAC, HL0, R, R0, SRUR, YI, YJ, YJJ,
     1   DMNORM, DFNORM, DBNORM
C-----------------------------------------------------------------------
C DPRJA is called by DSTODA to compute and process the matrix
C P = I - H*EL(1)*J , where J is an approximation to the Jacobian.
C Here J is computed by the user-supplied routine JAC if
C MITER = 1 or 4 or by finite differencing if MITER = 2 or 5.
C J, scaled by -H*EL(1), is stored in WM.  Then the norm of J (the
C matrix norm consistent with the weighted max-norm on vectors given
C by DMNORM) is computed, and J is overwritten by P.  P is then
C subjected to LU decomposition in preparation for later solution
C of linear systems with P as coefficient matrix.  This is done
C by DGEFA if MITER = 1 or 2, and by DGBFA if MITER = 4 or 5.
C
C In addition to variables described previously, communication
C with DPRJA uses the following:
C Y     = array containing predicted values on entry.
C FTEM  = work array of length N (ACOR in DSTODA).
C SAVF  = array containing f evaluated at predicted y.
C WM    = real work space for matrices.  On output it contains the
C         LU decomposition of P.
C         Storage of matrix elements starts at WM(3).
C         WM also contains the following matrix-related data:
C         WM(1) = SQRT(UROUND), used in numerical Jacobian increments.
C IWM   = integer work space containing pivot information, starting at
C         IWM(21).   IWM also contains the band parameters
C         ML = IWM(1) and MU = IWM(2) if MITER is 4 or 5.
C EL0   = EL(1) (input).
C PDNORM= norm of Jacobian matrix. (Output).
C IERPJ = output error flag,  = 0 if no trouble, .gt. 0 if
C         P matrix found to be singular.
C JCUR  = output flag = 1 to indicate that the Jacobian matrix
C         (or approximation) is now current.
C This routine also uses the Common variables EL0, H, TN, UROUND,
C MITER, N, NFE, and NJE.
C-----------------------------------------------------------------------
      NJE = NJE + 1
      IERPJ = 0
      JCUR = 1
      HL0 = H*EL0
C     GO TO (100, 200, 300, 400, 500), MITER
      IF (MITER .EQ. 1) GO TO 100
      IF (MITER .EQ. 2) GO TO 200
      IF (MITER .EQ. 3) GO TO 300
      IF (MITER .EQ. 4) GO TO 400
      IF (MITER .EQ. 5) GO TO 500
C If MITER = 1, call JAC and multiply by scalar. -----------------------
 100  LENP = N*N
      DO 110 I = 1,LENP
C110    WM(I+2) = 0.0D0
        WM(I+2) = 0.0D0
 110    CONTINUE
      CALL JAC (NEQ, TN, Y, 0, 0, WM(3), N)
      CON = -HL0
      DO 120 I = 1,LENP
C120    WM(I+2) = WM(I+2)*CON
        WM(I+2) = WM(I+2)*CON
 120    CONTINUE
      GO TO 240
C If MITER = 2, make N calls to F to approximate J. --------------------
 200  FAC = DMNORM (N, SAVF, EWT)
      R0 = 1000.0D0*ABS(H)*UROUND*N*FAC
      IF (R0 .EQ. 0.0D0) R0 = 1.0D0
      SRUR = WM(1)
      J1 = 2
      DO 230 J = 1,N
        YJ = Y(J)
        R = MAX(SRUR*ABS(YJ),R0/EWT(J))
        Y(J) = Y(J) + R
        FAC = -HL0/R
        CALL F (NEQ, TN, Y, FTEM)
        DO 220 I = 1,N
C220      WM(I+J1) = (FTEM(I) - SAVF(I))*FAC
          WM(I+J1) = (FTEM(I) - SAVF(I))*FAC
 220      CONTINUE
        Y(J) = YJ
        J1 = J1 + N
 230    CONTINUE
      NFE = NFE + N
 240  CONTINUE
C Compute norm of Jacobian. --------------------------------------------
      PDNORM = DFNORM (N, WM(3), EWT)/ABS(HL0)
C Add identity matrix. -------------------------------------------------
      J = 3
      NP1 = N + 1
      DO 250 I = 1,N
        WM(J) = WM(J) + 1.0D0
C250    J = J + NP1
        J = J + NP1
 250    CONTINUE
C Do LU decomposition on P. --------------------------------------------
      CALL DGEFA (WM(3), N, N, IWM(21), IER)
      IF (IER .NE. 0) IERPJ = 1
      RETURN
C Dummy block only, since MITER is never 3 in this routine. ------------
 300  RETURN
C If MITER = 4, call JAC and multiply by scalar. -----------------------
 400  ML = IWM(1)
      MU = IWM(2)
      ML3 = ML + 3
      MBAND = ML + MU + 1
      MEBAND = MBAND + ML
      LENP = MEBAND*N
      DO 410 I = 1,LENP
C410    WM(I+2) = 0.0D0
        WM(I+2) = 0.0D0
 410    CONTINUE
      CALL JAC (NEQ, TN, Y, ML, MU, WM(ML3), MEBAND)
      CON = -HL0
      DO 420 I = 1,LENP
C420    WM(I+2) = WM(I+2)*CON
        WM(I+2) = WM(I+2)*CON
 420    CONTINUE
      GO TO 570
C If MITER = 5, make MBAND calls to F to approximate J. ----------------
 500  ML = IWM(1)
      MU = IWM(2)
      MBAND = ML + MU + 1
      MBA = MIN(MBAND,N)
      MEBAND = MBAND + ML
      MEB1 = MEBAND - 1
      SRUR = WM(1)
      FAC = DMNORM (N, SAVF, EWT)
      R0 = 1000.0D0*ABS(H)*UROUND*N*FAC
      IF (R0 .EQ. 0.0D0) R0 = 1.0D0
      DO 560 J = 1,MBA
        DO 530 I = J,N,MBAND
          YI = Y(I)
          R = MAX(SRUR*ABS(YI),R0/EWT(I))
C530      Y(I) = Y(I) + R
          Y(I) = Y(I) + R
 530      CONTINUE
        CALL F (NEQ, TN, Y, FTEM)
        DO 550 JJ = J,N,MBAND
          Y(JJ) = YH(JJ,1)
          YJJ = Y(JJ)
          R = MAX(SRUR*ABS(YJJ),R0/EWT(JJ))
          FAC = -HL0/R
          I1 = MAX(JJ-MU,1)
          I2 = MIN(JJ+ML,N)
          II = JJ*MEB1 - ML + 2
          DO 540 I = I1,I2
C540        WM(II+I) = (FTEM(I) - SAVF(I))*FAC
            WM(II+I) = (FTEM(I) - SAVF(I))*FAC
 540        CONTINUE
 550      CONTINUE
 560    CONTINUE
      NFE = NFE + MBA
 570  CONTINUE
C Compute norm of Jacobian. --------------------------------------------
      PDNORM = DBNORM (N, WM(ML+3), MEBAND, ML, MU, EWT)/ABS(HL0)
C Add identity matrix. -------------------------------------------------
      II = MBAND + 2
      DO 580 I = 1,N
        WM(II) = WM(II) + 1.0D0
C580    II = II + MEBAND
        II = II + MEBAND
 580    CONTINUE
C Do LU decomposition of P. --------------------------------------------
      CALL DGBFA (WM(3), MEBAND, N, ML, MU, IWM(21), IER)
      IF (IER .NE. 0) IERPJ = 1
      RETURN
C----------------------- End of Subroutine DPRJA -----------------------
      END
*DECK DMNORM
      DOUBLE PRECISION FUNCTION DMNORM (N, V, W)
C-----------------------------------------------------------------------
C This function routine computes the weighted max-norm
C of the vector of length N contained in the array V, with weights
C contained in the array w of length N:
C   DMNORM = MAX(i=1,...,N) ABS(V(i))*W(i)
C-----------------------------------------------------------------------
      INTEGER N,   I
      DOUBLE PRECISION V, W,   VM
      DIMENSION V(N), W(N)
      VM = 0.0D0
      DO 10 I = 1,N
C10     VM = MAX(VM,ABS(V(I))*W(I))
        VM = MAX(VM,ABS(V(I))*W(I))
 10     CONTINUE
      DMNORM = VM
      RETURN
C----------------------- End of Function DMNORM ------------------------
      END
*DECK DFNORM
      DOUBLE PRECISION FUNCTION DFNORM (N, A, W)
C-----------------------------------------------------------------------
C This function computes the norm of a full N by N matrix,
C stored in the array A, that is consistent with the weighted max-norm
C on vectors, with weights stored in the array W:
C   DFNORM = MAX(i=1,...,N) ( W(i) * Sum(j=1,...,N) ABS(a(i,j))/W(j) )
C-----------------------------------------------------------------------
      INTEGER N,   I, J
      DOUBLE PRECISION A,   W, AN, SUM
      DIMENSION A(N,N), W(N)
      AN = 0.0D0
      DO 20 I = 1,N
        SUM = 0.0D0
        DO 10 J = 1,N
C10       SUM = SUM + ABS(A(I,J))/W(J)
          SUM = SUM + ABS(A(I,J))/W(J)
 10       CONTINUE
        AN = MAX(AN,SUM*W(I))
 20     CONTINUE
      DFNORM = AN
      RETURN
C----------------------- End of Function DFNORM ------------------------
      END
*DECK DBNORM
      DOUBLE PRECISION FUNCTION DBNORM (N, A, NRA, ML, MU, W)
C-----------------------------------------------------------------------
C This function computes the norm of a banded N by N matrix,
C stored in the array A, that is consistent with the weighted max-norm
C on vectors, with weights stored in the array W.
C ML and MU are the lower and upper half-bandwidths of the matrix.
C NRA is the first dimension of the A array, NRA .ge. ML+MU+1.
C In terms of the matrix elements a(i,j), the norm is given by:
C   DBNORM = MAX(i=1,...,N) ( W(i) * Sum(j=1,...,N) ABS(a(i,j))/W(j) )
C-----------------------------------------------------------------------
      INTEGER N, NRA, ML, MU
      INTEGER I, I1, JLO, JHI, J
      DOUBLE PRECISION A, W
      DOUBLE PRECISION AN, SUM
      DIMENSION A(NRA,N), W(N)
      AN = 0.0D0
      DO 20 I = 1,N
        SUM = 0.0D0
        I1 = I + MU + 1
        JLO = MAX(I-ML,1)
        JHI = MIN(I+MU,N)
        DO 10 J = JLO,JHI
C10       SUM = SUM + ABS(A(I1-J,J))/W(J)
          SUM = SUM + ABS(A(I1-J,J))/W(J)
 10       CONTINUE
        AN = MAX(AN,SUM*W(I))
 20     CONTINUE
      DBNORM = AN
      RETURN
C----------------------- End of Function DBNORM ------------------------
      END
*DECK DLSODA
      SUBROUTINE DLSODA (F, NEQ, Y, T, TOUT, ITOL, RTOL, ATOL, ITASK,
     1            ISTATE, IOPT, RWORK, LRW, IWORK, LIW, JAC, JT)
      EXTERNAL F, JAC
      INTEGER NEQ, ITOL, ITASK, ISTATE, IOPT, LRW, IWORK, LIW, JT
      DOUBLE PRECISION Y, T, TOUT, RTOL, ATOL, RWORK
      DIMENSION NEQ(*), Y(*), RTOL(*), ATOL(*), RWORK(LRW), IWORK(LIW)
C-----------------------------------------------------------------------
C This is the 12 November 2003 version of
C DLSODA: Livermore Solver for Ordinary Differential Equations, with
C         Automatic method switching for stiff and nonstiff problems.
C
C This version is in double precision.
C
C DLSODA solves the initial value problem for stiff or nonstiff
C systems of first order ODEs,
C     dy/dt = f(t,y) ,  or, in component form,
C     dy(i)/dt = f(i) = f(i,t,y(1),y(2),...,y(NEQ)) (i = 1,...,NEQ).
C
C This a variant version of the DLSODE package.
C It switches automatically between stiff and nonstiff methods.
C This means that the user does not have to determine whether the
C problem is stiff or not, and the solver will automatically choose the
C appropriate method.  It always starts with the nonstiff method.
C
C Authors:       Alan C. Hindmarsh
C                Center for Applied Scientific Computing, L-561
C                Lawrence Livermore National Laboratory
C                Livermore, CA 94551
C and
C                Linda R. Petzold
C                Univ. of California at Santa Barbara
C                Dept. of Computer Science
C                Santa Barbara, CA 93106
C
C References:
C 1.  Alan C. Hindmarsh,  ODEPACK, A Systematized Collection of ODE
C     Solvers, in Scientific Computing, R. S. Stepleman et al. (Eds.),
C     North-Holland, Amsterdam, 1983, pp. 55-64.
C 2.  Linda R. Petzold, Automatic Selection of Methods for Solving
C     Stiff and Nonstiff Systems of Ordinary Differential Equations,
C     Siam J. Sci. Stat. Comput. 4 (1983), pp. 136-148.
C-----------------------------------------------------------------------
C Summary of Usage.
C
C Communication between the user and the DLSODA package, for normal
C situations, is summarized here.  This summary describes only a subset
C of the full set of options available.  See the full description for
C details, including alternative treatment of the Jacobian matrix,
C optional inputs and outputs, nonstandard options, and
C instructions for special situations.  See also the example
C problem (with program and output) following this summary.
C
C A. First provide a subroutine of the form:
C               SUBROUTINE F (NEQ, T, Y, YDOT)
C               DOUBLE PRECISION T, Y(*), YDOT(*)
C which supplies the vector function f by loading YDOT(i) with f(i).
C
C B. Write a main program which calls Subroutine DLSODA once for
C each point at which answers are desired.  This should also provide
C for possible use of logical unit 6 for output of error messages
C by DLSODA.  On the first call to DLSODA, supply arguments as follows:
C F      = name of subroutine for right-hand side vector f.
C          This name must be declared External in calling program.
C NEQ    = number of first order ODEs.
C Y      = array of initial values, of length NEQ.
C T      = the initial value of the independent variable.
C TOUT   = first point where output is desired (.ne. T).
C ITOL   = 1 or 2 according as ATOL (below) is a scalar or array.
C RTOL   = relative tolerance parameter (scalar).
C ATOL   = absolute tolerance parameter (scalar or array).
C          the estimated local error in y(i) will be controlled so as
C          to be less than
C             EWT(i) = RTOL*ABS(Y(i)) + ATOL     if ITOL = 1, or
C             EWT(i) = RTOL*ABS(Y(i)) + ATOL(i)  if ITOL = 2.
C          Thus the local error test passes if, in each component,
C          either the absolute error is less than ATOL (or ATOL(i)),
C          or the relative error is less than RTOL.
C          Use RTOL = 0.0 for pure absolute error control, and
C          use ATOL = 0.0 (or ATOL(i) = 0.0) for pure relative error
C          control.  Caution: actual (global) errors may exceed these
C          local tolerances, so choose them conservatively.
C ITASK  = 1 for normal computation of output values of y at t = TOUT.
C ISTATE = integer flag (input and output).  Set ISTATE = 1.
C IOPT   = 0 to indicate no optional inputs used.
C RWORK  = real work array of length at least:
C             22 + NEQ * MAX(16, NEQ + 9).
C          See also Paragraph E below.
C LRW    = declared length of RWORK (in user's dimension).
C IWORK  = integer work array of length at least  20 + NEQ.
C LIW    = declared length of IWORK (in user's dimension).
C JAC    = name of subroutine for Jacobian matrix.
C          Use a dummy name.  See also Paragraph E below.
C JT     = Jacobian type indicator.  Set JT = 2.
C          See also Paragraph E below.
C Note that the main program must declare arrays Y, RWORK, IWORK,
C and possibly ATOL.
C
C C. The output from the first call (or any call) is:
C      Y = array of computed values of y(t) vector.
C      T = corresponding value of independent variable (normally TOUT).
C ISTATE = 2  if DLSODA was successful, negative otherwise.
C          -1 means excess work done on this call (perhaps wrong JT).
C          -2 means excess accuracy requested (tolerances too small).
C          -3 means illegal input detected (see printed message).
C          -4 means repeated error test failures (check all inputs).
C          -5 means repeated convergence failures (perhaps bad Jacobian
C             supplied or wrong choice of JT or tolerances).
C          -6 means error weight became zero during problem. (Solution
C             component i vanished, and ATOL or ATOL(i) = 0.)
C          -7 means work space insufficient to finish (see messages).
C
C D. To continue the integration after a successful return, simply
C reset TOUT and call DLSODA again.  No other parameters need be reset.
C
C E. Note: If and when DLSODA regards the problem as stiff, and
C switches methods accordingly, it must make use of the NEQ by NEQ
C Jacobian matrix, J = df/dy.  For the sake of simplicity, the
C inputs to DLSODA recommended in Paragraph B above cause DLSODA to
C treat J as a full matrix, and to approximate it internally by
C difference quotients.  Alternatively, J can be treated as a band
C matrix (with great potential reduction in the size of the RWORK
C array).  Also, in either the full or banded case, the user can supply
C J in closed form, with a routine whose name is passed as the JAC
C argument.  These alternatives are described in the paragraphs on
C RWORK, JAC, and JT in the full description of the call sequence below.
C
C-----------------------------------------------------------------------
C Example Problem.
C
C The following is a simple example problem, with the coding
C needed for its solution by DLSODA.  The problem is from chemical
C kinetics, and consists of the following three rate equations:
C     dy1/dt = -.04*y1 + 1.e4*y2*y3
C     dy2/dt = .04*y1 - 1.e4*y2*y3 - 3.e7*y2**2
C     dy3/dt = 3.e7*y2**2
C on the interval from t = 0.0 to t = 4.e10, with initial conditions
C y1 = 1.0, y2 = y3 = 0.  The problem is stiff.
C
C The following coding solves this problem with DLSODA,
C printing results at t = .4, 4., ..., 4.e10.  It uses
C ITOL = 2 and ATOL much smaller for y2 than y1 or y3 because
C y2 has much smaller values.
C At the end of the run, statistical quantities of interest are
C printed (see optional outputs in the full description below).
C
C     EXTERNAL FEX
C     DOUBLE PRECISION ATOL, RTOL, RWORK, T, TOUT, Y
C     DIMENSION Y(3), ATOL(3), RWORK(70), IWORK(23)
C     NEQ = 3
C     Y(1) = 1.
C     Y(2) = 0.
C     Y(3) = 0.
C     T = 0.
C     TOUT = .4
C     ITOL = 2
C     RTOL = 1.D-4
C     ATOL(1) = 1.D-6
C     ATOL(2) = 1.D-10
C     ATOL(3) = 1.D-6
C     ITASK = 1
C     ISTATE = 1
C     IOPT = 0
C     LRW = 70
C     LIW = 23
C     JT = 2
C     DO 40 IOUT = 1,12
C       CALL DLSODA(FEX,NEQ,Y,T,TOUT,ITOL,RTOL,ATOL,ITASK,ISTATE,
C    1     IOPT,RWORK,LRW,IWORK,LIW,JDUM,JT)
C       WRITE(6,20)T,Y(1),Y(2),Y(3)
C 20    FORMAT(' At t =',D12.4,'   Y =',3D14.6)
C       IF (ISTATE .LT. 0) GO TO 80
C 40    TOUT = TOUT*10.
C     WRITE(6,60)IWORK(11),IWORK(12),IWORK(13),IWORK(19),RWORK(15)
C 60  FORMAT(/' No. steps =',I4,'  No. f-s =',I4,'  No. J-s =',I4/
C    1   ' Method last used =',I2,'   Last switch was at t =',D12.4)
C     STOP
C 80  WRITE(6,90)ISTATE
C 90  FORMAT(///' Error halt.. ISTATE =',I3)
C     STOP
C     END
C
C     SUBROUTINE FEX (NEQ, T, Y, YDOT)
C     DOUBLE PRECISION T, Y, YDOT
C     DIMENSION Y(3), YDOT(3)
C     YDOT(1) = -.04*Y(1) + 1.D4*Y(2)*Y(3)
C     YDOT(3) = 3.D7*Y(2)*Y(2)
C     YDOT(2) = -YDOT(1) - YDOT(3)
C     RETURN
C     END
C
C The output of this program (on a CDC-7600 in single precision)
C is as follows:
C
C   At t =  4.0000e-01   y =  9.851712e-01  3.386380e-05  1.479493e-02
C   At t =  4.0000e+00   Y =  9.055333e-01  2.240655e-05  9.444430e-02
C   At t =  4.0000e+01   Y =  7.158403e-01  9.186334e-06  2.841505e-01
C   At t =  4.0000e+02   Y =  4.505250e-01  3.222964e-06  5.494717e-01
C   At t =  4.0000e+03   Y =  1.831975e-01  8.941774e-07  8.168016e-01
C   At t =  4.0000e+04   Y =  3.898730e-02  1.621940e-07  9.610125e-01
C   At t =  4.0000e+05   Y =  4.936363e-03  1.984221e-08  9.950636e-01
C   At t =  4.0000e+06   Y =  5.161831e-04  2.065786e-09  9.994838e-01
C   At t =  4.0000e+07   Y =  5.179817e-05  2.072032e-10  9.999482e-01
C   At t =  4.0000e+08   Y =  5.283401e-06  2.113371e-11  9.999947e-01
C   At t =  4.0000e+09   Y =  4.659031e-07  1.863613e-12  9.999995e-01
C   At t =  4.0000e+10   Y =  1.404280e-08  5.617126e-14  1.000000e+00
C
C   No. steps = 361  No. f-s = 693  No. J-s =  64
C   Method last used = 2   Last switch was at t =  6.0092e-03
C-----------------------------------------------------------------------
C Full description of user interface to DLSODA.
C
C The user interface to DLSODA consists of the following parts.
C
C 1.   The call sequence to Subroutine DLSODA, which is a driver
C      routine for the solver.  This includes descriptions of both
C      the call sequence arguments and of user-supplied routines.
C      following these descriptions is a description of
C      optional inputs available through the call sequence, and then
C      a description of optional outputs (in the work arrays).
C
C 2.   Descriptions of other routines in the DLSODA package that may be
C      (optionally) called by the user.  These provide the ability to
C      alter error message handling, save and restore the internal
C      Common, and obtain specified derivatives of the solution y(t).
C
C 3.   Descriptions of Common blocks to be declared in overlay
C      or similar environments, or to be saved when doing an interrupt
C      of the problem and continued solution later.
C
C 4.   Description of a subroutine in the DLSODA package,
C      which the user may replace with his/her own version, if desired.
C      this relates to the measurement of errors.
C
C-----------------------------------------------------------------------
C Part 1.  Call Sequence.
C
C The call sequence parameters used for input only are
C     F, NEQ, TOUT, ITOL, RTOL, ATOL, ITASK, IOPT, LRW, LIW, JAC, JT,
C and those used for both input and output are
C     Y, T, ISTATE.
C The work arrays RWORK and IWORK are also used for conditional and
C optional inputs and optional outputs.  (The term output here refers
C to the return from Subroutine DLSODA to the user's calling program.)
C
C The legality of input parameters will be thoroughly checked on the
C initial call for the problem, but not checked thereafter unless a
C change in input parameters is flagged by ISTATE = 3 on input.
C
C The descriptions of the call arguments are as follows.
C
C F      = the name of the user-supplied subroutine defining the
C          ODE system.  The system must be put in the first-order
C          form dy/dt = f(t,y), where f is a vector-valued function
C          of the scalar t and the vector y.  Subroutine F is to
C          compute the function f.  It is to have the form
C               SUBROUTINE F (NEQ, T, Y, YDOT)
C               DOUBLE PRECISION T, Y(*), YDOT(*)
C          where NEQ, T, and Y are input, and the array YDOT = f(t,y)
C          is output.  Y and YDOT are arrays of length NEQ.
C          Subroutine F should not alter Y(1),...,Y(NEQ).
C          F must be declared External in the calling program.
C
C          Subroutine F may access user-defined quantities in
C          NEQ(2),... and/or in Y(NEQ(1)+1),... if NEQ is an array
C          (dimensioned in F) and/or Y has length exceeding NEQ(1).
C          See the descriptions of NEQ and Y below.
C
C          If quantities computed in the F routine are needed
C          externally to DLSODA, an extra call to F should be made
C          for this purpose, for consistent and accurate results.
C          If only the derivative dy/dt is needed, use DINTDY instead.
C
C NEQ    = the size of the ODE system (number of first order
C          ordinary differential equations).  Used only for input.
C          NEQ may be decreased, but not increased, during the problem.
C          If NEQ is decreased (with ISTATE = 3 on input), the
C          remaining components of Y should be left undisturbed, if
C          these are to be accessed in F and/or JAC.
C
C          Normally, NEQ is a scalar, and it is generally referred to
C          as a scalar in this user interface description.  However,
C          NEQ may be an array, with NEQ(1) set to the system size.
C          (The DLSODA package accesses only NEQ(1).)  In either case,
C          this parameter is passed as the NEQ argument in all calls
C          to F and JAC.  Hence, if it is an array, locations
C          NEQ(2),... may be used to store other integer data and pass
C          it to F and/or JAC.  Subroutines F and/or JAC must include
C          NEQ in a Dimension statement in that case.
C
C Y      = a real array for the vector of dependent variables, of
C          length NEQ or more.  Used for both input and output on the
C          first call (ISTATE = 1), and only for output on other calls.
C          On the first call, Y must contain the vector of initial
C          values.  On output, Y contains the computed solution vector,
C          evaluated at T.  If desired, the Y array may be used
C          for other purposes between calls to the solver.
C
C          This array is passed as the Y argument in all calls to
C          F and JAC.  Hence its length may exceed NEQ, and locations
C          Y(NEQ+1),... may be used to store other real data and
C          pass it to F and/or JAC.  (The DLSODA package accesses only
C          Y(1),...,Y(NEQ).)
C
C T      = the independent variable.  On input, T is used only on the
C          first call, as the initial point of the integration.
C          on output, after each call, T is the value at which a
C          computed solution Y is evaluated (usually the same as TOUT).
C          on an error return, T is the farthest point reached.
C
C TOUT   = the next value of t at which a computed solution is desired.
C          Used only for input.
C
C          When starting the problem (ISTATE = 1), TOUT may be equal
C          to T for one call, then should .ne. T for the next call.
C          For the initial t, an input value of TOUT .ne. T is used
C          in order to determine the direction of the integration
C          (i.e. the algebraic sign of the step sizes) and the rough
C          scale of the problem.  Integration in either direction
C          (forward or backward in t) is permitted.
C
C          If ITASK = 2 or 5 (one-step modes), TOUT is ignored after
C          the first call (i.e. the first call with TOUT .ne. T).
C          Otherwise, TOUT is required on every call.
C
C          If ITASK = 1, 3, or 4, the values of TOUT need not be
C          monotone, but a value of TOUT which backs up is limited
C          to the current internal T interval, whose endpoints are
C          TCUR - HU and TCUR (see optional outputs, below, for
C          TCUR and HU).
C
C ITOL   = an indicator for the type of error control.  See
C          description below under ATOL.  Used only for input.
C
C RTOL   = a relative error tolerance parameter, either a scalar or
C          an array of length NEQ.  See description below under ATOL.
C          Input only.
C
C ATOL   = an absolute error tolerance parameter, either a scalar or
C          an array of length NEQ.  Input only.
C
C             The input parameters ITOL, RTOL, and ATOL determine
C          the error control performed by the solver.  The solver will
C          control the vector E = (E(i)) of estimated local errors
C          in y, according to an inequality of the form
C                      max-norm of ( E(i)/EWT(i) )   .le.   1,
C          where EWT = (EWT(i)) is a vector of positive error weights.
C          The values of RTOL and ATOL should all be non-negative.
C          The following table gives the types (scalar/array) of
C          RTOL and ATOL, and the corresponding form of EWT(i).
C
C             ITOL    RTOL       ATOL          EWT(i)
C              1     scalar     scalar     RTOL*ABS(Y(i)) + ATOL
C              2     scalar     array      RTOL*ABS(Y(i)) + ATOL(i)
C              3     array      scalar     RTOL(i)*ABS(Y(i)) + ATOL
C              4     array      array      RTOL(i)*ABS(Y(i)) + ATOL(i)
C
C          When either of these parameters is a scalar, it need not
C          be dimensioned in the user's calling program.
C
C          If none of the above choices (with ITOL, RTOL, and ATOL
C          fixed throughout the problem) is suitable, more general
C          error controls can be obtained by substituting a
C          user-supplied routine for the setting of EWT.
C          See Part 4 below.
C
C          If global errors are to be estimated by making a repeated
C          run on the same problem with smaller tolerances, then all
C          components of RTOL and ATOL (i.e. of EWT) should be scaled
C          down uniformly.
C
C ITASK  = an index specifying the task to be performed.
C          Input only.  ITASK has the following values and meanings.
C          1  means normal computation of output values of y(t) at
C             t = TOUT (by overshooting and interpolating).
C          2  means take one step only and return.
C          3  means stop at the first internal mesh point at or
C             beyond t = TOUT and return.
C          4  means normal computation of output values of y(t) at
C             t = TOUT but without overshooting t = TCRIT.
C             TCRIT must be input as RWORK(1).  TCRIT may be equal to
C             or beyond TOUT, but not behind it in the direction of
C             integration.  This option is useful if the problem
C             has a singularity at or beyond t = TCRIT.
C          5  means take one step, without passing TCRIT, and return.
C             TCRIT must be input as RWORK(1).
C
C          Note:  If ITASK = 4 or 5 and the solver reaches TCRIT
C          (within roundoff), it will return T = TCRIT (exactly) to
C          indicate this (unless ITASK = 4 and TOUT comes before TCRIT,
C          in which case answers at t = TOUT are returned first).
C
C ISTATE = an index used for input and output to specify the
C          the state of the calculation.
C
C          On input, the values of ISTATE are as follows.
C          1  means this is the first call for the problem
C             (initializations will be done).  See note below.
C          2  means this is not the first call, and the calculation
C             is to continue normally, with no change in any input
C             parameters except possibly TOUT and ITASK.
C             (If ITOL, RTOL, and/or ATOL are changed between calls
C             with ISTATE = 2, the new values will be used but not
C             tested for legality.)
C          3  means this is not the first call, and the
C             calculation is to continue normally, but with
C             a change in input parameters other than
C             TOUT and ITASK.  Changes are allowed in
C             NEQ, ITOL, RTOL, ATOL, IOPT, LRW, LIW, JT, ML, MU,
C             and any optional inputs except H0, MXORDN, and MXORDS.
C             (See IWORK description for ML and MU.)
C          Note:  A preliminary call with TOUT = T is not counted
C          as a first call here, as no initialization or checking of
C          input is done.  (Such a call is sometimes useful for the
C          purpose of outputting the initial conditions.)
C          Thus the first call for which TOUT .ne. T requires
C          ISTATE = 1 on input.
C
C          On output, ISTATE has the following values and meanings.
C           1  means nothing was done; TOUT = T and ISTATE = 1 on input.
C           2  means the integration was performed successfully.
C          -1  means an excessive amount of work (more than MXSTEP
C              steps) was done on this call, before completing the
C              requested task, but the integration was otherwise
C              successful as far as T.  (MXSTEP is an optional input
C              and is normally 500.)  To continue, the user may
C              simply reset ISTATE to a value .gt. 1 and call again
C              (the excess work step counter will be reset to 0).
C              In addition, the user may increase MXSTEP to avoid
C              this error return (see below on optional inputs).
C          -2  means too much accuracy was requested for the precision
C              of the machine being used.  This was detected before
C              completing the requested task, but the integration
C              was successful as far as T.  To continue, the tolerance
C              parameters must be reset, and ISTATE must be set
C              to 3.  The optional output TOLSF may be used for this
C              purpose.  (Note: If this condition is detected before
C              taking any steps, then an illegal input return
C              (ISTATE = -3) occurs instead.)
C          -3  means illegal input was detected, before taking any
C              integration steps.  See written message for details.
C              Note:  If the solver detects an infinite loop of calls
C              to the solver with illegal input, it will cause
C              the run to stop.
C          -4  means there were repeated error test failures on
C              one attempted step, before completing the requested
C              task, but the integration was successful as far as T.
C              The problem may have a singularity, or the input
C              may be inappropriate.
C          -5  means there were repeated convergence test failures on
C              one attempted step, before completing the requested
C              task, but the integration was successful as far as T.
C              This may be caused by an inaccurate Jacobian matrix,
C              if one is being used.
C          -6  means EWT(i) became zero for some i during the
C              integration.  Pure relative error control (ATOL(i)=0.0)
C              was requested on a variable which has now vanished.
C              The integration was successful as far as T.
C          -7  means the length of RWORK and/or IWORK was too small to
C              proceed, but the integration was successful as far as T.
C              This happens when DLSODA chooses to switch methods
C              but LRW and/or LIW is too small for the new method.
C
C          Note:  Since the normal output value of ISTATE is 2,
C          it does not need to be reset for normal continuation.
C          Also, since a negative input value of ISTATE will be
C          regarded as illegal, a negative output value requires the
C          user to change it, and possibly other inputs, before
C          calling the solver again.
C
C IOPT   = an integer flag to specify whether or not any optional
C          inputs are being used on this call.  Input only.
C          The optional inputs are listed separately below.
C          IOPT = 0 means no optional inputs are being used.
C                   default values will be used in all cases.
C          IOPT = 1 means one or more optional inputs are being used.
C
C RWORK  = a real array (double precision) for work space, and (in the
C          first 20 words) for conditional and optional inputs and
C          optional outputs.
C          As DLSODA switches automatically between stiff and nonstiff
C          methods, the required length of RWORK can change during the
C          problem.  Thus the RWORK array passed to DLSODA can either
C          have a static (fixed) length large enough for both methods,
C          or have a dynamic (changing) length altered by the calling
C          program in response to output from DLSODA.
C
C                       --- Fixed Length Case ---
C          If the RWORK length is to be fixed, it should be at least
C               MAX (LRN, LRS),
C          where LRN and LRS are the RWORK lengths required when the
C          current method is nonstiff or stiff, respectively.
C
C          The separate RWORK length requirements LRN and LRS are
C          as follows:
C          IF NEQ is constant and the maximum method orders have
C          their default values, then
C             LRN = 20 + 16*NEQ,
C             LRS = 22 + 9*NEQ + NEQ**2           if JT = 1 or 2,
C             LRS = 22 + 10*NEQ + (2*ML+MU)*NEQ   if JT = 4 or 5.
C          Under any other conditions, LRN and LRS are given by:
C             LRN = 20 + NYH*(MXORDN+1) + 3*NEQ,
C             LRS = 20 + NYH*(MXORDS+1) + 3*NEQ + LMAT,
C          where
C             NYH    = the initial value of NEQ,
C             MXORDN = 12, unless a smaller value is given as an
C                      optional input,
C             MXORDS = 5, unless a smaller value is given as an
C                      optional input,
C             LMAT   = length of matrix work space:
C             LMAT   = NEQ**2 + 2              if JT = 1 or 2,
C             LMAT   = (2*ML + MU + 1)*NEQ + 2 if JT = 4 or 5.
C
C                       --- Dynamic Length Case ---
C          If the length of RWORK is to be dynamic, then it should
C          be at least LRN or LRS, as defined above, depending on the
C          current method.  Initially, it must be at least LRN (since
C          DLSODA starts with the nonstiff method).  On any return
C          from DLSODA, the optional output MCUR indicates the current
C          method.  If MCUR differs from the value it had on the
C          previous return, or if there has only been one call to
C          DLSODA and MCUR is now 2, then DLSODA has switched
C          methods during the last call, and the length of RWORK
C          should be reset (to LRN if MCUR = 1, or to LRS if
C          MCUR = 2).  (An increase in the RWORK length is required
C          if DLSODA returned ISTATE = -7, but not otherwise.)
C          After resetting the length, call DLSODA with ISTATE = 3
C          to signal that change.
C
C LRW    = the length of the array RWORK, as declared by the user.
C          (This will be checked by the solver.)
C
C IWORK  = an integer array for work space.
C          As DLSODA switches automatically between stiff and nonstiff
C          methods, the required length of IWORK can change during
C          problem, between
C             LIS = 20 + NEQ   and   LIN = 20,
C          respectively.  Thus the IWORK array passed to DLSODA can
C          either have a fixed length of at least 20 + NEQ, or have a
C          dynamic length of at least LIN or LIS, depending on the
C          current method.  The comments on dynamic length under
C          RWORK above apply here.  Initially, this length need
C          only be at least LIN = 20.
C
C          The first few words of IWORK are used for conditional and
C          optional inputs and optional outputs.
C
C          The following 2 words in IWORK are conditional inputs:
C            IWORK(1) = ML     these are the lower and upper
C            IWORK(2) = MU     half-bandwidths, respectively, of the
C                       banded Jacobian, excluding the main diagonal.
C                       The band is defined by the matrix locations
C                       (i,j) with i-ML .le. j .le. i+MU.  ML and MU
C                       must satisfy  0 .le.  ML,MU  .le. NEQ-1.
C                       These are required if JT is 4 or 5, and
C                       ignored otherwise.  ML and MU may in fact be
C                       the band parameters for a matrix to which
C                       df/dy is only approximately equal.
C
C LIW    = the length of the array IWORK, as declared by the user.
C          (This will be checked by the solver.)
C
C Note: The base addresses of the work arrays must not be
C altered between calls to DLSODA for the same problem.
C The contents of the work arrays must not be altered
C between calls, except possibly for the conditional and
C optional inputs, and except for the last 3*NEQ words of RWORK.
C The latter space is used for internal scratch space, and so is
C available for use by the user outside DLSODA between calls, if
C desired (but not for use by F or JAC).
C
C JAC    = the name of the user-supplied routine to compute the
C          Jacobian matrix, df/dy, if JT = 1 or 4.  The JAC routine
C          is optional, but if the problem is expected to be stiff much
C          of the time, you are encouraged to supply JAC, for the sake
C          of efficiency.  (Alternatively, set JT = 2 or 5 to have
C          DLSODA compute df/dy internally by difference quotients.)
C          If and when DLSODA uses df/dy, it treats this NEQ by NEQ
C          matrix either as full (JT = 1 or 2), or as banded (JT =
C          4 or 5) with half-bandwidths ML and MU (discussed under
C          IWORK above).  In either case, if JT = 1 or 4, the JAC
C          routine must compute df/dy as a function of the scalar t
C          and the vector y.  It is to have the form
C               SUBROUTINE JAC (NEQ, T, Y, ML, MU, PD, NROWPD)
C               DOUBLE PRECISION T, Y(*), PD(NROWPD,*)
C          where NEQ, T, Y, ML, MU, and NROWPD are input and the array
C          PD is to be loaded with partial derivatives (elements of
C          the Jacobian matrix) on output.  PD must be given a first
C          dimension of NROWPD.  T and Y have the same meaning as in
C          Subroutine F.
C               In the full matrix case (JT = 1), ML and MU are
C          ignored, and the Jacobian is to be loaded into PD in
C          columnwise manner, with df(i)/dy(j) loaded into PD(i,j).
C               In the band matrix case (JT = 4), the elements
C          within the band are to be loaded into PD in columnwise
C          manner, with diagonal lines of df/dy loaded into the rows
C          of PD.  Thus df(i)/dy(j) is to be loaded into PD(i-j+MU+1,j).
C          ML and MU are the half-bandwidth parameters (see IWORK).
C          The locations in PD in the two triangular areas which
C          correspond to nonexistent matrix elements can be ignored
C          or loaded arbitrarily, as they are overwritten by DLSODA.
C               JAC need not provide df/dy exactly.  A crude
C          approximation (possibly with a smaller bandwidth) will do.
C               In either case, PD is preset to zero by the solver,
C          so that only the nonzero elements need be loaded by JAC.
C          Each call to JAC is preceded by a call to F with the same
C          arguments NEQ, T, and Y.  Thus to gain some efficiency,
C          intermediate quantities shared by both calculations may be
C          saved in a user Common block by F and not recomputed by JAC,
C          if desired.  Also, JAC may alter the Y array, if desired.
C          JAC must be declared External in the calling program.
C               Subroutine JAC may access user-defined quantities in
C          NEQ(2),... and/or in Y(NEQ(1)+1),... if NEQ is an array
C          (dimensioned in JAC) and/or Y has length exceeding NEQ(1).
C          See the descriptions of NEQ and Y above.
C
C JT     = Jacobian type indicator.  Used only for input.
C          JT specifies how the Jacobian matrix df/dy will be
C          treated, if and when DLSODA requires this matrix.
C          JT has the following values and meanings:
C           1 means a user-supplied full (NEQ by NEQ) Jacobian.
C           2 means an internally generated (difference quotient) full
C             Jacobian (using NEQ extra calls to F per df/dy value).
C           4 means a user-supplied banded Jacobian.
C           5 means an internally generated banded Jacobian (using
C             ML+MU+1 extra calls to F per df/dy evaluation).
C          If JT = 1 or 4, the user must supply a Subroutine JAC
C          (the name is arbitrary) as described above under JAC.
C          If JT = 2 or 5, a dummy argument can be used.
C-----------------------------------------------------------------------
C Optional Inputs.
C
C The following is a list of the optional inputs provided for in the
C call sequence.  (See also Part 2.)  For each such input variable,
C this table lists its name as used in this documentation, its
C location in the call sequence, its meaning, and the default value.
C The use of any of these inputs requires IOPT = 1, and in that
C case all of these inputs are examined.  A value of zero for any
C of these optional inputs will cause the default value to be used.
C Thus to use a subset of the optional inputs, simply preload
C locations 5 to 10 in RWORK and IWORK to 0.0 and 0 respectively, and
C then set those of interest to nonzero values.
C
C Name    Location      Meaning and Default Value
C
C H0      RWORK(5)  the step size to be attempted on the first step.
C                   The default value is determined by the solver.
C
C HMAX    RWORK(6)  the maximum absolute step size allowed.
C                   The default value is infinite.
C
C HMIN    RWORK(7)  the minimum absolute step size allowed.
C                   The default value is 0.  (This lower bound is not
C                   enforced on the final step before reaching TCRIT
C                   when ITASK = 4 or 5.)
C
C IXPR    IWORK(5)  flag to generate extra printing at method switches.
C                   IXPR = 0 means no extra printing (the default).
C                   IXPR = 1 means print data on each switch.
C                   T, H, and NST will be printed on the same logical
C                   unit as used for error messages.
C
C MXSTEP  IWORK(6)  maximum number of (internally defined) steps
C                   allowed during one call to the solver.
C                   The default value is 500.
C
C MXHNIL  IWORK(7)  maximum number of messages printed (per problem)
C                   warning that T + H = T on a step (H = step size).
C                   This must be positive to result in a non-default
C                   value.  The default value is 10.
C
C MXORDN  IWORK(8)  the maximum order to be allowed for the nonstiff
C                   (Adams) method.  the default value is 12.
C                   if MXORDN exceeds the default value, it will
C                   be reduced to the default value.
C                   MXORDN is held constant during the problem.
C
C MXORDS  IWORK(9)  the maximum order to be allowed for the stiff
C                   (BDF) method.  The default value is 5.
C                   If MXORDS exceeds the default value, it will
C                   be reduced to the default value.
C                   MXORDS is held constant during the problem.
C-----------------------------------------------------------------------
C Optional Outputs.
C
C As optional additional output from DLSODA, the variables listed
C below are quantities related to the performance of DLSODA
C which are available to the user.  These are communicated by way of
C the work arrays, but also have internal mnemonic names as shown.
C except where stated otherwise, all of these outputs are defined
C on any successful return from DLSODA, and on any return with
C ISTATE = -1, -2, -4, -5, or -6.  On an illegal input return
C (ISTATE = -3), they will be unchanged from their existing values
C (if any), except possibly for TOLSF, LENRW, and LENIW.
C On any error return, outputs relevant to the error will be defined,
C as noted below.
C
C Name    Location      Meaning
C
C HU      RWORK(11) the step size in t last used (successfully).
C
C HCUR    RWORK(12) the step size to be attempted on the next step.
C
C TCUR    RWORK(13) the current value of the independent variable
C                   which the solver has actually reached, i.e. the
C                   current internal mesh point in t.  On output, TCUR
C                   will always be at least as far as the argument
C                   T, but may be farther (if interpolation was done).
C
C TOLSF   RWORK(14) a tolerance scale factor, greater than 1.0,
C                   computed when a request for too much accuracy was
C                   detected (ISTATE = -3 if detected at the start of
C                   the problem, ISTATE = -2 otherwise).  If ITOL is
C                   left unaltered but RTOL and ATOL are uniformly
C                   scaled up by a factor of TOLSF for the next call,
C                   then the solver is deemed likely to succeed.
C                   (The user may also ignore TOLSF and alter the
C                   tolerance parameters in any other way appropriate.)
C
C TSW     RWORK(15) the value of t at the time of the last method
C                   switch, if any.
C
C NST     IWORK(11) the number of steps taken for the problem so far.
C
C NFE     IWORK(12) the number of f evaluations for the problem so far.
C
C NJE     IWORK(13) the number of Jacobian evaluations (and of matrix
C                   LU decompositions) for the problem so far.
C
C NQU     IWORK(14) the method order last used (successfully).
C
C NQCUR   IWORK(15) the order to be attempted on the next step.
C
C IMXER   IWORK(16) the index of the component of largest magnitude in
C                   the weighted local error vector ( E(i)/EWT(i) ),
C                   on an error return with ISTATE = -4 or -5.
C
C LENRW   IWORK(17) the length of RWORK actually required, assuming
C                   that the length of RWORK is to be fixed for the
C                   rest of the problem, and that switching may occur.
C                   This is defined on normal returns and on an illegal
C                   input return for insufficient storage.
C
C LENIW   IWORK(18) the length of IWORK actually required, assuming
C                   that the length of IWORK is to be fixed for the
C                   rest of the problem, and that switching may occur.
C                   This is defined on normal returns and on an illegal
C                   input return for insufficient storage.
C
C MUSED   IWORK(19) the method indicator for the last successful step:
C                   1 means Adams (nonstiff), 2 means BDF (stiff).
C
C MCUR    IWORK(20) the current method indicator:
C                   1 means Adams (nonstiff), 2 means BDF (stiff).
C                   This is the method to be attempted
C                   on the next step.  Thus it differs from MUSED
C                   only if a method switch has just been made.
C
C The following two arrays are segments of the RWORK array which
C may also be of interest to the user as optional outputs.
C For each array, the table below gives its internal name,
C its base address in RWORK, and its description.
C
C Name    Base Address      Description
C
C YH      21             the Nordsieck history array, of size NYH by
C                        (NQCUR + 1), where NYH is the initial value
C                        of NEQ.  For j = 0,1,...,NQCUR, column j+1
C                        of YH contains HCUR**j/factorial(j) times
C                        the j-th derivative of the interpolating
C                        polynomial currently representing the solution,
C                        evaluated at T = TCUR.
C
C ACOR     LACOR         array of size NEQ used for the accumulated
C         (from Common   corrections on each step, scaled on output
C           as noted)    to represent the estimated local error in y
C                        on the last step.  This is the vector E in
C                        the description of the error control.  It is
C                        defined only on a successful return from
C                        DLSODA.  The base address LACOR is obtained by
C                        including in the user's program the
C                        following 2 lines:
C                           COMMON /DLS001/ RLS(218), ILS(37)
C                           LACOR = ILS(22)
C
C-----------------------------------------------------------------------
C Part 2.  Other Routines Callable.
C
C The following are optional calls which the user may make to
C gain additional capabilities in conjunction with DLSODA.
C (The routines XSETUN and XSETF are designed to conform to the
C SLATEC error handling package.)
C
C     Form of Call                  Function
C   CALL XSETUN(LUN)          set the logical unit number, LUN, for
C                             output of messages from DLSODA, if
C                             the default is not desired.
C                             The default value of LUN is 6.
C
C   CALL XSETF(MFLAG)         set a flag to control the printing of
C                             messages by DLSODA.
C                             MFLAG = 0 means do not print. (Danger:
C                             This risks losing valuable information.)
C                             MFLAG = 1 means print (the default).
C
C                             Either of the above calls may be made at
C                             any time and will take effect immediately.
C
C   CALL DSRCMA(RSAV,ISAV,JOB) saves and restores the contents of
C                             the internal Common blocks used by
C                             DLSODA (see Part 3 below).
C                             RSAV must be a real array of length 240
C                             or more, and ISAV must be an integer
C                             array of length 46 or more.
C                             JOB=1 means save Common into RSAV/ISAV.
C                             JOB=2 means restore Common from RSAV/ISAV.
C                                DSRCMA is useful if one is
C                             interrupting a run and restarting
C                             later, or alternating between two or
C                             more problems solved with DLSODA.
C
C   CALL DINTDY(,,,,,)        provide derivatives of y, of various
C        (see below)          orders, at a specified point t, if
C                             desired.  It may be called only after
C                             a successful return from DLSODA.
C
C The detailed instructions for using DINTDY are as follows.
C The form of the call is:
C
C   CALL DINTDY (T, K, RWORK(21), NYH, DKY, IFLAG)
C
C The input parameters are:
C
C T         = value of independent variable where answers are desired
C             (normally the same as the T last returned by DLSODA).
C             For valid results, T must lie between TCUR - HU and TCUR.
C             (See optional outputs for TCUR and HU.)
C K         = integer order of the derivative desired.  K must satisfy
C             0 .le. K .le. NQCUR, where NQCUR is the current order
C             (see optional outputs).  The capability corresponding
C             to K = 0, i.e. computing y(T), is already provided
C             by DLSODA directly.  Since NQCUR .ge. 1, the first
C             derivative dy/dt is always available with DINTDY.
C RWORK(21) = the base address of the history array YH.
C NYH       = column length of YH, equal to the initial value of NEQ.
C
C The output parameters are:
C
C DKY       = a real array of length NEQ containing the computed value
C             of the K-th derivative of y(t).
C IFLAG     = integer flag, returned as 0 if K and T were legal,
C             -1 if K was illegal, and -2 if T was illegal.
C             On an error return, a message is also written.
C-----------------------------------------------------------------------
C Part 3.  Common Blocks.
C
C If DLSODA is to be used in an overlay situation, the user
C must declare, in the primary overlay, the variables in:
C   (1) the call sequence to DLSODA, and
C   (2) the two internal Common blocks
C         /DLS001/  of length  255  (218 double precision words
C                      followed by 37 integer words),
C         /DLSA01/  of length  31    (22 double precision words
C                      followed by  9 integer words).
C
C If DLSODA is used on a system in which the contents of internal
C Common blocks are not preserved between calls, the user should
C declare the above Common blocks in the calling program to insure
C that their contents are preserved.
C
C If the solution of a given problem by DLSODA is to be interrupted
C and then later continued, such as when restarting an interrupted run
C or alternating between two or more problems, the user should save,
C following the return from the last DLSODA call prior to the
C interruption, the contents of the call sequence variables and the
C internal Common blocks, and later restore these values before the
C next DLSODA call for that problem.  To save and restore the Common
C blocks, use Subroutine DSRCMA (see Part 2 above).
C
C-----------------------------------------------------------------------
C Part 4.  Optionally Replaceable Solver Routines.
C
C Below is a description of a routine in the DLSODA package which
C relates to the measurement of errors, and can be
C replaced by a user-supplied version, if desired.  However, since such
C a replacement may have a major impact on performance, it should be
C done only when absolutely necessary, and only with great caution.
C (Note: The means by which the package version of a routine is
C superseded by the user's version may be system-dependent.)
C
C (a) DEWSET.
C The following subroutine is called just before each internal
C integration step, and sets the array of error weights, EWT, as
C described under ITOL/RTOL/ATOL above:
C     Subroutine DEWSET (NEQ, ITOL, RTOL, ATOL, YCUR, EWT)
C where NEQ, ITOL, RTOL, and ATOL are as in the DLSODA call sequence,
C YCUR contains the current dependent variable vector, and
C EWT is the array of weights set by DEWSET.
C
C If the user supplies this subroutine, it must return in EWT(i)
C (i = 1,...,NEQ) a positive quantity suitable for comparing errors
C in y(i) to.  The EWT array returned by DEWSET is passed to the
C DMNORM routine, and also used by DLSODA in the computation
C of the optional output IMXER, and the increments for difference
C quotient Jacobians.
C
C In the user-supplied version of DEWSET, it may be desirable to use
C the current values of derivatives of y.  Derivatives up to order NQ
C are available from the history array YH, described above under
C optional outputs.  In DEWSET, YH is identical to the YCUR array,
C extended to NQ + 1 columns with a column length of NYH and scale
C factors of H**j/factorial(j).  On the first call for the problem,
C given by NST = 0, NQ is 1 and H is temporarily set to 1.0.
C NYH is the initial value of NEQ.  The quantities NQ, H, and NST
C can be obtained by including in DEWSET the statements:
C     DOUBLE PRECISION RLS
C     COMMON /DLS001/ RLS(218),ILS(37)
C     NQ = ILS(33)
C     NST = ILS(34)
C     H = RLS(212)
C Thus, for example, the current value of dy/dt can be obtained as
C YCUR(NYH+i)/H  (i=1,...,NEQ)  (and the division by H is
C unnecessary when NST = 0).
C-----------------------------------------------------------------------
C
C***REVISION HISTORY  (YYYYMMDD)
C 19811102  DATE WRITTEN
C 19820126  Fixed bug in tests of work space lengths;
C           minor corrections in main prologue and comments.
C 19870330  Major update: corrected comments throughout;
C           removed TRET from Common; rewrote EWSET with 4 loops;
C           fixed t test in INTDY; added Cray directives in STODA;
C           in STODA, fixed DELP init. and logic around PJAC call;
C           combined routines to save/restore Common;
C           passed LEVEL = 0 in error message calls (except run abort).
C 19970225  Fixed lines setting JSTART = -2 in Subroutine LSODA.
C 20010425  Major update: convert source lines to upper case;
C           added *DECK lines; changed from 1 to * in dummy dimensions;
C           changed names R1MACH/D1MACH to RUMACH/DUMACH;
C           renamed routines for uniqueness across single/double prec.;
C           converted intrinsic names to generic form;
C           removed ILLIN and NTREP (data loaded) from Common;
C           removed all 'own' variables from Common;
C           changed error messages to quoted strings;
C           replaced XERRWV/XERRWD with 1993 revised version;
C           converted prologues, comments, error messages to mixed case;
C           numerous corrections to prologues and internal comments.
C 20010507  Converted single precision source to double precision.
C 20010613  Revised excess accuracy test (to match rest of ODEPACK).
C 20010808  Fixed bug in DPRJA (matrix in DBNORM call).
C 20020502  Corrected declarations in descriptions of user routines.
C 20031105  Restored 'own' variables to Common blocks, to enable
C           interrupt/restart feature.
C 20031112  Added SAVE statements for data-loaded constants.
C
C-----------------------------------------------------------------------
C Other routines in the DLSODA package.
C
C In addition to Subroutine DLSODA, the DLSODA package includes the
C following subroutines and function routines:
C  DINTDY   computes an interpolated value of the y vector at t = TOUT.
C  DSTODA   is the core integrator, which does one step of the
C           integration and the associated error control.
C  DCFODE   sets all method coefficients and test constants.
C  DPRJA    computes and preprocesses the Jacobian matrix J = df/dy
C           and the Newton iteration matrix P = I - h*l0*J.
C  DSOLSY   manages solution of linear system in chord iteration.
C  DEWSET   sets the error weight vector EWT before each step.
C  DMNORM   computes the weighted max-norm of a vector.
C  DFNORM   computes the norm of a full matrix consistent with the
C           weighted max-norm on vectors.
C  DBNORM   computes the norm of a band matrix consistent with the
C           weighted max-norm on vectors.
C  DSRCMA   is a user-callable routine to save and restore
C           the contents of the internal Common blocks.
C  DGEFA and DGESL   are routines from LINPACK for solving full
C           systems of linear algebraic equations.
C  DGBFA and DGBSL   are routines from LINPACK for solving banded
C           linear systems.
C  DUMACH   computes the unit roundoff in a machine-independent manner.
C  XERRWD, XSETUN, XSETF, IXSAV, and IUMACH  handle the printing of all
C           error messages and warnings.  XERRWD is machine-dependent.
C Note:  DMNORM, DFNORM, DBNORM, DUMACH, IXSAV, and IUMACH are
C function routines.  All the others are subroutines.
C
C-----------------------------------------------------------------------
      EXTERNAL DPRJA, DSOLSY
      DOUBLE PRECISION DUMACH, DMNORM
      INTEGER INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH,
     1   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     2   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     3   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     4   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      INTEGER INSUFR, INSUFI, IXPR, IOWNS2, JTYP, MUSED, MXORDN, MXORDS
      INTEGER I, I1, I2, IFLAG, IMXER, KGO, LF0,
     1   LENIW, LENRW, LENWM, ML, MORD, MU, MXHNL0, MXSTP0
      INTEGER LEN1, LEN1C, LEN1N, LEN1S, LEN2, LENIWC, LENRWC
      DOUBLE PRECISION CONIT, CRATE, EL, ELCO, HOLD, RMAX, TESCO,
     1   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND
      DOUBLE PRECISION TSW, ROWNS2, PDNORM
      DOUBLE PRECISION ATOLI, AYI, BIG, EWTI, H0, HMAX, HMX, RH, RTOLI,
     1   TCRIT, TDIST, TNEXT, TOL, TOLSF, TP, SIZE, SUM, W0
      DIMENSION MORD(2)
      LOGICAL IHIT
      CHARACTER (LEN=60) MSG
      SAVE MORD, MXSTP0, MXHNL0
C-----------------------------------------------------------------------
C The following two internal Common blocks contain
C (a) variables which are local to any subroutine but whose values must
C     be preserved between calls to the routine ("own" variables), and
C (b) variables which are communicated between subroutines.
C The block DLS001 is declared in subroutines DLSODA, DINTDY, DSTODA,
C DPRJA, and DSOLSY.
C The block DLSA01 is declared in subroutines DLSODA, DSTODA, and DPRJA.
C Groups of variables are replaced by dummy arrays in the Common
C declarations in routines where those variables are not used.
C-----------------------------------------------------------------------
      COMMON /DLS001/ CONIT, CRATE, EL(13), ELCO(13,12),
     1   HOLD, RMAX, TESCO(3,12),
     2   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND,
     3   INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH,
     4   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     5   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     6   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     7   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
C
      COMMON /DLSA01/ TSW, ROWNS2(20), PDNORM,
     1   INSUFR, INSUFI, IXPR, IOWNS2(2), JTYP, MUSED, MXORDN, MXORDS
C
      DATA MORD(1),MORD(2)/12,5/, MXSTP0/500/, MXHNL0/10/
C-----------------------------------------------------------------------
C Block A.
C This code block is executed on every call.
C It tests ISTATE and ITASK for legality and branches appropriately.
C If ISTATE .gt. 1 but the flag INIT shows that initialization has
C not yet been done, an error return occurs.
C If ISTATE = 1 and TOUT = T, return immediately.
C-----------------------------------------------------------------------
      IF (ISTATE .LT. 1 .OR. ISTATE .GT. 3) GO TO 601
      IF (ITASK .LT. 1 .OR. ITASK .GT. 5) GO TO 602
      IF (ISTATE .EQ. 1) GO TO 10
      IF (INIT .EQ. 0) GO TO 603
      IF (ISTATE .EQ. 2) GO TO 200
      GO TO 20
 10   INIT = 0
      IF (TOUT .EQ. T) RETURN
C-----------------------------------------------------------------------
C Block B.
C The next code block is executed for the initial call (ISTATE = 1),
C or for a continuation call with parameter changes (ISTATE = 3).
C It contains checking of all inputs and various initializations.
C
C First check legality of the non-optional inputs NEQ, ITOL, IOPT,
C JT, ML, and MU.
C-----------------------------------------------------------------------
 20   IF (NEQ(1) .LE. 0) GO TO 604
      IF (ISTATE .EQ. 1) GO TO 25
      IF (NEQ(1) .GT. N) GO TO 605
 25   N = NEQ(1)
      IF (ITOL .LT. 1 .OR. ITOL .GT. 4) GO TO 606
      IF (IOPT .LT. 0 .OR. IOPT .GT. 1) GO TO 607
      IF (JT .EQ. 3 .OR. JT .LT. 1 .OR. JT .GT. 5) GO TO 608
      JTYP = JT
      IF (JT .LE. 2) GO TO 30
      ML = IWORK(1)
      MU = IWORK(2)
      IF (ML .LT. 0 .OR. ML .GE. N) GO TO 609
      IF (MU .LT. 0 .OR. MU .GE. N) GO TO 610
 30   CONTINUE
C Next process and check the optional inputs. --------------------------
      IF (IOPT .EQ. 1) GO TO 40
      IXPR = 0
      MXSTEP = MXSTP0
      MXHNIL = MXHNL0
      HMXI = 0.0D0
      HMIN = 0.0D0
      IF (ISTATE .NE. 1) GO TO 60
      H0 = 0.0D0
      MXORDN = MORD(1)
      MXORDS = MORD(2)
      GO TO 60
 40   IXPR = IWORK(5)
      IF (IXPR .LT. 0 .OR. IXPR .GT. 1) GO TO 611
      MXSTEP = IWORK(6)
      IF (MXSTEP .LT. 0) GO TO 612
      IF (MXSTEP .EQ. 0) MXSTEP = MXSTP0
      MXHNIL = IWORK(7)
      IF (MXHNIL .LT. 0) GO TO 613
      IF (MXHNIL .EQ. 0) MXHNIL = MXHNL0
      IF (ISTATE .NE. 1) GO TO 50
      H0 = RWORK(5)
      MXORDN = IWORK(8)
      IF (MXORDN .LT. 0) GO TO 628
      IF (MXORDN .EQ. 0) MXORDN = 100
      MXORDN = MIN(MXORDN,MORD(1))
      MXORDS = IWORK(9)
      IF (MXORDS .LT. 0) GO TO 629
      IF (MXORDS .EQ. 0) MXORDS = 100
      MXORDS = MIN(MXORDS,MORD(2))
      IF ((TOUT - T)*H0 .LT. 0.0D0) GO TO 614
 50   HMAX = RWORK(6)
      IF (HMAX .LT. 0.0D0) GO TO 615
      HMXI = 0.0D0
      IF (HMAX .GT. 0.0D0) HMXI = 1.0D0/HMAX
      HMIN = RWORK(7)
      IF (HMIN .LT. 0.0D0) GO TO 616
C-----------------------------------------------------------------------
C Set work array pointers and check lengths LRW and LIW.
C If ISTATE = 1, METH is initialized to 1 here to facilitate the
C checking of work space lengths.
C Pointers to segments of RWORK and IWORK are named by prefixing L to
C the name of the segment.  E.g., the segment YH starts at RWORK(LYH).
C Segments of RWORK (in order) are denoted  YH, WM, EWT, SAVF, ACOR.
C If the lengths provided are insufficient for the current method,
C an error return occurs.  This is treated as illegal input on the
C first call, but as a problem interruption with ISTATE = -7 on a
C continuation call.  If the lengths are sufficient for the current
C method but not for both methods, a warning message is sent.
C-----------------------------------------------------------------------
 60   IF (ISTATE .EQ. 1) METH = 1
      IF (ISTATE .EQ. 1) NYH = N
      LYH = 21
      LEN1N = 20 + (MXORDN + 1)*NYH
      LEN1S = 20 + (MXORDS + 1)*NYH
      LWM = LEN1S + 1
      IF (JT .LE. 2) LENWM = N*N + 2
      IF (JT .GE. 4) LENWM = (2*ML + MU + 1)*N + 2
      LEN1S = LEN1S + LENWM
      LEN1C = LEN1N
      IF (METH .EQ. 2) LEN1C = LEN1S
      LEN1 = MAX(LEN1N,LEN1S)
      LEN2 = 3*N
      LENRW = LEN1 + LEN2
      LENRWC = LEN1C + LEN2
      IWORK(17) = LENRW
      LIWM = 1
      LENIW = 20 + N
      LENIWC = 20
      IF (METH .EQ. 2) LENIWC = LENIW
      IWORK(18) = LENIW
      IF (ISTATE .EQ. 1 .AND. LRW .LT. LENRWC) GO TO 617
      IF (ISTATE .EQ. 1 .AND. LIW .LT. LENIWC) GO TO 618
      IF (ISTATE .EQ. 3 .AND. LRW .LT. LENRWC) GO TO 550
      IF (ISTATE .EQ. 3 .AND. LIW .LT. LENIWC) GO TO 555
      LEWT = LEN1 + 1
      INSUFR = 0
      IF (LRW .GE. LENRW) GO TO 65
      INSUFR = 2
      LEWT = LEN1C + 1
c     MSG='DLSODA-  Warning.. RWORK length is sufficient for now, but  '
c     CALL XERRWD (MSG, 60, 103, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG='      may not be later.  Integration will proceed anyway.   '
c     CALL XERRWD (MSG, 60, 103, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG = '      Length needed is LENRW = I1, while LRW = I2.'
c     CALL XERRWD (MSG, 50, 103, 0, 2, LENRW, LRW, 0, 0.0D0, 0.0D0)
      CALL XERRWD (103, 0)
 65   LSAVF = LEWT + N
      LACOR = LSAVF + N
      INSUFI = 0
      IF (LIW .GE. LENIW) GO TO 70
      INSUFI = 2
c     MSG='DLSODA-  Warning.. IWORK length is sufficient for now, but  '
c     CALL XERRWD (MSG, 60, 104, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG='      may not be later.  Integration will proceed anyway.   '
c     CALL XERRWD (MSG, 60, 104, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG = '      Length needed is LENIW = I1, while LIW = I2.'
c     CALL XERRWD (MSG, 50, 104, 0, 2, LENIW, LIW, 0, 0.0D0, 0.0D0)
      CALL XERRWD (104, 0)
 70   CONTINUE
C Check RTOL and ATOL for legality. ------------------------------------
      RTOLI = RTOL(1)
      ATOLI = ATOL(1)
      DO 75 I = 1,N
        IF (ITOL .GE. 3) RTOLI = RTOL(I)
        IF (ITOL .EQ. 2 .OR. ITOL .EQ. 4) ATOLI = ATOL(I)
        IF (RTOLI .LT. 0.0D0) GO TO 619
        IF (ATOLI .LT. 0.0D0) GO TO 620
 75     CONTINUE
      IF (ISTATE .EQ. 1) GO TO 100
C If ISTATE = 3, set flag to signal parameter changes to DSTODA. -------
      JSTART = -1
      IF (N .EQ. NYH) GO TO 200
C NEQ was reduced.  Zero part of YH to avoid undefined references. -----
      I1 = LYH + L*NYH
      I2 = LYH + (MAXORD + 1)*NYH - 1
      IF (I1 .GT. I2) GO TO 200
      DO 95 I = I1,I2
C95     RWORK(I) = 0.0D0
        RWORK(I) = 0.0D0
 95     CONTINUE
      GO TO 200
C-----------------------------------------------------------------------
C Block C.
C The next block is for the initial call only (ISTATE = 1).
C It contains all remaining initializations, the initial call to F,
C and the calculation of the initial step size.
C The error weights in EWT are inverted after being loaded.
C-----------------------------------------------------------------------
 100  UROUND = DUMACH()
      TN = T
      TSW = T
      MAXORD = MXORDN
      IF (ITASK .NE. 4 .AND. ITASK .NE. 5) GO TO 110
      TCRIT = RWORK(1)
      IF ((TCRIT - TOUT)*(TOUT - T) .LT. 0.0D0) GO TO 625
      IF (H0 .NE. 0.0D0 .AND. (T + H0 - TCRIT)*H0 .GT. 0.0D0)
     1   H0 = TCRIT - T
 110  JSTART = 0
      NHNIL = 0
      NST = 0
      NJE = 0
      NSLAST = 0
      HU = 0.0D0
      NQU = 0
      MUSED = 0
      MITER = 0
      CCMAX = 0.3D0
      MAXCOR = 3
      MSBP = 20
      MXNCF = 10
C Initial call to F.  (LF0 points to YH(*,2).) -------------------------
      LF0 = LYH + NYH
      CALL F (NEQ, T, Y, RWORK(LF0))
      NFE = 1
C Load the initial value vector in YH. ---------------------------------
      DO 115 I = 1,N
C115    RWORK(I+LYH-1) = Y(I)
        RWORK(I+LYH-1) = Y(I)
 115    CONTINUE
C Load and invert the EWT array.  (H is temporarily set to 1.0.) -------
      NQ = 1
      H = 1.0D0
      CALL DEWSET (N, ITOL, RTOL, ATOL, RWORK(LYH), RWORK(LEWT))
      DO 120 I = 1,N
        IF (RWORK(I+LEWT-1) .LE. 0.0D0) GO TO 621
C120    RWORK(I+LEWT-1) = 1.0D0/RWORK(I+LEWT-1)
        RWORK(I+LEWT-1) = 1.0D0/RWORK(I+LEWT-1)
 120    CONTINUE
C-----------------------------------------------------------------------
C The coding below computes the step size, H0, to be attempted on the
C first step, unless the user has supplied a value for this.
C First check that TOUT - T differs significantly from zero.
C A scalar tolerance quantity TOL is computed, as MAX(RTOL(i))
C if this is positive, or MAX(ATOL(i)/ABS(Y(i))) otherwise, adjusted
C so as to be between 100*UROUND and 1.0E-3.
C Then the computed value H0 is given by:
C
C   H0**(-2)  =  1./(TOL * w0**2)  +  TOL * (norm(F))**2
C
C where   w0     = MAX ( ABS(T), ABS(TOUT) ),
C         F      = the initial value of the vector f(t,y), and
C         norm() = the weighted vector norm used throughout, given by
C                  the DMNORM function routine, and weighted by the
C                  tolerances initially loaded into the EWT array.
C The sign of H0 is inferred from the initial values of TOUT and T.
C ABS(H0) is made .le. ABS(TOUT-T) in any case.
C-----------------------------------------------------------------------
      IF (H0 .NE. 0.0D0) GO TO 180
      TDIST = ABS(TOUT - T)
      W0 = MAX(ABS(T),ABS(TOUT))
      IF (TDIST .LT. 2.0D0*UROUND*W0) GO TO 622
      TOL = RTOL(1)
      IF (ITOL .LE. 2) GO TO 140
      DO 130 I = 1,N
C130    TOL = MAX(TOL,RTOL(I))
        TOL = MAX(TOL,RTOL(I))
 130    CONTINUE
 140  IF (TOL .GT. 0.0D0) GO TO 160
      ATOLI = ATOL(1)
      DO 150 I = 1,N
        IF (ITOL .EQ. 2 .OR. ITOL .EQ. 4) ATOLI = ATOL(I)
        AYI = ABS(Y(I))
        IF (AYI .NE. 0.0D0) TOL = MAX(TOL,ATOLI/AYI)
 150    CONTINUE
 160  TOL = MAX(TOL,100.0D0*UROUND)
      TOL = MIN(TOL,0.001D0)
      SUM = DMNORM (N, RWORK(LF0), RWORK(LEWT))
      SUM = 1.0D0/(TOL*W0*W0) + TOL*SUM**2
      H0 = 1.0D0/SQRT(SUM)
      H0 = MIN(H0,TDIST)
      H0 = SIGN(H0,TOUT-T)
C Adjust H0 if necessary to meet HMAX bound. ---------------------------
 180  RH = ABS(H0)*HMXI
      IF (RH .GT. 1.0D0) H0 = H0/RH
C Load H with H0 and scale YH(*,2) by H0. ------------------------------
      H = H0
      DO 190 I = 1,N
C190    RWORK(I+LF0-1) = H0*RWORK(I+LF0-1)
        RWORK(I+LF0-1) = H0*RWORK(I+LF0-1)
 190    CONTINUE
      GO TO 270
C-----------------------------------------------------------------------
C Block D.
C The next code block is for continuation calls only (ISTATE = 2 or 3)
C and is to check stop conditions before taking a step.
C-----------------------------------------------------------------------
 200  NSLAST = NST
C     GO TO (210, 250, 220, 230, 240), ITASK
      IF (ITASK .EQ. 1) GO TO 210
      IF (ITASK .EQ. 2) GO TO 250
      IF (ITASK .EQ. 3) GO TO 220
      IF (ITASK .EQ. 4) GO TO 230
      IF (ITASK .EQ. 5) GO TO 240

 210  IF ((TN - TOUT)*H .LT. 0.0D0) GO TO 250
      CALL DINTDY (TOUT, 0, RWORK(LYH), NYH, Y, IFLAG)
      IF (IFLAG .NE. 0) GO TO 627
      T = TOUT
      GO TO 420
 220  TP = TN - HU*(1.0D0 + 100.0D0*UROUND)
      IF ((TP - TOUT)*H .GT. 0.0D0) GO TO 623
      IF ((TN - TOUT)*H .LT. 0.0D0) GO TO 250
      T = TN
      GO TO 400
 230  TCRIT = RWORK(1)
      IF ((TN - TCRIT)*H .GT. 0.0D0) GO TO 624
      IF ((TCRIT - TOUT)*H .LT. 0.0D0) GO TO 625
      IF ((TN - TOUT)*H .LT. 0.0D0) GO TO 245
      CALL DINTDY (TOUT, 0, RWORK(LYH), NYH, Y, IFLAG)
      IF (IFLAG .NE. 0) GO TO 627
      T = TOUT
      GO TO 420
 240  TCRIT = RWORK(1)
      IF ((TN - TCRIT)*H .GT. 0.0D0) GO TO 624
 245  HMX = ABS(TN) + ABS(H)
      IHIT = ABS(TN - TCRIT) .LE. 100.0D0*UROUND*HMX
      IF (IHIT) T = TCRIT
      IF (IHIT) GO TO 400
      TNEXT = TN + H*(1.0D0 + 4.0D0*UROUND)
      IF ((TNEXT - TCRIT)*H .LE. 0.0D0) GO TO 250
      H = (TCRIT - TN)*(1.0D0 - 4.0D0*UROUND)
      IF (ISTATE .EQ. 2 .AND. JSTART .GE. 0) JSTART = -2
C-----------------------------------------------------------------------
C Block E.
C The next block is normally executed for all calls and contains
C the call to the one-step core integrator DSTODA.
C
C This is a looping point for the integration steps.
C
C First check for too many steps being taken, update EWT (if not at
C start of problem), check for too much accuracy being requested, and
C check for H below the roundoff level in T.
C-----------------------------------------------------------------------
 250  CONTINUE
      IF (METH .EQ. MUSED) GO TO 255
      IF (INSUFR .EQ. 1) GO TO 550
      IF (INSUFI .EQ. 1) GO TO 555
 255  IF ((NST-NSLAST) .GE. MXSTEP) GO TO 500
      CALL DEWSET (N, ITOL, RTOL, ATOL, RWORK(LYH), RWORK(LEWT))
      DO 260 I = 1,N
        IF (RWORK(I+LEWT-1) .LE. 0.0D0) GO TO 510
C260    RWORK(I+LEWT-1) = 1.0D0/RWORK(I+LEWT-1)
        RWORK(I+LEWT-1) = 1.0D0/RWORK(I+LEWT-1)
 260    CONTINUE
 270  TOLSF = UROUND*DMNORM (N, RWORK(LYH), RWORK(LEWT))
      IF (TOLSF .LE. 1.0D0) GO TO 280
      TOLSF = TOLSF*2.0D0
      IF (NST .EQ. 0) GO TO 626
      GO TO 520
 280  IF ((TN + H) .NE. TN) GO TO 290
      NHNIL = NHNIL + 1
      IF (NHNIL .GT. MXHNIL) GO TO 290
c     MSG = 'DLSODA-  Warning..Internal T (=R1) and H (=R2) are'
c     CALL XERRWD (MSG, 60, 101, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG='      such that in the machine, T + H = T on the next step  '
c     CALL XERRWD (MSG, 60, 101, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG = '     (H = step size). Solver will continue anyway.'
c     CALL XERRWD (MSG, 50, 101, 0, 0, 0, 0, 2, TN, H)
      CALL XERRWD (101, 0)
      IF (NHNIL .LT. MXHNIL) GO TO 290
c     MSG = 'DLSODA-  Above warning has been issued I1 times.  '
c     CALL XERRWD (MSG, 50, 102, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG = '     It will not be issued again for this problem.'
c     CALL XERRWD (MSG, 50, 102, 0, 1, MXHNIL, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (102, 0)
 290  CONTINUE
C-----------------------------------------------------------------------
C   CALL DSTODA(NEQ,Y,YH,NYH,YH,EWT,SAVF,ACOR,WM,IWM,F,JAC,DPRJA,DSOLSY)
C-----------------------------------------------------------------------
      CALL DSTODA (NEQ, Y, RWORK(LYH), NYH, RWORK(LYH), RWORK(LEWT),
     1   RWORK(LSAVF), RWORK(LACOR), RWORK(LWM), IWORK(LIWM),
     2   F, JAC, DPRJA, DSOLSY)
      KGO = 1 - KFLAG
C     GO TO (300, 530, 540), KGO
      IF (KGO .EQ. 1) GO TO 300
      IF (KGO .EQ. 2) GO TO 530
      IF (KGO .EQ. 3) GO TO 540

C-----------------------------------------------------------------------
C Block F.
C The following block handles the case of a successful return from the
C core integrator (KFLAG = 0).
C If a method switch was just made, record TSW, reset MAXORD,
C set JSTART to -1 to signal DSTODA to complete the switch,
C and do extra printing of data if IXPR = 1.
C Then, in any case, check for stop conditions.
C-----------------------------------------------------------------------
 300  INIT = 1
      IF (METH .EQ. MUSED) GO TO 310
      TSW = TN
      MAXORD = MXORDN
      IF (METH .EQ. 2) MAXORD = MXORDS
      IF (METH .EQ. 2) RWORK(LWM) = SQRT(UROUND)
      INSUFR = MIN(INSUFR,1)
      INSUFI = MIN(INSUFI,1)
      JSTART = -1
      IF (IXPR .EQ. 0) GO TO 310
      IF (METH .EQ. 2) THEN
c     MSG='DLSODA- A switch to the BDF (stiff) method has occurred     '
c     CALL XERRWD (MSG, 60, 105, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (105, 0)
      ENDIF
      IF (METH .EQ. 1) THEN
c     MSG='DLSODA- A switch to the Adams (nonstiff) method has occurred'
c     CALL XERRWD (MSG, 60, 106, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (106, 0)
      ENDIF
c     MSG='     at T = R1,  tentative step size H = R2,  step NST = I1 '
c     CALL XERRWD (MSG, 60, 107, 0, 1, NST, 0, 2, TN, H)
      CALL XERRWD (107, 0)
C310  GO TO (320, 400, 330, 340, 350), ITASK
 310  IF (ITASK .EQ. 1) GO TO 320
      IF (ITASK .EQ. 2) GO TO 400
      IF (ITASK .EQ. 3) GO TO 330
      IF (ITASK .EQ. 4) GO TO 340
      IF (ITASK .EQ. 5) GO TO 350

C ITASK = 1.  If TOUT has been reached, interpolate. -------------------
 320  IF ((TN - TOUT)*H .LT. 0.0D0) GO TO 250
      CALL DINTDY (TOUT, 0, RWORK(LYH), NYH, Y, IFLAG)
      T = TOUT
      GO TO 420
C ITASK = 3.  Jump to exit if TOUT was reached. ------------------------
 330  IF ((TN - TOUT)*H .GE. 0.0D0) GO TO 400
      GO TO 250
C ITASK = 4.  See if TOUT or TCRIT was reached.  Adjust H if necessary.
 340  IF ((TN - TOUT)*H .LT. 0.0D0) GO TO 345
      CALL DINTDY (TOUT, 0, RWORK(LYH), NYH, Y, IFLAG)
      T = TOUT
      GO TO 420
 345  HMX = ABS(TN) + ABS(H)
      IHIT = ABS(TN - TCRIT) .LE. 100.0D0*UROUND*HMX
      IF (IHIT) GO TO 400
      TNEXT = TN + H*(1.0D0 + 4.0D0*UROUND)
      IF ((TNEXT - TCRIT)*H .LE. 0.0D0) GO TO 250
      H = (TCRIT - TN)*(1.0D0 - 4.0D0*UROUND)
      IF (JSTART .GE. 0) JSTART = -2
      GO TO 250
C ITASK = 5.  See if TCRIT was reached and jump to exit. ---------------
 350  HMX = ABS(TN) + ABS(H)
      IHIT = ABS(TN - TCRIT) .LE. 100.0D0*UROUND*HMX
C-----------------------------------------------------------------------
C Block G.
C The following block handles all successful returns from DLSODA.
C If ITASK .ne. 1, Y is loaded from YH and T is set accordingly.
C ISTATE is set to 2, and the optional outputs are loaded into the
C work arrays before returning.
C-----------------------------------------------------------------------
 400  DO 410 I = 1,N
C410    Y(I) = RWORK(I+LYH-1)
        Y(I) = RWORK(I+LYH-1)
 410    CONTINUE
      T = TN
      IF (ITASK .NE. 4 .AND. ITASK .NE. 5) GO TO 420
      IF (IHIT) T = TCRIT
 420  ISTATE = 2
      RWORK(11) = HU
      RWORK(12) = H
      RWORK(13) = TN
      RWORK(15) = TSW
      IWORK(11) = NST
      IWORK(12) = NFE
      IWORK(13) = NJE
      IWORK(14) = NQU
      IWORK(15) = NQ
      IWORK(19) = MUSED
      IWORK(20) = METH
      RETURN
C-----------------------------------------------------------------------
C Block H.
C The following block handles all unsuccessful returns other than
C those for illegal input.  First the error message routine is called.
C If there was an error test or convergence test failure, IMXER is set.
C Then Y is loaded from YH and T is set to TN.
C The optional outputs are loaded into the work arrays before returning.
C-----------------------------------------------------------------------
C The maximum number of steps was taken before reaching TOUT. ----------
 500  MSG = 'DLSODA-  At current T (=R1), MXSTEP (=I1) steps   '
c     CALL XERRWD (MSG, 50, 201, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG = '      taken on this call before reaching TOUT     '
c     CALL XERRWD (MSG, 50, 201, 0, 1, MXSTEP, 0, 1, TN, 0.0D0)
      CALL XERRWD (201, 0)
      ISTATE = -1
      GO TO 580
C EWT(i) .le. 0.0 for some i (not at start of problem). ----------------
 510  EWTI = RWORK(LEWT+I-1)
c     MSG = 'DLSODA-  At T (=R1), EWT(I1) has become R2 .le. 0.'
c     CALL XERRWD (MSG, 50, 202, 0, 1, I, 0, 2, TN, EWTI)
      CALL XERRWD (202, 0)
      ISTATE = -6
      GO TO 580
C Too much accuracy requested for machine precision. -------------------
 520  MSG = 'DLSODA-  At T (=R1), too much accuracy requested  '
c     CALL XERRWD (MSG, 50, 203, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG = '      for precision of machine..  See TOLSF (=R2) '
c     CALL XERRWD (MSG, 50, 203, 0, 0, 0, 0, 2, TN, TOLSF)
      CALL XERRWD (203, 0)
      RWORK(14) = TOLSF
      ISTATE = -2
      GO TO 580
C KFLAG = -1.  Error test failed repeatedly or with ABS(H) = HMIN. -----
 530  MSG = 'DLSODA-  At T(=R1) and step size H(=R2), the error'
c     CALL XERRWD (MSG, 50, 204, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG = '      test failed repeatedly or with ABS(H) = HMIN'
c     CALL XERRWD (MSG, 50, 204, 0, 0, 0, 0, 2, TN, H)
      CALL XERRWD (204, 0)
      ISTATE = -4
      GO TO 560
C KFLAG = -2.  Convergence failed repeatedly or with ABS(H) = HMIN. ----
 540  MSG = 'DLSODA-  At T (=R1) and step size H (=R2), the    '
c     CALL XERRWD (MSG, 50, 205, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG = '      corrector convergence failed repeatedly     '
c     CALL XERRWD (MSG, 50, 205, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG = '      or with ABS(H) = HMIN   '
c     CALL XERRWD (MSG, 30, 205, 0, 0, 0, 0, 2, TN, H)
      CALL XERRWD (205, 0)
      ISTATE = -5
      GO TO 560
C RWORK length too small to proceed. -----------------------------------
 550  MSG = 'DLSODA-  At current T(=R1), RWORK length too small'
c     CALL XERRWD (MSG, 50, 206, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG='      to proceed.  The integration was otherwise successful.'
c     CALL XERRWD (MSG, 60, 206, 0, 0, 0, 0, 1, TN, 0.0D0)
      CALL XERRWD (206, 0)
      ISTATE = -7
      GO TO 580
C IWORK length too small to proceed. -----------------------------------
 555  MSG = 'DLSODA-  At current T(=R1), IWORK length too small'
c     CALL XERRWD (MSG, 50, 207, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG='      to proceed.  The integration was otherwise successful.'
c     CALL XERRWD (MSG, 60, 207, 0, 0, 0, 0, 1, TN, 0.0D0)
      CALL XERRWD (207, 0)
      ISTATE = -7
      GO TO 580
C Compute IMXER if relevant. -------------------------------------------
 560  BIG = 0.0D0
      IMXER = 1
      DO 570 I = 1,N
        SIZE = ABS(RWORK(I+LACOR-1)*RWORK(I+LEWT-1))
        IF (BIG .GE. SIZE) GO TO 570
        BIG = SIZE
        IMXER = I
 570    CONTINUE
      IWORK(16) = IMXER
C Set Y vector, T, and optional outputs. -------------------------------
 580  DO 590 I = 1,N
C590    Y(I) = RWORK(I+LYH-1)
        Y(I) = RWORK(I+LYH-1)
 590    CONTINUE
      T = TN
      RWORK(11) = HU
      RWORK(12) = H
      RWORK(13) = TN
      RWORK(15) = TSW
      IWORK(11) = NST
      IWORK(12) = NFE
      IWORK(13) = NJE
      IWORK(14) = NQU
      IWORK(15) = NQ
      IWORK(19) = MUSED
      IWORK(20) = METH
      RETURN
C-----------------------------------------------------------------------
C Block I.
C The following block handles all error returns due to illegal input
C (ISTATE = -3), as detected before calling the core integrator.
C First the error message routine is called.  If the illegal input
C is a negative ISTATE, the run is aborted (apparent infinite loop).
C-----------------------------------------------------------------------
 601  MSG = 'DLSODA-  ISTATE (=I1) illegal.'
c     CALL XERRWD (MSG, 30, 1, 0, 1, ISTATE, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (1, 0)
      IF (ISTATE .LT. 0) GO TO 800
      GO TO 700
 602  MSG = 'DLSODA-  ITASK (=I1) illegal. '
c     CALL XERRWD (MSG, 30, 2, 0, 1, ITASK, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (2, 0)
      GO TO 700
 603  MSG = 'DLSODA-  ISTATE .gt. 1 but DLSODA not initialized.'
c     CALL XERRWD (MSG, 50, 3, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (3, 0)
      GO TO 700
 604  MSG = 'DLSODA-  NEQ (=I1) .lt. 1     '
c     CALL XERRWD (MSG, 30, 4, 0, 1, NEQ(1), 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (4, 0)
      GO TO 700
 605  MSG = 'DLSODA-  ISTATE = 3 and NEQ increased (I1 to I2). '
c     CALL XERRWD (MSG, 50, 5, 0, 2, N, NEQ(1), 0, 0.0D0, 0.0D0)
      CALL XERRWD (5, 0)
      GO TO 700
 606  MSG = 'DLSODA-  ITOL (=I1) illegal.  '
c     CALL XERRWD (MSG, 30, 6, 0, 1, ITOL, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (6, 0)
      GO TO 700
 607  MSG = 'DLSODA-  IOPT (=I1) illegal.  '
c     CALL XERRWD (MSG, 30, 7, 0, 1, IOPT, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (7, 0)
      GO TO 700
 608  MSG = 'DLSODA-  JT (=I1) illegal.    '
c     CALL XERRWD (MSG, 30, 8, 0, 1, JT, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (8, 0)
      GO TO 700
 609  MSG = 'DLSODA-  ML (=I1) illegal: .lt.0 or .ge.NEQ (=I2) '
c     CALL XERRWD (MSG, 50, 9, 0, 2, ML, NEQ(1), 0, 0.0D0, 0.0D0)
      CALL XERRWD (9, 0)
      GO TO 700
 610  MSG = 'DLSODA-  MU (=I1) illegal: .lt.0 or .ge.NEQ (=I2) '
c     CALL XERRWD (MSG, 50, 10, 0, 2, MU, NEQ(1), 0, 0.0D0, 0.0D0)
      CALL XERRWD (10, 0)
      GO TO 700
 611  MSG = 'DLSODA-  IXPR (=I1) illegal.  '
c     CALL XERRWD (MSG, 30, 11, 0, 1, IXPR, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (11, 0)
      GO TO 700
 612  MSG = 'DLSODA-  MXSTEP (=I1) .lt. 0  '
c     CALL XERRWD (MSG, 30, 12, 0, 1, MXSTEP, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (12, 0)
      GO TO 700
 613  MSG = 'DLSODA-  MXHNIL (=I1) .lt. 0  '
c     CALL XERRWD (MSG, 30, 13, 0, 1, MXHNIL, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (13, 0)
      GO TO 700
 614  MSG = 'DLSODA-  TOUT (=R1) behind T (=R2)      '
c     CALL XERRWD (MSG, 40, 14, 0, 0, 0, 0, 2, TOUT, T)
c     MSG = '      Integration direction is given by H0 (=R1)  '
c     CALL XERRWD (MSG, 50, 14, 0, 0, 0, 0, 1, H0, 0.0D0)
      CALL XERRWD (14, 0)
      GO TO 700
 615  MSG = 'DLSODA-  HMAX (=R1) .lt. 0.0  '
c     CALL XERRWD (MSG, 30, 15, 0, 0, 0, 0, 1, HMAX, 0.0D0)
      CALL XERRWD (15, 0)
      GO TO 700
 616  MSG = 'DLSODA-  HMIN (=R1) .lt. 0.0  '
c     CALL XERRWD (MSG, 30, 16, 0, 0, 0, 0, 1, HMIN, 0.0D0)
      CALL XERRWD (16, 0)
      GO TO 700
 617  MSG='DLSODA-  RWORK length needed, LENRW (=I1), exceeds LRW (=I2)'
c     CALL XERRWD (MSG, 60, 17, 0, 2, LENRW, LRW, 0, 0.0D0, 0.0D0)
      CALL XERRWD (17, 0)
      GO TO 700
 618  MSG='DLSODA-  IWORK length needed, LENIW (=I1), exceeds LIW (=I2)'
c     CALL XERRWD (MSG, 60, 18, 0, 2, LENIW, LIW, 0, 0.0D0, 0.0D0)
      CALL XERRWD (18, 0)
      GO TO 700
 619  MSG = 'DLSODA-  RTOL(I1) is R1 .lt. 0.0        '
c     CALL XERRWD (MSG, 40, 19, 0, 1, I, 0, 1, RTOLI, 0.0D0)
      CALL XERRWD (19, 0)
      GO TO 700
 620  MSG = 'DLSODA-  ATOL(I1) is R1 .lt. 0.0        '
c     CALL XERRWD (MSG, 40, 20, 0, 1, I, 0, 1, ATOLI, 0.0D0)
      CALL XERRWD (20, 0)
      GO TO 700
 621  EWTI = RWORK(LEWT+I-1)
c     MSG = 'DLSODA-  EWT(I1) is R1 .le. 0.0         '
c     CALL XERRWD (MSG, 40, 21, 0, 1, I, 0, 1, EWTI, 0.0D0)
      CALL XERRWD (21, 0)
      GO TO 700
 622  MSG='DLSODA-  TOUT(=R1) too close to T(=R2) to start integration.'
c     CALL XERRWD (MSG, 60, 22, 0, 0, 0, 0, 2, TOUT, T)
      CALL XERRWD (22, 0)
      GO TO 700
 623  MSG='DLSODA-  ITASK = I1 and TOUT (=R1) behind TCUR - HU (= R2)  '
c     CALL XERRWD (MSG, 60, 23, 0, 1, ITASK, 0, 2, TOUT, TP)
      CALL XERRWD (23, 0)
      GO TO 700
 624  MSG='DLSODA-  ITASK = 4 or 5 and TCRIT (=R1) behind TCUR (=R2)   '
c     CALL XERRWD (MSG, 60, 24, 0, 0, 0, 0, 2, TCRIT, TN)
      CALL XERRWD (24, 0)
      GO TO 700
 625  MSG='DLSODA-  ITASK = 4 or 5 and TCRIT (=R1) behind TOUT (=R2)   '
c     CALL XERRWD (MSG, 60, 25, 0, 0, 0, 0, 2, TCRIT, TOUT)
      CALL XERRWD (25, 0)
      GO TO 700
 626  MSG = 'DLSODA-  At start of problem, too much accuracy   '
c     CALL XERRWD (MSG, 50, 26, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
c     MSG='      requested for precision of machine..  See TOLSF (=R1) '
c     CALL XERRWD (MSG, 60, 26, 0, 0, 0, 0, 1, TOLSF, 0.0D0)
      CALL XERRWD (26, 0)
      RWORK(14) = TOLSF
      GO TO 700
 627  MSG = 'DLSODA-  Trouble in DINTDY.  ITASK = I1, TOUT = R1'
c     CALL XERRWD (MSG, 50, 27, 0, 1, ITASK, 0, 1, TOUT, 0.0D0)
      CALL XERRWD (27, 0)
      GO TO 700
 628  MSG = 'DLSODA-  MXORDN (=I1) .lt. 0  '
c     CALL XERRWD (MSG, 30, 28, 0, 1, MXORDN, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (28, 0)
      GO TO 700
 629  MSG = 'DLSODA-  MXORDS (=I1) .lt. 0  '
c     CALL XERRWD (MSG, 30, 29, 0, 1, MXORDS, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (29, 0)
C
 700  ISTATE = -3
      RETURN
C
 800  MSG = 'DLSODA-  Run aborted.. apparent infinite loop.    '
c     CALL XERRWD (MSG, 50, 303, 2, 0, 0, 0, 0, 0.0D0, 0.0D0)
      CALL XERRWD (303, 2)
      RETURN
C----------------------- End of Subroutine DLSODA ----------------------
      END

## Discrete adjoint sensitivity analysis (fixed-step explicit RK4).
##
## The discrete adjoint differentiates the exact RK4 step map and transposes it,
## so its gradient equals the discrete forward sensitivity of the same scheme to
## machine precision (cf. Zhang, Abhyankar, Constantinescu & Anitescu, "Discrete
## Adjoint Sensitivity Analysis of Hybrid Dynamical Systems").  The adjoint is
## reverse-mode differentiation through the four RK4 stages using the symbolic F,
## dF/dy and dF/dtheta from .rxJacobian().  Reference/validation implementation.

#' Build symbolic evaluators for the discrete adjoint (RHS, Jacobian, forcing)
#'
#' Parses the model once and returns closures that evaluate the ODE RHS `F`, the
#' state Jacobian `dF/dy` and the parameter forcing `dF/dtheta` at an arbitrary
#' state (needed at interior RK4 stage points).  Pair with
#' [.rxDiscreteForwardSens()] and [.rxDiscreteAdjointGrad()].
#'
#' @param object model definition accepted by [rxode2::rxode2()].
#' @param calcSens character vector of parameters to differentiate w.r.t.
#' @return a list with `st` (state names), `ns`, `np`, `calcSens`, and evaluator
#'   closures `Fe(x, params)`, `FXe(x, params)` (`ns x ns`), `FPe(x, params)`
#'   (`ns x np`).
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxDiscreteAdjointBuild <- function(object, calcSens) {
  .model <- rxode2::rxS(rxode2::rxGetModel(object), TRUE, promoteLinSens = FALSE)
  .st <- rxode2::rxStateOde(.model); .ns <- length(.st); .np <- length(calcSens)
  if (.ns == 0L) stop("discrete adjoint requires a model with ODE states", call. = FALSE)
  invisible(rxode2::.rxJacobian(.model, c(.st, calcSens)))
  ## bind each symengine object to a local before rxFromSE (NSE gotcha)
  .fromSE <- function(nm) { .d <- get0(nm, envir = .model, inherits = FALSE); rxode2::rxFromSE(.d) }
  .Fexpr <- vapply(.st, function(i) .fromSE(paste0("rx__d_dt_", i, "__")), character(1))
  .FXexpr <- matrix("", .ns, .ns); .FPexpr <- matrix("", .ns, .np)
  for (i in seq_len(.ns)) for (j in seq_len(.ns))
    .FXexpr[i, j] <- .fromSE(paste0("rx__df_", .st[i], "_dy_", .st[j], "__"))
  for (i in seq_len(.ns)) for (p in seq_len(.np))
    .FPexpr[i, p] <- .fromSE(paste0("rx__df_", .st[i], "_dy_", calcSens[p], "__"))
  ## bioavailability F(theta) per dose compartment and d/dtheta, for the
  ## additive-bolus dose jump X[c] += F*amt
  .fStr <- list(); .dFdpStr <- list()
  for (.c in .st) {
    .fSE <- get0(paste0("rx_f_", .c, "_"), envir = .model, inherits = FALSE)
    if (!is.null(.fSE)) {
      .fStr[[.c]] <- rxode2::rxFromSE(.fSE)
      .dFdpStr[[.c]] <- vapply(calcSens, function(p) {
        .d <- symengine::D(.fSE, p); rxode2::rxFromSE(.d) }, character(1))
    }
  }
  ## pre-parse every expression once and evaluate the parsed forms against a
  ## single reusable environment updated in place per call
  .parse1 <- function(s) parse(text = s)[[1]]
  .FexprP  <- lapply(.Fexpr, .parse1)
  .FXexprP <- lapply(as.vector(.FXexpr), .parse1)
  .FPexprP <- lapply(as.vector(.FPexpr), .parse1)
  .fStrP   <- lapply(.fStr, .parse1)
  .dFdpStrP <- lapply(.dFdpStr, function(v) lapply(v, .parse1))
  .env <- new.env(parent = baseenv())
  assign("Rx_pow_di", function(a, b) a^b, .env); assign("Rx_pow", function(a, b) a^b, .env)
  assign("expit", function(x, a = 0, b = 1) a + (b - a) / (1 + exp(-x)), .env)
  assign("logit", function(x, a = 0, b = 1) log((x - a) / (b - x)), .env)
  .setState <- function(x) for (i in seq_len(.ns)) assign(.st[i], x[i], envir = .env)
  .setParams <- function(params) { .nm <- names(params); for (i in seq_along(params)) assign(.nm[i], params[[i]], envir = .env) }
  .evList <- function(pl) vapply(pl, eval, numeric(1), envir = .env)
  list(st = .st, ns = .ns, np = .np, calcSens = calcSens,
       fStr = .fStr, dFdpStr = .dFdpStr,
       ## evaluate F(theta) for a dose cmt at parameters only (no state)
       evP = function(cmtStr, params) { .setState(numeric(.ns)); .setParams(params); eval(cmtStr, .env) },
       Fe  = function(x, p) { .setState(x); .setParams(p); .evList(.FexprP) },
       FXe = function(x, p) { .setState(x); .setParams(p); matrix(.evList(.FXexprP), .ns, .ns) },
       FPe = function(x, p) { .setState(x); .setParams(p); matrix(.evList(.FPexprP), .ns, .np) },
       ## pre-parsed bioavailability forms for the dose jump
       fStrP = .fStrP, dFdpStrP = .dFdpStrP,
       evScalar = function(pexpr, params) { .setState(numeric(.ns)); .setParams(params); eval(pexpr, .env) })
}

## Resolve an additive-bolus dose jump (cmt index, F*amt magnitude, d(F*amt)/dtheta).
## dose = list(step = <0-based grid step>, cmt = <state name>, amt = <numeric>).
.rxDiscreteDoseJump <- function(build, dose, params) {
  .ci <- match(dose$cmt, build$st)
  if (dose$cmt %in% names(build$fStrP)) {
    .Fval <- build$evScalar(build$fStrP[[dose$cmt]], params)
    .dFdp <- vapply(build$dFdpStrP[[dose$cmt]], build$evScalar, numeric(1), params = params)
  } else { .Fval <- 1; .dFdp <- numeric(build$np) }
  list(ci = .ci, Fval = .Fval, dFdp = .dFdp, amt = dose$amt)
}

## one fixed-step explicit RK4 step; returns next state + the four stage states
.rxRk4Step <- function(build, Xn, params, h) {
  .k1 <- build$Fe(Xn, params);        .a2 <- Xn + h / 2 * .k1
  .k2 <- build$Fe(.a2, params);       .a3 <- Xn + h / 2 * .k2
  .k3 <- build$Fe(.a3, params);       .a4 <- Xn + h * .k3
  .k4 <- build$Fe(.a4, params)
  list(Xnext = Xn + h / 6 * (.k1 + 2 * .k2 + 2 * .k3 + .k4),
       stages = list(Xn, .a2, .a3, .a4))
}

#' Discrete forward sensitivity of a fixed-step RK4 solve
#'
#' Integrates the state and its exact derivative `S = dX/dtheta` through the RK4
#' map, storing the stages for the discrete adjoint.
#'
#' @param build object from [.rxDiscreteAdjointBuild()].
#' @param X0 named numeric initial state (order = `build$st`).
#' @param params named numeric parameter vector.
#' @param h fixed step size.
#' @param nStep number of RK4 steps (final time `= h * nStep`).
#' @param doses optional list of additive-bolus dose specs, each
#'   `list(step = <0-based grid step>, cmt = <state name>, amt = <numeric>)`;
#'   applied (with bioavailability `f(cmt)` if defined) at the start of the step.
#' @param S0 optional initial sensitivity (`ns x np`); default zero.
#' @return list with `XN`, `SN` (`ns x np`), `Sall` (per-step `S`), `stages`.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxDiscreteForwardSens <- function(build, X0, params, h, nStep, doses = NULL, S0 = NULL) {
  .ns <- build$ns; .np <- build$np
  .X <- X0; .S <- if (is.null(S0)) matrix(0, .ns, .np) else S0
  .Sall <- vector("list", nStep); .stages <- vector("list", nStep)
  .dstep <- if (length(doses)) split(doses, vapply(doses, function(d) d$step, numeric(1))) else list()
  for (n in seq_len(nStep)) {
    for (.d in .dstep[[as.character(n - 1L)]]) {   # dose at grid step n-1 (start of step)
      .j <- .rxDiscreteDoseJump(build, .d, params)
      .X[.j$ci] <- .X[.j$ci] + .j$Fval * .j$amt
      .S[.j$ci, ] <- .S[.j$ci, ] + .j$amt * .j$dFdp
    }
    .s4 <- .rxRk4Step(build, .X, params, h); .stages[[n]] <- .s4$stages
    .a1 <- .s4$stages[[1]]; .a2 <- .s4$stages[[2]]; .a3 <- .s4$stages[[3]]; .a4 <- .s4$stages[[4]]
    .dk1 <- build$FXe(.a1, params) %*% .S + build$FPe(.a1, params)
    .dk2 <- build$FXe(.a2, params) %*% (.S + h / 2 * .dk1) + build$FPe(.a2, params)
    .dk3 <- build$FXe(.a3, params) %*% (.S + h / 2 * .dk2) + build$FPe(.a3, params)
    .dk4 <- build$FXe(.a4, params) %*% (.S + h * .dk3) + build$FPe(.a4, params)
    .S <- .S + h / 6 * (.dk1 + 2 * .dk2 + 2 * .dk3 + .dk4)
    .X <- .s4$Xnext; .Sall[[n]] <- .S
  }
  list(XN = .X, SN = .S, Sall = .Sall, stages = .stages)
}

#' Discrete adjoint gradient of a trajectory objective
#'
#' Given the RK4 stages from [.rxDiscreteForwardSens()] and, per observation
#' step, the objective covector `c_n = dG/dX_{step_n}`, computes `dG/dtheta` by
#' reverse-mode differentiation through the RK4 steps -- equal to
#' `sum_n c_n^T S_{step_n}` to machine precision, in one backward pass.
#'
#' @param build object from [.rxDiscreteAdjointBuild()].
#' @param stages RK4 stage list from [.rxDiscreteForwardSens()].
#' @param params named numeric parameter vector.
#' @param h fixed step size (same as the forward solve).
#' @param obsSteps integer step indices at which observation covectors apply.
#' @param cov list of covector vectors (length `ns`), aligned with `obsSteps`.
#' @param doses optional list of additive-bolus dose specs (same format as
#'   [.rxDiscreteForwardSens()]); the exact transpose of the forward dose jump
#'   is applied, contributing `amt * dF/dtheta * lambda_c` to `dG/dtheta`.
#' @param lam0Cov optional initial-condition sensitivity `dX0/dtheta` (`ns x np`)
#'   to add the terminal transversality `lambda(t0)^T dX0/dtheta`; default none.
#' @return named numeric vector `dG/dtheta` over `build$calcSens`.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxDiscreteAdjointGrad <- function(build, stages, params, h, obsSteps, cov,
                                   doses = NULL, lam0Cov = NULL) {
  .ns <- build$ns; .np <- build$np; .nStep <- length(stages)
  .lam <- numeric(.ns); .mu <- numeric(.np)
  .obsMap <- match(seq_len(.nStep), obsSteps)
  .dstep <- if (length(doses)) split(doses, vapply(doses, function(d) d$step, numeric(1))) else list()
  for (n in .nStep:1) {
    if (!is.na(.obsMap[n])) .lam <- .lam + cov[[.obsMap[n]]]
    .a1 <- stages[[n]][[1]]; .a2 <- stages[[n]][[2]]; .a3 <- stages[[n]][[3]]; .a4 <- stages[[n]][[4]]
    .Xbar <- .lam
    .k1b <- h / 6 * .Xbar; .k2b <- h / 3 * .Xbar; .k3b <- h / 3 * .Xbar; .k4b <- h / 6 * .Xbar
    .a4b <- t(build$FXe(.a4, params)) %*% .k4b; .Xbar <- .Xbar + .a4b; .k3b <- .k3b + h * .a4b;     .mu <- .mu + t(build$FPe(.a4, params)) %*% .k4b
    .a3b <- t(build$FXe(.a3, params)) %*% .k3b; .Xbar <- .Xbar + .a3b; .k2b <- .k2b + h / 2 * .a3b; .mu <- .mu + t(build$FPe(.a3, params)) %*% .k3b
    .a2b <- t(build$FXe(.a2, params)) %*% .k2b; .Xbar <- .Xbar + .a2b; .k1b <- .k1b + h / 2 * .a2b; .mu <- .mu + t(build$FPe(.a2, params)) %*% .k2b
    .a1b <- t(build$FXe(.a1, params)) %*% .k1b; .Xbar <- .Xbar + .a1b;                              .mu <- .mu + t(build$FPe(.a1, params)) %*% .k1b
    .Ybar <- as.vector(.Xbar)   # dG/dY_{n-1}, Y = post-dose state at step n-1
    for (.d in .dstep[[as.character(n - 1L)]]) {   # transpose of the dose-parameter jump
      .j <- .rxDiscreteDoseJump(build, .d, params)
      .mu <- .mu + .j$amt * .j$dFdp * .Ybar[.j$ci]
    }
    .lam <- .Ybar    # additive bolus: dD/dX = I
  }
  .g <- as.vector(.mu)
  if (!is.null(lam0Cov)) .g <- .g + as.vector(.lam %*% lam0Cov)  # IC transversality
  stats::setNames(.g, build$calcSens)
}

## adjoint-expansion model builder for the in-engine discrete-adjoint `rk4s`
## method

#' Build the adjoint-expansion model text for an in-engine discrete-adjoint solve
#'
#' Exposes the state Jacobian `F_X = dF/dy` and parameter forcing
#' `F_p = dF/dtheta` as rxode2 lhs assignments spliced onto the base ODE, so the
#' C++ backward sweep can evaluate them at any stage state via `calc_lhs`.  The
#' lhs are emitted in a fixed order: all `F_X` entries row-major
#' (`rx__adjFX_i_j__` at index `i*ns + j`), then `F_p` (`rx__adjFP_i_p__` at
#' `ns*ns + i*np + p`).
#'
#' @inheritParams .rxDiscreteAdjointBuild
#' @return list with `text` (expanded model text), `st`, `ns`, `np`,
#'   `calcSens`, and lhs layout `fxOff` (=0), `fpOff` (=ns*ns), `nlhsAdj`.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjointExpand <- function(object, calcSens, stiff = FALSE) {
  # Prune if/else branches first (rxS cannot differentiate conditional-dosing
  # lines); the dosing modifiers (f/alag/rate/dur) remain as unconditional
  # assignments, each activated only when the data flags that dose.
  .model <- .rxLoadPrune(object, doConst = TRUE, promoteLinSens = FALSE)
  .st <- rxode2::rxStateOde(.model); .ns <- length(.st); .np <- length(calcSens)
  if (.ns == 0L) stop("discrete adjoint requires a model with ODE states", call. = FALSE)
  invisible(rxode2::.rxJacobian(.model, c(.st, calcSens)))
  .fromSE <- function(nm) { .d <- get0(nm, envir = .model, inherits = FALSE); rxode2::rxFromSE(.d) }
  .odeLines <- vapply(.st, function(i)
    sprintf("d/dt(%s)=%s", i, .fromSE(paste0("rx__d_dt_", i, "__"))), character(1))
  # Preserve modeled dosing modifiers (f/alag/rate/dur) so the forward primal
  # applies the same dosing as the real model.
  .fLines <- character(0)
  for (k in seq_len(.ns)) {
    .fSE <- get0(paste0("rx_f_", .st[k], "_"), envir = .model, inherits = FALSE)
    if (!is.null(.fSE)) .fLines <- c(.fLines, sprintf("f(%s)=%s", .st[k], rxode2::rxFromSE(.fSE)))
    .lSE <- get0(paste0("rx_lag_", .st[k], "_"), envir = .model, inherits = FALSE)
    if (!is.null(.lSE)) .fLines <- c(.fLines, sprintf("alag(%s)=%s", .st[k], rxode2::rxFromSE(.lSE)))
    .rSE <- get0(paste0("rx_rate_", .st[k], "_"), envir = .model, inherits = FALSE)
    if (!is.null(.rSE)) .fLines <- c(.fLines, sprintf("rate(%s)=%s", .st[k], rxode2::rxFromSE(.rSE)))
    .dSE <- get0(paste0("rx_dur_", .st[k], "_"), envir = .model, inherits = FALSE)
    if (!is.null(.dSE)) .fLines <- c(.fLines, sprintf("dur(%s)=%s", .st[k], rxode2::rxFromSE(.dSE)))
  }
  .fxLines <- character(0)
  # Analytic Jacobian for the stiff adjoint solvers only (Rosenbrock/implicit-RK
  # and the dop853s+ros4s composite).  Emit the nonzero base block (F_X) as real
  # df()/dy() lines so codegen fills calc_jac; the sens block is zero (d/dt==0)
  # and the solver zero-inits the Jacobian, so it stays 0.
  .jacLines <- character(0)
  if (isTRUE(stiff)) {
    for (i in seq_len(.ns)) for (j in seq_len(.ns)) {
      .je <- .fromSE(paste0("rx__df_", .st[i], "_dy_", .st[j], "__"))
      if (!identical(.je, "0") && nzchar(.je))
        .jacLines <- c(.jacLines, sprintf("df(%s)/dy(%s)=%s", .st[i], .st[j], .je))
    }
  }
  for (i in seq_len(.ns)) for (j in seq_len(.ns))
    .fxLines <- c(.fxLines, sprintf("rx__adjFX_%d_%d__=%s", i - 1L, j - 1L,
                                    .fromSE(paste0("rx__df_", .st[i], "_dy_", .st[j], "__"))))
  # F_p is built AFTER the delay scan so a param-dependent delay tau(p) can add
  # its breaking-point correction to the quadrature source F_p.
  .dtauList <- vector("list", .ns)   # per state i: list(stateJ, tau, djac, dtauByP)
  # DDE delayed Jacobian F_Xd[i][j] = d f_i / d(delay(y_j, tau)) and the delay
  # duration tau, as nBase x nBase lhs blocks, for the anticipating costate term.
  # Reuses the delayed-Jacobian machinery from .rxDelaySensAugment (R/dde.R).
  .fxdLines <- character(0); .tauLines <- character(0); .dtauLines <- character(0); .hasDelayAdj <- FALSE
  .fxdMat <- matrix("0", .ns, .ns); .tauMat <- matrix("0", .ns, .ns)
  .dtauMat <- vector("list", .ns)   # [[i]][[j]] = named dtau/dp over calcSens (dose-jump)
  # Collect every delay() subexpression by walking the R expression tree
  # (symengine's VecBasic accessors are masked after .rxJacobian).
  .findDelays <- function(e, acc = list()) {
    if (is.call(e)) {
      if (identical(e[[1L]], as.name("delay")) && length(e) == 3L) {
        acc[[length(acc) + 1L]] <- e
      }
      for (.a in as.list(e)[-1L]) acc <- .findDelays(.a, acc)
    }
    acc
  }
  # Replace every structurally-identical `target` node with `repl` in a tree
  # (spacing-robust; text gsub is unreliable against rxFromSE spacing).
  .substDelay <- function(e, target, repl) {
    if (identical(e, target)) return(repl)
    if (is.call(e)) for (.i in seq_along(e)) e[[.i]] <- .substDelay(e[[.i]], target, repl)
    e
  }
  for (i in seq_len(.ns)) {
    .fi <- get0(paste0("rx__d_dt_", .st[i], "__"), envir = .model, inherits = FALSE)
    if (is.null(.fi)) next
    .fiTxt <- rxode2::rxFromSE(.fi)
    .fullExpr <- parse(text = .fiTxt)[[1L]]
    .dcalls <- .findDelays(.fullExpr)
    if (length(.dcalls) == 0L) next
    .seen <- character(0)
    for (.dc in .dcalls) {
      .dcTxt <- deparse1(.dc)
      if (.dcTxt %in% .seen) next   # identical() subst already captured all copies
      .seen <- c(.seen, .dcTxt)
      .stateJ <- deparse1(.dc[[2L]]); .tau <- deparse1(.dc[[3L]])
      .j <- match(.stateJ, .st); if (is.na(.j)) next
      # d f_i / d(delay(y_j, tau)): substitute a plain symbol for delay(), then
      # differentiate w.r.t. it.
      .gName <- "rx__gdlyATMP__"
      .modTxt <- deparse1(.substDelay(.fullExpr, .dc, as.name(.gName)))
      .dj <- symengine::D(symengine::S(.modTxt), symengine::S(.gName))
      .djTxt <- gsub(.gName, paste0("delay(", .stateJ, ",", .tau, ")"),
                     rxode2::rxFromSE(.dj), fixed = TRUE)
      .fxdMat[i, .j] <- if (identical(.fxdMat[i, .j], "0")) .djTxt else
        paste0(.fxdMat[i, .j], "+(", .djTxt, ")")
      .tauMat[i, .j] <- .tau
      .hasDelayAdj <- TRUE
      # Param-dependent delay tau(p): d tau / d p per calcSens param (resolved in
      # the symengine env), feeding the breaking-point correction to F_p below.
      .dtauByP <- stats::setNames(rep("0", .np), calcSens)
      .tauRes <- tryCatch(eval(parse(text = .tau), envir = .model), error = function(e) NULL)
      if (!is.null(.tauRes) && inherits(.tauRes, "Basic")) {
        for (.pp in calcSens) {
          .dD <- tryCatch(symengine::D(.tauRes, symengine::S(.pp)), error = function(e) NULL)
          if (!is.null(.dD)) .dtauByP[.pp] <- rxode2::rxFromSE(.dD)
        }
      }
      if (any(.dtauByP != "0")) {
        .dtauList[[i]] <- c(.dtauList[[i]], list(list(stateJ = .stateJ, tau = .tau,
                                                      djac = .djTxt, dtauByP = .dtauByP)))
        if (is.null(.dtauMat[[i]])) .dtauMat[[i]] <- vector("list", .ns)
        .dtauMat[[i]][[.j]] <- .dtauByP   # for the dose-jump block
      }
    }
  }
  # F_p (quadrature source): explicit df_i/dp plus, for a param-dependent delay
  # tau(p), the breaking-point term -(d f_i/d delay(y_j,tau)) * rxDelayD(y_j,tau)
  # * d tau/dp (dual of the forward-sens term in .rxDelaySensAugment).
  .fpLines <- character(0)
  for (i in seq_len(.ns)) for (p in seq_len(.np)) {
    .fpExpr <- .fromSE(paste0("rx__df_", .st[i], "_dy_", calcSens[p], "__"))
    for (.z in .dtauList[[i]]) {
      .dt <- .z$dtauByP[[calcSens[p]]]
      if (!is.null(.dt) && !identical(.dt, "0"))
        .fpExpr <- paste0(.fpExpr, "-(", .z$djac, ")*rxDelayD(", .z$stateJ, ",", .z$tau, ")*(", .dt, ")")
    }
    .fpLines <- c(.fpLines, sprintf("rx__adjFP_%d_%d__=%s", i - 1L, p - 1L, .fpExpr))
  }
  if (.hasDelayAdj) {
    for (i in seq_len(.ns)) for (j in seq_len(.ns)) {
      .fxdLines <- c(.fxdLines, sprintf("rx__adjFXd_%d_%d__=%s", i - 1L, j - 1L, .fxdMat[i, j]))
      .tauLines <- c(.tauLines, sprintf("rx__adjTau_%d_%d__=%s", i - 1L, j - 1L, .tauMat[i, j]))
    }
    # d tau_ij / d p block (i,j,p) for the dose-induced breaking-point jump the
    # smooth rxDelayD term misses; applied per dose in the C++ backward sweep.
    for (i in seq_len(.ns)) for (j in seq_len(.ns)) for (p in seq_len(.np)) {
      .v <- if (!is.null(.dtauMat[[i]]) && !is.null(.dtauMat[[i]][[j]])) .dtauMat[[i]][[j]][[calcSens[p]]] else "0"
      if (is.null(.v)) .v <- "0"
      .dtauLines <- c(.dtauLines, sprintf("rx__adjDtau_%d_%d_%d__=%s", i - 1L, j - 1L, p - 1L, .v))
    }
  }
  # rx__sens_<state>_BY_<param>__ output-storage compartments (d/dt = 0): the
  # backward sweep writes dy_k(t_i)/dtheta_p here per observation.  Emitted after
  # the base ODEs so state order is [base, sens]; slot (k,p) is nBase + k*np + p.
  .sensLines <- character(0)
  for (k in seq_len(.ns)) for (p in seq_len(.np))
    .sensLines <- c(.sensLines, sprintf("d/dt(rx__sens_%s_BY_%s__)=0", .st[k], calcSens[p]))
  # dF/dtheta of per-compartment bioavailability F, as lhs (rx__adjdF_k_p__), for
  # the adjoint contribution mu += amt*dF/dtheta*lambda[c].
  .dfLines <- character(0)
  for (k in seq_len(.ns)) {
    .fSE <- get0(paste0("rx_f_", .st[k], "_"), envir = .model, inherits = FALSE)
    for (p in seq_len(.np)) {
      .expr <- if (is.null(.fSE)) "0" else { .d <- symengine::D(.fSE, calcSens[p]); rxode2::rxFromSE(.d) }
      .dfLines <- c(.dfLines, sprintf("rx__adjdF_%d_%d__=%s", k - 1L, p - 1L, .expr))
    }
  }
  # dlag/dtheta of a per-compartment modeled lag time, as lhs (rx__adjDlag_k_p__),
  # for the dose-time transversality jump mu += -amt*dlag_c/dtheta*(lambda^T F_X[:,c]).
  .dlagLines <- character(0); .hasLagAdj <- FALSE
  for (k in seq_len(.ns))
    if (!is.null(get0(paste0("rx_lag_", .st[k], "_"), envir = .model, inherits = FALSE))) .hasLagAdj <- TRUE
  if (.hasLagAdj) {
    for (k in seq_len(.ns)) {
      .lSE <- get0(paste0("rx_lag_", .st[k], "_"), envir = .model, inherits = FALSE)
      for (p in seq_len(.np)) {
        .expr <- if (is.null(.lSE)) "0" else { .d <- symengine::D(.lSE, calcSens[p]); rxode2::rxFromSE(.d) }
        .dlagLines <- c(.dlagLines, sprintf("rx__adjDlag_%d_%d__=%s", k - 1L, p - 1L, .expr))
      }
    }
  }
  # dR/dtheta of a per-compartment modeled infusion, as lhs (rx__adjDrate_k_p__),
  # for the infusion forcing quadrature and off-boundary transversality duals.  A
  # rate() cmt emits d(rate)/dtheta; a dur() cmt emits d(1/dur)/dtheta (runtime
  # multiplies by amt).  NB: rxFromSE is NSE, so assign the derivative to .d first.
  .drateLines <- character(0); .hasRateAdj <- FALSE
  for (k in seq_len(.ns))
    if (!is.null(get0(paste0("rx_rate_", .st[k], "_"), envir = .model, inherits = FALSE)) ||
        !is.null(get0(paste0("rx_dur_", .st[k], "_"), envir = .model, inherits = FALSE))) .hasRateAdj <- TRUE
  if (.hasRateAdj) {
    for (k in seq_len(.ns)) {
      .rSE <- get0(paste0("rx_rate_", .st[k], "_"), envir = .model, inherits = FALSE)
      .dSE <- get0(paste0("rx_dur_", .st[k], "_"), envir = .model, inherits = FALSE)
      for (p in seq_len(.np)) {
        .expr <- if (!is.null(.rSE)) {              # rate(): d(rate)/dtheta
          .d <- symengine::D(.rSE, calcSens[p]); rxode2::rxFromSE(.d)
        } else if (!is.null(.dSE)) {                # dur(): d(1/dur)/dtheta = -dur'/dur^2
          .d <- -symengine::D(.dSE, calcSens[p]) / (.dSE * .dSE); rxode2::rxFromSE(.d)
        } else "0"
        .drateLines <- c(.drateLines, sprintf("rx__adjDrate_%d_%d__=%s", k - 1L, p - 1L, .expr))
      }
    }
  }
  # STIFF (Rosenbrock) only: the stage matrix W = I/(h*gamma) - J depends on theta
  # and y through J, so the exact discrete adjoint needs rx__adjJp_ = dJ/dtheta
  # and rx__adjJy_ = dJ/dy (d2f/dy2, zero for state-linear f).
  .jpLines <- character(0); .jyLines <- character(0)
  if (isTRUE(stiff)) {
    for (i in seq_len(.ns)) for (j in seq_len(.ns)) {
      .dfx <- get0(paste0("rx__df_", .st[i], "_dy_", .st[j], "__"), envir = .model, inherits = FALSE)
      for (p in seq_len(.np)) {
        .expr <- if (is.null(.dfx)) "0" else { .d <- symengine::D(.dfx, calcSens[p]); rxode2::rxFromSE(.d) }
        .jpLines <- c(.jpLines, sprintf("rx__adjJp_%d_%d_%d__=%s", i - 1L, j - 1L, p - 1L, .expr))
      }
      for (m in seq_len(.ns)) {
        .expr <- if (is.null(.dfx)) "0" else { .d <- symengine::D(.dfx, .st[m]); rxode2::rxFromSE(.d) }
        .jyLines <- c(.jyLines, sprintf("rx__adjJy_%d_%d_%d__=%s", i - 1L, j - 1L, m - 1L, .expr))
      }
    }
  }
  # lhs layout (contiguous from 0): fx, fp, df, [jp, jy if stiff],
  # [fxd, tau, dtau if DDE]  (dtau is the (i,j,p) dose-jump block)
  .afterStiff <- if (isTRUE(stiff)) .ns * .ns + 2L * .ns * .np + .ns * .ns * .np + .ns * .ns * .ns
                 else .ns * .ns + 2L * .ns * .np
  # dlag block is appended last (after the DDE blocks) so it shifts no offset
  .afterDelay <- .afterStiff + (if (.hasDelayAdj) 2L * .ns * .ns + .ns * .ns * .np else 0L)
  .dlagLen <- if (.hasLagAdj) .ns * .np else 0L
  list(text = paste(c(.odeLines, .fLines, .sensLines, .jacLines, .fxLines, .fpLines, .dfLines,
                      .jpLines, .jyLines, .fxdLines, .tauLines, .dtauLines, .dlagLines,
                      .drateLines), collapse = "\n"),
       st = .st, ns = .ns, np = .np, calcSens = calcSens, stiff = isTRUE(stiff),
       fxOff = 0L, fpOff = .ns * .ns, dfOff = .ns * .ns + .ns * .np,
       jpOff = if (isTRUE(stiff)) .ns * .ns + 2L * .ns * .np else -1L,
       jyOff = if (isTRUE(stiff)) .ns * .ns + 2L * .ns * .np + .ns * .ns * .np else -1L,
       fxdOff = if (.hasDelayAdj) .afterStiff else -1L,
       tauOff = if (.hasDelayAdj) .afterStiff + .ns * .ns else -1L,
       dtauOff = if (.hasDelayAdj) .afterStiff + 2L * .ns * .ns else -1L,
       dlagOff = if (.hasLagAdj) .afterDelay else -1L,
       drateOff = if (.hasRateAdj) .afterDelay + .dlagLen else -1L,
       hasDelay = .hasDelayAdj,
       nlhsAdj = .ns * .ns + .ns * .np, sensOff = .ns)
}

## Cached build + solve convenience for the in-engine discrete adjoint ----------

.rxAdjointModelCache <- new.env(parent = emptyenv())

#' Build (and cache) the compiled adjoint-expansion model for rk4s
#'
#' Runs [.rxAdjointExpand()] and compiles the result, caching by the normalized
#' model text + `calcSens` so an optimizer that resolves the gradient many times
#' compiles once.
#'
#' @inheritParams .rxDiscreteAdjointBuild
#' @return list with `model` (compiled rxode2) and `info` (the
#'   [.rxAdjointExpand()] layout).
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjointModel <- function(object, calcSens, stiff = FALSE) {
  .txt <- if (is.character(object) && length(object) == 1L) object else rxode2::rxNorm(rxode2::rxModelVars(object))
  .key <- paste0(.txt, "\n##adj", if (isTRUE(stiff)) "S" else "", "##", paste(calcSens, collapse = ","))
  .hit <- get0(.key, envir = .rxAdjointModelCache, inherits = FALSE)
  if (!is.null(.hit)) return(.hit)
  .ex <- .rxAdjointExpand(object, calcSens, stiff = isTRUE(stiff))
  .ret <- list(model = rxode2::rxode2(.ex$text), info = .ex)
  assign(.key, .ret, envir = .rxAdjointModelCache)
  .ret
}

## Stiff (Rosenbrock / implicit-RK) adjoint methods: those needing the analytic
## Jacobian, so the expansion must be built with stiff=TRUE.  The dop853s+ros4s
## composite counts (its ros4s secondary needs the Jacobian).
.rxAdjointStiffMethods <- c("ros4s", "ros6s", "ros43s", "radauiia5s", "gauss6s",
                            "geng5s", "sdirk43s", "iiic6s", "backwardEulers")

#' Does an adjoint (`s`) solve method require the stiff analytic Jacobian?
#'
#' @param method adjoint method name(s) -- a single name, a composite vector
#'   `c("dop853s","ros4s")`, or a `"dop853s+ros4s"` string.
#' @return `TRUE` if any component is a stiff adjoint method.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjointMethodStiff <- function(method) {
  if (is.null(method)) return(FALSE)
  .m <- unlist(strsplit(as.character(method), "+", fixed = TRUE))
  any(.m %in% .rxAdjointStiffMethods)
}

#' Solve a model with an in-engine discrete-adjoint (`s`) method
#'
#' Applies the (cached) adjoint expansion, solves with the requested adjoint
#' method, and drops the internal `rx__adj*` lhs, returning the full-trajectory
#' `rx__sens_<state>_BY_<param>__` columns.  The stiff analytic Jacobian is
#' emitted into the expansion automatically, inferred from `method`.
#'
#' @inheritParams .rxDiscreteAdjointBuild
#' @param events an rxode2 event table / data set.
#' @param params optional named parameter vector or data frame (per subject).
#' @param method adjoint solve method (default `"rk4s"`).  May be a composite
#'   (`"dop853s+ros4s"`).
#' @param ... passed to [rxode2::rxSolve()].
#' @return the solved data frame with `rx__sens_<state>_BY_<param>__` columns.
#' @author Matthew L. Fidler
#' @export
rxSolveAdjointRk4 <- function(object, events, params = NULL, calcSens, method = "rk4s", ...) {
  .b <- .rxAdjointModel(object, calcSens, stiff = .rxAdjointMethodStiff(method))
  .df <- as.data.frame(rxode2::rxSolve(.b$model, events, params = params, method = method, ...))
  .drop <- grep("^rx__adj(FX|FP|dF|Jp|Jy)_", names(.df), value = TRUE)
  .df[, setdiff(names(.df), .drop), drop = FALSE]
}

rxTest({
  # Keep the ODE adjoint/sensitivity path under test: disable the automatic
  # ODE -> linCmt() translation, which would otherwise collapse these linear PK
  # models onto the analytic linCmt() solver and bypass the discrete-adjoint /
  # stiff-Jacobian machinery entirely.
  withr::local_options(rxode2.useLinCmt = FALSE)

  # Discrete adjoint sensitivity (fixed-step explicit RK4).
  #
  # Defining property (vs the continuous adjoint's ~1e-5): the discrete adjoint
  # gradient equals the DISCRETE FORWARD SENSITIVITY of the SAME RK4 scheme to
  # MACHINE PRECISION, because it is the exact reverse-mode transpose of the
  # identical numerical step map.

  mText <- "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center"
  cs <- c("ka", "cl", "v"); p <- c(ka = 1.2, cl = 3.5, v = 25)
  X0 <- c(depot = 100, center = 0)
  h <- 0.02; nStep <- 600L                       # t_N = 12
  B <- rxode2::.rxDiscreteAdjointBuild(mText, cs)
  fwd <- rxode2::.rxDiscreteForwardSens(B, X0, p, h, nStep)

  test_that("discrete adjoint of a terminal-state objective matches forward sens to machine precision", {
    skip_on_cran()
    skip_on_ci()
    for (kOut in seq_along(B$st)) {
      cov <- list(as.numeric(seq_along(B$st) == kOut))  # covector e_k at final step
      gAdj <- rxode2::.rxDiscreteAdjointGrad(B, fwd$stages, p, h, nStep, cov)
      gFwd <- fwd$SN[kOut, ]                            # e_k^T S_N
      expect_lt(max(abs(gAdj - gFwd)), 1e-10)
    }
  })

  test_that("discrete adjoint of a trajectory objective matches forward sens to machine precision", {
    skip_on_cran()
    skip_on_ci()
    fa <- rxode2::.rxDiscreteForwardSens(B, X0, p, h, nStep)  # provides Sall
    obsSteps <- c(50L, 100L, 200L, 300L, 450L, 600L)
    set.seed(1)
    cov <- lapply(seq_along(obsSteps), function(i) stats::rnorm(B$ns))
    gAdj <- rxode2::.rxDiscreteAdjointGrad(B, fa$stages, p, h, obsSteps, cov)
    gFwd <- Reduce(`+`, lapply(seq_along(obsSteps), function(i)
      as.vector(cov[[i]] %*% fa$Sall[[obsSteps[i]]])))
    expect_lt(max(abs(gAdj - gFwd)), 1e-10)
  })

  test_that("discrete adjoint with an additive-bolus (F) dose jump matches forward sens to machine precision", {
    skip_on_cran()
    skip_on_ci()
    dText <- "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\nf(depot)=Fbio"
    dcs <- c("ka", "cl", "v", "Fbio"); dp <- c(ka = 1.2, cl = 3.5, v = 25, Fbio = 0.7)
    dX0 <- c(depot = 0, center = 0)                 # dose supplies depot at t0
    dh <- 0.02; dN <- 600L
    doses <- list(list(step = 0L, cmt = "depot", amt = 100))
    DB <- rxode2::.rxDiscreteAdjointBuild(dText, dcs)
    dfwd <- rxode2::.rxDiscreteForwardSens(DB, dX0, dp, dh, dN, doses = doses)
    for (kOut in seq_along(DB$st)) {
      cov <- list(as.numeric(seq_along(DB$st) == kOut))
      gAdj <- rxode2::.rxDiscreteAdjointGrad(DB, dfwd$stages, dp, dh, dN, cov, doses = doses)
      gFwd <- dfwd$SN[kOut, ]
      expect_lt(max(abs(gAdj - gFwd)), 1e-10)
    }
    # Fbio gradient is genuinely nonzero (dose amount depends on it)
    covC <- list(as.numeric(DB$st == "center"))
    gC <- rxode2::.rxDiscreteAdjointGrad(DB, dfwd$stages, dp, dh, dN, covC, doses = doses)
    expect_gt(abs(gC[["Fbio"]]), 1)
  })

  test_that(".rxAdjointExpand exposes F_X/F_p as lhs matching the reference to machine precision", {
    skip_on_cran()
    skip_on_ci()
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    expect_equal(ex$ns, 2L); expect_equal(ex$np, 3L)
    expect_equal(ex$fxOff, 0L); expect_equal(ex$fpOff, 4L); expect_equal(ex$nlhsAdj, 10L)
    m <- rxode2::rxode2(ex$text)
    mv <- rxode2::rxModelVars(m)
    # first nlhsAdj lhs are the adjoint expansion, in the documented order
    expect_equal(mv$lhs[seq_len(ex$nlhsAdj)],
                 c("rx__adjFX_0_0__","rx__adjFX_0_1__","rx__adjFX_1_0__","rx__adjFX_1_1__",
                   "rx__adjFP_0_0__","rx__adjFP_0_1__","rx__adjFP_0_2__",
                   "rx__adjFP_1_0__","rx__adjFP_1_1__","rx__adjFP_1_2__"))
    xtest <- c(depot = 8.3, center = 41.2)
    s <- rxode2::rxSolve(m, et(0), params = p, inits = xtest, method = "liblsoda")
    sd <- as.data.frame(s)
    fx <- matrix(0, ex$ns, ex$ns); fp <- matrix(0, ex$ns, ex$np)
    for (i in seq_len(ex$ns)) for (j in seq_len(ex$ns))
      fx[i, j] <- sd[[sprintf("rx__adjFX_%d_%d__", i - 1L, j - 1L)]]
    for (i in seq_len(ex$ns)) for (pp in seq_len(ex$np))
      fp[i, pp] <- sd[[sprintf("rx__adjFP_%d_%d__", i - 1L, pp - 1L)]]
    expect_lt(max(abs(fx - B$FXe(xtest, p))), 1e-12)
    expect_lt(max(abs(fp - B$FPe(xtest, p))), 1e-12)
  })

  test_that("in-engine rk4s solver fills rx__sens_* columns matching FD of the RK4 solve", {
    skip_on_cran()
    skip_on_ci()
    tms <- c(1, 2, 4, 8, 12, 24)
    ev <- et(amt = 100, cmt = "depot") %>% et(tms)
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text)
    sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = "rk4s", cores = 1))
    # base states must match the identical-stepping rk4 primal
    mbase <- rxode2::rxode2(mText)
    solveBase <- function(pp) as.matrix(as.data.frame(
      rxode2::rxSolve(mbase, ev, params = pp, method = "rk4", cores = 1))[, c("depot", "center")])
    b0 <- solveBase(p)
    expect_lt(max(abs(as.matrix(sd[, c("depot", "center")]) - b0)), 1e-8)
    # adjoint sens columns vs central FD of the same RK4 map
    fdmax <- 0
    for (pn in cs) {
      hh <- p[[pn]] * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
      fd <- (solveBase(pp) - solveBase(pm)) / (2 * hh)
      for (k in seq_along(ex$st)) {
        adj <- sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]]
        fdmax <- max(fdmax, max(abs(adj - fd[, k])))
      }
    }
    expect_lt(fdmax, 1e-4)
  })

  test_that("steady-state (ss) bolus: forward-only adjoint primal matches base to machine precision", {
    skip_on_cran()
    skip_on_ci()
    # The ss pre-solve reuses the base method's single-point stepper for the
    # adjoint method codes (rk4s -> rk4), so the forward primal solves
    # steady-state dosing exactly (previously it left yp un-advanced -> huge
    # divergence).
    evSS <- et(amt = 100, cmt = "depot", ss = 1, ii = 12) %>% et(c(1, 2, 4, 8, 12))
    mbase <- rxode2::rxode2(mText)
    base <- as.matrix(as.data.frame(
      rxode2::rxSolve(mbase, evSS, params = p, method = "rk4", cores = 1))[, c("depot", "center")])
    mplain <- rxode2::rxode2(mText)
    fo <- as.matrix(as.data.frame(
      rxode2::rxSolve(mplain, evSS, params = p, method = "rk4s", cores = 1))[, c("depot", "center")])
    expect_lt(max(abs(fo - base)), 1e-10)
  })

  test_that("steady-state (ss) bolus: expanded adjoint sensitivities match forward sensitivities", {
    skip_on_cran()
    skip_on_ci()
    # The backward sweep adds the steady-state initial-condition sensitivity
    # dY_ss/dp via a one-period monodromy (rk4s.cpp), so the expanded adjoint
    # rx__sens_* columns reproduce the forward-sensitivity gradients at steady
    # state.  Compared against the analytic FORWARD-sensitivity model solved with
    # the SAME base discretization.
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text)
    mfwd <- rxode2::rxode2(mText, calcSens = cs)      # forward-sensitivity model
    chkFwd <- function(evSS, adjMeth, fwdMeth, tol) {
      a <- as.data.frame(suppressWarnings(rxode2::rxSolve(madj, evSS, params = p, method = adjMeth, cores = 1)))
      f <- as.data.frame(suppressWarnings(rxode2::rxSolve(mfwd, evSS, params = p, method = fwdMeth, cores = 1)))
      mx <- 0
      for (pn in cs) for (st in ex$st)
        mx <- max(mx, max(abs(a[[sprintf("rx__sens_%s_BY_%s__", st, pn)]] -
                              f[[sprintf("rx__sens_%s_BY_%s__", st, pn)]])))
      expect_lt(mx, tol)
    }
    evSS  <- et(amt = 100, cmt = "depot", ss = 1, ii = 12) %>% et(c(1, 2, 4, 8, 12))
    evSSa <- et(amt = 100, cmt = "depot", ss = 1, ii = 12, addl = 2) %>% et(seq(1, 40, by = 3))
    chkFwd(evSS,  "rk4s",    "rk4",    1e-5)   # fixed-step: matches to FD level
    chkFwd(evSS,  "vern98s", "vern98", 1e-5)   # high-order embedded RK
    chkFwd(evSS,  "dop54s",  "dop54",  1e-3)   # adaptive: base discretization differs
    chkFwd(evSSa, "rk4s",    "rk4",    1e-5)   # ss bolus + regular addl doses
    # fixed-rate periodic infusion steady state (dur = 100/10 = 10 < ii = 12):
    # the ss period (ON dur, OFF ii-dur) monodromy supplies the IC term.
    evInf  <- et(amt = 100, rate = 10, cmt = "depot", ss = 1, ii = 12) %>% et(c(1, 2, 4, 8, 12))
    evInfa <- et(amt = 100, rate = 10, cmt = "depot", ss = 1, ii = 12, addl = 2) %>% et(seq(1, 40, by = 3))
    chkFwd(evInf,  "rk4s",    "rk4",    1e-4)
    chkFwd(evInf,  "vern98s", "vern98", 1e-4)
    chkFwd(evInfa, "rk4s",    "rk4",    1e-4)   # ss infusion + regular addl doses
    # large-duration infusion (dur = 100/10 = 10 > ii = 8): overlapping
    # infusions -> two-phase periodic steady state ((numDoseInf+1)R, offTime)
    # then (numDoseInf*R, addTime).
    evLarge <- et(amt = 100, rate = 10, cmt = "depot", ss = 1, ii = 8) %>% et(c(1, 2, 4, 8, 12, 16))
    chkFwd(evLarge, "rk4s",    "rk4",    1e-4)
    chkFwd(evLarge, "vern98s", "vern98", 1e-4)
    # continuous infusion to a constant steady state (SSINF): dY_ss/dp is the
    # -J^{-1} df/dp linear solve rather than a monodromy.
    evCont <- et(amt = 0, rate = 10, cmt = "depot", ss = 1) %>% et(c(1, 2, 4, 8, 12))
    chkFwd(evCont, "rk4s",    "rk4",    1e-5)
    chkFwd(evCont, "vern98s", "vern98", 1e-5)
    # liblsodaadj (the DEFAULT adjoint driver) also handles continuous ss: the
    # ss pre-solve is recorded-paused so the window starts at Y_ss, then the same
    # -J^{-1} df/dp IC term is added in the multistep backward sweep.
    chkFwd(evCont, "liblsodaadj", "lsoda", 1e-4)
  })

  test_that("steady-state (ss): FD cross-check of the expanded rk4s adjoint sensitivities", {
    skip_on_cran()
    skip_on_ci()
    evSS <- et(amt = 100, cmt = "depot", ss = 1, ii = 12) %>% et(c(1, 2, 4, 8, 12))
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text)
    sd <- as.data.frame(suppressWarnings(
      rxode2::rxSolve(madj, evSS, params = p, method = "rk4s", cores = 1)))
    mbase <- rxode2::rxode2(mText)
    solveBase <- function(pp) as.matrix(as.data.frame(
      rxode2::rxSolve(mbase, evSS, params = pp, method = "rk4", cores = 1))[, c("depot", "center")])
    fdmax <- 0
    for (pn in cs) {
      hh <- p[[pn]] * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
      fd <- (solveBase(pp) - solveBase(pm)) / (2 * hh)
      for (k in seq_along(ex$st))
        fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
    }
    expect_lt(fdmax, 1e-4)
  })

  test_that("steady-state ss=2 (superposition, bolus): adjoint sensitivities match forward", {
    skip_on_cran()
    skip_on_ci()
    ex <- rxode2::.rxAdjointExpand(mText, cs); madj <- rxode2::rxode2(ex$text)
    mfwd <- rxode2::rxode2(mText, calcSens = cs)
    scol <- as.vector(outer(ex$st, cs, function(s, pn) sprintf("rx__sens_%s_BY_%s__", s, pn)))
    chk2 <- function(ev, adjMeth, fwdMeth, tol) {
      a <- as.data.frame(suppressWarnings(rxode2::rxSolve(madj, ev, params = p, method = adjMeth)))
      f <- as.data.frame(suppressWarnings(rxode2::rxSolve(mfwd, ev, params = p, method = fwdMeth)))
      mx <- 0; for (cn in scol) mx <- max(mx, max(abs(a[[cn]] - f[[cn]]), na.rm = TRUE))
      expect_lt(mx / max(1, max(abs(unlist(f[scol])), na.rm = TRUE)), tol)
    }
    # normal loading dose at t0, then a maintenance ss=2 superposition at t=12
    evA <- et(amt = 100, cmt = "depot") %>% et(amt = 50, cmt = "depot", ss = 2, ii = 12, time = 12) %>%
      et(c(1, 4, 8, 12, 16, 20, 24))
    chk2(evA, "rk4s",    "rk4",    1e-5)
    chk2(evA, "vern98s", "vern98", 1e-5)
    # two ss=2 superposition events (t=12 and t=36)
    evB <- et(amt = 100, cmt = "depot") %>% et(amt = 50, cmt = "depot", ss = 2, ii = 12, time = 12) %>%
      et(amt = 50, cmt = "depot", ss = 2, ii = 12, time = 36) %>% et(c(4, 12, 24, 36, 48))
    chk2(evB, "rk4s", "rk4", 1e-5)
    # interleaved loading ss=1 + maintenance ss=2 + interior ss=1 reset (the
    # nmtest id=125 pattern): ss1 at t0, ss2 at t12, ss1 (reset) at t24, ss2 at t36
    evC <- et(amt = 100, cmt = "depot", ss = 1, ii = 24) %>%
      et(amt = 50, cmt = "depot", ss = 2, ii = 24, time = 12) %>%
      et(amt = 100, cmt = "depot", ss = 1, ii = 24, time = 24) %>%
      et(amt = 50, cmt = "depot", ss = 2, ii = 24, time = 36) %>% et(c(4, 12, 20, 24, 30, 36, 44))
    chk2(evC, "rk4s",    "rk4",    1e-4)
    chk2(evC, "vern98s", "vern98", 1e-4)
    # ss=2 fixed-rate PERIODIC infusion superposition (dur = 50/10 = 5 < ii = 12)
    evD <- et(amt = 100, cmt = "central") %>%
      et(amt = 50, rate = 10, cmt = "central", ss = 2, ii = 12, time = 12) %>% et(c(1, 8, 12, 16, 24, 30))
    chk2(evD, "rk4s",    "rk4",    1e-4)
    chk2(evD, "vern98s", "vern98", 1e-4)
    # ss=2 LARGE-duration infusion (dur = 100/10 = 10 > ii = 8; two-phase monodromy)
    evE <- et(amt = 80, cmt = "central") %>%
      et(amt = 100, rate = 10, cmt = "central", ss = 2, ii = 8, time = 8) %>% et(c(1, 8, 12, 16, 24, 30))
    chk2(evE, "rk4s",    "rk4",    1e-4)
    chk2(evE, "vern98s", "vern98", 1e-4)
    # ss=2 FULL-interval infusion (dur = 120/10 = 12 == ii = 12; -J^-1 df/dp IC)
    evF <- et(amt = 80, cmt = "central") %>%
      et(amt = 120, rate = 10, cmt = "central", ss = 2, ii = 12, time = 12) %>% et(c(1, 8, 12, 16, 24, 30))
    chk2(evF, "rk4s",    "rk4",    1e-4)
    chk2(evF, "vern98s", "vern98", 1e-4)
  })

  test_that("steady-state on the composite (AutoSwitch) path: adjoint sensitivities match forward", {
    skip_on_cran()
    skip_on_ci()
    ex <- rxode2::.rxAdjointExpand(mText, cs); madj <- rxode2::rxode2(ex$text)
    mfwd <- rxode2::rxode2(mText, calcSens = cs)
    scol <- as.vector(outer(ex$st, cs, function(s, pn) sprintf("rx__sens_%s_BY_%s__", s, pn)))
    # the composite fill builds the ss IC (rk4sSsIc) with its primary explicit
    # tableau; compare dop853s+ros4s (composite adjoint) against dop853 (forward).
    chkC <- function(ev, tol) {
      a <- as.data.frame(suppressWarnings(rxode2::rxSolve(madj, ev, params = p, method = "dop853s+ros4s")))
      f <- as.data.frame(suppressWarnings(rxode2::rxSolve(mfwd, ev, params = p, method = "dop853")))
      mx <- 0; for (cn in scol) mx <- max(mx, max(abs(a[[cn]] - f[[cn]]), na.rm = TRUE))
      expect_lt(mx / max(1, max(abs(unlist(f[scol])), na.rm = TRUE)), tol)
    }
    smp <- c(1, 4, 8, 12, 16, 24)
    chkC(et(amt = 100, cmt = "depot", ss = 1, ii = 12) %>% et(smp), 1e-4)                       # ss1 bolus
    chkC(et(amt = 100, rate = 25, cmt = "central", ss = 1, ii = 12) %>% et(smp), 1e-4)          # ss1 periodic infusion
    chkC(et(amt = 120, rate = 10, cmt = "central", ss = 1, ii = 12) %>% et(smp), 1e-4)          # ss1 full-interval infusion
    chkC(et(amt = 100, cmt = "depot") %>% et(amt = 50, cmt = "depot", ss = 2, ii = 12, time = 12) %>% et(smp), 1e-4)          # ss2 bolus
    chkC(et(amt = 100, cmt = "central") %>% et(amt = 50, rate = 10, cmt = "central", ss = 2, ii = 12, time = 12) %>% et(smp), 1e-4)  # ss2 infusion
    chkC(et(amt = 100, cmt = "depot", ss = 1, ii = 24) %>%
           et(amt = 100, cmt = "depot", ss = 1, ii = 24, time = 24) %>% et(c(1, 12, 20, 24, 30, 44)), 1e-4)  # interior ss1 reset
  })

  test_that("steady-state (ss): not-yet-covered cases and drivers stay guarded", {
    skip_on_cran()
    skip_on_ci()
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text)
    evSS <- et(amt = 100, cmt = "depot", ss = 1, ii = 12) %>% et(c(1, 2, 4, 8, 12))
    # MODELED rate()/dur() ss is covered on the non-composite explicit path (below,
    # vs FD) but guarded on composite and for large-duration (dur>=ii) regimens.
    mMdl <- rxode2::rxode2(rxode2::.rxAdjointExpand(
      "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\ndur(center)=9*cl/3.5", c("ka", "cl", "v"))$text)
    expect_error(rxode2::rxSolve(mMdl, et(amt = 100, rate = -2, cmt = "center", ss = 1, ii = 24) %>% et(c(1, 8, 16)),
                                 params = p, method = "dop853s+ros4s", cores = 1), "steady-state")   # composite guarded
    expect_error(rxode2::rxSolve(mMdl, et(amt = 100, rate = -2, cmt = "center", ss = 1, ii = 8) %>% et(c(1, 4, 8)),
                                 params = p, method = "rk4s", cores = 1))   # large-dur modeled -> errors
    # liblsodaadj covers single ss==1 (below) but NOT ss==2 or interior/multiple ss==1
    expect_error(rxode2::rxSolve(madj, et(amt = 100, cmt = "depot", ss = 2, ii = 12) %>% et(c(1, 2, 4)),
                                 params = p, method = "liblsodaadj", cores = 1), "steady-state")
    expect_error(rxode2::rxSolve(madj, et(amt = 100, cmt = "depot", ss = 1, ii = 24) %>%
                                   et(amt = 100, cmt = "depot", ss = 1, ii = 24, time = 24) %>% et(c(1, 12, 30)),
                                 params = p, method = "liblsodaadj", cores = 1), "steady-state")
  })

  test_that("steady-state on the liblsodaadj multistep driver: adjoint sensitivities match forward", {
    skip_on_cran()
    skip_on_ci()
    ex <- rxode2::.rxAdjointExpand(mText, cs); madj <- rxode2::rxode2(ex$text)
    mfwd <- rxode2::rxode2(mText, calcSens = cs)
    scol <- as.vector(outer(ex$st, cs, function(s, pn) sprintf("rx__sens_%s_BY_%s__", s, pn)))
    # the driver re-records ONE ss period (recording-paused pre-solve) for the
    # monodromy IC (bolus/finite infusion) or a linear solve (continuous/full-int).
    chkL <- function(ev, tol) {
      a <- as.data.frame(suppressWarnings(rxode2::rxSolve(madj, ev, params = p, method = "liblsodaadj", cores = 1)))
      f <- as.data.frame(suppressWarnings(rxode2::rxSolve(mfwd, ev, params = p, method = "liblsoda", cores = 1)))
      mx <- 0; for (cn in scol) mx <- max(mx, max(abs(a[[cn]] - f[[cn]]), na.rm = TRUE))
      expect_lt(mx / max(1, max(abs(unlist(f[scol])), na.rm = TRUE)), tol)
    }
    smp <- c(1, 4, 8, 12, 16, 24)
    chkL(et(amt = 100, cmt = "depot", ss = 1, ii = 12, addl = 2) %>% et(smp), 1e-4)             # bolus (addl)
    chkL(et(amt = 100, rate = 25, cmt = "central", ss = 1, ii = 12, addl = 2) %>% et(smp), 1e-4) # periodic infusion
    chkL(et(amt = 100, rate = 10, cmt = "central", ss = 1, ii = 8) %>% et(smp), 1e-4)            # large-dur infusion
    chkL(et(amt = 120, rate = 10, cmt = "central", ss = 1, ii = 12) %>% et(smp), 1e-4)           # full-interval infusion
    chkL(et(amt = 0, rate = 8, cmt = "central", ss = 1) %>% et(smp), 1e-4)                       # continuous infusion
  })

  test_that("modeled rate()/dur() steady-state: adjoint sensitivities match FINITE DIFFERENCES", {
    skip_on_cran()
    skip_on_ci()
    # For a modeled rate/dur whose value DEPENDS on a sensitivity parameter, the ss
    # has a MOVING period boundary.  The discrete adjoint carries the dR/dp forcing
    # + transversality B terms; validate against central FD (rxode2's FORWARD
    # sensitivities are WRONG for this case -- they omit the moving-boundary term,
    # so FD, not forward-sens, is the oracle here).
    csM <- c("ka", "cl"); pM <- c(ka = 1.2, cl = 3.5, v = 25)
    chkFD <- function(mt, ev, tol) {
      exM <- rxode2::.rxAdjointExpand(mt, csM); madj <- rxode2::rxode2(exM$text); mb <- rxode2::rxode2(mt)
      a <- as.data.frame(suppressWarnings(rxode2::rxSolve(madj, ev, params = pM, method = "rk4s", cores = 1, atol = 1e-11, rtol = 1e-11)))
      mx <- 0
      for (pn in csM) {
        hh <- abs(pM[[pn]]) * 1e-6; pp <- pM; pm <- pM; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        sp <- as.data.frame(suppressWarnings(rxode2::rxSolve(mb, ev, params = pp, method = "rk4", cores = 1, atol = 1e-11, rtol = 1e-11)))
        sm <- as.data.frame(suppressWarnings(rxode2::rxSolve(mb, ev, params = pm, method = "rk4", cores = 1, atol = 1e-11, rtol = 1e-11)))
        for (st in exM$st) { fd <- (sp[[st]] - sm[[st]]) / (2 * hh)
          mx <- max(mx, max(abs(a[[sprintf("rx__sens_%s_BY_%s__", st, pn)]] - fd), na.rm = TRUE)) }
      }
      expect_lt(mx, tol)
    }
    smp <- c(1, 8, 16, 24, 30, 40)
    evM <- et(amt = 100, rate = -2, cmt = "central", ss = 1, ii = 24) %>% et(smp)   # modeled dur
    evR <- et(amt = 100, rate = -1, cmt = "central", ss = 1, ii = 24) %>% et(smp)   # modeled rate
    chkFD("d/dt(depot)=-ka*depot\nd/dt(central)=ka*depot-(cl/v)*central\ndur(central)=9*cl/3.5\ncp=central/(v/1000)",  evM, 1e-6)
    chkFD("d/dt(depot)=-ka*depot\nd/dt(central)=ka*depot-(cl/v)*central\nrate(central)=11*cl/3.5\ncp=central/(v/1000)", evR, 1e-6)
    # bioavailability (F != 1): CONSTANT F folds into the effective rate F*amt/dur,
    # and a PARAMETER-DEPENDENT F (dF != 0) adds a dF term to the infusion dual
    # (forcing for dur, off-boundary for rate).
    chkFD("d/dt(depot)=-ka*depot\nd/dt(central)=ka*depot-(cl/v)*central\nf(central)=0.61\ndur(central)=9*cl/3.5\ncp=central/(v/1000)",   evM, 1e-6)  # const F, dur
    chkFD("d/dt(depot)=-ka*depot\nd/dt(central)=ka*depot-(cl/v)*central\nf(central)=0.61*cl/3.5\ndur(central)=9\ncp=central/(v/1000)",    evM, 1e-6)  # dF!=0, dur
    chkFD("d/dt(depot)=-ka*depot\nd/dt(central)=ka*depot-(cl/v)*central\nf(central)=0.61*cl/3.5\nrate(central)=11\ncp=central/(v/1000)",  evR, 1e-6)  # dF!=0, rate
  })

  test_that("nmtest dosing scenarios: adjoint solution AND gradients match forward sensitivities (ka + elimination)", {
    skip_on_cran()
    skip_on_ci()
    skip_if_not_installed("nlmixr2data")
    d0 <- nlmixr2data::nmtest
    # nmtest 2-cmt model with ka + elimination (cl) as free sensitivity params,
    # INCLUDING the if(){} modeled-rate/dur conditional dosing (the adjoint
    # expansion prunes these branches, like the forward-sensitivity expansion).
    mt <- paste("d/dt(depot)=-ka*depot",
                "d/dt(central)=ka*depot-(cl/v)*central",
                "f(central)=bioav",
                "if (mode==1){", "rate(central)=rat2", "}",
                "if (mode==2){", "dur(central)=dur2", "}",
                "cp=central/(v/1000)", sep = "\n")
    ncs <- c("ka", "cl"); pp <- c(ka = 1.5, cl = 1.1, v = 20)
    nex <- rxode2::.rxAdjointExpand(mt, ncs); nmadj <- rxode2::rxode2(nex$text)
    nmfwd <- rxode2::rxode2(mt, calcSens = ncs)                # analytic forward sensitivities
    ncols <- as.vector(outer(nex$st, ncs, function(s, pn) sprintf("rx__sens_%s_BY_%s__", s, pn)))
    # Skipped: doses to cmt>=3 hit the sensitivity output compartments (which
    # differ between the adjoint output slots and the forward ODE states).
    # (Observations coincident with a full reset -- ids 23/24 -- are handled via
    # the boundaryDose disambiguation and are NOT skipped.)  Checked across EVERY
    # adjoint dense / adaptive one-step solver (the whole rk4s-framework family)
    # vs the matching base method -- they all share the explicit backward fill.
    tested <- 0
    for (mp in list(c("rk4s", "rk4"), c("dop853s", "dop853"), c("dop5s", "dop5"),
                    c("vern65s", "vern65"), c("vern76s", "vern76"), c("vern98s", "vern98"),
                    c("dop87s", "dop87"), c("f78s", "f78"), c("ck54s", "ck54"),
                    c("bs32s", "bs"), c("rk43s", "rk43"), c("tf65s", "tf65"),
                    c("v65rs", "v65r"), c("v76rs", "v76r"), c("v78s", "v78"),
                    c("v87rs", "v87r"), c("v89s", "v89"), c("dverk78s", "dverk78"),
                    c("tp75s", "tp75"))) {
      for (id in unique(d0$id)) {
        di <- d0[d0$id == id, ]
        if (any(di$evid != 0 & !(di$cmt %in% c(1, 2)))) next
        a <- tryCatch(as.data.frame(suppressWarnings(rxode2::rxSolve(
          nmadj, di, params = pp, method = mp[1], addlDropSs = TRUE, atol = 1e-10, rtol = 1e-10))),
          error = function(e) NULL)
        if (is.null(a)) next                   # guarded / not-yet-supported ss case
        f <- as.data.frame(suppressWarnings(rxode2::rxSolve(
          nmfwd, di, params = pp, method = mp[2], addlDropSs = TRUE, atol = 1e-10, rtol = 1e-10)))
        pd <- 0; for (st in nex$st) pd <- max(pd, max(abs(a[[st]] - f[[st]]), na.rm = TRUE))
        sd <- 0; for (cn in ncols) sd <- max(sd, max(abs(a[[cn]] - f[[cn]]), na.rm = TRUE))
        psc <- max(1, max(abs(unlist(f[nex$st])), na.rm = TRUE))
        ssc <- max(1, max(abs(unlist(f[ncols])), na.rm = TRUE))
        # The adaptive adjoint methods take dense-output steps for the recording,
        # so their primal can differ slightly from the non-dense base method on
        # hard steady-state cases (they are in fact a touch MORE accurate); the
        # gradient -- the adjoint-correctness check -- still matches tightly.
        expect_lt(pd / psc, 1e-2)              # solution (primal) matches
        expect_lt(sd / ssc, 6e-3)              # gradients match
        tested <- tested + 1
      }
    }
    expect_gt(tested, 900)                     # ~54 scenarios x 19 dense solvers
  })

  test_that("STIFF adjoint solvers (Rosenbrock/implicit): non-ss nmtest gradients match FD", {
    skip_on_cran()
    skip_on_ci()
    skip_if_not_installed("nlmixr2data")

    d0 <- nlmixr2data::nmtest
    mt <- paste("d/dt(depot)=-ka*depot", "d/dt(central)=ka*depot-(cl/v)*central",
                "f(central)=bioav", "if (mode==1){", "rate(central)=rat2", "}",
                "if (mode==2){", "dur(central)=dur2", "}", "cp=central/(v/1000)", sep = "\n")
    ncs <- c("ka", "cl"); pp <- c(ka = 1.5, cl = 1.1, v = 20)
    nex <- rxode2::.rxAdjointExpand(mt, ncs, stiff = TRUE); nmadj <- rxode2::rxode2(nex$text)
    mb <- rxode2::rxode2(mt)
    scol <- as.vector(outer(nex$st, ncs, function(s, pn) sprintf("rx__sens_%s_BY_%s__", s, pn)))
    # Pure stiff (Rosenbrock + fully-implicit RK) adjoint methods: ss dosing is
    # guarded on this family (the composite dop853s+ros4s covers stiff ss), so
    # only the NON-ss ids are checked here -- against central FD (method-agnostic).
    stiff <- c("ros4s", "ros6s", "ros43s", "geng5s", "gauss6s", "radauiia5s", "backwardEulers", "sdirk43s")
    tested <- 0
    for (id in unique(d0$id)) {
      di <- d0[d0$id == id, ]
      if (any(di$evid != 0 & !(di$cmt %in% c(1, 2)))) next
      if (any(di$ss != 0)) next                                    # ss guarded on pure stiff
      fd <- list()
      for (pn in ncs) {
        hh <- abs(pp[[pn]]) * 1e-6; up <- pp; dn <- pp; up[pn] <- up[pn] + hh; dn[pn] <- dn[pn] - hh
        sp <- as.data.frame(suppressWarnings(rxode2::rxSolve(mb, di, params = up, method = "lsoda", addlDropSs = TRUE, atol = 1e-11, rtol = 1e-11)))
        sm <- as.data.frame(suppressWarnings(rxode2::rxSolve(mb, di, params = dn, method = "lsoda", addlDropSs = TRUE, atol = 1e-11, rtol = 1e-11)))
        for (st in nex$st) fd[[sprintf("rx__sens_%s_BY_%s__", st, pn)]] <- (sp[[st]] - sm[[st]]) / (2 * hh)
      }
      for (meth in stiff) {
        a <- as.data.frame(suppressWarnings(rxode2::rxSolve(nmadj, di, params = pp, method = meth, addlDropSs = TRUE, atol = 1e-9, rtol = 1e-9)))
        mx <- 0; for (cn in scol) mx <- max(mx, max(abs(a[[cn]] - fd[[cn]]), na.rm = TRUE))
        ssc <- max(1, max(abs(unlist(a[scol])), na.rm = TRUE))
        expect_lt(mx / ssc, 5e-3)
        tested <- tested + 1
      }
    }
    expect_gt(tested, 100)                     # ~15 non-ss scenarios x 8 stiff solvers
  })

  test_that("STIFF adjoint solvers: steady-state sens match forward, using the ANALYTIC Jacobian", {
    skip_on_cran()
    skip_on_ci()
    # The stiff (Rosenbrock / implicit RK) adjoint solvers integrate the
    # adjoint-augmented system with an analytic Jacobian emitted by
    # .rxAdjointExpand(stiff=TRUE) -- df()/dy() of the base block (the sens states
    # have d/dt==0, so their rows/cols are zero).  Without it boost's rosenbrock4
    # (ros4s) cannot step the wide system at all (empty J -> "a new step size was
    # not found") and the others fall back to a numeric Jacobian.  This checks the
    # analytic-Jacobian path at steady state (bolus, fixed-rate infusion,
    # continuous, ss=2 superposition) against the forward-sensitivity reference.
    exS <- rxode2::.rxAdjointExpand(mText, cs, stiff = TRUE)
    madjS <- rxode2::rxode2(exS$text)
    mfwd <- rxode2::rxode2(mText, calcSens = cs)
    scol <- as.vector(outer(exS$st, cs, function(s, pn) sprintf("rx__sens_%s_BY_%s__", s, pn)))
    stiff <- c("ros4s", "ros6s", "ros43s", "radauiia5s", "gauss6s", "geng5s", "sdirk43s", "iiic6s")
    chkS <- function(ev, tol) {
      f <- as.data.frame(suppressWarnings(rxode2::rxSolve(mfwd, ev, params = p, method = "lsoda", cores = 1)))
      for (meth in stiff) {
        a <- as.data.frame(suppressWarnings(rxode2::rxSolve(madjS, ev, params = p, method = meth, cores = 1)))
        mx <- 0; for (cn in scol) mx <- max(mx, max(abs(a[[cn]] - f[[cn]]), na.rm = TRUE))
        expect_lt(mx / max(1, max(abs(unlist(f[scol])), na.rm = TRUE)), tol)
      }
    }
    # The ss monodromy IC is recorded AND transposed with the method's OWN stiff
    # stepper (rk4sSsIc precomputeStiff + ros/radauPeriodT), so the ss primal and
    # its sensitivity IC share one discretization -- tolerances are ~1e-6 here
    # (were ~1e-4 when the IC was recorded with a dop853 stand-in).
    chkS(et(amt = 100, cmt = "center", ss = 1, ii = 12) %>% et(c(4, 8, 12)), 1e-4)              # bolus ss=1
    chkS(et(amt = 100, rate = 10, cmt = "center", ss = 1, ii = 12) %>% et(c(4, 8, 12)), 1e-4)   # fixed-rate infusion ss=1
    chkS(et(amt = 0, rate = 10, cmt = "center", ss = 1) %>% et(c(4, 8, 12)), 1e-4)              # continuous ss=1
    chkS(et(amt = 100, cmt = "center") %>%
           et(amt = 50, cmt = "center", ss = 2, ii = 12, time = 12) %>% et(c(4, 12, 24)), 1e-4) # ss=2 superposition
  })

  test_that("rxSolveAdjointRk4 infers the stiff Jacobian from the method", {
    skip_on_cran()
    skip_on_ci()
    # .rxAdjointMethodStiff maps the adjoint method to whether the expansion needs
    # the analytic Jacobian (stiff Rosenbrock/implicit-RK + the composite).
    expect_true(.rxAdjointMethodStiff("ros4s"))
    expect_true(.rxAdjointMethodStiff("radauiia5s"))
    expect_true(.rxAdjointMethodStiff("dop853s+ros4s"))
    expect_false(.rxAdjointMethodStiff("rk4s"))
    expect_false(.rxAdjointMethodStiff("dop853s"))
    # the convenience wrapper auto-builds WITH the Jacobian for a stiff method and
    # WITHOUT for an explicit one -- both reproduce the forward sensitivities at ss.
    mtxt <- "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center"
    mfwd <- rxode2::rxode2(mtxt, calcSens = cs)
    evSS <- et(amt = 100, cmt = "center", ss = 1, ii = 12) %>% et(c(4, 8, 12))
    f <- as.data.frame(suppressWarnings(rxode2::rxSolve(mfwd, evSS, params = p, method = "lsoda")))
    scol <- as.vector(outer(c("depot", "center"), cs, function(s, pn) sprintf("rx__sens_%s_BY_%s__", s, pn)))
    for (meth in c("rk4s", "ros4s", "radauiia5s", "dop853s+ros4s")) {
      a <- suppressWarnings(rxode2::rxSolveAdjointRk4(mtxt, evSS, params = p, calcSens = cs, method = meth))
      expect_false(any(grepl("^rx__adj", names(a))))           # internal lhs dropped
      mx <- 0; for (cn in scol) mx <- max(mx, max(abs(a[[cn]] - f[[cn]]), na.rm = TRUE))
      expect_lt(mx / max(1, max(abs(unlist(f[scol])), na.rm = TRUE)), 3e-3)
    }
  })

  test_that("in-engine rk4s F/dose-jump: Fbio + all sens columns match FD of the RK4 solve", {
    skip_on_cran()
    skip_on_ci()
    fText <- "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\nf(depot)=Fbio"
    fcs <- c("ka", "cl", "v", "Fbio"); fp <- c(ka = 1.2, cl = 3.5, v = 25, Fbio = 0.7)
    tms <- c(1, 2, 4, 8, 12, 24)
    ev <- et(amt = 100, cmt = "depot") %>% et(tms)
    ex <- rxode2::.rxAdjointExpand(fText, fcs)
    madj <- rxode2::rxode2(ex$text)
    sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = fp, method = "rk4s", cores = 1))
    mbase <- rxode2::rxode2(fText)
    solveBase <- function(pp) as.matrix(as.data.frame(
      rxode2::rxSolve(mbase, ev, params = pp, method = "rk4", cores = 1))[, c("depot", "center")])
    # bioavailability preserved -> base states match the rk4 primal
    expect_lt(max(abs(as.matrix(sd[, c("depot", "center")]) - solveBase(fp))), 1e-8)
    fdmax <- 0
    for (pn in fcs) {
      hh <- fp[[pn]] * 1e-6; pp <- fp; pm <- fp; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
      fd <- (solveBase(pp) - solveBase(pm)) / (2 * hh)
      for (k in seq_along(ex$st))
        fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
    }
    expect_lt(fdmax, 1e-4)
    # the bioavailability gradient is genuinely nonzero
    expect_gt(max(abs(sd[["rx__sens_center_BY_Fbio__"]])), 1)
  })

  test_that("discrete-adjoint event jumps (reset/replace/multiply) match FD of the RK4 solve", {
    skip_on_cran()
    skip_on_ci()
    cs2 <- c("ka", "cl", "v"); p2 <- c(ka = 1.2, cl = 3.5, v = 25)
    ex <- rxode2::.rxAdjointExpand(mText, cs2)
    madj <- rxode2::rxode2(ex$text)
    mbase <- rxode2::rxode2(mText)
    solveBase <- function(pp, ev) as.matrix(as.data.frame(
      rxode2::rxSolve(mbase, ev, params = pp, method = "rk4", cores = 1))[, c("depot", "center")])
    evs <- list(
      reset    = et(amt = 100, cmt = "depot") %>% et(4, evid = 3) %>%
        et(amt = 100, cmt = "depot", time = 4) %>% et(c(1, 2, 6, 8, 12)),
      replace  = et(amt = 100, cmt = "depot") %>% et(amt = 40, cmt = "center", evid = 5, time = 4) %>%
        et(c(1, 2, 6, 8, 12)),
      multiply = et(amt = 100, cmt = "depot") %>% et(amt = 0.5, cmt = "center", evid = 6, time = 4) %>%
        et(c(1, 2, 6, 8, 12))
    )
    # reset/replace/multiply costate jumps: the discrete methods (rk4s/dop853s)
    # AND the continuous cvodesadj all carry them.
    meths <- c("rk4s", "dop853s", "cvodesadj")
    methForEv <- list(reset = meths, replace = meths, multiply = meths)
    for (nm in names(evs)) {
      ev <- evs[[nm]]
      for (meth in methForEv[[nm]]) {
        sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = p2, method = meth, cores = 1))
        fdmax <- 0
        for (pn in cs2) {
          hh <- p2[[pn]] * 1e-6; pp <- p2; pm <- p2; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
          fd <- (solveBase(pp, ev) - solveBase(pm, ev)) / (2 * hh)
          for (k in seq_along(ex$st))
            fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
        }
        # rk4s/dop853s match the RK4 map to ~FD precision; cvodesadj (CVODE forward)
        # differs from the rk4 reference at solver-tolerance level.
        expect_lt(fdmax, if (meth == "cvodesadj") 1e-3 else 1e-4)
      }
    }
    # bare reset (no redose): every post-reset sensitivity is identically 0
    ev0 <- et(amt = 100, cmt = "depot") %>% et(4, evid = 3) %>% et(c(1, 2, 6, 8, 12))
    for (meth in c("rk4s", "dop853s")) {
      s0 <- as.data.frame(rxode2::rxSolve(madj, ev0, params = p2, method = meth, cores = 1))
      post <- s0$time > 4
      for (pn in cs2) for (st in ex$st)
        expect_lt(max(abs(s0[[sprintf("rx__sens_%s_BY_%s__", st, pn)]][post])), 1e-8)
    }
  })

  test_that("adjoint expand preserves modeled alag/rate/dur (correct primal + ODE-param sens)", {
    skip_on_cran()
    skip_on_ci()
    # .rxAdjointExpand must keep modeled dosing modifiers so the adjoint forward
    # integrates the SAME primal as the real model; the ODE parameters (ka/cl/v)
    # then have correct sensitivities.  (The dosing PARAMETER's own sensitivity --
    # d/d tlag etc. -- needs the transversality jump term, a separate addition.)
    mods <- list(
      alag = list(txt = "alag(depot)=tlag", par = c(tlag = 0.7)),
      rate = list(txt = "rate(depot)=Rin",  par = c(Rin = 50)),
      dur  = list(txt = "dur(depot)=Dur",   par = c(Dur = 2))
    )
    for (nm in names(mods)) {
      mText2 <- paste0("d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\n", mods[[nm]]$txt)
      cs2 <- c("ka", "cl", "v", names(mods[[nm]]$par))
      p2 <- c(ka = 1.2, cl = 3.5, v = 25, mods[[nm]]$par)
      ex <- rxode2::.rxAdjointExpand(mText2, cs2)
      madj <- rxode2::rxode2(ex$text); mbase <- rxode2::rxode2(mText2)
      ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 2, 4, 8, 12))
      solveBase <- function(pp) as.matrix(as.data.frame(
        rxode2::rxSolve(mbase, ev, params = pp, method = "rk4", cores = 1))[, c("depot", "center")])
      s <- as.data.frame(rxode2::rxSolve(madj, ev, params = p2, method = "rk4s", cores = 1))
      # primal matches the real model
      expect_lt(max(abs(as.matrix(s[, c("depot", "center")]) - solveBase(p2))), 1e-6)
      # ODE-parameter (ka/cl/v) sensitivities match a central FD
      fdmax <- 0
      for (pn in c("ka", "cl", "v")) {
        hh <- p2[[pn]] * 1e-6; pp <- p2; pm <- p2; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveBase(pp) - solveBase(pm)) / (2 * hh)
        for (k in seq_along(ex$st))
          fdmax <- max(fdmax, max(abs(s[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
      }
      expect_lt(fdmax, 1e-4)
    }
  })

  test_that("modeled-alag transversality: d/d(lag) sensitivity matches FD (rk4s/dop853s)", {
    skip_on_cran()
    skip_on_ci()
    # A bolus into a compartment with a modeled alag() lands at t_dose+lag(theta);
    # shifting lag shifts the whole post-dose trajectory, so the lag parameter's
    # own sensitivity needs the transversality jump mu += -amt*dlag/dtheta*(lam^T Fx[:,c]).
    mText2 <- "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\nalag(depot)=tlag"
    cs2 <- c("ka", "cl", "v", "tlag"); p2 <- c(ka = 1.2, cl = 3.5, v = 25, tlag = 0.7)
    ex <- rxode2::.rxAdjointExpand(mText2, cs2)
    madj <- rxode2::rxode2(ex$text); mbase <- rxode2::rxode2(mText2)
    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 2, 4, 8, 12))
    solveBase <- function(pp, meth) as.matrix(as.data.frame(rxode2::rxSolve(
      mbase, ev, params = pp, method = if (meth == "cvodesadj") "liblsoda" else "rk4",
      atol = 1e-11, rtol = 1e-11, cores = 1))[, c("depot", "center")])
    for (meth in c("rk4s", "dop853s", "radauiia5s", "ros4s", "cvodesadj")) {
      exM <- if (meth %in% c("radauiia5s", "ros4s")) rxode2::.rxAdjointExpand(mText2, cs2, stiff = TRUE) else ex
      madjM <- if (identical(exM, ex)) madj else rxode2::rxode2(exM$text)
      s <- as.data.frame(rxode2::rxSolve(madjM, ev, params = p2, method = meth,
                                         atol = 1e-11, rtol = 1e-11, cores = 1))
      # every parameter, including the lag time tlag
      fdmax <- 0
      for (pn in cs2) {
        hh <- p2[[pn]] * 1e-6; pp <- p2; pm <- p2; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveBase(pp, meth) - solveBase(pm, meth)) / (2 * hh)
        for (k in seq_along(exM$st))
          fdmax <- max(fdmax, max(abs(s[[sprintf("rx__sens_%s_BY_%s__", exM$st[k], pn)]] - fd[, k])))
      }
      expect_lt(fdmax, if (meth == "cvodesadj") 1e-3 else 1e-4)
      # the lag-time gradient is genuinely nonzero
      expect_gt(max(abs(s[["rx__sens_center_BY_tlag__"]])), 1)
    }
  })

  test_that("modeled-rate infusion dual: d/d(rate) sensitivity matches FD (rk4s/dop853s)", {
    skip_on_cran()
    skip_on_ci()
    # A modeled-rate infusion of amt into cmt c runs over [t_on, t_on+amt/R(theta)]
    # adding R to the RHS; the rate parameter needs the forcing quadrature plus the
    # off-boundary transversality.  Rin=40 => off-time 100/40=2.5 is OFF the obs grid
    # (observing exactly at the off-boundary is a measure-zero FD kink).
    mText2 <- "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\nrate(depot)=Rin"
    cs2 <- c("ka", "cl", "v", "Rin"); p2 <- c(ka = 1.2, cl = 3.5, v = 25, Rin = 40)
    ex <- rxode2::.rxAdjointExpand(mText2, cs2)
    madj <- rxode2::rxode2(ex$text); mbase <- rxode2::rxode2(mText2)
    ev <- et(amt = 100, cmt = "depot", rate = -1) %>% et(c(1, 2, 4, 8, 12))
    solveBase <- function(pp, meth) as.matrix(as.data.frame(rxode2::rxSolve(
      mbase, ev, params = pp, method = if (meth == "cvodesadj") "liblsoda" else "rk4",
      atol = 1e-11, rtol = 1e-11, cores = 1))[, c("depot", "center")])
    for (meth in c("rk4s", "dop853s", "radauiia5s", "ros4s", "cvodesadj")) {
      exM <- if (meth %in% c("radauiia5s", "ros4s")) rxode2::.rxAdjointExpand(mText2, cs2, stiff = TRUE) else ex
      madjM <- if (identical(exM, ex)) madj else rxode2::rxode2(exM$text)
      s <- as.data.frame(rxode2::rxSolve(madjM, ev, params = p2, method = meth,
                                         atol = 1e-11, rtol = 1e-11, cores = 1))
      fdmax <- 0
      for (pn in cs2) {
        hh <- p2[[pn]] * 1e-6; pp <- p2; pm <- p2; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveBase(pp, meth) - solveBase(pm, meth)) / (2 * hh)
        for (k in seq_along(exM$st))
          fdmax <- max(fdmax, max(abs(s[[sprintf("rx__sens_%s_BY_%s__", exM$st[k], pn)]] - fd[, k])))
      }
      expect_lt(fdmax, if (meth == "cvodesadj") 1e-3 else 1e-4)
      expect_gt(max(abs(s[["rx__sens_center_BY_Rin__"]])), 0.5)
    }
  })

  test_that("modeled-dur infusion dual: d/d(dur) sensitivity matches FD (all methods)", {
    skip_on_cran()
    skip_on_ci()
    # A modeled-dur infusion has effective rate amt/dur, so dR/dtheta = amt *
    # d(1/dur)/dtheta -- same two-term dual as rate() with a durMult=amt runtime factor.
    # Dur=2.5 => off-time t_on+2.5 is OFF the obs grid (avoids the FD kink).
    mText2 <- "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\ndur(depot)=Dur"
    cs2 <- c("ka", "cl", "v", "Dur"); p2 <- c(ka = 1.2, cl = 3.5, v = 25, Dur = 2.5)
    ex <- rxode2::.rxAdjointExpand(mText2, cs2)
    madj <- rxode2::rxode2(ex$text); mbase <- rxode2::rxode2(mText2)
    ev <- et(amt = 100, cmt = "depot", rate = -2) %>% et(c(1, 2, 4, 8, 12))
    solveBase <- function(pp, meth) as.matrix(as.data.frame(rxode2::rxSolve(
      mbase, ev, params = pp, method = if (meth == "cvodesadj") "liblsoda" else "rk4",
      atol = 1e-11, rtol = 1e-11, cores = 1))[, c("depot", "center")])
    for (meth in c("rk4s", "dop853s", "radauiia5s", "ros4s", "cvodesadj")) {
      exM <- if (meth %in% c("radauiia5s", "ros4s")) rxode2::.rxAdjointExpand(mText2, cs2, stiff = TRUE) else ex
      madjM <- if (identical(exM, ex)) madj else rxode2::rxode2(exM$text)
      s <- as.data.frame(rxode2::rxSolve(madjM, ev, params = p2, method = meth,
                                         atol = 1e-11, rtol = 1e-11, cores = 1))
      fdmax <- 0
      for (pn in cs2) {
        hh <- p2[[pn]] * 1e-6; pp <- p2; pm <- p2; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveBase(pp, meth) - solveBase(pm, meth)) / (2 * hh)
        for (k in seq_along(exM$st))
          fdmax <- max(fdmax, max(abs(s[[sprintf("rx__sens_%s_BY_%s__", exM$st[k], pn)]] - fd[, k])))
      }
      expect_lt(fdmax, if (meth == "cvodesadj") 1e-3 else 1e-4)
      expect_gt(max(abs(s[["rx__sens_center_BY_Dur__"]])), 0.5)
    }
  })

  test_that("in-engine rk4s population solve (parallel) matches per-subject FD", {
    skip_on_cran()
    skip_on_ci()
    cs2 <- c("ka", "cl", "v")
    tms <- c(1, 4, 12, 24)
    pars <- data.frame(id = 1:3, ka = c(1.0, 1.2, 1.5), cl = c(3.0, 3.5, 4.0), v = c(20, 25, 30))
    ev <- do.call(rbind, lapply(1:3, function(i) {
      e <- et(amt = 100, cmt = "depot") %>% et(tms); e$id <- i; e
    }))
    ex <- rxode2::.rxAdjointExpand(mText, cs2)
    madj <- rxode2::rxode2(ex$text)
    s <- as.data.frame(rxode2::rxSolve(madj, ev, pars, method = "rk4s", cores = 2))
    mbase <- rxode2::rxode2(mText)
    maxerr <- 0
    for (i in 1:3) {
      pi <- c(ka = pars$ka[i], cl = pars$cl[i], v = pars$v[i])
      evi <- et(amt = 100, cmt = "depot") %>% et(tms)
      solveB <- function(pp) as.matrix(as.data.frame(
        rxode2::rxSolve(mbase, evi, params = pp, method = "rk4", cores = 1))[, c("depot", "center")])
      si <- s[s$id == i, ]
      for (pn in cs2) {
        hh <- pi[[pn]] * 1e-6; pp <- pi; pm <- pi; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveB(pp) - solveB(pm)) / (2 * hh)
        for (k in seq_along(ex$st))
          maxerr <- max(maxerr, max(abs(si[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
      }
    }
    expect_lt(maxerr, 1e-4)
  })


  test_that("rxSolveAdjointRk4 wrapper: clean full-trajectory gradient + cache", {
    skip_on_cran()
    skip_on_ci()
    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 4, 12, 24))
    df <- rxode2::rxSolveAdjointRk4(mText, ev, params = p, calcSens = cs)
    # internal adjoint lhs are dropped; rx__sens_* are present
    expect_false(any(grepl("^rx__adj(FX|FP|dF)_", names(df))))
    expect_true(all(vapply(cs, function(pn) all(vapply(c("depot", "center"), function(st)
      sprintf("rx__sens_%s_BY_%s__", st, pn) %in% names(df), logical(1))), logical(1))))
    # wrapper full-trajectory columns equal a direct rk4s solve
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    direct <- as.data.frame(rxode2::rxSolve(rxode2::rxode2(ex$text), ev, params = p, method = "rk4s", cores = 1))
    for (pn in cs) for (st in c("depot", "center")) {
      cn <- sprintf("rx__sens_%s_BY_%s__", st, pn)
      expect_equal(df[[cn]], direct[[cn]], tolerance = 1e-10)
    }
    # cache returns the identical compiled object
    expect_identical(rxode2::.rxAdjointModel(mText, cs)$model,
                     rxode2::.rxAdjointModel(mText, cs)$model)
  })

  test_that("stiff radauiia5s (fully-implicit Radau IIA) adjoint is exact (linear + nonlinear)", {
    skip_on_cran()
    skip_on_ci()
    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 4, 12, 24))
    chk <- function(mt, ccs, pp0, hmin) {
      ex <- rxode2::.rxAdjointExpand(mt, ccs)  # standard model: implicit RK needs only F_X/F_p
      madj <- rxode2::rxode2(ex$text)
      sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = pp0, method = "radauiia5s", hmin = hmin, cores = 1))
      solveB <- function(pp) as.matrix(as.data.frame(rxode2::rxSolve(
        madj, ev, params = pp, method = "radauiia5s", hmin = hmin, cores = 1))[, c("depot", "center")])
      fdmax <- 0
      for (pn in ccs) {
        hh <- pp0[[pn]] * 1e-6; pp <- pp0; pm <- pp0; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveB(pp) - solveB(pm)) / (2 * hh)
        for (kk in seq_along(ex$st))
          fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[kk], pn)]] - fd[, kk])))
      }
      expect_lt(fdmax, 1e-5)
    }
    chk(mText, cs, p, 0.1)
    chk("d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot - vmax*center/(km+center)",
        c("ka", "vmax", "km"), c(ka = 1.0, vmax = 8, km = 15), 0.05)
  })

  test_that("fully-implicit backwardEulers + gauss6s adjoints are exact vs FD", {
    skip_on_cran()
    skip_on_ci()
    # More fully-implicit RK methods on the Radau framework (first derivatives
    # only, no f'').  backwardEulers = implicit Euler (order 1, L-stable);
    # gauss6s = 3-stage Gauss-Legendre (order 6, symplectic, NOT stiffly accurate
    # -- the coupled-Newton framework allows b != last A row).  Both give an
    # EXACT discrete adjoint even where the primal is coarse (backwardEuler).
    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 4, 12, 24))
    chk <- function(method, hmin) {
      ex <- rxode2::.rxAdjointExpand(mText, cs)
      madj <- rxode2::rxode2(ex$text)
      sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = method, hmin = hmin, cores = 1))
      solveB <- function(pp) as.matrix(as.data.frame(rxode2::rxSolve(
        madj, ev, params = pp, method = method, hmin = hmin, cores = 1))[, c("depot", "center")])
      fdmax <- 0
      for (pn in cs) {
        hh <- p[[pn]] * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveB(pp) - solveB(pm)) / (2 * hh)
        for (kk in seq_along(ex$st))
          fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[kk], pn)]] - fd[, kk])))
      }
      expect_lt(fdmax, 1e-5)
    }
    chk("backwardEulers", 0.005)
    chk("gauss6s", 0.05)
    chk("sdirk43s", 0.01)   # SDIRK (lower-triangular A) on the same coupled framework
    chk("iiic6s", 0.05)     # Lobatto IIIC 4-stage 6th order
    chk("geng5s", 0.05)     # Geng5 3-stage 5th order (same b/c as Radau5, different A)
  })

  test_that("stiff ros4s (Rosenbrock) adjoint is exact for a linear model (dJ/dp term)", {
    skip_on_cran()
    skip_on_ci()
    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 4, 12, 24))
    ex <- rxode2::.rxAdjointExpand(mText, cs, stiff = TRUE)
    expect_true(ex$stiff); expect_gt(ex$jpOff, 0)
    madj <- rxode2::rxode2(ex$text)
    sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = "ros4s", hmin = 0.01, cores = 1))
    solveB <- function(pp) as.matrix(as.data.frame(rxode2::rxSolve(
      madj, ev, params = pp, method = "ros4s", hmin = 0.01, cores = 1))[, c("depot", "center")])
    fdmax <- 0
    for (pn in cs) {
      hh <- p[[pn]] * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
      fd <- (solveB(pp) - solveB(pm)) / (2 * hh)
      for (k in seq_along(ex$st))
        fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
    }
    expect_lt(fdmax, 1e-5)
  })

  test_that("stiff ros4s adjoint is exact for a NONLINEAR (Michaelis-Menten) model (f'' term)", {
    skip_on_cran()
    skip_on_ci()
    nlText <- "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot - vmax*center/(km+center)"
    ncs <- c("ka", "vmax", "km"); np9 <- c(ka = 1.0, vmax = 8, km = 15)
    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 4, 12, 24))
    ex <- rxode2::.rxAdjointExpand(nlText, ncs, stiff = TRUE)
    expect_gt(ex$jyOff, ex$jpOff)
    madj <- rxode2::rxode2(ex$text)
    sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = np9, method = "ros4s", hmin = 0.005, cores = 1))
    solveB <- function(pp) as.matrix(as.data.frame(rxode2::rxSolve(
      madj, ev, params = pp, method = "ros4s", hmin = 0.005, cores = 1))[, c("depot", "center")])
    fdmax <- 0
    for (pn in ncs) {
      hh <- np9[[pn]] * 1e-6; pp <- np9; pm <- np9; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
      fd <- (solveB(pp) - solveB(pm)) / (2 * hh)
      for (k in seq_along(ex$st))
        fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
    }
    expect_lt(fdmax, 1e-5)
  })

  test_that("stiff ros43s (GRK4A) + ros6s (ROW6A) adjoints are exact (linear + nonlinear)", {
    skip_on_cran()
    skip_on_ci()

    # More Rosenbrock methods on ros_backward_fill.  Each libode ROW-form set is
    # transformed to the Hairer-Wanner form this framework uses (GRK4A via a Ginv
    # transform; ROW6A via a plain 1/gamma scaling).  Primal-vs-liblsoda validates
    # the transform; reuses the dJ/dp (+ dJ/dy=f'') stiff-adjoint terms.
    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 4, 12, 24))
    chk <- function(method, mt, ccs, pp0, hmin) {
      ex <- rxode2::.rxAdjointExpand(mt, ccs, stiff = TRUE)
      madj <- rxode2::rxode2(ex$text)
      sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = pp0, method = method, hmin = hmin, cores = 1))
      solveB <- function(pp) as.matrix(as.data.frame(rxode2::rxSolve(
        madj, ev, params = pp, method = method, hmin = hmin, cores = 1))[, c("depot", "center")])
      fdmax <- 0
      for (pn in ccs) {
        hh <- pp0[[pn]] * 1e-6; pp <- pp0; pm <- pp0; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveB(pp) - solveB(pm)) / (2 * hh)
        for (kk in seq_along(ex$st))
          fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[kk], pn)]] - fd[, kk])))
      }
      expect_lt(fdmax, 1e-5)
    }
    nl <- "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot - vmax*center/(km+center)"
    nlcs <- c("ka", "vmax", "km"); nlp <- c(ka = 1.0, vmax = 8, km = 15)
    for (m in c("ros43s", "ros6s")) {
      chk(m, mText, cs, p, 0.01)
      chk(m, nl, nlcs, nlp, 0.005)
    }
  })

  test_that("DDE delay() adjoint converges to FD for explicit rk4s methods", {
    skip_on_cran()
    skip_on_ci()

    # The anticipating costate term (delayed Jacobian F_Xd) makes the backward
    # sweep account for the delayed-state coupling.  It is a continuous-adjoint
    # discretization -> converges to FD at the scheme rate (O(h)), not machine
    # exact, so we assert monotone convergence as the step shrinks.
    mTextD <- "d/dt(central) = -k * delay(central, tau)"
    exD <- rxode2::.rxAdjointExpand(mTextD, c("k"))
    expect_true(exD$hasDelay); expect_gt(exD$fxdOff, 0); expect_gt(exD$tauOff, exD$fxdOff)
    madjD <- rxode2::rxode2(exD$text)
    pD <- c(k = 0.3, tau = 2); evD <- et(amt = 10, cmt = "central") %>% et(seq(0, 20, by = 2))
    errAt <- function(h) {
      sc <- as.data.frame(rxode2::rxSolve(madjD, evD, params = pD, method = "rk4s", hmin = h, cores = 1))
      sb <- function(pp) as.data.frame(rxode2::rxSolve(madjD, evD, params = pp, method = "rk4s", hmin = h, cores = 1))$central
      hh <- pD[["k"]] * 1e-6; pp <- pD; pm <- pD; pp["k"] <- pp["k"] + hh; pm["k"] <- pm["k"] - hh
      max(abs(sc[["rx__sens_central_BY_k__"]] - (sb(pp) - sb(pm)) / (2 * hh)))
    }
    e1 <- errAt(0.05); e2 <- errAt(0.02); e3 <- errAt(0.005)
    expect_lt(e2, e1); expect_lt(e3, e2)   # converges as h shrinks
    expect_lt(e3, 0.05)                    # and is small at the finest step
  })

  test_that("param-dependent delay tau(p): adjoint d/dtau matches FD incl. dose-induced jump", {
    skip_on_cran()
    skip_on_ci()

    # F_p gets the smooth breaking-point correction -(F_Xd_ij)*rxDelayD(y_j,tau)*
    # dtau/dp, AND the backward sweep adds the dose-induced breaking-point JUMP
    # mu += -lam_i(t_dose+tau)*F_Xd_ij*[y_j]*dtau/dp.  Together the adjoint d/dtau
    # matches finite differences (which forward-sens' smooth-only term does not).
    exD <- rxode2::.rxAdjointExpand("d/dt(central) = -k * delay(central, tau)", c("k", "tau"))
    expect_gt(exD$dtauOff, exD$tauOff)
    madjD <- rxode2::rxode2(exD$text)
    p <- c(k = 0.3, tau = 2)
    # observations OFF the breaking points (2,4,6,...) so central FD does not
    # straddle a jump (at a breaking point the adjoint gives the one-sided limit).
    ev <- et(amt = 10, cmt = "central") %>% et(seq(0.7, 15, by = 1.7))
    errAt <- function(h) {
      sc <- as.data.frame(rxode2::rxSolve(madjD, ev, params = p, method = "rk4s", hmin = h, cores = 1))
      sb <- function(pp) as.data.frame(rxode2::rxSolve(madjD, ev, params = pp, method = "rk4s", hmin = h, cores = 1))$central
      hh <- p[["tau"]] * 5e-4; pp <- p; pm <- p; pp["tau"] <- pp["tau"] + hh; pm["tau"] <- pm["tau"] - hh
      max(abs(sc[["rx__sens_central_BY_tau__"]] - (sb(pp) - sb(pm)) / (2 * hh)))
    }
    e1 <- errAt(0.01); e2 <- errAt(0.002)
    expect_lt(e2, e1)      # converges as the step shrinks (jump location resolves)
    expect_lt(e2, 0.05)    # and is small at the fine step
  })

  test_that("DDE delay() adjoint converges for stiff (ros4s) and composite methods", {
    skip_on_cran()
    skip_on_ci()

    # The anticipating term is a per-step costate injection shared by every
    # backward fill (explicit, Rosenbrock, Radau, composite), so all of them
    # converge to FD as the step shrinks.
    exD <- rxode2::.rxAdjointExpand("d/dt(central) = -k * delay(central, tau)", c("k"), stiff = TRUE)
    madjD <- rxode2::rxode2(exD$text)
    pD <- c(k = 0.3, tau = 2); evD <- et(amt = 10, cmt = "central") %>% et(seq(0, 20, by = 2))
    errAt <- function(method, ctl) {
      sc <- as.data.frame(do.call(rxode2::rxSolve, c(list(madjD, evD, params = pD, method = method, cores = 1), ctl)))
      sb <- function(pp) as.data.frame(do.call(rxode2::rxSolve, c(list(madjD, evD, params = pp, method = method, cores = 1), ctl)))$central
      hh <- pD[["k"]] * 1e-6; pp <- pD; pm <- pD; pp["k"] <- pp["k"] + hh; pm["k"] <- pm["k"] - hh
      max(abs(sc[["rx__sens_central_BY_k__"]] - (sb(pp) - sb(pm)) / (2 * hh)))
    }
    expect_lt(errAt("ros4s", list(hmin = 0.01)), errAt("ros4s", list(hmin = 0.05)))
    expect_lt(errAt("ros4s", list(hmin = 0.01)), 0.15)
    expect_lt(errAt("dop853s+ros4s", list(hmax = 0.02, atol = 1e-9, rtol = 1e-9)), 0.15)
  })

  test_that("composite AutoSwitch dop853s+ros4s adjoint is exact across stiffness regimes", {
    skip_on_cran()
    skip_on_ci()

    # The composite freezes the per-interval switch decision and transposes each
    # step with the method (explicit dop853s OR Rosenbrock ros4s) that ran it.
    mTextC <- "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot - ke*center"
    csC <- c("ka", "ke"); pC <- c(ka = 80, ke = 0.3)
    exC <- rxode2::.rxAdjointExpand(mTextC, csC, stiff = TRUE)
    madjC <- rxode2::rxode2(exC$text)
    chk <- function(times) {
      ev <- et(amt = 100, cmt = "depot") %>% et(times)
      sc <- as.data.frame(rxode2::rxSolve(madjC, ev, params = pC, method = "dop853s+ros4s", hmin = 0.002, cores = 1))
      sl <- as.data.frame(rxode2::rxSolve(madjC, ev, params = pC, method = "liblsoda", cores = 1))
      expect_lt(max(abs(sc$center - sl$center), abs(sc$depot - sl$depot)), 1e-3)  # primal
      solveB <- function(pp) as.matrix(as.data.frame(rxode2::rxSolve(
        madjC, ev, params = pp, method = "dop853s+ros4s", hmin = 0.002, cores = 1))[, c("depot", "center")])
      fdmax <- 0
      for (pn in csC) {
        hh <- pC[[pn]] * 1e-6; pp <- pC; pm <- pC; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveB(pp) - solveB(pm)) / (2 * hh)
        for (k in seq_along(exC$st))
          fdmax <- max(fdmax, max(abs(sc[[sprintf("rx__sens_%s_BY_%s__", exC$st[k], pn)]] - fd[, k])))
      }
      expect_lt(fdmax, 1e-4)
    }
    chk(seq(3, 24, length.out = 6))     # all intervals stiff  -> all ros4s steps
    chk(c(0.02, 0.04, 0.06))            # all intervals nonstiff -> all dop853s steps
    chk(seq(0.05, 24, length.out = 8))  # mixed: early stiff (ros4s), late nonstiff (dop853s)
  })

  test_that("composite AutoSwitch requires 's' on both methods to be a sensitivity method", {
    skip_on_cran()
    skip_on_ci()
    # dop853+ros4 (no 's') is a plain forward composite, not an adjoint method.
    expect_false(grepl("dop853s", "dop853+ros4"))
  })

  test_that("adaptive methods: primal matches liblsoda and frozen-step adjoint matches FD", {
    skip_on_cran()
    skip_on_ci()

    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 4, 12, 24))
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text)
    mbase <- rxode2::rxode2(mText)
    ref <- as.data.frame(rxode2::rxSolve(mbase, ev, params = p, method = "liblsoda",
                                         atol = 1e-11, rtol = 1e-11))
    # per-method (tolerance, primal-tol) -- lower-order pairs need looser primal
    cfg <- list(dop5s = c(1e-11, 1e-5), dop853s = c(1e-11, 1e-5),
                ck54s = c(1e-11, 1e-5), bs32s = c(1e-8, 1e-4),
                rk43s = c(1e-9, 1e-4), vern65s = c(1e-11, 1e-5),
                vern76s = c(1e-11, 1e-5),
                dop87s = c(1e-11, 1e-5), f78s = c(1e-11, 1e-5))
    for (meth in names(cfg)) {
      tl <- cfg[[meth]][1]; ptol <- cfg[[meth]][2]
      sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = meth,
                                          atol = tl, rtol = tl, cores = 1))
      expect_lt(max(abs(as.matrix(sd[, c("depot", "center")]) - as.matrix(ref[, c("depot", "center")]))), ptol)
      solveB <- function(pp) as.matrix(as.data.frame(rxode2::rxSolve(
        madj, ev, params = pp, method = meth, atol = tl, rtol = tl, cores = 1))[, c("depot", "center")])
      fdmax <- 0
      for (pn in cs) {
        hh <- p[[pn]] * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveB(pp) - solveB(pm)) / (2 * hh)
        for (k in seq_along(ex$st))
          fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
      }
      expect_lt(fdmax, 1e-4)
    }
  })

  test_that("table-driven fixed-step methods (eulers/midpoints/heuns/rk3s) match FD of their own map", {
    skip_on_cran()
    skip_on_ci()

    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 4, 12, 24))
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text)
    for (meth in c("eulers", "midpoints", "heuns", "rk3s")) {
      sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = meth, cores = 1))
      solveB <- function(pp) as.matrix(as.data.frame(
        rxode2::rxSolve(madj, ev, params = pp, method = meth, cores = 1))[, c("depot", "center")])
      fdmax <- 0
      for (pn in cs) {
        hh <- p[[pn]] * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveB(pp) - solveB(pm)) / (2 * hh)
        for (k in seq_along(ex$st))
          fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
      }
      expect_lt(fdmax, 1e-4)
    }
  })

  test_that("discrete forward sensitivity agrees with a finite difference of the RK4 solve", {
    skip_on_cran()
    skip_on_ci()

    solveN <- function(pp) {
      X <- X0
      for (n in seq_len(nStep)) X <- .rxRk4Step(B, X, pp, h)$Xnext
      X
    }
    for (pn in cs) {
      hh <- p[[pn]] * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
      fd <- (solveN(pp) - solveN(pm)) / (2 * hh)
      an <- fwd$SN[, which(cs == pn)]
      expect_equal(unname(an), unname(fd), tolerance = 1e-5)
    }
  })

  # ---- CVODES adjoint (method="cvodesadj"): self-managed dense primal + plain
  # CVODES backward.  Handles every event type (unlike native ASA), so it is
  # validated against a central FD of the base (liblsoda) solve across single-dose,
  # multi-dose and infusion event tables. ----
  test_that("in-engine cvodesadj fills rx__sens_* matching FD across dose types", {
    skip_on_cran()
    skip_on_ci()

    cs2 <- c("ka", "cl", "v"); p2 <- c(ka = 1.2, cl = 3.5, v = 25)
    ex <- rxode2::.rxAdjointExpand(mText, cs2)
    madj <- rxode2::rxode2(ex$text)
    mbase <- rxode2::rxode2(mText)
    evs <- list(
      single   = et(amt = 100, cmt = "depot") %>% et(c(1, 2, 4, 8, 12, 24)),
      multi    = et(amt = 100, cmt = "depot", ii = 12, addl = 3) %>% et(c(1, 6, 13, 18, 25, 36, 48)),
      infusion = et(amt = 100, cmt = "center", rate = 10) %>% et(c(1, 4, 8, 12, 20))
    )
    for (nm in names(evs)) {
      ev <- evs[[nm]]
      solveBase <- function(pp) as.matrix(as.data.frame(rxode2::rxSolve(
        mbase, ev, params = pp, method = "liblsoda", atol = 1e-11, rtol = 1e-11, cores = 1))[, c("depot", "center")])
      sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = p2, method = "cvodesadj",
                                          atol = 1e-11, rtol = 1e-11, cores = 1))
      # base states match the reference solve
      expect_lt(max(abs(as.matrix(sd[, c("depot", "center")]) - solveBase(p2))), 1e-6)
      fdmax <- 0
      for (pn in cs2) {
        hh <- p2[[pn]] * 1e-6; pp <- p2; pm <- p2; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveBase(pp) - solveBase(pm)) / (2 * hh)
        for (k in seq_along(ex$st))
          fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
      }
      expect_lt(fdmax, 1e-4)
    }
  })

  test_that("cvodesadj reset (evid 3) costate jump: post-reset sens matches FD", {
    skip_on_cran()
    skip_on_ci()

    cs2 <- c("ka", "cl", "v"); p2 <- c(ka = 1.2, cl = 3.5, v = 25)
    ex <- rxode2::.rxAdjointExpand(mText, cs2)
    madj <- rxode2::rxode2(ex$text)
    mbase <- rxode2::rxode2(mText)
    # bare reset at t=4: the state is wiped to 0 (no redose), so every post-reset
    # sensitivity must be identically 0 (the costate zeroes at the reset).
    ev0 <- et(amt = 100, cmt = "depot") %>% et(4, evid = 3) %>% et(c(1, 2, 6, 8, 12))
    s0 <- as.data.frame(rxode2::rxSolve(madj, ev0, params = p2, method = "cvodesadj",
                                        atol = 1e-11, rtol = 1e-11, cores = 1))
    post <- s0$time > 4
    for (pn in cs2) for (st in ex$st)
      expect_lt(max(abs(s0[[sprintf("rx__sens_%s_BY_%s__", st, pn)]][post])), 1e-8)
    # reset + redose at t=4: after the reset the system is a fresh dose, so the
    # sensitivity must match a central FD of the base solve (dop853s does NOT get
    # this -- the reset costate jump is what cvodesadj adds).
    ev <- et(amt = 100, cmt = "depot") %>% et(4, evid = 3) %>%
      et(amt = 100, cmt = "depot", time = 4) %>% et(c(1, 2, 6, 8, 12))
    solveBase <- function(pp) as.matrix(as.data.frame(rxode2::rxSolve(
      mbase, ev, params = pp, method = "liblsoda", atol = 1e-11, rtol = 1e-11, cores = 1))[, c("depot", "center")])
    sd <- as.data.frame(rxode2::rxSolve(madj, ev, params = p2, method = "cvodesadj",
                                        atol = 1e-11, rtol = 1e-11, cores = 1))
    fdmax <- 0
    for (pn in cs2) {
      hh <- p2[[pn]] * 1e-6; pp <- p2; pm <- p2; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
      fd <- (solveBase(pp) - solveBase(pm)) / (2 * hh)
      for (k in seq_along(ex$st))
        fdmax <- max(fdmax, max(abs(sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
    }
    expect_lt(fdmax, 1e-4)
  })

  test_that("cvodesadj population solve matches the dop853s discrete adjoint", {
    skip_on_cran()
    skip_on_ci()

    cs2 <- c("ka", "cl", "v")
    ex <- rxode2::.rxAdjointExpand(mText, cs2)
    madj <- rxode2::rxode2(ex$text)
    pars <- data.frame(id = 1:3, ka = c(1.0, 1.2, 1.5), cl = c(3.0, 3.5, 4.0), v = c(20, 25, 30))
    ev <- do.call(rbind, lapply(1:3, function(i) {
      e <- et(amt = 100, cmt = "depot", ii = 12, addl = 1) %>% et(c(1, 4, 12, 20)); e$id <- i; e
    }))
    sA <- as.data.frame(rxode2::rxSolve(madj, ev, pars, method = "cvodesadj", atol = 1e-11, rtol = 1e-11, cores = 2))
    sB <- as.data.frame(rxode2::rxSolve(madj, ev, pars, method = "dop853s", atol = 1e-11, rtol = 1e-11, cores = 2))
    for (pn in cs2) for (st in ex$st) {
      col <- sprintf("rx__sens_%s_BY_%s__", st, pn)
      expect_lt(max(abs(sA[[col]] - sB[[col]])), 1e-5)
    }
  })

  # Adjoint auto-switch: solving an adjoint-expanded model with a BASE method
  # upgrades it to the adjoint variant (base + 200): rk4->rk4s, dop853->dop853s,
  # cvode->cvodesadj; a method without a direct variant (liblsoda) falls back to
  # dop853s.  Without the switch the rx__sens_* compartments (d/dt=0) integrate to
  # all-zero.
  test_that("adjoint auto-switch: base methods produce real sensitivities (cvode->cvodesadj)", {
    skip_on_cran()
    skip_on_ci()

    cs2 <- c("ka", "cl", "v"); p2 <- c(ka = 1.2, cl = 3.5, v = 25)
    ex <- rxode2::.rxAdjointExpand(mText, cs2)
    madj <- rxode2::rxode2(ex$text)
    mbase <- rxode2::rxode2(mText)
    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 4, 12, 24))
    solveBase <- function(pp) as.matrix(as.data.frame(rxode2::rxSolve(
      mbase, ev, params = pp, method = "liblsoda", atol = 1e-11, rtol = 1e-11, cores = 1))[, c("depot", "center")])
    fdref <- list()
    for (pn in cs2) {
      hh <- p2[[pn]] * 1e-6; pp <- p2; pm <- p2; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
      fd <- (solveBase(pp) - solveBase(pm)) / (2 * hh)
      for (k in seq_along(ex$st)) fdref[[paste0(k, pn)]] <- fd[, k]
    }
    for (meth in c("cvode", "rk4", "dop853", "liblsoda")) {
      s <- as.data.frame(rxode2::rxSolve(madj, ev, params = p2, method = meth,
                                         atol = 1e-11, rtol = 1e-11, cores = 1))
      m <- 0
      for (pn in cs2) for (k in seq_along(ex$st))
        m <- max(m, max(abs(s[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fdref[[paste0(k, pn)]])))
      expect_lt(m, 1e-4)
      # genuinely non-zero (the switch actually happened; not silent zeros)
      expect_gt(max(abs(s[["rx__sens_center_BY_cl__"]])), 1e-3)
    }
    # rxIsStiff classifies cvodesadj like cvode (both CVODE/BDF)
    expect_true(rxode2::rxIsStiff("cvodesadj"))
    expect_false(rxode2::rxIsNonStiff("cvodesadj"))
  })

  test_that("in-engine liblsodaadj (P1-P3) == discrete forward sensitivity of liblsoda's own step map", {
    skip_on_cran()
    skip_on_ci()

    # liblsodaadj is the EXACT reverse-mode transpose of liblsoda's Nordsieck
    # multistep step map, incl. variable ORDER (Nordsieck row add/drop) and variable
    # STEP (scaleh at the boundary + a bridge for the rejection rescaling).  Because
    # liblsoda's order/step schedule is param-ADAPTIVE, a finite difference of a
    # re-solved liblsoda moves the schedule and only agrees for params that do not
    # perturb it -- so the correctness test is a C-side self-check with two parts:
    #   (A) adjoint == discrete FORWARD sensitivity of the SAME recorded schedule to
    #       machine precision  (the defining property: an exact transpose), and
    #   (B) a PRIMAL replay of the same step-map reproduces liblsoda's recorded y to
    #       the corrector convergence level (O(tol))  (the model matches liblsoda).
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text)
    mbase <- rxode2::rxode2(mText)
    tms <- c(1, 2, 4, 8, 12, 24)
    ev <- et(amt = 100, cmt = "depot") %>% et(tms)

    op0 <- Sys.getenv("RX_LSADJ_SELFCHECK", unset = NA)
    on.exit(if (is.na(op0)) Sys.unsetenv("RX_LSADJ_SELFCHECK") else Sys.setenv(RX_LSADJ_SELFCHECK = op0), add = TRUE)
    selfcheck <- function(params, ctl, tolPrimal, minSwitch = 0L) {
      Sys.setenv(RX_LSADJ_SELFCHECK = "1")
      on.exit(Sys.unsetenv("RX_LSADJ_SELFCHECK"), add = TRUE)  # scope to this solve only
      msg <- capture.output(
        invisible(do.call(rxode2::rxSolve, c(list(madj, ev, params = params, method = "liblsodaadj", cores = 1), ctl))),
        type = "message")
      line <- grep("adjoint - discrete_forward_sens", msg, value = TRUE)
      expect_length(line, 1L)
      vAdj <- as.numeric(sub(".*forward_sens\\| = ([0-9.eE+-]+).*", "\\1", line))
      vPri <- as.numeric(sub(".*primal_replay_err = ([0-9.eE+-]+).*", "\\1", line))
      nSw  <- as.integer(sub(".*methodSwitches=([0-9]+).*", "\\1", line))
      expect_lt(vAdj, 1e-8)              # exact transpose (machine precision)
      expect_lt(vPri, tolPrimal)         # step-map model matches liblsoda to ~tol
      expect_gte(nSw, minSwitch)         # the intended regime is actually exercised
    }
    # P1: fixed order-1 / fixed step
    H <- 0.005
    selfcheck(p, list(maxordn = 1L, maxords = 1L, hmin = H, hmax = H, hini = H, atol = 1e-4, rtol = 1e-4), 1e-6)
    # P2+P3: variable order + adaptive step, tight tol
    selfcheck(p, list(maxordn = 5L, maxords = 5L, atol = 1e-9, rtol = 1e-9), 1e-6)
    selfcheck(p, list(maxordn = 5L, maxords = 5L, atol = 1e-11, rtol = 1e-11), 1e-8)
    # P4: a stiff problem forces an Adams->BDF method switch (liblsoda always starts
    # in Adams); the used el come from the elco table per method, so the switch needs
    # no special transpose handling -- assert a switch actually happened.
    selfcheck(c(ka = 200, cl = 1, v = 10), list(maxordn = 5L, maxords = 5L, atol = 1e-9, rtol = 1e-9),
              1e-6, minSwitch = 1L)

    # primal identical to plain liblsoda on the same fixed step (P1 config)
    sd <- as.data.frame(do.call(rxode2::rxSolve, c(list(madj, ev, params = p, method = "liblsodaadj",
      maxordn = 1L, maxords = 1L, hmin = H, hmax = H, hini = H, atol = 1e-4, rtol = 1e-4, cores = 1))))
    sb <- as.data.frame(do.call(rxode2::rxSolve, c(list(mbase, ev, params = p, method = "liblsoda",
      maxordn = 1L, maxords = 1L, hmin = H, hmax = H, hini = H, atol = 1e-4, rtol = 1e-4, cores = 1))))
    expect_lt(max(abs(as.matrix(sd[, ex$st]) - as.matrix(sb[, ex$st]))), 1e-8)

    # practical FD cross-check on schedule-INDEPENDENT params (cl, v: center = 0 at
    # t0, so they do not perturb liblsoda's fixed-step-order-1 start step).
    solveBase <- function(pp) as.matrix(as.data.frame(do.call(rxode2::rxSolve, c(list(mbase, ev, params = pp,
      method = "liblsoda", maxordn = 1L, maxords = 1L, hmin = H, hmax = H, hini = H, atol = 1e-4, rtol = 1e-4, cores = 1))))[, ex$st])
    for (pn in c("cl", "v")) {
      hh <- p[[pn]] * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
      fd <- (solveBase(pp) - solveBase(pm)) / (2 * hh)
      for (k in seq_along(ex$st)) {
        adj <- sd[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]]
        expect_lt(max(abs(adj - fd[, k])), 1e-4)
      }
    }
  })

  test_that("in-engine liblsodaadj (P5) interior event jumps: multi-dose / reset / replace / multiply", {
    skip_on_cran()
    skip_on_ci()

    # Interior events reset liblsoda's integrator (istateReset), so the trajectory
    # is a chain of Nordsieck segments joined by a state jump.  The segment re-init
    # yh0=[y0,h0*f(y0)] couples the Nordsieck rows through f, so the costate handed to
    # the previous segment is lam0[1] + h0*J(y0)^T*lam0[2], then the event's dPhi/dy^T
    # (bolus = I; reset -> 0; replace[c] -> comp c 0; multiply[c] -> comp c *= factor).
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text)
    p <- c(ka = 1.2, cl = 3.5, v = 25); obs <- c(1, 2, 6, 8, 12)
    evs <- list(
      multidose = et(et(amt = 100, cmt = "depot", ii = 6, addl = 3), c(2, 8, 14, 20)),
      reset     = et(et(et(et(amt = 100, cmt = "depot"), 4, evid = 3), amt = 100, cmt = "depot", time = 4), obs),
      replace   = et(et(et(amt = 100, cmt = "depot"), amt = 40, cmt = "center", evid = 5, time = 4), obs),
      multiply  = et(et(et(amt = 100, cmt = "depot"), amt = 0.5, cmt = "center", evid = 6, time = 4), obs))

    op0 <- Sys.getenv("RX_LSADJ_SELFCHECK", unset = NA)
    on.exit(if (is.na(op0)) Sys.unsetenv("RX_LSADJ_SELFCHECK") else Sys.setenv(RX_LSADJ_SELFCHECK = op0), add = TRUE)
    for (nm in names(evs)) {
      ev <- evs[[nm]]
      # (A) self-check: exact transpose (machine precision) + primal model match (O(tol))
      Sys.setenv(RX_LSADJ_SELFCHECK = "1")
      msg <- capture.output(
        invisible(rxode2::rxSolve(madj, ev, params = p, method = "liblsodaadj", cores = 1, atol = 1e-10, rtol = 1e-10)),
        type = "message")
      Sys.unsetenv("RX_LSADJ_SELFCHECK")
      line <- grep("adjoint - discrete_forward_sens", msg, value = TRUE)
      expect_length(line, 1L)
      expect_lt(as.numeric(sub(".*forward_sens\\| = ([0-9.eE+-]+).*", "\\1", line)), 1e-8)
      expect_lt(as.numeric(sub(".*primal_replay_err = ([0-9.eE+-]+).*", "\\1", line)), 1e-6)
      # (B) independent cross-check vs dop853s (both converge to the continuous
      # derivative at tight tol) -- validates the boundary jump itself.
      sL <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = "liblsodaadj", cores = 1, atol = 1e-10, rtol = 1e-10))
      sD <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = "dop853s",    cores = 1, atol = 1e-10, rtol = 1e-10))
      mx <- 0
      for (st in ex$st) for (pn in cs) { cn <- sprintf("rx__sens_%s_BY_%s__", st, pn); mx <- max(mx, max(abs(sL[[cn]] - sD[[cn]]))) }
      expect_lt(mx, 1e-5)
    }

    # bare reset (no redose): every post-reset sensitivity is identically 0
    evBare <- et(et(et(amt = 100, cmt = "depot"), 4, evid = 3), obs)
    sB <- as.data.frame(rxode2::rxSolve(madj, evBare, params = p, method = "liblsodaadj", cores = 1, atol = 1e-10, rtol = 1e-10))
    post <- sB$time > 4
    for (st in ex$st) for (pn in cs)
      expect_lt(max(abs(sB[[sprintf("rx__sens_%s_BY_%s__", st, pn)]][post])), 1e-10)
  })

  test_that("in-engine liblsodaadj (P5) modeled bioavailability F and lag-time alag jumps", {
    skip_on_cran()
    skip_on_ci()

    # A modeled f(depot)=Fbio adds mu += amt*dF/dp*lam[c] at the dose; a modeled
    # alag(depot)=tlag shifts the dose time, adding the transversality
    # mu += -amt*dlag/dp*(lam^T F_X[:,c]).  These apply at every segment start,
    # including the t0 dose (seg 0).  Cross-checked vs dop853s (same discrete-adjoint
    # family) at tight tol; the self-check confirms an exact transpose + model match.
    op0 <- Sys.getenv("RX_LSADJ_SELFCHECK", unset = NA)
    on.exit(if (is.na(op0)) Sys.unsetenv("RX_LSADJ_SELFCHECK") else Sys.setenv(RX_LSADJ_SELFCHECK = op0), add = TRUE)
    cases <- list(
      Fsingle = list(m = "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\nf(depot)=Fbio",
                     cs = c("ka","cl","v","Fbio"), p = c(ka=1.2, cl=3.5, v=25, Fbio=0.7),
                     ev = et(et(amt=100, cmt="depot"), c(1,2,4,8,12))),
      Fmulti  = list(m = "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\nf(depot)=Fbio",
                     cs = c("ka","cl","v","Fbio"), p = c(ka=1.2, cl=3.5, v=25, Fbio=0.7),
                     ev = et(et(amt=100, cmt="depot", ii=6, addl=2), c(1,7,13))),
      alag    = list(m = "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\nalag(depot)=tlag",
                     cs = c("ka","cl","v","tlag"), p = c(ka=1.2, cl=3.5, v=25, tlag=0.7),
                     ev = et(et(amt=100, cmt="depot"), c(1,2,4,8,12))))
    for (nm in names(cases)) {
      cc <- cases[[nm]]
      ex <- rxode2::.rxAdjointExpand(cc$m, cc$cs); madj <- rxode2::rxode2(ex$text)
      Sys.setenv(RX_LSADJ_SELFCHECK = "1")
      msg <- capture.output(
        sL <- as.data.frame(rxode2::rxSolve(madj, cc$ev, params = cc$p, method = "liblsodaadj", cores = 1, atol = 1e-10, rtol = 1e-10)),
        type = "message")
      Sys.unsetenv("RX_LSADJ_SELFCHECK")
      line <- grep("adjoint - discrete_forward_sens", msg, value = TRUE)
      expect_length(line, 1L)
      expect_lt(as.numeric(sub(".*forward_sens\\| = ([0-9.eE+-]+).*", "\\1", line)), 1e-8)
      expect_lt(as.numeric(sub(".*primal_replay_err = ([0-9.eE+-]+).*", "\\1", line)), 1e-6)
      sD <- as.data.frame(rxode2::rxSolve(madj, cc$ev, params = cc$p, method = "dop853s", cores = 1, atol = 1e-10, rtol = 1e-10))
      mx <- 0
      for (st in ex$st) for (pn in cc$cs) { cn <- sprintf("rx__sens_%s_BY_%s__", st, pn); mx <- max(mx, max(abs(sL[[cn]] - sD[[cn]]))) }
      expect_lt(mx, 1e-5)
      # the modeled-modifier parameter's own gradient is genuinely non-zero
      mp <- if (nm == "alag") "tlag" else "Fbio"
      expect_gt(max(abs(sL[[sprintf("rx__sens_center_BY_%s__", mp)]])), 1e-3)
    }
  })

  test_that("in-engine liblsodaadj (P5) infusions: constant rate + modeled rate()/dur() duals", {
    skip_on_cran()
    skip_on_ci()

    # A constant-rate infusion has no dR/dp and is transparent to the adjoint.  A
    # MODELED rate(c)=R or dur(c)=D adds the in-window forcing (F_p[c] += durMult*dR/dp
    # inside the infusion) and the moving off-boundary transversality (-amt/R*durMult),
    # with R = amt/(t_off - t_on) reconstructed from the infusion window's event times.
    # Validated vs dop853s at tight tol (the self-check replay does not carry the
    # runtime infusion rate, so it is skipped for infusion models).
    obs <- c(1, 2, 4, 6, 10)
    cases <- list(
      const   = list(m = "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center",
                     cs = c("ka","cl","v"), p = c(ka=1.2, cl=3.5, v=25),
                     ev = et(et(amt=100, cmt="center", rate=50), obs), gp = NULL),
      mrate   = list(m = "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\nrate(center)=Rin",
                     cs = c("ka","cl","v","Rin"), p = c(ka=1.2, cl=3.5, v=25, Rin=40),
                     ev = et(et(amt=100, cmt="center", rate=-1), obs), gp = "Rin"),
      mdur    = list(m = "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-(cl/v)*center\ndur(center)=Dur",
                     cs = c("ka","cl","v","Dur"), p = c(ka=1.2, cl=3.5, v=25, Dur=3),
                     ev = et(et(amt=100, cmt="center", rate=-2), obs), gp = "Dur"))
    for (nm in names(cases)) {
      cc <- cases[[nm]]
      ex <- rxode2::.rxAdjointExpand(cc$m, cc$cs); madj <- rxode2::rxode2(ex$text)
      sL <- as.data.frame(rxode2::rxSolve(madj, cc$ev, params = cc$p, method = "liblsodaadj", cores = 1, atol = 1e-9, rtol = 1e-9))
      sD <- as.data.frame(rxode2::rxSolve(madj, cc$ev, params = cc$p, method = "dop853s",    cores = 1, atol = 1e-9, rtol = 1e-9))
      mx <- 0
      for (st in ex$st) for (pn in cc$cs) { cn <- sprintf("rx__sens_%s_BY_%s__", st, pn); mx <- max(mx, max(abs(sL[[cn]] - sD[[cn]]))) }
      expect_lt(mx, 1e-5)
      # the modeled rate/dur parameter's own gradient is genuinely non-zero
      if (!is.null(cc$gp)) expect_gt(max(abs(sL[[sprintf("rx__sens_center_BY_%s__", cc$gp)]])), 1e-3)
    }
  })

  test_that("in-engine liblsodaadj (P6) auto-switch from liblsoda + parallel across subjects", {
    skip_on_cran()
    skip_on_ci()

    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text)
    ev <- et(amt = 100, cmt = "depot", ii = 6, addl = 2) %>% et(c(1, 4, 7, 10, 16))
    # (1) an adjoint model solved with the DEFAULT liblsoda method auto-switches to
    # liblsodaadj (its direct adjoint variant) rather than the generic dop853s.
    sAuto <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = "liblsoda",    cores = 1, atol = 1e-9, rtol = 1e-9))
    sExpl <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = "liblsodaadj", cores = 1, atol = 1e-9, rtol = 1e-9))
    for (cn in grep("rx__sens", names(sAuto), value = TRUE))
      expect_equal(sAuto[[cn]], sExpl[[cn]])
    # genuinely non-zero (not the silent all-zero base-liblsoda path)
    expect_gt(max(abs(sAuto[["rx__sens_center_BY_cl__"]])), 1e-3)

    # (2) parallel across subjects is bit-identical to serial.
    pd <- data.frame(id = 1:4, ka = c(1.0, 1.2, 1.4, 1.6), cl = c(3, 3.5, 4, 4.5), v = c(20, 25, 30, 22))
    evp <- et(amt = 100, cmt = "depot", ii = 6, addl = 2) %>% et(c(1, 4, 7, 10, 16)) %>% et(id = 1:4)
    s1 <- as.data.frame(rxode2::rxSolve(madj, evp, pd, method = "liblsodaadj", cores = 1, atol = 1e-9, rtol = 1e-9))
    s4 <- as.data.frame(rxode2::rxSolve(madj, evp, pd, method = "liblsodaadj", cores = 4, atol = 1e-9, rtol = 1e-9))
    for (cn in grep("rx__sens", names(s1), value = TRUE))
      expect_equal(s1[[cn]], s4[[cn]])
  })

  test_that("in-engine dop54s / dp54s / vern98s (explicit-RK tableau additions)", {
    skip_on_cran()
    skip_on_ci()

    # dop54/dp54 (OdeDoPri54) share the Dormand-Prince 5(4) tableau (= dop5s); vern98
    # is Verner's 9(8).  Both drop into the rk4s Butcher-tableau framework: the base
    # states must match the base method's own solve, and the sensitivities converge to
    # the continuous derivative (== dop853s) at tight tolerance.
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text); mbase <- rxode2::rxode2(mText)
    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 2, 4, 8, 12, 24))
    sD <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = "dop853s", cores = 1, atol = 1e-11, rtol = 1e-11))
    solveBase <- function(pp, m) as.matrix(as.data.frame(
      rxode2::rxSolve(mbase, ev, params = pp, method = m, cores = 1, atol = 1e-11, rtol = 1e-11))[, ex$st])
    for (info in list(c("dop54s", "dop54"), c("dp54s", "dop54"), c("vern98s", "vern98"))) {
      meth <- info[1]; base <- info[2]
      s <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = meth, cores = 1, atol = 1e-11, rtol = 1e-11))
      # base states track the base method's own adaptive solve
      expect_lt(max(abs(as.matrix(s[, ex$st]) - solveBase(p, base))), 1e-6)
      # sensitivities agree with dop853s (all -> continuous derivative at tight tol)
      mx <- 0
      for (st in ex$st) for (pn in cs) { cn <- sprintf("rx__sens_%s_BY_%s__", st, pn); mx <- max(mx, max(abs(s[[cn]] - sD[[cn]]))) }
      expect_lt(mx, 1e-6)
      # the base method on an adjoint model auto-switches to this variant (bit-identical)
      sAuto <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = base, cores = 1, atol = 1e-11, rtol = 1e-11))
      for (cn in grep("rx__sens", names(s), value = TRUE)) expect_equal(sAuto[[cn]], s[[cn]])
    }
  })

  test_that("in-engine libode explicit-RK family (discrete-adjoint tableau batch)", {
    skip_on_cran()
    skip_on_ci()

    # 28 libode explicit-RK methods get a discrete adjoint by dropping their Butcher
    # tableaus into the rk4s framework.  For each: the base states must track the base
    # method's own adaptive solve, and the sensitivities converge to the continuous
    # derivative (== dop853s) at tight tolerance.  Auto-switch (base -> variant) too.
    bases <- c("f45","t54","pp54","pp54b","bs54","ss54","dp65","c65","tp64","v65r",
               "dverk65","tf65","tp75","tmy7","tmy7s","v76r","ss76","v78","dverk78","dp85",
               "tp86","v87e","v87r","ev87","k87","v89","t98a","v98r","s98",
               "c108","b109","s1110a","o129")   # >16-stage (need enlarged rksTableau)
    # tmy7/tmy7s use the "adj" suffix (like cvodesadj/liblsodaadj) because "tmy7"+"s"
    # would collide with the base method tmy7s.
    adjNames <- c(tmy7 = "tmy7adj", tmy7s = "tmy7sadj")
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text); mbase <- rxode2::rxode2(mText)
    ev <- et(amt = 100, cmt = "depot") %>% et(c(1, 2, 4, 8, 12, 24))
    sD <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = "dop853s", cores = 1, atol = 1e-11, rtol = 1e-11))
    baseSolve <- function(m) as.matrix(as.data.frame(
      rxode2::rxSolve(mbase, ev, params = p, method = m, cores = 1, atol = 1e-11, rtol = 1e-11))[, ex$st])
    for (base in bases) {
      meth <- if (base %in% names(adjNames)) adjNames[[base]] else paste0(base, "s")
      s <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = meth, cores = 1, atol = 1e-11, rtol = 1e-11))
      expect_lt(max(abs(as.matrix(s[, ex$st]) - baseSolve(base))), 1e-6)  # base-state match
      mx <- 0
      for (st in ex$st) for (pn in cs) { cn <- sprintf("rx__sens_%s_BY_%s__", st, pn); mx <- max(mx, max(abs(s[[cn]] - sD[[cn]]))) }
      expect_lt(mx, 1e-6)                                                  # sens vs dop853s
      sAuto <- as.data.frame(rxode2::rxSolve(madj, ev, params = p, method = base, cores = 1, atol = 1e-11, rtol = 1e-11))
      expect_equal(sAuto[["rx__sens_center_BY_cl__"]], s[["rx__sens_center_BY_cl__"]])  # auto-switch
    }
  })

  test_that("in-engine abs (Adams-Bashforth) discrete adjoint == FD of the AB solve", {
    skip_on_cran()
    skip_on_ci()

    # abs is the exact reverse-mode transpose of ab.cpp's classical fixed-order/
    # fixed-step Adams-Bashforth (an f-history multistep, not liblsoda's Nordsieck).
    # Because the AB schedule (dt, order) is param-INDEPENDENT, the discrete adjoint
    # equals a central finite difference of the SAME AB solve to FD precision -- the
    # correct criterion (compare against the abs solve's own base states, which are
    # what the adjoint transposes).
    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text)
    ctl <- list(hmin = 0.02, cores = 1)
    for (ord in c(1L, 2L, 4L)) {
      cc <- c(ctl, list(maxordn = ord))
      for (ev in list(et(amt = 100, cmt = "depot") %>% et(c(1, 4, 8, 12)),
                      et(amt = 100, cmt = "depot", ii = 6, addl = 2) %>% et(c(1, 4, 7, 10, 16)))) {
        s <- as.data.frame(do.call(rxode2::rxSolve, c(list(madj, ev, params = p, method = "abs"), cc)))
        solveAbs <- function(pp) as.matrix(as.data.frame(
          do.call(rxode2::rxSolve, c(list(madj, ev, params = pp, method = "abs"), cc)))[, ex$st])
        mx <- 0
        for (pn in cs) {
          hh <- p[[pn]] * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
          fd <- (solveAbs(pp) - solveAbs(pm)) / (2 * hh)
          for (k in seq_along(ex$st))
            mx <- max(mx, max(abs(s[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k])))
        }
        expect_lt(mx, 1e-5)
        expect_gt(max(abs(s[["rx__sens_center_BY_cl__"]])), 1e-3)  # genuinely non-zero
      }
    }
  })

  test_that("in-engine abs interior event jumps (reset/replace/multiply) + parallel", {
    skip_on_cran()
    skip_on_ci()

    ex <- rxode2::.rxAdjointExpand(mText, cs)
    madj <- rxode2::rxode2(ex$text)
    ctl <- list(hmin = 0.02, maxordn = 4L, cores = 1); obs <- c(1, 2, 6, 8, 12)
    solveAbs <- function(pp, ev) as.matrix(as.data.frame(
      do.call(rxode2::rxSolve, c(list(madj, ev, params = pp, method = "abs"), ctl)))[, ex$st])
    evs <- list(
      reset    = et(et(et(et(amt = 100, cmt = "depot"), 4, evid = 3), amt = 100, cmt = "depot", time = 4), obs),
      replace  = et(et(et(amt = 100, cmt = "depot"), amt = 40, cmt = "center", evid = 5, time = 4), obs),
      multiply = et(et(et(amt = 100, cmt = "depot"), amt = 0.5, cmt = "center", evid = 6, time = 4), obs))
    for (nm in names(evs)) {
      ev <- evs[[nm]]
      s <- as.data.frame(do.call(rxode2::rxSolve, c(list(madj, ev, params = p, method = "abs"), ctl)))
      mx <- 0
      for (pn in cs) { hh <- p[[pn]] * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        fd <- (solveAbs(pp, ev) - solveAbs(pm, ev)) / (2 * hh)
        for (k in seq_along(ex$st)) mx <- max(mx, max(abs(s[[sprintf("rx__sens_%s_BY_%s__", ex$st[k], pn)]] - fd[, k]))) }
      expect_lt(mx, 1e-5)
    }
    # bare reset (no redose): post-reset sensitivities identically 0
    evBare <- et(et(et(amt = 100, cmt = "depot"), 4, evid = 3), obs)
    sB <- as.data.frame(do.call(rxode2::rxSolve, c(list(madj, evBare, params = p, method = "abs"), ctl)))
    post <- sB$time > 4
    for (st in ex$st) for (pn in cs)
      expect_lt(max(abs(sB[[sprintf("rx__sens_%s_BY_%s__", st, pn)]][post])), 1e-10)
    # parallel across subjects bit-identical to serial
    pd <- data.frame(id = 1:4, ka = c(1, 1.2, 1.4, 1.6), cl = c(3, 3.5, 4, 4.5), v = c(20, 25, 30, 22))
    evp <- et(amt = 100, cmt = "depot", ii = 6, addl = 2) %>% et(c(1, 4, 7, 10, 16)) %>% et(id = 1:4)
    s1 <- as.data.frame(do.call(rxode2::rxSolve, c(list(madj, evp, pd, method = "abs"), list(hmin = 0.02, maxordn = 4L, cores = 1))))
    s4 <- as.data.frame(do.call(rxode2::rxSolve, c(list(madj, evp, pd, method = "abs"), list(hmin = 0.02, maxordn = 4L, cores = 4))))
    for (cn in grep("rx__sens", names(s1), value = TRUE)) expect_equal(s1[[cn]], s4[[cn]])
  })
})

* The following changes were made:
*     - removed any printout information
*     - Made matvec take the arguments matvec( x, y, A, n)
*----------------------------------------------------------------------|
      subroutine DSPHIV( n, m, t, u, v, w, tol, anorm,
     .     wsp,lwsp, iwsp,liwsp, matvec, iflag, A, ideg, mxstep)

      implicit none
      integer n, m, lwsp, liwsp, iflag, iwsp(liwsp), ideg, mxstep
      double precision t, tol, anorm, u(n), v(n), w(n), wsp(lwsp)
      double precision A(n,n)
      external matvec

*-----Purpose----------------------------------------------------------|
*
*---  DSPHIV computes w = exp(t*A)v + t*phi(tA)u which is the solution 
*     of the nonhomogeneous linear ODE problem w' = Aw + u, w(0) = v.
*     phi(z) = (exp(z)-1)/z and A is a Symmetric matrix.
*
*     The method used is based on Krylov subspace projection
*     techniques and the matrix under consideration interacts only
*     via the external routine `matvec' performing the matrix-vector 
*     product (matrix-free method).
*
*-----Arguments--------------------------------------------------------|
*
*     n      : (input) order of the principal matrix A.
*                      
*     m      : (input) maximum size for the Krylov basis.
*                      
*     t      : (input) time at wich the solution is needed (can be < 0).
*   
*     u(n)   : (input) operand vector with respect to the phi function
*              (forcing term of the ODE problem).
*
*     v(n)   : (input) operand vector with respect to the exp function
*              (initial condition of the ODE problem).
*  
*     w(n)   : (output) computed approximation of exp(t*A)v + t*phi(tA)u 
* 
*     tol    : (input/output) the requested accuracy tolerance on w. 
*              If on input tol=0.0d0 or tol is too small (tol.le.eps)
*              the internal value sqrt(eps) is used, and tol is set to
*              sqrt(eps) on output (`eps' denotes the machine epsilon).
*              (`Happy breakdown' is assumed if h(j+1,j) .le. anorm*tol)
*
*     anorm  : (input) an approximation of some norm of A.
*
*   wsp(lwsp): (workspace) lwsp .ge. n*(m+1)+n+(m+3)^2+4*(m+3)^2+ideg+1
*                                   +---------+-------+---------------+
*              (actually, ideg=6)        V        H      wsp for PADE
*                   
* iwsp(liwsp): (workspace) liwsp .ge. m+3
*
*     matvec : external subroutine for matrix-vector multiplication.
*              synopsis: matvec( x, y, A, n)
*                        double precision x(*), y(*)
*              computes: y(1:n) <- A*x(1:n)
*                        where A is the principal matrix.
*
*     iflag  : (output) exit flag.
*              <0 - bad input arguments 
*               0 - no problem
*               1 - maximum number of steps reached without convergence
*               2 - requested tolerance was too high
*
*-----Accounts on the computation--------------------------------------|
*     Upon exit, an interested user may retrieve accounts on the 
*     computations. They are located in the workspace arrays wsp and 
*     iwsp as indicated below: 
*
*     location  mnemonic                 description
*     -----------------------------------------------------------------|
*     iwsp(1) = nmult, number of matrix-vector multiplications used
*     iwsp(2) = nexph, number of Hessenberg matrix exponential evaluated
*     iwsp(3) = nscale, number of repeated squaring involved in Pade
*     iwsp(4) = nstep, number of integration steps used up to completion 
*     iwsp(5) = nreject, number of rejected step-sizes
*     iwsp(6) = ibrkflag, set to 1 if `happy breakdown' and 0 otherwise
*     iwsp(7) = mbrkdwn, if `happy brkdown', basis-size when it occured
*     -----------------------------------------------------------------|
*     wsp(1)  = step_min, minimum step-size used during integration
*     wsp(2)  = step_max, maximum step-size used during integration
*     wsp(3)  = dummy
*     wsp(4)  = dummy
*     wsp(5)  = x_error, maximum among all local truncation errors
*     wsp(6)  = s_error, global sum of local truncation errors
*     wsp(7)  = tbrkdwn, if `happy breakdown', time when it occured
*     wsp(8)  = t_now, integration domain successfully covered
*
*----------------------------------------------------------------------|
*-----The following parameters may also be adjusted herein-------------|
*
      integer mxreject
      double precision delta, gamma
      parameter( mxreject = 0,
     .           delta    = 1.2d0,
     .           gamma    = 0.9d0 )

*     mxstep  : maximum allowable number of integration steps.
*               The value 0 means an infinite number of steps.
* 
*     mxreject: maximum allowable number of rejections at each step. 
*               The value 0 means an infinite number of rejections.
*
*     ideg    : the Pade approximation of type (ideg,ideg) is used as 
*               an approximation to exp(H).
*
*     delta   : local truncation error `safety factor'
*
*     gamma   : stepsize `shrinking factor'
*
*----------------------------------------------------------------------|
*     Roger B. Sidje (rbs@maths.uq.edu.au)
*     EXPOKIT: Software Package for Computing Matrix Exponentials.
*     ACM - Transactions On Mathematical Software, 24(1):130-156, 1998
*----------------------------------------------------------------------|
*
      integer i, j, k1, mh, mx, iv, ih, j1v, ns, ifree, lfree, iexph,
     .        ireject,ibrkflag,mbrkdwn, nmult, nreject, nexph, nscale,
     .        nstep, iphih
      double precision sgn, t_out, tbrkdwn, step_min,step_max, err_loc,
     .                 s_error, x_error, t_now, t_new, t_step, t_old,
     .                 xm, beta, break_tol, p1, p2, p3, eps, rndoff,
     .                 avnorm, hj1j, hjj, SQR1

      intrinsic AINT,ABS,DBLE,LOG10,MAX,MIN,NINT,SIGN,SQRT
      double precision DDOT, DNRM2

*---  check restrictions on input parameters ...
      iflag = 0
      if ( lwsp.lt.n*(m+3)+5*(m+3)**2+ideg+1 ) iflag = -1
      if ( iflag.ne.0 ) then
         call rexit('bad sizes (in input of DSPHIV; -1)')
      end if
      if ( liwsp.lt.m+3 ) iflag = -2
      if ( iflag.ne.0 ) then
         call rexit('bad sizes (in input of DSPHIV; -2)')
      end if
      if ( m.ge.n .or. m.le.0 ) iflag = -3
      if ( iflag.ne.0 ) then
         call rexit('bad sizes (in input of DSPHIV; -3)')
      end if
*
*---  initialisations ...
*
      k1 = 3
      mh = m + 3
      iv = 1 
      ih = iv + n*(m+1) + n
      ifree = ih + mh*mh
      lfree = lwsp - ifree + 1

      ibrkflag = 0
      mbrkdwn  = m
      nmult    = 0
      nreject  = 0
      nexph    = 0
      nscale   = 0

      t_out    = ABS( t )
      tbrkdwn  = 0.0d0
      step_min = t_out
      step_max = 0.0d0
      nstep    = 0
      s_error  = 0.0d0
      x_error  = 0.0d0
      t_now    = 0.0d0
      t_new    = 0.0d0

      p1 = 4.0d0/3.0d0
 1    p2 = p1 - 1.0d0
      p3 = p2 + p2 + p2
      eps = ABS( p3-1.0d0 )
      if ( eps.eq.0.0d0 ) go to 1
      if ( tol.le.eps ) tol = SQRT( eps )
      rndoff = eps*anorm
 
      break_tol = 1.0d-7
*>>>  break_tol = tol
*>>>  break_tol = anorm*tol

*
*---  step-by-step integration ...
*
      sgn = SIGN( 1.0d0,t )
      SQR1 = SQRT( 0.1d0 )
      call DCOPY( n, v,1, w,1 )

 100  if ( t_now.ge.t_out ) goto 500

      nmult =  nmult + 1
      call matvec( w, wsp(iv) , A, n)
      call DAXPY( n, 1.0d0, u,1, wsp(iv),1 )
      beta = DNRM2( n, wsp(iv),1 )
      if ( beta.eq.0.0d0 ) goto 500
      call DSCAL( n, 1.0d0/beta, wsp(iv),1 )
      do i = 1,mh*mh
         wsp(ih+i-1) = 0.0d0
      enddo

      if ( nstep.eq.0 ) then
*---     obtain the very first stepsize ...
         xm = 1.0d0/DBLE( m )
         p1 = tol*(((m+1)/2.72D0)**(m+1))*SQRT(2.0D0*3.14D0*(m+1))
         t_new = (1.0d0/anorm)*(p1/(4.0d0*beta*anorm))**xm
         p1 = 10.0d0**(NINT( LOG10( t_new )-SQR1 )-1)
         t_new = AINT( t_new/p1 + 0.55d0 ) * p1
      endif
      nstep = nstep + 1
      t_step = MIN( t_out-t_now, t_new )
*
*---  Lanczos loop ...
*
      j1v = iv + n
      do 200 j = 1,m
         nmult = nmult + 1
         call matvec( wsp(j1v-n), wsp(j1v), A, n)
         if ( j.gt.1 )
     .     call DAXPY(n,-wsp(ih+(j-1)*mh+j-2),wsp(j1v-2*n),1,wsp(j1v),1)
         hjj = DDOT( n, wsp(j1v-n),1, wsp(j1v),1 )
         call DAXPY( n, -hjj, wsp(j1v-n),1, wsp(j1v),1 )
         hj1j = DNRM2( n, wsp(j1v),1 )
         wsp(ih+(j-1)*(mh+1)) = hjj
*---     if `happy breakdown' go straightforward at the end ... 
         if ( hj1j.le.break_tol ) then
            k1 = 0
            ibrkflag = 1
            mbrkdwn = j
            tbrkdwn = t_now
            t_step = t_out-t_now
            goto 300
         endif
         wsp(ih+(j-1)*mh+j) = hj1j
         wsp(ih+j*mh+j-1) = hj1j
         call DSCAL( n, 1.0d0/hj1j, wsp(j1v),1 )
         j1v = j1v + n
 200  continue
      nmult = nmult + 1
      call matvec( wsp(j1v-n), wsp(j1v), A, n)
      avnorm = DNRM2( n, wsp(j1v),1 )
*
*---  set 1's for the 3-extended scheme ...
*
 300  continue
      wsp(ih+mh*mbrkdwn) = 1.0d0
      wsp(ih+m*mh+m-1)   = 0.0d0
      wsp(ih+(m-1)*mh+m) = 0.0d0
      do i = 1,k1-1
         wsp(ih+(m+i)*mh+m+i-1) = 1.0d0
      enddo
*
*---  loop while ireject<mxreject until the tolerance is reached ...
*
      ireject = 0
 401  continue
*
*---  compute w = beta*t_step*V*phi(t_step*H)*e1 + w
*
      nexph = nexph + 1
      mx = mbrkdwn + MAX( 1,k1 )

*---  irreducible rational Pade approximation ...
      call DGPADM( ideg, mx, sgn*t_step, wsp(ih),mh,
     .              wsp(ifree),lfree, iwsp, iexph, ns, iflag)
      iexph = ifree + iexph - 1
      iphih = iexph + mbrkdwn*mx
      nscale = nscale + ns
      wsp(iphih+mbrkdwn)   = hj1j*wsp(iphih+mx+mbrkdwn-1)
      wsp(iphih+mbrkdwn+1) = hj1j*wsp(iphih+2*mx+mbrkdwn-1)
 
* 
*---  error estimate ...
*
      if ( k1.eq.0 ) then
         err_loc = tol
      else
         p1 = ABS( wsp(iphih+m) )   * beta
         p2 = ABS( wsp(iphih+m+1) ) * beta * avnorm 
         if ( p1.gt.10.0d0*p2 ) then
            err_loc = p2
            xm = 1.0d0/DBLE( m+1 )
         elseif ( p1.gt.p2 ) then
            err_loc = (p1*p2)/(p1-p2)
            xm = 1.0d0/DBLE( m+1 )
         else
            err_loc = p1
            xm = 1.0d0/DBLE( m )
         endif
      endif
*
*---  reject the step-size if the error is not acceptable ...
*   
      if ( (k1.ne.0) .and. (err_loc.gt.delta*t_step*tol) .and. 
     .     (mxreject.eq.0 .or. ireject.lt.mxreject) ) then
         t_old = t_step
         t_step = gamma * t_step * (t_step*tol/err_loc)**xm
         p1 = 10.0d0**(NINT( LOG10( t_step )-SQR1 )-1)
         t_step = AINT( t_step/p1 + 0.55d0 ) * p1
         ireject = ireject + 1
         nreject = nreject + 1
         if ( mxreject.ne.0 .and. ireject.gt.mxreject ) then
            iflag = 2
            return
         endif
         goto 401
      endif
*
      mx = mbrkdwn + MAX( 0,k1-2 )
      call DGEMV( 'n', n,mx,beta,wsp(iv),n,wsp(iphih),1,1.0d0,w,1 )
*
*---  suggested value for the next stepsize ...
*
      t_new = gamma * t_step * (t_step*tol/err_loc)**xm
      p1 = 10.0d0**(NINT( LOG10( t_new )-SQR1 )-1)
      t_new = AINT( t_new/p1 + 0.55d0 ) * p1 

      err_loc = MAX( err_loc,rndoff )
*
*---  update the time covered ...
*
      t_now = t_now + t_step 
*
*---  display and keep some information ...
*
      step_min = MIN( step_min, t_step ) 
      step_max = MAX( step_max, t_step )
      s_error = s_error + err_loc
      x_error = MAX( x_error, err_loc )
 
      if ( mxstep.eq.0 .or. nstep.lt.mxstep ) goto 100
      iflag = 1
 
 500  continue

      iwsp(1) = nmult
      iwsp(2) = nexph
      iwsp(3) = nscale
      iwsp(4) = nstep
      iwsp(5) = nreject
      iwsp(6) = ibrkflag
      iwsp(7) = mbrkdwn

      wsp(1)  = step_min
      wsp(2)  = step_max
      wsp(3)  = 0.0d0
      wsp(4)  = 0.0d0
      wsp(5)  = x_error
      wsp(6)  = s_error
      wsp(7)  = tbrkdwn
      wsp(8)  = sgn*t_now
      END
*----------------------------------------------------------------------|

# Adaptive dosing function names that must be captured before symengine sees them
.adaptiveDosingFns <- c("evid_", "bolus", "infuse", "infuseDur",
                          "reset", "replace", "multiply", "phantom", "obs")

#' Replace strings with numbers for the strAssign
#'
#' @param lhs string for the left hand side of equation or variable
#'   that is trying to be replaced with an integer
#' @param ret expression that will be returned if no replacement is
#'   made
#' @param strAssign The `strAssign` list from the model variables
#' @return either `ret` or the integer that corresponds to the string
#'   assignment
#' @noRd
#' @author Matthew L. Fidler
.rxPruneStrAssign <- function(lhs, ret, strAssign=list()) {
  if (length(strAssign) == 0L) return(ret)
  .w <- which(lhs %in% names(strAssign))
  if (length(.w) == 1L) {
    # Replace with integer
    .w <- which(vapply(seq_along(strAssign[[.w]]),
                       function(i) {
                         identical(ret, strAssign[[.w]][i])
                       },
                       logical(1)))
    if (length(.w) == 1L) {
      return(as.numeric(.w))
    }
  }
  ret
}

#'  Internal Pruning function
#'
#' @param x List of quoted lines
#' @param envir Environment where information is stored
#' @param strAssign string assignment list from  model variables
#' @return Pruned model code
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
.rxPrune <- function(x, envir = parent.frame(),
                     strAssign=list()) {
  if (is.name(x) || is.atomic(x)) {
    if (is.character(x)) {
      return(deparse1(x))
    }
    return(as.character(x))
  } else if (is.call(x)) {
    if (identical(x[[1]], quote(`if`))) {
      .if <- envir$.if
      .if[length(.if) + 1] <- .rxPrune(x[[2]], envir = envir, strAssign=strAssign)
      envir$.if <- .if
      .x2 <- x[-(1:2)]
      if (length(.x2) == 2) {
        .ret1 <- .rxPrune(.x2[[1]], envir = envir, strAssign=strAssign)
        .if[length(.if)] <- paste0("1-(", .if[length(.if)], ")")
        envir$.if <- .if
        .else <- envir$.else
        envir$.else <- unique(c(findLhs(eval(parse(text = paste0("quote({", .ret1, "})"))))))
        .ret2 <- .rxPrune(.x2[[2]], envir = envir, strAssign=strAssign)
        envir$.else <- .else
        .ret <- paste0(.ret1, "\n", .ret2)
      } else if (length(.x2) == 1) {
        .ret <- .rxPrune(.x2[[1]], envir = envir, strAssign=strAssign)
      }
      .if <- .if[-length(.if)]
      envir$.if <- .if
      return(.ret)
    } else if (identical(x[[1]], quote(`(`))) {
      return(paste0("(", .rxPrune(x[[2]], envir = envir, strAssign=strAssign), ")"))
    } else if (identical(x[[1]], quote(`{`))) {
      .x2 <- x[-1]
      return(paste0(lapply(.x2, .rxPrune, envir = envir, strAssign=strAssign), collapse = "\n"))
    } else if (identical(x[[1]], quote(`==`)) ||
                 identical(x[[1]], quote(`!=`))) {
      ## These cases can be strings that are assigned to integers.
      ## Here we need to check left/right hand sides
      .x2 <- deparse1(x[[2]])
      .x3 <- deparse1(x[[3]])
      x[[2]] <- .rxPruneStrAssign(.x3, x[[2]], strAssign=strAssign)
      x[[3]] <- .rxPruneStrAssign(.x2, x[[3]], strAssign=strAssign)
      .ret <- paste0(
        .rxPrune(x[[2]], envir = envir, strAssign=strAssign), as.character(x[[1]]),
        .rxPrune(x[[3]], envir = envir, strAssign=strAssign))
      return(.ret)
    } else if (identical(x[[1]], quote(`>=`)) ||
                 identical(x[[1]], quote(`<=`)) ||
                 identical(x[[1]], quote(`>`)) ||
                 identical(x[[1]], quote(`<`)) ||
                 identical(x[[1]], quote(`&&`)) ||
                 identical(x[[1]], quote(`||`)) ||
                 identical(x[[1]], quote(`&`)) ||
                 identical(x[[1]], quote(`|`))) {
      .ret <- paste0(
        .rxPrune(x[[2]], envir = envir, strAssign=strAssign), as.character(x[[1]]),
        .rxPrune(x[[3]], envir = envir, strAssign=strAssign))
      return(.ret)
    } else if (identical(x[[1]], quote(`=`)) ||
                 identical(x[[1]], quote(`<-`)) ||
                 identical(x[[1]], quote(`~`))) {
      .lhs <- deparse1(x[[2]])
      x[[3]] <- .rxPruneStrAssign(.lhs, x[[3]], strAssign=strAssign)
      if (length(envir$.if > 0)) {
        .f2 <- .rxPrune(x[[2]], envir = envir, strAssign=strAssign)
        .if <- paste0(paste0("(", envir$.if, ")"), collapse = "*")
        if (any(envir$.def1 == .f2)) {
          .ret <- paste0(
            .f2, as.character(x[[1]]), .if, "*(",
            .rxPrune(x[[3]], envir = envir, strAssign), ")+(1-(", .if, "))*(",
            .f2, ")"
          )
        } else {
          .ret <- paste0(
            .f2, as.character(x[[1]]), .if, "*(",
            .rxPrune(x[[3]], envir = envir, strAssign=strAssign), ")",
            ifelse(any(envir$.else == .f2),
              paste0("+", .f2), ""
            )
          )
        }
        assign(".def1", unique(c(envir$.def1, .f2)), envir)
        return(.ret)
      } else {
        .f2 <- .rxPrune(x[[2]], envir = envir, strAssign=strAssign)
        assign(".def1", unique(c(envir$.def1, .f2)), envir)
        return(paste0(
          .f2, as.character(x[[1]]),
          .rxPrune(x[[3]], envir = envir, strAssign=strAssign)
        ))
      }
    } else if (identical(x[[1]], quote(`*`)) ||
      identical(x[[1]], quote(`^`)) ||
      identical(x[[1]], quote(`+`)) ||
      identical(x[[1]], quote(`-`)) ||
      identical(x[[1]], quote(`/`))) {
      if (length(x) == 3) {
        return(paste0(
          .rxPrune(x[[2]], envir = envir, strAssign=strAssign), as.character(x[[1]]),
          .rxPrune(x[[3]], envir = envir, strAssign=strAssign)
        ))
      } else {
        ## Unary Operators
        return(paste0(
          as.character(x[[1]]),
          .rxPrune(x[[2]], envir = envir, strAssign=strAssign)
        ))
      }
    } else if (identical(x[[1]], quote(`ifelse`))) {
      .f2 <- .rxPrune(x[[2]], envir = envir, strAssign=strAssign)
      .f3 <- .rxPrune(x[[3]], envir = envir, strAssign=strAssign)
      .f4 <- .rxPrune(x[[4]], envir = envir, strAssign=strAssign)
      return(paste0("((", .f2, ")*(", .f3, ")+(1-(", .f2, "))*(", .f4, "))"))
    } else if (identical(x[[1]], quote(`[`))) {
      .type <- toupper(as.character(x[[2]]))
      ## Since only THETA/ETA are allowed with rxode2 pruning
      ## only will take legal rxode2; Therefore just paste these.
      return(paste0(.type, "[", x[[3]], "]"))
    } else if (exists("..captureN", envir = envir, inherits = FALSE) &&
               as.character(x[[1]]) %in% .adaptiveDosingFns) {
      # Adaptive dosing call: replace with rxCaptureId<N> <- condition so that
      # symengine can process the model without encountering unknown functions.
      .cond <- if (length(envir$.if) > 0L) {
        paste0(paste0("(", envir$.if, ")"), collapse = "*")
      } else {
        "1"
      }
      envir$..captureN <- envir$..captureN + 1L
      .n <- envir$..captureN
      .var <- paste0("rxCaptureId", .n)
      envir$..capturedEvid[[.n]] <- list(
        id        = .n,
        original  = deparse1(x),
        condition = .cond,
        capVar    = .var
      )
      return(paste0(.var, " <- ", .cond))
    } else {
      .ret0 <- lapply(x, .rxPrune, envir = envir, strAssign=strAssign)
      .ret <- paste0(.ret0[[1]], "(")
      .ret0 <- .ret0[-1]
      .ret <- paste0(.ret, paste0(unlist(.ret0), collapse = ", "), ")")
      return(.ret)
    }
  } else { ## nocov start
    ## is.pairlist OR is.atomic OR unknown...
    stop("unsupported expression", call. = FALSE)
  } ## nocov end
}


#' Prune branches (ie if/else) from rxode2
#'
#' This prunes branches (ie if/else) from rxode2.
#'
#' @param x rxode2 model that can be accessed by rxNorm
#' @return Pruned rxode2 model text.
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
rxPrune <- function(x) {
  .env <- new.env(parent = emptyenv())
  .mv <- rxModelVars(x)
  .env$.if <- NULL
  .env$.def1 <- NULL
  .env$..captureN <- 0L
  .env$..capturedEvid <- list()
  .ret <- .rxPrune(eval(parse(text = paste0("quote({", rxNorm(x), "})"))), envir = .env, strAssign=.mv$strAssign)
  if (length(.env$..capturedEvid) > 0L) {
    attr(.ret, "capturedEvid") <- .env$..capturedEvid
  }
  return(.ret)
}

.regIdentifier1 <- rex::rex(
  one_of("a":"z", "A":"Z"),
  any_of("_", "a":"z", "A":"Z", "0":"9", ".")
)
.regIdentifier2 <- rex::rex(
  at_least(".", 1),
  one_of("_", "a":"z", "A":"Z"),
  any_of("_", "a":"z", "A":"Z", "0":"9", ".")
)
.regIdentifier <- rex::rex(or(.regIdentifier1, .regIdentifier2))
regSens <- rex::rex(
  "rx__sens_", capture(.regIdentifier), "_BY_",
  capture(.regIdentifier), "__"
)
regSensEtaTheta <- rex::rex(
  "rx__sens_", capture(.regIdentifier),
  "_BY_", capture(.regIdentifier),
  "_BY_", capture(.regIdentifier), "__"
)
regToSens1 <- rex::rex(
  capture(.regIdentifier), or("_", ".", ":"),
  capture(.regIdentifier)
)
regToSens2 <- rex::rex(
  "d/dt(d(", capture(.regIdentifier), ")/d(",
  capture(.regIdentifier), "))"
)
regFloat1 <- rex::rex(
  or(
    group(some_of("0":"9"), ".", any_of("0":"9")),
    group(any_of("0":"9"), ".", some_of("0":"9"))
  ),
  maybe(group(
    one_of("E", "e"),
    maybe(one_of("+", "-")),
    some_of("0":"9")
  ))
)
regFloat2 <- rex::rex(
  some_of("0":"9"), one_of("E", "e"),
  maybe(one_of("-", "+")), some_of("0":"9")
)
regDecimalint <- rex::rex(or("0", group("1":"9", any_of("0":"9"))))
regNum <- rex::rex(maybe("-"), or(
  regDecimalint, regFloat1,
  regFloat2
))
regDDt <- rex::rex(
  start, "rx__d_dt_",
  capture(anything), "__", end
)
regDfDy <- rex::rex(
  start, "rx__df_", capture(anything),
  "_dy_", capture(anything), "__", end
)
regThEt <- rex::rex(
  capture(or("TH", ""), "ETA"), "_",
  capture("1":"9", any_of("0":"9")), "_"
)
regDfDyTh <- rex::rex(
  start, "rx__df_", capture(anything),
  "_dy_", regThEt, "__", end
)
regEta <- rex::rex(
  start, "ETA[",
  capture("1":"9", any_of("0":"9")),
  "]"
)
regTheta <- rex::rex(
  start, "THETA[",
  capture("1":"9", any_of("0":"9")),
  "]"
)
regJac <- rex::rex(
  "df(", capture(.regIdentifier), ")/dy(",
  capture(or(
    .regIdentifier,
    group(
      or("THETA[", "ETA["), "1":"9",
      any_of("0":"9"), "]"
    )
  )),
  ")"
)
.regRate <- rex::rex(start, "rx_rate_", capture(anything), "_")
.regDur <- rex::rex(start, "rx_dur_", capture(anything), "_")
.regLag <- rex::rex(start, "rx_lag_", capture(anything), "_")
.regF <- rex::rex(start, "rx_f_", capture(anything), "_")
.knownPrint <- c(
  "printf", "Rprintf", "print",
  "jac_printf", "jac_Rprintf", "jac_print",
  "ode_printf", "ode_Rprintf", "ode_print",
  "jac0_printf", "jac0_Rprintf", "jac0_print",
  "ode_printf", "ode_Rprintf", "ode_print",
  "ode0_printf", "ode0_Rprintf", "ode0_print",
  "lhs_printf", "lhs_Rprintf", "lhs_print"
)

regPrint <- rex::rex(
  start, or(.knownPrint),
  or(
    group(
      "(", anything, ")",
      any_spaces, at_most(";", 1), any_spaces
    ),
    group(any_spaces, at_most(";", 1), any_spaces)
  ),
  end
)

regIni0 <- rex::rex(
  start, "rx_", capture(anything),
  "_ini_0__", end
)
regIni <- rex::rex(or(
  group(one_of("_."), "0"),
  "0", "(0)", "[0]", "{0}"
), end)
#' Expand if/else clauses into multiple different types of lines.
#'
#' @param model Model can be a character, or a rxode2 model.  It needs
#'     to have normalized syntax, that is `if (...){}` has to be
#'     on the same line.  The `else` statement must be on its
#'     own line with the closing bracket of the `if` statement
#'     on the previous line.  This `else` statment must also
#'     contain the opening bracket, like the code `else {}`
#'
#' @param removeInis A boolean indicating if parameter
#'     initializations should be removed from the model.
#'
#' @param removePrint A boolean indicating if printing statements
#'     should be removed from the model.
#'
#' @return A named character vector. The names of the vector are the
#'     logical conditions, the values are the lines that satisfy the
#'     logical conditions.
#'
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
rxExpandIfElse <- function(model, removeInis = TRUE, removePrint = TRUE) {
  ## expand if/else blocks into a list with lines for conditions that are true
  x <- strsplit(rxNorm(model, FALSE), "\n")[[1]]
  if (removeInis) {
    x <- .rxRmIni(x)
  }
  if (removePrint) {
    x <- .rxRmPrint(x)
  }
  model <- x
  w1 <- which(regexpr(regIfOrElse, model) != -1)
  w2 <- which(regexpr(regEnd, model) != -1)
  if (length(w1) > 0) {
    currExpr <- ""
    lst <- list()
    last <- ""
    known <- list()
    for (i in seq_along(model)) {
      if (any(i == w1)) {
        if (regexpr(regElse, model[i]) != -1) {
          currExpr[length(currExpr) + 1] <- sprintf("!(%s)", last)
        } else {
          currExpr[length(currExpr) + 1] <- gsub(regIf, "!(\\1)", model[i])
          known[[length(known) + 1]] <- c(
            paste(paste0("(", currExpr[-1], ")"),
              collapse = " && "
            ),
            currExpr[-1]
          )
          currExpr[length(currExpr)] <- gsub(regIf, "\\1", model[i])
          known[[length(known) + 1]] <- c(
            paste(paste0("(", currExpr[-1], ")"),
              collapse = " && "
            ),
            currExpr[-1]
          )
        }
        lst[[i]] <- "control"
      } else if (any(i == w2)) {
        last <- currExpr[length(currExpr)]
        currExpr <- currExpr[seq(1, length(currExpr) - 1)]
        lst[[i]] <- "control"
      } else {
        lst[[i]] <- currExpr
      }
    }
    ret <- list()
    rm <- NULL
    for (i in seq_along(known)) {
      mod <- NULL
      for (j in seq_along(model)) {
        if (identical(lst[[j]], "")) {
          mod[length(mod) + 1] <- model[j]
        } else {
          i1 <- lst[[j]][-1]
          i2 <- known[[i]][-1]
          i3 <- i2[seq(1, min(length(i1), length(i2)))]
          if (!identical(i2, i3)) {
            ## Find the expression to remove
            for (k in seq_along(known)) {
              i4 <- known[[k]][-1]
              if (identical(i4, i3)) {
                rm <- c(rm, known[[k]][1])
              }
            }
          }
          if (identical(i1, i3)) {
            mod[length(mod) + 1] <- model[j]
          }
        }
      }
      ret[[known[[i]][1]]] <- paste(mod, collapse = "\n")
    }
    ret <- unlist(ret)
    ret <- ret[!(names(ret) %in% rm)]
    return(ret)
  } else {
    return(paste(model, collapse = "\n"))
  }
}

#' Remove INIs
#'
#' @param x rxode2 list of lines to remove
#' @return rxode2 lines with inis removed.
#' @author Matthew L. Fidler
#' @keywords internal
#' @noRd
.rxRmIni <- function(x) {
  x <- x[regexpr(rex::rex(
    start, any_spaces,
    or(names(rxInits(x))), any_spaces,
    or("=", "~")
  ), x) == -1]
  x <- x[regexpr(rex::rex(
    start, any_spaces,
    or(names(rxInits(x))), "(0)",
    any_spaces, or("=", "~")
  ), x) == -1]
  return(x)
}

#' Remove print statements
#'
#' @param x rxode2 lines to remove
#' @return rxode2 with print lines removed.
#' @author Matthew L. Fidler
#' @noRd
.rxRmPrint <- function(x) {
  return(x[regexpr(getFromNamespace("regPrint", "rxode2"), x) == -1])
}

#' Add a return statment to a function.
#'
#' @param fn Function to deparse
#' @param ret boolean stating if a return statement will be added.
#' @return Function with parens removed and add a return statment.
#' @author Matthew L. Fidler
#' @noRd
rxAddReturn <- function(fn, ret = TRUE) {
  txt <- deparse(body(fn))
  if (txt[1] == "{") {
    txt <- txt[-c(1, length(txt))]
  }
  .v <- try(parse(text = txt[length(txt)]), silent = TRUE)
  while (inherits(.v, "try-error")) {
    .v0 <- txt[length(txt)]
    txt <- txt[-length(txt)]
    txt[length(txt)] <- paste(txt[length(txt)], .v0)
    .v <- try(parse(text = txt[length(txt)]), silent = TRUE)
  }
  ## FIXME, naieve assumption about functions.
  if (ret) {
    if (regexpr(rex::rex(
      or(boundary, start),
      "return(", anything, ")"
    ),
    txt[length(txt)],
    perl = TRUE
    ) == -1) {
      ## Add return statement
      if (regexpr(
        rex::rex(or("=", "~", "<-", "}")),
        txt[length(txt)]
      ) == -1) {
        txt[length(txt)] <- gsub(
          rex::rex(
            start, any_spaces, capture(anything),
            or(";", ""), any_spaces, end
          ),
          "return(\\1);", txt[length(txt)]
        )
      }
    }
  }
  return(paste(txt, collapse = "\n"))
}



## Start DSL based on http://adv-r.had.co.nz/dsl.html
## These operators are called to create the language and are
## not called in tests.
unaryOp <- function(left, right) {
  force(left)
  force(right)
  function(e1) {
    paste0(left, e1, right)
  }
}

binaryOp <- function(sep) {
  force(sep)
  function(e1, e2) {
    if (missing(e2)) {
      paste0(gsub(" ", "", sep), .rxRepRxQ(e1))
    } else {
      paste0(.rxRepRxQ(e1), sep, .rxRepRxQ(e2))
    }
  }
}

binaryOp2 <- function(sep) {
  force(sep)
  function(e1, e2) {
    paste0("(", .rxRepRxQ(e1), ")", sep, "(", .rxRepRxQ(e2), ")")
  }
}

divOp <- function() {
  function(e1, e2) {
    if (e1 == "d" && grepl(rex::rex(start, "__dt__"), e2)) {
      paste0("rx__d_dt_", gsub(rex::rex(start, "__dt__"), "", e2))
    } else if (grepl(rex::rex(start, "__df_", anything, "_", end), e1) &&
      grepl(rex::rex(start, "_dy_", anything, "__", end), e2)) {
      paste0("rx", substring(e1, 0, nchar(e1) - 1), e2)
    } else if (is(e1, "numeric")) {
      paste0("S(", e1, ")/", e2)
    } else {
      paste0(e1, "/", e2)
    }
  }
}

functionOp <- function(fn) {
  force(fn)
  function(...) {
    paste0(fn, "(", paste(unlist(list(...)), collapse = ", "), ")")
  }
}

functionOp2 <- function(fn, end) {
  force(fn)
  force(end)
  function(...) {
    paste0(fn, paste(unlist(list(...)), collapse = ", "), end)
  }
}

.dslStripParen <- function(x) {
  .stripIt <- function(x) {
    if (is.call(x)) {
      if (length(x) == 1) {
        return(x)
      } else if (identical(x[[1]], quote(`(`))) {
        return(.stripIt(x[[2]]))
      } else {
        return(x)
      }
    } else {
      return(x)
    }
  }
  ret <- eval(parse(text = sprintf(".stripIt(quote(%s))", x)))
  ret <- paste(deparse1(ret), collapse = " ")
  return(ret)
}

.dslToPow <- function(a, b) {
  a <- .dslStripParen(a)
  b <- .dslStripParen(b)
  num <- suppressWarnings({
    as.numeric(b)
  })
  if (is.na(num)) {
    return(sprintf("Rx_pow(%s, %s)", a, b))
  } else if (abs(num - round(num)) < .Machine$double.eps^0.5) {
    return(sprintf("Rx_pow_di(%s, %s)", a, b))
  } else if (abs(num - 0.5) < .Machine$double.eps^0.5) {
    return(sprintf("sqrt(%s)", a))
  } else {
    return(sprintf("Rx_pow(%s, %s)", a, b))
  }
}

## Add sympy->C mini DSL for omega parsing

symengineC <- new.env(parent = emptyenv())
symengineC[["**"]] <- .dslToPow
symengineC[["^"]] <- .dslToPow

symengineC[["S"]] <- function(x) {
  sprintf("%s", x)
}

for (f in c(
  "acos", "acosh", "asin", "atan", "atan2", "atanh", "beta",
  "cos", "cosh", "digamma", "erf", "erfc", "exp", "factorial",
  "gamma", "sin", "sinh", "sqrt", "tan",
  "tanh", "trigamma", "rxTBS", "rxTBSd"
)) {
  symengineC[[f]] <- functionOp(f)
}

symengineC[["("]] <- unaryOp("(", ")")

for (op in c("+", "-", "*")) {
  symengineC[[op]] <- binaryOp(paste0(" ", op, " "))
}

symengineC[["/"]] <- function(e1, e2) {
  sprintf("%s /( (%s == 0) ? %s : %s)", e1, e2, .Machine$double.eps, e2)
}


sympyTransit4 <- function(t, n, mtt, bio, podo = "podo", tlast = "tlast") {
  ktr <- paste0("((", n, " + 1)/(", mtt, "))")
  lktr <- paste0("(log((", n, ") + 1) - log(", mtt, "))")
  tc <- paste0("((", t, ")-(", tlast, "))")
  paste0(
    "exp(log((", bio, ") * (", podo, ")) + ", lktr, " + (",
    n, ") * ", "(", lktr, " + log(", t, ")) - ",
    ktr, " * (", t, ") - log(gamma(1 + (", n, "))))"
  )
}

allStrs <- function(x) {
  if (is.atomic(x)) {
    if (is.character(x)) {
      return(x)
    }
    return(character())
  } else if (is.name(x)) {
    return(as.character())
  } else if (is.call(x) || is.pairlist(x)) {
    children <- lapply(x[-1], allStrs)
    unique(unlist(children))
  } else {
    stop(sprintf(gettext("do not know how to handle type '%s'"), typeof(x)),
      call. = FALSE
    )
  }
}

## Start error function DSL
sumProdEnv <- new.env(parent = emptyenv())

rxSumProdSum <- FALSE
rxSumProdProd <- FALSE


sumProdEnv[["^"]] <- binaryOp("^")

sumProdEnv[["**"]] <- binaryOp("^")

sumProdEnv[["*"]] <- function(a, b) {
  if (rxSumProdProd) {
    a <- .dslStripParen(a)
    b <- .dslStripParen(b)
    ## log transformed
    sprintf("prod(%s, %s)", sub(rex::rex(start, "prod(", capture(anything), ")", end), "\\1", a), b)
  } else {
    sprintf("%s * %s", a, b)
  }
}

sumProdEnv[["/"]] <- function(a, b) {
  if (rxSumProdProd) {
    a <- .dslStripParen(a)
    b <- .dslStripParen(b)
    sprintf("prod(%s, 1/%s)", sub(rex::rex(start, "prod(", capture(anything), ")", end), "\\1", a), b)
  } else {
    sprintf("%s / %s", a, b)
  }
}

sumProdEnv[["+"]] <- function(a, b) {
  if (rxSumProdSum) {
    a <- .dslStripParen(a)
    if (!missing(b)) {
      b <- .dslStripParen(b)
      return(sprintf("sum(%s, %s)", sub(rex::rex(start, "sum(", capture(anything), ")", end), "\\1", a), b))
    } else {
      return(a)
    }
  } else {
    if (!missing(b)) {
      b <- .dslStripParen(b)
      return(sprintf("%s + %s", a, b))
    } else {
      return(a)
    }
  }
}

sumProdEnv[["-"]] <- function(a, b) {
  if (rxSumProdSum) {
    a <- .dslStripParen(a)
    if (!missing(b)) {
      b <- .dslStripParen(b)
      sprintf(
        "sum(%s, -(%s))",
        sub(
          rex::rex(start, "sum(", capture(anything), ")", end),
          "\\1", a
        ), b
      )
    } else {
      paste0("-", a)
    }
  } else {
    if (!missing(b)) {
      b <- .dslStripParen(b)
      return(sprintf("%s - %s", a, b))
    } else {
      return(paste0("-", a))
    }
  }
}

sumProdEnv[["=="]] <- function(a, b) {
  return(sprintf("(%s == %s)", a, b))
}

sumProdEnv[["!="]] <- function(a, b) {
  return(sprintf("(%s != %s)", a, b))
}

sumProdEnv[[">"]] <- function(a, b) {
  return(sprintf("(%s > %s)", a, b))
}

sumProdEnv[["<"]] <- function(a, b) {
  return(sprintf("(%s < %s)", a, b))
}

sumProdEnv[["<="]] <- function(a, b) {
  return(sprintf("(%s <= %s)", a, b))
}

sumProdEnv[[">="]] <- function(a, b) {
  return(sprintf("(%s >= %s)", a, b))
}

sumProdEnv[["("]] <- function(a) {
  return(sprintf("(%s)", a))
}

sumProdEnv[["["]] <- function(name, val) {
  n <- toupper(name)
  err <- gettext("rxode2 only supports THETA[#] and ETA[#] numbers")
  if (any(n == c("THETA", "ETA")) && is.numeric(val)) {
    if (round(val) == val && val > 0) {
      return(sprintf("%s[%s]", n, val))
    } else {
      stop(err, call. = FALSE)
    }
  } else {
    stop(err, call. = FALSE)
  }
}

sumProdRxEnv <- function(expr) {
  ## Known functions
  calls <- allCalls(expr)
  callList <- setNames(lapply(calls, functionOp), calls)
  callEnv <- list2env(callList)
  currEnv <- cloneEnv(sumProdEnv, callEnv)
  names <- allNames(expr)
  ## Replace time with t.
  n1 <- names
  n2 <- names
  symbolList <- setNames(as.list(n2), n1)
  symbolEnv <- list2env(symbolList, parent = currEnv)
  return(symbolEnv)
}

rxSumProd <- function(x) {
  return(eval(x, sumProdRxEnv(x)))
}
#' Recast model in terms of sum/prod
#'
#' @param model rxode2 model
#' @param expand Boolean indicating if the expression is expanded.
#' @param sum Use sum(...)
#' @param prod Use prod(...)
#' @return model string with prod(.) and sum(.) for all these
#'     operations.
#' @author Matthew L. Fidler
#' @export
rxSumProdModel <- function(model, expand = FALSE, sum = TRUE, prod = TRUE) {
  ## Sum for pairwise is equivalent to regular sum under 8 elements.
  rxReq("symengine")
  assignInMyNamespace("rxSumProdSum", sum)
  assignInMyNamespace("rxSumProdProd", prod)
  on.exit({
    assignInMyNamespace("rxSumProdSum", FALSE)
    assignInMyNamespace("rxSumProdProd", FALSE)
  })
  .lines <- str2lang(paste0("{",rxNorm(model), "}"))
  .ret <- character(length(.lines) - 1)
  for (i in seq_along(.lines)[-1]) {
    .cur <- .lines[[i]]
    if (identical(.cur[[1]], quote(`=`)) ||
          identical(.cur[[1]], quote(`~`))) {
      if (expand) {
        .l2 <- deparse1(.cur[[3]])
        .l2 <- rxToSE(.l2)
        .l2 <- symengine::S(.l2)
        .l2 <- rxFromSE(.l2)
        .l2 <- str2lang(.l2)
      } else {
        .l2 <- .cur[[3]]
      }
      .l3 <- eval(bquote(rxSumProd(quote(.(.l2)))))
      if (inherits(.l3, "character")) {
        .cur[[3]] <- str2lang(.l3)
      }
    }
    .ret[i-1] <- deparse1(.cur)
  }
  paste(.ret, collapse = "\n")
}

rm(f)
rm(op)

#' Print expanded information about the rxode2 object.
#'
#' This prints the expanded information about the rxode2 object.
#'
#' @param object rxode2 object
#' @param ... Ignored parameters
#' @return object is returned
#' @author Matthew L.Fidler
#' @export
summary.rxode2 <- function(object, ...) {
  print.rxode2(object, rxSuppress = TRUE)
  summary.rxDll(object$cmpMgr$rxDll(), noprint = TRUE)
  invisible(object)
}

#' @export
summary.rxC <- function(object, ...) {
  cat(sprintf("//C file: %s\n", object))
  cat("//\n")
  .rl <- suppressWarnings(paste0(paste(readLines(object), collapse = "\n")))
  cat(.rl, "\n")
  invisible(.rl)
}

#' Summary of rxDll object
#'
#' This gives expanded information about the rxDll object
#'
#' @param object RxDll object
#'
#' @param ... Other arguments.  Includes `noprint`, which is a
#'     logical telling if the object should print the rxDll object
#'     first. By default this is FALSE
#'
#' @return object is returned
#'
#' @keywords internal
#' @author Matthew L.Fidler
#' @export
summary.rxDll <- function(object, ...) {
  .args <- as.list(match.call(expand.dots = TRUE))
  if (any(names(.args) == "noprint")) {
    .noprint <- .args$noprint
  } else {
    .noprint <- FALSE
  }
  if (!.noprint) {
    print(object)
  }
  cat(sprintf("DLL: %s\n", getOption("rxode2.dll.print", rxode2::rxDll(object))))
  cat(sprintf(
    "Jacobian: %s\n",
    ifelse(rxode2::rxModelVars(object)$jac == "fulluser", "Full User Specified",
      "Full Internally Calculated"
    )
  ))
  print(coef(object))
  if (length(rxode2::rxLhs(object)) > 0) {
    cat("\nCalculated Variables:\n")
    print(rxode2::rxLhs(object))
  }
  .mv <- rxModelVars(object)
  if (length(.mv$indLin) > 0) {
    cat(cli::cli_format_method({
      .h2("Inductive Linearization Matrix/Matrices:")
    }), "\n")
    print(.mv$indLin)
  }
  if (!is.na(object$linCmtM)) {
    .tmp <- object$linCmtM
  } else {
    .tmp <- setNames(rxode2::rxModelVars(object)$model["normModel"], NULL)
  }
  class(.tmp) <- "rxModelText"
  print(.tmp)
  return(invisible(object))
}


#' @export
summary.rxSolve <- function(object, ...) {
  if (rxIs(object, "rxSolve")) {
    cat(cli::cli_format_method({
      .h2(crayon::bold("Summary of Solved rxode2 object"))
    }), sep = "\n")
    .model <- object$model
    print(.model, .summary = TRUE)
    print(object, .summary = TRUE, ...)
  } else {
    class(object) <- "data.frame"
    NextMethod("summary", object)
  }
}

# This handles the errors for simulations
.createSimLineObject <- function(x, line) {
  .predDf <- get("predDf", x)
  if (line > nrow(.predDf)) {
    return(NULL)
  }
  .predLine <- .predDf[line, ]
  .ret <- list(x, .predLine, line)
  class(.ret) <- c(paste(.predLine$distribution), "rxGetDistributionSimulationLines")
  .ret
}

#' This is a S3 method for getting the distribution lines for a rxode2 simulation
#'
#' @param line Parsed rxode2 model environment
#' @return Lines for the simulation of `ipred` and `dv`. This is based
#'   on the idea that the focei parameters are defined
#' @author Matthew Fidler
#' @keywords internal
#' @export
rxGetDistributionSimulationLines <- function(line) {
  UseMethod("rxGetDistributionSimulationLines")
}

.simulationFun <- list(
  "t"="rt",
  "pois"="rpois",
  "binom"="rbinom",
  "beta"="rbeta",
  "chisq"="rchisq",
  "dexp"="rexp",
  "f"="rf",
  "geom"="rgeom",
  "hyper"="rhyper",
  "unif"="runif",
  "weibull"="rweibull",
  "cauchy"="rcauchy",
  "dgamma"="rgamma",
  "ordinal"="rordinal",
  "nbinom"="rnbinom",
  "nbinomMu"="rnbinomMu"
)

.getQuotedDistributionAndSimulationArgs <- function(line) {
  env <- line[[1]]
  pred1 <- line[[2]]
  .dist <- as.character(pred1$distribution)
  if (.dist == "LL") {
    return(env$lstExpr[[pred1$line]][[3]])
  }
  .nargs <- max(.errDist[[.dist]])
  .cnd <- pred1$cond
  .argName <- .namedArgumentsToPredDf[[.dist]]
  .args <- vapply(seq(1:.nargs), function(.i) {
    .curDist <- .argName[.i]
    if (!is.na(pred1[[.curDist]])) {
      return(pred1[[.curDist]])
    } else {
      .curDist <- paste0(.dist, ifelse(.i == 1, "", .i))
      .w <- which(env$iniDf$err == .curDist & env$iniDf$condition == .cnd)
      if (length(.w) == 1) {
        return(env$iniDf$name[.w])
      } else {
        return("")
      }
    }
  }, character(1))

  as.call(lapply(c(.simulationFun[[.dist]], .args[.args != ""]), str2lang))
}

#' Get the ar() correlation as a quoted expression for an endpoint
#'
#' @param env parsed model environment
#' @param pred1 single predDf row
#' @return quoted correlation (parameter name, or the modeled correlation
#'   variable), or NULL when the endpoint has no ar() term
#' @author Matthew Fidler
#' @noRd
.rxGetArCorLang <- function(env, pred1) {
  # ar() correlations backed by a parameter live in the $iniDf with err == "ar"
  # (a literal ar(0.5) is parsed into an auto-generated FIX parameter, an
  # estimated ar(rho) keeps its name); there is no dedicated $predDf column
  .w <- which(env$iniDf$err == "ar" & env$iniDf$condition == pred1$cond)
  if (length(.w) == 1L) return(str2lang(env$iniDf$name[.w]))
  # otherwise the correlation is a modeled (calculated) variable, e.g.
  # `corv <- expit(tcor); cp ~ add(add.sd) + ar(corv)` -- recover it directly
  # from the endpoint's original error expression
  .rxFindArArg(env, pred1$cond)
}

#' Find the `ar()` argument in an endpoint's error expression
#'
#' Scans the model's stored expressions (`env$lstExpr`) for the `~` error line
#' whose left-hand side matches the endpoint `cond` and returns the (quoted)
#' argument of its `ar()` term, if any.  Used to resolve modeled `ar()`
#' correlations, which are not stored in `$iniDf`/`$predDf`.
#'
#' @param env parsed model environment (must carry `lstExpr`)
#' @param cond endpoint condition (the error expression's left-hand side)
#' @return the quoted `ar()` argument, or NULL when not found
#' @author Matthew L. Fidler
#' @noRd
.rxFindArArg <- function(env, cond) {
  .lst <- tryCatch(get("lstExpr", envir=env), error=function(e) NULL)
  if (is.null(.lst)) return(NULL)
  .findAr <- function(e) {
    if (is.call(e)) {
      if (identical(e[[1]], quote(ar)) && length(e) == 2L) return(e[[2]])
      for (.i in seq_along(e)) {
        .r <- .findAr(e[[.i]])
        if (!is.null(.r)) return(.r)
      }
    }
    NULL
  }
  for (.e in .lst) {
    if (is.call(.e) && identical(.e[[1]], quote(`~`)) && length(.e) == 3L) {
      .lhs <- .e[[2]]
      # strip conditioning (lhs | cond) down to the endpoint variable
      if (is.call(.lhs) && identical(.lhs[[1]], quote(`|`))) .lhs <- .lhs[[2]]
      # unwrap ll(x)/linCmt() the same way .errHandleLlOrLinCmt() sets the
      # endpoint condition, so ll(cp) ~ ... + ar(corv) still matches cond
      if (is.call(.lhs)) {
        if (identical(.lhs[[1]], quote(`ll`)) && length(.lhs) == 2L) {
          .lhs <- .lhs[[2]]
        } else if (identical(.lhs[[1]], quote(`linCmt`))) {
          .lhs <- quote(`rxLinCmt`)
        }
      }
      if (identical(deparse1(.lhs), as.character(cond))) {
        .arg <- .findAr(.e[[3]])
        if (!is.null(.arg)) return(.arg)
      }
    }
  }
  NULL
}

#' Build the continuous-time AR(1) simulation lines for an endpoint
#'
#' Implements a stationary continuous-time AR(1) (Ornstein-Uhlenbeck)
#' process on the residual: the innovation is the endpoint's usual unit
#' draw, scaled so the marginal variance stays rx_r_ for any correlation.
#' The lag-1 correlation decays as cor^(time difference).  State is carried
#' across observation records with the same lag()/lag0() form the estimation
#' likelihood uses: rx.arT.<var> carries the previous time, lag0() reads the
#' previous residual for the self-referential recurrence, and
#' `1 - is.na(lag(...))` is the NaN-safe first-record indicator (0 on the first
#' record so phi=0 -> the marginal draw, 1 after).
#'
#' @param cor quoted correlation (name or number)
#' @param var endpoint variable suffix (e.g. "cp")
#' @param innovation quoted unit residual draw (rxerr.<var> or an r*() call)
#' @return list of quoted model lines; the final residual is left in
#'   rx.arRes.<var> and the sim line back-transforms rx_pred_ + that residual
#' @author Matthew Fidler
#' @noRd
.rxArSimLines <- function(cor, var, innovation) {
  .res <- str2lang(paste0("rx.arRes.", var))
  .t <- str2lang(paste0("rx.arT.", var))
  .dt <- str2lang(paste0("rx.arDt.", var))
  .nf <- str2lang(paste0("rx.arNf.", var))
  .phi <- str2lang(paste0("rx.arPhi.", var))
  list(
    bquote(.(.t) <- time),
    bquote(.(.dt) <- time - lag0(.(.t), 1)),
    bquote(.(.nf) <- 1 - is.na(lag(.(.t), 1))),
    bquote(.(.phi) <- .(.nf) * .(cor)^.(.dt)),
    bquote(.(.res) <- .(.phi) * lag0(.(.res), 1) +
             sqrt(rx_r_ * (1 - .(.phi)^2)) * .(innovation)),
    bquote(sim <- rxTBSi(rx_pred_ + .(.res), rx_lambda_, rx_yj_, rx_low_, rx_hi_)))
}

#' @rdname rxGetDistributionSimulationLines
#' @export
rxGetDistributionSimulationLines.norm <- function(line) {
  env <- line[[1]]
  pred1 <- line[[2]]
  .errNum <- line[[3]]
  .err <- str2lang(paste0("rxerr.", pred1$var))
  .arCor <- .rxGetArCorLang(env, pred1)
  if (is.null(.arCor)) {
    .ret <- vector("list", 2)
    .ret[[1]] <- bquote(ipredSim <- rxTBSi(rx_pred_, rx_lambda_, rx_yj_, rx_low_, rx_hi_))
    .ret[[2]] <- bquote(sim <- rxTBSi(rx_pred_+sqrt(rx_r_) * .(.err), rx_lambda_, rx_yj_, rx_low_, rx_hi_))
  } else {
    .ret <- c(list(bquote(ipredSim <- rxTBSi(rx_pred_, rx_lambda_, rx_yj_, rx_low_, rx_hi_))),
              .rxArSimLines(.arCor, pred1$var, .err))
  }
  c(.handleSingleErrTypeNormOrTFoceiBase(env, pred1, .errNum, rxPredLlik=FALSE), .ret)
}

#' @rdname rxGetDistributionSimulationLines
#' @export
rxGetDistributionSimulationLines.dnorm <- rxGetDistributionSimulationLines.norm

#' @rdname rxGetDistributionSimulationLines
#' @export
rxGetDistributionSimulationLines.t <- function(line) {
  env <- line[[1]]
  pred1 <- line[[2]]
  .errNum <- line[[3]]
  .draw <- .getQuotedDistributionAndSimulationArgs(line)
  .arCor <- .rxGetArCorLang(env, pred1)
  if (is.null(.arCor)) {
    .ret <- vector("list", 2)
    .ret[[1]] <- bquote(ipredSim <- rxTBSi(rx_pred_, rx_lambda_, rx_yj_, rx_low_, rx_hi_))
    .ret[[2]] <- bquote(sim <- rxTBSi(rx_pred_+sqrt(rx_r_) * .(.draw), rx_lambda_, rx_yj_, rx_low_, rx_hi_))
  } else {
    .ret <- c(list(bquote(ipredSim <- rxTBSi(rx_pred_, rx_lambda_, rx_yj_, rx_low_, rx_hi_))),
              .rxArSimLines(.arCor, pred1$var, .draw))
  }
  c(.handleSingleErrTypeNormOrTFoceiBase(env, pred1, .errNum, rxPredLlik=FALSE), .ret)
}

#' @rdname rxGetDistributionSimulationLines
#' @export
rxGetDistributionSimulationLines.cauchy <- function(line) {
  env <- line[[1]]
  pred1 <- line[[2]]
  .errNum <- line[[3]]
  .draw <- .getQuotedDistributionAndSimulationArgs(line)
  .arCor <- .rxGetArCorLang(env, pred1)
  if (is.null(.arCor)) {
    .ret <- vector("list", 2)
    .ret[[1]] <- bquote(ipredSim <- rxTBSi(rx_pred_, rx_lambda_, rx_yj_, rx_low_, rx_hi_))
    .ret[[2]] <- bquote(sim <- rxTBSi(rx_pred_+sqrt(rx_r_) * .(.draw), rx_lambda_, rx_yj_, rx_low_, rx_hi_))
  } else {
    .ret <- c(list(bquote(ipredSim <- rxTBSi(rx_pred_, rx_lambda_, rx_yj_, rx_low_, rx_hi_))),
              .rxArSimLines(.arCor, pred1$var, .draw))
  }
  c(.handleSingleErrTypeNormOrTFoceiBase(env, pred1, .errNum, rxPredLlik=FALSE), .ret)
}

#' @rdname rxGetDistributionSimulationLines
#' @export
rxGetDistributionSimulationLines.ordinal <- function(line) {
  .env <- line[[1]]
  .pred1 <- line[[2]]
  .errNum <- line[[3]]
  .c <- .env$lstExpr[[.pred1$line[1]]][[3]]
  .ce <- try(eval(.c), silent=TRUE)
  if (inherits(.ce, "try-error")) {
  } else if (inherits(.ce, "numeric") &&
               !is.null(names(.ce))) {
    .n <- names(.ce)
    .ln <- length(.n)
    if (.n[.ln] != "") {
      stop("last element in ordinal simulation of c(p1=0, p2=0.5, ...) must be a number, not a named number",
           call.=FALSE)
    }
    .n <- .n[.n != ""]
    if (length(.n) != .ln - 1) {
      stop("names for ordinal simulation incorrect")
    }
    .ret <- vector("list", 3)
    .ret[[1]] <- quote(ipredSim <- NA)
    .ret[[2]] <- str2lang(paste0("rx_sim_~rxord(", paste(.n, collapse=", "), ")"))
    .ce <- setNames(.ce, NULL)

    .ret[[3]] <- str2lang(paste0("sim<-", paste(vapply(seq_along(.ce), function(i) {
      paste("(rx_sim_ == ", i, ")*", .ce[i])
   }, character(1), USE.NAMES=FALSE), collapse="+")))
    return(.ret)
  }
  .c[[1]] <- quote(`rxord`)
  .ret <- vector("list", 2)
  .ret[[1]] <- quote(ipredSim <- NA)
  .ret[[2]] <- bquote(sim <- .(.c))
  .ret
}

#' @rdname rxGetDistributionSimulationLines
#' @export
rxGetDistributionSimulationLines.default <- function(line) {
  env <- line[[1]]
  pred1 <- line[[2]]
  .errNum <- line[[3]]
  .ret <- vector("list", 1)
  .ret[[1]] <- bquote(sim <- .(.getQuotedDistributionAndSimulationArgs(line)))
  c(.handleSingleErrTypeNormOrTFoceiBase(env, pred1, .errNum, rxPredLlik=FALSE), .ret)
}

#' @rdname rxGetDistributionSimulationLines
#' @export
rxGetDistributionSimulationLines.rxUi <- function(line) {
  .predDf <- get("predDf", line)
  lapply(seq_along(.predDf$cond), function(c) {
    .mod <- .createSimLineObject(line, c)
    rxGetDistributionSimulationLines(.mod)
  })
}

#' @export
#' @rdname rxUiGet
rxUiGet.cmtLines <- function(x, ...) {
  .x <- x[[1]]
  .len <- length(.x$mv0$state)
  .predDf <- get("predDf", .x)
  lapply(.predDf[.predDf$cmt > .len, "cond"], function(cmt) {
    call("cmt", str2lang(cmt))
  })
}
attr(rxUiGet.cmtLines, "desc") <- "cmt lines for model"

#' @export
#' @rdname rxUiGet
rxUiGet.dvidLine <- function(x, ...) {
  .x <- x[[1]]
  as.call(c(list(quote(`dvid`)), as.numeric(.x$predDf$cmt)))
}
attr(rxUiGet.dvidLine, "desc") <- "dvid() line for model"
attr(rxUiGet.dvidLine, "rstudio") <- quote(dvid(1, 2, 3)) # for rstudio completion

#' @export
#' @rdname rxUiGet
rxUiGet.paramsLine <- function(x, ...) {
  .x <- x[[1]]
  .iniDf <- .x$iniDf
  .params <- c(.iniDf[is.na(.iniDf$neta1) | .iniDf$neta1 == .iniDf$neta2, "name"],
               .x$covariates)
  eval(parse(text=paste0("quote(params(", paste(.params, collapse=", "), "))")))
}
attr(rxUiGet.paramsLine, "desc") <- "params() line for model"
attr(rxUiGet.paramsLine, "rstudio") <- quote(params(ID, CL, V, KA)) # for rstudio completion

#' @export
#' @rdname rxUiGet
rxUiGet.interpLines <- function(x, ...){
  .ui <- x[[1]]
  .interp <- rxModelVars(.ui)$interp
  if (!is.factor(.interp) ||
        length(.interp) == 0) {
    return(NULL)
  }
  .lvl <- levels(.interp)
  if (length(.lvl) != 5L ||
        !identical(.lvl, c("default", "linear", "locf", "nocb", "midpoint"))) {
    return(NULL)
  }
  if (any(is.na(.interp))) {
    return(NULL)
  }
  .try <- try(all(.interp == "default"), silent=TRUE)
  if (inherits(.try, "try-error") ||
        !checkmate::testLogical(.try, any.missing=FALSE, len=1)) {
    return(NULL)
  }
  if (.try) {
    # use default
    return(NULL)
  }
  .ret <- list()
  .w <- which(.interp=="linear")
  if (length(.w) > 0) {
    .ret <- list(str2lang(paste("linear(",paste(names(.interp)[.w], collapse=", "), ")")))
  }
  .w <- which(.interp=="locf")
  if (length(.w) > 0) {
    .ret <- c(.ret, list(str2lang(paste("locf(",paste(names(.interp)[.w], collapse=", "), ")"))))
  }
  .w <- which(.interp=="nocb")
  if (length(.w) > 0) {
    .ret <- c(.ret, list(str2lang(paste("nocb(",paste(names(.interp)[.w], collapse=", "), ")"))))
  }
  .w <- which(.interp=="midpoint")
  if (length(.w) > 0) {
    .ret <- c(.ret, list(str2lang(paste("midpoint(",paste(names(.interp)[.w], collapse=", "), ")"))))
  }
  if (length(.ret) == 0) {
    return(NULL) #nocov
  }
  .ret
}
attr(rxUiGet.interpLines, "desc") <- "interpolation declaration line(s) for model"
attr(rxUiGet.interpLines, "rstudio") <- quote(linear(CP)) # for rstudio completion

#' @rdname rxUiGet
#' @export
rxUiGet.splitDose <- function(x, ...) {
  .ui <- x[[1]]
  .splitBolus <- rxModelVars(.ui)$splitBolus
  if (length(.splitBolus) < 2L) {
    return(NULL)
  }
  .state <- rxModelVars(.ui)$state
  if (length(.state) == 0L ||
        any(.splitBolus < 1L | .splitBolus > length(.state))) {
    return(NULL)
  }
  list(as.call(c(list(quote(`splitBolus`)),
                 lapply(.state[.splitBolus], as.name))))
}
attr(rxUiGet.splitDose, "desc") <- "split dose declaration line(s) for model"
attr(rxUiGet.splitDose, "rstudio") <- quote(splitBolus(depot, central))

#' @rdname rxUiGet
#' @export
rxUiGet.splitDoseLines <- rxUiGet.splitDose
attr(rxUiGet.splitDoseLines, "desc") <- attr(rxUiGet.splitDose, "desc")
attr(rxUiGet.splitDoseLines, "rstudio") <- attr(rxUiGet.splitDose, "rstudio")

#' @export
#' @rdname rxUiGet
rxUiGet.simulationSigma <- function(x, ...) {
  .x <- x[[1]]
  .exact <- x[[2]]
  .predDf <- get("predDf", .x)
  .sigmaNames <- vapply(seq_along(.predDf$var), function(i) {
    if (.predDf$distribution[i] %in% c("dnorm",  "norm")) {
      paste0("rxerr.", .predDf$var[i])
    } else {
      ""
    }
  }, character(1))
  .sigmaNames <- .sigmaNames[.sigmaNames != ""]
  .sigma <- diag(length(.sigmaNames))
  dimnames(.sigma) <- list(.sigmaNames, .sigmaNames)
  .sigma
}
attr(rxUiGet.simulationSigma, "desc") <- "simulation sigma"
attr(rxUiGet.simulationSigma, "rstudio") <- lotri::lotri(a+b ~ c(1, .1, 1))

.simulationModelAssignTOS <- function(ui, ret) {
  assign("theta", ui$theta, envir=ret)
  assign("omega", ui$omega, envir=ret)
  assign("simulationSigma", ui$simulationSigma, envir=ret)
  assign("uiFun", as.function(ui), envir=ret)
  class(ret) <- c("rxode2tos", "rxode2")
  ret
}

#' @export
#' @rdname rxUiGet
rxUiGet.simulationModel <- function(x, ...) {
  .x <- x[[1]]
  ## meta is a reference-type environment shared across all rxUiDecompress
  ## calls for the same model, so it is the right place for per-model caches.
  ## All cached values are invalidated automatically when the model is modified
  ## via piping/.copyUi(), because those operations create a new meta env.
  .meta <- .x$meta
  .cached <- .meta$.simModelBase
  if (!is.null(.cached)) {
    return(.cached)
  }
  ## Cold path: build the compiled simulation model and cache the full TOS.
  ## theta/omega/simulationSigma/uiFun are derived from the rxUi which is
  ## immutable until the next piping step, so they are safe to cache.
  .cached <- eval(getBaseSimModel(.x))
  assign("theta",           .x$theta,           envir=.cached)
  assign("omega",           .x$omega,           envir=.cached)
  assign("simulationSigma", .x$simulationSigma, envir=.cached)
  assign("uiFun",           as.function(.x),    envir=.cached)
  class(.cached) <- c("rxode2tos", "rxode2")
  .meta$.simModelBase <- .cached
  .cached
}
attr(rxUiGet.simulationModel, "desc") <- "simulation model from UI"
attr(rxUiGet.simulationModel, "rstudio") <- quote(rxode2()) # for rstudio completion

#' @export
#' @rdname rxUiGet
rxUiGet.symengineModelNoPrune <- function(x, ...) {
  # For rstudio completion we need to do something else here
  .x <- x[[1]]
  .exact <- x[[2]]
  .simulationModelAssignTOS(.x, eval(getBaseSymengineModel(.x)))
}
attr(rxUiGet.symengineModelNoPrune, "desc") <- "symengine model without pruning if/else from UI"
attr(rxUiGet.symengineModelNoPrune, "rstudio") <- quote(rxode2()) # for rstudio completion

#' @export
#' @rdname rxUiGet
rxUiGet.symengineModelPrune <- function(x, ...) {
  .x <- x[[1]]
  .tmp <- getBaseSymengineModel(.x)
  .tmp[[1]] <- quote(`rxModelVars`)
  .tmp <- eval(.tmp)
  .tmp <- rxode2(rxPrune(.tmp))
  .simulationModelAssignTOS(.x, .tmp)
}
attr(rxUiGet.symengineModelPrune, "desc") <- "symengine model with pruning if/else from UI"
attr(rxUiGet.symengineModelPrune, "rstudio") <- quote(rxode2()) # for rstudio completion

#' @export
#' @rdname rxUiGet
rxUiGet.simulationIniModel <- function(x, ...) {
  .x <- x[[1]]
  .exact <- x[[2]]
  .simulationModelAssignTOS(.x, eval(getBaseIniSimModel(.x)))
}
attr(rxUiGet.simulationIniModel, "desc") <- "simulation model with the ini values prepended (from UI)"
attr(rxUiGet.simulationIniModel, "rstudio") <- quote(rxode2()) # for rstudio completion

.rxModelNoErrorLines <- function(uiModel, prefixLines=NULL, paramsLine=NULL,
                                 modelVars=FALSE, cmtLines=TRUE,
                                 lstExpr=NULL) {
  if (is.null(lstExpr)) {
    .expr <- uiModel$lstExpr
  } else {
    .expr <- lstExpr
  }
  .cmtLines <- NULL
  if (cmtLines) {
    .cmtLines <- uiModel$cmtLines
  }

  .lenLines <- length(uiModel$lstExpr) + length(.cmtLines) + length(prefixLines)

  .k <- 2

  if (is.null(paramsLine)) {
  } else if (is.na(paramsLine)) {
    .lenLines <- .lenLines - 1
  }
  .ret <- vector("list", .lenLines + .k)
  .curErrLine <- 1
  .ret[[1]] <- quote(`{`)
  .k <- 3
  if (is.null(paramsLine)) {
    .ret[[2]] <- uiModel$paramsLine
  } else if (is.na(paramsLine)) {
    .k <- 2
  } else {
    .ret[[2]] <- paramsLine
  }
  for (.i in seq_along(prefixLines)) {
    .ret[[.k]] <- prefixLines[[.i]]
    .k <- .k + 1
  }
  for (.i in seq_along(.expr)) {
    .ret[[.k]] <- .expr[[.i]]
    .k <- .k + 1
  }
  for(.i in seq_along(.cmtLines)) {
    .ret[[.k]] <- .cmtLines[[.i]]
    .k <- .k + 1
  }
  if (modelVars) {
    as.call(list(quote(`rxModelVars`), as.call(.ret)))
  } else {
    as.call(list(quote(`rxode2`), as.call(.ret)))
  }
}
#' Filter out properties and adjust predDf
#'
#' @param predDf predDf from the ui model
#' @param lstExpr list of expressions for model or NULL, if NULL
#'   defaults to the model expressions accessible by ui$lstExpr
#' @param ui original ui model or NULL
#' @return a list containing the adjusted predDf and lstExpr, or if
#'   predDf is null simply the lstExpr
#' @noRd
#' @author Matthew L. Fidler
.rxFilterOutPropsAndAdjustPredDf <- function(ui, lstExpr, predDf=NULL) {
  if (is.null(lstExpr)) {
    .expr <- ui$lstExpr
  } else {
    .expr <- lstExpr
  }
  ## remove directive lines from .expr; these are re-added in a
  ## normalized position for simulation models.
  .f <- vapply(seq_along(.expr), function(i) {
    .e <- .expr[[i]]
    if (is.call(.e)) {
      identical(.e[[1]], quote(`linear`)) ||
        identical(.e[[1]], quote(`locf`)) ||
        identical(.e[[1]], quote(`nocb`)) ||
        identical(.e[[1]], quote(`midpoint`)) ||
        identical(.e[[1]], quote(`splitBolus`))
    } else {
      FALSE
    }
  }, logical(1))
  .df <- data.frame(filter =.f, origLine=seq_along(.f))
  .predDf <- predDf
  if (any(.df$filter)) {
    .expr <- lapply(seq_along(.expr)[!.df$filter], function(i) {
      .expr[[i]]
    })
    .df <- .df[!.df$filter, , drop=FALSE]
    .df$newLine <- seq_along(.df$filter)
    if (!is.null(.predDf)) {
      .predDf$line <- vapply(.predDf$line, function(i) {
        .df$newLine[.df$origLine == i]
      }, integer(1))
    }
  }
  if (is.null(.predDf)) return(.expr)
  list(predDf=.predDf, lstExpr=.expr)
}

#' Combine Error Lines and create rxode2 expression
#'
#' @param uiModel UI model
#' @param errLines Error lines; If missing, get the error lines from
#'   `rxGetDistributionSimulationLines()`
#' @param prefixLines Prefix lines, after param statement
#' @param paramsLine Params line, if not present.
#' @param modelVars Return model vars instead of rxode2 statement
#' @param cmtLines Include trailing `cmt` lines
#' @param dvidLine Include trailing `dvid()` specification
#' @param lstExpr A list of expressions for model, or NULL.  When NULL
#'   defaults to the model expressions accessible by
#'   `uiModel$lstExpr`.
#' @param useIf Use an `if (CMT == X)` for endpoints
#' @param interpLines Interpolation lines, if not present
#' @param splitDoseLines Split dose lines, if not present
#' @param levelLines Levels lines for assigned strings.  If not
#'   present, use the interpolation lines from the current model.
#' @return quoted expression that can be evaluated to compiled rxode2
#'   model
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
#' @details
#'
#' This is exported to allow other functions to mangle the error lines
#' to make other types of estimation methods (if needed)
#'
#' @examples
#'
#' \donttest{
#'
#' one.cmt <- function() {
#'   ini({
#'     ## You may label each parameter with a comment
#'     tka <- 0.45 # Log Ka
#'     tcl <- log(c(0, 2.7, 100)) # Log Cl
#'     ## This works with interactive models
#'     ## You may also label the preceding line with label("label text")
#'     tv <- 3.45; label("log V")
#'     ## the label("Label name") works with all models
#'     eta.ka ~ 0.6
#'     eta.cl ~ 0.3
#'     eta.v ~ 0.1
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl + eta.cl)
#'     v <- exp(tv + eta.v)
#'     linCmt() ~ add(add.sd)
#'   })
#' }
#'
#' f <- rxode2(one.cmt)
#'
#' # You can get the simulation model easily by
#' rxCombineErrorLines(f)
#'
#' # You can then get the compiled model by simply evaluting the model:
#' r <- eval(rxCombineErrorLines(f))
#'
#' # This also works with multile endpoint models:
#' pk.turnover.emax <- function() {
#'   ini({
#'     tktr <- log(1)
#'     tka <- log(1)
#'     tcl <- log(0.1)
#'     tv <- log(10)
#'     ##
#'     eta.ktr ~ 1
#'     eta.ka ~ 1
#'     eta.cl ~ 2
#'     eta.v ~ 1
#'     prop.err <- 0.1
#'     pkadd.err <- 0.1
#'     ##
#'     temax <- logit(0.8)
#'     tec50 <- log(0.5)
#'     tkout <- log(0.05)
#'     te0 <- log(100)
#'     ##
#'     eta.emax ~ .5
#'     eta.ec50  ~ .5
#'     eta.kout ~ .5
#'     eta.e0 ~ .5
#'     ##
#'     pdadd.err <- 10
#'   })
#'   model({
#'     ktr <- exp(tktr + eta.ktr)
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl + eta.cl)
#'     v <- exp(tv + eta.v)
#'     ##
#'     emax=expit(temax+eta.emax)
#'     ec50 =  exp(tec50 + eta.ec50)
#'     kout = exp(tkout + eta.kout)
#'     e0 = exp(te0 + eta.e0)
#'     ##
#'     DCP = center/v
#'     PD=1-emax*DCP/(ec50+DCP)
#'     ##
#'     effect(0) = e0
#'     kin = e0*kout
#'     ##
#'     d/dt(depot) = -ktr * depot
#'     d/dt(gut) =  ktr * depot -ka * gut
#'     d/dt(center) =  ka * gut - cl / v * center
#'     d/dt(effect) = kin*PD -kout*effect
#'     ##
#'     cp = center / v
#'     cp ~ prop(prop.err) + add(pkadd.err)
#'     effect ~ add(pdadd.err)
#'   })
#' }
#'
#' f <- rxode2(pk.turnover.emax)
#' rxCombineErrorLines(f)
#'
#' # Note that in the parsed form, you can also get the compiled rxode2
#' # model with $simulationModel
#'
#' f$simulationModel
#'
#' }
rxCombineErrorLines <- function(uiModel, errLines=NULL, prefixLines=NULL, paramsLine=NULL,
                                modelVars=FALSE, cmtLines=TRUE, dvidLine=TRUE,
                                lstExpr=NULL,
                                useIf=TRUE,
                                interpLines=NULL, splitDoseLines=NULL,
                                levelLines=NULL) {
  if(!inherits(uiModel, "rxUi")) {
    stop("uiModel must be a evaluated UI model by rxode2(modelFunction) or modelFunction()",
         call.=FALSE)
  }
  uiModel <- rxUiDecompress(uiModel)
  .predDf <- uiModel$predDf
  if (is.null(.predDf)) {
    return(.rxModelNoErrorLines(uiModel, prefixLines=prefixLines, paramsLine=paramsLine,
                                modelVars=modelVars, cmtLines=cmtLines,
                                lstExpr=lstExpr))
  }
  if (is.null(errLines)) {
    errLines <- rxGetDistributionSimulationLines(uiModel)
  }
  .if <- FALSE
  if (length(.predDf$line) > 1) {
    .lenLines <- length(.predDf$line)
    .if <- useIf
  } else {
    .lenLines <- sum(vapply(seq_along(errLines), function(i) {
      length(errLines[[i]])
    }, integer(1)))
  }
  .cmtLines <- NULL
  if (cmtLines) {
    .cmtLines <- uiModel$cmtLines
  }
  .predDf <- uiModel$predDf
  .tmp <- .rxFilterOutPropsAndAdjustPredDf(uiModel, predDf=.predDf, lstExpr=lstExpr)
  .predDf <- .tmp$predDf
  .expr <- .tmp$lstExpr

  .lenLines <- .lenLines + length(.expr) -
    length(.predDf$line) + length(.cmtLines) + length(prefixLines)
  .k <- 2 + dvidLine * 1

  if (is.null(paramsLine)) {
  } else if (is.na(paramsLine)) {
    .lenLines <- .lenLines - 1
    .k <- 1 + dvidLine * 1
  }
  if (is.null(levelLines)) {
    .levelLines <- rxUiGet.levels(list(uiModel))
    .lenLines <- .lenLines - length(.levelLines)
    .k <- .k + length(.levelLines)
  } else if (is.na(levelLines)) {
    .levelLines <- list()
  } else {
    .levelLines <- levelLines
    .lenLines <- .lenLines - length(.levelLines)
    .k <- .k + length(.levelLines)
  }
  if (is.null(interpLines)) {
    .interpLines <- rxUiGet.interpLines(list(uiModel))
    .lenLines <- .lenLines - length(.interpLines)
    .k <- .k + length(.interpLines)
  } else if (is.na(interpLines)) {
    .interpLines <- list()
  } else {
    .interpLines <- interpLines
    .lenLines <- .lenLines - length(.interpLines)
    .k <- .k + length(.interpLines)
  }
  if (is.null(splitDoseLines)) {
    .splitDoseLines <- rxUiGet.splitDoseLines(list(uiModel))
    .lenLines <- .lenLines - length(.splitDoseLines)
    .k <- .k + length(.splitDoseLines)
  } else if (is.na(splitDoseLines)) {
    .splitDoseLines <- list()
  } else {
    .splitDoseLines <- splitDoseLines
    .lenLines <- .lenLines - length(.splitDoseLines)
    .k <- .k + length(.splitDoseLines)
  }
  .ret <- vector("list", .lenLines + .k)
  .curErrLine <- 1
  .ret[[1]] <- quote(`{`)
  .k <- 3
  if (is.null(paramsLine)) {
    .ret[[2]] <- uiModel$paramsLine
  } else if (is.na(paramsLine)) {
    .k <- 2
  } else {
    .ret[[2]] <- paramsLine
  }
  if (length(.levelLines) > 0) {
    for (.i in seq_along(.levelLines)) {
      .ret[[.k]] <- .levelLines[[.i]]
      .k <- .k + 1
    }
  }
  if (length(.interpLines) > 0) {
    for (.i in seq_along(.interpLines)) {
      .ret[[.k]] <- .interpLines[[.i]]
      .k <- .k + 1
    }
  }
  if (length(.splitDoseLines) > 0) {
    for (.i in seq_along(.splitDoseLines)) {
      .ret[[.k]] <- .splitDoseLines[[.i]]
      .k <- .k + 1
    }
  }
  for (.i in seq_along(prefixLines)) {
    .ret[[.k]] <- prefixLines[[.i]]
    .k <- .k + 1
  }
  for (.i in seq_along(.expr)) {
    if (.i %in% .predDf$line) {
      .curErr <- errLines[[.curErrLine]]
      if (.if) {
        .ret[[.k]] <- as.call(list(quote(`if`),
                                   as.call(list(quote(`==`), quote(`CMT`),
                                                as.numeric(.predDf$cmt[.curErrLine]))),
                                   as.call(c(list(quote(`{`)), .curErr))))
        .k <- .k + 1
      } else {
        for (.j in seq_along(.curErr)) {
          .ret[[.k]] <- .curErr[[.j]]
          .k <- .k + 1
        }
      }
      .curErrLine <- .curErrLine + 1
    } else {
      .ret[[.k]] <- .expr[[.i]]
      .k <- .k + 1
    }

  }
  for (.i in seq_along(.cmtLines)) {
    .ret[[.k]] <- .cmtLines[[.i]]
    .k <- .k + 1
  }
  if (dvidLine) {
    .ret[[.k]] <- uiModel$dvidLine
  }
  .drop <- which(vapply(seq_along(.ret), function(.i) {
    identical(.ret[[.i]], quote(`_drop`))
  }, logical(1), USE.NAMES=FALSE))
  if (length(.drop) > 0) {
    .ret <- .ret[-.drop]
  }
  if (modelVars) {
    as.call(list(quote(`rxModelVars`), as.call(.ret)))
  } else {
    as.call(list(quote(`rxode2`), as.call(.ret)))
  }
}

#' Split bolus doses across compartments inside an rxode2 model
#'
#' @description
#' `splitBolus()` is a model-only directive that rewrites bolus doses
#' aimed at `cmt` into parallel bolus doses to the target compartments.
#' Each target compartment receives the full original amount. The
#' source-compartment bolus is replaced, not retained.
#'
#' This rewrite applies both to event tables translated by [etTrans()]
#' and to future doses scheduled during solving with [evid_()].
#'
#' The source compartment may also appear in the target list, but the
#' target compartments in `...` must be unique.
#'
#' @param cmt Source compartment name.
#' @param ... Target compartment names. Provide at least one target
#'   compartment.
#'
#' @return This function is only meaningful inside an rxode2 model; it
#'   errors when called directly from R.
#'
#' @export
splitBolus <- function(cmt, ...) {
  stop("'splitBolus()' can only be used inside an rxode2 model block", call. = FALSE)
}

## Collect additive terms from an expression tree, tracking sign.
## Returns list of {sign=+-1, expr}.
.collectAddTerms <- function(expr, sign = 1L) {
  if (!is.call(expr)) return(list(list(sign = sign, expr = expr)))
  .fn <- expr[[1]]
  if (identical(.fn, quote(`(`)) && length(expr) == 2L) {
    return(.collectAddTerms(expr[[2]], sign))
  }
  if (identical(.fn, quote(`+`)) && length(expr) == 3L) {
    return(c(.collectAddTerms(expr[[2]], sign), .collectAddTerms(expr[[3]], sign)))
  }
  if (identical(.fn, quote(`-`)) && length(expr) == 3L) {
    return(c(.collectAddTerms(expr[[2]], sign), .collectAddTerms(expr[[3]], -sign)))
  }
  if (identical(.fn, quote(`-`)) && length(expr) == 2L) {
    return(.collectAddTerms(expr[[2]], -sign))
  }
  list(list(sign = sign, expr = expr))
}

## Return all state variable names referenced anywhere in an expression.
.statesInExpr <- function(expr, states) {
  if (is.name(expr)) {
    .nm <- as.character(expr)
    return(if (.nm %in% states) .nm else character(0))
  }
  if (!is.call(expr)) return(character(0))
  if (identical(expr[[1]], quote(`(`)) && length(expr) == 2L) {
    return(.statesInExpr(expr[[2]], states))
  }
  unique(unlist(lapply(as.list(expr)[-1], .statesInExpr, states = states)))
}

## Return all free symbol names referenced in an expression, excluding the
## heads of calls (i.e. function names).  Used to gather the PK parameter
## names appearing in ODE rate coefficients so they can be passed explicitly
## to linCmt().
.freeSymbolsInExpr <- function(expr) {
  if (is.name(expr)) return(as.character(expr))
  if (!is.call(expr)) return(character(0))
  unique(unlist(lapply(as.list(expr)[-1L], .freeSymbolsInExpr)))
}

## Extract the coefficient from `coef * stateNm` or `stateNm * coef`,
## where coef does not reference any state variable.
## Returns the coefficient expression, or NULL if the term is not that form.
.extractMultCoef <- function(expr, stateNm, states) {
  if (is.name(expr)) {
    if (as.character(expr) == stateNm) return(quote(1))
    return(NULL)
  }
  if (!is.call(expr)) return(NULL)
  .fn <- expr[[1]]
  if (identical(.fn, quote(`(`)) && length(expr) == 2L) {
    return(.extractMultCoef(expr[[2]], stateNm, states))
  }
  if (!identical(.fn, quote(`*`))) return(NULL)
  .lhs <- expr[[2]]; .rhs <- expr[[3]]
  .lu <- length(.statesInExpr(.lhs, states)) > 0L
  .ru <- length(.statesInExpr(.rhs, states)) > 0L
  if (.lu && !.ru && is.name(.lhs) && as.character(.lhs) == stateNm) return(.rhs)
  if (!.lu && .ru && is.name(.rhs) && as.character(.rhs) == stateNm) return(.lhs)
  NULL
}

## Parse one additive term: returns {sign, coef, state} where state=NA means
## a non-state (input/forcing) term. Returns NULL if the term is nonlinear
## (references multiple states or uses a state nonlinearly).
.parseOneLinTerm <- function(sign, termExpr, states) {
  .refs <- .statesInExpr(termExpr, states)
  if (length(.refs) == 0L) {
    return(list(sign = sign, coef = termExpr, state = NA_character_))
  }
  if (length(.refs) > 1L) return(NULL)
  .state <- .refs[1L]
  .coef  <- .extractMultCoef(termExpr, .state, states)
  if (is.null(.coef)) return(NULL)
  list(sign = sign, coef = .coef, state = .state)
}

## Parse an ODE RHS into a flat list of {sign, coef, state} terms.
## Returns NULL if the RHS is not linear in all state variables.
.parseLinearRhs <- function(rhs, states) {
  .raw    <- .collectAddTerms(rhs)
  .parsed <- lapply(.raw, function(.t) .parseOneLinTerm(.t$sign, .t$expr, states))
  if (any(vapply(.parsed, is.null, logical(1)))) return(NULL)
  .parsed
}

## Net signed coefficient (sum of sign*coef) of state `s`, as a single
## expression; NULL when `s` is absent.  The parsed sign is folded into the
## coefficient (`-ka*depot` parses as a positive-sign term with coef `-ka`).
.odeToLinNetCoef <- function(terms, s) {
  .rel <- Filter(function(.t) !is.na(.t$state) && .t$state == s, terms)
  if (length(.rel) == 0L) return(NULL)
  .signed <- lapply(.rel, function(.t) if (.t$sign < 0L) bquote(-(.(.t$coef))) else .t$coef)
  Reduce(function(.a, .b) bquote(.(.a) + .(.b)), .signed)
}

## Verify mass balance for every non-central compartment: a genuine
## depot/peripheral loses exactly the flux central gains from it (the two net
## coefficients cancel).  A metabolite has an extra independent-elimination
## term that breaks the cancellation, so it must NOT convert.  Checked
## numerically at two distinct positive parameter assignments.
.odeToLinMassBalanced <- function(odes, central, others) {
  .byCmt <- setNames(odes, vapply(odes, function(.o) .o$cmt, character(1)))
  .centralOde <- .byCmt[[central]]
  if (is.null(.centralOde)) return(FALSE)
  for (.c in others) {
    .ode <- .byCmt[[.c]]
    if (is.null(.ode)) return(FALSE)
    .selfNet <- .odeToLinNetCoef(.ode$terms, .c)        # what C loses (a net outflow)
    .centNet <- .odeToLinNetCoef(.centralOde$terms, .c) # what central gains from C
    if (is.null(.selfNet) || is.null(.centNet)) return(FALSE)
    .syms <- unique(c(.freeSymbolsInExpr(.selfNet), .freeSymbolsInExpr(.centNet)))
    .balanced <- function(.offset) {
      .vals <- as.list(setNames(seq_along(.syms) + .offset, .syms))
      .v <- tryCatch(eval(bquote(.(.selfNet) + .(.centNet)), .vals, baseenv()),
                     error = function(e) NA_real_)
      length(.v) == 1L && is.finite(.v) && abs(.v) < 1e-8
    }
    if (!.balanced(1.5) || !.balanced(3.7)) return(FALSE)
  }
  TRUE
}

## Detect the topology of a linear ODE system and classify each compartment.
## odes: list of {cmt, terms}; outputCmt: the central compartment (from the
## output line).  Returns list(ncmt, oral0, central, depot, peripheral1,
## peripheral2) or NULL.
.odeToLinDetectTopology <- function(odes, outputCmt, cmtNames) {
  .n <- length(odes)
  if (.n == 0L || .n > 4L) return(NULL)

  ## Build inflow map: flowsIn[[cmt]] = list of {from, coef} for positive
  ## cross-compartment terms in cmt's ODE.
  .flowsIn  <- setNames(vector("list", .n), cmtNames)
  .flowsOut <- setNames(vector("list", .n), cmtNames)

  for (.ode in odes) {
    .cmt <- .ode$cmt
    for (.t in .ode$terms) {
      if (is.na(.t$state) || .t$state == .cmt) next
      if (.t$sign > 0L) {
        .flowsIn[[.cmt]] <- c(.flowsIn[[.cmt]],
                              list(list(from = .t$state, coef = .t$coef)))
      }
    }
  }
  ## Derive outflow map from inflow map.
  for (.cmt in cmtNames) {
    for (.f in .flowsIn[[.cmt]]) {
      .flowsOut[[.f$from]] <- c(.flowsOut[[.f$from]], list(list(to = .cmt)))
    }
  }

  .central <- outputCmt
  if (!.central %in% cmtNames) return(NULL)

  .depot       <- NULL
  .peripherals <- character(0)

  for (.cmt in cmtNames[cmtNames != .central]) {
    .nIn  <- length(.flowsIn[[.cmt]])
    .out  <- .flowsOut[[.cmt]]
    .nOut <- length(.out)

    .allOutToCentral  <- .nOut > 0L &&
      all(vapply(.out, function(.f) .f$to == .central, logical(1)))
    .allInFromCentral <- .nIn > 0L &&
      all(vapply(.flowsIn[[.cmt]], function(.f) .f$from == .central, logical(1)))

    if (.nIn == 0L && .nOut == 1L && .allOutToCentral) {
      ## Depot: no inflows, exactly one outflow to central
      if (!is.null(.depot)) return(NULL)
      .depot <- .cmt
    } else if (.nIn > 0L && .allInFromCentral && .allOutToCentral) {
      ## Peripheral: inflow from central, outflow to central
      .peripherals <- c(.peripherals, .cmt)
    } else {
      return(NULL)
    }
  }

  .ncmt <- 1L + length(.peripherals)
  if (.ncmt > 3L) return(NULL)

  ## Reject systems where a depot/peripheral has independent loss (e.g. a
  ## metabolite), which linCmt() cannot represent.
  .others <- c(.peripherals, if (is.null(.depot)) character(0) else .depot)
  if (!.odeToLinMassBalanced(odes, .central, .others)) return(NULL)

  list(
    ncmt        = .ncmt,
    oral0       = if (is.null(.depot)) 0L else 1L,
    central     = .central,
    depot       = .depot,
    peripheral1 = if (length(.peripherals) >= 1L) .peripherals[1L] else NULL,
    peripheral2 = if (length(.peripherals) >= 2L) .peripherals[2L] else NULL
  )
}

## Find a line of the form  var <- cmt / vExpr  where cmt is a state variable.
## Returns {var, cmt, vExpr, lineIdx} or NULL.
.odeToLinFindOutput <- function(lstExpr, states) {
  for (.i in seq_along(lstExpr)) {
    .e <- lstExpr[[.i]]
    if (!is.call(.e)) next
    if (!identical(.e[[1]], quote(`<-`)) && !identical(.e[[1]], quote(`=`))) next
    if (length(.e) < 3L || !is.name(.e[[2]])) next
    .rhs <- .e[[3]]
    if (!is.call(.rhs) || length(.rhs) != 3L) next
    if (!identical(.rhs[[1]], quote(`/`))) next
    .num <- .rhs[[2]]
    if (!is.name(.num)) next
    .cmtNm <- as.character(.num)
    if (!.cmtNm %in% states) next
    return(list(
      var     = as.character(.e[[2]]),
      cmt     = .cmtNm,
      vExpr   = .rhs[[3]],
      lineIdx = .i
    ))
  }
  NULL
}

## Check if an expression is d/dt(name).
## In R's AST, `d/dt(x)` parses as call("/", d, call(dt, x)), NOT call("d/dt", x).
.isDtExpr <- function(expr) {
  is.call(expr) && length(expr) == 3L &&
    identical(expr[[1]], quote(`/`)) &&
    is.name(expr[[2]]) && identical(expr[[2]], quote(d)) &&
    is.call(expr[[3]]) && length(expr[[3]]) == 2L &&
    is.name(expr[[3]][[1]]) && as.character(expr[[3]][[1]]) == "dt" &&
    is.name(expr[[3]][[2]])
}

## Extract compartment name from d/dt(name) expression.
.getDtCmt <- function(expr) as.character(expr[[3]][[2]])

## Collect the names of state variables referenced as *values* in an
## expression.  The compartment-position argument of an `f`/`rate`/`dur`/`alag`
## modifier or an adaptive-dosing call (`bolus`, `infuse`, ...) is NOT a value
## reference: those positions are rewritten to the standard linCmt compartment
## names by .odeToLinRenameExpr, so they never block conversion.  Anything else
## that names a state -- e.g. a peripheral observable `periph` in
## `Cp <- periph / vp` -- is a value reference.
.odeToLinValueStateRefs <- function(expr, states) {
  if (is.name(expr)) {
    .nm <- as.character(expr)
    return(if (.nm %in% states) .nm else character(0))
  }
  if (!is.call(expr)) return(character(0))
  .fn <- if (is.name(expr[[1L]])) as.character(expr[[1L]]) else ""
  ## f/rate/dur/alag(<cmt>, ...): the first argument is a compartment position.
  if (.fn %in% c("f", "rate", "dur", "alag")) {
    .rest <- as.list(expr)[-1L]
    if (length(.rest) >= 1L) .rest <- .rest[-1L]
    return(unique(unlist(lapply(.rest, .odeToLinValueStateRefs, states = states))))
  }
  ## Adaptive dosing calls: the compartment-position argument is rewritten too.
  ## Indices mirror .odeToLinRenameAdaptiveCall.
  .cmtIdx <- switch(.fn,
    bolus = 3L, replace = 3L, multiply = 3L, phantom = 3L,
    infuse = 4L, infuseDur = 4L,
    `evid_` = 5L,
    NULL)
  .idx <- seq_along(expr)[-1L]
  if (!is.null(.cmtIdx)) .idx <- setdiff(.idx, .cmtIdx)
  unique(unlist(lapply(.idx, function(.k) .odeToLinValueStateRefs(expr[[.k]], states))))
}

## TRUE when a compartment state is referenced as a value by some model line
## other than the ODE equations and the single central output line (e.g. a
## peripheral observable `Cp <- periph / vp`, or a state alias).  Such a
## reference cannot be served by a bare central-only linCmt() output, so the
## caller takes the coupled-conversion path (rename the compartments to their
## canonical linCmt names) when it is safe, and otherwise keeps the explicit
## ODEs.
.odeToLinStateReferencedElsewhere <- function(lstExpr, cmtNames, odeIdx, outputIdx) {
  for (.i in seq_along(lstExpr)) {
    if (.i %in% odeIdx || .i == outputIdx) next
    if (length(.odeToLinValueStateRefs(lstExpr[[.i]], cmtNames)) > 0L) {
      return(TRUE)
    }
  }
  FALSE
}

## TRUE when every residual/endpoint (`~`) line predicts the central output
## variable.  A linCmt() model can only observe the central concentration as an
## estimated endpoint: mapping a peripheral observable to its own endpoint
## (`Cp ~ ... | peripheral1`) collides with linCmt()'s internal peripheral
## compartment ("required for linCmt() but defined in ODE too").  So a coupled
## system converts only when the peripheral references are *output-only*.
.odeToLinAllEndpointsCentral <- function(lstExpr, outputVar) {
  for (.e in lstExpr) {
    if (!is.call(.e) || !identical(.e[[1]], quote(`~`)) || length(.e) < 3L) next
    .lhs <- .e[[2]]
    if (!is.name(.lhs) || as.character(.lhs) != outputVar) return(FALSE)
  }
  TRUE
}

## Recursively rename every reference to an ODE compartment name to its
## canonical linCmt name throughout an expression (value references included,
## e.g. `Cp <- periph / vp` -> `Cp <- peripheral1 / vp`).  Used only on the
## coupled-conversion path; compartment names live in a separate namespace from
## parameters, so a name match is unambiguous.
.odeToLinRenameValueRefs <- function(e, cmtMap) {
  if (is.name(e)) {
    .new <- cmtMap[as.character(e)]
    if (!is.na(.new)) return(as.name(.new))
    return(e)
  }
  if (!is.call(e)) return(e)
  for (.i in seq_along(e)) e[[.i]] <- .odeToLinRenameValueRefs(e[[.i]], cmtMap)
  e
}

## Map the central endpoint to the `central` compartment (`Cc ~ err` ->
## `Cc ~ err | central`).  Anchoring the endpoint to the existing central
## compartment stops rxode2 injecting a fresh `cmt(<pred>)` observation slot,
## which would otherwise flip the model into explicit-compartment mode and make
## in-model peripheral references resolve to 0.  Left unchanged if already
## conditioned with `|`.
.odeToLinAddCentralCond <- function(e) {
  .rhs <- e[[3]]
  if (is.call(.rhs) && identical(.rhs[[1]], quote(`|`))) return(e)
  e[[3]] <- call("|", .rhs, as.name("central"))
  e
}

## Names of model-defined variables whose value transitively depends on a
## compartment state -- e.g. the central concentration `Cc <- central / vc`,
## or anything built from it.  A term in an ODE RHS that references one of
## these is state-dependent even when it names no state directly, so it must
## block linCmt() conversion (see `.odeToLinDetect`).  Lines defining the ODEs
## themselves are excluded; only assignments to a plain name are tracked.
.odeToLinStateDerivedVars <- function(lstExpr, states, odeIdx) {
  .tainted <- character(0)
  repeat {
    .added <- FALSE
    for (.i in seq_along(lstExpr)) {
      if (.i %in% odeIdx) next
      .e <- lstExpr[[.i]]
      if (!is.call(.e)) next
      if (!identical(.e[[1]], quote(`<-`)) && !identical(.e[[1]], quote(`=`))) next
      if (length(.e) < 3L || !is.name(.e[[2]])) next
      .lhs <- as.character(.e[[2]])
      if (.lhs %in% .tainted) next
      if (any(all.vars(.e[[3]]) %in% c(states, .tainted))) {
        .tainted <- c(.tainted, .lhs)
        .added <- TRUE
      }
    }
    if (!.added) break
  }
  unique(.tainted)
}

## Attempt to detect whether ui is a linear compartment ODE model.
## Returns a list with topology + output info, or NULL if not convertible.
.odeToLinDetect <- function(ui) {
  .lstExpr <- ui$lstExpr
  .states  <- rxModelVars(ui)$state # nolint

  if (length(.states) == 0L) return(NULL)

  ## Gather ODE lines and their compartment names.
  .odeIdx  <- integer(0)
  .cmtNames <- character(0)
  for (.i in seq_along(.lstExpr)) {
    .e <- .lstExpr[[.i]]
    if (!is.call(.e)) next
    if (!identical(.e[[1]], quote(`<-`)) && !identical(.e[[1]], quote(`=`))) next
    if (length(.e) < 3L || !is.call(.e[[2]])) next
    if (!.isDtExpr(.e[[2]])) next
    .odeIdx   <- c(.odeIdx, .i)
    .cmtNames <- c(.cmtNames, .getDtCmt(.e[[2]]))
  }

  if (length(.odeIdx) == 0L || length(.odeIdx) > 4L) return(NULL)
  if (!all(.cmtNames %in% .states)) return(NULL)

  ## Parse each ODE RHS; bail if any is nonlinear in state variables.
  .odes <- lapply(seq_along(.odeIdx), function(.j) {
    .e <- .lstExpr[[.odeIdx[.j]]]
    .terms <- .parseLinearRhs(.e[[3]], .states)
    if (is.null(.terms)) return(NULL)
    list(cmt = .cmtNames[.j], terms = .terms)
  })
  if (any(vapply(.odes, is.null, logical(1)))) return(NULL)

  ## Bail when an ODE RHS depends on a state indirectly through a state-derived
  ## value (e.g. Michaelis-Menten via `Cc <- central/vc`), which the direct
  ## linearity check misses; keep the explicit ODE states.
  .stateDerived <- .odeToLinStateDerivedVars(.lstExpr, .states, .odeIdx)
  if (length(.stateDerived) > 0L &&
        any(vapply(.odes, function(.o) {
          any(vapply(.o$terms,
                     function(.t) any(all.vars(.t$coef) %in% .stateDerived),
                     logical(1)))
        }, logical(1)))) {
    return(NULL)
  }

  ## Find output line: var <- centralCmt / vExpr
  .out <- .odeToLinFindOutput(.lstExpr, .states)
  if (is.null(.out)) return(NULL)

  ## Classify topology.
  .topo <- .odeToLinDetectTopology(.odes, .out$cmt, .cmtNames)
  if (is.null(.topo)) return(NULL)

  ## Compartment states referenced as a value outside the ODEs and the central
  ## output line (e.g. `Cp <- periph/vp`): keep the model analytic by renaming
  ## the ODE compartments to their canonical linCmt names (the coupled path),
  ## valid only when every endpoint predicts the central output.
  .coupled <- FALSE
  if (.odeToLinStateReferencedElsewhere(.lstExpr, .cmtNames, .odeIdx, .out$lineIdx)) {
    if (!.odeToLinAllEndpointsCentral(.lstExpr, .out$var)) return(NULL)
    .coupled <- TRUE
  }

  ## PK parameter names from the rate coefficients and volume expression,
  ## passed explicitly to linCmt() so it can infer the parameterization even
  ## for ini()-only parameters.
  .params <- character(0)
  for (.ode in .odes) {
    for (.t in .ode$terms) {
      if (!is.na(.t$state)) {
        .params <- c(.params, .freeSymbolsInExpr(.t$coef))
      }
    }
  }
  .params <- c(.params, .freeSymbolsInExpr(.out$vExpr))
  .params <- setdiff(unique(.params),
                     c(.states, .out$var, "t", "time", "pi"))

  c(.topo, list(
    outputVar = .out$var,
    outputCmt = .out$cmt,
    vExpr     = .out$vExpr,
    outputIdx = .out$lineIdx,
    odeIdx    = .odeIdx,
    params    = .params,
    coupled   = .coupled
  ))
}

## Build a new lstExpr with ODE lines removed and the output line replaced by
## `outputVar <- linCmt()`.
## Map ODE compartment names to the standard linCmt names ("depot", "central").
## linCmt() models always use these two names regardless of what the ODE called them.
.odeToLinCmtMap <- function(info) {
  .map <- character(0)
  if (isTRUE(info$oral0 == 1L) && !is.null(info$depot)) {
    .map[info$depot] <- "depot"
  }
  if (!is.null(info$central)) {
    .map[info$central] <- "central"
  }
  ## On the coupled path the peripheral compartments are referenced by retained
  ## model lines (observables), so they too must map to their canonical linCmt
  ## names to resolve against the analytic solution's solved compartments.
  if (isTRUE(info$coupled)) {
    if (!is.null(info$peripheral1)) .map[info$peripheral1] <- "peripheral1"
    if (!is.null(info$peripheral2)) .map[info$peripheral2] <- "peripheral2"
  }
  .map
}

## Apply compartment renaming to a single f/rate/dur/alag assignment line.
## Returns the line unchanged if it is not a modifier or its compartment is
## not in the mapping.
.odeToLinRenameCmt <- function(e, cmtMap) {
  if (!is.call(e)) return(e)
  if (!identical(e[[1]], quote(`<-`)) && !identical(e[[1]], quote(`=`))) return(e)
  if (length(e) < 3L || !is.call(e[[2]])) return(e)
  .fn <- as.character(e[[2]][[1]])
  if (!(.fn %in% c("f", "rate", "dur", "alag"))) return(e)
  if (length(e[[2]]) < 2L) return(e)
  .cmt <- as.character(e[[2]][[2]])
  .newCmt <- cmtMap[.cmt]
  if (is.na(.newCmt)) return(e)
  e[[2]][[2]] <- as.name(.newCmt)
  e
}

## Rename the compartment argument in an adaptive dosing call.  cmt AST index:
## bolus/replace/multiply/phantom -> e[[3]], infuse/infuseDur -> e[[4]],
## evid_ -> e[[5]]; obs()/reset() have no cmt and are left unchanged.
.odeToLinRenameAdaptiveCall <- function(e, cmtMap) {
  if (!is.call(e)) return(e)
  .fn <- as.character(e[[1]])
  .cmtIdx <- switch(.fn,
    bolus = 3L, replace = 3L, multiply = 3L, phantom = 3L,
    infuse = 4L, infuseDur = 4L,
    `evid_` = 5L,
    NULL
  )
  if (is.null(.cmtIdx)) return(e)
  if (length(e) < .cmtIdx) return(e)
  .cmtArg <- e[[.cmtIdx]]
  if (!is.name(.cmtArg)) return(e)
  .cmtNm <- as.character(.cmtArg)
  .newNm <- cmtMap[.cmtNm]
  if (is.na(.newNm)) return(e)
  e[[.cmtIdx]] <- as.name(.newNm)
  e
}

## Recursively rename compartments throughout an expression tree.
## Handles f/rate/dur/alag assignments and adaptive dosing calls.
## Recurses into if/block/other constructs to find nested calls.
.odeToLinRenameExpr <- function(e, cmtMap) {
  if (!is.call(e)) return(e)
  .fn <- as.character(e[[1]])
  if (.fn %in% c("<-", "=") && length(e) >= 3L && is.call(e[[2]])) {
    .innerFn <- as.character(e[[2]][[1]])
    if (.innerFn %in% c("f", "rate", "dur", "alag")) {
      return(.odeToLinRenameCmt(e, cmtMap))
    }
  }
  if (.fn %in% c("bolus", "infuse", "infuseDur", "replace", "multiply", "phantom", "evid_")) {
    return(.odeToLinRenameAdaptiveCall(e, cmtMap))
  }
  for (.i in seq_along(e)) {
    .child <- e[[.i]]
    if (is.call(.child)) {
      e[[.i]] <- .odeToLinRenameExpr(.child, cmtMap)
    }
  }
  e
}

.odeToLinBuildExpr <- function(lstExpr, info) {
  ## Build `linCmt(<params>)` with the detected PK parameter names so the
  ## parameterization is inferred even when parameters live only in ini().
  ## linCmt() resolves its parameters by name, so the argument order does not
  ## matter.  Fall back to a bare linCmt() if no parameters were detected.
  .linCmtCall <- if (length(info$params) > 0L) {
    as.call(c(list(as.name("linCmt")), lapply(info$params, as.name)))
  } else {
    str2lang("linCmt()")
  }
  .linCmtLine <- call("<-", as.name(info$outputVar), .linCmtCall)
  .cmtMap <- .odeToLinCmtMap(info)
  .ret <- list()
  for (.i in seq_along(lstExpr)) {
    if (.i %in% info$odeIdx) {
      next  # remove ODE lines
    } else if (.i == info$outputIdx) {
      .ret[[length(.ret) + 1L]] <- .linCmtLine  # replace output with linCmt()
    } else if (isTRUE(info$coupled)) {
      ## Coupled path: rename every compartment reference (values included) to
      ## its canonical linCmt name, and anchor the central endpoint to the
      ## central compartment so no observation compartment is injected.
      .e <- lstExpr[[.i]]
      if (is.call(.e) && identical(.e[[1]], quote(`~`)) && length(.e) >= 3L &&
          is.name(.e[[2]]) && as.character(.e[[2]]) == info$outputVar) {
        .e <- .odeToLinAddCentralCond(.e)
      }
      .ret[[length(.ret) + 1L]] <- .odeToLinRenameValueRefs(.e, .cmtMap)
    } else {
      .e <- .odeToLinRenameExpr(lstExpr[[.i]], .cmtMap)
      .ret[[length(.ret) + 1L]] <- .e
    }
  }
  .ret
}

## Package-scope cache: maps model-text key -> compiled converted rxUi.
## Prevents recompilation when rxSolve(..., useLinCmt=TRUE) is called repeatedly
## with the same model.
.odeToLinCache <- new.env(parent = emptyenv(), hash = TRUE)

## Cheap cache key: normalized text of all model lines plus the initial
## estimates.  The converted model rebuilt by `.rebuildRxUiFromExpr()` bakes in
## `ui$iniFun`, so two models that share the same `model({})` equations but
## differ only in their `ini({})` block must map to distinct cache entries.
## Keying on the equations alone caused the second model to reuse the first
## model's parameters (see the rxSolve(useLinCmt=TRUE) regression).
.odeToLinCacheKey <- function(ui) {
  .model <- paste0(vapply(ui$lstExpr, deparse1, character(1)), collapse = "\n")
  .ini <- if (length(ui$iniDf$cond) > 0L) deparse1(ui$iniFun) else ""
  paste0(.model, "\n#### ini ####\n", .ini)
}

## Package-scope cache: maps the same model-text key used by `.odeToLinCache`
## to the character vector of compartment names the conversion renamed away.
## Deriving these requires `rxModelVars()` on both the original and converted
## models, which is constant per converted model, so it is computed once here
## rather than on every solve.
.odeToLinLostCache <- new.env(parent = emptyenv(), hash = TRUE)

## The compartment names the ODE->linCmt conversion renamed away (i.e. present
## in the original model but not the converted one).  Cached by model key.
.odeToLinLostStates <- function(cacheKey, original, converted) {
  if (exists(cacheKey, envir = .odeToLinLostCache, inherits = FALSE)) {
    return(get(cacheKey, envir = .odeToLinLostCache, inherits = FALSE))
  }
  .lost <- setdiff(rxModelVars(original)$state, rxModelVars(converted)$state)
  assign(cacheKey, .lost, envir = .odeToLinLostCache)
  .lost
}

## Is a converted linCmt() model safe to use for the given solve data?
##
## The ODE->linCmt conversion renames compartments (e.g. an ODE `centre`
## compartment becomes linCmt's `central`).  If the event data addresses a
## compartment (in a dose or an observation record) by the *name* of an original
## ODE compartment that the converted model no longer has, that record would
## silently be routed nowhere -- producing all-zero predictions.  In that case
## fall back to the original ODE model so behaviour matches solving without the
## linCmt optimization.  `lost` is the (pre-computed) set of renamed-away
## compartment names.  Numeric compartment indices are preserved by the
## conversion, and names that are not compartments in the original model (e.g.
## the special "(default)" placeholder in an event table) are handled
## identically by both models, so both are left untouched here.
.odeToLinCmtCompatible <- function(lost, data) {
  if (length(lost) == 0L) {
    return(TRUE)
  }
  if (is.null(data) || !is.data.frame(data)) {
    return(TRUE)
  }
  .nm <- names(data)
  .col <- .nm[tolower(.nm) == "cmt"]
  if (length(.col) == 0L) {
    return(TRUE)
  }
  .cmt <- data[[.col[1L]]]
  if (is.factor(.cmt)) {
    .cmt <- as.character(.cmt)
  }
  if (!is.character(.cmt)) {
    ## numeric indices are preserved by the conversion
    return(TRUE)
  }
  .cmt <- unique(.cmt[!is.na(.cmt)])
  if (length(.cmt) == 0L) {
    return(TRUE)
  }
  !any(.cmt %in% lost)
}

## Rebuild an rxUi from a modified lstExpr (following the linToOde pattern).
.rebuildRxUiFromExpr <- function(ui, expr) {
  .ls     <- ls(ui$meta, all.names = TRUE)
  .hasIni <- length(ui$iniDf$cond) > 0L
  .ret <- vector("list", length(.ls) + if (.hasIni) 3L else 2L)
  .ret[[1L]] <- quote(`{`)
  for (.i in seq_along(.ls)) {
    .ret[[.i + 1L]] <- rxUiDeparse(ui$meta[[.ls[.i]]], .ls[.i]) # nolint
  }
  .len <- length(.ls)
  if (.hasIni) {
    .ret[[.len + 2L]] <- ui$iniFun
    .ret[[.len + 3L]] <- bquote(model(.(as.call(c(quote(`{`), expr)))))
  } else {
    .ret[[.len + 2L]] <- bquote(model(.(as.call(c(quote(`{`), expr)))))
  }
  .fun <- function() {}
  body(.fun) <- as.call(.ret)
  if (is.function(ui$model)) environment(.fun) <- environment(ui$model)
  suppressMessages(as.rxUi(.fun)) # nolint
}

#' Convert ODE-based linear compartment models to analytical linCmt() form
#'
#' Detects whether a model's ODE equations form a standard 1-3 compartment PK
#' system and, if so, replaces the \code{d/dt()} equations and the central
#' output assignment (e.g. \code{cp <- central/v}) with \code{cp <- linCmt()},
#' preserving all other model lines.  Detection requires linear ODE
#' right-hand sides, 1-4 compartments in a standard depot/central/peripheral
#' topology, and one output line of the form \code{var <- central_cmt / v_expr}.
#' Named parameter assignments are retained so \code{linCmt()} can infer the
#' parameterization.
#'
#' @param ui rxUi-like model object (function, rxUi, or anything accepted by
#'   \code{as.rxUi}).
#' @return An rxUi model with ODE equations replaced by \code{linCmt()}, or
#'   (with a message) the original model unchanged if conversion is not possible.
#' @seealso \code{\link{linToOde}} for the inverse transformation.
#' @examples
#' oneCmtOde <- function() {
#'   ini({
#'     tka <- 0.45
#'     tcl <- log(2.7)
#'     tv <- 3.45
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka  <- exp(tka)
#'     cl  <- exp(tcl)
#'     v   <- exp(tv)
#'     d/dt(depot)   <- -ka * depot
#'     d/dt(central) <- ka * depot - cl/v * central
#'     cp <- central / v
#'     cp ~ add(add.sd)
#'   })
#' }
#' linCmtModel <- odeToLin(oneCmtOde)
#'
#' @author Matthew Fidler
#' @export
odeToLin <- function(ui) {
  .ui <- as.rxUi(ui) # nolint
  .ui <- rxUiDecompress(.ui) # nolint

  .info <- .odeToLinDetect(.ui)
  if (is.null(.info)) {
    message("model does not appear to be a linear compartment ODE; returning unchanged")
    return(rxUiCompress(.ui)) # nolint
  }

  .expr <- .odeToLinBuildExpr(.ui$lstExpr, .info)
  .rebuildRxUiFromExpr(.ui, .expr)
}

rxTest({

  ## Helper: tight-tolerance ODE vs linCmt comparison
  .chkConv <- function(mOde, ev, tol = 1e-5, seed = 1L) {
    .conv <- suppressMessages(odeToLin(mOde))
    set.seed(seed); .rOde <- suppressMessages(rxSolve(mOde,  ev, useLinCmt = FALSE))
    set.seed(seed); .rLin <- suppressMessages(rxSolve(.conv, ev))
    list(conv = .conv, diff = max(abs(.rOde$cp - .rLin$cp), na.rm = TRUE), tol = tol)
  }

  ## -----------------------------------------------------------------------
  ## Detection: models that SHOULD convert
  ## -----------------------------------------------------------------------

  test_that("odeToLin detects 1-cmt IV (no depot)", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tcl <- 1; tv <- 3.45; add.sd <- 0.7 })
      model({
        cl <- exp(tcl); v <- exp(tv)
        d/dt(central) <- -cl/v * central
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    .info <- .odeToLinDetect(.m)
    expect_equal(.info$ncmt,  1L)
    expect_equal(.info$oral0, 0L)
  })

  test_that("odeToLin detects 1-cmt oral", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tka <- 0.45; tcl <- 1; tv <- 3.45; add.sd <- 0.7 })
      model({
        ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
        d/dt(depot)   <- -ka * depot
        d/dt(central) <- ka * depot - cl/v * central
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    .info <- .odeToLinDetect(.m)
    expect_equal(.info$ncmt,  1L)
    expect_equal(.info$oral0, 1L)
  })

  test_that("odeToLin detects 2-cmt IV", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tcl <- 1; tv <- 3.45; tq <- 0.5; tvp <- 6; add.sd <- 0.7 })
      model({
        cl <- exp(tcl); v <- exp(tv); q <- exp(tq); vp <- exp(tvp)
        d/dt(central) <- -(cl+q)/v * central + q/vp * periph
        d/dt(periph)  <- q/v * central - q/vp * periph
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    .info <- .odeToLinDetect(.m)
    expect_equal(.info$ncmt,  2L)
    expect_equal(.info$oral0, 0L)
  })

  test_that("odeToLin detects 2-cmt oral", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tka <- 0.45; tcl <- 1; tv <- 3.45; tq <- 0.5; tvp <- 6; add.sd <- 0.7 })
      model({
        ka <- exp(tka); cl <- exp(tcl); v <- exp(tv); q <- exp(tq); vp <- exp(tvp)
        d/dt(depot)   <- -ka * depot
        d/dt(central) <- ka * depot - (cl+q)/v * central + q/vp * periph
        d/dt(periph)  <- q/v * central - q/vp * periph
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    .info <- .odeToLinDetect(.m)
    expect_equal(.info$ncmt,  2L)
    expect_equal(.info$oral0, 1L)
  })

  test_that("odeToLin detects 3-cmt IV", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tcl <- 1; tv <- 3; tq <- 0.5; tvp <- 6; tq2 <- 0.3; tvp2 <- 10; add.sd <- 0.7 })
      model({
        cl <- exp(tcl); v <- exp(tv); q <- exp(tq); vp <- exp(tvp)
        q2 <- exp(tq2); vp2 <- exp(tvp2)
        d/dt(central) <- -(cl+q+q2)/v * central + q/vp * periph1 + q2/vp2 * periph2
        d/dt(periph1) <- q/v * central - q/vp * periph1
        d/dt(periph2) <- q2/v * central - q2/vp2 * periph2
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    .info <- .odeToLinDetect(.m)
    expect_equal(.info$ncmt,  3L)
    expect_equal(.info$oral0, 0L)
  })

  test_that("odeToLin detects 3-cmt oral", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tka <- 0.45; tcl <- 1; tv <- 3; tq <- 0.5; tvp <- 6;
            tq2 <- 0.3; tvp2 <- 10; add.sd <- 0.7 })
      model({
        ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
        q <- exp(tq); vp <- exp(tvp); q2 <- exp(tq2); vp2 <- exp(tvp2)
        d/dt(depot)   <- -ka * depot
        d/dt(central) <- ka * depot - (cl+q+q2)/v * central + q/vp * periph1 + q2/vp2 * periph2
        d/dt(periph1) <- q/v * central - q/vp * periph1
        d/dt(periph2) <- q2/v * central - q2/vp2 * periph2
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    .info <- .odeToLinDetect(.m)
    expect_equal(.info$ncmt,  3L)
    expect_equal(.info$oral0, 1L)
  })

  ## -----------------------------------------------------------------------
  ## Detection: models that should NOT convert
  ## -----------------------------------------------------------------------

  test_that("odeToLin returns NULL for nonlinear ODE", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tcl <- 1; tv <- 3.45; km <- 10; add.sd <- 0.7 })
      model({
        cl <- exp(tcl); v <- exp(tv); km <- exp(km)
        d/dt(central) <- -cl * central / (km + central)
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    expect_null(.odeToLinDetect(.m))
  })

  test_that("odeToLin returns NULL for wrong topology (two depots)", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tka1 <- 0.4; tka2 <- 0.2; tcl <- 1; tv <- 3.45; add.sd <- 0.7 })
      model({
        ka1 <- exp(tka1); ka2 <- exp(tka2); cl <- exp(tcl); v <- exp(tv)
        d/dt(depot1)  <- -ka1 * depot1
        d/dt(depot2)  <- -ka2 * depot2
        d/dt(central) <- ka1 * depot1 + ka2 * depot2 - cl/v * central
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    expect_null(.odeToLinDetect(.m))
  })

  test_that("odeToLin returns NULL when output line is nonlinear in central", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tcl <- 1; tv <- 3.45; add.sd <- 0.7 })
      model({
        cl <- exp(tcl); v <- exp(tv)
        d/dt(central) <- -cl/v * central
        cp <- central^2 / v
        cp ~ add(add.sd)
      })
    }))
    expect_null(.odeToLinDetect(.m))
  })

  ## A genuinely nonlinear model whose nonlinearity is hidden behind a
  ## state-derived observable must NOT auto-linearize.  Here Michaelis-Menten
  ## elimination is written through the central concentration
  ## `Cc <- central / vc` -- `... - vmax * Cc * vc / (km + Cc)` -- so the
  ## linearity scan (which looks only for *direct* state references) saw no
  ## state and treated the term as a constant forcing input.  useLinCmt=TRUE
  ## (the default) then folded the 2-cmt system into linCmt(), dropped the MM
  ## term, and demoted k12/k21 to required inputs, so rxSolve aborted with
  ## "parameter(s) are required for solving: k21, k12".
  .mmObs <- function() {
    suppressMessages(rxode2(function() {
      ini({ lka <- -0.5; lcl <- -2; lvc <- 2; lvp <- 1.5; lq <- -1.6
            lvmax <- 1; lkm <- 0.2; addSd <- 0.02 })
      model({
        ka <- exp(lka); cl <- exp(lcl); vc <- exp(lvc); vp <- exp(lvp); q <- exp(lq)
        vmax <- exp(lvmax); km <- exp(lkm)
        Cc <- central / vc
        kel <- cl / vc; k12 <- q / vc; k21 <- q / vp
        d/dt(depot)       <- -ka * depot
        d/dt(central)     <-  ka * depot - kel * central - k12 * central +
          k21 * peripheral1 - vmax * Cc * vc / (km + Cc)
        d/dt(peripheral1) <-  k12 * central - k21 * peripheral1
        Cc ~ add(addSd)
      })
    }))
  }

  test_that("odeToLin returns NULL for MM elimination written via the central observable", {
    expect_null(.odeToLinDetect(.mmObs()))
  })

  test_that("default rxSolve keeps the explicit ODE states for an MM-via-observable model", {
    .m  <- .mmObs()
    .ev <- et(amt = 100, cmt = "depot") |> et(seq(0, 24, by = 4), cmt = "Cc")
    .rDef <- suppressMessages(rxSolve(.m, .ev))                  # default useLinCmt=TRUE
    .rOde <- suppressMessages(rxSolve(.m, .ev, useLinCmt = FALSE))
    # the coupled peripheral state survives and the nonlinear (MM) solve agrees
    expect_true("peripheral1" %in% names(.rDef))
    expect_equal(.rDef$Cc, .rOde$Cc, tolerance = 1e-6)
  })

  ## -----------------------------------------------------------------------
  ## Correctness: converted model matches ODE (tight tolerances)
  ## -----------------------------------------------------------------------

  test_that("odeToLin 1-cmt oral conversion is numerically correct", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tka <- 0.45; tcl <- 1; tv <- 3.45;
            eta.ka ~ 0.6; eta.cl ~ 0.3; eta.v ~ 0.1; add.sd <- 0.7 })
      model({
        ka <- exp(tka + eta.ka); cl <- exp(tcl + eta.cl); v <- exp(tv + eta.v)
        d/dt(depot)   <- -ka * depot
        d/dt(central) <- ka * depot - cl/v * central
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    .ev <- et(amt = 100, ii = 24, addl = 3) |> et(seq(0, 96, by = 1))
    .res <- .chkConv(.m, .ev)
    expect_equal(rxModelVars(.res$conv)$state, c("depot", "central"))
    expect_lt(.res$diff, .res$tol)
  })

  test_that("odeToLin 2-cmt oral conversion is numerically correct", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tka <- 0.45; tcl <- 1; tv <- 3; tq <- 0.5; tvp <- 6;
            eta.ka ~ 0.6; eta.cl ~ 0.3; eta.v ~ 0.1; add.sd <- 0.7 })
      model({
        ka <- exp(tka + eta.ka); cl <- exp(tcl + eta.cl); v <- exp(tv + eta.v)
        q <- exp(tq); vp <- exp(tvp)
        d/dt(depot)   <- -ka * depot
        d/dt(central) <- ka * depot - (cl+q)/v * central + q/vp * periph
        d/dt(periph)  <- q/v * central - q/vp * periph
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    .ev <- et(amt = 100, ii = 24, addl = 3) |> et(seq(0, 96, by = 1))
    .res <- .chkConv(.m, .ev)
    expect_equal(rxModelVars(.res$conv)$state, c("depot", "central"))
    expect_lt(.res$diff, .res$tol)
  })

  test_that("odeToLin 3-cmt oral conversion is numerically correct", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tka <- 0.45; tcl <- 1; tv <- 3; tq <- 0.5; tvp <- 6;
            tq2 <- 0.3; tvp2 <- 10;
            eta.ka ~ 0.6; eta.cl ~ 0.3; eta.v ~ 0.1; add.sd <- 0.7 })
      model({
        ka <- exp(tka + eta.ka); cl <- exp(tcl + eta.cl); v <- exp(tv + eta.v)
        q <- exp(tq); vp <- exp(tvp); q2 <- exp(tq2); vp2 <- exp(tvp2)
        d/dt(depot)   <- -ka * depot
        d/dt(central) <- ka * depot - (cl+q+q2)/v * central + q/vp * periph1 + q2/vp2 * periph2
        d/dt(periph1) <- q/v * central - q/vp * periph1
        d/dt(periph2) <- q2/v * central - q2/vp2 * periph2
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    .ev <- et(amt = 100, ii = 24, addl = 3) |> et(seq(0, 96, by = 1))
    .res <- .chkConv(.m, .ev)
    expect_equal(rxModelVars(.res$conv)$state, c("depot", "central"))
    expect_lt(.res$diff, .res$tol)
  })

  ## -----------------------------------------------------------------------
  ## Non-standard compartment names (renamed to depot/central)
  ## -----------------------------------------------------------------------

  test_that("odeToLin renames non-standard compartment names to linCmt standard", {
    .m <- suppressMessages(rxode2(function() {
      ini({
        tka <- 0.45; tcl <- 1; tv <- 3.45;
        eta.ka ~ 0.6; eta.cl ~ 0.3; eta.v ~ 0.1; add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka); cl <- exp(tcl + eta.cl); v <- exp(tv + eta.v)
        d/dt(depot)  <- -ka * depot
        d/dt(center) <- ka * depot - cl/v * center
        cp <- center / v
        cp ~ add(add.sd)
      })
    }))
    .conv <- suppressMessages(odeToLin(.m))
    expect_equal(rxModelVars(.conv)$state, c("depot", "central"))

    .ev <- et(amt = 100, ii = 24, addl = 2) |> et(seq(0, 72, by = 1))
    set.seed(1); .rOde <- suppressMessages(rxSolve(.m,    .ev, useLinCmt = FALSE))
    set.seed(1); .rLin <- suppressMessages(rxSolve(.conv, .ev))
    expect_lt(max(abs(.rOde$cp - .rLin$cp), na.rm = TRUE), 1e-5)
  })

  test_that("odeToLin renames f/rate modifiers for non-standard compartment names", {
    .m <- suppressMessages(rxode2(function() {
      ini({
        tka <- 0.45; trate <- 0.4; tcl <- 1; tv <- 3.45
        fDepot <- logit(0.5)
        eta.ka ~ 0.6; eta.cl ~ 0.3; eta.v ~ 0.1; add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka); cl <- exp(tcl + eta.cl); v <- exp(tv + eta.v)
        d/dt(depot)  <- -ka * depot
        d/dt(center) <- ka * depot - cl/v * center
        f(depot)     <- expit(fDepot)
        f(center)    <- 1 - expit(fDepot)
        rate(center) <- exp(trate)
        cp <- center / v
        cp ~ add(add.sd)
      })
    }))
    .conv <- tryCatch(suppressMessages(odeToLin(.m)), error = function(e) e)
    expect_false(inherits(.conv, "error"),
                 label = "odeToLin with f/rate on non-standard cmt name must not error")
    expect_equal(rxModelVars(.conv)$state, c("depot", "central"))
  })

  test_that("useLinCmt=TRUE solve works for model with non-standard cmt names + f/rate", {
    # Regression: test-et.R 'Fix for #732' -- model uses 'center' not 'central'
    .m <- suppressMessages(rxode2(function() {
      ini({
        tka <- 0.45; trate <- 0.4; tcl <- 1; tv <- 3.45
        fDepot <- logit(0.5)
        eta.ka ~ 0.6; eta.cl ~ 0.3; eta.v ~ 0.1; add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka); cl <- exp(tcl + eta.cl); v <- exp(tv + eta.v)
        d/dt(depot)  <- -ka * depot
        d/dt(center) <- ka * depot - cl/v * center
        f(depot)     <- expit(fDepot)
        f(center)    <- 1 - expit(fDepot)
        rate(center) <- exp(trate)
        cp <- center / v
        cp ~ add(add.sd)
      })
    }))
    .ev <- et(amt = 320, evid = 1, cmt = 1, time = 0) |>
      et(amt = 320, evid = 1, cmt = 2, time = 0) |>
      et(amt = 320, evid = 4, time = 72, cmt = 1) |>
      et(amt = 320, evid = 1, time = 72, cmt = 2) |>
      et(seq(0, 100, by = 1)) |>
      et(id = 1:4)
    .d <- suppressMessages(rxSolve(.m, .ev, addDosing = TRUE, useLinCmt = TRUE,
                                   seed = 123))
    .times <- sort(unique(.d[.d$evid == 1L & .d$cmt == 2L, "time"]))
    expect_equal(.times, c(0, 72))
  })

  ## -----------------------------------------------------------------------
  ## useLinCmt flag controls ODE vs linCmt path
  ## -----------------------------------------------------------------------

  test_that("useLinCmt flag controls whether ODE or linCmt solver is used", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tka <- 0.45; tcl <- 1; tv <- 3.45; add.sd <- 0.7 })
      model({
        ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
        d/dt(depot)   <- -ka * depot
        d/dt(central) <- ka * depot - cl/v * central
        cp <- central / v
        cp ~ add(add.sd)
      })
    }))
    .ev <- et(amt = 100, ii = 24, addl = 1) |> et(seq(0, 48, by = 1))
    .mLin <- suppressMessages(odeToLin(.m))

    .getDll <- function(r) {
      # normalizePath() canonicalizes the path so that the comparison is robust
      # to Windows 8.3 short-name mangling: the .rxd cache directory name is too
      # long for an 8.3 path, so one solve can report it in long form
      # (rx_..._x64_.rxd) while another reports the shortened form
      # (RX_...~1.RXD).  Both point to the same DLL; normalizing makes them equal.
      normalizePath(
        get("dll", envir = attr(attr(r, "class"), ".rxode2.env"), inherits = FALSE),
        winslash = "/", mustWork = FALSE)
    }

    # useLinCmt=TRUE must use the same DLL as explicit odeToLin conversion
    .rLinDirect <- suppressMessages(rxSolve(.mLin, .ev))
    .rLinAuto   <- suppressMessages(rxSolve(.m,    .ev, useLinCmt = TRUE))
    expect_equal(.getDll(.rLinAuto), .getDll(.rLinDirect),
                 label = "useLinCmt=TRUE must use same DLL as odeToLin conversion")

    # useLinCmt=FALSE must use a different DLL (the original ODE model)
    .rOde <- suppressMessages(rxSolve(.m, .ev, useLinCmt = FALSE))
    expect_false(identical(.getDll(.rOde), .getDll(.rLinAuto)),
                 label = "useLinCmt=FALSE must use original ODE DLL, not linCmt DLL")
  })

  ## -----------------------------------------------------------------------
  ## Explicit linCmt() parameters (works when params are only in ini())
  ## -----------------------------------------------------------------------

  test_that("odeToLin emits explicit linCmt() parameter names (1-cmt oral)", {
    # Parameters live only in ini() (no model-body assignment).  The
    # converted model must pass the parameter names explicitly to linCmt()
    # so the parameterization can still be inferred.
    .m <- suppressMessages(rxode2(function() {
      ini({ ka <- 0.5; cl <- 1; v <- 10 })
      model({
        d/dt(depot)   <- -ka * depot
        d/dt(central) <- ka * depot - cl/v * central
        cp <- central / v
      })
    }))
    .info <- .odeToLinDetect(.m)
    expect_setequal(.info$params, c("ka", "cl", "v"))

    .conv <- suppressMessages(odeToLin(.m))
    .linLine <- Find(function(x) grepl("linCmt", deparse1(x)), .conv$lstExpr)
    # Bare linCmt() (no args) would fail param inference for ini-only params
    expect_false(identical(deparse1(.linLine[[3L]]), "linCmt()"))
    expect_true(grepl("linCmt(", deparse1(.linLine), fixed = TRUE))

    .ev <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    .rOde <- suppressMessages(rxSolve(.m,    .ev, useLinCmt = FALSE))
    .rLin <- suppressMessages(rxSolve(.conv, .ev))
    expect_equal(.rOde$cp, .rLin$cp, tolerance = 1e-5)
  })

  test_that("odeToLin emits explicit linCmt() parameter names (2-cmt IV)", {
    .m <- suppressMessages(rxode2(function() {
      ini({ tcl <- log(1); tv <- log(10); tq <- log(0.5); tvp <- log(20) })
      model({
        cl <- exp(tcl); v <- exp(tv); q <- exp(tq); vp <- exp(tvp)
        d/dt(central)    <- -(cl+q)/v * central + q/vp * peripheral
        d/dt(peripheral) <-  q/v * central - q/vp * peripheral
        cp <- central / v
      })
    }))
    .info <- .odeToLinDetect(.m)
    expect_setequal(.info$params, c("cl", "v", "q", "vp"))

    .conv <- suppressMessages(odeToLin(.m))
    .ev <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    .rOde <- suppressMessages(rxSolve(.m,    .ev, useLinCmt = FALSE))
    .rLin <- suppressMessages(rxSolve(.conv, .ev))
    expect_equal(.rOde$cp, .rLin$cp, tolerance = 1e-5)
  })

  test_that("rxSolve(useLinCmt=TRUE) transparently solves ini-only-param ODE models", {
    # Regression: the transparent useLinCmt auto-conversion must not break an
    # otherwise-valid ODE model whose parameters are defined only in ini().
    .m <- suppressMessages(rxode2(function() {
      ini({ ka <- 0.5; cl <- 1; v <- 10 })
      model({
        d/dt(depot)   <- -ka * depot
        d/dt(central) <- ka * depot - cl/v * central
        cp <- central / v
      })
    }))
    .ev <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    .rAuto <- suppressMessages(rxSolve(.m, .ev))                   # useLinCmt=TRUE default
    .rOde  <- suppressMessages(rxSolve(.m, .ev, useLinCmt = FALSE))
    expect_true(nrow(.rAuto) > 0)
    expect_false(any(is.na(.rAuto$cp)))
    expect_equal(.rAuto$cp, .rOde$cp, tolerance = 1e-5)
  })

  ## -----------------------------------------------------------------------
  ## odeToLin + adaptive dosing
  ## -----------------------------------------------------------------------

  test_that("odeToLin preserves bolus() in model and converted model runs", {
    # An ODE model with in-model bolus push. odeToLin must convert the ODE to
    # linCmt AND keep the bolus() call intact. PK parameters must be model-body
    # assignments so linCmt() can infer the parameterization.
    .m <- suppressMessages(rxode2(function() {
      ini({ tka <- log(0.5); tcl <- log(1); tv <- log(10) })
      model({
        ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
        d/dt(depot)   <- -ka * depot
        d/dt(central) <- ka * depot - cl/v * central
        cp <- central / v
      })
    }))
    .mLin <- suppressMessages(odeToLin(.m))
    # Converted model should use linCmt
    .mLinTxt <- paste(vapply(.mLin$lstExpr, deparse1, character(1)), collapse = "\n")
    expect_true(grepl("linCmt", .mLinTxt))

    # Solve both without adaptive dosing to confirm trajectories agree
    .ev <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    .rOde <- suppressMessages(rxSolve(.m,    .ev, useLinCmt = FALSE))
    .rLin <- suppressMessages(rxSolve(.mLin, .ev))
    expect_equal(.rOde$cp, .rLin$cp, tolerance = 1e-5,
                 label = "odeToLin result matches ODE solution")
  })

  test_that("odeToLin preserves standard depot/central compartment in bolus()", {
    # Model uses standard names (depot, central) -- odeToLin should
    # compile and solve cleanly; adaptive dosing calls are preserved as-is.
    .m <- suppressMessages(rxode2(function() {
      ini({ tka <- log(0.5); tcl <- log(1); tv <- log(10); add.sd <- 0.1 })
      model({
        ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
        d/dt(depot)   <- -ka * depot
        d/dt(central) <- ka * depot - cl/v * central
        cp <- central / v
        if (t > 23 && t < 25) {
          bolus(50, cmt = depot)
        }
        cp ~ add(add.sd)
      })
    }))
    .mLin <- suppressMessages(odeToLin(.m))
    .mLinTxt <- paste(vapply(.mLin$lstExpr, deparse1, character(1)), collapse = "\n")
    expect_true(grepl("linCmt", .mLinTxt))

    .ev <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    .rLin <- suppressMessages(rxSolve(.mLin, .ev))
    expect_true(nrow(.rLin) > 0)
    expect_false(any(is.na(.rLin$cp)))
    expect_true(all(.rLin$cp >= 0))
    expect_true(all(.rLin$cp[.rLin$time > 0] > 0))
  })

  test_that("odeToLin renames non-standard ODE compartment names in bolus()", {
    # Model uses non-standard compartment names (abs, center).
    # odeToLin must rename them to depot/central in both linCmt() and bolus().
    .m <- suppressMessages(rxode2(function() {
      ini({ tka <- log(0.5); tcl <- log(1); tv <- log(10) })
      model({
        ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
        d/dt(abs)    <- -ka * abs
        d/dt(center) <- ka * abs - cl/v * center
        cp <- center / v
        if (t > 23 && t < 25) {
          bolus(50, cmt = abs)
        }
      })
    }))
    .mLin <- suppressMessages(odeToLin(.m))
    # Confirm the converted model no longer references the old compartment names
    .mLinTxt <- paste(vapply(.mLin$lstExpr, deparse1, character(1)), collapse = "\n")
    expect_true(grepl("linCmt", .mLinTxt))
    expect_false(grepl("\\babs\\b",    .mLinTxt),
                 label = "non-standard cmt name 'abs' should be renamed in converted model")

    .ev <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    .rLin <- suppressMessages(rxSolve(.mLin, .ev))
    expect_true(nrow(.rLin) > 0)
    expect_false(any(is.na(.rLin$cp)))
    expect_true(all(.rLin$cp >= 0))
    expect_true(all(.rLin$cp[.rLin$time > 0] > 0))
  })

  ## -----------------------------------------------------------------------
  ## Coupled / separately-observed states.
  ##
  ## A central sub-system can be linear-compartment-eligible yet have an
  ## additional state referenced by another model line (a peripheral observable
  ## like `Cp <- periph / vp`, possibly with its own endpoint).  When the
  ## coupled state is an output-only peripheral, odeToLin() converts it by
  ## renaming the ODE compartments to their canonical linCmt names and anchoring
  ## the central endpoint to the central compartment (see the conversion tests
  ## below).  When linCmt() cannot represent the system -- a compartment with
  ## independent loss (metabolite), or a peripheral with its own estimated
  ## endpoint -- detection returns NULL so the explicit ODE states are preserved.
  ## -----------------------------------------------------------------------

  test_that("odeToLin returns NULL for a metabolite-like coupled, observed state", {
    # Parent 2-cmt + observed metabolite with its own elimination.  The
    # metabolite exchanges with central (kbt) so the topology *looks* like a
    # peripheral, but it is separately observed (Cc_mhd) and carries an
    # independent elimination -- it must remain a solved ODE state.
    .m <- suppressMessages(rxode2(function() {
      ini({ lka <- log(1); lcl <- log(5); lvc <- log(30); lq <- log(2); lvp <- log(50)
            lclm <- log(3); lvcm <- log(20); lkbt <- log(0.5); propSd <- 0.1; propSdM <- 0.1 })
      model({
        ka <- exp(lka); cl <- exp(lcl); vc <- exp(lvc); q <- exp(lq); vp <- exp(lvp)
        cl_mhd <- exp(lclm); vc_mhd <- exp(lvcm); kbt <- exp(lkbt)
        kel <- cl/vc; kelm <- cl_mhd/vc_mhd; k12 <- q/vc; k21 <- q/vp
        d/dt(depot)       <- -ka*depot
        d/dt(central)     <-  ka*depot - (kel+k12)*central + k21*peripheral1 + kbt*central_mhd
        d/dt(peripheral1) <-  k12*central - k21*peripheral1
        d/dt(central_mhd) <-  kel*central - kelm*central_mhd - kbt*central_mhd
        f(depot) <- 1
        Cc     <- central / vc
        Cc_mhd <- central_mhd / vc_mhd
        Cc     ~ prop(propSd)
        Cc_mhd ~ prop(propSdM)
      })
    }))
    expect_null(.odeToLinDetect(.m))
  })

  test_that("odeToLin converts a coupled output-peripheral model and stays numerically correct", {
    # Regression: useLinCmt=TRUE (the default) used to auto-linearize this model,
    # drop `periph`, and abort with "parameter(s) are required for solving:
    # periph".  It now CONVERTS analytically instead: `periph` is renamed to the
    # canonical `peripheral1`, the central endpoint is anchored to the `central`
    # compartment (so no observation compartment is injected), and the peripheral
    # observable reads linCmt()'s solved peripheral amount -- so the default solve
    # stays fully analytic and agrees with the explicit ODE (useLinCmt=FALSE).
    .m <- suppressMessages(rxode2(function() {
      ini({ lcl <- log(5); lvc <- log(30); lq <- log(2); lvp <- log(50); propSd <- 0.1 })
      model({
        cl <- exp(lcl); vc <- exp(lvc); q <- exp(lq); vp <- exp(lvp)
        kel <- cl/vc; k12 <- q/vc; k21 <- q/vp
        d/dt(central) <- -kel*central - k12*central + k21*periph
        d/dt(periph)  <-  k12*central - k21*periph
        Cc <- central / vc
        Cp <- periph / vp
        Cc ~ prop(propSd)
      })
    }))
    .info <- .odeToLinDetect(.m)
    expect_true(isTRUE(.info$coupled))
    .conv    <- suppressMessages(odeToLin(.m))
    .convTxt <- paste(vapply(.conv$lstExpr, deparse1, character(1)), collapse = "\n")
    expect_true(grepl("linCmt", .convTxt))
    expect_true(grepl("peripheral1", .convTxt))   # periph renamed to canonical name
    expect_true(grepl("[|] *central", .convTxt))  # central endpoint anchored
    expect_false(grepl("d/dt", .convTxt))         # no ODE left

    .ev <- et(amt = 100, cmt = "central") |> et(seq(0, 24, by = 2))
    .rDef <- suppressMessages(rxSolve(.m, .ev))                  # default -> converts
    .rOde <- suppressMessages(rxSolve(.m, .ev, useLinCmt = FALSE))
    expect_true("Cp" %in% names(.rDef))
    expect_equal(.rDef$Cc, .rOde$Cc, tolerance = 1e-4)
    expect_equal(.rDef$Cp, .rOde$Cp, tolerance = 1e-4)           # peripheral observable correct
  })

  test_that("odeToLin converts a 2-cmt oral coupled model (depot amount + peripheral)", {
    .m <- suppressMessages(rxode2(function() {
      ini({ lka <- log(0.8); lcl <- log(5); lvc <- log(30); lq <- log(2); lvp <- log(50); propSd <- 0.1 })
      model({
        ka <- exp(lka); cl <- exp(lcl); vc <- exp(lvc); q <- exp(lq); vp <- exp(lvp)
        kel <- cl/vc; k12 <- q/vc; k21 <- q/vp
        d/dt(depot)   <- -ka*depot
        d/dt(central) <-  ka*depot - kel*central - k12*central + k21*periph
        d/dt(periph)  <-  k12*central - k21*periph
        Cc     <- central / vc
        Cp     <- periph / vp
        Adepot <- depot            # depot amount as a secondary output
        Cc ~ prop(propSd)
      })
    }))
    .info <- .odeToLinDetect(.m)
    expect_true(isTRUE(.info$coupled))
    .ev <- et(amt = 100, cmt = "depot") |> et(seq(0, 24, by = 2))
    .rDef <- suppressMessages(rxSolve(.m, .ev))
    .rOde <- suppressMessages(rxSolve(.m, .ev, useLinCmt = FALSE))
    expect_equal(.rDef$Cc,     .rOde$Cc,     tolerance = 1e-4)
    expect_equal(.rDef$Cp,     .rOde$Cp,     tolerance = 1e-4)
    expect_equal(.rDef$Adepot, .rOde$Adepot, tolerance = 1e-4)
  })

  test_that("odeToLin returns NULL for an output-only metabolite (mass-balance guard)", {
    # Single endpoint (would pass the endpoint guard), but the metabolite carries
    # an independent elimination (kelm) on top of the kbt it returns to central,
    # so mass is not conserved across the exchange.  linCmt() cannot represent
    # that, so the system must stay an explicit ODE.
    .m <- suppressMessages(rxode2(function() {
      ini({ lcl <- log(5); lvc <- log(30); lkpm <- log(1); lkbt <- log(0.5)
            lclm <- log(3); lvcm <- log(20); propSd <- 0.1 })
      model({
        cl <- exp(lcl); vc <- exp(lvc); kpm <- exp(lkpm); kbt <- exp(lkbt)
        cl_mhd <- exp(lclm); vc_mhd <- exp(lvcm)
        kel <- cl/vc; kelm <- cl_mhd/vc_mhd
        d/dt(central) <- -kel*central - kpm*central + kbt*met
        d/dt(met)     <-  kpm*central - kelm*met - kbt*met   # extra independent loss
        Cc <- central / vc
        Cm <- met / vc_mhd
        Cc ~ prop(propSd)
      })
    }))
    expect_null(.odeToLinDetect(.m))
  })

  test_that("default rxSolve handles a multi-endpoint coupled (fetus) model", {
    # Two endpoints (Cc and Cfetus); the second is keyed to a coupled state.
    .m <- suppressMessages(rxode2(function() {
      ini({ lcl <- log(5); lvc <- log(30); lq <- log(2); lvp <- log(50)
            lqmf <- log(1); lvf <- log(10); propSd <- 0.1; propSdF <- 0.1 })
      model({
        cl <- exp(lcl); vc <- exp(lvc); q <- exp(lq); vp <- exp(lvp)
        qmf <- exp(lqmf); vfetus <- exp(lvf)
        kel <- cl/vc; k12 <- q/vc; k21 <- q/vp; k1f <- qmf/vc; kf1 <- qmf/vfetus
        d/dt(central)     <- -kel*central - k12*central + k21*peripheral1 - k1f*central + kf1*fetus
        d/dt(peripheral1) <-  k12*central - k21*peripheral1
        d/dt(fetus)       <-  k1f*central - kf1*fetus
        Cc     <- central / vc
        Cfetus <- fetus / vfetus
        Cc     ~ prop(propSd)
        Cfetus ~ prop(propSdF)
      })
    }))
    expect_null(.odeToLinDetect(.m))
    .ev <- et(amt = 100, cmt = "central") |> et(seq(0, 24, by = 2), cmt = "Cc")
    .rDef <- suppressMessages(rxSolve(.m, .ev))
    .rOde <- suppressMessages(rxSolve(.m, .ev, useLinCmt = FALSE))
    expect_true(all(c("peripheral1", "fetus", "Cfetus") %in% names(.rDef)))
    expect_equal(.rDef$Cfetus, .rOde$Cfetus, tolerance = 1e-6)
  })

  test_that("odeToLin still converts when an observable derives from the central output", {
    # Boundary: a non-state quantity built from the central *output variable*
    # (cp, not the `central` state) must NOT block conversion.
    .m <- suppressMessages(rxode2(function() {
      ini({ tcl <- 1; tv <- 3.45; emax <- 10; ec50 <- 1; add.sd <- 0.7 })
      model({
        cl <- exp(tcl); v <- exp(tv)
        d/dt(central) <- -cl/v * central
        cp  <- central / v
        eff <- emax * cp / (ec50 + cp)   # references cp (output var), not the state
        cp ~ add(add.sd)
      })
    }))
    .info <- .odeToLinDetect(.m)
    expect_equal(.info$ncmt, 1L)
  })

  test_that("a hand-written linCmt() model reads a peripheral compartment under an error model", {
    # UI-system fix (rxUiGet.cmtLines): an endpoint injects a fresh observation
    # compartment, which previously shifted linCmt()'s materialized compartment
    # indices so an in-equation `peripheral1` read 0.  The peripheral observable
    # must now resolve to the solved amount even with an error model present and
    # WITHOUT anchoring the central endpoint to the central compartment.  This is
    # the general fix that also lets odeToLin keep coupled models analytic.
    .m <- suppressMessages(rxode2(function() {
      ini({ lcl <- log(5); lvc <- log(30); lq <- log(2); lvp <- log(50); propSd <- 0.1 })
      model({
        cl <- exp(lcl); vc <- exp(lvc); q <- exp(lq); vp <- exp(lvp)
        Cc <- linCmt()
        Cp <- peripheral1 / vp     # in-equation reference to a linCmt compartment
        Cc ~ prop(propSd)          # error model present, central endpoint NOT anchored
      })
    }))
    .mOde <- suppressMessages(rxode2(function() {
      ini({ lcl <- log(5); lvc <- log(30); lq <- log(2); lvp <- log(50) })
      model({
        cl <- exp(lcl); vc <- exp(lvc); q <- exp(lq); vp <- exp(lvp)
        kel <- cl/vc; k12 <- q/vc; k21 <- q/vp
        d/dt(central) <- -kel*central - k12*central + k21*periph
        d/dt(periph)  <-  k12*central - k21*periph
        Cc <- central / vc
        Cp <- periph / vp
      })
    }))
    .ev   <- et(amt = 100, cmt = "central") |> et(seq(0, 24, by = 2))
    .r    <- suppressMessages(rxSolve(.m,    .ev))
    .rOde <- suppressMessages(rxSolve(.mOde, .ev, useLinCmt = FALSE))
    expect_false(all(.r$Cp == 0))                    # the bug produced an all-zero Cp
    expect_equal(.r$Cp, .rOde$Cp, tolerance = 1e-4)  # peripheral amount now resolves
    expect_equal(.r$Cc, .rOde$Cc, tolerance = 1e-4)
  })

  test_that("odeToLin cache distinguishes models that differ only in ini()", {
    # Regression: the ODE->linCmt conversion cache was keyed on the model
    # equations alone.  Two models with identical model({}) but different
    # ini({}) collided, so the second rxSolve(useLinCmt=TRUE) reused the first
    # model's parameters -- yielding identical predictions regardless of ini().
    .mk <- function(tclVal, tvVal) {
      .f <- function() {
        ini({ tcl <- 0; tv <- 0; add.err <- 1 })
        model({
          cl <- exp(tcl); v <- exp(tv); k <- cl / v
          d/dt(centre) <- -k * centre
          cp <- centre / v
          cp ~ add(add.err)
        })
      }
      .f <- rxode2::ini(.f, tcl = tclVal, tv = tvVal)
      suppressMessages(rxode2(.f))
    }
    .u1 <- .mk(-4.75, 0.22)
    .u2 <- .mk(-2.60, -1.44)
    # Default (unnamed) dosing so the linCmt conversion is actually adopted; a
    # named `cmt = "centre"` would trigger the compatibility fallback and the
    # cache path under test would never run.
    .ev <- et(amt = 100) |> et(seq(0, 24, by = 4))

    .getDll <- function(r) {
      normalizePath(
        get("dll", envir = attr(attr(r, "class"), ".rxode2.env"), inherits = FALSE),
        winslash = "/", mustWork = FALSE)
    }

    # Solve u1 first (populates the cache), then u2 (previously collided)
    .s1 <- suppressWarnings(rxSolve(.u1, .ev))
    .s2 <- suppressWarnings(rxSolve(.u2, .ev))
    .g1 <- suppressWarnings(rxSolve(.u1, .ev, useLinCmt = FALSE))
    .g2 <- suppressWarnings(rxSolve(.u2, .ev, useLinCmt = FALSE))

    # The linCmt conversion must actually be in play, otherwise the cache path
    # is never exercised: the default solve must use a different DLL than the
    # ODE solve.
    expect_false(identical(.getDll(.s1), .getDll(.g1)),
                 label = "useLinCmt default must adopt the converted linCmt DLL")

    expect_false(isTRUE(all.equal(.s1$cp, .s2$cp)))  # must differ by ini()
    # Each must match its own explicit-parameter / ODE solve
    expect_equal(.s1$cp, .g1$cp, tolerance = 1e-4)
    expect_equal(.s2$cp, .g2$cp, tolerance = 1e-4)
  })

  test_that("auto linCmt conversion is skipped when data doses a named ODE compartment", {
    # Regression: odeToLin renames the ODE compartment (e.g. `centre`) to the
    # linCmt compartment (`central`).  Event data that doses the compartment by
    # its ODE name lost the dose after conversion, producing all-zero
    # predictions.  The default solve must fall back to the ODE model and match
    # useLinCmt = FALSE.
    .mod <- suppressMessages(rxode2(function() {
      ini({ tcl <- -4.75; tv <- 0.22; add.err <- 1 })
      model({
        cl <- exp(tcl); v <- exp(tv); k <- cl / v
        d/dt(centre) <- -k * centre
        cp <- centre / v
        cp ~ add(add.err)
      })
    }))
    .ev <- data.frame(ID = 1L,
                      TIME = c(0, 2, 4, 8, 12, 24),
                      AMT = c(100, 0, 0, 0, 0, 0),
                      EVID = c(1L, 0L, 0L, 0L, 0L, 0L),
                      CMT = "centre",
                      stringsAsFactors = FALSE)
    .rLin <- suppressWarnings(rxSolve(.mod, .ev))                     # default (useLinCmt=TRUE)
    .rOde <- suppressWarnings(rxSolve(.mod, .ev, useLinCmt = FALSE))
    expect_false(all(.rLin$cp == 0))            # the bug produced all-zero cp
    expect_equal(.rLin$cp, .rOde$cp, tolerance = 1e-4)
  })

})

#' Register a method for a suggested dependency
#'
#' Generally, the recommend way to register an S3 method is to use the
#' `S3Method()` namespace directive (often generated automatically by the
#' `@export` roxygen2 tag). However, this technique requires that the generic
#' be in an imported package, and sometimes you want to suggest a package,
#' and only provide a method when that package is loaded. `s3_register()`
#' can be called from your package's `.onLoad()` to dynamically register
#' a method only if the generic's package is loaded. (To avoid taking a
#' dependency on vctrs for this one function, please feel free to copy
#' and paste the function source into your own package.)
#'
#' For R 3.5.0 and later, `s3_register()` is also useful when demonstrating
#' class creation in a vignette, since method lookup no longer always involves
#' the lexical scope. For R 3.6.0 and later, you can achieve a similar effect
#' by using "delayed method registration", i.e. placing the following in your
#' `NAMESPACE` file:
#'
#' ```
#' if (getRversion() >= "3.6.0") {
#'   S3method(package::generic, class)
#' }
#' ```
#'
#' @param generic Name of the generic in the form `pkg::generic`.
#' @param class Name of the class
#' @param method Optionally, the implementation of the method. By default,
#'   this will be found by looking for a function called `generic.class`
#'   in the package environment.
#'
#'   Note that providing `method` can be dangerous if you use
#'   devtools. When the namespace of the method is reloaded by
#'   `devtools::load_all()`, the function will keep inheriting from
#'   the old namespace. This might cause crashes because of dangling
#'   `.Call()` pointers.
#' @export
#' @examples
#' # A typical use case is to dynamically register tibble/pillar methods
#' # for your class. That way you avoid creating a hard dependency on packages
#' # that are not essential, while still providing finer control over
#' # printing when they are used.
#'
#' .onLoad <- function(...) {
#'   .s3Register("pillar::pillar_shaft", "vctrs_vctr")
#'   .s3Register("tibble::type_sum", "vctrs_vctr")
#' }
#' @return nothing; called for side effects
#' @keywords internal
.s3register <- function(generic, class, method = NULL) {
  stopifnot(is.character(generic), length(generic) == 1)
  stopifnot(is.character(class), length(class) == 1)

  pieces <- strsplit(generic, "::")[[1]]
  stopifnot(length(pieces) == 2)
  package <- pieces[[1]]
  generic <- pieces[[2]]

  caller <- parent.frame()

  get_method_env <- function() {
    top <- topenv(caller)
    if (isNamespace(top)) {
      asNamespace(environmentName(top))
    } else {
      caller
    }
  }
  get_method <- function(method, env) {
    if (is.null(method)) {
      get(paste0(generic, ".", class), envir = get_method_env())
    } else {
      method
    }
  }

  method_fn <- get_method(method)
  stopifnot(is.function(method_fn))

  # Always register hook in case package is later unloaded & reloaded
  setHook(
    packageEvent(package, "onLoad"),
    function(...) {
      ns <- asNamespace(package)

      # Refresh the method, it might have been updated by `devtools::load_all()`
      method_fn <- get_method(method)

      registerS3method(generic, class, method_fn, envir = ns)
    }
  )

  # Avoid registration failures during loading (pkgload or regular)
  if (!isNamespaceLoaded(package)) {
    return(invisible())
  }

  envir <- asNamespace(package)

  # Only register if generic can be accessed
  if (exists(generic, envir)) {
    registerS3method(generic, class, method_fn, envir = envir)
  }

  invisible()
}

rxTest({
  # .onLoad() registers the S3 methods for suggested packages without
  # loading them.  .s3register() registers immediately when the other
  # namespace is already loaded and otherwise installs an onLoad hook, so
  # check both halves: the method is reachable, and it landed in the other
  # package's S3 methods table once that package is loaded.

  .rxS3Registered <- function(pkg, generic, class) {
    if (!isNamespaceLoaded(pkg)) return(NA)
    # Resolve the generic the way .s3register() does.  Prefer the exported
    # object so an unrelated same-named generic cannot be inherited from
    # base/stats and passed on; fall back to what the namespace sees, since
    # `units` has no `units<-` of its own and that call targets base's.
    .gen <- try(getExportedValue(pkg, generic), silent = TRUE)
    if (inherits(.gen, "try-error")) {
      .gen <- try(get(generic, envir = asNamespace(pkg)), silent = TRUE)
    }
    if (inherits(.gen, "try-error")) return(NA)
    # registerS3method() files a method under the environment that DEFINES the
    # generic, which is not always `pkg`: base for a primitive, and base for a
    # generic a package re-exports (`units` re-exports base's `units<-`)
    .defenv <- if (typeof(.gen) == "closure") environment(.gen) else asNamespace("base")
    .tbl <- try(get(".__S3MethodsTable__.", envir = .defenv, inherits = FALSE),
                silent = TRUE)
    if (inherits(.tbl, "try-error")) return(NA)
    exists(paste0(generic, ".", class), envir = .tbl, inherits = FALSE)
  }

  # Read the cases out of .onLoad() itself rather than repeating them, so
  # adding a registration cannot leave this file behind.  Match over the whole
  # body at once: a wrapped or doubled-up line would defeat a per-line regex.
  .rxS3Src <- paste(deparse(body(rxode2:::.onLoad), width.cutoff = 500L), # nolint
                    collapse = "\n")
  .rxS3Found <- regmatches(
    .rxS3Src,
    gregexec('\\.s3register\\("([^":]+)::([^"]+)", *"([^"]+)"\\)', .rxS3Src))[[1]]
  .rxS3Cases <- if (length(.rxS3Found) == 0L) list() else {
    apply(.rxS3Found, 2L,
          function(.m) list(pkg = .m[2], generic = .m[3], class = .m[4]))
  }

  test_that("every .onLoad() registration is covered below", {
    # a call the regex cannot read would otherwise be skipped silently
    .nCalls <- gregexpr(".s3register(", .rxS3Src, fixed = TRUE)[[1]]
    .nCalls <- if (identical(as.integer(.nCalls), -1L)) 0L else length(.nCalls)
    expect_equal(length(.rxS3Cases), .nCalls)
    expect_true(length(.rxS3Cases) > 20L)
    expect_false(any(vapply(.rxS3Cases,
                            function(.c) anyNA(unlist(.c)), logical(1))))
  })

  # `units<-.rxEvid` is the one case NAMESPACE also registers statically, so
  # for it this only shows dispatch will work, not that the hook fired.
  for (.case in .rxS3Cases) {
    test_that(sprintf("%s.%s is registered for %s",
                      .case$generic, .case$class, .case$pkg), {
      skip_if_not_installed(.case$pkg)
      # loading the namespace now must fire the deferred registration hook
      loadNamespace(.case$pkg)
      expect_true(isTRUE(.rxS3Registered(.case$pkg, .case$generic, .case$class)))
    })
  }

  test_that("suggested packages dispatch on rxode2 classes", {
    skip_if_not_installed("tibble")
    skip_if_not_installed("dplyr")
    .ev <- et(amt = 10) |> et(0:5)
    expect_s3_class(tibble::as_tibble(.ev), "tbl_df")
    # mutate.rxEt() dispatches and returns a tibble, not an rxEt
    .mut <- dplyr::mutate(.ev, .zz = 1)
    expect_true(".zz" %in% names(.mut))
    expect_equal(nrow(.mut), nrow(as.data.frame(.ev)))
  })
})

#' A list and description of rxode2 supported reserved keywords
#'
#' @format A data frame with 3 columns and 31 rows
#' \describe{
#' \item{Reserved Name}{Reserved Keyword Name}
#' \item{Meaning}{Reserved Keyword Meaning}
#' \item{Alias}{Keyword Alias}
#' }
"rxReservedKeywords"

#' @export
rxUdfUi.mix <- function(fun) {
  eval(fun)
}

#' Specify a mixture model of variables
#'
#'
#' @param ... Arguments to the mixture model.
#'
#' The first call to the mixture model function takes an odd number of
#' arguments (at least 3).
#'
#' For example the model we could have:
#'
#' cl = mix(cl1, p1, cl2, p2, cl3)
#'
#' Here there is a mixture of three clearance variables, `cl1`, `cl2`,
#' and `cl3`, at a probability of `p1`, `p2`, and the last one is
#' assumed to be `1 - p1 - p2`.
#'
#' For simulations this is selected randomly.  For estimations this is
#' selected by the data for each individual.
#'
#' After the first call when the number of populations has been
#' established, you can also call the mixture model with the number of
#' populations, for example:
#'
#' v = mix(v1, v2, v3)
#'
#' The ui function will translate this to the following model:
#'
#' v = mix(v1, p1, v2, p2, v3)
#'
#' This is because the first call to `mix()` sets the probabilities.
#' In rxode2/nlmixr2 these probabilities should be conserved between
#' the models.  These probabilities also have to be defined in the ini
#' block directly.
#'
#' @return The mixture model replacement for the underlying rxode2 model.
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' # This is an example of a mixture model
#' # Where there are 2 different clearance populations
#'
#' one.cmt <- function() {
#'   ini({
#'     tka <- 0.45 # Log Ka
#'     tcl1 <- log(c(0, 2.7, 100)) # Log Cl
#'     tcl2 <- log(c(0, 0.1, 120)) # Log Cl
#'     tv <- 3.45; label("log V")
#'     p1 <- 0.3
#'     eta.ka ~ 0.6
#'     eta.cl ~ 0.3
#'     eta.v ~ 0.1
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     # This is the example mixture model
#'     cl <- mix(exp(tcl1 + eta.cl), p1, exp(tcl2 + eta.cl))
#'     v <- exp(tv + eta.v)
#'     me <- mixest # This is the assigned mixture estimate
#'     mn <- mixnum # This is the number of mixture estimate in the model
#'     # This is the uniform mixture estimate used in simualtion to
#'     # determine the population
#'     mu <- mixunif
#'     linCmt() ~ add(add.sd)
#'   })
#' }
#'
#' \donttest{
#'
#' s <- rxSolve(one.cmt, et(amt=320, ii=12, addl=2, cmt=1) |>
#'                       et(seq(0, 72)) |>
#'                       et(id=1:20))
#'
#' plot(s, ipredSim)
#'
#' }
mix <- function(...) {
  .call <- match.call(expand.dots = TRUE)
  .args <- lapply(seq_along(.call)[-1L],
                  function(i) { .call[[i]] })
  .df <- rxUdfUiIniDf()
  if (is.null(.df)) {
    stop("mix() in a ui model must have an ini block",
         call.= FALSE)
  }
  if (all(is.na(.df$neta1))) {
    stop("mix() in a ui model must have an ini block with eta variables",
         call.= FALSE)
  }
  .ret <- list(replace=.call)
  if (is.null(rxUdfUiMv())) {
    if (identical(rxUdfUiNum(), 1L)) {
      if (length(.args) < 3) {
        stop("mix() requires at least three arguments")
      }
      .udfUiEnv$probs <- NULL
      .probs <- vapply(seq_along(.args),
                       function(i) {
                         if (i %% 2 == 0) {
                           as.character(.args[[i]])
                         } else {
                           ""
                         }
                       }, character(1))
      .udfUiEnv$probs <- .probs[nzchar(.probs)]
      .mp <- vapply(.udfUiEnv$probs,
                    function(p) {
                      p %in% .df$name
                    }, logical(1), USE.NAMES = TRUE)
      .w <- which(!.mp)
      if (length(.w) >= 1) {
        stop("the probabilities in a mixture must be in the model block, these variables were not: '",
             paste(names(.mp)[.w], collapse="', '"), "'")
      }
      .mp <- sum(vapply(.udfUiEnv$probs,
                        function(p) {
                          .w <- which(.df$name == p)
                          .df$est[.w]
                        }, numeric(1), USE.NAMES = FALSE))
      if (.mp >= 1 || .mp <= 0)  {
        stop("the probabilities in a mixture must sum to a number between 0 and 1, they sum to: ",
             .mp)
      }
    }
  }
  if (length(.args) == length(.udfUiEnv$probs)*2 + 1L) {
    .udfUiEnv$np <- 1L
    .matchProbs <- all(vapply(seq_along(.args),
                              function(i) {
                                if (i %% 2 == 0) {
                                  .ret <- identical(.udfUiEnv$probs[.udfUiEnv$np], as.character(.args[[i]]))
                                  .udfUiEnv$np <- .udfUiEnv$np + 1L
                                  .ret
                                } else {
                                  TRUE
                                }
                              }, logical(1)))
    if (!.matchProbs) {
      stop("the probabilities in a mixture must match throughout the problem",
           call.= FALSE)
    }
  }
  if (length(.args) == length(.udfUiEnv$probs) + 1L) {
    .udfUiEnv$na <- 1L
    .udfUiEnv$np <- 1L
    .ret <- list(replace=str2lang(paste0("mix(",
              paste(vapply(seq_len(length(.args)+ length(.udfUiEnv$probs)),
                           function(i) {
                             if ((i %% 2) == 0) {
                               .ret <- as.character(.udfUiEnv$probs[.udfUiEnv$np])
                               .udfUiEnv$np <- .udfUiEnv$np + 1L
                             } else {
                               .ret <- deparse1(.args[[.udfUiEnv$na]])
                               .udfUiEnv$na <- .udfUiEnv$na + 1L
                             }
                             .ret
                           }, character(1), USE.NAMES = FALSE),
                    collapse=", "),
              ")")))
  }
  .ret
}

#' Calculate the lambdas and coefficients of the two compartment model
#'
#' @param k10 elimination rate
#' @param k12 rate from central to peripheral compartment
#' @param k21 rate from peripheral to central compartment
#' @return List with `L` vector and matrices `C1` and `C2`
#' @export
#' @keywords internal
#' @author Matthew L. Fidler based on `wnl` package/paper, implemented
#'   in C/C++
#' @examples
#' .solComp2(k10=0.1, k12=3, k21=1)
.solComp2 <- function(k10, k12, k21) {
  checkmate::assertNumeric(k10, lower=0, len=1, any.missing=FALSE)
  checkmate::assertNumeric(k12, lower=0, len=1, any.missing=FALSE)
  checkmate::assertNumeric(k21, lower=0, len=1, any.missing=FALSE)
  .ret <- .Call(`_rxode2_solComp2`, k10, k12, k21)
  if (is.null(.ret)) {
    stop("roots must be distinct real values", call.=FALSE)
  }
  .ret
}
#' Calculate the lambdas and coefficients of the three compartment model
#'
#' @inheritParams .solComp2
#' @param k13 rate from central to peripheral compartment #2
#' @param k31 rate from peripheral compartment #2 to central
#' @return List with `L` vector and matrices `C1`, `C2` and `C3`
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
#' @examples
#' .solComp3(k10=0.1, k12=3, k21=1, k13=2, k31=0.5)
.solComp3 <- function(k10, k12, k21, k13, k31) {
  checkmate::assertNumeric(k10, lower=0, len=1, any.missing=FALSE)
  checkmate::assertNumeric(k12, lower=0, len=1, any.missing=FALSE)
  checkmate::assertNumeric(k21, lower=0, len=1, any.missing=FALSE)
  checkmate::assertNumeric(k13, lower=0, len=1, any.missing=FALSE)
  checkmate::assertNumeric(k31, lower=0, len=1, any.missing=FALSE)
  .ret <- .Call(`_rxode2_solComp3`, k10, k12, k21, k13, k31)
  if (is.null(.ret)) {
    stop("roots must be distinct real values", call.=FALSE)
  }
  .ret
}

#' This function asserts the requested rename makes sense
#'
#' It returns the new expression, old expression, new variable and old
#' variable
#'
#' @param line quoted call information line
#'
#' @param vars Variables contained within the mdel
#'
#' @return list(new, old, newChar, oldChar)
#'
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.assertRenameErrorModelLine <- function(line, vars) {
  .var.name <- NULL
  if (is.name(line[[2]])) {
    .var.name <- as.character(line[[2]])
  } else {
    stop("to rename a variable you need to use 'newName=oldName' syntax",
         call.=FALSE)
  }
  if (!identical(line[[1]], quote(`<-`))) {
    stop("to rename a variable you need to use '", .var.name, "=oldName' syntax",
         call.=FALSE)
  }
  if (!is.name(line[[3]])) {
    stop("to rename a variable you need to use '", .var.name, "=oldName' syntax, where oldName is a variable",
         call.=FALSE)
  } else {
    .var.name2 <- as.character(line[[3]])
  }
  if (.var.name %in% vars) {
    stop("the new variable '", .var.name,
         "' is already present in the model; cannot replace '",
         .var.name2, "' with '",
         .var.name, "'",
         call.=FALSE)
  }
  if (!(.var.name2 %in% vars)) {
    stop("the old variable '", .var.name2, "' is not present in the model and cannot be renamed to '", .var.name, "'",
         call.=FALSE)
  }
  list(line[[2]], line[[3]], .var.name, .var.name2)
}
#' Renames everything in one function
#'
#' @param item language item to process
#' @param lst list of renaming from .assertRenameErrorModelLine, ie list(new, old, newChar, oldChar)
#' @param isLhs is the expression the left handed side of the equation
#' @return expression renamed
#' @noRd
#' @author Matthew L. Fidler
.rxRenameRecursiveAll <- function(item, lst, isLhs=FALSE) {
  if (is.atomic(item)) {
    return(item)
  }
  if (is.name(item)) {
    .env <- new.env(parent=emptyenv())
    .env$new <- NULL
    lapply(seq_along(lst), function(i) {
      if (!is.null(.env$new)) return(NULL)
      .curLst <- lst[[i]]
      .old <- .curLst[[2]]
      if (identical(item, .old)) {
        .env$new <- .curLst[[1]]
      }
      return(NULL)
    })
    if (!is.null(.env$new)) {
      return(.env$new)
    }
    return(item)
  } else if (is.call(item)) {
    if (isLhs && identical(item[[1]], quote(`/`))) {
      # handle d/dt() differently so that d doesn't get renamed
      .num <- item[[2]]
      .denom <- item[[3]]
      if (is.call(.num)) .num <- as.call(lapply(.num, .rxRenameRecursiveAll, lst=lst, isLhs=TRUE))
      if (is.call(.denom)) .denom <- as.call(lapply(.denom, .rxRenameRecursiveAll, lst=lst, isLhs=TRUE))
      return(as.call(c(list(item[[1]]), .num, .denom)))
    } else if (isLhs && length(item) == 2L &&
                 is.numeric(item[[2]])) {
      .env <- new.env(parent=emptyenv())
      .env$new <- NULL
      lapply(seq_along(lst),
             function(i) {
               if (!is.null(.env$new)) return(NULL)
               .curLst <- lst[[i]]
               .old <- .curLst[[2]]
               if (identical(item[[1]], .old)) {
                 .env$new <- .curLst[[1]]
               }
               return(NULL)
             })
      if (!is.null(.env$new)) {
        # handle x(0) = items
        return(as.call(c(.env$new, lapply(item[-1], .rxRenameRecursiveAll, lst=lst, isLhs=isLhs))))
      }
    }
    if (identical(item[[1]], quote(`=`)) ||
          identical(item[[1]], quote(`<-`)) ||
          identical(item[[1]], quote(`~`))) {
      .elhs <- lapply(item[c(-1, -3)], .rxRenameRecursiveAll, lst=lst, isLhs=TRUE)
      .erhs <- lapply(item[c(-1, -2)], .rxRenameRecursiveAll, lst=lst, isLhs=FALSE)
      return(as.call(c(item[[1]], .elhs, .erhs)))
    } else {
      return(as.call(c(list(item[[1]]), lapply(item[-1], .rxRenameRecursiveAll, lst=lst, isLhs=isLhs))))
    }
  } else {
    stop("unknown expression", call.=FALSE)
  }
}
#' Rename all items in matrix dimnames
#'
#' @param mat matrix
#' @param lst list for renaming
#' @return renamed matrix
#' @noRd
#' @author Matthew L. Fidler
.rxRenameAllMat <- function(mat, lst) {
  .d <- dimnames(mat)[[1]]
  .d <- vapply(seq_along(.d), function(i) {
    .env <- new.env(parent=emptyenv())
    .env$new <- NULL
    .cur <- .d[i]
    lapply(seq_along(lst),
           function(j) {
             if (!is.null(.env$new)) return(NULL)
             .curLst <- lst[[j]]
             .old <- .curLst[[4]]
             if (.cur == .old) {
               .env$new <- .curLst[[3]]
             }
           })
    if (!is.null(.env$new)) return(.env$new)
    return(.cur)
  }, character(1), USE.NAMES=FALSE)
  dimnames(mat) <- list(.d, .d)
  mat
}

#' Rename all the items in the initialization data frame and model
#'
#' @param rxui the ui to process
#' @param lst the list of old and new expressions (like above)
#' @return Called for side effects
#' @noRd
#' @author Matthew L. Fidler
.rxRenameAll <- function(rxui, lst) {
  rxui <- rxUiDecompress(rxui)
  .iniDf <- rxui$iniDf
  .iniDf$name <- vapply(seq_along(.iniDf$name),
                        function(i) {
                          .env <- new.env(parent=emptyenv())
                          .env$new <- NULL
                          .cur <- .iniDf$name[i]
                          lapply(seq_along(lst),
                                 function(j) {
                                   if (!is.null(.env$new)) return(NULL)
                                   .curLst <- lst[[j]]
                                   .old <- .curLst[[4]]
                                   if (.cur == .old) {
                                     .env$new <- .curLst[[3]]
                                   }
                                 })
                          if (!is.null(.env$new)) return(.env$new)
                          return(.cur)
                        }, character(1), USE.NAMES=FALSE)
  rxui$iniDf <- .iniDf
  if (exists("sigma", rxui)) {
    assign("sigma", .rxRenameAllMat(get("sigma", envir=rxui), lst), envir=rxui)
  }
  if (exists("thetaMat", rxui)) {
    assign("thetaMat", .rxRenameAllMat(get("thetaMat", envir=rxui), lst), envir=rxui)
  }
  if (exists("meta", rxui)) {
    .meta <- get("meta", rxui)
    if (exists("sigma", .meta)) {
      assign("sigma", .rxRenameAllMat(get("sigma", envir=.meta), lst), envir=.meta)
    }
    if (exists("thetaMat", .meta)) {
      assign("thetaMat", .rxRenameAllMat(get("thetaMat", envir=.meta), lst), envir=.meta)
    }
  }
  rxui$lstExpr <- lapply(seq_along(rxui$lstExpr),
                         function(i) {
                           .rxRenameRecursiveAll(rxui$lstExpr[[i]], lst=lst)
                         })
}

#' Rename items inside of a `rxode2` ui model
#'
#' `rxRename()` changes the names of individual variables, lhs, and ode states using
#' `new_name = old_name` syntax
#'
#' @param .data rxode2 ui function, named data to be consistent with `dplyr::rename()`
#'
#' @param ... rename items
#'
#' @param envir Environment for evaluation
#'
#' @return New model with items renamed
#'
#' @author Matthew L. Fidler
#'
#' @details
#'
#' This is similar to `dplyr`'s `rename()` function.  When `dplyr` is
#' loaded, the `s3` methods work for the ui objects.
#'
#' Note that the `.rxRename()` is the internal function that is called
#' when renaming and is likely not what you need to call unless you
#' are writing your own extension of the function
#'
#' @export
#'
#' @examples
#'
#' ocmt <- function() {
#'   ini({
#'     tka <- exp(0.45) # Ka
#'     tcl <- exp(1) # Cl
#'     ## This works with interactive models
#'     ## You may also label the preceding line with label("label text")
#'     tv <- exp(3.45) # log V
#'     ## the label("Label name") works with all models
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- tka
#'     cl <- tcl
#'     v <- tv
#'     d/dt(depot) = -ka * depot
#'     d/dt(center) = ka * depot - cl / v * center
#'     cp = center / v
#'     cp ~ add(add.sd)
#'   })
#' }
#'
#' ocmt |> rxRename(cpParent=cp)
#'
rxRename <- function(.data, ..., envir=parent.frame()) {
  UseMethod("rxRename")
}

#' @rdname rxRename
#' @export
.rxRename <- function(.data, ..., envir=parent.frame()) {
  .inCompress <- FALSE
  if (inherits(.data, "rxUi") &&
        (is.list(.data) || inherits(.data, "raw"))) {
    .inCompress <- TRUE
  }
  rxui <- assertRxUi(.data)
  if (is.list(rxui) || inherits(rxui, "raw")) {
    rxui <- rxUiDecompress(rxui)
  }
  .vars <- unique(c(rxui$mv0$state, rxui$mv0$params, rxui$mv0$lhs, rxui$predDf$var, rxui$predDf$cond, rxui$iniDf$name))
  .modelLines <- .quoteCallInfoLines(match.call(expand.dots = TRUE)[-(1:2)], envir=envir)
  .lst <- lapply(seq_along(.modelLines), function(i) {
    .assertRenameErrorModelLine(.modelLines[[i]], .vars)
  })
  rxui <- .copyUi(rxui) # copy ui so effects do not affect original
  .rxRenameAll(rxui, .lst)
  .ret <- rxui$fun()
  if (inherits(.data, "rxUi")) {
    .ret <- .newModelAdjust(.ret, rxui, rename=TRUE)
    if (.inCompress) {
      .ret <- rxUiCompress(.ret)
    }
    .cls <- setdiff(class(.data), class(.ret))
    if (length(.cls) > 0) {
      class(.ret) <- c(.cls, class(.ret))
    }
  }
  .ret
}
#' @rdname rxRename
rename.rxUi <- function(.data, ...) {
  .lst <- as.list(match.call()[-1])
  .lst$.data <- .data
  do.call(.rxRename, c(.lst, list(envir=parent.frame(2))))
}
#' @rdname rxRename
rename.function <- function(.data, ...) {
  .lst <- as.list(match.call()[-1])
  .lst$.data <- .data
  do.call(.rxRename, c(.lst, list(envir=parent.frame(2))))
}
#' @export
#' @rdname rxRename
rxRename.rxUi <- function(.data, ...) {
  .lst <- as.list(match.call()[-1])
  .lst$.data <- .data
  do.call(.rxRename, c(.lst, list(envir=parent.frame(2))))
}
#' @export
#' @rdname rxRename
rxRename.function <- function(.data, ...) {
  .lst <- as.list(match.call()[-1])
  .lst$.data <- .data
  do.call(.rxRename, c(.lst, list(envir=parent.frame(2))))
}
#' @export
#' @rdname rxRename
rxRename.default <- function(.data, ...) {
  .lst <- as.list(match.call()[-1])
  .lst$.data <- .data
  do.call(.rxRename, c(.lst, list(envir=parent.frame(2))))
}

.pkgModelCurrent <- TRUE

.setPkgModels <- function(value) { ## For testing
  assignInMyNamespace(".pkgModelCurrent", value)
}

.norm2 <- function(obj) {
  if (inherits(obj, "rxode2")) {
    if (exists(".linCmtM", obj)) {
      return(get(".linCmtM", obj))
    }
  }
  setNames(rxode2::rxModelVars(obj)$model["normModel"], NULL)
}

.isWritable <- function(path) {
  if (checkmate::checkString(path)) {
    if (file.exists(path)) {
      return(file.access(path, mode = 2)[[1]] == 0)
    }
  }
  FALSE
}

.rxPkgInst <- function(obj) {
  .wd <- getwd()
  if (regexpr(obj$package, .wd) != -1) {
    .inst <- gsub(paste0("(", obj$package, ").*"), "\\1", .wd)
  } else {
    .inst <- system.file(package = obj$package)
  }
  if (.isWritable(.inst)) {
    if (regexpr("inst$", .inst) != -1) {
      return(.inst)
    }
    .inst2 <- file.path(.inst, "inst")
    if (file.exists(.inst2)) {
      return(.inst2)
    }
    .html <- file.path(.inst, "html")
    if (file.exists(.html)) {
      return(.inst)
    }
    return(.inst2)
  } else {
    .inst <- "~/.rxCache/"
    if (.isWritable(.inst)) {
      return(.inst)
    }
    return(rxTempDir())
  }
}

.rxPkgDir <- function(obj) {
  return(file.path(.rxPkgInst(obj), "rx"))
}

.rxPkgDll <- function(obj) {
  obj$mdir <- .rxPkgDir(obj)
  .pkgInfo <- getLoadedDLLs()[[obj$package]]
  if (!all(is.null(.pkgInfo))) {
    if (obj$isValid()) {
      .tmp <- .pkgInfo
      class(.tmp) <- "list"
      return(.tmp$path)
    } else {
      return(file.path(obj$mdir, basename(obj$rxDll$dll)))
    }
  } else {
    return(file.path(obj$mdir, basename(obj$rxDll$dll)))
  }
}

.rxNewMvStr <- function(obj) {
  gsub("[.].*", paste0("_new_", .Platform$r_arch, "_model_vars"), basename(obj$rxDll$dll))
}

.rxPkgLoaded <- function(pkg) {
  .si <- sessionInfo()
  return(length(intersect(pkg, c(
    names(.si$otherPkgs) ## ,names(.si$loadedOnly)
  ))) != 0)
}

.rxUseI <- new.env(parent = emptyenv())
# 1
.rxUseI$i <- 1L
.rxUseCdir <- ""

#' Use model object in your package
#' @param obj model to save.
#' @param internal If this is run internally.  By default this is FALSE
#' @inheritParams usethis::use_data
#' @return Nothing; This is used for its side effects and shouldn't be called by a user
#' @export
rxUse <- function(obj, overwrite = TRUE, compress = "bzip2",
                  internal = FALSE) {
  rxReq("usethis")
  rxReq("devtools")
  internal <- internal
  if (missing(obj)) {
    .env <- new.env()
    assign("internal", internal, .env)
    assign("overwrite", overwrite, .env)
    assign("compress", compress, .env)
    lapply(list.files(devtools::package_file("inst/rx"), full.names = TRUE),
           function(f) {
             unlink(f, force = TRUE, recursive = TRUE)
           })
    .models <- NULL
    for (.f in list.files(
      path = devtools::package_file("data"),
      pattern = "\\.rda$", full.names = TRUE
    )) {
      load(.f, envir = .env)
      .f2 <- basename(.f)
      .f2 <- substr(.f2, 0, nchar(.f2) - 4)
      if (is(.env[[.f2]], "rxode2")) {
        .env[[.f2]]$package <- NULL
        .minfo(sprintf("recompile '%s'", .f2))
        .models <- c(.models, .f2)
        eval(parse(text = sprintf("rxUse(%s, internal=internal, overwrite=overwrite, compress=compress)", .f2)),
          envir = .env
        )
        .docFile <- file.path(devtools::package_file("R"), paste0(.f2, "-doc.R"))
        if (!file.exists(.docFile)) {
          (sprintf("creating documentation '%s'", .docFile))
          sink(.docFile)
          .tmp <- .env[[.f2]]
          .mv <- rxModelVars(.tmp)
          cat(sprintf("#' %s rxode2 model\n", .f2))
          cat("#'\n")
          cat(sprintf(
            "#' @format An \\emph{rxode2} model with %s parameters, %s ODE states, and %s calc vars.\n",
            length(.tmp$params), length(.tmp$state) + .mv$extraCmt, length(.tmp$lhs)
          ))
          cat("#'\n")
          cat(sprintf("#'\\emph{Parameters (%s$params)}\n", .f2))
          cat("#'\n")
          cat("#' \\describe{\n")
          .def <- rxInits(.tmp)
          .defs <- paste0(" (default=", .def, ")")
          .defs[is.na(.def)] <- ""
          cat(paste(paste0("#'   \\item{", .tmp$params, "}{", .defs, "}\n"), collapse = ""))
          cat("#'}\n")
          .state <- .tmp$state
          ##
          if (.mv$extraCmt == 2) {
            .state <- c(.state, "depot", "central")
          } else if (.mv$extraCmt == 1) {
            .state <- c(.state, "central")
          }
          if (length(.state) > 0) {
            cat("#'\n")
            cat(sprintf("#' \\emph{State %s$state}\n", .f2))
            cat("#'\n")
            cat("#' \\describe{\n")
            cat(paste(paste0("#'   \\item{", .state, "}{ (=", seq_along(.state), ")}\n"), collapse = ""))
            cat("#' }\n")
          }
          .lhs <- .tmp$lhs
          if (length(.lhs) > 0) {
            cat("#'\n")
            cat(sprintf("#' \\emph{Calculated Variables %s$lhs}\n", .f2))
            cat("#'\n")
            cat("#' \\describe{\n")
            cat(paste(paste0("#'   \\item{", .lhs, "}{}\n"), collapse = ""))
            cat("#' }\n")
          }
          cat("#'\n")
          cat("#' \\emph{Model Code}\n") # sprintf(,.f2)
          cat("#'\n")
          .code <- deparse(body(eval(parse(text = paste("function(){", .norm2(.tmp), "}")))))
          .code[1] <- "rxode2({"
          .code[length(.code)] <- "})"
          cat(paste(paste0("#' ", .code, "\n"), collapse = ""))
          cat("#'\n")
          cat(paste(paste0("#' @seealso \\code{\\link[rxode2]{eventTable}}, \\code{\\link[rxode2]{et}}, \\code{\\link[rxode2]{rxSolve}}, \\code{\\link[rxode2]{rxode2}}\n")))
          cat("#' \n")
          cat("#' @examples\n")
          cat("#' ## Showing the model code\n")
          cat(sprintf("#' summary(%s)\n", .f2))
          cat("#'\n")
          cat(sprintf('"%s"\n', .f2))
          sink()
        }
      }
    }
    if (!dir.exists(devtools::package_file("src"))) {
      dir.create(devtools::package_file("src"), recursive = TRUE, showWarnings = FALSE)
    }
    .pkg <- basename(usethis::proj_get())
    .rx <- loadNamespace("rxode2")
    lapply(
      list.files(.rxUseCdir, pattern = "[.]c", full.names = TRUE),
      function(x) {
        .minfo(sprintf("copy '%s'", basename(x)))
        .env <- .rx$.rxUseI
        .rxUseI <- .env$i
        .f0 <- gsub(
          "^#define (.*) _rx(.*)$",
          paste0("#define \\1 _rxp", .rxUseI, "\\2"), readLines(x)
        )
        assign("i", .rxUseI + 1, envir = .env)
        .f0 <- c("#include <rxode2.h>\n#include <rxode2_model_shared.h>", .f0)
        .w <- which(.f0 == "#include \"extraC.h\"")
        if (length(.w) > 0) .f0 <- .f0[-.w[1]]
        writeLines(text = .f0, con = file.path(devtools::package_file("src"), basename(x)))
      }
    )
    .inits <- paste0("R_init0_", .pkg, "_", .models)
    .tmp <- paste0("{\"", .pkg, "_", .models, "_model_vars\", (DL_FUNC) &", .pkg, "_", .models, "_model_vars, 0},\\")
    .tmp[length(.tmp)] <- substr(.tmp[length(.tmp)], 0, nchar(.tmp[length(.tmp)]) - 1)
    .extraC <- c(
      "#define compiledModelCall \\",
      .tmp,
      paste0("SEXP ", .pkg, "_", .models, "_model_vars();"),
      paste0("void ", .inits, "();"),
      paste0("void R_init0_", .pkg, "_rxode2_models(){"),
      paste0("  ", .inits, "();"),
      "}"
    )
    sink(file.path(devtools::package_file("src"), paste0(.pkg, "_compiled.h")))
    if (.pkg == "rxode2") {
      cat("#include <R.h>\n#include <Rinternals.h>\n#include <stdlib.h> // for NULL\n#include <R_ext/Rdynload.h>\n#include \"../inst/include/rxode2.h\"\n#include \"../inst/include/rxode2_model_shared.h\"\n")
    } else {
      cat("#include <R.h>\n#include <Rinternals.h>\n#include <stdlib.h> // for NULL\n#include <R_ext/Rdynload.h>\n#include <rxode2.h>\n#include <rxode2_model_shared.h>\n")
    }
    cat(paste(.extraC, collapse = "\n"))
    cat("\n")
    sink()
    .files <- list.files(devtools::package_file("src"))
    .files <- .files[regexpr("rxode2_model_shared", .files) == -1]
    if (all(regexpr(paste0("^", .pkg), .files) != -1)) {
      .minfo(sprintf("only compiled models in this package, creating '%s_init.c'", .pkg))
      sink(file.path(devtools::package_file("src"), paste0(.pkg, "_init.c")))
      cat("#include <R.h>\n#include <Rinternals.h>\n#include <stdlib.h> // for NULL\n#include <R_ext/Rdynload.h>\n")
      cat("#include <rxode2.h>\n")
      cat("#include <rxode2_model_shared.h>\n")
      cat(paste0('#include "', .pkg, '_compiled.h"\n'))
      cat(sprintf("void R_init_%s(DllInfo *info){\n", .pkg))
      cat(sprintf("  R_init0_%s_rxode2_models();\n", .pkg))
      cat("  static const R_CallMethodDef callMethods[]  = {\n  compiledModelCall\n  {NULL, NULL, 0}\n  };\n")
      cat("  R_registerRoutines(info, NULL, callMethods, NULL, NULL);\n")
      cat("  R_useDynamicSymbols(info,FALSE);\n")
      cat("}\n")
      cat(paste(paste0("void R_unload_", .pkg, "_", .models, "(DllInfo *info);\n"), collapse = ""))
      cat(sprintf("void R_unload_%s(DllInfo *info){\n", .pkg))
      cat(paste(paste0("  R_unload_", .pkg, "_", .models, "(info);\n"), collapse = ""))
      cat("}\n")
      sink()
    }
    if (!file.exists(devtools::package_file("R/rxUpdated.R")) && .pkg != "rxode2") {
      sink(devtools::package_file("R/rxUpdated.R"))
      cat(".rxUpdated <- new.env(parent=emptyenv())\n")
      sink()
    }
    lapply(list.files(devtools::package_file("inst/rx"), full.names = TRUE),
           function(f) {
             unlink(f, force = TRUE, recursive = TRUE)
           })
    if (length(list.files(devtools::package_file("inst"))) == 0) {
      unlink(devtools::package_file("inst"), recursive = TRUE, force = TRUE)
    }
    return(invisible(TRUE))
  } else {
    .tempfile <- tempfile()
    .tempR <- tempfile(fileext=".R")
    .expr <- bquote({
      rxode2::.rxWithWd(.(getwd()), {
        .modName <- .(as.character(substitute(obj)))
        .pkg <- .(basename(usethis::proj_get()))
        .env <- new.env(parent = baseenv())
        assign(.modName, rxode2::rxode2(.(.norm2(obj)), package = .pkg, modName =.modName), .env)
        writeLines(dirname(rxode2::rxC(.env[[.modName]])), .(.tempfile))
        assign("internal", .(internal), .env)
        assign("overwrite", .(overwrite), .env)
        assign("compress", .(compress), .env)
        eval(parse(text = sprintf("usethis::use_data(%s, internal=internal, overwrite=overwrite, compress=compress)", .modName)), envir = .env)
      })
    })
    writeLines(paste(deparse(.expr), collapse="\n"), .tempR)
    .cmd <- file.path(R.home("bin"), "R")
    .args <- c("CMD", "BATCH", basename(.tempR))
    .rxWithWd(tempdir(), {
      .out <- sys::exec_internal(cmd = .cmd, args = .args, error = FALSE)
      message(paste(readLines(paste0(.tempR, "out")), collapse="\n"))
    })
    .stderr <- rawToChar(.out$stderr)
    if (!(all(.stderr == "") && length(.stderr) == 1)) {
      message(paste(.stderr, sep = "\n"))
    }
    if (!file.exists(.tempfile)) {
      stop("could not build model for inclusion in package",
           call.=FALSE)
    }
    assignInMyNamespace(".rxUseCdir", readLines(.tempfile))
    try(unlink(paste(.tempfile)), silent=TRUE)
  }
}

#' Creates a package from compiled rxode2 models
#'
#' @param ... Models to build a package from
#' @param package String of the package name to create
#' @param action Type of action to take after package is created
#' @param name Full name of author
#' @param license is the type of license for the package.
#' @inheritParams usethis::create_package
#' @inheritParams rxode2
#' @author Matthew Fidler
#' @return this function returns nothing and is used for its side effects
#' @export
rxPkg <- function(..., package,
                  wd = getwd(),
                  action = c("install", "build", "binary", "create"),
                  license = c("gpl3", "lgpl", "mit", "agpl3"),
                  name = "Firstname Lastname",
                  fields = list()) {
  if (missing(package)) {
    stop("'package' needs to be specified")
  }
  action <- match.arg(action)
  license <- match.arg(license)
  .owd <- getwd()
  .op <- options()
  on.exit({
    setwd(.owd)
    options(.op)
  })
  .dir <- wd
  if (!dir.exists(.dir)) {
    dir.create(.dir)
  }
  setwd(.dir)
  options(
    usethis.description = list(`Title` = "This is generated from rxode2"),
    usethis.full_name = ifelse(missing(name), getOption("usethis.full_name", "Firstname Lastname"), name)
  )
  .dir2 <- file.path(.dir, package)
  usethis::create_package(.dir2,
    fields = fields,
    rstudio = FALSE,
    roxygen = TRUE,
    check_name = TRUE,
    open = FALSE
  )
  setwd(.dir2)
  usethis::use_package("rxode2", "LinkingTo")
  usethis::use_package("rxode2", "Depends")
  if (license == "gpl3") {
    usethis::use_gpl3_license()
  } else if (license == "lgpl") {
    usethis::use_lgpl_license()
  } else if (license == "agpl3") {
    usethis::use_agpl3_license()
  } else if (license == "mit") {
    usethis::use_mit_license()
  }
  .p <- devtools::package_file("DESCRIPTION")
  writeLines(c(
    readLines(.p),
    "NeedsCompilation: yes",
    "Biarch: true"
  ), .p)
  ## Now use rxUse for each item
  .env <- new.env()
  .lst <- as.list(match.call()[-1])
  .w <- which(names(.lst) == "")
  .lst <- .lst[.w]

  for (.i in seq_along(.lst)) {
    .v <- as.character(deparse(.lst[[.i]]))
    assign(.v, eval(.lst[[.i]], envir = parent.frame(1)), .env)
    print(.env[[.v]])
    eval(parse(text = sprintf("rxUse(%s)", .v)), envir = .env)
  }

  ## Final rxUse to generate all code
  rxUse()

  .p <- file.path(devtools::package_file("R"), "rxUpdated.R")
  .f <- readLines(.p)
  .w <- which(regexpr("@useDynLib", .f) != -1)

  if (length(.w) == 0) {
    .f <- c(
      paste0("#' @useDynLib ", package, ", .registration=TRUE"),
      "#' @import rxode2",
      .f
    )
    writeLines(.f, .p)
  }
  devtools::document()
  if (!file.exists("configure.win")) {
    writeLines(c(
      "#!/bin/sh",
      "echo \"unlink('src', recursive=TRUE);rxode2::rxUse()\" > build.R",
      "${R_HOME}/bin/Rscript build.R",
      "rm build.R"
    ), "configure.win")
  }
  if (!file.exists("configure")) {
    writeLines(c(
      "#!/bin/sh",
      "echo \"unlink('src', recursive=TRUE);rxode2::rxUse()\" > build.R",
      "${R_HOME}/bin/Rscript build.R",
      "rm build.R"
    ), "configure")
    if (!file.exists("configure.ac")) {
      writeLines(
        "## dummy autoconf script",
        "configure.ac"
      )
    }
  }
  if (action == "install") {
    devtools::install()
  } else if (action == "build") {
    devtools::build()
  } else if (action == "binary") {
    devtools::build(binary = TRUE)
  }
  invisible()
}

.isIntegerish <- function(v) {
  checkmate::testIntegerish(v)
}
#' This gives the iCov index when generating keep from iCov
#'
#'
#' @param fullData Full data index information
#'
#'  This dataset has an "id" from the input event table and idx0 a
#'  0-based index of the full dataset
#'
#' @param iCovData This dataset has an "id" from the input `iCov`
#'   dataset and "idxi" which is a zero based `iCov` index
#'
#' @return An C-based index of where each item from the `iCov`
#'   data-frame to create the keep data-est as if the `iCov` came from
#'   the first dataset
#' @author Matthew L. Fidler
#' @noRd
.getIcovIdx <- function(fullData, iCovData) {
  .m <- merge(fullData, iCovData, by="id", all.x=TRUE)
  .m <- .m[order(.m$idx0),]
  .m$idxi
}

.convertExtra <- function(dat) {
  d <- as.data.frame(dat)
  .colNames0 <- colnames(d)
  .colNames <- toupper(.colNames0)
  ## Handle DATE TIME; DAT1 TIME; DAT2 TIME and DAT3 TIME

  ## Note NONMEM handles dates of the format DAY-MONTH and DAY as
  ## well for the DATE class of objects.

  ## It is too complex to handle, and not very common so it will
  ## throw an error

  .doDate <- FALSE
  .dupDate <- gettext("dates can only be specified by one of: 'DATE', 'DAT1', 'DAT2', 'DAT3' / 'TIME'")
  .checkBad <- function(d) {
    d <- paste(d)
    if (any(unlist(lapply(strsplit(d, "[^0-9]+"), length)) != 3)) {
      stop("dates formatted as MONTH-DAY or DAY alone are not supported in this conversion",
           call. = FALSE)
    }
    d
  }
  if (any(.colNames == "DATE")) {
    ##  Month Day Year
    .datReg2 <- rex::rex(start, any_spaces, capture(numbers), non_numbers,
                         capture(numbers), non_numbers, capture(number, number), any_spaces, end)
    .datReg4 <- rex::rex(start, any_spaces, capture(numbers), non_numbers,
                         capture(numbers), non_numbers, capture(number, number, number, number), any_spaces, end)
    dt <- .checkBad(d$DATE)
    d$DATE.TIME <- as.POSIXct(NA)
    w <- which(regexpr(.datReg2, dt) != -1)
    if (length(w) > 0) {
      d$DATE.TIME[w] <- as.POSIXct(paste(gsub(.datReg2, "\\1-\\2-\\3", dt[w]), d$TIME[w]), format = "%m-%d-%y %H:%M")
    }
    w <- which(regexpr(.datReg4, dt) != -1)
    if (length(w) > 0) {
      d$DATE.TIME[w] <- as.POSIXct(paste(gsub(.datReg4, "\\1-\\2-\\3", dt[w]), d$TIME[w]), format = "%m-%d-%Y %H:%M")
    }
    d <- d[, -which(names(d) == "DATE")]
    .doDate <- TRUE
  }
  if (any(.colNames == "DAT1")) {
    if (.doDate) {
      stop(.dupDate, call. = FALSE)
    }
    ## DAT1   day month year
    .datReg2 <- rex::rex(start, any_spaces, capture(numbers), non_numbers,
                         capture(numbers), non_numbers, capture(number, number), any_spaces, end)
    .datReg4 <- rex::rex(start, any_spaces, capture(numbers), non_numbers,
                         capture(numbers), non_numbers, capture(number, number, number, number), any_spaces, end)
    dt <- .checkBad(d$DAT1)
    d$DATE.TIME <- as.POSIXct(NA)
    w <- which(regexpr(.datReg2, dt) != -1)
    if (length(w) > 0) {
      d$DATE.TIME[w] <- as.POSIXct(paste(gsub(.datReg2, "\\1-\\2-\\3", dt[w]), d$TIME[w]), format = "%d-%m-%y %H:%M")
    }
    w <- which(regexpr(.datReg4, dt) != -1)
    if (length(w) > 0) {
      d$DATE.TIME[w] <- as.POSIXct(paste(gsub(.datReg4, "\\1-\\2-\\3", dt[w]), d$TIME[w]), format = "%d-%m-%Y %H:%M")
    }
    d <- d[, -which(names(d) == "DAT1")]
    .doDate <- TRUE
  }
  if (any(.colNames == "DAT2")) {
    ## DAT2   year month day
    if (.doDate) {
      stop(.dupDate, call. = FALSE)
    }
    .datReg2 <- rex::rex(start, any_spaces, capture(number, number), non_numbers,
                         capture(numbers), non_numbers, capture(numbers), any_spaces, end)
    .datReg4 <- rex::rex(start, any_spaces, capture(number, number, number, number), non_numbers,
                         capture(numbers), non_numbers, capture(numbers), any_spaces, end)
    dt <- .checkBad(d$DAT2)
    d$DATE.TIME <- as.POSIXct(NA)
    w <- which(regexpr(.datReg2, dt) != -1)
    if (length(w) > 0) {
      d$DATE.TIME[w] <- as.POSIXct(paste(gsub(.datReg2, "\\1-\\2-\\3", dt[w]), d$TIME[w]), format = "%y-%m-%d %H:%M")
    }
    w <- which(regexpr(.datReg4, dt) != -1)
    if (length(w) > 0) {
      d$DATE.TIME[w] <- as.POSIXct(paste(gsub(.datReg4, "\\1-\\2-\\3", dt[w]), d$TIME[w]), format = "%Y-%m-%d %H:%M")
    }
    d <- d[, -which(names(d) == "DAT2")]
    .doDate <- TRUE
  }
  if (any(.colNames == "DAT3")) {
    ## DAT3   year day month
    if (.doDate) {
      stop(.dupDate, call. = FALSE)
    }
    .datReg2 <- rex::rex(start, any_spaces, capture(number, number), non_numbers,
                         capture(numbers), non_numbers, capture(numbers), any_spaces, end)
    .datReg4 <- rex::rex(start, any_spaces, capture(number, number, number, number), non_numbers,
                         capture(numbers), non_numbers, capture(numbers), any_spaces, end)
    dt <- .checkBad(d$DAT3)
    d$DATE.TIME <- as.POSIXct(NA)
    w <- which(regexpr(.datReg2, dt) != -1)
    if (length(w) > 0) {
      d$DATE.TIME[w] <- as.POSIXct(paste(gsub(.datReg2, "\\1-\\2-\\3", dt[w]), d$TIME[w]), format = "%y-%d-%m %H:%M")
    }
    w <- which(regexpr(.datReg4, dt) != -1)
    if (length(w) > 0) {
      d$DATE.TIME[w] <- as.POSIXct(paste(gsub(.datReg4, "\\1-\\2-\\3", dt[w]), d$TIME[w]), format = "%Y-%d-%m %H:%M")
    }
    d <- d[, -which(names(d) == "DAT3")]
    .doDate <- TRUE
  }
  if (.doDate) {
    if (any(is.na(d$DATE.TIME))) {
      stop("date/time format was not correctly specified", call. = FALSE)
    }
  }
  if (.doDate) {
    ## Sort by date/time (though this should have been done already...)
    if (!any(names(d) == "ID")) {
      d$ID <- 1L
    }
    if (!any(names(d) == "EVID")) {
      d$EVID <- 0L
    }
    d <- d[order(d$ID, d$DATE.TIME, -d$EVID), ]
    d$TIME <- as.vector(unlist(sapply(unique(d$ID), function(id) {
      d0 <- d[d$ID == id, ]
      as.numeric(difftime(d0$DATE.TIME,
                          d0$DATE.TIME[1],
                          units = "hours"
                          ))
    })))
    d <- d[, -which(names(d) == "DATE.TIME")]
  }
  if (is(d$TIME, "numeric") || is(d$TIME, "integer")) {
    return(d)
  }
  stop("cannot figure out numeric time", call. = FALSE)
}

.setLastIdLvl <- function(idLvl) {
  .etInfo$lastIdLvl <- idLvl
}
#' Get the last `idLvl`
#'
#' @return Last `idLvl`
#' @author Matthew L. Fidler
#' @keywords internal
#' @examples
#' .getLastIdLvl()
#' @export
.getLastIdLvl <- function() {
  .etInfo$lastIdLvl
}

#' @export
as.data.frame.rxEtTran <- function(x, row.names = NULL, optional = FALSE, ...) {
  .Call(`_rxode2_rxEtTransAsDataFrame_`, x)
}

#' Get the real compartment numbers based on a linCmt()
#'
#' @param mv  object where model variables can be extracted
#' @param cmt compartment numbers to translate; default is 1...nState
#' @return Translated (and named) compartment numbers
#' @noRd
#' @author Matthew L. Fidler
.getCmtNum <- function(mv, cmt) {
  .mv <- rxModelVars(mv)
  if (missing(cmt)) {
    cmt <- seq_along(.mv$state)
  }
  .cmt <- getCmtNum_(cmt, .mv)
  setNames(.cmt, vapply(.cmt, function(x) {
    .mv$state[x]
  }, character(1), USE.NAMES = FALSE))
}

#' @rdname etTrans
#' @export
etTrans <- function(inData, obj, addCmt = FALSE, dropUnits = FALSE, allTimeVar = FALSE,
                    keepDosingOnly = FALSE, combineDvid = NULL, keep = character(0),
                    addlKeepsCov = FALSE, addlDropSs = TRUE, ssAtDoseTime = TRUE, iCov = NULL) {
  if (is.rxEt(inData)) {
    .env <- .rxEtEnv(inData)
    .chunks <- Filter(Negate(is.null), .env$chunks)
    if (length(.chunks) > 0L) {
      .raw <- as.data.frame(data.table::rbindlist(.chunks, fill = TRUE, use.names = TRUE))
      # deSolve-style: 'value' present but no 'amt' -> drop auto-added 'evid' so C++ uses method/var
      if (!is.null(.raw[["value"]]) && is.null(.raw[["amt"]])) {
        .raw[["evid"]] <- NULL
      }
      inData <- .raw
    } else {
      inData <- as.data.frame(inData, all = TRUE)
    }
  }
  .Call(`_rxode2_etTrans`, inData, obj, addCmt, dropUnits, allTimeVar, keepDosingOnly,
        combineDvid, keep, addlKeepsCov, addlDropSs, ssAtDoseTime, iCov)
}

rxTest({
  test_that("evid=3 reset time", {

    mod1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      C2 <- centr / V2
      C3 <- peri / V3
      d / dt(depot) <- -KA * depot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- 1
      printf("%f\n", time)
    })

    et_1 <-
      et(dose = 10000, addl = 0, ii = 0) |>
      et(0:8)

    et_reset <- et(evid = 3)

    et_2 <-
      et(dose = 20000, addl = 0, ii = 0) |>
      et(0:8)

    et <- dplyr::bind_rows(et_1, et_reset, et_2)

    tmp <- etTrans(et, mod1)

    expect_true(!identical(tmp$TIME, et$time))

    tmp <- as.data.frame(tmp)

    expect_equal(tmp$TIME, et$time)

    et2 <- rbind(data.frame(id=1, et[, names(et) != "id"]),
                 data.frame(id=2, et[, names(et) != "id"]))

    tmp <- etTrans(et2, mod1)

    expect_true(!identical(tmp$TIME, et2$time))

    tmp <- as.data.frame(tmp)

    expect_equal(tmp$TIME, et2$time)

    t <- tempfile("test-evid3", fileext = ".csv")

    suppressWarnings(.rxWithSink(t, {
      cat("t\n")
      x <- rxSolve(mod1, et)
    }))

    d <- read.csv(t)
    unlink(t)

    expect_true(all(d$t < 9))
    expect_true(all(x$time < 9))
    expect_true(!all(x$C2[x$resetno == 1] == x$C2[x$resetno == 2]))
    expect_true(x$eff[1] == 1)
  })


  test_that("evid=3 reset time mixed", {
    mod1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      C2 <- linCmt()
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- 1
      printf("%f\n", time)
    })

    et_1 <-
      et(dose = 10000, addl = 0, ii = 0) |>
      et(0:8)

    et_reset <- et(evid = 3)

    et_2 <-
      et(dose = 20000, addl = 0, ii = 0) |>
      et(0:8)

    et <- dplyr::bind_rows(et_1, et_reset, et_2)

    t <- tempfile("test-evid3", fileext = ".csv")

    suppressWarnings(.rxWithSink(t, {
      cat("t\n")
      x <- rxSolve(mod1, et)
    }))

    d <- read.csv(t)
    unlink(t)

    expect_true(all(d$t < 9))
    expect_true(all(x$time < 9))
    expect_true(!all(x$C2[x$resetno == 1] == x$C2[x$resetno == 2]))
    expect_true(x$eff[1] == 1)
  })



  test_that("evid=3 reset time linCmt", {

    mod1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      C2 <- linCmt()
      printf("%f\n", time)
    })

    et1 <-
      # et(amount.units='mg', time.units='hours') |>
      et(dose = 10000, addl = 0, ii = 0) |>
      # et(amt=20000, nbr.doses=5, start.time=120, dosing.interval=24) |>
      et(0:8)

    et_reset <- et(evid = 3)

    et_2 <-
      # et(amount.units='mg', time.units='hours') |>
      et(dose = 20000, addl = 0, ii = 0) |>
      # et(amt=20000, nbr.doses=5, start.time=120, dosing.interval=24) |>
      et(0:8)

    et <- dplyr::bind_rows(et1, et_reset, et_2)

    t <- tempfile("test-evid3", fileext = ".csv")

    suppressWarnings(.rxWithSink(t, {
      cat("t\n")
      x <- rxSolve(mod1, et)
    }))

    d <- read.csv(t)
    unlink(t)

    expect_true(all(d$t < 9))
    expect_true(all(x$time < 9))
    expect_true(!all(x$C2[x$resetno == 1] == x$C2[x$resetno == 2]))
  })

  test_that("warning for unsorted data with evid=3", {
    mod1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      C2 <- linCmt()
    })

    et1 <-
      et(dose = 10000, addl = 0, ii = 0) |>
      et(0:8)

    etReset <- et(evid = 3)

    et2 <-
      et(dose = 20000, addl = 0, ii = 0) |>
      et(0:8)

    et <- dplyr::bind_rows(et1, etReset, et2)

    et$time[5] <- 9

    expect_warning(x <- rxSolve(mod1, et))
    expect_false(any(names(x) == "resetno"))
  })

  test_that("tad is non-negative after SS dosing (EVID=3+110 reset)", {
    # Regression test: SS dosing inserts EVID=3 (time reset) + EVID=110 (SS bolus).
    # handleTlastInline was called for EVID=3 which has cmt=-1, corrupting tlast
    # and producing negative tad values for observations after the SS reset.

    # ODE model
    mod_ode <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL / V2 * centr
      cp <- centr / V2
      t_ad <- tad()
    })

    # linCmt model
    mod_lin <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      cp <- linCmt()
      t_ad <- tad()
    })

    # Multi-subject event table: SS bolus (ss=1) then repeated doses.
    # Using 3 subjects with slightly different observation schedules so
    # maxShift (computed globally) is > 0, which was the trigger for the bug.
    ev <- dplyr::bind_rows(
      data.frame(id = 1, time = 0,  evid = 4, amt = 320, ii = 12, ss = 1, cmt = 1),
      data.frame(id = 1, time = 0,  evid = 0, amt = 0,   ii = 0,  ss = 0, cmt = 1, dv = NA),
      data.frame(id = 1, time = 0.5, evid = 0, amt = 0,  ii = 0,  ss = 0, cmt = 1, dv = NA),
      data.frame(id = 1, time = 1,  evid = 0, amt = 0,   ii = 0,  ss = 0, cmt = 1, dv = NA),
      data.frame(id = 1, time = 12, evid = 101, amt = 320, ii = 0, ss = 0, cmt = 1),
      data.frame(id = 1, time = 12.5, evid = 0, amt = 0, ii = 0,  ss = 0, cmt = 1, dv = NA),
      data.frame(id = 2, time = 0,  evid = 4, amt = 320, ii = 12, ss = 1, cmt = 1),
      data.frame(id = 2, time = 0.3, evid = 0, amt = 0,  ii = 0,  ss = 0, cmt = 1, dv = NA),
      data.frame(id = 2, time = 1,  evid = 0, amt = 0,   ii = 0,  ss = 0, cmt = 1, dv = NA),
      data.frame(id = 2, time = 12, evid = 101, amt = 320, ii = 0, ss = 0, cmt = 1),
      data.frame(id = 2, time = 12.8, evid = 0, amt = 0, ii = 0,  ss = 0, cmt = 1, dv = NA),
      data.frame(id = 3, time = 0,  evid = 4, amt = 320, ii = 12, ss = 1, cmt = 1),
      data.frame(id = 3, time = 0.2, evid = 0, amt = 0,  ii = 0,  ss = 0, cmt = 1, dv = NA),
      data.frame(id = 3, time = 12, evid = 101, amt = 320, ii = 0, ss = 0, cmt = 1),
      data.frame(id = 3, time = 12.4, evid = 0, amt = 0, ii = 0,  ss = 0, cmt = 1, dv = NA)
    )

    suppressWarnings({
      x_ode <- rxSolve(mod_ode, ev)
      x_lin <- rxSolve(mod_lin, ev)
    })

    obs_ode <- x_ode[!is.na(x_ode$cp), ]
    obs_lin <- x_lin[!is.na(x_lin$cp), ]

    expect_true(all(obs_ode$t_ad >= 0),
                info = paste("ODE model: negative tad values found:",
                             paste(obs_ode$t_ad[obs_ode$t_ad < 0], collapse = ", ")))
    expect_true(all(obs_lin$t_ad >= 0),
                info = paste("linCmt model: negative tad values found:",
                             paste(obs_lin$t_ad[obs_lin$t_ad < 0], collapse = ", ")))
  })
})

rxTest({
  test_that("Issue RxODE#410", {
    mod <- rxode2("cmt(EFFECT);\nd/dt(EFFECT)=0;\nd/dt(rx__sens_EFFECT_BY_ETA_1___)=0;\nrx_yj_~2;\nrx_lambda_~1;\nrx_hi_~1;\nrx_low_~0;\nrx_expr_0~exp(THETA[4]);\nrx_expr_1~ETA[1]+THETA[3];\nrx_expr_2~SEX==\"F\";\nrx_expr_3~SEX==\"M\";\nrx_expr_4~exp(rx_expr_1);\nrx_expr_5~(rx_expr_2)*THETA[2];\nrx_expr_6~(rx_expr_3)*THETA[1];\nrx_expr_7~rx_expr_5+rx_expr_6;\nrx_expr_8~exp(rx_expr_7);\nrx_expr_9~rx_expr_4-rx_expr_8;\nrx_expr_10~(-2+NTS_birth)*(rx_expr_9);\nrx_expr_11~rx_expr_10/(-2+NTS_birth+rx_expr_0);\nrx_pred_=rx_expr_11+rx_expr_8;\nrx__sens_rx_pred__BY_ETA_1___=rx_expr_4*(-2+NTS_birth)/(-2+NTS_birth+rx_expr_0);\nrx_r_=Rx_pow_di((THETA[6]+THETA[5]*(rx_expr_11+rx_expr_8)),2);\nrx__sens_rx_r__BY_ETA_1___=2*rx_expr_4*(-2+NTS_birth)*(THETA[6]+THETA[5]*(rx_expr_11+rx_expr_8))*THETA[5]/(-2+NTS_birth+rx_expr_0);sexf=(SEX==\"F\");\nsexm=(SEX==\"M\");\ncmt(effect);\n")

    est <- c("THETA[1]" = 7.3132203870903, "THETA[2]" = 7.60090245954208, "THETA[3]" = 6.90775527898214, "THETA[4]" = 0.693147180559945, "THETA[5]" = 0.2, "THETA[6]" = 1, "ETA[1]" = 0)
    
    d <-
      data.frame(
        ID = rep(1:2, each = 2),
        CMT = "EFFECT",
        DV = 1000 + rnorm(4),
        NTS_birth = rep(2:3, 2),
        SEX = rep(c("F", "M"), each = 2),
        TIME = 1
      )
    
    e <- etTrans(d, mod)
    
    expect_error(rxSolve(mod, est, d), NA)
    
    s <- rxSolve(mod, est, d, addCov = TRUE)
    expect_s3_class(s$SEX, "factor")
    expect_equal(paste(d$SEX), paste(s$SEX))
    expect_equal((d$SEX == "F") * 1.0, s$sexf)
    expect_equal((d$SEX == "M") * 1.0, s$sexm)
  })
})

rxTest({
  test_that("deSolve events", {
    ## Data frame event from deSolve vignette

    derivs <- rxode2({
      v1(0) <- 1
      v2(0) <- 2
      d / dt(v1) <- 0
      d / dt(v2) <- 0
    })

    eventdat <- data.frame(
      var = c("v1", "v2", "v2", "v1"),
      time = c(1, 1, 5, 9),
      value = c(1, 2, 3, 4),
      method = c("add", "mult", "rep", "add")
    )

    ett <- etTrans(eventdat, derivs, keepDosingOnly = TRUE)

    ett2 <- etTrans(et(eventdat), derivs, keepDosingOnly = TRUE)

    expect_equal(ett$EVID, ett2$EVID)
    expect_equal(ett$AMT, ett2$AMT)
    expect_equal(ett$TIME, ett2$TIME)

    e <- et(eventdat) |>
      et(seq(0, 10, by = 0.1))

    tmp1 <- rxSolve(derivs, e)

    eventdat2 <- data.frame(
      cmt = c("v1", "v2", "v2", "v1"),
      time = c(1, 1, 5, 9),
      amt = c(1, 2, 3, 4),
      evid = c(1, 6, 5, 1)
    )

    e <- et(eventdat2) |>
      et(seq(0, 10, by = 0.1))

    tmp3 <- rxSolve(derivs, e)

    expect_equal(as.data.frame(tmp1), as.data.frame(tmp3))
  })
})

rxTest({
  test_that("Test Issue RxODE#169, NAs everywhere", {
    Book4 <- readRDS(test_path("test-issue-169.rds"))
    mod1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      C2 <- centr / V2
      C3 <- peri / V3
      d / dt(depot) <- -KA * depot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- 1
    })
    trans <- etTrans(Book4, mod1)
    
    expect_equal(trans$EVID, c(
      101L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 101L, 0L, 0L, 0L, 0L,
      0L, 0L, 0L, 0L
    ))
    expect_true(all(trans$II == 0.0))
  })
})

if (!.Call(`_rxode2_isIntel`)) {
  # lhs for different compartments make sense

  et <- et() |>
    et(amt = 3, addl = 5, ii = 8, cmt = "depot") |>
    et(seq(0, 48, length.out = 200), cmt = "depot") |>
    et(seq(0, 48, length.out = 200), cmt = "centr") |>
    et(seq(0, 48, length.out = 200), cmt = "peri")

  ode.2c.ka <- rxode2({
    C2 <- centr / V
    C3 <- peri / V2
    d / dt(depot) <- -KA * depot
    d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
    d / dt(peri) <- Q * C2 - Q * C3
    C4 <- CMT
    if (CMT == 1) {
      prd <- depot
    }
    if (CMT == 2) {
      prd <- centr
    }
    if (CMT == 3) {
      prd <- peri
    }
  })

  tmp <- etTrans(et, ode.2c.ka)

  test_that("multi-compartment solves", {
    s <- rxSolve(ode.2c.ka, params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), et)
    expect_equal(sort(unique(s$C4)), c(1, 2, 3))
    expect_equal(s[s$C4 == 1, "prd"], s[s$C4 == 1, "depot"])
    expect_equal(s[s$C4 == 2, "prd"], s[s$C4 == 2, "centr"])
    expect_equal(s[s$C4 == 3, "prd"], s[s$C4 == 3, "peri"])
  })

  ## Now change cmt to an un-ordered factor.
  et <- as.data.frame(et)
  et$cmt <- factor(et$cmt, c("peri", "depot", "centr"), c("peri", "depot", "centr"))

  tmp <- etTrans(et, ode.2c.ka)

  test_that("multi-compartment solves", {
    s <- rxSolve(ode.2c.ka, params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), et)
    expect_equal(sort(unique(s$C4)), c(1, 2, 3))
    expect_equal(s[s$C4 == 1, "prd"], s[s$C4 == 1, "depot"])
    expect_equal(s[s$C4 == 2, "prd"], s[s$C4 == 2, "centr"])
    expect_equal(s[s$C4 == 3, "prd"], s[s$C4 == 3, "peri"])
  })
}

## nocov start
g.y.log10 <- function(breaks = g.log.breaks.major, minor_breaks = g.log.breaks.minor, labels = scales::math_format(format = log10), ...) {
  g.log.breaks.minor <- function(y) {
    r1 <- range(log10(y))
    r <- r1
    r[1] <- floor(r[1])
    r[2] <- ceiling(r[2]) + 1
    breaks <- NULL
    for (i in seq(r[1], r[2])) {
      breaks <- c(breaks, seq(2 * 10^(i - 1), 10^i - 10^(i - 1), by = 10^(i - 1)))
    }
    breaks <- breaks[breaks <= 10^r1[2]]
    breaks <- breaks[breaks >= 10^r1[1]]
    return(breaks)
  }
  g.log.breaks.major <- function(y) {
    r1 <- range(log10(y))
    r <- r1
    r[1] <- floor(r[1])
    r[2] <- ceiling(r[2]) + 1
    breaks <- 10^seq(r[1], r[2])
    breaks <- breaks[breaks <= 10^r1[2]]
    breaks <- breaks[breaks >= 10^r1[1]]
    return(breaks)
  }
  ggplot2::scale_y_log10(..., labels = labels, breaks = breaks, minor_breaks = minor_breaks)
}

#' Use Shiny to help develop an rxode2 model
#'
#' @param object A rxode2 family of objects. If not supplied a
#'     2-compartment indirect effect model is used.  If it is
#'     supplied, use the model associated with the rxode2 object for
#'     the model exploration.
#' @param params Initial parameters for model
#' @param events Event information (currently ignored)
#' @param inits Initial estimates for model
#' @param ... Other arguments passed to rxShiny.  Currently doesn't
#'     do anything.
#' @param data Any data that you would like to plot.  If the data has
#'     a `time` variable as well as a compartment or calculated
#'     variable that matches the rxode2 model, the data will be added
#'     to the plot of a specific compartment or calculated variable.
#' @return Nothing; Starts a shiny server
#' @author Zufar Mulyukov and Matthew L. Fidler
#' @export
rxShiny <- function(object, params = NULL, events = NULL, inits = NULL, ..., data = data.frame()) {
  UseMethod("rxShiny")
}
#' @rdname rxShiny
#' @export
rxShiny.rxSolve <- function(object, params = NULL, events = NULL, inits = NULL, ..., data = data.frame()) {
  if (is.null(params)) {
    if (dim(object$params)[1] > 1) {
      warning("using the first solved parameters for rxShiny", call. = FALSE)
    }
    params <- setNames(unlist(object$params[1, ]), names(object$params))
  }
  if (length(inits) == 0) {
    inits <- object$inits
  }
  rxShiny.default(object = object, params = params, events = events, inits = inits, ..., data = data)
}
#' @rdname rxShiny
#' @export
rxShiny.default <- function(object = NULL, params = NULL, events = NULL, inits = NULL, ...,
                            data = data.frame()) {
  rxReq("shiny")
  rxReq("ggplot2")
  rxReq("scales")
  if (is.null(object)) {
    object <- "
C2 = centr/V2
C3 = peri/V3
d/dt(depot) =-KA*depot
d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3
d/dt(peri)  =                    Q*C2 - Q*C3
d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff
"
    inits <- c(eff = 1)
    params <- c(
      KA = .291, CL = 18.6,
      V2 = 40.2, Q = 10.5, V3 = 297.0,
      Kin = 1.0, Kout = 1.0, EC50 = 200.0
    )
  }
  lower.names <- tolower(names(data))
  w <- which(lower.names == "time")
  if (length(w) == 1) {
    names(data)[w] <- "time"
    new.names <- c(rxLhs(object), rxState(object), "time")
    data <- data[, which(names(data) %in% c(rxLhs(object), rxState(object), "time")), drop = FALSE]
  }

  ui <- eval(bquote(shiny::shinyUI(shiny::fluidPage(
    shiny::tags$style(shiny::HTML("input:invalid {background-color: #FFCCCC;}")),
    shiny::tags$script('
$(document).on("keyup", function(e) {
  if(e.keyCode == 13){
  shiny::Shiny.onInputChange("EnterPressed", Math.random());
  }});
  '),
    shiny::fluidRow(
      shiny::column(
        12,
        shiny::fluidRow(
          shiny::column(
            5,
            shiny::textAreaInput(
              inputId = "ode",
              label = "ODE",
              width = 360,
              height = 200,
              resize = "both",
              value = .(rxNorm(object))
            ),
            shiny::actionButton("goButton", "Compile", align = "right"),
            shiny::actionButton("goPlot", "Update Plots", align = "right"),
            shiny::checkboxInput("goLogy", "Log y"),
            shiny::hr(),
            shiny::h4("Time sampling"),
            shiny::tags$div(style = "display: inline-block;", shiny::textInput("tmin", "start", width = 60, value = 0)),
            shiny::tags$div(style = "display: inline-block;", shiny::textInput("tmax", "end", width = 60, value = 100)),
            shiny::tags$div(style = "display: inline-block;", shiny::textInput("step", "step", width = 60, value = 1)),
            shiny::hr(),
            shiny::h4("Dosing"),
            shiny::tags$div(style = "display: inline-block;", shiny::textInput("dose", "amount", width = 60, value = 1)),
            shiny::tags$div(style = "display: inline-block;", shiny::textInput("rate", "rate", width = 60)),
            shiny::div(style = "display: inline-block;", shiny::uiOutput("dosing_cmt")),
            shiny::br(),
            shiny::tags$div(style = "display: inline-block;", shiny::textInput("start", "start", width = 60, value = 0)),
            shiny::tags$div(style = "display: inline-block;", shiny::textInput("ndoses", "repeat", width = 60, value = 1)),
            shiny::tags$div(style = "display: inline-block;", shiny::textInput("interval", "interval", width = 60, value = 0))
          ),
          shiny::column(
            width = 7,
            shiny::fluidRow(
              shiny::h4("Initial values"),
              shiny::uiOutput("cmts"),
              shiny::br(),
              shiny::h4("Parameters"),
              shiny::uiOutput("pars"),
              shiny::uiOutput("message", placeholder = FALSE),
              shiny::uiOutput("plotTabs")
            )
          )
        )
      )
    )
  ))))

  server <- eval(bquote(function(input, output, session) {
    values <- reactiveValues()
    m1 <- NULL
    values$res <- NULL
    values$msg <- "NULL"
    values$logy <- TRUE
    ## tmp <- tempfile()
    shiny::observeEvent(input$goButton, {
      values$m1 <- NULL
      values$res <- NULL

      values$msg <- capture.output(
        tryCatch(
          {
            values$m1 <- rxode2(model = input$ode)
            values$cmts <- rxState(values$m1)
            values$pars <- .rxParams(values$m1)
          },
          error = function(e) {
            return(e)
          }
        )
      )
    })
    observe({
      if (!is.null(values$m1)) values$msg <- "Model is compiled. Enter parameters."
    })

    output$message <- shiny::renderPrint({
      if (length(values$msg) == 0) {
        return(invisible())
      }
      if (values$msg == "NULL") {
        return(invisible())
      }

      cat("<h3>Message:</h3>")
      cat(paste(gsub("<|>", "", values$msg), collapse = "<br>"))
    })

    output$dosing_cmt <- shiny::renderUI({
      if (is.null(values$cmts)) {
        return()
      }
      selectInput("into", "into", width = 80, choices = values$cmts, selected = values$cmts[1])
    })

    output$cmts <- shiny::renderUI({
      if (is.null(values$cmts)) {
        return()
      }
      if (length(values$cmts) == 0) {
        return("none")
      }

      tmp <- rxInits(values$m1, .(inits), rxState(values$m1), 0, TRUE)
      lapply(values$cmts, function(x) {
        shiny::div(style = "display: inline-block;", shiny::textInput(x, x, value = tmp[x], width = 60))
      })
    })

    output$pars <- shiny::renderUI({
      if (is.null(values$pars)) {
        return()
      }
      if (length(values$pars) == 0) {
        return("none")
      }
      tmp <- rxInits(values$m1, vec = .(params), req = rxParams(values$m1), defaultValue = 1)
      lapply(values$pars, function(x) {
        shiny::div(style = "display: inline-block;", shiny::textInput(x, x, value = tmp[x], width = 60))
      })
    })

    output$plotTabs <- shiny::renderUI({
      if (is.null(values$res)) {
        return(tabsetPanel())
      }
      dat <- values$res

      cmts <- colnames(values$res)[-1]

      sel.tab <- isolate(ifelse(is.null(input$plot.tabs), cmts[1], input$plot.tabs))

      tabs <-
        lapply(cmts, function(cmt) {
          plotname <- paste("plot", cmt, sep = "")
          output[[plotname]] <- renderPlot({
            tmp <- tolower(cmt)
            p <- ggplot2::ggplot(as.data.frame(dat), ggplot2::aes_(x = as.name("time"), y = as.name(cmt)))
            if (compareVersion(as.character(packageVersion("ggplot2")), "3.4.0") < 0) {
              p <- p + geom_line(size = 1.2)
            } else {
              p <- p + geom_line(linewidth = 1.2)
            }

            p <- p +
              ggplot2::theme_bw(base_size = 18)
            if (values$logy) {
              p <- p + g.y.log10()
            }
            data <- .(data)
            if (any(cmt == names(data))) {
              p <- p + ggplot2::geom_point(data = data)
            }
            p
          })
          tabPanel(cmt, plotOutput(plotname))
        })
      do.call(tabsetPanel, c(tabs, id = "plot.tabs", selected = sel.tab))
    })


    shiny::observeEvent(input$goLogy, {
      values$logy <- !(values$logy)
      values$msg <- capture.output(
        tryCatch(
          {
            solveODE()
          },
          error = function(e) {
            return(e)
          }
        )
      )
    })

    shiny::observeEvent(input$goPlot, {
      values$msg <- capture.output(
        tryCatch(
          {
            solveODE()
          },
          error = function(e) {
            return(e)
          }
        )
      )
    })

    shiny::observeEvent(
      {
        input$EnterPressed
      },
      {
        values$msg <- capture.output(
          tryCatch(
            {
              solveODE()
            },
            error = function(e) {
              return(e)
            }
          )
        )
      }
    )


    solveODE <- function() {
      values$res <- NULL
      if (is.null(values$m1)) {
        return()
      }

      stime <- as.numeric(input$tmin)
      etime <- as.numeric(input$tmax)
      tstep <- as.numeric(input$step)
      dose <- as.numeric(input$dose)
      rate <- as.numeric(input$rate)

      if (is.na(rate)) {
        rate <- NULL
      }
      if (is.na(dose)) {
        dose <- 0
      }

      ndoses <- as.numeric(input$ndoses)
      start <- as.numeric(input$start)
      interval <- as.numeric(input$interval)

      into <- ifelse(is.null(input$into), 1, match(input$into, values$cmts))

      ev <- eventTable() |>
        add.sampling(seq(stime, etime, tstep)) |>
        add.dosing(
          dose = dose, start.time = start,
          nbr.doses = ndoses, rate = rate,
          dosing.interval = interval,
          dosing.to = into
        )

      params <- .(params)


      cmts <- values$cmts

      inits <- NULL
      if (length(cmts) > 0) {
        init_str <- paste0(cmts, "=as.numeric(input$", cmts, ")")
        init_str <- paste("c(", toString(init_str), ")")
        inits <- eval(parse(text = init_str))
      }

      pars <- values$pars
      params <- NULL
      if (length(pars) > 0) {
        param_str <- paste0(pars, "=as.numeric(input$", pars, ")")
        param_str <- paste("c(", toString(param_str), ")")
        params <- eval(parse(text = param_str))
      }

      values$res <- values$m1$solve(params, ev, inits)
    }

    session$onSessionEnded(stopApp)
  }))


  shiny::shinyApp(ui = ui, server = server)
}
## nocov end

#' Exponential Linear Unit (ELU) Activation Function
#'
#' @family Activation Functions
#' @param x A numeric vector. All elements must be finite and
#'   non-missing.
#' @param alpha A numeric scalar. All elements must be finite and
#'  non-missing.
#' @return A numeric vector where the ReLU function has been applied
#'   to each element of `x`.
#' @author Matthew Fidler
#' @export
#' @examples
#'
#' ELU(c(-1, 0, 1, 2), 2)
#'
#' \donttest{
#' # Can also be used in rxode2:
#' x <- rxode2({
#'    r=ELU(time, 2)
#' })
#'
#' e <- et(c(-1, 0, 1, 2))
#'
#' rxSolve(x, e)
#' }
#'
ELU <- function(x, alpha=1) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  checkmate::assertNumeric(alpha, finite=TRUE, any.missing=FALSE)
  .df <- data.frame(x=x, alpha=alpha)
  .Call(`_rxode2_activationF2`, .df$x, .df$alpha, 1L)
}
#' Derivatives of the Exponential Linear Unit (ELU) Activation Function
#'
#'
#' @param x A numeric vector. All elements must be finite and
#'  non-missing.
#' @param alpha A numeric scalar. All elements must be finite and
#'  non-missing.
#' @return A numeric vector where the derivative(s) of the ELU function has been applied
#'  to each element of `x`.
#' @export
#' @author Matthew L. Fidler
#' @family Activation Functions
#' @examples
#' dELU(c(-1, 0, 1, 2), 2)
#' d2ELU(c(-1, 0, 1, 2), 2)
#' d2aELU(c(-1, 0, 1, 2), 2)
#' dELUa(c(-1, 0, 1, 2), 2)
#' d2ELUa(c(-1, 0, 1, 2), 2)
#'
#' \donttest{
#' # Can also be used in rxode2:
#' r <- rxode2({
#'   r1=dELU(time, 2)
#'   r2=d2ELU(time, 2)
#'   r2a=d2aELU(time, 2)
#'   ra=dELUa(time, 2)
#'   r2a=d2ELUa(time, 2)
#' })
#'
#' e <- et(c(-1, 0, 1, 2))
#' rxSolve(r, e)
#' }
dELU <- function(x, alpha=1) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  checkmate::assertNumeric(alpha, finite=TRUE, any.missing=FALSE)
  .df <- data.frame(x=x, alpha=alpha)
  .Call(`_rxode2_activationF2`, .df$x, .df$alpha, 2L)
}

#' @rdname dELU
#' @export
d2ELU <- function(x, alpha=1) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  checkmate::assertNumeric(alpha, finite=TRUE, any.missing=FALSE)
  .df <- data.frame(x=x, alpha=alpha)
  .Call(`_rxode2_activationF2`, .df$x, .df$alpha, 3L)
}

#' @rdname dELU
#' @export
d2aELU <- function(x, alpha=1) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  checkmate::assertNumeric(alpha, finite=TRUE, any.missing=FALSE)
  .df <- data.frame(x=x, alpha=alpha)
  .Call(`_rxode2_activationF2`, .df$x, .df$alpha, 4L)
}

#' @rdname dELU
#' @export
dELUa <- function(x, alpha=1) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  checkmate::assertNumeric(alpha, finite=TRUE, any.missing=FALSE)
  .df <- data.frame(x=x, alpha=alpha)
  .Call(`_rxode2_activationF2`, .df$x, .df$alpha, 5L)
}

#' @rdname dELU
#' @export
d2ELUa <- function(x, alpha=1) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  checkmate::assertNumeric(alpha, finite=TRUE, any.missing=FALSE)
  .df <- data.frame(x=x, alpha=alpha)
  .Call(`_rxode2_activationF2`, .df$x, .df$alpha, 6L)
}

rxTest({
  test_that("activation functions", {

    r <- rxode2({
      r <- ReLU(time)
      dr <- dReLU(time)
      #
      l <- lReLU(time)
      dl <- dlReLU(time)
      #
      p <- PReLU(time, alpha)
      dp <- dPReLU(time, alpha)
      dpa <- dPReLUa(time, alpha)
      dpa1 <- dPReLUa1(time, alpha)
      #
      g <- GELU(time)
      dg <- dGELU(time)
      d2g <- d2GELU(time)
      d3g <- d3GELU(time)
      d4g <- d4GELU(time)
      #
      s <- Swish(time)
      ds <- dSwish(time)
      #
      sp <- softplus(time)
      dsp <- dsoftplus(time)
      #
      el <- ELU(time, alpha)
      del <- dELU(time, alpha)
      d2el <- d2ELU(time, alpha)
      d2a <- dELUa(time, alpha)
      d2ael <- d2aELU(time, alpha)
      dael <- d2aELU(time, alpha)
      d2aela <- d2ELUa(time, alpha)
    })

    e <- et(seq(-10, 10, length.out = 41))

    s <- suppressWarnings(rxSolve(r, e, c(alpha=2)))

    expect_equal(ReLU(e$time), s$r)
    expect_equal(ReLU(e$time), vapply(e$time,
                                      function(x) {
                                        max(0, x)
                                      }, double(1)))
    expect_equal(dReLU(e$time), s$dr)
    expect_equal(dReLU(e$time), vapply(e$time,
                                      function(x) {
                                        (x > 0)*1
                                      }, double(1)))
    expect_equal(lReLU(e$time), s$l)
    expect_equal(lReLU(e$time), vapply(e$time,
                                       function(x) {
                                         (x > 0)*x + (x <= 0)*0.01*x
                                      }, double(1)))
    expect_equal(dlReLU(e$time), s$dl)
    expect_equal(dlReLU(e$time), vapply(e$time,
                                       function(x) {
                                         (x > 0) + (x <= 0)*0.01
                                       }, double(1)))

    expect_equal(PReLU(e$time, 2), s$p)
    expect_equal(PReLU(e$time, 2), vapply(e$time,
                                       function(x) {
                                         (x > 0)*x + (x <= 0)*2*x
                                       }, double(1)))

    expect_equal(dPReLU(e$time, 2), s$dp)
    expect_equal(dPReLU(e$time, 2), vapply(e$time,
                                          function(x) {
                                            (x > 0) + (x <= 0)*2
                                          }, double(1)))

    expect_equal(GELU(e$time), s$g)
    expect_equal(GELU(e$time), vapply(e$time,
                                      function(x) {
                                        phi(x)*x
                                      }, double(1)))

    expect_equal(dGELU(e$time), s$dg)
    expect_equal(dGELU(e$time), vapply(e$time,
                                       function(x) {
                                         M_SQRT1_2 <- 1/sqrt(2)
                                         1.0*x*exp(-x^2*M_SQRT1_2^2)*M_SQRT1_2/sqrt(pi) + 0.5*(1 + erf(x*M_SQRT1_2))
                                       }, double(1)))
    expect_equal(d2GELU(e$time), s$d2g)
    expect_equal(d2GELU(e$time),
                 vapply(e$time,
                        function(x) {
                          M_SQRT1_2 <- 1/sqrt(2)
                          2.0*exp(-x^2*M_SQRT1_2^2)*M_SQRT1_2/sqrt(pi) - 2.0*x^2*exp(-x^2*M_SQRT1_2^2)*M_SQRT1_2^3/sqrt(pi)
                        }, double(1)))

    expect_equal(d3GELU(e$time), s$d3g)
    expect_equal(d3GELU(e$time),
                 vapply(e$time,
                        function(x) {
                          M_SQRT1_2 <- 1/sqrt(2)
                          M_1_SQRT_2PI <- 1/sqrt(2*pi)
                          -8.0*x*exp(-x^2*0.5)*0.5*M_1_SQRT_2PI + 4.0*x^3*exp(-x^2*0.5)*0.25*M_1_SQRT_2PI
                        }, double(1)))
    expect_equal(d4GELU(e$time), s$d4g)
    expect_equal(d4GELU(e$time),
                 vapply(e$time,
                        function(x) {
                          M_SQRT1_2 <- 1/sqrt(2)
                          -8.0*exp(-x^2*M_SQRT1_2^2)*M_SQRT1_2^3/sqrt(pi) + 28.0*x^2*exp(-x^2*M_SQRT1_2^2)*M_SQRT1_2^5/sqrt(pi) - 8.0*x^4*exp(-x^2*M_SQRT1_2^2)*M_SQRT1_2^7/sqrt(pi)
                        }, double(1)))

    expect_equal(Swish(e$time), s$s)
    expect_equal(Swish(e$time), vapply(e$time,
                                       function(x) {
                                          x/(1 + exp(-x))
                                       }, double(1)))

    expect_equal(dSwish(e$time), s$ds)
    expect_equal(dSwish(e$time), vapply(e$time,
                                       function(x) {
                                         x*exp(-x)/(1 + exp(-x))^2 + (1 + exp(-x))^(-1)
                                       }, double(1)))

    expect_equal(softplus(e$time), s$sp)
    expect_equal(softplus(e$time), vapply(e$time,
                                          function(x) {
                                            log(1 + exp(x))
                                          }, double(1)))

    expect_equal(dsoftplus(e$time), s$dsp)
    expect_equal(dsoftplus(e$time), vapply(e$time,
                                          function(x) {
                                            exp(x)/(1 + exp(x))
                                          }, double(1)))

    expect_equal(ELU(e$time, 2), s$el)
    expect_equal(ELU(e$time, 2), vapply(e$time,
                                        function(x) {
                                          (x > 0)*x + (x <= 0)*2*(exp(x) - 1)
                                        }, double(1)))

    # dELU/dx
    expect_equal(dELU(e$time, 2), s$del)
    expect_equal(dELU(e$time, 2), vapply(e$time,
                                        function(x) {
                                          (x > 0) + (x <= 0)*2*exp(x)
                                        }, double(1)))

    # d2ELU/d2x
    expect_equal(d2ELU(e$time, 2), s$d2el)
    expect_equal(d2ELU(e$time, 2), vapply(e$time,
                                          function(x) {
                                            (x <= 0)*2*exp(x)
                                          }, double(1)))

    # dELU/dalpha
    expect_equal(dELUa(e$time, 2), s$d2a)
    expect_equal(dELUa(e$time, 2), vapply(e$time,
                                          function(x) {
                                            (x <= 0)*(exp(x) - 1)
                                          }, double(1)))

    # d2ELU/dalphad2x
    expect_equal(d2ELUa(e$time, 2), s$d2ael)
    expect_equal(d2ELUa(e$time, 2), vapply(e$time,
                                          function(x) {
                                            (x <= 0)*exp(x)
                                          }, double(1)))

  })
})

#' Validate rxode2
#' This allows easy validation/qualification of nlmixr by running the
#' testing suite on your system.
#'
#' @param type Type of test or filter of test type, When this is an
#'   expression, evaluate the contents, respecting `skipOnCran`.  In
#'   the expression form, stray progress messages are muffled unless
#'   `options(rxode2.test.verbose = TRUE)` is set.
#' @param skipOnCran when `TRUE` skip the test on CRAN.
#' @author Matthew L. Fidler
#' @return nothing
#' @export
rxValidate <- function(type = NULL, skipOnCran=TRUE) {
  if (is(substitute(type), "{")) {
    if (isTRUE(skipOnCran)) {
      if (!identical(Sys.getenv("NOT_CRAN"), "true") ||
            !identical(Sys.getenv("rxTest"), "")) {
        return(invisible())
      }
    }
    if (!.Call(`_rxode2_isIntel`)) {
      rxUnloadAll()
    }
    # Preserve the rxode2 parallel/internal RNG seed across each test block.
    # A test that calls rxSetSeed(<n>) leaves useRxSeed=TRUE set globally;
    # without this guard that state leaks into later test files and breaks
    # set.seed()-based reproducibility tests (e.g. test-random, test-rxRmvn).
    .rxSeed0 <- rxGetSeed()
    on.exit(rxSetSeed(.rxSeed0), add = TRUE)
    if (getOption("rxode2.test.verbose", FALSE)) {
      return(force(type))
    }
    # Muffle stray progress messages (cli alerts, rxCat) during tests;
    # expect_message()/verify_output() handlers run innermost and still
    # see what they assert on.  The rxParseSuppressMsg()/.rxSilentErr
    # probe message(" ") must pass through, or muffling is mistaken for
    # suppressMessages() and sets the C-level silent flag, hiding
    # parse-error output some tests assert on.
    return(withCallingHandlers(
      force(type),
      message = function(m) {
        if (!identical(conditionMessage(m), " \n")) {
          tryInvokeRestart("muffleMessage")
        }
      }))
  }
  pt <- proc.time()
  .filter <- NULL
  if (is.null(type)) type <- FALSE
  if (is.character(type)) {
    .filter <- type
    type <- TRUE
  }
  if (type == TRUE) {
    .oldCran <- Sys.getenv("NOT_CRAN")
    .oldRxTest <- Sys.getenv("rxTest")
    Sys.setenv("NOT_CRAN" = "true") # nolint
    Sys.setenv("rxTest" = "") # nolint
    on.exit(Sys.setenv("NOT_CRAN" = .oldCran, "rxTest"=.oldRxTest)) # nolint
  } else if (type == FALSE) {
    .oldCran <- Sys.getenv("NOT_CRAN")
    .oldRxTest <- Sys.getenv("rxTest")
    Sys.setenv("NOT_CRAN" = "false") # nolint
    Sys.setenv("rxTest" = "false") # nolint
    on.exit(Sys.setenv("NOT_CRAN" = .oldCran, "rxTest"=.oldRxTest)) # nolint
  }
  .rxWithOptions(list(testthat.progress.max_fails = 10000000000), {
    path <- file.path(system.file("tests", package = "rxode2"), "testthat")
    .rxWithWd(path, {
      try(devtools::test(path, filter = .filter))
      message("=================================================================================")
      print(proc.time() - pt)
      message("=================================================================================")
    })
  })
}

#' @rdname rxValidate
#' @export
rxTest <- rxValidate

## Adjoint (backward) sensitivity analysis -- R orchestration around the
## symbolic core in .rxAdjoint() (R/rxJacobian.R).  Reconstructs the full
## per-timepoint trajectory dy_k(t_i)/dp using a dense forward checkpoint fed to
## the backward model as covariates (avoids the unstable backward primal
## reconstruction).  Reference/validation implementation; the in-engine C driver
## uses each solver's own dense output.

#' Solve a model with adjoint (backward) sensitivities, drop-in for forward sens
#'
#' Solves `object` as [rxSolve()] would, then appends
#' `rx__sens_<state>_BY_<param>__` columns computed via adjoint (backward)
#' sensitivity analysis.  Same column names and output structure as
#' `rxSolve(object, ..., calcSens=)`.  Prefer [.rxAdjointGrad()] when only a
#' scalar objective gradient is needed.
#'
#' @param object model definition accepted by [rxode2::rxode2()].
#' @param params named numeric vector of parameter values.
#' @param events event table / data used to define dosing and sampling times;
#'   sampling (`evid==0`) times become the sensitivity output times.
#' @param calcSens character vector of parameter names to differentiate with
#'   respect to.
#' @param adjStates character vector of output states of interest; defaults to
#'   all ODE states (full forward-sensitivity parity).
#' @param denseBy,atol,rtol passed to the adjoint checkpoint solve; see
#'   [.rxAdjointSolveEvalC()].
#' @param ... additional arguments passed to the primal [rxSolve()] call.
#' @return the standard `rxSolve()` output (as `returnType="data.frame"`) with
#'   `rx__sens_<state>_BY_<param>__` columns appended.
#' @author Matthew L. Fidler
#' @export
rxSolveAdjoint <- function(object, params, events, calcSens, adjStates = NULL,
                           denseBy = 0.01, atol = 1e-08, rtol = 1e-06, ...) {
  .primal <- as.data.frame(rxode2::rxSolve(object, params = params, events,
                                           returnType = "data.frame", ...))
  .obsTimes <- sort(unique(.primal$time))
  .build <- .rxAdjointSolveBuild(object, calcSens, events, adjStates = adjStates)
  .sens <- .rxAdjointSolveEvalC(.build, params, .obsTimes, denseBy = denseBy,
                                atol = atol, rtol = rtol)
  .sensCols <- setdiff(names(.sens), "time")
  .idx <- match(.primal$time, .sens$time)
  for (.cn in .sensCols) .primal[[.cn]] <- .sens[[.cn]][.idx]
  .primal
}

## Reverse a backward-in-time "d/dt(L)=RHS" line to reversed-time forward form.
.rxAdjointNegLine <- function(line) {
  .m <- regmatches(line, regexec("^\\s*(d/dt\\([^)]*\\))\\s*=\\s*(.*)$", line))[[1]]
  paste0(.m[2], "=-(", .m[3], ")")
}

#' Solve the full-trajectory adjoint sensitivities for a model
#'
#' Reconstructs `dy_k(t_i)/dp` for every output state `k`, parameter `p` and
#' output time `t_i` via backward (adjoint) integration, returning the same
#' `rx__sens_<state>_BY_<param>__` columns that forward sensitivities produce.
#'
#' @param object model definition accepted by [rxode2::rxode2()] (text,
#'   `rxode2` object, or function/ui).
#' @param params named numeric vector of parameter values.
#' @param events event table / data used to define dosing.
#' @param calcSens character vector of parameter names to differentiate with
#'   respect to.
#' @param outTimes numeric vector of output (observation) times at which the
#'   sensitivities are requested.
#' @param adjStates character vector of output states of interest.  Defaults to
#'   all ODE states (full forward-sensitivity parity).
#' @param denseBy grid spacing for the forward checkpoint trajectory; smaller
#'   values reduce the covariate-interpolation error.
#' @param atol,rtol solver tolerances used for both the forward checkpoint and
#'   the backward sweeps.
#' @return data.frame with a `time` column and one
#'   `rx__sens_<state>_BY_<param>__` column per (state, param) pair.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjointSolve <- function(object, params, events, calcSens, outTimes,
                            adjStates = NULL, denseBy = 0.01,
                            atol = 1e-10, rtol = 1e-10) {
  .mText <- rxode2::rxNorm(rxode2::rxModelVars(object))
  .model <- rxode2::rxS(rxode2::rxGetModel(object), TRUE, promoteLinSens = FALSE)
  .st    <- rxode2::rxStateOde(.model)
  if (length(.st) == 0L) {
    stop("adjoint sensitivities require a model with ODE states", call. = FALSE)
  }
  if (is.null(adjStates)) adjStates <- .st
  adjStates <- intersect(.st, adjStates)
  invisible(rxode2::.rxJacobian(.model, c(.st, calcSens)))
  .adj <- rxode2::.rxAdjoint(.model, calcSens, adjStates)
  ## backward model in reversed time (forward states enter as covariates)
  .revMod <- rxode2::rxode2(paste(vapply(.adj, .rxAdjointNegLine, character(1)),
                                  collapse = "\n"))

  .nm <- function(state, p) paste0("rx__sens_", state, "_BY_", p, "__")
  .cols <- unlist(lapply(adjStates, function(k)
    vapply(calcSens, function(p) .nm(k, p), character(1))))

  ## dense forward checkpoint trajectory
  .denseT <- sort(unique(c(seq(0, max(outTimes), by = denseBy), outTimes)))
  .fev <- events |> rxode2::et(.denseT)
  .fwd <- as.data.frame(rxode2::rxSolve(rxode2::rxode2(.mText), params = params,
                                        .fev, returnType = "data.frame",
                                        atol = atol, rtol = rtol,
                                        addDosing = FALSE))
  .Y <- lapply(.st, function(s) stats::approx(.fwd$time, .fwd[[s]], .denseT,
                                              rule = 2)$y)
  names(.Y) <- .st

  ## terminal condition: costate matrix = identity, quadrature = 0
  .inits <- numeric(0)
  for (k in adjStates) for (i in .st)
    .inits[paste0("rx__adjLambda_", k, "_", i, "__")] <- as.numeric(k == i)
  for (k in adjStates) for (p in calcSens) .inits[.nm(k, p)] <- 0

  .out <- matrix(NA_real_, length(outTimes), length(.cols),
                 dimnames = list(NULL, .cols))
  for (.ii in seq_along(outTimes)) {
    .ti <- outTimes[.ii]
    .idx <- which(.denseT <= .ti + 1e-12)
    ## reversed time grid + time-reversed forward states as covariates
    .dat <- data.frame(time = .ti - rev(.denseT[.idx]), evid = 0L)
    for (.sn in .st) .dat[[.sn]] <- rev(.Y[[.sn]][.idx])
    .sol <- as.data.frame(rxode2::rxSolve(.revMod, params = params, .dat,
                                          inits = .inits, returnType = "data.frame",
                                          atol = atol, rtol = rtol,
                                          covsInterpolation = "linear",
                                          addDosing = FALSE))
    .out[.ii, ] <- unlist(.sol[nrow(.sol), .cols])
  }
  cbind(time = outTimes, as.data.frame(.out))
}

#' Build the symbolic part of a full-trajectory adjoint solve (compile once)
#'
#' Performs all symbolic work (Jacobian, dose-dual expressions, model
#' compilation) once and returns a reusable object for [.rxAdjointSolveEvalC()],
#' the C++-backed counterpart of [.rxAdjointSolve()].
#'
#' @inheritParams .rxAdjointSolve
#' @return an opaque list consumed by [.rxAdjointSolveEvalC()].
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjointSolveBuild <- function(object, calcSens, events, adjStates = NULL) {
  .mText <- rxode2::rxNorm(rxode2::rxModelVars(object))
  .model <- rxode2::rxS(rxode2::rxGetModel(object), TRUE, promoteLinSens = FALSE)
  .st <- rxode2::rxStateOde(.model)
  if (length(.st) == 0L) {
    stop("adjoint solve requires a model with ODE states", call. = FALSE)
  }
  if (is.null(adjStates)) adjStates <- .st
  adjStates <- intersect(.st, adjStates)
  invisible(rxode2::.rxJacobian(.model, c(.st, calcSens)))

  ## event detection (identical to .rxAdjointGradBuild)
  .ev <- as.data.frame(events)
  .dr <- which(!is.na(.ev$evid) & .ev$evid == 1L & !is.na(.ev$amt) & .ev$amt != 0)
  .rateCol <- if ("rate" %in% names(.ev)) .ev$rate[.dr] else rep(0, length(.dr))
  .rateCol[is.na(.rateCol)] <- 0
  .infusSym <- list()
  for (.k in which(.rateCol %in% c(-1, -2))) {
    .c <- as.character(.ev$cmt[.dr][.k])
    if (!(.c %in% .st)) next
    .t1 <- .ev$time[.dr][.k]; .amt <- .ev$amt[.dr][.k]
    if (.rateCol[.k] == -1) {
      .rSE <- get0(paste0("rx_rate_", .c, "_"), envir = .model, inherits = FALSE)
      if (is.null(.rSE)) next
      .valStr <- rxode2::rxFromSE(.rSE)
      .dStr <- vapply(calcSens, function(p) {
        .d <- symengine::D(.rSE, p); rxode2::rxFromSE(.d) }, character(1))
      .kind <- "rate"
    } else {
      .dSE <- get0(paste0("rx_dur_", .c, "_"), envir = .model, inherits = FALSE)
      if (is.null(.dSE)) next
      .valStr <- rxode2::rxFromSE(.dSE)
      .dStr <- vapply(calcSens, function(p) {
        .d <- symengine::D(.dSE, p); rxode2::rxFromSE(.d) }, character(1))
      .kind <- "dur"
    }
    .infusSym[[length(.infusSym) + 1L]] <- list(cmt = .c, t1 = .t1, amt = .amt,
                                                kind = .kind, valStr = .valStr, dStr = .dStr)
  }
  .doses <- data.frame(time = .ev$time[.dr], cmt = as.character(.ev$cmt[.dr]),
                       amt = .ev$amt[.dr], rate = .rateCol, stringsAsFactors = FALSE)
  .dFexpr <- list(); .dLagStr <- list(); .lagStr <- list()
  for (.c in unique(.doses$cmt[.doses$cmt %in% .st])) {
    .fSE <- get0(paste0("rx_f_", .c, "_"), envir = .model, inherits = FALSE)
    if (!is.null(.fSE)) .dFexpr[[.c]] <- vapply(calcSens, function(p) {
      .d <- symengine::D(.fSE, p); rxode2::rxFromSE(.d) }, character(1))
    .lSE <- get0(paste0("rx_lag_", .c, "_"), envir = .model, inherits = FALSE)
    if (!is.null(.lSE)) {
      .lagStr[[.c]] <- rxode2::rxFromSE(.lSE)
      .dLagStr[[.c]] <- vapply(calcSens, function(p) {
        .d <- symengine::D(.lSE, p); rxode2::rxFromSE(.d) }, character(1))
    }
  }
  .evR <- .ev[!is.na(.ev$evid) & .ev$evid %in% c(5L, 6L), , drop = FALSE]
  .evR$cmt <- as.character(.evR$cmt)
  .fLhs <- unlist(lapply(names(.dFexpr), function(.c)
    vapply(calcSens, function(p)
      paste0("rx__dFdp_", .c, "_", p, "__=", .dFexpr[[.c]][[p]]), character(1))))
  .needF <- length(.dLagStr) > 0L
  .fRHSlhs <- if (.needF) vapply(.st, function(i) {
    .d <- get0(paste0("rx__d_dt_", i, "__"), envir = .model, inherits = FALSE)
    paste0("rx__fRHS_", i, "__=", rxode2::rxFromSE(.d)) }, character(1)) else character(0)

  ## forward model emitting J(t), forcing df/dtheta and dose-dual quantities
  .jacLines <- unlist(lapply(.st, function(j) lapply(c(.st, calcSens), function(x) {
    .d <- get0(paste0("rx__df_", j, "_dy_", x, "__"), envir = .model, inherits = FALSE)
    paste0("rx__df_", j, "_dy_", x, "__=", rxode2::rxFromSE(.d))
  })))
  .cfmod <- rxode2::rxode2(paste(c(.mText, .jacLines, .fLhs, .fRHSlhs), collapse = "\n"))

  list(st = .st, calcSens = calcSens, adjStates = adjStates, cfmod = .cfmod,
       infusSym = .infusSym, doses = .doses, evR = .evR, dFexpr = .dFexpr,
       dLagStr = .dLagStr, lagStr = .lagStr, events = events)
}

#' Evaluate a prebuilt full-trajectory adjoint solve in C++
#'
#' Full-trajectory counterpart of [.rxAdjointGradEvalC()]: reconstructs
#' `dy_k(t_i)/dp` for every output state, time and parameter in one backward
#' sweep (`rxode2AdjointTrajSweep`), covering all dosing duals.  Numerically
#' equivalent to [.rxAdjointSolve()].
#'
#' @param build object returned by [.rxAdjointSolveBuild()].
#' @param params named numeric vector of parameter values.
#' @param outTimes numeric vector of output times.
#' @param denseBy,atol,rtol as in [.rxAdjointSolve()].
#' @return data.frame with a `time` column and one
#'   `rx__sens_<state>_BY_<param>__` column per (state, param) pair.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjointSolveEvalC <- function(build, params, outTimes, denseBy = 0.01,
                                 atol = 1e-10, rtol = 1e-10) {
  .st <- build$st; calcSens <- build$calcSens; adjStates <- build$adjStates
  .ns <- length(.st); .np <- length(calcSens); .nk <- length(adjStates)
  .doses <- build$doses
  .doses$tau <- .doses$time + vapply(seq_len(nrow(.doses)), function(i) {
    .c <- .doses$cmt[i]
    .lv <- if (.c %in% names(build$lagStr)) .rxAdjEvalNum(build$lagStr[[.c]], params) else 0
    if (length(.lv) && !is.na(.lv)) .lv else 0 }, numeric(1))
  .infus <- lapply(build$infusSym, function(z) {
    if (z$kind == "rate") {
      .Rv <- .rxAdjEvalNum(z$valStr, params)
      .dRp <- vapply(z$dStr, function(s) .rxAdjEvalNum(s, params), numeric(1))
    } else {
      .Dv <- .rxAdjEvalNum(z$valStr, params)
      .dDp <- vapply(z$dStr, function(s) .rxAdjEvalNum(s, params), numeric(1))
      .Rv <- z$amt / .Dv; .dRp <- -(z$amt / .Dv^2) * .dDp
    }
    list(cmt = z$cmt, t1 = z$t1, t2 = z$t1 + z$amt / .Rv, R = .Rv,
         dRdp = stats::setNames(.dRp, calcSens), amt = z$amt)
  })
  .infT <- unlist(lapply(.infus, function(z) c(z$t1, z$t2)))
  .denseT <- sort(unique(c(seq(0, max(outTimes), by = denseBy), outTimes,
                           build$evR$time, .doses$tau, .doses$tau - denseBy, .infT)))
  .denseT <- .denseT[.denseT >= 0]
  .fev <- build$events |> rxode2::et(.denseT)
  .fwd <- as.data.frame(rxode2::rxSolve(build$cfmod, params = params, .fev,
                                        returnType = "data.frame", atol = atol,
                                        rtol = rtol, addDosing = FALSE))
  .fwd <- .fwd[match(.denseT, .fwd$time), ]
  .nt <- length(.denseT)
  .J <- matrix(0.0, .nt, .ns * .ns)
  for (i in seq_len(.ns)) for (j in seq_len(.ns))
    .J[, (i - 1) * .ns + j] <- .fwd[[paste0("rx__df_", .st[i], "_dy_", .st[j], "__")]]
  .dP <- matrix(0.0, .nt, .ns * .np)
  for (i in seq_len(.ns)) for (p in seq_len(.np))
    .dP[, (i - 1) * .np + p] <- .fwd[[paste0("rx__df_", .st[i], "_dy_", calcSens[p], "__")]]
  for (.z in .infus) {
    .ci <- match(.z$cmt, .st)
    if (is.na(.ci)) next
    .inWin <- .denseT >= .z$t1 - 1e-9 & .denseT <= .z$t2 + 1e-9
    for (p in seq_len(.np))
      .dP[.inWin, (.ci - 1) * .np + p] <- .dP[.inWin, (.ci - 1) * .np + p] + .z$dRdp[[p]]
  }

  ## replace/multiply costate jumps
  .evR <- build$evR
  .cj <- if (nrow(.evR) > 0L) {
    .cmt0 <- match(.evR$cmt, .st) - 1L
    list(as.integer(match(.evR$time, .denseT) - 1L), as.integer(.cmt0),
         ifelse(.evR$evid == 5L, 0.0, .evR$amt))
  } else list(integer(0), integer(0), numeric(0))
  ## bolus dose duals: F and modeled lag
  .duals <- list()
  for (.j in which(.doses$rate == 0)) {
    .c <- .doses$cmt[.j]; .ci <- match(.c, .st); .tau <- .doses$tau[.j]
    if (is.na(.ci)) next
    .k <- which.min(abs(.denseT - .tau))
    if (.c %in% names(build$dFexpr)) {
      .w <- numeric(.ns); .w[.ci] <- .doses$amt[.j]
      .cc <- vapply(calcSens, function(p) .fwd[[paste0("rx__dFdp_", .c, "_", p, "__")]][.k],
                    numeric(1))
      .duals[[length(.duals) + 1L]] <- list(k = .k - 1L, w = .w, c = .cc)
    }
    if (.c %in% names(build$dLagStr)) {
      .kpost <- which(.denseT >= .tau - 1e-9)[1]
      .kpre  <- max(which(.denseT < .tau - 1e-9))
      .fp <- vapply(.st, function(i) .fwd[[paste0("rx__fRHS_", i, "__")]][.kpost], numeric(1))
      .fm <- vapply(.st, function(i) .fwd[[paste0("rx__fRHS_", i, "__")]][.kpre],  numeric(1))
      .cc <- vapply(build$dLagStr[[.c]], function(s) .rxAdjEvalNum(s, params), numeric(1))
      .duals[[length(.duals) + 1L]] <- list(k = .k - 1L, w = .fm - .fp, c = .cc)
    }
  }
  ## infusion moving-boundary dual
  for (.z in .infus) {
    .ci <- match(.z$cmt, .st)
    if (is.na(.ci)) next
    .k <- which.min(abs(.denseT - .z$t2))
    .w <- numeric(.ns); .w[.ci] <- .z$R
    .dTau2 <- -(.z$amt / .z$R^2) * .z$dRdp
    .duals[[length(.duals) + 1L]] <- list(k = .k - 1L, w = .w, c = .dTau2)
  }
  .dual <- if (length(.duals)) {
    list(as.integer(vapply(.duals, function(d) d$k, integer(1))),
         do.call(rbind, lapply(.duals, function(d) d$w)),
         do.call(rbind, lapply(.duals, function(d) d$c)))
  } else list(integer(0), numeric(0), numeric(0))

  .oidx <- match(outTimes, .denseT)
  .stateIdx <- match(adjStates, .st) - 1L
  .res <- .Call(`_rxode2_rxAdjointTrajSweep`, .denseT, .J, .dP,
               as.integer(.ns), as.integer(.np), as.integer(.oidx - 1L),
               as.integer(.stateIdx), .cj, .dual)
  ## res is flat nOut*nStates*np (column-major: o + s*nOut + p*nOut*nStates)
  .nOut <- length(outTimes)
  .out <- matrix(0.0, .nOut, .nk * .np)
  .cols <- character(.nk * .np)
  .col <- 1L
  for (.p in seq_len(.np)) for (.s in seq_len(.nk)) {
    .out[, .col] <- .res[(seq_len(.nOut) - 1L) + (.s - 1L) * .nOut + (.p - 1L) * .nOut * .nk + 1L]
    .cols[.col] <- paste0("rx__sens_", adjStates[.s], "_BY_", calcSens[.p], "__")
    .col <- .col + 1L
  }
  colnames(.out) <- .cols
  cbind(time = outTimes, as.data.frame(.out))
}

#' Adjoint gradient of an objective (single backward sweep)
#'
#' Computes `dG/dtheta` for every parameter with a single backward sweep.
#' Supports least squares (default,
#' `G = sum_i 1/2 weight_i (f_i - obs_i)^2`) or, when `errModel` is given, the
#' FOCEi -2 log-likelihood `G = sum_i r_i^2/v_i + log(v_i)` with
#' `v_i = add^2 + (prop*f_i)^2`.
#'
#' @inheritParams .rxAdjointSolve
#' @param pred character prediction expression `f` (function of states and
#'   parameters), e.g. `"center/v"`.
#' @param obsTimes numeric observation times.
#' @param obs numeric observed values aligned with `obsTimes`.
#' @param weight numeric scalar or vector of least-squares observation weights
#'   (ignored when `errModel` is supplied).
#' @param errModel `NULL` for least squares, or a list with character entries
#'   `add` and/or `prop` naming the additive / proportional residual-error
#'   parameters, selecting the FOCEi -2LL objective `v = add^2 + (prop*f)^2`.
#' @return named numeric vector `dG/dtheta` over `calcSens`.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjointGrad <- function(object, params, events, calcSens, pred, obsTimes, obs,
                           weight = 1, errModel = NULL,
                           denseBy = 0.01, atol = 1e-10, rtol = 1e-10) {
  .b <- .rxAdjointGradBuild(object, calcSens, pred, events, errModel)
  .rxAdjointGradEval(.b, params, obsTimes, obs, weight = weight,
                     denseBy = denseBy, atol = atol, rtol = rtol)
}

#' Adjoint gradient of a POPULATION objective (sum over subjects)
#'
#' Computes `dOFV/dtheta = sum_subjects dG_i/dtheta` with one backward sweep per
#' subject.  The symbolic model is built once ([.rxAdjointGradBuild()]) and
#' reused per subject.
#' Assumes a shared dosing regimen across subjects.
#'
#' @inheritParams .rxAdjointGrad
#' @param data data.frame of observations with columns `id`, `time`, and `dv`;
#'   one block of rows per subject.  Any additional columns are treated as
#'   per-subject covariates (first value per subject) merged into `params`.
#' @param useC use the C++ sweep ([.rxAdjointGradEvalC()]); otherwise the R
#'   reference ([.rxAdjointGradEval()]).
#' @return named numeric vector `dOFV/dtheta` over `calcSens`.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjointGradPop <- function(object, params, events, calcSens, pred, data,
                              errModel = NULL, denseBy = 0.01, useC = TRUE,
                              atol = 1e-10, rtol = 1e-10) {
  .build <- .rxAdjointGradBuild(object, calcSens, pred, events, errModel)
  .evalFn <- if (useC) .rxAdjointGradEvalC else .rxAdjointGradEval
  .covCols <- setdiff(names(data), c("id", "time", "dv"))
  .ids <- unique(data$id)
  .g <- stats::setNames(numeric(length(calcSens)), calcSens)
  for (.id in .ids) {
    .d <- data[data$id == .id, , drop = FALSE]
    .p <- params
    for (.cc in .covCols) .p[[.cc]] <- .d[[.cc]][1]   # per-subject covariates
    .g <- .g + .evalFn(.build, .p, .d$time, .d$dv, denseBy = denseBy,
                       atol = atol, rtol = rtol)
  }
  .g
}

#' Build the symbolic part of an adjoint objective gradient (compile once)
#'
#' Performs all symbolic work (symengine differentiation, `rxFromSE`, model
#' compilation) up front and returns a reusable object for
#' [.rxAdjointGradEval()], so an optimiser evaluates the gradient at many
#' parameter values without re-running symbolic code.  Build before any
#' `rxSolve` to avoid a `load_all`-only symengine dispatch fragility.
#'
#' @inheritParams .rxAdjointGrad
#' @return an opaque list consumed by [.rxAdjointGradEval()].
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjointGradBuild <- function(object, calcSens, pred, events, errModel = NULL) {
  .mText <- rxode2::rxNorm(rxode2::rxModelVars(object))
  .model <- rxode2::rxS(rxode2::rxGetModel(object), TRUE, promoteLinSens = FALSE)
  .st <- rxode2::rxStateOde(.model)
  if (length(.st) == 0L) {
    stop("adjoint gradient requires a model with ODE states", call. = FALSE)
  }
  invisible(rxode2::.rxJacobian(.model, c(.st, calcSens)))
  ## symbolic prediction, dh/dy_i and dh/dtheta_p (eval in the symengine env
  ## directly; with() would ignore enclos and mis-resolve `pred`).
  .ph <- eval(parse(text = pred), envir = .model)
  .predC <- rxode2::rxFromSE(.ph)
  .dhdy <- lapply(.st, function(i) { .d <- symengine::D(.ph, i); rxode2::rxFromSE(.d) })
  names(.dhdy) <- .st
  .dhdp <- lapply(calcSens, function(p) { .d <- symengine::D(.ph, p); rxode2::rxFromSE(.d) })
  names(.dhdp) <- calcSens

  ## event detection (needed before the backward model is built)
  .ev <- as.data.frame(events)
  .dr <- which(!is.na(.ev$evid) & .ev$evid == 1L & !is.na(.ev$amt) & .ev$amt != 0)
  .rateCol <- if ("rate" %in% names(.ev)) .ev$rate[.dr] else rep(0, length(.dr))
  .rateCol[is.na(.rateCol)] <- 0
  ## modeled infusions: capture symbolic dR/dtheta (rate flag -1) or dD/dtheta
  ## (flag -2) strings; numeric evaluation deferred to eval.
  .infusSym <- list()
  for (.k in which(.rateCol %in% c(-1, -2))) {
    .c <- as.character(.ev$cmt[.dr][.k])
    if (!(.c %in% .st)) next
    .t1 <- .ev$time[.dr][.k]; .amt <- .ev$amt[.dr][.k]
    if (.rateCol[.k] == -1) {
      .rSE <- get0(paste0("rx_rate_", .c, "_"), envir = .model, inherits = FALSE)
      if (is.null(.rSE)) next
      .valStr <- rxode2::rxFromSE(.rSE)
      .dStr <- vapply(calcSens, function(p) {
        .d <- symengine::D(.rSE, p); rxode2::rxFromSE(.d) }, character(1))
      .kind <- "rate"
    } else {
      .dSE <- get0(paste0("rx_dur_", .c, "_"), envir = .model, inherits = FALSE)
      if (is.null(.dSE)) next
      .valStr <- rxode2::rxFromSE(.dSE)
      .dStr <- vapply(calcSens, function(p) {
        .d <- symengine::D(.dSE, p); rxode2::rxFromSE(.d) }, character(1))
      .kind <- "dur"
    }
    .infusSym[[length(.infusSym) + 1L]] <- list(cmt = .c, t1 = .t1, amt = .amt,
                                                kind = .kind, valStr = .valStr, dStr = .dStr)
  }
  .infCmts <- unique(vapply(.infusSym, function(z) z$cmt, character(1)))

  ## reversed-time backward model (lambda + mu blocks); df/dy inlined as lhs over
  ## the covariate forward states.  mu quadrature adds the infusion forcing
  ## +lambda_c*dR/dtheta over the window (rxInfF_<cmt>_<p>__ covariate, 0 else).
  .lam <- function(i) paste0("rx__adjLam_", i, "__")
  .jacLines <- unlist(lapply(.st, function(j) lapply(c(.st, calcSens), function(x) {
    .d <- get0(paste0("rx__df_", j, "_dy_", x, "__"), envir = .model, inherits = FALSE)
    paste0("rx__df_", j, "_dy_", x, "__=", rxode2::rxFromSE(.d))
  })))
  .lamLines <- vapply(.st, function(i)
    paste0("d/dt(", .lam(i), ")=",
           paste(vapply(.st, function(j) paste0("rx__df_", j, "_dy_", i, "__*", .lam(j)),
                        character(1)), collapse = "+")), character(1))
  .muLines <- vapply(calcSens, function(p) {
    .base <- paste(vapply(.st, function(j) paste0(.lam(j), "*rx__df_", j, "_dy_", p, "__"),
                          character(1)), collapse = "+")
    .inf <- if (length(.infCmts))
      paste0("+", paste(vapply(.infCmts, function(cc)
        paste0(.lam(cc), "*rxInfF_", cc, "_", p, "__"), character(1)), collapse = "+")) else ""
    paste0("d/dt(rx__adjMu_", p, "__)=", .base, .inf)
  }, character(1))
  .revMod <- rxode2::rxode2(paste(c(.jacLines, .lamLines, .muLines), collapse = "\n"))

  ## bolus dose-jump duals (F + modeled lag): capture symbolic dF/dtheta and
  ## d(alag)/dtheta strings; numeric evaluation deferred to eval.
  .doses <- data.frame(time = .ev$time[.dr], cmt = as.character(.ev$cmt[.dr]),
                       amt = .ev$amt[.dr], rate = .rateCol, stringsAsFactors = FALSE)
  .dFexpr <- list(); .dLagStr <- list(); .lagStr <- list()
  for (.c in unique(.doses$cmt[.doses$cmt %in% .st])) {
    .fSE <- get0(paste0("rx_f_", .c, "_"), envir = .model, inherits = FALSE)
    if (!is.null(.fSE)) .dFexpr[[.c]] <- vapply(calcSens, function(p) {
      .d <- symengine::D(.fSE, p); rxode2::rxFromSE(.d) }, character(1))
    .lSE <- get0(paste0("rx_lag_", .c, "_"), envir = .model, inherits = FALSE)
    if (!is.null(.lSE)) {
      .lagStr[[.c]] <- rxode2::rxFromSE(.lSE)
      .dLagStr[[.c]] <- vapply(calcSens, function(p) {
        .d <- symengine::D(.lSE, p); rxode2::rxFromSE(.d) }, character(1))
    }
  }
  ## replace(evid5)/multiply(evid6) costate jumps (replace -> 0; multiply -> *alpha)
  .evR <- .ev[!is.na(.ev$evid) & .ev$evid %in% c(5L, 6L), , drop = FALSE]
  .evR$cmt <- as.character(.evR$cmt)
  .fLhs <- unlist(lapply(names(.dFexpr), function(.c)
    vapply(calcSens, function(p)
      paste0("rx__dFdp_", .c, "_", p, "__=", .dFexpr[[.c]][[p]]), character(1))))
  ## RHS f_i outputs for the pre/post-dose f(y-)/f(y+) transversality term
  .needF <- length(.dLagStr) > 0L
  .fRHSlhs <- if (.needF) vapply(.st, function(i) {
    .d <- get0(paste0("rx__d_dt_", i, "__"), envir = .model, inherits = FALSE)
    paste0("rx__fRHS_", i, "__=", rxode2::rxFromSE(.d)) }, character(1)) else character(0)

  ## forward checkpoint model (predictions, dh/dy, dh/dtheta, dF/dtheta, RHS f_i)
  .fmod <- rxode2::rxode2(paste(c(.mText,
    paste0("rx__pred__=", .predC),
    vapply(.st, function(i) paste0("rx__dhdy_", i, "__=", .dhdy[[i]]), character(1)),
    vapply(calcSens, function(p) paste0("rx__dhdp_", p, "__=", .dhdp[[p]]), character(1)),
    .fLhs, .fRHSlhs), collapse = "\n"))

  ## forward model that ALSO emits the Jacobian J (df_i/dy_j), forcing
  ## (df_i/dtheta_p), the dF/dtheta entries and the RHS f_i as outputs, so a
  ## single solve feeds the C++ sweep (incl. the dose-dual terms).
  .cfmod <- rxode2::rxode2(paste(c(.mText, .jacLines,
    paste0("rx__pred__=", .predC),
    vapply(.st, function(i) paste0("rx__dhdy_", i, "__=", .dhdy[[i]]), character(1)),
    vapply(calcSens, function(p) paste0("rx__dhdp_", p, "__=", .dhdp[[p]]), character(1)),
    .fLhs, .fRHSlhs), collapse = "\n"))

  list(st = .st, calcSens = calcSens, errModel = errModel, fmod = .fmod,
       revMod = .revMod, cfmod = .cfmod, infCmts = .infCmts, infusSym = .infusSym,
       doses = .doses, evR = .evR, dFexpr = .dFexpr, dLagStr = .dLagStr,
       lagStr = .lagStr, needF = .needF, events = events)
}

## numeric evaluation of an rxFromSE expression string at parameter values
.rxAdjEvalNum <- function(txt, params) {
  .env <- list2env(as.list(params), parent = baseenv())
  assign("Rx_pow_di", function(x, i) x^i, envir = .env)
  assign("Rx_pow", function(x, y) x^y, envir = .env)
  assign("expit", function(x, a = 0, b = 1) a + (b - a) / (1 + exp(-x)), envir = .env)
  assign("logit", function(x, a = 0, b = 1) log((x - a) / (b - x)), envir = .env)
  tryCatch(eval(parse(text = txt), envir = .env), error = function(e) NA_real_)
}

#' Evaluate a prebuilt adjoint objective gradient at given parameters
#'
#' @param build object returned by [.rxAdjointGradBuild()].
#' @inheritParams .rxAdjointGrad
#' @return named numeric vector `dG/dtheta` over the build's `calcSens`.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjointGradEval <- function(build, params, obsTimes, obs, weight = 1,
                               denseBy = 0.01, atol = 1e-10, rtol = 1e-10) {
  .st <- build$st; calcSens <- build$calcSens; errModel <- build$errModel
  .fmod <- build$fmod; .revMod <- build$revMod; .infCmts <- build$infCmts
  .evR <- build$evR; .dFexpr <- build$dFexpr
  .lam <- function(i) paste0("rx__adjLam_", i, "__")

  ## numeric infusion quantities from the cached symbolic strings
  .infus <- lapply(build$infusSym, function(z) {
    if (z$kind == "rate") {
      .Rv <- .rxAdjEvalNum(z$valStr, params)
      .dRp <- vapply(z$dStr, function(s) .rxAdjEvalNum(s, params), numeric(1))
    } else {
      .Dv <- .rxAdjEvalNum(z$valStr, params)
      .dDp <- vapply(z$dStr, function(s) .rxAdjEvalNum(s, params), numeric(1))
      .Rv <- z$amt / .Dv; .dRp <- -(z$amt / .Dv^2) * .dDp
    }
    list(cmt = z$cmt, t1 = z$t1, t2 = z$t1 + z$amt / .Rv, R = .Rv,
         dRdp = stats::setNames(.dRp, calcSens), amt = z$amt)
  })
  ## numeric lag values / d(alag)/dtheta from cached strings
  .lagVal <- stats::setNames(numeric(0), character(0)); .dLag <- list()
  for (.c in names(build$lagStr)) {
    .lagVal[.c] <- .rxAdjEvalNum(build$lagStr[[.c]], params)
    .dLag[[.c]] <- stats::setNames(
      vapply(build$dLagStr[[.c]], function(s) .rxAdjEvalNum(s, params), numeric(1)), calcSens)
  }
  .doses <- build$doses
  .doses$tau <- .doses$time + vapply(seq_len(nrow(.doses)), function(i) {
    .lv <- .lagVal[.doses$cmt[i]]; if (length(.lv) && !is.na(.lv)) .lv else 0 }, numeric(1))

  ## forward checkpoint + predictions/residuals at observations
  .infT <- unlist(lapply(.infus, function(z) c(z$t1, z$t2)))
  .denseT <- sort(unique(c(seq(0, max(obsTimes), by = denseBy), obsTimes,
                           .doses$tau, .doses$tau - denseBy,
                           .evR$time, .evR$time - denseBy, .infT)))
  .denseT <- .denseT[.denseT >= 0]
  .fev <- build$events |> rxode2::et(.denseT)
  .fwd <- as.data.frame(rxode2::rxSolve(.fmod, params = params, .fev,
                                        returnType = "data.frame", atol = atol,
                                        rtol = rtol, addDosing = FALSE))
  .Y <- lapply(.st, function(s) stats::approx(.fwd$time, .fwd[[s]], .denseT, rule = 2)$y)
  names(.Y) <- .st
  .atT <- function(col, t) .fwd[[col]][which.min(abs(.fwd$time - t))]
  .atObs <- function(col) vapply(obsTimes, function(t) .atT(col, t), numeric(1))
  .fobs  <- .atObs("rx__pred__")
  .resid <- .fobs - obs
  ## per-observation dG/df_i and (for -2LL) residual-error-parameter gradients
  .errGrad <- stats::setNames(numeric(0), character(0))
  if (is.null(errModel)) {
    .dLdf <- rep_len(weight, length(obsTimes)) * .resid       # d(1/2 w r^2)/df
  } else {
    .addV  <- if (!is.null(errModel$add))  params[[errModel$add]]  else 0
    .propV <- if (!is.null(errModel$prop)) params[[errModel$prop]] else 0
    .var    <- .addV^2 + (.propV * .fobs)^2                   # residual variance
    .dLdvar <- 1 / .var - .resid^2 / .var^2                   # dG/dv, G = r^2/v + log v
    .dvardf <- 2 * .propV^2 * .fobs                           # dv/df (proportional)
    .dLdf   <- 2 * .resid / .var + .dLdvar * .dvardf          # dG/df_i
    if (!is.null(errModel$add))
      .errGrad[errModel$add]  <- sum(.dLdvar * 2 * .addV)     # dv/d(add)  = 2 add
    if (!is.null(errModel$prop))
      .errGrad[errModel$prop] <- sum(.dLdvar * 2 * .propV * .fobs^2)  # dv/d(prop)
  }
  .cvec <- lapply(seq_along(obsTimes), function(i)
    vapply(.st, function(s) .dLdf[i] * .atObs(paste0("rx__dhdy_", s, "__"))[i],
           numeric(1)))

  ## pre/post-dose RHS at each actual dose time (for the lag transversality)
  .fPost <- function(t) vapply(.st, function(i)
    .fwd[[paste0("rx__fRHS_", i, "__")]][which(.fwd$time >= t - 1e-9)[1]], numeric(1))
  .fPre  <- function(t) { .k <- max(which(.fwd$time < t - 1e-9))
    vapply(.st, function(i) .fwd[[paste0("rx__fRHS_", i, "__")]][.k], numeric(1)) }

  ## single backward sweep, descending piecewise between breakpoints
  ## (observations and dose times): capture dose duals at lambda(tau+), then
  ## apply the observation lambda jump, then integrate down.
  .breaks <- sort(unique(c(obsTimes, .doses$tau, .evR$time, .infT, 0)), decreasing = TRUE)
  .breaks <- .breaks[.breaks <= max(obsTimes) + 1e-12]
  .state <- numeric(0)
  for (i in .st) .state[.lam(i)] <- 0
  for (p in calcSens) .state[paste0("rx__adjMu_", p, "__")] <- 0
  .gDose <- stats::setNames(numeric(length(calcSens)), calcSens)
  for (.bi in seq_along(.breaks)) {
    .tau <- .breaks[.bi]
    ## bolus dose duals (skip infusions: rate != 0 handled separately)
    .dHere <- which(abs(.doses$tau - .tau) < 1e-9 & .doses$rate == 0)
    for (.j in .dHere) {
      .c <- .doses$cmt[.j]
      ## bioavailability (F) dual
      if (.c %in% names(.dFexpr)) {
        .lamc <- .state[.lam(.c)]
        for (p in calcSens)
          .gDose[p] <- .gDose[p] + .lamc * .doses$amt[.j] *
            .atT(paste0("rx__dFdp_", .c, "_", p, "__"), .tau)
      }
      ## modeled-lag transversality dual
      if (.c %in% names(.dLag)) {
        .lamVec <- vapply(.st, function(i) .state[.lam(i)], numeric(1))
        .cross <- sum(.lamVec * (.fPre(.tau) - .fPost(.tau)))
        for (p in calcSens) .gDose[p] <- .gDose[p] + .cross * .dLag[[.c]][[p]]
      }
    }
    ## modeled-rate infusion moving-boundary term at tau2 = tau1 + amt/R
    for (.z in .infus) if (abs(.z$t2 - .tau) < 1e-9) {
      .lamc <- .state[.lam(.z$cmt)]
      .dTau2 <- -(.z$amt / .z$R^2) * .z$dRdp
      for (p in calcSens) .gDose[p] <- .gDose[p] + .z$R * .lamc * .dTau2[[p]]
    }
    ## replace/multiply costate jumps (lambda(tau+) -> lambda(tau-))
    for (.er in which(abs(.evR$time - .tau) < 1e-9)) {
      .cc <- .evR$cmt[.er]
      if (!(.cc %in% .st)) next
      if (.evR$evid[.er] == 5L) .state[.lam(.cc)] <- 0
      else if (.evR$evid[.er] == 6L) .state[.lam(.cc)] <- .evR$amt[.er] * .state[.lam(.cc)]
    }
    for (.o in which(abs(obsTimes - .tau) < 1e-9)) {
      .ci <- .cvec[[.o]]
      for (i in .st) .state[.lam(i)] <- .state[.lam(i)] + .ci[[i]]
    }
    if (.bi < length(.breaks)) {
      .tLo <- .breaks[.bi + 1]
      .idx <- which(.denseT >= .tLo - 1e-12 & .denseT <= .tau + 1e-12)
      .dat <- data.frame(time = .tau - rev(.denseT[.idx]), evid = 0L)
      for (.sn in .st) .dat[[.sn]] <- rev(.Y[[.sn]][.idx])
      ## infusion forcing covariate: dR/dtheta_p on [t1,t2], 0 outside
      if (length(.infCmts)) {
        .fwdt <- rev(.denseT[.idx])
        for (p in calcSens) for (.cc in .infCmts) {
          .val <- numeric(length(.fwdt))
          for (.z in .infus) if (.z$cmt == .cc) {
            .inw <- .fwdt >= .z$t1 - 1e-9 & .fwdt <= .z$t2 + 1e-9
            .val[.inw] <- .val[.inw] + .z$dRdp[[p]]
          }
          .dat[[paste0("rxInfF_", .cc, "_", p, "__")]] <- .val
        }
      }
      .sol <- as.data.frame(rxode2::rxSolve(.revMod, params = params, .dat,
                                            inits = .state, returnType = "data.frame",
                                            atol = atol, rtol = rtol,
                                            covsInterpolation = "linear", addDosing = FALSE))
      for (.nm in names(.state)) .state[.nm] <- .sol[[.nm]][nrow(.sol)]
    }
  }
  .gTraj <- vapply(calcSens, function(p) .state[[paste0("rx__adjMu_", p, "__")]], numeric(1))
  .gExpl <- vapply(calcSens, function(p)
    sum(.dLdf * .atObs(paste0("rx__dhdp_", p, "__"))), numeric(1))
  .gErr <- vapply(calcSens, function(p)
    if (p %in% names(.errGrad)) .errGrad[[p]] else 0, numeric(1))
  stats::setNames(.gTraj + .gExpl + .gDose[calcSens] + .gErr, calcSens)
}

#' Evaluate a prebuilt adjoint objective gradient in C++
#'
#' Covers continuous ODE, modeled bioavailability (F), modeled lag (alag),
#' replace/multiply events, and modeled-rate/duration infusions.  A single
#' forward solve emits the Jacobian, forcing and dose-dual quantities on a fine
#' grid; the backward costate + quadrature sweep, event jumps and gradient
#' accumulation run in C++ (`rxode2AdjointSweep`).  Numerically equivalent to
#' [.rxAdjointGradEval()].
#'
#' @inheritParams .rxAdjointGradEval
#' @return named numeric vector `dG/dtheta` over the build's `calcSens`.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjointGradEvalC <- function(build, params, obsTimes, obs, weight = 1,
                                denseBy = 0.01, atol = 1e-10, rtol = 1e-10) {
  ## all duals (F, modeled lag, replace/multiply, modeled rate/dur infusion) run
  ## through the C++ sweep.
  .st <- build$st; calcSens <- build$calcSens; errModel <- build$errModel
  .ns <- length(.st); .np <- length(calcSens)
  ## actual (possibly lagged) bolus dose times
  .doses <- build$doses
  .doses$tau <- .doses$time + vapply(seq_len(nrow(.doses)), function(i) {
    .c <- .doses$cmt[i]
    .lv <- if (.c %in% names(build$lagStr)) .rxAdjEvalNum(build$lagStr[[.c]], params) else 0
    if (length(.lv) && !is.na(.lv)) .lv else 0 }, numeric(1))
  ## numeric infusion quantities (R, dR/dtheta, [t1,t2]) from cached strings
  .infus <- lapply(build$infusSym, function(z) {
    if (z$kind == "rate") {
      .Rv <- .rxAdjEvalNum(z$valStr, params)
      .dRp <- vapply(z$dStr, function(s) .rxAdjEvalNum(s, params), numeric(1))
    } else {
      .Dv <- .rxAdjEvalNum(z$valStr, params)
      .dDp <- vapply(z$dStr, function(s) .rxAdjEvalNum(s, params), numeric(1))
      .Rv <- z$amt / .Dv; .dRp <- -(z$amt / .Dv^2) * .dDp
    }
    list(cmt = z$cmt, t1 = z$t1, t2 = z$t1 + z$amt / .Rv, R = .Rv,
         dRdp = stats::setNames(.dRp, calcSens), amt = z$amt)
  })
  .infT <- unlist(lapply(.infus, function(z) c(z$t1, z$t2)))
  .denseT <- sort(unique(c(seq(0, max(obsTimes), by = denseBy), obsTimes,
                           build$evR$time, .doses$tau, .doses$tau - denseBy, .infT)))
  .denseT <- .denseT[.denseT >= 0]
  .fev <- build$events |> rxode2::et(.denseT)
  .fwd <- as.data.frame(rxode2::rxSolve(build$cfmod, params = params, .fev,
                                        returnType = "data.frame", atol = atol,
                                        rtol = rtol, addDosing = FALSE))
  .fwd <- .fwd[match(.denseT, .fwd$time), ]
  .nt <- length(.denseT)
  ## J(k, i*ns+j) = df_i/dy_j ; dP(k, i*np+p) = df_i/dtheta_p
  .J <- matrix(0.0, .nt, .ns * .ns)
  for (i in seq_len(.ns)) for (j in seq_len(.ns))
    .J[, (i - 1) * .ns + j] <- .fwd[[paste0("rx__df_", .st[i], "_dy_", .st[j], "__")]]
  .dP <- matrix(0.0, .nt, .ns * .np)
  for (i in seq_len(.ns)) for (p in seq_len(.np))
    .dP[, (i - 1) * .np + p] <- .fwd[[paste0("rx__df_", .st[i], "_dy_", calcSens[p], "__")]]
  ## modeled-rate/dur infusion continuous forcing: add dR/dtheta_p to the
  ## infused state's forcing row over [t1,t2].
  for (.z in .infus) {
    .ci <- match(.z$cmt, .st)
    if (is.na(.ci)) next
    .inWin <- .denseT >= .z$t1 - 1e-9 & .denseT <= .z$t2 + 1e-9
    for (p in seq_len(.np))
      .dP[.inWin, (.ci - 1) * .np + p] <- .dP[.inWin, (.ci - 1) * .np + p] + .z$dRdp[[p]]
  }
  ## per-observation objective sensitivity dG/df_i and error-param gradients
  .oidx <- match(obsTimes, .denseT)
  .fobs <- .fwd[["rx__pred__"]][.oidx]
  .resid <- .fobs - obs
  .errGrad <- stats::setNames(numeric(0), character(0))
  if (is.null(errModel)) {
    .dLdf <- rep_len(weight, length(obsTimes)) * .resid
  } else {
    .addV  <- if (!is.null(errModel$add))  params[[errModel$add]]  else 0
    .propV <- if (!is.null(errModel$prop)) params[[errModel$prop]] else 0
    .var <- .addV^2 + (.propV * .fobs)^2
    .dLdvar <- 1 / .var - .resid^2 / .var^2
    .dLdf <- 2 * .resid / .var + .dLdvar * (2 * .propV^2 * .fobs)
    if (!is.null(errModel$add))  .errGrad[errModel$add]  <- sum(.dLdvar * 2 * .addV)
    if (!is.null(errModel$prop)) .errGrad[errModel$prop] <- sum(.dLdvar * 2 * .propV * .fobs^2)
  }
  ## observation covectors: cover(o, i) = dG/df_o * df_o/dy_i
  .cover <- matrix(0.0, length(obsTimes), .ns)
  for (i in seq_len(.ns))
    .cover[, i] <- .dLdf * .fwd[[paste0("rx__dhdy_", .st[i], "__")]][.oidx]
  ## replace(evid5)/multiply(evid6) costate jumps: lambda_c *= alpha (0 / mult)
  .evR <- build$evR
  .cj <- if (nrow(.evR) > 0L) {
    .cmt0 <- match(.evR$cmt, .st) - 1L
    list(as.integer(match(.evR$time, .denseT) - 1L), as.integer(.cmt0),
         ifelse(.evR$evid == 5L, 0.0, .evR$amt))
  } else list(integer(0), integer(0), numeric(0))
  ## bolus dose duals (F and modeled lag), out[p] += (lambda . w) * c
  .duals <- list()
  for (.j in which(.doses$rate == 0)) {
    .c <- .doses$cmt[.j]; .ci <- match(.c, .st); .tau <- .doses$tau[.j]
    if (is.na(.ci)) next
    .k <- which.min(abs(.denseT - .tau))
    if (.c %in% names(build$dFexpr)) {
      .w <- numeric(.ns); .w[.ci] <- .doses$amt[.j]
      .cc <- vapply(calcSens, function(p) .fwd[[paste0("rx__dFdp_", .c, "_", p, "__")]][.k],
                    numeric(1))
      .duals[[length(.duals) + 1L]] <- list(k = .k - 1L, w = .w, c = .cc)
    }
    if (.c %in% names(build$dLagStr)) {
      .kpost <- which(.denseT >= .tau - 1e-9)[1]
      .kpre  <- max(which(.denseT < .tau - 1e-9))
      .fp <- vapply(.st, function(i) .fwd[[paste0("rx__fRHS_", i, "__")]][.kpost], numeric(1))
      .fm <- vapply(.st, function(i) .fwd[[paste0("rx__fRHS_", i, "__")]][.kpre],  numeric(1))
      .cc <- vapply(build$dLagStr[[.c]], function(s) .rxAdjEvalNum(s, params), numeric(1))
      .duals[[length(.duals) + 1L]] <- list(k = .k - 1L, w = .fm - .fp, c = .cc)
    }
  }
  ## modeled-rate/dur infusion moving-boundary dual at tau2 = t1 + amt/R
  for (.z in .infus) {
    .ci <- match(.z$cmt, .st)
    if (is.na(.ci)) next
    .k <- which.min(abs(.denseT - .z$t2))
    .w <- numeric(.ns); .w[.ci] <- .z$R
    .dTau2 <- -(.z$amt / .z$R^2) * .z$dRdp
    .duals[[length(.duals) + 1L]] <- list(k = .k - 1L, w = .w, c = .dTau2)
  }
  .dual <- if (length(.duals)) {
    list(as.integer(vapply(.duals, function(d) d$k, integer(1))),
         do.call(rbind, lapply(.duals, function(d) d$w)),
         do.call(rbind, lapply(.duals, function(d) d$c)))
  } else list(integer(0), numeric(0), numeric(0))
  .gTraj <- .Call(`_rxode2_rxAdjointSweep`, .denseT, .J, .dP, .cover,
                  as.integer(.oidx - 1L), as.integer(.ns), as.integer(.np),
                  .cj, .dual)
  .gExpl <- vapply(calcSens, function(p)
    sum(.dLdf * .fwd[[paste0("rx__dhdp_", p, "__")]][.oidx]), numeric(1))
  .gErr <- vapply(calcSens, function(p)
    if (p %in% names(.errGrad)) .errGrad[[p]] else 0, numeric(1))
  stats::setNames(.gTraj + .gExpl + .gErr, calcSens)
}

rxTest({
  # Adjoint (backward) sensitivity analysis.
  #
  # .rxAdjoint() symbolically generates the continuous-adjoint ODE system
  #   costate:    d/dt(lambda_k)   = -J^T lambda_k
  #   quadrature: d/dt(sens_k_p)   = -lambda_k^T df/dp
  # reusing the rx__df_*_dy_*__ elemental derivatives that .rxJacobian
  # materialises (no automatic differentiation).  These tests verify (a) the
  # generated equations have the correct transpose-Jacobian / quadrature
  # structure and (b) that integrating them backward reproduces dy_k(T)/dp to
  # the same tolerance as the forward-sensitivity path and a finite difference.

  .nm <- function(state, p) paste0("rx__sens_", state, "_BY_", p, "__")

  mText <- "d/dt(depot)  = -ka*depot\nd/dt(center) =  ka*depot - (cl/v)*center"
  vars  <- c("ka", "cl", "v")
  p0    <- c(ka = 1.1, cl = 3.0, v = 20.0)
  Tfin  <- 12
  outState <- "center"

  model <- rxode2::rxS(rxode2::rxGetModel(mText), TRUE, promoteLinSens = FALSE)
  st    <- rxode2::rxStateOde(model)
  invisible(rxode2::.rxJacobian(model, c(st, vars)))
  s1    <- rxode2::.rxSens(model, vars)
  adj   <- rxode2::.rxAdjoint(model, vars, outState)

  test_that(".rxAdjoint emits one costate per (out-state,state) and one quadrature per param", {
    lam <- grep("^d/dt\\(rx__adjLambda_", adj, value = TRUE)
    quad <- grep("^d/dt\\(rx__sens_", adj, value = TRUE)
    expect_equal(length(lam), length(st))          # one lambda per state
    expect_equal(length(quad), length(vars))        # one dy/dp per param
    # quadrature reuses the forward-sensitivity output names exactly
    expect_true(all(vapply(vars, function(p)
      any(grepl(.nm(outState, p), quad, fixed = TRUE)), logical(1))))
    # costate for a decaying central compartment is +cl/v*lambda (transpose sign)
    expect_true(any(grepl("rx__adjLambda_center_center__)=cl*rx__adjLambda_center_center__/v",
                          adj, fixed = TRUE)))
    expect_null(NULL)
  })

  # ---- forward-sensitivity + finite-difference references --------------------
  fwdMod <- rxode2::rxode2(paste(c(mText, s1), collapse = "\n"))
  ev     <- rxode2::et(amt = 100, cmt = "depot") |> rxode2::et(Tfin)
  fwdRow <- as.data.frame(rxode2::rxSolve(fwdMod, params = p0, ev,
                                          returnType = "data.frame",
                                          atol = 1e-12, rtol = 1e-12))
  fwdRow <- fwdRow[fwdRow$time == Tfin, ]
  yT     <- c(depot = fwdRow$depot, center = fwdRow$center)
  refFwd <- vapply(vars, function(p) fwdRow[[.nm(outState, p)]], numeric(1))

  solveY <- function(p) {
    d <- as.data.frame(rxode2::rxSolve(rxode2::rxode2(mText), params = p, ev,
                                       returnType = "data.frame",
                                       atol = 1e-12, rtol = 1e-12))
    d[d$time == Tfin, outState]
  }
  refFD <- vapply(vars, function(p) {
    h <- p0[[p]] * 1e-5
    pp <- p0; pm <- p0; pp[p] <- pp[p] + h; pm[p] <- pm[p] - h
    (solveY(pp) - solveY(pm)) / (2 * h)
  }, numeric(1))

  # ---- adjoint via reverse-time (s = Tfin - t) reconstruction ----------------
  # dy/ds = -f(y); costate/quadrature flip sign from the backward-in-t form.
  negLine <- function(line) {
    m <- regmatches(line, regexec("^\\s*(d/dt\\([^)]*\\))\\s*=\\s*(.*)$", line))[[1]]
    paste0(m[2], "=-(", m[3], ")")
  }
  revPrimal <- vapply(strsplit(mText, "\n")[[1]], negLine, character(1),
                      USE.NAMES = FALSE)
  revAdj    <- vapply(adj, negLine, character(1), USE.NAMES = FALSE)
  revMod    <- rxode2::rxode2(paste(c(revPrimal, revAdj), collapse = "\n"))

  inits <- c(depot = yT[["depot"]], center = yT[["center"]])
  for (i in st) inits[paste0("rx__adjLambda_", outState, "_", i, "__")] <-
                  as.numeric(i == outState)
  for (p in vars) inits[.nm(outState, p)] <- 0
  revEv <- rxode2::et(seq(0, Tfin, length.out = 2001))
  revEnd <- as.data.frame(rxode2::rxSolve(revMod, params = p0, revEv,
                                          inits = inits, returnType = "data.frame",
                                          atol = 1e-12, rtol = 1e-12))
  revEnd  <- revEnd[nrow(revEnd), ]
  adjSens <- vapply(vars, function(p) revEnd[[.nm(outState, p)]], numeric(1))

  test_that("adjoint sensitivities match the forward-sensitivity path", {
    expect_equal(unname(adjSens), unname(refFwd), tolerance = 1e-6)
  })

  test_that("adjoint sensitivities match a central finite difference", {
    expect_equal(unname(adjSens), unname(refFD), tolerance = 1e-5)
  })

  # ---- full-trajectory adjoint (.rxAdjointSolve) over a long window ----------
  # Robustness: a 24 h window with fast absorption would make reverse-primal
  # reconstruction blow up like exp(+||J||*T); the checkpoint-interpolation
  # path stays bounded and matches forward sensitivities at every output time.
  fullOutTimes <- seq(1, 24, by = 1)
  fullP  <- c(ka = 1.5, cl = 4.0, v = 30.0)
  fullEv <- rxode2::et(amt = 100, cmt = "depot")

  fmod <- rxode2::rxode2(paste(c(mText, s1), collapse = "\n"))
  fev  <- rxode2::et(amt = 100, cmt = "depot") |> rxode2::et(fullOutTimes)
  fref <- as.data.frame(rxode2::rxSolve(fmod, params = fullP, fev,
                                        returnType = "data.frame",
                                        atol = 1e-11, rtol = 1e-11,
                                        addDosing = FALSE))
  fref <- fref[fref$time %in% fullOutTimes, ]
  allCols <- unlist(lapply(st, function(k)
    vapply(vars, function(p) .nm(k, p), character(1))))

  relTrajErr <- function(denseBy) {
    adjT <- rxode2::.rxAdjointSolve(mText, fullP, fullEv, vars, fullOutTimes,
                                    denseBy = denseBy)
    max(vapply(allCols, function(c)
      max(abs(adjT[[c]] - fref[[c]]) / (abs(fref[[c]]) + 1e-6)), numeric(1)))
  }

  test_that("full-trajectory adjoint matches forward sens over a long window", {
    expect_lt(relTrajErr(0.01), 1e-4)   # bounded (no reverse-primal blow-up)
  })

  test_that("full-trajectory adjoint error converges as the grid refines", {
    e_coarse <- relTrajErr(0.04)
    e_fine   <- relTrajErr(0.01)
    # O(denseBy^2) covariate-interpolation error: refining 4x cuts it markedly
    expect_lt(e_fine, e_coarse)
    expect_lt(e_fine, 5e-5)
  })

  # ---- functional-gradient adjoint (.rxAdjointGrad): the genuine win ---------
  # ONE backward sweep yields dG/dtheta for ALL theta, where
  #   G = sum_i 1/2 * (h(y(t_i), theta) - obs_i)^2,  h = center/v.
  test_that("adjoint objective gradient matches a central finite difference", {
    gP   <- c(ka = 1.2, cl = 3.5, v = 25.0)
    gCS  <- c("ka", "cl", "v")
    gPred <- "center/v"
    gObsT <- c(0.5, 1, 2, 4, 6, 8, 12, 18, 24)
    gEv  <- rxode2::et(amt = 100, cmt = "depot")

    truthMod <- rxode2::rxode2(paste0(mText, "\ncp=", gPred))
    gFev <- gEv |> rxode2::et(gObsT)
    truth <- as.data.frame(rxode2::rxSolve(truthMod, params = gP, gFev,
                                           returnType = "data.frame", addDosing = FALSE))
    gObs <- truth$cp[truth$time %in% gObsT] * 1.1 + 0.05   # nonzero residuals

    objG <- function(p) {
      d <- as.data.frame(rxode2::rxSolve(truthMod, params = p, gFev,
                                         returnType = "data.frame", addDosing = FALSE))
      sum(0.5 * (d$cp[d$time %in% gObsT] - gObs)^2)
    }
    gFD <- vapply(gCS, function(p) {
      h <- gP[[p]] * 1e-6; pp <- gP; pm <- gP
      pp[p] <- pp[p] + h; pm[p] <- pm[p] - h
      (objG(pp) - objG(pm)) / (2 * h)
    }, numeric(1))

    gAdj <- rxode2::.rxAdjointGrad(mText, gP, gEv, gCS, gPred, gObsT, gObs)
    expect_equal(unname(gAdj), unname(gFD), tolerance = 1e-3)
  })

  # ---- F (bioavailability) dose-jump dual ------------------------------------
  # A parameter entering ONLY through f(depot)=Fbio gets its whole gradient from
  # the dose-jump term  lambda_depot(t0+)*amt*dF/dtheta.
  test_that("adjoint gradient handles bioavailability (F) dose-jump dual", {
    fText <- paste0(mText, "\nf(depot)=Fbio")
    fP    <- c(ka = 1.2, cl = 3.5, v = 25.0, Fbio = 0.7)
    fCS   <- c("ka", "cl", "v", "Fbio")
    fPred <- "center/v"
    fObsT <- c(0.5, 1, 2, 4, 6, 8, 12, 18, 24)
    fEv   <- rxode2::et(amt = 100, cmt = "depot")

    truthMod <- rxode2::rxode2(paste0(fText, "\ncp=", fPred))
    fFev <- fEv |> rxode2::et(fObsT)
    truth <- as.data.frame(rxode2::rxSolve(truthMod, params = fP, fFev,
                                           returnType = "data.frame", addDosing = FALSE))
    fObs <- truth$cp[truth$time %in% fObsT] * 1.1 + 0.05

    objG <- function(p) {
      d <- as.data.frame(rxode2::rxSolve(truthMod, params = p, fFev,
                                         returnType = "data.frame", addDosing = FALSE))
      sum(0.5 * (d$cp[d$time %in% fObsT] - fObs)^2)
    }
    gFD <- vapply(fCS, function(p) {
      h <- fP[[p]] * 1e-6; pp <- fP; pm <- fP
      pp[p] <- pp[p] + h; pm[p] <- pm[p] - h
      (objG(pp) - objG(pm)) / (2 * h)
    }, numeric(1))

    gAdj <- rxode2::.rxAdjointGrad(fText, fP, fEv, fCS, fPred, fObsT, fObs)
    # Fbio gradient is entirely from the dose term and clearly nonzero
    expect_gt(abs(gAdj[["Fbio"]]), 1)
    expect_equal(unname(gAdj), unname(gFD), tolerance = 1e-3)
    # F dose dual also runs through the C++ sweep
    Bf <- rxode2::.rxAdjointGradBuild(fText, fCS, fPred, fEv)
    gCf <- rxode2::.rxAdjointGradEvalC(Bf, fP, fObsT, fObs)
    gRf <- rxode2::.rxAdjointGradEval(Bf, fP, fObsT, fObs)
    expect_equal(unname(gCf), unname(gRf), tolerance = 1e-4)
  })

  # ---- modeled lag (alag) transversality dual --------------------------------
  # A parameter entering through alag(depot)=tlag gets its gradient from the
  # time-triggered transversality term at the (lagged) dose time.
  test_that("adjoint gradient handles modeled lag (alag) transversality dual", {
    lText <- paste0(mText, "\nalag(depot)=tlag")
    lP    <- c(ka = 1.2, cl = 3.5, v = 25.0, tlag = 0.8)
    lCS   <- c("ka", "cl", "v", "tlag")
    lPred <- "center/v"
    lObsT <- c(1, 2, 4, 6, 8, 12, 18, 24)
    lEv   <- rxode2::et(amt = 100, cmt = "depot")

    truthMod <- rxode2::rxode2(paste0(lText, "\ncp=", lPred))
    lFev <- lEv |> rxode2::et(lObsT)
    truth <- as.data.frame(rxode2::rxSolve(truthMod, params = lP, lFev,
                                           returnType = "data.frame", addDosing = FALSE))
    lObs <- truth$cp[truth$time %in% lObsT] * 1.1 + 0.05

    objG <- function(p) {
      d <- as.data.frame(rxode2::rxSolve(truthMod, params = p, lFev,
                                         returnType = "data.frame", addDosing = FALSE))
      sum(0.5 * (d$cp[d$time %in% lObsT] - lObs)^2)
    }
    gFD <- vapply(lCS, function(p) {
      h <- lP[[p]] * 1e-6; pp <- lP; pm <- lP
      pp[p] <- pp[p] + h; pm[p] <- pm[p] - h
      (objG(pp) - objG(pm)) / (2 * h)
    }, numeric(1))

    gAdj <- rxode2::.rxAdjointGrad(lText, lP, lEv, lCS, lPred, lObsT, lObs, denseBy = 0.005)
    # the tlag transversality term is exact (point values); check it tightly
    expect_equal(gAdj[["tlag"]], gFD[["tlag"]], tolerance = 1e-3)
    # structural params limited by covariate interpolation of the sharp peak
    expect_equal(unname(gAdj), unname(gFD), tolerance = 1e-2)
    # lag transversality dual also runs through the C++ sweep
    Bl <- rxode2::.rxAdjointGradBuild(lText, lCS, lPred, lEv)
    gCl <- rxode2::.rxAdjointGradEvalC(Bl, lP, lObsT, lObs, denseBy = 0.005)
    gRl <- rxode2::.rxAdjointGradEval(Bl, lP, lObsT, lObs, denseBy = 0.005)
    expect_equal(unname(gCl), unname(gRl), tolerance = 1e-4)
  })

  # ---- replace(evid5) / multiply(evid6) costate jumps ------------------------
  # The costate jump is essential for correct structural-param gradients even
  # with a constant replace value (resetting/scaling lambda_c at the event).
  test_that("adjoint gradient handles replace/multiply costate jumps", {
    eP    <- c(ka = 1.2, cl = 3.5, v = 25.0)
    eCS   <- c("ka", "cl", "v")
    ePred <- "center/v"
    eObsT <- c(1, 2, 4, 6, 8, 12)
    tmod  <- rxode2::rxode2(paste0(mText, "\ncp=", ePred))

    chk <- function(eEv) {
      eFev <- eEv |> rxode2::et(eObsT)
      truth <- as.data.frame(rxode2::rxSolve(tmod, params = eP, eFev,
                                             returnType = "data.frame", addDosing = FALSE))
      eObs <- truth$cp[truth$time %in% eObsT] * 1.1 + 0.05
      objG <- function(p) {
        d <- as.data.frame(rxode2::rxSolve(tmod, params = p, eFev,
                                           returnType = "data.frame", addDosing = FALSE))
        sum(0.5 * (d$cp[d$time %in% eObsT] - eObs)^2)
      }
      gFD <- vapply(eCS, function(p) {
        h <- eP[[p]] * 1e-6; pp <- eP; pm <- eP
        pp[p] <- pp[p] + h; pm[p] <- pm[p] - h
        (objG(pp) - objG(pm)) / (2 * h)
      }, numeric(1))
      gAdj <- rxode2::.rxAdjointGrad(mText, eP, eEv, eCS, ePred, eObsT, eObs,
                                     denseBy = 0.005)
      expect_equal(unname(gAdj), unname(gFD), tolerance = 5e-3)
      # replace/multiply costate jumps also run through the C++ sweep
      B <- rxode2::.rxAdjointGradBuild(mText, eCS, ePred, eEv)
      gC <- rxode2::.rxAdjointGradEvalC(B, eP, eObsT, eObs, denseBy = 0.005)
      expect_equal(unname(gC), unname(gAdj), tolerance = 1e-4)
    }
    chk(rxode2::et(amt = 100, cmt = "depot") |>
          rxode2::et(time = 3, amt = 40, cmt = "center", evid = 5))   # replace
    chk(rxode2::et(amt = 100, cmt = "depot") |>
          rxode2::et(time = 3, amt = 0.5, cmt = "center", evid = 6))  # multiply
  })

  # ---- modeled-rate infusion continuous-forcing + moving-boundary dual -------
  # rate(depot)=Rin adds +R over [0, amt/R]; dG/dRin combines a continuous
  # forcing integral of lambda_depot and a moving-boundary term (amt/R moves).
  test_that("adjoint gradient handles modeled-rate infusion (forcing + boundary)", {
    iText <- paste0(mText, "\nrate(depot)=Rin")
    iP    <- c(ka = 1.2, cl = 3.5, v = 25.0, Rin = 50)
    iCS   <- c("ka", "cl", "v", "Rin")
    iPred <- "depot"                              # observe during infusion (strong signal)
    iObsT <- c(0.5, 1, 1.5, 2.5, 3, 4)
    iEv   <- rxode2::et(amt = 100, rate = -1, cmt = "depot")

    tmod <- rxode2::rxode2(paste0(iText, "\ncp=", iPred))
    iFev <- iEv |> rxode2::et(iObsT)
    truth <- as.data.frame(rxode2::rxSolve(tmod, params = iP, iFev,
                                           returnType = "data.frame", addDosing = FALSE))
    iObs <- truth$cp[truth$time %in% iObsT] * 1.1 + 0.5

    objG <- function(p) {
      d <- as.data.frame(rxode2::rxSolve(tmod, params = p, iFev,
                                         returnType = "data.frame", addDosing = FALSE))
      sum(0.5 * (d$cp[d$time %in% iObsT] - iObs)^2)
    }
    gFD <- vapply(iCS, function(p) {
      h <- iP[[p]] * 1e-5; pp <- iP; pm <- iP
      pp[p] <- pp[p] + h; pm[p] <- pm[p] - h
      (objG(pp) - objG(pm)) / (2 * h)
    }, numeric(1))

    gAdj <- rxode2::.rxAdjointGrad(iText, iP, iEv, iCS, iPred, iObsT, iObs, denseBy = 0.002)
    # Rin gradient (~ -3.2) is dominated by the infusion forcing+boundary dual
    expect_gt(abs(gAdj[["Rin"]]), 1)
    expect_equal(unname(gAdj), unname(gFD), tolerance = 5e-3)
    # infusion forcing + moving-boundary duals also run through the C++ sweep
    Bi <- rxode2::.rxAdjointGradBuild(iText, iCS, iPred, iEv)
    gCi <- rxode2::.rxAdjointGradEvalC(Bi, iP, iObsT, iObs, denseBy = 0.002)
    expect_equal(unname(gCi), unname(gAdj), tolerance = 1e-4)
  })

  # ---- modeled-duration infusion (rate = amt/D) ------------------------------
  test_that("adjoint gradient handles modeled-duration infusion", {
    dText <- paste0(mText, "\ndur(depot)=Dd")
    dP    <- c(ka = 1.2, cl = 3.5, v = 25.0, Dd = 2)
    dCS   <- c("ka", "cl", "v", "Dd")
    dPred <- "depot"
    dObsT <- c(0.5, 1, 1.5, 2.5, 3, 4)
    dEv   <- rxode2::et(amt = 100, rate = -2, cmt = "depot")

    tmod <- rxode2::rxode2(paste0(dText, "\ncp=", dPred))
    dFev <- dEv |> rxode2::et(dObsT)
    truth <- as.data.frame(rxode2::rxSolve(tmod, params = dP, dFev,
                                           returnType = "data.frame", addDosing = FALSE))
    dObs <- truth$cp[truth$time %in% dObsT] * 1.1 + 0.5

    objG <- function(p) {
      d <- as.data.frame(rxode2::rxSolve(tmod, params = p, dFev,
                                         returnType = "data.frame", addDosing = FALSE))
      sum(0.5 * (d$cp[d$time %in% dObsT] - dObs)^2)
    }
    gFD <- vapply(dCS, function(p) {
      h <- dP[[p]] * 1e-5; pp <- dP; pm <- dP
      pp[p] <- pp[p] + h; pm[p] <- pm[p] - h
      (objG(pp) - objG(pm)) / (2 * h)
    }, numeric(1))

    gAdj <- rxode2::.rxAdjointGrad(dText, dP, dEv, dCS, dPred, dObsT, dObs, denseBy = 0.002)
    expect_gt(abs(gAdj[["Dd"]]), 1)
    expect_equal(unname(gAdj), unname(gFD), tolerance = 5e-3)
    # modeled-dur duals also run through the C++ sweep
    Bd <- rxode2::.rxAdjointGradBuild(dText, dCS, dPred, dEv)
    gCd <- rxode2::.rxAdjointGradEvalC(Bd, dP, dObsT, dObs, denseBy = 0.002)
    expect_equal(unname(gCd), unname(gAdj), tolerance = 1e-4)
  })

  # ---- FOCEi -2LL objective gradient (structural + residual-error params) ----
  # The gradient FOCEi propagates: dG/dtheta of G = sum r^2/v + log(v),
  # v = add^2 + (prop*f)^2, over structural AND residual-error parameters, all
  # from ONE backward sweep.
  test_that("adjoint gradient of the FOCEi -2LL objective matches finite differences", {
    lP    <- c(ka = 1.2, cl = 3.5, v = 25.0, add = 0.3, prop = 0.1)
    lCS   <- c("ka", "cl", "v", "add", "prop")
    lPred <- "center/v"
    lObsT <- c(0.5, 1, 2, 4, 6, 8, 12, 18, 24)
    lEv   <- rxode2::et(amt = 100, cmt = "depot")
    tmod  <- rxode2::rxode2(paste0(mText, "\ncp=", lPred))
    lFev  <- lEv |> rxode2::et(lObsT)
    truth <- as.data.frame(rxode2::rxSolve(tmod, params = lP, lFev,
                                           returnType = "data.frame", addDosing = FALSE))
    set.seed(2)
    lObs <- truth$cp[truth$time %in% lObsT] * (1 + stats::rnorm(length(lObsT), 0, 0.1)) +
      stats::rnorm(length(lObsT), 0, 0.2)

    m2ll <- function(p) {
      d <- as.data.frame(rxode2::rxSolve(tmod, params = p, lFev,
                                         returnType = "data.frame", addDosing = FALSE))
      f <- d$cp[d$time %in% lObsT]; v <- p[["add"]]^2 + (p[["prop"]] * f)^2
      sum((f - lObs)^2 / v + log(v))
    }
    gFD <- vapply(lCS, function(p) {
      h <- abs(lP[[p]]) * 1e-6; pp <- lP; pm <- lP
      pp[p] <- pp[p] + h; pm[p] <- pm[p] - h
      (m2ll(pp) - m2ll(pm)) / (2 * h)
    }, numeric(1))
    gAdj <- rxode2::.rxAdjointGrad(mText, lP, lEv, lCS, lPred, lObsT, lObs,
                                   errModel = list(add = "add", prop = "prop"))
    expect_equal(unname(gAdj), unname(gFD), tolerance = 1e-3)
  })

  # ---- C++ eval (build once, evaluate many): matches R eval + FD -------------
  # The pure-numeric backward sweep runs in C++ (rxAdjointSweepC) for the
  # continuous case, matching the R eval to solver precision and the FOCEi -2LL
  # finite difference, with no symbolic work at evaluation time.
  test_that("C++ adjoint eval matches the R eval and finite differences", {
    cPred <- "center/v"
    cCS   <- c("ka", "cl", "v", "add", "prop")
    cObsT <- c(0.5, 1, 2, 4, 6, 8, 12, 18, 24)
    cEv   <- rxode2::et(amt = 100, cmt = "depot")
    B <- rxode2::.rxAdjointGradBuild(mText, cCS, cPred, cEv,
                                     errModel = list(add = "add", prop = "prop"))
    tmod <- rxode2::rxode2(paste0(mText, "\ncp=", cPred))
    cFev <- cEv |> rxode2::et(cObsT)
    for (cP in list(c(ka = 1.2, cl = 3.5, v = 25, add = 0.3, prop = 0.1),
                    c(ka = 0.8, cl = 5.0, v = 18, add = 0.5, prop = 0.05))) {
      truth <- as.data.frame(rxode2::rxSolve(tmod, params = cP, cFev,
                                             returnType = "data.frame", addDosing = FALSE))
      cObs <- truth$cp[truth$time %in% cObsT] * 1.15 + 0.1
      m2ll <- function(p) {
        d <- as.data.frame(rxode2::rxSolve(tmod, params = p, cFev,
                                           returnType = "data.frame", addDosing = FALSE))
        f <- d$cp[d$time %in% cObsT]; v <- p[["add"]]^2 + (p[["prop"]] * f)^2
        sum((f - cObs)^2 / v + log(v))
      }
      gFD <- vapply(cCS, function(p) {
        h <- abs(cP[[p]]) * 1e-6; pp <- cP; pm <- cP
        pp[p] <- pp[p] + h; pm[p] <- pm[p] - h
        (m2ll(pp) - m2ll(pm)) / (2 * h)
      }, numeric(1))
      gR <- rxode2::.rxAdjointGradEval(B, cP, cObsT, cObs, denseBy = 0.01)
      gC <- rxode2::.rxAdjointGradEvalC(B, cP, cObsT, cObs, denseBy = 0.01)
      expect_equal(unname(gC), unname(gR), tolerance = 1e-5)  # RK4 vs adaptive
      expect_equal(unname(gC), unname(gFD), tolerance = 2e-3)
    }
  })

  # ---- C++ full-trajectory sweep (rxode2AdjointTrajSweep / .rxAdjointSolveEvalC) --
  # dy_k(t_i)/dp for EVERY state of interest, output time, and param -- the
  # in-engine counterpart of forward sensitivity's full output structure.
  # IMPORTANT: this requires ONE independent backward sweep PER output time
  # (see the code comment in src/adjoint.cpp) -- it is NOT the O(1) win that
  # .rxAdjointGrad's scalar objective sweep is; it exists for output-structure
  # parity / validation, matching the plan's honest scoping note.
  # For a structurally-zero sensitivity (e.g. a state with no dependence on a
  # given param), the analytic answer is exactly 0 but a finite difference has
  # roundoff noise; the floor absorbs that noise before computing a relative
  # error so a genuinely tiny/zero true value does not blow up the ratio.
  relOrAbsErr <- function(cc, fd, absFloor = 0.02) {
    pmax(abs(cc - fd) - absFloor, 0) / (abs(fd) + absFloor)
  }

  test_that("C++ full-trajectory adjoint sweep matches the R reference (continuous)", {
    tCS <- c("ka", "cl", "v"); tP <- c(ka = 1.2, cl = 3.5, v = 25)
    tObsT <- c(1, 2, 4, 6, 8, 12, 18, 24)
    tEv <- rxode2::et(amt = 100, cmt = "depot")
    B <- rxode2::.rxAdjointSolveBuild(mText, tCS, tEv)
    resC <- rxode2::.rxAdjointSolveEvalC(B, tP, tObsT, denseBy = 0.01)
    resR <- rxode2::.rxAdjointSolve(mText, tP, tEv, tCS, tObsT, denseBy = 0.01)
    tCols <- setdiff(names(resC), "time")
    for (cc in tCols) {
      expect_lt(max(relOrAbsErr(resC[[cc]], resR[[cc]])), 1e-3)
    }
  })

  test_that("C++ full-trajectory adjoint sweep handles F and modeled-lag duals (vs FD)", {
    tOb <- c(1, 2, 4)
    tEv <- rxode2::et(amt = 100, cmt = "depot")
    chkTraj <- function(txt, cs, pp) {
      B <- rxode2::.rxAdjointSolveBuild(txt, cs, tEv)
      resC <- rxode2::.rxAdjointSolveEvalC(B, pp, tOb, denseBy = 0.005)
      tmod <- rxode2::rxode2(txt); tFev <- tEv |> rxode2::et(tOb)
      solveAt <- function(p) as.data.frame(rxode2::rxSolve(tmod, params = p, tFev,
                                                            returnType = "data.frame",
                                                            addDosing = FALSE))
      for (.stt in c("depot", "center")) for (.pp in cs) {
        h <- abs(pp[[.pp]]) * 1e-6; pp1 <- pp; pp2 <- pp
        pp1[.pp] <- pp1[.pp] + h; pp2[.pp] <- pp2[.pp] - h
        d1 <- solveAt(pp1); d2 <- solveAt(pp2)
        fd <- (d1[[.stt]] - d2[[.stt]]) / (2 * h)
        cc <- resC[[paste0("rx__sens_", .stt, "_BY_", .pp, "__")]]
        expect_lt(max(relOrAbsErr(cc, fd)), 2e-2)
      }
    }
    chkTraj(paste0(mText, "\nf(depot)=Fbio"), c("ka", "cl", "v", "Fbio"),
            c(ka = 1.2, cl = 3.5, v = 25, Fbio = 0.7))
    chkTraj(paste0(mText, "\nalag(depot)=tlag"), c("ka", "cl", "v", "tlag"),
            c(ka = 1.2, cl = 3.5, v = 25, tlag = 0.8))
  })

  # ---- rxSolveAdjoint(): drop-in rxSolve() wrapper with adjoint sens columns --
  # Same column names/structure as forward-sensitivity rxSolve(calcSens=), just
  # computed via the backward (adjoint) path -- output-structure parity.
  test_that("rxSolveAdjoint matches forward-sensitivity rxSolve output", {
    aP <- c(ka = 1.2, cl = 3.5, v = 25); aCS <- c("ka", "cl", "v")
    aEv <- rxode2::et(amt = 100, cmt = "depot") |> rxode2::et(c(1, 2, 4, 6, 8, 12))
    res <- rxode2::rxSolveAdjoint(mText, aP, aEv, aCS)

    model <- rxode2::rxS(rxode2::rxGetModel(mText), TRUE, promoteLinSens = FALSE)
    st <- rxode2::rxStateOde(model)
    invisible(rxode2::.rxJacobian(model, c(st, aCS)))
    s1 <- rxode2::.rxSens(model, aCS)
    fmod <- rxode2::rxode2(paste(c(mText, s1), collapse = "\n"))
    fref <- as.data.frame(rxode2::rxSolve(fmod, params = aP, aEv,
                                          returnType = "data.frame"))

    sensCols <- grep("^rx__sens_", names(fref), value = TRUE)
    expect_true(all(sensCols %in% names(res)))          # same column names
    expect_equal(res$depot, fref$depot, tolerance = 1e-4)     # same primal solve
    expect_equal(res$center, fref$center, tolerance = 1e-4)
    for (cc in sensCols) {
      expect_equal(res[[cc]], fref[[cc]], tolerance = 1e-3)   # same sens values
    }
  })

  # ---- population objective gradient (.rxAdjointGradPop) ---------------------
  # dOFV/dtheta = sum over subjects of each subject's adjoint -2LL gradient --
  # the exact quantity a gradient-based population estimator minimises, with one
  # backward sweep per subject regardless of the number of population parameters.
  test_that("adjoint population gradient matches a finite difference of the population -2LL", {
    pTh <- c(ka = 1.2, cl = 3.5, v = 25, add = 0.4, prop = 0.08)
    pCS <- c("ka", "cl", "v", "add", "prop")
    pErr <- list(add = "add", prop = "prop")
    pEv <- rxode2::et(amt = 100, cmt = "depot") |>
      rxode2::et(c(0.5, 1, 2, 4, 6, 8, 12, 18, 24))
    tmod <- rxode2::rxode2(paste0(mText, "\ncp=center/v"))
    set.seed(5); nsub <- 5
    truth <- as.data.frame(rxode2::rxSolve(tmod, params = pTh, pEv,
                                           returnType = "data.frame", addDosing = FALSE))
    data <- do.call(rbind, lapply(seq_len(nsub), function(i) {
      vv <- pTh[["add"]]^2 + (pTh[["prop"]] * truth$cp)^2
      data.frame(id = i, time = truth$time,
                 dv = truth$cp + stats::rnorm(length(truth$cp), 0, sqrt(vv)))
    }))
    m2ll <- function(th) {
      tot <- 0
      for (i in seq_len(nsub)) {
        di <- data[data$id == i, ]
        d <- as.data.frame(rxode2::rxSolve(tmod, params = th, pEv,
                                           returnType = "data.frame", addDosing = FALSE))
        f <- d$cp; vv <- th[["add"]]^2 + (th[["prop"]] * f)^2
        tot <- tot + sum((f - di$dv)^2 / vv + log(vv))
      }
      tot
    }
    gFD <- vapply(pCS, function(pn) {
      h <- abs(pTh[[pn]]) * 1e-6; p1 <- pTh; p2 <- pTh
      p1[pn] <- p1[pn] + h; p2[pn] <- p2[pn] - h
      (m2ll(p1) - m2ll(p2)) / (2 * h)
    }, numeric(1))
    gPop <- rxode2::.rxAdjointGradPop(mText, pTh, pEv, pCS, "center/v", data,
                                      pErr, denseBy = 0.005)
    expect_equal(unname(gPop), unname(gFD), tolerance = 5e-3)
  })

  test_that("adjoint population gradient handles per-subject covariates", {
    cText <- paste0("cl = exp(tcl + bwt*lwt)\n", "v  = exp(tv + bage*lage)\n", mText)
    cTh <- c(tcl = 1.2, tv = 3.2, ka = 1.1, bwt = 0.3, bage = -0.1, add = 0.4, prop = 0.08)
    cCS <- c("tcl", "tv", "ka", "bwt", "bage", "add", "prop")
    cErr <- list(add = "add", prop = "prop")
    cEv <- rxode2::et(amt = 100, cmt = "depot") |>
      rxode2::et(c(0.5, 1, 2, 4, 6, 8, 12, 18, 24))
    tmod <- rxode2::rxode2(paste0(cText, "\ncp=center/v"))
    set.seed(3); nsub <- 6
    covs <- data.frame(id = seq_len(nsub), lwt = stats::rnorm(nsub), lage = stats::rnorm(nsub))
    data <- do.call(rbind, lapply(seq_len(nsub), function(i) {
      p <- c(cTh, lwt = covs$lwt[i], lage = covs$lage[i])
      d <- as.data.frame(rxode2::rxSolve(tmod, params = p, cEv,
                                         returnType = "data.frame", addDosing = FALSE))
      vv <- p[["add"]]^2 + (p[["prop"]] * d$cp)^2
      data.frame(id = i, time = d$time,
                 dv = d$cp + stats::rnorm(length(d$cp), 0, sqrt(vv)),
                 lwt = covs$lwt[i], lage = covs$lage[i])
    }))
    m2ll <- function(th) {
      tot <- 0
      for (i in seq_len(nsub)) {
        di <- data[data$id == i, ]
        p <- c(th, lwt = di$lwt[1], lage = di$lage[1])
        d <- as.data.frame(rxode2::rxSolve(tmod, params = p, cEv,
                                           returnType = "data.frame", addDosing = FALSE))
        f <- d$cp; vv <- th[["add"]]^2 + (th[["prop"]] * f)^2
        tot <- tot + sum((f - di$dv)^2 / vv + log(vv))
      }
      tot
    }
    gFD <- vapply(cCS, function(pn) {
      h <- abs(cTh[[pn]]) * 1e-6; p1 <- cTh; p2 <- cTh
      p1[pn] <- p1[pn] + h; p2[pn] <- p2[pn] - h
      (m2ll(p1) - m2ll(p2)) / (2 * h)
    }, numeric(1))
    gPop <- rxode2::.rxAdjointGradPop(cText, cTh, cEv, cCS, "center/v", data,
                                      cErr, denseBy = 0.005)
    expect_equal(unname(gPop), unname(gFD), tolerance = 1e-2)
  })

  # ---- capstone: adjoint gradient drives a gradient-based fit (nlm-style) -----
  # Proves the functional-gradient adjoint is usable as the ONLY gradient source
  # for a BFGS optimisation that recovers the data-generating parameters.
  test_that("adjoint objective gradient drives a BFGS fit to recover parameters", {
    truePar <- c(ka = 1.2, cl = 3.5, v = 25.0)
    oPred <- "center/v"
    oObsT <- c(0.5, 1, 2, 4, 6, 8, 12, 18, 24)
    oEv   <- rxode2::et(amt = 100, cmt = "depot")
    tmod  <- rxode2::rxode2(paste0(mText, "\ncp=", oPred))
    oFev  <- oEv |> rxode2::et(oObsT)
    set.seed(1)
    truth <- as.data.frame(rxode2::rxSolve(tmod, params = truePar, oFev,
                                           returnType = "data.frame", addDosing = FALSE))
    oObs  <- truth$cp[truth$time %in% oObsT] * (1 + stats::rnorm(length(oObsT), 0, 0.02))

    obj <- function(lp) {
      p <- stats::setNames(exp(lp), c("ka", "cl", "v"))
      d <- as.data.frame(rxode2::rxSolve(tmod, params = p, oFev,
                                         returnType = "data.frame", addDosing = FALSE))
      sum(0.5 * (d$cp[d$time %in% oObsT] - oObs)^2)
    }
    gr <- function(lp) {
      p <- stats::setNames(exp(lp), c("ka", "cl", "v"))
      rxode2::.rxAdjointGrad(mText, p, oEv, c("ka", "cl", "v"), oPred, oObsT, oObs,
                             denseBy = 0.02) * p   # chain rule for log-params
    }
    fit <- stats::optim(log(c(2, 5, 15)), obj, gr, method = "BFGS",
                        control = list(reltol = 1e-8))
    expect_equal(unname(exp(fit$par)), unname(truePar), tolerance = 0.05)
  })
})

#' Is the dataset a memory summary?'
#'
#' @param dat data set
#' @return TRUE if the dataset is an \code{rxMemSummary}, FALSE otherwise
#' @noRd
#' @author Matthew L. Fidler
.isRxMemSummary <- function(dat) {
  inherits(dat, "rxMemSummary") ||
    (is.data.frame(dat) &&
       all(c("nobs", "ndoses") %in% names(dat)) &&
       !("evid" %in% names(dat)))
}

#' Create a per-ID event summary for memory estimation
#'
#' @param id   Integer or character vector of subject IDs (optional).
#' @param nobs Integer vector of observation counts per ID.
#' @param ndoses Integer vector of dose event counts per ID.
#' @return A data.frame with class \code{"rxMemSummary"}.
#' @examples
#'
#' # Three subjects with known observation and dose counts
#' rxMemSummary(nobs = c(48L, 96L, 48L), ndoses = c(7L, 14L, 7L))
#'
#' # Explicit subject IDs
#' rxMemSummary(nobs = c(10L, 20L), ndoses = c(5L, 5L), id = c(101L, 102L))
#' @export
rxMemSummary <- function(nobs, ndoses, id = seq_along(nobs)) {
  .ret <- data.frame(id = id, nobs = as.integer(nobs), ndoses = as.integer(ndoses))
  class(.ret) <- c("rxMemSummary", "data.frame")
  .ret
}
#' Summarize data for memory estimation from an event-table dataset
#'
#' It will be sent rxMemSummarizeDat()
#'
#' @param dat data to summarize
#' @return An \code{rxMemSummary} data.frame with columns \code{id}, \code{nobs}, and
#'  \code{ndoses}.
#' @noRd
#' @author Matthew L. Fidler
.rxMemSummarizeDat <- function(dat) {
  evid <- .  <- NULL  # nolint
  .groups <- attr(dat, "rxHomGroups", exact = TRUE)
  if (is.data.frame(dat) && !is.null(.groups) && "evid" %in% names(dat)) {
    .idCol <- grep("^id$", names(dat), ignore.case = TRUE, value = TRUE)[1]
    if (!is.na(.idCol)) {
      .repIds <- suppressWarnings(as.integer(dat[[.idCol]]))
      if (!anyNA(.repIds)) {
        .ids <- vector("list", 0L)
        .nobs <- vector("list", 0L)
        .ndoses <- vector("list", 0L)
        for (.i in seq_along(.groups)) {
          .groupIds <- .groups[[.i]]
          if (length(.groupIds) == 0L) next
          .rows <- .repIds == .i
          .nobsGroup <- sum(dat$evid[.rows] == 0L, na.rm = TRUE)
          .ndosesGroup <- sum(dat$evid[.rows] != 0L, na.rm = TRUE)
          .ids[[length(.ids) + 1L]] <- .groupIds
          .nobs[[length(.nobs) + 1L]] <- rep.int(.nobsGroup, length(.groupIds))
          .ndoses[[length(.ndoses) + 1L]] <- rep.int(.ndosesGroup, length(.groupIds))
        }
        if (length(.ids) > 0L) {
          return(rxMemSummary(
            id = unlist(.ids, use.names = FALSE),
            nobs = unlist(.nobs, use.names = FALSE),
            ndoses = unlist(.ndoses, use.names = FALSE)
          ))
        }
      }
    }
  }
  if (is.rxEt(dat)) {
    .groups <- .etGetGroups(.rxEtEnv(dat))
    if (length(.groups) == 0L) {
      return(rxMemSummary(nobs = integer(0), ndoses = integer(0), id = integer(0)))
    }
    .ids <- vector("list", length(.groups))
    .nobs <- vector("list", length(.groups))
    .ndoses <- vector("list", length(.groups))
    for (.i in seq_along(.groups)) {
      .g <- .groups[[.i]]
      .ids[[.i]] <- as.integer(.g$ids)
      .nobs[[.i]] <- rep.int(sum(.g$data$evid == 0L, na.rm = TRUE), length(.g$ids))
      .ndoses[[.i]] <- rep.int(sum(.g$data$evid != 0L, na.rm = TRUE), length(.g$ids))
    }
    return(rxMemSummary(
      id = unlist(.ids, use.names = FALSE),
      nobs = unlist(.nobs, use.names = FALSE),
      ndoses = unlist(.ndoses, use.names = FALSE)
    ))
  }
  .dt    <- data.table::as.data.table(dat)
  .idCol <- grep("^id$", names(.dt), ignore.case = TRUE, value = TRUE)[1]

  if (is.na(.idCol)) {
    .ret <- rxMemSummary(
      nobs   = sum(.dt[["evid"]] == 0L, na.rm = TRUE),
      ndoses = sum(.dt[["evid"]] != 0L, na.rm = TRUE)
    )
  } else {
    .agg <- .dt[, list(nobs = sum(evid == 0L, na.rm = TRUE),
                       ndoses = sum(evid != 0L, na.rm = TRUE)),
                by = .idCol]
    .ret <- rxMemSummary(id = .agg[[.idCol]], nobs = .agg$nobs, ndoses = .agg$ndoses)
  }
  .ret
}

.rxMemSolveLayoutStats <- function(dat, control = NULL, model = NULL) {
  .solveDat <- NULL
  .modelParams <- character(0)
  if (!is.null(model)) {
    .modelParams <- rxModelVars(model)$params
  }
  if (is.rxEt(dat)) {
    .solveInput <- dat
    if (!is.null(control) && !is.null(control$iCov) &&
        length(.etGroups(.rxEtEnv(dat))) > 0L) {
      .groupedSolve <- .etGroupedSolveDataICov(dat, control$iCov,
                                               keep = control$keep,
                                               modelParams = .modelParams)
      if (!is.null(.groupedSolve)) {
        .solveInput <- .groupedSolve$events
      }
    }
    .solveDat <- .etPrepareSolveEvents(.solveInput, control)
  } else if (is.data.frame(dat) && !is.null(attr(dat, "rxHomGroups", exact = TRUE))) {
    if (!is.null(control) && !is.null(control$iCov)) {
      .groupedSolve <- .etGroupedSolveDataFrameICov(dat, control$iCov,
                                                    keep = control$keep,
                                                    modelParams = .modelParams)
      if (!is.null(.groupedSolve)) {
        dat <- .groupedSolve$events
      }
    }
    .solveDat <- .etPrepareGroupedSolveData(dat, control)
  } else if (is.data.frame(dat) && "evid" %in% names(dat)) {
    .solveDat <- .etFixCmtForSolve(dat)
    if (nrow(.solveDat) > 0L &&
        all(.solveDat$evid != 0L, na.rm = TRUE) &&
        !is.null(control) &&
        any(vapply(control[c("from", "to", "by", "length.out")], Negate(is.null), logical(1)))) {
      .solveDat <- .etAddSolveObsRows(.solveDat, .etSolveObsTimes(.solveDat, control))
    }
  }
  if (is.null(.solveDat)) {
    return(NULL)
  }
  .solveNallTotal <- nrow(.solveDat)
  if (.solveNallTotal == 0L) {
    return(list(nallTotal = 0L, maxAllTimes = 0L))
  }
  .solveMaxAllTimes <- if ("id" %in% names(.solveDat)) {
    max(tabulate(match(.solveDat$id, unique(.solveDat$id))))
  } else {
    .solveNallTotal
  }
  list(
    nallTotal = as.integer(.solveNallTotal),
    maxAllTimes = as.integer(.solveMaxAllTimes)
  )
}

.rxMemResolveInput <- function(dat, control = NULL) {
  if (inherits(dat, "rxSolve")) {
    .env <- attr(class(dat), ".rxode2.env", exact = TRUE)
    if (is.environment(.env) && exists(".args.events", envir = .env, inherits = FALSE)) {
      .events <- get(".args.events", envir = .env, inherits = FALSE)
      if (is.null(control) && exists(".args", envir = .env, inherits = FALSE)) {
        control <- get(".args", envir = .env, inherits = FALSE)
      }
      return(list(dat = .events, control = control))
    }
  }
  if (is.character(dat) && length(dat) == 1L && .rxIsSerializedSolvePath(dat)) {
    .bundle <- .rxReadStateBundle(dat)
    if (is.null(.bundle$events)) {
      stop(sprintf("Serialized solve '%s' does not contain event data for memory estimation", dat),
           call. = FALSE)
    }
    return(list(dat = .bundle$events, control = control))
  }
  if (is.list(dat) && !is.data.frame(dat) && !is.null(dat$events)) {
    return(list(dat = dat$events, control = control))
  }
  list(dat = dat, control = control)
}
#' Extract model dimensions from model variables
#'
#'
#' @param model model variables to extract dimensions from
#'
#' @return A list with elements \code{neq}, \code{stateSize},
#'   \code{nlhs}, \code{npars}, \code{extraCmt}, \code{linB},
#'   \code{nMtime}, \code{nLlik}, and \code{nIndSim}. All are used to
#'   calculate memory useage.
#' @noRd
#' @author Matthew L. Fidler
.rxMemExtractModel <- function(model) {
  .mv    <- rxModelVars(model)
  .flags <- .mv[["flags"]]

  list(
    neq       = length(.mv[["state"]]),
    stateSize = length(.mv[["state"]]),
    nlhs      = length(.mv[["lhs"]]),
    npars     = length(.mv[["params"]]),
    extraCmt  = as.integer(.mv[["extraCmt"]]),
    linB      = as.integer(.flags["linB"]),
    nMtime    = as.integer(.mv[["nMtime"]]),
    nLlik     = as.integer(.flags["nLlik"]),
    nIndSim   = as.integer(.flags["nIndSim"])
  )
}
#' Extract memory-relevant options from an rxControl object
#'
#' @param ctrl An \code{rxControl} object.
#' @return A list with \code{cores}, \code{nsim}, \code{neta}, \code{neps},
#'   \code{nLlikAlloc}, and \code{nSubEff} (effective subject count;
#'   0 means use the data-derived count).
#' @noRd
#' @author Matthew L. Fidler
.rxMemExtractControl <- function(ctrl) {
  .cores <- as.integer(ctrl$cores)
  if (.cores <= 0L) .cores <- getRxThreads()
  .nsim  <- if (is.null(ctrl$nsim)) 1L else as.integer(ctrl$nsim)
  .neta  <- if (is.matrix(ctrl$omega)) nrow(ctrl$omega) else 0L
  .neps  <- if (is.matrix(ctrl$sigma)) nrow(ctrl$sigma) else 0L
  .nSub  <- if (is.null(ctrl$nSub))  1L else as.integer(ctrl$nSub)
  .nStud <- if (is.null(ctrl$nStud)) 1L else as.integer(ctrl$nStud)
  list(
    cores      = .cores,
    nsim       = .nsim,
    neta       = as.integer(.neta),
    neps       = as.integer(.neps),
    nLlikAlloc = ctrl$nLlikAlloc,
    nSub       = .nSub,
    nStud      = .nStud
  )
}
#' Detect total physical RAM in bytes
#'
#' @return Numeric bytes of total RAM, or \code{NA_real_} if unavailable.
#' @noRd
#' @author Matthew L. Fidler
.getRamBytes <- function() {
  .ram <- tryCatch(.Call(`_rxode2_rxRamBytes_`)[["total"]],
                   error = function(e) NA_real_)
  if (!is.na(.ram) && .ram > 0) return(.ram)
  NA_real_
}

.rxMemEstimateOutputData <- function(dat, summary, control, neq, nlhs, ncov,
                                     nsim, nsub, nobsTotal, nallTotal) {
  .addCov       <- TRUE
  .addDosing    <- FALSE
  .subsetNonmem <- TRUE
  .returnType   <- "rxSolve"
  .nkeep        <- 0L
  .ncov0        <- 0L
  .hasEvid2     <- FALSE
  .ptrBytes     <- .Machine$sizeof.pointer

  if (!is.null(control)) {
    if (!is.null(control$addCov)) {
      .addCov <- isTRUE(control$addCov)
    }
    if ("addDosing" %in% names(control)) {
      .addDosing <- control$addDosing
    }
    if (!is.null(control$subsetNonmem)) {
      .subsetNonmem <- isTRUE(control$subsetNonmem)
    }
    if (!is.null(control$returnType) && length(control$returnType) > 0L) {
      .returnType <- control$returnType[[1L]]
    }
    if (!is.null(control$keep)) {
      .nkeep <- length(control$keep)
    }
    if (!is.null(control$iCov)) {
      .ncov0 <- ncol(control$iCov)
    }
  }

  if (is.data.frame(dat) && "evid" %in% names(dat)) {
    .hasEvid2 <- any(dat$evid == 2L, na.rm = TRUE)
  }

  .doDose0 <- 0L
  if (is.null(.addDosing)) {
    .doDose0 <- -1L
  } else if (length(.addDosing) > 0L && isTRUE(is.na(.addDosing[[1L]]))) {
    .doDose0 <- 1L
  } else if (length(.addDosing) > 0L && isTRUE(.addDosing[[1L]])) {
    .doDose0 <- if (.subsetNonmem) 3L else 2L
  }

  .doDose <- as.integer(.doDose0 > 0L)
  .nmevid <- as.integer(.doDose0 %in% c(2L, 3L))
  .doTBS  <- as.integer(identical(.returnType, "data.frame.TBS"))
  .nidCols <- as.integer(nsub > 1L) + as.integer(nsim > 1L)
  .nevid2col <- as.integer(.doDose0 == 0L && .hasEvid2)

  .nr <- if (.doDose0 > 0L) {
    nallTotal * nsim
  } else {
    nobsTotal * nsim
  }

  .nIntCols <- .nidCols + .nevid2col + .doDose + 2L * .nmevid
  .nDblCols <- 1L + as.integer(neq) + as.integer(nlhs) +
    .doDose + 3L * .nmevid + .doTBS * 4L +
    as.integer(.addCov) * (as.integer(ncov) + .ncov0)

  .keepBytes <- 0
  if (.nkeep > 0L) {
    .keepNames <- control$keep
    .keepBytes <- sum(vapply(.keepNames, function(.nm) {
      if (!is.data.frame(dat) || !(.nm %in% names(dat))) {
        return(.nr * 8)
      }
      .col <- dat[[.nm]]
      if (is.character(.col)) {
        .nr * .ptrBytes
      } else if (is.logical(.col) || is.factor(.col) || is.integer(.col)) {
        .nr * 4
      } else {
        .nr * 8
      }
    }, numeric(1)))
  }

  .dataBytes <- .nr * (.nIntCols * 4 + .nDblCols * 8) + .keepBytes
  .listBytes <- (.nIntCols + .nDblCols + .nkeep) * .ptrBytes * 2

  .dataBytes + .listBytes
}

#' Detect currently available memory for allocation in bytes
#'
#' Uses the allocator preflight estimate (\code{rxAvailableMemoryBytes()}),
#' which may exceed physical RAM (page file on Windows, swap on Linux).
#'
#' @return Numeric bytes of available memory, or \code{NA_real_} if
#'   unavailable.
#' @noRd
#' @author Matthew L. Fidler
.getFreeRamBytes <- function() {
  .free <- tryCatch(.Call(`_rxode2_rxRamBytes_`)[["free"]],
                    error = function(e) NA_real_)
  if (!is.na(.free) && .free > 0) return(.free)
  NA_real_
}

#' Estimate memory required by rxSolve() for a given dataset and model
#'
#' Accepts either a pre-summarized per-ID table (an \code{\link{rxMemSummary}}
#' or any data.frame with \code{nobs} and \code{ndoses} columns) or a full
#' event-table data.frame with an \code{evid} column.  Model dimensions can
#' be supplied via a compiled rxode2 model object or overridden individually.
#'
#' The byte counts are computed by \code{rxMemoryComponents_()} which calls
#' the same \code{rxFillMemLayout()} used by the real allocator, so any
#' change to the allocation formulas propagates here automatically.
#'
#' @param dat A \code{\link{rxMemSummary}}, a data.frame with
#'   \code{nobs}/\code{ndoses} columns, or a full event-table
#'   data.frame with an \code{evid} column. This can also be a
#'   serialized solve state file path, a solve state bundle list
#'   containing an \code{events} element, or an \code{rxSolve}
#'   object (using its stored solve events).
#' @param model Optional rxode2 model object.  When supplied,
#'   \code{neq}, \code{nlhs}, \code{npars}, \code{extraCmt},
#'   \code{linB}, \code{nMtime}, \code{nLlik}, and \code{nIndSim} are
#'   extracted automatically.
#' @param control Optional \code{\link{rxControl}} object.  When
#'   supplied, \code{cores}, \code{nsim}, \code{neta} (from
#'   \code{omega}), \code{neps} (from \code{sigma}), and \code{nLlik}
#'   (adjusted by \code{nLlikAlloc}) are overridden automatically.
#' @param neq Number of ODE states.
#' @param stateSize Effective \code{state.size()} seen by the solver.
#'   Equals \code{neq} for pure ODE models; may differ for linCmt-only
#'   models.  Defaults to \code{neq}.
#' @param nlhs Number of LHS (calculated) output variables.
#' @param npars Number of model parameters (drives \code{gpars} size).
#' @param neta Number of random effects (etas).
#' @param neps Number of residual-error levels (epsilons).
#' @param ncov Number of time-varying covariates.
#' @param nsim Number of simulations.
#' @param cores Number of parallel OMP threads.
#' @param nMtime Number of model measurement times.
#' @param extraCmt Extra compartments (0, 1 = depot, 2 =
#'   depot+central).
#' @param linB \code{TRUE}/\code{1} if using a linear-compartment
#'   model.
#' @param nLlik Number of log-likelihood terms (FOCEi use).
#' @param nIndSim Per-individual simulation count.  Defaults to
#'   \code{neta + neps} when not supplied explicitly.
#' @param numLinSens Number of linear sensitivity parameters (FOCEi +
#'   linCmt).
#' @param numLin Number of linear compartment terms (FOCEi + linCmt).
#' @return A named list of class \code{"rxMemoryEstimate"} whose
#'   elements are raw byte counts plus \code{outputData},
#'   \code{ramBytes}, \code{freeRamBytes}, \code{total},
#'   \code{sizeofInd}, and \code{rxLlikSaveSize}.
#' @export
#'
#' @examples
#'
#' \donttest{
#'
#' mod <- rxode2::rxode2({
#'   d/dt(depot)  <- -ka * depot
#'   d/dt(center) <- ka * depot - cl / v * center
#'   cp           <- center / v
#' })
#'
#' ev <- rxode2::et(amt = 100, ii = 24, until = 168) |>
#'   rxode2::et(seq(0, 168, by = 1))
#'
#' # Basic estimate from event table and model
#' rxMemoryEstimate(as.data.frame(ev), model = mod)
#'
#' # With rxControl: population simulation with omega and 4 cores
#' ctrl <- rxode2::rxControl(
#'   cores = 4L,
#'   omega = lotri::lotri(eta.ka ~ 0.09, eta.cl ~ 0.04)
#' )
#' rxMemoryEstimate(as.data.frame(ev), model = mod, control = ctrl)
#'
#' }
rxMemoryEstimate <- function(
  dat,
  model     = NULL,
  control   = NULL,
  neq       = 1L,
  stateSize = neq,
  nlhs      = 0L,
  npars     = neq,
  neta      = 0L,
  neps      = 0L,
  ncov      = 0L,
  nsim      = 1L,
  cores     = 1L,
  nMtime    = 0L,
  extraCmt  = 0L,
  linB      = FALSE,
  nLlik     = 0L,
  nIndSim   = NULL,
  numLinSens = 0L,
  numLin    = 0L) {
  .resolved <- .rxMemResolveInput(dat, control)
  dat <- .resolved$dat
  control <- .resolved$control
  if (.isRxMemSummary(dat)) {
    .summary <- dat
    if (!inherits(.summary, "rxMemSummary")) {
      class(.summary) <- c("rxMemSummary", "data.frame")
    }
  } else if (is.data.frame(dat) && "evid" %in% names(dat)) {
    .summary <- .rxMemSummarizeDat(dat)
  } else {
    stop("'dat' must be an rxMemSummary, a data.frame with 'nobs'/'ndoses' ",
         "columns, or a data.frame with an 'evid' column")
  }

  if (!is.null(model)) {
    .mi       <- .rxMemExtractModel(model)
    neq       <- .mi$neq
    stateSize <- .mi$stateSize
    nlhs      <- .mi$nlhs
    npars     <- .mi$npars
    extraCmt  <- .mi$extraCmt
    linB      <- .mi$linB
    nMtime    <- .mi$nMtime
    nLlik     <- .mi$nLlik
    if (is.null(nIndSim)) nIndSim <- .mi$nIndSim
  }

  .ci <- NULL
  if (!is.null(control)) {
    .ci   <- .rxMemExtractControl(control)
    cores <- .ci$cores
    nsim  <- .ci$nsim
    if (.ci$neta > 0L) neta <- .ci$neta
    if (.ci$neps > 0L) neps <- .ci$neps
    if (!is.null(.ci$nLlikAlloc)) nLlik <- max(nLlik, as.integer(.ci$nLlikAlloc))
  }
  if (cores <= 0L) {
    cores <- 1L
  }
  if (is.null(nIndSim)) nIndSim <- neta + neps

  .nallVec     <- .summary$nobs + .summary$ndoses
  .nsub        <- nrow(.summary)
  .nobsTotal   <- sum(.summary$nobs)
  .nallTotal   <- sum(.nallVec)
  .maxAllTimes <- max(.nallVec)
  .solveStats  <- .rxMemSolveLayoutStats(dat, control, model)
  .solveNallTotal <- .nallTotal
  .solveMaxAllTimes <- .maxAllTimes

  if (!is.null(.ci) && (.ci$nSub > 1L || .ci$nStud > 1L)) {
    .subPerStudy <- if (.ci$nSub > 1L) .ci$nSub else .nsub
    .meanObsTimes <- .nobsTotal / .nsub
    .meanAllTimes <- .nallTotal / .nsub
    .nsub      <- .subPerStudy * .ci$nStud
    .nobsTotal <- .meanObsTimes * .nsub
    .nallTotal <- .meanAllTimes * .nsub
  } else if (!is.null(.solveStats)) {
    .solveNallTotal <- .solveStats$nallTotal
    .solveMaxAllTimes <- .solveStats$maxAllTimes
  }

  .raw <- rxMemoryComponents_(
    neq        = as.integer(neq),
    stateSize  = as.integer(stateSize),
    nlhs       = as.integer(nlhs),
    npars      = as.integer(npars),
    neta       = as.integer(neta),
    neps       = as.integer(neps),
    ncov       = as.integer(ncov),
    nsim       = as.integer(nsim),
    cores      = as.integer(cores),
    nMtime     = as.integer(nMtime),
    extraCmt   = as.integer(extraCmt),
    linB       = as.integer(linB),
    nLlik      = as.integer(nLlik),
    nIndSim    = as.integer(nIndSim),
    numLinSens = as.integer(numLinSens),
    numLin     = as.integer(numLin),
    nsub       = as.integer(.nsub),
    nallTotal  = as.double(.solveNallTotal),
    maxAllTimes = as.double(.solveMaxAllTimes)
  )

  .meta    <- c("sizeofInd", "rxLlikSaveSize")
  .sizes   <- .raw[!names(.raw) %in% .meta]
  .outputData <- .rxMemEstimateOutputData(
    dat       = dat,
    summary   = .summary,
    control   = control,
    neq       = neq,
    nlhs      = nlhs,
    ncov      = ncov,
    nsim      = nsim,
    nsub      = .nsub,
    nobsTotal = .nobsTotal,
    nallTotal = .nallTotal
  )
  .sizes   <- c(.sizes, outputData = .outputData)
  .wrapped <- lapply(.sizes, function(bytes) {
    structure(bytes, class = "rxRawBytes")
  })
  .total   <- Reduce(`+`, .wrapped)

  .ret <- c(list(total = .total), .wrapped,
            list(sizeofInd      = .raw[["sizeofInd"]],
                 rxLlikSaveSize = .raw[["rxLlikSaveSize"]],
                 ramBytes       = .getRamBytes(),
                 freeRamBytes   = .getFreeRamBytes(),
                 effectiveSubs  = .nsub))
  class(.ret) <- "rxMemoryEstimate"
  attr(.ret, "summary") <- .summary
  .ret
}

.rxOomChunkSize <- function(model, summary, control, safetyFactor = 0.80) {
  .est <- rxMemoryEstimate(summary, model = model, control = control)
  if (is.na(.est$freeRamBytes) || .est$freeRamBytes <= 0 || .est$effectiveSubs <= 0) {
    return(1L)
  }
  .memPerSub <- as.numeric(.est$total) / max(1L, .est$effectiveSubs)
  .avail <- .est$freeRamBytes * safetyFactor
  max(1L, as.integer(floor(.avail / .memPerSub)))
}

#' @export
print.rxMemoryEstimate <- function(x, ...) {
  .meta  <- c("total", "sizeofInd", "rxLlikSaveSize", "ramBytes", "freeRamBytes", "effectiveSubs")
  .comps <- x[!names(x) %in% .meta]

  .fmtSize <- function(v) {
    .b <- if (is.numeric(v)) v else unclass(v)
    if (.b >= 1e9)       sprintf("%.2f GB", .b / 1e9)
    else if (.b >= 1e6)  sprintf("%.2f MB", .b / 1e6)
    else if (.b >= 1e3)  sprintf("%.2f KB", .b / 1e3)
    else                 sprintf("%.0f B",  .b)
  }

  cat("rxSolve() memory estimate\n")
  cat(sprintf("  Total: %s\n\n", .fmtSize(x$total)))

  .bytes <- vapply(.comps, as.numeric, numeric(1))

  .ord <- order(.bytes, decreasing = TRUE)
  .nm  <- names(.comps)[.ord]
  .sz  <- .bytes[.ord]
  .tot <- sum(.sz)

  .labels <- c(
    gsolve        = "gsolve (double buffer total)",
    gsolve_n0     = "  |_ n0: ODE state output matrix",
    gon           = "gon (int buffer)",
    gall_times    = "gall_times (event time/dv/amt/ii/limit)",
    gevid         = "gevid (event IDs)",
    gcov          = "gcov (covariates)",
    gpars         = "gpars (parameters)",
    gomega        = "gomega (omega matrix)",
    gsigma        = "gsigma (sigma matrix)",
    gall_timesS   = "gall_timesS (extra sim times)",
    ordId         = "ordId (subject ordering)",
    gInfusionRate = "gInfusionRate (per-thread infusion)",
    inds_global   = "inds_global (per-subject structs)",
    outputData    = "outputData (estimated returned data)"
  )

  for (.i in seq_along(.nm)) {
    .n   <- .nm[.i]
    .lab <- if (!is.na(.labels[.n])) .labels[.n] else .n
    .pct <- if (.tot > 0) sprintf(" (%4.1f%%)", 100 * .sz[.i] / .tot) else ""
    cat(sprintf("  %-42s %s%s\n", .lab, .fmtSize(.comps[[.n]]), .pct))
  }

  .nsub <- if (!is.null(x$effectiveSubs)) as.integer(x$effectiveSubs) else nrow(attr(x, "summary"))
  cat(sprintf("\n  Subjects: %d  |  sizeof(rx_solving_options_ind): %d B",
              .nsub, as.integer(x$sizeofInd)))

  .ramBytes <- x$ramBytes
  .freeRamBytes <- x$freeRamBytes
  if (!is.null(.ramBytes) && !is.na(.ramBytes) && .ramBytes > 0) {
    .totalBytes <- as.numeric(x$total)
    cat(sprintf("  |  %.1f%% of RAM (%s)",
                100 * .totalBytes / .ramBytes, .fmtSize(.ramBytes)))
    if (!is.null(.freeRamBytes) && !is.na(.freeRamBytes) && .freeRamBytes > 0) {
      cat(sprintf("  |  %.1f%% of available memory (%s available)",
                  100 * .totalBytes / .freeRamBytes, .fmtSize(.freeRamBytes)))
    }
    cat("\n")
  } else {
    cat("\n")
  }
  invisible(x)
}

rxTest({
  test_that("rxMemSummary constructs correctly", {
    .s <- rxMemSummary(nobs = c(10L, 20L), ndoses = c(5L, 8L))
    expect_s3_class(.s, "rxMemSummary")
    expect_s3_class(.s, "data.frame")
    expect_equal(.s$nobs,   c(10L, 20L))
    expect_equal(.s$ndoses, c(5L, 8L))
    expect_equal(.s$id,     1:2)
  })

  test_that("rxMemSummary accepts explicit id", {
    .s <- rxMemSummary(nobs = 10L, ndoses = 5L, id = 99L)
    expect_equal(.s$id, 99L)
  })

  test_that("rxMemoryEstimate returns correct class", {
    .s   <- rxMemSummary(nobs = 100L, ndoses = 20L)
    .est <- rxMemoryEstimate(.s, neq = 2L, nlhs = 1L, npars = 3L)
    expect_s3_class(.est, "rxMemoryEstimate")
  })

  test_that("rxMemoryEstimate total equals sum of components", {
    .s     <- rxMemSummary(nobs = 100L, ndoses = 20L)
    .est   <- rxMemoryEstimate(.s, neq = 2L, nlhs = 1L, npars = 3L)
    .meta  <- c("total", "sizeofInd", "rxLlikSaveSize", "ramBytes", "freeRamBytes", "effectiveSubs")
    .comps <- .est[!names(.est) %in% .meta]
    expect_equal(as.numeric(.est$total), sum(vapply(.comps, as.numeric, numeric(1))))
  })

  test_that(".getRamBytes()/.getFreeRamBytes() query RAM natively", {
    .ram  <- .getRamBytes()
    .free <- .getFreeRamBytes()
    expect_true(is.numeric(.ram)  && length(.ram)  == 1L)
    expect_true(is.numeric(.free) && length(.free) == 1L)
    # Windows, macOS and Linux all have a native C path; none should be NA
    expect_gt(.ram, 0)
    expect_gt(.free, 0)
    expect_true(is.finite(.ram))
    expect_true(is.finite(.free))
    # no free <= total invariant: .getFreeRamBytes() is the allocation
    # budget, which may exceed physical RAM (page file on Windows, swap
    # on Linux)
  })

  test_that("rxMemoryEstimate contains memory availability metadata", {
    .s   <- rxMemSummary(nobs = 100L, ndoses = 20L)
    .est <- rxMemoryEstimate(.s, neq = 1L)
    expect_true("outputData" %in% names(.est))
    expect_gt(as.numeric(.est$outputData), 0)
    expect_true("ramBytes" %in% names(.est))
    expect_true("freeRamBytes" %in% names(.est))
    .rb <- .est$ramBytes
    .fb <- .est$freeRamBytes
    expect_true(is.numeric(.rb))
    expect_true(is.numeric(.fb))
    if (!is.na(.rb)) expect_gt(.rb, 0)
    if (!is.na(.fb)) expect_gt(.fb, 0)
  })

  test_that("rxMemoryEstimate accepts nobs/ndoses data.frame", {
    .df  <- data.frame(id = 1:3, nobs = c(10L, 20L, 30L), ndoses = c(2L, 4L, 6L))
    .est <- rxMemoryEstimate(.df, neq = 1L)
    expect_s3_class(.est, "rxMemoryEstimate")
    expect_equal(nrow(attr(.est, "summary")), 3L)
  })

  test_that("rxMemoryEstimate accepts evid event-table data.frame", {
    .df  <- data.frame(
      id   = c(1L, 1L, 1L, 2L, 2L),
      evid = c(1L, 0L, 0L, 1L, 0L)
    )
    .est <- rxMemoryEstimate(.df, neq = 1L)
    expect_s3_class(.est, "rxMemoryEstimate")
    .summ <- attr(.est, "summary")
    expect_equal(nrow(.summ), 2L)
    expect_equal(.summ$ndoses[.summ$id == 1L], 1L)
    expect_equal(.summ$nobs[.summ$id == 1L],   2L)
  })

  test_that("rxMemoryEstimate summarizes large compressed rxEt without losing ids", {
    .ev <- et(1, id = 1:2000)
    .est <- rxMemoryEstimate(.ev, neq = 1L)
    .summ <- attr(.est, "summary")
    expect_equal(nrow(.summ), 2000L)
    expect_equal(sum(.summ$nobs), 2000L)
    expect_equal(sum(.summ$ndoses), 0L)
  })

  test_that("grouped homogeneous rxEt lowers internal memory estimates", {
    .ev <- eventTable()
    .ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 24)
    .ev$add.sampling(seq(0, 48, by = 12))
    .ev <- et(.ev, id = 1:3)

    .grouped <- rxMemoryEstimate(.ev, neq = 2L, nlhs = 1L)
    .expanded <- rxMemoryEstimate(as.data.frame(.ev), neq = 2L, nlhs = 1L)

    expect_lt(as.numeric(.grouped$gall_times), as.numeric(.expanded$gall_times))
    expect_lt(as.numeric(.grouped$gevid), as.numeric(.expanded$gevid))
    expect_lt(as.numeric(.grouped$ordId), as.numeric(.expanded$ordId))
    expect_equal(as.numeric(.grouped$outputData), as.numeric(.expanded$outputData))
  })

  test_that("grouped dose-only rxEt lowers internal memory estimates", {
    .ev <- eventTable()
    .ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    .ev <- et(.ev, id = 1:4)
    .ctrl <- rxControl(from = 0, to = 24, by = 12)

    .grouped <- rxMemoryEstimate(.ev, neq = 2L, control = .ctrl)
    .expanded <- rxMemoryEstimate(as.data.frame(.ev), neq = 2L, control = .ctrl)

    expect_lt(as.numeric(.grouped$gall_times), as.numeric(.expanded$gall_times))
    expect_lt(as.numeric(.grouped$gevid), as.numeric(.expanded$gevid))
    expect_lt(as.numeric(.grouped$ordId), as.numeric(.expanded$ordId))
  })

  test_that("grouped dose-only rxEt with iCov keep stays compressed in memory estimate", {
    .mod <- rxode2({
      WT2 <- WT/70
      C2 <- centr / V2
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * WT2 * C2
    })
    .ev <- eventTable()
    .ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    .ev <- et(.ev, id = 1:4)
    .iCov <- data.frame(id = 1:4, WT = c(70, 70, 80, 80), grp = c("a", "a", "b", "b"))
    .ctrl <- rxControl(from = 0, to = 24, by = 12, iCov = .iCov, keep = "grp")

    .grouped <- rxMemoryEstimate(.ev, model = .mod, control = .ctrl)
    .expanded <- rxMemoryEstimate(as.data.frame(.ev), model = .mod, control = .ctrl)

    expect_lt(as.numeric(.grouped$gall_times), as.numeric(.expanded$gall_times))
    expect_lt(as.numeric(.grouped$gevid), as.numeric(.expanded$gevid))
    expect_lt(as.numeric(.grouped$ordId), as.numeric(.expanded$ordId))
    expect_equal(as.numeric(.grouped$outputData), as.numeric(.expanded$outputData))
  })

  test_that("grouped dose-only iCov keep affects solve layout without model", {
    .ev <- eventTable()
    .ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    .ev <- et(.ev, id = 1:4)
    .iCov <- data.frame(id = 1:4, grp = c("a", "a", "b", "b"))
    .ctrlNoKeep <- rxControl(from = 0, to = 24, by = 12, iCov = .iCov)
    .ctrlKeep <- rxControl(from = 0, to = 24, by = 12, iCov = .iCov, keep = "grp")

    .noKeep <- rxMemoryEstimate(.ev, neq = 2L, control = .ctrlNoKeep)
    .withKeep <- rxMemoryEstimate(.ev, neq = 2L, control = .ctrlKeep)
    .expanded <- rxMemoryEstimate(as.data.frame(.ev), neq = 2L, control = .ctrlKeep)

    expect_gt(as.numeric(.withKeep$gall_times), as.numeric(.noKeep$gall_times))
    expect_lt(as.numeric(.withKeep$gall_times), as.numeric(.expanded$gall_times))
  })

  test_that("grouped homogeneous data.frame preserves subject counts in memory estimate", {
    .ev <- eventTable()
    .ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    .ev$add.sampling(c(0, 12, 24))
    .ev <- et(.ev, id = 1:5)
    .groupedDf <- .etGroupedSolveData(.ev)

    .fromRxEt <- rxMemoryEstimate(.ev, neq = 2L)
    .fromGroupedDf <- rxMemoryEstimate(.groupedDf, neq = 2L)

    expect_equal(.fromGroupedDf$effectiveSubs, .fromRxEt$effectiveSubs)
    expect_equal(nrow(attr(.fromGroupedDf, "summary")), nrow(attr(.fromRxEt, "summary")))
    expect_equal(as.numeric(.fromGroupedDf$total), as.numeric(.fromRxEt$total))
    expect_equal(as.numeric(.fromGroupedDf$outputData), as.numeric(.fromRxEt$outputData))
  })

  test_that("rxMemoryEstimate errors on bad input", {
    expect_error(rxMemoryEstimate(data.frame(x = 1)), "'dat' must be")
  })

  test_that("print.rxMemoryEstimate runs without error", {
    .s   <- rxMemSummary(nobs = 50L, ndoses = 10L)
    .est <- rxMemoryEstimate(.s, neq = 1L)
    expect_output(print(.est), "rxSolve\\(\\) memory estimate")
    expect_output(print(.est), "Total:")
    if (!is.na(.est$freeRamBytes) && .est$freeRamBytes > 0) {
      expect_output(print(.est), "available memory")
    }
  })

  test_that("rxMemoryEstimate scales with subject count", {
    .s1 <- rxMemSummary(nobs = rep(50L, 10L),  ndoses = rep(5L, 10L))
    .s2 <- rxMemSummary(nobs = rep(50L, 100L), ndoses = rep(5L, 100L))
    .e1 <- rxMemoryEstimate(.s1, neq = 2L, npars = 3L)
    .e2 <- rxMemoryEstimate(.s2, neq = 2L, npars = 3L)
    expect_gt(.e2$total, .e1$total)
  })

  test_that("rxMemoryEstimate with compiled model", {
    skip_on_cran()
    .mod <- rxode2::rxode2({
      d/dt(depot)  <- -ka * depot
      d/dt(center) <- ka * depot - cl / v * center
    })
    .s   <- rxMemSummary(nobs = 100L, ndoses = 10L)
    .est <- rxMemoryEstimate(.s, model = .mod)
    expect_s3_class(.est, "rxMemoryEstimate")
    expect_gt(as.integer(.est$sizeofInd), 0L)
  })

  test_that("rxControl cores and nsim increase memory estimate", {
    .s     <- rxMemSummary(nobs = 100L, ndoses = 10L)
    .base  <- rxMemoryEstimate(.s, neq = 2L, npars = 3L)
    .ctrl4 <- rxControl(cores = 4L)
    .est4  <- rxMemoryEstimate(.s, neq = 2L, npars = 3L, control = .ctrl4)
    expect_gt(.est4$total, .base$total)
  })

  test_that("addDosing increases estimated output data memory", {
    .s       <- rxMemSummary(nobs = 100L, ndoses = 10L)
    .base    <- rxMemoryEstimate(.s, neq = 2L, nlhs = 1L)
    .ctrl    <- rxControl(addDosing = TRUE)
    .dosing  <- rxMemoryEstimate(.s, neq = 2L, nlhs = 1L, control = .ctrl)
    expect_gt(as.numeric(.dosing$outputData), as.numeric(.base$outputData))
  })

  test_that("rxControl omega sets neta, sigma sets neps", {
    .s    <- rxMemSummary(nobs = 100L, ndoses = 10L)
    .ctrl <- rxControl(
      omega = lotri::lotri(eta.ka ~ 0.09, eta.cl ~ 0.04)
    )
    .est  <- rxMemoryEstimate(.s, neq = 2L, npars = 3L, control = .ctrl)
    expect_gt(.est$gomega, 0)
  })

  test_that("rxControl nLlikAlloc raises nLlik floor", {
    .s     <- rxMemSummary(nobs = 100L, ndoses = 10L)
    .base  <- rxMemoryEstimate(.s, neq = 2L, nLlik = 1L)
    .ctrl  <- rxControl(nLlikAlloc = 5L)
    .est   <- rxMemoryEstimate(.s, neq = 2L, nLlik = 1L, control = .ctrl)
    expect_gt(.est$total, .base$total)
  })

  test_that("rxControl nSub overrides data subject count per study", {
    .s    <- rxMemSummary(nobs = rep(100L, 5L), ndoses = rep(10L, 5L))
    .base <- rxMemoryEstimate(.s, neq = 2L)
    .ctrl <- rxControl(nSub = 50L)
    .est  <- rxMemoryEstimate(.s, neq = 2L, control = .ctrl)
    expect_equal(.est$effectiveSubs, 50L)
  })

  test_that("rxControl nSub and nStud multiply: nSub overrides data subjects", {
    .s    <- rxMemSummary(nobs = rep(100L, 5L), ndoses = rep(10L, 5L))
    .ctrl <- rxControl(nSub = 10L, nStud = 5L)
    .est  <- rxMemoryEstimate(.s, neq = 2L, control = .ctrl)
    expect_equal(.est$effectiveSubs, 50L)
  })

  test_that("rxControl nSub=1 leaves subject count data-derived", {
    .s    <- rxMemSummary(nobs = rep(100L, 5L), ndoses = rep(10L, 5L))
    .ctrl <- rxControl(nSub = 1L)
    .est  <- rxMemoryEstimate(.s, neq = 2L, control = .ctrl)
    expect_equal(.est$effectiveSubs, 5L)
  })

  test_that("rxControl nStud multiplies data subject count", {
    .s    <- rxMemSummary(nobs = rep(100L, 5L), ndoses = rep(10L, 5L))
    .base <- rxMemoryEstimate(.s, neq = 2L)
    .ctrl <- rxControl(nStud = 100L)
    .est  <- rxMemoryEstimate(.s, neq = 2L, control = .ctrl)
    expect_equal(.est$effectiveSubs, 500L)
    expect_gt(.est$total, .base$total)
  })

  test_that("rxControl nStud with 1-subject dataset scales correctly", {
    .s    <- rxMemSummary(nobs = 100L, ndoses = 10L)
    .base <- rxMemoryEstimate(.s, neq = 2L)
    .ctrl <- rxControl(nStud = 100L)
    .est  <- rxMemoryEstimate(.s, neq = 2L, control = .ctrl)
    expect_equal(.est$effectiveSubs, 100L)
    expect_gt(.est$total, .base$total)
  })

  test_that("rxMemoryEstimate accepts serialized state files and bundles", {
    skip_on_cran()
    .mod <- rxode2({
      d/dt(depot) <- -ka * depot
      d/dt(centr) <- ka * depot - cl / v * centr
      cp <- centr / v
    })
    .theta <- c(ka = 1.5, cl = 10, v = 50)
    .ev <- eventTable()
    .ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    .ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    .ev <- et(.ev, id = 1:4)
    .stateFile <- tempfile(fileext = ".rxbin")
    rxSolve(.mod, .theta, .ev, serializeFile = .stateFile)
    .bundle <- .rxReadStateBundle(.stateFile)

    .fromFile <- rxMemoryEstimate(.stateFile, model = .mod)
    .fromBundle <- rxMemoryEstimate(.bundle, model = .mod)
    .fromEvents <- rxMemoryEstimate(.bundle$events, model = .mod)

    expect_equal(as.numeric(.fromFile$total), as.numeric(.fromEvents$total))
    expect_equal(as.numeric(.fromBundle$total), as.numeric(.fromEvents$total))
    expect_equal(as.numeric(.fromFile$outputData), as.numeric(.fromEvents$outputData))
    expect_equal(as.numeric(.fromBundle$outputData), as.numeric(.fromEvents$outputData))
  })

  test_that("rxMemoryEstimate accepts rxSolve objects", {
    skip_on_cran()
    .mod <- rxode2({
      d/dt(depot) <- -ka * depot
      d/dt(centr) <- ka * depot - cl / v * centr
      cp <- centr / v
    })
    .theta <- c(ka = 1.5, cl = 10, v = 50)
    .ev <- eventTable()
    .ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    .ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    .ev <- et(.ev, id = 1:4)
    .solved <- rxSolve(.mod, .theta, .ev)
    .env <- attr(class(.solved), ".rxode2.env")
    .fromSolve <- rxMemoryEstimate(.solved, model = .mod)
    .fromEvents <- rxMemoryEstimate(.env$.args.events, model = .mod, control = .env$.args)

    expect_equal(as.numeric(.fromSolve$total), as.numeric(.fromEvents$total))
    expect_equal(as.numeric(.fromSolve$outputData), as.numeric(.fromEvents$outputData))
  })

  test_that("rxMemoryEstimate infers control defaults from rxSolve objects", {
    skip_on_cran()
    .mod <- rxode2({
      d/dt(depot) <- -ka * depot
      d/dt(centr) <- ka * depot - cl / v * centr
      cp <- centr / v
    })
    .theta <- c(ka = 1.5, cl = 10, v = 50)
    .ev <- eventTable()
    .ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    .ev <- et(.ev, id = 1:4)
    .solved <- rxSolve(.mod, .theta, .ev, from = 0, to = 24, by = 12, nsim = 3)
    .env <- attr(class(.solved), ".rxode2.env")

    .fromSolve <- rxMemoryEstimate(.solved, model = .mod)
    .fromEvents <- rxMemoryEstimate(.env$.args.events, model = .mod, control = .env$.args)

    expect_equal(.fromSolve$effectiveSubs, .fromEvents$effectiveSubs)
    expect_equal(as.numeric(.fromSolve$total), as.numeric(.fromEvents$total))
    expect_equal(as.numeric(.fromSolve$gall_times), as.numeric(.fromEvents$gall_times))
  })

  test_that("rxMemoryEstimate explicit control overrides rxSolve defaults", {
    skip_on_cran()
    .mod <- rxode2({
      d/dt(depot) <- -ka * depot
      d/dt(centr) <- ka * depot - cl / v * centr
      cp <- centr / v
    })
    .theta <- c(ka = 1.5, cl = 10, v = 50)
    .ev <- eventTable()
    .ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    .ev <- et(.ev, id = 1:4)
    .solved <- rxSolve(.mod, .theta, .ev, from = 0, to = 24, by = 12, nsim = 3)
    .env <- attr(class(.solved), ".rxode2.env")

    .override <- rxControl(from = 0, to = 24, by = 12, nsim = 1)
    .fromSolveDefault <- rxMemoryEstimate(.solved, model = .mod)
    .fromSolveOverride <- rxMemoryEstimate(.solved, model = .mod, control = .override)
    .fromEventsOverride <- rxMemoryEstimate(.env$.args.events, model = .mod, control = .override)

    expect_equal(.fromSolveOverride$effectiveSubs, .fromEventsOverride$effectiveSubs)
    expect_equal(as.numeric(.fromSolveOverride$total), as.numeric(.fromEventsOverride$total))
    expect_equal(as.numeric(.fromSolveOverride$gall_times), as.numeric(.fromEventsOverride$gall_times))
    expect_false(isTRUE(all.equal(as.numeric(.fromSolveOverride$total),
                                  as.numeric(.fromSolveDefault$total))))
  })

  test_that("rxMemoryEstimate file/bundle/rxSolve parity for same dose-only grouped solve", {
    skip_on_cran()
    .mod <- rxode2({
      d/dt(depot) <- -ka * depot
      d/dt(centr) <- ka * depot - cl / v * centr
      cp <- centr / v
    })
    .theta <- c(ka = 1.5, cl = 10, v = 50)
    .ev <- eventTable()
    .ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    .ev <- et(.ev, id = 1:4)
    .ctrl <- rxControl(from = 0, to = 24, by = 12)
    .stateFile <- tempfile(fileext = ".rxbin")

    rxSolve(.mod, .theta, .ev, serializeFile = .stateFile)
    .bundle <- .rxReadStateBundle(.stateFile)
    .solved <- rxSolve(.mod, .theta, .ev)

    .fromFile <- rxMemoryEstimate(.stateFile, model = .mod, control = .ctrl)
    .fromBundle <- rxMemoryEstimate(.bundle, model = .mod, control = .ctrl)
    .fromSolve <- rxMemoryEstimate(.solved, model = .mod, control = .ctrl)

    expect_equal(as.numeric(.fromFile$total), as.numeric(.fromBundle$total))
    expect_equal(as.numeric(.fromFile$total), as.numeric(.fromSolve$total))
    expect_equal(as.numeric(.fromFile$gall_times), as.numeric(.fromBundle$gall_times))
    expect_equal(as.numeric(.fromFile$gall_times), as.numeric(.fromSolve$gall_times))
  })
})

.vname <- function(x) {
  .v <- paste0(deparse1(eval.parent(substitute(substitute(x)))),collapse = "\n")
  if (regexpr("[ >]+", .v) != -1) {
    return("model")
  }
  .v
}

#' Assert properties of the rxUi models
#'
#' @param ui Model to check
#'
#' @param extra Extra text to append to the error message (like
#'   "for focei")
#'
#' @details
#'
#' These functions have different types of assertions
#'
#' - `assertRxUi` -- Make sure this is a proper rxode2 model (if not throw error)
#'
#' - `assertRxUiSingleEndpoint` -- Make sure the rxode2 model is only
#'    a single endpoint model (if not throw error)
#'
#' - `assertRxUiTransformNormal` -- Make sure that the model residual
#'    distribution is normal or transformably normal
#'
#' - `assertRxUiNormal` -- Make sure that the model residual distribution is normal
#'
#' - `assertRxUiEstimatedResiduals` -- Make sure that the residual error
#'    parameters are estimated (not modeled).
#'
#' - `assertRxUiPopulationOnly` -- Make sure the model is the population only
#'    model (no mixed effects)
#'
#' - `assertRxUiMixedOnly` -- Make sure the model is a mixed effect model (not a
#'    population effect, only)
#'
#' - `assertRxUiPrediction` -- Make sure the model has predictions
#'
#' - `assertRxUiMuRefOnly` -- Make sure that all the parameters are mu-referenced
#'
#' - `assertRxUiRandomOnIdOnly` -- Make sure there are only random effects at the ID level
#'
#' - `assertRxUiIovNoCor` -- Make sure that the IOV model does not have any correlations
#'
#' - `assertRxUiNoMix` -- Make sure that the model does not have a mixture model inside it
#'
#' - `assertRxUiNoAutoregressive` -- Make sure the model does not have an
#'    autoregressive residual (ie `ar()`); used by estimation methods that do
#'    not support it
#'
#' @return the rxUi model
#'
#' @inheritParams checkmate::assertIntegerish
#'
#' @family Assertions
#' @author Matthew L. Fidler
#'
#' @export
#'
#' @examples
#'
#' \donttest{
#' one.cmt <- function() {
#'  ini({
#'    tka <- 0.45; label("Ka")
#'    tcl <- log(c(0, 2.7, 100)); label("Cl")
#'    tv <- 3.45; label("V")
#'    eta.ka ~ 0.6
#'    eta.cl ~ 0.3
#'    eta.v ~ 0.1
#'    add.sd <- 0.7
#'  })
#'  model({
#'    ka <- exp(tka + eta.ka)
#'    cl <- exp(tcl + eta.cl)
#'    v <- exp(tv + eta.v)
#'    linCmt() ~ add(add.sd)
#'  })
#' }
#'
#' assertRxUi(one.cmt)
#' # assertRxUi(rnorm) # will fail
#'
#' assertRxUiSingleEndpoint(one.cmt)
#' }
assertRxUi <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- try(as.rxUi(ui), silent = TRUE)
  if (inherits(ui, "try-error")) {
    stop("'", .var.name, "' needs to be a rxUi model", extra, call.=FALSE)
  }
  invisible(ui)
}
#' Test if rxode2 uses linear solved systems
#'
#' @param ui rxode2 model
#' @inheritParams assertRxUi
#' @return TRUE if the model uses linear solved systems, FALSE otherwise
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' one.cmt <- function() {
#'   ini({
#'    ## You may label each parameter with a comment
#'    tka <- 0.45 # Log Ka
#'    tcl <- log(c(0, 2.7, 100)) # Log Cl
#'    ## This works with interactive models
#'    ## You may also label the preceding line with label("label text")
#'    tv <- 3.45; label("log V")
#'    ## the label("Label name") works with all models
#'    eta.ka ~ 0.6
#'    eta.cl ~ 0.3
#'    eta.v ~ 0.1
#'    add.sd <- 0.7
#'  })
#'  model({
#'    ka <- exp(tka + eta.ka)
#'    cl <- exp(tcl + eta.cl)
#'    v <- exp(tv + eta.v)
#'    linCmt() ~ add(add.sd)
#'  })
#'}
#'
#' testRxLinCmt(one.cmt)
#'
testRxLinCmt <- function(ui, extra="", .var.name=.vname(ui)) {
  .ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  if (!is.null(.ui$.linCmtM)) {
    return(TRUE)
  }
  .predDf <- .ui$predDf
  if (any(.predDf$linCmt)) {
    return(TRUE)
  }
  FALSE
}

#' @describeIn testRxLinCmt Assert that the rxode2 uses linear solved systems
#' @export
assertRxLinCmt <- function(ui, extra="", .var.name=.vname(ui)) {
  .ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  if (testRxLinCmt(.ui)) {
    return(invisible(.ui))
  }
  stop("'", .var.name, "' needs to have 'linCmt()'", extra, call.=FALSE)
}

#' @export
#' @rdname assertRxUi
assertRxUiPrediction <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  .predDf <- ui$predDf
  if (is.null(.predDf)) {
    stop("there must be at least one prediction in the model({}) block", extra, ".  Use `~` for predictions",
         call.=FALSE)
  }
  invisible(ui)
}
#' @export
#' @rdname assertRxUi
assertRxUiIovNoCor <- function(ui, extra="", .var.name=.vname(ui)) {
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  .iniDf <- ui$iniDf
  .w <- which(!is.na(.iniDf$condition) &
                .iniDf$condition != "id" &
                 is.na(.iniDf$err) &
                 .iniDf$neta1 != .iniDf$neta2)
  if (length(.w) > 0) {
    stop("'", .var.name, "' cannot have covariance/correlation for IOV related components", extra, call.=FALSE)
  }
  invisible(ui)
}

#' @export
#' @rdname assertRxUi
assertRxUiNoMix <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  if (!is.null(ui$mixProbs)) {
    stop("'", .var.name, "' cannot have a mixture model (ie `mix()`)", extra, call.=FALSE)
  }
  invisible(ui)
}

#' Test if a model uses an autoregressive (`ar()`) residual
#'
#' @param ui rxode2 user interface model
#' @return logical, `TRUE` if any endpoint carries an `ar()` correlation
#'   (either an estimated correlation or a modeled/literal one)
#' @author Matthew L. Fidler
#' @export
rxHasAr <- function(ui) {
  ui <- assertRxUi(ui)
  .iniDf <- ui$iniDf
  # literal (auto-fixed) and estimated ar() correlations live in the $iniDf with
  # err == "ar"
  if (!is.null(.iniDf) && any(.iniDf$err == "ar", na.rm=TRUE)) return(TRUE)
  # a modeled correlation (e.g. corv <- expit(tcor); ar(corv)) is not a
  # parameter, so scan the endpoint error expressions for an ar() term
  .lst <- tryCatch(ui$lstExpr, error=function(e) NULL)
  if (is.null(.lst)) return(FALSE)
  .hasAr <- function(e) {
    if (is.call(e)) {
      if (identical(e[[1]], quote(ar))) return(TRUE)
      return(any(vapply(as.list(e), .hasAr, logical(1))))
    }
    FALSE
  }
  any(vapply(.lst, function(e) {
    is.call(e) && identical(e[[1]], quote(`~`)) && .hasAr(e)
  }, logical(1)))
}

#' @export
#' @rdname assertRxUi
assertRxUiNoAutoregressive <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  if (rxHasAr(ui)) {
    stop("'", .var.name, "' cannot have an autoregressive residual (ie `ar()`)", extra, call.=FALSE)
  }
  invisible(ui)
}

#' @export
#' @rdname assertRxUi
assertRxUiSingleEndpoint <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  assertRxUiPrediction(ui)
  .predDf <- ui$predDf
  .err <- FALSE
  if (length(.predDf$cond) > 1L) {
    stop("'", .var.name, "' needs to be a single endpoint model", extra, call.=FALSE)
  }
  invisible(ui)
}

#' @export
#' @rdname assertRxUi
assertRxUiTransformNormal <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  assertRxUiPrediction(ui)
  .predDf <- ui$predDf
  if (!all(.predDf$distribution == "norm")) {
    stop("'", .var.name, "' needs to be a (transformably) normal model", extra, call.=FALSE)
  }
  invisible(ui)
}

#' @export
#' @rdname assertRxUi
assertRxUiNormal <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  assertRxUiPrediction(ui)
  .predDf <- ui$predDf
  if (!all(.predDf$distribution == "norm" & .predDf$transform == "untransformed")) {
    stop("'", .var.name, "' needs to be a normal model", extra, call.=FALSE)
  }
  invisible(ui)
}


#' @export
#' @rdname assertRxUi
assertRxUiMuRefOnly <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  if (length(ui$nonMuEtas) != 0) {
    stop("'", .var.name, "' needs to be a completely mu-referenced model (ie tcl+eta.cl)", extra, call.=FALSE)
  }
}

#' @export
#' @rdname assertRxUi
assertRxUiEstimatedResiduals <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  assertRxUiPrediction(ui)
  .predDf <- ui$predDf
  if (!all(is.na(unlist(.predDf[ ,c("a", "b", "c", "d", "e", "f", "lambda")], use.names=FALSE)))) {
    stop("'", .var.name, "' residual parameters cannot depend on the model calculated parameters", extra, call.=FALSE)
  }
  invisible(ui)
}

#' @export
#' @rdname assertRxUi
assertRxUiPopulationOnly <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  .iniDf <- ui$iniDf
  if (any(!is.na(.iniDf$neta1))) {
    stop("'", .var.name, "' can only have population estimates", extra, call.=FALSE)
  }
  invisible(ui)
}

#' @export
#' @rdname assertRxUi
assertRxUiMixedOnly <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  .iniDf <- ui$iniDf
  if (all(is.na(.iniDf$neta1))) {
    stop("'", .var.name, "' needs to be a mixed effect model", extra, call.=FALSE)
  }
  invisible(ui)
}

#' @export
#' @rdname assertRxUi
assertRxUiRandomOnIdOnly <- function(ui, extra="", .var.name=.vname(ui)) {
  force(.var.name)
  ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  .iniDf <- ui$iniDf
  .eta <- .iniDf[!is.na(.iniDf$neta1), "condition"]
  if (!all(.eta == "id")) {
    stop("'", .var.name, "' can only have random effects on ID", extra, call.=FALSE)
  }
  invisible(ui)
}

#' Verify that a value is a valid nlmixr2 compartment name
#'
#' @param ui when needed, this is the rxode2/nlmixr2 model
#' @param x The value to test
#' @return The value or an error
#' @family Assertions
#' @author Bill Denney
#' @export
assertCompartmentName <- function(x) {
  checkmate::assertCharacter(
    x,
    pattern = "^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
    len = 1,
    any.missing = FALSE,
    min.chars = 1,
    .var.name = paste0(deparse(eval.parent(substitute(substitute(x))), width.cutoff = 500L), collapse = "\n")
  )
}

#' Verify that a compartment would be new to the model
#'
#' @param ui is the model to test that a model paramet exists
#' @param x The value to test
#' @return The value or an error
#' @family Assertions
#' @author Matthew Fidler & Bill Denney
#' @export
assertCompartmentNew <- function(ui, x) {
  .vn <- as.character(substitute(x))
  .tmp <- try(force(x), silent=TRUE)
  if (!inherits(.tmp, "try-error")) {
    if (is.character(x)) {
      .vn <- x
    }
  }
  checkmate::assertCharacter(
    .vn,
    pattern = "^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
    len = 1,
    any.missing = FALSE,
    min.chars = 1,
    .var.name = paste0(deparse(eval.parent(substitute(substitute(x))), width.cutoff = 500L), collapse = "\n")
  )

  .ui <-rxode2::assertRxUi(ui)
  if (.vn %in% c(rxode2::rxModelVars(.ui)$state)) {
    stop("compartment '", .vn, "' already exists in the model",
         call.=FALSE)
  }
  return(invisible())
}

#' Verify that the compartment exists in a model
#'
#' @param ui is the model to test
#' @param x The value to test (can be a vector of strings)
#' @return the value of the compartment that exists; if it is a vector
#'   returns the first item that matches
#' @family Assertions
#' @author Matthew Fidler & Bill Denney
#' @export
assertCompartmentExists <- function(ui, x) {
  .all <- as.character(substitute(x))
  .tmp <- try(force(x), silent=TRUE)
  if (!inherits(.tmp, "try-error")) {
    if (is.character(x)) {
      .all <- x
    }
  }
  .ui <-rxode2::assertRxUi(ui)
  .state <- rxode2::rxModelVars(.ui)$state
  for (.vn in .all) {
    checkmate::assertCharacter(
      .vn,
      pattern = "^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
      len = 1,
      any.missing = FALSE,
      min.chars = 1,
      .var.name = paste0(deparse(eval.parent(substitute(substitute(x))), width.cutoff = 500L), collapse = "\n")
    )

    if (.vn %in% .state) return(invisible(.vn))
  }
  stop("'", paste(.all, collapse="', '"), "' compartment is not in the model",
       call.=FALSE)
}

#' @describeIn assertCompartmentExists Test if compartment exists
#' @export
testCompartmentExists <- function(ui, x) {
  .vn <- as.character(substitute(x))
  .tmp <- try(force(x), silent=TRUE)
  if (!inherits(.tmp, "try-error")) {
    if (is.character(x)) {
      .vn <- x
    }
  }
  checkmate::assertCharacter(
    .vn,
    pattern = "^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
    len = 1,
    any.missing = FALSE,
    min.chars = 1,
    .var.name = paste0(deparse(eval.parent(substitute(substitute(x))), width.cutoff = 500L), collapse = "\n")
  )

  .ui <-rxode2::assertRxUi(ui)
  (.vn %in% c(rxode2::rxModelVars(.ui)$state))
}
#' @describeIn assertCompartmentName Verify that a value is a valid
#'   nlmixr2 variable name
#' @export
assertVariableName <- assertCompartmentName

#' Assert a variable exists in the model
#'
#' @param ui rxode2 ui model
#' @param x does the `x` variable exist in the model.  If it is a
#'   vector of variable check to see if any exists, but all must be
#'   valid nlmixr2 variable names
#' @return variable that matches, in the case of multiple variables,
#'   the first that matches.  If nothing matches return error
#' @export
#' @family Assertions
#' @author Matthew L. Fidler
assertVariableExists <- function(ui, x) {
  .all <- as.character(substitute(x))
  .tmp <- try(force(x), silent=TRUE)
  if (!inherits(.tmp, "try-error")) {
    if (is.character(x)) {
      .all <- x
    }
  }
  .ui <-rxode2::assertRxUi(ui)
  .mv <- rxode2::rxModelVars(.ui)

  for (.vn in .all) {
    checkmate::assertCharacter(
      .vn,
      pattern = "^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
      len = 1,
      any.missing = FALSE,
      min.chars = 1,
      .var.name = paste0(deparse(eval.parent(substitute(substitute(x))), width.cutoff = 500L), collapse = "\n")
    )
    if (.vn %in% c(.mv$lhs, .mv$params)) {
      return(invisible(.vn))
    }
  }
  stop("variable '", paste(.all, collapse="', '"), "' not in the model",
       call.=FALSE)
}

#' @describeIn assertVariableExists Test if variable exists
#' @export
testVariableExists <- function(ui, x) {
  .all <- as.character(substitute(x))
  .tmp <- try(force(x), silent=TRUE)
  if (!inherits(.tmp, "try-error")) {
    if (is.character(x)) {
      .all <- x
    }
  }
  .ui <-rxode2::assertRxUi(ui)
  .mv <- rxode2::rxModelVars(.ui)
  for (.vn in .all) {
    checkmate::assertCharacter(
      .vn,
      pattern = "^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
      len = 1,
      any.missing = FALSE,
      min.chars = 1,
      .var.name = paste0(deparse(eval.parent(substitute(substitute(x))), width.cutoff = 500L), collapse = "\n")
    )
    if (.vn %in% c(.mv$lhs, .mv$params)) return(TRUE)
  }
  FALSE
}

#' Assert a variable would be new to the model
#'
#' @param ui rxode2 ui model
#' @param x would the variable `x` variable be new in the model
#' @return nothing, but will error if `x` would not be new
#' @export
#' @family Assertions
#' @author Matthew L. Fidler
assertVariableNew <- function(ui, x) {
  .vn <- as.character(substitute(x))
  .tmp <- try(force(x), silent=TRUE)
  if (!inherits(.tmp, "try-error")) {
    if (is.character(x)) {
      .vn <- x
    }
  }
  checkmate::assertCharacter(
    .vn,
    pattern = "^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
    len = 1,
    any.missing = FALSE,
    min.chars = 1,
    .var.name = paste0(deparse(eval.parent(substitute(substitute(x))), width.cutoff = 500L), collapse = "\n")
  )

  .ui <-rxode2::assertRxUi(ui)
  .mv <- rxode2::rxModelVars(.ui)
  if (.vn %in% c(.mv$lhs, .mv$params))  {
    stop("variable '", .vn, "' is already in the model",
         call.=FALSE)
  }
  invisible()
}

#' @describeIn assertCompartmentName Verify that a value is a valid
#'   nlmixr2 parameter value
#' @export
assertParameterValue <- function(x) {
  checkmate::assertNumeric(
    x,
    len=1,
    any.missing=FALSE,
    finite = TRUE,
    .var.name = paste0(deparse(eval.parent(substitute(substitute(x))), width.cutoff = 500L), collapse = "\n")
  )
}

#' @describeIn assertCompartmentName Assert compartment/variable exists
#' @export
assertExists <- function(ui, x) {
  .all <- as.character(substitute(x))
  .tmp <- try(force(x), silent=TRUE)
  if (!inherits(.tmp, "try-error")) {
    if (is.character(x)) {
      .all <- x
    }
  }
  for (.vn in .all) {
    checkmate::assertCharacter(
      .vn,
      pattern = "^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
      len = 1,
      any.missing = FALSE,
      min.chars = 1,
      .var.name = paste0(deparse(eval.parent(substitute(substitute(x))), width.cutoff = 500L), collapse = "\n")
    )

    .ui <-rxode2::assertRxUi(ui)
    .mv <- rxode2::rxModelVars(.ui)
    if (.vn %in% c(.mv$lhs, .mv$params, .mv$state)) return(invisible(.vn))
  }
  stop("'", paste(.all, collapse="', '"), "' not in the model",
       call.=FALSE)
}

#' @describeIn assertCompartmentName Test compartment/variable exists
#' @export
testExists <- function(ui, x) {
  .vn <- as.character(substitute(x))
  .tmp <- try(force(x), silent=TRUE)
  if (!inherits(.tmp, "try-error")) {
    if (is.character(x)) {
      .vn <- x
    }
  }
  checkmate::assertCharacter(
    .vn,
    pattern = "^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$",
    len = 1,
    any.missing = FALSE,
    min.chars = 1,
    .var.name = paste0(deparse(eval.parent(substitute(substitute(x))), width.cutoff = 500L), collapse = "\n")
  )

  .ui <-rxode2::assertRxUi(ui)
  .mv <- rxode2::rxModelVars(.ui)
  if (.vn %in% c(.mv$lhs, .mv$params, .mv$state)) return(TRUE)
  FALSE
}

#' Check if parameters have user boundaries different than defaults
#'
#' @param ui rxode2 ui
#' @param extra extra information to append to the error message
#' @param .var.name variable name
#' @return a named logical vector indicating whether each parameter is bounded
#' @noRd
#' @author Matthew L. Fidler
.getRxBounded <- function(ui, extra="", .var.name=.vname(ui)) {
  .ui <- assertRxUi(ui, extra=extra, .var.name=.var.name)
  .iniDf <- .ui$iniDf
  .theta <- .iniDf[which(!is.na(.iniDf$ntheta)),]
  setNames(vapply(seq_along(.theta$name),
               function(i) {
                 .t <- .theta[i,]
                 if (is.na(.t$err)) {
                   return(is.finite(.t$upper) || is.finite(.t$lower))
                 }
                 .err <- .errDistArgRanges[[.t$err]]
                 return (!identical(.t$lower, .err[1]) ||
                           !identical(.t$upper, .err[2]))
               }, logical(1), USE.NAMES=FALSE), .theta$name)
}
#' Test if the rxode2 model has any parameters with user defined boundaries
#'
#' @param ui rxode2 ui
#' @param extra extra information to append to the error message
#' @param .var.name variable name
#' @return boolean indicating if any parameters have user defined boundaries
#' @family Assertions
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' one.cmt <- function() {
#'   ini({
#'     tka <- 0.45; label("Ka")
#'     tcl <- log(c(0, 2.7, 100)); label("Cl")
#'     tv <- 3.45; label("V")
#'     eta.ka ~ 0.6
#'     eta.cl ~ 0.3
#'     eta.v ~ 0.1
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl + eta.cl)
#'     v <- exp(tv + eta.v)
#'     linCmt() ~ add(add.sd)
#'   })
#' }
#'
#' testRxUnbounded(one.cmt)
#'
#' try(assertRxUnbounded(one.cmt))
#'
#' warnRxBounded(one.cmt)
testRxUnbounded <- function(ui) {
  !any(.getRxBounded(ui))
}

#' @describeIn testRxUnbounded Assert that the rxode2 model has any parameters with user defined boundaries
#' @export
assertRxUnbounded <- function(ui, extra="", .var.name=.vname(ui)) {
  if (testRxUnbounded(ui)) {
    return(invisible(ui))
  }
  stop("'", .var.name, "' can not have user defined boundaries", extra, call.=FALSE)
}

#' @describeIn testRxUnbounded Warn that the rxode2 model has any parameters with user defined boundaries
#' @export
warnRxBounded <- function(ui, extra="", .var.name=.vname(ui)) {
  .bound <- .getRxBounded(ui, extra=extra, .var.name=.var.name)
  .w <- which(.bound)
  if (length(.w) > 0) {
    warning("'", .var.name, "' has the following user-defined boundaries: ",
         paste(names(.bound)[.w], collapse=", "),
         extra, call.=FALSE)
  }
  invisible()
}

#' This function tests if this object is a iniDf as needed by the UI
#'
#'
#' @param iniDf the object to test if it is a rxode2 ui `iniDf` data.frame
#' @param extra information to append to the error message
#' @inheritParams checkmate::testDataFrame
#' @return boolean, indicating if the object is a valid initialization data frame
#' @export
#' @author Matthew L. Fidler
#' @family Assertions
#' @examples
#' testIniDf(TRUE)
testIniDf <- function(iniDf) {
  if (checkmate::testDataFrame(iniDf)) {
    checkmate::testSubset(names(iniDf),
                          c("ntheta", "neta1", "neta2", "name", "lower", "est", "upper",
                            "fix", "label", "backTransform", "condition", "err"))
  } else {
    FALSE
  }
}
#' @describeIn testIniDf Assert that the object is a valid rxode2 ui initialization data frame
#' @export
assertIniDf <- function(iniDf, extra="", .var.name=.vname(iniDf), null.ok = FALSE) {
  if (testIniDf(iniDf)) {
    return(invisible(iniDf))
  }
  if (null.ok && is.null(iniDf)) {
    return(invisible(NULL))
  }
  stop("'", .var.name, "' is not a rxode2 ui initial conditions data.frame", extra, call.=FALSE)
}

rxTest({
  test_that("assert properties of rxUi models", {
    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    pk.turnover.emax <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)

        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1

        temax <- logit(0.8)
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)

        eta.emax ~ .5
        eta.ec50  ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5

        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        ##
        #poplogit = log(temax/(1-temax))
        emax=expit(temax+eta.emax)
        #logit=temax+eta.emax
        ec50 =  exp(tec50 + eta.ec50)
        kout = exp(tkout + eta.kout)
        e0 = exp(te0 + eta.e0)
        ##
        DCP = center/v
        PD=1-emax*DCP/(ec50+DCP)
        ##
        effect(0) = e0
        kin = e0*kout
        ##
        d/dt(depot) = -ktr * depot
        d/dt(gut) =  ktr * depot -ka * gut
        d/dt(center) =  ka * gut - cl / v * center
        d/dt(effect) = kin*PD -kout*effect
        ##
        cp = center / v
        cp ~ prop(prop.err) + add(pkadd.err)
        effect ~ add(pdadd.err)
      })
    }

    suppressMessages(
      expect_error(
        assertRxUi(rnorm),
        "needs to be a rxUi model"
      )
    )
    expect_error(assertRxUi(one.cmt), NA)

    expect_error(assertRxUiSingleEndpoint(pk.turnover.emax))

    expect_error(assertRxUiSingleEndpoint(one.cmt), NA)

    expect_error(assertRxUiNormal(one.cmt), NA)

    expect_error(assertRxUiTransformNormal(one.cmt), NA)

    one.cmt.t <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
        nu <- 3
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd) + dt(nu)
      })
    }

    expect_error(assertRxUiNormal(one.cmt.t))

    expect_error(assertRxUiTransformNormal(one.cmt.t))

    expect_error(assertRxUiEstimatedResiduals(one.cmt.t), NA)

    one.cmt.t.est <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        nu <- 3
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        add.sd <- 3 + ka
        linCmt() ~ add(add.sd) + dt(nu)
      })
    }

    expect_error(assertRxUiEstimatedResiduals(one.cmt.t.est))
    expect_error(assertRxUiEstimatedResiduals(one.cmt.t), NA)

    expect_error(assertRxUiMixedOnly(one.cmt.t), NA)
    expect_error(assertRxUiPopulationOnly(one.cmt.t))

    one.cmt.pop <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        add.sd <- 0.7
        nu <- 3
      })
      model({
        ka <- exp(tka)
        cl <- exp(tcl)
        v <- exp(tv)
        linCmt() ~ add(add.sd) + dt(nu)
      })
    }

    expect_error(assertRxUiMixedOnly(one.cmt.pop))
    expect_error(assertRxUiPopulationOnly(one.cmt.pop), NA)
  })

  test_that("There must be at least one prediction assertion", {
    uif <- function() {
      ini({
        tka <- 4
        tcl <- exp(-3.2)
        tv <- exp(1)
        eta.ka ~ 0.1
        eta.cl ~ 0.2
      })
      model({
        ka <- tka + eta.ka
        cl <- tcl + eta.cl

        v <- tv
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
      })
    }

    tmp <- rxode2(uif)

    expect_error(
      assertRxUiPrediction(tmp),
      regexp="there must be at least one prediction"
    )
  })

  test_that("Transformably and non-transformably normal", {
    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
        lambda <- c(-2, 1, 2)
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd) + boxCox(lambda)
      })
    }

    expect_error(assertRxUiNormal(one.cmt))
    expect_error(assertRxUiTransformNormal(one.cmt), NA)
  })

  test_that("mu ref only", {
    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
        lambda <- c(-2, 1, 2)
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd) + boxCox(lambda)
      })
    }

    expect_error(assertRxUiMuRefOnly(one.cmt), NA)

    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
        lambda <- c(-2, 1, 2)
      })
      model({
        ka <- tka * exp(eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd) + boxCox(lambda)
      })
    }

    expect_error(assertRxUiMuRefOnly(one.cmt))

    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6 | occ
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    expect_error(
      assertRxUiRandomOnIdOnly(one.cmt),
      regexp = "can only have random effects on ID"
    )

  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {

    test_that("no warnings/errors without boundaries", {

      one.cmt <- function() {
        ini({
          tka <- 0.45; label("Ka")
          tcl <- log(2.7); label("Cl")
          tv <- 3.45; label("V")
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.sd <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          linCmt() ~ add(add.sd)
        })
      }

      expect_true(testRxUnbounded(one.cmt))
      expect_error(assertRxUnbounded(one.cmt), NA)
      expect_warning(warnRxBounded(one.cmt), NA)
    })
    test_that("assert or testRxLinCmt and boundaries", {

      one.cmt <- function() {
        ini({
          tka <- 0.45; label("Ka")
          tcl <- log(c(0, 2.7, 100)); label("Cl")
          tv <- 3.45; label("V")
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.sd <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          linCmt() ~ add(add.sd)
        })
      }

      expect_false(testRxUnbounded(one.cmt))
      expect_error(assertRxUnbounded(one.cmt))
      expect_warning(warnRxBounded(one.cmt))

      expect_error(assertRxLinCmt(one.cmt), NA)
      expect_true(testRxLinCmt(one.cmt))

      mod <- function() {
        ini({
          cl <- 1.1
          v <- 20
          ka <- 1.5
        })
        model({
          d/dt(depot) <- -ka*depot
          d/dt(central) <- ka*depot - (cl/v)*central
          f(central) <- bioav
          if (mode == 1) rate(central) <- rat2
          if (mode == 2) dur(central) <- dur2
          cp <- central/(v/1000)
        })
      }

      expect_error(assertRxLinCmt(mod))
      expect_false(testRxLinCmt(mod))

    })
    test_that("assertRxUiRandomOnIdOnly", {
      one.cmt <- function() {
        ini({
          tka <- 0.45; label("Ka")
          tcl <- log(c(0, 2.7, 100)); label("Cl")
          tv <- 3.45; label("V")
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.sd <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          linCmt() ~ add(add.sd)
        })
      }
      expect_equal(
        assertRxUiRandomOnIdOnly(one.cmt),
        as.rxUi(one.cmt)
      )
    })

    test_that("assert/test Compartment/Var", {

      mod <- function() {
        ini({
          cl <- 1.1
          v <- 20
          ka <- 1.5
        })
        model({
          d/dt(depot) <- -ka*depot
          d/dt(central) <- ka*depot - (cl/v)*central
          f(central) <- bioav
          if (mode == 1) rate(central) <- rat2
          if (mode == 2) dur(central) <- dur2
          cp <- central/(v/1000)
        })
      }

      mod <- mod()

      expect_equal(assertVariableExists(mod, c("ka", "cl")), "ka")

      expect_equal(assertVariableExists(mod, c("cl", "ka")), "cl")

      expect_error(assertVariableExists(mod, c("depot", "central")),
                   "variable 'depot', 'central' not in the model")

      expect_equal(assertCompartmentExists(mod, c("depot", "central")), "depot")
      expect_equal(assertCompartmentExists(mod, c("central", "depot")), "central")
      expect_equal(assertCompartmentExists(mod, c("ack", "central")), "central")

      expect_equal(assertExists(mod, c("depot", "ka")), "depot")

      expect_equal(assertExists(mod, c("ka", "depot")), "ka")

      expect_error(assertCompartmentExists(mod, "central"), NA)
      expect_error(assertCompartmentNew(mod, "central"),
                   "compartment 'central' already exists in the model")
      expect_true(testCompartmentExists(mod, "central"))
      expect_error(assertCompartmentExists(mod, "funny"),
                   "'funny' compartment is not in the model")
      expect_error(assertCompartmentNew(mod, "funny"), NA)
      expect_false(testCompartmentExists(mod, "funny"))
      expect_error(assertCompartmentExists(mod, funny),
                   "'funny' compartment is not in the model")
      expect_error(assertCompartmentNew(mod, funny), NA)
      expect_false(testCompartmentExists(mod, funny))
      expect_error(assertCompartmentExists(mod, central), NA)
      expect_error(assertCompartmentNew(mod, central),
                   "compartment 'central' already exists in the model")
      expect_true(testCompartmentExists(mod, central))

      # now variables
      expect_error(assertVariableExists(mod, "cp"), NA)
      expect_error(assertExists(mod, "cp"), NA)
      expect_true(testVariableExists(mod, "cp"))
      expect_true(testExists(mod, "cp"))
      expect_error(assertVariableNew(mod, "cp"),
                   "variable 'cp' is already in the model")

      expect_error(assertVariableExists(mod, cp), NA)
      expect_true(testVariableExists(mod, cp))
      expect_error(assertVariableNew(mod, cp),
                   "variable 'cp' is already in the model")

      expect_error(assertVariableExists(mod,"funny"),
                   "variable 'funny' not in the model")
      expect_false(testVariableExists(mod, "funny"))
      expect_error(assertVariableNew(mod, "funny"),NA)

      expect_error(assertVariableExists(mod,funny),
                   "variable 'funny' not in the model")
      expect_false(testVariableExists(mod, funny))
      expect_error(assertVariableNew(mod, funny),NA)
    })

    test_that("assertCompartmentName", {
      expect_equal(assertCompartmentName("x"), "x")
      expect_equal(assertCompartmentName("x.y"), "x.y")
      expect_equal(assertCompartmentName("x_y"), "x_y")
      # This is a valid R variable name, but it does not work with rxode2
      ## f <- rxode2("d/dt(.) = . + 3")
      ## rxode2 model syntax error:
      ## ===================================================
      ## rxode2 syntax error:
      ## :001: d/dt(.) = . + 3
      ##            ^
      expect_error(assertCompartmentName("."))
      expect_error(assertCompartmentName("9"))
      expect_error(assertCompartmentName(9))
      expect_error(assertCompartmentName(NULL))
      expect_error(assertCompartmentName(c("A", "B")))
    })

    test_that("assertVariableName", {
      expect_equal(assertVariableName("x"), "x")
      expect_equal(assertVariableName("x.y"), "x.y")
      expect_equal(assertVariableName("x_y"), "x_y")
      # This is a valid R variable name, but it does not work
      # with rxode2
      # rxode2 model syntax error:
      # ===================================================
      # rxode2 syntax error:
      # :001: . = 3
      #       ^

      expect_error(assertVariableName("."))
      expect_error(assertVariableName("9"))
      expect_error(assertVariableName(9))
      expect_error(assertVariableName(NULL))
      expect_error(assertVariableName(c("A", "B")))
    })

    test_that("assertParameterValue", {
      expect_equal(assertParameterValue(1), 1)
      expect_equal(assertParameterValue(-9), -9)
      expect_equal(assertParameterValue(0), 0)
      expect_error(assertParameterValue(Inf))
      expect_error(assertParameterValue(NA))
    })
  }

  test_that("no warnings/errors without boundaries", {

    one.cmt <- function() {
      ini({
        tka <- 0.45; label("Ka")
        tcl <- log(2.7); label("Cl")
        tv <- 3.45; label("V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        iov.cl ~ 0.1 | occ
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl + iov.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    expect_error(suppressWarnings(assertRxUiIovNoCor(one.cmt)), NA)

    one.cmt <- function() {
      ini({
        tka <- 0.45; label("Ka")
        tcl <- log(2.7); label("Cl")
        tv <- 3.45; label("V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        iov.cl ~ 0.1 | occ
        iov.v ~ 0.1 | occ
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl + iov.cl)
        v <- exp(tv + eta.v + iov.v)
        linCmt() ~ add(add.sd)
      })
    }

    expect_error(suppressWarnings(assertRxUiIovNoCor(one.cmt)), NA)

    one.cmt <- function() {
      ini({
        tka <- 0.45; label("Ka")
        tcl <- log(2.7); label("Cl")
        tv <- 3.45; label("V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        iov.cl + iov.v ~ c(0.1,
                           0.001, 0.1) | occ
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl + iov.cl)
        v <- exp(tv + eta.v + iov.v)
        linCmt() ~ add(add.sd)
      })
    }
    expect_error(suppressWarnings(assertRxUiIovNoCor(one.cmt)))

  })

})

#' Find the assignments in R expression
#'
#' @param x R expression
#' @return list of assigned parameters
#' @author Hadley Wickham and Matthew L. Fidler
#' @keywords internal
#' @export
findLhs <- function(x) {
  ## Modified from http://adv-r.had.co.nz/Expressions.html find_assign4
  if (is.atomic(x) || is.name(x)) {
    character()
  } else if (is.call(x)) {
    if ((identical(x[[1]], quote(`<-`)) ||
           identical(x[[1]], quote(`=`)) ||
           identical(x[[1]], quote(`~`))) &&
          is.name(x[[2]])) {
      .lhs <- as.character(x[[2]])
    } else {
      .lhs <- character()
    }
    unique(c(.lhs, unlist(lapply(x, rxode2::findLhs))))
  } else if (is.pairlist(x)) {
    unique(unlist(lapply(x, rxode2::findLhs)))
  } else {
    stop(sprintf("do not know how to handle type '%s'", typeof(x)),
      call. = FALSE
    )
  }
}

#' Get the linear compartment model true function
#'
#' @inheritParams rxode2
#' @return model with linCmt() replaced with linCmtA()
#' @author Matthew Fidler
#' @export
rxGetLin <- function(model, linCmtSens = c("linCmtA", "linCmtB"),
                     verbose = FALSE) {
  .mv <- rxGetModel(model) # nolint
  if (.Call(`_rxode2_isLinCmt`) == 1L) { # nolint
    .vars <- c(.mv$params, .mv$lhs, .mv$slhs)
    .Call(
      `_rxode2_linCmtGen`, # nolint
      length(.mv$state),
      .vars,
      setNames(
        c(
          "linCmtA" = 1L, "linCmtB" = 2L
        )[match.arg(linCmtSens)],
        NULL
      ), verbose)
  } else {
    model
  }
}
#' Check if an expression is a linCmt call
#'
#' @param expr expression to check (a line in the model)
#'
#' @return `TRUE` if the expression is a linCmt call, `FALSE`
#'   otherwise
#'
#' @noRd
#' @author Matthew L. Fidler
.isLinCmtCall <- function(expr) {
  if (!is.call(expr)) {
    return(FALSE)
  }
  if (!identical(expr[[1]], quote(`=`)) &&
        !identical(expr[[1]], quote(`<-`)) &&
        !identical(expr[[1]], quote(`~`))) {
    return(FALSE)
  }
  .rhs <- expr[[3]]
  is.call(.rhs) && (as.character(.rhs[[1]]) %in% c("linCmtA", "linCmtB"))
}
#' Linear compartment model to ODE model expression conversion
#'
#' @param ui rxUi model object
#'
#' @return list of R expressions representing the ODE model equivalent to the linCmt model
#' @noRd
#' @author Matthew L. Fidler
.linToOdeLinExpr <- function(ui) {
  if (is.null(ui$mvL)) {
    return(list())
  }
  # the mvL gives the linear compartment translation
  .mvLExpr <- as.list(str2lang(paste0("{",
                                      rxNorm(ui$mvL), # nolint
                                      "}")))[-1]
  .mvLExpr <- Filter(function(expr) {
    is.call(expr)
  }, .mvLExpr)
  Filter(.isLinCmtCall, .mvLExpr)
}
#' This converts the linCmtA/linCmtB
#'
#' @param expr the actual linCmtA or linCmtB expression, which is used
#'   to extract the parameters for the ODE translation
#' @return list of parameters for the ODE translation
#' @noRd
#' @author Matthew L. Fidler
.linToOdeArgs <- function(expr) {
  .rhs <- expr[[3]]
  .args <- as.list(.rhs)[-1]
  list(ncmt = as.integer(eval(.args[[4]], envir = baseenv())),
       oral0 = as.integer(eval(.args[[5]], envir = baseenv())),
       trans = as.integer(eval(.args[[7]], envir = baseenv())),
       p1 = .args[[8]],
       v1 = .args[[9]],
       p2 = .args[[10]],
       p3 = .args[[11]],
       p4 = .args[[12]],
       p5 = .args[[13]],
       ka = .args[[14]])
}
#' This builds the micro parameters for the one compartment model
#'
#' @param args arguments extracted from the linCmtA/linCmtB
#'   expression, which are used
#' @return list of parameters for the ODE translation
#' @noRd
#' @author Matthew L. Fidler
.linToOdeBuildMicro1 <- function(args) {
  .ret <- list(ncmt = 1L, oral0 = args$oral0, ka = args$ka)
  if (args$trans == 1L) {
    .ret$v <- args$v1
    .ret$k <- bquote(.(args$p1) / .(args$v1))
  } else if (args$trans %in% c(2L, 11L)) {
    .ret$v <- args$v1
    .ret$k <- args$p1
  } else if (args$trans == 10L) {
    .ret$v <- bquote(1 / .(args$v1))
    .ret$k <- args$p1
  } else {
    stop("unsupported 1-cmt linCmt translation", call. = FALSE)
  }
  .ret
}
#' This builds the micro parameters for the 2 compartment model
#'
#' @param the actual linCmtA or linCmtB expression, which is used
#'   to extract the parameters for the ODE translation
#'
#' @return list of parameters for the ODE translation
#'
#' @noRd
#'
#' @author Matthew L. Fidler
#'
.linToOdeBuildMicro2 <- function(args) {
  .ret <- list(ncmt = 2L, oral0 = args$oral0, ka = args$ka)
  if (args$trans == 1L) {
    .ret$v <- args$v1
    .ret$k <- bquote(.(args$p1) / .(args$v1))
    .ret$k12 <- bquote(.(args$p2) / .(args$v1))
    .ret$k21 <- bquote(.(args$p2) / .(args$p3))
  } else if (args$trans == 2L) {
    .ret$v <- args$v1
    .ret$k <- args$p1
    .ret$k12 <- args$p2
    .ret$k21 <- args$p3
  } else if (args$trans == 3L) {
    .ret$v <- args$v1
    .ret$k <- bquote(.(args$p1) / .(args$v1))
    .ret$k12 <- bquote(.(args$p2) / .(args$v1))
    .ret$k21 <- bquote(.(args$p2) / (.(args$p3) - .(args$v1)))
  } else if (args$trans == 4L) {
    .ret$v <- args$v1
    .ret$k21 <- args$p3
    .ret$k <- bquote((.(args$p1) * .(args$p2)) / .(.ret$k21))
    .ret$k12 <- bquote(.(args$p1) + .(args$p2) - .(.ret$k21) - .(.ret$k))
  } else if (args$trans == 5L) {
    .ret$v <- args$v1
    .ret$k21 <- bquote((.(args$p3) * .(args$p2) + .(args$p1)) / (.(args$p3) + 1))
    .ret$k <- bquote((.(args$p1) * .(args$p2)) / .(.ret$k21))
    .ret$k12 <- bquote(.(args$p1) + .(args$p2) - .(.ret$k21) - .(.ret$k))
  } else if (args$trans == 11L) {
    .A <- bquote(1 / .(args$v1)) # nolint
    .B <- args$p3 # nolint
    .alpha <- args$p1
    .beta <- args$p2
    .ret$v <- bquote(1 / (.(.A) + .(.B)))
    .ret$k21 <- bquote((.(.A) * .(.beta) + .(.B) * .(.alpha)) * .(.ret$v))
    .ret$k <- bquote((.(.alpha) * .(.beta)) / .(.ret$k21))
    .ret$k12 <- bquote(.(.alpha) + .(.beta) - .(.ret$k21) - .(.ret$k))
  } else if (args$trans == 10L) {
    .A <- args$v1 # nolint
    .B <- args$p3 # nolint
    .alpha <- args$p1
    .beta <- args$p2
    .ret$v <- bquote(1 / (.(.A) + .(.B)))
    .ret$k21 <- bquote((.(.A) * .(.beta) + .(.B) * .(.alpha)) * .(.ret$v))
    .ret$k <- bquote((.(.alpha) * .(.beta)) / .(.ret$k21))
    .ret$k12 <- bquote(.(.alpha) + .(.beta) - .(.ret$k21) - .(.ret$k))
  } else {
    stop("unsupported 2-cmt linCmt translation", call. = FALSE)
  }
  .ret
}
#' Build the micro parameters for the 3 compartment model
#'
#' @param args arguments extracted from the linCmtA/linCmtB
#'   expression, which are used
#'
#' @return list of parameters for the ODE translation
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.linToOdeBuildMicro3 <- function(args) {
  .ret <- list(ncmt = 3L, oral0 = args$oral0, ka = args$ka)
  if (args$trans == 1L) {
    .ret$v <- args$v1
    .ret$k <- bquote(.(args$p1) / .(args$v1))
    .ret$k12 <- bquote(.(args$p2) / .(args$v1))
    .ret$k21 <- bquote(.(args$p2) / .(args$p3))
    .ret$k13 <- bquote(.(args$p4) / .(args$v1))
    .ret$k31 <- bquote(.(args$p4) / .(args$p5))
  } else if (args$trans == 2L) {
    .ret$v <- args$v1
    .ret$k <- args$p1
    .ret$k12 <- args$p2
    .ret$k21 <- args$p3
    .ret$k13 <- args$p4
    .ret$k31 <- args$p5
  } else if (args$trans %in% c(10L, 11L)) {
    if (args$trans == 11) {
      .A <- bquote(1 / .(args$v1)) # nolint
    } else {
      .A <- args$v1 # nolint
    }
    .B <- args$p3 # nolint
    .C <- args$p5 # nolint
    .alpha <- args$p1
    .beta <- args$p2
    .gamma <- args$p4
    .ret$v <- bquote(1 / (.(.A) + .(.B) + .(.C)))
    .btemp <- bquote(-(.(.alpha) * .(.C) + .(.alpha) * .(.B) +
                         .(.gamma) * .(.A) + .(.gamma) * .(.B) +
                         .(.beta) * .(.A) + .(.beta) * .(.C)) * .(.ret$v))
    .ctemp <- bquote((.(.alpha) * .(.beta) * .(.C) +
                        .(.alpha) * .(.gamma) * .(.B) +
                        .(.beta) * .(.gamma) * .(.A)) * .(.ret$v))
    .dtemp <- bquote(sqrt((.(.btemp)) * (.(.btemp)) - 4 * (.(.ctemp))))
    .ret$k21 <- bquote(0.5 * (-(.(.btemp)) + .(.dtemp)))
    .ret$k31 <- bquote(0.5 * (-(.(.btemp)) - .(.dtemp)))
    .ret$k <- bquote((.(.alpha) * .(.beta) * .(.gamma)) / .(.ret$k21) / .(.ret$k31))
    .ret$k12 <- bquote(((.(.beta) * .(.gamma) + .(.alpha) * .(.beta) + .(.alpha) * .(.gamma)) -
                          .(.ret$k21) * (.(.alpha) + .(.beta) + .(.gamma)) -
                          .(.ret$k) * .(.ret$k31) + .(.ret$k21) * .(.ret$k21)) /
                         (.(.ret$k31) - .(.ret$k21)))
    .ret$k13 <- bquote(.(.alpha) + .(.beta) + .(.gamma) -
                         (.(.ret$k) + .(.ret$k12) + .(.ret$k21) + .(.ret$k31)))
  } else {
    stop("unsupported 3-cmt linCmt translation", call. = FALSE)
  }
  .ret
}
#' Return the micro-constants for the ODE translation of a linCmt model
#'
#' @param expr the actual linCmtA or linCmtB expression, which is used
#'
#' @return list of micro-constants for the ODE translation
#' @noRd
#' @author Matthew L. Fidler
.linToOdeBuildMicro <- function(expr) {
  .args <- .linToOdeArgs(expr)
  if (.args$ncmt == 1L) {
    .linToOdeBuildMicro1(.args)
  } else if (.args$ncmt == 2L) {
    .linToOdeBuildMicro2(.args)
  } else if (.args$ncmt == 3L) {
    .linToOdeBuildMicro3(.args)
  } else {
    stop("unsupported linCmt compartment count", call. = FALSE)
  }
}
#' convert the predDf linCmt() line to the ODE equivalent
#'
#' @param ui rxode2 ui model
#' @param i line number to check for linCmt() call
#' @return data frame row with the ODE equivalent of the linCmt() call, or `NULL`
#' @noRd
#' @author Matthew L. Fidler
.linToOdePredLine <- function(ui, i) {
  if (is.null(ui$predDf)) {
    return(NULL)
  }
  .w <- which(ui$predDf$line == i)
  if (length(.w) == 1L) {
    ui$predDf[.w, , drop = FALSE]
  } else {
    NULL
  }
}
#' Convert the micro-constants to ODE lines for the ODE translation of a linCmt model
#'
#' @param micro list of the micro constant
#' @return a list of R expressions for the ode model
#' @noRd
#' @author Matthew L. Fidler
.linToOdeOdeLines <- function(micro) {
  .ret <- list()
  .input <- "0"
  .kTxt <- paste0("(", deparse1(micro$k), ")")
  if (micro$oral0 == 1) {
    .kaTxt <- paste0("(", deparse1(micro$ka), ")")
    .ret[[length(.ret) + 1L]] <- str2lang(paste0("d/dt(depot) <- -", .kaTxt, " * depot"))
    .input <- paste0(.kaTxt, " * depot")
  }
  if (micro$ncmt == 1) {
    .ret[[length(.ret) + 1L]] <- str2lang(paste0("d/dt(central) <- ", .input,
                                                 " - ", .kTxt, " * central"))
  } else if (micro$ncmt == 2L) {
    .k12Txt <- paste0("(", deparse1(micro$k12), ")")
    .k21Txt <- paste0("(", deparse1(micro$k21), ")")
    .ret[[length(.ret) + 1L]] <- str2lang(paste0("d/dt(central) <- ", .input,
                                                 " - ", .kTxt, " * central - ", .k12Txt,
                                                 " * central + ", .k21Txt, " * peripheral1"))
    .ret[[length(.ret) + 1L]] <- str2lang(paste0("d/dt(peripheral1) <- ", .k12Txt,
                                                 " * central - ", .k21Txt, " * peripheral1"))
  } else if (micro$ncmt == 3L) {
    .k12Txt <- paste0("(", deparse1(micro$k12), ")")
    .k21Txt <- paste0("(", deparse1(micro$k21), ")")
    .k13Txt <- paste0("(", deparse1(micro$k13), ")")
    .k31Txt <- paste0("(", deparse1(micro$k31), ")")
    .ret[[length(.ret) + 1L]] <- str2lang(paste0("d/dt(central) <- ", .input,
                                                 " - ", .kTxt, " * central - ", .k12Txt,
                                                 " * central + ", .k21Txt, " * peripheral1 - ",
                                                 .k13Txt, " * central + ", .k31Txt, " * peripheral2"))
    .ret[[length(.ret) + 1L]] <- str2lang(paste0("d/dt(peripheral1) <- ", .k12Txt,
                                                 " * central - ", .k21Txt, " * peripheral1"))
    .ret[[length(.ret) + 1L]] <- str2lang(paste0("d/dt(peripheral2) <- ", .k13Txt,
                                                 " * central - ", .k31Txt, " * peripheral2"))
  }
  .ret
}
#' Convert the micro expressions to lhs values for ode solving
#'
#' @param expr linCmt expression
#' @param predLine the pred line for the linCmt() model
#' @param micro list of micro constants
#' @return a list of lhs expressions for ode conversion
#' @noRd
#' @author Matthew L. Fidler
.linToOdeLhsLines <- function(expr, predLine, micro) {
  if (!is.null(predLine) && isTRUE(predLine$linCmt)) {
    .lhs <- predLine$var
  } else {
    .lhs <- as.character(expr[[2]])
  }
  .lhsExpr <- str2lang(.lhs)
  .vExpr <- str2lang(paste0("(", deparse1(micro$v), ")"))
  .ret <- list(call("<-", .lhsExpr, call("/", str2lang("central"), .vExpr)))
  if (!is.null(predLine) && isTRUE(predLine$linCmt)) {
    .ret[[length(.ret) + 1L]] <- str2lang(sub("linCmt\\s*\\(\\s*\\)",
                                              .lhs, deparse1(expr), perl = TRUE))
  }
  .ret
}
#' Render the linCmt() system as an ode system
#'
#' @param expr  linCmt expression
#' @param predLine predicion line data frame
#' @param mvExpr linCmtA/linCmtB expression
#' @return nothing, called for side effects
#' @noRd
#' @author Matthew L. Fidler
.linToOdeRender <- function(expr, predLine, mvExpr) {
  .micro <- .linToOdeBuildMicro(mvExpr)
  c(.linToOdeOdeLines(.micro), .linToOdeLhsLines(expr, predLine, .micro))
}
#' Convert the linear compartment models to ode expressions.
#'
#' @param ui rxUi model object
#' @return list of R expressions representing the ODE model equivalent to the linCmt model
#' @noRd
#' @author Matthew L. Fidler
.linToOdeExpr <- function(ui) {
  .linExpr <- .linToOdeLinExpr(ui)
  if (length(.linExpr) == 0L) {
    return(ui$lstExpr)
  }
  .linCur <- 1L
  .ret <- list()
  for (i in seq_along(ui$lstExpr)) {
    .expr <- ui$lstExpr[[i]]
    .predLine <- .linToOdePredLine(ui, i)
    if (regexpr("linCmt\\s*\\(", deparse1(.expr), perl = TRUE) != -1) {
      .ret <- c(.ret, .linToOdeRender(.expr, .predLine, .linExpr[[.linCur]]))
      .linCur <- .linCur + 1L
    } else {
      .ret[[length(.ret) + 1L]] <- .expr
    }
  }
  .ret
}

#' Convert linCmt rxUi models to ODE rxUi models
#'
#' @param ui rxUi-like model object
#'
#' @return rxUi model with `linCmt()` translated to explicit ODEs
#' @examples
#' oneCmt <- function() {
#'   ini({
#'     tka <- 0.45
#'     tcl <- log(2.7)
#'     tv <- 3.45
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- exp(tka)
#'     cl <- exp(tcl)
#'     v <- exp(tv)
#'     cp <- linCmt()
#'     cp ~ add(add.sd)
#'   })
#' }
#'
#' oneCmtOde <- linToOde(oneCmt)
#'
#'
#'
#' pkpd <- function() {
#'   ini({
#'     ka <- 1
#'     cl <- 2
#'     v <- 20
#'     kin <- 1
#'     kout <- 1
#'     ec50 <- 2
#'   })
#'   model({
#'     cp <- linCmt(ka, cl, v)
#'     eff(0) <- 1
#'     d/dt(eff) <- kin - kout * (1 - cp/(ec50 + cp)) * eff
#'   })
#' }
#'
#' pkpdOde <- linToOde(pkpd)
#'
#' @author Matthew Fidler
#' @export
linToOde <- function(ui) {
  .ui <- as.rxUi(ui) # nolint
  .ui <- rxUiDecompress(.ui) # nolint
  .expr <- .linToOdeExpr(.ui)
  if (identical(.expr, .ui$lstExpr)) {
    return(rxUiCompress(.ui)) # nolint
  }
  .ls <- ls(.ui$meta, all.names=TRUE)
  .ret <- vector("list", length(.ls) + ifelse(length(.ui$iniDf$cond) > 0, 3, 2))
  .ret[[1]] <- quote(`{`)
  for (.i in seq_along(.ls)) {
    .var <- .ls[.i]
    .ret[[.i + 1]] <- rxUiDeparse(.ui$meta[[.var]], .var) # nolint
  }
  .len <- length(.ls)
  if (length(.ui$iniDf$cond) > 0) {
    .ret[[.len + 2]] <- .ui$iniFun
    .ret[[.len + 3]] <- bquote(model(.(as.call(c(quote(`{`), .expr)))))
  } else {
    .ret[[.len + 2]] <- bquote(model(.(as.call(c(quote(`{`), .expr)))))
  }
  .fun <- function() {
  }
  body(.fun) <- as.call(.ret)
  if (is.function(.ui$model)) {
    environment(.fun) <- environment(.ui$model)
  }
  suppressMessages(as.rxUi(.fun)) # nolint
}

#' Parameters specified by the model
#'
#' This returns the model's parameters that are required to solve the
#' ODE system, and can be used to pipe parameters into an rxode2 solve
#'
#' @inheritParams rxModelVars
#'
#' @param constants is a boolean indicting if constants should be
#'     included in the list of parameters. Currently rxode2 parses
#'     constants into variables in case you wish to change them
#'     without recompiling the rxode2 model.
#'
#' @inheritParams rxControl
#'
#' @return When extracting the parameters from an rxode2 model, a
#'     character vector listing the parameters in the model.
#'
#' @author Matthew L.Fidler
#' @family Query model information
#' @export
rxParams <- function(obj, ...) {
  UseMethod("rxParams")
}

#' @rdname rxParams
#' @export
rxParams.rxode2 <- function(obj, constants = TRUE, ...,
                           params = NULL, inits = NULL, iCov = NULL,
                           keep = NULL,
                           thetaMat = NULL,
                           omega = NULL, dfSub = NULL,
                           sigma = NULL, dfObs = NULL,
                           nSub = NULL, nStud = NULL) {
  if (!is.null(iCov)) {
    stop("'iCov' in a pipline is no longer supported", call. = FALSE)
  }
  .ret <- list(
    params = params, inits = inits, keep = keep,
    thetaMat = thetaMat,
    omega = omega, dfSub = dfSub,
    sigma = sigma, dfObs = dfObs,
    nSub = nSub, nStud = nStud
  )
  if (all(sapply(seq_along(.ret), function(x) {
    is.null(.ret[[x]])
  }))) {
    if (length(list(...)) > 0) {
      .clearPipe()
      .asFunctionEnv$rx <- NULL
      stop("unknown arguments in 'rxParams'", call. = FALSE)
    }
    return(rxParams.default(obj, constants = constants))
  } else {
    .lst <- list(...)
    if (length(.lst) > 0) {
      .clearPipe()
      .asFunctionEnv$rx <- NULL
      stop(sprintf(
        gettext("unknown arguments in 'rxParams': %s\ntry piping to 'rxSolve'"),
        paste(names(.lst), collapse = ", ")
      ), call. = FALSE)
    }
    ## Most likely
    ## rxode2() |> rxParams() |>
    rxode2::.pipeRx(obj)
    rxode2::.pipeInits(NULL)
    rxode2::.pipeEvents(NULL)
    rxode2::.pipeParams(NULL)
    rxode2::.pipeKeep(NULL)
    rxode2::.pipeThetaMat(NULL)
    rxode2::.pipeOmega(NULL)
    rxode2::.pipeSigma(NULL)
    rxode2::.pipeDfObs(NULL)
    rxode2::.pipeDfSub(NULL)
    rxode2::.pipeNSub(NULL)
    rxode2::.pipeNStud(NULL)
    class(.ret) <- "rxParams"
    return(.ret)
  }
}

#' @rdname rxParams
#' @export
rxParams.rxSolve <- function(obj, constants = TRUE, ...,
                             params = NULL, inits = NULL, iCov = NULL,
                             keep = NULL,
                             thetaMat = NULL,
                             omega = NULL, dfSub = NULL,
                             sigma = NULL, dfObs = NULL,
                             nSub = NULL, nStud = NULL) {
  if (!is.null(iCov)) {
    stop("'iCov' in a pipline is no longer supported", call. = FALSE)
  }
  .ret <- list(
    params = params, inits = inits, keep = keep,
    thetaMat = thetaMat,
    omega = omega, dfSub = dfSub,
    sigma = sigma, dfObs = dfObs,
    nSub = nSub, nStud = nStud
  )
  if (all(sapply(seq_along(.ret), function(x) {
    is.null(.ret[[x]])
  }))) {
    if (length(list(...)) > 0) {
      .clearPipe()
      .asFunctionEnv$rx <- NULL
      stop("unknown arguments in 'rxParams'", call. = FALSE)
    }
    return(rxParams.default(obj, constants = constants))
  } else {
    .lst <- list(...)
    if (length(.lst) > 0) {
      .clearPipe()
      .asFunctionEnv$rx <- NULL
      stop(sprintf(
        gettext("unknown arguments in 'rxParams': %s\ntry piping to 'rxSolve'"),
        paste(names(.lst), collapse = ", ")
      ),
      call. = FALSE
      )
    }
    ## Most likely
    ## solveObject |> rxParams() |>
    .x <- obj
    ## Assign prior information
    ## Need to extract:
    ## 1. rxode2 model
    rxode2::.pipeRx(.x$.args.object)
    ## Events
    rxode2::.pipeEvents(.x$.args.events)
    ## 2. rxode2 parameters
    rxode2::.pipeParams(.x$.args.par0)
    ## 3. rxode2 inits
    rxode2::.pipeInits(.x$.args.inits)
    ## 4. rxode2 thetaMat
    rxode2::.pipeThetaMat(.x$.args$thetaMat)
    ## 5. rxode2 omega
    rxode2::.pipeOmega(.x$.args$omega)
    ## 6. rxode2 sigma
    rxode2::.pipeSigma(.x$.args$sigma)
    ## 7. rxode2 dfObs
    rxode2::.pipeDfObs(.x$env$.args$dfObs)
    ## 8. rxode2 dfSub
    rxode2::.pipeDfSub(.x$env$.args$dfSub)
    class(.ret) <- "rxParams"
    return(.ret)
  }
}

#' @rdname rxParams
#' @export
rxParams.rxEt <- function(obj, ...,
                          params = NULL, inits = NULL, iCov = NULL,
                          keep = NULL,
                          thetaMat = NULL,
                          omega = NULL, dfSub = NULL,
                          sigma = NULL, dfObs = NULL,
                          nSub = NULL, nStud = NULL) {
  if (!is.null(iCov)) {
    stop("'iCov' in a pipline is no longer supported", call. = FALSE)
  }
  # et() |> rxParams() |>
  rxode2::.pipeEvents(obj)
  .lst <- list(...)
  if (length(.lst) > 0) {
    .clearPipe()
    .asFunctionEnv$rx <- NULL
    stop(sprintf(gettext("unknown arguments in 'rxParams': %s\ntry piping to 'rxSolve'"), paste(names(.lst), collapse = ", ")),
      call. = FALSE
    )
  }
  .ret <- list(
    params = params, inits = inits, keep = keep,
    thetaMat = thetaMat, omega = omega, dfSub = dfSub,
    sigma = sigma, dfObs = dfObs,
    nSub = nSub, nStud = nStud
  )
  class(.ret) <- "rxParams"
  return(.ret)
}

.rxParams <- function(obj, constants = TRUE) {
  .ret <- rxParams_(obj)
  if (!constants) {
    .init <- rxode2::rxInit(obj)
    .ret <- .ret[!(.ret %in% names(.init))]
  }
  return(.ret)
}
#' @export
rxParams.default <- function(obj, ..., constants = TRUE) {
  if (!missing(obj)) {
    return(.rxParams(obj, constants))
  } else {
    .lst <- list(...)
    .nm <- c(
      "cov", "params", "inits", "keep",
      "thetaMat", "omega", "dfSub",
      "sigma", "dfObs", "nSub", "nStud"
    )
    .ret <- lapply(.nm, function(x) {
      return(.lst[[x]])
    })
    names(.ret) <- .nm
    class(.ret) <- "rxParams"
    return(.ret)
  }
}

#' @rdname rxParams
#' @export
rxParam <- rxParams

# rxEt environment and chunk helpers -------------------------------------


#' Is the homogenous event table path active?
#'
#'
#' @return logical; TRUE if homogenous event table path is active,
#'   FALSE otherwise, gets from option "rxode2.homogenous" with
#'   default TRUE
#' @noRd
#' @author Matthew L. Fidler
.rxGetHomogenous <- function() {
  isTRUE(getOption("rxode2.homogenous", TRUE))
}

#' Extract mutable .env from either new-style (data.frame subclass + .rxEtEnv attr)
#' or internal mini-rxEt (1-element named list where [[1L]] is the env).
#'
#' @param x object to extract data from
#'
#' @return environment or NULL if not found
#'
#' @noRd
.rxEtEnv <- function(x) {
  .e <- attr(x, ".rxEtEnv", exact = TRUE)
  if (!is.null(.e)) return(.e)
  .ux <- unclass(x)
  if (is.list(.ux) && ".env" %in% names(.ux)) return(.ux[[".env"]])
  if (length(.ux) >= 1L && is.environment(.ux[[1L]])) return(.ux[[1L]])
  NULL
}

#' Default show flags (controls which columns print/as.data.frame expose)
#'
#' This is kept here so that it can be specified once and referred to
#' by multiple locations.
#'
#' @return named logical vector of default show flags
#'
#' @noRd
.etDefaultShow <- function() {
  c(id = FALSE, low = FALSE, time = TRUE, high = FALSE, cmt = FALSE,
    amt = FALSE, rate = FALSE, ii = FALSE, addl = FALSE, evid = TRUE,
    ss = FALSE, dur = FALSE)
}
#' Drop the units for the check
#'
#' @param df NULL, data.frame or non-data frame object.
#'   from for internal use
#'
#' @return this function returns NULL or the non-materialized data, or
#'   the data.frame with units columns converted to numeric for
#'   internal use
#'
#' @noRd
#' @author Matthew L. Fidler
.etDropUnitsForChunk <- function(df) {
  if (is.null(df) || !is.data.frame(df)) {
    return(df)
  }
  for (.nm in names(df)) {
    if (inherits(df[[.nm]], "units")) {
      df[[.nm]] <- as.numeric(df[[.nm]])
    }
  }
  df
}
#' Get the groups from the rxEt environment
#'
#' @param envRef The environmental reference for the rxEt groups
#' @return A list of groups, where each group is a list containing 'ids' and 'data'
#' @noRd
#' @author Matthew L. Fidler
.etGroups <- function(envRef) {
  .groups <- envRef$groups
  if (is.null(.groups)) .groups <- list()
  .groups
}
#' Check if two sets of group IDs are equal
#'
#' @param x first group to check if it is equal.
#' @param y second group to check if it is equal.
#' @return logical value indicating whether the two sets of group IDs are equal
#' @noRd
#' @author Matthew L. Fidler
.etGroupIdsEqual <- function(x, y) {
  identical(sort.int(as.integer(x), method = "quick"),
            sort.int(as.integer(y), method = "quick"))
}
#' Process a chunk of data for grouping
#'
#' This function takes a data frame and an optional set of IDs, and
#' processes the data frame for grouping.
#'
#'
#' @param df data frame to process for grouping
#' @param ids optional vector of IDs to consider for grouping; if NULL, the function will
#' @return processed data frame
#' @noRd
#' @author Matthew L. Fidler
.etGroupChunk <- function(df, ids = NULL) {
  if (!is.data.frame(df)) {
    df <- .etExpandObsChunk(df)
  }
  df <- .etDropUnitsForChunk(df)
  if (!is.null(ids) && "id" %in% names(df)) {
    .id <- unique(as.integer(df$id))
    if (length(.id) == 1L && length(ids) >= 1L) {
      df$id <- NULL
    }
  }
  df
}

#' Get the groups from the rxEt environment, with fallback to chunks
#' if groups are not set
#'
#' @param envRef The environment reference group
#' @return A list of groups, where each group is a list containing
#'   'ids' and 'data'. If groups are not set, it will return an empty
#'   list or process the chunks to create groups.
#' @noRd
#' @author Matthew L. Fidler
.etGetGroups <- function(envRef) {
  .groups <- .etGroups(envRef)
  if (length(.groups) > 0L) return(.groups)

  .chunks <- envRef$chunks
  if (length(.chunks) == 0L) return(list())

  .ret <- vector("list", 0L)
  for (.i in seq_along(.chunks)) {
    .chunk <- .chunks[[.i]]
    if (is.null(.chunk)) next
    .ids <- if (!is.null(names(.chunk)) && "id" %in% names(.chunk)) {
      unique(as.integer(.chunk$id))
    } else if (length(envRef$ids) == 1L) {
      as.integer(envRef$ids)
    } else {
      as.integer(.i)
    }
    .df <- .etGroupChunk(.chunk, .ids)
    .ret[[length(.ret) + 1L]] <- list(ids = .ids, data = .df)
  }
  .ret
}

.etSetGroups <- function(envRef, groups) {
  envRef$groups <- groups
  if (length(groups) > 0L) {
    envRef$chunks <- list()
  }
  invisible(NULL)
}

.etPresentIds <- function(envRef) {
  .groups <- .etGroups(envRef)
  if (length(.groups) > 0L) {
    return(sort(unique(as.integer(unlist(lapply(.groups, `[[`, "ids"), use.names = FALSE)))))
  }
  .chunks <- envRef$chunks
  if (length(.chunks) == 0L) return(integer(0))
  .ids <- integer(0)
  for (.i in seq_along(.chunks)) {
    .chunk <- .chunks[[.i]]
    if (is.null(.chunk)) next
    if (!is.null(names(.chunk)) && "id" %in% names(.chunk)) {
      .ids <- c(.ids, unique(as.integer(.chunk$id)))
    } else if (length(envRef$ids) == 1L) {
      .ids <- c(.ids, as.integer(envRef$ids))
    } else {
      .ids <- c(.ids, as.integer(.i))
    }
  }
  sort(unique(.ids))
}

#' Per-row id column of the materialized table, without materializing
#'
#' Mirrors the row counts of .etMaterialize() (obs-chunk expansion plus
#' the windowed-addl expansion of .etExpandAddl(windows=TRUE)) but never
#' draws window times, so reading ev$id does not touch the RNG (#1154).
#'
#' @param et rxEt object
#' @return integer vector matching .etMaterialize(et)$id
#' @noRd
.etIdColumn <- function(et) {
  .env <- .rxEtEnv(et)
  .groups <- .etGetGroups(.env)
  if (length(.groups) == 0L) return(integer(0))
  .ids <- integer(0)
  .counts <- integer(0)
  for (.g in .groups) {
    .df <- .g$data
    if (!is.data.frame(.df)) .df <- .etExpandObsChunk(.df)
    if (!is.data.frame(.df) || nrow(.df) == 0L) next
    # rows added by the windowed-addl expansion in .etMaterialize
    .extra <- if (is.null(.df$evid) || is.null(.df$addl) || is.null(.df$low)) {
      rep.int(0L, nrow(.df))
    } else {
      ifelse(!is.na(.df$evid) & .df$evid != 0L &
               !is.na(.df$addl) & .df$addl > 0L & !is.na(.df$low),
             as.integer(.df$addl), 0L)
    }
    if (!is.null(.df$id)) {
      .rowIds <- as.integer(.df$id)
      .tab <- tapply(1L + .extra, .rowIds, sum)
      .ids <- c(.ids, as.integer(names(.tab)))
      .counts <- c(.counts, as.integer(.tab))
    } else {
      .n <- nrow(.df) + sum(.extra)
      .gIds <- as.integer(.g$ids)
      .ids <- c(.ids, .gIds)
      .counts <- c(.counts, rep.int(as.integer(.n), length(.gIds)))
    }
  }
  if (length(.ids) == 0L) return(integer(0))
  # total rows per unique id, in ascending id order (materialize sorts by id)
  .tot <- vapply(split(.counts, .ids), sum, integer(1))
  .uid <- as.integer(names(.tot))
  .ord <- order(.uid)
  rep(.uid[.ord], .tot[.ord])
}

.etResetCountsFromGroups <- function(envRef) {
  .groups <- .etGetGroups(envRef)
  if (length(.groups) == 0L) {
    envRef$nobs <- 0L
    envRef$ndose <- 0L
    return(invisible(NULL))
  }
  envRef$nobs <- as.integer(sum(vapply(.groups, function(.g) {
    sum(.g$data$evid == 0L, na.rm = TRUE) * length(.g$ids)
  }, numeric(1))))
  envRef$ndose <- as.integer(sum(vapply(.groups, function(.g) {
    sum(.g$data$evid != 0L, na.rm = TRUE) * length(.g$ids)
  }, numeric(1))))
  invisible(NULL)
}

.etChunkIndex <- function(chunks, id) {
  .idx <- as.integer(id)
  if (is.na(.idx) || .idx < 1L) {
    stop("invalid event table id", call. = FALSE)
  }
  .len <- length(chunks)
  if (.idx > .len) {
    length(chunks) <- .idx
  }
  chunks
}

.etSplitGroupToChunks <- function(chunks, group, randomType = NA_integer_) {
  .df <- .etDropUnitsForChunk(group$data)
  if (!is.data.frame(.df) || nrow(.df) == 0L || length(group$ids) == 0L) {
    return(chunks)
  }
  .hasWin <- !is.na(.df$low) & !is.na(.df$high)
  for (.id in as.integer(group$ids)) {
    .row <- .df
    if (any(.hasWin)) {
      if (!is.na(randomType) && randomType == 3L) {
        .row$time[.hasWin] <- stats::rnorm(sum(.hasWin), .row$low[.hasWin], .row$high[.hasWin])
      } else if (!is.na(randomType) && randomType == 2L) {
        .row$time[.hasWin] <- stats::runif(sum(.hasWin), .row$low[.hasWin], .row$high[.hasWin])
      }
    }
    .row$id <- .id
    chunks <- .etChunkIndex(chunks, .id)
    chunks[[.id]] <- .row
  }
  chunks
}

.etSimulateRepresentation <- function(envRef) {
  .groups <- .etGroups(envRef)
  .chunks <- envRef$chunks
  .randomType <- envRef$randomType
  .hasWin <- FALSE
  .needsChunkSplit <- length(.chunks) > 0L

  if (length(.groups) > 0L) {
    for (.g in .groups) {
      .df <- .etDropUnitsForChunk(.g$data)
      if (is.data.frame(.df) && nrow(.df) > 0L) {
        .groupHasWin <- !is.na(.df$low) & !is.na(.df$high)
        if (any(.groupHasWin)) {
          .hasWin <- TRUE
          if (length(.g$ids) > 1L) {
            .needsChunkSplit <- TRUE
          }
        }
      }
    }
  }

  if (!.hasWin && length(.chunks) > 0L) {
    for (.i in seq_along(.chunks)) {
      .df <- .etDropUnitsForChunk(.chunks[[.i]])
      if (is.data.frame(.df) && nrow(.df) > 0L && any(!is.na(.df$low) & !is.na(.df$high))) {
        .hasWin <- TRUE
        break
      }
    }
  }

  if (!.hasWin) {
    return(list(hasWin = FALSE, groups = .groups, chunks = .chunks))
  }

  if (!is.na(.randomType) && .randomType == 1L) {
    return(list(hasWin = FALSE, groups = .groups, chunks = .chunks))
  }

  if (.needsChunkSplit) {
    .outChunks <- vector("list", max(length(.chunks), suppressWarnings(max(as.integer(envRef$ids), 0L)), na.rm = TRUE))
    if (length(.groups) > 0L) {
      for (.g in .groups) {
        .outChunks <- .etSplitGroupToChunks(.outChunks, .g, .randomType)
      }
    }
    if (length(.chunks) > 0L) {
      for (.i in seq_along(.chunks)) {
        .df <- .etDropUnitsForChunk(.chunks[[.i]])
        if (!is.data.frame(.df) || nrow(.df) == 0L) next
        .hasWinI <- !is.na(.df$low) & !is.na(.df$high)
        if (any(.hasWinI)) {
          if (!is.na(.randomType) && .randomType == 3L) {
            .df$time[.hasWinI] <- stats::rnorm(sum(.hasWinI), .df$low[.hasWinI], .df$high[.hasWinI])
          } else if (!is.na(.randomType) && .randomType == 2L) {
            .df$time[.hasWinI] <- stats::runif(sum(.hasWinI), .df$low[.hasWinI], .df$high[.hasWinI])
          }
        }
        .id <- if ("id" %in% names(.df)) unique(as.integer(.df$id))[1L] else as.integer(.i)
        .outChunks <- .etChunkIndex(.outChunks, .id)
        .outChunks[[.id]] <- .df
      }
    }
    return(list(hasWin = TRUE, groups = list(), chunks = .outChunks))
  }

  .outGroups <- lapply(.groups, function(.g) {
    .df <- .etDropUnitsForChunk(.g$data)
    .hasWinG <- !is.na(.df$low) & !is.na(.df$high)
    if (any(.hasWinG)) {
      if (!is.na(.randomType) && .randomType == 3L) {
        .df$time[.hasWinG] <- stats::rnorm(sum(.hasWinG), .df$low[.hasWinG], .df$high[.hasWinG])
      } else if (!is.na(.randomType) && .randomType == 2L) {
        .df$time[.hasWinG] <- stats::runif(sum(.hasWinG), .df$low[.hasWinG], .df$high[.hasWinG])
      }
    }
    list(ids = .g$ids, data = .df)
  })
  list(hasWin = TRUE, groups = .outGroups, chunks = list())
}

.formatIds <- function(ids) {
  ids <- sort(unique(as.integer(ids)))
  if (length(ids) == 0L) return(NA_character_)
  if (length(ids) == 1L) return(ids)
  if (all(diff(ids) == 1L)) {
    return(paste0(ids[1L], ":", ids[length(ids)]))
  }
  paste(ids, collapse = ",")
}

.etPreviewData <- function(envRef, subset = c("all", "dosing", "sampling")) {
  subset <- match.arg(subset)
  .groups <- .etGroups(envRef)
  if (length(.groups) == 0L) {
    .mat <- .etMaterialize(structure(list(env = envRef), class = "rxEt"))
    if (nrow(.mat) == 0L) return(NULL)
    if (subset == "dosing") {
      .mat <- .mat[.mat$evid != 0L, , drop = FALSE]
    } else if (subset == "sampling") {
      .mat <- .mat[.mat$evid == 0L, , drop = FALSE]
    }
    if (nrow(.mat) == 0L) return(NULL)
    rownames(.mat) <- seq_len(nrow(.mat))
    return(.mat)
  }

  .ret <- vector("list", 0L)
  .meta <- vector("list", 0L)
  for (.i in seq_along(.groups)) {
    .g <- .groups[[.i]]
    .df <- .etDropUnitsForChunk(.g$data)
    if (subset == "dosing") {
      .df <- .df[.df$evid != 0L, , drop = FALSE]
    } else if (subset == "sampling") {
      .df <- .df[.df$evid == 0L, , drop = FALSE]
    }
    if (nrow(.df) == 0L) next
    if (subset == "all") {
      .formattedId <- if (length(.g$ids) == 1L) as.integer(.g$ids) else .formatIds(.g$ids)
      .df <- cbind(id = .formattedId, .df)
    }
    .ret[[length(.ret) + 1L]] <- .df
    .meta[[length(.meta) + 1L]] <- list(
      ids = as.integer(.g$ids),
      nSub = length(.g$ids),
      nRow = nrow(.df)
    )
  }
  if (length(.ret) == 0L) return(NULL)
  .out <- as.data.frame(data.table::rbindlist(.ret, fill = TRUE))
  rownames(.out) <- seq_len(nrow(.out))
  attr(.out, "rxEtPreviewGroups") <- .meta
  attr(.out, "rxEtShow") <- envRef$show
  attr(.out, "rxEtExtraCols") <- .etExtraCols(envRef) # nolint
  attr(.out, "rxEtMarkedCols") <- names(.out)
  .tu <- envRef$units["time"]
  if (!is.na(.tu) && nchar(.tu) > 0 && requireNamespace("units", quietly = TRUE)) {
    for (.col in c("time", "ii", "low", "high", "dur")) {
      if (.col %in% names(.out)) {
        .out[[.col]] <- units::set_units(.out[[.col]], .tu, mode = "standard")
      }
    }
  }
  # class last: `[[<-` on a marked frame drops the display marking
  class(.out) <- c("rxEtPreview", class(.out))
  .out
}

.etMaterializeGroup <- function(group) {
  .df <- .etDropUnitsForChunk(group$data)
  if (!is.data.frame(.df) || nrow(.df) == 0L || length(group$ids) == 0L) {
    return(.etEmptyDf())
  }
  .n <- nrow(.df)
  .idx <- rep(seq_len(.n), times = length(group$ids))
  .id <- rep(as.integer(group$ids), each = .n)
  .ret <- .df[.idx, , drop = FALSE]
  .ret$id <- .id
  .ret
}

.etShiftChunk <- function(df, timeDelta) {
  .df <- .etDropUnitsForChunk(df)
  if (!is.data.frame(.df) || nrow(.df) == 0L || identical(timeDelta, 0)) {
    return(.df)
  }
  .df$time <- .df$time + timeDelta
  if (!is.null(.df$low) && any(!is.na(.df$low))) {
    .df$low[!is.na(.df$low)] <- .df$low[!is.na(.df$low)] + timeDelta
  }
  if (!is.null(.df$high) && any(!is.na(.df$high))) {
    .df$high[!is.na(.df$high)] <- .df$high[!is.na(.df$high)] + timeDelta
  }
  if (!is.null(.df$ii)) .df$ii[is.na(.df$ii)] <- 0.0
  .df
}

.etExpandGroupData <- function(df, envRef) {
  .df <- .etDropUnitsForChunk(df)
  if (!is.data.frame(.df) || nrow(.df) == 0L) {
    return(.df)
  }
  .tmp <- .df
  .tmp$id <- 1L
  .expanded <- .etExpandAddl(.tmp, envRef)
  .expanded$id <- NULL
  .expanded
}

#' Add rows of a data.frame to the ID-indexed chunks list
#'
#' Assigns id column and appends to \code{chunks[[id]]} for each ID in
#' .ids.
#'
#' @param .envRef mutable env from rxEt
#'
#' @param .df data.frame of rows to add (no 'id' column yet)
#'
#' @param .ids integer vector of ids; NULL/empty defaults to 1L
#'
#' @noRd
.etAddChunk <- function(envRef, df, ids = NULL) {
  .rt <- attr(df, ".randomType")
  if (!is.null(.rt) && !is.na(.rt)) {
    if (is.na(envRef$randomType) || .rt > envRef$randomType)
      envRef$randomType <- .rt
    if (!isTRUE(envRef$show["low"])) {
      envRef$show["low"]  <- TRUE
      envRef$show["high"] <- TRUE
    }
  }
  if (!is.data.frame(df)) {
    df <- .etExpandObsChunk(df)
  }
  if (nrow(df) == 0) {
    return(invisible(NULL))
  }
  if (is.null(ids) || length(ids) == 0L) {
    ids <-  1L
  } else {
    ids <- as.integer(ids)
  }
  .posIds <- ids[ids > 0L]
  if (length(.posIds) == 0L) return(invisible(NULL))
  if (!"id" %in% names(df)) {
    .groups <- .etGetGroups(envRef)
    .chunk <- .etGroupChunk(df, .posIds)
    if (length(.groups) > 0L) {
      .newGroups <- vector("list", 0L)
      .remainingIds <- sort.int(as.integer(.posIds), method = "quick")
      for (.g in .groups) {
        .overlap <- intersect(.g$ids, .remainingIds)
        if (length(.overlap) == 0L) {
          .newGroups[[length(.newGroups) + 1L]] <- .g
          next
        }
        .keepIds <- setdiff(.g$ids, .overlap)
        if (length(.keepIds) > 0L) {
          .newGroups[[length(.newGroups) + 1L]] <- list(ids = .keepIds, data = .g$data)
        }
        .newGroups[[length(.newGroups) + 1L]] <- list(
          ids = sort.int(as.integer(.overlap), method = "quick"),
          data = as.data.frame(data.table::rbindlist(list(.g$data, .chunk), fill = TRUE))
        )
        .remainingIds <- setdiff(.remainingIds, .overlap)
      }
      if (length(.remainingIds) > 0L) {
        .newGroups[[length(.newGroups) + 1L]] <- list(ids = .remainingIds, data = .chunk)
      }
      .etSetGroups(envRef, .newGroups)
      return(invisible(NULL))
    }
    if (length(.posIds) > 1L) {
      .match <- which(vapply(.groups, function(.g) .etGroupIdsEqual(.g$ids, .posIds), logical(1)))
      if (length(.match) == 1L) {
        .groups[[.match]]$data <- as.data.frame(
          data.table::rbindlist(list(.groups[[.match]]$data, .chunk), fill = TRUE)
        )
      } else {
        .groups[[length(.groups) + 1L]] <- list(ids = sort.int(as.integer(.posIds), method = "quick"),
                                                data = .chunk)
      }
      .etSetGroups(envRef, .groups)
      return(invisible(NULL))
    }
  }
  for (.i in .posIds) {
    .row <- .etDropUnitsForChunk(df)
    .row$id <- .i
    if (.i <= length(envRef$chunks)) {
      .cur <- envRef$chunks[[.i]] # units already dropped
      envRef$chunks[[.i]] <- as.data.frame(data.table::rbindlist(list(.cur, .row), fill = TRUE))
    } else {
      envRef$chunks[[.i]] <- .row
    }
  }
  invisible(NULL)
}

#' Merge a data.frame (with id column) into an ID-indexed chunks list
#'
#' Used by etSeq/etRbind when accumulating materialized data.frames.
#'
#' @param chunks list indexed by ID integer value
#'
#' @param df data.frame with 'id' column already set
#'
#' @return updated chunks
#'
#' @noRd
.addRowsToChunks <- function(chunks, df) {
  if (nrow(df) == 0L) return(chunks)
  .ids <- unique(as.integer(df$id))
  for (.i in .ids) {
    .rows <- .etDropUnitsForChunk(df[df$id == .i, , drop = FALSE])
    if (.i <= length(chunks)) {
      .cur <- chunks[[.i]] # units already dropped
      chunks[[.i]] <- as.data.frame(data.table::rbindlist(list(.cur, .rows), fill = TRUE))
    } else {
      chunks[[.i]] <- .rows
    }
  }
  chunks
}

# rxEt object construction ------------------------------------------------

#' Create a new empty rxEt object
#'
#' @param amountUnits character dose unit, e.g. "mg"
#'
#' @param timeUnits character time unit, e.g. "hours"
#'
#' @return rxEt object
#'
#' @noRd
.newRxEt <- function(amountUnits = NA_character_, timeUnits = NA_character_) {
  .env <- new.env(parent = emptyenv())
  .env$chunks     <- list()
  .env$groups     <- list()
  .env$units      <- c(dosing = amountUnits, time = timeUnits)
  .env$show       <- .etDefaultShow()
  .env$ids        <- 1L
  .env$nobs       <- 0L
  .env$ndose      <- 0L
  .env$randomType <- NA_integer_
  .env$canResize  <- TRUE
  .env$extraCols  <- character(0)
  .env$methods <- .etBuildMethods(.env)
  .obj <- list(
    id = integer(0), low = numeric(0), time = numeric(0), high = numeric(0),
    cmt = integer(0), amt = numeric(0), rate = numeric(0), ii = numeric(0),
    addl = integer(0), evid = integer(0), ss = integer(0), dur = numeric(0)
  )
  attr(.obj, "class")     <- c("rxEt", "data.frame")
  attr(.obj, "row.names") <- integer(0)
  attr(.obj, ".rxEtEnv")  <- .env
  .obj
}

#' Sync the materialized data.frame shell with the rxEt environment
#'
#' @param x rxEt object
#'
#' @return materialized synced rxEt data frame
#'
#' @noRd
.rxEtSyncData <- function(x) {
  .env <- .rxEtEnv(x)
  if (!is.environment(.env)) return(x)
  .nTot <- .env$nobs + .env$ndose
  .ret <- if (.nTot <= 1000L) .etMaterialize(x) else .etEmptyDf()
  attr(.ret, ".rxEtEnv") <- .env
  .rt <- .env$randomType
  .cls <- c("rxEt", "data.frame")
  if (!is.null(.rt) && !is.na(.rt)) {
    .cls <- structure(.cls, ".rxode2.lst" = list(randomType = as.integer(.rt)))
  }
  class(.ret) <- .cls
  .ret
}

# rxEt materialization and dose expansion --------------------------------

#' Expand addl doses into individual records
#'
#' @param df materialized data.frame from .etMaterialize()
#'
#' @param env environment for randomType and show flags
#'
#' @param windows logical; if TRUE, only expand windows (low/high) by
#'   ii on each subsequent dose based on random type in env
#'
#' @return data.frame with addl expanded into individual dose records
#'
#' @noRd
.etExpandAddl <- function(df, env=NULL, windows=FALSE) {
  if (windows) {
    .doseRows <- df[df$evid != 0L & df$addl > 0L & !is.na(df$low), , drop = FALSE]
  } else {
    .doseRows <- df[df$evid != 0L & df$addl > 0L, , drop = FALSE]
  }
  if (nrow(.doseRows) == 0L) return(df)
  if (is.environment(env) &&
        exists("randomType", envir = env, inherits = FALSE)) {
    .rt <- env$randomType
    if (is.na(.rt)) .rt <- 1L
  } else {
    .rt <- 1
  }
  .extras <- vector("list", nrow(.doseRows))
  for (.i in seq_len(nrow(.doseRows))) {
    .row   <- .doseRows[.i, ]
    .n     <- .row$addl
    .extra <- .row[rep(1L, .n), , drop = FALSE]
    # This works for normal and uniform cases
    if (!is.na(.row$low)) {
      .extra$low <- .row$low + seq_len(.n) * .row$ii
    }
    if (.rt == 3) {
      # normal; sd remains the same on expansion
      .extra$time <- vapply(seq_len(.n), function(i) {
        stats::rnorm(1L, .extra$low[i], .extra$high[i])
      }, numeric(1), USE.NAMES = FALSE)
    } else if (!is.na(.row$high)) {
      # uniform or window expand window by ii on each subsequent dose
      .extra$high <- .row$high + seq_len(.n) * .row$ii
      .extra$time <- vapply(seq_len(.n), function(i) {
        stats::runif(1L, .extra$low[i], .extra$high[i])
      }, numeric(1), USE.NAMES = FALSE)
    } else {
      .extra$time <- .row$time + seq_len(.n) * .row$ii
    }


    .extra$high <- .row$high + seq_len(.n) * .row$ii
    .extra$addl <- 0L
    .extra$ii   <- 0.0
    .extras[[.i]] <- .extra
  }
  df$addl[df$evid != 0L] <- 0L
  df$ii[df$evid   != 0L] <- 0.0
  .combined <- do.call(rbind, c(list(df), .extras))
  .evidSort <- ifelse(.combined$evid == 3L, -1L, as.integer(.combined$evid))
  .combined[.order3(.combined$id, .combined$time, .evidSort), ] # nolint
}

# Canonical column order matching C++ etEmpty()
.etColOrder <- c("id", "low", "time", "high", "cmt", "amt", "rate", "ii", "addl", "evid", "ss", "dur")

#' Convert character cmt column to integer for C++ solver
#'
#' The C++ etTran code handles integer cmt as direct compartment indices,
#' but character cmt values are looked up as compartment NAMES. Numeric
#' strings like "2" do NOT match compartment names ("amt2"), so they get
#' assigned as extra compartments beyond state.size(). This function converts
#' all-convertible character cmt columns to integer before passing to C++.
#'
#' "(default)" and "(obs)" map to compartment 1; NA stays NA_integer_.
#' If any value is a non-numeric, non-sentinel string (a compartment name),
#' the column stays character so C++ name lookup works normally.
#'
#' @param df materialized data.frame from .etMaterialize()
#'
#' @return df with cmt column possibly converted to integer
#'
#' @noRd
.etFixCmtForSolve <- function(df) {
  .cmt <- df[["cmt"]]
  if (is.null(.cmt) || !is.character(.cmt)) return(df)
  .isNA      <- is.na(.cmt)
  .isSentinel <- !.isNA & (.cmt == "(default)" | .cmt == "(obs)")
  .numericOk  <- !.isNA & !.isSentinel & suppressWarnings(!is.na(as.integer(.cmt)))
  if (all(.isNA | .isSentinel | .numericOk)) {
    .int <- integer(length(.cmt))
    .int[.isNA]       <- NA_integer_
    .int[.isSentinel] <- 1L
    .int[.numericOk]  <- as.integer(.cmt[.numericOk])
    df[["cmt"]] <- .int
  }
  df
}

.etGroupedSolveData <- function(x) {
  .env <- .rxEtEnv(x)
  .groups <- .etGetGroups(.env)
  if (length(.groups) == 0L) {
    return(NULL)
  }
  .dat <- vector("list", 0L)
  .ids <- vector("list", 0L)
  for (.i in seq_along(.groups)) {
    .g <- .groups[[.i]]
    .df <- .etFixCmtForSolve(.etDropUnitsForChunk(.g$data))
    if (!is.data.frame(.df) || nrow(.df) == 0L || length(.g$ids) == 0L) next
    .df$id <- rep.int(.i, nrow(.df))
    .dat[[length(.dat) + 1L]] <- .df
    .ids[[length(.ids) + 1L]] <- as.integer(.g$ids)
  }
  if (length(.dat) == 0L) {
    return(.etEmptyDf())
  }
  .out <- as.data.frame(data.table::rbindlist(.dat, fill = TRUE, use.names = TRUE))
  if ("evid" %in% names(.out)) {
    .doseIdx <- which(.out$evid != 0L)
    if (length(.doseIdx) > 0L) {
      for (.col in c("rate", "ii", "dur")) {
        if (.col %in% names(.out)) {
          .vals <- .out[[.col]]
          .naInDose <- is.na(.vals[.doseIdx])
          if (any(.naInDose)) {
            .vals[.doseIdx[.naInDose]] <- 0.0
            .out[[.col]] <- .vals
          }
        }
      }
      for (.col in c("addl", "ss")) {
        if (.col %in% names(.out)) {
          .vals <- .out[[.col]]
          .naInDose <- is.na(.vals[.doseIdx])
          if (any(.naInDose)) {
            .vals[.doseIdx[.naInDose]] <- 0L
            .out[[.col]] <- .vals
          }
        }
      }
    }
  }
  if ("time" %in% names(.out) && "id" %in% names(.out)) {
    .evidSort <- if ("evid" %in% names(.out)) {
      ifelse(!is.na(.out$evid) & .out$evid == 3L, -1L, as.integer(.out$evid))
    } else {
      rep(0L, nrow(.out))
    }
    .ord <- order(.out$id, as.numeric(.out$time), .evidSort)
    .out <- .out[.ord, ]
  }
  attr(.out, "rxHomGroups") <- .ids
  attr(.out, "rxHomIdLevels") <- as.character(unlist(.ids, use.names = FALSE))
  .tu <- .env$units["time"]
  if (!is.na(.tu) && nchar(.tu) > 0 && requireNamespace("units", quietly = TRUE)) {
    .out[["time"]] <- units::set_units(.out[["time"]], .tu, mode = "standard")
  }
  .out
}

.etGroupedSolveSplitKey <- function(df) {
  if (!is.data.frame(df) || ncol(df) == 0L) {
    return(rep.int("1", if (is.data.frame(df)) nrow(df) else 0L))
  }
  do.call(interaction, c(unname(df), list(drop = TRUE, lex.order = TRUE)))
}

.etGroupedSolveDataFrameICov <- function(events, iCov, keep = NULL, modelParams = character(0)) {
  .groups <- attr(events, "rxHomGroups", exact = TRUE)
  if (!is.data.frame(events) || is.null(.groups) || !inherits(iCov, "data.frame")) {
    return(NULL)
  }
  .idCol <- which(tolower(names(iCov)) == "id")
  if (length(.idCol) != 1L) {
    return(NULL)
  }
  .idName <- names(iCov)[.idCol]
  .keep <- as.character(keep)
  .modelParams <- as.character(modelParams)
  .icovIdRaw <- if (is.factor(iCov[[.idName]])) as.character(iCov[[.idName]]) else iCov[[.idName]]
  .icovId <- suppressWarnings(as.integer(.icovIdRaw))
  if (length(.icovId) != nrow(iCov) || anyNA(.icovId)) {
    return(NULL)
  }
  if (anyDuplicated(.icovId)) {
    return(NULL)
  }
  .eventId <- suppressWarnings(as.integer(if (is.factor(events$id)) as.character(events$id) else events$id))
  if (length(.eventId) != nrow(events) || anyNA(.eventId)) {
    return(NULL)
  }
  .eventRows <- vector("list", 0L)
  .icovRows <- vector("list", 0L)
  .outGroups <- vector("list", 0L)
  .outId <- 1L
  for (.i in seq_along(.groups)) {
    .idsInGroup <- as.integer(.groups[[.i]])
    .idx <- match(.idsInGroup, .icovId)
    if (anyNA(.idx)) {
      return(NULL)
    }
    .subIc <- iCov[.idx, , drop = FALSE]
    .splitCols <- setdiff(names(.subIc), .idName)
    .splitNeeded <- unique(c(tolower(.modelParams), tolower(.keep)))
    .splitCols <- .splitCols[tolower(.splitCols) %in% .splitNeeded]
    .splitKey <- if (length(.splitCols) == 0L) {
      factor(rep.int("1", nrow(.subIc)))
    } else {
      .etGroupedSolveSplitKey(.subIc[.splitCols])
    }
    .split <- split(seq_len(nrow(.subIc)), .splitKey, drop = TRUE)
    .df <- events[.eventId == .i, , drop = FALSE]
    if (!is.data.frame(.df) || nrow(.df) == 0L) {
      return(NULL)
    }
    for (.grpIdx in .split) {
      .ids <- as.integer(.idsInGroup[.grpIdx])
      .row <- .df
      .row$id <- rep.int(.outId, nrow(.row))
      .eventRows[[length(.eventRows) + 1L]] <- .row
      .icovRow <- .subIc[.grpIdx[1L], , drop = FALSE]
      .icovRow[[.idName]] <- .outId
      .icovRows[[length(.icovRows) + 1L]] <- .icovRow
      .outGroups[[length(.outGroups) + 1L]] <- .ids
      .outId <- .outId + 1L
    }
  }
  if (length(.eventRows) == 0L) {
    return(NULL)
  }
  .events <- as.data.frame(data.table::rbindlist(.eventRows, fill = TRUE, use.names = TRUE))
  .iCovOut <- as.data.frame(data.table::rbindlist(.icovRows, fill = TRUE, use.names = TRUE))
  .iCovOut[[.idName]] <- suppressWarnings(as.integer(.iCovOut[[.idName]]))
  if (anyNA(.iCovOut[[.idName]])) {
    return(NULL)
  }
  attr(.events, "rxHomGroups") <- .outGroups
  attr(.events, "rxHomIdLevels") <- as.character(unlist(.outGroups, use.names = FALSE))
  list(
    events = .events,
    iCov = .iCovOut
  )
}

.etGroupedSolveDataICov <- function(x, iCov, keep = NULL, modelParams = character(0)) {
  .env <- .rxEtEnv(x)
  .groups <- .etGetGroups(.env)
  if (length(.groups) == 0L) {
    return(NULL)
  }
  .etGroupedSolveDataFrameICov(.etGroupedSolveData(x), iCov,
                               keep = keep, modelParams = modelParams)
}

.etSolveObsValue <- function(x, name) {
  if (is.null(x)) {
    return(NULL)
  }
  if (length(x) != 1L) {
    stop(sprintf("'%s' must be of length 1", name), call. = FALSE)
  }
  as.numeric(x)[1L]
}

.etSolveObsTimes <- function(events, ctl = NULL) {
  .from <- .etSolveObsValue(ctl$from, "from")
  if (is.null(.from)) {
    .from <- 0.0
  }
  .to <- .etSolveObsValue(ctl$to, "to")
  if (is.null(.to)) {
    .maxTime <- suppressWarnings(max(as.numeric(events$time), na.rm = TRUE))
    if (!is.finite(.maxTime)) {
      .maxTime <- .from
    }
    .to <- .maxTime + 24
  }
  .by <- .etSolveObsValue(ctl$by, "by")
  .lengthOut <- .etSolveObsValue(ctl$length.out, "length.out")
  if (!is.null(.lengthOut)) {
    .lengthOut <- as.integer(.lengthOut)
    if (!is.null(.by)) {
      stop("cannot use both 'by' and 'length.out' for rxode2 simulations", call. = FALSE)
    }
    .by <- (.to - .from)/(.lengthOut - 1)
  } else if (is.null(.by)) {
    .lengthOut <- 200L
    .by <- (.to - .from)/(.lengthOut - 1)
  } else {
    .lengthOut <- as.integer((.to - .from)/.by + 1.0)
  }
  .by * seq.int(0L, .lengthOut - 1L) + .from
}

.etAddSolveObsRows <- function(events, obsTimes) {
  .groups <- attr(events, "rxHomGroups", exact = TRUE)
  .idLevels <- attr(events, "rxHomIdLevels", exact = TRUE)
  .events <- .etFixCmtForSolve(events)
  .events$time <- as.numeric(.events$time)
  .ids <- sort(unique(as.integer(.events$id)))
  .obs <- vector("list", length(.ids))
  for (.i in seq_along(.ids)) {
    .obs[[.i]] <- data.frame(
      id = rep.int(.ids[.i], length(obsTimes)),
      time = obsTimes,
      evid = 0L
    )
  }
  .out <- as.data.frame(data.table::rbindlist(c(list(.events), .obs), fill = TRUE, use.names = TRUE))
  if (!is.null(.groups)) {
    attr(.out, "rxHomGroups") <- .groups
  }
  if (!is.null(.idLevels)) {
    attr(.out, "rxHomIdLevels") <- .idLevels
  }
  .out
}

.etPrepareGroupedSolveData <- function(events, ctl = NULL) {
  .groups <- attr(events, "rxHomGroups", exact = TRUE)
  .events <- .etFixCmtForSolve(events)
  if (is.null(.groups)) {
    return(.events)
  }
  if (nrow(.events) == 0L) {
    return(.events)
  }
  if ("evid" %in% names(.events) && any(.events$evid == 0L, na.rm = TRUE)) {
    return(.events)
  }
  .etAddSolveObsRows(.events, .etSolveObsTimes(.events, ctl))
}

.etPrepareSolveEvents <- function(events, ctl = NULL) {
  if (!is.rxEt(events)) {
    return(.etPrepareGroupedSolveData(events, ctl))
  }
  .env <- .rxEtEnv(events)
  if (is.null(ctl$iCov) && length(.etGroups(.env)) > 0L) {
    return(.etPrepareGroupedSolveData(.etGroupedSolveData(events), ctl))
  }
  if (isTRUE(.env$nobs == 0L)) {
    return(.etFixCmtForSolve(.etMaterialize(events)))
  }
  .etFixCmtForSolve(.etMaterialize(events))
}

#' Empty data.frame skeleton matching materialized format
#'
#' @return empty data.frame with canonical columns and types
#' @noRd
.etEmptyDf <- function() {
  data.frame(
    id    = integer(0),
    low   = numeric(0),
    time  = numeric(0),
    high  = numeric(0),
    cmt   = character(0),
    amt   = numeric(0),
    rate  = numeric(0),
    ii    = numeric(0),
    addl  = integer(0),
    evid  = integer(0),
    ss    = integer(0),
    dur   = numeric(0),
    stringsAsFactors = FALSE
  )
}

#' Materialize all chunks into a sorted data.frame
#'
#' Chunks is a list indexed by ID integer value; \code{chunks[[i]]} is
#' a data.frame for ID i (with 'id' column set), or NULL if no records
#' for that ID.
#'
#' @param et rxEt object
#'
#' @return data.frame with canonical columns, sorted by (id, time,
#'   evid-adjusted)
#'
#' @noRd
.etMaterialize <- function(et) {
  id <- NULL # for R cmd check since data.table uses id := ...
  .env <- .rxEtEnv(et)
  .groups <- .etGetGroups(.env)

  if (length(.groups) > 0L) {
    .nonNull <- Filter(function(.g) !is.null(.g$data) && nrow(.g$data) > 0L, .groups)
    if (length(.nonNull) == 0L) return(.etEmptyDf())
    .dtList <- lapply(.nonNull, function(.g) {
      .df <- .g$data
      if (!is.data.frame(.df)) {
        .df <- .etExpandObsChunk(.df)
      }
      if ("id" %in% names(.df)) {
        return(data.table::as.data.table(.df))
      }
      .n <- nrow(.df)
      if (.n == 0L || length(.g$ids) == 0L) {
        return(data.table::as.data.table(.etEmptyDf()))
      }
      .idx <- rep.int(seq_len(.n), times = length(.g$ids))
      .id <- rep(as.integer(.g$ids), each = .n)
      .ret <- data.table::as.data.table(.df[.idx, , drop = FALSE])
      .ret[, id := .id]
      .ret
    })
    .dt <- data.table::rbindlist(.dtList, fill = TRUE, use.names = TRUE)
  } else {
    .chunks <- .env$chunks

    if (length(.chunks) == 0L) return(.etEmptyDf())

    # Each element is a data.frame for one ID; filter NULL entries
    .nonNull <- Filter(Negate(is.null), .chunks)
    if (length(.nonNull) == 0L) return(.etEmptyDf())

    # rbindlist: fast sparse bind across ID data.frames
    .dt <- data.table::rbindlist(.nonNull, fill = TRUE, use.names = TRUE)
  }

  # ---- Fill column defaults ----
  if (is.null(.dt[["id"]]))   data.table::set(.dt, j = "id",   value = 1L)
  if (is.null(.dt[["evid"]])) data.table::set(.dt, j = "evid", value = 0L)
  if (is.null(.dt[["low"]]))  data.table::set(.dt, j = "low",  value = NA_real_)
  if (is.null(.dt[["high"]])) data.table::set(.dt, j = "high", value = NA_real_)
  if (is.null(.dt[["cmt"]]))  data.table::set(.dt, j = "cmt",  value = "(default)")

  .isDose <- .dt$evid != 0L
  if (is.null(.dt[["amt"]]))  data.table::set(.dt, j = "amt",  value = NA_real_)
  if (is.null(.dt[["rate"]])) data.table::set(.dt, j = "rate", value = ifelse(.isDose, 0.0, NA_real_))
  if (is.null(.dt[["ii"]]))   data.table::set(.dt, j = "ii",   value = ifelse(.isDose, 0.0, NA_real_))
  if (is.null(.dt[["addl"]])) data.table::set(.dt, j = "addl", value = ifelse(.isDose, 0L,  NA_integer_))
  if (is.null(.dt[["ss"]]))   data.table::set(.dt, j = "ss",   value = ifelse(.isDose, 0L,  NA_integer_))
  if (is.null(.dt[["dur"]]))  data.table::set(.dt, j = "dur",  value = ifelse(.isDose, 0.0, NA_real_))

  # Patch NA cells in dose rows for numeric dose fields
  .doseIdx <- which(.dt$evid != 0L)
  if (length(.doseIdx) > 0L) {
    for (.col in c("rate", "ii", "dur")) {
      if (.col %in% names(.dt)) {
        .vals <- .dt[[.col]]
        .naInDose <- is.na(.vals[.doseIdx])
        if (any(.naInDose)) {
          .vals[.doseIdx[.naInDose]] <- 0.0
          data.table::set(.dt, j = .col, value = .vals)
        }
      }
    }
    for (.col in c("addl", "ss")) {
      if (.col %in% names(.dt)) {
        .vals <- .dt[[.col]]
        .naInDose <- is.na(.vals[.doseIdx])
        if (any(.naInDose)) {
          .vals[.doseIdx[.naInDose]] <- 0L
          data.table::set(.dt, j = .col, value = .vals)
        }
      }
    }
  }

  # Sort: id ASC, time ASC, evid=3 sorts before others at same (id, time)
  .evidSort <- ifelse(.dt$evid == 3L, -1L, as.integer(.dt$evid))
  .ord <- .order3(.dt$id, .dt$time, .evidSort) # nolint
  .dt <- .dt[.ord, ]

  # Ensure canonical column order
  .missing <- setdiff(.etColOrder, names(.dt))
  for (.col in .missing) data.table::set(.dt, j = .col, value = NA)
  data.table::setcolorder(.dt, .etColOrder)

  .df <- as.data.frame(.dt)
  .df <- .etExpandAddl(.df, env = .env, windows = TRUE)

  # Apply units to columns so C++ solver can propagate them to output
  .tu <- .env$units["time"]
  .du <- .env$units["dosing"]
  .hasTimeU <- !is.na(.tu) && nchar(.tu) > 0
  .hasDoseU <- !is.na(.du) && nchar(.du) > 0
  if (.hasTimeU || .hasDoseU) {
    if (.hasTimeU) {
      for (.col in c("time", "ii", "low", "high", "dur")) {
        .df[[.col]] <- units::set_units(.df[[.col]], .tu, mode = "standard")
      }
    }
    if (.hasDoseU) {
      .df[["amt"]] <- units::set_units(.df[["amt"]], .du, mode = "standard")
      if (.hasTimeU) {
        .df[["rate"]] <- units::set_units(.df[["rate"]], paste0(.du, "/", .tu), mode = "standard")
      }
    }
  }
  .df
}

#' Check if object is an rxEt event table
#'
#' @param x object to test
#'
#' @return logical
#'
#' @export
is.rxEt <- function(x) {
  if (!inherits(x, "rxEt")) return(FALSE)
  .env <- .rxEtEnv(x)
  is.environment(.env)
}

# rxEt chunk builders and shell rebuilders -------------------------------

#' @noRd
.etWindowTime <- function(time, allow1 = FALSE) {
  .randomType <- NA_integer_
  if (inherits(time, "list")) {
    .nw <- length(time)
    .low  <- numeric(.nw)
    .mid  <- numeric(.nw)
    .high <- numeric(.nw)
    # This is for the normal variability case that is list(c(4, 2, NA))
    .hasNormal <- any(vapply(time,
                             function(.w) {
                               length(.w) == 3L && is.na(.w[3L])
                             }, logical(1L)))
    .has3 <- any(vapply(time,
                        function(.w) {
                          length(.w) == 3L && !is.na(.w[3L])
                        }, logical(1L)))
    for (.i in seq_len(.nw)) {
      .w <- time[[.i]]
      if (length(.w) == 1L) {
        if (!allow1) stop("each window must be c(low, high) or c(low, mid, high)", call. = FALSE)
        .low[.i]  <- NA_real_
        .mid[.i]  <- as.numeric(.w)
        .high[.i] <- NA_real_
      } else if (length(.w) == 2L) {
        if (.hasNormal) {
          .low[.i]  <- .w[1]
          .mid[.i]  <- stats::rnorm(1, .w[1], .w[2])
          .high[.i] <- .w[2]
        } else if (.has3) {
          stop("Cannot mix 2 and 3 element windows", call. = FALSE)
        } else {
          if (.w[1] > .w[2]) {
            stop("window bounds must be ordered c(low, high)", call. = FALSE)
          }
          .low[.i]  <- .w[1]
          .mid[.i]  <- stats::runif(1, .w[1], .w[2])
          .high[.i] <- .w[2]
        }
      } else if (length(.w) == 3L) {
        if (is.na(.w[3L])) {
          .low[.i]  <- .w[1]
          .mid[.i]  <- stats::rnorm(1, .w[1], .w[2])
          .high[.i] <- .w[2]
        } else {
          if (.w[1] > .w[2] || .w[2] > .w[3]) stop("window bounds must be ordered c(low, mid, high)", call. = FALSE)
          .low[.i]  <- .w[1]
          .mid[.i]  <- .w[2]
          .high[.i] <- .w[3]
        }
      } else {
        stop("each window must be c(low, high) or c(low, mid, high)", call. = FALSE)
      }
    }
    if (.hasNormal) {
      .randomType <- 3L
    } else if (.has3) {
      .randomType <- 1L
    } else {
      .randomType <- 2L
    }
    return(list(time = .mid, low = .low, high = .high, randomType = .randomType))
  }
  list(time = as.numeric(time), low = NA_real_, high = NA_real_, randomType = NA_integer_)
}

#' @noRd
.etDoseUntil <- function(until, time, ii, addl, isList) {
  if (is.null(until)) return(addl)
  if (ii <= 0) stop("'until' requires a positive 'ii'", call. = FALSE)
  .tmp <- until - time - ii
  if (any(.tmp > 0, na.rm = TRUE)) {
    .ratio <- .tmp / ii
    .ceil <- ceiling(.ratio)
    addl <- ifelse(.ceil == .ratio, .ratio + 1, .ceil)
    addl <- as.integer(addl)
  } else {
    addl <- 0L
    if (!isList) {
      warning("'time'+'ii' is greater than 'until', no additional doses added", call. = FALSE)
    }
  }
  if (any(addl < 0L, na.rm = TRUE)) {
    warning("'until' is before 'time'; setting addl=0", call. = FALSE)
    addl[addl < 0L] <- 0L
  }
  addl
}

#' @noRd
.etDoseValidate <- function(amt, ii, addl, ss, rate) {
  if (ii > 0 && ss == 0L && all(addl == 0L, na.rm = TRUE)) {
    warning(sprintf(
      "'ii' requires non zero additional doses ('addl') or steady state dosing ('ii': %f, 'ss': %d; 'addl': %d), reset 'ii' to zero", # nolint
      ii, ss, max(addl, na.rm = TRUE)
    ), call. = FALSE)
    ii <- 0.0
  }

  if (any(addl > 0L, na.rm = TRUE) && ii == 0.0)
    stop("'addl' > 0 requires a positive inter-dose interval ('ii')", call. = FALSE)

  if (ss > 0L) {
    if (rate < -1.0 && ii == 0.0)
      stop("cannot use duration flag (rate=-2) with steady-state dosing", call. = FALSE)
    if (ss == 2L && ii == 0.0)
      stop("ss=2 requires a positive inter-dose interval ('ii')", call. = FALSE)
    if (rate > 0 && ii > 0 && amt == 0)
      stop("cannot combine constant infusion (rate>0) with dose interval (ii>0) for steady-state; use ii=0 for constant infusion SS", call. = FALSE) # nolint
  }
  ii
}

#' Build an observation chunk list (no id column; caller uses .etAddChunk)

#'
#' Returns a sparse named list where scalar fields (evid) are stored once and
#' vector fields (time) hold per-row values. Callers use length(.df$time) for
#' row count, not nrow().
#'
#' @param time numeric vector of sample times, OR a list of c(low,high)
#'   or c(low,mid,high) windows.
#'
#' @param cmt compartment name or number (optional scalar)
#'
#' @param id integer vector of ids to attach (optional)
#'
#' @return sparse named list with evid=0L scalar and time vector
#'
#' @noRd
.etObsChunk <- function(time, cmt = NULL, id = NULL) {
  .time <- .etWindowTime(time)
  .df <- list(evid = 0L, time = .time$time, low = .time$low, high = .time$high)
  if (!is.null(cmt)) .df$cmt <- cmt
  if (!is.null(id))  .df$id  <- as.integer(id)
  attr(.df, ".randomType") <- .time$randomType
  .df
}

#' Expand a sparse obs chunk list to a full data.frame
#'
#' Replicates scalar fields to match the length of the time vector.
#' @param .sparse sparse list from .etObsChunk
#' @return data.frame with all columns same length
#' @noRd
.etExpandObsChunk <- function(.sparse) {
  .n <- length(.sparse$time)
  as.data.frame(
    lapply(.sparse, function(.v) if (length(.v) == 1L) rep(.v, .n) else .v),
    stringsAsFactors = FALSE
  )
}

#' Build a dosing chunk data.frame (no id column; caller uses .etAddChunk)
#'
#' @param time numeric start time(s)
#' @param amt numeric dose amount(s)
#' @param evid integer event ID (default 1L)
#' @param cmt compartment name/number (default "(default)")
#' @param ii inter-dose interval (default 0)
#' @param addl additional doses (default 0L)
#' @param ss steady-state flag (default 0L)
#' @param rate infusion rate (default 0)
#' @param dur infusion duration (default 0.0)
#' @param until time of last dose; overrides addl
#' @param nbr.doses number of doses (legacy); sets addl = nbr.doses - 1
#' @param dosing.interval inter-dose interval (legacy alias for ii)
#' @return data.frame
#' @noRd
.etDoseChunk <- function(time = 0, amt, evid = 1L, cmt = "(default)",
                         ii = 0.0, addl = 0L, ss = 0L, rate = 0.0, dur = 0.0,
                         until = NULL, nbr.doses = NULL, dosing.interval = NULL) {
  if (!is.null(dosing.interval)) ii <- as.numeric(dosing.interval)
  if (!is.null(nbr.doses))       addl <- as.integer(nbr.doses) - 1L

  .time <- .etWindowTime(time, allow1 = TRUE)
  .timeDose <- .time$time
  .lowDose <- .time$low
  .highDose <- .time$high
  .randomType <- .time$randomType

  addl <- .etDoseUntil(until, .timeDose, ii, addl, inherits(time, "list"))
  ii <- .etDoseValidate(amt, ii, addl, ss, rate)

  if (dur > 0 && rate == 0.0) {
    rate <- amt / dur
    dur  <- 0.0
  }

  if (length(amt) > 1L || length(time) > 1L) {
    if (length(.timeDose) == 1L) .timeDose <- rep(.timeDose, length(amt))
    if (length(amt)  == 1L) amt  <- rep(amt,  length(.timeDose))
    if (length(.timeDose) != length(amt)) stop("'time' and 'amt' must have the same length", call. = FALSE)
  }

  .res <- data.frame(
    time = as.numeric(.timeDose), amt  = as.numeric(amt),
    evid = as.integer(evid), cmt  = as.character(cmt),
    ii   = as.numeric(ii),   addl = as.integer(addl),
    ss   = as.integer(ss),   rate = as.numeric(rate),
    dur  = as.numeric(dur),
    low  = as.numeric(.lowDose),
    high = as.numeric(.highDose),
    stringsAsFactors = FALSE
  )
  attr(.res, ".randomType") <- .randomType
  .res
}

#' Rebuild an rxEt shell after data-frame style mutation
#'
#' @param x rxEt object
#'
#' @param df materialized data.frame from .etMaterialize() after mutation
#'
#' @noRd
.rxEtRebuildShell <- function(x, df) {
  .env0 <- .rxEtEnv(x)
  .et <- .newRxEt(amountUnits = .env0$units["dosing"], timeUnits = .env0$units["time"])
  .env <- .rxEtEnv(.et)
  if (nrow(df) > 0L) {
    .env$methods$import.EventTable(df)
  }
  .env$show[names(.env0$show)] <- .env$show[names(.env0$show)] | .env0$show
  .env$randomType <- .env0$randomType
  .env$canResize <- .env0$canResize
  .etAddExtraCols(.env, .etExtraCols(.env0)) # nolint
  .et
}

#' @export
dim.rxEt <- function(x) {
  dim(as.data.frame(x, all = TRUE))
}

rxTest({ # mostly tested in 'rxode2et'
  test_that("warfarin model", {

    skip_if_not_installed("nlmixr2data")
    warfarin <- nlmixr2data::warfarin

    mod <- rxode2({
      lka <- log(0.1) # log Ka
      lv <- log(10) # Log Vc
      lcl <- log(4) # Log Cl
      lq <- log(10) # log Q
      lvp <- log(20) # Log Vp
      eta.ka <- 0
      eta.v <- 0.1
      eta.cl <- 0.1
      ka <- exp(lka + eta.ka + sex + age + dvid)
      cl <- exp(lcl + eta.cl)
      v <- exp(lv + eta.v)
      q <- exp(lq)
      vp <- exp(lvp)
      sf <- (sex == "female")
      sm <- (sex == "male")
      d.cp <- (dvid == "cp")
      d.pca <- (dvid == "pca")
      cp <- linCmt()
    })

    t <- rxSolve(mod, warfarin, keep = c("sex", "age", "dvid"))

    expect_equal(sort(unique(t$sf)), c(0, 1))
    expect_equal(sort(unique(t$sm)), c(0, 1))
    expect_equal(sort(unique(t$d.cp)), c(0, 1))
    expect_equal(sort(unique(t$d.pca)), c(0, 1))

    expect_s3_class(t$sex, "factor")
    expect_s3_class(t$dvid, "factor")

    expect_equal(as.double((t$sex == "male") * 1), t$sm)
    expect_equal(as.double((t$sex == "female") * 1), t$sf)

    t <- rxSolve(mod, warfarin, addCov = TRUE)

    expect_equal(as.double((t$sex == "male") * 1), t$sm)
    expect_equal(as.double((t$sex == "female") * 1), t$sf)
    ## expect_equal(as.double((t$dvid == "cp") * 1), t$d.cp)
    ## expect_equal(as.double((t$pca == "pca") * 1), t$d.pca)

    warfarin$sex <- paste(warfarin$sex)

    t <- rxSolve(mod, warfarin, keep = c("sex", "age", "dvid"))

    expect_equal(sort(unique(t$sf)), c(0, 1))
    expect_equal(sort(unique(t$sm)), c(0, 1))
    expect_equal(sort(unique(t$d.cp)), c(0, 1))
    expect_equal(sort(unique(t$d.pca)), c(0, 1))

    expect_s3_class(t$sex, "factor")
    expect_s3_class(t$dvid, "factor")

    expect_equal(as.double((t$sex == "male") * 1), t$sm)
    expect_equal(as.double((t$sex == "female") * 1), t$sf)

    t <- rxSolve(mod, warfarin, addCov = TRUE)

    expect_equal(as.double((t$sex == "male") * 1), t$sm)
    expect_equal(as.double((t$sex == "female") * 1), t$sf)

  })

  test_that("etTrans na time evid=2", {

    mod <- rxode2parse("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
d/dt(blood)     = a*intestine - b*blood
")

    et <- structure(list(time = c(0, 0.05, 0.1, 0.2, 0.3, NA),
                         cmt = c("(default)", "(obs)", "intestine", "-intestine", "intestine", "out"),
                         amt = c(0.0833333333333333, NA, 3, NA, 3, 3),
                         rate = c(2, 0, 0, 0, 0, 0),
                         ii = c(1, 0, 3, 0, 3, 0),
                         addl = c(9L, 0L, 0L, 0L, 0L, 0L),
                         evid = c(1L, 2L, 1L, 2L, 1L, 1L),
                         ss = c(0L, 0L, 1L, 0L, 2L, 0L)),
                    class = "data.frame",
                    row.names = c(NA,-6L))
    skip_on_cran()

  expect_warning(expect_false(any(is.na(etTrans(et, mod)$TIME))))

  })

  test_that("etTrans uses model string comparison ordering for event data covariates", {
    mod <- rxode2parse('c = cov == "alpha"; d = "beta" == cov')

    dat <- data.frame(
      id = c(1, 1, 2, 2, 3, 3),
      time = c(0, 1, 0, 1, 0, 1),
      amt = c(1, NA, 1, NA, 1, NA),
      evid = c(1, 0, 1, 0, 1, 0),
      cov = c("beta", "beta", "gamma", "gamma", "alpha", "alpha")
    )

    tmp <- etTrans(dat, mod, keep = "cov")
    tmp2 <- attr(class(tmp), ".rxode2.lst")
    class(tmp2) <- NULL

    expect_equal(tmp2$levelInfo$cov, c("alpha", "beta", "gamma"))
    expect_equal(as.numeric(tmp2$cov1$cov), c(2, 3, 1))
  })

  test_that("homogeneous solve events stay compressed through etTrans", {
    mod <- rxode2parse("
      ka = 1
      d/dt(depot) = -ka * depot
    ")

    ev <- et(time = 0, amt = 100, cmt = 1) |>
      et(0:4) |>
      et(id = 1:3)

    prep <- .etPrepareSolveEvents(ev, rxControl())
    grp <- .etGetGroups(.rxEtEnv(ev))

    expect_equal(nrow(prep), nrow(grp[[1]]$data))
    expect_equal(attr(prep, "rxHomIdLevels"), c("1", "2", "3"))

    tr <- etTrans(prep, mod)
    expect_equal(nrow(as.data.frame(tr)), nrow(prep))
    expect_equal(attr(tr, "rxHomIdLevels"), c("1", "2", "3"))
    expect_equal(as.integer(attr(tr, "rxHomGroups")[[1]]), 1:3)

    trInfo <- attr(class(tr), ".rxode2.lst")
    expect_equal(trInfo$idLvl, c("1", "2", "3"))
  })

  test_that("homogeneous grouped solve matches expanded solve output", {
    mod <- rxode2({
      KA <- 1
      CL <- 1
      V <- 10
      C2 <- centr / V
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    ev <- et(ev, id = 1:3)

    got <- as.data.frame(rxSolve(mod, ev))
    want <- as.data.frame(rxSolve(mod, as.data.frame(ev)))

    expect_equal(
      got[, c("id", "time", "depot", "centr")],
      want[, c("id", "time", "depot", "centr")]
    )
  })

  test_that("homogeneous grouped dose-only solve keeps grouped ids", {
    mod <- rxode2({
      KA <- 1
      CL <- 1
      V <- 10
      C2 <- centr / V
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev <- et(ev, id = 1:3)

    got <- as.data.frame(rxSolve(mod, ev, from = 0, to = 24, by = 12))
    want <- as.data.frame(rxSolve(mod, as.data.frame(ev), from = 0, to = 24, by = 12))

    expect_equal(
      got[, c("id", "time", "depot", "centr")],
      want[, c("id", "time", "depot", "centr")]
    )
  })

  test_that("homogeneous grouped dose-only solve prep keeps grouped attrs", {
    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev <- et(ev, id = 1:3)

    prep <- .etPrepareSolveEvents(ev, rxControl(from = 0, to = 24, by = 12))

    expect_equal(attr(prep, "rxHomGroups"), list(1:3))
    expect_equal(attr(prep, "rxHomIdLevels"), c("1", "2", "3"))
    expect_equal(sort(unique(prep$id)), 1L)
    expect_equal(sum(prep$evid == 0L), 3L)
  })

  test_that("grouped solve prep handles empty grouped data frames", {
    ev <- data.frame(id = integer(0), time = numeric(0), evid = integer(0))
    attr(ev, "rxHomGroups") <- list(1:2)
    attr(ev, "rxHomIdLevels") <- c("1", "2")

    prep <- .etPrepareGroupedSolveData(ev, rxControl())

    expect_true(is.data.frame(prep))
    expect_equal(nrow(prep), 0L)
    expect_equal(attr(prep, "rxHomGroups"), list(1:2))
  })

  test_that("homogeneous grouped dose-only solve prep supports model iCov compression", {
    mod <- rxode2({
      WT2 <- WT/70
      C2 <- centr / V
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * WT2 * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev <- et(ev, id = 1:4)
    iCov <- data.frame(id = 1:4, WT = c(70, 70, 80, 80))
    ctl <- rxControl(iCov = iCov, from = 0, to = 24, by = 12)
    prep <- .etGroupedSolveDataICov(ev, iCov, modelParams = rxModelVars(mod)$params)
    prepEvents <- .etPrepareSolveEvents(prep$events, ctl)
    ctl$iCov <- prep$iCov

    got <- as.data.frame(rxSolveSEXP(mod, ctl, NULL, list(),
                                              c(KA = 1, CL = 7, V = 40),
                                              prepEvents, NULL, FALSE))
    want <- as.data.frame(rxSolve(mod, as.data.frame(ev), params = c(KA = 1, CL = 7, V = 40), iCov = iCov,
                                  from = 0, to = 24, by = 12))

    expect_equal(
      got[, c("id", "time", "depot", "centr")],
      want[, c("id", "time", "depot", "centr")]
    )
  })

  test_that("homogeneous grouped dose-only solve prep handles unit time defaults", {
    mod <- rxode2({
      WT2 <- WT/70
      C2 <- centr / V
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * WT2 * C2
    })

    ev <- et(amount.units = "mg", time.units = "hours")
    ev <- et(ev, amt = 10000, cmt = 1)
    ev <- et(ev, id = 1:4)
    iCov <- data.frame(id = 1:4, WT = c(70, 70, 80, 80))

    expect_no_error(
      .sol <- solve(mod, ev, iCov = iCov, params = c(KA = 1, CL = 7, V = 40))
    )
    expect_true(nrow(as.data.frame(.sol)) > 0L)
  })

  test_that("homogeneous grouped dose-only solve keeps iCov keep-column output", {
    mod <- rxode2({
      WT2 <- WT/70
      C2 <- centr / V
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * WT2 * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev <- et(ev, id = 1:4)
    iCov <- data.frame(id = 1:4, WT = c(70, 70, 80, 80), grp = c("a", "a", "b", "b"))

    got <- suppressWarnings(
      as.data.frame(
        rxSolve(mod, ev, params = c(KA = 1, CL = 7, V = 40), iCov = iCov, keep = "grp",
                from = 0, to = 24, by = 12)
      )
    )
    want <- suppressWarnings(
      as.data.frame(
        rxSolve(mod, as.data.frame(ev), params = c(KA = 1, CL = 7, V = 40), iCov = iCov, keep = "grp",
                from = 0, to = 24, by = 12)
      )
    )

    expect_equal(
      got[, c("id", "time", "depot", "centr", "grp")],
      want[, c("id", "time", "depot", "centr", "grp")]
    )
  })

  test_that("homogeneous grouped solve matches expanded output when nsim is requested", {
    mod <- rxode2({
      KA <- 1
      CL <- 1
      V <- 10
      C2 <- centr / V
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    ev <- et(ev, id = 1:3)

    got <- as.data.frame(rxSolve(mod, ev, nsim = 2))
    want <- as.data.frame(rxSolve(mod, as.data.frame(ev), nsim = 2))

    expect_equal(
      got[, c("id", "time", "depot", "centr")],
      want[, c("id", "time", "depot", "centr")]
    )
  })

  test_that("homogeneous rxEt routes nsim through nSub instead of nStud", {
    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev <- et(ev, id = 1:3)

    groupedCtl <- rxControl(nsim = 2, events = ev)
    expandedCtl <- rxControl(nsim = 2, events = as.data.frame(ev))

    expect_equal(groupedCtl$nSub, 2L)
    expect_equal(groupedCtl$nStud, 1L)
    expect_equal(expandedCtl$nSub, 1L)
    expect_equal(expandedCtl$nStud, 2L)
  })

  test_that("rxSolve updates preserve grouped solve event attributes", {
    on.exit(rxClean(), add = TRUE)
    mod <- rxode2({
      KA <- 1
      CL <- 1
      V <- 10
      C2 <- centr / V
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    ev <- et(ev, id = 1:4)

    s1 <- rxSolve(mod, ev)
    .e1 <- attr(class(s1), ".rxode2.env")
    .ev1 <- .e1$.args.events
    .groups <- attr(.ev1, "rxHomGroups", exact = TRUE)
    .idLvls <- attr(.ev1, "rxHomIdLevels", exact = TRUE)
    expect_false(is.null(.groups))

    s2 <- rxSolveUpdate(s1, "KA", rep(0.5, nrow(.ev1)))
    .e2 <- attr(class(s2), ".rxode2.env")
    .ev2 <- .e2$.args.events
    expect_equal(attr(.ev2, "rxHomGroups"), .groups)
    expect_equal(attr(.ev2, "rxHomIdLevels"), .idLvls)

    s3 <- rxSolveUpdate(s2, "KA", 1)
    .e3 <- attr(class(s3), ".rxode2.env")
    .ev3 <- .e3$.args.events
    expect_equal(attr(.ev3, "rxHomGroups"), .groups)
    expect_equal(attr(.ev3, "rxHomIdLevels"), .idLvls)
  })

  test_that("rxSolve updates preserve grouped iCov keep-column output", {
    on.exit(rxClean(), add = TRUE)
    mod <- rxode2({
      WT2 <- WT/70
      C2 <- centr / V2
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * WT2 * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev <- et(ev, id = 1:4)
    iCov <- data.frame(id = 1:4, WT = c(70, 70, 80, 80), grp = c("a", "a", "b", "b"))

    s1 <- suppressWarnings(
      rxSolve(mod, ev, params = c(KA = 1, CL = 7, V2 = 40),
              iCov = iCov, keep = "grp", from = 0, to = 24, by = 12)
    )
    .e1 <- attr(class(s1), ".rxode2.env")
    .ev1 <- .e1$.args.events
    expect_equal(sort(unique(as.data.frame(s1)$grp)), c("a", "b"))

    s2 <- suppressWarnings(rxSolveUpdate(s1, "KA", rep(0.5, nrow(.ev1))))
    .e2 <- attr(class(s2), ".rxode2.env")
    .ev2 <- .e2$.args.events
    expect_equal(attr(.ev2, "rxHomGroups"), attr(.ev1, "rxHomGroups"))
    expect_equal(attr(.ev2, "rxHomIdLevels"), attr(.ev1, "rxHomIdLevels"))
    expect_equal(sort(unique(as.data.frame(s2)$grp)), c("a", "b"))

    s3 <- suppressWarnings(rxSolveUpdate(s2, "KA", 1))
    .e3 <- attr(class(s3), ".rxode2.env")
    .ev3 <- .e3$.args.events
    expect_equal(attr(.ev3, "rxHomGroups"), attr(.ev1, "rxHomGroups"))
    expect_equal(attr(.ev3, "rxHomIdLevels"), attr(.ev1, "rxHomIdLevels"))
    expect_equal(sort(unique(as.data.frame(s3)$grp)), c("a", "b"))
  })

  test_that("homogeneous grouped solve supports model iCov compression", {
    mod <- rxode2({
      WT2 <- WT/70
      C2 <- centr / V2
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * WT2 * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    ev <- et(ev, id = 1:4)
    iCov <- data.frame(id = 1:4, WT = c(70, 70, 80, 80))

    got <- as.data.frame(rxSolve(mod, ev, params = c(KA = 1, CL = 7, V2 = 40), iCov = iCov))
    want <- as.data.frame(rxSolve(mod, as.data.frame(ev), params = c(KA = 1, CL = 7, V2 = 40), iCov = iCov))

    expect_equal(
      got[, c("id", "time", "depot", "centr")],
      want[, c("id", "time", "depot", "centr")]
    )
  })

  test_that("grouped solve data.frame infers iCov ids from homogeneous id levels", {
    mod <- rxode2({
      WT2 <- WT/70
      C2 <- centr / V2
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * WT2 * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    ev <- et(ev, id = 101:104)

    groupedDf <- .etGroupedSolveData(ev)
    iCovNoId <- data.frame(WT = c(70, 70, 80, 80))

    got <- as.data.frame(rxSolve(mod, groupedDf, params = c(KA = 1, CL = 7, V2 = 40), iCov = iCovNoId))
    want <- as.data.frame(rxSolve(mod, as.data.frame(ev), params = c(KA = 1, CL = 7, V2 = 40), iCov = iCovNoId))

    expect_equal(
      got[, c("id", "time", "depot", "centr")],
      want[, c("id", "time", "depot", "centr")]
    )
  })

  test_that("grouped solve data.frame iCov split accepts factor representative ids", {
    mod <- rxode2({
      WT2 <- WT/70
      C2 <- centr / V2
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * WT2 * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    ev <- et(ev, id = 1:4)
    groupedDf <- .etGroupedSolveData(ev)
    groupedDf$id <- factor(groupedDf$id)
    iCov <- data.frame(id = 1:4, WT = c(70, 70, 80, 80))

    got <- as.data.frame(rxSolve(mod, groupedDf, params = c(KA = 1, CL = 7, V2 = 40), iCov = iCov))
    want <- as.data.frame(rxSolve(mod, as.data.frame(ev), params = c(KA = 1, CL = 7, V2 = 40), iCov = iCov))

    expect_equal(
      got[, c("id", "time", "depot", "centr")],
      want[, c("id", "time", "depot", "centr")]
    )
  })

  test_that("homogeneous grouped solve keeps iCov keep-column output", {
    mod <- rxode2({
      WT2 <- WT/70
      C2 <- centr / V2
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * WT2 * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    ev <- et(ev, id = 1:4)
    iCov <- data.frame(id = 1:4, WT = c(70, 70, 80, 80), grp = c("a", "a", "b", "b"))

    got <- suppressWarnings(
      as.data.frame(rxSolve(mod, ev, params = c(KA = 1, CL = 7, V2 = 40), iCov = iCov, keep = "grp"))
    )
    want <- suppressWarnings(
      as.data.frame(rxSolve(mod, as.data.frame(ev), params = c(KA = 1, CL = 7, V2 = 40), iCov = iCov, keep = "grp"))
    )

    expect_equal(
      got[, c("id", "time", "depot", "centr", "grp")],
      want[, c("id", "time", "depot", "centr", "grp")]
    )
  })

  test_that("homogeneous grouped iCov keep columns stay on grouped solve prep", {
    mod <- rxode2({
      WT2 <- WT/70
      C2 <- centr / V2
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * WT2 * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    ev <- et(ev, id = 1:4)
    iCov <- data.frame(id = 1:4, WT = c(70, 70, 80, 80), grp = c("a", "a", "b", "b"))

    prep <- .etGroupedSolveDataICov(ev, iCov, keep = "grp", modelParams = rxModelVars(mod)$params)

    expect_false(is.null(prep))
    expect_equal(attr(prep$events, "rxHomGroups"), list(1:2, 3:4))
    expect_equal(prep$iCov$id, 1:2)
    expect_equal(prep$iCov$grp, c("a", "b"))
  })

  test_that("homogeneous grouped iCov split ignores unrelated iCov columns", {
    mod <- rxode2({
      WT2 <- WT/70
      C2 <- centr / V2
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * WT2 * C2
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev <- et(ev, id = 1:4)
    iCov <- data.frame(id = 1:4, WT = c(70, 70, 70, 70), grp = c("a", "b", "c", "d"))

    prep <- .etGroupedSolveDataICov(ev, iCov, modelParams = rxModelVars(mod)$params)

    expect_false(is.null(prep))
    expect_equal(attr(prep$events, "rxHomGroups"), list(1:4))
    expect_equal(prep$iCov$id, 1L)
    expect_equal(prep$iCov$WT, 70)
  })

  test_that("grouped iCov split aborts on incomplete representative groups", {
    ev <- data.frame(id = 1L, time = 0, evid = 1L, amt = 100)
    attr(ev, "rxHomGroups") <- list(1:2, 3:4)
    attr(ev, "rxHomIdLevels") <- c("1", "2", "3", "4")
    iCov <- data.frame(id = 1:4, WT = c(70, 70, 80, 80))

    prep <- .etGroupedSolveDataFrameICov(ev, iCov, modelParams = "WT")

    expect_null(prep)
  })

  test_that("grouped iCov split aborts on duplicate iCov ids", {
    ev <- data.frame(id = c(1L, 2L), time = c(0, 0), evid = c(1L, 1L), amt = c(100, 100))
    attr(ev, "rxHomGroups") <- list(1:2)
    attr(ev, "rxHomIdLevels") <- c("1", "2")
    iCov <- data.frame(id = c(1, 1, 2), WT = c(70, 75, 80))

    prep <- .etGroupedSolveDataFrameICov(ev, iCov, modelParams = "WT")

    expect_null(prep)
  })

  test_that("grouped iCov split aborts on coercion-colliding iCov ids", {
    ev <- data.frame(id = c(1L, 2L), time = c(0, 0), evid = c(1L, 1L), amt = c(100, 100))
    attr(ev, "rxHomGroups") <- list(1:2)
    attr(ev, "rxHomIdLevels") <- c("1", "2")
    iCov <- data.frame(id = c("1", "01", "2"), WT = c(70, 75, 80))

    prep <- .etGroupedSolveDataFrameICov(ev, iCov, modelParams = "WT")

    expect_null(prep)
  })

  test_that("keep from iCov is case-insensitive and does not warn missing", {
    mod <- rxode2({
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL / V * centr
      cp <- centr / V
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    ev <- et(ev, id = 1:2)
    iCov <- data.frame(id = 1:2, wt = c(70, 80), sex = c("M", "F"))

    expect_no_warning(
      .sol <- rxSolve(mod, ev, params = c(KA = 1, CL = 7, V = 40),
                      iCov = iCov, keep = c("WT", "SEX"))
    )

    .nm <- tolower(names(as.data.frame(.sol)))
    expect_true("wt" %in% .nm)
    expect_true("sex" %in% .nm)
  })

  test_that("missing keep columns still warn when only subset is in iCov", {
    mod <- rxode2({
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL / V * centr
      cp <- centr / V
    })

    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    ev <- et(ev, id = 1:2)
    iCov <- data.frame(id = 1:2, wt = c(70, 80))

    .warnings <- character()
    withCallingHandlers(
      rxSolve(mod, ev, params = c(KA = 1, CL = 7, V = 40),
              iCov = iCov, keep = c("WT", "MISSING_COL")),
      warning = function(.w) {
        .warnings <<- c(.warnings, conditionMessage(.w))
        invokeRestart("muffleWarning")
      }
    )
    expect_true(any(grepl("Cannot keep missing columns", .warnings, fixed = TRUE)))
    expect_true(any(grepl("MISSING_COL", .warnings, fixed = TRUE)))
    expect_false(any(grepl("\\bWT\\b", .warnings)))
  })

  test_that("splitBolus expands source bolus doses to all target compartments", {
    modSplit <- rxode2parse("
      splitBolus(depot, depot, central, peripheral)
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - cl / v * central
      d/dt(peripheral) <- 0
    ")

    modBase <- rxode2parse("
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - cl / v * central
      d/dt(peripheral) <- 0
    ")

    eSplit <- et(time = 0, amt = 10, cmt = "depot", ii = 12, addl = 1)
    eBase <- et(time = 0, amt = 10, cmt = "depot", ii = 12, addl = 1) |>
      et(time = 0, amt = 10, cmt = "central", ii = 12, addl = 1) |>
      et(time = 0, amt = 10, cmt = "peripheral", ii = 12, addl = 1)

    got <- as.data.frame(etTrans(eSplit, modSplit, addCmt = TRUE, keepDosingOnly = TRUE))
    want <- as.data.frame(etTrans(eBase, modBase, addCmt = TRUE, keepDosingOnly = TRUE))

    got <- got[order(got$TIME, got$CMT, got$EVID), c("TIME", "CMT", "EVID", "AMT", "II")]
    want <- want[order(want$TIME, want$CMT, want$EVID), c("TIME", "CMT", "EVID", "AMT", "II")]

    expect_equal(got, want)
  })

  test_that("splitBolus can transfer a source bolus to one target compartment", {
    modSplit <- rxode2parse("
      splitBolus(depot, central)
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - cl / v * central
    ")

    modBase <- rxode2parse("
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - cl / v * central
    ")

    eSplit <- et(time = 0, amt = 10, cmt = "depot", ii = 12, addl = 1)
    eBase <- et(time = 0, amt = 10, cmt = "central", ii = 12, addl = 1)

    got <- as.data.frame(etTrans(eSplit, modSplit, addCmt = TRUE, keepDosingOnly = TRUE))
    want <- as.data.frame(etTrans(eBase, modBase, addCmt = TRUE, keepDosingOnly = TRUE))

    got <- got[order(got$TIME, got$CMT, got$EVID), c("TIME", "CMT", "EVID", "AMT", "II")]
    want <- want[order(want$TIME, want$CMT, want$EVID), c("TIME", "CMT", "EVID", "AMT", "II")]

    expect_equal(got, want)
  })

  test_that("splitBolus allows source repeats but rejects duplicate targets", {
    expect_no_error(rxode2parse("
      splitBolus(depot, central)
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - cl / v * central
    "))

    expect_no_error(rxode2parse("
      splitBolus(depot, depot, central)
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - cl / v * central
    "))

    expect_true(inherits(try(rxode2parse("
      splitBolus(depot, central, central)
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - cl / v * central
    "), silent = TRUE), "try-error"))
  })

  test_that("model vars expose empty splitBolus when unused", {
    mod <- rxode2parse("
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - cl / v * central
    ")

    expect_equal(length(rxModelVars(mod)$splitBolus), 0)
  })

  .Call(`_rxode2_etTransEvidIsObs`, FALSE)
  for (radi in 1:2) {

    forderForceBase(switch(radi,TRUE, FALSE))

    radix <- switch(radi,
                    "base::order",
                    "data.table::forder")

    # context(sprintf("etTrans checks (radix: %s)", radix))
    rxSetIni0(FALSE)

    mod <- rxode2parse("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
d/dt(blood)     = a*intestine - b*blood
")

    et <- structure(list(time = c(0, 0.05, 0.1, 0.2, 0.3, 0.5),
                         cmt = c("(default)", "(obs)", "intestine", "-intestine", "intestine", "out"),
                         amt = c(0.0833333333333333, NA, 3, NA, 3, 3),
                         rate = c(2, 0, 0, 0, 0, 0),
                         ii = c(1, 0, 3, 0, 3, 0),
                         addl = c(9L, 0L, 0L, 0L, 0L, 0L),
                         evid = c(1L, 2L, 1L, 2L, 1L, 1L),
                         ss = c(0L, 0L, 1L, 0L, 2L, 0L)),
                    class = "data.frame",
                    row.names = c(NA, -6L))

    ## et <- eventTable()
    ## et$add.dosing(
    ##   dose = 2 / 24, rate = 2, start.time = 0,
    ##   nbr.doses = 10, dosing.interval = 1
    ## )

    ## et <- et |>
    ##   et(0.05, evid = 2) |>
    ##   et(amt = 3, time = 0.5, cmt = out) |>
    ##   et(amt = 3, time = 0.1, cmt = intestine, ss = 1, ii = 3) |>
    ##   et(amt = 3, time = 0.3, cmt = intestine, ss = 2, ii = 3) |>
    ##   et(time = 0.2, cmt = "-intestine") |>
    ##   as.data.frame()

    test_that("error for empty data", {
      expect_error(suppressWarnings({
        etTrans(et, mod)
      }))
    })

    ett1 <- etTrans(et, mod, keepDosingOnly = TRUE)
    tmp1 <- sort(unique(ett1$EVID))

    et$cmt <- factor(et$cmt)
    ett2 <- etTrans(et, mod, keepDosingOnly = TRUE)

    test_that("factor and character give same compartment information", {
      expect_equal(attr(class(ett2), ".rxode2.lst")$cmtInfo, attr(class(ett1), ".rxode2.lst")$cmtInfo)
      expect_equal(attr(class(ett2), ".rxode2.lst")$cmtInfo, c("intestine", "blood", "out"))
    })

    test_that("factor and character give same evids", {
      expect_equal(ett1$EVID, ett2$EVID)
    })

    et0 <- et

    et$cmt <- paste(et$cmt)
    et$cmt[1:2] <- NA_character_

    ett1 <- etTrans(et, mod, keepDosingOnly = TRUE, addCmt = TRUE)

    test_that("string NA gives 1 for default compartment", {
      expect_equal(ett1$EVID, ett2$EVID)
    })

    et <- et0
    et$cmt[1:2] <- NA_integer_
    ett2 <- etTrans(et, mod, keepDosingOnly = TRUE, addCmt = TRUE)

    test_that("factor NA gives 1 for default compartment", {
      expect_equal(ett2$EVID, ett1$EVID)
    })

    et$cmt <- as.integer(et$cmt)

    et$cmt[1:2] <- NA_integer_

    ett2 <- etTrans(et, mod, keepDosingOnly = TRUE, addCmt = TRUE)

    test_that("factor NA gives 1 for default compartment", {
      expect_equal(ett2$EVID[1:2], ett1$EVID[1:2])
    })

    et <- structure(list(time = c(0, 0.05, 0.5), cmt = c("(default)", "(obs)",
                                                         "-out"), amt = c(0.0833333333333333, NA, NA), rate = c(2, 0,
                                                                                                                0), ii = c(1, 0, 0), addl = c(9L, 0L, 0L), evid = c(1L, 2L, 2L
                                                                                                                                                                    )), class = "data.frame", row.names = c(NA, -3L))
    ## et <- eventTable()
    ## et$add.dosing(
    ##   dose = 2 / 24, rate = 2, start.time = 0,
    ##   nbr.doses = 10, dosing.interval = 1
    ## )
    ## et <- et |>
    ##   et(0.05, evid = 2) |>
    ##   et(amt = 3, time = 0.5, cmt = "-out") |>
    ##   as.data.frame()

    et <- structure(list(time = c(0, 0.05, 0.5),
                         cmt = c("(default)", "(obs)", "-out"),
                         amt = c(0.0833333333333333, NA, NA),
                         rate = c(2, 0, 0),
                         ii = c(1, 0, 0),
                         addl = c(9L, 0L, 0L),
                         evid = c(1L, 2L, 2L)),
                    class = "data.frame",
                    row.names = c(NA, -3L))

    test_that("error for negative non ODE compartments", {
      expect_error(etTrans(et, mod, keepDosingOnly = TRUE))
      et$cmt <- factor(et$cmt)
      expect_error(etTrans(et, mod, keepDosingOnly = TRUE))
    })

    et <- structure(list(time = c(0, 0.05, 0.25, 0.5),
                         cmt = c("(default)", "(obs)", "out", "-out"),
                         amt = c(0.0833333333333333, NA, 3, NA),
                         rate = c(2, 0, 0, 0),
                         ii = c(1, 0, 0, 0),
                         addl = c(9L, 0L, 0L, 0L),
                         evid = c(1L, 2L, 1L, 2L)),
                    class = "data.frame", row.names = c(NA, -4L))

    test_that("error for negative non ODE compartments after defined compartment", {
      expect_error(etTrans(et, mod, keepDosingOnly = TRUE))
      et$cmt <- factor(et$cmt)
      expect_error(etTrans(et, mod, keepDosingOnly = TRUE))
    })

    et <- structure(list(time = c(0, 0.24), amt = c(NA_real_, 3),
                         evid = c(0L, 4L)), class = "data.frame", row.names = c(NA, -2L))

    test_that("EVID=4 makes sense", {
        expect_equal(
          etTrans(et, mod, keepDosingOnly = TRUE)$EVID,
          c(0L, 3L, 101L)
        )
    })


    mod <- rxode2parse("    CO = (187 * WT^0.81) * 60/1000
    QHT = 4 * CO/100
    QBR = 12 * CO/100
    QMU = 17 * CO/100
    QAD = 5 * CO/100
    QSK = 5 * CO/100
    QSP = 3 * CO/100
    QPA = 1 * CO/100
    QLI = 25.5 * CO/100
    QST = 1 * CO/100
    QGU = 14 * CO/100
    QHA = QLI - (QSP + QPA + QST + QGU)
    QBO = 5 * CO/100
    QKI = 19 * CO/100
    QRB = CO - (QHT + QBR + QMU + QAD + QSK + QLI + QBO + QKI)
    QLU = QHT + QBR + QMU + QAD + QSK + QLI + QBO + QKI + QRB
    VLU = (0.76 * WT/100)/1.051
    VHT = (0.47 * WT/100)/1.03
    VBR = (2 * WT/100)/1.036
    VMU = (40 * WT/100)/1.041
    VAD = (21.42 * WT/100)/0.916
    VSK = (3.71 * WT/100)/1.116
    VSP = (0.26 * WT/100)/1.054
    VPA = (0.14 * WT/100)/1.045
    VLI = (2.57 * WT/100)/1.04
    VST = (0.21 * WT/100)/1.05
    VGU = (1.44 * WT/100)/1.043
    VBO = (14.29 * WT/100)/1.99
    VKI = (0.44 * WT/100)/1.05
    VAB = (2.81 * WT/100)/1.04
    VVB = (5.62 * WT/100)/1.04
    VRB = (3.86 * WT/100)/1.04
    BP = 0.61
    fup = 0.028
    fub = fup/BP
    KbLU = exp(0.8334)
    KbHT = exp(1.1205)
    KbSK = exp(-0.5238)
    KbSP = exp(0.3224)
    KbPA = exp(0.3224)
    KbLI = exp(1.7604)
    KbST = exp(0.3224)
    KbGU = exp(1.2026)
    KbKI = exp(1.3171)
    S15 = VVB * BP/1000
    C15 = Venous_Blood/S15
    lnC15 = log(C15)
    d/dt(Lungs) = QLU * (Venous_Blood/VVB - Lungs/KbLU/VLU)
    d/dt(Heart) = QHT * (Arterial_Blood/VAB - Heart/KbHT/VHT)
    d/dt(Brain) = QBR * (Arterial_Blood/VAB - Brain/KbBR/VBR)
    d/dt(Muscles) = QMU * (Arterial_Blood/VAB - Muscles/KbMU/VMU)
    d/dt(Adipose) = QAD * (Arterial_Blood/VAB - Adipose/KbAD/VAD)
    d/dt(Skin) = QSK * (Arterial_Blood/VAB - Skin/KbSK/VSK)
    d/dt(Spleen) = QSP * (Arterial_Blood/VAB - Spleen/KbSP/VSP)
    d/dt(Pancreas) = QPA * (Arterial_Blood/VAB - Pancreas/KbPA/VPA)
    d/dt(Liver) = QHA * Arterial_Blood/VAB + QSP * Spleen/KbSP/VSP + QPA * Pancreas/KbPA/VPA + QST * Stomach/KbST/VST + QGU * Gut/KbGU/VGU - CLint * fub * Liver/KbLI/VLI - QLI * Liver/KbLI/VLI
    d/dt(Stomach) = QST * (Arterial_Blood/VAB - Stomach/KbST/VST)
    d/dt(Gut) = QGU * (Arterial_Blood/VAB - Gut/KbGU/VGU)
    d/dt(Bones) = QBO * (Arterial_Blood/VAB - Bones/KbBO/VBO)
    d/dt(Kidneys) = QKI * (Arterial_Blood/VAB - Kidneys/KbKI/VKI)
    d/dt(Arterial_Blood) = QLU * (Lungs/KbLU/VLU - Arterial_Blood/VAB)
    d/dt(Venous_Blood) = QHT * Heart/KbHT/VHT + QBR * Brain/KbBR/VBR + QMU * Muscles/KbMU/VMU + QAD * Adipose/KbAD/VAD + QSK * Skin/KbSK/VSK + QLI * Liver/KbLI/VLI + QBO * Bones/KbBO/VBO + QKI * Kidneys/KbKI/VKI + QRB * Rest_of_Body/KbRB/VRB - QLU * Venous_Blood/VVB
    d/dt(Rest_of_Body) = QRB * (Arterial_Blood/VAB - Rest_of_Body/KbRB/VRB)")


    test_that("strange rate doesn't affect model", {
      et1 <- test_path("etTrans1.rds")
      skip_if_not(file.exists(et1))
      dat <- readRDS(et1)
      expect_false(any(etTrans(dat, mod)$AMT < 0, na.rm = TRUE))
    })

    test_that("Missing evid gives the same results", {
      skip_if_not_installed("nlmixr2data")
      theoSd <- nlmixr2data::theo_sd
      d <- theoSd[, names(theoSd) != "EVID"]

      mod <- rxode2parse("
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        cp <- linCmt()
      ", linear=TRUE)

      t1 <- etTrans(theoSd, mod)
      t2 <- etTrans(d, mod)

      expect_equal(t1$ID, t2$ID)
      expect_equal(t1$TIME, t2$TIME)
      expect_equal(t1$EVID, t2$EVID)
      expect_equal(t1$AMT, t2$AMT)
      expect_equal(t1$II, t2$II)
      expect_equal(t1$DV, t2$DV)
    })

    ## Test non-standard inputs
    tmp <- structure(list(time = 0.24, amt = 3, evid = 4L), class = "data.frame", row.names = c(NA, -1L))

    for (col in c("ss", "evid", "dur", "amt", "addl", "dv", "mdv", "rate", "ii")) {
      et <- data.frame(col = "a", tmp[, names(tmp) != col], stringsAsFactors = FALSE)
      names(et)[1] <- col
      test_that(sprintf("Non-numeric inputs raise errors (%s)", col), {
        expect_error(etTrans(et, mod), col)
      })
    }

    ## Test dates
    d1 <- data.frame(DATE = c("10-1-86", "10-1-86", "10-2-86"), TIME = c("9:15", "14:40", "8:30"), stringsAsFactors = F)
    d1$DV <- 0

    d2 <- rbind(data.frame(ID = 1, d1, stringsAsFactors = F), data.frame(ID = 2, d1, stringsAsFactors = F))
    d2[d2$ID == 2, "DATE"] <- gsub("^10", "11", d2[d2$ID == 2, "DATE"])

    d3 <- d1
    d3$DATE <- c("10-1-1986", "10-1-1986", "10-2-1986")

    d4 <- d1
    d4$DATE <- c("10 1 1986", "10/1/86", "10-2-1986")

    test_that("DATE conversion works correctly", {
      tmp <- etTrans(d1, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d2, mod)
      expect_equal(c(0, 5.41666666666667, 23.25, 0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d3, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d4, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
    })

    ## Dat1= day month year
    d1 <- data.frame(DV = 0,
                     DAT1 = c("1-10-86", "1-10-86", "2-10-86"),
                     TIME = c("9:15", "14:40", "8:30"),
                     stringsAsFactors = FALSE)

    d2 <- rbind(data.frame(ID = 1, d1, stringsAsFactors = F),
                data.frame(ID = 2, d1, stringsAsFactors = F))
    d2[d2$ID == 2, "DAT1"] <- gsub("-10-", "-11-", d2[d2$ID == 2, "DAT1"])

    d3 <- d1
    d3$DAT1 <- c("1-10-1986", "1-10-1986", "2-10-1986")

    d4 <- d1
    d4$DAT1 <- c("1-10-1986", "1-10-86", "2-10-1986")

    test_that("DAT1 conversion works correctly", {
      tmp <- etTrans(d1, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d2, mod)
      expect_equal(c(0, 5.41666666666667, 23.25, 0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d3, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d4, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
    })

    ## Dat2 = year month day
    d1 <- data.frame(DAT2 = c("86-10-1", "86-10-1", "86-10-2"),
                     TIME = c("9:15", "14:40", "8:30"),
                     stringsAsFactors = FALSE)
    d1$DV <- 0

    d2 <- rbind(data.frame(ID = 1, d1, stringsAsFactors = F),
                data.frame(ID = 2, d1, stringsAsFactors = F))
    d2[d2$ID == 2, "DAT2"] <- gsub("-10-", "-11-", d2[d2$ID == 2, "DAT2"])

    d3 <- d1
    d3$DAT2 <- c("1986-10-1", "1986-10-1", "1986-10-2")

    d4 <- d1
    d4$DAT2 <- c("1986-10-1", "86-10-1", "1986-10-2")

    test_that("DAT2 conversion works correctly", {
      tmp <- etTrans(d1, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d2, mod)
      expect_equal(c(0, 5.41666666666667, 23.25, 0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d3, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d4, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
    })

    ## DAT3 conversion
    d1 <- data.frame(DAT3 = c("86-1-10", "86-1-10", "86-2-10"), TIME = c("9:15", "14:40", "8:30"), stringsAsFactors = F)
    d1$DV <- 0

    d2 <- rbind(data.frame(ID = 1, d1, stringsAsFactors = F), data.frame(ID = 2, d1, stringsAsFactors = F))
    d2[d2$ID == 2, "DAT3"] <- gsub("-10$", "-11", d2[d2$ID == 2, "DAT3"])

    d3 <- d1
    d3$DAT3 <- c("1986-1-10", "1986-1-10", "1986-2-10")

    d4 <- d1
    d4$DAT3 <- c("1986-1-10", "86-1-10", "1986-2-10")

    test_that("DAT3 conversion works correctly", {
      tmp <- etTrans(d1, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d2, mod)
      expect_equal(c(0, 5.41666666666667, 23.25, 0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d3, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
      tmp <- etTrans(d4, mod)
      expect_equal(c(0, 5.41666666666667, 23.25), tmp$TIME)
    })

    d1 <- data.frame(DV = 0,
                     DATE = c("10-1-86", "10-1-86", "10-2-86"),
                     TIME = c("9:15", "14:40", "8:30"),
                     stringsAsFactors = FALSE)

    d2 <- d1
    d2$DAT1 <- d2$DATE

    d3 <- d1
    d3$DAT2 <- d3$DATE

    d4 <- d1
    d4$DAT3 <- d4$DATE

    test_that("Multiple DATE errors", {
      expect_error(etTrans(d2, mod))
      expect_error(etTrans(d3, mod))
      expect_error(etTrans(d4, mod))
    })


    d1 <- data.frame(DV = 0,
                     DATE = c("10-1-86", "10-1-86", "10-2-86"),
                     TIME = c("9.15", "14:40", "8:30"),
                     stringsAsFactors = FALSE)

    test_that("Bad Date/Time combination", {
      expect_error(etTrans(d1, mod))
    })

    ## Test mixed classic rxode2 and NONMEM inputs
    test_that("mixed rxode2/NONMEM EVID/data gives a warning", {

      mod <- rxode2parse("
        d1 <- exp(td1 + eta.d1)
        cl <- exp(tcl + eta.cl)
        d/dt(center) <- -cl / v * center
        dur(center) <- d1
        cp <- center / v
      ")


      d <- structure(list(
        ID = c(1L, 1L, 1L),
        TIME = c(0, 0, 0.25),
        DV = c(0, 0.74, 2.84),
        AMT = c(319.992, 0, 0),
        EVID = c(101L, 0L, 0L),
        WT = c(79.6, 79.6, 79.6),
        RATE = c(-2, 0, 0)),
        row.names = c(NA, 3L), class = "data.frame")

      expect_warning(etTrans(d, mod), "'rate'")

      d <- structure(list(
        ID = c(1L, 1L, 1L),
        TIME = c(0, 0, 0.25),
        DV = c(0, 0.74, 2.84),
        AMT = c(319.992, 0, 0),
        EVID = c(101L, 0L, 0L),
        WT = c(79.6, 79.6, 79.6),
        DUR = c(-2, 0, 0)),
        row.names = c(NA, 3L), class = "data.frame")

      expect_warning(etTrans(d, mod), "'dur'")

      d <- structure(list(
        ID = c(1L, 1L, 1L),
        TIME = c(0, 0, 0.25),
        DV = c(0, 0.74, 2.84),
        AMT = c(319.992, 0, 0),
        EVID = c(101L, 0L, 0L),
        WT = c(79.6, 79.6, 79.6),
        SS = c(1, 0, 0),
        II = c(24, 0, 0)),
        row.names = c(NA, 3L), class = "data.frame")

      expect_warning(etTrans(d, mod), "'ss'")

    })

    mod <- rxode2parse("    x1(0) = x10\n    d/dt(x1) = a * x1\n    Volume = x1;\ncmt(Volume);\n\n    nlmixr_pred <- Volume")

    test_that("DV=NA; issue #106", {

      RawData2 <- data.frame(
        ID = c(1, 1, 1, 1, 2, 2, 2, 2),
        TIME = c(0, 3, 4, 5, 0, 3, 4, 5),
        DV = c(NA, 30, 80, 250, NA, 40, 150, 400))

      dat1 <- etTrans(RawData2, mod)

      RawData2a <- data.frame(
        ID = c(1, 1, 1, 1, 2, 2, 2, 2),
        TIME = c(0, 3, 4, 5, 0, 3, 4, 5),
        DV = c(NA, 30, 80, 250, NA, 40, 150, 400),
        AMT = c(NA, NA, NA, NA, NA, NA, NA, NA))

      dat1a <- etTrans(RawData2a, mod)

      RawData2b <- data.frame(
        ID = c(1, 1, 1, 1, 2, 2, 2, 2),
        TIME = c(0, 3, 4, 5, 0, 3, 4, 5),
        DV = c(NA, 30, 80, 250, NA, 40, 150, 400),
        AMT = c(0, 0, 0, 0, 0, 0, 0, 0))

      dat1b <- etTrans(RawData2b, mod)

      RawData2c <- data.frame(
        ID = c(1, 1, 1, 1, 2, 2, 2, 2),
        TIME = c(0, 3, 4, 5, 0, 3, 4, 5),
        DV = c(NA, 30, 80, 250, NA, 40, 150, 400),
        AMT = c(1, 0, 0, 0, 1, 0, 0, 0))

      dat1c <- etTrans(RawData2c, mod)

      expect_equal(dat1a$EVID, c(2L, 0L, 0L, 0L, 2L, 0L, 0L, 0L))
      expect_equal(dat1a$EVID, dat1b$EVID)
      expect_equal(dat1c$EVID, c(101L, 0L, 0L, 0L, 101L, 0L, 0L, 0L))

    })

    RawData3 <- data.frame(
      ID = c(1, 1, 1, 1, 2, 2, 2, 2),
      TIME = c(0, 3, 4, 5, 0, 3, 4, 5),
      DV = c(0, 30, 80, 250, 0, 40, 150, 400),
      EVID = c(2, 0, 0, 0, 2, 0, 0, 0))

    dat2 <- etTrans(RawData3, mod)

    RawData4 <- data.frame(
      ID = c(1, 1, 1, 1, 2, 2, 2, 2),
      TIME = c(0, 3, 4, 5, 0, 3, 4, 5),
      DV = c(0, 30, 80, 250, 0, 40, 150, 400),
      EVID = c(2, 0, 0, 0, 2, 0, 0, 0),
      CMT = c(1, 0, 0, 0, 1, 0, 0, 0))

    dat3 <- etTrans(RawData4, mod)

    RawData5 <- data.frame(
      ID = c(1, 1, 1, 1, 2, 2, 2, 2),
      TIME = c(0, 3, 4, 5, 0, 3, 4, 5),
      DV = c(0, 30, 80, 250, 0, 40, 150, 400),
      EVID = c(2, 0, 0, 0, 2, 0, 0, 0),
      CMT = c(2, 0, 0, 0, 2, 0, 0, 0)
    )

    dat4 <- etTrans(RawData5, mod)

    test_that("dat2=dat4", {
      dat2 <- as.data.frame(dat2)
      dat4 <- as.data.frame(dat4)
      dat4 <- dat4[, names(dat2)]
      expect_equal(dat2, dat4)
    })

    test_that("dat3 has evid w/amt 0", {
      expect_equal(dat3$EVID, c(160L, 2L, 0L, 0L, 0L, 160L, 2L, 0L, 0L, 0L))
      expect_equal(dat3$AMT, c(0, NA, NA, NA, NA, 0, NA, NA, NA, NA))
    })

    test_that("X(0) should be at time zero; see issue #105", {

      mod <- rxode2parse("    x1(0) = x10\n    d/dt(x1) = a * x1\n    Volume = x1;\ncmt(Volume);\n\n    nlmixr_pred <- Volume")

      rxSetIni0(FALSE)

      RawData2 <- data.frame(
        ID = c(1, 1, 1, 2, 2, 2),
        TIME = c(3, 4, 5, 3, 4, 5),
        DV = c(30, 80, 250, 40, 150, 400))

      expect_warning(dat1 <- etTrans(RawData2, mod))

      expect_equal(dat1$TIME, RawData2$TIME)

      rxSetIni0(TRUE)
      dat1 <- etTrans(RawData2, mod)

      expect_equal(dat1$TIME, c(0, 3, 4, 5, 0, 3, 4, 5))
      expect_equal(dat1$EVID, c(9L, 0L, 0L, 0L, 9L, 0L, 0L, 0L))

    })

    rxSetIni0(TRUE)
    test_that("censoring checks", {

      mod <- rxode2parse("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
d/dt(blood)     = a*intestine - b*blood
")

      et <- structure(list(time = c(0, 0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7,
                                    0.8, 0.9, 1, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2,
                                    2.1, 2.2, 2.3, 2.4, 2.5, 2.6, 2.7, 2.8, 2.9, 3, 3.1, 3.2, 3.3,
                                    3.4, 3.5, 3.6, 3.7, 3.8, 3.9, 4, 4.1, 4.2, 4.3, 4.4, 4.5, 4.6,
                                    4.7, 4.8, 4.9, 5, 5.1, 5.2, 5.3, 5.4, 5.5, 5.6, 5.7, 5.8, 5.9,
                                    6, 6.1, 6.2, 6.3, 6.4, 6.5, 6.6, 6.7, 6.8, 6.9, 7, 7.1, 7.2,
                                    7.3, 7.4, 7.5, 7.6, 7.7, 7.8, 7.9, 8, 8.1, 8.2, 8.3, 8.4, 8.5,
                                    8.6, 8.7, 8.8, 8.9, 9, 9.1, 9.2, 9.3, 9.4, 9.5, 9.6, 9.7, 9.8,
                                    9.9, 10, 10.1, 10.2, 10.3, 10.4, 10.5, 10.6, 10.7, 10.8, 10.9,
                                    11, 11.1, 11.2, 11.3, 11.4, 11.5, 11.6, 11.7, 11.8, 11.9, 12,
                                    12.1, 12.2, 12.3, 12.4, 12.5, 12.6, 12.7, 12.8, 12.9, 13, 13.1,
                                    13.2, 13.3, 13.4, 13.5, 13.6, 13.7, 13.8, 13.9, 14, 14.1, 14.2,
                                    14.3, 14.4, 14.5, 14.6, 14.7, 14.8, 14.9, 15, 15.1, 15.2, 15.3,
                                    15.4, 15.5, 15.6, 15.7, 15.8, 15.9, 16, 16.1, 16.2, 16.3, 16.4,
                                    16.5, 16.6, 16.7, 16.8, 16.9, 17, 17.1, 17.2, 17.3, 17.4, 17.5,
                                    17.6, 17.7, 17.8, 17.9, 18, 18.1, 18.2, 18.3, 18.4, 18.5, 18.6,
                                    18.7, 18.8, 18.9, 19, 19.1, 19.2, 19.3, 19.4, 19.5, 19.6, 19.7,
                                    19.8, 19.9, 20, 20.1, 20.2, 20.3, 20.4, 20.5, 20.6, 20.7, 20.8,
                                    20.9, 21, 21.1, 21.2, 21.3, 21.4, 21.5, 21.6, 21.7, 21.8, 21.9,
                                    22, 22.1, 22.2, 22.3, 22.4, 22.5, 22.6, 22.7, 22.8, 22.9, 23,
                                    23.1, 23.2, 23.3, 23.4, 23.5, 23.6, 23.7, 23.8, 23.9, 24),
                           amt = c(NA,
                                   0.0833333333333333, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                   NA, NA, NA, NA, NA),
                           rate = c(NA, 2, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA),
                           ii = c(NA, 1, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA),
                           addl = c(NA,
                                    9L, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                    NA),
                           evid = c(0L, 1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                    0L, 0L, 0L, 0L, 0L, 0L)),
                      class = "data.frame", row.names = c(NA, -242L))

      tmp <- et
      tmp$cens <- 0
      tmp$cens[1] <- 2

      expect_error(etTrans(tmp, mod))

      tmp <- et
      tmp$cens <- 0
      tmp$dv <- 3
      tmp$cens[2] <- 1

      ret <- suppressWarnings(etTrans(tmp, mod))
      expect_false(any(names(ret) == "CENS"))
      expect_equal(attr(class(ret), ".rxode2.lst")$censAdd, 0L)
      expect_equal(attr(class(ret), ".rxode2.lst")$limitAdd, 0L)

      tmp <- et
      tmp$cens <- 0
      tmp$dv[1] <- 2
      tmp$cens <- 0
      tmp$cens[1] <- 1

      ret <- etTrans(tmp, mod)
      expect_true(any(names(ret) == "CENS"))
      expect_equal(attr(class(ret), ".rxode2.lst")$censAdd, 1L)
      expect_equal(attr(class(ret), ".rxode2.lst")$limitAdd, 0L)

      tmp$limit <- 0

      ret <- etTrans(tmp, mod)
      expect_true(any(names(ret) == "CENS"))
      expect_true(any(names(ret) == "LIMIT"))
      expect_equal(attr(class(ret), ".rxode2.lst")$censAdd, 1L)
      expect_equal(attr(class(ret), ".rxode2.lst")$limitAdd, 1L)
    })

    test_that("rxode2 constant infusion taken to steady state", {

      et <- structure(list(time = 0, amt = 0, rate = 10, ii = 0, evid = 1L,
                           ss = 1L), class = "data.frame", row.names = c(NA, -1L))

      trn1 <- etTrans(et, mod, keepDosingOnly = TRUE) |> as.data.frame()

      expect_equal(structure(list(
        ID = structure(1L, class = "factor", levels = "1"),
        TIME = 0, EVID = 10140L, AMT = 10, II = 0, DV = NA_real_
      ),
      class = "data.frame", row.names = c(NA, -1L)
      ), trn1)

      et <- structure(list(time = 0, amt = 0, rate = -1, ii = 0, evid = 1L,
                           ss = 1L), class = "data.frame", row.names = c(NA, -1L))

      trn1 <- etTrans(et, mod, keepDosingOnly = TRUE) |> as.data.frame()

      expect_equal(structure(list(
        ID = structure(1L, class = "factor", levels = "1"),
        TIME = 0, EVID = 90140L, AMT = 0, II = 0, DV = NA_real_
      ),
      class = "data.frame", row.names = c(NA, -1L)
      ), trn1)
    })

    ## etTrans example from xgxr + nlmixr + ggpmx
    test_that("etTrans", {
      lst <- readRDS(test_path("test-etTrans-1.rds"))
      events2 <- lst$events
      events2 <- events2[, names(events2) != "CENS"]

      # suppressWarnings() is used on the outside because the rxSetIni0(FALSE)
      # warning only occurs once per session
      t0 <- suppressWarnings(etTrans(events2, rxode2parse(lst$object),
                                          FALSE, FALSE, FALSE, FALSE, NULL, character(0)))
      expect_s3_class(t0, "rxEtTran")

      t1 <- etTrans(events2, rxode2parse(lst$object),
                         FALSE, FALSE, FALSE, TRUE, NULL, character(0))
      expect_s3_class(t1, "rxEtTran")
    })

    test_that("etTrans drop levels are correct", {

      dat <- readRDS(test_path("etTrans-drop.rds"))

      mod <- rxode2parse("
        lka <- log(0.1) # log Ka
        lv <- log(10) # Log Vc
        lcl <- log(4) # Log Cl
        lq <- log(10) # log Q
        lvp <- log(20) # Log Vp
        eta.ka <- 0
        eta.v <- 0.1
        eta.cl <- 0.1
        ka <- exp(lka + eta.ka)
        cl <- exp(lcl + eta.cl)
        v <- exp(lv + eta.v)
        q <- exp(lq)
        vp <- exp(lvp)
        cp <- linCmt()
      ", linear = TRUE)

      # suppressWarnings() is used on the outside because the rxSetIni0(FALSE)
      # warning only occurs once per session
      suppressWarnings(expect_warning(expect_warning(
        tmp <- etTrans(dat, mod),
        regexp="while censoring is included"), regexp="IDs without observations"
        ))
      lvls <- c(
        "32", "33", "35", "36", "37", "40", "41", "42", "43", "47",
        "48", "49", "50", "51", "54", "55", "57", "59", "61", "62", "63",
        "64", "65", "66", "67", "68", "69", "70", "71", "72", "73", "74",
        "75", "76", "77", "78", "79", "80", "81", "82", "83", "84", "85",
        "86", "87", "88", "89", "90", "91", "92", "93", "94", "95", "96",
        "97", "98", "99", "100", "101", "102", "103", "104", "105", "106",
        "107", "108", "109", "110", "111", "112", "113", "114", "115",
        "116", "117", "118", "119", "120", "121", "122", "123", "124",
        "125", "126", "127", "128", "129", "130", "131", "132", "133",
        "134", "135", "136", "137", "138", "139", "140", "141", "142",
        "143", "144", "145", "146", "147", "148", "149", "150", "151",
        "152", "153", "154", "155", "156", "157", "158", "159", "160",
        "161", "162", "163", "164", "165", "166", "167", "168", "169",
        "170", "171", "172", "173", "174", "175", "176", "177", "178",
        "179", "180"
      )
      expect_equal(attr(class(tmp), ".rxode2.lst")$idLvl, lvls)
      expect_equal(levels(tmp$ID), lvls)
    })

    test_that("phantom doses", {

      mod <- rxode2parse("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
d/dt(blood)     = a*intestine - b*blood
")

      d <- structure(list(time = c(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10),
                          cmt = c(2L, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA),
                          amt = c(3, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA),
                          evid = c(7L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L)),
                     class = "data.frame", row.names = c(NA, -11L))

      tran <- etTrans(d, mod)

      expect_equal(tran$EVID[1], 250L)

    })

    test_that("non time varying covariates with evid=9", {

      rx <- rxode2parse("
        param(lkng, ltau, lec50, kmax, propErr, addErr, cp, tumor0)
        kng <- exp(lkng)
        tau <- exp(ltau)
        taulast <- tau
        ec50 <- exp(lec50)
        edrug <- kmax * cp/(ec50 + cp)
        tumor(0) <- tumor0
        d/dt(transit1) <- (edrug - transit1)/tau
        d/dt(transit2) <- (transit1 - transit2)/tau
        d/dt(transit3) <- (transit2 - transit3)/tau
        d/dt(transitlast) <- transit3/tau - transitlast/taulast
        d/dt(tumor) <- kng * tumor - transitlast * tumor
        rx_yj_ ~ 2
        rx_lambda_ ~ 1
        rx_low_ ~ 0
        rx_hi_ ~ 1
        rx_pred_f_ ~ tumor
        rx_pred_ ~ rx_pred_f_
        rx_r_ ~ (addErr)^2 + (rx_pred_f_)^2 * (propErr)^2
        ipredSim <- rxTBSi(rx_pred_, rx_lambda_, rx_yj_, rx_low_,
                           rx_hi_)
        sim <- rxTBSi(rx_pred_ + sqrt(rx_r_) * err.tumor, rx_lambda_,
                      rx_yj_, rx_low_, rx_hi_)
        dvid(5)
      ")

      prepfit <- readRDS(test_path("etTrans-prepfit.rds"))

      trans <- etTrans(prepfit,rx)

      expect_true(all(names(trans) != "tumor0"))
      expect_true(any(names(attr(class(trans), ".rxode2.lst")$cov1) =="tumor0"))

    })
  }

  .Call(`_rxode2_etTransEvidIsObs`, TRUE)

  test_that("test etTran on addl ss items", {

    rx <- rxode2parse("
      cp <- linCmt(ka, cl, v)
    ", linear=TRUE)

    e <- structure(list(time = c(0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
                                 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26,
                                 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42,
                                 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58,
                                 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74,
                                 75, 76, 77, 78, 79, 80),
                        cmt = c(NA, 2L, NA, NA, NA, NA, NA,
                                NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA),
                        amt = c(NA, 100,
                                NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA
                                ),
                        rate = c(NA, 10, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                 NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                 NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                 NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                 NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                 NA, NA, NA, NA, NA),
                        ii = c(NA, 24, NA, NA, NA, NA, NA, NA, NA,
                               NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                               NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                               NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                               NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                               NA, NA, NA, NA, NA, NA, NA, NA, NA),
                        addl = c(NA, 3L, NA, NA,
                                 NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                 NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                 NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                 NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                 NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA),
                        evid = c(0L,
                                 1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                 0L),
                        ss = c(NA, 1L, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                               NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                               NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                               NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                               NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                               NA, NA, NA, NA, NA)),
                   class = "data.frame",
                   row.names = c(NA, -82L))

    # should not drop the off infusion record
    t <- etTrans(e, rx, addlDropSs=FALSE)

    expect_equal(t$TIME[length(t$TIME)], 82)
    expect_equal(t$AMT[length(t$AMT)], -10)
    expect_equal(t$EVID[length(t$EVID)], 10210L)
    expect_equal(t$II[length(t$II)], 0)

    t2 <- t |> dplyr::filter(AMT>0)

    expect_equal(t2$TIME, c(0, 24, 48, 72))
    expect_true(all(t2$AMT == 10))
    expect_true(all(t2$EVID == 10210L))
  })


  test_that("warning for all na", {

    mod1 <- rxode2parse("
      d/dt(A_centr) <- -A_centr * (CLI / V1I + 204 / V1I) + 204 * A_periph / V2I
      d/dt(A_periph) <- 204 * A_centr / V1I - 204 * A_periph / V2I
      d/dt(A_circ) <- -4 * A_circ * exp(-ETA[2] - THETA[2]) + 4 * A_tr3 * exp(-ETA[2] - THETA[2])
      A_circ(0) <- exp(ETA[1] + THETA[1])
      d/dt(A_prol) <- 4 * A_prol * Rx_pow(exp(ETA[1] + THETA[1]) / A_circ, exp(THETA[4])) * (-A_centr * exp(ETA[3] + THETA[3]) / V1I + 1) * exp(-ETA[2] - THETA[2]) - 4 * A_prol * exp(-ETA[2] - THETA[2])
      A_prol(0) <- exp(ETA[1] + THETA[1])
      d/dt(A_tr1) <- 4 * A_prol * exp(-ETA[2] - THETA[2]) - 4 * A_tr1 * exp(-ETA[2] - THETA[2])
      A_tr1(0) <- exp(ETA[1] + THETA[1])
      d/dt(A_tr2) <- 4 * A_tr1 * exp(-ETA[2] - THETA[2]) - 4 * A_tr2 * exp(-ETA[2] - THETA[2])
      A_tr2(0) <- exp(ETA[1] + THETA[1])
      d/dt(A_tr3) <- 4 * A_tr2 * exp(-ETA[2] - THETA[2]) - 4 * A_tr3 * exp(-ETA[2] - THETA[2])
      A_tr3(0) <- exp(ETA[1] + THETA[1])
    ")

    d3na <- data.frame(
      ID = c(1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L),
      TIME = c(0, 0, 2.99270072992701, 192, 336, 456, 0, 0, 3.07272727272727, 432),
      AMT = c(137L, 0L, -137L, 0L, 0L, 0L, 110L, 0L, -110L, 0L),
      V2I = c(909L, NA_integer_, 909L, 909L, 909L, 909L, 942L, 942L, 942L, 942L),
      V1I = c(545L, 545L, 545L, 545L, 545L, 545L, NA_integer_, NA_integer_, NA_integer_, NA_integer_),
      CLI = c(471L, 471L, 471L, 471L, NA_integer_, 471L, 405L, 405L, 405L, 405L),
      EVID = c(10101L, 0L, 10101L, 0L, 0L, 0L, 10101L, 0L, 10101L, 0L)
    )

    expect_warning(etTrans(d3na, mod1, addlDropSs=FALSE), "column 'V1I' has only 'NA' values for id '2'")

    d3na <-
      data.frame(
        ID = c(1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L),
        TIME = c(0, 0, 2.99270072992701, 192, 336, 456, 0, 0, 3.07272727272727, 432),
        AMT = c(137L, 0L, -137L, 0L, 0L, 0L, 110L, 0L, -110L, 0L),
        V2I = c(909L, NA_integer_, 909L, 909L, 909L, 909L, 942L, 942L, 942L, 942L),
        V1I = c(545L, 545L, 545L, 545L, 545L, 545L, 306L, 306L, 306L, NA_integer_),
        CLI = c(471L, 471L, 471L, 471L, NA_integer_, 471L, 405L, 405L, 405L, 405L),
        EVID = c(10101L, 0L, 10101L, 0L, 0L, 0L, 10101L, 0L, 10101L, 0L)
      )

    expect_warning(etTrans(d3na, mod1, addlDropSs=FALSE), NA)

  })

  test_that("na ids give error", {

    mod1 <- rxode2parse("
    mw_anon <- 50000
    mw_convert_anon <- 1 / mw_anon * 1e3
    kel_anon <- log(2)/(hl_anon/60/24)
    kel_target <- log(2)/hl_target
    kform_target <- conc_target_ss*kel_target
    kd_anon_target_umolL <- kd_anon_target/1000
    d/dt(depot_anon) <- -ka_anon*depot_anon
    d/dt(central_anon) <- ka_anon*depot_anon - kel_anon*central_anon
    # Calculate bound concentration
    central_anon_umolL <- central_anon/vc_anon*mw_convert_anon # Unit conversion from mg/L to umol/L
    totalconc <- central_anon_umolL + central_target + kd_anon_target_umolL
    bound_umolL <- (totalconc - sqrt(totalconc^2 - 4*central_anon_umolL*central_target))/2
    free_central_target <- central_target - bound_umolL
    d/dt(central_target) <- kform_target - kel_target*free_central_target - kel_anon*bound_umolL # Units are umol/L
    f(depot_anon) <- f_anon
    central_target(0) <- conc_target_ss
    ")

    mydata <-
      structure(list(conc_target_ss = c(NA, 0.12, NA, NA, NA, NA,
                                        NA, NA, NA, NA, NA, NA, NA, 0.12, NA, NA, NA, NA, NA, NA, NA,
                                        NA, NA, NA), hl_target = c(NA, 5, NA, NA, NA, NA, NA, NA, NA,
                                                                   NA, NA, NA, NA, 5, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA), kd_anon_target = c(NA,
                                                                                                                                                  1, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 1, NA, NA, NA,
                                                                                                                                                  NA, NA, NA, NA, NA, NA, NA), ID = c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
                                                                                                                                                                                      1L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L), CMT = c(NA,
                                                                                                                                                                                                                                                   "central_anon", NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                                                                                                                                                                                                                                   "central_anon", NA, NA, NA, NA, NA, NA, NA, NA, NA, NA), EVID = c(0, 1, 0, 0, 0,
                                                                                                                                                                                                                                                                                                                     0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0), TIME = c(-0.01,
                                                                                                                                                                                                                                                                                                                                                                                        0, 0, 0.001, 0.00215443469003188, 0.00464158883361278, 0.01,
                                                                                                                                                                                                                                                                                                                                                                                        0.0215443469003188, 0.0464158883361278, 0.1, 0.215443469003188,
                                                                                                                                                                                                                                                                                                                                                                                        0.464158883361278, -0.01, 0, 0, 0.001, 0.00215443469003188, 0.00464158883361278,
                                                                                                                                                                                                                                                                                                                                                                                        0.01, 0.0215443469003188, 0.0464158883361278, 0.1, 0.215443469003188,
                                                                                                                                                                                                                                                                                                                                                                                        0.464158883361278), AMT = c(NA, 100, NA, NA, NA, NA, NA, NA,
                                                                                                                                                                                                                                                                                                                                                                                                                    NA, NA, NA, NA, NA, 100, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                                                                                                                                                                                                                                                                                                                                                                                                    NA), II = c(NA, 7, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                                                                                                                                                                                                                                                                                                                                                                                                                14, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA), ADDL = c(NA, 11,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 5, NA, NA, NA, NA,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      NA, NA, NA, NA, NA, NA)), row.names = c(NA, -24L), class = c("tbl_df",
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   "tbl", "data.frame"))

    mydata$ID[mydata$EVID == 0] <- NA

    expect_error(etTrans(mydata, mod1, addlDropSs=FALSE))

    mydata$ID <- as.double(mydata$ID)

    expect_error(etTrans(mydata, mod1, addlDropSs=FALSE))

    mydata$ID <- as.character(mydata$ID)

    expect_error(etTrans(mydata, mod1, addlDropSs=FALSE))

  })


  test_that("error with missing 'amt' but dosing evid", {

    mod <- rxode2parse("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
d/dt(blood)     = a*intestine - b*blood
")

    dSimple <-
      data.frame(ID = 1, EVID = c(1, 0), cmt = c("depot", "central"), DV = c(NA, 1), TIME = 0:1)
    expect_error(etTrans(dSimple, mod, addlDropSs=FALSE), "EVID=1")

    dSimple <-
      data.frame(ID = 1, EVID = c(1, 0), cmt = c("depot", "central"), DV = c(NA, 1), TIME = 0:1, AMT=NA)
    expect_error(etTrans(dSimple, mod, addlDropSs=FALSE), "evid: 1")

    dSimple <-
      data.frame(ID = 1, EVID = c(7, 0), cmt = c("depot", "central"), DV = c(NA, 1), TIME = 0:1)
    expect_error(etTrans(dSimple, mod, addlDropSs=FALSE), "EVID=7")

    dSimple <-
      data.frame(ID = 1, EVID = c(7, 0), cmt = c("depot", "central"), DV = c(NA, 1), TIME = 0:1, amt=NA)
    expect_error(etTrans(dSimple, mod, addlDropSs=FALSE), "evid: 7")

    dSimple <-
      data.frame(ID = 1,
                 EVID = c(0, 4, 0),
                 cmt = c("central", "depot", "central"),
                 DV = c(1, NA, 1),
                 TIME = c(0, 0, 1))

    expect_error(etTrans(dSimple, mod, addlDropSs=FALSE), "EVID=4")

    dSimple <-
      data.frame(ID = 1, EVID = c(0, 4, 0), cmt = c("central", "depot", "central"), DV = c(1, NA, 1), TIME = c(0, 0:1), amt=NA)
    expect_error(etTrans(dSimple, mod, addlDropSs=FALSE), "evid: 4")


    dSimple <-
      data.frame(ID = 1, EVID = c(5, 0), cmt = c("depot", "central"), DV = c(NA, 1), TIME = 0:1)
    expect_error(etTrans(dSimple, mod, addlDropSs=FALSE), "EVID=5")

    dSimple <-
      data.frame(ID = 1, EVID = c(5, 0), cmt = c("depot", "central"), DV = c(NA, 1), TIME = 0:1, amt=NA)
    expect_error(etTrans(dSimple, mod, addlDropSs=FALSE), "evid: 5")

    dSimple <-
      data.frame(ID = 1, EVID = c(6, 0), cmt = c("depot", "central"), DV = c(NA, 1), TIME = 0:1)
    expect_error(etTrans(dSimple, mod, addlDropSs=FALSE), "EVID=6")

    dSimple <-
      data.frame(ID = 1, EVID = c(7, 0), cmt = c("depot", "central"), DV = c(NA, 1), TIME = 0:1, amt=NA)
    expect_error(etTrans(dSimple, mod, addlDropSs=FALSE), "evid: 7")
  })

  test_that("etTrans lag ss", {
    mod <- rxode2parse("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
alag(intestine) = lag
d/dt(blood)     = a*intestine - b*blood
")

    d <- data.frame(time=c(0, 1),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(1, 0))

    tmp <- etTrans(d, mod)

    expect_equal(tmp$TIME, c(0, 0, 1))
    expect_equal(tmp$EVID, c(109L, 101L, 0L))
    expect_equal(tmp$II, c(24, 0, 0))
    expect_equal(tmp$AMT, c(100, 100, NA))


    d <- data.frame(time=c(0, 1),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(2, 0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 1))
    expect_equal(tmp$EVID, c(119L, 101L, 0L))
    expect_equal(tmp$II, c(24, 0, 0))
    expect_equal(tmp$AMT, c(100, 100, NA))

    d <- data.frame(time=c(0, 1),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(1, 0),
                    rate=c(5, 0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 0, 1, 20))
    expect_equal(tmp$EVID, c(10109L, 10108L, 10101L, 0L, 10101L))
    expect_equal(tmp$II, c(24, 24, 0, 0, 0))
    expect_equal(tmp$AMT, c(5, -5, 5, NA, -5))


    d <- data.frame(time=c(0, 1),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(1, 0),
                    dur=c(20, 0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 0, 1, 20))
    expect_equal(tmp$EVID, c(20109L, 20108L, 20101L, 0L, 20101L))
    expect_equal(tmp$II, c(24, 24, 0, 0, 0))
    expect_equal(tmp$AMT, c(5, -5, 5, NA, -5))

    d <- data.frame(time=c(0, 1),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(2, 0),
                    rate=c(5, 0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 0, 1, 20))
    expect_equal(tmp$EVID, c(10119L, 10108L, 10101L, 0L, 10101L))
    expect_equal(tmp$II, c(24, 24, 0, 0, 0))
    expect_equal(tmp$AMT, c(5, -5, 5, NA, -5))

    d <- data.frame(time=c(0, 1),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(2, 0),
                    dur=c(20, 0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 0, 1, 20))
    expect_equal(tmp$EVID, c(20119L, 20108L, 20101L, 0L, 20101L))
    expect_equal(tmp$II, c(24, 24, 0, 0, 0))
    expect_equal(tmp$AMT, c(5, -5, 5, NA, -5))

    d <- data.frame(time=c(0, 200),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(1, 0),
                    rate=c(5, 0),
                    addl=c(3, 0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 0, 20, 24, 24, 24, 44, 48, 48, 48, 68, 72, 72, 72, 92, 200))
    expect_equal(tmp$EVID, c(10109L, 10108L, 10101L, 10101L, 10109L, 10108L, 10101L, 10101L,
                             10109L, 10108L, 10101L, 10101L, 10109L, 10108L, 10101L, 10101L, 0L))
    expect_equal(tmp$AMT, c(5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, NA))
    expect_equal(tmp$II, c(24, 24, 0, 0, 24, 24, 0, 0, 24, 24, 0, 0, 24, 24, 0, 0, 0))


    d <- data.frame(time=c(0, 200),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(1, 0),
                    dur=c(20, 0),
                    addl=c(3, 0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 0, 20, 24, 24, 24, 44, 48, 48, 48, 68, 72, 72, 72, 92, 200))
    expect_equal(tmp$EVID, c(20109L, 20108L, 20101L, 20101L, 20109L, 20108L, 20101L, 20101L,
                             20109L, 20108L, 20101L, 20101L, 20109L, 20108L, 20101L, 20101L, 0L))
    expect_equal(tmp$AMT, c(5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, NA))
    expect_equal(tmp$II, c(24, 24, 0, 0, 24, 24, 0, 0, 24, 24, 0, 0, 24, 24, 0, 0, 0))

    d <- data.frame(time=c(0, 200),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(2, 0),
                    rate=c(5, 0),
                    addl=c(3, 0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 0, 20, 24, 24, 24, 44, 48, 48, 48, 68, 72, 72, 72, 92, 200))
    expect_equal(tmp$EVID, c(10119L, 10108L, 10101L, 10101L, 10119L, 10108L, 10101L, 10101L,
                             10119L, 10108L, 10101L, 10101L, 10119L, 10108L, 10101L, 10101L, 0L))
    expect_equal(tmp$AMT, c(5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, NA))
    expect_equal(tmp$II, c(24, 24, 0, 0, 24, 24, 0, 0, 24, 24, 0, 0, 24, 24, 0, 0, 0))


    d <- data.frame(time=c(0, 200),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(2, 0),
                    dur=c(20, 0),
                    addl=c(3, 0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 0, 20, 24, 24, 24, 44, 48, 48, 48, 68, 72, 72, 72, 92, 200))
    expect_equal(tmp$EVID, c(20119L, 20108L, 20101L, 20101L, 20119L, 20108L, 20101L, 20101L,
                             20119L, 20108L, 20101L, 20101L, 20119L, 20108L, 20101L, 20101L, 0L))
    expect_equal(tmp$AMT, c(5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, NA))
    expect_equal(tmp$II, c(24, 24, 0, 0, 24, 24, 0, 0, 24, 24, 0, 0, 24, 24, 0, 0, 0))

    # addl on bolus

    d <- data.frame(time=c(0, 200),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(1, 0),
                    addl=c(3,0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 24, 24, 48, 48, 72, 72, 200))
    expect_equal(tmp$EVID, c(109L, 101L, 109L, 101L, 109L, 101L, 109L, 101L, 0L))
    expect_equal(tmp$AMT, c(100, 100, 100, 100, 100, 100, 100, 100, NA))
    expect_equal(tmp$II, c(24, 0, 24, 0, 24, 0, 24, 0, 0))

    d <- data.frame(time=c(0, 200),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(2, 0),
                    addl=c(3,0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 24, 24, 48, 48, 72, 72, 200))
    expect_equal(tmp$EVID, c(119L, 101L, 119L, 101L, 119L, 101L, 119L, 101L, 0L))
    expect_equal(tmp$AMT, c(100, 100, 100, 100, 100, 100, 100, 100, NA))
    expect_equal(tmp$II, c(24, 0, 24, 0, 24, 0, 24, 0, 0))


    mod <- rxode2parse("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
alag(intestine) = lag
dur(intestine) = di
rate(intestine) = ri
d/dt(blood)     = a*intestine - b*blood
")

    d <- data.frame(time=c(0, 1),
                    amt=c(100, 0),
                    ii=c(24, 0),
                    evid=c(1,0),
                    ss=c(1, 0),
                    rate=c(-1, 0))

    tmp <- etTrans(d, mod, addlDropSs=FALSE)

    expect_equal(tmp$TIME, c(0, 0, 0, 1))
    expect_equal(tmp$EVID, c(90109L, 90101L, 70101L, 0L))
    expect_equal(tmp$AMT, c(100, 100, 100, NA))
    expect_equal(tmp$II, c(24, 0, 0, 0))
  })

  test_that("cmt translation gives appropriate results #938", {
    rxWithSeed(5447, {
      withr::with_seed(5447, {

        my_model <- function() {
          ini({
            # Typical Value of System Parameters
            TV_F1           = fixed(c(.Machine$double.eps, 0.744, .Machine$double.xmax))
            TV_ka           = c(.Machine$double.eps, 0.282, .Machine$double.xmax)
            TV_CL           =   log(c(.Machine$double.eps, 0.200, .Machine$double.xmax))
            TV_Vc           =   log(c(.Machine$double.eps, 3.61, .Machine$double.xmax))
            TV_Vp           =   log(c(.Machine$double.eps, 2.75, .Machine$double.xmax))
            TV_Q            =   log(c(.Machine$double.eps, 0.747, .Machine$double.xmax))
            TV_MW           = fixed(c(.Machine$double.eps, 140, .Machine$double.xmax))
            TV_BM_IC        = fixed(c(.Machine$double.eps, 1000, .Machine$double.xmax))
            TV_kdeg_BM      = fixed(c(.Machine$double.eps, 0.1, .Machine$double.xmax))
            TV_Emax         = fixed(c(.Machine$double.eps, 100.0, .Machine$double.xmax))
            TV_EC50         = fixed(c(.Machine$double.eps, 300, .Machine$double.xmax))

            # Between-subject variability:
            ETAka +     ETACL +     ETAVc +     ETAVp +     ETAQ +      ETABM_IC +  ETAEmax ~
              c(0.416,
                0,          0.09875,
                0,          0.0786,     0.116,
                0,          0.0619,     0.0377,     0.0789,
                0,          0,          0,          0,          0.699,
                0,          0,          0,          0,          0,          0.05,
                0,          0,          0,          0,          0,          0,          0.10)

            # Error model parameters
            prop_err_pk  =  c(.Machine$double.eps, 0.1, .Machine$double.xmax)
            add_err_pk   =  c(.Machine$double.eps, 0.1, .Machine$double.xmax)

            prop_err_pd  =  c(.Machine$double.eps, 0.1, .Machine$double.xmax)
            add_err_pd   =  c(.Machine$double.eps, 0.1, .Machine$double.xmax)
          })
          model({
            # System Parameters
            F1          = TV_F1
            ka          = TV_ka*exp(ETAka)
            CL          = exp(TV_CL + ETACL)
            Vc          = exp(TV_Vc + ETAVc)
            # AMTIFY:Cc
            Cc          = Ac/(Vc)
            Vp          = exp(TV_Vp + ETAVp)
            Q           = exp(TV_Q + ETAQ)
            MW          = TV_MW
            BM_IC       = TV_BM_IC*exp(ETABM_IC)
            kdeg_BM     = TV_kdeg_BM
            Emax        = TV_Emax*exp(ETAEmax)
            EC50        = TV_EC50

            # Static Secondary Parameters
            WTTV        = 70
            CL_IND      = CL*(1.0+sex*.1)*(1.0+subtype*.1)
            kel         = CL_IND/Vc*((wt/WTTV))^(-0.35)
            kcp         = Q/Vc*((wt/WTTV))^(-0.35)
            kpc         = Q/Vp*((wt/WTTV))^(-0.35)
            ksyn_BM     = kdeg_BM*BM_IC

            # Dynamic Secondary Parameters
            STIM        = 1.0+ Emax*Cc/(EC50+Cc)

            # non-zero initial conditions
            BM(0)           = BM_IC

            # Defining ODEs
            d/dt(At)    = (-ka*At)
            d/dt(Ac)    = ((ka*At*F1/Vc  -kel*Cc - kcp*Cc  + kpc*Cp*Vp/Vc))*Vc
            d/dt(Cp)    = (kcp*Cc*Vc/Vp - kpc*Cp)
            d/dt(BM)    = ksyn_BM*STIM-(kdeg_BM*BM)

            # Outputs and error models
            C_ng_ml     = Cc*MW
            C_ng_ml ~ add(add_err_pk) + prop(prop_err_pk)

            BM_obs      = BM
            BM_obs  ~ add(add_err_pd) + prop(prop_err_pd)

          })
        }


        sd_sam <- c(c(15, 30)/60/24,
                   c(1, 3, 6, 12)/24,
                   c(1, 3, 7, 14, 21))



        ev <- et()

        # Converting mg to ng
        conv <- 1e6

        nsub <- 10
        scens <- list(
          sd3iv = list(
            desc    = "SD 3 mg IV",
            cmt     = "Ac",
            dtimes  = 0,
            tsam    = sd_sam,
            damts   = 3),
          sd3sc = list(
            desc = "SD 3 mg SC",
            cmt  = "At",
            dtimes  = 0,
            tsam = sd_sam,
            damts= 30),
          sd10iv = list(
            desc    = "SD 30 mg IV",
            cmt     = "Ac",
            dtimes  = 0,
            tsam    = sd_sam,
            damts   = 30)
        )

        max_sub <- 0
        rxWithSeed(5447, {
          ev   <- et()
          iCov <- NULL
          for(scen in names(scens)){
            # Generating subject IDs
            scen_subs      <- c(1:nsub)+max_sub

            # Generating covariates
            scen_sex       <- sample(c(0,1),nsub,replace=TRUE)
            scen_subtype   <- sample(c(0,1),nsub,replace=TRUE)
            scen_wt        <- exp(rnorm(nsub, 0, sd=.1))*70
            for(tmp_sub_idx in 1:length(scen_subs)){
              tmp_ev <-
                et(id   = scen_subs[tmp_sub_idx],
                   amt  = scens[[scen]][["damts"]]*conv,
                   time = scens[[scen]][["dtimes"]],
                   cmt  = scens[[scen]][["cmt"]]
                   ) |>
                et(id   = scen_subs[tmp_sub_idx],
                   time = unique(c(scens[[scen]][["tsam"]])),
                   cmt   = "C_ng_ml"
                   ) |>
                et(id   = scen_subs[tmp_sub_idx],
                   time = unique(c(scens[[scen]][["tsam"]])),
                   cmt   = "BM_obs"
                   )
              ev <- etRbind(ev, tmp_ev)
              iCov <- rbind(iCov,
                           data.frame(
                             id      = scen_subs[tmp_sub_idx],
                             sex     = scen_sex[[tmp_sub_idx]],
                             wt      = scen_wt[[tmp_sub_idx]],
                             subtype = scen_subtype[[tmp_sub_idx]]))
            }
            max_sub <- max(scen_subs)
          }
          sim <- rxSolve(my_model, events=ev, iCov=iCov)

          expect_true(!any(is.na(sim$ipredSim)))
        }, rxseed = 5447)

      })
    })
  })
})


rxTest({
  test_that("warning on translation (#780)", {

    p <- test_path("test-etTrans-780.rds")
    skip_if_not(file.exists(p))
    dat <- readRDS(p)

    m <- rxode2parse("
      param(Kpm_pop, V_pop, k_pop, k12_pop, k21_pop, ka_pop,
            km_pop, a1_Cp, b1_Cp, a2_Cm, b2_Cm, omega_Kpm, omega_V,
            omega_k, omega_k12, omega_k21, omega_ka, omega_km)
      cmt(depot)
      cmt(central)
      cmt(cmt2)
      cmt(cmt3)
      Kpm = exp(Kpm_pop + omega_Kpm)
      V = exp(V_pop + omega_V)
      k = exp(k_pop + omega_k)
      k12 = exp(k12_pop + omega_k12)
      k21 = exp(k21_pop + omega_k21)
      ka = exp(ka_pop + omega_ka)
      km = exp(km_pop + omega_km)
      d/dt(depot) = -ka * depot
      d/dt(central) = -k12 * central + k21 * cmt2 + ka * depot -
        k * central - Kpm * central
      Cp = central/V
      d/dt(cmt2) = +k12 * central - k21 * cmt2
      d/dt(cmt3) = +Kpm * central - km * cmt3
      Cm = cmt3/V
      y1_Cp = Cp
      if (CMT == 5) {
        rx_yj_ ~ 2
        rx_lambda_ ~ 1
        rx_low_ ~ 0
        rx_hi_ ~ 1
        rx_pred_f_ ~ y1_Cp
        rx_pred_ ~ rx_pred_f_
        rx_r_ ~ ((a1_Cp) + (rx_pred_f_) * (b1_Cp))^2
        ipredSim = rxTBSi(rx_pred_, rx_lambda_, rx_yj_, rx_low_,
                          rx_hi_)
        sim = rxTBSi(rx_pred_ + sqrt(rx_r_) * rxerr.y1_Cp,
                     rx_lambda_, rx_yj_, rx_low_, rx_hi_)
      }
      y2_Cm = Cm
      if (CMT == 6) {
        rx_yj_ ~ 2
        rx_lambda_ ~ 1
        rx_low_ ~ 0
        rx_hi_ ~ 1
        rx_pred_f_ ~ y2_Cm
        rx_pred_ ~ rx_pred_f_
        rx_r_ ~ ((a2_Cm) + (rx_pred_f_) * (b2_Cm))^2
        ipredSim = rxTBSi(rx_pred_, rx_lambda_, rx_yj_, rx_low_,
                          rx_hi_)
        sim = rxTBSi(rx_pred_ + sqrt(rx_r_) * rxerr.y2_Cm,
                     rx_lambda_, rx_yj_, rx_low_, rx_hi_)
      }
      iwres = (DV - rx_pred_)/sqrt(rx_r_)
      ires = DV - rx_pred_
      cmt(y1_Cp)
      cmt(y2_Cm)
      dvid(5, 6)
    ")

    expect_warning(etTrans(dat, m), NA)

    test_that("etTrans(combineDvid=TRUE) does not read past a length-1 vector", {
      ## Before the fix combineDvid was read at index [1] of the (length-1) logical,
      ## warning "subscript out of bounds (index 1 >= vector size 1)" and using an
      ## out-of-bounds value as the dvid-combine flag.
      mod <- rxode2({
        d/dt(center) <- -kel * center
        cp <- center / v
      })
      d <- data.frame(ID = 1L, TIME = c(0, 1, 2), DV = c(0, 5, 3),
                      AMT = c(100, 0, 0), EVID = c(1, 0, 0))
      expect_warning(etTrans(d, mod, combineDvid = TRUE), NA)
    })

  })
})

rxTest({
  test_that("rxEt bind_rows coerces to canonical data.frame", {
    skip_if_not_installed("dplyr")

    ev1 <- et(dose = 100, addl = 0, ii = 0) |>
      et(0:2)
    evReset <- et(evid = 3)
    ev2 <- et(dose = 200, addl = 0, ii = 0) |>
      et(0:2)

    got <- dplyr::bind_rows(ev1, evReset, ev2)

    expect_s3_class(got, "data.frame")
    expect_false(inherits(got, "rxEt"))
    expect_equal(names(got), c("id", "low", "time", "high", "cmt", "amt",
                               "rate", "ii", "addl", "evid", "ss", "dur"))
    expect_equal(sum(got$evid == 3L, na.rm = TRUE), 1L)
  })

  test_that("rxEt vec_ptype2 and vec_cast round-trip through data.frame", {
    ev <- et(dose = 100, addl = 0, ii = 0, amountUnits = "mg", timeUnits = "hours") |>
      et(0:2)
    df <- as.data.frame(ev, all = TRUE)

    ptype <- vctrs::vec_ptype2(ev, ev)
    expect_s3_class(ptype, "data.frame")
    expect_false(inherits(ptype, "rxEt"))
    expect_equal(names(ptype), names(df))

    restored <- vctrs::vec_cast(df, et(amountUnits = "mg", timeUnits = "hours"))
    expect_true(is.rxEt(restored))
    expect_equal(as.data.frame(restored, all = TRUE), df)
  })

  test_that("rxEt double dispatch works with data.table and tibble", {
    ev <- et(dose = 100, addl = 0, ii = 0) |>
      et(0:2)
    df <- as.data.frame(ev, all = TRUE)

    dt <- data.table::as.data.table(df)
    expect_true(inherits(vctrs::vec_ptype2(ev, dt), "data.table"))
    expect_true(inherits(vctrs::vec_cast(ev, dt), "data.table"))
    expect_true(is.rxEt(vctrs::vec_cast(df, ev)))

    skip_if_not_installed("tibble")
    tb <- tibble::as_tibble(df)
    expect_s3_class(vctrs::vec_ptype2(ev, tb), "tbl_df")
    expect_s3_class(vctrs::vec_cast(ev, tb), "tbl_df")
  })
})

test_that("rxEt objects with compatible units can be combined", {
  skip_if_not_installed("units")
  library(units)

  # Create two event tables with different but compatible dosing units
  et1 <- et(amount.units="mg", time.units="hr") %>% et(amt=1, time=1)
  et2 <- et(amount.units="g", time.units="hr") %>% et(amt=1, time=2)

  # Combine them using vctrs::vec_c (which is used by dplyr::bind_rows)
  # Based on existing design, this returns a data.frame
  res <- vctrs::vec_c(et1, et2)

  expect_s3_class(res, "data.frame")
  
  # The second dose (1g) should be converted to 1000mg in the units-aware column
  expect_equal(as.numeric(res$amt), c(1, 1000))
  
  # Now check time units
  et3 <- et(amount.units="mg", time.units="hr") %>% et(amt=1, time=1)
  et4 <- et(amount.units="mg", time.units="min") %>% et(amt=1, time=60)
  
  res2 <- vctrs::vec_c(et3, et4)
  # The second time (60 min) should be converted to 1 hr
  expect_equal(as.numeric(res2$time), c(1, 1))
})

test_that("rxEt objects with incompatible units throw error", {
  skip_if_not_installed("units")
  et1 <- et(amount.units="mg", time.units="hr") %>% et(amt=1, time=1)
  et2 <- et(amount.units="m", time.units="hr") %>% et(amt=1, time=1)
  
  expect_error(vctrs::vec_c(et1, et2), "incompatible dose units")
})

## Event ("jump") sensitivities: build-time index map relating dosing/event
## compartments to sensitivity compartments, plus the symbolic total
## derivatives of the dosing parameters (alag/F/rate/dur) used by the runtime
## jump injection.

#' Resolve the event-sensitivity calculation mode
#'
#' `"jump"` = analytic jump sensitivities, `"fd"` = finite differences (the
#' backward-compatible opt-out), `"both"` = compute both for cross-checking.
#' The default comes from `getOption("rxode2.eventSens")`.
#'
#' @param mode One of `"jump"`, `"fd"`, `"both"`, or `NULL` to use the option.
#' @return The resolved mode string.
#' @noRd
.rxEventSensMode <- function(mode = NULL) {
  if (is.null(mode)) mode <- getOption("rxode2.eventSens", "jump")
  mode <- as.character(mode)[1L]
  if (!mode %in% c("jump", "fd", "both", "fdAll")) {
    stop("'eventSens' must be one of \"jump\", \"fd\", \"both\", or \"fdAll\"", call. = FALSE)
  }
  mode
}

#' Split a first-order sensitivity state name into (state, param)
#'
#' Sensitivity compartments are named `rx__sens_<state>_BY_<param>__`.  Both the
#' state and parameter names may themselves contain underscores, so the split is
#' anchored on the known physical state names (the longest matching prefix) and
#' the `_BY_` separator rather than naive string splitting.
#'
#' @param sens Character vector of sensitivity compartment names.
#' @param states Character vector of physical (non-sensitivity) state names.
#' @return data.frame with columns `sens`, `state`, `param`.  Second- (and
#'   higher-) order sensitivities (two or more `_BY_`) yield `NA` and are dropped
#'   by the caller; they are handled in a later phase.
#' @noRd
.rxEventSensSplit <- function(sens, states) {
  .pre <- "rx__sens_"
  .core <- sub("__$", "", sub(paste0("^", .pre), "", sens))
  ## order states longest-first so e.g. a state named "central_x" wins over
  ## "central" when both exist.
  .states <- states[order(nchar(states), decreasing = TRUE)]
  .state <- rep(NA_character_, length(.core))
  .param <- rep(NA_character_, length(.core))
  for (.i in seq_along(.core)) {
    for (.s in .states) {
      .head <- paste0(.s, "_BY_")
      if (startsWith(.core[.i], .head)) {
        .rest <- substring(.core[.i], nchar(.head) + 1L)
        ## first-order only: the remainder must not contain another `_BY_`
        if (!grepl("_BY_", .rest, fixed = TRUE)) {
          .state[.i] <- .s
          .param[.i] <- .rest
        }
        break
      }
    }
  }
  data.frame(sens = sens, state = .state, param = .param,
             stringsAsFactors = FALSE)
}

#' Split a second-order sensitivity state name into (state, p, q)
#'
#' Second-order sensitivity compartments are named
#' `rx__sens_<state>_BY_<p>_BY_<q>__`.  Anchored on the known physical state
#' names (longest match) and the two `_BY_` separators; parameter names never
#' contain `_BY_`, so the remainder splits cleanly into `(p, q)`.
#'
#' @param sens Character vector of sensitivity compartment names.
#' @param states Character vector of physical (non-sensitivity) state names.
#' @return data.frame(sens, state, p, q); first-order names (one `_BY_`) and
#'   non-matches yield `NA` and are dropped by the caller.
#' @noRd
.rxEventSensSplit2 <- function(sens, states) {
  .pre <- "rx__sens_"
  .core <- sub("__$", "", sub(paste0("^", .pre), "", sens))
  .states <- states[order(nchar(states), decreasing = TRUE)]
  .state <- rep(NA_character_, length(.core))
  .p <- rep(NA_character_, length(.core))
  .q <- rep(NA_character_, length(.core))
  for (.i in seq_along(.core)) {
    for (.s in .states) {
      .head <- paste0(.s, "_BY_")
      if (startsWith(.core[.i], .head)) {
        .rest <- substring(.core[.i], nchar(.head) + 1L)
        .parts <- strsplit(.rest, "_BY_", fixed = TRUE)[[1]]
        ## second-order only: exactly two parts (p, q)
        if (length(.parts) == 2L) {
          .state[.i] <- .s
          .p[.i] <- .parts[1]
          .q[.i] <- .parts[2]
        }
        break
      }
    }
  }
  data.frame(sens = sens, state = .state, p = .p, q = .q,
             stringsAsFactors = FALSE)
}

#' Split a third-order sensitivity state name into (state, p, q, r)
#'
#' Third-order sensitivity compartments are named
#' `rx__sens_<state>_BY_<p>_BY_<q>_BY_<r>__` (`rxExpandSens3_`).
#' Anchored the same way as `.rxEventSensSplit2()`.
#'
#' @param sens Character vector of sensitivity compartment names.
#' @param states Character vector of physical (non-sensitivity) state names.
#' @return data.frame(sens, state, p, q, r); non-third-order names (not
#'   exactly three `_BY_` segments) yield `NA` and are dropped by the caller.
#' @noRd
.rxEventSensSplit3 <- function(sens, states) {
  .pre <- "rx__sens_"
  .core <- sub("__$", "", sub(paste0("^", .pre), "", sens))
  .states <- states[order(nchar(states), decreasing = TRUE)]
  .state <- rep(NA_character_, length(.core))
  .p <- rep(NA_character_, length(.core))
  .q <- rep(NA_character_, length(.core))
  .r <- rep(NA_character_, length(.core))
  for (.i in seq_along(.core)) {
    for (.s in .states) {
      .head <- paste0(.s, "_BY_")
      if (startsWith(.core[.i], .head)) {
        .rest <- substring(.core[.i], nchar(.head) + 1L)
        .parts <- strsplit(.rest, "_BY_", fixed = TRUE)[[1]]
        ## third-order only: exactly three parts (p, q, r)
        if (length(.parts) == 3L) {
          .state[.i] <- .s
          .p[.i] <- .parts[1]
          .q[.i] <- .parts[2]
          .r[.i] <- .parts[3]
        }
        break
      }
    }
  }
  data.frame(sens = sens, state = .state, p = .p, q = .q, r = .r,
             stringsAsFactors = FALSE)
}

#' Build the event-sensitivity index map for a model
#'
#' Relates each dosing/event compartment to the first-order sensitivity
#' compartments that its events must jump.  Pure model-vars bookkeeping.
#'
#' @param obj Anything `rxModelVars()` accepts (rxUi, rxode2, modelVars).
#' @return A list with:
#'   * `states`     -- physical (non-sensitivity) state names, in compartment order.
#'   * `nState`     -- number of physical states.
#'   * `stateCmt`   -- named integer: physical state -> 1-based compartment index.
#'   * `sensParams` -- parameters the first-order sensitivities are taken wrt.
#'   * `map`        -- data.frame(state, param, stateCmt, sensCmt): for each
#'                     (physical state, sensitivity parameter) the 1-based
#'                     compartment of `rx__sens_<state>_BY_<param>__`.
#'   * `lagCmt`/`fCmt`/`rateCmt`/`durCmt` -- 1-based compartments carrying a
#'                     modeled alag/F/rate/dur (the event compartments).
#'   Returns `NULL` when the model has no first-order sensitivity compartments.
#' @noRd
.rxEventSensMap <- function(obj) {
  .mv <- rxModelVars(obj)
  .sens <- .mv$sens
  if (is.null(.sens) || length(.sens) == 0L) return(NULL)
  .states <- .mv$normal.state
  .ord <- .mv$stateOrd
  .split <- .rxEventSensSplit(.sens, .states)
  .split <- .split[!is.na(.split$state), , drop = FALSE]
  if (nrow(.split) == 0L) return(NULL)
  .split$sensCmt <- unname(.ord[.split$sens])
  .split$stateCmt <- unname(.ord[.split$state])
  .sensParams <- unique(.split$param)
  ## Only states with sensitivity compartments count: the runtime jump formula
  ## needs a contiguous block of exactly nState states starting at index 0.
  .statesWithSens <- unique(.split$state)
  ## Preserve compartment ordering from .map$stateCmt (ascending).
  .statesWithSens <- .statesWithSens[order(unname(.ord[.statesWithSens]))]
  .map <- .split[, c("state", "param", "stateCmt", "sensCmt"), drop = FALSE]
  .map <- .map[order(.map$param, .map$stateCmt), , drop = FALSE]
  rownames(.map) <- NULL
  ## Second-order sensitivity compartments (Hessian path), if present.
  .split2 <- .rxEventSensSplit2(.sens, .states)
  .split2 <- .split2[!is.na(.split2$state), , drop = FALSE]
  .map2 <- NULL
  if (nrow(.split2) > 0L) {
    .split2$sensCmt <- unname(.ord[.split2$sens])
    .split2$stateCmt <- unname(.ord[.split2$state])
    .map2 <- .split2[, c("state", "p", "q", "stateCmt", "sensCmt"), drop = FALSE]
    ## Deliberately NOT re-sorted: .rxEventSensCLines() recovers each
    ## parameter's ordinal position in calcSens/calcSens2 from
    ## unique(.map2$p)/unique(.map2$q) first-occurrence order, which must stay
    ## the compiled rxExpandSens2_ layout order; re-sorting writes 2nd-order
    ## jump values into the wrong compartment.
    rownames(.map2) <- NULL
  }
  ## Third-order compartments: same deliberately-unsorted-row-order
  ## requirement as .map2 (.pIdx/.qIdx/.rIdx use first-occurrence order).
  .split3 <- .rxEventSensSplit3(.sens, .states)
  .split3 <- .split3[!is.na(.split3$state), , drop = FALSE]
  .map3 <- NULL
  if (nrow(.split3) > 0L) {
    .split3$sensCmt <- unname(.ord[.split3$sens])
    .split3$stateCmt <- unname(.ord[.split3$state])
    .map3 <- .split3[, c("state", "p", "q", "r", "stateCmt", "sensCmt"), drop = FALSE]
    rownames(.map3) <- NULL
  }
  ## event compartments: those carrying a modeled alag/F/rate/dur
  ## (mv$alag plus stateProp bit flags; see .rxEventSensProp)
  .prop <- .rxEventSensProp(.mv)
  list(
    states = .statesWithSens,
    nState = length(.statesWithSens),
    stateCmt = stats::setNames(unname(.ord[.statesWithSens]), .statesWithSens),
    sensParams = .sensParams,
    map = .map,
    map2 = .map2,
    map3 = .map3,
    lagCmt = .prop$lagCmt,
    fCmt = .prop$fCmt,
    rateCmt = .prop$rateCmt,
    durCmt = .prop$durCmt
  )
}

#' Detect an ODE compartment name colliding with a linCmt() reserved name
#'
#' An explicit `d/dt()` on a linCmt()-reserved compartment name
#' (depot/central/peripheralN) conflates the ODE and linCmt compartments -- the
#' ODE state loses its sensitivity expansion (its `rx__sens_<state>_BY_*`
#' compartment is never generated), so both the continuous sensitivity ODE and
#' the analytic jump come out wrong.
#'
#' @param obj Anything `rxModelVars()`/`rxNorm()` accepts.
#' @return character vector of colliding compartment names (empty when none).
#' @noRd
.rxLinCmtNameCollision <- function(obj) {
  .mv <- rxModelVars(obj)
  if (.rxLinNcmt(.mv)["numLin"] <= 0L) return(character(0))
  .reservedPhys <- grep("^rx__sens_", .rxLinCmt(.mv), value = TRUE, invert = TRUE)
  if (length(.reservedPhys) == 0L) return(character(0))
  .norm <- rxNorm(obj)
  .reservedPhys[vapply(.reservedPhys, function(.nm)
    grepl(paste0("d/dt(", .nm, ")"), .norm, fixed = TRUE), logical(1))]
}

#' Restrict jump sensitivities to ODE states for mixed ODE+linCmt models
#'
#' For pure linCmt models (no ODE states), event sensitivities are handled by the
#' finite-difference linCmt path, so jump metadata is disabled (`NULL`).  For
#' mixed models, keep only ODE-scoped states/parameters in the jump map.
#'
#' @param obj Model object accepted by `rxModelVars()`.
#' @param map `.rxEventSensMap(obj)` result.
#' @return Filtered map list, or `NULL` when jump should be disabled.
#' @noRd
.rxEventSensFilterMap <- function(obj, map) {
  .mv <- rxModelVars(obj)
  .lin <- .rxLinNcmt(.mv)
  if (.lin["numLin"] <= 0L) return(map)
  ## Defensive: a linCmt reserved-name collision (see .rxLinCmtNameCollision)
  ## yields silently-incorrect sensitivities; disable jump.  (linCmt models now
  ## downgrade to FD upstream via .rxEventSensEffectiveMode, so this is a guard
  ## for any path that still reaches here.)
  if (length(.rxLinCmtNameCollision(obj)) > 0L) return(NULL)
  .odeStates <- setdiff(.mv$normal.state, .rxLinCmt(.mv))
  if (length(.odeStates) == 0L) return(map)
  .stateCmt <- unname(map$stateCmt[.odeStates])
  ## The jump runtime assumes ODE states are the leading contiguous block
  ## [1..nState]. If not, keep behavior safe by disabling jump for this model.
  if (!identical(.stateCmt, seq_along(.stateCmt))) return(NULL)
  .map1 <- map$map[map$map$state %in% .odeStates, , drop = FALSE]
  if (nrow(.map1) == 0L) return(NULL)
  .sensParams <- unique(.map1$param)
  .map2 <- map$map2
  if (!is.null(.map2) && nrow(.map2) > 0L) {
    .map2 <- .map2[
      .map2$state %in% .odeStates & .map2$p %in% .sensParams,
      , drop = FALSE
    ]
    if (nrow(.map2) == 0L) .map2 <- NULL
  }
  .map3 <- map$map3
  if (!is.null(.map3) && nrow(.map3) > 0L) {
    .map3 <- .map3[
      .map3$state %in% .odeStates & .map3$p %in% .sensParams,
      , drop = FALSE
    ]
    if (nrow(.map3) == 0L) .map3 <- NULL
  }
  list(
    states = .odeStates,
    nState = length(.odeStates),
    stateCmt = stats::setNames(seq_along(.odeStates), .odeStates),
    sensParams = .sensParams,
    map = .map1,
    map2 = .map2,
    map3 = .map3,
    lagCmt = map$lagCmt[map$lagCmt %in% .stateCmt],
    fCmt = map$fCmt[map$fCmt %in% .stateCmt],
    rateCmt = map$rateCmt[map$rateCmt %in% .stateCmt],
    durCmt = map$durCmt[map$durCmt %in% .stateCmt]
  )
}

#' Resolve effective event-sensitivity mode for linCmt-containing models
#'
#' `fdAll` is the explicit full finite-difference fallback. Models that include
#' linCmt() always resolve to `fd` regardless of the requested mode (see the
#' comment in the body); the requested mode is honored otherwise.
#'
#' @param requested Requested mode from `.rxEventSensMode()`.
#' @param mv Parsed model vars.
#' @return Effective mode string (`jump`, `fd`, or `both`).
#' @noRd
.rxEventSensEffectiveMode <- function(requested, mv) {
  if (identical(requested, "fdAll")) return("fd")
  ## linCmt() models downgrade to finite differences for event-timing
  ## sensitivities.  The analytic moving-boundary (jump) sensitivity for a
  ## modeled alag()/f() on a linCmt compartment is not implemented (see
  ## nlmixr2/rxode2#1119); rather than emit an incomplete/incorrect analytic
  ## jump, all linCmt models use FD for event sensitivities (the FOCEi/nlmixr2est
  ## finite-difference event path).  Structural-parameter linCmt sensitivities
  ## are unaffected (they are continuous and handled by linCmtB directly).
  if (.rxLinNcmt(mv)["numLin"] > 0L) return("fd")
  requested
}

#' Decode which compartments carry modeled alag/F/rate/dur
#'
#' `mv$stateProp` is a named integer of per-compartment property bit flags.
#' `mv$alag` directly lists the lag compartments; this helper returns all four
#' event-property compartment sets, falling back to the bit flags when needed.
#'
#' @param mv A model-vars list.
#' @return list(lagCmt, fCmt, rateCmt, durCmt) of 1-based compartment integers.
#' @noRd
.rxEventSensProp <- function(mv) {
  .ord <- mv$stateOrd
  .prop <- mv$stateProp
  ## stateProp bit flags (src/tran.h): propF=2, propAlag=4, propRate=8, propDur=16
  .bit <- function(flag) {
    if (is.null(.prop)) return(integer(0))
    .nm <- names(.prop)[bitwAnd(as.integer(.prop), flag) != 0L]
    sort(unname(.ord[.nm]))
  }
  .lag <- .bit(4L)
  if (length(.lag) == 0L && !is.null(mv$alag)) .lag <- sort(as.integer(mv$alag))
  list(lagCmt = .lag, fCmt = .bit(2L), rateCmt = .bit(8L), durCmt = .bit(16L))
}

#' Identify (linCmt compartment, event-timing parameter) pairs needing sensitivities
#'
#' For a linCmt() model, a modeled alag()/f() on a linCmt compartment driven by a
#' `calcSens` parameter has an event-timing (moving-boundary) sensitivity that the
#' structural linCmt Jacobian (wrt p1/v1/ka/...) does not capture.  This returns
#' the (linCmt state, driving parameter) pairs so the linCmt solve can carry the
#' extra sensitivity columns (Part B of nlmixr2/rxode2#1119).
#'
#' @param obj Built model (rxUi, rxode2, modelVars).
#' @param calcSens Character vector of first-order sensitivity parameters.
#' @return data.frame(state, param, kind, cmt) (1-based `cmt`), or `NULL` when
#'   the model has no linCmt event-timing sensitivities.
#' @noRd
.rxLinCmtEventSensPairs <- function(obj, calcSens) {
  if (is.null(calcSens) || length(calcSens) == 0L) return(NULL)
  .mv <- rxModelVars(obj)
  if (.rxLinNcmt(.mv)["numLin"] <= 0L) return(NULL)
  ## linCmt physical compartments (reserved names that are not sens compartments)
  .linPhys <- grep("^rx__sens_", .rxLinCmt(.mv), value = TRUE, invert = TRUE)
  if (length(.linPhys) == 0L) return(NULL)
  .norm <- strsplit(rxNorm(obj), "\n", fixed = TRUE)[[1]]
  .rows <- list()
  for (.kind in c("alag", "f")) {
    .re <- paste0("^", .kind, "\\(([^)]+)\\)=(.*);$")
    for (.line in grep(.re, .norm, value = TRUE)) {
      .cmt <- sub(.re, "\\1", .line)
      if (!(.cmt %in% .linPhys)) next
      .rhs <- sub(.re, "\\2", .line)
      .vars <- tryCatch(all.vars(str2lang(.rhs)), error = function(e) character(0))
      for (.p in intersect(.vars, calcSens)) {
        .rows[[length(.rows) + 1L]] <-
          data.frame(state = .cmt, param = .p, kind = .kind,
                     stringsAsFactors = FALSE)
      }
    }
  }
  if (length(.rows) == 0L) return(NULL)
  .df <- do.call(rbind, .rows)
  .df$cmt <- unname(.mv$stateOrd[.df$state])
  .df[!duplicated(.df[c("state", "param")]), , drop = FALSE]
}

#' Free symbols of a dosing expression in symengine (SE-mangled) names
#'
#' `map$sensParams` are SE-mangled names (e.g. `ETA_3_`); `all.vars()` of the
#' R-syntax text collapses an indexed `ETA[3]` to `"ETA"` and never matches,
#' silently dropping the derivative -- use symengine's `free_symbols` directly.
#'
#' @param sym symengine expression (or numeric constant).
#' @return character vector of free-symbol names (empty for constants).
#' @noRd
.rxEventSensFreeSyms <- function(sym) {
  if (is.numeric(sym)) return(character(0))
  tryCatch(
    vapply(symengine::free_symbols(sym), as.character, character(1)),
    error = function(e) character(0))
}

#' Total derivative of one dosing-parameter expression wrt a parameter
#'
#' `d(g)/dp = partial g/partial p + sum_l (partial g/partial x_l) * S^p_l`,
#' where `g` is a modeled alag/F/rate/dur expression (`rx_<kind>_<cmt>_` in the
#' symengine env) and `S^p_l = rx__sens_<x_l>_BY_<p>__`.
#'
#' @param model symengine model environment (e.g. from `.rxLoadPrune()`).
#' @param sym symengine expression for the dosing parameter `g`.
#' @param param Parameter name `p` to differentiate wrt.
#' @param states Physical state names `x_l`.
#' @return `rxFromSE` expression text for `d(g)/dp` (the string `"0"` when zero).
#' @noRd
.rxEventSensDExpr <- function(model, sym, param, states) {
  ## guard on free symbols: skips zero terms and avoids D() on constants (errors)
  .vars <- .rxEventSensFreeSyms(sym)
  ## assign symengine results before rxFromSE (NSE capture; see R/dde.R)
  .tot <- NULL
  if (param %in% .vars) {
    .tot <- symengine::D(sym, symengine::S(param))
  }
  for (.l in states) {
    if (!(.l %in% .vars)) next
    .dl <- symengine::D(sym, symengine::S(.l))
    .dlTxt <- rxFromSE(.dl)
    if (.dlTxt != "0" && .dlTxt != "0.0") {
      .S <- symengine::S(paste0("rx__sens_", .l, "_BY_", param, "__"))
      .term <- .dl * .S
      .tot <- if (is.null(.tot)) .term else .tot + .term
    }
  }
  if (is.null(.tot)) return("0")
  rxFromSE(.tot)
}

#' First-order total derivative of a dosing expression as a symengine object
#'
#' Same quantity as `.rxEventSensDExpr()` but returns the symengine expression
#' (or `NULL` when zero) instead of `rxFromSE` text, so it can be differentiated
#' again for the second-order total derivative.
#'
#' @return symengine expression for `d(g)/dp`, or `NULL` if identically zero.
#' @noRd
.rxEventSensDSym <- function(sym, param, states) {
  .vars <- .rxEventSensFreeSyms(sym)
  .tot <- NULL
  if (param %in% .vars) {
    .tot <- symengine::D(sym, symengine::S(param))
  }
  for (.l in states) {
    if (!(.l %in% .vars)) next
    .dl <- symengine::D(sym, symengine::S(.l))
    .dlTxt <- rxFromSE(.dl)
    if (.dlTxt != "0" && .dlTxt != "0.0") {
      .S <- symengine::S(paste0("rx__sens_", .l, "_BY_", param, "__"))
      .term <- .dl * .S
      .tot <- if (is.null(.tot)) .term else .tot + .term
    }
  }
  .tot
}

#' Second-order total derivative of a dosing-parameter expression
#'
#' Computes `d^2(g)/dp/dq` as a total derivative: the direct partial plus the
#' state-coupling (`* S^q_l`) and first-order-sensitivity-coupling
#' (`* S^{pq}_l`) terms.
#'
#' @param sym symengine expression for the dosing parameter `g`.
#' @param p,q Parameter names to differentiate wrt.
#' @param states Physical state names `x_l`.
#' @return `rxFromSE` text for `d^2(g)/dp/dq` (the string `"0"` when zero).
#' @noRd
.rxEventSensD2Expr <- function(sym, p, q, states) {
  .tot <- .rxEventSensD2Sym(sym, p, q, states)
  if (is.null(.tot)) return("0")
  rxFromSE(.tot)
}

#' Second-order total derivative as a symengine object
#'
#' Same quantity as `.rxEventSensD2Expr()` but returns the symengine
#' expression (or `NULL` when zero) instead of `rxFromSE` text, so it can be
#' differentiated again for the third-order total derivative.
#'
#' @inheritParams .rxEventSensD2Expr
#' @return symengine expression for `d^2(g)/dp/dq`, or `NULL` if identically zero.
#' @noRd
.rxEventSensD2Sym <- function(sym, p, q, states) {
  .dgp <- .rxEventSensDSym(sym, p, states)
  if (is.null(.dgp)) return(NULL)
  ## SE-mangled free symbols (see .rxEventSensFreeSyms)
  .vars <- .rxEventSensFreeSyms(.dgp)
  .tot <- NULL
  ## direct partial wrt q
  if (q %in% .vars) {
    .tot <- symengine::D(.dgp, symengine::S(q))
  }
  ## state-coupling: d/dx_l * S^q_l
  for (.l in states) {
    if (!(.l %in% .vars)) next
    .dxl <- symengine::D(.dgp, symengine::S(.l))
    if (rxFromSE(.dxl) %in% c("0", "0.0")) next
    .Sq <- symengine::S(paste0("rx__sens_", .l, "_BY_", q, "__"))
    .term <- .dxl * .Sq
    .tot <- if (is.null(.tot)) .term else .tot + .term
  }
  ## first-order-sensitivity coupling: d/d(S^p_l) * S^{pq}_l
  for (.l in states) {
    .Spl <- paste0("rx__sens_", .l, "_BY_", p, "__")
    if (!(.Spl %in% .vars)) next
    .dSpl <- symengine::D(.dgp, symengine::S(.Spl))
    if (rxFromSE(.dSpl) %in% c("0", "0.0")) next
    .Spq <- symengine::S(paste0("rx__sens_", .l, "_BY_", p, "_BY_", q, "__"))
    .term <- .dSpl * .Spq
    .tot <- if (is.null(.tot)) .term else .tot + .term
  }
  .tot
}

#' Third-order total derivative of a dosing-parameter expression
#'
#' One level deeper than `.rxEventSensD2Sym()`: the direct partial plus the
#' state-coupling and the p-, q-, and pq-chain sensitivity couplings.
#' `S^{pr}_l`/`S^{qr}_l` are valid second-order compartments because
#' `calcSens3 subset calcSens2 subset calcSens`.
#'
#' @param sym symengine expression for the dosing parameter `g`.
#' @param p,q,r Parameter names to differentiate wrt (p in calcSens, q in
#'   calcSens2, r in calcSens3).
#' @param states Physical state names `x_l`.
#' @return `rxFromSE` text for `d^3(g)/dp/dq/dr` (the string `"0"` when zero).
#' @noRd
.rxEventSensD3Expr <- function(sym, p, q, r, states) {
  .dgpq <- .rxEventSensD2Sym(sym, p, q, states)
  if (is.null(.dgpq)) return("0")
  .vars <- .rxEventSensFreeSyms(.dgpq)
  .tot <- NULL
  ## direct partial wrt r
  if (r %in% .vars) {
    .tot <- symengine::D(.dgpq, symengine::S(r))
  }
  ## state-coupling: d/dx_l * S^r_l
  for (.l in states) {
    if (!(.l %in% .vars)) next
    .dxl <- symengine::D(.dgpq, symengine::S(.l))
    if (rxFromSE(.dxl) %in% c("0", "0.0")) next
    .Sr <- symengine::S(paste0("rx__sens_", .l, "_BY_", r, "__"))
    .term <- .dxl * .Sr
    .tot <- if (is.null(.tot)) .term else .tot + .term
  }
  ## p-chain coupling: d/d(S^p_l) * S^{pr}_l
  for (.l in states) {
    .Spl <- paste0("rx__sens_", .l, "_BY_", p, "__")
    if (!(.Spl %in% .vars)) next
    .dSpl <- symengine::D(.dgpq, symengine::S(.Spl))
    if (rxFromSE(.dSpl) %in% c("0", "0.0")) next
    .Spr <- symengine::S(paste0("rx__sens_", .l, "_BY_", p, "_BY_", r, "__"))
    .term <- .dSpl * .Spr
    .tot <- if (is.null(.tot)) .term else .tot + .term
  }
  ## q-chain coupling: d/d(S^q_l) * S^{qr}_l
  for (.l in states) {
    .Sql <- paste0("rx__sens_", .l, "_BY_", q, "__")
    if (!(.Sql %in% .vars)) next
    .dSql <- symengine::D(.dgpq, symengine::S(.Sql))
    if (rxFromSE(.dSql) %in% c("0", "0.0")) next
    .Sqr <- symengine::S(paste0("rx__sens_", .l, "_BY_", q, "_BY_", r, "__"))
    .term <- .dSql * .Sqr
    .tot <- if (is.null(.tot)) .term else .tot + .term
  }
  ## pq-chain coupling: d/d(S^{pq}_l) * S^{pqr}_l
  for (.l in states) {
    .Spql <- paste0("rx__sens_", .l, "_BY_", p, "_BY_", q, "__")
    if (!(.Spql %in% .vars)) next
    .dSpql <- symengine::D(.dgpq, symengine::S(.Spql))
    if (rxFromSE(.dSpql) %in% c("0", "0.0")) next
    .Spqr <- symengine::S(paste0("rx__sens_", .l, "_BY_", p, "_BY_", q, "_BY_", r, "__"))
    .term <- .dSpql * .Spqr
    .tot <- if (is.null(.tot)) .term else .tot + .term
  }
  if (is.null(.tot)) return("0")
  rxFromSE(.tot)
}

#' Symbolic total-derivative tables for modeled alag / F
#'
#' For each event compartment carrying a modeled `alag` (resp. `F`) and each
#' first-order sensitivity parameter, emit the total derivative
#' `d(alag_c)/dp` (resp. `d(F_c)/dp`).  These feed the runtime jump injection:
#' `d(alag)/dp` scales the `dxk/dtau` rows and `d(F)/dp` the `dxk/ddelta` rows.
#'
#' @param obj Anything `rxModelVars()`/`.rxLoadPrune()` accepts.
#' @param map Optional precomputed `.rxEventSensMap(obj)`.
#' @return list with `lag` and `f` data.frames `(cmt, cmtName, param, expr)`
#'   holding only the non-zero derivatives; `NULL` when the model has no
#'   first-order sensitivities.
#' @noRd
.rxEventSensDerivs <- function(obj, map = NULL) {
  if (is.null(map)) map <- .rxEventSensMap(obj)
  if (is.null(map)) return(NULL)
  .model <- .rxLoadPrune(obj)
  .states <- map$states
  .params <- map$sensParams
  .cmtName <- function(cmt) map$states[match(cmt, map$stateCmt)]
  .build <- function(cmts, kind) {
    .rows <- list()
    for (.c in cmts) {
      .nm <- .cmtName(.c)
      .symName <- paste0("rx_", kind, "_", .nm, "_")
      if (!exists(.symName, envir = .model)) next
      .sym <- get(.symName, envir = .model)
      for (.p in .params) {
        .e <- .rxEventSensDExpr(.model, .sym, .p, .states)
        if (.e != "0" && .e != "0.0") {
          .rows[[length(.rows) + 1L]] <-
            data.frame(cmt = .c, cmtName = .nm, param = .p, expr = .e,
                       stringsAsFactors = FALSE)
        }
      }
    }
    if (length(.rows) == 0L) {
      return(data.frame(cmt = integer(0), cmtName = character(0),
                        param = character(0), expr = character(0),
                        stringsAsFactors = FALSE))
    }
    do.call(rbind, .rows)
  }
  ## 2nd-order tables (Hessian jump path), indexed (cmt, p, q); built only
  ## when map2 is present
  .build2 <- function(cmts, kind) {
    if (is.null(map$map2)) {
      return(data.frame(cmt = integer(0), cmtName = character(0),
                        p = character(0), q = character(0), expr = character(0),
                        stringsAsFactors = FALSE))
    }
    .p2 <- unique(map$map2$p)
    .q2 <- unique(map$map2$q)
    .rows <- list()
    for (.c in cmts) {
      .nm <- .cmtName(.c)
      .symName <- paste0("rx_", kind, "_", .nm, "_")
      if (!exists(.symName, envir = .model)) next
      .sym <- get(.symName, envir = .model)
      for (.p in .p2) {
        for (.q in .q2) {
          .e <- .rxEventSensD2Expr(.sym, .p, .q, .states)
          if (.e != "0" && .e != "0.0") {
            .rows[[length(.rows) + 1L]] <-
              data.frame(cmt = .c, cmtName = .nm, p = .p, q = .q, expr = .e,
                         stringsAsFactors = FALSE)
          }
        }
      }
    }
    if (length(.rows) == 0L) {
      return(data.frame(cmt = integer(0), cmtName = character(0),
                        p = character(0), q = character(0), expr = character(0),
                        stringsAsFactors = FALSE))
    }
    do.call(rbind, .rows)
  }
  ## 3rd-order table: additive-bolus `F` row only, indexed (cmt, p, q, r)
  .build3 <- function(cmts, kind) {
    if (is.null(map$map3)) {
      return(data.frame(cmt = integer(0), cmtName = character(0),
                        p = character(0), q = character(0), r = character(0),
                        expr = character(0), stringsAsFactors = FALSE))
    }
    .p3 <- unique(map$map3$p)
    .q3 <- unique(map$map3$q)
    .r3 <- unique(map$map3$r)
    .rows <- list()
    for (.c in cmts) {
      .nm <- .cmtName(.c)
      .symName <- paste0("rx_", kind, "_", .nm, "_")
      if (!exists(.symName, envir = .model)) next
      .sym <- get(.symName, envir = .model)
      for (.p in .p3) {
        for (.q in .q3) {
          for (.r in .r3) {
            .e <- .rxEventSensD3Expr(.sym, .p, .q, .r, .states)
            if (.e != "0" && .e != "0.0") {
              .rows[[length(.rows) + 1L]] <-
                data.frame(cmt = .c, cmtName = .nm, p = .p, q = .q, r = .r,
                           expr = .e, stringsAsFactors = FALSE)
            }
          }
        }
      }
    }
    if (length(.rows) == 0L) {
      return(data.frame(cmt = integer(0), cmtName = character(0),
                        p = character(0), q = character(0), r = character(0),
                        expr = character(0), stringsAsFactors = FALSE))
    }
    do.call(rbind, .rows)
  }
  ## d(F)/dq table: the first-order total derivative evaluated at the
  ## calcSens2 parameters, indexed (cmt, q) in calcSens2's own index space so
  ## the C buffer shares the same qIdx as d2Lag/d2F (no runtime remap)
  .buildQ <- function(cmts, kind, qParams) {
    .rows <- list()
    for (.c in cmts) {
      .nm <- .cmtName(.c)
      .symName <- paste0("rx_", kind, "_", .nm, "_")
      if (!exists(.symName, envir = .model)) next
      .sym <- get(.symName, envir = .model)
      for (.q in qParams) {
        .e <- .rxEventSensDExpr(.model, .sym, .q, .states)
        if (.e != "0" && .e != "0.0") {
          .rows[[length(.rows) + 1L]] <-
            data.frame(cmt = .c, cmtName = .nm, param = .q, expr = .e,
                       stringsAsFactors = FALSE)
        }
      }
    }
    if (length(.rows) == 0L) {
      return(data.frame(cmt = integer(0), cmtName = character(0),
                        param = character(0), expr = character(0),
                        stringsAsFactors = FALSE))
    }
    do.call(rbind, .rows)
  }
  ## d(J[k][c])/dq table: the total derivative of the physical Jacobian column
  ## wrt a calcSens2 parameter; J[k][c] = d(rx__d_dt_<k>__)/dx_c is available
  ## symbolically for every model type.  Indexed (cmt, k, q): cmt = the
  ## lag-carrying compartment, k = 0-based physical-state row.
  .buildJacQ <- function(cmts) {
    if (is.null(map$map2)) {
      return(data.frame(cmt = integer(0), k = integer(0), q = character(0),
                        expr = character(0), stringsAsFactors = FALSE))
    }
    .q2 <- unique(map$map2$q)
    .rows <- list()
    for (.c in cmts) {
      .cName <- .cmtName(.c)
      for (.kIdx in seq_along(.states)) {
        .kName <- .states[.kIdx]
        .fSymName <- paste0("rx__d_dt_", .kName, "__")
        if (!exists(.fSymName, envir = .model)) next
        .fSym <- get(.fSymName, envir = .model)
        .Jkc <- tryCatch(symengine::D(.fSym, symengine::S(.cName)),
                         error = function(e) NULL)
        if (is.null(.Jkc)) next
        .JkcTxt <- rxFromSE(.Jkc)
        if (.JkcTxt == "0" || .JkcTxt == "0.0") next
        for (.q in .q2) {
          .dJ <- .rxEventSensDSym(.Jkc, .q, .states)
          if (is.null(.dJ)) next
          .e <- rxFromSE(.dJ)
          if (.e != "0" && .e != "0.0") {
            .rows[[length(.rows) + 1L]] <-
              data.frame(cmt = .c, k = .kIdx - 1L, q = .q, expr = .e,
                         stringsAsFactors = FALSE)
          }
        }
      }
    }
    if (length(.rows) == 0L) {
      return(data.frame(cmt = integer(0), k = integer(0), q = character(0),
                        expr = character(0), stringsAsFactors = FALSE))
    }
    do.call(rbind, .rows)
  }
  .q2All <- if (is.null(map$map2)) character(0) else unique(map$map2$q)
  list(lag = .build(map$lagCmt, "lag"), f = .build(map$fCmt, "f"),
       rate = .build(map$rateCmt, "rate"), dur = .build(map$durCmt, "dur"),
       f2 = .build2(map$fCmt, "f"), lag2 = .build2(map$lagCmt, "lag"),
       rate2 = .build2(map$rateCmt, "rate"), dur2 = .build2(map$durCmt, "dur"),
       f3 = .build3(map$fCmt, "f"),
       fq = .buildQ(map$fCmt, "f", .q2All),
       lagJacQ = .buildJacQ(map$lagCmt),
       ## d(alag)/dq safety guard: when q also drives the same event's alag,
       ## the product-rule 2nd-order dtau row misses a Leibniz/moving-boundary
       ## term (dS^p_k/dt * dLag_q[c]); this table lets the runtime skip those
       ## (cmt, q) pairs rather than inject a wrong nonzero value.
       lagQ = .buildQ(map$lagCmt, "lag", .q2All),
       ## d(dur)/dq for the quotient-rule 2nd derivative of rate=F*amt/dur,
       ## in calcSens2's own index space (like fq/lagQ)
       durQ = .buildQ(map$durCmt, "dur", .q2All))
}

#' Generate the C assignment lines for the dLag / dF functions
#'
#' Produces the body assignment lines writing each dosing-parameter total
#' derivative into a flat per-subject scratch buffer indexed
#' `(cmt0 * nSensParam + paramIdx)`.  The lines are inserted into the generated
#' C verbatim; the only rewrite needed is nlmixr2's indexed
#' `THETA[n]`/`ETA[n]` parameters, which `.rxEventSensCExpr()` maps to the
#' codegen locals `_THETA_n_`/`_ETA_n_`.
#'
#' @param info An `.rxEventSensInfo()` result (mode + map + derivs).
#' @return list with `nSensParam`, `paramIdx` (named 0-based), and character
#'   vectors `lag` and `f` of C assignment lines; `NULL` if `info` is `NULL`.
#' @noRd
.rxEventSensCExpr <- function(expr, plainParams = character(0)) {
  # THETA[n]/ETA[n] map to the codegen locals _THETA_n_/_ETA_n_ unless the
  # model declares the plain name THETA_n_ itself (then use it, no leading _).
  .rw <- function(expr, kind) {
    .toks <- unique(regmatches(expr, gregexpr(paste0(kind, "\\[[0-9]+\\]"), expr))[[1]])
    for (.tok in .toks) {
      .n <- sub(paste0(kind, "\\[([0-9]+)\\]"), "\\1", .tok)
      .plain <- paste0(kind, "_", .n, "_")
      .repl <- if (.plain %in% plainParams) .plain else paste0("_", .plain)
      expr <- gsub(.tok, .repl, expr, fixed = TRUE)
    }
    expr
  }
  .rw(.rw(expr, "THETA"), "ETA")     # THETA before ETA (ETA[ nests inside THETA[)
}

.rxEventSensCLines <- function(info) {
  if (is.null(info)) return(NULL)
  .pp <- info$params                     # declared param names (plain THETA_n_ vs indexed)
  .params <- info$map$sensParams
  .np <- length(.params)
  .pIdx <- stats::setNames(seq_along(.params) - 1L, .params)
  .lines <- function(tab, buf) {
    if (is.null(tab) || nrow(tab) == 0L) return(character(0))
    .cmt0 <- tab$cmt - 1L                      # 0-based, matches _alag[_cmt]
    .idx <- .cmt0 * .np + .pIdx[tab$param]
    sprintf("  %s[%d] = %s;", buf, .idx, .rxEventSensCExpr(tab$expr, .pp))
  }
  ## 2nd-order buffer: (cmt0*(np*np2) + pIdx*np2 + qIdx)
  .q2 <- if (is.null(info$map$map2)) character(0) else unique(info$map$map2$q)
  .np2 <- length(.q2)
  .qIdx <- stats::setNames(seq_along(.q2) - 1L, .q2)
  .lines2 <- function(tab, buf) {
    if (is.null(tab) || nrow(tab) == 0L) return(character(0))
    .cmt0 <- tab$cmt - 1L
    .idx <- .cmt0 * (.np * .np2) + .pIdx[tab$p] * .np2 + .qIdx[tab$q]
    sprintf("  %s[%d] = %s;", buf, .idx, .rxEventSensCExpr(tab$expr, .pp))
  }
  ## 3rd-order buffer: (cmt0*(np*np2*np3) + pIdx*(np2*np3) + qIdx*np3 + rIdx);
  ## F row only
  .r3 <- if (is.null(info$map$map3)) character(0) else unique(info$map$map3$r)
  .np3 <- length(.r3)
  .rIdx <- stats::setNames(seq_along(.r3) - 1L, .r3)
  .lines3 <- function(tab, buf) {
    if (is.null(tab) || nrow(tab) == 0L) return(character(0))
    .cmt0 <- tab$cmt - 1L
    .idx <- .cmt0 * (.np * .np2 * .np3) + .pIdx[tab$p] * (.np2 * .np3) +
      .qIdx[tab$q] * .np3 + .rIdx[tab$r]
    sprintf("  %s[%d] = %s;", buf, .idx, .rxEventSensCExpr(tab$expr, .pp))
  }
  ## d(F)/dq buffer: (cmt0*np2 + qIdx), q in calcSens2's own index space
  .linesQ <- function(tab, buf) {
    if (is.null(tab) || nrow(tab) == 0L) return(character(0))
    .cmt0 <- tab$cmt - 1L
    .idx <- .cmt0 * .np2 + .qIdx[tab$param]
    sprintf("  %s[%d] = %s;", buf, .idx, .rxEventSensCExpr(tab$expr, .pp))
  }
  ## d(J[k][c])/dq buffer: (cmt0*(nState*np2) + k*np2 + qIdx), sized
  ## nState*nState*np2 (every possible cmt slot, like the other buffers)
  .ns <- info$map$nState
  .linesJacQ <- function(tab, buf) {
    if (is.null(tab) || nrow(tab) == 0L) return(character(0))
    .cmt0 <- tab$cmt - 1L
    .idx <- .cmt0 * (.ns * .np2) + tab$k * .np2 + .qIdx[tab$q]
    sprintf("  %s[%d] = %s;", buf, .idx, .rxEventSensCExpr(tab$expr, .pp))
  }
  list(
    nSensParam = .np,
    paramIdx = .pIdx,
    nSensParam2 = .np2,
    paramIdx2 = .qIdx,
    nSensParam3 = .np3,
    paramIdx3 = .rIdx,
    lag = .lines(info$derivs$lag, "_dLagSave"),
    f = .lines(info$derivs$f, "_dFSave"),
    rate = .lines(info$derivs$rate, "_dRateSave"),
    dur = .lines(info$derivs$dur, "_dDurSave"),
    f2 = .lines2(info$derivs$f2, "_d2FSave"),
    lag2 = .lines2(info$derivs$lag2, "_d2LagSave"),
    rate2 = .lines2(info$derivs$rate2, "_d2RateSave"),
    dur2 = .lines2(info$derivs$dur2, "_d2DurSave"),
    f3 = .lines3(info$derivs$f3, "_d3FSave"),
    fq = .linesQ(info$derivs$fq, "_dFQSave"),
    lagJacQ = .linesJacQ(info$derivs$lagJacQ, "_dLagJacSave"),
    lagQ = .linesQ(info$derivs$lagQ, "_dLagQSave"),
    durQ = .linesQ(info$derivs$durQ, "_dDurQSave")
  )
}

#' dLag/dF/dRate/dDur/d2F/d2Lag/d2Rate/d2Dur/d3F/dFQ/dLagJac/dLagQ/dDurQ C body lines for codegen
#'
#' Returns the 13 body-line strings (empty when none), passed as `.Call`
#' arguments so the lines reach codegen in the same package instance (robust
#' under `pkgload::load_all`).
#'
#' @param info An `.rxEventSensInfo()` result, or `NULL`.
#' @return character(13): the dLag, dF, dRate, dDur, d2F, d2Lag, d2Rate,
#'   d2Dur, d3F, dFQ, dLagJac, dLagQ, and dDurQ body lines.
#' @noRd
.rxEventSensCodeStrings <- function(info) {
  .cl <- .rxEventSensCLines(info)
  .join <- function(x) if (is.null(.cl) || length(x) == 0L) "" else paste(x, collapse = "\n")
  c(.join(.cl$lag), .join(.cl$f), .join(.cl$rate), .join(.cl$dur), .join(.cl$f2),
    .join(.cl$lag2), .join(.cl$rate2), .join(.cl$dur2), .join(.cl$f3),
    .join(.cl$fq), .join(.cl$lagJacQ), .join(.cl$lagQ), .join(.cl$durQ))
}

#' Does this model need the `calc_jac`-based dtau/lag Jacobian column?
#'
#' matExp()/indLin() models have no functional `dydt()`, so `handle_evid`'s
#' central-difference Jacobian column is always zero for them;
#' `rxSensMatExp()` emits explicit `df()/dy()` lines and the runtime should
#' read `calc_jac` instead.
#'
#' @param object Anything `rxModelVars()` accepts.
#' @return `TRUE`/`FALSE`.
#' @noRd
.rxEventSensUseCalcJac <- function(object) {
  length(rxModelVars(object)$indLin) > 0L
}

#' Push the event-sensitivity runtime dims to the solver before a solve
#'
#' Reads the model's `eventSensInfo` (attached by `rxode2()` when
#' `eventSens != "fd"`) and sets the C-side runtime gate: `active` (1 when jump
#' injection should run), `nState` (physical states), `nParam` (first-order
#' sensitivity parameters).  A no-op (`active = 0`) for `fd` models, models
#' without sensitivities, or objects that carry no `eventSensInfo`.
#'
#' @param object A solve target (rxode2 model env, UI, etc.).
#' @return invisibly `NULL`.
#' @noRd
.rxSetEventSensDims <- function(object) {
  .info <- tryCatch(object$eventSensInfo, error = function(e) NULL)
  if (is.null(.info) || identical(.info$mode, "fd")) {
    .Call(`_rxode2_setEventSensUseCalcJac`, 0L)
    .Call(`_rxode2_setEventSensNParam3`, 0L)
    return(invisible(.Call(`_rxode2_setEventSensDims`, 0L, 0L, 0L, 0L)))
  }
  .nState <- .info$map$nState
  .nParam <- length(.info$map$sensParams)
  ## number of second-order (calcSens2) parameters; 0 when no Hessian path
  .nParam2 <- if (is.null(.info$map$map2)) 0L else length(unique(.info$map$map2$q))
  ## number of third-order (calcSens3) parameters; 0 when no Phase H1 path
  .nParam3 <- if (is.null(.info$map$map3)) 0L else length(unique(.info$map$map3$r))
  .Call(`_rxode2_setEventSensUseCalcJac`, as.integer(.rxEventSensUseCalcJac(object)))
  .Call(`_rxode2_setEventSensNParam3`, as.integer(.nParam3))
  invisible(.Call(`_rxode2_setEventSensDims`, 1L,
                  as.integer(.nState), as.integer(.nParam), as.integer(.nParam2)))
}

#' Point the rxode2 event-sensitivity globals at a jump-sensitivity model
#'
#' For downstream packages (e.g. nlmixr2est's FOCEi) that solve a sensitivity
#' model through a direct C++ `ind_solve()` loop, bypassing `rxSolve()`: sets
#' rxode2's event ("jump") sensitivity function pointers and runtime dims to
#' `model` and turns the jumps on.  The jump blocks are bounds-guarded, so
#' smaller models solved afterwards skip the injection safely.  Pair with
#' `rxEventSensDeactivate()` after the run.
#'
#' @param model A built jump-sensitivity model (carrying `eventSensInfo`).
#' @return invisibly `TRUE` when the jumps were activated, `FALSE` otherwise
#'   (fd model / no sensitivities).
#' @export
#' @keywords internal
rxEventSensLoadModel <- function(model) {
  .info <- tryCatch(model$eventSensInfo, error = function(e) NULL)
  if (is.null(.info) || identical(.info$mode, "fd")) return(invisible(FALSE))
  .trans <- rxModelVars(model)$trans
  .nState <- .info$map$nState
  .nParam <- length(.info$map$sensParams)
  .nParam2 <- if (is.null(.info$map$map2)) 0L else length(unique(.info$map$map2$q))
  .nParam3 <- if (is.null(.info$map$map3)) 0L else length(unique(.info$map$map3$r))
  .Call(`_rxode2_eventSensLoad`, .trans, 1L, as.integer(.nState),
        as.integer(.nParam), as.integer(.nParam2))
  .Call(`_rxode2_setEventSensUseCalcJac`, as.integer(.rxEventSensUseCalcJac(model)))
  .Call(`_rxode2_setEventSensNParam3`, as.integer(.nParam3))
  invisible(TRUE)
}

#' Turn off the rxode2 event-sensitivity jump injection
#'
#' Resets the runtime gate set by [rxEventSensLoadModel()] (active = 0) after a
#' direct C++ solve loop completes, so later unrelated solves are unaffected.
#'
#' @return invisibly `NULL`.
#' @export
#' @keywords internal
rxEventSensDeactivate <- function() {
  .Call(`_rxode2_setEventSensUseCalcJac`, 0L)
  .Call(`_rxode2_setEventSensNParam3`, 0L)
  invisible(.Call(`_rxode2_setEventSensDims`, 0L, 0L, 0L, 0L))
}

#' Assemble the event-sensitivity information for a built model
#'
#' Combines the resolved mode, the index map, and the symbolic total-derivative
#' tables; `NULL` for `mode = "fd"` or models without first-order sensitivities.
#'
#' @param obj A built model (rxUi, rxode2, modelVars).
#' @param mode Resolved mode from `.rxEventSensMode()`.
#' @return A list `(mode, map, derivs)` or `NULL`.
#' @noRd
.rxEventSensInfo <- function(obj, mode) {
  if (identical(mode, "fd")) return(NULL)
  .map <- .rxEventSensMap(obj)
  if (is.null(.map)) return(NULL)
  .map <- .rxEventSensFilterMap(obj, .map)
  if (is.null(.map)) return(NULL)
  list(mode = mode, map = .map, derivs = .rxEventSensDerivs(obj, map = .map),
       params = rxModelVars(obj)$params)   # declared param names (plain vs indexed)
}

rxTest({
  # Event ("jump") sensitivities -- Phase A: build-time index map and the
  # symbolic total-derivatives of the dosing parameters (alag, F).
  # See ~/src/rxode2-event-sensitivities-plan.md.

  .mkMod <- function(modStr) {
    # depot+central first-order absorption with modeled alag + F, sensitivities
    # wrt two etas, full Jacobian (as eventSens="jump" will force).
    rxode2(modStr, calcSens = c("eta_ka", "eta_lag"), calcJac = TRUE)
  }

  .modConstF <- "
    ka <- exp(tka + eta_ka)
    cl <- exp(tcl)
    v  <- exp(tv)
    alag(depot) <- exp(tlag + eta_lag)
    f(depot)    <- expit(tf)
    d/dt(depot)   <- -ka * depot
    d/dt(central) <-  ka * depot - cl / v * central
  "

  .modStateF <- "
    ka <- exp(tka + eta_ka)
    cl <- exp(tcl)
    v  <- exp(tv)
    alag(depot) <- exp(tlag + eta_lag)
    f(depot)    <- 1 / (1 + exp(-(tf + 0.1 * central)))
    d/dt(depot)   <- -ka * depot
    d/dt(central) <-  ka * depot - cl / v * central
  "

  test_that(".rxEventSensMap relates sens compartments to states/params", {
    im <- .rxEventSensMap(.mkMod(.modConstF))
    expect_equal(im$states, c("depot", "central"))
    expect_equal(im$nState, 2L)
    expect_equal(unname(im$stateCmt[c("depot", "central")]), c(1L, 2L))
    expect_setequal(im$sensParams, c("eta_ka", "eta_lag"))
    # one map row per (state, param): 2 states x 2 params
    expect_equal(nrow(im$map), 4L)
    # rx__sens_depot_BY_eta_ka__ must be a real compartment mapped to depot
    .row <- im$map[im$map$state == "depot" & im$map$param == "eta_ka", ]
    expect_equal(.row$stateCmt, 1L)
    expect_true(.row$sensCmt > im$nState) # sens compartments come after states
    # depot carries modeled alag + F, not rate/dur
    expect_equal(im$lagCmt, 1L)
    expect_equal(im$fCmt, 1L)
    expect_length(im$rateCmt, 0L)
    expect_length(im$durCmt, 0L)
  })

  test_that(".rxEventSensMap returns NULL without sensitivities", {
    m <- rxode2("ka <- 1\nd/dt(depot) <- -ka * depot")
    expect_null(.rxEventSensMap(m))
  })

  test_that(".rxEventSensSplit handles underscored state/param names", {
    s <- c("rx__sens_depot_BY_eta_ka__", "rx__sens_central_BY_eta_lag__")
    sp <- .rxEventSensSplit(s, c("depot", "central"))
    expect_equal(sp$state, c("depot", "central"))
    expect_equal(sp$param, c("eta_ka", "eta_lag"))
    # second-order names are not first-order: state/param NA
    sp2 <- .rxEventSensSplit("rx__sens_depot_BY_eta_ka_BY_eta_lag__",
                             c("depot", "central"))
    expect_true(is.na(sp2$state))
  })

  test_that("alag total-derivative is the plain partial when state-independent", {
    d <- .rxEventSensDerivs(.mkMod(.modConstF))
    # alag depends only on eta_lag (no states) -> single non-zero row, no coupling
    expect_equal(nrow(d$lag), 1L)
    expect_equal(d$lag$param, "eta_lag")
    expect_false(any(grepl("rx__sens_", d$lag$expr)))
    # d(exp(eta_lag+tlag))/d(eta_lag) = exp(eta_lag+tlag)
    expect_equal(eval(parse(text = d$lag$expr),
                      list(eta_lag = 0.3, tlag = log(2))),
                 exp(0.3 + log(2)))
  })

  test_that("F total-derivative picks up the state-coupling term", {
    d <- .rxEventSensDerivs(.mkMod(.modStateF))
    # F depends on `central` -> d(F)/d(eta) is purely the coupling term
    # (no direct eta in F), nonzero for BOTH etas via S_central.
    expect_setequal(d$f$param, c("eta_ka", "eta_lag"))
    expect_true(all(grepl("rx__sens_central_BY_", d$f$expr)))
  })

  test_that(".rxEventSensMode resolves and validates the mode", {
    expect_equal(.rxEventSensMode("jump"), "jump")
    expect_equal(.rxEventSensMode("both"), "both")
    expect_equal(.rxEventSensMode("fd"), "fd")
    expect_equal(.rxEventSensMode("fdAll"), "fdAll")
    expect_error(.rxEventSensMode("bogus"), "eventSens")
    withr::with_options(list(rxode2.eventSens = "jump"), {
      expect_equal(.rxEventSensMode(NULL), "jump")
    })
    withr::with_options(list(rxode2.eventSens = NULL), {
      expect_equal(.rxEventSensMode(NULL), "jump")
    })
  })

  test_that("eventSens='jump' (default) is a no-op on rxode2()", {
    m <- rxode2("d/dt(depot) <- -ka*depot\nka <- 1",
                calcSens = "ka")
    expect_equal(m$eventSens, "jump")
    expect_null(m$eventSensInfo)
  })

  test_that("eventSens='jump' forces calcJac and attaches Phase-A info", {
    m <- rxode2(.modConstF, calcSens = c("eta_ka", "eta_lag"),
                eventSens = "jump")
    expect_equal(m$eventSens, "jump")
    expect_true(isTRUE(m$calcJac))
    info <- m$eventSensInfo
    expect_equal(info$mode, "jump")
    expect_equal(nrow(info$map$map), 4L)
    # constant F (expit(tf)) -> no eta-dependent F derivative; alag depends on
    # eta_lag -> exactly one lag derivative row.
    expect_equal(nrow(info$derivs$lag), 1L)
    expect_equal(nrow(info$derivs$f), 0L)
  })

  test_that("eventSens='jump' on a model without sensitivities is a no-op", {
    m <- rxode2("d/dt(depot) <- -depot", eventSens = "jump")
    expect_null(m$eventSensInfo)
  })

  test_that("eventSens on a linCmt()+mtime() model downgrades to fd", {
    m <- rxode2({
      C2 <- linCmt(CL, V)
      mtime(t1) <- mt1
      mtime(t2) <- mt2
    }, eventSens = "jump", linCmtSens = "linCmtA", linCmtSensType = "A")
    # linCmt models downgrade to finite differences for event sensitivities (#1119)
    expect_equal(m$eventSens, "fd")
    expect_null(m$eventSensInfo)
    e <- eventTable() |>
      add.dosing(dose = 3, nbr.doses = 2, dosing.interval = 8) |>
      add.sampling(0:16)
    s <- rxSolve(m, e, params = c(V = 20, CL = 25, mt1 = 0.5, mt2 = 1.75))
    expect_true(all(c(0.5, 1.75) %in% s$time))
  })

  test_that("mixed ODE+linCmt downgrades to fd for event sensitivities", {
    m <- rxode2({
      C2 <- linCmt(CL, V)
      alag(eff) <- exp(tlag + eta_lag)
      d/dt(eff) <- kin - kout * eff
    }, calcSens = "eta_lag", eventSens = "fd",
    linCmtSens = "linCmtA", linCmtSensType = "A")
    # any linCmt model downgrades to FD (#1119): no jump metadata
    expect_equal(m$eventSens, "fd")
    expect_null(m$eventSensInfo)
  })

  test_that("eventSens='fdAll' keeps full finite-difference fallback", {
    m <- rxode2({
      C2 <- linCmt(CL, V)
      d/dt(eff) <- kin - kout * eff
    }, eventSens = "fdAll", linCmtSens = "linCmtA", linCmtSensType = "A")
    expect_equal(m$eventSens, "fd")
    expect_null(m$eventSensInfo)
  })

  test_that("eventSens='fdAll' works correctly in a population (multi-subject) solve", {
    # Regression for a previously-untested combination (plan Section 0.4 gap):
    # confirm fdAll's per-subject solving (each subject independently, no
    # jump machinery/eventSensInfo at all) isn't corrupted by the population
    # (multi-subject/parallel) solve path -- each subject's trajectory must
    # match an INDEPENDENT single-subject solve with that subject's own
    # modeled-lag parameter.
    mLin <- rxode2({
      C2 <- linCmt(CL, V)
      alag(central) <- exp(tlag + eta_lag)
    }, calcSens = "eta_lag", eventSens = "fdAll",
    linCmtSens = "linCmtA", linCmtSensType = "A")
    expect_equal(mLin$eventSens, "fd")
    expect_null(mLin$eventSensInfo)

    th <- c(V = 20, CL = 25, tlag = 0)
    ev <- data.frame(eta_lag = c(0, 0.3, -0.2))
    e <- et(amt = 100, cmt = "central", id = 1:3) |> et(seq(0, 12, 0.5), id = 1:3)

    sPop <- rxSolve(mLin, e, th, iCov = ev)
    for (.id in 1:3) {
      e1 <- eventTable() |>
        add.dosing(dose = 100, cmt = "central") |>
        add.sampling(seq(0, 12, 0.5))
      s1 <- rxSolve(mLin, e1, c(th, eta_lag = ev$eta_lag[.id]))
      expect_equal(sPop$central[sPop$id == .id], s1$central,
        tolerance = 1e-8, info = paste0("id: ", .id)
      )
    }
  })

  test_that("pure linCmt with modeled alag downgrades to fd", {
    m <- rxode2({
      C2 <- linCmt(CL, V)
      alag(central) <- exp(tlag + eta_lag)
    }, calcSens = "eta_lag", eventSens = "fd",
    linCmtSens = "linCmtA", linCmtSensType = "A")
    # pure linCmt downgrades to FD (#1119)
    expect_equal(m$eventSens, "fd")
    expect_null(m$eventSensInfo)
  })

  test_that("pure linCmt hybrid event path matches the ODE equivalent", {
    mLin <- rxode2({
      C2 <- linCmt(CL, V)
      alag(central) <- exp(tlag + eta_lag)
    }, calcSens = "eta_lag", eventSens = "fd",
    linCmtSens = "linCmtA", linCmtSensType = "A")
    mOde <- rxode2({
      d/dt(central) <- -CL / V * central
      alag(central) <- exp(tlag + eta_lag)
    }, calcSens = "eta_lag", eventSens = "jump")

    e <- eventTable() |>
      add.dosing(dose = 100, cmt = "central") |>
      add.sampling(seq(0, 12, 0.5))
    p <- c(V = 20, CL = 25, tlag = 0, eta_lag = 0)

    sLin <- rxSolve(mLin, e, params = p)
    sOde <- rxSolve(mOde, e, params = p)

    expect_equal(sLin$central, sOde$central, tolerance = 1e-5)
  })

  test_that("mixed ODE+linCmt with explicit jump still downgrades to fd", {
    # Even eventSens="jump" downgrades to FD when linCmt() is present (#1119):
    # the linCmt-compartment moving-boundary jump is not implemented, so all
    # linCmt models use finite differences for event sensitivities.
    m <- rxode2({
      C2 <- linCmt(CL, V)
      alag(eff) <- exp(tlag + eta_lag)
      d/dt(eff) <- kin - kout * eff
    }, calcSens = "eta_lag", eventSens = "jump",
    linCmtSens = "linCmtA", linCmtSensType = "A")
    expect_equal(m$eventSens, "fd")
    expect_null(m$eventSensInfo)
  })

  test_that("ODE d/dt() colliding with a linCmt reserved compartment name warns", {
    # Naming an ODE compartment `depot` alongside an oral/multi-cmt linCmt (which
    # reserves `depot`) conflates the two compartments: the ODE state loses its
    # sensitivity expansion, so its sensitivities are silently incorrect.  Warn.
    expect_warning(
      rxode2({
        ka <- exp(tka)
        alag(depot) <- 2 * exp(eta_lag)
        d/dt(depot)   <- -ka * depot
        d/dt(peri)    <-  ka * depot - 0.3 * peri
        C2 <- linCmt(CL, V)
      }, calcSens = "eta_lag", eventSens = "jump",
      linCmtSens = "linCmtA", linCmtSensType = "A"),
      "share a name with linCmt"
    )
    # a valid (non-colliding) linCmt model: no warning; downgrades to fd (#1119)
    mOk <- suppressWarnings(rxode2({
      ka <- exp(tka)
      alag(gut) <- 2 * exp(eta_lag)
      d/dt(gut) <- -ka * gut
      d/dt(eff) <-  ka * gut - 0.3 * eff
      C2 <- linCmt(CL, V)
    }, calcSens = "eta_lag", eventSens = "fd",
    linCmtSens = "linCmtA", linCmtSensType = "A"))
    expect_equal(mOk$eventSens, "fd")
    expect_null(mOk$eventSensInfo)
  })

  test_that("mixed ODE+linCmt hybrid event path matches the ODE equivalent", {
    mLin <- rxode2({
      C2 <- linCmt(CL, V)
      alag(central) <- exp(tlag + eta_lag)
      d/dt(eff) <- kin - kout * eff
    }, calcSens = "eta_lag", eventSens = "fd",
    linCmtSens = "linCmtA", linCmtSensType = "A")
    mOde <- rxode2({
      d/dt(central) <- -CL / V * central
      alag(central) <- exp(tlag + eta_lag)
      d/dt(eff) <- kin - kout * eff
    }, calcSens = "eta_lag", eventSens = "jump")

    e <- eventTable() |>
      add.dosing(dose = 100, cmt = "central") |>
      add.sampling(seq(0, 12, 0.5))
    p <- c(V = 20, CL = 25, kin = 1, kout = 0.2, tlag = 0, eta_lag = 0)

    sLin <- rxSolve(mLin, e, params = p)
    sOde <- rxSolve(mOde, e, params = p)

    expect_equal(sLin$central, sOde$central, tolerance = 1e-5)
    expect_equal(sLin$eff, sOde$eff, tolerance = 1e-5)
  })

  test_that("eventSens='jump' works with ODE+mtime() models", {
    .mod <- "
      d/dt(depot) <- -ka*depot
      d/dt(central) <- ka*depot-kel*central
      ka <- exp(tka)
      kel <- exp(tkel)
      mtime(t1) <- mt1
      mtime(t2) <- mt2
    "
    mj <- rxode2(.mod, calcSens = "tka", eventSens = "jump")
    mfd <- rxode2(.mod, calcSens = "tka", eventSens = "fd")
    e <- eventTable() |>
      add.dosing(dose = 100, cmt = "depot") |>
      add.sampling(0:12)
    .p <- c(tka = 0, tkel = log(0.2), mt1 = 0.5, mt2 = 1.75)
    sj <- rxSolve(mj, e, params = .p)
    sfd <- rxSolve(mfd, e, params = .p)
    expect_true(isTRUE(mj$calcJac))
    expect_equal(sj$central, sfd$central, tolerance = 1e-5)
  })

  test_that(".rxEventSensCLines emits correctly-indexed C assignment lines", {
    m <- rxode2(.modStateF, calcSens = c("eta_ka", "eta_lag"),
                eventSens = "jump")
    cl <- .rxEventSensCLines(m$eventSensInfo)
    expect_equal(cl$nSensParam, 2L)
    expect_equal(unname(cl$paramIdx[c("eta_ka", "eta_lag")]), c(0L, 1L))
    # alag(depot): cmt0=0, eta_lag idx 1 -> _dLagSave[0*2+1] = _dLagSave[1]
    expect_length(cl$lag, 1L)
    expect_match(cl$lag, "_dLagSave\\[1\\] = exp\\(eta_lag\\+tlag\\);", fixed = FALSE)
    # state-dependent F: both etas via S_central, indices 0 and 1
    expect_length(cl$f, 2L)
    expect_true(any(grepl("_dFSave[0]", cl$f, fixed = TRUE)))
    expect_true(any(grepl("_dFSave[1]", cl$f, fixed = TRUE)))
    # right-hand sides reference the exact local names the F function declares
    expect_true(all(grepl("rx__sens_central_BY_", cl$f)))
  })

  test_that(".rxEventSensCLines is NULL for fd / no-sens", {
    expect_null(.rxEventSensCLines(NULL))
  })

  test_that(".rxEventSensCExpr maps indexed vs plainly-declared params (1st + 2nd order)", {
    # nlmixr2's indexed THETA[n]/ETA[n] map to the codegen locals _THETA_n_/_ETA_n_
    # (populated from _PP[]); this is what the dF/d2F assignment lines reference.
    expect_equal(.rxEventSensCExpr("exp(THETA[2])"), "exp(_THETA_2_)")
    expect_equal(.rxEventSensCExpr("THETA[1]*ETA[3]"), "_THETA_1_*_ETA_3_")
    # But a model may declare the parameter under its own plain name THETA_n_/ETA_n_
    # (e.g. nlmixr2est's augmented outer-gradient model, whose direction params are
    # named THETA_n_/ETA_n_); the preamble then declares them verbatim, so the dose
    # (and 2nd-order dose) derivatives must reference the plain name -- otherwise the
    # emitted _THETA_n_ is undeclared and the model fails to compile.
    expect_equal(.rxEventSensCExpr("exp(THETA[2])", plainParams = "THETA_2_"), "exp(THETA_2_)")
    expect_equal(.rxEventSensCExpr("THETA[1]*ETA[3]", plainParams = c("THETA_1_", "ETA_3_")),
                 "THETA_1_*ETA_3_")
    # mixed: only the plainly-declared params drop the leading underscore
    expect_equal(.rxEventSensCExpr("THETA[1]*ETA[3]", plainParams = "THETA_1_"), "THETA_1_*_ETA_3_")
    # info$params carries the declared names so .rxEventSensCLines picks the right form
    m <- rxode2(.modStateF, calcSens = c("eta_ka", "eta_lag"), eventSens = "jump")
    expect_true("params" %in% names(m$eventSensInfo))
  })

  test_that("eventSens='jump' model with state-dependent F compiles and solves", {
    # Exercises the full lines channel: R-generated dLag/dF assignment lines are
    # pushed to codegen, spliced into the function bodies (with the state-coupling
    # term referencing populated sens locals), and the model compiles + solves.
    # (The exact emitted C is asserted at the R level by the .rxEventSensCLines
    # tests above; here we verify the end-to-end build/solve.)
    m <- rxode2(.modStateF, calcSens = c("eta_ka", "eta_lag"),
                eventSens = "jump")
    expect_s3_class(m, "rxode2")
    e <- et(amt = 100, cmt = "depot")
    e <- et(e, seq(0, 12, 4))
    ini <- c(tka = 0, tcl = 1, tv = 2, tlag = 0, tf = 0,
             eta_ka = 0, eta_lag = 0)
    s <- rxSolve(m, e, ini)
    expect_true(nrow(s) > 0L)
    # jump-mode solve matches fd-mode for the physical states: the jump injection
    # only touches the sensitivity compartments, never the physical state.  The
    # match is to solver tolerance (not bit-identical) because jump forces
    # calcJac=TRUE -> the analytic Jacobian gives LSODA a slightly different step
    # path than fd's numerical Jacobian (~1e-7 on the trajectory).
    mfd <- rxode2(.modStateF, calcSens = c("eta_ka", "eta_lag"),
                  eventSens = "fd")
    sfd <- rxSolve(mfd, e, ini)
    expect_equal(s$central, sfd$central, tolerance = 1e-5)
  })

  test_that("additive-bolus F jump (ddelta row) matches finite differences", {
    # Constant alag (no eta) + F = expit(tf + eta_f): the only event-sensitivity
    # contribution at the dose is the ddelta row, d(delta)/d(eta_f) = amt*dF.
    # The analytic jump-mode sensitivity must match a central difference of the
    # central state wrt eta_f; the fd-mode sens ODE alone (no jump) must NOT.
    .mod <- "
      ka <- exp(tka + eta_ka)
      cl <- exp(tcl); v <- exp(tv)
      alag(depot) <- 1
      f(depot)    <- expit(tf + eta_f)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0, tcl = 1, tv = 2, tf = 0.3, eta_ka = 0, eta_f = 0)
    e <- et(amt = 100, cmt = "depot")
    e <- et(e, seq(0, 12, 2))
    .central <- function(p, mode) {
      m <- rxode2(.mod, calcSens = c("eta_ka", "eta_f"), eventSens = mode)
      rxSolve(m, e, p)
    }
    sj <- .central(pars, "jump")
    analytic <- sj[["rx__sens_central_BY_eta_f__"]]
    h <- 1e-4
    pp <- pars; pp["eta_f"] <- pars["eta_f"] + h
    pm <- pars; pm["eta_f"] <- pars["eta_f"] - h
    fd <- (.central(pp, "fd")$central - .central(pm, "fd")$central) / (2 * h)
    # analytic jump matches FD to FD precision
    expect_equal(analytic, fd, tolerance = 1e-4)
    # fd-mode sens ODE alone is missing the dF contribution -> clearly different
    sfd <- .central(pars, "fd")[["rx__sens_central_BY_eta_f__"]]
    expect_gt(max(abs(sfd - fd)), 1)
  })

  test_that("additive-bolus lag jump (dtau row) matches finite differences", {
    # alag = exp(tlag + eta_lag) depends on eta_lag, F constant: the jump is the
    # dtau row, -J[k][c]*delta*d(alag)/d(eta_lag), with J taken from a central
    # difference of dydt.  Observe AWAY from the (lagged) dose time, since at the
    # exact event time the sensitivity is discontinuous and a central FD straddles
    # the jump (the analytic value is the correct right-limit).
    .mod <- "
      ka <- exp(tka + eta_ka)
      cl <- exp(tcl); v <- exp(tv)
      alag(depot) <- exp(tlag + eta_lag)
      f(depot)    <- expit(tf)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2, tf = 0.5, tlag = 0,
              eta_ka = 0, eta_lag = 0)
    e <- et(amt = 100, cmt = "depot")
    e <- et(e, seq(0.5, 12, 1)) # off the dose time (alag = 1)
    .central <- function(p, mode) {
      m <- rxode2(.mod, calcSens = c("eta_ka", "eta_lag"), eventSens = mode)
      rxSolve(m, e, p)
    }
    sj <- .central(pars, "jump")
    h <- 1e-4
    pp <- pars; pp["eta_lag"] <- h
    pm <- pars; pm["eta_lag"] <- -h
    fd <- (.central(pp, "fd")$central - .central(pm, "fd")$central) / (2 * h)
    expect_equal(sj[["rx__sens_central_BY_eta_lag__"]], fd, tolerance = 1e-4)
    # eta_ka has no lag/F dependence -> no jump; must still match FD
    ppk <- pars; ppk["eta_ka"] <- h
    pmk <- pars; pmk["eta_ka"] <- -h
    fdK <- (.central(ppk, "fd")$central - .central(pmk, "fd")$central) / (2 * h)
    expect_equal(sj[["rx__sens_central_BY_eta_ka__"]], fdK, tolerance = 1e-4)
  })

  test_that("additive-bolus lag jump: coincident-time output uses one-sided limits", {
    # Regression for the doubled-sensitivity bug when an observation time exactly
    # equals a modeled-lag dose arrival.  alag(depot) = 2 * exp(eta_lag): with
    # eta_lag = 0 the dose (t = 0) arrives at EXACTLY t = 2, where an observation
    # is placed.  At that coincident time the sensitivity has a genuine jump
    # discontinuity; the reported value must match the STATE's reported side:
    #   - central (not dosed, continuous): reported state is the pre-arrival
    #     value, so its sensitivity must be the pre-jump (right/forward) limit.
    #   - depot (dosed): reported state is post-dose, so its sensitivity keeps the
    #     post-jump (left/backward) limit.
    # Before the fix the coincident central sensitivity was the post-jump
    # (backward) limit -- the wrong side, ~2x a symmetric FD -- which inflated the
    # FOCEi objective on models whose observations land on lagged arrivals.
    .mod <- "
      ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
      alag(depot) <- 2 * exp(eta_lag)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = log(1.15), tcl = log(0.135), tv = log(8), eta_lag = 0)
    e <- et(amt = 100, cmt = "depot") |> et(c(1.9, 2.0, 2.1, 4))
    m <- rxode2(.mod, calcSens = "eta_lag", eventSens = "jump")
    sj <- rxSolve(m, e, pars)
    h <- 1e-5
    pp <- pars; pp["eta_lag"] <-  h
    pm <- pars; pm["eta_lag"] <- -h
    s0 <- rxSolve(m, e, pars)
    sP <- rxSolve(m, e, pp)
    sM <- rxSolve(m, e, pm)
    .at <- function(s, tm, col) s[[col]][which(abs(s$time - tm) < 1e-8)[1]]
    # one-sided finite differences of the STATE at the coincident time t = 2
    fwdCentral <- (.at(sP, 2, "central") - .at(s0, 2, "central")) / h   # right
    bwdCentral <- (.at(s0, 2, "central") - .at(sM, 2, "central")) / h   # left
    bwdDepot   <- (.at(s0, 2, "depot")   - .at(sM, 2, "depot"))   / h
    ajCentral <- .at(sj, 2, "rx__sens_central_BY_eta_lag__")
    ajDepot   <- .at(sj, 2, "rx__sens_depot_BY_eta_lag__")
    # central: analytic == pre-jump (forward/right) limit (~0 here), and clearly
    # NOT the old doubled (backward/left) value.
    expect_equal(ajCentral, fwdCentral, tolerance = 1e-3)
    expect_true(abs(ajCentral - bwdCentral) > 0.1 * abs(bwdCentral))
    # depot: dosed compartment keeps the post-jump (backward/left) limit.
    expect_equal(ajDepot, bwdDepot, tolerance = 1e-2)
    # off-dose observations still match a two-sided central difference.
    for (tm in c(1.9, 2.1, 4)) {
      fd <- (.at(sP, tm, "central") - .at(sM, tm, "central")) / (2 * h)
      expect_equal(.at(sj, tm, "rx__sens_central_BY_eta_lag__"), fd,
                   tolerance = 1e-3)
    }
  })

  test_that("additive-bolus jumps accumulate across a multi-dose regimen", {
    # alag(depot)=exp(tlag+eta_lag), F=expit(tf+eta_f): every dose contributes
    # both a dtau and a ddelta jump.  The rows use method "add", so the jumps
    # must accumulate across the regimen.  A multi-dose schedule must still match
    # a central difference of the solution wrt each eta.
    .mod <- "
      ka <- exp(tka + eta_ka)
      cl <- exp(tcl); v <- exp(tv)
      alag(depot) <- exp(tlag + eta_lag)
      f(depot)    <- expit(tf + eta_f)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2, tf = 0.3, tlag = 0,
              eta_ka = 0, eta_lag = 0, eta_f = 0)
    e <- et(amt = 100, cmt = "depot", ii = 6, addl = 3) |>
      et(seq(0.5, 30, 1.5))
    .central <- function(p, mode) {
      m <- rxode2(.mod, calcSens = c("eta_lag", "eta_f"), eventSens = mode)
      rxSolve(m, e, p)
    }
    sj <- .central(pars, "jump")
    h <- 1e-4
    for (.eta in c("eta_lag", "eta_f")) {
      pp <- pars; pp[.eta] <- h
      pm <- pars; pm[.eta] <- -h
      fd <- (.central(pp, "fd")$central - .central(pm, "fd")$central) / (2 * h)
      expect_equal(sj[[paste0("rx__sens_central_BY_", .eta, "__")]], fd,
                   tolerance = 1e-3)
    }
  })

  test_that("replacement event jump (dp_j -> 0) matches finite differences", {
    # Replace central with a constant (50) at t=5 via an evid=5 event.  The
    # replaced state's value no longer depends on eta_ka, so d(central)/d(eta_ka)
    # must jump to 0 at the event and rebuild afterwards.  The analytic jump must
    # match a central difference of the solution; the fd-mode sensitivity ODE
    # (which never zeroes the sens compartment) must NOT.
    .mod <- "
      ka <- exp(tka + eta_ka)
      cl <- exp(tcl); v <- exp(tv)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2, eta_ka = 0)
    e <- et(amt = 100, cmt = "depot") |>
      et(time = 5, amt = 50, cmt = "central", evid = 5) |>
      et(seq(0.5, 12, 1))
    .central <- function(p, mode) {
      m <- rxode2(.mod, calcSens = "eta_ka", eventSens = mode)
      rxSolve(m, e, p)
    }
    sj <- .central(pars, "jump")
    h <- 1e-4
    pp <- pars; pp["eta_ka"] <- h
    pm <- pars; pm["eta_ka"] <- -h
    fd <- (.central(pp, "fd")$central - .central(pm, "fd")$central) / (2 * h)
    # observations at/after the replace must match FD (analytic jump correct)
    .post <- sj$time >= 5
    expect_equal(sj[["rx__sens_central_BY_eta_ka__"]][.post], fd[.post],
                 tolerance = 1e-4)
    # fd-mode sens ODE alone never zeroes the replaced state's sens -> differs
    sfd <- .central(pars, "fd")[["rx__sens_central_BY_eta_ka__"]]
    expect_gt(max(abs(sfd[.post] - fd[.post])), 1)
  })

  test_that("multiplicative event jump (dp_j *= alpha) matches finite differences", {
    # Multiply central by alpha=0.5 at t=5 via an evid=6 event.  d(central)/deta
    # must be scaled by the same 0.5 at the event.  Analytic jump matches FD; the
    # fd-mode sens ODE (no scaling of the sens compartment) does not.
    .mod <- "
      ka <- exp(tka + eta_ka)
      cl <- exp(tcl); v <- exp(tv)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2, eta_ka = 0)
    e <- et(amt = 100, cmt = "depot") |>
      et(time = 5, amt = 0.5, cmt = "central", evid = 6) |>
      et(seq(0.5, 12, 1))
    .central <- function(p, mode) {
      m <- rxode2(.mod, calcSens = "eta_ka", eventSens = mode)
      rxSolve(m, e, p)
    }
    sj <- .central(pars, "jump")
    h <- 1e-4
    pp <- pars; pp["eta_ka"] <- h
    pm <- pars; pm["eta_ka"] <- -h
    fd <- (.central(pp, "fd")$central - .central(pm, "fd")$central) / (2 * h)
    .post <- sj$time >= 5
    expect_equal(sj[["rx__sens_central_BY_eta_ka__"]][.post], fd[.post],
                 tolerance = 1e-4)
    sfd <- .central(pars, "fd")[["rx__sens_central_BY_eta_ka__"]]
    expect_gt(max(abs(sfd[.post] - fd[.post])), 0.1)
  })

  test_that("replace/multiply zero the 2nd-order (Hessian) compartment too", {
    # Extends Phase F's additive-bolus-only 2nd-order jump (the "dx1/dp_j"
    # row, NOT the dtau row) to replace/multiply: a constant replacement
    # value has an identically zero 2nd derivative wrt any parameter pair
    # (same reasoning as 1st order), and alpha (treated as parameter-fixed
    # for this row, same convention as 1st order) scales the 2nd-order
    # compartment the same way. Validated against a finite difference of the
    # analytic 1st-order sensitivity (the plan's standard 2nd-order
    # validation strategy -- a raw 2nd central difference of the solution is
    # too imprecise to catch a jump-sized correction).
    .mod <- "
      ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2)
    m1 <- rxode2(.mod, calcSens = "tka", eventSens = "jump")
    m2 <- rxode2(.mod, calcSens = "tka", calcSens2 = "tka", eventSens = "jump")
    h <- 1e-4
    pp <- pars; pp["tka"] <- pars["tka"] + h
    pm <- pars; pm["tka"] <- pars["tka"] - h

    e_replace <- et(amt = 100, cmt = "depot") |>
      et(time = 5, amt = 50, cmt = "central", evid = 5) |>
      et(seq(0.5, 12, 1))
    fd_replace <- (rxSolve(m1, e_replace, pp)$rx__sens_central_BY_tka__ -
      rxSolve(m1, e_replace, pm)$rx__sens_central_BY_tka__) / (2 * h)
    s_replace <- rxSolve(m2, e_replace, pars)
    expect_equal(s_replace$rx__sens_central_BY_tka_BY_tka__, fd_replace, tolerance = 1e-3)

    e_mult <- et(amt = 100, cmt = "depot") |>
      et(time = 5, amt = 0.5, cmt = "central", evid = 6) |>
      et(seq(0.5, 12, 1))
    fd_mult <- (rxSolve(m1, e_mult, pp)$rx__sens_central_BY_tka__ -
      rxSolve(m1, e_mult, pm)$rx__sens_central_BY_tka__) / (2 * h)
    s_mult <- rxSolve(m2, e_mult, pars)
    expect_equal(s_mult$rx__sens_central_BY_tka_BY_tka__, fd_mult, tolerance = 1e-3)
  })

  test_that("raw event-table replace with a modeled lag gets a dtau row (Phase B, B2)", {
    # Regression (Phase B "B2" gap, plan Section 4): a raw et(..., evid=5)
    # replace record on a compartment with a MODELED alag was found to
    # completely omit the dx1/dtau row (paper Table 1) -- the modeled lag
    # correctly shifts the event's OWN time (confirmed: alag applies to
    # evid=5/6 records exactly like normal doses), but the sensitivity was
    # silently exactly zero even though the true value is clearly nonzero
    # (confirmed via a hand-derived closed form and FD, ~20.2 here). The
    # in-model replace()/multiply() plugins were previously found NOT to need
    # this row (captured-dosing re-emits correctly) -- this gap is specific
    # to raw event-table evid=5/6 records.
    .mod <- "
      cl <- exp(tcl); v <- exp(tv)
      alag(central) <- exp(tlag + eta_lag)
      d/dt(central) <- -cl / v * central
    "
    pars <- c(tcl = 1, tv = 2, tlag = log(1.1), eta_lag = 0)
    e <- et(time = 5, amt = 50, cmt = "central", evid = 5) |> et(seq(0, 10, 0.05))
    mj <- rxode2(.mod, calcSens = "eta_lag", eventSens = "jump")
    mfd <- rxode2(.mod, calcSens = "eta_lag", eventSens = "fd")
    sj <- rxSolve(mj, e, pars)
    h <- 1e-4
    pp <- pars; pp["eta_lag"] <- pars["eta_lag"] + h
    pm <- pars; pm["eta_lag"] <- pars["eta_lag"] - h
    # tight atol/rtol on the FD reference solves: the default LSODA tolerance
    # is loose enough, relative to h, that its own step-to-step numerical
    # noise dominates a naive central difference here (~1% spurious offset
    # observed with default tolerances; confirmed to vanish under atol/rtol
    # 1e-12 -- same class of pitfall as too-small an h, not a real signal).
    fd <- (rxSolve(mfd, e, pp, atol = 1e-12, rtol = 1e-12)$central -
             rxSolve(mfd, e, pm, atol = 1e-12, rtol = 1e-12)$central) / (2 * h)
    # skip the sample landing exactly on the lagged event time (t=6.1): a
    # central FD straddling a true discontinuity is a known artifact, not a
    # correctness signal (documented in the plan's infusion-jump validation
    # note; the same subtlety applies here).
    .post <- sj$time >= 6.15 & sj$time < 10
    expect_equal(sj[["rx__sens_central_BY_eta_lag__"]][.post], fd[.post], tolerance = 1e-3)
    expect_gt(max(abs(sj[["rx__sens_central_BY_eta_lag__"]])), 15)
  })

  test_that("raw event-table multiply with a modeled lag gets a dtau row (Phase B, B2)", {
    # Same gap as above, for evid=6 (multiply): checked on a compartment OTHER
    # than the multiplied one (depot's multiply shifts central's future
    # absorption input), since a single-compartment homogeneous-linear decay
    # makes the paper's dx1/dtau row (J_cc*x1 - f_c) identically (and
    # correctly) zero -- not a useful regression target for THIS row.
    .mod <- "
      ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
      alag(depot) <- exp(tlag + eta_lag)
      depot(0) <- 100
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.3, tcl = 1, tv = 2, tlag = log(1.1), eta_lag = 0)
    e <- et(time = 5, amt = 0.6, cmt = "depot", evid = 6) |> et(seq(0, 15, 0.1))
    mj <- rxode2(.mod, calcSens = "eta_lag", eventSens = "jump")
    mfd <- rxode2(.mod, calcSens = "eta_lag", eventSens = "fd")
    sj <- rxSolve(mj, e, pars)
    h <- 1e-3
    pp <- pars; pp["eta_lag"] <- pars["eta_lag"] + h
    pm <- pars; pm["eta_lag"] <- pars["eta_lag"] - h
    sp <- rxSolve(mfd, e, pp, atol = 1e-11, rtol = 1e-11)
    sm <- rxSolve(mfd, e, pm, atol = 1e-11, rtol = 1e-11)
    fd <- (sp$central - sm$central) / (2 * h)
    .post <- sj$time >= 6.15 & sj$time < 15
    expect_equal(sj[["rx__sens_central_BY_eta_lag__"]][.post], fd[.post], tolerance = 1e-3)
    expect_gt(max(abs(sj[["rx__sens_central_BY_eta_lag__"]])), 0.01)
  })

  test_that("constant-rate infusion needs no jump (sens ODE alone is correct)", {
    # An infusion rate that does not depend on any parameter (and a fixed start
    # time) enters dydt as a parameter-independent forcing, so the symbolic
    # sensitivity ODE already captures it -- no jump contribution is required.
    # Both modes must agree with each other and with a finite difference; this
    # also guards that future infusion jumps stay zero in the constant-rate case.
    .mod <- "
      ka <- exp(tka + eta_ka)
      cl <- exp(tcl); v <- exp(tv)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2, eta_ka = 0)
    e <- et(amt = 100, rate = 20, cmt = "depot") |> et(seq(0.5, 12, 1))
    .central <- function(p, mode) {
      m <- rxode2(.mod, calcSens = "eta_ka", eventSens = mode)
      rxSolve(m, e, p)
    }
    sj <- .central(pars, "jump")[["rx__sens_central_BY_eta_ka__"]]
    sf <- .central(pars, "fd")[["rx__sens_central_BY_eta_ka__"]]
    h <- 1e-4
    pp <- pars; pp["eta_ka"] <- h
    pm <- pars; pm["eta_ka"] <- -h
    fd <- (.central(pp, "fd")$central - .central(pm, "fd")$central) / (2 * h)
    expect_equal(sj, sf)                    # jump adds nothing for constant rate
    expect_equal(sj, fd, tolerance = 1e-3)  # and both match the finite difference
  })

  test_that("modeled-rate infusion with NO eventSens/calcSens requested does not crash", {
    # Regression for a real C stack overflow: `_esInfusionDydtPre()`
    # (inst/include/rxode2parseHandleEvid.h) used to call `dydtEs()` (an
    # alias for the model's OWN compiled dydt()) UNCONDITIONALLY at every
    # MODEL_RATE_ON/MODEL_DUR_ON event, before any `_rxEsActive` check.
    # Since `handle_evid()` is itself invoked FROM WITHIN `dydt()` for
    # modeled-rate/duration events, this created unbounded re-entrant
    # recursion into `dydt()` on EVERY solve of a model with modeled
    # rate()/dur() -- entirely independent of whether eventSens/calcSens
    # was ever requested (found via `git bisect`, culprit commit
    # ce61e9a99, then reproduced standalone on a completely unrelated,
    # pre-existing test file with no calcSens= anywhere). This model
    # deliberately requests NEITHER calcSens NOR eventSens, so
    # `_rxEsActive` is 0 for the whole solve -- the crash fired
    # regardless.
    .mod <- "
      ka <- 0.5; cl <- 0.2; v <- 10
      rate(central) <- 20
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    m <- rxode2(.mod)
    e <- et(amt = 100, cmt = "central", rate = -1) |> et(seq(0.5, 12, 1))
    expect_no_error(rxSolve(m, e))
    expect_no_error(rxSolve(m, e, method = "lsode"))
  })

  test_that("in-model evid_() dosing plugins match finite differences (jump)", {
    # bolus(), replace(), multiply(), reset() and constant-rate infuse()/
    # infuseDur() pushed from inside the model must all yield jump-mode
    # sensitivities that match a central difference of the solution.  (Modeled
    # rate/dur -- where the infusion magnitude depends on a parameter -- is the
    # separate continuous-forcing case covered elsewhere.)
    .skel <- function(dose) paste0("
      ka <- exp(tka + eta_ka); cl <- exp(tcl); v <- exp(tv); rt <- 5
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
      if (t >= rt && t < rt + 0.4) { ", dose, " }")
    .doses <- c(
      bolus      = "bolus(50, central, 0, 0, 0)",
      infuse     = "infuse(100, 50, central, 0, 0, 0)",
      infuseDur  = "infuseDur(100, 2, central, 0, 0, 0)",
      replace    = "if (central > 0) replace(40, central)",
      multiply   = "if (central > 0) multiply(0.5, central)",
      reset      = "reset()"
    )
    pars <- c(tka = 0.2, tcl = 1, tv = 2, eta_ka = 0)
    e <- et(amt = 100, cmt = "depot") |> et(seq(0.5, 14, 0.5))
    h <- 1e-4
    for (.nm in names(.doses)) {
      .mod <- .skel(.doses[[.nm]])
      .central <- function(p, mode) {
        rxSolve(rxode2(.mod, calcSens = "eta_ka", eventSens = mode), e, p)
      }
      sj <- .central(pars, "jump")
      pp <- pars; pp["eta_ka"] <- h
      pm <- pars; pm["eta_ka"] <- -h
      fd <- (.central(pp, "fd")$central - .central(pm, "fd")$central) / (2 * h)
      .post <- sj$time >= 5
      expect_equal(sj[["rx__sens_central_BY_eta_ka__"]][.post], fd[.post],
                   tolerance = 1e-3,
                   info = paste0("plugin: ", .nm))
    }
  })

  test_that("modeled-rate infusion continuous forcing matches finite differences", {
    # rate(central) = exp(tr + eta_r) with a rate=-1 dose: the infusion rate
    # depends on eta_r.  The rate is solver-applied forcing (not a symbolic term
    # in f), so the symbolic sens ODE misses d(rate)/dp -- the jump injects it as
    # a forcing on the sensitivity compartment over the infusion.  The analytic
    # jump must match a central difference; the fd-mode sens ODE must NOT.
    .mod <- "
      ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
      rate(central) <- exp(tr + eta_r)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2, tr = 0.7, eta_r = 0)
    e <- et(amt = 100, cmt = "central", rate = -1) |> et(seq(0.25, 14, 0.25))
    .central <- function(p, mode) {
      rxSolve(rxode2(.mod, calcSens = "eta_r", eventSens = mode), e, p)
    }
    sj <- .central(pars, "jump")[["rx__sens_central_BY_eta_r__"]]
    sf <- .central(pars, "fd")[["rx__sens_central_BY_eta_r__"]]
    h <- 1e-4
    pp <- pars; pp["eta_r"] <- h
    pm <- pars; pm["eta_r"] <- -h
    fd <- (.central(pp, "fd")$central - .central(pm, "fd")$central) / (2 * h)
    expect_equal(sj, fd, tolerance = 1e-3)
    expect_gt(max(abs(sf - fd)), 1)   # fd-mode misses the forcing sensitivity
  })

  test_that("modeled-duration infusion (forcing + moving boundary) matches FD", {
    # dur(central) = exp(tr + eta_r) with a rate=-2 dose: rate = amt/dur depends
    # on eta_r AND the infusion end tau2 = tau1 + dur(eta_r) moves with eta_r.
    # The jump needs BOTH the continuous d(rate)/dp forcing and the moving-
    # boundary state jump rate*d(dur)/dp at the infusion end.  Validated vs FD;
    # the fd-mode sens ODE (neither contribution) is badly wrong.
    .mod <- "
      ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
      dur(central) <- exp(tr + eta_r)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2, tr = 0.7, eta_r = 0)
    e <- et(amt = 100, cmt = "central", rate = -2) |> et(seq(0.25, 14, 0.25))
    .central <- function(p, mode) {
      rxSolve(rxode2(.mod, calcSens = "eta_r", eventSens = mode), e, p)
    }
    sj <- .central(pars, "jump")[["rx__sens_central_BY_eta_r__"]]
    sf <- .central(pars, "fd")[["rx__sens_central_BY_eta_r__"]]
    h <- 1e-4
    pp <- pars; pp["eta_r"] <- h
    pm <- pars; pm["eta_r"] <- -h
    fd <- (.central(pp, "fd")$central - .central(pm, "fd")$central) / (2 * h)
    expect_equal(sj, fd, tolerance = 1e-2)
    expect_gt(max(abs(sf - fd)), 1)   # fd-mode misses forcing + boundary
  })

  test_that("modeled-dur infusion with estimated F and duration matches FD", {
    # dur(central)=exp(tr+eta_r) AND f(central)=expit(tf+eta_f): rate=F*amt/dur
    # depends on BOTH parameters.  The continuous forcing carries both
    # d(rate)/dp pieces ((amt/dur)*dF and -(rate/dur)*dDur) and the moving
    # boundary applies only to the duration (tau2 = tau1 + dur, independent of F).
    # Both eta sensitivities must match a central difference.
    .mod <- "
      ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
      dur(central) <- exp(tr + eta_r)
      f(central)   <- expit(tf + eta_f)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2, tr = 0.7, tf = 0.3,
              eta_r = 0, eta_f = 0)
    e <- et(amt = 100, cmt = "central", rate = -2) |> et(seq(0.25, 14, 0.25))
    .central <- function(p, mode) {
      rxSolve(rxode2(.mod, calcSens = c("eta_r", "eta_f"), eventSens = mode), e, p)
    }
    sj <- .central(pars, "jump")
    h <- 1e-4
    for (.eta in c("eta_r", "eta_f")) {
      pp <- pars; pp[.eta] <- h
      pm <- pars; pm[.eta] <- -h
      fd <- (.central(pp, "fd")$central - .central(pm, "fd")$central) / (2 * h)
      expect_equal(sj[[paste0("rx__sens_central_BY_", .eta, "__")]], fd,
                   tolerance = 1e-2, info = paste0("param: ", .eta))
    }
  })

  test_that("infusion with a modeled lag matches FD (all infusion types)", {
    # alag(central) = exp(tl + eta_l): the whole infusion window [tau1, tau2]
    # shifts by d(alag)/dp (tau1 = t0 + alag, tau2 = tau1 + dur), giving moving
    # start/stop boundary jumps [S] = -/+ rate*d(alag)/dp.  Validated for fixed-
    # rate, modeled-rate and modeled-duration infusions.  An irregular grid keeps
    # observations off the (discontinuous) boundary times so the FD is well-posed.
    .obs <- seq(0.1, 14, 0.37)
    .lagLine <- "alag(central) <- exp(tl + eta_l)"
    .ode <- "d/dt(depot) <- -ka * depot\n      d/dt(central) <- ka * depot - cl / v * central"
    .cases <- list(
      fixed = list(extra = "", dose = et(amt = 100, rate = 30, cmt = "central"),
                   p = c(tka = 0.2, tcl = 1, tv = 2, tl = log(1.1), eta_l = 0)),
      mrate = list(extra = "rate(central) <- exp(tr)",
                   dose = et(amt = 100, rate = -1, cmt = "central"),
                   p = c(tka = 0.2, tcl = 1, tv = 2, tl = log(1.1), tr = log(30), eta_l = 0)),
      mdur  = list(extra = "dur(central) <- exp(td)",
                   dose = et(amt = 100, rate = -2, cmt = "central"),
                   p = c(tka = 0.2, tcl = 1, tv = 2, tl = log(1.1), td = log(3), eta_l = 0))
    )
    h <- 1e-4
    for (.nm in names(.cases)) {
      .cs <- .cases[[.nm]]
      .mod <- paste("ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)",
                    .lagLine, .cs$extra, .ode, sep = "\n      ")
      e <- .cs$dose |> et(.obs)
      .central <- function(p, mode) {
        rxSolve(rxode2(.mod, calcSens = "eta_l", eventSens = mode), e, p)
      }
      sj <- .central(.cs$p, "jump")[["rx__sens_central_BY_eta_l__"]]
      pp <- .cs$p; pp["eta_l"] <- .cs$p["eta_l"] + h
      pm <- .cs$p; pm["eta_l"] <- .cs$p["eta_l"] - h
      fd <- (.central(pp, "fd")$central - .central(pm, "fd")$central) / (2 * h)
      expect_equal(sj, fd, tolerance = 1e-2, info = paste0("infusion: ", .nm))
    }
  })

  test_that("jump sensitivities are correct per subject (population solve)", {
    # FOCEi solves many subjects at once (in parallel).  Each subject has its
    # own etas, so its own lagged dose time and bioavailability; the jump
    # injection must use per-subject state/dose.  Three subjects with distinct
    # etas must each match a per-subject central difference.
    .mod <- "
      ka <- exp(tka + eta_ka); cl <- exp(tcl); v <- exp(tv)
      alag(depot) <- exp(tlag + eta_lag)
      f(depot)    <- expit(tf + eta_f)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    th <- c(tka = 0.2, tcl = 1, tv = 2, tf = 0.3, tlag = log(1.1))
    ev <- data.frame(eta_ka = c(0, 0.3, -0.2),
                     eta_lag = c(0, 0.1, -0.15),
                     eta_f = c(0, -0.2, 0.25))
    e <- et(amt = 100, cmt = "depot", id = 1:3) |> et(seq(0.13, 12, 0.41), id = 1:3)
    mj <- rxode2(.mod, calcSens = c("eta_lag", "eta_f"), eventSens = "jump")
    mfd <- rxode2(.mod, calcSens = c("eta_lag", "eta_f"), eventSens = "fd")
    sj <- rxSolve(mj, e, th, iCov = ev)
    h <- 1e-4
    for (.sp in c("eta_lag", "eta_f")) {
      evp <- ev; evp[[.sp]] <- evp[[.sp]] + h
      evm <- ev; evm[[.sp]] <- evm[[.sp]] - h
      fd <- (rxSolve(mfd, e, th, iCov = evp)$central -
             rxSolve(mfd, e, th, iCov = evm)$central) / (2 * h)
      expect_equal(sj[[paste0("rx__sens_central_BY_", .sp, "__")]], fd,
                   tolerance = 1e-2, info = paste0("param: ", .sp))
    }
  })

  test_that(".rxEventSensMap exposes a second-order (Hessian) compartment map", {
    .mod <- "
      ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
      f(depot) <- expit(tf + eta_f)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    m <- rxode2(.mod, calcSens = "eta_f", calcSens2 = "eta_f")
    im <- .rxEventSensMap(m)
    expect_false(is.null(im$map2))
    # one (state, p, q) row per physical state for the single (eta_f, eta_f) pair
    expect_setequal(im$map2$state, c("depot", "central"))
    expect_true(all(im$map2$p == "eta_f" & im$map2$q == "eta_f"))
    # the depot 2nd-order compartment exists and comes after the 1st-order ones
    .row <- im$map2[im$map2$state == "depot", ]
    expect_true(.row$sensCmt > nrow(im$map))
  })

  test_that(".rxEventSensCLines indexes multi-parameter calcSens2 consistently with calcSens", {
    # Regression: .rxEventSensMap() used to sort `map2` alphabetically by
    # (p, q, stateCmt); .rxEventSensCLines()'s .qIdx (built from
    # unique(map2$q)) then inherited that alphabetical order instead of
    # calcSens2's *as-passed* order (matching the compiled compartment
    # layout rxExpandSens2_ actually used) -- for any calcSens2 with more
    # than one parameter where the as-passed order isn't alphabetical, the
    # 2nd-order derivative values were written into the WRONG compartment
    # (found while validating the 2nd-order infusion jump, 2026-06-30;
    # masked in every earlier test, which all used a single-parameter
    # calcSens2 where reordering is a no-op). trate/tlag deliberately spelled
    # so alphabetical order ("tlag","trate") differs from the as-passed order
    # ("trate","tlag").
    .mod <- "
      ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
      alag(depot) <- exp(tlag)
      rate(depot) <- exp(trate)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    m <- rxode2(.mod, calcSens = c("trate", "tlag"), calcSens2 = c("trate", "tlag"),
                eventSens = "jump")
    info <- m$eventSensInfo
    expect_equal(names(info$map$sensParams), NULL) # sanity: sensParams unnamed char vec
    expect_equal(info$map$sensParams, c("trate", "tlag"))
    # .qIdx must be built in the SAME (as-passed) order as .pIdx/sensParams,
    # not alphabetically -- this is the actual bug: previously "tlag" (unique(map2$q)'s
    # alphabetically-first value) would come first here instead of "trate".
    expect_equal(unique(info$map$map2$q), c("trate", "tlag"))
    expect_equal(unique(info$map$map2$p), c("trate", "tlag"))
  })

  test_that(".rxEventSensD2Expr is the correct second total derivative", {
    # d2(F)/d(eta_f)^2 for F = expit(tf + eta_f) must equal expit''(tf+eta_f)
    # = p(1-p)(1-2p).  State-independent F -> only the direct second partial.
    m <- rxode2("ka<-exp(tka)\nf(depot)<-expit(tf+eta_f)\nd/dt(depot)<- -ka*depot\nd/dt(central)<-ka*depot-central",
                calcSens = "eta_f", calcSens2 = "eta_f")
    mdl <- .rxLoadPrune(m)
    e2 <- .rxEventSensD2Expr(get("rx_f_depot_", envir = mdl), "eta_f", "eta_f",
                             .rxEventSensMap(m)$states)
    expect_true(nzchar(e2) && e2 != "0")
    val <- eval(parse(text = e2),
                list(eta_f = 0, tf = 0.3, Rx_pow_di = function(a, b) a^b,
                     expit = function(x) 1 / (1 + exp(-x))))
    p <- 1 / (1 + exp(-0.3))
    expect_equal(val, p * (1 - p) * (1 - 2 * p), tolerance = 1e-8)
  })

  test_that("calcSens2 generates correct second-order sensitivity ODEs", {
    # rxode2(..., calcSens2=) emits the 2nd-order sensitivity compartments
    # rx__sens_<state>_BY_<p>_BY_<q>__ (the Hessian path).  For a parameter that
    # does not touch dosing there is no jump, so the 2nd-order sensitivity comes
    # purely from the continuous variational ODEs.  S^{pq} = d(S^p)/dq, so it must
    # match a (precise) first central difference of the first-order sensitivity.
    .mod <- "
      ka <- exp(tka + eta_ka); cl <- exp(tcl); v <- exp(tv)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2, eta_ka = 0)
    e <- et(amt = 100, cmt = "depot") |> et(seq(0.5, 12, 0.5))
    m1 <- rxode2(.mod, calcSens = "eta_ka")
    m2 <- rxode2(.mod, calcSens = "eta_ka", calcSens2 = "eta_ka")
    expect_true("rx__sens_central_BY_eta_ka_BY_eta_ka__" %in% rxModelVars(m2)$state)
    s2 <- rxSolve(m2, e, pars)[["rx__sens_central_BY_eta_ka_BY_eta_ka__"]]
    h <- 1e-5
    .s1 <- function(p) rxSolve(m1, e, p)[["rx__sens_central_BY_eta_ka__"]]
    pp <- pars; pp["eta_ka"] <- h
    pm <- pars; pm["eta_ka"] <- -h
    fd <- (.s1(pp) - .s1(pm)) / (2 * h)
    expect_equal(s2, fd, tolerance = 1e-2)
  })

  test_that("calcSens3 generates correct third-order sensitivity ODEs (Phase H0)", {
    # rxode2(..., calcSens2=, calcSens3=) emits the 3rd-order sensitivity
    # compartments rx__sens_<state>_BY_<p>_BY_<q>_BY_<r>__ via
    # rxExpandSens3_() (PR #1092), exposed at the rxode2() build level for
    # the first time here (previously only reachable through the DDE-
    # specific delay() sensitivity path). No jump-sensitivity content is
    # involved -- this is pure continuous-ODE 3rd-order sensitivity, so it
    # is validated standalone against a nested finite difference of the
    # analytic 2nd-order sensitivity (the plan's H0 acceptance criterion),
    # with calcSens2 == calcSens3 == calcSens (the common "full Hessian/
    # third-order" case, matching how calcSens2 is used everywhere else).
    .mod <- "
      ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2)
    e <- et(amt = 100, cmt = "depot") |> et(seq(0.5, 12, 1))
    m2 <- rxode2(.mod, calcSens = "tka", calcSens2 = "tka")
    m3 <- rxode2(.mod, calcSens = "tka", calcSens2 = "tka", calcSens3 = "tka")
    expect_true("rx__sens_central_BY_tka_BY_tka_BY_tka__" %in% rxModelVars(m3)$state)
    s3 <- rxSolve(m3, e, pars)[["rx__sens_central_BY_tka_BY_tka_BY_tka__"]]
    h <- 1e-4
    .s2 <- function(p) rxSolve(m2, e, p)[["rx__sens_central_BY_tka_BY_tka__"]]
    pp <- pars; pp["tka"] <- pars["tka"] + h
    pm <- pars; pm["tka"] <- pars["tka"] - h
    fd <- (.s2(pp) - .s2(pm)) / (2 * h)
    expect_equal(s3, fd, tolerance = 1e-3)

    # calcSens3 requires calcSens2
    expect_error(rxode2(.mod, calcSens = "tka", calcSens3 = "tka"))
  })

  test_that("second-order additive-bolus F jump matches finite differences", {
    # F = expit(tf + eta_f) feeds the dose only (ka/cl/v independent of eta_f),
    # so the second-order sensitivity S^{ff} of the dosed compartment is *entirely*
    # the explicit jump amt*d2F/d eta_f^2 -- without it S^{ff} is exactly 0.  The
    # jump-mode 2nd-order sensitivity must match d(S^f)/d eta_f (a precise first
    # central difference of the jump-correct first-order sensitivity).
    .mod <- "
      ka <- exp(tka); cl <- exp(tcl); v <- exp(tv)
      alag(depot) <- 1
      f(depot)    <- expit(tf + eta_f)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(tka = 0.2, tcl = 1, tv = 2, tf = 0.3, eta_f = 0)
    e <- et(amt = 100, cmt = "depot") |> et(seq(0.5, 12, 0.5))
    m1 <- rxode2(.mod, calcSens = "eta_f", eventSens = "jump")
    m2 <- rxode2(.mod, calcSens = "eta_f", calcSens2 = "eta_f", eventSens = "jump")
    s2 <- rxSolve(m2, e, pars)[["rx__sens_central_BY_eta_f_BY_eta_f__"]]
    h <- 1e-5
    .s1 <- function(p) rxSolve(m1, e, p)[["rx__sens_central_BY_eta_f__"]]
    pp <- pars; pp["eta_f"] <- h
    pm <- pars; pm["eta_f"] <- -h
    fd <- (.s1(pp) - .s1(pm)) / (2 * h)
    expect_equal(s2, fd, tolerance = 1e-2)
    # without the 2nd-order jump the contribution would be entirely missing
    expect_gt(max(abs(fd)), 1)
  })

  test_that("third-order additive-bolus F jump matches nested finite differences (Phase H1)", {
    # F = expit(tf) feeds the dose only, so the third-order sensitivity
    # S^{fff} of the dosed compartment is entirely the explicit jump
    # amt*d3F/dtf^3 -- validate against a central difference of the
    # jump-correct *second*-order sensitivity (mirrors the 2nd-order F-jump
    # test's own validation strategy one level deeper).
    ode_code_f <- "
      f(depot)    <- expit(tf)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars_f <- c(ka = 0.5, cl = 0.2, v = 10, tf = qlogis(0.7))
    e <- et(amt = 100, cmt = "depot") |> et(seq(0, 10, by = 0.5))
    m2 <- rxode2(ode_code_f, calcSens = c("tf", "ka"), calcSens2 = c("tf", "ka"),
                 eventSens = "jump")
    m3 <- rxode2(ode_code_f, calcSens = c("tf", "ka"), calcSens2 = c("tf", "ka"),
                 calcSens3 = "tf", eventSens = "jump")
    expect_true("rx__sens_central_BY_tf_BY_tf_BY_tf__" %in% rxModelVars(m3)$state)
    s3 <- rxSolve(m3, e, pars_f)[["rx__sens_central_BY_tf_BY_tf_BY_tf__"]]
    eps <- 1e-4
    p1 <- pars_f; p1["tf"] <- pars_f["tf"] + eps
    p2 <- pars_f; p2["tf"] <- pars_f["tf"] - eps
    .s2 <- function(p) rxSolve(m2, e, p)[["rx__sens_central_BY_tf_BY_tf__"]]
    fd3 <- (.s2(p1) - .s2(p2)) / (2 * eps)
    expect_equal(s3, fd3, tolerance = 1e-4)
    # without the 3rd-order jump the contribution would be entirely missing
    expect_gt(max(abs(fd3)), 1)
  })

  test_that("third-order additive-bolus F jump: replace/multiply scale/zero the 3rd-order compartment", {
    # Regression mirroring the 2nd-order replace/multiply fix: a raw
    # event-table replace() zeros every sensitivity order of the replaced
    # state (constant replacement value); a multiply() scales every order by
    # the same alpha.  Both validated against nested FD of the jump-correct
    # 2nd-order sensitivity.
    ode_code_f <- "
      f(depot)    <- expit(tf)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars_f <- c(ka = 0.5, cl = 0.2, v = 10, tf = qlogis(0.7))
    m2 <- rxode2(ode_code_f, calcSens = c("tf", "ka"), calcSens2 = c("tf", "ka"),
                 eventSens = "jump")
    m3 <- rxode2(ode_code_f, calcSens = c("tf", "ka"), calcSens2 = c("tf", "ka"),
                 calcSens3 = "tf", eventSens = "jump")
    for (.evid in c(5, 6)) {
      .amt <- if (.evid == 5) 50 else 0.5
      e <- et(amt = 100, cmt = "depot") |>
        et(time = 5, amt = .amt, cmt = "central", evid = .evid) |>
        et(seq(0.5, 12, 1))
      s3 <- rxSolve(m3, e, pars_f)[["rx__sens_central_BY_tf_BY_tf_BY_tf__"]]
      eps <- 1e-4
      p1 <- pars_f; p1["tf"] <- pars_f["tf"] + eps
      p2 <- pars_f; p2["tf"] <- pars_f["tf"] - eps
      .s2 <- function(p) rxSolve(m2, e, p)[["rx__sens_central_BY_tf_BY_tf__"]]
      fd3 <- (.s2(p1) - .s2(p2)) / (2 * eps)
      expect_equal(s3, fd3, tolerance = 1e-4, info = paste0("evid: ", .evid))
    }
  })

  test_that("second-order dtau/lag row: q unrelated to the modeled alag (Phase H1-dtau)", {
    # tlag drives alag; tf drives F only (independent of tlag).  Differentiating
    # the 1st-order dtau jump (-J[k][c]*delta*dLag_p[c]) by the product rule
    # wrt q=tf needs three terms: d(J)/dq (here 0, linear system), d(delta)/dq
    # = amt*dFQ[c][q] (nonzero, F depends on tf), and d2Lag[p][q] (here 0,
    # alag doesn't depend on tf).  Only the middle term survives, and it is
    # exactly what makes S^{tlag,tf} nonzero (without it, entirely missing).
    ode_code <- "
      alag(depot) <- exp(tlag)
      f(depot)    <- expit(tf)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl / v * central
    "
    pars <- c(ka = 0.5, cl = 0.2, v = 10, tlag = log(0.5), tf = qlogis(0.7))
    # avoid the observation landing exactly on the (parameter-dependent) lag
    # time (t=0.5): a known artifact source, see the plan's validation notes.
    e <- et(amt = 100, cmt = "depot") |> et(seq(0.55, 10, by = 0.5))
    m1 <- rxode2(ode_code, calcSens = c("tlag", "tf"), eventSens = "jump")
    m2 <- rxode2(ode_code, calcSens = c("tlag", "tf"), calcSens2 = c("tlag", "tf"),
                 eventSens = "jump")
    s2 <- rxSolve(m2, e, pars, atol = 1e-12, rtol = 1e-12)[["rx__sens_central_BY_tlag_BY_tf__"]]
    eps <- 1e-5
    p1 <- pars; p1["tf"] <- pars["tf"] + eps
    p2 <- pars; p2["tf"] <- pars["tf"] - eps
    .s1 <- function(p) rxSolve(m1, e, p, atol = 1e-12, rtol = 1e-12)[["rx__sens_central_BY_tlag__"]]
    fd <- (.s1(p1) - .s1(p2)) / (2 * eps)
    expect_equal(s2, fd, tolerance = 1e-6)
    expect_gt(max(abs(fd)), 1)
  })

  test_that("second-order dtau/lag row: d(J[k][c])/dq Jacobian-coupling term (nonlinear model)", {
    # Michaelis-Menten elimination makes the physical Jacobian state- and
    # parameter-dependent, exercising the d(J)/dq term (zero for the linear
    # models above) in isolation: J[central][central] = -Vm*Km/(Km+central)^2
    # depends explicitly on Vm, so d(J)/dtvm != 0.
    ode_code <- "
      alag(central) <- exp(tlag)
      vm <- exp(tvm)
      km <- exp(tkm)
      d/dt(central) = -vm*central/(km+central)
    "
    pars <- c(tvm = log(2), tkm = log(5), tlag = log(0.5))
    e <- et(amt = 10, cmt = "central") |> et(seq(0.55, 8, by = 0.5))
    m1 <- rxode2(ode_code, calcSens = c("tlag", "tvm"), eventSens = "jump")
    m2 <- rxode2(ode_code, calcSens = c("tlag", "tvm"), calcSens2 = c("tlag", "tvm"),
                 eventSens = "jump")
    info <- m2$eventSensInfo
    expect_true(nrow(info$derivs$lagJacQ) > 0L)
    s2 <- rxSolve(m2, e, pars, atol = 1e-12, rtol = 1e-12)[["rx__sens_central_BY_tlag_BY_tvm__"]]
    eps <- 1e-5
    p1 <- pars; p1["tvm"] <- pars["tvm"] + eps
    p2 <- pars; p2["tvm"] <- pars["tvm"] - eps
    .s1 <- function(p) rxSolve(m1, e, p, atol = 1e-12, rtol = 1e-12)[["rx__sens_central_BY_tlag__"]]
    fd <- (.s1(p1) - .s1(p2)) / (2 * eps)
    expect_equal(s2, fd, tolerance = 1e-6)
    expect_gt(max(abs(fd)), 0.1)
  })

  test_that("second-order dtau/lag row: Leibniz/moving-boundary term when q ALSO drives the same alag (Phase H1-dtau)", {
    # Differentiating the 1st-order dtau row by the product rule alone is
    # PROVEN INCOMPLETE (by FD) whenever q ALSO shifts the same event's time
    # (e.g. the (tlag,tlag) diagonal, or two lag parameters combined
    # additively in one alag() expression): it misses a Leibniz/moving-
    # boundary term `-[dS^p_k/dt|post - dS^p_k/dt|pre] * dLag_q[c]`, the same
    # class of problem that originally blocked the infusion 2nd-order
    # attempt. `dS^p_k/dt` is read directly from the compiled dydt() output
    # (pre-jump state vs. a scratch post-jump copy), needing no separate
    # Jacobian-matrix construction. Validated against FD for both the
    # diagonal and a two-lag-parameter model.
    ode_code_diag <- "
      alag(depot) <- exp(tlag)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl / v * central
    "
    pars_diag <- c(ka = 0.5, cl = 0.2, v = 10, tlag = log(0.5))
    e_diag <- et(amt = 100, cmt = "depot") |> et(seq(0.55, 10, by = 0.5))
    m1d <- rxode2(ode_code_diag, calcSens = "tlag", eventSens = "jump")
    m2d <- rxode2(ode_code_diag, calcSens = "tlag", calcSens2 = "tlag", eventSens = "jump")
    s2d <- rxSolve(m2d, e_diag, pars_diag, atol = 1e-12, rtol = 1e-12)[["rx__sens_central_BY_tlag_BY_tlag__"]]
    eps <- 1e-5
    p1 <- pars_diag; p1["tlag"] <- pars_diag["tlag"] + eps
    p2 <- pars_diag; p2["tlag"] <- pars_diag["tlag"] - eps
    .s1d <- function(p) rxSolve(m1d, e_diag, p, atol = 1e-12, rtol = 1e-12)[["rx__sens_central_BY_tlag__"]]
    fdd <- (.s1d(p1) - .s1d(p2)) / (2 * eps)
    expect_equal(s2d, fdd, tolerance = 1e-6)
    expect_gt(max(abs(fdd)), 1)

    ode_code_2p <- "
      alag(depot) <- exp(tlag1 + tlag2)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl / v * central
    "
    pars_2p <- c(ka = 0.5, cl = 0.2, v = 10, tlag1 = log(0.5) / 2, tlag2 = log(0.5) / 2)
    e_2p <- et(amt = 100, cmt = "depot") |> et(seq(0.55, 10, by = 0.25))
    m1p <- rxode2(ode_code_2p, calcSens = c("tlag1", "tlag2"), eventSens = "jump")
    m2p <- rxode2(ode_code_2p, calcSens = c("tlag1", "tlag2"), calcSens2 = c("tlag1", "tlag2"),
                  eventSens = "jump")
    s2p <- rxSolve(m2p, e_2p, pars_2p, atol = 1e-12, rtol = 1e-12)[["rx__sens_central_BY_tlag1_BY_tlag2__"]]
    p1p <- pars_2p; p1p["tlag2"] <- pars_2p["tlag2"] + eps
    p2p <- pars_2p; p2p["tlag2"] <- pars_2p["tlag2"] - eps
    .s1p <- function(p) rxSolve(m1p, e_2p, p, atol = 1e-12, rtol = 1e-12)[["rx__sens_central_BY_tlag1__"]]
    fdp <- (.s1p(p1p) - .s1p(p2p)) / (2 * eps)
    expect_equal(s2p, fdp, tolerance = 1e-6)
    expect_gt(max(abs(fdp)), 1)
  })

  test_that("second-order dtau/lag row: mirror compartment symmetry (Phase H1-dtau)", {
    # S^{p,q} and S^{q,p} must be identical (Schwarz).  Before the Leibniz
    # term was added, S^{tlag,tf} validated correctly (tlag drives alag, so
    # the outer p-loop ran) while S^{tf,tlag} stayed silently at 0 (tf does
    # not drive alag, so the outer p-loop skipped it entirely and never had
    # a chance to apply the Leibniz correction to S^tf's own post-ddelta-jump
    # value). The p-loop now also runs whenever ANY calcSens2 parameter
    # drives this compartment's alag, regardless of the current p's own
    # dLag_p, fixing this.
    ode_code <- "
      alag(depot) <- exp(tlag)
      f(depot)    <- expit(tf)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl / v * central
    "
    pars <- c(ka = 0.5, cl = 0.2, v = 10, tlag = log(0.5), tf = qlogis(0.7))
    e <- et(amt = 100, cmt = "depot") |> et(seq(0.55, 10, by = 0.5))
    m2 <- rxode2(ode_code, calcSens = c("tlag", "tf"), calcSens2 = c("tlag", "tf"),
                 eventSens = "jump")
    s2 <- rxSolve(m2, e, pars, atol = 1e-12, rtol = 1e-12)
    expect_equal(s2$rx__sens_central_BY_tlag_BY_tf__, s2$rx__sens_central_BY_tf_BY_tlag__,
                 tolerance = 1e-8)
    expect_gt(max(abs(s2$rx__sens_central_BY_tlag_BY_tf__)), 1)
  })

  test_that("second-order infusion-boundary jump (fixed-rate/duration infusion + modeled alag)", {
    # Re-derives the same "jump condition one level up" trick used for the
    # additive-bolus dtau row, applied to a forcing DISCONTINUITY (not a
    # state jump): a fixed-rate/duration infusion whose start/stop time is
    # shifted by a modeled alag jumps InfusionRate[cmt] by a fixed amount at
    # the lag-shifted boundary. The naive product-rule-only term
    # (tmp*d2Lag[p][q]) matched FD DURING the infusion window and diverged
    # right after it ended (documented in the plan as "found wrong, reverted"
    # in an earlier session) -- fixed here by adding the missing Leibniz
    # term, validated through BOTH the start and stop boundaries.
    ode_code <- "
      alag(central) <- exp(tlag)
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars <- c(ka = 0.5, cl = 0.2, v = 10, tlag = log(0.5))
    e <- et(amt = 100, cmt = "central", rate = 20) |> et(seq(0.55, 15, by = 0.5))
    m1 <- rxode2(ode_code, calcSens = "tlag", eventSens = "jump")
    m2 <- rxode2(ode_code, calcSens = "tlag", calcSens2 = "tlag", eventSens = "jump")
    s2 <- rxSolve(m2, e, pars, atol = 1e-12, rtol = 1e-12)[["rx__sens_central_BY_tlag_BY_tlag__"]]
    eps <- 1e-5
    p1 <- pars; p1["tlag"] <- pars["tlag"] + eps
    p2 <- pars; p2["tlag"] <- pars["tlag"] - eps
    .s1 <- function(p) rxSolve(m1, e, p, atol = 1e-12, rtol = 1e-12)[["rx__sens_central_BY_tlag__"]]
    fd <- (.s1(p1) - .s1(p2)) / (2 * eps)
    expect_equal(s2, fd, tolerance = 1e-6)
    # crosses the stop boundary at tlag+amt/rate=5.5; matched only DURING the
    # window before the Leibniz term was added, so require post-window rows
    # to be present and nonzero (not just the pre-window ones).
    expect_gt(sum(e$time[e$evid == 0] > 6), 5)
    expect_gt(max(abs(fd)), 1)
  })

  test_that("second-order infusion-boundary jump propagates to a COUPLED (non-infused) compartment", {
    # The dosed/lag-shifted compartment's OWN sensitivity jump [S^p]_cmt is
    # the only *value* discontinuity at 1st order (other states only pick up
    # a kink -- their derivative, not their value, jumps, since it's a
    # forcing discontinuity, not a state jump like the additive-bolus case).
    # But at 2nd order, differentiating that kink's own (parameter-dependent)
    # strength wrt a SECOND parameter that ALSO shifts the boundary time
    # reintroduces a genuine Leibniz value-jump for the coupled compartment
    # too -- missed in an earlier version of this fix that only updated the
    # k=cmt (dosed) compartment, because every prior validation happened to
    # observe the SAME compartment being infused (an accidental blind spot).
    # This model infuses "depot" (coupled to "central" via ka) and checks
    # central's OWN 2nd-order sensitivity, exercising exactly that path.
    ode_code <- "
      alag(depot) <- tlag
      f(depot)    <- doseAmt
      dur(depot)  <- tinf
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars <- c(ka = 1, cl = 6, v = 60, tlag = 10, doseAmt = 200, tinf = 10)
    e <- et(amt = 1, cmt = "depot", rate = -2) |> et(seq(0.53, 60, by = 0.5))
    m1 <- rxode2(ode_code, calcSens = c("tlag", "doseAmt", "tinf"), eventSens = "jump")
    m2 <- rxode2(ode_code, calcSens = c("tlag", "doseAmt", "tinf"),
                 calcSens2 = c("tlag", "doseAmt", "tinf"), eventSens = "jump")
    s2 <- rxSolve(m2, e, pars, atol = 1e-11, rtol = 1e-11)[["rx__sens_central_BY_tlag_BY_tlag__"]]
    eps <- 1e-4
    p1 <- pars; p1["tlag"] <- pars["tlag"] + eps
    p2 <- pars; p2["tlag"] <- pars["tlag"] - eps
    .s1 <- function(p) rxSolve(m1, e, p, atol = 1e-11, rtol = 1e-11)[["rx__sens_central_BY_tlag__"]]
    fd <- (.s1(p1) - .s1(p2)) / (2 * eps)
    expect_equal(s2, fd, tolerance = 1e-6)
    expect_gt(max(abs(fd)), 1)
  })

  test_that("additive-bolus F jump fires for indexed THETA[n]/ETA[n] params", {
    # The nlmixr2 FOCEi inner model writes sensitivities wrt the indexed
    # parameters ETA[n]/THETA[n] (compartments rx__sens_<state>_BY_ETA_n___).
    # rxFromSE renders these as `ETA[3]`, whose all.vars() collapses to "ETA" and
    # never matched the SE-mangled sensParam "ETA_3_" -- the dosing-parameter
    # derivatives were silently dropped and no jump was injected.  Guard both the
    # symbolic detection (free_symbols) and the codegen rewrite (ETA[n]->_ETA_n_).
    code <- paste(
      "param(THETA[1],THETA[2],THETA[3],THETA[4],ETA[1],ETA[3]);",
      "cmt(depot);",
      "cmt(central);",
      "f(depot)=1/(1+exp(-(ETA[3]+THETA[4])));",
      "d/dt(depot)=-exp(ETA[1]+THETA[1])*depot;",
      "d/dt(central)=exp(ETA[1]+THETA[1])*depot-exp(THETA[2]-THETA[3])*central;",
      "d/dt(rx__sens_depot_BY_ETA_3___)=-exp(ETA[1]+THETA[1])*rx__sens_depot_BY_ETA_3___;",
      "d/dt(rx__sens_central_BY_ETA_3___)=exp(ETA[1]+THETA[1])*rx__sens_depot_BY_ETA_3___-exp(THETA[2]-THETA[3])*rx__sens_central_BY_ETA_3___;",
      sep = "\n")
    mj <- rxode2(code, eventSens = "jump")
    # the F derivative wrt ETA[3] must be detected (non-empty derivs table)
    .df <- mj$eventSensInfo$derivs$f
    expect_true(nrow(.df) >= 1L)
    expect_true("ETA_3_" %in% .df$param)

    p <- c("THETA[1]" = 0.45, "THETA[2]" = 1, "THETA[3]" = 3.45,
           "THETA[4]" = 0.9, "ETA[1]" = 0, "ETA[3]" = 0.2)
    # sample immediately after the dose to capture the jump peak before decay
    e <- et(amt = 100, cmt = "depot") |> et(c(1e-4, seq(0.13, 24, length.out = 30)))
    s <- rxSolve(mj, p, e, atol = 1e-10, rtol = 1e-10)
    # the additive-bolus F jump makes the depot sensitivity wrt ETA[3] non-zero;
    # analytic value at t=0+ is amt*F*(1-F) with F = expit(0.2 + 0.9).
    .F <- 1 / (1 + exp(-(0.2 + 0.9)))
    expect_equal(max(s[["rx__sens_depot_BY_ETA_3___"]]), 100 * .F * (1 - .F),
                 tolerance = 1e-2)
  })

  test_that("eventSens mode is folded into the model cache key", {
    mfd <- rxode2(.modConstF, calcSens = c("eta_ka", "eta_lag"),
                  eventSens = "fd")
    mj <- rxode2(.modConstF, calcSens = c("eta_ka", "eta_lag"),
                 eventSens = "jump")
    # same model text, different mode -> distinct parsed md5 (distinct DLLs)
    expect_false(identical(
      unname(rxModelVars(mfd)$md5["parsed_md5"]),
      unname(rxModelVars(mj)$md5["parsed_md5"])
    ))
    # the cache key is reset after each build (no leak to later non-jump builds)
    expect_identical(.rxEventSensCacheKey, "")
  })

  test_that("reproduces the paper's PK/PD IMAX infusion example (dEffect/dtlag, ddose, dtinf)", {
    # Kaschek & Fidler, "Forward Sensitivity Equations in the Presence of
    # Events" (~/src/EventSensitivities/paper.tex, Sec. Example): a depot-
    # absorption model with a zero-order infusion into the depot compartment,
    # parameterized directly by lag time (tlag), infusion duration (tinf), and
    # dose (not etas) -- alag(depot)=tlag, dur(depot)=tinf, f(depot)=doseAmt
    # (bioavailability standing in for the dose, routed through the
    # duration-fixed F/amt convention: rate = F*amt/dur = doseAmt*1/tinf,
    # matching the paper's r1 = dose/tinf exactly). Effect is a downstream LHS
    # of central, so dEffect/dp = dEffect/dcentral * dcentral/dp (closed-form
    # chain rule below) using the jump-computed rx__sens_central_BY_p__.
    .mod <- "
      ka <- exp(lka)
      cl <- exp(lcl)
      v  <- exp(lv)
      imax <- exp(limax)
      ic50 <- exp(lic50)
      e0 <- exp(le0)
      alag(depot) <- tlag
      dur(depot)  <- tinf
      f(depot)    <- doseAmt
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
      effect <- e0 * (1 - (central / v) * imax / (ic50 + (central / v)))
    "
    pars <- c(
      lka = log(1), lcl = log(6), lv = log(60), limax = log(1), lic50 = log(1),
      le0 = log(15), tlag = 10, doseAmt = 200, tinf = 10
    )
    mj <- rxode2(.mod, calcSens = c("tlag", "doseAmt", "tinf"), eventSens = "jump")
    e <- et(amt = 1, cmt = "depot", rate = -2) |> et(seq(0, 60, by = 0.5))
    sj <- rxSolve(mj, e, pars)

    dEffect_dCentral <- with(sj, -e0 * imax * ic50 / (v * (ic50 + central / v)^2))
    d_tlag <- dEffect_dCentral * sj$rx__sens_central_BY_tlag__
    d_tinf <- dEffect_dCentral * sj$rx__sens_central_BY_tinf__
    d_dose <- dEffect_dCentral * sj$rx__sens_central_BY_doseAmt__

    # (1) matches a true finite difference of the solved "effect" output
    # (the plan's primary correctness gate, Section 5) for all three params.
    .h <- 1e-4
    .fd <- function(pname) {
      .pp <- pars; .pp[pname] <- pars[pname] + .h
      .pm <- pars; .pm[pname] <- pars[pname] - .h
      (rxSolve(mj, e, .pp)$effect - rxSolve(mj, e, .pm)$effect) / (2 * .h)
    }
    expect_equal(d_tlag, .fd("tlag"), tolerance = 1e-3)
    expect_equal(d_tinf, .fd("tinf"), tolerance = 1e-3)
    expect_equal(d_dose, .fd("doseAmt"), tolerance = 1e-3)

    # (2) qualitative claims from the paper (Sec. Example):
    # "higher doses lead to stronger inhibition" -- Effect decreases (relative
    # to E0) as dose increases, i.e. dEffect/ddose <= 0 everywhere.
    expect_true(all(d_dose <= 1e-8))
    # "a longer infusion time will decrease the inhibition during the
    # infusion and increase it afterwards. The same holds for [tlag]" --
    # Effect increases (dEffect/dp > 0) during [tlag, tlag+tinf] and
    # decreases (dEffect/dp < 0) after tlag+tinf, for both tlag and tinf.
    .during <- sj$time > pars["tlag"] + 0.5 & sj$time < pars["tlag"] + pars["tinf"] - 0.5
    .after <- sj$time > pars["tlag"] + pars["tinf"] + 0.5 & sj$time < 40
    expect_true(all(d_tinf[.during] > 0))
    expect_true(all(d_tinf[.after] < 0))
    expect_true(all(d_tlag[.during] > 0))
    expect_true(all(d_tlag[.after] < 0))
    # "the impact of the lag time is five times higher than the impact of the
    # infusion time" -- loose bound around the paper's approximate reading.
    .ratio <- max(abs(d_tlag)) / max(abs(d_tinf))
    expect_true(.ratio > 3 && .ratio < 10)
  })

  test_that("reproduces the paper's PK/PD IMAX infusion example at 2nd order (S^{tlag,tlag})", {
    # Same model as the 1st-order reproduction above, extended to calcSens2:
    # exercises the infusion-boundary Leibniz term together with the
    # additive-bolus dtau row's own Leibniz term in the paper's actual
    # headline scenario (depot infused, central observed, alag+F+dur all
    # modeled and all parameter-dependent). Offset time grid avoids landing
    # an observation exactly on the tlag/tlag+tinf boundaries (a known FD
    # artifact source, not a model error).
    .mod <- "
      ka <- exp(lka)
      cl <- exp(lcl)
      v  <- exp(lv)
      alag(depot) <- tlag
      dur(depot)  <- tinf
      f(depot)    <- doseAmt
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(lka = log(1), lcl = log(6), lv = log(60), tlag = 10, doseAmt = 200, tinf = 10)
    m1 <- rxode2(.mod, calcSens = c("tlag", "doseAmt", "tinf"), eventSens = "jump")
    m2 <- rxode2(.mod, calcSens = c("tlag", "doseAmt", "tinf"),
                 calcSens2 = c("tlag", "doseAmt", "tinf"), eventSens = "jump")
    e <- et(amt = 1, cmt = "depot", rate = -2) |> et(seq(0.53, 60, by = 0.5))
    s2 <- rxSolve(m2, e, pars, atol = 1e-11, rtol = 1e-11)[["rx__sens_central_BY_tlag_BY_tlag__"]]
    eps <- 1e-4
    p1 <- pars; p1["tlag"] <- pars["tlag"] + eps
    p2 <- pars; p2["tlag"] <- pars["tlag"] - eps
    .s1 <- function(p) rxSolve(m1, e, p, atol = 1e-11, rtol = 1e-11)[["rx__sens_central_BY_tlag__"]]
    fd <- (.s1(p1) - .s1(p2)) / (2 * eps)
    expect_equal(s2, fd, tolerance = 1e-6)
    expect_gt(max(abs(fd)), 1)
  })

  test_that("second-order modeled-DUR boundary: combined alag+dur cross term S^{tlag,tinf}", {
    # alag(depot)+dur(depot), NO F: isolates the MODEL_DUR_ON/OFF boundary's
    # combined derivation (tau2 = tau1(alag) + dur(tinf) shifts with BOTH
    # mechanisms at once) from the additive-bolus dtau row's own Leibniz
    # term. An earlier implementation handled alag and dur as two SEPARATE
    # per-mechanism terms and got this cross term wrong (not just missing) --
    # confirmed by FD, ~8% relative error -- because dtau2/dp and dtau2/dq
    # must be SUMMED (dLag_p+dDur_p, dLagQ_q+dDurQ_q) before use, not treated
    # independently. Checks BOTH the infused (depot) and downstream
    # (central) compartments through and after the boundary.
    .mod <- "
      alag(depot) <- tlag
      dur(depot)  <- tinf
      d/dt(depot)   = -ka * depot
      d/dt(central) =  ka * depot - cl/v * central
    "
    pars <- c(ka = 1, cl = 6, v = 60, tlag = 10, tinf = 10)
    m1 <- rxode2(.mod, calcSens = c("tlag", "tinf"), eventSens = "jump")
    m2 <- rxode2(.mod, calcSens = c("tlag", "tinf"), calcSens2 = c("tlag", "tinf"),
                 eventSens = "jump")
    e <- et(amt = 1, cmt = "depot", rate = -2) |> et(seq(0.53, 60, by = 0.5))
    r2 <- rxSolve(m2, e, pars, atol = 1e-11, rtol = 1e-11)
    eps <- 1e-4
    p1 <- pars; p1["tinf"] <- pars["tinf"] + eps
    p2 <- pars; p2["tinf"] <- pars["tinf"] - eps
    r1a <- rxSolve(m1, e, p1, atol = 1e-11, rtol = 1e-11)
    r1b <- rxSolve(m1, e, p2, atol = 1e-11, rtol = 1e-11)
    fd_central <- (r1a$rx__sens_central_BY_tlag__ - r1b$rx__sens_central_BY_tlag__) / (2 * eps)
    fd_depot <- (r1a$rx__sens_depot_BY_tlag__ - r1b$rx__sens_depot_BY_tlag__) / (2 * eps)
    expect_equal(r2$rx__sens_central_BY_tlag_BY_tinf__, fd_central, tolerance = 1e-6)
    expect_equal(r2$rx__sens_depot_BY_tlag_BY_tinf__, fd_depot, tolerance = 1e-6)
    expect_gt(max(abs(fd_central)), 0.05)
  })

  test_that("second-order modeled-DUR boundary: alag x F cross term S^{tlag,doseAmt}", {
    # Full alag+dur+F IMAX model: exercises the SAME combined boundary
    # derivation as the alag+dur-only test above, but with q (doseAmt)
    # driving F instead of dur. A first fix (validated against the
    # alag+dur-only isolate) still had this pair wrong by a full sign flip:
    # the dRateQq helper (d(rate)/dq used in both the MODEL_DUR_ON tau1-
    # boundary term and the MODEL_DUR_OFF tau2-boundary term) had its dur
    # component's sign backwards relative to the already-validated 1st-order
    # d(rate)/dp = (amt*dF + tmp*dDur)/dur formula -- it happened to cancel
    # out for a pure-dur q (like tinf above) but flipped the answer for a
    # pure-F q (doseAmt), so both cross terms are needed to catch this class
    # of bug.
    .mod <- "
      ka <- exp(lka)
      cl <- exp(lcl)
      v  <- exp(lv)
      alag(depot) <- tlag
      dur(depot)  <- tinf
      f(depot)    <- doseAmt
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
    "
    pars <- c(lka = log(1), lcl = log(6), lv = log(60), tlag = 10, doseAmt = 200, tinf = 10)
    m1 <- rxode2(.mod, calcSens = c("tlag", "doseAmt", "tinf"), eventSens = "jump")
    m2 <- rxode2(.mod, calcSens = c("tlag", "doseAmt", "tinf"),
                 calcSens2 = c("tlag", "doseAmt", "tinf"), eventSens = "jump")
    e <- et(amt = 1, cmt = "depot", rate = -2) |> et(seq(0.53, 60, by = 0.5))
    s2 <- rxSolve(m2, e, pars, atol = 1e-11, rtol = 1e-11)[["rx__sens_central_BY_tlag_BY_doseAmt__"]]
    eps <- 1e-4
    p1 <- pars; p1["doseAmt"] <- pars["doseAmt"] + eps
    p2 <- pars; p2["doseAmt"] <- pars["doseAmt"] - eps
    .s1 <- function(p) rxSolve(m1, e, p, atol = 1e-11, rtol = 1e-11)[["rx__sens_central_BY_tlag__"]]
    fd <- (.s1(p1) - .s1(p2)) / (2 * eps)
    expect_equal(s2, fd, tolerance = 1e-6)
    expect_gt(max(abs(fd)), 0.05)
  })

  test_that("modeled dur() STEADY STATE forward jump-sensitivities match FD (solveSSinf moving boundary)", {
    # solveSSinf re-expresses a modeled dur() ss infusion as a fixed-rate window;
    # its classic OFF used to skip the MODEL_DUR_OFF sens forcing-removal + moving-
    # boundary jump, so post-boundary sensitivities were ~30% wrong.  handleSS now
    # arms a marker so handle_evid runs that logic at the re-expressed OFF.
    mt <- "d/dt(depot)=-ka*depot\nd/dt(central)=ka*depot-(cl/v)*central\ndur(central)=9*cl/3.5\ncp=central/(v/1000)"
    cs <- c("ka", "cl"); p <- c(ka = 1.2, cl = 3.5, v = 25)
    mb <- rxode2(mt); mj <- rxode2(mt, calcSens = cs, eventSens = "jump")
    ev <- et(amt = 100, rate = -2, cmt = "central", ss = 1, ii = 24) |> et(c(4, 8, 9.5, 12, 16, 24, 30))
    f <- as.data.frame(suppressWarnings(rxSolve(mj, ev, params = p, atol = 1e-11, rtol = 1e-11)))
    mx <- 0
    for (pn in cs) {
      hh <- abs(p[[pn]]) * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
      sp <- as.data.frame(suppressWarnings(rxSolve(mb, ev, params = pp, atol = 1e-11, rtol = 1e-11)))
      sm <- as.data.frame(suppressWarnings(rxSolve(mb, ev, params = pm, atol = 1e-11, rtol = 1e-11)))
      for (st in c("depot", "central")) {
        fd <- (sp[[st]] - sm[[st]]) / (2 * hh)
        mx <- max(mx, max(abs(f[[sprintf("rx__sens_%s_BY_%s__", st, pn)]] - fd), na.rm = TRUE))
      }
    }
    # 1e-4 keeps power against the ~30% skip bug; 1e-5 flakes on macOS ARM64 (3.5e-5)
    expect_lt(mx, 1e-4)
  })

  test_that("modeled rate() moving-boundary forward jump-sensitivities match FD (non-ss AND ss)", {
    # The modeled rate() OFF boundary tau2 = tau1 + F*amt/rate(p) moves with the
    # parameters; that transversality jump [S] = amt*dF - (F*amt/rate)*d(rate)/dp
    # was previously DEFERRED (MODEL_RATE_OFF only removed the continuous forcing),
    # so both regular AND steady-state modeled-rate sensitivities were ~30% wrong.
    cs <- c("ka", "cl"); p <- c(ka = 1.2, cl = 3.5, v = 25)
    chk <- function(mt, ev, drop = NA) {
      st0 <- rxode2::.rxAdjointExpand(mt, cs)$st
      mb <- rxode2(mt); mj <- rxode2(mt, calcSens = cs, eventSens = "jump")
      a0 <- list(mj, ev, params = p, atol = 1e-11, rtol = 1e-11); if (!is.na(drop)) a0$addlDropSs <- drop
      f <- as.data.frame(suppressWarnings(do.call(rxSolve, a0)))
      mx <- 0
      for (pn in cs) {
        hh <- abs(p[[pn]]) * 1e-6; pp <- p; pm <- p; pp[pn] <- pp[pn] + hh; pm[pn] <- pm[pn] - hh
        a1 <- list(mb, ev, params = pp, atol = 1e-11, rtol = 1e-11); a2 <- list(mb, ev, params = pm, atol = 1e-11, rtol = 1e-11)
        if (!is.na(drop)) { a1$addlDropSs <- drop; a2$addlDropSs <- drop }
        sp <- as.data.frame(suppressWarnings(do.call(rxSolve, a1)))
        sm <- as.data.frame(suppressWarnings(do.call(rxSolve, a2)))
        for (st in st0) { fd <- (sp[[st]] - sm[[st]]) / (2 * hh)
          mx <- max(mx, max(abs(f[[sprintf("rx__sens_%s_BY_%s__", st, pn)]] - fd), na.rm = TRUE)) }
      }
      expect_lt(mx, 5e-5)   # FD-truncation-safe; the deferred-boundary bug was ~1e1
    }
    mt  <- "d/dt(depot)=-ka*depot\nd/dt(central)=ka*depot-(cl/v)*central\nrate(central)=11*cl/3.5\ncp=central/(v/1000)"
    mtF <- "d/dt(depot)=-ka*depot\nd/dt(central)=ka*depot-(cl/v)*central\nf(central)=0.61\nrate(central)=11*cl/3.5\ncp=central/(v/1000)"
    chk(mt,  et(amt = 100, rate = -1, cmt = "central") |> et(c(1, 4, 8, 12, 16, 20)))                              # non-ss
    chk(mtF, et(amt = 100, rate = -1, cmt = "central") |> et(c(1, 4, 8, 12, 16, 20)))                              # non-ss, F != 1
    chk(mt,  et(amt = 100, rate = -1, cmt = "central", ss = 1, ii = 24) |> et(c(4, 8, 12, 16, 24, 30)), TRUE)      # ss
  })
})

# compatibility function for units

.setUnitsMode <- function() {
  if (requireNamespace("units", quietly = TRUE)) {
    return(units::units_options("set_units_mode"))
  } else {
    return(NULL)
  }
}

.unitless <- function() {
  if (requireNamespace("units", quietly = TRUE)) {
    return(units::unitless)
  } else {
    return(NULL)
  }
}

.amtTrans <- function(data) {
  if (!any(names(data) == "amt")) {
    stop("need 'amt' aesthetic")
  } else if (!any(names(data) == "x") && any(names(data) == "time")) {
    data$x <- data$time
  } else if (!any(names(data) == "x")) {
    stop("need 'x' aesthetic")
  }
  ret <- data[!is.na(data$amt), c("x", "amt")]
  ret$xend <- ret$x
  ret$y <- -Inf
  ret$yend <- Inf
  ret
}

GeomAmt <- NULL

StatAmt <- ggplot2::ggproto("StatAmt", ggplot2::Stat,
  compute_group = function(data, scales) {
    .amtTrans(data)
  },
  required_aes = c("x", "amt")
)


#' Dosing/Amt geom/stat
#'
#' This is a dosing geom that shows the vertical lines where a dose occurs
#'
#' Requires the following aesthetics:
#'
#' \itemize{
#' \item x representing the x values, usually time
#' \item amt representing the dosing values;  They are missing or zero when no dose is given
#' }
#'
#' @return This returns a stat_amt in context of a ggplot2 plot
#'
#' @examples
#'
#' \donttest{
#' library(rxode2)
#' library(units)
#'
#'
#' ## Model from RxODE tutorial
#' mod1 <- function() {
#'   ini({
#'     KA <- 2.94E-01
#'     CL <- 1.86E+01
#'     V2 <- 4.02E+01
#'     Q  <- 1.05E+01
#'     V3 <- 2.97E+02
#'     Kin <- 1
#'     Kout <- 1
#'     EC50 <- 200
#'   })
#'   model({
#'     C2 <- centr/V2
#'     C3 <- peri/V3
#'     d/dt(depot) <- -KA*depot
#'     d/dt(centr) <-  KA*depot - CL*C2 - Q*C2 + Q*C3
#'     d/dt(peri)  <-                    Q*C2 - Q*C3
#'     d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
#'   })
#' }
#'
#' ## These are making the more complex regimens of the rxode2 tutorial
#'
#' ## bid for 5 days
#' bid <- et(timeUnits="hr") |>
#'   et(amt=10000,ii=12,until=set_units(5, "days"))
#'
#' ## qd for 5 days
#' qd <- et(timeUnits="hr") |>
#'       et(amt=20000,ii=24,until=set_units(5, "days"))
#'
#' ## bid for 5 days followed by qd for 5 days
#'
#' et <- seq(bid,qd) |>
#'       et(seq(0,11*24,length.out=100))
#'
#' bidQd <- rxSolve(mod1, et, addDosing=TRUE)
#'
#' # by default dotted and under-stated
#' plot(bidQd, C2) + geom_amt(aes(amt=amt))
#'
#' # of course you can make it a bit more visible
#'
#' plot(bidQd, C2) + geom_amt(aes(amt=amt), col="red", lty=1, linewidth=1.2)
#'
#' }
#' @export
#' @inheritParams ggplot2::stat_identity
stat_amt <- function(mapping = NULL, data = NULL,
                     position = "identity", show.legend = NA,
                     inherit.aes = TRUE, ...) {
  ggplot2::layer(
    stat = StatAmt, data = data, mapping = mapping, geom = GeomAmt,
    position = position, show.legend = show.legend, inherit.aes = inherit.aes,
    params = list(na.rm = TRUE, ...)
  )
}

#' @rdname stat_amt
#' @export
geom_amt <- stat_amt

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that(".amtTrans", {
      expect_error(
        .amtTrans(data.frame()),
        regexp = "need 'amt' aesthetic"
      )
      expect_error(
        .amtTrans(data.frame(amt = 1)),
        regexp = "need 'x' aesthetic"
      )
      # No error for missing 'x' if 'time' is present
      expect_error(
        .amtTrans(data.frame(amt = 1, time = 1)),
        regexp = NA
      )
      expect_equal(
        .amtTrans(data.frame(amt = 1, time = 1)),
        data.frame(x = 1, amt = 1, xend = 1, y = -Inf, yend = Inf)
      )
      # NA rows are dropped
      expect_equal(
        .amtTrans(data.frame(amt = c(1, NA), time = 1:2)),
        data.frame(x = 1, amt = 1, xend = 1, y = -Inf, yend = Inf)
      )
    })
  }
})

# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxnorm <- function(mean = 0, sd = 1) {
    .mean <- as.character(substitute(mean))
    .dp <- deparse1(substitute(mean))
    .tmp <- suppressWarnings(try(force(mean), silent = TRUE))
    .mean <- .uiArg(.mean, .tmp, .dp)
    .sd <- as.character(substitute(sd))
    .dp <- deparse1(substitute(sd))
    .tmp <- suppressWarnings(try(force(sd), silent = TRUE))
    .sd <- .uiArg(.sd, .tmp, .dp)
    list(replace = paste0("rxnorm(", .mean, ", ", .sd, ")"))
}

#'@export
rxUdfUi.rxnorm <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxpois <- function(lambda) {
    .lambda <- as.character(substitute(lambda))
    .dp <- deparse1(substitute(lambda))
    .tmp <- suppressWarnings(try(force(lambda), silent = TRUE))
    .lambda <- .uiArg(.lambda, .tmp, .dp)
    list(replace = paste0("rxpois(", .lambda, ")"))
}

#'@export
rxUdfUi.rxpois <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxt <- function(df) {
    .df <- as.character(substitute(df))
    .dp <- deparse1(substitute(df))
    .tmp <- suppressWarnings(try(force(df), silent = TRUE))
    .df <- .uiArg(.df, .tmp, .dp)
    list(replace = paste0("rxt(", .df, ")"))
}

#'@export
rxUdfUi.rxt <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxunif <- function(min = 0, max = 1) {
    .min <- as.character(substitute(min))
    .dp <- deparse1(substitute(min))
    .tmp <- suppressWarnings(try(force(min), silent = TRUE))
    .min <- .uiArg(.min, .tmp, .dp)
    .max <- as.character(substitute(max))
    .dp <- deparse1(substitute(max))
    .tmp <- suppressWarnings(try(force(max), silent = TRUE))
    .max <- .uiArg(.max, .tmp, .dp)
    list(replace = paste0("rxunif(", .min, ", ", .max, ")"))
}

#'@export
rxUdfUi.rxunif <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxweibull <- function(shape, scale = 1) {
    .shape <- as.character(substitute(shape))
    .dp <- deparse1(substitute(shape))
    .tmp <- suppressWarnings(try(force(shape), silent = TRUE))
    .shape <- .uiArg(.shape, .tmp, .dp)
    .scale <- as.character(substitute(scale))
    .dp <- deparse1(substitute(scale))
    .tmp <- suppressWarnings(try(force(scale), silent = TRUE))
    .scale <- .uiArg(.scale, .tmp, .dp)
    list(replace = paste0("rxweibull(", .shape, ", ", .scale, 
        ")"))
}

#'@export
rxUdfUi.rxweibull <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxgeom <- function(prob) {
    .prob <- as.character(substitute(prob))
    .dp <- deparse1(substitute(prob))
    .tmp <- suppressWarnings(try(force(prob), silent = TRUE))
    .prob <- .uiArg(.prob, .tmp, .dp)
    list(replace = paste0("rxgeom(", .prob, ")"))
}

#'@export
rxUdfUi.rxgeom <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxbeta <- function(shape1, shape2) {
    .shape1 <- as.character(substitute(shape1))
    .dp <- deparse1(substitute(shape1))
    .tmp <- suppressWarnings(try(force(shape1), silent = TRUE))
    .shape1 <- .uiArg(.shape1, .tmp, .dp)
    .shape2 <- as.character(substitute(shape2))
    .dp <- deparse1(substitute(shape2))
    .tmp <- suppressWarnings(try(force(shape2), silent = TRUE))
    .shape2 <- .uiArg(.shape2, .tmp, .dp)
    list(replace = paste0("rxbeta(", .shape1, ", ", .shape2, 
        ")"))
}

#'@export
rxUdfUi.rxbeta <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxgamma <- function(shape, rate = 1) {
    .shape <- as.character(substitute(shape))
    .dp <- deparse1(substitute(shape))
    .tmp <- suppressWarnings(try(force(shape), silent = TRUE))
    .shape <- .uiArg(.shape, .tmp, .dp)
    .rate <- as.character(substitute(rate))
    .dp <- deparse1(substitute(rate))
    .tmp <- suppressWarnings(try(force(rate), silent = TRUE))
    .rate <- .uiArg(.rate, .tmp, .dp)
    list(replace = paste0("rxgamma(", .shape, ", ", .rate, ")"))
}

#'@export
rxUdfUi.rxgamma <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxf <- function(df1, df2) {
    .df1 <- as.character(substitute(df1))
    .dp <- deparse1(substitute(df1))
    .tmp <- suppressWarnings(try(force(df1), silent = TRUE))
    .df1 <- .uiArg(.df1, .tmp, .dp)
    .df2 <- as.character(substitute(df2))
    .dp <- deparse1(substitute(df2))
    .tmp <- suppressWarnings(try(force(df2), silent = TRUE))
    .df2 <- .uiArg(.df2, .tmp, .dp)
    list(replace = paste0("rxf(", .df1, ", ", .df2, ")"))
}

#'@export
rxUdfUi.rxf <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxexp <- function(rate) {
    .rate <- as.character(substitute(rate))
    .dp <- deparse1(substitute(rate))
    .tmp <- suppressWarnings(try(force(rate), silent = TRUE))
    .rate <- .uiArg(.rate, .tmp, .dp)
    list(replace = paste0("rxexp(", .rate, ")"))
}

#'@export
rxUdfUi.rxexp <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxchisq <- function(df) {
    .df <- as.character(substitute(df))
    .dp <- deparse1(substitute(df))
    .tmp <- suppressWarnings(try(force(df), silent = TRUE))
    .df <- .uiArg(.df, .tmp, .dp)
    list(replace = paste0("rxchisq(", .df, ")"))
}

#'@export
rxUdfUi.rxchisq <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxcauchy <- function(location = 0, scale = 1) {
    .location <- as.character(substitute(location))
    .dp <- deparse1(substitute(location))
    .tmp <- suppressWarnings(try(force(location), silent = TRUE))
    .location <- .uiArg(.location, .tmp, .dp)
    .scale <- as.character(substitute(scale))
    .dp <- deparse1(substitute(scale))
    .tmp <- suppressWarnings(try(force(scale), silent = TRUE))
    .scale <- .uiArg(.scale, .tmp, .dp)
    list(replace = paste0("rxcauchy(", .location, ", ", .scale, 
        ")"))
}

#'@export
rxUdfUi.rxcauchy <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rxbinom <- function(size, prob) {
    .size <- as.character(substitute(size))
    .dp <- deparse1(substitute(size))
    .tmp <- suppressWarnings(try(force(size), silent = TRUE))
    .size <- .uiArg(.size, .tmp, .dp)
    .prob <- as.character(substitute(prob))
    .dp <- deparse1(substitute(prob))
    .tmp <- suppressWarnings(try(force(prob), silent = TRUE))
    .prob <- .uiArg(.prob, .tmp, .dp)
    list(replace = paste0("rxbinom(", .size, ", ", .prob, ")"))
}

#'@export
rxUdfUi.rxbinom <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rinorm <- function(mean = 0, sd = 1) {
    .mean <- as.character(substitute(mean))
    .dp <- deparse1(substitute(mean))
    .tmp <- suppressWarnings(try(force(mean), silent = TRUE))
    .mean <- .uiArg(.mean, .tmp, .dp)
    .sd <- as.character(substitute(sd))
    .dp <- deparse1(substitute(sd))
    .tmp <- suppressWarnings(try(force(sd), silent = TRUE))
    .sd <- .uiArg(.sd, .tmp, .dp)
    list(replace = paste0("rinorm(", .mean, ", ", .sd, ")"))
}

#'@export
rxUdfUi.rinorm <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.ripois <- function(lambda) {
    .lambda <- as.character(substitute(lambda))
    .dp <- deparse1(substitute(lambda))
    .tmp <- suppressWarnings(try(force(lambda), silent = TRUE))
    .lambda <- .uiArg(.lambda, .tmp, .dp)
    list(replace = paste0("ripois(", .lambda, ")"))
}

#'@export
rxUdfUi.ripois <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rit <- function(df) {
    .df <- as.character(substitute(df))
    .dp <- deparse1(substitute(df))
    .tmp <- suppressWarnings(try(force(df), silent = TRUE))
    .df <- .uiArg(.df, .tmp, .dp)
    list(replace = paste0("rit(", .df, ")"))
}

#'@export
rxUdfUi.rit <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.riunif <- function(min = 0, max = 1) {
    .min <- as.character(substitute(min))
    .dp <- deparse1(substitute(min))
    .tmp <- suppressWarnings(try(force(min), silent = TRUE))
    .min <- .uiArg(.min, .tmp, .dp)
    .max <- as.character(substitute(max))
    .dp <- deparse1(substitute(max))
    .tmp <- suppressWarnings(try(force(max), silent = TRUE))
    .max <- .uiArg(.max, .tmp, .dp)
    list(replace = paste0("riunif(", .min, ", ", .max, ")"))
}

#'@export
rxUdfUi.riunif <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.riweibull <- function(shape, scale = 1) {
    .shape <- as.character(substitute(shape))
    .dp <- deparse1(substitute(shape))
    .tmp <- suppressWarnings(try(force(shape), silent = TRUE))
    .shape <- .uiArg(.shape, .tmp, .dp)
    .scale <- as.character(substitute(scale))
    .dp <- deparse1(substitute(scale))
    .tmp <- suppressWarnings(try(force(scale), silent = TRUE))
    .scale <- .uiArg(.scale, .tmp, .dp)
    list(replace = paste0("riweibull(", .shape, ", ", .scale, 
        ")"))
}

#'@export
rxUdfUi.riweibull <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rigeom <- function(prob) {
    .prob <- as.character(substitute(prob))
    .dp <- deparse1(substitute(prob))
    .tmp <- suppressWarnings(try(force(prob), silent = TRUE))
    .prob <- .uiArg(.prob, .tmp, .dp)
    list(replace = paste0("rigeom(", .prob, ")"))
}

#'@export
rxUdfUi.rigeom <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.ribeta <- function(shape1, shape2) {
    .shape1 <- as.character(substitute(shape1))
    .dp <- deparse1(substitute(shape1))
    .tmp <- suppressWarnings(try(force(shape1), silent = TRUE))
    .shape1 <- .uiArg(.shape1, .tmp, .dp)
    .shape2 <- as.character(substitute(shape2))
    .dp <- deparse1(substitute(shape2))
    .tmp <- suppressWarnings(try(force(shape2), silent = TRUE))
    .shape2 <- .uiArg(.shape2, .tmp, .dp)
    list(replace = paste0("ribeta(", .shape1, ", ", .shape2, 
        ")"))
}

#'@export
rxUdfUi.ribeta <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rigamma <- function(shape, rate = 1) {
    .shape <- as.character(substitute(shape))
    .dp <- deparse1(substitute(shape))
    .tmp <- suppressWarnings(try(force(shape), silent = TRUE))
    .shape <- .uiArg(.shape, .tmp, .dp)
    .rate <- as.character(substitute(rate))
    .dp <- deparse1(substitute(rate))
    .tmp <- suppressWarnings(try(force(rate), silent = TRUE))
    .rate <- .uiArg(.rate, .tmp, .dp)
    list(replace = paste0("rigamma(", .shape, ", ", .rate, ")"))
}

#'@export
rxUdfUi.rigamma <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.rif <- function(df1, df2) {
    .df1 <- as.character(substitute(df1))
    .dp <- deparse1(substitute(df1))
    .tmp <- suppressWarnings(try(force(df1), silent = TRUE))
    .df1 <- .uiArg(.df1, .tmp, .dp)
    .df2 <- as.character(substitute(df2))
    .dp <- deparse1(substitute(df2))
    .tmp <- suppressWarnings(try(force(df2), silent = TRUE))
    .df2 <- .uiArg(.df2, .tmp, .dp)
    list(replace = paste0("rif(", .df1, ", ", .df2, ")"))
}

#'@export
rxUdfUi.rif <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.riexp <- function(rate) {
    .rate <- as.character(substitute(rate))
    .dp <- deparse1(substitute(rate))
    .tmp <- suppressWarnings(try(force(rate), silent = TRUE))
    .rate <- .uiArg(.rate, .tmp, .dp)
    list(replace = paste0("riexp(", .rate, ")"))
}

#'@export
rxUdfUi.riexp <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.richisq <- function(df) {
    .df <- as.character(substitute(df))
    .dp <- deparse1(substitute(df))
    .tmp <- suppressWarnings(try(force(df), silent = TRUE))
    .df <- .uiArg(.df, .tmp, .dp)
    list(replace = paste0("richisq(", .df, ")"))
}

#'@export
rxUdfUi.richisq <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.ricauchy <- function(location = 0, scale = 1) {
    .location <- as.character(substitute(location))
    .dp <- deparse1(substitute(location))
    .tmp <- suppressWarnings(try(force(location), silent = TRUE))
    .location <- .uiArg(.location, .tmp, .dp)
    .scale <- as.character(substitute(scale))
    .dp <- deparse1(substitute(scale))
    .tmp <- suppressWarnings(try(force(scale), silent = TRUE))
    .scale <- .uiArg(.scale, .tmp, .dp)
    list(replace = paste0("ricauchy(", .location, ", ", .scale, 
        ")"))
}

#'@export
rxUdfUi.ricauchy <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.ribinom <- function(size, prob) {
    .size <- as.character(substitute(size))
    .dp <- deparse1(substitute(size))
    .tmp <- suppressWarnings(try(force(size), silent = TRUE))
    .size <- .uiArg(.size, .tmp, .dp)
    .prob <- as.character(substitute(prob))
    .dp <- deparse1(substitute(prob))
    .tmp <- suppressWarnings(try(force(prob), silent = TRUE))
    .prob <- .uiArg(.prob, .tmp, .dp)
    list(replace = paste0("ribinom(", .size, ", ", .prob, ")"))
}

#'@export
rxUdfUi.ribinom <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.logit <- function(x, low = 0, high = 1) {
    .x <- as.character(substitute(x))
    .dp <- deparse1(substitute(x))
    .tmp <- suppressWarnings(try(force(x), silent = TRUE))
    .x <- .uiArg(.x, .tmp, .dp)
    .low <- as.character(substitute(low))
    .dp <- deparse1(substitute(low))
    .tmp <- suppressWarnings(try(force(low), silent = TRUE))
    .low <- .uiArg(.low, .tmp, .dp)
    .high <- as.character(substitute(high))
    .dp <- deparse1(substitute(high))
    .tmp <- suppressWarnings(try(force(high), silent = TRUE))
    .high <- .uiArg(.high, .tmp, .dp)
    list(replace = paste0("logit(", .x, ", ", .low, ", ", .high, 
        ")"))
}

#'@export
rxUdfUi.logit <- rxUdfUi.rxpois

## nocov end
# This file is generated by .generateRandomUiFuns() in build.R
## nocov start
.expit <- function(x, low = 0, high = 1) {
    .x <- as.character(substitute(x))
    .dp <- deparse1(substitute(x))
    .tmp <- suppressWarnings(try(force(x), silent = TRUE))
    .x <- .uiArg(.x, .tmp, .dp)
    .low <- as.character(substitute(low))
    .dp <- deparse1(substitute(low))
    .tmp <- suppressWarnings(try(force(low), silent = TRUE))
    .low <- .uiArg(.low, .tmp, .dp)
    .high <- as.character(substitute(high))
    .dp <- deparse1(substitute(high))
    .tmp <- suppressWarnings(try(force(high), silent = TRUE))
    .high <- .uiArg(.high, .tmp, .dp)
    list(replace = paste0("expit(", .x, ", ", .low, ", ", .high, 
        ")"))
}

#'@export
rxUdfUi.expit <- rxUdfUi.rxpois

## nocov end

library(rxode2)
library(testthat)
library(data.table)
# CRAN work-arounds
if (!identical(Sys.getenv("NOT_CRAN"), "true")) {
  setRxThreads(2L)
  setDTthreads(2L)
  Sys.setenv(OMP_NUM_THREADS = "2")
  Sys.setenv(MKL_NUM_THREADS = "2")
  if (identical(Sys.info()["sysname"], "Darwin")) {
    rxode2::rxUnloadAll(set=FALSE)
  }
}
#test_check("rxode2", reporter = testthat::LocationReporter,
#          stop_on_failure = FALSE)
## test_check("rxode2", reporter = testthat::LocationReporter,
##            stop_on_failure = TRUE)
withr::with_options(list(rxode2.useLinCmt=FALSE), {
  test_check("rxode2")
})

# This get/sets a control option from the rxode2 UI


#' rxRemoveControl options for UI object
#'
#' @param ui rxode2 ui object
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @export
rxRemoveControl <- function(ui) {
  if (is.list(ui) || inherits(ui, "raw")) {
    stop("cannot remove from compressed 'rxUi'\nfirst decompress with `rxode2::rxUiDecompress()'",
         call.=FALSE)
  }
  if (exists("control", ui))
    rm("control", envir=ui)
  invisible()
}


#' rxSetControl options for UI object
#'
#'
#' @param ui rxode2 ui object
#' @param control Default value
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @export
rxSetControl <- function(ui, control) {
  if (inherits(control, "list")) {
    assign("control", control, envir=ui)
  } else {
    stop("control must be a list-like object")
  }
  invisible()
}

#'  Assign Control Variable
#'
#' @param ui rxode2 ui function
#' @param option Option name in the control to modify
#' @param value Value of control to modify
#' @return Nothing; called for the side effects
#' @author Matthew L. Fidler
#' @export
rxAssignControlValue <- function(ui, option, value) {
  if (is.list(ui)) {
    stop("cannot assign value to compressed 'rxUi'\nfirst decompress with `rxode2::rxUiDecompress()'",
         call.=FALSE)
  }
  if (exists("control", envir=ui)) {
    .ctl <- get("control", envir=ui)
  } else {
    .ctl <- list()
  }
  .ctl[[option]] <- value
  assign("control", .ctl, envir=ui)
  invisible()
}
#'  rxGetControl option from ui
#'
#'
#' @param ui rxode2 ui object
#' @param option Option to get
#' @param default Default value
#' @return Option (if present) or default value
#' @author Matthew L. Fidler
#' @export
rxGetControl <- function(ui, option, default) {
  ui <- rxUiDecompress(ui)
  if (!exists("control", envir=ui)) return(default)
  .ctl <- get("control", envir=ui)
  if (option %in% names(.ctl)) return(.ctl[[option]])
  default
}

## devtools::load_all("~/src/rxode2")

## devtools::load_all("~/src/nlmixr2est")
library(nlmixr2)

## f <- function() {
##   ini({ # Where initial conditions/variables are specified
##     # '<-' or '=' defines population parameters
##     # Simple numeric expressions are supported
##     lCl <- 1.6      #log Cl (L/hr)
##     lVc <- 4.5      #log V (L)
##     # Bounds may be specified by c(lower, est, upper), like NONMEM:
##     # Residuals errors are assumed to be population parameters
##     prop.err <- c(0, 0.3, 1)
##     # Between subject variability estimates are specified by '~'
##     # Semicolons are optional
##     eta.Vc ~ 0.1   #IIV V
##     eta.Cl ~ 0.1   #IIV Cl
##   })
##   model({ # Where the model is specified
##     # The model uses the ini-defined variable names
##     Vc <- exp(lVc + eta.Vc)
##     Cl <- exp(lCl + eta.Cl)
##     cp <- linCmt()
##     cp ~ prop(prop.err)
##   })
## }

## f2 <- function() {
##   ini({ # Where initial conditions/variables are specified
##     # '<-' or '=' defines population parameters
##     # Simple numeric expressions are supported
##     lCl <- 1.6      #log Cl (L/hr)
##     lVc <- 4.5      #log V (L)
##     # Bounds may be specified by c(lower, est, upper), like NONMEM:
##     # Residuals errors are assumed to be population parameters
##     prop.err <- c(0, 0.3, 1)
##     # Between subject variability estimates are specified by '~'
##     # Semicolons are optional
##     eta.Vc ~ 0.1   #IIV V
##     eta.Cl ~ 0.1   #IIV Cl
##   })
##   model({ # Where the model is specified
##     # The model uses the ini-defined variable names
##     Vc <- exp(lVc + eta.Vc)
##     Cl <- exp(lCl + eta.Cl)
##     d/dt(central) <- -Cl/Vc * central
##     cp <- central/Vc
##     cp ~ prop(prop.err)
##   })
## }


## datr <- nlmixr2data::Bolus_1CPT
## dat <- datr[datr$SD == 1, ]
## dat <- dat[, names(dat) != "SS"]

## fit <- nlmixr2(f, dat, est="focei")

## fitS <- nlmixr2(f, dat, est="saem")



## fit2 <- nlmixr2(f2, dat, est="focei")

## fit2S <- nlmixr2(f2, dat, est="saem")

## print(c("linear"=sum(fit$time), "ode"=sum(fit2$time)))

## print(c("linear"=sum(fitS$time), "ode"=sum(fit2S$time)))

## f <- function(){
##   ini({ # Where initial conditions/variables are specified
##     # '<-' or '=' defines population parameters
##     # Simple numeric expressions are supported
##     lCl <- 1.8      #log Cl (L/hr)
##     lVc <- 4.7      #log V (L)
##     lKA <- 0.2      #log V (L)
##     # Bounds may be specified by c(lower, est, upper), like NONMEM:
##     # Residuals errors are assumed to be population parameters
##     prop.err <- c(0, 0.3, 1)
##     # Between subject variability estimates are specified by '~'
##     # Semicolons are optional
##     eta.Cl ~ 0.15
##     eta.Vc ~ 0.15
##     eta.KA ~ 0.15
##   })
##   model({ # Where the model is specified
##     # The model uses the ini-defined variable names
##     Cl <- exp(lCl + eta.Cl)
##     Vc <- exp(lVc + eta.Vc)
##     KA <- exp(lKA + eta.KA)
##     cp <- linCmt()
##     cp ~ prop(prop.err)
##   })
## }

## f2 <- function() {
##   ini({ # Where initial conditions/variables are specified
##     # '<-' or '=' defines population parameters
##     # Simple numeric expressions are supported
##     lCl <- 1.8      #log Cl (L/hr)
##     lVc <- 4.7      #log V (L)
##     lKA <- 0.2      #log V (L)
##     # Bounds may be specified by c(lower, est, upper), like NONMEM:
##     # Residuals errors are assumed to be population parameters
##     prop.err <- c(0, 0.3, 1)
##     # Between subject variability estimates are specified by '~'
##     # Semicolons are optional
##     eta.Cl ~ 0.15
##     eta.Vc ~ 0.15
##     eta.KA ~ 0.15
##   })
##   model({ # Where the model is specified
##     # The model uses the ini-defined variable names
##     Cl <- exp(lCl + eta.Cl)
##     Vc <- exp(lVc + eta.Vc)
##     KA <- exp(lKA + eta.KA)
##     # RxODE-style differential equations are supported
##     d/dt(depot)    = -KA*depot;
##     d/dt(centr)  =  KA*depot-(Cl/Vc)*centr;
##     ## Concentration is calculated
##     cp = centr / Vc;
##     # And is assumed to follow proportional error estimated by prop.err
##     cp ~ prop(prop.err)
##   })
## }

## datr <- nlmixr2data::Oral_1CPT
## dat <- datr[datr$SD == 1, ]
## dat <- dat[, names(dat) != "SS"]

## fit <- nlmixr2(f, dat, est="focei")
## fitS <- nlmixr2(f, dat, est="saem")

## fit2 <- nlmixr2(f2, dat, est="focei")
## fit2S <- nlmixr2(f2, dat, est="saem")

## print(c("linear"=sum(fitS$time), "ode"=sum(fit2S$time)))

## print(c("linear"=sum(fit$time), "ode"=sum(fit2$time)))


f <- function() {
  ini({ # Where initial conditions/variables are specified
    # '<-' or '=' defines population parameters
    # Simple numeric expressions are supported
    lCl <- 1.6      #log Cl (L/hr)
    lVc <- 4.5      #log Vc (L)
    lQ  <- 1.6      #log Q (L/hr)
    lVp <- 4        #log Vp (L)

    # Bounds may be specified by c(lower, est, upper), like NONMEM:
    # Residuals errors are assumed to be population parameters
    prop.err <- c(0, 0.3, 1)
    # Between subject variability estimates are specified by '~'
    # Semicolons are optional
    eta.Vc ~ 0.15
    eta.Cl ~ 0.15
    eta.Vp ~ 0.15
    eta.Q  ~ 0.15
  })
  model({ # Where the model is specified
    # The model uses the ini-defined variable names
    Vc <- exp(lVc + eta.Vc)
    Cl <- exp(lCl + eta.Cl)
    Vp <- exp(lVp + eta.Vp)
    Q <- exp(lQ + eta.Q)
    # And is assumed to follow proportional error estimated by prop.err
    cp <- linCmt()
    cp ~ prop(prop.err)
  })
}

f2 <- function() {
  ini({ # Where initial conditions/variables are specified
    # '<-' or '=' defines population parameters
    # Simple numeric expressions are supported
    lCl <- 1.6      #log Cl (L/hr)
    lVc <- 4.5      #log Vc (L)
    lQ  <- 1.6      #log Q (L/hr)
    lVp <- 4        #log Vp (L)

    # Bounds may be specified by c(lower, est, upper), like NONMEM:
    # Residuals errors are assumed to be population parameters
    prop.err <- c(0, 0.3, 1)
    # Between subject variability estimates are specified by '~'
    # Semicolons are optional
    eta.Vc ~ 0.15
    eta.Cl ~ 0.15
    eta.Vp ~ 0.15
    eta.Q  ~ 0.15
  })
  model({ # Where the model is specified
    # The model uses the ini-defined variable names
    Vc <- exp(lVc + eta.Vc)
    Cl <- exp(lCl + eta.Cl)
    Vp <- exp(lVp + eta.Vp)
    Q <- exp(lQ + eta.Q)
    # RxODE-style differential equations are supported
    K10<- Cl/Vc
    K12<- Q/Vc
    K21<- Q/Vp
    d/dt(centr)  = K21*periph-K12*centr-K10*centr;
    d/dt(periph) =-K21*periph+K12*centr;
    ## Concentration is calculated
    cp = centr / Vc;
    # And is assumed to follow proportional error estimated by prop.err
    cp ~ prop(prop.err)
  })
}

datr <- nlmixr2data::Bolus_2CPT
dat <- datr[datr$SD == 1, ]
dat <- dat[, names(dat) != "SS"]

fit <- nlmixr2(f, dat, est="focei")

## fitS <- nlmixr2(f, dat, est="saem")

## fit2 <- nlmixr2(f2, dat, est="focei")

## fit2S <- nlmixr2(f2, dat, est="saem")


## print(c("linear"=sum(fit$time), "ode"=sum(fit2$time)))

## print(c("linear"=sum(fitS$time), "ode"=sum(fit2S$time)))


## theta <- setNames(fit$theta,paste0("THETA[", seq_along(fit$theta), "]"))

## eta <- fit$eta[,-1]
## names(eta) <- paste0("ETA[", seq_along(eta), "]")



## ###################################################################


## devtools::load_all("~/src/rxode2")

## datr <- nlmixr2data::Bolus_1CPT
## dat <- datr[datr$SD == 1, ]
## dat <- dat[, names(dat) != "SS"]



## theta <- p[[1]]

## eta <- p[[2]]

## pars <- data.frame(as.data.frame(t(theta), check.names=FALSE),eta, check.names=FALSE)

## rx <- rxode2({
##   param(THETA[1], THETA[2], THETA[3], ETA[1], ETA[2])
##   rx_yj_ ~ 2
##   rx_lambda_ ~ 1
##   rx_hi_ ~ 1
##   rx_low_ ~ 0
##   rx_expr_0 ~ ETA[2] + THETA[1]
##   rx_expr_1 ~ ETA[1] + THETA[2]
##   rx_expr_2 ~ exp(rx_expr_0)
##   rx_expr_3 ~ exp(rx_expr_1)
##   rx_pred_ = linCmtB(rx__PTR__, t, 1, 1, 0, -1, -1, 1, rx_expr_2,
##                      rx_expr_3, 0, 0, 0, 0, 0)
##   rx__sens_rx_pred__BY_ETA_1___ = rx_expr_3 * linCmtB(rx__PTR__,
##                                                       t, 1, 1, 0, -2, 1, 1, rx_expr_2, rx_expr_3, 0, 0, 0,
##                                                       0, 0)
##   rx__sens_rx_pred__BY_ETA_2___ = rx_expr_2 * linCmtB(rx__PTR__,
##                                                       t, 1, 1, 0, -2, 0, 1, rx_expr_2, rx_expr_3, 0, 0, 0,
##                                                       0, 0)
##   rx_r_ = Rx_pow_di((linCmtB(rx__PTR__, t, 1, 1, 0, -1, -1,
##                              1, rx_expr_2, rx_expr_3, 0, 0, 0, 0, 0) * THETA[3]),
##                     2)
##   rx__sens_rx_r__BY_ETA_1___ = 2 * rx_expr_3 * (linCmtB(rx__PTR__,
##                                                         t, 1, 1, 0, -1, -1, 1, rx_expr_2, rx_expr_3, 0, 0, 0,
##                                                         0, 0) * THETA[3]) * linCmtB(rx__PTR__, t, 1, 1, 0, -2,
##                                                                                     1, 1, rx_expr_2, rx_expr_3, 0, 0, 0, 0, 0) * THETA[3]
##   rx__sens_rx_r__BY_ETA_2___ = 2 * rx_expr_2 * (linCmtB(rx__PTR__,
##                                                         t, 1, 1, 0, -1, -1, 1, rx_expr_2, rx_expr_3, 0, 0, 0,
##                                                         0, 0) * THETA[3]) * linCmtB(rx__PTR__, t, 1, 1, 0, -2,
##                                                                                     0, 1, rx_expr_2, rx_expr_3, 0, 0, 0, 0, 0) * THETA[3]
##   cmt(cp)
##   dvid(2)
## })


## tmp <- rxSolve(rx, pars, dat)

#' mlogit -- Convert multiple probabilities to log-scale numbers
#'
#' These multiple probabilities need to add up to be less than 1.
#'
#' Once converted to log-scale numbers, they can be used in the
#' in the `mexpit()` to get the probabilities back with the following equation
#'
#'  \deqn{p_i = \frac{e^{x_i}}{1+\sum_{j=1}^{N-1} e^{x_j}}}
#'
#' This ensures one remaining probability will add to one, that is
#'
#' \deqn{p_N = \frac{1}{1+\sum_{j=1}^{N-1} e^{x_j}}}
#'
#' Unfortunately, the log-scale inverse cannot be solved analytically,
#' so it is solved with the `rootSolve::multiroot()` function.
#'
#' You may adjust some of the root finding options when using this
#' function.
#'
#' When running `nlmixr2` with mixture models (ie. `mix()` models),
#' the `mlogit()` function is called in the probabilities and the
#' log-based values are used in the optimization problem.  The
#' probabilities are determined by the `mexpit()` function.
#'
#' @param ... numeric probabilities to convert to log-scale numbers.
#'   These probabilities must add to a number less than 1 and are used
#'   in the mix() estimation algorithm.
#'
#' @inheritParams rootSolve::multiroot
#'
#' @param returnRoot logical; If TRUE, return the root object,
#'   otherwise return the root itself.
#'
#' @return A numeric vector of the log-scale numbers for use in
#'   regressions where the sum of a set of probabilities must add to
#'   be one.
#'
#' @export
#'
#' @author Matthew L. Fidler
#'
#' @examples
#'
#' mlogit(0.1, 0.2, 0.3)
#'
#' mlogit(0.1, 0.2, 0.3, returnRoot = TRUE)
#'
mlogit <- function(..., maxiter = 10000, rtol = 1e-10, atol = 1e-12,
                   ctol = 1e-12, returnRoot=FALSE) {
  rxReq("rootSolve")
  checkmate::assertLogical(returnRoot, len = 1, any.missing = FALSE,
                          null.ok = FALSE)
  .n <- as.numeric(unlist(list(...)))
  checkmate::assertNumeric(.n, any.missing = FALSE, min.len = 1,
                          finite = TRUE, lower = 0, upper=1)
  .s <- sum(.n)
  if (.s >= 1) {
    stop("the sum of the probabilities must be less than 1",
         call. = FALSE)
  }
  if (length(.n) == 1) {
    return(logit(.n))
  }
  f <- function(x) {
    .Call(`_rxode2_mlogit_f`, x, .n)
  }
  j <- function(x) {
    .Call(`_rxode2_mlogit_j`, x)
  }
  .init <- rep(0, length(.n))
  .ret <- try(suppressWarnings(rootSolve::multiroot(f = f, start = .init,
                               jacfunc = j,
                               maxiter = maxiter, rtol = rtol, atol = atol,
                               ctol = ctol)), silent = TRUE)
  if (inherits(.ret, "try-error")) {
    .init <- logit(.init)
    .ret <- try(suppressWarnings(rootSolve::multiroot(f = f, start = .init,
                                                      jacfunc = j,
                                                      maxiter = maxiter, rtol = rtol, atol = atol,
                                                      ctol = ctol)), silent = TRUE)
    if (inherits(.ret, "try-error")) {
      .init <- rep(3, length(.n))
      .ret <- try(suppressWarnings(rootSolve::multiroot(f = f, start = .init,
                                                        jacfunc = j,
                                                        maxiter = maxiter, rtol = rtol, atol = atol,
                                                        ctol = ctol)), silent = TRUE)
      if (inherits(.ret, "try-error")) {
        .init <- rep(-3, length(.n))
        .ret <- try(suppressWarnings(rootSolve::multiroot(f = f, start = .init,
                                                          jacfunc = j,
                                                          maxiter = maxiter, rtol = rtol, atol = atol,
                                                          ctol = ctol)), silent = TRUE)
        if (inherits(.ret, "try-error")) {
          stop("mlogit failed to find a solution.  Please check your input probabilities.",
call. = FALSE)
        }
      }
    }
  }
  if (returnRoot) {
    .ret
  } else {
    .ret$root
  }
}
#' mexpit -- Convert log-scale numbers to probabilities
#'
#' This function converts log-scale numbers to probabilities.
#'
#' The probabilities are calculated using the following equation:
#'
#' \deqn{p_i = \frac{e^{x_i}}{1+\sum_{j=1}^{N-1} e^{x_j}}}
#'
#' This ensures one remaining probability will add to one, that is
#'
#' \deqn{p_N = \frac{1}{1+\sum_{j=1}^{N-1} e^{x_j}}}
#'
#' For the function `dmexpit()`, the element-wise derivatives are calculated; that is, it returns the diagonal of the Jacobian matrix, \eqn{d p_i / d x_i}, not the full Jacobian with off-diagonal terms.
#'
#'
#' @param ...  numeric log-scale numbers to convert to probabilities.
#' @return Probabilities that add up to a number less than 1.
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' m <- mlogit(0.1, 0.2, 0.3)
#' mexpit(m)
#'
#' # derivatives
#' dmexpit(m)
#'
#' p <- mexpit(-3, 0.5, 3)
#' mlogit(p)
#'
#'
mexpit <- function(...) {
  .n <- as.numeric(unlist(list(...)))
  if (length(.n) == 1) {
    return(expit(.n))
  }
  .Call(`_rxode2_mexpit`, .n)
}

#' @rdname mexpit
#' @export
dmexpit <- function(...) {
  .n <- as.numeric(unlist(list(...)))
  .Call(`_rxode2_dmexpit`, .n)
}

rxTest({

  for (i in seq(0.05, 0.95, by=0.05)) {
    test_that(paste0("mlogit", i), {
      expect_equal(mlogit(i), logit(i))
      expect_equal(mexpit(mlogit(i)), i)
    })
  }

  df <- expand.grid(p1=seq(0.05, 0.95, by=0.05), p2=seq(0.05, 0.95, by=0.05))
  df <- df[df$p1+df$p2 < 1, ]
  for (i in seq_along(df$p1)) {
    test_that(paste0("mlogit ", df$p1[i], ", ", df$p2[i]), {
      expect_equal(mexpit(mlogit(df$p1[i], df$p2[i])), c(df$p1[i], df$p2[i]))
    })
  }

  df <- expand.grid(p1=seq(0.05, 0.95, by=0.05), p2=seq(0.05, 0.95, by=0.05),
                    p3=seq(0.05, 0.95, by=0.05))
  df <- df[df$p1+df$p2 + df$p3 < 1, ]
  for (i in seq_along(df$p1)) {
    test_that(paste0("mlogit ", df$p1[i], ", ", df$p2[i], ", ", df$p3[i]), {
      expect_equal(mexpit(mlogit(df$p1[i], df$p2[i], df$p3[i])),
                   c(df$p1[i], df$p2[i], df$p3[i]))
    })
  }

  df <- expand.grid(p1=seq(0.05, 0.95, by=0.05), p2=seq(0.05, 0.95, by=0.05),
                    p3=seq(0.05, 0.95, by=0.05))
  df <- df[df$p1+df$p2 + df$p3 < 1, ]
  for (i in seq_along(df$p1)) {
    ml <- mlogit(df$p1[i], df$p2[i], df$p3[i])
    dml <- dmexpit(ml)
    dml2 <- vapply(seq_along(ml), function(j) {
      exp(ml[j])/(1+sum(exp(ml))) - exp(2*ml[j])/(1+sum(exp(ml)))^2
    }, numeric(1), USE.NAMES=FALSE)
    test_that(paste0("dmexpit ", df$p1[i], ", ", df$p2[i], ", ", df$p3[i]), {
      expect_equal(dml, dml2)
    })
  }
})

#' rxEt vctrs helpers
#'
#' `rxEt` behaves like a data-frame subclass for vctrs/dplyr operations, but it
#' stores its mutable state in an attached environment. These helpers round-trip
#' through a canonical materialized data set and rebuild a valid env-backed
#' object on restore instead of copying stale attributes onto a plain frame.
#'
#' @noRd
#' @keywords internal
#' @importFrom vctrs vec_cast vec_ptype2 vec_proxy vec_restore
NULL

#' Get the meta data for an `rxEt` object
#'
#' @param x `rxEt` object
#' @return A list showing the units, show settings, randomType, and
#'   canResize properties of the `rxEt` object
#' @noRd
#' @author Matthew L. Fidler
.rxEtMeta <- function(x) {
  .meta <- list(
    units = c(dosing = NA_character_, time = NA_character_),
    show = .etDefaultShow(), # nolint
    randomType = NA_integer_,
    canResize = TRUE,
    extraCols = character(0)
  )
  .env <- tryCatch(.rxEtEnv(x), error = function(e) NULL) # nolint
  if (is.environment(.env)) {
    .meta$units <- .env$units
    .meta$show <- .env$show
    .meta$randomType <- .env$randomType
    .meta$canResize <- .env$canResize
    .meta$extraCols <- .etExtraCols(.env) # nolint
  }
  .lst <- attr(attr(x, "class"), ".rxode2.lst", exact = TRUE)
  if (is.na(.meta$randomType) && !is.null(.lst$randomType)) {
    .meta$randomType <- as.integer(.lst$randomType)
  }
  .meta
}
#' Combine the units of two `rxEt` objects, checking for consistency
#'
#' This is used when combining two `rxEt` objects to ensure that their
#' dosing and time units are compatible. If one of the objects has
#' missing or NA units, it will take the units from the other
#' object. If both objects have non-missing units that are not
#' identical, an error will be thrown.
#'
#' @param x The first `rxEt` object (or its meta data)
#'
#' @param y The second `rxEt` object (or its meta data)
#'
#' @param what A string indicating whether we are combining "dose" or
#'   "time" units, used for error messages
#'
#' @return The combined units, which will be the non-missing units if one is
#'  missing, or the common units if both are non-missing and identical
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.rxEtMetaCombineUnits <- function(x, y, what) {
  if (is.null(x) || length(x) == 0L || isTRUE(is.na(x)) || identical(x, "")) {
    return(y)
  }
  if (is.null(y) || length(y) == 0L || isTRUE(is.na(y)) || identical(y, "")) {
    return(x)
  }
  if (!identical(unname(x), unname(y))) {
    if (requireNamespace("units", quietly = TRUE)) {
      .canConvert <- try(units::ud_are_convertible(x, y), silent = TRUE)
      if (inherits(.canConvert, "try-error") || !.canConvert) {
        stop(sprintf("cannot combine rxEt objects with incompatible %s units ('%s' vs '%s')",
                     what, x, y),
             call. = FALSE)
      }
    } else {
      stop(sprintf("cannot combine rxEt objects with different %s units ('%s' vs '%s')",
                   what, x, y),
           call. = FALSE)
    }
  }
  x
}
#' Combine the meta data of two `rxEt` objects
#'
#'
#' @param x The first `rxEt` object (or its meta data)
#'
#' @param y The second `rxEt` object (or its meta data)
#'
#' @return A list containing the combined meta data, where the units
#'   are combined using `.rxEtMetaCombineUnits`, the show settings are
#'   combined using a logical OR, the randomType is the maximum of the
#'   two (ignoring NA), and canResize is TRUE only if both are TRUE.
#'
#' @noRd
#' @author Matthew L. Fidler
.rxEtMetaCombine <- function(x, y) {
  .x <- .rxEtMeta(x)
  .y <- .rxEtMeta(y)
  .show <- .etDefaultShow() # nolint
  .show[names(.x$show)] <- as.logical(.x$show)
  .show[names(.y$show)] <- .show[names(.y$show)] | as.logical(.y$show)
  list(
    units = c(
      dosing = unname(.rxEtMetaCombineUnits(.x$units["dosing"], .y$units["dosing"], "dose")),
      time = unname(.rxEtMetaCombineUnits(.x$units["time"], .y$units["time"], "time"))
    ),
    show = .show,
    randomType = suppressWarnings(max(c(.x$randomType, .y$randomType), na.rm = TRUE)),
    canResize = isTRUE(.x$canResize) && isTRUE(.y$canResize),
    extraCols = unique(c(.x$extraCols, .y$extraCols))
  )
}

#' Make sure the meta data list has all the expected fields and types, filling in
#' defaults where necessary.
#'
#' @param meta A list containing meta data fields such as units, show,
#'   randomType, and canResize.
#'
#' @return normalized meta data list with all expected fields and types
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.rxEtNormalizeMeta <- function(meta) {
  .ret <- list(
    units = c(dosing = NA_character_, time = NA_character_),
    show = .etDefaultShow(), # nolint
    randomType = NA_integer_,
    canResize = TRUE,
    extraCols = character(0)
  )
  if (is.null(meta)) return(.ret)
  if (!is.null(meta$units)) .ret$units[names(meta$units)] <- meta$units
  if (!is.null(meta$show)) {
    .ret$show[names(meta$show)] <- as.logical(meta$show)
  }
  if (!is.null(meta$randomType)) .ret$randomType <- as.integer(meta$randomType)
  if (!is.null(meta$canResize)) .ret$canResize <- isTRUE(meta$canResize)
  if (!is.null(meta$extraCols)) .ret$extraCols <- as.character(meta$extraCols)
  if (is.infinite(.ret$randomType)) .ret$randomType <- NA_integer_
  .ret
}
#' Convert the `rxEt` object to a full data frame with all expected columns.
#'
#' @param x An `rxEt` object or a data frame that can be coerced to one.
#'
#' @return data.frame
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.rxEtAsFullDataFrame <- function(x) {
  if (is.rxEt(x)) { # nolint
    return(as.data.frame(x, all = TRUE))
  }
  as.data.frame(x, stringsAsFactors = FALSE)
}

#' Rebuild the `rxEt` object from a data frame and meta data,
#' restoring the environment and class attributes.
#'
#' @param x A data frame containing the event table data, with columns like id, time, amt, etc.
#' @param to An `rxEt` object or its class attribute, used to extract the meta data for rebuilding.
#' @return An `rxEt` object with the data from `x` and the meta data from `to`.
#' @noRd
#' @author Matthew L. Fidler
.rxEtRebuild <- function(x, to) {
  .meta <- .rxEtNormalizeMeta(.rxEtMeta(to))
  .df <- .rxEtAsFullDataFrame(x)
  .et <- .newRxEt(amountUnits = .meta$units["dosing"], timeUnits = .meta$units["time"]) # nolint
  .env <- .rxEtEnv(.et) # nolint
  if (nrow(.df) > 0L) {
    .env$methods$import.EventTable(.df)
  }
  .env$show[names(.meta$show)] <- .env$show[names(.meta$show)] | .meta$show
  if (!is.na(.meta$randomType)) {
    .env$randomType <- .meta$randomType
  }
  .env$canResize <- .meta$canResize
  .etAddExtraCols(.env, .meta$extraCols) # nolint
  .proxy <- .etMaterialize(.et) # nolint
  attr(.proxy, "class") <- c("rxEt", "data.frame")
  attr(.proxy, ".rxEtEnv") <- .env
  .proxy
}
#' Convert to data.table
#'
#' @param x rxEt object
#' @return data.table
#' @noRd
#' @author Matthew L. Fidler
.rxEtAsDataTable <- function(x) {
  data.table::as.data.table(.rxEtAsFullDataFrame(x))
}
#' Convert to tibble
#'
#' @param x rxEt object
#' @return tibble
#' @noRd
#' @author Matthew L. Fidler
.rxEtAsTibble <- function(x) {
  rxReq("tibble") # nolint
  tibble::as_tibble(.rxEtAsFullDataFrame(x))
}

#' @export
vec_proxy.rxEt <- function(x, ...) {
  .df <- .etMaterialize(x) # nolint
  attr(.df, ".rxEtMeta") <- .rxEtMeta(x)
  .df
}

#' @export
vec_restore.rxEt <- function(x, to, ...) {
  .rxEtRebuild(x, to)
}

#' @export
vec_ptype2.rxEt.rxEt <- function(x, y, ...) {
  .rxEtMetaCombine(x, y)
  vctrs::vec_ptype2(.rxEtAsFullDataFrame(x), .rxEtAsFullDataFrame(y), ...)
}

#' @export
vec_ptype2.rxEt.data.frame <- function(x, y, ...) {
  vctrs::vec_ptype2(.rxEtAsFullDataFrame(x), y, ...)
}

#' @export
vec_ptype2.data.frame.rxEt <- function(x, y, ...) {
  vctrs::vec_ptype2(x, .rxEtAsFullDataFrame(y), ...)
}

#' @export
vec_ptype2.rxEt.data.table <- function(x, y, ...) {
  vctrs::vec_ptype2(.rxEtAsDataTable(x), y, ...)
}

#' @export
vec_ptype2.data.table.rxEt <- function(x, y, ...) {
  vctrs::vec_ptype2(x, .rxEtAsDataTable(y), ...)
}

#' @exportS3Method vctrs::vec_ptype2 rxEt.tbl_df
vec_ptype2.rxEt.tbl_df <- function(x, y, ...) {
  vctrs::vec_ptype2(.rxEtAsTibble(x), y, ...)
}

#' @exportS3Method vctrs::vec_ptype2 tbl_df.rxEt
vec_ptype2.tbl_df.rxEt <- function(x, y, ...) {
  vctrs::vec_ptype2(x, .rxEtAsTibble(y), ...)
}

#' @export
vec_cast.rxEt.rxEt <- function(x, to, ...) {
  .rxEtRebuild(.rxEtAsFullDataFrame(x), to)
}

#' @export
vec_cast.rxEt.data.frame <- function(x, to, ...) {
  .rxEtRebuild(as.data.frame(x, stringsAsFactors = FALSE), to)
}

#' @export
vec_cast.data.frame.rxEt <- function(x, to, ...) {
  vctrs::vec_cast(.rxEtAsFullDataFrame(x), to, ...)
}

#' @export
vec_cast.rxEt.data.table <- function(x, to, ...) {
  .rxEtRebuild(as.data.frame(x), to)
}

#' @export
vec_cast.data.table.rxEt <- function(x, to, ...) {
  vctrs::vec_cast(.rxEtAsDataTable(x), to, ...)
}

#' @exportS3Method vctrs::vec_cast rxEt.tbl_df
vec_cast.rxEt.tbl_df <- function(x, to, ...) {
  .rxEtRebuild(as.data.frame(x), to)
}

#' @exportS3Method vctrs::vec_cast tbl_df.rxEt
vec_cast.tbl_df.rxEt <- function(x, to, ...) {
  vctrs::vec_cast(.rxEtAsTibble(x), to, ...)
}

#' @export
vec_cast.rxEt.tibble <- function(x, to, ...) {
  vec_cast.rxEt.tbl_df(x, to, ...)
}

#' @export
vec_cast.tibble.rxEt <- function(x, to, ...) {
  vec_cast.tbl_df.rxEt(x, to, ...)
}

dplyr_reconstruct.rxEt <- function(data, template) {
  if (inherits(data, "rxEt")) {
    class(data) <- setdiff(class(data), "rxEt")
  }
  data
}

# dplyr keeps the class but drops the marker attributes a display-only
# print needs, so degrade to a plain data frame the way rxEt does.
dplyr_reconstruct.rxEtPreview <- function(data, template) {
  if (inherits(data, "rxEtPreview")) {
    class(data) <- setdiff(class(data), "rxEtPreview")
  }
  data
}

#' @export
print.rxEtTran <- function(x, ...) {
  print(as.data.frame(x))
  .cls <- class(x)
  .lst <- attr(.cls, ".rxode2.lst")
  cat("\nCovariates (non time-varying):\n")
  print(.lst$cov1)
  cat("\nCompartment translation:\n")
  print(data.frame(
    "Compartment Name" = .lst$cmtInfo,
    "Compartment Number" = seq_along(.lst$cmtInfo),
    check.names = FALSE
  ))
}

#' @export
print.rxHidden <- function(x, ...) {
  cat("\r")
}

#' @rdname rxEvid
#' @export
print.rxEvid <- function(x, ...) {
  cat(paste(.colorFmt.rxEvid(x), collapse = "\n"), "\n")
  return(invisible(x))
}

#' @export
print.rxRateDur <- function(x, ...) {
  cat(paste(.colorFmt.rxRateDur(x), collapse = "\n"), "\n")
  return(invisible(x))
}

.h2 <- function(x) {
  cli::cli_text(crayon::bold(paste0(cli::symbol$line, cli::symbol$line, " ", x, " ", cli::symbol$line, cli::symbol$line)))
}

.etPreviewGroupLabel <- function(ids) {
  ids <- sort(unique(as.integer(ids)))
  if (length(ids) == 0L) return("0 individuals")
  if (length(ids) == 1L) return(sprintf("1 individual (id %s)", ids))
  if (all(diff(ids) == 1L)) {
    return(sprintf("%s individuals (ids %s:%s)", length(ids), ids[1], ids[length(ids)]))
  }
  sprintf("%s individuals", length(ids))
}

#' @export
print.rxEtPreview <- function(x, ...) {
  .groups <- attr(x, "rxEtPreviewGroups", exact = TRUE)
  if (length(.groups) > 0L) {
    if (length(.groups) == 1L) {
      cat(sprintf("-- compressed preview for %s --\n", .etPreviewGroupLabel(.groups[[1]]$ids)))
    } else {
      cat(sprintf("-- compressed preview for %s groups --\n", length(.groups)))
      .n <- min(5L, length(.groups))
      for (.i in seq_len(.n)) {
        cat(sprintf("   group %s: %s\n", .i, .etPreviewGroupLabel(.groups[[.i]]$ids)))
      }
      if (length(.groups) > 5L) {
        cat(sprintf("   ... (and %s more groups)\n", length(.groups) - 5L))
      }
    }
  }
  .df <- x
  class(.df) <- "data.frame"
  if (!("id" %in% names(.df)) && length(.groups) > 0L) {
    .idList <- lapply(.groups, function(.g) {
      .formattedId <- if (length(.g$ids) == 1L) as.integer(.g$ids) else .formatIds(.g$ids)
      rep(.formattedId, .g$nRow)
    })
    .idCol <- unlist(.idList)
    if (length(.idCol) == nrow(.df)) {
      .df <- cbind(id = .idCol, .df)
    }
  }
  .show <- attr(x, "rxEtShow", exact = TRUE)
  if (!is.null(.show)) {
    .df <- .df[, .etKeepCols(names(.df), # nolint
                             attr(x, "rxEtMarkedCols", exact = TRUE), .show,
                             attr(x, "rxEtExtraCols", exact = TRUE)),
               drop = FALSE]
  }
  print(tibble::as_tibble(.df), ...)
  invisible(x)
}

#' @export
print.rxEt <- function(x, ...) {
  if (.isRxEt(x)) {
    bound <- .getBound(x, parent.frame(2))
    .et1 <- paste0("EventTable with ", x$nobs + x$ndose, " records")
    .et2 <- NULL
    .units <- x$.units
    .maxId <- length(x$ids)
    if (.maxId != 1) {
      .et2 <- sprintf("   %s individuals", .maxId)
    }
    .et3 <- sprintf(
      "   %s dosing records (see %s$%s(); add with %s or %s)",
      x$ndose, bound, "get.dosing", "add.dosing", "et"
    )
    .et4 <- sprintf(
      "   %s observation times (see %s$%s(); add with %s or %s)",
      x$nobs, bound, "get.sampling", "add.sampling",
      "et"
    )
    .et5 <- NULL
    if (x$show["addl"]) {
      .et5 <- sprintf(
        "   multiple doses in `addl` columns, expand with %s$%s(); or %s(%s)",
        bound, "expand", "etExpand", bound
      )
    }
    .et <- c(.et2, .et3, .et4, .et5)
    .df <- data.frame(et = .et, stringsAsFactors = FALSE)
    names(.df) <- .et1
    class(.df) <- c(
      sprintf("EventTable Info: %s", bound),
      "paged_df", "data.frame"
    )
    .out <- utils::capture.output({
      print(.df)
    })
    .nb <- TRUE
    if (length(.out) > 0) {
      .nb <- FALSE
      cat(cli::cli_format_method({
        .h2(.et1)
        cli::cli_text(sprintf(
          "   %s dosing records (see %s$%s(); add with %s or %s)\n",
          x$ndose, crayon::yellow(bound), crayon::blue("get.dosing"),
          crayon::blue("add.dosing"), crayon::blue("et")))
        cli::cli_text(sprintf(
          "   %s observation times (see %s$%s(); add with %s or %s)\n",
          x$nobs, crayon::yellow(bound), crayon::blue("get.sampling"),
          crayon::blue("add.sampling"), crayon::blue("et")))
        if (x$show["addl"]) {
          cli::cli_text(sprintf(
            "   multiple doses in `addl` columns, expand with %s$%s(); or %s(%s)\n",
            crayon::yellow(bound), crayon::blue("expand"),
            crayon::blue("etExpand"), crayon::yellow(bound)
          ))
        }
      }), sep = "\n")


    }
    if (x$nobs != 0 || x$ndose != 0) {
      if (!.nb) {
        cat(cli::cli_format_method({
          .h2(paste0("First part of ", crayon::yellow(bound), ":"))
        }), sep = "\n")
      }
      .preview <- .etPreviewData(.rxEtEnv(x), "all")
      # re-tagging columns is a `[[<-`, which drops the display marking
      .previewCls <- class(.preview)
      if (!is.null(.preview[["evid"]])) {
        class(.preview[["evid"]]) <- c("rxEvid", class(.preview[["evid"]]))
      }
      for (.nm in c("rate", "dur")) {
        if (!is.null(.preview[[.nm]])) {
          class(.preview[[.nm]]) <- c("rxRateDur", class(.preview[[.nm]]))
        }
      }
      class(.preview) <- .previewCls
      .env <- .rxEtEnv(x)
      .showCols <- .etDisplayCols(names(.preview), .env$show, .etExtraCols(.env)) # nolint
      if (inherits(.preview, "rxEtPreview") && .maxId > 1L) {
        print(.preview)
      } else {
        .df <- .preview[, .showCols, drop = FALSE]
        class(.df) <- "data.frame"
        print(tibble::as_tibble(.df))
      }
    }
    invisible(x)
  } else {
    print.data.frame(x)
  }
}

#' Print information about the rxode2 object.
#'
#' This prints the model name and its status for being able to be solved
#'
#' @param x An rxode2 object
#' @param ... Ignored parameters
#' @author Matthew L.Fidler
#' @return original object
#' @export
#' @keywords internal
print.rxode2 <- function(x, ...) {
  rxModelVars(x)
  x <- .getReal(x)
  .bound <- .getBound(x, parent.frame(2))
  .valid <- x$isValid()
  .msg2 <- ""
  if (!.valid) {
    .msg <- crayon::red$bold("invalid")
    .msg2 <- paste0(" re-create with ", crayon::blue("rxode2::"), crayon::yellow("rxode2"))
    .ico <- crayon::red(cli::symbol$cross)
  } else {
    .loaded <- x$isLoaded()
    if (.loaded) {
      .msg <- crayon::green$bold("ready")
      .ico <- crayon::green(cli::symbol$tick)
    } else {
      .msg <- crayon::yellow$bold("unloaded")
      .ico <- crayon::yellow(cli::symbol$warning)
      .msg2 <- paste0(" reload with ", crayon::blue("rxode2::"), crayon::yellow("rxLoad"))
    }
  }
  if (.useUtf()) {
    .ico <- paste0(.ico, " ")
  } else {
    .ico <- ""
  }
  .dll <- getOption("rxode2.basename.print", basename(rxode2::rxDll(x)))
  .env <- attr(x, ".env")
  .pkg <- ""
  .new <- ""
  if (!is.null(x$package)) {
    .dll <- .bound
    .pkg <- crayon::blue(paste0(x$package, "::"))
    if (regexpr("_new", .dll) != -1) {
      .dll <- gsub("_new", "", .dll)
      .new <- " (updated)"
    }
  } else {
    .dll <- substr(.dll, 1, nchar(.dll) - nchar(.Platform$dynlib.ext) - 1)
  }
  cat(paste0(
    crayon::bold("rxode2 "), .rxVersion["version"], " model named ", .pkg,
    crayon::yellow$bold(getOption("rxode2.dll.print", .dll)), .new, " model (", .ico, .msg,
    .msg2, ")."
  ), "\n")
  .indLin <- rxModelVars(x)$indLin
  if (length(.indLin) > 0) {
    cat(crayon::bold("  indLin: "))
    if (.indLin[["fullIndLin"]]) {
      if (is.null(.indLin[["f"]])) {
        cat(paste0(crayon::yellow("homogenous matrix exponential + Inductive Linearization")))
      } else {
        cat(paste0(crayon::red("in"), crayon::yellow("homogenous matrix exponential + Inductive Linearization")))
      }
    } else {
      if (is.null(.indLin[["f"]])) {
        cat(crayon::yellow("homogenous matrix exponential"))
      } else {
        cat(paste0(crayon::red("in"), crayon::yellow("homogenous matrix exponential")))
      }
    }
    cat("\n")
  }
  if (!any(names(list(...)) == "rxSuppress") && .valid) {
    .cur <- rxode2::rxState(x)
    if (length(.cur) > 0) {
      cat(paste0(crayon::yellow(.bound), crayon::blue$bold("$state"), ": ", paste(.cur, collapse = ", "), "\n"))
    }
    .cur <- x$stateExtra
    if (length(.cur) > 0) {
      cat(paste0(crayon::yellow(.bound), crayon::blue$bold("$stateExtra"), ": ", paste(.cur, collapse = ", "), "\n"))
    }
    .cur <- rxode2::rxParams(x)
    if (length(.cur) > 0) {
      cat(paste0(crayon::yellow(.bound), crayon::blue$bold("$params"), ": ", paste(.cur, collapse = ", "), "\n"))
    }
    .cur <- rxode2::rxLhs(x)
    if (length(.cur) > 0) {
      cat(paste0(crayon::yellow(.bound), crayon::blue$bold("$lhs"), ": ", paste(.cur, collapse = ", "), "\n"))
    }
  }
  invisible(x)
}

#' Print the rxCoef object
#'
#' This prints out the user supplied arguments for rxCoef object
#'
#' @param x rxCoef object
#'
#' @keywords internal
#' @author Matthew L.Fidler
#' @return original object
#' @export
print.rxCoef <- function(x, ...) {
  .rxDllObj <- x$rxode2
  if (length(rxParams(.rxDllObj)) > 0) {
    cat(cli::cli_format_method({
      .h2("User supplied parameters:")
    }), "\n")
    print(rxode2::rxInits(.rxDllObj, NULL, rxode2::rxParams(.rxDllObj), NA, TRUE))
    cat(cli::cli_format_method({
      .h2("User initial conditions:")
    }), "\n")
    .tmp <- rxode2::rxInits(.rxDllObj, NULL, rxode2::rxState(.rxDllObj), 0, TRUE)
    if (length(x$sens) > 0) {
      .tmp <- .tmp[regexpr(getFromNamespace("regSens", "rxode2"), names(.tmp)) == -1]
    }
    print(.tmp)
  }
  cat(cli::cli_format_method({
    .h2("Compartments:")
  }), "\n")
  .tmp <- rxode2::rxState(.rxDllObj)
  if (length(.tmp) > 0) {
    names(.tmp) <- paste0("cmt=", seq_along(.tmp))
    if (length(x$sens) > 0) {
      .tmp1 <- .tmp[regexpr(getFromNamespace("regSens", "rxode2"), .tmp) == -1]
      print(.tmp1)
      cat(cli::cli_format_method({
        .h2("Sensitivities:")
      }), "\n")
      .tmp2 <- gsub(
        getFromNamespace("regSens", "rxode2"), "d/dt(d(\\1)/d(\\2))",
        .tmp[regexpr(regSens, .tmp) != -1]
      )
      print(.tmp2)
    } else {
      print(.tmp)
    }
  } else {
    cat("No ODEs in this DLL.\n")
  }
  return(invisible())
}

#' @export
print.rxC <- function(x, ...) {
  cat(sprintf("C file: %s  ('summary' for code)\n", getOption("rxode2.c.print", x)))
}


#' Print rxDll object
#'
#' This tells if the rxDll is loaded, ready and/or deleted.
#'
#' @keywords internal
#' @author Matthew L.Fidler
#' @return original object
#' @export
print.rxDll <- function(x, ...) {
  if (file.exists(x$dll)) {
    cat(sprintf("rxode2 DLL named \"%s\"", getOption("rxode2.basename.print", basename(x$dll))))
    if (rxDllLoaded(x)) {
      cat(" is loaded and ready to use.\n")
    } else {
      cat(" is not loaded now.\n")
    }
  } else {
    cat(sprintf("rxode2 DLL named \"%s\" has been deleted.\n", getOption("rxode2.basename.print", basename(x$dll))))
  }
  invisible(x)
}



#' @export
print.rxSolveCovs <- function(x, ...) {
  .args <- as.list(match.call(expand.dots = TRUE))
  if (any(names(.args) == "bound")) {
    .bound <- .args$bound
  } else {
    .bound <- .getBound(x, parent.frame(2))
  }
  NextMethod()
}

#' @export
print.boundInits <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
}

#' @export
print.rxSolveInits <- function(x, ...) {
  .args <- as.list(match.call(expand.dots = TRUE))
  if (any(names(.args) == "bound")) {
    .bound <- .args$bound
  } else {
    .bound <- .getBound(x, parent.frame(2))
  }
  print(structure(.bound, class = "boundInits"), ...)
  .df <- x$inits
  print(.df)
  NextMethod()
}

#' @export
print.rxSolveSimType <- function(x, ...) {
  if (any(names(x) == "sim.id")) {
    cat(format(x, ...), sep = "\n")
  } else if (!is.null(x$omegaList)) {
    cat(format(x, ...), sep = "\n")
  } else if (!is.null(x$sigmaList)) {
    cat(format(x, ...), sep = "\n")
  } else if (!is.null(x$thetaMat)) {
    cat(format(x, ...), sep = "\n")
  }
}


#' @export
print.rxSolve <- function(x, ...) {
  if (rxIs(x, "rxSolve")) {
    .nb <- TRUE
    .args <- as.list(match.call(expand.dots = TRUE))
    if (any(names(.args) == "n")) {
      .n <- .args$n
    } else {
      .n <- 6L
    }
    if (any(names(.args) == "width")) {
      .width <- .args$width
    } else {
      .width <- NULL
    }
    if (any(names(.args) == "bound")) {
      .bound <- .args$bound
    } else {
      .bound <- .getBound(x, parent.frame(2))
      assignInMyNamespace(".getBoundRemember", .bound)
      on.exit({
        assignInMyNamespace(".getBoundRemember", NULL)
      }, add=TRUE)
    }
    if (.nb) {
      .df <- x$pars
      if (rxIs(.df, "data.frame")) {
        .cls <- c(
          paste0("Parameters ", .bound, "$params"),
          "paged_df", "data.frame"
        )
        class(.df) <- .cls
        .out <- utils::capture.output({
          print(.df)
        })
        if (length(.out) > 0) .nb <- FALSE
      }
    }
    if (.nb) {
      .df <- x$inits
      .df <- as.data.frame(t(x$inits))
      .cls <- c(
        paste0("Initial\u00A0State ", .bound, "$inits"),
        "paged_df", "data.frame"
      )
      class(.df) <- .cls
      .out <- utils::capture.output({
        print(.df)
      })
      if (length(.out) > 0) .nb <- FALSE
    }
    if (.nb) {
      print.rxSolveSimType(x)
      .df <- x
      .cls <- c(
        paste0("Solved\u00A0Data: ", .bound),
        "paged_df", "data.frame"
      )
      class(.df) <- .cls
      print(.df)
      return(invisible(x))
    } else {
      .summary <- any(names(.args) == ".summary")
      if (!.summary) {
        cat(cli::cli_format_method({
          .h2(crayon::bold("Solved rxode2 object"))
        }), sep = "\n")
      }
      NextMethod()
      if (.summary) {
        cat(cli::cli_format_method({
          .h2(crayon::bold("Summary of data (object):"))
        }), sep = "\n")
        print(summary.data.frame(x))
      } else {
        cat(cli::cli_format_method({
          .h2(crayon::bold("First part of data (object):"))
        }), sep = "\n")
        .isDplyr <- requireNamespace("tibble", quietly = TRUE) &&
          getOption("rxode2.display.tbl", TRUE)
        if (!.isDplyr) {
          print(head(as.data.frame(x), n = .n))
        } else {
          print(tibble::as_tibble(x), n = .n, width = .width)
        }
      }
    }
  } else {
    print.data.frame(x)
  }
}

#' @export
print.rxModelText <- function(x, ...) {
  .args <- as.list(match.call(expand.dots = TRUE))
  .summary <- any(names(.args) == ".summary")
  if (any(names(.args) == "bound")) {
    .bound <- .args$bound
  } else {
    .bound <- .getBound(x, parent.frame(2))
  }
  .code <- deparse(body(eval(parse(text = paste("function() {", as.vector(x), "}")))))
  .code[1] <- "rxode2({"
  .code[length(.code)] <- "})"
  if (.summary) {
    cat(cli::cli_format_method({
      .h2(.fmt3("Model", .bound, "model"))
    }), sep = "\n")
  } else {
    cat(cli::cli_format_method({
      .h2(crayon::bold("rxode2 Model Syntax"))
    }), sep = "\n")
  }
  cat(paste(.code, collapse = "\n"), "\n")
}

#' @export
print.boundParams <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
}

#' @export
print.rxSolveParams <- function(x, ..., n = 0L) {
  .args <- as.list(match.call(expand.dots = TRUE))
  if (any(names(.args) == "bound")) {
    .bound <- .args$bound
  } else {
    .bound <- .getBound(x, parent.frame(2))
  }
  class(.bound) <- "boundParams"
  .df <- x$.params.single
  if (length(.df) > 0) {
    print(.bound, ...)
    print(.df, ...)
  } else {
    .df <- x$pars
    if (rxIs(.df, "data.frame")) {
      print(.bound, ...)
      print(tibble::as_tibble(.df), ...)
    }
  }
  NextMethod()
}

#' @export
print.rxSymInvCholEnv <- function(x, ...) {
  if (is.null(x$theta)) {
    cat(sprintf("Uninitialized $theta, please assign (requires %s arguments)!\n", x$ntheta))
  } else {
    cat(sprintf("$theta=c(%s) for:\n\n", paste(x$theta, collapse = ", ")))
    print(x$invobj$fmat)
    cat("\nThis allows accessing $omegaInv, $omega, etc. For a full list see str(.)\n")
  }
}

#' @inherit base::print
#' @return This returns invisibly the model variables object
#' @export
print.rxModelVars <- function(x, ...) {
  .bound <- .getBound(x, parent.frame(2))
  cat("rxode2 model variables (see str to see all variables)\n")
  .cur <- x$state
  if (length(.cur) > 0) {
    cat(paste0(crayon::yellow(.bound), crayon::blue$bold("$state"), ": ", paste(.cur, collapse = ", "), "\n"))
  }
  .cur <- x$stateExtra
  if (length(.cur) > 0) {
    cat(paste0(crayon::yellow(.bound), crayon::blue$bold("$stateExtra"), ": ", paste(.cur, collapse = ", "), "\n"))
  }
  .cur <- x$params
  if (length(.cur) > 0) {
    cat(paste0(crayon::yellow(.bound), crayon::blue$bold("$params"), ": ", paste(.cur, collapse = ", "), "\n"))
  }
  .cur <- x$lhs
  if (length(.cur) > 0) {
    cat(paste0(crayon::yellow(.bound), crayon::blue$bold("$lhs"), ": ", paste(.cur, collapse = ", "), "\n"))
  }
  invisible(x)
}

#' Softplus Activation Function
#'
#' @param x numeric vector
#' @return numeric vector
#' @family Activation Functions
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' softplus(c(-1, 0, 1, 2))
#'
#' \donttest{
#' # You can use rxode2 too:
#'
#' r <- rxode2({
#'  s <- softplus(time)
#' })
#'
#' e <- et(c(-1, 0, 1, 2))
#'
#' rxSolve(r, e)
#' }
#'
softplus <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 3L)
}

#' Default Softplus Activation Function
#'
#' @param x numeric vector
#' @return numeric vector
#' @family Activation Functions
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' dsoftplus(c(-1, 0, 1, 2))
#'
#' \donttest{
#' # You can use rxode2 too:
#'
#' r <- rxode2({
#'  s1 <- dsoftplus(time)
#' })
#'
#' e <- et(c(-1, 0, 1, 2))
#'
#' rxSolve(r, e)
#' }
#'
dsoftplus <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 13L)
}

#' @rdname dsoftplus
d2softplus <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 14L)
}

#' @rdname dsoftplus
d3softplus <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 15L)
}

#' @rdname dsoftplus
d4softplus <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 16L)
}

#' Assign the model block in the rxode2 related object
#'
#' @param x rxode2 related object
#' @param envir Environment where assignment occurs
#' @param value Value of the object
#' @return rxode2 related object
#' @export
#' @author Matthew L. Fidler
`model<-` <- function(x, envir=environment(x), value) {
  UseMethod("model<-")
}
#' @export
`model<-.default` <- function(x, envir=environment(x), value) {
  .ret <- try(as.rxUi(x), silent = TRUE)
  if (inherits(.ret, "try-error")) {
    stop("cannot figure out what to do with model assignment", call.=FALSE)
  }
  .model <- as.model(force(value))
  .ini <- .ret$iniFun
  .fun <- function() {} # nolint
  body(.fun) <- as.call(list(quote(`{`), .ini, .model))
  rxode2(x) <- .fun
  x
}

#' Assign the ini block in the rxode2 related object
#'
#' @param x rxode2 related object
#' @param envir Environment where assignment occurs
#' @param value Value of the object
#' @return rxode2 related object
#' @export
#' @author Matthew L. Fidler
`ini<-` <- function(x, envir=environment(x), value) {
  UseMethod("ini<-")
}
#' @export
`ini<-.default` <- function(x, envir=environment(x), value) {
  .ret <- try(as.rxUi(x), silent = TRUE)
  if (inherits(.ret, "try-error")) {
    stop("cannot figure out what to do with ini assignment", call.=FALSE)
  }
  .ini <- as.ini(force(value))
  .model <- .ret$modelFun
  .fun <- function() {} # nolint
  body(.fun) <- as.call(list(quote(`{`), .ini, .model))
  rxode2(x) <- .fun
  x
}
#' This gets all the significant items in the model
#'
#' @param model uncompresseg model to check
#' @return all significant items to assign
#' @noRd
#' @author Matthew L. Fidler
.getAllSigEnv <- function(model) {
  .lsModel <- ls(envir=model)
  setdiff(.lsModel, .rxUiBlessed)
}

#' This gets the dropped items if a significant item changed
#'
#' @param model uncompressed model to check
#' @return Character vector of items to be dropped
#' @noRd
#' @author Matthew L. Fidler
.getDropEnv <- function(model) {
  .lsModel <- ls(envir=model)
  setdiff(.lsModel, c(.rxUiBlessed, model$sticky))
}
#' This gets the additional items kept if a significant item changed
#'
#' @param model uncompressed model to check
#' @return Character vector of additional items to be kept
#' @noRd
#' @author Matthew L. Fidler
.getKeepEnv <- function(model) {
  .lsModel <- ls(envir=model)
  .ret <- setdiff(.lsModel, .rxUiBlessed)
  .ret[.ret %in% model$sticky]
}
#' Checks to see if the models are nearly the same
#'
#' @param newModel New rxode2 model
#' @param oldModel old rxode2 model
#' @return boolean to say if the models are nearly the same
#' @noRd
#' @details
#'
#' Nearly the same models has the same model block, the same
#' estimates, but can change in fixed/unfixed, upper/lower limits and
#' labels.
#'
#' @author Matthew L. Fidler
.modelsNearlySame <- function(newModel, oldModel) {
  # first check to see if this is a "significant" change.
  if (identical(newModel$modelFun, oldModel$modelFun)) {
    .pre <- oldModel$iniDf[,c("name", "est")]
    names(.pre)[2] <- "estPre"
    .post <- newModel$iniDf[,c("name", "est")]
    .both <- merge(.pre, .post, all.x=TRUE, all.y=TRUE, by="name")
    if (any(is.na(.both$est))) return(FALSE)
    if (any(is.na(.both$estPre))) return(FALSE)
    if (all(abs(.both$est - .both$estPre) < 1e-10)) {
      return(TRUE)
    }
  }
  FALSE
}
#' Adjust new and old model
#'
#' @param newModel new model to adjust (uncompressed)
#' @param oldModel old model to get information from (uncompressed)
#' @param rename if this operation is a rename operation, it is
#'   considered an insignificant operation
#' @return new model adjusted to match oldModel as much as reasonable.
#'   The class needs to be adjusted and needs to be compressed
#' @noRd
#' @author Matthew L. Fidler
.newModelAdjust <- function(newModel, oldModel, rename=FALSE) {
  newModel <- rxUiDecompress(newModel)
  oldModel <- rxUiDecompress(oldModel)
  lapply(c("meta", "sticky", "model", "modelName"), function(x) {
    if (exists(x, envir=oldModel)) {
      if (x == "meta") {
        # meta must be forked, not shared by reference: the simulation-model
        # cache (`.simModelBase`) lives in meta, so a shared env lets either
        # model pick up the other's cached simulation model after piping.
        # This is the invariant rxUiGet.simulationModel documents and
        # .copyUi() already follows.  The fork is shallow (see
        # .copyMetaForPiping) so metadata carried over keeps the reference
        # semantics it had when meta was shared.
        assign(x, .copyMetaForPiping(get(x, envir=oldModel)), envir=newModel)
      } else {
        assign(x, get(x, envir=oldModel), envir=newModel)
      }
    }
  })
  if (rename || .modelsNearlySame(newModel, oldModel)) {
    lapply(.getAllSigEnv(oldModel),
           function(x) {
             assign(x, get(x, envir=oldModel), envir=newModel)
           })
    return(newModel)
  }
  .drop <- .getDropEnv(oldModel)
  .keep <- .getKeepEnv(oldModel)
  lapply(.keep, function(v) {
    assign(v, get(v, envir=oldModel), envir=newModel)
  })
  lapply(.drop, function(v) {
    if (exists(v, envir=newModel)) {
      rm(list=v, envir=newModel)
    }
  })
  if (length(.drop) > 0) {
    cli::cli_alert("significant model change detected")
    if (length(.keep) > 0) {
      cli::cli_alert(sprintf("kept in model: '%s'",
                             paste(paste0("$", .keep), collapse="', '")))
    }
    cli::cli_alert(sprintf("removed from model: '%s'",
                           paste(paste0("$", .drop), collapse="', '")))
  }
  newModel
}

#' Set the body of the rxode2 related function
#'
#' @param fun function for setting the body
#'
#' @param envir environment where this is assigned
#'
#' @param value value that will be assigned
#'
#' @return new rxode2 function with meta information retained
#'
#' @noRd
#' @author Matthew L. Fidler & Bill Denney
.bodySetRxUi <- function(x, envir = parent.frame(), value) {
  if (is.function(value)) {
    value <- body(value)
  }
  .clsModel <- class(x)
  .model <- rxUiDecompress(x)
  .modelFun <- .model$fun # don't use as-function to avoid environment issues
  if (!inherits(.modelFun, "function")) stop("wrong input for 'x' in .bodySetRxUi", call.=FALSE)
  body(.modelFun) <- value
  .modelFun()
}

#' Set the function body of an rxUi object while retaining other object
#' information (like data)
#'
#' @param x The rxUi object
#' @param envir environment where the assignment ocurs
#' @param value the value that will be assigned
#' @return The rxode2 ui/function
#' @export
#' @examples
#'
#' one.compartment <- function() {
#'   ini({
#'     tka <- log(1.57); label("Ka")
#'     tcl <- log(2.72); label("Cl")
#'     tv <- log(31.5); label("V")
#'     eta.ka ~ 0.6
#'     eta.cl ~ 0.3
#'     eta.v ~ 0.1
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl + eta.cl)
#'     v <- exp(tv + eta.v)
#'     d/dt(depot) = -ka * depot
#'     d/dt(center) = ka * depot - cl / v * center
#'     cp = center / v
#'     cp ~ add(add.sd)
#'   })
#' }
#'
#' two.compartment <- function() {
#'   ini({
#'     lka <- 0.45 ; label("Absorption rate (Ka)")
#'     lcl <- 1 ; label("Clearance (CL)")
#'     lvc  <- 3 ; label("Central volume of distribution (V)")
#'     lvp  <- 5 ; label("Peripheral volume of distribution (Vp)")
#'     lq  <- 0.1 ; label("Intercompartmental clearance (Q)")
#'     propSd <- 0.5 ; label("Proportional residual error (fraction)")
#'   })
#'   model({
#'     ka <- exp(lka)
#'     cl <- exp(lcl)
#'     vc <- exp(lvc)
#'     vp <- exp(lvp)
#'     q  <- exp(lq)
#'     kel <- cl/vc
#'     k12 <- q/vc
#'     k21 <- q/vp
#'     d/dt(depot) <- -ka*depot
#'     d/dt(central) <-  ka*depot - kel*central - k12*central + k21*peripheral1
#'     d/dt(peripheral1) <- k12*central - k21*peripheral1
#'     cp <- central / vc
#'     cp ~ prop(propSd)
#'   })
#' }
#'
#' ui <- rxode2(one.compartment)
#'
#' rxode2(ui) <- two.compartment
#'
`rxode2<-` <- function(x, envir=environment(x), value) {
  UseMethod("rxode2<-")
}
#' @rdname rxode2-set
#' @export
`rxode2<-.function` <- function(x, envir=environment(x), value) {
  .val <- force(value)
  if (inherits(.val, "{")) {
    .fun <- function() {} #nolint
    body(.fun) <- .val
    return(.fun)
  } else if (!inherits(value, "function")) {
    stop("cannot figure out how to assign this to the with rxode()<-",
         call.=FALSE)
  }
  return(force(value))
}

#' @rdname rxode2-set
#' @export
`rxode2<-.default` <- function(x, envir=environment(x), value) {
  force(value)
  .v <- value
  if (inherits(value, "function")) {
    value <- body(value)
  } else if (inherits(value, "rxUi")) {
    value <- body(as.function(value))
  } else if (!inherits(value, "{")) {
    stop("do not know how to assign this", call.=FALSE)
  }
  .ret <- .bodySetRxUi(x, envir = parent.frame(), value)
  if (inherits(x, "rxUi")) {
    .ret <- .newModelAdjust(.ret, x)
  }
  .ret <- rxUiCompress(.ret)
  if (inherits(x, "rxUi")) {
    .cls <- setdiff(class(x), class(.ret))
    if (length(.cls) > 0) {
      class(.ret) <- c(.cls, class(.ret))
    }
  }
  .ret
}

#'@rdname rxode2-set
#'@export
`rxode<-` <- function(x, envir=environment(x), value) {
  UseMethod("rxode2<-")
}

#'@rdname rxode2-set
#'@export
`RxODE<-` <- function(x, envir=environment(x), value) {
  UseMethod("rxode2<-")
}

#' @export
`$<-.rxUi` <- function(x, name, value) {
  .raw <- is.list(x) || inherits(x, "raw")
  if (!.raw) {
    assign(name, value, envir=x)
    return(x)
  }
  .x <- x
  if (name %in% c("ini", "iniDf")) {
    ini(x) <- value
    return(x)
  }
  if (name == "model") {
    model(x) <- value
    return(x)
  }
  .x <- rxUiDecompress(.x)
  if (exists(name, .x)) {
    stop("'", name, "' is a fixed UI component and should not be overwritten",
         call.=FALSE)
  }
  .meta <- get("meta", .x)
  assign(name, value, envir=.meta)
  .x <- rxUiCompress(.x)
  .x
}

## This is a list of supported distributions with the number of arguments they currently support.
.errDist <- list(
  "dpois" = 1,
  "pois" = 1,
  "dbinom" = 1:2,
  "binom"=1:2,
  "dbern" = 1,
  "bern" = 1,
  "dbeta" = 2, # non-central isn't supported by stan so drop support
  "beta" = 2,
  "dt" = 1, # non-central isn't supported by stan, so drop
  "t" = 1,
  ##
  ## "dnbinom"=2:3,  ## dnbinom is in R; FIXME: how does ot compare to dneg_binomial
  ## "dneg_binomial", ## not in base R (but in glnmm2)
  ##
  ## Available as external package http://ugrad.stat.ubc.ca/R/library/rmutil/html/BetaBinom.html
  ## "dbetabinomial", ## not in base R (but in glnmm2)
  "add" = 1,
  "norm" = 0,
  "dnorm" = 0,
  "prop" = 1,
  "propT" = 1,
  "propF" = 2,
  "pow" = 2,
  "powT" = 2,
  "powF"=3,
  "tbs" = 1,
  "boxCox" = 1,
  "tbsYj" = 1,
  "yeoJohnson" = 1,
  "logn" = 1,
  "lnorm" = 1,
  "dlnorm" = 1,
  "dlogn" = 1,
  "logitNorm" = 1:3,
  "probitNorm" = 1:3,
  "combined1"=0,
  "combined2"=0,
  "var"=0,
  "dv"=0,
  "ar"=1,
  "comb1"=0,
  "comb2"=0,
  "dchisq"=1,
  "chisq"=1,
  "dexp"=0:1,
  "df"=2:3,
  "f"=2:3,
  "dgeom"=1,
  "geom"=1,
  #  "dhyper"=3,
  #  "hyper"=3,
  "dunif"=0:2,
  "unif"=0:2,
  "dweibull"=1:2,
  "weibull"=1:2,
  "cauchy"= 0,
  "dcauchy"= 0:2,
  "dgamma"=1:2,
  "nbinom"=2,
  "dnbinom"=2,
  "nbinomMu"=2,
  "dnbinomMu"=2
)

.errDistsPositive <- c("add", "prop", "propT", "pow", "powT", "logn", "dlogn", "lnorm",
                       "dlnorm", "logitNorm", "probitNorm")

.errUnsupportedDists <- "nlmixrDist"

.errAddDists <- c("add", "prop", "propT", "propF", "norm", "pow", "powT", "powF",
                  "dnorm", "logn", "lnorm", "dlnorm", "tbs", "tbsYj", "boxCox",
                  "yeoJohnson", "logitNorm", "probitNorm", "combined1", "combined2",
                  "comb1", "comb2", "t", "cauchy", "norm", "var", "dv", "ar")

.errIdenticalDists <- list(
  "lnorm"=c("logn", "dlogn", "dlnorm"),
  "dnorm"="norm",
  "boxCox"="tbs",
  "yeoJohnson"="tbsYj",
  "pois"="dpois",
  "binom"="dbinom",
  "bern"="dbern",
  "beta"="dbeta",
  "t"="dt",
  "combined1"="comb1",
  "combined2"="comb2",
  "chisq"="dchisq", #6
  "f"="df", #8
  "geom"="dgeom", #9
  "unif"="dunif", #10
  "weibull"="dweibull", #11
  "cauchy"="dcauchy", #12
  "nbinom"="dnbinom",
  "nbinomMu"="dnbinomMu")

.errDistArgRanges <- list(
  "add"=c(0, Inf),
  "prop"=c(0, Inf),
  "propT"=c(0, Inf),
  "propF"=c(0, Inf),
  "pow"=c(0, Inf),
  "powT"=c(0, Inf),
  "powF"=c(0, Inf),
  "pow2"=c(-Inf, Inf),
  "powT2"=c(-Inf, Inf),
  "powF2"=c(-Inf, Inf),
  "lnorm"=c(0, Inf),
  "boxCox"=c(-Inf, Inf),
  "yeoJohnson"=c(-Inf, Inf),
  "pois"=c(0, Inf),
  "binom"=c(0, Inf),
  "binom2"=c(0, 1),
  "nbinom"=c(0, Inf),
  "nbinom2"=c(0, 1),
  "nbinomMu"=c(0, Inf),
  "nbinomMu2"=c(0, Inf),
  "bern"=c(0, 1),
  "logitNorm"=c(0, Inf),
  "probitNorm"=c(0, Inf),
  "chisq"=c(0, Inf),
  "chisq2"=c(0, Inf),
  "dexp"=c(0, Inf),
  "f"=c(0, Inf),
  "f2"=c(0, Inf),
  "f3"=c(0, Inf),
  "geom"=c(0, 1),
  ## "hyper"=c(0, Inf),
  ## "hyper2"=c(0, Inf),
  ## "hyper3"=c(0, Inf),
  "dunif"=c(-Inf, Inf),
  "dunif2"=c(-Inf, Inf),
  "weibull"=c(0, Inf),
  "weibull2"=c(0, Inf),
  "cauchy"=c(-Inf, Inf),
  "cauchy2"=c(0, Inf),
  "dgamma"=c(0, Inf),
  "dgamma2"=c(0, Inf),
  "geom"=c(0, 1),
  "beta" = c(0, Inf),
  "beta2"=c(0, Inf),
  "t"=c(0, Inf),
  "t2"=c(0, Inf),
  "ar"=c(0, 1)
)

## the desired outcome for each expression is to capture the condition
## when the multiple endpoint occurs, the lower and upper for the
## transformation (if any) and the error type of the model.  The
## errors should be labeled "add" for an additive error and pow, pow2
## for the first and second argument of the power distribution.  The
## ini block should have already been processed by lotri, so we have
## some information about the model.  The predictions will be replaced
## by rxPred__# for easy replacement with the correct information for
## simulating or estimating in different nlmixr algorithms and
## different rxode2 simulation routines

## Currently you can support the following types of expressions:
##
## f ~ add(add.sd)
##
## This assumes a named DVID of f
##
## or
##
## g ~ add(add.sd) | namedDvid
##
## In this case the named DVID is namedDvid
##
## In the final rxode2 model expression these expressions will look like:
##
## rxPred__1 = f
## rxPred__2 = g
##
## With this change, there is sufficient information to send to the
## mu-referencing routine


#' Change distribution name to the preferred distribution name term
#'
#' This is determined by the internal preferred condition name list
#' `.errIdenticalDists`
#'
#' @param dist This is the input distribution
#' @return Preferred distribution term
#' @author Matthew Fidler
#'
#' @examples
#'
#' rxPreferredDistributionName("dt")
#'
#' rxPreferredDistributionName("add")
#'
#' # can be vectorized
#'
#' rxPreferredDistributionName(c("add","dt"))
#'
#' @export
rxPreferredDistributionName <- function(dist) {
  if (length(dist) == 1) {
    .names <- names(.errIdenticalDists)
    for(.n in .names) {
      if (dist == .n) return(.n)
      else if (dist %in% .errIdenticalDists[[.n]]) return(.n)
    }
    dist
  } else {
    vapply(dist, rxPreferredDistributionName, character(1))
  }
}

.incompatibleTransformations <-
  list(boxCox=c("yeoJohnson", "lnorm"),
       yeoJohnson=c("boxCox", "lnorm"),
       lnorm=c("yeoJohnson", "boxCox", "logit", "logitNorm", "probit", "probitNorm", "logit + yeoJohnson"),
       logit=c("probit", "probitNorm", "lnorm"),
       probit=c("logit", "lnorm"),
       "logit + yeoJohnson"=c("boxCox", "probit", "probitNorm", "lnorm"),
       "probit + yeoJohnson"=c("boxCox", "logit", "logitNorm", "lnorm"),
       "logit + boxCox"=c("yeoJohnson", "probit", "probitNorm", "lnorm"),
       "probit + boxCox"=c("yeoJohnson", "logit", "logitNorm", "lnorm"))

.rxTransformCombineLevels <- c(
  "boxCox", # 1
  "yeoJohnson", # 2
  "untransformed", # 3
  "lnorm", # 4
  "logit", # 5
  "logit + yeoJohnson", # 6
  "probit", # 7
  "probit + yeoJohnson", #8
  "logit + boxCox", # 9
  "probit + boxCox" # 10
)


.rxTransformHasALambdaParameter <- function(distribution) {
  (as.integer(distribution) %in% c(1L, 2L, 6L, 8L, 9L, 10L))
}

.rxTransformHasBounds <- function(distribution) {
  (as.integer(distribution) %in% 5:14)
}

.rxAddPropLevels <- c(
  "combined1", # 1
  "combined2", # 2
  "default" # 3
)

.incompatibleAddProp <- list(
  combined1="combined2",
  combined2="combined1"
)

.rxErrType <- c(
  "add", # 1
  "prop", # 2
  "pow", # 3
  "add + prop", # 4
  "add + pow", # 5
  "none" # 6
)

.rxDistributionType <- c(
  "norm",    #  1
  "pois",    #  2
  "binom",   #  3
  "beta",    #  4
  "t",       #  5
  "chisq",   #  6
  "dexp",    #  7
  "f",       #  8
  "geom",    #  9
  "hyper",   # 10; hyper will not be supported since all the inputs are integers
  "unif",    # 11
  "weibull", # 12
  "cauchy",  # 13
  "dgamma",  # 14
  "ordinal", # 15
  "LL",      # 16
  "dnorm",   # 17
  "nbinom",  # 18 size, prob
  "nbinomMu"  # 19 size, mu
)

#' Demote the error type
#'
#' @param errType Error type factor
#' @return Demoted Error Type
#' @author Matthew Fidler
#' @export
#' @examples
#' rxErrTypeCombine("add") |>
#'   rxErrTypeCombine("prop")
#'
#' # This removes the internal additive error
#' rxErrTypeCombine("add") |>
#'   rxErrTypeCombine("prop") |>
#'   rxDemoteAddErr()
#'
#' # This is used for logitNorm(NA), the additive portion is stripped
#'
#' @keywords internal
rxDemoteAddErr <- function(errType) {
  if (inherits(errType, "factor")) {
    return(structure(switch(as.character(errType),
                            add=6L,
                            "add + prop"=2L,
                            "add + pow"=3L,
                            as.integer(errType)),
                     levels=.rxErrType,
                     class="factor"))
  } else if (inherits(errType, "rxCombinedErrorList")) {
    return(.rxTransformCombineListOrChar(list(transform=errType$transform,
                                              errType=rxDemoteAddErr(errType$errType),
                                              errTypeF=errType$errTypeF,
                                              addProp=errType$addProp)))
  }
}

.incompatibleErrType <- list(prop=c("pow", "powT", "powF"),
                             propT=c("pow", "powT", "powF"),
                             propF=c("pow", "powT", "powF"),
                             pow=c("prop", "propT", "propF"),
                             powT=c("prop", "propT", "propF"),
                             powF=c("prop", "propT", "propF"))


.rxErrTypeF <- c(
  "untransformed", # 1
  "transformed", # 2
  "f", # 3
  "none")

.incompatibleErrTypeF <- list(prop=c("propT", "propF", "powT", "powF"),
                              propT=c("prop", "propF", "pow", "powF"),
                              propF=c("prop", "propT", "pow", "powT"),
                              pow=c("propF", "propT", "powF", "powT"),
                              powT=c("prop", "propF", "pow", "powF"),
                              powF=c("prop", "propT", "pow", "powT"))

.incompatibleErr <- function(err1, err2) {
  .errs <- sort(c(err1, err2))
  paste0("`", .errs[1], "` and `", .errs[2], "` are incompatible")
}
#' Combine error types to get the model F type
#'
#' @param newAddProp New error type
#' @param oldAddProp old error type
#' @return factor of the error type function OR a string indicating a syntax error
#' @author Matthew Fidler
#' @noRd
.rxCombineAddProp <- function(newAddProp, oldAddProp="default") {
  .tmp <- as.character(oldAddProp)
  .w <- which(names(.incompatibleAddProp) == .tmp)
  if (length(.w) == 1L) {
    if (newAddProp %in% .incompatibleAddProp[[.w]]) {
      return(.incompatibleErr(.tmp, newAddProp))
    }
  }
  structure(switch(newAddProp,
                   combined1 =1L,
                   combined2=2L,
                   default=3L,
                   ifelse(inherits(oldAddProp, "character"),
                          switch(oldAddProp,
                                 combined1 =1L,
                                 combined2=2L,
                                 default=3L,
                                 3L),
                          as.integer(oldAddProp))),
            levels=.rxAddPropLevels,
            class="factor")
}

#' Combine error types to get the model F type
#'
#' @param newErrTypeF New error type
#' @param oldErrTypeF old error type
#' @return factor of the error type function OR a string indicating a syntax error
#' @author Matthew Fidler
#' @noRd
.rxCombineErrTypeF <- function(newErrTypeF, oldErrTypeF="none") {
  .tmp <- as.character(oldErrTypeF)
  .w <- which(names(.incompatibleErrTypeF) == .tmp)
  if (length(.w) == 1L) {
    if (newErrTypeF %in% .incompatibleErrTypeF[[.w]]) {
      return(.incompatibleErr(.tmp, newErrTypeF))
    }
  }
  structure(switch(newErrTypeF,
                   prop =1L,
                   propT=2L,
                   propF=3L,
                   pow=1L,
                   powT=2L,
                   powF=3L,
                   ifelse(inherits(oldErrTypeF, "character"),
                          switch(oldErrTypeF,
                                 prop =1L,
                                 propT=2L,
                                 propF=3L,
                                 pow=1L,
                                 powT=2L,
                                 powF=3L,
                                 4L),
                          as.integer(oldErrTypeF))),
            levels=.rxErrTypeF,
            class="factor")
}


#' Combine error model
#'
#' @param newErrType New error portion combined with current proportion
#' @param oldErrType Old error information
#' @return A factor for error type OR a string with error information
#' @author Matthew Fidler
#' @noRd
.rxCombineErrType <- function(newErrType, oldErrType="none") {
  .tmp <- as.character(oldErrType)
  .w <- which(names(.incompatibleErrType) == .tmp)
  if (length(.w) == 1L) {
    if (newErrType %in% .incompatibleErrType[[.w]]) {
      return(.incompatibleErr(.tmp, newErrType))
    }
  }
  structure(switch(.tmp,
                   add=switch(newErrType,
                              prop=4L,
                              propT=4L,
                              propF=4L,
                              pow=5L,
                              powT=5L,
                              powF=5L,
                              1L),
                   prop=switch(newErrType,
                               add=4L,
                               probit=4L,
                               probitNorm=4L,
                               logit=4L,
                               logitNorm=4L,
                               lnorm=4L,
                               2L),
                   pow=switch(newErrType,
                              add=5L,
                              probit=5L,
                              probitNorm=5L,
                              logit=5L,
                              logitNorm=5L,
                              lnorm=5L,
                              3L),
                   none=structure(switch(newErrType,
                                         add=1L,
                                         lnorm=1L,
                                         logit=1L,
                                         logitNorm=1L,
                                         probit=1L,
                                         probitNorm=1L,
                                         prop=2L,
                                         propT=2L,
                                         propF=2L,
                                         pow=3L,
                                         powT=3L,
                                         powF=3L,
                                         6L)),
                   as.integer(oldErrType)),
            levels=.rxErrType,
            class="factor")
}
#' Combine transformations
#'
#' @param newTransform New error structure added together
#' @param oldTransform Old transformation added together
#' @return A factor describing the transformation type
#' @author Matthew Fidler
#' @noRd
.rxCombineTransform <- function(newTransform, oldTransform="untransformed") {
  .tmp <- as.character(oldTransform)
  .w <- which(names(.incompatibleTransformations) == .tmp)
  if (length(.w) == 1L) {
    if (newTransform %in% .incompatibleTransformations[[.w]]) {
      return(.incompatibleErr(.tmp, newTransform))
    }
  }
  structure(switch(.tmp,
                   boxCox=switch(newTransform,
                                 logitNorm=9L,
                                 probitNorm=10L,
                                 1L),
                   yeoJohnson=switch(newTransform,
                                     logitNorm=6L,
                                     probitNorm=8L,
                                     2L),
                   logit=switch(newTransform,
                                boxCox=9L,
                                yeoJohnson=6L,
                                5L),
                   probit=switch(newTransform,
                                 boxCox=10L,
                                 yeoJohnson=8L,
                                 7L),
                   untransformed=structure(switch(newTransform,
                                                  boxCox=1L,
                                                  yeoJohnson=2L,
                                                  lnorm=4L,
                                                  logitNorm=5L,
                                                  probitNorm=7L,
                                                  3L),
                                           levels=.rxTransformCombineLevels,
                                           class="factor"),
                   as.integer(oldTransform)),
            levels=.rxTransformCombineLevels,
            class="factor")
}

#' This is a wrapper to make sure that the transformation combination
#' returns the correct value
#'
#' @param inputList This is either an input list or character vector of length one
#'
#' @return Either a complete list, or a character vector which
#'   represents the parsed error that was encountered
#'
#' @author Matthew Fidler
#' @noRd
.rxTransformCombineListOrChar <- function(inputList) {
  if (inherits(inputList, "character")) return(inputList)
  .err  <- NULL
  for (i in names(inputList)) {
    if (inherits(inputList[[i]], "character")) {
      .err <- c(.err, inputList[[i]])
    }
  }
  if (is.null(.err)) {
    .ret <- inputList
    class(.ret) <- "rxCombinedErrorList"
    return(.ret)
  } else {
    return(paste(.err, collapse="\n"))
  }
}

#' Combine transformations and error structures
#'
#' Combine error information to figure out what transformation is
#' being applied for the current endpoint
#'
#' @param oldErrType This is the old transformation, by default is
#'   zero representing no prior transformation. This parameter is
#'   first to allow piping. When the parameter `addTransform` is
#'   missing and `oldErrType` is a character value, this functions
#'   swaps `oldErrType` and `addTransform` and assigns
#'   `oldErrType` to zero assuming that there is no prior
#'   distribution.
#'
#' @param newErrType This is the new distribution that is being
#'   "added" to the current transformation.  These assumes the inputs
#'   are in the preferred distribution name, as determined by
#'   `rxPreferredDistributionName()`
#'
#' @return The new transformation as a factor
#'
#' @author Matthew Fidler
#'
#' @examples
#'
#' rxErrTypeCombine("probitNorm")
#'
#' rxErrTypeCombine("probitNorm") |>
#'   rxErrTypeCombine("boxCox")
#'
#'
#' @export
#' @keywords internal
rxErrTypeCombine <- function(oldErrType, newErrType) {
  if (missing(newErrType) && inherits(oldErrType, "character")) {
    return(.rxTransformCombineListOrChar(list(transform=.rxCombineTransform(oldErrType),
                                              errType=.rxCombineErrType(oldErrType),
                                              errTypeF=.rxCombineErrTypeF(oldErrType),
                                              addProp=.rxCombineAddProp(oldErrType))))
  } else if (inherits(oldErrType, "rxCombinedErrorList")) {
    return(.rxTransformCombineListOrChar(list(transform=.rxCombineTransform(newErrType, oldErrType$transform),
                                              errType=.rxCombineErrType(newErrType, oldErrType$errType),
                                              errTypeF=.rxCombineErrTypeF(newErrType, oldErrType$errTypeF),
                                              addProp=.rxCombineAddProp(newErrType, oldErrType$addProp))))
  } else {
    stop("old transform not in the proper format", call.=FALSE)
  }
}

#' Checks to see if an expression is numeric
#'
#' @param expression quoted expression
#' @param env Environment to store result in `env$.numeric`
#' @return TRUE if this is an expression containing a positive or negative expression or FALSE if it is an expression that doesn't contain an expression.
#' @author Matthew Fidler
#' @noRd
.is.numeric <- function(expression, env) {
  if (is.numeric(expression)) {
    env$.numeric <- expression
    return(TRUE)
  } else if (length(expression) == 2L) {
    if (identical(expression[[1]], quote(`-`)) &&
          is.numeric(expression[[2]])) {
      env$.numeric <- -(expression[[2]])
      return(TRUE)
    } else if (identical(expression[[1]], quote(`+`)) &&
                 is.numeric(expression[[2]])) {
      env$.numeric <- expression[[2]]
      return(TRUE)
    }
  }
  return(FALSE)
}

.allowDemoteAddDistributions <- c("lnorm", "probitNorm", "logitNorm")

.namedArgumentsToPredDf <- list(
  add="a",
  lnorm="a",
  boxCox="lambda",
  yeoJohnson="lambda",
  pow=c("b", "c"),
  powT=c("b", "c"),
  powF=c("b", "c", "f"),
  prop="b",
  propT="b",
  propF=c("b", "f"),
  t=c("d", "e"),
  pois=c("a"),
  binom=c("a", "b"),
  beta=c("a", "b",  "c"),
  chisq=c("a", "b"), #6
  dexp=c("a"), #7
  f=c("a", "b", "c"), #8
  geom=c("a"), #9
  #  hyper=c("a", "b", "c"), #10
  unif=c("a", "b"), #11
  weibull=c("a", "b"), #12
  cauchy=c("a", "b"),
  dgamma=c("a", "b"),
  nbinom=c("a", "b"),
  nbinomMu=c("a", "b")
)

.allowEstimatedParameters <- "ordinal"

#' Convert a numeric literal error-model argument into a FIX `$iniDf` parameter
#'
#' A residual literal such as `add(0.7)`, `prop(0.1)` or `ar(0.5)` is turned
#' into an auto-generated, uniquely named, `rx`-prefixed **FIX** parameter that
#' is appended to `env$df` (the working `$iniDf`) and linked to the current
#' endpoint through the `err`/`condition` columns -- exactly as a user-supplied
#' estimated (fixed) residual parameter would be.  This keeps residual literals
#' out of `$predDf` (so no dedicated columns like the former `ar` are needed)
#' and lets them flow through every downstream path (estimation, simulation,
#' export) unchanged.
#'
#' The generated name is `rx.<endpoint>.<err>` where `<err>` is the endpoint's
#' `err` specification (the function name for argument 1, the function name with
#' the argument index appended otherwise, e.g. `pow` / `pow2`), so multiple
#' fixed literals in one function get distinct names.  It is de-duplicated with
#' a numeric suffix if that name already exists.
#'
#' @param argumentNumber distribution argument index (1-based); argument 1 uses
#'   the bare function name for `err`, later arguments append the index (matching
#'   the named-parameter convention in [.errHandleSingleDistributionArgument]).
#' @param funName error distribution function name
#' @param value numeric literal value
#' @param env parse environment (holds `df`, the working `$iniDf`, and
#'   `curCondition`, the current endpoint variable)
#' @return NULL, called for its side effect of appending a row to `env$df`
#' @author Matthew L. Fidler
#' @noRd
.rxErrLiteralToFixParam <- function(argumentNumber, funName, value, env) {
  .df <- env$df
  # the name mirrors the `err` specification (funName for argument 1, funName
  # with the argument index appended otherwise), so a function with more than
  # one fixed literal (e.g. pow(1.5, 2)) gets distinct names (rx.cp.pow,
  # rx.cp.pow2) rather than colliding on the bare function name
  .errName <- if (argumentNumber == 1L) funName else paste0(funName, argumentNumber)
  .base <- paste0("rx.", env$curCondition, ".", .errName)
  .name <- .base
  .i <- 0L
  while (.name %in% .df$name) {
    .i <- .i + 1L
    .name <- paste0(.base, ".", .i)
  }
  .ntheta <- suppressWarnings(max(.df$ntheta, na.rm=TRUE))
  if (!is.finite(.ntheta)) .ntheta <- 0L
  # positive-support residual parameters get a lower bound of 0; ar() is a
  # correlation in [0, 1), so its upper bound is the strict-below-1 sup rather
  # than an inclusive 1; everything else stays unbounded (the value is FIX
  # regardless)
  .lower <- -Inf
  .upper <- Inf
  if (funName == "ar") {
    .lower <- 0
    .upper <- 1 - .Machine$double.eps
  } else if (funName %in% .errDistsPositive) {
    .lower <- 0
  }
  # build a one-row frame that matches env$df's columns exactly, then fill it
  .row <- .df[0, , drop=FALSE]
  .row[1, ] <- NA
  .row$ntheta <- .ntheta + 1L
  .row$neta1 <- NA_integer_
  .row$neta2 <- NA_integer_
  .row$name <- .name
  .row$lower <- .lower
  .row$est <- value
  .row$upper <- .upper
  .row$fix <- TRUE
  .row$condition <- env$curCondition
  .row$err <- .errName
  env$df <- rbind(.df, .row)
  assign("lastDistAssign", .name, envir=env)
  invisible(NULL)
}

#' This handles the error distribution for a single argument.
#'
#' @param argumentNumber The argument number of the distribution being processed
#' @param funName Function name string of the distribution name
#' @param expression Function expression (including the function name)
#' @param env Environment where the names and calculations are made for the user interface.
#' @return None, called for the side effects.
#' @author Matthew Fidler
#' @noRd
.errHandleSingleDistributionArgument <- function(argumentNumber, funName, expression, env) {
  .cur <- expression[[argumentNumber + 1]]
  .isLogitOrProbit <- (funName %in% c("logitNorm", "probitNorm"))
  if (is.name(.cur)) {
    .curName <- as.character(.cur)
    .w <- which(env$df$name == .curName)
    if (.isLogitOrProbit && argumentNumber > 2) {
      env$err <- c(env$err,
                   paste0("'", funName, "()' requires numeric bounds"))
    } else if (length(.w) == 1L) {
      .df  <- env$df
      .df$err[.w] <- ifelse(argumentNumber == 1, funName, paste0(funName, argumentNumber))
      if (funName == "ar" && !is.na(.df$est[.w]) &&
            (.df$est[.w] < 0 || .df$est[.w] >= 1)) {
        env$err <- c(env$err,
                     paste0("the initial estimate of 'ar(", .curName,
                            ")' must be in [0, 1)"))
      }
      if (!is.na(.df$condition[.w])) {
        assign("dupErr", c(env$dupErr, .df$name[.w]), envir=env)
      }
      .df$condition[.w] <- env$curCondition
      assign("df", .df, envir=env)
      assign("lastDistAssign", .curName, envir=env)
    } else if (funName == "ar") {
      # ar(<modeled variable>): the correlation is a model-calculated quantity
      # (e.g. corv <- expit(tcor)) rather than an estimated parameter or a
      # literal.  Nothing is stored in $iniDf/$predDf; .rxGetArCorLang() recovers
      # the correlation directly from the endpoint's error expression.  The
      # referenced variable is validated as a model quantity later, like other
      # modeled residual references.
      assign("lastDistAssign", .curName, envir=env)
      return(NULL)
    } else {
      .w <- which(names(.namedArgumentsToPredDf) == funName)
      if (length(.w) == 1) {
        .curLst <- .namedArgumentsToPredDf[[.w]]
        if (argumentNumber <= length(.curLst)) {
          assign(.curLst[argumentNumber], .curName, envir=env)
          return(NULL)
        }
      }
      if (env$estNotAllowed) {
        if (!(funName %in% .allowEstimatedParameters)) {
          env$err <- c(env$err,
                       paste0("in the error expression, the variable '", .curName, "' must be estimated, not calculated"))
        }
      }
    }
  } else if (.is.numeric(.cur, env)) {
    if (.isLogitOrProbit && (argumentNumber == 2 || argumentNumber == 3)) {
      # logitNorm()/probitNorm() transformation bounds are structural
      # constants, not parameters
      env$trLimit[argumentNumber - 1] <- env$.numeric
    } else if (funName == "ar" && (env$.numeric < 0 || env$.numeric >= 1)) {
      env$err <- c(env$err,
                   "'ar()' correlation must be in [0, 1)")
    } else {
      # A numeric literal supplied to an error-model function (e.g. add(0.7),
      # prop(0.1), ar(0.5)) is converted to an auto-generated, uniquely named
      # FIX parameter in the $iniDf -- exactly as if the user had written
      # `rx.<endpoint>.<func> <- fix(<value>)` and referenced it by name.  This
      # keeps residual literals out of $predDf (no dedicated columns needed) and
      # routes them through the same machinery as an estimated residual
      # parameter.
      .rxErrLiteralToFixParam(argumentNumber, funName, env$.numeric, env)
    }
  } else if (is.na(.cur)) {
    if (argumentNumber == 1 && funName %in% .allowDemoteAddDistributions) {
      env$needToDemoteAdditiveExpression <- TRUE
    } else {
      env$err <- c(env$err,
                   paste0("NA in '", funName, "()' cannot be used here"))
    }
  }
}

#' Checks and modifies the error term as needed in the user interface function.
#'
#' @param funName Name of the distributional error
#' @param expression The R expression (including the function name)
#' @param env The environment that holds information about the error structure
#' @return Nothing; Called for the side effects
#' @author Matthew Fidler
#' @noRd
.errHandleSingleDistributionTerm <- function(funName, expression, env) {
  .nargs <- length(expression) - 1
  .doIt <- FALSE
  if (funName == "ordinal") {
    if (.nargs == 0) {
      assign("err", c(env$err,
                      paste0("ordinal errors require at least 1 argument (i.e. err ~ c(err1))", envir=env)))
      return(invisible())
    }
    .doIt <- TRUE
  } else {
    .errDistArgs <- .errDist[[funName]]
    .doIt <- .nargs %in% .errDistArgs
  }
  if (.doIt) {
    if (.nargs > 0) {
      lapply(seq(1, .nargs), .errHandleSingleDistributionArgument, funName=funName, expression=expression, env=env)
    }
    env$errTypeInfo <- rxErrTypeCombine(env$errTypeInfo, funName)
    env$needsToBeAnErrorExpression <- TRUE
  } else {
    .min <- range(.errDistArgs)
    .max <- .min[2]
    .min <- .min[1]
    if (.min == .max) {
      assign("err", c(env$err,
                      paste0("`", funName, "` requires ",
                             .max, " argument(s), you specified ", .nargs)), envir=env)
    } else {
      assign("err", c(env$err,
                      paste0("`", funName, "` requires ",
                             .min, " to ", .max, " argument(s), you specified ", .nargs)),
             envir=env)
    }
    if (.nargs > 0) {
      lapply(seq(1, .nargs), .errHandleSingleDistributionArgument, funName=funName, expression=expression, env=env)
    }
  }
}
#' This handles a function that is not an error term.
#'
#' @param funName Function that is being called, as a character
#' @param expression Expression (including function)
#' @param env Environment that stores the information about errors
#' @return Nothing, called for the side effects
#' @author Matthew Fidler
#' @noRd
.errHandleSingleTerm <- function(funName, expression, env) {
  env$hasNonErrorTerm <- TRUE
}

#' Handle the error structure term
#'
#' @param expression The error structure term
#'
#' @param env The environment that holds information about the error
#'   structure
#'
#' @return Nothing, called for the side efects
#' @author Matthew Fidler
#' @noRd
.errHandleErrorStructure <- function(expression, env) {
  .isAdd <- FALSE
  if (is.name(expression) || is.atomic(expression)) {
    .isAdd <- FALSE
  } else if (identical(expression[[1]], quote(`+`))) {
    .isAdd <- TRUE
  }
  if (.isAdd) {
    env$isAnAdditiveExpression <- TRUE
    .errHandleErrorStructure(expression[[2]], env)
    .errHandleErrorStructure(expression[[3]], env)
  } else if (env$isAnAdditiveExpression) {
    .currErr <- rxPreferredDistributionName(deparse1(expression[[1]]))
    if (.currErr %in% .errAddDists) {
      if (.currErr == "var") {
        env$var <- TRUE
        return(invisible())
      }
      if (.currErr == "dv") {
        env$dv <- TRUE
        return(invisible())
      }
      if (.currErr == "ar") {
        if (env$hasAr) {
          env$err <- c(env$err, "'ar()' can only be specified once per endpoint")
          return(invisible())
        }
        if (length(expression) - 1 != 1) {
          env$err <- c(env$err, "'ar()' requires 1 argument (the correlation)")
          return(invisible())
        }
        env$hasAr <- TRUE
        .errHandleSingleDistributionArgument(1, "ar", expression, env)
        env$needsToBeAnErrorExpression <- TRUE
        return(invisible())
      }
      if (.currErr == "t") {
        if (env$distribution == "cauchy") {
          stop("you cannot combine 't' and 'cauchy' distributions")
        }
        if (env$distribution == "dnorm") {
          stop("you cannot combine 't' and 'dnorm' distributions")
        }
        env$distribution <- "t"
      } else if (.currErr == "cauchy") {
        if (env$distribution == "t") {
          stop("you cannot combine 't' and 'cauchy' distributions")
        }
        if (env$distribution == "dnorm") {
          stop("you cannot combine 'dnorm' and 'cauchy' distributions")
        }
        env$distribution <- "cauchy"
      } else if (.currErr == "dnorm") {
        if (env$distribution == "t") {
          stop("you cannot combine 't' and 'dnorm' distributions")
        }
        if (env$distribution == "cauchy") {
          stop("you cannot combine 'cauchy' and 'dnorm' distributions")
        }
        env$distribution <- "dnorm"
      }
      .errHandleSingleDistributionTerm(.currErr, expression, env)
    } else if (.currErr %in% names(.errDist)) {
      assign("err", c(env$err,
                      paste0("`", .currErr, "` is incorrectly added to an error expression")), envir=env)
    } else {
      .errHandleSingleTerm(.currErr, expression, env)
    }
  } else {
    if (is.name(expression) || is.atomic(expression)) {
      .currErr <- deparse1(expression)
      .isC <- FALSE
      .isErrDist <- FALSE
    } else {
      .currErr <- deparse1(expression[[1]])
      .isC <- .currErr == "c"
      .isErrDist <- .currErr %in% names(.errDist)
    }
    if (.isC) {
      env$distribution <- "ordinal"
      .errHandleSingleDistributionTerm("ordinal", expression, env)
    } else if (identical(rxPreferredDistributionName(.currErr), "ar")) {
      env$err <- c(env$err,
                   "'ar()' must be combined with a transformably normal error model (e.g. add(add.sd) + ar(cor))")
    } else if (.isErrDist) {
      .currErr <- rxPreferredDistributionName(.currErr)
      env$distribution <- .currErr
      .errHandleSingleDistributionTerm(.currErr, expression, env)
    } else {
      .errHandleSingleTerm(.currErr, expression, env)
    }
  }
}

#' Handle the right hand side conditional expression (if it exists)
#'
#' @param expression A right hand side of the tilde equation
#' @param env Environment for storing information about the expression
#' @return An expression without a conditional statement.
#'
#' @details
#'
#' In addition to stripping the conditional statement out of the
#' expression, the environment is modified when a conditional
#' expression is present.  First, the environment variable
#' `needsToBeAnErrorExpression` is changed to `TRUE`.  Second, the
#' expression `curCondition` is modified to match the information
#' within the conditional statement.
#'
#' @author Matthew Fidler
#' @noRd
.errHandleCondition <- function(expression, env) {
  if (is.name(expression) || is.atomic(expression)) return(expression)
  if (identical(expression[[1]], quote(`|`))) {
    env$needsToBeAnErrorExpression  <- TRUE
    if (length(expression[[3]]) != 1 &&
          !is.symbol(expression[[3]])) {
      env$err <- c(env$err,
                   paste0("the condition '", deparse1(expression[[3]]), "' must be a simple name"))
      env$earlyErr <- TRUE
    }
    env$curCondition <- deparse1(expression[[3]])
    return(expression[[2]])
  }
  expression
}

#' Handle LL equivalent for ll or `linCmt()` statements for lhs
#'
#' @param expression Left handed side of the equation
#'
#' @param env Environment for storing information about the error
#'   structure
#'
#' @return Nothing called for side effects
#'
#' @details
#'
#' The takes the expression
#'
#' ll(var) ~  log(...)
#'
#' And strips the `ll` and sets the flag `env$ll` to `TRUE`
#'
#' Otherwise it leaves the `expression` alone and returns the value
#'
#' This takes the expression:
#'
#' linCmt() ~ add() ...
#'
#' And returns a `rxLinCmt` as the lhs value and sets the linCmt flag to TRUE
#'
#'
#' @author Matthew Fidler
#' @noRd
.errHandleLlOrLinCmt <- function(expression, env) {
  if (is.call(expression)) {
    if (identical(expression[[1]], quote(`ll`)) &&
          length(expression) == 2L) {
      env$ll <- TRUE
      return(expression[[2]])
    } else if (identical(expression[[1]], quote(`linCmt`))) {
      env$linCmt <- TRUE
      return(quote(`rxLinCmt`))
    } else {
      stop("the left handed side of the error expression (function: '", as.character(expression[[1]]), "') can only be functions with 'linCmt' or 'll'",
           call.=FALSE)
    }
  }
  expression
}


#' Handle the error expressions
#'
#' @param expression Single tilde error expression
#' @param env Environment with initial estimate data.frame
#' @return nothing, called for side effects
#' @author Matthew Fidler
#' @noRd
.errHandleTilde <- function(expression, env) {
  env$ll <- FALSE
  env$estNotAllowed <- TRUE
  env$linCmt <- FALSE
  env$var <- FALSE
  env$dv <- FALSE
  env$hasAr <- FALSE
  .left <- .errHandleLlOrLinCmt(expression[[2]], env)
  env$trLimit <- c(-Inf, Inf)
  env$a <- env$b <- env$c <- env$d <- env$e <- env$f <- env$lambda <- NA_character_
  env$curCondition <- env$curVar <- deparse1(.left)
  env$hasNonErrorTerm <- FALSE
  env$needsToBeAnErrorExpression <- FALSE
  env$needToDemoteAdditiveExpression <- FALSE
  .right <- .errHandleCondition(expression[[3]], env)
  env$isAnAdditiveExpression <- FALSE
  env$errTypeInfo <- rxErrTypeCombine("")
  env$distribution <- "norm"
  .errHandleErrorStructure(.right, env)
  if (env$hasAr && !(env$distribution %in% c("norm", "dnorm", "t", "cauchy"))) {
    env$err <- c(env$err,
                 "'ar()' is only allowed with normal, t, or cauchy based error models")
  }
  if (inherits(env$errTypeInfo, "character")) {
    env$err <- c(env$err, env$errTypeInfo)
    env$errTypeInfo <- rxErrTypeCombine("")
  } else if (env$needToDemoteAdditiveExpression) {
    env$errTypeInfo <- rxDemoteAddErr(env$errTypeInfo)
  }
  if (env$hasNonErrorTerm && env$needsToBeAnErrorExpression) {
    assign("errGlobal", c(env$errGlobal, "cannot mix error expression with algebraic expressions"),
           envir=env)
  } else if (env$hasNonErrorTerm) {
    if (env$ll) {
      env$distribution <- "LL"
      env$predDf <- rbind(env$predDf,
                          data.frame(cond=env$curCondition,
                                     var=env$curVar, dvid=env$curDvid,
                                     trLow=env$trLimit[1],
                                     trHi=env$trLimit[2],
                                     transform=env$errTypeInfo$transform,
                                     errType=env$errTypeInfo$errType,
                                     errTypeF=env$errTypeInfo$errTypeF,
                                     addProp=env$errTypeInfo$addProp,
                                     distribution=factor(env$distribution, levels=.rxDistributionType),
                                     line=env$line,
                                     a=env$a,
                                     b=env$b,
                                     c=env$c,
                                     d=env$d,
                                     e=env$e,
                                     f=env$f,
                                     lambda=env$lambda,
                                     linCmt=env$linCmt,
                                     variance=env$var,
                                     dv=env$dv))
      env$curDvid <- env$curDvid + 1L

    }
  } else if (!env$hasNonErrorTerm) {
    if (env$ll) {
      assign("errGlobal", c(env$errGlobal, "a -2 log-likelihood expression cannot use abbreviated error codes like add() + prop() "),
             envir=env)
    } else {
      if (!(env$distribution %in% .rxDistributionType)) {
        env$distribution <- "norm"
      }
      .tmp <- data.frame(cond=env$curCondition, var=env$curVar, dvid=env$curDvid,
                         trLow=env$trLimit[1], trHi=env$trLimit[2],
                         transform=env$errTypeInfo$transform,
                         errType=env$errTypeInfo$errType,
                         errTypeF=env$errTypeInfo$errTypeF,
                         addProp=env$errTypeInfo$addProp,
                         distribution=factor(env$distribution, levels=.rxDistributionType),
                         line=env$line,
                         a=env$a,
                         b=env$b,
                         c=env$c,
                         d=env$d,
                         e=env$e,
                         f=env$f,
                         lambda=env$lambda,
                         linCmt=env$linCmt,
                         variance=env$var,
                         dv=env$dv)
      env$predDf <- rbind(env$predDf, .tmp)
      env$curDvid <- env$curDvid + 1L
    }
  }
}
#' Add the Error Names
#'
#' @param errs Input endpoints
#' @return The endpoints that the user enters
#' @author Matthew L. Fidler
#' @noRd
.userEndpointNames <- function(endpoints) {
  vapply(endpoints, function(e) {
    if (e == "rxLinCmt") {
      return("linCmt()")
    }
    e
  }, character(1), USE.NAMES=FALSE)
}

#' Check for error exceptions
#'
#' @param env Environment to check for error exceptions
#' @return Nothing called for side effects
#' @author Matthew L. Fidler
#' @noRd
.checkForMissingOrDupliacteInitials <- function(env) {
  .predDf <- env$predDf
  .iniDf <- env$iniDf
  .err <- env$dupErr
  for (.i in seq_along(.predDf$cond)) {
    if (!any(!is.na(.predDf[.i, c("a", "b", "c", "d", "e", "f", "lambda")]))) {
      if (!(.predDf[.i, "distribution"] %in% c("LL", "ordinal"))) {
        .cnd <- .predDf$cond[.i]
        .w <- which(.iniDf$condition == .cnd)
        if (length(.w) == 0L) {
          if (!any(.errDist[[paste(.predDf$distribution[.i])]] == 0)) {
            .err <- c(.err, .cnd)
          }
        }
      }
    }
  }
  if (length(.err) > 0) {
    stop("endpoint parameter(s) missing, duplicated, or defined with '~': ", paste(.userEndpointNames(.err), collapse=", "),
         call.=FALSE)
  }
}

#' Get a blank, theta1, or eta1 initialization block for iniDf
#'
#' @param type type of initialization block to return
#' @return A data.frame with the appropriate number/type of columns.
#'
#' For type="empty", the data.frame will have 0 rows but all the correct types.
#'
#' For type="theta", the data.frame will have 1 row with the correct
#' types and default values.  The "name" and "est" will likely need to
#' be updated.
#'
#' For type="eta", the data.frame will have 1 row with the correct
#' types and default values for the a single eta being added.  The
#' "name" and "est" will likely need to be updated.
#'
#'
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
#' @examples
#'
#' .rxBlankIni("empty")
#'
#' .rxBlankIni("theta")
#'
#' .rxBlankIni("eta")
#'
.rxBlankIni <- function(type=c("empty", "theta", "eta")) {
  type <- match.arg(type)
  if (type == "empty") {
    data.frame(ntheta=integer(0),
               neta1=integer(0),
               neta2=integer(0),
               name=character(0),
               lower=numeric(0),
               est=numeric(0),
               upper=numeric(0),
               fix=logical(0),
               err=character(0),
               label=character(0),
               stringsAsFactors=FALSE)
  } else if (type == "theta") {
    data.frame(ntheta=1L,
               neta1=NA_integer_,
               neta2=NA_integer_,
               name=NA_character_,
               lower=-Inf,
               est=0,
               upper=Inf,
               fix=FALSE,
               err=NA_character_,
               label=NA_character_,
               stringsAsFactors=FALSE)
  } else {
    data.frame(ntheta=NA_integer_,
               neta1=1L,
               neta2=1L,
               name=NA_character_,
               lower=0,
               est=0.1,
               upper=Inf,
               fix=FALSE,
               err=NA_character_,
               label=NA_character_,
               stringsAsFactors=FALSE)
  }
}
#' UI errors detected. Called in multiple places.
#'
#' @param env environment where error information is located
#' @return nothing, called for side effects
#' @noRd
#' @author Matthew L. Fidler
.handleErrs <- function(env) {
  if (env$hasErrors) {
    .errMsg <- paste0(crayon::bold$blue("\nmodel"), "({}) errors:\n",
                      paste(vapply(seq_along(env$lstExpr),
                                   function(i) {
                                     sprintf(paste0("%s", crayon::bold("%03d:"), " %s"),
                                             ifelse(is.null(env$lstErr[[i]]), "",
                                                    sprintf(paste0(crayon::bold("%s"), "\n"), env$lstErr[[i]])),
                                             i, deparse1(env$lstExpr[[i]]))
                                   }, character(1), USE.NAMES=FALSE), collapse="\n"))
    message(.errMsg)
  }
  if (length(env$err) > 0) {
    stop(paste0(ifelse(env$hasErrors, "syntax/parsing errors (see above) and additionally:\n", "syntax/parsing errors:\n"),
                paste(env$err, collapse="\n")),
         call.=FALSE)
  } else if (env$hasErrors) {
    stop("syntax/parsing errors, see above", call.=FALSE)
  }
  invisible(NULL)
}

#' Process the errors in the quoted expression
#'
#' @param x Quoted expression for parsing
#' @param ini Initialization block
#' @param linCmtSens Type of linCmt sensitivity
#' @param verbose verbose
#' @param checkMissing Check for missing arguments in the ini({}) block
#' @param df lotri data.frame of estimates
#' @return Environment with error information setup.
#' @author Matthew Fidler
#' @examples
#' lmat <- lotri({
#'  ## You may label each parameter with a comment
#'  tka <- 0.45 # Log Ka
#'  tcl <- log(c(0, 2.7, 100)) # Log Cl
#'  ## This works with interactive models
#'  ## You may also label the preceding line with label("label text")
#'  tv <- 3.45; label("log V")
#'  tvp <- 3.45; label("log V")
#'  cl.wt <- 0.1
#'  v.wt <- 0.1
#'  cl.sex <- 0.1
#'  v.sex <- 0.1
#'  cl.age <- 0.1
#'  v.age <- 0.1
#'  vp.wt <- 1
#'  vp.sex <- 1
#'  vp.age <- 1
#'  ## the label("Label name") works with all models
#'  eta.ka ~ 0.6
#'  eta.cl ~ 0.3
#'  eta.v ~ 0.1
#'  add.sd <- 0.7
#'})
#'
#'
#' iniDf <- as.data.frame(lmat)
#'
#' .errProcessExpression()
#' @noRd
.errProcessExpression <- function(x, ini,
                                  linCmtSens = c("linCmtA", "linCmtB", "linCmtC"),
                                  verbose=FALSE, checkMissing=TRUE, mv=rxUdfUiMv()) {
  on.exit({
    .udfUiEnv$num <- 1L
    .udfUiEnv$iniDf <- NULL
    .udfUiEnv$lhs <- NULL
    .udfUiEnv$parsing <- FALSE
  })
  .udfUiEnv$probs <- NULL
  .udfUiEnv$parsing <- TRUE
  # ntheta neta1 neta2   name lower       est   upper   fix  err  label
  # backTransform condition trLow trHi
  .env <- new.env(parent=emptyenv())
  .env$uiUseData <- FALSE
  .env$uiUseMv <- FALSE
  rxUdfUiEst(NULL)
  rxUdfUiMv(mv)
  .env$rxUdfUiCount <- new.env(parent=emptyenv())
  .env$before <- list()
  .env$after <- list()
  .env$level <- NULL
  if (is.matrix(ini)) {
    .env$eta <- dimnames(ini)[[1]]
  } else {
    .env$eta <- dimnames(ini$id)[[1]]
    ## Get the levels of the ini block
    for (v in names(ini)) {
      if (v == "id") next
      .env$level <- c(.env$level, dimnames(ini[[v]])[[1]])
    }
  }
  .env$top <- TRUE
  if (!inherits(ini, "lotriFix") && inherits(ini, "matrix")) {
    class(ini) <- c("lotriFix", class(ini))
  }
  .env$df <- as.data.frame(ini)
  .env$dupErr <- NULL
  .env$err <- NULL
  .env$errGlobal <- NULL
  # Add error structure like nlmixr ui had before transitioning to rxode2
  if (length(.env$df$ntheta) > 0) {
    .env$df$err <- NA_character_
  } else {
    .env$df$err <- character(0)
  }
  #.env$df$trLow <- .env$df$trHi <- NA_real_
  .env$curDvid <- 1L
  # Pred df needs to be finalized with compartment information from parsing the raw rxode2 model
  .env$predDf  <- NULL
  .env$earlyErr <- FALSE
  .env$lastDistAssign <- ""
  if (is.call(x)) {
    if (.env$top && identical(x[[1]], quote(`{`))) {
      .env$top <- FALSE
      .y <- x[-1]
      .env$lstChr <- character(length(.y))
      .env$lstErr <- vector(length(.y), mode="list")
      .env$lstExpr <- vector(length(.y), mode="list")
      .env$hasErrors <- FALSE
      .i <- 1L
      .y <- lapply(seq_along(.y), function(i) {
        .y[[i]]
      })
      while(.i <= length(.y)) {
        .env$line <- .i
        if (length(.y[[.i]]) == 1 &&
              is.name(.y[[.i]])) {
          .env$err <- c(.env$err,
                        paste0("the symbol '", deparse1(.y[[.i]]), "' cannot be by itself"))
          .env$earlyErr <- TRUE
        } else if (identical(.y[[.i]][[1]], quote(`~`))) {
          .errHandleTilde(.y[[.i]], .env)
        } else {
          .env$redo <- FALSE
          .cur <- .y[[.i]]
          if (length(.cur) >= 3 && identical(.cur[[1]], quote(`<-`))) {
            .env$lhs <- .cur[[2]]
          } else if (length(.cur)>= 3 && identical(.cur[[1]], quote(`=`))) {
            .env$lhs <- .cur[[2]]
          } else {
            .env$lhs <- NULL
          }
          .cur <- .handleUdfUi(.cur, .env)
          .len <- length(.y)
          .y <- c(lapply(seq_len(.i - 1),
                         function(i) {
                           .y[[i]]
                         }), .env$before, .cur, .env$after,
                  lapply(seq_len(.len - .i),
                         function(i) {
                           .y[[i + .i]]
                         }))
          if (length(.y) != .len) {
            # Update the lengths of lstChr, lstErr, lstExpr
            .len <- length(.env$before) + length(.env$after)
            .env$lstChr <- c(.env$lstChr, character(.len))
            .env$lstErr <- c(.env$lstErr, vector(.len, mode="list"))
            .env$lstExpr <- c(.env$lstExpr, vector(.len, mode="list"))
            .env$before <- list()
            .env$after <- list()
            # redo the parsing since the length of the expression has changed
            next
          } else if (.env$redo) {
            next
          }
        }
        .env$lstChr[[.i]] <- deparse1(.y[[.i]])
        .env$lstExpr[[.i]] <- .y[[.i]]
        if (!is.null(.env$err)) {
          .env$lstErr[[.i]] <- paste(paste(" ", .env$err), collapse="\n")
          .env$err <- NULL
          .env$hasErrors <- TRUE
        }
        .i <- .i + 1L
      }
      .env$iniDf <- .env$df
      if (is.null(.env$predDf)) {
        ## .env$errGlobal <- c(.env$errGlobal,
        ##                     "there must be at least one prediction in the model({}) block.  Use `~` for predictions")
      } else if (length(.env$predDf[, 1]) == 0L) {
        .env$predDf <- NULL
        ## .env$errGlobal <- c(.env$errGlobal,
        ##                     "there must be at least one prediction in the model({}) block.  Use `~` for predictions")
      }
      if (!is.null(.env$errGlobal)) {
        stop(paste(.env$errGlobal, collapse="\n"), call.=FALSE)
      }
      if (!is.null(.env$predDf)) {
        .lstChr <- .env$lstChr
        .lines <- .env$predDf$line
        .w <- which(.env$predDf$distribution == "LL")
        if (length(.w) > 0) {
          .lstExpr <- .env$lstExpr
          .w2 <- .env$predDf$line[.w]
          .lstChr <- vapply(seq_along(.lstExpr),
                            function(i) {
                              .cur <- .lstExpr[[i]]
                              if (i %in% .w2) {
                                paste0("rxLL ~ ", deparse1(.cur[[3]]))
                              } else {
                                deparse1(.cur)
                              }
                            }, character(1), USE.NAMES=FALSE)
          .lines <- .lines[-.w]
          if (length(.lines) > 0) {
            .lstChr <- .lstChr[-.lines]
          }
        } else {
          .lstChr <- .lstChr[-.lines]
        }
        if (length(.lstChr) == 0) {
          stop("a rxode2 ui model must have more than error definition(s) in the `model({})` block",
               call.=FALSE)
        }
        if (.env$earlyErr) {
          .handleErrs(.env)
        }
        if (any(.env$predDf$linCmt)) {
          .env$mv0 <- rxModelVars(paste(c(.lstChr, "rxLinCmt ~ linCmt()"), collapse="\n"))
        } else {
          .env$mv0 <- rxModelVars(paste(.lstChr, collapse="\n"))
        }
      } else {
        if (.env$earlyErr) {
          .handleErrs(.env)
        }
        .env$mv0 <- rxModelVars(paste(.env$lstChr, collapse="\n"))
      }
      if (isTRUE(.env$uiUseMv) && is.null(mv)) {
        # ui function requests model variables, so re-process
        on.exit({
          rxUdfUiMv(NULL)
        })
        return(.errProcessExpression(x=x, ini=ini,
                                     linCmtSens = linCmtSens,
                                     verbose=verbose, checkMissing=checkMissing,
                                     mv=.env$mv0))
      }
      .env$errParams0 <- rxUiGet.errParams(list(.env, TRUE))
      if (.Call(`_rxode2_isLinCmt`) == 1L) {
        .env$.linCmtM <- rxNorm(.env$mv0)
        .ini <- .env$mv0$ini
        .ini <- which(!is.na(.ini))
        .vars <- c(.env$mv0$params, .env$mv0$lhs, .env$mv0$slhs, names(.ini))
        .env$mvL <- rxGetModel(.Call(
          `_rxode2_linCmtGen`,
          length(.env$mv0$state),
          .vars,
          setNames(
            c("linCmtA" = 1L, "linCmtB" = 2L)[match.arg(linCmtSens)],
            NULL
          ), verbose
        ))
      } else {
        .env$mvL <- NULL
      }
      .env$curCmt <- length(.env$mv0$state)
      .env$extraCmt <- NULL
      if (!is.null(.env$predDf)) {
        .env$predDf$cmt <- vapply(seq_along(.env$predDf$line),
                                  function(i) {
                                    .cmtName <- .env$predDf$cond[i]
                                    .w <- which(.cmtName == .env$mv0$state)
                                    if (length(.w) == 1L) return(.w)
                                    assign("curCmt", .env$curCmt + 1L, envir=.env)
                                    assign("extraCmt", c(.env$extraCmt,
                                                         paste0("cmt(", .cmtName, ")")),
                                           envir=.env)
                                    .env$curCmt
                                  }, integer(1))
      }
      .env$extraDvid <- paste0("dvid(", paste(.env$predDf$cmt, collapse = ","), ")")
      .env$mixProbs <- .udfUiEnv$probs
      # Cleanup the environment
      .rm <- intersect(c("curCondition", "curDvid", "curVar", "df",
                         "errTypeInfo", "err", "hasNonErrorTerm", "isAnAdditiveExpression",
                         "lastDistAssign", "line", "needsToBeAnErrorExpression",
                         "needToDemoteAdditiveExpression",
                         "top", "trLimit", ".numeric", "a", "b", "c", "d", "e", "f",  "lambda",
                         "curCmt", "errGlobal", "linCmt", "ll", "distribution", "rxUdfUiCount", "before", "after",
                         "lhs", "earlyErr", "var", "dv", "hasAr"),
                       ls(envir=.env, all.names=TRUE))
      if (length(.rm) > 0) rm(list=.rm, envir=.env)
      if (checkMissing) .checkForMissingOrDupliacteInitials(.env)
      return(.env)
    }
  }
  return(invisible(NULL))
}

#' Determine if expression is a rxode2 error expression
#'
#' Currently only handles simple error expressions, not ll(var) ~ ...
#'
#' @param expr Expression to test
#' @param uiEnv Return the uiEnv for the single expression
#' @return Boolean that says if the expression is an error expression or not
#' @author Matthew L. Fidler
#' @examples
#'
#' .isErrorExpression(quote(ipre~add(add.sd)))
#' .isErrorExpression(quote(ipre~add(add.sd)+3))
#'
#' @noRd
.isErrorExpression <- function(expr, uiEnv=FALSE) {
  if (!identical(expr[[1]], quote(`~`))) return(FALSE)
  if (identical(expr[[2]], quote(linCmt()))) {
    .pre <- "linCmt()"
    .var <- "cl=abc\nv=d\nrxDummyVarNotUsedInModel"
    .lin <- TRUE
  } else {
    .expr2 <- expr[[2]]
    if (length(.expr2) == 2L) {
      if (identical(.expr2[[1]], quote(`-`)) &&
            is.name(.expr2[[2]])) {
        return(TRUE)
      }
    }
    .pre <- allNames(.expr2)
    if (length(.pre) != 1L) return(FALSE)
    .var <- .pre
    .lin <- FALSE
  }
  .mod <- eval(parse(text=paste0("quote({", .var, "=1+2\n", deparse1(expr), "})")))
  .ini <- as.data.frame(eval(parse(text=paste0("lotri({\n",
                                               paste(paste(allNames(expr[[3]]), "<- 1"), collapse="\n"),
                                               "\n})"))))
  .env <- try(.errProcessExpression(.mod, .ini, checkMissing=FALSE), silent=TRUE)
  if (inherits(.env, "try-error")) return(FALSE)
  if (uiEnv) return(.env)
  ifelse(.lin, "rxLinCmt" == .env$predDf$var, .pre == .env$predDf$var)
}
#'  Is Normal, Cauchy or t distribution model specification?
#'
#' @param expr Expression
#'
#' @return TRUE if this is a normal/t/cauchy model, FALSE otherwise
#' @author Matthew L. Fidler
#' @noRd
.isNormOrTErrorExpression <- function(expr) {
  .env <- .isErrorExpression(expr, TRUE)
  if (inherits(.env, "logical")) return(FALSE)
  any(.env$predDf$distribution == c("norm", "t", "cauchy"))
}
#' Throw an error if the error expression  is invalid
#'
#' @param expr Error expression from a model({}) or related pipe
#' @return Nothing, called for side effect of an error for a bad pipe
#' @author Matthew L. Fidler
#' @noRd
.throwIfInvalidTilde <- function(expr) {
  .env <- .isErrorExpression(expr, TRUE)
  .env$isAnAdditiveExpression <- FALSE
  .env$ll <- FALSE
  .env$linCmt <- FALSE
  .left <- .errHandleLlOrLinCmt(expr[[2]], .env)
  .env$trLimit <- c(-Inf, Inf)
  .env$a <- .env$b <- .env$c <- .env$d <- .env$e <- .env$f <- .env$lambda <- NA_character_
  .env$hasAr <- FALSE
  .env$curCondition <- .env$curVar <- deparse1(.left)
  .env$hasNonErrorTerm <- FALSE
  .env$needsToBeAnErrorExpression <- FALSE
  .env$needToDemoteAdditiveExpression <- FALSE
  .right <- .errHandleCondition(expr[[3]], .env)
  .env$isAnAdditiveExpression <- FALSE
  .env$errTypeInfo <- rxErrTypeCombine("")
  .env$distribution <- "norm"
  .env$estNotAllowed <- FALSE
  .errHandleErrorStructure(.right, .env)
  if (!is.null(.env$err)) {
    stop(paste(.env$err, collapse="\n"), call.=FALSE)
  }
}

#' Get defined residual information for parsing elsewhere
#'
#' @return A list of error distribution information
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.getErrDist <- function() {
  .errDist
}
#' @export
#' @rdname dot-getErrDist
.getNamedArgumentsToPredDf <- function() {
  .namedArgumentsToPredDf
}

rxTest({

  lmat <- lotri({
    ## You may label each parameter with a comment
    tka <- 0.45 # Log Ka
    tcl <- log(c(0, 2.7, 100)) # Log Cl
    ## This works with interactive models
    ## You may also label the preceding line with label("label text")
    tv <- 3.45; label("log V")
    tvp <- 3.45; label("log V")
    cl.wt <- 0.1
    v.wt <- 0.1
    cl.sex <- 0.1
    v.sex <- 0.1
    cl.age <- 0.1
    v.age <- 0.1
    vp.wt <- 1
    vp.sex <- 1
    vp.age <- 1
    ## the label("Label name") works with all models
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
    pow.sd <- 0.2
    pow  <- 1
    lambda  <- 0.5
  })

  expect_err2 <- function(x, extra=FALSE) {
    if (is.na(extra)){
      expect_false(x$hasErrors)
    } else {
      expect_true(x$hasErrors)
    }
  }

  test_that("log-likelihood variables are captured", {
    expect_error(
      .errProcessExpression(quote({
        ka <- exp(tka)
        cl <- exp(tcl)
        v <- exp(tv)
        cp <- linCmt()
        ll(err) ~ -log(add.sd) - 0.5*log(2*pi) - 0.5*((DV-cp)/add.sd)^2
      }),
      lotri({
        ## You may label each parameter with a comment
        tka <- 0.45 # Ka
        tcl <- log(c(0, 2.7, 100)) # Log Cl
        ## This works with interactive models
        ## You may also label the preceding line with label("label text")
        tv <- 3.45; label("log V")
        ## the label("Label name") works with all models
        add.sd <- 0.7
      })),
      NA
    )
  })

  test_that("When checking for distributions, don't consider if the parameter is defined", {
    expect_true(.isErrorExpression(ipre ~ add(f2) + propF(prop.sd, f2)))
  })

  test_that("error when errors have too many arguments", {
    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ add(add.sd, pow1) + pow(pow.sd, pow2) + boxCox(lambda) | cond
    }), lmat))
  })

  test_that("error when adding distributions that do not support the additive notation", {
    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ dbinom(add.sd, pow) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat))
  })

  test_that("error for specifying distributions that have multiple numbers of arguments", {
    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      p <- pow
      l <- lamba
      cp ~ dbinom(add.sd, p, l)
      center ~ pow(pow.sd, p) + boxCox(l) | cond
    }), lmat))
  })

  test_that("error when adding algebraic expressions to known distributional abbreviations", {
    expect_error(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ add(add.sd) + pow(pow.sd, pow) + boxCox(lambda) + tan(vp)| cond
    }), lmat))
  })

  test_that("The distribution names will transform to the preferred distributions", {
    expect_equal(rxPreferredDistributionName("add"), "add")
    expect_equal(rxPreferredDistributionName("logitNorm"), "logitNorm")
  })

  test_that("non-numeric bounds for logitNorm and probitNorm", {

    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ logitNorm(add.sd, lower, upper) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat))

    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ logitNorm(add.sd) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat), NA)

    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ logitNorm(add.sd, 0) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat), NA)

    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ logitNorm(add.sd, 0, 5) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat), NA)

    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ logitNorm(add.sd, -5, 5) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat), NA)

    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ logitNorm(add.sd, -Inf, Inf) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat), NA)

    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ probitNorm(add.sd, lower, upper) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat))

  })

  test_that("rxTransformCombine", {

    testCombine0 <- function(transforms) {
      .cmb <- rxErrTypeCombine("matt")
      for (i in transforms) {
        .cmb <- rxErrTypeCombine(.cmb, i)
      }
      .cmb
    }

    testCombine <- function(transforms) {
      .cmb <- testCombine0(transforms)
      .cmb2 <- testCombine0(rev(transforms))
      expect_equal(.cmb, .cmb2)
      .cmb
    }

    ## add + prop
    expect_equal(testCombine(c("add", "prop"))$errType,
                 testCombine(c("add", "propT"))$errType)

    expect_equal(testCombine(c("add", "prop"))$errType,
                 testCombine(c("add", "propF"))$errType)

    expect_equal(testCombine(c("logitNorm", "prop"))$errType,
                 testCombine(c("add", "prop"))$errType)

    expect_equal(testCombine(c("logitNorm", "propT"))$errType,
                 testCombine(c("add", "prop"))$errType)

    expect_equal(testCombine(c("logitNorm", "propF"))$errType,
                 testCombine(c("add", "prop"))$errType)

    expect_equal(testCombine(c("probitNorm", "prop"))$errType,
                 testCombine(c("add", "prop"))$errType)

    expect_equal(testCombine(c("probitNorm", "propT"))$errType,
                 testCombine(c("add", "prop"))$errType)

    expect_equal(testCombine(c("probitNorm", "propF"))$errType,
                 testCombine(c("add", "prop"))$errType)

    ## add + pow
    expect_equal(testCombine(c("add", "pow"))$errType,
                 testCombine(c("add", "powF"))$errType)

    expect_equal(testCombine(c("lnorm", "pow"))$errType,
                 testCombine(c("add", "pow"))$errType)

    expect_equal(testCombine(c("logitNorm", "pow"))$errType,
                 testCombine(c("add", "pow"))$errType)

    expect_equal(testCombine(c("probitNorm", "pow"))$errType,
                 testCombine(c("add", "pow"))$errType)

    # Test Error type F
    expect_equal(testCombine(c("add", "powF"))$errTypeF,
                 testCombine(c("add", "propF"))$errTypeF)

    expect_equal(testCombine(c("add", "powT"))$errTypeF,
                 testCombine(c("add", "propT"))$errTypeF)

    expect_equal(testCombine(c("add", "pow"))$errTypeF,
                 testCombine(c("add", "prop"))$errTypeF)

    expect_equal(testCombine("powF")$errTypeF,
                 testCombine("propF")$errTypeF)

    expect_equal(testCombine("powT")$errTypeF,
                 testCombine("propT")$errTypeF)

    expect_equal(testCombine("pow")$errTypeF,
                 testCombine("prop")$errTypeF)

    # Test Transformation Type
    expect_equal(testCombine(c("add", "prop", "boxCox"))$transform,
                 testCombine(c("add", "propT", "boxCox"))$transform)

    expect_equal(testCombine(c("add", "prop", "boxCox"))$transform,
                 testCombine(c("add", "propF", "boxCox"))$transform)

    expect_equal(testCombine(c("add", "pow", "boxCox"))$transform,
                 testCombine(c("add", "propF", "boxCox"))$transform)

    expect_equal(testCombine(c("pow", "boxCox"))$transform,
                 testCombine(c("add", "boxCox"))$transform)

    expect_equal(testCombine(c("prop", "boxCox"))$transform,
                 testCombine(c("add", "boxCox"))$transform)

    expect_equal(testCombine(c("lnorm", "prop", "boxCox")),
                 testCombine(c("lnorm", "propT", "boxCox")))

    expect_equal(testCombine(c("lnorm", "prop", "boxCox")),
                 testCombine(c("lnorm", "propF", "boxCox")))

    expect_equal(testCombine(c("lnorm", "pow", "boxCox")),
                 testCombine(c("lnorm", "propF", "boxCox")))

    expect_equal(testCombine(c("logitNorm", "prop"))$transform,
                 testCombine(c("logitNorm", "propT"))$transform)

    expect_equal(testCombine(c("logitNorm", "add"))$transform,
                 testCombine(c("logitNorm", "propT"))$transform)

    expect_equal(testCombine(c("logitNorm", "prop", "boxCox"))$transform,
                 testCombine(c("logitNorm", "propT", "boxCox"))$transform)

    expect_equal(testCombine(c("logitNorm", "prop", "boxCox"))$transform,
                 testCombine(c("logitNorm", "propF", "boxCox"))$transform)

    expect_equal(testCombine(c("logitNorm", "pow", "boxCox"))$transform,
                 testCombine(c("logitNorm", "propF", "boxCox"))$transform)

    expect_equal(testCombine(c("probitNorm", "prop", "boxCox"))$transform,
                 testCombine(c("probitNorm", "propT", "boxCox"))$transform)

    expect_equal(testCombine(c("probitNorm", "prop", "boxCox"))$transform,
                 testCombine(c("probitNorm", "propF", "boxCox"))$transform)

    expect_equal(testCombine(c("probitNorm", "pow", "boxCox"))$transform,
                 testCombine(c("probitNorm", "propF", "boxCox"))$transform)

    # Yeo Johnson
    expect_equal(testCombine(c("add", "prop", "yeoJohnson"))$transform,
                 testCombine(c("add", "propT", "yeoJohnson"))$transform)

    expect_equal(testCombine(c("add", "prop", "yeoJohnson"))$transform,
                 testCombine(c("add", "propF", "yeoJohnson"))$transform)

    expect_equal(testCombine(c("add", "pow", "yeoJohnson"))$transform,
                 testCombine(c("add", "propF", "yeoJohnson"))$transform)

    expect_equal(testCombine(c("pow", "yeoJohnson"))$transform,
                 testCombine(c("add", "yeoJohnson"))$transform)

    expect_equal(testCombine(c("prop", "yeoJohnson"))$transform,
                 testCombine(c("add", "yeoJohnson"))$transform)

    expect_equal(testCombine(c("lnorm", "prop", "yeoJohnson")),
                 testCombine(c("lnorm", "propT", "yeoJohnson")))

    expect_equal(testCombine(c("lnorm", "prop", "yeoJohnson")),
                 testCombine(c("lnorm", "propF", "yeoJohnson")))

    expect_equal(testCombine(c("lnorm", "pow", "yeoJohnson")),
                 testCombine(c("lnorm", "propF", "yeoJohnson")))

    expect_equal(testCombine(c("logitNorm", "prop", "yeoJohnson"))$transform,
                 testCombine(c("logitNorm", "propT", "yeoJohnson"))$transform)

    expect_equal(testCombine(c("logitNorm", "prop", "yeoJohnson"))$transform,
                 testCombine(c("logitNorm", "propF", "yeoJohnson"))$transform)

    expect_equal(testCombine(c("logitNorm", "pow", "yeoJohnson"))$transform,
                 testCombine(c("logitNorm", "propF", "yeoJohnson"))$transform)

    expect_equal(testCombine(c("probitNorm", "prop", "yeoJohnson"))$transform,
                 testCombine(c("probitNorm", "propT", "yeoJohnson"))$transform)

    expect_equal(testCombine(c("probitNorm", "prop", "yeoJohnson"))$transform,
                 testCombine(c("probitNorm", "propF", "yeoJohnson"))$transform)

    expect_equal(testCombine(c("probitNorm", "pow", "yeoJohnson"))$transform,
                 testCombine(c("probitNorm", "propF", "yeoJohnson"))$transform)

    # Box Cox and Yeo Johnson are not compatible
    expect_equal(testCombine(c("boxCox", "prop", "yeoJohnson")),
                 testCombine(c("boxCox", "propT", "yeoJohnson")))

    # Prop and pow are not compatible
    expect_equal(testCombine(c("pow", "prop")),
                 testCombine(c("pow", "prop")))

    # Make sure demotion works
    expect_equal(rxDemoteAddErr(testCombine(c("add", "prop"))),
                 testCombine("prop"))

    expect_equal(rxDemoteAddErr(testCombine(c("add", "pow"))),
                 testCombine("pow"))

    .errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ logitNorm(add.sd) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat) -> mod

    expect_equal(mod$predDf$errType, testCombine(c("pow", "add"))$errType)

    .errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v + add.sd
      cp ~ logitNorm(NA) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat) -> mod

    expect_equal(mod$predDf$errType, testCombine("pow")$errType)

    .errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v + add.sd + lambda
      cp ~ lnorm(NA) + pow(pow.sd, pow) | cond
    }), lmat) -> mod

    expect_equal(mod$predDf$errType, testCombine("pow")$errType)

    .errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v + add.sd + lambda
      cp ~ lnorm(add.sd) + pow(pow.sd, pow)  | cond
    }), lmat) -> mod

    expect_equal(mod$predDf$errType, testCombine(c("add", "pow"))$errType)

    .errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v + add.sd
      cp ~ probitNorm(NA) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat) -> mod

    expect_equal(mod$predDf$errType, testCombine("pow")$errType)

    .errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ probitNorm(add.sd) + pow(pow.sd, pow) + boxCox(lambda) | cond
    }), lmat) -> mod

    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ probitNorm(add.sd) + pow(pow.sd, pow) + boxCox(lambda) + var() | cond
    }), lmat), NA)

    expect_equal(mod$predDf$errType, testCombine(c("add", "pow"))$errType)

    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ probitNorm(add.sd) + pow(pow.sd, pow) + boxCox(lambda) + var() | cond
    }), lmat), NA)

  })

  test_that("categorical expressions", {

    expect_err2(.errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ c(add.sd, pow.sd, pow, lambda) | cond
    }), lmat), NA)

    .errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ c(add.sd, pow.sd, pow, lambda) | cond
    }), lmat) -> mod

    testOrd <- mod$iniDf[which(mod$iniDf$condition == "cond"),c("name","err")]
    row.names(testOrd) <- NULL

    expect_equal(
      testOrd,
      structure(
        list(
          name = c("add.sd", "pow.sd", "pow", "lambda"),
          err = c("ordinal", "ordinal2", "ordinal3", "ordinal4")),
        row.names = c(NA, -4L),
        class = "data.frame")
    )

  })

  test_that("theta/eta/eps problems", {

    .errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v + add.sd + pow.sd + pow + lambda
      # Nonsense parameters just to check calculated parameters being used in the error expression
      a = tka + eta.ka
      b = tka + eta.ka
      c = tka + eta.ka
      d = tka + eta.ka
      e = tka + eta.ka
      f = tka + eta.ka
      l = tka + eta.ka
      cp ~ add(a) + powF(b, c, f) + t(d, e) + boxCox(l)| cond
    }), lmat) -> mod

    expect_equal(mod$predDf[, c("a", "b", "c", "d", "e", "lambda")],
                 structure(list(a = "a", b = "b", c = "c", d = "d", e = "e", lambda = "l"), class = "data.frame", row.names = c(NA, -1L)))

    .errProcessExpression(quote({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v+ add.sd + pow.sd + pow + lambda
      cp ~ add(a) + powF(b, c, f) + t(d, e) + boxCox(l)| cond
    }), lmat) -> mod

  })

  test_that("multiple endpoint parsing", {

    lmat <- lotri({
      tktr <- log(1)
      tka <- log(1)
      tcl <- log(0.1)
      tv <- log(10)
      ##
      eta.ktr ~ 1
      eta.ka ~ 1
      eta.cl ~ 2
      eta.v ~ 1
      prop.err <- 0.1
      pkadd.err <- 0.1
      ##
      temax <- logit(0.8)
      #temax <- 7.5
      tec50 <- log(0.5)
      tkout <- log(0.05)
      te0 <- log(100)
      ##
      eta.emax ~ .5
      eta.ec50  ~ .5
      eta.kout ~ .5
      eta.e0 ~ .5
      ##
      pdadd.err <- 10
    })

    .errProcessExpression(quote({
      ktr <- exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      ##
      #poplogit = log(temax/(1-temax))
      emax=expit(temax+eta.emax)
      #logit=temax+eta.emax
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin = e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ add(pdadd.err)
    }), lmat) -> mod

    expect_equal(mod$predDf[, c("cond", "var", "dvid", "cmt")],
                 structure(list(cond = c("cp", "effect"), var = c("cp", "effect"), dvid = 1:2, cmt = 5:4), class = "data.frame", row.names = c(NA, -2L)))

    .errProcessExpression(quote({
      ktr <- exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      ##
      emax=expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin = e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err) | center
      effect ~ add(pdadd.err)
    }), lmat) -> mod

    expect_equal(mod$predDf[, c("cond", "var", "dvid", "cmt")],
                 structure(list(cond = c("center", "effect"), var = c("cp", "effect"), dvid = 1:2, cmt = 3:4),
                           class = "data.frame", row.names = c(NA, -2L)))

    .errProcessExpression(quote({
      ktr <- exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin = e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(mod$predDf[, c("cond", "var", "dvid", "cmt")],
                 structure(list(cond = c("cp", "pca"), var = c("cp", "effect"), dvid = 1:2, cmt = 5:6),
                           class = "data.frame", row.names = c(NA, -2L)))

  })

  # Try hidden expressions combined with error expression
  test_that("test expressions that are hidden with ~ vs error expressions with ~", {

    lmat <- lotri({
      tktr <- log(1)
      tka <- log(1)
      tcl <- log(0.1)
      tv <- log(10)
      ##
      eta.ktr ~ 1
      eta.ka ~ 1
      eta.cl ~ 2
      eta.v ~ 1
      prop.err <- 0.1
      pkadd.err <- 0.1
      ##
      temax <- logit(0.8)
      #temax <- 7.5
      tec50 <- log(0.5)
      tkout <- log(0.05)
      te0 <- log(100)
      ##
      eta.emax ~ .5
      eta.ec50  ~ .5
      eta.kout ~ .5
      eta.e0 ~ .5
      ##
      pdadd.err <- 10
    })

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(mod$predDf[, c("cond", "var", "dvid", "cmt")],
                 structure(list(cond = c("cp", "pca"), var = c("cp", "effect"), dvid = 1:2, cmt = 5:6),
                           class = "data.frame", row.names = c(NA, -2L)))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(DCP)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(mod$predDf[, c("cond", "var", "dvid", "cmt")],
                 structure(list(cond = c("cp", "pca"), var = c("cp", "effect"), dvid = 1:2, cmt = 5:6),
                           class = "data.frame", row.names = c(NA, -2L)))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(
      mod$predDf[, c("cond", "var", "dvid", "cmt")],
      structure(
        list(cond = c("cp", "pca"),
             var = c("cp", "effect"),
             dvid = 1:2,
             cmt = 5:6
             ),
        class = "data.frame", row.names = c(NA, -2L)
      )
    )

  })

  test_that("test different distributions", {

    lmat <- lotri({
      tktr <- log(1)
      tka <- log(1)
      tcl <- log(0.1)
      tv <- log(10)
      ##
      eta.ktr ~ 1
      eta.ka ~ 1
      eta.cl ~ 2
      eta.v ~ 1
      prop.err <- 0.1
      pkadd.err <- 0.1
      ##
      temax <- logit(0.8)
      #temax <- 7.5
      tec50 <- log(0.5)
      tkout <- log(0.05)
      te0 <- log(100)
      ##
      eta.emax ~ .5
      eta.ec50  ~ .5
      eta.kout ~ .5
      eta.e0 ~ .5
      ##
      pdadd.err <- 10
    })

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ dchisq(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("norm", "chisq"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ chisq(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("norm", "chisq"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ dexp(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("norm", "dexp"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ df(prop.err, pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("f", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ f(prop.err, pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("f", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v + pdadd.err
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ pois(lambda)
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("norm", "pois"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v + pdadd.err
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ dpois(lambda)
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("norm", "pois"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v + pdadd.err
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ dbinom(pdadd.err)
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("norm", "binom"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v + pdadd.err
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ binom(pdadd.err)
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("norm", "binom"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ dbeta(prop.err, pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("beta", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ beta(prop.err, pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("beta", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ geom(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("norm", "geom"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err)
      effect ~ dgeom(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("norm", "geom"))

    ## .errProcessExpression(quote({
    ##   ktr ~ exp(tktr + eta.ktr)
    ##   ka <- exp(tka + eta.ka)
    ##   cl <- exp(tcl + eta.cl)
    ##   v <- exp(tv + eta.v)
    ##   emax = expit(temax+eta.emax)
    ##   ec50 =  exp(tec50 + eta.ec50)
    ##   kout = exp(tkout + eta.kout)
    ##   e0 = exp(te0 + eta.e0)
    ##   ##
    ##   DCP = center/v
    ##   tmp ~ exp(eta.tr)
    ##   PD=1-emax*DCP/(ec50+DCP)
    ##   ##
    ##   effect(0) = e0
    ##   kin ~ e0*kout
    ##   ##
    ##   d/dt(depot) = -ktr * depot
    ##   d/dt(gut) =  ktr * depot -ka * gut
    ##   d/dt(center) =  ka * gut - cl / v * center
    ##   d/dt(effect) = kin*PD -kout*effect
    ##   ##
    ##   cp = center / v
    ##   cp ~ hyper(prop.err, pkadd.err, pdadd.err)
    ##   effect ~ dpois(cp) | pca
    ## }), lmat) -> mod

    ## expect_equal(paste(mod$predDf$distribution), c("hyper", "pois"))

    ## "unif", #11

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ unif(prop.err, pkadd.err)
      effect ~ add(cp) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("unif", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ dunif(prop.err, pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("unif", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ weibull(prop.err, pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("weibull", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ dweibull(prop.err, pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("weibull", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ c(prop.err, pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("ordinal", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ c(prop.err, pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("ordinal", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ dt(prop.err, pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("t", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ t(prop.err, pkadd.err)
      effect ~ add(pdadd.err) | pca
    }), lmat) -> mod

    expect_equal(paste(mod$predDf$distribution), c("t", "norm"))

    lmat2 <- lotri({
      tktr <- log(1)
      tka <- log(1)
      tcl <- log(0.1)
      tv <- log(10)
      ##
      eta.ktr ~ 1
      eta.ka ~ 1
      eta.cl ~ 2
      eta.v ~ 1
      prop.err <- 0.1
      pkadd.err <- 0.1
      ##
      temax <- logit(0.8)
      #temax <- 7.5
      tec50 <- log(0.5)
      tkout <- log(0.05)
      te0 <- log(100)
      ##
      eta.emax ~ .5
      eta.ec50  ~ .5
      eta.kout ~ .5
      eta.e0 ~ .5
      ##
      pdadd.err <- 10
      df <- c(1, 10)
    })

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err) + dt(df)
      effect ~ add(pdadd.err) | pca
    }), lmat2) -> mod

    expect_equal(paste(mod$predDf$distribution), c("t", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      cp ~ prop(prop.err) + add(pkadd.err) + dt(df)
      effect ~ add(pdadd.err) | pca
    }), lmat2) -> mod

    expect_equal(paste(mod$predDf$distribution), c("t", "norm"))

    .errProcessExpression(quote({
      ktr ~ exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      emax = expit(temax+eta.emax)
      ec50 =  exp(tec50 + eta.ec50)
      kout = exp(tkout + eta.kout)
      e0 = exp(te0 + eta.e0)
      ##
      DCP = center/v
      tmp ~ exp(eta.tr)
      PD=1-emax*DCP/(ec50+DCP)
      ##
      effect(0) = e0
      kin ~ e0*kout
      ##
      d/dt(depot) = -ktr * depot
      d/dt(gut) =  ktr * depot -ka * gut
      d/dt(center) =  ka * gut - cl / v * center
      d/dt(effect) = kin*PD -kout*effect
      ##
      cp = center / v
      ll(cp) ~ log(prop.err) + log(cp)
      effect ~ add(pdadd.err) | pca
    }), lmat2) -> mod

  })

  test_that(".isErrorExpression", {
    expect_true(.isErrorExpression(quote(-cp ~ .)))
    expect_false(.isErrorExpression(quote(lipre~add(add.err) + 3*x | matt)))
    expect_true(.isErrorExpression(quote(lipre~add(add.err) | matt)))
    expect_true(.isErrorExpression(quote(lipre~add(add.err) + prop(prop.sd) + boxCox(lambda)| matt)))
    expect_true(.isErrorExpression(quote(lipre~add(add.err) + prop(prop.sd) + boxCox(lambda))))
    expect_false(.isErrorExpression(quote(lipre <- add(add.err) + prop(prop.sd) + boxCox(lambda))))
    expect_false(.isErrorExpression(quote(lipre + 2 <- add(add.err) + prop(prop.sd) + boxCox(lambda))))
    expect_true(.isErrorExpression(quote(linCmt() ~ add(add.err) + prop(prop.err))))

  })

  test_that("symbol alone", {

    whoops <- function() {
      ini({
        paramAlone <- 1
        paramUsed <- 2
      })
      model({
        x <- paramUsed
        paramAlone
      })
    }

    expect_message(expect_error(rxode2(whoops)),
                  "the symbol 'paramAlone' cannot be by itself")

  })

  test_that("complex error condition", {

    f <- function() {
      ini({
        paramUsed <- 1
        add.sd <- 0.1
        add.sd2 <- 0.2
      })
      model({
        x <- paramUsed
        x1 <- x
        x2 <- x
        x1 ~ add(add.sd) | cmt(1)
        x2 ~ add(add.sd2) | cmt(2)
      })
    }

    expect_message(expect_error(rxode2(f)),
                   "the condition 'cmt\\(2\\)' must be a simple name")

  })


})

*----------------------------------------------------------------------|
      subroutine DGPADM( ideg,m,t,H,ldh,wsp,lwsp,ipiv,iexph,ns,iflag )

      implicit none
      integer ideg, m, ldh, lwsp, iexph, ns, iflag, ipiv(m)
      double precision t, H(ldh,m), wsp(lwsp)

*-----Purpose----------------------------------------------------------|
*
*     Computes exp(t*H), the matrix exponential of a general matrix in
*     full, using the irreducible rational Pade approximation to the 
*     exponential function exp(x) = r(x) = (+/-)( I + 2*(q(x)/p(x)) ),
*     combined with scaling-and-squaring.
*
*-----Arguments--------------------------------------------------------|
*
*     ideg      : (input) the degre of the diagonal Pade to be used.
*                 a value of 6 is generally satisfactory.
*
*     m         : (input) order of H.
*
*     H(ldh,m)  : (input) argument matrix.
*
*     t         : (input) time-scale (can be < 0).
*                  
*     wsp(lwsp) : (workspace/output) lwsp .ge. 4*m*m+ideg+1.
*
*     ipiv(m)   : (workspace)
*
*>>>> iexph     : (output) number such that wsp(iexph) points to exp(tH)
*                 i.e., exp(tH) is located at wsp(iexph ... iexph+m*m-1)
*                       ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
*                 NOTE: if the routine was called with wsp(iptr), 
*                       then exp(tH) will start at wsp(iptr+iexph-1).
*
*     ns        : (output) number of scaling-squaring used.
*
*     iflag     : (output) exit flag.
*                      0 - no problem
*                     <0 - problem
*
*----------------------------------------------------------------------|
*     Roger B. Sidje (rbs@maths.uq.edu.au)
*     EXPOKIT: Software Package for Computing Matrix Exponentials.
*     ACM - Transactions On Mathematical Software, 24(1):130-156, 1998
*----------------------------------------------------------------------|
*
      integer mm,i,j,k,ih2,ip,iq,iused,ifree,iodd,icoef,iput,iget
      double precision hnorm,scale,scale2,cp,cq

      intrinsic INT,ABS,DBLE,LOG,MAX

*---  check restrictions on input parameters ...
      mm = m*m
      iflag = 0
      if ( ldh.lt.m ) iflag = -1
      if ( lwsp.lt.4*mm+ideg+1 ) iflag = -2
      if ( iflag.ne.0 ) call rexit('bad sizes (in input of DGPADM)')
*
*---  initialise pointers ...
*
      icoef = 1
      ih2 = icoef + (ideg+1)
      ip  = ih2 + mm
      iq  = ip + mm
      ifree = iq + mm
*
*---  scaling: seek ns such that ||t*H/2^ns|| < 1/2; 
*     and set scale = t/2^ns ...
*
      do i = 1,m
         wsp(i) = 0.0d0
      enddo
      do j = 1,m
         do i = 1,m
            wsp(i) = wsp(i) + ABS( H(i,j) )
         enddo
      enddo
      hnorm = 0.0d0
      do i = 1,m
         hnorm = MAX( hnorm,wsp(i) )
      enddo
      hnorm = ABS( t*hnorm )
      if (hnorm.eq.0.0d0 ) then
         call rexit('Error - null H in input of DGPADM.')
      endif
      ns = MAX( 0,INT(LOG(hnorm)/LOG(2.0d0))+2 )
      scale = t / DBLE(2**ns)
      scale2 = scale*scale
*
*---  compute Pade coefficients ...
*
      i = ideg+1
      j = 2*ideg+1
      wsp(icoef) = 1.0d0
      do k = 1,ideg
         wsp(icoef+k) = (wsp(icoef+k-1)*DBLE( i-k ))/DBLE( k*(j-k) )
      enddo
*
*---  H2 = scale2*H*H ...
*
      call DGEXX( 'n','n',m,m,m,scale2,H,ldh,H,ldh,0.0d0,wsp(ih2),m )
*
*---  initialize p (numerator) and q (denominator) ...
*
      cp = wsp(icoef+ideg-1)
      cq = wsp(icoef+ideg)
      do j = 1,m
         do i = 1,m
            wsp(ip + (j-1)*m + i-1) = 0.0d0
            wsp(iq + (j-1)*m + i-1) = 0.0d0
         enddo
         wsp(ip + (j-1)*(m+1)) = cp
         wsp(iq + (j-1)*(m+1)) = cq
      enddo
*
*---  Apply Horner rule ...
*
      iodd = 1
      k = ideg - 1
 100  continue
      iused = iodd*iq + (1-iodd)*ip
      call DGEXX( 'n','n',m,m,m, 1.0d0,wsp(iused),m,
     .             wsp(ih2),m, 0.0d0,wsp(ifree),m )
      do j = 1,m
         wsp(ifree+(j-1)*(m+1)) = wsp(ifree+(j-1)*(m+1))+wsp(icoef+k-1)
      enddo
      ip = (1-iodd)*ifree + iodd*ip
      iq = iodd*ifree + (1-iodd)*iq
      ifree = iused
      iodd = 1-iodd
      k = k-1
      if ( k.gt.0 )  goto 100
*
*---  Obtain (+/-)(I + 2*(p\q)) ...
*
      if ( iodd .eq. 1 ) then
         call DGEXX( 'n','n',m,m,m, scale,wsp(iq),m,
     .                H,ldh, 0.0d0,wsp(ifree),m )
         iq = ifree
      else
         call DGEXX( 'n','n',m,m,m, scale,wsp(ip),m,
     .                H,ldh, 0.0d0,wsp(ifree),m )
         ip = ifree
      endif
      call DAXPY( mm, -1.0d0,wsp(ip),1, wsp(iq),1 )
      call DGESV( m,m, wsp(iq),m, ipiv, wsp(ip),m, iflag )
      if ( iflag.ne.0 ) call rexit('Problem in DGESV (within DGPADM)')
      call DSCAL( mm, 2.0d0, wsp(ip), 1 )
      do j = 1,m
         wsp(ip+(j-1)*(m+1)) = wsp(ip+(j-1)*(m+1)) + 1.0d0
      enddo
      iput = ip
      if ( ns.eq.0 .and. iodd.eq.1 ) then
         call DSCAL( mm, -1.0d0, wsp(ip), 1 )
         goto 200
      endif
*
*--   squaring : exp(t*H) = (exp(t*H))^(2^ns) ...
*
      iodd = 1
      do k = 1,ns
         iget = iodd*ip + (1-iodd)*iq
         iput = (1-iodd)*ip + iodd*iq
         call DGEXX( 'n','n',m,m,m, 1.0d0,wsp(iget),m, wsp(iget),m,
     .                0.0d0,wsp(iput),m )
         iodd = 1-iodd
      enddo
 200  continue
      iexph = iput
      END
*----------------------------------------------------------------------|

rxTest({
  test_that("rxS for tlast() shouldn't error", {
    expect_error(rxS("d/dt(depot)=exp(log(bio*podo())+log(ktr)+n*log(ktr*tad(depot))-ktr*tad(depot)-lgamma(n+1))-ka*depot"), NA)
    expect_error(rxS("d/dt(depot)=exp(log(bio*podo())+log(ktr)+n*log(ktr*tad)-ktr*tad-lgamma(n+1))-ka*depot"), NA)
  })
})

# Test large model compiles (RxODE#117)
rxTest({
  mod <- rxode2("
##### define heart failure parameters. Allow them to increase from baseline over time #####
if (Heart_failure_link == 1) {
R_art0		  = (R_art0_initial-R_art0_initial*heart_failure_resistance_aorta_scale)/exp(sim_time/(24*5))+R_art0_initial*heart_failure_resistance_aorta_scale;
R_per0	  	  = (R_per0_initial-R_per0_initial*heart_failure_resistance_prepheral_scale)/exp(sim_time/(24*5))+R_per0_initial*heart_failure_resistance_prepheral_scale;
R_al0 = R_al0_scale*R_per0;    #0.66*R_per0
R_cap0= R_cap0_scale*R_per0;    #0.25*R_per0
R_vn0 = R_vn0_scale*R_per0;   #0.09*R_per0
R_ven0	  	= (R_ven0_initial-R_ven0_initial*heart_failure_resistance_venous_scale)/exp(sim_time/(24*5))+R_ven0_initial*heart_failure_resistance_venous_scale;
R_art_pulm	= (R_art_pulm_initial-R_art_pulm_initial*heart_failure_resistance_R_art_pulm_scale)/exp(sim_time/(24*5))+R_art_pulm_initial*heart_failure_resistance_R_art_pulm_scale;
R_ven_pulm	= (R_ven_pulm_initial-R_ven_pulm_initial*heart_failure_resistance_R_ven_pulm_scale)/exp(sim_time/(24*5))+R_ven_pulm_initial*heart_failure_resistance_R_ven_pulm_scale;

C_art	    	= (C_art_initial-C_art_initial*heart_failure_compliance_aorta_scale)/exp(sim_time/(24*5))+C_art_initial*heart_failure_compliance_aorta_scale;
C_ven0	  	= (C_ven0_initial-C_ven0_initial*heart_failure_compliance_venous_scale)/exp(sim_time/(24*5))+C_ven0_initial*heart_failure_compliance_venous_scale;
C_pulm_ven  = (C_pulm_ven_initial-C_pulm_ven_initial*heart_failure_compliance_pulm_ven_scale)/exp(sim_time/(24*5))+C_pulm_ven_initial*heart_failure_compliance_pulm_ven_scale;
C_pulm_art  = (C_pulm_art_initial-C_pulm_art_initial*heart_failure_compliance_pulm_art_scale)/exp(sim_time/(24*5))+C_pulm_art_initial*heart_failure_compliance_pulm_art_scale;

cf            = (cf_initial-cf_initial*cf_scale)/exp(sim_time/(24))+cf_initial*cf_scale;
contractility = (contractility_initial-contractility_initial*contractility_HF_scale)/exp(sim_time/(24))+contractility_initial*contractility_HF_scale;

Sodium_protein_filtration_rate_Kf = (Sodium_protein_filtration_rate_Kf_nom-Sodium_protein_filtration_rate_Kf_nom*kf_scale)/exp(sim_time/(24))+Sodium_protein_filtration_rate_Kf_nom*kf_scale;
vascular_responsiveness_scale =  (vascular_responsiveness_scale_nom-vascular_responsiveness_scale_nom*vascular_responsiveness_scale_nom_scale)/exp(sim_time/(24))+vascular_responsiveness_scale_nom*vascular_responsiveness_scale_nom_scale;
}
else {  #If no heart failure, parameters just equation baseline values
R_art0		  = R_art0_initial	;
R_per0		  = R_per0_initial	;
R_al0 = R_al0_scale*R_per0;    #0.66*R_per0
R_cap0= R_cap0_scale*R_per0;    #0.25*R_per0
R_vn0 = R_vn0_scale*R_per0;   #0.09*R_per0
R_ven0		  = R_ven0_initial	;
R_art_pulm	= R_art_pulm_initial;
R_ven_pulm	= R_ven_pulm_initial;
C_art		    = C_art_initial 	;
C_ven0	  	= C_ven0_initial	;
C_pulm_ven  = C_pulm_ven_initial;
C_pulm_art  = C_pulm_art_initial;

cf = cf_initial;
contractility = contractility_initial;
Sodium_protein_filtration_rate_Kf = Sodium_protein_filtration_rate_Kf_nom;
vascular_responsiveness_scale = vascular_responsiveness_scale_nom;
}
CO_nom = CO_nom_initial;

plasma_protein_concentration =plasma_protein_amount / (blood_volume_L * L_dL);
ISF_protein_concentration = ISF_protein_amount / (interstitial_fluid_volume * L_dL);


if(test_ISF_pressure==1){
ISF_pressure =  -0.000000002316554*interstitial_fluid_volume^6
+ 0.000000910096686*interstitial_fluid_volume^5 - 0.000141950241878*interstitial_fluid_volume^4
+ 0.011247538513374*interstitial_fluid_volume^3 - 0.471694546620885*interstitial_fluid_volume^2
+ 9.92251133652023*interstitial_fluid_volume - 81.2426416630404+5.09732;

}else{
ISF_pressure=ISF_pressure_initial;
}

########### feedback of the cardiac output perturbation#################

#If not testing cardiac reserve:
if (Heart_failure_link == 0 & cardiac_reserve == 0  ) {
tissue_autoregulation_signal = max(0.1,1+tissue_autoreg_scale*(Kp_CO*(CO_delayed - CO_nom*CO_species_scale)+Ki_CO*CO_error));
peripheral_resistance_multiplier = disease_effect_on_TPR_peripheral_resistance * B2sna_effect_on_TPR*A1sna_effect_on_TPR*tissue_autoregulation_signal;
peripheral_resistance_multiplier_adjusted = max(dilation_scale,time_TPR_scale*(1+vascular_responsiveness_scale*(peripheral_resistance_multiplier-1)));#*max(0,(-peripheral_resistance_multiplier_adjusted+dilation_scale)/dilation_scale);
heart_rate = HR_heart_rate * BB_HR_effect ;

} else if (Heart_failure_link == 0 & cardiac_reserve == 1){ #If testing cardiac reserve

tissue_autoregulation_signal = max(0.1,1+tissue_autoreg_scale*(Kp_CO*(CO_delayed - CO_nom*CO_species_scale)+Ki_CO*CO_error));
HR_autoregulation_signal = 1+tissue_autoreg_scale*(HRP_CO*(CO_delayed - CO_nom*CO_species_scale)+HRI_CO*CO_error);

heart_rate_multiplier_adjusted = time_HR_scale*(1-HR_autoregulation_signal);
Heart_rate_increasing_ratio =  min(Mag_HR_changing_ratio,(Heart_rate_increasing_ratio + heart_rate_multiplier_adjusted));#*max(0,(-heart_rate_multiplier_adjusted+Mag_HR_changing_ratio)/Mag_HR_changing_ratio);
heart_rate = (HR_heart_rate * BB_HR_effect*(1+Heart_rate_increasing_ratio));

peripheral_resistance_multiplier = disease_effect_on_TPR_peripheral_resistance * B2sna_effect_on_TPR*A1sna_effect_on_TPR*tissue_autoregulation_signal;
peripheral_resistance_multiplier_adjusted = max(dilation_scale,time_TPR_scale*(1+vascular_responsiveness_scale*(peripheral_resistance_multiplier-1)));#*max(0,(-peripheral_resistance_multiplier_adjusted+dilation_scale)/dilation_scale);

} else if (Heart_failure_link == 1 & cardiac_reserve == 0){

tissue_autoregulation_signal = max(0.1,1+tissue_autoreg_scale*(Kp_HF_CO*(CO_delayed - CO_nom*CO_species_scale)+Ki_HF_CO*CO_error));
HR_autoregulation_signal = 1+tissue_autoreg_scale*(HRP_HF_CO*(CO_delayed - CO_nom*CO_species_scale)+HRI_HF_CO*CO_error);

peripheral_resistance_multiplier = disease_effect_on_TPR_peripheral_resistance * B2sna_effect_on_TPR*A1sna_effect_on_TPR*tissue_autoregulation_signal;
peripheral_resistance_multiplier_adjusted =max(dilation_HF_Chronic_scale,(1+vascular_responsiveness_scale*(peripheral_resistance_multiplier-1)));#*max(0,(-peripheral_resistance_multiplier_adjusted+dilation_HF_scale)/dilation_HF_scale);

heart_rate_multiplier_adjusted = (1-HR_autoregulation_signal);
Heart_rate_increasing_ratio = min(Mag_HR_HF_Chronic_ratio,(Heart_rate_increasing_ratio + heart_rate_multiplier_adjusted));#*max(0,(-heart_rate_multiplier_adjusted+Mag_HR_HF_changing_ratio)/Mag_HR_HF_changing_ratio);
heart_rate = (HR_heart_rate * BB_HR_effect*(1+Heart_rate_increasing_ratio));
} else if (Heart_failure_link == 1& cardiac_reserve == 1){

tissue_autoregulation_signal = max(0.1,1+tissue_autoreg_scale*(Kp_HF_CO*(CO_delayed - CO_nom*CO_species_scale)+Ki_HF_CO*CO_error));
HR_autoregulation_signal = 1+tissue_autoreg_scale*(HRP_HF_CO*(CO_delayed - CO_nom*CO_species_scale)+HRI_HF_CO*CO_error);

peripheral_resistance_multiplier = disease_effect_on_TPR_peripheral_resistance * B2sna_effect_on_TPR*A1sna_effect_on_TPR*tissue_autoregulation_signal;
peripheral_resistance_multiplier_adjusted =max(dilation_HF_Acute_scale,(1+vascular_responsiveness_scale*(peripheral_resistance_multiplier-1)));#*max(0,(-peripheral_resistance_multiplier_adjusted+dilation_HF_scale)/dilation_HF_scale);

heart_rate_multiplier_adjusted = (1-HR_autoregulation_signal);
Heart_rate_increasing_ratio = min(Mag_HR_HF_Acute_ratio,(Heart_rate_increasing_ratio + heart_rate_multiplier_adjusted));#*max(0,(-heart_rate_multiplier_adjusted+Mag_HR_HF_changing_ratio)/Mag_HR_HF_changing_ratio);
heart_rate = (HR_heart_rate * BB_HR_effect*(1+Heart_rate_increasing_ratio));
}
########### feedback of the cardiac output perturbation #################


peripheral_resistance =TPR_scale_peripheral_resistance * R_al0*peripheral_resistance_multiplier_adjusted;

arterial_dis_resistance=peripheral_resistance;
capillary_resistance=R_cap0;
venules_resistance=R_vn0;


beat_duration = min_sec / heart_rate ;
beat_time = sim_time/beat_duration - floor(sim_time/beat_duration);
periods = floor(sim_time/beat_duration);

blood_volume = blood_volume_L/1000;
time_step = 0.01;

#Neurohormonal effects - right now they are all turned off   SNA''sympathetic nerve activity''
sna_effect_on_contractility=1;
sna_effect_on_HR = 1;
AngII_effect_on_venous_compliance=1;


# ANP_effect_on_venous_compliance=1;
SNA_effect_on_venous_compliance=1;
B2sna_effect_on_TPR = 1;
A1sna_effect_on_TPR =  1;
Ang_II_effect_on_systemic_resistance =1;
aldo_effect_on_systemic_resistances = 1;
CCB_effect_on_systemic_arterial_resistance = 1;


# ANP_effect_on_peripheral_resistance = 1;
glu_eff_1 = 0;
angII_eff_1 = 0;
aldo_eff_1 = 0;

#Beta-blockers are drugs that bind to beta-adrenoceptors
#and thereby block the binding of norepinephrine and epinephrine to these receptors.
beta_blocker_effect_on_contractility = BB_contractility_effect; # BB_contractility_effect = 1
BP_effect_on_compliance=1;




##### Volume unit conversions #####
LV_volume_mL = LV_volume * m3_mL;
arterial_volume_mL = arterial_volume * m3_mL;
arterial_dis_circulation_volume_mL 	=arterial_dis_circulation_volume *m3_mL;
capillary_circulation_volume_mL	 	=capillary_circulation_volume	 *m3_mL;
venules_circulation_volume_mL		=venules_circulation_volume		 *m3_mL;

############

RV_volume_mL = RV_volume * m3_mL;
pulmonary_arterial_volume_mL = pulmonary_arterial_volume * m3_mL;
venous_volume_mL = venous_volume * m3_mL;
#total_blood_volume_mL = LV_volume_mL + arterial_volume_mL + peripheral_volume_mL + RV_volume_mL + pulmonary_arterial_volume_mL + pulmonary_venous_volume * m3_mL + venous_volume_mL;
total_blood_volume_mL = LV_volume_mL + arterial_volume_mL + arterial_dis_circulation_volume_mL+capillary_circulation_volume_mL+venules_circulation_volume_mL + RV_volume_mL + pulmonary_arterial_volume_mL + pulmonary_venous_volume * m3_mL + venous_volume_mL;



########## Cardiac tissue composition ################
# V_w_0 = 0.00012
# Baseline_Interstitial_Fibrosis = 0.000003
#Baseline_Replacement_Fibrosis = 0.000003
#Baseline_Interstitial_Tissue = 0.000032
#Baseline_Segmental_Fibrosis = 0

baseline_total_myocyte_volume = V_w_0 - Baseline_Interstitial_Fibrosis - Baseline_Replacement_Fibrosis - Baseline_Interstitial_Tissue;  ## baseline myocyte volume determined by V_w_0
baseline_single_myocyte_volume = baseline_total_myocyte_volume/Baseline_Myocyte_Number;
baseline_myocyte_diameter = 2*sqrt(baseline_single_myocyte_volume/(Pi*Baseline_Myocyte_Length));	## baseline myocyte diameter determined by V_w_0 & Baseline_Myocyte_Length, NOT by Baseline_Myocyte_Diameter

myocyte_length = Baseline_Myocyte_Length + change_in_myocyte_length;						## change_in_myocyte_length depends on passive stress levels
myocyte_diameter = baseline_myocyte_diameter + change_in_myocyte_diameter;					## change_in_myocyte_diameter depends on active stress levels - HYPERTROPHY
single_myocyte_volume = myocyte_length * Pi * (myocyte_diameter ^ 2) / 4;					## myocyte volume calculated as a cylinder

#Baseline_Myocyte_Number = 3.3e+9
number_of_live_myocytes = Baseline_Myocyte_Number;
total_myocyte_volume = single_myocyte_volume * number_of_live_myocytes;
total_nonmyocyte_volume = Baseline_Interstitial_Fibrosis + Baseline_Interstitial_Tissue + Baseline_Replacement_Fibrosis;
LV_wall_volume = total_myocyte_volume + total_nonmyocyte_volume;


level_of_hypertrophy = LV_wall_volume / (baseline_total_myocyte_volume + total_nonmyocyte_volume);		## a measure of how much LV wall wolume has grown
outward_growth = LV_cavity_volume / LV_V0_baseline;										## a measure of how much the LV chamber has grown in volume

pct_change_in_myocyte_diameter = change_in_myocyte_diameter / baseline_myocyte_diameter * 100;
pct_change_in_myocyte_length = change_in_myocyte_length / Baseline_Myocyte_Length * 100;


#################################################################################################
#################################################################################################

######### Cardiac Mechanics #########
## Muscle fiber stress and strain are approximately homogeneously distributed, so that they may be approximated by single values.
## Microscopic constitutive laws for fiber stress and radial stress are used to model active and passive fiber stress.

LV_cavity_volume = LV_V0_baseline * (1 + myo_L_scale * change_in_myocyte_length / Baseline_Myocyte_Length) ^ 3 * (1 - myo_D_scale * change_in_myocyte_diameter / baseline_myocyte_diameter) ^ 2;
LV_fiber_stretch =((LV_volume + (LV_wall_volume/3)) / (LV_cavity_volume + (LV_wall_volume / 3)))^0.3333333;

LV_sarcomere_length = ls_0_passive_LV_sarcomere_length * LV_fiber_stretch;
LV_sarcomere_contraction_velocity = (LV_sarcomere_length - LV_sarcomere_length_delayed) / time_step;
contraction_velocity_effect_in_LV = (1 - LV_sarcomere_contraction_velocity / v0_LV_contraction_velocity_effect_in_LV) / (1 + Cv_contraction_velocity_effect_in_LV * LV_sarcomere_contraction_velocity / v0_LV_contraction_velocity_effect_in_LV);

if (LV_sarcomere_length > ls_a0) {
sarcomere_length_effect_in_LV = ((LV_sarcomere_length - ls_a0) / (ls_ar_sarcomere_length_effect_in_LV - ls_a0));
} else {
sarcomere_length_effect_in_LV = 0;
}

chamber_radius = ((LV_cavity_volume * 3 / 4 / Pi) ^ 0.3333333) * m_mm ;			## approximating the LV as a spherical shell
chamber_diameter = 2 * chamber_radius;

outer_radius = (((LV_cavity_volume + LV_wall_volume) * 3 / 4 / Pi) ^ 0.3333333) * m_mm ;
h_wall = outer_radius - chamber_radius;			## wall thickness

h_over_r = h_wall / chamber_radius;				## a measure of the LV chamber growth; it's the ratio of wall thickness to chamber radius


EDV_chamber_radius = ((LV_EDV* 3 / 4 / Pi) ^ 0.33333333) * m_mm;
EDV_chamber_diameter = 2*EDV_chamber_radius;

EDV_outer_radius = (((LV_EDV + LV_wall_volume) * 3 / 4 / Pi) ^ 0.3333333) * m_mm ;
EDV_h_wall = EDV_outer_radius - EDV_chamber_radius;			## wall thickness
EDV_h_over_r = EDV_h_wall / EDV_chamber_radius;

LV_mass = 1000000*LV_wall_volume*1.05 ;			##wall volume*[(cm->m)^3]*density
## density= 1.05 g/cm^3
#LV_mass = ( 0.8*(1.04*((chamber_diameter + h_wall + h_wall)^3 - chamber_diameter^3)) + 0.6 ) / 1000;	# LV mass in g

#### Cardiac excitation ####

RV_twitch_duration = RV_systolic_time_fraction * beat_duration;

t_d = tau_d_LV_twitch_shape;
t_r = tau_r_LV_twitch_shape;
t_twitch = t_r + t_d;

if (beat_time <= t_r) {
sin_signal=(sin(Pi*beat_time/t_twitch))^n_r_LV_twitch_shape;
}else {
sin_signal=(sin(Pi*beat_time/t_twitch))^n_r_LV_twitch_shape;
}

LV_twitch_shape = sin_signal;
if (beat_time < 0) {
LV_twitch_shape = 0;
}
if (beat_time > t_twitch) {
LV_twitch_shape=0;
}

RV_twitch_shape = (sin(Pi * beat_time / RV_twitch_duration)) ^ 2;
if (beat_time < 0) {
RV_twitch_shape=0;
}
if (beat_time > RV_twitch_duration) {
RV_twitch_shape = 0;
}

LV_active_stress = contractility_scale_LV_active_stress * contractility * sigma_ar * sarcomere_length_effect_in_LV * LV_twitch_shape * contraction_velocity_effect_in_LV * beta_blocker_effect_on_contractility *sna_effect_on_contractility;

#Hypertrophy increases ventricular stiffness
hypertrophy_effect_on_Cf = hypertrophy_Cf_slope*(level_of_hypertrophy - 1);
C_f = cf*(1+hypertrophy_effect_on_Cf);


stretch_zero_S = stretch_min_LV_passive_stress_along_fiber - stretch_scale_LV_passive_stress_along_fiber;
if (LV_fiber_stretch >= stretch_zero_S) {
LV_passive_stress_along_fiber = s_f0 * (exp(C_f * (LV_fiber_stretch - stretch_zero_S)) - 1);
} else {
LV_passive_stress_along_fiber = 0;
}


LV_radial_stretch = 1/ (LV_fiber_stretch * LV_fiber_stretch);

if (LV_radial_stretch >= 1) {
LV_passive_radial_stress = s_r0 * (exp(c_r_LV * (LV_radial_stretch - 1)) - 1);
} else {
LV_passive_radial_stress = 0;
}

LV_total_stress = LV_active_stress + LV_passive_stress_along_fiber - 2 * LV_passive_radial_stress;


if (LV_volume > LV_V0_min) {
rel_volume_LV = 1+LV_wall_volume/LV_volume;
} else {
rel_volume_LV = 1+LV_wall_volume/LV_V0_min;
}

LV_pressure = LV_total_stress * log(rel_volume_LV)/ 3;



#################################################################################################
#################################################################################################

#peripheral_pressure = P_ven0 + (peripheral_circulation_volume - V_per0) / C_per;
arterial_dis_pressure = P_al0 + (arterial_dis_circulation_volume-V_al0)/C_al;
capillary_pressure = P_cap0 + (capillary_circulation_volume-V_cap0)/C_cap;
venules_pressure = P_vn0 + (venules_circulation_volume-V_vn0)/C_vn;

# venous_compliance=C_ven0*AngII_effect_on_venous_compliance*ANP_effect_on_venous_compliance*SNA_effect_on_venous_compliance;
venous_compliance=C_ven0*AngII_effect_on_venous_compliance*SNA_effect_on_venous_compliance;
venous_pressure = P_ven0 + (venous_volume - V_ven0) / venous_compliance; ################################################################################# was P_ven0

RV_Cavity_Volume = RV_V0;
RV_wall_volume = V_w_0_RV;
RV_fiber_stretch = ((RV_volume + V_w_0_RV/3) / (RV_Cavity_Volume + RV_wall_volume/3))^(0.333);
RV_sarcomere_length = ls_a0_RV * RV_fiber_stretch;

if (RV_sarcomere_length > ls_a0_RV) {
sarcomere_length_effect_in_RV = (RV_sarcomere_length - ls_a0_RV) / (0.000002 - ls_a0_RV);
} else {
sarcomere_length_effect_in_RV = 0;
}


RV_sarcomere_contraction_velocity = (RV_sarcomere_length - RV_sarcomere_length_delayed) / time_step;
contraction_velocity_effect_in_RV = (1 - RV_sarcomere_contraction_velocity / v0_RV_contraction_velocity_effect_in_RV) / (1 + 0 * RV_sarcomere_contraction_velocity / v0_RV_contraction_velocity_effect_in_RV);
RV_active_stress_multiplier = contractility_RV*sna_effect_on_contractility;
RV_active_stress = contractility_RV * sigma_ar_RV * sarcomere_length_effect_in_RV * RV_twitch_shape * contraction_velocity_effect_in_RV*sna_effect_on_contractility;

RV_radial_stretch = 1/ (RV_fiber_stretch * RV_fiber_stretch);

if (RV_radial_stretch >= 1) {
RV_passive_radial_stress = s_r0_RV * (exp(c_r_RV * (RV_radial_stretch - 1)) - 1);
} else {
RV_passive_radial_stress = 0;
}


if (RV_fiber_stretch >= 1) {
RV_passive_stress_along_fiber = s_f0_RV * (exp(cf_RV * (RV_fiber_stretch - 1)) - 1);
} else {
RV_passive_stress_along_fiber = 0;
}

RV_total_stress = RV_active_stress + RV_passive_stress_along_fiber - 2 * RV_passive_radial_stress;

if (RV_volume > RV_V0_min) {
rel_volume = (1 + RV_wall_volume / RV_volume);
} else {
rel_volume = (1 + RV_wall_volume / RV_V0_min);
}

RV_pressure = RV_total_stress * log(rel_volume) / 3;

######### Circulatory hemodynamics #########
venous_flow = (venules_pressure - venous_pressure) / (R_ven0);

##allow blood volume link between heart and kidney to be turned on/off
if (heart_renal_link == 1) {
venous_volume_target =  blood_volume - LV_volume - arterial_volume - arterial_dis_circulation_volume-capillary_circulation_volume-venules_circulation_volume - RV_volume - pulmonary_arterial_volume - pulmonary_venous_volume;
} else {
venous_volume_target = venous_volume;
}

tricuspid_valve_flow_rate = max((venous_pressure - RV_pressure) / R_r_atrium,min_flux);

pulmonary_arterial_pressure = ( pulmonary_arterial_volume - V_pulm_art0 )/C_pulm_art + P_art0;
pulmonary_venous_pressure = P_ven0 + (pulmonary_venous_volume  - V_pulm_ven0 )/C_pulm_ven;
pulmonary_arterial_blood_flow = (pulmonary_arterial_pressure  - pulmonary_venous_pressure )/ R_ven_pulm ;

dP = RV_pressure - pulmonary_arterial_pressure;
Zn = L_pulm + time_step * R_art_pulm;
pulmonary_blood_flow = (pulmonary_blood_flow_delayed * L_pulm + dP * time_step) / Zn;

pulmonary_valve_flow_rate = max(pulmonary_blood_flow,min_flux);


#Allow regurgitation if pressure differential across mitral valve exceeds threshold (set threshold really high to eliminate any regurgitation)
if (pulmonary_venous_pressure > LV_pressure) {
mitral_valve_flow_rate = max  (  (pulmonary_venous_pressure - LV_pressure)/R_left_atrium  , min_flux ) ;
} else {
if (LV_pressure - pulmonary_venous_pressure < mitral_regurgitation_pressure_diff) {
mitral_valve_flow_rate = min_flux;
} else {
#Allow negative flow back out of the ventricle
mitral_valve_flow_rate = (pulmonary_venous_pressure - LV_pressure)/R_left_atrium;
}

}

pulmonary_valve_flow_rate = max(pulmonary_blood_flow,min_flux);


## arterial_pressure computed without taking account compliance
## arterial_compliance = compliance_scale_arterial_compliance * (C_art ) * BP_effect_on_compliance;
## arterial_pressure = (arterial_volume - V_art0) / arterial_compliance + P_art0 ;

## arterial_pressure computed taking account compliance
## Constants taken from Safar, M. E., et al. Stiffness of carotid artery wall material and blood pressure in humans: application to antihypertensive therapy and stroke prevention. Stroke 31.3 (2000): 782-790.
## We assume a linear effect as follows.
## BP_effect_on_stiffness = (arterial_pressure - 85)*Stiffness_BP_slope;
## Solving the following system

Stiffness0=1/C_art;
arterial_stiffness = Stiffness0*(1+ (MAP_delayed - nominal_map_setpoint)*Stiffness_BP_slope);
arterial_compliance = 1/arterial_stiffness;
arterial_pressure = (arterial_volume - V_art0) / arterial_compliance + P_art0 ;


#peripheral_pressure = P_ven0 + (peripheral_circulation_volume - V_per0) / C_per;
systemic_blood_flow = (arterial_pressure - arterial_dis_pressure) / arterial_dis_resistance;
arterial_dis_blood_flow = (arterial_dis_pressure-capillary_pressure)/capillary_resistance;
capillary_blood_flow = (capillary_pressure-venules_pressure)/venules_resistance;
venules_blood_flow = (venules_pressure-venous_pressure)/R_ven0;
dP_1 = LV_pressure - arterial_pressure;
Zn_1 = L_art + R_art0 * time_step;
aortic_blood_flow = (aortic_blood_flow_delayed * L_art*L_scale + dP_1 * time_step) / Zn_1;

aortic_valve_flow_rate = max(aortic_blood_flow,min_flux);



#############################################
### Relating BNP and NTP to End Diastolic Stress###

BNP = exp(BNP_factor*((LV_EDS+1736)/5.094)+3.14);
NTP = exp((log(BNP)+1.4919)/1.0694);

###Assume ANP proportional to BNP
# ANP = BNP;

## Relate ANP with other variables following Karaaslan et al. Long-term mathematical model involving renal sympathetic nerve activity, arterial pressure, and sodium excretion. Annals of biomedical engineering 33.11 (2005): 1607-1630.
if (heart_renal_link  == 1) {
# Right atrial pressure (Pra) - Frank-Starling law
Pra = 0.2787*exp(CO_delayed*0.2281); # [cardiac output]=l/min, [Pra]=mmHg
# Normalized atrial natriuretic peptide concentration
Canp = max(0,7.427 - 6.554 / ( 1 + exp( Pra - 3.762) ) + deltaCanp); # For human beings the actual value at normal conditions is about 36 ng/l
lambdaANP = -0.1*Canp+1.1199;
} else {
lambdaANP = 1;
}


#############################################
### Find Diastolic and Systolic values ###

if (beat_time >= (1 - .01/beat_duration) && beat_time < 1) {
LV_pressure_diastolic_max = LV_pressure;
LV_stress_diastolic_max = LV_passive_stress_along_fiber;
LV_volume_maximum = LV_volume;

} else {
LV_pressure_diastolic_max = LV_EDP;
LV_stress_diastolic_max = LV_EDS;
LV_volume_maximum = LV_EDV;
}

LV_EDP_old = LV_pressure_diastolic_max;
LV_EDS_old = LV_stress_diastolic_max;
LV_EDV_old = LV_volume_maximum;



## Method to find Systolic and Diastolic blood pressure
## use current and last two time steps to find local maxima and minima
if (arterial_pressure_delayed < arterial_pressure_bigger_delay) {
systemic_pressure_minimum_1 = arterial_pressure_delayed;
} else {
systemic_pressure_minimum_1 = diastolic_pressure;
}
if (arterial_pressure_delayed < arterial_pressure) {
systemic_pressure_minimum = systemic_pressure_minimum_1;
} else {
systemic_pressure_minimum = diastolic_pressure;
}

if (arterial_pressure_delayed > arterial_pressure_bigger_delay) {
systemic_pressure_maximum_1 = arterial_pressure_delayed;
} else {
systemic_pressure_maximum_1 = systolic_pressure;
}
if (arterial_pressure_delayed > arterial_pressure) {
systemic_pressure_maximum= systemic_pressure_maximum_1;
} else {
systemic_pressure_maximum= systolic_pressure;
}

systolic_pressure_old = systemic_pressure_maximum;
diastolic_pressure_old = systemic_pressure_minimum;



#############################################################
### Find Peak systolic stress in the LV ###

if (beat_time >= (t_r*0.8) && beat_time < (t_r*0.85)) {
LV_peak_stress = LV_active_stress;
} else {
LV_peak_stress = LV_active_stress_peak;
}

## fixes a problem where it was finding peaks that were very close to zero
if (LV_active_stress > 1) {
LV_active_stress_peak_old = LV_peak_stress;
} else {
LV_active_stress_peak_old = LV_active_stress_peak;
}


################################Find Capillary and Venous Pressures

###Method to compute the systolic and diastolic venous_pressure
if (venous_pressure_delayed < venous_pressure_bigger_delay) {
venous_pressure_minimum_1 = venous_pressure_delayed;
} else {
venous_pressure_minimum_1 = venous_diastolic_pressure;
}
if (venous_pressure_delayed < venous_pressure) {
venous_pressure_minimum = venous_pressure_minimum_1;
} else {
venous_pressure_minimum = venous_diastolic_pressure;
}

if (venous_pressure_delayed > venous_pressure_bigger_delay) {
venous_pressure_maximum_1 = venous_pressure_delayed;
} else {
venous_pressure_maximum_1 = venous_systolic_pressure;
}
if (venous_pressure_delayed > venous_pressure) {
venous_pressure_maximum= venous_pressure_maximum_1;
} else {
venous_pressure_maximum= venous_systolic_pressure;
}

venous_systolic_pressure_old = venous_pressure_maximum;
venous_diastolic_pressure_old = venous_pressure_minimum;

###Method to compute the systolic and diastolic peripheral_pressure

if (capillary_pressure_delayed < capillary_pressure_bigger_delay) {
capillary_pressure_minimum_1 = capillary_pressure_delayed;
} else {
capillary_pressure_minimum_1 = capillary_diastolic_pressure;
}
if (capillary_pressure_delayed < capillary_pressure) {
capillary_pressure_minimum = capillary_pressure_minimum_1;
} else {
capillary_pressure_minimum = capillary_diastolic_pressure;
}

if (capillary_pressure_delayed > capillary_pressure_bigger_delay) {
capillary_pressure_maximum_1 = capillary_pressure_delayed;
} else {
capillary_pressure_maximum_1 = capillary_systolic_pressure;
}
if (capillary_pressure_delayed > capillary_pressure) {
capillary_pressure_maximum= capillary_pressure_maximum_1;
} else {
capillary_pressure_maximum= capillary_systolic_pressure;
}

capillary_systolic_pressure_old = capillary_pressure_maximum;
capillary_diastolic_pressure_old = capillary_pressure_minimum;



#Calculate mean pressures

mean_capillary_pressure = (capillary_systolic_pressure/3+capillary_diastolic_pressure*2/3)*Pa_mmHg;



if (heart_renal_link  == 1) {
mean_arterial_pressure_MAP = (systolic_pressure/3+diastolic_pressure*2/3)*Pa_mmHg;
mean_venous_pressure = (venous_systolic_pressure+venous_diastolic_pressure*2)/3*Pa_mmHg;
} else {
mean_arterial_pressure_MAP = nominal_map_setpoint;
mean_venous_pressure = P_venous;
}




#################################################################################################
#################################################################################################

######################### Hypertrophy ########################################
## growth occurs when the peak of the active stress exceeds the threshhold
## use the same method as above to locate the peaks of the active stress

######################### Hypertrophy ########################################

if (LV_active_stress_peak > LV_active_stress_threshhold) {		## myocyte diameter grows if peak active stress exceeds the threshhold
kD_hypertrophy =  (kD_HYPERTROPHY*C_renal_CV_timescale) * max(0, ((max_myocyte_diameter_increase) - change_in_myocyte_diameter)/(max_myocyte_diameter_increase));	## progression of hypertophic growth - myocyte diameter increase
} else {																							## there is a limit on how much the diameter of the myocyte can increase (right now: 40%)
kD_hypertrophy = (kD_HYPERTROPHY*C_renal_CV_timescale);		## regression of hypertrophic growth (negative will come from ODE below)
}

if (LV_EDS > LV_passive_stress_along_fiber_threshhold) {
kL_hypertrophy =  (kL_HYPERTROPHY*C_renal_CV_timescale) * max(0, ((max_myocyte_length_increase) - change_in_myocyte_length)/(max_myocyte_length_increase));		## progression of growth - myocyte length increase
} else {              ##adding a length threshhold to limit volumetric remodeling
kL_hypertrophy = 0;		## myocyte length does not decrease once stretched
}

#################################################################################################
#################################################################################################

LVID=((6*LV_EDV)/3.14159)^(1/3);

#############################################################################


############################ Kidney Model ###############################################

#Disease effects on nephrons

#For now we limit glomeruli loss to 30% of Kf damage - somewhat arbitrary
number_of_functional_glomeruli = baseline_nephrons*(1 - 0.3*disease_effects_decreasing_Kf);

#Assume that other nephrons are lost due to tubular effects, which do not affect the glomerulus
number_of_functional_tubules = baseline_nephrons*(1-disease_effect_on_nephrons);


###AT1-bound AngII constricts the preafferent, afferent, and efferent arterioles
AT1_preaff_int = 1 - AT1_preaff_scale/2;
AT1_effect_on_preaff = AT1_preaff_int + AT1_preaff_scale/(1+exp(-(AT1_bound_AngII - nominal_equilibrium_AT1_bound_AngII)/AT1_preaff_slope));

AT1_aff_int = 1 - AT1_aff_scale/2;
AT1_effect_on_aff = AT1_aff_int + AT1_aff_scale/(1+exp(-(AT1_bound_AngII - nominal_equilibrium_AT1_bound_AngII)/AT1_aff_slope));

AT1_eff_int = 1 - AT1_eff_scale/2;
AT1_effect_on_eff = AT1_eff_int + AT1_eff_scale/(1+exp(-(AT1_bound_AngII - nominal_equilibrium_AT1_bound_AngII)/AT1_eff_slope));


# ANP_aff_int = 1 + ANP_aff_scale/2;
# ANP_effect_on_aff = ANP_aff_int - ANP_aff_scale/(1+exp(-(ANP - nom_ANP)/ANP_aff_slope));

### RSNA constricts the preafferent vasculature
rsna_preaff_int = 1 - rsna_preaff_scale/2;
rsna_effect_on_preaff = rsna_preaff_int + rsna_preaff_scale/(1+exp(-(renal_sympathetic_nerve_activity - nom_rsna)/rsna_preaff_slope));

### ANP may dilate the preafferent, afferent, and/or efferent arterioles
# ANP_preaff_int = 1 + ANP_preaff_scale/2;
# ANP_effect_on_preaff = ANP_preaff_int - ANP_preaff_scale/(1+exp(-(normalized_atrial_NP_concentration - nom_ANP)/ANP_preaff_slope));

# ANP_aff_int = 1 + ANP_aff_scale/2;
# ANP_effect_on_aff = ANP_aff_int - ANP_aff_scale/(1+exp(-(normalized_atrial_NP_concentration - nom_ANP)/ANP_aff_slope));

# ANP_eff_int = 1 + ANP_eff_scale/2;
# ANP_effect_on_eff= ANP_eff_int - ANP_eff_scale/(1+exp(-(normalized_atrial_NP_concentration - nom_ANP)/ANP_eff_slope));


######Preafferent Resistance


if(renal_blood_flow_L_min_delayed < 1 & interstitial_fluid_volume < 15){
IF_Venous_RIHP_Effect_int = 1- IF_Venous_RIHP_Effect_scale/2;
IF_Venous_RIHP_Effect =(IF_Venous_RIHP_Effect_int+IF_Venous_RIHP_Effect_scale/(1+exp(-(renal_blood_flow_L_min_delayed-nom_renal_blood_flow_L_min)/IF_Venous_RIHP_Effect_slope)));
}else{
IF_Venous_RIHP_Effect=1;
}


preaff_arteriole_signal_multiplier = AT1_effect_on_preaff*rsna_effect_on_preaff*preafferent_pressure_autoreg_signal*CCB_effect_on_preafferent_resistance;
preaff_arteriole_adjusted_signal_multiplier = (1/(1+exp(preaff_signal_nonlin_scale*(1-preaff_arteriole_signal_multiplier)))+0.5);
preafferent_arteriole_resistance = IF_Venous_RIHP_Effect*nom_preafferent_arteriole_resistance*preaff_arteriole_adjusted_signal_multiplier;

###### Afferent Arteriole Resistance
#The afferent arteriole responses the tubuloglomerular feedback (calculated later), as well as to AT1-bound AngII and ANP.
#It may respond myogenically as well. Some studies suggest the upstream portion responds myogenically while the distal portion responds to TGF. Thus, one could consider the
#myogenically responsive portion as part of the preafferent resistance.
#The dilation/constriction of the arterioles is limited, and thus the total combined effect of all regulators must saturate

# afferent_arteriole_signal_multiplier = tubulo_glomerular_feedback_effect * AT1_effect_on_aff * ANP_effect_on_aff*glomerular_pressure_autoreg_signal*CCB_effect_on_afferent_resistance;
afferent_arteriole_signal_multiplier = tubulo_glomerular_feedback_effect * AT1_effect_on_aff *glomerular_pressure_autoreg_signal*CCB_effect_on_afferent_resistance;
afferent_arteriole_adjusted_signal_multiplier = (1/(1+exp(afferent_signal_nonlin_scale*(1-afferent_arteriole_signal_multiplier)))+0.5);
afferent_arteriole_resistance = IF_Venous_RIHP_Effect*nom_afferent_arteriole_resistance*afferent_arteriole_adjusted_signal_multiplier*(1-ANP_effect_on_Arterial_Resistance);

###### Efferent Arteriole Resistance
#The efferent arteriole responses to AT1-bound AngII and ANP.
#The dilation/constriction of the arterioles is limited, and thus the total combined effect of all regulators must saturate
#efferent_arteriole_signal_multiplier = AT1_effect_on_eff * ANP_effect_on_eff *CCB_effect_on_efferent_resistance;

efferent_arteriole_signal_multiplier = AT1_effect_on_eff * CCB_effect_on_efferent_resistance;
efferent_arteriole_adjusted_signal_multiplier = 1/(1+exp(efferent_signal_nonlin_scale*(1-efferent_arteriole_signal_multiplier)))+0.5;
efferent_arteriole_resistance =  IF_Venous_RIHP_Effect*nom_efferent_arteriole_resistance*efferent_arteriole_adjusted_signal_multiplier; #

######Peritubular Resistance
#Autoregulation of peritubular resistance allows RBF to be autoregulated separately from GFR
#This is exploratory for now. By default, this effect is turned off by setting RBF_autoreg_scale to zero
RBF_autoreg_int = 1-RBF_autoreg_scale/2;
peritubular_autoreg_signal = RBF_autoreg_int + RBF_autoreg_scale/(1+exp((nom_renal_blood_flow_L_min - renal_blood_flow_L_min_delayed)/RBF_autoreg_steepness));
autoregulated_peritubular_resistance = peritubular_autoreg_signal*nom_peritubular_resistance;

###### Renal Vascular Resistance
renal_vascular_resistance = (preafferent_arteriole_resistance + (afferent_arteriole_resistance + efferent_arteriole_resistance) / number_of_functional_glomeruli + autoregulated_peritubular_resistance);

###Renal blood flow
renal_blood_flow_L_min = ((mean_arterial_pressure_MAP - P_venous) / renal_vascular_resistance);
renal_blood_flow_ml_hr = renal_blood_flow_L_min * 1000 * 60;


###Renal Vasculature Pressures
preafferent_pressure = mean_arterial_pressure_MAP - renal_blood_flow_L_min*preafferent_arteriole_resistance;
glomerular_pressure = (mean_arterial_pressure_MAP  - renal_blood_flow_L_min * (preafferent_arteriole_resistance + afferent_arteriole_resistance / number_of_functional_glomeruli));
postglomerular_pressure = (mean_arterial_pressure_MAP  - renal_blood_flow_L_min * (preafferent_arteriole_resistance + (afferent_arteriole_resistance+efferent_arteriole_resistance) / number_of_functional_glomeruli));


#Autoregulatory signals for preafferent and afferent resistances
preaff_autoreg_int = 1 - preaff_autoreg_scale/2;
preafferent_pressure_autoreg_function = preaff_autoreg_int+preaff_autoreg_scale/(1+exp((nom_preafferent_pressure - preafferent_pressure)/myogenic_steepness));
gp_autoreg_int = 1 - gp_autoreg_scale/2;
glomerular_pressure_autoreg_function = gp_autoreg_int+gp_autoreg_scale/(1+exp((nom_glomerular_pressure - glomerular_pressure)/myogenic_steepness));

######################## Glomerular Filtration ######################

# Assume glomerulosclerosis causes a decrease in Kf over time, and also a loss of the renal vasculature (afferent and efferent arterioles)
# as glomeruli become completely sclerotic

#Glomerular hypertrophy resulting in increased surface area and thus increased Kf is assumed to occur
#in response to elevated glomerular pressure. A 2 mmHg buffer is built in (i.e. glomerular pressure must be at least 2 mmHg above normal for hypertrophy to begin
#The increase in Kf saturates and cannot exceed the fractional increase set by maximal_glom_surface_area_increase
GP_effect_increasing_Kf = (maximal_glom_surface_area_increase - disease_effects_increasing_Kf) * max(glomerular_pressure/(nom_glomerular_pressure+2) - 1,0) / (T_glomerular_pressure_increases_Kf/C_renal_CV_timescale);
glomerular_hydrostatic_conductance_Kf = nom_Kf*(1+disease_effects_increasing_Kf)*(1-disease_effects_decreasing_Kf);

###Glomerular Fitlration Rate
#Calculation of P_bowmans are described later

net_filtration_pressure = glomerular_pressure - oncotic_pressure_difference - P_bowmans;

if (net_filtration_pressure <= 0) { #Prevent crashing if NFP becomes negative
SNGFR_nL_min = 0.001;
} else {
SNGFR_nL_min = glomerular_hydrostatic_conductance_Kf * (glomerular_pressure - oncotic_pressure_difference - P_bowmans);
}

GFR =  (SNGFR_nL_min / 1000 / 1000000 * number_of_functional_tubules);
GFR_ml_min = GFR * 1000;

filtration_fraction = GFR/renal_blood_flow_L_min;

serum_creatinine_concentration = serum_creatinine/blood_volume_L;
creatinine_clearance_rate = GFR_ml_min * dl_ml * serum_creatinine_concentration; #Units: mg/min

####################### Protein filtering ####################
GPdiff = max(0, glomerular_pressure - (nom_GP_seiving_damage));
GP_effect_on_Seiving = Emax_seiving * GPdiff ^ Gamma_seiving / (GPdiff ^ Gamma_seiving + Km_seiving ^ Gamma_seiving);

#Dean and Lazzara 2006 - Seiving coefficient decreases as GFR increases

nom_glomerular_albumin_sieving_coefficient = seiving_inf/(1-(1-seiving_inf)*exp(-c_albumin*SNGFR_nL_min));
glomerular_albumin_sieving_coefficient = nom_glomerular_albumin_sieving_coefficient*(1 + GP_effect_on_Seiving);

SN_albumin_filtration_rate = plasma_albumin_concentration*SNGFR_nL_min*1e-6*glomerular_albumin_sieving_coefficient; #mg/min

SN_albumin_excretion_rate = max(0, SN_albumin_filtration_rate - SN_albumin_reabsorptive_capacity)+nom_albumin_excretion_rate;
albumin_excretion_rate = SN_albumin_excretion_rate*number_of_functional_tubules;

#albumin_filtation_rate = plasma_albumin_concentration*SNGFR_nL_min*glomerular_albumin_sieving_coefficient;
#albumin_excretion_rate = max(0, albumin_filtation_rate - max_PT_albumin_reabsorption_rate);



###Oncotic pressure difference
#Landis Pappenheimer equation used to calculate oncotic pressure at entrance and exit to glomerulus
#Oncotic pressure is approximated as varying linearly along the glomerulus. Oncotic pressure in the Bowman's space is zero
#Thus the average pressure difference is the average of the entrance and exit oncotic pressure
#We do not consider filtration equilibrium
Oncotic_pressure_in = 1.629*plasma_protein_concentration+0.2935*(plasma_protein_concentration^2);
SNRBF_nl_min = 1e6*1000*renal_blood_flow_L_min/number_of_functional_glomeruli;
plasma_protein_concentration_out = (SNRBF_nl_min*plasma_protein_concentration-SN_albumin_filtration_rate)/(SNRBF_nl_min-SNGFR_nL_min);
Oncotic_pressure_out = 1.629*plasma_protein_concentration_out+0.2935*(plasma_protein_concentration_out^2);
oncotic_pressure_avg = (Oncotic_pressure_in+Oncotic_pressure_out)/2;



##################### Plasma sodium concentration and vasopressin secretion
###Plasma sodium concentration
Na_concentration = sodium_amount / blood_volume_L;
IF_Na_concentration = IF_sodium_amount/interstitial_fluid_volume;

sodium_storate_rate = Q_Na_store*((max_stored_sodium - stored_sodium)/max_stored_sodium)*(IF_Na_concentration - ref_Na_concentration);

###Control of vasopressin secretion
#A proportional-integral controller is used to ensure there is no steady state error in sodium concentration
#Relative gains of the P and I controller must be chosen carefully.
#In order to permit a steady-state error, the integral controller can be removed. But care should be given then in choosing the proportional gain
Na_water_controller = Na_controller_gain*(Kp_VP*(Na_concentration - ref_Na_concentration)+Ki_VP*Na_concentration_error);

###Vasopressin
#Vasopressin is critical in the model, because it allows water excretion to be decoupled from sodium excretion in the collecting duct
normalized_vasopressin_concentration = 1 + Na_water_controller;
vasopressin_concentration = nominal_vasopressin_conc * normalized_vasopressin_concentration;

#Effect of vasopressin on water intake
water_intake_vasopressin_int = 1-water_intake_vasopressin_scale/2;
water_intake = water_intake_species_scale*(nom_water_intake/60/24)*(water_intake_vasopressin_int + water_intake_vasopressin_scale/(1+exp((normalized_vasopressin_concentration_delayed-1)/water_intake_vasopressin_slope)));
daily_water_intake = (water_intake * 24 * 60);


##################### Tubular Flow and Reabsorption######################

#Length of tubular segments
L_pt_s1 = L_pt_s1_nom*(1+tubular_length_increase);
L_pt_s2 = L_pt_s2_nom*(1+tubular_length_increase);
L_pt_s3 = L_pt_s3_nom*(1+tubular_length_increase);
Dc_pt = Dc_pt_nom*(1+tubular_diameter_increase);
L_pt = L_pt_s1+L_pt_s2 + L_pt_s3;

SN_filtered_Na_load = (SNGFR_nL_min / 1000 / 1000000)*Na_concentration;
filtered_Na_load = SN_filtered_Na_load*number_of_functional_tubules;

######Regulatory effects on reabsorption
###Pressure natriuresis effects


pressure_natriuresis_f_int = 1- pressure_natriuresis_f_scale/2;
pressure_natriuresis_p_int = 1- pressure_natriuresis_p_scale/2;

pressure_natriuresis_signal_f = ((pressure_natriuresis_f_int+pressure_natriuresis_f_scale/(1+exp(-(renal_blood_flow_L_min_delayed-nom_renal_blood_flow_L_min)/pressure_natriuresis_f_slope))));  #+0.85 +0.3

pressure_natriuresis_signal_p =(pressure_natriuresis_p_int+pressure_natriuresis_p_scale/(1+exp(-(RIHP_delayed-RIHP0)/pressure_natriuresis_p_slope)));

pressure_natriuresis_IF_int = 1- pressure_natriuresis_IF_scale/2;
pessure_natriuresis_signal_IF = ((pressure_natriuresis_IF_int+pressure_natriuresis_IF_scale/(1+exp(-(Net_oncotic_pressure_diff-Nom_Net_oncotic_pressure_diff)/pressure_natriuresis_IF_slope))));


pressure_natriuresis_signal = pressure_natriuresis_signal_f*pressure_natriuresis_signal_p*pessure_natriuresis_signal_IF;

pressure_natriuresis_PT_int = 1 - pressure_natriuresis_PT_scale/2;
pressure_natriuresis_PT_effect = max(0.001,pressure_natriuresis_PT_int + pressure_natriuresis_PT_scale / (1 + exp(pressure_natriuresis_signal-1)));

pressure_natriuresis_LoH_int = 1 - pressure_natriuresis_LoH_scale/2;
pressure_natriuresis_LoH_effect = max(0.001,pressure_natriuresis_LoH_int + pressure_natriuresis_LoH_scale / (1 + exp((postglomerular_pressure_delayed - RIHP0) / pressure_natriuresis_LoH_slope)));

pressure_natriuresis_DCT_magnitude = max(0,pressure_natriuresis_DCT_scale );
pressure_natriuresis_DCT_int = 1 - pressure_natriuresis_DCT_magnitude/2;
pressure_natriuresis_DCT_effect = max(0.001,pressure_natriuresis_DCT_int + pressure_natriuresis_DCT_magnitude/ (1 + exp((postglomerular_pressure_delayed - RIHP0) / pressure_natriuresis_DCT_slope)));

pressure_natriuresis_CD_magnitude = max(0,pressure_natriuresis_CD_scale *(1+disease_effects_decreasing_CD_PN));
pressure_natriuresis_CD_int = 1 - pressure_natriuresis_CD_magnitude/2;
pressure_natriuresis_CD_effect = max(0.001,pressure_natriuresis_CD_int + pressure_natriuresis_CD_magnitude/ (1 + exp(pressure_natriuresis_signal-1)));

#smax(0.001,pressure_natriuresis_signal);

###AT1-bound AngII effect on PT reabsorption
AT1_PT_int = 1 - AT1_PT_scale/2;
AT1_effect_on_PT = AT1_PT_int + AT1_PT_scale/(1+exp(-(AT1_bound_AngII - nominal_equilibrium_AT1_bound_AngII)/AT1_PT_slope));
### Aldosterone effect on DCT and CD reabsorption



### RSNA effect on PT and CD sodium reabsorption
#RSNA effect on CD is turned off by default
rsna_PT_int = 1 - rsna_PT_scale/2;

#######NEED To either change rsna_delayed consistently throughout or revert back
rsna_effect_on_PT = 1;
rsna_CD_int = 1 - rsna_CD_scale/2;
rsna_effect_on_CD= rsna_CD_int + rsna_CD_scale/(1+exp((1 - renal_sympathetic_nerve_activity)/rsna_CD_slope));


###Aldosterone effect on distal and collecting duct sodium reabsorption
aldosterone_concentration = normalized_aldosterone_level* nominal_aldosterone_concentration;
Aldo_MR_normalised_effect = normalized_aldosterone_level*(1 - pct_target_inhibition_MRA);

aldo_DCT_int = 1 - aldo_DCT_scale/2;
aldo_effect_on_DCT = aldo_DCT_int + aldo_DCT_scale/(1+exp((1 - Aldo_MR_normalised_effect)/aldo_DCT_slope));

aldo_CD_int = 1 - aldo_CD_scale/2;
aldo_effect_on_CD= aldo_CD_int + aldo_CD_scale/(1+exp((1 - Aldo_MR_normalised_effect)/aldo_CD_slope));

###ANP effect on collecting duct sodium reabsorption
# anp_CD_int = 1 - anp_CD_scale/2;
# anp_effect_on_CD= anp_CD_int + anp_CD_scale/(1+exp((1 - normalized_atrial_NP_concentration)/anp_CD_slope));

#Assume insulin has effect on NHE3. Use RUGE as surrogate for insulin effect. When RUGE goes up, insulin effect goes down.
NHE3inhib = Anhe3*RUGE_delayed;
pt_multiplier = AT1_effect_on_PT * rsna_effect_on_PT *pressure_natriuresis_PT_effect*(1-NHE3inhib);

e_pt_sodreab = min(1,nominal_pt_na_reabsorption_nonSGLT * pt_multiplier);# AT1_effect_on_PT * rsna_effect_on_PT *pressure_natriuresis_PT_effect;


e_dct_sodreab = min(1,nominal_dt_na_reabsorption * aldo_effect_on_DCT*pressure_natriuresis_DCT_effect ); #HCTZ_effect_on_DT_Na_reabs

# cd_multiplier = anp_effect_on_CD *aldo_effect_on_CD*rsna_effect_on_CD*pressure_natriuresis_CD_effect;
cd_multiplier = aldo_effect_on_CD*rsna_effect_on_CD*pressure_natriuresis_CD_effect;
#Ensure that CD reabsorption smoothly approaches upper limit set by max_cd_reabs_rate
cd_scale = max_cd_reabs_rate/nominal_cd_na_reabsorption-1;
#if (cd_multiplier > 1) {
#	cd_multiplier = 1+cd_scale - cd_scale*exp((1-cd_multiplier)/.028);
#}
e_cd_sodreab = min(0.9999,nominal_cd_na_reabsorption*cd_multiplier*lambdaANP);

##################################### Proximal Tubule #########################################
###Glucose Filtration and reabsorption in PT
#Assume glucose reabsorption depends only on availability of SGLT1/2
#Assume constant amount of reabsorption per unit length through SGLT2 in convoluted PT
#Assume constant amount of reabsorption per unit length through SGLT1 in straight/recta PT

#Chosen so that UGE becomes non-zero for plasma_glucose concentration ~8.5 mmol/l
glucose_reabs_per_unit_length_s1 = nom_glucose_reabs_per_unit_length_s1*diabetic_adaptation*SGLT2_inhibition_delayed*(1+RTg_compensation);
glucose_reabs_per_unit_length_s2 = nom_glucose_reabs_per_unit_length_s2*diabetic_adaptation*SGLT2_inhibition_delayed*(1+RTg_compensation);
glucose_reabs_per_unit_length_s3 = nom_glucose_reabs_per_unit_length_s3*diabetic_adaptation*(1+RTg_compensation)*SGLT1_inhibition;

SN_filtered_glucose_load = glucose_concentration*SNGFR_nL_min / 1000 / 1000000;  #mmol/min
glucose_pt_out_s1 = max(0,SN_filtered_glucose_load-glucose_reabs_per_unit_length_s1*L_pt_s1); #mmol/min
glucose_pt_out_s2 = max(0,glucose_pt_out_s1-glucose_reabs_per_unit_length_s2*L_pt_s2); #mmol/min
glucose_pt_out_s3 = max(0,glucose_pt_out_s2-glucose_reabs_per_unit_length_s3*L_pt_s3); #mmol/min
RUGE = glucose_pt_out_s3*number_of_functional_tubules*180; #RUGE in mg/min


excess_glucose_increasing_RTg = (maximal_RTg_increase - RTg_compensation) * max(RUGE,0) / (T_glucose_RTg/C_renal_CV_timescale);

osmotic_natriuresis_effect_pt = 1-min(1,RUGE *glucose_natriuresis_effect_pt);
osmotic_natriuresis_effect_cd = 1-min(1,RUGE *glucose_natriuresis_effect_cd);
osmotic_diuresis_effect_pt = 1-min(1,RUGE *glucose_diuresis_effect_pt);
osmotic_diuresis_effect_cd = 1-min(1,RUGE *glucose_diuresis_effect_cd);


###PT Sodium filtration and reabsorption
# Sodium reabsorbed 1:1 with glucose in S1 and S2
# Sodium reabsorbed 2:1 with glucose in S3
# Assume for non-SGLT reabsorption, sodium reabsorbed at a constant RATE along the tubule
# (represents glomerulotubular balance)
SN_filtered_Na_load = (SNGFR_nL_min / 1000 / 1000000)*Na_concentration;  	#mmol/min
SGTL2_Na_reabs_mmol_s1 = SN_filtered_glucose_load-glucose_pt_out_s1;   		#mmol/min
SGTL2_Na_reabs_mmol_s2 = glucose_pt_out_s1-glucose_pt_out_s2;			#mmol/min
SGTL1_Na_reabs_mmol = 2*(glucose_pt_out_s2-glucose_pt_out_s3);			#mmol/min
total_SGLT_Na_reabs = SGTL2_Na_reabs_mmol_s1+SGTL2_Na_reabs_mmol_s2+SGTL1_Na_reabs_mmol; 	#mmol/min



Na_reabs_per_unit_length = -log(1-e_pt_sodreab)/(L_pt_s1+L_pt_s2+L_pt_s3); #non-SGLT2 reabs	#mmol/min
Na_pt_s1_reabs = min(max_s1_Na_reabs, SN_filtered_Na_load*(1-exp(-Na_reabs_per_unit_length*L_pt_s1)));
Na_pt_out_s1 = SN_filtered_Na_load - Na_pt_s1_reabs - SGTL2_Na_reabs_mmol_s1 ;

Na_pt_s2_reabs = min(max_s2_Na_reabs, Na_pt_out_s1*(1-exp(-Na_reabs_per_unit_length*L_pt_s2)));
Na_pt_out_s2 = Na_pt_out_s1 - Na_pt_s2_reabs - SGTL2_Na_reabs_mmol_s2;

Na_pt_s3_reabs = min(max_s3_Na_reabs, Na_pt_out_s2*(1-exp(-Na_reabs_per_unit_length*L_pt_s3)));
Na_pt_out_s3 = Na_pt_out_s2 - Na_pt_s3_reabs - SGTL1_Na_reabs_mmol;

PT_Na_reabs_fraction = 1-Na_pt_out_s3/SN_filtered_Na_load;

###PT Urea filtration and reabsorption
SN_filtered_urea_load = (SNGFR_nL_min / 1000 / 1000000)*plasma_urea;
urea_out_s1 = SN_filtered_urea_load - urea_permeability_PT*(SN_filtered_urea_load/(0.5*((SNGFR_nL_min / 1000 / 1000000)+water_out_s1_delayed))-plasma_urea)*water_out_s1_delayed; #For now, assuming only reabsorbed at the end
urea_out_s2 = urea_out_s1 - urea_permeability_PT*(urea_out_s1/(0.5*(water_out_s1_delayed+water_out_s2_delayed))-plasma_urea)*water_out_s2_delayed; #For now, assuming only reabsorbed at the end
urea_out_s3 = urea_out_s2 - urea_permeability_PT*(urea_out_s2/(0.5*(water_out_s2_delayed+water_out_s3_delayed))-plasma_urea)*water_out_s3_delayed; #For now, assuming only reabsorbed at the end
urea_reabsorption_fraction = 1-urea_out_s3/SN_filtered_urea_load;


###PT Water Reabsorption
osmoles_out_s1 = 2*Na_pt_out_s1 + glucose_pt_out_s1 + urea_out_s1;
water_out_s1 = (((SNGFR_nL_min / 1000 / 1000000)/(2*SN_filtered_Na_load+SN_filtered_glucose_load+ SN_filtered_urea_load)))*osmoles_out_s1;

osmoles_out_s2 = 2*Na_pt_out_s2 + glucose_pt_out_s2 + urea_out_s2;
water_out_s2 = (water_out_s1/osmoles_out_s1)*osmoles_out_s2;

osmoles_out_s3 = 2*Na_pt_out_s3 + glucose_pt_out_s3 + urea_out_s3;
water_out_s3 = (water_out_s2/osmoles_out_s2)*osmoles_out_s3;

PT_water_reabs_fraction = 1-water_out_s3/(SNGFR_nL_min / 1000 / 1000000);

###Concentrations out of PT
Na_concentration_out_s1 = Na_pt_out_s1/water_out_s1;
Na_concentration_out_s2 = Na_pt_out_s2/water_out_s2;
Na_concentration_out_s3 = Na_pt_out_s3/water_out_s3;

glucose_concentration_out_s1 = glucose_pt_out_s1/water_out_s1;
glucose_concentration_out_s2 = glucose_pt_out_s2/water_out_s2;
glucose_concentration_out_s3 = glucose_pt_out_s3/water_out_s3;

urea_concentration_out_s1 = urea_out_s1/water_out_s1;
urea_concentration_out_s2 = urea_out_s2/water_out_s2;
urea_concentration_out_s3 = urea_out_s3/water_out_s3;

osmolality_out_s1 = osmoles_out_s1/water_out_s1;
osmolality_out_s2 = osmoles_out_s2/water_out_s2;
osmolality_out_s3 = osmoles_out_s3/water_out_s3;

PT_Na_outflow = Na_pt_out_s3*number_of_functional_tubules;

#Tubular sodium reabsorption per unit SA as the driver of tubular hypertrophy
PT_Na_reab_perUnitSA = SN_filtered_Na_load*e_pt_sodreab/(3.14*Dc_pt*(L_pt_s1+L_pt_s2+L_pt_s3));

normalized_PT_reabsorption_density = PT_Na_reab_perUnitSA/PT_Na_reab_perUnitSA_0;
PT_Na_reabs_effect_increasing_tubular_length = 0;#(maximal_tubule_length_increase - tubular_length_increase) * max(normalized_PT_reabsorption_density - 1,0) / (T_PT_Na_reabs_PT_length/C_renal_CV_timescale);
PT_Na_reabs_effect_increasing_tubular_diameter = 0;#(maximal_tubule_diameter_increase - tubular_diameter_increase) * max(normalized_PT_reabsorption_density - 1,0) / (T_PT_Na_reabs_PT_diameter/C_renal_CV_timescale);


##################################### Loop of Henle #########################################

#####Descending Loop of Henle

water_in_DescLoH = water_out_s3; # L/min
Na_in_DescLoH = Na_pt_out_s3;
urea_in_DescLoH = urea_out_s3;
glucose_in_DescLoH = glucose_pt_out_s3;
osmoles_in_DescLoH = osmoles_out_s3;

Na_concentration_in_DescLoH = Na_concentration_out_s3;
Urea_concentration_in_DescLoH = urea_concentration_out_s3;
glucose_concentration_in_DescLoH = glucose_concentration_out_s3;
osmolality_in_DescLoH = osmoles_out_s3/water_out_s3;

#No solute reabsorption in descending limb
Na_out_DescLoH = Na_in_DescLoH;
urea_out_DescLoH = urea_in_DescLoH;
glucose_out_DescLoH = glucose_in_DescLoH;
osmoles_out_DescLoH = osmoles_in_DescLoH;


#For LoH, baseline osmoles reabsorbed per unit length is calculated from nominal fractional sodium reabsorption (see baseline parameters file)
#The rate of reabsorption per unit length may be flow-dependent, and may be modulated by tubular pressure-natriuresis
# If LoH_flow_dependence = 0, then no flow dependence.

deltaLoH_NaFlow = min(max_deltaLoH_reabs,LoH_flow_dependence*(Na_out_DescLoH-nom_Na_in_AscLoH));
AscLoH_Reab_Rate =(2*nominal_loh_na_reabsorption*(nom_Na_in_AscLoH+deltaLoH_NaFlow)*loop_diuretic_effect)/L_lh_des; #osmoles reabsorbed per unit length per minute. factor of 2 because osmoles = 2

effective_AscLoH_Reab_Rate =AscLoH_Reab_Rate*pressure_natriuresis_LoH_effect; #osmoles reabsorbed per unit length per minute. factor of 2 because osmoles = 2*Na


#Min function necesssary to ensure that the LoH does not reabsorb more Na than is delivered to it
osmolality_out_DescLoH = osmolality_in_DescLoH*exp(min(effective_AscLoH_Reab_Rate*L_lh_des,2*Na_in_DescLoH)/(water_in_DescLoH*osmolality_in_DescLoH));
water_out_DescLoH = water_in_DescLoH*osmolality_in_DescLoH/osmolality_out_DescLoH;

Na_concentration_out_DescLoH = Na_out_DescLoH/water_out_DescLoH;
glucose_concentration_out_DescLoH = glucose_out_DescLoH/water_out_DescLoH;
urea_concentration_out_DescLoH = urea_out_DescLoH/water_out_DescLoH;

#####Ascending Loop of Henle

Na_in_AscLoH = Na_out_DescLoH;
urea_in_AscLoH_before_secretion = urea_out_DescLoH;
glucose_in_AscLoH = glucose_out_DescLoH;
osmoles_in_AscLoH_before_secretion = osmoles_out_DescLoH;
water_in_AscLoH = water_out_DescLoH;
Na_concentration_in_AscLoH = Na_concentration_out_DescLoH;
#Urea Secretion --> Assume all urea reabsorbed and secreted only at tip of loop
urea_in_AscLoH = urea_in_AscLoH_before_secretion + reabsorbed_urea_cd_delayed;
urea_concentration_in_AscLoH = urea_in_AscLoH/water_out_DescLoH;
osmoles_in_AscLoH = osmoles_in_AscLoH_before_secretion  + reabsorbed_urea_cd_delayed;

osmolality_in_AscLoH = osmoles_in_AscLoH/water_in_AscLoH;

#Osmolality descreased due to sodium reabsorption along ascending loop
#min function necessary so that LoH doesn't reabsorb more sodium than is delivered to it
osmolality_out_AscLoH = osmolality_in_AscLoH - min(L_lh_des*effective_AscLoH_Reab_Rate, 2*Na_in_DescLoH)*(exp(min(L_lh_des*effective_AscLoH_Reab_Rate, 2*Na_in_DescLoH)/(water_in_DescLoH*osmolality_in_DescLoH))/water_in_DescLoH);
osmoles_reabsorbed_AscLoH = (osmolality_in_AscLoH - osmolality_out_AscLoH)*water_in_AscLoH;
Na_reabsorbed_AscLoH = osmoles_reabsorbed_AscLoH/2;
Na_out_AscLoH = max(0,Na_in_AscLoH - Na_reabsorbed_AscLoH);



#Water, glucose, and urea are not reabsorbed along the ascending limb
urea_out_AscLoH = urea_in_AscLoH; #urea secretion accounted for above
glucose_out_AscLoH = glucose_in_AscLoH;
water_out_AscLoH = water_in_AscLoH;

osmoles_out_AscLoH = osmolality_out_AscLoH*water_out_AscLoH;


Na_concentration_out_AscLoH = Na_out_AscLoH/water_out_AscLoH;
glucose_concentration_out_AscLoH = glucose_out_AscLoH/water_out_AscLoH;
urea_concentration_out_AscLoH = urea_out_AscLoH/water_out_AscLoH;

LoH_reabs_fraction = 1-Na_out_AscLoH/Na_in_AscLoH;

SN_macula_densa_Na_flow = Na_out_AscLoH;
MD_Na_concentration = Na_concentration_out_AscLoH;
TGF0_tubulo_glomerular_feedback = 1 - S_tubulo_glomerular_feedback/2;
tubulo_glomerular_feedback_signal = (TGF0_tubulo_glomerular_feedback + S_tubulo_glomerular_feedback / (1 + exp((MD_Na_concentration_setpoint - MD_Na_concentration)/ F_md_scale_tubulo_glomerular_feedback)));


#############################Distal Convoluted Tubule #######################


water_in_DCT = water_out_AscLoH;
Na_in_DCT = Na_out_AscLoH;
urea_in_DCT = urea_out_AscLoH;
glucose_in_DCT = glucose_out_AscLoH;
osmoles_in_DCT = osmoles_out_AscLoH;

Na_concentration_in_DCT = Na_concentration_out_AscLoH;
urea_concentration_in_DCT = urea_concentration_out_AscLoH;
glucose_concentration_in_DCT = glucose_concentration_out_AscLoH;
osmolality_in_DCT = osmolality_out_AscLoH;

#Assume only sodium reabsorbed along DCT, no water, glucose, or urea reabsorption
urea_out_DCT = urea_in_DCT;
glucose_out_DCT = glucose_in_DCT;
water_out_DCT = water_in_DCT;

urea_concentration_out_DCT = urea_out_DCT/water_out_DCT;
glucose_concentration_out_DCT = glucose_out_DCT/water_out_DCT;

#Assume sodium reabsorption at a constant fraction of delivery
R_dct = -log(1-e_dct_sodreab)/L_dct;

Na_out_DCT = Na_in_DCT*exp(-R_dct*L_dct);
Na_concentration_out_DCT = Na_out_DCT/water_out_DCT;

osmolality_out_DCT = 2*Na_concentration_out_DCT + glucose_concentration_out_DescLoH + urea_concentration_in_AscLoH;
osmoles_out_DCT = osmolality_out_DCT*water_out_DCT;

DCT_Na_reabs_fraction = 1-Na_out_DCT/Na_in_DCT;

Na_reabsorbed_DCT = Na_in_DCT - Na_out_DCT;




################################################Collecting Duct###############################

water_in_CD = water_out_DCT;
Na_in_CD = Na_out_DCT;
urea_in_CD = urea_out_DCT;
glucose_in_CD = glucose_out_DCT;

osmoles_in_CD = osmoles_out_DCT;

#Use this to turn off osmotic diuresis effect
#osmoles_in_CD = osmoles_out_DCT - glucose_in_CD;

osmolality_in_CD = osmoles_in_CD/water_in_CD;

Na_concentration_in_CD = Na_concentration_out_DCT;
urea_concentration_in_CD = urea_concentration_out_DCT;
glucose_concentration_in_CD = glucose_concentration_out_DCT;

osmotic_diuresis_effect_cd = 1-min(1,RUGE *glucose_diuresis_effect_cd);

####Assume sodium reabsorbed, then water follows
#### Then urea reabsorbed at end
#### Then additional water reabsorbed following urea reabsorption

#Assume sodium reabsorbed at fractional rate eta
e_cd_sodreab_adj = e_cd_sodreab*osmotic_natriuresis_effect_cd;
R_cd = -log(1-e_cd_sodreab_adj)/L_cd;
Na_reabsorbed_CD = min(Na_in_CD*(1-exp(-R_cd*L_cd)),CD_Na_reabs_threshold);
Na_out_CD = Na_in_CD-Na_reabsorbed_CD;

CD_Na_reabs_fraction = 1-Na_out_CD/Na_in_CD;

Na_reabsorbed_CT=Na_in_CD-Na_out_CD;

ADH_water_permeability_old = min(0.99999,max(0,nom_ADH_water_permeability*normalized_vasopressin_concentration));
ADH_water_permeability = normalized_vasopressin_concentration/(0.15+normalized_vasopressin_concentration);


osmoles_out_CD = osmoles_in_CD-2*(Na_in_CD - Na_out_CD);

osmolality_out_CD_before_osmotic_reabsorption = osmoles_out_CD/water_in_CD;
water_reabsorbed_CD = ADH_water_permeability*osmotic_diuresis_effect_cd*water_in_CD*(1-osmolality_out_CD_before_osmotic_reabsorption/osmolality_out_DescLoH);
water_out_CD = water_in_CD-water_reabsorbed_CD;
Na_concentration_out_CD = Na_out_CD/water_out_CD;
osmolality_out_CD_after_osmotic_reabsorption = osmoles_out_CD/water_out_CD;
glucose_concentration_after_urea_reabsorption = glucose_in_CD/water_out_CD;

urine_flow_rate = water_out_CD*number_of_functional_tubules;

daily_urine_flow = (urine_flow_rate * 60 * 24);

Na_excretion_via_urine = Na_out_CD*number_of_functional_tubules;
Na_balance = Na_intake_rate - Na_excretion_via_urine;
water_balance = daily_water_intake - daily_urine_flow;


total_NA_reabsorbed = (total_SGLT_Na_reabs +Na_reabsorbed_AscLoH +Na_reabsorbed_DCT +Na_reabsorbed_CT  )*number_of_functional_tubules;

Na_concentration_average_PT = (Na_concentration_out_s1 + Na_concentration_out_s2+Na_concentration_out_s3)/3;
Na_concentration_average_DescLoH = 0.5*(Na_concentration_in_DescLoH + Na_concentration_out_DescLoH);
Na_concentration_average_AscLoH = 0.5*(Na_concentration_in_AscLoH + Na_concentration_out_AscLoH );
Na_concentration_average_DCT = 0.5*(Na_concentration_in_DCT + Na_concentration_out_DCT);
Na_concentration_average_CD = 0.5*(Na_concentration_in_CD + Na_concentration_out_CD);

Na_concentration_average_tubule = 0.2*(Na_concentration_average_PT+Na_concentration_average_DescLoH+Na_concentration_average_AscLoH+Na_concentration_average_DCT+Na_concentration_average_CD);

Oncotic_pressure_tubule = Na_concentration_average_tubule*19.3*2;

Na_amount_in_renal_interstitium = total_NA_reabsorbed;
Na_concentration_renal_interstitium = Na_amount_in_renal_interstitium/(0.03831672); #Total kidney volume 202ml. Interstitium = 0.13*KV
Oncotic_pressure_renal_interstitium = Na_concentration_renal_interstitium*19.3*2;

Net_oncotic_pressure = Oncotic_pressure_tubule-Oncotic_pressure_renal_interstitium;

FENA = Na_excretion_via_urine/filtered_Na_load;

PT_fractional_glucose_reabs = (SN_filtered_glucose_load - glucose_pt_out_s3)/SN_filtered_glucose_load;
PT_fractional_Na_reabs = (SN_filtered_Na_load - Na_pt_out_s3)/SN_filtered_Na_load;
PT_fractional_urea_reabs = ( SN_filtered_urea_load - urea_out_s3)/SN_filtered_urea_load;
PT_fractional_water_reabs = ((SNGFR_nL_min / 1000 / 1000000) - water_out_s3)/(SNGFR_nL_min / 1000 / 1000000);

LoH_fractional_Na_reabs = (Na_in_DescLoH - Na_out_AscLoH)/Na_in_DescLoH;
LoH_fractional_urea_reabs = (urea_in_DescLoH-urea_out_AscLoH)/urea_in_DescLoH;
LoH_fractional_water_reabs = (water_in_DescLoH - water_out_AscLoH)/water_in_DescLoH;

DCT_fractional_Na_reabs = (Na_in_DCT - Na_out_DCT)/Na_in_DCT;

CD_fractional_Na_reabs = (Na_in_CD - Na_out_CD)/Na_in_CD;
#CD_fractional_urea_reabs = (urea_in_CD - urea_out_CD)/urea_in_CD;
CD_OM_fractional_water_reabs = (water_in_CD - water_out_CD)/water_in_CD;


#####################Renal Interstitial Hydrostatic pressure

######RIHP can be approximated from Starling's equation for the peritubular capillaries
### Flow out of the capillary = Kf_peritubular*(Peritubular pressure - RIHP - oncotic pressure difference)
### Assume that any fluid reabsorbed reenters the capillary.
### Therefore, RIHP = Peritubular Pressure - (oncotic pressure in peritubular capillary - interstitial oncotic pressure) + tubular reabsorption/KF
#Peritubular pressure is assumed to equal postglomerular pressure

#Oncotic pressure at the entrance to the peritubular circulation equals oncotic pressure at the exit of the glomerular
Oncotic_pressure_peritubular_in = Oncotic_pressure_out;
plasma_protein_concentration_peritubular_out = (SNRBF_nl_min)*plasma_protein_concentration/(SNRBF_nl_min-urine_flow_rate*1e6*1000/number_of_functional_glomeruli);
Oncotic_pressure_peritubular_out = 1.629*plasma_protein_concentration_peritubular_out+0.2935*(plasma_protein_concentration_peritubular_out^2);
oncotic_pressure_peritubular_avg = (Oncotic_pressure_peritubular_in+Oncotic_pressure_peritubular_out)/2;

Na_concentration_peritubular_cap = (sodium_amount-filtered_Na_load)/blood_volume_L;
oncotic_pressure_peritubular_cap_Na = 0;#Na_concentration_peritubular_cap*19.3*2;
oncotic_pressure_peritubular = oncotic_pressure_peritubular_avg+oncotic_pressure_peritubular_cap_Na;

tubular_reabsorption = GFR_ml_min/1000 - urine_flow_rate;

Volume = interstitial_fluid_volume;

volume_RIHP_int = 1- volume_RIHP_scale/2;
IF_effect_RIHP = (volume_RIHP_int+volume_RIHP_scale/(1+exp(-(Volume-IF_nom)/volume_RIHP_slope)));

oncotic_int = 1-IF_Interonvotic_Effect_scale/2;
IF_effect_oncotic = (oncotic_int+IF_Interonvotic_Effect_scale/(1+exp((Volume-IF_nom)/IF_Interonvotic_Effect_slope)));

#########################################################################

Renal_plasma_amount= 2.5 * RISF_nom*0.01;  # plasma amount in renal interstitium   # Plasma protein concentration = 7
RISF_plasma_protein_concentration = Renal_plasma_amount / (RISF*10);
interstitial_oncotic_pressure = 1.629*RISF_plasma_protein_concentration+0.2935*(RISF_plasma_protein_concentration^2);
RIHP = RISF/C_RISF;
capillary_filtration = nom_peritubular_cap_Kf*(RIHP - postglomerular_pressure - (interstitial_oncotic_pressure -oncotic_pressure_peritubular));

#RIHP = IF_effect_RIHP*(postglomerular_pressure - (oncotic_pressure_peritubular - interstitial_oncotic_pressure*IF_effect_oncotic) + tubular_reabsorption/nom_peritubular_cap_Kf);


################# Tubular Pressure #####################

#####See written documentation for derivation of the equations below
#flow rates expressed in m3/min, rather than L/min


mmHg_Nperm2_conv = 133.32;
Pc_pt_s1 = Pc_pt_s1_mmHg*mmHg_Nperm2_conv;
Pc_pt_s2 = Pc_pt_s2_mmHg*mmHg_Nperm2_conv;
Pc_pt_s3 = Pc_pt_s3_mmHg*mmHg_Nperm2_conv;
Pc_lh_des = Pc_lh_des_mmHg*mmHg_Nperm2_conv;
Pc_lh_asc = Pc_lh_asc_mmHg*mmHg_Nperm2_conv;
Pc_dt = Pc_dt_mmHg*mmHg_Nperm2_conv;
Pc_cd = Pc_cd_mmHg*mmHg_Nperm2_conv;
P_interstitial = (RIHP)*mmHg_Nperm2_conv;# 4.9*mmHg_Nperm2_conv;
pi=3.14;

###CD
B1 = (4*tubular_compliance+1)*128*gamma/pi;
mean_cd_water_flow = (water_in_CD-water_out_CD)/2;
B2_cd = (Pc_cd^(4*tubular_compliance))/(Dc_cd^4);
P_in_cd = (0^(4*tubular_compliance+1)+B1*B2_cd*(mean_cd_water_flow/1e3)*L_cd)^(1/(4*tubular_compliance+1));
P_in_cd_mmHg = (P_in_cd+P_interstitial)/mmHg_Nperm2_conv;



### DCT
B2_dt = (Pc_dt^(4*tubular_compliance))/(Dc_dt^4);
P_in_dt = (P_in_cd^(4*tubular_compliance+1)+B1*B2_dt*(water_in_DCT/1e3)*L_dct)^(1/(4*tubular_compliance+1));
P_in_dt_mmHg = (P_in_dt+P_interstitial)/mmHg_Nperm2_conv;



### Asc LoH
B2_lh_asc = (Pc_lh_asc^(4*tubular_compliance))/(Dc_lh^4);
P_in_lh_asc = (P_in_dt^(4*tubular_compliance+1)+B1*B2_lh_asc*(water_in_AscLoH/1e3)*L_lh_asc)^(1/(4*tubular_compliance+1));
P_in_lh_asc_mmHg = (P_in_lh_asc+P_interstitial)/mmHg_Nperm2_conv;

### Desc LoH
A_lh_des = effective_AscLoH_Reab_Rate/(water_in_DescLoH*osmolality_in_DescLoH);
B2_lh_des = (Pc_lh_des^(4*tubular_compliance))*(water_in_DescLoH/1e3)/((Dc_lh^4)*A_lh_des);
P_in_lh_des = (P_in_lh_asc^(4*tubular_compliance+1)+B1*B2_lh_des*(1-exp(-A_lh_des*L_lh_des)))^(1/(4*tubular_compliance+1));
P_in_lh_des_mmHg = (P_in_lh_des+P_interstitial)/mmHg_Nperm2_conv;

### PT

#Treat urea as if reabsorbed linearly along whole length of PT
Rurea = (SN_filtered_urea_load - urea_out_s3)/(L_pt_s1+L_pt_s2+L_pt_s3);
urea_in_s2 = SN_filtered_urea_load - Rurea*L_pt_s1;
urea_in_s3 = SN_filtered_urea_load - Rurea*(L_pt_s1+L_pt_s2);
A_na = Na_reabs_per_unit_length;

flow_integral_s3 = 2*(Na_pt_out_s2/A_na)*(1-exp(-A_na*L_pt_s3)) - (3/2)*glucose_pt_out_s2*L_pt_s3^2 + urea_in_s3*L_pt_s3 - (1/2)*Rurea*(L_pt_s3^2);
flow_integral_s2 = 2*(Na_pt_out_s1/A_na)*(1-exp(-A_na*L_pt_s2)) - (1/2)*glucose_pt_out_s1*L_pt_s2^2 + urea_in_s2*L_pt_s2 - (1/2)*Rurea*(L_pt_s2^2);
flow_integral_s1 = 2*(SN_filtered_Na_load/A_na)*(1-exp(-A_na*L_pt_s1)) - (1/2)*SN_filtered_glucose_load*L_pt_s1^2 + SN_filtered_urea_load*L_pt_s1 - (1/2)*Rurea*(L_pt_s1^2);


#S3 segment
B2_pt_s3 = (Pc_pt_s3^(4*tubular_compliance))/(Dc_pt^4);
B3_pt_s3 = (water_out_s2/1e3)/osmoles_out_s2;
P_in_pt_s3= (P_in_lh_des^(4*tubular_compliance+1)+B1*B2_pt_s3*B3_pt_s3*flow_integral_s3)^(1/(4*tubular_compliance+1));
P_in_pt_s3_mmHg = (P_in_pt_s3+P_interstitial)/mmHg_Nperm2_conv;

B2_pt_s2 = (Pc_pt_s3^(4*tubular_compliance))/(Dc_pt^4);
B3_pt_s2 = (water_out_s1/1e3)/osmoles_out_s1;
P_in_pt_s2= (P_in_pt_s3^(4*tubular_compliance+1)+B1*B2_pt_s2*B3_pt_s2*flow_integral_s2)^(1/(4*tubular_compliance+1));
P_in_pt_s2_mmHg = (P_in_pt_s2+P_interstitial)/mmHg_Nperm2_conv;

B2_pt_s1 = (Pc_pt_s1^(4*tubular_compliance))/(Dc_pt^4);
B3_pt_s1 = (SNGFR_nL_min / 1e12)/(2*SN_filtered_Na_load+SN_filtered_glucose_load+ SN_filtered_urea_load);
P_in_pt_s1= (P_in_pt_s2^(4*tubular_compliance+1)+B1*B2_pt_s1*B3_pt_s1*flow_integral_s1)^(1/(4*tubular_compliance+1));
P_in_pt_s1_mmHg = (P_in_pt_s1+P_interstitial)/mmHg_Nperm2_conv;




####################### Aldosterone and Renin Secretion

###Aldosterone is secreted in response to AT1-bound AngII and changes in potassium or sodium concentration
#Potassium concentration is treated as a constant for now
#Empirircal relationship for Karaaslan 2005


AT1_aldo_int = 1 - AT1_aldo_slope*nominal_equilibrium_AT1_bound_AngII;
AngII_effect_on_aldo = AT1_aldo_int + AT1_aldo_slope*AT1_bound_AngII;
N_als = (K_Na_ratio_effect_on_aldo * AngII_effect_on_aldo );

###Renin is secreted in response to decreases in AT1-bound AngII and decreases in MD sodium flow

###Renin is secreted in response to decreases in AT1-bound AngII, decreases in MD sodium flow, or increases in RSNA
#RSNA effect on renin secretion
rsna_renin_intercept = 1-rsna_renin_slope;
rnsa_effect_on_renin_secretion =  rsna_renin_slope * renal_sympathetic_nerve_activity + rsna_renin_intercept;


#Macula Densa Sodium flow effect on renin secretion
#This relationship is known to be non-linear, and md_renin_tau can be calibrated based on data on changes in renin as a functoin of sodium intake
md_effect_on_renin_secretion = md_renin_A*exp(-md_renin_tau*(SN_macula_densa_Na_flow_delayed*baseline_nephrons - nom_LoH_Na_outflow));


#AT1-bound AngII feedback on renin secretion
AT1_bound_AngII_effect_on_PRA = (10 ^ (AT1_PRC_slope * log10(AT1_bound_AngII / nominal_equilibrium_AT1_bound_AngII) + AT1_PRC_yint));


#Aldo effect on renin secretion
aldo_renin_intercept = 1-aldo_renin_slope;
aldo_effect_on_renin_secretion =  aldo_renin_intercept + aldo_renin_slope*Aldo_MR_normalised_effect;

#Plasma renin activity
plasma_renin_activity = concentration_to_renin_activity_conversion_plasma* plasma_renin_concentration*(1-pct_target_inhibition_DRI);

#Renin secretion
renin_secretion_rate = (log(2)/renin_half_life)*nominal_equilibrium_PRC*AT1_bound_AngII_effect_on_PRA*md_effect_on_renin_secretion*HCTZ_effect_on_renin_secretion*aldo_effect_on_renin_secretion*BB_renin_secretion_effect;

#RAAS degradation rates
renin_degradation_rate = log(2)/renin_half_life;
AngI_degradation_rate = log(2)/AngI_half_life;
AngII_degradation_rate = log(2)/AngII_half_life;
AT1_bound_AngII_degradation_rate =  log(2)/AT1_bound_AngII_half_life;
AT2_bound_AngII_degradation_rate = log(2)/AT2_bound_AngII_half_life;

#RAAS rate constants
ACE_activity = nominal_ACE_activity*(1 - pct_target_inhibition_ACEi);
chymase_activity = nominal_chymase_activity;
AT1_receptor_binding_rate = nominal_AT1_receptor_binding_rate*(1 - pct_target_inhibition_ARB);
AT2_receptor_binding_rate = nominal_AT2_receptor_binding_rate;



Blood_volume_protein_osmotic_pressure = 1.629*plasma_protein_concentration + 0.2935*plasma_protein_concentration^2;
ISF_protein_osmotic_pressure = 1.629*ISF_protein_concentration + 0.2935*ISF_protein_concentration^2;

Blood_volume_osmotic_pressure = Blood_volume_protein_osmotic_pressure+ Na_concentration*19.3*2;
ISF_osmotic_pressire = IF_Na_concentration*19.3*2 + ISF_protein_osmotic_pressure;
Protein_sodium_filtration_pressure_grad = (mean_capillary_pressure - ISF_pressure - Blood_volume_osmotic_pressure + ISF_osmotic_pressire);

Kidney_disconnect_heart = Q_water*(Na_concentration - IF_Na_concentration);
Kidney_connect_heart = -Sodium_protein_filtration_rate_Kf*(Protein_sodium_filtration_pressure_grad)*0.001;

if (heart_renal_link == 1) {
Fluid_exchanging_function=Kidney_connect_heart;
} else{
Fluid_exchanging_function=Kidney_connect_heart;
}

peripheral_volume_change=arterial_dis_circulation_volume+capillary_circulation_volume+venules_circulation_volume;
#d/dt(HR_ratio) = heart_rate_constant*heart_rate_multiplier_adjusted;
d/dt(venous_volume) = venous_flow + C_renal_CV_timescale*(venous_volume_target - venous_volume) - tricuspid_valve_flow_rate ;
d/dt(LV_volume) = mitral_valve_flow_rate - aortic_valve_flow_rate;
d/dt(arterial_volume) = (aortic_valve_flow_rate) - (systemic_blood_flow);
#d/dt(peripheral_circulation_volume)= systemic_blood_flow - venous_flow;
d/dt(RV_volume) = (tricuspid_valve_flow_rate) - (pulmonary_valve_flow_rate);
d/dt(pulmonary_arterial_volume) = pulmonary_valve_flow_rate - pulmonary_arterial_blood_flow;
#d/dt(peripheral_circulation_volume) = (systemic_blood_flow) - venous_flow;

d/dt(arterial_dis_circulation_volume)=systemic_blood_flow - capillary_blood_flow;
d/dt(capillary_circulation_volume)=arterial_dis_blood_flow-venules_blood_flow;
d/dt(venules_circulation_volume)=capillary_blood_flow-venous_flow;

d/dt(pulmonary_venous_volume) = pulmonary_arterial_blood_flow - mitral_valve_flow_rate;
d/dt(aortic_blood_flow_delayed) = C_cycle2 * (aortic_blood_flow - aortic_blood_flow_delayed);
d/dt(pulmonary_blood_flow_delayed) = C_cycle2 * (pulmonary_blood_flow - pulmonary_blood_flow_delayed);

#################################################################################################
#################################################################################################
## Hypertrophy
d/dt(change_in_myocyte_length) = kL_hypertrophy * (LV_EDS / LV_passive_stress_along_fiber_threshhold - 1);
d/dt(change_in_myocyte_diameter) = kD_hypertrophy * (LV_active_stress_peak / LV_active_stress_threshhold - 1);
d/dt(LV_active_stress_peak) = C_cycle3 *(LV_active_stress_peak_old - LV_active_stress_peak);
d/dt(sim_time)=1;
d/dt(LV_sarcomere_length_delayed) = C_cycle* (LV_sarcomere_length - LV_sarcomere_length_delayed);
d/dt(RV_sarcomere_length_delayed) = C_cycle* (RV_sarcomere_length - RV_sarcomere_length_delayed);
d/dt(LV_EDV) = C_cycle2 * (LV_EDV_old - LV_EDV);
d/dt(LV_EDP) = C_cycle2 *(LV_EDP_old - LV_EDP);
#d/dt(LV_PSP) = C_cycle2 *(LV_PSP_old - LV_PSP);
d/dt(LV_EDS) = C_cycle2 *(LV_EDS_old - LV_EDS);

d/dt(arterial_pressure_delayed) = C_cycle2 * (arterial_pressure - arterial_pressure_delayed);
d/dt(arterial_pressure_bigger_delay) = C_cycle2 * (arterial_pressure_delayed - arterial_pressure_bigger_delay);
d/dt(systolic_pressure) = C_cycle2 * (systolic_pressure_old - systolic_pressure);
d/dt(diastolic_pressure) = C_cycle2 * (diastolic_pressure_old - diastolic_pressure);

######################computing the systolic and diastolic venous pressure##########
d/dt(venous_pressure_delayed)=C_cycle2*(venous_pressure-venous_pressure_delayed);
d/dt(venous_pressure_bigger_delay)=C_cycle2*(venous_pressure_delayed-venous_pressure_bigger_delay);
d/dt(venous_systolic_pressure)=C_cycle2*(venous_systolic_pressure_old-venous_systolic_pressure);
d/dt(venous_diastolic_pressure)=C_cycle2*(venous_diastolic_pressure_old-venous_diastolic_pressure);
d/dt(mean_venous_pressure_delayed) = 1*(mean_venous_pressure - mean_venous_pressure_delayed);

######################computing the systolic and diastolic peripheral pressure##########
d/dt(capillary_pressure_delayed)=C_cycle2*(capillary_pressure-capillary_pressure_delayed);
d/dt(capillary_pressure_bigger_delay)=C_cycle2*(capillary_pressure_delayed-capillary_pressure_bigger_delay);
d/dt(capillary_systolic_pressure)=C_cycle2*(capillary_systolic_pressure_old-capillary_systolic_pressure);
d/dt(capillary_diastolic_pressure)=C_cycle2*(capillary_diastolic_pressure_old-capillary_diastolic_pressure);
d/dt(mean_capillary_pressure_delayed) = 1*(mean_capillary_pressure - mean_capillary_pressure_delayed);


d/dt(CO) = C_co*(aortic_valve_flow_rate*60/L_m3 - CO);
d/dt(CO_delayed) = C_co_delay*(CO - CO_delayed);

#RAAS Pathway
d/dt(AngI) = plasma_renin_activity - (AngI) * (chymase_activity + ACE_activity) - (AngI) * AngI_degradation_rate;
d/dt(AngII) = AngI * (chymase_activity + ACE_activity) - AngII * AngII_degradation_rate - AngII*AT1_receptor_binding_rate - AngII* (AT2_receptor_binding_rate);
d/dt(AT1_bound_AngII) = AngII * (AT1_receptor_binding_rate) - AT1_bound_AngII_degradation_rate*AT1_bound_AngII;
d/dt(AT2_bound_AngII) = AngII * (AT2_receptor_binding_rate) - AT2_bound_AngII_degradation_rate*AT2_bound_AngII;
d/dt(plasma_renin_concentration) = renin_secretion_rate - plasma_renin_concentration * renin_degradation_rate;

#Change in Interstitial fluid volume over time is determined by the different between water intake and urine outflow
d/dt(blood_volume_L) = C_renal_CV_timescale *(water_intake- urine_flow_rate+Fluid_exchanging_function);  #
d/dt(interstitial_fluid_volume) = -C_renal_CV_timescale *Fluid_exchanging_function;

#Change in total body sodium over time is determined by the different between sodium intake and excretion
d/dt(sodium_amount) = C_renal_CV_timescale * (Na_intake_rate - Na_excretion_via_urine + Q_Na*(IF_Na_concentration - Na_concentration));
d/dt(IF_sodium_amount) = C_renal_CV_timescale *(Q_Na*(Na_concentration - IF_Na_concentration) - sodium_storate_rate);

d/dt(stored_sodium) = C_renal_CV_timescale *sodium_storate_rate;

#These equations serve only to delay the input variable by one timestep. This allows the previous value of the input variable to be used in an equation that appears
#in the code before the input variable was defined
d/dt(tubulo_glomerular_feedback_effect) =  C_renal_CV_timescale *(tubulo_glomerular_feedback_signal-tubulo_glomerular_feedback_effect);
d/dt(normalized_aldosterone_level) =  C_renal_CV_timescale *C_aldo_secretion * (N_als-normalized_aldosterone_level);
d/dt(preafferent_pressure_autoreg_signal) = C_renal_CV_timescale *100*(preafferent_pressure_autoreg_function - preafferent_pressure_autoreg_signal);
d/dt(glomerular_pressure_autoreg_signal) = 0;#C_glomerular_pressure_autoreg_signal*(glomerular_pressure_autoreg_function - glomerular_pressure_autoreg_signal);

#Error signals for PI controllers of cardiac output and sodium concentration
d/dt(CO_error) = C_renal_CV_timescale*C_co_error*(CO_delayed-CO_nom);
d/dt(Na_concentration_error) = C_renal_CV_timescale *C_Na_error*(Na_concentration - ref_Na_concentration);

#This equation allows a delay between the secretion of vasopression and its effect on water intake and tubular water reabsorption
d/dt(normalized_vasopressin_concentration_delayed)= C_renal_CV_timescale *C_vasopressin_delay*(normalized_vasopressin_concentration - normalized_vasopressin_concentration_delayed);

#TGF resetting. If C_tgf_reset = 0, no TGF resetting occurs. If it is greater than zero, the setpoint will change over time and will eventually
#come to equal the ambient MD sodium flow rate.
d/dt(F0_TGF) = C_renal_CV_timescale *C_tgf_reset*(SN_macula_densa_Na_flow*baseline_nephrons - F0_TGF);

#As above, these equations allow a variable to be used in equations that appear in the code before the variable was first defined.
d/dt(P_bowmans) = C_renal_CV_timescale *100*(P_in_pt_s1_mmHg - P_bowmans);
d/dt(oncotic_pressure_difference) = 100*(oncotic_pressure_avg - oncotic_pressure_difference);
d/dt(renal_blood_flow_L_min_delayed)=C_renal_CV_timescale*C_rbf*(renal_blood_flow_L_min - renal_blood_flow_L_min_delayed);
d/dt(SN_macula_densa_Na_flow_delayed) = C_renal_CV_timescale * C_md_flow*( SN_macula_densa_Na_flow - SN_macula_densa_Na_flow_delayed);
d/dt(rsna_delayed) = C_renal_CV_timescale *C_rsna*(renal_sympathetic_nerve_activity - rsna_delayed);

###Disease effects (turned off by default)
#Glomerular hypertrophy
d/dt(disease_effects_increasing_Kf) = GP_effect_increasing_Kf;
#Loss of CD pressure natriuresis response over time
d/dt(disease_effects_decreasing_CD_PN) = CD_PN_loss_rate;

#Tubular hypertrophy
d/dt(tubular_length_increase) = PT_Na_reabs_effect_increasing_tubular_length;
d/dt(tubular_diameter_increase) = PT_Na_reabs_effect_increasing_tubular_diameter;


d/dt(water_out_s1_delayed) = C_renal_CV_timescale * C_pt_water*(water_out_s1 - water_out_s1_delayed);
d/dt(water_out_s2_delayed) = C_renal_CV_timescale * C_pt_water*(water_out_s2 - water_out_s2_delayed);
d/dt(water_out_s3_delayed) = C_renal_CV_timescale * C_pt_water*(water_out_s3 - water_out_s3_delayed);
d/dt(reabsorbed_urea_cd_delayed) = 0;#C_pt_water*(reabsorbed_urea_cd - reabsorbed_urea_cd_delayed);

#Urinary glucose excretion
d/dt(UGE) = C_renal_CV_timescale * RUGE;

#Serum Creatinine
d/dt(serum_creatinine) = C_renal_CV_timescale*(creatinine_synthesis_rate - creatinine_clearance_rate);
d/dt(cumNaExcretion) = C_renal_CV_timescale*Na_excretion_via_urine;
d/dt(cumWaterExcretion) = C_renal_CV_timescale*urine_flow_rate;
d/dt(cumCreatinineExcretion) = C_renal_CV_timescale*creatinine_clearance_rate;
d/dt(RTg_compensation) = C_renal_CV_timescale*excess_glucose_increasing_RTg;
d/dt(SGLT2_inhibition_delayed) = C_renal_CV_timescale*C_sglt2_delay*(SGLT2_inhibition - SGLT2_inhibition_delayed);
d/dt(RUGE_delayed) = C_renal_CV_timescale*C_ruge*(RUGE - RUGE_delayed);
d/dt(postglomerular_pressure_delayed) = C_renal_CV_timescale*C_postglomerular_pressure*(postglomerular_pressure - postglomerular_pressure_delayed); #necessary to prevent large oscillations
d/dt(postglomerular_pressure_error) = C_renal_CV_timescale*(postglomerular_pressure - RIHP0);
d/dt(renal_flow_rate_error) = C_renal_CV_timescale*(renal_blood_flow_L_min - nom_renal_blood_flow_L_min);
d/dt(MAP_delayed) = C_renal_CV_timescale*C_cycle2*(mean_arterial_pressure_MAP - MAP_delayed); #necessary to prevent large oscillations;
d/dt(RIHP_delayed)=C_renal_CV_timescale*C_cycle2*(RIHP - RIHP_delayed);
d/dt(Net_oncotic_pressure_diff) = C_renal_CV_timescale*C_cycle2*(Net_oncotic_pressure - Net_oncotic_pressure_diff);
d/dt(RISF) = tubular_reabsorption - capillary_filtration;
")

  test_that("large models compile", {
    expect_s3_class(mod, "rxode2")
  })
})

## Decay test: dx/dt = -x, x(0)=1 over [0,1]. Exact solution: x(1) = exp(-1).
## Mirrors rklib's rk_decay.F90 test. Tolerance 1e-4 (same as rklib).
rxTest({
  .mod <- rxode2::rxode2({ d/dt(x) <- -x })
  .ev  <- rxode2::et(1)
  .target <- exp(-1)

  ## --- Fixed-step methods (hmin = step size) --------------------------------
  .fixedMethods <- list(
    euler    = list(method = "euler",    hmin = 1e-4),  # O(h): needs h<=2e-4 for err<1e-4
    midpoint = list(method = "midpoint", hmin = 1e-3),
    heun     = list(method = "heun",     hmin = 1e-3),
    ssp22  = list(method = "ssp22",  hmin = 1e-3),
    rk3      = list(method = "rk3",      hmin = 1e-2),
    ssp53  = list(method = "ssp53",  hmin = 1e-2),
    s4     = list(method = "s4",     hmin = 1e-2),
    r4     = list(method = "r4",     hmin = 1e-2),
    ls44   = list(method = "ls44",   hmin = 1e-2),
    ls54   = list(method = "ls54",   hmin = 1e-2),
    ssp54  = list(method = "ssp54",  hmin = 1e-2),
    s5     = list(method = "s5",     hmin = 1e-1),
    rk5    = list(method = "rk5",    hmin = 1e-1),
    c5     = list(method = "c5",     hmin = 1e-1),
    l5     = list(method = "l5",     hmin = 1e-1),
    lk5a   = list(method = "lk5a",   hmin = 1e-1),
    lk5b   = list(method = "lk5b",   hmin = 1e-1),
    b6     = list(method = "b6",     hmin = 1e-1),
    s7     = list(method = "s7",     hmin = 1e-1),
    s8_10  = list(method = "s8_10",  hmin = 1e-1),
    cv8    = list(method = "cv8",    hmin = 1e-1),
    s8_12  = list(method = "s8_12",  hmin = 1e-1),
    s10    = list(method = "s10",    hmin = 1.0),
    z10    = list(method = "z10",    hmin = 1.0),
    o10    = list(method = "o10",    hmin = 1.0),
    h10    = list(method = "h10",    hmin = 1.0)
  )

  for (.nm in names(.fixedMethods)) {
    .cfg <- .fixedMethods[[.nm]]
    test_that(paste("decay test:", .nm), {
      .s <- rxode2::rxSolve(.mod, .ev, inits = c(x = 1),
                             method = .cfg$method, hmin = .cfg$hmin, cores = 1)
      expect_lt(abs(.s$x[1] - .target), 1e-4,
                label = paste(.nm, "decay error"))
    })
  }

  ## --- Adaptive (variable-step) methods: atol=rtol=1e-6 --------------------
  .adaptive_methods <- c(
    ## aliases for existing implementations
    "ck54", "dp54", "v65e", "v76e", "dp87", "v98e", "ssp33",
    ## new variable-step methods
    "bs32", "ssp43", "f45",
    "t54", "s54", "pp54", "pp54b", "bs54", "ss54",
    "dp65", "c65", "tp64", "v65r", "v65", "dverk65", "tf65",
    "tp75", "tmy7", "tmy7s", "v76r", "ss76", "v78", "dverk78",
    "dp85", "tp86", "v87e", "v87r", "ev87", "k87", "f89", "v89",
    "t98a", "v98r", "s98",
    "f108", "c108", "b109", "s1110a", "f1210", "o129", "f1412"
  )

  for (.meth in .adaptive_methods) {
    test_that(paste("decay test:", .meth), {
      .s <- rxode2::rxSolve(.mod, .ev, inits = c(x = 1),
                             method = .meth, atol = 1e-8, rtol = 1e-6, cores = 1)
      expect_lt(abs(.s$x[1] - .target), 1e-4,
                label = paste(.meth, "decay error"))
    })
  }
})

rxTest({
  test_that("state-dep alag uses initial state at sort phase (not NA)", {
    # alag(depot) = central * 0.1
    # central=0 -> lag=0; central=5 -> lag=0.5 hours
    mod <- rxode2({
      alag(depot) <- central * 0.1
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - kel * central
    })
    et <- et(amt = 100) |> et(seq(0, 24, by = 1))
    s0 <- rxSolve(mod, et, c(ka = 0.5, kel = 0.2),
                  inits = c(depot = 0, central = 0))
    s5 <- rxSolve(mod, et, c(ka = 0.5, kel = 0.2),
                  inits = c(depot = 0, central = 5))
    # Previously produced NA due to ypNA placeholder; now should work
    expect_false(any(is.na(s0$central)))
    expect_false(any(is.na(s5$central)))
    # Different initial lag -> different trajectories
    expect_false(isTRUE(all.equal(s0$central, s5$central)))
  })

  test_that("dur(cmt) <- state with constant alag: stop event at absolute lagged time", {
    # alag=0.5, dur=state-dep -> absolute stop = (t + 0.5) + dur(central)
    mod <- rxode2({
      alag(depot) <- 0.5
      dur(depot) <- central * 0.1 + 1.0
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - kel * central
    })
    et <- et(amt = 100, rate = -2) |> et(seq(0, 24, by = 1))
    s <- rxSolve(mod, et, c(ka = 0.5, kel = 0.2),
                 inits = c(depot = 0, central = 3))
    expect_false(any(is.na(s$central)))
    expect_false(any(is.na(s$depot)))
  })

  test_that("state-dep alag: runtime refresh updates lag for multiple doses", {
    mod <- rxode2({
      alag(depot) <- central * 0.1
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - kel * central
    })
    et <- et() |>
      et(amt = 100, time = 0) |>
      et(amt = 100, time = 6) |>
      et(seq(0, 24, by = 1))
    # State changes between doses -> lag differs for second dose; no NA
    s <- rxSolve(mod, et, c(ka = 0.5, kel = 0.2),
                 inits = c(depot = 0, central = 5))
    expect_false(any(is.na(s$central)))
  })

  test_that("state-dep alag: dose fires at dosing_time + state_value (constant lag, numeric)", {
    # Use d/dt(state_val)=0 so lag = state_val stays constant -> analytically verifiable.
    # With ka=0.5 (mono-exp depot), depot(t) after a dose at t_fire = 100*exp(-0.5*(t-t_fire)).
    # dose1 raw_t=0, lag=state_val -> fires at t=state_val.
    # dose2 raw_t=3, lag=state_val -> fires at t=3+state_val.
    # At each obs time, depot = sum of contributions from fired doses.
    mod <- rxode2({
      alag(depot) <- state_val
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - kel * central
      d/dt(state_val) <- 0
    })
    params <- c(ka = 0.5, kel = 0.2)
    et2dose <- et(amt = 100, time = 0, cmt = "depot") |>
      et(amt = 100, time = 3, cmt = "depot") |>
      et(seq(0, 10, by = 1))

    for (sv in c(0, 1, 2)) {
      s <- rxSolve(mod, et2dose, params,
                   inits = c(depot = 0, central = 0, state_val = sv))
      t_fire1 <- sv          # dose1 fires at t=sv
      t_fire2 <- 3 + sv      # dose2 fires at t=3+sv
      for (tobs in seq(0, 10)) {
        expected <- 0
        if (tobs >= t_fire1) expected <- expected + 100 * exp(-0.5 * (tobs - t_fire1))
        if (tobs >= t_fire2) expected <- expected + 100 * exp(-0.5 * (tobs - t_fire2))
        got <- s$depot[s$time == tobs]
        expect_equal(got, expected, tolerance = 1e-3,
                     label = sprintf("state_val=%d, t=%d: depot", sv, tobs))
      }
    }
  })
})

rxTest({

  ode <- rxode2({
    param(THETA[1], THETA[2], THETA[3], THETA[4], ETA[1], ETA[2],
          ETA[3])
    cmt(depot)
    cmt(central)
    rx_expr_0 ~ ETA[3] + THETA[3]
    rx_expr_3 ~ exp(rx_expr_0)
    d/dt(depot) = -rx_expr_3 * depot
    rx_expr_1 ~ ETA[1] + THETA[1]
    rx_expr_2 ~ ETA[2] + THETA[2]
    rx_expr_6 ~ rx_expr_3 * depot
    rx_expr_8 ~ rx_expr_1 - (rx_expr_2)
    rx_expr_9 ~ exp(rx_expr_8)
    rx_expr_11 ~ rx_expr_9 * central
    d/dt(central) = rx_expr_6 - rx_expr_11
    d/dt(rx__sens_depot_BY_ETA_1___) = -rx_expr_3 * rx__sens_depot_BY_ETA_1___
    d/dt(rx__sens_central_BY_ETA_1___) = rx_expr_3 * rx__sens_depot_BY_ETA_1___ -
      rx_expr_11 - rx_expr_9 * rx__sens_central_BY_ETA_1___
    d/dt(rx__sens_depot_BY_ETA_2___) = -rx_expr_3 * rx__sens_depot_BY_ETA_2___
    d/dt(rx__sens_central_BY_ETA_2___) = rx_expr_3 * rx__sens_depot_BY_ETA_2___ +
      rx_expr_11 - rx_expr_9 * rx__sens_central_BY_ETA_2___
    rx_expr_12 ~ rx_expr_3 * rx__sens_depot_BY_ETA_3___
    d/dt(rx__sens_depot_BY_ETA_3___) = -rx_expr_3 * depot - rx_expr_12
    d/dt(rx__sens_central_BY_ETA_3___) = rx_expr_6 + rx_expr_12 -
      rx_expr_9 * rx__sens_central_BY_ETA_3___
    rx_yj_ ~ 2
    rx_lambda_ ~ 1
    rx_hi_ ~ 1
    rx_low_ ~ 0
    rx_expr_4 ~ exp(-(rx_expr_2))
    rx_expr_7 ~ rx_expr_4 * central
    rx_pred_ = rx_expr_7
    rx__sens_rx_pred__BY_ETA_1___ = rx_expr_4 * rx__sens_central_BY_ETA_1___
    rx__sens_rx_pred__BY_ETA_2___ = -rx_expr_4 * central + rx_expr_4 *
      rx__sens_central_BY_ETA_2___
    rx__sens_rx_pred__BY_ETA_3___ = rx_expr_4 * rx__sens_central_BY_ETA_3___
    rx_expr_10 ~ rx_expr_7 * THETA[4]
    rx_r_ = Rx_pow_di((rx_expr_10), 2)
    rx_expr_5 ~ 2 * rx_expr_4
    rx__sens_rx_r__BY_ETA_1___ = rx_expr_5 * rx__sens_central_BY_ETA_1___ *
      (rx_expr_10) * THETA[4]
    rx__sens_rx_r__BY_ETA_2___ = -2 * rx_expr_4 * central * (rx_expr_10) *
      THETA[4] + rx_expr_5 * rx__sens_central_BY_ETA_2___ *
      (rx_expr_10) * THETA[4]
    rx__sens_rx_r__BY_ETA_3___ = rx_expr_5 * rx__sens_central_BY_ETA_3___ *
      (rx_expr_10) * THETA[4]
    cmt(cp)
    dvid(3)
  })

  lin <- rxode2({
    param(THETA[1], THETA[2], THETA[3], THETA[4], ETA[1], ETA[2],
        ETA[3])
    rx_yj_ ~ 2
    rx_lambda_ ~ 1
    rx_hi_ ~ 1
    rx_low_ ~ 0
    rx_expr_0 ~ ETA[1] + THETA[1]
    rx_expr_1 ~ ETA[2] + THETA[2]
    rx_expr_2 ~ ETA[3] + THETA[3]
    rx_expr_3 ~ exp(rx_expr_0)
    rx_expr_4 ~ exp(rx_expr_1)
    rx_expr_5 ~ exp(rx_expr_2)
    rx_pred_ = linCmtB(rx__PTR__, t, 2, 1, 1, -1, -1, 1, rx_expr_3,
        rx_expr_4, 0, 0, 0, 0, rx_expr_5)
    rx__sens_rx_pred__BY_ETA_1___ = rx_expr_3 * linCmtB(rx__PTR__,
        t, 2, 1, 1, -2, 0, 1, rx_expr_3, rx_expr_4, 0, 0, 0,
        0, rx_expr_5)
    rx__sens_rx_pred__BY_ETA_2___ = rx_expr_4 * linCmtB(rx__PTR__,
        t, 2, 1, 1, -2, 1, 1, rx_expr_3, rx_expr_4, 0, 0, 0,
        0, rx_expr_5)
    rx__sens_rx_pred__BY_ETA_3___ = rx_expr_5 * linCmtB(rx__PTR__,
        t, 2, 1, 1, -2, 2, 1, rx_expr_3, rx_expr_4, 0, 0, 0,
        0, rx_expr_5)
    rx_r_ = Rx_pow_di((linCmtB(rx__PTR__, t, 2, 1, 1, -1, -1,
        1, rx_expr_3, rx_expr_4, 0, 0, 0, 0, rx_expr_5) * THETA[4]),
        2)
    rx__sens_rx_r__BY_ETA_1___ = 2 * rx_expr_3 * linCmtB(rx__PTR__,
        t, 2, 1, 1, -2, 0, 1, rx_expr_3, rx_expr_4, 0, 0, 0,
        0, rx_expr_5) * (linCmtB(rx__PTR__, t, 2, 1, 1, -1, -1,
        1, rx_expr_3, rx_expr_4, 0, 0, 0, 0, rx_expr_5) * THETA[4]) *
  THETA[4]
    rx__sens_rx_r__BY_ETA_2___ = 2 * rx_expr_4 * linCmtB(rx__PTR__,
        t, 2, 1, 1, -2, 1, 1, rx_expr_3, rx_expr_4, 0, 0, 0,
        0, rx_expr_5) * (linCmtB(rx__PTR__, t, 2, 1, 1, -1, -1,
        1, rx_expr_3, rx_expr_4, 0, 0, 0, 0, rx_expr_5) * THETA[4]) *
  THETA[4]
    rx__sens_rx_r__BY_ETA_3___ = 2 * rx_expr_5 * linCmtB(rx__PTR__,
        t, 2, 1, 1, -2, 2, 1, rx_expr_3, rx_expr_4, 0, 0, 0,
        0, rx_expr_5) * (linCmtB(rx__PTR__, t, 2, 1, 1, -1, -1,
        1, rx_expr_3, rx_expr_4, 0, 0, 0, 0, rx_expr_5) * THETA[4]) *
  THETA[4]
    cmt(rxLinCmt)
    dvid(3)
  })


  ev     <- et(amt=100) |> et(c(0.5,1,2,4,8,12,24))
  params <- data.frame("THETA[1]"=log(4),"THETA[2]"=log(70),"THETA[3]"=log(1),
                     "THETA[4]"=0.1,"ETA[1]"=0.1,"ETA[2]"=-0.1,"ETA[3]"=0.05,
                     check.names=FALSE)

  outOde    <- rxSolve(ode, params=params, events=ev)
  outLin    <- rxSolve(lin, params=params, events=ev)


  test_that("linCmtB and ode solve the same", {
    expect_equal(outOde$rx_pred_, outLin$rx_pred_, tolerance=1e-5)
    expect_equal(outOde$rx_r_, outLin$rx_r_, tolerance=1e-5)
  })

  sensCols <- grep("sens_rx_pred.*ETA", names(outOde), value=TRUE)
  for (col in sensCols) {
    test_that(paste0("linCmtB and ode solve the same for ", col), {
      expect_equal(outOde[[col]], outLin[[col]], tolerance=1e-5)
    })
  }

  test_that("linCmtB forward-AD solves are thread safe (parallel == serial)", {
    skip_on_cran()
    # Without >= 2 available threads the cores = 2 request cannot exercise the
    # parallel path, so the comparison would pass vacuously; skip instead.
    # rxode2 exposes no per-solve "cores actually used" telemetry, so gate on
    # thread availability (this model's linCmtB carries the forward-AD flag that
    # the cores = 2 request now honors -- see rxData.cpp thread decision).
    skip_if_not(rxode2::getRxThreads() >= 2L)
    .nsub <- 60
    set.seed(1)
    .params <- data.frame(
      id = seq_len(.nsub),
      "THETA[1]" = log(4), "THETA[2]" = log(70), "THETA[3]" = log(1),
      "THETA[4]" = 0.1,
      "ETA[1]" = rnorm(.nsub, 0, 0.3),
      "ETA[2]" = rnorm(.nsub, 0, 0.3),
      "ETA[3]" = rnorm(.nsub, 0, 0.3),
      check.names = FALSE)
    .ev <- et(amt = 100, ii = 12, until = 48) |>
      et(seq(0, 48, by = 1)) |>
      et(id = seq_len(.nsub))
    .s1 <- rxSolve(lin, params = .params, events = .ev, cores = 1,
                   returnType = "data.frame")
    .s2 <- rxSolve(lin, params = .params, events = .ev, cores = 2,
                   returnType = "data.frame")
    .cols <- grep("sens_rx_pred.*ETA|^rx_pred_$", names(.s1), value = TRUE)
    for (.col in .cols) {
      expect_identical(.s1[[.col]], .s2[[.col]])
    }
  })
})

rxTest({
  # Dual lhs/param values (Issue #135)

  test_that("Two defined variables", {
    mod1 <- rxode2({
      k <- k + 3
      km <- km + 4
    })

    expect_equal(mod1$lhs, c("k", "km"))
    expect_equal(mod1$params, c("k", "km"))

    expect_equal(as.vector(rxSolve(mod1, et(0), params = c(k = 3, km = 4), returnType = "matrix")), as.double(c(0, 6, 8)))
  })


  test_that("Two defined variables with ini", {
    mod1 <- rxode2({
      k <- 3
      k <- k + 3
      km <- 4
      km <- km + 4
    })
    expect_equal(mod1$lhs, c("k", "km"))
    expect_equal(mod1$params, c("k", "km"))
    expect_equal(rxInits(mod1), c(k = 3, km = 4))
    expect_equal(as.vector(rxSolve(mod1, et(0), returnType = "matrix")), as.double(c(0, 6, 8)))
  })

  test_that("lhs/params changes", {
    mod4 <- rxode2({
      j <- k + m
      k <- j + 3
    })
    expect_equal(mod4$lhs, c("j", "k"))
    expect_equal(mod4$params, c("k", "m"))
  })

  test_that("one variable/param", {
    mod1 <- rxode2({
      k <- k + 3
    })
    expect_equal(mod1$lhs, "k")
    expect_equal(mod1$params, "k")
  })

  test_that("Sc is not lhs", {
    tmp <- rxode2({
      Sc <- L^2 * ((g * e) / (g + e)) * (1 + (km * L / v))
      d / dt(L) <- ((1 / (3 * L^2))) * (((v / g) * Sc) - km * L^3)
    })
    expect_false(any(tmp$params == "Sc"))
  })


  test_that("Last item of last line still counts", {
    mod4 <- rxode2({
      j <- k
      k <- j + 3
    })
    expect_equal(mod4$lhs, c("j", "k"))
    expect_equal(mod4$params, "k")

    mod4 <- rxode2({
      j <- k + 4
      k <- j + 3
    })
    expect_equal(mod4$lhs, c("j", "k"))
    expect_equal(mod4$params, "k")
  })

  test_that("Make sure CL is correctly identified as hidden lhs not a dual", {
    mod2 <- rxode2({
      C2 <- centr / V2
      C3 ~ peri / V3
      CL ~ TCL * exp(eta.Cl)
      d / dt(depot) ~ -KA * depot
      d / dt(centr) ~ KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) ~ Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- 1000
      e1 <- err1
      e2 <- err2
      resp <- eff + e1
      pk <- C2 * exp(e2)
    })
    expect_false(any("CL" == mod2$params))
    expect_false(any("CL" == mod2$lhs))
  })

  test_that("newind variables not identified as dual parameters", {
    ode.1c <- rxode2({
      V <- 20
      Cl <- 1
      fc <- 1
      C2 <- center / V
      ni <- newind
      ni2 <- NEWIND
      d / dt(center) ~ -Cl * C2
      f(center) <- fc
    })

    expect_equal(ode.1c$params, c("V", "Cl", "fc"))
    expect_equal(ode.1c$lhs, c("C2", "ni", "ni2"))
  })

  test_that("suppressed assignments give correct variables", {
    mod1 <- rxode2({
      k ~ k + 3
      km ~ km + 4
      ret <- k + km
    })

    expect_equal(mod1$lhs, "ret")
    expect_equal(mod1$params, c("k", "km"))
    expect_equal(rxSolve(mod1, c(k = 1, km = 2), et(0))$ret, as.double(10))
  })

  test_that("a=NA gives correct variables", {
    mod1 <- rxode2("a=NA;\nb=2;\nc=a+b")
    expect_equal(mod1$params, "b")
    expect_equal(mod1$lhs, c("a", "c"))

    mod1 <- rxode2("a=2;\nb=NA;\nc=a+b")
    expect_equal(mod1$params, "a")
    expect_equal(mod1$lhs, c("b", "c"))

    mod1 <- rxode2("a~NA;\nb~2;\nc=a+b")
    expect_equal(mod1$params, character(0))
    expect_equal(mod1$lhs, "c")

    mod1 <- rxode2("a~2;\nb~NA;\nc=a+b")
    expect_equal(mod1$params, character(0))
    expect_equal(mod1$lhs, "c")
  })

  test_that("dual order lhs mixed with non dual-order gives right order", {

    skip_if_not_installed("nlmixr2data")
    m <- rxode2({
      param(THETA[1], THETA[2], THETA[3], THETA[4], Nominal)
      rx_yj_ ~ 162
      rx_lambda_ ~ 1
      rx_hi_ ~ 1
      rx_low_ ~ 0
      rx_expr_0 ~ exp(THETA[2])
      rx_expr_1 ~ exp(THETA[3])
      rx_expr_2 ~ exp(THETA[1])
      rx_pred_ = linCmtA(rx__PTR__, t, 2, 1, 1, -1, 1, rx_expr_0,
                         rx_expr_1, 0, 0, 0, 0, rx_expr_2)
      rx_r_ = Rx_pow_di(THETA[4], 2)
      tka = THETA[1]
      tcl = THETA[2]
      tv = THETA[3]
      add.sd = THETA[4]
      ka = rx_expr_2
      cl = rx_expr_0
      v = rx_expr_1
      Nominal = Nominal
      tad = tad()
      dosenum = dosenum()
      cmt(rxLinCmt)
      dvid(3)
    })

    expect_equal(m$lhs,
                 c("rx_pred_", "rx_r_", "tka", "tcl", "tv", "add.sd", "ka", "cl", "v", "Nominal", "tad", "dosenum"))

    theo_sd2 <- nlmixr2data::theo_sd
    theo_sd2$Nominal <- 300

    s <- rxSolve(m, theo_sd2, c(`THETA[1]` = 0.440455057067531, `THETA[2]` = 0.962169566880609,
                                `THETA[3]` = 3.49144561224211, `THETA[4]` = 1.39422305101919),
                 addCov=FALSE)

    expect_true(all(s$Nominal == 300))
    expect_false(any(s$rx_pred_ == 300))

  })
})

rxTest({

  m <- rxode2({
    v  <- exp(tv+eta.v)
    cl <- exp(tcl+eta.cl)
    ka <- exp(tka+eta.ka)
    d/dt(depot) <- -ka*depot
    d/dt(centr) <- ka*depot - cl/v*centr
    cp <- centr/v
    if (is.na(C_max)) {
      C_max <- 0
      C_min <- Inf
      T_max <- 0
      T_min <- Inf
    }
    if (cp > C_max) {
      C_max <- cp
      T_max <- time
      # Needs to be after C_max is acheived
      C_min <- Inf
      T_min <- Inf
    } else if (cp < C_min || C_min == 0) {
      C_min <- cp
      T_min <- time
    }
  })

  test_that("sticky vars work", {

    ev <- et(amt=100) |>
      et(seq(0, 24)) |>
      et(id=1:5)

    s <- rxSolve(m, ev, param=c(tv=log(10), tcl=log(1), tka=log(1)),
                 omega=lotri::lotri({
                   eta.v ~ 0.1
                   eta.cl ~ 0.1
                   eta.ka ~ 0.01
                 }))

    for (i in unique(s$id)) {
      si <- s[s$id == i, ]
      cmax <- max(si$cp)
      w <- which(si$cp == cmax)
      tmax <- si$time[w[length(w)]]

      expect_equal(cmax, si$C_max[nrow(si)])
      expect_equal(tmax, si$T_max[nrow(si)])
      expect_equal(si$T_min[nrow(si)], si$time[nrow(si)])
      expect_equal(si$C_min[nrow(si)], si$cp[nrow(si)])
    }

  })

  test_that("sticky vars do not accumulate across separate rxSolve calls", {

    # First solve: large dose -- C_max ~100 (amt=1000, V=10)
    ev_large <- et(amt=1000) |> et(seq(0, 24))
    s1 <- rxSolve(m, ev_large,
                  param=c(tv=log(10), tcl=log(1), tka=log(1),
                          eta.v=0, eta.cl=0, eta.ka=0))

    # Second solve: small dose -- C_max ~1 (amt=10)
    ev_small <- et(amt=10) |> et(seq(0, 24))
    s2 <- rxSolve(m, ev_small,
                  param=c(tv=log(10), tcl=log(1), tka=log(1),
                          eta.v=0, eta.cl=0, eta.ka=0))

    # If thread-local lhs bleeds across calls, s2$C_max would start at ~100
    # (from s1) instead of NA, producing wrong final C_max.
    expect_equal(s2$C_max[length(s2$C_max)], max(s2$cp))
    # Sanity: the two C_max values differ by ~100x (linear PK, 100x dose ratio)
    expect_true(s1$C_max[length(s1$C_max)] > s2$C_max[length(s2$C_max)] * 50)

  })

})

rxTest({
  test_that("rxode2 threading doesn't disturb some solves; Issue RxODE#430", {
    skip_if_not(file.exists("test-issue-430.rds"))

    model <- rxode2({
      KA <- THETA_KA * exp(ETA_KA)
      CL <- THETA_CL * exp(ETA_CL)
      V2 <- THETA_V2 * exp(ETA_V2)
      V3 <- THETA_V3 * exp(ETA_V3)
      Q <- THETA_Q * exp(ETA_Q)
      S2 <- V2
      ALAG1 <- THETA_ALAG1 * exp(ETA_ALAG1 + IOV_ALAG1)
      d / dt(A_DEPOT) <- -KA * A_DEPOT
      d / dt(A_CENTRAL) <- KA * A_DEPOT + Q * A_PERIPHERAL / V3 + (-CL / V2 - Q / V2) * A_CENTRAL
      d / dt(A_PERIPHERAL) <- -Q * A_PERIPHERAL / V3 + Q * A_CENTRAL / V2
      d / dt(A_OUTPUT) <- CL * A_CENTRAL / V2
      lag(A_DEPOT) <- ALAG1
      CP <- A_CENTRAL / S2
    })

    theta <- c(THETA_KA = 1, THETA_CL = 5, THETA_V2 = 80, THETA_V3 = 20, THETA_Q = 4, THETA_ALAG1 = 5)

    dataset <- readRDS("test-issue-430.rds") # This dataset contains 3 boluses given at time 0, 24 & 48

    r2 <- rxode2::rxSolve(object = model, params = theta, omega = NULL, sigma = NULL, events = dataset, cores = 2, returnType = "data.frame")
    r1 <- rxode2::rxSolve(object = model, params = theta, omega = NULL, sigma = NULL, events = dataset, cores = 1, returnType = "data.frame")

    expect_equal(r1, r2)
  })
})

rxTest({
  # The extra-dosing pools (ignoredDoses/pendingDoses/extraDose*) are indexed by
  # _rxTid(), which is bounded by op->cores, but they were only ever sized once
  # at load from omp_get_max_threads().  omp_get_max_threads() honors
  # OMP_NUM_THREADS, while `rxSolve(cores=)` overrides it through the
  # num_threads clause -- so an environment that sets OMP_NUM_THREADS below the
  # requested cores left every thread past that count writing off the end of
  # those arrays, corrupting the heap.  CRAN check machines set
  # OMP_NUM_THREADS=2, which is how this reached a reverse dependency.
  #
  # The pools are sized at load, so this has to run in a fresh process with the
  # environment variable already set; it cannot be provoked in-process.
  test_that("solving with more cores than OMP_NUM_THREADS does not corrupt the heap", {
    skip_on_cran()
    if (!isTRUE(rxode2::rxCores() > 2L)) {
      skip("needs more than 2 available cores")
    }
    # resolved HERE, not in the child: inside the child OMP_NUM_THREADS caps
    # rxCores() at 2, which is exactly the mismatch we need to avoid asking for
    .cores <- as.integer(rxode2::rxCores())
    .script <- tempfile(fileext = ".R")
    on.exit(unlink(.script), add = TRUE)
    writeLines(c(
      sprintf(".libPaths(%s)", paste(deparse(.libPaths()), collapse = "")),
      "suppressMessages(library(rxode2))",
      "m <- rxode2({",
      "  ka <- exp(lka + eta.ka)",
      "  cl <- exp(lcl + eta.cl)",
      "  v <- exp(lv)",
      "  d/dt(depot) <- -ka*depot",
      "  d/dt(cent) <- ka*depot - cl/v*cent",
      "  Cc <- cent/v",
      "})",
      "ev <- et(et(amt=100, ii=24, addl=6), seq(0, 168, by=2))",
      "om <- lotri::lotri(eta.ka ~ 0.09, eta.cl ~ 0.09)",
      "th <- c(lka=0.5, lcl=1, lv=3)",
      "tm <- diag(3)*0.01",
      "dimnames(tm) <- list(names(th), names(th))",
      "set.seed(1)",
      "for (i in 1:40) {",
      "  s <- rxSolve(m, ev, params=th, omega=om, thetaMat=tm, nStud=4,",
      sprintf("               nSub=25, simVariability=TRUE, cores=%dL)", .cores),
      "  d <- as.data.frame(s)",
      # allocate hard right after each solve -- a corrupted heap dies here
      "  for (k in 1:20) invisible(sum(rnorm(nrow(d)))) ",
      "}",
      "cat('RXODE2-OK\\n')"
    ), .script)
    .out <- suppressWarnings(
      system2(file.path(R.home("bin"), "Rscript"),
              args = c("--vanilla", shQuote(.script)),
              env = c("OMP_NUM_THREADS=2", "MKL_NUM_THREADS=2",
                      "NOT_CRAN=true"),
              stdout = TRUE, stderr = TRUE))
    .status <- attr(.out, "status")
    # a corrupted heap shows up as a non-zero exit plus glibc's own complaint
    # ("free(): invalid pointer" / "malloc(): ...") or an outright segfault
    expect_true(any(grepl("RXODE2-OK", .out, fixed = TRUE)),
                info = paste(utils::tail(.out, 15), collapse = "\n"))
    expect_false(any(grepl("free\\(\\)|malloc\\(\\)|corrupt|segfault", .out)),
                 info = paste(utils::tail(.out, 15), collapse = "\n"))
    expect_true(is.null(.status) || identical(as.integer(.status), 0L))
  })
})

rxTest({
  test_that("RxODE#352", {
    rx <- rxode2({
      cmt(A1)
      d / dt(A1) <- -exp(eta.cl + tcl - (eta.v + tv)) * A1
      rx_yj_ ~ 2
      rx_lambda_ ~ 1
      rx_hi_ ~ 1
      rx_low_ ~ 0
      rx_pred_ ~ exp(-(eta.v + tv)) * A1
      rx_r_ ~ add.err
      cmt(cp)
      ipred <- rxTBSi(rx_pred_, rx_lambda_, rx_yj_, rx_low_, rx_hi_)
      sim <- rxTBSi(rx_pred_ + rx_r_, rx_lambda_, rx_yj_, rx_low_, rx_hi_)
    })

    simdata <-
      et(
        amt = 25,
        time = 0,
        evid = 1
      ) |>
      # Observations
      et(0:24)

    omega <- lotri(eta.cl ~ 0, eta.v ~ 0)

    sigma <- lotri(add.err ~ 0)

    params <- c(tcl = -4.98894153987034, tv = 0.341202932945698)

    suppressMessages(expect_error(
      rxSolve(rx, params = params, events = simdata, omega = omega, sigma = sigma),
      NA
    ))
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("mdv means EVID=2 when amt=0", {
      skip_if_not_installed("nlmixr2data")
      theoSd <- nlmixr2data::theo_sd

      d <- theoSd[, names(theoSd) != "EVID"]
      d$MDV <- ifelse(d$AMT == 0, 0, 1)
      d <- d[, names(d) != "WT"]

      d2 <- expand.grid(ID = 1:12, TIME = c(0.125, 25), DV = 0, AMT = 0, CMT = 2, MDV = 1)

      d <- rbind(d, d2)

      mod <- rxode2({
        tka <- 1
        tcl <- 2
        tv <- 3
        ka <- exp(tka)
        cl <- exp(tcl)
        v <- exp(tv)
        cp <- linCmt()
      })

      tmp <- rxSolve(mod, d)

      expect_true(any(tmp$time == 0.125))
      expect_true(any(tmp$time == 25))
    })
  }
})

rxTest({
  test_that("Test a matrix that needs nearPD", {

    rx1 <- RxODE({
      cl <- tcl*(1+crcl.cl*(CLCR-65)) * exp(eta.v)
      v <- tv * WT * exp(eta.v)
      ka <- tka * exp(eta.ka)
      ipred <- linCmt()
      obs <- ipred * (1 + prop.sd) + add.sd
    })

    theta <- c(tcl=2.63E+01, tv=1.35E+00, tka=4.20E+00, tlag=2.08E-01,
               prop.sd=2.05E-01, add.sd=1.06E-02, crcl.cl=7.17E-03,
               ## Note that since we are using the separation strategy the ETA variances are here too
               eta.cl=7.30E-02,  eta.v=3.80E-02, eta.ka=1.91E+00)

    thetaMat <- lotri(
      tcl + tv + tka + tlag + prop.sd + add.sd + crcl.cl + eta.cl + eta.v + eta.ka ~
        c(7.95E-01,
          ##      2.05E-02, 1.92E-03, --> Here I am assuming that the tv was fixed during estimation
          ##    so that nonmem cov output is a low triangular matrix with a zero row
          0, 0,
          7.22E-02, -8.30E-03, 6.55E-01,
          -3.45E-03, -6.42E-05, 3.22E-03, 2.47E-04,
          8.71E-04, 2.53E-04, -4.71E-03, -5.79E-05, 5.04E-04,
          6.30E-04, -3.17E-06, -6.52E-04, -1.53E-05, -3.14E-05, 1.34E-05,
          -3.30E-04, 5.46E-06, -3.15E-04, 2.46E-06, 3.15E-06, -1.58E-06, 2.88E-06,
          -1.29E-03, -7.97E-05, 1.68E-03, -2.75E-05, -8.26E-05, 1.13E-05, -1.66E-06, 1.58E-04,
          -1.23E-03, -1.27E-05, -1.33E-03, -1.47E-05, -1.03E-04, 1.02E-05, 1.67E-06, 6.68E-05, 1.56E-04,
          7.69E-02, -7.23E-03, 3.74E-01, 1.79E-03, -2.85E-03, 1.18E-05, -2.54E-04, 1.61E-03, -9.03E-04, 3.12E-01))

    thetaMat1 <- thetaMat

    # Quick fix idea: add a very small (epsilon) diagonal term for thetaMat
    # so that there is "very small" uncertainty around tv
    thetaMat1[2, 2] <- 1e-06

    evw <- et(amount.units="mg", time.units="hours") |>
      et(amt=100) |>
      ## For this problem we will simulate with sampling windows
      et(list(c(0, 0.5),
              c(0.5, 1),
              c(1, 3),
              c(3, 6),
              c(6, 12))) |>
      et(id=1:1000)

    skip_on_os("windows")

    expect_error(rxSolve(rx1, theta, evw,  nSub=100, nStud=10,
                    thetaMat=thetaMat1,
                    ## Match boundaries of problem
                    thetaLower=0,
                    sigma=c("prop.sd", "add.sd"), ## Sigmas are standard deviations
                    sigmaXform="identity", # default sigma xform="identity"
                    omega=c("eta.cl", "eta.v", "eta.ka"), ## etas are variances
                    omegaXform="variance", # default omega xform="variance"
                    iCov=data.frame(WT=rnorm(1000, 70, 15), CLCR=rnorm(1000, 65, 25)),
                    dfSub=74, dfObs=476),
                 NA)
  })
})

rxTest({
  # NA tests

  test_that("can assign NA", {
    mod1 <- rxode2("a=NA;\nb=2;\nc=a+b")
    ret <- rxSolve(mod1, et(0), returnType = "data.frame")
    expect_true(all(is.na(ret$a)))
    expect_true(all(is.na(ret$b)))
  })

  test_that("logic NA", {
    mod1 <- rxode2("a=NA;\nif(is.na(a)){a=3}b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")

    expect_equal(ret$a, as.double(3))
    expect_equal(ret$c, 3 + 2 + exp(3))

    mod1 <- rxode2("a=NA;\nif(!is.na(a)){a=4} else { a=3 }; b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")
    expect_equal(ret$c, 3 + 2 + exp(3))
  })

  # Inf tests

  test_that("can assign Inf", {
    mod1 <- rxode2("a=Inf;\nb=2;\nc=a+b")
    ret <- rxSolve(mod1, et(0), returnType = "data.frame")
    expect_true(all(is.infinite(ret$a)))
    expect_true(all(ret$a > 0))
    expect_true(all(ret$b > 0))
  })

  test_that("can assign -Inf", {
    mod1 <- rxode2("a=-Inf;\nb=2;\nc=a+b")
    ret <- rxSolve(mod1, et(0), returnType = "data.frame")
    expect_true(all(is.infinite(ret$a)))
    expect_true(all(ret$a < 0))
    expect_true(all(ret$b < 0))
  })

  test_that("logic Inf", {
    mod1 <- rxode2("a=Inf;\nif(is.infinite(a)){a=3}b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")

    expect_equal(ret$a, as.double(3))
    expect_equal(ret$c, 3 + 2 + exp(3))

    mod1 <- rxode2("a=Inf;\nif(!is.finite(a)){a=3}b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")

    expect_equal(ret$a, as.double(3))
    expect_equal(ret$c, 3 + 2 + exp(3))

    mod1 <- rxode2("a=Inf;\nif(!is.infinite(a)){a=4} else { a=3 }; b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")
    expect_equal(ret$c, 3 + 2 + exp(3))

    mod1 <- rxode2("a=Inf;\nif(is.finite(a)){a=4} else { a=3 }; b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")
    expect_equal(ret$c, 3 + 2 + exp(3))
  })

  test_that("logic -Inf", {
    mod1 <- rxode2("a=-Inf;\nif(is.infinite(a)){a=3}b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")

    expect_equal(ret$a, as.double(3))
    expect_equal(ret$c, 3 + 2 + exp(3))

    mod1 <- rxode2("a=-Inf;\nif(!is.finite(a)){a=3}b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")

    expect_equal(ret$a, as.double(3))
    expect_equal(ret$c, 3 + 2 + exp(3))

    mod1 <- rxode2("a=-Inf;\nif(!is.infinite(a)){a=4} else { a=3 }; b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")
    expect_equal(ret$c, 3 + 2 + exp(3))

    mod1 <- rxode2("a=-Inf;\nif(is.finite(a)){a=4} else { a=3 }; b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")
    expect_equal(ret$c, 3 + 2 + exp(3))
  })

  # NaN tests

  test_that("can assign NaN", {
    mod1 <- rxode2("a=NaN;\nb=2;\nc=a+b")
    ret <- rxSolve(mod1, et(0), returnType = "data.frame")
    expect_true(all(is.nan(ret$a)))
    expect_true(all(is.nan(ret$b)))
  })

  test_that("logic NaN", {
    mod1 <- rxode2("a=NaN;\nif(is.nan(a)){a=3}b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")

    expect_equal(ret$a, as.double(3))
    expect_equal(ret$c, 3 + 2 + exp(3))

    mod1 <- rxode2("a=NaN;\nif(!is.nan(a)){a=4} else { a=3 }; b=2;\nc=a+b+exp(3)")

    ret <- rxSolve(mod1, et(0), returnType = "data.frame")
    expect_equal(ret$c, 3 + 2 + exp(3))
  })
})

rxTest({
  # This test is for issue #999
  # 3-compartment TMDD model (matches the pharmacokinetic simulation in this project)
  model <- rxode2({
    Kel     <- CL / Vcentral
    K12     <- Qcp1 / Vcentral
    K21     <- Qcp1 / Vperiph1
    K13     <- Qcp2 / Vcentral
    K31     <- Qcp2 / Vperiph2
    KelTMDD <- CLTMDD / Vcentral * CENT / Vcentral / (C50TMDD + CENT / Vcentral)
    d/dt(GUT)  <- -Ka * GUT
    d/dt(CENT) <- Ka * GUT - (Kel + KelTMDD + K12 + K13) * CENT + K21 * P1 + K31 * P2
    d/dt(P1)   <- K12 * CENT - K21 * P1
    d/dt(P2)   <- K13 * CENT - K31 * P2
    CP_noerror <- CENT / Vcentral
  })

  make_dataset <- function(n_subjects) {
    obs_times <- c(0, 0.5, 1, 2, 3, 4, 8, 12, 16, 24, 36, 48)

    doses <- data.frame(
      id   = seq_len(n_subjects),
      time = 0,
      evid = 1,
      amt  = 1,
      cmt  = "CENT",  # IV bolus
      rate = 0
    )

    obs <- expand.grid(id = seq_len(n_subjects), time = obs_times)
    obs$evid <- 0; obs$amt <- 0; obs$cmt <- "CENT"; obs$rate <- 0

    events <- rbind(doses, obs)
    events <- events[order(events$id, events$time, -events$evid), ]

    # Vary parameters across subjects to match simulation diversity.
    # CLTMDD is set to a non-zero fraction of CL to activate the nonlinear
    # Michaelis-Menten TMDD elimination term, which is required to trigger
    # the segfault. Setting CLTMDD = 0 does NOT reproduce the crash.
    set.seed(42)
    cl_vals <- sample(c(0.25, 1, 4), n_subjects, replace = TRUE)
    params <- data.frame(
      id       = seq_len(n_subjects),
      Ka       = 0,
      CL       = cl_vals,
      CLTMDD   = cl_vals * sample(c(0.25, 1, 4), n_subjects, replace = TRUE),
      C50TMDD  = 1,
      Vcentral = 1,
      Vperiph1 = sample(c(0.25, 1, 4), n_subjects, replace = TRUE),
      Vperiph2 = sample(c(0.25, 1, 4), n_subjects, replace = TRUE),
      Qcp1     = sample(c(0.25, 1, 4), n_subjects, replace = TRUE),
      Qcp2     = sample(c(0.25, 1, 4), n_subjects, replace = TRUE)
    )

    list(events = events, params = params)
  }

  test_that("rxSolve handles large datasets without crashing", {
    skip_on_os("windows")
    skip_on_os("mac")
    for (n_sub in c(4000, 8000, 12000, 16000, 20000, 26208)) {
      dat <- make_dataset(n_sub)
      tryCatch({
        result <- expect_error(rxSolve(model, params = dat$params, events = dat$events,
                             atol = 1e-50, rtol = 1e-8), NA)
        rm(result); gc()
      }, error = function(e) {
        cat(n_sub, "subjects: ERROR -", conditionMessage(e), "\n")
      })
    }
  })

  # Test that nSize = nsim * nsub is computed correctly for the VPC simulation
  # path (nsim > 1).  The nSize overflow protection (int64_t check) guards
  # against integer overflow when nsim * nsub would exceed INT_MAX; at those
  # scales allocation itself would fail, so the overflow check fires first and
  # rxSolve() throws an informative error rather than segfaulting.
  test_that("rxSolve nSize is correct for multi-simulation (VPC) path", {

    m2 <- rxode2({
      CL <- TVCL * exp(eta.CL)
      C2 <- centr / V2
      d/dt(centr) <- -CL * C2
    })

    omega <- matrix(0.04, 1, 1, dimnames = list("eta.CL", "eta.CL"))

    ev2 <- et(amt = 100, addl = 4, ii = 24) |>
      et(0:120) %>%
      et(id=1:10)

    # nSub=10, nStud=5 => nSize = 5*10 = 50; verify no crash and correct dims
    result <- expect_error(
      rxSolve(m2, params = c(TVCL = 1, V2 = 10), events = ev2,
              omega = omega, nStud = 5, cores = 1),
      NA
    )

    expect_true(nrow(result) > 0)
    # sim.id should range from 1 to nStud=5
    expect_equal(sort(unique(result$sim.id)), 1:5)
    # id should range from 1 to nSub=10
    expect_equal(length(unique(result$id)), 10)

    ev2 <- et(amt = 100, addl = 4, ii = 24) |>
      et(0:120)

    result <- expect_error(
      rxSolve(m2, params = c(TVCL = 1, V2 = 10), events = ev2,
              omega = omega, nSub=10, nStud = 5, cores = 1),
      NA
    )

    expect_true(nrow(result) > 0)
    # sim.id should range from 1 to nSub*nStud=50 (nSub=10 subjects x nStud=5 studies)
    expect_equal(sort(unique(result$sim.id)), 1:50)

    rm(result); gc()
  })

  # Test that the rxSimThetaOmega overflow guard provides an informative error
  # rather than a segfault when nSub * nStud > INT_MAX.
  # 46342 * 46342 = 2,147,580,964 > 2,147,483,647 (INT_MAX).
  # Before the unsigned-type fix, this segfaulted via NumericVector(negative_size).
  # After the fix, an informative error is thrown before allocation.
  # Test that the nr (output row count) overflow guard provides an informative
  # error rather than a crash when nobs * nStud > INT_MAX.
  # nobs = 46,342 timepoints for 1 subject; nStud = 46,342 -> nr = 2,147,580,964 > INT_MAX.
  # The overflow guard fires before any ODE solving (fast, lightweight test).
  # Before the fix: "negative length vectors are not allowed".
  # After the fix: informative "too large" error.
  test_that("rxSolve nr overflow guard gives informative error not segfault", {
    m_nr <- rxode2({ d/dt(x) <- -k * x })
    ev_nr <- et(0:46341)
    omega_nr <- matrix(0.04, 1, 1, dimnames = list("k", "k"))
    expect_error(
      rxSolve(m_nr, params = c(k = 0.1), events = ev_nr,
              omega = omega_nr, nStud = 46342L, cores = 1),
      "too large"
    )
  })

  test_that("rxSimThetaOmega overflow guard gives informative error not segfault", {
    m3 <- rxode2({
      CL <- TVCL * exp(eta.CL)
      C2 <- centr / V2
      d/dt(centr) <- -CL * C2
    })
    omega <- matrix(0.04, 1, 1, dimnames = list("eta.CL", "eta.CL"))
    ev3 <- et(amt = 100, addl = 0) |> et(0:10)
    expect_error(
      rxSolve(m3, params = c(TVCL = 1, V2 = 10), events = ev3,
              omega = omega, nSub = 46342L, nStud = 46342L, cores = 1),
      "too large"
    )
  })
})

rxTest({
  test_that("Zero variances; RxODE#299", {

    mod <- rxode2({
      eff(0) <- 1
      C2 <- centr / V2
      C3 <- peri / V3
      CL <- TCl * exp(eta.Cl) ## This is coded as a variable in the model
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d/dt(peri) <- Q * C2 - Q * C3
      d/dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      e <- eff + eff.err
      cp <- centr * (1 + cp.err)
    })

    ev <- et(amount.units = "mg", time.units = "hours") |>
      et(amt = 10000, cmt = "centr") |>
      et(0, 48, length.out = 100)

    ## Variability
    theta <- c(
      KA = 2.94E-01, TCl = 1.86E+01, V2 = 4.02E+01, # central
      Q = 1.05E+01, V3 = 2.97E+02, # peripheral
      Kin = 1, Kout = 1, EC50 = 200
    ) # effects

    omega <- lotri(eta.Cl ~ 0.4^2)
    sigma <- lotri(eff.err ~ 0.1, cp.err ~ 0.1)
    tmp <- matrix(rnorm(8^2), 8, 8)
    tMat <- tcrossprod(tmp, tmp) / (8^2)
    dimnames(tMat) <- list(NULL, names(theta))

    sim <- rxSolve(mod, theta, ev,
                   omega = omega, nSub = 100, sigma = sigma, thetaMat = tMat, nStud = 10,
                   dfSub = 10, dfObs = 100
                   )


    omega <- lotri(eta.Cl ~ 0)
    sigma <- lotri(eff.err ~ 0, cp.err ~ 0)

    tMat <- matrix(0, 8, 8)
    dimnames(tMat) <- list(NULL, names(theta))

    suppressMessages(expect_error(
      x <-
        rxSolve(
          mod, theta, ev,
          omega = omega, nSub = 100, sigma = sigma, thetaMat = tMat, nStud = 10,
          dfSub = 10, dfObs = 100
        ),
      NA
    ))

    expect_true(isNullZero(x$thetaMat))
    expect_true(isNullZero(x$omegaList))
    expect_true(isNullZero(x$sigmaList))
    expect_true(isNullZero(NULL))
  })
})

rxTest({
  test_that("CMT translation matches input", {
    filename <- test_path("433.rds")
    skip_if_not(file.exists(filename))

    lst <- readRDS(filename)

    rx <- rxode2({
      cmt(parent)
      cmt(m1)
      parent(0) <- ETA[1] + THETA[1]
      d / dt(parent) <- -exp(ETA[2] + THETA[2]) * parent
      d / dt(m1) <- -exp(ETA[3] + THETA[3]) * m1 + 1 * exp(ETA[2] +
                                                             THETA[2]) * parent / (1 + exp(-(ETA[4] + THETA[4])))
      rx_yj_ ~ 2 * (1 - (CMT == 2)) * (CMT == 1) +
        2 * (CMT == 2)
      rx_lambda_ ~ 1 * (1 - (CMT == 2)) * (CMT == 1) + 1 * (CMT == 2)
      rx_hi_ ~ 1 * (1 - (CMT == 2)) * (CMT == 1) + 1 * (CMT == 2)
      rx_low_ ~ 0
      rx_pred_ <- (m1 * (CMT == 2) + parent * (1 - (CMT == 2)) *
                     (CMT == 1)) * (CMT == 2) + (m1 * (CMT == 2) +
                                                   parent * (1 - (CMT == 2)) * (CMT == 1)) * (1 - (CMT == 2)) * (CMT == 1)
      rx_r_ <- (Rx_pow_di(((m1 * (CMT == 2) + parent * (1 - (CMT == 2)) * (CMT == 1)) * (CMT == 2) + (m1 * (CMT == 2) + parent *
                                                                                                        (1 - (CMT == 2)) * (CMT == 1)) * (1 - (CMT == 2)) * (CMT == 1)), 2) * Rx_pow_di(THETA[8], 2) +
                  Rx_pow_di(THETA[7], 2)) * (CMT == 2) + (Rx_pow_di(((m1 * (CMT == 2) + parent * (1 - (CMT == 2)) * (CMT == 1)) * (CMT == 1)), 2) *
                                                            Rx_pow_di(THETA[6], 2) + Rx_pow_di(THETA[5], 2)) * (1 - (CMT == 2)) * (CMT == 1)
      parent_0 <- THETA[1]
      log_k_parent <- THETA[2]
      log_k_m1 <- THETA[3]
      f_parent_qlogis <- THETA[4]
      sigma_low_parent <- THETA[5]
      rsd_high_parent <- THETA[6]
      sigma_low_m1 <- THETA[7]
      rsd_high_m1 <- THETA[8]
      eta.parent_0 <- ETA[1]
      eta.log_k_parent <- ETA[2]
      eta.log_k_m1 <- ETA[3]
      eta.f_parent_qlogis <- ETA[4]
      parent_0_model <- ETA[1] + THETA[1]
      k_parent <- exp(ETA[2] + THETA[2])
      k_m1 <- exp(ETA[3] + THETA[3])
      f_parent_to_m1 <- 1 / (1 + exp(-(ETA[4] + THETA[4])))
      tad <- tad()
      dosenum <- dosenum()
    })

    lst <- c(list(rx), lst)

    s <- do.call(rxSolve, lst)

    expect_equal(
      subset(lst[[3]], time == 0 & CMT == 2, "dv"),
      subset(s, time == 0 & CMT == 2, "dv")
    )
  })
})

rxTest({
  test_that("issue RxODE#229", {
    model <- rxode2({
      cmt(SC_AKR)
      cmt(Cent_AKR)
      rx_expr_0 ~ exp(THETA[1])
      d / dt(SC_AKR) <- -rx_expr_0 * SC_AKR
      rx_expr_2 ~ THETA[2] - THETA[3]
      rx_expr_4 ~ exp(rx_expr_2)
      d / dt(Cent_AKR) <- rx_expr_0 * SC_AKR - rx_expr_4 * Cent_AKR
      f(SC_AKR) <- THETA[4]
      rx_yj_ ~ 2
      rx_lambda_ ~ 1
      rx_hi_ ~ 1
      rx_low_ ~ 0
      rx_expr_1 ~ exp(-THETA[3])
      rx_expr_3 ~ 1000 * rx_expr_1
      rx_expr_5 ~ rx_expr_3 * Cent_AKR
      rx_pred_ <- rx_expr_5
      rx_r_ <- 1e+06 * Rx_pow_di((rx_expr_1 * Cent_AKR), 2) * Rx_pow_di(
        THETA[5],
        2
      ) + Rx_pow_di(THETA[6], 2)
      lKa <- THETA[1]
      lCL <- THETA[2]
      lVc <- THETA[3]
      F_SC <- THETA[4]
      prop_err <- THETA[5]
      add_err <- THETA[6]
      Ka <- rx_expr_0
      CL <- exp(THETA[2])
      Vc <- exp(THETA[3])
      kel <- rx_expr_4
      Cp_AKR <- rx_expr_5
      cmt(Cp_AKR)
      tad <- tad()
      dosenum <- dosenum()
    })
    
    parm <- setNames(
      c(-3.54456675092961, -2.30258509299405, 1.09861228866811, 0.3, 0.2, 10),
      paste0("THETA[", 1:6, "]")
    )
    
    d_mod <- readRDS(test_path("test-issue-299.rds"))
    
    s <- suppressWarnings(rxSolve(model, parm, d_mod))
    
    expect_true(all(diff(order(s$id, s$time)) == 1))
    
    d2 <- d_mod[d_mod$ID == 43, ]
    
    s <- suppressWarnings(rxSolve(model, parm, d2, addDosing = NA))
  })
})

#' Determine if Rstudio completion 2025+ is running
#'
#' @return boolean, is Rstudio completion 2025+ running
#' @export
#' @keywords internal
#' @author Matthew L. Fidler
#' @examples
#' # Determines if rstudio is running completion
#' .rstudioComplete()
.rstudioComplete <- function() {
  ## sys.calls() is expensive; autocomplete can only fire in interactive sessions
  if (!interactive()) return(FALSE)
  .sc <- try(sys.calls(), silent=TRUE)
  if (inherits(.sc, "try-error")) {
    return(FALSE)
  }
  .sc <- try(.sc[[1]], silent=TRUE)
  if (inherits(.sc, "try-error")) {
    return(FALSE)
   }
  .sc <- try(.sc[[1]], silent=TRUE)
  if (inherits(.sc, "try-error")) {
    return(FALSE)
  }
  identical(quote(`.rs.rpc.get_completions`), .sc)
}

.rxModelVarsLast <- NULL

#' Internal translation to get model variables list
#'
#'
#' @param model Model (either file name or string)
#' @param linear boolean indicating if linear compartment model should
#'   be generated from `linCmt()` (default FALSE)
#' @param linCmtSens Linear compartment model sensitivity type
#' @param verbose is a boolean indicating the type of model detected
#'   with `linCmt()` parsing
#' @param code is a file name where the c code is written to (for
#'   testing purposes mostly, it needs `rxode2` to do anything fancy)
#' @param envir is the environment to look for R user functions
#'   (defaults to parent environment)
#' @return A rxModelVars object that has the model variables of a
#'   rxode2 syntax expression
#' @export
#' @importFrom Rcpp evalCpp
#' @importFrom dparser dparse
#' @importFrom utils capture.output
#' @importFrom stats setNames
#' @importFrom methods is
#' @importFrom utils assignInMyNamespace
#' @examples
#' rxode2parse("a=3")
rxode2parse <- function(model, linear=FALSE, linCmtSens = c("linCmtA", "linCmtB"), verbose=FALSE,
                        code=NULL, envir=parent.frame()) {
  rxParseSuppressMsg()
  .udfEnvSet(envir)
  checkmate::assertCharacter(model, len=1, any.missing=FALSE)
  if (file.exists(model)) {
    .isStr <- 0L
  } else {
    .isStr <- 1L
  }
  modelPrefix <- ""
  fullPrint <- FALSE
  md5 <-   digest::digest(model)
  .ret <- .Call(
    `_rxode2_trans`, model, modelPrefix, md5, .isStr,
    as.integer(crayon::has_color()),
    "", .rxSupportedFuns(),
    fullPrint
  )
  if (linear && .isLinCmt()) {
    .vars <- c(.ret$params, .ret$lhs, .ret$slhs)
    .ret <- .Call(`_rxode2_linCmtGen`,length(.ret$state), .vars,
                  setNames(
                    c(
                      "linCmtA" = 1L, "linCmtB" = 2L
                    )[match.arg(linCmtSens)],
                    NULL
                  ), verbose)
    md5 <- digest::digest(.ret)
    .ret <- .Call(`_rxode2_trans`, .ret, modelPrefix, md5, .isStr,
                  as.integer(crayon::has_color()),
                  "", .rxSupportedFuns(),
                  fullPrint)
  }
  md5 <- c(file_md5 = md5, parsed_md5 = digest::digest(c(
    .ret$model,
    .ret$ini,
    .ret$state,
    .ret$params,
    .ret$lhs,
    .ret$alag
  )))
  .ret$timeId <- -1L
  .ret$md5 <- md5
  if (.isStr == 1L) {
    ## Now update trans.
    .prefix <- paste0("rx_", md5["parsed_md5"], "_", .Platform$r_arch, "_")
    .libName <- substr(.prefix, 0, nchar(.prefix) - 1)
    .ret <- .Call(`_rxode2_rxUpdateTrans_`, .ret, .prefix, .libName)
  }
  ## dparser::dpReload();
  ## rxReload()
  if (is.character(code)) {
    .libname <- gsub("[.]c$", "", code)
    .libname <- c(.libname, .libname)
    .ret[[17]] <- list()
    .Call(
      `_rxode2_codegen`, code, .prefix, .libname,
            md5["parsed_md5"], "-1",
      .ret, .rxSupportedFuns(), "", "", "", "", "", "", "", "", "", "", "", "", "")
  }
  .ret
}

rxode2parseFuns <- function() {
  if (!requireNamespace("devtools", quietly = TRUE)) {
    stop("this requires devtools", call.=FALSE)
  }
  message("rebuild parseFuns.R from rxode2")
  try(source(devtools::package_file("build/refresh.R")), silent=TRUE)
  message("done")
  ""
}

#' This assigns the c level linkages for a roxde2 model
#'
#' @param df data frame containing the character column names rxFun,
#'   fun, type, package, packageFun and the integer column names
#'   argMin and argMax
#' @return Nothing called for side effects
#' @author Matthew L. Fidler
#' @export
#' @examples
#'
#' rxode2parseAssignTranslation(rxode2parseGetTranslation())
#'
rxode2parseAssignTranslation <- function(df) {
  .char <- c("rxFun", "fun", "type", "package", "packageFun")
  .int <- c("argMin", "argMax", "threadSafe")
  .df <- df[,c(.char, .int)]
  for (.c in .char) {
    .df[[.c]] <- as.character(.df[[.c]])
  }
  for (.i in .int) {
    .df[[.i]] <- as.integer(.df[[.i]])
  }
  assign(".rxode2parseDf", .df, envir=.parseEnv)
  invisible(.df)
}

#' This function gets the currently assigned translations
#'
#' @return The currently assigned translations
#' @author Matthew L. Fidler
#' @export
#' @examples
#' rxode2parseGetTranslation()
rxode2parseGetTranslation <- function() {
  .parseEnv$.rxode2parseDf
}

rxode2parseGetTranslationBuiltin <- function() {
  data.frame(n=names(.parseEnv$.parseNum), i=as.integer(setNames(.parseEnv$.parseNum, NULL)))
}


.parseEnv$.packagesToLoad <- c("rxode2ll", "lotri")

#'@rdname rxode2parseAssignPackagesToLoad
#'@export
rxode2parseGetPackagesToLoad <- function() {
  .parseEnv$.packagesToLoad
}

#' Control the packages that are loaded when a `rxode2` model dll is loaded
#'
#' @param pkgs The packages to make sure are loaded every time you load an rxode2 model.
#' @return List of packages to load
#' @author Matthew Fidler
#' @examples
#'
#' rxode2parseGetPackagesToLoad()
#'
#' rxode2parseAssignPackagesToLoad(rxode2parseGetPackagesToLoad())
#' @export
rxode2parseAssignPackagesToLoad <- function(pkgs=rxode2parseGetPackagesToLoad()) {
  assign(".packagesToLoad", pkgs, envir=.parseEnv)
  pkgs
}


.parseEnv$.rxode2parsePointerAssignment <- "rxode2"

#' This function gets the currently assigned function pointer assignments
#'
#' @return The currently assigned pointer assignments
#' @author Matthew L. Fidler
#' @export
#' @examples
#' rxode2parseGetTranslation()
rxode2parseGetPointerAssignment <- function() {
  .parseEnv$.rxode2parsePointerAssignment
}


#' This sets function gets the currently assigned function pointer assignments
#'
#' @param var List of packages where pointer assignment will be called.
#'
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
#' @examples
#' rxode2parseAssignPointerTranslation("rxode2")
rxode2parseAssignPointerTranslation <- function(var) {
  checkmate::assertCharacter(var)
  assign(".rxode2parsePointerAssignment", var, envir=.parseEnv)
  invisible()
}

rxTest({

  test_that("test lag-time information parsing", {

    m1 <- rxode2parse("KA=2.94E-01;
    CL=1.86E+01;
    V2=4.02E+01;
    Q=1.05E+01;
    V3=2.97E+02;
    Kin=1;
    Kout=1;
    EC50=200;
    fdepot = 1;
    durDepot = 8;
    rateDepot = 1250;
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    f(depot) = fdepot
    dur(depot) = durDepot
    rate(depot) = rateDepot
    alag(depot) = alagDepot
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    eff(0) = 1")
    expect_equal(m1$alag, 1L)

    m1 <- rxode2parse("KA=2.94E-01;
    CL=1.86E+01;
    V2=4.02E+01;
    Q=1.05E+01;
    V3=2.97E+02;
    Kin=1;
    Kout=1;
    EC50=200;
    fdepot = 1;
    durDepot = 8;
    rateDepot = 1250;
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    f(depot) = fdepot
    dur(depot) = durDepot
    rate(depot) = rateDepot
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    alag(centr) = alagDepot
    d/dt(peri)  =                    Q*C2 - Q*C3;
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    eff(0) = 1")
    expect_equal(m1$alag, 2L)

    m1 <- rxode2parse("KA=2.94E-01;
    CL=1.86E+01;
    V2=4.02E+01;
    Q=1.05E+01;
    V3=2.97E+02;
    Kin=1;
    Kout=1;
    EC50=200;
    fdepot = 1;
    durDepot = 8;
    rateDepot = 1250;
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    f(depot) = fdepot
    dur(depot) = durDepot
    rate(depot) = rateDepot
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    alag(peri) = alagDepot
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    eff(0) = 1")
    expect_equal(m1$alag, 3L)

    m1 <- rxode2parse("KA=2.94E-01;
    CL=1.86E+01;
    V2=4.02E+01;
    Q=1.05E+01;
    V3=2.97E+02;
    Kin=1;
    Kout=1;
    EC50=200;
    fdepot = 1;
    durDepot = 8;
    rateDepot = 1250;
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    f(depot) = fdepot
    dur(depot) = durDepot
    rate(depot) = rateDepot
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    alag(eff) = alagDepot
    eff(0) = 1")
    expect_equal(m1$alag, 4L)

    m1 <- rxode2parse("KA=2.94E-01;
    CL=1.86E+01;
    V2=4.02E+01;
    Q=1.05E+01;
    V3=2.97E+02;
    Kin=1;
    Kout=1;
    EC50=200;
    fdepot = 1;
    durDepot = 8;
    rateDepot = 1250;
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    f(depot) = fdepot
    dur(depot) = durDepot
    rate(depot) = rateDepot
    alag(depot) = alagDepot
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    alag(centr) = alagDepot
    d/dt(peri)  =                    Q*C2 - Q*C3;
    alag(peri) = alagDepot
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    alag(eff) = alagDepot
    eff(0) = 1")
    expect_equal(m1$alag, 1:4)

    m1 <- rxode2parse("KA=2.94E-01;
    CL=1.86E+01;
    V2=4.02E+01;
    Q=1.05E+01;
    V3=2.97E+02;
    Kin=1;
    Kout=1;
    EC50=200;
    fdepot = 1;
    durDepot = 8;
    rateDepot = 1250;
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    f(depot) = fdepot
    dur(depot) = durDepot
    rate(depot) = rateDepot
    alag(depot) = alagDepot
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    alag(peri) = alagDepot
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    alag(eff) = alagDepot
    eff(0) = 1")
    expect_equal(m1$alag, c(1L, 3:4))

    m1 <- rxode2parse("KA=2.94E-01;
    CL=1.86E+01;
    V2=4.02E+01;
    Q=1.05E+01;
    V3=2.97E+02;
    Kin=1;
    Kout=1;
    EC50=200;
    fdepot = 1;
    durDepot = 8;
    rateDepot = 1250;
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    f(depot) = fdepot
    dur(depot) = durDepot
    rate(depot) = rateDepot
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    alag(centr) = alagDepot
    d/dt(peri)  =                    Q*C2 - Q*C3;
    alag(peri) = alagDepot
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    alag(eff) = alagDepot
    eff(0) = 1")
    expect_equal(m1$alag, 2:4)

    m1 <- rxode2parse("KA=2.94E-01;
    CL=1.86E+01;
    V2=4.02E+01;
    Q=1.05E+01;
    V3=2.97E+02;
    Kin=1;
    Kout=1;
    EC50=200;
    fdepot = 1;
    durDepot = 8;
    rateDepot = 1250;
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    f(depot) = fdepot
    dur(depot) = durDepot
    rate(depot) = rateDepot
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    eff(0) = 1")
    expect_equal(m1$alag, integer(0))

  })


  et <- et(1:10)
  et$b <- 1:10

  test_that("lag()", {

    suppressMessages(expect_error(rxode2({
      a <- lag()
    })))

    suppressMessages(expect_error(rxode2({
      a <- lag(b, c)
    })))

    m1 <- rxode2({
      c <- b + 2
      a <- lag(b, 3)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(NA, NA, NA, 1, 2, 3, 4, 5, 6, 7))

    m1 <- rxode2({
      a <- lag(b, 3)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(NA, NA, NA, 1, 2, 3, 4, 5, 6, 7))

    m1 <- rxode2({
      a <- lag(b, -1)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(2, 3, 4, 5, 6, 7, 8, 9, 10, NA))

    m1 <- rxode2({
      a <- lag(b, 0)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, 1:10)

    m1 <- rxode2({
      a <- lag(b)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(NA, 1, 2, 3, 4, 5, 6, 7, 8, 9))

    m1 <- rxode2({
      a <- b
      c <- lag(a)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$c, c(NA, 1, 2, 3, 4, 5, 6, 7, 8, 9))

    m1 <- rxode2({
      a <- b
      c <- lag(a, 1)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$c, c(NA, 1, 2, 3, 4, 5, 6, 7, 8, 9))

    m1 <- rxode2({
      a <- b
      c <- lead(a, -1)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$c, c(NA, 1, 2, 3, 4, 5, 6, 7, 8, 9))

    m1 <- rxode2({
      c <- b + 3
      a <- lag(b)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(NA, 1, 2, 3, 4, 5, 6, 7, 8, 9))
  })

  test_that("lead()", {
    suppressMessages(expect_error(rxode2({
      a <- lead()
    })))

    suppressMessages(expect_error(rxode2({
      a <- lead(b, c)
    })))

    m1 <- rxode2({
      c <- b + 2
      a <- lead(b, 3)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(4, 5, 6, 7, 8, 9, 10, NA, NA, NA))

    m1 <- rxode2({
      a <- lead(b, 3)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(4, 5, 6, 7, 8, 9, 10, NA, NA, NA))

    m1 <- rxode2({
      a <- lead(b, -1)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(NA, 1, 2, 3, 4, 5, 6, 7, 8, 9))

    m1 <- rxode2({
      a <- lead(b, 0)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, 1:10)

    m1 <- rxode2({
      a <- lead(b)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(2:10, NA))

    m1 <- rxode2({
      c <- b + 3
      a <- lead(b)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(2:10, NA))
  })

  test_that("first()", {
    suppressMessages(expect_error(rxode2({
      a <- first()
    })))

    suppressMessages(expect_error(rxode2({
      a <- first(b, 1)
    })))

    suppressMessages(expect_error(rxode2({
      a <- first(b, 1, 2)
    })))

    m1 <- rxode2({
      a <- first(b)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_true(all(x1$a == 1))

    m1 <- rxode2({
      c <- b + 3
      a <- first(b)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_true(all(x1$a == 1))
  })

  test_that("last()", {
    suppressMessages(expect_error(rxode2({
      a <- last()
    })))

    suppressMessages(expect_error(rxode2({
      a <- last(b, 1)
    })))

    suppressMessages(expect_error(rxode2({
      a <- last(b, 1, 2)
    })))

    m1 <- rxode2({
      a <- last(b)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_true(all(x1$a == 10))

    m1 <- rxode2({
      c <- b + 3
      a <- last(b)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_true(all(x1$a == 10))
  })


  et <- et(1:10)
  et$b <- 2^(1:10)

  test_that("diff()", {
    suppressMessages(expect_error(rxode2({
      a <- diff()
    })))

    suppressMessages(expect_error(rxode2({
      a <- diff(b, 1, 2)
    })))

    suppressMessages(expect_error(rxode2({
      a <- diff(b, 1.2)
    })))

    suppressMessages(expect_error(rxode2({
      a <- diff(b, c)
    })))

    suppressMessages(expect_error(rxode2({
      a <- diff(b, -1)
    })))

    suppressMessages(expect_error(rxode2({
      a <- diff(b, 0)
    })))

    m1 <- rxode2({
      a <- diff(b)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(NA, 2, 4, 8, 16, 32, 64, 128, 256, 512))

    m1 <- rxode2({
      c <- b
      a <- diff(c)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(NA, 2, 4, 8, 16, 32, 64, 128, 256, 512))

    m1 <- rxode2({
      c <- b
      a <- diff(c, 1)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(NA, 2, 4, 8, 16, 32, 64, 128, 256, 512))

    m1 <- rxode2({
      a <- diff(b, 2)
    })

    expect_s3_class(m1, "rxode2")

    x1 <- m1 |> rxSolve(et)

    expect_equal(x1$a, c(NA, NA, 6, 12, 24, 48, 96, 192, 384, 768))
  })

  test_that("bad lag() types", {
    suppressMessages(expect_error(rxode2({
      a ~ c + d
      b <- lag(a)
    })))

    suppressMessages(expect_error(rxode2({
      d / dt(a) <- 3
      b <- lag(a)
    })))

    suppressMessages(expect_error(rxode2({
      a <- a + 3
      b <- lag(a)
    })))

    suppressMessages(expect_error(rxode2({
      a <- 13 + b
      b <- lag(a, 3)
    })))

    suppressMessages(expect_error(rxode2({
      a <- 13 + b
      b <- lead(a)
    })))
  })

  test_that("test sticky lhs", {
    mod1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      C2 <- centr / V2
      C3 <- peri / V3
      d / dt(depot) <- -KA * depot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      isna <- is.na(amt)
      if (!is.na(amt)) {
        tdose <- time
      } else {
        tad <- time - tdose
      }
    })

    ev <- et(amountUnits = "mg", timeUnits = "hours") |>
      et(amt = 10000, addl = 9, ii = 12, cmt = "depot") |>
      et(time = 120, amt = 2000, addl = 4, ii = 14, cmt = "depot") |>
      et(0, 240)

    r1 <- rxSolve(mod1, ev, addDosing = TRUE)

    expect_equal(max(r1$tad, na.rm = TRUE), 64)

    r2 <- rxSolve(mod1, ev, addDosing = FALSE)

    expect_equal(max(r2$tad, na.rm = TRUE), 64)
  })

  test_that("newind", {
    mod1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      C2 <- centr / V2
      C3 <- peri / V3
      d / dt(depot) <- -KA * depot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      if (!is.na(amt)) {
        tdose <- time
      } else {
        tad <- time - tdose
      }
      if (newind <= 1) {
        first <- 0
      } else if (tad > 24) {
        first <- 24
      }
    })

    ev <- et(amountUnits = "mg", timeUnits = "hours") |>
      et(amt = 10000, addl = 9, ii = 12, cmt = "depot") |>
      et(time = 120, amt = 2000, addl = 4, ii = 14, cmt = "depot") |>
      et(0, 240)

    r1 <- rxSolve(mod1, ev, addDosing = TRUE)
    expect_equal(unique(r1$first), c(0, 24))

    r1 <- rxSolve(mod1, ev, addDosing = FALSE)
    expect_equal(unique(r1$first), c(0, 24))
  })
})

#' @rdname rxEvid
#' @export
format.rxEvid <- function(x, ...) {
  .x <- unclass(x)
  format(as.character.rxEvid(.x), align = "left", width = 12)
}

#' @rdname rxEvid
#' @export
format.rxRateDur <- function(x, ...) {
  .x <- unclass(x)
  format(as.character.rxRateDur(.x), align = "left")
}

.fmt3 <- function(name, bound, access) {
  paste0(
    crayon::bold(name), " (", crayon::yellow(bound),
    crayon::bold$blue(paste0("$", access)), "):"
  )
}
#' @export
format.boundParams <- function(x, ...) {
  cli::cli_format_method({
    .h2(.fmt3("Parameters", x, "params"))
  })
}

#' @export
format.boundInits <- function(x, ...) {
  cli::cli_format_method({
    .h2(.fmt3("Initial Conditions", x, "inits"))
  })
}


#' @export
format.rxSolveSimType <- function(x, ...) {
  .args <- as.list(match.call(expand.dots = TRUE))
  if (any(names(.args) == "bound")) {
    .bound <- .args$bound
  } else {
    .bound <- .getBound(x, parent.frame(2))
  }
  .uncert <- character(0)
  if (!isNullZero(x$thetaMat)) {
    .uncert <- c(.uncert, paste0("parameters (", crayon::yellow(.bound), crayon::bold$blue("$thetaMat"), " for changes)"))
  }
  if (!isNullZero(x$omegaList)) {
    .uncert <- c(.uncert, paste0("omega matrix (", crayon::yellow(.bound), crayon::bold$blue("$omegaList"), ")"))
  }
  if (!isNullZero(x$sigmaList)) {
    .uncert <- c(.uncert, paste0("sigma matrix (", crayon::yellow(.bound), crayon::bold$blue("$sigmaList"), ")"))
  }
  if (length(.uncert) == 0L) {
    .first <- paste0("\nSimulation ", crayon::bold("without uncertainty"), " in parameters, omega, or sigma matricies\n\n")
  } else if (length(.uncert) == 1L) {
    .first <- paste0("\nSimulation ", crayon::bold("with uncertainty"), " in ", paste(.uncert, collapse = ", "), "\n")
  } else {
    .first <- paste0("\nSimulation ", crayon::bold("with uncertainty"), " in:")
    return(cli::cli_format_method({
      cli::cli_text("")
      cli::cli_text(.first)
      cli::cli_ul(.uncert)
      cli::cli_text("")
    }))
  }
  return(cli::cli_format_method({
    cli::cli_text("")
    cli::cli_text(.first)
    cli::cli_text("")
  }))
}

.ggplot2Fix <- function() {
  .ggplot2 <- loadNamespace("ggplot2")
  if (any(ls(.ggplot2) == "guide_none")) {
    assignInMyNamespace("guide_none", .ggplot2$guide_none)
  }
  if (compareVersion(as.character(packageVersion("ggplot2")), "3.3.6.9000") < 0) {
    assignInMyNamespace("GeomAmt",
                        ggplot2::ggproto("GeomAmt", ggplot2::GeomSegment,
                                         required_aes = c("x", "y", "xend", "yend"),
                                         default_aes = ggplot2::aes(colour = "black", linetype = "dotted", size = 0.5, alpha = 1, fill = "black")))
  } else {
    assignInMyNamespace("GeomAmt",
                        ggplot2::ggproto("GeomAmt", ggplot2::GeomSegment,
                                         required_aes = c("x", "y", "xend", "yend"),
                                         default_aes = ggplot2::aes(colour = "black", linetype = "dotted", linewidth = 0.5, alpha = 1, fill = "black")))
  }
}
.hasUnits <- FALSE
#' Get the rxode2 function pointers
#'
#' This function is used to get the function pointers for rxode2.  This is
#' used to allow rxode2 to have binary linkage to nlmixr2est.
#'
#' @return a list of function pointers
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' .rxode2ptrs()
.rxode2ptrs <- function() {
  .Call(`_rxode2_rxode2Ptr`, PACKAGE = "rxode2")
}

## nocov start
.onLoad <- function(libname, pkgname) {
  .ver <- .rxVersion
  .ver["version"] <- as.character(utils::packageVersion("rxode2"))
  assignInMyNamespace(".rxVersion", .ver)
  ## data.table is an Imports and is used unconditionally, so keep it eager
  requireNamespace("data.table", quietly=TRUE)
  ## `.s3register()` registers immediately when the other namespace is
  ## already loaded and otherwise installs an onLoad hook, so the suggested
  ## packages must not be loaded here -- doing so made `library(rxode2)`
  ## pay for pillar/arrow/units/dplyr on every session.
  .s3register("pillar::type_sum", "rxEvid")
  .s3register("pillar::type_sum", "rxRateDur")
  .s3register("pillar::pillar_shaft", "rxEvid")
  .s3register("pillar::pillar_shaft", "rxRateDur")
  .s3register("tibble::as_tibble", "rxEt")
  .s3register("tibble::as_tibble", "rxSolveOom")
  .s3register("data.table::as.data.table", "rxEt")
  .s3register("data.table::as.data.table", "rxSolveOom")
  .s3register("arrow::as_arrow_table", "rxSolveOom")
  .s3register("dplyr::dplyr_reconstruct", "rxEt")
  .s3register("dplyr::dplyr_reconstruct", "rxEtPreview")
  .s3register("dplyr::filter", "rxEt")
  .s3register("dplyr::select", "rxEt")
  .s3register("dplyr::rename", "rxEt")
  .s3register("dplyr::mutate", "rxEt")
  .s3register("dplyr::rename", "rxUi")
  .s3register("dplyr::rename", "function")
  .s3register("nlme::fixef", "rxUi")
  .s3register("nlme::fixef", "function")
  .s3register("units::set_units", "rxEt")
  .s3register("units::set_units", "rxRateDur")
  .s3register("units::drop_units", "rxEt")
  .s3register("units::units<-", "rxEvid")
  .s3register("units::drop_units", "rxSolve")
  .s3register("digest::sha1", "rxUi")
  assignInMyNamespace(".hasUnits", nzchar(system.file(package = "units")))
  .s3register("lotri::as.lotri", "call")
  backports::import(pkgname)
  ## Setup rxode2.prefer.tbl
  .Call(`_rxode2_setRstudio`, Sys.getenv("RSTUDIO") == "1")
  rxSyncOptions("permissive")
  rxTempDir()
  if (!interactive()) {
    setProgSupported(0)
  }
  .ggplot2Fix()
  .linkAll()
  forderForceBase(FALSE)
} ## nocov end

.iniLotriPtrs <- function() {
  .Call(`_iniLotriPtr`, lotri::.lotriPointers())
}

.iniPreciseSumsPtr <- function() {
  .Call(`_iniPreciseSumsPtr`, PreciseSums::.preciseSumsPtr())
}

.iniDparserPtr <- function() {
  .Call(`_rxode2_iniDparserPtr`, dparser::.dparsePtr())
}

.linkAll <- function() {
  .iniLotriPtrs()
  .iniPreciseSumsPtr()
  .iniDparserPtr()
}


.onAttach <- function(libname, pkgname) {
  .ver <- .rxVersion
  .ver["version"] <- as.character(utils::packageVersion("rxode2"))
  assignInMyNamespace(".rxVersion", .ver)
  .Call(`_rxode2_setRstudio`, Sys.getenv("RSTUDIO") == "1")
  rxSyncOptions("permissive")
  if (!interactive()) {
    setProgSupported(0)
  }
  .linkAll()

  rxTempDir()
  .ggplot2Fix()
  v <- utils::packageVersion("rxode2")
  packageStartupMessage(
    "rxode2 ", v, " using ", getRxThreads(verbose = FALSE),
    " threads (see ?getRxThreads)",
    ifelse(.cacheIsTemp, "\n  no cache: create with `rxCreateCache()`", "")
  )
  if (!.Call(`_rxHasOpenMp`)) {
    packageStartupMessage(
      "========================================\n",
      "rxode2 has not detected OpenMP support and will run in single-threaded mode\n",
      if (Sys.info()["sysname"] == "Darwin") {
        "This is a Mac. Please read https://mac.r-project.org/openmp/"
      } else {
        paste0("The system is ", Sys.info()["sysname"], "; To get best performance enable OpenMP")
      },
      "\n========================================\n"
    )
  }
  forderForceBase(FALSE)
}

.onUnload <- function(libpath) {
  ## nocov start
  rxUnloadAll()
  gc() # Force garbage collection finalization
  library.dynam.unload("rxode2", libpath)
  ## nocov end
}

.mkCache <- function(.tmp) {
  if (!file.exists(.tmp)) {
    dir.create(.tmp, recursive = TRUE, showWarnings = FALSE)
  } else if (!file.exists(file.path(.tmp, paste0(rxode2.md5, ".md5")))) {
    if (!.cacheIsTemp) packageStartupMessage("detected new version of rxode2, cleaning cache")
    unlink(.tmp, recursive = TRUE, force = TRUE)
    dir.create(.tmp, recursive = TRUE, showWarnings = FALSE)
    writeLines("rxode2", file.path(.tmp, paste0(rxode2.md5, ".md5")))
  }
}

.cacheIsTemp <- TRUE
.rxTempDir0 <- NULL
.cacheDefault <- NULL
#' Get the rxode2 temporary directory
#'
#' @return rxode2 temporary directory.
#' @export
rxTempDir <- function() {
  if (is.null(getFromNamespace(".rxTempDir0", "rxode2"))) {
    .tmp <- Sys.getenv("rxTempDir")
    .rxUserDir <- R_user_dir("rxode2", "cache")
    assignInMyNamespace(".cacheIsTemp", FALSE)
    if (!file.exists(.rxUserDir)) {
      .rxUserDir <- file.path(tempdir(), "rxode2")
      assignInMyNamespace(".cacheIsTemp", TRUE)
    }
    if (.tmp == "") {
      if (is.null(.cacheDefault)) {
        assignInMyNamespace(".cacheDefault", .rxUserDir)
      }
      if (getOption("rxode2.cache.directory", .cacheDefault) != ".") {
        .tmp <- getOption("rxode2.cache.directory", .cacheDefault)
      } else {
        .tmp <- .rxUserDir
      }
    }
    .mkCache(.tmp)
    .tmp <- .normalizePath(.tmp)
    Sys.setenv(rxTempDir = .tmp)
    utils::assignInMyNamespace(".rxTempDir0", .tmp)
    options(rxode2.cache.directory = .tmp)
    return(.tmp)
  } else {
    .tmp <- getFromNamespace(".rxTempDir0", "rxode2")
    .mkCache(.tmp)
    options(rxode2.cache.directory = .tmp)
    return(.tmp)
  }
}
#' This will create the cache directory for rxode2 to save between sessions
#'
#' When run, if the `R_user_dir` for rxode2's cache isn't present,
#' create the cache
#'
#' @return nothing
#'
#' @author Matthew Fidler
#'
#' @export
rxCreateCache <- function() {
  .tmp <- R_user_dir("rxode2", "cache")
  assignInMyNamespace(".cacheDefault", R_user_dir("rxode2", "cache"))
  .mkCache(.tmp)
  .tmp <- .normalizePath(.tmp)
  Sys.setenv(rxTempDir = .tmp)
  utils::assignInMyNamespace(".rxTempDir0", .tmp)
  options(rxode2.cache.directory=.tmp)
  invisible()
}

#' Clear memoise cache for rxode2
#'
#' @author Matthew L. Fidler
#' @return nothing; called for side effects
#' @keywords internal
#' @export
## Cache of memoised function names; computed once on first rxForget() call.
.rxMemoisedFns <- NULL

#' Clear memoise caches for rxode2
#'
#' @return nothing; called for side effects
#' @keywords internal
#' @export
rxForget <- function() {
  if (is.null(.rxMemoisedFns)) {
    .ns <- getNamespace("rxode2")
    .fns <- Filter(
      function(fn) {
        .v <- get(fn, envir=.ns, inherits=FALSE)
        is.function(.v) && memoise::is.memoised(.v)
      },
      ls(envir=.ns)
    )
    assignInMyNamespace(".rxMemoisedFns", .fns)
  }
  for (.fn in .rxMemoisedFns) {
    memoise::forget(getFromNamespace(.fn, "rxode2"))
  }
}

.isTestthat <- function() {
  return(regexpr("/tests/testthat/", getwd(), fixed = TRUE) != -1) # nolint
}
#' Respect suppress messages
#'
#' This turns on the silent REprintf in C when `suppressMessages()` is
#' turned on. This makes the `REprintf` act like `messages` in R,
#' they can be suppressed with `suppressMessages()`
#'
#' @return Nothing
#' @author Matthew Fidler
#' @export
#' @examples
#'
#' # rxSupressMsg() is called with rxode2()
#'
#' # Note the errors are output to the console
#'
#' try(rxode2("d/dt(matt)=/3"), silent = TRUE)
#'
#' # When using suppressMessages, the output is suppressed
#'
#' suppressMessages(try(rxode2("d/dt(matt)=/3"), silent = TRUE))
#'
#' # In rxode2, we use REprintf so that interrupted threads do not crash R
#' # if there is a user interrupt. This isn't captured by R's messages, but
#' # This interface allows the `suppressMessages()` to suppress the C printing
#' # as well
#'
#' # If you  want to suppress messages from rxode2 in other packages, you can use
#' # this function
rxSuppressMsg <- function() {
  if (requireNamespace("knitr", quietly = TRUE)) {
    if (!is.null(knitr::opts_knit$get("rmarkdown.pandoc.to"))) {
      return(invisible(NULL))
    } else {
      rxSetSilentErr(as.integer(length(capture.output(message(" "), type = "message")) == 0L))
    }
  } else {
    rxSetSilentErr(as.integer(length(capture.output(message(" "), type = "message")) == 0L))
  }
  invisible(NULL)
}

#' Sync options with rxode2 variables
#'
#'
#' @param setDefaults This will setup rxode2's default solving options with the following options:
#'
#' - `"none"` leave the options alone
#' - `"permissive"` This is a permissive option set similar to R language specifications.
#' - `"strict"` This is a strict option set similar to the original
#'    rxode2(). It requires semicolons at the end of lines and equals for
#'    assignment
#'
#' @author Matthew L. Fidler
#' @return nothing; called for side effects
#' @export
rxSyncOptions <- function(setDefaults = c("none", "permissive", "strict")) {
  setDefaults <- match.arg(setDefaults)
  if (setDefaults == "permissive") {
    options(rxode2.syntax.allow.ini=TRUE)
  } else if (setDefaults == "strict") {
    options(rxode2.syntax.allow.ini=FALSE)
  }
}

## Vendor SUNDIALS C source files and public headers from the sundialr source
## tarball into src/ (sources) and src/sundials_inc/ (headers).
##
## The sources and headers MUST come from the same SUNDIALS release; vendoring
## both keeps them in lock-step (https://github.com/nlmixr2/rxode2/issues/1155).
##
## Run this script from the package root when:
##   - upgrading to a new SUNDIALS/sundialr version
##   - re-generating vendored files after a clean checkout that lost them
##   - the committed src/sundials_*.c or src/sundials_inc files need refreshing
##
## Usage (from the package root):
##   Rscript build/vendor-sundials.R
##   # or from within R:
##   source("build/vendor-sundials.R")
##
## After running, review any diff in src/ and commit the updated files.

if (!nzchar(Sys.getenv("RSCRIPT_IS_MAIN"))) {
  ## Allow sourcing without side-effects during devtools::document()
}

.pkg_root <- tryCatch(devtools::package_file(), error = function(e) getwd())
.old <- setwd(.pkg_root)
on.exit(setwd(.old), add = TRUE)

## sundialr is NOT a dependency of rxode2; it is only the vendoring source.
## Pick the version via the SUNDIALR_VERSION env var, or fall back to the
## locally installed sundialr if there is one.
.sdr_ver <- Sys.getenv("SUNDIALR_VERSION", "")
if (!nzchar(.sdr_ver)) {
  .sdr_ver <- tryCatch(as.character(packageVersion("sundialr")),
                       error = function(e) {
                         stop("Set SUNDIALR_VERSION (e.g. 'SUNDIALR_VERSION=0.1.7 ",
                              "Rscript build/vendor-sundials.R') or install sundialr.",
                              call. = FALSE)
                       })
}
message("Vendoring SUNDIALS from sundialr ", .sdr_ver, " ...")

## File mapping: path inside the SUNDIALS source tree -> dst in src/
.map <- c(
  "cvode/cvode.c"                                    = "sundials_cvode.c",
  "cvode/cvode_diag.c"                               = "sundials_cvode_diag.c",
  "cvode/cvode_io.c"                                 = "sundials_cvode_io.c",
  "cvode/cvode_ls.c"                                 = "sundials_cvode_ls.c",
  "cvode/cvode_nls.c"                                = "sundials_cvode_nls.c",
  "cvode/cvode_proj.c"                               = "sundials_cvode_proj.c",
  "cvode/cvode_impl.h"                               = "cvode_impl.h",
  "cvode/cvode_diag_impl.h"                          = "cvode_diag_impl.h",
  "cvode/cvode_ls_impl.h"                            = "cvode_ls_impl.h",
  "cvode/cvode_proj_impl.h"                          = "cvode_proj_impl.h",
  "nvector/serial/nvector_serial.c"                  = "sundials_nvector_serial.c",
  "sunmemory/system/sundials_system_memory.c"        = "sundials_system_memory.c",
  "sunmatrix/band/sunmatrix_band.c"                  = "sundials_sunmatrix_band.c",
  "sunmatrix/dense/sunmatrix_dense.c"                = "sundials_sunmatrix_dense.c",
  "sunmatrix/sparse/sunmatrix_sparse.c"              = "sundials_sunmatrix_sparse.c",
  "sunlinsol/band/sunlinsol_band.c"                  = "sundials_sunlinsol_band.c",
  "sunlinsol/dense/sunlinsol_dense.c"                = "sundials_sunlinsol_dense.c",
  "sunlinsol/spgmr/sunlinsol_spgmr.c"               = "sunlinsol_spgmr.c",
  "sunlinsol/spbcgs/sunlinsol_spbcgs.c"             = "sunlinsol_spbcgs.c",
  "sunlinsol/sptfqmr/sunlinsol_sptfqmr.c"           = "sunlinsol_sptfqmr.c",
  "sunnonlinsol/newton/sunnonlinsol_newton.c"        = "sundials_sunnonlinsol_newton.c",
  "sunnonlinsol/fixedpoint/sunnonlinsol_fixedpoint.c"= "sundials_sunnonlinsol_fixedpoint.c",
  "sundials/sundials_adiak_metadata.h"               = "sundials_adiak_metadata.h",
  "sundials/sundials_cli.h"                          = "sundials_cli.h",
  "sundials/sundials_datanode.h"                     = "sundials_datanode.h",
  "sundials/sundials_hashmap_impl.h"                 = "sundials_hashmap_impl.h",
  "sundials/sundials_iterative_impl.h"               = "sundials_iterative_impl.h",
  "sundials/sundials_logger_impl.h"                  = "sundials_logger_impl.h",
  "sundials/sundials_macros.h"                       = "sundials_macros.h",
  "sundials/sundials_profiler_impl.h"                = "sundials_profiler_impl.h",
  "sundials/sundials_utils.h"                        = "sundials_utils.h",
  "sundials/sundials_debug.h"                        = "sundials_debug.h",
  "sundials/sundials_band.c"                         = "sundials_sundials_band.c",
  "sundials/sundials_cli.c"                          = "sundials_sundials_cli.c",
  "sundials/sundials_context.c"                      = "sundials_sundials_context.c",
  "sundials/sundials_dense.c"                        = "sundials_sundials_dense.c",
  "sundials/sundials_direct.c"                       = "sundials_sundials_direct.c",
  "sundials/sundials_errors.c"                       = "sundials_sundials_errors.c",
  "sundials/sundials_hashmap.c"                      = "sundials_sundials_hashmap.c",
  "sundials/sundials_iterative.c"                    = "sundials_sundials_iterative.c",
  "sundials/sundials_linearsolver.c"                 = "sundials_sundials_linearsolver.c",
  "sundials/sundials_logger.c"                       = "sundials_sundials_logger.c",
  "sundials/sundials_math.c"                         = "sundials_sundials_math.c",
  "sundials/sundials_matrix.c"                       = "sundials_sundials_matrix.c",
  "sundials/sundials_memory.c"                       = "sundials_sundials_memory.c",
  "sundials/sundials_nonlinearsolver.c"              = "sundials_sundials_nonlinearsolver.c",
  "sundials/sundials_nvector.c"                      = "sundials_sundials_nvector.c",
  "sundials/sundials_profiler.c"                     = "sundials_sundials_profiler.c",
  "sundials/sundials_version.c"                      = "sundials_sundials_version.c",
  "sundials/stl/sunstl_vector.h"                     = file.path("stl", "sunstl_vector.h")
)

## ---------------------------------------------------------------------------
## Locate or download the sundialr SOURCE tarball.
## Prefer a cached copy; fall back to a direct CRAN download so we always get
## the true source (not a posit RSPM binary served at src/contrib/).
## ---------------------------------------------------------------------------

.tarball_name <- paste0("sundialr_", .sdr_ver, ".tar.gz")
.search_dirs  <- c(tempdir(),
                   Sys.getenv("TEMP", unset = ""),
                   Sys.getenv("TMP",  unset = ""),
                   tools::R_user_dir("cache", which = "cache"),
                   getwd())
.search_dirs  <- .search_dirs[nzchar(.search_dirs)]
.tarball      <- Filter(file.exists, file.path(.search_dirs, .tarball_name))
.tarball      <- if (length(.tarball)) .tarball[1] else character(0)

if (!length(.tarball) && !nzchar(Sys.getenv("SUNDIALR_VERSION"))) {
  ## Try download.packages() with the configured repos first (may be RSPM).
  ## If that produces a binary (no nested tarball), we fall back below.
  ## Skipped when SUNDIALR_VERSION is set: download.packages() always fetches
  ## the repo's CURRENT release and would silently replace the pinned version.
  .tarball <- tryCatch(
    utils::download.packages("sundialr", destdir = tempdir(),
                             repos = getOption("repos"), type = "source")[1, 2],
    error = function(e) character(0)
  )
  if (length(.tarball) && basename(.tarball) != .tarball_name) .tarball <- character(0)
}

if (!length(.tarball) || !file.exists(.tarball)) {
  ## Fall back to the canonical CRAN source URLs -- guaranteed to be the real
  ## source tarball regardless of which mirror/RSPM the session is using.
  ## Superseded versions live under src/contrib/Archive/sundialr/.
  .dest <- file.path(tempdir(), .tarball_name)
  .cran_urls <- paste0("https://cran.r-project.org/src/contrib/",
                       c("", "Archive/sundialr/"), .tarball_name)
  .ok <- FALSE
  for (.u in .cran_urls) {
    message("Downloading ", .u)
    .ok <- tryCatch({
      download.file(.u, .dest, mode = "wb", quiet = FALSE)
      TRUE
    }, error = function(e) FALSE)
    if (.ok) break
  }
  if (!.ok) {
    stop("Could not download the sundialr ", .sdr_ver, " source tarball from ",
         "CRAN (tried src/contrib and src/contrib/Archive).", call. = FALSE)
  }
  .tarball <- .dest
}

## ---------------------------------------------------------------------------
## Extract and locate the SUNDIALS source tree.
##
## sundialr packages the SUNDIALS source in one of two ways:
##   (a) as a nested tarball  sundialr/src/sundials-mod-<ver>.tar.gz
##   (b) as flat source files directly under sundialr/src/sundials-<ver>/
## We handle both.
## ---------------------------------------------------------------------------

.td <- tempfile("sundialr-vendor-")
dir.create(.td)
on.exit(unlink(.td, recursive = TRUE), add = TRUE)

message("Extracting ", .tarball, " ...")
utils::untar(.tarball, exdir = .td)

## Strategy (a): nested tarball
.nested <- list.files(file.path(.td, "sundialr", "src"),
                      pattern = "^sundials-mod-.*\\.tar\\.gz$",
                      full.names = TRUE)
if (length(.nested)) {
  message("Found nested SUNDIALS tarball: ", basename(.nested[1]))
  utils::untar(.nested[1], exdir = .td)
}

## Find the extracted SUNDIALS source root (either from nested or flat layout)
.sun_dirs <- list.dirs(.td, recursive = FALSE, full.names = TRUE)
.sun_dirs <- .sun_dirs[grepl("sundials[-_][0-9]", basename(.sun_dirs))]
if (!length(.sun_dirs)) {
  ## Also try one level deeper in case it was inside sundialr/
  .sun_dirs <- list.dirs(file.path(.td, "sundialr"), recursive = FALSE, full.names = TRUE)
  .sun_dirs <- .sun_dirs[grepl("sundials[-_][0-9]", basename(.sun_dirs))]
}
if (!length(.sun_dirs)) {
  stop("Unable to locate SUNDIALS source tree inside sundialr tarball.\n",
       "Checked: ", .td, "\n",
       "Run 'list.dirs(\"", .td, "\", recursive=TRUE)' to inspect the layout.",
       call. = FALSE)
}
.sun <- .sun_dirs[1]
message("Using SUNDIALS source tree: ", basename(.sun))

## ---------------------------------------------------------------------------
## Copy files
## ---------------------------------------------------------------------------

suppressWarnings(dir.create("src/stl", recursive = TRUE, showWarnings = FALSE))

.copied  <- character(0)
.missing <- character(0)
for (.rel in names(.map)) {
  .src <- file.path(.sun, "src", .rel)
  .dst <- file.path("src", .map[[.rel]])
  if (!file.exists(.src)) {
    .missing <- c(.missing, .rel)
    next
  }
  dir.create(dirname(.dst), recursive = TRUE, showWarnings = FALSE)
  file.copy(.src, .dst, overwrite = TRUE)
  .copied <- c(.copied, .map[[.rel]])
}

if (length(.missing)) {
  warning("The following source files were not found in the SUNDIALS tree and were skipped:\n  ",
          paste(.missing, collapse = "\n  "), call. = FALSE)
}
message("Copied ", length(.copied), " files to src/.")

## ---------------------------------------------------------------------------
## Vendor the complete SUNDIALS public header tree into src/sundials_inc/.
## sundialr ships it as inst/include; the subdirectory structure must be
## preserved because the sources use the '#include <cvode/cvode.h>' form.
## The sundialr-specific R helper headers at the top level are not part of
## SUNDIALS and are excluded.
## ---------------------------------------------------------------------------

.inc_src <- file.path(.td, "sundialr", "inst", "include")
if (!dir.exists(.inc_src)) {
  stop("Unable to locate the SUNDIALS public headers (sundialr/inst/include) ",
       "inside the sundialr tarball.", call. = FALSE)
}
unlink("src/sundials_inc", recursive = TRUE)
dir.create("src/sundials_inc", recursive = TRUE)
.hdrs <- list.files(.inc_src, recursive = TRUE, all.files = FALSE)
.hdrs <- .hdrs[!.hdrs %in% c("check_retval.h", "rhs_func.h", "sundials_err_handler.h")]
for (.h in .hdrs) {
  .dst <- file.path("src", "sundials_inc", .h)
  dir.create(dirname(.dst), recursive = TRUE, showWarnings = FALSE)
  file.copy(file.path(.inc_src, .h), .dst, overwrite = TRUE)
}
message("Copied ", length(.hdrs), " headers to src/sundials_inc/.")

## ---------------------------------------------------------------------------
## Strip sensitivity-only functions from the nonlinear solver files.
## CVODE (not CVODES) does not need sensitivity wrapping, and
## sundials_nvector_senswrapper.h is not shipped in the installed sundialr headers.
## ---------------------------------------------------------------------------

.strip_sens <- function(.path, .fn_name) {
  if (!file.exists(.path)) return(invisible(NULL))
  .lines <- readLines(.path)
  .lines <- .lines[!grepl("sundials_nvector_senswrapper\\.h", .lines)]
  .start <- grep(paste0("SUNNonlinearSolver\\s+", .fn_name, "\\s*\\("), .lines)
  if (length(.start)) {
    .cb <- .start[1] - 1L
    while (.cb >= 1L && !grepl("^/\\*={10,}", .lines[.cb])) .cb <- .cb - 1L
    .cb <- max(.cb, 1L)
    .end <- .start[1]
    while (.end <= length(.lines) && !grepl("^\\}", .lines[.end])) .end <- .end + 1L
    .lines <- .lines[-seq(.cb, .end)]
  }
  .out <- file(.path, "wb")
  writeLines(.lines, .out, sep = "\n")
  close(.out)
}
.strip_sens("src/sundials_sunnonlinsol_newton.c",    "SUNNonlinSol_NewtonSens")
.strip_sens("src/sundials_sunnonlinsol_fixedpoint.c","SUNNonlinSol_FixedPointSens")

## ---------------------------------------------------------------------------
## Replace sprintf(name, "...") with snprintf(name, N, "...") so CRAN's
## _FORTIFY_SOURCE check does not flag ___sprintf_chk in the compiled .so.
## The buffer size N matches the malloc(N * sizeof(char)) call in each file.
## ---------------------------------------------------------------------------

.fix_sprintf <- function(.path, .bufsz) {
  if (!file.exists(.path)) return(invisible(NULL))
  .lines <- readLines(.path)
  .lines <- gsub(
    sprintf("\\bsprintf\\(name,"),
    sprintf("snprintf(name, %d,", .bufsz),
    .lines, perl = TRUE
  )
  .out <- file(.path, "wb")
  writeLines(.lines, .out, sep = "\n")
  close(.out)
  message("Fixed sprintf -> snprintf (buf=", .bufsz, ") in ", .path)
}

.fix_sprintf("src/sundials_cvode_diag.c", 30L)
.fix_sprintf("src/sundials_cvode_io.c",   24L)
.fix_sprintf("src/sundials_cvode_ls.c",   30L)

## ---------------------------------------------------------------------------
## malloc -> calloc in the *NewEmpty constructors. If the headers ever gain
## struct fields the vendored sources do not know about, calloc leaves them
## NULL (handled by the existing 'if (ops->foo)' guards) instead of garbage.
## ---------------------------------------------------------------------------

.fix_calloc <- function(.path) {
  if (!file.exists(.path)) return(invisible(NULL))
  .lines <- readLines(.path)
  .new <- gsub("malloc\\(sizeof \\*([A-Za-z_]+)\\)", "calloc(1, sizeof *\\1)", .lines)
  .new <- gsub("malloc\\(sizeof\\(struct (SUNMemoryHelper_Ops_|SUNMemoryHelper_|SUNMemory_)\\)\\)",
               "calloc(1, sizeof(struct \\1))", .new)
  if (!identical(.new, .lines)) {
    .out <- file(.path, "wb")
    writeLines(.new, .out, sep = "\n")
    close(.out)
    message("Fixed malloc -> calloc in ", .path)
  }
}
for (.f in c("src/sundials_sundials_nonlinearsolver.c",
             "src/sundials_sundials_linearsolver.c",
             "src/sundials_sundials_matrix.c",
             "src/sundials_sundials_nvector.c",
             "src/sundials_sundials_memory.c",
             "src/sundials_sunnonlinsol_newton.c",
             "src/sundials_sunlinsol_dense.c")) {
  .fix_calloc(.f)
}

message("Done. Review 'git diff src/' and commit any updated files.")

#' Turn into an ini block for initialization
#'
#' @param x Item to convert to a rxode2/nlmixr2 ui ini expression
#' @return rxode2 ini expression
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' ini <- quote(ini({
#'   tka <- log(1.57)
#'   tcl <- log(2.72)
#'   tv <- log(31.5)
#'   eta.ka ~ 0.6
#'   eta.cl ~ 0.3
#'   eta.v ~ 0.1
#'   add.sd <- 0.7
#' }))
#'
#' as.ini(ini)
#'
#' l <- quote(lotri({
#'   tka <- log(1.57)
#'   tcl <- log(2.72)
#'   tv <- log(31.5)
#'   eta.ka ~ 0.6
#'   eta.cl ~ 0.3
#'   eta.v ~ 0.1
#'   add.sd <- 0.7
#'  }))
#'
#' as.ini(l)
#'
#' m <- lotri({
#'    eta.ka ~ 0.6
#'    eta.cl ~ 0.3
#'    eta.v ~ 0.1
#' })
#'
#' as.ini(m)
#'
#' one.compartment <- function() {
#'   ini({
#'     tka <- log(1.57)
#'     tcl <- log(2.72)
#'     tv <- log(31.5)
#'     eta.ka ~ 0.6
#'     eta.cl ~ 0.3
#'     eta.v ~ 0.1
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl + eta.cl)
#'     v <- exp(tv + eta.v)
#'     d/dt(depot) = -ka * depot
#'     d/dt(center) = ka * depot - cl / v * center
#'     cp = center / v
#'     cp ~ add(add.sd)
#'   })
#' }
#'
#' as.ini(one.compartment)
#'
#' ui <- one.compartment()
#'
#' as.ini(ui)
#'
#' ui$iniDf
#'
#' as.ini(ui$iniDf)
#'
#' ini <- c("ini({",
#'           "tka <- log(1.57)",
#'           "tcl <- log(2.72)",
#'           "tv <- log(31.5)",
#'           "eta.ka ~ 0.6",
#'           "eta.cl ~ 0.3",
#'           "eta.v ~ 0.1",
#'           "add.sd <- 0.7",
#'           "})")
#'
#' as.ini(ini)
#'
#' ini <- paste(ini, collapse="\n")
#'
#' as.ini(ini)
#'
as.ini <- function(x) {
  UseMethod("as.ini")
}

#' @rdname as.ini
#' @export
as.ini.character <- function(x) {
  as.ini(str2lang(paste(x, collapse="\n")))
}

#' @rdname as.ini
#' @export
as.ini.data.frame <- function(x) {
 lotri::lotriDataFrameToLotriExpression(x, useIni = TRUE)
}

#' @rdname as.ini
#' @export
as.ini.call <- function(x) {
  if (!identical(x[[1]], quote(`ini`)) &&
        !identical(x[[1]], quote(`lotri`))) {
    stop("unsupported expression for ini({}) block",
         call.=FALSE)
  }
  if (identical(x[[1]], quote(`ini`))) {
    .tmp <- x
    .tmp[[1]] <- quote(`lotri`)
    .tmp <- eval(.tmp)
    return(as.ini(.tmp))
  } else if (identical(x[[1]], quote(`lotri`))) {
    .tmp <- eval(x)
    return(as.ini(.tmp))
  }
  stop("unsupported expression of ini({}) block",
       call.=FALSE)
}

#' @rdname as.ini
#' @export
as.ini.lotriFix <- function(x) {
  .ret <- as.expression(x)
  .ret[[1]] <- quote(`ini`)
  return(.ret)
}

#' @rdname as.ini
#' @export
as.ini.matrix <- function(x) {
  .ret <- x
  class(.ret) <- c("lotriFix", class(.ret))
  as.ini(.ret)
}

#' @rdname as.ini
#' @export
as.ini.default <- function(x) {
  if (is.null(x)) {
    return(quote(ini({}))) # nolint
  }
  .ini <- try(as.rxUi(x), silent=TRUE)
  if (inherits(.ini, "try-error")) {
    stop("do not know how to convert this to an `ini` expression",
         call.=FALSE)
  }
  ini(.ini)
}

#' Variance-covariance (non-Cholesky) Omega parameterization derivatives
#'
#' A non-Cholesky `Omega` path that differentiates with respect to the
#' variance-covariance entries directly (for reporting SEs on the natural scale
#' or building an analytic covariance over the `Omega` elements).  Returns
#' `Omega^{-1}`, `log|Omega|`, and their first (and optionally second)
#' derivatives with respect to each free lower-triangular element
#' `omega_{ab}`, via
#'
#' \deqn{\partial \Omega^{-1}/\partial \omega_{ab} = -\Omega^{-1} E_{ab} \Omega^{-1}}
#' \deqn{\partial \log|\Omega|/\partial \omega_{ab} = \mathrm{tr}(\Omega^{-1} E_{ab})}
#'
#' where \eqn{E_{ab}} is the symmetric single-entry basis matrix.
#'
#' @param omega symmetric positive-definite random-effects covariance matrix.
#' @param order integer; `1` for first derivatives only, `2` (default) to also
#'   return the second derivatives needed for the covariance Hessian.
#' @return a list with `omegaInv`, `logDet`, the free-element index matrix
#'   `elements` (each row `c(a, b)`, `a >= b`), first derivatives
#'   `dOmegaInv` / `dLogDet`, and (when `order = 2`) second derivatives
#'   `d2OmegaInv` / `d2LogDet`.
#' @author Hidde van de Beek
#' @export
rxOmegaVarCovDeriv <- function(omega, order = 2L) {
  omega <- as.matrix(omega)
  storage.mode(omega) <- "double"
  # Heavy lifting (the E-basis matrix calculus) is done in C++/RcppArmadillo so
  # this can be called from the analytic-covariance C path; see src/omegaVarCov.cpp.
  rxOmegaVarCovDeriv_(omega, as.integer(order))
}

rxTest({
  # Non-Cholesky (variance-covariance) Omega derivatives must match finite
  # differences, including off-diagonal (block) elements.
  L <- matrix(c(0.6, 0, 0,  0.1, 0.4, 0,  -0.05, 0.08, 0.3), 3, byrow = TRUE)
  Om <- L %*% t(L)
  r <- rxOmegaVarCovDeriv(Om, order = 2L)
  np <- nrow(r$elements)
  perturb <- function(O, k, hh) { a <- r$elements[k, 1]; b <- r$elements[k, 2]
    O[a, b] <- O[a, b] + hh; if (a != b) O[b, a] <- O[b, a] + hh; O }
  h <- 1e-5

  test_that("first derivatives of Omega^-1 and log|Omega| match FD", {
    e1Oi <- 0; e1ld <- 0
    for (k in seq_len(np)) {
      fdOi <- (solve(perturb(Om, k, h)) - solve(perturb(Om, k, -h))) / (2 * h)
      e1Oi <- max(e1Oi, max(abs(r$dOmegaInv[[k]] - fdOi)))
      fdld <- (determinant(perturb(Om, k, h), logarithm = TRUE)$modulus -
               determinant(perturb(Om, k, -h), logarithm = TRUE)$modulus) / (2 * h)
      e1ld <- max(e1ld, abs(r$dLogDet[k] - as.numeric(fdld)))
    }
    expect_lt(e1Oi, 1e-4)
    expect_lt(e1ld, 1e-4)
  })

  test_that("second derivative of log|Omega| matches FD", {
    e2 <- 0
    for (j in seq_len(np)) for (k in seq_len(np)) {
      pp <- perturb(perturb(Om, j, h), k, h); pm <- perturb(perturb(Om, j, h), k, -h)
      mp <- perturb(perturb(Om, j, -h), k, h); mm <- perturb(perturb(Om, j, -h), k, -h)
      fd <- (as.numeric(determinant(pp, logarithm = TRUE)$modulus) -
             as.numeric(determinant(pm, logarithm = TRUE)$modulus) -
             as.numeric(determinant(mp, logarithm = TRUE)$modulus) +
             as.numeric(determinant(mm, logarithm = TRUE)$modulus)) / (4 * h * h)
      e2 <- max(e2, abs(r$d2LogDet[j, k] - fd))
    }
    expect_lt(e2, 1e-3)
  })
})

# rxEt internal methods ---------------------------------------------------

#' This is the $add.dosing type of argument to the event tables
#'
#' @param env The metadata environment for the event table, which is
#'   modified by reference
#'
#' @param dose numeric dose amount
#'
#' @param nbr.doses number of doses (1 means just the initial dose at
#'   start.time)
#'
#' @param dosing.interval dosing interval in time units (ignored if
#'   nbr.doses is 1)
#'
#' @param dosing.to compartment number to dose to (default 1); ignored
#'   if cmt column is provided in ...
#'
#' @param rate numeric infusion rate (amount/time); if provided, dur is ignored and
#'  dur is calculated as amt/rate
#'
#' @param amount.units character dose unit, e.g. "mg"; if provided,
#'   overrides existing amount unit
#'
#' @param start.time numeric time of first dose (default 0)
#'
#' @param do.sampling logical; if TRUE, adds sampling records at each
#'   dosing time (default FALSE)
#'
#' @param time.units character time unit, e.g. "hours"; if provided,
#'   overrides existing time unit
#'
#' @param evid integer event ID to use for dosing records; if NULL,
#'   defaults to 1 for non-zero dose and 0 for zero dose
#' @param strt.time deprecated alias for start.time
#' @param ... additional arguments passed to et() when creating the dosing records
#'
#' @return nothing, called for side effect of modifying env by reference
#' @noRd
#' @author Matthew L. Fidler
.etMethodAddDosing <- function(env, dose, nbr.doses = 1L, dosing.interval = 24,
                               dosing.to = 1L, rate = NULL,
                               amount.units = NA_character_,
                               start.time = 0.0, do.sampling = FALSE,
                               time.units = NA_character_,
                               evid = NULL, strt.time = NULL, ...) {
  if (!is.null(strt.time)) start.time <- strt.time
  .et <- structure(list(env = env), class = "rxEt")
  .args <- list(
    x = .et,
    amt = dose,
    time = start.time,
    ii = if (nbr.doses > 1L) dosing.interval else 0.0,
    addl = as.integer(nbr.doses) - 1L,
    addSampling = do.sampling
  )
  if (!is.null(rate)) .args$rate <- rate
  if (!is.na(amount.units)) .args$amountUnits <- amount.units
  if (!is.na(time.units)) .args$timeUnits <- time.units
  if (!is.null(evid)) .args$evid <- evid
  .extra <- list(...)
  if (is.null(.extra$cmt) && is.null(.extra$dosing.to) &&
        !identical(dosing.to, 1L)) {
    .args$cmt <- dosing.to
  }
  .ret <- do.call(et, c(.args, .extra)) # nolint
  .retEnv <- .rxEtEnv(.ret)
  env$chunks <- .retEnv$chunks
  env$groups <- .retEnv$groups
  env$units <- .retEnv$units
  env$show <- .retEnv$show
  env$ids <- .retEnv$ids
  env$nobs <- .retEnv$nobs
  env$ndose <- .retEnv$ndose
  env$randomType <- .retEnv$randomType
  env$canResize <- .retEnv$canResize
  .etAddExtraCols(env, .etExtraCols(.retEnv)) # nolint
  invisible(NULL)
}

#' $get.EventTable()
#'
#' @param env environment to get the event table
#' @return event table
#' @noRd
#' @author Matthew L. Fidler
.etMethodGetEventTable <- function(env) {
  if (.rxGetHomogenous()) { # nolint
    .mat <- .etPreviewData(env, "all")
  } else {
    .mat <- .etMaterialize(structure(list(env = env), class = "rxEt")) # nolint
  }
  if (is.null(.mat) || nrow(.mat) == 0L) return(NULL)
  .cols <- .etDisplayCols(names(.mat), env$show, .etExtraCols(env)) # nolint
  .ret <- .mat[, .cols, drop = FALSE]
  rownames(.ret) <- seq_len(nrow(.ret))
  .ret
}

#' Clear dosing method
#'
#' @param env environment to clear dosing from
#' @return data frame with dosing cleared
#' @noRd
#' @author Matthew L. Fidler
.etMethodGetDosing <- function(env) {
  if (.rxGetHomogenous()) { # nolint
    .d <- .etPreviewData(env, "dosing")
  } else {
    .full <- .etMaterialize(structure(list(env = env), class = "rxEt")) # nolint
    if (is.null(.full) || nrow(.full) == 0L) return(NULL)
    .d <- .full[.full$evid != 0L, , drop = FALSE]
  }
  if (is.null(.d)) return(NULL)
  if (nrow(.d) == 0L) {
    NULL
  } else {
    rownames(.d) <- seq_len(nrow(.d))
    .etMarkDisplay(.d, env) # nolint
  }
}

#' This clears the sampling in the attached event table
#'
#' Called with $clear.sampling()
#'
#' @param env environment for clearing the samples
#' @return nothing, called for side effect of modifying env by reference
#' @noRd
#' @author Matthew L. Fidler
.etMethodClearSampling <- function(env) {
  .groups <- .etGetGroups(env) # nolint
  if (length(.groups) > 0L) {
    .groups <- Filter(Negate(is.null), lapply(.groups, function(.g) {
      .df <- .g$data[.g$data$evid != 0L, , drop = FALSE]
      if (nrow(.df) == 0L) return(NULL)
      list(ids = .g$ids, data = .df)
    }))
    .etSetGroups(env, .groups) # nolint
    .etResetCountsFromGroups(env) # nolint
    return(invisible(NULL))
  }
  for (.i in seq_along(env$chunks)) {
    if (!is.null(env$chunks[[.i]])) {
      .df <- env$chunks[[.i]]
      .df <- .df[.df$evid != 0L, , drop = FALSE]
      if (nrow(.df) == 0L) env$chunks[.i] <- list(NULL) else env$chunks[[.i]] <- .df
    }
  }
  env$nobs <- 0L
  invisible(NULL)
}

# rxEt import helpers -----------------------------------------------------

#' Auto-detect time and dosing units from columns with units class
#'
#' @param env environment to update with detected units (modified by reference)
#' @param df data.frame to check for units columns
#'
#' @return list with detected time and dosing units and flags
#'   indicating if they were already set in env or auto-detected from
#'   columns
#' @noRd
#' @author Matthew L. Fidler
.etImportAutoUnits <- function(env, df) {
  .tu <- env$units["time"]
  .du <- env$units["dosing"]
  .hasTimeU <- !is.na(.tu) && nchar(.tu) > 0
  .hasDoseU <- !is.na(.du) && nchar(.du) > 0
  # Auto-detect units from columns if not already set
  if (requireNamespace("units", quietly = TRUE)) {
    if (!.hasTimeU) {
      for (.nmCheck in c("time", "ii")) {
        if (!is.null(df[[.nmCheck]]) && inherits(df[[.nmCheck]], "units")) {
          .tu <- units::deparse_unit(df[[.nmCheck]])
          env$units["time"] <- .tu
          .hasTimeU <- TRUE
          break
        }
      }
    }
    if (!.hasDoseU) {
      if (!is.null(df[["amt"]]) && inherits(df[["amt"]], "units")) {
        .du <- units::deparse_unit(df[["amt"]])
        env$units["dosing"] <- .du
        .hasDoseU <- TRUE
      } else if (!is.null(df[["rate"]]) && inherits(df[["rate"]], "units")) {
        # udunits format: "ug s-1" - amount unit has no digit exponent
        .rateStr <- units::deparse_unit(df[["rate"]])
        .parts <- strsplit(trimws(.rateStr), "\\s+")[[1L]]
        .amtParts <- .parts[!grepl("[0-9]", .parts)]
        if (length(.amtParts) >= 1L) {
          .du <- .amtParts[[1L]]
          env$units["dosing"] <- .du
          .hasDoseU <- TRUE
        }
      }
    }
  }
  list(tu = .tu, du = .du, hasTimeU = .hasTimeU, hasDoseU = .hasDoseU)
}

#' Import an event table from a data.frame, converting units if needed
#'
#' @param df data.frame to import; may have units class on time, ii,
#'   amt, and/or rate columns
#'
#' @param tu time unit to convert to (if df has time/ii columns with units class)
#'
#' @param du dose unit to convert to (if df has amt and/or rate columns with units class)
#'
#' @param hasTimeU logical indicating if time unit was already set in
#'   env or auto-detected from df
#'
#' @param hasDoseU logical indicating if dose unit was already set in
#'   env or auto-detected from df
#'
#' @return data.frame with units columns converted to numeric and time/amt/rate
#'
#' @noRd
#' @author Matthew L. Fidler
.etImportConvertUnits <- function(df, tu, du, hasTimeU, hasDoseU) {
  .cols <- lapply(names(df), function(.nm) {
    .col <- df[[.nm]]
    if (!inherits(.col, "units")) return(.col)
    if (requireNamespace("units", quietly = TRUE)) {
      # Only convert when et had prior units; auto-detected units just strip label
      if (.nm %in% c("time", "ii") && hasTimeU) {
        return(as.numeric(units::set_units(.col, tu, mode = "standard")))
      }
      if (.nm == "amt" && hasDoseU) {
        return(as.numeric(units::set_units(.col, du, mode = "standard")))
      }
      if (.nm == "rate" && hasDoseU && hasTimeU) {
        .rateU <- paste0(du, "/", tu)
        return(as.numeric(units::set_units(.col, .rateU, mode = "standard")))
      }
    }
    as.numeric(.col)
  })
  names(.cols) <- names(df)
  as.data.frame(.cols, stringsAsFactors = FALSE)
}

#' This standardizes the id and evid columns in the imported data.frame
#'
#' @param df input data.frame
#' @return data.frame with standardized id and evid columns
#' @noRd
#' @author Matthew L. Fidler
.etImportStandardizeIdEvid <- function(df) {
  if (is.null(df$evid)) {
    df$evid <- if (!is.null(df$amt)) ifelse(!is.na(df$amt) & as.numeric(df$amt) != 0, 1L, 0L) else 0L
  }
  df$evid <- as.integer(df$evid)
  if (is.null(df$id)) df$id <- 1L
  df$id <- as.integer(df$id)
  df
}

#' Import, updating which columns are shown
#'
#' @param env environment to update with new data and show flags (modified by reference)
#'
#' @param df data.frame to import, with standardized id and evid columns
#'
#' @return nothing, called for side effect of modifying env by reference
#'
#' @noRd
#'
#' @author Matthew L. Fidler
#'
.etImportUpdateShow <- function(env, df) {
  if (length(env$ids) > 1L) env$show["id"] <- TRUE
  if (sum(df$evid != 0L, na.rm = TRUE) > 0L) env$show["amt"] <- TRUE
  if (!is.null(df$rate) && any(df$rate[df$evid != 0L] != 0, na.rm = TRUE))
    env$show["rate"] <- TRUE
  if (!is.null(df$dur) && any(df$dur[df$evid != 0L] != 0, na.rm = TRUE))
    env$show["dur"] <- TRUE
  if (!is.null(df$ii) && any(df$ii != 0, na.rm = TRUE)) {
    env$show["ii"] <- TRUE
    env$show["addl"] <- TRUE
  }
}

#' Import and update the environment
#'
#' @param env environment to update with new data and show flags
#'   (modified by reference)
#' @param df data.frame to import, with standardized id and evid
#'   columns and units converted to numeric
#' @return nothing, called for side effect of modifying env by reference
#' @noRd
#' @author Matthew L. Fidler
.etImportUpdateEnv <- function(env, df) {
  env$ids  <- sort(unique(df$id))
  env$nobs  <- env$nobs  + sum(df$evid == 0L, na.rm = TRUE)
  env$ndose <- env$ndose + sum(df$evid != 0L, na.rm = TRUE)
  env$groups <- list()
  env$chunks <- .addRowsToChunks(env$chunks, df)
  .etImportUpdateShow(env, df)
}

#' Import the event tables and normalize column names to lowercase,
#' converting NONMEM-style uppercase
#'
#' @param df data.frame to normalize column names
#' @return normalized data frame
#' @noRd
#' @author Matthew L. Fidler
.etImportNormalizeNames <- function(df) {
  # Normalize UPPERCASE/mixed-case NONMEM-style column names to lowercase
  .colMap <- c(ID="id", TIME="time", CMT="cmt", AMT="amt", EVID="evid",
               RATE="rate", II="ii", ADDL="addl", SS="ss", DUR="dur",
               LOW="low", HIGH="high")
  .nms <- names(df)
  .upper <- toupper(.nms)
  for (.i in seq_along(.nms)) {
    if (.upper[.i] %in% names(.colMap) && .nms[.i] != .colMap[.upper[.i]]) {
      names(df)[.i] <- .colMap[.upper[.i]]
    }
  }
  # rename the assigned-column tag with the columns it points at (#1154)
  .extra <- .etExtraColsAttr(df) # nolint
  if (length(.extra) > 0L) {
    .map <- stats::setNames(names(df), .nms)[.extra]
    attr(df, "rxEtExtraCols") <- unname(ifelse(is.na(.map), .extra, .map))
  }
  df
}

#' Import ID column as integer, converting from character if needed
#'
#' @param df data.frame to import the ID column from
#' @return data.frame with ID column converted to integer if it was character
#' @noRd
#' @author Matthew L. Fidler
.etImportIdToInteger <- function(df) {
  # Convert character ID to sequential integers (silently)
  if (!is.null(df$id) && !is.numeric(df$id) && !is.integer(df$id)) {
    .uniq <- unique(df$id)
    df$id <- as.integer(factor(df$id, levels = .uniq))
  }
  df
}

#' Drop rows with NA time, warning about how many were dropped
#'
#' @param df input data frame
#' @return data frame with NA time dropped
#' @noRd
#' @author Matthew L. Fidler
.etImportDropNaTime <- function(df) {
  # Drop rows with NA time (warn)
  if (!is.null(df$time) && any(is.na(as.numeric(df$time)))) {
    .nDrop <- sum(is.na(as.numeric(df$time)))
    warning(sprintf("dropping %d row(s) with NA time", .nDrop), call. = FALSE)
    df <- df[!is.na(as.numeric(df$time)), , drop = FALSE]
  }
  df
}

#' Import a data frame
#'
#' @param env environment to update with new data and show flags
#'   (modified by reference)
#'
#' @param df data.frame to import, with standardized id and evid
#'   columns and units converted to numeric
#'
#' @return nothing, called for side effects
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.etMethodImportEventTable <- function(env, df) {
  if (!is.data.frame(df)) {
    stop("'df' must be a data.frame", call. = FALSE)
  }
  .extra <- .etExtraColsAttr(df) # nolint
  .u <- .etImportAutoUnits(env, df)
  df <- .etImportConvertUnits(df, .u$tu, .u$du, .u$hasTimeU, .u$hasDoseU)
  df <- .etImportStandardizeIdEvid(df)
  .etImportUpdateEnv(env, df)
  .etAddExtraCols(env, intersect(.extra, names(df))) # nolint
  invisible(NULL)
}

#'  Import an event table from a data.frame.
#'
#' This method auto-detects units and normalization of column names
#' and id/evid columns; delegates to .etMethodImportEventTable after
#' processing
#'
#' @param env environment to update with new data and show flags
#'   (modified by reference)
#'
#' @param df data.frame to import
#'
#' @param ... additional arguments (ignored)
#'
#' @return nothing, called for side effects
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.etMethodImportEventTable2 <- function(env, df, ...) {
  if (!is.data.frame(df)) stop("'df' must be a data.frame", call. = FALSE)
  .extra <- .etExtraColsAttr(df) # nolint
  df <- as.data.frame(df)
  # as.data.frame() may drop it; re-tag first so the rename below tracks it
  if (length(.extra) > 0L) attr(df, "rxEtExtraCols") <- .extra
  df <- .etImportNormalizeNames(df)
  df <- .etImportIdToInteger(df)
  df <- .etImportDropNaTime(df)
  # Delegate to import.EventTable for units handling + storage
  .etMethodImportEventTable(env, df)
  invisible(NULL)
}

# rxEt query and mutation methods -----------------------------------------

#' $expand() method
#'
#' @param env environment to update with expanded data and show flags
#'   (modified by reference)
#'
#' @return nothing, called for side effect of modifying env by reference
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.etMethodExpand <- function(env) {
  .groups <- .etGetGroups(env) # nolint
  if (length(.groups) > 0L) {
    env$groups <- lapply(.groups, function(.g) {
      list(ids = as.integer(.g$ids), data = .etExpandGroupData(.g$data, env)) # nolint
    })
    env$chunks <- list()
  } else {
    .et <- structure(list(env = env), class = "rxEt")
    .mat <- .etMaterialize(.et)
    .expanded <- .etExpandAddl(.mat, env)
    env$groups <- list()
    env$chunks <- list()
    if (nrow(.expanded) > 0L) {
      .ids <- unique(as.integer(.expanded$id))
      for (.i in .ids) {
        env$chunks[[.i]] <- .expanded[.expanded$id == .i, , drop = FALSE]
      }
      env$ids <- sort(.ids)
    } else {
      env$ids <- 1L
    }
  }
  .etResetCountsFromGroups(env) # nolint
  if (length(.etGroups(env)) == 0L) { # nolint
    env$nobs <- sum(.expanded$evid == 0L, na.rm = TRUE)
    env$ndose <- sum(.expanded$evid != 0L, na.rm = TRUE)
  }
  env$show["id"] <- length(env$ids) > 1L
  if (length(.etGroups(env)) > 0L) { # nolint
    env$show["addl"] <- any(vapply(env$groups, function(.g) {
      !is.null(.g$data$addl) && any(.g$data$addl != 0L, na.rm = TRUE)
    }, logical(1)))
  } else {
    env$show["addl"] <- !is.null(.expanded$addl) && any(.expanded$addl != 0L, na.rm = TRUE)
  }
  env$randomType <- NA_integer_
  env$canResize <- FALSE
  invisible(NULL)
}

#' This get the observation records (evid == 0) from the event table
#'
#' @param env  environment to get the observation records from
#' @return logical of observation records (evid == 0)
#' @noRd
#' @author Matthew L. Fidler
.etMethodGetObsRec <- function(env) {
  .mat <- .etMaterialize(structure(list(env = env), class = "rxEt"))
  .mat$evid == 0L
}

#' $copy() method
#'
#' Creates a new rxEt object with a copy of the environment and methods, so
#' that mutations to the new copy do not affect the original.
#'
#' @param env environment to copy from
#' @return new rxEt object with copied environment and methods
#' @noRd
#' @author Matthew L. Fidler
.etMethodCopy <- function(env) {
  .newEnv <- new.env(parent = emptyenv())
  .newEnv$chunks     <- env$chunks
  .newEnv$groups     <- env$groups
  .newEnv$units      <- env$units
  .newEnv$show       <- env$show
  .newEnv$ids        <- env$ids
  .newEnv$nobs       <- env$nobs
  .newEnv$ndose      <- env$ndose
  .newEnv$randomType <- env$randomType
  .newEnv$canResize  <- env$canResize
  .newEnv$extraCols  <- .etExtraCols(env) # nolint
  .newEnv$methods <- .etBuildMethods(.newEnv)
  .cp <- list()
  attr(.cp, "names")     <- character(0)
  attr(.cp, "class")     <- c("rxEt", "data.frame")
  attr(.cp, "row.names") <- integer(0)
  attr(.cp, ".rxEtEnv")  <- .newEnv
  .cp
}

#'  $simulate() method
#'
#' @param env environment to update with simulated data (modified by
#'   reference)
#'
#' @param seed random seed for reproducibility (currently ignored, as
#'   simulation is done in-place on the event table)
#'
#' @param ... additional arguments passed to the simulation method
#'   (currently ignored, as simulation is done in-place on the event
#'   table)
#'
#' @return nothing, called for side effects
#'
#' @noRd
#' @author Matthew L. Fidler
.etMethodSimulate <- function(env, seed = NULL, ...) {
  if (!is.null(seed)) set.seed(seed)
  .sim <- .etSimulateRepresentation(env) # nolint
  if (!isTRUE(.sim$hasWin)) {
    env$groups <- .sim$groups
    env$chunks <- .sim$chunks
    env$randomType <- NA_integer_
    env$canResize <- FALSE
    warning("simulating event table without windows returns identical event table", call. = FALSE)
  } else {
    env$groups <- .sim$groups
    env$chunks <- .sim$chunks
    env$randomType <- NA_integer_
    env$canResize <- FALSE
  }
  invisible(NULL)
}

.etMethodClearDosing <- function(env) {
  .groups <- .etGetGroups(env) # nolint
  if (length(.groups) > 0L) {
    .groups <- Filter(Negate(is.null), lapply(.groups, function(.g) {
      .df <- .g$data[.g$data$evid == 0L, , drop = FALSE]
      if (nrow(.df) == 0L) return(NULL)
      list(ids = .g$ids, data = .df)
    }))
    .etSetGroups(env, .groups) # nolint
    .etResetCountsFromGroups(env) # nolint
    if (env$ndose == 0L) {
      env$show[c("amt", "rate", "ii", "addl", "ss", "dur")] <- FALSE
    }
    return(invisible(NULL))
  }
  for (.i in seq_along(env$chunks)) {
    if (!is.null(env$chunks[[.i]])) {
      .df <- env$chunks[[.i]]
      .df <- .df[.df$evid == 0L, , drop = FALSE]
      if (nrow(.df) == 0L) env$chunks[.i] <- list(NULL) else env$chunks[[.i]] <- .df
    }
  }
  env$ndose <- 0L
  env$show[c("amt", "rate", "ii", "addl", "ss", "dur")] <- FALSE
  invisible(NULL)
}

.etMethodGetSampling <- function(env) {
  if (.rxGetHomogenous()) { # nolint
    .s <- .etPreviewData(env, "sampling")
  } else {
    .full <- .etMaterialize(structure(list(env = env), class = "rxEt")) # nolint
    if (is.null(.full) || nrow(.full) == 0L) return(NULL)
    .s <- .full[.full$evid == 0L, , drop = FALSE]
  }
  if (is.null(.s)) return(NULL)
  if (nrow(.s) == 0L) {
    NULL
  } else {
    rownames(.s) <- seq_len(nrow(.s))
    .etMarkDisplay(.s, env) # nolint
  }
}

.etMethodAddSampling <- function(env, time, time.units = NA_character_) {
  .time <- time
  if (is.list(.time)) {
    .time <- lapply(.time, function(.window) {
      if (inherits(.window, "units") && requireNamespace("units", quietly = TRUE)) {
        .tu <- env$units["time"]
        if (!is.na(.tu) && nchar(.tu) > 0) {
          return(as.numeric(units::set_units(.window, .tu, mode = "standard")))
        }
        return(as.numeric(.window))
      }
      .window
    })
  } else if (inherits(.time, "units") && requireNamespace("units", quietly = TRUE)) {
    .tu <- env$units["time"]
    if (!is.na(.tu) && nchar(.tu) > 0) {
      .time <- as.numeric(units::set_units(.time, .tu, mode = "standard"))
    } else {
      .time <- as.numeric(.time)
    }
  }
  .df <- .etObsChunk(.time)
  .etAddChunk(env, .df, env$ids)
  env$nobs   <- env$nobs + length(.df$time) * length(env$ids)
  if (!is.na(time.units)) env$units["time"] <- time.units
  invisible(NULL)
}

# rxEt method registry ----------------------------------------------------

#' Attach method functions to an rxEt\'s list structure
#'
#' Each method captures env by reference so mutations are shared.
#'
#' @param env environment
#'
#' @return named list of functions
#'
#' @noRd
.etBuildMethods <- function(env) {
  .lst <- list(
    add.dosing = function(dose, nbr.doses = 1L, dosing.interval = 24,
                          dosing.to = 1L, rate = NULL,
                          amount.units = NA_character_,
                          start.time = 0.0, do.sampling = FALSE,
                          time.units = NA_character_,
                          evid = NULL, strt.time = NULL, ...) {
      .etMethodAddDosing(env, dose, nbr.doses, dosing.interval, dosing.to, rate,
                         amount.units, start.time, do.sampling, time.units,
                         evid, strt.time, ...)
    },

    add.sampling = function(time, time.units = NA_character_) {
      .etMethodAddSampling(env, time, time.units)
    },

    get.units = function() env$units,
    getUnits  = function() env$units,

    get.nobs  = function() env$nobs,
    get.EventTable = function() {
      .etMethodGetEventTable(env)
    },
    get.obs.rec = function() {
      .etMethodGetObsRec(env)
    },
    get.dosing = function() {
      .etMethodGetDosing(env)
    },
    get.sampling = function() {
      .etMethodGetSampling(env)
    },
    clear.sampling = function() {
      .etMethodClearSampling(env)
    },
    clear.dosing = function() {
      .etMethodClearDosing(env)
    },
    copy = function() {
      .etMethodCopy(env)
    },
    import.EventTable = function(df) {
      .etMethodImportEventTable(env, df)
    },

    importEventTable = function(df, ...) {
      .etMethodImportEventTable2(env, df, ...)
    },

    expand = function() {
      .etMethodExpand(env)
    },
    simulate = function(object, nsim = 1, seed = NULL, ...) {
      .etMethodSimulate(env, seed, ...)
    }
  )
  .lst$addDosing      <- .lst[["add.dosing"]]
  .lst$add_dosing     <- .lst[["add.dosing"]]
  .lst$addSampling    <- .lst[["add.sampling"]]
  .lst$add_sampling   <- .lst[["add.sampling"]]
  .lst$getDosing      <- .lst[["get.dosing"]]
  .lst$getSampling    <- .lst[["get.sampling"]]
  .lst$getEventTable  <- .lst[["get.EventTable"]]
  .lst$clearDosing    <- .lst[["clear.dosing"]]
  .lst$clear_dosing   <- .lst[["clear.dosing"]]
  .lst$clearSampling  <- .lst[["clear.sampling"]]
  .lst$clear_sampling <- .lst[["clear.sampling"]]
  .lst$get_units      <- .lst[["get.units"]]
  .lst
}

rxTest({
  ## rxode2 instance 1
  m1 <-
    rxode2(
      model = "
         C2 = centr/V2;
         C3 = peri/V3;
         d/dt(depot) =-KA*depot;
         d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
         d/dt(peri)  =                    Q*C2 - Q*C3;
         d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;"
    )

  test_that("rxode2 instance 1 is created", {
    expect_s3_class(m1, "rxode2")
  })

  et1 <- eventTable(amount.units = "ug", time.units = "hours")
  et1$add.dosing(dose = 10000, nbr.doses = 5, dosing.interval = 24)
  et1$add.sampling(0:24)
  et1$add.sampling(24)
  et1$add.sampling(seq(from = 24 + 8, to = 5 * 24, by = 8))

  test_that("rxode2 event table 1 was created", {
    expect_s3_class(et1, "rxEt")
    expect_equal(et1$get.nobs(), 38)
    expect_equal(length(et1$get.dosing()[, 1]), 1)
  })

  o1.first <-
    suppressWarnings(rxSolve(
      m1,
      params = c(
        KA = .291, CL = 18.6, V2 = 40.2, Q = 10.5, V3 = 297.0,
        Kin = 1.0, Kout = 1.0, EC50 = 200.0
      ),
      events = et1,
      inits = c(0, 0, 0, 1)
    ))

  o1.df <- as.data.frame(o1.first)
  o1.df2 <- tibble::as_tibble(o1.first)

  test_that("Numeric Data frame lookup operators [] make sense", {
    expect_equal(o1.first[], o1.df[])
    expect_equal(o1.first[, 3], o1.df[, 3])
    expect_equal(o1.first[, 3, drop = FALSE], o1.df[, 3, drop = FALSE])
    expect_equal(o1.first[3, ], o1.df[3, ])
    expect_equal(o1.first[, c(1, 3)], o1.df[, c(1, 3)])
    expect_equal(o1.first[c(1, 3), ], o1.df[c(1, 3), ])
    expect_equal(o1.first[1, 3], o1.df[1, 3])
    expect_equal(o1.first[c(1, 3), c(1, 3)], o1.df[c(1, 3), c(1, 3)])
  })

  test_that("as_data_frame produces reasonable results.", {
    expect_equal(tibble::as_tibble(o1.df), o1.df2)
  })

  test_that("Character data frame lookup operators [] make sense", {
    expect_equal(o1.first[, "centr"], o1.df[, "centr"])
    expect_equal(o1.first[, "centr", drop = FALSE], o1.df[, "centr", drop = FALSE])
    expect_equal(o1.first[, c("centr", "depot")], o1.df[, c("centr", "depot")])
  })

  test_that("Character data frame assignment operators [] make sense", {
    o1.assign <- o1.first
    expect_s3_class(o1.assign, "data.frame")
    o1.assign[, "depot"] <- 0
    expect_equal(rep(0, times = length(as.data.frame(o1.assign)[, 1])), as.data.frame(o1.assign)[, "depot"])
    expect_equal(rep(0, times = length(o1.assign$depot)), o1.assign$depot)
    expect_false(any(class(o1.assign) == "rxSolve"))
  })

  test_that("Numeric data frame lookup operators [[]] make sense", {
    expect_equal(o1.first[[1]], o1.df[[1]])
    expect_equal(o1.first[[3]], o1.df[[3]])
  })

  test_that("Character data frame lookup operators [[]] make sense", {
    expect_equal(o1.first[["depot"]], o1.df[["depot"]])
    expect_equal(o1.first[["de"]], o1.df[["de"]])
    expect_equal(o1.first[["de"]], NULL)
    expect_equal(o1.first[["de", exact = FALSE]], o1.df[["de", exact = FALSE]])
    expect_equal(o1.first[["de", exact = FALSE]], o1.df[["depot"]])
    expect_warning(o1.first[["de", exact = NA]])
  })

  test_that("Character data frame assignment operators [[]] make sense", {
    o1.assign <- o1.first
    expect_s3_class(o1.assign, "data.frame")
    o1.assign[["depot"]] <- 0
    expect_equal(rep(0, times = length(as.data.frame(o1.assign)[, 1])), as.data.frame(o1.assign)[["depot"]])
    expect_equal(rep(0, times = length(o1.assign$depot)), o1.assign$depot)
    expect_false(any(class(o1.assign) == "rxSolve"))
  })

  test_that("Character data frame lookup operators $ make sense", {
    expect_equal(o1.first$centr, o1.df$centr)
    expect_equal(o1.first$depot, o1.df$depot)
  })

  test_that("Character data frame assignment operators $ make sense", {
    o1.assign <- o1.first
    expect_s3_class(o1.assign, "rxSolve")
    o1.assign$depot <- 0
    expect_equal(rep(0, times = length(as.data.frame(o1.assign)[, 1])), as.data.frame(o1.assign)$depot)
    expect_equal(rep(0, times = length(o1.assign$depot)), o1.assign$depot)
    expect_false(any(class(o1.assign) == "rxSolve"))
  })

  test_that("rownames lookup & assignment makes sense", {
    expect_equal(rownames(o1.first), paste(seq_len(length(o1.first[, 1]))))
    rownames(o1.first) <- paste("row", seq_len(length(o1.first$depot)))
    expect_equal(rownames(o1.first), paste("row", seq_len(length(o1.first$depot))))
    rownames(o1.first) <- NULL
    expect_equal(rownames(o1.first), paste(seq_len(length(o1.first[, 1]))))
  })
})

rxTest({
  library(withr)
  test_that("new rxEt constructor", {
    ev <- .newRxEt()
    .env <- .rxEtEnv(ev)
    expect_true(inherits(ev, "rxEt"))
    expect_true(is.rxEt(ev))
    expect_equal(length(.env$chunks), 0L)
    expect_equal(.env$nobs, 0L)
    expect_equal(.env$ndose, 0L)
    expect_true(is.na(.env$units["dosing"]))
    expect_true(is.na(.env$units["time"]))
  })

  test_that("new rxEt constructor with units", {
    ev <- .newRxEt(amountUnits = "mg", timeUnits = "hours")
    .env <- .rxEtEnv(ev)
    expect_equal(.env$units["dosing"], c(dosing = "mg"))
    expect_equal(.env$units["time"], c(time = "hours"))
  })

  test_that("is.rxEt", {
    expect_true(is.rxEt(.newRxEt()))
    expect_false(is.rxEt(list()))
    expect_false(is.rxEt(data.frame()))
  })

  test_that("is.rxEt handles class-only rxEt (no .env)", {
    # Should return FALSE -- wrong structure, not a valid rxEt
    expect_false(is.rxEt(structure(list(), class = "rxEt")))
  })

  test_that(".etMaterialize on empty et", {
    ev <- .newRxEt()
    df <- .etMaterialize(ev)
    expect_equal(nrow(df), 0L)
    expect_true(all(c("id","low","time","high","cmt","amt","rate","ii","addl","evid","ss","dur") %in% names(df)))
  })

  test_that(".etMaterialize single obs chunk", {
    ev <- .newRxEt()
    .e <- .rxEtEnv(ev)
    .e$chunks <- list(list(time = c(0, 1, 2), evid = 0L))
    .e$nobs <- 3L
    df <- .etMaterialize(ev)
    expect_equal(nrow(df), 3L)
    expect_equal(df$time, c(0, 1, 2))
    expect_true(all(df$evid == 0L))
  })

  test_that("et stores homogeneous ids as one compressed group", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      ev <- et(1, id = 1:10)
      .e <- .rxEtEnv(ev)
      expect_equal(length(.e$groups), 1L)
      expect_equal(length(.e$chunks), 0L)
      expect_equal(.e$groups[[1]]$ids, 1:10)
      expect_false("id" %in% names(.e$groups[[1]]$data))
      expect_equal(.e$nobs, 10L)
    })
  })

  test_that("large homogeneous id resize stays grouped without chunk cloning", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      ev <- et(1, id = 1:70000)
      .e1 <- .rxEtEnv(ev)
      expect_equal(length(.e1$groups), 1L)
      expect_equal(length(.e1$chunks), 0L)
      expect_equal(length(.e1$groups[[1]]$ids), 70000L)

      ev2 <- et(ev, id = 70001:140000)
      .e2 <- .rxEtEnv(ev2)
      expect_equal(length(.e2$groups), 1L)
      expect_equal(length(.e2$chunks), 0L)
      expect_equal(length(.e2$groups[[1]]$ids), 70000L)
      expect_equal(min(.e2$groups[[1]]$ids), 70001L)
      expect_equal(max(.e2$groups[[1]]$ids), 140000L)
      expect_equal(.e2$nobs, 70000L)
    })
  })

  test_that("homogeneous compressed group expands on as.data.frame", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      ev <- et(1, id = 1:3)
      df <- as.data.frame(ev, all = TRUE)
      expect_equal(nrow(df), 3L)
      expect_equal(df$id, 1:3)
      expect_equal(df$time, c(1, 1, 1))
      expect_equal(df$evid, c(0L, 0L, 0L))
    })
  })

  test_that("homogeneous previews stay compressed for print and query helpers", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      ev <- et(amt = 100, ii = 24, addl = 2, id = 1:5) |>
        et(seq(0, 72, by = 24))

      dosing <- ev$get.dosing()
      sampling <- ev$get.sampling()

      expect_s3_class(dosing, "rxEtPreview")
      expect_s3_class(sampling, "rxEtPreview")
      expect_equal(nrow(dosing), 1L)
      expect_equal(nrow(sampling), 4L)
      expect_false("id" %in% names(dosing))
      expect_false("id" %in% names(sampling))

      out <- paste(capture.output(print(ev)), collapse = "\n")
      expect_match(out, "5 individuals")
      expect_match(out, "compressed preview")
    })
  })

  test_that("etRbind keeps non-overlapping homogeneous groups compressed", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      e1 <- et(amt = 100, ii = 24, addl = 2, id = 1:3)
      e2 <- et(amt = 50, ii = 12, addl = 1, id = 4:5)
      e3 <- rbind(e1, e2)
      .e <- .rxEtEnv(e3)

      expect_equal(length(.e$groups), 2L)
      expect_equal(length(.e$chunks), 0L)
      expect_equal(.e$groups[[1]]$ids, 1:3)
      expect_equal(.e$groups[[2]]$ids, 4:5)

      dosing <- e3$get.dosing()
      expect_s3_class(dosing, "rxEtPreview")
      expect_equal(nrow(dosing), 2L)
      expect_false("id" %in% names(dosing))

      out <- paste(capture.output(print(e3)), collapse = "\n")
      expect_match(out, "compressed preview for 2 groups")
      expect_match(out, "group 1: 3 individuals")
      expect_match(out, "group 2: 2 individuals")
    })
  })

  test_that("etRbind keeps matching homogeneous ids compressed", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      e1 <- et(amt = 100, id = 1:3)
      e2 <- et(amt = 50, time = 24, id = 1:3)
      e3 <- rbind(e1, e2)
      .e <- .rxEtEnv(e3)

      expect_equal(length(.e$groups), 1L)
      expect_equal(length(.e$chunks), 0L)
      expect_equal(.e$groups[[1]]$ids, 1:3)

      dosing <- e3$get.dosing()
      expect_s3_class(dosing, "rxEtPreview")
      expect_equal(nrow(dosing), 2L)
      expect_false("id" %in% names(dosing))
      expect_equal(dosing$time, c(0, 24))
      expect_equal(dosing$amt, c(100, 50))

      df <- as.data.frame(e3, all = TRUE)
      expect_equal(nrow(df), 6L)
      expect_equal(df$time[df$id == 1], c(0, 24))
      expect_equal(df$amt[df$id == 1], c(100, 50))
    })
  })

  test_that("etSeq keeps same-id homogeneous groups compressed", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
    e1 <- et(amt = 100, ii = 24, addl = 1, id = 1:3)
    e2 <- et(amt = 50, id = 1:3)
    e3 <- etSeq(e1, e2)
    .e <- .rxEtEnv(e3)

    expect_equal(length(.e$groups), 1L)
    expect_equal(length(.e$chunks), 0L)
    expect_equal(.e$groups[[1]]$ids, 1:3)

    dosing <- e3$get.dosing()
    expect_s3_class(dosing, "rxEtPreview")
    expect_equal(nrow(dosing), 2L)
    expect_false("id" %in% names(dosing))
    expect_equal(dosing$time, c(0, 48))
    expect_equal(dosing$amt, c(100, 50))

    df <- as.data.frame(e3, all = TRUE)
    expect_equal(nrow(df), 6L)
    expect_equal(sort(unique(df$id)), 1:3)
    expect_equal(df$time[df$id == 1], c(0, 48))
    expect_equal(df$amt[df$id == 1], c(100, 50))
    })
  })

  test_that("subset dose edits split homogeneous groups", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      ev <- et(amt = 10, id = 1:5) |> et(amt = 20, id = 4:5)
      .e <- .rxEtEnv(ev)

      expect_equal(length(.e$groups), 2L)
      expect_equal(.e$groups[[1]]$ids, 1:3)
      expect_equal(.e$groups[[2]]$ids, 4:5)
      expect_equal(.e$groups[[1]]$data$amt, 10)
      expect_equal(.e$groups[[2]]$data$amt, c(10, 20))

      dosing <- ev$get.dosing()
      expect_s3_class(dosing, "rxEtPreview")
      expect_equal(nrow(dosing), 3L)
      expect_false("id" %in% names(dosing))

      df <- as.data.frame(ev, all = TRUE)
      expect_equal(df$amt[df$id == 1], 10)
      expect_equal(df$amt[df$id == 4], c(10, 20))
      expect_equal(df$amt[df$id == 5], c(10, 20))
    })
  })

  test_that("id-only resize keeps added ids in grouped homogeneous tables", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      ev <- et(amt = 10, id = 1:3) |> et(amt = 20, id = 4:5)
      .templateId <- .rxEtEnv(ev)$ids[[1L]]
      .template <- as.data.frame(ev, all = TRUE)
      .template <- .template[.template$id == .templateId, c("time", "evid", "amt"), drop = FALSE]
      ev2 <- et(ev, id = 1:6)
      .e2 <- .rxEtEnv(ev2)

      expect_equal(length(.e2$groups), 2L)
      expect_true(any(vapply(.e2$groups, function(.g) 6L %in% .g$ids, logical(1))))

      df <- as.data.frame(ev2, all = TRUE)
      expect_true(6L %in% unique(df$id))
      .actual <- df[df$id == 6L, c("time", "evid", "amt"), drop = FALSE]
      rownames(.actual) <- NULL
      rownames(.template) <- NULL
      expect_equal(.actual, .template)
    })
  })

  test_that("id-only resize with disjoint ids keeps grouped template events", {
    ev <- et(amt = 10, id = 1:3) |> et(amt = 20, id = 4:5)
    .templateId <- .rxEtEnv(ev)$ids[[1L]]
    .template <- as.data.frame(ev, all = TRUE)
    .template <- .template[.template$id == .templateId, c("time", "evid", "amt"), drop = FALSE]

    ev2 <- et(ev, id = 6:7)
    .e2 <- .rxEtEnv(ev2)

    expect_equal(length(.e2$groups), 1L)
    expect_equal(sort(.e2$groups[[1]]$ids), 6:7)

    df <- as.data.frame(ev2, all = TRUE)
    expect_equal(sort(unique(df$id)), 6:7)
    .actual <- df[df$id == 6L, c("time", "evid", "amt"), drop = FALSE]
    rownames(.actual) <- NULL
    rownames(.template) <- NULL
    expect_equal(.actual, .template)
  })

  test_that("id-only resize to empty clears grouped tables", {
    ev <- et(amt = 10, id = 1:3) |> et(amt = 20, id = 4:5)
    ev0 <- et(ev, id = integer(0))
    .e0 <- .rxEtEnv(ev0)

    expect_equal(length(.e0$ids), 0L)
    expect_equal(length(.e0$groups), 0L)
    expect_equal(sum(vapply(.e0$chunks, Negate(is.null), logical(1))), 0L)
    expect_equal(.e0$nobs, 0L)
    expect_equal(.e0$ndose, 0L)
    expect_equal(nrow(as.data.frame(ev0, all = TRUE)), 0L)
  })

  test_that("id-only resize deduplicates repeated ids", {
    ev <- et(amt = 10, id = 1:3)
    ev2 <- et(ev, id = c(2, 2, 3, 3))
    .e2 <- .rxEtEnv(ev2)

    expect_equal(.e2$ids, c(2L, 3L))
    expect_equal(length(.e2$groups), 1L)
    expect_equal(.e2$groups[[1]]$ids, c(2L, 3L))

    df <- as.data.frame(ev2, all = TRUE)
    expect_equal(sort(unique(df$id)), c(2L, 3L))
    expect_equal(nrow(df), 2L)
  })

  test_that("id handling rejects non-integer-like values", {
    ev <- et(1, id = 1:2)

    expect_error(
      et(ev, id = c("a", "2")),
      "'id' must contain integer-like values"
    )

    expect_no_warning(
      ev2 <- et(ev, id = as.character(c(3, 4)))
    )
    expect_equal(sort(unique(as.data.frame(ev2, all = TRUE)$id)), c(3L, 4L))
  })

  test_that("simulate keeps grouped homogeneous tables compressed when unchanged", {
    ev <- et(1, id = 1:5)

    ev2 <- suppressWarnings(simulate(ev, seed = 42))
    .e2 <- .rxEtEnv(ev2)
    expect_equal(length(.e2$groups), 1L)
    expect_equal(length(.e2$chunks), 0L)
    expect_equal(.e2$groups[[1]]$ids, 1:5)

    ev$simulate(seed = 42)
    .e <- .rxEtEnv(ev)
    expect_equal(length(.e$groups), 1L)
    expect_equal(length(.e$chunks), 0L)
    expect_equal(.e$groups[[1]]$ids, 1:5)
  })

  test_that("simulate splits grouped homogeneous windows by id", {
    ev <- et(list(c(4, 0.5, NA))) |> et(id = 1:10)

    ev2 <- simulate(ev, seed = 42)
    .e2 <- .rxEtEnv(ev2)
    expect_equal(length(.e2$groups), 0L)
    expect_equal(sum(vapply(.e2$chunks, Negate(is.null), logical(1))), 10L)

    df2 <- as.data.frame(ev2)
    expect_equal(sort(unique(df2$id)), 1:10)
    expect_gt(stats::sd(df2$time), 0)

    ev$simulate(seed = 42)
    .e <- .rxEtEnv(ev)
    expect_equal(length(.e$groups), 0L)
    expect_equal(sum(vapply(.e$chunks, Negate(is.null), logical(1))), 10L)
  })

  test_that(".etMaterialize sorts by id then time", {
    ev <- .newRxEt()
    .e <- .rxEtEnv(ev)
    .e$chunks <- list(
      list(time = c(2, 0, 1), evid = 0L, id = 1L),
      list(time = c(0, 1), evid = 0L, id = 2L)
    )
    .e$nobs <- 5L
    df <- .etMaterialize(ev)
    expect_equal(df$time[df$id == 1], c(0, 1, 2))
    expect_equal(df$id, c(1L, 1L, 1L, 2L, 2L))
  })

  test_that(".etMaterialize evid=3 sorts before others at same time", {
    ev <- .newRxEt()
    .e <- .rxEtEnv(ev)
    .e$chunks <- list(
      list(time = 0, evid = 1L, amt = 100, cmt = "(default)", ii = 0, addl = 0L, ss = 0L, rate = 0, dur = 0),
      list(time = 0, evid = 3L)
    )
    .e$ndose <- 1L
    .e$nobs  <- 1L
    df <- .etMaterialize(ev)
    expect_equal(df$evid[1], 3L)
  })

  test_that(".etMaterialize mixed dose+obs fills dose defaults correctly", {
    .ev <- .newRxEt()
    .envRef <- .rxEtEnv(.ev)
    # Obs chunk (sparse -- no dose columns)
    .envRef$chunks <- list(
      list(time = c(1, 2), evid = 0L),
      list(time = 0, evid = 1L, amt = 100, cmt = "(default)")
      # dose chunk missing rate/ii/addl/ss/dur
    )
    .envRef$nobs  <- 2L
    .envRef$ndose <- 1L
    df <- .etMaterialize(.ev)
    .doseRow <- df[df$evid == 1L, ]
    expect_equal(.doseRow$rate, 0.0)
    expect_equal(.doseRow$ii,   0.0)
    expect_equal(.doseRow$addl, 0L)
    expect_equal(.doseRow$ss,   0L)
    expect_equal(.doseRow$dur,  0.0)
    # Obs rows should still be NA for dose cols
    .obsRows <- df[df$evid == 0L, ]
    expect_true(all(is.na(.obsRows$rate)))
  })

  test_that(".etObsChunk basic times", {
    chunk <- .etObsChunk(c(0, 1, 2, 4, 8))
    expect_equal(chunk$time, c(0, 1, 2, 4, 8))
    expect_equal(chunk$evid, 0L)
    expect_null(chunk$amt)   # obs chunks are sparse
    expect_null(chunk$rate)
  })

  test_that(".etObsChunk with cmt", {
    chunk <- .etObsChunk(c(0, 1), cmt = "central")
    expect_equal(chunk$cmt, "central")
  })

  test_that(".etObsChunk with id vector", {
    chunk <- .etObsChunk(c(0, 1), id = c(1L, 2L))
    expect_equal(chunk$id, c(1L, 2L))
  })

  test_that(".etObsChunk window list c(low,high) returns window chunk", {
    rxWithSeed(42, {
      chunk <- .etObsChunk(list(c(0, 2), c(4, 8)))
      expect_equal(chunk$low,  c(0, 4))
      expect_equal(chunk$high, c(2, 8))
      expect_equal(chunk$evid, 0L)
    })
  })

  test_that(".etObsChunk window list c(low,mid,high) uses mid as time", {
    chunk <- .etObsChunk(list(c(0, 1, 2), c(4, 6, 8)))
    expect_equal(chunk$time, c(1, 6))
    expect_equal(chunk$low,  c(0, 4))
    expect_equal(chunk$high, c(2, 8))
  })

  test_that(".etDoseChunk basic dose", {
    chunk <- .etDoseChunk(time = 0, amt = 100)
    expect_equal(chunk$time, 0)
    expect_equal(chunk$amt,  100)
    expect_equal(chunk$evid, 1L)
    expect_equal(chunk$ii,   0.0)
    expect_equal(chunk$addl, 0L)
    expect_equal(chunk$ss,   0L)
    expect_equal(chunk$rate, 0.0)
    expect_equal(chunk$dur,  0.0)
    expect_equal(chunk$cmt,  "(default)")
  })

  test_that(".etDoseChunk with addl", {
    chunk <- .etDoseChunk(time = 0, amt = 100, ii = 24, addl = 4L)
    expect_equal(chunk$ii,   24)
    expect_equal(chunk$addl, 4L)
  })

  test_that(".etDoseChunk nbr.doses interface", {
    chunk <- .etDoseChunk(time = 0, amt = 100,
                           nbr.doses = 5L, dosing.interval = 24)
    expect_equal(chunk$addl, 4L)
    expect_equal(chunk$ii,   24)
  })

  test_that(".etDoseChunk with until", {
    chunk <- .etDoseChunk(time = 0, amt = 100, ii = 24, until = 120)
    expect_equal(chunk$addl, 5L)
  })

  test_that(".etDoseChunk windowed time works with until", {
    rxWithSeed(42, {
      chunk <- .etDoseChunk(time = list(c(0, 6)), amt = 100, ii = 12, until = 48)
      expect_equal(chunk$low, 0)
      expect_equal(chunk$high, 6)
      expect_equal(chunk$addl, 3L)
    })
  })

  test_that("et until without additional doses matches main branch shape", {
    expect_warning(ev <- et(amt = 100, ii = 24, until = 24))
    expect_equal(names(as.data.frame(ev)), c("time", "amt", "evid"))
  })

  test_that(".etDoseChunk dur converts to rate", {
    chunk <- .etDoseChunk(time = 0, amt = 100, dur = 2)
    expect_equal(chunk$rate, 50)
    expect_equal(chunk$dur,  0.0)
  })

  test_that(".etDoseChunk rate=-1 modeled rate", {
    chunk <- .etDoseChunk(time = 0, amt = 100, rate = -1)
    expect_equal(chunk$rate, -1)
  })

  test_that(".etDoseChunk ss flag", {
    chunk <- .etDoseChunk(time = 0, amt = 100, ss = 1L)
    expect_equal(chunk$ss, 1L)
  })

  test_that(".etDoseChunk vector amt creates data.frame", {
    chunk <- .etDoseChunk(time = c(0, 24), amt = c(100, 50))
    expect_equal(nrow(chunk), 2L)
    expect_equal(chunk$amt, c(100, 50))
    expect_equal(chunk$time, c(0, 24))
  })

  test_that("et() empty creates rxEt", {
    ev <- et()
    expect_true(is.rxEt(ev))
    .envRef <- .rxEtEnv(ev)
    expect_equal(.envRef$nobs, 0L)
    expect_equal(.envRef$ndose, 0L)
  })

  test_that("et(time) creates obs record", {
    ev <- et(time = c(0, 1, 2, 4, 8))
    .envRef <- .rxEtEnv(ev)
    expect_equal(.envRef$nobs, 5L)
    df <- .etMaterialize(ev)
    expect_equal(nrow(df), 5L)
    expect_equal(df$time, c(0, 1, 2, 4, 8))
    expect_true(all(df$evid == 0L))
  })

  test_that("add.sampling converts unit-valued times to event table units", {
    ev <- eventTable(time.units = "hr")
    ev$add.sampling(units::set_units(c(1, 2), days))
    df <- as.data.frame(ev)
    expect_equal(as.numeric(df$time), c(24, 48))
  })

  test_that("etRbind unique can append observation times", {
    bid <- et(timeUnits = "hr") |> et(amt = 10000, ii = 12, until = units::set_units(1, days))
    qd <- et(timeUnits = "hr") |> et(amt = 20000, ii = 24, until = units::set_units(1, days))
    ev <- etRbind(bid, qd, id = "unique") |> et(seq(0, 24, length.out = 5))
    df <- as.data.frame(ev)
    expect_equal(sort(unique(df$id)), 1:2)
    expect_true(any(df$evid == 0L))
  })

  test_that("et(0, 24) two-arg sequence", {
    ev <- et(0, 24)
    df <- .etMaterialize(ev)
    expect_equal(df$time, as.numeric(0:24))
  })

  test_that("et(0, 24, by=4) generates seq", {
    ev <- et(0, 24, by = 4)
    df <- .etMaterialize(ev)
    expect_equal(df$time, seq(0, 24, by = 4))
  })

  test_that("et(amt=100) dose record", {
    ev <- et(amt = 100)
    .envRef <- .rxEtEnv(ev)
    expect_equal(.envRef$ndose, 1L)
    df <- .etMaterialize(ev)
    expect_equal(df$amt[df$evid == 1L], 100)
  })

  test_that("et pipe: obs then dose", {
    ev <- et(amt = 100) |> et(time = c(0, 1, 2, 4))
    .envRef <- .rxEtEnv(ev)
    expect_equal(.envRef$ndose, 1L)
    expect_equal(.envRef$nobs, 4L)
    df <- .etMaterialize(ev)
    expect_equal(nrow(df), 5L)
  })

  test_that("et(amountUnits, timeUnits)", {
    ev <- et(amountUnits = "mg", timeUnits = "hours")
    .envRef <- .rxEtEnv(ev)
    expect_equal(.envRef$units["dosing"], c(dosing = "mg"))
    expect_equal(.envRef$units["time"],   c(time   = "hours"))
  })

  test_that("et evid=obs alias", {
    ev <- et(time = 0, evid = obs)
    df <- .etMaterialize(ev)
    expect_equal(df$evid[1], 0L)
  })

  test_that("et cmt string", {
    ev <- et(amt = 100, cmt = "depot")
    df <- .etMaterialize(ev)
    expect_equal(df$cmt[df$evid == 1L], "depot")
  })

  test_that("as.data.frame shows only visible columns", {
    ev <- et(amt = 100, ii = 24, addl = 4) |> et(time = c(0, 1, 2))
    df <- as.data.frame(ev)
    expect_true("time"  %in% names(df))
    expect_true("evid"  %in% names(df))
    expect_true("amt"   %in% names(df))
    expect_true("ii"    %in% names(df))
    expect_true("addl"  %in% names(df))
    expect_false("id"   %in% names(df))   # single ID, hidden
    expect_false("low"  %in% names(df))
  })

  test_that("print.rxEt matches main branch summary output", {
    local_options(cli.unicode = FALSE, crayon.enabled = FALSE, width = 80)
    ev <- et(amt = 100) |> et(time = c(0, 1, 2))
    out <- paste(capture.output(print(ev)), collapse = "\n")
    expect_match(out, "EventTable with 4 records")
    expect_match(out, "1 dosing records \\(see \\$get\\.dosing\\(\\); add with add\\.dosing or et\\)")
    expect_match(out, "3 observation times \\(see \\$get\\.sampling\\(\\); add with add\\.sampling or et\\)")
    expect_match(out, "# A tibble: 4 x 3")
    expect_match(out, "<evid>")
    expect_match(out, "time\\s+amt\\s+evid")
  })

  test_that("print.rxEt after piping keeps main branch columns", {
    local_options(cli.unicode = FALSE, crayon.enabled = FALSE, width = 80)
    ev <- et(timeUnits = "hr") |>
      et(amt = 100, ii = 12, until = 24) |>
      et(seq(0, 24, by = 6))
    out <- paste(capture.output(print(ev)), collapse = "\n")
    expect_match(out, "EventTable with 6 records")
    expect_match(out, "multiple doses in `addl` columns, expand with \\$expand\\(\\); or etExpand\\(\\)")
    expect_match(out, "# A tibble: 6 x 5")
    expect_match(out, "<evid>")
    expect_match(out, "time\\s+amt\\s+ii\\s+addl\\s+evid")
    expect_false(grepl("\\bid\\b", out))
  })

  test_that("as.data.table.rxEt returns data.table", {
    ev <- et(time = c(0, 1, 2))
    dt <- as.data.table(ev)
    expect_true(inherits(dt, "data.table"))
  })

  test_that("dplyr::filter.rxEt converts through tibble", {
    skip_if_not_installed("dplyr")
    ev <- et(amt = 100) |> et(time = c(0, 1, 2))
    out <- dplyr::filter(ev, evid == 1)
    expect_s3_class(out, "tbl_df")
    expect_equal(nrow(out), 1L)
    expect_equal(out$amt, 100)
    expect_equal(out$evid, 1L)
  })

  test_that("$.rxEt add.dosing closure", {
    ev <- et()
    .env <- .rxEtEnv(ev)
    ev$add.dosing(dose = 100, nbr.doses = 5, dosing.interval = 24)
    expect_equal(.env$ndose, 1L)
    df <- as.data.frame(ev)
    expect_equal(df$amt[1], 100)
    expect_equal(df$addl[1], 4L)
    expect_equal(df$ii[1], 24)
  })

  test_that("$.rxEt add.sampling closure", {
    ev <- et()
    .env <- .rxEtEnv(ev)
    ev$add.sampling(c(0, 1, 2, 4, 8))
    expect_equal(.env$nobs, 5L)
  })

  test_that("$.rxEt get.units returns units", {
    ev <- et(amountUnits = "mg", timeUnits = "hours")
    expect_equal(ev$get.units(), c(dosing = "mg", time = "hours"))
  })

  test_that("$.rxEt get.dosing homogenous path returns chunk columns with units and without id", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      skip_if_not_installed("units")
      ev <- et(timeUnits = "hr") |>
        et(amt = 100, ii = 12, until = 24) |>
        et(seq(0, 24, by = 6))
      df <- ev$get.dosing()
      expect_equal(rownames(df), "1")
      expect_true(inherits(df$time, "units"))
      # Homogenous path returns chunk data: no id expansion, no id column
      expect_equal(
        names(df),
        c("time", "amt", "evid", "cmt", "ii", "addl", "ss", "rate", "dur", "low", "high")
      )
      expect_equal(as.numeric(df$ii), 12)
      expect_equal(df$addl, 2L)
    })
  })

  test_that("$.rxEt get.dosing non-homogenous path returns all records with id", {
    withr::with_options(list(rxode2.homogenous = FALSE), {
      ev <- et(timeUnits = "hr") |>
        et(amt = 100, ii = 12, until = 24) |>
        et(seq(0, 24, by = 6))
      df <- ev$get.dosing()
      expect_equal(rownames(df), "1")
      # Non-homogenous path retains units from timeUnits specification
      expect_true(inherits(df$time, "units"))
      # Non-homogenous path materializes with id column
      expect_equal(
        names(df),
        c("id", "low", "time", "high", "cmt", "amt", "rate", "ii", "addl", "evid", "ss", "dur")
      )
      expect_equal(as.numeric(df$ii), 12)
      expect_equal(df$addl, 2L)
    })
  })

  test_that("$.rxEt get.dosing reindexes rows after piping tables together", {
    skip_if_not_installed("units")
    bid <- et(timeUnits = "hr") |> et(amt = 100, ii = 12, until = 24)
    qd <- et(timeUnits = "hr") |> et(amt = 200, ii = 24, until = 24)
    ev <- etRbind(bid, qd, id = "unique") |>
      et(seq(0, 24, by = 12))
    df <- ev$get.dosing()
    expect_equal(rownames(df), c("1", "2"))
    expect_equal(df$amt, c(100, 200))
    expect_equal(df$low, units::set_units(c(NA, NA), "hr"))
    expect_equal(df$time, units::set_units(c(0, 0), "hr"))
    expect_equal(df$high, units::set_units(c(NA, NA), "hr"))
    expect_equal(df$ii, units::set_units(c(12, 0), "hr"))
    expect_equal(df$addl, c(2L, 0L))
  })

  test_that("$.rxEt nobs and ndose", {
    ev <- et(amt = 100) |> et(time = c(0, 1))
    expect_equal(ev$nobs,  2L)
    expect_equal(ev$ndose, 1L)
  })

  test_that("$.rxEt time column", {
    ev <- et(time = c(0, 1, 2))
    expect_equal(ev$time, c(0, 1, 2))
  })

  test_that("$.rxEt show property", {
    ev <- et()
    expect_false(ev$show["amt"])
    ev <- et(amt = 100)
    expect_true(ev$show["amt"])
  })

  test_that("etRbind two tables merges rows", {
    ev1 <- et(amt = 100) |> et(time = c(0, 1, 2))
    ev2 <- et(amt = 50)  |> et(time = c(3, 4, 5))
    ev  <- etRbind(ev1, ev2)
    df  <- as.data.frame(ev)
    expect_equal(nrow(df), 8L)   # 2 doses + 6 obs
  })

  test_that("etRbind id=unique renumbers IDs", {
    ev1 <- et(amt = 100, id = 1L) |> et(time = c(0, 1))
    ev2 <- et(amt =  50, id = 1L) |> et(time = c(0, 1))
    ev  <- etRbind(ev1, ev2, id = "unique")
    df  <- as.data.frame(ev)
    expect_equal(sort(unique(df$id)), c(1L, 2L))
  })

  test_that("eventTable() creates empty rxEt with units", {
    ev <- eventTable(amount.units = "mg", time.units = "hours")
    expect_true(is.rxEt(ev))
    .env <- .rxEtEnv(ev)
    expect_equal(.env$units["dosing"], c(dosing = "mg"))
  })

  test_that("eventTable add.dosing/add.sampling pipe", {
    qd <- eventTable(amount.units = "mg", time.units = "days") |>
      add.dosing(dose = 50, nbr.doses = 5, dosing.interval = 1) |>
      add.sampling(seq(from = 0, to = 5, by = 0.5))
    df <- as.data.frame(qd)
    expect_equal(sum(df$evid == 1L), 1L)  # lazy: 1 dose row with addl=4
    expect_true(sum(df$evid == 0L) > 0L)
  })

  test_that("etExpand expands addl doses", {
    ev  <- et(amt = 100, ii = 24, addl = 4)
    ev2 <- etExpand(ev)
    df  <- as.data.frame(ev2)
    expect_equal(sum(df$evid == 1L), 5L)
    expect_true(all(df$addl[df$evid == 1L] == 0L))
  })

  test_that("etExpand keeps homogeneous groups compressed", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      ev <- et(amt = 100, ii = 24, addl = 2, id = 1:3)
      ev2 <- etExpand(ev)
      .e <- .rxEtEnv(ev2)

      expect_equal(length(.e$groups), 1L)
      expect_equal(length(.e$chunks), 0L)
      expect_equal(.e$groups[[1]]$ids, 1:3)
      expect_equal(.e$groups[[1]]$data$time, c(0, 24, 48))
      expect_true(all(.e$groups[[1]]$data$addl == 0L))

      dosing <- ev2$get.dosing()
      expect_s3_class(dosing, "rxEtPreview")
      expect_equal(nrow(dosing), 3L)
      expect_false("id" %in% names(dosing))

      df <- as.data.frame(ev2, all = TRUE)
      expect_equal(nrow(df), 9L)
      expect_equal(df$time[df$id == 1], c(0, 24, 48))
    })
  })

  test_that("in-place expand updates addl print metadata", {
    local_options(cli.unicode = FALSE, crayon.enabled = FALSE, width = 80)
    ev <- et(amt = 100, ii = 24, addl = 4)
    ev$expand()
    expect_equal(ev$ndose, 5L)
    expect_false(ev$show["addl"])
    expect_equal(names(as.data.frame(ev)), c("time", "amt", "ii", "evid"))
    expect_false(grepl("multiple doses in `addl` columns", paste(capture.output(print(ev)), collapse = "\n"), fixed = TRUE))
  })

  test_that("in-place expand keeps homogeneous groups compressed", {
    ev <- et(amt = 100, ii = 24, addl = 2, id = 1:3)
    ev$expand()
    .e <- .rxEtEnv(ev)

    expect_equal(length(.e$groups), 1L)
    expect_equal(length(.e$chunks), 0L)
    expect_equal(.e$groups[[1]]$data$time, c(0, 24, 48))
    expect_true(all(.e$groups[[1]]$data$addl == 0L))
    expect_false(ev$show["addl"])

    df <- as.data.frame(ev, all = TRUE)
    expect_equal(nrow(df), 9L)
    expect_equal(df$time[df$id == 1], c(0, 24, 48))
  })

  test_that("clear.dosing updates grouped homogeneous tables", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      ev <- et(amt = 10, ii = 24, addl = 1, id = 1:3) |>
        et(seq(0, 24, by = 24))
      ev$clear.dosing()
      .e <- .rxEtEnv(ev)

      expect_equal(length(.e$groups), 1L)
      expect_equal(length(.e$chunks), 0L)
      expect_equal(.e$ndose, 0L)
      expect_equal(.e$nobs, 6L)
      expect_false(ev$show["addl"])

      sampling <- ev$get.sampling()
      expect_s3_class(sampling, "rxEtPreview")
      expect_equal(nrow(sampling), 2L)
      expect_equal(ev$get.dosing(), NULL)

      df <- as.data.frame(ev, all = TRUE)
      expect_equal(nrow(df), 6L)
      expect_true(all(df$evid == 0L))
    })
  })

  test_that("clear.sampling updates grouped homogeneous tables", {
    withr::with_options(list(rxode2.homogenous = TRUE), {
      ev <- et(amt = 10, ii = 24, addl = 1, id = 1:3) |>
        et(seq(0, 24, by = 24))
      ev$clear.sampling()
      .e <- .rxEtEnv(ev)

      expect_equal(length(.e$groups), 1L)
      expect_equal(length(.e$chunks), 0L)
      expect_equal(.e$ndose, 3L)
      expect_equal(.e$nobs, 0L)
      expect_equal(ev$get.sampling(), NULL)

      dosing <- ev$get.dosing()
      expect_s3_class(dosing, "rxEtPreview")
      expect_equal(nrow(dosing), 1L)

      df <- as.data.frame(ev, all = TRUE)
      expect_equal(nrow(df), 3L)
      expect_true(all(df$evid != 0L))
    })
  })

  test_that("etRep repeats event table", {
    ev  <- et(amt = 100, ii = 24, addl = 4) |> et(time = c(0, 24))
    ev3 <- etRep(ev, times = 3, samples = "use")
    df  <- as.data.frame(ev3)
    expect_equal(sum(df$evid == 1L), 3L)
    expect_equal(sum(df$evid == 0L), 6L)
  })

  test_that("etRep with wait", {
    ev  <- et(amt = 100) |> et(time = c(0, 24))
    ev3 <- etRep(ev, times = 3, wait = 24, samples = "use")
    df  <- as.data.frame(ev3)
    expect_true(max(df$time) >= 48)
  })

  test_that("etSeq offsets times of second table", {
    ev1 <- et(amt = 100) |> et(time = c(0, 24))
    ev2 <- et(amt = 100) |> et(time = c(0, 24))
    ev  <- etSeq(ev1, ev2, samples = "use")
    df  <- as.data.frame(ev)
    # second table should start after first table's last event (at 24),
    # so the second table's obs at time 24 becomes 24+24=48
    expect_true(max(df$time) > 24)
  })

  test_that("et() handles NONMEM-style uppercase column names", {
    df <- data.frame(ID = c(1L, 2L), TIME = c(0, 0), EVID = c(1L, 1L),
                     CMT = c(1L, 1L), AMT = c(10, 50))
    ev <- et(df)
    out <- as.data.frame(ev)
    expect_true("id" %in% names(out))
    expect_equal(sort(unique(out$id)), c(1L, 2L))
  })

  test_that("et() shows id column even for single-ID data frame input", {
    df <- data.frame(ID = 1L, TIME = 0, EVID = 1L, CMT = 1L, AMT = 10)
    ev <- et(df)
    out <- dplyr::as_tibble(ev)
    expect_true("id" %in% names(out))
  })

  test_that("et() accepts tibble without warnings (no ii/rate/var columns)", {
    skip_if_not_installed("tibble")
    tbl <- tibble::tibble(time = c(0, 1, 2), amt = c(100, NA, NA), evid = c(1L, 0L, 0L))
    expect_no_warning(et(tbl))
    ev <- et(tbl)
    out <- as.data.frame(ev)
    expect_equal(nrow(out), 3L)
    expect_equal(sum(out$evid == 1L), 1L)
  })

  test_that("et() accepts tibble with ii column", {
    skip_if_not_installed("tibble")
    tbl <- tibble::tibble(time = 0, amt = 100, ii = 24, addl = 4L, evid = 1L)
    expect_no_warning(et(tbl))
    ev <- et(tbl)
    out <- as.data.frame(ev)
    expect_equal(out$ii[1], 24)
  })

  test_that("et() accepts tibble with rate column", {
    skip_if_not_installed("tibble")
    tbl <- tibble::tibble(time = 0, amt = 100, rate = 50, evid = 1L)
    expect_no_warning(et(tbl))
    ev <- et(tbl)
    out <- as.data.frame(ev)
    expect_equal(out$rate[1], 50)
  })

  test_that("et() pipe from tibble works without warnings", {
    skip_if_not_installed("tibble")
    tbl <- tibble::tibble(time = c(0, 24), amt = c(100, NA), evid = c(1L, 0L))
    expect_no_warning({
      ev <- tbl |> rxode2::et()
    })
    out <- as.data.frame(ev)
    expect_equal(nrow(out), 2L)
  })

  test_that("import.EventTable accepts tibble without warnings", {
    skip_if_not_installed("tibble")
    tbl <- tibble::tibble(time = c(0, 1), amt = c(100, NA), evid = c(1L, 0L))
    ev <- et()
    expect_no_warning(ev$import.EventTable(tbl))
    out <- as.data.frame(ev)
    expect_equal(nrow(out), 2L)
  })

  test_that("et() loop with new IDs adds doses and observations to all IDs", {
    nSub <- 5
    dose <- c(45, 48, 52, 50, 47)
    ev <- et()
    for (i in seq_len(nSub)) {
      ev <- ev |> et(id = i, amt = dose[i], time = 0, addl = 9, ii = 12)
    }
    ev <- ev |> add.sampling(time = 0:240)
    df <- as.data.frame(ev)
    expect_equal(sort(unique(df$id)), 1:5)
    tbl_evid <- table(df$id, df$evid)
    expect_true(all(tbl_evid[, "0"] == 241))
    expect_true(all(tbl_evid[, "1"] == 1))
  })
})

if (!.Call(`_rxode2_isIntel`)) {
  rxTest({
    # (Regression) test 3 multiple instances of rxode2 objects to ensure
    # C symbols and operations don't conflict.

    # Make sure C operations and symbols don't conflict

    test.dir <- tempfile("Rxmult-")

    ## rxode2 instance 1
    m1 <-
      rxode2(
        model = "
         C2 = centr/V2;
         C3 = peri/V3;
         d/dt(depot) =-KA*depot;
         d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
         d/dt(peri)  =                    Q*C2 - Q*C3;
         d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;",
        modName = "inst1",
        wd = test.dir
      )

    test_that("rxode2 instance 1 is created", {
      expect_s3_class(m1, "rxode2")
    })

    et1 <- eventTable(amount.units = "ug", time.units = "hours")
    et1$add.dosing(dose = 10000, nbr.doses = 5, dosing.interval = 24)
    et1$add.sampling(0:24)
    et1$add.sampling(seq(from = 24 + 8, to = 5 * 24, by = 8))

    test_that("rxode2 event table 1 was created", {
      expect_s3_class(et1, "rxEt")
      expect_equal(et1$get.nobs(), 37)
      expect_equal(length(et1$get.dosing()[, 1]), 1)
    })

    o1.first <- NULL
    test_that("warning", {
      expect_warning({
        o1.first <<- m1$solve(
          params = c(
            KA = .291, CL = 18.6, V2 = 40.2, Q = 10.5, V3 = 297.0,
            Kin = 1.0, Kout = 1.0, EC50 = 200.0
          ),
          events = et1,
          inits = c(0, 0, 0, 1)
        )
      })
    })

    test_that("rxode2 event 1 solved.", {
      expect_equal(length(o1.first[, 1]), et1$get.nobs())
    })

    ## rxode2 instance 2 (complete example)
    m2 <-
      rxode2(
        model = "d/dt(y) = r * y * (1.0 - y/K);",
        modName = "inst2",
        wd = test.dir
      )

    test_that("rxode2 instance 2 was created", {
      expect_s3_class(m1, "rxode2")
    })

    et2 <- eventTable(time.units = NA)
    et2$add.sampling(seq(from = 0, to = 20, by = 0.2))

    test_that("rxode2 event table 2 was created", {
      expect_s3_class(et1, "rxEt")
      expect_equal(et2$get.nobs(), 101)
      expect_equal(length(et2$get.dosing()[, 1]), 0)
    })

    o2.s <-
      m2$solve(params = c(r = 1, K = 10), events = et2, inits = c(y = 2), method = "lsoda")

    test_that("rxode2 instance 2 was solved", {
      expect_equal(length(o2.s[, 1]), et2$get.nobs())
    })

    ## rxode2 instance 3 (complete example)
    m3 <-
      rxode2(
        model = "
         d/dt(X) = a*X + Y*Z;
         d/dt(Y) = b*(Y - Z);
         d/dt(Z) = -X*Y + c*Y - Z;",
        modName = "inst3",
        wd = tempdir() # don't pollute "./tests"
      )

    test_that("rxode2 instance 3 is created", {
      expect_s3_class(m3, "rxode2")
    })

    et3 <- eventTable() # default time units
    et3$add.sampling(seq(from = 0, to = 100, by = 0.01))

    test_that("rxode2 instance 3 event table is created", {
      expect_s3_class(et3, "rxEt")
      expect_equal(et3$get.nobs(), 10001)
    })

    o3 <-
      m3$solve(
        params = c(a = -8 / 3, b = -10, c = 28),
        events = et3,
        inits = c(X = 1, Y = 1, Z = 1)
      )

    test_that("rxode2 instance 3 was solved", {
      expect_equal(et3$get.nobs(), length(o3[, 1]))
    })

    ## Now go back to model 1 for (same) integration
    o1.second <- NULL
    test_that("Warning when inits is unnamed", {
      expect_warning({
        o1.second <<- m1$solve(
          params = c(
            KA = .291, CL = 18.6, V2 = 40.2, Q = 10.5, V3 = 297.0,
            Kin = 1.0, Kout = 1.0, EC50 = 200.0
          ),
          events = et1,
          inits = c(0, 0, 0, 1)
        )
      })
    })

    test_that("Second solve of model 1 gives the same results", {
      expect_equal(o1.first, o1.second)
    })

    ## and go back to model 2 for different ode solver
    o2.ns <-
      m2$solve(params = c(r = 1, K = 10), events = et2, inits = c(y = 12), method = "dop853")

    test_that("Differnt solve of Model 1 still works", {
      expect_equal(length(o2.ns[, 1]), length(o2.s[, 1]))
    })

    ## Inspect the internal compilation manager in each of the rxode2 objects
    prt <-
      function(obj, expect) {
        ## cat(
        ##     sprintf('Model name: %s\nDyn lib: %s\node_solver symbol: %s\n',
        ##             obj$modName, obj$cmpMgr$dllfile, obj$cmpMgr$ode_solver
        ##             ))
        expect_equal(obj$modName, expect)
        expect_match(obj$cmpMgr$dllfile, expect)
      }

    test_that("Compilation manager elements make sense.", {
      prt(m1, "inst1")
      prt(m2, "inst2")
      prt(m3, "inst3")
    })

    ## Unload object code (this is platform-dependent, as documented in the
    ## "Note" section of help("dyn.load"). Remove test.dir.
    test_that("Unloading will allow removal of directory", {
      m1$dynUnload()
      m2$dynUnload()
      m3$dynUnload()
      .ret <- unlink(test.dir, recursive = TRUE)
      skip_on_os("windows")
      expect_equal(.ret, 0) # 0==success, 1==failed
    })
  })
}

rxTest({

  test_that("sync covariates", {

    dat <- structure(list(ID = c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
                                 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
                                 1L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                                 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 3L, 3L, 3L, 3L,
                                 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L,
                                 3L, 3L, 3L, 3L, 3L, 3L, 3L),
                          TIME = c(0, 0, 0, 0.5, 0.5, 1, 1,
                                   1.5, 1.5, 2, 2, 2.5, 2.5, 3, 3, 3.5, 3.5, 4, 4, 4.5, 4.5, 5,
                                   5, 5.5, 5.5, 6, 6, 0, 0, 0, 0.5, 0.5, 1, 1, 1.5, 1.5, 2, 2, 2.5,
                                   2.5, 3, 3, 3.5, 3.5, 4, 4, 4.5, 4.5, 5, 5, 5.5, 5.5, 6, 6, 0,
                                   0, 0, 0.5, 0.5, 1, 1, 1.5, 1.5, 2, 2, 2.5, 2.5, 3, 3, 3.5, 3.5,
                                   4, 4, 4.5, 4.5, 5, 5, 5.5, 5.5, 6, 6),
                          EVID = c(80101L, 60101L,
                                   0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                   0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 80101L, 60101L, 0L, 0L, 0L,
                                   0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                   0L, 0L, 0L, 0L, 0L, 0L, 80101L, 60101L, 0L, 0L, 0L, 0L, 0L, 0L,
                                   0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                                   0L, 0L, 0L),
                          AMT = c(10, 10, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, 10, 10, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 10, 10, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                                  NA, NA, NA, NA, NA, NA, NA),
                          II = c(0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0),
                          DV = c(NA, NA, 0, 3.8, 0.9, 5.8,
                                 2.5, 3.2, 3.2, 1.8, 2.9, 1, 2.2, 0.5, 1.6, 0.3, 1.1, 0.2, 0.8,
                                 0.1, 0.5, 0, 0.3, 0, 0.2, 0, 0.1, NA, NA, 0, 3.4, 0.7, 7.3, 2.4,
                                 2.8, 2.5, 1.3, 2.7, 1, 1.6, 0.6, 1.1, 0.3, 0.8, 0.2, 0.8, 0.1,
                                 0.5, 0, 0.3, 0, 0.2, 0, 0.2, NA, NA, 0, 2.8, 1.1, 6.7, 2.7, 3.9,
                                 3.6, 1.5, 3.1, 0.7, 2.4, 0.5, 1.5, 0.3, 1.2, 0.1, 0.5, 0.1, 0.6,
                                 0.1, 0.4, 0, 0.2, 0, 0.2),
                          CMT = c(1L, 1L, 4L, 3L, 4L, 3L, 4L,
                                  3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L,
                                  3L, 4L, 3L, 4L, 1L, 1L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L,
                                  4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 1L,
                                  1L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L,
                                  3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L, 3L, 4L),
                          nlmixrRowNums =
                            c(1,
                              NA, 14, 2, 15, 3, 16, 4, 17, 5, 18, 6, 19, 7, 20, 8, 21, 9, 22,
                              10, 23, 11, 24, 12, 25, 13, 26, 27, NA, 40, 28, 41, 29, 42, 30,
                              43, 31, 44, 32, 45, 33, 46, 34, 47, 35, 48, 36, 49, 37, 50, 38,
                              51, 39, 52, 53, NA, 66, 54, 67, 55, 68, 56, 69, 57, 70, 58, 71,
                              59, 72, 60, 73, 61, 74, 62, 75, 63, 76, 64, 77, 65, 78)),
                     class = "data.frame", row.names = c(NA, -81L))

    par <- structure(c(0, 0, 0, 0.693147180559945, 0.693147180559945, 0.693147180559945,
                       2.30258509299405, 2.30258509299405, 2.30258509299405, 2.30258509299405,
                       2.30258509299405, 2.30258509299405, 0, 0, 0, 1, 1, 1, 1, 1, 1,
                       -1.86782087447842, 0, 0, -2.44813462994314, 0, 0, 0.0831313226164289,
                       0, 0, -2.28843604434224, 0, 0, 1.97816823869187e-05, 0, 0),
                     dim = c(3L, 12L),
                     dimnames = list(NULL,
                                     c("THETA[1]", "THETA[2]", "THETA[3]",
                                       "THETA[4]", "THETA[5]", "THETA[6]", "THETA[7]", "ETA[1]", "ETA[2]",
                                       "ETA[3]", "ETA[4]", "ETA[5]")))


    rx <- rxode2({
      param(THETA[1], THETA[2], THETA[3], THETA[4], THETA[5], THETA[6],
            THETA[7], ETA[1], ETA[2], ETA[3], ETA[4], ETA[5])
      cmt(center)
      cmt(meta)
      rx_expr_5 ~ ETA[3] + THETA[1]
      rx_expr_6 ~ ETA[1] + THETA[2]
      rx_expr_7 ~ ETA[2] + THETA[4]
      rx_expr_10 ~ exp(rx_expr_5)
      d/dt(center) = -rx_expr_10 * center - exp(rx_expr_6 - (rx_expr_7)) *
        center
      rx_expr_8 ~ ETA[5] + THETA[5]
      rx_expr_11 ~ exp(rx_expr_8)
      dur(center) = rx_expr_11
      rx_expr_9 ~ ETA[4] + THETA[3]
      d/dt(meta) = rx_expr_10 * center - exp(rx_expr_9 - (rx_expr_7)) *
        meta
      center(0) = 0
      meta(0) = 0
      rx_expr_0 ~ CMT == 4
      rx_expr_1 ~ CMT == 3
      rx_expr_2 ~ 1 - (rx_expr_0)
      rx_yj_ ~ 2 * (rx_expr_2) * (rx_expr_1) + 2 * (rx_expr_0)
      rx_expr_3 ~ (rx_expr_0)
      rx_expr_4 ~ (rx_expr_2)
      rx_expr_12 ~ rx_expr_4 * (rx_expr_1)
      rx_lambda_ ~ rx_expr_12 + rx_expr_3
      rx_hi_ ~ rx_expr_12 + rx_expr_3
      rx_low_ ~ 0
      rx_expr_13 ~ exp(-(rx_expr_7))
      rx_expr_14 ~ rx_expr_13 * meta
      rx_expr_15 ~ rx_expr_13 * center
      rx_expr_17 ~ rx_expr_14 * (rx_expr_0)
      rx_expr_18 ~ rx_expr_15 * (rx_expr_2)
      rx_expr_19 ~ rx_expr_18 * (rx_expr_1)
      rx_pred_ = (rx_expr_0) * (rx_expr_17 + rx_expr_19) + rx_expr_18 *
        Rx_pow_di((rx_expr_1), 2)
      rx_r_ = (rx_expr_0) * Rx_pow_di(((rx_expr_17 + rx_expr_19) *
                                         THETA[7]), 2) + (rx_expr_2) * Rx_pow_di((rx_expr_15 *
                                                                                    THETA[6] * (rx_expr_1)), 2) * (rx_expr_1)
      tkm = THETA[1]
      tcl = THETA[2]
      tclm = THETA[3]
      tv = THETA[4]
      tdur0 = THETA[5]
      prop.err = THETA[6]
      prop.err2 = THETA[7]
      eta.cl = ETA[1]
      eta.v = ETA[2]
      eta.km = ETA[3]
      eta.clm = ETA[4]
      eta.dur0 = ETA[5]
      km = rx_expr_10
      cl = exp(rx_expr_6)
      clm = exp(rx_expr_9)
      v = exp(rx_expr_7)
      dur0 = rx_expr_11
      cp = rx_expr_15
      cm = rx_expr_14
      tad = tad()
      dosenum = dosenum()
      cmt(cp)
      cmt(cm)
      dvid(3, 4)
    })

    s <- rxSolve(rx, dat, par, returnType="data.frame",
                 keep=c("nlmixrRowNums", "DV"), subsetNonmem=TRUE, addCov=TRUE)

    expect_equal(sort(unique(as.integer(s[s$time==0.5,"CMT"]))), 3:4)

  })

  skip_if_not_installed("units")

  for (.homogenous in c(TRUE, FALSE)) {
    withr::with_options(list(rxode2.homogenous = .homogenous), {

      for (meth in .methodsCov) {

        # context(sprintf("Simple test for time-varying covariates (%s)", meth))

        ode <- rxode2({
          b <- -1
          d/dt(X) <- a * X + Y * Z
          d/dt(Y) <- b * (Y - Z)
          d/dt(Z) <- -X * Y + c * Y - Z
          printf("%.16f,%.16f\n", t, c)
        })

        odeLin <- rxode2({
          linear(c)
          b <- -1
          d/dt(X) <- a * X + Y * Z
          d/dt(Y) <- b * (Y - Z)
          d/dt(Z) <- -X * Y + c * Y - Z
          printf("%.16f,%.16f\n", t, c)
        })

        odeLocf <- rxode2({
          locf(c)
          b <- -1
          d/dt(X) <- a * X + Y * Z
          d/dt(Y) <- b * (Y - Z)
          d/dt(Z) <- -X * Y + c * Y - Z
          printf("%.16f,%.16f\n", t, c)
        })

        odeNocb <- rxode2({
          nocb(c)
          b <- -1
          d/dt(X) <- a * X + Y * Z
          d/dt(Y) <- b * (Y - Z)
          d/dt(Z) <- -X * Y + c * Y - Z
          printf("%.16f,%.16f\n", t, c)
        })

        odeMidpoint <- rxode2({
          midpoint(c)
          b <- -1
          d/dt(X) <- a * X + Y * Z
          d/dt(Y) <- b * (Y - Z)
          d/dt(Z) <- -X * Y + c * Y - Z
          printf("%.16f,%.16f\n", t, c)
        })

        et <- eventTable(time.units = "hr") # default time units
        et$add.sampling(seq(from = 0, to = 10, by = 0.5))

        cov <- data.frame(c = as.numeric(et$get.EventTable()$time) + 1)

        et0 <- et

        et <- cbind(et, cov)

        ## rxode2 uses isSameTime(a,b): |a-b| <= 2*eps*max(|a|,|b|) for
        ## covariate lookup.  When an integrator evaluates the RHS at a
        ## time within that tolerance of an observation knot, rxode2's
        ## fast path returns the knot value directly.  The approxfun
        ## comparison must match that behavior.  The snap is applied
        ## ONLY when the actual printed c value already equals the
        ## approxfun value at the snapped time (confirming the fast
        ## path fired); if actual differs from the snapped expected,
        ## the slow path was used and approxfun at the original t
        ## already agrees -- so no snap is needed.
        .obs_times <- as.numeric(et$time)
        .snap_to_obs <- function(.t) {
          ## Vectorized: find nearest obs knot using findInterval
          .ni <- findInterval(.t, .obs_times, all.inside = TRUE)
          .nt_l <- .obs_times[pmax(.ni, 1L)]
          .nt_r <- .obs_times[pmin(.ni + 1L, length(.obs_times))]
          .nt <- ifelse(abs(.t - .nt_l) <= abs(.t - .nt_r), .nt_l, .nt_r)
          .tol <- 2 * .Machine$double.eps *
            pmax(abs(.t), abs(.nt), .Machine$double.xmin)
          ifelse(abs(.t - .nt) <= .tol, .nt, .t)
        }
        ## Conditional approxfun: snap t to nearest obs only when
        ## the actual value already matches the snapped expected.
        .cov_approx <- function(.t, .c, .fun) {
          .st <- .snap_to_obs(.t)
          .c_s <- .fun(.st)
          ifelse(.st != .t & .c == .c_s, .c_s, .fun(.t))
        }

        cov.lin <- approxfun(et$time, et$c, yleft = et$c[1], yright = et$c[length(cov$c)])

        t <- tempfile("temp", fileext = ".csv")
        suppressWarnings(.rxWithSink(t, {
          cat("t,c\n")
          out <- rxSolve(ode,
                         params = c(a = -8 / 3, b = -10),
                         events = et,
                         inits = c(X = 1, Y = 1, Z = 1),
                         addCov = TRUE,
                         covsInterpolation = "linear",
                         method = meth
                         )
        }))

        lin.interp <- read.csv(t)
        unlink(t)

        lin.interp$c2 <- .cov_approx(lin.interp$t, lin.interp$c, cov.lin)

        test_that(paste0("time varying covariates output covariate in data frame (", meth, ";", .homogenous, ")"), {
          expect_equal(cov$c, out$c)
        })

        test_that(paste0("Linear Approximation matches approxfun (", meth, ";", .homogenous, ")"), {
          expect_equal(lin.interp$c, lin.interp$c2)
        })

        t <- tempfile("temp", fileext = ".csv")
        suppressWarnings(.rxWithSink(t, {
          cat("t,c\n")
          out <- rxSolve(odeLin,
                         params = c(a = -8 / 3, b = -10),
                         events = et,
                         inits = c(X = 1, Y = 1, Z = 1),
                         addCov = TRUE,
                         covsInterpolation = "nocb",
                         method = meth
                         )
        }))

        lin.interp <- read.csv(t)
        unlink(t)

        lin.interp$c2 <- .cov_approx(lin.interp$t, lin.interp$c, cov.lin)

        test_that(paste0("Linear Approximation matches approxfun and overrides covsInterpolation (",
                       meth, ")"), {
          expect_equal(lin.interp$c, lin.interp$c2)
        })

        ## NONMEM interpolation
        suppressWarnings(.rxWithSink(t, {
          cat("t,c\n")
          out <- rxSolve(ode,
                         params = c(a = -8 / 3, b = -10),
                         events = et,
                         inits = c(X = 1, Y = 1, Z = 1),
                         covsInterpolation = "nocb", addCov = TRUE,
                         method = meth
                         )
        }))

        lin.interp <- read.csv(t)
        unlink(t)

        cov.lin <- approxfun(out$time, out$c,
                             yleft = cov$c[1], yright = cov$c[length(cov$c)],
                             method = "constant", f = 1
                             )
        lin.interp$c2 <- .cov_approx(lin.interp$t, lin.interp$c, cov.lin)

        test_that(paste0("NOCB Approximation similar to approxfun (", meth, ";", .homogenous, ")"), {
          expect_equal(lin.interp$c, lin.interp$c2)
        })

        suppressWarnings(.rxWithSink(t, {
          cat("t,c\n")
          out <- rxSolve(odeNocb,
                         params = c(a = -8 / 3, b = -10),
                         events = et,
                         inits = c(X = 1, Y = 1, Z = 1),
                         covsInterpolation = "locf", addCov = TRUE,
                         method = meth
                         )
        }))

        lin.interp <- read.csv(t)
        unlink(t)

        cov.lin <- approxfun(out$time, out$c,
                             yleft = cov$c[1], yright = cov$c[length(cov$c)],
                             method = "constant", f = 1
                             )
        lin.interp$c2 <- .cov_approx(lin.interp$t, lin.interp$c, cov.lin)

        test_that(paste0("NOCB Approximation similar to approxfun and overrides locf when in ode (",
                       meth, ")"), {
          expect_equal(lin.interp$c, lin.interp$c2)
        })

        ## midpoint interpolation
        suppressWarnings(.rxWithSink(t, {
          cat("t,c\n")
          out <- rxSolve(ode,
                         params = c(a = -8 / 3, b = -10),
                         events = et,
                         inits = c(X = 1, Y = 1, Z = 1),
                         covsInterpolation = "midpoint", addCov = TRUE,
                         method = meth
                         )
        }))
        lin.interp <- read.csv(t)
        unlink(t)

        cov.lin <- approxfun(out$time, out$c,
                             yleft = cov$c[1], yright = cov$c[length(cov$c)],
                             method = "constant", f = 0.5
                             )

        lin.interp$c2 <- .cov_approx(lin.interp$t, lin.interp$c, cov.lin)

        test_that(paste0("midpoint Approximation similar to approxfun (", meth, ";", .homogenous, ")"), {
          expect_equal(lin.interp$c, lin.interp$c2)
        })


        ## midpoint interpolation
        suppressWarnings(.rxWithSink(t, {
          cat("t,c\n")
          out <- rxSolve(odeMidpoint,
                         params = c(a = -8 / 3, b = -10),
                         events = et,
                         inits = c(X = 1, Y = 1, Z = 1),
                         covsInterpolation = "locf", addCov = TRUE,
                         method = meth
                         )
        }))
        lin.interp <- read.csv(t)
        unlink(t)

        cov.lin <- approxfun(out$time, out$c,
                             yleft = cov$c[1], yright = cov$c[length(cov$c)],
                             method = "constant", f = 0.5
                             )

        lin.interp$c2 <- .cov_approx(lin.interp$t, lin.interp$c, cov.lin)

        test_that(paste0("midpoint Approximation similar to approxfun and overrides covsInterpolation method (", meth, ";", .homogenous, ")"), {
          expect_equal(lin.interp$c, lin.interp$c2)
        })

        ## covs_interpolation
        suppressWarnings(.rxWithSink(t, {
          cat("t,c\n")
          out <- rxSolve(ode,
                         params = c(a = -8 / 3, b = -10),
                         events = et,
                         inits = c(X = 1, Y = 1, Z = 1),
                         covsInterpolation = "locf", addCov = TRUE,
                         method = meth
                         )
        }))

        lin.interp <- read.csv(t)
        unlink(t)

        cov.lin <- approxfun(out$time, out$c,
                             yleft = cov$c[1], yright = cov$c[length(cov$c)],
                             method = "constant")

        lin.interp$c2 <- .cov_approx(lin.interp$t, lin.interp$c, cov.lin)

        test_that(paste0("Constant Approximation similar to approxfun (", meth, ";", .homogenous, ")"), {
          expect_equal(lin.interp$c, lin.interp$c2)
        })

        ## covs_interpolation
        suppressWarnings(.rxWithSink(t, {
          cat("t,c\n")
          out <- rxSolve(odeLocf,
                         params = c(a = -8 / 3, b = -10),
                         events = et,
                         inits = c(X = 1, Y = 1, Z = 1),
                         covsInterpolation = "nocb", addCov = TRUE,
                         method = meth
                         )
        }))

        lin.interp <- read.csv(t)
        unlink(t)

        cov.lin <- approxfun(out$time, out$c,
                             yleft = cov$c[1], yright = cov$c[length(cov$c)],
                             method = "constant")

        lin.interp$c2 <- .cov_approx(lin.interp$t, lin.interp$c, cov.lin)

        test_that(paste0("Constant Approximation similar to approxfun and overrides covsInterpolation: ", meth), {
          expect_equal(lin.interp$c, lin.interp$c2)
        })

        out <- as.data.frame(out)
        out <- out[, names(out) != "c"]

        suppressWarnings(.rxWithSink(t, {
          out1 <-
            rxSolve(ode,
                    params = c(a = -8 / 3, b = -10, c = 0),
                    events = et,
                    inits = c(X = 1, Y = 1, Z = 1), addCov = TRUE,
                    method = meth
                    )
        }))
        unlink(t)

        out1 <- as.data.frame(out1)

        test_that(paste0("time varying covariates produce different outputs: ", meth), {
          expect_false(isTRUE(all.equal(out, out1)))
        })

        .t0 <- as.numeric(et0$get.EventTable()$time)
        cov <- data.frame(c = .t0 + 1, a = -.t0 / 100)

        et <- cbind(et0, cov)

        suppressWarnings(.rxWithSink(t, {
          out <- rxSolve(ode,
                         params = c(a = -8 / 3, b = -10),
                         events = et,
                         inits = c(X = 1, Y = 1, Z = 1),
                         addCov = TRUE,
                         method = meth
                         )

          out3 <- rxSolve(ode,
                          params = c(a = -8 / 3, b = -10),
                          events = et,
                          inits = c(X = 1, Y = 1, Z = 1),
                          addCov = TRUE,
                          method = meth
                          )
        }))
        unlink(t)

        test_that(paste0("time varying covariates output covariate in data frame: ", meth), {
          expect_equal(cov$c[et0$get.obs.rec()], out$c)
          expect_equal(cov$a[et0$get.obs.rec()], out$a)
        })

        cov <- data.frame(c = as.numeric(et0$get.EventTable()$time) + 1)
        et <- cbind(et0, cov)

        suppressWarnings(.rxWithSink(t, {
          out2 <- rxSolve(ode,
                          params = c(a = -8 / 3, b = -10),
                          events = et,
                          inits = c(X = 1, Y = 1, Z = 1),
                          addCov = TRUE,
                          method = meth
                          )
        }))
        unlink(t)

        test_that(paste0("Before assinging the time varying to -8/3, out and out2 should be different: ", meth), {
          expect_false(isTRUE(all.equal(out, out2)))
        })

        # context(sprintf("Test First Assignment (%s)", meth))

        ## Assign a time-varying to a simple parameter
        suppressWarnings(.rxWithSink(t, {
          out$a <- -8 / 3
        }))
        unlink(t)

        test_that(paste0("The out$a=-8/3 works: ", meth), {
          expect_equal(as.data.frame(out), as.data.frame(out2))
        })

        # context(sprintf("Test Second Assignment (%s)", meth))

        suppressWarnings(.rxWithSink(t, {
          out$a <- out3$a
        }))
        unlink(t)

        test_that(paste0("the out$a = time varying covariate works: ", meth), {
          expect_equal(as.data.frame(out), as.data.frame(out3))
        })

        # context(sprintf("Covariate solve with data frame event table (%s)", meth))

        ## Covariate solve for data frame
        d3 <- data.frame(
          TIME = c(0, 0, 2.99270072992701, 192, 336, 456),
          AMT = c(137L, 0L, -137L, 0L, 0L, 0L),
          V2I = c(909L, 909L, 909L, 909L, 909L, 909L),
          V1I = c(545L, 545L, 545L, 545L, 545L, 545L),
          CLI = c(471L, 471L, 471L, 471L, 471L, 471L),
          EVID = c(10101L, 0L, 10101L, 0L, 0L, 0L)
        )

        mod1 <- rxode2({
          d/dt(A_centr) <- -A_centr * (CLI / V1I + 204 / V1I) + 204 * A_periph / V2I
          d/dt(A_periph) <- 204 * A_centr / V1I - 204 * A_periph / V2I
          d/dt(A_circ) <- -4 * A_circ * exp(-ETA[2] - THETA[2]) + 4 * A_tr3 * exp(-ETA[2] - THETA[2])
          A_circ(0) <- exp(ETA[1] + THETA[1])
          d/dt(A_prol) <- 4 * A_prol * Rx_pow(exp(ETA[1] + THETA[1]) / A_circ, exp(THETA[4])) * (-A_centr * exp(ETA[3] + THETA[3]) / V1I + 1) * exp(-ETA[2] - THETA[2]) - 4 * A_prol * exp(-ETA[2] - THETA[2])
          A_prol(0) <- exp(ETA[1] + THETA[1])
          d/dt(A_tr1) <- 4 * A_prol * exp(-ETA[2] - THETA[2]) - 4 * A_tr1 * exp(-ETA[2] - THETA[2])
          A_tr1(0) <- exp(ETA[1] + THETA[1])
          d/dt(A_tr2) <- 4 * A_tr1 * exp(-ETA[2] - THETA[2]) - 4 * A_tr2 * exp(-ETA[2] - THETA[2])
          A_tr2(0) <- exp(ETA[1] + THETA[1])
          d/dt(A_tr3) <- 4 * A_tr2 * exp(-ETA[2] - THETA[2]) - 4 * A_tr3 * exp(-ETA[2] - THETA[2])
          A_tr3(0) <- exp(ETA[1] + THETA[1])
        })

        tmp <-
          rxSolve(
            mod1, d3,
            setNames(
              c(2.02103, 4.839305, 3.518676, -1.391113, 0.108127023, -0.064170725, 0.087765769),
              c(sprintf("THETA[%d]", 1:4), sprintf("ETA[%d]", 1:3))
            ),
            addCov = TRUE,
            method = meth
          )

        test_that(paste0("Data Frame single subject solve: ", meth), {
          expect_equal(
            tmp |>
              dplyr::select(CLI, V1I, V2I) |> as.data.frame(),
            d3 |> dplyr::filter(EVID == 0) |>
              dplyr::select(CLI, V1I, V2I) |> as.data.frame()
          )
          expect_equal(names(tmp$params), mod1$params[-(1:3)])
        })

        d3 <- data.frame(
          ID = c(1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L),
          TIME = c(0, 0, 2.99270072992701, 192, 336, 456, 0, 0, 3.07272727272727, 432),
          AMT = c(137L, 0L, -137L, 0L, 0L, 0L, 110L, 0L, -110L, 0L),
          V2I = c(909L, 909L, 909L, 909L, 909L, 909L, 942L, 942L, 942L, 942L),
          V1I = c(545L, 545L, 545L, 545L, 545L, 545L, 306L, 306L, 306L, 306L),
          CLI = c(471L, 471L, 471L, 471L, 471L, 471L, 405L, 405L, 405L, 405L),
          EVID = c(10101L, 0L, 10101L, 0L, 0L, 0L, 10101L, 0L, 10101L, 0L)
        )

        par2 <-
          matrix(
            c(
              2.02103, 4.839305, 3.518676, -1.391113, 0.108127023, -0.064170725, 0.087765769,
              2.02103, 4.839305, 3.518676, -1.391113, -0.064170725, 0.087765769, 0.108127023
            ),
            nrow = 2, byrow = T,
            dimnames = list(NULL, c(sprintf("THETA[%d]", 1:4), sprintf("ETA[%d]", 1:3)))
          )

        tmp <- rxSolve(mod1, d3, par2, addCov = TRUE, cores = 2, method = meth)

        test_that(paste0("Data Frame multi subject solve", meth), {
          expect_equal(
            tmp |>
              dplyr::select(CLI, V1I, V2I) |> as.data.frame(),
            d3 |>
              dplyr::filter(EVID == 0) |>
              dplyr::select(CLI, V1I, V2I) |> as.data.frame()
          )
          expect_equal(names(tmp$params)[-1], mod1$params[-(1:3)])
        })

        ## Now check missing covariate values.

        d3na <-
          data.frame(
            ID = c(1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L),
            TIME = c(0, 0, 2.99270072992701, 192, 336, 456, 0, 0, 3.07272727272727, 432),
            AMT = c(137L, 0L, -137L, 0L, 0L, 0L, 110L, 0L, -110L, 0L),
            V2I = c(909L, NA_integer_, 909L, 909L, 909L, 909L, 942L, 942L, 942L, 942L),
            V1I = c(545L, 545L, 545L, 545L, 545L, 545L, 306L, 306L, 306L, NA_integer_),
            CLI = c(471L, 471L, 471L, 471L, NA_integer_, 471L, 405L, 405L, 405L, 405L),
            EVID = c(10101L, 0L, 10101L, 0L, 0L, 0L, 10101L, 0L, 10101L, 0L)
          )

        expect_warning(
          tmp <- rxSolve(mod1, d3na, par2, addCov = TRUE, cores = 2, method = meth),
          NA)

        tmp2 <- rxSolve(mod1, d3, par2, addCov = TRUE, cores = 2, method = meth)

        # context(sprintf("Test NA extrapolation for %s solving", meth))
        test_that(paste0("NA solve is the same", meth), {
          for (i in c("id", "time", "A_centr", "A_periph", "A_circ", "A_prol", "A_tr1", "A_tr2", "A_tr3")) {
            expect_equal(tmp[[i]], tmp2[[i]])
          }
        })

        d3na <- data.frame(
          ID = c(1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L),
          TIME = c(0, 0, 2.99270072992701, 192, 336, 456, 0, 0, 3.07272727272727, 432),
          AMT = c(137L, 0L, -137L, 0L, 0L, 0L, 110L, 0L, -110L, 0L),
          V2I = c(909L, NA_integer_, 909L, 909L, 909L, 909L, 942L, 942L, 942L, 942L),
          V1I = c(545L, 545L, 545L, 545L, 545L, 545L, NA_integer_, NA_integer_, NA_integer_, NA_integer_),
          CLI = c(471L, 471L, 471L, 471L, NA_integer_, 471L, 405L, 405L, 405L, 405L),
          EVID = c(10101L, 0L, 10101L, 0L, 0L, 0L, 10101L, 0L, 10101L, 0L)
        )

        test_that(paste0("All covariates are NA give a warning", meth), {
          suppressWarnings(expect_warning(
            rxSolve(mod1, d3na, par2, addCov = TRUE, cores = 2, method = meth),
            "column 'V1I' has only 'NA' values for id '2'"))
        })
      }

    }) ## withr::with_options
  } ## for (.homogenous in ...)

  # time-varying covariates work with ODEs

  test_that("time varying covariates lhs", {

    dfadvan <- data.frame(
      ID = c(
        1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
        1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
        2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
        2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L
      ),
      TIME = c(
        0L, 1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 12L, 13L, 14L, 15L,
        16L, 17L, 18L, 19L, 20L, 21L, 22L, 23L, 24L, 0L, 1L, 2L, 3L,
        4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 12L, 13L, 14L, 15L, 16L,
        17L, 18L, 19L, 20L, 21L, 22L, 23L, 24L
      ), AMT = c(
        100L, 0L, 0L,
        0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 100L, 0L, 0L, 0L, 0L,
        0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 100L, 0L, 0L, 0L, 0L, 0L, 0L,
        0L, 0L, 0L, 0L, 0L, 0L, 100L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
        0L, 0L, 0L, 0L
      ), MDV = c(
        1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
        0L, 0L, 0L, 0L, 1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
        0L, 1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 1L, 0L,
        0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L
      ), CLCR = c(
        120L, 120L,
        120L, 120L, 120L, 120L, 120L, 120L, 120L, 120L, 120L, 120L, 120L,
        120L, 120L, 120L, 30L, 30L, 30L, 30L, 30L, 30L, 30L, 30L, 30L,
        30L, 30L, 30L, 30L, 30L, 30L, 120L, 120L, 120L, 120L, 120L, 120L,
        120L, 120L, 120L, 120L, 120L, 120L, 120L, 120L, 120L, 120L, 120L,
        120L, 120L, 120L, 120L
      )
    )

    mod <- rxode2({
      CLpop <- 2 # clearance
      Vpop <- 10 # central volume of distribution
      CL <- CLpop * (CLCR / 100)
      V <- Vpop
    })

    mod2 <- rxode2({
      CLpop <- 2 # clearance
      Vpop <- 10 # central volume of distribution
      CL <- CLpop * (CLCR / 100)
      V <- Vpop
      d / dt(matt) <- 0
    })

    x1 <- rxSolve(mod, dfadvan, keep = "CL")
    x2 <- rxSolve(mod2, dfadvan, keep = "CL")
    expect_equal(x1$CL, x2$CL)
  })
})

#' Format rxSolve and related objects as html.
#'
#' @param x rxode2 object
#' @param ... Extra arguments sent to kable
#' @return html code for rxSolve object
#' @author Matthew L. Fidler
#' @export
rxHtml <- function(x, ...) {
  UseMethod("rxHtml")
}
#' @rdname rxHtml
#' @export
rxHtml.rxSolve <- function(x, ...) {
  rxode2::rxReq("knitr")
  bound <- .getBound(x, parent.frame(2))
  cat("<style>
         th,td{
             padding:2px 5px 2px 5px;
         }
    </style>")
  cat("<table style=\"border-radius: 10px 10px 10px 10px; -moz-border-radius: 10px 10px 10px 10px; -webkit-border-radius: 10px 10px 10px 10px; border: 5px solid #7b3e3e;\"><tr><td style=\"text-align: center; padding: 5px;\"><strong>Solved rxode2 object</strong></td><tr><td><strong>Parameters</strong> (<span style=\"color: Orange;\">")
  cat(bound)
  cat("</span><strong \"color: Blue;\">&#36;params</strong>):</td></tr><tr><td>")
  cat(knitr::kable(x$params, "html", ...))
  df <- x$covs
  if (!is.null(df)) {
    cat(paste0("</td></tr><tr><td><strong>Covariates</strong> (<span style=\"color: Orange;\">", bound, "</span><strong \"color: Blue;\">&#36;covs</strong>):"))
    cat(knitr::kable(head(as.matrix(df)), "html", ...))
  }
  cat("</td></tr><tr><td><strong>Initial Conditions</strong> (<span style=\"color: Orange;\">", bound, "</span><strong \"color: Blue;\">&#36;inits</strong>):</td></tr><tr><td>")
  cat(paste0(knitr::kable(data.frame(t(x$inits)))))
  cat("<strong>First part of data (object):</strong>")
  cat(knitr::kable(head(as.data.frame(x)), "html", ...))
  cat("</td></tr></table>")
}

#' rxode2 progress bar functions
#'
#' `rxProgress` sets up the progress bar
#'
#' `rxTick` is a progress bar tick
#'
#' `rxProgressStop` stop progress bar
#'
#' `rxProgressAbort` shows an abort if `rxProgressStop`
#' wasn't called.
#'
#' @param num Tot number of operations to track
#' @param core Number of cores to show.  If below 1, don't show
#'     number of cores
#' @param clear Boolean telling if you should clear the progress bar
#'     after completion (as if it wasn't displayed).  By default this is TRUE
#' @param error With rxProgressAbort this is the error that is displayed
#' @return All return NULL invisibly.
#' @author Matthew L. Fidler
#' @examples
#' 
#' f <- function() {
#'   on.exit({
#'     rxProgressAbort()
#'   })
#'   rxProgress(100)
#'   for (i in 1:100) {
#'     rxTick()
#'     Sys.sleep(1 / 100)
#'   }
#'   rxProgressStop()
#' }
#' \donttest{
#' f()
#' }
#'
#' @export
rxProgress <- function(num, core = 0L) {
  invisible(.Call(`_rxProgress`, as.integer(num), as.integer(core)))
}

#' @rdname rxProgress
#' @export
rxTick <- function() {
  invisible(.Call(`_rxTick`))
}

#' @rdname rxProgress
#' @export
rxProgressStop <- function(clear = TRUE) {
  invisible(.Call(`_rxProgressStop`, as.integer(clear)))
}

#' @rdname rxProgress
#' @export
rxProgressAbort <- function(error = "Aborted calculation") {
  invisible(.Call(`_rxProgressAbort`, error))
}

#' Internal messaging statements
#'
#' @param text Text
#' @param ... Other arguments
#' @param .envir Environment to evaluate in
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.minfo <- function(text, ..., .envir = parent.frame()) {
  cli::cli_alert_info(gettext(text), ..., .envir = .envir)
}
#' @rdname dot-minfo
#' @export
.malert <- function(text, ..., .envir = parent.frame()) {
  cli::cli_alert(gettext(text), ..., .envir = .envir)
}
#' @rdname dot-minfo
#' @export
.mwarn <- function(text, ..., .envir = parent.frame()) {
  cli::cli_alert_warning(gettext(text), ..., .envir = .envir)
}
#' @rdname dot-minfo
#' @export
.msuccess <- function(text, ..., .envir = parent.frame()) {
  cli::cli_alert_success(gettext(text), ..., .envir = .envir)
}
#' Internal function to tell if the linCmt() is the model variables
#'
#'
#' @return 0 or 1
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
.rxIsLinCmt <- function() {
  .Call(`_rxode2_isLinCmt`)
}
#'  Internal function to generate the model variables for a linCmt() model
#'
#'
#' @param lenState Length of the state
#' @param vars Variables in the model
#' @return Model variables of expanded linCmt model
#' @author Matthew L. Fidler
#' @export
.rxLinCmtGen <- function(lenState, vars) {
  rxGetModel(.Call(
    `_rxode2_linCmtGen`,
    lenState, vars, 1L, FALSE))
}

.normalizePath <- function(path, ...) {
  ifelse(.Platform$OS.type == "windows",
    suppressWarnings(utils::shortPathName(normalizePath(path, ...))),
    ifelse(regexpr("^[/~]", path) != -1,
      suppressWarnings(normalizePath(path, ...)),
      suppressWarnings(normalizePath(file.path(getwd(), path), ...))
    )
  )
}

#' Use cat when rxode2.verbose is TRUE
#'
#' @param ... Parameters sent to cat
#' @author Matthew L. Fidler
#' @keywords internal
#' @return nothing
#' @export
rxCat <- function(a, ...) {
  ## nocov start
  if (getOption("rxode2.verbose", TRUE)) {
    if (is(a, "rxode2")) {
      message(rxode2::rxNorm(a), appendLF = FALSE)
    } else {
      message(a, ..., appendLF = FALSE)
    }
  }
  ## nocov end
}

#' Cleanup anonymous DLLs by unloading them
#'
#' This cleans up any rxode2 loaded DLLs
#'
#' @param wd What directory should be cleaned; (DEPRECIATED), this no
#'     longer does anything.
#'
#' This unloads all rxode2 anonymous dlls.
#'
#' @return TRUE if successful
#'
#' @author Matthew L. Fidler
#' @export
rxClean <- function(wd) {
  if (!missing(wd)) warning("'wd' is depreciated")
  rxUnloadAll()
  unlink(rxTempDir(), recursive = TRUE, force = TRUE)
  suppressMessages(.mkCache(rxTempDir()))
}

#' Defunct setting of sum
#'
#' @param type used to be type of product
#'
#' @return nothing
#'
#' @export
rxSetSum <- function(type = c("pairwise", "fsum", "kahan", "neumaier", "c")) {
  stop("'rxSetSum' has been moved to rxSolve(...,sum=)", call. = FALSE)
}

#' Defunct setting of product
#'
#' @param type used to be type of product
#' @return nothing
#'
#' @export
rxSetProd <- function(type = c("long double", "double", "logify")) {
  stop("'rxSetProd' has been moved to rxSolve(...,sum=)", call. = FALSE)
}

#' Set timing for progress bar
#'
#' @param seconds This sets the number of seconds that need to elapse
#'     before drawing the next segment of the progress bar.  When
#'     this is zero or below this turns off the progress bar.
#'
#' @return nothing, used for side effects
#'
#' @export
#' @author Matthew Fidler
rxSetProgressBar <- function(seconds = 1.0) {
  invisible(.Call(`_rxParProgress`, as.double(seconds)))
}



#' Error function
#'
#'
#' @param x vector or real values
#' @return erf of x
#' @author Matthew L. Fidler
#' @examples
#' erf(1.0)
#' @export
erf <- function(x) {
  checkmate::assertNumeric(x)
  .Call(`_rxode2_rxErf`, x, PACKAGE = "rxode2")
}
#' Gammap: normalized lower incomplete gamma function
#'
#' This is the gamma_p from the boost library
#'
#' @param a The numeric 'a' parameter in the normalized lower
#'   incomplete gamma
#'
#' @param z The numeric 'z' parameter in the normalized lower
#'   incomplete gamma
#'
#' @details
#'
#' The gamma p function is given by:
#'
#' gammap = lowergamma(a, z)/gamma(a)
#'
#' @return gammap results
#' @author Matthew L. Fidler
#' @examples
#'
#' gammap(1, 3)
#' gammap(1:3, 3)
#' gammap(1, 1:3)
#' @export
gammap <- function(a, z) {
  .Call(`_gammap`, a, z, PACKAGE = "rxode2")
}

#' Gammaq: normalized upper incomplete gamma function
#'
#' This is the gamma_q from the boost library
#'
#' @param a The numeric 'a' parameter in the normalized upper
#'   incomplete gamma
#'
#' @param z The numeric 'z' parameter in the normalized upper
#'   incomplete gamma
#'
#' @details
#'
#' The gamma q function is given by:
#'
#' gammaq = uppergamma(a, z)/gamma(a)
#'
#' @return gammaq results
#' @author Matthew L. Fidler
#' @examples
#'
#' gammaq(1, 3)
#' gammaq(1:3, 3)
#' gammaq(1, 1:3)
#' @export
gammaq <- function(a, z) {
  .Call(`_gammaq`, a, z, PACKAGE = "rxode2")
}
#' uppergamma:  upper incomplete gamma function
#'
#' This is the tgamma from the boost library
#'
#' @param a The numeric 'a' parameter in the upper
#'   incomplete gamma
#'
#' @param z The numeric 'z' parameter in the upper
#'   incomplete gamma
#'
#' @details
#'
#' The uppergamma function is given by:
#'
#' \eqn{uppergamma(a, z) = \int_{z}^{\infty}t^{a-1}\cdot e^{-t} dt}
#'
#' @return uppergamma results
#'
#' @author Matthew L. Fidler
#'
#' @examples
#'
#' uppergamma(1, 3)
#'
#' uppergamma(1:3, 3)
#'
#' uppergamma(1, 1:3)
#' @export
uppergamma <- function(a, z) {
  .Call(`_uppergamma`, a, z, PACKAGE = "rxode2")
}

#' lowergamma:  upper incomplete gamma function
#'
#' This is the tgamma_lower from the boost library
#'
#' @param a The numeric 'a' parameter in the upper
#'   incomplete gamma
#'
#' @param z The numeric 'z' parameter in the upper
#'   incomplete gamma
#'
#' @details
#'
#' The lowergamma function is given by:
#'
#' \deqn{lowergamma(a, z) = \int_{0}^{z}t^{a-1}\cdot e^{-t} dt}
#'
#' @return lowergamma results
#'
#' @author Matthew L. Fidler
#'
#' @examples
#'
#' lowergamma(1, 3)
#'
#' lowergamma(1:3, 3)
#'
#' lowergamma(1, 1:3)
#' @export
lowergamma <- function(a, z) {
  .Call(`_lowergamma`, a, z, PACKAGE = "rxode2")
}

#' gammapDer:  derivative of gammap
#'
#' This is the gamma_p_derivative from the boost library
#'
#' @param a The numeric 'a' parameter in the upper
#'   incomplete gamma
#'
#' @param z The numeric 'z' parameter in the upper
#'   incomplete gamma
#'
#' @return lowergamma results
#'
#' @author Matthew L. Fidler
#'
#' @examples
#'
#' gammapDer(1:3, 3)
#'
#' gammapDer(1, 1:3)
#' @export
gammapDer <- function(a, z) {
  .Call(`_gammapDer`, a, z, PACKAGE = "rxode2")
}

#' gammaqInv and gammaqInva:  Inverses of normalized gammaq function
#'
#' @param a The numeric 'a' parameter in the upper
#'   incomplete gamma
#'
#' @param x The numeric 'x' parameter in the upper incomplete gamma
#'
#' @param q The numeric 'q' parameter in the upper
#'   incomplete gamma
#'
#' @details
#'
#' With the equation:
#'
#' q = gammaq(a, x)
#'
#' The 'gammaqInv' function returns a value 'x' that satisfies the
#' equation above
#'
#' The 'gammaqInva' function returns a value 'a' that satisfies the
#' equation above
#'
#' NOTE: gammaqInva is slow
#'
#' @return inverse gammaq results
#'
#' @author Matthew L. Fidler
#'
#' @examples
#'
#' gammaqInv(1:3, 0.5)
#'
#' gammaqInv(1, 1:3 / 3)
#'
#' gammaqInv(1:3, 1:3 / 3.1)
#'
#' gammaqInva(1:3, 1:3 / 3.1)
#' @export
gammaqInv <- function(a, q) {
  .Call(`_gammaqInv`, a, q, PACKAGE = "rxode2")
}

#' @rdname gammaqInv
#' @export
gammaqInva <- function(x, q) {
  .Call(`_gammaqInva`, x, q, PACKAGE = "rxode2")
}


#' gammapInv and gammapInva:  Inverses of normalized gammap function
#'
#' @param a The numeric 'a' parameter in the upper
#'   incomplete gamma
#'
#' @param x The numeric 'x' parameter in the upper incomplete gamma
#'
#' @param p The numeric 'p' parameter in the upper incomplete gamma
#'
#' @details
#'
#' With the equation:
#'
#' p = gammap(a, x)
#'
#' The 'gammapInv' function returns a value 'x' that satisfies the
#' equation above
#'
#' The 'gammapInva' function returns a value 'q' that satisfies the
#' equation above
#'
#' NOTE: gammapInva is slow
#'
#' @return inverse gammap results
#'
#' @author Matthew L. Fidler
#'
#' @examples
#'
#' gammapInv(1:3, 0.5)
#'
#' gammapInv(1, 1:3 / 3.1)
#'
#' gammapInv(1:3, 1:3 / 3.1)
#'
#' gammapInva(1:3, 1:3 / 3.1)
#' @export
gammapInv <- function(a, p) {
  .Call(`_gammapInv`, a, p, PACKAGE = "rxode2")
}

#' @rdname gammapInv
#' @export
gammapInva <- function(x, p) {
  .Call(`_gammapInva`, x, p, PACKAGE = "rxode2")
}
#' rxode2 general transformation function
#'
#' @param x value that will be transformed
#' @param lambda lambda value for the transformation
#' @param transform transformation to use (can be integer or string
#'   matching supported transformations)
#' @param low lower bound for the transformation
#' @param high upper bound for the transformation
#' @param inverse boolean if the inverse transformation should be performed
#' @return transformed value
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
#' @examples
#'
#' logit(0.25)
#'
#' .rxTransform(0.25, transform="logit")
#'
#' expit(-1.09)
#'
#' .rxTransform(-1.09, transform="logit", inverse=TRUE)
#'
.rxTransform <- function(x, lambda=1.0,
                         low = 0.0, high = 1.0,
                         transform=c("boxCox", "yeoJohnson", "untransformed",
                                     "lnorm", "logit", "logit + yeoJohnson",
                                     "probit", "probit + yeoJohnson",
                                     "logit + boxCox", "probit + boxCox"),
                         inverse=FALSE) {
  .w <- which(is.na(x))
  if (length(.w) > 0L) {
    .ret <- numeric(length(x))
    .ret[.w] <- NA_real_
    .ret[-.w] <- .rxTransform(x[-.w], lambda, low, high, transform, inverse)
    return(.ret)
  }
  if (is.integer(transform)) {
  } else {
    transform <- factor(match.arg(transform),
                        levels=c("boxCox", "yeoJohnson", "untransformed",
                                 "lnorm", "logit", "logit + yeoJohnson",
                                 "probit", "probit + yeoJohnson", "logit + boxCox",
                                 "probit + boxCox"))
    transform <- as.integer(transform)-1L
  }
  if (length(lambda) > 1 ||
        length(low) > 1 ||
        length(high) > 1 ||
        length(transform) > 1 ||
        length(inverse) > 1) {
    .df <- data.frame(x = x, lambda = lambda, low = low, high = high,
                      transform=transform, inverse=inverse)
    vapply(seq_len(nrow(.df)),
           function(i) {
             .rxTransform(.df$x[i], .df$lambda[i], .df$low[i], .df$high[i],
                          .df$transform[i], .df$inverse[i])
    }, numeric(1), USE.NAMES = FALSE)
  } else {
    checkmate::assertNumeric(x, any.missing = FALSE)
    checkmate::assertNumeric(lambda, any.missing = FALSE)
    checkmate::assertNumeric(low, any.missing = FALSE)
    checkmate::assertNumeric(high, any.missing = FALSE)
    checkmate::assertInteger(transform, any.missing = FALSE)
    checkmate::assertLogical(inverse, any.missing = FALSE)
    .Call(`_rxode2_powerD`, x, low, high, lambda, transform, inverse)
  }
}

#' Transform log-Jacobian (powerL) and its lambda gradient (powerDL)
#'
#' Internal test hook; the underlying C functions are only shared through
#' `R_RegisterCCallable()` (consumed by nlmixr2est) and are otherwise
#' unreachable from R.
#'
#' @param x numeric vector on the transformed scale
#' @param lambda transform parameter
#' @param low,high transform bounds
#' @param transform integer transform code (same coding as `.rxTransform`)
#' @param dLambda when `TRUE` return the lambda gradient (`powerDL`) instead
#'   of the log-Jacobian (`powerL`)
#' @return numeric vector
#' @author Matthew L. Fidler
#' @noRd
.rxTransformL <- function(x, lambda = 1.0, low = 0.0, high = 1.0,
                          transform = 2L, dLambda = FALSE) {
  .Call(`_rxode2_powerLDL`, as.double(x), as.double(low), as.double(high),
        as.double(lambda), as.integer(transform), as.integer(dLambda))
}

#' logit and inverse logit (expit) functions
#'
#' @param x Input value(s) in range \[low,high\] to translate -Inf to
#'   Inf
#'
#' @param alpha Infinite value(s) to translate to range of \[low,
#'   high\]
#'
#' @param low Lowest value in the range
#'
#' @param high Highest value in the range
#'
#' @param mean logit-scale mean
#'
#' @param sd logit-scale standard deviation
#'
#' @inheritParams stats::integrate
#'
#' @param ... other parameters passed to `integrate()`
#'
#' @return  values from logit and expit
#'
#' @details
#'
#' logit is given by:
#'
#' logit(p) = -log(1/p-1)
#'
#' where:
#'
#' p = x-low/high-low
#'
#'  expit is given by:
#'
#' expit(p, low, high) = (high-low)/(1+exp(-alpha)) + low
#'
#' The `logitNormInfo()` gives the mean, variance and coefficient of
#' variability on the untransformed scale.
#'
#' @examples
#'
#' logit(0.25)
#'
#' expit(-1.09)
#'
#' logitNormInfo(logit(0.25), sd = 0.1)
#'
#' logitNormInfo(logit(1, 0, 10), sd = 1, low = 0, high = 10)
#'
#' @export
logit <- function(x, low = 0, high = 1) {
  .rxTransform(x, 1.0, low, high, 4L, FALSE)
}
#' @rdname logit
#' @export
expit <- function(alpha, low = 0, high = 1) {
  .rxTransform(alpha, 1.0, low, high, 4L, TRUE)
}

#' Handle arguments for ui functions
#'
#' Note this is an internal function but it is exported in case it is
#' useful.
#'
#' @param char This is the character equivalent of the argument
#' @param f This is the forced equivalent of the argument
#' @param dp This is deparsed expression
#' @return character representing the underlying rxode2 code for the argument
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
#' @examples
#'
#' .uiArg("1.0", 1.0, "1.0")
.uiArg <- function(char, f, dp) {
  if (!inherits(f, "try-error")) {
    if (is.numeric(f)) {
      return(as.character(f))
    }
    if (is.character(f)) {
      return(f)
    }
  }
  if (length(char) > 1) {
    dp
  } else {
    char
  }
}

#' @rdname logit
#' @export
logitNormInfo <- function(mean = 0, sd = 1, low = 0, high = 1, abs.tol = 1e-6, ...) {
  .fM1 <- function(x) {
    expit(x, low, high) * dnorm(x, mean = mean, sd = sd)
  }
  .m <- integrate(.fM1, -Inf, Inf, abs.tol = abs.tol, ...)$value
  .fV <- function(x){
    (expit(x, low, high) - .m)^2 * dnorm(x, mean = mean, sd = sd)
  }
  .v <- integrate(.fV, -Inf, Inf, abs.tol = abs.tol, ...)$value
  c(mean = .m, var = .v, cv = sqrt(.v) / .m)
}

#' probit and inverse probit functions
#'
#' @inheritParams logit
#' @return values from probit, probitInv and probitNormInfo
#' @examples
#'
#' probit(0.25)
#'
#' probitInv(-0.674)
#'
#' probitNormInfo(probit(0.25), sd = 0.1)
#'
#' probitNormInfo(probit(1, 0, 10), sd = 1, low = 0, high = 10)
#' @export
probit <- function(x, low = 0, high = 1) {
  .rxTransform(x, 1.0, low, high, 6L, FALSE)
}

#' @rdname probit
#' @export
probitInv <- function(x, low = 0, high = 1) {
  .rxTransform(x, 1.0, low, high, 6L, TRUE)
}


#' @rdname logit
#' @export
probitNormInfo <- function(mean = 0, sd = 1, low = 0, high = 1, abs.tol = 1e-6, ...) {
  .fM1 <- function(x) probitInv(x, low, high) * dnorm(x, mean = mean, sd = sd)
  .m <- integrate(.fM1, -Inf, Inf, abs.tol = abs.tol, ...)$value
  .fV <- function(x) (probitInv(x, low, high) - .m)^2 * dnorm(x, mean = mean, sd = sd)
  .v <- integrate(.fV, -Inf, Inf, abs.tol = abs.tol, ...)$value
  c(mean = .m, var = .v, cv = sqrt(.v) / .m)
}

#' boxCox/yeoJohnson and inverse boxCox/yeoJohnson functions
#'
#' @param x input value(s) to transform
#' @param lambda lambda value for the transformation
#' @return values from boxCox and boxCoxInv
#' @export
#' @examples
#'
#' boxCox(10, 0.5)
#'
#' boxCoxInv(4.32, 0.5)
#'
#' yeoJohnson(10, 0.5)
#'
#' yeoJohnsonInv(4.32, 0.5)
#'
boxCox <- function(x, lambda = 1.0) {
  checkmate::assertNumeric(x, lower=0.0, any.missing=FALSE)
  .rxTransform(x, lambda, low=0.0, high=1.0, 0L, FALSE)
}
#' @rdname boxCox
#' @export
boxCoxInv <- function(x, lambda = 1.0) {
  .rxTransform(x, lambda, low=0.0, high=1.0, 0L, TRUE)
}
#' @rdname boxCox
#' @export
yeoJohnson <- function(x, lambda = 1.0) {
  .rxTransform(x, lambda, low=0.0, high=1.0, 1L, FALSE)
}
#' @rdname boxCox
#' @export
yeoJohnsonInv <- function(x, lambda = 1.0) {
  .rxTransform(x, lambda, low=0.0, high=1.0, 1L, TRUE)
}
#' Get/Set the number of threads that rxode2 uses
#'
#' @param threads NULL (default) rereads environment variables
#'   specified by OpenMP (`OMP_THREAD_LIMIT`).  If they are not
#'   specified, this determines based the logical cores available. 0
#'   means to use all logical CPU threads available. Otherwise a number >= 1
#'
#' @param percent If provided it should be a number between 2 and
#'   100; the percentage of logical CPUs to use. By default on
#'   startup, 50 percent.
#'
#' @param throttle 2 (default) means that, roughly speaking, a
#'   single thread will be used when number subjects solved for is <=2, 2 threads when
#'   the number of all points is <=4, etc. The throttle is to speed up small data
#'   tasks (especially when repeated many times) by not incurring the
#'   overhead of managing multiple threads.
#'
#'   The throttle will also suppress sorting which ID will be solved first
#'   when there are (nsubject solved)*throttle <= nthreads.  In
#'   `rxode2` this sorting occurs to minimize the time for waiting for
#'   another thread to finish. If the last item solved is has a long
#'   solving time, all the other solving have to wait for that last
#'   costly solving to occur. If the items which are likely to take
#'   more time are solved first, this wait is less likely to have an
#'   impact on the overall solving time.
#'
#'   In rxode2 the ids are sorted by the individual number of solving
#'   points (largest first). It also has a C interface that allows
#'   these ids to be resorted by total time spent solving the
#'   equation.  This allows packages like nlmixr to sort by solving
#'   time if needed.
#'
#'   Overall the the number of threads is throttled (restricted) for
#'   small tasks and sorting for ids are suppressed.
#'
#' @param verbose Display the value of relevant OpenMP settings
#' @return number of threads that rxode2 uses
#' @export
getRxThreads <- function(verbose = FALSE) {
  .Call(`getRxThreads_R`, verbose)
}

#' @rdname getRxThreads
#' @export
setRxThreads <- function(threads = NULL, percent = NULL, throttle = NULL) {
  if (!missing(percent)) {
    if (!missing(threads)) stop("provide either threads= or percent= but not both")
    if (length(percent) != 1) stop("percent= is provided but is length ", length(percent))
    percent <- as.integer(percent)
    if (is.na(percent) || percent < 2L || percent > 100L) stop("percent==", percent, " but should be a number between 2 and 100")
    invisible(.Call(`setRxthreads`, percent, TRUE, as.integer(throttle)))
  } else {
    invisible(.Call(`setRxthreads`, as.integer(threads), FALSE, as.integer(throttle)))
  }
}

#' @rdname getRxThreads
#' @export
rxCores <- getRxThreads

.rxUnloadAllEnv <- new.env(parent=emptyenv())
.rxUnloadAllEnv$reallyUnload <- TRUE

#' Unloads all rxode2 compiled DLLs
#'
#' @param set If specified and `TRUE`, unloads all rxode2 dlls.  If
#'   specified and `FALSE`, block a simple `rxUnloadAll()` will not
#'   actually unload all dlls (helps with CRAN ASAN check)
#'
#' @return boolean telling if `rxUnloadAll()` completed the unloading procedure
#'
#' @examples
#'
#' # print(rxUnloadAll())
#' @export
#'
rxUnloadAll <- function(set=TRUE) {
  if (missing(set) && !.rxUnloadAllEnv$reallyUnload) {
    return(invisible(FALSE))
  }
  if (!missing(set) && isTRUE(set)) {
    .rxUnloadAllEnv$reallyUnload <- TRUE
  }
  if (.rxUnloadAllEnv$reallyUnload  && isTRUE(set)) {
    #try(rxUnloadAll_(), silent = TRUE)
  } else if (isTRUE(set)) {
    return(invisible(FALSE))
  } else if (isFALSE(set)) {
    .rxUnloadAllEnv$reallyUnload <- FALSE
    return(invisible(FALSE))
  }
  # First slice down on the keep
  .nKeep <- getOption("rxode2.dontUnload", 10)
  if (checkmate::testIntegerish(.nKeep, lower = 0, len = 1)) {
    .nKeep <- as.integer(.nKeep)
  } else {
    .nKeep <- 10L
  }
  .nKeep <- as.integer(.nKeep)
  if (.nKeep <= 0L) {
    .rxLastModels <- NULL
  } else if (length(.rxLastModels) < .nKeep) {
    .rxLastModels <- .rxLastModels[!is.na(.rxLastModels)]
    .rxLastModels <- .rxLastModels[seq(1, .nKeep)]
    assignInMyNamespace(".rxLastModels", .rxLastModels)
  }
  .ret <- try(rxUnloadAll_(), silent = TRUE)
  # Now Look for orphan rxode2 DLLs
  .loadedDLLs <- getLoadedDLLs()
  .checkOrphans <- getOption("rxode2.checkOrphans", 75)
  if (length(.loadedDLLs) >= .checkOrphans) {
    .orphans <- character(0)
    .n <- 0L
    for (i in seq_along(.loadedDLLs)) {
      .dll <- .loadedDLLs[[i]]
      .name <- .dll[["name"]]
      .path <- .dll[["path"]]
      .mv <- paste0(.name, "_model_vars")

      .mv <- try(eval(str2lang(paste0(".", "Call(", deparse1(.mv), ")")), envir=globalenv()), silent=TRUE)
      if (!inherits(.mv, "try-error")) {
        # This is a rxode2 DLL
        .md5 <- .mv$md5["parsed_md5"]
        if (.rxShouldUnload(.md5)) {
          .n <- .n + 1L
          # These are orphan models that should be unloaded
          # unload
          dyn.unload(.path)
          # and remove information
          .info <- .rxGetModelInfoFromDll(.path)
          for (i in .info) {
            .t <- try(exists(i, envir = .rxModels), silent=TRUE)
            if (isTRUE(.t)) {
              rm(list = i, envir = .rxModels)
            } else if (inherits(.t, "try-error")) {
              try(rm(list = i, envir = .rxModels), silent = TRUE)
            }
          }
        } else {
          # This is a model for consideration later
          .orphans <- c(.orphans, .path)
        }
      }
    }
    .nKeep <- .nKeep - .n
    if (length(.orphans) > .nKeep && .nKeep >= 0L) {
      # If there are more orphans than the number of models to keep,
      # then we will remove the last nKeep orphans
      .orphans <- .orphans[-seq_len(.nKeep)]

      for (o in .orphans) {
        # Remove the orphan DLLs
        dyn.unload(o)
        # and remove information
        .info <- .rxGetModelInfoFromDll(.path)
        for (i in .info) {
          .t <- try(exists(i, envir = .rxModels), silent=TRUE)
          if (isTRUE(.t)) {
            rm(list = i, envir = .rxModels)
          } else if (inherits(.t, "try-error")) {
            try(rm(list = i, envir = .rxModels), silent = TRUE)
          }
        }
      }
    }
  }
  invisible(TRUE)
}
#' With one sink, then release
#'
#' @param file the path to the file sink while running the `code`
#'
#' @param code The code to run during the sink
#'
#' @return Will return the results of the `code` section
#'
#' @details
#'
#' `.rxWithSink` captures output from `cat`
#'
#' `.rxWithSinkBoth` captures output from `cat` and `message`
#'
#' @export
#'
#' @keywords internal
#'
#' @author Matthew Fidler
#'
#' @examples
#'
#' t <- tempfile()
#' .rxWithSink(t, cat("message\n"))
#' cat("cat2\n") # now you can see the cat2
#' lines <- readLines(t)
#' unlink(t)
.rxWithSink <- function(file, code) {
  sink(file) # nolint
  on.exit(sink()) # nolint
  force(code)
}

#' @rdname dot-rxWithSink
#' @export
.rxWithSinkBoth <- function(file, code) {
  zz <- file(file, open = "wt")
  sink(zz) # nolint
  sink(zz, type = "message") # nolint
  on.exit({
    sink() # nolint
    sink(type = "message") # nolint
    close(zz)
  })
  force(code)
}


#' Temporarily set options then restore them while running code
#'
#' @param ops list of options that will be temporarily set for the
#'   `code`
#'
#' @inheritParams .rxWithSink
#'
#' @return value of code
#'
#' @export
#' @examples
#'
#' .rxWithOptions(list(digits = 21), {
#'   print(pi)
#' })
#'
#' print(pi)
.rxWithOptions <- function(ops, code) {
  .old <- options() # nolint
  do.call(options, as.list(ops)) # nolint
  on.exit({
    options(.old) # nolint
  })
  force(code)
}


#' Temporarily set options then restore them while running code
#'
#' @param wd working directory to temporarily set the system to while
#'   evaluating the code
#'
#' @return value of code
#'
#' @inheritParams .rxWithSink
#'
#' @export
#' @examples
#'
#' .rxWithWd(tempdir(), {
#'   getwd()
#' })
#'
#' getwd()
.rxWithWd <- function(wd, code) {
  .old <- getwd() # nolint
  on.exit({
    setwd(.old) # nolint
  })
  setwd(wd) # nolint
  force(code)
}

.qassert <- function(x, rules, .var.name = checkmate::vname(x)) {
  .val <- try(checkmate::qassert(x, rules, .var.name = .var.name), silent = TRUE)
  if (inherits(.val, "try-error")) {
    return(attr(.val, "condition")$message)
  }
  ""
}

use.utf <- function() {
  opt <- getOption("cli.unicode", NULL)
  if (!is.null(opt)) {
    isTRUE(opt)
  } else {
    l10n_info()$`UTF-8` && !is.latex()
  }
}

is.latex <- function() {
  if (!("knitr" %in% loadedNamespaces())) {
    return(FALSE)
  }
  get("is_latex_output", asNamespace("knitr"))()
}


.nsToLoad <- function() {
  vapply(rxode2parseGetPackagesToLoad(),
         function(pkg) {
           requireNamespace(pkg, quietly = TRUE)
         }, logical(1))
}

#' Check if a language object matches a template language object
#'
#' \itemize{
#'   \item{If \code{template == str2lang(".")}, it will match anything.}
#'   \item{If \code{template == str2lang(".name")}, it will match any name.}
#'   \item{If \code{template == str2lang(".call()")}, it will match any call.}
#' }
#'
#' @param x The object to check
#' @param template The template object it should match
#' @return TRUE if it matches, FALSE, otherwise
#' @keywords Internal
#' @examples
#' .matchesLangTemplate(str2lang("d/dt(foo)"), str2lang("d/dt(.name)"))
#' .matchesLangTemplate(str2lang("d/dt(foo)"), str2lang("d/foo(.name)"))
#' .matchesLangTemplate(str2lang("d/dt(foo)"), str2lang("d/."))
#' @export
.matchesLangTemplate <- function(x, template) {
  if (identical(template, as.name("."))) {
    ret <- TRUE
  } else if (is.name(x) && identical(template, as.name(".name"))) {
    ret <- TRUE
  } else if (is.call(x) && identical(template, str2lang(".call()"))) {
    ret <- TRUE
  } else {
    # A more specific match is needed
    ret <- all(class(x) == class(template))
    if (ret) {
      if (length(x) == length(template)) {
        if (length(x) > 1) {
          for (idx in seq_along(x)) {
            ret <- ret && .matchesLangTemplate(x[[idx]], template[[idx]])
          }
        } else if (is.name(x)) {
          # Check for a value if the name is not ".name"
          ret <- x == template
        } else {
          # Require identical for one-length calls (e.g. `linCmt()`), numeric,
          # character, etc.
          ret <- identical(x, template)
        }
      } else {
        ret <- FALSE
      }
    }
  }
  ret
}
#' Print out a table in the documentation
#'
#' @param table data frame
#' @param caption a character vector representing the caption for the latex table
#' @return based on the `knitr` context:
#' - output a `kableExtra::kbl` for `latex` output
#' - output a `DT::datatable` for html output
#' - otherwise output a `knitr::kable`
#' @keywords internal
#' @export
#' @author Matthew L. Fidler
#' @examples
#' .rxDocTable(rxReservedKeywords)
.rxDocTable <- function(table, caption="none") {
  rxReq("knitr")
  if (knitr::is_latex_output()) {
    rxReq("kableExtra")
    kableExtra::kbl(table, longtable=TRUE, booktabs=TRUE, caption=caption) |>
      kableExtra::kable_styling(latex_options=c("repeat_header", "striped", "hold_position"))
  } else if (knitr::is_html_output(excludes = "gfm")) {
    rxReq("DT")
    DT::datatable(table, rownames = FALSE, filter="top",  options=list(pageLength = 5, scrollX=TRUE))
  } else {
    knitr::kable(table)
  }
}

#' Calculate expected confidence bands or prediction intreval with normal or t sampling distribution
#'
#' The generic function `meanProbs` produces expected confidence bands
#' under either the t distribution or the normal sampling
#' distribution. This uses `qnorm()` or `qt()` with the mean and
#' standard deviation.
#'
#' For a single probability, p, it uses either:
#'
#' mean + qt(p, df=n)*sd/sqrt(n)
#'
#' or
#'
#' mean + qnorm(p)*sd/sqrt(n)
#'
#' The smallest observation corresponds to a probability of 0 and the
#' largest to a probability of 1 and the mean corresponds to 0.5.
#'
#' The mean and standard deviation of the sample is calculated based
#' on Welford's method for a single pass.
#'
#' This is meant to perform in the same way as `quantile()` so it can
#' be a drop in replacement for code using `quantile()` but using
#' distributional assumptions.
#'
#' @param x numeric vector whose mean and probability based confidence
#'   values are wanted, NA and NaN values are not allowed in numeric
#'   vectors unless 'na.rm' is 'TRUE'.
#' @param probs numeric vector of probabilities with values in the
#'   interval from 0 to 1 .
#' @param na.rm logical; if true, any NA and NaN's are removed from
#'   `x` before the quantiles are computed.
#' @param names logical; if true, the result has a names attribute.
#' @param useT logical; if true, use the t-distribution to calculate
#'   the confidence-based estimates. If false use the normal
#'   distribution to calculate the confidence based estimates.
#' @param onlyProbs logical; if true, only return the probability
#'   based confidence interval estimates, otherwise return
#' @param pred logical; if true use the prediction interval instead of
#'   the confidence interval
#' @param n integer/integerish; this is the n used to calculate the
#'   prediction or confidence interval.  When `n=0` (default) use the
#'   number of non-`NA` observations.
#' @param ... Arguments passed to default method, allows many
#'   different methods to be applied.
#' @return By default the return has the probabilities as names (if
#'   named) with the points where the expected distribution are
#'   located given the sampling mean and standard deviation. If
#'   `onlyProbs=FALSE` then it would prepend mean, variance, standard
#'   deviation, minimum, maximum and number of non-NA observations.
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' quantile(x<- rnorm(1001))
#' meanProbs(x)
#'
#' # Can get some extra statistics if you request onlyProbs=FALSE
#' meanProbs(x, onlyProbs=FALSE)
#'
#' x[2] <- NA_real_
#'
#' meanProbs(x, onlyProbs=FALSE)
#'
#' quantile(x<- rnorm(42))
#'
#' meanProbs(x)
#'
#' meanProbs(x, useT=FALSE)
#'
meanProbs <- function(x, ...) {
  UseMethod("meanProbs")
}

#' @rdname meanProbs
#' @export
meanProbs.default <- function(x, probs=seq(0, 1, 0.25), na.rm=FALSE,
                              names=TRUE, useT=TRUE, onlyProbs=TRUE, pred=FALSE,
                              n=0L, ...) {
  checkmate::assertNumeric(x)
  checkmate::assertNumeric(probs, min.len=1, any.missing = FALSE, lower=0.0, upper=1.0)
  checkmate::assertLogical(na.rm, any.missing=FALSE, len=1)
  checkmate::assertLogical(names, any.missing=FALSE, len=1)
  checkmate::assertLogical(useT, any.missing=FALSE, len=1)
  checkmate::assertLogical(onlyProbs, any.missing=FALSE, len=1)
  checkmate::assertLogical(pred, any.missing=FALSE, len=1)
  checkmate::assertIntegerish(n, min.len=1, max.len=1, any.missing=FALSE, lower=0)
  n <- as.integer(n)
  .ret <- .Call(`_rxode2_meanProbs_`, x, probs, na.rm, useT, pred, n)
  .names <- NULL
  if (names) {
    .names <- paste0(probs*100, "%")
  }
  if (onlyProbs) {
    .ret <- .ret[-1L:-6L]
    if (names) {
      names(.ret) <- .names
    }
  } else if (names) {
    names(.ret) <- c("mean","var", "sd", "min", "max", "n", .names)
  }
  .ret
}

#' Calculate expected confidence bands with binomial sampling distribution
#'
#' This is meant to perform in the same way as `quantile()` so it can
#' be a drop in replacement for code using `quantile()` but using
#' distributional assumptions.
#'
#' It is used for confidence intervals with rxode2 solved objects using
#' `confint(mean="binom")`
#'
#' @param x numeric vector whose mean and probability based confidence
#'   values are wanted, NA and NaN values are not allowed in numeric
#'   vectors unless `na.rm` is `TRUE`.
#'
#' @param probs numeric vector of probabilities with values in the
#'   interval 0 to 1, inclusive. When 0, it represents the maximum
#'   observed, when 1, it represents the maximum observed. When 0.5 it
#'   represents the expected probability (mean).
#'
#' @param na.rm logical; if true, any NA and NaN's are removed from
#'   `x` before the quantiles are computed.
#'
#' @param names logical; if true, the result has a names attribute.
#'
#' @param onlyProbs logical; if true, only return the probability
#'   based confidence interval/prediction interval estimates,
#'   otherwise return extra statistics.
#'
#' @param n integer/integerish; this is the n used to calculate the
#'   prediction or confidence interval.  When `n=0` (default) use the
#'   number of non-`NA` observations.  When calculating the prediction
#'   interval, this represents the number of observations used in the
#'   input ("true") distribution.
#'
#' @param pred Use a prediction interval instead of a confidence
#'   interval.  By default this is `FALSE`.
#'
#' @param m integer.  When using the prediction interval this
#'   represents the number of samples that will be observed in the
#'   future for the prediction interval.
#'
#' @param piMethod gives the prediction interval method (currently only lim) from Lu 2020
#'
#' @param M number of simulations to run for the LIM PI.
#'
#' @param tol tolerance of root finding in the LIM prediction interval
#'
#' @param ciMethod gives the method for calculating the confidence
#'   interval.
#'
#'  Can be:
#'
#'  - "argestiCoull" or "ac" -- Agresti-Coull method. For a 95\% confidence
#'     interval, this method does not use the concept   of "adding 2
#'     successes and 2 failures," but rather uses the formulas explicitly
#'     described in the following link:
#'
#' https://en.wikipedia.org/wiki/Binomial_proportion_confidence_interval#Agresti-Coull_Interval.
#'
#'   - "wilson" -- Wilson Method
#'
#'   - "wilsonCorrect" or "wc" -- Wilson method with continuity correction
#'
#'   - "wald" -- Wald confidence interval or standard z approximation.
#'
#' @param ... Arguments passed to default method, allows many
#'   different methods to be applied.
#'
#' @return By default the return has the probabilities as names (if
#'   named) with the points where the expected distribution are
#'   located given the sampling mean and standard deviation. If
#'   `onlyProbs=FALSE` then it would prepend mean, variance, standard
#'   deviation, minimum, maximum and number of non-NA observations.
#'
#' @export
#' @author Matthew L. Fidler
#' @references
#'
#' - Newcombe, R. G. (1998). "Two-sided confidence intervals for the single
#'   proportion: comparison of seven methods". Statistics
#'   in Medicine. 17 (8):
#'   857-872. doi:10.1002/(SICI)1097-0258(19980430)17:8<857::AID-SIM777>3.0.CO;2-E. PMID
#'   9595616.
#'
#' - Hezhi Lu, Hua Jin,
#'   A new prediction interval for binomial random variable based on inferential models,
#'   Journal of Statistical Planning and Inference,
#'   Volume 205,
#'   2020,
#'   Pages 156-174,
#'   ISSN 0378-3758,
#'   https://doi.org/10.1016/j.jspi.2019.07.001.
#' @examples
#'
#' x<- rbinom(7001, p=0.375, size=1)
#' binomProbs(x)
#'
#' # you can also use the prediction interval
#' \donttest{
#' binomProbs(x, pred=TRUE)
#' }
#'
#' # Can get some extra statistics if you request onlyProbs=FALSE
#' binomProbs(x, onlyProbs=FALSE)
#'
#' x[2] <- NA_real_
#'
#' binomProbs(x, onlyProbs=FALSE)
#'
#' binomProbs(x, na.rm=TRUE)
#'
binomProbs <- function(x, ...) {
  UseMethod("binomProbs")
}

#' @rdname binomProbs
#' @export
binomProbs.default <- function(x, probs=c(0.025, 0.05, 0.5, 0.95, 0.975), na.rm=FALSE,
                               names=TRUE, onlyProbs=TRUE, n=0L, m=0L,
                               pred=FALSE,
                               piMethod="lim", M=500000,
                               tol=.Machine$double.eps^0.25,
                               ciMethod=c("wilson", "wilsonCorrect", "agrestiCoull", "wald", "wc", "ac"), ...) {
  checkmate::assertNumeric(x, min.len=1, lower=0.0, upper=1.0)
  x <- as.double(x)
  checkmate::assertIntegerish(n, min.len=1, lower=0, any.missing=FALSE)
  n <- as.integer(n)
  checkmate::assertIntegerish(m, min.len=1, lower=0, any.missing=FALSE)
  m <- as.integer(m)
  checkmate::assertNumeric(probs, min.len=1, any.missing = FALSE, lower=0.0, upper=1.0)
  checkmate::assertLogical(na.rm, any.missing=FALSE, len=1)
  checkmate::assertLogical(names, any.missing=FALSE, len=1)
  checkmate::assertLogical(onlyProbs, any.missing=FALSE, len=1)
  if (pred) {
    .m <- mean(x, na.rm=na.rm)
    if (is.na(.m)) {
      .ret <- stats::quantile(NULL,probs=probs)
      if (!onlyProbs) {
        .ret <- c("mean"=NA_real_,"var"=NA_real_, "sd"=NA_real_, "n"=NA_real_,
                  .ret)
      }
    } else {
      .nC <- sum(!is.na(x))
      if (n == 0L) n <- as.integer(.nC)
      if (m == 0L) m <- as.integer(.nC)
      .Y <- round(.nC * .m) # number of successes
      .ret <- stats::quantile(.Call(`_rxode2_binomProbsPredVec_`, n, m, .Y, M, TRUE, tol),
                       probs=probs)
      if (!onlyProbs) {
        .ret <- c("mean"=.m,"var"=.m * (1.0 - .m), "sd"=sqrt(.m * (1.0 - .m)), "n"=.nC,
                  .ret)
      }
    }
    if (!names) {
      names(.ret) <- NULL
    }
    .ret
  } else {
    ciMethod <- match.arg(ciMethod)
    ciMethod <- setNames(c("wilson"=1L, "wilsonCorrect"=0L, "agrestiCoull"=3L, "wald"=2L, "ac"=3L, "wc"=0L)[ciMethod], NULL)
    .ret <- .Call(`_rxode2_binomProbs_`, x, probs, na.rm, n, ciMethod)
    .names <- NULL
    if (names) {
      .names <- paste0(probs*100, "%")
    }
    if (onlyProbs) {
      .ret <- .ret[-1L:-4L]
      if (names) {
        names(.ret) <- .names
      }
    } else if (names) {
      names(.ret) <- c("mean","var", "sd", "n", .names)
    }
    .ret
  }
}



#' Convert a factor/char to an id
#'
#' @param a value to convert to an id
#' @return id factor
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
#' @examples
#' .convertId("a")
.convertId <- function(a) {
  .Call(`_rxode2_convertId_`, a)
}

#' Get the internal breakdown of an evid
#'
#' @param i evid to breakdown
#' @return named evid integer vector
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
#' @examples
#'
#' .getWh(1001)
#' .getWh(10401)
#'
.getWh <- function(i) {
  checkmate::assertIntegerish(i,len=1, any.missing=FALSE)
  .Call(`_rxode2_getWh`, as.integer(i))
}

#' This converts NONMEM-style EVIDs to classic RxODE events
#'
#' @param cmt compartment flag
#' @param amt dose amount
#' @param rate dose rate
#' @param dur dose duration
#' @param ii inter-dose interval
#' @param evid event id
#' @param ss steady state
#' @return classic evids, excluding evids that are added (you need to
#'   add them manually) or simply use etTran.  This is mostly for
#'   testing and really shouldn't be used directly.
#' @export
#' @author Matthew L. Fidler
#' @examples
#' .toClassicEvid(cmt=10, amt=3, evid=1)
#' .toClassicEvid(cmt=10, amt=3, rate=2, evid=1)
#' .toClassicEvid(cmt=10, amt=3, rate=-1, evid=1)
#' .toClassicEvid(cmt=10, amt=3, rate=-2, evid=1)
#' .toClassicEvid(cmt=10, amt=3, dur=2, evid=1)
#' .toClassicEvid(cmt=304, amt=3, dur=2, evid=1)
#' .toClassicEvid(cmt=7, amt=0, rate=2, evid=1, ss=1)
#' .toClassicEvid(cmt=-10, amt=3, evid=1)
#' .toClassicEvid(cmt=10, amt=3, evid=5)
#' .toClassicEvid(cmt=6, amt=3, evid=6)
#' .toClassicEvid(cmt=6, amt=3, evid=7)
#' .toClassicEvid(evid=2)
#' .toClassicEvid(evid=4)
.toClassicEvid <- function(cmt=1L, amt=0.0, rate=0.0, dur=0.0, ii=0.0, evid=0L, ss=0.0) {
  .w <- which(is.na(cmt))
  if (length(.w) > 0) cmt[.w] <- 1
  checkmate::assertIntegerish(cmt)
  checkmate::assertIntegerish(evid, any.missing=FALSE)
  checkmate::assertNumeric(amt)
  checkmate::assertNumeric(dur, any.missing=FALSE)
  checkmate::assertNumeric(ii)
  checkmate::assertNumeric(ss)
  .df <- data.frame(cmt=as.integer(cmt), evid=as.integer(evid), amt=as.double(amt),
                    rate=as.double(rate), dur=as.double(dur),
                    ii=as.double(ii),
                    ss=as.double(ss))
  .Call(`_rxode2_getClassicEvid`,
        .df$cmt, .df$amt, .df$rate, .df$dur,
        .df$ii, .df$evid, .df$ss)
}

.rxDerivedReg <- rex::rex(
  start,
  or(
    group(or("V", "Q", "VP", "VT", "CLD"), number),
    "KA", "VP", "VT", "CLD", "V", "VC", "CL", "VSS", "K", "KE", "KEL",
    "Q", "VT", group("K", number, number), "AOB", "ALPHA", "BETA", "GAMMA",
    "A", "B", "C"
  ),
  end
)


#' Calculate derived parameters for the 1-, 2-, and 3- compartment
#' linear models.
#'
#' This calculates the derived parameters based on what is provided
#' in a data frame or arguments
#'
#' @param ... The input can be:
#'
#'
#'  * A data frame with PK parameters in it; This should ideally
#'  be a data frame with one pk parameter per row since it will
#'  output a data frame with one PK parameter per row.
#'
#'  * PK parameters as either a vector or a scalar
#'
#'
#' @param verbose boolean that when TRUE provides a message about the detected pk parameters
#'   and the detected compartmental model.  By default this is `FALSE`.
#'
#' @param digits represents the number of significant digits for the
#'   output; If the number is zero or below (default), do not round.
#'
#' @return Return a data.frame of derived PK parameters for a 1-, 2-,
#'   or 3-compartment linear model given provided clearances and
#'   volumes based on the inferred model type.
#'
#' The model parameters that will be provided in the data frame are:
#'
#' * `vc`: Central Volume (for 1-, 2- and 3-
#'   compartment models)
#'
#' * `kel`: First-order elimination rate (for 1-, 2-, and
#'   3-compartment models)
#'
#' * `k12`: First-order rate of transfer from central to
#'   first peripheral compartment; (for 2- and 3-compartment models)
#'
#' * `k21`: First-order rate of transfer from first
#'   peripheral to central compartment, (for 2- and 3-compartment
#'   models)
#'
#' * `k13`: First-order rate of transfer from central to
#'   second peripheral compartment; (3-compartment model)
#'
#' * `k31`: First-order rate of transfer from second
#'   peripheral to central compartment (3-compartment model)
#'
#' * `vp`: Peripheral Volume (for 2- and 3- compartment models)
#'
#' * `vp2`: Peripheral Volume for 3rd compartment (3- compartment model)
#'
#' * `vss`: Volume of distribution at steady state; (1-, 2-, and 3-compartment models)
#'
#' * `t12alpha`: \eqn{t_{1/2,\alpha}}; (1-, 2-, and 3-compartment models)
#'
#' * `t12beta`: \eqn{t_{1/2,\beta}}; (2- and 3-compartment models)
#'
#' * `t12gamma`: \eqn{t_{1/2,\gamma}}; (3-compartment model)
#'
#' * `alpha`: \eqn{\alpha}; (1-, 2-, and 3-compartment models)
#'
#' * `beta`: \eqn{\beta}; (2- and 3-compartment models)
#'
#' * `gamma`: \eqn{\beta}; (3-compartment model)
#'
#' * `A`: true `A`; (1-, 2-, and 3-compartment models)
#'
#' * `B`: true `B`; (2- and 3-compartment models)
#'
#' * `C`: true `C`; (3-compartment model)
#'
#' * `fracA`: fractional A; (1-, 2-, and 3-compartment models)
#'
#' * `fracB`: fractional B; (2- and 3-compartment models)
#'
#' * `fracC`: fractional C; (3-compartment model)
#'
#' @author Matthew Fidler and documentation from Justin Wilkins, \email{justin.wilkins@@occams.com}
#'
#' @references Shafer S. L. `CONVERT.XLS`
#'
#' @references Rowland M, Tozer TN. Clinical Pharmacokinetics and Pharmacodynamics: Concepts and Applications (4th). Clipping Williams & Wilkins, Philadelphia, 2010.
#'
#' @examples
#'
#' ## Note that rxode2 parses the names to figure out the best PK parameter
#'
#' params <- rxDerived(cl = 29.4, v = 23.4, Vp = 114, vp2 = 4614, q = 270, q2 = 73)
#'
#' ## That is why this gives the same results as the value before
#'
#' params <- rxDerived(CL = 29.4, V1 = 23.4, V2 = 114, V3 = 4614, Q2 = 270, Q3 = 73)
#'
#' ## You may also use micro-constants alpha/beta etc.
#'
#' params <- rxDerived(k12 = 0.1, k21 = 0.2, k13 = 0.3, k31 = 0.4, kel = 10, v = 10)
#'
#' ## or you can mix vectors and scalars
#'
#' params <- rxDerived(CL = 29.4, V = 1:3)
#'
#' ## If you want, you can round to a number of significant digits
#' ## with the `digits` argument:
#'
#' params <- rxDerived(CL = 29.4, V = 1:3, digits = 2)
#' @export
rxDerived <- function(..., verbose = FALSE, digits = 0) {
  .lst <- list(...)
  if (inherits(.lst[[1]], "data.frame")) {
    .lst <- .lst[[1]]
  }
  .namesU <- toupper(names(.lst))
  .w <- which(regexpr(.rxDerivedReg, .namesU) != -1)
  if (length(.w) > 1L) {
    if (verbose) {
      message("parameters: ", paste(names(.lst)[.w], collapse = ","))
    }
    .lst <- as.data.frame(.lst)
    .linCmt <- .Call(
      `_linCmtParse`, names(.lst)[.w],
      c(
        "with(.lst,.Call(`_rxode2_calcDerived`, ", "list(", "0, 0, 0, 0, ",
        ", 0, 0, 0, 0),digits))"
      ),
      verbose
    )$str
    .env <- environment()
    eval(parse(text = .linCmt), envir = .env)
  } else {
    stop("cannot figure out PK parameters to convert", call. = FALSE)
  }
}

#' Get the information about the rxode2 derived parameter transformation
#'
#'
#' @param ... Parameters translated, should be unquoted and not assigned to anything.
#' @return Translation information; This list contains:
#'
#' - `$str` A named string of the parameters as seen in the underlying C/C++
#'   code. The parameters that are NA are not used in the linear
#'   compartment model calculations.
#'
#' - `$ncmt` the number of compartments in the model
#'
#' - `$trans` the rxode2 translation number of the parameterization
#'
#' This contains the linCmt()
#'   translation number, the number of compartments and the parameters
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
#' @examples
#'
#' .rxTransInfo(cl, v , Vp, vp2, q, q2)
#'
#' .rxTransInfo(k12, k21, k13, k31, kel, v)
#'
#' .rxTransInfo(k12, k21, k13, k31, kel, v, ka)
#'
#' .rxTransInfo(CL, V)
#'
.rxTransInfo <- function(...) {
  .args <- as.list(match.call(expand.dots = TRUE))[-1]
  .args <- as.character(.args)
  .namesU <- toupper(as.character(.args))
  .w <- which(regexpr(.rxDerivedReg, .namesU) != -1)
  if (length(.w) > 1L) {
    .linCmt <- .Call(
      `_linCmtParse`, .args[.w],
      c(
        "", "", "tlag, F, rate1, dur1, ",
        ", tlag2, F2, rate2, dur2"
      ),
      FALSE
    )
    .str <- .linCmt$str
    .str <- strsplit(.str, ", +")[[1]]
    .str <- .str[-(1:2)]
    .str <- .str[c(1:6, 11)]
    .str <- vapply(seq_along(.str), function(i) {
      .num <- suppressWarnings(as.numeric(.str[i]))
      if (is.na(.num)) return(.str[i])
      NA_character_
    }, character(1), USE.NAMES=FALSE)
    names(.str) <- c("p1", "v1", "p2", "p3","p4", "p5", "ka")
    .linCmt$str <- .str
    .linCmt
  } else {
    stop("cannot figure out PK parameters to use", call. = FALSE)
  }
}

## nocov start
.dummy <- function() {
  #dummy import to make check() and CRAN happy
  .r <- rex::rex(start, end)
  .d <- data.table::data.table(a=1)
}
## nocov end


#' Get the number of linear compartments
#'
#' @param obj rxode2 object
#' @return A named integer vector with the number of linear compartments (numLin, numLinSens and depotLin)
#' @export
#' @keywords internal
#' @author Matthew L. Fidler
.rxLinNcmt <- function(obj) {
  .mv <- rxModelVars(obj)
  .flag <- setNames(.mv$flags["linCmtFlg"], NULL)
  if (.flag <= 0) {
    c(numLin=0L, numLinSens=0L, depotLin=0L)
  } else {
    .numLinSens <- floor(.flag/100)
    .numLin <- floor((.flag - .numLinSens*100)/10)
    .depotLin <- floor((.flag - .numLinSens*100- .numLin*10))
    c(numLin=.numLin, numLinSens=.numLinSens, depotLin=.depotLin)
  }
}

#' Get the linear compartment model states
#'
#' @param obj rxode2 type of object
#' @return compartment names
#' @export
#' @keywords internal
#' @author Matthew L. Fidler
.rxLinCmt <- function(obj) {
  .ncmt <- .rxLinNcmt(obj)
  .ret <- character(0)
  if (.ncmt["numLin"] <= 0) {
    return(.ret)
  }
  .vars <- c("p1", "v1","p2", "p3", "p4", "p5")
  .vars <- .vars[seq(1, 2*.ncmt["numLin"])]
  if (.ncmt["depotLin"] > 0) {
    .ret <- c(.ret, "depot")
    .vars <- c(.vars, "ka")
  }
  if (.ncmt["numLin"] > 0L) {
    .ret <- c(.ret, "central")
    if (.ncmt["numLin"] > 1L) {
        .ret <- c(.ret, paste0("peripheral", seq_len(.ncmt["numLin"] - 1L)))
    }
  }
  if (.ncmt["numLinSens"] > 0L) {
    .ret <- c(.ret,
              paste0("rx__sens_central_BY_", .vars))
    if (.ncmt["numLin"] > 1L) {
      .ret <- c(.ret,
                do.call(`c`,
                        lapply(paste0("peripheral", seq_len(.ncmt["numLin"] - 1L)),
                               function(x) {
                                 paste0("rx__sens_", x, "_BY_", .vars)
                               })))
    }
    if (.ncmt["depotLin"] > 0) {
      .ret <- c(.ret,
                "rx__sens_depot_BY_ka")
    }
  }
  .ret
}

#' Get the ODE states only
#'
#' @param obj rxode2 object
#' @return ODE states only
#' @export
#' @author Matthew L. Fidler
#' @examples
#' \donttest{
#' mod <- rxode2({
#'   Cp <- linCmt(Cl, V, Q2, V2, Q3, V3)
#'   ke0 <- log(2)/(50)
#'   d/dt(Ce) <- (Cp-Ce)*ke0
#' })
#'
#' rxStateOde(mod)
#'
#' rxState(mod)
#'
#' mod <- rxode2({
#'   Cp <- linCmt(Cl, V, Q2, V2, Q3, V3, ka)
#'   ke0 <- log(2)/(50)
#'   d/dt(Ce) <- (Cp-Ce)*ke0
#' }, linCmtSens="linCmtB")
#'
#' rxStateOde(mod)
#'
#' rxState(mod)
#' }
#'
rxStateOde <- function(obj) {
  .state <- rxState(obj)
  .mv <- rxModelVars(obj)
  if (length(.mv$indLin) > 0L) {
    .state <- setdiff(.state, "output")
  }
  setdiff(.state, .rxLinCmt(obj))
}

rxTest({
  test_that("drop support functions", {
    expect_equal(.getModelLineEquivalentLhsExpression(quote(-cl)), quote(cl))
    expect_equal(.getModelLineEquivalentLhsExpression(quote(-lag(matt))), quote(lag(matt)))
    expect_equal(.getModelLineEquivalentLhsExpression(quote(-alag(matt))), quote(alag(matt)))
    expect_equal(.getModelLineEquivalentLhsExpression(quote(-F(matt))), quote(F(matt)))
    expect_equal(.getModelLineEquivalentLhsExpression(quote(-f(matt))), quote(f(matt)))
    expect_equal(.getModelLineEquivalentLhsExpression(quote(-rate(matt))), quote(rate(matt)))
    expect_equal(.getModelLineEquivalentLhsExpression(quote(-dur(matt))), quote(dur(matt)))
    expect_equal(.getModelLineEquivalentLhsExpression(quote(-matt(0))), quote(matt(0)))
    expect_equal(.getModelLineEquivalentLhsExpression(quote(-d/dt(matt))), quote(d/dt(matt)))
    expect_equal(.getModelLineEquivalentLhsExpression(quote(-cp ~ .)), quote(cp))

    expect_true(.isDropExpression(quote(-v)))
    expect_false(.isDropExpression(quote(-v+3)))
    expect_false(.isDropExpression(quote(-3)))

    expect_false(.isDropExpression(quote(x <- y)))
    expect_false(.isDropExpression(quote(x + y ~ c(1, 0.01, 1))))

    expect_true(.isDropExpression(quote(-f(depot))))
    expect_true(.isDropExpression(quote(-F(depot))))
    expect_true(.isDropExpression(quote(-alag(depot))))
    expect_true(.isDropExpression(quote(-lag(depot))))
    expect_true(.isDropExpression(quote(-rate(depot))))
    expect_true(.isDropExpression(quote(-dur(depot))))
    expect_false(.isDropExpression(quote(-matt(depot))))
    expect_false(.isDropExpression(quote(-f(depot + central))))

    expect_true(.isDropExpression(quote(-depot(0))))
    expect_true(.isDropExpression(quote(-d/dt(depot))))

    expect_true(.isDropExpression(quote(-cp~.)))
  })

  test_that("drop from model before single endpoint model", {
    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        f(depot) <- 3
        cp <- center / v
        cp2 <- cp * cl
        cp ~ add(add.err)
      })
    }

    f2 <- one.compartment |> model(-cp2)

    expect_equal(f2$lstExpr[[8]], quote(cp ~ add(add.err)))
    expect_length(f2$lstExpr, 8L)
  })

  test_that("drop from model after single endpoint model", {
    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        f(depot) <- 3
        cp <- center / v
        cp ~ add(add.err)
        cp2 <- cp * cl
      })
    }

    f2 <- one.compartment |> model(-cp2)
    expect_equal(f2$lstExpr[[8]], quote(cp ~ add(add.err)))
    expect_length(f2$lstExpr, 8L)
  })

  test_that("drop endpoint from  multiple endpoint model", {
    pk.turnover.emax <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        ##
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        ##
        temax <- logit(0.8)
        #temax <- 7.5
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        ##
        eta.emax ~ .5
        eta.ec50  ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        ##
        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        ##
        #poplogit = log(temax/(1-temax))
        emax=expit(temax+eta.emax)
        #logit=temax+eta.emax
        ec50 =  exp(tec50 + eta.ec50)
        kout = exp(tkout + eta.kout)
        e0 = exp(te0 + eta.e0)
        ##
        DCP = center/v
        PD=1-emax*DCP/(ec50+DCP)
        ##
        effect(0) = e0
        kin = e0*kout
        ##
        d/dt(depot) = -ktr * depot
        d/dt(gut) =  ktr * depot -ka * gut
        d/dt(center) =  ka * gut - cl / v * center
        d/dt(effect) = kin*PD -kout*effect
        ##
        cp = center / v
        cp ~ prop(prop.err) + add(pkadd.err)
        effect ~ add(pdadd.err)
      })
    }

    suppressMessages(
      f2 <- pk.turnover.emax |> model(-cp)
    )

    expect_length(f2$predDf$cond, 1)
    expect_equal(f2$predDf$cond, "effect")
    expect_length(f2$lstExpr, 17)
  })

  test_that("drop compartment and compartment-related properties", {
    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) <- -ka * depot
        d/dt(center) <- ka * depot - cl / v * center
        f(depot) <- 3
        cp <- center / v
        cp ~ add(add.err)
        cp2 <- cp * cl
      })
    }

    suppressMessages(
      f2 <- one.compartment |> model(-d/dt(depot))
    )
    expect_equal(f2$mv0$state, "center")
    expect_length(f2$lstExpr, 7L)

    f2 <- one.compartment |> model(-f(depot))
    expect_equal(f2$mv0$state, c("depot", "center"))
    expect_length(f2$lstExpr, 8L)
  })

  test_that("drop endpoint test", {
    ocmt <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) <- -ka * depot
        d/dt(center) <- ka * depot - cl / v * center
        f(depot) <- 3
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    suppressMessages(
      f2 <- ocmt |> model(-cp ~ .)
    )
    expect_true(is.null(f2$predDf))
    expect_equal(f2$theta, c(tka = 0.45, tcl = 1, tv = 3.45))

    suppressMessages(
      f3 <- f2 |> model(cp ~ add(add.sd), append=TRUE)
    )

    expect_false(is.null(f3$predDf))
    expect_equal(f3$theta, c(tka = 0.45, tcl = 1, tv = 3.45, add.sd=1))
  })
})

rxTest({
  test_that("Multiple endpoint parsing", {

    pk.turnover.emax <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        ##
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        ##
        temax <- logit(0.8)
        #temax <- 7.5
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        ##
        eta.emax ~ .5
        eta.ec50  ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        ##
        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        ##
        #poplogit = log(temax/(1-temax))
        emax=expit(temax+eta.emax)
        #logit=temax+eta.emax
        ec50 =  exp(tec50 + eta.ec50)
        kout = exp(tkout + eta.kout)
        e0 = exp(te0 + eta.e0)
        ##
        DCP = center/v
        PD=1-emax*DCP/(ec50+DCP)
        ##
        effect(0) = e0
        kin = e0*kout
        ##
        d/dt(depot) = -ktr * depot
        d/dt(gut) =  ktr * depot -ka * gut
        d/dt(center) =  ka * gut - cl / v * center
        d/dt(effect) = kin*PD -kout*effect
        ##
        cp = center / v
        cp ~ prop(prop.err) + add(pkadd.err)
        effect ~ add(pdadd.err)
      })
    }
    f <- rxode2(pk.turnover.emax)

    expect_equal(f$paramsLine,
                 quote(params(tktr, tka, tcl, tv, prop.err, pkadd.err, temax, tec50,
                              tkout, te0, pdadd.err, eta.ktr, eta.ka, eta.cl, eta.v, eta.emax,
                              eta.ec50, eta.kout, eta.e0)))

    expect_equal(f$cmtLines,
                 list(quote(cmt(cp))))

    expect_equal(f$dvidLine, quote(dvid(5, 4)))

    pk.turnover.emax2 <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        ##
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        ##
        temax <- logit(0.8)
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        ##
        eta.emax ~ .5
        eta.ec50  ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        ##
        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        ##
        emax=expit(temax+eta.emax)
        ec50 =  exp(tec50 + eta.ec50)
        kout = exp(tkout + eta.kout)
        e0 = exp(te0 + eta.e0)
        ##
        DCP = center/v
        PD=1-emax*DCP/(ec50+DCP)
        ##
        effect(0) = e0
        kin = e0*kout
        ##
        d/dt(depot) = -ktr * depot
        d/dt(gut) =  ktr * depot -ka * gut
        d/dt(center) =  ka * gut - cl / v * center
        d/dt(effect) = kin*PD -kout*effect
        ##
        cp = center / v
        cp ~ prop(prop.err) + add(pkadd.err) | center
        effect ~ add(pdadd.err)
      })
    }

    ui2 <- rxode2(pk.turnover.emax2)

    expect_equal(ui2$paramsLine,
                 quote(params(tktr, tka, tcl, tv, prop.err, pkadd.err, temax, tec50,
                              tkout, te0, pdadd.err, eta.ktr, eta.ka, eta.cl, eta.v, eta.emax,
                              eta.ec50, eta.kout, eta.e0)))

    expect_equal(ui2$cmtLines,
                 list())

    expect_equal(ui2$dvidLine, quote(dvid(3, 4)))

    pk.turnover.emax3 <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        ##
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        ##
        temax <- logit(0.8)
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        ##
        eta.emax ~ .5
        eta.ec50  ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        ##
        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        emax = expit(temax+eta.emax)
        ec50 =  exp(tec50 + eta.ec50)
        kout = exp(tkout + eta.kout)
        e0 = exp(te0 + eta.e0)
        ##
        DCP = center/v
        PD=1-emax*DCP/(ec50+DCP)
        ##
        effect(0) = e0
        kin = e0*kout
        ##
        d/dt(depot) = -ktr * depot
        d/dt(gut) =  ktr * depot -ka * gut
        d/dt(center) =  ka * gut - cl / v * center
        d/dt(effect) = kin*PD -kout*effect
        ##
        cp = center / v
        cp ~ prop(prop.err) + add(pkadd.err)
        effect ~ add(pdadd.err) | pca
      })
    }

    ui3 <- rxode2(pk.turnover.emax3)

    expect_equal(ui3$paramsLine,
                 quote(params(tktr, tka, tcl, tv, prop.err, pkadd.err, temax, tec50,
                              tkout, te0, pdadd.err, eta.ktr, eta.ka, eta.cl, eta.v, eta.emax,
                              eta.ec50, eta.kout, eta.e0)))

    expect_equal(ui3$cmtLines,
                 list(quote(cmt(cp)),
                      quote(cmt(pca))))

    expect_equal(ui3$dvidLine, quote(dvid(5, 6)))

    pk.turnover.emax4 <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        ##
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        ##
        temax <- logit(0.8)
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        ##
        eta.emax ~ .5
        eta.ec50  ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        ##
        pdadd.err <- 10
        covWt <- 1
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl + wt * covWt)
        v <- exp(tv + eta.v)
        emax = expit(temax+eta.emax)
        ec50 =  exp(tec50 + eta.ec50)
        kout = exp(tkout + eta.kout)
        e0 = exp(te0 + eta.e0)
        ##
        DCP = center/v
        PD=1-emax*DCP/(ec50+DCP)
        ##
        effect(0) = e0
        kin = e0*kout
        ##
        d/dt(depot) = -ktr * depot
        d/dt(gut) =  ktr * depot -ka * gut
        d/dt(center) =  ka * gut - cl / v * center
        d/dt(effect) = kin*PD -kout*effect
        ##
        cp = center / v
        cp ~ prop(prop.err) + add(pkadd.err)
        effect ~ add(pdadd.err) | pca
      })
    }

    ui4 <- rxode2(pk.turnover.emax4)

    expect_equal(ui4$paramsLine,
                 quote(params(tktr, tka, tcl, tv, prop.err, pkadd.err, temax, tec50,
                              tkout, te0, pdadd.err, covWt, eta.ktr, eta.ka, eta.cl, eta.v,
                              eta.emax, eta.ec50, eta.kout, eta.e0, wt)))

    expect_equal(ui4$cmtLines,
                 list(quote(cmt(cp)),
                      quote(cmt(pca))))

    expect_equal(ui4$dvidLine, quote(dvid(5, 6)))

  })

  test_that("test constants are not considered a covariate ", {

    f <- function() {
      ini({
        t.fub <- 1
        cov.wt <- 0
        eta.fub ~ 0.1
        add.sd <- 0.01
      })
      model({
        BP = 0.61      # Blood:plasma partition coefficient
        fup = 0.028 + eta.fub    # Fraction unbound in plasma
        fub = fup/BP + t.fub + wt * cov.wt   # Fraction unbound in blood
        fub ~ add(add.sd)
      })
    }

    f <- rxode2(f)

    expect_equal(f$covariates, "wt")

  })
})

rxTest({

  test_that("logit ui test", {

    f <- function() {
      model({
        a <- logit(x, high=10)
      })
    }

    expect_equal((f() |> modelExtract()),
                 "a <- logit(x, 0, 10)")

    f <- function() {
      model({
        a <- expit(x, high=10+1)
      })
    }

    expect_equal((f() |> modelExtract()),
                 "a <- expit(x, 0, 11)")

    f <- function() {
      model({
        a <- expit(high=10+1, low=a, x)
      })
    }

    expect_equal((f() |> modelExtract()),
                 "a <- expit(x, a, 11)")


  })

  test_that("logit tests", {
    expect_equal(
      rxToSE("logit(a)"),
      "-log(1/(a)-1)"
    )
    expect_equal(
      rxToSE("logit(a,b)"),
      "-log(1/(((a)-(b))/(1.0-(b)))-1)"
    )
    expect_equal(
      rxToSE("logit(a,b,c)"),
      "-log(1/(((a)-(b))/((c)-(b)))-1)"
    )
    expect_error(rxToSE("logit()"))
    expect_error(rxToSE("logit(a,b,c,d)"))

    mod <- rxode2({
      b <- logit(a)
    })

    expect_equal(
      logit(seq(0.01, 0.99, length.out = 10)),
      rxSolve(mod, et(0), params = data.frame(a = seq(0.01, 0.99, length.out = 10)))$b
    )

    mod <- rxode2({
      b <- logit(a, -0.5)
    })

    expect_equal(
      logit(seq(0.01, 0.99, length.out = 10), -0.5),
      rxSolve(mod, et(0), params = data.frame(a = seq(0.01, 0.99, length.out = 10)))$b
    )

    mod <- rxode2({
      b <- logit(a, -0.5, 1.5)
    })

    expect_equal(
      logit(seq(0.01, 0.99, length.out = 10), -0.5, 1.5),
      rxSolve(mod, et(0), params = data.frame(a = seq(0.01, 0.99, length.out = 10)))$b
    )

    expect_error({
      b <- logit()
    })
    expect_error({
      b <- logit(a, b, c, d)
    })

    expect_equal(logit(1:10, 0L, 11L), logit(as.double(1:10), 0.0, 11.0))

    expect_equal(logit(7, c(1, 2), c(10, 114)),
                 c(logit(7, 1, 10),
                   logit(7, 2, 114)))

    expect_error(.Call(`_rxode2_powerD`, 0.5, c(1, 2), 3, 4, 4L, TRUE))
    expect_error(.Call(`_rxode2_powerD`, 0.5, 1, c(3, 4), 4, 4L, TRUE))
    expect_error(.Call(`_rxode2_powerD`, 0.5, 1, 3, 4, c(4L, 5L), TRUE))
    expect_error(.Call(`_rxode2_powerD`, 0.5, 1, 3, 4, 4L, c(TRUE, FALSE)))

    expect_error(logit(0.5, 1, -2))
  })

  test_that("expit", {
    expect_equal(
      rxToSE("expit(a)"),
      "1/(1+exp(-(a)))"
    )
    expect_equal(
      rxToSE("expit(a,b)"),
      "(1.0-(b))*(1/(1+exp(-(a))))+(b)"
    )
    expect_equal(
      rxToSE("expit(a,b,c)"),
      "((c)-(b))*(1/(1+exp(-(a))))+(b)"
    )
    expect_error(rxToSE("expit()"))
    expect_error(rxToSE("expit(a,b,c,d)"))

    mod <- rxode2({
      b <- expit(a)
    })

    expect_equal(
      expit(seq(-4, 4, length.out = 10)),
      rxSolve(mod, et(0), params = data.frame(a = seq(-4, 4, length.out = 10)))$b
    )

    mod <- rxode2({
      b <- expit(a, 0.5)
    })

    expect_equal(
      expit(seq(-4, 4, length.out = 10), 0.5),
      rxSolve(mod, et(0), params = data.frame(a = seq(-4, 4, length.out = 10)))$b
    )

    mod <- rxode2({
      b <- expit(a, 0.5, 1.5)
    })

    suppressMessages(expect_error(rxode2({
      b <- expit()
    })))
    suppressMessages(expect_error(rxode2({
      b <- expit(a, b, c, d)
    })))

    expect_equal(
      expit(seq(-4, 4, length.out = 10), 0.5, 1.5),
      rxSolve(mod, et(0), params = data.frame(a = seq(-4, 4, length.out = 10)))$b
    )

    expect_equal(expit(1:10, 0L, 11L), expit(as.double(1:10), 0.0, 11.0))
  })
})

if (!.Call(`_rxode2_isIntel`)) {
  test_that(".matchesLangTemplate", {
    expect_false(.matchesLangTemplate(str2lang("d/dt(foo)"), str2lang("d/foo(.name)")))
    # . matches anything
    expect_true(.matchesLangTemplate(str2lang("'foo'"), str2lang(".")))
    # .name matches any name
    expect_true(.matchesLangTemplate(str2lang("d/dt(foo)"), str2lang("d/dt(.name)")))
    # .call() matches any call but it must be given as a call (and not a name)
    expect_true(.matchesLangTemplate(str2lang("d/(foo)"), str2lang("d/.call()")))
    expect_false(.matchesLangTemplate(str2lang("d/(foo)"), str2lang("d/.call")))
  })

  test_that(".rxTransform() handles NA", {
    expect_equal(.rxTransform(NA), NA_real_)
  })
}

rxTest({
  test_that("single or multiple endpoint model", {

    # Test for rxode2 issue #17
    f <- function() {
      ini({
        tke <- 0.5
        eta.ke ~ 0.04
        prop.sd <- sqrt(0.1)
      })
      model({
        ke <- tke * exp(eta.ke)
        ipre <- 10 * exp(-ke * t)
        f2 <- ipre / (ipre + 5)
        f3 <- f2 * 3
        lipre <- log(ipre)
        ipre ~ prop(prop.sd)
      })
    }

    f <- rxode2(f)

    expect_equal("ipre", f$predDf$var)
    suppressMessages(
      expect_warning(
        .tmp <- f |> model(lipre ~ add(log.add.sd)),
        "with single endpoint model prediction 'ipre' is changed to 'lipre'"
      )
    )
    expect_equal("lipre", .tmp$predDf$var)

    expect_error(f |> model(PD ~ add(log.add.sd)))

    fo <- function() {
      ini({
        tke <- 0.5
        eta.ke ~ 0.04
        prop.sd <- sqrt(0.1)
      })
      model({
        ke <- tke * exp(eta.ke)
        d/dt(ipre) <- -ke * ipre
        f2 <- ipre / (ipre + 5)
        f3 <- f2 * 3
        lipre <- log(ipre)
        ipre ~ prop(prop.sd)
      })
    }

    fo <- rxode2(fo)

    expect_equal("ipre", fo$predDf$var)
    suppressMessages(
      expect_warning(
        .tmp <- fo |> model(lipre ~ add(log.add.sd)),
        "with single endpoint model prediction 'ipre' is changed to 'lipre'"
      )
    )
    expect_equal("lipre", .tmp$predDf$var)

    expect_error(fo |> model(PD ~ add(log.add.sd)))

    pk.turnover.emax2 <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        ##
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        ##
        temax <- logit(0.8)
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        ##
        eta.emax ~ .5
        eta.ec50  ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        ##
        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        ##
        emax=expit(temax+eta.emax)
        ec50 =  exp(tec50 + eta.ec50)
        kout = exp(tkout + eta.kout)
        e0 = exp(te0 + eta.e0)
        ##
        DCP = center/v
        PD=1-emax*DCP/(ec50+DCP)
        ##
        effect(0) = e0
        kin = e0*kout
        ##
        d/dt(depot) = -ktr * depot
        d/dt(gut) =  ktr * depot -ka * gut
        d/dt(center) =  ka * gut - cl / v * center
        d/dt(effect) = kin*PD -kout*effect
        ##
        cp = center / v
        cp ~ prop(prop.err) + add(pkadd.err) | center
        effect ~ add(pdadd.err)
      })
    }

    multiple <- rxode2(pk.turnover.emax2)

    expect_error(multiple |> model(PD ~ add(add.sd)))
    suppressMessages(
      expect_error(multiple |> model(effect ~ add(add.sd)), NA)
    )
  })
})

rxTest({
  test_that("incomplete gamma", {
    gp <- rxode2({
      x <- gammap(a, z)
    })
    
    ## Numbers taken from pracma
    expect_equal(lowergamma(2, 1.5), 0.4421746)
    expect_equal(uppergamma(2, 1.5), 0.5578254)
    expect_equal(gammap(2, 1.5), 0.4421746)
    expect_equal(lowergamma(3, 1.5), 0.3823063, tolerance = 1e-6)
    expect_equal(uppergamma(3, 1.5), 1.6176937, tolerance = 1e-6)
    expect_equal(uppergamma(3, 1.5), 1.6176937, tolerance = 1e-6)
    ## note -1.5 doesn't work with boost, but works with pracma...
    expect_equal(gammaq(2, 1.5), 0.5578254, tolerance = 1e-6)
    
    expect_equal(
      rxFromSE("Derivative(gammap(x,y),y)"),
      "gammapDer(x,y)"
    )
  })
})

rxTest({
  test_that("plot tests", {

    skip_if_not_installed("units")
    skip_if_not_installed("ggplot2", minimum_version="3.3.5")
    skip_if_not_installed("vdiffr")
    rxUnloadAll()

    # vdiffr doesn't cause errors on CRAN, so skip there.
    skip_on_cran()

    ## Model from rxode2 tutorial
    m1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      ## Added modeled bioavaiblity, duration and rate
      fdepot <- 1
      durDepot <- 8
      rateDepot <- 1250
      C2 <- centr / V2
      C3 <- peri / V3
      d / dt(depot) <- -KA * depot
      f(depot) <- fdepot
      dur(depot) <- durDepot
      rate(depot) <- rateDepot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- 1
    })

    ev <- et(timeUnits = "hr") |>
      et(amt = 10000, ii = 12, until = 24) |>
      et(seq(0, 24, length.out = 100))

    evR <- et(timeUnits = "hr") |> et(evid = 3)

    evR2 <- et(timeUnits = "hr") |>
      et(amt = 5000, ii = 12, until = 24) |>
      et(seq(0, 24, length.out = 100))

    evR <- dplyr::bind_rows(as.data.frame(ev), as.data.frame(evR), as.data.frame(evR2))

    evR4 <- rbind(data.frame(id = 1, evR), data.frame(id = 2, evR), data.frame(id = 3, evR), data.frame(id = 4, evR))

    s <- rxSolve(m1, ev)

    sR <- rxSolve(m1, evR)

    ev <- et(timeUnits = "hr") |>
      et(amt = 10000, until = units::set_units(3, days), ii = 12) |>
      # loading doses
      et(seq(0, 48, length.out = 200)) |>
      et(id = 1:4)

    rxWithSeed(
      42,
      s2 <- rxSolve(m1, ev, params = data.frame(KA = 0.294 * exp(rnorm(4)), 18.6 * exp(rnorm(4))))
    )

    rxWithSeed(
      42,
      s2R <- rxSolve(m1, evR4, params = data.frame(KA = 0.294 * exp(rnorm(4)), 18.6 * exp(rnorm(4))))
    )

    rxWithSeed(
      42,
      s20 <- rxSolve(m1, ev, params = data.frame(KA = 0.294 * exp(rnorm(20)), 18.6 * exp(rnorm(20))))
    )

    rxWithSeed(
      42,
      s20R <- rxSolve(m1, evR4, params = data.frame(KA = 0.294 * exp(rnorm(20)), 18.6 * exp(rnorm(20))))
    )

    m2 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01 * exp(eta.Cl)
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      ## Added modeled bioavaiblity, duration and rate
      fdepot <- 1
      durDepot <- 8
      rateDepot <- 1250
      C2 <- centr / V2
      C3 <- peri / V3
      d / dt(depot) <- -KA * depot
      f(depot) <- fdepot
      dur(depot) <- durDepot
      rate(depot) <- rateDepot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- 1
    })

    omega <- lotri(eta.Cl ~ 0.4^2)

    ev <- et(amount.units = "mg", time.units = "hours") |>
      et(amt = 10000, cmt = "centr", until = 48, ii = 8) |>
      et(0, 48, length.out = 100)

    evR <- et(amount.units = "mg", time.units = "hours") |> et(evid = 3)

    evR2 <- et(amount.units = "mg", time.units = "hours") |>
      et(amt = 5000, cmt = "centr", until = 48, ii = 8) |>
      et(0, 48, length.out = 100)

    evR <- dplyr::bind_rows(as.data.frame(ev), as.data.frame(evR), as.data.frame(evR2))

    ev$extra <- c(rep(1, 50), rep(0, 51))

    sim <- rxSolve(m2, ev, omega = omega, nSub = 100, keep="extra")

    simR <- rxSolve(m2, evR, omega = omega, nSub = 100)

    expect_error(plot(sim, log = "f"))

    ev2 <- et() |>
      et(amt = 10000, cmt = "centr", until = 48, ii = 8) |>
      et(0, 48, length.out = 100)

    ev2R <- et(evid = 3)

    ev2R2 <- et() |>
      et(amt = 5000, cmt = "centr", until = 48, ii = 8) |>
      et(0, 48, length.out = 100)

    ev2R <- dplyr::bind_rows(as.data.frame(ev2), as.data.frame(ev2R), as.data.frame(ev2R2))

    sim3 <- rxSolve(m2, ev2, omega = omega, nSub = 3)
    sim3R <- rxSolve(m2, ev2R, omega = omega, nSub = 3)

    .rxWithOptions(list(rxode2.theme = TRUE), {
      #vdiffr::expect_doppelganger("sim.id-unitless", plot(sim3, C2))
      expect_error(plot(sim3, C2), NA)
    })

    .rxWithOptions(list(rxode2.theme = FALSE), {
      #vdiffr::expect_doppelganger("sim.id-unitless-notheme", plot(sim3, C2))
      expect_error(plot(sim3, C2), NA)
    })

    .rxWithOptions(list(rxode2.theme = TRUE), {

      ci1.C2 <- confint(sim, "C2", ci=0.99)

      ci1.C3 <- confint(sim,  ci=0.99)

      ci1.C2.e <- confint(sim, "C2", by="extra")

      ci1.C2.eff <- confint(sim, c("C2", "eff"))

      ci1.C2.eff.e <- confint(sim, c("C2", "eff"), by="extra")

      sim2 <- rxSolve(m2, ev, omega = omega, nSub = 2500, keep="extra")

      sim2R <- rxSolve(m2, evR, omega = omega, nSub = 2500)

      ci2.C2 <- confint(sim2, "C2")

      ci2.C2.e <- confint(sim2, "C2", by="extra")

      ci2.C2.eff <- confint(sim2, c("C2", "eff"))

      ci2.C2.eff.e <- confint(sim2, c("C2", "eff"), by="extra")

      f <- function(xgxr = FALSE, repel = FALSE) {
        if (xgxr) {
          .xgxtxt <- "xgxr-"
          .xgxOp <- list(rxode2.xgxr = TRUE)
        } else {
          .xgxtxt <- ""
          .xgxOp <- list(rxode2.xgxr = FALSE)
        }
        .rxWithOptions(.xgxOp, {
          #vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "C2"), suppressWarnings(s |> plot(C2)))
          expect_error(suppressWarnings(s |> plot(C2)), NA)

          #vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "C2-log-x"), suppressWarnings(s |> plot(C2, log = "x")))
          expect_error(suppressWarnings(s |> plot(C2, log = "x")), NA)

          #vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "C2-log-y"), suppressWarnings(s |> plot(C2, log = "y")))
          expect_error(suppressWarnings(s |> plot(C2, log = "y")), NA)

          #vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "C2-log-xy"), suppressWarnings(s |> plot(C2, log = "xy")))
          expect_error(suppressWarnings(s |> plot(C2, log = "xy")), NA)

          #vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "C2-log-yx"), suppressWarnings(s |> plot(C2, log = "yx")))

          expect_error(suppressWarnings(s |> plot(C2, log = "yx")), NA)

          #vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "all"), suppressWarnings(s |> plot()))
          expect_error(suppressWarnings(s |> plot()), NA)

          #vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "all-log-x"), suppressWarnings(s |> plot(log = "x")))
          expect_error(suppressWarnings(s |> plot(log = "x")), NA)

          #vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "all-log-y"), suppressWarnings(s |> plot(log = "y")))
          expect_error(suppressWarnings(s |> plot(log = "y")), NA)

          #vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "all-log-xy"), suppressWarnings(s |> plot(log = "xy")))
          expect_error(suppressWarnings(s |> plot(log = "xy")), NA)

          #vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "all-log-yx"), suppressWarnings(s |> plot(log = "yx")))
          expect_error(suppressWarnings(s |> plot(log = "yx")), NA)


          #vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "C2-r"), suppressWarnings(sR |> plot(C2)))
          expect_error(suppressWarnings(sR |> plot(C2)), NA)

          # vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "C2-log-x-r"), suppressWarnings(sR |> plot(C2, log = "x")))
          expect_error(suppressWarnings(sR |> plot(C2, log = "x")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "C2-log-y-r"), suppressWarnings(sR |> plot(C2, log = "y")))
          expect_error(suppressWarnings(sR |> plot(C2, log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "C2-log-xy-r"), suppressWarnings(sR |> plot(C2, log = "xy")))
          expect_error(suppressWarnings(sR |> plot(C2, log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "C2-log-yx-r"), suppressWarnings(sR |> plot(C2, log = "yx")))
          expect_error(suppressWarnings(sR |> plot(C2, log = "yx")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "all-r"), suppressWarnings(sR |> plot()))
          expect_error(suppressWarnings(sR |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "all-log-x-r"), suppressWarnings(sR |> plot(log = "x")))
          expect_error(suppressWarnings(sR |> plot(log = "x")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "all-log-y-r"), suppressWarnings(sR |> plot(log = "y")))
          expect_error(suppressWarnings(sR |> plot(log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "all-log-xy-r"), suppressWarnings(sR |> plot(log = "xy")))
          expect_error(suppressWarnings(sR |> plot(log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-", .xgxtxt, "all-log-yx-r"), suppressWarnings(sR |> plot(log = "yx")))
          expect_error(suppressWarnings(sR |> plot(log = "yx")), NA)


          # vdiffr::expect_doppelganger(paste0("plot-ci1c2", .xgxtxt), suppressWarnings(ci1.C2 |> plot()))
          expect_error(suppressWarnings(ci1.C2 |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c3", .xgxtxt, "-centr"), suppressWarnings(ci1.C3 |> plot("centr")))
          expect_error(suppressWarnings(ci1.C3 |> plot("centr")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c3", .xgxtxt, "-full"), suppressWarnings(ci1.C3 |> plot()))
          expect_error(suppressWarnings(ci1.C3 |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2", .xgxtxt, "log-x"), suppressWarnings(ci1.C2 |> plot(log = "x")))
          expect_error(suppressWarnings(ci1.C2 |> plot(log = "x")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2", .xgxtxt, "log-y"), suppressWarnings(ci1.C2 |> plot(log = "y")))
          expect_error(suppressWarnings(ci1.C2 |> plot(log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2", .xgxtxt, "log-xy"), suppressWarnings(ci1.C2 |> plot(log = "xy")))
          expect_error(suppressWarnings(ci1.C2 |> plot(log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2", .xgxtxt, "log-yx"), suppressWarnings(ci1.C2 |> plot(log = "yx")))
          expect_error(suppressWarnings(ci1.C2 |> plot(log = "yx")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-e", .xgxtxt), suppressWarnings(ci1.C2.e |> plot()))
          expect_error(suppressWarnings(ci1.C2.e |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-e", .xgxtxt, "log-x"), suppressWarnings(ci1.C2.e |> plot(log = "x")))
          expect_error(suppressWarnings(ci1.C2.e |> plot(log = "x")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-e", .xgxtxt, "log-y"), suppressWarnings(ci1.C2.e |> plot(log = "y")))
          expect_error(suppressWarnings(ci1.C2.e |> plot(log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-e", .xgxtxt, "log-xy"), suppressWarnings(ci1.C2.e |> plot(log = "xy")))
          expect_error(suppressWarnings(ci1.C2.e |> plot(log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-e", .xgxtxt, "log-yx"), suppressWarnings(ci1.C2.e |> plot(log = "yx")))
          expect_error(suppressWarnings(ci1.C2.e |> plot(log = "yx")), NA)


          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-eff", .xgxtxt), suppressWarnings(ci1.C2.eff |> plot()))
          expect_error(suppressWarnings(ci1.C2.eff |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-eff", .xgxtxt, "log-x"), suppressWarnings(ci1.C2.eff |> plot(log = "x")))
          expect_error(suppressWarnings(ci1.C2.eff |> plot(log = "x")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-eff", .xgxtxt, "log-y"), suppressWarnings(ci1.C2.eff |> plot(log = "y")))

          expect_error(suppressWarnings(ci1.C2.eff |> plot(log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-eff", .xgxtxt, "log-xy"), suppressWarnings(ci1.C2.eff |> plot(log = "xy")))

          expect_error(suppressWarnings(ci1.C2.eff |> plot(log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-eff", .xgxtxt, "log-yx"), suppressWarnings(ci1.C2.eff |> plot(log = "yx")))
          expect_error(suppressWarnings(ci1.C2.eff |> plot(log = "yx")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-eff-e", .xgxtxt), suppressWarnings(ci1.C2.eff.e |> plot()))
          expect_error(suppressWarnings(ci1.C2.eff.e |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-eff-e", .xgxtxt, "log-x"), suppressWarnings(ci1.C2.eff.e |> plot(log = "x")))
          expect_error(suppressWarnings(ci1.C2.eff.e |> plot(log = "x")), NA)

          #vdiffr::expect_doppelganger(paste0("plot-ci1c2-eff-e", .xgxtxt, "log-y"), suppressWarnings(ci1.C2.eff.e |> plot(log = "y")))
          expect_error(suppressWarnings(ci1.C2.eff.e |> plot(log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-eff-e", .xgxtxt, "log-xy"), suppressWarnings(ci1.C2.eff.e |> plot(log = "xy")))
          expect_error(suppressWarnings(ci1.C2.eff.e |> plot(log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci1c2-eff-e", .xgxtxt, "log-yx"), suppressWarnings(ci1.C2.eff.e |> plot(log = "yx")))
          expect_error(suppressWarnings(ci1.C2.eff.e |> plot(log = "yx")), NA)


          # vdiffr::expect_doppelganger(paste0("plot-ci2c2", .xgxtxt), suppressWarnings(ci2.C2 |> plot()))
          expect_error(suppressWarnings(ci2.C2 |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2", .xgxtxt, "log-x"), suppressWarnings(ci2.C2 |> plot(log = "x")))
          expect_error(suppressWarnings(ci2.C2 |> plot(log = "x")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2", .xgxtxt, "log-y"), suppressWarnings(ci2.C2 |> plot(log = "y")))
          expect_error(suppressWarnings(ci2.C2 |> plot(log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2", .xgxtxt, "log-xy"), suppressWarnings(ci2.C2 |> plot(log = "xy")))
          expect_error(suppressWarnings(ci2.C2 |> plot(log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2", .xgxtxt, "log-yx"), suppressWarnings(ci2.C2 |> plot(log = "yx")))
          expect_error(suppressWarnings(ci2.C2 |> plot(log = "yx")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-e", .xgxtxt), suppressWarnings(ci2.C2.e |> plot()))
          expect_error(suppressWarnings(ci2.C2.e |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-e", .xgxtxt, "log-x"), suppressWarnings(ci2.C2.e |> plot(log = "x")))
          expect_error(suppressWarnings(ci2.C2.e |> plot(log = "x")), NA)


          ## vdiffr::expect_doppelganger(paste0("plot-ci2c2-e", .xgxtxt, "log-y"), suppressWarnings(ci2.C2.e |> plot(log = "y")))
          expect_error(suppressWarnings(ci2.C2.e |> plot(log = "y")), NA)


          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-e", .xgxtxt, "log-xy"), suppressWarnings(ci2.C2.e |> plot(log = "xy")))
          expect_error(suppressWarnings(ci2.C2.e |> plot(log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-e", .xgxtxt, "log-yx"), suppressWarnings(ci2.C2.e |> plot(log = "yx")))
          expect_error(suppressWarnings(ci2.C2.e |> plot(log = "yx")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-eff", .xgxtxt), suppressWarnings(ci2.C2.eff |> plot()))
          expect_error(suppressWarnings(ci2.C2.eff |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-eff", .xgxtxt, "log-x"), suppressWarnings(ci2.C2.eff |> plot(log = "x")))
          expect_error(suppressWarnings(ci2.C2.eff |> plot(log = "x")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-eff", .xgxtxt, "log-y"), suppressWarnings(ci2.C2.eff |> plot(log = "y")))
          expect_error(suppressWarnings(ci2.C2.eff |> plot(log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-eff", .xgxtxt, "log-xy"), suppressWarnings(ci2.C2.eff |> plot(log = "xy")))
          expect_error(suppressWarnings(ci2.C2.eff |> plot(log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-eff", .xgxtxt, "log-yx"), suppressWarnings(ci2.C2.eff |> plot(log = "yx")))
          expect_error(suppressWarnings(ci2.C2.eff |> plot(log = "yx")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-eff-e", .xgxtxt), suppressWarnings(ci2.C2.eff.e |> plot()))
          expect_error(suppressWarnings(ci2.C2.eff.e |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-eff-e", .xgxtxt, "log-x"), suppressWarnings(ci2.C2.eff.e |> plot(log = "x")))
          expect_error(suppressWarnings(ci2.C2.eff.e |> plot(log = "x")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-eff-e", .xgxtxt, "log-y"), suppressWarnings(ci2.C2.eff.e |> plot(log = "y")))
          expect_error(suppressWarnings(ci2.C2.eff.e |> plot(log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-ci2c2-eff-e", .xgxtxt, "log-xy"), suppressWarnings(ci2.C2.eff.e |> plot(log = "xy")))
          expect_error(suppressWarnings(ci2.C2.eff.e |> plot(log = "xy")), NA)


          ## vdiffr::expect_doppelganger(paste0("plot-ci2c2-eff-e", .xgxtxt, "log-yx"), suppressWarnings(ci2.C2.eff.e |> plot(log = "yx")))
          expect_error(suppressWarnings(ci2.C2.eff.e |> plot(log = "yx")), NA)

          ##

          ## large
          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "C2"), suppressWarnings(s20 |> plot(C2)))
          expect_error(suppressWarnings(s20 |> plot(C2)), NA)


          #vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "C2-log-x"), suppressWarnings(s20 |> plot(C2, log = "x")))
          expect_error(suppressWarnings(s20 |> plot(C2, log = "x")), NA)

          #vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "C2-log-y"), suppressWarnings(s20 |> plot(C2, log = "y")))
          expect_error(suppressWarnings(s20 |> plot(C2, log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "C2-log-xy"), suppressWarnings(s20 |> plot(C2, log = "xy")))
          expect_error(suppressWarnings(s20 |> plot(C2, log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "C2-log-yx"), suppressWarnings(s20 |> plot(C2, log = "yx")))
          expect_error(suppressWarnings(s20 |> plot(C2, log = "yx")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "all"), suppressWarnings(s20 |> plot()))
          expect_error(suppressWarnings(s20 |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "all-log-x"), suppressWarnings(s20 |> plot(log = "x")))
          expect_error(suppressWarnings(s20 |> plot(log = "x")), NA)

          ## vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "all-log-y"), suppressWarnings(s20 |> plot(log = "y")))
          expect_error(suppressWarnings(s20 |> plot(log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "all-log-xy"), suppressWarnings(s20 |> plot(log = "xy")))
          expect_error(suppressWarnings(s20 |> plot(log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "all-log-yx"), suppressWarnings(s20 |> plot(log = "yx")))
          expect_error(suppressWarnings(s20 |> plot(log = "yx")), NA)


          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "C2-r"), suppressWarnings(s20R |> plot(C2)))
          expect_error(suppressWarnings(s20R |> plot(C2)), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "C2-log-x-r"), suppressWarnings(s20R |> plot(C2, log = "x")))
          expect_error(suppressWarnings(s20R |> plot(C2, log = "x")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "C2-log-y-r"), suppressWarnings(s20R |> plot(C2, log = "y")))
          expect_error(suppressWarnings(s20R |> plot(C2, log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "C2-log-xy-r"), suppressWarnings(s20R |> plot(C2, log = "xy")))
          expect_error(suppressWarnings(s20R |> plot(C2, log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "C2-log-yx-r"), suppressWarnings(s20R |> plot(C2, log = "yx")))
          expect_error(suppressWarnings(s20R |> plot(C2, log = "yx")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "all-r"), suppressWarnings(s20R |> plot()))
          expect_error(suppressWarnings(s20R |> plot()), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "all-log-x-r"), suppressWarnings(s20R |> plot(log = "x")))
          expect_error(suppressWarnings(s20R |> plot(log = "x")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "all-log-y-r"), suppressWarnings(s20R |> plot(log = "y")))
          expect_error(suppressWarnings(s20R |> plot(log = "y")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "all-log-xy-r"), suppressWarnings(s20R |> plot(log = "xy")))
          expect_error(suppressWarnings(s20R |> plot(log = "xy")), NA)

          # vdiffr::expect_doppelganger(paste0("plot-sp-", .xgxtxt, "all-log-yx-r"), suppressWarnings(s20R |> plot(log = "yx")))
          expect_error(suppressWarnings(s20R |> plot(log = "yx")), NA)

          for (repel in c(TRUE, FALSE)) {
            if (repel) {
              .repel <- "repel-"
              .repelOp <- list(rxode2.ggrepel = TRUE)
            } else {
              .repel <- ""
              .repelOp <- list(rxode2.ggrepel = FALSE)
            }
            .rxWithOptions(.repelOp, {

              #vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "C2"), suppressWarnings(s2 |> plot(C2)))
              expect_error(suppressWarnings(s2 |> plot(C2)), NA)

              #vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "C2-log-x"), suppressWarnings(s2 |> plot(C2, log = "x")))
              expect_error(suppressWarnings(s2 |> plot(C2, log = "x")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "C2-log-y"), suppressWarnings(s2 |> plot(C2, log = "y")))
              expect_error(suppressWarnings(s2 |> plot(C2, log = "y")), NA)


              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "C2-log-xy"), suppressWarnings(s2 |> plot(C2, log = "xy")))
              expect_error(suppressWarnings(s2 |> plot(C2, log = "xy")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "C2-log-yx"), suppressWarnings(s2 |> plot(C2, log = "yx")))
              expect_error(suppressWarnings(s2 |> plot(C2, log = "yx")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "all"), suppressWarnings(s2 |> plot()))
              expect_error(suppressWarnings(s2 |> plot()), NA)

              ## vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "all-log-x"), suppressWarnings(s2 |> plot(log = "x")))
              expect_error(suppressWarnings(s2 |> plot(log = "x")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "all-log-y"), suppressWarnings(s2 |> plot(log = "y")))
              expect_error(suppressWarnings(s2 |> plot(log = "y")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "all-log-xy"), suppressWarnings(s2 |> plot(log = "xy")))
              expect_error(suppressWarnings(s2 |> plot(log = "xy")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "all-log-yx"), suppressWarnings(s2 |> plot(log = "yx")))
              expect_error(suppressWarnings(s2 |> plot(log = "yx")), NA)

              ## Issue #284
              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "284-log-yx"), suppressWarnings(s2 |> plot(C2, eff, log = "yx")))
              expect_error(suppressWarnings(s2 |> plot(C2, eff, log = "yx")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "284-log-y"), suppressWarnings(s2 |> plot(C2, eff, log = "y")))
              expect_error(suppressWarnings(s2 |> plot(C2, eff, log = "y")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "284-log-x"), suppressWarnings(s2 |> plot(C2, eff, log = "x")))
              expect_error(suppressWarnings(s2 |> plot(C2, eff, log = "x")), NA)

              # reset
              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "C2-r"), suppressWarnings(s2R |> plot(C2)))
              expect_error(suppressWarnings(s2R |> plot(C2)), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "C2-log-x-r"), suppressWarnings(s2R |> plot(C2, log = "x")))

              expect_error(suppressWarnings(s2R |> plot(C2, log = "x")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "C2-log-y-r"), suppressWarnings(s2R |> plot(C2, log = "y")))
              expect_error(suppressWarnings(s2R |> plot(C2, log = "y")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "C2-log-xy-r"), suppressWarnings(s2R |> plot(C2, log = "xy")))
              expect_error(suppressWarnings(s2R |> plot(C2, log = "xy")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "C2-log-yx-r"), suppressWarnings(s2R |> plot(C2, log = "yx")))
              expect_error(suppressWarnings(s2R |> plot(C2, log = "yx")), NA)


              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "all-r"), suppressWarnings(s2R |> plot()))
              expect_error(suppressWarnings(s2R |> plot()), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "all-log-x-r"), suppressWarnings(s2R |> plot(log = "x")))
              expect_error(suppressWarnings(s2R |> plot(log = "x")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "all-log-y-r"), suppressWarnings(s2R |> plot(log = "y")))
              expect_error(suppressWarnings(s2R |> plot(log = "y")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "all-log-xy-r"), suppressWarnings(s2R |> plot(log = "xy")))
              expect_error(suppressWarnings(s2R |> plot(log = "xy")), NA)

              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "all-log-yx-r"), suppressWarnings(s2R |> plot(log = "yx")))
              expect_error(suppressWarnings(s2R |> plot(log = "yx")), NA)

              ## Issue #284
              # vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "284-log-yx-r"), suppressWarnings(s2R |> plot(C2, eff, log = "yx")))
              expect_error(suppressWarnings(s2R |> plot(C2, eff, log = "yx")), NA)

              #vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "284-log-y-r"), suppressWarnings(s2R |> plot(C2, eff, log = "y")))

              expect_error(suppressWarnings(s2R |> plot(C2, eff, log = "y")), NA)

              #vdiffr::expect_doppelganger(paste0("plot-multi-", .repel, .xgxtxt, "284-log-x-r"), suppressWarnings(s2R |> plot(C2, eff, log = "x")))
              expect_error(suppressWarnings(s2R |> plot(C2, eff, log = "x")), NA)

            })
          }
        })
      }
      f(TRUE)
      f(FALSE)
    })
  })
})

rxTest({
  test_that("ordinal simulation", {
    # large simulation model #1
    ord <- function() {
      ini({
        b1 <- 7.86
        slope <- log(0.045)
        b2 <- c(-Inf, -1.73, 0)
        b3 <- c(-Inf, -1.95, 0)
        b4 <- c(-Inf, -1.55, 0)
        b5 <- c(-Inf, -1.54, 0)
        b6 <- c(-Inf, -1.51, 0)
        b7 <- c(-Inf, -1.49, 0)
        b8 <- c(-Inf, -1.80, 0)
        b9 <- c(-Inf, -2.22, 0)
        b10 <- c(-Inf, -2.09, 0)
        eta ~ 11.1
        eta.slope ~ 0.09
      })
      model({
        slp <- exp(slope + eta.slope) # modified to be mu-referenced
        drg <- dose * slp # drug-effect
        lge1 <- b1 + eta - drg
        lge2 <- b2 + lge1
        lge3 <- b3 + lge2
        lge4 <- b4 + lge3
        lge5 <- b5 + lge4
        lge6 <- b6 + lge5
        lge7 <- b7 + lge6
        lge8 <- b8 + lge7
        lge9 <- b9 + lge8
        lge10 <- b10 + lge9

        # Probabilities y >= X
        pge1 <- exp(lge1)/(1 + exp(lge1))
        pge2 <- exp(lge2)/(1 + exp(lge2))
        pge3 <- exp(lge3)/(1 + exp(lge3))
        pge4 <- exp(lge4)/(1 + exp(lge4))
        pge5 <- exp(lge5)/(1 + exp(lge5))
        pge6 <- exp(lge6)/(1 + exp(lge6))
        pge7 <- exp(lge7)/(1 + exp(lge7))
        pge8 <- exp(lge8)/(1 + exp(lge8))
        pge9 <- exp(lge9)/(1 + exp(lge9))
        pge10 <- exp(lge10)/(1 + exp(lge10))

        # Probabilities of y == X
        p0 <- (1    - pge1)
        p1 <- (pge1 - pge2)
        p2 <- (pge2 - pge3)
        p3 <- (pge3 - pge4)
        p4 <- (pge4 - pge5)
        p5 <- (pge5 - pge6)
        p6 <- (pge6 - pge7)
        p7 <- (pge7 - pge8)
        p8 <- (pge8 - pge9)
        p9 <- (pge9 - pge10)
        p10 <- pge10

        sp <- p0 + p1 + p2 + p3 + p4 + p5 + p6 + p7 + p8 + p9 + p10
        y ~ c(p0=0, p1=1, p2=2, p3=3, p4=4, p5=5, p6=6, p7=7, p8=8, p9=9, 10)
      })
    }

    tmp <- ord()
    expect_error(tmp$simulationModel, NA)

    ord <- function() {
      ini({
        b1 <- 7.86
        slope <- log(0.045)
        b2 <- c(-Inf, -1.73, 0)
        b3 <- c(-Inf, -1.95, 0)
        b4 <- c(-Inf, -1.55, 0)
        b5 <- c(-Inf, -1.54, 0)
        b6 <- c(-Inf, -1.51, 0)
        b7 <- c(-Inf, -1.49, 0)
        b8 <- c(-Inf, -1.80, 0)
        b9 <- c(-Inf, -2.22, 0)
        b10 <- c(-Inf, -2.09, 0)
        eta ~ 11.1
        eta.slope ~ 0.09
      })
      model({
        slp <- exp(slope + eta.slope) # modified to be mu-referenced
        drg <- dose * slp # drug-effect
        lge1 <- b1 + eta - drg
        lge2 <- b2 + lge1
        lge3 <- b3 + lge2
        lge4 <- b4 + lge3
        lge5 <- b5 + lge4
        lge6 <- b6 + lge5
        lge7 <- b7 + lge6
        lge8 <- b8 + lge7
        lge9 <- b9 + lge8
        lge10 <- b10 + lge9

        # Probabilities y >= X
        pge1 <- exp(lge1)/(1 + exp(lge1))
        pge2 <- exp(lge2)/(1 + exp(lge2))
        pge3 <- exp(lge3)/(1 + exp(lge3))
        pge4 <- exp(lge4)/(1 + exp(lge4))
        pge5 <- exp(lge5)/(1 + exp(lge5))
        pge6 <- exp(lge6)/(1 + exp(lge6))
        pge7 <- exp(lge7)/(1 + exp(lge7))
        pge8 <- exp(lge8)/(1 + exp(lge8))
        pge9 <- exp(lge9)/(1 + exp(lge9))
        pge10 <- exp(lge10)/(1 + exp(lge10))

        # Probabilities of y == X
        p0 <- (1    - pge1)
        p1 <- (pge1 - pge2)
        p2 <- (pge2 - pge3)
        p3 <- (pge3 - pge4)
        p4 <- (pge4 - pge5)
        p5 <- (pge5 - pge6)
        p6 <- (pge6 - pge7)
        p7 <- (pge7 - pge8)
        p8 <- (pge8 - pge9)
        p9 <- (pge9 - pge10)
        p10 <- pge10

        sp <- p0 + p1 + p2 + p3 + p4 + p5 + p6 + p7 + p8 + p9 + p10
        y ~ c(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9)
      })
    }

    tmp <- ord()
    expect_error(tmp$simulationModel, NA)

    f <- function() {
      ini({
        tkel <- 0.1
        tp0 <- -1
        eta.p ~ 0.02
        add.sd <- 0.2
      })
      model({
        kel <- tkel
        d/dt(kpd) <- -kel * kpd
        p1 <- expit(tp0 + eta.p)
        kpd ~ add(add.sd)
        cac ~ c(p1)
      })
    }

     tmp <- rxode2(f)

     expect_error(tmp$simulationModel, NA)
     expect_error(tmp$simulationIniModel, NA)

     tmp1 <- tmp$simulationModel

     tmp2 <- tmp$simulationIniModel

     expect_true(inherits(as.function(tmp1), "function"))
     expect_true(inherits(as.function(tmp2), "function"))

     expect_true(inherits(as.rxUi(tmp1), "rxUi"))
     expect_true(inherits(as.rxUi(tmp2), "rxUi"))

    ev <- et(amt=0.7, ii=24, until=7 * 24, cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=2) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithSeed(42, {
      s <- rxSolve(tmp, ev,
                   returnType="tibble", addCov=TRUE)

      s <- s |> dplyr::filter(CMT == 2)
      expect_equal(length(as.numeric(table(s$sim))), 2)

      expect_equal(sort(unique(s$sim)), c(1, 2))
    })

    rxWithSeed(42, {
      s <- rxSolve(tmp1, ev,
                   returnType="tibble", addCov=TRUE)

      s <- s |> dplyr::filter(CMT == 2)
      expect_equal(length(as.numeric(table(s$sim))), 2)

      expect_equal(sort(unique(s$sim)), c(1, 2))
    })

    rxWithSeed(42, {
      s <- rxSolve(tmp2, ev,
                   returnType="tibble", addCov=TRUE)

      s <- s |> dplyr::filter(CMT == 2)
      expect_equal(length(as.numeric(table(s$sim))), 2)

      expect_equal(sort(unique(s$sim)), c(1, 2))
    })

    f <- function() {
      ini({
        tkel <- 0.1
        tp0 <- -0.01
        eta.p ~ 0.02
        add.sd <- 0.2
      })
      model({
        kel <- tkel
        d/dt(kpd) <- -kel * kpd
        p1 <- expit(tp0 + eta.p)
        kpd ~ add(add.sd)
        cac ~ c(p1=0, 0.5)
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    ev <- et(amt=0.7, ii=24, until=7 * 24, cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=2) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithSeed(42, {

      s <- rxSolve(tmp, ev,
                   returnType="tibble", addCov=TRUE)

      s <- s |> dplyr::filter(CMT == 2)
      expect_equal(length(as.numeric(table(s$sim))), 2)

      expect_equal(sort(unique(s$sim)), c(0, 0.5))
    })
  })

  test_that("logLik simulations", {
    f <- function() {
      ini({
        tkel <- 0.1
        tp0 <- -3
        eta.p ~ 0.02
        add.sd <- 0.2
      })
      model({
        kel <- tkel
        d/dt(kpd) <- -kel * kpd
        p1 <- expit(tp0 + eta.p)
        kpd ~ add(add.sd)
        p2 <- -2 * log(p1)
        ll(lik) ~ p2
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    ev <-
      et(amt=0.7, ii=24, until=7 * 24, cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=2) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("normal simulations with dnorm()", {
    f <- function() {
      ini({
        tcl <- log(0.008)
        tv <-  log(0.6)
        eta.cl + eta.v ~ c(1,
                           0.01, 1)
        add.err <- 0.1
        lambda <- 0.5
      })
      model({
        cl <- exp(tcl + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v)    # individual value of volume
        ke <- cl / v            # elimination rate constant
        d/dt(A1) = - ke * A1    # model differential equation
        cp = A1 / v             # concentration in plasma
        cp ~ add(add.err) + boxCox(lambda) + dnorm() # define error model
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    ev <-
      et(amt=0.7, ii=24, until=7 * 24, cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=2) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("normal simulations", {
    f <- function() {
      ini({
        tcl <- log(0.008)
        tv <-  log(0.6)
        eta.cl + eta.v ~ c(1,
                           0.01, 1)
        add.err <- 0.1
        lambda <- 0.5
      })
      model({
        cl <- exp(tcl + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v)    # individual value of volume
        ke <- cl / v            # elimination rate constant
        d/dt(A1) = - ke * A1    # model differential equation
        cp = A1 / v             # concentration in plasma
        cp ~ add(add.err) + boxCox(lambda)# define error model
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    ev <-
      et(amt=0.7, ii=24, until=7 * 24, cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=2) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("t simulations", {
     f <- function() {
      ini({
        tcl <- log(0.008)
        tv <-  log(0.6)
        eta.cl + eta.v ~ c(1,
                           0.01, 1)
        add.err <- 0.1
        lambda <- 0.5
        nu <- 3
      })
      model({
        cl <- exp(tcl + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v)    # individual value of volume
        ke <- cl / v            # elimination rate constant
        d/dt(A1) = - ke * A1    # model differential equation
        cp = A1 / v             # concentration in plasma
        cp ~ prop(add.err) + boxCox(lambda) + dt(nu)# define error model
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("rxt[(]nu[)]", rxNorm(tmp$simulationModel)) != -1)

    ev <-
      et(amt=0.7, ii=24, until=7 * 24, cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=2) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("pois simulations", {
    f <- function() {
      ini({
        tlambda <- 0.5
        eta.lambda ~ 0.01
      })
      model({
        lambda <- exp(tlambda + eta.lambda)
        err ~ pois(lambda)
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("rpois[(]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("binom simulations", {
    f <- function() {
      ini({
        tn <- 0.5
        eta.n ~ 0.01
        prob <- logit(0.45)
      })
      model({
        n <- exp(tn + eta.n)
        p <- expit(prob)
        err ~ dbinom(n, p) | tmp
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("rxbinom[(]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("beta simulations", {
    f <- function() {
      ini({
        talpha <- 0.5
        eta.alpha ~ 0.01
        tbeta <- 3
        eta.beta ~ 0.01
      })
      model({
        alpha <- exp(talpha + eta.alpha)
        beta <- exp(tbeta + eta.beta)
        err ~ beta(alpha, beta)
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("rbeta[(]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("chisq simulations", {
    f <- function() {
      ini({
        tdf <- 0.5
        eta.df ~ 0.01
      })
      model({
        nu <- exp(tdf + eta.df)
        err ~ chisq(nu)
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("rchisq[(]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("dexp simulations", {
    f <- function() {
      ini({
        trate <- 0.5
        eta.rate ~ 0.01
      })
      model({
        r <- exp(trate + eta.rate)
        err ~ dexp(r)
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("rexp[(]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  ## Hyperbolic simulation not supported yet..
  ## test_that("dhyper simulations", {
  ##   f <- function() {
  ##     ini({
  ##       tm <- 0.5
  ##       eta.m ~ 0.01
  ##       tn <- 0.5
  ##       eta.n ~ 0.01
  ##       tk <- 0.7
  ##       eta.k ~ 0.01
  ##     })
  ##     model({
  ##       m <- exp(tm + eta.m)
  ##       n <- exp(tn + eta.n)
  ##       k <- exp(tk + eta.k)
  ##       err ~ dhyper(m, n, k)
  ##     })
  ##   }

  ##   tmp <- rxode2(f)

  ##   expect_error(tmp$simulationModel, NA)

  ##   expect_true(regexpr("rf[(]", rxNorm(tmp$simulationModel)) != -1)

  ##   ev <- et(seq(0.1, 24 * 8, by=12)) |>
  ##     et(id=1:20) |>
  ##     dplyr::as_tibble()

  ##   rxWithPreserveSeed({
  ##     expect_error(rxSolve(tmp, ev,
  ##                          returnType="tibble", addCov=TRUE), NA)
  ##   })

  ## })


  #  "unif"="runif",

  test_that("unif simulations", {

    f <- function() {
      ini({
        ta <- 0.5
        eta.a ~ 0.01
        tb <- 0.5
        eta.b ~ 0.01
      })
      model({
        a <- exp(ta + eta.a)
        b <- exp(tb + eta.b)
        err ~ dunif(a, b)
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("runif[(]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  #  "weibull"="rweibull",
  test_that("rweibull simulations", {
    f <- function() {
      ini({
        ta <- 0.5
        eta.a ~ 0.01
        tb <- 0.5
        eta.b ~ 0.01
      })
      model({
        a <- exp(ta + eta.a)
        b <- exp(tb + eta.b)
        err ~ dweibull(a, b)
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("rweibull[(]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("rcauchy simulations", {
    f <- function() {
      ini({
        tcl <- log(0.008)
        tv <-  log(0.6)
        eta.cl + eta.v ~ c(1,
                           0.01, 1)
        add.err <- 10
        lambda <- 0.5
      })
      model({
        cl <- exp(tcl + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v)    # individual value of volume
        ke <- cl / v            # elimination rate constant
        d/dt(A1) = - ke * A1    # model differential equation
        cp = A1 / v             # concentration in plasma
        cp ~ prop(add.err) + boxCox(lambda) + dcauchy()# define error model
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("rcauchy[(]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("rgamma simulations", {
    f <- function() {
      ini({
        ta <- 0.5
        eta.a ~ 0.01
        tb <- 0.5
        eta.b ~ 0.01
      })
      model({
        a <- exp(ta + eta.a)
        b <- exp(tb + eta.b)
        err ~ dgamma(a, b)
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("rgamma[(]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("rgeom simulations", {
    f <- function() {
      ini({
        ta <- logit(0.5)
        eta.a ~ 0.01
      })
      model({
        a <- expit(ta + eta.a)
        err ~ dgeom(a)
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("rgeom[(]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("omega/sigma=NA simulations", {
    f <- function() {
      ini({
        tcl <- log(0.008)
        tv <-  log(0.6)
        eta.cl + eta.v ~ c(1,
                           0.01, 1)
        add.err <- 0.1
        lambda <- 0.5
        nu <- 3
      })
      model({
        cl <- exp(tcl + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v)    # individual value of volume
        ke <- cl / v            # elimination rate constant
        d/dt(A1) = - ke * A1    # model differential equation
        cp = A1 / v             # concentration in plasma
        cp ~ add(add.err) + boxCox(lambda) + dt(nu)# define error model
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)

    expect_true(regexpr("rxt[(]nu[)]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(amt=0.7, ii=24, until=7 * 24, cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=1) |>
      et(seq(0.1, 24 * 8, by=12), cmt=2) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      .rx1 <- rxSolve(tmp, ev, addCov=TRUE)
      expect_true(all(.rx1$ipredSim != .rx1$sim))
      expect_true(all(.rx1$params$eta.cl != 0))
      expect_true(all(.rx1$params$eta.v != 0))

      suppressWarnings(
        .rx2 <- rxSolve(tmp, ev, omega=NA, addCov=TRUE)
      )
      expect_true(all(.rx2$ipredSim != .rx2$sim))
      expect_true(all(.rx2$params$eta.cl == 0))
      expect_true(all(.rx2$params$eta.v == 0))

      .rx3 <- rxSolve(tmp, ev, omega=NA, sigma=NA, addCov=TRUE)
      expect_true(any(names(.rx3) == "pred"))
      expect_true(all(.rx3$params$eta.cl == 0))
      expect_true(all(.rx3$params$eta.v == 0))
    })
  })

  test_that("negative binomial simulation", {
    f <- function() {
      ini({
        tn <- 0.5
        eta.n ~ 0.01
        prob <- logit(0.45)
      })
      model({
        n <- exp(tn + eta.n)
        p <- expit(prob)
        err ~ dnbinom(n, p)
      })
    }

    tmp <- rxode2(f)
    expect_error(tmp$simulationModel, NA)
    expect_true(regexpr("rxnbinom[(]n[,] *p[)]", rxNorm(tmp$simulationModel)) != -1)

    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("negative binomial simulation", {
    f <- function() {
      ini({
        tn <- 0.5
        eta.n ~ 0.01
        prob <- logit(0.45)
      })
      model({
        n <- exp(tn + eta.n)
        p <- expit(prob)
        err ~ dnbinomMu(n, p)
      })
    }

    tmp <- rxode2(f)

    expect_error(tmp$simulationModel, NA)
    expect_true(regexpr("rxnbinomMu[(]n[,] *p[)]", rxNorm(tmp$simulationModel)) != -1)


    ev <- et(seq(0.1, 24 * 8, by=12)) |>
      et(id=1:20) |>
      dplyr::as_tibble()

    rxWithPreserveSeed({
      expect_error(rxSolve(tmp, ev,
                           returnType="tibble", addCov=TRUE), NA)
    })
  })

  test_that("model without params() works",{

    mod <- function() {
      model({
        ## Table 3 from Savic 2007
        cl = 17.2 # (L/hr)
        vc = 45.1 # L
        ka = 0.38 # 1/hr
        mtt = 0.37 # hr
        bio=1
        n = 20.1
        k = cl/vc
        ktr = (n+1)/mtt
        ## note that lgammafn is the same as lgamma in R.
        d/dt(depot) = exp(log(bio*podo(depot))+log(ktr)+n*log(ktr*tad(depot))-
                            ktr*tad(depot)-lgammafn(n+1))-ka*depot
        d/dt(cen) = ka*depot-k*cen
      })
    }

    mod <- mod()

    expect_error(mod$simulationModel, NA)
    expect_error(mod$simulationIniModel, NA)
    expect_error(mod$symengineModelNoPrune, NA)
    expect_error(mod$symengineModelPrune, NA)

  })
})

rxTest({
  test_that("bounded functions needs numeric bounds", {
    lmat <- lotri({
      tka <- 0.2
      tcl <- 0.2
      tv <- 0.1
      eta.ka ~ 0.1
      eta.cl ~ 0.1
      eta.v ~ 0.1
      add.sd <- 0.1
    })

    testBounded <- function(type="expit") {
      expect_error(.rxMuRef(paste0("a=", type, "(tka + eta.ka, a, b)"), lmat))
      expect_error(.rxMuRef(paste0("a=", type, "(tka + eta.ka, 1, b)"), lmat))
      expect_error(.rxMuRef(paste0("a=", type, "(tka + eta.ka, 1, b)"), lmat))
      expect_error(.rxMuRef(paste0("a=", type, "(tka + eta.ka, 1, 2)"), lmat), NA)
      expect_error(.rxMuRef(paste0("a=", type, "(tka + eta.ka, 2, 1)"), lmat))
      expect_error(.rxMuRef(paste0("a=", type, "(tka + eta.ka, 0.5)"), lmat), NA)
      expect_error(.rxMuRef(paste0("a=", type, "(tka + eta.ka, a)"), lmat))
      expect_error(.rxMuRef(paste0("a=", type, "(tka + eta.ka, 4)"), lmat))
    }

    testBounded("logit")
    testBounded("expit")
    testBounded("probit")
    testBounded("probitInv")
  })

  lmat <- lotri({
    theta1 <- 1
    theta2 <- 1
    theta3 <- 1
    eta1 ~ 0.1
  })

  test_that("bad mu referencing examples (throw error)", {
    expect_error(.rxMuRef("a=theta1+theta2+theta3*wt+eta1", lmat))
    expect_error(.rxMuRef("a=theta1+theta2*wt+theta3*wt+eta1", lmat))
    # the 2+ population parameter error must name the parameters that
    # were actually summed (rxode2#471)
    expect_error(.rxMuRef("a=theta2+theta3+eta1", lmat), "'theta2', 'theta3'")
  })

  test_that("summing population parameters without an eta is not a mu-ref error (rxode2#471)", {
    # 'sqrt(sigma.1. + sigma.2.)'-style combined residual error sums two
    # population parameters with no random effect; this is legitimate and
    # must not be flagged as a mu-referenced expression
    expect_error(.rxMuRef("a=theta1+theta2", lmat), NA)
    expect_error(.rxMuRef("a=sqrt(theta1+theta2)", lmat), NA)

    f <- function() {
      ini({
        theta1 <- c(0, 1.08, 10)
        theta2 <- c(0, 0.814, 10)
        theta3 <- fix(0.179)
        eta1 ~ fix(0)
        sigma.1. <- 0.422
        sigma.2. <- 0.0364
      })
      model({
        TVKGA <- 0
        if (BACT == 2) TVKGA <- theta1
        TVKGP <- 0
        if (BACT == 1) TVKGP <- theta2
        TVKK <- theta3
        KGDXS <- 1
        KGA <- KGDXS * TVKGA * exp(eta1)
        KGP <- KGDXS * TVKGP * exp(eta1)
        IPRED <- KGA + KGP + TVKK
        W <- sqrt(sigma.1. + sigma.2.)
        Y <- IPRED + W * eps1
      })
    }
    expect_error(suppressMessages(rxode2(f)), NA)
  })

  testEnv <- function(env, ref) {
    lapply(names(ref), function(n) {
      expect_equal(get(n, envir=env), ref[[n]])
    })
    invisible()
  }

  listEnv <- function(env) {
    refNames <- c("muRefCovariateDataFrame", "muRefCovariateEmpty", "muRefCurEval", "muRefDataFrame",
                  "muRefDropParameters", "muRefExtra", "muRefExtraEmpty", "nonMuEtas")
    setNames(lapply(refNames, function(n) {
      get(n, env)
    }), refNames)
  }

  messageEnv <- function(env) {
    message(paste0(deparse(listEnv(env)), collapse="\n"))
  }

  test_that("simple mu referencing", {
    lmat <- lotri({
      tka <- 0.45
      tcl <- log(c(0, 2.7, 100))
      tv <- 3.45
      eta.ka ~ 0.6
      eta.cl ~ 0.3
      eta.v ~ 0.1
      add.sd <- 0.7
    })

    env <- .rxMuRef(rxode2({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- tv + eta.v
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      ## cp ~ add(add.sd)
    }), lmat)

    testEnv(env,
            list(muRefCovariateDataFrame = structure(list(theta = character(0),
                                                          covariate = character(0), covariateParameter = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefCovariateEmpty = c("tka", "tcl", "tv"), muRefCurEval = structure(list(
                   parameter = c("eta.ka", "tka", "eta.cl", "tcl", "eta.v",
                                 "tv"), curEval = c("exp", "exp", "exp", "exp", "", ""
                                                    ), low = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_,
                                                               NA_real_), hi = c(NA_real_, NA_real_, NA_real_, NA_real_,
                                                                                 NA_real_, NA_real_)), row.names = c(NA, -6L), class = "data.frame"),
                 muRefDataFrame = structure(list(theta = c("tka", "tcl", "tv"
                                                           ), eta = c("eta.ka", "eta.cl", "eta.v"), level = c("id",
                                                                                                              "id", "id")), row.names = c(NA, -3L), class = "data.frame"),
                 muRefDropParameters = structure(list(parameter = character(0),
                                                      term = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtra = structure(list(parameter = character(0), extra = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtraEmpty = c("tka", "tcl", "tv"), nonMuEtas = NULL))
  })

  test_that("shared eta is not mu referencing", {
    lmat <- lotri({
      t.EmaxA = 1
      t.EmaxB = 2
      t.EmaxC = 3
      eta.emax ~ 0.1
      add.sd <- 0.7
    })

    ## Test a duplicated eta; It shouldn't be counted as mu-referenced
    env <- .rxMuRef(rxode2({
      EmaxA <- exp(t.EmaxA + eta.emax)
      EmaxB <- exp(t.EmaxB + eta.emax)
      EmaxC <- exp(t.EmaxC + eta.emax)
    }), lmat)

    testEnv(env,
            list(muRefCovariateDataFrame = structure(list(theta = character(0),
                                                          covariate = character(0), covariateParameter = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefCovariateEmpty = c("t.EmaxA", "t.EmaxB", "t.EmaxC"),
                 muRefCurEval = structure(list(parameter = c("eta.emax", "t.EmaxA",
                                                             "t.EmaxB", "t.EmaxC"), curEval = c("exp", "exp", "exp", "exp"
                                                                                                ), low = c(NA_real_, NA_real_, NA_real_, NA_real_), hi = c(NA_real_,
                                                                                                                                                           NA_real_, NA_real_, NA_real_)), row.names = c(NA, -4L), class = "data.frame"),
                 muRefDataFrame = structure(list(theta = character(0), eta = character(0),
                                                 level = character(0)), row.names = integer(0), class = "data.frame"),
                 muRefDropParameters = structure(list(parameter = character(0),
                                                      term = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtra = structure(list(parameter = character(0), extra = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtraEmpty = c("t.EmaxA", "t.EmaxB", "t.EmaxC"), nonMuEtas = "eta.emax"))

    env <- .rxMuRef(rxode2({
      EmaxA <- exp(t.EmaxA + eta.emax)
      EmaxB <- exp(t.EmaxB + eta.emax)
      EmaxC <- t.EmaxC + eta.emax
    }), lmat)

    testEnv(env,
            list(muRefCovariateDataFrame = structure(list(theta = character(0),
                                                          covariate = character(0), covariateParameter = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefCovariateEmpty = c("t.EmaxA", "t.EmaxB", "t.EmaxC"),
                 muRefCurEval = structure(list(parameter = c("eta.emax", "t.EmaxA",
                                                             "t.EmaxB", "t.EmaxC"), curEval = c("", "exp", "exp", ""),
                                               low = c(NA_real_, NA_real_, NA_real_, NA_real_), hi = c(NA_real_,
                                                                                                       NA_real_, NA_real_, NA_real_)), row.names = c(NA, -4L
                                                                                                                                                     ), class = "data.frame"), muRefDataFrame = structure(list(
                                                                                                                                                       theta = character(0), eta = character(0), level = character(0)), row.names = integer(0), class = "data.frame"),
                 muRefDropParameters = structure(list(parameter = character(0),
                                                      term = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtra = structure(list(parameter = character(0), extra = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtraEmpty = c("t.EmaxA", "t.EmaxB", "t.EmaxC"), nonMuEtas = "eta.emax"))

    env <- .rxMuRef(rxode2({
      EmaxB <- t.EmaxB + eta.emax
      EmaxA <- exp(t.EmaxA + eta.emax)
    }), lmat)

    testEnv(env,
            list(muRefCovariateDataFrame = structure(list(theta = character(0),
                                                          covariate = character(0), covariateParameter = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefCovariateEmpty = c("t.EmaxB", "t.EmaxA"), muRefCurEval = structure(list(
                   parameter = c("eta.emax", "t.EmaxB", "t.EmaxA"), curEval = c("",
                                                                                "", "exp"), low = c(NA_real_, NA_real_, NA_real_), hi = c(NA_real_,
                                                                                                                                          NA_real_, NA_real_)), row.names = c(NA, -3L), class = "data.frame"),
                 muRefDataFrame = structure(list(theta = character(0), eta = character(0),
                                                 level = character(0)), row.names = integer(0), class = "data.frame"),
                 muRefDropParameters = structure(list(parameter = character(0),
                                                      term = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtra = structure(list(parameter = character(0), extra = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtraEmpty = c("t.EmaxB", "t.EmaxA"), nonMuEtas = "eta.emax"))

  })

  test_that("composite ode expressions", {

    lmat <- lotri({
      tka <- 0.45
      tcl <- log(c(0, 2.7, 100))
      tv <- 3.45
      eta.ka ~ 0.6
      eta.cl ~ 0.3
      eta.v ~ 0.1
      add.sd <- 0.7
    })

    env <- .rxMuRef(rxode2({
      d/dt(depot) = -exp(tka + eta.ka) * depot
      d/dt(center) = exp(tka + eta.ka) * depot - exp(tcl + eta.cl)/exp(tv + eta.v) * center
      cp = center/exp(tv + eta.v)
      #cp ~ add(add.sd)
    }), lmat)

    testEnv(
      env,
      list(
        muRefCovariateDataFrame =
          structure(list(
            theta = character(0),
            covariate = character(0), covariateParameter = character(0)), class = "data.frame", row.names = integer(0)
            ),
        muRefCovariateEmpty = c("tka", "tcl", "tv"),
        muRefCurEval = structure(list(
          parameter = c("eta.ka", "tka", "eta.cl", "tcl", "eta.v", "tv"),
          curEval = c("exp", "exp", "exp", "exp", "exp", "exp"),
          low = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_),
          hi = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_)),
          row.names = c(NA, -6L),
          class = "data.frame"
          ),
        muRefDataFrame = structure(list(
          theta = c("tka", "tcl", "tv"),
          eta = c("eta.ka", "eta.cl", "eta.v"),
          level = c("id", "id", "id")
        ),
        row.names = c(NA, -3L),
        class = "data.frame"
        ),
        muRefDropParameters = structure(list(
          parameter = character(0), term = character(0)), class = "data.frame", row.names = integer(0)),
        muRefExtra = structure(list(parameter = character(0), extra = character(0)), class = "data.frame", row.names = integer(0)),
        muRefExtraEmpty = c("tka", "tcl", "tv"), nonMuEtas = NULL)
    )

  })

  test_that("old style tka*eta(eta.ka)", {

    lmat <- lotri({
      tka <- 0.45
      tcl <- log(c(0, 2.7, 100))
      tv <- 3.45
      eta.ka ~ 0.6
      eta.cl ~ 0.3
      eta.v ~ 0.1
      add.sd <- 0.7
    })

    env <- .rxMuRef(rxode2({
      ka <- tka * exp(eta.ka + 0)
      cl <- tcl * exp(eta.cl + 0)
      v <- tv * exp(eta.v + 0)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      ## cp ~ add(add.sd)
    }), lmat)

    testEnv(env,
            list(muRefCovariateDataFrame = structure(list(theta = character(0),
                                                          covariate = character(0), covariateParameter = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefCovariateEmpty = NULL, muRefCurEval = structure(list(
                   parameter = c("tka", "eta.ka", "tcl", "eta.cl", "tv",
                                 "eta.v"), curEval = c("*", "exp", "*", "exp", "*", "exp"
                                                       ), low = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_,
                                                                  NA_real_), hi = c(NA_real_, NA_real_, NA_real_, NA_real_,
                                                                                    NA_real_, NA_real_)), row.names = c(NA, -6L), class = "data.frame"),
                 muRefDataFrame = structure(list(eta = character(0), theta = character(0),
                                                 level = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefDropParameters = structure(list(parameter = character(0),
                                                      term = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtra = structure(list(parameter = c("eta.ka", "eta.cl",
                                                           "eta.v"), extra = c("0", "0", "0")), row.names = c(NA, -3L
                                                                                                              ), class = "data.frame"), muRefExtraEmpty = NULL, nonMuEtas = c("eta.ka",
                                                                                                                                                                              "eta.cl", "eta.v")))

    env <- .rxMuRef(rxode2({
      ka <- tka * exp(eta.ka + 0)
      cl <- tcl * exp(eta.cl + 0)
      v <- tv * exp(eta.v + 0)
      v2 <- tv + eta.v
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      ## cp ~ add(add.sd)
    }), lmat)

    testEnv(env,
            list(muRefCovariateDataFrame = structure(list(theta = character(0),
                                                          covariate = character(0), covariateParameter = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefCovariateEmpty = "tv", muRefCurEval = structure(list(
                   parameter = c("tka", "eta.ka", "tcl", "eta.cl", "tv",
                                 "eta.v"), curEval = c("*", "exp", "*", "exp", "", ""),
                   low = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_,
                           NA_real_), hi = c(NA_real_, NA_real_, NA_real_, NA_real_,
                                             NA_real_, NA_real_)), row.names = c(NA, -6L), class = "data.frame"),
                 muRefDataFrame = structure(list(eta = character(0), theta = character(0),
                                                 level = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefDropParameters = structure(list(parameter = character(0),
                                                      term = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtra = structure(list(parameter = c("eta.ka", "eta.cl",
                                                           "eta.v"), extra = c("0", "0", "0")), row.names = c(NA, -3L
                                                                                                              ), class = "data.frame"), muRefExtraEmpty = "tv", nonMuEtas = c("eta.ka",
                                                                                                                                                                              "eta.cl", "eta.v")))

    env <- .rxMuRef(rxode2({
      ka <- tka * exp(eta.ka)
      cl <- tcl * exp(eta.cl)
      v <- tv * exp(eta.v)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      ## cp ~ add(add.sd)
    }), lmat)

    testEnv(env,
            list(muRefCovariateDataFrame = structure(list(theta = character(0),
                                                          covariate = character(0), covariateParameter = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefCovariateEmpty = NULL, muRefCurEval = structure(list(
                   parameter = c("tka", "eta.ka", "tcl", "eta.cl", "tv",
                                 "eta.v"), curEval = c("*", "exp", "*", "exp", "*", "exp"
                                                       ), low = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_,
                                                                  NA_real_), hi = c(NA_real_, NA_real_, NA_real_, NA_real_,
                                                                                    NA_real_, NA_real_)), row.names = c(NA, -6L), class = "data.frame"),
                 muRefDataFrame = structure(list(eta = character(0), theta = character(0),
                                                 level = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefDropParameters = structure(list(parameter = character(0),
                                                      term = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtra = structure(list(parameter = character(0), extra = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtraEmpty = NULL, nonMuEtas = c("eta.ka", "eta.cl",
                                                       "eta.v")))

    env <- .rxMuRef(rxode2({
      ka <- tka * exp(eta.ka)
      cl <- tcl * exp(eta.cl)
      v <- tv * exp(eta.v)
      etv <- eta.v
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      ## cp ~ add(add.sd)
    }), lmat)

    testEnv(env,
            list(muRefCovariateDataFrame = structure(list(theta = character(0),
                                                          covariate = character(0), covariateParameter = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefCovariateEmpty = NULL, muRefCurEval = structure(list(
                   parameter = c("tka", "eta.ka", "tcl", "eta.cl", "tv",
                                 "eta.v"), curEval = c("*", "exp", "*", "exp", "*", "exp"
                                                       ), low = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_,
                                                                  NA_real_), hi = c(NA_real_, NA_real_, NA_real_, NA_real_,
                                                                                    NA_real_, NA_real_)), row.names = c(NA, -6L), class = "data.frame"),
                 muRefDataFrame = structure(list(eta = character(0), theta = character(0),
                                                 level = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefDropParameters = structure(list(parameter = character(0),
                                                      term = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtra = structure(list(parameter = character(0), extra = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtraEmpty = NULL, nonMuEtas = c("eta.ka", "eta.cl",
                                                       "eta.v")))

  })


  test_that("curEval for theta only", {

    lmat <- lotri({
      tka <- 0.45
      tcl <- log(c(0, 2.7, 100))
      tv <- 3.45
      eta.cl ~ 0.3
      eta.v ~ 0.1
      add.sd <- 0.7
    })

    env <- .rxMuRef(rxode2({
      ka <- exp(tka)
      cl <- exp(tcl + eta.cl)
      v <- tv * exp(eta.v)
      etv <- eta.v
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      ## cp ~ add(add.sd)
    }), lmat)

    testEnv(env,
            list(muRefCovariateDataFrame = structure(list(theta = character(0),
                                                          covariate = character(0), covariateParameter = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefCovariateEmpty = "tcl", muRefCurEval = structure(list(
                   parameter = c("tka", "eta.cl", "tcl", "tv", "eta.v"),
                   curEval = c("exp", "exp", "exp", "*", "exp"),
                   low = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_),
                   hi = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_)),
                   row.names = c(NA, -5L), class = "data.frame"),
                 muRefDataFrame = structure(list(theta = "tcl", eta = "eta.cl", level = "id"),
                                            row.names = c(NA, -1L), class = "data.frame"), muRefDropParameters = structure(list(parameter = character(0), term = character(0)),
                                                                                                                           class = "data.frame", row.names = integer(0)),
                 muRefExtra = structure(list(parameter = character(0), extra = character(0)), class = "data.frame", row.names = integer(0)),
                 muRefExtraEmpty = "tcl", nonMuEtas = "eta.v"))

  })

  test_that("test covariates", {

    lmat <- lotri({
      tka <- 0.45
      tcl <- log(c(0, 2.7, 100))
      tv <- 3.45
      tvp <- 3.45
      cl.wt <- 0.1
      v.wt <- 0.1
      cl.sex <- 0.1
      v.sex <- 0.1
      cl.age <- 0.1
      v.age <- 0.1
      vp.wt <- 1
      vp.sex <- 1
      vp.age <- 1
      eta.ka ~ 0.6
      eta.cl ~ 0.3
      eta.v ~ 0.1
      add.sd <- 0.7
    })

    env <- .rxMuRef(rxode2({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt2 / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      ## cp ~ add(add.sd)
    }), lmat)

    testEnv(env,
            list(muRefCovariateDataFrame =
                   structure(list(theta = c("tcl", "tcl", "tv", "tv", "tv", "tvp", "tvp", "tvp"),
                                  covariate = c("age", "sex", "age", "sex", "wt", "age", "sex", "wt"),
                                  covariateParameter = c("cl.age", "cl.sex", "v.age", "v.sex", "v.wt", "vp.age", "vp.sex", "vp.wt")),
                             row.names = c(NA, -8L),
                             class = "data.frame"),
                 muRefCovariateEmpty = "tka",
                 muRefCurEval = structure(list(parameter = c("eta.ka", "tka", "eta.cl", "tcl", "eta.v", "tv", "tvp"),
                                               curEval = c("exp", "exp", "exp", "exp", "exp", "exp", "exp"),
                                               low = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_),
                                               hi = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_)),
                                          row.names = c(NA, -7L),
                                          class = "data.frame"),
                 muRefDataFrame = structure(list(theta = c("tka", "tcl", "tv"),
                                                 eta = c("eta.ka", "eta.cl", "eta.v"),
                                                 level = c("id", "id", "id")),
                                            row.names = c(NA, -3L),
                                            class = "data.frame"),
                 muRefDropParameters = structure(list(parameter = character(0),
                                                      term = character(0)),
                                                 class = "data.frame",
                                                 row.names = integer(0)),
                 muRefExtra = structure(list(parameter = c("tcl", "tv"),
                                             extra = c("3", "2")),
                                        row.names = c(NA, -2L), class = "data.frame"),
                 muRefExtraEmpty = c("tka", "tvp"), nonMuEtas = NULL))


    # This one tv is used in 2 covariate references
    env <- .rxMuRef(rxode2({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt2 / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tv + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      ## cp ~ add(add.sd)
    }), lmat)

    testEnv(env, list(muRefCovariateDataFrame =
                        structure(list(theta = c("tcl", "tcl"),
                                       covariate = c("age", "sex"),
                                       covariateParameter = c("cl.age", "cl.sex")),
                                  row.names = c(NA_integer_, -2L), class = "data.frame"),
                      muRefCovariateEmpty = c("tka", "tv"),
                      muRefCurEval =
                        structure(list(parameter = c("eta.ka","tka", "eta.cl", "tcl", "eta.v", "tv"),
                                       curEval = c("exp", "exp", "exp", "exp", "", ""),
                                       low = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_), hi = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_)),
                                  row.names = c(NA, -6L), class = "data.frame"),
                      muRefDataFrame =
                        structure(list(theta = c("tka", "tcl"),
                                       eta = c("eta.ka", "eta.cl"),
                                       level = c("id", "id")),
                                  row.names = c(NA, -2L), class = "data.frame"),
                      muRefDropParameters =
                        structure(list(parameter = c("tv", "tv", "tv", "tv", "tv", "tv", "tv"),
                                       term = c("age*vp.age", "sex*vp.sex", "wt*vp.wt", "age*v.age", "sex*v.sex", "wt*v.wt", "2")),
                                  row.names = c(NA, -7L), class = "data.frame"),
                      muRefExtra = structure(list(parameter = "tcl",
                                                  extra = "3"),
                                             row.names = c(NA_integer_, -1L),
                                             class = "data.frame"),
                      muRefExtraEmpty =
                        c("tka", "tv"),
                      nonMuEtas = "eta.v"))

    #env$nonMuEtas

    #expect_equal(env$nonMuEtas, "eta.v")

  })

  ## ## Composite expressions should be extracted to their own lines
  ## rxMuRef(rxode2({
  ##   ratio <- exp(t.EmaxA + eta.emaxA) / exp(t.EmaxB + eta.emaxB)
  ## }), theta=c("tka", "tcl", "tv", "add.sd"),
  ## eta=c("eta.ka", "eta.cl", "eta.v"))

  ## ## This should be expanded to eta.ka mu-referenced variables
  ## rxMuRef(rxode2({
  ##   ka <- tka * exp(eta.ka)
  ##   cl <- tcl * exp(eta.cl)
  ##   v <- tv * exp(eta.v)
  ##   d/dt(depot) = -ka * depot
  ##   d/dt(center) = ka * depot - cl/v * center
  ##   cp = center/v
  ##   ## cp ~ add(add.sd)
  ## }), theta=c("tka", "tcl", "tv", "add.sd"),
  ## eta=c("eta.ka", "eta.cl", "eta.v"))


  test_that("mu-ref detects curEval correctly", {

    pk.turnover.emax3 <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        ##
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        ##
        temax <- logit(0.8)
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        ##
        eta.emax ~ .5
        eta.ec50  ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        ##
        pdadd.err <- 10
        ##
        tdepot <- 1
        eta.depot ~ 0.1
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        emax = expit(temax+eta.emax)
        ec50 =  exp(tec50 + eta.ec50)
        kout = exp(tkout + eta.kout)
        e0 = te0 + eta.e0
        ##
        DCP = center/v
        PD=1-emax*DCP/(ec50+DCP)
        ##
        effect(0) = e0
        kin = e0*kout
        ##
        d/dt(depot) = -ktr * depot
        depot(0) = tdepot + eta.depot
        d/dt(gut) =  ktr * depot -ka * gut
        d/dt(center) =  ka * gut - cl / v * center
        d/dt(effect) = kin*PD -kout*effect
        ##
        cp = center / v
        cp ~ prop(prop.err) + add(pkadd.err)
        effect ~ add(pdadd.err) | pca
      })
    }

    ui <- rxode(pk.turnover.emax3)

    curEval <- ui$muRefCurEval

    w <- which(curEval$parameter == "tdepot")

    expect_equal(curEval$curEval[w], "")

    w <- which(curEval$parameter == "eta.depot")

    expect_equal(curEval$curEval[w], "")

  })

  test_that("test mu-reference covariate degradation", {

    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        cl.wt <- 0
        v.wt <- 0
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl + WT * cl.wt)
        v <- exp(tv + eta.v) + WT ^ 2 * v.wt
        linCmt() ~ add(add.sd)
      })
    }

    ui <- rxode(one.cmt)

    expect_equal(length(ui$muRefCovariateDataFrame$theta), 0)
    expect_true("eta.ka" %in% ui$muRefDataFrame$eta)
    expect_true("eta.cl" %in% ui$muRefDataFrame$eta)
    expect_true("eta.v" %in% ui$muRefDataFrame$eta)

    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        cl.wt <- 0
        v.wt <- 0
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)+ WT ^ 2* cl.wt
        v <- exp(tv + eta.v+ WT * v.wt)
        linCmt() ~ add(add.sd)
      })
    }

    ui <- rxode(one.cmt)

    expect_equal(length(ui$muRefCovariateDataFrame$theta), 0)
    expect_true("eta.ka" %in% ui$muRefDataFrame$eta)
    expect_true("eta.cl" %in% ui$muRefDataFrame$eta)
    expect_true("eta.v" %in% ui$muRefDataFrame$eta)


    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        cl.wt <- 0
        v.wt <- 0
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl + WT * cl.wt)
        v <- exp(tv + eta.v)
        v <- v + WT ^ 2 * v.wt
        linCmt() ~ add(add.sd)
      })
    }

    ui <- rxode(one.cmt)

    expect_equal(length(ui$muRefCovariateDataFrame$theta), 0)
    expect_true("eta.ka" %in% ui$muRefDataFrame$eta)
    expect_true("eta.cl" %in% ui$muRefDataFrame$eta)
    expect_true("eta.v" %in% ui$muRefDataFrame$eta)

  })
})

rxTest({

  # helper: pooled within-subject lag-k autocorrelation of a residual vector
  .arLagCor <- function(res, id, k=1) {
    .prev <- ave(res, id, FUN=function(x) c(rep(NA, k), head(x, -k)))
    .ok <- !is.na(.prev)
    cor(res[.ok], .prev[.ok])
  }

  test_that("ar() parses and stores per-endpoint information", {

    .estCor <- function() {
      ini({tcl <- log(1); tv <- log(10); add.sd <- 0.5; ar1.cor <- 0.5})
      model({
        cl <- exp(tcl); v <- exp(tv)
        d/dt(central) <- -cl / v * central
        cp <- central / v
        cp ~ add(add.sd) + ar(ar1.cor)
      })
    }
    ui <- rxode2(.estCor)
    # estimated correlation lives in iniDf (err == "ar"); there is no $predDf
    # ar column (it must never be added -- reverse dependencies rely on the
    # $predDf schema)
    expect_false("ar" %in% colnames(ui$predDf))
    .w <- which(ui$iniDf$name == "ar1.cor")
    expect_equal(ui$iniDf$err[.w], "ar")
    expect_equal(ui$iniDf$condition[.w], "cp")
    expect_equal(ui$iniDf$lower[.w], 0)
    expect_equal(ui$iniDf$upper[.w], 1)

    # a numeric literal correlation becomes an auto-generated FIX parameter in
    # the $iniDf (rx.<endpoint>.ar), not a $predDf entry
    .litCor <- function() {
      ini({tcl <- log(1); add.sd <- 0.5})
      model({cl <- exp(tcl); cp <- cl; cp ~ add(add.sd) + ar(0.5)})
    }
    .lit <- rxode2(.litCor)
    expect_false("ar" %in% colnames(.lit$predDf))
    .wl <- which(.lit$iniDf$name == "rx.cp.ar")
    expect_equal(length(.wl), 1L)
    expect_true(.lit$iniDf$fix[.wl])
    expect_equal(.lit$iniDf$est[.wl], 0.5)
    expect_equal(.lit$iniDf$err[.wl], "ar")
    expect_equal(.lit$iniDf$condition[.wl], "cp")

    # numeric literals in other error functions (add()/prop()) are likewise
    # turned into FIX $iniDf params (rx.<endpoint>.<func>) rather than being
    # silently dropped
    .litAddProp <- function() {
      ini({tcl <- log(1)})
      model({cl <- exp(tcl); cp <- cl; cp ~ add(0.7) + prop(0.1)})
    }
    .lap <- rxode2(.litAddProp)
    .wa <- which(.lap$iniDf$name == "rx.cp.add")
    expect_equal(length(.wa), 1L)
    expect_true(.lap$iniDf$fix[.wa])
    expect_equal(.lap$iniDf$est[.wa], 0.7)
    expect_equal(.lap$iniDf$err[.wa], "add")
    expect_equal(.lap$iniDf$condition[.wa], "cp")
    expect_equal(.lap$iniDf$lower[.wa], 0)
    .wp <- which(.lap$iniDf$name == "rx.cp.prop")
    expect_equal(length(.wp), 1L)
    expect_true(.lap$iniDf$fix[.wp])
    expect_equal(.lap$iniDf$est[.wp], 0.1)
    expect_equal(.lap$iniDf$err[.wp], "prop")
    expect_equal(.lap$iniDf$condition[.wp], "cp")

    # multi-argument error functions (pow(alpha, exponent)): a literal in a
    # non-first argument is fixed into $iniDf and named after its `err` spec
    # (rx.<endpoint>.<func><argno>), keeping an estimated first argument
    .litPow <- function() {
      ini({tcl <- log(1); pw <- 0.5})
      model({cl <- exp(tcl); cp <- cl; cp ~ pow(pw, 1.5)})
    }
    .lp <- rxode2(.litPow)
    # the estimated first argument stays a normal parameter
    .we <- which(.lp$iniDf$name == "pw")
    expect_equal(length(.we), 1L)
    expect_false(.lp$iniDf$fix[.we])
    expect_equal(.lp$iniDf$err[.we], "pow")
    # the literal exponent becomes a FIX param named rx.cp.pow2 (err == "pow2")
    .wpow2 <- which(.lp$iniDf$name == "rx.cp.pow2")
    expect_equal(length(.wpow2), 1L)
    expect_true(.lp$iniDf$fix[.wpow2])
    expect_equal(.lp$iniDf$est[.wpow2], 1.5)
    expect_equal(.lp$iniDf$err[.wpow2], "pow2")
    expect_equal(.lp$iniDf$condition[.wpow2], "cp")

    # both pow() arguments literal: distinct names keyed to their err specs
    # (rx.cp.pow / rx.cp.pow2), no collision on the bare function name
    .litPow2 <- function() {
      ini({tcl <- log(1)})
      model({cl <- exp(tcl); cp <- cl; cp ~ pow(1.5, 2)})
    }
    .lp2 <- rxode2(.litPow2)
    .wp1 <- which(.lp2$iniDf$name == "rx.cp.pow")
    .wp2 <- which(.lp2$iniDf$name == "rx.cp.pow2")
    expect_equal(length(.wp1), 1L)
    expect_equal(length(.wp2), 1L)
    expect_equal(.lp2$iniDf$est[.wp1], 1.5)
    expect_equal(.lp2$iniDf$est[.wp2], 2)
    expect_equal(.lp2$iniDf$err[.wp1], "pow")
    expect_equal(.lp2$iniDf$err[.wp2], "pow2")
    expect_true(all(.lp2$iniDf$fix[c(.wp1, .wp2)]))

    # a modeled correlation (a calculated model variable) is neither a parameter
    # nor a $predDf column; it is recovered from the endpoint error expression
    .modCor <- function() {
      ini({tcl <- log(1); tv <- log(10); add.sd <- 0.5; tcor <- 0.3})
      model({
        cl <- exp(tcl); v <- exp(tv); corv <- expit(tcor)
        d/dt(central) <- -cl / v * central
        cp <- central / v
        cp ~ add(add.sd) + ar(corv)
      })
    }
    .mod <- rxode2(.modCor)
    expect_false("ar" %in% colnames(.mod$predDf))
    expect_false("corv" %in% .mod$iniDf$name)
    expect_true(rxHasAr(.mod))
  })

  test_that("ar() works with a variety of transformably-normal error models", {
    # each case pairs an ini block (only the params it uses) with an rhs
    .cases <- list(
      c("add.sd <- 0.5; prop.sd <- 0.1; ar1.cor <- 0.4",
        "add(add.sd) + prop(prop.sd) + ar(ar1.cor)"),
      c("add.sd <- 0.5; ar1.cor <- 0.4",
        "lnorm(add.sd) + ar(ar1.cor)"),
      c("add.sd <- 0.5; df <- 5; lam <- 0.5; ar1.cor <- 0.4",
        "add(add.sd) + dt(df) + yeoJohnson(lam) + ar(ar1.cor)"),
      c("add.sd <- 0.5; lam <- 0.5; ar1.cor <- 0.4",
        "add(add.sd) + boxCox(lam) + ar(ar1.cor)"),
      c("ar1.cor <- 0.4",
        "cauchy() + ar(ar1.cor)"))
    for (.case in .cases) {
      .txt <- paste0(
        "function() {\n",
        "  ini({tcl <- log(1); tv <- log(10); ", .case[1], "})\n",
        "  model({\n",
        "    cl <- exp(tcl); v <- exp(tv)\n",
        "    d/dt(central) <- -cl / v * central\n",
        "    cp <- central / v\n",
        "    cp ~ ", .case[2], "\n",
        "  })\n}")
      .f <- eval(parse(text=.txt))
      expect_error(rxode2(.f), NA)
    }
  })

  test_that("ar() rejects invalid usage", {
    # ar alone (no normal error term)
    .arAlone <- function() {
      ini({tcl <- log(1); ar1.cor <- 0.5})
      model({cl <- exp(tcl); cp <- cl; cp ~ ar(ar1.cor)})
    }
    expect_error(rxode2(.arAlone))

    # ar twice on one endpoint
    .arTwice <- function() {
      ini({tcl <- log(1); add.sd <- 0.5; c1 <- 0.5; c2 <- 0.3})
      model({cl <- exp(tcl); cp <- cl; cp ~ add(add.sd) + ar(c1) + ar(c2)})
    }
    expect_error(rxode2(.arTwice))

    # ar with a generalized-likelihood distribution
    .arPois <- function() {
      ini({tcl <- log(1); ar1.cor <- 0.5})
      model({cl <- exp(tcl); cp <- cl; cp ~ pois(cp) + ar(ar1.cor)})
    }
    expect_error(rxode2(.arPois))

    # out-of-range numeric literal
    .arBadLit <- function() {
      ini({tcl <- log(1); add.sd <- 0.5})
      model({cl <- exp(tcl); cp <- cl; cp ~ add(add.sd) + ar(1.5)})
    }
    expect_error(rxode2(.arBadLit))
  })

  test_that("ar() generates lag0()-based simulation lines", {
    .f <- function() {
      ini({tcl <- log(1); tv <- log(10); add.sd <- 0.5; ar1.cor <- 0.8})
      model({
        cl <- exp(tcl); v <- exp(tv)
        d/dt(central) <- -cl / v * central
        cp <- central / v
        cp ~ add(add.sd) + ar(ar1.cor)
      })
    }
    .txt <- vapply(rxCombineErrorLines(rxode2(.f)), deparse1, character(1))
    .all <- paste(.txt, collapse="\n")
    # the self-referential AR(1) residual recurrence via lag0(), mirroring the
    # estimation likelihood (previous residual/time via lag0(), NaN-safe
    # first-record indicator via 1 - is.na(lag(...)))
    expect_true(grepl("rx.arRes.cp", .all, fixed=TRUE))
    expect_true(grepl("lag0(rx.arRes.cp, 1)", .all, fixed=TRUE))
    expect_true(grepl("is.na(lag(rx.arT.cp, 1))", .all, fixed=TRUE))
  })

  test_that("ar() simulation reproduces the target autocorrelation and stationary variance", {
    .f <- function() {
      ini({tcl <- log(1); tv <- log(10); add.sd <- 2; ar1.cor <- 0.8})
      model({
        cl <- exp(tcl); v <- exp(tv)
        d/dt(central) <- -cl / v * central
        cp <- central / v + 50
        cp ~ add(add.sd) + ar(ar1.cor)
      })
    }
    ui <- rxode2(.f)
    ev <- et(amt=100, cmt="central") %>% et(0:200)
    s <- rxWithSeed(42, rxSolve(ui, ev, nSub=200, returnType="data.frame"))
    .res <- s$sim - s$ipredSim
    # stationary marginal variance stays add.sd (pooled, unbiased)
    expect_equal(sd(.res), 2, tolerance=0.05)
    # continuous-time decay: lag-k correlation = cor^k on the unit grid
    expect_equal(.arLagCor(.res, s$sim.id, 1), 0.8, tolerance=0.03)
    expect_equal(.arLagCor(.res, s$sim.id, 2), 0.64, tolerance=0.03)
  })

  test_that("ar(0) reduces to iid residuals", {
    .f <- function() {
      ini({tcl <- log(1); tv <- log(10); add.sd <- 2; ar1.cor <- 0})
      model({
        cl <- exp(tcl); v <- exp(tv)
        d/dt(central) <- -cl / v * central
        cp <- central / v + 50
        cp ~ add(add.sd) + ar(ar1.cor)
      })
    }
    ui <- suppressMessages(rxode2(.f) |> ini(ar1.cor=0))
    ev <- et(amt=100, cmt="central") %>% et(0:100)
    s <- rxWithSeed(7, rxSolve(ui, ev, nSub=200, returnType="data.frame"))
    .res <- s$sim - s$ipredSim
    expect_equal(.arLagCor(.res, s$sim.id, 1), 0, tolerance=0.03)
  })

  test_that("ar() correlation decays with the actual time gap (continuous AR(1))", {
    .f <- function() {
      ini({tcl <- log(1); tv <- log(10); add.sd <- 2; ar1.cor <- 0.8})
      model({
        cl <- exp(tcl); v <- exp(tv)
        d/dt(central) <- -cl / v * central
        cp <- central / v + 50
        cp ~ add(add.sd) + ar(ar1.cor)
      })
    }
    ui <- rxode2(.f)
    # gaps of 1 then 2 repeated: 0,1,3,4,6,7,...
    .t <- as.vector(rbind(seq(0, 300, by=3), seq(1, 301, by=3)))
    .t <- sort(.t[.t <= 300])
    ev <- et(amt=100, cmt="central") %>% et(.t)
    s <- rxWithSeed(11, rxSolve(ui, ev, nSub=300, returnType="data.frame"))
    s$res <- s$sim - s$ipredSim
    .prev <- ave(s$res, s$sim.id, FUN=function(x) c(NA, head(x, -1)))
    .dt <- ave(s$time, s$sim.id, FUN=function(x) c(NA, diff(x)))
    # correlation across a gap of 1 vs a gap of 2 should be 0.8 vs 0.8^2
    .c1 <- cor(s$res[!is.na(.prev) & .dt == 1], .prev[!is.na(.prev) & .dt == 1])
    .c2 <- cor(s$res[!is.na(.prev) & .dt == 2], .prev[!is.na(.prev) & .dt == 2])
    expect_equal(.c1, 0.8, tolerance=0.04)
    expect_equal(.c2, 0.64, tolerance=0.04)
  })

  test_that("ar() is independent per endpoint in a multiple-endpoint model", {
    .f <- function() {
      ini({tcl <- log(1); tv <- log(10); a1 <- 2; a2 <- 3; c1 <- 0.9; c2 <- 0.3})
      model({
        cl <- exp(tcl); v <- exp(tv)
        d/dt(central) <- -cl / v * central
        cp <- central / v + 50
        ef <- 100 - central / v
        cp ~ add(a1) + ar(c1)
        ef ~ add(a2) + ar(c2)
      })
    }
    ui <- rxode2(.f)
    ev <- et(amt=100, cmt="central") %>%
      et(0:120, cmt="cp") %>%
      et(0:120, cmt="ef")
    s <- rxWithSeed(5, rxSolve(ui, ev, nSub=150, returnType="data.frame"))
    s$res <- s$sim - s$ipredSim
    .cp <- s[s$CMT == 2, ]
    .ef <- s[s$CMT == 3, ]
    expect_equal(.arLagCor(.cp$res, .cp$sim.id, 1), 0.9, tolerance=0.04)
    expect_equal(.arLagCor(.ef$res, .ef$sim.id, 1), 0.3, tolerance=0.04)
  })

})

rxTest({
  test_that("probitInv/probit can compile", {
    expect_error(rxode2("a=probit(b)"), NA)
    expect_error(rxode2("a=probit(b, c)"), NA)
    expect_error(rxode2("a=probit(b, c, d)"), NA)
    expect_error(rxode2("a=probitInv(b)"), NA)
    expect_error(rxode2("a=probitInv(b, c)"), NA)
    expect_error(rxode2("a=probitInv(b, c, d)"), NA)
  })

  test_that("probitInv/probit work", {
    p <- seq(0.01, 0.99, by = 0.01)
    x <- qnorm(p)
    f <- rxode2("a=probit(time);c=probitInv(a)")
    s <- rxSolve(f, et(p))
    expect_equal(x, s$a)
    expect_equal(p, s$c)
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("markov llik", {

      markov <- function() {
        ini({
          logitp02 <- logit(0.2)
          label("Probablity of transition from 0 to 2")
          logitp20 <- logit(0.2)
          label("Probablity of transition from 2 to 0")
        })
        model({
          p02 <- expit(logitp02)
          p00 <- 1 - p02
          p20 <- expit(logitp20)
          p22 <- 1 - p20
          p <-
            p02*(PDV == 0 & DV == 2) +
            p00*(PDV == 0 & DV == 0) +
            p20*(PDV == 2 & DV == 0) +
            p22*(PDV == 2 & DV == 2)
          ll(err) ~ log(p)
        })
      }

      f <- rxode2(markov)

      expect_true(all(f$muRefCurEval$low == 0))
      expect_true(all(f$muRefCurEval$hi == 1))

    })
  }
})

rxTest({

  test_that("plogis ui translation normalizes to expit expressions", {
    f <- function() {
      model({
        a <- plogis(x, scale = 10 + 1, location = a)
        b <- plogis(x, lower.tail = FALSE)
        c <- plogis(logitx, lower.tail = FALSE, log.p = TRUE)
      })
    }

    expect_equal(
      f() |> modelExtract(),
      c(
        "a <- expit((x - a)/11, 0, 1)",
        "b <- expit(-x, 0, 1)",
        "c <- log(expit(-logitx, 0, 1))"
      )
    )
  })

  test_that("plogis string parsing lowers to expit-compatible models", {
    mod <- function() {
      model({
        a=plogis(time, 1, 2, 0, 0)
        b=plogis(time, 1, 2, 1, 1)
      })
    }
    et <- et(seq(0, 8, length.out = 9))
    s <- suppressWarnings(rxSolve(mod, et))
    expect_equal(s$a, stats::plogis(s$time, location = 1, scale = 2, lower.tail = FALSE, log.p = FALSE))
    expect_equal(s$b, stats::plogis(s$time, location = 1, scale = 2, lower.tail = TRUE, log.p = TRUE))
  })

})

rxTest({

  # The Yeo-Johnson transform is monotone increasing, so rxTBSd() must be
  # positive everywhere.  On the negative branch lambda == 2 is a special case
  # (yj(x) = -log(1 - x)); it used to return the derivative with the wrong sign,
  # which also made it disagree with the general formula in the limit.

  .yjModel <- function() {
    model({
      v <- rxTBS(xv, lam, 1, 1, 0)
      d1 <- rxTBSd(xv, lam, 1, 1, 0)
      d2 <- rxTBSd2(xv, lam, 1, 1, 0)
    })
  }

  .yj <- function(x, lambda) {
    .et <- et(seq_along(x))
    .et$xv <- x
    .et$lam <- lambda
    as.data.frame(rxSolve(.yjModel, .et, cores = 1L))[, c("v", "d1", "d2")]
  }

  test_that("yeoJohnson derivatives at lambda = 2 have the right sign (negative branch)", {
    x <- c(-3, -1, -0.5)
    r <- .yj(x, 2.0)
    # yj(x) = -log(1 - x); yj'(x) = 1/(1 - x); yj''(x) = 1/(1 - x)^2
    expect_equal(r$v, -log1p(-x))
    expect_equal(r$d1, 1 / (1 - x))
    expect_equal(r$d2, 1 / ((1 - x) * (1 - x)))
    expect_true(all(r$d1 > 0))
    expect_true(all(r$d2 > 0))
  })

  test_that("yeoJohnson derivatives are continuous in lambda at lambda = 2", {
    x <- c(-3, -1, -0.5)
    at2 <- .yj(x, 2.0)
    lo <- .yj(x, 2.0 - 1e-6)
    hi <- .yj(x, 2.0 + 1e-6)
    expect_equal(at2$d1, lo$d1, tolerance = 1e-5)
    expect_equal(at2$d1, hi$d1, tolerance = 1e-5)
    expect_equal(at2$d2, lo$d2, tolerance = 1e-5)
    expect_equal(at2$d2, hi$d2, tolerance = 1e-5)
  })

  test_that("yeoJohnson derivatives match finite differences on both branches", {
    x <- c(-3, -1, -0.5, 0.5, 2)
    h <- 1e-4
    for (lambda in c(2.0, 0.5, 1.5, 0)) {
      r <- .yj(x, lambda)
      fp <- .yj(x + h, lambda)$v
      fm <- .yj(x - h, lambda)$v
      expect_equal(r$d1, (fp - fm) / (2 * h), tolerance = 1e-5)
      expect_equal(r$d2, (fp - 2 * r$v + fm) / (h * h), tolerance = 1e-3)
    }
  })

  # General transform model: yj code and bounds as covariates.  Codes: 4 logit,
  # 5 logit + yeoJohnson, 6 probit, 7 probit + yeoJohnson; x in (lo, hi).  For
  # yj 5/7, x below the midpoint gives a negative inner logit/probit value, so
  # these also exercise the negative Yeo-Johnson branch (including lambda == 2).
  .tbsModel <- function() {
    model({
      v <- rxTBS(xv, lam, yjc, lo, hi)
      d1 <- rxTBSd(xv, lam, yjc, lo, hi)
      d2 <- rxTBSd2(xv, lam, yjc, lo, hi)
      vi <- rxTBSi(v, lam, yjc, lo, hi)
    })
  }

  .tbs <- function(x, lambda, yjc, lo = 0, hi = 1) {
    .et <- et(seq_along(x))
    .et$xv <- x
    .et$lam <- lambda
    .et$yjc <- yjc
    .et$lo <- lo
    .et$hi <- hi
    as.data.frame(rxSolve(.tbsModel, .et, cores = 1L))[, c("v", "d1", "d2", "vi")]
  }

  test_that("logit/probit (+ yeoJohnson) derivatives match finite differences", {
    x <- c(0.15, 0.35, 0.5, 0.75, 0.9)
    h <- 1e-6
    for (yjc in c(4, 5, 6, 7)) {
      for (lambda in c(2.0, 1.0, 0.5, 0)) {
        r <- .tbs(x, lambda, yjc)
        fp <- .tbs(x + h, lambda, yjc)
        fm <- .tbs(x - h, lambda, yjc)
        expect_equal(r$d1, (fp$v - fm$v) / (2 * h), tolerance = 1e-4)
        expect_equal(r$d2, (fp$d1 - fm$d1) / (2 * h), tolerance = 1e-4)
      }
    }
  })

  test_that("rxTBSi() inverts rxTBS() for the composed transforms", {
    x <- c(0.15, 0.35, 0.5, 0.75, 0.9)
    for (yjc in c(4, 5, 6, 7)) {
      for (lambda in c(2.0, 1.0, 0.5, 0)) {
        r <- .tbs(x, lambda, yjc)
        expect_equal(r$vi, x, tolerance = 1e-6)
      }
    }
  })

  # boxCox (yj = 0) and lnorm (yj = 3) clamp x at sqrt(.Machine$double.eps)
  # for stability; the clamped x is then run through the usual derivative
  # formula, so the derivatives are continuous at the clamp boundary (they
  # used to return the clamp constant ~1.5e-8 itself below the boundary).
  test_that("boxCox/lnorm derivatives clamp x and evaluate the formula", {
    .eps <- sqrt(.Machine$double.eps)
    x <- c(0, .eps, 0.5, 2)
    xc <- pmax(x, .eps)
    r <- .tbs(x, 1.0, 3)
    expect_equal(r$d1, 1 / xc)
    expect_equal(r$d2, -1 / (xc * xc))
    for (lambda in c(0, 0.5, 2.0)) {
      r <- .tbs(x, lambda, 0)
      if (lambda == 0) {
        expect_equal(r$d1, 1 / xc)
        expect_equal(r$d2, -1 / (xc * xc))
      } else {
        expect_equal(r$d1, xc^(lambda - 1))
        expect_equal(r$d2, (lambda - 1) * xc^(lambda - 2))
      }
    }
  })

  test_that("clamped boxCox/lnorm derivatives are continuous at the boundary", {
    .eps <- sqrt(.Machine$double.eps)
    x <- c(0, .eps * (1 + 1e-8))
    r <- .tbs(x, 0.5, 0)
    expect_equal(r$d1[1], r$d1[2], tolerance = 1e-6)
    expect_equal(r$d2[1], r$d2[2], tolerance = 1e-6)
    r <- .tbs(x, 1.0, 3)
    expect_equal(r$d1[1], r$d1[2], tolerance = 1e-6)
    expect_equal(r$d2[1], r$d2[2], tolerance = 1e-6)
  })

  # powerL (log-Jacobian) and powerDL (its lambda gradient) are only shared
  # via R_RegisterCCallable (nlmixr2est), so they are pinned through the
  # internal .rxTransformL() test hook.
  test_that("powerL/powerDL log-Jacobian and lambda gradient", {
    .pL <- function(x, lambda, yj) {
      rxode2:::.rxTransformL(x, lambda, transform = yj)
    }
    .pDL <- function(x, lambda, yj) {
      rxode2:::.rxTransformL(x, lambda, transform = yj, dLambda = TRUE)
    }
    # yeoJohnson log-Jacobian: (lambda-1)*log1p(x) for x >= 0,
    # (1-lambda)*log1p(-x) for x < 0
    expect_equal(.pL(c(2, -0.5, -2), 0.5, 1),
                 c(-0.5 * log1p(2), 0.5 * log1p(0.5), 0.5 * log1p(2)))
    # lambda gradient log1p(x) / -log1p(-x), including at exactly lambda == 1
    # (used to return 0 there) and x < -1 (used to be NaN)
    for (lam in c(0.25, 1, 1.75)) {
      expect_equal(.pDL(c(2, -0.5, -2), lam, 1),
                   c(log1p(2), -log1p(0.5), -log1p(2)))
    }
    # boxCox gradient log(x), including at exactly lambda == 1
    expect_equal(.pDL(2, 1, 0), log(2))
    # lambda-free transforms have a zero gradient (norm, lnorm, logit, probit)
    for (yj in c(2, 3, 4, 6)) {
      expect_equal(.pDL(0.5, 0.7, yj), 0)
    }
    # composed transforms chain through the inner logit/probit value
    # (yj = 7 used to return NA)
    expect_equal(.pDL(0.25, 0.5, 5), -log1p(log(1 / 0.25 - 1)))
    expect_equal(.pDL(0.25, 0.5, 7), -log1p(-qnorm(0.25)))
    # powerDL is the lambda derivative of powerL
    h <- 1e-6
    for (yj in c(0, 1, 5, 7)) {
      x <- c(2, -0.5, 0.35, 0.35)[match(yj, c(0, 1, 5, 7))]
      fd <- (.pL(x, 0.75 + h, yj) - .pL(x, 0.75 - h, yj)) / (2 * h)
      expect_equal(.pDL(x, 0.75, yj), fd, tolerance = 1e-6)
    }
    # logit log-Jacobian is finite at the upper bound (clamped)
    expect_true(is.finite(.pL(1, 1, 4)))
    expect_error(rxode2:::.rxTransformL(0.5, low = 1, high = 0, transform = 4L),
                 "'high' must be greater than 'low'")
  })

})

## Refactor Jacobian calculation
#' Faster expand.grid
#'
#' Only support x and y as characters right now
#'
#' @param x first element (must be character)
#' @param y second element (must be character)
#' @param type Internal type=0L is traditional expand grid and
#'     type=1L is Jacobian expand grid (adds symbols)
#' @return Expand grid
#' @author Matthew Fidler
#' @keywords internal
#' @examples
#'
#' ##
#' rxExpandGrid(letters, letters)
#'
#' ## Another fast method; See
#' ## https://stackoverflow.com/questions/10405637/use-outer-instead-of-expand-grid
#'
#' expand.grid.jc <- function(seq1, seq2) {
#'   cbind(
#'     Var1 = rep.int(seq1, length(seq2)),
#'     Var2 = rep.int(seq2, rep.int(length(seq1), length(seq2)))
#'   )
#' }
#' \donttest{
#' microbenchmark::microbenchmark(rxExpandGrid(letters, letters), expand.grid.jc(letters, letters))
#' }
#' @export
rxExpandGrid <- function(x, y, type = 0L) {
  .Call(`_rxode2_rxExpandGrid_`, x, y, type)
}

## Assumes model is loaded.
#'  Internal function for calculating the Jacobian
#'
#'
#' @param model symengine environment
#' @return Jacobian information
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxInjectMatExpOdes <- function(model) {
  .mv <- get("..mv", envir = model, inherits = FALSE)
  if (!is.list(.mv$indLin) || length(.mv$indLin) != 4L) {
    return(invisible(FALSE))
  }
  .states <- rxStateOde(model)
  if (length(.states) == 0L) {
    return(invisible(FALSE))
  }
  if (exists(paste0("rx__d_dt_", .states[1L], "__"), envir = model, inherits = FALSE)) {
    return(invisible(FALSE))
  }
  .ddt <- setNames(replicate(length(.states), symengine::S("0"), simplify = FALSE), .states)
  .sym <- function(name) get(name, envir = model, inherits = FALSE)
  for (.p in ls(envir = model, all.names = TRUE)) {
    .m <- regexec("^k[_.]([^_.]+)[_.]([^_.]+)$", .p)[[1L]]
    if (length(.m) == 1L) next
    .from <- substring(.p, .m[2L], .m[2L] + attr(.m, "match.length")[2L] - 1L)
    .to <- substring(.p, .m[3L], .m[3L] + attr(.m, "match.length")[3L] - 1L)
    if (!(.from %in% .states)) next  # check names, not integer values
    .rate <- .sym(.p)
    .ddt[[.from]] <- .ddt[[.from]] - .rate * .sym(.from)
    if (.to %in% .states) {
      .ddt[[.to]] <- .ddt[[.to]] + .rate * .sym(.from)
    }
  }
  ## Add any indLin() forcing functions (e.g. Michaelis-Menten elimination)
  ## captured during the symengine load.  Each is stored as a per-state symengine
  ## variable rx__indLinForce_<state>__; they add to the matrix-derived d/dt().
  for (.s in .states) {
    .forceName <- paste0("rx__indLinForce_", .s, "__")
    if (exists(.forceName, envir = model, inherits = FALSE)) {
      .ddt[[.s]] <- .ddt[[.s]] + get(.forceName, envir = model, inherits = FALSE)
    }
  }
  for (.s in .states) {
    assign(paste0("rx__d_dt_", .s, "__"), .ddt[[.s]], envir = model)
  }
  invisible(TRUE)
}

#'  Internal function for calculating the Jacobian
#'
#' @param model symengine environment
#' @param vars Variables
#' @return Jacobian information
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxJacobian <- function(model, vars = TRUE) {
  .rxInjectMatExpOdes(model)
  if (rxIs(vars, "logical")) {
    if (vars) {
      .pars <- .rxParams(model, TRUE)
      .pars <- setdiff(.pars, rxLhs(model))
      if (any(.pars == "ETA[1]")) {
        .pars <- .pars[regexpr(rex::rex(start, "ETA[", any_numbers, "]"), .pars) != -1]
      }
      .jac <- rxExpandGrid(
        rxStateOde(model),
        c(rxStateOde(model), .pars),
        1L
      )
    } else {
      .pars <- NULL
      .jac <- rxExpandGrid(
        rxStateOde(model),
        rxStateOde(model),
        1L
      )
    }
  } else if (rxIs(vars, "character")) {
    .pars <- vars
    .jac <- rxExpandGrid(
      rxStateOde(model),
      c(rxStateOde(model), vars),
      1L
    )
  }
  assign("..vars", .pars, envir = model)
  ## .malert("calculate jacobian")
  rxProgress(dim(.jac)[1])
  on.exit({
    rxProgressAbort()
  })
  .ret <- apply(.jac, 1, function(x) {
    .l <- x["line"]
    .l <- eval(parse(text = .l))
    rxTick()
    paste0(x["rx"], "=", rxFromSE(.l))
  })
  assign("..jacobian", .ret, envir = model)
  rxProgressStop()
  return(.ret)
}

#' Full-system Jacobian for a first-order forward-sensitivity model, reusing F_X
#'
#' Builds the full df()/dy() block for the sensitivity-expanded system (base
#' states + `rx__sens_*` variational compartments) from the `F_X`/`F_p` pieces
#' `.rxJacobian(model)` already produced.  The base x base and sens x sens blocks
#' are `F_X` (reused, no differentiation); only the sens x base block needs new
#' `symengine::D` and it vanishes for state-independent (linear) systems.
#' Requires `.rxJacobian(model)` (vars=TRUE) to have run.
#'
#' @param model symengine model environment (post `.rxJacobian`)
#' @param state base ODE state names
#' @param params sensitivity parameter names (`calcSens`)
#' @return character vector of `df(state)/dy(state)` lines for the full system
#' @author Matthew L. Fidler
#' @keywords internal
.rxFwdSensJacBlock <- function(model, state, params) {
  .ns <- length(state); .np <- length(params)
  .fx <- function(i, j) get0(paste0("rx__df_", state[i], "_dy_", state[j], "__"),
                             envir = model, inherits = FALSE)
  .fp <- function(i, p) get0(paste0("rx__df_", state[i], "_dy_", params[p], "__"),
                             envir = model, inherits = FALSE)
  .sn <- function(i, p) paste0("rx__sens_", state[i], "_BY_", params[p], "__")
  .lines <- character(0)
  .emit <- function(lhsSt, rhsSt, e) {
    if (is.null(e)) return(invisible())
    .txt <- rxode2::rxFromSE(e)
    if (!identical(.txt, "0") && nzchar(.txt))
      .lines[[length(.lines) + 1L]] <<- sprintf("df(%s)/dy(%s)=%s", lhsSt, rhsSt, .txt)
  }
  ## base x base and sens x sens (block-diagonal per parameter): both are F_X,
  ## re-emitted without differentiation.
  for (i in seq_len(.ns)) for (j in seq_len(.ns)) .emit(state[i], state[j], .fx(i, j))
  for (p in seq_len(.np)) for (i in seq_len(.ns)) for (k in seq_len(.ns))
    .emit(.sn(i, p), .sn(k, p), .fx(i, k))
  ## sens x base: the only block that needs new derivatives; identically zero
  ## whenever F_X and F_p are state-independent (linear systems).
  for (i in seq_len(.ns)) for (j in seq_len(.ns)) {
    .dF_X_k <- list()
    for (k in seq_len(.ns)) {
      .fxik <- .fx(i, k); if (is.null(.fxik)) next
      .d <- symengine::D(.fxik, state[j])
      if (paste(.d) != "0") {
        .dF_X_k[[k]] <- .d
      }
    }
    for (p in seq_len(.np)) {
      .acc <- NULL
      for (k in seq_along(.dF_X_k)) {
        .d <- .dF_X_k[[k]]
        if (!is.null(.d)) {
          .term <- .d * symengine::Symbol(.sn(k, p))
          .acc <- if (is.null(.acc)) .term else .acc + .term
        }
      }
      .fpip <- .fp(i, p)
      if (!is.null(.fpip)) {
        .d <- symengine::D(.fpip, state[j])
        if (paste(.d) != "0") .acc <- if (is.null(.acc)) .d else .acc + .d
      }
      .emit(.sn(i, p), state[j], .acc)
    }
  }
  .lines
}

## Assumes .rxJacobian called on model c(state,vars)
#'  Sensitivity for model
#'
#' @param model symengine model environment
#' @param vars Variables for single sensitivity
#' @param vars2 if present, 2 parameter sensitivity
#' @return Sensitivity
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxSens <- function(model, vars, vars2, vars3) {
  .state <- rxStateOde(model)
  if (length(.state) > 0L) {
    if (missing(vars)) vars <- get("..vars", envir = model)
    if (missing(vars2)) {
      ## delay differential equations: reject state-dependent delays before the
      ## progress bar opens (so the error message is not masked).
      .rxDelayValidateTauSE(model)
    } else {
      ## second- and higher-order: parameter-dependent delays are not yet
      ## supported (moving breaking points -> jump discontinuities); reject
      ## before the progress bar so the message is not masked.
      .rxDelayValidateHigherOrderSE(model, union(union(vars, vars2),
                                                 if (missing(vars3)) NULL else vars3),
                                    thirdOrder = !missing(vars3))
      ## third-order only covers linear delays; reject nonlinear ones early too.
      if (!missing(vars3)) .rxDelayValidate3rdLinearSE(model)
    }
    if (!missing(vars3)) {
      .grd <- rxode2::rxExpandSens3_(.state, vars, vars2, vars3)
    } else if (!missing(vars2)) {
      .grd <- rxode2::rxExpandSens2_(.state, vars, vars2)
    } else {
      .grd <- rxode2::rxExpandSens_(.state, vars)
    }
    .malert("calculate sensitivities")
    rxProgress(dim(.grd)[1])
    on.exit({
      rxProgressAbort()
    })
    lapply(c(.grd$ddtS, .grd$ddS2), function(x) {
      assign(x, symengine::Symbol(x), envir = model)
    })
    .ret <- apply(.grd, 1, function(x) {
      .l <- x["line"]
      .l <- eval(parse(text = .l))
      .ret <- paste0(x["ddt"], "=", rxFromSE(.l))
      if (exists(x["s0"], envir = model)) {
        .l <- x["s0D"]
        .l <- eval(parse(text = .l))
        if (paste(.l) != "0") {
          .ret <- paste0(
            .ret, "\n", x["s0r"], "=", rxFromSE(.l),
            "+0.0"
          )
        }
      }
      rxTick()
      return(.ret)
    })
    if (!missing(vars3)) {
      ## Delay differential equations: add the third-order delayed (variational)
      ## terms (constant-delay) to each third-order sensitivity ODE.
      .ret <- .rxDelaySensAugment3(model, .ret, union(union(vars, vars2), vars3))
      assign("..sens3", .ret, envir = model)
    } else if (missing(vars2)) {
      ## Delay differential equations: add the delayed (variational) terms
      ## d f_i/d[delay(y_j,T)] * delay(S_j,T) to each first-order sensitivity
      ## ODE.  No-op when the model has no delay() terms.  Placed here so both
      ## rxGetModel(calcSens=) and nlmixr2est's foceiEtaS path are covered.
      .ret <- .rxDelaySensAugment(model, .ret, vars)
      assign("..sens", .ret, envir = model)
    } else {
      ## Delay differential equations: add the second-order delayed (variational)
      ## terms (constant-delay) to each second-order sensitivity ODE.  No-op
      ## without delay() terms.
      .ret <- .rxDelaySensAugment2(model, .ret, union(vars, vars2))
      assign("..sens2", .ret, envir = model)
    }
    rxProgressStop()
  } else {
    assign("..sens", NULL, envir = model)
    .ret <- NULL
  }
  return(.ret)
}


## Assumes .rxJacobian called on model with c(state, vars)
#' Adjoint (backward) sensitivity equations for a model
#'
#' Symbolically generates the continuous-adjoint ODE system mirroring the
#' forward sensitivity output (`.rxSens`), reusing the same
#' `rx__sens_<state>_BY_<param>__` output names.  For each output state `k` it
#' emits two backward-in-time blocks: a costate
#' `d/dt(rx__adjLambda_<k>_<i>__) = -(J^T lambda)_i` and a quadrature
#' `d/dt(rx__sens_<k>_BY_<p>__) = -(lambda^T df/dp)_p`.  The derivatives are
#' taken from the `rx__df_*` symbols `.rxJacobian()` already materialised, so no
#' new differentiation is done; the costate symbols are registered as bare
#' symengine `Symbol`s.
#'
#' @param model symengine model environment (as returned by `rxS`), with
#'   `.rxJacobian(model, c(rxStateOde(model), vars))` already called.
#' @param vars character vector of parameter names to differentiate with
#'   respect to (the `p` in `dy/dp`).  Defaults to `model$..vars`.
#' @param states character vector of output states of interest (the `k`).
#'   Defaults to all ODE states, matching the full coverage of forward
#'   sensitivities.
#' @return character vector of `d/dt(...) = ...` lines (costate block followed
#'   by quadrature block).  The costate (`rx__adjLambda_*`) lines are internal
#'   scaffolding; only the quadrature (`rx__sens_*`) lines carry user-visible
#'   output.  Also stashes the result in `model$..adjoint`.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxAdjoint <- function(model, vars, states) {
  .state <- rxStateOde(model)
  if (length(.state) == 0L) {
    assign("..adjoint", NULL, envir = model)
    return(NULL)
  }
  if (missing(vars)) vars <- get("..vars", envir = model)
  if (missing(states) || is.null(states)) states <- .state
  states <- intersect(.state, states)
  if (length(states) == 0L) {
    stop("adjoint 'states' of interest must be a subset of the ODE states",
      call. = FALSE)
  }
  ## register the costate (lambda) compartments as bare symbols so they survive
  ## assembly of the backward right-hand sides
  .lamName <- function(k, i) paste0("rx__adjLambda_", k, "_", i, "__")
  for (.k in states) {
    for (.i in .state) {
      .nm <- .lamName(.k, .i)
      assign(.nm, symengine::Symbol(.nm), envir = model)
    }
  }
  .malert("calculate adjoint sensitivities")
  rxProgress(length(states) * (length(.state) + length(vars)))
  on.exit({
    rxProgressAbort()
  })
  .ret <- character(0)
  for (.k in states) {
    ## costate block: for each state i, -sum_j (df_j/dy_i) * lambda_{k,j}
    for (.i in .state) {
      .terms <- vapply(.state, function(.j) {
        paste0("rx__df_", .j, "_dy_", .i, "__*", .lamName(.k, .j))
      }, character(1L))
      .expr <- paste0("with(model,-(", paste(.terms, collapse = "+"), "))")
      .l <- eval(parse(text = .expr))
      .ret <- c(.ret, paste0("d/dt(", .lamName(.k, .i), ")=", rxFromSE(.l)))
      rxTick()
    }
    ## quadrature block: for each param p, -sum_j lambda_{k,j} * (df_j/dp)
    for (.p in vars) {
      .terms <- vapply(.state, function(.j) {
        paste0(.lamName(.k, .j), "*rx__df_", .j, "_dy_", .p, "__")
      }, character(1L))
      .expr <- paste0("with(model,-(", paste(.terms, collapse = "+"), "))")
      .l <- eval(parse(text = .expr))
      .ret <- c(.ret, paste0("d/dt(rx__sens_", .k, "_BY_", .p, "__)=",
                             rxFromSE(.l)))
      rxTick()
    }
  }
  assign("..adjoint", .ret, envir = model)
  rxProgressStop()
  .ret
}


## Check for good functions for predfn and pkpars and error functions
.goodFns <- c(".GlobalEnv", "package:rxode2", "package:nlmixr")
.checkGood <- function(x) {
  .tmp <- suppressWarnings({
    utils::find(deparse1(substitute(x)))
  })
  if (!identical(.tmp, character())) {
    if (!any(.tmp == .goodFns)) {
      stop(sprintf(gettext("'%s' is from '%s' and cannot be used in this context"), deparse1(substitute(x)), .tmp),
        call. = FALSE
      )
    }
  }
}

#' Get the state information for the current model
#'
#' @param obj rxode2 model that can take rxModelVars
#' @return character vector of initial preserved states (state),
#'     extra states (statef) and dvid translation information
#'     (dvid). This is used in generating the final rxode2 model.
#' @author Matthew Fidler
#' @noRd
.rxGenFunState <- function(obj) {
  .mv0 <- rxModelVars(obj)
  .curDvid <- .mv0$dvid
  .state0 <- .mv0$state
  if (.rxLinNcmt(.mv0)["numLin"] > 0L) {
    .state0 <- rxStateOde(.mv0)
  }
  if (length(.state0) > 0) {
    .state0 <- paste(paste0("cmt(", .state0, ");\n"), collapse = "")
  } else {
    .state0 <- ""
  }
  .statef <- .mv0$stateExtra
  if (length(.statef) > 0) {
    .statef <- paste0(paste(paste0("\ncmt(", .mv0$stateExtra, ");"), collapse = ""), "\n")
  } else {
    .statef <- ""
  }
  if (length(.curDvid) > 1) {
    .dvidF <- paste0(
      "\ndvid(",
      paste(.curDvid, collapse = ","), ");\n"
    )
  } else {
    .dvidF <- ""
  }
  return(c(state = .state0, statef = .statef, dvid = .dvidF))
}

.rxFixR <- function(.newmod, addProp) {
  if (!exists("..fixR", envir = .newmod)) {
    if (exists("rx_r_", envir = .newmod)) {
      ## Breaks focei for non-trivial examples
      ## if (!rxErrEnv.hasAdd) {
      ##   ## Convert abs() to abs1()
      ##   .r <- get("rx_r_", envir = .newmod)
      ##   .r <- paste0("abs1(", rxFromSE(.r), ")")
      ##   .r <- symengine::S(rxToSE(.r))
      ##   assign("rx_r_", .r, envir=.newmod)
      ## }
      if (addProp == "combined1") {
        assign("rx_r_", get("rx_r_", envir = .newmod)^2, envir = .newmod)
      }
      assign("..fixR", TRUE, envir = .newmod)
    }
  }
}

#' Build restore lines for captured adaptive dosing calls
#'
#' @param captures List of capture records from `rxPrune()` attribute `"capturedEvid"`
#' @return Character vector of restore lines, one per captured call
#' @noRd
.restoreAdaptiveDosing <- function(captures) {
  vapply(captures, function(cap) {
    paste0("if (", cap$capVar, ") { ", cap$original, " }")
  }, character(1L), USE.NAMES = FALSE)
}

.rxLoadPrune <- function(mod, doConst = TRUE, promoteLinSens = TRUE, fullModel = FALSE,
                         addProp = c("combined2", "combined1")) {
  addProp <- match.arg(addProp)
  ## if (fullModel) {
  ##   .malert("pruning branches ({.code if}/{.code else}) of full model...")
  ## } else {
  ##   .malert("pruning branches ({.code if}/{.code else})...")
  ## }
  .pruned <- rxPrune(mod)
  .captures <- attr(.pruned, "capturedEvid")
  .newmod <- rxGetModel(.pruned)
  ## .msuccess("done")
  ## message("Loading into symengine environment...", appendLF=FALSE)
  ## if (fullModel) {
  ##   .malert("loading full model into {.pkg symengine} environment...")
  ## } else {
  ##   .malert("loading into {.pkg symengine} environment...")
  ## }
  .newmod <- rxS(.newmod, doConst, promoteLinSens = promoteLinSens)
  if (length(.captures) > 0L) {
    .newmod$..capturedEvid <- .captures
    .newmod$..restoreLines <- .restoreAdaptiveDosing(.captures)
  }
  .rxFixR(.newmod, addProp)
  ## .msuccess("done")
  return(.newmod)
}

.rxIsOp <- function(x) {
  (identical(x, quote(`*`)) ||
    identical(x, quote(`**`)) ||
    identical(x, quote(`^`)) ||
    identical(x, quote(`+`)) ||
    identical(x, quote(`-`)) ||
    identical(x, quote(`/`)) ||
    identical(x, quote(`==`)) ||
    identical(x, quote(`!=`)) ||
    identical(x, quote(`>=`)) ||
    identical(x, quote(`<=`)) ||
    identical(x, quote(`<`)) ||
    identical(x, quote(`>`)) ||
    identical(x, quote(`&&`)) ||
    identical(x, quote(`&`)) ||
    identical(x, quote(`||`)) ||
    identical(x, quote(`|`)))
}
## norm <- rxode2("
## d/dt(y)  = dy
## d/dt(dy) = mu*(1-y^2)*dy - y
## ## Initial conditions
## y(0) = 2
## dy(0) = 0
## ## mu
## mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
## ")

rxTest({
  # Analytic third-order forward sensitivities (rxExpandSens3_ / .rxSens vars3)
  # must match a central finite difference of the analytic second-order
  # sensitivities (which differentiates an exact derivative, so the agreement
  # is finite-difference-step limited rather than formula limited).

  .nm <- function(cS, ...) paste0("rx__sens_", cS, "_BY_",
                                  paste(c(...), collapse = "_BY_"), "__")

  m <- "d/dt(depot)  = -ka*depot\nd/dt(center) = ka*depot - (cl/v)*center"
  model <- rxode2::rxS(rxode2::rxGetModel(m), TRUE, promoteLinSens = FALSE)
  vars  <- c("ka", "cl", "v")
  st    <- rxode2::rxStateOde(model)

  invisible(rxode2::.rxJacobian(model, c(st, vars)))
  s1 <- rxode2::.rxSens(model, vars)              # first order
  s2 <- rxode2::.rxSens(model, vars, vars)        # second order
  s3 <- rxode2::.rxSens(model, vars, vars, vars)  # third order (new path)

  test_that("rxExpandSens3_ stores ..sens3 and emits one ODE per triple/state", {
    expect_true(!is.null(model$..sens3))
    expect_equal(length(s3), length(st) * length(vars)^3)
  })

  mod <- rxode2::rxode2(paste(c(
    "d/dt(depot)  = -ka*depot",
    "d/dt(center) =  ka*depot - (cl/v)*center",
    s1, s2, s3), collapse = "\n"))

  p0     <- c(ka = 1.1, cl = 3.0, v = 20.0)
  ev     <- rxode2::et(amt = 100, cmt = "depot") |> rxode2::et(seq(0.25, 12, by = 0.25))
  solve1 <- function(p) as.data.frame(rxode2::rxSolve(mod, params = p, ev,
                                                      returnType = "data.frame",
                                                      atol = 1e-12, rtol = 1e-12))
  base <- solve1(p0)

  # d^3 x / d(pa) d(pb) d(pc) vs central FD of d^2 x / d(pb) d(pc) over pa
  relErr <- function(pa, pb, pc, state, h = 1e-4) {
    pp <- p0; pm <- p0; pp[pa] <- pp[pa]*(1+h); pm[pa] <- pm[pa]*(1-h); hh <- p0[pa]*h
    d <- (solve1(pp)[[.nm(state, pb, pc)]] - solve1(pm)[[.nm(state, pb, pc)]]) / (2*hh)
    a <- base[[.nm(state, pa, pb, pc)]]
    max(abs(a - d) / (abs(a) + 1e-8))
  }

  combos <- list(c("ka","ka","ka","center"), c("cl","cl","cl","center"),
                 c("v","v","v","center"),    c("cl","ka","ka","center"),
                 c("v","cl","ka","center"),  c("ka","cl","v","center"),
                 c("ka","ka","ka","depot"),  c("v","v","cl","center"))

  test_that("analytic third-order sensitivities match finite differences", {
    err <- max(vapply(combos, function(z) relErr(z[1], z[2], z[3], z[4]), numeric(1)))
    expect_true(err < 1e-4)
  })

  test_that("mixed third derivatives are symmetric in their variables", {
    # d^3/dka dcl dv is independent of ordering
    expect_equal(base[[.nm("center","ka","cl","v")]],
                 base[[.nm("center","v","cl","ka")]], tolerance = 1e-8)
    expect_equal(base[[.nm("center","ka","cl","v")]],
                 base[[.nm("center","cl","ka","v")]], tolerance = 1e-8)
  })
})

rxTest({
  expand.grid.jc <- function(seq1, seq2) {
    cbind(
      Var1 = rep.int(seq1, length(seq2)),
      Var2 = rep.int(seq2, rep.int(length(seq1), length(seq2)))
    )
  }

  skip_if_not_installed("microbenchmark")
  x <- microbenchmark::microbenchmark(rxExpandGrid(letters, letters), expand.grid.jc(letters, letters))

  x <- as.data.frame(x) |>
    dplyr::group_by(expr) |>
    dplyr::summarize(x = median(time)) |>
    dplyr::arrange(expr)

  ## test_that("rxExpandGrid is faster that expand.grid.jc", {
  ##     expect_true(x$x[1] < x$x[2]);
  ## })

  f <- function() {
    tmp <- setNames(data.frame(expand.grid.jc(letters, letters),
                               stringsAsFactors = FALSE
                               ), c("s1", "s2"))
    tmp <- cbind(tmp, with(tmp, data.frame(
      rx = paste0("df(", s1, ")/dy(", s1, ")"),
      sym = paste0("__d_df_", s1, "_dy_", s2, "__"),
      line = paste0("__d_df_", s1, "_dy_", s2, "__=diff(rx__d_dt_", s1, "__, ", s2, ")")
    )))
    return(tmp)
  }

  x <- microbenchmark::microbenchmark(
    rxExpandGrid(letters, letters, 1L),
    f()
  )

  x <- as.data.frame(x) |>
    dplyr::group_by(expr) |>
    dplyr::summarize(x = median(time)) |>
    dplyr::arrange(expr)

  test_that("rxExpandGrid is faster than printing out letters", {
    skip_if_not(x$x[1] < x$x[2],"rxExpandGrid is not faster than printing out letters (in codecov)")
    expect_true(x$x[1] < x$x[2])
  })
})

#' Determines if the object is a valid object for serialization
#'
#' Currently this is only data frames and rxModelVars objects
#'
#' @param obj object to test for validity
#' @return boolean, TRUE if valid for serialization in rxode2
#' @keywords internal
#' @noRd
#' @author Matthew L. Fidler
.validSerializationObject <- function(obj) {
  .cls <- class(obj)
  if (length(.cls) == 1L &&
        (inherits(obj, "rxModelVars") ||
           inherits(obj, "data.frame"))) {
    return(TRUE)
  }
  FALSE
}
#' Get the Default Serialization Type
#'
#' @return string indicating the default serialization type
#'
#' @export
#'
#' @author Matthew L. Fidler
#'
#' @examples
#'
#' rxGetDefaultSerialize()
#'
rxGetDefaultSerialize <- function() {
  op <- getOption("rxode2.serialize.type", "bzip2")
  if (!op %in% c("base", "bzip2", "xz")) {
    stop("option 'rxode2.serialize.type' must be one of 'base', 'bzip2' or 'xz'", call.=FALSE)
  }
  op
}
#' Serialize an R Object to a Raw Vector
#'
#' @param x object to serialize
#'
#' @param type serialization type; one of "base", "xz" or "bzip2".
#'
#' @return raw vector
#'
#' @export
#'
#' @author Matthew L. Fidler
#' @keywords internal
#' @examples
#'
#' rxSerialize(mtcars)
#'
#' rxRawToC(mtcars)
#'
rxSerialize <- function(x, type=c("xz", "bzip2", "base")) {
  ## Suggested for security reasons to limit what can be deserialized
  if (missing(type)) {
    type <- rxGetDefaultSerialize()
  }
  if (!.validSerializationObject(x)) {
    .cls <- class(x)
    stop("serialization object of class ",
         paste(.cls, collapse=", "),
         " is not supported")
  }
  switch(match.arg(type),
         bzip2 = {
           memCompress(serialize(x, NULL), type="bzip2")
         },
         xz = {
           memCompress(serialize(x, NULL), type="xz")
         },
         base = {
           serialize(x, NULL)
         },
         stop("unknown serialization type") # nocov
         )
}
#' Deserialize a Raw Vector or String to an R Object
#'
#' @param x raw vector or string to deserialize
#' @return R object
#' @keywords internal
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' rxDeserialize(rxSerialize(mtcars))
#'
rxDeserialize <- function(x) {
  if (checkmate::testCharacter(x, len=1L, any.missing=FALSE)) {
    .x <- try(qs2::base91_decode(x), silent=TRUE)
    if (inherits(.x, "try-error")) {
      stop("Input must be a raw vector or base91 encoded string")
    }
    x <- .x
  }
  if (!inherits(x, "raw")) {
    stop("Input must be a raw vector or base91 encoded string")
  }
  .type <- .Call(`_rxode2_rxGetSerialType_`, x)
  # 'qs2'/'qdata' are only ever read, never written; they come from objects
  # stored while 'qs2' was still an allowed 'rxode2.serialize.type'
  .ret <- try(switch(.type,
                     qs2 = {
                       qs2::qs_deserialize(x)
                     },
                     qdata = {
                       qs2::qd_deserialize(x)
                     },
                     qs = {
                       rxReq("qs")
                       .Call(`_rxode2_qsDes`, x)
                     },
                     bzip2 = {
                        unserialize(memDecompress(x, type="bzip2"))
                     },
                     xz = {
                       unserialize(memDecompress(x, type="xz"))
                     },
                     base = {
                       unserialize(x)
                     },
                     stop("Unknown serialization type")), silent=TRUE)
  if (inherits(.ret, "try-error")) {
    stop("Deserialization failed")
  }
  .cls <- class(.ret)
  ## Suggested for security reasons to limit what can be deserialized
  if (.validSerializationObject(.ret)) {
    return(.ret)
  }
  stop("Deserialized object of class ",
       paste(.cls, collapse=", "),
       " is not supported")
}
#' Convert a Raw Vector or R object to C Code
#'
#' This function converts a raw vector or R object to C code that
#' is used in the rxode2 model
#'
#' @return raw vector for conversion to C in the codegen procedure
#'
#' @keywords internal
#' @inherit rxSerialize
#' @param  raw vector or R object to convert
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' rxRawToC(mtcars)
#'
rxRawToC <- function(raw, type=c("xz", "base", "bzip2")) {
  if (missing(type)) {
    type <- rxGetDefaultSerialize()
  }
  if (inherits(raw, "raw")) {
    raw
  } else {
    rxRawToC(rxSerialize(raw, type=type))
  }
}

#' Deserialize a Raw Vector using qs
#'
#' @param x raw vector to deserialize using qs (not on CRAN any more)
#' @return qs::deserialize (if qs is available) of x
#' @export
#' @keywords internal
#' @author Matthew L. Fidler
rxOldQsDes <- function(x) {
  rxReq("qs")
  .Call(`_rxode2_qsDes`, x)
}

rxTest({

  mv <- rxModelVars("a=b")

  test_that("rxSerialize and rxDeserialize work with model variables", {
    r.base <- rxSerialize(mv, "base")
    expect_equal(rxDeserialize(r.base), mv)
    r.base <- rxSerialize(mv, "bzip2")
    expect_equal(rxDeserialize(r.base), mv)
    r.base <- rxSerialize(mv, "xz")
    expect_equal(rxDeserialize(r.base), mv)
  })

  df <- data.frame(a=1:10, b=11:20)

  test_that("rxSerialize and rxDeserialize work with simple data", {
    r.base <- rxSerialize(df, "base")
    expect_equal(rxDeserialize(r.base), df)
    r.base <- rxSerialize(df, "bzip2")
    expect_equal(rxDeserialize(r.base), df)
    r.base <- rxSerialize(df, "xz")
    expect_equal(rxDeserialize(r.base), df)
  })

  test_that("rxSerialize and rxDeserialize fail with bad input", {

    badInput <- "not a good object"

    expect_error(rxSerialize(badInput))

    expect_error(rxDeserialize(badInput))

    expect_error(rxDeserialize("av"))

    expect_error(rxSerialize(df, "qs2"))

  })

  test_that("serial type", {
    expect_equal(rxGetSerialType_(as.raw("0")), "unknown")
    # qs/qs2 are not dependencies, test by header only
    expect_equal(rxGetSerialType_(as.raw(c(0xFD, 0x37, 0x7A, 0x58, 0x5A, 0x00))), "xz")
    expect_equal(rxGetSerialType_(charToRaw("BZh")), "bzip2")
    expect_equal(rxGetSerialType_(as.raw(c(0x0B, 0x0E, 0x0A, 0x0C))), "qs")
    expect_equal(rxGetSerialType_(as.raw(c(0x0B, 0x0E, 0x0A, 0xC1))), "qs2")
    expect_equal(rxGetSerialType_(as.raw(c(0x0B, 0x0E, 0x0A, 0xCD))), "qdata")
    expect_equal(rxGetSerialType_(serialize("matt", NULL)), "base")
    expect_error(rxGetSerialType_(4))
    expect_equal(rxGetSerialType_(as.raw(c(0x58, 0x0A))), "base")
    expect_equal(rxGetSerialType_(as.raw(c(0x41, 0x0A))), "base")
    expect_equal(rxGetSerialType_(as.raw(c(0x42, 0x0A))), "base")
    expect_equal(rxGetSerialType_(as.raw(c(0x43, 0x0A))), "unknown")
    expect_equal(rxGetSerialType_(as.raw(c(0x43, 0x0A, 0x0A))), "unknown")
  })

  test_that("legacy qs2 objects can still be read back", {
    # qs2 is only ever read, never written; these come from objects stored
    # while 'qs2' was an allowed rxode2.serialize.type.
    #
    # Guarded even though qs2 is an Import: it pulls in stringfish, whose CRAN
    # macOS binary currently fails to *load* (undefined TBB symbol against the
    # installed RcppParallel) even though it installs.  requireNamespace()
    # returns FALSE for that, so this skips rather than erroring.  rxode2
    # itself is unaffected -- it never loads qs2 outside rxDeserialize().
    skip_if_not_installed("qs2")
    expect_equal(rxDeserialize(qs2::qs_serialize(mv)), mv)
    expect_equal(rxDeserialize(qs2::qd_serialize(mv)), mv)
    expect_equal(rxDeserialize(qs2::qs_serialize(df)), df)
    expect_equal(rxDeserialize(qs2::qd_serialize(df)), df)
    expect_error(rxDeserialize(qs2::qs_serialize("not a good object")))
  })

})

.rbind2rxSove <- function(v, v2) {
  .tmp <- v$add.dosing # make sure info is calculated
  .tmp <- v2$add.dosing
  .cls <- class(v)
  .env1 <- attr(class(v), ".rxode2.env")
  .env2 <- attr(class(v2), ".rxode2.env")
  if (length(.env1$.check.names) != length(.env2$.check.names) &&
        !all(.env1$.check.names == .env2$.check.names)) {
    stop("cannot rbind these 2 rxSolve objects", call.=FALSE)
  }
  ## if (is.null(.env1$.params.single) || is.null(.env2$.params.single)) {
  ##   stop("cannot rbind single solve environments", call.=FALSE)
  ## }
  .cloneEnv <- new.env(parent=emptyenv())
  for (.v in ls(.env1, all.names=TRUE)) {
    assign(.v, get(.v, envir=.env1), envir=.cloneEnv)
  }
  .nStud1 <- .cloneEnv$.args$nStud
  .nStud2 <- .env2$.args$nStud
  if (.nStud1 > 1 && .nStud2 > 1) {
    .args <- .cloneEnv$.args
    .args$nStud <- .nStud1 + .nStud2
    .cloneEnv$.args <- .args
    .v <- v
    class(.v) <- "data.frame"
    .v2 <- v2
    .v2$sim.id <- .nStud1 + .v2$sim.id
    class(.v2) <- "data.frame"
    .v <- rbind(.v, .v2)
    .cloneEnv$.nsub <- .cloneEnv$.nsub + .env2$.nsub
    .cloneEnv$.et <- NULL
    .cloneEnv$.args.params <- NULL
    .cloneEnv$.args.inits <- NULL
    .cloneEnv$.args.object <- rxModelVars(.cloneEnv$.args.object)
    .cloneEnv$.args.par0 <- NULL
    .cloneEnv$.args.params <- NULL
    .cloneEnv$.check.nrow <- .cloneEnv$.check.nrow + .env2$.check.nrow
    .cloneEnv$nobs <- .cloneEnv$nobs + .env2$nobs
    .cloneEnv$.nsim <- .cloneEnv$.nsim + .env2$.nsim
    .cloneEnv$.nsub <- NULL
    .cloneEnv$.dadt.counter <- 0L
    .cloneEnv$.init.dat <- setNames(rep(NA_real_, length(.cloneEnv$.init.dat)),
                                    names(.cloneEnv$.init.dat))
    .cloneEnv$.jac.counter <- 0L
    .cloneEnv$.nsub <- NA_integer_
    .cloneEnv$.par.pos <- NULL
    .cloneEnv$.par.pos.ini <- NULL
    .pd <- .env2$.params.dat
    .pd$sim.id <- .nStud1 + .pd$sim.id

    .cloneEnv$.params.dat <- rbind(.cloneEnv$.params.dat, .pd)
    .cloneEnv$.params.single <- NULL
    .cloneEnv$.par.pos <- NULL
    .cloneEnv$.par.pos.ini <- NULL
    .cloneEnv$.et <- NULL
    if (inherits(.v$id, "factor")) {
      .cloneEnv$.idLevels <- levels(.v$id)
    }
    .cloneEnv$.jac.counter <- .cloneEnv$.slvr.counter <- 0L
    .cloneEnv$.nsim <- .cloneEnv$.nsim + .env2$.nsim
    .cloneEnv$.real.update <- FALSE
    .cloneEnv$.sigma <- NULL
    .fun <- function(...) {
      stop("functions don't work on rbound rxSolve", call.=FALSE)
    }
    .cloneEnv$.replace.sampling <- .fun
    .cloneEnv$add.dosing <- .fun
    .cloneEnv$add.sampling <- .fun
    .cloneEnv$clear.dosing <- .fun
    .cloneEnv$clear.sampling <- .fun
    .cloneEnv$get.dosing <- .fun
    .cloneEnv$get.EventTable <- .fun
    .cloneEnv$get.nobs <- .fun
    .cloneEnv$get.obs.rec <- .fun
    .cloneEnv$get.sampling <- .fun
    .cloneEnv$get.units <- .fun
    .cloneEnv$import <- .fun
    .cloneEnv$counts.EventTable <- NULL
    .cloneEnv$get.units <- NULL
    .cloneEnv$units <- c(dosing="NA", time="NA")
    .cloneEnv$dll <- NULL
    attr(.cls, ".rxode2.env") <- .cloneEnv
    class(.v) <- .cls
    return(.v)
  }
}

#' @export
rbind.rxSolve <- function(..., deparse.level = 1) {
  .lst <- list(...)
  if (length(.lst) >= 2) {
    .ret <- try(.rbind2rxSove(.lst[[1]], .lst[[2]]),
                silent=TRUE)
    if (inherits(.ret, "try-error")) {
      return(do.call(rbind.data.frame, .lst))
    }
    if (length(.lst) == 2) {
      return(.ret)
    }
    return(do.call(rbind.rxSolve,
                   c(list(.ret),
                     lapply(seq_along(.lst)[-(1:2)],
                            function(i){
                              .lst[[i]]
                            }))))
  }
  if (length(.lst) == 1) return(.lst[[1]])
  stop("called rbind.rxSolve() with no arguments",
       call.=FALSE)
}

#' A list and description of Rode supported syntax functions
#'
#' @format A data frame with 3 columns and 106 rows
#' \describe{
#' \item{Function}{Reserved function Name}
#' \item{Description}{Description of function}
#' \item{Aliases}{Function Aliases}
#' }
"rxSyntaxFunctions"

.asFunctionEnv <- new.env(parent=emptyenv())
.asFunctionEnv$rx <- NULL
#' @export
as.function.rxUi <- function(x, ...) {
  x$fun
}


#' @export
as.function.rxode2tos <- function(x, ...) {
  .asFunctionEnv$rx <- x
  suppressMessages(x$uiFun)
}

#' @export
as.function.rxode2 <- function(x, ...) {
  eval(str2lang(paste(c("function() {",
    "model({",
    rxNorm(x),
    "})",
    "}"), collapse="\n")))
}

#' @export
as.function.character <- function(x, ...) {
  if (length(x) != 1) {
    x <- paste(x, collapse="\n")
  }
  eval(str2lang(paste(c("function() {",
                        "model({",
                        rxNorm(x),
                        "})",
                        "}"), collapse="\n")))
}


#' @export
as.function.rxModelVars <- as.function.rxode2

#' Require namespace, otherwise throw error.
#'
#' @param pkg Package required for function to work.
#' @return Nothing
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
rxReq <- function(pkg) {
  ## nocov start
  if (!requireNamespace(pkg, quietly = TRUE)) {
    stop(sprintf(gettext("package \"%s\" needed for this function to work"), pkg),
      call. = FALSE
    )
  }
  ## nocov end
}

#' Respect suppress messages
#'
#' This turns on the silent REprintf in C when `suppressMessages()` is
#' turned on. This makes the `REprintf` act like `messages` in R,
#' they can be suppressed with `suppressMessages()`
#'
#' @return Nothing
#' @author Matthew Fidler
#' @export
#' @examples
#'
#' # rxParseSuppressMsg() is called with rxode2()
#'
#' # Note the errors are output to the console
#'
#' try(rxode2parse("d/dt(matt)=/3"), silent = TRUE)
#'
#' # When using suppressMessages, the output is suppressed
#'
#' suppressMessages(try(rxode2parse("d/dt(matt)=/3"), silent = TRUE))
#'
#' # In rxode2, we use REprintf so that interrupted threads do not crash R
#' # if there is a user interrupt. This isn't captured by R's messages, but
#' # This interface allows the `suppressMessages()` to suppress the C printing
#' # as well
#'
#' # If you  want to suppress messages from rxode2 in other packages, you can use
#' # this function
rxParseSuppressMsg <- function() {
  if (requireNamespace("knitr", quietly = TRUE)) {
    if (!is.null(knitr::opts_knit$get("rmarkdown.pandoc.to"))) {
      return(invisible(NULL))
    } else {
      rxParseSetSilentErr(as.integer(length(capture.output(message(" "), type = "message")) == 0L))
    }
  } else {
    rxParseSetSilentErr(as.integer(length(capture.output(message(" "), type = "message")) == 0L))
  }
  invisible(NULL)
}

#' Replace comments with label()
#'
#' @param src source function
#' @return character vector with labels("") replaced in function
#' @author Matthew Fidler
#' @noRd
.rxReplaceCommentWithLabel <- function(src) {
  .env <- new.env(parent=emptyenv())#match.call(expand.dots = TRUE)[-(1:2)]
  .env$inIni <- FALSE
  .env$convertLabel <- FALSE
  .regIni <- rex::rex(boundary, or("ini", "lotri"), "(", any_spaces, "{", any_spaces)
  .regOther1 <- rex::rex(any_spaces, "}", any_spaces, ")", any_spaces)
  .regOther2 <- rex::rex(any_spaces, "model", any_spaces, "(", any_spaces, "{", any_spaces)
  .regCommentOnBlankLine <- "^ *#+ *(.*) *$"
  .regLabel <- "^( *[^\n\"]+) *#+ *(.*) *$"
  .ret <- vapply(src,
                 function(line) {
                   if (regexpr(.regIni, line, perl=TRUE) != -1) {
                     assign("inIni", TRUE, envir=.env)
                   } else if (regexpr(.regOther1, line, perl=TRUE) != -1) {
                     assign("inIni", FALSE, envir=.env)
                   } else if (regexpr(.regOther2, line, perl=TRUE) != -1) {
                     assign("inIni", FALSE, envir=.env)
                   } else if (.env$inIni) {
                     if (regexpr(.regCommentOnBlankLine, line) != -1) {
                     } else if (regexpr(.regLabel, line) != -1) {
                       .env$convertLabel <- TRUE
                       .label <- deparse1(sub(.regLabel, "\\2", line))
                       return(sub(.regLabel, paste0("\\1; label(", .label, ")"), line))
                     }
                   }
                   line
                 }, character(1), USE.NAMES = FALSE)
  if (.env$convertLabel) {
    cli::cli_alert_info("parameter labels from comments will be replaced by 'label()'")
  }
  .ret <- deparse(eval(parse(text=paste(.ret, collapse="\n"), keep.source=FALSE)))
  .ret
}

#' Convert rxode2/nlmixr model function to a string
#'
#' @param fun function name for parsing
#' @return Modified function
#' @author Matthew Fidler
#' @noRd
.rxFunction2string <- function(fun) {
  .srcRef <- attr(fun, "srcref")
  if (is.null(.srcRef)) {
    cli::cli_alert_info("parameter labels from comments are typically ignored in non-interactive mode")
    cli::cli_alert_info("Need to run with the source intact to parse comments")
    .ret <- deparse(fun)
  } else {
    .ret <- .rxReplaceCommentWithLabel(as.character(.srcRef, useSource = TRUE))
  }
  .ret
}
#' Rearrange function
#'
#' @param fun Function to rearrange so that `model` is at the end (if needed)
#' @return A more normalized function
#' @author Matthew L. Fidler
#' @noRd
.rxFunctionRearrange <- function(fun) {
  .lst <- as.list(body(fun)[-1])
  .idx <- seq_along(.lst)
  .w <- which(vapply(.idx, function(x) {
    identical(.lst[[x]][[1]], quote(`ini`))
  }, logical(1), USE.NAMES=TRUE))
  if (length(.w) == 0) {
  } else if (length(.w) != 1) {
    stop("rxode2 model function can only have one 'ini({})' block",
         call.=FALSE)
  }
  if (identical(.lst[[length(.lst)]][[1]], quote(`model`))) {
    return(fun)
  }
  .w <- which(vapply(.idx, function(x) {
    identical(.lst[[x]][[1]], quote(`model`))
  }, logical(1), USE.NAMES=TRUE))
  if (length(.w) != 1) {
    stop("rxode2 model function requires one 'model({})' block",
         call.=FALSE)
  }
  warning("'model({})' is not on the last line of the function, rearranging; function cannot be called directly to produce model object",
          call.=FALSE)
  .fun2 <- function() {
  }
  body(.fun2) <- as.call(lapply(c(-1L, .idx[-.w], .w), function(i) {
    if (i == -1L) return(quote(`{`))
    .lst[[i]]
  }))
  .fun2
}

.rxFunction2ui <- function(fun) {
  .fun <- .rxFunctionRearrange(eval(parse(text=paste(.rxFunction2string(fun), collapse="\n"))))
  .ret <- .fun()
  # Save $model like nlmixr UI used to...
  .ret <- rxUiDecompress(.ret)
  assign("model", fun, envir=.ret)
  rxUiCompress(.ret)
}

.lastIni <- NULL
.lastIniQ <- NULL

#' Ini block for rxode2/nlmixr models
#'
#' The ini block controls initial conditions for 'theta' (fixed effects),
#' 'omega' (random effects), and 'sigma' (residual error) elements of the model.
#'
#' The \code{ini()} function is used in two different ways.  The main way that
#' it is used is to set the initial conditions and associated attributes
#' (described below) in a model.  The other way that it is used is for updating
#' the initial conditions in a model, often using the pipe operator.
#'
#' 'theta' and 'sigma' can be set using either \code{<-} or \code{=} such as
#' \code{tvCL <- 1} or equivalently \code{tvCL = 1}.  'omega' can be set with a
#' \code{~} such as \code{etaCL ~ 0.1}.
#'
#' Parameters can be named or unnamed (though named parameters are preferred).
#' A named parameter is set using the name on the left of the assignment while
#' unnamed parameters are set without an assignment operator.  \code{tvCL <- 1}
#' would set a named parameter of \code{tvCL} to \code{1}.  Unnamed parameters
#' are set using just the value, such as \code{1}.
#'
#' For some estimation methods, lower and upper bounds can be set for 'theta'
#' and 'sigma' values.  To set a lower and/or upper bound, use a vector of
#' values.  The vector is \code{c(lower, estimate, upper)}.  The vector may be
#' given with just the estimate (\code{estimate}), the lower bound and
#' estimate (\code{c(lower, estimate)}), or all three (\code{c(lower, estimate,
#' upper)}).  To set an estimate and upper bound without a lower bound, set the
#' lower bound to \code{-Inf}, \code{c(-Inf, estimate, upper)}.  When an
#' estimation method does not support bounds, the bounds will be ignored with a
#' warning.
#'
#' 'omega' values can be set as a single value or as the values of a
#' lower-triangular matrix.  The values may be set as either a
#' variance-covariance matrix (the default) or as a correlation matrix
#' for the off-diagonals with the standard deviations on the
#' diagonals.  Names may be set on the left side of the \code{~}.  To
#' set a variance-covariance matrix with variance values of 2 and 3
#' and a covariance of -2.5 use \code{~c(2, 2.5, 3)}.  To set the same
#' matrix with names of \code{iivKa} and \code{iivCL}, use \code{iivKa
#' + iivCL~c(2, 2.5, 3)}.  To set a correlation matrix with standard
#' deviations on the diagonal, use \code{cor()} like \code{iivKa +
#' iivCL~cor(2, -0.5, 3)}.  As of rxode2 3.0 you can also use
#' \code{iivKa ~ 2, iivCL ~ c(2.5, 3)} for covariance matrices as
#' well.
#'
#' Values may be fixed (and therefore not estimated) using either the name
#' \code{fixed} at the end of the assignment or by calling \code{fixed()} as a
#' function for the value to fix.  For 'theta' and 'sigma', either the estimate
#' or the full definition (including lower and upper bounds) may be included in
#' the fixed setting.  For example, the following are all effectively equivalent
#' to set a 'theta' or 'sigma' to a fixed value (because the lower and upper
#' bounds are ignored for a fixed value): \code{tvCL <- fixed(1)}, \code{tvCL <-
#' fixed(0, 1)}, \code{tvCL <- fixed(0, 1, 2)}, \code{tvCL <- c(0, fixed(1),
#' 2)}, or \code{tvCL <- c(0, 1, fixed)}.  For 'omega' assignment, the full
#' block or none of the block must be set as \code{fixed}.  Examples of setting
#' an 'omega' value as fixed are: \code{iivKa~fixed(1)}, \code{iivKa +
#' iivCL~fixed(1, 2, 3)}, or \code{iivKa + iivCL~c(1, 2, 3, fixed)}.  Anywhere
#' that \code{fixed} is used, \code{FIX}, \code{FIXED}, or \code{fix} may be
#' used equivalently.
#'
#' For any value, standard mathematical operators or functions may be used to
#' define the value.  For example, \code{log(2)} and \code{24*30} may be used to
#' define a value anywhere that a number can be used (e.g. lower bound,
#' estimate, upper bound, variance, etc.).
#'
#' Values may be labeled using the \code{label()} function after the assignment.
#' Labels are are used to make reporting easier by giving a human-readable
#' description of the parameter, but the labels do not have any effect on
#' estimation.  The typical way to set a label so that the parameter \code{tvCL}
#' has a label of "Typical Value of Clearance (L/hr)" is \code{tvCL <- 1;
#' label("Typical Value of Clearance (L/hr)")}.
#'
#' Off diagonal values of 'omega' can be set to zero using the
#' \code{diag()} to remove all off-diagonals can be removed with
#' `ini(diag())`.  To remove covariances of 'omega' item with `iivKa`,
#' you can use `|> ini(diag(iivKa))`.  Or to remove covariances that
#' contain either `iivKa` or `iivCl` you can use `|> ini(diag(iivKa,
#' iivCl))`.  For finer control you can remove the covariance between
#' two items (like `iivKa` and `iivCl`) by `|> ini(-cov(iivKa, iivCl))
#'
#' \code{rxode2}/\code{nlmixr2} will attempt to determine some
#' back-transformations for the user.  For example, \code{CL <- exp(tvCL)} will
#' detect that \code{tvCL} must be back-transformed by \code{exp()} for easier
#' interpretation.  When you want to control the back-transformation, you can
#' specify the back-transformation using \code{backTransform()} after the
#' assignment.  For example, to set the
#' back-transformation to \code{exp()}, you can use \code{tvCL <- 1;
#' backTransform(exp())}.
#'
#' @param x expression
#' @param ... Other expressions for `ini()` function
#' @param envir the `environment` in which unevaluated model
#'   expressions is to be evaluated.  May also be `NULL`, a list, a
#'   data frame, a pairlist or an integer as specified to `sys.call`.
#' @inheritParams .iniHandleAppend
#' @return ini block
#' @author Matthew Fidler
#' @family Initial conditions
#' @examples
#' # Set the ini() block in a model
#' one.compartment <- function() {
#'   ini({
#'     tka <- log(1.57); label("Ka")
#'     tcl <- log(2.72); label("Cl")
#'     tv <- log(31.5); label("V")
#'     eta.ka ~ 0.6
#'     eta.cl ~ 0.3
#'     eta.v ~ 0.1
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl + eta.cl)
#'     v <- exp(tv + eta.v)
#'     d/dt(depot) = -ka * depot
#'     d/dt(center) = ka * depot - cl / v * center
#'     cp = center / v
#'     cp ~ add(add.sd)
#'   })
#' }
#'
#' # Use piping to update initial conditions
#' one.compartment |> ini(tka <- log(2))
#' one.compartment |> ini(tka <- label("Absorption rate, Ka (1/hr)"))
#' # Move the tka parameter to be just below the tv parameter (affects parameter
#' # summary table, only)
#' one.compartment |> ini(tka <- label("Absorption rate, Ka (1/hr)"), append = "tv")
#' # When programming with rxode2/nlmixr2, it may be easier to pass strings in
#' # to modify the ini
#' one.compartment |> ini("tka <- log(2)")
#' @export
ini <- function(x, ..., envir = parent.frame(), append = NULL) {
  if (is(substitute(x), "{")) {
    .ini <- eval(bquote(lotri::lotri(.(substitute(x)),
                                     cov=TRUE, rcm=TRUE)),
                 envir=envir)
    assignInMyNamespace(".lastIni", .ini)
    assignInMyNamespace(".lastIniQ", bquote(.(substitute(x))))
    return(invisible(.ini))
  }
  UseMethod("ini")
}

#' Model block for rxode2/nlmixr models
#'
#' @param x model expression
#'
#' @param ... Other arguments
#'
#' @param append This is a boolean to determine if the lines are
#'   appended in piping.  The possible values for this is:
#'
#'  - `TRUE` which is when the lines are appended to the model instead of replaced
#'  - `FALSE` when the lines are replaced in the model (default)
#'  - `NA` is when the lines are pre-pended to the model instead of replaced
#'  - `lhs expression`, which will append the lines after the last observed line of the expression `lhs`
#'
#' @param auto This boolean tells if piping automatically selects the
#'   parameters should be characterized as a population parameter,
#'   between subject variability, or a covariate.  When `TRUE` this
#'   automatic selection occurs.  When `FALSE` this automatic
#'   selection is turned off and everything is added as a covariate
#'   (which can be promoted to a parameter with the `ini` statement).
#'   By default this is `TRUE`, but it can be changed by
#'   `options(rxode2.autoVarPiping=FALSE)`.
#'
#'
#' @param cov is a character vector of variables that should be
#'   assumed to be covariates.  This will override automatic promotion
#'   to a population parameter estimate (or an eta)
#'
#' @param envir the `environment` in which unevaluated model
#'   expressions is to be evaluated.  May also be `NULL`, a list, a
#'   data frame, a pairlist or an integer as specified to `sys.call`.
#'
#' @return Model block with ini information included.  `ini` must be
#'   called before `model` block
#'
#' @author Matthew Fidler
#'
#' @export
model <- function(x, ..., append=FALSE, auto=getOption("rxode2.autoVarPiping", TRUE),
                  cov=NULL, envir=parent.frame()) {
  if (is(substitute(x), "{")) {
    .funName <- try(as.character(as.list(with(envir, match.call()))[[1]]), silent=TRUE)
    if (inherits(.funName, "try-error")) {
      .funName <- NULL
    } else if (length(.funName) == 1L && exists(.funName, envir=parent.env(envir))) {
      .udfEnvSet(parent.env(envir))
    }
    .ini <- .lastIni
    .iniQ <- .lastIniQ
    if (is.null(.ini)) {
      .ini <- data.frame(ntheta=integer(0),
                         neta1=numeric(0),
                         neta2=numeric(0),
                         name=character(0),
                         lower=numeric(0),
                         est=numeric(0),
                         upper=numeric(0),
                         fix=logical(0),
                         label=character(0),
                         backTransform=character(0),
                         condition=character(0),
                         err=character(0))
      .iniQ <- NULL
      ## stop("ini({}) block must be called before the model block",
      ##      call.=FALSE)
    }
    assignInMyNamespace(".lastIni", NULL)
    assignInMyNamespace(".lastIniQ", NULL)
    .mod <- .rxMuRef(eval(bquote(.errProcessExpression(quote(.(substitute(x))), .ini))))
    .meta <- new.env(parent=emptyenv())
    if (!identical(envir, globalenv())) {
      for (.i in ls(envir, all.names=TRUE)) {
        if (.i != ".simModelBase") {
          assign(.i, get(.i, envir), .meta)
        }
      }
    }
    .mod$meta <- .meta
    .mod$sticky <- character(0)
    .w <- which(!is.na(.mod$iniDf$err) & !is.na(.mod$iniDf$neta1))
    if (length(.w) > 0) {
      stop("the parameter(s) '", paste(.mod$iniDf$name[.w], collapse="', '"), "' cannot be an error and between subject variability",
           call.=FALSE)
    }
    assign("modelName", .funName, envir=.mod)
    class(.mod) <- "rxUi"
    return(rxUiCompress(.mod))
  }
  on.exit({.varSelect$cov <- NULL})
  UseMethod("model")
}

#' @export
#' @rdname model
model.default <- function(x, ..., append=FALSE, cov=NULL, envir=parent.frame()) {
  stop("rxode2 does not know how to handle this model statement")
}

#' @export
print.rxUi <-function(x, ...) {
  .md <- x$modelDesc
  cat(cli::cli_format_method({
    cli::cli_h1("{.md}")
  }), "\n")
  .theta <- x$theta
  .omega <- x$omega
  if (length(x$iniDf$cond) > 0) {
    cat(cli::cli_format_method({
      cli::cli_h2("Initalization:")
    }), "\n")
  }
  if (length(.theta) > 0) {
    cat(paste0(crayon::bold("Fixed Effects"), " (", crayon::bold$blue("$theta"), "):"), "\n")
    print(.theta)
  }
  if (!is.null(dim(.omega))) {
    if (dim(.omega)[1] > 0) {
      cat(paste0("\n", crayon::bold("Omega"), " (", crayon::bold$blue("$omega"), "):"), "\n")
      print(.omega)
    }
  }
  if (length(x$state) > 0) {
    cat(paste0(crayon::bold("\nStates"), " (", crayon::bold$blue("$state"), " or ", crayon::bold$blue("$stateDf"), "):"), "\n")
    print(rxUiGet.stateDf(list(x,TRUE)))
  }
  # Multiple Endpoint
  .me <- x$multipleEndpoint
  if (!is.null(.me)) {
    .met <- crayon::bold("Multiple Endpoint Model")
    .med <- crayon::bold$blue("$multipleEndpoint")
    cat(cli::cli_format_method({
      cli::cli_h2("{.met} ({.med}):")
    }), "\n")
    print(.me)
    if (getOption("rxode2.combine.dvid", TRUE)) {
      cat("  * If dvids are outside this range, all dvids are re-numered sequentially, ie 1,7, 10 becomes 1,2,3 etc\n")
    }
    cat("\n")
  }

  # muRefTable
  .mu <- x$muRefTable
  if (!is.null(.mu)) {
    .muU <- crayon::bold(paste0(ifelse(use.utf(), "\u03bc", "mu"), "-referencing"))
    .muR <- crayon::bold$blue("$muRefTable")
    cat(cli::cli_format_method({
      cli::cli_h2("{.muU} ({.muR}):")
    }), "\n")
    print(.mu)
    cat("\n")
  }
  cat(cli::cli_format_method({
    cli::cli_h2("Model (Normalized Syntax):")
  }))
  cat("\nfunction() ")
  print(as.call(x$funPrint))
  return(invisible(x))
}

#' Compress/Decompress `rxode2` ui
#'
#'
#' @param ui rxode2 ui object
#' @return A compressed or decompressed rxui object
#' @author Matthew L. Fidler
#' @export
#' @examples
#' one.cmt <- function() {
#'   ini({
#'     ## You may label each parameter with a comment
#'     tka <- 0.45 # Log Ka
#'     tcl <- log(c(0, 2.7, 100)) # Log Cl
#'     ## This works with interactive models
#'     ## You may also label the preceding line with label("label text")
#'     tv <- 3.45; label("log V")
#'     ## the label("Label name") works with all models
#'     eta.ka ~ 0.6
#'     eta.cl ~ 0.3
#'     eta.v ~ 0.1
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl + eta.cl)
#'     v <- exp(tv + eta.v)
#'     linCmt() ~ add(add.sd) | tmp
#'   })
#' }
#'
#' f <- rxode2(one.cmt)
#' print(class(f))
#' print(is.environment(f))
#'
#' f  <- rxUiDecompress(f)
#' print(class(f))
#' print(is.environment(f))
#'
#' f  <- rxUiCompress(f)
#' print(class(f))
#' print(is.environment(f))
#'
rxUiDecompress <- function(ui) {
  if (!inherits(ui, "rxUi")) return(ui)
  if (is.environment(ui))  return(ui)
  if (inherits(ui, "raw")) {
    rxReq("qs")
    warning("decompression of an rxUi object from rxode2 < 4.0 requires qs which is not on CRAN",
            call.=FALSE)
    .ret <- .Call(`_rxode2_qsDes`, ui)
  } else if (is.list(ui)) {
    .ret <- list2env(ui, parent=emptyenv())
  }
  class(.ret) <- "rxUi"
  .ret
}

#' @rdname rxUiDecompress
#' @export
rxUiCompress <- function(ui) {
  if (!inherits(ui, "rxUi")) return(ui)
  if (is.environment(ui)) {
    .ls <- ls(ui, all.names=TRUE)
    .ret <- lapply(.ls, function(nm) {
      get(nm, ui)
    })
    names(.ret) <- .ls
    class(.ret) <- c("rxUi", "list")
    return(.ret)
  }
  ui
}

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("label replacement", {
      .pheno <- c("function() {", "  # Pheno with covariance", "  ini({", "    tcl <- log(0.008) # typical value of clearance", "    tv <-  log(0.6)   # typical value of volume", "    ## var(eta.cl)", "    eta.cl + eta.v ~ c(1,", "                       0.01, 1) ## cov(eta.cl, eta.v), var(eta.v)", "    # interindividual variability on clearance and volume", "    add.err <- 0.1    # residual variability", "  })", "  model({", "    cl <- exp(tcl + eta.cl) # individual value of clearance", "    v <- exp(tv + eta.v)    # individual value of volume",  "    ke <- cl / v            # elimination rate constant", "    d/dt(A1) = - ke * A1    # model differential equation", "    cp = A1 / v             # concentration in plasma", "    cp ~ add(add.err)       # define error model", "  })", "}")
      suppressMessages(
        expect_error(.rxReplaceCommentWithLabel(.pheno), NA)
      )
    })
  }
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("limCmt raises error", {

      M2 <- function() {
        ini({
          tD1 <- log(0.8)
          tcl <- log(51.2)
          tv <- log(344)
          eta.D1 ~ 0.4
          eta.cl ~ 0.4
          eta.v ~ 0.4
          add.err <- 1
          prop.err <- 0.2
        })
        model({
          D1 <- exp(tD1 + eta.D1)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          limCmt() ~ add(add.err) + prop(prop.err)
        })
      }

      expect_error(rxode2(M2), "linCmt")
    })
  }
})

rxTest({
  test_that("rxode2 to ui promotion", {

    mod1 <- rxode2({
      C2 <- centr/V2
      C3 <- peri/V3
      d/dt(depot) <- -KA*depot
      d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
      d/dt(peri)  <- Q*C2 - Q*C3
      d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
    })

    suppressMessages(
      expect_error(assertRxUi(mod1), NA)
    )
    suppressMessages(
      mod <- assertRxUi(mod1)
    )
    expect_true(inherits(mod, "rxUi"))

    suppressMessages(
      expect_error(mod1 |> rxRename(C4=C2), NA)
    )
    suppressMessages(
      mod <- mod1 |> rxRename(C4=C2)
    )
    expect_true(inherits(mod, "rxUi"))

    suppressMessages(
      mod <- mod1 |> ini(V2=1)
    )
    expect_true(inherits(mod, "rxUi"))

    suppressMessages(
      mod <- mod1 |> model(C2 <- centr)
    )
    expect_true(inherits(mod, "rxUi"))
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("ui props", {

      fun_odes  <- function() {
        description <- "Dupilumab PK model (Kovalenko 2020)"
        reference <- "Kovalenko P, Davis JD, Li M, et al. Base and Covariate Population Pharmacokinetic Analyses of Dupilumab Using Phase 3 Data. Clinical Pharmacology in Drug Development. 2020;9(6):756-767. doi:10.1002/cpdd.780"
        # Model 1 from table 1 and supplementary Table 2 in the publication and its
        # supplement.
        covariateData <-
          list(
            WT = "Body weight in kg"
          )
        ini({
          lvc <- log(2.48); label("central volume (L)")
          lke <- log(0.0534); label("elimination rate (1/d)")
          lkcp <- log(0.213); label("central-to-peripheral rate (1/d)")
          Mpc <- 0.686; label("ratio of kcp and kpc (kpc is peripheral to central rate with units of 1/d)")
          lka <- log(0.256); label("absorption rate (1/d)")
          lMTT <- log(0.105); label("mean transit time (d)")
          lVm <- log(1.07); label("maximum target-mediated rate of elimination (mg/L/d)")
          Km <- fixed(0.01); label("Michaelis-Menten constant (mg/L)")
          lfdepot <- log(0.643); label("Bioavailability (fraction)")
          e_wt_vc <- 0.711; label("Exponent of weight on central volume (unitless)")

          etalvc ~ 0.192
          etalke ~ 0.285
          etalka ~ 0.474
          etalvm ~ 0.236
          etamtt ~ 0.525 # etamtt is assumed to be on log-scale MTT to prevent negative values; this is a difference relative to Supplementary Table 2

          cppropSd <- 0.15; label("Proportional residual error (fraction)")
          cpaddSd <- fixed(0.03); label("Additive residual error (mg/L)")
        })
        model({
          # Weight normalization to 75 kg is assumed based on prior publication.  It
          # is not specified in the current publication:
          # Kovalenko P, DiCioccio AT, Davis JD, et al. Exploratory Population PK
          # Analysis of Dupilumab, a Fully Human Monoclonal Antibody Against
          # IL-4Ralpha, in Atopic Dermatitis Patients and Normal Volunteers. CPT
          # Pharmacometrics Syst Pharmacol. 2016;5(11):617-624. doi:10.1002/psp4.12136
          vc <- exp(lvc + etalvc)*(WT/75)^e_wt_vc
          ke <- exp(lke + etalke)
          kcp <- exp(lkcp)
          ka <- exp(lka + etalka)
          MTT <- exp(lMTT + etamtt)
          Vm <- exp(lVm + etalvm)

          # Derived parameters
          kpc <- kcp/Mpc
          ktr <- (3 + 1)/MTT

          d/dt(depot) <- -ktr*depot
          d/dt(transit1) <- ktr*(depot - transit1)
          d/dt(transit2) <- ktr*(transit1 - transit2)
          d/dt(transit3) <- ktr*transit2 - ka*transit3
          # Linear and Michaelis-Menten clearance
          d/dt(central) <-                 ka*transit3 - ke*central - kcp*central + kpc*periph - central*(Vm/(Km + central/vc))
          d/dt(periph) <-                                             kcp*central - kpc*periph

          f(depot) <- exp(lfdepot)
          # No unit conversion is required to change mg/L (dosing amount/central
          # volume unit) to mg/L (measurement unit)
          Cc <- central/vc
          Cc ~ add(cpaddSd) + prop(cppropSd)
        })
      }

      fun_ana <- function() {
        description <- "Two compartment PK model with linear clearance for average monoclonal antibodies (Davda 2014)"
        reference <- "Davda JP, Dodds MG, Gibbs MA, Wisdom W, Gibbs JP. A model-based meta-analysis of monoclonal antibody pharmacokinetics to guide optimal first-in-human study design. MAbs. 2014;6(4):1094-1102. doi:10.4161/mabs.29095"
        ini({
          lfdepot <- log(0.744) ; label("Subcutaneous bioavailability (fraction)")
          lka <- log(0.282) ; label("Absorption rate (Ka, 1/day)")
          lcl <- log(0.200) ; label("Clearance (CL, L/day)")
          lv  <- log(3.61) ; label("Central volume of distribution (V, L)")
          lvp  <- log(2.75) ; label("Peripheral volume of distribution (Vp, L)")
          lq  <- log(0.747) ; label("Intercompartmental clearance (Q, L/day)")

          allocl <- 0.865 ; label("Allometric exponent on clearance and intercompartmental clearance (unitless)")
          allov <- 0.957 ; label("Allometric exponent on volumes of distribution (unitless)")

          etafdepot ~ 0
          etaka ~ 0.416
          etacl + etav + etavp ~ c(0.0987,
                                   0.0786, 0.116,
                                   0.0377, 0.0619, 0.0789)
          etaq ~ 0.699

          prop.err <- 0.144 ; label("Proportional residual error (fraction)")
        })
        model({
          # WT is body weight in kg
          fdepot <- exp(lfdepot + etafdepot)
          ka <- exp(lka + etaka)
          wtnorm <- log(WT/70)
          cl <- exp(lcl + allocl*wtnorm + etacl)
          q  <- exp(lq + allocl*wtnorm + etaq)
          v <- exp(lv + allov*wtnorm + etav)
          vp <- exp(lvp + allov*wtnorm + etavp)
          Cc <- linCmt()

          f(depot) <- fdepot  # Units are dosing units/L (typically mg/L = ug/mL)
          Cc ~ prop(prop.err)

        })
      }

      obj_ana = rxode2(fun_ana)
      obj_odes = rxode2(fun_odes)

      expect_equal(obj_ana$props,
                   list(pop = c("lfdepot", "lka", "lcl", "lv", "lvp", "lq", "allocl", "allov"),
                        resid = "prop.err",
                        group = list(id = c("etafdepot", "etaka", "etacl", "etav", "etavp", "etaq")),
                        linCmt = TRUE,
                        cmt = c("depot", "central"),
                        output = list(primary = c("fdepot", "ka", "wtnorm", "cl", "q", "v", "vp"),
                                      secondary = character(0),
                                      endpoint = "Cc",
                                      state = c("depot", "central")),
                        cmtProp=data.frame(Compartment = "depot", Property = "f")))

      expect_equal(obj_odes$props,
                   list(pop = c("lvc", "lke", "lkcp", "Mpc", "lka", "lMTT", "lVm", "Km", "lfdepot", "e_wt_vc"),
                        resid = c("cppropSd", "cpaddSd"),
                        group = list(id = c("etalvc", "etalke", "etalka", "etalvm", "etamtt")),
                        linCmt = FALSE,
                        cmt = c("depot", "transit1", "transit2", "transit3", "central", "periph"),
                        output = list(primary = c("vc", "ke", "kcp", "ka", "MTT", "Vm", "kpc"),
                                      secondary = "ktr",
                                      endpoint = "Cc",
                                      state = c("depot", "transit1", "transit2", "transit3", "central", "periph")),
                        cmtProp=data.frame(Compartment="depot", Property="f")))


      fun_ana2 <- function() {
        description <- "Two compartment PK model with linear clearance for average monoclonal antibodies (Davda 2014)"
        reference <- "Davda JP, Dodds MG, Gibbs MA, Wisdom W, Gibbs JP. A model-based meta-analysis of monoclonal antibody pharmacokinetics to guide optimal first-in-human study design. MAbs. 2014;6(4):1094-1102. doi:10.4161/mabs.29095"
        ini({
          lfdepot <- log(0.744) ; label("Subcutaneous bioavailability (fraction)")
          lka <- log(0.282) ; label("Absorption rate (Ka, 1/day)")
          lcl <- log(0.200) ; label("Clearance (CL, L/day)")
          lv  <- log(3.61) ; label("Central volume of distribution (V, L)")
          lvp  <- log(2.75) ; label("Peripheral volume of distribution (Vp, L)")
          lq  <- log(0.747) ; label("Intercompartmental clearance (Q, L/day)")

          allocl <- 0.865 ; label("Allometric exponent on clearance and intercompartmental clearance (unitless)")
          allov <- 0.957 ; label("Allometric exponent on volumes of distribution (unitless)")

          etafdepot ~ 0
          etaka ~ 0.416
          etacl + etav + etavp ~ c(0.0987,
                                   0.0786, 0.116,
                                   0.0377, 0.0619, 0.0789)
          etaq ~ 0.699

          prop.err <- 0.144 ; label("Proportional residual error (fraction)")
        })
        model({
          # WT is body weight in kg
          fdepot <- exp(lfdepot + etafdepot)
          ka <- exp(lka + etaka)
          wtnorm <- log(WT/70)
          cl <- exp(lcl + allocl*wtnorm + etacl)
          q  <- exp(lq + allocl*wtnorm + etaq)
          v <- exp(lv + allov*wtnorm + etav)
          vp <- exp(lvp + allov*wtnorm + etavp)
          linCmt() ~ prop(prop.err)
          f(depot) <- fdepot  # Units are dosing units/L (typically mg/L = ug/mL)

        })
      }

      tmp <- fun_ana2()

      expect_equal(tmp$props,
                   list(pop = c("lfdepot", "lka", "lcl", "lv", "lvp", "lq", "allocl", "allov"),
                        resid = "prop.err",
                        group = list(id = c("etafdepot", "etaka", "etacl", "etav", "etavp", "etaq")),
                        linCmt = TRUE,
                        cmt = c("depot", "central"),
                        output = list(primary = c("fdepot", "ka", "wtnorm", "cl", "q", "v", "vp"),
                                      secondary = character(0),
                                      endpoint = character(0),
                                      state = c("depot", "central")),
                        cmtProp=data.frame(Compartment = "depot", Property = "f")))

    })

    test_that("linCmt single compartment parses correctly", {

      mod <- function() {
        ini({
          cl <- 0.1
          vc <- 4
          err.sd <- 0.1
        })
        model({
          Cc <- linCmt(cl, vc)
          Cc ~ add(err.sd)
        })
      }

      mod <- mod()

      expect_error(mod$props, NA)

      expect_equal(mod$props,
                   list(pop = c("cl", "vc"),
                        resid = "err.sd",
                        group = structure(list(), names = character(0)),
                        linCmt = TRUE,
                        cmt = "central",
                        output = list(primary = character(0),
                                      secondary = character(0),
                                      endpoint = "Cc",
                                      state = "central"),
                        cmtProp=NULL))

    })

    test_that("state based endpoint", {

      oncology_sdm_lobo_2002 <- function() {
        description <- "Signal transduction model for delayed concentration effects on cancer cell growth"
        reference <- "Lobo ED, Balthasar JP. Pharmacodynamic modeling of chemotherapeutic effects: Application of a transit compartment model to characterize methotrexate effects in vitro. AAPS J. 2002;4(4):212-222. doi:10.1208/ps040442"
        depends<-"Cc"
        units<-list(time="hr")
        # Values for lkng, ltau, lec50, and kmax are for methotrexate from Lobo 2002,
        # Table 2.  propErr and addErr are added as reasonable values though not from
        # Lobo 2002 where no value is apparent in the paper.
        ini({
          lkng <- log(0.02) ; label("Cell net growth rate (growth minus death) (1/hr)")
          ltau <- log(34.1) ; label("Mean transit time of each transit compartment (hr)")
          lec50 <- log(0.1) ; label("Drug concentration reducing the cell growth by 50% (ug/mL)")
          kmax <- 0.29 ; label("Maximum drug-related reduction in cell growth (1/hr)")
          tumorVolpropSd <- c(0, 0.3) ; label("Proportional residual error (fraction)")
          tumorVoladdSd <- c(0, 50, 1000) ; label("Additive residual error (tumor volume units)")
        })
        model({
          # Cc is the drug concentration
          kng <- exp(lkng)
          tau <- exp(ltau)
          ec50 <- exp(lec50)
          drugEffectTumorVol <- kmax*Cc/(ec50 + Cc)
          tumorVol(0) <- tumorVol0
          d/dt(tumorVol) <- kng*tumorVol - transit4*tumorVol
          d/dt(transit1) <- (drugEffectTumorVol - transit1)/tau
          d/dt(transit2) <- (transit1 - transit2)/tau
          d/dt(transit3) <- (transit2 - transit3)/tau
          d/dt(transit4) <- (transit3 - transit4)/tau
          tumorVol ~ prop(tumorVolpropSd) + add(tumorVoladdSd)
        })
      }

      rx_obj = rxode2::rxode2(oncology_sdm_lobo_2002)

      expect_equal(rx_obj$props$output$endpoint, "tumorVol")
    })

    test_that("bad ui props", {

      f <- function() {
        dosing <- c("central", "depot")
        ini({
          lcl <- 1
          label("Clearance (CL)")
          lvc <- 3.45
          label("Central volume of distribution (V)")
          propSd <- c(0, 0.5)
          label("Proportional residual error (fraction)")
          lwa <- 0.1
          label("Weibull absorption alpha (wa)")
          lwb <- 0.1
          label("Weibull absorption beta (wa)")
          uEk <- 0.1
          label("untransformed slope (Ek)")
          effectSd <- c(0, 0.1)
          label("additive error for effect")
        })
        model({
          Ek <- uEk
          wa <- exp(lwa)
          wb <- exp(lwb)
          cl <- exp(lcl)
          vc <- exp(lvc)
          kel <- cl/vc
          d/dt(depot) <- (wb/wa) * (tad0(depot)/wa)^(wb - 1) *
            depot
          d/dt(central) <- (wb/wa) * (tad0(depot)/wa)^(wb - 1) *
            depot - kel * central
          Cc <- central/vc
          Cc ~ prop(propSd)
          effect <- Ek * log(Cc)
          effect ~ add(effectSd)
        })
      }

      f <- f()

      expect_error(f$props, NA)
    })
  }
})

rxTest({

  ## devtools::load_all()

  skip_if_not_installed("nlmixr2data")
  d <- nlmixr2data::nmtest
  # internally rxode2 treats lag time evids differently than
  # non-lagged events

  # 2 different models to test then
  f <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    d/dt(depot) <- -ka*depot
    d/dt(central) <- ka*depot - (cl/v)*central
    f(central) <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central) <- dur2
    cp <- central/(v/1000)
  })

  lf <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    f(central) <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central) <- dur2
    cp <- linCmt()*1000
  })

  lbf <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    f(central) <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central) <- dur2
    cp <- linCmt()*1000
  }, linCmtSens = "linCmtB")


  elf <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    f(central) <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central) <- dur2
    cp <- linCmt()*1000
    ke0 <- log(2)/(50)
    d/dt(Ce) <- (Cp-Ce)*ke0
  })

  elbf <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    f(central) <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central) <- dur2
    cp <- linCmt()*1000
    ke0 <- log(2)/(50)
    d/dt(Ce) <- (Cp-Ce)*ke0
  }, linCmtSens = "linCmtB")

  fl <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    d/dt(depot) <- -ka*depot
    d/dt(central) <- ka*depot - (cl/v)*central
    lag(central) <- lagt
    f(central) <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central) <- dur2
    cp <- central/(v/1000)
  })

  lfl <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    lag(central) <- lagt
    f(central) <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central) <- dur2
    cp <- linCmt()*1000
  })

  lbfl <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    lag(central) <- lagt
    f(central) <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central) <- dur2
    cp <- linCmt()*1000
  }, linCmtSens = "linCmtB")

  elfl <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    lag(central) <- lagt
    f(central) <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central) <- dur2
    cp <- linCmt()*1000
    ke0 <- log(2)/(50)
    d/dt(Ce) <- (Cp-Ce)*ke0
  })

  elbfl <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    lag(central) <- lagt
    f(central) <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central) <- dur2
    cp <- linCmt()*1000
    ke0 <- log(2)/(50)
    d/dt(Ce) <- (Cp-Ce)*ke0
  }, linCmtSens = "linCmtB")


  library(ggplot2)

  solveEqual <- function(id, plot = p, meth="A",
                         modifyData = c("none", "dur", "rate"),
                         addlKeepsCov = TRUE, addlDropSs=TRUE,
                         ss2cancelAllPending=FALSE) {
    # Preserve the caller's method name for the test titles.  `meth` is
    # rewritten below (e.g. the dense composite "ddop853+dros4" becomes
    # "dop853+ros4" with a separate `dense` flag), which would otherwise
    # mislabel dense-composite failures as the non-dense method.
    .methLabel <- meth
    if (meth == "Bs" || meth == "As") {
      ssSolved <- FALSE
    } else {
      ssSolved <- TRUE
    }
    dense <- FALSE
    if (meth == "A" || meth == "Ao" || meth == "As") {
      lin <- "A"
      meth <- "liblsoda"
    } else if (meth == "B" || meth == "Bo" || meth == "Bs") {
      lin <- "B"
      meth <- "liblsoda"
    } else  if (meth == "ddop853") {
      lin <- "ode"
      meth <- "dop853"
      dense <- TRUE
    } else  if (meth == "dcvode") {
      lin <- "ode"
      meth <- "cvode"
      dense <- TRUE
    } else  if (meth == "ddop5") {
      lin <- "ode"
      meth <- "dop5"
      dense <- TRUE
    } else  if (meth == "dbs") {
      lin <- "ode"
      meth <- "bs"
      dense <- TRUE
    } else  if (meth == "dros4") {
      lin <- "ode"
      meth <- "ros4"
      dense <- TRUE
    } else if (meth == "ddop853+dros4") {
      lin <- "ode"
      meth <- "dop853+ros4"
      dense <- TRUE
    } else  if (meth == "Ad") {
      lin <- "A"
      meth <- "dop853"
    } else  if (meth == "Bd") {
      lin <- "B"
      meth <- "dop853"
    } else  if (meth == "Al") {
      lin <- "A"
      meth <- "lsoda"
    } else  if (meth == "Bl") {
      lin <- "B"
      meth <- "lsoda"
    } else if (meth == "B" || meth == "Bo" || meth == "Bs") {
      lin <- "B"
      meth <- "liblsoda"
    } else {
      lin <- "ode"
    }

    noLag <-  all(d[d$id == id & d$evid != 0,]$lagt == 0)
    hasRate <- any(d[d$id == id & d$evid != 0,]$rate != 0)
    hasModeledRate <- any(d[d$id == id & d$evid != 0,]$mode == 1)
    hasModeledDur  <- any(d[d$id == id & d$evid != 0, ]$mode == 2)
    hasChangedF <- any(d[d$id == id & d$evid != 0, ]$bioav != 1)
    modifyData <- match.arg(modifyData)
    d <- d[d$id == id,]
    rate <- unlist(as.vector(d[d$evid != 0, "rate"]))
    ii0 <- all(d$ii == 0)
    oneRate <- (length(rate) == 1L)
    dose1 <- all(d[d$evid != 0, ]$cmt == 1)
    if (!addlDropSs) {
      if (any(d$ss == 2)) {
        return()
      }
    }
    if (modifyData == "rate" && hasRate && !hasModeledRate && !hasModeledDur && oneRate && !ii0 && !dose1) {
      if (p) message("modified rate to be modeled")
      rate <- d[d$evid != 0, "rate", drop=FALSE]
      rate <- rate$rate
      if (length(rate) == 1) {
        d$rat2 <- rate
        d$rate <- ifelse(d$rate==0, 0, -1)
        d$mode <- 1
      }
      ## print(etTrans(d, f, addlDropSs=addlDropSs))
    } else if (modifyData == "dur" && hasRate && !hasModeledRate && !hasModeledDur && !hasChangedF && oneRate && !ii0 && !dose1) {
      if (p) message("modified dur to be modeled")
      rate <- as.numeric(d[d$evid != 0, "rate"])
      amt <- as.numeric(d[d$evid != 0, "amt"])
      if (length(rate) == 1) {
        d$dur2 <- amt/rate
        d$rate <- ifelse(d$rate==0, 0, -2)
        d$mode <- 2
        assign(".d", d, envir=globalenv())
      }
    } else if (any(modifyData == c("dur", "rate"))) {
      if (p) {
        message("skipping because cannot be modified")
        print(list(modifyData=modifyData,
                   noLag=noLag,
                   hasRate=hasRate,
                   hasModeledRate=hasModeledRate,
                   hasModeledDur= hasModeledDur,
                   hasChangedF=hasChangedF))
      }
      return(invisible())
    }
    if (plot) {
      print(ggplot(d, aes(time, cp)) +
              geom_point(col="red") +
              rxode2::rxTheme() +
              ggtitle(paste0("id=", id)))
      ## print(etTrans(d, fl))
      if (lin == "A") {
        s1 <- rxSolve(lfl, d, method=meth, addlKeepsCov = addlKeepsCov,
                      addlDropSs=addlDropSs,
                      ss2cancelAllPending=ss2cancelAllPending,
                      ssSolved=ssSolved, dense=dense)
      } else if (lin == "B") {
        s1 <- rxSolve(lbfl, d, method=meth, addlKeepsCov = addlKeepsCov,
                      addlDropSs=addlDropSs,
                      ss2cancelAllPending=ss2cancelAllPending,
                      ssSolved=ssSolved, dense=dense)
      } else if (lin == "Ao") {
        s1 <- rxSolve(elfl, d, method=meth, addlKeepsCov = addlKeepsCov,
                      addlDropSs=addlDropSs,
                      ss2cancelAllPending=ss2cancelAllPending,
                      ssSolved=ssSolved, dense=dense)
      } else if (lin == "Bo") {
        s1 <- rxSolve(elbfl, d, method=meth, addlKeepsCov = addlKeepsCov,
                      addlDropSs=addlDropSs,
                      ss2cancelAllPending=ss2cancelAllPending,
                      ssSolved=ssSolved, dense=dense)
      } else {
        s1 <- rxSolve(fl, d, method=meth, addlKeepsCov = addlKeepsCov,
                      addlDropSs=addlDropSs,
                      ss2cancelAllPending=ss2cancelAllPending,
                      ssSolved=ssSolved, dense=dense)
      }
      if (!noLag) {
        print(plot(s1, cp) +
                geom_point(data=d, aes(x=time, y=cp), col="red") +
                ggtitle(paste0("id=", id, "(lag)")))
      } else {
        message("================================================== ")
        message("f without lag")
        message("================================================== ")
        if (lin == "A") {
          s2 <- rxSolve(lf, d, method=meth, addlKeepsCov = addlKeepsCov,
                        addlDropSs=addlDropSs,
                        ss2cancelAllPending=ss2cancelAllPending,
                        ssSolved=ssSolved, dense=dense)
        } else if (lin == "B") {
          s2 <- rxSolve(lbf, d, method=meth, addlKeepsCov = addlKeepsCov,
                        addlDropSs=addlDropSs,
                        ss2cancelAllPending=ss2cancelAllPending,
                        ssSolved=ssSolved, dense=dense)
        } else if (lin == "Ao") {
          s2 <- rxSolve(elf, d, method=meth, addlKeepsCov = addlKeepsCov,
                        addlDropSs=addlDropSs,
                        ss2cancelAllPending=ss2cancelAllPending,
                        ssSolved=ssSolved, dense=dense)
        } else if (lin == "Bo") {
          s2 <- rxSolve(elbf, d, method=meth, addlKeepsCov = addlKeepsCov,
                        addlDropSs=addlDropSs,
                        ss2cancelAllPending=ss2cancelAllPending,
                        ssSolved=ssSolved, dense=dense)
        } else {
          s2 <- rxSolve(f, d, method=meth, addlKeepsCov = addlKeepsCov,
                        addlDropSs=addlDropSs,
                        ss2cancelAllPending=ss2cancelAllPending,
                        ssSolved=ssSolved, dense=dense)
        }
        return(plot(s1, cp) +
                 geom_point(data=d, aes(x=time, y=cp), col="red") +
                 geom_line(data=s2, aes(x=time, y=cp), col="blue", alpha=0.5, linewidth=2) +
                 ggtitle(paste0("id=", id, "(nolag)")))
      }
      ## print(etTrans(d, fl))
    } else {
      sub <- 0
      if (id %in% c(410, 411, 409, 415, 709, 510, 610)) {
        sub <- 24
      }
      if (id %in% c(809, 909, 1009)) {
        sub <- 48
      }
      if (id %in% 825) {
        sub <- 96
      }
      if (noLag) {
        test_that(paste0("nmtest id:", id, " no alag; method: ", .methLabel, "; modifyData:", modifyData, "; addlDropSs: ", addlDropSs, "; lin=", lin, "; ssSolved=", ssSolved),
        {
          if (lin == "A") {
            s1 <- rxSolve(lf, d, method=meth,
                          addlKeepsCov = addlKeepsCov,
                          addlDropSs=addlDropSs, dense=dense)
          } else if (lin == "B") {
            s1 <- rxSolve(lbf, d, method=meth,
                          addlKeepsCov = addlKeepsCov,
                          addlDropSs=addlDropSs, dense=dense)
          } else {
            s1 <- rxSolve(f, d, method=meth,
                          addlKeepsCov = addlKeepsCov,
                          addlDropSs=addlDropSs, dense=dense)
          }
          expect_equal(s1$cp[s1$time >= sub],
                       d[d$id == id & d$evid == 0 & d$time >= sub,]$cp,
                       tolerance = 0.1)
        })
      }
      test_that(paste0("nmtest id:", id, " alag; method: ", .methLabel, "; modifyData:", modifyData,"; addlDropSs: ", addlDropSs, "; lin=", lin, "; ssSolved=", ssSolved),
      {
        if (lin == "A") {
          s1 <- rxSolve(lfl, d, method=meth,
                        addlKeepsCov = addlKeepsCov,
                        addlDropSs=addlDropSs, dense=dense)
        } else if (lin == "B") {
          s1 <- rxSolve(lfl, d, method=meth,
                        addlKeepsCov = addlKeepsCov,
                        addlDropSs=addlDropSs, dense=dense)
        } else if (lin == "Ao") {
          s1 <- rxSolve(elfl, d, method=meth,
                        addlKeepsCov = addlKeepsCov,
                        addlDropSs=addlDropSs, dense=dense)
        } else if (lin == "Bo") {
          s1 <- rxSolve(elfl, d, method=meth,
                        addlKeepsCov = addlKeepsCov,
                        addlDropSs=addlDropSs, dense=dense)
        } else {
          s1 <- rxSolve(fl, d, method=meth,
                        addlKeepsCov = addlKeepsCov,
                        addlDropSs=addlDropSs, dense=dense)
        }
        expect_equal(s1$cp[s1$time >= sub],
                     d[d$id == id & d$evid == 0 & d$time >= sub,]$cp,
                     tolerance = 0.1)
      })
    }
  }

  p <- TRUE

  id <- unique(d$id)

  # id = 12


  p <- FALSE

  meths <- c(.methods0, .methods1, .methods2)
  modDat <- c("none", "rate", "dur")
  # One method per test: pair each method with a round-robin id.  Running the
  # full methods x ids cross-product is prohibitively slow for the ~60 adjoint
  # solver variants, so instead every method is exercised on one dosing
  # scenario (id) and -- because there are more methods than ids -- every id is
  # still visited at least once (dosing-scenario coverage is preserved while
  # method coverage becomes one-per-test).
  lapply(seq_along(meths), function(.mi) {
    meth <- meths[.mi]
    i <- id[((.mi - 1L) %% length(id)) + 1L]
    for (modifyData in modDat) {
      for (addlDropSs in c(TRUE, FALSE)) {
        solveEqual(i, meth=meth, modifyData=modifyData, addlDropSs=addlDropSs)
      }
    }
  })

  ## invisible(lapply(unique(d$id), function(i){message(i);solveEqual(i);Sys.sleep(1)}))

  ## need to check id=25 type for infusions too (static, modeled) need
  ## to check for steady state when the infusion is still going at the
  ## time of steady-state release

  ## Need to check steady state infusion as well as infusions where ii=dur with lag times

  ## modeled equivalents of 425, 525

})

rxTest({
  test_that("evid4", {

    df <- readRDS(test_path("nmtest-evid4.rds"))


    evid4 <- function() {
      ini({
        # Where initial conditions/variables are specified
        lka  <- log(2.097)   #log ka (1/h)
        lcl  <- log(2.568724)#log Cl (L/h)
        lv   <- log(69.1925) #log V (L)
        allocl <- fix(0.75)  #WT on Cl
        allov  <- fix(1.00)  #WT on V
      })
      model({
        # Where the model is specified
        cl <- exp(lcl + allocl*(log(WT/70)))
        v  <- exp(lv + allov*(log(WT/70)))
        ka <- exp(lka)
        lin <- linCmt()
      })
    }

    evid4ode <- function() {
      ini({
        # Where initial conditions/variables are specified
        lka  <- log(2.097)   #log ka (1/h)
        lcl  <- log(2.568724)#log Cl (L/h)
        lv   <- log(69.1925) #log V (L)
        allocl <- fix(0.75)  #WT on Cl
        allov  <- fix(1.00)  #WT on V
      })
      model({
        # Where the model is specified
        cl <- exp(lcl + allocl*(log(WT/70)))
        v  <- exp(lv + allov*(log(WT/70)))
        ka <- exp(lka)
        kel <- cl/v
        d/dt(depot)   <- -ka*depot
        d/dt(central) <- ka*depot - kel*central
        lin <- central/v
      })
    }


    f1 <- rxSolve(evid4, df[1:5,], addDosing=TRUE, returnType="data.frame")

    f1o <- rxSolve(evid4ode, df[1:5,], addDosing=TRUE, returnType="data.frame")

    f2 <- rxSolve(evid4, df, addDosing=TRUE, returnType="data.frame")

    f2o <- rxSolve(evid4ode, df, addDosing=TRUE, returnType="data.frame")

    expect_equal(f1$lin, f2$lin[seq_along(f1$lin)])

    expect_equal(f1o$lin, f2o$lin[seq_along(f1o$lin)])

    expect_equal(f1$lin, f1o$lin, tolerance = 1e-6)

    expect_equal(f2$lin, f2o$lin, tolerance = 1e-6)

  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("as.ini ini expression", {

      is.ini <- function(x) {
        expect_true(is.call(x))
        expect_true(identical(x[[1]], quote(`ini`)))
      }

      ini <- quote(ini({
        tka <- log(1.57)
        tcl <- log(2.72)
        tv <- log(31.5)
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      }))

      is.ini(as.ini(ini))

      suppressMessages(expect_error(ini |> ini(tka=3), NA))

      l <- quote(lotri({
        tka <- log(1.57)
        tcl <- log(2.72)
        tv <- log(31.5)
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      }))

      is.ini(as.ini(l))

      suppressMessages(expect_error(l |> ini(tka=3), NA))

      m <- lotri({
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
      })

      is.ini(as.ini(m))

      suppressMessages(expect_error(m |> ini(eta.ka=3), NA))

      suppressMessages(expect_true(inherits(m |> ini(eta.ka=3), "matrix")))

      one.compartment <- function() {
        ini({
          tka <- log(1.57)
          tcl <- log(2.72)
          tv <- log(31.5)
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.sd <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
          cp ~ add(add.sd)
        })
      }

      is.ini(as.ini(one.compartment))

      ui <- one.compartment()

      is.ini(as.ini(ui))

      is.ini(as.ini(ui$iniDf))

      ini <- c("ini({",
               "tka <- log(1.57)",
               "tcl <- log(2.72)",
               "tv <- log(31.5)",
               "eta.ka ~ 0.6",
               "eta.cl ~ 0.3",
               "eta.v ~ 0.1",
               "add.sd <- 0.7",
               "})")

      is.ini(as.ini(ini))

      suppressMessages(expect_error(ini |> ini(tka=3), NA))

      suppressMessages(expect_true(inherits(ini |> ini(tka=3) |> lotri::as.lotri(.),
                                            "lotriFix")))

      ini <- paste(ini, collapse="\n")

      is.ini(as.ini(ini))

      suppressMessages(expect_error(ini |> ini(tka=3), NA))

      l <- lotri({
        tka <- log(1.57)
        tcl <- log(2.72)
        tv <- log(31.5)
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })

      suppressMessages(expect_error(l |> ini(tka=3), NA))
      suppressMessages(expect_true(inherits(l |> ini(tka=3), "lotriFix")))

    })
  }
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("test $modelName", {
      one.cmt <- function() {
        ini({
          tka <- 0.45
          tcl <- c(-Inf, 0.993251773010283, 4.60517018598809)
          tv <- 3.45
          label("log V")
          add.sd <- c(0, 0.7)
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          linCmt() ~ add(add.sd)
        })
      }

      f <- rxode2(one.cmt)

      expect_equal(f$modelName, "one.cmt")

      f <- one.cmt()

      expect_equal(f$modelName, "one.cmt")
    })
  }
})

rxTest({
  test_that("nopred ui", {
    pheno2 <- function() {
      ini({
        tcl <- log(0.008)
        tv <-  log(0.6)
        eta.cl + eta.v ~ c(1,
                           0.01, 1)
      })
      model({
        cl <- exp(tcl + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v)    # individual value of volume
        ke <- cl / v            # elimination rate constant
        d/dt(A1) = - ke * A1    # model differential equation
        cp = A1 / v             # concentration in plasma
      })
    }

    expect_error(rxode2(pheno2), NA)

    mod <- rxode2(pheno2)

    et <- et(amt=1) |> et(1:20) |> et(id=1:20)

    expect_error(rxWithSeed(42, rxSolve(mod, et)), NA)
  })
})

rxTest({
  f <- function() {
    ini({
      tke <- 0.5
      eta.ke ~ 0.04
      prop.sd <- sqrt(0.1)
    })
    model({
      ke <- tke * exp(eta.ke)
      ipre <- 10 * exp(-ke * t)
      f2 <- ipre / (ipre + 5)
      f3 <- f2 * 3
      lipre <- log(ipre)
      ipre ~ prop(prop.sd)
    })
  }

  expect_equal_quietly <- function(ui, expected) {
    uiUp <- suppressMessages(force(ui))
    expect_equal(.rxGetVarianceForErrorType(uiUp, uiUp$predDf[1, ]), expected)
  }

  test_that("constant error model tests", {
    # Constant:  y=f+a*err
    # Var = a^2

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd)),
      quote((add.sd) ^ 2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2)),
      quote((f2) ^ 2)
    )
  })

  test_that("prop model tests", {
    # Proportional:  y=f+b*f*err
    # Var = (f*b)^2

    expect_equal_quietly(
      f |> model(ipre ~ prop(prop.sd)),
      quote((rx_pred_f_ * prop.sd) ^ 2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ prop(f2)),
      quote((rx_pred_f_ * f2) ^ 2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ propT(prop.sd)),
      quote((rx_pred_ * prop.sd) ^ 2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ propT(f2)),
      quote((rx_pred_ * f2) ^ 2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ propF(prop.sd, f3)),
      quote((f3 * prop.sd) ^ 2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ propF(f2, f3)),
      quote((f3 * f2) ^ 2)
    )
  })

  test_that("power models", {
    # Power:  y=f+b*f^c*err
    # Var = (f^c*b)^2

    # pow(sd,pw)
    expect_equal_quietly(
      f |> model(ipre ~ pow(prop.sd, pw)),
      quote(((rx_pred_f_)^(pw) * prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ pow(f2, pw)),
      quote(((rx_pred_f_)^(pw) * f2)^2)
    )

    # powT(sd,pw)
    expect_equal_quietly(
      f |> model(ipre ~ powT(prop.sd, pw)),
      quote(((rx_pred_)^(pw) * prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ powT(f2, pw)),
      quote(((rx_pred_)^(pw) * f2)^2)
    )

    # powF(sd,pw,f)
    expect_equal_quietly(
      f |> model(ipre ~ powF(prop.sd, pw, f3)),
      quote(((f3)^(pw) * prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ powF(f2, pw, f3)),
      quote(((f3)^(pw) * f2)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ powF(f2, lipre, f3)),
      quote(((f3)^(lipre) * f2)^2)
    )
  })

  test_that("add+prop combined1", {
    # Combined1:  y=f + (a+b*f^c)*err
    # Var = (a + b*f^c)^2

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + prop(prop.sd) + combined1()),
      quote(((add.sd) + (rx_pred_f_) * (prop.sd))^2)
    )

    expect_equal_quietly(
      f |>
        model(ipre ~ add(f2) + prop(prop.sd) + combined1()),
      quote(((f2) + (rx_pred_f_) * (prop.sd))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + prop(f2) + combined1()),
      quote(((add.sd) + (rx_pred_f_) * (f2))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + prop(f3) + combined1()),
      quote(((f2) + (rx_pred_f_) * (f3))^2)
    )

    # propT
    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + propT(prop.sd) + combined1()),
      quote(((add.sd) + (rx_pred_) * (prop.sd))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + propT(prop.sd) + combined1()),
      quote(((f2) + (rx_pred_) * (prop.sd))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + propT(f2) + combined1()),
      quote(((add.sd) + (rx_pred_) * (f2))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + propT(f3) + combined1()),
      quote(((f2) + (rx_pred_) * (f3))^2)
    )

    # propF
    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + propF(prop.sd, f3) + combined1()),
      quote(((add.sd) + (f3) * (prop.sd))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(lipre) + propF(prop.sd, f3) + combined1()),
      quote(((lipre) + (f3) * (prop.sd))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + propF(lipre, f3) + combined1()),
      quote(((add.sd) + (f3) * (lipre))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(lipre) + propF(f2, f3) + combined1()),
      quote(((lipre) + (f3) * (f2))^2)
    )
  })

  test_that("add+pow combined1", {
    # Combined1:  y=f + (a+b*f^c)*err
    # Var = (a + b*f^c)^2

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + pow(prop.sd, c) + combined1()),
      quote(((add.sd) + (rx_pred_f_) ^ (c) * (prop.sd))^2)
    )

    expect_equal_quietly(
      f |>
        model(ipre ~ add(f2) + pow(prop.sd, c) + combined1()),
      quote(((f2) + (rx_pred_f_) ^ (c) * (prop.sd))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + pow(f2, c) + combined1()),
      quote(((add.sd) + (rx_pred_f_) ^ (c)* (f2))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + pow(f3, c) + combined1()),
      quote(((f2) + (rx_pred_f_) ^ (c) * (f3))^2)
    )

    # propT
    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + powT(prop.sd, c) + combined1()),
      quote(((add.sd) + (rx_pred_) ^ (c) * (prop.sd))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + powT(prop.sd, c) + combined1()),
      quote(((f2) + (rx_pred_) ^ (c)* (prop.sd))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + powT(f2, c) + combined1()),
      quote(((add.sd) + (rx_pred_) ^ (c)* (f2))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + powT(f3, c) + combined1()),
      quote(((f2) + (rx_pred_) ^ (c) * (f3))^2)
    )

    # propF
    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + powF(prop.sd, c, f3) + combined1()),
      quote(((add.sd) + (f3) ^ (c) * (prop.sd))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(lipre) + powF(prop.sd, c, f3) + combined1()),
      quote(((lipre) + (f3) ^ (c)* (prop.sd))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + powF(lipre, c, f3) + combined1()),
      quote(((add.sd) + (f3) ^ (c) * (lipre))^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(lipre) + powF(f2, c, f3) + combined1()),
      quote(((lipre) + (f3) ^ (c)* (f2))^2)
    )
  })

  # combined2
  test_that("add+prop combined2", {
    # Combined2:  y=f + sqrt(a^2 + b^2*(f^c)^2)*err
    # Var = a^2 + b^2*(f^c)^2

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + prop(prop.sd) + combined2()),
      quote((add.sd)^2 + (rx_pred_f_)^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + prop(prop.sd) + combined2()),
      quote((f2)^2 + (rx_pred_f_)^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + prop(f2) + combined2()),
      quote((add.sd)^2 + (rx_pred_f_)^2 * (f2)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + prop(f3) + combined2()),
      quote((f2)^2 + (rx_pred_f_)^2 * (f3)^2)
    )

    # propT
    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + propT(prop.sd) + combined2()),
      quote((add.sd)^2 + (rx_pred_)^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + propT(prop.sd) + combined2()),
      quote((f2)^2 + (rx_pred_)^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + propT(f2) + combined2()),
      quote((add.sd)^2 + (rx_pred_)^2 * (f2)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + propT(f3) + combined2()),
      quote((f2)^2 + (rx_pred_)^2 * (f3)^2)
    )

    # propF
    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + propF(prop.sd, f3) + combined2()),
      quote((add.sd)^2 + (f3)^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(lipre) + propF(prop.sd, f3) + combined2()),
      quote((lipre)^2 + (f3)^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + propF(lipre, f3) + combined2()),
      quote((add.sd)^2 + (f3)^2 * (lipre)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(lipre) + propF(f2, f3) + combined2()),
      quote((lipre)^2 + (f3)^2 * (f2)^2)
    )
  })


  test_that("add+pow combined2", {
    # Combined2:  y=f + (a+b*f^c)*err
    # Var = (a + b*f^c)^2

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + pow(prop.sd, c) + combined2()),
      quote((add.sd)^2 + ((rx_pred_f_)^(c))^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + pow(prop.sd, c) + combined2()),
      quote((f2)^2 + ((rx_pred_f_)^(c))^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + pow(f2, c) + combined2()),
      quote((add.sd)^2 + ((rx_pred_f_)^(c))^2 * (f2)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + pow(f3, c) + combined2()),
      quote((f2)^2 + ((rx_pred_f_)^(c))^2 * (f3)^2)
    )

    # propT
    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + powT(prop.sd, c) + combined2()),
      quote((add.sd)^2 + ((rx_pred_)^(c))^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + powT(prop.sd, c) + combined2()),
      quote((f2)^2 + ((rx_pred_)^(c))^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + powT(f2, c) + combined2()),
      quote((add.sd)^2 + ((rx_pred_)^(c))^2 * (f2)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(f2) + powT(f3, c) + combined2()),
      quote((f2)^2 + ((rx_pred_)^(c))^2 * (f3)^2)
    )

    # propF
    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + powF(prop.sd, c, f3) + combined2()),
      quote((add.sd)^2 + ((f3)^(c))^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(lipre) + powF(prop.sd, c, f3) + combined2()),
      quote((lipre)^2 + ((f3)^(c))^2 * (prop.sd)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(add.sd) + powF(lipre, c, f3) + combined2()),
      quote((add.sd)^2 + ((f3)^(c))^2 * (lipre)^2)
    )

    expect_equal_quietly(
      f |> model(ipre ~ add(lipre) + powF(f2, c, f3) + combined2()),
      quote((lipre)^2 + ((f3)^(c))^2 * (f2)^2)
    )
  })
})

rxTest({

  test_that("lag()/diff() of an lhs variable return the previous record value", {
    f <- function() {
      ini({tcl <- 1})
      model({
        cl <- exp(tcl)
        myr <- cl * time
        plag <- lag(myr, 1)
        plag1 <- lag(myr)
        pdiff <- diff(myr, 1)
        y <- plag
      })
    }
    ui <- rxode2(f)
    r <- rxSolve(ui, et(c(0, 2, 5, 9)), returnType = "data.frame")
    # previous record's myr, NA on the first record
    expect_equal(r$plag, c(NA, r$myr[1:3]))
    expect_equal(r$plag1, c(NA, r$myr[1:3]))
    # diff = current - previous
    expect_equal(r$pdiff, c(NA, diff(r$myr)))
  })

  test_that("lag()/diff() reset per individual", {
    f <- function() {
      ini({tcl <- 1})
      model({
        cl <- exp(tcl)
        myr <- cl * time
        plag <- lag(myr, 1)
        y <- plag
      })
    }
    ui <- rxode2(f)
    ev <- et(c(0, 2, 5)) %>% et(id = 1:2)
    r <- rxSolve(ui, ev, returnType = "data.frame")
    # each individual starts with NA (no carry-over across subjects)
    expect_true(all(is.na(r$plag[r$time == 0])))
    for (.id in unique(r$id)) {
      .s <- r[r$id == .id, ]
      expect_equal(.s$plag, c(NA, .s$myr[1:2]))
    }
  })

  test_that("lag() of a computed lhs survives the estimation (symengine) path", {
    f <- function() {
      ini({tcl <- 1; add.sd <- 0.5})
      model({
        cl <- exp(tcl)
        myr <- cl * time
        p <- lag(myr, 1)
        cp <- cl + p
        cp ~ add(add.sd)
      })
    }
    ui <- rxode2(f)
    # the pruned/symengine estimation model must still contain the lag
    expect_error(ui$symengineModelPrune, NA)
  })

  test_that("lag(x, n) with n > 1 is rejected for an lhs variable", {
    f <- function() {
      ini({tcl <- 1})
      model({cl <- exp(tcl); myr <- cl * time; p <- lag(myr, 2); y <- p})
    }
    expect_error(rxode2(f))
  })

  test_that("a variable may lag itself (first-order recurrence)", {
    # b = phi*lag0(b, 1) + 1 is a legal recurrence: lag0() reads the PREVIOUS
    # record's stored value, not the current uninitialized one.
    f <- function() {
      ini({phi <- 0.5})
      model({
        b <- phi * lag0(b, 1) + 1
        y <- b
      })
    }
    ui <- rxode2(f)
    r <- rxSolve(ui, et(1:6), returnType = "data.frame")
    # b_1 = 1 (lag0 = 0 on first record); b_i = 0.5*b_{i-1} + 1
    .expect <- Reduce(function(prev, .) 0.5 * prev + 1, seq_len(5), accumulate = TRUE, 1)
    expect_equal(r$y, .expect)
  })

  test_that("self-lag resets per individual", {
    f <- function() {
      ini({phi <- 0.5})
      model({b <- phi * lag0(b, 1) + 1; y <- b})
    }
    ui <- rxode2(f)
    r <- rxSolve(ui, et(1:3) %>% et(id = 1:2), returnType = "data.frame")
    for (.id in unique(r$id)) {
      expect_equal(r$y[r$id == .id], c(1, 1.5, 1.75))
    }
  })

  test_that("a non-lag self-reference is still treated as a parameter", {
    # b = b*2 (no lag) reads the current value -> b is a required input param,
    # NOT a recurrence; must not be silently turned into a self-lag.
    f <- function() {
      ini({tcl <- 1})
      model({cl <- exp(tcl); b <- b * 2; y <- b})
    }
    ui <- rxode2(f)
    expect_true("b" %in% ui$mv0$params)
  })

  test_that("lag()/diff() of a time-varying covariate return the previous value", {
    f <- function() {
      ini({tcl <- 1})
      model({
        cl <- exp(tcl)
        pw <- lag(WT, 1)
        dw <- diff(WT, 1)
        y <- cl + WT
      })
    }
    ui <- rxode2(f)
    ev <- et(c(0, 2, 5, 9))
    ev$WT <- c(70, 72, 75, 80)
    r <- rxSolve(ui, ev, returnType = "data.frame")
    expect_equal(r$pw, c(NA, 70, 72, 75))
    expect_equal(r$dw, c(NA, 2, 3, 5))
  })

})

rxTest({
  test_that("proper errors from event tables with single id", {

    m <- function() {
      ini({
        tCL20 <-  c(10)
        tV1 <- 20
        tCL12 <- 30
        tV2 <- 40
        tCL23 <- 50
        tCL32 <- 60
        betaWTCL20 <- 0.1
        betaSexV1 <- 0.2
        prop.err <- c(0, 0.268724820805969)
        eta.V1 + eta.CL12 + eta.V2 + eta.CL20 + eta.CL23 + eta.CL32 ~
          cor(sd(1,
                 0.25, 1.10,
                 -0.25, 0.25, 1.2,
                 -0.35, 0.35, -0.45, 1.3,
                 -0.2,  0.2,  0.2,   -0.2, 1.4,
                 -0.1,  0.1, -0.1, 0.1, -0.1, 1.4))
      })
      model({
        CL20 <- exp(tCL20 + eta.CL20) * ((WTBL/85)^betaWTCL20)
        V1 <- exp(tV1 + eta.V1) * (1 + betaSexV1 * SEXN)
        CL12 <- exp(tCL12 + eta.CL12)
        V2 <- exp(tV2 + eta.V2)
        CL23 <- exp(tCL23 + eta.CL23)
        CL32 <- exp(tCL32 + eta.CL32)
        k12 <- CL12/V1
        k20 <- CL20/V2
        k23 <- CL23/V2
        k32 <- CL32/V1
        d/dt(central) = -k12 * central
        d/dt(tissue) = k12 * central - (k20 + k23) * tissue +    k32 * central_ext
        d/dt(central_ext) = k23 * tissue - k32 * central_ext
        d/dt(central_tot) = (central + central_ext)/V1
        conc = (central + central_ext)/V1
        IPRED = conc
        IPRED ~ prop(prop.err)
      })
    }


    ev <- eventTable() |>
      add.dosing(dose=900630, nbr.doses=1, rate=900630/2) |>
      et(0, 600, length.out=10)

    expect_error(rxSolve(m, ev,

                 # Create individual covariate data-frame
                 iCov=data.frame(ID=1:10,
                                 WTBL=runif(n = 1000, min = 50, max = 150),
                                 SEXN=rbinom(n = 1000, size = 1, prob = 0.5)),
                 # in this case it would be useful to keep the WT in the output dataset
                 keep=c("WTBL","SEXN")),
                 "the 'id' in the iCov must have 1 unique match to the event table")
  })


  test_that("iCov", {

    nsub <- 10 # 100 sub-problems

    mod1 <- rxode2({
      C2 = centr/V2
      C3 = peri/V3
      d/dt(depot) = -KA*depot
      d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3
      d/dt(peri) = Q*C2 - Q*C3
      d/dt(eff) = Kin - Kout*(1-C2/(EC50+C2))*eff
      eff(0) = 1
    })


    ev <- et() |>
      et(amt=10000, addl=9,ii=12) |>
      et(time=120, amt=20000, addl=4, ii=24) |>
      et(0:240) |> ## Add Sampling
      et(id=1:nsub) ## Add id

    sigma <- matrix(c(0.09,0.08,0.08,0.25),2,2) # IIV covariance matrix
    mv <- rxRmvn(n=nsub, rep(0,2), sigma) # Sample from covariance matrix
    CL <- 7*exp(mv[,1])
    V2 <- 40*exp(mv[,2])
    params.all <- cbind(KA=0.3, CL=CL, V2=V2, Q=10, V3=300,
                        Kin=0.2, Kout=0.2, EC50=8)
    paramsDf <- as.data.frame(params.all)
    paramsDf$ptS=paste(1:nsub, "pt")
    paramsDf$ptF <- factor(paramsDf$ptS)
    paramsDf$ptI <- 1:nsub
    paramsDf$ptL <- paramsDf$ptI <=50

    expect_error(etTrans(ev, mod1, iCov=paramsDf, keep=c("CL", "V2")),
                 "'iCov' must have an 'id' column")

    paramsDf$id <- 1:nsub+1
    expect_error(etTrans(ev, mod1, iCov=paramsDf, keep=c("CL", "V2")),
                 "some of the 'id' values do not match the input event table")


    paramsDf$id <- 1:nsub

    paramsDf2 <- paramsDf

    paramsDf2$id <- paste("pt ",paramsDf2$id)

    expect_error(etTrans(ev, mod1, iCov=paramsDf2),
                 "data 'id' column is an integer; 'iCov' 'id' also needs to be an integer")

    paramsDf2 <- paramsDf[-1,]

    expect_error(etTrans(ev, mod1, iCov=paramsDf2),
                 "the 'id' in the iCov must have 1 unique match to the event table")

    paramsDf2 <- rbind(paramsDf[1,], paramsDf)

    expect_error(etTrans(ev, mod1, iCov=paramsDf2),
                 "the 'id' in the iCov must have 1 unique match to the event table")

    # Now check for proper behavior

    tmp <- etTrans(ev, mod1, iCov=paramsDf)
    tmp2 <- attr(class(tmp), ".rxode2.lst")
    class(tmp2) <- NULL

    for (v in c("EC50", "Kout", "Kin", "V3", "Q", "V2", "CL", "KA")) {
      expect_equal(tmp2$cov1[[v]], paramsDf[[v]])
      expect_equal(tmp2$pars[v,], paramsDf[[v]])
    }

    # Check keep of various sorts in iCov
    tmp <- rxSolve(mod1, ev, iCov=paramsDf, keep=c("CL", "V2", "ptS", "ptF", "ptI", "ptL"))
    expect_true(is.character(tmp$ptS))
    expect_true(all(paste(1:nsub, "pt") %in% unique(tmp$ptS)))
    expect_true(is.factor(tmp$ptF))
    expect_true(all(paste(1:nsub, "pt") %in% unique(tmp$ptF)))
    expect_true(is.integer(tmp$ptI))
    expect_true(is.logical(tmp$ptL))
    expect_true(all((tmp$ptI <= 50) == tmp$ptL))

    for (v in
         c("evid",  "time", "amt", "value", "cmt", "ytype", "state", "var", "dv",
           "rate","dur", "addl", "ii", "mdv", "cens", "limit", "method")) {
      names(paramsDf)[9] <- v
      expect_error(etTrans(ev, mod1, iCov=paramsDf),
                   paste0("cannot specify '", v,"' in 'iCov'"))
    }

  })

  test_that("iCov string comparison covariates use model ordering", {
    mod <- rxode2parse('c = cov == "alpha"; d = "beta" == cov')

    ev <- et() |>
      et(amt = 1, time = 0) |>
      et(time = 1) |>
      et(id = 1:3)

    ic <- data.frame(
      id = 1:3,
      cov = factor(c("beta", "gamma", "alpha"),
                   levels = c("gamma", "beta", "alpha"))
    )

    tmp <- etTrans(ev, mod, iCov = ic, keep = "cov")
    tmp2 <- attr(class(tmp), ".rxode2.lst")
    class(tmp2) <- NULL

    expect_equal(as.numeric(tmp2$cov1$cov), c(2, 3, 1))
    expect_equal(as.numeric(tmp2$pars["cov", ]), c(2, 3, 1))
  })
})

rxTest({
  test_that("rename for ui makes sense", {

    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        tv <- exp(3.45)
        add.sd <- 0.7
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    expect_error(rxRename(ocmt, d/dt(cp1) ~ cp))
    expect_error(rxRename(ocmt, cp1 ~ cp))
    expect_error(rxRename(ocmt, cp1=3))
    expect_error(rxRename(ocmt, cp=cp1))
    expect_error(rxRename(ocmt, cp2=cp1))

    f <- rxRename(ocmt, cpParent=cp, parentDepot=depot, parentCentral=center, parentKa=ka,
                  parentTka=tka, parent.sd=add.sd)

    expect_true("parent.sd" %in% f$iniDf$name)
    expect_true("parentTka" %in% f$iniDf$name)
    expect_true("parentKa" %in% f$mv0$lhs)
    expect_true("parentCentral" %in% f$mv0$state)
    expect_true("parentDepot" %in% f$mv0$state)
    expect_true("cpParent" %in% f$mv0$lhs)
    expect_true("cpParent" %in% f$predDf$var)
    expect_true("cpParent" %in% f$predDf$cond)

    f <- dplyr::rename(ocmt, cp.parent=cp)
    expect_true("cp.parent" %in% f$mv0$lhs)
    expect_true("cp.parent" %in% f$predDf$var)
    expect_true("cp.parent" %in% f$predDf$cond)

    f2 <- dplyr::rename(f, depot.parent=depot)
    expect_true("depot.parent" %in% f2$mv0$state)

    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        tv <- exp(3.45)
        add.sd <- 0.7
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd) | parent
      })
    }

    f <- rxRename(ocmt, metabolite=parent)

    expect_true("metabolite" %in% f$predDf$cond)

    # now test that functions remain intact

    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        tv <- exp(3.45)
        lag <- 1
        add.sd <- 0.7
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        lag(depot) <- lag
        cp = center / v
        cp ~ add(add.sd) | parent
      })
    }

    f <- dplyr::rename(ocmt, alag=lag)

    expect_equal(f$lstExpr[[6]], quote(lag(depot) <- alag))

    # now test that d/dt(item) remains intact


    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        tv <- exp(3.45)
        d <- 1
        add.sd <- 0.7
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        lag(depot) <- d
        cp = center / v
        cp ~ add(add.sd) | parent
      })
    }

    f <- dplyr::rename(ocmt, alag=d)
    expect_equal(f$lstExpr[[6]], quote(lag(depot) <- alag))

    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        tv <- exp(3.45)
        add.sd <- 0.7
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        depot(0) <- 1
        rate(depot) <- 1
        dur(depot) <- 1
        alag(depot) <- 1
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    tmp <- rxRename(rxode2(ocmt), dcmt=depot)

    expect_equal(tmp$lstExpr[[5]], quote(dcmt(0) <- 1))

    expect_equal(tmp$lstExpr[[6]], quote(rate(dcmt) <- 1))
    expect_equal(tmp$lstExpr[[7]], quote(dur(dcmt) <- 1))
    expect_equal(tmp$lstExpr[[8]], quote(alag(dcmt) <- 1))

  })

  test_that("rename with sigma and thetaMat", {

    f <- function() {
      description <- "BOLUS_2CPT_CLV1QV2 SINGLE DOSE FOCEI (120 Ind/2280 Obs) runODE032"
      dfObs <- 2280
      dfSub <- 120
      sigma <- lotri({
        eps1 ~ 1
      })
      thetaMat <- lotri({
        theta1 + theta2 + theta3 + theta4 + RSV + eps1 + eta1 +
          omega.2.1 + eta2 + omega.3.1 + omega.3.2 + eta3 +
          omega.4.1 + omega.4.2 + omega.4.3 + eta4 ~
            c(0.000887681,
              -0.00010551, 0.000871409, 0.000184416, -0.000106195,
              0.00299336, -0.000120234, -5.06663e-05, 0.000165252,
              0.00121347, 5.2783e-08, -1.56562e-05, 5.99331e-06,
              -2.53991e-05, 9.94218e-06, 0, 0, 0, 0, 0, 0, -4.71273e-05,
              4.69667e-05, -3.64271e-05, 2.54796e-05, -8.16885e-06,
              0, 0.000169296, 0, 0, 0, 0, 0, 0, 0, 0, -7.37156e-05,
              2.56634e-05, -8.08349e-05, 1.37e-05, -4.36564e-06,
              0, 8.75181e-06, 0, 0.00015125, 0, 0, 0, 0, 0, 0,
              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.63383e-05,
              -8.19002e-05, 0.000548985, 0.000168356, 1.59122e-06,
              0, 3.48714e-05, 0, 4.31593e-07, 0, 0, 0.000959029,
              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
              0, 0, 0, 0, 0, 0, 0, 0, -9.49661e-06, 0.000110108,
              -0.000306537, -9.12897e-05, 3.1877e-06, 0, 1.36628e-05,
              0, -1.95096e-05, 0, 0, -0.00012977, 0, 0, 0, 0.00051019)
      })
      ini({
        theta1 <- 1.37034036528946
        label("log Cl")
        theta2 <- 4.19814911033061
        label("log Vc")
        theta3 <- 1.38003493562413
        label("log Q")
        theta4 <- 3.87657341967489
        label("log Vp")
        RSV <- c(0, 0.196446108190896, 1)
        label("RSV")
        eta1 ~ 0.101251418415006
        eta2 ~ 0.0993872449483344
        eta3 ~ 0.101302674763154
        eta4 ~ 0.0730497519364148
      })
      model({
        cmt(CENTRAL)
        cmt(PERI)
        cl <- exp(theta1 + eta1)
        v <- exp(theta2 + eta2)
        q <- exp(theta3 + eta3)
        v2 <- exp(theta4 + eta4)
        v1 <- v
        scale1 <- v
        k21 <- q/v2
        k12 <- q/v
        d/dt(CENTRAL) <- k21 * PERI - k12 * CENTRAL - cl * CENTRAL/v1
        d/dt(PERI) <- -k21 * PERI + k12 * CENTRAL
        f <- CENTRAL/scale1
        ipred <- f
        rescv <- RSV
        w <- ipred * rescv
        ires <- DV - ipred
        iwres <- ires/w
        y <- ipred + w * eps1
      })
    }

    f <- f()

    f2 <- rxRename(f, eta.cl=eta1, eta.v=eta2, eta.q=eta3, eta.v2=eta4, eps=eps1)

    expect_equal(sort(intersect(dimnames(f2$omega)[[1]],
                                c("eta.cl", "eta.v", "eta.q", "eta.v2"))),
                 c("eta.cl", "eta.q", "eta.v", "eta.v2"))

    expect_equal(sort(intersect(dimnames(f2$thetaMat)[[1]],
                                c("eta.cl", "eta.v", "eta.q", "eta.v2", "eps"))),
                 c("eps", "eta.cl", "eta.q", "eta.v", "eta.v2"))

    expect_equal(dimnames(f2$sigma)[[1]], "eps")

    f <- rxUiDecompress(f)
    f$sigma <- f$meta$sigma
    f$thetaMat <- f$meta$thetaMat
    rm("sigma", envir=f$meta)
    rm("thetaMat", envir=f$meta)
    f <- rxUiCompress(f)

    f3 <- rxRename(f, eta.cl=eta1, eta.v=eta2, eta.q=eta3, eta.v2=eta4, eps=eps1)

    expect_equal(sort(intersect(dimnames(f3$omega)[[1]],
                                c("eta.cl", "eta.v", "eta.q", "eta.v2"))),
                 c("eta.cl", "eta.q", "eta.v", "eta.v2"))

    expect_equal(sort(intersect(dimnames(f3$thetaMat)[[1]],
                                c("eta.cl", "eta.v", "eta.q", "eta.v2", "eps"))),
                 c("eps", "eta.cl", "eta.q", "eta.v", "eta.v2"))

    expect_equal(dimnames(f3$sigma)[[1]], "eps")

  })

  test_that("rename doesn't change parent ui", {

    f <- function() {
      description <- "BOLUS_2CPT_CLV1QV2 SINGLE DOSE FOCEI (120 Ind/2280 Obs) runODE032"
      dfObs <- 2280
      dfSub <- 120
      sigma <- lotri({
        eps1 ~ 1
      })
      thetaMat <- lotri({
        theta1 ~ c(theta1 = 0.000887681)
        theta2 ~ c(theta1 = -0.00010551, theta2 = 0.000871409)
        theta3 ~ c(theta1 = 0.000184416, theta2 = -0.000106195,
                   theta3 = 0.00299336)
        theta4 ~ c(theta1 = -0.000120234, theta2 = -5.06663e-05,
                   theta3 = 0.000165252, theta4 = 0.00121347)
        RSV ~ c(theta1 = 5.2783e-08, theta2 = -1.56562e-05, theta3 = 5.99331e-06,
                theta4 = -2.53991e-05, RSV = 9.94218e-06)
        eps1 ~ c(theta1 = 0, theta2 = 0, theta3 = 0, theta4 = 0,
                 RSV = 0, eps1 = 0)
        eta1 ~ c(theta1 = -4.71273e-05, theta2 = 4.69667e-05,
                 theta3 = -3.64271e-05, theta4 = 2.54796e-05, RSV = -8.16885e-06,
                 eps1 = 0, eta1 = 0.000169296)
        omega.2.1 ~ c(theta1 = 0, theta2 = 0, theta3 = 0, theta4 = 0,
                      RSV = 0, eps1 = 0, eta1 = 0, omega.2.1 = 0)
        eta2 ~ c(theta1 = -7.37156e-05, theta2 = 2.56634e-05,
                 theta3 = -8.08349e-05, theta4 = 1.37e-05, RSV = -4.36564e-06,
                 eps1 = 0, eta1 = 8.75181e-06, omega.2.1 = 0, eta2 = 0.00015125)
        omega.3.1 ~ c(theta1 = 0, theta2 = 0, theta3 = 0, theta4 = 0,
                      RSV = 0, eps1 = 0, eta1 = 0, omega.2.1 = 0, eta2 = 0,
                      omega.3.1 = 0)
        omega.3.2 ~ c(theta1 = 0, theta2 = 0, theta3 = 0, theta4 = 0,
                      RSV = 0, eps1 = 0, eta1 = 0, omega.2.1 = 0, eta2 = 0,
                      omega.3.1 = 0, omega.3.2 = 0)
        eta3 ~ c(theta1 = 6.63383e-05, theta2 = -8.19002e-05,
                 theta3 = 0.000548985, theta4 = 0.000168356, RSV = 1.59122e-06,
                 eps1 = 0, eta1 = 3.48714e-05, omega.2.1 = 0, eta2 = 4.31593e-07,
                 omega.3.1 = 0, omega.3.2 = 0, eta3 = 0.000959029)
        omega.4.1 ~ c(theta1 = 0, theta2 = 0, theta3 = 0, theta4 = 0,
                      RSV = 0, eps1 = 0, eta1 = 0, omega.2.1 = 0, eta2 = 0,
                      omega.3.1 = 0, omega.3.2 = 0, eta3 = 0, omega.4.1 = 0)
        omega.4.2 ~ c(theta1 = 0, theta2 = 0, theta3 = 0, theta4 = 0,
                      RSV = 0, eps1 = 0, eta1 = 0, omega.2.1 = 0, eta2 = 0,
                      omega.3.1 = 0, omega.3.2 = 0, eta3 = 0, omega.4.1 = 0,
                      omega.4.2 = 0)
        omega.4.3 ~ c(theta1 = 0, theta2 = 0, theta3 = 0, theta4 = 0,
                      RSV = 0, eps1 = 0, eta1 = 0, omega.2.1 = 0, eta2 = 0,
                      omega.3.1 = 0, omega.3.2 = 0, eta3 = 0, omega.4.1 = 0,
                      omega.4.2 = 0, omega.4.3 = 0)
        eta4 ~ c(theta1 = -9.49661e-06, theta2 = 0.000110108,
                 theta3 = -0.000306537, theta4 = -9.12897e-05, RSV = 3.1877e-06,
                 eps1 = 0, eta1 = 1.36628e-05, omega.2.1 = 0, eta2 = -1.95096e-05,
                 omega.3.1 = 0, omega.3.2 = 0, eta3 = -0.00012977,
                 omega.4.1 = 0, omega.4.2 = 0, omega.4.3 = 0, eta4 = 0.00051019)
      })
      validation <- c("IPRED relative difference compared to Nonmem IPRED: 0%; 95% percentile: (0%,0%); rtol=6.43e-06",
                      "IPRED absolute difference compared to Nonmem IPRED: 95% percentile: (2.19e-05, 0.0418); atol=0.00167",
                      "IWRES relative difference compared to Nonmem IWRES: 0%; 95% percentile: (0%,0.01%); rtol=8.99e-06",
                      "IWRES absolute difference compared to Nonmem IWRES: 95% percentile: (1.82e-07, 4.63e-05); atol=3.65e-06",
                      "PRED relative difference compared to Nonmem PRED: 0%; 95% percentile: (0%,0%); rtol=6.41e-06",
                      "PRED absolute difference compared to Nonmem PRED: 95% percentile: (1.41e-07,0.00382) atol=6.41e-06")
      ini({
        theta1 <- 1.37034036528946
        label("log Cl")
        theta2 <- 4.19814911033061
        label("log Vc")
        theta3 <- 1.38003493562413
        label("log Q")
        theta4 <- 3.87657341967489
        label("log Vp")
        RSV <- c(0, 0.196446108190896, 1)
        label("RSV")
        eta1 ~ 0.101251418415006
        eta2 ~ 0.0993872449483344
        eta3 ~ 0.101302674763154
        eta4 ~ 0.0730497519364148
      })
      model({
        cmt(CENTRAL)
        cmt(PERI)
        cl <- exp(theta1 + eta1)
        v <- exp(theta2 + eta2)
        q <- exp(theta3 + eta3)
        v2 <- exp(theta4 + eta4)
        v1 <- v
        scale1 <- v
        k21 <- q/v2
        k12 <- q/v
        d/dt(CENTRAL) <- k21 * PERI - k12 * CENTRAL - cl * CENTRAL/v1
        d/dt(PERI) <- -k21 * PERI + k12 * CENTRAL
        f <- CENTRAL/scale1
        ipred <- f
        rescv <- RSV
        w <- ipred * rescv
        ires <- DV - ipred
        iwres <- ires/w
        y <- ipred + w * eps1
      })
    }

    f <- f()

    f2 <- f |> rxRename(err1=eps1)
    expect_equal(dimnames(f2$sigma)[[1]], "err1")
    expect_equal(dimnames(f$sigma)[[1]], "eps1")
    expect_true(any(dimnames(f2$thetaMat)[[1]] == "err1"))
    expect_false(any(dimnames(f2$thetaMat)[[1]] == "eps1"))
    expect_false(any(dimnames(f$thetaMat)[[1]] == "err1"))
    expect_true(any(dimnames(f$thetaMat)[[1]] == "eps1"))

    f3 <- f |> rxRename(Vp=theta4)
    expect_true(any(dimnames(f3$thetaMat)[[1]] == "Vp"))
    expect_false(any(dimnames(f3$thetaMat)[[1]] == "theta4"))
    expect_false(any(dimnames(f$thetaMat)[[1]] == "Vp"))
    expect_true(any(dimnames(f$thetaMat)[[1]] == "theta4"))

  })


})

rxTest({
  # AutoSwitch composite "primary+stiff": the stiff secondary's analytical
  # Jacobian must be generated and used whenever the secondary is an implicit
  # (Rosenbrock / implicit-RK) method.  These methods are exactly the ones for
  # which rxIsImplicit() is TRUE and which the C solver's _jacAvailable check
  # expects: ros4(13), iem(14), ros43(31), ros6(32), backwardEuler(33),
  # gauss6(34), iiic6(35), radauiia5(36), geng5(37), sdirk43(38).

  ## A well-conditioned stiff linear system (eigenvalues -100, -1).
  .stiff <- rxode2({
    d/dt(a) <- -100 * a + 99 * b
    d/dt(b) <- a - b
    a(0) <- 1
    b(0) <- 0
  })
  .ev <- et(seq(0, 3, by = 0.5))
  .ref <- rxSolve(.stiff, .ev, method = "lsoda", atol = 1e-10, rtol = 1e-10)

  .implicit <- c("ros4", "iem", "ros43", "ros6", "backwardEuler",
                 "gauss6", "iiic6", "radauiia5", "geng5", "sdirk43")

  test_that("rxIsImplicit() flags exactly the Jacobian-needing methods", {
    expect_true(all(rxIsImplicit(.implicit)))
    ## solvers that supply their own Jacobian internally are NOT flagged
    expect_false(any(rxIsImplicit(c("dop853", "lsoda", "liblsoda", "cvode", "bdf"))))
  })

  test_that("dop853 + <implicit> composites generate and use the Jacobian", {
    for (.s in .implicit) {
      .m <- paste0("dop853+", .s)
      .x <- rxSolve(.stiff, .ev, method = .m, atol = 1e-8, rtol = 1e-8)
      expect_false(any(is.na(.x$a)),
                   info = paste(.m, "produced NA (Jacobian not hooked up?)"))
      expect_true(max(abs(.x$a - .ref$a)) < 1e-5,
                  info = paste(.m, "did not match the reference solution"))
    }
  })

  test_that("the dense dop853+ros4 composite generates and uses the Jacobian", {
    .x <- rxSolve(.stiff, .ev, method = "dop853+ros4", dense = TRUE,
                  atol = 1e-8, rtol = 1e-8)
    expect_false(any(is.na(.x$a)))
    expect_true(max(abs(.x$a - .ref$a)) < 1e-5)
  })

  test_that("the dense dop853+ros4 composite solves steady state (SS-path regression)", {
    ## Regression for the SS-path AutoSwitch gotcha.  solveWith1Pt -- the
    ## single-interval solver that steady-state dosing advances repeatedly --
    ## once kept an interval-length Gershgorin stiffness pre-check after the
    ## main-solve paths had dropped it.  On the long tau-sized SS intervals
    ## that check over-estimated the spectral radius and spuriously toggled
    ## ind->autoMethod; the toggled state leaked into the dense main solve and
    ## forced ros4 segments that returned stale (unchanged) state, so every
    ## observation after the first SS dose was corrupted.  Making solveWith1Pt
    ## reactive (like the main-solve paths) fixes it.  This pins the behavior
    ## for the dense composite, which is otherwise only covered indirectly --
    ## and was mislabeled -- in the large nmtest suite.
    .ssEv <- et() |>
      et(amt = 10, ii = 1, ss = 1, cmt = "a") |>
      et(seq(0, 1, by = 0.1))
    .refss <- rxSolve(.stiff, .ssEv, method = "lsoda", atol = 1e-10, rtol = 1e-10)
    .xss <- rxSolve(.stiff, .ssEv, method = "dop853+ros4", dense = TRUE,
                    atol = 1e-8, rtol = 1e-8)
    expect_false(any(is.na(.xss$a)))
    ## not stale: the steady-state trajectory genuinely varies across the
    ## interval (the bug pinned every post-dose observation to one value).
    expect_true(stats::sd(.xss$a) > 1e-6)
    ## and it matches the reference steady-state solution.
    expect_true(max(abs(.xss$a - .refss$a)) < 1e-4)
  })

  test_that("the non-dense dop853+ros4 composite switches to ros4 mid-solve", {
    ## A stiff Robertson problem with widely spaced output times that overwhelms
    ## the non-stiff dop853 primary.  The composite must switch to ros4 per
    ## interval and solve it; pure dop853 cannot.
    .rob <- rxode2({
      d/dt(a)  <- -0.04 * a + 1e4 * b * cc
      d/dt(b)  <-  0.04 * a - 1e4 * b * cc - 3e7 * b * b
      d/dt(cc) <-  3e7 * b * b
      a(0) <- 1
      b(0) <- 0
      cc(0) <- 0
    })
    .evr <- et(c(0.1, 1, 10, 100))
    .refr <- rxSolve(.rob, .evr, method = "lsoda", atol = 1e-10, rtol = 1e-10)

    ## pure dop853 cannot solve it ...
    expect_error(rxSolve(.rob, .evr, method = "dop853", atol = 1e-8, rtol = 1e-8))

    ## ... but the non-dense composite (no dense=TRUE) does, matching lsoda.
    .xr <- rxSolve(.rob, .evr, method = "dop853+ros4", atol = 1e-8, rtol = 1e-8)
    expect_false(any(is.na(.xr$a)))
    expect_true(max(abs(.xr$a - .refr$a)) < 1e-5)
  })

  test_that("dop853+ros4 composite solves a STIFF FORWARD-SENSITIVITY system with the analytic Jacobian", {
    withr::local_options(rxode2.useLinCmt = FALSE)
    # Robertson with free rate constants + first-order sensitivities: the
    # augmented (states + rx__sens_*) system is stiff, so pure dop853 cannot solve
    # it, and the ros4 secondary is only viable because the FULL forward-sensitivity
    # Jacobian is emitted (calcJac=TRUE) -- without it boost's rosenbrock4 could not
    # step the wide augmented system.  This exercises the composite ACTUALLY
    # switching to the stiff secondary on a forward-sens model (not just the primal).
    .rob <- "d/dt(a) = -k1*a + k2*b*cc\nd/dt(b) = k1*a - k2*b*cc - k3*b*b\nd/dt(cc) = k3*b*b"
    .cs <- c("k1", "k2", "k3"); .p <- c(k1 = 0.04, k2 = 1e4, k3 = 3e7); .ini <- c(a = 1, b = 0, cc = 0)
    .ev <- et(c(0.1, 1, 10, 100))
    .scol <- as.vector(outer(c("a", "b", "cc"), .cs, function(s, pn) sprintf("rx__sens_%s_BY_%s__", s, pn)))
    .ref <- as.data.frame(suppressWarnings(rxSolve(rxode2(.rob, calcSens = .cs), .ev,
                                                   params = .p, inits = .ini, method = "lsoda", atol = 1e-10, rtol = 1e-10)))
    .mj <- rxode2(.rob, calcSens = .cs, calcJac = TRUE)
    ## pure dop853 fails on the stiff augmented system ...
    expect_error(suppressWarnings(rxSolve(.mj, .ev, params = .p, inits = .ini, method = "dop853", atol = 1e-8, rtol = 1e-8)))
    ## ... but the composite switches to ros4 (analytic Jacobian) and matches lsoda.
    .x <- as.data.frame(suppressWarnings(rxSolve(.mj, .ev, params = .p, inits = .ini,
                                                 method = "dop853+ros4", atol = 1e-8, rtol = 1e-8)))
    expect_false(any(is.na(.x$a)))
    .sref <- max(1, max(abs(unlist(.ref[.scol])), na.rm = TRUE)); .mx <- 0
    for (cn in .scol) .mx <- max(.mx, max(abs(.x[[cn]] - .ref[[cn]]), na.rm = TRUE))
    expect_lt(.mx / .sref, 1e-3)
  })
})

rxTest({
  test_that("multiple-endpoint", {

    one.compartment.saem <- function() {
      ini({
        tka <- .5 ; label("Log Ka")
        tcl <- -3.2 ; label("Log Cl")
        tv <- -1 ; label("Log V")
        extra <- 20
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        add.err <- 0.1
        add.err2 <- 0.1
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp2 <- cp + extra
        cp ~ add(add.err) | center
        cp2 ~ add(add.err2) | c20
      })
    }

    tmp <- rxode2(one.compartment.saem)

    expect_equal(tmp$predDf$cmt, c(2L, 3L))
    expect_equal(tmp$predDf$dvid, c(1L, 2L))
    expect_equal(tmp$predDf$cond, c("center", "c20"))
    expect_equal(tmp$predDf$var, c("cp", "cp2"))

    ## tmp2 <- rxode2::etTrans(df,tmp$rxode,TRUE)

    one.compartment.saem <- function() {
      ini({
        tka <- .5 ; label("Log Ka")
        tcl <- -3.2 ; label("Log Cl")
        tv <- -1 ; label("Log V")
        extra <- 20
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        add.err <- 0.1
        add.err2 <- 0.1
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp2 <- cp + extra
        cp ~ add(add.err)
        cp2 ~ add(add.err2)
      })
    }

    tmp <- rxode2(one.compartment.saem)
    expect_s3_class(tmp, "rxUi")

    one.compartment.saem <- function() {
      ini({
        tka <- .5 ; label("Log Ka")
        tcl <- -3.2 ; label("Log Cl")
        tv <- -1 ; label("Log V")
        extra <- 20
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        add.err <- 0.1
        add.err2 <- 0.1
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp2 <- cp + extra
        cp ~ add(add.err) | center
        cp2 ~ add(add.err2)
      })
    }

    d <- rxode2(one.compartment.saem)
    expect_s3_class(d, "rxUi")

    one.compartment.saem <- function() {
      ini({
        tka <- .5 ; label("Log Ka")
        tcl <- -3.2 ; label("Log Cl")
        tv <- -1 ; label("Log V")
        extra <- 20
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        add.err <- 0.1
        add.err2 <- 0.1
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp2 <- cp + extra
        cp ~ add(add.err)
        cp2 ~ add(add.err2)
      })
    }

    d <- rxode2(one.compartment.saem)
    expect_s3_class(d, "rxUi")

    pk.turnover.emax <- function() {
      ini({
        tktr <- log(0.00001)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(1)

        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 1
        pkadd.err <- 0.00002

        poplogit <- 2
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)

        eta.emax ~ .5
        eta.ec50 ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5

      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)

        # poplogit = log(temax/(1-temax))
        logit <- exp(poplogit + eta.emax)
        # logit=temax+eta.emax
        emax <- logit / (1 + logit)
        ec50 <- exp(tec50 + eta.ec50)
        kout <- exp(tkout + eta.kout)
        e0 <- exp(te0 + eta.e0)

        DCP <- center / v
        PD <- 1 - emax * DCP / (ec50 + DCP)

        effect(0) <- e0
        kin <- e0 * kout

        d / dt(depot) <- -ktr * depot
        d / dt(gut) <- ktr * depot - ka * gut
        d / dt(center) <- ka * gut - cl / v * center
        d / dt(effect) <- kin * PD - kout * effect

        cp <- center / v
        cp ~ prop(prop.err) + add(pkadd.err)
        pca ~ add(pkadd.err)
      })
    }

    ## Now you can share estimates between endpoints, if they  are in the model, not estimated
    expect_error(rxode2(pk.turnover.emax))

    pk.turnover.emax3 <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        ##
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        poplogit <- 2
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        eta.emax ~ .5
        eta.ec50 ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        ##
        # poplogit = log(temax/(1-temax))
        logit <- exp(poplogit + eta.emax)
        # logit=temax+eta.emax
        emax <- logit / (1 + logit)
        ec50 <- exp(tec50 + eta.ec50)
        kout <- exp(tkout + eta.kout)
        e0 <- exp(te0 + eta.e0)
        ##
        DCP <- center / v
        PD <- 1 - emax * DCP / (ec50 + DCP)
        ##
        effect(0) <- e0
        kin <- e0 * kout
        ##
        d / dt(depot) <- -ktr * depot
        d / dt(gut) <- ktr * depot - ka * gut
        d / dt(center) <- ka * gut - cl / v * center
        d / dt(effect) <- kin * PD - kout * effect
        ##
        cp <- center / v
        cp ~ prop(prop.err) + add(pkadd.err)
        effect ~ add(pdadd.err) | pca
      })
    }

    expect_error(rxode2(pk.turnover.emax3), NA)

    pk.turnover.emax4 <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        poplogit <- 2
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        eta.emax ~ .5
        eta.ec50 ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        ##
        # poplogit = log(temax/(1-temax))
        logit <- exp(poplogit + eta.emax)
        # logit=temax+eta.emax
        emax <- logit / (1 + logit)
        ec50 <- exp(tec50 + eta.ec50)
        kout <- exp(tkout + eta.kout)
        e0 <- exp(te0 + eta.e0)
        ##
        DCP <- center / v
        PD <- 1 - emax * DCP / (ec50 + DCP)
        ##
        effect(0) <- e0
        kin <- e0 * kout
        ##
        d / dt(depot) <- -ktr * depot
        d / dt(gut) <- ktr * depot - ka * gut
        d / dt(center) <- ka * gut - cl / v * center
        d / dt(effect) <- kin * PD - kout * effect
        ##
        cp <- center / v
        ## Who would use this...
        log(cp) ~ prop(prop.err) + add(pkadd.err) | center
        effect * 1 / (1 + cp) ~ add(pdadd.err) | pca
      })
    }

    expect_error(rxode2(pk.turnover.emax4))

    pk.turnover.emax4 <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        poplogit <- 2
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        eta.emax ~ .5
        eta.ec50 ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        ##
        # poplogit = log(temax/(1-temax))
        logit <- exp(poplogit + eta.emax)
        # logit=temax+eta.emax
        emax <- logit / (1 + logit)
        ec50 <- exp(tec50 + eta.ec50)
        kout <- exp(tkout + eta.kout)
        e0 <- exp(te0 + eta.e0)
        ##
        DCP <- center / v
        PD <- 1 - emax * DCP / (ec50 + DCP)
        ##
        effect(0) <- e0
        kin <- e0 * kout
        ##
        d / dt(depot) <- -ktr * depot
        d / dt(gut) <- ktr * depot - ka * gut
        d / dt(center) <- ka * gut - cl / v * center
        d / dt(effect) <- kin * PD - kout * effect
        ##
        cp <- center / v
        ## Who would use this...
        log(cp) ~ prop(prop.err) + add(pkadd.err)
        effect * 1 / (1 + cp) ~ add(pdadd.err) | pca
      })
    }

    ##   Error: multiple compartment models with expressions need to be conditioned by `|`
    ## ie log(cp) ~ add(err) | cmt
    ## The following endpoints need to be corrected: log(cp)

    expect_error(rxode2(pk.turnover.emax4))

    pk.turnover.emax4 <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        poplogit <- 2
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        eta.emax ~ .5
        eta.ec50 ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        ##
        # poplogit = log(temax/(1-temax))
        logit <- exp(poplogit + eta.emax)
        # logit=temax+eta.emax
        emax <- logit / (1 + logit)
        ec50 <- exp(tec50 + eta.ec50)
        kout <- exp(tkout + eta.kout)
        e0 <- exp(te0 + eta.e0)
        ##
        DCP <- center / v
        PD <- 1 - emax * DCP / (ec50 + DCP)
        ##
        effect(0) <- e0
        kin <- e0 * kout
        ##
        d / dt(depot) <- -ktr * depot
        d / dt(gut) <- ktr * depot - ka * gut
        d / dt(center) <- ka * gut - cl / v * center
        d / dt(effect) <- kin * PD - kout * effect
        ##
        cp <- center / v
        ## Who would use this...
        log(cp) ~ prop(prop.err) + add(pkadd.err) | center
        log(pca) ~ add(pdadd.err) | cmt
      })
    }

    ## Error in rxModelVars_(obj) :
    ## Evaluation error: 'cmt' cannot be a state or lhs expression.

    expect_error(rxode2(pk.turnover.emax4))

  })
})

rxTest({
  one.cmt <- function() {
    ini({
      tka <- 0.45
      tcl <- log(c(0, 2.7, 100))
      tv <- 3.45
      eta.ka ~ 0.6
      eta.cl ~ 0.3
      eta.v ~ 0.1
      add.sd <- 0.7
    })
    model({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      linCmt() ~ add(add.sd)
    })
  }

  skip_if_not_installed("nlmixr2data")
  d <- nlmixr2data::theo_sd
  d$SEX <- ifelse(d$ID < 7, "M", "F")
  d$fSEX <- factor(d$SEX)

  simOrig <- rxSolve(one.cmt, events = d, keep = c("SEX", "fSEX"))
  expect_error({
    simOrig$eta.v <- NULL
  }, NA)

  expect_true(inherits(simOrig, "data.frame"))
  expect_false(inherits(simOrig, "rxSolve"))
  expect_false(any(names(simOrig) == "eta.v"))

  expect_error({
      simOrig[["eta.v"]] <- NULL
  }, NA)

  expect_true(inherits(simOrig, "data.frame"))
  expect_false(inherits(simOrig, "rxSolve"))
  expect_false(any(names(simOrig) == "eta.v"))

})

rxTest({
  # Individual keep AGE==AGE2
  if (!.Call(`_rxode2_isIntel`)) {

    test_that("Make sure the keep gives the right values", {
      TVQ <- 4
      TVV3 <- 7
      dat <- data.frame(AGE = c(20, 30), ID = c(10, 20)) |>
        dplyr::mutate(id = seq(from = 1, to = dplyr::n())) |>
        dplyr::rename(NMID = ID)


      par.tab <- data.frame(
        ThetaKa = c(0.7, 0.9),
        ThetaCl = c(4.6, 5.4),
        ThetaV2 = c(7, 8),
        ID = c(10, 20)
      ) |>
        dplyr::mutate(id = seq(from = 1, to = dplyr::n())) |>
        dplyr::mutate(ThetaQ = TVQ, ThetaV3 = TVV3) |>
        dplyr::rename(NMID = ID)

      tabtot <- dat |>
        dplyr::left_join(par.tab, by = c("NMID", "id"))

      mod1 <- rxode2({
        ## PK parameters
        CL <- exp(ThetaCl)
        KA <- exp(ThetaKa)
        V2 <- exp(ThetaV2)
        Q <- exp(ThetaQ)
        V3 <- exp(ThetaV3)

        K20 <- CL / V2
        K23 <- Q / V2
        K32 <- Q / V3

        CP <- A2 / V2

        ##
        d / dt(A1) <- -KA * A1
        d / dt(A2) <- KA * transit3 - K23 * A2 + K32 * A3 - K20 * A2
        d / dt(A3) <- K23 * A2 - K32 * A3

        d / dt(transit1) <- KA * A1 - KA * transit1
        d / dt(transit2) <- KA * transit1 - KA * transit2
        d / dt(transit3) <- KA * transit2 - KA * transit3

        f(A1) <- 1

        d / dt(AUC) <- CP
        A1(0) <- 0
        A2(0) <- 0
        A3(0) <- 0

        AGE2 <- AGE
      })

      NSubj <- length(tabtot$id)

      dose_ref <- 8000 ##

      ev_ref <- eventTable() |>
        et(dose = dose_ref / 1000, time = seq(0, 24, 24)) |>
        et(id = 1:NSubj) |>
        add.sampling(seq(0, 24, 1)) |>
        dplyr::mutate(DOSE = dose_ref) |>
        dplyr::group_by(id) |>
        tidyr::fill(DOSE, .direction = "downup") |>
        dplyr::ungroup() |>
        dplyr::mutate(Cycle = dplyr::case_when(
          time <= 12 ~ 1, #
          time >= 12 ~ 2, #
          TRUE ~ 0
        )) |>
        dplyr::as_tibble()

      ev_ref <- ev_ref |>
        dplyr::left_join(tabtot, by = "id") |>
        dplyr::as_tibble()

      PK.ev_ref2 <- rxSolve(mod1,
                            events = ev_ref, cores = 2,
                            seed = 123, addCov = TRUE, keep = c("Cycle", "AGE")
                            )

      expect_equal(PK.ev_ref2$AGE, PK.ev_ref2$AGE2)
    })

    test_that("rxSolve 'keep' maintains character output (#190/#622)", {

      skip_if_not_installed("nlmixr2data")
      one.cmt <- function() {
        ini({
          tka <- 0.45
          tcl <- log(c(0, 2.7, 100))
          tv <- 3.45
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.sd <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          linCmt() ~ add(add.sd)
        })
      }

      d <- nlmixr2data::theo_sd
      d$SEX <- ifelse(d$ID < 7, "M", "F")
      d$fSEX <- factor(d$SEX)
      d$oSEX <- d$fSEX
      class(d$oSEX) <- c("ordered", "factor")
      d$iSEX <- as.integer(d$fSEX)
      d$lSEX <- as.logical(d$iSEX == 1)
      d$dSEX <- d$iSEX + 0.5
      library(units)
      d$uSEX <- set_units(d$dSEX, kg)
      d$eSEX <- lapply(d$SEX, function(e) {
        str2lang(e)
      })

      sim <- rxSolve(one.cmt, events = d, keep = c("SEX", "fSEX", "iSEX", "dSEX", "oSEX", "uSEX", "lSEX"))

      expect_type(sim$SEX, "character")
      expect_s3_class(sim$fSEX, "factor")
      expect_equal(levels(sim$fSEX), c("F", "M"))
      expect_type(sim$iSEX, "integer")
      expect_type(sim$dSEX, "double")
      expect_true(inherits(sim$oSEX, "ordered"))
      expect_true(inherits(sim$uSEX, "units"))
      expect_true(is.logical(sim$lSEX))

      d <- nlmixr2data::theo_sd
      d$SEX <- ifelse(d$ID < 7, "M", "F")
      d$SEX[4] <- NA_character_
      d$fSEX <- factor(d$SEX)
      d$fSEX[4] <- NA_integer_
      d$iSEX <- as.integer(d$fSEX)
      d$iSEX[4] <- NA_integer_
      d$lSEX <- as.logical(d$iSEX == 1)
      d$lSEX[4] <- NA
      d$dSEX <- d$iSEX + 0.5
      d$eSEX <- lapply(d$SEX, function(e) {
        str2lang(e)
      })

      sim <- rxSolve(one.cmt, events = d, keep = c("SEX", "fSEX", "iSEX", "dSEX", "lSEX"))

      expect_true(is.na(d$SEX[4]))
      expect_true(is.na(d$fSEX[4]))
      expect_true(is.na(d$iSEX[4]))
      expect_true(is.na(d$lSEX[4]))

      expect_error(rxSolve(one.cmt, events = d, keep = c("eSEX")))

    })

    test_that("rxSolve 'keep' does not crash and keeps correct values #756", {

      qs <- test_path("keep-756.rds")
      skip_if_not(file.exists(qs), "Test file not found")

      d <- readRDS(qs)

      mod <- function() {
        ini({
          ka_anon1 <- 1
          f_anon1 <- 0.8
          vc_anon1 <- 3
          hl_anon1 <- 5
        })
        model({
          mw_anon1 <- 55000
          mw_convert_anon1 <- 1 / mw_anon1 * 1e3
          kel_anon1 <- log(2)/(hl_anon1/60/24)
          kel_target <- log(2)/1.2
          kform_target <- 1.4*kel_target
          kd_anon1_target_umolL <- 1.5/1000
          d/dt(depot_anon1) <- -ka_anon1*depot_anon1
          d/dt(central_anon1) <- ka_anon1*depot_anon1 - kel_anon1*central_anon1
          central_anon1_umolL <- central_anon1/vc_anon1*mw_convert_anon1
          totalconc <- central_anon1_umolL + central_target + kd_anon1_target_umolL
          bound_umolL <- (totalconc - sqrt(totalconc^2 - 4*central_anon1_umolL*central_target))/2
          free_central_target <- central_target - bound_umolL
          d/dt(central_target) <- kform_target - kel_target*free_central_target - kel_anon1*bound_umolL - 1.1 * (1.3*central_target - peripheral_target)
          d/dt(peripheral_target) <- 1.1 * (1.3*central_target - peripheral_target)
          d/dt(cleared_amount_bound_anon1_umol) <- kel_anon1*bound_umolL
          f(depot_anon1) <- f_anon1
          central_target(0) <- 1.4
          peripheral_target(0) <- 1.4*1.3
        })
      }

      expect_error(rxSolve(mod, d, keep="target_name"), NA)

      tmp <- rxSolve(mod, d, keep="target_name", addDosing=FALSE)

      expect_false(any(is.na(tmp$target_name)))

      tmp <- rxSolve(mod, d, keep="target_name", addDosing=TRUE, keepInterpolation="na")

      expect_true(any(is.na(tmp$target_name)))

      tmp <- rxSolve(mod, d, keep="target_name", addDosing=TRUE, keepInterpolation="locf")

      expect_false(any(is.na(tmp$target_name)))

      tmp <- rxSolve(mod, d, keep="target_name", addDosing=TRUE, keepInterpolation="nocb")

      expect_false(any(is.na(tmp$target_name)))


      ## print(head(tmp[,c("id", "amt", "target_name")]))

      et <- etTrans(d, mod, keep="target_name")
      et2 <- attr(class(et), ".rxode2.lst")
      class(et2) <- NULL

      expect_equal(length(et2$keepL$keepL[[1]]),
                   length(et$ID))

      class(et) <- "data.frame"
      et <- cbind(et, k=et2$keepL$keepL[[1]])

      d2 <- d |> dplyr::filter(EVID==1) |> dplyr::select(ID, TIME) |>
        dplyr::mutate(i=1)

      # only variables in the dataset are considered NA
      expect_true(merge(et, d2, all.x=TRUE) |>
                    dplyr::filter(i!= 1) |> dplyr::pull(k) |>
                    is.na() |> all())

      ## s <- rxSolve(mod, d, keep="target_name")

    })
  }

  test_that("keep column with leading/trailing NA for the first subject (no OOB)", {
    ## Before the fix, get_fkeep's LOCF/NOCB scan could exit at i = fid-1 (reads
    ## vals[-1] for the first subject) or i = fid + n_all_times (past the id
    ## block) and dereference that out-of-bounds element.  Here the first
    ## subject's leading/trailing keep values are NA, forcing the scan off the
    ## end of the id block.
    mod <- rxode2({
      d/dt(center) <- -kel * center
      cp <- center / v
    })
    d <- data.frame(ID = 1L, TIME = 0:3, AMT = c(100, 0, 0, 0),
                    EVID = c(1L, 0L, 0L, 0L), WT = c(NA, NA, 70, 72))
    s <- rxSolve(mod, d, params = c(kel = 0.2, v = 10),
                 keep = "WT", keepInterpolation = "locf")
    ## leading NA at the first observation forward-fills to the first available
    ## value (70), not garbage read from vals[-1]
    expect_equal(s$WT, c(70, 70, 72))

    d2 <- data.frame(ID = 1L, TIME = 0:3, AMT = c(100, 0, 0, 0),
                     EVID = c(1L, 0L, 0L, 0L), WT = c(70, 72, NA, NA))
    s2 <- rxSolve(mod, d2, params = c(kel = 0.2, v = 10),
                  keep = "WT", keepInterpolation = "nocb")
    ## trailing NA falls back to the last available value (72)
    expect_equal(s2$WT, c(72, 72, 72))
  })
})

rxTest({

  test_that("test +dv()", {

    one.cmt <- function() {
      ini({
        ## You may label each parameter with a comment
        tka <- 0.45 # Log Ka
        tcl <- log(c(0, 2.7, 100)) # Log Cl
        ## This works with interactive models
        ## You may also label the preceding line with label("label text")
        tv <- 3.45; label("log V")
        ## the label("Label name") works with all models
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        cp <- linCmt()
        cp ~ lnorm(add.sd)
      })
    }
    one.cmt <- one.cmt()

    v <- .handleSingleErrTypeNormOrTFoceiBase(one.cmt, one.cmt$predDf[1,])

    expect_true(identical(v[[6]],
                          quote(rx_pred_ ~ log(rx_pred_f_))))

    expect_true(identical(v[[1]],
                          quote(rx_yj_ ~ 3)))

    one.cmt <- one.cmt |>
      model(cp ~ lnorm(add.sd) + dv())

    v <- .handleSingleErrTypeNormOrTFoceiBase(one.cmt, one.cmt$predDf[1,])

    expect_true(identical(v[[6]],
                          quote(rx_pred_ ~ rx_pred_f_)))

    expect_true(identical(v[[1]],
                          quote(rx_yj_ ~ 3)))

    one.cmt <- one.cmt |>
      model(cp ~ logitNorm(add.sd))

    v <- .handleSingleErrTypeNormOrTFoceiBase(one.cmt, one.cmt$predDf[1,])

    expect_true(identical(v[[1]],
                          quote(rx_yj_ ~ 4)))

    expect_true(identical(v[[6]],
                          quote(rx_pred_ ~ rxTBS(rx_pred_f_, rx_lambda_, rx_yj_, rx_low_, rx_hi_))))


    one.cmt <- one.cmt |>
      model(cp ~ logitNorm(add.sd) + dv())

    v <- .handleSingleErrTypeNormOrTFoceiBase(one.cmt, one.cmt$predDf[1,])

    expect_true(identical(v[[6]],
                          quote(rx_pred_ ~ rx_pred_f_)))

    expect_true(identical(v[[1]],
                          quote(rx_yj_ ~ 4)))

  })


})

rxTest({
  test_that("rxStack preserves factors", {
    pheno2 <- function() {
      ini({
        tcl <- log(0.008)
        tv <-  log(0.6)
        eta.cl + eta.v ~ c(1,
                           0.01, 1)
        add.err <- 0.1
      })
      model({
        cl <- exp(tcl + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v)    # individual value of volume
        ke <- cl / v            # elimination rate constant
        d/dt(A1) = - ke * A1    # model differential equation
        cp = A1 / v             # concentration in plasma
        cp ~ add(add.err)  # define error model
      })
    }

    simdata <- data.frame(time=1:10, ID=factor(c("A", "B")))

    sim <- rxSolve(pheno2, events=simdata)

    expect_equal(levels(sim$id), c("A", "B"))
  })
})

rxTest({
  test_that("Now test fix and unfixing", {
    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl + eta.v ~ c(0.3,
                           0.01, 0.1)
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        f(depot) <- 3
        cp <- center / v
        cp2 <- cp * cl
        cp ~ add(add.err)
      })
    }

    suppressMessages(
      f2 <- one.compartment |> ini(tka=fix)
    )
    w <- which(f2$iniDf$name == "tka")
    expect_true(f2$iniDf$fix[w])

    suppressMessages(
      f3 <- f2 |> ini(tka=unfix)
    )
    w <- which(f3$iniDf$name == "tka")
    expect_false(f3$iniDf$fix[w])

    suppressMessages(
      f2 <- one.compartment |> ini(tka=fix(3))
    )
    w <- which(f2$iniDf$name == "tka")
    expect_true(f2$iniDf$fix[w])

    suppressMessages(
      f3 <- f2 |> ini(tka=unfix(3))
    )
    w <- which(f3$iniDf$name == "tka")
    expect_false(f3$iniDf$fix[w])

    # should fix the entire block
    suppressMessages(
      f2 <- one.compartment |> ini(eta.cl=fix)
    )
    expect_equal(f2$iniDf[!is.na(f2$iniDf$neta1), "fix"], c(FALSE, TRUE, TRUE, TRUE))

    # should unfix the entire block
    suppressMessages(
      f3 <- f2 |> ini(eta.cl=unfix)
    )
    expect_equal(f3$iniDf[!is.na(f3$iniDf$neta1), "fix"], c(FALSE, FALSE, FALSE, FALSE))

    suppressMessages(
      f2 <- one.compartment |> ini(fix(eta.cl))
    )
    expect_equal(f2$iniDf[!is.na(f2$iniDf$neta1), "fix"], c(FALSE, TRUE, TRUE, TRUE))

    suppressMessages(
      f3 <- f2 |> ini(unfixed(eta.cl))
    )
    expect_equal(f3$iniDf[!is.na(f3$iniDf$neta1), "fix"], c(FALSE, FALSE, FALSE, FALSE))
  })
})

rxTest({

  test_that("bad ui", {

    f <- function() {
      ini({
        sd <- 1
      })
      model({
        err ~ add(sd)
      })
    }
    expect_error(f(), "model")

  })
  test_that("issue nlmixr#501", {

    nlmixr_threecmt_mm_no_add_wtcl_pdtg_kout_delay <- function() {
      ini({
        tf_sc <- log(999)
        tf_infilt <- log(999)
        tka_sc <- log(999)
        tka_infilt <- log(999)
        tcl_low <- log(999)
        tcl_high <- log(999)
        tcl_c50 <- log(3000)
        e_wt_cl <- fixed(999)
        tv <- log(999)
        tq1 <- log(999)
        tvp1 <- log(10)
        tq2 <- log(999)
        tvp2 <- log(20)
        eta_cl~999
        eta_v~999
        prop_err <- 999

        tg_bl <- log(999)
        eta_tg_bl~999
        tg_kel <- log(999)
        tg_ec50 <- log(5000)
        tg_emax_kel <- log(2)
        ktr_tg <- log(999)
        prop_err_tg <- 999
      })
      model({
        # PK setup
        f_sc <- exp(tf_sc)
        f_infilt <- exp(tf_infilt)
        ka_sc <- exp(tka_sc)
        ka_infilt <- exp(tka_infilt)
        cl_low <- exp(tcl_low + eta_cl)*(WEIGHT_BL/85)^e_wt_cl
        cl_high <- exp(tcl_high + eta_cl)*(WEIGHT_BL/85)^e_wt_cl
        cl_c50 <- exp(tcl_c50)
        v <- exp(tv + eta_v)
        q1 <- exp(tq1)
        vp1 <- exp(tvp1)
        q2 <- exp(tq2)
        vp2 <- exp(tvp2)

        # PK micro-parameters
        ke_low <- cl_low/v
        ke_high <- cl_high/v
        kc_p1 <- q1/v
        kp1_c <- q1/vp1
        kc_p2 <- q2/v
        kp2_c <- q2/vp2

        # differential equations
        cp <- CENTRAL/v*1e3 # 1e3 is for unit conversion
        ke <- ke_low + (ke_high - ke_low)*cp/(cp + cl_c50)
        d/dt(IVINFILT) =             - ka_infilt * IVINFILT
        d/dt(SC)       = -ka_sc * SC
        d/dt(CENTRAL)  =  ka_sc * SC + ka_infilt * IVINFILT - ke*CENTRAL - kc_p1*CENTRAL + kp1_c*P1 - kc_p2*CENTRAL + kp2_c*P2
        d/dt(P1) =                                                         kc_p1*CENTRAL - kp1_c*P1
        d/dt(P2) =                                                                                    kc_p2*CENTRAL - kp2_c*P2

        f(SC) <- f_sc
        f(IVINFILT) <- f_infilt

        # TG Emax model
        tgbl <- exp(tg_bl + eta_tg_bl)
        kin_tg <- tgbl*exp(tg_kel)
        TG(0) <- tgbl
        ktr_TG <- exp(ktr_tg)
        d/dt(TG_TR) = ktr_tg*cp - ktr_tg*TG_TR
        kout_tg <- exp(tg_kel) + exp(tg_emax_kel)*TG_TR/(TG_TR + exp(tg_ec50))
        d/dt(TG) = kin_tg - kout_tg*TG

        # Residual error models
        cp ~ prop(prop_err)
        TG ~ prop(prop_err_tg)
      })
    }

    expect_error(rxode2(nlmixr_threecmt_mm_no_add_wtcl_pdtg_kout_delay), NA)

    one.compartment <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
      keep = "WT"
      drop = "depot"
    }

    expect_warning(rxode2(one.compartment))

    one.compartment <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
      keep = "WT"
      drop = "depot"
    }

    expect_error(rxode2(one.compartment), "ini")

    one.compartment <- function() {
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        add.sd <- 4
        cp ~ add(add.sd)
      })
      keep = "WT"
      drop = "depot"
    }

    expect_warning(
      expect_error(rxode2(one.compartment), NA),
      regexp = "function cannot be called directly to produce model object"
    )

    one.compartment <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      keep = "WT"
      drop = "depot"
    }

    expect_error(
      rxode2(one.compartment),
      regexp = "rxode2 model function requires one 'model({})' block",
      fixed = TRUE
    )


    one.compartment <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
      keep = "WT"
      drop = "depot"
    }

    expect_error(
      rxode2(one.compartment),
      regexp = "rxode2 model function requires one 'model({})' block",
      fixed = TRUE
    )
  })

  test_that("Duplicate parameters raise errors", {

    uif <- function() {
      ini({
        lCL <- 1.37
        lV <- 4.19
        lCLD <- 1.37
        lVT <- 3.87
        prop.err <- 1
        eta.Cl ~ 0.1
        eta.V ~ 0.1
        ## Duplicate CLs
        eta.Cl ~ 0.1
        eta.VT ~ 0.1
      })
      model({
        CL <- exp(lCL + eta.Cl)
        V <- exp(lV + eta.V)
        CLD <- exp(lCLD + eta.Cl)
        VT <- exp(lVT + eta.VT)
        linCmt() ~ prop(prop.err)
      })
    }

    expect_error(rxode2(uif), rex::rex("duplicated parameter(s): 'eta.Cl'"))
  })

  test_that("Un-estimated paramteres raise errors", {

    uif.ode <- function() {
      ini({
        lCL <- 1.37
        lV <- 4.19
        lCLD <- 1.37
        lVT <- 3.87
        ## Prop error isn't estimated
        prop.err <- 1
        add.err <- 0.1
        eta.Cl + eta.V ~ c(
          0.1,
          0.01, 0.01
        )
      })
      model({
        CL <- exp(lCL + eta.Cl)
        V <- exp(lV + eta.V)
        CLD <- exp(lCLD)
        VT <- exp(lVT)
        K10 <- CL / V
        K12 <- CLD / V
        K21 <- CLD / VT
        d / dt(centr) <- K21 * periph - K12 * centr - K10 * centr
        d / dt(periph) <- -K21 * periph + K12 * centr
        cp <- centr / V
        cp ~ add(add.err)
      })
    }

    expect_error(rxode2(uif.ode), rex::rex("the following parameter(s) were in the ini block but not in the model block: prop.err"))

    uif <- function() {
      ini({
        tka <- exp(0.5)
        tcl <- exp(-3.2)
        tv <- exp(1)
        eta.ka ~ 0.1
        ## Should be eta.cl
        eta.v ~ 0.2
        add.err ~ 0.1
      })
      model({
        ka <- tka + eta.ka
        cl <- tcl + eta.cl
        v <- tv
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    expect_error(rxode2(uif), rex::rex("endpoint parameter(s) missing, duplicated, or defined with '~'"))


    uif <- function() {
      ini({
        tka <- exp(0.5)
        tcl <- exp(-3.2)
        tv <- exp(1)
        eta.ka ~ 0.1
        ## Should be eta.cl
        eta.v ~ 0.2
        add.err <- 0.1
      })
      model({
        ka <- tka + eta.ka
        cl <- tcl + eta.cl
        v <- tv
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    expect_warning(
      expect_error(rxode2(uif), rex::rex("the following parameter(s) were in the ini block but not in the model block: eta.v")),
      regexp = "some etas defaulted to non-mu referenced"
    )
  })

  test_that("Residuals are population parameters", {

    uif <- function() {
      ini({
        tka <- exp(0.5)
        tcl <- exp(-3.2)
        tv <- exp(1)
        eta.ka ~ 0.1
        eta.cl ~ 0.2
        add.err ~ 0.1
      })
      model({
        ka <- tka + eta.ka
        cl <- tcl + eta.cl
        v <- tv
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    expect_error(rxode2(uif), rex::rex("endpoint parameter(s) missing, duplicated, or defined with '~'"))

  })

  test_that("Parameters need to be named", {

    uif <- function() {
      ini({
        tka <- exp(0.5)
        tcl <- exp(-3.2)
        tv <- exp(1)
        eta.ka ~ 0.1
        eta.cl ~ 0.2
        ## Should be assign since it is a THETa, should I support it....?
        0.1
      })
      model({
        ka <- tka + eta.ka
        cl <- tcl + eta.cl
        v <- tv
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }
    ## , rex::rex("The following THETAs are unnamed: THETA[4]")

    expect_error(
      expect_message(
        rxode2(uif),
        regexp="bad matrix specification"
      ),
      regexp="lotri syntax errors above"
    )

    uif <- function() {
      ini({
        tka <- exp(0.5)
        tcl <- exp(-3.2)
        tv <- exp(1)
        eta.ka ~ 0.1
        ~0.2
        ## Should be assign since it is a THETa, should I support it....?
        add.err <- 0.1
      })
      model({
        ka <- tka + eta.ka
        cl <- tcl + eta.cl
        v <- tv
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    ## rex::rex("The following ETAs are unnamed: ETA[2]")
    expect_error(
      expect_message(
        rxode2(uif),
        rex::rex("matrix expression should be 'name ~ c(lower-tri)'")
      ),
      regexp="lotri syntax errors above"
    )

  })

  test_that("Parameters cannot be missing or Infinite", {

    uif <- function() {
      ini({
        tka <- 1 / 0
        tcl <- exp(-3.2)
        tv <- exp(1)
        eta.ka ~ 0.1
        eta.cl ~ 0.2
        add.err <- 4
      })
      model({
        ka <- tka + eta.ka
        cl <- tcl + eta.cl
        v <- tv
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    expect_error(rxode2(uif), rex::rex("infinite/NA initial parameters: tka"))

    uif <- function() {
      ini({
        tka <- NA
        tcl <- exp(-3.2)
        tv <- exp(1)
        eta.ka ~ 0.1
        eta.cl ~ 0.2
        add.err <- 4
      })
      model({
        ka <- tka + eta.ka
        cl <- tcl + eta.cl
        v <- tv
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    expect_error(
      expect_message(
        rxode2(uif),
        regexp=rex::rex("estimate syntax unsupported: tka <- NA")
      ),
      regexp="lotri syntax errors above"
    )

    uif <- function() {
      ini({
        tka <- 3
        tcl <- exp(-3.2)
        tv <- exp(1)
        eta.ka ~ 0.1
        eta.cl ~ 0.2
      })
      model({
        ka <- tka + eta.ka
        cl <- tcl + eta.cl
        v <- tv
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    expect_error(rxode2(uif), rex::rex("endpoint 'cp' needs the following parameters estimated or modeled"))

    uif <- function() {
      ini({
        tka <- 3
        tcl <- exp(-3.2)
        tv <- exp(1)
        eta.ka ~ 0.1
        eta.cl ~ 0.2
      })
      model({
        ka <- tka + eta.ka
        cl <- tcl + eta.cl
        v <- tv
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        add.err <- ka + cl
        cp ~ add(add.err)
      })
    }

    expect_error(rxode2(uif), NA)

  })


  test_that("modeled endpoints", {
    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        eta.v ~ 0.01
        add.sd <- 0.7
        add.sd2 <- 0.7
        tprop <- 0.5
        prop.eta ~ 0.01
      })
      model({
        ka <- tka
        cl <- tcl
        v <- eta.v
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        prop.sd <- exp(tprop + prop.eta)
        cp ~ add(add.sd)
        cp2 ~ add(add.sd2)
      })
    }

    expect_warning(
      expect_error(
        ocmt(),
        regexp = "endpoint 'cp2' is not defined in the model"
      ),
      regexp = "some etas defaulted to non-mu referenced"
    )
  })

  test_that("if/else in endpoints should error", {
    
    myModel_OS <- function(){           # TTE model for OS. Weibull baseline hazard.
      ini({

        tLAM <- log(0.001)               # scale parameter
        tSHP <- log(2)                   # shape parameter
        tBSLD0 <- log(0.1)               # parameter for SLD (sum of longest diameter) at enrolment
        tBTSR <- log(0.1)                # parameter for TSR(t) (tumour size ratio)
        tBNWLS <- log(0.1)               # parameter for NewLesion(t)
        tBECOG <- log(0.1)               # parameter for ECOG at enrolment
        eta.LAM ~ 0.01                   # inter-individual variability in scale parameter
        add.err <- 0

      })
      model({

        LAM = exp(tLAM + eta.LAM)
        SHP = exp(tSHP)
        BSLD0 = exp(tBSLD0)
        BTSR = exp(tBTSR)
        BNWLS = exp(tBNWLS)
        BECOG = exp(tBECOG)

        #Time constant covariates
        TVSLD0 = 70                      # average SLD at enrolment
        NSLD0 = SLD0/TVSLD0              # normalised SLD at enrolment
        IECOG=ECOG                       # ECOG at enrolment

        # Model for dSLD(t)
        A(0) = IBASE*1000                # SLD baseline
        MMBAS = IBASE*1000               # Parameter of the SLD(t) model were estimated previously (IPP approach)
        d/dt(A) = KG/1000 * A - (KD0/1000 * AUC0 + + KD1/100 * AUC1) * A
        TUM = A
        TSR = (TUM-MMBAS)/MMBAS

        if(time==0){                     #TSR(t) for t<= week 12 and TSR(week12) for t> week 12
          WTS = 0
          TM12 = 0
        }
        if(time<=84){
          WTS = TSR
          TM12 = WTS
        }
        else {
          WTS = WTS
        }

        # survival model
        DEL = 1E-6
        d/dt(A2) = LAM*SHP*(LAM*(T+DEL))**(SHP-1) *exp(BSLD0*NSLD0+BTSR*WTS+BNWLS*NWLS+BECOG*IECOG)

        # Death hazard
        CHZ = A2
        SUR=exp(-CHZ)
        DELX = 1E-6
        xTUM = A
        xTSR = (xTUM-MMBAS)/MMBAS
        if(time==0){
          XWTS=0
          XTM12=0
        }

        if(time<=84){
          XWTS = XTSR
          XTM12 = XWTS
        }
        else {
          XWTS=XWTS
        }

        HAZN = LAM*SHP*(LAM*(TIME+DELx))**(SHP-1)*exp(BSLD0*NSLD0+BTSR*XWTS+BNWLS*NWLS+BECOG*IECOG)
        # prediction
        if(FLAG == 9 & EVID == 0 & OSCENS == 1){
          IPRED=SUR                      # probability of survival (censored event)
          Y = IPRED                      # Y is probability for TTE data
          Y ~ add(add.err)
        }
        if(FLAG == 9 & EVID == 0 & OSCENS == 0){
          IPRED=SUR*HAZN                # probability of event (death) at time=TIME
          Y = IPRED                     # Y is probability for TTE data
        }
        Y ~ add(add.err)
      })
    }
    
    expect_error(suppressMessages(rxode2(myModel_OS)))
    
    
    expect_error(suppressMessages(myModel_OS()))
  })
})

rxTest({

  ## Reading materialized linCmt() compartment amounts (central / peripheral1 /
  ## peripheral2 / depot) in model equations *under an error model*.  An
  ## endpoint injects an observation compartment that used to shift the linCmt
  ## compartments' solved-state indices, so an in-equation reference read an
  ## unwritten slot (== 0).  The __DDT declaration-index fix makes these resolve
  ## to the solved amount.  Each case is checked against the equivalent explicit
  ## ODE solved with useLinCmt = FALSE (the analytic and numeric solutions must
  ## agree), and the referenced columns must not be identically zero (the bug).

  .tol <- 1e-4
  .chk <- function(lin, ode, ev, cols) {
    .rL <- suppressMessages(rxSolve(lin, ev))
    .rO <- suppressMessages(rxSolve(ode, ev, useLinCmt = FALSE))
    for (.c in cols) {
      expect_true(.c %in% names(.rL), info = paste0(.c, " present in linCmt solve"))
      expect_false(all(.rL[[.c]] == 0),
                   label = paste0(.c, " must not be identically zero (the bug)"))
      expect_equal(.rL[[.c]], .rO[[.c]], tolerance = .tol,
                   label = paste0(.c, " (linCmt) vs ", .c, " (ODE)"))
    }
  }
  .evIV <- et(amt = 100, cmt = "central") |> et(seq(0, 48, by = 2))
  .evPO <- et(amt = 100, cmt = "depot")   |> et(seq(0, 48, by = 2))

  test_that("2-cmt IV: peripheral1 amount and concentration resolve under an error model", {
    .lin <- suppressMessages(rxode2(function() {
      ini({ lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50); pSd<-0.1 })
      model({
        cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp)
        Cc  <- linCmt()
        Ap1 <- peripheral1        # solved peripheral amount
        Cp1 <- peripheral1 / vp   # peripheral concentration
        Cc ~ prop(pSd)
      })
    }))
    .ode <- suppressMessages(rxode2(function() {
      ini({ lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50) })
      model({
        cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp)
        kel<-cl/vc; k12<-q/vc; k21<-q/vp
        d/dt(central)     <- -kel*central - k12*central + k21*peripheral1
        d/dt(peripheral1) <-  k12*central - k21*peripheral1
        Cc<-central/vc; Ap1<-peripheral1; Cp1<-peripheral1/vp
      })
    }))
    .chk(.lin, .ode, .evIV, c("Cc", "Ap1", "Cp1"))
  })

  test_that("2-cmt IV: central amount resolves under an error model", {
    .lin <- suppressMessages(rxode2(function() {
      ini({ lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50); pSd<-0.1 })
      model({
        cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp)
        Cc <- linCmt()
        Ac <- central             # solved central amount
        Cc ~ prop(pSd)
      })
    }))
    .ode <- suppressMessages(rxode2(function() {
      ini({ lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50) })
      model({
        cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp)
        kel<-cl/vc; k12<-q/vc; k21<-q/vp
        d/dt(central)     <- -kel*central - k12*central + k21*peripheral1
        d/dt(peripheral1) <-  k12*central - k21*peripheral1
        Cc<-central/vc; Ac<-central
      })
    }))
    .chk(.lin, .ode, .evIV, c("Cc", "Ac"))
  })

  test_that("3-cmt IV: peripheral1 and peripheral2 resolve under an error model", {
    .lin <- suppressMessages(rxode2(function() {
      ini({ lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50); lq2<-log(1); lvp2<-log(100); pSd<-0.1 })
      model({
        cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp); q2<-exp(lq2); vp2<-exp(lvp2)
        Cc  <- linCmt()
        Cp1 <- peripheral1 / vp
        Cp2 <- peripheral2 / vp2
        Cc ~ prop(pSd)
      })
    }))
    .ode <- suppressMessages(rxode2(function() {
      ini({ lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50); lq2<-log(1); lvp2<-log(100) })
      model({
        cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp); q2<-exp(lq2); vp2<-exp(lvp2)
        kel<-cl/vc; k12<-q/vc; k21<-q/vp; k13<-q2/vc; k31<-q2/vp2
        d/dt(central)     <- -kel*central - k12*central + k21*peripheral1 - k13*central + k31*peripheral2
        d/dt(peripheral1) <-  k12*central - k21*peripheral1
        d/dt(peripheral2) <-  k13*central - k31*peripheral2
        Cc<-central/vc; Cp1<-peripheral1/vp; Cp2<-peripheral2/vp2
      })
    }))
    .chk(.lin, .ode, .evIV, c("Cc", "Cp1", "Cp2"))
  })

  test_that("2-cmt oral: depot amount and peripheral1 resolve under an error model", {
    .lin <- suppressMessages(rxode2(function() {
      ini({ lka<-log(0.8); lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50); pSd<-0.1 })
      model({
        ka<-exp(lka); cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp)
        Cc   <- linCmt()
        Adep <- depot             # depot amount
        Cp1  <- peripheral1 / vp
        Cc ~ prop(pSd)
      })
    }))
    .ode <- suppressMessages(rxode2(function() {
      ini({ lka<-log(0.8); lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50) })
      model({
        ka<-exp(lka); cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp)
        kel<-cl/vc; k12<-q/vc; k21<-q/vp
        d/dt(depot)       <- -ka*depot
        d/dt(central)     <-  ka*depot - kel*central - k12*central + k21*peripheral1
        d/dt(peripheral1) <-  k12*central - k21*peripheral1
        Cc<-central/vc; Adep<-depot; Cp1<-peripheral1/vp
      })
    }))
    .chk(.lin, .ode, .evPO, c("Cc", "Adep", "Cp1"))
  })

  test_that("3-cmt oral: depot, central, peripheral1 and peripheral2 all resolve together", {
    .lin <- suppressMessages(rxode2(function() {
      ini({ lka<-log(0.8); lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50)
            lq2<-log(1); lvp2<-log(100); pSd<-0.1 })
      model({
        ka<-exp(lka); cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp); q2<-exp(lq2); vp2<-exp(lvp2)
        Cc   <- linCmt()
        Adep <- depot
        Ac   <- central
        Cp1  <- peripheral1 / vp
        Cp2  <- peripheral2 / vp2
        Cc ~ prop(pSd)
      })
    }))
    .ode <- suppressMessages(rxode2(function() {
      ini({ lka<-log(0.8); lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50)
            lq2<-log(1); lvp2<-log(100) })
      model({
        ka<-exp(lka); cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp); q2<-exp(lq2); vp2<-exp(lvp2)
        kel<-cl/vc; k12<-q/vc; k21<-q/vp; k13<-q2/vc; k31<-q2/vp2
        d/dt(depot)       <- -ka*depot
        d/dt(central)     <-  ka*depot - kel*central - k12*central + k21*peripheral1 - k13*central + k31*peripheral2
        d/dt(peripheral1) <-  k12*central - k21*peripheral1
        d/dt(peripheral2) <-  k13*central - k31*peripheral2
        Cc<-central/vc; Adep<-depot; Ac<-central; Cp1<-peripheral1/vp; Cp2<-peripheral2/vp2
      })
    }))
    .chk(.lin, .ode, .evPO, c("Cc", "Adep", "Ac", "Cp1", "Cp2"))
  })

  test_that("the same references work without an error model (control)", {
    ## Always worked (no observation compartment injected); guards against the
    ## fix regressing the non-error path.
    .lin <- suppressMessages(rxode2(function() {
      ini({ lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50) })
      model({
        cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp)
        Cc<-linCmt(); Cp1<-peripheral1/vp
      })
    }))
    .ode <- suppressMessages(rxode2(function() {
      ini({ lcl<-log(5); lvc<-log(30); lq<-log(2); lvp<-log(50) })
      model({
        cl<-exp(lcl); vc<-exp(lvc); q<-exp(lq); vp<-exp(lvp)
        kel<-cl/vc; k12<-q/vc; k21<-q/vp
        d/dt(central)     <- -kel*central - k12*central + k21*peripheral1
        d/dt(peripheral1) <-  k12*central - k21*peripheral1
        Cc<-central/vc; Cp1<-peripheral1/vp
      })
    }))
    .chk(.lin, .ode, .evIV, c("Cc", "Cp1"))
  })

})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("infusion + bolus works correctly", {

      d <- test_path("test-infusion-bolus.rds")
      skip_if_not(file.exists(d))
      dataset <- readRDS(d)

      base_model <- function() {
        ini({
          POP_CL <- c(0.0, 3.84718, Inf)
          POP_VC <- c(0.0, 19.4133, Inf)
          POP_QP1 <- c(0.0, 3.84718, 999999.0)
          POP_VP1 <- c(0.0, 0.970665, 999999.0)
          ETA_1 + ETA_2 + ETA_3 + ETA_4 ~ c(0.145644, 0.0211085, 0.0202481, 0.011449, 0.0042689, 0.09, 0.011449, 0.0042689, 0.009, 0.09)
          sigma <- 0.0888837
        })
        model({
          VP1 <- POP_VP1*exp(ETA_4)
          QP1 <- POP_QP1*exp(ETA_3)
          CL <- POP_CL*exp(ETA_1)
          VC <- POP_VC*exp(ETA_2)
          V1 <- VC
          Q <- QP1
          V2 <- VP1
          d/dt(A_CENTRAL) = Q*A_PERIPHERAL/V2 + (-CL/V1 - Q/V1)*A_CENTRAL
          d/dt(A_PERIPHERAL) = -Q*A_PERIPHERAL/V2 + Q*A_CENTRAL/V1
          Fv <- A_CENTRAL
          IPRED <- A_CENTRAL/VC
          if (IPRED == 0) {
            IPREDADJ <- 2.22500000000000E-16
          } else {
            IPREDADJ <- IPRED
          }
          Y <- IPREDADJ
          Y ~ prop(sigma)
        })
      }

      f <- base_model()

      etTrans(dataset, f$simulationModel)


      s <- rxSolve(f, dataset, addDosing = TRUE)
      expect_true(any(names(s) == "rate"))
      expect_equal(names(s)[2], "resetno")

    })
  }
})

test_that("statePropDf test", {

  m1 <- function() {
    ini({
      KA   <- 2.94E-01
      CL   <- 1.86E+01
      V2   <- 4.02E+01
      Q    <- 1.05E+01
      V3   <- 2.97E+02
      Kin  <- 1
      Kout <- 1
      EC50 <- 200
      ## Added modeled bioavaiblity, duration and rate
      fdepot <- 1
      durDepot <- 8
      rateDepot <- 1250
    })
    model({
      C2 <- centr/V2
      C3 <- peri/V3
      d/dt(depot) <- -KA*depot
      depot(0)<- 0
      f(depot) <- fdepot
      dur(depot) <- durDepot
      rate(depot) <- rateDepot
      d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
      d/dt(peri)  <-                    Q*C2 - Q*C3
      d/dt(eff)  <- Kin - Kout*(1-C2/(EC50+C2))*eff
      eff(0) <- 1
    })
  }

  m1 <- m1()

  e1 <- data.frame(Compartment = c("depot", "depot", "depot", "depot", "eff"),
                   Property=c("ini", "f", "rate", "dur", "ini"))

  expect_equal(m1$statePropDf,e1)

  m2 <- m1 |> model(-depot(0))

  .noRow <- function(d) {
    row.names(d) <- NULL
    d
  }

  expect_equal(m2$statePropDf,
               .noRow(e1[-which(e1$Compartment == "depot" & e1$Property == "ini"), ]))

  m2 <- m1 |> model(-f(depot))

  expect_equal(m2$statePropDf,
               .noRow(e1[-which(e1$Compartment == "depot" & e1$Property == "f"), ]))

  m2 <- m1 |> model(-dur(depot))

  expect_equal(m2$statePropDf,
               .noRow(e1[-which(e1$Compartment == "depot" & e1$Property == "dur"), ]))

  m2 <- m1 |> model(-rate(depot))

  expect_equal(m2$statePropDf,
               .noRow(e1[-which(e1$Compartment == "depot" & e1$Property == "rate"), ]))



  m1 <- function() {
    ini({
      KA   <- 2.94E-01
      CL   <- 1.86E+01
      V2   <- 4.02E+01
      Q    <- 1.05E+01
      V3   <- 2.97E+02
    })
    model({
      C2 <- centr/V2
      C3 <- peri/V3
      d/dt(depot) <- -KA*depot
      d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
      d/dt(peri)  <-                    Q*C2 - Q*C3
    })
  }

  m2 <- m1()
  expect_equal(m2$statePropDf,
               NULL)

})

rxTest({
  test_that("test lhs string assign lhs expression information", {
    p <- rxode2parse('a="oh no"')
    expect_equal(p$lhs, "a")
    expect_equal(p$slhs, character(0))
    expect_equal(p$strAssign, list(a = "oh no"))

    p <- rxode2parse('a<-"oh no"')
    expect_equal(p$lhs, "a")
    expect_equal(p$slhs, character(0))
    expect_equal(p$strAssign, list(a = "oh no"))

    p <- rxode2parse('a~"oh no"')
    expect_equal(p$slhs, "a")
    expect_equal(p$lhs, character(0))
    expect_equal(p$strAssign, list(a = "oh no"))

    p <- rxode2parse('a<-"oh no"\nb<-3+40')
    expect_equal(p$lhs, c("a", "b"))
    expect_equal(p$slhs, character(0))
    expect_equal(p$strAssign, list(a = "oh no"))
    expect_equal(p$lhsStr, c(a = TRUE, b = FALSE))

    p <- rxode2parse('
if (APGAR == 10 || APGAR == 8 || APGAR == 9) {
    tAPGAR <- "High"
  } else if (APGAR == 1 || APGAR == 2 || APGAR == 3) {
    tAPGAR <- "Low"
  } else if (APGAR == 4 || APGAR == 5 || APGAR == 6 || APGAR == 7) {
    tAPGAR <- "Med"
  } else {
    tAPGAR<- "Med"
  }
')

    expect_equal(p$strAssign, list(tAPGAR = c("High", "Low", "Med")))
    expect_equal(p$lhsStr, c(tAPGAR = TRUE))
    expect_equal(p$lhs, "tAPGAR")
    expect_equal(p$slhs, character(0))
    expect_equal(p$params , "APGAR")
    expect_equal(p$model["normModel"],
                 c(normModel = "if (APGAR==10||APGAR==8||APGAR==9){\ntAPGAR <-\"High\";\n}\nelse {\nif (APGAR==1||APGAR==2||APGAR==3){\ntAPGAR <-\"Low\";\n}\nelse {\nif (APGAR==4||APGAR==5||APGAR==6||APGAR==7){\ntAPGAR <-\"Med\";\n}\nelse {\ntAPGAR <-\"Med\";\n}\n}\n}\n"))


    p <- rxode2parse('
if (APGAR == 10 || APGAR == 8 || APGAR == 9) {
    tAPGAR <- "High"
  } else if (APGAR == 1 || APGAR == 2 || APGAR == 3) {
    tAPGAR <- "Low"
  } else if (APGAR == 4 || APGAR == 5 || APGAR == 6 || APGAR == 7) {
    tAPGAR <- "Med"
  } else {
    tAPGAR<- "Med"
  }
   tAPGAR <- 1
')

    p <- rxode2parse('levels(tAPGAR) <- c("High", "Med", "Low")\ntAPGAR <- "Low"')

    expect_equal(p$strAssign, list(tAPGAR = c("High", "Med", "Low")))
    expect_equal(p$lhsStr, c(tAPGAR = TRUE))
    expect_equal(p$lhs, "tAPGAR")
    expect_equal(p$slhs, character(0))
    expect_equal(p$model["normModel"],
                 c(normModel = "levels(tAPGAR) <- c(\"High\", \"Med\", \"Low\");\ntAPGAR <-\"Low\";\n"))

  })


  f <- function() {
    expect_error(rxode2parse('a <- "matt"; a<- 2'))
    expect_error(rxode2parse('a <- "matt"; a<- 1'), NA)
    expect_error(rxode2parse('a <- "matt"; a <- "rxode2"; a<- 2'), NA)
    expect_error(rxode2parse('a <- "matt"; a <- "rxode2"; a<- 3'))
    expect_error(rxode2parse('a <- "matt"; a(0)<- 2'))
    expect_error(rxode2parse('a <- b; a <- "str"'))
    expect_error(rxode2parse('a <- 1; a <- "str"'))
    expect_error(rxode2parse("d/dt(a) <- -kel; a <- \"str\""))
    expect_error(rxode2parse("rate(a) <- -kel; a <- \"str\""))
    expect_error(rxode2parse("dur(a) <- -kel; a <- \"str\""))
    expect_error(rxode2parse("alag(a) <- -kel; a <- \"str\""))
    expect_error(rxode2parse("a(0) <- -kel; a <- \"str\""))
    expect_error(rxode2parse("a(0) <- 1; a <- \"str\""))
    expect_error(rxode2parse('a <- "matt"; d/dt(a)<- 2'))
    expect_error(rxode2parse('a <- "matt"; rate(a)<- 2'))
    expect_error(rxode2parse('a <- "matt"; dur(a)<- 2'))
    expect_error(rxode2parse('a <- "matt"; alag(a)<- 2'))
    expect_error(rxode2parse("a <- \"str\"; a(0) <- -kel"))
    expect_error(rxode2parse("a <- \"str\"; a(0) <- 1"))
    # so that pruned expressions can work
    expect_error(rxode2parse("a <- \"str\"; a <- 1+5"), NA)
    expect_error(rxode2parse("a <- \"str\"; a <- -1+5"), NA)
    expect_error(rxode2parse("a <- \"str\"; a <- +1+5"), NA)
  }

  test_that("test lhs string assign rxode2.syntax.allow.ini=TRUE", {
    withr::with_options(list(rxode2.syntax.allow.ini=TRUE,
                             rxode2.syntax.require.ode.first = FALSE), {
                               f()
                             })
  })

  test_that("test lhs string assign rxode2.syntax.allow.ini=FALSE", {
    withr::with_options(list(rxode2.syntax.allow.ini=FALSE,
                             rxode2.syntax.require.ode.first = FALSE), {
                               f()
                             })
  })

  test_that("lhs solve; tests lhs assign & str equals with lhs", {

    rx <- rxode2({
      if (t < 10) {
        a <- "<10"
      } else {
        a <- ">=10"
      }
      b <- 1
      if (a == "<10") {
        b <- 0;
      }
    })

    e <- et(1:20)

    s <-rxSolve(rx, e, returnType = "data.frame")

    expect_true(all(s$a[s$time < 10] == "<10"))
    expect_true(all(s$a[s$time >= 10] == ">=10"))
    expect_true(all(s$b[s$time < 10] == 0))
    expect_true(all(s$b[s$time >= 10] == 1))
  })

  test_that("out of bounds solve gives NA for factors", {

    rx <- rxode2({
      if (t < 10) {
        a <- "<10"
      } else {
        a <- ">=10"
      }
      a <- 1-3
      b <- 1
      if (a == "<10") {
        b <- 0;
      }
    })

    e <- et(1:20)

    s <-rxSolve(rx, e, returnType = "data.frame")

    expect_true(all(is.na(s$a)))

    rx <- rxode2({
      if (t < 10) {
        a <- "<10"
      } else {
        a <- ">=10"
      }
      a <- 1+20
      b <- 1
      if (a == "<10") {
        b <- 0;
      }
    })

    s <-rxSolve(rx, e, returnType = "data.frame")

    expect_true(all(is.na(s$a)))


  })


  test_that("lhs solve; tests lhs levels & str equals with lhs", {

    rx <- rxode2({
      levels(a) <- c("<10", ">=10")
      if (t < 10) {
        a <- 1
      } else {
        a <- 2
      }
      b <- 1
      if (a == "<10") {
        b <- 0;
      }
    })

    e <- et(1:20)

    s <-rxSolve(rx, e, returnType = "data.frame")

    expect_true(all(s$a[s$time < 10] == "<10"))
    expect_true(all(s$a[s$time >= 10] == ">=10"))
    expect_true(all(s$b[s$time < 10] == 0))
    expect_true(all(s$b[s$time >= 10] == 1))

  })


  test_that("levels1 statement solve", {

    rx <- rxode2({
      levels(a) <- "<10"
      if (t < 10) {
        a <- 1
      } else {
        a <- ">=10"
      }
      b <- 1
      if (a == "<10") {
        b <- 0;
      }
    })

    e <- et(1:20)

    s <-rxSolve(rx, e, returnType = "data.frame")

    expect_true(all(s$a[s$time < 10] == "<10"))
    expect_true(all(s$a[s$time >= 10] == ">=10"))
    expect_true(all(s$b[s$time < 10] == 0))
    expect_true(all(s$b[s$time >= 10] == 1))

  })

  test_that("levels extraction", {

    rx <- function() {
      model({
        levels(a) <- c("<10", ">=10")
        if (t < 10) {
          a <- 1
        } else {
          a <- 2
        }
        b <- 1
        if (a == "<10") {
          b <- 2;
        }
      })
    }

    rx <- rx()

    expect_equal(rx$levels,
                 list(str2lang("levels(a) <- c(\"<10\", \">=10\")")))

    rx <- function() {
      model({
        levels(a) <- c("<10", ">=10")
        levels(b) <- c("low", "high")
        if (t < 10) {
          a <- 1
        } else {
          a <- 2
        }
        b <- 1
        if (a == "<10") {
          b <- 2;
        }
      })
    }

    rx <- rx()

    expect_equal(rx$levels,
                 list(str2lang("levels(a) <- c(\"<10\", \">=10\")"),
                      str2lang("levels(b) <- c(\"low\", \"high\")")))


    rx <- function() {
      model({
        levels(a) <- c("<10", ">=10")
        levels(b) <- c("low", "high")
        levels(c) <- c("funny")
        if (t < 10) {
          a <- 1
        } else {
          a <- 2
        }
        b <- 1
        if (a == "<10") {
          b <- 2;
        }
        c <- 1
      })
    }

    rx <- rx()

    expect_equal(rx$levels,
                 list(str2lang("levels(a) <- c(\"<10\", \">=10\")"),
                      str2lang("levels(b) <- c(\"low\", \"high\")"),
                      str2lang("levels(c) <- \"funny\"")))

  })

  test_that("test symengine translation to integers", {

    v <-rxModelVars("
levels(a) <- c(\"<10\", \">=10\")
b <- (a == \"<10\")*1 + (a == \">=10\")*2
")

    s <- rxS(v)
    v <- as.character(s$b)
    expect_error(rxFromSE(v), NA)

    expect_equal(.rxPrune(str2lang("{a<-'a'; b<-1}"),
                          strAssign=list(a = "a")),
                 "a<-1\nb<-1")

    expect_equal(.rxPrune(str2lang("{a<-'a'; b<-1}"),
                          strAssign=list(a = c("b", "a"))),
                 "a<-2\nb<-1")

    expect_equal(.rxPrune(str2lang("{a<-'a'; b<-1}"),
                          strAssign=list(c = c("b", "a"))),
                 "a<-\"a\"\nb<-1")

    expect_equal(.rxPrune(str2lang("{b <- (a == 'a')}"),
                          strAssign=list(c = c("b", "a"))),
                 "b<-(a==\"a\")")

    expect_equal(.rxPrune(str2lang("{b <- (a == 'a')}"),
                          strAssign=list(a = c("b", "a"))),
                 "b<-(a==2)")

    expect_equal(.rxPrune(str2lang("{b <- (a == 'a')}"),
                          strAssign=list(a = c("a"))),
                 "b<-(a==1)")

    ## No test the other direction
    expect_equal(.rxPrune(str2lang("{b <- ('a' == a)}"),
                          strAssign=list(c = c("b", "a"))),
                 "b<-(\"a\"==a)")

    expect_equal(.rxPrune(str2lang("{b <- ('a' == a)}"),
                          strAssign=list(a = c("b", "a"))),
                 "b<-(2==a)")

    expect_equal(.rxPrune(str2lang("{b <- ('a' == a)}"),
                          strAssign=list(a = c("a"))),
                 "b<-(1==a)")


    ## neq
    expect_equal(.rxPrune(str2lang("{b <- (a != 'a')}"),
                          strAssign=list(c = c("b", "a"))),
                 "b<-(a!=\"a\")")

    expect_equal(.rxPrune(str2lang("{b <- (a != 'a')}"),
                          strAssign=list(a = c("b", "a"))),
                 "b<-(a!=2)")

    expect_equal(.rxPrune(str2lang("{b <- (a != 'a')}"),
                          strAssign=list(a = c("a"))),
                 "b<-(a!=1)")

    ## No test the other direction
    expect_equal(.rxPrune(str2lang("{b <- ('a' != a)}"),
                          strAssign=list(c = c("b", "a"))),
                 "b<-(\"a\"!=a)")

    expect_equal(.rxPrune(str2lang("{b <- ('a' != a)}"),
                          strAssign=list(a = c("b", "a"))),
                 "b<-(2!=a)")

    expect_equal(.rxPrune(str2lang("{b <- ('a' != a)}"),
                          strAssign=list(a = c("a"))),
                 "b<-(1!=a)")

  })


  test_that("simulation model will include string information", {

    f <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        cl.wt <- 0
        v.wt <- 0
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        if (time > 10) {
          timeText <- "time > 10"
        } else {
          timeText <- "time <= 10"
        }
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)+ WT ^ 2* cl.wt
        v <- exp(tv + eta.v+ WT * v.wt + b + c + d)
        linCmt() ~ add(add.sd)
      })
    }

    ui <- rxode(f)

    expect_error(ui$simulationModel, NA)

    mod <- ui$simulationModel

    expect_equal(rxModelVars(mod)$strAssign,
                 list(timeText = c("time > 10", "time <= 10")))

    expect_error(ui$simulationIniModel, NA)

    mod <- ui$simulationIniModel

    expect_equal(rxModelVars(mod)$strAssign,
                 list(timeText = c("time > 10", "time <= 10")))

  })

  test_that("CENS column does not produce malformed factor LHS columns", {
    rx <- rxode2({
      ka <- exp(tka)
      cl <- exp(tcl)
      v  <- exp(tv)
      d/dt(depot) = -ka * depot
      d/dt(cent)  = ka * depot - cl / v * cent
      cp = cent / v
    })
    # 3 subjects x 6 rows = 18 events; dfN = 7 (ID, TIME, EVID, AMT, DV, CMT, CENS).
    # glhs_str starts at gcens[2*7=14]; events 14-17 have CENS=1, which corrupts
    # glhs_str[0..3] = lhs_str for ka, cl, v, cp.
    dat <- data.frame(
      ID   = rep(1:3, each = 6),
      TIME = rep(c(0, 0.5, 1.0, 1.5, 2.0, 2.5), 3),
      EVID = rep(c(1L, 0L, 0L, 0L, 0L, 0L), 3),
      AMT  = rep(c(10, NA, NA, NA, NA, NA), 3),
      DV   = rep(c(NA, 3.0, 2.0, 1.5, 1.0, 0.5), 3),
      CMT  = rep(c(1L, 2L, 2L, 2L, 2L, 2L), 3),
      CENS = c(rep(0L, 14), rep(1L, 4))
    )
    params <- c(tka = log(0.5), tcl = log(0.3), tv = log(8))
    ret <- rxSolve(rx, params, dat)
    df  <- as.data.frame(ret)
    for (col in c("ka", "cl", "v", "cp")) {
      expect_true(is.numeric(df[[col]]),
                  label = paste0(col, " must be numeric, not a factor"))
      expect_true(any(!is.na(df[[col]])),
                  label = paste0(col, " must not be all NA"))
    }
  })
})

rxTest({
  # Make sure there are lower bounds for appropriate errors

  test_that("add/prop have appropriate errors", {

    run7.mod <- function() {
      ini({
        tcl <- log(0.008) ; label("typical value of clearance")
        tv <- log(0.6) ; label("typical value of volume")
        all.cl <- 1 ; label("allometric exponent on CL")
        eta.cl + eta.v ~ c(
          1,
          0.01, 1
        ) ; label("interindividual variability on clearance and volume")
        add.err <- 0.1 ; label("residual variability")
        prop.err <- 0.1
      })
      model({
        cl <- exp(tcl + all.cl * log_allo_wt + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v) # individual value of volume
        ke <- cl / v # elimination rate constant
        d / dt(A1) <- -ke * A1 # model differential equation
        cp <- A1 / v # concentration in plasma
        cp ~ add(add.err) + prop(prop.err) # define error model
      })
    }

    f <- rxode2(run7.mod)
    expect_equal(f$iniDf$lower, c(-Inf, -Inf, -Inf, 0, 0, -Inf, -Inf, -Inf))

    run7.mod <- function() {
      ini({
        tcl <- log(0.008) ; label("typical value of clearance")
        tv <- log(0.6) ; label("typical value of volume")
        all.cl <- 1 ; label("allometric exponent on CL")
        eta.cl + eta.v ~ c(
          1,
          0.01, 1
        ) ; label("interindividual variability on clearance and volume")
        add.err <- c(0.001, 0.1) ; label("residual variability")
        prop.err <- c(0.001, 0.1)
      })
      model({
        cl <- exp(tcl + all.cl * log_allo_wt + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v) # individual value of volume
        ke <- cl / v # elimination rate constant
        d / dt(A1) <- -ke * A1 # model differential equation
        cp <- A1 / v # concentration in plasma
        cp ~ add(add.err) + prop(prop.err) # define error model
      })
    }
    f <- rxode2(run7.mod)
    expect_equal(f$iniDf$lower, c(-Inf, -Inf, -Inf, 0.001, 0.001, -Inf, -Inf, -Inf))

    run7.mod <- function() {
      ini({
        tcl <- log(0.008) ; label("typical value of clearance")
        tv <- log(0.6) ; label("typical value of volume")
        all.cl <- 1 ; label("allometric exponent on CL")
        eta.cl + eta.v ~ c(
          1,
          0.01, 1
        ) ; label("interindividual variability on clearance and volume")
        add.err <- -1 ; label("residual variability")
        prop.err <- c(0.001, 0.1)
      })
      model({
        cl <- exp(tcl + all.cl * log_allo_wt + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v) # individual value of volume
        ke <- cl / v # elimination rate constant
        d / dt(A1) <- -ke * A1 # model differential equation
        cp <- A1 / v # concentration in plasma
        cp ~ add(add.err) + prop(prop.err) # define error model
      })
    }

    expect_error(rxode2(run7.mod))


    run7.mod <- function() {
      ini({
        tcl <- log(0.008) ; label("typical value of clearance")
        tv <- log(0.6) ; label("typical value of volume")
        all.cl <- 1 ; label("allometric exponent on CL")
        eta.cl + eta.v ~ c(
          1,
          0.01, 1
        ) ; label("interindividual variability on clearance and volume")
        add.err <- 2 ; label("residual variability")
      })
      model({
        cl <- exp(tcl + all.cl * log_allo_wt + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v) # individual value of volume
        ke <- cl / v # elimination rate constant
        d / dt(A1) <- -ke * A1 # model differential equation
        cp <- A1 / v # concentration in plasma
        cp ~ bern(add.err)  # define error model
      })
    }
    expect_error(rxode2(run7.mod))

    run7.mod <- function() {
      ini({
        tcl <- log(0.008) ; label("typical value of clearance")
        tv <- log(0.6) ; label("typical value of volume")
        all.cl <- 1 ; label("allometric exponent on CL")
        eta.cl + eta.v ~ c(
          1,
          0.01, 1
        ) ; label("interindividual variability on clearance and volume")
        add.err <- c(-1, 0.5, 1) ; label("residual variability")
      })
      model({
        cl <- exp(tcl + all.cl * log_allo_wt + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v) # individual value of volume
        ke <- cl / v # elimination rate constant
        d / dt(A1) <- -ke * A1 # model differential equation
        cp <- A1 / v # concentration in plasma
        cp ~ bern(add.err)  # define error model
      })
    }

    expect_warning(rxode2(run7.mod), NA)

    run7.mod <- function() {
      ini({
        tcl <- log(0.008) ; label("typical value of clearance")
        tv <- log(0.6) ; label("typical value of volume")
        all.cl <- 1 ; label("allometric exponent on CL")
        eta.cl + eta.v ~ c(
          1,
          0.01, 1
        ) ; label("interindividual variability on clearance and volume")
        add.err <- c(0, 0.5, 1) ; label("residual variability")
      })
      model({
        cl <- exp(tcl + all.cl * log_allo_wt + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v) # individual value of volume
        ke <- cl / v # elimination rate constant
        d / dt(A1) <- -ke * A1 # model differential equation
        cp <- A1 / v # concentration in plasma
        cp ~ bern(add.err)  # define error model
      })
    }

    expect_warning(rxode2(run7.mod), NA)

  })
})

rxTest({
  # Make sure the right state is ignored.
  ## https://cran.r-project.org/web/packages/diffEq/vignettes/ODEinR.pdf p11

  ## 6.1
  mod <- rxode2({
    a <- 6
    b <- 0.6
    d / dt(intestine) <- -a * intestine
    d / dt(blood) <- a * intestine - b * blood
  })

  test_that("Nothing ignored", {
    expect_equal(rxModelVars(mod)$state.ignore, c(0L, 0L))
  })

  mod <- rxode2({
    a <- 6
    b <- 0.6
    d / dt(intestine) ~ -a * intestine
    d / dt(blood) <- a * intestine - b * blood
  })

  mod.f <- rxode2({
    a <- 6
    b <- 0.6
    f(intestine) <- 1
    d / dt(intestine) ~ -a * intestine
    d / dt(blood) <- a * intestine - b * blood
  })

  mod.alag <- rxode2({
    a <- 6
    b <- 0.6
    alag(intestine) <- 0
    d / dt(intestine) ~ -a * intestine
    d / dt(blood) <- a * intestine - b * blood
  })

  mod.rate <- rxode2({
    a <- 6
    b <- 0.6
    rate(intestine) <- 2
    d / dt(intestine) ~ -a * intestine
    d / dt(blood) <- a * intestine - b * blood
  })

  mod.dur <- rxode2({
    a <- 6
    b <- 0.6
    dur(intestine) <- 2
    d / dt(intestine) ~ -a * intestine
    d / dt(blood) <- a * intestine - b * blood
  })

  test_that("Ignore first comparment", {
    expect_equal(rxModelVars(mod)$state.ignore, c(1L, 0L))
    expect_equal(rxModelVars(mod.f)$state.ignore, c(1L, 0L))
    expect_equal(rxModelVars(mod.alag)$state.ignore, c(1L, 0L))
    expect_equal(rxModelVars(mod.rate)$state.ignore, c(1L, 0L))
    expect_equal(rxModelVars(mod.dur)$state.ignore, c(1L, 0L))
  })

  mod <- rxode2({
    a <- 6
    b <- 0.6
    d / dt(intestine) <- -a * intestine
    d / dt(blood) ~ a * intestine - b * blood
  })

  mod.f <- rxode2({
    a <- 6
    b <- 0.6
    f(blood) <- 1
    d / dt(intestine) <- -a * intestine
    d / dt(blood) ~ a * intestine - b * blood
  })

  mod.alag <- rxode2({
    a <- 6
    b <- 0.6
    alag(blood) <- 0
    d / dt(intestine) <- -a * intestine
    d / dt(blood) ~ a * intestine - b * blood
  })

  mod.rate <- rxode2({
    a <- 6
    b <- 0.6
    rate(blood) <- 2
    d / dt(intestine) <- -a * intestine
    d / dt(blood) ~ a * intestine - b * blood
  })

  mod.dur <- rxode2({
    a <- 6
    b <- 0.6
    dur(blood) <- 2
    d / dt(intestine) <- -a * intestine
    d / dt(blood) ~ a * intestine - b * blood
  })

  test_that("Nothing ignored", {
    expect_equal(rxModelVars(mod)$state.ignore, c(0L, 1L))
    expect_equal(rxModelVars(mod)$state, c("intestine", "blood"))
    expect_equal(rxModelVars(mod.f)$state.ignore, c(0L, 1L))
    expect_equal(rxModelVars(mod.f)$state, c("intestine", "blood"))
    expect_equal(rxModelVars(mod.alag)$state.ignore, c(0L, 1L))
    expect_equal(rxModelVars(mod.alag)$state, c("intestine", "blood"))
    expect_equal(rxModelVars(mod.rate)$state.ignore, c(0L, 1L))
    expect_equal(rxModelVars(mod.rate)$state, c("intestine", "blood"))
    expect_equal(rxModelVars(mod.dur)$state.ignore, c(0L, 1L))
    expect_equal(rxModelVars(mod.dur)$state, c("intestine", "blood"))
  })

  mod.f <- rxode2({
    a <- 6
    b <- 0.6
    d / dt(intestine) <- -a * intestine
    f(blood) <- 1
    d / dt(blood) ~ a * intestine - b * blood
  })

  mod.alag <- rxode2({
    a <- 6
    b <- 0.6
    d / dt(intestine) <- -a * intestine
    alag(blood) <- 0
    d / dt(blood) ~ a * intestine - b * blood
  })

  mod.rate <- rxode2({
    a <- 6
    b <- 0.6
    d / dt(intestine) <- -a * intestine
    rate(blood) <- 2
    d / dt(blood) ~ a * intestine - b * blood
  })

  mod.dur <- rxode2({
    a <- 6
    b <- 0.6
    d / dt(intestine) <- -a * intestine
    dur(blood) <- 2
    d / dt(blood) ~ a * intestine - b * blood
  })


  test_that("Ignore Compartment #2", {
    expect_equal(rxModelVars(mod)$state.ignore, c(0L, 1L))
    expect_equal(rxModelVars(mod)$state, c("intestine", "blood"))
    expect_equal(rxModelVars(mod.f)$state.ignore, c(0L, 1L))
    expect_equal(rxModelVars(mod.f)$state, c("intestine", "blood"))
    expect_equal(rxModelVars(mod.alag)$state.ignore, c(0L, 1L))
    expect_equal(rxModelVars(mod.alag)$state, c("intestine", "blood"))
    expect_equal(rxModelVars(mod.rate)$state.ignore, c(0L, 1L))
    expect_equal(rxModelVars(mod.rate)$state, c("intestine", "blood"))
    expect_equal(rxModelVars(mod.dur)$state.ignore, c(0L, 1L))
    expect_equal(rxModelVars(mod.dur)$state, c("intestine", "blood"))
  })

  test_that("ignore state corruption issue #857", {
    d_clean <-
      structure(list(
        amt = c(0, 150, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
        evid = c(0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
        dv = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_),
        mdv = c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L),
        cmt = c(2, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2),
        cens = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_, NA_real_),
        id = c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L),
        ii = c(0, 24, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
        addl = c(0, 28, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
        time = c(0, 0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5, 5.5, 6, 6.5, 7, 7.5, 8),
        FED = c(0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L),
        FORMTAB = c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L)
      ), row.names = c(NA, -18L), class = c("tbl_df", "tbl", "data.frame"))

    fit_clean <- function () {
      ini({
        e_f_fed <- fix(0)
        e_f_form <- fix(0)
        lka <- -0.4
        e_ka_fed <- -2
        e_ka_form <- -2
        lcl <- -0.3
        lvc <- 2.5
        lvp <- 5
        lq <- 0.3
        prop_sd <- 0.4
        add_sd <- 17
        bsv_ka ~ 0.3
        bsv_cl ~ 0.5
        bsv_vc ~ 0.5
      })
      model({
        f_depot <- exp(e_f_fed * FED + e_f_form * FORMTAB)
        ka <- exp(lka + bsv_ka + e_ka_fed * FED + e_ka_form *
                    FORMTAB)
        cl <- exp(lcl + bsv_cl)
        vc <- exp(lvc + bsv_vc)
        vp <- exp(lvp)
        q <- exp(lq)
        kel <- cl/vc
        k12 <- q/vc
        k21 <- q/vp
        d/dt(depot) <- -ka * depot
        d/dt(central) <- ka * f_depot * depot - kel * central -
          k12 * central + k21 * peripheral1
        d/dt(peripheral1) <- k12 * central - k21 * peripheral1
        Cc <- central/vc * 1000
        Cc ~ prop(prop_sd) + add(add_sd)
      })
    }
    expect_error(rxSolve(fit_clean, d_clean), NA)
  })

})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("rxode2<- and other rxUi methods", {

      one.compartment <- function() {
        ini({
          tka <- log(1.57)
          tcl <- log(2.72)
          tv <- log(31.5)
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.sd <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
          cp ~ add(add.sd)
        })
      }

      one.compartment2 <- function() {
        ini({
          tka <- 1.57
          tcl <- 2.72
          tv <- 31.5
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.sd <- 0.7
        })
        model({
          ka <- tka * exp(eta.ka)
          cl <- tcl *exp(eta.cl)
          v <- tv*exp(eta.v)
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
          cp ~ add(add.sd)
        })
      }

      two.compartment <- function() {
        ini({
          lka <- 0.45
          lcl <- 1
          lvc  <- 3
          lvp  <- 5
          lq  <- 0.1
          propSd <- 0.5
        })
        model({
          ka <- exp(lka)
          cl <- exp(lcl)
          vc <- exp(lvc)
          vp <- exp(lvp)
          q  <- exp(lq)
          kel <- cl/vc
          k12 <- q/vc
          k21 <- q/vp
          d/dt(depot) <- -ka*depot
          d/dt(central) <-  ka*depot - kel*central - k12*central + k21*peripheral1
          d/dt(peripheral1) <- k12*central - k21*peripheral1
          cp <- central / vc
          cp ~ prop(propSd)
        })
      }

      uiOne <- rxode2(one.compartment)
      uiTwo <- uiOne
      rxode2(uiTwo) <- body(two.compartment)
      expect_equal(body(uiOne$fun), body(rxode2(one.compartment)$fun))
      expect_equal(body(uiTwo$fun), body(rxode2(two.compartment)$fun))

      fun1 <- one.compartment
      fun2 <- two.compartment
      rxode2(fun1) <- fun2
      expect_equal(fun1, fun2)

      fun1 <- one.compartment
      rxode2(fun1) <- body(fun2)
      expect_equal(deparse1(fun1), deparse1(fun2))

      expect_error({
        rxode2(fun1) <- "matt"
      })

      uiOne <- rxode2(one.compartment)
      uiTwo <- uiOne
      rxode2(uiTwo) <- two.compartment

      expect_equal(body(uiOne$fun), body(rxode2(one.compartment)$fun))
      expect_equal(body(uiTwo$fun), body(rxode2(two.compartment)$fun))

      uiOne <- rxode2(one.compartment)
      uiOne$model <- model(one.compartment2)
      expect_equal(model(uiOne), model(one.compartment2))
      expect_equal(ini(uiOne), ini(one.compartment))


      uiOne <- rxode2(one.compartment)

      model(uiOne) <-  model(one.compartment2)
      expect_equal(model(uiOne), model(one.compartment2))
      expect_equal(ini(uiOne), ini(one.compartment))

      uiOne <- rxode2(one.compartment)
      ini(uiOne) <-  ini(one.compartment2)
      expect_equal(model(uiOne), model(one.compartment))
      expect_equal(ini(uiOne), ini(one.compartment2))

      # now lets add something to the model that should be kept and dropped
      uiOne <- rxUiDecompress(rxode2(one.compartment))
      uiOne$sticky <- "matt"
      uiOne$matt <- "f"
      uiOne$f <- "matt"
      class(uiOne) <- c("uiOne", class(uiOne))

      # this makes "insignificant" changes
      iniNew <- quote(ini({
        tka <- c(-Inf, 0.451075619360217, 2.99573227355399)
        tcl <- fix(1.00063188030791)
        tv <- 3.44998754583159
        label("tv")
        add.sd <- c(0, 0.7)
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
      }))

      ini(uiOne) <-  iniNew
      expect_equal(ini(uiOne), iniNew)
      expect_equal(uiOne$matt, "f")
      expect_equal(uiOne$f, "matt")
      expect_true(inherits(uiOne, "uiOne"))

      # order is also an insignificant change
      iniNew <- quote(ini({
        tcl <- fix(1.00063188030791)
        tka <- c(-Inf, 0.451075619360217, 2.99573227355399)
        tv <- 3.44998754583159
        label("tv")
        add.sd <- c(0, 0.7)
        eta.ka ~ 0.6
        eta.v ~ 0.1
        eta.cl ~ 0.3
      }))

      ini(uiOne) <-  iniNew

      expect_equal(ini(uiOne), iniNew)
      expect_equal(uiOne$matt, "f")
      expect_equal(uiOne$f, "matt")
      expect_true(inherits(uiOne, "uiOne"))

      ## changing an estimate is a significant change
      iniNew <- quote(ini({
        tka <- c(-Inf, 0.451075619360217, 2.99573227355399)
        tcl <- fix(1.00063188030791)
        tv <- 3.44998754583159
        label("tv")
        add.sd <- c(0, 0.7)
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 1
      }))

      ini(uiOne) <-  iniNew

      expect_equal(ini(uiOne), iniNew)
      expect_equal(uiOne$matt, "f")
      expect_equal(uiOne$f, NULL)
      expect_true(inherits(uiOne, "uiOne"))

      # now test changing model() should be a significant change
      uiOne <- rxUiDecompress(rxode2(one.compartment))
      uiOne$sticky <- "matt"
      uiOne$matt <- "f"
      uiOne$f <- "matt"
      class(uiOne) <- c("uiOne", class(uiOne))

      model(uiOne) <-  model(one.compartment2)

      expect_equal(model(uiOne), model(one.compartment2))
      expect_equal(uiOne$matt, "f")
      expect_equal(uiOne$f, NULL)
      expect_true(inherits(uiOne, "uiOne"))

      # now test piping
      uiOne <- rxUiDecompress(rxode2(one.compartment))
      uiOne$sticky <- "matt"
      uiOne$matt <- "f"
      uiOne$f <- "matt"
      class(uiOne) <- c("uiOne", class(uiOne))

      uiTwo <- uiOne |>
        ini(tka=fix)

      expect_equal(uiTwo$matt, "f")
      expect_equal(uiTwo$f, "matt")
      expect_true(inherits(uiTwo, "uiOne"))

      # now a significant change
      uiTwo <- uiOne |>
        ini(tcl=77)

      expect_equal(uiTwo$matt, "f")
      expect_equal(uiTwo$f, NULL)
      expect_true(inherits(uiTwo, "uiOne"))

      # nothing change in input ui
      expect_equal(uiOne$matt, "f")
      expect_equal(uiOne$f, "matt")
      expect_true(inherits(uiOne, "uiOne"))

      uiTwo <- uiOne |>
        model(ka <- tka * exp(eta.ka))

      expect_equal(uiTwo$matt, "f")
      expect_equal(uiTwo$f, NULL)
      expect_true(inherits(uiTwo, "uiOne"))

      # rename something in the model block, insignificant
      uiTwo <- uiOne |>
        rxRename(isKa=ka)

      expect_equal(uiTwo$matt, "f")
      expect_equal(uiTwo$f, "matt")
      expect_true(inherits(uiTwo, "uiOne"))


      # rename something in the ini block is also an insignificant change
      uiTwo <- uiOne |>
        rxRename(isKa=tka)

      expect_equal(uiTwo$matt, "f")
      expect_equal(uiTwo$f, "matt")
      expect_true(inherits(uiTwo, "uiOne"))



    })

    test_that("ini(model) <- NULL drops", {
      one.compartment <- function() {
        ini({
          tka <- log(1.57)
          tcl <- log(2.72)
          tv <- log(31.5)
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
        })
      }

      uiOne <- one.compartment()
      ini(uiOne) <- NULL
      expect_length(uiOne$iniDf$ntheta, 0L)
      expect_equal(as.ini(NULL), quote(ini({}))) #nolint

      # try with $ini assignment
      uiOne <- one.compartment()
      uiOne$ini <- NULL
      expect_length(uiOne$iniDf$ntheta, 0L)
      expect_equal(as.ini(NULL), quote(ini({}))) #nolint
    })

    test_that("assign model changes meta information", {

      one.compartment <- function() {
        ini({
          tka <- log(1.57)
          tcl <- log(2.72)
          tv <- log(31.5)
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
        })
      }

      uiOne <- one.compartment()

      uiOne$matt <- "matt"

      expect_equal(uiOne$meta$matt, "matt")

      expect_equal(uiOne$matt, "matt")

    })
  }
})

rxTest({
  test_that("batched multi-subject linCmt() + modeled alag() does not leak amounts between subjects (#1153)", {
    mod <- function() {
      ini({
        tcl <- log(3)
        tv <- log(30)
        tka <- log(1)
        tlag <- log(0.7)
      })
      model({
        cl <- exp(tcl)
        v <- exp(tv)
        ka <- exp(tka)
        alag(depot) <- exp(tlag)
        cp <- linCmt(cl, v, ka)
      })
    }

    tms <- c(0.9, 1.7, 3, 6, 10)
    ev <- et(amt = 100, cmt = "depot") %>% et(tms)
    pars <- data.frame(tcl = log(3) + c(-0.5, 0, 0.5), tv = log(30),
                       tka = log(1), tlag = log(0.7))

    # reference: each subject solved alone
    ref <- unlist(lapply(seq_len(nrow(pars)), function(i) {
      rxSolve(mod, params = pars[i, ], events = ev,
              returnType = "data.frame", addDosing = FALSE)$cp
    }))

    # batched: subjects share one thread when cores < nSub; before the fix the
    # per-thread linCmt() save buffer leaked subject k-1's amounts into k
    for (cr in 1:2) {
      o <- rxSolve(mod, params = pars, events = ev, cores = cr,
                   returnType = "data.frame", addDosing = FALSE)
      expect_equal(o$cp, ref, tolerance = 1e-8)
    }
  })

  test_that("nonzero linCmt() initial amounts via inits= are still honored", {
    mod <- function() {
      ini({
        tcl <- log(3)
        tv <- log(30)
        tka <- log(1)
      })
      model({
        cl <- exp(tcl)
        v <- exp(tv)
        ka <- exp(tka)
        cp <- linCmt(cl, v, ka)
      })
    }
    tms <- c(0.5, 1, 2, 4)
    ev <- et(tms) # no dose; only the initial amount can make cp nonzero
    s <- rxSolve(mod, events = ev, inits = c(central = 50),
                 returnType = "data.frame")
    expect_equal(s$cp, 50 / 30 * exp(-0.1 * tms), tolerance = 1e-6)
  })
})

rxTest({
  test_that("rxFix", {

    One.comp.transit.allo <- function() {
      ini({
        # Where initial conditions/variables are specified
        lktr <- log(1.15)  #log k transit (/h)
        lcl  <- log(0.15)  #log Cl (L/hr)
        lv   <- log(7)     #log V (L)
        ALLC <- fix(0.75)  #allometric exponent cl
        ALLV <- fix(1.00)  #allometric exponent v
        prop.err <- 0.15   #proportional error (SD/mean)
        add.err <- 0.6     #additive error (mg/L)
        eta.ktr ~ 0.5
        eta.cl ~ 0.1
        eta.v ~ 0.1
      })
      model({
        #Allometric scaling on weight
        cl <- exp(lcl + eta.cl + ALLC * logWT70)
        v  <- exp(lv + eta.v + ALLV * logWT70)
        ktr <- exp(lktr + eta.ktr)
        # RxODE-style differential equations are supported
        d/dt(depot)   = -ktr * depot
        d/dt(central) =  ktr * trans - (cl/v) * central
        d/dt(trans)   =  ktr * depot - ktr * trans
        ## Concentration is calculated
        cp = central/v
        # And is assumed to follow proportional and additive error
        cp ~ prop(prop.err) + add(add.err)
      })
    }

    tmp <- rxFixPop(One.comp.transit.allo)

    expect_equal(tmp$theta,
                 c(lktr = 0.139761942375159, lcl = -1.89711998488588, lv = 1.94591014905531,
                   prop.err = 0.15, add.err = 0.6))

    tmp2 <- rxFixPop(tmp)

    expect_equal(tmp2$theta,
                 c(lktr = 0.139761942375159, lcl = -1.89711998488588, lv = 1.94591014905531,
                   prop.err = 0.15, add.err = 0.6))

    tmp3 <- rxFixPop(tmp2, returnNull = TRUE)

    expect_true(is.null(tmp3))

    One.comp.transit.allo <- function() {
      ini({
        # Where initial conditions/variables are specified
        lktr <- log(1.15)  #log k transit (/h)
        lcl  <- log(0.15)  #log Cl (L/hr)
        lv   <- log(7)     #log V (L)
        ALLC <- 0.75  #allometric exponent cl
        ALLV <- 1.00  #allometric exponent v
        prop.err <- fix(0.15)   #proportional error (SD/mean)
        add.err <- fix(0.6)     #additive error (mg/L)
        eta.ktr ~ 0.5
        eta.cl ~ 0.1
        eta.v ~ 0.1
      })
      model({
        #Allometric scaling on weight
        cl <- exp(lcl + eta.cl + ALLC * logWT70)
        v  <- exp(lv + eta.v + ALLV * logWT70)
        ktr <- exp(lktr + eta.ktr)
        # RxODE-style differential equations are supported
        d/dt(depot)   = -ktr * depot
        d/dt(central) =  ktr * trans - (cl/v) * central
        d/dt(trans)   =  ktr * depot - ktr * trans
        ## Concentration is calculated
        cp = central/v
        # And is assumed to follow proportional and additive error
        cp ~ prop(prop.err) + add(add.err)
      })
    }

    tmp <- rxFixRes(One.comp.transit.allo)

    nlmixr_threecmt_mm_no_add_wtcl_pdtg_kout_delay2 <- function() {
      ini({
        tf_sc <- log(999)
        tf_infilt <- log(999)
        tka_sc <- log(999)
        tka_infilt <- log(999)
        tcl_low <- log(999)
        tcl_high <- log(999)
        tcl_c50 <- log(3000)
        e_wt_cl <- fixed(999)
        tv <- log(999)
        tq1 <- log(999)
        tvp1 <- log(10)
        tq2 <- log(999)
        tvp2 <- log(20)
        eta_cl~999
        eta_v~999
        prop_err <- 999
        tg_bl <- log(999)
        eta_tg_bl~999
        tg_kel <- log(999)
        tg_ec50 <- log(5000)
        tg_emax_kel <- log(2)
        ktr_tg <- log(999)
        prop_err_tg <- 999
      })
      model({
        # PK setup
        f_sc <- exp(tf_sc)
        f_infilt <- exp(tf_infilt)
        ka_sc <- exp(tka_sc)
        ka_infilt <- exp(tka_infilt)
        cl_low <- exp(tcl_low + eta_cl)*(WEIGHT_BL/85)^e_wt_cl
        cl_high <- exp(tcl_high + eta_cl)*(WEIGHT_BL/85)^e_wt_cl
        cl_c50 <- exp(tcl_c50)
        v <- exp(tv + eta_v)
        q1 <- exp(tq1)
        vp1 <- exp(tvp1)
        q2 <- exp(tq2)
        vp2 <- exp(tvp2)
        # PK micro-parameters
        ke_low <- cl_low/v
        ke_high <- cl_high/v
        kc_p1 <- q1/v
        kp1_c <- q1/vp1
        kc_p2 <- q2/v
        kp2_c <- q2/vp2
        # TG setup
        tgbl <- exp(tg_bl + eta_tg_bl)
        kin_tg <- tgbl*exp(tg_kel)
        ktr_TG <- exp(ktr_tg)
        TG(0) <- tgbl
        # differential equations
        cp <- CENTRAL/v*1e3 # 1e3 is for unit conversion
        ke <- ke_low + (ke_high - ke_low)*cp/(cp + cl_c50)
        kout_tg <- exp(tg_kel) + exp(tg_emax_kel)*TG_TR/(TG_TR + exp(tg_ec50))
        d/dt(IVINFILT) =             - ka_infilt * IVINFILT
        d/dt(SC)       = -ka_sc * SC
        d/dt(CENTRAL)  =  ka_sc * SC + ka_infilt * IVINFILT - ke*CENTRAL - kc_p1*CENTRAL + kp1_c*P1 - kc_p2*CENTRAL + kp2_c*P2
        d/dt(P1) =                                                         kc_p1*CENTRAL - kp1_c*P1
        d/dt(P2) =                                                                                    kc_p2*CENTRAL - kp2_c*P2
        f(SC) <- f_sc
        f(IVINFILT) <- f_infilt
        # TG transit model
        d/dt(TG_TR) = ktr_tg*cp - ktr_tg*TG_TR
        d/dt(TG) = kin_tg - kout_tg*TG
        # Residual error models
        cp ~ prop(prop_err)
        TG ~ prop(prop_err_tg)
      })
    }

    tmp <- rxFixPop(nlmixr_threecmt_mm_no_add_wtcl_pdtg_kout_delay2)

    expect_equal(tmp$theta,
                 c(tf_sc = 6.90675477864855, tf_infilt = 6.90675477864855, tka_sc = 6.90675477864855,
                   tka_infilt = 6.90675477864855, tcl_low = 6.90675477864855, tcl_high = 6.90675477864855,
                   tcl_c50 = 8.00636756765025, tv = 6.90675477864855, tq1 = 6.90675477864855,
                   tvp1 = 2.30258509299405, tq2 = 6.90675477864855, tvp2 = 2.99573227355399,
                   prop_err = 999, tg_bl = 6.90675477864855, tg_kel = 6.90675477864855,
                   tg_ec50 = 8.51719319141624, tg_emax_kel = 0.693147180559945,
                   ktr_tg = 6.90675477864855, prop_err_tg = 999))


  })

  test_that("rxFixPop excludes a fixed mixture proportion (mix() needs it named)", {
    # a fixed mixture proportion must NOT be literally substituted -- mix() requires
    # the proportion to be a named model-block variable, so re-parsing the
    # substituted model would throw from mix() (the error leaked to the console when
    # a downstream caller wrapped rxFixPop in try()).
    .mixMod <- function() {
      ini({
        tcl1 <- log(2); tcl2 <- log(0.5); tv <- log(32)
        eta.v ~ 0.1
        p1 <- fix(0.5)
        add.sd <- 0.7
      })
      model({
        v <- exp(tv + eta.v)
        cl <- mix(exp(tcl1), p1, exp(tcl2))
        ke <- cl / v
        d/dt(depot) <- -depot
        d/dt(center) <- depot - ke * center
        cp <- center / v
        cp ~ add(add.sd)
      })
    }
    .ui <- rxode2(.mixMod)
    # the only fixed theta is the mixture proportion, so there is nothing to
    # substitute -> NULL, and crucially no error from re-parsing mix().
    expect_error(.res <- rxFixPop(.ui, returnNull = TRUE), NA)
    expect_null(.res)
    # p1 is still a named parameter in the model
    expect_true("p1" %in% .ui$iniDf$name)
  })

})

rxTest({
  test_that("ordinal style estimation log-likelihood line check", {
    
    f <- function() {
      ini({
        tkel <- 0.1
        tp0 <- -3
        eta.p ~ 0.02
        add.sd <- 0.2
      })
      model({
        kel <- tkel
        d/dt(kpd) <- -kel * kpd
        p1 <- expit(tp0 + eta.p)
        kpd ~ add(add.sd)
        cac ~ c(p1)
      })
    }

    tmp <- rxode2(f)

    tst1 <- .handleSingleErrTypeNormOrTFoceiBase(tmp, tmp$predDf[2,])
    
    w <- which(vapply(seq_along(tst1), function(i) {
      identical(tst1[[i]][[2]],  quote(`rx_pred_`))
    }, logical(1)))

    expect_equal(deparse1(tst1[[w]]), "rx_pred_ ~ log((DV == 1) * (p1) + (DV == 2) * (1 - (p1)))")


    f <- function() {
      ini({
        tkel <- 0.1
        tp0 <- -3
        eta.p ~ 0.02
        add.sd <- 0.2
      })
      model({
        kel <- tkel
        d/dt(kpd) <- -kel * kpd
        p1 <- expit(tp0 + eta.p)
        p2 <- 0.005
        p3 <- 0.005
        kpd ~ add(add.sd)
        cac ~ c(p1, p2, p3)
      })
    }

    tmp <- rxode2(f)

    tst1 <- .handleSingleErrTypeNormOrTFoceiBase(tmp, tmp$predDf[2,])
    
    w <- which(vapply(seq_along(tst1), function(i) {
      identical(tst1[[i]][[2]],  quote(`rx_pred_`))
    }, logical(1)))

    expect_equal(deparse1(tst1[[w]]), "rx_pred_ ~ log((DV == 1) * (p1) + (DV == 2) * (p2) + (DV == 3) * (p3) + (DV == 4) * (1 - (p1) - (p2) - (p3)))")

    f <- function() {
      ini({
        tkel <- 0.1
        tp0 <- -3
        eta.p ~ 0.02
        add.sd <- 0.2
      })
      model({
        kel <- tkel
        d/dt(kpd) <- -kel * kpd
        p1 <- expit(tp0 + eta.p)
        p2 <- 0.005
        p3 <- 0.005
        kpd ~ add(add.sd)
        cac ~ c(p1=0, p2=1, p3=2, 3)
      })
    }

    tmp <- rxode2(f)

    tst1 <- .handleSingleErrTypeNormOrTFoceiBase(tmp, tmp$predDf[2,])


    w <- which(vapply(seq_along(tst1), function(i) {
      identical(tst1[[i]][[2]],  quote(`rx_pred_`))
    }, logical(1)))

    expect_equal(deparse1(tst1[[w]]),
                 "rx_pred_ ~ log((DV == 0) * (p1) + (DV == 1) * (p2) + (DV == 2) * (p3) + (DV == 3) * (1 - p1 - p2 - p3))")

  })
})

rxTest({
  test_that("modelExtract and related functions", {

    one.compartment <- function() {
      ini({
        tka <- 0.45 # Log Ka
        tcl <- 1 # Log Cl
        tv <- 3.45    # Log V
        fDepot <-1
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        f(depot) <- fDepot
        depot(0) <- fDepot
        dur(depot) <- fDepot
        rate(depot) <- fDepot
        lag(depot) <- fDepot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    f <- one.compartment()

    tmp <- "d/dt(center)"

    expect_equal(modelExtract(f, tmp),
                 "d/dt(center) = ka * depot - cl/v * center")

    tmp <- list(tmp="d/dt(center)")
    expect_equal(modelExtract(f, tmp$tmp),
                 "d/dt(center) = ka * depot - cl/v * center")


    tmp <- list(tmp=list(tmp="d/dt(center)"))
    expect_equal(modelExtract(f, tmp$tmp$tmp),
                 "d/dt(center) = ka * depot - cl/v * center")

    expect_equal(modelExtract(f, "f(depot)"),
                 "f(depot) <- fDepot")

    expect_equal(modelExtract(f, "alag(depot)"),
                 "lag(depot) <- fDepot")

    expect_equal(modelExtract(f, "lag(depot)"),
                 "lag(depot) <- fDepot")

    expect_equal(modelExtract(f, "dur(depot)"),
                 "dur(depot) <- fDepot")

    expect_equal(modelExtract(f, "rate(depot)"),
                 "rate(depot) <- fDepot")

    expect_equal(modelExtract(f, "depot(0)"),
                 "depot(0) <- fDepot")

    one.compartment <- function() {
      ini({
        tka <- 0.45 # Log Ka
        tcl <- 1 # Log Cl
        tv <- 3.45    # Log V
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    f <- one.compartment()

    expect_equal(modelExtract(f, endpoint=NA, lines=TRUE, expression=TRUE),
                 structure(list(quote(ka <- exp(tka + eta.ka)),
                                quote(cl <- exp(tcl + eta.cl)),
                                quote(v <- exp(tv + eta.v)),
                                str2lang("d/dt(depot) = -ka * depot"),
                                str2lang("d/dt(center) = ka * depot - cl/v * center"),
                                str2lang("cp = center/v"),
                                quote(cp ~ add(add.sd))),
                           lines = 1:7))

    expect_equal(modelExtract(f, "ka", expression=FALSE, endpoint=FALSE, lines=TRUE),
                 structure("ka <- exp(tka + eta.ka)", lines = 1L))

    expect_equal(modelExtract(f, cl, expression=TRUE),
                 list(quote(cl <- exp(tcl + eta.cl))))

    expect_equal(modelExtract(f, cl, expression=FALSE),
                 "cl <- exp(tcl + eta.cl)")

    expect_equal(modelExtract(f, "cp", expression=FALSE, endpoint=NA),
                 c("cp = center/v", "cp ~ add(add.sd)"))

    expect_equal(modelExtract(f, "cp", expression=FALSE, endpoint=TRUE),
                 "cp ~ add(add.sd)")

    expect_equal(modelExtract(f, "cp", expression=FALSE, endpoint=FALSE),
                 "cp = center/v")

    one.compartment <- function() {
      ini({
        tka <- 0.45 # Log Ka
        tcl <- 1 # Log Cl
        tv <- 3.45    # Log V
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- tcl
        cl <- cl*exp(eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    expect_equal(modelExtract(one.compartment, "cl", expression=FALSE, endpoint=FALSE),
                 c("cl <- tcl", "cl <- cl * exp(eta.cl)"))

    mod <- rxode2({
      ka <- exp(tka + eta.ka)
      cl <- tcl
      cl <- cl*exp(eta.cl)
      v <- exp(tv + eta.v)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl / v * center
      cp = center / v
    })

    expect_equal(modelExtract(mod, "v", expression=FALSE, endpoint=FALSE),
                 "v = exp(tv + eta.v)")

  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("as.model expression", {

      is.model <- function(x) {
        expect_true(is.call(x))
        expect_true(identical(x[[1]], quote(`model`)))
      }

      model <- quote(model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      }))

      is.model(as.model(model))

      one.compartment <- function() {
        ini({
          tka <- log(1.57)
          tcl <- log(2.72)
          tv <- log(31.5)
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.sd <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
          cp ~ add(add.sd)
        })
      }

      is.model(as.model(one.compartment))

      ui <- one.compartment()

      is.model(as.model(ui))

      model <- c("model({",
                 "ka <- exp(tka + eta.ka)",
                 "cl <- exp(tcl + eta.cl)",
                 "v <- exp(tv + eta.v)",
                 "d/dt(depot) = -ka * depot",
                 "d/dt(center) = ka * depot - cl / v * center",
                 "cp = center / v",
                 "cp ~ add(add.sd)",
                 "})")

      is.model(as.model(model))

      model <- paste(model, collapse="\n")

      is.model(as.model(model))

    })
  }
})

rxTest({

  .isAltrep <- function(x) {
    rxIs(x, "altrep")
  }

  .isRepint <- function(x) {
    rxIs(x, "repint")
  }

  .isSeqrep <- function(x) {
    rxIs(x, "seqrep")
  }

  .isRepstr <- function(x) {
    rxIs(x, "repstr")
  }

  for (rt in c("data.frame", "tibble", "rxSolve")) {
    for (ad in c(TRUE, FALSE, NA))  {

      test_that(sprintf("repeated simulation event columns use ALTREP repetition %s, %s",
                        rt, ad), {

        mod <- rxode2({
          d/dt(depot) <- -ka * depot
          d/dt(centr) <- ka * depot - cl / v * centr
          cp <- centr / v
        })

        ev <- et(amt = 100, cmt = 1, ii = 12, addl = 1) |>
          et(seq(0, 24, by = 1))

        p <- data.frame(id = 1:2, ka = c(1, 1), cl = c(1, 1), v = c(10, 10))
        out <- rxSolve(mod, p, ev,
                       addDosing = ad,
                       returnType = rt)

          expect_true(.isAltrep(out$sim.id))
          # simSubjectPath: id column is absent (subjects identified by sim.id)
          expect_false("id" %in% names(out))
          expect_true(.isAltrep(out$time))
          if (!identical(ad, FALSE)) {
            expect_true(.isAltrep(out$evid))
            expect_true(.isAltrep(out$amt))
          }
      })

      test_that(sprintf("runtime event mutation fallback keeps event columns materialized %s %s", rt, ad), {

        mod <- rxode2({
          d/dt(x) <- -x
          if (t < 1) evid_(t + 0.1, 1, 10, 1, 0, 0, 0, 0)
        })
        ev <- et(amt = 1, time = 0) |>
          et(seq(0, 1, by = 0.1))

        p <- data.frame(id = 1:2, x = c(1, 1))
        out <- rxSolve(mod, p, ev,
                       addDosing = ad, returnType = rt)

        expect_false(.isAltrep(out$time))
      })

      # -- covariate (addCov = TRUE) columns --------------------------------------

      test_that(sprintf("covariate column is ALTREP via homogenous event-table path (%s %s)",
                        rt, ad), {

        # Identical subjects -> homogenous ET optimization -> rx->nsim > 1
        mod <- rxode2({
          d/dt(depot) <- -ka * depot
          d/dt(centr) <- ka * depot - (cl * (wt / 70)^0.75) / v * centr
          cp <- centr / v
        })
        ev <- et(amt = 100, cmt = 1, ii = 12, addl = 1) |>
          et(seq(0, 24, by = 1))
        evWt <- cbind(ev, wt = 70.0)   # same covariate for every subject
        p <- data.frame(id = 1:3, ka = 0.5, cl = 1, v = 10)

        out <- rxSolve(mod, p, evWt, addCov = TRUE, returnType = rt)

        expect_true(.isAltrep(out$wt))
        # simSubjectPath: id column is absent (subjects identified by sim.id)
        expect_false("id" %in% names(out))

        expect_true(all(out$wt == 70))
      })

      test_that(sprintf("covariate columns are ALTREP via explicit nsim (stochastic) path (%s %s)",
                        rt, ad), {

        # Multiple virtual studies from omega draws; covariate values are fixed
        # per-subject across sims -> each sim block is identical -> ALTREP.
        pkCov <- function() {
          ini({
            tka <- 0.5
            tcl <- 0
            tv <- log(10)
            eta.ka ~ 0.3
            eta.cl ~ 0.2
            eta.v ~ 0.1
          })
          model({
            ka <- exp(tka + eta.ka)
            cl <- exp(tcl + eta.cl) * (wt / 70)^0.75 * (age / 40)^0.3
            v  <- exp(tv + eta.v)
            d/dt(depot) <- -ka * depot
            d/dt(centr) <- ka * depot - cl / v * centr
            cp <- centr / v
          })
        }
        evDf <- as.data.frame(
          et(amt = 100, cmt = 1) |> et(seq(0, 24, by = 1)) |> et(id = 1:3)
        )
        n <- sum(evDf$id == 1L)
        evDf$wt  <- rep(c(70, 80, 90), each = n)
        evDf$age <- rep(c(30, 40, 50), each = n)

        set.seed(42)
        out <- rxSolve(pkCov, evDf, nsim = 3, addCov = TRUE,
                       returnType = rt)

        if (!identical(rt, "data.table")) {
          expect_true(.isAltrep(out$wt))
          expect_true(.isAltrep(out$age))
        }
        for (.id in 1:3) {
          expect_equal(unique(out$wt[out$id  == .id]), c(70, 80, 90)[.id])
          expect_equal(unique(out$age[out$id == .id]), c(30, 40, 50)[.id])
        }
      })

      # -- keep columns -----------------------------------------------------------

      test_that(paste0("keep columns (double, integer, logical, factor) are ALTREP via homogenous event-table path, %s %s", rt, ad), {
        mod <- rxode2({
          d/dt(depot) <- -ka * depot
          d/dt(centr) <- ka * depot - cl / v * centr
          cp <- centr / v
        })
        ev <- et(amt = 100, cmt = 1, ii = 12, addl = 1) |> et(seq(0, 24, by = 1))
        # cbind onto a single-subject event table; identical for every subject
        evDf <- cbind(ev, score = 1.5, code = 2L, flag = TRUE, label = "low")
        evDf$grp <- factor(rep("A", nrow(evDf)), levels = c("A", "B"))
        p <- data.frame(id = 1:3, ka = 0.5, cl = 1, v = 10)

        out <- rxSolve(mod, p, evDf, returnType = rt,
                       keep = c("score", "code", "flag", "grp", "label"))

        if (!identical(rt, "data.table")) {
          expect_true(.isAltrep(out$score))
          expect_true(.isAltrep(out$code))
          expect_true(.isAltrep(out$flag))
          expect_true(.isAltrep(out$grp))
          # Factor levels must survive the ALTREP wrapping
          expect_equal(levels(out$grp), c("A", "B"))
          expect_true(.isAltrep(out$label))
        }
        # All values must be correct
        expect_true(all(out$score == 1.5))
        expect_true(all(out$label == "low"))
      })

      test_that(sprintf("nStud > 1 makes event-table and keep columns ALTREP (theo_sd / WT) %s %s", rt, ad), {

        skip_if_not_installed("nlmixr2data")
        one.cmt <- function() {
          ini({
            tka <- 0.45
            tcl <- log(c(0, 2.7, 100))
            tv <- 3.45
            eta.ka ~ 0.6
            eta.cl ~ 0.3
            eta.v ~ 0.1
            add.sd <- 0.7
          })
          model({
            ka <- exp(tka + eta.ka)
            cl <- exp(tcl + eta.cl)
            v  <- exp(tv + eta.v)
            linCmt() ~ add(add.sd)
          })
        }
        d <- nlmixr2data::theo_sd

        set.seed(42)
        out <- rxSolve(one.cmt, d, nStud = 4, keep = "WT",
                       addDosing = ad, returnType = rt)


        # Event-table columns that repeat identically across studies must be ALTREP.
        expect_true(.isAltrep(out$sim.id))
        expect_true(max(out$sim.id) == 4)
        expect_true(.isSeqrep(out$id))
        expect_true(.isAltrep(out$time))

        if (!identical(ad, FALSE)) {
          expect_true(.isAltrep(out$evid))
          expect_true(.isAltrep(out$amt))
        }
        # The kept WT covariate must also be ALTREP.
        expect_true(.isAltrep(out$WT))


        # WT must be constant within each output id and match the source data.
        wtRef <- setNames(vapply(unique(d$ID),
                                 function(i) {
                                   d$WT[d$ID == i][1]
                                 }, numeric(1)),
                          unique(d$ID))
        nOrig <- length(unique(d$ID))
        for (.id in unique(out$id)) {
          # id cycles 1..nOrig across each study block
          origId <- ((.id - 1L) %% nOrig) + 1L
          wtVals <- unique(out$WT[out$id == .id])
          expect_length(wtVals, 1)
          expect_equal(wtVals, wtRef[[as.character(origId)]])
        }
      })

      test_that(paste0("keep columns (double, integer, logical, factor) are ALTREP via explicit nsim (stochastic) path %s %s", rt, ad), {
        # Per-subject varying keep values; same values in every sim block -> ALTREP.
        pkKeep <- function() {
          ini({
            tka <- 0.5
            tcl <- 0
            tv <- log(10)
            eta.ka ~ 0.3
            eta.cl ~ 0.2
            eta.v ~ 0.1
          })
          model({
            ka <- exp(tka + eta.ka)
            cl <- exp(tcl + eta.cl)
            v  <- exp(tv + eta.v)
            d/dt(depot) <- -ka * depot
            d/dt(centr) <- ka * depot - cl / v * centr
            cp <- centr / v
          })
        }
        evDf <- as.data.frame(
          et(amt = 100, cmt = 1) |> et(seq(0, 24, by = 1)) |> et(id = 1:3)
        )
        n <- sum(evDf$id == 1L)
        evDf$score <- rep(c(1.5, 2.5, 3.5),          each = n)
        evDf$code  <- rep(c(1L,  2L,  3L),            each = n)
        evDf$flag  <- rep(c(TRUE, FALSE, TRUE),        each = n)
        evDf$grp   <- factor(rep(c("A", "B", "A"),    each = n), levels = c("A", "B"))
        evDf$label <- rep(c("low", "high", "low"),     each = n)

        set.seed(42)
        out <- rxSolve(pkKeep, evDf, nsim = 3,
                       keep = c("score", "code", "flag", "grp", "label"),
                       returnType = rt)

        expect_true(.isAltrep(out$score))
        expect_true(.isAltrep(out$code))
        expect_true(.isAltrep(out$flag))
        expect_true(.isAltrep(out$grp))
        # Factor levels must survive the ALTREP wrapping
        expect_equal(levels(out$grp), c("A", "B"))
        expect_true(.isAltrep(out$label))
        # Values correct for each subject across all sims
        for (.id in 1:3) {
          .rows <- out$id == .id
          expect_equal(unique(out$score[.rows]), c(1.5, 2.5, 3.5)[.id])
          expect_equal(unique(out$code[.rows]),  c(1L,  2L,  3L)[.id])
          expect_equal(unique(out$label[.rows]), c("low", "high", "low")[.id])
        }
      })
    }
  }

  # -- TBS columns (returnType = "data.frame.TBS") ---------------------------

  for (ad in c(TRUE, FALSE, NA)) {

    test_that(sprintf("TBS columns are ALTREP via homogeneous event-table path (addDosing=%s)", ad), {
      mod <- rxode2({
        d/dt(depot) <- -ka * depot
        d/dt(centr) <- ka * depot - cl / v * centr
        cp <- centr / v
      })
      ev <- et(amt = 100, cmt = 1, ii = 12, addl = 1) |> et(seq(0, 24, by = 1))
      p <- data.frame(id = 1:3, ka = 0.5, cl = 1, v = 10)

      out <- rxSolve(mod, p, ev, addDosing = ad, returnType = "data.frame.TBS")

      expect_true(.isAltrep(out$rxLambda))
      expect_true(.isAltrep(out$rxYj))
      expect_true(.isAltrep(out$rxLow))
      expect_true(.isAltrep(out$rxHi))
      # Values must be replicated correctly
      expect_true(all(out$rxLambda == out$rxLambda[1]))
      expect_true(all(out$rxYj    == out$rxYj[1]))
    })

    test_that(sprintf("TBS columns are ALTREP via nStud path (addDosing=%s)", ad), {
      skip_if_not_installed("nlmixr2data")
      one.cmt <- function() {
        ini({
          tka <- 0.45
          tcl <- log(c(0, 2.7, 100))
          tv <- 3.45
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.sd <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v  <- exp(tv + eta.v)
          linCmt() ~ add(add.sd)
        })
      }
      d <- nlmixr2data::theo_sd

      set.seed(42)
      out <- rxSolve(one.cmt, d, nStud = 3, addDosing = ad,
                     returnType = "data.frame.TBS")

      expect_true(.isAltrep(out$rxLambda))
      expect_true(.isAltrep(out$rxYj))
      expect_true(.isAltrep(out$rxLow))
      expect_true(.isAltrep(out$rxHi))
      expect_true(all(out$rxLambda == out$rxLambda[1]))
      expect_true(all(out$rxYj    == out$rxYj[1]))
    })
  }

  # -- Sequential integer IDs should never be a factor ----------------------

  test_that("sequential integer IDs are never a factor, regardless of event-table uniformity", {
    mod <- rxode2({
      d/dt(depot) <- -ka * depot
      d/dt(centr) <- ka * depot - cl / v * centr
      cp <- centr / v
    })

    # Non-uniform event tables (arm1 vs arm2 style) -- non-ALTREP path.
    # Previously this produced a factor id due to a stale guard; verify it
    # now returns a plain integer.
    ev1 <- as.data.frame(et(amt = 1000, cmt = 1, time = c(0, 24, 48)) |> et(seq(0, 72, by = 5)))
    ev2 <- as.data.frame(et(amt = 2000, cmt = 1, time = c(5, 29, 53)) |> et(seq(0, 77, by = 6)))
    ev1$id <- 1L; ev2$id <- 2L
    ev3 <- ev1; ev3$id <- 3L; ev4 <- ev2; ev4$id <- 4L
    evDf <- rbind(ev1, ev2, ev3, ev4)
    p <- data.frame(id = 1:4, ka = 0.5, cl = 1, v = 10)

    out <- rxSolve(mod, p, evDf, returnType = "data.frame")
    expect_false(is.factor(out$id))
    expect_equal(sort(unique(out$id)), 1:4)

    # Uniform event tables -- ALTREP seqrep path.  Should also not be a factor.
    ev <- et(amt = 100, cmt = 1, ii = 12, addl = 1) |> et(seq(0, 24, by = 1))
    evDf2 <- rbind(
      cbind(as.data.frame(ev), id = 1L),
      cbind(as.data.frame(ev), id = 2L),
      cbind(as.data.frame(ev), id = 3L),
      cbind(as.data.frame(ev), id = 4L)
    )
    out2 <- rxSolve(mod, p, evDf2, returnType = "data.frame")
    expect_false(is.factor(out2$id))
    expect_equal(sort(unique(out2$id)), 1:4)
  })

  # -- Non-sequential integer IDs --------------------------------------------

  test_that("non-sequential integer IDs produce ALTREP repint id column (nStud > 1)", {
    pkMod <- function() {
      ini({
        tka <- 0.5
        tcl <- 0
        tv  <- log(10)
        eta.ka ~ 0.3
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl)
        v  <- exp(tv)
        d/dt(depot) <- -ka * depot
        d/dt(centr) <- ka * depot - cl / v * centr
        cp <- centr / v
      })
    }
    # IDs must come from the event table; the parameter data frame is matched
    # by position, not by its id column.
    evDf <- as.data.frame(
      et(amt = 100, cmt = 1, ii = 12, addl = 1) |>
        et(seq(0, 24, by = 1)) |>
        et(id = c(5L, 10L, 15L))
    )
    set.seed(42)
    out <- rxSolve(pkMod, evDf, nStud = 2)

    # nStud=2: fill loop writes lvlI values directly per subject; re-wrap block
    # creates rep_int(base, times=2) ALTREP since the pattern is uniform across sims.
    expect_true(.isAltrep(out$id))
    expect_true(.isRepint(out$id))
    expect_false(is.factor(out$id))
    expect_equal(sort(unique(out$id)), c(5L, 10L, 15L))
    for (.id in c(5L, 10L, 15L)) {
      expect_true(all(out$id[out$id == .id] == .id))
    }
  })

  # -- String keep columns -- dedicated rx_rep_str ALTREP tests --------------

  test_that("string keep column is rx_rep_str ALTREP via homogeneous path", {
    mod  <- rxode2({
      d/dt(depot) <- -ka * depot
      d/dt(centr) <- ka * depot - cl / v * centr
      cp <- centr / v
    })
    ev   <- et(amt = 100, cmt = 1, ii = 12, addl = 1) |> et(seq(0, 24, by = 1))
    evDf <- cbind(ev, label = "low")
    p    <- data.frame(id = 1:3, ka = 0.5, cl = 1, v = 10)

    out <- rxSolve(mod, p, evDf, keep = "label")

    expect_true(.isAltrep(out$label))
    expect_true(.isRepstr(out$label))
    expect_true(all(out$label == "low"))
    expect_null(attr(out$label, "class"))
  })

  test_that("string keep column is rx_rep_str ALTREP via stochastic nsim path", {
    pkKeep <- function() {
      ini({ tka <- 0.5; tcl <- 0; tv <- log(10); eta.ka ~ 0.3 })
      model({
        ka <- exp(tka + eta.ka); cl <- exp(tcl); v <- exp(tv)
        d/dt(depot) <- -ka * depot
        d/dt(centr) <- ka * depot - cl / v * centr
        cp <- centr / v
      })
    }
    evDf <- as.data.frame(
      et(amt = 100, cmt = 1) |> et(seq(0, 24, by = 1)) |> et(id = 1:3)
    )
    n <- sum(evDf$id == 1L)
    evDf$label <- rep(c("low", "high", "low"), each = n)

    set.seed(42)
    out <- rxSolve(pkKeep, evDf, nsim = 3, keep = "label")

    expect_true(.isAltrep(out$label))
    expect_true(.isRepstr(out$label))
    for (.id in 1:3) {
      expect_equal(unique(out$label[out$id == .id]),
                   c("low", "high", "low")[.id])
    }
  })

  test_that("single-sim string keep column is NOT ALTREP", {
    mod  <- rxode2({ d/dt(A) <- -k * A })
    ev   <- et(seq(0, 12, by = 1)) |> et(amt = 100)
    evDf <- cbind(ev, label = "ctrl")

    out <- rxSolve(mod, c(k = 0.1), evDf, keep = "label")

    expect_false(.isAltrep(out$label))
    expect_equal(unique(out$label), "ctrl")
  })

  test_that("string ALTREP column materialises correctly", {
    mod  <- rxode2({
      d/dt(depot) <- -ka * depot
      d/dt(centr) <- ka * depot - cl / v * centr
    })
    ev   <- et(amt = 100, cmt = 1) |> et(seq(0, 24, by = 1))
    evDf <- cbind(ev, grp = "A")
    p    <- data.frame(id = 1:4, ka = 0.5, cl = 1, v = 10)

    out <- rxSolve(mod, p, evDf, keep = "grp")
    expect_true(.isAltrep(out$grp))

    # Force full materialisation via DATAPTR then verify values are correct
    mat <- as.character(out$grp)
    expect_true(all(mat == "A"))
    expect_equal(length(mat), nrow(out))
  })

  test_that("element access on string ALTREP column exercises Elt without materialising", {
    mod  <- rxode2({ d/dt(A) <- -k * A })
    ev   <- et(seq(0, 10, by = 1)) |> et(amt = 1)
    evDf <- cbind(ev, trt = "drug")
    p    <- data.frame(id = 1:5, k = 0.2)

    out <- rxSolve(mod, p, evDf, keep = "trt")
    expect_true(.isAltrep(out$trt))

    expect_equal(out$trt[[1L]], "drug")
    expect_equal(out$trt[[nrow(out)]], "drug")
    expect_true(.isAltrep(out$trt))
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("lhs ui accessors", {

      oneCmtAllo <- function() {
        ini({
          lka <- log(0.1)
          label("Absorption rate (Ka)")
          lcl <- log(0.2)
          label("Clearance (CL)")
          lvc <- log(1)
          label("Central volume of distribution (V)")
          cppropSd <- c(0, 0.5)
          allo_cl <- fix(0.75)
          allo_vc <- fix(1)
          etalcl ~ 0.1
        })
        model({
          ka <- exp(lka)
          cl <- exp(lcl + allo_cl * log(WEIGHT_BL/100) + etalcl)
          vc <- exp(lvc + allo_vc * log(WEIGHT_BL/100))
          cp <- 1000 * linCmt()
          cp ~ prop(cppropSd)
        })
      }

      f <- rxode2(oneCmtAllo)


      expect_equal(f$lhsVar,
                   c(cl = "etalcl", ka = "lka", cl = "lcl", cl = "allo_cl", vc = "lvc", vc = "allo_vc", cl = "WEIGHT_BL", vc = "WEIGHT_BL"))

      expect_equal(f$varLhs,
                   c(etalcl = "cl", lka = "ka", lcl = "cl", allo_cl = "cl", lvc = "vc", allo_vc = "vc", WEIGHT_BL = "cl", WEIGHT_BL = "vc"))


      expect_equal(f$etaLhs, c(etalcl = "cl"))
      expect_equal(f$lhsEta, c(cl = "etalcl"))

      expect_equal(f$thetaLhs, c(lka = "ka", lcl = "cl", allo_cl = "cl", lvc = "vc", allo_vc = "vc"))
      expect_equal(f$lhsTheta, c(ka = "lka", cl = "lcl", cl = "allo_cl", vc = "lvc", vc = "allo_vc"))

      expect_equal(f$covLhs,
                   c(WEIGHT_BL = "cl", WEIGHT_BL = "vc"))

      expect_equal(f$lhsCov,
                   c(cl = "WEIGHT_BL", vc = "WEIGHT_BL"))

    })
  }
})

rxTest({

  expect_plotlog <- function(o, timex, logx, logy, dat) {
    # Checking for the correct type for logx and logy is nontrivial, so j
    expect_named(o, c("timex", "logx", "logy", "dat"))
    if (is.null(timex)) {
      expect_null(o$timex)
    } else {
      expect_equal(o$timex, timex)
    }
    if (is.null(logx)) {
      expect_null(o$logx)
    } else {
      expect_type(o$logx, "list")
      expect_length(o$logx, 1)
      expect_s3_class(o$logx[[1]], logx)
    }
    if (is.null(logy)) {
      expect_null(o$logy)
    } else {
      expect_type(o$logy, "list")
      expect_length(o$logy, 1)
      expect_s3_class(o$logy[[1]], logy)
    }
    expect_equal(o$dat, dat)
  }

  test_that(".plotLog works without xgxr", {
    skip_on_os("mac")
    d <- data.frame(time=-1:2, conc=0:3)
    expect_equal(
      .plotLog(.dat=d, .timex="A", log=""),
      list(
        timex="A",
        logx=NULL,
        logy=NULL,
        dat=d
      )
    )
    expect_plotlog(
      .plotLog(.dat=d, .timex="A", log="x"),
      timex=NULL, logx="ScaleContinuousPosition", logy=NULL, dat=d[-(1:2), ]
    )
    expect_plotlog(
      .plotLog(.dat=d, .timex="A", log="y"),
      timex="A", logx=NULL, logy="ScaleContinuousPosition", dat=d
    )
    expect_plotlog(
      .plotLog(.dat=d, .timex="A", log="xy"),
      timex=NULL, logx="ScaleContinuousPosition", logy="ScaleContinuousPosition", dat=d[-(1:2), ]
    )
  })

  test_that(".plotLog works with xgxr", {
    skip_if_not_installed("xgxr")
    skip_on_os("mac")
    skip("See https://github.com/Novartis/xgxr/issues/50 for why xgxr is not uniquely tested as of 2022-03-15")
    current_xgxr_option <- getOption("rxode2.xgxr")
    withr::with_options(
      list(rxode2.xgxr=TRUE), {
        d <- data.frame(time=-1:2, conc=0:3)
        # Insert tests here
      }
    )
  })

  test_that(".plotLog gives expected errors", {
    skip_on_os("mac")
    expect_error(.plotLog(log=c("x", "y")))
    expect_error(.plotLog(log=1))
    expect_error(.plotLog(log="foo"))
    expect_error(.plotLog(log=c("x", "y")))
    expect_error(.plotLog(.dat=data.frame(A=1), log="x", .timex="A"))
    # Time column is only required when log="x"
    expect_silent(.plotLog(.dat=data.frame(A=1), log="y", .timex="A"))
  })

  test_that("plot() with invalid component throws an error", {
    skip_on_os("mac")

    pheno2 <- function() {
      ini({
        tcl <- log(0.008)
        tv <-  log(0.6)
        eta.cl + eta.v ~ c(1,
                           0.01, 1)
        add.err <- 0.1
      })
      model({
        cl <- exp(tcl + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v)    # individual value of volume
        ke <- cl / v            # elimination rate constant
        d/dt(A1) = - ke * A1    # model differential equation
        cp = A1 / v             # concentration in plasma
        cp ~ add(add.err)  # define error model
      })
    }

    simdata <- data.frame(time=1:10, ID=factor(c("A", "B")))

    sim <- rxSolve(pheno2, events=simdata)

    expect_error(plot(sim, "foo"))
    expect_warning(plot(sim, sim, "foo"))
    expect_error(plot(sim), NA)
  })

})

rxTest({
  test_that("simulate zeros tests", {

    # from nonmem2R
    f <- function() {
      description <- "PK"
      ini({
        t.CL <- c(0, 27.7322)
        label("1 CL L/h parent")
        t.V1 <- c(0, 112.773)
        label("2 V1 L parent")
        t.V2 <- c(0, 98.2478)
        label("3 V2 L parent")
        t.Q1 <- c(0, 20.8085)
        label("4 Q1 parent")
        theta5 <- c(0, 0.168248)
        label("5  ERROR parent")
        t.MTT <- c(0, 0.938162)
        label("6 MTT")
        F1 <- fix(1)
        label("7 F1")
        t.CLM <- c(0, 15.6045)
        label("8 CLM metabolite")
        t.VM1 <- c(0, 9.69614)
        label("9 VM1 L metabolite")
        t.VM2 <- c(0, 44.5183)
        label("10 VM2 L metabolite")
        t.Q2 <- c(0, 5.55961)
        label("11 Q2 metabolite")
        theta12 <- c(0, 0.145536)
        label("12 ERROR metabolite")
        theta13 <- c(0, 1.65686)
        label("13 Add ERROR parent")
        theta14 <- c(0, 1.07706)
        label("14 Add ERROR metabolite")
        NTR <- c(1, 1.95835, 12)
        label("15 NTR")
        eta1 ~ 0.0704585
        e.Q1 ~ 0.0386847
        eta3 ~ 0.0430261
        e.V2 ~ fix(0)
        e.KTR ~ 0.146879
        e.NTR ~ fix(0)
        Bioavailability ~ 0.0506769
      })
      model({
        cmt(cent)
        cmt(centmet)
        cmt(peri)
        cmt(perimet)
        if (newind <= 1) {
          tdos <- -1000
          pd <- 0
        }
        if (amt > 0) {
          tdos <- t
          pd <- amt
        }
        tad2 <- t - tdos
        e1 <- exp(eta1)
        e2 <- exp(e.Q1)
        e3 <- exp(eta3)
        e4 <- exp(e.V2)
        e5 <- exp(e.KTR)
        e6 <- exp(e.NTR)
        e7 <- exp(Bioavailability)
        tvcl <- t.CL
        cl <- tvcl * e1
        tvv1 <- t.V1
        v1 <- tvv1
        tvv2 <- t.V2
        v2 <- tvv2
        tvq1 <- t.Q1
        q1 <- tvq1 * e2
        tvclm <- t.CLM
        clm <- tvclm * e3
        tvvm1 <- t.VM1
        vm1 <- tvvm1
        tvvm2 <- t.VM2
        vm2 <- tvvm2 * e4
        tvq2 <- t.Q2
        q2 <- tvq2
        mtt <- t.MTT * e5
        nt <- 1 + NTR * e6
        ktr <- nt/mtt
        nn <- nt - 1
        if (nn >= 0.558) {
          l <- log(2.5066) + (nn + 0.5) * log(nn) - nn + log(1 +
                                                               1/(12 * nn))
        } else {
          l <- -0.1171157 + 0.569 * ((0.558 - nn)^2) - 0.1122 *
            (0.558 - nn)
        }
        k13 <- q1/v1
        k31 <- q1/v2
        TVFM <- 0.22
        fm <- TVFM
        k10 <- cl * (1 - fm)/v1
        k12 <- fm * cl/v1
        k24 <- q2/vm1
        k42 <- q2/vm2
        k20 <- clm/vm1
        scale1 <- v1
        scale2 <- vm1
        rxf.rxddta1. <- 0
        f(cent) <- rxf.rxddta1.
        bio <- F1 * e7
        X <- 1e-07
        if (t >= tdos) {
          d/dt(cent) <- bio * pd * ktr * exp(nn * log(ktr *
                                                        (t - tdos) + X) - ktr * (t - tdos) - l) - k10 *
                                         cent - k12 * cent - k13 * cent + k31 * peri
        } else {
          d/dt(cent) <- -k12 * cent - k13 * cent + k31 * peri
        }
        d/dt(centmet) <- k12 * cent - k20 * centmet - k24 * centmet +
          k42 * perimet
        d/dt(peri) <- k13 * cent - k31 * peri
        d/dt(perimet) <- k24 * centmet - k42 * perimet
        f <- cent/scale1
        strt <- FLAG
        eps1 <- rxnorm()
        if (FLAG <= 1) {
          ipred <- cent/scale1
          w <- sqrt((theta5 * ipred)^2 + theta13^2)
          y <- ipred + w * eps1
        } else {
          ipred <- centmet/scale2
          w <- sqrt((theta12 * ipred)^2 + theta14^2)
          y <- ipred + w * eps1
        }
        cp <- cent/scale1
        cm <- centmet/scale2
        del <- 0
        if (w == 0)
          del <- 1
        ires <- DV - ipred
        iwres <- ires/(w + del)
      })
    }
    f <- f()

    e <- et(amt=100) |> et(seq(0,20))
    e$FLAG <- 1

    expect_error(rxSolve(f, e), NA)

    m <- f$simulationModel

    expect_error(rxSolve(m, params=f$theta, events=e, omega=f$omega), NA)

    # ok now try just the control

    .ctl <- rxControl(omega=lotri::lotri(eta1~c(0.0)))
    expect_equal(.ctl$omega, NULL)
    expect_equal(.ctl$.zeros, "eta1")

    .ctl <- rxControl(omega=lotri::lotri(eta1+eta2~c(0.0, 0.0, 1)))
    expect_equal(.ctl$omega, lotri::lotri(eta2~1.0))
    expect_equal(.ctl$.zeros, "eta1")

    .ctl <- rxControl(omega=lotri::lotri(eta1+eta2~c(0.0, 0.0, 1)),
                      omegaLower=c(-1, -1), omegaUpper=c(1,1))
    expect_equal(.ctl$omega, lotri::lotri(eta2~1.0))
    expect_equal(.ctl$.zeros, "eta1")
    expect_equal(.ctl$omegaLower, c(eta1=-1, eta2=-1))
    expect_equal(.ctl$omegaUpper, c(eta1=1, eta2=1))

    # sigma
    .ctl <- rxControl(sigma=lotri::lotri(eps1~c(0.0)))
    expect_equal(.ctl$sigma, NULL)
    expect_equal(.ctl$.zeros, "eps1")

    .ctl <- rxControl(sigma=lotri::lotri(eps1+eps2~c(0.0, 0.0, 1)))
    expect_equal(.ctl$sigma, lotri::lotri(eps2~1.0))
    expect_equal(.ctl$.zeros, "eps1")

    .ctl <- rxControl(sigma=lotri::lotri(eps1+eps2~c(0.0, 0.0, 1)),
                      sigmaLower=c(-1, -1), sigmaUpper=c(1,1))
    expect_equal(.ctl$sigma, lotri::lotri(eps2~1.0))
    expect_equal(.ctl$.zeros, "eps1")
    expect_equal(.ctl$sigmaLower, c(eps1=-1, eps2=-1))
    expect_equal(.ctl$sigmaUpper, c(eps1=1, eps2=1))

  })
})

rxTest({
  test_that("pred model errors", {
    fn1 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        linCmt() ~ pois()
      })
    }

    expect_error(rxode2(fn1), "linCmt()")

    fn2 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        Prop.Err2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        linCmt() ~ pow(Prop.Err, Prop.Err2)
      })
    }

    expect_error(rxode2(fn2), "Prop.Err")

    fn2 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        Prop.Err <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        linCmt() ~ pow(Prop.Err, Prop.Err2)
      })
    }

    expect_error(rxode2(fn2), "Prop.Err2")

    fn2 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        Prop.Err <- 1
        Prop.Err2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        linCmt() ~ pow(Prop.Err, Prop.Err2)
      })
    }

    expect_error(rxode2(fn2), NA)

    fn2 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        Prop.Err <- 1
        Prop.Err2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        linCmt() ~ prop(Prop.Err, Prop.Err2)
      })
    }

    expect_error(suppressMessages(rxode2(fn2)))


    fn3 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        linCmt() ~ prop()
      })
    }

    expect_message(
      expect_error(
        rxode2(fn3),
        "syntax/parsing errors, see above"
      ),
      regexp = "`prop` requires 1 argument(s), you specified 0",
      fixed = TRUE
    )

    fn4 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        p1 <- 1
        p2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        linCmt() ~ pois(p1, p2)
      })
    }

    expect_error(suppressMessages(rxode2(fn4)))

    fn5 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        linCmt() ~ pois()
      })
    }

    expect_error(rxode2(fn5))

    fn6 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        linCmt() ~ dpois(1, 2)
      })
    }

    expect_error(rxode2(fn6))

    fn7 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        linCmt() ~ dbinom()
      })
    }

    expect_error(rxode2(fn7))

    fn8 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ dbinom(13)
      })
    }

    expect_error(rxode2(f8))

    fn9 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ dbinom(13, 0.5, 32)
      })
    }

    expect_error(rxode2(fn9))

    fn10 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ dbeta()
      })
    }

    expect_error(rxode2(fn10))

    fn11 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ dbeta(par1)
      })
    }

    suppressMessages(
      expect_error(
        expect_warning(
          rxode2(fn11),
          regexp = "`beta` requires 2 argument(s), you specified 1"
        ),
        regexp = "syntax/parsing errors, see above"
      )
    )

    fn12 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
        par3 <- 1
        par4 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ dbeta(par1, par2, par3, par4)
      })
    }

    suppressMessages(
      expect_error(expect_message(rxode2(fn12)))
    )

    fn13 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ dt()
      })
    }

    expect_error(expect_message(rxode2(fn13)))

    fn14 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ dt(par1, par2, par3)
      })
    }

    suppressMessages(
      expect_error(expect_message(rxode2(fn14)))
    )

    fn15 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ binom()
      })
    }

    expect_error(expect_message(rxode2(fn15)))

    fn15 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ dexp()
      })
    }

    expect_error(rxode2(fn15), NA)

    fn16 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ binom(par1)
      })
    }

    expect_error(rxode2(fn16), NA)

    fn17 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
        par3 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ binom(par1, par2, par3)
      })
    }

    suppressMessages(
      expect_error(expect_message(rxode2(fn17)))
    )

    fn18 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ beta()
      })
    }

    expect_error(rxode2(fn18))

    fn19 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ beta(par1)
      })
    }

    suppressMessages(
      expect_error(expect_message(rxode2(fn19)))
    )

    fn20 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
        par3 <- 1
        par4 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ beta(par1, par2, par3, par4)
      })
    }

    suppressMessages(
      expect_error(expect_message(rxode2(fn20)))
    )

    fn21 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ t()
      })
    }

    expect_error(rxode2(fn21))

    fn22 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
        par3 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ t(par1, par2, par3)
      })
    }

    suppressMessages(
      expect_error(rxode2(fn22))
    )

    fn23 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ add()
      })
    }

    suppressMessages(
      expect_error(rxode2(fn23))
    )

    fn24 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ add(par1, par2)
      })
    }

    suppressMessages(
      expect_error(expect_message(rxode2(fn24)))
    )

    fn25 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ prop()
      })
    }

    suppressMessages(
      expect_error(rxode2(fn25))
    )

    fn26 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ prop(par1, par2)
      })
    }

    suppressMessages(
      expect_error(expect_message(rxode2(fn26)))
    )

    fn27 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ df()
      })
    }

    suppressMessages(
      expect_error(rxode2(fn27))
    )

    fn28 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ norm(par1, par2)
      })
    }

    suppressMessages(
      expect_error(expect_message(rxode2(fn28)))
    )

    fn29 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ dnorm()
      })
    }

    expect_error(rxode2(fn29))

    fn30 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ dnorm(par1, par2)
      })
    }
    suppressMessages(
      expect_error(rxode2(fn30))
    )

    fn31 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ pois()
      })
    }
    expect_error(rxode2(fn31))

    fn32 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ add(par1) + pois(par2)
      })
    }

    suppressMessages(
      expect_error(expect_message(rxode2(fn32)))
    )

    fn33 <- function() {
      ini({
        tKA <- 1
        tCL <- 1
        eta.CL ~ 0.1
        eta.KA ~ 0.1
        eta.V ~ 0.1
        par1 <- 1
        par2 <- 1
      })
      model({
        KA <- tKA + eta.KA
        CL <- tCL + eta.CL
        V <- tV + eta.V
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        linCmt() ~ prop(par1) + add(par2) + pois(par3)
      })
    }

    suppressMessages(
      expect_error(expect_message(rxode2(fn33)))
    )

    fn1 <- function() {
      ini({
        KA <- c(0, 1)
        CL <- c(0, 0.5)
      })
      model({
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        v1 <- 1
        linCmt() ~ pois()
      })
    }

    expect_error(rxode2(fn1))

    fn34 <- function() {
      ini({
        KA <- c(0, 1)
        CL <- c(0, 0.5)
        par1 <- 1
        par2 <- 2
      })
      model({
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        v1 <- 1
        linCmt() ~ add(par1) + prop(par2)
      })
    }

    expect_error(rxode2(fn34), NA)

    fn35 <- function() {
      ini({
        KA <- c(0, 1)
        CL <- c(0, 0.5)
        par1 <- 1
        par2 <- 2
      })
      model({
        KA <- KA + eta.KA
        CL <- CL + eta.CL
        v1 <- 1
        linCmt() ~ prop(par1) + add(par2)
      })
    }

    expect_error(rxode2(fn35), NA)
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("mu2 referencing", {

      f <- function() {
        ini({
          tka <- 0.45644793862596
          label("Log Ka")
          tcl <- 1.0158342238421
          label("Log Cl")
          tv <- 3.45168722356501
          label("Log V")
          add.sd <- c(0, 0.693865857699365)
          wt_cl <- fix(0.75)
          eta.ka ~ 0.413242125585966
          eta.cl ~ 0.0700506402691836
          eta.v ~ 0.0190586604340742
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl + wt_cl * log(WT/70.5))
          v <- exp(tv + eta.v)
          linCmt() ~ add(add.sd)
        })
      }

      f <- f()

      model.BASE <- function() {
        ini({
          tka <- log(1) # Log Ka
          tcl <- log(2) # Log Cl
          tv <- log(20)    # Log V
          eta.ka ~ 0.5
          eta.cl ~ 0.1
          eta.v ~ 0.1
          add.sd <- 0.5
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          linCmt() ~ add(add.sd)
        })
      }


      m2 <- model.BASE |>
        model(cl <- exp(tcl + eta.cl + wt_cl*log(WT/70.5))) |>
        ini(wt_cl <- fix(0.75))

      expect_equal(f$mu2RefCovariateReplaceDataFrame,
                   m2$mu2RefCovariateReplaceDataFrame)

    })
  }
})

rxTest({
  test_that("fixef()", {
    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd) + dnorm()
      })
    }

    expect_equal(nlme::fixef(one.cmt),
                 c(tka = 0.45, tcl = 0.993251773010283, tv = 3.45, add.sd = 0.7))

    f <- one.cmt()

    expect_equal(nlme::fixef(f),
                 c(tka = 0.45, tcl = 0.993251773010283, tv = 3.45, add.sd = 0.7))
  })

  test_that("coef()", {
    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd) + dnorm()
      })
    }

    f <- one.cmt()

    .theta <- c(tka = 0.45, tcl = 0.993251773010283, tv = 3.45, add.sd = 0.7)

    # default and explicit fixed levels return theta
    expect_equal(coef(f), .theta)
    expect_equal(coef(f, level = "theta"), .theta)
    expect_equal(coef(f, level = "fixed"), .theta)
    # coef() also works on the model function directly
    expect_equal(coef(one.cmt), .theta)

    # random-effect level returns the omega matrix
    expect_equal(coef(f, level = "omega"), f$omega)
    expect_equal(coef(f, level = "random"), f$omega)

    # all/both returns a list of fixed and random effects
    .all <- coef(f, level = "all")
    expect_equal(.all$theta, .theta)
    expect_equal(.all$omega, f$omega)
    expect_equal(coef(f, level = "both"), .all)

    # invalid level is an error
    expect_error(coef(f, level = "matt"))

    # models without random effects return NULL for omega
    no.re <- function() {
      ini({
        tka <- 0.5
      })
      model({
        ka <- tka
        d/dt(depot) <- -ka * depot
      })
    }
    expect_null(coef(no.re(), level = "omega"))
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("single theta", {
      f <- function() {
        ini({
          tka <- 0.45
          label("Ka")
          tcl <- 1
          label("Cl")
          tv <- 3.45
          label("V")
          add.sd <- c(0, 0.7)
          twt <- 0
          eta.ka ~ 0.6
          eta.v ~ 0.1
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + wt * twt)
          v <- exp(tv + eta.v)
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl/v * center
          cp = center/v
          cp ~ add(add.sd)
        })
      }

      f1 <- f()

      expect_equal(f1$singleTheta, c("tka", "tcl", "tv"))

      f <- function() {
        ini({
          tka <- 0.45
          label("Ka")
          tcl <- 1
          label("Cl")
          tv <- 3.45
          label("V")
          add.sd <- c(0, 0.7)
          twt <- 0
          eta.ka ~ 0.6
          eta.v ~ 0.1
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + wt * twt/70)
          v <- exp(tv + eta.v)
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl/v * center
          cp = center/v
          cp ~ add(add.sd)
        })
      }

      f1 <- f()

      expect_equal(f1$singleTheta, c("tka", "tcl", "tv"))

    })
  }
})

rxTest({
  test_that("model-defined IOV simulates with occasion data; RxODE#982", {
    two.compartment.tacro.test.Polimorf <- function() {
      ini({
        tka <- log(2.49)
        tcl <- log(19.84)
        tvc <- log(330.4)
        tvp <- log(118.2)
        tvq <- log(35.44)
        eta.ka ~ 0.82
        eta.cl ~ 0.358
        eta.vc ~ 1.186
        eta.q ~ 0.58
        iov.cl ~ 0.25^2 | occ
        prop.sd <- 0.204
      })
      model({
        cl <- exp(tcl + eta.cl + iov.cl)
        ka <- exp(tka + eta.ka)
        vc <- exp(tvc + eta.vc)
        vp <- exp(tvp)
        q <- exp(tvq + eta.q)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / vc * center
        d / dt(periph) <- q / vc * depot - q / vp * periph
        cp <- center / vc
        cp ~ prop(prop.sd)
      })
    }

    ev <- et(amt = 1000, ii = 24, addl = 6) |>
      et(seq(0, 24, by = 0.5)) |>
      et(seq(0, 24, by = 0.5) + 24 * 3) |>
      et(seq(0, 24, by = 0.5) + 24 * 6) |>
      et(id = 1:8)

    ev <- as.data.frame(ev)
    ev$occ <- 1
    ev$occ[ev$time > 25] <- 2
    ev$occ[ev$time > 24 * 5 + 1] <- 3

    rxWithSeed(982, {
      sol <- suppressWarnings(rxSolve(two.compartment.tacro.test.Polimorf, ev))

      expect_s3_class(sol, "rxSolve")
      expect_true(any(names(sol) == "iov.cl"))
      expect_true(any(names(sol) == "sim"))
      expect_equal(as.character(sort(unique(sol$occ))), c("1", "2", "3"))
      expect_true(all(is.finite(sol$iov.cl)))
      expect_true(any(!is.na(sol$sim)))

    })
  })
})

rxTest({
  test_that("comments are parsed correctly", {
    cmt <- c("function() {", "      ini({", "        ## You may label each parameter with a comment",
             "        tka <- 0.45 # Log Ka", "        tcl <- log(c(0, 2.7, 100)) # Log Cl",
             "        ## This works with interactive models", "        ## You may also label the preceding line with label(\"label text\")",
             "        tv <- 3.45; label(\"log V\")", "        ## the label(\"Label name\") works with all models",
             "        eta.ka ~ 0.6", "        eta.cl ~ 0.3", "        eta.v ~ 0.1",
             "        add.sd <- 0.7", "      })", "      model({", "        ka <- exp(tka + eta.ka)",
             "        cl <- exp(tcl + eta.cl)", "        v <- exp(tv + eta.v)",
             "        linCmt() ~ add(add.sd)", "      })", "    }")

    eq <- c("function () ", "{", "    ini({", "        tka <- 0.45", "        label(\"Log Ka\")",
            "        tcl <- log(c(0, 2.7, 100))", "        label(\"Log Cl\")",
            "        tv <- 3.45", "        label(\"log V\")", "        eta.ka ~ 0.6",
            "        eta.cl ~ 0.3", "        eta.v ~ 0.1", "        add.sd <- 0.7",
            "    })", "    model({", "        ka <- exp(tka + eta.ka)", "        cl <- exp(tcl + eta.cl)",
            "        v <- exp(tv + eta.v)", "        linCmt() ~ add(add.sd)",
            "    })", "}")

    suppressMessages(
      expect_equal(.rxReplaceCommentWithLabel(cmt), eq)
    )

    # Leave comment labels in here as they are required for equality testing below
    one.cmt <- function() {
      ini({
        ## You may label each parameter with a comment
        tka <- 0.45 # Log Ka
        tcl <- log(c(0, 2.7, 100)) # Log Cl
        ## This works with interactive models
        ## You may also label the preceding line with label("label text")
        tv <- 3.45; label("log V")
        ## the label("Label name") works with all models
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    suppressMessages(
      str <- .rxFunction2string(one.cmt)
    )
    if (!is.null(attr(one.cmt, "srcref"))) {
      expect_equal(str, eq)
      attr(one.cmt, "srcref") <- NULL
    }

    # Leave comment labels in here as they are required for equality testing below
    one.cmt <- function() {
      ini({
        ## You may label each parameter with a comment
        tka <- 0.45 # Log Ka
        tcl <- log(c(0, 2.7, 100)) # Log Cl
        ## This works with interactive models
        ## You may also label the preceding line with label("label text")
        tv <- 3.45; label("log V")
        ## the label("Label name") works with all models
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd) | tmp
      })
    }

    suppressMessages(
      mkstr <- .rxFunction2string(one.cmt)
    )
    expect_equal(mkstr,
                 c("function () ", "{", "    ini({", "        tka <- 0.45", "        label(\"Log Ka\")",
                   "        tcl <- log(c(0, 2.7, 100))", "        label(\"Log Cl\")",
                   "        tv <- 3.45", "        label(\"log V\")", "        eta.ka ~ 0.6",
                   "        eta.cl ~ 0.3", "        eta.v ~ 0.1", "        add.sd <- 0.7",
                   "    })", "    model({", "        ka <- exp(tka + eta.ka)", "        cl <- exp(tcl + eta.cl)",
                   "        v <- exp(tv + eta.v)", "        linCmt() ~ add(add.sd) | tmp",
                   "    })", "}"))

  })

  test_that("meta information parsing", {

    one.cmt <- function() {
      meta1 <- "meta"
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      meta2 <- "meta2"
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    tmp1 <- one.cmt()

    expect_true(inherits(as.function(tmp1), "function"))

    expect_equal(tmp1$meta$meta1, "meta")
    expect_equal(tmp1$meta$meta2, "meta2")

    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
        lambda <- c(-2, 1, 2)
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd) + boxCox(lambda) | tmp
      })
    }

    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        lambda <- 3 + v
        linCmt() ~ add(add.sd) + boxCox(lambda) | tmp
      })
    }

    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ lnorm(add.sd) | tmp
      })
    }

    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
        bLambda <- c(0, 3)
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ lnorm(add.sd) | tmp
        tmp2 ~ dpois(bLambda)
      })
    }

    cov <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        tvp <- 3.45
        cl.wt <- 0.1
        v.wt <- 0.1
        cl.sex <- 0.1
        v.sex <- 0.1
        cl.age <- 0.1
        v.age <- 0.1
        vp.wt <- 1
        vp.sex <- 1
        vp.age <- 1
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
        v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
        vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl/v * center
        cp = center/v
        cp ~ add(add.sd)
      })
    }

    pk.turnover.emax <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        ##
        eta.ktr ~ 1
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        prop.err <- 0.1
        pkadd.err <- 0.1
        ##
        temax <- logit(0.8)
        #temax <- 7.5
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        ##
        eta.emax ~ .5
        eta.ec50  ~ .5
        eta.kout ~ .5
        eta.e0 ~ .5
        ##
        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr + eta.ktr)
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        ##
        #poplogit = log(temax/(1-temax))
        emax=expit(temax+eta.emax)
        #logit=temax+eta.emax
        ec50 =  exp(tec50 + eta.ec50)
        kout = exp(tkout + eta.kout)
        e0 = exp(te0 + eta.e0)
        ##
        DCP = center/v
        PD=1-emax*DCP/(ec50+DCP)
        ##
        effect(0) = e0
        kin = e0*kout
        ##
        d/dt(depot) = -ktr * depot
        d/dt(gut) =  ktr * depot -ka * gut
        d/dt(center) =  ka * gut - cl / v * center
        d/dt(effect) = kin*PD -kout*effect
        ##
        cp = center / v
        cp ~ prop(prop.err) + add(pkadd.err)
        effect ~ add(pdadd.err)
      })
    }

    turnover.emax.noeta <- function() {
      ini({
        tktr <- log(1)
        tka <- log(1)
        tcl <- log(0.1)
        tv <- log(10)
        ##
        prop.err <- 0.1
        pkadd.err <- 0.1
        ##
        temax <- logit(0.8)
        #temax <- 7.5
        tec50 <- log(0.5)
        tkout <- log(0.05)
        te0 <- log(100)
        ##
        pdadd.err <- 10
      })
      model({
        ktr <- exp(tktr)
        ka <- exp(tka)
        cl <- exp(tcl)
        v <- exp(tv)
        ##
        #poplogit = log(temax/(1-temax))
        emax=expit(temax)
        ec50 =  exp(tec50)
        kout = exp(tkout)
        e0 = exp(te0)
        ##
        DCP = center/v
        PD=1-emax*DCP/(ec50+DCP)
        ##
        effect(0) = e0
        kin = e0*kout
        ##
        d/dt(depot) = -ktr * depot
        d/dt(gut) =  ktr * depot -ka * gut
        d/dt(center) =  ka * gut - cl / v * center
        d/dt(effect) = kin*PD -kout*effect
        ##
        cp = center / v
        cp ~ prop(prop.err) + add(pkadd.err)
        effect ~ add(pdadd.err)
      })
    }

    f <- function() {
      ini({
        lKA <- log(0.294)
        CL <- 18.6
        V2 <- 40.2
        Q <- 10.5
        V3 <- 297
        Kin <- 1
        Kout <- 1
        EC50 <- 200
        eta.ka ~ 0.12
        prop.sd ~ 0.2
      })
      model({
        KA <- exp(lKA + eta.ka)
        C2 <- centr/V2
        C3 <- peri/V3
        d/dt(depot) <- -KA*depot
        d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
        d/dt(peri) <- Q*C2 - Q*C3
        d/dt(eff) <- Kin - Kout*(1-C2/(EC50+C2))*eff
        eff(0) <- 1
        C2 ~ prop(prop.sd)
      })
    }

    expect_error(f(), "prop.sd")

    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka + eta.cl ~ c(0.6,
                            0.001, 0.3)
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd) | tmp
        vv ~ add(add.sd)
      })
    }

    expect_error(one.cmt())
  })


  test_that("model only", {

    one.cmt <- function() {
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        add.sd <- 4 + 3
        linCmt() ~ add(add.sd)
      })
    }

    expect_error(one.cmt(), NA)


    one.cmt <- function() {
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        add.sd <- 4
        linCmt() ~ add(add.sd)
      })
    }

    expect_error(one.cmt(), NA)

  })

  test_that("linCmt ui normalization expands to ode systems", {

    pure.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(2.7)
        tv <- 3.45
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka)
        cl <- exp(tcl)
        v <- exp(tv)
        linCmt() ~ add(add.sd) | tmp
      })
    }

    mixed.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(2.7)
        tv <- 3.45
        kin <- 1
        kout <- 1
        ec50 <- 2
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka)
        cl <- exp(tcl)
        v <- exp(tv)
        cp <- linCmt()
        eff(0) <- 1
        d/dt(eff) <- kin - kout * (1 - cp/(ec50 + cp)) * eff
        cp ~ add(add.sd)
      })
    }

    pure.ui <- suppressMessages(pure.cmt())
    expect_true(any(pure.ui$predDf$linCmt))
    expect_true(grepl("linCmt\\s*\\(", paste(deparse(as.function(pure.ui)), collapse = "\n")))
    pure.ode <- suppressMessages(linToOde(pure.ui))
    pure.fun <- paste(deparse(as.function(pure.ode)), collapse = "\n")

    expect_false(grepl("linCmt\\s*\\(", pure.fun))
    expect_true(grepl("d/dt\\(depot\\)", pure.fun))
    expect_true(grepl("d/dt\\(central\\)", pure.fun))
    expect_true(grepl("rxLinCmt <- central/", pure.fun, fixed = TRUE))
    expect_true(grepl("rxLinCmt ~ add\\(add\\.sd\\) \\| tmp", pure.fun))

    pure.ui2 <- suppressMessages(as.function(pure.ode)())
    expect_false(any(pure.ui2$predDf$linCmt))

    mixed.ui <- suppressMessages(mixed.cmt())
    expect_true(any(mixed.ui$predDf$linCmt == FALSE))
    mixed.ode <- suppressMessages(linToOde(mixed.ui))
    mixed.fun <- paste(deparse(as.function(mixed.ode)), collapse = "\n")

    expect_false(grepl("linCmt\\s*\\(", mixed.fun))
    expect_true(grepl("d/dt\\(depot\\)", mixed.fun))
    expect_true(grepl("d/dt\\(central\\)", mixed.fun))
    expect_true(grepl("cp <- central/", mixed.fun, fixed = TRUE))
    expect_true(grepl("d/dt\\(eff\\) <- kin - kout \\* \\(1 - cp/\\(ec50 \\+ cp\\)\\) \\* eff", mixed.fun))

    mixed.ui2 <- suppressMessages(as.function(mixed.ode)())
    expect_false(any(mixed.ui2$predDf$linCmt))
  })

  test_that("linToOde covers all supported linCmt translations", {
    .makeLinToOdeUi <- function(params, withDepot=FALSE) {
      .params <- params
      if (withDepot) {
        .params <- c(.params, "ka")
      }
      .lines <- c(vapply(seq_along(.params), function(i) {
        sprintf("%s <- %s", .params[i], i)
      }, character(1), USE.NAMES = FALSE),
      "add.sd <- 0.7")
      .txt <- paste(c(
        "function() {",
        "  model({",
        paste0("    ", .lines),
        "    cp <- linCmt()",
        "    cp ~ add(add.sd)",
        "  })",
        "}"
      ), collapse = "\n")
      suppressMessages(eval(parse(text=.txt))())
    }

    .linMeta <- function(ui) {
      .expr <- as.list(str2lang(paste0("{", rxNorm(ui$mvL), "}")))[-1]
      .w <- which(vapply(.expr, function(x) {
        is.call(x) &&
          (identical(x[[1]], quote(`=`)) || identical(x[[1]], quote(`<-`))) &&
          is.call(x[[3]]) &&
          as.character(x[[3]][[1]]) %in% c("linCmtA", "linCmtB")
      }, logical(1), USE.NAMES = FALSE))
      expect_length(.w, 1)
      .rhs <- .expr[[.w]][[3]]
      list(
        ncmt = as.integer(eval(.rhs[[5]], envir=baseenv())),
        oral0 = as.integer(eval(.rhs[[6]], envir=baseenv())),
        trans = as.integer(eval(.rhs[[8]], envir=baseenv()))
      )
    }

    .cases <- list(
      list(name="1c trans1", ncmt=1L, trans=1L, params=c("cl", "v")),
      list(name="1c trans2", ncmt=1L, trans=2L, params=c("k", "v")),
      list(name="1c trans10", ncmt=1L, trans=10L, params=c("alpha", "a")),
      list(name="1c trans11", ncmt=1L, trans=11L, params=c("alpha", "v")),
      list(name="2c trans1", ncmt=2L, trans=1L, params=c("cl", "v", "q", "vp")),
      list(name="2c trans2", ncmt=2L, trans=2L, params=c("k", "v", "k12", "k21")),
      list(name="2c trans3", ncmt=2L, trans=3L, params=c("cl", "v", "q", "vss")),
      list(name="2c trans4", ncmt=2L, trans=4L, params=c("alpha", "v", "beta", "k21")),
      list(name="2c trans5", ncmt=2L, trans=5L, params=c("alpha", "v", "beta", "aob")),
      list(name="2c trans10", ncmt=2L, trans=10L, params=c("alpha", "a", "beta", "b")),
      list(name="2c trans11", ncmt=2L, trans=11L, params=c("alpha", "v", "beta", "b")),
      list(name="3c trans1", ncmt=3L, trans=1L, params=c("cl", "v", "q", "vp", "q2", "vp2")),
      list(name="3c trans2", ncmt=3L, trans=2L, params=c("k", "v", "k12", "k21", "k13", "k31")),
      list(name="3c trans10", ncmt=3L, trans=10L, params=c("alpha", "a", "beta", "b", "gamma", "c")),
      list(name="3c trans11", ncmt=3L, trans=11L, params=c("alpha", "v", "beta", "b", "gamma", "c"))
    )

    for (.case in .cases) {
      for (.depot in c(FALSE, TRUE)) {
        .lbl <- sprintf("%s %s", .case$name, ifelse(.depot, "with depot", "without depot"))
        .ui <- .makeLinToOdeUi(.case$params, withDepot=.depot)
        .meta <- .linMeta(.ui)
        expect_identical(.meta$ncmt, .case$ncmt, info=.lbl)
        expect_identical(.meta$trans, .case$trans, info=.lbl)
        expect_identical(.meta$oral0, as.integer(.depot), info=.lbl)

        .ode <- suppressMessages(linToOde(.ui))
        .fun <- paste(deparse(as.function(.ode)), collapse = "\n")
        .odeUi <- suppressMessages(as.function(.ode)())

        expect_false(grepl("linCmt\\s*\\(", .fun), info=.lbl)
        expect_true(grepl("cp <- central/", .fun, fixed = TRUE), info=.lbl)
        expect_true(grepl("d/dt\\(central\\)", .fun), info=.lbl)
        expect_identical(grepl("d/dt\\(depot\\)", .fun), .depot, info=.lbl)
        expect_identical(grepl("d/dt\\(peripheral1\\)", .fun), .case$ncmt >= 2L, info=.lbl)
        expect_identical(grepl("d/dt\\(peripheral2\\)", .fun), .case$ncmt >= 3L, info=.lbl)
        expect_false(any(.odeUi$predDf$linCmt), info=.lbl)
      }
    }
  })

  test_that("linToOde handles assigned oral linCmt ui forms", {
    .makeAssignedLinUi <- function(linCmtLine) {
      .txt <- paste(c(
        "function() {",
        "  ini({",
        "    tka <- 0.45",
        "    tcl <- log(2.7)",
        "    tv <- 3.45",
        "    add.sd <- 0.7",
        "  })",
        "  model({",
        "    ka <- exp(tka)",
        "    cl <- exp(tcl)",
        "    v <- exp(tv)",
        paste0("    ", linCmtLine),
        "    cp ~ add(add.sd)",
        "  })",
        "}"
      ), collapse = "\n")
      suppressMessages(eval(parse(text=.txt))())
    }

    .cases <- c("cp <- linCmt()", "cp <- linCmt(ka, cl, v)")

    for (case in .cases) {
      .ui <- .makeAssignedLinUi(case)
      .uiFun <- paste(deparse(as.function(.ui)), collapse = "\n")
      .ode <- suppressMessages(linToOde(.ui))
      .odeFun <- paste(deparse(as.function(.ode)), collapse = "\n")
      .odeUi <- suppressMessages(as.function(.ode)())

      expect_true(grepl(case, .uiFun, fixed = TRUE), info=case)
      expect_false(grepl("linCmt\\s*\\(", .odeFun), info=case)
      expect_true(grepl("d/dt\\(depot\\)", .odeFun), info=case)
      expect_true(grepl("d/dt\\(central\\)", .odeFun), info=case)
      expect_true(grepl("cp <- central/", .odeFun, fixed = TRUE), info=case)
      expect_false(any(.odeUi$predDf$linCmt), info=case)
    }
  })

  test_that("linToOde handles assigned oral linCmt ui with ini parameters", {
    .ui <- suppressMessages(eval(parse(text = paste(c(
      "function() {",
      "  ini({",
      "    ka <- 1",
      "    cl <- 2",
      "    v <- 3",
      "    add.sd <- 0.7",
      "  })",
      "  model({",
      "    cp <- linCmt(ka, cl, v)",
      "    cp ~ add(add.sd)",
      "  })",
      "}"
    ), collapse = "\n")))())

    .uiFun <- paste(deparse(as.function(.ui)), collapse = "\n")
    .ode <- suppressMessages(linToOde(.ui))
    .odeFun <- paste(deparse(as.function(.ode)), collapse = "\n")
    .odeUi <- suppressMessages(as.function(.ode)())

    expect_true(grepl("cp <- linCmt\\(ka, cl, v\\)", .uiFun), info="ini oral params")
    expect_false(grepl("linCmt\\s*\\(", .odeFun), info="ini oral params")
    expect_true(grepl("d/dt\\(depot\\)", .odeFun), info="ini oral params")
    expect_true(grepl("d/dt\\(central\\)", .odeFun), info="ini oral params")
    expect_true(grepl("cp <- central/", .odeFun, fixed = TRUE), info="ini oral params")
    expect_false(any(.odeUi$predDf$linCmt), info="ini oral params")
  })

  test_that("iov covariates handled correctly", {

    one.cmt <- function() {
      ini({
        ## You may label each parameter with a comment
        tka <- 0.45 # Log Ka
        tcl <- log(c(0, 2.7, 100)) # Log Cl
        ## This works with interactive models
        ## You may also label the preceding line with label("label text")
        tv <- 3.45; label("log V")
        ## the label("Label name") works with all models
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        iov.cl ~ 0.1 | occ
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl + iov.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    o <- suppressWarnings(one.cmt())

    expect_equal(o$covariates, character(0))

  })

  test_that("ui desc doesn't include compartments", {
    one.cmt <- function() {
      ini({
        ## You may label each parameter with a comment
        tka <- 0.45 # Log Ka
        tcl <- log(c(0, 2.7, 100)) # Log Cl
        ## This works with interactive models
        ## You may also label the preceding line with label("label text")
        tv <- 3.45; label("log V")
        ## the label("Label name") works with all models
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }
    m <- one.cmt()
    expect_equal(m$modelDesc, "rxode2-based solved PK 1-compartment model")
  })
})

rxTest({
  test_that("ini will strip covariances and renumber if needed", {
    one.cmt <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    suppressMessages(
      f <- rxode2(one.cmt) |>
        ini(eta.v + eta.ka + eta.cl ~ c(1,
                                        0.01, 1,
                                        -0.01, 0.01, 1))
    )
    expect_equal(dimnames(f$omega)[[1]],
                 c("eta.v", "eta.ka", "eta.cl"))


    suppressMessages(
      f <- rxode2(one.cmt) |>
        ini(eta.v + eta.cl ~ c(1,
                               0.01, 1))
    )
    expect_equal(dimnames(f$omega)[[1]],
                 c("eta.ka", "eta.v", "eta.cl"))
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    # Was nlmixr issue #161
    test_that("Log-scaled vs Back-transformed parameters", {

      run7.mod <- function() {
        ini({
          tcl <- log(0.008) ; label("typical value of clearance")
          tv <- log(0.6) ; label("typical value of volume")
          all.cl <- 1 ; label("allometric exponent on CL")
          eta.cl ~ 1 ; label("interindividual variability on clearance")
          add.err <- 0.1 ; label("residual variability")
        })
        model({
          cl <- exp(tcl + all.cl * log_allo_wt + eta.cl) # individual value of clearance
          v <- exp(tv) # individual value of volume
          ke <- cl / v # elimination rate constant
          d / dt(A1) <- -ke * A1 # model differential equation
          cp <- A1 / v # concentration in plasma
          cp ~ add(add.err) # define error model
        })
      }

      ui <- rxode2(run7.mod)

      expectTransform <- function(ui, name, transform) {
        curEval <- ui$muRefCurEval
        w <- which(curEval$parameter == name)
        if (length(w) != 1) return(expect_equal(transform, ""))
        expect_equal(transform, curEval$curEval[w])
      }

      # make sure that all.cl and eta.cl add.err are not log-transformed parameters
      expectTransform(ui, "tcl", "exp")
      expectTransform(ui, "tv", "exp")
      expectTransform(ui, "all.cl", "")
      expectTransform(ui, "add.err", "")

    })
  }
})

rxTest({

  test_that("binding together two models without inis", {

    ocmt <- function() {
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
      })
    }

    idr <- function() {
      model({
        kin <- exp(tkin)
        kout <- exp(tkout)
        ic50 <- exp(tic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_output(
      expect_error(print(m1), NA),
      "Normalized Syntax"
    )

    expect_equal(m1$theta, setNames(numeric(0), character(0)))

  })

  test_that("binding together with first model missing an ini", {

    ocmt <- function() {
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
      })
      model({
        kin <- exp(tkin)
        kout <- exp(tkout)
        ic50 <- exp(tic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
        eff ~ add(idr.sd)
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_output(
      expect_error(print(m1), NA),
      "Normalized Syntax"
    )
    expect_true("idr.sd" %in% m1$iniDf$name)

  })
  test_that("binding together second model missing ini", {

    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        tv <- exp(3.45)
        add.sd <- 0.7
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      model({
        kin <- exp(tkin)
        kout <- exp(tkout)
        ic50 <- exp(tic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_output(
      expect_error(print(m1), NA),
      "Normalized Syntax"
    )
    expect_true("tv" %in% m1$iniDf$name)

  })

  test_that("binding together", {

    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        tv <- exp(3.45)
        add.sd <- 0.7
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
      })
      model({
        kin <- exp(tkin)
        kout <- exp(tkout)
        ic50 <- exp(tic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
        eff ~ add(idr.sd)
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_output(
      expect_error(print(m1), NA),
      "Normalized Syntax"
    )
    expect_true("idr.sd" %in% m1$iniDf$name)
    expect_true("tv" %in% m1$iniDf$name)

    expect_warning(rxAppendModel(ocmt, idr), "no variables in common")

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
      })
      model({
        kin <- exp(tkin)
        kout <- exp(tkout)
        ic50 <- exp(tic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
        eff ~ add(idr.sd)
      })
    }

    suppressMessages(
      expect_error(
        idr |> model({
          eff2 <- eff + 3
          eff2 ~ add(idr.sd2)
        }, append=TRUE),
        NA
      )
    )

    suppressMessages(
      addModelLine <-
        idr |> model({
          eff2 <- eff + 3
          eff2 ~ add(idr.sd2)
        },
        append=TRUE
        )
    )
    expect_true(any(addModelLine$iniDf$name == "idr.sd2"))
    expect_false(any(addModelLine$iniDf$name == "eff"))
    expect_false(any(addModelLine$iniDf$name == "eff2"))

    suppressMessages(
      expect_error(
        idr |> model({
          eff2 <- eff + 3
          eff2 ~ add(idr.sd2) | matt
        },
        append=TRUE),
        NA
      )
    )

    suppressMessages(
      addModelLine <-
        idr |> model({
          eff2 <- eff + 3
          eff2 ~ add(idr.sd2) | matt
        },
        append=TRUE
        )
    )

    expect_true(any(addModelLine$iniDf$name == "idr.sd2"))
    expect_false(any(addModelLine$iniDf$name == "eff"))
    expect_false(any(addModelLine$iniDf$name == "eff2"))
    expect_false(any(addModelLine$iniDf$name == "matt"))
  })

  test_that("bind together functions where population parameters overlap", {

    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        tv <- exp(3.45)
        add.sd <- 0.7
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
        tv <- 3
      })
      model({
        kin <- exp(tkin)
        kout <- exp(tkout)
        ic50 <- exp(tic50)
        v <- exp(tv)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma) * v)
        eff ~ add(idr.sd)
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_equal(m1$theta,
                 c(tka = 1.56831218549017, tcl = 2.71828182845905, tv = 31.5003923087479, add.sd = 0.7, tkin = 0, tkout = 0, tic50 = 2.30258509299405, gamma = 1, idr.sd = 1))

  })


  test_that("bind together functions where population parameters where all parameters overlap", {

    ocmt <- function() {
      ini({
        tv <- exp(3.45)
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
      })
    }

    idr <- function() {
      ini({
        tv <- 3
      })
      model({
        kin <- exp(tkin)
        kout <- exp(tkout)
        ic50 <- exp(tic50)
        v <- exp(tv)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma) * v)
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_equal(m1$theta,
                 c(tv = 31.5003923087479))

  })

  test_that("etas in first model but not in second", {

    ocmt <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.1
        eta.cl ~ 0.1
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
      })
      model({
        kin <- exp(tkin)
        kout <- exp(tkout)
        ic50 <- exp(tic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
        eff ~ add(idr.sd)
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_equal(m1$omega,
                 lotri({
                   eta.ka ~ 0.1
                   eta.cl ~ 0.1
                   eta.v ~ 0.1
                 }))

    expect_equal(m1$theta,
                 c(tka = 0.45, tcl = 1, tv = 3.45, add.sd = 0.7, tkin = 0, tkout = 0, tic50 = 2.30258509299405, gamma = 1, idr.sd = 1))

  })

  test_that("etas in second model but not first", {

    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        tv <- exp(3.45)
        add.sd <- 0.7
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
        eta.kin ~ 0.1
        eta.kout ~ 0.1
        eta.ic50 ~ 0.1
      })
      model({
        kin <- exp(tkin + eta.kin)
        kout <- exp(tkout + eta.kout)
        ic50 <- exp(tic50 + eta.ic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
        eff ~ add(idr.sd)
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_equal(m1$omega,
                 lotri({
                   eta.kin ~ 0.1
                   eta.kout ~ 0.1
                   eta.ic50 ~ 0.1
                 }))

    expect_equal(m1$theta,
                 c(tka = 1.56831218549017, tcl = 2.71828182845905, tv = 31.5003923087479, add.sd = 0.7, tkin = 0, tkout = 0, tic50 = 2.30258509299405, gamma = 1, idr.sd = 1))

  })

  test_that("bind together 2 models with etas", {

    ocmt <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.1
        eta.cl ~ 0.1
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
        eta.kin ~ 0.1
        eta.kout ~ 0.1
        eta.ic50 ~ 0.1
      })
      model({
        kin <- exp(tkin + eta.kin)
        kout <- exp(tkout + eta.kout)
        ic50 <- exp(tic50 + eta.ic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
        eff ~ add(idr.sd)
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_equal(m1$omega,
                 lotri({
                   eta.ka ~ 0.1
                   eta.cl ~ 0.1
                   eta.v ~ 0.1
                   eta.kin ~ 0.1
                   eta.kout ~ 0.1
                   eta.ic50 ~ 0.1
                 }))

    expect_equal(m1$theta,
                 c(tka = 0.45, tcl = 1, tv = 3.45, add.sd = 0.7, tkin = 0, tkout = 0, tic50 = 2.30258509299405, gamma = 1, idr.sd = 1))


  })

  test_that("bind together 2 models with etas with overlapping etas", {

    ocmt <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.1
        eta.cl ~ 0.1
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
        eta.kin ~ 0.1
        eta.kout ~ 0.1
        eta.ic50 ~ 0.1
        eta.v ~ 1
      })
      model({
        kin <- exp(tkin + eta.kin)
        kout <- exp(tkout + eta.kout)
        ic50 <- exp(tic50 + eta.ic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma) + eta.v)
        eff ~ add(idr.sd)
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_equal(m1$omega,
                 lotri({
                   eta.ka ~ 0.1
                   eta.cl ~ 0.1
                   eta.v ~ 0.1
                   eta.kin ~ 0.1
                   eta.kout ~ 0.1
                   eta.ic50 ~ 0.1
                 }))

    expect_equal(m1$theta,
                 c(tka = 0.45, tcl = 1, tv = 3.45, add.sd = 0.7, tkin = 0, tkout = 0, tic50 = 2.30258509299405, gamma = 1, idr.sd = 1))

  })

  test_that("bind together 2 models with etas with overlapping etas w/cov in 1", {

    ocmt <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka + eta.cl ~ c(0.1,
                            0.001, 0.1)
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
        eta.kin ~ 0.1
        eta.kout ~ 0.1
        eta.ic50 ~ 0.1
        eta.v ~ 1
      })
      model({
        kin <- exp(tkin + eta.kin)
        kout <- exp(tkout + eta.kout)
        ic50 <- exp(tic50 + eta.ic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma) + eta.v)
        eff ~ add(idr.sd)
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_equal(m1$omega,
                 lotri({
                   eta.ka + eta.cl ~ c(0.1, 0.001, 0.1)
                   eta.v ~ 0.1
                   eta.kin ~ 0.1
                   eta.kout ~ 0.1
                   eta.ic50 ~ 0.1
                 }))

    expect_equal(m1$theta,
                 c(tka = 0.45, tcl = 1, tv = 3.45, add.sd = 0.7, tkin = 0, tkout = 0, tic50 = 2.30258509299405, gamma = 1, idr.sd = 1))

  })


  test_that("bind together 2 models with etas with overlapping etas w/cov in 2", {

    ocmt <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.1
        eta.cl ~ 0.1
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
        eta.kin + eta.kout~ c(0.1,
                              0.01, 0.1)
        eta.ic50 ~ 0.1
        eta.v ~ 1
      })
      model({
        kin <- exp(tkin + eta.kin)
        kout <- exp(tkout + eta.kout)
        ic50 <- exp(tic50 + eta.ic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma) + eta.v)
        eff ~ add(idr.sd)
      })
    }

    m1 <- rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr)

    expect_equal(m1$omega,
                 lotri({
                   eta.ka ~ 0.1
                   eta.cl ~ 0.1
                   eta.v ~ 0.1
                   eta.kin + eta.kout ~ c(0.1, 0.01, 0.1)
                   eta.ic50 ~ 0.1
                 }))

    expect_equal(m1$theta,
                 c(tka = 0.45, tcl = 1, tv = 3.45, add.sd = 0.7, tkin = 0, tkout = 0, tic50 = 2.30258509299405, gamma = 1, idr.sd = 1))

  })

  test_that("bind together 2 models with etas with overlapping etas w/cov in 1", {

    ocmt <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.1
        eta.v + eta.cl~ c(0.1,
                          0.01, 0.1)
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
        eta.kin ~ 0.1
        eta.kout ~ 0.1
        eta.ic50 ~ 0.1
        eta.v ~ 1
      })
      model({
        kin <- exp(tkin + eta.kin)
        kout <- exp(tkout + eta.kout)
        ic50 <- exp(tic50 + eta.ic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma) + eta.v)
        eff ~ add(idr.sd)
      })
    }

    expect_error(rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr))
  })

  test_that("bind together 2 models with etas with overlapping etas w/cov in 2", {

    ocmt <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.1
        eta.v  ~ 0.1
        eta.cl ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
        eta.kin ~ 0.1
        eta.kout ~ 0.1
        eta.ic50 + eta.v ~ c(0.1,
                             0.001, 1)
      })
      model({
        kin <- exp(tkin + eta.kin)
        kout <- exp(tkout + eta.kout)
        ic50 <- exp(tic50 + eta.ic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma) + eta.v)
        eff ~ add(idr.sd)
      })
    }

    expect_error(rxAppendModel(ocmt |> model(ceff=cp,append=TRUE), idr))

  })

  test_that("combine models without anything in common", {

    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        tv <- exp(3.45)
        add.sd <- 0.7
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
      })
      model({
        kin <- exp(tkin)
        kout <- exp(tkout)
        ic50 <- exp(tic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
        eff ~ add(idr.sd)
      })
    }

    m1 <- rxAppendModel(ocmt, idr, common=FALSE)

    expect_true("idr.sd" %in% m1$iniDf$name)
    expect_true("tv" %in% m1$iniDf$name)

  })

  test_that("combine 3 models", {

    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        tv <- exp(3.45)
        add.sd <- 0.7
      })
      model({
        ka <- tka
        cl <- tcl
        v <- tv
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
    }

    d1 <- ocmt |> rxRename(tkaD1=tka,
                           tclD1=tcl,
                           tvD1=tv,
                           add.sd.d1=add.sd,
                           kaD1=ka,
                           clD1=cl,
                           vD1=v,
                           depotD1=depot,
                           centerD1=center,
                           cpD1=cp)

    d2 <- ocmt |> rxRename(tkad2=tka,
                           tcld2=tcl,
                           tvd2=tv,
                           add.sd.d2=add.sd,
                           kad2=ka,
                           cld2=cl,
                           vd2=v,
                           depotd2=depot,
                           centerd2=center,
                           cpd2=cp)

    idr <- function() {
      ini({
        tkin <- log(1)
        tkout <- log(1)
        tic50 <- log(10)
        gamma <- fix(1)
        idr.sd <- 1
      })
      model({
        kin <- exp(tkin)
        kout <- exp(tkout)
        ic50 <- exp(tic50)
        d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
        eff ~ add(idr.sd)
      })
    }

    idr <- idr |> model(ceff=cpD1 + cpd2, append=NA)

    full <- rxAppendModel(d1, d2, idr, common=FALSE)

    expect_equal(full$theta,
                 c(tkaD1 = 1.56831218549017, tclD1 = 2.71828182845905, tvD1 = 31.5003923087479, add.sd.d1 = 0.7, tkad2 = 1.56831218549017, tcld2 = 2.71828182845905, tvd2 = 31.5003923087479, add.sd.d2 = 0.7, tkin = 0, tkout = 0, tic50 = 2.30258509299405, gamma = 1, idr.sd = 1))

  })
})

rxTest({
  test_that("model properties after are parsed OK", {
    one.compartment <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        cp ~ add(add.sd)
      })
      keep = "WT"
      drop = "depot"
    }

    expect_warning(rxode2(one.compartment))

    expect_s3_class(suppressWarnings(rxode2(one.compartment)), "rxUi")

    expect_true(inherits(one.compartment(), "character"))
  })

  test_that("This model parses k0 as a covariate", {
    one.compartment.saem <- function() {
      ini({
        tka <- .5 ; label("Log Ka")
        tcl <- -3.2 ; label("Log Cl")
        tv <- -1 ; label("Log V")
        eta.ka ~ 1
        eta.cl ~ 2
        eta.v ~ 1
        add.err <- 0.1
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot + exp(-k0 * t)
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    mod <- rxode2(one.compartment.saem)

    expect_equal(mod$covariates, "k0")
    expect_equal(mod$all.covs, "k0") # backward compatible
  })

  test_that("complex models that used to raise errors but should not", {
    two.cmt.pd <- function() {
      ini({
        tKa <- log(0.64)
        tCl <- log(5.22)
        tV2 <- log(41.3)
        tV3 <- log(115)
        tQ <- log(11.96)
        BWef <- log(1.87)
        tSlope <- log(10) ; label("add for PD estimation")
        tIntercept <- log(1) ; label("add for PD estimation")
        eta.Ka ~ 1.18
        eta.Cl ~ 0.09
        eta.V2 ~ 0.2
        eta.V3 ~ 0.12
        eta.Q ~ 0.12
        eta.Slope ~ 0.1 ; label("add for PD estimation")
        eta.Intercept ~ 0.1 ; label("add for PD estimation")

        prop.err1 <- 0.1 ; label("Cp")
        prop.err2 <- 0.3 ; label("Ef")
      })
      model({
        Ka <- exp(tKa + eta.Ka)
        Cl <- exp(tCl + BWef * log.BW.70 + eta.Cl)
        V2 <- exp(tV2 + eta.V2)
        V3 <- exp(tV3 + eta.V3)
        Q <- exp(tQ + eta.Q)
        Slope <- exp(tSlope + eta.Slope) ## add for PD estimation
        Intercept <- exp(tIntercept + eta.Intercept) ## add for PD estimation

        d / dt(depot) <- -Ka * depot
        d / dt(center) <- Ka * depot - Cl / V2 * center + Q / V3 * periph - Q / V2 * center
        d / dt(periph) <- Q / V2 * center - Q / V3 * periph

        Cp <- center / V2
        Ef <- Cp * Slope + Intercept ## add for PD estimation

        Cp ~ prop(prop.err1) | center
        Ef ~ prop(prop.err2) ## add for PD estimation
      })
    }

    expect_s3_class(rxode2(two.cmt.pd), "rxUi")

    one.compartment.IV.model <- function() {
      ini({ # Where initial conditions/variables are specified
        # '<-' or '=' defines population parameters
        # Simple numeric expressions are supported
        Cl <- 1.6 # Cl (L/hr)
        Vc <- 4.5 # V (L)
        # Bounds may be specified by c(lower, est, upper), like NONMEM:
        # Residuals errors are assumed to be population parameters
        prop.err <- c(0, 0.3, 1)
        # Between subject variability estimates are specified by '~'
        # Semicolons are optional
        # eta.Vc ~ 0.1   #IIV V
        # eta.Cl ~ 0.1   #IIV Cl
      })
      model({ # Where the model is specified
        # The model uses the ini-defined variable names
        # Vc <- exp(lVc + eta.Vc)
        # Cl <- exp(lCl + eta.Cl)
        # RxODE-style differential equations are supported
        d / dt(centr) <- -(Cl / Vc) * centr
        ## Concentration is calculated
        cp <- centr / Vc
        # And is assumed to follow proportional error estimated by prop.err
        cp ~ prop(prop.err)
      })
    }

    expect_s3_class(suppressMessages(rxode2(one.compartment.IV.model)), "rxUi")

    model1 <- function() {
      ini({
        CL <- 2.2
        V <- 65
        add.err <- 0.01
        prop.err <- 0.01
      })
      model({
        kel <- CL / V
        X(0) <- 0
        d / dt(X) <- -kel * X
        cp <- X / V
        cp ~ add(add.err) + prop(prop.err)
      })
    }

    f <- rxode2(model1)

    expect_s3_class(f, "rxUi")
  })

  test_that("modeled expressions don't have to be in the model if non-normal", {
    ocmt <- function() {
      ini({
        tka <- exp(0.45)
        tcl <- exp(1)
        eta.v ~ 0.01
        lower <- 0.1
        upper <- 0.9
        prop.eta ~ 0.01
      })
      model({
        ka <- tka
        cl <- tcl
        v <- eta.v
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl / v * center
        cp = center / v
        prop.sd <- exp(tprop + prop.eta)
        cp2 ~ dunif(lower, upper)
      })
    }

    expect_warning(
      expect_error(ocmt(), NA),
      regexp = "some etas defaulted to non-mu referenced"
    )
  })

  test_that("only specifying residual error", {
    one.cmt <- function() {
      ini({
        add.sd <- 4
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    expect_error(one.cmt(), NA)
  })

  test_that("one cmt noeta", {
    one.cmt.ll.noeta <- function() {
      ini({
        tka <- 0.45
        tcl <- log(c(0, 2.7, 100))
        tv <- 3.45
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka)
        cl <- exp(tcl)
        v <- exp(tv)
        cp <- linCmt()
        ll(err) ~ -log(add.sd) - 0.5*log(2*pi) - 0.5*((DV-cp)/add.sd)^2
      })
    }

    expect_error(one.cmt.ll.noeta(), NA)
  })

  test_that("no theta, but eta (rxode2#433)", {
    fun <- function() {
      ini({
        eta1 ~ 0.2
        eta2 ~ 0.2
        eta3 ~ 0.2
      })
      model({
        IETA1 <- 0 
        IETA2 <- 0
        IETA3 <- 0
        ETCL <- eta1 + IETA1
        ETVC <- eta2 + IETA2
        ETKA <- eta3 + IETA3
        TVCL <- 4.0
        TVVC <- 70.0
        TVKA <- 1.0
        CL <- TVCL * exp(ETCL)
        VC <- TVVC * exp(ETVC)
        KA <- TVKA * exp(ETKA)
        K20 <- CL / VC
        scale2 <- VC
        d/dt(rxddta1) <-  - KA * rxddta1
        d/dt(rxddta2) <- KA * rxddta1 - K20 * rxddta2
        DEL <- 0
        if (F == 0) DEL <- 1
        W <- F + DEL
        Y <- F + W * eps1
        IPRED <- F
        IRES <- DV - IPRED
        IWRES <- IRES / W
      })}
    expect_error(fun(), NA)
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("ini piping string", {
      one.compartment <- function() {
        ini({
          tka <- 0.45
          tcl <- 1
          tv <- 3.45
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.err <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          d / dt(depot) <- -ka * depot
          d / dt(center) <- ka * depot - cl / v * center
          F(depot) <- 3
          cp <- center / v
          cp ~ add(add.err)
        })
      }

      tmp <- rxode2(one.compartment)

      theta <- tmp$theta

      fixItems <- setNames(rep("fix",length(theta)), names(theta))
      suppressMessages(
        ui <- tmp |> ini(fixItems)
      )

      expect_true(all(ui$iniDf$fix[!is.na(ui$iniDf$ntheta)]))

      unfixItems <- setNames(rep("unfix",length(theta)), names(theta))

      suppressMessages(
        ui2 <- ui |> ini(unfixItems)
      )

      expect_true(all(!ui2$iniDf$fix[!is.na(ui2$iniDf$ntheta)]))
    })
  }
})

if (!.Call(`_rxode2_isIntel`)) {
  test_that("noop has identical mu-ref", {
    pheno <- function() {
      ini({
        tcl <- log(0.008)
        tv <-  log(0.6)
        eta.cl + eta.v ~ c(1,
                           0.01, 1)
        add.err <- 0.1
      })
      model({
        cl <- exp(tcl + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v)    # individual value of volume
        ke <- cl / v            # elimination rate constant
        d/dt(A1) = - ke * A1    # model differential equation
        cp = A1 / v             # concentration in plasma
        cp ~ add(add.err)  # define error model
      })
    }

    p1 <- rxode2(pheno)

    pheno2 <- function() {
      ini({
        tcl <- log(0.008)
        tv <-  log(0.6)
        eta.cl + eta.v ~ c(1,
                           0.01, 1)
        add.err <- 0.1
      })
      model({
        cl <- exp(+tcl + eta.cl) # individual value of clearance
        v <- exp(tv + eta.v)    # individual value of volume
        ke <- cl / v            # elimination rate constant
        d/dt(A1) = - ke * A1    # model differential equation
        cp = A1 / v             # concentration in plasma
        cp ~ add(add.err)  # define error model
      })
    }

    p2 <- rxode2(pheno2)

    expect_equal(p1$muRefCovariateDataFrame, p2$muRefCovariateDataFrame)
    expect_equal(p1$muRefCovariateEmpty, p2$muRefCovariateEmpty)
    expect_equal(p1$muRefCurEval, p2$muRefCurEval)
    expect_equal(p1$muRefDataFrame, p2$muRefDataFrame)
    expect_equal(p1$muRefDropParameters, p2$muRefDropParameters)
    expect_equal(p1$muRefExtra, p2$muRefExtra)
    expect_equal(p1$muRefExtraEmpty, p2$muRefExtraEmpty)
    expect_equal(p1$nonMuEtas, p2$nonMuEtas)
  })
}

rxTest({

  test_that("mix() in normal rxode2 model", {
    expect_error(rxModelVars("a = mix(a)"))
    expect_error(rxModelVars("a = mix(a, b)"))
    expect_error(rxModelVars("a = mix(a, b, c)"), NA)
    expect_error(rxModelVars("a = mix(a, b, c, d)"))
    expect_error(rxModelVars("a = mix(a, p1, c); c = mix(a, p1, c, p2, e)"))
    expect_error(rxModelVars("a = mix(a, p1, c); c = mix(a, p1, d)"), NA)
    expect_error(rxModelVars("mixest <- 1"))
    expect_error(rxModelVars("mixnum <- 1"))
    expect_error(rxModelVars("mixunif <- 1"))
  })

  test_that("good mix() in ui model", {
    f <- function() {
      ini({
        p1 <- 0.1
        p2 <- 0.2
        eta.a ~ 0.1
      })
      model({
        a <- mix(a1, p1, a2, p2, a3) + eta.a
        b <- mix(b1, b2, b3)
      })
    }

    expect_error(rxModelVars(f()), NA)

    f <- f()
    expect_equal(f$mixProbs, c("p1", "p2"))
  })

  test_that("mix() requires between subject variability", {
    f <- function() {
      ini({
        p1 <- 0.1
        p2 <- 0.2
      })
      model({
        a <- mix(a1, p1, a2, p2, a3)
        b <- mix(b1, b2, b3)
      })
    }
    expect_error(f())
  })

  test_that("mix() requires require probabilities in the ini block", {
    f <- function() {
      ini({
        a1 <- 0.1
        a2 <- 0.2
        eta.a1 ~ 0.1
      })
      model({
        a <- mix(a1, p1, a2, p2, a3) + eta.a1
        b <- mix(b1, b2, b3)
      })
    }

    expect_error(f())
  })

  test_that("mix() requires require probabilities to sum to a number between 0 and 1", {

    f <- function() {
      ini({
        p1 <- -10
        p2 <- 0.2
        eta.a1 ~ 0.1
      })
      model({
        a <- mix(a1, p1, a2, p2, a3) + eta.a1
        b <- mix(b1, b2, b3)
      })
    }

    expect_error(f())
  })

  test_that("mix() requires the same probabilities in each proportion", {

    f <- function() {
      ini({
        p1 <- 0.1
        p2 <- 0.2
        eta.a1 ~ 0.1
      })
      model({
        a <- mix(a1, p1, a2, p2, a3) + eta.a1
        b <- mix(b1, p2, b2, p1, b3)
      })
    }

    expect_error(f())

    f <- function() {
      ini({
        p1 <- 0.1
        p2 <- 0.2
        eta.a1 ~ 0.1
      })
      model({
        a <- mix(a1, p1, a2, p2, a3) + eta.a1
        b <- mix(b1, p1, b2, p2, b3)
      })
    }

    expect_error(f(), NA)

    f <- f()

    # Gets the probabilities
    expect_equal(f$mixProbs, c("p1", "p2"))

    # Throws error that this is a mixed model
    expect_error(assertRxUiNoMix(f))

    one.cmt <- function() {
      ini({
        ## You may label each parameter with a comment
        tka <- 0.45 # Log Ka
        tcl <- log(c(0, 2.7, 100)) # Log Cl
        ## This works with interactive models
        ## You may also label the preceding line with label("label text")
        tv <- 3.45; label("log V")
        ## the label("Label name") works with all models
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    expect_error(assertRxUiNoMix(one.cmt), NA)

  })

  test_that("test dsl to change mix()", {
    expect_equal(rxToSE("mix(a1, p1, b)"),
                 "(rxEq(mixest, 1)*(a1)+rxEq(mixest, 2)*(b))")

    expect_equal(rxToSE("mix(a1, p1, b, p2, c)"),
                 "(rxEq(mixest, 1)*(a1)+rxEq(mixest, 2)*(b)+rxEq(mixest, 3)*(c))")
  })

  test_that("mix() simulation", {
    rxWithSeed(42, {

      one.cmt <- function() {
        ini({
          ## You may label each parameter with a comment
          tka <- 0.45 # Log Ka
          tcl1 <- log(c(0, 2.7, 100)) # Log Cl
          tcl2 <- log(c(0, 0.1, 120)) # Log Cl
          ## This works with interactive models
          ## You may also label the preceding line with label("label text")
          tv <- 3.45; label("log V")
          p1 <- 0.3
          ## the label("Label name") works with all models
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.sd <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- mix(exp(tcl1 + eta.cl), p1, exp(tcl2 + eta.cl))
          v <- exp(tv + eta.v)
          me <- mixest
          mn <- mixnum
          mu <- mixunif
          linCmt() ~ add(add.sd)
        })
      }

      s <- rxSolve(one.cmt, et(amt=320, ii=12, addl=2, cmt=1) |>
                              et(seq(0, 72)) |>
                              et(id=1:20))

      expect_false(all(s$me == 1))
      expect_false(all(s$me == 2))

      expect_true(any(s$me == 1))
      expect_true(any(s$me == 2))
    })
  })

  test_that("mix() simulation, and the re-estimate with same mixest", {

    one.cmt <- function() {
      ini({
        ## You may label each parameter with a comment
        tka <- 0.45 # Log Ka
        tcl1 <- log(c(0, 2.7, 100)) # Log Cl
        tcl2 <- log(c(0, 0.1, 120)) # Log Cl
        ## This works with interactive models
        ## You may also label the preceding line with label("label text")
        tv <- 3.45; label("log V")
        p1 <- 0.3
        ## the label("Label name") works with all models
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- mix(exp(tcl1 + eta.cl), p1, exp(tcl2 + eta.cl))
        v <- exp(tv + eta.v)
        me <- mixest
        mu <- mixunif
        mn <- mixnum
        linCmt() ~ add(add.sd)
      })
    }

    s0 <- s <- rxSolve(one.cmt, et(amt=320, ii=12, addl=2, cmt=1) |>
                                  et(seq(0, 72)) |>
                                  et(id=1:20), addDosing=TRUE) |>
      dplyr::rename(mixest=me, dv=sim) |>
      dplyr::select(id, mixest, evid, cmt, amt, time, dv, mu)

    trn <- etTrans(s, one.cmt)
    lst <- attr(class(trn), ".rxode2.lst")
    class(lst) <- NULL
    expect_true(all(lst$mixUnif >= 1))
    expect_true(length(lst$mixUnif) == length(lst$idLvl))
    expect_length(lst$covParPos, 0)
    expect_length(lst$covParPos0, 0)

    s2 <- s
    s2$mixest <- NULL

    trn <- etTrans(s2, one.cmt)
    lst2 <- attr(class(trn), ".rxode2.lst")
    class(lst2) <- NULL

    for (n in names(lst)) {
      if (n %in% c("mixUnif", "lib_name")) next
      expect_equal(lst[[n]], lst2[[n]], info=n)
    }

    # Now error with mixtures above nmix in the model
    s$mixest[s$id == 1] <- 3
    expect_error(etTrans(s, one.cmt))

    ## mixest below 1
    s$mixest[s$id == 1] <- 0
    expect_error(etTrans(s, one.cmt))

    # non integer mixest
    s$mixest[s$id == 1] <- 0.5
    expect_error(etTrans(s, one.cmt))

    # Time varying mixest
    s$mixest <- seq_along(s$mixest) %% 2 + 1

    expect_error(etTrans(s, one.cmt))

    expect_error(rxSolve(one.cmt, s0, addDosing=TRUE, nStud=100))

    s2 <- rxSolve(one.cmt, s0, addDosing=TRUE) |>
      dplyr::rename(mixest=me, dv=sim) |>
      dplyr::select(id, mixest, evid, cmt, amt, time, dv, mu)

    # keeps all the mixest
    expect_true(all(s2$mixest == s0$mixest))

    # mixunif is not simulated, so the values are not the same
    expect_false(all(s2$mu == s0$mu))


    s0 <- rxSolve(one.cmt, et(amt=320, ii=12, addl=2, cmt=1) |>
                             et(seq(0, 72)) |>
                             et(id=1:20), addDosing=TRUE) |>
      dplyr::rename(mixunif=mu, dv=sim) |>
      dplyr::select(id, mixunif, evid, cmt, amt, time, dv, me)

    s2 <- rxSolve(one.cmt, s0, addDosing=TRUE) |>
      dplyr::rename(mixunif=mu, dv=sim) |>
      dplyr::select(id, mixunif, evid, cmt, amt, time, dv, me)

    # keeps all the mixest
    expect_true(all(s2$me == s0$me))

    # mixunif is not simulated, so the values are not the same
    expect_true(all(s2$mixunif == s0$mixunif))


    s0 <- rxSolve(one.cmt, et(amt=320, ii=12, addl=2, cmt=1) |>
                             et(seq(0, 72)) |>
                             et(id=1:20), addDosing=TRUE) |>
      dplyr::rename(mixunif=mu, mixest=me, dv=sim) |>
      dplyr::select(id, mixunif, evid, cmt, amt, time, dv, mixest)

    expect_error(rxSolve(one.cmt, s0, addDosing=TRUE))

  })

  test_that("test mixture models load with rxS()", {

 expect_error(rxS("tka=THETA[1];\ntcl1=THETA[2];\ntcl2=THETA[3];\ntv=THETA[4];\np1=THETA[5];\nadd.sd=THETA[6];\neta.ka=ETA[1];\neta.cl=ETA[2];\neta.v=ETA[3];\nka=exp(tka+eta.ka);\ncl=mix(exp(tcl1+eta.cl),p1,exp(tcl2+eta.cl));\nv=exp(tv+eta.v);\nme=mixest;\nmn=mixnum;\nmu=mixunif;\nrx_yj_~2;\nrx_lambda_~1;\nrx_low_~0;\nrx_hi_~1;\nrx_pred_f_~linCmtA(rx__PTR__,t,2,1,1,-1,1,cl,v,0.0,0.0,0.0,0.0,ka);\nrx_pred_~rx_pred_f_;\nrx_r_~(add.sd)^2;\n"),
                 NA)

    expect_error(
      rxS("tka=THETA[1];\ntcl1=THETA[2];\ntcl2=THETA[3];\ntv=THETA[4];\np1=THETA[5];\nadd.sd=THETA[6];\neta.cl=ETA[1];\nka=exp(tka);\ncl=mix(expit(tcl1+eta.cl,0.1,200),p1,expit(tcl2+eta.cl,0.1,200));\nv=exp(tv);\nrx_yj_~2;\nrx_lambda_~1;\nrx_low_~0;\nrx_hi_~1;\nrx_pred_f_~linCmtA(rx__PTR__,t,2,1,1,-1,1,cl,v,0.0,0.0,0.0,0.0,ka);\nrx_pred_~rx_pred_f_;\nrx_r_~(add.sd)^2;\n"),
      NA
    )
  })
})

rxTest({

  testPipeQuote <- function(..., envir=parent.frame(), iniDf = NULL) {
    rxUnloadAll()
    gc()
    .quoteCallInfoLines(match.call(expand.dots = TRUE)[-1], envir=envir, iniDf=iniDf)
  }


  rxTest({

    test_that("nse evaluation", {

      tmp <- "d/dt(depot)"
      expect_equal(testPipeQuote(tmp),
                   list(quote(d/dt(depot))))


      t <- c("-d/dt(peripheral1)", "-d/dt(peripheral2)")
      expect_equal(testPipeQuote(t),
                   list(quote(-d/dt(peripheral1)),
                        quote(-d/dt(peripheral2))))

      t <- c(a="x", b="y")

      expect_equal(testPipeQuote(t),
                   list(quote(a <- x), quote(b <- y)))

      tmp <- list(tmp="d/dt(depot)")

      expect_equal(testPipeQuote(tmp$tmp),
                   list(quote(d/dt(depot))))

      tmp <- list(tmp=list(tmp="d/dt(depot)"))

      expect_equal(testPipeQuote(tmp$tmp$tmp),
                   list(quote(d/dt(depot))))

    })

    test_that("equivalent drop statements", {

      expect_equal(.changeDropNullLine(quote(a <- NULL)),
                   quote(-a))

      expect_equal(.changeDropNullLine(quote(a ~ NULL)),
                   quote(-a))
      expect_equal(.changeDropNullLine(str2lang("a = NULL")),
                   quote(-a))

      expect_equal(.changeDropNullLine(quote(d/dt(a) <- NULL)),
                   quote(-d/dt(a)))
      expect_equal(.changeDropNullLine(quote(d/dt(a) ~ NULL)),
                   quote(-d/dt(a)))
      expect_equal(.changeDropNullLine(str2lang("d/dt(a) = NULL")),
                   quote(-d/dt(a)))

      expect_equal(.changeDropNullLine(quote(lag(a) <- NULL)),
                   quote(-lag(a)))
      expect_equal(.changeDropNullLine(quote(lag(a) ~ NULL)),
                   quote(-lag(a)))
      expect_equal(.changeDropNullLine(str2lang("lag(a) = NULL")),
                   quote(-lag(a)))

      expect_equal(.changeDropNullLine(quote(alag(a) <- NULL)),
                   quote(-alag(a)))
      expect_equal(.changeDropNullLine(quote(alag(a) ~ NULL)),
                   quote(-alag(a)))
      expect_equal(.changeDropNullLine(str2lang("alag(a) = NULL")),
                   quote(-alag(a)))

      expect_equal(.changeDropNullLine(quote(F(a) <- NULL)),
                   quote(-F(a)))
      expect_equal(.changeDropNullLine(quote(F(a) ~ NULL)),
                   quote(-F(a)))
      expect_equal(.changeDropNullLine(str2lang("F(a) = NULL")),
                   quote(-F(a)))

      expect_equal(.changeDropNullLine(quote(f(a) <- NULL)),
                   quote(-f(a)))
      expect_equal(.changeDropNullLine(quote(f(a) ~ NULL)),
                   quote(-f(a)))
      expect_equal(.changeDropNullLine(str2lang("f(a) = NULL")),
                   quote(-f(a)))

      expect_equal(.changeDropNullLine(quote(rate(a) <- NULL)),
                   quote(-rate(a)))
      expect_equal(.changeDropNullLine(quote(rate(a) ~ NULL)),
                   quote(-rate(a)))
      expect_equal(.changeDropNullLine(str2lang("rate(a) = NULL")),
                   quote(-rate(a)))

      expect_equal(.changeDropNullLine(quote(dur(a) <- NULL)),
                   quote(-dur(a)))
      expect_equal(.changeDropNullLine(quote(dur(a) ~ NULL)),
                   quote(-dur(a)))
      expect_equal(.changeDropNullLine(str2lang("dur(a) = NULL")),
                   quote(-dur(a)))

      expect_equal(.changeDropNullLine(quote(a(0) <- NULL)),
                   quote(-a(0)))
      expect_equal(.changeDropNullLine(quote(a(0) ~ NULL)),
                   quote(-a(0)))
      expect_equal(.changeDropNullLine(str2lang("a(0) = NULL")),
                   quote(-a(0)))

    })

    test_that("test fix/unfix for eta", {

      expect_equal(testPipeQuote(a~fix),
                   list(quote(a<-fix)))
      expect_equal(testPipeQuote(a~unfix),
                   list(quote(a<-unfix)))

    })

    test_that("test as formula", {

      expect_equal(testPipeQuote(as.formula(a~b)),
                   list(quote(a~b)))

    })

    test_that("test of standard quoting of piping arguments", {

      expect_equal(testPipeQuote(-ka, tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
        cl = exp(tcl + eta.cl)
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, "tv10=3"), list(quote(-ka),
                         quote(tka <- 0.5),
                         quote(tv <- 3),
                         quote(tcl <- 10),
                         quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
                         quote(cl <- exp(tcl + eta.cl)),
                         quote(eta.ka ~ 3),
                         quote(eta.ka ~ 3),
                         quote(tv <- 3),
                         quote(tcl <- 10),
                         quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
                         quote(tv10 <- 3)))

      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2)))

      expect_equal(testPipeQuote({
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2),
      list(quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2)))

      # Test c()
      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, c(tka=1, tv=3, tcl=4)),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(tka <- 1),
           quote(tv <- 3),
           quote(tcl <- 4))
      )

      # test list()
      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, list(tka=1, tv=3, tcl=4)),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(tka <- 1),
           quote(tv <- 3),
           quote(tcl <- 4))
      )

      .tmp <- list(tcl = 3, tv = 4)
      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, .tmp),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(tcl <- 3),
           quote(tv <- 4))
      )

      .tmp <- list(tcl = 3, tv = 4)
      .tmp2 <- c(fun1=3, fun2=4)
      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, .tmp, ~.tmp2),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(tcl <- 3),
           quote(tv <- 4),
           quote(fun1 ~ 3),
           quote(fun2 ~ 4))
      )

      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, ~.tmp, .tmp2),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(tcl ~ 3),
           quote(tv ~ 4),
           quote(fun1 <- 3),
           quote(fun2 <- 4))
      )


      .tmp <- list(tcl = 3, tv = 4)
      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, ~.tmp),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(tcl ~ 3),
           quote(tv ~ 4))
      )


      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, ~.tmp),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(tcl ~ 3),
           quote(tv ~ 4))
      )

      .tmp <- c(tcl = 3, tv = 4)

      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, .tmp),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(tcl <- 3),
           quote(tv <- 4))
      )

      .tmp <- quote({
        ka = exp(tka)
      })

      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, .tmp),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(ka <- exp(tka)))
      )

      .tmp <- quote(ka <- 8)

      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, .tmp),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(ka <- 8))
      )

      .tmp <- quote(ka4 ~ 8)

      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, .tmp),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(ka4 ~ 8))
      )

      .tmp <- quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1))))

      expect_equal(testPipeQuote(tka=0.5, {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.ka ~ 3, eta.ka ~ 3,
      {
        tv = 3
        tcl = 10
        eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1)))
      }, eta.v ~ 0.2, .tmp),
      list(quote(tka <- 0.5),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.ka ~ 3),
           quote(eta.ka ~ 3),
           quote(tv <- 3),
           quote(tcl <- 10),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))),
           quote(eta.v ~ 0.2),
           quote(eta.v + eta.cl ~ unfix(cor(sd(0.3, 0.02, 0.1)))))
      )

    })

    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("Model Line from Expression, assign", {

      expect_equal(.getModelLineFromExpression(quote(ka), f), 1L)
      expect_equal(.getModelLineFromExpression(quote(d/dt(depot)), f), 4L)

      expect_equal(.getModelLineFromExpression(quote(f(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(F(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(lag(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(alag(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(rate(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(dur(depot)), f), -4L)

    })

    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        f(depot) <- 3
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("Model Line from Expression, lower case f()", {

      expect_equal(.getModelLineFromExpression(quote(ka), f), 1L)
      expect_equal(.getModelLineFromExpression(quote(d/dt(depot)), f), 4L)

      expect_equal(.getModelLineFromExpression(quote(f(depot)), f), 6L)
      expect_equal(.getModelLineFromExpression(quote(F(depot)), f), 6L)

      expect_equal(.getModelLineFromExpression(quote(lag(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(alag(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(rate(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(dur(depot)), f), -4L)

    })

    one.compartment <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        F(depot) <- 3
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("Model Line from Expression, upper case F()", {

      expect_equal(.getModelLineFromExpression(quote(ka), f), 1L)
      expect_equal(.getModelLineFromExpression(quote(d/dt(depot)), f), 4L)

      expect_equal(.getModelLineFromExpression(quote(f(depot)), f), 6L)
      expect_equal(.getModelLineFromExpression(quote(F(depot)), f), 6L)

      expect_equal(.getModelLineFromExpression(quote(lag(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(alag(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(rate(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(dur(depot)), f), -4L)

    })

    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        lag(depot) <- 3
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("Model Line from Expression, lag()", {

      expect_equal(.getModelLineFromExpression(quote(ka), f), 1L)
      expect_equal(.getModelLineFromExpression(quote(d/dt(depot)), f), 4L)

      expect_equal(.getModelLineFromExpression(quote(f(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(F(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(lag(depot)), f), 6L)
      expect_equal(.getModelLineFromExpression(quote(alag(depot)), f), 6L)

      expect_equal(.getModelLineFromExpression(quote(rate(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(dur(depot)), f), -4L)

    })

    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        alag(depot) <- 3
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("Model Line from Expression, alag()", {

      expect_equal(.getModelLineFromExpression(quote(ka), f), 1L)
      expect_equal(.getModelLineFromExpression(quote(d/dt(depot)), f), 4L)

      expect_equal(.getModelLineFromExpression(quote(f(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(F(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(lag(depot)), f), 6L)
      expect_equal(.getModelLineFromExpression(quote(alag(depot)), f), 6L)

      expect_equal(.getModelLineFromExpression(quote(rate(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(dur(depot)), f), -4L)

    })

    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        rate(depot) <- 3
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("Model Line from Expression, rate()", {

      expect_equal(.getModelLineFromExpression(quote(ka), f), 1L)
      expect_equal(.getModelLineFromExpression(quote(d/dt(depot)), f), 4L)

      expect_equal(.getModelLineFromExpression(quote(f(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(F(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(lag(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(alag(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(rate(depot)), f), 6L)
      expect_equal(.getModelLineFromExpression(quote(dur(depot)), f), -4L)

    })

    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        dur(depot) <- 3
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("Model Line from Expression, dur()", {

      expect_equal(.getModelLineFromExpression(quote(ka), f), 1L)
      expect_equal(.getModelLineFromExpression(quote(d/dt(depot)), f), 4L)

      expect_equal(.getModelLineFromExpression(quote(f(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(F(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(lag(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(alag(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(rate(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(dur(depot)), f), 6L)

      expect_equal(.getModelLineFromExpression(quote(not), f), NA_integer_)

    })

    # look at duplicate lines
    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d/dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("Model Line from Expression, duplicate d/dt(depot)", {

      expect_equal(.getModelLineFromExpression(quote(ka), f), 1L)

      expect_equal(.getModelLineFromExpression(quote(d/dt(depot)), f), NULL)

      expect_equal(.getModelLineFromExpression(quote(f(depot)), f), -5L)
      expect_equal(.getModelLineFromExpression(quote(F(depot)), f), -5L)

      expect_equal(.getModelLineFromExpression(quote(lag(depot)), f), -5L)
      expect_equal(.getModelLineFromExpression(quote(alag(depot)), f), -5L)

      expect_equal(.getModelLineFromExpression(quote(rate(depot)), f), -5L)
      expect_equal(.getModelLineFromExpression(quote(dur(depot)), f), -5L)

      expect_equal(.getModelLineFromExpression(quote(not), f), NA_integer_)

    })

    # look at duplicate lines
    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        f(depot) <- 3
        F(depot) <- 1
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("Model Line from Expression, duplicate f(depot)", {

      expect_equal(.getModelLineFromExpression(quote(ka), f), 1L)
      expect_equal(.getModelLineFromExpression(quote(d/dt(depot)), f), 4L)

      expect_equal(.getModelLineFromExpression(quote(f(depot)), f), NULL)
      expect_equal(.getModelLineFromExpression(quote(F(depot)), f), NULL)

      expect_equal(.getModelLineFromExpression(quote(lag(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(alag(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(rate(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(dur(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(not), f), NA_integer_)

    })

    # look at duplicate lag()
    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        lag(depot) <- 3
        alag(depot) <- 1
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("Model Line from Expression, duplicate f(depot)", {

      expect_equal(.getModelLineFromExpression(quote(ka), f), 1L)
      expect_equal(.getModelLineFromExpression(quote(d/dt(depot)), f), 4L)

      expect_equal(.getModelLineFromExpression(quote(f(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(F(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(lag(depot)), f), NULL)
      expect_equal(.getModelLineFromExpression(quote(alag(depot)), f), NULL)

      expect_equal(.getModelLineFromExpression(quote(rate(depot)), f), -4L)
      expect_equal(.getModelLineFromExpression(quote(dur(depot)), f), -4L)

      expect_equal(.getModelLineFromExpression(quote(not), f), NA_integer_)
      expect_equal(.getModelLineFromExpression(quote(cp), f), 8L)

      expect_equal(.getModelLineFromExpression(quote(cp), f, TRUE), 9L)

    })

    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    testEst <- function(ui, par, lower, value, upper, fix=FALSE) {
      uiForce <- suppressMessages(force(ui))
      .ini <- ui$iniDf
      .w <- which(.ini$name == par)
      expect_equal(length(.w), 1)
      expect_equal(.ini$lower[.w], lower)
      expect_equal(.ini$est[.w], value)
      expect_equal(.ini$upper[.w], upper)
      expect_equal(.ini$fix[.w], fix)
    }

    test_that("simple ini piping, uncorrelated model", {

      testEst(f, "tka", -Inf, 0.45, Inf, FALSE)
      testEst(f |> ini(tka=0.5), "tka", -Inf, 0.5, Inf, FALSE)
      testEst(f |> ini(tka=fix), "tka", -Inf, 0.45, Inf, TRUE)

      testEst(f |> ini(tka=c(0, 0.5)), "tka", 0, 0.5, Inf, FALSE)
      testEst(f |> ini(tka=c(0, 0.5, 1)), "tka", 0, 0.5, 1, FALSE)

      expect_error(f |> ini(tka=c(0, 0.5, 1, 4)), "tka")

      expect_error(f |> ini(tka=c(3,2,1)), "tka")

      suppressMessages(
        fFix <- f |> ini(tka=fix)
      )
      testEst(fFix, "tka", -Inf, 0.45, Inf, TRUE)
      testEst(fFix |> ini(tka=unfix), "tka", -Inf, 0.45, Inf, FALSE)
      testEst(fFix |> ini(tka=unfix(0.5)), "tka", -Inf, 0.5, Inf, FALSE)

      testEst(f |> ini(eta.v ~ 0.2), "eta.v", -Inf, 0.2, Inf, FALSE)

      testEst(f |> ini(eta.cl+eta.v~c(0.3, 0.02, 0.1)), "eta.cl", -Inf, 0.3, Inf, FALSE)
      testEst(f |> ini(eta.cl+eta.v~c(0.3, 0.02, 0.1)), "eta.v", -Inf, 0.1, Inf, FALSE)
      testEst(f |> ini(eta.cl+eta.v~c(0.3, 0.02, 0.1)), "(eta.cl,eta.v)", -Inf, 0.02, Inf, FALSE)

      testEst(f |> ini(eta.cl+eta.v~cor(0.3, 0.02, 0.1)), "eta.cl", -Inf, 0.3, Inf, FALSE)
      testEst(f |> ini(eta.cl+eta.v~cor(0.3, 0.02, 0.1)), "eta.v", -Inf, 0.1, Inf, FALSE)
      testEst(f |> ini(eta.cl+eta.v~cor(0.3, 0.02, 0.1)), "(eta.cl,eta.v)", -Inf, 0.02*(sqrt(0.3)*sqrt(0.1)), Inf, FALSE)

      testEst(f |> ini(eta.cl+eta.v~fix(cor(sd(0.3,0.02,0.1)))), "eta.cl", -Inf, 0.3 * 0.3, Inf, TRUE)
      testEst(f |> ini(eta.cl+eta.v~fix(cor(sd(0.3,0.02,0.1)))), "eta.v", -Inf, 0.1 * 0.1, Inf, TRUE)
      testEst(f |> ini(eta.cl+eta.v~fix(cor(sd(0.3,0.02,0.1)))), "(eta.cl,eta.v)", -Inf, 0.1 * 0.3 * 0.02, Inf, TRUE)

      # Test adding matrix directly

      .omega <- lotri::lotri(eta.cl+eta.v~c(0.3, 0.02, 0.1))

      testEst(f |> ini(.omega), "eta.cl", -Inf, 0.3, Inf, FALSE)
      testEst(f |> ini(.omega), "eta.v", -Inf, 0.1, Inf, FALSE)
      testEst(f |> ini(.omega), "(eta.cl,eta.v)", -Inf, 0.02, Inf, FALSE)

      expect_warning(expect_warning(
        testEst(f |> ini(eta.cl+eta.v~unfix(cor(sd(0.3,0.02,0.1)))), "eta.cl", -Inf, 0.3 * 0.3, Inf, FALSE),
        regexp="unfix.*eta.cl"), regexp="unfix.*eta.v"
        )
      expect_warning(expect_warning(
        testEst(f |> ini(eta.cl+eta.v~unfix(cor(sd(0.3,0.02,0.1)))), "eta.v", -Inf, 0.1 * 0.1, Inf, FALSE),
        regexp="unfix.*eta.cl"), regexp="unfix.*eta.v"
        )
      expect_warning(expect_warning(
        testEst(f |> ini(eta.cl+eta.v~unfix(cor(sd(0.3,0.02,0.1)))), "(eta.cl,eta.v)", -Inf, 0.1 * 0.3 * 0.02, Inf, FALSE),
        regexp="unfix.*eta.cl"), regexp="unfix.*eta.v"
        )

    })

    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl + eta.v ~ sd(cor(0.3,
                                -0.7, 0.1))
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("simple ini piping, correlated model", {

      testEst(f, "tka", -Inf, 0.45, Inf, FALSE)
      testEst(f |> ini(tka=0.5), "tka", -Inf, 0.5, Inf, FALSE)
      testEst(f |> ini(tka=fix), "tka", -Inf, 0.45, Inf, TRUE)

      testEst(f |> ini(tka=c(0, 0.5)), "tka", 0, 0.5, Inf, FALSE)
      testEst(f |> ini(tka=c(0, 0.5, 1)), "tka", 0, 0.5, 1, FALSE)

      expect_error(f |> ini(tka=c(0, 0.5, 1, 4)), "tka")

      expect_error(f |> ini(tka=c(3,2,1)), "tka")

      suppressMessages(
        fFix <- f |> ini(tka=fix)
      )
      testEst(fFix, "tka", -Inf, 0.45, Inf, TRUE)
      testEst(fFix |> ini(tka=unfix), "tka", -Inf, 0.45, Inf, FALSE)
      testEst(fFix |> ini(tka=unfix(0.5)), "tka", -Inf, 0.5, Inf, FALSE)

      testEst(f |> ini(eta.v ~ 0.2), "eta.v", -Inf, 0.2, Inf, FALSE)

      testEst(f |> ini(eta.cl+eta.v~c(0.3, 0.02, 0.1)), "eta.cl", -Inf, 0.3, Inf, FALSE)
      testEst(f |> ini(eta.cl+eta.v~c(0.3, 0.02, 0.1)), "eta.v", -Inf, 0.1, Inf, FALSE)
      testEst(f |> ini(eta.cl+eta.v~c(0.3, 0.02, 0.1)), "(eta.cl,eta.v)", -Inf, 0.02, Inf, FALSE)

      testEst(f |> ini(eta.cl+eta.v~cor(0.3, 0.02, 0.1)), "eta.cl", -Inf, 0.3, Inf, FALSE)
      testEst(f |> ini(eta.cl+eta.v~cor(0.3, 0.02, 0.1)), "eta.v", -Inf, 0.1, Inf, FALSE)
      testEst(f |> ini(eta.cl+eta.v~cor(0.3, 0.02, 0.1)), "(eta.cl,eta.v)", -Inf, 0.02*(sqrt(0.3)*sqrt(0.1)), Inf, FALSE)

      testEst(f |> ini(eta.cl+eta.v~fix(cor(sd(0.3,0.02,0.1)))), "eta.cl", -Inf, 0.3 * 0.3, Inf, TRUE)
      testEst(f |> ini(eta.cl+eta.v~fix(cor(sd(0.3,0.02,0.1)))), "eta.v", -Inf, 0.1 * 0.1, Inf, TRUE)
      testEst(f |> ini(eta.cl+eta.v~fix(cor(sd(0.3,0.02,0.1)))), "(eta.cl,eta.v)", -Inf, 0.1 * 0.3 * 0.02, Inf, TRUE)

      expect_warning(expect_warning(
        testEst(f |> ini(eta.cl+eta.v~unfix(cor(sd(0.3,0.02,0.1)))), "eta.cl", -Inf, 0.3 * 0.3, Inf, FALSE),
        regexp="unfix.*eta.cl"), regexp="unfix.*eta.v"
        )
      expect_warning(expect_warning(
        testEst(f |> ini(eta.cl+eta.v~unfix(cor(sd(0.3,0.02,0.1)))), "eta.v", -Inf, 0.1 * 0.1, Inf, FALSE),
        regexp="unfix.*eta.cl"), regexp="unfix.*eta.v"
        )
      expect_warning(expect_warning(
        testEst(f |> ini(eta.cl+eta.v~unfix(cor(sd(0.3,0.02,0.1)))), "(eta.cl,eta.v)", -Inf, 0.1 * 0.3 * 0.02, Inf, FALSE),
        regexp="unfix.*eta.cl"), regexp="unfix.*eta.v"
        )

    })

    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl + eta.v ~ fix(sd(cor(0.3,
                                    -0.7, 0.1)))
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("simple ini piping, fixed correlated model", {

      testEst(f, "tka", -Inf, 0.45, Inf, FALSE)
      testEst(f |> ini(tka=0.5), "tka", -Inf, 0.5, Inf, FALSE)
      testEst(f |> ini(tka=fix), "tka", -Inf, 0.45, Inf, TRUE)

      testEst(f |> ini(tka=c(0, 0.5)), "tka", 0, 0.5, Inf, FALSE)
      testEst(f |> ini(tka=c(0, 0.5, 1)), "tka", 0, 0.5, 1, FALSE)

      expect_error(f |> ini(tka=c(0, 0.5, 1, 4)), "tka")

      expect_error(f |> ini(tka=c(3,2,1)), "tka")

      suppressMessages(
        fFix <- f |> ini(tka=fix)
      )
      testEst(fFix, "tka", -Inf, 0.45, Inf, TRUE)
      testEst(fFix |> ini(tka=unfix), "tka", -Inf, 0.45, Inf, FALSE)
      testEst(fFix |> ini(tka=unfix(0.5)), "tka", -Inf, 0.5, Inf, FALSE)

      # should warn? Modify fixed value
      testEst(f |> ini(eta.v ~ 0.2), "eta.v", -Inf, 0.2, Inf, TRUE)

      testEst(f |> ini(eta.cl+eta.v~c(0.3, 0.02, 0.1)), "eta.cl", -Inf, 0.3, Inf, TRUE)
      testEst(f |> ini(eta.cl+eta.v~c(0.3, 0.02, 0.1)), "eta.v", -Inf, 0.1, Inf, TRUE)
      testEst(f |> ini(eta.cl+eta.v~c(0.3, 0.02, 0.1)), "(eta.cl,eta.v)", -Inf, 0.02, Inf, TRUE)

      testEst(f |> ini(eta.cl+eta.v~cor(0.3, 0.02, 0.1)), "eta.cl", -Inf, 0.3, Inf, TRUE)
      testEst(f |> ini(eta.cl+eta.v~cor(0.3, 0.02, 0.1)), "eta.v", -Inf, 0.1, Inf, TRUE)
      testEst(f |> ini(eta.cl+eta.v~cor(0.3, 0.02, 0.1)), "(eta.cl,eta.v)", -Inf, 0.02*(sqrt(0.3)*sqrt(0.1)), Inf, TRUE)

      expect_warning(expect_warning(
        testEst(f |> ini(eta.cl+eta.v~fix(cor(sd(0.3,0.02,0.1)))), "eta.cl", -Inf, 0.3 * 0.3, Inf, TRUE),
        regexp="fix.*eta.cl"), regexp="fix.*eta.v"
        )
      expect_warning(expect_warning(
        testEst(f |> ini(eta.cl+eta.v~fix(cor(sd(0.3,0.02,0.1)))), "eta.v", -Inf, 0.1 * 0.1, Inf, TRUE),
        regexp="fix.*eta.cl"), regexp="fix.*eta.v"
        )
      expect_warning(expect_warning(
        testEst(f |> ini(eta.cl+eta.v~fix(cor(sd(0.3,0.02,0.1)))), "(eta.cl,eta.v)", -Inf, 0.1 * 0.3 * 0.02, Inf, TRUE),
        regexp="fix.*eta.cl"), regexp="fix.*eta.v"
        )

      testEst(f |> ini(eta.cl+eta.v~unfix(cor(sd(0.3,0.02,0.1)))), "eta.cl", -Inf, 0.3 * 0.3, Inf, FALSE)
      testEst(f |> ini(eta.cl+eta.v~unfix(cor(sd(0.3,0.02,0.1)))), "eta.v", -Inf, 0.1 * 0.1, Inf, FALSE)
      testEst(f |> ini(eta.cl+eta.v~unfix(cor(sd(0.3,0.02,0.1)))), "(eta.cl,eta.v)", -Inf, 0.1 * 0.3 * 0.02, Inf, FALSE)
    })

    # |> ini(tka=0.5)
    # |> ini(tka=fix)
    # |> ini(tka=unfix)
    # |> ini(eta.v~0.2)

    # Try with |>
    # |> ini(eta.cl+eta.v~c(0.3, 0.02, 0.1))
    # |> ini(eta.cl+eta.v~cor(0.3, 0.02, 0.1))
    # |> ini(eta.v+eta.cl~fix(cor(sd(0.3,0.02,0.1))))
    # |> ini(eta.v+eta.cl~unfix(cor(sd(0.3,0.02,0.1))))

    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    testUi <- function(ui, has = NULL, exclude = NULL, values = NULL) {
      uiForce <- suppressMessages(force(ui))
      if (!is.null(has)) {
        expect_true(all(has %in% paste(uiForce$ini$name)))
      }
      if (!is.null(values) && !is.null(names(values))) {
        .vals <- setNames(uiForce$ini$est, paste(uiForce$ini$name))
        .vals <- .vals[names(values)]
        expect_equal(values, .vals)
      }
      if (!is.null(exclude)) {
        expect_false(any(exclude %in% paste(uiForce$ini$name)))
      }
      ## General UI properties
      expect_true(all(!is.na(uiForce$ini$fix)))
      expect_true(all(!is.na(uiForce$ini$lower)))
      expect_true(all(!is.na(uiForce$ini$upper)))
    }

    test_that("update: Test Base model", {

      testUi(f, c("tka", "tcl", "tv", "eta.ka", "eta.cl", "eta.v", "add.err"),
             "matt", c(tka = 0.45, tcl = 1, tv = 3.45, eta.ka = 0.6, eta.cl = 0.3, eta.v = 0.1, add.err = 0.7))

    })

    test_that("UI updates work correctly", {

      # context("update: Multiple component change with c()")
      testUi(
        f |> update(tka = 4, cl = exp(tcl), ka = exp(tka), c(tcl = 3, tv = 4)),
        c("tka", "tcl", "tv", "eta.v", "add.err"),
        c("eta.ka", "eta.cl"),
        c(tka = 4, tcl = 3, tv = 4, eta.v = 0.1, add.err = 0.7)
      )

      # context("update: Multiple component change with list()")
      testUi(
        f |> update(tka = 4, cl = exp(tcl), ka = exp(tka), list(tcl = 3, tv = 4)),
        c("tka", "tcl", "tv", "eta.v", "add.err"),
        c("eta.ka", "eta.cl"),
        c(tka = 4, tcl = 3, tv = 4, eta.v = 0.1, add.err = 0.7)
      )

      # context("update: Multiple component change with assigned .tmp=list()")
      .tmp <- list(tcl = 3, tv = 4)
      testUi(
        f |> update(tka = 4, cl = exp(tcl), ka = exp(tka), .tmp),
        c("tka", "tcl", "tv", "eta.v", "add.err"),
        c("eta.ka", "eta.cl"),
        c(tka = 4, tcl = 3, tv = 4, eta.v = 0.1, add.err = 0.7)
      )

      # context("update: Multiple component change with assigned .tmp=c()")
      .tmp <- c(tcl = 3, tv = 4)
      testUi(
        f |> update(tka = 4, cl = exp(tcl), ka = exp(tka), .tmp),
        c("tka", "tcl", "tv", "eta.v", "add.err"),
        c("eta.ka", "eta.cl"),
        c(tka = 4, tcl = 3, tv = 4, eta.v = 0.1, add.err = 0.7)
      )

      # context("update: Multiple component change with assigned .tmp={}")
      .tmp <- quote({
        ka <- exp(tka)
      })
      testUi(
        f |> update(tka = 4, cl = exp(tcl), .tmp, c(tcl = 3, tv = 4)),
        c("tka", "tcl", "tv", "eta.v", "add.err"),
        c("eta.ka", "eta.cl"),
        c(tka = 4, tcl = 3, tv = 4, eta.v = 0.1, add.err = 0.7)
      )

      testUi(
        f |> update(
          tka = 4,
          cl = exp(tcl),
          {
            ka <- exp(tka)
          },
          c(tcl = 3, tv = 4)
        ),
        c("tka", "tcl", "tv", "eta.v", "add.err"),
        c("eta.ka", "eta.cl"),
        c(tka = 4, tcl = 3, tv = 4, eta.v = 0.1, add.err = 0.7)
      )

      testUi(
        f |> update(ka = exp(tka)),
        c("tka", "tcl", "tv", "eta.cl", "eta.v", "add.err"),
        "eta.ka", c(tka = 0.45, tcl = 1, tv = 3.45, eta.cl = 0.3, eta.v = 0.1, add.err = 0.7)
      )

      ## Now test linCmt() issue #166
      one.cmt <- function() {
        ini({
          tka <- 0.45 ; label("Log Ka")
          tcl <- 1 ; label("Log Cl")
          tv <- 3.45 ; label("Log V")
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.err <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          linCmt() ~ add(add.err)
        })
      }

      suppressMessages(
        .ui <- one.cmt |> update({
          linCmt() ~ add(add.err) + prop(prop.err)
        })
      )
      expect_s3_class(.ui, "rxUi")

    })

    # piping looks through parent environments
    test_that("Looks through prior frames for the correct object", {

      fit <- rxode2(one.compartment)
      fits <- lapply(seq(-1, -0.1, 0.1), function(kainit) {
        suppressMessages(
          rxode2(update(fit, tka = kainit))
        )
      })

      expect_type(fits, "list")
      expect_error(lapply(seq(-1, -0.1, 0.1), function(kainit) {
        suppressMessages(
          rxode2(update(fit, tka = matt))
        )
      }))

    })

    one.compartment <- function() {
      ini({
        tka <- 0.45 ; label("Log Ka")
        tcl <- 1 ; label("Log Cl")
        tv <- 3.45 ; label("Log V")
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    test_that("piping works for correlations #1", {

      testUi(f |> ini(eta.ka + eta.cl ~ c(
        0.2,
        0.01, 0.2
      )),
      has = c("tka", "tcl", "tv", "eta.ka", "eta.cl", "eta.v", "add.err", "(eta.ka,eta.cl)"),
      exclude = "matt",
      values = c(
        tka = 0.45, tcl = 1, tv = 3.45, eta.ka = 0.2, eta.cl = 0.2, eta.v = 0.1, add.err = 0.7,
        `(eta.ka,eta.cl)` = 0.01
      ))

    })

    test_that("piping works for correlations #2", {

      suppressMessages(
        expect_error(
          f |>
            ini(eta.ka + eta.matt ~ c(0.2,
                                      0.01, 0.2)
                )))

    })

    test_that("piping works for correlations #3", {

      testUi(
        f |> update(eta.ka + eta.cl ~ c(
          0.2,
          0.01, 0.2
        )),
        c("tka", "tcl", "tv", "eta.ka", "eta.cl", "eta.v", "add.err", "(eta.ka,eta.cl)"),
        "matt", c(
          tka = 0.45, tcl = 1, tv = 3.45, eta.ka = 0.2, eta.cl = 0.2, eta.v = 0.1, add.err = 0.7,
          `(eta.ka,eta.cl)` = 0.01))

    })

    test_that("piping works for correlations #4", {

      suppressMessages(
        expect_error(
          f |>
            update(eta.ka + eta.matt ~ c(0.2,
                                         0.01, 0.2))))

    })

    test_that("expected piping errors", {

      f <- function() {
        ini({
          ke <- 0.5
          eta.ke ~ 0.04
          prop.sd <- sqrt(0.1)
        })
        model({
          ke <- ke * exp(eta.ke)
          ipre <- 10 * exp(-ke * t)
          ipre ~ prop(prop.sd)
        })
      }

      f <- rxode2::rxode2(f)

      suppressMessages(
        expect_error(f |> model(ipre ~ add(add.sd)) |> ini(add.sd=sqrt(0.1)), NA))

    })

    test_that("new ipre", {

      f <- function() {
        ini({
          tke <- 0.5
          eta.ke ~ 0.04
          prop.sd <- sqrt(0.1)
        })
        model({
          ke <- tke * exp(eta.ke)
          ipre <- 10 * exp(-ke * t)
          f2 <- ipre / (ipre + 5)
          ipre ~ prop(prop.sd)
        })
      }

      f <- rxode2(f)

      trans <- function(f) {
        suppressMessages(
          f |> model(ipre ~ propF(prop.sd, f2)) |> ini(prop.sd=sqrt(0.1))
        )
      }

      f2 <- trans(f)

      expect_true(!any(f2$iniDf$name %in% "f2"))

    })

    test_that("piping looks in the right environment for variables with fix()", {

      f <- function() {
        ini({
          tke <- 0.5
          eta.ke ~ 0.04
          prop.sd <- sqrt(0.1)
        })
        model({
          ke <- tke * exp(eta.ke)
          ipre <- 10 * exp(-ke * t)
          f2 <- ipre / (ipre + 5)
          ipre ~ prop(prop.sd)
        })
      }

      intke <- 5
      suppressMessages(
        tmp <- f |> ini(tke=fix(intke))
      )

      expect_true(tmp$iniDf[tmp$iniDf$name == "tke","fix"])
      expect_equal(tmp$iniDf[tmp$iniDf$name == "tke","est"], 5)
      rm(list="intke")

      f2 <- function() {
        f <- function() {
          ini({
            tke <- 0.5
            eta.ke ~ 0.04
            prop.sd <- sqrt(0.1)
          })
          model({
            ke <- tke * exp(eta.ke)
            ipre <- 10 * exp(-ke * t)
            f2 <- ipre / (ipre + 5)
            ipre ~ prop(prop.sd)
          })
        }

        intke <- 5
        f |> ini(tke=fix(intke))

      }

      suppressMessages(
        tmp <- f2()
      )

      expect_true(tmp$iniDf[tmp$iniDf$name == "tke","fix"])
      expect_equal(tmp$iniDf[tmp$iniDf$name == "tke","est"], 5)

      expect_false(any(ls() == "intke"))

    })

    test_that("invalid model pipe (more arguments than expected) throws an error", {

      f <- function() {
        ini({
          tke <- 0.5
          eta.ke ~ 0.04
          prop.sd <- sqrt(0.1)
        })
        model({
          ke <- tke * exp(eta.ke)
          ipre <- 10 * exp(-ke * t)
          f2 <- ipre / (ipre + 5)
          f3 <- f2 * 3
          lipre <- log(ipre)
          ipre ~ prop(prop.sd)
        })
      }

      expect_error(f |> model(ipre~prop(f2,f3,c)))

    })

    test_that("Add an eta to a model that does not have an eta will work", {

      ocmt <- function() {
        ini({
          tka <- exp(0.45)
          tcl <- exp(1)
          tv <- exp(3.45)
          add.sd <- 0.7
        })
        model({
          ka <- tka
          cl <- tcl
          v <- tv
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
          cp ~ add(add.sd)
        })
      }

      suppressMessages(
        expect_error(
          ocmt |>
            model(ka <- exp(tka + eta.ka)),
          NA))

    })

    test_that("Add covariate to model works", {

      ocmt <- function() {
        ini({
          tka <- exp(0.45)
          tcl <- exp(1)
          tv <- exp(3.45)
          add.sd <- 0.7
        })
        model({
          ka <- tka
          cl <- tcl
          v <- tv
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
          cp ~ add(add.sd)
        })
      }


      suppressMessages(
        expect_error(
          ocmt |>
            model(ka <- exp(tka + covKa * wt + eta.ka)),
          NA))

      suppressMessages(
        tmp <-
          ocmt |>
          model(ka <- exp(tka + covKaWt * wt + eta.ka)))
      expect_equal(tmp$allCovs, "wt")

      expect_true("covKaWt" %in% tmp$iniDf$name)
      expect_true("tka" %in% tmp$iniDf$name)
      expect_true("eta.ka" %in% tmp$iniDf$name)

      suppressMessages(
        tmp <-
          ocmt |>
          model(ka <- exp(covKaWt * wt + eta.ka)))

      expect_equal(tmp$allCovs, "wt")
      expect_true("covKaWt" %in% tmp$iniDf$name)
      expect_false("tka" %in% tmp$iniDf$name)
      expect_true("eta.ka" %in% tmp$iniDf$name)

      suppressMessages(
        tmp <-
          tmp |>
          model(ka <- exp(tka + covKaWt * wt + eta.ka)))
      expect_equal(tmp$allCovs, "wt")
      expect_true("covKaWt" %in% tmp$iniDf$name)
      expect_true("tka" %in% tmp$iniDf$name)
      expect_true("eta.ka" %in% tmp$iniDf$name)

    })

    test_that("Appending or pre-pending items to a model works", {

      ocmt <- function() {
        ini({
          tka <- exp(0.45)
          tcl <- exp(1)
          tv <- exp(3.45)
          add.sd <- 0.7
        })
        model({
          ka <- tka
          cl <- tcl
          v <- tv
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
          cp ~ add(add.sd)
        })
      }

      f <- rxode2(ocmt)
      f2 <- f |> model(cp1 <- cp, append=TRUE)

      expect_true("cp1" %in% f2$mv0$lhs)
      expect_equal(f2$lstExpr[[length(f2$lstExpr)]], quote(cp1 <- cp))

      f <- rxode2(ocmt)
      f2 <- f |> model(cp1 <- cp, append=Inf)

      expect_true("cp1" %in% f2$mv0$lhs)
      expect_equal(f2$lstExpr[[length(f2$lstExpr)]], quote(cp1 <- cp))

      f <- rxode2(ocmt)
      f2 <- f |> model(cp1 <- cp, append=100)

      expect_true("cp1" %in% f2$mv0$lhs)
      expect_equal(f2$lstExpr[[length(f2$lstExpr)]], quote(cp1 <- cp))

      f2 <- f |> model(f2 <- 3 * 2, append=NA)
      expect_true("f2" %in% f2$mv0$lhs)
      expect_equal(f2$lstExpr[[1]], quote(f2 <- 3 * 2))

    })

    test_that("ini promotion works", {

      ocmt <- function() {
        ini({
          tka <- 0.45
          tcl <- 1
          add.sd <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
          cp ~ add(add.sd)
        })
      }

      f <- rxode2(ocmt)

      expect_equal(f$allCovs, c("eta.ka", "eta.cl", "tv", "eta.v"))
      expect_equal(f$theta, c(tka=0.45, tcl=1, add.sd=0.7))
      expect_equal(f$omega, NULL)

      # now promote tv
      suppressMessages(
        f2 <- f |> ini(tv=0.5))
      expect_equal(f2$allCovs, c("eta.ka", "eta.cl", "eta.v"))
      expect_equal(f2$theta, c(tka=0.45, tcl=1, add.sd=0.7, tv=0.5))
      expect_equal(f2$omega, NULL)

      # now promote eta.ka
      suppressMessages(
        f3 <- f2 |> ini(eta.ka ~ 0.01))

      expect_equal(f3$allCovs, c("eta.cl", "eta.v"))
      expect_equal(f3$theta, c(tka=0.45, tcl=1, add.sd=0.7, tv=0.5))
      expect_equal(f3$omega, matrix(0.01, dimnames=list("eta.ka", "eta.ka")))

      # now promote a correlation between eta.cl and eta.v
      suppressMessages(
        f4 <- f2 |> ini(eta.cl + eta.v ~ c(1,
                                            0.01, 1)))
      expect_equal(f4$allCovs, "eta.ka")
      expect_equal(f4$theta, c(tka=0.45, tcl=1, add.sd=0.7, tv=0.5))

      expect_equal(f4$omega, lotri(eta.cl + eta.v ~ c(1,
                                                      0.01, 1)))

      # Now promote independent eta block
      suppressMessages(
        f5 <- f3 |> ini(eta.cl + eta.v ~ c(1,
                                            0.01, 1)))
      expect_length(f5$allCovs, 0)
      expect_equal(f5$theta, c(tka=0.45, tcl=1, add.sd=0.7, tv=0.5))
      expect_equal(f5$omega, lotri(eta.ka ~ 0.01,
                                   eta.cl + eta.v ~ c(1,
                                                      0.01, 1)))

      # Now promote eta block that includes prior eta information
      suppressMessages(
        f6 <- f3 |> ini(eta.ka + eta.cl + eta.v ~ c(1,
                                                     0.01, 1,
                                                     -0.01, 0.01, 1)))
      expect_length(f6$allCovs, 0)
      expect_equal(f6$theta, c(tka=0.45, tcl=1, add.sd=0.7, tv=0.5))
      expect_equal(f6$omega, lotri(eta.ka + eta.cl + eta.v ~ c(1,
                                                               0.01, 1,
                                                               -0.01, 0.01, 1)))

    })

    test_that("Ignoring auto-selected parameter types work", {

      ocmt <- function() {
        ini({
          tka <- exp(0.45)
          tcl <- exp(1)
          eta.v ~ 0.01
          add.sd <- 0.7
        })
        model({
          ka <- tka
          cl <- tcl
          v <- eta.v
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
          cp ~ add(add.sd)
        })
      }

      suppressWarnings(
        f <- rxode2(ocmt))

      expect_equal(f$allCovs, character(0))
      expect_equal(f$theta, c(tka=exp(0.45), tcl=exp(1), add.sd=0.7))
      expect_equal(f$omega, matrix(0.01, dimnames=list("eta.v", "eta.v")))

      suppressMessages(suppressWarnings(
        f2 <- f |> model(ka <- tka * exp(eta.ka), auto=FALSE)
      ))

      expect_equal(f2$allCovs, "eta.ka")
      expect_equal(f2$theta, c(tka=exp(0.45), tcl=exp(1), add.sd=0.7))
      expect_equal(f2$omega, matrix(0.01, dimnames=list("eta.v", "eta.v")))

      suppressMessages(suppressWarnings(
        f2 <-
          f |>
          model(ka <- tka * exp(eta.ka), auto=FALSE) |>
          ini(eta.ka ~ 0.02)))
      expect_equal(f2$allCovs, character(0))
      expect_equal(f2$theta, c(tka=exp(0.45), tcl=exp(1), add.sd=0.7))
      expect_equal(f2$omega, lotri(eta.v ~ 0.01,
                                   eta.ka ~ 0.02))

      suppressMessages(suppressWarnings(
        f2 <- f |> model(v <- tv + eta.v, auto=FALSE)
      ))
      expect_equal(f2$allCovs, "tv")
      expect_equal(f2$theta, c(tka=exp(0.45), tcl=exp(1), add.sd=0.7))
      expect_equal(f2$omega, lotri(eta.v ~ 0.01))

      suppressMessages(suppressWarnings(
        f2 <- f |> model(v <- tv + eta.v, auto=FALSE) |>
          ini(tv=0.2)
      ))
      expect_equal(f2$allCovs, character(0))
      expect_equal(f2$theta, c(tka=exp(0.45), tcl=exp(1), add.sd=0.7, tv=0.2))
      expect_equal(f2$omega, lotri(eta.v ~ 0.01))

    })

    test_that("Ignoring auto-selected parameter types work", {

      ocmt <- function() {
        ini({
          tka <- exp(0.45)
          tcl <- exp(1)
          eta.v ~ 0.01
          add.sd <- 0.7
          tprop <- 0.5
          prop.eta ~ 0.01
        })
        model({
          ka <- tka
          cl <- tcl
          v <- eta.v
          d/dt(depot) = -ka * depot
          d/dt(center) = ka * depot - cl / v * center
          cp = center / v
          prop.sd <- exp(tprop + prop.eta)
          cp ~ add(add.sd)
        })
      }

      suppressWarnings(
        f1 <- ocmt()
      )
      suppressWarnings(
        f2 <- ocmt |> model(cp ~ add(add.sd) + prop(prop.sd))
      )
      expect_equal(f2$theta, f1$theta)
      expect_equal(f2$omega, f1$omega)

    })

    test_that("Pre-declaring list of covariates works", {

      rxSetCovariateNamesForPiping(c("WT","HT", "TC"))

      # Note this is case sensitive
      one.compartment <- function() {
        ini({
          tka <- 0.45
          tcl <- 1
          tv <- 3.45
          eta.ka ~ 0.6
          eta.cl ~ 0.3
          eta.v ~ 0.1
          add.err <- 0.7
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          d / dt(depot) <- -ka * depot
          d/dt(depot) <- -ka * depot
          d / dt(center) <- ka * depot - cl / v * center
          cp <- center / v
          cp ~ add(add.err)
        })
      }

      # now TC is detected as a covariate instead of a population parameter
      suppressMessages(
        mod <-
          one.compartment |>
          model({ka <- exp(tka + eta.ka + TC * cov_C)})
      )
      expect_true("cov_C" %in% mod$iniDf$name)
      expect_false("TC" %in% mod$iniDf$name)

      rxSetCovariateNamesForPiping()

      suppressMessages(
        mod <-
          one.compartment |>
          model({ka <- exp(tka + eta.ka + TC * cov_C)})
      )
      expect_true("cov_C" %in% mod$iniDf$name)
      expect_true("TC" %in% mod$iniDf$name)

    })
  })

  test_that("eff(0) piping should work", {

    mod1 <- rxode2({
      C2 <- centr/V2
      C3 <- peri/V3
      d/dt(depot) <- -KA*depot
      d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
      d/dt(peri)  <- Q*C2 - Q*C3
      d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
    })

    suppressMessages(
      expect_error(
        mod1 |>
          model(KA<-exp(tka+eta.ka), append=NA) |> # Prepend a line by append=NA
          ini(tka=log(2.94E-01),
              eta.ka=0.2,
              CL=1.86E+01, V2=4.02E+01, # central
              Q=1.05E+01,  V3=2.97E+02, # peripheral
              Kin=1, Kout=1, EC50=200) |>
          model(eff(0) <- 1),
        NA))

  })


  test_that("auto with studid==", {

    one.compartment <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv+eta.v)
        cp <- linCmt()
        cp ~ add(add.sd)
      })
    }


    i <- rxode2(one.compartment)

    j <- i |>
      model({
        f(central)  <- 1 + f_study1*(STUDYID==1)
      },
      append=NA,
      auto=FALSE)

    expect_false(any(j$iniDf$name == "f_study1"))
    expect_false(any(j$iniDf$name == "STUDYID"))

  })

  test_that("piping with append=lhs", {

    ocmt_rx0 <- rxode2( {
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl / v * center
      cp = center / v
    })

    m1 <- ocmt_rx0 |> model( cl <- tvcl*2, append = NA)

    expect_true(identical(m1$lstExpr[[1]], quote(cl <- tvcl * 2)))

    m2 <- ocmt_rx0 |> model( cl <- tvcl*2, append = d/dt(depot))

    expect_true(identical(m2$lstExpr[[2]], quote(cl <- tvcl * 2)))

    expect_error(ocmt_rx0 |> model( cl <- tvcl*2, append = notFound))

    m3 <- ocmt_rx0 |> model( cl <- tvcl*2, append = cp)

    expect_true(identical(m3$lstExpr[[4]], quote(cl <- tvcl * 2)))

    test_that("piping ui functions", {

      m1 <- function() {
        ini({
          tka <- 0.463613555325211
          label("Ka")
          tcl <- c(-Inf, 1.01211464338867, 4.60517018598809)
          label("Log Cl")
          tv <- 3.46039743010498
          label("log V")
          add.sd <- c(0, 0.694761430696633)
          eta.ka ~ 0.400673718508127
          eta.cl ~ 0.069154564934726
          eta.v ~ 0.0191298379535425
        })
        model({
          ka <- exp(tka + eta.ka)
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          linCmt() ~ add(add.sd)
        })
      }

      m1 <- m1()

      m2 <- function() {
        ini({
          tcl <- c(-Inf, 1.01211464338867, 4.60517018598809)
          label("Log Cl")
          tv <- 3.46039743010498
          label("log V")
          add.sd <- c(0, 0.694761430696633)
          eta.cl ~ 0.069154564934726
          eta.v ~ 0.0191298379535425
        })
        model({
          cl <- exp(tcl + eta.cl)
          v <- exp(tv + eta.v)
          linCmt() ~ add(add.sd)
        })
      }

      m2 <- m2()

      expect_equal(testPipeQuote(m1, iniDf=m2$iniDf),
                   list(quote(tcl <- c(-Inf, 1.01211464338867, 4.60517018598809)),
                        quote(tv <- 3.46039743010498),
                        quote(add.sd <- c(0, 0.694761430696633)),
                        quote(eta.cl ~ 0.069154564934726),
                        quote(eta.v ~ 0.0191298379535425)))

      expect_equal(testPipeQuote(m2, iniDf=m1$iniDf),
                   list(quote(tcl <- c(-Inf, 1.01211464338867, 4.60517018598809)),
                        quote(tv <- 3.46039743010498),
                        quote(add.sd <- c(0, 0.694761430696633)),
                        quote(eta.cl ~ 0.069154564934726),
                        quote(eta.v ~ 0.0191298379535425)))

      m4 <- function() {
        ini({
          tcl <- c(-Inf, 1.01211464338867, 4.60517018598809)
          label("Log Cl")
          tv <- 3.46039743010498
          label("log V")
          add.sd <- c(0, 0.694761430696633)
          eta..cl ~ 0.069154564934726
          eta..v ~ 0.0191298379535425
        })
        model({
          cl <- exp(tcl + eta..cl)
          v <- exp(tv + eta..v)
          linCmt() ~ add(add.sd)
        })
      }

      # no etas

      m4 <- m4()

      expect_equal(testPipeQuote(m4, iniDf=m1$iniDf),
                   list(quote(tcl <- c(-Inf, 1.01211464338867, 4.60517018598809)),
                        quote(tv <- 3.46039743010498),
                        quote(add.sd <- c(0, 0.694761430696633))))

      expect_equal(testPipeQuote(m1, iniDf=m4$iniDf),
                   list(quote(tcl <- c(-Inf, 1.01211464338867, 4.60517018598809)),
                        quote(tv <- 3.46039743010498),
                        quote(add.sd <- c(0, 0.694761430696633))))

      # no thetas

      m5 <- function() {
        ini({
          t.cl <- c(-Inf, 1.01211464338867, 4.60517018598809)
          label("Log Cl")
          t.v <- 3.46039743010498
          label("log V")
          add..sd <- c(0, 0.694761430696633)
          eta.cl ~ 0.069154564934726
          eta.v ~ 0.0191298379535425
        })
        model({
          cl <- exp(t.cl + eta.cl)
          v <- exp(t.v + eta.v)
          linCmt() ~ add(add..sd)
        })
      }

      m5 <- m5()

      expect_equal(testPipeQuote(m5, iniDf=m1$iniDf),
                   list(quote(eta.cl ~ 0.069154564934726),
                        quote(eta.v ~ 0.0191298379535425)))

      expect_equal(testPipeQuote(m1, iniDf=m5$iniDf),
                   list(quote(eta.cl ~ 0.069154564934726),
                        quote(eta.v ~ 0.0191298379535425)))


      m6 <- function() {
        ini({
          t.cl <- c(-Inf, 1.01211464338867, 4.60517018598809)
          label("Log Cl")
          t.v <- 3.46039743010498
          label("log V")
          add..sd <- c(0, 0.694761430696633)
          eta..cl ~ 0.069154564934726
          eta..v ~ 0.0191298379535425
        })
        model({
          cl <- exp(t.cl + eta..cl)
          v <- exp(t.v + eta..v)
          linCmt() ~ add(add..sd)
        })
      }

      m6 <- m6()

      expect_equal(testPipeQuote(m6, iniDf=m1$iniDf),
                   list())

      expect_equal(testPipeQuote(m1, iniDf=m6$iniDf),
                   list())
    })
    test_that("model piping that shares err parameter#427", {
      u <- function() {
        ini({
          b <- 3
          err.sd <- 2
        })
        model({
          a <- x + err.sd
          c <- 1+b
          c ~ add(err.sd)
        })
      }

      expect_error(u |> model(-a), NA)
    })

    test_that("adding a constant does not add to the ini block", {
      u <- function() {
        ini({
          b <- 3
          err.sd <- 2
        })
        model({
          a <- x + err.sd
          c <- 1+b
          c ~ add(err.sd)
        })
      }

      n <- u |> model(aa <- pi+4, append=c)

      expect_false(any(n$iniDf$name == "pi"))
    })

    test_that("adding a line with a defined constant doesn't add to ini()", {
      u <- function() {
        ini({
          b <- 3
          err.sd <- 2
        })
        model({
          a <- x + err.sd
          aa <- 3
          c <- 1+b
          c ~ add(err.sd)
        })
      }

      n <- u |> model(aaa <- aa+4, append=c)

      expect_false(any(n$iniDf$name == "aa"))
    })
  })


  test_that("test ui appending of derived variables like `sim` can work", {

    one.compartment <- function() {
      ini({
        tka <- 0.45
        tcl <- 1
        tv <- 3.45
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        add.err <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d / dt(depot) <- -ka * depot
        d / dt(center) <- ka * depot - cl / v * center
        cp <- center / v
        cp ~ add(add.err)
      })
    }

    f <- rxode2(one.compartment)

    expect_error(model(f$simulationModel, sim2=sim+1, append=sim), NA)

  })


  test_that("off-diagonal piping issue #518", {

    mod <- function() {
      ini({
        a <- 1
        b <- 2
        etaa + etab ~ c(3, 0.1, 4)
        c <- 5
        etac ~ 6
        d <- 7
        f <- 9
        etad + etaf ~ c(8, 0.2, 10)
      })
      model({
        g <- (a + etaa)/(b + etab)
        h <- (c + etac)
        i <- (d + etad)
        j <- f + etaf
      })
    }

    modNew <-
      ini(
        rxode2(mod),
        etab + etac + etad ~
          c(7,
            0.2, 8,
            0.3, 0.4, 9),
        etaa ~ 0
      )

    expect_error(modNew$omega, NA)

  })

  test_that("piping append", {

    mod <- function() {
      ini({
        tka <- 0.45
        label("Ka")
        tcl <- 1
        label("Cl")
        tv <- 3.45
        label("V")
        add.sd <- c(0, 0.7)
        eta.cl ~ 0.3
        eta.v ~ 0.1
      })
      model({
        ka <- exp(tka)
        cl <- exp(tcl + eta.cl)
        v <- exp(tv + eta.v)
        d/dt(depot) = -ka * depot
        d/dt(center) = ka * depot - cl/v * center
        cp = center/v
        cp ~ add(add.sd)
      })
    }

    t <- c("-cp","-d/dt(depot)")
    expect_error(mod |> model(t), NA)

    t <- c("cp <- NULL","d/dt(depot) = NULL")
    expect_error(mod |> model(t), NA)

    t <- c("cp <- NULL","d/dt(depot) ~ NULL")
    expect_error(mod |> model(t), NA)

    mod5 <- mod |>
      model({
        PD <- 1-emax*cp/(ec50+cp)
        ##
        effect(0) <- e0
        kin <- e0*kout
        d/dt(effect) <- kin*PD -kout*effect
      }, append=d/dt(center))

    expect_equal(mod5$theta, c(tka = 0.45, tcl = 1, tv = 3.45, add.sd = 0.7))

    mod5 <- mod |>
      model({
        PD <- 1-emax*cp/(ec50+cp)
        ##
        effect(0) <- e0
        kin <- e0*kout
        d/dt(effect) <- kin*PD -kout*effect
      }, append="d/dt(center)")

    expect_equal(mod5$theta, c(tka = 0.45, tcl = 1, tv = 3.45, add.sd = 0.7))

    mod6 <- mod5 |>
      model({
        emax <- exp(temax)
        e0 <- exp(te0 + eta.e0)
        ec50 <- exp(tec50)
        kin <- exp(tkin)
        kout <- exp(tkout)
      }, append=NA)

    expect_equal(mod6$theta,
                 c(tka = 0.45, tcl = 1, tv = 3.45, add.sd = 0.7, temax = 1, te0 = 1, tec50 = 1, tkin = 1, tkout = 1))

    expect_equal(
      mod6$omega,
      lotri({
        eta.cl ~ 0.3
        eta.v ~ 0.1
        eta.e0 ~ 1
      }))

    mod6 <- mod5 |>
      model({
        emax <- exp(temax)
        e0 <- exp(te0 + eta.e0)
        ec50 <- exp(tec50)
        kin <- exp(tkin)
        kout <- exp(tkout)
      }, append=FALSE)

    expect_equal(
      mod6$omega,
      lotri({
        eta.cl ~ 0.3
        eta.v ~ 0.1
        eta.e0 ~ 1
      }))

    expect_equal(mod6$theta,
                 c(tka = 0.45, tcl = 1, tv = 3.45, add.sd = 0.7, temax = 1, te0 = 1, tec50 = 1, tkin = 1, tkout = 1))

    mod6 <- mod5 |>
      model({
        emax <- exp(temax)
        e0 <- exp(te0 + eta.e0)
        ec50 <- exp(tec50)
        kin <- exp(tkin)
        kout <- exp(tkout)
      }, append=0)

    expect_equal(
      mod6$omega,
      lotri({
        eta.cl ~ 0.3
        eta.v ~ 0.1
        eta.e0 ~ 1
      }))

    expect_equal(mod6$theta,
                 c(tka = 0.45, tcl = 1, tv = 3.45, add.sd = 0.7, temax = 1, te0 = 1, tec50 = 1, tkin = 1, tkout = 1))

    # make sure auto model piping turns off

    withr::with_options(list(rxode2.autoVarPiping=FALSE),
                        mod7 <- mod5 |>
                          model({
                            emax <- exp(temax)
                            e0 <- exp(te0 + eta.e0)
                            ec50 <- exp(tec50)
                            kin <- exp(tkin)
                            kout <- exp(tkout)
                          }, append=NA))

    expect_equal(mod7$theta,
                 c(tka = 0.45, tcl = 1, tv = 3.45, add.sd = 0.7))

    expect_equal(
      mod7$omega,
      lotri({
        eta.cl ~ 0.3
        eta.v ~ 0.1
      }))


  })

  test_that("piping appends a state after the simulation model is cached (persistent meta)", {
    # Realizing $simulationModel populates meta$.simModelBase.  Piping copies
    # the previous model's meta env (to keep sticky items); the copy must not
    # share the env by reference -- otherwise either model can pick up the
    # other's cached simulation model, in whichever order they are solved.
    f <- function() {
      ini({
        ke <- 1
        sd <- 0.1
      })
      model({
        d/dt(CENTRAL) <- -ke * CENTRAL
        cp <- CENTRAL
        cp ~ add(sd)
      })
    }
    m <- rxode2(f)
    # prime the cached simulation model before modifying the model
    expect_false(is.null(m$simulationModel))

    m2 <- m %>% model(d/dt(AUC) <- cp, append = TRUE)
    # the meta env must not be shared by reference after piping
    expect_false(identical(rxUiDecompress(m)$meta, rxUiDecompress(m2)$meta))
    # the appended state must be present in the (rebuilt) simulation model
    expect_true("AUC" %in% rxModelVars(m2$simulationModel)$state)

    .ev <- et(amt = 100, cmt = "CENTRAL") %>% et(0:5)
    s <- suppressWarnings(rxSolve(m2, .ev))
    expect_true("AUC" %in% names(s))

    # solving the piped model must not contaminate the original model
    expect_false("AUC" %in% rxModelVars(m$simulationModel)$state)
    s1 <- suppressWarnings(rxSolve(m, .ev))
    expect_false("AUC" %in% names(s1))

    # reverse order: solve the original first, then the piped model
    m3 <- rxode2(f)
    expect_false(is.null(m3$simulationModel))
    m4 <- m3 %>% model(d/dt(AUC) <- cp, append = TRUE)
    s3 <- suppressWarnings(rxSolve(m3, .ev))
    expect_false("AUC" %in% names(s3))
    s4 <- suppressWarnings(rxSolve(m4, .ev))
    expect_true("AUC" %in% names(s4))

    # ini() piping must not leak the piped estimates into the original's
    # cached simulation model (theta is stored on the cached TOS)
    m5 <- rxode2(f)
    expect_false(is.null(m5$simulationModel))
    m6 <- m5 %>% ini(ke = 5)
    expect_false(is.null(m6$simulationModel))
    expect_equal(m5$simulationModel$theta[["ke"]], 1)
    expect_equal(m6$simulationModel$theta[["ke"]], 5)
  })

  test_that("forking the meta env on piping preserves metadata semantics", {
    # Only the meta env itself is new; metadata carried over must behave as it
    # did when meta was shared -- nested envs stay shared (reference
    # semantics), closures keep working, and parent chains are intact.
    f <- function() {
      ini({
        ke <- 1
        sd <- 0.1
      })
      model({
        d/dt(CENTRAL) <- -ke * CENTRAL
        cp <- CENTRAL
        cp ~ add(sd)
      })
    }
    m <- rxode2(f)

    .ctx <- new.env(parent = baseenv())
    evalq({
      k <- 1
      getK <- function() k
      expr <- quote(sum(1, 2))
    }, .ctx)
    m$ctx <- .ctx
    m$plainMeta <- "kept"

    m2 <- m %>% model(d/dt(AUC) <- cp, append = TRUE)

    # plain metadata carries over
    expect_equal(m2$ctx$k, 1)
    expect_equal(m2$plainMeta, "kept")
    # closures stored in meta still evaluate
    expect_equal(m2$ctx$getK(), 1)
    # parent chain of a nested env survives (baseenv() -> sum() visible)
    expect_equal(eval(m2$ctx$expr, m2$ctx), 3)
    # nested envs keep reference semantics: still the same env, not a clone
    expect_true(identical(m2$ctx, .ctx))
    m2$ctx$k <- 5
    expect_equal(.ctx$getK(), 5)
  })
})

.rxode2 <- function(...) {
  suppressMessages(rxode2::rxode2(...))
}

# Helper functions for the test cases of the methods.
# general ode methods used
# euler excluded: 1st-order accuracy requires hmin << 0.01 to pass covariate tests
# indLin excluded for now
.methods0 <- c("liblsoda", "lsoda", "dop853", "f78", "rk4", "ck54", "ab", "abm", "dop5", "bs", "ros4", "iem", "sem", "sb3a", "sb3am4", "mm", "em", "cvode", "trapz", "ssp3", "f32", "rk43", "dop54", "vern65", "vern76", "dop87", "vern98", "ros43", "ros6", "backwardEuler", "gauss6", "iiic6", "radauiia5", "geng5", "sdirk43", "euler", "midpoint", "heun", "ssp22", "rk3", "ssp53", "s4", "r4", "ls44", "ls54", "ssp54", "s5", "rk5", "c5", "l5", "lk5a", "lk5b", "b6", "s7", "s8_10", "cv8", "s8_12", "s10", "z10", "o10", "h10", "dp54", "v65e", "v76e", "dp87", "v98e", "ssp33", "bs32", "ssp43", "f45", "t54", "s54", "pp54", "pp54b", "bs54", "ss54", "dp65", "c65", "tp64", "v65r", "v65", "dverk65", "tf65", "tp75", "tmy7", "tmy7s", "v76r", "ss76", "v78", "dverk78", "dp85", "tp86", "v87e", "v87r", "ev87", "k87", "f89", "v89", "t98a", "v98r", "s98", "f108", "c108", "b109", "s1110a", "f1210", "o129", "f1412", "lsode", "bdf", "t54+sdirk43")
# AutoSwitch composite methods: dop853 primary paired with each stiff
# secondary that needs a Jacobian.  1st-order secondaries (iem,
# backwardEuler) are intentionally excluded -- when dop853 falls back to
# them they cannot meet the accuracy tolerances, which defeats the purpose
# of a high-order primary.  This set exercises the non-dense composite
# autoswitch dispatch for every higher-order stiff secondary.
.methods0 <- c("dop853+ros4", "dop853+ros43", "dop853+ros6",
               "dop853+gauss6", "dop853+iiic6",
               "dop853+radauiia5", "dop853+geng5", "dop853+sdirk43")

.methods0 <- c("lsode", "bdf", "liblsoda", "dop853", "t54+sdirk43", "cvode")
.methods1 <- c("ddop853", "ddop5", "dbs", "dros4", "ddop853+dros4", "dcvode")
# All dense composite methods (ros4 is the only dense-output stiff secondary).
.methods1 <- c("ddop853+dros4")
.methods2 <- c("A", "B", "Ao", "Bo", "As", "Bs", "Ad", "Bd", "Al", "Bl")

# ---------------------------------------------------------------------------
# Discrete-adjoint solver variants (method codes 200-304).  On a PLAIN model
# (no rx__sens_* expansion) the explicit/multistep variants run FORWARD-ONLY
# -- so pointing the covariate (test-cov) and dosing (test-nmtest) regression
# suites at them exercises exactly the covariate-interpolation + event/dosing
# handling of every adjoint forward path.  This has historically been a
# regression miss.  The implicit/Rosenbrock/CVODES adjoint variants (ros4s,
# ros43s, ros6s, backwardEulers, gauss6s, iiic6s, radauiia5s, geng5s,
# sdirk43s, cvodesadj) size their LU factorization from the adjoint-augmented
# system and cannot run forward-only, so they are intentionally omitted here
# (their forward IS the well-tested base implicit method; the adjoint sweep is
# validated on the expanded models in test-adjoint-discrete.R).
# .methods1/.methods2 are cleared so nmtest cycles over .methods0 alone.
.methods0 <- c("dop853s", "liblsodaadj", "f78s", "rk4s", "ck54s", "abs",
               "dop5s", "rk43s", "dop54s", "vern65s", "vern76s", "dop87s",
               "vern98s", "eulers", "midpoints", "heuns", "rk3s", "bs32s",
               "f45s", "t54s", "pp54s", "pp54bs", "bs54s", "ss54s", "dp65s",
               "c65s", "tp64s", "v65rs", "dverk65s", "tf65s", "tp75s",
               "tmy7adj", "tmy7sadj", "v76rs", "ss76s", "v78s", "dverk78s",
               "dp85s", "tp86s", "v87es", "v87rs", "ev87s", "k87s", "v89s",
               "t98as", "v98rs", "s98s", "c108s", "b109s", "s1110as", "o129s")
.methods1 <- NULL
.methods2 <- NULL

# test-cov validates the (method-independent) covariate-interpolation logic by
# exact-matching each integrator's RHS-evaluation trace against an approxfun
# reference.  That comparison is tied to the integrator's step schedule, so it
# can only run on the base integrators it was tuned for -- the adjoint variants
# record extra internal steps and would produce spurious knot-boundary
# mismatches (their covariate handling is the same shared code and is verified
# via solution-parity + the nmtest dosing sweep above).
.methodsCov <- c("lsode", "bdf", "liblsoda", "dop853", "t54+sdirk43", "cvode")

#' Determine if expression x is "clean"
#'
#' Clean means that it is free from calculated variables (lhs), and state.
#'
#' The expression needs to be clean of these calculated variables and
#' states to safely extract into a mu expression at the top of the
#' model.
#'
#' @param x The expression that is evaluated; must include an `info`
#'   that has a list of defined states and lhs values
#' @param env Environment for assigning information
#' @return A boolean indicating if the environment is clean from
#'   confounding elements
#' @author Matthew Fidler
#' @examples
#'
#' env <- new.env(parent=emptyenv())
#'
#' env$info <- list(state= c("depot", "center"), lhs=c("ka", "cl", "v", "cp"))
#'
#' .rxMuRefIsClean(quote(exp(tka + eta.ka)), env)
#' .rxMuRefIsClean(quote(exp(tka + eta.ka + depot)), env)
#' .rxMuRefIsClean(quote(exp(tka + eta.ka + cp)), env)
#'
#' @noRd
.rxMuRefIsClean <- function(x, env) {
  if (is.name(x) ) {
    .n <- as.character(x)
    if (any(.n == env$info$state)) {
      return(FALSE)
    } else if (any(.n == env$info$lhs)) {
      return(FALSE)
    }
    return(TRUE)
  } else if (is.call(x)) {
    return(all(unlist(lapply(x[-1], .rxMuRefIsClean, env=env))))
  } else {
    return(TRUE)
  }
}
#' Does this have an eta parameter or covariate in the expression?
#'
#' @inheritParams .rxMuRefIsClean
#' @return boolean saying if the expression has an eta parameter or
#'   covariate
#' @author Matthew Fidler
#' @examples
#'
#' env <- new.env(parent=emptyenv())
#'
#' env$info <- list(theta="tka", eta="eta.ka", cov="wt")
#'
#' .rxMuRefHasThetaEtaOrCov(quote(exp(tka + eta.ka)/v), env)
#' .rxMuRefHasThetaEtaOrCov(quote(cp/v), env)
#'
#' @noRd
.rxMuRefHasThetaEtaOrCov <- function(x, env) {
  if (is.name(x)) {
    .n <- as.character(x)
    .lhs <- deparse1(env$curLhs)
    if (any(.n == env$info$eta)) {
      .w <- which(env$etaLhsDf$lhs == .lhs &
                    env$etaLhsDf$eta == .n)
      if (length(.w) == 0) {
        env$etaLhsDf <- rbind(env$etaLhsDf,
                              data.frame(lhs=.lhs, eta=.n))
      }
      return(TRUE)
    } else if (any(.n == env$info$theta)) {
      .w <- which(env$thetaLhsDf$lhs == .lhs &
                    env$thetaLhsDf$theta == .n)
      if (length(.w) == 0) {
        env$thetaLhsDf <- rbind(env$thetaLhsDf,
                              data.frame(lhs=.lhs, theta=.n))
      }
      return(TRUE)
    } else if (any(.n == env$info$cov)) {
      .w <- which(env$covLhsDf$lhs == .lhs &
                    env$covLhsDf$cov == .n)
      if (length(.w) == 0) {
        env$covLhsDf <- rbind(env$covLhsDf,
                                data.frame(lhs=.lhs, cov=.n))
      }
      return(TRUE)
    } else if (any(.n == env$info$level)) {
      .w <- which(env$levelLhsDf$lhs == .lhs &
                    env$levelLhsDf$level == .n)
      if (length(.w) == 0) {
        env$levelLhsDf <- rbind(env$levelLhsDf,
                              data.frame(lhs=.lhs, level=.n))
      }
      return(TRUE)
    }
    return(FALSE)
  } else if (is.call(x)) {
    return(any(unlist(lapply(x[-1], .rxMuRefHasThetaEtaOrCov, env=env))))
  } else {
    return(FALSE)
  }
}

#' This determines if a whole line is "clean"
#'
#' This not only determines if the lhs of the line is independent of
#' any prior declarations (by `.rxMuRefIsClean()`), but it makes sure
#' that the rhs is not a special rxode2 expression like `d/dt(depot)`
#' or `rate(depot)` `depot(0)` etc.
#'
#' This also assigns the line status in the enviroment `env` to
#' `.curLineClean` and also resets the current evaluation function to
#' `.curEval`
#'
#' @inheritParams .rxMuRefIsClean
#' @return boolean indicating if the line is clean
#' @author Matthew Fidler
#' @noRd
.rxMuRefLineIsClean <- function(x, env) {
  # First figure out if the mu reference line is clean
  .clean <- FALSE
  if (length(x[[2]]) == 1L && is.name(x[[2]])) {
    env$info$lhs <- c(as.character(x[[2]]), env$info$lhs)
    .clean <- TRUE
  }
  if (.clean) .clean <- .rxMuRefIsClean(x[[3]], env)
  assign(".curLineClean", .clean, env)
  assign(".curEval", "", env)
  return(env$.curLineClean)
}

#' Does this line have an eta?
#'
#' @inheritParams .rxMuRefIsClean
#' @return boolean indicating the line has etas
#' @author Matthew Fidler
#' @noRd
.rxMuRefLineHasEta <- function(x, env) {
  if (is.name(x)) {
    .n <- as.character(x)
    if (any(.n == env$info$eta)) {
      return(TRUE)
    }
    return(FALSE)
  } else if (is.call(x)) {
    return(any(unlist(lapply(x[-1], .rxMuRefLineHasEta, env=env))))
  } else {
    return(FALSE)
  }
}

.rxIsLogicalOp <- function(x) {
  (identical(x[[1]], quote(`==`)) ||
     identical(x[[1]], quote(`>`)) ||
     identical(x[[1]], quote(`<`)) ||
     identical(x[[1]], quote(`<=`)) ||
     identical(x[[1]], quote(`>=`)) ||
     identical(x[[1]], quote(`!=`)) ||
     identical(x[[1]], quote(`&&`)) ||
     identical(x[[1]], quote(`||`)) ||
     identical(x[[1]], quote(`|`)) ||
     identical(x[[1]], quote(`&`)))
}
#' Handle the limit for logit/expit types of functions
#'
#' @param x This is the parsed expression tree (from R) that has limits
#' @param env environment where parsing information is saved
#' @return Nothing; Called for its side effects
#' @author Matthew Fidler
#' @noRd
.rxMuRefHandleLimits <- function(x, env) {
  if (length(x) == 4) {
    # expit(x, 1, 2)
    if (is.numeric(x[[3]])) {
      assign("curLow", as.numeric(x[[3]]), envir=env)
    } else {
      assign("err", unique(c(env$err,
                             paste0("syntax error '", deparse1(x),
                                    "': limits must be numeric"))))
      assign("curLow", -Inf, envir=env)
    }
    if (is.numeric(x[[4]])) {
      assign("curHi", as.numeric(x[[4]]), envir=env)
    } else {
      env$err <- unique(c(env$err,
                          paste0("syntax error '", deparse1(x),
                                 "': limits must be numeric")))
      assign("curHi", Inf, envir=env)
    }
    x <- x[1:2]
  } else if (length(x) == 3) {
    # expit(x, 1)
    if (is.numeric(x[[3]])) {
      assign("curLow", as.numeric(x[[3]]), envir=env)
    } else {
      assign("err",
             unique(c(env$err,
                      paste0("syntax error '", deparse1(x),
                             "': limits must be numeric"))),
             envir=env)
      assign("curLow", -Inf, envir=env)
    }
    assign("curHi", 1, envir=env)
    x <- x[1:2]
  } else {
    assign("curLow", 0, envir=env)
    assign("curHi", 1, envir=env)
  }
  if (env$curLow >= env$curHi) {
    assign("err",
           unique(c(env$err,
                    paste0("syntax error '", deparse1(x),
                           "': limits must be lower, higher"))),
           envir=env)
  }
  x
}

.muRefExtractTheta <- function(x, env) {
  if (is.name(x)) {
    .n <- as.character(x)
    .lhs <- deparse1(env$curLhs)
    if (any(.n == env$info$theta)) {
      return(.n)
    }
    return(NULL)
  } else if (is.call(x)) {
    return(do.call(`c`, lapply(x[-1], .muRefExtractTheta, env=env)))
  } else {
    return(NULL)
  }
}

#' Extract single variable names from a expression
#'
#' @param x Expression
#' @param names Names to prepend to the final names output
#' @param env Environment that contains a boolean `.found` that
#'   indicates if an interesting expression has been found
#' @return character vector of names that are uncomplicated, like
#'   "a", "b"; If the names are part of a larger expression these
#'   names are skipped.
#' @author Matthew Fidler
#' @noRd
.muRefExtractSingleVariableNames <- function(x, names, env) {
  c(names, do.call(`c`, lapply(x, function(y) {
    if(is.name(y)) {
      env$found <- TRUE
      .th <- as.character(y)
      if (.th %in% env$info$theta &&
            !(.th %in% env$singleTheta)) {
        env$singleTheta <- c(env$singleTheta, .th)
      }
      return(as.character(y))
    }
    return(NULL)
  })))
}
#' Extract mu-style covariates that is theta + eta + cov*theta.cov
#'
#' @param x expression to evaluate
#'
#' @param doubleNames A list of the covariates with the estimates
#'   attached.  This is in a single expression so wt*theta.cov1 +
#'   wt*theta.cov2 will add an error at the end of the expression
#'
#' @param env Environment with $info that has information about the
#'   parsed model
#'
#' @return A list of covariates with estimates attached
#'
#' @author Matthew Fidler
#'
#' @noRd
.muRefExtractMultiplyMuCovariates <- function(x, doubleNames, env) {
  c(doubleNames, do.call(`c`, lapply(x, function(y) {
    if(is.call(y) && identical(y[[1]], quote(`*`))) {
      .y2 <- y[-1]
      if (length(.y2) == 2) {
        if (is.name(.y2[[1]]) && is.name(.y2[[2]])) {
          .y2 <- vapply(.y2, as.character, character(1))
          if (any(.y2[1] == env$info$cov) &&
                any(.y2[2] == env$info$theta)) {
            if (any(.y2[1] == names(doubleNames))) {
              env$err <- unique(c(env$err,
                                  paste0("syntax error: covariate '", .y2[1],
                                         "' is duplicated in mu-referenced expression for '",
                                         .y2[2], "' and '", doubleNames[[.y2[1]]], "'")))
            }
            env$.found <- TRUE
            return(setNames(list(.y2[2]), .y2[1]))
          } else if (any(.y2[2] == env$info$cov) &&
                       any(.y2[1] == env$info$theta)) {
            if (any(.y2[2] == names(doubleNames))) {
              env$err <- unique(c(env$err,
                                  paste0("syntax error: covariate '", .y2[2],
                                         "' is duplicated in mu-referenced expression for '",
                                         .y2[1], "' and '", doubleNames[[.y2[2]]], "'")))
            }
            env$.found <- TRUE
            return(setNames(list(.y2[1]), .y2[2]))
          }
        }
      }
      .thetas <- try(.muRefExtractTheta(y, env), silent=TRUE)
      if (inherits(.thetas, "try-error")) .thetas <- NULL
      if (length(.thetas) == 1L &&
            grepl("\\b(delay|past)\\s*\\(", deparse1(y), perl = TRUE)) {
        ## A delay()/past() call is an ODE state's history: the call value is
        ## never a theta*covariate, and the whole term cannot be parsed in
        ## isolation (the `rxdummyLhs=<term>` fragment has no d/dt() for that
        ## state, which the C parser reports directly to the console even under
        ## try(silent=TRUE)).  But its DURATION argument may carry a
        ## mu-reference (individual/covariate delay), so look inside the call and
        ## analyse the duration argument(s) instead of the un-parseable term.
        .taus <- list()
        .collectTau <- function(z) {
          if (is.call(z)) {
            if ((identical(z[[1L]], quote(delay)) || identical(z[[1L]], quote(past))) &&
                  length(z) >= 3L) {
              .taus[[length(.taus) + 1L]] <<- z[[3L]]
            } else {
              for (.i in seq_along(z)) .collectTau(z[[.i]])
            }
          }
        }
        .collectTau(y)
        if (length(.taus) == 0L) return(NULL)
        .inner <- .muRefExtractMultiplyMuCovariates(.taus, list(0), env)
        if (length(.inner) > 1L) return(.inner[-1L])
        return(NULL)
      }
      if (length(.thetas) == 1L) {
        .d <- try(symengine::D(get("rxdummyLhs", rxS(paste0("rxdummyLhs=", deparse1(y)))), .thetas), silent=TRUE)
        .extra <- try(str2lang(rxFromSE(.d)), silent=TRUE)
        .thetaD <- try(.muRefExtractTheta(.extra, env), silent=TRUE)
        if (inherits(.thetaD, "try-error")) .thetaD <- NULL
        if (is.null(.thetaD)) {
          # mu2 expression
          env$.found <- TRUE
          env$mu2RefCovariateReplaceDataFrame <-
            rbind(env$mu2RefCovariateReplaceDataFrame,
                  data.frame(covariate=deparse1(.extra),
                             covariateParameter=.thetas,
                             modelExpression=deparse1(y)))
          return(setNames(list(.thetas), deparse1(.extra)))
        }
      }
      return(NULL)
    }
    return(NULL)
  })))
}

.muRefNextAdditiveExpression <- function(x) {
  .expr <- NULL
  for (i in seq_along(length(x))) {
    if(is.call(x[[i]])) {
      .expr <- x[[i]]
      if (identical(.expr[[1]], quote(`+`))) {
        if (length(.expr) == 2L) {
          .expr <- NULL
        } else {
          .expr <- .expr[-1]
        }
      } else {
        .expr <- NULL
      }
    }
  }
  .expr
}
#' Downgrade the mu reference to additive
#'
#' @param .we which eta values
#' @param .wt which theta values
#' @param .names names
#' @param env rxode2 ui environment (incomplete)
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @noRd
.muRefDowngradeEvalToAdditive <- function(.we, .wt, .names, env) {
  .w <- c(.we, .wt)
  if (length(.w) > 0) {
    lapply(.w, function(i) {
      .muRefSetCurEval(.names[i], env, "")
    })
  }
}
#'  Mu reference handle single theta covariates
#'
#' @param .we Which etas
#' @param .wt which thetas
#' @param .names which  names
#' @param .doubleNames These are the double names
#' @param .extraItems Extra items
#' @param env rxode2 incomplete environment
#' @return Nothing called for side effects
#' @author Matthew L. Fidler
#' @noRd
.muRefHandleSingleThetaCovOnly <- function(.we, .wt, .names, .doubleNames, .extraItems, env) {
  .w <- which(env$muRefCovariateDataFrame$theta == .names[.wt])
  .w0 <- which(env$muRefCovariateEmpty == .names[.wt])

  if (length(.doubleNames) == 0) {
    # here mu referenced covariates are not present.
    if (length(.w0) == 0) {
      env$muRefCovariateEmpty <- c(env$muRefCovariateEmpty, .names[.wt])
    }
    if (length(.w) != 0) {
      # Previously, this was thought to have mu referenced
      # variables, could be an error in coding, perhaps a warning
      # should be issued?
      env$muRefDropParameters <- rbind(env$muRefDropParameters,
                                       data.frame(parameter=.names[.wt],
                                                  term=with(env$muRefCovariateDataFrame[.w, ],
                                                            paste0(covariate, "*", covariateParameter))))
      env$muRefCovariateDataFrame <- env$muRefCovariateDataFrame[-.w, ]
      .muRefDowngradeEvalToAdditive(.we, .wt, .names, env)
    }
  } else if (length(.w0) != 0) {
    # Here double names are present, but had been previously flagged
    # as not having covariates.  Perhaps a warning should be
    # issued for this condition
    .covariate <- names(.doubleNames)
    .covariateParameter <- setNames(unlist(.doubleNames), NULL)
    env$muRefDropParameters <- rbind(env$muRefDropParameters,
                                     data.frame(parameter=.names[.wt], term=paste0(.covariate, "*", .covariateParameter)))
    .muRefDowngradeEvalToAdditive(.we, .wt, .names, env)
  } else {
    .covariate <- names(.doubleNames)
    .covariateParameter <- setNames(unlist(.doubleNames), NULL)
    if (length(.w) > 0) {
      # Here there are covariates defined, but perhaps not the same
      # ones as defined in the past.  Check the intersect of what is
      # defined now and what was defined previously
      .covariatePrior <- env$muRefCovariateDataFrame[.w, ]
      .covariateParameterPrior <- .covariatePrior$covariateParameter
      .covariatePrior <- .covariatePrior$covariate
      .multPrior <- paste0(.covariatePrior, "*", .covariateParameterPrior)
      .mult <- paste0(.covariate, "*", .covariateParameter)
      .multBoth <- intersect(.mult, .multPrior)
      if (length(.multBoth) == 0) {
        # Nothing in common, assume that there are no mu referenced covariates from now on
        env$muRefDropParameters <- rbind(env$muRefDropParameters,
                                         data.frame(parameter=.names[.wt],
                                                    term=c(.mult, .multPrior)))

        env$muRefCovariateEmpty <- c(env$muRefCovariateEmpty, .names[.wt])
        env$muRefCovariateDataFrame <- env$muRefCovariateDataFrame[-.w,, drop = FALSE]
        .muRefDowngradeEvalToAdditive(.we, .wt, .names, env)
      } else {
        .w2 <- .w[which(!(.multPrior %in% .multBoth))]
        if (length(.w2) > 0) {
          # Maybe warn that these are dropped
          env$muRefDropParameters <-
            rbind(env$muRefDropParameters,
                  data.frame(parameter=.names[.wt],
                             term=with(env$muRefCovariateDataFrame[.w2, ],
                                       paste0(covariate, "*", covariateParameter))))
          env$muRefCovariateDataFrame <- env$muRefCovariateDataFrame[-.w2,, drop = FALSE]
          .muRefDowngradeEvalToAdditive(.we, .wt, .names, env)
        }
      }
    } else {
      .df <- data.frame(theta=.names[.wt],
                        covariate=.covariate,
                        covariateParameter=.covariateParameter)
      env$muRefCovariateDataFrame <- rbind(env$muRefCovariateDataFrame, .df)
    }
  }
}
#' Handle the extra items that are not part of the mu referencing, but may be added
#'
#' @param .we This is the index of the etas that are being processed
#' @param .wt This is the index of the thetas that are being processed
#' @param .names This is the names of the single items that need to be processed
#' @param .doubleNames This is the double names of the mu referenced covariates
#' @param .extraItems This is the extra items in the mu reference extressions
#' @param env This represents the mu referencing environment where things are stored and modified
#' @param eta When `FALSE` this processes the thetas, when `TRUE` this processes the etas
#' @return Nothing, called for the side effects
#' @author Matthew Fidler
#' @noRd
.muRefHandleSingleThetaExtraOnly <- function(.we, .wt, .names, .doubleNames, .extraItems, env, eta=FALSE)  {
  .wcur <- .wt
  if (eta) .wcur <- .we
  .w <- which(env$muRefExtra$parameter == .names[.wcur])
  .w0 <- which(env$muRefExtraEmpty == .names[.wcur])
  if (length(.extraItems) == 0) {
    # here mu referenced covariates are not present.
    if (length(.w0) == 0) {
      env$muRefExtraEmpty <- c(env$muRefExtraEmpty, .names[.wcur])
    }
    if (length(.w) != 0) {
      # Previously, this was thought to have mu referenced
      # variables, could be an error in coding, perhaps a warning
      # should be issued?
      .extraDrop <- data.frame(parameter=.names[.wcur],
                               term=env$muRefExtra$extra[.w])
      env$muRefDropParameters <- rbind(env$muRefDropParameters,
                                       .extraDrop)
      env$muRefExtra <- env$muRefExtra[-.w, ]
      .muRefDowngradeEvalToAdditive(.we, .wt, .names, env)
    }
  } else if (length(.w0) != 0) {
    # Here extra items are present, but had been previously flagged
    # as not having double names.  Perhaps a warning should be
    # issued for this condition
    env$muRefDropParameters <- rbind(env$muRefDropParameters,
                                     data.frame(parameter=.names[.wcur],
                                                term=.extraItems))
    .muRefDowngradeEvalToAdditive(.we, .wt, .names, env)
  } else {
    if (length(.w) > 0) {
      # Here there are extra items defined, but perhaps not the same
      # ones as defined in the past.  Check the intersect of what is
      # defined now and what was defined previously
      .extraItemsPrior <- env$muRefExtra$extra[.w]
      .extraBoth <- intersect(.extraItems, .extraItemsPrior)
      if (length(.extraBoth) == 0) {
        # Nothing in common, assume that there are no mu referenced covariates from now on
        env$muRefDropParameters <- rbind(env$muRefDropParameters,
                                         data.frame(parameter=.names[.wcur],
                                                    term=c(.extraItems, .extraItemsPrior)))

        env$muRefExtraEmpty <- c(env$muRefExtraEmpty, .names[.wcur])
        env$muRefExtra <- env$muRefExtra[-.w,, drop = FALSE]
        .muRefDowngradeEvalToAdditive(.we, .wt, .names, env)
      } else {
        .w2 <- .w[which(!(.extraItemsPrior %in% .extraBoth))]
        if (length(.w2) > 0) {
          # Maybe warn that these are dropped
          env$muRefDropParameters <- rbind(env$muRefDropParameters,
                                           data.frame(parameter=.names[.wcur],
                                                      term=env$muRefExtra[.w2]))
          env$muRefExtra <- env$muRefExtra[-.w2,, drop = FALSE]
          .muRefDowngradeEvalToAdditive(.we, .wt, .names, env)
        }
      }
    } else {
      .df <- data.frame(parameter=.names[.wcur],
                        extra=.extraItems)
      env$muRefExtra <- rbind(env$muRefExtra, .df)
    }
  }
}

.muRefHandleSingleThetaCovAndExtra <- function(.we, .wt, .names, .doubleNames, .extraItems, env) {
  if (length(.wt) == 1) {
    .muRefSetCurEval(.names[.wt], env)
    .doubleNames <- .doubleNames[names(.doubleNames) != ""]
    .muRefHandleSingleThetaCovOnly(.we, .wt, .names, .doubleNames, .extraItems, env)
    .muRefHandleSingleThetaExtraOnly(.we, .wt, .names, .doubleNames, .extraItems, env)
    .wmu <- which(env$muRefDataFrame$theta == .names[.wt] &
                    env$muRefDataFrame$level == "id")
    if (length(.wmu) == 1 && length(.we) == 0L) {
      .curEta <- env$muRefDataFrame$eta[.wmu]
      .muRefSetNonMuEta(.curEta, env)
      .muRefSetCurEval(.curEta, env, set="")
    }
  }
}

#' This sets the non mu etas
#'
#' @param .curEta The eta to set as a non-mu eta
#' @param env mu ref environtment
#' @return Nothing, called for its side effects
#' @author Matthew Fidler
#' @noRd
.muRefSetNonMuEta <- function(.curEta, env) {
  if (!is.null(env$levels) && any(env$levels == .curEta)) return(invisible())
  if (!any(env$nonMuEtas == .curEta)) {
    env$nonMuEtas <- c(env$nonMuEtas, .curEta)
    .wEtaInDf <- which(env$muRefDataFrame$eta == .curEta)
    if (length(.wEtaInDf) > 0) {
      env$muRefDataFrame <- env$muRefDataFrame[-.wEtaInDf,, drop = FALSE]
    }
  }
}
# This function handles the extra information in a theta based mu referenced

#' This handles the case where there is a mu referenced single theta
#' with a reference population eta
#'
#' @param .we this represents the item number in `.names` that is the
#'   mu-referenced population eta
#'
#' @param .wt this represents the item number in `.names` that is the
#'   mu-referenced population theta
#'
#' @param .names This is the names of the single item names in the
#'   additive expression that may be a mu-referenced expression
#'
#' @param .doubleNames The double names of the covariate and
#'   estimated theta for the mu referenced covariate.
#'
#' The double names is a list. The elements of this list are the
#' population parameter estimates for a data-based covariate. The
#' names of the list are the covariate that the population parameters
#' are estimating an effect for.
#'
#' @param .extraItems This represents the extra items that are in the
#'   additive expression that may be part of the full additive
#'   expression
#'
#' @param env This is the mu-reference environment where mu reference
#'   information is saved
#'
#' @return Nothing; Called for its side effects and modifications in
#'   the mu reference environment `env`
#'
#' @author Matthew Fidler
#' @noRd
.muRefHandleSingleThetaMuRef <- function(.we, .wt, .names, .doubleNames, .extraItems, env) {
  # Here the mu reference is possible
  if (length(.we) == 1) {
    # Simple theta/eta mu-referencing
    .curEta <- .names[.we]
    .muRefSetCurEval(.curEta, env)
    if (!any(.curEta == env$nonMuEtas)) {
      .wEtaInDf <- which(env$muRefDataFrame$eta == .curEta)
      if (length(.wEtaInDf) > 0) {
        # duplicated ETAs, if everything is not the same then it isn't really mu-referenced

        # The duplicated ETAs can occur for shared etas for 2
        # different population parameters.  This can make sense for 2
        # different treatments with a similar between subject
        # variability that can be pooled, like:
        #
        # emaxA = tv.emaxA + eta.emax
        # emaxB = tv.emaxB + eta.emax
        #
        # In this case, the eta.emax is no longer mu-referenced
        #
        if (!all(env$muRefDataFrame$theta[.wEtaInDf] == .names[.wt]) ||
              !all(env$muRefDataFrame$eta[.wEtaInDf] == .curEta)) {
          .muRefSetNonMuEta(.curEta, env)
        }
      } else {
        env$muRefDataFrame <-
          rbind(env$muRefDataFrame,
                data.frame(theta=.names[.wt],
                           eta=.names[.we],
                           level="id"))
      }
    }
  } else if (length(.we) != 0) {
    # Mu reference can only occur on id-referenced items currently
    stop("currently do not theta + eta1 + eta2")
  }
  .muRefHandleSingleThetaCovAndExtra(.we, .wt, .names, .doubleNames, .extraItems, env)
}

#' Handle unary plus as a noop. Also recursively handle calls that are not additive expressions
#'
#' @param expr parameter expression
#' @param env environment for assigning items
#' @return return plus
#' @author Matthew L. Fidler
#' @noRd
.muRefHandlePlusNoop <- function(expr, env) {
  if (is.symbol(expr)) return(expr)
  if (identical(expr[[1]], quote(`+`)) &&
        length(expr) == 2L) {
    return(expr[[2]])
  } else if (is.call(expr) && !identical(expr[[1]], quote(`+`))) {
    # save curEval
    .curEval <- env$.curEval
    .rxMuRefHandleNonPlusCall(expr, env)
    assign(".curEval", .curEval, envir=env) # restore curEval
  }
  expr
}


#' Handle the + expressions to determine mu-reference expressions
#'
#' @param x additive Call Expression
#' @param env Environment information
#' @return nothing
#' @author Matthew Fidler
#' @noRd
.muRefHandlePlus <- function(x, env) {
  .x2 <- x[-1]
  .names <- NULL
  .doubleNames <- list(0)
  .extraItems <- NULL
  while (!is.null(.x2)) {
    env$.found <- FALSE
    .x2[[1]] <- .muRefHandlePlusNoop(.x2[[1]], env)
    .names <- .muRefExtractSingleVariableNames(.x2, .names, env)
    .doubleNames <- .muRefExtractMultiplyMuCovariates(.x2, .doubleNames, env)
    if (length(.x2) >= 2) {
      if (!env$.found && !is.name(.x2[[2]])) {
        .extraItems <- c(.extraItems, deparse1(.x2[[2]]))
      }
    }
    .x2 <- .muRefNextAdditiveExpression(.x2)
  }
  .wt <- which(.names %in% env$info$theta)
  .we <- which(.names %in% env$info$eta)
  .wl <- if (!is.null(env$info$level)) which(.names %in% env$info$level) else integer(0)
  if (length(.wt) >= 2 && length(.we) >= 1) {
    # Only a genuine mu-referenced expression (ie one that also has an
    # eta) with 2+ population parameters is ambiguous.  Summing
    # population parameters with no random effect (eg a combined
    # residual error 'sqrt(sigma.1. + sigma.2.)') is legitimate and must
    # not be flagged.
    env$err <- unique(c(env$err,
                        paste0("syntax error: 2+ single population parameters in a single mu-referenced expression: '",
                               paste(.names[.wt], collapse="', '"),
                               "'\nthis could occur when a between subject variability parameter is not initialized with a '~'")))
  } else if (length(.wt) == 1) {
    if (!is.null(.extraItems)) {
      .ord <- order(vapply(.extraItems, nchar, integer(1)))
      .extraItems <- .extraItems[.ord]
    }
    .muRefHandleSingleThetaMuRef(.we, .wt, .names, .doubleNames, .extraItems, env)
  } else if (length(.we) == 1) {
    .curEta <- .names[.we]
    .muRefSetCurEval(.curEta, env)
    .muRefSetNonMuEta(.curEta, env)
    .muRefHandleSingleThetaExtraOnly(.we, .wt, .names, .doubleNames, .extraItems, env, eta=TRUE)
  }
  if (length(.wl) > 0) {
    .iovs <- .names[.wl]
    lapply(.iovs, function(iov) {
      .muRefSetCurEval(iov, env)
    })
  }
  invisible()
}
#'  Handle single eta value if it exists
#'
#' @param expr r language expression
#' @param env environment for saving information
#' @return Called for side effects
#' @author Matthew L. Fidler
#' @noRd
.handleSingleEtaIfExists <- function(expr, env) {
  .curEta <- deparse1(expr)
  if (any(.curEta == env$info$eta)) {
    ## eta singlet, not mu referenced
    .muRefSetCurEval(.curEta, env)
    .muRefSetNonMuEta(.curEta, env)
  } else if (any(.curEta == env$info$theta)) {
    .muRefSetCurEval(.curEta, env)
  }
}
#' Handle non plus calls
#'
#' @param x expression
#' @param env Environment for saving information
#' @author Matthew L. Fidler
#' @noRd
.rxMuRefHandleNonPlusCall <- function(x, env) {
  .curEval <- as.character(x[[1]])
  assign(".curEval", .curEval, env)
  env$curHi <- NA_real_
  env$curLow <- NA_real_
  if (env$.curEval == "probitInv" ||
        env$.curEval == "expit" ||
        env$.curEval == "logit" ||
        env$.curEval == "probit") {
    x <- .rxMuRefHandleLimits(x, env)
  }
  .handleSingleEtaIfExists(x[[2]], env)
  lapply(x[-1], .rxMuRef0, env=env)
}

.rxMuRef0 <- function(x, env) {
  if (env$top) {
    env$top <- FALSE
    y <- x
    for (.i in seq_along(y)) {
      x <- y[[.i]]
      if (identical(x[[1]], quote(`=`)) ||
            identical(x[[1]], quote(`~`))) {
        env$curLhs <- x[[2]]
        #.handleSingleEtaIfExists(x[[3]], env)
        if (.rxMuRefHasThetaEtaOrCov(x[[3]], env)) {
          # This line has etas or covariates and might need to be
          # separated into mu-referenced line
          .rxMuRefLineIsClean(x, env)
          if (getOption("rxode2.debug", FALSE)) {
            .tmp <- lapply(x, function(y, env) {
              assign(".curEval", "", env)
              .rxMuRef0(y, env=env)
            }, env=env)
          } else {
            .tmp <- try(lapply(x, function(y, env) {
              assign(".curEval", "", env)
              .rxMuRef0(y, env=env)
            }, env=env), silent=TRUE)
          }
          if (inherits(.tmp, "try-error")) {
            .msg <- paste0("mu-ref err: ", attr(.tmp,"condition")$message)
            if (!is.null(env$lstErr[[.i]])) {
              .msg <- paste(env$lstErr[[.i]], "\n", .msg)
            }
            env$lstErr[[.i]] <- .msg
            env$hasErrors <- TRUE
          }
        } else {
          # This line does not depend on etas or covariates
          # simply add to the body
          env$body <- c(env$body, list(x))
        }
      } else {
        ## This line is a special statement, simply add to the body
        env$body <- c(env$body, list(x))
      }
    }
  } else if (is.call(x)) {
    if (identical(x[[1]], quote(`+`))) {
      .muRefHandlePlus(x, env)
    } else {
      .rxMuRefHandleNonPlusCall(x, env)
    }
  } else if (is.name(x)) {
    .name <- as.character(x)
    if (any(.name == env$info$theta)) {
      env$oneTheta <- unique(.name, env$oneTheta)
    }
  }
}

#' This sets the current evaluation for a single parameter, either a theta or an eta
#'
#' @param parameter The parameter is the string for theta or eta
#'   value
#' @param env Mu referencing parsing environment
#' @param set if a character, set the evalutaion to this value
#'   regardless of the internal value of `env$.curEval`
#' @return Nothing, called for its side effects
#' @author Matthew Fidler
#' @noRd
.muRefSetCurEval <- function(parameter, env, set=NULL) {
  .w <- which(env$muRefCurEval == parameter)
  .curEval <- env$.curEval
  .blankEval <- ""
  if (inherits(set, "character")) {
    if (length(.w) == 0L) {
      env$muRefCurEval <-
        rbind(env$muRefCurEval,
              data.frame(parameter=parameter, curEval=set, low=env$curLow, hi=env$curHi))
    } else {
      env$muRefCurEval$curEval[.w] <- set
    }
  } else {
    if (length(.w) == 0L) {
      if (is.null(env$curLow)) {
        env$curLow <- NA_real_
      }
      if (is.null(env$curHi)) {
        env$curHi <- NA_real_
      }
      env$muRefCurEval <-
        rbind(env$muRefCurEval,
              data.frame(parameter=parameter, curEval=.curEval, low=env$curLow, hi=env$curHi))
    } else if (env$muRefCurEval$curEval[.w] != env$.curEval) {
      env$muRefCurEval$curEval[.w] <- .blankEval
    }
  }
}

#'  Setup the initial environment for mu-referencing calculations
#'
#' @param mod Model environment
#' @param ini parsed lotri ini
#' @return model environment to setup rxUi
#' @author Matthew L. Fidler
#' @noRd
.rxMuRefSetupInitialEnvironment <- function(mod, ini=NULL) {
  if (is.null(ini)) {
    .eta <- mod$eta
    .iniDf <- mod$iniDf
    .mv <- mod$mv0
    .env <- mod
    .level <- mod$level
  } else {
    .level <- NULL
    if (is.list(ini) && any(names(ini) == "id")) {
      .eta <- dimnames(ini$id)[[1]]
      for (v in names(ini)) {
        if (v != "id") {
          .level <- c(.level, dimnames(ini[[v]])[[1]])
        }
      }
    } else {
      .eta <- dimnames(ini)[[1]]
    }
    .iniDf <- as.data.frame(ini)
    .mv  <- rxModelVars(mod)
    .env <- new.env(parent=emptyenv())
  }
  .theta <- .iniDf$name[!is.na(.iniDf$ntheta)]
  .expr <- eval(parse(text=paste0("quote({",rxNorm(.mv),"})")))
  .expr <- .expr[-1]
  .state <- .mv$state

  .params <- .mv$params
  .lhs <- .mv$lhs

  # Covariates are model based parameters not described by theta/eta
  .info <- list(state=.state,
                lhs=NULL,
                theta=.theta,
                eta=.eta,
                level=.level,
                cov=setdiff(.params, c(.theta, .eta, .level, names(rxInits(.mv)))))
  .env$param <- list()
  .env$singleTheta <- NULL
  .env$body <- list()
  .env$info <- .info
  .env$top <- TRUE
  if (!exists("hasErrors", envir=.env)) {
    .env$hasErrors <- FALSE
  }

  # probit/probitInv
  .env$probit.theta.low <- NULL
  .env$probit.theta.hi <- NULL
  .env$probit.theta <- NULL

  .env$probitInv.theta.low <- NULL
  .env$probitInv.theta.hi <- NULL
  .env$probitInv.theta <- NULL

  # logit/expit
  .env$logit.theta <- NULL
  .env$logit.theta.low <- NULL
  .env$logit.theta.hi <- NULL

  .env$expit.theta <- NULL
  .env$expit.theta.low <- NULL
  .env$expit.theta.hi <- NULL

  .env$log.theta <- NULL
  .env$exp.theta <- NULL

  .env$cov.ref <- NULL
  .env$err <- NULL
  .env$.expr <- .expr
  # This records the evaluation status of single parameters like thetas and etass
  .env$muRefCurEval <- data.frame(parameter=character(0), curEval=character(0))
  # This records the mu reference relationship between eta
  # variability and the population parameters
  .env$muRefDataFrame <- data.frame(eta=character(0), theta=character(0), level=character(0))
  .env$muRefExtra <- data.frame(parameter=character(0), extra=character(0))
  .env$muRefExtraEmpty <- NULL
  .env$muRefCovariateDataFrame <- data.frame(theta=character(0), covariate=character(0), covariateParameter=character(0))
  .env$mu2RefCovariateReplaceDataFrame <- data.frame(covariate=character(0), covariateParameter=character(0), modelExpression=character(0))
  .env$muRefDropParameters <- data.frame(parameter=character(0), term=character(0))
  .env$muRefCovariateEmpty <- NULL
  .env$nonMuEtas <- NULL
  .env$covariates <- .info$cov
  .env$etaLhsDf <- data.frame(lhs=character(0), eta=character(0))
  .env$thetaLhsDf <- data.frame(lhs=character(0), theta=character(0))
  .env$covLhsDf <- data.frame(lhs=character(0), cov=character(0))
  .env$levelLhsDf <- data.frame(lhs=character(0), level=character(0))
  .env$curLhs <- NULL
  return(.env)
}
#' Check/adjust ranges for error parameters
#'
#' This will throw an error of an error parameter is defined out of
#' range. It also will update `iniDf` to match the parameter description
#'
#' @param env rxode ui environment (likely before fianlization)
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @noRd
.checkAndAdjustErrInformation <- function(env) {
  .iniDf <- env$iniDf
  .w <- which(!is.na(.iniDf$err))
  for (.err in .w) {
    .curErr <- .iniDf$err[.err]
    .wN <- which(names(.errDistArgRanges) == .curErr)
    if (length(.wN) == 1L) {
      .range <- .errDistArgRanges[[.curErr]]
      .est <- .iniDf$est[.err]
      .lower <- .iniDf$lower[.err]
      .name <- .iniDf$name[.err]
      .upper <- .iniDf$upper[.err]
      if (.range[1] > .est) {
        env$err <- c(env$err,
                     paste0("'", .name, "' estimate (",
                            .est, ") needs to be above ", .range[1]))
      }
      if (.range[2] < .est) {
        env$err <- c(env$err,
                     paste0("'", .name, "' estimate (",
                            .est, ") needs to be below ", .range[2]))
      }
      if (.lower < .range[1]) {
        if (isTRUE(getOption("rxode2.verbose.pipe", TRUE)) && is.finite(.lower)) {
          .minfo(paste0("'", .name, "' lower bound (",
                        .lower, ") needs to be equal or above ", .range[1],
                        "; adjusting"))
        }
        .lower <- .range[1]
      }
      if (.upper > .range[2]) {
        if (isTRUE(getOption("rxode2.verbose.pipe", TRUE)) && is.finite(.upper)) {
          .minfo(paste0("'", .name, "' upper bound (", .upper,
                        ") needs to be equal or below ", .range[2],
                        "; adjusting"))
        }
        .upper <- .range[2]
      }
      .iniDf$lower[.err] <- .lower
      .iniDf$upper[.err] <- .upper
    }
  }
  env$iniDf <- .iniDf
  invisible()
}
#' Check for ini({}) parameters that are missing in the model({})
#'
#'
#' @param ui rxode2 ui model
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @noRd
.checkForIniParametersMissingFromModelBlock <- function(ui) {
  .iniDf <- ui$iniDf
  .mv <- ui$mv0
  if (is.null(ui$predDf)) {
    .errEsts <- NULL
  } else {
    .errEsts <- .iniDf[.iniDf$condition %in% ui$predDf$cond, "name"]
  }
  .estName <- .iniDf$name[!is.na(.iniDf$ntheta) |
                            (!is.na(.iniDf$neta1) & .iniDf$neta1 == .iniDf$neta2)]
  .estName <- c(.estName, .errEsts)
  .missingPars <- setdiff(.estName, c(.mv$params, .errEsts))
  if (length(.missingPars) > 0) {
    ui$err <-
      c(ui$err,
        paste0("the following parameter(s) were in the ini block but not in the model block: ",
               paste(.missingPars, collapse=", ")))
  }
  invisible()
}

#' Check for infinite or missing parameters
#'
#' @param ui rxode2 ui function
#' @return Nothing called for side effects
#' @author Matthew L. Fidler
#' @noRd
.checkForInfiniteOrNaParameters <- function(ui) {
  .iniDf <- ui$iniDf
  .bad <- .iniDf$name[is.infinite(.iniDf$est) | is.na(.iniDf$est)]
  if (length(.bad) > 0) {
    ui$err <-
      c(ui$err,
        paste0("infinite/NA initial parameters: ", paste(.bad, collapse=", ")))
  }
  invisible()
}

.checkForAtLeastOneEstimatedOrModeledParameterPerEndpoint <- function(ui) {
  .iniDf <- ui$iniDf
  .predDf <- ui$predDf
  if (is.null(.predDf)) return(NULL)
  .mv <- ui$mv0
  lapply(seq_along(.predDf$cond), function(i) {
    .cond <- .predDf$cond[i]
    .w <- which(.iniDf$condition == .cond)
    # endpoint has no estimated parameters, see if they are modeled
    # parameters
    .ret <- as.character(.predDf[i, c("a", "b", "c", "d", "e", "f", "lambda")])
    .ret <- .ret[!is.na(.ret)]
    .ini <- .mv$ini
    .ini <- .ini[!is.na(.ini)]
    .names <- c(.mv$lhs, names(.ini))
    .ret <- setdiff(.ret, .names)
    if (length(.ret) > 0L) {
      ui$err <- c(ui$err,
                  paste0("endpoint '", .userEndpointNames(.predDf$cond[i]), "' needs the following parameters estimated or modeled: ",
                         paste(.ret, collapse=", ")))
    }
    if (.predDf$distribution[i] %in% c("norm", "t") && !(.predDf$var[i] %in% c(.mv$lhs, .mv$state, "rxLinCmt"))) {
      ui$err <- c(ui$err,
                  paste0("endpoint '", .userEndpointNames(.predDf$cond[i]), "' is not defined in the model"))
    }
  })
}


.muRefSeparateCalculatedMuRefCovs <- function(.env) {
  if (length(.env$mu2RefCovariateReplaceDataFrame$covariate) == 0L) {
    .env$mu2RefCovariateReplaceDataFrame$theta <- character(0)
  } else {
    .env$mu2RefCovariateReplaceDataFrame$theta <- NA_character_
    lapply(seq_along(.env$mu2RefCovariateReplaceDataFrame$covariate),
           function(i) {
             .cov <- .env$mu2RefCovariateReplaceDataFrame$covariate[i]
             .covPar <- .env$mu2RefCovariateReplaceDataFrame$covariateParameter[i]
             .wmu <- which(.env$muRefCovariateDataFrame$covariate == .cov &
                             .env$mu2RefCovariateReplaceDataFrame$covariateParameter[i] == .covPar)
             if (length(.wmu) == 1L) {
               .theta <- .env$muRefCovariateDataFrame$theta[.wmu]
               .env$muRefCovariateDataFrame <- .env$muRefCovariateDataFrame[-.wmu,]
               .env$mu2RefCovariateReplaceDataFrame$theta[i] <- .theta
             }
           })
  }
  .env$mu2RefCovariateReplaceDataFrame <- .env$mu2RefCovariateReplaceDataFrame[,c("theta", "covariate", "covariateParameter", "modelExpression")]
  row.names(.env$muRefCovariateDataFrame) <- NULL
  row.names(.env$mu2RefCovariateReplaceDataFrame) <- NULL
}

#' Get mu-referencing model from model variables
#'
#' The rxMuRef is the core of the nlmixr ui functions
#'
#' This function takes the initialization values from `lotri()` the
#' parsed rxode2 model to generate mu-referenced models adding
#' mu-references for etas that do not have them to allow saem to
#' support non mu-referenced models by a parsing trick.
#'
#' @param mod Model
#' @param ini lotri initialization information
#' @return parsed environment
#' @author Matthew Fidler
#' @examples
#'
#' # First lets get a lotri initialization block:
#'
#' ini <- lotri({
#'   ## You may label each parameter with a comment
#'   tka <- 0.45 # Log Ka
#'   tcl <- log(c(0, 2.7, 100)) # Log Cl
#'   ## This works with interactive models
#'   ## You may also label the preceding line with label("label text")
#'   tv <- 3.45; label("log V")
#'   ## the label("Label name") works with all models
#'   eta.ka ~ 0.6
#'   eta.cl + eta.v ~ c(0.3,
#'                      0.001, 0.1)
#'   add.sd <- 0.7
#' })
#'
#'
#' ini2 <- lotri({
#'   ## You may label each parameter with a comment
#'   tka <- 0.45 # Log Ka
#'   tcl <- log(c(0, 2.7, 100)) # Log Cl
#'   ## This works with interactive models
#'   ## You may also label the preceding line with label("label text")
#'   tv <- 3.45; label("log V")
#'   add.sd <- 0.7
#' })
#'
#' @noRd
.rxMuRef <- function(mod, ini=NULL) {
  .env <- .rxMuRefSetupInitialEnvironment(mod, ini)
  .rxMuRef0(.env$.expr, env=.env)
  .checkAndAdjustErrInformation(.env)
  .checkForIniParametersMissingFromModelBlock(.env)
  .checkForInfiniteOrNaParameters(.env)
  .checkForAtLeastOneEstimatedOrModeledParameterPerEndpoint(.env)
  .muRefDowngrade(.env)
  .muRefSeparateCalculatedMuRefCovs(.env)
  .handleErrs(.env)
  .rm <- intersect(c(".curEval", ".curLineClean", ".expr", ".found", "body", "cov.ref",
                     "err", "exp.theta", "expit.theta", "expit.theta.hi", "expit.theta.low",
                     "found", "info", "log.theta", "logit.theta", "logit.theta.hi",
                     "logit.theta.low", "param", "probit.theta", "probit.theta.hi",
                     "probit.theta.low", "probitInv.theta", "probitInv.theta.hi",
                     "probitInv.theta.low", "top", "dupErr", "lstErr", "lstChr", "curLhs"),
                   ls(envir=.env, all.names=TRUE))
  if (length(.rm) > 0) rm(list=.rm, envir=.env)
  return(invisible(.env))
}

#devtools::install("~/src/rxode2")

library(nlmixr2)

## one.cmt <- function() {
##   ini({
##     ## You may label each parameter with a comment
##     tka <- 0.45 # Log Ka
##     tcl <- log(c(0, 2.7, 100)) # Log Cl
##     ## This works with interactive models
##     ## You may also label the preceding line with label("label text")
##     tv <- 3.45; label("log V")
##     ## the label("Label name") works with all models
##     eta.ka ~ 0.6
##     eta.cl ~ 0.3
##     eta.v ~ 0.1
##     add.sd <- 0.7
##   })
##   model({
##     ka <- exp(tka + eta.ka)
##     cl <- exp(tcl + eta.cl)
##     v <- exp(tv + eta.v)
##     linCmt() ~ add(add.sd)
##   })
## }

## fit.c3 <- nlmixr(one.cmt, theo_sd, est="focei",
##                  control=foceiControl(rxControl=rxControl(linCmtSensType="forward3")))

## fit <- nlmixr(one.cmt, theo_sd, est="focei")

## fit.ad <- nlmixr(one.cmt, theo_sd, est="focei",
##  control=foceiControl(rxControl=rxControl(linCmtSensType="AD")))

## fit.c5 <- nlmixr(one.cmt, theo_sd, est="focei",
##   control=foceiControl(rxControl=rxControl(linCmtSensType="endpoint5")))


## f <- function(){
##   ini({ # Where initial conditions/variables are specified
##     # '<-' or '=' defines population parameters
##     # Simple numeric expressions are supported
##     lCl <- 1.8      #log Cl (L/hr)
##     lVc <- 4.7      #log V (L)
##     lKA <- 0.2      #log V (L)
##     # Bounds may be specified by c(lower, est, upper), like NONMEM:
##     # Residuals errors are assumed to be population parameters
##     prop.err <- c(0, 0.3, 1)
##     # Between subject variability estimates are specified by '~'
##     # Semicolons are optional
##     eta.Cl ~ 0.15
##     eta.Vc ~ 0.15
##     eta.KA ~ 0.15
##   })
##   model({ # Where the model is specified
##     # The model uses the ini-defined variable names
##     Cl <- exp(lCl + eta.Cl)
##     Vc <- exp(lVc + eta.Vc)
##     KA <- exp(lKA + eta.KA)
##     cp <- linCmt()
##     cp ~ prop(prop.err)
##   })
## }

## datr <- nlmixr2data::Oral_1CPT
## dat <- datr[datr$SD == 1, ]
## dat <- dat[, names(dat) != "SS"]

## fit1 <- nlmixr2(f, dat, est="focei")

## fit1.c3 <- nlmixr2(f, dat, est="focei",
##                    control=foceiControl(rxControl=rxControl(linCmtSensType="forward3")))

## fit1.c5 <- nlmixr2(f, dat, est="focei",
##                    control=foceiControl(rxControl=rxControl(linCmtSensType="endpoint5")))

## fit1.ad <- nlmixr2(f, dat, est="focei",  control=foceiControl(rxControl=rxControl(linCmtSensType="AD")))

## c(fit1=sum(fit1$time),
##   fit1.c3=sum(fit1.c3$time),
##   fit1.c5=sum(fit1.c5$time),
##   fit1.ad=sum(fit1.ad$time))


f <- function() {
  ini({ # Where initial conditions/variables are specified
    # '<-' or '=' defines population parameters
    lCl <- 1.6      #log Cl (L/hr)
    lVc <- 4.5      #log Vc (L)
    lQ  <- 1.6      #log Q (L/hr)
    lVp <- 4        #log Vp (L)
    lKA <- 0.2      #log V (L)
    # Bounds may be specified by c(lower, est, upper), like NONMEM:
    # Residuals errors are assumed to be population parameters
    prop.err <- c(0, 0.3, 1)
    # Between subject variability estimates are specified by '~'
    # Semicolons are optional
    eta.Vc ~ 0.15
    eta.Cl ~ 0.15
    eta.Vp ~ 0.15
    eta.Q  ~ 0.15
    eta.KA ~ 0.15
  })
  model({ # Where the model is specified
    # The model uses the ini-defined variable names
    Vc <- exp(lVc + eta.Vc)
    Cl <- exp(lCl + eta.Cl)
    Vp <- exp(lVp + eta.Vp)
    Q  <- exp(lQ + eta.Q)
    KA <- exp(lKA + eta.KA)
    # And is assumed to follow proportional error estimated by prop.err
    cp <- linCmt()
    cp ~ prop(prop.err)
  })
}

datr <- nlmixr2data::Oral_2CPT
dat <- datr[datr$SD == 1, ]
dat <- dat[, names(dat) != "SS"]

## fit2 <- nlmixr2(f, dat, est="focei")

## fit2.ad <- nlmixr2(f, dat, est="focei",
##  control=foceiControl(rxControl=rxControl(linCmtSensType="AD")))

## fit2.c3 <- nlmixr2(f, dat, est="focei",
##                    control=foceiControl(rxControl=rxControl(linCmtSensType="forward3")))

## fit2.c <- nlmixr2(f, dat, est="focei",
##                   control=foceiControl(rxControl=rxControl(linCmtSensType="central")))

## fit2.ch <- nlmixr2(f, dat, est="focei",
##                   control=foceiControl(rxControl=rxControl(linCmtSensType="centralH")))

fit2.f <- nlmixr2(f, dat, est="focei",
                  control=foceiControl(rxControl=rxControl(linCmtSensType="forward"),
                                       outerOpt="lbfgsb3c"))

fit2.f3 <- nlmixr2(f, dat, est="focei",
                   control=foceiControl(rxControl=rxControl(linCmtSensType="forward3"),
                                        outerOpt="nlminb"))


fit2.fg <- nlmixr2(f, dat, est="focei",                  control=foceiControl(rxControl=rxControl(linCmtSensType="forwardG"), outerOpt="lbfgsb3c"))

fit2n.f <- nlmixr2(f, dat, est="focei",
                  control=foceiControl(rxControl=rxControl(linCmtSensType="forward"),
                                       outerOpt="nlminb"))

fit2n.fg <- nlmixr2(f, dat, est="focei",                  control=foceiControl(rxControl=rxControl(linCmtSensType="forwardG"), outerOpt="nlminb"))

## fit2.c <- nlmixr2(f, dat, est="focei",                  control=foceiControl(rxControl=rxControl(linCmtSensType="central")))


#fit2 <- nlmixr2(f, dat, est="focei")


## fit2.c5 <- nlmixr2(f, dat, est="focei",
##  control=foceiControl(rxControl=rxControl(linCmtSensType="endpoint5")))

## c(fit2=sum(fit2$time),
##   fit2.c3=sum(fit2.c3$time),
##   fit2.c5=sum(fit2.c5$time),
##   fit2.ad=sum(fit2.ad$time))

## ## fit <- nlmixr2(f, dat, est="focei")


## f <- function() {
##   ini({ # Where initial conditions/variables are specified
##     # '<-' or '=' defines population parameters
##     lCl <- 1.6      #log Cl (L/hr)
##     lVc <- 4.5      #log Vc (L)
##     lQ  <- 1.6      #log Q (L/hr)
##     lVp <- 4        #log Vp (L)
##     lQ2 <- 2
##     lVp2 <- 6
##     lKA <- 0.2      #log V (L)
##     # Bounds may be specified by c(lower, est, upper), like NONMEM:
##     # Residuals errors are assumed to be population parameters
##     prop.err <- c(0, 0.3, 1)
##     # Between subject variability estimates are specified by '~'
##     # Semicolons are optional
##     eta.Vc ~ 0.15
##     eta.Cl ~ 0.15
##     eta.KA ~ 0.15
##   })
##   model({ # Where the model is specified
##     # The model uses the ini-defined variable names
##     Vc <- exp(lVc + eta.Vc)
##     Cl <- exp(lCl + eta.Cl)
##     Vp <- exp(lVp)
##     Q  <- exp(lQ)
##     KA <- exp(lKA + eta.KA)
##     Q2 <- exp(lQ2)
##     Vp2 <- exp(lVp2)
##     # And is assumed to follow proportional error estimated by prop.err
##     cp <- linCmt()
##     cp ~ prop(prop.err)
##   })
## }


## fit3 <- nlmixr2(f, dat, est="focei")

## fit3.ad <- nlmixr2(f, dat, est="focei",
##  control=foceiControl(rxControl=rxControl(linCmtSensType="AD")))

## fit3.c3 <- nlmixr2(f, dat, est="focei",
##  control=foceiControl(rxControl=rxControl(linCmtSensType="forward3")))

## fit3.c5 <- nlmixr2(f, dat, est="focei",
##  control=foceiControl(rxControl=rxControl(linCmtSensType="endpoint5")))


## c(fit1=sum(fit1$time),
##   fit1.c3=sum(fit1.c3$time),
##   fit1.c5=sum(fit1.c5$time),
##   fit1.ad=sum(fit1.ad$time))

## c(fit2=sum(fit2$time),
##   fit2.c3=sum(fit2.c3$time),
##   fit2.c5=sum(fit2.c5$time),
##   fit2.ad=sum(fit2.ad$time))

## c(fit3=sum(fit3$time),
##   fit3.c3=sum(fit3.c3$time),
##   fit3.c5=sum(fit3.c5$time),
##   fit3.ad=sum(fit3.ad$time))

## Delay differential equation (DDE) helpers for forward sensitivities: catalog
## the delay(state, T) terms and splice the variational delayed term
## d f_i / d[delay(y_j, T)] * delay(S_j, T) into each sensitivity ODE.

#' Top-level statements of a model's normalized text
#'
#' `rxNorm()` emits `if`/`else` with `}` and `else` on separate top-level
#' lines, which is only valid R inside a `{ }` block; wrap before parsing
#' (#1151).
#'
#' @param model anything `rxNorm()` accepts.
#' @return list of top-level statements (language objects).
#' @noRd
.rxNormStatements <- function(model) {
  .e <- parse(text = paste0("{\n", paste(rxNorm(model), collapse = "\n"), "\n}"))[[1L]]
  as.list(.e)[-1L]
}

#' Is a statement a `past(state, tau) <- expr` assignment?
#'
#' @param x a language object.
#' @return logical.
#' @noRd
.rxIsPastAssign <- function(x) {
  is.call(x) && (identical(x[[1L]], quote(`=`)) ||
                   identical(x[[1L]], quote(`<-`))) &&
    is.call(x[[2L]]) && identical(x[[2L]][[1L]], quote(past)) &&
    length(x[[2L]]) == 3L
}

#' Catalog the delay(state, T) terms in a model
#'
#' Returns one row per distinct delayed term with a surrogate symbol name so
#' symengine can treat the delayed value as an independent variable when
#' differentiating.
#'
#' @param model anything `rxNorm()` accepts (rxode2 model, ui, model vars).
#' @return `NULL` when the model has no delay() terms, otherwise a data.frame
#'   with columns `state`, `tau` (duration expression text) and `surrogate`.
#' @author Matthew L. Fidler
#' @keywords internal
#' @noRd
.rxDelayTerms <- function(model) {
  .e <- .rxNormStatements(model)
  .found <- list()
  .walk <- function(x) {
    if (is.call(x)) {
      if (identical(x[[1]], quote(delay)) && length(x) == 3L) {
        .found[[length(.found) + 1L]] <<- list(
          state = deparse1(x[[2]]),
          tau = deparse1(x[[3]])
        )
      }
      for (.i in seq_along(x)) .walk(x[[.i]])
    }
  }
  for (.i in seq_along(.e)) .walk(.e[[.i]])
  if (length(.found) == 0L) {
    return(NULL)
  }
  .df <- do.call(rbind, lapply(.found, function(z) {
    data.frame(state = z$state, tau = z$tau, stringsAsFactors = FALSE)
  }))
  .df <- unique(.df)
  rownames(.df) <- NULL
  .df$surrogate <- paste0("rx__dly_", .df$state, "_", seq_len(nrow(.df)), "__")
  .df
}

#' Base past(state, tau) <- expr history lines from a symengine env
#'
#' Rebuilds the base `past(state,tau)=expr` line(s) from the stored
#' `rx__pastRhs_STATE__` / `rx__pastTau_STATE__` text; used by gradient-free
#' estimators (SAEM) whose symengine env is built without sensitivities.
#'
#' @param model a symengine environment (as from `.loadSymengine`/`rxS`).
#' @return character vector of past() lines, or NULL if none.
#'
#' @keywords internal
#'
#' @export
.rxPastBaseLinesFromEnv <- function(model) {
  .states <- tryCatch(rxode2::rxStateOde(model), error = function(e) character(0))
  .lines <- character(0)
  for (.si in .states) {
    .rhsTxt <- base::mget(paste0("rx__pastRhs_", .si, "__"), envir = model,
                          ifnotfound = list(NULL))[[1L]]
    if (is.null(.rhsTxt)) next
    .tauTxt <- base::mget(paste0("rx__pastTau_", .si, "__"), envir = model,
                          ifnotfound = list(NULL))[[1L]]
    ## resolve through the env so the injected line references root parameters
    .rhsB <- tryCatch(eval(parse(text = .rhsTxt), envir = model),
                      error = function(e) NULL)
    .rhsOut <- if (!is.null(.rhsB) && inherits(.rhsB, "Basic")) rxFromSE(.rhsB) else .rhsTxt
    .lines <- c(.lines, sprintf("past(%s,%s)=%s", .si, .tauTxt, .rhsOut))
  }
  if (length(.lines)) .lines else NULL
}

#' Extract past(state, tau) <- expr non-constant-history terms from a model
#'
#' @param model anything `rxNorm()` accepts.
#' @return list of {state, tau, expr} (character), or NULL if none.
#' @noRd
.rxPastTerms <- function(model) {
  .e <- .rxNormStatements(model)
  .found <- list()
  for (.i in seq_along(.e)) {
    .st <- .e[[.i]]
    ## past(state, tau) = expr  parses as `=`(past(state, tau), expr)
    if (.rxIsPastAssign(.st)) {
      .lhs <- .st[[2L]]
      .found[[length(.found) + 1L]] <- list(state = deparse1(.lhs[[2L]]),
                                            tau = deparse1(.lhs[[3L]]),
                                            expr = deparse1(.st[[3L]]))
    }
  }
  if (length(.found) == 0L) return(NULL)
  .found
}

#' Validate past(state, tau) <- expr non-constant-history lines
#'
#' The state must be a delayed ODE state, the duration must match one of its
#' delay() terms, and the history expression may not reference an ODE state;
#' machine-generated sensitivity histories (rx__sens_*) are skipped.
#'
#' @param model anything `rxNorm()` accepts.
#' @return invisibly NULL; errors on an invalid past() line.
#' @noRd
.rxValidatePast <- function(model) {
  ## a past() history inside an if/else branch is invisible to the top-level
  ## scan (and its conditional C emission would leave the history undefined on
  ## the other branch): reject it outright
  .assertNoNestedPast <- function(x) {
    if (is.call(x)) {
      if (.rxIsPastAssign(x)) {
        stop(sprintf("past(%s, %s) must be at the top level of the model (not inside if/else)",
                     deparse1(x[[2L]][[2L]]), deparse1(x[[2L]][[3L]])),
             call. = FALSE)
      }
      for (.i in seq_along(x)) .assertNoNestedPast(x[[.i]])
    }
  }
  for (.st in .rxNormStatements(model)) {
    if (.rxIsPastAssign(.st)) next
    .assertNoNestedPast(.st)
  }
  .past <- .rxPastTerms(model)
  if (is.null(.past)) return(invisible(NULL))
  .states <- rxode2::rxStateOde(model)
  .delays <- .rxDelayTerms(model)
  for (.p in .past) {
    if (grepl("^rx__sens_", .p$state)) next          # machine-generated, trusted
    if (!(.p$state %in% .states)) {
      stop(sprintf("past(%s, %s): '%s' is not an ODE state (define d/dt(%s))",
                   .p$state, .p$tau, .p$state, .p$state), call. = FALSE)
    }
    .sd <- if (is.null(.delays)) NULL else .delays[.delays$state == .p$state, , drop = FALSE]
    if (is.null(.sd) || nrow(.sd) == 0L) {
      stop(sprintf("past(%s, %s): '%s' has no delay(%s, ...) term (a past history is only used by delay())",
                   .p$state, .p$tau, .p$state, .p$state), call. = FALSE)
    }
    if (!(.p$tau %in% .sd$tau)) {
      stop(sprintf("past(%s, %s): duration '%s' does not match any delay(%s, ...) (found: %s)",
                   .p$state, .p$tau, .p$tau, .p$state, paste(unique(.sd$tau), collapse = ", ")),
           call. = FALSE)
    }
    .refs <- tryCatch(all.vars(parse(text = .p$expr)[[1L]]), error = function(e) character(0))
    .bad <- intersect(.refs, .states)
    if (length(.bad) > 0L) {
      stop(sprintf("past(%s, %s): history may not reference ODE state(s) '%s' (it is a function of t and parameters only)",
                   .p$state, .p$tau, paste(.bad, collapse = "', '")), call. = FALSE)
    }
  }
  invisible(NULL)
}

#' Named list of explicit assignments (lhs = rhs) in a model
#'
#' Skips ODE (`d/dt(...)`) and compartment-property lines; used to resolve a
#' delay-duration expression down to its root symbols.  Recurses into
#' `if`/`else` branches; an lhs assigned more than once maps to a
#' `(rhs1)+(rhs2)` union of its rhs texts so root-variable resolution stays
#' conservative (any branch may execute).
#'
#' @param model anything `rxNorm()` accepts.
#' @return named character vector mapping each assigned lhs to its rhs text.
#' @noRd
.rxModelDefs <- function(model) {
  .defs <- character(0)
  .collect <- function(x) {
    if (!is.call(x)) return(invisible(NULL))
    if (identical(x[[1L]], quote(`=`)) || identical(x[[1L]], quote(`<-`))) {
      .lhs <- x[[2L]]
      ## only simple `name = rhs` assignments (skip d/dt(x), f(x), etc.)
      if (is.name(.lhs)) {
        .nm <- as.character(.lhs)
        .rhs <- deparse1(x[[3L]])
        if (.nm %in% names(.defs) && !identical(.defs[[.nm]], .rhs)) {
          .defs[[.nm]] <<- paste0("(", .defs[[.nm]], ")+(", .rhs, ")")
        } else {
          .defs[[.nm]] <<- .rhs
        }
      }
    } else if (identical(x[[1L]], quote(`if`)) || identical(x[[1L]], quote(`{`))) {
      for (.i in seq_along(x)[-1L]) .collect(x[[.i]])
    }
    invisible(NULL)
  }
  for (.st in .rxNormStatements(model)) .collect(.st)
  .defs
}

#' Root symbols an expression depends on, resolving through model definitions
#'
#' Transitively expands the free variables of `exprText` through the model's
#' explicit assignments so, e.g., `tau` defined as `exp(eta_tau)` resolves to
#' `eta_tau`.
#'
#' @param exprText expression text (e.g. a delay duration).
#' @param defs named vector from `.rxModelDefs()`.
#' @return character vector of root symbol names.
#' @noRd
.rxResolveRootVars <- function(exprText, defs) {
  .seen <- character(0)
  .roots <- character(0)
  .stack <- all.vars(parse(text = exprText))
  while (length(.stack) > 0L) {
    .v <- .stack[[1L]]
    .stack <- .stack[-1L]
    if (.v %in% .seen) next
    .seen <- c(.seen, .v)
    if (.v %in% names(defs)) {
      .stack <- c(.stack, all.vars(parse(text = defs[[.v]])))
    } else {
      .roots <- c(.roots, .v)
    }
  }
  unique(.roots)
}

#' Root symbols an expression depends on, resolving through a symengine env
#'
#' Env-based analogue of `.rxResolveRootVars()` used inside `.rxSens()`.
#' Bindings are read with `get0(envir=)` (bare `get()` is masked by symengine).
#'
#' @param exprText expression text (e.g. a delay duration).
#' @param model symengine environment.
#' @return character vector of root symbol names.
#' @noRd
.rxResolveRootVarsSE <- function(exprText, model) {
  .seen <- character(0)
  .roots <- character(0)
  .stack <- all.vars(parse(text = exprText))
  while (length(.stack) > 0L) {
    .v <- .stack[[1L]]
    .stack <- .stack[-1L]
    if (.v %in% .seen) next
    .seen <- c(.seen, .v)
    .def <- get0(.v, envir = model, inherits = FALSE)
    if (is.null(.def)) {
      .roots <- c(.roots, .v)
      next
    }
    .fv <- tryCatch(all.vars(parse(text = rxFromSE(.def))), error = function(e) .v)
    if (length(.fv) == 0L) {
      ## numeric constant: not a parameter root
    } else if (length(.fv) == 1L && .fv == .v) {
      .roots <- c(.roots, .v) # symbol defined as itself
    } else {
      .stack <- c(.stack, .fv)
    }
  }
  unique(.roots)
}

#' Validate that delay durations do not depend on a state (env)
#'
#' Parameter/eta-dependent delay durations are supported; a duration depending
#' on a *state* would need the state's own sensitivity inside the duration and
#' is rejected.  Used inside `.rxSens()`.
#'
#' @param model symengine environment from the model loader.
#' @return invisibly `TRUE`; stops with an informative error otherwise.
#' @noRd
.rxDelayValidateTauSE <- function(model) {
  .states <- rxStateOde(model)
  ## walk each RHS as rxFromSE text (symengine intercepts VecBasic `[[`)
  for (.si in .states) {
    .f <- get0(paste0("rx__d_dt_", .si, "__"), envir = model, inherits = FALSE)
    if (is.null(.f)) next
    .e <- parse(text = rxFromSE(.f))
    .walk <- function(x) {
      if (is.call(x)) {
        if (identical(x[[1L]], quote(delay)) && length(x) == 3L) {
          .stateJ <- deparse1(x[[2L]])
          .tau <- deparse1(x[[3L]])
          .bad <- intersect(.rxResolveRootVarsSE(.tau, model), .states)
          if (length(.bad) > 0L) {
            stop("delay duration 'delay(", .stateJ, ", ", .tau,
                 ")' depends on the state(s) ", paste(.bad, collapse = ", "),
                 "; state-dependent delays are not supported for sensitivities",
                 call. = FALSE)
          }
        }
        for (.i in seq_along(x)) .walk(x[[.i]])
      }
    }
    for (.i in seq_along(.e)) .walk(.e[[.i]])
  }
  invisible(TRUE)
}

#' Augment forward-sensitivity equations with the delayed (variational) terms
#'
#' Each forward-sensitivity ODE gains
#' `+ (d f_i / d[delay(y_j, T)]) * delay(rx__sens_<y_j>_BY_<p>__, T)`; the
#' delayed Jacobian comes from substituting the delay subexpression with a
#' fresh symbol and differentiating.
#'
#' @param model symengine environment from `.rxLoadPrune()` (holds the original
#'   ODE RHS as `rx__d_dt_<state>__`).
#' @param sensVec the `..sens` character vector produced by `.rxSens()`.
#' @param params character vector of sensitivity parameters.
#' @return `sensVec` with the delayed terms spliced into each matching equation.
#' @author Matthew L. Fidler
#' @noRd
.rxDelaySensAugment <- function(model, sensVec, params) {
  if (length(sensVec) == 0L) return(sensVec)
  .states <- rxStateOde(model)
  ## Per original state, the delay terms in its RHS and their delayed Jacobians.
  .delayJac <- lapply(.states, function(.si) {
    .f <- get(paste0("rx__d_dt_", .si, "__"), envir = model)
    .fns <- tryCatch(symengine::function_symbols(.f), error = function(e) NULL)
    if (is.null(.fns) || length(.fns) == 0L) {
      return(list())
    }
    .out <- list()
    for (.k in seq_along(.fns)) {
      .fn <- .fns[[.k]]
      ## identify delay() symbols via rxFromSE text (as.character()/get_args()
      ## are intercepted here)
      .fnTxt <- rxFromSE(.fn)
      if (!grepl("^delay\\(", .fnTxt)) next
      .call <- parse(text = .fnTxt)[[1L]]
      .stateJ <- deparse1(.call[[2L]])
      .tau <- deparse1(.call[[3L]])
      .gName <- paste0("rx__gdly", .k, "TMP__")
      .g <- symengine::S(.gName)
      .dj <- symengine::D(symengine::subs(.f, .fn, .g), .g)
      .djTxt <- rxFromSE(.dj)
      ## restore the substituted symbol back to the delay() subexpression
      .djTxt <- gsub(.gName, paste0("delay(", .stateJ, ",", .tau, ")"),
                     .djTxt, fixed = TRUE)
      ## param-dependent delay: precompute d tau/d p here (symengine work stays
      ## in this lapply; the splice below is pure string assembly)
      .dtauByP <- stats::setNames(rep("0", length(params)), params)
      ## eval the duration text in the env to resolve intermediates (S() on a
      ## function expression is intercepted here)
      .tauRes <- tryCatch(eval(parse(text = .tau), envir = model),
                          error = function(e) NULL)
      if (!is.null(.tauRes) && inherits(.tauRes, "Basic")) {
        for (.pp in params) {
          ## assign before rxFromSE, which captures its argument (NSE)
          .psym <- symengine::S(.pp)
          .dD <- tryCatch(symengine::D(.tauRes, .psym), error = function(e) NULL)
          if (!is.null(.dD)) .dtauByP[.pp] <- rxFromSE(.dD)
        }
      }
      .out[[length(.out) + 1L]] <- list(stateJ = .stateJ, tau = .tau,
                                        djac = .djTxt, dtauByP = .dtauByP)
    }
    .out
  })
  names(.delayJac) <- .states
  ## Non-constant pre-history: re-add the base past() line and emit the
  ## per-sensitivity-compartment history
  ## past(rx__sens_<state>_BY_<p>__, tau) = d(expr)/d(p).
  .baseLines <- character(0)   # base state history (also needed by gradient-free SAEM)
  .pastLines <- character(0)   # base + per-sensitivity-compartment histories
  for (.si in .states) {
    .rhsTxt <- base::mget(paste0("rx__pastRhs_", .si, "__"), envir = model,
                          ifnotfound = list(NULL))[[1L]]
    if (is.null(.rhsTxt)) next
    .tauTxt <- base::mget(paste0("rx__pastTau_", .si, "__"), envir = model,
                          ifnotfound = list(NULL))[[1L]]
    ## resolve through the env so the line references root parameters
    ## (past()-only intermediates are dead-code eliminated from the model)
    .rhsB <- tryCatch(eval(parse(text = .rhsTxt), envir = model),
                      error = function(e) NULL)
    .rhsOut <- if (!is.null(.rhsB) && inherits(.rhsB, "Basic")) rxFromSE(.rhsB) else .rhsTxt
    .base <- sprintf("past(%s,%s)=%s", .si, .tauTxt, .rhsOut)
    .baseLines <- c(.baseLines, .base)
    .pastLines <- c(.pastLines, .base)
    ## sens-compartment pre-history: d(history)/d(param)
    if (is.null(.rhsB) || !inherits(.rhsB, "Basic")) next
    for (.p in params) {
      .dp <- tryCatch(symengine::D(.rhsB, symengine::S(.p)), error = function(e) NULL)
      if (is.null(.dp)) next
      .dpTxt <- rxFromSE(.dp)
      if (identical(.dpTxt, "0")) next
      .pastLines <- c(.pastLines,
                      sprintf("past(rx__sens_%s_BY_%s__,%s)=%s",
                              .si, .p, .tauTxt, .dpTxt))
    }
  }
  ## append; unique dedups the base past() line shared by the 1st/2nd-order augments
  .prevBase <- base::mget("..pastBaseLines", envir = model, ifnotfound = list(NULL))[[1L]]
  .baseLines <- unique(c(.prevBase, .baseLines))
  assign("..pastBaseLines", if (length(.baseLines)) .baseLines else NULL, envir = model)
  .prevPast <- base::mget("..pastLines", envir = model, ifnotfound = list(NULL))[[1L]]
  .pastLines <- unique(c(.prevPast, .pastLines))
  assign("..pastLines", if (length(.pastLines)) .pastLines else NULL, envir = model)
  if (all(lengths(.delayJac) == 0L)) {
    assign("..sensDelayAlagF", NULL, envir = model)
    return(sensVec)
  }
  ## Dose-induced breaking-point jump (param-dependent delay only): reproduce
  ## [S_i]=-(djac)*[y_j]*dtau/dp with a modeled bolus on the sensitivity
  ## compartment (alag=tau, f=-(djac)*dtau/dp); a no-op unless rxSolve() adds
  ## the mirroring doses.
  .alagf <- character(0); .seenCmt <- character(0)
  for (.si in .states) {
    .dj <- .delayJac[[.si]]
    if (is.null(.dj) || length(.dj) == 0L) next
    for (.p in params) for (z in .dj) {
      .dtau <- z$dtauByP[[.p]]
      if (is.null(.dtau) || identical(.dtau, "0")) next
      .sensCmt <- paste0("rx__sens_", .si, "_BY_", .p, "__")
      if (.sensCmt %in% .seenCmt) next   # one delay term per state/param (per-cmt alag/f)
      .seenCmt <- c(.seenCmt, .sensCmt)
      .alagf <- c(.alagf, sprintf("alag(%s)=%s", .sensCmt, z$tau),
                          sprintf("f(%s)=-(%s)*(%s)", .sensCmt, z$djac, .dtau))
    }
  }
  assign("..sensDelayAlagF", if (length(.alagf)) .alagf else NULL, envir = model)
  vapply(sensVec, function(.entry) {
    .m <- regmatches(.entry, regexec("^d/dt\\(rx__sens_(.+?)_BY_(.+)__\\)=", .entry))[[1L]]
    if (length(.m) != 3L) {
      return(.entry)
    }
    .si <- .m[2L]
    .p <- .m[3L]
    .dj <- .delayJac[[.si]]
    if (is.null(.dj) || length(.dj) == 0L) {
      return(.entry)
    }
    .add <- vapply(.dj, function(z) {
      ## delay(S_j, tau): the value-sensitivity of the delayed state.
      .term <- paste0("+(", z$djac, ")*delay(rx__sens_", z$stateJ, "_BY_", .p,
                      "__,", z$tau, ")")
      ## param-dependent delay adds -ydot_j(t-tau)*dtau/dp, with
      ## ydot_j(t-tau) = rxDelayD(y_j, tau)
      .dtau <- z$dtauByP[[.p]]
      if (!is.null(.dtau) && !identical(.dtau, "0")) {
        .term <- paste0(.term, "-(", z$djac, ")*rxDelayD(", z$stateJ, ",", z$tau,
                        ")*(", .dtau, ")")
      }
      .term
    }, character(1L))
    .add <- paste(.add, collapse = "")
    ## insert before the initial-condition line (if any), otherwise append
    .nl <- regexpr("\n", .entry, fixed = TRUE)
    if (.nl > 0L) {
      paste0(substr(.entry, 1L, .nl - 1L), .add, substr(.entry, .nl, nchar(.entry)))
    } else {
      paste0(.entry, .add)
    }
  }, character(1L), USE.NAMES = FALSE)
}

#' Dose-induced breaking-point jump for forward delay sensitivities
#'
#' A dose propagated through `delay(y_j, T(p))` makes the 1st-order sensitivity
#' jump at `t = t_dose + T` by `[S_i] = -(d f_i/d delay(y_j,T)) * [y_j] * dT/dp`,
#' reproduced here as a modeled bolus of `[y_j]` on the sensitivity compartment
#' with modeled lag `T` and bioavailability `-(d f_i/d delay)*dT/dp`.
#' `.rxDelaySensJumpMap()` does the model-only symengine analysis (cached via
#' `.rxDelaySensJumpMapCached()`); `.rxDelaySensJumpEvents()` rbinds the
#' mirroring doses per solve; `.rxDelaySensJump()` does both.
#'
#' @param model base ODE model (anything `rxNorm()` accepts).
#' @param calcSens character vector of sensitivity parameters.
#' @param events an rxode2 event table (to mirror the state-j doses onto the
#'   sensitivity compartments).
#' @param jumpMap the jump map from `.rxDelaySensJumpMap()` (`$jumpMap`).
#' @param st ODE state names (`$st` from `.rxDelaySensJumpMap()`), for numeric-cmt
#'   resolution.
#' @return `.rxDelaySensJumpMap()`: list with `alagf`, `jumpMap`, and `st`, or
#'   `NULL` when the model has no parameter-dependent delay.  `.rxDelaySensJump()`:
#'   list with `alagf` and `events`, or `NULL`.
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.rxDelaySensJumpMap <- function(model, calcSens) {
  .m <- rxode2::rxS(rxode2::rxGetModel(model), TRUE, promoteLinSens = FALSE)
  .st <- rxode2::rxStateOde(.m); .ns <- length(.st)
  .findDelays <- function(e, acc = list()) {
    if (is.call(e)) {
      if (identical(e[[1L]], as.name("delay")) && length(e) == 3L) acc[[length(acc) + 1L]] <- e
      for (.a in as.list(e)[-1L]) acc <- .findDelays(.a, acc)
    }
    acc
  }
  .substDelay <- function(e, target, repl) {
    if (identical(e, target)) return(repl)
    if (is.call(e)) for (.i in seq_along(e)) e[[.i]] <- .substDelay(e[[.i]], target, repl)
    e
  }
  .alagf <- character(0); .jumpMap <- list(); .seenCmt <- character(0)
  for (i in seq_len(.ns)) {
    # skip sensitivity compartments when applied to an already-augmented model
    # (their d/dt carries delay(rx__sens_*, tau) which must not spawn its own jump)
    if (grepl("^rx__sens_", .st[i])) next
    .fi <- get0(paste0("rx__d_dt_", .st[i], "__"), envir = .m, inherits = FALSE)
    if (is.null(.fi)) next
    .fiTxt <- rxode2::rxFromSE(.fi); .full <- parse(text = .fiTxt)[[1L]]
    .seen <- character(0)
    for (.dc in .findDelays(.full)) {
      .dcTxt <- deparse1(.dc); if (.dcTxt %in% .seen) next; .seen <- c(.seen, .dcTxt)
      .stateJ <- deparse1(.dc[[2L]]); .tau <- deparse1(.dc[[3L]])
      if (is.na(match(.stateJ, .st))) next
      .g <- "rx__gdlyJTMP__"
      .dj <- symengine::D(symengine::S(deparse1(.substDelay(.full, .dc, as.name(.g)))), symengine::S(.g))
      .djTxt <- gsub(.g, paste0("delay(", .stateJ, ",", .tau, ")"), rxode2::rxFromSE(.dj), fixed = TRUE)
      .tauRes <- tryCatch(eval(parse(text = .tau), envir = .m), error = function(e) NULL)
      for (.p in calcSens) {
        .dt <- "0"
        if (!is.null(.tauRes) && inherits(.tauRes, "Basic")) {
          .dD <- tryCatch(symengine::D(.tauRes, symengine::S(.p)), error = function(e) NULL)
          if (!is.null(.dD)) .dt <- rxode2::rxFromSE(.dD)
        }
        if (identical(.dt, "0")) next
        .sensCmt <- paste0("rx__sens_", .st[i], "_BY_", .p, "__")
        if (.sensCmt %in% .seenCmt)
          stop("forward-sens dose-jump supports one delay term per state/param; '",
               .sensCmt, "' has more than one", call. = FALSE)
        .seenCmt <- c(.seenCmt, .sensCmt)
        # jump [S_i] = -(F_Xd_ij) * [y_j] * dtau/dp  ==  bolus [y_j] with lag tau,
        # bioavailability -(F_Xd_ij)*dtau/dp, on the sensitivity compartment.
        .alagf <- c(.alagf,
                    sprintf("alag(%s)=%s", .sensCmt, .tau),
                    sprintf("f(%s)=-(%s)*(%s)", .sensCmt, .djTxt, .dt))
        .jumpMap[[length(.jumpMap) + 1L]] <- list(sensCmt = .sensCmt, stateJ = .stateJ)
      }
    }
  }
  if (length(.jumpMap) == 0L) return(NULL)
  list(alagf = .alagf, jumpMap = .jumpMap, st = .st)
}

#' @rdname dot-rxDelaySensJumpMap
#' @export
#' @keywords internal
.rxDelaySensJumpEvents <- function(jumpMap, st, events) {
  if (is.null(jumpMap) || length(jumpMap) == 0L) return(events)
  # mirror each state-j dose onto its sensitivity compartment(s); no symengine
  .ev <- as.data.frame(events)
  .isDose <- if (!is.null(.ev$evid)) .ev$evid != 0 else rep(FALSE, nrow(.ev))
  .cmtName <- function(c) if (is.numeric(c)) st[c] else as.character(c)
  .add <- .ev[0, , drop = FALSE]
  for (.jm in jumpMap) {
    for (.r in which(.isDose)) {
      if (!identical(.cmtName(.ev$cmt[.r]), .jm$stateJ)) next
      .row <- .ev[.r, , drop = FALSE]; .row$cmt <- .jm$sensCmt
      .add <- rbind(.add, .row)
    }
  }
  if (nrow(.add)) rbind(.ev, .add) else .ev
}

#' @rdname dot-rxDelaySensJumpMap
#' @export
#' @keywords internal
.rxDelaySensJump <- function(model, calcSens, events) {
  .map <- .rxDelaySensJumpMap(model, calcSens)
  if (is.null(.map)) return(NULL)
  list(alagf = .map$alagf, events = .rxDelaySensJumpEvents(.map$jumpMap, .map$st, events))
}

#' Second-order breaking-point jump: inject the modeled boluses
#'
#' The 2nd-order jump `[S_i^{ab}](xi) = JD_ij * [ydot_j](t_break) * dTa * dTb`
#' at `xi = t_break + T` is delivered as a modeled bolus:
#' `.rxDelaySensAugment2()` emits the common `F = JD * dTa * dTb` and
#' `alag = T`; the magnitude factor is the injected dose amount (`f_j(IC)` at
#' `t0` for the initial history; each user dose on a coupled state `k` mirrored
#' with amount `A * df_j/dy_k`).  `.rxDelaySensJump2Cmts()` finds the 2nd-order
#' jump compartments; `.rxDelaySensJump2Map()` derives each compartment's
#' delayed state, history amount, and couplings.
#'
#' @param norm normalized model text (`rxNorm()` output).
#' @param model anything `rxNorm()`/`rxGetModel()` accept.
#' @param map the map from `.rxDelaySensJump2Map()`.
#' @param events event table (anything `as.data.frame()` accepts).
#' @return `.rxDelaySensJump2Cmts()`: compartment names.  `.rxDelaySensJump2Map()`:
#'   `NULL` or a list with `st` and `entries`.  `.rxDelaySensJump2Events()`:
#'   `events` with the jump boluses added.
#' @author Matthew L. Fidler
#' @noRd
.rxDelaySensJump2Cmts <- function(norm) {
  .lines <- strsplit(norm, "\n", fixed = TRUE)[[1L]]
  .hit <- regmatches(.lines, regexpr("alag\\(rx__sens_[^)]+__\\)", .lines))
  if (length(.hit) == 0L) return(character(0))
  .cmt <- sub("^alag\\((rx__sens_[^)]+__)\\)$", "\\1", .hit)
  ## keep only 2nd-order compartments (exactly two _BY_ groups)
  .cmt <- .cmt[lengths(gregexpr("_BY_", .cmt, fixed = TRUE)) == 2L]
  unique(.cmt)
}

## text -> numeric (via symengine simplification, so e.g. "-ke*0+0" -> 0)
.rxToNum <- function(txt) {
  if (is.null(txt)) return(NA_real_)
  .v <- suppressWarnings(as.numeric(txt))
  if (!is.na(.v)) return(.v)
  .s <- tryCatch(as.character(symengine::S(txt)), error = function(e) NA_character_)
  suppressWarnings(as.numeric(.s))
}

## f_j(IC): the delayed state's RHS with states/delays replaced by their initial
## conditions; numeric for constant history, NA when parameter-dependent.
.rxDelayFjICval <- function(m, j) {
  .f <- get0(paste0("rx__d_dt_", j, "__"), envir = m, inherits = FALSE)
  if (is.null(.f)) return(NA_real_)
  .st <- rxode2::rxStateOde(m)
  .icOf <- function(s) {
    .v <- get0(paste0("rx_", s, "_ini_0__"), envir = m, inherits = FALSE)
    if (is.null(.v)) "0" else rxode2::rxFromSE(.v)
  }
  .subIC <- function(x) {
    if (is.call(x)) {
      if (identical(x[[1L]], quote(delay)) && length(x) == 3L) {
        return(str2lang(.icOf(deparse1(x[[2L]]))))
      }
      for (.i in seq_along(x)) x[[.i]] <- .subIC(x[[.i]]); return(x)
    }
    if (is.name(x) && as.character(x) %in% .st) return(str2lang(.icOf(as.character(x))))
    x
  }
  .rxToNum(deparse1(.subIC(parse(text = rxode2::rxFromSE(.f))[[1L]])))
}

.rxDelaySensJump2Map <- function(model) {
  .norm <- rxNorm(model)
  .cmts <- .rxDelaySensJump2Cmts(.norm)
  if (length(.cmts) == 0L) return(NULL)
  .m <- rxode2::rxS(rxode2::rxGetModel(model), TRUE, promoteLinSens = FALSE)
  .st <- rxode2::rxStateOde(.m)
  .fjTxt <- function(s) {
    .f <- get0(paste0("rx__d_dt_", s, "__"), envir = .m, inherits = FALSE)
    if (is.null(.f)) NULL else rxode2::rxFromSE(.f)
  }
  .fjIC <- function(j) .rxDelayFjICval(.m, j)
  ## couplings df_j/dy_k (delay differentiates to 0 -> instantaneous coupling)
  .coupl <- function(j) {
    .f <- get0(paste0("rx__d_dt_", j, "__"), envir = .m, inherits = FALSE)
    if (is.null(.f)) return(list())
    .out <- list()
    for (.k in .st) {
      .d <- tryCatch(symengine::D(.f, symengine::S(.k)), error = function(e) NULL)
      if (is.null(.d)) next
      .t <- rxode2::rxFromSE(.d)
      if (!identical(.t, "0")) .out[[.k]] <- .t
    }
    .out
  }
  .entries <- list()
  for (.cmt in .cmts) {
    .mm <- regmatches(.cmt, regexec("^rx__sens_(.+?)_BY_(.+?)_BY_(.+)__$", .cmt))[[1L]]
    if (length(.mm) != 4L) next
    .fi <- .fjTxt(.mm[2L]); if (is.null(.fi)) next
    .dd <- NULL
    .walk <- function(x) {
      if (is.call(x)) {
        if (identical(x[[1L]], quote(delay)) && length(x) == 3L) .dd <<- x
        for (.i in seq_along(x)) .walk(x[[.i]])
      }
    }
    .walk(parse(text = .fi)[[1L]])
    if (is.null(.dd)) next
    .j <- deparse1(.dd[[2L]])
    .entries[[.cmt]] <- list(cmt = .cmt, stateJ = .j, histAmt = .fjIC(.j),
                             couplings = .coupl(.j))
  }
  if (length(.entries) == 0L) return(NULL)
  list(st = .st, entries = .entries)
}

.rxDelaySensJump2Events <- function(map, events) {
  if (is.null(map) || length(map$entries) == 0L) return(events)
  .ev <- as.data.frame(events)
  if (nrow(.ev) == 0L) return(.ev)
  if (is.null(.ev$evid)) .ev$evid <- 0L
  if (is.null(.ev$amt)) .ev$amt <- NA_real_
  if (is.null(.ev$cmt)) .ev$cmt <- 1L
  .ev$cmt <- as.character(.ev$cmt)
  .idCol <- intersect(c("id", "ID"), names(.ev))
  .idCol <- if (length(.idCol)) .idCol[1L] else NULL
  .st <- map$st
  .cmtName <- function(c) {
    .i <- suppressWarnings(as.integer(c))
    if (!is.na(.i) && .i >= 1L && .i <= length(.st)) .st[.i] else as.character(c)
  }
  .mkRow <- function(.template, .t, .amt, .cmt) {
    .row <- .template[1L, , drop = FALSE]
    .set <- function(col, val) if (!is.null(.row[[col]])) .row[[col]] <<- val
    .row$time <- .t; .row$evid <- 1L; .row$amt <- .amt; .row$cmt <- .cmt
    .set("ss", 0L); .set("ii", 0); .set("addl", 0L); .set("rate", 0); .set("dur", 0)
    .set("dv", NA_real_)
    .row
  }
  .grps <- if (is.null(.idCol)) list(.ev) else split(.ev, .ev[[.idCol]])
  .add <- .ev[0, , drop = FALSE]
  for (.sub in .grps) {
    .t0 <- min(.sub$time, na.rm = TRUE)
    .dose <- .sub[!is.na(.sub$evid) & .sub$evid != 0, , drop = FALSE]
    for (.e in map$entries) {
      ## initial-history breaking point: bolus of amount f_j(IC) at t0
      .ha <- .e$histAmt
      if (is.numeric(.ha) && !is.na(.ha) && .ha != 0) {
        .add <- rbind(.add, .mkRow(.sub, .t0, .ha, .e$cmt))
      }
      ## dose-induced breaking points: mirror each user dose on a coupled state k
      if (nrow(.dose)) for (.r in seq_len(nrow(.dose))) {
        .k <- .cmtName(.dose$cmt[.r])
        .cp <- .e$couplings[[.k]]
        if (is.null(.cp)) next
        .cpn <- .rxToNum(.cp)
        if (is.na(.cpn)) next            # nonlinear/param coupling: skip (gradient stays exact)
        .A <- .dose$amt[.r]
        if (is.na(.A) || .A == 0) next
        .add <- rbind(.add, .mkRow(.dose[.r, , drop = FALSE], .dose$time[.r], .A * .cpn, .e$cmt))
      }
    }
  }
  if (nrow(.add)) rbind(.ev, .add) else .ev
}

# The jump map depends only on the model, so optimizer inner loops compute the
# symengine analysis once; the wrapper list distinguishes a cached NULL from a miss.
.rxDelaySensJumpCache <- new.env(parent = emptyenv())

#' @rdname dot-rxDelaySensJumpMap
#' @param keyTxt normalized model text used as the cache key (pass `rxNorm(model)`
#'   if already computed, else it is derived).
#' @export
#' @keywords internal
.rxDelaySensJumpMapCached <- function(model, calcSens, keyTxt = NULL) {
  if (is.null(keyTxt)) keyTxt <- rxode2::rxNorm(model)
  .key <- paste0(keyTxt, "\n##cs##", paste(calcSens, collapse = ","))
  .hit <- get0(.key, envir = .rxDelaySensJumpCache, inherits = FALSE)
  if (!is.null(.hit)) return(.hit$map)
  .map <- .rxDelaySensJumpMap(model, calcSens)
  assign(.key, list(map = .map), envir = .rxDelaySensJumpCache)
  .map
}

#' Validate that delay durations are constant for second-order sensitivities
#'
#' Parameter-dependent delays move the DDE breaking points with the parameter,
#' putting jump discontinuities in the second- and higher-order sensitivities;
#' unsupported cases are rejected (first-order sensitivities stay continuous
#' and are supported).
#'
#' @param model symengine environment from the model loader.
#' @param params character vector of sensitivity parameters.
#' @return invisibly `TRUE`; stops with an informative error otherwise.
#' @noRd
.rxDelayValidateHigherOrderSE <- function(model, params, thirdOrder = TRUE) {
  for (.si in rxStateOde(model)) {
    .f <- get0(paste0("rx__d_dt_", .si, "__"), envir = model, inherits = FALSE)
    if (is.null(.f)) next
    .e <- parse(text = rxFromSE(.f))
    ## collect the parameter-dependent delay terms in this state's RHS
    .pdep <- list()
    .walk <- function(x) {
      if (is.call(x)) {
        if (identical(x[[1L]], quote(delay)) && length(x) == 3L) {
          .bad <- intersect(.rxResolveRootVarsSE(deparse1(x[[3L]]), model), params)
          if (length(.bad) > 0L) {
            .pdep[[length(.pdep) + 1L]] <<- list(state = deparse1(x[[2L]]),
                                                 tau = deparse1(x[[3L]]), bad = .bad)
          }
        }
        for (.i in seq_along(x)) .walk(x[[.i]])
      }
    }
    for (.i in seq_along(.e)) .walk(.e[[.i]])
    if (length(.pdep) == 0L) next
    ## 3rd order: param-dependent delays unsupported.  2nd order: a single
    ## param-dependent delay per state is handled; multiple are rejected.
    if (thirdOrder || length(.pdep) > 1L) {
      .d <- .pdep[[1L]]
      .ord <- if (thirdOrder) "third-order" else "second-order"
      .why <- if (!thirdOrder && length(.pdep) > 1L)
        paste0("multiple parameter-dependent delays on state '", .si,
               "' are not yet supported for ", .ord, " sensitivities")
      else
        paste0("parameter-dependent delay 'delay(", .d$state, ", ", .d$tau,
               ")' is not yet supported for analytic ", .ord, " sensitivities")
      stop(.why, ": the delay duration depends on ",
           paste(unique(unlist(lapply(.pdep, `[[`, "bad"))), collapse = ", "),
           ", which moves the DDE breaking points and introduces jump ",
           "discontinuities in the ", .ord, " sensitivities.  The first-order ",
           "sensitivities (the gradient) are exact, so fit these models with a ",
           "numeric or Gauss-Newton Hessian (the default in nlmixr2 FOCEi).",
           call. = FALSE)
    }
    ## 2nd order: the initial-history jump amount f_j(IC) is injected as a
    ## numeric dose amount, so it must be constant.
    if (!thirdOrder && is.na(.rxDelayFjICval(model, .pdep[[1L]]$state))) {
      stop("parameter-dependent delay 'delay(", .pdep[[1L]]$state, ", ",
           .pdep[[1L]]$tau, ")' is not yet supported for analytic second-order ",
           "sensitivities: the delayed state's initial rate depends on ",
           "parameters (a non-constant breaking-point jump).  The first-order ",
           "sensitivities (the gradient) are exact, so fit these models with a ",
           "numeric or Gauss-Newton Hessian (the default in nlmixr2 FOCEi).",
           call. = FALSE)
    }
  }
  invisible(TRUE)
}

#' Augment second-order forward-sensitivity equations with the delayed terms
#'
#' Second-order analogue of `.rxDelaySensAugment()`: treating each
#' `delay(y_j, T)` as a surrogate `g`, splices in the missing terms built from
#' `JD = df/dg`, `H_gy = d^2 f/dg dy`, `H_gg = d^2 f/dg dg'`, and
#' `H_gp = d^2 f/dg dp`, with the delayed sensitivities generalized to
#' surrogate sensitivities `SG` (`rxDelayD`/`rxDelayD2` corrections weighted by
#' `dT/dp`; all corrections vanish for a constant delay).  Unsupported
#' parameter-dependent cases are rejected upstream
#' (`.rxDelayValidateHigherOrderSE()`).
#'
#' @param model symengine environment from the model loader.
#' @param sensVec the second-order `..sens` vector (`rxExpandSens2_` output).
#' @param params character vector of sensitivity parameters.
#' @return `sensVec` with the delayed terms spliced into each matching equation.
#' @author Matthew L. Fidler
#' @noRd
.rxDelaySensAugment2 <- function(model, sensVec, params) {
  if (length(sensVec) == 0L) return(sensVec)
  .states <- rxStateOde(model)
  ## 2nd-order pre-history: past(rx__sens_s_BY_p_BY_q__, tau) = d^2 expr/dp dq
  .pastLines2 <- character(0)
  for (.si in .states) {
    .rhsTxt <- base::mget(paste0("rx__pastRhs_", .si, "__"), envir = model,
                          ifnotfound = list(NULL))[[1L]]
    if (is.null(.rhsTxt)) next
    .tauTxt <- base::mget(paste0("rx__pastTau_", .si, "__"), envir = model,
                          ifnotfound = list(NULL))[[1L]]
    .rhsB <- tryCatch(eval(parse(text = .rhsTxt), envir = model),
                      error = function(e) NULL)
    if (is.null(.rhsB) || !inherits(.rhsB, "Basic")) next
    .cmts <- regmatches(sensVec,
                        regexpr(paste0("rx__sens_", .si, "_BY_[^,)]+_BY_[^,)]+__"),
                                sensVec))
    for (.cmt in unique(.cmts[nzchar(.cmts)])) {
      .mm <- regmatches(.cmt, regexec(
        paste0("^rx__sens_", .si, "_BY_(.+)_BY_(.+)__$"), .cmt))[[1L]]
      if (length(.mm) != 3L) next
      .d2 <- tryCatch(symengine::D(symengine::D(.rhsB, symengine::S(.mm[2L])),
                                   symengine::S(.mm[3L])),
                      error = function(e) NULL)
      if (is.null(.d2)) next
      .d2Txt <- rxFromSE(.d2)
      if (identical(.d2Txt, "0")) next
      .pastLines2 <- c(.pastLines2,
                       sprintf("past(%s,%s)=%s", .cmt, .tauTxt, .d2Txt))
    }
  }
  if (length(.pastLines2)) {
    .prevPast <- base::mget("..pastLines", envir = model, ifnotfound = list(NULL))[[1L]]
    assign("..pastLines", unique(c(.prevPast, .pastLines2)), envir = model)
  }
  .delayJac <- lapply(.states, function(.si) {
    .f <- get0(paste0("rx__d_dt_", .si, "__"), envir = model, inherits = FALSE)
    if (is.null(.f)) return(NULL)
    ## Find delay() terms as Basic function symbols directly on .f -- re-eval'ing
    ## the rxFromSE text in envir=model fails (ETA[n]/THETA[n] are not bound there).
    .fns <- tryCatch(symengine::function_symbols(.f), error = function(e) NULL)
    .terms <- list()
    if (!is.null(.fns)) {
      for (.k in seq_along(.fns)) {
        .fn <- .fns[[.k]]
        .fnTxt <- rxFromSE(.fn)
        if (!grepl("^delay\\(", .fnTxt)) next
        .call <- parse(text = .fnTxt)[[1L]]
        .terms[[length(.terms) + 1L]] <- list(
          fn = .fn, stateJ = deparse1(.call[[2L]]), tau = deparse1(.call[[3L]]),
          gName = paste0("rx__gdly", length(.terms) + 1L, "TMP__"))
      }
    }
    if (length(.terms) == 0L) return(NULL)
    ## Substitute every delay() Basic with its own surrogate symbol, all into
    ## the SAME .fsub so cross derivatives between two delay() terms (hgg) see
    ## both surrogates.
    .fsub <- .f
    for (.t in .terms) .fsub <- symengine::subs(.fsub, .t$fn, symengine::S(.t$gName))
    .restore <- function(txt) {
      for (.t in .terms) {
        txt <- gsub(.t$gName, paste0("delay(", .t$stateJ, ",", .t$tau, ")"),
                    txt, fixed = TRUE)
      }
      txt
    }
    .nz <- function(e) {
      .txt <- rxFromSE(e)
      if (identical(.txt, "0")) NULL else .restore(.txt)
    }
    .out <- lapply(.terms, function(.t) {
      .g <- symengine::S(.t$gName)
      .jdE <- symengine::D(.fsub, .g)                       # JD = df/dg
      .hgy <- list()
      for (.mState in .states) {                            # H_gy = d^2 f/dg dy
        .ym <- symengine::S(.mState)
        .v <- .nz(symengine::D(.jdE, .ym))
        if (!is.null(.v)) .hgy[[.mState]] <- .v
      }
      .hgg <- lapply(.terms, function(.tp) {                # H_gg = d^2 f/dg dg'
        .gp <- symengine::S(.tp$gName)
        .nz(symengine::D(.jdE, .gp))
      })
      .hgp <- list()
      for (.pp in params) {                                 # H_gp = d^2 f/dg dp
        .v <- .nz(symengine::D(.jdE, symengine::S(.pp)))
        if (!is.null(.v)) .hgp[[.pp]] <- .v
      }
      ## param-dependent delay: d tau/dp and d^2 tau/dp dq weight the
      ## rxDelayD/rxDelayD2 corrections below ("0" for a constant delay)
      .dtau <- stats::setNames(rep("0", length(params)), params)
      .d2tau <- list()
      .tauRes <- tryCatch(eval(parse(text = .t$tau), envir = model),
                          error = function(e) NULL)
      if (!is.null(.tauRes) && inherits(.tauRes, "Basic")) {
        .dE <- list()
        for (.pp in params) {
          .psym <- symengine::S(.pp)
          .dpp <- tryCatch(symengine::D(.tauRes, .psym), error = function(e) NULL)
          if (!is.null(.dpp)) {
            .dtau[.pp] <- rxFromSE(.dpp)
            .dE[[.pp]] <- .dpp
          }
        }
        for (.p1 in params) {
          if (is.null(.dE[[.p1]]) || identical(.dtau[[.p1]], "0")) next
          for (.p2 in params) {
            .d2 <- tryCatch(symengine::D(.dE[[.p1]], symengine::S(.p2)),
                            error = function(e) NULL)
            if (!is.null(.d2)) {
              .txt <- rxFromSE(.d2)
              if (!identical(.txt, "0")) .d2tau[[paste0(.p1, "|", .p2)]] <- .txt
            }
          }
        }
      }
      list(stateJ = .t$stateJ, tau = .t$tau, jd = .restore(rxFromSE(.jdE)),
           hgy = .hgy, hgg = .hgg, hgp = .hgp, dtau = .dtau, d2tau = .d2tau)
    })
    .out
  })
  names(.delayJac) <- .states
  if (all(vapply(.delayJac, is.null, logical(1L)))) {
    assign("..sens2DelayAlagF", NULL, envir = model)
    assign("..sens2JumpCmts", NULL, envir = model)
    return(sensVec)
  }
  ## 2nd-order breaking-point jump (param-dependent delay only): S_i^{ab} jumps
  ## at xi1 = t0 + T by (df_i/d delay(y_j,T)) * f_j(t0) * dT/da * dT/db, which
  ## the smooth rxDelayD/rxDelayD2 terms miss; reproduce it with a modeled bolus
  ## on the 2nd-order sens compartment (unit dose at t0 from rxSolve, alag=T,
  ## F=jump magnitude).  Constant delay: dT/dp=0, no jump emitted.
  .alagf2 <- character(0); .jump2Cmts <- character(0); .seen2 <- character(0)
  .nzt0 <- function(x) !is.null(x) && !identical(x, "0")
  vapply(sensVec, function(.entry) {
    .m <- regmatches(.entry,
                     regexec("^d/dt\\(rx__sens_(.+?)_BY_(.+?)_BY_(.+)__\\)=", .entry))[[1L]]
    if (length(.m) != 4L) return(.entry)
    .si <- .m[2L]; .a <- .m[3L]; .b <- .m[4L]
    .dj <- .delayJac[[.si]]
    if (is.null(.dj)) return(.entry)
    .sensCmt2 <- paste0("rx__sens_", .si, "_BY_", .a, "_BY_", .b, "__")
    for (z in .dj) {
      .ta <- z$dtau[[.a]]; .tb <- z$dtau[[.b]]
      if (!.nzt0(.ta) || !.nzt0(.tb)) next          # constant in a or b -> no jump
      if (.sensCmt2 %in% .seen2) next                # one delay term per 2nd-order cmt
      .seen2 <- c(.seen2, .sensCmt2)
      ## common modeled F = JD_ij * dTa * dTb; the [ydot_j](t_break) magnitude
      ## factor is carried by the injected dose amounts (.rxDelaySensJump2Events)
      .alagf2 <<- c(.alagf2,
                    sprintf("alag(%s)=%s", .sensCmt2, z$tau),
                    sprintf("f(%s)=(%s)*(%s)*(%s)", .sensCmt2, z$jd, .ta, .tb))
      .jump2Cmts <<- c(.jump2Cmts, .sensCmt2)
    }
    .Sx <- function(st, ord) paste0("rx__sens_", st, "_BY_", ord, "__")
    .nzt <- function(x) !is.null(x) && !identical(x, "0")
    ## first-order surrogate sensitivity SG_k^p = delay(S_j^p, tau)
    ##   - rxDelayD(y_j, tau) * dtau/dp     (the second term only when tau(p))
    .sg1 <- function(z, p) {
      .s <- paste0("delay(", .Sx(z$stateJ, p), ",", z$tau, ")")
      .dt <- z$dtau[[p]]
      if (.nzt(.dt)) {
        .s <- paste0("(", .s, "-rxDelayD(", z$stateJ, ",", z$tau, ")*(", .dt, "))")
      }
      .s
    }
    ## second-order surrogate sensitivity SG_k^{ab} (the variational delayed
    ## second derivative; reduces to delay(S_j^{ab}, tau) for constant tau).
    .sg2 <- function(z) {
      .s <- paste0("delay(", .Sx(z$stateJ, paste0(.a, "_BY_", .b)), ",", z$tau, ")")
      .ta <- z$dtau[[.a]]; .tb <- z$dtau[[.b]]
      .d2 <- z$d2tau[[paste0(.a, "|", .b)]]
      if (is.null(.d2)) .d2 <- z$d2tau[[paste0(.b, "|", .a)]]
      .corr <- character(0)
      if (.nzt(.tb)) .corr <- c(.corr, paste0("-rxDelayD(", .Sx(z$stateJ, .a), ",",
                                              z$tau, ")*(", .tb, ")"))
      if (.nzt(.ta)) .corr <- c(.corr, paste0("-rxDelayD(", .Sx(z$stateJ, .b), ",",
                                              z$tau, ")*(", .ta, ")"))
      if (.nzt(.ta) && .nzt(.tb)) {
        .corr <- c(.corr, paste0("+rxDelayD2(", z$stateJ, ",", z$tau, ")*(", .ta,
                                 ")*(", .tb, ")"))
      }
      if (.nzt(.d2)) .corr <- c(.corr, paste0("-rxDelayD(", z$stateJ, ",", z$tau,
                                              ")*(", .d2, ")"))
      if (length(.corr) > 0L) .s <- paste0("(", .s, paste(.corr, collapse = ""), ")")
      .s
    }
    .parts <- character(0)
    for (.ki in seq_along(.dj)) {
      z <- .dj[[.ki]]
      .SGa <- .sg1(z, .a); .SGb <- .sg1(z, .b)
      ## pure second-order term: JD * SG_k^{ab}
      .parts <- c(.parts, paste0("+(", z$jd, ")*", .sg2(z)))
      ## H_gy: SG_k^a*S_m^b + S_m^a*SG_k^b
      for (.mState in names(z$hgy)) {
        .parts <- c(.parts,
                    paste0("+(", z$hgy[[.mState]], ")*(", .SGa, "*", .Sx(.mState, .b),
                           "+", .Sx(.mState, .a), "*", .SGb, ")"))
      }
      ## H_gg: SG_k^a * SG_{k'}^b
      for (.kp in seq_along(z$hgg)) {
        .h <- z$hgg[[.kp]]
        if (is.null(.h)) next
        .parts <- c(.parts, paste0("+(", .h, ")*", .SGa, "*", .sg1(.dj[[.kp]], .b)))
      }
      ## H_gp: H_gp[b]*SG_k^a + H_gp[a]*SG_k^b
      if (!is.null(z$hgp[[.b]])) .parts <- c(.parts, paste0("+(", z$hgp[[.b]], ")*", .SGa))
      if (!is.null(z$hgp[[.a]])) .parts <- c(.parts, paste0("+(", z$hgp[[.a]], ")*", .SGb))
    }
    .add <- paste(.parts, collapse = "")
    .nl <- regexpr("\n", .entry, fixed = TRUE)
    if (.nl > 0L) {
      paste0(substr(.entry, 1L, .nl - 1L), .add, substr(.entry, .nl, nchar(.entry)))
    } else {
      paste0(.entry, .add)
    }
  }, character(1L), USE.NAMES = FALSE) -> .res
  assign("..sens2DelayAlagF", if (length(.alagf2)) .alagf2 else NULL, envir = model)
  assign("..sens2JumpCmts", if (length(.jump2Cmts)) unique(.jump2Cmts) else NULL,
         envir = model)
  .res
}

#' Reject nonlinear delays for third-order sensitivities (early, env)
#'
#' `.rxDelaySensAugment3()` only covers delays that appear linearly; run the
#' same check early so a nonlinear delay errors with a clear message.
#'
#' @param model symengine environment from the model loader.
#' @return invisibly `TRUE`; stops otherwise.
#' @noRd
.rxDelayValidate3rdLinearSE <- function(model) {
  .states <- rxStateOde(model)
  for (.si in .states) {
    .f <- get0(paste0("rx__d_dt_", .si, "__"), envir = model, inherits = FALSE)
    if (is.null(.f)) next
    .e <- parse(text = rxFromSE(.f))[[1L]]
    .terms <- list()
    .walk <- function(x) {
      if (is.call(x)) {
        if (identical(x[[1L]], quote(delay)) && length(x) == 3L) {
          .key <- deparse1(x)
          if (!any(vapply(.terms, function(z) z$key == .key, logical(1L)))) {
            .terms[[length(.terms) + 1L]] <<- list(
              key = .key, stateJ = deparse1(x[[2L]]), tau = deparse1(x[[3L]]),
              gName = paste0("rx__gdly", length(.terms) + 1L, "TMP__"))
          }
        }
        for (.i in seq_along(x)) .walk(x[[.i]])
      }
    }
    .walk(.e)
    if (length(.terms) == 0L) next
    .subst <- function(x) {
      if (is.call(x)) {
        if (identical(x[[1L]], quote(delay)) && length(x) == 3L) {
          .key <- deparse1(x)
          for (.t in .terms) if (identical(.t$key, .key)) return(as.name(.t$gName))
        }
        for (.i in seq_along(x)) x[[.i]] <- .subst(x[[.i]])
      }
      x
    }
    .fsubTxt <- deparse1(.subst(.e))
    for (.t in .terms) assign(.t$gName, symengine::S(.t$gName), envir = model)
    .fsub <- eval(parse(text = .fsubTxt), envir = model)
    for (.t in .terms) {
      .g <- symengine::S(.t$gName)
      .jdE <- symengine::D(.fsub, .g)
      for (.zName in c(.states, vapply(.terms, function(z) z$gName, character(1L)))) {
        .zsym <- symengine::S(.zName)
        .d <- symengine::D(.jdE, .zsym)
        if (!identical(rxFromSE(.d), "0")) {
          stop("nonlinear delay 'delay(", .t$stateJ, ", ", .t$tau,
               ")' (the delayed value multiplies a state or another delayed ",
               "value) is not yet supported for third-order sensitivities",
               call. = FALSE)
        }
      }
    }
  }
  invisible(TRUE)
}

#' Augment third-order forward-sensitivity equations with the delayed terms
#'
#' Constant-delay third-order analogue of `.rxDelaySensAugment2()` for delays
#' that appear linearly: splices the missing terms built from `JD = df/dg`,
#' `H_gp = d^2 f/dg dp`, and `H_gpp = d^3 f/dg dp dq`.  Nonlinear delays are
#' rejected here; parameter-dependent delays upstream
#' (`.rxDelayValidateHigherOrderSE()`).
#'
#' @param model symengine environment from the model loader.
#' @param sensVec the third-order `..sens` vector (`rxExpandSens3_()` output).
#' @param params character vector of sensitivity parameters.
#' @return `sensVec` with the delayed terms spliced into each matching equation.
#' @author Matthew L. Fidler
#' @noRd
.rxDelaySensAugment3 <- function(model, sensVec, params) {
  if (length(sensVec) == 0L) return(sensVec)
  .states <- rxStateOde(model)
  .delayJac <- lapply(.states, function(.si) {
    .f <- get0(paste0("rx__d_dt_", .si, "__"), envir = model, inherits = FALSE)
    if (is.null(.f)) return(NULL)
    .e <- parse(text = rxFromSE(.f))[[1L]]
    .terms <- list()
    .walk <- function(x) {
      if (is.call(x)) {
        if (identical(x[[1L]], quote(delay)) && length(x) == 3L) {
          .key <- deparse1(x)
          if (!any(vapply(.terms, function(z) z$key == .key, logical(1L)))) {
            .terms[[length(.terms) + 1L]] <<- list(
              key = .key, stateJ = deparse1(x[[2L]]), tau = deparse1(x[[3L]]),
              gName = paste0("rx__gdly", length(.terms) + 1L, "TMP__"))
          }
        }
        for (.i in seq_along(x)) .walk(x[[.i]])
      }
    }
    .walk(.e)
    if (length(.terms) == 0L) return(NULL)
    .subst <- function(x) {
      if (is.call(x)) {
        if (identical(x[[1L]], quote(delay)) && length(x) == 3L) {
          .key <- deparse1(x)
          for (.t in .terms) if (identical(.t$key, .key)) return(as.name(.t$gName))
        }
        for (.i in seq_along(x)) x[[.i]] <- .subst(x[[.i]])
      }
      x
    }
    .fsubTxt <- deparse1(.subst(.e))
    for (.t in .terms) assign(.t$gName, symengine::S(.t$gName), envir = model)
    .fsub <- eval(parse(text = .fsubTxt), envir = model)
    .restore <- function(txt) {
      for (.t in .terms) {
        txt <- gsub(.t$gName, paste0("delay(", .t$stateJ, ",", .t$tau, ")"),
                    txt, fixed = TRUE)
      }
      txt
    }
    lapply(.terms, function(.t) {
      .g <- symengine::S(.t$gName)
      .jdE <- symengine::D(.fsub, .g)
      ## reject nonlinear delays; assign symengine results before rxFromSE
      ## (which captures its argument)
      for (.mState in .states) {
        .msym <- symengine::S(.mState)
        .dm <- symengine::D(.jdE, .msym)
        if (!identical(rxFromSE(.dm), "0")) {
          stop("nonlinear delay 'delay(", .t$stateJ, ", ", .t$tau,
               ")' (the delayed value multiplies a state) is not yet supported ",
               "for third-order sensitivities", call. = FALSE)
        }
      }
      for (.tp in .terms) {
        .gsym <- symengine::S(.tp$gName)
        .dg <- symengine::D(.jdE, .gsym)
        if (!identical(rxFromSE(.dg), "0")) {
          stop("product of delayed values is not yet supported for third-order ",
               "sensitivities", call. = FALSE)
        }
      }
      .hgp <- list()
      .dE <- list()
      for (.pp in params) {
        .d <- symengine::D(.jdE, symengine::S(.pp))
        .txt <- rxFromSE(.d)
        if (!identical(.txt, "0")) { .hgp[[.pp]] <- .restore(.txt); .dE[[.pp]] <- .d }
      }
      .hgpp <- list()
      for (.p1 in names(.dE)) {
        for (.p2 in params) {
          .d2 <- symengine::D(.dE[[.p1]], symengine::S(.p2))
          .txt <- rxFromSE(.d2)
          if (!identical(.txt, "0")) .hgpp[[paste0(.p1, "|", .p2)]] <- .restore(.txt)
        }
      }
      list(stateJ = .t$stateJ, tau = .t$tau, jd = .restore(rxFromSE(.jdE)),
           hgp = .hgp, hgpp = .hgpp)
    })
  })
  names(.delayJac) <- .states
  if (all(vapply(.delayJac, is.null, logical(1L)))) return(sensVec)
  vapply(sensVec, function(.entry) {
    .m <- regmatches(.entry, regexec(
      "^d/dt\\(rx__sens_(.+?)_BY_(.+?)_BY_(.+?)_BY_(.+)__\\)=", .entry))[[1L]]
    if (length(.m) != 5L) return(.entry)
    .si <- .m[2L]; .a <- .m[3L]; .b <- .m[4L]; .c <- .m[5L]
    .dj <- .delayJac[[.si]]
    if (is.null(.dj)) return(.entry)
    .dS <- function(st, tau, ord) paste0("delay(rx__sens_", st, "_BY_", ord, "__,", tau, ")")
    .gpp <- function(z, p, q) {
      .v <- z$hgpp[[paste0(p, "|", q)]]
      if (is.null(.v)) z$hgpp[[paste0(q, "|", p)]] else .v
    }
    .parts <- character(0)
    for (z in .dj) {
      .parts <- c(.parts, paste0("+(", z$jd, ")*",
                                 .dS(z$stateJ, z$tau, paste0(.a, "_BY_", .b, "_BY_", .c))))
      ## H_gp paired with a second-order delayed sensitivity
      for (.pr in list(c(.c, .a, .b), c(.b, .a, .c), c(.a, .b, .c))) {
        .h <- z$hgp[[.pr[1L]]]
        if (!is.null(.h)) {
          .parts <- c(.parts, paste0("+(", .h, ")*",
                                     .dS(z$stateJ, z$tau, paste0(.pr[2L], "_BY_", .pr[3L]))))
        }
      }
      ## H_gpp paired with a first-order delayed sensitivity
      for (.pr in list(c(.b, .c, .a), c(.a, .c, .b), c(.a, .b, .c))) {
        .h <- .gpp(z, .pr[1L], .pr[2L])
        if (!is.null(.h)) {
          .parts <- c(.parts, paste0("+(", .h, ")*", .dS(z$stateJ, z$tau, .pr[3L])))
        }
      }
    }
    .add <- paste(.parts, collapse = "")
    .nl <- regexpr("\n", .entry, fixed = TRUE)
    if (.nl > 0L) {
      paste0(substr(.entry, 1L, .nl - 1L), .add, substr(.entry, .nl, nchar(.entry)))
    } else {
      paste0(.entry, .add)
    }
  }, character(1L), USE.NAMES = FALSE)
}

#' Validate that delay durations do not depend on the sensitivity parameters
#'
#' Requires `d tau / d p == 0` for every sensitivity parameter; errors naming
#' the offending term.
#'
#' @param model anything `rxNorm()` accepts.
#' @param params character vector of parameters sensitivities are taken w.r.t.
#' @param terms optional pre-computed `.rxDelayTerms()` result.
#' @return invisibly `TRUE` when valid; otherwise stops.
#' @noRd
.rxDelayValidateTau <- function(model, params, terms = NULL) {
  if (is.null(terms)) terms <- .rxDelayTerms(model)
  if (is.null(terms)) {
    return(invisible(TRUE))
  }
  .defs <- .rxModelDefs(model)
  for (.i in seq_len(nrow(terms))) {
    .roots <- .rxResolveRootVars(terms$tau[.i], .defs)
    .bad <- intersect(.roots, params)
    if (length(.bad) > 0L) {
      stop("delay duration 'delay(", terms$state[.i], ", ", terms$tau[.i],
           ")' depends on the sensitivity parameter(s) ",
           paste(.bad, collapse = ", "),
           "; parameter-dependent delays are not yet supported for sensitivities",
           call. = FALSE)
    }
  }
  invisible(TRUE)
}

#' Delayed state for delay differential equations
#'
#' `delay(state, T)` evaluates an ODE state at the past time `t - T`, turning
#' an ordinary differential equation model into a delay differential equation
#' (DDE). The semantics match the `delay()` function of Monolix.
#'
#' @param state An ODE state (compartment) defined in the model whose
#'   past value is required.
#' @param T The delay duration.  May be a constant, a parameter, a
#'   covariate, or any model expression.  The value returned is the state
#'   at time `t - T`.
#'
#' @return Inside an rxode2 model, the value of `state` at the past time
#'   `t - T`.  Before the start of integration the constant
#'   initial-condition history is used.
#'
#' @details
#' Delayed states are interpolated from the solver's dense output, so delay
#' models are solved on a dense path: the default method becomes the dense
#' AutoSwitch composite `"dop853+ros4"`, and dense methods such as `"dop853"`
#' or `"ros4"` also work.  The step size is capped to the smallest delay, and
#' methods that cannot record dense history raise an error.  The dense-output
#' and delay-history machinery is adapted from the `dde` package by Rich
#' FitzJohn and Wes Hinsley (Imperial College of Science, Technology and
#' Medicine).
#'
#' @seealso [rxSolve()]
#'
#' @examples
#' \donttest{
#' # Classic linear delay differential equation y'(t) = -y(t - 1)
#' dde <- rxode2({
#'   y(0) <- 1
#'   d/dt(y) <- -delay(y, 1)
#' })
#'
#' s <- rxSolve(dde, et(seq(0, 5, by = 0.1)))
#'
#' # Delayed (Hutchinson) logistic growth
#' hutch <- rxode2({
#'   r <- 0.5
#'   K <- 10
#'   tau <- 1
#'   N(0) <- 2
#'   d/dt(N) <- r * N * (1 - delay(N, tau) / K)
#' })
#'
#' s2 <- rxSolve(hutch, et(seq(0, 40, by = 0.5)))
#' }
#'
#' @author Matthew L. Fidler
#' @export
delay <- function(state, T) {
  stop("'delay()' can only be used inside an rxode2 model block", call. = FALSE)
}

rxTest({

  # Forward-sensitivity support for delay() models (DDE).  Phase A: cataloging
  # delayed terms and validating that delay durations do not depend on the
  # sensitivity parameters.
  test_that(".rxDelayTerms catalogs distinct delayed terms", {
    .m <- .rxode2({
      d/dt(depot) <- -ka * depot
      d/dt(cen) <- ka * depot - cl / v * cen + kin * delay(cen, tau)
      ka <- exp(tka); cl <- exp(tcl); v <- exp(tv); kin <- 0.1; tau <- 1.5
    })
    .t <- .rxDelayTerms(.m)
    expect_s3_class(.t, "data.frame")
    expect_equal(nrow(.t), 1L)
    expect_equal(.t$state, "cen")
    expect_equal(.t$tau, "tau")
    expect_match(.t$surrogate, "^rx__dly_cen_1__$")
  })

  test_that(".rxDelayTerms returns NULL when there is no delay()", {
    .m <- .rxode2({
      d/dt(x) <- -k * x
      k <- 0.1
    })
    expect_null(.rxDelayTerms(.m))
  })

  test_that(".rxDelayTerms distinguishes multiple delays / durations", {
    .m <- .rxode2({
      d/dt(a) <- -a + delay(a, 1) + delay(a, 2)
      d/dt(b) <- -b + delay(b, 1)
    })
    .t <- .rxDelayTerms(.m)
    expect_equal(nrow(.t), 3L)
    expect_setequal(paste(.t$state, .t$tau), c("a 1", "a 2", "b 1"))
    expect_equal(length(unique(.t$surrogate)), 3L)
  })

  test_that(".rxResolveRootVars resolves through model assignments", {
    .m <- .rxode2({
      d/dt(cen) <- -cl / v * cen + delay(cen, tau)
      cl <- exp(tcl); v <- exp(tv); tau <- exp(ltau)
    })
    .defs <- .rxModelDefs(.m)
    expect_setequal(.rxResolveRootVars("tau", .defs), "ltau")
    expect_setequal(.rxResolveRootVars("cl / v", .defs), c("tcl", "tv"))
  })

  test_that(".rxDelayValidateTau passes for constant / covariate delays", {
    .m <- .rxode2({
      d/dt(cen) <- -cl / v * cen + delay(cen, tau)
      cl <- exp(tcl); v <- exp(tv); tau <- 1.5
    })
    expect_true(.rxDelayValidateTau(.m, c("tcl", "tv")))
    ## covariate-dependent delay is allowed (covariate is not a sens parameter)
    .mc <- .rxode2({
      d/dt(cen) <- -cl / v * cen + delay(cen, tau)
      cl <- exp(tcl); v <- exp(tv); tau <- WT / 70
    })
    expect_true(.rxDelayValidateTau(.mc, c("tcl", "tv")))
  })

  test_that("delay() sensitivity ODEs include the delayed variational term", {
    .m <- .rxode2({
      d/dt(cen) <- -k * cen + kin * delay(cen, tau)
      cen(0) <- 10
      k <- 0.2; kin <- 0.5; tau <- 1
    })
    .ms <- .rxode2(.m, calcSens = c("k", "kin"))
    .n <- rxNorm(.ms)
    ## dS_k/dt has the delayed sensitivity of cen w.r.t. k
    expect_match(.n, "delay(rx__sens_cen_BY_k__,tau)", fixed = TRUE)
    expect_match(.n, "delay(rx__sens_cen_BY_kin__,tau)", fixed = TRUE)
    ## and the sens states inherit the DDE flag once re-parsed
    expect_equal(unname(rxModelVars(.ms)$flags[["hasDelay"]]), 1L)
  })

  test_that("delay() forward sensitivities match central finite differences", {
    .mk <- function(k, kin) {
      .rxode2(paste0("d/dt(cen) <- -k*cen + kin*delay(cen, tau)\n",
                    "cen(0) <- 10\nk <- ", k, "; kin <- ", kin, "; tau <- 1"))
    }
    .ev <- et(seq(0, 8, by = 0.5))
    .s <- rxSolve(.rxode2(.mk(0.2, 0.5), calcSens = c("k", "kin")), .ev,
                  atol = 1e-12, rtol = 1e-12)
    .h <- 1e-4
    .fdk <- (rxSolve(.mk(0.2 + .h, 0.5), .ev, atol = 1e-12, rtol = 1e-12)$cen -
             rxSolve(.mk(0.2 - .h, 0.5), .ev, atol = 1e-12, rtol = 1e-12)$cen) / (2 * .h)
    .fdkin <- (rxSolve(.mk(0.2, 0.5 + .h), .ev, atol = 1e-12, rtol = 1e-12)$cen -
               rxSolve(.mk(0.2, 0.5 - .h), .ev, atol = 1e-12, rtol = 1e-12)$cen) / (2 * .h)
    ## non-trivial sensitivities (delay feedback amplifies them)
    expect_gt(diff(range(.s[["rx__sens_cen_BY_k__"]])), 1)
    expect_lt(max(abs(.s[["rx__sens_cen_BY_k__"]] - .fdk)), 1e-2)
    expect_lt(max(abs(.s[["rx__sens_cen_BY_kin__"]] - .fdkin)), 1e-2)
  })

  test_that("delay() sensitivities are correct in a 2-state model (delay on one state)", {
    ## depot is not delayed; cen is.  The depot sens must have no delay term and
    ## the cen sens must include delay(S_cen, tau).
    .mk <- function(ka, ke) {
      .rxode2(paste0("d/dt(depot) <- -ka*depot\n",
                    "d/dt(cen) <- ka*depot - ke*cen + 0.3*delay(cen, 1)\n",
                    "depot(0) <- 100\nka <- ", ka, "; ke <- ", ke))
    }
    .ev <- et(seq(0, 10, by = 0.5))
    .ms <- .rxode2(.mk(1.0, 0.3), calcSens = c("ka", "ke"))
    .n <- rxNorm(.ms)
    expect_match(.n, "delay(rx__sens_cen_BY_ka__,1)", fixed = TRUE)
    expect_false(grepl("delay(rx__sens_depot", .n, fixed = TRUE))
    .s <- rxSolve(.ms, .ev, atol = 1e-12, rtol = 1e-12)
    .h <- 1e-4
    .fdke <- (rxSolve(.mk(1.0, 0.3 + .h), .ev, atol = 1e-12, rtol = 1e-12)$cen -
              rxSolve(.mk(1.0, 0.3 - .h), .ev, atol = 1e-12, rtol = 1e-12)$cen) / (2 * .h)
    expect_lt(max(abs(.s[["rx__sens_cen_BY_ke__"]] - .fdke)), 1e-2)
  })

  test_that("delay() sensitivity is correct when the initial condition depends on a parameter", {
    ## cen(0) = base, so S_base(0) = 1 and the delayed pre-history of the sens
    ## state (for t < tau) must return that initial sensitivity.
    .mk <- function(base) {
      .rxode2(paste0("d/dt(cen) <- -k*cen + kin*delay(cen, tau)\n",
                    "cen(0) <- base\nk <- 0.2; kin <- 0.5; tau <- 1; base <- ", base))
    }
    .ev <- et(seq(0, 6, by = 0.5))
    .ms <- .rxode2(.mk(10), calcSens = "base")
    expect_match(rxNorm(.ms), "rx__sens_cen_BY_base__(0)=1", fixed = TRUE)
    .s <- rxSolve(.ms, .ev, atol = 1e-12, rtol = 1e-12)
    expect_equal(.s[["rx__sens_cen_BY_base__"]][1], 1, tolerance = 1e-8)
    .h <- 1e-3
    .fd <- (rxSolve(.mk(10 + .h), .ev, atol = 1e-12, rtol = 1e-12)$cen -
            rxSolve(.mk(10 - .h), .ev, atol = 1e-12, rtol = 1e-12)$cen) / (2 * .h)
    expect_lt(max(abs(.s[["rx__sens_cen_BY_base__"]] - .fd)), 1e-4)
  })

  test_that(".rxDelayValidateTau errors for parameter-dependent delays", {
    .m <- .rxode2({
      d/dt(cen) <- -cl / v * cen + delay(cen, tau)
      cl <- exp(tcl); v <- exp(tv); tau <- exp(ltau)
    })
    expect_error(.rxDelayValidateTau(.m, c("tcl", "tv", "ltau")), "not yet supported")
    ## direct dependence (tau is itself the parameter)
    .m2 <- .rxode2({
      d/dt(cen) <- -k * cen + delay(cen, tau)
      k <- 0.1
    })
    expect_error(.rxDelayValidateTau(.m2, c("tau")), "parameter")
  })

  test_that("parameter-dependent delay sensitivities match finite differences", {
    ## tau depends on the sensitivity parameter ltau (tau = exp(ltau)).  The
    ## sensitivity gains the delayed-derivative correction -ydot(t-tau)*dtau/dp.
    .m <- .rxode2("d/dt(cen) <- -k*cen + kin*delay(cen, tau)
cen(0) <- 10
k <- 0.2; kin <- 0.5; tau <- exp(ltau)")
    .ms <- .rxode2(.m, calcSens = "ltau")
    ## the correction term uses the analytic delayed time-derivative rxDelayD()
    expect_match(rxNorm(.ms), "rxDelayD(cen,exp(ltau))", fixed = TRUE)
    .ev <- et(seq(0, 8, by = 0.25))
    .l0 <- log(1.5)
    .s <- rxSolve(.ms, .ev, params = c(ltau = .l0), atol = 1e-12, rtol = 1e-12)
    .h <- 1e-4
    .fd <- (rxSolve(.ms, .ev, params = c(ltau = .l0 + .h), atol = 1e-12, rtol = 1e-12)$cen -
            rxSolve(.ms, .ev, params = c(ltau = .l0 - .h), atol = 1e-12, rtol = 1e-12)$cen) / (2 * .h)
    .an <- .s[["rx__sens_cen_BY_ltau__"]]
    expect_gt(diff(range(.an)), 1) # non-trivial
    ## analytic delayed derivative: matches the FD reference to its own accuracy
    expect_lt(max(abs(.an - .fd)), 1e-3)
  })

  test_that("rxDelayD() is the exact time-derivative of the delayed state", {
    ## d/dt(w) = rxDelayD(y, 1) with y = 10 exp(-t) integrates the delayed
    ## derivative ydot(t-1), so w(t) = y(t-1) - y(prehistory) = 10 exp(-(t-1)) - 10
    ## for t > 1 and 0 for t <= 1.  Validates the analytic dense-interpolant
    ## derivative against a closed form (machine precision).
    .m <- .rxode2({
      d/dt(y) <- -y
      d/dt(w) <- rxDelayD(y, 1)
      y(0) <- 10
      w(0) <- 0
    })
    .tt <- seq(0, 4, by = 0.5)
    .s <- rxSolve(.m, et(.tt), atol = 1e-12, rtol = 1e-12)
    .expect <- ifelse(.tt > 1, 10 * exp(-(.tt - 1)) - 10, 0)
    expect_lt(max(abs(.s$w - .expect)), 1e-7)
  })

  test_that("rxDelayD2/rxDelayD3 are the exact higher time-derivatives", {
    ## y = 10 exp(-t): yddot = y, ydddot = -y, so for tau = 1.5 and t > 1.5
    ##   d/dt(b2) = rxDelayD2(y,1.5) = y(t-1.5)  -> b2 = 10 - 10 exp(-(t-1.5))
    ##   d/dt(b3) = rxDelayD3(y,1.5) = -y(t-1.5) -> b3 = 10 exp(-(t-1.5)) - 10
    .m <- .rxode2({
      d/dt(y) <- -y
      d/dt(b2) <- rxDelayD2(y, 1.5)
      d/dt(b3) <- rxDelayD3(y, 1.5)
      y(0) <- 10; b2(0) <- 0; b3(0) <- 0
    })
    .tt <- seq(0, 5, by = 0.5)
    .s <- rxSolve(.m, et(.tt), atol = 1e-12, rtol = 1e-12)
    expect_lt(max(abs(.s$b2 - ifelse(.tt > 1.5, 10 - 10 * exp(-(.tt - 1.5)), 0))), 1e-6)
    expect_lt(max(abs(.s$b3 - ifelse(.tt > 1.5, 10 * exp(-(.tt - 1.5)) - 10, 0))), 1e-6)
  })

  test_that("state-dependent delays are rejected for sensitivities", {
    .m <- .rxode2({
      d/dt(cen) <- -k * cen + delay(cen, tau)
      tau <- 1 + 0.01 * cen
      k <- 0.2
    })
    expect_error(.rxode2(.m, calcSens = "k"), "state-dependent")
  })

  ## Second-order (constant-delay) sensitivities -- Phase A.  There is no
  ## production assembly path for 2nd-order sensitivities yet (that arrives with
  ## the 3rd-order work), so build the augmented model directly from the pieces.
  .build2nd <- function(rhs, pars) {
    .m <- .rxode2(paste0(rhs, "\ncen(0) <- 10"))
    .mod <- .rxLoadPrune(rxModelVars(.m), FALSE)
    invisible(.rxJacobian(.mod, c("cen", pars)))
    invisible(.rxSens(.mod, pars))
    invisible(.rxSens(.mod, pars, pars))
    .st <- rxStateOde(.mod)
    .rxode2(paste(c(paste0("cmt(", .st, ");"), .mod$..ddt, .mod$..jacobian,
                   .mod$..sens, .mod$..sens2), collapse = "\n"))
  }

  test_that("second-order delay() sensitivities match finite differences", {
    .m2 <- .build2nd("d/dt(cen) <- -k*cen + kin*delay(cen, 1)", c("k", "kin"))
    expect_match(rxNorm(.m2), "delay(rx__sens_cen_BY_k_BY_kin__,1)", fixed = TRUE)
    .ev <- et(seq(0, 6, by = 0.5))
    ## the default composite can stall on the larger 2nd-order DDE system; the
    ## pure dense dop853 path is exact to the method order.
    .sv <- function(k, kin) {
      rxSolve(.m2, .ev, params = c(k = k, kin = kin), method = "dop853",
              atol = 1e-10, rtol = 1e-10)
    }
    .h <- 1e-4
    .s <- .sv(0.2, 0.5)
    ## S^{k,kin} = d(S^k)/dkin : central diff of the analytic first-order S^k
    .fd <- (.sv(0.2, 0.5 + .h)[["rx__sens_cen_BY_k__"]] -
            .sv(0.2, 0.5 - .h)[["rx__sens_cen_BY_k__"]]) / (2 * .h)
    .an <- .s[["rx__sens_cen_BY_k_BY_kin__"]]
    expect_gt(diff(range(.an)), 1)
    expect_lt(max(abs(.an - .fd)), 0.05)
    ## pure second S^{kin,kin}
    .fd2 <- (.sv(0.2, 0.5 + .h)[["rx__sens_cen_BY_kin__"]] -
             .sv(0.2, 0.5 - .h)[["rx__sens_cen_BY_kin__"]]) / (2 * .h)
    expect_lt(max(abs(.s[["rx__sens_cen_BY_kin_BY_kin__"]] - .fd2)), 0.05)
  })

  test_that("param-dependent delay with a param-dependent initial rate is rejected for second-order sensitivities", {
    ## A single parameter-dependent delay is now handled via the xi1 breaking-point
    ## jump, BUT only when the jump amount is constant.  Here tau depends on ltau and
    ## the delayed state's initial rate (-k*cen + kin*delay(cen, tau) at t0) depends
    ## on k and kin, so the breaking-point jump amount is non-constant and the
    ## analytic second-order path is rejected cleanly (numeric Hessian instead).
    .m <- .rxode2("d/dt(cen) <- -k*cen + kin*delay(cen, tau)\ncen(0)<-10\ntau<-exp(ltau)")
    .mod <- .rxLoadPrune(rxModelVars(.m), FALSE)
    invisible(.rxJacobian(.mod, c("cen", "k", "kin", "ltau")))
    invisible(.rxSens(.mod, c("k", "ltau")))
    expect_error(.rxSens(.mod, c("k", "ltau"), c("k", "ltau")),
                 "initial rate depends on")
  })

  ## Third-order (constant-delay) sensitivities.  The general third-order
  ## expansion is interim (R, to be superseded by the C++ one in
  ## nlmixr2/rxode2#1092); the delay augmentation is the part being tested.
  .build3rd <- function(rhs, pars) {
    .m <- .rxode2(paste0(rhs, "\ncen(0) <- 10"))
    .mod <- .rxLoadPrune(rxModelVars(.m), FALSE)
    invisible(.rxJacobian(.mod, c("cen", pars)))
    invisible(.rxSens(.mod, pars))
    invisible(.rxSens(.mod, pars, pars))
    invisible(.rxSens(.mod, pars, pars, pars))
    .st <- rxStateOde(.mod)
    .rxode2(paste(c(paste0("cmt(", .st, ");"), .mod$..ddt, .mod$..jacobian,
                   .mod$..sens, .mod$..sens2, .mod$..sens3), collapse = "\n"))
  }

  test_that("third-order delay() sensitivities match finite differences", {
    .m3 <- .build3rd("d/dt(cen) <- -k*cen + kin*delay(cen, 1)", c("k", "kin"))
    expect_match(rxNorm(.m3), "delay(rx__sens_cen_BY_k_BY_kin_BY_kin__,1)", fixed = TRUE)
    .ev <- et(seq(0, 6, by = 0.5))
    .sv <- function(k, kin) {
      rxSolve(.m3, .ev, params = c(k = k, kin = kin), method = "dop853",
              atol = 1e-10, rtol = 1e-10)
    }
    .h <- 1e-4
    .s <- .sv(0.2, 0.5)
    ## S^{k,kin,kin} = d(S^{k,kin})/dkin : central diff of the analytic 2nd order
    .fd <- (.sv(0.2, 0.5 + .h)[["rx__sens_cen_BY_k_BY_kin__"]] -
            .sv(0.2, 0.5 - .h)[["rx__sens_cen_BY_k_BY_kin__"]]) / (2 * .h)
    .an <- .s[["rx__sens_cen_BY_k_BY_kin_BY_kin__"]]
    expect_gt(diff(range(.an)), 1)
    expect_lt(max(abs(.an - .fd)), 0.05)
  })

  test_that("nonlinear delays are rejected for third-order sensitivities", {
    ## the delayed value multiplies a state -> extra third-order tensor terms
    .m <- .rxode2("d/dt(cen) <- -k*cen*delay(cen, 1)\ncen(0)<-10")
    .mod <- .rxLoadPrune(rxModelVars(.m), FALSE)
    invisible(.rxJacobian(.mod, c("cen", "k")))
    invisible(.rxSens(.mod, "k"))
    invisible(.rxSens(.mod, "k", "k"))
    expect_error(.rxSens(.mod, "k", "k", "k"), "nonlinear")
  })

  # ---- second-order breaking-point JUMP for parameter-dependent delays ----
  # For a constant history the delayed state y_j has a derivative discontinuity at
  # t0; propagated through delay(y_j, T(p)) it makes the SECOND-order sensitivity
  # S_i^{ab} jump at xi1 = t0 + T by  JD_ij * ydot_j(t0) * (dT/da) * (dT/db).  This
  # is reproduced by a modeled bolus on the 2nd-order sens compartment
  # (.rxDelaySensAugment2 emits alag()/f(); rxSolve injects the t0 dose).

  test_that("parameter-dependent delay is allowed for 2nd-order sensitivities", {
    ## previously rejected; the xi1 jump now covers a single param-dependent delay
    expect_s3_class(
      suppressMessages(.rxode2("tau<-exp(L)\n d/dt(y) <- delay(y, tau)\n y(0) <- 1",
                               calcSens = "L", calcSens2 = "L")), "rxode2")
    ## the 2nd-order jump alag()/f() lines are emitted
    .m <- suppressMessages(.rxode2("tau<-exp(L)\n d/dt(y) <- delay(y, tau)\n y(0) <- 1",
                                   calcSens = "L", calcSens2 = "L"))
    expect_true(any(grepl("alag(rx__sens_y_BY_L_BY_L__)", rxNorm(.m), fixed = TRUE)))
  })

  test_that("constant delay emits no 2nd-order jump and still matches (regression)", {
    .m <- suppressMessages(.rxode2("d/dt(y) <- -k*y + delay(y, 1.3)\n y(0) <- 1",
                                   calcSens = "k", calcSens2 = "k"))
    ## no jump alag()/f() for a constant delay (dT/dp == 0)
    expect_false(any(grepl("alag(rx__sens_y_BY_k_BY_k__)", rxNorm(.m), fixed = TRUE)))
  })

  test_that("2nd-order jump sensitivity matches finite differences", {
    ## reference ydot=delay(y,tau), tau=exp(L): S^{LL} jumps by +T^2 at xi1=T.
    .tt <- seq(0, 3, by = 0.02)
    .mk <- function(L) suppressMessages(.rxode2(paste0(
      "tau<-exp(", L, ")\n d/dt(y) <- delay(y, tau)\n y(0) <- 1")))
    .h <- 1e-3
    .g <- function(d) rxSolve(.mk(d), et(.tt), method = "dop853",
                              atol = 1e-10, rtol = 1e-10)$y
    .fd2 <- (.g(.h) - 2 * .g(0) + .g(-.h)) / .h^2
    .ms <- suppressMessages(.rxode2("tau<-exp(L)\n d/dt(y) <- delay(y, tau)\n y(0) <- 1",
                                    calcSens = "L", calcSens2 = "L"))
    .s <- rxSolve(.ms, et(.tt), params = c(L = 0), method = "dop853",
                  atol = 1e-10, rtol = 1e-10)
    .an <- .s[["rx__sens_y_BY_L_BY_L__"]]
    ## compare away from the breaking points (FD straddles the discontinuity there)
    .keep <- abs(.tt - 1) > 0.05 & abs(.tt - 2) > 0.05
    expect_lt(max(abs(.an - .fd2)[.keep], na.rm = TRUE), 0.02)
    ## and the jump is actually present (S^{LL} is large after xi1)
    expect_gt(max(abs(.an), na.rm = TRUE), 1)
  })

  test_that("dose-induced 2nd-order jump matches finite differences", {
    ## delay(cen,tau) driven by a DOSED upstream compartment: the depot dose makes
    ## ydot_cen jump at t0, so S^{LL}_cen jumps at xi1=t0+T by JD*(df_cen/ddepot)*A*(dT/dL)^2.
    ## The user dose is mirrored onto the 2nd-order sens compartment.
    .txt <- paste0("tau<-exp(L)\n ke<-0.7\n d/dt(depot) <- -depot\n",
                   " d/dt(cen) <- depot - ke*cen + 0.5*delay(cen, tau)\n cen(0)<-0")
    .tt <- seq(0, 4, by = 0.02)
    .ev <- et(et(.tt), amt = 10, cmt = "depot", time = 0)
    .mk <- function(L) suppressMessages(.rxode2(gsub("exp(L)", paste0("exp(", L, ")"),
                                                    .txt, fixed = TRUE)))
    .h <- 1e-3
    .g <- function(d) rxSolve(.mk(d), .ev, method = "dop853", atol = 1e-10, rtol = 1e-10)$cen
    .fd2 <- (.g(.h) - 2 * .g(0) + .g(-.h)) / .h^2
    .ms <- suppressMessages(.rxode2(.txt, calcSens = "L", calcSens2 = "L"))
    ## common F = JD*(dT/dL)^2 (magnitude comes from the mirrored dose amount)
    expect_true(any(grepl("f(rx__sens_cen_BY_L_BY_L__)=(0.5)*(exp(L))*(exp(L))",
                          rxNorm(.ms), fixed = TRUE)))
    .an <- rxSolve(.ms, .ev, params = c(L = 0), method = "dop853",
                   atol = 1e-10, rtol = 1e-10)[["rx__sens_cen_BY_L_BY_L__"]]
    .keep <- rep(TRUE, length(.tt))
    for (.b in 1:3) .keep <- .keep & abs(.tt - .b) > 0.08   # exclude FD spikes at xi_n
    expect_lt(max(abs(.an - .fd2)[.keep], na.rm = TRUE), 0.03)
    expect_gt(max(abs(.an), na.rm = TRUE), 1)               # the dose-induced jump is present
  })

  test_that("2nd-order jump is rejected when the delayed state has a parameter baseline", {
    ## f_j(t0) depends on a parameter (non-constant jump amount) -> numeric Hessian.
    expect_error(
      suppressMessages(.rxode2("R0<-exp(lr0)\n tau<-exp(L)\n d/dt(R) <- k0 - delay(R, tau)\n R(0) <- R0",
                               calcSens = c("L", "lr0"), calcSens2 = c("L", "lr0"))),
      "initial rate depends on")
  })
})

rxTest({
  # Delay differential equations: delay(state, T) gives the value of `state`
  # at the past time t - T (Monolix semantics).

  ## Classic linear DDE  y'(t) = -delay(y, 1),  history y(t) = 1 for t <= 0.
  ## Method of steps gives the exact piecewise solution:
  ##   [0,1]: y = 1 - t
  ##   [1,2]: y = t^2/2 - 2t + 3/2
  .exact <- function(t) ifelse(t <= 1, 1 - t, t^2 / 2 - 2 * t + 3 / 2)

  .dde <- rxode2({
    y(0) <- 1
    d/dt(y) <- -delay(y, 1)
  })

  test_that("delay() parses and round-trips in the normalized model", {
    expect_true(grepl("delay(y,1)", rxNorm(.dde), fixed = TRUE))
    expect_equal(unname(rxModelVars(.dde)$flags[["hasDelay"]]), 1L)
  })

  test_that("delay()/rxDelayD() are valid off a d/dt() line (keyed to the delayed state)", {
    ## delay()'s history slot is keyed to the DELAYED STATE (its first argument),
    ## whose per-state __DDT__ define is emitted model-wide -- so delay() is valid
    ## anywhere the state is, not only on a d/dt() right-hand side.  This lets the
    ## discrete-adjoint machinery use delay() in plain lhs (rx__adjFP_*) assignments.
    expect_s3_class(suppressMessages(rxode2({
      d/dt(y) <- -y
      y(0) <- 10
      z <- delay(y, 1)
    })), "rxode2")
    expect_s3_class(suppressMessages(rxode2({
      d/dt(y) <- -y
      y(0) <- 10
      z <- rxDelayD(y, 1)
    })), "rxode2")
    ## the delayed model still reports hasDelay so the dense-history path engages
    .m <- rxode2({
      d/dt(y) <- -y
      y(0) <- 10
      z <- delay(y, 1)
    })
    expect_equal(unname(rxModelVars(.m)$flags[["hasDelay"]]), 1L)
  })

  test_that("delay()'s first argument must be an ODE state, not a parameter", {
    ## delay() interpolates the dense history of a differential state, so its
    ## first argument must have a d/dt() defined.  Passing a parameter
    ## (kel here) used to be silently swallowed as an algebraic compartment;
    ## it must error instead.
    expect_error(rxode2({
      d/dt(central) <- -kel * central + delay(kel, 5)
      conc <- central / v
    }), "syntax error")
    expect_output(
      try(rxode2({
        d/dt(central) <- -kel * central + delay(kel, 5)
        conc <- central / v
      }), silent = TRUE),
      "must be an ODE state"
    )
    ## the same expression with a genuine state is fine
    expect_s3_class(rxode2({
      d/dt(central) <- -kel * central + kel * delay(central, 5)
      conc <- central / v
    }), "rxode2")
  })

  test_that("delay differential equation matches the analytic solution", {
    .tt <- seq(0, 2, by = 0.1)
    .s <- rxSolve(.dde, et(.tt), atol = 1e-10, rtol = 1e-10)
    expect_true(max(abs(.s$y - .exact(.s$time))) < 1e-6)
  })

  test_that("delay models default to the dense dop853+ros4 composite", {
    .tt <- seq(0, 2, by = 0.5)
    ## no method specified -> auto dense dop853+ros4
    .s <- rxSolve(.dde, et(.tt), atol = 1e-10, rtol = 1e-10)
    expect_true(max(abs(.s$y - .exact(.s$time))) < 1e-6)
    ## explicit dop853 also works (dense enabled automatically)
    .s2 <- rxSolve(.dde, et(.tt), method = "dop853", atol = 1e-10, rtol = 1e-10)
    expect_true(max(abs(.s2$y - .exact(.s2$time))) < 1e-6)
    ## explicit composite works
    .s3 <- rxSolve(.dde, et(.tt), method = "dop853+ros4", atol = 1e-10, rtol = 1e-10)
    expect_true(max(abs(.s3$y - .exact(.s3$time))) < 1e-6)
  })

  test_that("small delays (below the natural step size) stay accurate", {
    ## y'(t) = -delay(y, 0.5), history 1.  Method of steps:
    ##   [0,0.5]:   y = 1 - t
    ##   [0.5,1]:   y = t^2/2 - 1.5 t + 1.125
    .exact05 <- function(t) {
      ifelse(t <= 0.5, 1 - t, t^2 / 2 - 1.5 * t + 1.125)
    }
    .dde05 <- rxode2({
      y(0) <- 1
      d/dt(y) <- -delay(y, 0.5)
    })
    .tt <- seq(0, 1, by = 0.1)
    ## an intentionally huge hmax must be overridden by the delay step cap
    .s <- rxSolve(.dde05, et(.tt), atol = 1e-10, rtol = 1e-10, hmax = 100)
    expect_true(max(abs(.s$y - .exact05(.tt))) < 1e-6)
  })

  test_that("delay models reject solvers that cannot record dense history", {
    .tt <- seq(0, 2, by = 0.5)
    for (.m in c("lsoda", "dop5", "bs", "cvode")) {
      expect_error(rxSolve(.dde, et(.tt), method = .m), "dense")
    }
  })

  test_that("stiff delay models solve with the dense ros4 path", {
    .tt <- seq(0, 2, by = 0.1)
    ## ros4 dense output is recorded and interpolated for delay(); the linear
    ## DDE must match the analytic solution on the ros4 path too
    .s <- rxSolve(.dde, et(.tt), method = "ros4", atol = 1e-10, rtol = 1e-10)
    expect_true(max(abs(.s$y - .exact(.tt))) < 1e-5)

    ## a genuinely stiff DDE that overwhelms dop853 (too many steps) solves
    ## fine with ros4
    .stiff <- rxode2({
      y(0) <- 1
      d/dt(y) <- -1e5 * (y - delay(y, 0.2)) - delay(y, 0.2)
    })
    .ss <- rxSolve(.stiff, et(seq(0, 3, by = 0.5)), method = "ros4")
    expect_false(any(is.na(.ss$y)))
    expect_true(all(.ss$y > 0))
  })

  test_that("delay() works with an expression for the delay duration", {
    .m <- rxode2({
      tau <- 0.5
      y(0) <- 2
      d/dt(y) <- -delay(y, tau * 2)
    })
    ## delay duration is 1; on [0,1] history is constant 2 so y' = -2, y = 2 - 2t
    .s <- rxSolve(.m, et(seq(0, 1, by = 0.25)), atol = 1e-10, rtol = 1e-10)
    expect_equal(.s$y, 2 - 2 * .s$time, tolerance = 1e-6)
  })

  test_that("the dop853+ros4 composite switches to ros4 mid-solve for stiff DDEs", {
    ## A very stiff delay model that pure dop853 cannot handle (too many steps).
    .st <- rxode2({
      y(0) <- 1
      d/dt(y) <- -1e5 * (y - delay(y, 0.2)) - delay(y, 0.2)
    })
    .ev <- et(seq(0, 3, by = 0.5))
    ## pure dop853 fails on this stiff problem ...
    expect_error(rxSolve(.st, .ev, method = "dop853"))
    ## ... but the composite default solves it by switching to ros4, matching
    ## the pure ros4 solution.
    .comp <- rxSolve(.st, .ev)
    .ros  <- rxSolve(.st, .ev, method = "ros4")
    expect_false(any(is.na(.comp$y)))
    expect_equal(.comp$y, .ros$y, tolerance = 1e-4)

    ## Mixed stiffness: non-stiff early (dop853), very stiff after t=1.5 (ros4).
    ## The composite must switch mid-solve and mix dop853/ros4 dense history.
    .mix <- rxode2({
      kk <- 0.5 + 1e5 / (1 + exp(-(t - 1.5) * 40))
      y(0) <- 1
      d/dt(y) <- -kk * (y - delay(y, 0.2)) - delay(y, 0.2)
    })
    .evm <- et(seq(0, 3, by = 0.25))
    expect_error(rxSolve(.mix, .evm, method = "dop853")) # dop853 alone fails
    .cm <- rxSolve(.mix, .evm)                           # composite switches
    .rm <- rxSolve(.mix, .evm, method = "ros4")
    expect_false(any(is.na(.cm$y)))
    expect_equal(.cm$y, .rm$y, tolerance = 1e-3)
  })

  test_that("multi-subject DDE solves keep per-subject history isolated", {
    .tt <- seq(0, 2, by = 0.25)
    .ev <- et(.tt)
    .ev$id <- 1
    .e3 <- do.call(rbind, lapply(1:3, function(i) {
      d <- .ev
      d$id <- i
      d
    }))
    .s3 <- rxSolve(.dde, .e3, atol = 1e-10, rtol = 1e-10)
    expect_true(max(abs(.s3$y - .exact(.s3$time))) < 1e-6)
  })

  test_that("a delayed-logistic DDE solves without error", {
    ## Hutchinson's equation: N'(t) = r N (1 - N(t-tau)/K)
    .log <- rxode2({
      r <- 0.5
      K <- 10
      tau <- 1
      N(0) <- 2
      d/dt(N) <- r * N * (1 - delay(N, tau) / K)
    })
    .s <- rxSolve(.log, et(seq(0, 20, by = 0.5)))
    expect_false(any(is.na(.s$N)))
    expect_true(all(.s$N > 0))
  })

  test_that("dense history is recorded only for the states delay() looks back on", {
    ## delay() references the SECOND state; the first state is not delayed.  The
    ## solver records dense history only for delayed states (propDelay bit in
    ## stateProp), so this exercises the compaction + the state-index -> history
    ## column map in _rxDelay.  The delayed state's trajectory must match the
    ## same equation solved on its own.
    .two <- rxode2({
      d/dt(a) <- -a
      d/dt(b) <- -0.5 * b + 0.3 * delay(b, 1)
      a(0) <- 5
      b(0) <- 1
    })
    .one <- rxode2({
      d/dt(b) <- -0.5 * b + 0.3 * delay(b, 1)
      b(0) <- 1
    })
    ## stateProp carries the delay bit only on the delayed state (b, index 1).
    .sp <- rxModelVars(.two)$stateProp
    expect_equal(bitwAnd(.sp[1], 262144L), 0L)        # a: not delayed
    expect_equal(bitwAnd(.sp[2], 262144L), 262144L)   # b: delayed
    .ev <- et(seq(0, 10, by = 0.5))
    .s2 <- rxSolve(.two, .ev, atol = 1e-10, rtol = 1e-10)
    .s1 <- rxSolve(.one, .ev, atol = 1e-10, rtol = 1e-10)
    ## the compacted column maps back to the right state: b matches the
    ## standalone solve, and the undelayed state a still decays as exp(-t).
    expect_true(max(abs(.s2$b - .s1$b)) < 1e-6)
    expect_true(max(abs(.s2$a - 5 * exp(-.s2$time))) < 1e-5)
  })

  test_that("param-dependent delay: forward-sens dose-induced jump matches FD", {
    # .rxDelaySensJump reproduces the dose-induced breaking-point jump in the
    # 1st-order tau-sensitivity as a modeled-lag/F bolus on the sens compartment
    # (no runtime break events).  Assemble base + smooth sens + the alag/f lines,
    # compare the AUTO-WIRED forward sensitivity d/dtau to finite diffs.
    base <- "d/dt(central) = -k * delay(central, tau)"
    p <- c(k = 0.3, tau = 2)
    ev <- et(amt = 10, cmt = "central") %>% et(seq(0.7, 15, by = 1.7))
    # the build auto-emits the alag()/f() jump lines on the sens compartment
    .m <- rxode2::rxode2(base, calcSens = c("k", "tau"))
    expect_true(any(grepl("alag(rx__sens_central_BY_tau__)=tau", strsplit(rxode2::rxNorm(.m), "\n")[[1]], fixed = TRUE)))
    # and rxSolve auto-adds the mirroring sens-compartment dose -> jump captured
    sj <- as.data.frame(rxode2::rxSolve(.m, ev, params = p, method = "dop853",
                                        atol = 1e-10, rtol = 1e-10, cores = 1))
    ex <- rxode2::.rxAdjointExpand(base, c("k", "tau")); madj <- rxode2::rxode2(ex$text)
    sb <- function(pp) as.data.frame(rxode2::rxSolve(madj, ev, params = pp, method = "dop853",
                                                     atol = 1e-10, rtol = 1e-10, cores = 1))$central
    # A dose-induced delay breaking point puts a KINK in central(tau), so a
    # central difference straddling it is O(1)-noisy and gets WORSE as h->0 (h=1e-3
    # lands right in that regime -> ~6e-3; h=1e-2 averages over the kink -> ~7e-4).
    # Use the well-conditioned h to validate the analytic sensitivity.
    hh <- 1e-2; pp <- p; pm <- p; pp["tau"] <- pp["tau"] + hh; pm["tau"] <- pm["tau"] - hh
    fd <- (sb(pp) - sb(pm)) / (2 * hh)
    expect_lt(max(abs(sj[["rx__sens_central_BY_tau__"]] - fd)), 5e-3)
  })

  test_that("an lhs reading delay() is reported correctly in the output (#1140)", {
    # The output data frame recalculates lhs after the solve; the delay history
    # used to be freed at the end of each subject's solve, so such an lhs
    # reported the constant pre-history (0) at every record even though the
    # delayed value drove the ODE.
    .m <- rxode2("a = 0.5 * delay(x, tau)\nd/dt(x) = -k * x + a")
    .ev <- et(amt = 10, cmt = "x") %>% et(seq(0, 24, by = 0.5))
    # tau = 2 lands t - tau on the output grid, so a[t] must match 0.5 * x[t - 2]
    .check <- function(.s) {
      .d <- as.data.frame(.s)
      # at t = tau the lookup time equals the solve start (td <= delayT0), so
      # the boundary record still legitimately reports the pre-history (0)
      expect_equal(.d$a[.d$time <= 2], rep(0, sum(.d$time <= 2)), tolerance = 1e-12)
      # match the delayed lookup by time rather than a row offset
      .j <- match(.d$time - 2, .d$time)
      .i <- which(.d$time > 2)
      expect_equal(.d$a[.i], 0.5 * .d$x[.j[.i]], tolerance = 1e-6)
    }
    .check(rxSolve(.m, c(k = 0.3, tau = 2), .ev))                    # dop853+ros4 default
    .check(rxSolve(.m, c(k = 0.3, tau = 2), .ev, method = "ros4"))   # stiff path
    # multi-subject parallel solve: each subject keeps its own history
    .p <- data.frame(k = seq(0.1, 0.9, length.out = 4), tau = 2)
    .d <- rxSolve(.m, .p, .ev, cores = 2)
    for (.di in split(as.data.frame(.d), .d$sim.id)) .check(.di)
  })

  test_that("delay()/past() models with an if/else block solve (#1151)", {
    # rxNorm() emits `}` and `else` on separate top-level lines, which is only
    # valid R inside a `{ }` block; .rxDelayTerms()/.rxPastTerms()/.rxModelDefs()
    # used to parse the bare normalized text and choke on the `else`.
    .m <- rxode2({
      if (FLAG == 1) {
        kx <- 0
      } else {
        kx <- 5
      }
      d/dt(A1) <- 1 - kg * A1
      d/dt(A2) <- kx * delay(A1, tau1)
    })
    .dt <- .rxDelayTerms(.m)
    expect_equal(.dt$state, "A1")
    expect_equal(.dt$tau, "tau1")
    expect_null(.rxPastTerms(.m))
    # branch assignments are collected as a conservative (rhs1)+(rhs2) union
    expect_equal(.rxModelDefs(.m), c(kx = "(0)+(5)"))

    # past() line alongside the if/else still catalogs at the top level
    .mp <- rxode2({
      if (FLAG == 1) {
        kx <- 0
      } else {
        kx <- 5
      }
      past(A1, tau1) <- 7
      kel <- kg * 2
      d/dt(A1) <- 1 - kg * A1
      d/dt(A2) <- kx * delay(A1, tau1)
    })
    .pt <- .rxPastTerms(.mp)
    expect_equal(.pt[[1L]]$state, "A1")
    expect_equal(.pt[[1L]]$tau, "tau1")
    expect_equal(.pt[[1L]]$expr, "7")
    expect_equal(.rxModelDefs(.mp), c(kx = "(0)+(5)", kel = "kg * 2"))

    # end-to-end solve of the issue reproducer
    f <- function() {
      ini({
        kg <- 0.4
        k4 <- 0.3
        tau1 <- 5
        prop.sd <- 0.1
      })
      model({
        if (FLAG == 1) {
          kx <- 0
        } else {
          kx <- 5
        }
        d/dt(A1) <- 1 - kg * A1
        d/dt(A2) <- k4 * A1 - kx * delay(A1, tau1)
        cp <- A2
        cp ~ prop(prop.sd)
      })
    }
    .ev <- et(seq(0, 20, by = 1))
    .ev$FLAG <- 1
    .s <- suppressWarnings(rxSolve(rxode2(f), .ev))
    expect_true(all(is.finite(.s$cp)))

    # and with an explicit past() history line
    .s2 <- suppressWarnings(rxSolve(.mp, c(kg = 0.4, tau1 = 5), .ev))
    expect_true(all(is.finite(.s2$A2)))
  })

  test_that("past() inside an if/else branch is rejected, not silently skipped (#1151)", {
    # a conditional past() is invisible to the top-level scan and its C emission
    # would leave the history undefined on the other branch
    .m <- rxode2({
      if (FLAG == 1) {
        past(A1, tau1) <- 7
      } else {
        past(A1, tau1) <- 5
      }
      d/dt(A1) <- 1 - kg * A1
      d/dt(A2) <- delay(A1, tau1)
    })
    expect_error(.rxValidatePast(.m), "top level")
    .ev <- et(seq(0, 10, by = 1))
    .ev$FLAG <- 1
    expect_error(suppressWarnings(rxSolve(.m, c(kg = 0.4, tau1 = 5), .ev)),
                 "top level")
  })

  test_that("a delay duration alias assigned inside if/else is still validated (#1151)", {
    # tau resolves through a branch assignment; the sensitivity-parameter check
    # must still see the dependence
    .m <- rxode2({
      if (FLAG == 1) {
        tau1 <- exp(ltau)
      } else {
        tau1 <- 1
      }
      d/dt(A1) <- 1 - kg * A1
      d/dt(A2) <- delay(A1, tau1)
    })
    expect_error(.rxDelayValidateTau(.m, "ltau"), "not yet supported")
    expect_true(.rxDelayValidateTau(.m, "kg"))
  })
})

regIf <- rex::rex(start, any_spaces, "if", any_spaces, "(", capture(anything), ")", any_spaces, "{", any_spaces, end)
regElse <- rex::rex(start, any_spaces, "else", any_spaces, "{", any_spaces, end)
regEnd <- rex::rex(start, any_spaces, "}", any_spaces, end)
regIfOrElse <- rex::rex(or(regIf, regElse))

## first.arg second.arg and type of function
## rxode2->symengine
.rxSEsingle <- list(
  "gammafn" = c("gamma(", ")", "gamma"),
  "lgammafn" = c("lgamma(", ")", "lgamma"),
  "lgamma" = c("lgamma(", ")", "lgamma"),
  "loggamma" = c("lgamma(", ")", "lgamma"),
  "digamma" = c("polygamma(0,", ")", "psigamma"),
  "trigamma" = c("polygamma(1,", ")", "psigamma"),
  "tetragamma" = c("polygamma(2,", ")", "psigamma"),
  "pentagamma" = c("polygamma(3,", ")", "psigamma"),
  "cospi" = c("cos(pi*(", "))", "cos"),
  "sinpi" = c("sin(pi*(", "))", "sin"),
  "tanpi" = c("tan(pi*(", "))", "tan"),
  "log1p" = c("log(1+", ")", "log"),
  "expm1" = c("(exp(", ")-1)", "exp"),
  "factorial" = c("gamma(", "+1)", "gamma"),
  "lfactorial" = c("lgamma(", "+1)", "lgamma"),
  "lgamma1p" = c("lgamma(", "+1)", "lgamma"),
  "expm1" = c("(exp(", ")-1)", "exp"),
  "log10" = c("log(", ")/log(10)", "log"),
  "log2" = c("log(", ")/log(2)", "log"),
  "log1pexp" = c("log(1+exp(", "))", "log1pexp"),
  "!" = c("rxNot(", ")", ""),
  "phi" = c("0.5*(1+erf((", ")/sqrt(2)))"),
  "pnorm" = c("0.5*(1+erf((", ")/sqrt(2)))"),
  "normcdf" = c("0.5*(1+erf((", ")/sqrt(2)))"),
  "qnorm" = c("sqrt(2)*erfinv(2*(", ")-1)"),
  "fabs" = c("abs0(", ")")
)

.SEsingle <- list(
  "rxNot" = c("(!(", "))"),
  "loggamma" = c("lgamma(", ")")
)

.rxSEdouble <- list(
  "pow" = c("(", ")^(", ")"),
  "R_pow" = c("(", ")^(", ")"),
  "R_pow_di" = c("(", ")^(", ")"),
  "Rx_pow_di" = c("(", ")^(", ")"),
  "Rx_pow" = c("(", ")^(", ")"),
  "lbeta" = c("log(beta(", ",", "))"),
  "==" = c("rxEq(", ",", ")"),
  "!=" = c("rxNeq(", ",", ")"),
  ">=" = c("rxGeq(", ",", ")"),
  "<=" = c("rxLeq(", ",", ")"),
  "<" = c("rxLt(", ",", ")"),
  ">" = c("rxGt(", ",", ")"),
  "&&" = c("rxAnd(", ",", ")"),
  "||" = c("rxOr(", ",", ")"),
  "&" = c("rxAnd(", ",", ")"),
  "|" = c("rxOr(", ",", ")"),
  "%%" = c("rxMod(", ",", ")")
)

.SEdouble <- list(
  "lbeta" = c("lbeta(", ",", ")"),
  "rxMod" = c("(", "%%", ")"),
  "rxEq" = c("(", "==", ")"),
  "rxNeq" = c("(", "!=", ")"),
  "rxGeq" = c("(", ">=", ")"),
  "rxLeq" = c("(", "<=", ")"),
  "rxGt" = c("(", ">", ")"),
  "rxLt" = c("(", "<", ")"),
  "rxAnd" = c("(", "&&", ")"),
  "rxOr" = c("(", "||", ")"),
  # raw R comparison/logical operators map to themselves so rxFromSE() can
  # re-consume its own output; the Subs handler re-parses converted text with
  # R's parser, turning rxGt()/rxEq() back into `>`/`==` operator calls
  "==" = c("(", "==", ")"),
  "!=" = c("(", "!=", ")"),
  ">=" = c("(", ">=", ")"),
  "<=" = c("(", "<=", ")"),
  ">" = c("(", ">", ")"),
  "<" = c("(", "<", ")"),
  "&&" = c("(", "&&", ")"),
  "&" = c("(", "&&", ")"),
  "||" = c("(", "||", ")"),
  "|" = c("(", "||", ")"),
  "%%" = c("(", "%%", ")"),
  "R_pow"=c("(", ")^(", ")"),
  "R_pow_di"=c("(", ")^(", ")"),
  "Rx_pow"=c("(", ")^(", ")"),
  "Rx_pow_di"=c("(", ")^(", ")"),
  ## Bare R relational/logical operators.  These are not produced by symengine
  ## directly (it uses rxEq/rxGt/...), but a Subs() over a Derivative whose body
  ## carries a relational -- e.g. from abs() or an occasion indicator -- converts
  ## rxGt/rxEq to R `>`/`==` and then re-runs .rxFromSE(); handle those infix so
  ## the relational survives the second pass instead of tripping the unknown-user-
  ## function check.
  ">"=c("(", ">", ")"),
  "<"=c("(", "<", ")"),
  ">="=c("(", ">=", ")"),
  "<="=c("(", "<=", ")"),
  "=="=c("(", "==", ")"),
  "!="=c("(", "!=", ")"),
  "&&"=c("(", "&&", ")"),
  "||"=c("(", "||", ")")
)

## atan2
.rxSEeq <- c(
  "lgamma" = 1,
  "abs" = 1,
  "acos" = 1,
  "acosh" = 1,
  "asin" = 1,
  "asinh" = 1,
  "atan" = 1,
  "atan2" = 2,
  "atanh" = 1,
  "beta" = 2,
  "cos" = 1,
  "cosh" = 1,
  "erf" = 1,
  "erfc" = 1,
  "exp" = 1,
  "gamma" = 1,
  "log" = 1,
  "polygamma" = 2,
  "rxTBS" = 5,
  "rxTBSi" = 5,
  "rxTBSd" = 5,
  "rxTBSd2" = 5,
  "rxTBSdL" = 5,
  "rxTBSdL2" = 5,
  "rxTBSdLx" = 5,
  "sin" = 1,
  "sinh" = 1,
  "sqrt" = 1,
  "tan" = 1,
  "tanh" = 1,
  "gammap" = 2,
  ## C's math.h library
  "floor" = 1,
  "round" = 1,
  "ceil" = 1,
  "trunc" = 1,
  ## Special R functions
  "bessel_i" = 3,
  "bessel_j" = 2,
  "bessel_k" = 3,
  "bessel_y" = 2,
  "logspace_add" = 2,
  "logspace_sub" = 2,
  "fmax2" = 2,
  "fmin2" = 2,
  "sign" = 1,
  "fsign" = 2,
  "fprec" = 2,
  "fround" = 2,
  "ftrunc" = 2,
  "transit" = NA,
  "gammaq" = 2,
  "gammapDer" = 2,
  "gammapInv" = 2,
  "gammapInva" = 2,
  "gammaqInv" = 2,
  "gammaqInva" = 2,
  "lowergamma" = 2,
  "uppergamma" = 2,
  "linCmtA"=14,
  "linCmtB"=15,
  "max" = NA,
  "min" = NA,
  "logit" = NA,
  "expit" = NA,
  "probit" = NA,
  "probitInv" = NA,
  "tlast" = NA,
  "tlast0" = NA,
  "tfirst" = NA,
  "tfirst0" = NA,
  "lag" = NA,
  "lead" = NA,
  "lag0" = NA,
  "lead0" = NA,
  "dose" =NA,
  "podo" =NA,
  "dose0" =NA,
  "podo0" =NA,
  "dabs" = 1,
  "dabs2" = 1,
  "abs1" = 1,
  "dabs1" = 1,
  "erfinv" = 1,
  "abs0" = 1,
  "dosenum" = 0,
  "first" = 1,
  "last" = 1,
  "diff" = 1,
  "diff0" = 1,
  "is.nan" = 1,
  "is.na" = 1,
  "is.finite" = 1,
  "is.infinite" = 1,
  "llikPois"=2,
  "llikPoisDlambda"=2,
  "llikBinom"=3,
  "llikBinomDprob"=3,
  "llikNbinom"=3,
  "llikNbinomDprob"=3,
  "llikNbinomMu"=3,
  "llikNbinomMuDmu"=3,
  "llikBeta"=3,
  "llikBetaDshape1"=3,
  "llikBetaDshape2"=3,
  "llikT"=4,
  "llikTDdf"=4,
  "llikTDmean"=4,
  "llikTDsd"=4,
  "llikChisq"=2,
  "llikChisqDdf"=2,
  "llikExp"=2,
  "llikExpDrate"=2,
  "llikF"=3,
  "llikFDdf1"=3,
  "llikFDdf2"=3,
  "llikGeom"=2,
  "llikGeomDprob"=2,
  "llikUnif"=3,
  "llikUnifDalpha"=3,
  "llikUnifDbeta"=3,
  "llikWeibull"=3,
  "llikWeibullDshape"=3,
  "llikWeibullDscale"=3,
  "llikGamma"=3,
  "llikGammaDshape"=3,
  "llikGammaDrate"=3,
  "llikCauchy"=3,
  "llikCauchyDlocation"=3,
  "llikCauchyDscale"=3,
  "llikNorm"=3,
  "llikNormDmean"=3,
  "llikNormDsd"=3,
  # Now the llikX variety for saving comp. time
  "llikXPois"=3,
  "llikXPoisDlambda"=3,
  "llikXBinom"=4,
  "llikXBinomDprob"=4,
  "llikXNbinomMu"=4,
  "llikXNbinomMuDmu"=4,
  "llikXNbinom"=4,
  "llikXNbinomDprob"=4,
  "llikXBeta"=4,
  "llikXBetaDshape1"=4,
  "llikXBetaDshape2"=4,
  "llikXT"=5,
  "llikXTDdf"=5,
  "llikXTDmean"=5,
  "llikXTDsd"=5,
  "llikXChisq"=3,
  "llikXChisqDdf"=3,
  "llikXExp"=3,
  "llikXExpDrate"=3,
  "llikXF"=4,
  "llikXFDdf1"=4,
  "llikXFDdf2"=4,
  "llikXGeom"=3,
  "llikXGeomDprob"=3,
  "llikXUnif"=4,
  "llikXUnifDalpha"=4,
  "llikXUnifDbeta"=4,
  "llikXWeibull"=4,
  "llikXWeibullDshape"=4,
  "llikXWeibullDscale"=4,
  "llikXGamma"=4,
  "llikXGammaDshape"=4,
  "llikXGammaDrate"=4,
  "llikXCauchy"=4,
  "llikXCauchyDlocation"=4,
  "llikXCauchyDscale"=4,
  "llikXNorm"=4,
  "llikXNormDmean"=4,
  "llikXNormDsd"=4,
  "ReLU"=1,
  "dReLU"=1,
  "GELU"=1,
  "dGELU"=1,
  "d2GELU"=1,
  "d3GELU"=1,
  "d4GELU"=1,
  "ELU"=2,
  "dELU"=2,
  "d2ELU"=2,
  "d2aELU"=2,
  "dELUa"=2,
  "d2ELUa"=2,
  "softplus"=1,
  "dsoftplus"=1,
  "d2softplus"=1,
  "d3softplus"=1,
  "d4softplus"=1,
  "SELU"=1,
  "dSELU"=1,
  "lReLU"=1,
  "dlReLU"=1,
  "PReLU"=2,
  "dPReLU"=2,
  "d2PReLU"=2,
  "dPReLUa"=2,
  "dPReLUa1"=2,
  "Swish"=1,
  "dSwish"=1
)

.rxOnly <- c(
  ## Now random number generators
  "rnorm" = NA,
  "rxnorm" = NA,
  "rxbinom" = 2,
  "rbinom" = 2,
  "rxcauchy" = NA,
  "rcauchy" = NA,
  "rchisq" = 1,
  "rxchisq" = 1,
  "rexp" = 1,
  "rxexp" = 1,
  "rbeta" = 2,
  "rxbeta" = 2,
  "rgeom" = 1,
  "rxgeom" = 1,
  "rxpois" = 1,
  "rpois" = 1,
  "rxt" = 1,
  "rt" = 1
)

#' Add/Create C functions for use in rxode2
#'
#' @inheritParams rxFunParse
#'
#' @param name This can either give the name of the user function or
#'   be a simple R function that you wish to convert to C.  If you
#'   have rxode2 convert the R function to C, the name of the function
#'   will match the function name provided and the number of arguments
#'   will match the R function provided.  Hence, if you are providing
#'   an R function for conversion to C, the rest of the arguments are
#'   implied.
#' @export
#' @examples
#' \donttest{
#' # Right now rxode2 is not aware of the function fun
#' # Therefore it cannot translate it to symengine or
#' # Compile a model with it.
#'
#' try(rxode2("a=fun(a,b,c)"))
#'
#' # Note for this approach to work, it cannot interfere with C
#' # function names or reserved rxode2 special terms.  Therefore
#' # f(x) would not work since f is an alias for bioavailability.
#'
#' fun <- "
#' double fun(double a, double b, double c) {
#'   return a*a+b*a+c;
#' }
#' " # C-code for function
#'
#' rxFun("fun", c("a", "b", "c"), fun) ## Added function
#'
#' # Now rxode2 knows how to translate this function to symengine
#'
#' rxToSE("fun(a,b,c)")
#'
#' # And will take a central difference when calculating derivatives
#'
#' rxFromSE("Derivative(fun(a,b,c),a)")
#'
#' ## Of course, you could specify the derivative table manually
#' rxD("fun", list(
#'   function(a, b, c) {
#'     paste0("2*", a, "+", b)
#'   },
#'   function(a, b, c) {
#'     return(a)
#'   },
#'   function(a, b, c) {
#'     return("0.0")
#'   }
#' ))
#'
#' rxFromSE("Derivative(fun(a,b,c),a)")
#'
#' # You can also remove the functions by `rxRmFun`
#'
#' rxRmFun("fun")
#'
#' # you can also use R functions directly in rxode2
#'
#'
#' gg <- function(x, y) {
#'   x + y
#' }
#'
#' f <- rxode2({
#'  z = gg(x, y)
#' })
#'
#'
#' e <- et(1:10) |> as.data.frame()
#'
#' e$x <- 1:10
#' e$y <- 21:30
#'
#' rxSolve(f, e)
#'
#' # Note that since it touches R, it can only run single-threaded.
#' # There are also requirements for the function:
#' #
#' # 1. It accepts one value per argument (numeric)
#' #
#' # 2. It returns one numeric value
#'
#' # If it is a simple function (like gg) you can also convert it to C
#' # using rxFun and load it into rxode2
#'
#' rxFun(gg)
#'
#' rxSolve(f, e)
#'
#' # to stop the recompile simply reassign the function
#' f <- rxode2(f)
#'
#' rxSolve(f, e)
#'
#' rxRmFun("gg")
#' rm(gg)
#' rm(f)
#'
#'
#' # You can also automatically convert a R function to R code (and
#' # calculate first derivatives)
#'
#' fun <- function(a, b, c) {
#'   a^2+b*a+c
#' }
#'
#' rxFun(fun)
#'
#' # You can see the R code if you want with rxC
#'
#' message(rxC("fun"))
#'
#' # you can also remove both the function and the
#' # derivatives with rxRmFun("fun")
#'
#' rxRmFun("fun")
#'
#' }
rxFun <- function(name, args, cCode) {
  if (missing(args) && missing(cCode)) {
    .funName <- as.character(substitute(name))
    .lst <- rxFun2c(name, name=.funName)
    .env <- new.env(parent=emptyenv())
    .env$d <- list()
    lapply(seq_along(.lst), function(i) {
      .cur <- .lst[[i]]
      do.call(rxode2::rxFunParse, .cur[1:3])
      message("converted R function '", .cur$name, "' to C (will now use in rxode2)")
      ## message(.cur$cCode)
      if (length(.cur) == 4L) {
        .env$d <- c(.env$d, list(.cur[[4]]))
      }
    })
    if (length(.env$d) > 0) {
      message("Added derivative table for '", .lst[[1]]$name, "'")
       rxD(.lst[[1]]$name, .env$d)
    }
    return(invisible())
  }
  rxFunParse(name, args, cCode)
}

#' @rdname rxFun
#' @export
rxRmFun <- function(name) {
  rxRmFunParse(name)
}

.SE1p <- c(
  "loggamma" = "lgamma1p",
  "lgamma" = "lgamma1p",
  "log" = "log1p"
)

.SE1m <- c(
  "cos" = "cospi",
  "sin" = "sinpi",
  "tan" = "tanpi"
)
## "rxTBS", "rxTBSd"
.rxSEcnt <- c(
  "M_E" = "E",
  "M_PI" = "pi",
  "M_PI_2" = "pi/2",
  "M_PI_4" = "pi/4",
  "M_1_PI" = "1/pi",
  "M_2_PI" = "2/pi",
  "M_2PI" = "2*pi",
  "M_SQRT_PI" = "sqrt(pi)",
  "M_2_SQRTPI" = "2/sqrt(pi)",
  "M_1_SQRT_2PI" = "1/sqrt(2*pi)",
  "M_SQRT2" = "sqrt(2)",
  "M_SQRT_3" = "sqrt(3)",
  "M_SQRT_32" = "sqrt(32)",
  "M_SQRT_2dPI" = "sqrt(2/pi)",
  "M_LN_SQRT_PI" = "log(sqrt(pi))",
  "M_LN_SQRT_2PI" = "log(sqrt(2*pi))",
  "M_LN_SQRT_PId2" = "log(sqrt(pi/2))",
  "M_LOG10_2" = "log(2)/log(10)",
  "M_LOG2E" = "1/log(2)",
  "M_LOG10E" = "1/log(10)",
  "M_LN2" = "log(2)",
  "M_LN10" = "log(10)"
)
## "rxTBS", "rxTBSd"

.rxSEreserved <- list(
  "e" = 2.718281828459045090796,
  "E" = 2.718281828459045090796,
  "EulerGamma" = 0.57721566490153286060651209008240243104215933593992,
  "Catalan" = 0.915965594177219015054603514932384110774,
  "GoldenRatio" = 2.118033988749894902526,
  "I" = 1i
)

#' Add to rxode2's derivative tables
#'
#' @param name Function Name
#' @param derivatives A list of functions. Each function takes the
#'     same number of arguments as the original function.  The first
#'     function will construct the derivative with respect to the
#'     first argument; The second function will construct the
#'     derivitive with respect to the second argument, and so on.
#' @return nothing
#' @author Matthew Fidler
#' @export
#' @examples
#' ## Add an arbitrary list of derivative functions
#' ## In this case the fun(x,y) is assumed to be 0.5*x^2+0.5*y^2
#'
#' rxD("fun", list(
#'   function(x, y) {
#'     return(x)
#'   },
#'   function(x, y) {
#'     return(y)
#'   }
#' ))
rxD <- function(name, derivatives) {
  if (!inherits(derivatives, "list") || length(derivatives) == 0L) {
    stop("derivatives must be a list of functions with at least 1 element", call. = FALSE)
  }
  if (!all(sapply(derivatives, function(x) (inherits(x, "function") || is.null(x))))) {
    stop("derivatives must be a list of functions with at least 1 element", call. = FALSE)
  }
  .rxD <- rxode2::rxode2parseD()
  if (exists(name, envir = .rxD)) {
    warning(sprintf(gettext("replacing defined derivatives for '%s'"), name), call. = FALSE)
  }
  assign(name, derivatives, envir = .rxD)
  return(invisible())
}

.rxToSE.envir <- new.env(parent=emptyenv())
.rxToSE.envir$envir <- NULL

.promoteLinB <- FALSE
#' rxode2 to symengine environment
#'
#' @param x expression
#'
#' @param envir default is `NULL`; Environment to put symengine
#'     variables in.
#'
#' @param progress shows progress bar if true.
#'
#' @param promoteLinSens Promote solved linear compartment systems to
#'     sensitivity-based solutions.
#'
#' @param unknownDerivatives When handling derivatives from unknown
#'     functions, the translator will translate into different types
#'     of numeric derivatives.  The currently supported methods are:
#'
#'     - `forward` for forward differences
#'     - `central` for central differences
#'     - `error` for throwing an error for unknown derivatives
#'
#' @param parent is the parent environment to look for R-based user functions
#'
#' @return An rxode2 symengine environment
#' @author Matthew L. Fidler
#' @export
rxToSE <- function(x, envir = NULL, progress = FALSE,
                   promoteLinSens = TRUE, parent = parent.frame()) {
  .udfEnvSet(parent)
  .rxToSE.envir$parent <- parent
  .rxToSElinCmt$linCmt <- NULL # no linCmt() found
  if (exists("t", envir=.rxToSElinCmt, inherits=FALSE)) rm("t", envir=.rxToSElinCmt)
  assignInMyNamespace(".promoteLinB", promoteLinSens)
  assignInMyNamespace(".rxIsLhs", FALSE)
  assignInMyNamespace(".rxLastAssignedDdt", "")
  if (is(substitute(x), "character")) {
    force(x)
  } else if (is(substitute(x), "{")) {
    x <- deparse1(substitute(x))
    if (x[1] == "{") {
      x <- x[-1]
      x <- x[-length(x)]
    }
    x <- paste(x, collapse = "\n")
  } else {
    .xc <- as.character(substitute(x))
    x <- substitute(x)
    if (length(.xc) == 1) {
      .found <- FALSE
      .frames <- seq(1, sys.nframe())
      .frames <- .frames[.frames != 0]
      for (.f in .frames) {
        .env <- parent.frame(.f)
        if (exists(.xc, envir = .env)) {
          .val2 <- try(get(.xc, envir = .env), silent = TRUE)
          if (inherits(.val2, "character")) {
            .val2 <- eval(parse(text = paste0("quote({", .val2, "})")))
            return(.rxToSE(.val2, envir=envir, progress=progress))
          } else if (inherits(.val2, "numeric") || inherits(.val2, "integer")) {
            return(sprintf("%s", .val2))
          }
        }
      }
    }
    return(.rxToSE(x, envir=envir, progress=progress))
  }
  return(.rxToSE(eval(parse(text = paste0("quote({", x, "})"))), envir, progress))
}

## adapted from URLencode
.rxStrEncode <- function (str) {
  paste0("rxQ__",
         vapply(str, function(str) {
           OK <- "[^ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789]"
           x <- strsplit(str, "")[[1L]]
           z <- grep(OK, x)
           if (length(z)) {
             y <- vapply(x[z], function(x)
               paste0("_", toupper(as.character(charToRaw(x))),
                      collapse = ""), "")
             x[z] <- y
           }
           paste(x, collapse = "")
         }, character(1), USE.NAMES = FALSE),
         "__rxQ")
}

.rxStrDecode <- function(x) {
  .nchr <- nchar(x)
  if (.nchr > 10) {
    if (substr(x, 1, 5) == "rxQ__") {
      .x <- charToRaw(substr(x, 6, .nchr - 5))
      .pc <- charToRaw("_")
      .out <- raw(0L)
      .i <- 1L
      while (.i <= length(.x)) {
        if (.x[.i] != .pc) {
          .out <- c(.out, .x[.i])
          .i <- .i + 1L
        }
        else {
          .y <- as.integer(.x[.i + 1L:2L])
          .y[.y > 96L] <- .y[.y > 96L] - 32L
          .y[.y > 57L] <- .y[.y > 57L] - 7L
          .y <- sum((.y - 48L) * c(16L, 1L))
          .out <- c(.out, as.raw(as.character(.y)))
          .i <- .i + 3L
        }
      }
      return(rawToChar(.out))
    }
  }
  return(x)
}


.rxChrToSym <- function(x) {
  str2lang(.rxStrEncode(x))
}

.rxRepRxQ <- function(x) {
  .nchr <- nchar(x)
  if (.nchr > 10) {
    if (substr(x, 1, 5) == "rxQ__") {
      return(deparse1(.rxStrDecode(x)))
    }
  }
  return(x)
}

.rxToSEDualVarFunction <- c("tlast", "tlast0", "tad", "tad0", "tafd", "tafd0",
                            "tfirst", "tfirst0", "dose", "podo", "dose0", "podo0",
                            "dosenum", "dosenum0")

#' Change rxode2 linCmt states to `linCmtA/B` syntax
#'
#' @param x character expression
#' @param envir Current evaluating environment
#' @param progress Progress information
#' @param isEnvir Tell if this is an environment parse
#' @return Changed expression (as character), or NULL if not
#' @author Matthew L. Fidler
#' @noRd
.rxToSEtoLinCmt <- function(x, envir=NULL, progress=FALSE, isEnv=TRUE) {
  if (is.null(.rxToSElinCmt$linCmt)) {
    return(NULL)
  }
  .linCmtB <- .rxToSElinCmt$linCmt == "linCmtB"
  .oral0 <- as.numeric(paste(.rxToSElinCmt$oral0)) == 1
  .ncmt <- as.numeric(paste(.rxToSElinCmt$ncmt))
  .which1 <- NA_integer_
  if (.oral0) {
    if (.ncmt == 1) {
      if (x == "depot") {
        .which1 <- 0L
      } else if (x == "central") {
        .which1 <- 1L
      }
      # The sensitivity equations
      if (.linCmtB) {
        if (x == "rx__sens_central_BY_p1") {
          .which1 <- 2L
        } else if (x == "rx__sens_central_BY_v1") {
          .which1 <- 3L
        } else if (x == "rx__sens_central_BY_ka") {
          .which1 <- 4L
        } else if (x == "rx__sens_depot_BY_ka") {
          .which1 <- 5L
        }
      }
    } else if (.ncmt == 2) {
      if (x == "depot") {
        .which1 <- 0L
      } else if (x == "central") {
        .which1 <- 1L
      } else if (x == "peripheral1") {
        .which1 <- 2L
      }
      if (.linCmtB) {
        if (x == "rx__sens_central_BY_p1") {
          .which1 <- 3L
        } else if (x == "rx__sens_central_BY_v1") {
          .which1 <- 4L
        } else if (x == "rx__sens_central_BY_p2") {
          .which1 <- 5L
        } else if (x == "rx__sens_central_BY_p3") {
          .which1 <- 6L
        } else if (x == "rx__sens_central_BY_ka") {
          .which1 <- 7L
        } else if (x == "rx__sens_peripheral1_BY_p1") {
          .which1 <- 8L
        } else if (x == "rx__sens_peripheral1_BY_v1") {
          .which1 <- 9L
        } else if (x == "rx__sens_peripheral1_BY_p2") {
          .which1 <- 10L
        } else if (x == "rx__sens_peripheral1_BY_p3") {
          .which1 <- 11L
        } else if (x == "rx__sens_peripheral1_BY_ka") {
          .which1 <- 12L
        } else if (x == "rx__sens_depot_BY_ka") {
          .which1 <- 13L
        }
      }
    } else {
      if (x == "depot") {
        .which1 <- 0L
      } else if (x == "central") {
        .which1 <- 1L
      } else if (x == "peripheral1") {
        .which1 <- 2L
      } else if (x == "peripheral2") {
        .which1 <- 3L
      }
      if (.linCmtB) {
        if (x == "rx__sens_central_BY_p1") {
          .which1 <- 4L
        } else if (x == "rx__sens_central_BY_v1") {
          .which1 <- 5L
        } else if (x == "rx__sens_central_BY_p2") {
          .which1 <- 6L
        } else if (x == "rx__sens_central_BY_p3") {
          .which1 <- 7L
        } else if (x == "rx__sens_central_BY_p4") {
          .which1 <- 8L
        } else if (x == "rx__sens_central_BY_p5") {
          .which1 <- 9L
        } else if (x == "rx__sens_central_BY_ka") {
          .which1 <- 10L
        } else if (x == "rx__sens_peripheral1_BY_p1") {
          .which1 <- 11L
        } else if (x == "rx__sens_peripheral1_BY_v1") {
          .which1 <- 12L
        } else if (x == "rx__sens_peripheral1_BY_p2") {
          .which1 <- 13L
        } else if (x == "rx__sens_peripheral1_BY_p3") {
          .which1 <- 14L
        } else if (x == "rx__sens_peripheral1_BY_p4") {
          .which1 <- 15L
        } else if (x == "rx__sens_peripheral1_BY_p5") {
          .which1 <- 16L
        } else if (x == "rx__sens_peripheral1_BY_ka") {
          .which1 <- 17L
        } else if (x == "rx__sens_peripheral2_BY_p1") {
          .which1<- 18L
        } else if (x == "rx__sens_peripheral2_BY_v1") {
          .which1 <- 19L
        } else if (x == "rx__sens_peripheral2_BY_p2") {
          .which1 <- 20L
        } else if (x == "rx__sens_peripheral2_BY_p3") {
          .which1 <- 21L
        } else if (x == "rx__sens_peripheral2_BY_p4") {
          .which1 <- 22L
        } else if (x == "rx__sens_peripheral2_BY_p5") {
          .which1 <- 23L
        } else if (x == "rx__sens_peripheral2_BY_ka") {
          .which1 <- 24L
        } else if (x == "rx__sens_depot_BY_ka") {
          .which1 <- 25L
        }
      }
    }
  } else {
    if (.ncmt == 1) {
      if (x == "central") {
        .which1 <- 0L
      }
      if (.linCmtB) {
        if (x == "rx__sens_central_BY_p1") {
          .which1 <- 1L
        } else if (x == "rx__sens_central_BY_v1") {
          .which1 <- 2L
        }
      }
    } else if (.ncmt == 2) {
      if (x == "central") {
        .which1 <- 0L
      } else if (x == "peripheral1") {
        .which1 <- 1L
      }
      if (.linCmtB) {
        if (x == "rx__sens_central_BY_p1") {
          .which1 <- 2L
        } else if (x == "rx__sens_central_BY_v1") {
          .which1 <- 3L
        } else if (x == "rx__sens_central_BY_p2") {
          .which1 <- 4L
        } else if (x == "rx__sens_central_BY_p3") {
          .which1 <- 5L
        } else if (x == "rx__sens_peripheral1_BY_p1") {
          .which1 <- 6L
        } else if (x == "rx__sens_peripheral1_BY_v1") {
          .which1 <- 7L
        } else if (x == "rx__sens_peripheral1_BY_p2") {
          .which1 <- 8L
        } else if (x == "rx__sens_peripheral1_BY_p3") {
          .which1 <- 9L
        }
      }
    } else if (.ncmt == 3) {
      if (x == "central") {
        .which1 <- 0L
      } else if (x == "peripheral1") {
        .which1 <- 1L
      } else if (x == "peripheral2") {
        .which1 <- 2L
      }
      if (.linCmtB) {
        if (x == "rx__sens_central_BY_p1") {
          .which1 <- 3L
        } else if (x == "rx__sens_central_BY_v1") {
          .which1 <- 4L
        } else if (x == "rx__sens_central_BY_p2") {
          .which1 <- 5L
        } else if (x == "rx__sens_central_BY_p3") {
          .which1 <- 6L
        } else if (x == "rx__sens_central_BY_p4") {
          .which1 <- 7L
        } else if (x == "rx__sens_central_BY_p5") {
          .which1 <- 8L
        } else if (x == "rx__sens_peripheral1_BY_p1") {
          .which1 <- 9L
        } else if (x == "rx__sens_peripheral1_BY_v1") {
          .which1 <- 10L
        } else if (x == "rx__sens_peripheral1_BY_p2") {
          .which1 <- 11L
        } else if (x == "rx__sens_peripheral1_BY_p3") {
          .which1 <- 12L
        } else if (x == "rx__sens_peripheral1_BY_p4") {
          .which1 <- 13L
        } else if (x == "rx__sens_peripheral1_BY_p5") {
          .which1 <- 14L
        } else if (x == "rx__sens_peripheral2_BY_p1") {
          .which1 <- 15L
        } else if (x == "rx__sens_peripheral2_BY_v1") {
          .which1 <- 16L
        } else if (x == "rx__sens_peripheral2_BY_p2") {
          .which1 <- 17L
        } else if (x == "rx__sens_peripheral2_BY_p3") {
          .which1 <- 18L
        } else if (x == "rx__sens_peripheral2_BY_p4") {
          .which1 <- 19L
        } else if (x == "rx__sens_peripheral2_BY_p5") {
          .which1 <- 20L
        }
      }
    }
  }
  if (is.na(.which1)) {
    return(NULL)
  }
  if (!.linCmtB) {
    # linCmtA(rx__PTR__, t, linCmt, ncmt, oral0, which, trans,
    #         p1, v1, p2, p3, p4, p5, ka)
    .rxToSElinCmt$which <- .which1
    return(sprintf("linCmtA(rx__PTR__, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)",
                   .rxToSElinCmt$t, .rxToSElinCmt$linCmtN, .rxToSElinCmt$ncmt,
                   .rxToSElinCmt$oral0, .rxToSElinCmt$which,
                   .rxToSElinCmt$trans, .rxToSElinCmt$p1,
                   .rxToSElinCmt$v1, .rxToSElinCmt$p2, .rxToSElinCmt$p3,
                   .rxToSElinCmt$p4, .rxToSElinCmt$p5,
                   .rxToSElinCmt$ka))
  } else {
    .rxToSElinCmt$which1 <- .which1
    .rxToSElinCmt$which2 <- -2L
    return(sprintf("linCmtB(rx__PTR__, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)",
                   .rxToSElinCmt$t,
                   .rxToSElinCmt$linCmtN,
                   .rxToSElinCmt$ncmt,
                   .rxToSElinCmt$oral0,
                   .rxToSElinCmt$which1,
                   .rxToSElinCmt$which2,
                   .rxToSElinCmt$trans,
                   .rxToSElinCmt$p1,
                   .rxToSElinCmt$v1,
                   .rxToSElinCmt$p2,
                   .rxToSElinCmt$p3,
                   .rxToSElinCmt$p4,
                   .rxToSElinCmt$p5,
                   .rxToSElinCmt$ka))
  }
}

#' Change rxode2 syntax to symengine syntax for symbols and numbers
#'
#' @param x expression
#' @param envir Current evaluating environment
#' @param progress Progress information
#' @param isEnvir Tell if this is an environment parse
#' @return Changed expression (as character)
#' @author Matthew L. Fidler
#' @noRd
.rxToSENameOrAtomic <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  .cnst <- names(.rxSEreserved)
  if (is.character(x)) {
    .ret <- .rxChrToSym(x)
    .ret2 <- as.character(.ret)
    if (isEnv) {
      .ret2 <- as.character(.ret)
      assign(.ret2, symengine::Symbol(.ret2), envir = envir)
    }
    return(.ret)
  } else {
    .ret <- as.character(x)
    if (!.rxIsLhs && .ret %in% .rxToSEDualVarFunction) {
      ## A bare dose-history name (tad, dosenum, tlast, ...) used as a *value*
      ## expands to its functional form (e.g. tad -> (t-tlast())).  When the same
      ## name is an assignment *target* (e.g. `tad = tad()`, which FOCEi's FD model
      ## emits) it must stay a plain symbol -- otherwise the lhs is rewritten into
      ## the function body and produces an invalid assignment like
      ## `(t-tlast()) = t-tlast()`, which breaks analytic Jacobian (calcJac)
      ## generation for the model (notably delay()/DDE models, which force the
      ## Jacobian-requiring dense solver).
      .ret <- paste0(.ret, "()")
      .ret <- rxToSE(.ret)
      return(.ret)
    } else if (any(.ret == .cnst)) {
      .ret <- paste0("rx_SymPy_Res_", .ret)
      if (isEnv && is.name(x)) {
        if (substr(x, 1, 1) != ".") {
          if (!exists(.ret, envir = envir)) {
            assign(.ret, symengine::Symbol(.ret), envir = envir)
          }
        }
      }
      return(.ret)
    } else {
      .lin <- .rxToSEtoLinCmt(.ret, envir=envir, progress=progress, isEnv=isEnv)
      if (!is.null(.lin)) {
        return(.lin)
      }
      .ret0 <- .rxSEcnt[.ret]
      if (is.na(.ret0)) {
        if (isEnv && is.name(x)) {
          ## message(.ret)
          if (substr(x, 1, 1) != ".") {
            if (!exists(.ret, envir = envir) && is.name(x)) {
              assign(.ret, symengine::Symbol(.ret), envir = envir)
            }
          }
        }
        return(.ret)
      }
      return(setNames(.ret0, NULL))
    }
  }
}

.rxToSECurlyBrace <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  .x2 <- x[-1]
  if (progress) {
    rxProgress(length(.x2))
    on.exit({
      rxProgressAbort()
    })
    .ret <- paste(lapply(.x2, function(x) {
      rxTick()
      .rxToSE(x, envir = envir)
    }), collapse = "\n")
    rxProgressStop()
  } else {
    .ret <- paste(lapply(.x2, .rxToSE, envir = envir),
                  collapse = "\n")
  }
  ## Assign and evaluate deferred items.
  if (isEnv) {
    for (.var in names(envir$..ddt..)) {
      .expr <- envir$..ddt..[[.var]]
      .expr <- eval(parse(text = .expr))
      assign(.var, .expr, envir = envir)
      .rx <- paste0(
        rxFromSE(.var), "=",
        rxFromSE(.expr)
      )
      assign("..ddt", c(envir$..ddt, .rx),
             envir = envir
             )
    }
    for (.var in names(envir$..sens0..)) {
      .expr <- envir$..sens0..[[.var]]
      .expr <- eval(parse(text = .expr))
      assign(.var, .expr, envir = envir)
      .rx <- paste0(
        rxFromSE(.var), "=",
        rxFromSE(.expr))
      assign("..sens0", c(envir$..sens0, .rx),
             envir = envir)
    }
    for (.var in names(envir$..jac0..)) {
      .expr <- envir$..jac0..[[.var]]
      .expr <- eval(parse(text = .expr))
      assign(.var, .expr, envir = envir)
      .rx <- paste0(
        rxFromSE(.var), "=",
        rxFromSE(.expr))
      assign("..jac0", c(envir$..jac0, .rx),
             envir = envir)
    }
  }
  return(.ret)
}

.rxLastAssignedDdt <- ""
.rxIsLhs <- FALSE

.rxToSEArithmeticOperators <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  if (length(x) == 3) {
    if (identical(x[[1]], quote(`/`))) {
      .x2 <- x[[2]]
      .x3 <- x[[3]]
      ## df(%s)/dy(%s)
      if (identical(.x2, quote(`d`)) &&
            identical(.x3[[1]], quote(`dt`))) {
        if (length(.x3[[2]]) == 1) {
          .state <- as.character(.x3[[2]]) # .rxToSE(.x3[[2]], envir = envir)
        } else {
          .state <- .rxToSE(.x3[[2]], envir = envir)
        }
        if (.rxIsLhs) {
          assignInMyNamespace(".rxLastAssignedDdt", .state)
        }
        return(paste0("rx__d_dt_", .state, "__"))
      } else {
        if (length(.x2) == 2 && length(.x3) == 2) {
          if (identical(.x2[[1]], quote(`df`)) &&
                identical(.x3[[1]], quote(`dy`))) {
            if (length(.x2[[2]]) == 1) {
              .state <- as.character(.x2[[2]])
            } else {
              .state <- .rxToSE(.x2[[2]], envir = envir)
            }
            if (length(.x3[[2]]) == 1) {
              .var <- as.character(.x3[[2]])
            } else {
              .var <- .rxToSE(.x3[[2]], envir = envir)
            }
            return(paste0(
              "rx__df_", .state,
              "_dy_", .var, "__"
            ))
          }
        }
        .ret <- paste0(
          .rxToSE(.x2, envir = envir),
          as.character(x[[1]]),
          .rxToSE(.x3, envir = envir)
        )
      }
    } else {
      .ret <- paste0(
        .rxToSE(x[[2]], envir = envir),
        as.character(x[[1]]),
        .rxToSE(x[[3]], envir = envir)
      )
    }
    return(.ret)
  } else {
    ## Unary Operators
    return(paste(
      as.character(x[[1]]),
      .rxToSE(x[[2]], envir = envir)
    ))
  }
}

.rxToSEAssignOperators <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  assignInMyNamespace(".rxIsLhs", TRUE)
  .var <- .rxToSE(x[[2]], envir = envir)
  assignInMyNamespace(".rxIsLhs", FALSE)
  .isNum <- FALSE
  if (isEnv) {
    if (length(x[[2]]) == 2) {
      if (as.character(x[[2]][[1]]) == "mtime") {
        .mtimeVar <- .rxToSE(x[[2]][[2]], envir = envir)
        envir$..mtimeVars <- unique(c(.mtimeVar, envir$..mtimeVars))
        return(invisible(NULL))
      }
      if (as.character(x[[2]][[1]]) == "indLin") {
        ## indLin(state) <- forcing : store the forcing as a per-state symengine
        ## variable (rx__indLinForce_<state>__) so .rxInjectMatExpOdes() can add
        ## it to the matrix-derived d/dt().  The LHS itself has no symengine
        ## variable (.rxToSE returns "" for indLin), so it must be handled here
        ## before the generic assignment below.  A named variable (not a list)
        ## is used so symengine's masked `[[<-` cannot coerce the container.
        .indLinState <- as.character(x[[2]][[2]])
        .force <- eval(parse(text = paste0(
          "with(envir,", .rxToSE(x[[3]], envir = envir), ")")))
        assign(paste0("rx__indLinForce_", .indLinState, "__"), .force,
               envir = envir)
        return(invisible(NULL))
      }
      if (any(as.character(x[[2]][[1]]) == c("alag", "lag", "F", "f", "rate", "dur"))) {
        envir$..eventVars <- unique(c(.var, envir$..eventVars))
      }
    }
    if (length(x[[2]]) == 3 && as.character(x[[2]][[1]]) == "past") {
      ## past(state, tau) <- expr : non-constant delay() pre-history.  Store the
      ## RHS (rx__past_<state>__, for sensitivity histories) and the delay tau
      ## text (rx__pastTau_<state>__) as per-state symengine env variables, then
      ## return: the line is retained verbatim in the normalized model and must
      ## not enter Jacobian/sensitivity differentiation here (state is the delay
      ## history, differentiated only via its own past() expression).
      .pastState <- as.character(x[[2]][[2]])
      ## Store the history RHS and tau as plain rxode2-syntax TEXT (not symengine
      ## Basics): reading Basics back out of the model env in .rxDelaySensAugment
      ## is subject to symengine's masked get/[[.  The RHS is re-parsed there with
      ## eval(parse(text=), envir=model) for differentiation (the safe pattern).
      assign(paste0("rx__pastRhs_", .pastState, "__"),
             paste(deparse(x[[3]]), collapse = " "), envir = envir)
      assign(paste0("rx__pastTau_", .pastState, "__"),
             paste(deparse(x[[2]][[3]]), collapse = " "), envir = envir)
      return(invisible(NULL))
    }
  }
  if (inherits(x[[3]], "numeric") || inherits(x[[3]], "integer")) {
    .isNum <- TRUE
    if (isEnv) {
      if (envir$..doConst) {
        assign(.var, x[[3]], envir = envir)
      }
    }
    .expr <- x[[3]]
  }
  if (isEnv) {
    .expr <- paste0(
      "with(envir,",
      .rxToSE(x[[3]],
              envir = envir
              ), ")"
    )
    # A variable that is referenced inside a history function (lag/lead/diff/...)
    # must be emitted as an lhs and bound as a bare symbol, so downstream uses
    # (including the history reference) stay symbolic and the defining line is not
    # inlined or dead-code eliminated.
    if (!is.null(envir$..laggedVars) && any(.var == envir$..laggedVars) &&
          !identical(x[[1]], quote(`~`))) {
      .val <- try(eval(parse(text = .expr)), silent = TRUE)
      if (!inherits(.val, "try-error")) {
        .rx <- paste0(rxFromSE(.var), "=", rxFromSE(.val))
        assign("..lhs", c(envir$..lhs, .rx), envir = envir)
        assign(.var, symengine::S(.var), envir = envir)
        return(invisible(NULL))
      }
    }
    if (regexpr(rex::rex(or(
      .regRate,
      .regDur,
      .regLag,
      .regF,
      regIni0,
      regDDt
    )), .var) != -1) {
      ## A d/dt(state) whose state is read by delay() is a genuine ODE, not a
      ## derivable sensitivity -- keep it in ..ddt.. even if its name matches the
      ## sensitivity convention.
      .ddtState <- if (regexpr(rex::rex(regDDt), .var) != -1) {
        sub(rex::rex(start, "rx__d_dt_", capture(anything), "__", end), "\\1", .var)
      } else {
        NA_character_
      }
      .isDelayedOde <- !is.na(.ddtState) &&
        !is.null(envir$..delayedStates) && (.ddtState %in% envir$..delayedStates)
      if (!.isDelayedOde && regexpr(
        rex::rex(or(regSens, regSensEtaTheta)),
        .var
      ) != -1) {
        .lst <- get("..sens0..", envir = envir)
        .lst[[.var]] <- .expr
        assign("..sens0..", .lst, envir = envir)
      } else {
        .lst <- get("..ddt..", envir = envir)
        .lst[[.var]] <- .expr
        assign("..ddt..", .lst, envir = envir)
      }
    } else if (regexpr(rex::rex(or(
      regDfDy,
      regDfDyTh
    )), .var) != -1) {
      .lst <- get("..jac0..", envir = envir)
      .lst[[.var]] <- .expr
      assign("..jac0..", .lst, envir = envir)
    } else if (!identical(x[[1]], quote(`~`))) {
      if (is.call(x[[3]]) && identical(as.character(x[[3]][[1]]), "linCmt")) {
        assign(.var, symengine::S(.var), envir = envir)
        .name <- rxFromSE(.var)
        .rx <- paste0(.name, "=", paste(deparse(x[[3]]), collapse = ""))
        if (regexpr("^(nlmixr|rx)_", .var) == -1) {
          assign("..lhs", c(envir$..lhs, .rx), envir = envir)
        }
        return(invisible(NULL))
      }
      .expr <- try(eval(parse(text = .expr)), silent = TRUE)
      .isNum <- (inherits(.expr, "numeric") || inherits(.expr, "integer"))
      if ((.isNum && envir$..doConst) ||
            (!.isNum)) {
        assign(.var, .expr, envir = envir)
      }
      .name <- rxFromSE(.var)
      .rx <- paste0(
        .name, "=",
        rxFromSE(.expr)
      )
      if (regexpr("^(nlmixr|rx)_", .var) == -1) {
        if (.isNum) {
          names(.rx) <- .name
          assign("..lhs0", c(envir$..lhs0, .rx),
                 envir = envir
                 )
        } else {
          if (any(names(envir$..lhs0) == .name)) {
            .tmp <- envir$..lhs0
            .tmp <- .tmp[names(.tmp) != .name]
            assign("..lhs0", .tmp, envir = envir)
          }
          assign("..lhs", c(envir$..lhs, .rx),
                 envir = envir
                 )
        }
      }
    } else {
      # Suppressed (`~`) intermediate: never an output column, it exists only to
      # be substituted into downstream expressions.  Always bind it so references
      # resolve -- even a literal constant (e.g. `rx_expr_3 ~ 0`) with
      # doConst=FALSE, which otherwise leaks as an undefined phantom parameter.
      .expr <- eval(parse(text = .expr))
      assign(.var, .expr, envir = envir)
      .rx <- paste0(
        rxFromSE(.var), "=",
        rxFromSE(.expr)
      )
    }
  }
}

.rxToSESquareBracket <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  .type <- toupper(as.character(x[[2]]))
  if (any(.type == c("THETA", "ETA"))) {
    if (is.numeric(x[[3]])) {
      .num <- x[[3]]
      if (round(.num) == .num) {
        if (.num > 0) {
          if (isEnv) {
            if (.type == "THETA") {
              if (exists("..maxTheta", envir = envir)) {
                .m <- get("..maxTheta", envir = envir)
              } else {
                .m <- 0
              }
              .m <- max(.m, .num)
              .funs <- envir$..curCall
              .funs <- .funs[.funs != ""]
              if (length(.funs) > 0) {
                .funs <- paste(.funs, collapse = ".")
                envir$..extraTheta[[.funs]] <-
                  unique(c(
                    envir$..extraTheta[[.funs]],
                    .num
                  ))
              }
              assign("..maxTheta", .m, envir = envir)
            } else {
              if (exists("..maxEta", envir = envir)) {
                .m <- get("..maxEta", envir = envir)
              } else {
                .m <- 0
              }
              .m <- max(.m, .num)
              .funs <- envir$..curCall
              .funs <- .funs[.funs != ""]
              if (length(.funs) > 0) {
                .funs <- paste(.funs, collapse = ".")
                envir$..extraEta[[.funs]] <-
                  unique(c(
                    envir$..extraEta[[.funs]],
                    .num
                  ))
              }
              assign("..maxEta", .m, envir = envir)
            }
          }
          return(paste0(.type, "_", .num, "_"))
        } else {
          stop("only 'THETA[#]' or 'ETA[#]' are supported", call. = FALSE)
        }
      } else {
        stop("only 'THETA[#]' or 'ETA[#]' are supported", call. = FALSE)
      }
    } else {
      stop("only 'THETA[#]' or 'ETA[#]' are supported", call. = FALSE)
    }
  } else {
    stop("only 'THETA[#]' or 'ETA[#]' are supported", call. = FALSE)
  }
}

.rxToSETad0 <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  .len <- length(x)
  if (.len == 1L) {
  } else if (.len == 2L) {
    if (length(x[[2]]) != 1) {
      stop(as.character(x[[1]]), "() must be used with a state", call. = FALSE)
    }
    return(paste0("(t-tlast0(", as.character(x[[2]]), "))"))
  } else {
    stop(as.character(x[[1]]), "() can have 0-1 arguments", call. = FALSE)
  }
  return(paste0("(t-tlast0())"))
}

.rxToSETad <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  .len <- length(x)
  if (.len == 1L) {
  } else if (.len == 2L) {
    if (length(x[[2]]) != 1) {
      stop(as.character(x[[1]]), "() must be used with a state", call. = FALSE)
    }
    return(paste0("(t-tlast(", as.character(x[[2]]), "))"))
  } else {
    stop(as.character(x[[1]]), "() can have 0-1 arguments", call. = FALSE)
  }
  return(paste0("(t-tlast())"))
}

.rxToSELagOrLead <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  .len <- length(x)
  .fun <- as.character(x[[1]])
  # lag()/lead() reference the previous/next record value of a *variable*, so the
  # variable must stay a symbol.  In a symengine environment the assignment eval
  # (`with(envir, ...)`) would otherwise inline an lhs variable's definition and
  # break the call, so wrap the variable in symengine::S("var") to keep it a raw
  # symbol; it round-trips back to the bare name in rxFromSE().
  .vref <- function(v) if (isEnv) paste0("symengine::S(\"", v, "\")") else v
  if (.len == 1L) {
    stop(.fun, "() takes 1-2 arguments")
  } else if (.len == 2L) {
    if (length(x[[2]]) != 1) {
      stop(.fun, "() must be used with a variable", call. = FALSE)
    }
    return(paste0(.fun, "(", .vref(as.character(x[[2]])), ")"))
  } else if (.len == 3L) {
    if (length(x[[2]]) != 1) {
      stop(.fun, "() must be used with a variable", call. = FALSE)
    }
    if (length(x[[3]]) != 1) {
      stop(.fun, "(", as.character(x[[2]]), ", #) must have an integer for the number of lagged doses", call. = FALSE)
    }
    if (regexpr(rex::rex(maybe(one_of("-", "+")), regDecimalint), as.character(x[[3]]), perl = TRUE) == -1) {
      stop(.fun, "(", as.character(x[[2]]), ", #) must have an integer for the number of lagged doses", call. = FALSE)
    }
    return(paste0(.fun, "(", .vref(as.character(x[[2]])), ", ", as.character(x[[3]]), ")"))
  } else {
    stop(as.character(x[[1]]), "() can have 0-1 arguments", call. = FALSE)
  }
  return(paste0("(t-tfirst())"))
}

.rxToSETlastOrTafd <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  .len <- length(x)
  if (.len == 1L) {
  } else if (.len == 2L) {
    if (length(x[[2]]) != 1) {
      stop(as.character(x[[1]]), "() must be used with a state", call. = FALSE)
    }
    return(paste0("(t-tfirst(", as.character(x[[2]]), "))"))
  } else {
    stop(as.character(x[[1]]), "() can have 0-1 arguments", call. = FALSE)
  }
  return(paste0("(t-tfirst())"))
}

.rxToSETlastOrTafd0 <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  .len <- length(x)
  if (.len == 1L) {
  } else if (.len == 2L) {
    if (length(x[[2]]) != 1) {
      stop(as.character(x[[1]]), "() must be used with a state", call. = FALSE)
    }
    return(paste0("(t-tfirst0(", as.character(x[[2]]), "))"))
  } else {
    stop(as.character(x[[1]]), "() can have 0-1 arguments", call. = FALSE)
  }
  return(paste0("(t-tfirst0())"))
}

.rxToSETlastOrTfirst <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  .len <- length(x)
  if (.len == 1L) {
    if (identical(x[[1]], quote(`podo`))) {
      return(paste0("podo(", .rxLastAssignedDdt, ")"))
    }
    if (identical(x[[1]], quote(`podo0`))) {
      return(paste0("podo0(", .rxLastAssignedDdt, ")"))
    }
  } else if (.len == 2L) {
    if (length(x[[2]]) != 1) {
      stop(as.character(x[[1]]), "() must be used with a state", call. = FALSE)
    }
    return(paste0(as.character(x[[1]]), "(", as.character(x[[2]]), ")"))
  } else {
    stop(as.character(x[[1]]), "() can have 0-1 arguments", call. = FALSE)
  }
  return(paste0(as.character(x[[1]]), "()"))
}

.rxToSEDelay <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  # delay(state, T) is the value of `state` at the past time t - T.  It is kept
  # as a function call so it round-trips back to delay() in any generated code,
  # and its derivative with respect to any variable is zero (a delayed/past
  # state does not depend on the current state); see the lag/lead handling in
  # rxFromSE().
  if (length(x) != 3L) {
    stop("'delay' takes 2 arguments 'delay(state, T)'", call. = FALSE)
  }
  if (length(x[[2]]) != 1) {
    stop("the first argument to 'delay' must be an ODE state 'delay(state, T)'", call. = FALSE)
  }
  .t <- .rxToSE(x[[3]], envir = envir)
  ## also handles rxDelayD() (the delayed time-derivative); keep the function
  ## name so it round-trips and differentiates to 0 like delay().
  paste0(as.character(x[[1]]), "(", as.character(x[[2]]), ", ", .t, ")")
}

.rxToSEPsigamma <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  if (length(x == 3)) {
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..curCall <- c(envir$..curCall, "psigamma")
    }
    .a <- .rxToSE(x[[2]], envir = envir)
    .b <- .rxToSE(x[[3]], envir = envir)
    if (isEnv) envir$..curCall <- .lastCall
    return(paste0("polygamma(", .b, ",", .a, ")"))
  } else {
    stop("'psigamma' takes 2 arguments", call. = FALSE)
  }
}

.rxToSELog1pmx <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  if (length(x == 2)) {
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..curCall <- c(envir$..curCall, "log")
    }
    .a <- .rxToSE(x[[2]], envir = envir)
    if (isEnv) envir$..curCall <- .lastCall
    return(paste0("(log(1+", .a, ")-(", .a, "))"))
  } else {
    stop("'log1pmx' only takes 1 argument", call. = FALSE)
  }
}

.rxToSEChoose <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  if (length(x) == 3) {
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..curCall <- c(envir$..curCall, "gamma")
    }
    .n <- .rxToSE(x[[2]], envir = envir)
    .k <- .rxToSE(x[[3]], envir = envir)
    if (isEnv) envir$..curCall <- .lastCall
    return(paste0(
      "gamma(", .n, "+1)/(gamma(",
      .k, "+1)*gamma(", .n, "-(", .k, ")+1))"
    ))
  } else {
    stop("'choose' takes 2 arguments", call. = FALSE)
  }
}

.rxToSELchoose <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  if (length(x) == 3) {
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..curCall <- c(envir$..curCall, "lgamma")
    }
    .n <- .rxToSE(x[[2]], envir = envir)
    .k <- .rxToSE(x[[3]], envir = envir)
    if (isEnv) envir$..curCall <- .lastCall
    return(paste0("(lgamma(", .n, "+1)-lgamma(", .k, "+1)-lgamma(", .n, "-(", .k, ")+1))"))
  } else {
    stop("'lchoose' takes 2 arguments", call. = FALSE)
  }
}

.rxToSEPnorm <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  if (length(x) == 4) {
    ## pnorm(q, mean, sd)
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..currCall <- c(envir$..curCall, "erf")
    }
    .q <- .rxToSE(x[[2]], envir = envir)
    .mean <- .rxToSE(x[[3]], envir = envir)
    .sd <- .rxToSE(x[[4]], envir = envir)
    return(paste0("0.5*(1+erf((((", .q, ")-(", .mean, "))/(", .sd, "))/sqrt(2)))"))
  } else if (length(x) == 3) {
    ## pnorm(q, mean)
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..currCall <- c(envir$..curCall, "erf")
    }
    .q <- .rxToSE(x[[2]], envir = envir)
    .mean <- .rxToSE(x[[3]], envir = envir)
    return(paste0("0.5*(1+erf((((", .q, ")-(", .mean, ")))/sqrt(2)))"))
  } else if (length(x) == 2) {
    ## pnorm(q)
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..currCall <- c(envir$..curCall, "erf")
    }
    .q <- .rxToSE(x[[2]], envir = envir)
    return(paste0("0.5*(1+erf((", .q, ")/sqrt(2)))"))
  } else {
    stop("'pnorm' can only take 1-3 arguments", call. = FALSE)
  }
}

.rxToSEd4GELU <- function(x, envir=NULL, progress=FALSE, isEnv=TRUE) {
  if (length(x) == 2) {
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..curCall <- c(envir$..curCall, "d4GELU")
    }
    .x <- .rxToSE(x[[2]], envir = envir)
    if (isEnv) envir$..curCall <- .lastCall
    return(
      paste0("exp(-(", .x, ")^2/2)*(7*(", .x, ")^2 - 4 - (", .x, ")^4)/sqrt(2*pi)")
    )
  } else {
    stop("'d4GELU' can only take 1 argument", call. = FALSE)
  }
}

.rxToSEd4softplus <- function(x, envir=NULL, progress=FALSE, isEnv=TRUE) {
  if (length(x) == 2) {
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..curCall <- c(envir$..curCall, "d4softplus")
    }
    .x <- .rxToSE(x[[2]], envir = envir)
    if (isEnv) envir$..curCall <- .lastCall
    .ex1 <- paste0("(1.0 + exp(-(", .x, ")))")
    return(
      paste0("6.0*exp(-3.0*(", .x, "))/((", .ex1,
             ")^4) - 6.0*exp(-2.0*(", .x, "))/((", .ex1,
             ")^3) + exp(-(", .x, "))/((", .ex1, ")^2)")
    )
  } else {
    stop("'d4softplus' can only take 1 argument", call. = FALSE)
  }
}

.rxToSEdSELU <- function(x, envir=NULL, progress=FALSE, isEnv=TRUE) {
  if (length(x) == 2) {
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..curCall <- c(envir$..curCall, "dSELU")
    }
    .x <- .rxToSE(x[[2]], envir = envir)
    if (isEnv) envir$..curCall <- .lastCall
    return(
      paste0("(rxGt(", .x, ", 0)*1.0507009873554804934193349852946 + 1.0507009873554804934193349852946*1.6732632423543772848170429916717*exp(", .x, ")*rxLeq(", .x, ", 0))")
    )
  } else {
    stop("'dSELU' can only take 1 argument", call. = FALSE)
  }

}

.rxToSEdSwish <- function(x, envir=NULL, progress=FALSE, isEnv=TRUE) {
  if (length(x) == 2) {
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..curCall <- c(envir$..curCall, "dSwish")
    }
    .x <- .rxToSE(x[[2]], envir = envir)
    if (isEnv) envir$..curCall <- .lastCall
    # x*exp(-x)/(1.0 + exp(-x))^2 + (1.0 + exp(-x))^(-1);
    return(
      paste0("((", .x, ")*exp(-(", .x, "))/(1.0 + exp(-(", .x,
             ")))^2 + 1.0/(1.0 + exp(-(", .x, ")))")
      )
  } else {
    stop("'dSwish' can only take 1 argument", call. = FALSE)
  }

}


# Convert the mix() expression to a if clause with mixest for symengine
.rxToSEMix <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  .expr <- vapply(seq_along(x), function(i) {
    if (i %% 2 == 0)  {
      # Noting that argument 1 is the function name,
      # The even arguments are the mixture values
      paste0("rxEq(mixest, ", i/2, ")*(", .rxToSE(x[[i]], envir = envir, progress = progress), ")")
    } else {
      ""
    }
  }, character(1L), USE.NAMES = FALSE)
  paste0("(", paste(.expr[nzchar(.expr)], collapse="+"), ")")
}

.rxToSETransit <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  if (length(x) == 4) {
    ## transit(n, mtt, bio)
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..curCall <- c(envir$..curCall, "lgamma")
    }
    .n <- .rxToSE(x[[2]], envir = envir)
    if (isEnv) {
      envir$..curCall <- .lastCall
      .lastCall <- envir$..curCall
      envir$..curCall <- c(envir$..curCall, "log")
    }
    .mtt <- .rxToSE(x[[3]], envir = envir)
    .bio <- .rxToSE(x[[4]], envir = envir)
    if (isEnv) envir$..curCall <- .lastCall
    return(paste0(
      "exp(log((", .bio, ")*(podo0(", .rxLastAssignedDdt, ")))+log(",
      .n, " + 1)-log(", .mtt, ")+(", .n,
      ")*((log(", .n, "+1)-log(", .mtt,
      "))+log(t-tlast0(", .rxLastAssignedDdt, ")))-((", .n, "+1)/(", .mtt,
      "))*(t-tlast0(", .rxLastAssignedDdt, "))-lgamma(1+", .n, "))"
    ))
  } else if (length(x) == 3) {
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..curCall <- c(envir$..curCall, "lgamma")
    }
    .n <- .rxToSE(x[[2]], envir = envir)
    .mtt <- .rxToSE(x[[3]], envir = envir)
    if (isEnv) envir$..curCall <- .lastCall
    return(paste0("exp(log(podo0(", .rxLastAssignedDdt, "))+(log(", .n, "+1)-log(", .mtt, "))+(", .n, ")*((log(", .n, "+1)-log(", .mtt, "))+ log(t-tlast0(", .rxLastAssignedDdt, ")))-((", .n, " + 1)/(", .mtt, "))*(t-tlast0(",.rxLastAssignedDdt, "))-lgamma(1+", .n, "))"))
  } else {
    stop("'transit' can only take 2-3 arguments", call. = FALSE)
  }
}

.rxToSEMax <- function(x, min=FALSE) {
  # Based on https://stackoverflow.com/questions/30738923/max-implemented-with-basic-operators
  if (length(x) == 0) return("")
  if (length(x) == 1) return(paste0("(", x[1], ")"))
  .x2 <- x[1:2]
  .xrest <- x[-(1:2)]
  .a <- paste0(.x2[1])
  .b <- paste0(.x2[2])
  .cmp <- ifelse(min, "rxLt", "rxGt")
  .av <- suppressWarnings(as.numeric(.x2[1]))
  .bv <- suppressWarnings(as.numeric(.x2[2]))
  if (identical(.av, 0)) {
    return(paste0("((", .b, ")*", .cmp, "(", .b, ",0))"))
  }
  if (identical(.bv, 0)) {
    return(paste0("((", .a, ")*", .cmp, "(", .a,",0))"))
  }
  .ret <- paste0("(((", .a, ")-(", .b, "))*", .cmp, "(", .a, ",", .b, ")+(", .b, "))")
  if (length(.xrest) == 0) return(.ret)
  return(.rxToSEMax(c(.ret, .xrest), min=min))
}

.rxToSElinCmt <- new.env(parent=emptyenv())
.rxToSElinCmt$linCmt <- NULL


#' This saves the linCmt information to the .rxToSElinCmt environment
#'
#' This is useful when trying to convert the depot and other states
#' defined by the linear compartment to `linCmtA` or `linCmtB` to the
#' function.
#'
#' @param expr language based expression to save information for.
#'
#' @return nothing, called for side effects
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.rxSEsaveLinCmt <- function(expr) {
  if (is.null(.rxToSElinCmt$linCmt)) stop(".rxToSElinCmt$linCmt is NULL")
  if (!exists("t", envir=.rxToSElinCmt)) {
    if (.rxToSElinCmt$linCmt == "linCmtA") {
      # [[1]] == linCmtA
      # [[2]] == rx__PTR__
      .rxToSElinCmt$t <- expr[[3]]
      .rxToSElinCmt$linCmtN <- expr[[4]]
      .rxToSElinCmt$ncmt <- expr[[5]]
      .rxToSElinCmt$oral0 <- expr[[6]]
      # [[7]] = which
      .rxToSElinCmt$trans <- expr[[8]]
      .rxToSElinCmt$p1 <- expr[[9]]
      .rxToSElinCmt$v1 <- expr[[10]]
      .rxToSElinCmt$p2 <- expr[[11]]
      .rxToSElinCmt$p3 <- expr[[12]]
      .rxToSElinCmt$p4 <- expr[[13]]
      .rxToSElinCmt$p5 <- expr[[14]]
      .rxToSElinCmt$ka <- expr[[15]]
      # linCmtA(rx__PTR__, t, linCmt, ncmt, oral0, which, trans,
      #         p1, v1, p2, p3, p4, p5, ka)
    } else {
      # linCmtB(rx__PTR__, t, linCmt, ncmt, oral0, which1, which2, trans,
      #         p1, v1, p2, p3, p4, p5, ka)
      # [[1]] == linCmtB
      # [[2]] == rx__PTR__
      .rxToSElinCmt$t <- expr[[3]]
      .rxToSElinCmt$linCmtN <- expr[[4]]
      .rxToSElinCmt$ncmt <- expr[[5]]
      .rxToSElinCmt$oral0 <- expr[[6]]
      # [[7]] = which1
      # [[8]] = which2
      .rxToSElinCmt$trans <- expr[[9]]
      .rxToSElinCmt$p1 <- expr[[10]]
      .rxToSElinCmt$v1 <- expr[[11]]
      .rxToSElinCmt$p2 <- expr[[12]]
      .rxToSElinCmt$p3 <- expr[[13]]
      .rxToSElinCmt$p4 <- expr[[14]]
      .rxToSElinCmt$p5 <- expr[[15]]
      .rxToSElinCmt$ka <- expr[[16]]
    }
  } else {
    if (.rxToSElinCmt$linCmt != expr[[1]]) {
      stop("linCmtA and linCmtB cannot be used in the same model", call. = FALSE)
    }
    if (.rxToSElinCmt$linCmt == "linCmtA") {
      # [[1]] == linCmtA
      # [[2]] == rx__PTR__
      err <- NULL
      if (!identical(.rxToSElinCmt$t, expr[[3]])) {
        err <- c(err, "t")
      }
      if (!identical(.rxToSElinCmt$linCmtN, expr[[4]])) {
        err <- c(err, "linCmtN")
      }
      if (!identical(.rxToSElinCmt$ncmt, expr[[5]])) {
        err <- c(err, "ncmt")
      }
      if (!identical(.rxToSElinCmt$oral0, expr[[6]])) {
        err <- c(err, "oral0")
      }
      if (!identical(.rxToSElinCmt$trans, expr[[8]])) {
        err <- c(err, "trans")
      }
      if (!identical(.rxToSElinCmt$p1, expr[[9]])) {
        err <- c(err, "p1")
      }
      if (!identical(.rxToSElinCmt$v1, expr[[10]])) {
        err <- c(err, "v1")
      }
      if (!identical(.rxToSElinCmt$p2, expr[[11]])) {
        err <- c(err, "p2")
      }
      if (!identical(.rxToSElinCmt$p3, expr[[12]])) {
        err <- c(err, "p3")
      }
      if (!identical(.rxToSElinCmt$p4, expr[[13]])) {
        err <- c(err, "p4")
      }
      if (!identical(.rxToSElinCmt$p5, expr[[14]])) {
        err <- c(err, "p5")
      }
      if (!identical(.rxToSElinCmt$ka, expr[[15]])) {
        err <- c(err, "ka")
      }
      if (!is.null(err)) {
        stop(paste0("linCmtA needs exactly the same arguments: ",
                    paste(err, collapse = ", ")),
             call. = FALSE)
      }
    } else {
      # linCmtB(rx__PTR__, t, linCmt, ncmt, oral0, which1, which2, trans,
      #         p1, v1, p2, p3, p4, p5, ka)
      # [[1]] == linCmtB
      # [[2]] == rx__PTR__
      err <- NULL
      if (!identical(.rxToSElinCmt$t, expr[[3]])) {
        err <- c(err, "t")
      }
      if (!identical(.rxToSElinCmt$linCmtN, expr[[4]])) {
        err <- c(err, "linCmtN")
      }
      if (!identical(.rxToSElinCmt$ncmt, expr[[5]])) {
        err <- c(err, "ncmt")
      }
      if (!identical(.rxToSElinCmt$oral0, expr[[6]])) {
        err <- c(err, "oral0")
      }
      if (!identical(.rxToSElinCmt$trans, expr[[9]])) {
        err <- c(err, "trans")
      }
      if (!identical(.rxToSElinCmt$p1, expr[[10]])) {
        err <- c(err, "p1")
      }
      if (!identical(.rxToSElinCmt$v1, expr[[11]])) {
        err <- c(err, "v1")
      }
      if (!identical(.rxToSElinCmt$p2, expr[[12]])) {
        err <- c(err, "p2")
      }
      if (!identical(.rxToSElinCmt$p3, expr[[13]])) {
        err <- c(err, "p3")
      }
      if (!identical(.rxToSElinCmt$p4, expr[[14]])) {
        err <- c(err, "p4")
      }
      if (!identical(.rxToSElinCmt$p5, expr[[15]])) {
        err <- c(err, "p5")
      }
      if (!identical(.rxToSElinCmt$ka, expr[[16]])) {
        err <- c(err, "ka")
      }
      if (!is.null(err)) {
        stop(paste0("linCmtB needs exactly the same arguments: ",
                    paste(err, collapse = ", ")),
             call. = FALSE)
      }
    }
  }
}

#' This is used converting the rxode2 functions to a symengine call
#'
#' @param x language object that is a call
#' @param envir environment that has information
#' @param progress boolean telling if the progress bar should be shown
#' @param isEnv boolean to indicate if the environment is a rxode2 environment
#' @return string representing the translation
#' @noRd
#' @author Matthew L. Fidler
.rxToSECall <- function(x, envir = NULL, progress = FALSE, isEnv=TRUE) {
  if (identical(x[[1]], quote(`(`))) {
    return(paste0("(", .rxToSE(x[[2]], envir = envir), ")"))
  } else if (identical(x[[1]], quote(`{`))) {
    return(.rxToSECurlyBrace(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`*`)) ||
               identical(x[[1]], quote(`^`)) ||
               identical(x[[1]], quote(`+`)) ||
               identical(x[[1]], quote(`-`)) ||
               identical(x[[1]], quote(`/`))) {
    return(.rxToSEArithmeticOperators(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`=`)) ||
               identical(x[[1]], quote(`<-`)) ||
               identical(x[[1]], quote(`~`))) {
    if (length(x[[2]]) == 2 &&
          identical(x[[2]][[1]], quote(`levels`))) {
      return("")
    }
    return(.rxToSEAssignOperators(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`[`))) {
    return(.rxToSESquareBracket(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`tad`))) {
    return(.rxToSETad(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`tad0`))) {
    return(.rxToSETad0(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`lag`)) ||
               identical(x[[1]], quote(`lead`)) ||
               identical(x[[1]], quote(`lag0`)) ||
               identical(x[[1]], quote(`lead0`))) {
    return(.rxToSELagOrLead(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`delay`)) ||
               identical(x[[1]], quote(`rxDelayD`)) ||
               identical(x[[1]], quote(`rxDelayD2`)) ||
               identical(x[[1]], quote(`rxDelayD3`))) {
    return(.rxToSEDelay(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`tafd`))) {
    return(.rxToSETlastOrTafd(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`tafd0`))) {
    return(.rxToSETlastOrTafd0(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`tlast`)) ||
               identical(x[[1]], quote(`tfirst`)) ||
               identical(x[[1]], quote(`tlast0`)) ||
               identical(x[[1]], quote(`tfirst0`)) ||
               identical(x[[1]], quote(`dose`)) ||
               identical(x[[1]], quote(`podo`)) ||
               identical(x[[1]], quote(`dose0`)) ||
               identical(x[[1]], quote(`podo0`))) {
    return(.rxToSETlastOrTfirst(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`psigamma`))) {
    return(.rxToSEPsigamma(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`log1pmx`))) {
    return(.rxToSELog1pmx(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`choose`))) {
    return(.rxToSEChoose(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`lchoose`))) {
    return(.rxToSELchoose(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if ((identical(x[[1]], quote(`pnorm`))) ||
               (identical(x[[1]], quote(`normcdf`))) ||
               (identical(x[[1]], quote(`phi`)))) {
    return(.rxToSEPnorm(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`transit`))) {
    return(.rxToSETransit(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`mix`))) {
    return(.rxToSEMix(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`abs`)) ||
               identical(x[[1]], quote(`fabs`)) ||
               identical(x[[1]], quote(`abs0`))) {
    if (length(x) != 2) stop("abs only takes 1 argument", call.=FALSE)
    .r <- .rxToSE(x[[2]], envir = envir)
    return(paste0("(2.0*(", .r, ")*rxGt(", .r, ",0.0)-(", .r, "))"))
  } else if (identical(x[[1]], quote(`d4GELU`))) {
    return(.rxToSEd4GELU(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`d4softplus`))) {
    return(.rxToSEd4softplus(x, envir = envir, progress = progress, isEnv=isEnv))
  } else if (identical(x[[1]], quote(`dSELU`))) {
    .rxToSEdSELU(x, envir=envir, progress=progress, isEnv=isEnv)
  } else if (identical(x[[1]], quote(`dSwish`))) {
    .rxToSEdSwish(x, envir=envir, progress=progress, isEnv=isEnv)
  } else {
    if (length(x[[1]]) == 1) {
      .x1 <- as.character(x[[1]])
      .xc <- .rxSEsingle[[.x1]]
      if (!is.null(.xc)) {
        if (length(x) == 2) {
          if (isEnv) {
            .lastCall <- envir$..curCall
            envir$..curCall <- c(envir$..curCall, .xc[3])
          }
          .ret <- paste0(
            .xc[1], .rxToSE(x[[2]], envir = envir),
            .xc[2]
          )
          if (isEnv) envir$..curCall <- .lastCall
          return(.ret)
        } else {
          stop(sprintf("'%s' only accepts 1 argument", .x1), call. = FALSE)
        }
      }
      .xc <- .rxSEdouble[[.x1]]
      if (!is.null(.xc)) {
        .x1 <- as.character(x[[1]])
        if (length(x) == 3) {
          .ret <- paste0(
            .xc[1], .rxToSE(x[[2]], envir = envir),
            .xc[2],
            .rxToSE(x[[3]], envir = envir),
            .xc[3]
          )
          return(.ret)
        } else {
          stop(sprintf("'%s' only acceps 2 arguments", .x1), call. = FALSE)
        }
      }
    }
    if (isEnv) {
      .lastCall <- envir$..curCall
      envir$..curCall <- c(
        envir$..curCall,
        as.character(x[[1]])
      )
    }
    .fun0 <- as.character(x[[1]])
    if (.fun0 %in% c("dur", "rate", "lag", "alag", "f", "F")) {
      .ret0 <- c(list(as.character(x[[1]])),
                 lapply(x[-1], as.character, envir = envir))
    } else {
      .ret0 <- c(list(as.character(x[[1]])),
                 lapply(x[-1], .rxToSE, envir = envir))
    }
    if (isEnv) envir$..curCall <- .lastCall
    .SEeq <- c(.rxSEeq, rxode2::.rxSEeqUsr())
    .curName <- paste(.ret0[[1]])
    .nargs <- .SEeq[.curName]
    if (.curName == "linCmtA") {
      # Save the information about the model so that
      # depot and the rest can be converted to the right equivalent
      # linear compartment model
      if (.promoteLinB) {
        # linCmtA(rx__PTR__, t, linCmt, ncmt, oral0, which, trans,
        #         p1, v1, p2, p3, p4, p5, ka)
        # linCmtB(rx__PTR__, t, linCmt, ncmt, oral0, which1, which2, trans,
        #         p1, v1, p2, p3, p4, p5, ka)
        .rxToSElinCmt$linCmt <- "linCmtB"
        .which <- eval(str2lang(.ret0[[7]]))
        if (.which == -1) {
          # Linear compartment solution
          .ret0 <- c(
            list("linCmtB"),
            .ret0[2:6],
            list("-1", "-1"),
            .ret0[-c(1, 2:7)]
          )
        } else {
          # Compartment value, convert to the corresponding compartment
          # value
          .ret0 <- c(
            list("linCmtB"),
            .ret0[2:6],
            list(paste0(.which), "-2"),
            .ret0[-c(1, 2:7)]
          )
        }
        .nargs <- .nargs + 1
      } else {
        .rxToSElinCmt$linCmt <- "linCmtA"
      }
      .rxSEsaveLinCmt(.ret0)
    }
    if (!is.na(.nargs)) {
      if (.nargs == length(.ret0) - 1) {
        .ret <- paste0(.ret0[[1]], "(")
        .ret0 <- .ret0[-1]
        .ret <- paste0(.ret, paste(unlist(.ret0), collapse = ","), ")")
        if (.ret == "exp(1)") {
          return("E")
        }
        return(.ret)
      } else {
        stop(sprintf(
          gettext("'%s' takes %s arguments (has %s)"),
          paste(.ret0[[1]]),
          .nargs, length(.ret0) - 1
        ), call. = FALSE)
      }
    } else {
      .fun <- paste(.ret0[[1]])
      .ret0 <- .ret0[-1]
      if (.fun == "linCmt") {
        return(paste0("linCmt(", paste(unlist(.ret0), collapse = ", "), ")"))
      }
      if (.fun == "past") {
        ## past(state, tau) <- expr : non-constant delay() pre-history.  Keyed by
        ## state (one past() per state); tau is carried separately.  Returns a
        ## flat per-state symbol so it never reaches symengine differentiation.
        return(paste0("rx_past_", .ret0[[1]], "_"))
      }
      if (length(.ret0) == 1L) {
        if (any(.fun == c("alag", "lag"))) {
          return(paste0("rx_lag_", .ret0[[1]], "_"))
        } else if (any(.fun == c("F", "f"))) {
          return(paste0("rx_f_", .ret0[[1]], "_"))
        } else if (any(.fun == c("rate", "dur"))) {
          return(paste0("rx_", .fun, "_", .ret0[[1]], "_"))
        }
      }
      .ret <- paste0("(", paste(unlist(.ret0), collapse = ","), ")")
      if (.ret == "(0)") {
        return(paste0("rx_", .fun, "_ini_0__"))
      } else if (any(.fun == c("cmt", "dvid", "mtime", "matExp", "indLin"))) {
        return("")
      } else if (.fun == "max") {
        .ret <- .rxToSEMax(unlist(.ret0), min=FALSE)
      } else if (.fun == "min") {
        .ret <- .rxToSEMax(unlist(.ret0), min=TRUE)
      } else if (.fun == "sum") {
        .ret <- paste0("(", paste(paste0("(", unlist(.ret0), ")"), collapse = "+"), ")")
      } else if (.fun == "prod") {
        .ret <- paste0("(", paste(paste0("(", unlist(.ret0), ")"), collapse = "*"), ")")
      } else if (.fun == "probitInv") {
        ## erf <- function(x) 2 * pnorm(x * sqrt(2)) - 1 (probitInv=pnorm)
        if (length(.ret0) == 1) {
          .ret <- paste0("0.5*(1+erf((", unlist(.ret0)[1], ")/sqrt(2)))")
        } else if (length(.ret0) == 2) {
          .ret0 <- unlist(.ret0)
          .p <- paste0("0.5*(1+erf((", .ret0[1], ")/sqrt(2)))")
          ## return (high-low)*p+low;
          .ret <- paste0(
            "(1.0-(", .ret0[2], "))*(", .p,
            ")+(", .ret0[2], ")"
          )
        } else if (length(.ret0) == 3) {
          .ret0 <- unlist(.ret0)
          .p <- paste0("0.5*(1+erf((", .ret0[1], ")/sqrt(2)))")
          .ret <- paste0(
            "((", .ret0[3], ")-(", .ret0[2], "))*(", .p,
            ")+(", .ret0[2], ")"
          )
        } else {
          stop("'probitInv' requires 1-3 arguments",
               call. = FALSE
               )
        }
      } else if (.fun == "probit") {
        ## erfinv <- function (x) qnorm((1 + x)/2)/sqrt(2) (probit=qnorm )
        if (length(.ret0) == 1) {
          .ret <- paste0("sqrt(2)*erfinv(2*(", unlist(.ret0), ")-1)")
        } else if (length(.ret0) == 2) {
          .ret0 <- unlist(.ret0)
          .p <- paste0(
            "((", .ret0[1], ")-(", .ret0[2], "))/(1.0-",
            "(", .ret0[2], "))"
          )
          .ret <- paste0("sqrt(2)*erfinv(2*(", .p, ")-1)")
        } else if (length(.ret0) == 3) {
          .ret0 <- unlist(.ret0)
          ## (x-low)/(high-low)
          .p <- paste0(
            "((", .ret0[1], ")-(", .ret0[2],
            "))/((", .ret0[3], ")-(", .ret0[2], "))"
          )
          .ret <- paste0("sqrt(2)*erfinv(2*(", .p, ")-1)")
        } else {
          stop("'probit' requires 1-3 arguments",
               call. = FALSE
               )
        }
      } else if (.fun == "logit") {
        if (length(.ret0) == 1) {
          .ret <- paste0("-log(1/(", unlist(.ret0), ")-1)")
        } else if (length(.ret0) == 2) {
          .ret0 <- unlist(.ret0)
          .p <- paste0(
            "((", .ret0[1], ")-(", .ret0[2], "))/(1.0-",
            "(", .ret0[2], "))"
          )
          .ret <- paste0("-log(1/(", .p, ")-1)")
        } else if (length(.ret0) == 3) {
          .ret0 <- unlist(.ret0)
          ## (x-low)/(high-low)
          .p <- paste0(
            "((", .ret0[1], ")-(", .ret0[2],
            "))/((", .ret0[3], ")-(", .ret0[2], "))"
          )
          .ret <- paste0("-log(1/(", .p, ")-1)")
        } else {
          stop("'logit' requires 1-3 arguments",
               call. = FALSE
               )
        }
      } else if (any(.fun == c("expit", "invLogit", "logitInv"))) {
        if (length(.ret0) == 1) {
          .ret <- paste0("1/(1+exp(-(", unlist(.ret0)[1], ")))")
        } else if (length(.ret0) == 2) {
          .ret0 <- unlist(.ret0)
          .p <- paste0("1/(1+exp(-(", .ret0[1], ")))")
          ## return (high-low)*p+low;
          .ret <- paste0(
            "(1.0-(", .ret0[2], "))*(", .p,
            ")+(", .ret0[2], ")"
          )
        } else if (length(.ret0) == 3) {
          .ret0 <- unlist(.ret0)
          .p <- paste0("1/(1+exp(-(", .ret0[1], ")))")
          .ret <- paste0(
            "((", .ret0[3], ")-(", .ret0[2], "))*(", .p,
            ")+(", .ret0[2], ")"
          )
        } else {
          stop("'expit' requires 1-3 arguments",
               call. = FALSE
               )
        }
      } else {
        if (.fun %in% c("param", "dvid", "cmt", "locf", "nocb",
                        "midpoint", "linear", "splitBolus", "matExp", "indLin")) return(NULL)
        if (.fun %in% c("printf", "Rprintf", "print")) {
          return(paste(deparse(x), collapse=""))
        }
        .udf <- try(get(.fun, envir = .rxToSE.envir$parent, mode="function"), silent =TRUE)
        if (inherits(.udf, "try-error")) {
          .udf <- try(get(.fun, envir = rxode2::.udfEnvSet(NULL), mode="function"), silent =TRUE)
        }
        if (inherits(.udf, "try-error")) {
          stop(sprintf(gettext("function '%s' or its derivatives are not supported in rxode2"), .fun),
               call. = FALSE
               )
        } else {
          .f <- formals(.udf)
          if (any(names(.f)  == "...")) {
            stop(sprintf(gettext("R user function '%s' has variable number of arguments with'...' and is not supported in rxode2"), .fun),
                 call. = FALSE
                 )
          } else if (length(.ret0) == length(.f)) {
            if (is.environment(envir)) {
              assign(.fun, .rxFunction(.fun), envir = envir)
            }
            .ret0 <- unlist(.ret0)
            .ret <- paste0(.fun, "(",paste(.ret0, collapse=", "), ")")
          } else {
            stop(sprintf(gettext("user function '%s' requires %d arguments (supplied %d)"), .fun,
                         length(.f), length(.ret0)),
                 call. = FALSE
                 )
          }
        }
      }
    }
  }
}

#' @rdname rxToSE
#' @export
.rxToSE <- function(x, envir = NULL, progress = FALSE) {
  rxReq("symengine")
  .isEnv <- inherits(envir, "rxS") || inherits(envir, "environment")
  if (is.name(x) || is.atomic(x)) {
    return(.rxToSENameOrAtomic(x, envir=envir, progress=progress, isEnv=.isEnv))
  } else if (is.call(x)) {
    return(.rxToSECall(x, envir = envir, progress = progress, isEnv=.isEnv))
  } else {
    stop("unsupported expression", call. = FALSE)
  }
}

.rxUnXi <- function(x) {
  gsub("_xi_([[:digit:]]*)", "rx_xi_\\1", x)
}


## 0L = error
## 1L = forward
## 2L = central
.rxFromNumDer <- 0L
.rxDelta <- (.Machine$double.eps)^(1 / 3)

.rxFromSE.envir <- new.env(parent=emptyenv())
.rxFromSE.envir$parent <- NULL

#' @rdname rxToSE
#' @export
rxFromSE <- function(x, unknownDerivatives = c("forward", "central", "error"),
                     parent=parent.frame()) {
  rxReq("symengine")
  .udfEnvSet(parent)
  .rxFromSE.envir$parent <- parent
  .unknown <- c("central" = 2L, "forward" = 1L, "error" = 0L)
  assignInMyNamespace(".rxFromNumDer", .unknown[match.arg(unknownDerivatives)])
  if (is(substitute(x), "character")) {
    return(.rxFromSE(eval(parse(text = paste0("quote({", .rxUnXi(x), "})")))))
  } else if (is(substitute(x), "{")) {
    x <- deparse1(substitute(x))
    if (x[1] == "{") {
      x <- x[-1]
      x <- x[-length(x)]
    }
    x <- .rxUnXi(paste(x, collapse = "\n"))
    .ret <- .rxFromSE(eval(parse(text = paste0("quote({", x, "})"))))
    return(.ret)
  } else {
    .xc <- as.character(substitute(x))
    x <- substitute(x)
    if (length(.xc) == 1) {
      .found <- FALSE
      .frames <- seq(1, sys.nframe())
      .frames <- .frames[.frames != 0]
      for (.f in .frames) {
        .env <- parent.frame(.f)
        if (exists(.xc, envir = .env)) {
          .val2 <- try(get(.xc, envir = .env), silent = TRUE)
          if (inherits(.val2, "character")) {
            .val2 <- eval(parse(text = paste0("quote({", .rxUnXi(.val2), "})")))
            .ret <- .rxFromSE(.val2)
            return(.ret)
          } else if (inherits(.val2, "numeric") || inherits(.val2, "integer")) {
            return(sprintf("%s", .val2))
          } else {
            if (!is.null(attr(class(.val2), "package"))) {
              if (attr(class(.val2), "package") == "symengine") {
                .val2 <- eval(parse(text = paste0("quote({", .rxUnXi(as.character(.val2)), "})")))
                .ret <- .rxFromSE(.val2)
                return(.ret)
              }
            }
          }
        }
      }
    }
    x <- eval(parse(text = paste("quote(", .rxUnXi(paste(deparse1(x), collapse = " ")), ")")))
    .ret <- .rxFromSE(x)
    return(.ret)
  }
  x <- .rxUnXi(x)
  .ret <- .rxFromSE(eval(parse(text = paste0("quote({", x, "})"))))
  return(.ret)
}

.stripP <- function(x) {
  if (is.call(x)) {
    if (length(x) == 1) {
      return(x)
    } else if (identical(x[[1]], quote(`(`))) {
      return(.stripP(x[[2]]))
    } else {
      return(x)
    }
  } else {
    return(x)
  }
}

.rxM1rmF <- function(x) {
  .env <- new.env(parent = emptyenv())
  .env$found <- FALSE
  .f <- function(x, envir) {
    if (is.call(x)) {
      if (identical(x[[1]], quote(`*`))) {
        if (length(x) == 3) {
          .x2 <- as.character(x[[2]])
          .x3 <- as.character(x[[3]])
          if (length(.x3) == 1) {
            if (.x3 == "pi") {
              envir$found <- TRUE
              return(.rxFromSE(.stripP(x[[2]])))
            }
          }
          if (length(.x2) == 1) {
            if (.x2 == "pi") {
              envir$found <- TRUE
              return(.rxFromSE(.stripP(x[[3]])))
            }
          }
          return(paste0(
            .f(x[[2]], envir), "*",
            .f(x[[3]], envir)
          ))
        } else {
          return(.rxFromSE(x))
        }
      } else if (identical(x[[1]], quote(`/`))) {
        .x2 <- as.character(x[[2]])
        if (length(.x2) == 1) {
          if (.x2 == "pi") {
            envir$found <- TRUE
            x[[2]] <- 1
            return(.rxFromSE(x))
          }
          return(paste0(
            .f(x[[2]], envir), "/",
            .f(x[[3]], envir)
          ))
        }
        ## Compound numerator (eg sin(2*(time-a)/period)): recurse into both
        ## sides so a `pi` factor buried in the numerator is still stripped and
        ## the argument is not dropped (else the whole arg became NULL -> sin()).
        ## Parenthesize the numerator unless it is already parenthesized so the
        ## `/` precedence is preserved after any pi-factor stripping.
        .num <- .f(x[[2]], envir)
        if (!identical(x[[2]][[1]], quote(`(`))) {
          .num <- paste0("(", .num, ")")
        }
        return(paste0(.num, "/", .f(x[[3]], envir)))
      } else {
        return(.rxFromSE(x))
      }
    } else {
      return(.rxFromSE(x))
    }
  }
  .ret <- .f(x, .env)
  return(list(.ret, .env$found))
}

.rxP1rmF <- function(x) {
  .env <- new.env(parent = emptyenv())
  .env$found <- FALSE
  .f <- function(x, envir) {
    if (is.call(x)) {
      if (identical(x[[1]], quote(`+`))) {
        if (length(x) == 3) {
          if (inherits(x[[3]], "numeric")) {
            if (x[[3]] == 1) {
              envir$found <- TRUE
              return(.rxFromSE(.stripP(x[[2]])))
            } else {
              return(paste0(
                .f(x[[2]], envir),
                "+", as.character(x[[3]])
              ))
            }
          }
          if (inherits(x[[2]], "numeric")) {
            if (x[[2]] == 1) {
              envir$found <- TRUE
              return(.rxFromSE(.stripP(x[[3]])))
            } else {
              return(paste0(
                as.character(x[[2]]), "+",
                .f(x[[3]], envir)
              ))
            }
          }
          return(paste0(
            .f(x[[2]], envir), "+",
            .f(x[[3]], envir)
          ))
        } else {
          return(.rxFromSE(x))
        }
      } else if (identical(x[[1]], quote(`-`))) {
        if (length(x) == 3) {
          if (inherits(x[[2]], "numeric")) {
            if (x[[2]] == 1) {
              x <- x[-2]
              envir$found <- TRUE
              return(.rxFromSE(x))
            } else {
              return(.rxFromSE(x))
            }
          } else {
            return(.rxFromSE(x))
          }
        } else {
          return(.rxFromSE(x))
        }
      } else {
        return(.rxFromSE(x))
      }
    } else {
      return(.rxFromSE(x))
    }
  }
  .ret <- .f(x, .env)
  return(list(.ret, .env$found))
}

.rxFromSEnum <- function(x) {
  .ret <- as.character(x)
  .retl <- nchar(.ret)
  if (length(.retl) == 1L && .retl > 5) {
    .op <- options()
    options(digits = 22)
    on.exit(options(.op))
    .rx <- names(.rxSEcnt)
    .val <- setNames(.rxSEcnt, NULL)
    E <- exp(1)
    for (.i in seq_along(.rx)) {
      .tmp <- try(eval(parse(text = paste0("substr(paste(", .val[.i], "), 1, .retl)"))), silent = TRUE)
      if (!inherits(.tmp, "try-error")) {
        if (.ret == .tmp) {
          return(.rx[.i])
        }
      }
    }
  }
  return(.ret)
}

#' @export
#' @rdname rxToSE
.rxFromSE <- function(x) {
  rxReq("symengine")
  .cnst <- setNames(
    names(.rxSEreserved),
    paste0("rx_SymPy_Res_", names(.rxSEreserved))
  )
  if (is.name(x) || is.atomic(x)) {
    .ret <- .rxFromSEnum(x)
    .ret <- .rxRepRxQ(.ret)
    .ret0 <- .cnst[.ret]
    if (!is.na(.ret0)) {
      return(.ret0)
    }
    .ret0 <- .rxSEreserved[[.ret]]
    if (!is.null(.ret0)) {
      if (is.character(.ret0)) {
        return(.ret0)
      } else if (is.numeric(.ret0)) {
        return(sprintf("%.16f", .ret0))
      }
    }
    if (.ret == "E") {
      return("M_E")
    }
    .ret <- sub(
      .regRate, "rate(\\1)",
      sub(
        .regDur, "dur(\\1)",
        sub(
          .regLag, "alag(\\1)",
          sub(
            .regF, "f(\\1)",
            sub(
              regIni0, "\\1(0)",
              sub(
                regDfDy, "df(\\1)/dy(\\2)",
                sub(
                  regDfDyTh, "df(\\1)/dy(\\2[\\3])",
                  sub(
                    regDDt, "d/dt(\\1)",
                    sub(
                      rex::rex(start, regThEt, end),
                      "\\1[\\2]", .ret
                    )
                  )
                )
              )
            )
          )
        )
      )
    )
    .ret <- sub("[(]rx_SymPy_Res_", "(", .ret)

    return(.ret)
  } else if (is.call(x)) {
    if (identical(x[[1]], quote(`(`))) {
      return(paste0("(", .rxFromSE(x[[2]]), ")"))
    } else if (identical(x[[1]], quote(`{`))) {
      .x2 <- x[-1]
      return(paste(lapply(.x2, function(x) {
        .ret <- .rxFromSE(x)
        return(.ret)
      }),
      collapse = "\n"
      ))
    } else if (identical(x[[1]], quote(`*`)) ||
      identical(x[[1]], quote(`^`)) ||
      identical(x[[1]], quote(`+`)) ||
      identical(x[[1]], quote(`-`)) ||
      identical(x[[1]], quote(`/`))) {
      if (length(x) == 3) {
        .x1 <- as.character(x[[1]])
        .x2 <- x[[2]]
        .x2 <- .rxFromSE(.x2)
        .x3 <- x[[3]]
        .x3 <- .rxFromSE(.x3)
        # eval in baseenv() so only numeric-constant subexpressions
        # canonicalize; the default frame's scope chain reaches the user's
        # global environment, leaking workspace variables into the
        # conversion (nlmixr2/rxode2#1109)
        .x3v <- try(eval(parse(text = .x3), envir = baseenv()), silent = TRUE)
        if (inherits(.x3v, "numeric") && length(.x3v) == 1L) {
          .x3 <- .rxFromSEnum(.x3v)
        }
        if (.x1 == "^" && .x3 == "1") {
          return(.x2)
        }
        if (.x1 == "^" && .x3 == "-1") {
          return(paste0("(1/(", .x2, "))"))
        }
        if (.x1 == "^" && is.numeric(x[[3]])) {
          if (round(x[[3]]) == x[[3]]) {
            return(paste0("Rx_pow_di(", .x2, ",", .x3, ")"))
          }
          if (.x3 == 0.5) {
            if (any(.x2 == c("pi", "M_PI"))) {
              return("M_SQRT_PI")
            } else if (any(.x2 == c("M_2_PI", "(M_2_PI)"))) {
              return("M_SQRT_2dPI")
            } else {
              return(paste0("sqrt(", .x2, ")"))
            }
          } else {
            return(paste0("Rx_pow(", .x2, ",", .x3, ")"))
          }
        }
        .ret <- paste0(.x2, .x1, .x3)
        ## FIXME parsing to figure out if *2 or *0.5 *0.4 is in
        ## expression
        if (any(.ret == c("pi*2", "2*pi", "M_PI*2", "2*M_PI"))) {
          return("M_2PI")
        }
        if (any(.ret == c("pi/2", "pi*0.5", "0.5*pi", "M_PI/2", "M_PI*0.5", "0.5*M_PI"))) {
          return("M_PI_2")
        }
        if (any(.ret == c("pi/4", "pi*0.25", "0.25*pi", "M_PI/4", "M_PI*0.25", "0.25*M_PI"))) {
          return("M_PI_4")
        }
        if (any(.ret == c("1/pi", "1/M_PI"))) {
          return("M_1_PI")
        }
        if (any(.ret == c("2/pi", "2/M_PI"))) {
          return("M_2_PI")
        }
        if (any(.ret == c(
          "(M_2_PI)^0.5", "(M_2_PI)^(1/2)",
          "M_2_PI^0.5", "M_2_PI^(1/2)",
          "sqrt((M_2_PI))"
        ))) {
          return("M_SQRT_2dPI")
        }
        if (any(.ret == c(
          "(pi)^0.5", "(pi)^(1/2)",
          "pi^0.5", "pi^(1/2)",
          "(M_PI)^0.5", "(M_PI)^(1/2)",
          "M_PI^0.5", "M_PI^(1/2)"
        ))) {
          return("M_SQRT_PI")
        }
        if (.ret == "log(2)/log(10)") {
          return("M_LOG10_2")
        }
        if (.ret == "1/log(10)") {
          return("M_LOG10E")
        }
        if (.ret == "1/log(2)") {
          return("M_LOG2E")
        }
        if (any(.ret == c("2/M_SQRT_PI", "2/(M_SQRT_PI)"))) {
          return("M_2_SQRTPI")
        }
        if (any(.ret == c(
          "1/sqrt(M_2PI)",
          "1/(sqrt((M_2PI)))",
          "1/(M_2PI^0.5)", "1/(M_2PI^(1/2))",
          "1/((M_2PI)^0.5)", "1/((M_2PI)^(1/2))"
        ))) {
          return("M_1_SQRT_2PI")
        }
        ## if (.x1 == "^") {
        ##   return(paste0("Rx_pow(", .x2, ",", .x3, ")"))
        ## }
        return(.ret)
      } else {
        ## Unary Operators
        .ret <- paste0(
          as.character(x[[1]]),
          .rxFromSE(x[[2]])
        )
        return(.ret)
      }
    } else if (identical(x[[1]], quote(`=`)) ||
      identical(x[[1]], quote(`<-`)) ||
      identical(x[[1]], quote(`~`))) {
      .var <- .rxFromSE(x[[2]])
      .val <- .rxFromSE(x[[3]])
    } else if (identical(x[[1]], quote(`[`))) {
      if (any(as.character(x[[2]]) == c("THETA", "ETA"))) {
        return(paste0(x[[2]], "[", x[[3]], "]"))
      }
      stop("[...] expressions not supported",
        call. = FALSE
      )
    } else if (identical(x[[1]], quote(`lag`)) ||
      identical(x[[1]], quote(`lead`)) ||
      identical(x[[1]], quote(`lag0`)) ||
      identical(x[[1]], quote(`lead0`))) {
      .a <- .rxFromSE(x[[2]])
      .fun <- as.character(x[[1]])
      if (length(x) == 3) {
        return(paste0(.fun, "(", .a, ",", .rxFromSE(x[[3]]), ")"))
      } else {
        return(paste0(.fun, "(", .a, ")"))
      }
    } else if (identical(x[[1]], quote(`delay`))) {
      return(paste0("delay(", .rxFromSE(x[[2]]), ", ", .rxFromSE(x[[3]]), ")"))
    } else if (identical(x[[1]], quote(`rxDelayD`))) {
      return(paste0("rxDelayD(", .rxFromSE(x[[2]]), ", ", .rxFromSE(x[[3]]), ")"))
    } else if (identical(x[[1]], quote(`rxDelayD2`))) {
      return(paste0("rxDelayD2(", .rxFromSE(x[[2]]), ", ", .rxFromSE(x[[3]]), ")"))
    } else if (identical(x[[1]], quote(`rxDelayD3`))) {
      return(paste0("rxDelayD3(", .rxFromSE(x[[2]]), ", ", .rxFromSE(x[[3]]), ")"))
    } else if (identical(x[[1]], quote(`polygamma`))) {
      if (length(x == 3)) {
        .a <- .rxFromSE(x[[2]])
        .b <- .rxFromSE(x[[3]])
        if (.a == "0") {
          return(paste0("digamma(", .b, ")"))
        } else if (.a == "1") {
          return(paste0("trigamma(", .b, ")"))
        } else if (.a == "2") {
          return(paste0("tetragamma(", .b, ")"))
        } else if (.a == "3") {
          return(paste0("pentagamma(", .b, ")"))
        } else {
          return(paste0(
            "psigamma(", .b, ",",
            .a, ")"
          ))
        }
      } else {
        stop("'polygamma' takes 2 arguments",
          call. = FALSE
        )
      }
    } else {
      if (length(x[[1]]) == 1) {
        .x1 <- as.character(x[[1]])
        .xc <- .SEsingle[[.x1]]
        if (!is.null(.xc)) {
          if (length(x) == 2) {
            .x2 <- x[[2]]
            if (length(.x2) != 1) {
              if (identical(.x2[[1]], quote(`+`))) {
                .tmp0 <- .SE1p[.x1]
                if (!is.na(.tmp0)) {
                  .ret <- .rxP1rmF(.x2)
                  if (.ret[[2]]) {
                    .r1 <- .ret[[1]]
                    return(paste0(
                      .tmp0, "(",
                      .r1,
                      ")"
                    ))
                  }
                }
                return(paste0(.xc[1], .x2[[1]], .xc[2]))
              }
            }
            return(paste0(.xc[1], .rxFromSE(x[[2]]), .xc[2]))
          } else {
            stop(sprintf("'%s' only acceps 1 argument", .x1),
              call. = FALSE
            )
          }
        }
        .xc <- .SEdouble[[.x1]]
        if (!is.null(.xc)) {
          if (length(x) == 3) {
            .x1 <- .rxFromSE(x[[1]])
            return(paste0(
              .xc[1], .rxFromSE(x[[2]]), .xc[2],
              .rxFromSE(x[[3]]),
              .xc[3]
            ))
          } else {
            stop(sprintf("'%s' only acceps 2 arguments", .x1),
              call. = FALSE
            )
          }
        }
      }
      if (length(x) == 2) {
        .isnan <- try(is.nan(x[[2]]), silent=TRUE)
        if (inherits(.isnan, "try-error")) .isnan <- FALSE
        if (.isnan) {
          if (as.character(x[[1]]) %in% .rxToSEDualVarFunction) {
            return(paste0(as.character(x[[1]]), "()"))
          }
        }
        if (identical(x[[1]], quote(`log`))) {
          if (length(x[[2]]) == 3) {
            if (identical(x[[2]][[1]], quote(`beta`))) {
              .tmp <- x[[2]]
              .tmp[[1]] <- quote(`lbeta`)
              return(.rxFromSE(.tmp))
            }
          }
        }
      }
      .ret0 <- lapply(lapply(x, .stripP), .rxFromSE)
      .SEeq <- c(.rxSEeq, .rxSEeqUsr())
      .nargs <- .SEeq[paste(.ret0[[1]])]
      if (!is.na(.nargs)) {
        if (.nargs == length(.ret0) - 1) {
          .x1 <- as.character(.ret0[[1]])
          .tmp0 <- .x1
          if (.nargs == 1) {
            .tmp0 <- .SE1p[.x1]
            .x2 <- x[[2]]
            if (!is.na(.tmp0)) {
              .ret <- .rxP1rmF(.x2)
              if (.ret[[2]]) {
                if (.tmp0 == "log1p") {
                  .tmp <- eval(parse(text = paste0("quote(", .ret[[1]], ")")))
                  if (length(.tmp) > 1) {
                    if (identical(.tmp[[1]], quote(`exp`))) {
                      .tmp <- .tmp[[-1]]
                      .tmp0 <- "log1pexp"
                      .ret[[1]] <- .rxFromSE(.tmp)
                    }
                  }
                }
                return(paste0(
                  .tmp0, "(",
                  .ret[[1]],
                  ")"
                ))
              } else {
                .ret <- paste0(
                  .x1, "(",
                  .ret[[1]],
                  ")"
                )
                if (.ret == "log(2)") {
                  return("M_LN2")
                }
                if (.ret == "log(10)") {
                  return("M_LN10")
                }
                if (.ret == "log(M_SQRT_PI)") {
                  return("M_LN_SQRT_PI")
                }
                if (any(.ret == c(
                  "log(sqrt((M_PI_2)))",
                  "log(sqrt(M_PI_2))",
                  "log((M_PI_2)^(1/2))",
                  "log((M_PI_2)^0.5)",
                  "log(M_PI_2^(1/2))",
                  "log(M_PI_2^0.5)"
                ))) {
                  return("M_LN_SQRT_PId2")
                }
                if (any(.ret == c(
                  "log(sqrt((M_2PI)))",
                  "log(sqrt(M_2PI))",
                  "log((M_2PI)^0.5)",
                  "log((M_2PI)^(1/2))",
                  "log(M_2PI^0.5)",
                  "log(M_2PI^(1/2))"
                ))) {
                  return("M_LN_SQRT_2PI")
                }
                return(.ret)
              }
            }
            .tmp0 <- .SE1m[.x1]
            if (!is.na(.tmp0)) {
              .ret <- .rxM1rmF(.x2)
              if (.ret[[2]]) {
                return(paste0(
                  .tmp0, "(",
                  .ret[[1]],
                  ")"
                ))
              } else {
                return(paste0(
                  .x1, "(",
                  .ret[[1]],
                  ")"
                ))
              }
            }
            .tmp0 <- .x1
          }
          .ret <- paste0(.tmp0, "(")
          .ret0 <- .ret0[-1]
          .ret <- paste0(
            .ret, paste(unlist(.ret0), collapse = ","),
            ")"
          )
          if (.ret == "exp(1)") {
            return("M_E")
          }
          if (.ret == "sin(pi)") {
            return("0")
          }
          if (.ret == "cos(pi)") {
            return("1")
          }
          if (.ret == "tan(pi)") {
            return("0")
          }
          if (.ret == "sqrt(3)") {
            return("M_SQRT_3")
          }
          if (.ret == "sqrt(2)") {
            return("M_SQRT2")
          }
          if (.ret == "sqrt(32)") {
            return("M_SQRT_32")
          }
          if (.ret == "sqrt(pi)") {
            return("M_SQRT_PI")
          }
          if (any(.ret == c("sqrt(M_2_PI)", "sqrt((M_2_PI))"))) {
            return("M_SQRT_2dPI")
          }
          return(.ret)
        } else {
          stop(sprintf(
            gettext("'%s' takes %s arguments"),
            paste(.ret0[[1]]),
            .nargs
          ))
        }
      } else if (identical(x[[1]], quote(`Derivative`))) {
        ## delay-family functions (delay/rxDelayD/lag/lead) have zero derivatives
        ## of every order, so Derivative(<delay-family>, ...) collapses to 0.
        ## This also covers the higher-order (length(x) > 3) derivatives produced
        ## when the sensitivity machinery differentiates a delayed term more than
        ## once -- symengine keeps Derivative(delay(...), v1, v2) unevaluated.
        if (length(x) >= 3 && is.call(x[[2]]) &&
              any(as.character(x[[2]][[1L]]) == c("lead", "lag", "delay", "rxDelayD", "rxDelayD2", "rxDelayD3"))) {
          return("0")
        }
        if (length(x) == 3) {
          .fun <- as.character(x[[2]])
          .var <- .rxFromSE(x[[3]])
          if (length(.fun) == 1) {
            if (.fun == "abs0") {
              return(paste0("abs(", .var, ")"))
            }
          }
          ## Convert the differentiated function's arguments from the PARSED
          ## expressions (as.list(x[[2]])[-1]), not as.character(x[[2]])[-1]:
          ## the latter deparses each argument to a whole string, so a nested
          ## symengine name (e.g. THETA_3_ inside sqrt(THETA_3_^2)) is treated as
          ## an atomic value and the start/end-anchored THETA_n_ -> THETA[n]
          ## conversion never fires, leaking the mangled name into the derivative
          ## (e.g. llikNormDmean(DV, y, sqrt(THETA_3_^2))).  Matches the
          ## length(x) > 3 branch below.
          .args <- lapply(as.list(x[[2]])[-1], .rxFromSE)
          .with <- which(.var == .args)
          .errD <- function(force = FALSE) {
            if (!force && .rxFromNumDer != 0L) {
              ## Can calculate forward or central
              ## difference instead.
              ## Warn
              if (.rxFromNumDer == 1L) {
                ## Forward
                .a1 <- .args
                .fn <- .fun[1]
                .a1[.with] <- paste0("(", .a1[.with], ")+", .rxDelta)
                .a2 <- .args
                return(paste0(
                  "(", .fn, "(", paste0(.a1, collapse = ","), ")-",
                  .fn, "(", paste0(.a2, collapse = ","), "))/", .rxDelta
                ))
              } else if (.rxFromNumDer == 2L) {
                ## Central
                .a1 <- .args
                .fn <- .fun[1]
                .a1[.with] <- paste0(.a1[.with], "-", (0.5 * .rxDelta))
                .a2 <- .args
                .a2[.with] <- paste0(.a2[.with], "+", (0.5 * .rxDelta))
                return(paste0(
                  "(", .fn, "(", paste0(.a1, collapse = ","), ")-",
                  .fn, "(", paste0(.a2, collapse = ","), "))/", .rxDelta
                ))
              } else {
                stop("only forward and central differences are supported", call. = FALSE)
              }
            } else {
              stop(sprintf(gettext("cannot figure out the '%s' derivative with respect to '%s'"), .fun[1], .var[1]))
            }
          }
          if (length(.with) != 1) {
            .errD(force = TRUE)
          }
          if (any(.fun[1] == c("lead", "lag", "delay", "rxDelayD", "rxDelayD2", "rxDelayD3"))) {
            return("0")
          }
          .rxD <- rxode2parseD()
          if (exists(.fun[1], envir = .rxD)) {
            .funLst <- get(.fun[1], envir = .rxD)
            if (length(.funLst) < .with) {
              return(.errD())
            }
            .derFun <- .funLst[[.with]]
            if (is.null(.derFun)) {
              return(.errD())
            }
            .ret <- try(do.call(.derFun, as.list(.args)), silent = TRUE)
            if (inherits(.ret, "try-error")) {
              warning(
                "an error occurred looking up the derivative for '",
                .fun[1], "' using numerical differences instead"
              )
              return(.errD())
            } else {
              return(.ret)
            }
          } else {
            if (.rxFromNumDer == 0L) {
              stop(sprintf(gettext("rxode2/symengine does not know how to take a derivative of '%s'"), .fun[1]),
                call. = FALSE
              )
            } else {
              return(.errD())
            }
          }
        } else if (length(x) > 3) {
          ## symengine collapses repeated same-argument differentiation
          ## D(D(fun, v), v)... into Derivative(fun, v, v, ...).  Apply the registered
          ## derivative chain iteratively (e.g. rxTBS -> rxTBSd -> rxTBSd2), carrying the
          ## arguments through the parsed result so each step uses the correct
          ## derivative-argument order; args are converted from symengine to rxode2 syntax
          ## once (first step) and reused as-is thereafter.
          .rxD <- rxode2parseD()
          .fun <- as.character(x[[2]][[1L]])
          .args <- lapply(as.list(x[[2]])[-1], .rxFromSE)
          .vars <- lapply(x[-(1:2)], .rxFromSE)
          .ok <- TRUE
          .res <- NULL
          for (.k in seq_along(.vars)) {
            .with <- which(.vars[[.k]] == unlist(.args))
            if (length(.with) != 1 || !exists(.fun, envir = .rxD)) {
              .ok <- FALSE
              break
            }
            .funLst <- get(.fun, envir = .rxD)
            if (length(.funLst) < .with || is.null(.funLst[[.with]])) {
              .ok <- FALSE
              break
            }
            .res <- try(do.call(.funLst[[.with]], as.list(.args)), silent = TRUE)
            if (inherits(.res, "try-error")) {
              .ok <- FALSE
              break
            }
            .call <- str2lang(.res)
            .fun <- as.character(.call[[1L]])
            .args <- lapply(as.list(.call)[-1],
                            function(.z) if (is.character(.z)) .z else paste(deparse(.z), collapse = ""))
          }
          if (.ok && !is.null(.res)) {
            return(.res)
          }
          stop("'Derivative' conversion only takes one function and one argument",
            call. = FALSE
          )
        } else {
          stop("'Derivative' conversion only takes one function and one argument",
            call. = FALSE
          )
        }
      } else if (identical(x[[1]], quote(`Subs`))) {
        .fun <- eval(parse(text = paste0("quote(", .rxFromSE(x[[2]]), ")")))
        .what <- .stripP(x[[3]])
        .with <- .stripP(x[[4]])
        .subs <- function(x) {
          if (identical(x, .what)) {
            return(.with)
          } else if (is.call(x)) {
            as.call(lapply(x, .subs))
          } else if (is.pairlist(x)) {
            as.pairlist(lapply(x, .subs))
          } else {
            return(x)
          }
        }
        .ret <- .subs(.fun)
        return(.rxFromSE(.ret))
      } else if (any(paste(.ret0[[1]]) == c("max", "min"))) {
        .x1 <- as.character(.ret0[[1]])
        .ret <- paste0(.x1, "(")
        .ret0 <- .ret0[-1]
        .ret <- paste0(
          .ret, paste(unlist(.ret0), collapse = ","),
          ")"
        )
        return(.ret)
      } else if (any(paste(.ret0[[1]]) == c("tlast", "tfirst", "dose", "podo",
                                            "tlast0", "first0", "dose0", "podo0"))) {
        if (length(.ret0) == 1L) {
          return(paste0(.ret0[[1]], "()"))
        } else if (length(.ret0) == 2L) {
          if (length(.ret0[[2]]) == 1L) {
            return(paste0(.ret0[[1]], "(", .ret0[[2]], ")"))
          }
        }
        stop(paste0(.ret0[[1]], "() takes 0-1 arguments"))
      } else {
        .fun <- paste(.ret0[[1]])
        .g <- try(get(.fun, envir=.rxFromSE.envir$parent, mode="function"), silent=TRUE)
        if (inherits(.g, "try-error")) {
          .g <- try(get(.fun, envir=.udfEnvSet(NULL), mode="function"), silent=TRUE)
        }
        if (inherits(.g, "try-error")) {
          stop(sprintf(gettext("'%s' not supported in symengine->rxode2"), .fun),
               call. = FALSE
               )
        } else {
          .f <- formals(.g)
          if (any(names(.f)  == "...")) {
            stop(sprintf(gettext("R user function '%s' has variable number of arguments with'...' and is not supported in rxode2"), .fun),
                 call. = FALSE
                 )
          } else if (length(.ret0) - 1 == length(.f)) {
            .ret <- unlist(.ret0)
            .fun <- .ret[1]
            .args <- .ret[-1]
            return(paste0(.fun, "(", paste(.args, collapse = ", "), ")"))
          } else {
            stop(sprintf(gettext("user function '%s' requires %d arguments (supplied %d)"), .fun,
                         length(.f), length(.ret0) - 1),
                 call. = FALSE)

          }
        }
      }
    }
  } else {
    stop("unsupported expression", call. = FALSE)
  }
}

.rxFunction <- function(name) {
  .f <- function(...) {
    1
  }
  body(.f) <- bquote({
    .args <- unlist(list(...))
    if (length(.args) == 0) {
      .args <- NaN
    }
    return(symengine::FunctionSymbol(.(name), .args))
  })
  return(.f)
}

#' Load a model into a symengine environment
#'
#' @param x rxode2 object
#' @param doConst Load constants into the environment as well.
#' @inheritParams rxToSE
#' @return rxode2/symengine environment
#' @author Matthew Fidler
#' @export
rxS <- function(x, doConst = TRUE, promoteLinSens = FALSE, envir=parent.frame()) {
  .udfEnvSet(envir)
  rxReq("symengine")
  .cnst <- names(.rxSEreserved)
  .env <- new.env(parent = loadNamespace("symengine"))
  .env$..mv <- rxModelVars(x)
  # States read by delay() are genuine ODEs (delay(X, T) requires d/dt(X)) even
  # when their name matches the sensitivity convention (rx__sens_<s>_BY_<v>__);
  # their d/dt() must stay in ..ddt.. and not be reclassified as a derivable
  # sensitivity that later passes strip (e.g. nlmixr2est's analytic-cov
  # augmented model supplies delayed sensitivity ODEs directly).
  .env$..delayedStates <- if (isTRUE(.env$..mv$flags[["hasDelay"]] == 1L)) {
    tryCatch(.rxDelayTerms(.env$..mv)$state, error = function(e) character(0))
  } else {
    character(0)
  }
  .env$..jac0 <- NULL
  .env$..jac0.. <- list()
  .env$..ddt <- NULL
  .env$..ddt.. <- list()
  .env$..sens0 <- NULL
  .env$..sens0.. <- list()
  .env$..lhs <- NULL
  .env$..lhs0 <- NULL
  .env$..doConst <- doConst
  .rxD <- rxode2parseD()
  for (.f in c(
    ls(.symengineFs()),
    ls(.rxD), "linCmtA", "linCmtB",
    "rxEq", "rxNeq", "rxGeq", "rxLeq", "rxLt",
    "rxGt", "rxAnd", "rxOr", "rxNot", "rxTBS", "rxTBSd", "rxTBSd2",
    "rxTBSdL", "rxTBSdL2", "rxTBSdLx", "lag", "lead",
    "lag0", "lead0", "diff0",
    "delay", "rxDelayD", "rxDelayD2", "rxDelayD3", "rxTBSi"
  )) {
    assign(.f, .rxFunction(.f), envir = .env)
  }

  for (.v in seq_along(.rxSEreserved)) {
    assign(names(.rxSEreserved)[.v], .rxSEreserved[[.v]], envir = .env)
  }
  .env$..s0 <- symengine::S("0")
  .env$..extraTheta <- list()
  .env$..extraEta <- list()
  .env$..curCall <- character(0)
  .env$..eventVars <- NULL
  .env$..mtimeVars <- NULL
  .env$polygamma <- function(a, b) {
    ## symengine::subs(symengine::subs(..polygamma, ..a, a), ..b,  b)
    symengine::psigamma(b, a)
  }
  # "tlast"
  .pars <- c(
    rxParams(x), rxState(x),
    "t", "time",  "rx1c", "rx__PTR__",
    "mixnum", "mixest", "mixunif")

  ## default lambda/yj values
  .env$rx_lambda_ <- symengine::S("1")
  .env$rx_yj_ <- symengine::S("2")
  .env$rx_low_ <- symengine::S("0")
  .env$rx_hi_ <- symengine::S("1")
  if (!is.null(.rxSEeqUsr())) {
    sapply(names(.rxSEeqUsr()), function(x) {
      assign(x, .rxFunction(x), envir = .env)
    })
  }
  ## EulerGamma=0.57721566490153286060651209008240243104215933593992
  ## S("I")
  ## S("pi")
  ## S("E")
  ## S("") # EulerGamma
  ## S("Catalan") = 0.915965594177219015054603514932384110774
  ## S("GoldenRatio") = 1+sqrt(5)/2
  ## S("inf")
  ## S("nan")
  sapply(.pars, function(x) {
    if (any(.cnst == x)) {
      .tmp <- paste0("rx_SymPy_Res_", x)
      assign(.tmp, symengine::Symbol(.tmp), envir = .env)
    } else {
      .tmp <- rxToSE(x, envir=.env)
      assign(.tmp, symengine::Symbol(.tmp), envir = .env)
      assign(x, symengine::Symbol(x), envir = .env)
    }
  })
  assignInMyNamespace(".promoteLinB", promoteLinSens)
  .expr <- eval(parse(text = paste0("quote({", rxNorm(x), "})")))
  # variables referenced inside lag()/lead()/diff()/first()/last() must be kept
  # as emitted lhs and bound as symbols (not inlined or dead-code eliminated), so
  # the history function still references a defined variable in the output model
  .env$..laggedVars <- .rxCollectLaggedVars(.expr)
  .ret <- .rxToSE(.expr, envir=.env)
  class(.env) <- "rxS"
  return(.env)
}

#' Collect the variables referenced inside history functions (lag/lead/diff/...)
#'
#' Only the RHS of assignments is scanned so the dosing `lag(cmt) <- ...` form
#' (a compartment property, not a history reference) is not collected.
#'
#' @param expr a quoted model (or sub-expression)
#' @return character vector of variable names used as the first argument of a
#'   history function
#' @author Matthew Fidler
#' @noRd
.rxCollectLaggedVars <- function(expr) {
  .acc <- character(0)
  .histFn <- c("lag", "lead", "diff", "first", "last", "lag0", "lead0", "diff0")
  .walk <- function(e) {
    if (is.call(e)) {
      .f <- e[[1]]
      if (is.name(.f) &&
            (identical(.f, quote(`<-`)) || identical(.f, quote(`=`)) ||
               identical(.f, quote(`~`)))) {
        # A variable defined directly as a history function (eg
        # rx_arEp <- lag0(rx_arE, 1)) is itself a lagged quantity: keep it
        # symbolic too (not inlined), so downstream differentiation w.r.t. it
        # stays valid (eg the AR(1) exact eta/theta gradient differentiates the
        # whitened likelihood by the previous residual).
        .rhs <- e[[3]]
        if (!identical(.f, quote(`~`)) && is.name(e[[2]]) && is.call(.rhs) &&
              is.name(.rhs[[1]]) && as.character(.rhs[[1]]) %in% .histFn) {
          .acc[[length(.acc) + 1L]] <<- as.character(e[[2]])
        }
        # only scan the RHS (skip dosing lag(cmt) on the lhs)
        .walkRhs(e[[3]])
      } else {
        for (.i in seq_along(e)) .walk(e[[.i]])
      }
    }
  }
  .walkRhs <- function(e) {
    if (is.call(e)) {
      .f <- e[[1]]
      if (is.name(.f) && as.character(.f) %in% .histFn &&
            length(e) >= 2L && is.name(e[[2]])) {
        .acc[[length(.acc) + 1L]] <<- as.character(e[[2]])
      }
      for (.i in seq_along(e)[-1]) .walkRhs(e[[.i]])
    }
  }
  .walk(expr)
  unique(.acc)
}

symengineC <- new.env(parent = emptyenv())
symengineC$"**" <- .dslToPow
symengineC$"^" <- .dslToPow

symengineC$S <- function(x) {
  sprintf("%s", x)
}

for (f in names(.rxSEeq)) {
  symengineC[[f]] <- functionOp(f)
}
symengineC$"(" <- unaryOp("(", ")")
for (op in c("+", "-", "*")) {
  symengineC[[op]] <- binaryOp(paste0(" ", op, " "))
}

symengineC[["/"]] <- function(e1, e2) {
  sprintf("%s /( (%s == 0) ? %s : %s)", e1, e2, .Machine$double.eps, e2)
}

unknownCsymengine <- function(op) {
  force(op)
  function(...) {
    stop(sprintf("rxode2 doesn't support '%s' translation for 'Omega' translation", op),
      call. = FALSE
    )
  }
}

symengineCEnv <- function(expr) {
  ## Known functions
  calls <- allCalls(expr)
  callList <- setNames(lapply(calls, unknownCsymengine), calls)
  callEnv <- list2env(callList)
  rxSymPyFEnv <- cloneEnv(symengineC, callEnv)
  names <- allNames(expr)
  ## Replace time with t.
  n1 <- names
  n2 <- names
  n2 <- gsub(rex::rex("t", capture(numbers)), "REAL(theta)[\\1]", n2)
  n2 <- gsub(rex::rex("pi"), "M_PI", n2)
  n2 <- gsub(rex::rex("rx_SymPy_Res_"), "", n2)
  n2 <- gsub("None", "NA_REAL", n2)
  w <- n2[n2 == "t"]
  symbol.list <- setNames(as.list(n2), n1)
  symbol.env <- list2env(symbol.list, parent = symengineC)
  return(symbol.env)
}

seC <- function(x) {
  expr <- eval(parse(text = sprintf("quote(%s)", as.character(x))))
  .ret <- eval(expr, symengineCEnv(expr))
}


## nocov end

allNames <- function(x) {
  if (is.atomic(x)) {
    character()
  } else if (is.name(x)) {
    as.character(x)
  } else if (is.call(x) || is.pairlist(x)) {
    children <- lapply(x[-1], allNames)
    unique(unlist(children))
  } else {
    stop("do not know how to handle type ", typeof(x),
      call. = FALSE
    )
  }
}

allCalls <- function(x) {
  if (is.atomic(x) || is.name(x)) {
    character()
  } else if (is.call(x)) {
    fname <- as.character(x[[1]])
    children <- lapply(x[-1], allCalls)
    unique(c(fname, unlist(children)))
  } else if (is.pairlist(x)) {
    unique(unlist(lapply(x[-1], allCalls), use.names = FALSE))
  } else {
    stop("do not know how to handle type ", typeof(x), call. = FALSE)
  }
}

cloneEnv <- function(env, parent = parent.env(env)) {
  list2env(as.list(env), parent = parent)
}

.exists2 <- function(x, where) {
  .nc <- try(nchar(x) < 1000, silent = TRUE)
  if (inherits(.nc, "try-error")) .nc <- FALSE
  if (rxIs(.nc, "logical")) .nc <- FALSE
  if (.nc) {
    return(exists(x, where))
  } else {
    return(FALSE)
  }
}

## Start error function DSL
rxErrEnvF <- new.env(parent = emptyenv())
for (op in c(
  "+", "-", "*", "/", "^", "**",
  "!=", "==", "&", "&&", "|", "||"
)) {
  op2 <- op
  if (op == "**") {
    op2 <- "^"
  }
  rxErrEnvF[[op]] <- binaryOp(paste0(" ", op2, " "))
}
for (op in c("=", "~", "<-")) {
  rxErrEnvF[[op]] <- binaryOp(" = ")
}
rxErrEnvF$"{" <- function(...) {
  return(sprintf("{\n%s\n}", paste(unlist(list(...)), collapse = "\n")))
}
rxErrEnvF$"(" <- unaryOp("(", ")")
rxErrEnvF$"[" <- function(name, val) {
  n <- toupper(name)
  err <- gettext("rxode2 only supports THETA[#] and ETA[#] numbers")
  if (any(n == c("THETA", "ETA")) && is.numeric(val)) {
    if (round(val) == val && val > 0) {
      if (n == "THETA" && as.numeric(val) <= length(rxErrEnv.init)) {
        return(sprintf("THETA[%s]", val))
      } else {
        return(sprintf("%s[%s]", n, val))
      }
    } else {
      stop(err)
    }
  } else {
    stop(err)
  }
}

rxErrEnvF$"if" <- function(lg, tr, fl) {
  if (missing(fl)) {
    return(sprintf("if (%s) %s", lg, tr))
  } else {
    return(sprintf("if (%s) %s else %s", lg, tr, fl))
  }
}
rxErrEnv.theta <- 1
rxErrEnv.diag.est <- NULL
rxErrEnv.ret <- "rx_r_"
rxErrEnv.init <- NULL
rxErrEnv.lambda <- NULL
rxErrEnv.yj <- NULL
rxErrEnv.combined <- "^2"
rxErrEnv.hasAdd <- FALSE
rxErrEnv.hi <- "1"
rxErrEnv.low <- "0"

.rxErrEnvInit <- function() {
  assignInMyNamespace("rxErrEnv.hasAdd", FALSE)
  assignInMyNamespace("rxErrEnv.theta", 1)
  assignInMyNamespace("rxErrEnv.diag.est", NULL)
  assignInMyNamespace("rxErrEnv.ret", "rx_r_")
  assignInMyNamespace("rxErrEnv.init", NULL)
  assignInMyNamespace("rxErrEnv.lambda", NULL)
  assignInMyNamespace("rxErrEnv.yj", NULL)
  assignInMyNamespace("rxErrEnv.combined", "^2")
  assignInMyNamespace("rxErrEnv.hi", "1")
  assignInMyNamespace("rxErrEnv.low", "0")
}


rxErrEnvF$lnorm <- function(est) {
  if (rxErrEnv.ret != "rx_r_") {
    stop("'lnorm' can only be in an error function", call. = FALSE)
  }
  if (!is.null(rxErrEnv.lambda)) {
    if (rxErrEnv.lambda != "0" && rxErrEnv.yj != "0") {
      stop("'lnorm' cannot be used with other data transformations", call. = FALSE)
    }
  }
  if (is.na(est)) {
    assignInMyNamespace("rxErrEnv.lambda", "0")
    assignInMyNamespace("rxErrEnv.yj", "0")
    return("")
  } else {
    estN <- suppressWarnings(as.numeric(est))
    assignInMyNamespace("rxErrEnv.hasAdd", TRUE)
    if (is.na(estN)) {
      ret <- (sprintf("(%s)%s", est, rxErrEnv.combined))
      assignInMyNamespace("rxErrEnv.lambda", "0")
      assignInMyNamespace("rxErrEnv.yj", "0")
    } else {
      theta <- sprintf("THETA[%s]", rxErrEnv.theta)
      est <- estN
      theta.est <- theta
      ret <- (sprintf("(%s)%s", theta.est, rxErrEnv.combined))
      tmp <- rxErrEnv.diag.est
      tmp[sprintf("THETA[%s]", rxErrEnv.theta)] <- as.numeric(est)
      assignInMyNamespace("rxErrEnv.diag.est", tmp)
      assignInMyNamespace("rxErrEnv.theta", rxErrEnv.theta + 1)
      assignInMyNamespace("rxErrEnv.lambda", "0")
      assignInMyNamespace("rxErrEnv.yj", "0")
    }
  }
  return(ret)
}

rxErrEnvF$dlnorm <- rxErrEnvF$lnorm
rxErrEnvF$logn <- rxErrEnvF$lnorm

rxErrEnvF$logitNorm <- function(est, low = "0", hi = "1") {
  if (rxErrEnv.ret != "rx_r_") {
    stop("'logitNorm' can only be in an error function", call. = FALSE)
  }
  if (!is.null(rxErrEnv.lambda)) {
    if (rxErrEnv.yj != "1") {
      if (rxErrEnv.yj != "4" && rxErrEnv.yj != "5") {
        stop("'logitNorm' cannot be used with other data transformations", call. = FALSE)
      }
    }
  }
  if (is.na(est)) {
    if (is.null(rxErrEnv.yj)) {
      assignInMyNamespace("rxErrEnv.yj", "4")
    } else if (rxErrEnv.yj == "1") {
      assignInMyNamespace("rxErrEnv.yj", "5")
    } else {
      assignInMyNamespace("rxErrEnv.yj", "4")
    }
    assignInMyNamespace("rxErrEnv.hi", hi)
    assignInMyNamespace("rxErrEnv.low", low)
    return("")
  } else {
    estN <- suppressWarnings(as.numeric(est))
    assignInMyNamespace("rxErrEnv.hasAdd", TRUE)
    if (is.na(estN)) {
      ret <- (sprintf("(%s)%s", est, rxErrEnv.combined))
      assignInMyNamespace("rxErrEnv.lambda", "0")
      if (is.null(rxErrEnv.yj)) {
        assignInMyNamespace("rxErrEnv.yj", "4")
      } else if (rxErrEnv.yj == "1") {
        assignInMyNamespace("rxErrEnv.yj", "5")
      } else {
        assignInMyNamespace("rxErrEnv.yj", "4")
      }
      assignInMyNamespace("rxErrEnv.hi", hi)
      assignInMyNamespace("rxErrEnv.low", low)
    } else {
      theta <- sprintf("THETA[%s]", rxErrEnv.theta)
      est <- estN
      theta.est <- theta
      ret <- (sprintf("(%s)%s", theta.est, rxErrEnv.combined))
      tmp <- rxErrEnv.diag.est
      tmp[sprintf("THETA[%s]", rxErrEnv.theta)] <- as.numeric(est)
      assignInMyNamespace("rxErrEnv.diag.est", tmp)
      assignInMyNamespace("rxErrEnv.theta", rxErrEnv.theta + 1)
      assignInMyNamespace("rxErrEnv.lambda", "0")
      if (is.null(rxErrEnv.yj)) {
        assignInMyNamespace("rxErrEnv.yj", "4")
      } else if (rxErrEnv.yj == "1") {
        assignInMyNamespace("rxErrEnv.yj", "5")
      } else {
        assignInMyNamespace("rxErrEnv.yj", "4")
      }
      assignInMyNamespace("rxErrEnv.hi", hi)
      assignInMyNamespace("rxErrEnv.low", low)
    }
  }
  return(ret)
}

rxErrEnvF$probitNorm <- function(est, low = "0", hi = "1") {
  if (rxErrEnv.ret != "rx_r_") {
    stop("'probitNorm' can only be in an error function", call. = FALSE)
  }
  if (!is.null(rxErrEnv.lambda)) {
    if (rxErrEenv.yj != "1") {
      if (rxErrEnv.yj != "6" &&rxErrEnv.yj != "7") {
        print(rxErrEnv.yj)
        stop("'probitNorm' cannot be used with other data transformations", call. = FALSE)
      }
    }
  }
  if (is.na(est)) {
    if (is.null(rxErrEnv.yj)) {
      assignInMyNamespace("rxErrEnv.yj", "6")
    } else if (rxErrEnv.yj == "1") {
      assignInMyNamespace("rxErrEnv.yj", "7")
    } else {
      assignInMyNamespace("rxErrEnv.yj", "6")
    }
    assignInMyNamespace("rxErrEnv.hi", hi)
    assignInMyNamespace("rxErrEnv.low", low)
    return("")
  } else {
    estN <- suppressWarnings(as.numeric(est))
    assignInMyNamespace("rxErrEnv.hasAdd", TRUE)
    if (is.na(estN)) {
      ret <- (sprintf("(%s)%s", est, rxErrEnv.combined))
      assignInMyNamespace("rxErrEnv.lambda", "0")
      if (is.null(rxErrEnv.yj)) {
        assignInMyNamespace("rxErrEnv.yj", "6")
      } else if (rxErrEnv.yj == "1") {
        assignInMyNamespace("rxErrEnv.yj", "7")
      } else {
        assignInMyNamespace("rxErrEnv.yj", "6")
      }
      assignInMyNamespace("rxErrEnv.hi", hi)
      assignInMyNamespace("rxErrEnv.low", low)
    } else {
      theta <- sprintf("THETA[%s]", rxErrEnv.theta)
      est <- estN
      theta.est <- theta
      ret <- (sprintf("(%s)%s", theta.est, rxErrEnv.combined))
      tmp <- rxErrEnv.diag.est
      tmp[sprintf("THETA[%s]", rxErrEnv.theta)] <- as.numeric(est)
      assignInMyNamespace("rxErrEnv.diag.est", tmp)
      assignInMyNamespace("rxErrEnv.theta", rxErrEnv.theta + 1)
      assignInMyNamespace("rxErrEnv.lambda", "0")
      if (is.null(rxErrEnv.yj)) {
        assignInMyNamespace("rxErrEnv.yj", "6")
      } else if (rxErrEnv.yj == "1") {
        assignInMyNamespace("rxErrEnv.yj", "7")
      } else {
        assignInMyNamespace("rxErrEnv.yj", "6")
      }
      assignInMyNamespace("rxErrEnv.hi", hi)
      assignInMyNamespace("rxErrEnv.low", low)
    }
  }
  return(ret)
}

rxErrEnvF$tbs <- function(lambda) {
  if (rxErrEnv.ret != "rx_r_") {
    stop("'boxCox' can only be in an error function", call. = FALSE)
  }
  if (!is.null(rxErrEnv.lambda)) {
    if (rxErrEnv.yj != "0" && rxErrEnv.lambda != "0" && rxErrEnv.lambda != "1") {
      stop("'boxCox' cannot be used with other data transformations", call. = FALSE)
    }
  }
  estN <- suppressWarnings(as.numeric(lambda))
  if (is.na(estN)) {
    assignInMyNamespace("rxErrEnv.lambda", lambda)
    assignInMyNamespace("rxErrEnv.yj", "0")
  } else {
    tmp <- rxErrEnv.diag.est
    tmp[sprintf("THETA[%s]", rxErrEnv.theta)] <- estN
    assignInMyNamespace("rxErrEnv.lambda", sprintf("THETA[%s]", rxErrEnv.theta))
    assignInMyNamespace("rxErrEnv.diag.est", tmp)
    assignInMyNamespace("rxErrEnv.theta", rxErrEnv.theta + 1)
    assignInMyNamespace("rxErrEnv.yj", "0")
  }
  return("0")
}

rxErrEnvF$boxCox <- rxErrEnvF$tbs

rxErrEnvF$tbsYj <- function(lambda) {
  if (rxErrEnv.ret != "rx_r_") {
    stop("'yeoJohnson' can only be in an error function", call. = FALSE)
  }
  if (!is.null(rxErrEnv.lambda)) {
    if ((rxErrEnv.yj != "1" && rxErrEnv.yj != "4" && rxErrEnv.yj != "6")) {
      stop("'yeoJohnson' cannot be used with other data transformations", call. = FALSE)
    }
  }
  estN <- suppressWarnings(as.numeric(lambda))
  if (is.na(estN)) {
    assignInMyNamespace("rxErrEnv.lambda", lambda)
    if (is.null(rxErrEnv.yj)) {
      assignInMyNamespace("rxErrEnv.yj", "1")
    } else if (rxErrEnv.yj == "4") {
      assignInMyNamespace("rxErrEnv.yj", "5")
    } else if (rxErrEnv.yj == "6") {
      assignInMyNamespace("rxErrEnv.yj", "7")
    } else {
      assignInMyNamespace("rxErrEnv.yj", "1")
    }
  } else {
    tmp <- rxErrEnv.diag.est
    tmp[sprintf("THETA[%s]", rxErrEnv.theta)] <- estN
    assignInMyNamespace("rxErrEnv.lambda", sprintf("THETA[%s]", rxErrEnv.theta))
    assignInMyNamespace("rxErrEnv.diag.est", tmp)
    assignInMyNamespace("rxErrEnv.theta", rxErrEnv.theta + 1)
    if (is.null(rxErrEnv.yj)) {
      assignInMyNamespace("rxErrEnv.yj", "1")
    } else if (rxErrEnv.yj == "4") {
      assignInMyNamespace("rxErrEnv.yj", "5")
    } else if (rxErrEnv.yj == "6") {
      assignInMyNamespace("rxErrEnv.yj", "7")
    } else {
      assignInMyNamespace("rxErrEnv.yj", "1")
    }
  }
  return("0")
}

rxErrEnvF$yeoJohnson <- rxErrEnvF$tbsYj

rxErrEnvF$add <- function(est) {
  if (rxErrEnv.ret != "rx_r_") {
    stop("'add' can only be in an error function", call. = FALSE)
  }
  assignInMyNamespace("rxErrEnv.hasAdd", TRUE)
  estN <- suppressWarnings(as.numeric(est))
  if (is.na(estN)) {
    ret <- (sprintf("(%s)%s", est, rxErrEnv.combined))
  } else {
    theta <- sprintf("THETA[%s]", rxErrEnv.theta)
    est <- estN
    theta.est <- theta
    ret <- (sprintf("(%s)%s", theta.est, rxErrEnv.combined))
    tmp <- rxErrEnv.diag.est
    tmp[sprintf("THETA[%s]", rxErrEnv.theta)] <- as.numeric(est)
    assignInMyNamespace("rxErrEnv.diag.est", tmp)
    assignInMyNamespace("rxErrEnv.theta", rxErrEnv.theta + 1)
  }
  return(ret)
}

rxErrEnvF$norm <- rxErrEnvF$add
rxErrEnvF$dnorm <- rxErrEnvF$add

rxErrEnvF$"for" <- function(...) {
  stop("'for' is not supported", call. = FALSE)
}
rxErrEnvF$`return` <- function(est) {
  if (rxErrEnv.ret == "") {
    stop("The PK function should not return anything", call. = FALSE)
  }
  .extra <- ""
  force(est)
  if (rxErrEnv.ret == "rx_r_") {
    .hi <- rxErrEnv.hi
    .low <- rxErrEnv.low
    if (is.null(rxErrEnv.lambda)) {
      .lambda <- "1"
    } else {
      .lambda <- rxErrEnv.lambda
    }
    if (is.null(rxErrEnv.yj)) {
      .yj <- "0"
    } else {
      .yj <- rxErrEnv.yj
    }
    if (.yj == "0" && .lambda == "1") {
      .yj <- "2"
      .lambda <- "1"
    }
    if (.yj == "0" && .lambda == "0") {
      .yj <- "3"
      .lambda <- "0"
    }
    .extra <- sprintf("rx_yj_~%s;\nrx_lambda_~%s;\nrx_hi_~%s\nrx_low_~%s\n", .yj, .lambda, .hi, .low)
  }
  return(sprintf("%s%s=%s;", .extra, rxErrEnv.ret, est))
}


rxErrEnvF$`|` <- binaryOp(" | ")
rxErrEnvF$`||` <- binaryOp(" || ")
rxErrEnvF$`&&` <- binaryOp(" && ")
rxErrEnvF$`<=` <- binaryOp(" <= ")
rxErrEnvF$`>=` <- binaryOp(" >= ")
rxErrEnvF$`==` <- binaryOp(" == ")

rxErrEnvF$prop <- function(est) {
  if (rxErrEnv.ret != "rx_r_") {
    stop("'prop' can only be in an error function")
  }
  estN <- suppressWarnings(as.numeric(est))
  if (is.na(estN)) {
    ret <- (sprintf("(rx_pred_f_)%s * (%s)%s", rxErrEnv.combined, est, rxErrEnv.combined))
  } else {
    est <- estN
    ret <- ""
    theta <- sprintf("THETA[%s]", rxErrEnv.theta)
    theta.est <- theta
    ret <- (sprintf("(rx_pred_f_)%s*(%s)%s", rxErrEnv.combined, theta.est, rxErrEnv.combined))
    tmp <- rxErrEnv.diag.est
    tmp[sprintf("THETA[%s]", rxErrEnv.theta)] <- as.numeric(est)
    assignInMyNamespace("rxErrEnv.diag.est", tmp)
    assignInMyNamespace("rxErrEnv.theta", rxErrEnv.theta + 1)
  }
  return(ret)
}

rxErrEnvF$propT <- function(est) {
  if (rxErrEnv.ret != "rx_r_") {
    stop("'propT' can only be in an error function")
  }
  estN <- suppressWarnings(as.numeric(est))
  if (is.na(estN)) {
    ret <- (sprintf("(rx_pred_)%s * (%s)%s", rxErrEnv.combined, est, rxErrEnv.combined))
  } else {
    est <- estN
    ret <- ""
    theta <- sprintf("THETA[%s]", rxErrEnv.theta)
    theta.est <- theta
    ret <- (sprintf("(rx_pred_)%s*(%s)%s", rxErrEnv.combined, theta.est, rxErrEnv.combined))
    tmp <- rxErrEnv.diag.est
    tmp[sprintf("THETA[%s]", rxErrEnv.theta)] <- as.numeric(est)
    assignInMyNamespace("rxErrEnv.diag.est", tmp)
    assignInMyNamespace("rxErrEnv.theta", rxErrEnv.theta + 1)
  }
  return(ret)
}

rxErrEnvF$pow <- function(est, pow) {
  if (rxErrEnv.ret != "rx_r_") {
    stop("'pow' can only be in an error function", call. = FALSE)
  }
  estN <- suppressWarnings(as.numeric(est))
  if (is.na(estN)) {
    ret <- (sprintf(
      "(rx_pred_f_)^(%s%s) * (%s)%s", ifelse(rxErrEnv.combined == "^2", "2*", ""),
      pow, est, rxErrEnv.combined
    ))
  } else {
    est <- estN
    ret <- ""
    theta <- sprintf("THETA[%s]", rxErrEnv.theta)
    theta2 <- sprintf("THETA[%s]", rxErrEnv.theta + 1)
    theta.est <- theta
    ret <- (sprintf("(rx_pred_f_)^(%s%s) * (%s)%s", ifelse(rxErrEnv.combined == "^2", "2*", ""), theta2, theta.est, rxErrEnv.combined))
    tmp <- rxErrEnv.diag.est
    tmp[sprintf("THETA[%s]", rxErrEnv.theta)] <- as.numeric(est)
    tmp[sprintf("THETA[%s]", rxErrEnv.theta + 1)] <- as.numeric(pow)
    assignInMyNamespace("rxErrEnv.diag.est", tmp)
    assignInMyNamespace("rxErrEnv.theta", rxErrEnv.theta + 2)
  }
  return(ret)
}

rxErrEnvF$powT <- function(est, pow) {
  if (rxErrEnv.ret != "rx_r_") {
    stop("'powT' can only be in an error function", call. = FALSE)
  }
  estN <- suppressWarnings(as.numeric(est))
  if (is.na(estN)) {
    ret <- (sprintf(
      "(rx_pred_)^(%s%s) * (%s)%s", ifelse(rxErrEnv.combined == "^2", "2*", ""),
      pow, est, rxErrEnv.combined
    ))
  } else {
    est <- estN
    ret <- ""
    theta <- sprintf("THETA[%s]", rxErrEnv.theta)
    theta2 <- sprintf("THETA[%s]", rxErrEnv.theta + 1)
    theta.est <- theta
    ret <- (sprintf("(rx_pred_)^(%s%s) * (%s)%s", ifelse(rxErrEnv.combined == "^2", "2*", ""), theta2, theta.est, rxErrEnv.combined))
    tmp <- rxErrEnv.diag.est
    tmp[sprintf("THETA[%s]", rxErrEnv.theta)] <- as.numeric(est)
    tmp[sprintf("THETA[%s]", rxErrEnv.theta + 1)] <- as.numeric(pow)
    assignInMyNamespace("rxErrEnv.diag.est", tmp)
    assignInMyNamespace("rxErrEnv.theta", rxErrEnv.theta + 2)
  }
  return(ret)
}

.convStr <- function(x) {
  if (is.atomic(x) || is.name(x)) {
    if (is.character(x)) {
      .rxChrToSym(x)
    } else {
      return(x)
    }
  } else if (is.call(x) || is.pairlist(x)) {
    return(as.call(lapply(x, .convStr)))
  } else {
    stop("do not know how to handle type ", typeof(x),
      call. = FALSE
    )
  }
}

rxErrEnv <- function(expr) {
  calls <- allCalls(expr)
  callList <- setNames(lapply(calls, functionOp), calls)
  callEnv <- list2env(callList)

  ## Known functions
  rxErrFEnv <- cloneEnv(rxErrEnvF, callEnv)

  ## Symbols
  names <- allNames(expr)
  n1 <- names
  n2 <- names
  n2[n2 == "time"] <- "t"
  if (any(n2 == "err")) {
    stop("use 'return' for errors", call. = FALSE)
  }
  if (any(n2 == "error")) {
    stop("use 'return' for errors", call. = FALSE)
  }
  if (any(n2 == "rx_r")) {
    stop("use 'return' for errors", call. = FALSE)
  }
  ## n2[n2 == "err"] <- "rx_r_";
  ## n2[n2 == "error"] <- "rx_r_";
  ## n2[n2 == "f"] <- "rx_pred_f_";
  symbol.list <- setNames(as.list(n2), n1)
  symbol.env <- list2env(symbol.list, parent = rxErrFEnv)
  return(symbol.env)
}

#' Parse PK function for inclusion in rxode2
#'
#' @param x PK function
#' @inheritParams rxParseErr
#' @return rxode2 transformed text.
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
rxParsePk <- function(x, init = NULL) {
  return(rxParseErr(x, init = init, ret = ""))
}
#' Prepare Pred function for inclusion in rxode2
#'
#' @param x pred function
#' @inheritParams rxParseErr
#' @return rxode2 transformed text.
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
rxParsePred <- function(x, init = NULL, err = NULL, addProp = c("combined2", "combined1")) {
  if (is.null(err)) {
    return(rxParseErr(x, ret = "rx_pred_", init = init, addProp = addProp))
  } else {
    .ini <- attr(err, "ini")
    .errs <- rxExpandIfElse(rxGetModel(err))
    .prd <- rxParseErr(x, ret = "rx_pred_", init = init)
    .prd <- rxExpandIfElse(rxGetModel(.prd))
    if (length(.prd) > 1) {
      if (length(.prd) == length(.errs)) {
        .prd <- .prd[names(.errs)]
        if (any(is.na(.prd))) {
          stop("errors and predictions need to have the same conditions ('if'/'then' statements)",
            call. = FALSE
          )
        }
      } else if (length(.errs) != 1) {
        stop("do not know how to handle this error/pred combination",
          call. = FALSE
        )
      }
    }
    .ret <- sapply(seq(1, max(length(.errs), length(.prd))), function(en) {
      .e <- .errs[min(length(.errs), en)]
      .p <- .prd[min(length(.prd), en)]
      .reg <- rex::rex("rx_pred_", any_spaces, "=", any_spaces, capture(except_any_of(";\n")), any_of(";\n"))
      if (regexpr(rex::rex("rx_yj_~2;\nrx_lambda_~1;\n"), .e) != -1) {
        .p <- gsub(.reg, "rx_pred_f_~\\1;\nrx_pred_ = \\1", .p)
      } else if (regexpr(rex::rex("rx_yj_~3;\nrx_lambda_~0;\n"), .e) != -1) {
        .p <- gsub(.reg, "rx_pred_f_~\\1;\nrx_pred_ = log(\\1)", .p)
      } else {
        .p <- gsub(.reg, "rx_pred_f_~\\1;\nrx_pred_ = rxTBS(\\1, rx_lambda_, rx_yj_, rx_low_, rx_hi_)", .p)
      }
      return(gsub("rx_r_", sprintf("%s\nrx_r_", .p), .e))
    })
    if (length(.ret) == 1L) {
      .ret <- setNames(.ret, NULL)
      attr(.ret, "ini") <- .ini
      return(.ret)
    } else {
      if (length(.errs) >= length(.prd)) {
        .n <- names(.errs)
      } else {
        .n <- names(.prd)
      }
      .ord <- order(sapply(.n, nchar))
      .n <- .n[.ord]
      .ret <- .ret[.ord]
      .ret <- paste(sapply(seq_along(.n), function(x) {
        if (x > 1) {
          if (.n[x] == sprintf("(!%s)", .n[x - 1])) {
            return(sprintf("else {\n%s\n}", .ret[x]))
          }
        }
        return(sprintf("if %s {\n%s\n}", .n[x], .ret[x]))
      }), collapse = "\n")
      attr(.ret, "ini") <- .ini
      return(.ret)
    }
  }
}
#' Prepare Error function for inclusion in rxode2
#'
#' @param x error function
#' @param baseTheta Base theta to start numbering add(.) and prop(.) from.
#' @param ret Internal return type.  Should not be changed by the user...
#' @param init Initialization vector
#' @return rxode2 transformed text
#' @keywords internal
#' @author Matthew L. Fidler
#' @export
rxParseErr <- function(x, baseTheta, ret = "rx_r_", init = NULL,
                       addProp = c("combined2", "combined1")) {
  addProp <- match.arg(addProp)
  if (!missing(baseTheta)) {
    assignInMyNamespace("rxErrEnv.theta", baseTheta)
  }
  if (!missing(ret)) {
    assignInMyNamespace("rxErrEnv.ret", ret)
  }
  if (!missing(init)) {
    assignInMyNamespace("rxErrEnv.init", init)
  }
  if (!missing(init)) {
    assignInMyNamespace("rxErrEnv.init", init)
  }
  if (addProp == "combined2") {
    assignInMyNamespace("rxErrEnv.combined", "^2")
  } else {
    assignInMyNamespace("rxErrEnv.combined", "")
  }
  if (is(x, "function")) {
    x <- rxAddReturn(x, ret != "")
  }
  if (is(substitute(x), "character")) {
    ret <- eval(parse(text = sprintf("rxode2:::rxParseErr(quote({%s}),addProp=\"%s\")", x, addProp)))
    ret <- substring(ret, 3, nchar(ret) - 2)
    assignInMyNamespace("rxErrEnv.diag.est", NULL)
    assignInMyNamespace("rxErrEnv.theta", 1)
    assignInMyNamespace("rxErrEnv.ret", "rx_r_")
    assignInMyNamespace("rxErrEnv.init", NULL)
    return(ret)
  } else if (is(substitute(x), "name")) {
    ret <- eval(parse(text = sprintf("rxode2:::rxParseErr(%s, addProp=\"%s\")", deparse1(x), addProp)))
    assignInMyNamespace("rxErrEnv.diag.est", NULL)
    assignInMyNamespace("rxErrEnv.theta", 1)
    assignInMyNamespace("rxErrEnv.ret", "rx_r_")
    assignInMyNamespace("rxErrEnv.init", NULL)
    return(ret)
  } else {
    ret <- NULL
    if (is(x, "character")) {
      ret <- eval(parse(text = sprintf("rxode2:::rxParseErr(quote({%s}))", paste(x, collapse = "\n"))))
      ret <- substring(ret, 3, nchar(ret) - 2)
    } else {
      x <- .convStr(x)
      ret <- eval(x, rxErrEnv(x))
    }
    attr(ret, "ini") <- rxErrEnv.diag.est
    assignInMyNamespace("rxErrEnv.diag.est", NULL)
    assignInMyNamespace("rxErrEnv.theta", 1)
    assignInMyNamespace("rxErrEnv.ret", "rx_r_")
    assignInMyNamespace("rxErrEnv.init", NULL)
    return(ret)
  }
}

rxSimpleExprP <- function(x) {
  if (is.name(x) || is.atomic(x)) {
    return(TRUE)
  } else {
    return(FALSE)
  }
}

#' This function splits a function based on + or - terms
#'
#' It uses the parser and does not disturb terms within other
#' functions.  For example:
#'
#' a*exp(b+c)+d*log(e-f)-g*f
#'
#' would return
#'
#' c("a * exp(b + c)", "d * log(e - f)", "- g * f")
#'
#' @param x Quoted R expression for splitting
#' @param level Internal level of parsing
#' @param mult boolean to split based on * and / expressions instead.
#'     By default this is turned off.
#' @return character vector of the split expressions
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
rxSplitPlusQ <- function(x, level = 0, mult = FALSE) {
  if (is(x, "character") && level == 0) {
    return(eval(parse(text = sprintf("rxSplitPlusQ(quote(%s))", x))))
  }
  if (is.name(x) || is.atomic(x)) {
    if (level == 0) {
      return(paste(deparse1(x), collapse = ""))
    } else {
      return(character())
    }
  } else if (is.call(x)) { # Call recurse_call recursively
    if ((mult && ((identical(x[[1]], quote(`*`)) ||
      identical(x[[1]], quote(`/`))) && level == 0)) ||
      (!mult && ((identical(x[[1]], quote(`+`)) ||
        identical(x[[1]], quote(`-`))) && level == 0))) {
      if (length(x) == 3) {
        if (identical(x[[1]], quote(`+`))) {
          one <- paste(deparse1(x[[3]]), collapse = "")
        } else if (!mult) {
          one <- paste("-", paste(deparse1(x[[3]]), collapse = ""))
        } else if (identical(x[[1]], quote(`*`))) {
          one <- paste(deparse1(x[[3]]), collapse = "")
        } else if (mult) {
          one <- paste("1/", paste(deparse1(x[[3]]), collapse = ""))
        }
        tmp <- rxSplitPlusQ(x[[2]], level = 0, mult = mult)
        if (length(tmp) > 0) {
          return(c(tmp, one))
        } else {
          tmp <- paste(deparse1(x[[2]]), collapse = "")
          return(c(tmp, one))
        }
      } else {
        ## Unary + or -
        if (identical(x[[1]], quote(`+`))) {
          one <- paste(deparse1(x[[2]]), collapse = "")
        } else {
          one <- paste("-", paste(deparse1(x[[2]]), collapse = ""))
        }
        return(one)
      }
    } else {
      tmp <- unlist(lapply(x, rxSplitPlusQ, level = 1, mult = mult))
      if (level == 0) {
        if (length(tmp) == 0) {
          tmp <- paste(deparse1(x), collapse = "")
        }
      }
      return(tmp)
    }
  } else if (is.pairlist(x)) {
    ## Call recurse_call recursively
    tmp <- unlist(lapply(x, rxSplitPlusQ, level = level, mult = mult))
    if (level == 0) {
      if (length(tmp) == 0) {
        tmp <- paste(deparse1(x), collapse = "")
      }
    }
    return(tmp)
  } else { # User supplied incorrect input
    stop("Don't know how to handle type '", typeof(x), "'.",
      call. = FALSE
    )
  }
}

.rxSupportedFunsExtra <- FALSE
.rxSupportedFuns <- function(extra = .rxSupportedFunsExtra) {
  .ret <- c(
    names(.rxSEsingle), names(.rxSEdouble), names(.rxSEeq),
    "linCmt", names(.rxOnly), ls(.symengineFs())
  )
  if (extra) {
    .ret <- c(.ret, c("linCmtA","linCmtB",
      "rxEq", "rxNeq", "rxGeq", "rxLeq", "rxLt",
      "rxGt", "rxAnd", "rxOr", "rxNot", "dabs", "dabs2", "abs0",
      "dabs1", "abs1"
    ))
  }

  .df <- rxode2parseGetTranslation()
  .w <- which(.df$package %in% c("rxode2", "rxode2ll"))
  .df <- .df[-.w, ,drop=FALSE]

  .ret <- c(.ret, .df$rxFun)

  # remove operators
  .ret <- setdiff(.ret,
                  c("==", "!=", ">=", "<=", "<", ">", "&&", "||", "&", "|", "!", "+", "-", "*", "**", "^", "/"))
  .ret
}


#' Get list of supported functions
#'
#' @return list of supported functions in rxode2
#' @examples
#' rxSupportedFuns()
#' @export
rxSupportedFuns <- function() {
  .rxSupportedFuns(FALSE)
}

.rxFunEq <- c(
  "Rx_pow_di"=2,
  "Rx_pow"=2,
  "R_pow_di"=2,
  "R_pow"=2,
  "lgamma" = 1,
  "abs" = 1,
  "acos" = 1,
  "acosh" = 1,
  "asin" = 1,
  "asinh" = 1,
  "atan" = 1,
  "atan2" = 2,
  "atanh" = 1,
  "beta" = 2,
  "cos" = 1,
  "cosh" = 1,
  "erf" = 1,
  "erfc" = 1,
  "exp" = 1,
  "log" = 1,
  "sin" = 1,
  "sinh" = 1,
  "sqrt" = 1,
  "tan" = 1,
  "tanh" = 1,
  ## C's math.h library
  "floor" = 1,
  "round" = 1,
  "ceil" = 1,
  "trunc" = 1,
  ## Special R functions
  "bessel_i" = 3,
  "bessel_j" = 2,
  "bessel_k" = 3,
  "bessel_y" = 2,
  "logspace_add" = 2,
  "logspace_sub" = 2,
  "fmax2" = 2,
  "fmin2" = 2,
  "sign" = 1,
  "fsign" = 2,
  "fprec" = 2,
  "fround" = 2,
  "ftrunc" = 2,
  "transit" = NA,
  "gammaq" = 2,
  "gammapDer" = 2,
  "gammapInv" = 2,
  "gammapInva" = 2,
  "gammaqInv" = 2,
  "gammaqInva" = 2,
  "lowergamma" = 2,
  "uppergamma" = 2)

.rxOnly <- c(
  ## Now random number generators
  "rnorm" = NA,
  "rxnorm" = NA,
  "rxbinom" = 2,
  "rbinom" = 2,
  "rxcauchy" = NA,
  "rcauchy" = NA,
  "rchisq" = 1,
  "rxchisq" = 1,
  "rexp" = 1,
  "rxexp" = 1,
  "rbeta" = 2,
  "rxbeta" = 2,
  "rgeom" = 1,
  "rxgeom" = 1,
  "rxpois" = 1,
  "rpois" = 1,
  "rxt" = 1,
  "rt" = 1
)


.rxFun2cNameOrAtomic <- function(x, envir) {
  # see if it is a reserved rxode2 name/function for name clashes
  x <- as.character(x)
  if (!exists("res", envir=envir)) {
    envir$res <- c(rxSupportedFuns(),
                    rxode2::rxReservedKeywords[, 1],
                    strsplit(paste(rxode2::rxReservedKeywords[, 3],collapse=","),"[,]+")[[1]])
  }
  if (x %in% envir$funs) {
    return(paste0("_qf_", x))
  }
  x
}

.rxFun2cArithmeticOperators <- function(x, envir) {
  if (length(x) == 3) {
    if (identical(x[[1]], quote(`/`))) {
      .x2 <- x[[2]]
      .x3 <- x[[3]]
      ## df(%s)/dy(%s)
      if (identical(.x2, quote(`d`)) &&
            identical(.x3[[1]], quote(`dt`))) {
        if (length(.x3[[2]]) == 1) {
          .state <- as.character(.x3[[2]]) # .rxToSE(.x3[[2]], envir = envir)
        } else {
          .state <- .rxFun2c(.x3[[2]], envir = envir)
        }
        stop("d/dt(", .state, ") not supported in functions for translation")
      } else {
        if (length(.x2) == 2 && length(.x3) == 2) {
          if (identical(.x2[[1]], quote(`df`)) &&
                identical(.x3[[1]], quote(`dy`))) {
            if (length(.x2[[2]]) == 1) {
              .state <- as.character(.x2[[2]])
            } else {
              .state <- .rxFun2c(.x2[[2]], envir = envir)
            }
            if (length(.x3[[2]]) == 1) {
              .var <- as.character(.x3[[2]])
            } else {
              .var <- .rxFun2c(.x3[[2]], envir = envir)
            }
            stop("df(", .state, ")/dy(", .var, ") statements are not supported in translation",
                 call. = FALSE)
          }
        }
        .ret <- paste0(
          .rxFun2c(.x2, envir = envir),
          as.character(x[[1]]),
          .rxFun2c(.x3, envir = envir)
        )
      }
    } else if (identical(x[[1]], quote(`^`)) ||
                 identical(x[[1]], quote(`**`))) {
      if (is.numeric(x[[3]]) &&
            checkmate::checkIntegerish(x[[3]])) {
        return(paste0("R_pow_di(", .rxFun2c(x[[2]], envir = envir), ",",
                     as.character(x[[3]]), ")"))
      } else {
        return(paste0("R_pow(", .rxFun2c(x[[2]], envir=envir), ",",
                     .rxFun2c(x[[3]], envir=envir), ")"))
      }
    } else {
      .ret <- paste0(
        .rxFun2c(x[[2]], envir = envir),
        as.character(x[[1]]),
        .rxFun2c(x[[3]], envir = envir)
      )
    }
    return(.ret)
  } else {
    ## Unary Operators
    return(paste(
      as.character(x[[1]]),
      .rxFun2c(x[[2]], envir = envir)
    ))
  }
}
.rxFun2cAssignOperators <- function(x, envir = envir) {
  ## if (!envir$isRx && identical(x[[1]], quote(`~`))) {
  ##   stop("formulas or other expressions with '~` are not supported in translation",
  ##        call.=FALSE)
  ## }
  if (as.character(x[[2]]) %in% envir$args) {
    stop("cannot assign argument '", as.character(x[[2]]),
         "' in functions converted to C",
         call.=FALSE)
  }
  .lhs <- .rxFun2cNameOrAtomic(x[[2]], envir=envir)
  if (!(.lhs %in% envir$args)) {
    envir$vars <- c(envir$vars, .lhs)
  }
  envir$didAssign <- TRUE
  .pre <- paste0(rep(" ", envir$n), collapse="")
  if (envir$isRx) {
    paste0(.lhs, " <- ",
           .rxFun2c(x[[3]], envir=envir), "\n",
           "rxLastValue <-", .lhs, "\n")
  } else {
    paste0(.pre, "_lastValue = ", .lhs, " = ",
                  .rxFun2c(x[[3]], envir=envir), ";\n")
  }
}

.rxFun2cSquareBracket <- function(x, envir) {
  stop("bracket expressions (ie ret[3]) are not supported in translation",
       call. = FALSE)
}

.rxFun2cLogic <- function(x, envir) {
  if (identical(x[[1]], quote(`!`))) {
    return(paste0("!(", .rxFun2c(x[[2]], envir=envir), ")"))
  } else if (identical(x[[1]], quote(`&`))) {
    return(paste0(.rxFun2c(x[[2]], envir=envir), " && ",
                  .rxFun2c(x[[3]], envir=envir)))
  } else if (identical(x[[1]], quote(`|`))) {
    return(paste0(.rxFun2c(x[[2]], envir=envir), " || ",
                  .rxFun2c(x[[3]], envir=envir)))
  } else {
    return(paste0(.rxFun2c(x[[2]], envir=envir), " ", as.character(x[[1]]), " ",
                  .rxFun2c(x[[3]], envir=envir)))
  }
}

.rxFun2cIf <- function(x, envir) {
  .logic <- .rxFun2c(x[[2]], envir=envir)
  .pre <- paste0(rep(" ", envir$n), collapse="")
  .ret <- paste(.pre, "if (", .logic, ") ")
  .ret <- paste0(.ret, .rxFun2c(x[[3]], envir=envir))

  if (length(x) == 3) {
    envir$isExpr <- TRUE
    return(.ret)
  }
  .ret <- sub("else +if", "else if", paste0(.ret, .pre, "else ", .rxFun2c(x[[4]], envir=envir)))
  envir$isExpr <- TRUE
  return(.ret)
}

.rxFun2cCall <- function(x, envir) {
  if (identical(x[[1]], quote(`{`))) {
    .ret <- "{\n"
    envir$n <- envir$n + 2
    .ret <- paste0(.ret,
                   paste(vapply(seq_along(x)[-1],
                                function(i) {
                                  .cur <- x[[i]]
                                  .last <- envir$didAssign
                                  .expr <-  envir$isExpr
                                  on.exit({
                                    assign("isExpr", .expr, envir=envir)
                                    assign("didAssign", .last, envir=envir)
                                  })
                                  envir$didAssign <- FALSE
                                  .cur <- .rxFun2c(.cur, envir=envir)
                                  if(!envir$didAssign && !envir$isExpr) {
                                    .pre <- paste0(rep(" ", envir$n), collapse="")
                                    if (envir$isRx) {
                                      return(paste0(.pre, "rxLastValue <- ", .cur, "\n"))
                                    } else {
                                      return(paste0(.pre, "_lastValue = ", .cur, ";\n"))
                                    }
                                  }
                                  .cur
                                }, character(1), USE.NAMES = FALSE),
                         collapse=""))
    envir$n <- envir$n - 2
    .pre <- paste0(rep(" ", envir$n), collapse="")
    .ret <- paste0(.ret, .pre, "}\n")
    return(.ret)
  } else if (identical(x[[1]], quote(`(`))) {
    return(paste0("(", .rxFun2c(x[[2]], envir = envir), ")"))
  } else if (identical(x[[1]], quote(`&`)) ||
               identical(x[[1]], quote(`&&`)) ||
               identical(x[[1]], quote(`==`)) ||
               identical(x[[1]], quote(`||`)) ||
               identical(x[[1]], quote(`|`)) ||
               identical(x[[1]], quote(`>`)) ||
               identical(x[[1]], quote(`<`)) ||
               identical(x[[1]], quote(`<=`)) ||
               identical(x[[1]], quote(`>=`)) ||
               identical(x[[1]], quote(`!=`)) ||
               identical(x[[1]], quote(`!`))
               ) {
    return(.rxFun2cLogic(x, envir=envir))
  } else if (identical(x[[1]], quote(`*`)) ||
               identical(x[[1]], quote(`**`)) ||
               identical(x[[1]], quote(`^`)) ||
               identical(x[[1]], quote(`+`)) ||
               identical(x[[1]], quote(`-`)) ||
               identical(x[[1]], quote(`/`))) {
    return(.rxFun2cArithmeticOperators(x, envir = envir))
  } else if (identical(x[[1]], quote(`=`)) ||
               identical(x[[1]], quote(`<-`)) ||
               identical(x[[1]], quote(`~`))) {
    return(.rxFun2cAssignOperators(x, envir))
  } else if (identical(x[[1]], quote(`[`))) {
    return(.rxFun2cSquareBracket(x, envir = envir))
  } else if (identical(x[[1]], quote(`if`))) {
    return(.rxFun2cIf(x, envir = envir))
  } else {
    # supported functions
    if (identical(x[[1]], quote(`return`))) {
      envir$hasReturn <- TRUE
      .pre <- paste0(rep(" ", envir$n), collapse="")
      envir$didAssign <- TRUE
      return(paste0(.pre, "return (", .rxFun2c(x[[2]], envir=envir), ");\n"))
    }
    .ret0 <- lapply(x, .stripP)
    .FunEq <- c(.rxFunEq, .rxSEeqUsr())
    .curName <- paste(.ret0[[1]])
    .nargs <- .FunEq[.curName]
    if (!is.na(.nargs)) {
      if (.nargs == length(.ret0) - 1) {
        return(paste0(.curName, "(",
                      paste(vapply(seq_along(.ret0)[-1],
                                  function(i) {
                                    .rxFun2c(.ret0[[i]], envir=envir)
                                  }, character(1), USE.NAMES=FALSE),
                            collapse=","),
                      ")"))
      }
    }
    stop("cannot translate function '", .curName, "'",
         call.=FALSE)
  }
}

.rxFun2c <- function(x, envir) {
  if (is.name(x) || is.atomic(x)) {
    return(.rxFun2cNameOrAtomic(x, envir=envir))
  } else if (is.call(x)) {
    return(.rxFun2cCall(x, envir = envir))
  } else {
    stop("unsupported expression", call. = FALSE)
  }
}

#' Calculate derivatives/C code from a R function
#'
#' @param fun Function to convert to C
#'
#' @param name function name to convert to C, implied if needed
#'
#' @param onlyF Only calculate the C for the function, don't calculate the derivatives
#'
#' @return A list with C code and derivative information
#'
#' @keywords internal
#'
#' @noRd
rxFun2c <- function(fun, name, onlyF=FALSE) {
  .env <- new.env(parent=emptyenv())
  .env$vars <- character(0)
  if (!missing(name)) {
    .funName <- name
  } else {
    .funName <- as.character(substitute(fun))
  }
  .f <- formals(fun)
  .env$args <- names(.f)
  .env$n <- 2
  .env$isExpr <- FALSE
  .env$isRx <- FALSE
  .env$hasReturn <- FALSE
  if (any(.env$args == "...")) {
    stop("functions with ... in them are not supported",
         call. =FALSE)
  }
  .start <- paste0("double ", .funName, "(", paste(paste("double ", .env$args), collapse=", "),
                   ") {\n")

  .body <- as.list(body(fun))
  .body <- paste(vapply(seq_along(.body)[-1], function(i) {
    .extra <- .extra2 <- ""
    .cur <- .body[[i]]
    .env$didAssign <- FALSE
    .cur <- .rxFun2c(.cur, envir=.env)
    if(!.env$didAssign && !.env$isExpr) {
      .pre <- paste0(rep(" ", .env$n), collapse="")
      if (.env$isRx) {
        return(paste0(.pre, "rxLastValue <- ", .cur, "\n"))
      } else {
        return(paste0(.pre, "_lastValue = ", .cur, ";\n"))
      }
    }
    .env$isExpr <- FALSE
    .cur
  },
  character(1), USE.NAMES=FALSE), collapse="")

  .start <- paste0(.start,
                   paste0("  double ",
                          paste(paste0(c("_lastValue", unique(.env$vars)), "=NA_REAL"), collapse=","),
                          ";\n"))
  .stop <- "  return _lastValue;\n}\n"
  .cCode <- paste0(.start, .body, .stop)
  .ret <- list(name=.funName,
               args=.env$args,
               cCode=.cCode)

  if (onlyF) {
    return(.ret)
  }
  if (!.env$hasReturn) {
    # Can calculate derivatives
    # Firs create an rxode2 like model:
    .env <- new.env(parent=emptyenv())
    .env$isRx <- TRUE
    .env$args <- names(.f)
    .env$n <- 2
    .env$isExpr <- FALSE
    .env$hasReturn <- FALSE
    .body <- as.list(body(fun))
    .body <- paste(vapply(seq_along(.body)[-1], function(i) {
      .extra <- .extra2 <- ""
      .cur <- .body[[i]]
      .env$didAssign <- FALSE
      .cur <- .rxFun2c(.cur, envir=.env)
      if(!.env$didAssign && !.env$isExpr) {
        .pre <- paste0(rep(" ", .env$n), collapse="")
        return(paste0(.pre, "rxLastValue = ", .cur, ";\n"))
      }
      .env$isExpr <- FALSE
      .cur
    },
    character(1), USE.NAMES=FALSE), collapse="")
    # take out if/else
    .body <- rxPrune(.body)
    .s <- rxS(.body)
    .lastValue <- .s$rxLastValue
    return(c(list(.ret),
      lapply(.env$args, function(v) {
      .v <- symengine::D(.lastValue, symengine::S(v))
      .v <- paste0("function(", paste(.env$args, collapse=", "), ") {\n", rxOptExpr(paste0("rxLastValue=", rxFromSE(.v)), msg=paste0("d(", .funName, ")/d(", v, ")")),
                   "\nrxLastValue}")
      .v <- eval(str2lang(.v))
      .dName <-  paste0("rx_", .funName, "_d_", v)
      .v <- rxFun2c(.v, .dName, onlyF=TRUE)
      .v2 <- paste0("function(", paste(.env$args, collapse=", "), "){\n",
                  "paste0(\"", .dName, "(\", ",
                  paste(.env$args, collapse=", \", \", "), ", \")\")",
                  "}")
      .v2 <- eval(str2lang(.v2))
      c(.v, list(.v2))
    })))
  } else {
    message("function contains return statement; derivatives not calculated")
  }
  return(list(.ret))
}

## Tests for external C package integration as described in
## vignettes/articles/rxode2-pkg-exported-funs.Rmd
##
## Strategy: build the complete 'mm' package (exactly as the vignette
## describes), install it to a temporary library, attach it, run all
## assertions, then detach it and verify the cleanup hooks fire correctly.

rxTest({

  ## -- package-building helpers ---------------------------------------------

  .writeMmPackage <- function(pkgDir) {

    dir.create(file.path(pkgDir, "R"),   recursive = TRUE, showWarnings = FALSE)
    dir.create(file.path(pkgDir, "src"), recursive = TRUE, showWarnings = FALSE)

    ## DESCRIPTION ---------------------------------------------------------
    writeLines(c(
      "Package: mm",
      "Version: 0.1.0",
      "Title: Michaelis-Menten rxode2 Functions",
      "Description: Test package for rxode2 external C integration.",
      "Authors@R: person('Test', 'Author', role = c('aut', 'cre'),",
      "    email = 'test@test.com')",
      "License: MIT + file LICENSE",
      "Imports: rxode2",
      "LinkingTo: rxode2",
      "Encoding: UTF-8"
    ), file.path(pkgDir, "DESCRIPTION"))

    ## LICENSE stub (required by MIT + file LICENSE) -----------------------
    writeLines("YEAR: 2026\nCOPYRIGHT HOLDER: Test Author",
               file.path(pkgDir, "LICENSE"))

    ## NAMESPACE -----------------------------------------------------------
    writeLines(c(
      "useDynLib(mm, .registration = TRUE)",
      "importFrom(rxode2, rxD)",
      "importFrom(rxode2, rxRmFun)",
      "importFrom(rxode2, rxUdfUi)",
      "importFrom(rxode2, rxode2parseAssignTranslation)",
      "importFrom(rxode2, rxode2parseGetTranslation)",
      "export(mm)",
      "export(mm_dC)",
      "export(mm_dVmax)",
      "export(mm_dKm)",
      "S3method(rxUdfUi, mm)"
    ), file.path(pkgDir, "NAMESPACE"))

    ## src/myfuns.c --------------------------------------------------------
    writeLines(c(
      "#define STRICT_R_HEADERS",
      "#include <R.h>",
      "#include <Rinternals.h>",
      "#include <R_ext/RS.h>",
      "",
      "/* Scalar functions ----------------------------------------- */",
      "double mm(double C, double Vmax, double Km) {",
      "  return Vmax * C / (Km + C);",
      "}",
      "",
      "double mm_dC(double C, double Vmax, double Km) {",
      "  double den = Km + C;",
      "  return Vmax * Km / (den * den);",
      "}",
      "",
      "double mm_dVmax(double C, double Vmax, double Km) {",
      "  (void)Vmax;",
      "  return C / (Km + C);",
      "}",
      "",
      "double mm_dKm(double C, double Vmax, double Km) {",
      "  double den = Km + C;",
      "  return -Vmax * C / (den * den);",
      "}",
      "",
      "/* SEXP vector wrappers ---------------------------------------- */",
      "SEXP _mm_mm(SEXP C, SEXP Vmax, SEXP Km) {",
      "  int n = LENGTH(C);",
      "  SEXP out = PROTECT(allocVector(REALSXP, n));",
      "  double *c = REAL(C), *vmax = REAL(Vmax), *km = REAL(Km);",
      "  double *res = REAL(out);",
      "  for (int i = 0; i < n; i++) res[i] = mm(c[i], vmax[i], km[i]);",
      "  UNPROTECT(1); return out;",
      "}",
      "SEXP _mm_mm_dC(SEXP C, SEXP Vmax, SEXP Km) {",
      "  int n = LENGTH(C);",
      "  SEXP out = PROTECT(allocVector(REALSXP, n));",
      "  double *c = REAL(C), *vmax = REAL(Vmax), *km = REAL(Km);",
      "  double *res = REAL(out);",
      "  for (int i = 0; i < n; i++) res[i] = mm_dC(c[i], vmax[i], km[i]);",
      "  UNPROTECT(1); return out;",
      "}",
      "SEXP _mm_mm_dVmax(SEXP C, SEXP Vmax, SEXP Km) {",
      "  int n = LENGTH(C);",
      "  SEXP out = PROTECT(allocVector(REALSXP, n));",
      "  double *c = REAL(C), *vmax = REAL(Vmax), *km = REAL(Km);",
      "  double *res = REAL(out);",
      "  for (int i = 0; i < n; i++) res[i] = mm_dVmax(c[i], vmax[i], km[i]);",
      "  UNPROTECT(1); return out;",
      "}",
      "SEXP _mm_mm_dKm(SEXP C, SEXP Vmax, SEXP Km) {",
      "  int n = LENGTH(C);",
      "  SEXP out = PROTECT(allocVector(REALSXP, n));",
      "  double *c = REAL(C), *vmax = REAL(Vmax), *km = REAL(Km);",
      "  double *res = REAL(out);",
      "  for (int i = 0; i < n; i++) res[i] = mm_dKm(c[i], vmax[i], km[i]);",
      "  UNPROTECT(1); return out;",
      "}"
    ), file.path(pkgDir, "src", "myfuns.c"))

    ## src/init.c ----------------------------------------------------------
    writeLines(c(
      "#include <R.h>",
      "#include <Rinternals.h>",
      "#include <R_ext/Rdynload.h>",
      "#include <stdlib.h>",
      "",
      "extern double mm      (double, double, double);",
      "extern double mm_dC   (double, double, double);",
      "extern double mm_dVmax(double, double, double);",
      "extern double mm_dKm  (double, double, double);",
      "",
      "extern SEXP _mm_mm      (SEXP, SEXP, SEXP);",
      "extern SEXP _mm_mm_dC   (SEXP, SEXP, SEXP);",
      "extern SEXP _mm_mm_dVmax(SEXP, SEXP, SEXP);",
      "extern SEXP _mm_mm_dKm  (SEXP, SEXP, SEXP);",
      "",
      "void R_init_mm(DllInfo *dll) {",
      "  static const R_CallMethodDef callMethods[] = {",
      "    {\"_mm_mm\",       (DL_FUNC) &_mm_mm,       3},",
      "    {\"_mm_mm_dC\",    (DL_FUNC) &_mm_mm_dC,    3},",
      "    {\"_mm_mm_dVmax\", (DL_FUNC) &_mm_mm_dVmax, 3},",
      "    {\"_mm_mm_dKm\",   (DL_FUNC) &_mm_mm_dKm,   3},",
      "    {NULL, NULL, 0}",
      "  };",
      "  R_registerRoutines(dll, NULL, callMethods, NULL, NULL);",
      "  R_useDynamicSymbols(dll, FALSE);",
      "",
      "  R_RegisterCCallable(\"mm\", \"mm\",       (DL_FUNC) &mm);",
      "  R_RegisterCCallable(\"mm\", \"mm_dC\",    (DL_FUNC) &mm_dC);",
      "  R_RegisterCCallable(\"mm\", \"mm_dVmax\", (DL_FUNC) &mm_dVmax);",
      "  R_RegisterCCallable(\"mm\", \"mm_dKm\",   (DL_FUNC) &mm_dKm);",
      "}"
    ), file.path(pkgDir, "src", "init.c"))

    ## R/mm.R --------------------------------------------------------------
    writeLines(c(
      "mm <- function(C, Vmax, Km) {",
      "  df <- data.frame(C = C, Vmax = Vmax, Km = Km)",
      "  .Call(`_mm_mm`, as.double(df$C), as.double(df$Vmax), as.double(df$Km))",
      "}",
      "mm_dC <- function(C, Vmax, Km) {",
      "  df <- data.frame(C = C, Vmax = Vmax, Km = Km)",
      "  .Call(`_mm_mm_dC`, as.double(df$C), as.double(df$Vmax), as.double(df$Km))",
      "}",
      "mm_dVmax <- function(C, Vmax, Km) {",
      "  df <- data.frame(C = C, Vmax = Vmax, Km = Km)",
      "  .Call(`_mm_mm_dVmax`, as.double(df$C), as.double(df$Vmax), as.double(df$Km))",
      "}",
      "mm_dKm <- function(C, Vmax, Km) {",
      "  df <- data.frame(C = C, Vmax = Vmax, Km = Km)",
      "  .Call(`_mm_mm_dKm`, as.double(df$C), as.double(df$Vmax), as.double(df$Km))",
      "}",
      "",
      "rxUdfUi.mm <- function(fun) {",
      "  .dummy <- function(C, Vmax, Km) {}",
      "  .mc <- match.call(.dummy, fun)",
      "  .C    <- deparse1(.mc$C)",
      "  .Vmax <- deparse1(.mc$Vmax)",
      "  .Km   <- deparse1(.mc$Km)",
      "  list(replace = paste0('mm(', .C, ',', .Vmax, ',', .Km, ')'))",
      "}"
    ), file.path(pkgDir, "R", "mm.R"))

    ## R/zzz.R -------------------------------------------------------------
    writeLines(c(
      ".onAttach <- function(libname, pkgname) {",
      "  .cur <- rxode2parseGetTranslation()",
      "  .newRows <- data.frame(",
      "    rxFun      = c('mm',          'mm_dC',       'mm_dVmax',    'mm_dKm'),",
      "    fun        = c('mm',          'mm_dC',        'mm_dVmax',   'mm_dKm'),",
      "    type       = rep('rxode2_fn3', 4),",
      "    package    = rep('mm', 4),",
      "    packageFun = c('mm',          'mm_dC',        'mm_dVmax',   'mm_dKm'),",
      "    argMin     = 3L,",
      "    argMax     = 3L,",
      "    threadSafe = 1L,",
      "    stringsAsFactors = FALSE",
      "  )",
      "  .mmNames   <- .newRows$rxFun",
      "  .conflicts <- .cur[.cur$rxFun %in% .mmNames, , drop = FALSE]",
      "  if (nrow(.conflicts) > 0L) {",
      "    .fromOther <- .conflicts[.conflicts$package != 'mm', , drop = FALSE]",
      "    .fromSelf  <- .conflicts[.conflicts$package == 'mm', , drop = FALSE]",
      "    if (nrow(.fromOther) > 0L) {",
      "      stop(",
      "        \"Cannot load package 'mm': the following rxode2 function name(s) are \",",
      "        \"already registered by another package:\\n\",",
      "        paste0(\"  \", .fromOther$rxFun, \" (registered by '\",",
      "               .fromOther$package, \"')\", collapse = \"\\n\"),",
      "        call. = FALSE",
      "      )",
      "    }",
      "    if (nrow(.fromSelf) > 0L) {",
      "      packageStartupMessage(",
      "        \"Package 'mm' rxode2 function registration was already present \",",
      "        \"(package reloaded); re-registering.\"",
      "      )",
      "    }",
      "  }",
      "  .cur <- .cur[.cur$package != 'mm', , drop = FALSE]",
      "  rxode2parseAssignTranslation(rbind(.cur, .newRows))",
      "",
      "  rxD('mm', list(",
      "    function(C, Vmax, Km) paste0('mm_dC(',    C, ',', Vmax, ',', Km, ')'),",
      "    function(C, Vmax, Km) paste0('mm_dVmax(', C, ',', Vmax, ',', Km, ')'),",
      "    function(C, Vmax, Km) paste0('mm_dKm(',   C, ',', Vmax, ',', Km, ')')",
      "  ))",
      "",
      "  rxD('mm_dC', list(",
      "    function(C, Vmax, Km)",
      "      paste0('(-2.0*(',Vmax,')*(',Km,'))/pow((',Km,')+(',C,'),3.0)'),",
      "    function(C, Vmax, Km)",
      "      paste0('(',Km,')/pow((',Km,')+(',C,'),2.0)'),",
      "    function(C, Vmax, Km)",
      "      paste0('(',Vmax,')*((',C,')-(',Km,'))/pow((',Km,')+(',C,'),3.0)')",
      "  ))",
      "",
      "  rxD('mm_dVmax', list(",
      "    function(C, Vmax, Km)",
      "      paste0('(',Km,')/pow((',Km,')+(',C,'),2.0)'),",
      "    function(C, Vmax, Km) '0.0',",
      "    function(C, Vmax, Km)",
      "      paste0('-(',C,')/pow((',Km,')+(',C,'),2.0)')",
      "  ))",
      "",
      "  rxD('mm_dKm', list(",
      "    function(C, Vmax, Km)",
      "      paste0('(',Vmax,')*((',C,')-(',Km,'))/pow((',Km,')+(',C,'),3.0)'),",
      "    function(C, Vmax, Km)",
      "      paste0('-(',C,')/pow((',Km,')+(',C,'),2.0)'),",
      "    function(C, Vmax, Km)",
      "      paste0('2.0*(',Vmax,')*(',C,')/pow((',Km,')+(',C,'),3.0)')",
      "  ))",
      "  invisible()",
      "}",
      "",
      ".onUnload <- function(libpath) {",
      "  .cur <- try(rxode2parseGetTranslation(), silent = TRUE)",
      "  if (!inherits(.cur, 'try-error')) {",
      "    try(rxode2parseAssignTranslation(",
      "      .cur[.cur$package != 'mm', , drop = FALSE]",
      "    ), silent = TRUE)",
      "  }",
      "  suppressWarnings(try(rxRmFun('mm'),       silent = TRUE))",
      "  suppressWarnings(try(rxRmFun('mm_dC'),    silent = TRUE))",
      "  suppressWarnings(try(rxRmFun('mm_dVmax'), silent = TRUE))",
      "  suppressWarnings(try(rxRmFun('mm_dKm'),   silent = TRUE))",
      "}"
    ), file.path(pkgDir, "R", "zzz.R"))

    invisible(pkgDir)
  }

  ## -- main test ------------------------------------------------------------

  test_that("mm package: install, attach, verify integration, detach, verify cleanup", {

    ## Create package skeleton + temp library dir
    .pkgDir <- tempfile("mm_src_")
    .libDir <- tempfile("mm_lib_")
    dir.create(.pkgDir, recursive = TRUE, showWarnings = FALSE)
    dir.create(.libDir, recursive = TRUE, showWarnings = FALSE)
    on.exit({
      try(detach("package:mm", unload = TRUE, force = TRUE), silent = TRUE)
      unlink(.pkgDir, recursive = TRUE)
      unlink(.libDir, recursive = TRUE)
    }, add = TRUE)

    .writeMmPackage(.pkgDir)

    ## -- install ----------------------------------------------------------
    .out <- system2(
      file.path(R.home("bin"), "R"),
      args   = c("CMD", "INSTALL", "--no-multiarch",
                 paste0("--library=", .libDir), .pkgDir),
      env    = paste0("R_LIBS=",
                      paste(c(.libDir, .libPaths()), collapse = .Platform$path.sep)),
      stdout = TRUE, stderr = TRUE
    )
    if (!any(grepl("DONE \\(mm\\)", .out))) {
      skip(paste0("mm package failed to install (C compiler issue?):\n",
                  paste(.out, collapse = "\n")))
    }

    ## -- attach -----------------------------------------------------------
    .oldPaths <- .libPaths()
    on.exit(.libPaths(.oldPaths), add = TRUE)
    .libPaths(c(.libDir, .oldPaths))

    ## Capture startup messages from .onAttach
    expect_no_error(
      suppressPackageStartupMessages(library(mm, lib.loc = .libDir))
    )

    ## -- translation table ------------------------------------------------
    .trans <- rxode2::rxode2parseGetTranslation()
    .mm    <- .trans[.trans$package == "mm", , drop = FALSE]

    expect_equal(nrow(.mm), 4L)
    expect_equal(sort(.mm$rxFun),
                 sort(c("mm", "mm_dC", "mm_dVmax", "mm_dKm")))
    expect_true(all(.mm$type       == "rxode2_fn3"))
    expect_true(all(.mm$argMin     == 3L))
    expect_true(all(.mm$argMax     == 3L))
    expect_true(all(.mm$threadSafe == 1L))
    expect_true(all(.mm$packageFun == .mm$rxFun))

    ## mm functions are visible to the parser
    .funs <- rxode2::rxSupportedFuns()
    expect_true("mm"       %in% .funs)
    expect_true("mm_dC"    %in% .funs)
    expect_true("mm_dVmax" %in% .funs)
    expect_true("mm_dKm"   %in% .funs)

    ## -- R vector interface -----------------------------------------------
    ## mm(C, Vmax, Km) = Vmax * C / (Km + C)
    expect_equal(mm(2,  10, 2), 10 * 2  / (2 + 2))     # 5
    expect_equal(mm(0,  10, 2), 0)
    expect_equal(
      mm(c(1, 2, 5), 10, 2),
      10 * c(1, 2, 5) / (2 + c(1, 2, 5))
    )

    ## mm_dC(C, Vmax, Km) = Vmax * Km / (Km + C)^2
    expect_equal(mm_dC(2, 10, 2), 10 * 2 / (2 + 2)^2)

    ## mm_dVmax(C, Vmax, Km) = C / (Km + C)
    expect_equal(mm_dVmax(2, 10, 2), 2 / (2 + 2))

    ## mm_dKm(C, Vmax, Km) = -Vmax * C / (Km + C)^2
    expect_equal(mm_dKm(2, 10, 2), -10 * 2 / (2 + 2)^2)

    ## -- rxUdfUi named-argument reordering --------------------------------
    ## Positional call -- already canonical
    .c1 <- quote(mm(Cp, Vmax, Km))
    class(.c1) <- "mm"
    expect_equal(rxode2::rxUdfUi(.c1)$replace, "mm(Cp,Vmax,Km)")

    ## Named in canonical order
    .c2 <- quote(mm(C = Cp, Vmax = Vmax, Km = Km))
    class(.c2) <- "mm"
    expect_equal(rxode2::rxUdfUi(.c2)$replace, "mm(Cp,Vmax,Km)")

    ## Named in different order -- must be reordered to C, Vmax, Km
    .c3 <- quote(mm(Km = Km, Vmax = Vmax, C = Cp))
    class(.c3) <- "mm"
    expect_equal(rxode2::rxUdfUi(.c3)$replace, "mm(Cp,Vmax,Km)")

    ## -- ODE simulation ---------------------------------------------------
    .mod <- rxode2::rxode2({
      Cp       <- A1 / Vc
      d/dt(A1) <- -mm(Cp, Vmax, Km)
    })
    .ev  <- rxode2::et(amt = 100) |> rxode2::et(0, 24, by = 0.5)
    .sim <- rxode2::rxSolve(.mod, .ev, params = c(Vc = 10, Vmax = 5, Km = 1))

    .A1 <- .sim$A1

    ## Starts at 100 immediately post-dose
    expect_equal(.A1[1], 100, tolerance = 1e-3)

    ## Monotonically decreasing (saturable elimination)
    expect_true(all(diff(.A1) <= 0))

    ## Substantially cleared after 24 h (> 90% eliminated)
    expect_lt(.A1[length(.A1)], 0.1 * .A1[1])

    ## -- conflict detection -----------------------------------------------
    ## Attempting to add the same names from a different package is
    ## detected by the conflict-check logic in .onAttach.
    .cur <- rxode2::rxode2parseGetTranslation()
    .other <- data.frame(
      rxFun = "mm", fun = "mm", type = "rxode2_fn3",
      package = "other_pkg", packageFun = "mm",
      argMin = 3L, argMax = 3L, threadSafe = 1L,
      stringsAsFactors = FALSE
    )
    ## The conflict rows (from "mm", not "other_pkg") would be detected
    .conflicts <- .cur[.cur$rxFun %in% .other$rxFun, , drop = FALSE]
    .fromOther  <- .conflicts[.conflicts$package != "other_pkg", , drop = FALSE]
    expect_true(nrow(.fromOther) > 0L)
    expect_equal(.fromOther$package[1], "mm")

    ## -- detach and verify cleanup ----------------------------------------
    detach("package:mm", unload = TRUE, force = TRUE)

    .trans2 <- rxode2::rxode2parseGetTranslation()
    expect_false("mm"       %in% .trans2$rxFun)
    expect_false("mm_dC"    %in% .trans2$rxFun)
    expect_false("mm_dVmax" %in% .trans2$rxFun)
    expect_false("mm_dKm"   %in% .trans2$rxFun)

    ## mm no longer visible to the parser
    .funs2 <- rxode2::rxSupportedFuns()
    expect_false("mm"    %in% .funs2)
    expect_false("mm_dC" %in% .funs2)

    ## R functions gone from the search path
    expect_false("package:mm" %in% search())
  })

  ## -- reload (re-attach) test -----------------------------------------------

  test_that("mm package: re-attach emits startup message, does not error", {

    .pkgDir <- tempfile("mm_src2_")
    .libDir <- tempfile("mm_lib2_")
    dir.create(.pkgDir, recursive = TRUE, showWarnings = FALSE)
    dir.create(.libDir, recursive = TRUE, showWarnings = FALSE)
    on.exit({
      try(detach("package:mm", unload = TRUE, force = TRUE), silent = TRUE)
      unlink(.pkgDir, recursive = TRUE)
      unlink(.libDir, recursive = TRUE)
    }, add = TRUE)

    .writeMmPackage(.pkgDir)

    .out <- system2(
      file.path(R.home("bin"), "R"),
      args   = c("CMD", "INSTALL", "--no-multiarch",
                 paste0("--library=", .libDir), .pkgDir),
      env    = paste0("R_LIBS=",
                      paste(c(.libDir, .libPaths()), collapse = .Platform$path.sep)),
      stdout = TRUE, stderr = TRUE
    )
    if (!any(grepl("DONE \\(mm\\)", .out))) {
      skip(paste0("mm package failed to install:\n", paste(.out, collapse = "\n")))
    }

    .oldPaths <- .libPaths()
    on.exit(.libPaths(.oldPaths), add = TRUE)
    .libPaths(c(.libDir, .oldPaths))

    ## First attach -- no startup message about re-registering
    suppressPackageStartupMessages(library(mm, lib.loc = .libDir))

    ## Simulate a "reload" by directly calling .onAttach again; the
    ## existing rows from "mm" are still in the table so the re-register
    ## path should fire (emitting a packageStartupMessage).
    ## rxD() also warns "replacing defined derivatives" on re-registration
    ## -- those warnings are expected and suppressed here.
    expect_message(
      suppressWarnings(mm:::.onAttach("", "mm")),
      "re-registering"
    )

    ## After re-registration, table still has exactly 4 mm rows
    .mm <- rxode2::rxode2parseGetTranslation()
    .mm <- .mm[.mm$package == "mm", , drop = FALSE]
    expect_equal(nrow(.mm), 4L)
  })

})

rxTest({
  # Test symengine<->rxode2 dsl
  test_that("d/dt(x) parsing", {
    expect_equal(rxToSE(d / dt(matt)), "rx__d_dt_matt__")
    expect_equal(rxFromSE(rx__d_dt_matt__), "d/dt(matt)")
    expect_equal(rxToSE(d / dt(matt)), "rx__d_dt_matt__")
    expect_equal(rxFromSE("rx__d_dt_matt__"), "d/dt(matt)")
    tmp <- symengine::Symbol("rx__d_dt_matt__")
    expect_equal(rxFromSE(tmp), "d/dt(matt)")
    expect_equal(rxToSE(d / dt(E)), "rx__d_dt_E__")
    expect_equal(rxFromSE("rx__d_dt_E__"), "d/dt(E)")
    tmp <- symengine::Symbol("rx__d_dt_E__")
    expect_equal(rxFromSE(tmp), "d/dt(E)")
  })

  test_that("rxFromSE() round-trips raw comparison/logical operators", {
    # The Subs handler re-parses converted text with R's parser, which turns
    # rxGt()/rxEq() back into raw `>`/`==` operator calls; .rxFromSE() must
    # accept those so a comparison inside a Derivative(linCmtB(...)) does not
    # error with "user function '>' requires 0 arguments" (nlmixr2/nlmixr2#390).
    expect_equal(rxFromSE("a > b"), "(a>b)")
    expect_equal(rxFromSE("a < b"), "(a<b)")
    expect_equal(rxFromSE("a >= b"), "(a>=b)")
    expect_equal(rxFromSE("a <= b"), "(a<=b)")
    expect_equal(rxFromSE("a == b"), "(a==b)")
    expect_equal(rxFromSE("a != b"), "(a!=b)")
    expect_equal(rxFromSE("exp(x + (a > 0))"), "exp(x+((a>0)))")
    # the derivative of a linCmtB() whose parameter contains a comparison
    # (as produced for IOV + linCmt() FOCEi models) must convert without error
    .x <- paste0("Subs(Derivative(linCmtB(rx__PTR__, t, 2.0, 1.0, 1.0, -1.0, ",
                 "-1.0, 1.0, exp(THETA_2_ + rxGt(THETA_5_, 0.0)*rxEq(occ, 1.0)), ",
                 "exp(THETA_3_), 0.0, 0.0, 0.0, 0.0, _xi_15), _xi_15), ",
                 "(_xi_15), (exp(THETA_1_)))")
    expect_error(rxFromSE(.x), NA)
  })

  test_that("df(x)/dy(x) parsing", {
    expect_equal(rxToSE(df(matt) / dy(ruth)), "rx__df_matt_dy_ruth__")
    expect_equal(rxFromSE(rx__df_matt_dy_ruth__), "df(matt)/dy(ruth)")
    expect_equal(rxToSE(df(matt) / dy(ruth)), "rx__df_matt_dy_ruth__")
    expect_equal(rxFromSE("rx__df_matt_dy_ruth__"), "df(matt)/dy(ruth)")
    tmp <- symengine::Symbol("rx__df_matt_dy_ruth__")
    expect_equal(rxFromSE(tmp), "df(matt)/dy(ruth)")
  })

  test_that("function and constant translation", {
    expect_equal(rxFromSE("x^1"), "x")
    expect_equal(rxToSE(gammafn(a)), "gamma(a)")
    expect_error(rxToSE(gammafn(a, b)))
    expect_error(rxFromSE(gamma(a, b)))

    expect_equal(rxToSE(lgammafn(a)), "lgamma(a)")
    expect_equal(rxToSE(lgammafn(a)), "lgamma(a)")
    expect_equal(rxFromSE(loggamma(a)), "lgamma(a)")
    expect_equal(rxFromSE(loggamma(1 + a)), "lgamma1p(a)")
    expect_equal(rxFromSE(lgamma(1 + a)), "lgamma1p(a)")

    expect_equal(rxToSE(digamma(a)), "polygamma(0,a)")
    expect_equal(rxFromSE(polygamma(0, a)), "digamma(a)")

    expect_equal(rxToSE(trigamma(a)), "polygamma(1,a)")
    expect_equal(rxFromSE(polygamma(1, a)), "trigamma(a)")

    expect_equal(rxToSE(tetragamma(a)), "polygamma(2,a)")
    expect_equal(rxFromSE(polygamma(2, a)), "tetragamma(a)")

    expect_equal(rxToSE(pentagamma(a)), "polygamma(3,a)")
    expect_equal(rxFromSE(polygamma(3, a)), "pentagamma(a)")

    expect_equal(rxToSE(lbeta(a, b)), "log(beta(a,b))")
    expect_equal(rxFromSE(log(beta(a, b))), "lbeta(a,b)")

    expect_equal(rxToSE(lgamma1p(a)), "lgamma(a+1)")
    expect_equal(rxFromSE(lgamma(a + 1)), "lgamma1p(a)")
    expect_equal(rxFromSE(lgamma(1 + a)), "lgamma1p(a)")
    expect_equal(rxFromSE(lgamma(a + 1 + b)), "lgamma1p(a+b)")

    expect_equal(rxToSE(log1p(a)), "log(1+a)")
    expect_equal(rxFromSE(log(a + 1)), "log1p(a)")

    expect_equal(rxToSE(cospi(a)), "cos(pi*(a))")
    expect_equal(rxFromSE(cos(pi * a)), "cospi(a)")
    expect_equal(rxFromSE(cos(pi * (a))), "cospi(a)")

    expect_equal(rxFromSE(cos(b * pi * a)), "cospi(b*a)")

    expect_equal(rxToSE(sinpi(a)), "sin(pi*(a))")
    expect_equal(rxFromSE(sin(pi * a)), "sinpi(a)")

    expect_equal(rxToSE(tanpi(a)), "tan(pi*(a))")
    expect_equal(rxFromSE(tan(pi * a)), "tanpi(a)")
    expect_equal(rxFromSE("tan(pi/2)"), "tanpi(1/2)")

    ## nlmixr2est#513: a trig argument that is a compound expression divided by
    ## something (eg sin(2*3.14*(time-mtime1)/period)) must keep its argument;
    ## the `/` branch used to fall through and drop the whole argument -> sin()
    expect_equal(rxFromSE("sin(2 * 3.14 * (time-mtime1)/period)"),
                 "sin((2*3.14*(time-mtime1))/period)")
    expect_equal(rxFromSE("cos((time-mtime1)/period)"),
                 "cos((time-mtime1)/period)")
    ## a pi factor buried in a compound numerator still folds to sinpi and keeps
    ## precedence-correct parentheses
    expect_equal(rxFromSE("sin(pi * (time-mtime1)/period)"),
                 "sinpi((time-mtime1)/period)")

    expect_equal(rxToSE(log1pmx(a)), "(log(1+a)-(a))")
    expect_equal(rxToSE(expm1(a)), "(exp(a)-1)")
    expect_equal(rxToSE(pow(a, b)), "(a)^(b)")
    expect_error(rxToSE(pow(a, b, c)))
    expect_error(rxToSE(pow(a)))
    expect_equal(rxToSE(R_pow(a, b)), "(a)^(b)")
    expect_equal(rxToSE(Rx_pow_di(a, b)), "(a)^(b)")
    expect_equal(rxToSE(Rx_pow(a, b)), "(a)^(b)")
    expect_equal(rxToSE(R_pow_di(a, b)), "(a)^(b)")
    expect_equal(rxToSE(factorial(n)), "gamma(n+1)")

    expect_equal(rxToSE(beta(a, b)), "beta(a,b)")
    expect_error(rxToSE(beta(a)))

    expect_equal(rxToSE(choose(n, k)), "gamma(n+1)/(gamma(k+1)*gamma(n-(k)+1))")
    expect_equal(rxToSE(lchoose(n, k)), "(lgamma(n+1)-lgamma(k+1)-lgamma(n-(k)+1))")

    expect_equal(rxFromSE(log(-x + 1)), "log1p(-x)")
    expect_equal(rxFromSE(log(1 - x)), "log1p(-x)")

    expect_equal(rxToSE(log1pexp(x)), "log(1+exp(x))")
    expect_equal(rxFromSE(log(1 + exp(x))), "log1pexp(x)")

    ## expect_equal(rxFromSymPy(log((1 + x)^2)), "log1p(Rx_pow_di(x, 2)+2 * x)")
    ## expect_equal(rxFromSymPy(log((0.75 + x)^2)), "log1p(Rx_pow_di(x, 2)+1.5 * x-0.4375)")

    expect_equal(rxFromSE(E), "2.7182818284590451")
    expect_equal(rxToSE(M_E), "E")

    expect_equal(rxToSE(exp(1)), "E")
    expect_equal(rxFromSE(exp(1)), "M_E")

    expect_equal(rxFromSE(log(2)), "M_LN2")
    expect_equal(rxFromSE(log(10)), "M_LN10")

    expect_equal(rxToSE("M_LN2"), "log(2)")
    expect_equal(rxFromSE("log(2)"), "M_LN2")

    expect_equal(rxToSE("M_LN10"), "log(10)")
    expect_equal(rxFromSE("log(10)"), "M_LN10")

    expect_equal(rxToSE("M_LN_SQRT_PI"), "log(sqrt(pi))")

    expect_equal(rxFromSE("log(sqrt(pi))"), "M_LN_SQRT_PI")
    expect_equal(rxFromSE("log(pi**0.5)"), "M_LN_SQRT_PI")
    expect_equal(rxFromSE("log(pi^0.5)"), "M_LN_SQRT_PI")
    expect_equal(rxFromSE("log(pi^(1/2))"), "M_LN_SQRT_PI")
    expect_equal(rxFromSE("log(pi**(1/2))"), "M_LN_SQRT_PI")

    expect_equal(rxToSE("M_LN_SQRT_PId2"), "log(sqrt(pi/2))")
    expect_equal(rxFromSE(log(sqrt(pi / 2))), "M_LN_SQRT_PId2")
    expect_equal(rxFromSE(log(sqrt(pi * 0.5))), "M_LN_SQRT_PId2")
    expect_equal(rxFromSE(log(sqrt(0.5 * pi))), "M_LN_SQRT_PId2")

    expect_equal(rxFromSE(log((pi / 2)^0.5)), "M_LN_SQRT_PId2")
    expect_equal(rxFromSE(log((pi * 0.5)^0.5)), "M_LN_SQRT_PId2")
    expect_equal(rxFromSE(log((0.5 * pi)^0.5)), "M_LN_SQRT_PId2")

    expect_equal(rxFromSE(log((pi / 2)**0.5)), "M_LN_SQRT_PId2")
    expect_equal(rxFromSE(log((pi * 0.5)**0.5)), "M_LN_SQRT_PId2")
    expect_equal(rxFromSE(log((0.5 * pi)**0.5)), "M_LN_SQRT_PId2")

    expect_equal(rxFromSE(log((pi / 2)**(1 / 2))), "M_LN_SQRT_PId2")
    expect_equal(rxFromSE(log((pi * 0.5)**(1 / 2))), "M_LN_SQRT_PId2")
    expect_equal(rxFromSE(log((0.5 * pi)**(1 / 2))), "M_LN_SQRT_PId2")

    expect_equal(rxFromSE(log((pi / 2)^(1 / 2))), "M_LN_SQRT_PId2")
    expect_equal(rxFromSE(log((pi * 0.5)^(1 / 2))), "M_LN_SQRT_PId2")
    expect_equal(rxFromSE(log((0.5 * pi)^(1 / 2))), "M_LN_SQRT_PId2")

    expect_equal(rxToSE("M_LN_SQRT_2PI"), "log(sqrt(2*pi))")
    expect_equal(rxFromSE(log((pi * 2)^0.5)), "M_LN_SQRT_2PI")
    expect_equal(rxFromSE(log((2 * pi)^0.5)), "M_LN_SQRT_2PI")

    expect_equal(rxFromSE(log((pi * 2)**0.5)), "M_LN_SQRT_2PI")
    expect_equal(rxFromSE(log((2 * pi)**0.5)), "M_LN_SQRT_2PI")

    expect_equal(rxFromSE(log((pi * 2)**(1 / 2))), "M_LN_SQRT_2PI")
    expect_equal(rxFromSE(log((2 * pi)**(1 / 2))), "M_LN_SQRT_2PI")

    expect_equal(rxFromSE(log((pi * 2)^(1 / 2))), "M_LN_SQRT_2PI")
    expect_equal(rxFromSE(log((2 * pi)^(1 / 2))), "M_LN_SQRT_2PI")


    expect_equal(rxToSE("M_SQRT_3"), "sqrt(3)")
    expect_equal(rxFromSE("sqrt(3)"), "M_SQRT_3")

    expect_equal(rxToSE("M_SQRT2"), "sqrt(2)")
    expect_equal(rxFromSE("sqrt(2)"), "M_SQRT2")

    expect_equal(rxToSE("M_SQRT_32"), "sqrt(32)")
    expect_equal(rxFromSE("sqrt(32)"), "M_SQRT_32")

    expect_equal(rxToSE("M_SQRT_PI"), "sqrt(pi)")
    expect_equal(rxFromSE("sqrt(pi)"), "M_SQRT_PI")

    expect_equal(rxToSE("M_SQRT_2dPI"), "sqrt(2/pi)")
    expect_equal(rxFromSE("sqrt(2/pi)"), "M_SQRT_2dPI")
    expect_equal(rxFromSE("(2/pi)^0.5"), "M_SQRT_2dPI")
    expect_equal(rxFromSE("(2/pi)^(1/2)"), "M_SQRT_2dPI")

    expect_equal(rxToSE("M_PI_2"), "pi/2")
    expect_equal(rxFromSE("pi*0.5"), "M_PI_2")
    expect_equal(rxFromSE("0.5*pi"), "M_PI_2")
    expect_equal(rxFromSE("pi/2"), "M_PI_2")

    expect_equal(rxToSE("M_PI_4"), "pi/4")
    expect_equal(rxFromSE("pi*0.25"), "M_PI_4")
    expect_equal(rxFromSE("0.25*pi"), "M_PI_4")
    expect_equal(rxFromSE("pi/4"), "M_PI_4")

    expect_equal(rxFromSE("1/pi"), "M_1_PI")
    expect_equal(rxToSE("M_1_PI"), "1/pi")

    expect_equal(rxFromSE("2/pi"), "M_2_PI")
    expect_equal(rxToSE("M_2_PI"), "2/pi")

    expect_equal(rxToSE("M_2_SQRTPI"), "2/sqrt(pi)")
    expect_equal(rxFromSE("2/sqrt(pi)"), "M_2_SQRTPI")
    expect_equal(rxFromSE("2/(pi^0.5)"), "M_2_SQRTPI")

    expect_equal(rxToSE(M_1_SQRT_2PI), "1/sqrt(2*pi)")
    expect_equal(rxFromSE("1/sqrt(2*pi)"), "M_1_SQRT_2PI")
    expect_equal(rxFromSE("1/sqrt(pi*2)"), "M_1_SQRT_2PI")
    expect_equal(rxFromSE("1/((pi*2)^(1/2))"), "M_1_SQRT_2PI")
    expect_equal(rxFromSE("1/((pi*2)^0.5)"), "M_1_SQRT_2PI")

    expect_equal(rxToSE(log10(a)), "log(a)/log(10)")
    expect_equal(rxToSE(log2(a)), "log(a)/log(2)")
    ## FIXME log10 log2? fromSE?


    ## expect_equal(rxFromSymPy("3 + 4*3+2+2*matt*pi"), "3 + 4 * 3 + 2 + matt * M_2PI")
    ## expect_equal(rxFromSymPy("3 + 4*3+2+pi*matt*2"), "3 + 4 * 3 + 2 + matt * M_2PI")
  })

  test_that("transit compartment translation.", {

    expect_equal(
      rxToSE(transit(n, mtt, bio)),
      "exp(log((bio)*(podo0()))+log(n + 1)-log(mtt)+(n)*((log(n+1)-log(mtt))+log(t-tlast0()))-((n+1)/(mtt))*(t-tlast0())-lgamma(1+n))")

    expect_equal(
      rxToSE(transit(n, mtt)),
      "exp(log(podo0())+(log(n+1)-log(mtt))+(n)*((log(n+1)-log(mtt))+ log(t-tlast0()))-((n + 1)/(mtt))*(t-tlast0())-lgamma(1+n))")

    tmp <- rxode("d/dt(depot) <- transit(n, mtt, bio)-ka*depot\nd/dt(center)=ka*depot-kel*center")

    tmp2 <- rxS(tmp)
    tmp3 <- tmp2$rx__d_dt_depot__
    expect_equal(rxFromSE(tmp3), "-ka*depot+exp(n*(-log(mtt)+log1p(n)+log(t-tlast0(depot)))-(1+n)*(t-tlast0(depot))/mtt-log(mtt)+log(bio*podo0(depot))+log1p(n)-lgamma1p(n))")

    tmp <- rxode("d/dt(depot) <- transit(n, mtt) - ka*depot\nd/dt(center)=ka*depot-kel*center")
    tmp2 <- rxS(tmp)
    tmp3 <- tmp2$rx__d_dt_depot__
    expect_equal(rxFromSE(tmp3), "-ka*depot+exp(n*(-log(mtt)+log1p(n)+log(t-tlast0(depot)))-(1+n)*(t-tlast0(depot))/mtt-log(mtt)+log1p(n)+log(podo0(depot))-lgamma1p(n))")

  })

  test_that("unknown functions throw errors. rxToSE", {
    expect_error(rxToSE(matt(3)))
  })

  test_that("Theta/eta conversion rxToSE", {
    expect_equal(rxToSE(THETA[1]), "THETA_1_")
    expect_equal(rxToSE(ETA[1]), "ETA_1_")
    expect_equal(rxToSE(df(matt) / dy(THETA[1])), "rx__df_matt_dy_THETA_1___")
    expect_equal(rxToSE(df(matt) / dy(ETA[1])), "rx__df_matt_dy_ETA_1___")
    expect_error(rxToSE(THETA[0]))
    expect_error(rxToSE(ETA[0]))
    expect_error(rxToSE(THETA[0.5]))
    expect_error(rxToSE(ETA[0.5]))
    expect_error(rxToSE(THETA[a]))
    expect_error(rxToSE(ETA[a]))
    expect_error(rxToSE(THETA["b"]))
    expect_error(rxToSE(ETA["b"]))
  })

  test_that("Extra Derivative Table Test", {
    expect_equal(rxFromSE("Derivative(rxTBS(a, b, c, d, f), a)"), "rxTBSd(a,b,c,d,f)")
    expect_equal(rxFromSE("Derivative(rxTBSd(a, b, c, d, f), a)"), "rxTBSd2(a,b,c,d,f)")
    expect_error(rxFromSE("Derivative(rxTBSd2(a, b, c, d, f), a)", unknownDerivatives = "error"))
    expect_error(rxFromSE("Derivative(rxTBS(a, b, c, d, f), d)", unknownDerivatives = "error"))
    expect_error(rxFromSE("Derivative(f(a, b, c), a)", unknownDerivatives = "forward"))
    expect_equal(
      rxFromSE("(2*a + b)*Subs(Derivative(rxTBS(_xi_1, b, c, d, f), _xi_1), (_xi_1), (a*b + a^2))"),
      "(2*a+b)*rxTBSd(a*b+Rx_pow_di(a,2),b,c,d,f)"
    )
    ## A Subs() over a Derivative whose body carries a relational (rxGt/rxEq,
    ## e.g. from abs() or an occasion indicator, as generated for FOCEi IOV
    ## models) must not error: the Derivative conversion turns the relational
    ## into an R `>`/`==`, and the substituted expression is converted a second
    ## time, so those bare operators have to be recognized rather than treated as
    ## unknown user functions (previously "user function '>' requires 0 arguments").
    expect_equal(
      rxFromSE("Subs(Derivative(rxTBS(_xi_1, rxGt(b, 0), c, d, f), _xi_1), (_xi_1), (a))"),
      "rxTBSd(a,(b>0),c,d,f)"
    )
    expect_equal(
      rxFromSE("Subs(Derivative(rxTBS(_xi_1, rxEq(occ, 1), c, d, f), _xi_1), (_xi_1), (a))"),
      "rxTBSd(a,(occ==1),c,d,f)"
    )
  })

  test_that("NN Activation functions derivatives", {

    expect_equal(rxFromSE("Derivative(ReLU(x), x)"), "dReLU(x)")
    expect_equal(rxFromSE("Derivative(dReLU(x), x)"), "0")

    expect_equal(rxFromSE("Derivative(GELU(x), x)"), "dGELU(x)")
    expect_equal(rxFromSE("Derivative(dGELU(x), x)"), "d2GELU(x)")
    expect_equal(rxFromSE("Derivative(d2GELU(x), x)"), "d3GELU(x)")
    expect_equal(rxFromSE("Derivative(d3GELU(x), x)"), "d4GELU(x)")

    expect_equal(rxToSE("d4GELU(x)"), "exp(-(x)^2/2)*(7*(x)^2 - 4 - (x)^4)/sqrt(2*pi)")

    expect_equal(rxFromSE("Derivative(ELU(x, alpha), x)"), "dELU(x, alpha)")
    expect_equal(rxFromSE("Derivative(ELU(x, alpha), alpha)"), "dELUa(x, alpha)")

    expect_equal(rxFromSE("Derivative(dELU(x, alpha), x)"), "d2ELU(x, alpha)")
    expect_equal(rxFromSE("Derivative(dELU(x, alpha), alpha)"), "d2aELU(x, alpha)")

    expect_equal(rxFromSE("Derivative(dELUa(x, alpha), x)"), "d2ELUa(x, alpha)")
    expect_equal(rxFromSE("Derivative(dELUa(x, alpha), alpha)"), "0")

    expect_equal(rxFromSE("Derivative(d2ELUa(x, alpha), x)"), "d2ELUa(x, alpha)")
    expect_equal(rxFromSE("Derivative(d2ELUa(x, alpha), alpha)"), "0")

    expect_equal(rxFromSE("Derivative(d2aELU(x, alpha), x)"), "d2aELU(x, alpha)")
    expect_equal(rxFromSE("Derivative(d2aELU(x, alpha), alpha)"), "0")

    expect_equal(rxFromSE("Derivative(softplus(x), x)"),   "dsoftplus(x)")
    expect_equal(rxFromSE("Derivative(dsoftplus(x), x)"),  "d2softplus(x)")
    expect_equal(rxFromSE("Derivative(d2softplus(x), x)"), "d3softplus(x)")
    expect_equal(rxFromSE("Derivative(d3softplus(x), x)"), "d4softplus(x)")

    expect_equal(rxToSE("d4softplus(x)"),
                 "6.0*exp(-3.0*(x))/(((1.0 + exp(-(x))))^4) - 6.0*exp(-2.0*(x))/(((1.0 + exp(-(x))))^3) + exp(-(x))/(((1.0 + exp(-(x))))^2)")

    expect_equal(rxFromSE("Derivative(SELU(x), x)"), "dSELU(x)")

    expect_equal(rxToSE("dSELU(x)"),
                 "(rxGt(x, 0)*1.0507009873554804934193349852946 + 1.0507009873554804934193349852946*1.6732632423543772848170429916717*exp(x)*rxLeq(x, 0))")

    expect_equal(rxFromSE("Derivative(lReLU(x), x)"), "dlReLU(x)")
    expect_equal(rxFromSE("Derivative(dlReLU(x), x)"), "0")

    expect_equal(rxFromSE("Derivative(PReLU(x, alpha), x)"), "dPReLU(x,alpha)")
    expect_equal(rxFromSE("Derivative(PReLU(x, alpha), alpha)"), "dPReLUa(x,alpha)")

    expect_equal(rxFromSE("Derivative(dPReLU(x, alpha), x)"), "0")
    expect_equal(rxFromSE("Derivative(dPReLU(x, alpha), alpha)"), "dPReLUa1(x,alpha)")

    expect_equal(rxFromSE("Derivative(dPReLUa(x, alpha), x)"), "dPReLUa1(x,alpha)")
    expect_equal(rxFromSE("Derivative(dPReLUa(x, alpha), alpha)"), "0")

    expect_equal(rxFromSE("Derivative(dPReLUa1(x, alpha), x)"), "0")
    expect_equal(rxFromSE("Derivative(dPReLUa1(x, alpha), alpha)"), "0")

    expect_equal(rxFromSE("Derivative(Swish(x), x)"), "dSwish(x)")

    expect_equal(rxToSE("dSwish(x)"),
                 "((x)*exp(-(x))/(1.0 + exp(-(x)))^2 + 1.0/(1.0 + exp(-(x)))")

  })

  test_that("mod", {
    expect_equal(rxFromSE("rxMod(a,b)"), "(a%%b)")
    expect_equal(rxToSE("a%%b"), "rxMod(a,b)")
    expect_equal(rxFromSE("Derivative(rxMod(a,b), a)"), "1")
    expect_equal(rxFromSE("Derivative(rxMod(a,b), b)"), "0")
  })

  test_that("logic tests", {
    expect_equal(rxFromSE("rxEq(a,b)"), "(a==b)")
    expect_equal(rxFromSE("rxNeq(a,b)"), "(a!=b)")
    expect_equal(rxFromSE("rxLt(a,b)"), "(a<b)")
    expect_equal(rxFromSE("rxGt(a,b)"), "(a>b)")
    expect_equal(rxFromSE("rxGeq(a,b)"), "(a>=b)")
    expect_equal(rxFromSE("rxLeq(a,b)"), "(a<=b)")
    expect_equal(rxFromSE("rxAnd(a,b)"), "(a&&b)")
    expect_equal(rxFromSE("rxOr(a,b)"), "(a||b)")
    expect_equal(rxFromSE("rxNot(rxOr(a,b))"), "(!((a||b)))")

    expect_equal(rxToSE("(a==b)"), "(rxEq(a,b))")
    expect_equal(rxToSE("(a!=b)"), "(rxNeq(a,b))")
    expect_equal(rxToSE("(a>=b)"), "(rxGeq(a,b))")
    expect_equal(rxToSE("(a<=b)"), "(rxLeq(a,b))")
    expect_equal(rxToSE("(a<b)"), "(rxLt(a,b))")
    expect_equal(rxToSE("(a>b)"), "(rxGt(a,b))")
    expect_equal(rxToSE("(a&b)"), "(rxAnd(a,b))")
    expect_equal(rxToSE("(a&&b)"), "(rxAnd(a,b))")
    expect_equal(rxToSE("(a|b)"), "(rxOr(a,b))")
    expect_equal(rxToSE("(a||b)"), "(rxOr(a,b))")
    expect_equal(rxToSE("!(a||b)"), "rxNot((rxOr(a,b)))")

    expect_equal(
      rxFromSE("Derivative(rxEq(a,b), a)"),
      "(-20*tanh(10*(a-b))+20*tanh(10*(a-b))^3)"
    )
    expect_equal(
      rxFromSE("Derivative(rxEq(a,b), b)"),
      "(20*tanh(10*(a-b))-20*tanh(10*(a-b))^3)"
    )

    ## Inequality derivatives are the nascent-delta bump +/-(k/2)*sech^2(k*(a-b))
    ## centered on the boundary a==b (no atanh(2*tol-1) shift); see R/d.R.
    expect_equal(
      rxFromSE("Derivative(rxGeq(a,b), a)"),
      "(5-5*tanh(10*(a-b))^2)"
    )

    expect_equal(
      rxFromSE("Derivative(rxGeq(a,b), b)"),
      "(-5+5*tanh(10*(a-b))^2)"
    )

    expect_equal(
      rxFromSE("Derivative(rxLeq(a,b), a)"),
      "(-5+5*tanh(10*(a-b))^2)"
    )
    expect_equal(
      rxFromSE("Derivative(rxLeq(a,b), b)"),
      "(5-5*tanh(10*(a-b))^2)"
    )

    expect_equal(
      rxFromSE("Derivative(rxLt(a,b), a)"),
      "(-5+5*tanh(10*(a-b))^2)"
    )
    expect_equal(
      rxFromSE("Derivative(rxLt(a,b), b)"),
      "(5-5*tanh(10*(a-b))^2)"
    )

    expect_equal(
      rxFromSE("Derivative(rxGt(a,b), a)"),
      "(5-5*tanh(10*(a-b))^2)"
    )
    expect_equal(
      rxFromSE("Derivative(rxGt(a,b), b)"),
      "(-5+5*tanh(10*(a-b))^2)"
    )

    expect_equal(
      rxFromSE("Derivative(rxOr(a,b), a)"),
      "(1-(b))"
    )
    expect_equal(
      rxFromSE("Derivative(rxOr(a,b), b)"),
      "(1-(a))"
    )

    expect_equal(
      rxFromSE("Derivative(rxNot(a), a)"),
      "(-1)"
    )
  })

  test_that("abs/min/max derivatives use the centered relational bump", {
    ## Composite operators expand through rxGt/rxLt, so their derivatives must
    ## carry the centered bump.  (`get` is masked in a symengine env; capture
    ## the Basic from assign()'s return value -- see CLAUDE.md.)
    .s <- rxS("rx_pred_ = max(a, b)")
    .d <- eval(parse(text = "assign(\".d\", with(.s, D(rx_pred_, a)), envir = .s)"))
    expect_equal(rxFromSE(.d), "(a-b)*(5-5*tanh(10*(a-b))^2)+(a>b)")

    .s <- rxS("rx_pred_ = min(a, b)")
    .d <- eval(parse(text = "assign(\".d\", with(.s, D(rx_pred_, a)), envir = .s)"))
    expect_equal(rxFromSE(.d), "(a-b)*(-5+5*tanh(10*(a-b))^2)+(a<b)")

    .s <- rxS("rx_pred_ = abs(a)")
    .d <- eval(parse(text = "assign(\".d\", with(.s, D(rx_pred_, a)), envir = .s)"))
    expect_equal(rxFromSE(.d), "-1+2*a*(5-5*tanh(10*(a-0))^2)+2*(a>0)")
  })

  # Test factor expansion by `rxSplitPlusQ'
  test_that("rxSplitPlusQ", {
    expect_equal(rxSplitPlusQ(quote(a * exp(b + c) + d * log(e - f) - g * f)), c("a * exp(b + c)", "d * log(e - f)", "- g * f"))
    expect_equal(rxSplitPlusQ(quote(-a * exp(b + c) + d * log(e - f) - g * f)), c("-a * exp(b + c)", "d * log(e - f)", "- g * f"))
    expect_equal(rxSplitPlusQ(quote(+a * exp(b + c) + d * log(e - f) - g * f)), c("+a * exp(b + c)", "d * log(e - f)", "- g * f"))
    expect_equal(rxSplitPlusQ(quote(+a * exp(b + c))), "+a * exp(b + c)")
    expect_equal(rxSplitPlusQ(quote(center)), "center")
    expect_equal(rxSplitPlusQ(quote(0)), "0")
  })

  # Test Error DSLs

  pk <- function() {
    KA <- exp(THETA[1])
    CL <- exp(THETA[2] + ETA[1])
    V <- exp(THETA[3] + ETA[2])
    return(1)
  }

  pred <- function() {
    if (cmt == 1) {
      return(centr)
    } else if (cmt == 2) {
      return(depot)
    } else {
      return(perip)
    }
  }


  pred.for <- function() {
    for (i in 1:10) {
      pred <- 1
    }
  }

  err <- function() {
    add(0.2) + prop(0.3)
  }

  test_that("PK/Pred/Error function parsing.", {
    expect_error(rxParsePk(pk))
    expect_error(rxParsePred(pred), NA)
    expect_error(rxParsePred(pred.for))
    expect_error(rxParsePk(err))
    expect_error(rxParsePred(err))
  })


  pk2 <- function() {
    KA <- exp(THETA[1])
    CL <- exp(THETA[2] + ETA[1])
    V <- exp(THETA[3] + ETA[2])
  }

  test_that("sum/prod testing", {
    expect_equal(rxToSE("sum(a,b,c)"), "((a)+(b)+(c))")
    expect_equal(rxToSE("prod(a,b,c)"), "((a)*(b)*(c))")
  })

  test_that("tlast", {
    expect_error(rxToSE("tlast(matt+3)"))
    expect_error(rxToSE("tlast(matt, 3)"))
    expect_equal(rxToSE("tlast(matt)"), "tlast(matt)")
    expect_equal(rxToSE("tlast()"), "tlast()")
    expect_equal(rxFromSE("Derivative(tlast(matt), matt)"), "0")
    expect_equal(rxFromSE("tlast(matt)"), "tlast(matt)")

    expect_error(rxFromSE("tlast(matt,2)"))
    expect_equal(rxFromSE("tlast(matt)"), "tlast(matt)")
    expect_equal(rxFromSE("tlast()"), "tlast()")
  })

  test_that("tfirst", {
    expect_error(rxToSE("tfirst(matt+3)"))
    expect_error(rxToSE("tfirst(matt, 3)"))
    expect_equal(rxToSE("tfirst(matt)"), "tfirst(matt)")
    expect_equal(rxToSE("tfirst()"), "tfirst()")
    expect_equal(rxFromSE("Derivative(tfirst(matt), matt)"), "0")
    expect_equal(rxFromSE("tfirst(matt)"), "tfirst(matt)")

    expect_error(rxFromSE("tfirst(matt,2)"))
    expect_equal(rxFromSE("tfirst(matt)"), "tfirst(matt)")
    expect_equal(rxFromSE("tfirst()"), "tfirst()")
  })

  test_that("dosenum", {
    expect_error(rxToSE("dosenum(a)"))
    expect_equal(rxToSE("dosenum()"), "dosenum()")
    expect_equal(rxFromSE("dosenum()"), "dosenum()")
  })

  test_that("tad()", {
    expect_equal(rxToSE("tad()"), "(t-tlast())")
    expect_equal(rxToSE("tad(matt)"), "(t-tlast(matt))")
    expect_error(rxToSE("tad(matt,f)"))
    expect_error(rxToSE("tad(matt+f)"))
  })

  test_that("tad0()", {
    expect_equal(rxToSE("tad0()"), "(t-tlast0())")
    expect_equal(rxToSE("tad0(matt)"), "(t-tlast0(matt))")
    expect_error(rxToSE("tad0(matt,f)"))
    expect_error(rxToSE("tad0(matt+f)"))
  })

  test_that("tafd()", {
    expect_equal(rxToSE("tafd()"), "(t-tfirst())")
    expect_equal(rxToSE("tafd(matt)"), "(t-tfirst(matt))")
    expect_error(rxToSE("tafd(matt,f)"))
    expect_error(rxToSE("tafd(matt+f)"))
  })

  test_that("lag()", {
    expect_error(rxToSE("lag()"))
    expect_equal(rxToSE("lag(b)"), "lag(b)")
    expect_error(rxToSE("lag(b+3)"))
    expect_equal(rxToSE("lag(b, 3)"), "lag(b, 3)")
    expect_error(rxToSE("lag(b, c)"))
    expect_error(rxToSE("lag(b, 3+4)"))
    expect_equal(rxFromSE("Derivative(lag(a,b), a)"), "0")
    expect_equal(rxFromSE("Derivative(lag(a,b), b)"), "0")
    expect_equal(rxFromSE("Derivative(lag(a), a)"), "0")
    expect_equal(rxFromSE("lag(a,b)"), "lag(a,b)")
    expect_equal(rxFromSE("lag(a)"), "lag(a)")
  })

  test_that("lead()", {
    expect_error(rxToSE("lead()"))
    expect_equal(rxToSE("lead(b)"), "lead(b)")
    expect_error(rxToSE("lead(b+3)"))
    expect_equal(rxToSE("lead(b, 3)"), "lead(b, 3)")
    expect_error(rxToSE("lead(b, c)"))
    expect_error(rxToSE("lead(b, 3+4)"))
    expect_equal(rxFromSE("Derivative(lead(a,b), a)"), "0")
    expect_equal(rxFromSE("Derivative(lead(a,b), b)"), "0")
    expect_equal(rxFromSE("Derivative(lead(a), a)"), "0")
    expect_equal(rxFromSE("lead(a,b)"), "lead(a,b)")
    expect_equal(rxFromSE("lead(a)"), "lead(a)")
  })

  test_that("first", {
    expect_error(rxToSE("first()"))
    expect_equal(rxToSE("first(v)"), "first(v)")
    expect_equal(rxFromSE("Derivative(first(a),a)"), "0")
    expect_equal(rxFromSE("first(v)"), "first(v)")
  })

  test_that("last", {
    expect_error(rxToSE("last()"))
    expect_equal(rxToSE("last(v)"), "last(v)")
    expect_equal(rxFromSE("Derivative(last(a),a)"), "0")
    expect_equal(rxFromSE("last(v)"), "last(v)")
  })

  test_that("diff", {
    expect_error(rxToSE("diff()"))
    expect_equal(rxToSE("diff(v)"), "diff(v)")
    expect_equal(rxFromSE("Derivative(diff(a),a)"), "0")
    expect_equal(rxFromSE("diff(v)"), "diff(v)")
  })

  test_that("is.nan", {
    expect_error(rxToSE("is.nan()"))
    expect_equal(rxToSE("is.nan(v)"), "is.nan(v)")
    expect_equal(rxFromSE("Derivative(is.nan(a),a)"), "0")
    expect_equal(rxFromSE("is.nan(v)"), "is.nan(v)")
  })

  test_that("is.na", {
    expect_error(rxToSE("is.na()"))
    expect_equal(rxToSE("is.na(v)"), "is.na(v)")
    expect_equal(rxFromSE("Derivative(is.na(a),a)"), "0")
    expect_equal(rxFromSE("is.na(v)"), "is.na(v)")
  })

  test_that("is.finite", {
    expect_error(rxToSE("is.finite()"))
    expect_equal(rxToSE("is.finite(v)"), "is.finite(v)")
    expect_equal(rxFromSE("Derivative(is.finite(a),a)"), "0")
    expect_equal(rxFromSE("is.finite(v)"), "is.finite(v)")
  })

  test_that("is.infinite", {
    expect_error(rxToSE("is.infinite()"))
    expect_equal(rxToSE("is.infinite(v)"), "is.infinite(v)")
    expect_equal(rxFromSE("Derivative(is.infinite(a),a)"), "0")
    expect_equal(rxFromSE("is.infinite(v)"), "is.infinite(v)")
  })

  test_that("dual function/variable promotion",{
    expect_equal(rxToSE("tlast"), "tlast()")
    expect_equal(rxFromSE("tlast()"), "tlast()")
    expect_equal(rxToSE("tad"), "(t-tlast())")
    expect_equal(rxFromSE("tlast(NaN)"), "tlast()")
  })

  test_that("max/min tests", {
    expect_equal(rxToSE("max(a)"), "(a)")
    expect_equal(rxToSE("max(a,0)"), "((a)*rxGt(a,0))")
    expect_equal(rxToSE("max(0,a)"), "((a)*rxGt(a,0))")
    expect_equal(rxToSE("max(a,b)"), "(((a)-(b))*rxGt(a,b)+(b))")
    expect_equal(rxToSE("max(a,b,c)"),
                 "((((((a)-(b))*rxGt(a,b)+(b)))-(c))*rxGt((((a)-(b))*rxGt(a,b)+(b)),c)+(c))")
    expect_equal(rxToSE("max()"), "")
    expect_equal(rxToSE("min(a)"), "(a)")
    expect_equal(rxToSE("min(a,0)"), "((a)*rxLt(a,0))")
    expect_equal(rxToSE("min(0,a)"), "((a)*rxLt(a,0))")
    expect_equal(rxToSE("min(a,b)"), "(((a)-(b))*rxLt(a,b)+(b))")
    expect_equal(rxToSE("min(a,b,c)"),
                 "((((((a)-(b))*rxLt(a,b)+(b)))-(c))*rxLt((((a)-(b))*rxLt(a,b)+(b)),c)+(c))")
    expect_equal(rxToSE("min()"), "")
  })

  test_that("abs tests", {
    expect_equal(rxToSE("abs(a)"), "(2.0*(a)*rxGt(a,0.0)-(a))")
    expect_equal(rxToSE("fabs(a)"), "(2.0*(a)*rxGt(a,0.0)-(a))")
    expect_equal(rxToSE("abs0(a)"), "(2.0*(a)*rxGt(a,0.0)-(a))")
  })

  test_that("parsing errors", {

    test <- "E0=THETA[1];\nEm=0.5;\nE50=THETA[2];\ng=2;\nv=E0+Em*t^g/(E50^g+t^g);\nrx_yj_~152;\nrx_lambda_~1;\nrx_low_~0;\nrx_hi_~1;\nrx_r_~0;\nrx_pred_~DV*v-log(1+exp(v));\n"

    expect_error(rxS(test), NA)

  })

  test_that("Promote linear solution linCmtA to linear solution linCmtB -- linCmt() sol", {
    expect_equal(str2lang(rxToSE("linCmtA(rx__PTR__,t,2,1,1,-1,1,cl,v,q,v2,q2,v3,ka)",
                    promoteLinSens = TRUE)),
                 str2lang("linCmtB(rx__PTR__, t, 2, 1, 1, -1, -1, 1, cl, v, q, v2, q2, v3, ka)"))
  })

  test_that("Promote linear solution linCmtA to linear solution linCmtB -- amts", {
    for (i in 0:7) {
      str <- paste0("linCmtA(rx__PTR__,t,2,1,1,",i,",1,cl,v,q,v2,q2,v3,ka)")
      expect_equal(str2lang(rxToSE(str, promoteLinSens = TRUE)),
                   str2lang(paste0("linCmtB(rx__PTR__, t, 2, 1, 1, ", i, ", -2, 1, cl, v, q, v2, q2, v3, ka)")))
    }
  })

  testVars <- c("b=depot",
                "c=central",
                "d=peripheral1",
                "ee=peripheral2",
                "f=rx__sens_central_BY_p1",
                "g=rx__sens_central_BY_v1",
                "h=rx__sens_central_BY_p2",
                "i=rx__sens_central_BY_p3",
                "j=rx__sens_central_BY_p4",
                "k=rx__sens_central_BY_p5",
                "l=rx__sens_central_BY_ka",
                "m=rx__sens_peripheral1_BY_p1",
                "n=rx__sens_peripheral1_BY_v1",
                "o=rx__sens_peripheral1_BY_p2",
                "p=rx__sens_peripheral1_BY_p3",
                "q=rx__sens_peripheral1_BY_p4",
                "r=rx__sens_peripheral1_BY_p5",
                "s=rx__sens_peripheral1_BY_ka",
                "tt=rx__sens_peripheral2_BY_p1",
                "u=rx__sens_peripheral2_BY_v1",
                "v=rx__sens_peripheral2_BY_p2",
                "w=rx__sens_peripheral2_BY_p3",
                "x=rx__sens_peripheral2_BY_p4",
                "y=rx__sens_peripheral2_BY_p5",
                "z=rx__sens_peripheral2_BY_ka",
                "aa=rx__sens_depot_BY_ka")

  # Testing function for translation
  f <- function(w, num=NULL, s=s) {
    w <- get(w, envir=s)
    v <- rxFromSE(w)
    l <- str2lang(v)
    if (is.symbol(l)) {
      if (is.null(num)) {
        return(TRUE)
      } else {
        return(FALSE)
      }
    }
    if (is.call(l) && (identical(l[[1]], quote(`linCmtA`)) ||
                         identical(l[[1]], quote(`linCmtB`)))) {
      return(l[[7]] == num)
    }
    FALSE
  }

  test_that("3 compartment oral states translated in symengine", {

    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,3,1,-1,1,cl,v,q,v2,q2,v3,ka)",
                 testVars), collapse="\n")
    s <- rxS(s)

    expect_true(f("b", 0, s=s)) # depot
    expect_true(f("c", 1, s=s)) # central
    expect_true(f("d", 2, s=s)) # peripheral1
    expect_true(f("ee", 3, s=s)) # peripheral2
    expect_true(f("f", s=s))
    expect_true(f("g", s=s))
    expect_true(f("h", s=s))
    expect_true(f("i", s=s))
    expect_true(f("j", s=s))
    expect_true(f("k", s=s))
    expect_true(f("l", s=s))
    expect_true(f("m", s=s))
    expect_true(f("n", s=s))
    expect_true(f("o", s=s))
    expect_true(f("p", s=s))
    expect_true(f("q", s=s))
    expect_true(f("r", s=s))
    expect_true(f("s", s=s))
    expect_true(f("tt", s=s))
    expect_true(f("u", s=s))
    expect_true(f("v", s=s))
    expect_true(f("w", s=s))
    expect_true(f("x", s=s))
    expect_true(f("y", s=s))
    expect_true(f("z", s=s))
    expect_true(f("aa", s=s))

    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,3,1,-1,1,cl,v,q,v2,q2,v3,ka)",
                 testVars), collapse="\n")
    s <- rxS(s, promoteLinSens = TRUE)

    expect_true(f("b", 0, s=s)) # depot
    expect_true(f("c", 1, s=s)) # central
    expect_true(f("d", 2, s=s)) # peripheral1
    expect_true(f("ee", 3, s=s)) # peripheral2
    expect_true(f("f", 4, s=s))
    expect_true(f("g", 5, s=s))
    expect_true(f("h", 6, s=s))
    expect_true(f("i", 7, s=s))
    expect_true(f("j", 8, s=s))
    expect_true(f("k", 9, s=s))
    expect_true(f("l", 10, s=s))
    expect_true(f("m", 11, s=s))
    expect_true(f("n", 12, s=s))
    expect_true(f("o", 13, s=s))
    expect_true(f("p", 14, s=s))
    expect_true(f("q", 15, s=s))
    expect_true(f("r", 16, s=s))
    expect_true(f("s", 17, s=s))
    expect_true(f("tt", 18, s=s))
    expect_true(f("u", 19, s=s))
    expect_true(f("v", 20, s=s))
    expect_true(f("w", 21, s=s))
    expect_true(f("x", 22, s=s))
    expect_true(f("y", 23, s=s))
    expect_true(f("z", 24, s=s))
    expect_true(f("aa", 25, s=s))
  })

  test_that("3 compartment states translated in symengine", {

    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,3,0,-1,1,cl,v,q,v2,q2,v3,0)",
                 testVars), collapse="\n")
    s <- rxS(s)

    expect_true(f("b", s=s)) # depot
    expect_true(f("c", 0, s=s)) # central
    expect_true(f("d", 1, s=s)) # peripheral1
    expect_true(f("ee", 2, s=s)) # peripheral2
    expect_true(f("f", s=s))
    expect_true(f("g", s=s))
    expect_true(f("h", s=s))
    expect_true(f("i", s=s))
    expect_true(f("j", s=s))
    expect_true(f("k", s=s))
    expect_true(f("l", s=s))
    expect_true(f("m", s=s))
    expect_true(f("n", s=s))
    expect_true(f("o", s=s))
    expect_true(f("p", s=s))
    expect_true(f("q", s=s))
    expect_true(f("r", s=s))
    expect_true(f("s", s=s))
    expect_true(f("tt", s=s))
    expect_true(f("u", s=s))
    expect_true(f("v", s=s))
    expect_true(f("w", s=s))
    expect_true(f("x", s=s))
    expect_true(f("y", s=s))
    expect_true(f("z", s=s))
    expect_true(f("aa", s=s))

    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,3,0,-1,1,cl,v,q,v2,q2,v3,0)",
                 testVars), collapse="\n")
    s <- rxS(s, promoteLinSens = TRUE)

    expect_true(f("b", s=s)) # depot
    expect_true(f("c", 0, s=s)) # central
    expect_true(f("d", 1, s=s)) # peripheral1
    expect_true(f("ee", 2, s=s)) # peripheral2
    expect_true(f("f", 3, s=s)) # rx__sens_central_BY_p1
    expect_true(f("g", 4, s=s)) # rx__sens_central_BY_v1
    expect_true(f("h", 5, s=s)) # rx__sens_central_BY_p2
    expect_true(f("i", 6, s=s)) # rx__sens_central_BY_p3
    expect_true(f("j", 7, s=s)) # rx__sens_central_BY_p4
    expect_true(f("k", 8, s=s)) # rx__sens_central_BY_p5
    expect_true(f("l", s=s))    # rx__sens_central_BY_ka
    expect_true(f("m", 9, s=s)) # rx__sens_peripheral1_BY_p1
    expect_true(f("n", 10, s=s))# rx__sens_peripheral1_BY_v1
    expect_true(f("o", 11, s=s))# rx__sens_peripheral1_BY_p2
    expect_true(f("p", 12, s=s))# rx__sens_peripheral1_BY_p3
    expect_true(f("q", 13, s=s))# rx__sens_peripheral1_BY_p4
    expect_true(f("r", 14, s=s))# rx__sens_peripheral1_BY_p5
    expect_true(f("s", s=s))    # rx__sens_peripheral1_BY_ka
    expect_true(f("tt", 15, s=s))#rx__sens_peripheral2_BY_p1
    expect_true(f("u", 16, s=s))# rx__sens_peripheral2_BY_v1
    expect_true(f("v", 17, s=s))# rx__sens_peripheral2_BY_p2
    expect_true(f("w", 18, s=s))# rx__sens_peripheral2_BY_p3
    expect_true(f("x", 19, s=s))# rx__sens_peripheral2_BY_p4
    expect_true(f("y", 20, s=s))# rx__sens_peripheral2_BY_p5
    expect_true(f("z", s=s))# rx__sens_peripheral2_BY_ka
    expect_true(f("aa", s=s))#rx__sens_depot_BY_ka

  })

  test_that("2 compartment oral states translated in symengine", {

    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,2,1,-1,1,cl,v,q,v2,0,0,ka)",
                 testVars), collapse="\n")
    s <- rxS(s)

    expect_true(f("b", 0, s=s)) # depot
    expect_true(f("c", 1, s=s)) # central
    expect_true(f("d", 2, s=s)) # peripheral1
    expect_true(f("ee", s=s)) # peripheral2
    expect_true(f("f", s=s)) # rx__sens_central_BY_p1
    expect_true(f("g", s=s)) # rx__sens_central_BY_v1
    expect_true(f("h", s=s)) # rx__sens_central_BY_p2
    expect_true(f("i", s=s)) # rx__sens_central_BY_p3
    expect_true(f("j", s=s)) # rx__sens_central_BY_p4
    expect_true(f("k", s=s)) # rx__sens_central_BY_p5
    expect_true(f("l", s=s)) # rx__sens_central_BY_ka
    expect_true(f("m", s=s))# rx__sens_peripheral1_BY_p1
    expect_true(f("n", s=s))# rx__sens_peripheral1_BY_v1
    expect_true(f("o", s=s))# rx__sens_peripheral1_BY_p2
    expect_true(f("p", s=s))# rx__sens_peripheral1_BY_p3
    expect_true(f("q", s=s))# rx__sens_peripheral1_BY_p4
    expect_true(f("r", s=s))# rx__sens_peripheral1_BY_p5
    expect_true(f("s", s=s))# rx__sens_peripheral1_BY_ka
    expect_true(f("tt", s=s))#rx__sens_peripheral2_BY_p1
    expect_true(f("u", s=s))# rx__sens_peripheral2_BY_v1
    expect_true(f("v", s=s))# rx__sens_peripheral2_BY_p2
    expect_true(f("w", s=s))# rx__sens_peripheral2_BY_p3
    expect_true(f("x", s=s))# rx__sens_peripheral2_BY_p4
    expect_true(f("y", s=s))# rx__sens_peripheral2_BY_p5
    expect_true(f("z", s=s))# rx__sens_peripheral2_BY_ka
    expect_true(f("aa", s=s))#rx__sens_depot_BY_ka

    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,2,1,-1,1,cl,v,q,v2,0,0,ka)",
                 testVars), collapse="\n")
    s <- rxS(s, promoteLinSens = TRUE)
    expect_true(f("b", 0, s=s)) # depot
    expect_true(f("c", 1, s=s)) # central
    expect_true(f("d", 2, s=s)) # peripheral1
    expect_true(f("ee", s=s)) # peripheral2
    expect_true(f("f", 3, s=s)) # rx__sens_central_BY_p1
    expect_true(f("g", 4, s=s)) # rx__sens_central_BY_v1
    expect_true(f("h", 5, s=s)) # rx__sens_central_BY_p2
    expect_true(f("i", 6, s=s)) # rx__sens_central_BY_p3
    expect_true(f("j", s=s))    # rx__sens_central_BY_p4
    expect_true(f("k", s=s))    # rx__sens_central_BY_p5
    expect_true(f("l", 7, s=s)) # rx__sens_central_BY_ka
    expect_true(f("m", 8, s=s)) # rx__sens_peripheral1_BY_p1
    expect_true(f("n", 9, s=s)) # rx__sens_peripheral1_BY_v1
    expect_true(f("o", 10, s=s))# rx__sens_peripheral1_BY_p2
    expect_true(f("p", 11, s=s))# rx__sens_peripheral1_BY_p3
    expect_true(f("q", s=s))    # rx__sens_peripheral1_BY_p4
    expect_true(f("r", s=s))    # rx__sens_peripheral1_BY_p5
    expect_true(f("s", 12, s=s))# rx__sens_peripheral1_BY_ka
    expect_true(f("tt", s=s))   # rx__sens_peripheral2_BY_p1
    expect_true(f("u", s=s))    # rx__sens_peripheral2_BY_v1
    expect_true(f("v", s=s))    # rx__sens_peripheral2_BY_p2
    expect_true(f("w", s=s))    # rx__sens_peripheral2_BY_p3
    expect_true(f("x", s=s))    # rx__sens_peripheral2_BY_p4
    expect_true(f("y", s=s))    # rx__sens_peripheral2_BY_p5
    expect_true(f("z", s=s))    # rx__sens_peripheral2_BY_ka
    expect_true(f("aa", 13, s=s))#rx__sens_depot_BY_ka
  })

  test_that("2 compartment states translated in symengine", {

    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,2,0,-1,1,cl,v,q,v2,0,0,0)",
                 testVars), collapse="\n")
    s <- rxS(s)

    expect_true(f("b", s=s)) # depot
    expect_true(f("c", 0, s=s)) # central
    expect_true(f("d", 1, s=s)) # peripheral1
    expect_true(f("ee", s=s)) # peripheral2
    expect_true(f("f", s=s)) # rx__sens_central_BY_p1
    expect_true(f("g", s=s)) # rx__sens_central_BY_v1
    expect_true(f("h", s=s)) # rx__sens_central_BY_p2
    expect_true(f("i", s=s)) # rx__sens_central_BY_p3
    expect_true(f("j", s=s)) # rx__sens_central_BY_p4
    expect_true(f("k", s=s)) # rx__sens_central_BY_p5
    expect_true(f("l", s=s)) # rx__sens_central_BY_ka
    expect_true(f("m", s=s))# rx__sens_peripheral1_BY_p1
    expect_true(f("n", s=s))# rx__sens_peripheral1_BY_v1
    expect_true(f("o", s=s))# rx__sens_peripheral1_BY_p2
    expect_true(f("p", s=s))# rx__sens_peripheral1_BY_p3
    expect_true(f("q", s=s))# rx__sens_peripheral1_BY_p4
    expect_true(f("r", s=s))# rx__sens_peripheral1_BY_p5
    expect_true(f("s", s=s))# rx__sens_peripheral1_BY_ka
    expect_true(f("tt", s=s))#rx__sens_peripheral2_BY_p1
    expect_true(f("u", s=s))# rx__sens_peripheral2_BY_v1
    expect_true(f("v", s=s))# rx__sens_peripheral2_BY_p2
    expect_true(f("w", s=s))# rx__sens_peripheral2_BY_p3
    expect_true(f("x", s=s))# rx__sens_peripheral2_BY_p4
    expect_true(f("y", s=s))# rx__sens_peripheral2_BY_p5
    expect_true(f("z", s=s))# rx__sens_peripheral2_BY_ka
    expect_true(f("aa", s=s))#rx__sens_depot_BY_ka

    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,2,0,-1,1,cl,v,q,v2,0,0,ka)",
                 testVars), collapse="\n")
    s <- rxS(s, promoteLinSens = TRUE)

    expect_true(f("b", s=s)) # depot
    expect_true(f("c", 0, s=s)) # central
    expect_true(f("d", 1, s=s)) # peripheral1
    expect_true(f("ee", s=s)) # peripheral2
    expect_true(f("f", 2, s=s)) # rx__sens_central_BY_p1
    expect_true(f("g", 3, s=s)) # rx__sens_central_BY_v1
    expect_true(f("h", 4, s=s)) # rx__sens_central_BY_p2
    expect_true(f("i", 5, s=s)) # rx__sens_central_BY_p3
    expect_true(f("j", s=s))    # rx__sens_central_BY_p4
    expect_true(f("k", s=s))    # rx__sens_central_BY_p5
    expect_true(f("l", s=s))    # rx__sens_central_BY_ka
    expect_true(f("m", 6, s=s)) # rx__sens_peripheral1_BY_p1
    expect_true(f("n", 7, s=s)) # rx__sens_peripheral1_BY_v1
    expect_true(f("o", 8, s=s)) # rx__sens_peripheral1_BY_p2
    expect_true(f("p", 9, s=s))# rx__sens_peripheral1_BY_p3
    expect_true(f("q", s=s))    # rx__sens_peripheral1_BY_p4
    expect_true(f("r", s=s))    # rx__sens_peripheral1_BY_p5
    expect_true(f("s", s=s))    # rx__sens_peripheral1_BY_ka
    expect_true(f("tt", s=s))   # rx__sens_peripheral2_BY_p1
    expect_true(f("u", s=s))    # rx__sens_peripheral2_BY_v1
    expect_true(f("v", s=s))    # rx__sens_peripheral2_BY_p2
    expect_true(f("w", s=s))    # rx__sens_peripheral2_BY_p3
    expect_true(f("x", s=s))    # rx__sens_peripheral2_BY_p4
    expect_true(f("y", s=s))    # rx__sens_peripheral2_BY_p5
    expect_true(f("z", s=s))    # rx__sens_peripheral2_BY_ka
  })

  test_that("1 compartment oral states translated in symengine", {
    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,1,1,-1,1,cl,v,0,0,0,0,ka)",
                 testVars), collapse="\n")
    s <- rxS(s)

    expect_true(f("b", 0, s=s)) # depot
    expect_true(f("c", 1, s=s)) # central
    expect_true(f("d", s=s))  # peripheral1
    expect_true(f("ee", s=s)) # peripheral2
    expect_true(f("f", s=s)) # rx__sens_central_BY_p1
    expect_true(f("g", s=s)) # rx__sens_central_BY_v1
    expect_true(f("h", s=s)) # rx__sens_central_BY_p2
    expect_true(f("i", s=s)) # rx__sens_central_BY_p3
    expect_true(f("j", s=s)) # rx__sens_central_BY_p4
    expect_true(f("k", s=s)) # rx__sens_central_BY_p5
    expect_true(f("l", s=s)) # rx__sens_central_BY_ka
    expect_true(f("m", s=s))# rx__sens_peripheral1_BY_p1
    expect_true(f("n", s=s))# rx__sens_peripheral1_BY_v1
    expect_true(f("o", s=s))# rx__sens_peripheral1_BY_p2
    expect_true(f("p", s=s))# rx__sens_peripheral1_BY_p3
    expect_true(f("q", s=s))# rx__sens_peripheral1_BY_p4
    expect_true(f("r", s=s))# rx__sens_peripheral1_BY_p5
    expect_true(f("s", s=s))# rx__sens_peripheral1_BY_ka
    expect_true(f("tt", s=s))#rx__sens_peripheral2_BY_p1
    expect_true(f("u", s=s))# rx__sens_peripheral2_BY_v1
    expect_true(f("v", s=s))# rx__sens_peripheral2_BY_p2
    expect_true(f("w", s=s))# rx__sens_peripheral2_BY_p3
    expect_true(f("x", s=s))# rx__sens_peripheral2_BY_p4
    expect_true(f("y", s=s))# rx__sens_peripheral2_BY_p5
    expect_true(f("z", s=s))# rx__sens_peripheral2_BY_ka
    expect_true(f("aa", s=s))#rx__sens_depot_BY_ka

    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,1,1,-1,1,cl,v,0,0,0,0,ka)",
                 testVars), collapse="\n")
    s <- rxS(s, promoteLinSens = TRUE)

    expect_true(f("b", 0, s=s)) # depot
    expect_true(f("c", 1, s=s)) # central
    expect_true(f("d", s=s)) # peripheral1
    expect_true(f("ee", s=s)) # peripheral2
    expect_true(f("f", 2, s=s)) # rx__sens_central_BY_p1
    expect_true(f("g", 3, s=s)) # rx__sens_central_BY_v1
    expect_true(f("h", s=s))    # rx__sens_central_BY_p2
    expect_true(f("i", s=s))    # rx__sens_central_BY_p3
    expect_true(f("j", s=s))    # rx__sens_central_BY_p4
    expect_true(f("k", s=s))    # rx__sens_central_BY_p5
    expect_true(f("l", 4, s=s)) # rx__sens_central_BY_ka
    expect_true(f("m", s=s))    # rx__sens_peripheral1_BY_p1
    expect_true(f("n", s=s))    # rx__sens_peripheral1_BY_v1
    expect_true(f("o", s=s))    # rx__sens_peripheral1_BY_p2
    expect_true(f("p", s=s))    # rx__sens_peripheral1_BY_p3
    expect_true(f("q", s=s))    # rx__sens_peripheral1_BY_p4
    expect_true(f("r", s=s))    # rx__sens_peripheral1_BY_p5
    expect_true(f("s", s=s))    # rx__sens_peripheral1_BY_ka
    expect_true(f("tt", s=s))   # rx__sens_peripheral2_BY_p1
    expect_true(f("u", s=s))    # rx__sens_peripheral2_BY_v1
    expect_true(f("v", s=s))    # rx__sens_peripheral2_BY_p2
    expect_true(f("w", s=s))    # rx__sens_peripheral2_BY_p3
    expect_true(f("x", s=s))    # rx__sens_peripheral2_BY_p4
    expect_true(f("y", s=s))    # rx__sens_peripheral2_BY_p5
    expect_true(f("z", s=s))    # rx__sens_peripheral2_BY_ka
    expect_true(f("aa", 5, s=s))#rx__sens_depot_BY_ka
  })

  test_that("1 compartment states translated in symengine", {

    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,1,0,-1,1,cl,v,0,0,0,0,0)",
                 testVars), collapse="\n")
    s <- rxS(s)

    expect_true(f("b", s=s)) # depot
    expect_true(f("c", 0, s=s)) # central
    expect_true(f("d", s=s))  # peripheral1
    expect_true(f("ee", s=s)) # peripheral2
    expect_true(f("f", s=s)) # rx__sens_central_BY_p1
    expect_true(f("g", s=s)) # rx__sens_central_BY_v1
    expect_true(f("h", s=s)) # rx__sens_central_BY_p2
    expect_true(f("i", s=s)) # rx__sens_central_BY_p3
    expect_true(f("j", s=s)) # rx__sens_central_BY_p4
    expect_true(f("k", s=s)) # rx__sens_central_BY_p5
    expect_true(f("l", s=s)) # rx__sens_central_BY_ka
    expect_true(f("m", s=s))# rx__sens_peripheral1_BY_p1
    expect_true(f("n", s=s))# rx__sens_peripheral1_BY_v1
    expect_true(f("o", s=s))# rx__sens_peripheral1_BY_p2
    expect_true(f("p", s=s))# rx__sens_peripheral1_BY_p3
    expect_true(f("q", s=s))# rx__sens_peripheral1_BY_p4
    expect_true(f("r", s=s))# rx__sens_peripheral1_BY_p5
    expect_true(f("s", s=s))# rx__sens_peripheral1_BY_ka
    expect_true(f("tt", s=s))#rx__sens_peripheral2_BY_p1
    expect_true(f("u", s=s))# rx__sens_peripheral2_BY_v1
    expect_true(f("v", s=s))# rx__sens_peripheral2_BY_p2
    expect_true(f("w", s=s))# rx__sens_peripheral2_BY_p3
    expect_true(f("x", s=s))# rx__sens_peripheral2_BY_p4
    expect_true(f("y", s=s))# rx__sens_peripheral2_BY_p5
    expect_true(f("z", s=s))# rx__sens_peripheral2_BY_ka
    expect_true(f("aa", s=s))#rx__sens_depot_BY_ka

    s <- paste(c("linc=linCmtA(rx__PTR__,t,2,1,0,-1,1,cl,v,0,0,0,0,0)",
                 testVars), collapse="\n")
    s <- rxS(s, promoteLinSens = TRUE)

    expect_true(f("b", s=s)) # depot
    expect_true(f("c", 0, s=s)) # central
    expect_true(f("d", s=s)) # peripheral1
    expect_true(f("ee", s=s)) # peripheral2
    expect_true(f("f", 1, s=s)) # rx__sens_central_BY_p1
    expect_true(f("g", 2, s=s)) # rx__sens_central_BY_v1
    expect_true(f("h", s=s))    # rx__sens_central_BY_p2
    expect_true(f("i", s=s))    # rx__sens_central_BY_p3
    expect_true(f("j", s=s))    # rx__sens_central_BY_p4
    expect_true(f("k", s=s))    # rx__sens_central_BY_p5
    expect_true(f("l", s=s))    # rx__sens_central_BY_ka
    expect_true(f("m", s=s))    # rx__sens_peripheral1_BY_p1
    expect_true(f("n", s=s))    # rx__sens_peripheral1_BY_v1
    expect_true(f("o", s=s))    # rx__sens_peripheral1_BY_p2
    expect_true(f("p", s=s))    # rx__sens_peripheral1_BY_p3
    expect_true(f("q", s=s))    # rx__sens_peripheral1_BY_p4
    expect_true(f("r", s=s))    # rx__sens_peripheral1_BY_p5
    expect_true(f("s", s=s))    # rx__sens_peripheral1_BY_ka
    expect_true(f("tt", s=s))   # rx__sens_peripheral2_BY_p1
    expect_true(f("u", s=s))    # rx__sens_peripheral2_BY_v1
    expect_true(f("v", s=s))    # rx__sens_peripheral2_BY_p2
    expect_true(f("w", s=s))    # rx__sens_peripheral2_BY_p3
    expect_true(f("x", s=s))    # rx__sens_peripheral2_BY_p4
    expect_true(f("y", s=s))    # rx__sens_peripheral2_BY_p5
    expect_true(f("z", s=s))    # rx__sens_peripheral2_BY_ka
    expect_true(f("aa", s=s))   #rx__sens_depot_BY_ka

  })

  f <- function(txt, var, w1, w2) {
    .txt <- paste0("Derivative(", txt, ",", var, ")")
    .txt <- str2lang(rxFromSE(.txt))
    if (is.null(w1)) {
      expect_equal(0, .txt)
    } else {
      expect_equal(eval(.txt[[7]]), w1)
      expect_equal(eval(.txt[[8]]), w2)
    }
  }

  test_that("test linCmtB 3 compartment oral derivatives", {

    # linCmt()
    f("linCmtB(rx__PTR__,t,2,3,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "cl", -2, 0)
    f("linCmtB(rx__PTR__,t,2,3,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v",  -2, 1)
    f("linCmtB(rx__PTR__,t,2,3,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "q",  -2, 2)
    f("linCmtB(rx__PTR__,t,2,3,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v2", -2, 3)
    f("linCmtB(rx__PTR__,t,2,3,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "q2", -2, 4)
    f("linCmtB(rx__PTR__,t,2,3,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v3", -2, 5)
    f("linCmtB(rx__PTR__,t,2,3,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "ka", -2, 6)

    # state values

    # depot:
    f("linCmtB(rx__PTR__,t,2,3,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 0, 0)
    f("linCmtB(rx__PTR__,t,2,3,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  0, 1)
    f("linCmtB(rx__PTR__,t,2,3,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  0, 2)
    f("linCmtB(rx__PTR__,t,2,3,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", 0, 3)
    f("linCmtB(rx__PTR__,t,2,3,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", 0, 4)
    f("linCmtB(rx__PTR__,t,2,3,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", 0, 5)
    f("linCmtB(rx__PTR__,t,2,3,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", 0, 6)

    # central:
    f("linCmtB(rx__PTR__,t,2,3,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 1, 0)
    f("linCmtB(rx__PTR__,t,2,3,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  1, 1)
    f("linCmtB(rx__PTR__,t,2,3,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  1, 2)
    f("linCmtB(rx__PTR__,t,2,3,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", 1, 3)
    f("linCmtB(rx__PTR__,t,2,3,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", 1, 4)
    f("linCmtB(rx__PTR__,t,2,3,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", 1, 5)
    f("linCmtB(rx__PTR__,t,2,3,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", 1, 6)

    # peripharal1
    f("linCmtB(rx__PTR__,t,2,3,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 2, 0)
    f("linCmtB(rx__PTR__,t,2,3,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  2, 1)
    f("linCmtB(rx__PTR__,t,2,3,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  2, 2)
    f("linCmtB(rx__PTR__,t,2,3,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", 2, 3)
    f("linCmtB(rx__PTR__,t,2,3,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", 2, 4)
    f("linCmtB(rx__PTR__,t,2,3,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", 2, 5)
    f("linCmtB(rx__PTR__,t,2,3,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", 2, 6)

    # peripharal2
    f("linCmtB(rx__PTR__,t,2,3,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 3, 0)
    f("linCmtB(rx__PTR__,t,2,3,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  3, 1)
    f("linCmtB(rx__PTR__,t,2,3,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  3, 2)
    f("linCmtB(rx__PTR__,t,2,3,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", 3, 3)
    f("linCmtB(rx__PTR__,t,2,3,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", 3, 4)
    f("linCmtB(rx__PTR__,t,2,3,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", 3, 5)
    f("linCmtB(rx__PTR__,t,2,3,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", 3, 6)

  })

  test_that("test linCmtB 3 compartment derivatives", {

    # linCmt()
    f("linCmtB(rx__PTR__,t,2,3,0,-1,-1,1,cl,v,q,v2,q2,v3,0)", "cl", -2, 0)
    f("linCmtB(rx__PTR__,t,2,3,0,-1,-1,1,cl,v,q,v2,q2,v3,0)", "v",  -2, 1)
    f("linCmtB(rx__PTR__,t,2,3,0,-1,-1,1,cl,v,q,v2,q2,v3,0)", "q",  -2, 2)
    f("linCmtB(rx__PTR__,t,2,3,0,-1,-1,1,cl,v,q,v2,q2,v3,0)", "v2", -2, 3)
    f("linCmtB(rx__PTR__,t,2,3,0,-1,-1,1,cl,v,q,v2,q2,v3,0)", "q2", -2, 4)
    f("linCmtB(rx__PTR__,t,2,3,0,-1,-1,1,cl,v,q,v2,q2,v3,0)", "v3", -2, 5)
    f("linCmtB(rx__PTR__,t,2,3,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

    # state values

    # central:
    f("linCmtB(rx__PTR__,t,2,3,0,0,-2,1,cl,v,q,v2,q2,v3,0)", "cl", 0, 0)
    f("linCmtB(rx__PTR__,t,2,3,0,0,-2,1,cl,v,q,v2,q2,v3,0)", "v",  0, 1)
    f("linCmtB(rx__PTR__,t,2,3,0,0,-2,1,cl,v,q,v2,q2,v3,0)", "q",  0, 2)
    f("linCmtB(rx__PTR__,t,2,3,0,0,-2,1,cl,v,q,v2,q2,v3,0)", "v2", 0, 3)
    f("linCmtB(rx__PTR__,t,2,3,0,0,-2,1,cl,v,q,v2,q2,v3,0)", "q2", 0, 4)
    f("linCmtB(rx__PTR__,t,2,3,0,0,-2,1,cl,v,q,v2,q2,v3,0)", "v3", 0, 5)
    f("linCmtB(rx__PTR__,t,2,3,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

    # peripharal1:
    f("linCmtB(rx__PTR__,t,2,3,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 1, 0)
    f("linCmtB(rx__PTR__,t,2,3,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  1, 1)
    f("linCmtB(rx__PTR__,t,2,3,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  1, 2)
    f("linCmtB(rx__PTR__,t,2,3,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", 1, 3)
    f("linCmtB(rx__PTR__,t,2,3,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", 1, 4)
    f("linCmtB(rx__PTR__,t,2,3,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", 1, 5)
    f("linCmtB(rx__PTR__,t,2,3,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

    # peripharal2
    f("linCmtB(rx__PTR__,t,2,3,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 2, 0)
    f("linCmtB(rx__PTR__,t,2,3,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  2, 1)
    f("linCmtB(rx__PTR__,t,2,3,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  2, 2)
    f("linCmtB(rx__PTR__,t,2,3,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", 2, 3)
    f("linCmtB(rx__PTR__,t,2,3,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", 2, 4)
    f("linCmtB(rx__PTR__,t,2,3,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", 2, 5)
    f("linCmtB(rx__PTR__,t,2,3,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

  })


  test_that("test linCmtB 2 compartment oral derivatives", {

    # linCmt()
    f("linCmtB(rx__PTR__,t,2,2,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "cl", -2, 0)
    f("linCmtB(rx__PTR__,t,2,2,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v",  -2, 1)
    f("linCmtB(rx__PTR__,t,2,2,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "q",  -2, 2)
    f("linCmtB(rx__PTR__,t,2,2,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v2", -2, 3)
    f("linCmtB(rx__PTR__,t,2,2,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "ka", -2, 4)

    # state values

    # depot:
    f("linCmtB(rx__PTR__,t,2,2,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 0, 0)
    f("linCmtB(rx__PTR__,t,2,2,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  0, 1)
    f("linCmtB(rx__PTR__,t,2,2,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  0, 2)
    f("linCmtB(rx__PTR__,t,2,2,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", 0, 3)
    f("linCmtB(rx__PTR__,t,2,2,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", 0, 4)

    # central:
    f("linCmtB(rx__PTR__,t,2,2,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 1, 0)
    f("linCmtB(rx__PTR__,t,2,2,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  1, 1)
    f("linCmtB(rx__PTR__,t,2,2,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  1, 2)
    f("linCmtB(rx__PTR__,t,2,2,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", 1, 3)
    f("linCmtB(rx__PTR__,t,2,2,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", 1, 4)

    # peripharal1
    f("linCmtB(rx__PTR__,t,2,2,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 2, 0)
    f("linCmtB(rx__PTR__,t,2,2,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  2, 1)
    f("linCmtB(rx__PTR__,t,2,2,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  2, 2)
    f("linCmtB(rx__PTR__,t,2,2,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", 2, 3)
    f("linCmtB(rx__PTR__,t,2,2,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", 2, 4)

    # peripharal2
    f("linCmtB(rx__PTR__,t,2,2,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,2,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

  })

  test_that("test linCmtB 2 compartment derivatives", {

    # linCmt()
    f("linCmtB(rx__PTR__,t,2,2,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "cl", -2, 0)
    f("linCmtB(rx__PTR__,t,2,2,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v",  -2, 1)
    f("linCmtB(rx__PTR__,t,2,2,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "q",  -2, 2)
    f("linCmtB(rx__PTR__,t,2,2,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v2", -2, 3)
    f("linCmtB(rx__PTR__,t,2,2,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

    # state values

    # central:
    f("linCmtB(rx__PTR__,t,2,2,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 0, 0)
    f("linCmtB(rx__PTR__,t,2,2,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  0, 1)
    f("linCmtB(rx__PTR__,t,2,2,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  0, 2)
    f("linCmtB(rx__PTR__,t,2,2,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", 0, 3)
    f("linCmtB(rx__PTR__,t,2,2,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

    # peripharal1
    f("linCmtB(rx__PTR__,t,2,2,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 1, 0)
    f("linCmtB(rx__PTR__,t,2,2,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  1, 1)
    f("linCmtB(rx__PTR__,t,2,2,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  1, 2)
    f("linCmtB(rx__PTR__,t,2,2,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", 1, 3)
    f("linCmtB(rx__PTR__,t,2,2,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,1,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

    # peripharal2
    f("linCmtB(rx__PTR__,t,2,2,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,2,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

  })

  test_that("test linCmtB 1 compartment oral derivatives", {

    # linCmt()
    f("linCmtB(rx__PTR__,t,2,1,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "cl", -2, 0)
    f("linCmtB(rx__PTR__,t,2,1,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v",  -2, 1)
    f("linCmtB(rx__PTR__,t,2,1,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "q",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "ka", -2, 2)

    # state values

    # depot:
    f("linCmtB(rx__PTR__,t,2,1,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 0, 0)
    f("linCmtB(rx__PTR__,t,2,1,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  0, 1)
    f("linCmtB(rx__PTR__,t,2,1,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,0,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", 0, 2)

    # central:
    f("linCmtB(rx__PTR__,t,2,1,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 1, 0)
    f("linCmtB(rx__PTR__,t,2,1,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  1, 1)
    f("linCmtB(rx__PTR__,t,2,1,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,1,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", 1, 2)

    # peripharal1
    f("linCmtB(rx__PTR__,t,2,1,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,2,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

    # peripharal2
    f("linCmtB(rx__PTR__,t,2,1,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,1,1,3,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

  })

  test_that("test linCmtB 1 compartment derivatives", {

    # linCmt()
    f("linCmtB(rx__PTR__,t,2,1,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "cl", -2, 0)
    f("linCmtB(rx__PTR__,t,2,1,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v",  -2, 1)
    f("linCmtB(rx__PTR__,t,2,1,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "q",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,-1,-1,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

    # state values

    # central
    f("linCmtB(rx__PTR__,t,2,1,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", 0, 0)
    f("linCmtB(rx__PTR__,t,2,1,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  0, 1)
    f("linCmtB(rx__PTR__,t,2,1,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,0,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

    # peripharal1
    f("linCmtB(rx__PTR__,t,2,1,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,2,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

    # peripharal2
    f("linCmtB(rx__PTR__,t,2,1,0,3,-2,1,cl,v,q,v2,q2,v3,ka)", "cl", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,3,-2,1,cl,v,q,v2,q2,v3,ka)", "q",  NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,3,-2,1,cl,v,q,v2,q2,v3,ka)", "q2", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,3,-2,1,cl,v,q,v2,q2,v3,ka)", "v3", NULL)
    f("linCmtB(rx__PTR__,t,2,1,0,3,-2,1,cl,v,q,v2,q2,v3,ka)", "ka", NULL)

  })


  for (prop in c("dur", "rate", "lag", "alag", "f", "F", "0")) {
    for (cmt in c("depot", "central", "peripheral1", "peripheral2")) {
      val <- ifelse(prop=="0",
                    sprintf("%s(0)",cmt),
                    sprintf("%s(%s)", prop, cmt))
      test <- c("cl=exp(tvcl)",
                "v=exp(tvv)",
                "q=exp(tvq)",
                "v2=exp(tvv2)",
                "q2=exp(tvq2)",
                "v3=exp(tvv3)",
                "mat=exp(tvmat)",
                "D1=mat*(1-expit(tvfrd1,0,1))",
                "ka=1/(mat*expit(tvfrd1,0,1))",
                "cp=linCmtA(rx__PTR__,t,2,3,1,-1,1,cl,v,q,v2,q2,v3,ka)",
                sprintf("%s=D1", val),
                "rx_pred_=cp")

      m <- paste(test, collapse = "\n")
      shouldError <- prop == "0" || (cmt %in% c("peripheral1", "peripheral2") &&
                                       prop %in% c("dur", "rate"))

      if (shouldError) {
        test_that(sprintf("rxS errors in %s parameter in linCmtA()", val), {
          expect_error(rxS(m))
        })
      } else {
        test_that(sprintf("rxS processes %s parameter in linCmtA()", val), {
          expect_error(rxS(m), NA)
        })
      }
    }
  }

  test_that(".rxFromSEnum handles empty input (#1109)", {
    expect_equal(.rxFromSEnum(numeric(0)), character(0))
    expect_equal(.rxFromSEnum(character(0)), character(0))
  })

  test_that("rxFromSE does not leak user-workspace variables (#1109)", {
    # a variable in the global environment named like a model symbol must
    # neither error (zero-length) nor be substituted into the conversion
    assign("rxFromSE1109center", numeric(0), envir = globalenv())
    on.exit(rm("rxFromSE1109center", envir = globalenv()), add = TRUE)
    .x <- symengine::S("exp(ETA_3_ + THETA_3_ - (ETA_3_ + THETA_3_))*rx__sens_x_BY_ETA_1___/rxFromSE1109center")
    expect_equal(
      rxFromSE(.x),
      "exp(ETA[3]+THETA[3]-(ETA[3]+THETA[3]))*rx__sens_x_BY_ETA_1___/rxFromSE1109center")
    assign("rxFromSE1109center", 42, envir = globalenv())
    .y <- symengine::S("rx__sens_x_BY_ETA_1___/rxFromSE1109center")
    expect_equal(rxFromSE(.y), "rx__sens_x_BY_ETA_1___/rxFromSE1109center")
  })

})

rxTest({
  # test id=="item"

  f1 <- rxode2({
    a <- 1
    if (id == "matt") {
      a <- 2
    }
  })

  df <- data.frame(
    ID = c("matt", "not"),
    TIME = c(0, 1)
  )

  ref <- rxSolve(f1, df)

  test_that("id==matt #1", {
    expect_equal(ref$a, c(2, 1))
  })

  df <- data.frame(
    ID = c("not", "matt"),
    TIME = c(0, 1)
  )

  ref <- rxSolve(f1, df)

  test_that("id==matt #2", {
    expect_equal(ref$a, c(1, 2))
  })

  f2 <- rxode2({
    a <- 1
    if ("matt" == ID) {
      a <- 2
    }
  })

  df <- data.frame(
    ID = c("matt", "not"),
    TIME = c(0, 1)
  )

  ref <- rxSolve(f2, df)

  test_that("id==matt #3", {
    expect_equal(ref$a, c(2, 1))
  })

  df <- data.frame(
    ID = c("not", "matt"),
    TIME = c(0, 1)
  )

  ref <- rxSolve(f2, df)

  test_that("id==mat #4", {
    expect_equal(ref$a, c(1, 2))
  })

  f1 <- rxode2({
    a <- 1
    if (id != "matt") {
      a <- 2
    }
  })

  df <- data.frame(
    ID = c("matt", "not"),
    TIME = c(0, 1)
  )

  ref <- rxSolve(f1, df)

  test_that("id!=matt #1", {
    expect_equal(ref$a, c(1, 2))
  })

  df <- data.frame(
    ID = c("not", "matt"),
    TIME = c(0, 1)
  )

  ref <- rxSolve(f1, df)

  test_that("id!=matt #2", {
    expect_equal(ref$a, c(2, 1))
  })

  f1 <- rxode2({
    a <- 1
    if (id == "100") {
      a <- 2
    }
  })

  df <- data.frame(
    ID = c(100, 200),
    TIME = c(0, 1)
  )

  ref <- rxSolve(f1, df)

  test_that("id2", {
    expect_equal(ref$a, c(2, 1))
  })

  test_that("id equality error", {
    suppressMessages(expect_error(rxode2({
      a <- 1
      if (id == 100) {
        a <- 2
      }
    })))
  })

  f1 <- rxode2({
    a <- 1 + (id == "100") * 2
  })

  ref <- rxSolve(f1, df)

  test_that("id3", {
    expect_equal(ref$a, c(3, 1))
  })

  f1 <- rxode2({
    a <- 1 + (id == "100") * 2
  })

  ref <- rxSolve(f1, df)

  test_that("id3", {
    expect_equal(ref$a, c(3, 1))
  })

  # test cmt=="item"/no.

  f3 <- rxode2({
    a <- 3
    if (cmt == "depot") {
      a <- 2
    }
  })

  df <- data.frame(
    ID = c("not", "matt"),
    TIME = c(0, 1),
    cmt = structure(1:2, levels = c("depot", "central"), class = "factor")
  )

  tmp <- rxSolve(f3, df)

  test_that("cmt==\"depot\"", {
    expect_equal(tmp$a, c(2, 3))
  })

  f3 <- rxode2({
    a <- 3
    if (cmt == 1) {
      a <- 2
    }
  })

  tmp <- rxSolve(f3, df)

  test_that("cmt==1", {
    expect_equal(tmp$a, c(2, 3))
  })

  f3 <- rxode2({
    a <- 3
    if (cmt != "depot") {
      a <- 2
    }
  })

  tmp <- rxSolve(f3, df)

  test_that("cmt!=\"depot\"", {
    expect_equal(tmp$a, c(3, 2))
  })

  f3 <- rxode2({
    a <- 3
    if (cmt != 1) {
      a <- 2
    }
  })

  test_that("cmt!=\"depot\"", {
    expect_equal(tmp$a, c(3, 2))
  })

  f3 <- rxode2({
    a <- 1
    if ("depot" == cmt) {
      a <- 2
    }
  })

  tmp <- rxSolve(f3, df)

  test_that("cmt==\"depot\"", {
    expect_equal(tmp$a, c(2, 1))
  })

  f3 <- rxode2({
    a <- 1
    if ("depot" != cmt) {
      a <- 2
    }
  })

  tmp <- rxSolve(f3, df)

  test_that("cmt==\"depot\"", {
    expect_equal(tmp$a, c(1, 2))
  })

  test_that("translation to and from SE", {
    a <- rxToSE("id==\"matt\"")
    expect_equal(a, "rxEq(id,rxQ__matt__rxQ)")
    b <- rxFromSE(a)
    expect_equal(b, "(id==\"matt\")")
  })

  tmp <- "C2=centr/V;\nC3=peri/V2;\nd/dt(depot)=-KA*depot;\nd/dt(centr)=KA*depot-CL*C2-Q*C2+Q*C3;\nd/dt(peri)=Q*C2-Q*C3;\nC4=CMT;\nif(CMT==\"depot\"){\nprd=depot;\n}\nif(CMT==\"centr\"){\nprd=centr;\n}\nif(CMT==\"peri\"){\nprd=peri;\n}\n"

  test_that("prune checks", {
    expect_equal(
      rxPrune(tmp),
      "C2=centr/V\nC3=peri/V2\nd/dt(depot)=-KA*depot\nd/dt(centr)=KA*depot-CL*C2-Q*C2+Q*C3\nd/dt(peri)=Q*C2-Q*C3\nC4=CMT\nprd=(CMT==\"depot\")*(depot)\nprd=(CMT==\"centr\")*(centr)+(1-((CMT==\"centr\")))*(prd)\nprd=(CMT==\"peri\")*(peri)+(1-((CMT==\"peri\")))*(prd)"
    )
  })
})

rxTest({
  test_that("user function tests", {
    suppressMessages(expect_error(rxode2("a=fun(d,b,c)")))
    expect_error(rxFromSE("Derivative(fun(a,b,c),a)"))

    fun <- "
 double fun(double a, double b, double c) {
   return a*a+b*a+c;
 }
"
    expect_error(rxFun(list("fun"), c("a", "b", "c"), fun))
    expect_error(rxFun("fun", c("a", "b", "c")))
    rxFun("fun", c("a", "b", "c"), fun)
    expect_error(rxFun("fun", c("a", "b", "c"), fun))

    tmp <- rxode2("a=fun(d,b,c)")

    tmp <- rxSolve(tmp, c(c = 2, b = 4, d = 8), et(1))
    expect_equal(tmp$a, 8^2 + 4 * 8 + 2)

    expect_equal(rxToSE("fun(a,b,c)"), "fun(a,b,c)")
    expect_equal(rxFromSE("fun(a,b,c)"), "fun(a,b,c)")

    expect_error(rxToSE("fun(a)"))
    expect_error(rxFromSE("fun(a)"))

    expect_equal(
      rxFromSE("Derivative(fun(a,b,c),a)", unknownDerivatives = "central"),
      "(fun(a-3.02772722619667e-06,b,c)-fun(a+3.02772722619667e-06,b,c))/6.05545445239334e-06"
    )
    expect_equal(
      rxFromSE("Derivative(fun(a,b,c),a)", unknownDerivatives = "forward"),
      "(fun((a)+6.05545445239334e-06,b,c)-fun(a,b,c))/6.05545445239334e-06"
    )
    expect_error(rxFromSE("Derivative(fun(a,b,c),a)", unknownDerivatives = "error"))

    ## now add derivative table
    rxD("fun", list(
      function(a, b, c) {
        paste0("2*", a, "+", b)
      },
      function(a, b, c) {
        return(a)
      },
      function(a, b, c) {
        return("0.0")
      }
    ))

    expect_equal(rxFromSE("Derivative(fun(a1,b1,c1),a1)"), "2*a1+b1")
    expect_equal(rxFromSE("Derivative(fun(a1,b1,c1),b1)"), "a1")
    expect_equal(rxFromSE("Derivative(fun(a1,b1,c1),c1)"), "0.0")

    expect_warning(rxD("fun", list(function(a, b, c) {
      paste0("2*", a, "+", b)
    })))

    expect_equal(rxFromSE("Derivative(fun(a1,b1,c1),a1)"), "2*a1+b1")
    expect_equal(
      rxFromSE("Derivative(fun(a,b,c),b)", unknownDerivatives = "central"),
      "(fun(a,b-3.02772722619667e-06,c)-fun(a,b+3.02772722619667e-06,c))/6.05545445239334e-06"
    )

    expect_warning(rxD("fun", list(NULL, function(a, b, c) {
      paste0(a)
    })))

    expect_equal(
      rxFromSE("Derivative(fun(a,b,c),b)"),
      "a"
    )

    expect_equal(
      rxFromSE("Derivative(fun(a,b,c),a)", unknownDerivatives = "central"),
      "(fun(a-3.02772722619667e-06,b,c)-fun(a+3.02772722619667e-06,b,c))/6.05545445239334e-06"
    )

    expect_error(rxD("fun", "matt"))
    expect_error(rxD("fun", list()))
    expect_error(rxD("fun", list(NULL, "a", function(x) {
      x
    })))

    ## Errors do not replace derivative table
    expect_equal(
      rxFromSE("Derivative(fun(a,b,c),b)"),
      "a"
    )
    expect_equal(
      rxFromSE("Derivative(fun(a,b,c),a)", unknownDerivatives = "central"),
      "(fun(a-3.02772722619667e-06,b,c)-fun(a+3.02772722619667e-06,b,c))/6.05545445239334e-06"
    )

    expect_error(rxRmFun(list("a")))
    expect_warning(rxRmFun("a"))

    rxRmFun("fun")
    suppressMessages(expect_error(rxode2("a=fun(d,b,c)")))
    expect_error(rxFromSE("Derivative(fun(a,b,c),a)"))
  })

})

#' Create a file-backed event table reference
#'
#' @param path path to the file containing the event table
#' @param format file format: "auto" (detect from extension), "parquet", "csv", "fst", or "rds"
#' @param id name of the subject ID column
#' @return An \code{rxEtFile} object
#' @export
rxEventTableFile <- function(path, format = c("auto", "parquet", "csv", "fst", "rds"),
                              id = "id") {
  format <- match.arg(format)
  if (format == "auto") {
    .ext <- tolower(tools::file_ext(path))
    format <- switch(.ext, parquet = "parquet", csv = "csv", fst = "fst", rds = "rds", "rds")
  }
  structure(list(path = path, format = format, id_col = id), class = "rxEtFile")
}

.rxEtFileReadFull <- function(x) {
  switch(x$format,
    parquet = {
      if (!requireNamespace("arrow", quietly = TRUE))
        stop("package 'arrow' is required for parquet format")
      as.data.frame(arrow::read_parquet(x$path))
    },
    csv     = data.table::fread(x$path, data.table = FALSE),
    fst     = {
      if (!requireNamespace("fst", quietly = TRUE))
        stop("package 'fst' is required for fst format")
      fst::read_fst(x$path)
    },
    rds     = readRDS(x$path)
  )
}

.rxEtFileReadCols <- function(x, cols) {
  switch(x$format,
    parquet = {
      if (!requireNamespace("arrow", quietly = TRUE))
        stop("package 'arrow' is required for parquet format")
      as.data.frame(arrow::read_parquet(x$path, col_select = dplyr::all_of(cols)))
    },
    csv     = data.table::fread(x$path, select = cols, data.table = FALSE),
    fst     = {
      if (!requireNamespace("fst", quietly = TRUE))
        stop("package 'fst' is required for fst format")
      fst::read_fst(x$path, columns = cols)
    },
    rds     = {
      .d <- readRDS(x$path)
      .d[, intersect(cols, names(.d)), drop = FALSE]
    }
  )
}

.rxEtFileReadChunk <- function(x, ids) {
  .dat <- .rxEtFileReadFull(x)
  .dat[.dat[[x$id_col]] %in% ids, , drop = FALSE]
}

# First n rows without materializing the whole file (rds has no lazy read).
.rxEtFileReadHead <- function(x, n = 100L) {
  switch(x$format,
    parquet = {
      if (!requireNamespace("arrow", quietly = TRUE))
        stop("package 'arrow' is required for parquet format")
      as.data.frame(utils::head(arrow::open_dataset(x$path), n))
    },
    csv     = data.table::fread(x$path, nrows = n, data.table = FALSE),
    fst     = {
      if (!requireNamespace("fst", quietly = TRUE))
        stop("package 'fst' is required for fst format")
      .n <- min(as.integer(n), fst::metadata_fst(x$path)$nrOfRows)
      fst::read_fst(x$path, from = 1L, to = .n)
    },
    rds     = utils::head(readRDS(x$path), n)
  )
}

rxMemSummary.rxEtFile <- function(x, ...) {
  .cols <- c(x$id_col, "evid")
  .dat <- .rxEtFileReadCols(x, .cols)
  .rxMemSummarizeDat(.dat)
}

# -- Main OOM solve loop -------------------------------------------------------

.rxSolveOom <- function(object, params, events, inits, .ctl, .envir = parent.frame()) {
  .prefix   <- .ctl$file
  .nDaemons <- if (is.null(.ctl$parallel)) 0L else as.integer(.ctl$parallel)
  .useMirai <- .nDaemons > 0L && requireNamespace("mirai", quietly = TRUE)

  # Normalize: if events is an rxEt but params is not, events is the params; swap
  if ((is.rxEt(params) || rxIs(params, "rx.event") || inherits(params, "rxEtFile")) &&
        !is.rxEt(events)) {
    .tmp  <- events
    events <- params
    params <- .tmp
  }

  # Build memory summary from events
  .summary <- if (inherits(events, "rxEtFile")) {
    rxMemSummary.rxEtFile(events)
  } else {
    .evDf <- if (is.rxEt(events)) as.data.frame(events) else as.data.frame(events)
    .rxMemSummarizeDat(.evDf)
  }

  # Chunk size
  .chunkSize <- if (!is.null(.ctl$chunkSize)) {
    as.integer(.ctl$chunkSize)
  } else {
    .rxOomChunkSize(object, .summary, .ctl)
  }

  # Capture or assign base seed
  .baseSeed <- rxGetSeed()
  if (.baseSeed == -1L) {
    .baseSeed <- sample.int(.Machine$integer.max, 1L)
    rxSetSeed(.baseSeed)
  }

  # Split IDs into chunks
  .allIds    <- .summary$id
  .nSub      <- length(.allIds)
  .nChunks   <- ceiling(.nSub / .chunkSize)
  .chunkList <- split(.allIds, ceiling(seq_len(.nSub) / .chunkSize))

  .manifest <- list(
    version = 1L, prefix = .prefix,
    chunks  = character(.nChunks), nrows = integer(.nChunks),
    paramChunks = character(.nChunks), inits = NULL,
    seed    = .baseSeed
  )
  .outFiles    <- character(.nChunks)
  .paramFiles  <- character(.nChunks)
  .cumSub      <- 0L

  # Control args forwarded to each chunk rxSolve call (strip OOM-specific fields)
  .fwdCtlArgs <- as.list(.ctl)
  .fwdCtlArgs$file      <- NULL
  .fwdCtlArgs$chunkSize <- NULL
  .fwdCtlArgs$parallel  <- NULL
  .fwdCtlArgs$serializeFile <- NULL

  # Pre-draw ALL subjects' etas once using the base seed so that chunked solves
  # reproduce the same etas as a single full rxSolve(seed=baseSeed) call.
  #
  # rxSolve_ calls seedEng(op->cores) BEFORE rxSimThetaOmega, advancing rxSeed
  # by 2*ncores.  We replicate that here with rxSeedEng() so our standalone
  # rxSimThetaOmega sees the same effective seed as the internal call in rxSolve_.
  .preDrawnParams <- NULL
  if (!is.null(.ctl$omega)) {
    .ncores <- if (!is.null(.ctl$cores) && .ctl$cores > 0L) {
      as.integer(.ctl$cores)
    } else {
      getRxThreads()
    }
    rxSetSeed(.baseSeed)
    rxSeedEng(.ncores)
    .preDrawnParams <- rxSimThetaOmega(
      params          = params,
      omega           = .ctl$omega,
      omegaDf         = .ctl$omegaDf,
      omegaLower      = if (!is.null(.ctl$omegaLower))  .ctl$omegaLower  else -Inf,
      omegaUpper      = if (!is.null(.ctl$omegaUpper))  .ctl$omegaUpper  else  Inf,
      omegaIsChol     = if (!is.null(.ctl$omegaIsChol)) .ctl$omegaIsChol else FALSE,
      omegaSeparation = if (!is.null(.ctl$omegaSeparation)) .ctl$omegaSeparation else "auto",
      omegaXform      = if (!is.null(.ctl$omegaXform))  .ctl$omegaXform  else 1L,
      nSub            = .nSub,
      nCoresRV        = 1L,
      nStud           = 1L
    )
    # Strip omega from forwarded args -- etas are now baked into per-chunk params
    .fwdCtlArgs$omega           <- NULL
    .fwdCtlArgs$omegaDf         <- NULL
    .fwdCtlArgs$omegaLower      <- NULL
    .fwdCtlArgs$omegaUpper      <- NULL
    .fwdCtlArgs$omegaIsChol     <- NULL
    .fwdCtlArgs$omegaSeparation <- NULL
    .fwdCtlArgs$omegaXform      <- NULL
  }

  # Normalize: ensure id column is always present so chunks can be rbind'd.
  # Single-subject solves drop the id column; stamp it back from .chunkIds.
  .normalizeResult <- function(.result, .chunkIds) {
    .df <- as.data.frame(.result)
    if (!("id" %in% names(.df))) {
      .nPerSub <- nrow(.df) %/% max(length(.chunkIds), 1L)
      .df <- cbind(id = rep(.chunkIds, each = .nPerSub), .df)
    }
    .df
  }

  .writeResult <- function(.result, .chunkIds) {
    .df <- .normalizeResult(.result, .chunkIds)
    if (.rxOomHasArrow()) {
      .f <- tempfile(fileext = ".parquet")
      arrow::write_parquet(.df, .f)
    } else {
      .f <- tempfile(fileext = ".rds")
      saveRDS(.df, .f)
    }
    .f
  }

  # Persist the per-subject parameter table (res$params) for one chunk. The id
  # column is stamped from .chunkIds so the chunk param tables concatenate
  # cleanly (single-subject solves drop the id column).
  .writeParams <- function(.result, .chunkIds) {
    .pars <- tryCatch(as.data.frame(.result$params), error = function(e) NULL)
    if (is.null(.pars) || nrow(.pars) == 0L) return(NA_character_)
    if (!("id" %in% names(.pars)) && nrow(.pars) == length(.chunkIds)) {
      .pars <- cbind(id = .chunkIds, .pars)
    }
    if (.rxOomHasArrow()) {
      .f <- tempfile(fileext = ".parquet")
      arrow::write_parquet(.pars, .f)
    } else {
      .f <- tempfile(fileext = ".rds")
      saveRDS(.pars, .f)
    }
    .f
  }

  .extractChunkEvents <- function(.chunkIds) {
    if (inherits(events, "rxEtFile")) {
      .rxEtFileReadChunk(events, .chunkIds)
    } else {
      .evDf  <- if (is.rxEt(events)) as.data.frame(events) else as.data.frame(events)
      .idCol <- grep("^id$", names(.evDf), ignore.case = TRUE, value = TRUE)[1]
      if (is.na(.idCol)) {
        .evDf
      } else {
        .evDf[.evDf[[.idCol]] %in% .chunkIds, , drop = FALSE]
      }
    }
  }

  if (.useMirai) {
    .modelObj <- if (inherits(object, c("rxode2", "rxDll"))) object else rxode2(object)
    mirai::daemons(.nDaemons)
    on.exit(mirai::daemons(0), add = TRUE)
    .chunkEvList   <- vector("list", .nChunks)
    .chunkParamsList <- vector("list", .nChunks)
    for (.i in seq_len(.nChunks)) {
      .chunkEvList[[.i]] <- .extractChunkEvents(.chunkList[[.i]])
      .nThis <- length(.chunkList[[.i]])
      .chunkParamsList[[.i]] <- if (!is.null(.preDrawnParams)) {
        .preDrawnParams[(.cumSub + 1L):(.cumSub + .nThis), , drop = FALSE]
      } else {
        params
      }
      .cumSub <- .cumSub + .nThis
    }
    .cumSub <- 0L
    .chunkIdsList <- .chunkList
    # Resolve the backend in the parent (where the rxode2.oom.backend option
    # lives) and forward it to the daemons: the option for faithful propagation,
    # and the already-resolved write decision because the daemon closure cannot
    # reach the unexported .rxOomHasArrow() helper.
    .backendOpt    <- .rxOomBackendOpt()
    .useArrowWrite <- .rxOomHasArrow()
    .tasks <- mirai::mirai_map(
      seq_len(.nChunks),
      function(.i, .modelObj, .chunkEvList, .chunkIdsList, .chunkParamsList, .inits, .fwdCtlArgs, .mainTmp, .backendOpt, .useArrowWrite) {
        library(rxode2)
        options(rxode2.oom.backend = .backendOpt)
        .result <- do.call(rxSolve,
                           c(list(object = .modelObj, params = .chunkParamsList[[.i]],
                                  events = .chunkEvList[[.i]], inits = .inits),
                             .fwdCtlArgs))
        .df <- as.data.frame(.result)
        if (!("id" %in% names(.df))) {
          .ids <- .chunkIdsList[[.i]]
          .nPerSub <- nrow(.df) %/% max(length(.ids), 1L)
          .df <- cbind(id = rep(.ids, each = .nPerSub), .df)
        }
        # Write to the parent process's tempdir (shared filesystem). A daemon's
        # own session tempdir is removed when the daemon shuts down, which would
        # leave the manifest pointing at deleted chunk files.
        # Per-subject parameter table; stamp id so chunks concatenate cleanly.
        .pars <- tryCatch(as.data.frame(.result$params), error = function(e) NULL)
        .ids  <- .chunkIdsList[[.i]]
        if (!is.null(.pars) && nrow(.pars) > 0L &&
            !("id" %in% names(.pars)) && nrow(.pars) == length(.ids)) {
          .pars <- cbind(id = .ids, .pars)
        }
        if (.useArrowWrite) {
          .f <- tempfile(fileext = ".parquet", tmpdir = .mainTmp)
          arrow::write_parquet(.df, .f)
          .pf <- if (!is.null(.pars) && nrow(.pars) > 0L) {
            .p <- tempfile(fileext = ".parquet", tmpdir = .mainTmp)
            arrow::write_parquet(.pars, .p)
            .p
          } else NA_character_
        } else {
          .f <- tempfile(fileext = ".rds", tmpdir = .mainTmp)
          saveRDS(.df, .f)
          .pf <- if (!is.null(.pars) && nrow(.pars) > 0L) {
            .p <- tempfile(fileext = ".rds", tmpdir = .mainTmp)
            saveRDS(.pars, .p)
            .p
          } else NA_character_
        }
        list(file = .f, nrows = nrow(.df), paramFile = .pf,
             inits = tryCatch(.result$inits, error = function(e) NULL))
      },
      .args = list(.modelObj = .modelObj,
                   .chunkEvList = .chunkEvList, .chunkIdsList = .chunkIdsList,
                   .chunkParamsList = .chunkParamsList,
                   .inits = inits, .fwdCtlArgs = .fwdCtlArgs,
                   .mainTmp = tempdir(),
                   .backendOpt = .backendOpt, .useArrowWrite = .useArrowWrite)
    )
    for (.i in seq_len(.nChunks)) {
      .r <- .tasks[[.i]][]
      if (inherits(.r, "miraiError") || inherits(.r, "errorValue") || is.null(.r$file)) {
        stop(sprintf("parallel chunk %d failed in a mirai daemon: %s", .i,
                     tryCatch(conditionMessage(.r),
                              error = function(e) paste(utils::head(unclass(.r), 1L), collapse = ""))),
             call. = FALSE)
      }
      .outFiles[.i] <- .r$file
      .paramFiles[.i] <- if (is.null(.r$paramFile)) NA_character_ else .r$paramFile
      .manifest$nrows[.i] <- .r$nrows
      if (is.null(.manifest$inits) && !is.null(.r$inits)) {
        .manifest$inits <- .r$inits
      }
    }
  } else {
    for (.i in seq_len(.nChunks)) {
      .chunkIds <- .chunkList[[.i]]
      .nThis    <- length(.chunkIds)
      .chunkEvents <- .extractChunkEvents(.chunkIds)
      .chunkParams <- if (!is.null(.preDrawnParams)) {
        .preDrawnParams[(.cumSub + 1L):(.cumSub + .nThis), , drop = FALSE]
      } else {
        rxSetSeed(as.integer(
          (as.double(.baseSeed) + as.double(.cumSub)) %% .Machine$integer.max
        ))
        params
      }
      .result <- do.call(rxSolve,
                         c(list(object = object, params = .chunkParams,
                                events = .chunkEvents, inits = inits,
                                envir = .envir), .fwdCtlArgs))
      .outFiles[.i]   <- .writeResult(.result, .chunkIds)
      .paramFiles[.i] <- .writeParams(.result, .chunkIds)
      .manifest$nrows[.i] <- nrow(.result)
      if (is.null(.manifest$inits)) {
        .manifest$inits <- tryCatch(.result$inits, error = function(e) NULL)
      }
      .cumSub <- .cumSub + .nThis
    }
  }

  for (.i in seq_len(.nChunks)) {
    .manifest$chunks[.i] <- .outFiles[.i]
  }
  # Drop param chunks that failed to write (NA); keep only valid files.
  .manifest$paramChunks <- .paramFiles[!is.na(.paramFiles)]

  saveRDS(.manifest, paste0(.prefix, "_manifest.rds"))
  .rxSolveOomFromManifest(.manifest)
}

# -- rxSolveOom return object --------------------------------------------------

.rxSolveOomFromManifest <- function(manifest) {
  structure(list(), class = "rxSolveOom", manifest = manifest)
}

.rxOomHasParquet <- function(manifest) {
  any(grepl("\\.parquet$", manifest$chunks))
}

# -- DuckDB lazy query layer over the parquet chunks --------------------------
#
# DuckDB is preferred for lazy access (head, single column, schema) because it
# pushes the LIMIT / column projection into the parquet reader instead of
# materializing whole files.  Everything is guarded so the arrow (and rds)
# fallbacks remain when duckdb/DBI are unavailable.
#
# The backend can be pinned with the `rxode2.oom.backend` option, which makes
# every code path deterministically exercisable (e.g. in tests):
#   "auto"   - duckdb if installed, else arrow, else rds (default)
#   "duckdb" - duckdb query layer over arrow-written parquet
#   "arrow"  - arrow reads/writes, no duckdb
#   "rds"    - plain rds files, no arrow or duckdb
# A requested engine that is not installed silently degrades (duckdb -> arrow
# -> rds), so the option is a preference cap, never a hard requirement.

.rxOomBackendOpt <- function() {
  match.arg(getOption("rxode2.oom.backend", "auto"),
            c("auto", "duckdb", "arrow", "rds"))
}

.rxOomHasDuckdb <- function() {
  if (!(.rxOomBackendOpt() %in% c("auto", "duckdb"))) return(FALSE)
  requireNamespace("duckdb", quietly = TRUE) &&
    requireNamespace("DBI", quietly = TRUE)
}

# Wrapper around the arrow availability check.  Pulling it out (instead of
# calling requireNamespace() inline) lets the `rxode2.oom.backend` option pin
# the rds / arrow / duckdb code paths.
.rxOomHasArrow <- function() {
  if (.rxOomBackendOpt() == "rds") return(FALSE)
  requireNamespace("arrow", quietly = TRUE)
}

# parquet files for a given set of paths
.rxOomParquetFiles <- function(files) {
  files[grepl("\\.parquet$", files)]
}

# Build a DuckDB SQL list literal of parquet paths: ['a.parquet','b.parquet']
.rxOomDuckFileList <- function(files) {
  paste0("[", paste0("'", gsub("'", "''", files), "'", collapse = ", "), "]")
}

# Run a SELECT against a set of parquet files via an in-memory DuckDB and return
# a data.frame.  `sql` must reference the placeholder {tbl}, which is replaced
# with read_parquet([...]).
.rxOomDuckQuery <- function(files, sql) {
  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE), add = TRUE)
  .tbl <- paste0("read_parquet(", .rxOomDuckFileList(files), ")")
  DBI::dbGetQuery(con, gsub("{tbl}", .tbl, sql, fixed = TRUE))
}

# Read the persisted per-subject parameter table (res$params).
.rxOomParams <- function(manifest) {
  .pc <- manifest$paramChunks
  .pq <- .rxOomParquetFiles(.pc)
  if (length(.pq) > 0L) {
    if (.rxOomHasDuckdb()) {
      return(.rxOomDuckQuery(.pq, "SELECT * FROM {tbl}"))
    }
    if (.rxOomHasArrow()) {
      return(do.call(rbind, lapply(.pq, function(.f)
        as.data.frame(arrow::read_parquet(.f)))))
    }
  }
  do.call(rbind, lapply(.pc, readRDS))
}

.rxOomInits <- function(manifest) manifest$inits

# Backend label for the print footer.
.rxOomBackend <- function(manifest) {
  if (.rxOomHasParquet(manifest)) {
    if (.rxOomHasDuckdb()) return(" [DuckDB/Arrow-backed]")
    if (.rxOomHasArrow()) return(" [Arrow-backed]")
  }
  ""
}

#' @export
print.rxSolveOom <- function(x, ...) {
  .m <- attr(x, "manifest")
  .args <- as.list(match.call(expand.dots = TRUE))
  .n <- if (any(names(.args) == "n")) .args$n else 6L
  .bound <- .getBound(x, parent.frame(2))

  cat(cli::cli_format_method({
    .h2(crayon::bold("Solved rxode2 object"))
  }), sep = "\n")

  # Parameters (res$params)
  cat(format.boundParams(.bound), sep = "\n")
  .pars <- .rxOomParams(.m)
  if (requireNamespace("tibble", quietly = TRUE)) {
    print(tibble::as_tibble(.pars))
  } else {
    print(utils::head(.pars))
  }

  # Initial Conditions (res$inits)
  cat(format.boundInits(.bound), sep = "\n")
  print(.rxOomInits(.m))

  # First part of data (object)
  cat(cli::cli_format_method({
    .h2(crayon::bold("First part of data (object):"))
  }), sep = "\n")
  .isDplyr <- requireNamespace("tibble", quietly = TRUE) &&
    getOption("rxode2.display.tbl", TRUE)
  .head <- utils::head(x, n = .n)
  if (.isDplyr) {
    print(tibble::as_tibble(.head), n = .n)
  } else {
    print(.head)
  }

  # Footer: chunk / backend note
  cat(sprintf("<rxSolveOom: %d chunks, %d total rows, prefix='%s'%s>\n",
              length(.m$chunks), sum(.m$nrows), .m$prefix,
              .rxOomBackend(.m)))
  invisible(x)
}

#' Convert an rxSolveOom result to an Arrow Table
#'
#' Reads all parquet chunk files and concatenates them into a single in-memory
#' Arrow Table using \code{arrow::concat_tables()}.  Requires the \code{arrow}
#' package.
#'
#' @param x An \code{rxSolveOom} object.
#' @param ... Ignored.
#' @return An \code{arrow::Table}.
#' @keywords internal
#' @noRd
as_arrow_table.rxSolveOom <- function(x, ...) {
  if (!requireNamespace("arrow", quietly = TRUE))
    stop("package 'arrow' is required for as_arrow_table()")
  .m <- attr(x, "manifest")
  .pq <- .m$chunks[grepl("\\.parquet$", .m$chunks)]
  if (length(.pq) == 0L)
    return(arrow::as_arrow_table(as.data.frame(x)))
  # read_parquet() returns a tibble by default; concat_tables() needs Arrow
  # Tables, so read with as_data_frame = FALSE. concat_tables() also takes the
  # tables as individual `...` arguments, not a list, hence do.call().
  do.call(arrow::concat_tables,
          lapply(.pq, function(.f) arrow::read_parquet(.f, as_data_frame = FALSE)))
}

#' Convert an rxSolveOom result to a lazy Arrow Dataset
#'
#' Opens all parquet chunk files as a single lazy \code{arrow::Dataset} using
#' \code{arrow::open_dataset()}.  The dataset can be filtered and selected with
#' dplyr verbs before calling \code{dplyr::collect()} to materialise.  Requires
#' the \code{arrow} package.
#'
#' This is an rxode2 generic: \pkg{arrow} provides \code{open_dataset()} but no
#' lazy-dataset coercion generic to dispatch on, so rxode2 defines its own.
#'
#' @param x An \code{rxSolveOom} object.
#' @param ... Ignored.
#' @return An \code{arrow::Dataset}.
#' @export
as.arrow <- function(x, ...) UseMethod("as.arrow")

#' @rdname as.arrow
#' @export
as.arrow.rxSolveOom <- function(x, ...) {
  if (!requireNamespace("arrow", quietly = TRUE))
    stop("package 'arrow' is required for as.arrow()")
  .m <- attr(x, "manifest")
  .pq <- .m$chunks[grepl("\\.parquet$", .m$chunks)]
  if (length(.pq) == 0L)
    stop("No parquet chunk files found. Re-run rxSolve() with the arrow package installed.")
  arrow::open_dataset(.pq)
}

#' @export
as.data.frame.rxSolveOom <- function(x, ...) {
  .m <- attr(x, "manifest")
  .total <- sum(.m$nrows)
  if (.total > 1e6)
    message(sprintf("Materializing %.0f rows into memory", .total))
  if (.rxOomHasArrow() && .rxOomHasParquet(.m))
    return(as.data.frame(as_arrow_table.rxSolveOom(x)))
  do.call(rbind, lapply(.m$chunks, readRDS))
}

as_tibble.rxSolveOom <- function(x, ...) {
  if (.rxOomHasArrow() && .rxOomHasParquet(attr(x, "manifest")))
    return(tibble::as_tibble(as_arrow_table.rxSolveOom(x)))
  tibble::as_tibble(as.data.frame(x))
}

#' @export
as.data.table.rxSolveOom <- function(x, keep.rownames = FALSE, ...) {
  if (.rxOomHasArrow() && .rxOomHasParquet(attr(x, "manifest")))
    return(data.table::as.data.table(as.data.frame(as_arrow_table.rxSolveOom(x))))
  data.table::as.data.table(as.data.frame(x), keep.rownames = keep.rownames)
}

#' First rows of an out-of-memory solved object
#'
#' Reads only the first \code{n} rows from the parquet (or rds) chunks,
#' preferring DuckDB (which pushes the row limit into the parquet reader) and
#' falling back to arrow / rds.
#'
#' @param x An \code{rxSolveOom} object.
#' @param n Number of rows to return.
#' @param ... Ignored.
#' @return A \code{data.frame} with the first \code{n} rows.
#' @keywords internal
#' @export
head.rxSolveOom <- function(x, n = 6L, ...) {
  .m <- attr(x, "manifest")
  .pq <- .rxOomParquetFiles(.m$chunks)
  if (length(.pq) > 0L && .rxOomHasDuckdb()) {
    return(.rxOomDuckQuery(.pq, sprintf("SELECT * FROM {tbl} LIMIT %d", as.integer(n))))
  }
  if (length(.pq) > 0L && .rxOomHasArrow()) {
    # Walk chunks until we have n rows; only the first chunk(s) are read.
    .acc <- vector("list", 0L)
    .got <- 0L
    for (.f in .pq) {
      .d <- as.data.frame(arrow::read_parquet(.f))
      .acc[[length(.acc) + 1L]] <- utils::head(.d, n - .got)
      .got <- .got + nrow(.acc[[length(.acc)]])
      if (.got >= n) break
    }
    return(do.call(rbind, .acc))
  }
  .acc <- vector("list", 0L)
  .got <- 0L
  for (.f in .m$chunks) {
    .d <- readRDS(.f)
    .acc[[length(.acc) + 1L]] <- utils::head(.d, n - .got)
    .got <- .got + nrow(.acc[[length(.acc)]])
    if (.got >= n) break
  }
  do.call(rbind, .acc)
}

#' @export
`$.rxSolveOom` <- function(x, name) {
  .m <- attr(x, "manifest")
  if (name %in% c("params", "par", "pars", "param")) {
    return(.rxOomParams(.m))
  }
  if (name %in% c("inits", "init")) {
    return(.rxOomInits(.m))
  }
  .pq <- .rxOomParquetFiles(.m$chunks)
  if (length(.pq) > 0L && .rxOomHasDuckdb()) {
    .r <- .rxOomDuckQuery(.pq, sprintf('SELECT "%s" FROM {tbl}', gsub('"', '""', name)))
    return(.r[[1L]])
  }
  if (length(.pq) > 0L && .rxOomHasArrow()) {
    .cols <- lapply(.pq, function(.f)
      arrow::read_parquet(.f, col_select = name)[[1L]])
    return(unlist(.cols, use.names = FALSE))
  }
  unlist(lapply(.m$chunks, function(.f) readRDS(.f)[[name]]), use.names = FALSE)
}

# Column count from the first chunk's schema (cheap: header-only for parquet).
.rxOomNcol <- function(manifest) {
  .pq <- .rxOomParquetFiles(manifest$chunks)
  if (length(.pq) > 0L && .rxOomHasDuckdb()) {
    return(nrow(.rxOomDuckQuery(.pq[1L], "DESCRIBE SELECT * FROM {tbl}")))
  }
  if (length(.pq) > 0L && .rxOomHasArrow()) {
    return(length(arrow::open_dataset(.pq[1L])$schema$names))
  }
  if (length(manifest$chunks) > 0L) return(ncol(readRDS(manifest$chunks[1L])))
  NA_integer_
}

# nrow() is not an S3 generic, so a `nrow.rxSolveOom` method is never
# dispatched (and exporting it as a plain function trips an "undocumented
# code object" check).  base::nrow(x) is dim(x)[1L], and dim() *is* generic,
# so dim.rxSolveOom() below already makes nrow() return the right value.
#' @export
dim.rxSolveOom <- function(x) {
  .m <- attr(x, "manifest")
  c(sum(.m$nrows), .rxOomNcol(.m))
}

# -- User-facing convenience wrapper ------------------------------------------

#' Solve an ODE model in memory-safe chunks
#'
#' Splits subjects into chunks sized to fit in available RAM, solves each chunk,
#' and writes output to parquet (or rds) files. Returns an \code{rxSolveOom}
#' object that lazily reads chunks on demand.
#'
#' The storage/query engine used by the resulting \code{rxSolveOom} object is
#' controlled by the \code{rxode2.oom.backend} option:
#' \describe{
#'   \item{\code{"auto"}}{(default) DuckDB if installed, else arrow, else rds.}
#'   \item{\code{"duckdb"}}{lazy DuckDB SQL queries over arrow-written parquet.}
#'   \item{\code{"arrow"}}{arrow parquet reads/writes, no DuckDB.}
#'   \item{\code{"rds"}}{plain rds files, no arrow or DuckDB.}
#' }
#' A requested engine that is not installed silently degrades
#' (duckdb \eqn{\to} arrow \eqn{\to} rds).
#'
#' @param object rxode2 model
#' @param params model parameters
#' @param events event table or \code{rxEtFile}
#' @param inits initial conditions
#' @param ... additional arguments passed to \code{rxControl()}
#' @param chunkSize number of subjects per chunk (auto-computed from free RAM if omitted)
#' @param seed random seed (sets before solving if not NULL)
#' @param parallel number of mirai daemons for parallel chunk solving (0 = serial)
#' @return An \code{rxSolveOom} object
#' @export
rxSolveChunked <- function(object, params = NULL, events = NULL, inits = NULL, ...,
                            chunkSize, seed = NULL, parallel = 0L) {
  # Normalize params/events to match rxSolve convention
  if ((is.rxEt(params) || rxIs(params, "rx.event") || inherits(params, "rxEtFile")) &&
        !is.rxEt(events)) {
    .tmp   <- events
    events <- params
    params <- .tmp
  }
  if (!missing(chunkSize) && !is.null(chunkSize)) {
    .chunkSize <- as.integer(chunkSize)
  } else {
    .chunkSize <- NULL
  }
  if (!is.null(seed)) rxSetSeed(seed)
  .ctl <- rxControl(...,
    file      = tempfile("rxChunk"),
    chunkSize = .chunkSize,
    parallel  = as.integer(parallel))
  .rxSolveOom(object, params = params, events = events, inits = inits,
              .ctl = .ctl, .envir = parent.frame())
}

test_that("rxSolveChunked is reproducible with the same seed", {
  rxTest({
    mod <- rxode2({
      k <- exp(lk + eta.k)
      d/dt(A) <- -k * A
    })
    et_pop <- et(seq(0, 24, by = 1)) |> et(amt = 100) |> et(id = 1:20)

    chnk1 <- rxSolveChunked(mod, c(lk = log(0.1)), et_pop, seed = 42,
                             omega = lotri::lotri(eta.k ~ 0.09),
                             chunkSize = 5)
    chnk2 <- rxSolveChunked(mod, c(lk = log(0.1)), et_pop, seed = 42,
                             omega = lotri::lotri(eta.k ~ 0.09),
                             chunkSize = 5)
    expect_s3_class(chnk1, "rxSolveOom")
    expect_equal(nrow(chnk1), 20L * 25L)

    got1 <- as.data.frame(chnk1)
    got2 <- as.data.frame(chnk2)
    expect_equal(got1$A, got2$A, tolerance = 1e-6)
  })
})

test_that("rxSolveChunked matches rxSolve with same rxSetSeed", {
  rxTest({
    mod <- rxode2({
      k <- exp(lk + eta.k)
      d/dt(A) <- -k * A
    })
    et_pop <- et(seq(0, 24, by = 1)) |> et(amt = 100) |> et(id = 1:20)

    rxSetSeed(42)
    full <- rxSolve(mod, c(lk = log(0.1)), et_pop,
                    omega = lotri::lotri(eta.k ~ 0.09))
    chnk <- rxSolveChunked(mod, c(lk = log(0.1)), et_pop, seed = 42,
                            omega = lotri::lotri(eta.k ~ 0.09),
                            chunkSize = 5)

    full_df <- as.data.frame(full)
    chnk_df <- as.data.frame(chnk)
    full_df <- full_df[order(full_df$id, full_df$time), ]
    chnk_df <- chnk_df[order(chnk_df$id, chnk_df$time), ]
    expect_equal(full_df$A, chnk_df$A, tolerance = 1e-6)
  })
})

test_that("rxSolve with file returns rxSolveOom and manifest is written", {
  rxTest({
    mod <- rxode2({
      d/dt(A) <- -k * A
    })
    et_pop <- et(seq(0, 24, by = 1)) |> et(amt = 100) |> et(id = 1:10)

    .prefix <- tempfile("oomsim")
    out <- rxSolve(mod, c(k = 0.1), et_pop,
                   file = .prefix, chunkSize = 5L)
    expect_s3_class(out, "rxSolveOom")
    expect_true(file.exists(paste0(.prefix, "_manifest.rds")))

    print_out <- capture.output(print(out))
    expect_true(any(grepl("rxSolveOom", print_out)))
  })
})

test_that("$.rxSolveOom extracts a column across all chunks", {
  rxTest({
    mod <- rxode2({
      d/dt(A) <- -k * A
    })
    et_pop <- et(seq(0, 24, by = 1)) |> et(amt = 100) |> et(id = 1:10)

    chnk <- rxSolveChunked(mod, c(k = 0.1), et_pop, seed = 99, chunkSize = 3)

    .A_chnk <- chnk$A
    expect_length(.A_chnk, nrow(as.data.frame(chnk)))
  })
})

test_that("as_arrow_table.rxSolveOom returns an Arrow Table", {
  rxTest({
    skip_if_not_installed("arrow")
    mod <- rxode2({
      d/dt(A) <- -k * A
    })
    et_pop <- et(seq(0, 24, by = 1)) |> et(amt = 100) |> et(id = 1:10)
    chnk <- rxSolveChunked(mod, c(k = 0.1), et_pop, chunkSize = 3)

    tbl <- arrow::as_arrow_table(chnk)
    expect_true(inherits(tbl, "Table"))
    expect_equal(as.integer(tbl$num_rows), nrow(chnk))
    expect_true("A" %in% names(tbl))
  })
})

test_that("as.arrow.rxSolveOom returns a lazy Arrow Dataset", {
  rxTest({
    skip_if_not_installed("arrow")
    skip_if_not_installed("dplyr")
    mod <- rxode2({
      d/dt(A) <- -k * A
    })
    et_pop <- et(seq(0, 24, by = 1)) |> et(amt = 100) |> et(id = 1:10)
    chnk <- rxSolveChunked(mod, c(k = 0.1), et_pop, chunkSize = 3)

    ds <- as.arrow(chnk)
    expect_true(inherits(ds, "Dataset"))

    collected <- dplyr::collect(ds)
    full_df   <- as.data.frame(chnk)
    collected <- collected[order(collected$id, collected$time), ]
    full_df   <- full_df[order(full_df$id, full_df$time), ]
    expect_equal(collected$A, full_df$A, tolerance = 1e-6)
  })
})

test_that("as.data.frame.rxSolveOom is consistent with and without arrow", {
  rxTest({
    mod <- rxode2({
      d/dt(A) <- -k * A
    })
    et_pop <- et(seq(0, 24, by = 1)) |> et(amt = 100) |> et(id = 1:10)
    chnk <- rxSolveChunked(mod, c(k = 0.1), et_pop, chunkSize = 3)

    df <- as.data.frame(chnk)
    expect_s3_class(df, "data.frame")
    expect_equal(nrow(df), 10L * 25L)
    expect_true("A" %in% names(df))
  })
})

test_that("as_tibble.rxSolveOom returns a tibble", {
  rxTest({
    skip_if_not_installed("tibble")
    mod <- rxode2({
      d/dt(A) <- -k * A
    })
    et_pop <- et(seq(0, 24, by = 1)) |> et(amt = 100) |> et(id = 1:10)
    chnk <- rxSolveChunked(mod, c(k = 0.1), et_pop, chunkSize = 3)

    tib <- tibble::as_tibble(chnk)
    expect_true(inherits(tib, "tbl_df"))
    expect_equal(nrow(tib), 10L * 25L)
  })
})

test_that("as.data.table.rxSolveOom returns a data.table", {
  rxTest({
    skip_if_not_installed("data.table")
    mod <- rxode2({
      d/dt(A) <- -k * A
    })
    et_pop <- et(seq(0, 24, by = 1)) |> et(amt = 100) |> et(id = 1:10)
    chnk <- rxSolveChunked(mod, c(k = 0.1), et_pop, chunkSize = 3)

    dt <- data.table::as.data.table(chnk)
    expect_true(data.table::is.data.table(dt))
    expect_equal(nrow(dt), 10L * 25L)
  })
})

test_that("print.rxSolveOom shows [Arrow-backed] when arrow is installed", {
  rxTest({
    skip_if_not_installed("arrow")
    mod <- rxode2({
      d/dt(A) <- -k * A
    })
    et_pop <- et(seq(0, 24, by = 1)) |> et(amt = 100) |> et(id = 1:10)
    chnk <- rxSolveChunked(mod, c(k = 0.1), et_pop, chunkSize = 3)

    out <- capture.output(print(chnk))
    expect_true(any(grepl("Arrow-backed", out)))
  })
})

test_that("rxEventTableFile works with rds format", {
  rxTest({
    mod <- rxode2({
      d/dt(A) <- -k * A
    })
    et <- et(seq(0, 12, by = 1)) |> et(amt = 100)
    .etDf <- as.data.frame(et)

    .tf <- tempfile(fileext = ".rds")
    saveRDS(.etDf, .tf)

    .etf <- rxEventTableFile(.tf, format = "rds")
    expect_s3_class(.etf, "rxEtFile")
  })
})

test_that("rxEtFile is recognized as events when passed positionally", {
  rxTest({
    mod <- rxode2({
      d/dt(A) <- -k * A
    })
    et <- et(seq(0, 12, by = 1)) |> et(amt = 100) |> et(id = 1:4)

    .tf <- tempfile(fileext = ".rds")
    saveRDS(as.data.frame(et), .tf)
    .etf <- rxEventTableFile(.tf, format = "rds")

    # positional (params slot) must swap to events like rxEt/data.frame do
    .res <- rxSolve(mod, .etf, params = c(k = 0.1),
                    file = tempfile("rxPos"), chunkSize = 2)
    expect_s3_class(.res, "rxSolveOom")
    .df <- as.data.frame(.res)
    expect_equal(sort(unique(.df$id)), 1:4)

    # ui function models re-parse from a head preview of the file
    fun <- function() {
      ini({
        k <- 0.1
      })
      model({
        d/dt(A) <- -k * A
      })
    }
    .res2 <- rxSolve(fun, .etf, file = tempfile("rxPosUi"), chunkSize = 2)
    expect_s3_class(.res2, "rxSolveOom")
  })
})

test_that("rxSolve(parallel=) mirai path matches the serial chunked solve", {
  rxTest({
    skip_if_not_installed("mirai")
    skip_if_not_installed("arrow")
    mod <- rxode2({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v  <- exp(tv)
      d/dt(depot)  <- -ka * depot
      d/dt(center) <- ka * depot - cl / v * center
      cp           <- center / v
    })
    et_pop <- et(amt = 100) |> et(seq(0, 24, by = 4)) |> et(id = 1:6)
    pars   <- c(tka = log(0.5), tcl = log(4), tv = log(70))
    omega  <- lotri::lotri(eta.ka ~ 0.09, eta.cl ~ 0.04)

    rxSetSeed(42)
    ser <- rxSolve(mod, pars, et_pop, omega = omega,
                   file = tempfile("rxSer"), chunkSize = 2)
    rxSetSeed(42)
    par <- rxSolve(mod, pars, et_pop, omega = omega,
                   file = tempfile("rxPar"), chunkSize = 2, parallel = 2)

    expect_s3_class(par, "rxSolveOom")
    ser_df <- as.data.frame(ser)
    par_df <- as.data.frame(par)
    ser_df <- ser_df[order(ser_df$id, ser_df$time), ]
    par_df <- par_df[order(par_df$id, par_df$time), ]
    expect_equal(nrow(par_df), nrow(ser_df))
    expect_equal(sort(unique(par_df$id)), 1:6)
    # parallel chunking must not change the eta draws -> identical solve
    expect_equal(par_df$cp, ser_df$cp, tolerance = 1e-8)
  })
})

test_that("rxode2.oom.backend option is forwarded to the mirai workers", {
  rxTest({
    skip_if_not_installed("mirai")
    withr::local_options(rxode2.oom.backend = "rds")
    mod <- rxode2({
      cl <- exp(lcl + eta.cl)
      v  <- exp(lv)
      d/dt(central) <- -cl / v * central
      cp <- central / v
    })
    et_pop <- et(amt = 100) |> et(seq(0, 12, 4)) |> et(id = 1:4)

    par <- rxSolve(mod, c(lcl = 1, lv = 3.45), et_pop,
                   omega = lotri::lotri(eta.cl ~ 0.04),
                   file = tempfile("rxParRds"), chunkSize = 2, parallel = 2)

    expect_s3_class(par, "rxSolveOom")
    m <- attr(par, "manifest")
    # Daemons honored the "rds" backend instead of writing parquet.
    expect_true(all(grepl("\\.rds$", m$chunks)))
    expect_true(all(grepl("\\.rds$", m$paramChunks)))
    expect_equal(nrow(par), 4L * 4L)
    expect_equal(nrow(par$params), 4L)
  })
})

test_that("rxSolveOom persists params and inits like an rxSolve object", {
  rxTest({
    mod <- rxode2({
      ka <- exp(lka + eta.ka)
      cl <- exp(lcl + eta.cl)
      v  <- exp(lv)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - cl / v * central
      cp <- central / v
    })
    pars   <- c(lka = 0.45, lcl = 1, lv = 3.45)
    et_pop <- et(amt = 100) |> et(seq(0, 24, 4)) |> et(id = 1:6)
    omega  <- lotri::lotri(eta.ka ~ 0.09, eta.cl ~ 0.04)

    chnk <- rxSolveChunked(mod, pars, et_pop, seed = 7,
                            omega = omega, chunkSize = 2)

    # res$params: one row per subject, NOT NULL
    .pars <- chnk$params
    expect_s3_class(.pars, "data.frame")
    expect_equal(nrow(.pars), 6L)
    expect_true(all(c("id", "eta.ka", "eta.cl") %in% names(.pars)))

    # res$inits: named numeric vector matching the model states
    .inits <- chnk$inits
    expect_true(is.numeric(.inits))
    expect_setequal(names(.inits), c("depot", "central"))
  })
})

test_that("head.rxSolveOom reads only the first rows", {
  rxTest({
    mod <- rxode2({
      d/dt(A) <- -k * A
    })
    et_pop <- et(seq(0, 24, by = 1)) |> et(amt = 100) |> et(id = 1:10)
    chnk <- rxSolveChunked(mod, c(k = 0.1), et_pop, chunkSize = 3)

    h <- head(chnk, 4)
    expect_equal(nrow(h), 4L)
    expect_true(all(c("id", "time", "A") %in% names(h)))
    # ncol via dim() resolves from the schema (not NA)
    expect_equal(ncol(chnk), ncol(as.data.frame(chnk)))
  })
})

test_that("rxSolveOom print mirrors rxSolve output with a chunk footer", {
  rxTest({
    mod <- rxode2({
      cl <- exp(lcl + eta.cl)
      v  <- exp(lv)
      d/dt(central) <- -cl / v * central
      cp <- central / v
    })
    et_pop <- et(amt = 100) |> et(seq(0, 24, 4)) |> et(id = 1:6)
    chnk <- rxSolveChunked(mod, c(lcl = 1, lv = 3.45), et_pop, seed = 3,
                            omega = lotri::lotri(eta.cl ~ 0.04), chunkSize = 2)

    out <- capture.output(print(chnk))
    expect_true(any(grepl("Solved rxode2 object", out)))
    expect_true(any(grepl("Parameters", out)))
    expect_true(any(grepl("Initial Conditions", out)))
    expect_true(any(grepl("First part of data", out)))
    expect_true(any(grepl("rxSolveOom", out)))
  })
})

# The `rxode2.oom.backend` option pins the storage/query engine so each of the
# three interfaces (plain rds, arrow parquet, duckdb over parquet) can be
# exercised deterministically regardless of what is installed by default.
for (.oomBackend in c("rds", "arrow", "duckdb")) {
  local({
    .backend <- .oomBackend
    test_that(sprintf("rxSolveOom round-trips through the '%s' backend", .backend), {
      rxTest({
        if (.backend %in% c("arrow", "duckdb")) skip_if_not_installed("arrow")
        if (.backend == "duckdb") {
          skip_if_not_installed("duckdb")
          skip_if_not_installed("DBI")
        }
        withr::local_options(rxode2.oom.backend = .backend)

        mod <- rxode2({
          cl <- exp(lcl + eta.cl)
          v  <- exp(lv)
          d/dt(central) <- -cl / v * central
          cp <- central / v
        })
        et_pop <- et(amt = 100) |> et(seq(0, 12, 4)) |> et(id = 1:4)
        chnk <- rxSolveChunked(mod, c(lcl = 1, lv = 3.45), et_pop, seed = 5,
                                omega = lotri::lotri(eta.cl ~ 0.04), chunkSize = 2)

        # chunks are written in the format the active backend dictates
        m <- attr(chnk, "manifest")
        .ext <- if (.backend == "rds") "\\.rds$" else "\\.parquet$"
        expect_true(all(grepl(.ext, m$chunks)))
        expect_true(all(grepl(.ext, m$paramChunks)))

        # data-frame-like behavior is identical across all three backends
        .df <- as.data.frame(chnk)
        expect_equal(nrow(chnk), nrow(.df))
        expect_equal(ncol(chnk), ncol(.df))
        expect_equal(nrow(head(chnk, 3)), 3L)
        expect_length(chnk$cp, nrow(chnk))
        expect_equal(chnk$cp, .df$cp, tolerance = 1e-8)

        # params / inits round-trip
        expect_equal(nrow(chnk$params), 4L)
        expect_setequal(names(chnk$inits), "central")

        # the print footer reflects the active backend
        out <- capture.output(print(chnk))
        if (.backend == "rds") {
          expect_false(any(grepl("Arrow|DuckDB", out)))
        } else if (.backend == "arrow") {
          expect_true(any(grepl("\\[Arrow-backed\\]", out)))
        } else {
          expect_true(any(grepl("\\[DuckDB/Arrow-backed\\]", out)))
        }
      })
    })
  })
}

# A possibly-out-of-memory rxSolveOom object must be queryable with dplyr
# verbs *lazily* -- the filtering/aggregation is pushed down to the
# on-disk parquet chunks and only the reduced result is materialized, so
# the full data set is never loaded into memory.
test_that("dplyr filter/select on rxSolveOom stays lazy (no full materialization)", {
  rxTest({
    skip_if_not_installed("arrow")
    skip_if_not_installed("dplyr")
    mod <- rxode2({
      cl <- exp(lcl)
      v  <- exp(lv)
      d/dt(central) <- -cl / v * central
      cp <- central / v
    })
    et_pop <- et(amt = 100) |> et(seq(0, 12, 4)) |> et(id = 1:6)
    chnk <- rxSolveChunked(mod, c(lcl = 1, lv = 3.45), et_pop, chunkSize = 2)

    ds <- as.arrow(chnk)
    # The data stays on disk: a file-backed Dataset, not an in-memory Table.
    expect_true(inherits(ds, "Dataset"))
    expect_false(inherits(ds, "Table"))

    # Building the query does NOT execute it -- it is a deferred arrow query,
    # not a materialized data.frame.
    q <- ds |>
      dplyr::filter(id == 1L) |>
      dplyr::select(id, time, cp)
    expect_s3_class(q, "arrow_dplyr_query")
    expect_false(is.data.frame(q))

    # Only the matching rows are pulled into memory on collect().
    got <- dplyr::collect(q)
    expect_true(is.data.frame(got))
    expect_lt(nrow(got), nrow(chnk))
    expect_setequal(unique(got$id), 1L)
    expect_setequal(names(got), c("id", "time", "cp"))

    # ... and they match the reference subset.
    ref <- as.data.frame(chnk)
    ref <- ref[ref$id == 1L, c("id", "time", "cp")]
    got <- got[order(got$time), ]
    ref <- ref[order(ref$time), ]
    expect_equal(got$cp, ref$cp, tolerance = 1e-8)
  })
})

test_that("dplyr group_by/summarise on rxSolveOom is pushed to the arrow backend", {
  rxTest({
    skip_if_not_installed("arrow")
    skip_if_not_installed("dplyr")
    mod <- rxode2({
      cl <- exp(lcl + eta.cl)
      v  <- exp(lv)
      d/dt(central) <- -cl / v * central
      cp <- central / v
    })
    et_pop <- et(amt = 100) |> et(seq(0, 12, 4)) |> et(id = 1:6)
    chnk <- rxSolveChunked(mod, c(lcl = 1, lv = 3.45), et_pop, seed = 11,
                            omega = lotri::lotri(eta.cl ~ 0.04), chunkSize = 2)

    agg <- as.arrow(chnk) |>
      dplyr::group_by(id) |>
      dplyr::summarise(mcp = mean(cp), n = dplyr::n())
    # aggregation is deferred until collect()
    expect_s3_class(agg, "arrow_dplyr_query")

    out <- dplyr::collect(agg)
    out <- out[order(out$id), ]
    expect_equal(nrow(out), 6L)              # one row per subject, not per record
    expect_true(all(out$n == 4L))            # 4 time points each

    ref <- as.data.frame(chnk)
    ref <- tapply(ref$cp, ref$id, mean)
    expect_equal(out$mcp, as.numeric(ref[as.character(out$id)]), tolerance = 1e-8)
  })
})

test_that("rxSolveOom supports lazy dplyr queries through a DuckDB connection", {
  rxTest({
    skip_if_not_installed("arrow")
    skip_if_not_installed("duckdb")
    skip_if_not_installed("dbplyr")
    skip_if_not_installed("dplyr")
    mod <- rxode2({
      cl <- exp(lcl)
      v  <- exp(lv)
      d/dt(central) <- -cl / v * central
      cp <- central / v
    })
    et_pop <- et(amt = 100) |> et(seq(0, 12, 4)) |> et(id = 1:6)
    chnk <- rxSolveChunked(mod, c(lcl = 1, lv = 3.45), et_pop, chunkSize = 2)

    # Register the on-disk parquet chunks as a lazy DuckDB-backed table.
    tbl <- arrow::to_duckdb(as.arrow(chnk))
    expect_s3_class(tbl, "tbl_lazy")
    expect_false(is.data.frame(tbl))

    q <- tbl |> dplyr::filter(time == 0)
    expect_s3_class(q, "tbl_lazy")

    got <- dplyr::collect(q)
    expect_lt(nrow(got), nrow(chnk))
    expect_true(all(got$time == 0))
    expect_equal(nrow(got), 6L)              # one t=0 row per subject
  })
})

#' Switch Activation Function
#'
#' The switch activation function is defined as:
#'
#' \deqn{f(x) = x \cdot \text{sigmoid}(x)}
#'
#' @family Activation Functions
#'
#' @param x A numeric vector. All elements must be finite and
#'   non-missing.
#'
#' @return A numeric vector where the ReLU function has been applied
#'   to each element of `x`.
#'
#' @author Matthew Fidler
#'
#' @export
#'
#' @examples
#'
#' Swish(c(-1, 0, 1, 2))
#'
#' \donttest{
#' # Can also be used in rxode2:
#' x <- rxode2({
#'    r<- Swish(time)
#' })
#'
#' e <- et(c(-1, 0, 1, 2))
#'
#' rxSolve(x, e)
#' }
#'
Swish <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 6L)
}

#' Derivative of the Swish Activation Function
#'
#' @param x A numeric vector. All elements must be finite and
#'  non-missing.
#' @return A numeric vector where the derivative of the SELU function
#'  has been applied to each element of `x`.
#' @author Matthew Fidler
#' @family Activation Functions
#' @export
#' @examples
#' dSwish(c(-1, 0, 1, 2))
#'
#' \donttest{
#' # Can also be used in rxode2:
#' x <- rxode2({
#'   r <- dSwish(time)
#' })
#' e <- et(c(-1, 0, 1, 2))
#' rxSolve(x, e)
#' }
dSwish <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 18L)
}

#' @title Convert event data to trial duration data
#' A helper function to create a custom event table.
#' The observation time will start from the first event time (baseline) and end at trial duration.
#' The interval is the spacing between each observation.
#' @param ev event data
#' @param trialEnd extend trial duration. Must be same time unit as event data
#' @param interval observation interval. Must be same time unit as event data
#' @param writeDir if not NULL, write the output to a csv file
#' @importFrom stats ave
#' @importFrom utils write.csv
#' @examples
#'  # Create event table with unique time for each ID
#'  ev = et(data.frame(id = rep(1:10, 3),  time = runif(min = 10, max = 20, n = 30)))
#'
#'  # select the duration and spacing interval (assuming time is in years)
#'  toTrialDuration(ev, trialEnd = 1.5, interval = 0.2)
#'
#' @author Omar Elashkar
#' @export
toTrialDuration <- function(ev, trialEnd, interval, writeDir = NULL) {
  checkmate::assertClass(ev, "rxEt")
  reg <- as.data.frame(ev[,c("id", "time")])
  reg <- reg[reg$time == ave(reg$time, reg$id, FUN=min), ]

  reg <- Map(function(id, time) {
    data.frame(id = id, time = seq(time, time+trialEnd, interval))
  }, id = reg$id, time = reg$time
  )
  reg <- do.call(rbind, reg) |>
    merge(unique(ev[, names(ev) != "time", drop = FALSE]), by = "id", all.x = T)
  reg <- et(reg)
  if(is.character(writeDir)) {
    write.csv(reg, writeDir, row.names = F)
    }

  reg
}

.addExpr <- function(.ret) {
  .new <- .rxOptEnv$.rep[[.ret]]
  if (!is.null(.new)) {
    if (length(.rxOptEnv$.exclude) != 1) .rxOptEnv$.exclude <- ""
    if (.new == .rxOptEnv$.exclude) {
      return(.ret)
    } else {
      .rxOptEnv$.new <- c(.rxOptEnv$.new, .new)
      return(.new)
    }
  } else {
    if (is.null(.rxOptEnv$.list[[.ret]])) {
      .rxOptEnv$.list[[.ret]] <- 1L
    } else {
      .rxOptEnv$.list[[.ret]] <- .rxOptEnv$.list[[.ret]] + 1L
    }
  }
  return(.ret)
}

.rxOptFn <- function(fn) {
  force(fn)
  function(...) {
    .ret <- paste0(fn, "(", paste(unlist(list(...)), collapse = ", "), ")")
    return(.addExpr(.ret))
  }
}
.rxOptBin <- function(sep) {
  force(sep)
  function(e1, e2) {
    .add <- TRUE
    if (missing(e2)) {
      if (sep == "+") {
        .ret <- paste0(e1)
      } else {
        .ret <- paste0(gsub(" ", "", sep), e1)
      }
      if (regexpr(rex::rex(start, any_spaces, regNum, any_spaces, end),
        .ret,
        perl = TRUE
      ) != -1) {
        .add <- FALSE
      }
    } else {
      if (sep == "^" && isTRUE(checkmate::checkIntegerish(suppressWarnings(as.numeric(e2)),
        lower = 2, any.missing = FALSE
      ))) {
        .ret <- paste0("(", paste(rep(paste0("(", e1, ")"), as.numeric(e2)), collapse = "*"), ")")
      } else {
        if ((regexpr(rex::rex(start, any_spaces, regNum, any_spaces, end),
          paste0(e1),
          perl = TRUE
        ) != -1) &&
          (regexpr(rex::rex(start, any_spaces, regNum, any_spaces, end),
            paste0(e2),
            perl = TRUE
          ) != -1)) {
          .add <- FALSE
        }
        .ret <- paste0(e1, sep, e2)
      }
    }
    if (.add) {
      return(.addExpr(.ret))
    } else {
      return(.ret)
    }
  }
}

.rxOptEnv <- new.env(parent = emptyenv())
.rxOptEnv[["^"]] <- .rxOptBin("^")
.rxOptEnv[["**"]] <- .rxOptBin("^")

.rxOptEnv[["*"]] <- .rxOptBin("*")
.rxOptEnv[["/"]] <- .rxOptBin("/")
.rxOptEnv[["+"]] <- .rxOptBin("+")
.rxOptEnv[["-"]] <- .rxOptBin("-")
.rxOptEnv[["&&"]] <- .rxOptBin("&&")
.rxOptEnv[["||"]] <- .rxOptBin("||")
.rxOptEnv[["|"]] <- .rxOptBin("|")
.rxOptEnv[["&"]] <- .rxOptBin("&")
.rxOptEnv[["<="]] <- .rxOptBin("<=")
.rxOptEnv[[">="]] <- .rxOptBin(">=")
.rxOptEnv[["<"]] <- .rxOptBin("<")
.rxOptEnv[[">"]] <- .rxOptBin(">")
.rxOptEnv[["=="]] <- .rxOptBin("==")
.rxOptEnv[["!="]] <- .rxOptBin("!=")
.rxOptEnv[["["]] <- function(name, val) {
  .n <- toupper(name)
  .err <- gettext("rxode2 only supports THETA[#] and ETA[#] numbers")
  if (any(.n == c("THETA", "ETA")) && is.numeric(val)) {
    if (round(val) == val && val > 0) {
      return(sprintf("%s[%s]", .n, val))
    } else {
      stop(.err, call. = FALSE)
    }
  } else {
    stop(.err, call. = FALSE)
  }
}
.rxOptEnv[["{"]] <- function(...) {
  return(sprintf("{\n%s\n}", paste(unlist(list(...)), collapse = "\n")))
}
.rxOptEnv[["["]] <- function(name, val) {
  .n <- toupper(name)
  .err <- gettext("rxode2 only supports THETA[#] and ETA[#] numbers")
  if (any(.n == c("THETA", "ETA")) && is.numeric(val)) {
    if (round(val) == val && val > 0) {
      return(sprintf("%s[%s]", .n, val))
    } else {
      stop(err, call. = FALSE)
    }
  } else {
    stop(err, call. = FALSE)
  }
}

.rxOptEnv[["("]] <- unaryOp("(", ")")

.rxOptEnv$.list <- list()
.rxOptEnv$.rep <- list()
.rxOptEnv$.exclude <- ""
.rxOptEnv$.new <- NULL
.rxOptEnv$.added <- NULL
.rxOptGetEnv <- function(expr) {
  ## Known functions
  .calls <- allCalls(expr)
  .callList <- setNames(lapply(.calls, .rxOptFn), .calls)
  .callEnv <- list2env(.callList)
  .currEnv <- cloneEnv(.rxOptEnv, .callEnv)
  .names <- allNames(expr)
  .n1 <- .names
  .n2 <- .names
  .symbolList <- setNames(as.list(.n2), .n1)
  .symbolEnv <- list2env(.symbolList, parent = .currEnv)
  return(.symbolEnv)
}

.rxOptExpr <- function(x) {
  x <- .convStr(x)
  .ret <- eval(x, .rxOptGetEnv(x))
  .ret <- eval(parse(text = paste0("quote(", .ret, ")")))
  return(..rxOpt(.ret))
}

..rxOptLhs <- function(x) {
  if (is.atomic(x) || is.name(x)) {
    return(as.character(x))
  } else if (is.call(x)) {
    if (identical(x[[1]], quote(`/`))) {
      return(paste0(..rxOptLhs(x[[2]]), "/", ..rxOptLhs(x[[3]])))
    } else if (identical(x[[1]], quote(`(`))) {
      return(paste0("(", ..rxOptLhs(x[[2]]), ")"))
    } else if (identical(x[[1]], quote(`dt`))) {
      return(paste0("dt(", ..rxOptLhs(x[[2]]), ")"))
    } else if (identical(x[[1]], quote(`f`))) {
      return(paste0("f(", ..rxOptLhs(x[[2]]), ")"))
    } else if (identical(x[[1]], quote(`F`))) {
      return(paste0("F(", ..rxOptLhs(x[[2]]), ")"))
    } else if (identical(x[[1]], quote(`rate`))) {
      return(paste0("rate(", ..rxOptLhs(x[[2]]), ")"))
    } else if (identical(x[[1]], quote(`alag`))) {
      return(paste0("alag(", ..rxOptLhs(x[[2]]), ")"))
    } else if (identical(x[[1]], quote(`lag`))) {
      return(paste0("alag(", ..rxOptLhs(x[[2]]), ")"))
    } else if (identical(x[[1]], quote(`dur`))) {
      return(paste0("dur(", ..rxOptLhs(x[[2]]), ")"))
    } else if (identical(x[[1]], quote(`past`))) {
      return(paste0("past(", ..rxOptLhs(x[[2]]), ",", ..rxOptLhs(x[[3]]), ")"))
    } else if (identical(x[[1]], quote(`dy`))) {
      return(paste0("dy(", ..rxOptLhs(x[[2]]), ")"))
    } else if (identical(x[[1]], quote(`df`))) {
      return(paste0("df(", ..rxOptLhs(x[[2]]), ")"))
    } else if (identical(x[[2]], 0)) {
      return(paste0(as.character(x[[1]]), "(0)"))
    } else {
      print(x)
      stop("unsupported lhs in optimize expression")
    }
  }
}

..rxOpt <- function(x, progress = FALSE) {
  if (is.atomic(x)) {
    return(as.character(x))
  } else if (is.name(x)) {
    return(.rxRepRxQ(as.character(x)))
  } else if (is.call(x)) {
    .x2 <- x[-1]
    if (identical(x[[1]], quote(`{`))) {
      if (progress) {
        rxProgress(length(.x2))
        on.exit({
          rxProgressAbort("stopped optimizing duplicate expressions")
        })
        .ret <- unlist(lapply(.x2, function(x) {
          rxTick()
          ..rxOpt(x)
        }))
        rxProgressStop()
        return(.ret)
      } else {
        return(unlist(lapply(.x2, ..rxOpt)))
      }
    } else if (identical(x[[1]], quote(`(`))) {
      .x2 <- x[[2]]
      .test <- TRUE
      while (.test) {
        if (length(.x2) == 2) {
          if (identical(.x2[[1]], quote(`(`))) {
            .x2 <- .x2[[2]]
          } else {
            .test <- FALSE
          }
        } else {
          .test <- FALSE
        }
      }
      return(paste0("(", ..rxOpt(.x2), ")"))
    } else if (identical(x[[1]], quote(`*`)) ||
      identical(x[[1]], quote(`^`)) ||
      identical(x[[1]], quote(`+`)) ||
      identical(x[[1]], quote(`-`)) ||
      identical(x[[1]], quote(`/`)) ||
      identical(x[[1]], quote(`==`)) ||
      identical(x[[1]], quote(`>=`)) ||
      identical(x[[1]], quote(`<=`)) ||
      identical(x[[1]], quote(`>`)) ||
      identical(x[[1]], quote(`<`)) ||
      identical(x[[1]], quote(`!=`)) ||
      identical(x[[1]], quote(`&&`)) ||
      identical(x[[1]], quote(`||`)) ||
      identical(x[[1]], quote(`&`)) ||
      identical(x[[1]], quote(`|`))) {
      if (length(x) == 3) {
        if (length(x[[2]]) == 2) {
          if (identical(x[[2]][[1]], quote(`-`)) &&
            is.atomic(x[[2]][[2]])) {
            if (is.atomic(x[[3]])) {
              if (identical(x[[1]], quote(`/`))) {
                return(as.character(-x[[2]][[2]] / x[[3]]))
              } else if (identical(x[[1]], quote(`+`))) {
                return(as.character(-x[[2]][[2]] + x[[3]]))
              } else if (identical(x[[1]], quote(`-`))) {
                return(as.character(-x[[2]][[2]] - x[[3]]))
              } else if (identical(x[[1]], quote(`*`))) {
                return(as.character(-x[[2]][[2]] * x[[3]]))
              }
            }
          }
          if (x[[2]][[2]] == 1 &&
            identical(x[[1]], quote(`*`))) {
            return(paste0("-", ..rxOpt(x[[3]])))
          }
        }
        if (is.atomic(x[[2]]) && is.atomic(x[[3]])) {
          if (identical(x[[1]], quote(`/`))) {
            return(as.character(x[[2]] / x[[3]]))
          } else if (identical(x[[1]], quote(`+`))) {
            return(as.character(x[[2]] + x[[3]]))
          } else if (identical(x[[1]], quote(`-`))) {
            return(as.character(x[[2]] - x[[3]]))
          } else if (identical(x[[1]], quote(`*`))) {
            return(as.character(x[[2]] * x[[3]]))
          }
        }
        if (is.atomic(x[[2]])) {
          if (x[[2]] == 1) {
            if (identical(x[[1]], quote(`*`))) {
              return(..rxOpt(x[[3]]))
            }
          }
          if (x[[2]] == 0) {
            if (identical(x[[1]], quote(`*`))) {
              return("0")
            } else if (identical(x[[1]], quote(`+`))) {
              return(..rxOpt(x[[3]]))
            } else if (identical(x[[1]], quote(`-`))) {
              return(paste0("-", ..rxOpt(x[[3]])))
            } else if (identical(x[[1]], quote(`/`))) {
              return("0")
            }
          }
        }
        if (is.atomic(x[[3]])) {
          if (x[[3]] == 1) {
            if (identical(x[[1]], quote(`*`))) {
              return(..rxOpt(x[[2]]))
            }
          }
          if (x[[3]] == 0) {
            if (identical(x[[1]], quote(`*`))) {
              return("0")
            } else if (identical(x[[1]], quote(`+`))) {
              return(..rxOpt(x[[2]]))
            } else if (identical(x[[1]], quote(`-`))) {
              return(..rxOpt(x[[2]]))
            } else if (identical(x[[2]], quote(`/`))) {
              stop("cannot divide by zero", call. = FALSE)
            }
          }
        }
        return(paste0(
          ..rxOpt(x[[2]]), as.character(x[[1]]),
          ..rxOpt(x[[3]])
        ))
      } else {
        ## Unary Operators
        return(paste0(
          as.character(x[[1]]),
          ..rxOpt(x[[2]])
        ))
      }
    } else if (identical(x[[1]], quote(`~`)) ||
      identical(x[[1]], quote(`=`)) ||
      identical(x[[1]], quote(`<-`))) {
      .rxOptEnv$.new <- NULL
      .x3 <- .rxOptExpr(x[[3]])
      if (length(.x3) == 3) {
        if (any(.x3[1] == c("/", "*", "+", "-"))) {
          .x3 <- paste0(.x3[2], .x3[1], .x3[3])
        }
      }
      if (length(.x3) != 1) {
        stop("error optimizing expression, try 'optExpression=FALSE'")
      }
      .ret <- paste0(
        ..rxOptLhs(x[[2]]),
        ifelse(identical(x[[1]], quote(`<-`)),
          "=", as.character(x[[1]])
        ),
        .x3
      )
      .extra <- NULL
      if (length(.rxOptEnv$.new) > 0) {
        for (.i in seq_along(.rxOptEnv$.rep)) {
          if (any(.rxOptEnv$.rep[[.i]] == .rxOptEnv$.new) &&
            !any(.rxOptEnv$.rep[[.i]] == .rxOptEnv$.added)) {
            .cur <- .rxOptEnv$.rep[[.i]]
            if (.i != 1) {
              for (.j in seq(1, .i - 1)) {
                while (!any(.rxOptEnv$.rep[[.j]] == .rxOptEnv$.added) &&
                  regexpr(
                    rex::rex(or(.cur)),
                    names(.rxOptEnv$.rep)[.j]
                  ) != -1) {
                  .extra <- c(.extra, paste0(
                    .rxOptEnv$.rep[[.j]],
                    "~", names(.rxOptEnv$.rep)[.j]
                  ))
                  .rxOptEnv$.added <- c(
                    .rxOptEnv$.added,
                    .rxOptEnv$.rep[.j]
                  )
                }
              }
            }
            .extra <- c(
              .extra,
              paste0(
                .rxOptEnv$.rep[[.i]], "~",
                ..rxOpt(eval(parse(text = paste0("quote(", names(.rxOptEnv$.rep)[.i], ")"))))
              )
            )
            .rxOptEnv$.added <- c(
              .rxOptEnv$.added,
              .rxOptEnv$.rep[.i]
            )
          }
        }
      }
      return(paste(c(.extra, .ret), collapse = "\n"))
    } else if (identical(x[[1]], quote(`[`))) {
      return(paste0(..rxOpt(x[[2]]), "[", ..rxOpt(x[[3]]), "]"))
    } else {
      .ret0 <- lapply(x, ..rxOpt)
      .ret <- paste0(.ret0[[1]], "(")
      if (.ret == "((") .ret <- "("
      .ret0 <- .ret0[-1]
      .ret <- paste0(.ret, paste(unlist(.ret0), collapse = ", "), ")")
      return(.ret)
    }
  }
}

# -- Chunked optimization ------------------------------------------------------

# Split the model into contiguous chunks of roughly `targetChars` characters.
# Optimizing a chunk is strongly superlinear in its size, so the largest chunk
# dominates the total; balancing on characters rather than lines minimizes it,
# since a sensitivity equation costs far more than a short assignment.
.rxBalancedChunks <- function(lines, targetChars) {
  .w <- pmax(1L, nchar(lines))
  .k <- max(1L, min(length(lines), as.integer(ceiling(sum(.w) / max(1, targetChars)))))
  if (.k <= 1L) return(list(lines))
  unname(split(lines, findInterval(cumsum(.w), seq_len(.k - 1L) * sum(.w) / .k)))
}

# A state initial condition (state(0)=), a dosing modifier (f/F/alag/lag/rate/dur(cmt)=)
# or a delay pre-history (past(cmt, tau)=)
# is a syntax error in a chunk that lacks the matching d/dt() ("'W(0)' present, but
# d/dt(W) not defined").  Such a line cannot simply be moved to the chunk that has the
# d/dt(): its right-hand side may read a variable that a later line reassigns, so its
# position is load-bearing.  Rewrite its left-hand side to a unique plain name *in place*
# instead -- which parses anywhere as an ordinary assignment -- and undo that with
# .rxRestoreCmt() once the chunks are optimized.  The line never moves, so semantics are
# preserved exactly.  Whitespace is kept verbatim (the caller's text need not be
# canonical, since the whole model is deliberately never normalized): a canonical
# left-hand side embeds the compartment name bare (readable), while a spacing variant
# the grammar also accepts (`depot (0)=`, `f( depot )=`) is hex-encoded whole, so the
# disguised name is always a valid identifier and the restore is byte-exact either way.
# past(cmt, tau)= always takes the hex form: its second argument is an arbitrary
# expression (`past(G, exp(THETA[8]))`), which no canonical bare-name form could carry.
# A construct not recognised here is left alone; its chunk then fails and falls back to
# the whole model.
.rxDisguiseCmt <- function(modTxt) {
  .ln <- strsplit(modTxt, "\n", fixed = TRUE)[[1]]
  .eq <- regexpr("=", .ln, fixed = TRUE)
  .raw <- ifelse(.eq > 0L, substr(.ln, 1L, .eq - 1L), "")
  .lhs <- trimws(.raw)
  .lead <- sub("^([ \t]*).*$", "\\1", .raw)
  .trail <- substr(.raw, nchar(.lead) + nchar(.lhs) + 1L, nchar(.raw))
  .rhs <- ifelse(.eq > 0L, substr(.ln, .eq, nchar(.ln)), "")
  .id <- "[a-zA-Z][a-zA-Z0-9_.]*"
  .icRe <- paste0("^(", .id, ")[ \t]*\\(0\\)$")
  .modRe <- paste0("^([fF]|alag|lag|rate|dur)[ \t]*\\([ \t]*(", .id, ")[ \t]*\\)$")
  .pastRe <- paste0("^past[ \t]*\\([ \t]*", .id, "[ \t]*,.*\\)$")
  .isIc <- .eq > 0L & grepl(.icRe, .lhs)
  .isMod <- .eq > 0L & grepl(.modRe, .lhs)
  .isPast <- .eq > 0L & grepl(.pastRe, .lhs)
  .icCan <- .isIc & grepl(paste0("^", .id, "\\(0\\)$"), .lhs)
  .modCan <- .isMod & grepl(paste0("^([fF]|alag|lag|rate|dur)\\(", .id, "\\)$"), .lhs)
  .hex <- (.isIc | .isMod | .isPast) & !(.icCan | .modCan)
  .new <- .lhs
  .new[.icCan] <- paste0("rx__disg_ic__", sub("\\(0\\)$", "", .lhs[.icCan]), "__")
  .mm <- regmatches(.lhs[.modCan], regexec("^([a-zA-Z]+)\\((.*)\\)$", .lhs[.modCan]))
  .new[.modCan] <- vapply(.mm, function(.m) paste0("rx__disg_mod__", .m[2L], "__", .m[3L], "__"),
                          character(1))
  .new[.hex] <- vapply(.lhs[.hex], function(.l) {
    paste0("rx__disg_lhs__", paste(as.character(charToRaw(.l)), collapse = ""), "__")
  }, character(1), USE.NAMES = FALSE)
  paste(ifelse(.isIc | .isMod | .isPast, paste0(.lead, .new, .trail, .rhs), .ln),
        collapse = "\n")
}

# Reverse .rxDisguiseCmt().  Only the left-hand side is rewritten (split at the first
# "="), so the optimized right-hand side is kept verbatim.  The rx__disg_lhs__ form is
# hex-decoded back to the original left-hand side, byte for byte.
.rxRestoreCmt <- function(txt) {
  .ln <- strsplit(txt, "\n", fixed = TRUE)[[1]]
  .disg <- grepl("^[ \t]*rx__disg_(ic|mod|lhs)__", .ln)
  if (any(.disg)) {
    .eq <- regexpr("=", .ln[.disg], fixed = TRUE)
    .raw <- substr(.ln[.disg], 1L, .eq - 1L)
    .rest <- substr(.ln[.disg], .eq, nchar(.ln[.disg]))
    .lead <- sub("^([ \t]*).*$", "\\1", .raw)
    .tok <- trimws(.raw)
    .trail <- substr(.raw, nchar(.lead) + nchar(.tok) + 1L, nchar(.raw))
    .tok <- sub("^rx__disg_ic__(.*)__$", "\\1(0)", .tok)
    .tok <- sub("^rx__disg_mod__([a-zA-Z]+)__(.*)__$", "\\1(\\2)", .tok)
    .isHex <- grepl("^rx__disg_lhs__([0-9a-f][0-9a-f])+__$", .tok)
    .tok[.isHex] <- vapply(sub("^rx__disg_lhs__([0-9a-f]+)__$", "\\1", .tok[.isHex]),
                           function(.h) {
                             rawToChar(as.raw(strtoi(substring(.h, seq(1L, nchar(.h), 2L),
                                                               seq(2L, nchar(.h), 2L)), 16L)))
                           }, character(1), USE.NAMES = FALSE)
    .ln[.disg] <- paste0(.lead, .tok, .trail, .rest)
  }
  paste(.ln, collapse = "\n")
}

# delay(state, T) -- and its sensitivity derivatives rxDelayD/rxDelayD2/rxDelayD3 --
# parses only where d/dt(state) is defined, so a chunk boundary separating the two makes
# the chunk a syntax error: the parser's :ERR: block prints mid-fit and the whole model
# falls back to the unchunked pass.  Unlike the left-hand sides above, a delay() call is
# a right-hand-side *expression* whose repeated occurrences are worth factoring (each is
# a dense-history interpolation at runtime), so it is only disguised where it must be:
# in a chunk that lacks the matching d/dt(), the whole call is replaced in place by a
# unique plain name -- which parses anywhere -- and restored verbatim by
# .rxRestoreDelay() once the chunks are reassembled.  A chunk holding both keeps the
# call and its factoring.  A call whose state has no d/dt() anywhere in the model is
# left alone, so a malformed model still fails its chunk and reaches the whole-model
# fallback, which raises exactly what the unchunked call raises.  Likewise a call this
# scan cannot delimit (spanning lines, or a non-name first argument) is left alone.
.rxDisguiseDelayChunks <- function(chunks) {
  .ddtRe <- "^[ \t]*d[ \t]*/[ \t]*dt[ \t]*\\([ \t]*([a-zA-Z][a-zA-Z0-9_.]*)[ \t]*\\)[ \t]*(=|<-|~)"
  .ddtOf <- function(.ln) {
    .m <- regmatches(.ln, regexec(.ddtRe, .ln))
    unique(vapply(.m[lengths(.m) == 3L], `[`, character(1), 2L))
  }
  .allDdt <- unique(unlist(lapply(chunks, .ddtOf)))
  .fnRe <- "(?<![a-zA-Z0-9_.])(rxDelayD3|rxDelayD2|rxDelayD|delay)[ \t]*\\("
  .map <- character(0) # id -> original call text, shared across chunks
  .oneLine <- function(.l, .ddt) {
    .pos <- 1L
    repeat {
      .m <- regexpr(.fnRe, substr(.l, .pos, nchar(.l)), perl = TRUE)
      if (.m == -1L) break
      .start <- .pos + as.integer(.m) - 1L
      .open <- .start + attr(.m, "match.length") - 1L
      .ch <- strsplit(substr(.l, .open, nchar(.l)), "", fixed = TRUE)[[1]]
      .rel <- which(cumsum((.ch == "(") - (.ch == ")")) == 0L)[1L]
      if (is.na(.rel)) { # no matching ")" on this line: leave the call alone
        .pos <- .open + 1L
        next
      }
      .end <- .open + .rel - 1L
      .inner <- substr(.l, .open + 1L, .end - 1L)
      .st <- regmatches(.inner,
                        regexec("^[ \t]*([a-zA-Z][a-zA-Z0-9_.]*)[ \t]*,", .inner))[[1]]
      # keep scanning inside the arguments (nested calls) when this call stays
      if (length(.st) != 2L || .st[2L] %in% .ddt || !(.st[2L] %in% .allDdt)) {
        .pos <- .open + 1L
        next
      }
      .call <- substr(.l, .start, .end)
      .hit <- match(.call, .map)
      if (is.na(.hit)) {
        .id <- sprintf("rx__disg_delay_%d__", length(.map) + 1L)
        .map[[.id]] <<- .call
      } else {
        .id <- names(.map)[.hit]
      }
      .l <- paste0(substr(.l, 1L, .start - 1L), .id, substr(.l, .end + 1L, nchar(.l)))
      .pos <- .start + nchar(.id)
    }
    .l
  }
  .chunks <- lapply(chunks, function(.ln) {
    .ddt <- .ddtOf(.ln)
    vapply(.ln, .oneLine, character(1), .ddt = .ddt, USE.NAMES = FALSE)
  })
  list(chunks = .chunks, map = .map)
}

# Reverse .rxDisguiseDelayChunks(): each unique name is put back as the original call
# text, byte for byte, wherever the optimizer left it (in place, or hoisted into a
# rx_expr_ temporary's definition).
.rxRestoreDelay <- function(txt, map) {
  for (.id in names(map)) {
    txt <- gsub(.id, map[[.id]], txt, fixed = TRUE)
  }
  txt
}

# A chunk is itself a (smaller) model, so optimize it with rxOptExpr() and chunking off.
# Each chunk restarts its rx_expr_ counter at zero, so the names one chunk introduces would
# collide with another's once reassembled; prefix them per chunk.
#
# Only the names *this* call introduced may be renamed.  A rx_expr_ name already present in
# the chunk -- the model was optimized before, say -- must be left exactly as it is: its
# definition and its uses can sit in different chunks, and renaming them per chunk would
# rename them differently and pull them apart.  Matching whole words likewise keeps a
# variable that merely contains "rx_expr_" (say `my_rx_expr_var`) from being rewritten.
#
# Errors are deliberately not caught here: a chunk that cannot be optimized must reach
# .rxOptExprChunked(), which falls back to the whole model.
.rxOptExprChunk <- function(i, chunks, msg) {
  .txt <- paste(chunks[[i]], collapse = "\n")
  .o <- suppressMessages(rxOptExpr(.txt, msg, chunkLines = 0L))
  .re <- "\\brx_expr_[0-9]+\\b"
  .new <- setdiff(unique(regmatches(.o, gregexpr(.re, .o))[[1]]),
                  unique(regmatches(.txt, gregexpr(.re, .txt))[[1]]))
  for (.v in .new) {
    .o <- gsub(paste0("\\b", .v, "\\b"),
               sub("^rx_expr_", sprintf("rx_expr_c%d_", i), .v), .o)
  }
  .o
}

# Chunked (optionally parallel) common subexpression optimization.  Subexpressions are only
# shared within a chunk, so the optimized text is not the text the whole-model call would
# produce -- it carries more temporaries -- but it is an equivalent model, with the same
# states and parameters and the same solution, and a malformed model still raises the same
# error.  What is optimized changes; what the model *means* does not.
.rxOptExprChunked <- function(x, msg = "model", chunkLines = 40L, parallel = 0L) {
  # Never rxNorm() the whole model here.  Normalizing (i.e. parsing) is itself strongly
  # superlinear in model size, and is what actually dominates optimizing a large model: on a
  # 275-line augmented model the whole-model call is ~113s, of which the common subexpression
  # search is only ~15s.  Chunking is fast precisely because rxOptExpr() normalizes each
  # chunk on its own, so normalizing the whole model here would pay the very cost this is
  # avoiding.  Text is already line-oriented and is split as-is; only an object needs rxNorm.
  #
  # Mirror how rxModelVars() reads a character, in its order: a length-1 string may be a
  # filename (the file holds the model text; read it) or a registered model name (it has no
  # "=", "<-" or "~"; only rxNorm() can resolve it); anything else is literal model text.
  if (is.character(x) && length(x) == 1L &&
        isTRUE(tryCatch(file.exists(x), error = function(e) FALSE,
                        warning = function(w) FALSE))) {
    x <- readLines(x, warn = FALSE)
  } else if (is.character(x) && length(x) == 1L && !grepl("[=~]|<-", x)) {
    x <- rxNorm(x)
  }
  .txt <- if (is.character(x)) paste(x, collapse = "\n") else rxNorm(x)
  .ln <- strsplit(.txt, "\n", fixed = TRUE)[[1]]
  if (length(.ln) <= chunkLines) {
    return(rxOptExpr(.txt, msg = msg, chunkLines = 0L))
  }
  # Chunking introduces names of its own into the model's namespace: rx_expr_c<i>_ for the
  # temporaries a chunk contributes, and rx__disg_ while a compartment-scoped line or a
  # boundary-split delay() call is disguised.  A model that already uses such a name would
  # have it silently captured --
  # renaming or restoring would rewrite the model's own variable -- so do not chunk it.
  if (grepl("rx_expr_c[0-9]", .txt) || grepl("rx__disg_", .txt, fixed = TRUE)) {
    return(rxOptExpr(.txt, msg = msg, chunkLines = 0L))
  }

  # Disguise compartment-scoped left-hand sides so that every chunk parses standalone.
  .chunks <- .rxBalancedChunks(strsplit(.rxDisguiseCmt(.txt), "\n", fixed = TRUE)[[1]],
                               mean(pmax(1L, nchar(.ln))) * chunkLines)
  # Disguise delay() calls that a chunk boundary separated from their d/dt().
  .delay <- .rxDisguiseDelayChunks(.chunks)
  .chunks <- .delay$chunks
  .nChunks <- length(.chunks)

  # `parallel` carries rxControl(cores=)'s semantics: 0 means the rxode2 thread setting
  # (rxCores(), which setRxThreads()/OMP_THREAD_LIMIT control -- so CRAN and users tune
  # this with the same knob as the solver), n > 0 means n.  Never use more daemons than
  # there are chunks, nor more than that thread setting; a single daemon has no
  # parallelism to offer, only dispatch overhead, so 1 runs serially.
  .nDaemons <- as.integer(parallel)
  if (is.na(.nDaemons) || .nDaemons < 0L) .nDaemons <- 0L
  if (.nDaemons == 0L) .nDaemons <- max(1L, as.integer(rxCores()))
  .nDaemons <- min(.nDaemons, .nChunks, max(1L, as.integer(rxCores())))
  .useMirai <- .nDaemons > 1L
  # A caller's existing mirai pool is used as-is and never shut down; a pool of our own
  # is only worth starting (a few seconds: each daemon loads rxode2) when there are
  # enough chunks for the parallel win to beat that startup.
  .ownDaemons <- FALSE
  if (.useMirai) {
    .have <- tryCatch(sum(mirai::status()$connections) > 0L, error = function(e) FALSE)
    if (!.have) {
      .ownDaemons <- .nChunks >= 4L
      .useMirai <- .ownDaemons
    }
  }
  .malert(sprintf("optimizing duplicate expressions in %s (%d chunks%s)...", msg, .nChunks,
                  if (.useMirai) sprintf(", %d daemons", .nDaemons) else ""))

  .opt <- tryCatch({
    if (.useMirai) {
      if (.ownDaemons) {
        mirai::daemons(.nDaemons)
        on.exit(mirai::daemons(0), add = TRUE)
      }
      # `.rxOptExprChunk` is passed as an argument, carrying the rxode2 namespace as its
      # environment, so a chunk is optimized by the same code path serially and in parallel.
      .tasks <- mirai::mirai_map(
        seq_len(.nChunks),
        function(.i, .chunks, .msg, .optOne) {
          library(rxode2)
          .optOne(.i, .chunks, .msg)
        },
        .args = list(.chunks = .chunks, .msg = msg, .optOne = .rxOptExprChunk)
      )
      # A chunk that failed in a daemon comes back as an error object rather than throwing,
      # and would otherwise be pasted into the model text; raise it so the fallback below
      # optimizes the whole model instead.
      .res <- character(.nChunks)
      for (.i in seq_len(.nChunks)) {
        .r <- .tasks[[.i]][]
        if (inherits(.r, "miraiError") || inherits(.r, "errorValue") || !is.character(.r)) {
          stop(sprintf("parallel chunk %d failed in a mirai daemon: %s", .i,
                       tryCatch(conditionMessage(.r),
                                error = function(e) paste(utils::head(unclass(.r), 1L),
                                                          collapse = ""))),
               call. = FALSE)
        }
        .res[.i] <- .r
      }
      .res
    } else {
      vapply(seq_len(.nChunks), .rxOptExprChunk, character(1), chunks = .chunks, msg = msg)
    }
  }, error = function(e) NULL)

  # A chunk is only a fragment of the model, so it can fail to optimize where the whole
  # model would not -- it may hold a compartment-scoped line the disguise did not recognise,
  # or only part of a statement.  It can equally fail because the model is malformed, and
  # the chunk alone cannot tell those apart.  The whole model can: it optimizes cleanly for
  # a mere fragment, and raises exactly what the unchunked call raises for a broken model.
  # (A chunk with nothing left to reduce returns its text and does not error, so an ordinary
  # model never lands here.)
  if (is.null(.opt)) {
    return(rxOptExpr(.txt, msg = msg, chunkLines = 0L))
  }
  .rxRestoreCmt(.rxRestoreDelay(paste(.opt, collapse = "\n"), .delay$map))
}

#' Optimize rxode2 for computer evaluation
#'
#' This optimizes rxode2 code for computer evaluation by only
#' calculating redundant expressions once.
#'
#' @param x rxode2 model that can be accessed by rxNorm
#'
#' @param msg This is the name of type of object that rxode2 is
#'     optimizing that will in the message when optimizing.  For
#'     example "model" will produce the following message while
#'     optimizing the model:
#'
#'  finding duplicate expressions in model...
#'
#' @param chunkLines Integer; when positive (the default is 40),
#'     a model longer than this many lines is optimized in contiguous
#'     cost-balanced chunks of roughly this many lines instead of in a
#'     single pass; a model at or under it is optimized whole, exactly
#'     as before.  `0` always optimizes the whole model at once.
#'
#'     Chunking pays off for a large machine-generated model -- a
#'     sensitivity- or Jacobian-augmented model, say.  Normalizing a
#'     model (`rxNorm()`, i.e. parsing it) is strongly superlinear in
#'     its size, and for such a model it, not the common subexpression
#'     search, is what dominates: optimizing a 275-line augmented model
#'     takes ~113s, of which the subexpression search is only ~15s.
#'     Chunking amortizes that parse -- `rxOptExpr()` normalizes each
#'     chunk on its own -- taking the same model to ~11s:
#'
#'     \tabular{rrrr}{
#'       lines \tab whole \tab chunked \tab \cr
#'       34 \tab 0.5s \tab 0.5s \tab (a typical model: not chunked) \cr
#'       119 \tab 2.0s \tab 0.8s \tab 2.5x \cr
#'       149 \tab 22.2s \tab 3.9s \tab 5.7x \cr
#'       275 \tab 112.7s \tab 10.6s \tab 10.7x \cr
#'     }
#'
#'     Common subexpressions are then only shared within a chunk, so the
#'     model is equivalent but carries more temporaries.  That costs no
#'     measurable solve time, but it does make the C compilation of the
#'     model somewhat slower, which partly offsets the gain.
#'
#'     A chunk is a fragment, so it can fail to optimize where the whole
#'     model would not.  If any chunk fails, the whole model is optimized
#'     instead, so a malformed model still raises the error the unchunked
#'     call raises; falling back costs the unchunked time only on that
#'     rare path.
#'
#'     Chunking therefore does not give the same optimized text as the
#'     whole-model call -- it shares fewer subexpressions and so carries
#'     more temporaries -- but it gives an equivalent model: the same
#'     states and parameters, the same solution, and the same errors.
#'
#' @param parallel Integer; number of `mirai` daemons used to optimize
#'     the chunks in parallel.  Only used when the model is chunked.  It
#'     carries the same semantics as `rxControl(cores=)`: `0` (the
#'     default) means the rxode2 thread setting `rxCores()`, so CRAN and
#'     users tune it with the same knob as the solver (`setRxThreads()`,
#'     `OMP_THREAD_LIMIT`) or by passing `parallel=` directly; `1` runs
#'     the chunks serially.
#'     It is capped by the number of chunks and by `rxCores()`, so it
#'     will not oversubscribe past the threads the user asked for.
#'
#'     An existing `mirai` daemon pool is used as-is and left running.
#'     Otherwise a pool is started for the call and shut down when it
#'     returns; that startup (loading rxode2 into each daemon) costs a
#'     few seconds, so a pool is only started when the model splits into
#'     at least 4 chunks, where the parallel win covers it.
#'
#' @return Optimized rxode2 model text.  The order and type lhs and
#'     state variables is maintained while the evaluation is sped up.
#'     While parameters names are maintained, their order may be
#'     modified.
#'
#' @author Matthew L. Fidler
#' @export
rxOptExpr <- function(x, msg = "model", chunkLines = 40L,
                      parallel = 0L) {
  .chunkLines <- as.integer(chunkLines)
  if (!is.na(.chunkLines) && .chunkLines > 0L) {
    return(.rxOptExprChunked(x, msg = msg, chunkLines = .chunkLines, parallel = parallel))
  }
  .oldOpts <- options()
  options(digits = 22)
  on.exit(options(.oldOpts))
  .mv <- rxModelVars(x)
  .params <- .mv$params
  .rxOptEnv$.list <- list()
  .rxOptEnv$.rep <- list()
  .rxOptEnv$.added <- NULL
  .rxOptEnv$.exclude <- ""
  .malert(sprintf("finding duplicate expressions in %s...", msg))
  .p <- eval(parse(text = paste0("quote({", rxNorm(x), "})")))
  .lines <- ..rxOpt(.p, progress = TRUE)
  .rxOptEnv$.list <- .rxOptEnv$.list[which(unlist(.rxOptEnv$.list) > 1L)]
  .exprs <- names(.rxOptEnv$.list)[order(nchar(names(.rxOptEnv$.list)))]
  .exprs <- .exprs[regexpr(rex::rex(start, regNum, end), .exprs,
    perl = TRUE
  ) == -1]
  .thetaEtaR <- rex::rex(start, or("THETA[", "ETA["), any_numbers, "]", end)
  .exprs <- .exprs[regexpr(.thetaEtaR, .exprs, perl = TRUE) == -1]
  if (length(.exprs) > 0) {
    ## Take out unary [-] that way
    ## expr1=-ka       #nolint
    ## expr2 = expr-ka #nolint
    ## will not become expr = exprka where exprka isn't defined.
    .exprs <- .exprs[regexpr("^[-]", .exprs) == -1]
    if (length(.exprs) == 0) {
      return(x)
    }
    .rp <- rxOptRep_(.exprs)
    .rxOptEnv$.rep <- as.list(.rp[[1]])
    .rxOptEnv$.exclude <- ""
    .malert(sprintf("optimizing duplicate expressions in %s...", msg))
    .opt <- ..rxOpt(.p, progress = TRUE)
    return(paste(.opt, collapse = "\n"))
  } else {
    return(paste(.lines, collapse = "\n"))
  }
}

rxTest({
  test_that("simple expression optimization", {
    exp1 <- "rx_yj_~2;\nrx_lambda_~1;\nrx_pred_=10*exp(-THETA[1]*t*exp(ETA[1]));\nrx__sens_rx_pred__BY_ETA_1___=-10*THETA[1]*t*exp(ETA[1])*exp(-THETA[1]*t*exp(ETA[1]));\nrx_r_=100*Rx_pow_di(THETA[2],2)*exp(-2*THETA[1]*t);\ndvid(3,4)\n"
    suppressMessages(expect_equal(
      rxOptExpr(exp1),
      "rx_yj_~2\nrx_lambda_~1\nrx_expr_0~exp(ETA[1])\nrx_expr_1~exp(-THETA[1]*t*rx_expr_0)\nrx_pred_=10*rx_expr_1\nrx__sens_rx_pred__BY_ETA_1___=-10*THETA[1]*t*rx_expr_0*rx_expr_1\nrx_r_=100*Rx_pow_di(THETA[2], 2)*exp(-2*THETA[1]*t)\ndvid(3, 4)"
    ))
  })

  test_that("expression optimization errors", {
    suppressMessages(expect_error(
      rxOptExpr("A1=exp(-k10*(tau - tinf))*r1*(1.0 - exp(-k10*tinf))/(k10*(1.0 - exp(-tau*k10)))"),
      NA
    ))
    suppressMessages(expect_error(
      rxOptExpr("A1=r1/ka\nA1ka=-r1/ka^2\nA1k20=0\nA1b1=0\nA1r1=ka^(-1)\nA2=r1/k20\nA2ka=0\nA2k20=-r1/k20^2\nA2b1=0\nA2r1=k20^(-1)"),
      NA
    ))
  })

  suppressMessages(rxOptExpr("a=1+(-1/2)*b"))

  suppressMessages(rxOptExpr("a=-1*exp(b)"))

  suppressMessages(rxOptExpr("a=1+(((-1/2)))*b"))

  suppressMessages(rxOptExpr("a=1+(1/2)*b; c=d^(1/2); e=(1/2)*f^(1/2)"))

  test_that("simple expression optimization", {
    expect_equal(length(..rxOpt(quote(exp(ETA[1] + THETA[4]) + 0))), 1L)
  })

  ## chunked optimization -------------------------------------------------

  # a model big enough to chunk, carrying both kinds of compartment-scoped construct:
  # a parameter-dependent initial condition and dosing modifiers.  Optimizing a chunk
  # that held one of these *without* its d/dt() used to be a syntax error, which is
  # what the disguise/restore round trip below exists to prevent.
  .chunkModel <- function(n = 60L) {
    .s <- vapply(seq_len(n), function(i) {
      sprintf("v%d=exp(THETA[1]+ETA[1])*exp(THETA[2]*%d)+sin(THETA[3]*t)", i, i)
    }, character(1))
    paste(c("d/dt(depot)=-exp(THETA[1]+ETA[1])*depot",
            "d/dt(center)=exp(THETA[1]+ETA[1])*depot-exp(THETA[2])*center",
            "depot(0)=exp(THETA[4])",          # parameter-dependent initial condition
            "f(depot)=exp(THETA[5])",          # dosing modifiers
            "alag(depot)=exp(THETA[6])",
            .s,
            "rx_pred_=center*exp(THETA[3])"), collapse = "\n")
  }

  test_that("the default chunks a large model; chunkLines=0 optimizes it whole", {
    .m <- .chunkModel()
    .def <- suppressMessages(rxOptExpr(.m, "model"))
    # per-chunk rx_expr_c<i>_ names prove the default took the chunked path
    expect_true(grepl("rx_expr_c[0-9]+_", .def))
    expect_identical(.def, suppressMessages(rxOptExpr(.m, "model", chunkLines = 40L)))
    # chunkLines = 0 still forces the single whole-model pass
    .whole <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 0L))
    expect_false(grepl("rx_expr_c[0-9]+_", .whole))
  })

  test_that("chunked optimization yields an equivalent model", {
    .m <- .chunkModel()
    .whole <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 0L))
    .chunk <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 40L))
    # the compartment-scoped lines survive, in place, and the chunked model still builds
    expect_true(grepl("depot(0)=", .chunk, fixed = TRUE))
    expect_true(grepl("f(depot)=", .chunk, fixed = TRUE))
    expect_true(grepl("alag(depot)=", .chunk, fixed = TRUE))
    expect_error(rxModelVars(.chunk), NA)
    # same states, and same parameters (chunking must not leave an undefined variable
    # behind as a spurious input parameter)
    .rv <- function(.x) {
      .mv <- rxModelVars(.x)
      list(state = .mv$state, params = sort(.mv$params[!grepl("^rx_expr_", .mv$params)]))
    }
    expect_identical(.rv(.whole), .rv(.chunk))
  })

  test_that("a model at or below chunkLines is untouched by the default", {
    .m <- "d/dt(depot)=-ka*depot\nd/dt(center)=ka*depot-cl/v*center"
    suppressMessages(expect_identical(rxOptExpr(.m, "model"),
                                      rxOptExpr(.m, "model", chunkLines = 0L)))
  })

  test_that("compartment-scoped disguise/restore round-trips exactly", {
    # byte-exact for every spacing variant, so a chunk boundary can never alter the
    # surrounding whitespace, and idempotent in both directions
    .lines <- c("depot(0)=W0", "  depot(0)=W0", "\tdepot(0)=W0", "depot(0) = W0",
                "depot (0)=W0", "f(depot)=ff", "f( depot )=ff", "f (depot)=ff",
                "F(depot)=ff", "alag( central )=tlag",
                "rate(depot)=r0", "dur(depot)=d0", "lag(depot)=tl",
                "rx__sens_W_BY_ETA_1___(0)=exp(x)",
                "d/dt(depot)=-ka*depot", "ka=exp(tka)")
    for (.l in .lines) {
      expect_identical(.rxRestoreCmt(.rxDisguiseCmt(.l)), .l)
    }
    .txt <- paste(.lines, collapse = "\n")
    expect_identical(.rxRestoreCmt(.rxDisguiseCmt(.txt)), .txt)
    expect_identical(.rxDisguiseCmt(.rxDisguiseCmt(.txt)), .rxDisguiseCmt(.txt))
    expect_identical(.rxRestoreCmt(.rxRestoreCmt(.txt)), .txt)
    # every disguised left-hand side is a syntactically valid identifier, even for the
    # spacing variants (`depot (0)=`, `f( depot )=`), so the chunk it lands in parses
    .lhs <- trimws(sub("=.*$", "", strsplit(.rxDisguiseCmt(.txt), "\n", fixed = TRUE)[[1]]))
    .disg <- grepl("^rx__disg_", .lhs)
    expect_true(all(grepl("^[a-zA-Z][a-zA-Z0-9_.]*$", .lhs[.disg])))
    # nothing compartment-scoped is left for a chunk to orphan
    expect_false(any(endsWith(.lhs, "(0)")))
    expect_false(any(grepl("^([fF]|alag|lag|rate|dur)[ \t]*\\(", .lhs) & endsWith(.lhs, ")")))
  })

  test_that("grammar-accepted spacing variants chunk without falling back", {
    # `depot (0)=` and `f( depot )=` used to be disguised into an identifier containing
    # spaces, a syntax error that silently sent every such model down the whole-model
    # fallback; they are now hex-encoded, so the chunked path handles them.
    .pad <- vapply(1:60, function(i) {
      sprintf("v%d=exp(THETA[1]+ETA[1])*exp(THETA[2]*%d)", i, i)
    }, character(1))
    .m <- paste(c("d/dt(depot)=-exp(THETA[1]+ETA[1])*depot",
                  "d/dt(center)=exp(THETA[1]+ETA[1])*depot-exp(THETA[2])*center",
                  "depot (0)=exp(THETA[4])",
                  "f( depot )=exp(THETA[5])",
                  "alag (depot)=exp(THETA[6])",
                  .pad, "rx_pred_=center"), collapse = "\n")
    .out <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 40L))
    # per-chunk rx_expr_c<i>_ names prove the chunked path ran (a fallback has none)
    expect_true(grepl("rx_expr_c[0-9]+_", .out))
    # the spacing-variant lines are restored byte-exact and the model still parses
    expect_true(grepl("depot (0)=", .out, fixed = TRUE))
    expect_true(grepl("f( depot )=", .out, fixed = TRUE))
    expect_true(grepl("alag (depot)=", .out, fixed = TRUE))
    expect_error(rxModelVars(.out), NA)
  })

  test_that("a chunk boundary separating delay() from its d/dt() no longer falls back", {
    # delay(state, T) parses only where d/dt(state) is defined; a boundary between the
    # two used to fail the chunk, print the parser's :ERR: block, and lose the chunking
    # speedup to the whole-model fallback.  The call is now disguised across the
    # boundary and restored byte-exact.
    .pad <- vapply(1:60, function(i) {
      sprintf("v%d=exp(THETA[1]+ETA[1])*exp(THETA[2]*%d)", i, i)
    }, character(1))
    .m <- paste(c("d/dt(x)=-exp(THETA[1])*x",
                  .pad,
                  "a=delay(x, exp(THETA[2]))+exp(THETA[1]+ETA[1])",
                  "b=rxDelayD(x, exp(THETA[2]))*2+exp(THETA[1]+ETA[1])",
                  "rx_pred_=a+b"), collapse = "\n")
    .chunk <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 40L))
    # per-chunk rx_expr_c<i>_ names prove the chunked path ran (a fallback has none)
    expect_true(grepl("rx_expr_c[0-9]+_", .chunk))
    # the calls are restored byte-exact, nothing disguised is left, and it still parses
    expect_true(grepl("delay(x, exp(THETA[2]))", .chunk, fixed = TRUE))
    expect_true(grepl("rxDelayD(x, exp(THETA[2]))", .chunk, fixed = TRUE))
    expect_false(grepl("rx__disg_delay_", .chunk, fixed = TRUE))
    expect_error(rxModelVars(.chunk), NA)
    # same states and parameters as the whole-model pass
    .whole <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 0L))
    .rv <- function(.x) {
      .mv <- rxModelVars(.x)
      list(state = .mv$state, params = sort(.mv$params[!grepl("^rx_expr_", .mv$params)]))
    }
    expect_identical(.rv(.whole), .rv(.chunk))
  })

  test_that("a chunk holding delay() and its d/dt() keeps optimizing the call", {
    # co-located calls are NOT disguised, so their arguments still join the chunk's
    # common-subexpression pool
    .pad <- vapply(1:60, function(i) {
      sprintf("v%d=exp(THETA[1]+ETA[1])*exp(THETA[2]*%d)", i, i)
    }, character(1))
    .m <- paste(c("d/dt(x)=-exp(THETA[1])*x",
                  "a=delay(x, exp(THETA[2]))+1",
                  "b=delay(x, exp(THETA[2]))*2",
                  .pad,
                  "rx_pred_=a+b"), collapse = "\n")
    .chunk <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 40L))
    expect_true(grepl("rx_expr_c[0-9]+_", .chunk))
    # the delay argument was factored in place of the literal exp(THETA[2])
    expect_true(grepl("delay\\(x, rx_expr_c[0-9]+_[0-9]+\\)", .chunk))
  })

  test_that("a delay() state with no d/dt() anywhere raises the whole-model error", {
    # the disguise only hides a call whose d/dt() exists in another chunk; a malformed
    # model must still reach the whole-model fallback and raise the unchunked error
    .pad <- vapply(1:60, function(i) {
      sprintf("v%d=exp(THETA[1]+ETA[1])*exp(THETA[2]*%d)", i, i)
    }, character(1))
    .m <- paste(c(.pad, "a=delay(x, exp(THETA[2]))+1", "rx_pred_=a"), collapse = "\n")
    .err <- function(.chunkLines) {
      tryCatch({
        .z <- capture.output(suppressMessages(
          rxOptExpr(.m, "model", chunkLines = .chunkLines)), type = "output")
        NA_character_
      }, error = function(e) conditionMessage(e))
    }
    .chunk <- .err(40L)
    expect_false(is.na(.chunk))
    expect_identical(.chunk, .err(0L))
  })

  test_that("delay disguise/restore round-trips and is chunk-local", {
    .chunks <- list(
      c("d/dt(x)=-k*x", "d/dt(y)=k*x-ke*y",
        "a=delay(x, tau)+1"),                      # d/dt(x) here: kept as-is
      c("b=delay(x, exp(tau))*2",                  # split from d/dt(x): disguised
        "c=rxDelayD2(y, tau+delay(x, exp(tau)))",  # nested call inside a disguised one
        "d=delay(z, tau)"))                        # no d/dt(z) anywhere: left alone
    .d <- .rxDisguiseDelayChunks(.chunks)
    expect_identical(.d$chunks[[1]], .chunks[[1]])
    # every delay-family call split from its d/dt() is hidden in chunk 2 ...
    expect_false(any(grepl("delay\\(x|rxDelayD2\\(y", .d$chunks[[2]])))
    # ... while the call whose state has no d/dt() anywhere survives untouched
    expect_true(grepl("delay(z, tau)", .d$chunks[[2]][3], fixed = TRUE))
    # identical calls share one name; restore is byte-exact
    expect_identical(.d$chunks[[2]][1],
                     sub("delay(x, exp(tau))", names(.d$map)[match("delay(x, exp(tau))", .d$map)],
                         .chunks[[2]][1], fixed = TRUE))
    .joined <- vapply(.d$chunks, paste, character(1), collapse = "\n")
    expect_identical(.rxRestoreDelay(paste(.joined, collapse = "\n"), .d$map),
                     paste(vapply(.chunks, paste, character(1), collapse = "\n"),
                           collapse = "\n"))
  })

  test_that("a filename is read as a file when chunking, like the unchunked call", {
    .f <- tempfile(fileext = ".rx")
    on.exit(unlink(.f), add = TRUE)
    writeLines(.chunkModel(), .f)
    suppressMessages(expect_identical(
      rxOptExpr(.f, "model", chunkLines = 40L),
      rxOptExpr(.chunkModel(), "model", chunkLines = 40L)))
  })

  test_that("chunks are cost balanced and lossless", {
    .ln <- c("aaaaaaaaaaaaaaaaaaaa", "b", "c", "dddddddddddddddddddd", "e", "f")
    .ch <- .rxBalancedChunks(.ln, 20)
    expect_identical(unlist(.ch, use.names = FALSE), .ln)  # order and content preserved
    expect_true(length(.ch) > 1L)
  })

  test_that("a model already using the chunker's own generated names is not chunked", {
    # Chunking introduces names into the model's namespace -- rx_expr_c<i>_ for the
    # temporaries a chunk contributes, and rx__disg_ while a compartment-scoped line is
    # disguised.  rxode2 does not reserve these, so a model may legitimately use one; it
    # would then be captured (renaming or restoring would rewrite the model's own variable)
    # and the model would silently change.  Such a model is left to the whole-model call.
    .pad <- vapply(1:45, function(i) sprintf("a%d=exp(THETA[1])*%d", i, i), character(1))
    .ode <- c("d/dt(depot)=-exp(THETA[1])*depot",
              "d/dt(center)=exp(THETA[1])*depot-exp(THETA[2])*center")
    for (.nm in c("rx_expr_c1_0", "rx__disg_ic__depot__", "rx__disg_mod__f__depot__")) {
      .m <- paste(c(.ode, paste0(.nm, "=exp(THETA[3])"), .pad,
                    paste0("cp=center*", .nm)), collapse = "\n")
      .chunked <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 40L))
      # falls back to the whole-model call, so the model's own variable is untouched
      expect_identical(.chunked, suppressMessages(rxOptExpr(.m, "model", chunkLines = 0L)))
      expect_true(grepl(.nm, .chunked, fixed = TRUE))
    }
  })

  test_that("only the rx_expr_ names a chunk introduces are renamed", {
    # The chunks are optimized independently, so each restarts its rx_expr_ counter and the
    # names must be prefixed per chunk.  But renaming *every* rx_expr_ occurrence would
    # break two cases: a name already in the model (it was optimized before) whose
    # definition and uses land in different chunks would be renamed differently and pulled
    # apart; and a variable that merely contains "rx_expr_" would be rewritten.
    .pad <- vapply(1:45, function(i) sprintf("a%d=exp(THETA[1]+ETA[1])*%d", i, i), character(1))
    .base <- c("d/dt(depot)=-exp(THETA[1]+ETA[1])*depot",
               "d/dt(center)=exp(THETA[1]+ETA[1])*depot-exp(THETA[2])*center")

    # (a) re-optimizing an already-optimized model
    .once <- suppressMessages(rxOptExpr(paste(c(.base, .pad, "cp=center"), collapse = "\n")))
    expect_true(grepl("rx_expr_", .once))
    .twice <- suppressMessages(rxOptExpr(.once, "model", chunkLines = 40L))
    expect_error(rxModelVars(.twice), NA)
    # no rx_expr_ left dangling as an input parameter
    expect_false(any(grepl("^rx_expr_", rxModelVars(.twice)$params)))

    # (b) a user variable whose name merely contains "rx_expr_"
    .m <- paste(c(.base, "my_rx_expr_var=exp(THETA[3])", .pad,
                  "cp=center*my_rx_expr_var"), collapse = "\n")
    .o <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 40L))
    expect_true(grepl("my_rx_expr_var", .o, fixed = TRUE))   # left exactly as it was
    expect_error(rxModelVars(.o), NA)
    expect_false(any(grepl("^rx_expr_", rxModelVars(.o)$params)))
  })

  test_that("a failing chunk falls back to the whole model, so chunking changes nothing", {
    # A chunk is a fragment, so it can fail where the whole model would not.  But it can
    # equally fail because the model is genuinely malformed, and the chunk alone cannot tell
    # the two apart -- so any chunk failure falls back to optimizing the whole model, which
    # can.  Chunking must therefore return the same model, and raise the same error, as the
    # unchunked call in BOTH cases.
    .pad <- vapply(1:60, function(i) sprintf("v%d=exp(THETA[1])*%d", i, i), character(1))

    # (a) valid model whose chunk orphans a compartment-scoped line.  `f(depot) <- ...`
    # (assignment with `<-`) is accepted by rxode2, but the disguise only splits a line at
    # its first "=", so it is not disguised and its chunk cannot be optimized on its own.
    .frag <- paste(c("d/dt(depot)=-exp(THETA[1])*depot",
                     "d/dt(center)=exp(THETA[1])*depot-exp(THETA[2])*center",
                     .pad, "f(depot) <- exp(THETA[3])", "rx_pred_=center"), collapse = "\n")
    .out <- suppressMessages(rxOptExpr(.frag, "model", chunkLines = 40L))
    expect_identical(.out, suppressMessages(rxOptExpr(.frag, "model", chunkLines = 0L)))
    expect_true(grepl("rx_expr_", .out))   # fell back, so it is still fully optimized
    expect_error(rxModelVars(.out), NA)

    # (b) genuinely malformed model: chunking must raise, exactly as the unchunked call does
    .bad <- paste(c("d/dt(depot)=-ka*depot", "y = = 3 +", .pad), collapse = "\n")
    suppressMessages(expect_error(rxOptExpr(.bad, "model")))
    suppressMessages(expect_error(rxOptExpr(.bad, "model", chunkLines = 40L)))
  })

  test_that("parallel chunking gives the identical model and leaves no daemons behind", {
    # daemons only start when the model splits into at least 4 chunks, so use a model
    # large enough to cross that threshold
    .m <- .chunkModel(150L)
    .seq <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 40L, parallel = 1L))
    .par <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 40L, parallel = 2L))
    # optimizing the chunks in daemons must not change the model at all
    expect_identical(.seq, .par)
    # the pool is started for the call and shut down when it returns
    expect_equal(sum(mirai::status()$connections), 0L)
    # the default parallel (0 = the rxode2 thread setting) is the same model too,
    # and equally leaves no pool behind
    .def <- suppressMessages(rxOptExpr(.m, "model"))
    expect_identical(.def, .seq)
    expect_equal(sum(mirai::status()$connections), 0L)
  })

  test_that("too few chunks do not pay for a daemon pool", {
    # .chunkModel() splits into only 2 chunks: even with parallel forced on, no pool is
    # started (its startup would cost more than it saves) and none is left behind
    .m <- .chunkModel()
    .par <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 40L, parallel = 2L))
    .seq <- suppressMessages(rxOptExpr(.m, "model", chunkLines = 40L, parallel = 1L))
    expect_identical(.par, .seq)
    expect_equal(sum(mirai::status()$connections), 0L)
  })
})

rxTest({
  # past(state, tau) <- expr : user-specified non-constant delay() pre-history.
  # delay(state, tau) returns expr evaluated at t-tau for t<t0 instead of the
  # constant initial condition.

  test_that("past() parses and round-trips through rxNorm", {
    .m <- rxode2("y(0) <- 1\nd/dt(y) <- -delay(y, 1)\npast(y, 1) <- exp(0.5*t)\n")
    .n <- rxNorm(.m)
    expect_true(grepl("past(y,1)=exp(0.5 * t)", .n, fixed = TRUE) ||
                grepl("past(y,1)=exp(0.5*t)", .n, fixed = TRUE))
  })

  test_that("constant past() value drives the pre-history (not the IC)", {
    .ev <- et(seq(0, 1, by = 0.25))
    ## for t<1, delay(y,1) = past value; y' = -past
    .p2 <- rxSolve(rxode2("y(0) <- 1\nd/dt(y) <- -delay(y, 1)\npast(y, 1) <- 2\n"), .ev)
    .p0 <- rxSolve(rxode2("y(0) <- 1\nd/dt(y) <- -delay(y, 1)\npast(y, 1) <- 0.5\n"), .ev)
    ## past=2 -> y decreases by 2 per unit time; past=0.5 -> by 0.5
    expect_equal(.p2$y, 1 - 2 * .p2$time, tolerance = 1e-6)
    expect_equal(.p0$y, 1 - 0.5 * .p0$time, tolerance = 1e-6)
  })

  test_that("non-constant past() matches the analytic solution", {
    ## y'(t) = -delay(y,1), history h(t)=exp(0.5 t); for t<=1 the delayed value
    ## is exp(0.5(t-1)), so y(t) = 1 - 2*(exp(0.5(t-1)) - exp(-0.5)).
    .ev <- et(seq(0, 1, by = 0.05))
    .s <- rxSolve(rxode2("y(0) <- 1\nd/dt(y) <- -delay(y, 1)\npast(y, 1) <- exp(0.5*t)\n"),
                  .ev, atol = 1e-10, rtol = 1e-10)
    .ana <- 1 - 2 * (exp(0.5 * (.ev$get.EventTable()$time - 1)) - exp(-0.5))
    expect_equal(.s$y, .ana, tolerance = 1e-6)
  })

  test_that("past() first-order sensitivities match finite differences", {
    .ev <- et(seq(0, 3, by = 0.25))
    .mstr <- function(k, a, b) sprintf(
      "k=%.12g\na=%.12g\nb=%.12g\ny(0) <- a\nd/dt(y) <- -k*delay(y, 1)\npast(y, 1) <- a*exp(b*t)\n",
      k, a, b)
    .s <- rxSolve(rxode2(.mstr(0.3, 1, 0.5), calcSens = c("k", "a", "b")),
                  .ev, atol = 1e-11, rtol = 1e-11)
    .fwd <- function(k, a, b) rxSolve(rxode2(.mstr(k, a, b)), .ev,
                                      atol = 1e-11, rtol = 1e-11)$y
    .eps <- 1e-4
    .base <- c(k = 0.3, a = 1, b = 0.5)
    for (.p in c("k", "a", "b")) {
      .hi <- .base; .hi[.p] <- .hi[.p] + .eps
      .lo <- .base; .lo[.p] <- .lo[.p] - .eps
      .fd <- (.fwd(.hi[1], .hi[2], .hi[3]) - .fwd(.lo[1], .lo[2], .lo[3])) / (2 * .eps)
      expect_equal(.s[[paste0("rx__sens_y_BY_", .p, "__")]], .fd, tolerance = 1e-4)
    }
  })

  test_that("past() history is validated", {
    .ev <- et(seq(0, 2, by = 0.5))
    ## history may not reference an ODE state
    expect_error(rxSolve(rxode2("y(0)<-1\nd/dt(y)<- -delay(y,1)\npast(y,1)<-0.5*y\n"), .ev),
                 "may not reference ODE state")
    ## duration must match the delay
    expect_error(rxSolve(rxode2("y(0)<-1\nd/dt(y)<- -delay(y,2)\npast(y,1)<-exp(0.5*t)\n"), .ev),
                 "does not match any delay")
    ## past on a non-delayed state (while another state is delayed) is rejected
    expect_error(rxSolve(rxode2("y(0)<-1\nz(0)<-1\nd/dt(y)<- -y\nd/dt(z)<- -delay(z,1)\npast(y,1)<-exp(t)\n"), .ev),
                 "has no delay")
    ## a valid history solves without error
    .s <- suppressMessages(rxSolve(rxode2("y(0)<-1\nd/dt(y)<- -delay(y,1)\npast(y,1)<-exp(0.5*t)\n"), .ev))
    expect_true(all(is.finite(.s$y)))
  })

  test_that("past() second-order sensitivities match finite differences", {
    .ev <- et(seq(0, 3, by = 0.5))
    .mstr <- function(k, a, b) sprintf(
      "k=%.12g\na=%.12g\nb=%.12g\ny(0) <- a\nd/dt(y) <- -k*delay(y, 1)\npast(y, 1) <- a*exp(b*t)\n",
      k, a, b)
    .pars <- c("k", "a", "b")
    .s <- rxSolve(rxode2(.mstr(0.3, 1, 0.5), calcSens = .pars, calcSens2 = .pars),
                  .ev, atol = 1e-12, rtol = 1e-12)
    .fwd <- function(m) rxSolve(rxode2(.mstr(m[1], m[2], m[3])), .ev,
                                atol = 1e-12, rtol = 1e-12)$y
    .base <- c(k = 0.3, a = 1, b = 0.5); names(.base) <- .pars
    .h <- 1e-3
    .same <- function(p) {
      .hi <- .base; .hi[p] <- .hi[p] + .h; .lo <- .base; .lo[p] <- .lo[p] - .h
      (.fwd(.hi) - 2 * .fwd(.base) + .fwd(.lo)) / (.h * .h)
    }
    .cross <- function(p, q) {
      .pp <- .base; .pp[p] <- .pp[p] + .h; .pp[q] <- .pp[q] + .h
      .pm <- .base; .pm[p] <- .pm[p] + .h; .pm[q] <- .pm[q] - .h
      .mp <- .base; .mp[p] <- .mp[p] - .h; .mp[q] <- .mp[q] + .h
      .mm <- .base; .mm[p] <- .mm[p] - .h; .mm[q] <- .mm[q] - .h
      (.fwd(.pp) - .fwd(.pm) - .fwd(.mp) + .fwd(.mm)) / (4 * .h * .h)
    }
    for (.i in seq_along(.pars)) for (.j in .i:length(.pars)) {
      .p <- .pars[.i]; .q <- .pars[.j]
      .col <- paste0("rx__sens_y_BY_", .p, "_BY_", .q, "__")
      if (is.null(.s[[.col]])) .col <- paste0("rx__sens_y_BY_", .q, "_BY_", .p, "__")
      .fd <- if (.p == .q) .same(.p) else .cross(.p, .q)
      expect_equal(.s[[.col]], .fd, tolerance = 1e-3)
    }
  })

  # past(state, tau) without a d/dt(state) is a user error.  Reporting it used to
  # append nothing to the error buffer and then trim a trailing "', " that was
  # never written, moving the write offset before the start of the buffer; the
  # heap corruption surfaced as a double free on a *later*, unrelated parse.
  test_that("past() without d/dt() errors cleanly and leaves the parser usable", {
    .bad <- "a=1\nb=0.5\nT=12.8\npast(G, T) = a*exp(b*t)\nR2 = a\n"
    expect_error(suppressMessages(rxModelVars(.bad)), "syntax error")
    # the message names the offending property rather than a mangled fragment
    expect_output(try(suppressMessages(rxModelVars(.bad)), silent = TRUE),
                  "'past(G)' present, but d/dt(G) not defined", fixed = TRUE)
    # parsing again, and parsing a good model afterwards, must still work
    expect_error(suppressMessages(rxModelVars(.bad)), "syntax error")
    expect_true(inherits(rxModelVars("d/dt(G)=-G\n"), "rxModelVars"))
  })

  # A past() line is compartment-scoped: it only parses in a chunk that also has
  # the matching d/dt().  .rxDelaySensAugment() appends past() after every d/dt(),
  # so on a model long enough to chunk it lands in the last chunk, away from its
  # d/dt() -- .rxDisguiseCmt() must hide it for the chunk to parse standalone.
  test_that("past() survives chunked expression optimization", {
    .filler <- paste(sprintf("v%d = %d * exp(kg * t) + sin(v0 * %d)", 1:45, 1:45, 1:45),
                     collapse = "\n")
    .mod <- paste0(
      "a = 1\nb = 0.5\nT = 12.8\nk3 = 5\nkg = 0.4\nk4 = 0.3\nk5 = 0.1\nv0 = 2\n",
      "G(0) = a\nd/dt(G) = k3 - kg * G\n",
      "I(0) = 0\nd/dt(I) = k4 * G - k4 * delay(G, T)\n",
      "D(0) = 0\nd/dt(D) = k4 * delay(G, T) - k5 * D\n",
      .filler, "\n", "R2 = D + v1 + v45\n",
      "past(G, T) = a * exp(b * t)\n")

    .dg <- rxode2:::.rxDisguiseCmt(.mod)
    # the past line is hidden as an ordinary assignment, so a chunk parses alone
    expect_false(any(grepl("^past", strsplit(.dg, "\n")[[1]])))
    expect_true(any(grepl("^rx__disg_lhs__", strsplit(.dg, "\n")[[1]])))
    # ... and is restored byte-exactly, spacing included
    expect_equal(strsplit(rxode2:::.rxRestoreCmt(.dg), "\n")[[1]],
                 strsplit(.mod, "\n")[[1]])

    .o <- suppressMessages(rxOptExpr(.mod, "m", chunkLines = 40L))
    expect_true(any(grepl("^past\\(G, *T\\)", strsplit(.o, "\n")[[1]])))
    expect_false(grepl("rx__disg_", .o, fixed = TRUE))
    # optimizing must not change what the model means
    .ev <- et(seq(0, 30, by = 1))
    .s1 <- rxSolve(rxode2(.mod), .ev, method = "dop853", atol = 1e-10, rtol = 1e-10,
                   dense = TRUE)
    .s2 <- rxSolve(rxode2(.o), .ev, method = "dop853", atol = 1e-10, rtol = 1e-10,
                   dense = TRUE)
    expect_equal(.s1$R2, .s2$R2, tolerance = 1e-8)
  })
})

.codeLoaded <- function() {
  .Call(`_rxode2_codeLoaded`)
}

.codegen <- function(c_file, prefix, libname, pMd5, timeId, lastMv, goodFun,
                     eventSensCode = rep("", 13L)) {
  .Call(`_rxode2_codegen`, c_file, prefix, libname, pMd5, timeId, lastMv, goodFun,
        eventSensCode[1], eventSensCode[2], eventSensCode[3], eventSensCode[4],
        eventSensCode[5], eventSensCode[6], eventSensCode[7], eventSensCode[8],
        eventSensCode[9], eventSensCode[10], eventSensCode[11], eventSensCode[12],
        eventSensCode[13])
}

.parseModel <- function(type) {
  .Call(`_rxode2_parseModel`, type)
}

.isLinCmt <- function() {
  .Call(`_rxode2_isLinCmt`)
}

.trans <- function(parse_file, prefix, model_md5, parseStr, isEscIn, inME, goodFuns, fullPrintIn) {
  .Call(`_rxode2_trans`,
        parse_file, prefix, model_md5, parseStr, isEscIn, inME, goodFuns, fullPrintIn)
}

.linCmtParse <- function(vars, inStr, verbose) {
  .Call(`_linCmtParse`, vars, inStr, verbose)
}

.linCmtGen <- function(linCmt, vars, linCmtSens, verbose) {
  .Call(`_rxode2_linCmtGen`, linCmt, vars, linCmtSens, verbose)
}

.parseFreeSexp  <- function(last) {
  .Call(`_rxode2_parseFreeSexp`, last)
}

## .calcDerived <- function(ncmtSXP, transSXP, inp, sigdigSXP) {
##   .Call(`_calcDerived`, ncmtSXP, transSXP, inp, sigdigSXP)
## }

#' As rxode2 ui
#'
#' @param x Object to convert to `rxUi` object
#' 
#' @return rxUi object (or error if it cannot be converted)
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' mod1 <- function() {
#'  ini({
#'    # central 
#'    KA=2.94E-01
#'    CL=1.86E+01
#'    V2=4.02E+01
#'    # peripheral
#'    Q=1.05E+01
#'    V3=2.97E+02
#'    # effects
#'    Kin=1
#'    Kout=1
#'    EC50=200 
#'  })
#'  model({
#'    C2 <- centr/V2
#'    C3 <- peri/V3
#'    d/dt(depot) <- -KA*depot
#'    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
#'    d/dt(peri)  <- Q*C2 - Q*C3
#'    eff(0) <- 1
#'    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
#'  })
#' }
#' 
#' as.rxUi(mod1)
#' 
as.rxUi <- function(x) {
  UseMethod("as.rxUi")
}
#' @rdname as.rxUi
#' @export
as.rxUi.rxode2 <- function(x) {
  as.rxUi(as.function(x))
}

#' @rdname as.rxUi
#' @export
as.rxUi.rxode2tos <- function(x) {
  suppressMessages(as.rxUi(as.function(x)))
}
#' @export
#' @rdname as.rxUi
as.rxUi.rxModelVars <- function(x) {
  as.rxUi(as.function(x))
}
#' @export
#' @rdname as.rxUi
as.rxUi.function <- function(x) {
  model <- try(rxode2(x), silent=TRUE)
  if (inherits(model, "try-error")) {
    stop("cannot convert to rxUi object", call.=FALSE)
  }
  model
}
#' @export
#' @rdname as.rxUi 
as.rxUi.rxUi <- function(x) {
  x
}
#' @export
#' @rdname as.rxUi
as.rxUi.default <- function(x) {
  stop("could not convert to rxUi object", call.=FALSE)
}

#' A description of Rode2 supported residual errors
#'
#' @format A data frame with 6 columns and 183 rows
#' \describe{
#' \item{Error model}{A description of the type of residual error}
#' \item{Functional Form}{For additive and proportional what functional form is used}
#' \item{Transformation}{The type of transformation that is done on the DV and the prediction}
#' \item{code}{Example code for the residual error type}
#' \item{addProp}{The type of add+prop residual error default that would be equivalent}
#' \item{lhs}{what the left handed side of the specification represents, either a response variable, or a compartment specification}
#' }
#' @references
#' The transformation-based and autocorrelated (`ar()`, AR(1)) residual
#' error models follow Karlsson MO, Beal SL, Sheiner LB. Three new residual
#' error models for population PK/PD analyses. J Pharmacokinet Biopharm.
#' 1995;23(6):651-672. doi:10.1007/BF02353466.
"rxResidualError"

## nocov start

genDefine <- function() {

  mod1 <- rxode2parse("
    C2 = centr/V2
    C3 = peri/V3
    d/dt(depot) =-KA*depot
    alag(depot) = 3
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3
    d/dt(peri)  =                    Q*C2 - Q*C3
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff
  ")

  mod <- rxode2parse("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
d/dt(blood)     = a*intestine - b*blood
")

  mv <- mod1

  .ctl <- rxode2::rxControl()

  .n <- gsub("[.]","_",names(.ctl))
  sink(devtools::package_file("inst/include/rxode2parse_control.h")) # nolint
  cat("#pragma once\n")
  cat("#ifndef __rxode2parse_control_H__\n#define __rxode2parse_control_H__\n")
  cat(paste(paste0("#define ", "Rxc_", .n, " ", seq_along(.n)-1),collapse="\n"))

  .mv <- mod1

  .nmv <- gsub("[.]", "_", names(.mv))
  cat("\n")
  cat(paste(paste0("#define RxMv_", .nmv, " ", seq_along(.nmv)-1),collapse="\n"))
  .nmvf <- names(.mv$flag)
  cat("\n")
  cat(paste(paste0("#define RxMvFlag_", .nmvf, " ", seq_along(.nmvf)-1),collapse="\n"))
  cat("\n")

  .nmvt <- gsub("[.]", "_", names(.mv$trans))

  cat("\n")
  cat(paste(paste0("#define RxMvTrans_", .nmvt, " ",
                   seq_along(.nmvt)-1),collapse="\n"))
  cat("\n")

  et <- structure(list(time = c(0, 0.05, 0.1, 0.2, 0.3, 0.5), cmt = c("(default)", "(obs)", "intestine", "-intestine", "intestine", "out"), amt = c(0.0833333333333333, NA, 3, NA, 3, 3), rate = c(2, 0, 0, 0, 0, 0), ii = c(1, 0, 3, 0, 3, 0), addl = c(9L, 0L, 0L, 0L, 0L, 0L), evid = c(1L, 2L, 1L, 2L, 1L, 1L), ss = c(0L, 0L, 1L, 0L, 2L, 0L)), class = "data.frame", row.names = c(NA, -6L))

  ett1 <- etTrans(et, mod, keepDosingOnly=TRUE)
  .n <- gsub("[.]", "_", names(attr(class(ett1), ".rxode2")))

  cat(paste(paste0("#define RxTrans_", .n, " ", seq_along(.n)-1),collapse="\n"))
  cat(paste0("\n#define RxTransNames CharacterVector _en(", length(.n), ");",
             paste(paste0("_en[",seq_along(.n)-1,']="', .n, '";'), collapse=""),"e.names() = _en;"))
  cat("\n")
  cat("\n#endif // __rxode2parse_control_H__\n")
  sink() # nolint
}

.generateRandomUiFun <- function(fun, args, vals=NULL) {
  .ret <- as.call(c(
    quote(`{`),
    do.call(`c`,
            lapply(args, function(arg) {
              .dotArg <- str2lang(paste0(".", arg))
              .arg <- str2lang(arg)
              .ret <- bquote({
                .(.dotArg) <- as.character(substitute(.(.arg)))
                .dp <- deparse1(substitute(.(.arg)))
                .tmp <- suppressWarnings(try(force(.(.arg)), silent=TRUE))
                .(.dotArg) <- .uiArg(.(.dotArg), .tmp, .dp)
              })
              lapply(seq_along(.ret)[-1], function(i) {
                .ret[[i]]
              })
            })),
    str2lang(paste0("list(replace=paste0('", fun, "(', ", paste(paste0(".", args), collapse=", ', ', "), ", ')'))"))
  ))
  if (is.null(vals)) {
    .f <- paste0(".", fun, " <- function(", paste(args, collapse=", "), ") ",
                 paste(deparse(.ret), collapse="\n"),
                 "\n")
  } else {
    .f <- paste0(".", fun, " <- function(", paste0(paste0(args, ifelse(is.na(vals), "", " = "), ifelse(is.na(vals), "", vals)),
                                                  collapse=", "), ") ",
                 paste(deparse(.ret), collapse="\n"),
                 "\n")
  }
  .f <- paste0(.f, "\n#'@export\nrxUdfUi.", fun, " <- rxUdfUi.rxpois\n\n")
  .f
}

.generateRandomUiFuns <- function() {
  .lst <- list("rxnorm"=c("mean"=0, "sd"=1),
               "rxpois"="lambda",
               "rxt"="df",
               "rxunif"=c("min"=0, "max"=1),
               "rxweibull"=c("shape"=NA, "scale"=1),
               "rxgeom"="prob",
               "rxbeta"=c("shape1", "shape2"),
               "rxgamma"=c("shape"=NA, "rate"=1),
               "rxf"=c("df1","df2"),
               "rxexp"="rate",
               "rxchisq"="df",
               "rxcauchy"=c(location = 0, scale = 1),
               "rxbinom"=c("size", "prob"))
  .lst2 <- .lst
  names(.lst2) <- gsub("rx", "ri", names(.lst2))
  .lst <- c(.lst, .lst2,
            list("logit"=c("x"=NA, "low"=0, "high"=1),
                 "expit"=c("x"=NA, "low"=0, "high"=1)))
  paste0("# This file is generated by .generateRandomUiFuns() in build.R\n## nocov start\n",
         vapply(names(.lst), function(fun) {
           if (is.null(names(.lst[[fun]]))) {
             .generateRandomUiFun(fun, .lst[[fun]])
           } else {
             .generateRandomUiFun(fun, names(.lst[[fun]]), .lst[[fun]])
           }
         }, character(1), USE.NAMES=FALSE),
         "## nocov end")
}


.rxodeBuildCode <- function() {
  # This builds the code needed for rxode2
  message("Generate rxode2 random named arguments option")

  .l <- .generateRandomUiFuns()
  .R <- file(devtools::package_file("R/rxrandomui.R"), "wb")
  writeLines(.l, .R)
  close(.R)

  message("done")
  message("Generate grammar include file")
  dparser::mkdparse(devtools::package_file("inst/tran.g"),
                    devtools::package_file("src/"),
                    grammar_ident="rxode2parse")
  l <- readLines(devtools::package_file("src/tran.g.d_parser.c"))
  .w <- which(grepl("#line ", l))
  if (.w > 1L) {
    .w <- .w[1L]
    l[.w] <- sub("[#]line([^\"]*\").*(src.*)", "#line\\1\\2",l[.w])
  }
  tran.g.h <- file(devtools::package_file("src/tran.g.d_parser.h"), "wb")
  writeLines(l, tran.g.h)
  close(tran.g.h)
  unlink(devtools::package_file("src/tran.g.d_parser.c"))


  # generate control
  try({
    message("generate defines")
    sink(devtools::package_file("inst/include/rxode2_control.h"))
    cat("#pragma once\n")
    cat("#ifndef __rxode2_control_H__\n#define __rxode2_control_H__\n")
    cat('#include "rxode2parse_control.h"\n')
    cat("\n#endif // __rxode2_control_H__\n")
    sink()
    message("Copy header to inst directory")
    file.copy(devtools::package_file("src/rxode2_types.h"),
              devtools::package_file("inst/include/rxode2_types.h"),
              overwrite=TRUE)
    .createRxUiBlessedList()
  },silent=TRUE)
  message("generate rxResidualError and update documentation")
  rxResidualError <- utils::read.csv(devtools::package_file("inst/residualErrors.csv"),
                              check.names=FALSE)
  usethis::use_data(rxResidualError, overwrite = TRUE)
  .l <- readLines(devtools::package_file("R/rxResidualError.R"))
  .l <- sub("[#][']\\s*@format\\s*.*",
            sprintf("#' @format A data frame with %d columns and %d rows",
                    dim(rxResidualError)[2], dim(rxResidualError)[1]), .l)
  .R <- file(devtools::package_file("R/rxResidualError.R"), "wb")
  writeLines(.l, .R)
  close(.R)
  message("done")
  message("generate rxReservedKeywords and update documentation")
  rxReservedKeywords <- utils::read.csv(devtools::package_file("inst/reserved-keywords.csv"))
  names(rxReservedKeywords)[1] <- "Reserved Name"
  usethis::use_data(rxReservedKeywords, overwrite=TRUE)
  .l <- readLines(devtools::package_file("R/rxReservedKeywords.R"))
  .l <- sub("[#][']\\s*@format\\s*.*",
            sprintf("#' @format A data frame with %d columns and %d rows",
                    dim(rxReservedKeywords)[2], dim(rxReservedKeywords)[1]), .l)
  .R <- file(devtools::package_file("R/rxReservedKeywords.R"), "wb")
  writeLines(.l, .R)
  close(.R)
  message("generate rxSyntaxFunctions and update documentation")
  rxSyntaxFunctions <- utils::read.csv(devtools::package_file("inst/syntax-functions.csv"))
  usethis::use_data(rxSyntaxFunctions, overwrite=TRUE)
  .l <- readLines(devtools::package_file("R/rxSyntaxFunctions.R"))
  .l <- sub("[#][']\\s*@format\\s*.*",
            sprintf("#' @format A data frame with %d columns and %d rows",
                    dim(rxSyntaxFunctions)[2], dim(rxSyntaxFunctions)[1]), .l)
  .R <- file(devtools::package_file("R/rxSyntaxFunctions.R"), "wb")
  writeLines(.l, .R)
  close(.R)
  message("done")
  message("generate rxode2_control.h")
  .n <- gsub("[.]","_",names(rxControl()))
  sink(devtools::package_file("inst/include/rxode2_control.h"))
  cat("#pragma once\n")
  cat("#ifndef __rxode2_control_H__\n#define __rxode2_control_H__\n")
  cat('#include "rxode2parse_control.h"\n')
  cat(paste(paste0("#define ", "Rxc_", .n, " ", seq_along(.n)-1),collapse="\n"))
  cat("\n#endif // __rxode2_control_H__\n")
  sink()
  message("done")
  return(invisible(""))
}
#' This creates the list of "blessed" rxode2 items
#'
#' @return nothing, called for side effects
#' @noRd
#' @author Matthew L. Fidler
.createRxUiBlessedList <- function() {
  message("querying default rxode2 object contents")
  tka <- log(1.57)
  tcl <- log(2.72)
  tv <- log(31.5)
  tv2 <- 3
  eta.ka <- 0.6
  eta.cl <- 0.3
  eta.v <- 0.1
  add.sd <- 0.7
  depot <- center <- NULL
  `/<-` <- function(...) {} # nolint
  dt <- function(...) {} #nolint
  .f <- function() {
    ini({
      tka <- log(1.57)
      tcl <- log(2.72)
      tv <- log(31.5)
      tv2 <- 3
      eta.ka ~ 0.6
      eta.cl ~ 0.3
      eta.v ~ 0.1
      add.sd <- 0.7
    })
    model({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      v2 <- tv2
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl / v * center
      cp = center / v
      cp ~ add(add.sd)
    })
  }
  .f <- .f()
  .f <- rxUiDecompress(.f)
  .blessed <- sort(unique(c("model", "modelName", ls(.f, all.names=TRUE))))
  .blessed <- deparse(str2lang(paste0(".rxUiBlessed <- ",
                      paste(deparse(.blessed), collapse="\n"))))
  writeLines(c("## created by .createRxUiBlessedList() in ui-assign-parts.R edit there",
               .blessed), devtools::package_file("R/rxUiBlessed.R"))
  message("saved!")
  message("Generating parseFuns.R")
  .var <- deparse(rxSupportedFuns())
  .num <- deparse(.rxSEeq)
  .var[1] <- paste0(".parseEnv$.parseFuns <- ", .var[1])
  .num[1] <- paste0(".parseEnv$.parseNum <- ", .num[1])
  .pf <- devtools::package_file("R/parseFuns.R")
  unlink(.pf)
  parseFuns.R <- file(.pf, "wb")
  writeLines(.var, parseFuns.R)
  writeLines(.num, parseFuns.R)
  close(parseFuns.R)

  message("rebuild rxode2parse_control.h")
  genDefine()
  message("done")
  invisible("")
}


## nocov end

#' Extract model lines from a rxui model
#'
#' @param x model to extract lines from
#'
#' @param ... variables to extract. When it is missing, it will
#'   extract the entire model (conditioned on the endpoint option
#'   below)
#'
#' @param expression return expressions (if `TRUE`) or strings (if
#'   `FALSE`)
#'
#' @param endpoint include endpoint.  This can be:
#'
#'  - `NA`    -- Missing means include both the endpoint and non-endpoint lines
#'
#'  - `TRUE`  -- Only include endpoint lines
#'
#'  - `FALSE` -- Only include non-endpoint lines
#'
#' @param lines is a boolean.  When `TRUE` this will add the lines as
#'   an attribute to the output value ie `attr(, "lines")`
#'
#' @param envir Environment for evaluating variables
#'
#' @return expressions or strings of extracted lines. Note if there is
#'   a duplicated lhs expression in the line, it will return both
#'   lines
#'
#' @export
#'
#' @author Matthew L. Fidler
#'
#' @examples
#'
#' one.compartment <- function() {
#'   ini({
#'     tka <- 0.45 # Log Ka
#'     tcl <- 1 # Log Cl
#'     tv <- 3.45    # Log V
#'     eta.ka ~ 0.6
#'     eta.cl ~ 0.3
#'     eta.v ~ 0.1
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl + eta.cl)
#'     v  <- exp(tv + eta.v)
#'     d/dt(depot)  <- -ka * depot
#'     d/dt(center) <-  ka * depot - cl / v * center
#'     cp <- center / v
#'     cp ~ add(add.sd)
#'   })
#'  }
#'
#'  f <- one.compartment()
#'
#'  modelExtract(f, cp)
#'
#'  modelExtract(one.compartment, d/dt(depot))
#'
#'  # from variable
#'  var <- "d/dt(depot)"
#'
#'  modelExtract(one.compartment, var)
#'
#'  modelExtract(f, endpoint=NA, lines=TRUE, expression=TRUE)
#'
modelExtract <- function(x, ..., expression=FALSE, endpoint=FALSE, lines=FALSE, envir=parent.frame()) {
  checkmate::assertLogical(expression, any.missing=FALSE, len=1)
  checkmate::assertLogical(lines, any.missing=FALSE, len=1)
  checkmate::assertLogical(endpoint, any.missing=TRUE, len=1)
  UseMethod("modelExtract")
}
#' Common extract model lines
#'
#' @param modelLines Model lines, in this case it is the variables
#' @param rxui rxode2 parsed ui
#' @param expression Should an expression list be returned
#' @param endpoint Should this be an endpoint (yes: TRUE, no: FALSE, both: TRUE)
#' @noRd
#' @author Matthew L. Fidler
.modelExtractCommon <- function(modelLines, rxui, expression=FALSE, endpoint=FALSE, lines=FALSE) {
  .lstExpr <- rxui$lstExpr
  .isNull <- length(modelLines) == 0L ||
    all(vapply(seq_along(modelLines),
               function(i) {
                 is.null(modelLines[[i]])
               }, logical(1)))
  if (.isNull) {
    .ret <- seq_along(.lstExpr)
  } else {
    .ret <- do.call(`c`, lapply(seq_along(modelLines),
                                function(i) {
                                  .w <- .getModelLineFromExpression(modelLines[[i]],
                                                                    rxui, errorLine=FALSE,
                                                                    returnAllLines=TRUE)
                                  .w <- .w[.w>0]
                                  .w
                                }))

  }
  .ret <- sort(unique(.ret))
  .endPointLines <- rxui$predDf
  if (!is.null(.endPointLines)) {
    .endPointLines <- .endPointLines$line
    if (is.na(endpoint)) {
      # do both
    } else if (endpoint) {
      .ret <- .ret[.ret %in% .endPointLines]
    } else {
      .ret <- .ret[!(.ret %in% .endPointLines)]
    }
  }
  .lines <- .ret
  .ret <- lapply(.ret,
                 function(i) {
                   .lstExpr[[i]]
                 })
  if (!expression) {
    .ret <- vapply(seq_along(.ret),
                   function(i) {
                     deparse1(.ret[[i]])
                   }, character(1))
  }
  if (lines) {
    attr(.ret, "lines") <- .lines
  }
  .ret
}
#' Quote call info for extraction variables
#'
#' @param callInfo Call information
#' @return list of expressions
#' @noRd
#' @author Matthew L. Fidler
.quoteCallVars <- function(callInfo, ..., envir=parent.frame()) {
  if (length(callInfo) == 0L) return(NULL)
  .env <- new.env(parent=emptyenv())
  .env$alag <- list()
  .env$lag <- list()
  c(lapply(seq_along(callInfo),
         function(i) {
           .name <- names(callInfo)[i]
           .cur <- callInfo[[i]]
           if (is.name(.cur)) {
             .curChar <- as.character(.cur)
             if (exists(.curChar, envir=envir)) {
               .cur <- get(.curChar, envir=envir)
             }
           }
           if (is.list(.cur)) {
             .tmp <- eval(.cur)
             .tmp <- as.vector(.cur)
             .tmp <- setNames(unlist(.tmp), NULL)
             .cur <- .tmp
           }
           if (is.call(.cur) &&
                 identical(.cur[[1]], quote(`$`))) {
             .list <- list(...)
             .tmp <- .list[[i]]
             .cur <- .tmp
           }
           if (inherits(.cur, "character")) {
             .cur <- str2lang(.cur)
           }
           if (is.null(.name)) {
           } else if (.name %in% c("expression",  "endpoint", "envir", "lines")) {
             return(NULL)
           }
           if (is.name(.cur)) {
             return(str2lang(paste0("-",deparse1(.cur))))
           } else if (is.call(.cur) &&
                        (.matchesLangTemplate(.cur, str2lang("d/dt(.name)")) ||
                           .matchesLangTemplate(.cur, str2lang("f(.name)")) ||
                           .matchesLangTemplate(.cur, str2lang(".name(0)")) ||
                           .matchesLangTemplate(.cur, str2lang("rate(.name)")) ||
                           .matchesLangTemplate(.cur, str2lang("dur(.name)")))) {
             return(str2lang(paste0("-", deparse1(.cur))))
           } else if (is.call(.cur) &&
                        .matchesLangTemplate(.cur, str2lang("alag(.name)"))) {
             .env$lag <- c(.env$lag,
                           list(str2lang(paste0("-", sub("alag", "lag", deparse1(.cur))))))
              return(str2lang(paste0("-", deparse1(.cur))))
           } else if (is.call(.cur) &&
                        .matchesLangTemplate(.cur, str2lang("lag(.name)"))) {
             .env$alag <- c(.env$alag,
                            list(str2lang(paste0("-", sub("lag", "alag", deparse1(.cur))))))
             return(str2lang(paste0("-", deparse1(.cur))))
           }
           stop("unknown variable expression: ", deparse1(.cur),
                call.=FALSE)
         }),
    .env$alag,
    .env$lag)
}

#' @export
#' @rdname modelExtract
modelExtract.function <- function(x, ..., expression=FALSE, endpoint=FALSE, lines=FALSE, envir=parent.frame()) {
  .modelLines <- .quoteCallVars(match.call(expand.dots = TRUE)[-(1:2)], ..., envir=envir)
  .ret <- rxode2(x)
  .modelExtractCommon(.modelLines, .ret, expression=expression, endpoint=endpoint, lines=lines)
}
#' @export
#' @rdname modelExtract
modelExtract.rxUi <- function(x, ..., expression=FALSE, endpoint=FALSE, lines=FALSE, envir=parent.frame()) {
  .modelLines <- .quoteCallVars(match.call(expand.dots = TRUE)[-(1:2)], ..., envir=envir)
  .modelExtractCommon(.modelLines, x, expression=expression, endpoint=endpoint, lines=lines)
}
#' @export
#' @rdname modelExtract
modelExtract.rxode2 <- function(x, ..., expression=FALSE, endpoint=FALSE, lines=FALSE, envir=parent.frame()) {
  .modelLines <- .quoteCallVars(match.call(expand.dots = TRUE)[-(1:2)], ..., envir=envir)
  x <- as.function(x)
  .ret <- rxode2(x)
  .modelExtractCommon(.modelLines, .ret, expression=expression, endpoint=endpoint, lines=lines)
}
#' @export
#' @rdname modelExtract
modelExtract.rxModelVars <- modelExtract.rxode2

#' @export
#' @rdname modelExtract
modelExtract.default <- function(x, ..., expression=FALSE, endpoint=FALSE, lines=FALSE, envir=parent.frame()) {
  stop("rxode2 does not know how to handle this modelExtract object",
       call.=FALSE)
}

#' Get the lambda value based on the pred information
#'
#' @param env Environment that has the environment
#' @param pred1 Single error data frame
#' @return Lambda expression
#' @author Matthew Fidler
#' @keywords internal
#' @export
.rxGetLambdaFromPred1AndIni <- function(env, pred1) {
  if (!is.na(pred1$lambda)) {
    return(str2lang(pred1$lambda))
  }
  if (.rxTransformHasALambdaParameter(pred1$transform)) {
    .cnd <- pred1$cond
    .w <- which(env$iniDf$err %in% c("boxCox", "yeoJohnson") & env$iniDf$condition == .cnd)
    if (length(.w) == 1L) {
      return(str2lang(env$iniDf$name[.w]))
    } else {
      stop("cannot find lambda", call.=FALSE)
    }
  }
  return(1)
}
#' Get the lower boundary condition when the transformation requires it
#'
#' @param env  Environment for the parsed model
#' @param pred1 The `data.frame` of the current error
#' @return Lower Boundary
#' @author Matthew Fidler
#' @keywords internal
#' @export
.rxGetLowBoundaryPred1AndIni <- function(env, pred1) {
  if (.rxTransformHasBounds(pred1$transform)) {
    return(pred1$trLow)
  }
  return(0)
}
#' Get the upper boundary condition when the transformation it
#'
#' @param env Environment for the parsed model
#' @param pred1 The `data.frame` of the current error
#' @return Upper Boundary
#' @author Matthew Fidler
#' @keywords internal
#' @export
.rxGetHiBoundaryPred1AndIni <- function(env, pred1) {
  if (.rxTransformHasBounds(pred1$transform)) {
    return(pred1$trHi)
  }
  return(1)
}

#' Get the prediction name
#'
#' @param env Environment for the parsed model
#' @param pred1 The `data.frame` of the current error
#' @return The prediction symbol
#' @author Matthew Fidler
#' @export
#' @keywords internal
.rxGetPredictionF <- function(env, pred1) {
  .f <- pred1$var
  if (.f == "rxLinCmt") {
    return(quote(linCmt()))
  }
  str2lang(.f)
}

#' Get the prediction transformation
#'
#' @param env Environment for the parsed model
#' @param pred1 The `data.frame` of the current error
#' @param yj The transformation number for the current error
#' @return The transformation expression
#' @author Matthew Fidler
#' @keywords internal
#' @export
.rxGetPredictionFTransform <- function(env, pred1, yj) {
  if (yj == 2 || pred1$dv) {
    return(quote(rx_pred_f_))
  } else if (yj == 3) {
    return(quote(log(rx_pred_f_)))
  } else {
    return(quote(rxTBS(rx_pred_f_, rx_lambda_, rx_yj_, rx_low_, rx_hi_)))
  }
}

#' Get the Observation transformation
#'
#' @param env Environment for the parsed model
#' @param pred1 The `data.frame` of the current error
#' @param yj The transformation number for the current error
#' @return The transformation expression
#' @author Matthew Fidler
#' @keywords internal
#' @export
.rxGetPredictionDVTransform <- function(env, pred1, yj) {
  if (yj == 2) {
    return(quote(DV))
  } else if (yj == 3) {
    return(quote(log(DV)))
  } else {
    return(quote(rxTBS(DV, rx_lambda_, rx_yj_, rx_low_, rx_hi_)))
  }
}

#' Get the additive transformation
#'
#' @param env Environment for the parsed model
#' @param pred1 The `data.frame` of the current error
#' @return The quoted symbolic name of the additive standard deviation
#' @author Matthew Fidler
#' @noRd
.rxGetVarianceForErrorAdd <- function(env, pred1) {
  if (!is.na(pred1$a)) {
    .p1 <- str2lang(pred1$a)
  } else {
    .cnd <- pred1$cond
    .w <- which(env$iniDf$err %in% c("add", "lnorm", "logitNorm", "probitNorm") & env$iniDf$condition == .cnd)
    if (length(.w) == 1L) {
      .p1 <- str2lang(env$iniDf$name[.w])
    } else {
      stop("cannot find additive standard deviation for '", .cnd, "'",
           ifelse(length(env$predDf$condition) == 1L, "", "; this parameter could be estimated by another endpoint, to fix move outside of error expression."), call.=FALSE)
    }
  }
  if (pred1$variance) {
    bquote(.(.p1))
  } else {
    bquote((.(.p1)) ^ 2)
  }
}

#' Based on current error get the F that is used for prop or pow expressions
#'
#' @param env Environment of the parsed model
#' @param pred1 The `data.frame` of the current error
#' @return The f expression
#' @author Matthew Fidler
#' @noRd
.rxGetVarianceForErrorPropOrPowF <- function(env, pred1) {
  .f <- pred1$f
  .type <- as.character(pred1$errTypeF)
  if (.type == "f") {
    if (is.na(.f)) {
      stop("for propF() or powF(), f must be part of the model and not estimated",
           call.=FALSE)
    }
  }
  switch(.type,
         untransformed=quote(rx_pred_f_),
         transformed=quote(rx_pred_),
         f=str2lang(.f),
         none=quote(rx_pred_f_))
}

#' Get Variance for proportional error
#'
#' @param env Environment for the parsed model
#' @param pred1 The `data.frame` of the current error
#' @return The quoted proportional error
#' @author Matthew Fidler
#' @noRd
.rxGetVarianceForErrorProp <- function(env, pred1) {
  .f <- .rxGetVarianceForErrorPropOrPowF(env, pred1)
  if (!is.na(pred1$b)) {
    .p1 <- str2lang(pred1$b)
  } else {
    .cnd <- pred1$cond
    .w <- which(env$iniDf$err %in% c("prop", "propF", "propT") & env$iniDf$condition == .cnd)
    if (length(.w) == 1L) {
      .p1 <- str2lang(env$iniDf$name[.w])
    } else {
      stop("cannot find proportional standard deviation", call.=FALSE)
    }
  }
  if (pred1$variance) {
    bquote(.(.f)^2 * .(.p1))
  } else {
    bquote((.(.f) * .(.p1))^2)
  }
}

#' Get the Variance for pow error model
#'
#' @param env Environment for the parsed model
#' @param pred1 The `data.frame` of the current error
#' @return The quoted additive + proportional expression
#' @author Matthew Fidler
#' @noRd
.rxGetVarianceForErrorPow <- function(env, pred1) {
  .f <- .rxGetVarianceForErrorPropOrPowF(env, pred1)
  .cnd <- pred1$cond
  if (!is.na(pred1$b)) {
    .p1 <- str2lang(pred1$b)
  } else {
    .w <- which(env$iniDf$err %in% c("pow", "powF", "powT") & env$iniDf$condition == .cnd)
    if (length(.w) == 1L) {
      .p1 <- str2lang(env$iniDf$name[.w])
    } else {
      stop("cannot find power standard deviation", call.=FALSE)
    }
  }
  if (!is.na(pred1$c)) {
    .p2 <- str2lang(pred1$c)
  } else {
    .w <- which(env$iniDf$err %in% c("pow2", "powF2", "powT2") & env$iniDf$condition == .cnd)
    if (length(.w) == 1L) {
      .p2 <- str2lang(env$iniDf$name[.w])
    } else {
      stop("cannot find exponent of power expression", call.=FALSE)
    }
  }
  if (pred1$variance) {
    bquote((.(.f))^(2*.(.p2)) * .(.p1))
  } else {
    bquote(((.(.f))^(.(.p2)) * .(.p1))^2)
  }
}

#' Get Variance for proportional error
#'
#' @param env Environment for the parsed model
#' @param pred1 The `data.frame` of the current error
#' @return The add + prop error expression
#' @author Matthew Fidler
#' @noRd
.rxGetVarianceForErrorAddProp <- function(env, pred1) {
  if (!is.na(pred1$a)) {
    .p1 <- str2lang(pred1$a)
  } else {
    .cnd <- pred1$cond
    .w <- which(env$iniDf$err %in% c("add", "lnorm", "probitNorm", "logitNorm") & env$iniDf$condition == .cnd)
    if (length(.w) == 1L) {
      .p1 <- str2lang(env$iniDf$name[.w])
    } else {
      stop("cannot find additive standard deviation", call.=FALSE)
    }
  }
  .f <- .rxGetVarianceForErrorPropOrPowF(env, pred1)
  if (!is.na(pred1$b)) {
    .p2 <- str2lang(pred1$b)
  } else {
    .cnd <- pred1$cond
    .w <- which(env$iniDf$err %in% c("prop", "propT", "propF") & env$iniDf$condition == .cnd)
    if (length(.w) == 1L) {
      .p2 <- str2lang(env$iniDf$name[.w])
    } else {
      stop("cannot find proportional standard deviation", call.=FALSE)
    }
  }
  if (pred1$addProp == "default") {
    .addProp <- rxGetControl(env, "addProp", getOption("rxode2.addProp", "combined2"))
  } else {
    .addProp <- pred1$addProp
  }
  if (pred1$variance) {
    if (.addProp == "combined2") {
      bquote((.(.p1))+ (.(.f))^2*.(.p2))
    } else {
      bquote( ( (sqrt(.(.p1))) + (.(.f)) * (sqrt(.(.p2))) ) ^ 2)
    }
  } else {
    if (.addProp == "combined2") {
      bquote((.(.p1))^2+ (.(.f))^2*(.(.p2))^2)
    } else {
      bquote( ( (.(.p1)) + (.(.f)) * (.(.p2)) ) ^ 2)
    }
  }

}

#' Additive + Power
#'
#' @param env Environment for the parsed model
#' @param pred1 The `data.frame` of the current error
#' @return additive + power
#' @author Matthew Fidler
#' @noRd
.rxGetVarianceForErrorAddPow <- function(env, pred1) {
  if (!is.na(pred1$a)) {
    .p1 <- str2lang(pred1$a)
  } else {
    .cnd <- pred1$cond
    .w <- which(env$iniDf$err %in% c("add", "lnorm", "logitNorm", "probitNorm") & env$iniDf$condition == .cnd)
    if (length(.w) == 1L) {
      .p1 <- str2lang(env$iniDf$name[.w])
    } else {
      stop("cannot find additive standard deviation", call.=FALSE)
    }
  }
  .f <- .rxGetVarianceForErrorPropOrPowF(env, pred1)
  if (!is.na(pred1$b)) {
    .p2 <- str2lang(pred1$b)
  } else {
    .cnd <- pred1$cond
    .w <- which(env$iniDf$err %in% c("pow", "powF", "powT") & env$iniDf$condition == .cnd)
    if (length(.w) == 1L) {
      .p2 <- str2lang(env$iniDf$name[.w])
    } else {
      stop("cannot find pow standard deviation", call.=FALSE)
    }
  }
  if (!is.na(pred1$c)) {
    .p3 <- str2lang(pred1$c)
  } else {
    .cnd <- pred1$cond
    .w <- which(env$iniDf$err %in% c("pow2", "powF2", "powT2") & env$iniDf$condition == .cnd)
    if (length(.w) == 1L) {
      .p3 <- str2lang(env$iniDf$name[.w])
    } else {
      stop("cannot find pow exponent", call.=FALSE)
    }
  }
  if (pred1$addProp == "default") {
    .addProp <- rxGetControl(env, "addProp", getOption("rxode2.addProp", "combined2"))
  } else {
    .addProp <- pred1$addProp
  }
  if (pred1$variance) {
    if (.addProp == "combined2") {
      return(bquote( (.(.p1)) + ( (.(.f))^(.(.p3)) )^2 * .(.p2)))
    } else {
      return(bquote( ( sqrt(.(.p1)) + (.(.f)) ^ (.(.p3))* sqrt(.(.p2)) ) ^ 2))
    }
  } else {
    if (.addProp == "combined2") {
      return(bquote( (.(.p1))^2 + ( (.(.f))^(.(.p3)) )^2 * (.(.p2))^2))
    } else {
      return(bquote( ( (.(.p1)) + (.(.f)) ^ (.(.p3))* (.(.p2)) ) ^ 2))
    }
  }
}
#' Get Variance for error type
#'
#' @param env Environment
#' @param pred1 Pred for one end-point
#' @return Variance error type
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
.rxGetVarianceForErrorType <- function(env, pred1) {
  switch(as.character(pred1$errType),
         "add"=.rxGetVarianceForErrorAdd(env, pred1), # 1
         "prop"=.rxGetVarianceForErrorProp(env, pred1), # 2
         "pow"=.rxGetVarianceForErrorPow(env, pred1), # 3
         "add + prop"=.rxGetVarianceForErrorAddProp(env, pred1),# 4
         "add + pow"=.rxGetVarianceForErrorAddPow(env, pred1) # 5
         )
}

#' Build the per-observation log-likelihood call for an AR(1) endpoint
#'
#' @param env parsed model environment
#' @param pred1 single predDf row
#' @param errNum endpoint number
#' @param type distribution type (norm/dnorm/t/cauchy)
#' @param dvTrans quoted transformed observed value
#' @return function(meanLang, sdLang) returning the quoted llik call
#' @author Matthew Fidler
#' @noRd
.rxArBuildLlik <- function(env, pred1, errNum, type, dvTrans) {
  if (type == "t") {
    .iniDf <- env$iniDf
    .cnd <- pred1$cond
    .w <- which(.iniDf$err == "t" & .iniDf$condition == .cnd)
    if (length(.w) == 1) {
      .nu <- str2lang(paste(.iniDf$name[.w]))
    } else {
      if (is.na(pred1$d)) {
        stop("t distribution needs a proper degrees of freedom specified", call. = FALSE)
      }
      .nu <- str2lang(pred1$d)
    }
    function(.mean, .sd) {
      if (errNum == 1) bquote(llikT(.(dvTrans), .(.nu), .(.mean), .(.sd)))
      else bquote(llikXT(.(errNum - 1), .(dvTrans), .(.nu), .(.mean), .(.sd)))
    }
  } else if (type == "cauchy") {
    function(.mean, .sd) {
      if (errNum == 1) bquote(llikCauchy(.(dvTrans), .(.mean), .(.sd)))
      else bquote(llikXCauchy(.(errNum - 1), .(dvTrans), .(.mean), .(.sd)))
    }
  } else {
    function(.mean, .sd) {
      if (errNum == 1) bquote(llikNorm(.(dvTrans), .(.mean), .(.sd)))
      else bquote(llikXNorm(.(errNum - 1), .(dvTrans), .(.mean), .(.sd)))
    }
  }
}

#' Generate the estimation lines for an AR(1) (`ar()`) endpoint
#'
#' Emits a continuous-time AR(1) conditional log-likelihood.  The marginal
#' residual `rx.arE.<var> = DVtrans - pred` and last time `rx.arT.<var>` are
#' plain lhs variables; their previous-record values come from `lag()` (which
#' survives the symengine estimation prune), and `ifelse(is.na(...))` handles the
#' first record per individual (marginal likelihood).  A normal endpoint is
#' packed as `dnorm` so the engine takes the explicit-likelihood path.
#'
#' @param env parsed model environment
#' @param pred1 single predDf row
#' @param errNum endpoint number
#' @param type distribution type (norm/dnorm/t/cauchy)
#' @param yj transform code for this endpoint
#' @param cor quoted AR(1) correlation
#' @return list of quoted model lines
#' @author Matthew Fidler
#' @noRd
.rxArEstLlikLines <- function(env, pred1, errNum, type, yj, cor) {
  .distInt <- if (type == "norm") {
    as.integer(factor("dnorm", levels = levels(pred1$distribution)))
  } else {
    as.integer(pred1$distribution)
  }
  .dvTrans <- .rxGetPredictionDVTransform(env, pred1, yj)
  .buildLlik <- .rxArBuildLlik(env, pred1, errNum, type, .dvTrans)
  # use dot-free names: the symengine sensitivity/derivative path mishandles
  # dotted variable names inside a llik() argument, which breaks analytic
  # gradients for the estimation model
  .var <- gsub("[^A-Za-z0-9]", "_", pred1$var)
  .e <- str2lang(paste0("rx_arE_", .var))
  .t <- str2lang(paste0("rx_arT_", .var))
  .ep <- str2lang(paste0("rx_arEp_", .var))
  .dt <- str2lang(paste0("rx_arDt_", .var))
  .nf <- str2lang(paste0("rx_arNf_", .var))
  .phi <- str2lang(paste0("rx_arPhi_", .var))
  # First observation per subject/endpoint uses the marginal likelihood.  lag0()
  # keeps the previous residual and time gap FINITE on the first record (0 and
  # time), and `1 - is.na(lag(...))` is a NaN-safe first-record indicator (0 on
  # the first record, 1 after) -- so rx_arNf * cor^dt = 0*finite = 0 on the first
  # record (marginal), with no NaN (which the pruned-ifelse 0*NaN would produce).
  list(
    bquote(rx_yj_ ~ .(yj + 10 * (.distInt - 1))),
    bquote(rx_lambda_ ~ .(.rxGetLambdaFromPred1AndIni(env, pred1))),
    bquote(rx_low_ ~ .(.rxGetLowBoundaryPred1AndIni(env, pred1))),
    bquote(rx_hi_ ~ .(.rxGetHiBoundaryPred1AndIni(env, pred1))),
    bquote(rx_pred_f_ ~ .(.rxGetPredictionF(env, pred1))),
    bquote(rx_pred_ ~ .(.rxGetPredictionFTransform(env, pred1, yj))),
    bquote(rx_rll_ ~ sqrt(.(.rxGetVarianceForErrorType(env, pred1)))),
    bquote(.(.t) <- time),
    bquote(.(.e) <- .(.dvTrans) - rx_pred_),
    bquote(.(.ep) <- lag0(.(.e), 1)),
    bquote(.(.dt) <- time - lag0(.(.t), 1)),
    bquote(.(.nf) <- 1 - is.na(lag(.(.t), 1))),
    bquote(.(.phi) <- .(.nf) * .(cor)^.(.dt)),
    bquote(rx_pred_ ~ .(.buildLlik(bquote(rx_pred_ + .(.phi) * .(.ep)),
                                   bquote(rx_rll_ * sqrt(1 - .(.phi)^2))))),
    quote(rx_r_ ~ 0))
}

#' AR(1) whitened residual in Gaussian norm (mean/variance) form
#'
#' Norm sibling of `.rxArEstLlikLines()` for the FOCEi inner model: instead of an
#' explicit `llik`, the whitened conditional likelihood is expressed as a plain
#' `norm` endpoint whose mean is `rx_pred_ + phi*e_prev` and whose variance is
#' `R*(1 - phi^2)`.  Keeping the endpoint on the `norm` path lets FOCEi build the
#' exact Almquist eta-Hessian (the `llik` path only has a first-order gradient).
#'
#' inheritParams .rxArEstLlikLines
#'
#' @return list of model lines
#' @author Matthew Fidler
#' @keywords internal
#' @noRd
.rxArEstNormLines <- function(env, pred1, errNum, type, yj, cor) {
  .dvTrans <- .rxGetPredictionDVTransform(env, pred1, yj)
  # dot-free names (the symengine sensitivity path mishandles dotted names)
  .var <- gsub("[^A-Za-z0-9]", "_", pred1$var)
  .e <- str2lang(paste0("rx_arE_", .var))
  .t <- str2lang(paste0("rx_arT_", .var))
  .ep <- str2lang(paste0("rx_arEp_", .var))
  .dt <- str2lang(paste0("rx_arDt_", .var))
  .nf <- str2lang(paste0("rx_arNf_", .var))
  .phi <- str2lang(paste0("rx_arPhi_", .var))
  # rx_arE_ uses the structural (transformed) prediction; rx_pred_ is then
  # reassigned to the conditional mean and rx_r_ to the conditional variance.
  # First record: rx_arNf_=0 => phi=0 => mean=pred, var=R (marginal), NaN-safe.
  list(
    bquote(rx_yj_ ~ .(yj + 10 * (as.integer(pred1$distribution) - 1))),
    bquote(rx_lambda_ ~ .(.rxGetLambdaFromPred1AndIni(env, pred1))),
    bquote(rx_low_ ~ .(.rxGetLowBoundaryPred1AndIni(env, pred1))),
    bquote(rx_hi_ ~ .(.rxGetHiBoundaryPred1AndIni(env, pred1))),
    bquote(rx_pred_f_ ~ .(.rxGetPredictionF(env, pred1))),
    bquote(rx_pred_ ~ .(.rxGetPredictionFTransform(env, pred1, yj))),
    bquote(.(.t) <- time),
    bquote(.(.e) <- .(.dvTrans) - rx_pred_),
    bquote(.(.ep) <- lag0(.(.e), 1)),
    bquote(.(.dt) <- time - lag0(.(.t), 1)),
    bquote(.(.nf) <- 1 - is.na(lag(.(.t), 1))),
    bquote(.(.phi) <- .(.nf) * .(cor)^.(.dt)),
    bquote(rx_pred_ ~ rx_pred_ + .(.phi) * .(.ep)),
    bquote(rx_r_ ~ .(.rxGetVarianceForErrorType(env, pred1)) * (1 - .(.phi)^2)))
}

#' Handle the single error for normal or t distributions
#'
#' @param env Environment for the parsed model
#'
#' @param pred1 The `data.frame` of the current error
#'
#' @param errNum The number of the error specification in the nlmixr2 model
#'
#' @param rxPredLlik A boolean indicating if the log likelihood should
#'   be calculated for non-normal distributions.  By default `TRUE`.
#'
#' @param arNorm A boolean; when `TRUE`, emit the AR(1) norm
#'   (conditional mean/variance) form used by the focei inner model.
#'   By default `FALSE`.
#'
#' @return A list of the lines added.  The lines will contain
#'
#' - `rx_yj_` which is an integer that corresponds to the
#'   transformation type.
#'
#' - `rx_lambda_` is the transformation lambda
#'
#' - `rx_low_` The lower boundary of the transformation
#'
#' - `rx_hi_` The upper boundary of the transformation
#'
#' - `rx_pred_f_` The prediction function
#'
#' - `rx_pred_` The transformed prediction function
#'
#' - `rx_r_` The transformed variance
#'
#' @author Matthew Fidler
#' @export
.handleSingleErrTypeNormOrTFoceiBase <- function(env, pred1, errNum=1L, rxPredLlik=TRUE, arNorm=FALSE) {
  type <- pred1$distribution
  if (type %in% c("norm", "t", "cauchy", "dnorm")) {
    .yj <- as.double(pred1$transform) - 1
    # AR(1), norm form (focei inner model): keep the endpoint on the Gaussian
    # norm path (mean + variance) so the exact eta-Hessian is used.
    if (arNorm && type == "norm") {
      .arCorN <- .rxGetArCorLang(env, pred1)
      if (!is.null(.arCorN)) {
        return(.rxArEstNormLines(env, pred1, errNum, type, .yj, .arCorN))
      }
    }
    .arCor <- if (rxPredLlik) .rxGetArCorLang(env, pred1) else NULL
    if (!is.null(.arCor)) {
      # AR(1) endpoint: emit a per-observation conditional log-likelihood
      # whitened with lag() (a symengine-safe previous-residual accessor) and
      # routed through the explicit-likelihood path
      return(.rxArEstLlikLines(env, pred1, errNum, type, .yj, .arCor))
    }
    .ret <- vector("list", ifelse(type == "norm", 7, ifelse(rxPredLlik, 9, 7)))
    .ret[[1]] <- bquote(rx_yj_ ~ .(.yj + 10*(as.integer(pred1$distribution)-1)))
    .ret[[2]] <- bquote(rx_lambda_~.(.rxGetLambdaFromPred1AndIni(env, pred1)))
    .ret[[3]] <- bquote(rx_low_ ~ .(.rxGetLowBoundaryPred1AndIni(env, pred1)))
    .ret[[4]] <- bquote(rx_hi_ ~ .(.rxGetHiBoundaryPred1AndIni(env, pred1)))
    .ret[[5]] <- bquote(rx_pred_f_ ~ .(.rxGetPredictionF(env, pred1)))
    .ret[[6]] <- bquote(rx_pred_ ~ .(.rxGetPredictionFTransform(env, pred1, .yj)))
    if (type == "norm") {
      .ret[[7]] <- bquote(rx_r_ ~ .(.rxGetVarianceForErrorType(env, pred1)))
    } else {
      if (rxPredLlik) {
        .ret[[7]] <- bquote(rx_rll_ ~ sqrt(.(.rxGetVarianceForErrorType(env, pred1))))
        if (type == "t") {
          .iniDf <- env$iniDf
          .cnd <- pred1$cond
          .w <- which(.iniDf$err == "t" & .iniDf$condition == .cnd)
          if (length(.w) == 1) {
            .nu <- str2lang(paste(.iniDf$name[.w]))
          } else {
            if (is.na(pred1$d)) {
              stop("t distribution needs a proper degrees of freedom specified",
                   call. = FALSE)
            }
            .nu <- str2lang(pred1$d)
          }
          if (errNum == 1) {
            .ret[[8]] <- bquote(rx_pred_ ~ llikT(.(.rxGetPredictionDVTransform(env, pred1, .yj)),
                                                 .(.nu), rx_pred_, rx_rll_))
          } else {
            .ret[[8]] <- bquote(rx_pred_ ~ llikXT(.(errNum - 1),
                                                  .(.rxGetPredictionDVTransform(env, pred1, .yj)),
                                                 .(.nu), rx_pred_, rx_rll_))
          }
        } else if (type == "cauchy") {
          if (errNum == 1) {
            .ret[[8]] <- bquote(rx_pred_ ~ llikCauchy(.(.rxGetPredictionDVTransform(env, pred1, .yj)),
                                                      rx_pred_, rx_rll_))
          } else {
            .ret[[8]] <- bquote(rx_pred_ ~ llikXCauchy(.(errNum - 1),
                                                       .(.rxGetPredictionDVTransform(env, pred1, .yj)),
                                                       rx_pred_, rx_rll_))
          }
        } else if (type == "dnorm") {
          if (errNum == 1) {
            .ret[[8]] <- bquote(rx_pred_ ~ llikNorm(.(.rxGetPredictionDVTransform(env, pred1, .yj)),
                                                    rx_pred_, rx_rll_))
          } else {
            .ret[[8]] <- bquote(rx_pred_ ~ llikXNorm(.(errNum - 1),
                                                     .(.rxGetPredictionDVTransform(env, pred1, .yj)),
                                                    rx_pred_, rx_rll_))
          }
        }
        .ret[[9]] <- quote(rx_r_ ~ 0)
      } else {
        .ret[[7]] <- bquote(rx_r_ ~ .(.rxGetVarianceForErrorType(env, pred1)))
      }
    }
    .ret
  } else {
    .yj <- 2
    .ret <- vector("list", 6)
    .yj <- as.double(pred1$transform) - 1
    .ret[[1]] <- bquote(rx_yj_ ~ .(.yj + 10 * (as.integer(pred1$distribution) - 1)))
    .ret[[2]] <- bquote(rx_lambda_ ~ 1)
    .ret[[3]] <- bquote(rx_low_ ~ 0)
    .ret[[4]] <- bquote(rx_hi_ ~ 1)
    .ret[[5]] <- bquote(rx_r_ ~ 0)
    # This is for the non-normal cases
    .ret[[6]] <- bquote(rx_pred_ ~ .(.getQuotedDistributionAndLlikArgs(env, pred1, errNum)))
    .ret
  }
}

.foceEstLLFun <- list(
  "t"="llikT",
  "pois"="llikPois",
  "binom"="llikBinom",
  "nbinom"="llikNbinom",
  "nbinomMu"="llikNbinomMu",
  "beta"="llikBeta",
  "chisq"="llikChisq",
  "dexp"="llikExp",
  "f"="llikF",
  "geom"="llikGeom",
#  "hyper"="rhyper",
  "unif"="llikUnif",
  "weibull"="llikWeibull",
  "cauchy"="llikCauchy",
  "dgamma"="llikGamma",
  "ordinal"="rordinal"
)

.getOrdinalDistPred <- function(env, pred1) {
  # final should be log((DV==1)*prob1 + (DV==2)*prob2 + ... + (DV==n)*(1 - prob1 - prob2 - ...))
  .c <- env$lstExpr[[pred1$line[1]]][[3]]
  .ce <- try(eval(.c), silent=TRUE)
  if (inherits(.ce, "try-error")) {
  } else if (inherits(.ce, "numeric") &&
               !is.null(names(.ce))) {
    .n <- names(.ce)
    .ln <- length(.n)
    if (.n[.ln] != "") {
      stop("last element in ordinal simulation of c(p1=0, p2=0.5, ...) must be a number, not a named number",
           call.=FALSE)
    }
    .n <- .n[.n != ""]
    if (length(.n) != .ln - 1) {
      stop("names for ordinal simulation incorrect")
    }
    .first <- vapply(seq_along(.n), function(i) {
      paste0("(DV==",.ce[i],")*(", .n[i], ")")
    }, character(1), USE.NAMES=FALSE)
    .last <- paste0("(DV==", .ce[length(.n) + 1], ")*(1.0 -",paste(.n, collapse="-"),")")
    .ret <- paste0("log(", paste(c(.first, .last), collapse="+"),")")
    return(str2lang(.ret))
  }
  if (length(.c) >= 2) {
    .first <- vapply(seq(2, length(.c)), function(i) {
      paste0("(DV==",(i-1),")*(",
             deparse1(.c[[i]]),")")

    }, character(1), USE.NAMES=FALSE)
    .last <- vapply(seq(2, length(.c)), function(i) {
      paste0("(", deparse1(.c[[i]]), ")")
    }, character(1), USE.NAMES=FALSE)
    .last <- paste0("(DV==", length(.c), ")*(1.0 -",paste(.last, collapse="-"),")")
    .ret <- paste0("log(", paste(c(.first, .last), collapse="+"),")")
    return(str2lang(.ret))
  } else {
    stop("not enough categories to estimate")
  }
}

.getQuotedDistributionAndLlikArgs <- function(env, pred1, errNum=1L) {
  .dist <- as.character(pred1$distribution)
  if (.dist == "LL") {
    return(env$lstExpr[[pred1$line]][[3]])
  } else if (.dist == "ordinal") {
    return(.getOrdinalDistPred(env, pred1))
  }
  .nargs <- max(.errDist[[.dist]])
  .cnd <- pred1$cond
  .argName <- .namedArgumentsToPredDf[[.dist]]
  .args <- vapply(seq(1:.nargs), function(.i) {
    .curDist <- .argName[.i]
    if (!is.na(pred1[[.curDist]])) {
      return(pred1[[.curDist]])
    } else {
      .curDist <- paste0(.dist, ifelse(.i == 1, "", .i))
      .w <- which(env$iniDf$err == .curDist & env$iniDf$condition == .cnd)
      if (length(.w) == 1) {
        return(env$iniDf$name[.w])
      } else {
        return("")
      }
    }
  }, character(1))
  .d1 <- .foceEstLLFun[[.dist]]
  if (errNum != 1) {
    .d1 <- sub("^llik", "llikX", .d1)
    .args <- c(.d1, paste0(errNum - 1), "DV", .args[.args != ""])
  } else {
    .args <- c(.d1, "DV", .args[.args != ""])
  }
  as.call(lapply(.args, str2lang))
}

#' This asserts arguments that should not be used together and returns the resolved ID
#'
#' @param dotArgs The list of the ... arguments passed to the main
#'   et() function, used to check for deprecated or alias arguments
#'
#' @param id the id argument passed to et(), which may be modified if
#'   an uppercase "ID" is found in dotArgs
#'
#' @param ii the ii argument passed to et()
#'
#' @param amt the amt argument passed to et()
#'
#' @param timeUnits the timeUnits argument passed to et()
#'
#' @param time the time argument passed to et()
#'
#' @param dur the dur argument passed to et()
#'
#' @param idMissing a logical indicating whether the id argument was
#'   missing
#'
#' @param iiMissing a logical indicating whether the ii argument was
#'   missing
#'
#' @param amtMissing a logical indicating whether the amt argument was
#'   missing
#'
#' @param cmtMissing a logical indicating whether the cmt argument was
#'   missing
#'
#' @param timeUnitsMissing a logical indicating whether the timeUnits
#'   argument was missing
#'
#' @param timeMissing a logical indicating whether the time argument
#'   was missing
#'
#' @param durMissing a logical indicating whether the dur argument was
#'   missing
#'
#' @return the resolved id value after checking for aliases and conflicts
#' @noRd
.etAssertArgsAndReturnId <- function(dotArgs, id, ii, amt, cmt, timeUnits, time, dur,
                                     idMissing, iiMissing, amtMissing, cmtMissing,
                                     timeUnitsMissing, timeMissing, durMissing) {
  # Uppercase ID alias -> id
  if (!is.null(dotArgs[["ID"]]) && idMissing) {
    id <- dotArgs[["ID"]]
  }
  if (!is.null(dotArgs[["dosing.interval"]]) && !iiMissing)
    stop("cannot specify both 'ii' and 'dosing.interval'", call. = FALSE)
  if (!is.null(dotArgs[["dose"]]) && !amtMissing)
    stop("cannot specify both 'amt' and 'dose'", call. = FALSE)
  if (!is.null(dotArgs[["dosing.to"]]) && !cmtMissing)
    stop("cannot specify both 'cmt' and 'dosing.to'", call. = FALSE)
  if (!is.null(dotArgs[["dose.to"]]) && !cmtMissing)
    stop("cannot specify both 'cmt' and 'dose.to'", call. = FALSE)
  if (!is.null(dotArgs[["state"]]) && !cmtMissing)
    stop("cannot specify both 'cmt' and 'state'", call. = FALSE)
  if (!is.null(dotArgs[["amt.units"]]) && !is.null(dotArgs[["dose.units"]]))
    stop("cannot specify both 'amt.units' and 'dose.units'", call. = FALSE)
  if (!is.null(dotArgs[["time.units"]]) && !timeUnitsMissing)
    stop("cannot specify both 'time.units' and 'timeUnits'", call. = FALSE)
  if (!is.null(dotArgs[["start.time"]]) && !timeMissing)
    stop("cannot specify both 'time' and 'start.time'", call. = FALSE)
  if (!is.null(dotArgs[["nbr.doses"]]) && !is.null(dotArgs[["nbrDoses"]]))
    stop("cannot specify both 'nbr.doses' and 'nbrDoses'", call. = FALSE)
  if (!is.null(dotArgs[["duration"]]) && !durMissing)
    stop("cannot specify both 'dur' and 'duration'", call. = FALSE)
  id
}
#' This function handles the sequence type of arguments
#'
#' @param by the by argument passed to et(), used to specify the
#'   increment for a sequence of observation times.
#'
#' @param length.out the length.out argument passed to et(), used to specify the number
#'  of points in a sequence of observation times.
#'
#' @param xIsRxEt a logical indicating whether the x argument is an
#'   rxEt object, which affects how the from/to values for the
#'   sequence are resolved.
#'
#' @param envRef a reference to the internal environment of the rxEt
#'   object being constructed, used to access and modify the current
#'   state of the event table as new chunks are added.
#'
#' @param x the x argument passed to et(), which may be used to
#'   resolve the from/to values for the sequence if xIsRxEt is FALSE
#'
#' @param ... the ... arguments passed to et(), which may contain
#'   named or unnamed arguments used to resolve the from/to values for
#'   the sequence, such as from/to or the first two unnamed numeric
#'   arguments.
#'
#' @param envir the environment in which to evaluate any expressions needed to
#'  resolve the from/to values for the sequence.
#'
#' @param time the time argument passed to et(), which is used to
#'   check for conflicts with the sequence arguments.
#'
#' @param et the current state of the event table being constructed,
#'   which may be modified by adding a new chunk of observation times
#'   if the sequence arguments are successfully resolved.
#'
#' @param xMissing a logical indicating whether the x argument was
#'   missing, which affects how the from/to values for the sequence
#'   are resolved.
#'
#' @param timeMissing a logical indicating whether the time argument
#'   was missing, which is used to check for conflicts with the
#'   sequence arguments.
#'
#' @return a list with components done (a logical indicating whether
#'   the sequence was successfully handled and the main et() function
#'   should return immediately) and et (the possibly modified event
#'   table after handling the sequence arguments)
#'
#' @noRd
#'
#' @author Matthew L. Fidler
#'
.etHandleSeq <- function(by, length.out, xIsRxEt, envRef, x, ..., envir,
                         time, et, xMissing, timeMissing) {
  if (!is.null(by) && !is.null(length.out))
    stop("cannot specify both 'by' and 'length.out'", call. = FALSE)
  if (!is.null(by) || !is.null(length.out)) {
    if (xIsRxEt && length(envRef$ids) > 0L) {
      .seqTargetIds <- envRef$ids
    } else {
      .seqTargetIds <- NULL
    }
    .fromVal <- NULL
    .toVal <- NULL
    .seqDots <- list(...)

    if (!xMissing && !xIsRxEt) {
      .xVal <- x
    } else {
      .xVal <- NULL
    }
    .dotsNum <- Filter(function(.v) {
      is.numeric(.v) || is.integer(.v)
    }, .seqDots)

    # Also check named from/to in ...
    if (!is.null(.seqDots[["from"]])) {
      .fromVal <- .seqDots[["from"]]
    } else if (!is.null(.xVal) && (is.numeric(.xVal) || is.integer(.xVal))) {
      .fromVal <- .xVal
    } else if (xIsRxEt && length(.dotsNum) >= 1L) {
      .fromVal <- .dotsNum[[1L]]
    }

    if (!is.null(.seqDots[["to"]])) {
      .toVal <- .seqDots[["to"]]
    } else if (!xMissing && !xIsRxEt && length(.dotsNum) >= 1L) {
      .toVal <- .dotsNum[[1L]]
    } else if (xIsRxEt && length(.dotsNum) >= 2L) {
      .toVal <- .dotsNum[[2L]]
    }

    if (!is.null(.fromVal) &&
          (is.numeric(.fromVal) || is.integer(.fromVal))) {
      .from <- as.numeric(.fromVal)
      if (length(.from) != 1L) {
        stop("'from' must be scalar", call. = FALSE)
      }
      if (!is.null(.toVal)) {
        .to <- as.numeric(.toVal)
        if (length(.to) != 1L) {
          stop("'to' must be scalar", call. = FALSE)
        }
        if (!is.null(by)) {
          .resolvedTime <- seq(from = .from, to = .to, by = by)
        } else {
          .resolvedTime <- seq(from = .from, to = .to, length.out = length.out)
        }
      } else if (!is.null(by)) {
        .resolvedTime <- seq(from = .from, by = by)
      } else {
        .resolvedTime <- seq(from = .from, length.out = length.out)
      }
    } else if (!is.null(time)) {
      stop("'by'/'length.out' requires both a 'from' and 'to' value", call. = FALSE)
    } else {
      .resolvedTime <- numeric(0)
    }
    .df <- .etObsChunk(.resolvedTime) # nolint
    .etAddChunk(envRef, .df, .seqTargetIds) # nolint
    envRef$nobs <- envRef$nobs + length(.resolvedTime) * max(1L, length(.seqTargetIds))
    return(list(done = TRUE, et = et))
  }
  list(done = FALSE)
}

#' Get the range sequence with et(1, 20) to seq(1, 20)
#'
#'
#' @param xVal x value
#'
#' @param dots the dots arguments
#'
#' @param envRef the environmental reference for the rxEt
#'
#' @param et event table that may be modified
#'
#' @return Either NULL or a list with `done` and `et` components
#'
#' @noRd
#' @author Matthew L. Fidler
.etHandlePositionalRange <- function(xVal, dots, envRef, et) {
  if (length(dots) == 1 &&
        !is.null(xVal) &&
        length(xVal) == 1 &&
        (is.numeric(xVal) || is.integer(xVal)) &&
        (is.numeric(dots[[1]]) || is.integer(dots[[1]]))) {
    #et(1, 20) add obs seq(1, 20)
    .resolvedTime <- seq(from = as.numeric(xVal), to = as.numeric(dots[[1]]))
    .df <- .etObsChunk(.resolvedTime) # nolint
    .etAddChunk(envRef, .df, NULL) # nolint
    envRef$nobs <- envRef$nobs + length(.resolvedTime)
    return(list(done = TRUE, et = et))
  }
  NULL
}
#' Handle the positional data.frame arguments
#'
#' This includes modifying deSolve event data frames to their proper
#' value
#'
#' @param xVal the xVal is the data frame
#'
#' @param envRef environment reference
#'
#' @param et event table
#'
#' @return list of done and et events
#'
#' @noRd
#'
#' @author Matthew L. Fidler
#'
.etHandlePositionalDataFrame <- function(xVal, envRef, et) {
  .df <- as.data.frame(xVal)
  # as.data.frame() may drop it; re-tag first so the rename below tracks it
  .extra <- .etExtraColsAttr(xVal) # nolint
  if (length(.extra) > 0L) attr(.df, "rxEtExtraCols") <- .extra
  .df <- .etImportNormalizeNames(.df)
  # Convert deSolve-style (var/value/method) to canonical rxEt format
  if (!is.null(.df$var) && !is.null(.df$value) && is.null(.df$amt) && is.null(.df$evid)) {
    .df$cmt   <- .df$var
    .df$var   <- NULL
    .df$amt   <- .df$value
    .df$value <- NULL
    if (!is.null(.df$method)) {
      .df$evid <- ifelse(.df$method == "rep", 5L,
                         ifelse(.df$method == "mult", 6L, 1L))
      .df$method <- NULL
    } else {
      .df$evid <- 1L
    }
  }
  if (is.null(.df$evid)) {
    if (!is.null(.df$amt)) {
      .df$evid <- ifelse(!is.na(.df$amt) & as.numeric(.df$amt) != 0, 1L, 0L)
    } else {
      .df$evid <- 0L
    }
  }
  .df$evid <- as.integer(.df$evid)
  .hadId <- !is.null(.df$id)
  if (is.null(.df$id)) {
    .df$id <- 1L
  }
  .df$id <- as.integer(.df$id)
  .obsIdx <- .df$evid == 0L
  if (any(.obsIdx)) {
    if (!is.null(.df$rate)) {
      .df$rate[.obsIdx] <- NA_real_
    }
    if (!is.null(.df$amt)) {
      .df$amt[.obsIdx]  <- NA_real_
    }
  }
  envRef$ids    <- sort(unique(.df$id))
  envRef$nobs   <- envRef$nobs  + sum(.obsIdx)
  envRef$ndose  <- envRef$ndose + sum(!.obsIdx)
  if (length(envRef$ids) > 1L || .hadId) {
    envRef$show["id"] <- TRUE
  }
  if (sum(!.obsIdx) > 0L) {
    envRef$show["amt"] <- TRUE
  }
  if (!is.null(.df$rate) &&
        any(.df$rate[!.obsIdx] != 0, na.rm = TRUE)) {
    envRef$show["rate"] <- TRUE
  }
  if (!is.null(.df$ii) && any(.df$ii != 0, na.rm = TRUE)) {
    envRef$show["ii"]   <- TRUE
    envRef$show["addl"] <- TRUE
  }
  envRef$chunks <- .addRowsToChunks(envRef$chunks, .df) # nolint
  .etImportExtraCols(envRef, .df, names(.df)) # nolint
  list(done = TRUE, et = et)
}

#' Handle the position arguments (ie ...)
#'
#'
#' @param x the x argument passed to et(), which may be used to
#'   resolve the from/to values for a sequence of observation times if
#'   xIsRxEt is FALSE
#'
#' @param ... the ... arguments passed to et(), to be resolved
#'
#' @param time the time argument passed to et()
#'
#' @param xIsRxEt a logical indicating whether the x argument is an
#'   rxEt object
#'
#' @param envir the environment in which to evaluate any expressions
#'   needed to
#'
#' @param envRef the reference to the internal environment of the rxEt
#'   object being constructed
#'
#' @param et the current state of the event table being constructed,
#'   which may be modified by adding a new chunk of observation times
#'   if the sequence arguments are successfully resolved.
#'
#' @param xMissing Is the x argument missing?
#'
#' @param timeMissing Is the time argument missing?
#'
#' @return list with components done (a logical indicating whether the
#'   position arguments were successfully handled and the main et()
#'   function should return immediately), posCmt (the resolved cmt
#'   value from the position arguments, if any), listObs (a list of
#'   observation times from the position arguments, if any), time (the
#'   resolved time value from the position arguments, if any), and et
#'   (the possibly modified event table after handling the position
#'   arguments)
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.etHandlePositional <- function(x, ..., time, xIsRxEt, envir, envRef, et,
                                xMissing, timeMissing) {
  .posCmt <- NULL
  .listObs <- NULL
  if (is.null(time) && !xMissing && !xIsRxEt) {
    .xVal <- x
    .dots <- list(...)
    if (length(.dots) >= 2L) {
      # possibly sequence of event tables or waiting times
      if (any(vapply(.dots, is.rxEt, logical(1))) || # nolint
            any(vapply(.dots, is.numeric, logical(1)))) {
        # defer to etSeq handled in HandlePiping
        return(list(done = FALSE, posCmt = NULL, listObs = NULL, time = NULL))
      }
      stop("unused positional arguments", call. = FALSE)
    }
    if (length(.dots) == 1L && !is.null(names(.dots)) && any(names(.dots) != "")) {
      # Has named dots, let standard handler take it
    } else {
      .res <- .etHandlePositionalRange(.xVal, .dots, envRef, et)
      if (!is.null(.res)) return(.res)
      if (is.data.frame(.xVal)) return(.etHandlePositionalDataFrame(.xVal, envRef, et))
      if (is.list(.xVal)) {
        .listObs <- .xVal
      } else if (!is.null(.xVal)) {
        # single positional arg becomes time
        time <- .xVal
        # Second positional arg in ... treated as compartment (old API compat)
        if (length(.dots) == 1L && is.character(.dots[[1]])) {
          .posCmt <- .dots[[1]]
        }
      }
    }
  }
  list(done = FALSE, posCmt = .posCmt, listObs = .listObs, time = time)
}

#' This function handles any units sent to the et() function.
#'
#'
#' @param envRef this is the reference to the internal environment of
#'   the rxEt object being constructed
#'
#' @param amountUnits the amountUnits argument passed to et()
#'
#' @param timeUnits the timeUnits argument passed to et()
#'
#' @param dotArgs The .. arguments passed to et(), which may contain
#'   named or unnamed arguments
#'
#' @param amountUnitsMissing a logical indicating whether the
#'   amountUnits argument was missing
#'
#' @param timeUnitsMissing a logical indicating whether the timeUnits
#'   argument was missing
#'
#' @return nothing, called for side effects
#' @noRd
#' @author Matthew L. Fidler
.etHandleUnits <- function(envRef, amountUnits, timeUnits, dotArgs,
                           amountUnitsMissing, timeUnitsMissing) {
  if (!is.null(amountUnits)) envRef$units["dosing"] <- amountUnits
  if (!is.null(timeUnits))   envRef$units["time"]   <- timeUnits
  # Handle dot-style and underscore aliases from ...
  if (!is.null(dotArgs[["amount.units"]]) &&
        is.null(amountUnits))
    envRef$units["dosing"] <- dotArgs[["amount.units"]]
  if (!is.null(dotArgs[["time.units"]]) &&
        is.null(timeUnits))
    envRef$units["time"] <- dotArgs[["time.units"]]
  if (!is.null(dotArgs[["time_units"]]) &&
        is.null(timeUnits) &&
        is.null(dotArgs[["time.units"]]))
    envRef$units["time"] <- dotArgs[["time_units"]]
  if (!is.null(dotArgs[["amount_units"]]) &&
        is.null(amountUnits) &&
        is.null(dotArgs[["amount.units"]]))
    envRef$units["dosing"] <- dotArgs[["amount_units"]]
}
#' `et(id=)` handling
#'
#' @param id The id argument
#'
#' @param envRef The `et()` environment
#'
#' @param xIsRxEt Is the object a rxode2 et object?
#'
#' @param envir evaluation environment
#'
#' @return id value for the event table object
#'
#' @noRd
#'
#' @author Matthew L. Fidler
#'
.etHandleId <- function(id, envRef, xIsRxEt, envir, addingEvents = FALSE) {
  .resolvedId  <- NULL
  .addedIds    <- integer(0)
  .removedIds  <- integer(0)
  .existingIds <- integer(0)
  .doResize    <- FALSE
  if (!is.null(id)) {
    .idInput <- if (is.factor(id)) as.character(id) else id
    .idVal <- suppressWarnings(as.integer(.idInput))
    if (anyNA(.idVal)) {
      stop("'id' must contain integer-like values", call. = FALSE)
    }
    .posIds      <- .idVal[.idVal > 0L]
    .negIds      <- abs(.idVal[.idVal < 0L])
    .existingIds <- envRef$ids
    if (length(.posIds) > 0L &&
        !xIsRxEt &&
        identical(sort(as.integer(envRef$ids)), 1L) &&
        envRef$nobs == 0L &&
        envRef$ndose == 0L &&
        length(.etGroups(envRef)) == 0L &&
        length(envRef$chunks) == 0L) {
      .addedIds   <- setdiff(.posIds, .existingIds)
      .removedIds <- setdiff(.existingIds, .posIds)
      envRef$ids  <- sort(unique(.posIds))
    } else if (xIsRxEt && envRef$canResize && length(.idVal) == 0L) {
      .removedIds <- .existingIds
      .addedIds <- integer(0)
      envRef$ids <- integer(0)
      .doResize <- length(.removedIds) > 0L
    } else if (length(.posIds) > 0L && xIsRxEt && envRef$canResize &&
               (!addingEvents || (envRef$nobs == 0L && envRef$ndose == 0L &&
                                  length(envRef$chunks) == 0L && length(.etGroups(envRef)) == 0L))) {
      # canResize mode: positive ids define the exact target set, replacing existing
      .removedIds <- setdiff(.existingIds, .posIds)
      .addedIds   <- setdiff(.posIds, .existingIds)
      envRef$ids <- sort(unique(.posIds))
      .doResize   <- length(.addedIds) > 0L || length(.removedIds) > 0L
    } else {
      .addedIds   <- setdiff(.posIds, .existingIds)
      .removedIds <- intersect(.negIds, .existingIds)
      if (length(.posIds) > 0L) {
        envRef$ids <- sort(unique(c(envRef$ids, .posIds)))
      }
      if (length(.negIds) > 0L) {
        envRef$ids <- setdiff(envRef$ids, .negIds)
      }
      .doResize   <- xIsRxEt && (length(.addedIds) > 0L || length(.removedIds) > 0L)
    }
    envRef$show["id"] <- TRUE
    .resolvedId <- .posIds
  }
  .targetIds <- .resolvedId
  if (is.null(.targetIds) && xIsRxEt && length(envRef$ids) > 0L) {
    .targetIds <- envRef$ids
  }
  list(resolvedId = .resolvedId, targetIds = .targetIds, doResize = .doResize,
       addedIds = .addedIds, removedIds = .removedIds, existingIds = .existingIds)
}

#' Handle the EVID expression
#'
#' @param evidExpr The evid expression generated from `substitute(evid)`
#'
#' @param evidSym The evid symbol generated from `as.character(evidExpr)`
#'
#' @param envir environment where the evaluation occurs
#'
#' @param evidMissing Is the `evid` argument to `et()` missing
#'
#' @return evid value or NULL
#' @noRd
#' @author Matthew L. Fidler
.etHandleEvid <- function(evidExpr, evidSym, envir, evidMissing) {
  .evidVal <- NULL
  if (!evidMissing) {
    .evidVal <- switch(evidSym,
      obs       = 0L,
      `0`       = 0L,
      dose      = 1L,
      `1`       = 1L,
      other     = 2L,
      `2`       = 2L,
      reset     = 3L,
      `3`       = 3L,
      doseReset = 4L,
      resetDose = 4L,
      `4`       = 4L,
      as.integer(tryCatch(eval(evidExpr, envir = envir),
                          error = function(e) as.integer(evidSym)))
    )
  }
  .evidVal
}
#' Handle cmt in et(cmt=...)
#'
#' @param cmtExpr Compartment expression (from `substitute(cmt)`)
#'
#' @param cmtSym Compartment symbol (from `as.character(cmtExpr)`)
#'
#' @param posCmt When et(1, "cmt") this is the position of the cmt
#'   argument
#'
#' @param envir environment where evaluations are made
#'
#' @param cmtMissing logical value that tells if the cmt argument of
#'   `et()` is missing
#'
#' @return The compartment value
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.etHandleCmt <- function(cmtExpr, cmtSym, posCmt, envir, cmtMissing) {
  .cmtVal <- posCmt
  if (!cmtMissing) {
    .cmtTry <- tryCatch(eval(cmtExpr, envir = envir), error = function(e) cmtSym)
    .cmtVal <- if (is.character(.cmtTry) || is.numeric(.cmtTry)) .cmtTry else cmtSym
  }
  .cmtVal
}
#' Handle unit conversion for time and dosing valuess
#'
#' @param val the value to convert, which may be a numeric or a units
#'   object
#'
#' @param unitType a character indicating the type of unit to convert,
#'   either "time" or "dosing"
#'
#' @param envRef The environment for the event table
#'
#' @return the numeric value after converting units if necessary, or
#'   the original numeric value if no conversion is needed
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.etConvertUnits <- function(val, unitType, envRef) {
  if (inherits(val, "units") && requireNamespace("units", quietly = TRUE)) {
    .u <- envRef$units[unitType]
    if (!is.na(.u) && nchar(.u) > 0) {
      return(as.numeric(units::set_units(val, .u, mode = "standard")))
    }
    return(as.numeric(val))
  }
  as.numeric(val)
}
#' Handle the dose value, including unit conversion if needed
#'
#' @param expr the expression for the dose value
#'
#' @param missing a logical indicating whether the dose value is missing
#'
#' @param envRef the reference to the internal environment of the rxEt
#'   object being constructed, used to access unit information for
#'   conversion
#'
#' @param envir the environment in which to evaluate the dose expression
#'
#' @param unitType a character indicating the type of unit to convert
#'   to, either "time" or "rate"
#'
#' @return the numeric dose value after evaluating the expression and
#'   converting units if necessary
#'
#' @noRd
#' @author Matthew L. Fidler
.etHandleDoseValue <- function(expr, missing, envRef, envir, unitType = "time") {
  if (missing) return(0.0)
  .val <- eval(expr, envir = envir)
  if (unitType == "rate") {
    if (inherits(.val, "units") && requireNamespace("units", quietly = TRUE)) {
      .du <- envRef$units["dosing"]
      .tu <- envRef$units["time"]
      if (!is.na(.du) && nchar(.du) > 0 && !is.na(.tu) && nchar(.tu) > 0) {
        return(as.numeric(units::set_units(.val, paste0(.du, "/", .tu), mode = "standard")))
      }
    }
    return(as.numeric(.val))
  }
  .etConvertUnits(.val, unitType, envRef)
}
#' Handle dose and related arguments
#'
#' @param amt Amount argument
#'
#' @param amtExpr Amount expression
#'
#' @param dotArgs the ... args (list)
#'
#' @param envRef the rxEt environment
#'
#' @param envir the evaluation environment
#'
#' @param time the time argument
#'
#' @param timeExpr the time expression
#'
#' @param iiExpr the ii expression
#'
#' @param addlExpr the addl expression
#'
#' @param ssExpr the ss expression
#'
#' @param rateExpr the rate expression
#'
#' @param durExpr the dur expression
#'
#' @param untilExpr the until expression
#'
#' @param evidVal the evid value
#'
#' @param cmtVal the cmt value
#'
#' @param resolvedId the resolved id value. When missing it defaults
#'   to the ids present in the id environment
#'
#' @param targetIds The target ids
#'
#' @param addSampling the add sampling argument
#'
#' @param et the current state of the event table being constructed,
#'   which may be modified by adding a new chunk of observation times
#'   if the sequence arguments are successfully resolved.
#'
#' @param rateSym rate symbol (ie character)
#'
#' @param amtMissing is the amt argument missing
#'
#' @param timeMissing is the time argument missing
#'
#' @param iiMissing is the ii argument missing
#'
#' @param addlMissing is the addl argument missing
#'
#' @param ssMissing is the ss argument missing
#'
#' @param rateMissing is the rate argument missing
#'
#' @param durMissing is the duration argument missing?
#'
#' @param untilMissing is the until argument missing?
#'
#' @param addSamplingMissing is the addSampling argument missing?
#'
#' @return A list with components done (a logical indicating whether
#'   the dose handling was successful and the main et() function
#'   should return immediately) and et (the possibly modified event
#'   table after handling the dose arguments)
#'
#' @noRd
#'
#' @author Matthew L. Fidler
#'
.etHandleDose <- function(amt, amtExpr, dotArgs, envRef, envir, time,
                          timeExpr, iiExpr, addlExpr, ssExpr, rateExpr,
                          durExpr, untilExpr, evidVal, cmtVal, resolvedId,
                          targetIds, addSampling, et, rateSym, amtMissing,
                          timeMissing, iiMissing, addlMissing, ssMissing,
                          rateMissing, durMissing, untilMissing,
                          addSamplingMissing) {
  if (!is.null(amt) || !is.null(dotArgs[["dose"]])) {
    if (!is.null(amt)) {
      .amtVal <- amt
    } else {
      .amtVal <- dotArgs[["dose"]]
    }
    .amtVal <- .etConvertUnits(.amtVal, "dosing", envRef)

    if (!is.null(time)) {
      .timeVal <- time
    } else {
      .timeVal <- 0
    }
    .iiVal   <- .etHandleDoseValue(iiExpr, iiMissing, envRef, envir, "time")
    .addlVal <- if (!addlMissing) as.integer(eval(addlExpr, envir = envir)) else 0L
    .ssVal   <- if (!ssMissing) as.integer(eval(ssExpr, envir = envir)) else 0L

    if (!rateMissing) {
      if (rateSym == "model") {
        .rateVal <- -1.0 # model -> -1, matching NONMEM conventions
      } else if (rateSym == "dur") {
        .rateVal <- -2.0 # dur -> -2, matching NONMEM conventions
      } else {
        .rateVal <- .etHandleDoseValue(rateExpr, FALSE, envRef, envir, "rate")
      }
    } else {
      .rateVal <- 0.0
    }
    if (!durMissing) {
      .durVal   <-  as.numeric(eval(durExpr, envir = envir))
    } else {
      .durVal <- 0.0
    }
    if (!untilMissing) {
      .untilVal <- .etHandleDoseValue(untilExpr, FALSE, envRef, envir, "time")
    } else {
      .untilVal <- NULL
    }

    .df <- .etDoseChunk(time = .timeVal, # nolint
                        amt = .amtVal,
                        evid = ifelse((!is.null(evidVal)), evidVal, 1L),
                        cmt  = ifelse((!is.null(cmtVal)), cmtVal, "(default)"),
                        ii   = .iiVal, addl = .addlVal, ss = .ssVal,
                        rate = .rateVal, dur = .durVal,
                        until = .untilVal)

    if (!durMissing && rateMissing && any(.durVal > 0, na.rm = TRUE)) {
      .df$rate <- rep_len(0.0, nrow(.df))
      .df$dur <- rep_len(.durVal, nrow(.df))
    }
    .pairDoseIds <- !is.null(resolvedId) &&
      nrow(.df) > 1L &&
      nrow(.df) == length(resolvedId)
    if (.pairDoseIds) {
      .df$id <- as.integer(resolvedId)
      envRef$chunks <- .addRowsToChunks(envRef$chunks, .df) # nolint
    } else {
      .etAddChunk(envRef, .df, targetIds) # nolint
    }
    .pairIdsN <- if (.pairDoseIds) 1L else max(1L, length(targetIds))
    envRef$ndose  <- envRef$ndose + max(1L, nrow(.df)) * .pairIdsN
    envRef$show["amt"] <- TRUE
    if (!is.null(.df$ii) && any(.df$ii > 0, na.rm = TRUE)) {
      envRef$show["ii"] <- TRUE
    }
    if (!is.null(.df$addl) && any(.df$addl > 0L, na.rm = TRUE)) {
      envRef$show["addl"] <- TRUE
    }
    if (!is.null(.ssVal) && .ssVal > 0L) {
      envRef$show["ss"] <- TRUE
    }
    if (!is.null(.df$rate) && any(.df$rate != 0, na.rm = TRUE)) {
      envRef$show["rate"] <- TRUE
    }
    if (!is.null(.df$dur) && any(.df$dur != 0, na.rm = TRUE)) {
      envRef$show["dur"] <- TRUE
    }
    if (!is.null(cmtVal) && cmtVal != "(default)") {
      envRef$show["cmt"] <- TRUE
    }
    if (!addSamplingMissing && isTRUE(addSampling)) {
      .obsChunk <- .etObsChunk(.timeVal) # nolint
      .etAddChunk(envRef, .obsChunk, targetIds) # nolint
      envRef$nobs   <- envRef$nobs + length(.obsChunk$time) * max(1L, length(targetIds))
    }
    return(list(done = TRUE, et = et))
  }
  list(done = FALSE)
}
#' This handles infusions where amount isn't specified
#'
#' @param amt the amount argument passed to et(), which is expected to
#'   be NULL for this handler to do anything
#'
#' @param amtExpr the expression for the amount argument, generated
#'   from `substitute(amt)`
#'
#' @param rateExpr the expression for the rate argument, generated
#'   from `substitute(rate)`
#'
#' @param ssExpr the expression for the ss argument, generated from
#'   `substitute(ss)`
#'
#' @param envRef the reference to the internal environment of the rxEt
#'   object being constructed, used to access unit information for
#'   conversion
#'
#' @param envir the environment in which to evaluate any expressions
#'   needed to handle the infusion dose
#'
#' @param time the time argument passed to `et()`, which may be used as
#'   the time for the infusion dose
#'
#' @param timeExpr the expression for the time argument, generated
#'   from `substitute(time)`
#'
#' @param iiExpr the expression for the ii argument, generated from
#'   `substitute(ii)`
#'
#' @param durExpr the expression for the dur argument, generated from
#'   `substitute(dur)`
#'
#' @param evidVal the evid value to use for the infusion dose, which
#'   may be NULL
#'
#' @param cmtVal the cmt value to use for the infusion dose, which may
#'   be NULL
#'
#' @param targetIds the target ids for the infusion dose, which may be
#'   NULL to indicate
#'
#' @param et the current state of the event table being constructed, which may
#'  be modified by adding a new chunk for the infusion dose if the
#' handling is successful
#'
#' @param rateSym the rate symbol (ie character) for the infusion dose
#'
#' @param amtMissing is the amt argument missing?
#'
#' @param rateMissing is the rate argument missing?
#'
#' @param ssMissing is the ss argument missing?
#'
#' @param timeMissing is the time argument missing?
#'
#' @param iiMissing is the ii argument missing?
#'
#' @param durMissing is the dur argument missing?
#'
#' @return A list with components done
#'
#' @noRd
#' @author Matthew L. Fidler
.etHandleInfusionNoAmt <- function(amt, amtExpr, rateExpr, ssExpr, envRef, envir,
                                   time, timeExpr, iiExpr, durExpr, evidVal,
                                   cmtVal, targetIds, et, rateSym,
                                   amtMissing, rateMissing, ssMissing,
                                   timeMissing, iiMissing, durMissing) {
  if (is.null(amt) && (!rateMissing || !ssMissing)) {
    .timeVal <- if (!is.null(time)) time else 0.0
    .iiVal   <- if (!iiMissing)   as.numeric(eval(iiExpr, envir = envir))   else 0.0
    .ssVal   <- if (!ssMissing)   as.integer(eval(ssExpr, envir = envir))   else 0L
    if (!rateMissing) {
      .rateVal <- switch(rateSym,
        model = -1.0,
        dur   = -2.0,
        as.numeric(eval(rateExpr, envir = envir))
      )
    } else {
      .rateVal <- 0.0
    }
    if (!durMissing) {
      .durVal <- as.numeric(eval(durExpr, envir = envir))
    } else {
      .durVal <- 0.0
    }
    if (!is.null(evidVal)) {
      .evid2 <- evidVal
    } else {
      .evid2 <- 1L
    }
    if (!is.null(cmtVal)) {
      .cmtVal <- cmtVal
    } else {
      .cmtVal <- "(default)"
    }
    .df <- .etDoseChunk( # nolint
      time = .timeVal, amt = 0.0,
      evid = .evid2,
      cmt  = .cmtVal,
      ii = .iiVal, addl = 0L, ss = .ssVal,
      rate = .rateVal, dur = .durVal
    )
    .etAddChunk(envRef, .df, targetIds) # nolint
    envRef$ndose  <- envRef$ndose + max(1L, nrow(.df)) * max(1L, length(targetIds))
    envRef$show["amt"]  <- TRUE
    if (!is.null(.ssVal) && .ssVal > 0L)   envRef$show["ss"]   <- TRUE
    if (!is.null(.df$rate) && any(.df$rate != 0, na.rm = TRUE)) envRef$show["rate"] <- TRUE
    return(list(done = TRUE, et = et))
  }
  list(done = FALSE)
}
#' Handle observation times specified in the position arguments or time
#'
#'
#' @param time the time argument passed to et(), which may be used as
#'   the observation time if not NULL
#'
#' @param timeExpr the expression for the time argument, generated
#'   from `substitute(time)`, which may be evaluated to get the
#'   observation time if time is NULL
#'
#' @param envir the environment in which to evaluate any expressions
#'
#' @param envRef the reference to the internal environment of the rxEt
#'   object
#'
#' @param evidVal the evid value to use for the observation events,
#'   which may be NULL
#'
#' @param cmtVal the cmt value to use for the observation events,
#'   which may be NULL
#'
#' @param targetIds the target ids for the observation events, which may be NULL to
#'  indicate all ids
#'
#' @param et the current state of the event table being constructed, which may
#'  be modified by adding a new chunk for the observation events
#'
#' @param timeMissing is the time argument missing?
#'
#' @return A list with components done
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.etHandleObs <- function(time, timeExpr, envir, envRef, evidVal, cmtVal,
                         targetIds, et, timeMissing) {
  if (!is.null(time) || !timeMissing) {
    if (!is.null(time)) {
      .timeVal <-  time
    } else {
      .timeVal <- eval(timeExpr, envir = envir)
    }

    if (!is.list(.timeVal) &&
          inherits(.timeVal, "units") &&
          requireNamespace("units", quietly = TRUE)) {
      .tu2 <- envRef$units["time"]
      if (!is.na(.tu2) && nchar(.tu2) > 0) {
        .timeVal <- as.numeric(units::set_units(.timeVal, .tu2, mode = "standard"))
      } else {
        .timeVal <- as.numeric(.timeVal)
      }
    }
    if (!is.null(evidVal)) {
      .evid2 <- evidVal
    } else {
      .evid2 <- 0L
    }
    .df <- .etObsChunk(.timeVal, cmt = cmtVal) # nolint
    if (.evid2 != 0L) {
      .df$evid <- as.integer(.evid2)
    }
    .etAddChunk(envRef, .df, targetIds) # nolint
    envRef$nobs <- envRef$nobs + length(.df$time) * max(1L, length(targetIds))
    if (!is.null(cmtVal)) {
      envRef$show["cmt"] <- TRUE
    }
    return(list(done = TRUE, et = et))
  }
  list(done = FALSE)
}
#' Handle piping types of calls
#'
#'
#' @param xIsRxEt is the x argument an rxEt object, implying piping of some sort
#'
#' @param x the x argument passed to et()
#'
#' @param time the time argument passed to et()
#'
#' @param timeExpr the expression for the time argument, generated from `substitute(time)`
#'
#' @param amt the amt argument passed to et()
#'
#' @param amtExpr the expression for the amt argument, generated from `substitute(amt)`
#'
#' @param dotArgs the ... arguments passed to et(), which may contain
#'   named or unnamed arguments
#'
#' @param cmtVal the cmt value to use for any events created from the
#'   piping arguments, which may be NULL
#'
#' @param targetIds the target ids for any events created from the
#'   piping arguments, which may be NULL to indicate all ids
#'
#' @param evidVal the evid value to use for any events created from the piping
#'
#' @param envRef the reference to the internal environment of the rxEt object
#'
#' @param et the current state of the event table
#'
#' @param timeMissing is the time argument missing?
#'
#' @param amtMissing is the amt argument missing?
#'
#' @param envir the environment in which to evaluate any expressions needed to handle
#'   the piping arguments
#'
#' @return A list with components done and et, where done is a logical
#'   indicating whether the piping arguments were successfully handled
#'   and the main et()
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.etHandlePiping <- function(xIsRxEt, x, time, timeExpr, amt, amtExpr, # nolint
                            dotArgs, cmtVal, targetIds, evidVal,
                            envRef, et, timeMissing, amtMissing, envir) {
  if (xIsRxEt &&
        is.null(time) &&
        is.null(amt) && timeMissing && amtMissing) {
    if (length(dotArgs) >= 1) {
      .firstDot <- dotArgs[[1]]
      if (is.rxEt(.firstDot) || # nolint
            any(vapply(dotArgs, is.rxEt, logical(1)))) {
        # Sequence of event tables
        .ret <- do.call(etSeq, c(list(x), dotArgs)) # nolint
        return(list(done = TRUE, et = .ret))
      }
      if (length(dotArgs) > 2) {
        stop("unused positional arguments", call. = FALSE)
      }
      if (is.list(.firstDot) && !is.data.frame(.firstDot)) {
        .df <- .etObsChunk(.firstDot, cmt = cmtVal) # nolint
        .etAddChunk(envRef, .df, targetIds) # nolint
        envRef$nobs <- envRef$nobs + length(.df$time) * max(1L, length(targetIds))
        if (!is.null(cmtVal)) {
          envRef$show["cmt"] <- TRUE
        }
        return(list(done = TRUE, et = et))
      } else if (is.numeric(.firstDot) || is.integer(.firstDot)) {
        if (length(dotArgs) >= 2L &&
              length(.firstDot) == 1L &&
              (is.numeric(dotArgs[[2]]) || is.integer(dotArgs[[2]])) &&
              length(dotArgs[[2]]) == 1L) {
          # et(0, 10) add obs at times 0, 1, ..., 10
          .timeVec <- seq(from = as.numeric(.firstDot), to = as.numeric(dotArgs[[2]]))
        } else if (inherits(.firstDot, "units") &&
                     requireNamespace("units", quietly = TRUE)) {
          .tu2 <- envRef$units["time"]
          if (!is.na(.tu2) && nchar(.tu2) > 0) {
            .timeVec <- as.numeric(units::set_units(.firstDot, .tu2, mode = "standard"))
          } else {
            .timeVec <- as.numeric(.firstDot)
          }
        } else {
          .timeVec <- as.numeric(.firstDot)
        }
        .df <- .etObsChunk(.timeVec, cmt = cmtVal) # nolint
        if (!is.null(evidVal) && evidVal != 0L) {
          .df$evid <- as.integer(evidVal)
        }
        .etAddChunk(envRef, .df, targetIds) # nolint
        envRef$nobs <- envRef$nobs + length(.timeVec) * max(1L, length(targetIds))
        return(list(done = TRUE, et = et))
      }
    }
  }
  list(done = FALSE)
}
#' Handle etSeq() calls
#'
#'
#' @param item the item argument passed to etSeq(), which is expected
#'   to be an rxEt object
#'
#' @param units the units argument passed to etSeq(), which may be NULL
#'
#' @param show the show argument passed to etSeq(), which may be NULL
#'
#' @param ids the ids argument passed to etSeq(), which may be NULL
#'
#' @param timeDelta the timeDelta argument passed to etSeq(), which is
#'   expected to be a numeric value
#'
#' @param samples the samples argument passed to etSeq(), which is
#'   expected to be a character value of either "use" or "dose"
#'
#' @param chunks the chunks argument passed to etSeq(), which is
#'   expected to be a list of data frames
#'
#' @param nobs the nobs argument passed to etSeq(), which is expected
#'   to be an integer value
#'
#' @param ndose the ndose argument passed to etSeq(), which is expected to be an
#'   integer value
#'
#' @param explicitIi the explicitIi argument passed to etSeq(), which
#'   is expected to be a logical value
#'
#' @param ii the ii argument passed to etSeq(), which is expected to be a numeric value
#'
#' @return a list with components units, show, ids, chunks, nobs, ndose, timeDelta,
#'  lastIi, and lastDose, which represent the updated state of the event
#'
#' @noRd
#'
#' @author Matthew L. Fidler
.etSeqHandleRxEt <- function(item, units, show, ids, timeDelta, samples, chunks, groups,
                             nobs, ndose, explicitIi, ii) {
  env <- .rxEtEnv(item) # nolint
  if (is.null(units)) {
    units <- env$units
    show  <- env$show
  } else {
    show <- show | env$show
  }
  .itemGroups <- .etGetGroups(env) # nolint
  .keepGrouped <- length(chunks) == 0L &&
    length(.itemGroups) == 1L &&
    (length(groups) == 0L ||
       (length(groups) == 1L && .etGroupIdsEqual(groups[[1]]$ids, .itemGroups[[1]]$ids))) # nolint

  if (!.keepGrouped && length(groups) > 0L) {
    for (.g in groups) {
      chunks <- .addRowsToChunks(chunks, .etMaterializeGroup(.g)) # nolint
    }
    groups <- list()
  }

  ids <- sort(unique(c(ids, env$ids)))
  if (.keepGrouped) {
    .summaryDf <- .etShiftChunk(.itemGroups[[1]]$data, timeDelta) # nolint
    if (samples == "use") {
      .groupDf <- .summaryDf
      nobs <- nobs + env$nobs
    } else {
      .groupDf <- .summaryDf[.summaryDf$evid != 0L, , drop = FALSE]
    }
    if (nrow(.groupDf) > 0L) {
      if (length(groups) == 0L) {
        groups <- list(list(ids = as.integer(.itemGroups[[1]]$ids), data = .groupDf))
      } else {
        groups[[1]]$data <- as.data.frame(
          data.table::rbindlist(list(groups[[1]]$data, .groupDf), fill = TRUE)
        )
      }
    }
    mat <- .summaryDf
  } else {
    mat <- .etShiftChunk(.etMaterialize(item), timeDelta) # nolint
    if (samples == "use") {
      chunks <- .addRowsToChunks(chunks, mat) # nolint
      nobs   <- nobs + env$nobs
    } else {
      doseOnly <- mat[mat$evid != 0L, , drop = FALSE]
      chunks <- .addRowsToChunks(chunks, doseOnly) # nolint
    }
  }

  lastIi <- 0.0
  lastDose <- 0.0
  doseRows <- mat[mat$evid != 0L, , drop = FALSE]
  if (nrow(doseRows) > 0L) {
    lastDoseRow <- doseRows[nrow(doseRows), ]
    if (is.null(lastDoseRow$addl) || is.na(lastDoseRow$addl)) {
      addlVal <- 0L
    } else {
      addlVal <- as.integer(lastDoseRow$addl)
    }
    if (addlVal > 0L) {
      lastIi <- lastDoseRow$ii
    }
    lastDose <- lastDoseRow$time + addlVal * lastIi
  }
  maxTime <- max(mat$time, na.rm = TRUE)
  ndose <- ndose + env$ndose
  # Advance past last dose period; also respect max obs time (for samples="use")
  if (!(explicitIi && identical(ii, 0))) {
    if (lastIi > 0) {
      effectiveIi <- lastIi
    } else {
      effectiveIi <- ii
    }
    timeDelta <- max(maxTime, lastDose + effectiveIi)
  }
  list(units = units, show = show, ids = ids, chunks = chunks, groups = groups,
       nobs = nobs, ndose = ndose, timeDelta = timeDelta,
       lastIi = lastIi, lastDose = lastDose)
}
#' Handle wait
#'
#'
#' @param item the item argument passed to etSeq(), which is expected
#'   to be a numeric value representing the wait time
#'
#' @param waitType the waitType argument passed to etSeq(), which is
#'   expected to be a character value of either "+ii" or "smart"
#'
#' @param lastDose the lastDose value from the previous item in the
#'   etSeq() sequence
#'
#' @param lastIi the lastIi value from the previous item in the
#'   etSeq() sequence
#'
#' @param ii the ii argument passed to etSeq(), which is expected to
#'   be a numeric value representing the inter-dose interval
#'
#' @param timeDelta the current timeDelta value in the etSeq() sequence, which may be
#'   updated by this function to reflect the wait time
#'
#' @return the updated timeDelta value after applying the wait time
#'   according to the specified waitType
#' @noRd
#'
#' @author Matthew L. Fidler
.etSeqHandleWait <- function(item, waitType, lastDose, lastIi, ii, timeDelta) {
  wait        <- as.numeric(item)
  effectiveIi <- if (lastIi > 0) lastIi else ii
  if (waitType == "+ii") {
    timeDelta <- lastDose + effectiveIi + wait
  } else {  # smart
    if (wait < effectiveIi) {
      timeDelta <- lastDose + effectiveIi
    } else {
      timeDelta <- lastDose + wait
    }
  }
  timeDelta
}

#' Parametric ReLU Activation Function
#'
#' @family Activation Functions
#' @param x A numeric vector. All elements must be finite and
#'   non-missing.
#' @param alpha A numeric scalar. All elements must be finite and
#'  non-missing.
#' @return A numeric vector where the ReLU function has been applied
#'   to each element of `x`.
#' @author Matthew Fidler
#' @export
#' @examples
#'
#' PReLU(c(-1, 0, 1, 2), 2)
#'
#' \donttest{
#' # Can also be used in rxode2:
#' x <- rxode2({
#'    r=PReLU(time, 2)
#' })
#'
#' e <- et(c(-1, 0, 1, 2))
#'
#' rxSolve(x, e)
#' }
#'
PReLU <- function(x, alpha=1) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  checkmate::assertNumeric(alpha, finite=TRUE, any.missing=FALSE)
  .df <- data.frame(x=x, alpha=alpha)
  .Call(`_rxode2_activationF2`, .df$x, .df$alpha, 7L)
}
#' Derivatives Parametric ReLU Activation Function
#'
#'
#' @param x A numeric vector. All elements must be finite and
#'  non-missing.
#' @param alpha A numeric scalar. All elements must be finite and
#'  non-missing.
#' @return A numeric vector where the derivative(s) of the ELU function has been applied
#'  to each element of `x`.
#' @export
#' @author Matthew L. Fidler
#' @family Activation Functions
#' @examples
#'
#' dPReLU(c(-1, 0, 1, 2), 2)
#' dPReLUa(c(-1, 0, 1, 2), 2)
#' dPReLUa1(c(-1, 0, 1, 2), 2)
#'
#'
#' \donttest{
#' # Can also be used in rxode2:
#' r <- rxode2({
#'   r1=dPReLU(time, 2)
#'   r2a=dPReLUa(time, 2)
#'   ra=dPReLUa1(time, 2)
#' })
#'
#' e <- et(c(-1, 0, 1, 2))
#' rxSolve(r, e)
#' }
dPReLU <- function(x, alpha=1) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  checkmate::assertNumeric(alpha, finite=TRUE, any.missing=FALSE)
  .df <- data.frame(x=x, alpha=alpha)
  .Call(`_rxode2_activationF2`, .df$x, .df$alpha, 8L)
}

#' @rdname dPReLU
#' @export
dPReLUa <- function(x, alpha=1) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  checkmate::assertNumeric(alpha, finite=TRUE, any.missing=FALSE)
  .df <- data.frame(x=x, alpha=alpha)
  .Call(`_rxode2_activationF2`, .df$x, .df$alpha, 9L)
}

#' @rdname dPReLU
#' @export
dPReLUa1 <- function(x, alpha=1) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  checkmate::assertNumeric(alpha, finite=TRUE, any.missing=FALSE)
  .df <- data.frame(x=x, alpha=alpha)
  .Call(`_rxode2_activationF2`, .df$x, .df$alpha, 10L)
}

*DECK DGEFA
      SUBROUTINE DGEFA (A, LDA, N, IPVT, INFO)
C***BEGIN PROLOGUE  DGEFA
C***PURPOSE  Factor a matrix using Gaussian elimination.
C***CATEGORY  D2A1
C***TYPE      DOUBLE PRECISION (SGEFA-S, DGEFA-D, CGEFA-C)
C***KEYWORDS  GENERAL MATRIX, LINEAR ALGEBRA, LINPACK,
C             MATRIX FACTORIZATION
C***AUTHOR  Moler, C. B., (U. of New Mexico)
C***DESCRIPTION
C
C     DGEFA factors a double precision matrix by Gaussian elimination.
C
C     DGEFA is usually called by DGECO, but it can be called
C     directly with a saving in time if  RCOND  is not needed.
C     (Time for DGECO) = (1 + 9/N)*(Time for DGEFA) .
C
C     On Entry
C
C        A       DOUBLE PRECISION(LDA, N)
C                the matrix to be factored.
C
C        LDA     INTEGER
C                the leading dimension of the array  A .
C
C        N       INTEGER
C                the order of the matrix  A .
C
C     On Return
C
C        A       an upper triangular matrix and the multipliers
C                which were used to obtain it.
C                The factorization can be written  A = L*U  where
C                L  is a product of permutation and unit lower
C                triangular matrices and  U  is upper triangular.
C
C        IPVT    INTEGER(N)
C                an integer vector of pivot indices.
C
C        INFO    INTEGER
C                = 0  normal value.
C                = K  if  U(K,K) .EQ. 0.0 .  This is not an error
C                     condition for this subroutine, but it does
C                     indicate that DGESL or DGEDI will divide by zero
C                     if called.  Use  RCOND  in DGECO for a reliable
C                     indication of singularity.
C
C***REFERENCES  J. J. Dongarra, J. R. Bunch, C. B. Moler, and G. W.
C                 Stewart, LINPACK Users' Guide, SIAM, 1979.
C***ROUTINES CALLED  DAXPY, DSCAL, IDAMAX
C***REVISION HISTORY  (YYMMDD)
C   780814  DATE WRITTEN
C   890831  Modified array declarations.  (WRB)
C   890831  REVISION DATE from Version 3.2
C   891214  Prologue converted to Version 4.0 format.  (BAB)
C   900326  Removed duplicate information from DESCRIPTION section.
C           (WRB)
C   920501  Reformatted the REFERENCES section.  (WRB)
C***END PROLOGUE  DGEFA
      INTEGER LDA,N,IPVT(*),INFO
      DOUBLE PRECISION A(LDA,*)
C
      DOUBLE PRECISION T
      INTEGER IDAMAX,J,K,KP1,L,NM1
C
C     GAUSSIAN ELIMINATION WITH PARTIAL PIVOTING
C
C***FIRST EXECUTABLE STATEMENT  DGEFA
      INFO = 0
      NM1 = N - 1
      IF (NM1 .LT. 1) GO TO 70
      DO 60 K = 1, NM1
         KP1 = K + 1
C
C        FIND L = PIVOT INDEX
C
         L = IDAMAX(N-K+1,A(K,K),1) + K - 1
         IPVT(K) = L
C
C        ZERO PIVOT IMPLIES THIS COLUMN ALREADY TRIANGULARIZED
C
         IF (A(L,K) .EQ. 0.0D0) GO TO 40
C
C           INTERCHANGE IF NECESSARY
C
            IF (L .EQ. K) GO TO 10
               T = A(L,K)
               A(L,K) = A(K,K)
               A(K,K) = T
   10       CONTINUE
C
C           COMPUTE MULTIPLIERS
C
            T = -1.0D0/A(K,K)
            CALL DSCAL(N-K,T,A(K+1,K),1)
C
C           ROW ELIMINATION WITH COLUMN INDEXING
C
            DO 30 J = KP1, N
               T = A(L,J)
               IF (L .EQ. K) GO TO 20
                  A(L,J) = A(K,J)
                  A(K,J) = T
   20          CONTINUE
               CALL DAXPY(N-K,T,A(K+1,K),1,A(K+1,J),1)
   30       CONTINUE
         GO TO 50
   40    CONTINUE
            INFO = K
   50    CONTINUE
   60 CONTINUE
   70 CONTINUE
      IPVT(N) = N
      IF (A(N,N) .EQ. 0.0D0) INFO = N
      RETURN
      END
*DECK DGESL
      SUBROUTINE DGESL (A, LDA, N, IPVT, B, JOB)
C***BEGIN PROLOGUE  DGESL
C***PURPOSE  Solve the real system A*X=B or TRANS(A)*X=B using the
C            factors computed by DGECO or DGEFA.
C***CATEGORY  D2A1
C***TYPE      DOUBLE PRECISION (SGESL-S, DGESL-D, CGESL-C)
C***KEYWORDS  LINEAR ALGEBRA, LINPACK, MATRIX, SOLVE
C***AUTHOR  Moler, C. B., (U. of New Mexico)
C***DESCRIPTION
C
C     DGESL solves the double precision system
C     A * X = B  or  TRANS(A) * X = B
C     using the factors computed by DGECO or DGEFA.
C
C     On Entry
C
C        A       DOUBLE PRECISION(LDA, N)
C                the output from DGECO or DGEFA.
C
C        LDA     INTEGER
C                the leading dimension of the array  A .
C
C        N       INTEGER
C                the order of the matrix  A .
C
C        IPVT    INTEGER(N)
C                the pivot vector from DGECO or DGEFA.
C
C        B       DOUBLE PRECISION(N)
C                the right hand side vector.
C
C        JOB     INTEGER
C                = 0         to solve  A*X = B ,
C                = nonzero   to solve  TRANS(A)*X = B  where
C                            TRANS(A)  is the transpose.
C
C     On Return
C
C        B       the solution vector  X .
C
C     Error Condition
C
C        A division by zero will occur if the input factor contains a
C        zero on the diagonal.  Technically this indicates singularity
C        but it is often caused by improper arguments or improper
C        setting of LDA .  It will not occur if the subroutines are
C        called correctly and if DGECO has set RCOND .GT. 0.0
C        or DGEFA has set INFO .EQ. 0 .
C
C     To compute  INVERSE(A) * C  where  C  is a matrix
C     with  P  columns
C           CALL DGECO(A,LDA,N,IPVT,RCOND,Z)
C           IF (RCOND is too small) GO TO ...
C           DO 10 J = 1, P
C              CALL DGESL(A,LDA,N,IPVT,C(1,J),0)
C        10 CONTINUE
C
C***REFERENCES  J. J. Dongarra, J. R. Bunch, C. B. Moler, and G. W.
C                 Stewart, LINPACK Users' Guide, SIAM, 1979.
C***ROUTINES CALLED  DAXPY, DDOT
C***REVISION HISTORY  (YYMMDD)
C   780814  DATE WRITTEN
C   890831  Modified array declarations.  (WRB)
C   890831  REVISION DATE from Version 3.2
C   891214  Prologue converted to Version 4.0 format.  (BAB)
C   900326  Removed duplicate information from DESCRIPTION section.
C           (WRB)
C   920501  Reformatted the REFERENCES section.  (WRB)
C***END PROLOGUE  DGESL
      INTEGER LDA,N,IPVT(*),JOB
      DOUBLE PRECISION A(LDA,*),B(*)
C
      DOUBLE PRECISION DDOT,T
      INTEGER K,KB,L,NM1
C***FIRST EXECUTABLE STATEMENT  DGESL
      NM1 = N - 1
      IF (JOB .NE. 0) GO TO 50
C
C        JOB = 0 , SOLVE  A * X = B
C        FIRST SOLVE  L*Y = B
C
         IF (NM1 .LT. 1) GO TO 30
         DO 20 K = 1, NM1
            L = IPVT(K)
            T = B(L)
            IF (L .EQ. K) GO TO 10
               B(L) = B(K)
               B(K) = T
   10       CONTINUE
            CALL DAXPY(N-K,T,A(K+1,K),1,B(K+1),1)
   20    CONTINUE
   30    CONTINUE
C
C        NOW SOLVE  U*X = Y
C
         DO 40 KB = 1, N
            K = N + 1 - KB
            B(K) = B(K)/A(K,K)
            T = -B(K)
            CALL DAXPY(K-1,T,A(1,K),1,B(1),1)
   40    CONTINUE
      GO TO 100
   50 CONTINUE
C
C        JOB = NONZERO, SOLVE  TRANS(A) * X = B
C        FIRST SOLVE  TRANS(U)*Y = B
C
         DO 60 K = 1, N
            T = DDOT(K-1,A(1,K),1,B(1),1)
            B(K) = (B(K) - T)/A(K,K)
   60    CONTINUE
C
C        NOW SOLVE TRANS(L)*X = Y
C
         IF (NM1 .LT. 1) GO TO 90
         DO 80 KB = 1, NM1
            K = N - KB
            B(K) = B(K) + DDOT(N-K,A(K+1,K),1,B(K+1),1)
            L = IPVT(K)
            IF (L .EQ. K) GO TO 70
               T = B(L)
               B(L) = B(K)
               B(K) = T
   70       CONTINUE
   80    CONTINUE
   90    CONTINUE
  100 CONTINUE
      RETURN
      END
*DECK DGBFA
      SUBROUTINE DGBFA (ABD, LDA, N, ML, MU, IPVT, INFO)
C***BEGIN PROLOGUE  DGBFA
C***PURPOSE  Factor a band matrix using Gaussian elimination.
C***CATEGORY  D2A2
C***TYPE      DOUBLE PRECISION (SGBFA-S, DGBFA-D, CGBFA-C)
C***KEYWORDS  BANDED, LINEAR ALGEBRA, LINPACK, MATRIX FACTORIZATION
C***AUTHOR  Moler, C. B., (U. of New Mexico)
C***DESCRIPTION
C
C     DGBFA factors a double precision band matrix by elimination.
C
C     DGBFA is usually called by DGBCO, but it can be called
C     directly with a saving in time if  RCOND  is not needed.
C
C     On Entry
C
C        ABD     DOUBLE PRECISION(LDA, N)
C                contains the matrix in band storage.  The columns
C                of the matrix are stored in the columns of  ABD  and
C                the diagonals of the matrix are stored in rows
C                ML+1 through 2*ML+MU+1 of  ABD .
C                See the comments below for details.
C
C        LDA     INTEGER
C                the leading dimension of the array  ABD .
C                LDA must be .GE. 2*ML + MU + 1 .
C
C        N       INTEGER
C                the order of the original matrix.
C
C        ML      INTEGER
C                number of diagonals below the main diagonal.
C                0 .LE. ML .LT.  N .
C
C        MU      INTEGER
C                number of diagonals above the main diagonal.
C                0 .LE. MU .LT.  N .
C                More efficient if  ML .LE. MU .
C     On Return
C
C        ABD     an upper triangular matrix in band storage and
C                the multipliers which were used to obtain it.
C                The factorization can be written  A = L*U  where
C                L  is a product of permutation and unit lower
C                triangular matrices and  U  is upper triangular.
C
C        IPVT    INTEGER(N)
C                an integer vector of pivot indices.
C
C        INFO    INTEGER
C                = 0  normal value.
C                = K  if  U(K,K) .EQ. 0.0 .  This is not an error
C                     condition for this subroutine, but it does
C                     indicate that DGBSL will divide by zero if
C                     called.  Use  RCOND  in DGBCO for a reliable
C                     indication of singularity.
C
C     Band Storage
C
C           If  A  is a band matrix, the following program segment
C           will set up the input.
C
C                   ML = (band width below the diagonal)
C                   MU = (band width above the diagonal)
C                   M = ML + MU + 1
C                   DO 20 J = 1, N
C                      I1 = MAX(1, J-MU)
C                      I2 = MIN(N, J+ML)
C                      DO 10 I = I1, I2
C                         K = I - J + M
C                         ABD(K,J) = A(I,J)
C                10    CONTINUE
C                20 CONTINUE
C
C           This uses rows  ML+1  through  2*ML+MU+1  of  ABD .
C           In addition, the first  ML  rows in  ABD  are used for
C           elements generated during the triangularization.
C           The total number of rows needed in  ABD  is  2*ML+MU+1 .
C           The  ML+MU by ML+MU  upper left triangle and the
C           ML by ML  lower right triangle are not referenced.
C
C***REFERENCES  J. J. Dongarra, J. R. Bunch, C. B. Moler, and G. W.
C                 Stewart, LINPACK Users' Guide, SIAM, 1979.
C***ROUTINES CALLED  DAXPY, DSCAL, IDAMAX
C***REVISION HISTORY  (YYMMDD)
C   780814  DATE WRITTEN
C   890531  Changed all specific intrinsics to generic.  (WRB)
C   890831  Modified array declarations.  (WRB)
C   890831  REVISION DATE from Version 3.2
C   891214  Prologue converted to Version 4.0 format.  (BAB)
C   900326  Removed duplicate information from DESCRIPTION section.
C           (WRB)
C   920501  Reformatted the REFERENCES section.  (WRB)
C***END PROLOGUE  DGBFA
      INTEGER LDA,N,ML,MU,IPVT(*),INFO
      DOUBLE PRECISION ABD(LDA,*)
C
      DOUBLE PRECISION T
      INTEGER I,IDAMAX,I0,J,JU,JZ,J0,J1,K,KP1,L,LM,M,MM,NM1
C
C***FIRST EXECUTABLE STATEMENT  DGBFA
      M = ML + MU + 1
      INFO = 0
C
C     ZERO INITIAL FILL-IN COLUMNS
C
      J0 = MU + 2
      J1 = MIN(N,M) - 1
      IF (J1 .LT. J0) GO TO 30
      DO 20 JZ = J0, J1
         I0 = M + 1 - JZ
         DO 10 I = I0, ML
            ABD(I,JZ) = 0.0D0
   10    CONTINUE
   20 CONTINUE
   30 CONTINUE
      JZ = J1
      JU = 0
C
C     GAUSSIAN ELIMINATION WITH PARTIAL PIVOTING
C
      NM1 = N - 1
      IF (NM1 .LT. 1) GO TO 130
      DO 120 K = 1, NM1
         KP1 = K + 1
C
C        ZERO NEXT FILL-IN COLUMN
C
         JZ = JZ + 1
         IF (JZ .GT. N) GO TO 50
         IF (ML .LT. 1) GO TO 50
            DO 40 I = 1, ML
               ABD(I,JZ) = 0.0D0
   40       CONTINUE
   50    CONTINUE
C
C        FIND L = PIVOT INDEX
C
         LM = MIN(ML,N-K)
         L = IDAMAX(LM+1,ABD(M,K),1) + M - 1
         IPVT(K) = L + K - M
C
C        ZERO PIVOT IMPLIES THIS COLUMN ALREADY TRIANGULARIZED
C
         IF (ABD(L,K) .EQ. 0.0D0) GO TO 100
C
C           INTERCHANGE IF NECESSARY
C
            IF (L .EQ. M) GO TO 60
               T = ABD(L,K)
               ABD(L,K) = ABD(M,K)
               ABD(M,K) = T
   60       CONTINUE
C
C           COMPUTE MULTIPLIERS
C
            T = -1.0D0/ABD(M,K)
            CALL DSCAL(LM,T,ABD(M+1,K),1)
C
C           ROW ELIMINATION WITH COLUMN INDEXING
C
            JU = MIN(MAX(JU,MU+IPVT(K)),N)
            MM = M
            IF (JU .LT. KP1) GO TO 90
            DO 80 J = KP1, JU
               L = L - 1
               MM = MM - 1
               T = ABD(L,J)
               IF (L .EQ. MM) GO TO 70
                  ABD(L,J) = ABD(MM,J)
                  ABD(MM,J) = T
   70          CONTINUE
               CALL DAXPY(LM,T,ABD(M+1,K),1,ABD(MM+1,J),1)
   80       CONTINUE
   90       CONTINUE
         GO TO 110
  100    CONTINUE
            INFO = K
  110    CONTINUE
  120 CONTINUE
  130 CONTINUE
      IPVT(N) = N
      IF (ABD(M,N) .EQ. 0.0D0) INFO = N
      RETURN
      END
*DECK DGBSL
      SUBROUTINE DGBSL (ABD, LDA, N, ML, MU, IPVT, B, JOB)
C***BEGIN PROLOGUE  DGBSL
C***PURPOSE  Solve the real band system A*X=B or TRANS(A)*X=B using
C            the factors computed by DGBCO or DGBFA.
C***CATEGORY  D2A2
C***TYPE      DOUBLE PRECISION (SGBSL-S, DGBSL-D, CGBSL-C)
C***KEYWORDS  BANDED, LINEAR ALGEBRA, LINPACK, MATRIX, SOLVE
C***AUTHOR  Moler, C. B., (U. of New Mexico)
C***DESCRIPTION
C
C     DGBSL solves the double precision band system
C     A * X = B  or  TRANS(A) * X = B
C     using the factors computed by DGBCO or DGBFA.
C
C     On Entry
C
C        ABD     DOUBLE PRECISION(LDA, N)
C                the output from DGBCO or DGBFA.
C
C        LDA     INTEGER
C                the leading dimension of the array  ABD .
C
C        N       INTEGER
C                the order of the original matrix.
C
C        ML      INTEGER
C                number of diagonals below the main diagonal.
C
C        MU      INTEGER
C                number of diagonals above the main diagonal.
C
C        IPVT    INTEGER(N)
C                the pivot vector from DGBCO or DGBFA.
C
C        B       DOUBLE PRECISION(N)
C                the right hand side vector.
C
C        JOB     INTEGER
C                = 0         to solve  A*X = B ,
C                = nonzero   to solve  TRANS(A)*X = B , where
C                            TRANS(A)  is the transpose.
C
C     On Return
C
C        B       the solution vector  X .
C
C     Error Condition
C
C        A division by zero will occur if the input factor contains a
C        zero on the diagonal.  Technically this indicates singularity
C        but it is often caused by improper arguments or improper
C        setting of LDA .  It will not occur if the subroutines are
C        called correctly and if DGBCO has set RCOND .GT. 0.0
C        or DGBFA has set INFO .EQ. 0 .
C
C     To compute  INVERSE(A) * C  where  C  is a matrix
C     with  P  columns
C           CALL DGBCO(ABD,LDA,N,ML,MU,IPVT,RCOND,Z)
C           IF (RCOND is too small) GO TO ...
C           DO 10 J = 1, P
C              CALL DGBSL(ABD,LDA,N,ML,MU,IPVT,C(1,J),0)
C        10 CONTINUE
C
C***REFERENCES  J. J. Dongarra, J. R. Bunch, C. B. Moler, and G. W.
C                 Stewart, LINPACK Users' Guide, SIAM, 1979.
C***ROUTINES CALLED  DAXPY, DDOT
C***REVISION HISTORY  (YYMMDD)
C   780814  DATE WRITTEN
C   890531  Changed all specific intrinsics to generic.  (WRB)
C   890831  Modified array declarations.  (WRB)
C   890831  REVISION DATE from Version 3.2
C   891214  Prologue converted to Version 4.0 format.  (BAB)
C   900326  Removed duplicate information from DESCRIPTION section.
C           (WRB)
C   920501  Reformatted the REFERENCES section.  (WRB)
C***END PROLOGUE  DGBSL
      INTEGER LDA,N,ML,MU,IPVT(*),JOB
      DOUBLE PRECISION ABD(LDA,*),B(*)
C
      DOUBLE PRECISION DDOT,T
      INTEGER K,KB,L,LA,LB,LM,M,NM1
C***FIRST EXECUTABLE STATEMENT  DGBSL
      M = MU + ML + 1
      NM1 = N - 1
      IF (JOB .NE. 0) GO TO 50
C
C        JOB = 0 , SOLVE  A * X = B
C        FIRST SOLVE L*Y = B
C
         IF (ML .EQ. 0) GO TO 30
         IF (NM1 .LT. 1) GO TO 30
            DO 20 K = 1, NM1
               LM = MIN(ML,N-K)
               L = IPVT(K)
               T = B(L)
               IF (L .EQ. K) GO TO 10
                  B(L) = B(K)
                  B(K) = T
   10          CONTINUE
               CALL DAXPY(LM,T,ABD(M+1,K),1,B(K+1),1)
   20       CONTINUE
   30    CONTINUE
C
C        NOW SOLVE  U*X = Y
C
         DO 40 KB = 1, N
            K = N + 1 - KB
            B(K) = B(K)/ABD(M,K)
            LM = MIN(K,M) - 1
            LA = M - LM
            LB = K - LM
            T = -B(K)
            CALL DAXPY(LM,T,ABD(LA,K),1,B(LB),1)
   40    CONTINUE
      GO TO 100
   50 CONTINUE
C
C        JOB = NONZERO, SOLVE  TRANS(A) * X = B
C        FIRST SOLVE  TRANS(U)*Y = B
C
         DO 60 K = 1, N
            LM = MIN(K,M) - 1
            LA = M - LM
            LB = K - LM
            T = DDOT(LM,ABD(LA,K),1,B(LB),1)
            B(K) = (B(K) - T)/ABD(M,K)
   60    CONTINUE
C
C        NOW SOLVE TRANS(L)*X = Y
C
         IF (ML .EQ. 0) GO TO 90
         IF (NM1 .LT. 1) GO TO 90
            DO 80 KB = 1, NM1
               K = N - KB
               LM = MIN(ML,N-K)
               B(K) = B(K) + DDOT(LM,ABD(M+1,K),1,B(K+1),1)
               L = IPVT(K)
               IF (L .EQ. K) GO TO 70
                  T = B(L)
                  B(L) = B(K)
                  B(K) = T
   70          CONTINUE
   80       CONTINUE
   90    CONTINUE
  100 CONTINUE
      RETURN
      END
*DECK XSETF
      SUBROUTINE XSETF (MFLAG)
C***BEGIN PROLOGUE  XSETF
C***PURPOSE  Reset the error print control flag.
C***CATEGORY  R3A
C***TYPE      ALL (XSETF-A)
C***KEYWORDS  ERROR CONTROL
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***DESCRIPTION
C
C   XSETF sets the error print control flag to MFLAG:
C      MFLAG=1 means print all messages (the default).
C      MFLAG=0 means no printing.
C
C***SEE ALSO  XERRWD, XERRWV
C***REFERENCES  (NONE)
C***ROUTINES CALLED  IXSAV
C***REVISION HISTORY  (YYMMDD)
C   921118  DATE WRITTEN
C   930329  Added SLATEC format prologue. (FNF)
C   930407  Corrected SEE ALSO section. (FNF)
C   930922  Made user-callable, and other cosmetic changes. (FNF)
C***END PROLOGUE  XSETF
C
C Subroutines called by XSETF.. None
C Function routine called by XSETF.. IXSAV
C-----------------------------------------------------------------------
C**End
      INTEGER MFLAG, JUNK, IXSAV
C
C***FIRST EXECUTABLE STATEMENT  XSETF
      IF (MFLAG .EQ. 0 .OR. MFLAG .EQ. 1) JUNK = IXSAV (2,MFLAG,.TRUE.)
      RETURN
C----------------------- End of Subroutine XSETF -----------------------
      END
*DECK XSETUN
      SUBROUTINE XSETUN (LUN)
C***BEGIN PROLOGUE  XSETUN
C***PURPOSE  Reset the logical unit number for error messages.
C***CATEGORY  R3B
C***TYPE      ALL (XSETUN-A)
C***KEYWORDS  ERROR CONTROL
C***DESCRIPTION
C
C   XSETUN sets the logical unit number for error messages to LUN.
C
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***SEE ALSO  XERRWD, XERRWV
C***REFERENCES  (NONE)
C***ROUTINES CALLED  IXSAV
C***REVISION HISTORY  (YYMMDD)
C   921118  DATE WRITTEN
C   930329  Added SLATEC format prologue. (FNF)
C   930407  Corrected SEE ALSO section. (FNF)
C   930922  Made user-callable, and other cosmetic changes. (FNF)
C***END PROLOGUE  XSETUN
C
C Subroutines called by XSETUN.. None
C Function routine called by XSETUN.. IXSAV
C-----------------------------------------------------------------------
C**End
      INTEGER LUN, JUNK, IXSAV
C
C***FIRST EXECUTABLE STATEMENT  XSETUN
      IF (LUN .GT. 0) JUNK = IXSAV (1,LUN,.TRUE.)
      RETURN
C----------------------- End of Subroutine XSETUN ----------------------
      END
*DECK IXSAV
      INTEGER FUNCTION IXSAV (IPAR, IVALUE, ISET)
C***BEGIN PROLOGUE  IXSAV
C***SUBSIDIARY
C***PURPOSE  Save and recall error message control parameters.
C***CATEGORY  R3C
C***TYPE      ALL (IXSAV-A)
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***DESCRIPTION
C
C  IXSAV saves and recalls one of two error message parameters:
C    LUNIT, the logical unit number to which messages are printed, and
C    MESFLG, the message print flag.
C  This is a modification of the SLATEC library routine J4SAVE.
C
C  Saved local variables..
C   LUNIT  = Logical unit number for messages.  The default is obtained
C            by a call to IUMACH (may be machine-dependent).
C   MESFLG = Print control flag..
C            1 means print all messages (the default).
C            0 means no printing.
C
C  On input..
C    IPAR   = Parameter indicator (1 for LUNIT, 2 for MESFLG).
C    IVALUE = The value to be set for the parameter, if ISET = .TRUE.
C    ISET   = Logical flag to indicate whether to read or write.
C             If ISET = .TRUE., the parameter will be given
C             the value IVALUE.  If ISET = .FALSE., the parameter
C             will be unchanged, and IVALUE is a dummy argument.
C
C  On return..
C    IXSAV = The (old) value of the parameter.
C
C***SEE ALSO  XERRWD, XERRWV
C***ROUTINES CALLED  IUMACH
C***REVISION HISTORY  (YYMMDD)
C   921118  DATE WRITTEN
C   930329  Modified prologue to SLATEC format. (FNF)
C   930915  Added IUMACH call to get default output unit.  (ACH)
C   930922  Minor cosmetic changes. (FNF)
C   010425  Type declaration for IUMACH added. (ACH)
C***END PROLOGUE  IXSAV
C
C Subroutines called by IXSAV.. None
C Function routine called by IXSAV.. IUMACH
C-----------------------------------------------------------------------
C**End
      LOGICAL ISET
      INTEGER IPAR, IVALUE
C-----------------------------------------------------------------------
      INTEGER IUMACH, LUNIT, MESFLG
C-----------------------------------------------------------------------
C The following Fortran-77 declaration is to cause the values of the
C listed (local) variables to be saved between calls to this routine.
C-----------------------------------------------------------------------
      SAVE LUNIT, MESFLG
      DATA LUNIT/-1/, MESFLG/1/
C
C***FIRST EXECUTABLE STATEMENT  IXSAV
      IF (IPAR .EQ. 1) THEN
        IF (LUNIT .EQ. -1) LUNIT = IUMACH()
        IXSAV = LUNIT
        IF (ISET) LUNIT = IVALUE
        ENDIF
C
      IF (IPAR .EQ. 2) THEN
        IXSAV = MESFLG
        IF (ISET) MESFLG = IVALUE
        ENDIF
C
      RETURN
C----------------------- End of Function IXSAV -------------------------
      END
*DECK IUMACH
      INTEGER FUNCTION IUMACH()
C***BEGIN PROLOGUE  IUMACH
C***PURPOSE  Provide standard output unit number.
C***CATEGORY  R1
C***TYPE      INTEGER (IUMACH-I)
C***KEYWORDS  MACHINE CONSTANTS
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***DESCRIPTION
C *Usage:
C        INTEGER  LOUT, IUMACH
C        LOUT = IUMACH()
C
C *Function Return Values:
C     LOUT : the standard logical unit for Fortran output.
C
C***REFERENCES  (NONE)
C***ROUTINES CALLED  (NONE)
C***REVISION HISTORY  (YYMMDD)
C   930915  DATE WRITTEN
C   930922  Made user-callable, and other cosmetic changes. (FNF)
C***END PROLOGUE  IUMACH
C
C*Internal Notes:
C  The built-in value of 6 is standard on a wide range of Fortran
C  systems.  This may be machine-dependent.
C**End
C***FIRST EXECUTABLE STATEMENT  IUMACH
      IUMACH = 6
C
      RETURN
C----------------------- End of Function IUMACH ------------------------
      END

## created by .createRxUiBlessedList() in ui-assign-parts.R edit there
.rxUiBlessed <- c("covLhsDf", "covariates", "curHi", "curLow", 
    "errParams0", "estNotAllowed", "eta", "etaLhsDf", "extraCmt", 
    "extraDvid", "hasErrors", "iniDf", "level", "levelLhsDf", 
    "lstExpr", "meta", "mixProbs", "model", "modelName", "mu2RefCovariateReplaceDataFrame", 
    "muRefCovariateDataFrame", "muRefCovariateEmpty", "muRefCurEval", 
    "muRefDataFrame", "muRefDropParameters", "muRefExtra", "muRefExtraEmpty", 
    "mv0", "mvL", "nonMuEtas", "oneTheta", "predDf", "redo", 
    "singleTheta", "sticky", "thetaLhsDf", "uiUseData", "uiUseMv")

## nocov start
#' Generate an example (template) of a dosing regimen shiny app
#'
#' Create a complete shiny application for exploring dosing regimens
#' given a (hardcoded) PK/PD model.
#'
#' @param appDir a string with a directory where to store the shiny
#'     app, by default is `"shinyExample"`. The directory
#'     `appDir` will be created if it does not exist.
#'
#'
#' @param verbose logical specifying whether to write messages as the
#'     shiny app is generated. Defaults to `TRUE`.
#'
#' @param statevars List of statevars passed to to the [write.template.ui()] function.  This usually isn't called directly.
#'
#' A PK/PD model is defined using [rxode2()], and
#' a set of parameters and initial values are defined.  Then
#' the appropriate R scripts for the shiny's user interface `ui.R`
#' and the server logic `server.R` are created in the
#' directory `appDir`.
#'
#' The function evaluates the following PK/PD model by default:
#' \preformatted{
#'     C2 = centr/V2;
#'     C3 = peri/V3;
#'     d/dt(depot) =-KA*depot;
#'     d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
#'     d/dt(peri)  =                    Q*C2 - Q*C3;
#'     d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
#' }
#'
#' This can be changed by the `ODE.config` parameter.
#'
#' To launch the shiny app, simply issue the `runApp(appDir)`
#' R command.
#'
#' @param ODE.config model name compiled and list of parameters sent to [rxSolve()].
#'
#' @return None, these functions are used for their side effects.
#'
#' @note These functions create a simple, but working example of a
#'     dosing regimen simulation web application. Users may want to
#'     modify the code to experiment creating shiny applications for
#'     their specific `rxode2` models.
#' @seealso [rxode2()],[eventTable()], and the package \pkg{shiny} (<https://shiny.posit.co>).
#'
#' @examples
#' \donttest{
#' # create in a temp dir so nothing is left in the working directory
#' .appDir <- tempfile("myapp")
#' on.exit(unlink(.appDir, recursive = TRUE, force = TRUE))
#' # create the shiny app example (template)
#' genShinyApp.template(appDir = .appDir)
#' # run the shiny app
#' if (requireNamespace("shiny", quietly=TRUE)) {
#'   library(shiny)
#'   # runApp(.appDir) # Won't launch in environments without browsers
#' }
#' }
#' @keywords simulation nonlinear
#' @concept PK/PD
#' @concept pharmacometrics
#' @export genShinyApp.template
genShinyApp.template <-
  function(appDir = "shinyExample", verbose = TRUE,
           ODE.config = list(ode = "model", params = c(KA = 0.294), inits = c(eff = 1), method = "lsoda", atol = 1e-8, rtol = 1e-6)) {
   if (!requireNamespace("shiny", quietly=TRUE)) {
     warning("install.packages('shiny') to use genShinyApp.template()")
     return(NULL)
   }
   if (missing(ODE.config)) {
      ODE.config <- list(
        ode = "
   C2 = centr/V2;
   C3 = peri/V3;
   d/dt(depot) =-KA*depot;
   d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
   d/dt(peri)  =                    Q*C2 - Q*C3;
   d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
   ",
        params =
          c(
            KA = .294, CL = 18.6, V2 = 40.2, Q = 10.5, V3 = 297.0,
            Kin = 1.0, Kout = 1.0, EC50 = 200.0
          ),
        inits = c(depot = 0, centr = 0, pari = 0, eff = 1),
        method = "lsoda",
        atol = 1e-08,
        rtol = 1e-06
      )
    }
    if (!file.exists(appDir)) {
      dir.create(appDir, recursive = TRUE, showWarnings = FALSE)
    }
    # if(.Platform$OS.type=="windows") appDir <- gsub("\\\\", "/", utils::shortPathName(.normalizePath(appDir)))  # safe pathname

    pkpd <- ODE.config$ode

    if (verbose) {
      cat("\nGenerating an example (template) for a dosing regimen shiny app\n\n")
      cat("Using the following PK/PD model:")
      cat(pkpd, "\n")

      cat("Translating the PK/PD ODE model into C, compiling, etc.\n\n")
    }

    mod1 <- rxode2(pkpd, modName = "mod1", wd = appDir)

    stateVars <- mod1$cmpMgr$get.modelVars()$state
    stateVarStr <- paste(sprintf("\"%s\"", stateVars), collapse = ", ")

    params <- ODE.config$params
    inits <- ODE.config$inits

    if (verbose) {
      cat("\nParameters and their values:\n")
      print(params)
      cat("\nInitial values in each compartment:\n")
      print(inits)
      cat("\n")
    }

    # save the model, parameters, init values, etc. in the
    # file rx_shiny_data.rda to be loaded by the server.R

    fn <- file.path(appDir, "rx_shiny_data.rda")
    method <- ODE.config$method
    atol <- ODE.config$atol
    rtol <- ODE.config$rtol
    save(mod1, params, inits, method, atol, rtol, file = fn)

    # write the shiny server.R and ui.R files
    write.template.server(appDir)
    write.template.ui(appDir, statevars = c(stateVarStr, stateVars[1]))

    if (verbose) {
      cat("Shiny files (ui.R, server.R) plus R data saved.\n")
    }

    cat("\nTo launch the Shiny app, type the following two R commands:\n\n")
    cat("\tlibrary(shiny)\n")
    cat(sprintf('\trunApp("%s")\n\n', appDir))
  }
#' @rdname genShinyApp.template
#' @export write.template.server
write.template.server <-
  function(appDir) {
    # create a shiny server interface server.R script in the appDir

    server <- file.path(appDir, "server.R")

    server.code <-
      sprintf(appDir, fmt = '
      #
      # Dosing regimen template generated by rxode2::genShinyApp.template()
      #

      debug = TRUE
      #wd = sprintf("%%s/../", getwd())
      #setwd(wd)

      # Server inputs: Dose, dosing regimen, dosing frequency,
      # dosing cycle definition, number of dosing cycles

      library(shiny)
      library(rxode2)

      # read objects from "rx_shiny_data.rda" in the  AppDir folder,
      # objects include, mod1, params, inits, method, atol, rtol.]

      load("./rx_shiny_data.rda")
      if (!rxDynLoad(mod1)) mod1 <- rxode2(mod1, modName="mod1")
      # Define server logic
      shinyServer(function(input, output) {

        get.cp <- reactive({
          ds <- input$Dose
          reg <- switch(input$regimen, "QD"=1, "BID"=2)
          cyc <- switch(input$cycle,
              "continous"=c(7,0),
              "1wkon 1wkoff"=c(7,7),
              "2wkon 1wkoff"=c(14,7),
              "3wkon 1wkoff"=c(21,7)
          )
          cyc <- rep(1:0, cyc)
          ncyc <- input$ncyc
          lcyc <- length(cyc)

          ev <- eventTable()
          for (i in 1:ncyc) ev$add.dosing(
              dose=ds,
              nbr.doses=sum(cyc)*reg,
              dosing.interval=24/reg,
              start.time=(i-1)*lcyc*24
          )
          ev$get.EventTable()
          ev$add.sampling(0:(ncyc*lcyc*24))

          mod1$solve(params, ev, inits, method=method, atol=atol, rtol=rtol)
        })


        output$summary <- renderPrint({
          x <- get.cp()
          print(x[1:4,])
          if (debug) print(getwd())
        })

        output$CpPlot <- renderPlot({
          x <- get.cp()
          cmp <- input$compartment
          plot(x[,c("time", cmp)], xlab = "Time", ylab = "Drug amount",
               main = cmp, type = "l")
        })
      })
   ', appDir)
    writeLines(server.code, con = server)
  }
#' @rdname genShinyApp.template
#' @export write.template.ui
write.template.ui <-
  function(appDir, statevars) {
    # create a shiny user interface ui.R script in the appDir

    ui <- file.path(appDir, "ui.R")

    ui.code <-
      sprintf(fmt = '
      #
      # Dosing regimen template automatically generated by
      # rxode2::genShinyApp.template()
      #

      library(rxode2)
      library(shiny)

      # Define UI for dataset viewer application
      shinyUI(pageWithSidebar(

        # Application title.
        headerPanel("Regimen simulator"),

        # Sidebar with controls to select a dataset and specify the number
        # of observations to view. The helpText function is also used to
        # include clarifying text. Most notably, the inclusion of a
        # submitButton defers the rendering of output until the user
        # explicitly clicks the button (rather than doing it immediately
        # when inputs change). This is useful if the computations required
        # to render output are inordinately time-consuming.
        sidebarPanel(
          sliderInput("Dose", "Dose:",
                        min=100, max=1000, value=400, step=50),

          selectInput("regimen", "Regimen:",
                      choices = c("QD", "BID"),
                      selected="QD"),

          selectInput("cycle", "Cycle:",
                      choices = c(
                              "continous",
                              "1wkon 1wkoff",
                              "2wkon 1wkoff",
                              "3wkon 1wkoff"),
                      selected="2wkon 1wkoff"),

          sliderInput("ncyc", "# cycles:",
                        min=1, max=5, value=3, step=1),

          radioButtons("compartment", "Select PK compartment for plotting:",
                       choices=c(%s),
                       selected = "%s")
        ),

        # Show a summary of the dataset and an HTML table with the requested
        # number of observations. Note the use of the h4 function to provide
        # an additional header above each output section.
        mainPanel(
          h4("State variables"),
          verbatimTextOutput("summary"),
          h4("Concentration time course"),
          plotOutput("CpPlot")
        )
      ))
   ', statevars[1], statevars[2])

    writeLines(ui.code, con = ui)
  }
## nocov end

.udfEnv <- new.env(parent=emptyenv())
.udfEnv$fun <- list()
.udfEnv$udf <- integer(0)
.udfEnv$envir <- NULL
.udfEnv$envList <- list()
.udfEnv$searchList <- list()
.udfEnv$hashSearchList <- new.env(parent = emptyenv())
.udfEnv$rxSEeqUsr <- NULL
.udfEnv$rxCcode <- NULL
.udfEnv$symengineFs <- new.env(parent = emptyenv())
.udfEnv$extraCnow <- ""
.udfEnv$bestFun <- NULL
.udfEnv$bestFunEnv <- NULL
.udfEnv$bestFunHasDots <- FALSE
.udfEnv$bestNargs <- NA_integer_
.udfEnv$bestEqArgs <- FALSE
#' Get the udf strings for creating model md5
#'
#' @return string vector
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.udfMd5Info <- function() {
  .tmp <- ls(.udfEnv$symengineFs, all.names=TRUE)
  .env <- new.env(parent=emptyenv())
  .env$found <- FALSE
  .ret <- vapply(.tmp, function(x) {
    .cur <- .udfEnv$fun[[x]]
    if (!is.null(.cur)) {
      .env$found <- TRUE
    }
    x
  }, character(1), USE.NAMES = FALSE)
  if (.env$found) {
    .ret <- c(.ret, data.table::address(.udfEnv$envir),
              # don't cache md5 changes every run:
              as.character(Sys.time()))
  }
  .ret
}
#' Generate extraC information for rxode2 models
#'
#' @param extraC Additional extraC from rxode2 compile optioioins
#' @return Nothing, called for side effects
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.extraC <- function(extraC = NULL) {
  if (!is.null(extraC)) {
    if (file.exists(extraC)) {
      .ret <- sprintf("#include \"%s\"\n", extraC)
    } else {
      .ret <- paste(extraC, collapse = "\n")
    }
  } else {
    .ret <- ""
  }
  if (length(.udfEnv$rxCcode) > 0L) {
    .ret <- sprintf("%s\n%s\n", .ret, paste(.udfEnv$rxCcode, collapse = "\n"))
  }
  .udfEnv$extraCnow <- .ret
  return(invisible())
}
#' Get the extraCnow for compiling
#'
#'
#' @return string of extraC information
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.extraCnow <- function() {
  .udfEnv$extraCnow
}

#' Add user function to rxode2
#'
#' This adds a user function to rxode2 that can be called.  If needed,
#' these functions can be differentiated by numerical differences or
#' by adding the derivatives to rxode2's internal derivative table
#' with rxode2's `rxD` function
#'
#' @param name This gives the name of the user function
#' @param args This gives the arguments of the user function
#' @param cCode This is the C-code for the new function
#' @return nothing
#' @author Matthew L. Fidler
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
rxFunParse <- function(name, args, cCode) {
  if (!is.character(name) || length(name) != 1L) {
    stop("name argument must be a length-one character vector", call. = FALSE)
  }
  if (missing(cCode)) stop("a new function requires a C function so it can be used in rxode2", call. = FALSE)
  if (any(name == names(.udfEnv$rxSEeqUsr))) {
    stop("already defined user function '", name, "', remove it fist ('rxRmFun')",
         call. = FALSE
         )
  }
  suppressWarnings(rxRmFunParse(name))
  .udfEnv$rxSEeqUsr <- c(.udfEnv$rxSEeqUsr, setNames(length(args), name))
  .udfEnv$rxCcode <- c(.udfEnv$rxCcode, setNames(cCode, name))
  assign(name, symengine::Function(name), envir = .udfEnv$symengineFs)
  return(invisible())
}
#' Return the equivalents symengine user functions from C
#'
#' @return equivalent symengine user functions
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.rxSEeqUsr <- function() {
  .udfEnv$rxSEeqUsr
}

#' Return symengineFs from user functions
#'
#' @return symengineFs from user functions
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.symengineFs <- function() {
  .udfEnv$symengineFs
}

#' Return the C code of an internal function
#'
#' @param fun is the string of a function that you wish to get the C
#'   code for
#' @return C code if found (as a string) or NULL if not found
#' @export
#' @author Matthew Fider
#' @keywords internal
.rxC <- function(fun) {
  .w <- which(names(.udfEnv$rxCcode) == fun)
  if (length(.w) == 1) return(setNames(.udfEnv$rxCcode[fun], NULL))
  NULL
}

#' @rdname rxFunParse
#' @export
rxRmFunParse <- function(name) {
  if (!is.character(name) || length(name) != 1L) {
    stop("name argument must be a length-one character vector",
         call. = FALSE)
  }
  if (!any(name == names(.udfEnv$rxSEeqUsr))) {
    warning("no user function '", name, "' to remove", call. = FALSE)
  }
  .w <- which(name == names(.udfEnv$rxSEeqUsr))
  if (length(.w) == 1L) {
    .udfEnv$rxSEeqUsr <- .udfEnv$rxSEeqUsr[-.w]
  }
  .w <- which(name == names(.udfEnv$rxCcode))
  if (length(.w) == 1L) {
    .udfEnv$rxCcode <- .udfEnv$rxCcode[-.w]
  }
  .rxD <- rxode2parseD()
  if (exists(name, envir = .rxD)) {
    if (!grepl("^rx_", name)) {
      .d <- get(name, envir=.rxD)
      lapply(names(formals(.d[[1]])), function(v) {
        suppressWarnings(rxRmFunParse(paste0("rx_", name, "_d_", v)))
      })
    }
    rm(list = name, envir = .rxD)
  }
  if (exists(name, envir = .udfEnv$symengineFs)) {
    rm(list = name, envir = .udfEnv$symengineFs)
  }
  return(invisible())
}



.udfAddToSearch <- function(envir) {
  if (is.list(envir)) {
    lapply(seq_along(envir),
           function(i) {
             .udfAddToSearch(envir[[i]])
           })
    return(invisible())
  }
  if (length(.udfEnv$searchList) == 0L) {
    .udfEnv$searchList <- list(envir)
    .udfEnv$hashSearchList[[data.table::address(envir)]] <- TRUE
  }
  if (exists(data.table::address(envir), envir = .udfEnv$hashSearchList, inherits = FALSE)) {
    return(invisible())
  }
  .udfEnv$searchList <- c(.udfEnv$searchList, list(envir))
  .udfEnv$hashSearchList[[data.table::address(envir)]] <- TRUE
  invisible()
}

#' Setup the UDF environment (for querying user defined funtions)
#'
#' @param env environment where user defined functions are queried. If NULL return current environment
#' @return environment
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.udfEnvSet <- function(env) {
  if (is.null(.udfEnv$envir)) {
    if (is.list(env)) {
      .udfEnv$envir <- env[[1]]
    } else {
      .udfEnv$envir <- env
    }
  }
  .udfAddToSearch(env)
  return(invisible(.udfEnv$envir))
}
#' Lock/Unlock environment for getting R user functions
#'
#' @param lock logical to see if environment to look for user defined
#'   functions is locked.  If it is locked then environments are not
#'   assigned.  When NULL returns lock status
#'
#' @return lock status
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.udfEnvReset <- function(lock=TRUE) {
  .udfEnv$fun <- list()
  .udfEnv$searchList <- list()
  .udfEnv$hashSearchList <- new.env(parent = emptyenv())
}
#' See if the UI function exists in given environment.
#'
#' If other functions have been declared, make sure they exist too.
#'
#' @param fun Function to check
#' @param nargs Number of args to check
#' @param envir Environment to check
#' @param doList A boolean to see if the functions in .udfEnv$fun
#'   should be checked too.  By default TRUE, but this is called
#'   recursively for each function (and set to FALSE)
#' @return logical declaring if the udf function exists in this environment
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.udfExists <- function(fun, nargs, envir, doList=TRUE) {
  if (is.null(envir)) return(FALSE)
  .e <- exists(fun, mode="function", envir=envir)
  if (!.e) return(FALSE)
  # ok now see if it makes sense
  .fun <- get(fun, mode="function", envir=envir)
  .f <- formals(.fun)
  .bestHasDots <- any(names(.f) == "...")
  .nargs <- length(.f)
  .bestEqArgs <- .nargs == nargs
  if (.bestEqArgs) { # We want the function to match the declared number of arguments
    if (!.bestHasDots) { # We don't want ... arguments
      if (doList) {
        # In the case of multiple user functions, make sure the other
        # user functions also exist in this environment
        if (!all(vapply(seq_along(.udfEnv$fun), function(i) {
          .info <- .udfEnv$fun[[i]]
          return(.udfExists(.info[[1]], .info[[2]], envir=envir, doList=FALSE))
        }, logical(1), USE.NAMES = FALSE))) {
          if (is.null(.udfEnv$bestFun)) {
            .udfEnv$bestFun <- .fun
          }
          return(FALSE)
        }
        # Success, save function and environment
        .udfEnv$bestFun <- .fun
        .udfEnv$bestFunEnv <- envir
        .udfEnv$bestFunHasDots <- FALSE
        .udfEnv$bestEqArgs <- TRUE
        .udfEnv$bestNargs <- nargs
      }
      return(TRUE)
    }
  }
  if (doList && is.null(.udfEnv$bestFun)) {
    .udfEnv$bestFun <- .fun
    .udfEnv$bestFunEnv <- envir
    .udfEnv$bestFunHasDots <- .bestHasDots
    .udfEnv$bestEqArgs <- .bestEqArgs
    .udfEnv$bestNargs <- .nargs
  }
  FALSE
}
#' While parsing or setting up the solving, get information about the
#' user defined function
#'
#' @param fun function (character) to get information about
#' @param nargs Preferred number of arguments
#' @return A list with two elements
#'   - nargs = `NA` if the user function isn't supported, or the number of arguments suported
#'   - string = Error message when `NA` or function string
#' @noRd
#' @author Matthew L. Fidler
.getUdfInfo <- function(fun, nargs) {
  .nargs <- .rxUdfUiNarg(fun)
  if (is.integer(.nargs) && !is.na(.nargs)) {
    if (.nargs > 0L && nargs != .nargs) {
      return(list(nargs=NA_integer_,
                  sprintf("rxode2 ui user defined R function has %d arguments, but supplied %d",
                          .nargs, nargs)))
    } else if (.nargs <= 0L) {
      return(list(nargs=NA_integer_,
                  "rxode2 ui user defined R needs to be setup with a positive number of arguments"))
    } else {
      return(list(nargs=-42L,
                  ".rxUiUdfNone"))
    }
  }
  if (is.null(.udfEnv$envir)) {
    return(list(nargs=NA_integer_,
                "rxode2 cannot determine which environment the user defined functions are located"))
  }
  .udfEnv$bestFun <- NULL
  .udfEnv$bestFunHasDots <- FALSE
  .udfEnv$bestEqArgs <- TRUE
  .found <- FALSE
  if (!.udfExists(fun, nargs, .udfEnv$envir)) {
    # search prior environments with UDFs, assign the first one in the environments that match
    if (length(.udfEnv$searchList) > 0L) {
      if (any(vapply(seq_along(.udfEnv$searchList), function(i) {
        .udfExists(fun, nargs, .udfEnv$searchList[[i]])
      },  logical(1), USE.NAMES = FALSE))) {
        .found <- TRUE
      }
    }
  } else {
    .found <- TRUE
  }
  if (.udfEnv$bestFunHasDots) {
    return(list(nargs=NA_integer_,
                "rxode2 user defined R cannot have '...' arguments"))
  }
  if (!.udfEnv$bestEqArgs) {
    return(list(nargs=NA_integer_,
                sprintf("rxode2 user defined R function has %d arguments, but supplied %d",
                        .udfEnv$bestNargs, nargs)))
  }
  if (!.found) {
    .msg <- sprintf("function '%s' is not supported; user function not found",
                    fun)
    return(list(nargs=NA_integer_, .msg))
  }

  .fun <- .udfEnv$bestFun
  .udfEnv$envir <- .udfEnv$bestFunEnv
  .udfEnv$fun[[fun]] <- list(fun, nargs)
  .w <- which(names(.udfEnv$udf) == fun)
  if (length(.w) == 0L) {
    .udfEnv$udf <- c(.udfEnv$udf, setNames(nargs, fun))
  }
  return(list(nargs=nargs,
              fun))
}

#' This function is run before starting a rxode2 solve to make sure
#' the R-based user functions are setup correctly.
#'
#' This function also resets the udf-based run-time errors
#'
#' @param iv Named Integer Vector with the names representing the
#'   functions and the integers representing the number of arguments
#'   that were present when the model was compiled
#' @return nothing, called for side effect
#' @noRd
#' @author Matthew L. Fidler
.setupUdf <- function(iv) {
  if (!is.environment(.udfEnv$envir)) return(FALSE)
  .w <- which(is.na(iv))
  iv <- iv[-.w]
  .n <- names(iv)
  .env <- new.env(parent=emptyenv())
  .env$needRecompile <- FALSE
  lapply(.n,
         function(n) {
           .oldArg <- iv[n]
           .new <- .getUdfInfo(n, .oldArg)
           if (any(names(.udfEnv$rxSEeqUsr) == n)) {
             .c <- .udfEnv$rxSEeqUsr[n]
             if (.c == .new[[1]]) {
               message("compiled with R user function '", n, "'; now there is a clashing C user function")
               .env$needRecompile <- TRUE
               message("triggered a recompile to use the C user function (they are always preferred)")
             } else {
               stop("there is both C and R user functions '", n, "' with a different number of arguments\n  since rxode2 prefers C, you will need to rename your R user function to use it")

             }
           }
           if (is.na(.new[[1]])) {
             stop(.new[[2]], call.=FALSE)
           } else if (.new[[1]] != .oldArg) {
             stop("'", n,
                  "' had ", .oldArg, " arguments when model was compiled, now it has ",
                  .new[[1]], " arguments",
                  call.=FALSE)
           }
           NULL
         })
  .env$needRecompile
}
#' Reset the tracking of user defined functions
#'
#' This is called during parsing reset
#'
#' @return Nothing, called for side effects
#' @noRd
#' @author Matthew L. Fidler
.udfReset <- function() {
  .udfEnv$udf <- integer(0)
}

#' This gets the user defined functions information for incorporation
#' in the model variables
#'
#' @return A integer vector; The values are the number of arguments;
#'   the names are the function names
#' @author Matthew L. Fidler
#' @noRd
.udfInfo <- function() {
  if (length(.udfEnv$udf) == 0) return(integer(0))
  if (!is.environment(.udfEnv$envir)) return(integer(0))
  .addr <- data.table::address(.udfEnv$envir)
  .udfEnv$envList[[.addr]] <- .udfEnv$envir
  c(.udfEnv$udf, setNames(NA_integer_, .addr))
}

#' Use the udf model variable information to get the environment where
#' the functions exists
#'
#' @param udf modelVars$udf, integer vector with NA_integer_ for the
#'   address of the environment where the functions exist
#' @return nothing called for side effects
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.udfEnvSetUdf <- function(udf) {
  if (length(udf) == 0L) return(invisible())
  .w <- which(is.na(udf))
  .addr <- names(udf)[.w]
  .env <- .udfEnv$envList[[.addr]]
  if (is.environment(.env)) {
    .udfAddToSearch(.env)
    ## .udfEnv$envir <- .env
  } else {
    stop("environment were user functions were defined is no longer present")
  }
  invisible()
}
#' Get the function name with the current arguments as a string
#'
#' @param fun function name
#' @param args  arguments
#' @return string of the form 'fun(arg1, arg2)':
#' @export
#' @author Matthew L. Fidler
#' @keywords internal
.udfCallFunArg <- function(fun, args) {
  paste0("'", fun, "(",
         paste(vapply(seq_along(args),
                function(i) {
                  as.character(args[[i]])
                }, character(1), USE.NAMES=FALSE), collapse=", "),
         ")': ")
}
#' This is the function that is always called for every user function in rxode2
#'
#' @param fun A character vector representing the function
#' @param args A list of double numbers that will be used as the
#'   function arguments
#' @return A double numeric value, including `NA_real` when the
#'   function isn't working as expected
#' @noRd
#' @author Matthew L. Fidler
.udfCall <- function(fun, args) {
  .ret <- try(do.call(fun, args, envir=.udfEnv$envir), silent=TRUE)
  if (inherits(.ret, "try-error")) {
    .msg <- try(attr(.ret, "condition")$message, silent=TRUE)
    if (inherits(.msg, "try-error")) .msg <- "Unknown Error"
    # This can error since it isn't threaded
    stop(paste0(.udfCallFunArg(fun, args), .msg), call.=FALSE)
  }
  if (checkmate::testNumeric(.ret, len=1)) {
    return(as.double(.ret))
  }
  stop(paste0(.udfCallFunArg(fun, args), "needs to return a length 1 numeric"),
       call.=FALSE)
  .ret
}

#' Push a future dose or observation event from within an rxode2 model
#'
#' @description
#' `evid_()` is a model-only function that schedules a future dosing or
#' observation event during ODE solving.  It is evaluated at each output
#' time point and inserts the requested event into the individual's event
#' timeline for future processing.
#'
#' This function has no meaning outside an rxode2 model block and will
#' throw an error if called directly from R.
#'
#' @param time Numeric. The time at which the event should occur.  Must
#'   be greater than the current model time `t`; events in the past are
#'   silently counted and a warning is issued after solving.
#'
#' @param evid Integer event ID.  Follows NONMEM/rxode2 conventions:
#'   \describe{
#'     \item{`0`}{Observation (adds an output row, no dose).}
#'     \item{`1`}{Dose (bolus or infusion depending on `rate`).}
#'     \item{`2`}{Other type observation (passed through).}
#'     \item{`3`}{Reset all compartments.}
#'     \item{`4`}{Reset then dose.}
#'     \item{`5`}{Replace compartment amount.}
#'     \item{`6`}{Multiply compartment amount.}
#'     \item{`7`}{Phantom/transit dose.}
#'     \item{`>= 100`}{Classic rxode2 internal evid; passed through verbatim.}
#'   }
#'
#' @param amt Numeric dose amount (for dose events) or `0` for
#'   observations.  When `rate > 0` this is interpreted as the total
#'   infusion amount and the infusion duration is `amt / rate`.
#'
#' @param cmt Integer compartment number (1-based) to which the dose is
#'   applied.  Default is 1
#' @param rate Numeric infusion rate.
#'   \describe{
#'     \item{`0`}{Bolus dose (default).}
#'     \item{`> 0`}{Fixed infusion rate; duration = `amt / rate`.}
#'     \item{`-1`}{Rate defined by the model (`rate_<cmt>` variable).}
#'     \item{`-2`}{Duration defined by the model (`dur_<cmt>` variable).}
#'   }
#' @param ii Numeric inter-dose interval.  Used together with `addl` to
#'   schedule repeat doses at `time`, `time + ii`, `time + 2*ii`, ...,
#'   `time + addl*ii`.  Also required when `ss > 0`. Default 0
#' @param addl Integer number of *additional* doses beyond the first.
#'   The total number of doses pushed is `addl + 1`, spaced `ii` apart.
#'   Each dose is pushed as a standalone event (not as a periodic
#'   schedule in the event table).
#' @param ss Integer steady-state flag applied to the *first* dose only
#'   (`addl` repetitions always use `ss = 0`).
#'   \describe{
#'     \item{`0`}{No steady-state (default).}
#'     \item{`1`}{Steady-state additive: add SS solution to current state.}
#'     \item{`2`}{Steady-state replace: replace current state with SS solution.}
#'   }
#'
#' @return This function is only meaningful inside an rxode2 model; it
#'   returns `NULL` invisibly if called from R directly (after signaling
#'   an error).
#'
#' @details
#' ## Behavior inside a model
#'
#' `evid_()` is evaluated at every output time point (when the solver is
#' exactly at a scheduled event time).  The pushed event is inserted into
#' the individual's event timeline and the solver visits it at the
#' specified future time.
#'
#' The number of events that may be pushed per individual is limited by
#' the `maxExtra` argument of [rxSolve()].  When `maxExtra = 0`
#' (the default) there is no limit.  Exceeding the limit causes an
#' error.
#'
#' Past-time pushes (where `time <= t`) are silently ignored and counted;
#' a warning is issued after solving.
#'
#' ## Relationship to NONMEM event columns
#'
#' The argument order and names mirror the standard NONMEM dataset
#' columns: `TIME`, `EVID`, `AMT`, `CMT`, `RATE`, `II`, `ADDL`, `SS`.
#'
#' @seealso [rxSolve()] for the `maxExtra` control argument.  Other
#'   more convenient ways to interact with adaptive observations and
#'   events include [bolus()] [infuse()], [infuseDur()],
#'   [reset()].
#'
#' @examples
#' \donttest{
#' # Push a single bolus of 50 mg to compartment 1 at t + 12
#' m <- rxode2({
#'   d/dt(depot)   <- -ka * depot
#'   d/dt(central) <- ka * depot - cl / vd * central
#'   cp <- central / vd
#'   if (t < 24) {
#'     evid_(t + 12, 1, 50, 1, 0, 0, 0, 0)
#'   }
#' })
#'
#' # Push three boluses (addl = 2) at t+6, t+18, t+30 (ii = 12)
#' m2 <- rxode2({
#'   d/dt(depot)   <- -ka * depot
#'   d/dt(central) <- ka * depot - cl / vd * central
#'   cp <- central / vd
#'   if (t < 1) {
#'     evid_(t + 6, 1, 50, 1, 0, 12, 2, 0)
#'   }
#' })
#' }
#'
#' @export
evid_ <- function(time, evid, amt, cmt=1, rate=0, ii=0, addl=0, ss=0.0) {
  stop("'evid_()' can only be used inside an rxode2 model block", call. = FALSE)
}

## Normalize a compartment argument from a matched adaptive-dosing call.
## Accepts the unevaluated `match.call()` element for `cmt` and returns the
## compartment as a bare token for the rxode2 grammar:
##  - missing/NULL          -> default compartment "1"
##  - a string ("depot")    -> the bare name `depot` (quotes stripped)
##  - a name/number/expr    -> its deparsed text
## The string case matters because `deparse1("depot")` keeps the quotes,
## which the grammar then mis-parses (e.g. bolus(50, "depot, 0, 0, 0)).
.rxUdfUiCmt <- function(cmt) {
  if (is.null(cmt)) return("1")
  if (is.character(cmt)) return(cmt)
  deparse1(cmt)
}

#' @export
#' @keywords internal
#' @rdname rxUdfUi
rxUdfUi.evid_ <- function(fun) {
  .dummy <- function(time, evid, amt, cmt, rate, ii, addl, ss) {}
  .mc <- match.call(.dummy, fun)
  .time    <- deparse1(.mc$time)
  .evid <- deparse1(.mc$evid)
  .amt   <- deparse1(.mc$amt)

  .cmt  <- .rxUdfUiCmt(.mc$cmt)

  .rate   <- deparse1(.mc$rate)
  if (.rate == "NULL") {
    .rate <- "0"
  }

  .ii   <- deparse1(.mc$ii)
  if (.ii == "NULL") {
    .ii <- "0"
  }

  .addl   <- deparse1(.mc$addl)
  if (.addl == "NULL") {
    .addl <- "0"
  }

  .ss   <- deparse1(.mc$ss)
  if (.ss == "NULL") {
    .ss <- "0"
  }

  list(replace = paste0("evid_(", .time, ",", .evid, ",", .amt, ",", .cmt,
                        ",", .rate, ",", .ii, ",", .addl, ",", .ss, ")"))
}
#' Administer a bolus dose inside a rxode2 model
#'
#' @inheritParams evid_
#'
#' @return This function is only meaningful inside an rxode2 model; it
#'   returns `NULL` invisibly if called from R directly (after signaling
#'   an error).
#'
#' @details
#' ## Behavior inside a model
#'
#' `bolus()` is evaluated at every output time point (when the solver is
#' exactly at a scheduled event time).  The pushed event is inserted into
#' the individual's event timeline and the solver visits it now or at the
#' specified future time.
#'
#' The number of events that may be pushed per individual is limited by
#' the `maxExtra` argument of [rxSolve()].  When `maxExtra = 0`
#' (the default) there is no limit.  Exceeding the limit causes an
#' error.
#'
#' Past-time pushes (where `time < t`) are silently ignored and counted;
#' a warning is issued after solving.
#'
#' @export
#' @author Matthew L. Fidler
bolus <- function(amt, cmt = 1, ii = 0, addl = 0, ss = 0) {
  stop("'bolus()' can only be used inside an rxode2 model block", call. = FALSE)
}

#' @export
#' @keywords internal
#' @rdname rxUdfUi
rxUdfUi.bolus <- function(fun) {
  .dummy <- function(amt, cmt, ii, addl, ss) {}
  .mc <- match.call(.dummy, fun)
  .amt   <- deparse1(.mc$amt)

  .cmt  <- .rxUdfUiCmt(.mc$cmt)

  .ii   <- deparse1(.mc$ii)
  if (.ii == "NULL") {
    .ii <- "0"
  }

  .addl   <- deparse1(.mc$addl)
  if (.addl == "NULL") {
    .addl <- "0"
  }

  .ss   <- deparse1(.mc$ss)
  if (.ss == "NULL") {
    .ss <- "0"
  }

  list(replace = paste0("bolus(", .amt, ",", .cmt, ",", .ii, ",", .addl, ",", .ss, ")"))
}
#' Administer a infusion (with rate being fixed) inside a rxode2 model
#'
#' @inheritParams evid_
#'
#' @return This function is only meaningful inside an rxode2 model; it
#'   returns `NULL` invisibly if called from R directly (after signaling
#'   an error).
#'
#' @details
#' ## Behavior inside a model
#'
#' `infuse()` is evaluated at every output time point (when the solver is
#' exactly at a scheduled event time).  The pushed event is inserted into
#' the individual's event timeline and the solver visits it now or at the
#' specified future time.
#'
#' The number of events that may be pushed per individual is limited by
#' the `maxExtra` argument of [rxSolve()].  When `maxExtra = 0`
#' (the default) there is no limit.  Exceeding the limit causes an
#' error.
#'
#' Past-time pushes (where `time < t`) are silently ignored and counted;
#' a warning is issued after solving.
#'
#' @export
#' @author Matthew L. Fidler
infuse <- function(amt, rate, cmt = 1, ii = 0, addl = 0, ss = 0) {
  stop("'infuse()' can only be used inside an rxode2 model block", call. = FALSE)
}

#' @export
#' @keywords internal
#' @rdname rxUdfUi
rxUdfUi.infuse <- function(fun) {
  .dummy <- function(amt, rate, cmt, ii, addl, ss) {}
  .mc <- match.call(.dummy, fun)
  .amt <- deparse1(.mc$amt)
  .rate <- deparse1(.mc$rate)
  .cmt <- .rxUdfUiCmt(.mc$cmt)
  .ii <- deparse1(.mc$ii)
  if (.ii == "NULL") .ii <- "0"
  .addl <- deparse1(.mc$addl)
  if (.addl == "NULL") .addl <- "0"
  .ss <- deparse1(.mc$ss)
  if (.ss == "NULL") .ss <- "0"
  list(replace = paste0("infuse(", .amt, ",", .rate, ",", .cmt, ",", .ii, ",", .addl, ",", .ss, ")"))
}

#' Administer a infusion (with duration being fixed) inside a rxode2 model
#'
#' @inheritParams evid_
#'
#' @param dur Numeric infusion duration.
#'
#' @return This function is only meaningful inside an rxode2 model; it
#'   returns `NULL` invisibly if called from R directly (after signaling
#'   an error).
#'
#' @details
#' ## Behavior inside a model
#'
#' `infuseDur()` is evaluated at every output time point (when the solver is
#' exactly at a scheduled event time).  The pushed event is inserted into
#' the individual's event timeline and the solver visits it now or at the
#' specified future time.
#'
#' The number of events that may be pushed per individual is limited by
#' the `maxExtra` argument of [rxSolve()].  When `maxExtra = 0`
#' (the default) there is no limit.  Exceeding the limit causes an
#' error.
#'
#' Past-time pushes (where `time < t`) are silently ignored and counted;
#' a warning is issued after solving.
#'
#' @export
#' @author Matthew L. Fidler
infuseDur <- function(amt, dur, cmt = 1, ii = 0, addl = 0, ss = 0) {
  stop("'infuseDur()' can only be used inside an rxode2 model block", call. = FALSE)
}

#' @export
#' @keywords internal
#' @rdname rxUdfUi
rxUdfUi.infuseDur <- function(fun) {
  .dummy <- function(amt, dur, cmt, ii, addl, ss) {}
  .mc <- match.call(.dummy, fun)
  .amt <- deparse1(.mc$amt)
  .dur <- deparse1(.mc$dur)
  .cmt <- .rxUdfUiCmt(.mc$cmt)
  .ii <- deparse1(.mc$ii)
  if (.ii == "NULL") .ii <- "0"
  .addl <- deparse1(.mc$addl)
  if (.addl == "NULL") .addl <- "0"
  .ss <- deparse1(.mc$ss)
  if (.ss == "NULL") .ss <- "0"
  list(replace = paste0("infuseDur(", .amt, ",", .dur, ",", .cmt, ",", .ii, ",", .addl, ",", .ss, ")"))
}

#' Reset the rxode2 system in the model
#'
#' @return This function is only meaningful inside an rxode2 model; it
#'   returns `NULL` invisibly if called from R directly (after signaling
#'   an error).
#'
#' @details
#' ## Behavior inside a model
#'
#' `reset()` is evaluated at every output time point (when the solver is
#' exactly at a scheduled event time).  The pushed event is inserted into
#' the individual's event timeline and the solver visits it now or at the
#' specified future time.
#'
#' The number of events that may be pushed per individual is limited by
#' the `maxExtra` argument of [rxSolve()].  When `maxExtra = 0`
#' (the default) there is no limit.  Exceeding the limit causes an
#' error.
#'
#' Past-time pushes (where `time < t`) are silently ignored and counted;
#' a warning is issued after solving.
#'
reset <- function() {
  stop("'reset()' can only be used inside an rxode2 model block", call. = FALSE)
}

#' Replace dose the rxode2 system in the model
#'
#' @inheritParams evid_
#'
#' @return This function is only meaningful inside an rxode2 model; it
#'   returns `NULL` invisibly if called from R directly (after signaling
#'   an error).
#'
#' @details
#' ## Behavior inside a model
#'
#' `replace()` is evaluated at every output time point (when the solver is
#' exactly at a scheduled event time).  The pushed event is inserted into
#' the individual's event timeline and the solver visits it now or at the
#' specified future time.
#'
#' The number of events that may be pushed per individual is limited by
#' the `maxExtra` argument of [rxSolve()].  When `maxExtra = 0`
#' (the default) there is no limit.  Exceeding the limit causes an
#' error.
#'
#' Past-time pushes (where `time < t`) are silently ignored and counted;
#' a warning is issued after solving.
#'
replace <- function(amt, cmt = 1) {
  stop("'replace()' can only be used inside an rxode2 model block", call. = FALSE)
}


#' @export
#' @keywords internal
#' @rdname rxUdfUi
rxUdfUi.replace <- function(fun) {
  .dummy <- function(amt, cmt=1) {}
  .mc <- match.call(.dummy, fun)
  .amt <- deparse1(.mc$amt)
  .cmt <- .rxUdfUiCmt(.mc$cmt)
  list(replace = paste0("replace(", .amt, ",", .cmt, ")"))
}


#' Multiply dose the rxode2 system in the model
#'
#' @inheritParams evid_
#'
#' @return This function is only meaningful inside an rxode2 model; it
#'   returns `NULL` invisibly if called from R directly (after signaling
#'   an error).
#'
#' @details
#' ## Behavior inside a model
#'
#' `multiply()` is evaluated at every output time point (when the solver is
#' exactly at a scheduled event time).  The pushed event is inserted into
#' the individual's event timeline and the solver visits it now or at the
#' specified future time.
#'
#' The number of events that may be pushed per individual is limited by
#' the `maxExtra` argument of [rxSolve()].  When `maxExtra = 0`
#' (the default) there is no limit.  Exceeding the limit causes an
#' error.
#'
#' Past-time pushes (where `time < t`) are silently ignored and counted;
#' a warning is issued after solving.
#'
multiply <- function(amt, cmt = 1) {
  stop("'multiply()' can only be used inside an rxode2 model block", call. = FALSE)
}


#' @export
#' @keywords internal
#' @rdname rxUdfUi
rxUdfUi.multiply <- function(fun) {
  .dummy <- function(amt, cmt=1) {}
  .mc <- match.call(.dummy, fun)
  .amt <- deparse1(.mc$amt)
  .cmt <- .rxUdfUiCmt(.mc$cmt)
  list(replace = paste0("multiply(", .amt, ",", .cmt, ")"))
}

#' Administer a phantom/transit dose inside a rxode2 model
#'
#' @inheritParams evid_
#'
#' @return This function is only meaningful inside an rxode2 model; it
#'   returns `NULL` invisibly if called from R directly (after signaling
#'   an error).
#'
#' @details
#' ## Behavior inside a model
#'
#' `phantom()` is evaluated at every output time point (when the solver is
#' exactly at a scheduled event time).  The pushed event is inserted into
#' the individual's event timeline and the solver visits it now or at the
#' specified future time.
#'
#' The number of events that may be pushed per individual is limited by
#' the `maxExtra` argument of [rxSolve()].  When `maxExtra = 0`
#' (the default) there is no limit.  Exceeding the limit causes an
#' error.
#'
#' Past-time pushes (where `time < t`) are silently ignored and counted;
#' a warning is issued after solving.
#'
#' @export
#' @author Matthew L. Fidler
phantom <- function(amt, cmt = 1, ii = 0, addl = 0, ss = 0) {
  stop("'phantom()' can only be used inside an rxode2 model block", call. = FALSE)
}

#' @export
#' @keywords internal
#' @rdname rxUdfUi
rxUdfUi.phantom <- function(fun) {
  .dummy <- function(amt, cmt, ii, addl, ss) {}
  .mc <- match.call(.dummy, fun)
  .amt   <- deparse1(.mc$amt)

  .cmt  <- .rxUdfUiCmt(.mc$cmt)

  .ii   <- deparse1(.mc$ii)
  if (.ii == "NULL") {
    .ii <- "0"
  }

  .addl   <- deparse1(.mc$addl)
  if (.addl == "NULL") {
    .addl <- "0"
  }

  .ss   <- deparse1(.mc$ss)
  if (.ss == "NULL") {
    .ss <- "0"
  }

  list(replace = paste0("phantom(", .amt, ",", .cmt, ",", .ii, ",", .addl, ",", .ss, ")"))
}

#' Add observations to a model
#'
#'
#' @param ... Numeric values to be added as observations after the
#'   current time `t`.  They would be added as events with `evid = 0`
#'   and `amt = 0` at the specified future time points.  The time
#'   points are relative to the current model time `t` (i.e.,
#'   `obs(12)` adds an observation at `t + 12`).  If this is a ui
#'   model, you can also specify observations with obs(seq(0, 24,
#'   by=0.5)) to add observations every 0.5 time units from t to t+24
#'   since it will resolve to the obs() inside of the final rxode2
#'   model.
#'
#' @return This function is only meaningful inside an rxode2 model; it
#'   returns `NULL` invisibly if called from R directly (after signaling
#'   an error).
#'
#' @details
#' ## Behavior inside a model
#'
#' `obs()` is evaluated at every output time point (when the solver is
#' exactly at a scheduled event time).  The pushed event is inserted into
#' the individual's event timeline and the solver visits it now or at the
#' specified future time.
#'
#' The number of events that may be pushed per individual is limited by
#' the `maxExtra` argument of [rxSolve()].  When `maxExtra = 0`
#' (the default) there is no limit.  Exceeding the limit causes an
#' error.
#'
#' Past-time pushes (where `time < t`) are silently ignored and counted;
#' a warning is issued after solving.
#'
#' @export
#' @author Matthew L. Fidler
obs <- function(...) {
  stop("'obs()' can only be used inside an rxode2 model block", call. = FALSE)
}

#' @export
#' @keywords internal
#' @rdname rxUdfUi
rxUdfUi.obs <- function(fun) {
  .mc1 <- fun
  .mc1[[1]] <- quote(`list`)
  .mc1 <- try(eval(.mc1), silent=TRUE)
  if (inherits(.mc1, "try-error")) {
    list(replace=deparse1(fun))
  } else {
    .mc1 <- try(unlist(.mc1, recursive=TRUE, use.names=FALSE), silent=TRUE)
    if (inherits(.mc1, "try-error")) {
      list(replace=deparse1(fun))
    } else if (is.numeric(.mc1) || is.integer(.mc1)) {
      .obs <- as.character(.mc1)
      list(replace = paste0("obs(", paste(.obs, collapse=", "), ")"))
    } else {
      list(replace=deparse1(fun))
    }
  }
}

rxTest({
  test_that("exceeding maxExtra raises an error (single subject)", {
    # This model unconditionally pushes a new event at t+0.1 on every ODE
    # evaluation step.  Each integration interval (0.1 wide) produces one push,
    # so seq(0, 2, by=0.1) creates 20 intervals -> 20 pushes, which exceeds
    # maxExtra=10 and must raise an error.
    mCascade <- rxode2({
      d/dt(x) <- -x
      evid_(t + 0.1, 1, 10, 1, 0, 0, 0, 0)
    })
    e <- et(amt = 1, time = 0) |> et(seq(0, 2, by = 0.1))
    expect_error(
      rxSolve(mCascade, c(x = 1), e, maxExtra = 10L),
      regexp = "maxExtra"
    )

    # Now with linear compartment model -- bolus
    mLinCascade <- rxode2({
      cp <- linCmt(cl, v)
      evid_(t+0.1, 1, 10, 1, 0, 0, 0, 0)
    })
    e <- et(amt = 1, time = 0) |> et(seq(0, 2, by = 0.1))
    expect_error(
      rxSolve(mLinCascade, c(cl = 2, v=1), e, maxExtra = 10L),
      regexp = "maxExtra")

    # Oral
    mLinCascade <- rxode2({
      cp <- linCmt(ka, cl, v)
      evid_(t+0.1, 1, 10, 1, 0, 0, 0, 0)
    })
    e <- et(amt = 1, time = 0) |> et(seq(0, 2, by = 0.1))
    expect_error(
      rxSolve(mLinCascade, c(ka=0.02, cl = 2, v=1), e, maxExtra = 10L),
      regexp = "maxExtra")

  })

  test_that("exceeding maxExtra does not crash R with parallel subjects", {
    # Same cascading model, multiple subjects solved in parallel.
    # Must raise an error rather than segfaulting.
    mCascade <- rxode2({
      d/dt(x) <- -x
      evid_(t + 0.1, 1, 10, 1, 0, 0, 0, 0)
    })
    nSub <- 4L
    params <- data.frame(id = seq_len(nSub), x = rep(1, nSub))
    e <- et(amt = 1, time = 0) |> et(seq(0, 2, by = 0.1))
    expect_error(
      rxSolve(mCascade, params, e, maxExtra = 10L, cores = 2L),
      regexp = "maxExtra"
    )

    # Try with linear compartment model
    mLinCascade <- rxode2({
      cp <- linCmt(cl, v)
      evid_(t+0.1, 1, 10, 1, 0, 0, 0, 0)
    })

    params <- data.frame(id = seq_len(nSub), cl = rep(2, nSub),
                         v=rep(1, nSub))

    expect_error(
      rxSolve(mCascade, params, e, maxExtra = 10L, cores = 2L),
      regexp = "maxExtra"
    )

    mLinCascade <- rxode2({
      cp <- linCmt(ka, cl, v)
      evid_(t+0.1, 1, 10, 1, 0, 0, 0, 0)
    })

    params <- data.frame(id = seq_len(nSub), cl = rep(2, nSub),
                         v=rep(1, nSub), ka=rep(0.02, nSub))

    expect_error(
      rxSolve(mCascade, params, e, maxExtra = 10L, cores = 2L),
      regexp = "maxExtra"
    )
  })

  test_that("maxExtra=0 allows unlimited pushes (no error)", {
    # Limited cascade: condition prevents indefinite growth.
    # maxExtra=0 disables the guard entirely; should succeed without error.
    m <- rxode2({
      d/dt(x) <- -x
      evid_(t + 0.1, 1, 10, 1, 0, 0, 0, 0)
    })
    e <- et(amt = 1, time = 0) |> et(seq(0, 2, by = 0.1))
    expect_no_error(rxSolve(m, c(x = 1), e, maxExtra = 0L))
  })

  for (meth in .methods0) {
    test_that(paste0("bolus push via evid_() adds extra timepoints and affects trajectory (",
                     meth,
                     ")"), {

                       m <- rxode2({
                         d/dt(depot) <- -ka * depot
                         d/dt(central) <- ka * depot - cl/vd * central
                         cp <- central / vd
                         if (t < 24) {
                           evid_(t + 12, 1, 50, 1, 0, 0, 0, 0)  # bolus 50 units to cmt 1 at t+12
                         }
                       })
                       e <- et(amt = 100, time = 0) |>
                         et(seq(0, 24, by = 1))
                       p <- c(ka = 0.5, cl = 1, vd = 10)
                       r <- rxSolve(m, p, e, method = meth)
                       # The pushed dose at t+12 should create an extra event, affecting cp after t=12
                       expect_true(nrow(r) > 0)
                       # The bolus at t=12 should be visible -- cp should rise after t=12
                       cp12 <- r$cp[r$time == 12]
                       cp14 <- r$cp[r$time == 14]
                       expect_true(length(cp12) > 0 && length(cp14) > 0)
                       expect_true(cp14 > cp12)

                       m2 <- rxode2({
                         d/dt(depot) <- -ka * depot
                         d/dt(central) <- ka * depot - cl/vd * central
                         cp <- central / vd
                       })
                       e2 <- et(amt = 100, time = 0) |>
                         et(seq(0, 24, by = 1)) |>
                         et(amt = 50, time = 12, ii=1, until=30)

                       p <- c(ka = 0.5, cl = 1, vd = 10)
                       r2 <- rxSolve(m2, p, e2, method = meth)

                       expect_equal(r$cp, r2$cp, tolerance = 1e-5)
                     })
  }
  test_that(paste0("bolus push via evid_() adds extra timepoints and affects trajectory (linCmt)"), {
    # Now try with linCmt() only
    m <- rxode2({
      cp <- linCmt(ka, cl, v)
      if (t < 24) {
        evid_(t + 12, 1, 50, 1, 0, 0, 0, 0)  # bolus 50 units to cmt 1 at t+12
      }
    })
    e <- et(amt = 100, time = 0) |>
      et(seq(0, 24, by = 1))
    p <- c(ka = 0.5, cl = 1, v = 10)
    rLin <- rxSolve(m, p, e)
    # rLin$time >= 19 should be non-zero
    expect_true(all(rLin$cp[rLin$time >= 19] > 0))
    # The pushed dose at t+12 should create an extra event, affecting cp after t=12
    expect_true(nrow(rLin) > 0)
    # The bolus at t=12 should be visible -- cp should rise after t=12
    cp12 <- rLin$cp[rLin$time == 12]
    cp14 <- rLin$cp[rLin$time == 14]
    expect_true(length(cp12) > 0 && length(cp14) > 0)
    expect_true(cp14 > cp12)

    m2 <- rxode2({
      cp <- linCmt(ka, cl, v)
    })
    e2 <- et(amt = 100, time = 0) |>
      et(seq(0, 24, by = 1)) |>
      et(amt = 50, time = 12, ii=1, until=30)

    rLin2 <- rxSolve(m, p, e)

    expect_equal(rLin$cp, rLin2$cp)
  })
  for (meth in .methods0) {
    test_that(paste0("infusion push via evid_() adds start+stop events (", meth, ")"), {
      m <- rxode2({
        d/dt(central) <- -cl / vd * central
        cp <- central / vd
        if (t >= 14 && t < 15) {
          evid_(15, 1, 100, 1, 10, 0, 0, 0)  # rate=10, so dur=10h; push from t=14
        }
      })
      e <- et(amt = 100, time = 0) |> et(seq(0, 30, by = 1))
      p <- c(cl = 1, vd = 10)
      r <- rxSolve(m, p, e, method=meth)
      expect_true(nrow(r) > 0)
      # Central should rise during the pushed infusion window (t=15 to t=25)
      cp7  <- r$cp[r$time == 7]
      cp20 <- r$cp[r$time == 20]
      expect_true(length(cp7) > 0 && length(cp20) > 0)
      # At t=7, infusion is active; at t=20, it has stopped and cp is declining
      expect_true(cp7 > 0)

      m2 <- rxode2({
        d/dt(central) <- -cl / vd * central
        cp <- central / vd
      })

      e2 <- et(amt = 100, time = 0) |>
        et(seq(0, 30, by = 1)) |>
        et(amt=100, time=15, rate=10)

      r2 <- rxSolve(m2, p, e2, method=meth)

      expect_equal(r$cp, r2$cp, tolerance = 1e-5)
    })
  }

  for (meth in .methods0) {
    test_that(paste0("in-model infuse() changes duration with bioavailability (",
                     meth,
                     ")"), {
                       obs <- c(0, 1e-8, seq(0.5, 12, by = 0.5))
                       ref <- rxode2({
                         d/dt(central) <- -cl / v * central
                         cp <- central / v
                       })

                       mod <- rxode2({
                         d/dt(central) <- -cl / v * central
                         f(central) <- fc
                         cp <- central / v
                         if (t < 1e-8) {
                           infuse(100, 10, 1, 0, 0, 0)
                         }
                       })
                       p <- c(cl = 1, v = 10, fc = 0.5)
                       got <- rxSolve(mod, p, et(obs), method = meth)
                       want <- rxSolve(ref, c(cl = 1, v = 10),
                                       et(amt = 50, time = 0, rate = 10) |> et(obs),
                                       method = meth)
                       expect_equal(got$time, want$time)
                       expect_equal(got$cp, want$cp, tolerance = 1e-5)
                     })
  }

  for (meth in .methods0) {
    test_that(paste0("in-model infuseDur() changes rate with bioavailability (",
                     meth, ")"), {
                       obs <- c(0, 1e-8, seq(0.5, 12, by = 0.5))
                       ref <- rxode2({
                         d/dt(central) <- -cl / v * central
                         cp <- central / v
                       })

                       mod <- rxode2({
                         d/dt(central) <- -cl / v * central
                         f(central) <- fc
                         cp <- central / v
                         if (t < 1e-8) {
                           infuseDur(100, 10, 1, 0, 0, 0)
                         }
                       })
                       p <- c(cl = 1, v = 10, fc = 0.5)
                       got <- rxSolve(mod, p, et(obs), method = meth)
                       want <- rxSolve(ref, c(cl = 1, v = 10),
                                       et(amt = 50, time = 0, dur = 10) |> et(obs),
                                       method = meth)
                       expect_equal(got$time, want$time)
                       expect_equal(got$cp, want$cp, tolerance = 1e-5)
                     })
  }

  for (meth in .methods0) {
    test_that(paste0("in-model reset() matches an evid=3 reset event (",
                     meth,
                     ")"), {
                       obs <- seq(0, 24, by = 1)
                       e <- et(amt = 100, time = 0) |>
                         et(amt = 50, time = 18) |>
                         et(obs)
                       eRef <- et(amt = 100, time = 0) |>
                         et(time = 12, evid = 3) |>
                         et(amt = 50, time = 18) |>
                         et(obs)

                       mod <- rxode2({
                         mtime(resetAt) <- 12
                         d/dt(depot) <- -ka * depot
                         d/dt(central) <- ka * depot - cl / v * central
                         cp <- central / v
                         if (t >= resetAt && t < resetAt + 0.5 && cp > 0) {
                           reset()
                         }
                       })

                       ref <- rxode2({
                         d/dt(depot) <- -ka * depot
                         d/dt(central) <- ka * depot - cl / v * central
                         cp <- central / v
                       })

                       p <- c(ka = 0.5, cl = 1, v = 10)
                       got <- rxSolve(mod, p, e, method = meth)
                       want <- rxSolve(ref, p, eRef, method = meth)
                       gotReset <- got[got$time >= 13, ]
                       wantReset <- want[want$time >= 13, ]

                       expect_equal(sum(got$time == 12), 2)
                       expect_true(all(got$cp[got$time == 12] > 0))
                       expect_true(all(gotReset$cp[gotReset$time < 18] == 0))
                       expect_equal(gotReset$time, wantReset$time)
                       expect_equal(gotReset$depot, wantReset$depot, tolerance = 1e-5)
                       expect_equal(gotReset$central, wantReset$central, tolerance = 1e-5)
                       expect_equal(gotReset$cp, wantReset$cp, tolerance = 1e-5)
                     })
  }

  test_that(paste0("in-model reset() matches an evid=3 reset event (linCmt)"), {
    obs <- seq(0, 24, by = 1)
    e <- et(amt = 100, time = 0) |>
      et(amt = 50, time = 18) |>
      et(obs)
    eRef <- et(amt = 100, time = 0) |>
      et(time = 12, evid = 3) |>
      et(amt = 50, time = 18) |>
      et(obs)

    modLin <- rxode2({
      mtime(resetAt) <- 12
      cp <- linCmt(ka, cl, v)
      if (t >= resetAt && t < resetAt + 0.5 && cp > 0) {
        reset()
      }
    })

    refLin <- rxode2({
      cp <- linCmt(ka, cl, v)
    })

    p <- c(ka = 0.5, cl = 1, v = 10)
    gotLin <- rxSolve(modLin, p, e)
    wantLin <- rxSolve(refLin, p, eRef)
    gotLinReset <- gotLin[gotLin$time >= 13, ]
    wantLinReset <- wantLin[wantLin$time >= 13, ]

    expect_equal(sum(gotLin$time == 12), 2)
    expect_true(all(gotLin$cp[gotLin$time == 12] == 0))
    expect_true(all(gotLinReset$cp[gotLinReset$time < 18] == 0))
    expect_equal(gotLinReset$time, wantLinReset$time)
    expect_equal(gotLinReset$cp, wantLinReset$cp, tolerance = 1e-5)
  })

  test_that("in-model replace() pushes a replacement event", {
    obs <- seq(0, 24, by = 1)
    e <- et(amt = 100, time = 0) |>
      et(amt = 50, time = 18) |>
      et(obs)
    eRef <- et(amt = 100, time = 0) |>
      et(time = 13, amt = 25, cmt = 1, evid = 5) |>
      et(amt = 50, time = 18) |>
      et(obs)

    for (meth in .methods0) {
      mod <- rxode2({
        mtime(replaceAt) <- 12
        d/dt(depot) <- -ka * depot
        d/dt(central) <- ka * depot - cl / v * central
        cp <- central / v
        if (t >= replaceAt && t < replaceAt + 0.5 && depot > 0) {
          replace(25, depot)
        }
      })

      ref <- rxode2({
        d/dt(depot) <- -ka * depot
        d/dt(central) <- ka * depot - cl / v * central
        cp <- central / v
      })

      p <- c(ka = 0.5, cl = 1, v = 10)
      got <- rxSolve(mod, p, e, method = meth)
      want <- rxSolve(ref, p, eRef, method = meth)
      gotReplace <- got[got$time >= 13, ]
      wantReplace <- want[want$time >= 13, ]

      expect_equal(sum(got$time == 12), 2)
      expect_equal(gotReplace$time, wantReplace$time)
      expect_equal(gotReplace$depot, wantReplace$depot, tolerance = 1e-5)
      expect_equal(gotReplace$central, wantReplace$central, tolerance = 1e-5)
      expect_equal(gotReplace$cp, wantReplace$cp, tolerance = 1e-5)
    }
  })

  for (meth in .methods0) {
    test_that(paste0("in-model multiply() pushes a multiplication event (",
                     meth,
                     ")"), {
                       obs <- seq(0, 24, by = 1)
                       e <- et(amt = 100, time = 0) |>
                         et(amt = 50, time = 18) |>
                         et(obs)
                       eRef <- et(amt = 100, time = 0) |>
                         et(time = 13, amt = 2, cmt = 1, evid = 6) |>
                         et(amt = 50, time = 18) |>
                         et(obs)

                       mod <- rxode2({
                         mtime(multAt) <- 12
                         d/dt(depot) <- -ka * depot
                         d/dt(central) <- ka * depot - cl / v * central
                         cp <- central / v
                         if (t >= multAt && t < multAt + 0.5 && depot > 0) {
                           multiply(2, depot)
                         }
                       })

                       ref <- rxode2({
                         d/dt(depot) <- -ka * depot
                         d/dt(central) <- ka * depot - cl / v * central
                         cp <- central / v
                       })

                       p <- c(ka = 0.5, cl = 1, v = 10)
                       got <- rxSolve(mod, p, e, method = meth)
                       want <- rxSolve(ref, p, eRef, method = meth)
                       gotMult <- got[got$time >= 13, ]
                       wantMult <- want[want$time >= 13, ]

                       expect_equal(sum(got$time == 12), 2)
                       expect_equal(gotMult$time, wantMult$time)
                       expect_equal(gotMult$depot, wantMult$depot, tolerance = 1e-5)
                       expect_equal(gotMult$central, wantMult$central, tolerance = 1e-5)
                       expect_equal(gotMult$cp, wantMult$cp, tolerance = 1e-5)
                     })
  }

  for (meth in .methods0) {
    test_that(paste0("in-model phantom() pushes a phantom event (", meth, ")"), {
      obs <- seq(0, 20, by = 1)
      e <- et(obs)
      eRef <- et(time = 12, amt = 20, cmt = 1, evid = 7) |>
        et(obs)

      mod <- rxode2({
        mtime(phantomAt) <- 12
        d/dt(depot) <- 0
        d/dt(x) <- -k * x
        pd <- podo(depot)
        td <- tad(depot)
        if (t >= phantomAt && t < phantomAt + 0.5) {
          phantom(20, depot, 0, 0, 0)
        }
      })

      ref <- rxode2({
        d/dt(depot) <- 0
        d/dt(x) <- -k * x
        pd <- podo(depot)
        td <- tad(depot)
      })

      p <- c(k = 0.1)
      got <- rxSolve(mod, p, e, inits = c(depot = 0, x = 1), method = meth)
      want <- rxSolve(ref, p, eRef, inits = c(depot = 0, x = 1), method = meth)
      gotPhantom <- got[got$time >= 13, ]
      wantPhantom <- want[want$time >= 13, ]

      expect_equal(sum(got$time == 12), 2)
      expect_equal(gotPhantom$time, wantPhantom$time)
      expect_equal(gotPhantom$depot, wantPhantom$depot, tolerance = 1e-5)
      expect_equal(gotPhantom$x, wantPhantom$x, tolerance = 1e-5)
      expect_equal(gotPhantom$pd, wantPhantom$pd, tolerance = 1e-5)
      expect_equal(gotPhantom$td, wantPhantom$td, tolerance = 1e-5)
    })
  }

  for (meth in .methods0) {
    test_that(paste0("in-model bolus() passes ii to repeated bolus events (", meth, ")"), {
      obs <- seq(0, 40, by = 1)
      e <- et(amt = 100, time = 0) |>
        et(obs)
      eRef <- et(amt = 100, time = 0) |>
        et(time = 13, amt = 50, cmt = 1, evid = 1) |>
        et(time = 24, amt = 50, cmt = 1, evid = 1) |>
        et(time = 36, amt = 50, cmt = 1, evid = 1) |>
        et(obs)

      mod <- rxode2({
        mtime(bolusAt) <- 12
        d/dt(depot) <- -ka * depot
        d/dt(central) <- ka * depot - cl / v * central
        cp <- central / v
        if (t >= bolusAt && t < bolusAt + 0.5 && depot > 0) {
          bolus(50, depot, 12, 2, 0)
        }
      })

      ref <- rxode2({
        d/dt(depot) <- -ka * depot
        d/dt(central) <- ka * depot - cl / v * central
        cp <- central / v
      })

      p <- c(ka = 0.5, cl = 1, v = 10)
      got <- rxSolve(mod, p, e, method = meth)
      want <- rxSolve(ref, p, eRef, method = meth)
      gotBolus <- got[got$time >= 13, ]
      wantBolus <- want[want$time >= 13, ]

      expect_equal(sum(got$time == 12), 2)
      expect_equal(gotBolus$time, wantBolus$time)
      expect_equal(gotBolus$depot, wantBolus$depot, tolerance = 1e-5)
      expect_equal(gotBolus$central, wantBolus$central, tolerance = 1e-5)
      expect_equal(gotBolus$cp, wantBolus$cp, tolerance = 1e-5)
    })
  }

  test_that("past-time evid_() produces a warning", {
    m3 <- rxode2({
      d/dt(x) <- -x
      if (t < 5) {
        evid_(t - 1, 1, 10, 1, 0, 0, 0, 0)  # past time -- should warn
      }
    })
    e <- et(amt = 1, time = 0) |> et(seq(0, 5, by = 1))
    expect_warning(
      rxSolve(m3, c(x = 0), e),
      regexp = "evid_\\(\\).*time"
    )
  })

  test_that("evid=0 observation push adds an output row without dosing", {
    m4 <- rxode2({
      d/dt(central) <- -cl / vd * central
      cp <- central / vd
      if (t < 12) {
        evid_(t + 5.5, 0, 0, 1, 0, 0, 0, 0)  # push observation at t+5.5
      }
    })
    e <- et(amt = 100, time = 0) |> et(c(0, 6, 12))
    p <- c(cl = 1, vd = 10)
    r <- rxSolve(m4, p, e)
    expect_equal(nrow(r), 5) # 5 observations should be in the output
  })

  test_that("obs() pushes multiple observation rows", {
    mObs <- rxode2({
      d/dt(central) <- -cl / vd * central
      cp <- central / vd
      if (t < 1) {
        obs(0, 1, 2, 4, 6)
      }
    })
    e <- et(amt = 100, time = 0) |> et(c(0, 12))
    p <- c(cl = 1, vd = 10)
    r <- rxSolve(mObs, p, e)
    expect_equal(nrow(r), 7)
    expect_equal(sort(unique(r$time)), c(0, 1, 2, 4, 6, 12))
  })

  test_that("addl doses via evid_() push multiple doses at ii intervals", {
    # evid_(t+6, 1, 50, 1, 0, 12, 2, 0) pushes 3 boluses: t+6, t+18, t+30
    m6 <- rxode2({
      d/dt(depot)   <- -ka * depot
      d/dt(central) <- ka * depot - cl / vd * central
      cp <- central / vd
      if (t < 1) {
        evid_(t + 6, 1, 50, 1, 0, 12, 2, 0)
      }
    })
    e <- et(amt = 100, time = 0) |> et(seq(0, 36, by = 1))
    p <- c(ka = 0.5, cl = 1, vd = 10)
    r <- rxSolve(m6, p, e)
    expect_true(nrow(r) > 0)
    # At t=28: long after t=18 dose, before t=30 dose -- cp should be low
    # At t=31: 1h after t=30 pushed dose -- cp should be rising
    cp28 <- r$cp[r$time == 28]
    cp31 <- r$cp[r$time == 31]
    expect_true(length(cp28) > 0 && length(cp31) > 0)
    expect_true(cp31 > cp28)
  })

  test_that("evid_() with depot", {
    # evid_(t+6, 1, 50, 1, 0, 12, 2, 0) pushes 3 boluses: t+6, t+18, t+30
    m6 <- rxode2({
      d/dt(depot)   <- -ka * depot
      d/dt(central) <- ka * depot - cl / vd * central
      cp <- central / vd
      if (t < 1) {
        evid_(t + 6, 1, 50, depot, 0, 12, 2, 0)
      }
    })
    e <- et(amt = 100, time = 0) |> et(seq(0, 36, by = 1))
    p <- c(ka = 0.5, cl = 1, vd = 10)
    r <- rxSolve(m6, p, e)
    expect_true(nrow(r) > 0)
    # At t=28: long after t=18 dose, before t=30 dose -- cp should be low
    # At t=31: 1h after t=30 pushed dose -- cp should be rising
    cp28 <- r$cp[r$time == 28]
    cp31 <- r$cp[r$time == 31]
    expect_true(length(cp28) > 0 && length(cp31) > 0)
    expect_true(cp31 > cp28)
  })

  test_that("evid_() with 'depot'", {
    # evid_(t+6, 1, 50, 1, 0, 12, 2, 0) pushes 3 boluses: t+6, t+18, t+30
    m6 <- rxode2({
      d/dt(depot)   <- -ka * depot
      d/dt(central) <- ka * depot - cl / vd * central
      cp <- central / vd
      if (t < 1) {
        evid_(t + 6, 1, 50, 'depot', 0, 12, 2, 0)
      }
    })
    e <- et(amt = 100, time = 0) |> et(seq(0, 36, by = 1))
    p <- c(ka = 0.5, cl = 1, vd = 10)
    r <- rxSolve(m6, p, e)
    expect_true(nrow(r) > 0)
    # At t=28: long after t=18 dose, before t=30 dose -- cp should be low
    # At t=31: 1h after t=30 pushed dose -- cp should be rising
    cp28 <- r$cp[r$time == 28]
    cp31 <- r$cp[r$time == 31]
    expect_true(length(cp28) > 0 && length(cp31) > 0)
    expect_true(cp31 > cp28)
  })

  test_that("evid_() with \"depot\"", {
    # evid_(t+6, 1, 50, 1, 0, 12, 2, 0) pushes 3 boluses: t+6, t+18, t+30
    m6 <- rxode2({
      d/dt(depot)   <- -ka * depot
      d/dt(central) <- ka * depot - cl / vd * central
      cp <- central / vd
      if (t < 1) {
        evid_(t + 6, 1, 50, 'depot', 0, 12, 2, 0)
      }
    })
    e <- et(amt = 100, time = 0) |> et(seq(0, 36, by = 1))
    p <- c(ka = 0.5, cl = 1, vd = 10)
    r <- rxSolve(m6, p, e)
    expect_true(nrow(r) > 0)
    # At t=28: long after t=18 dose, before t=30 dose -- cp should be low
    # At t=31: 1h after t=30 pushed dose -- cp should be rising
    cp28 <- r$cp[r$time == 28]
    cp31 <- r$cp[r$time == 31]
    expect_true(length(cp28) > 0 && length(cp31) > 0)
    expect_true(cp31 > cp28)
  })

  test_that("classic rxode2 internal evid >= 100 passes through verbatim", {
    # internal evid for a bolus to cmt 1: cmt100=0, rateI=0, cmt99=1, flg=1 => evid=101
    m5 <- rxode2({
      d/dt(depot)   <- -ka * depot
      d/dt(central) <- ka * depot - cl / vd * central
      cp <- central / vd
      if (t < 24) {
        evid_(t + 12, 101, 50, 1, 0, 0, 0, 0)
      }
    })
    e <- et(amt = 100, time = 0) |> et(seq(0, 24, by = 1))
    p <- c(ka = 0.5, cl = 1, vd = 10)
    r <- rxSolve(m5, p, e)
    expect_true(nrow(r) > 0)
    cp12 <- r$cp[r$time == 12]
    cp14 <- r$cp[r$time == 14]
    expect_true(length(cp12) > 0 && length(cp14) > 0)
    expect_true(cp14 > cp12)
  })

  for (meth in .methods0) {
    test_that(paste0("splitBolus applies to boluses pushed by evid_() [",meth, "]"), {
      mSplit <- rxode2({
        splitBolus(depot, depot, central)
        d/dt(depot) <- -ka * depot
        d/dt(central) <- ka * depot - cl / v * central
        cp <- central / v
        if (t < 1) {
          evid_(t + 6, 1, 50, 1, 0, 12, 1, 0)
        }
      })

      mBase <- rxode2({
        d/dt(depot) <- -ka * depot
        d/dt(central) <- ka * depot - cl / v * central
        cp <- central / v
      })

      e <- et(amt = 100, time = 0) |>
        et(seq(0, 30, by = 1))
      eBase <- e |>
        et(amt = 100, time = 0, cmt = 2) |>
        et(amt = 50, time = 6, cmt = 1) |>
        et(amt = 50, time = 6, cmt = 2) |>
        et(amt = 50, time = 18, cmt = 1) |>
        et(amt = 50, time = 18, cmt = 2)

      p <- c(ka = 0.5, cl = 1, v = 10)
      rSplit <- rxSolve(mSplit, p, e, method = meth)
      rBase <- rxSolve(mBase, p, eBase, method = meth)

      expect_equal(rSplit$depot, rBase$depot, tolerance = 1e-5)
      expect_equal(rSplit$central, rBase$central, tolerance = 1e-5)
      expect_equal(rSplit$cp, rBase$cp, tolerance = 1e-5)
    })
  }
  test_that("evid_() ui changes work", {

    f <- function() {
      model({
        evid_(t + 12, 101, 50)
      })
    }

    f <- f()
    expect_equal(setNames(rxModelVars(f)$model["normModel"], NULL),
                 "evid_(t + 12, 101, 50, 1, 0, 0, 0, 0);\n")

  })

  test_that("obs() ui expands seq() inputs", {

    f <- function() {
      model({
        obs(seq(0, 6, by = 2))
      })
    }

    f <- f()
    expect_equal(setNames(rxModelVars(f)$model["normModel"], NULL),
                 "obs(0, 2, 4, 6);\n")

  })

  test_that("reset() ui changes work", {

    f <- function() {
      model({
        reset()
      })
    }

    f <- f()
    expect_equal(setNames(rxModelVars(f)$model["normModel"], NULL),
                 "reset();\n")

  })

  test_that("replace() ui changes work", {

    f <- function() {
      model({
        replace(10, depot)
      })
    }

    f <- f()
    expect_equal(setNames(rxModelVars(f)$model["normModel"], NULL),
                 "replace(10, depot);\n")

  })

  test_that("multiply() ui changes work", {

    f <- function() {
      model({
        multiply(10, depot)
      })
    }

    f <- f()
    expect_equal(setNames(rxModelVars(f)$model["normModel"], NULL),
                 "multiply(10, depot);\n")

  })

  test_that("phantom() ui changes work", {

    f <- function() {
      model({
        phantom(10, depot, 0, 0, 0)
      })
    }

    f <- f()
    expect_equal(setNames(rxModelVars(f)$model["normModel"], NULL),
                 "phantom(10, depot, 0, 0, 0);\n")

  })

  test_that("splitBolus() ui changes work", {
    f <- function() {
      model({
        splitBolus(depot, depot, central, peripheral)
      })
    }

    f <- f()
    expect_equal(tail(names(rxModelVars(f)), 5),
                 c("lhsOrd", "splitBolus", "strCmpParams", "timeId", "md5"))
    expect_equal(unname(rxModelVars(f)$splitBolus), c(1L, 1L, 2L, 3L))
    expect_equal(setNames(rxModelVars(f)$model["normModel"], NULL),
                 "splitBolus(depot,depot,central,peripheral);\n")
  })

  test_that("splitBolus() ui allows transfer to one target compartment", {
    f <- function() {
      model({
        splitBolus(depot, central)
      })
    }

    f <- f()
    expect_equal(unname(rxModelVars(f)$splitBolus), c(1L, 2L))
    expect_equal(setNames(rxModelVars(f)$model["normModel"], NULL),
                 "splitBolus(depot,central);\n")
  })

  test_that("splitBolus() exposes split dose lines on the ui", {
    f <- function() {
      ini({
        ka <- 1
        cl <- 1
        v <- 10
      })
      model({
        splitBolus(depot, depot, central, peripheral)
        d/dt(depot) <- -ka * depot
        d/dt(central) <- ka * depot - cl / v * central
        d/dt(peripheral) <- 0
        cp <- central / v
      })
    }

    ui <- rxode(f)

    expect_equal(rxUiGet.splitDose(list(ui)),
                 list(str2lang("splitBolus(depot, depot, central, peripheral)")))
    expect_equal(ui$splitDoseLines,
                 list(str2lang("splitBolus(depot, depot, central, peripheral)")))

    f <- function() {
      ini({
        ka <- 1
        cl <- 1
        v <- 10
      })
      model({
        d/dt(depot) <- -ka * depot
        d/dt(central) <- ka * depot - cl / v * central
        cp <- central / v
      })
    }

    ui <- rxode(f)

    expect_null(rxUiGet.splitDose(list(ui)))
    expect_null(ui$splitDoseLines)
  })

  test_that("splitBolus() is preserved in simulation models", {
    f <- function() {
      ini({
        ka <- 1
        cl <- 1
        v <- 10
      })
      model({
        splitBolus(depot, depot, central, peripheral)
        d/dt(depot) <- -ka * depot
        d/dt(central) <- ka * depot - cl / v * central
        d/dt(peripheral) <- 0
        cp <- central / v
      })
    }

    ui <- rxode(f)

    expect_error(ui$simulationModel, NA)
    mod <- ui$simulationModel
    expect_equal(unname(rxModelVars(mod)$splitBolus), c(1L, 1L, 2L, 3L))

    expect_error(ui$simulationIniModel, NA)
    mod <- ui$simulationIniModel
    expect_equal(unname(rxModelVars(mod)$splitBolus), c(1L, 1L, 2L, 3L))
  })

  for (meth in .methods0) {
    test_that(paste0("splitBolus applies to a one-target bolus pushed by evid_() [",
                     meth, "]"), {
                       mSplit <- rxode2({
                         splitBolus(depot, central)
                         d/dt(depot) <- -ka * depot
                         d/dt(central) <- ka * depot - cl / v * central
                         cp <- central / v
                         if (t < 1) {
                           evid_(t + 6, 1, 50, 1, 0, 12, 1, 0)
                         }
                       })

                       mBase <- rxode2({
                         d/dt(depot) <- -ka * depot
                         d/dt(central) <- ka * depot - cl / v * central
                         cp <- central / v
                       })

                       e <- et(seq(0, 30, by = 1))
                       eBase <- e |>
                         et(amt = 50, time = 6, cmt = 2) |>
                         et(amt = 50, time = 18, cmt = 2)

                       p <- c(ka = 0.5, cl = 1, v = 10)
                       rSplit <- rxSolve(mSplit, p, e, method = meth)
                       rBase <- rxSolve(mBase, p, eBase, method = meth)

                       expect_equal(rSplit$depot, rBase$depot, tolerance = 1e-5)
                       expect_equal(rSplit$central, rBase$central, tolerance = 1e-5)
                       expect_equal(rSplit$cp, rBase$cp, tolerance = 1e-5)
                     })
  }

  ## -------------------------------------------------------------------------
  ## bolus() with linCmt models (positional and functional forms)
  ## -------------------------------------------------------------------------

  test_that("bolus() positional form works with linCmt 1-cmt oral model", {
    m <- rxode2({
      cp <- linCmt(ka, cl, v)
      if (t > 23 && t < 25) {
        bolus(50, depot, 0, 0, 0)
      }
    })
    e <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    p <- c(ka = 0.5, cl = 1, v = 10)
    r <- suppressMessages(rxSolve(m, p, e))
    expect_true(nrow(r) > 0)
    expect_false(any(is.na(r$cp)))
    # cp >= 0 (0 at t=0 before absorption, positive thereafter)
    expect_true(all(r$cp >= 0))
    cp30 <- r$cp[r$time == 30]
    expect_true(length(cp30) > 0)
    expect_true(cp30 > 0)
  })

  test_that("bolus() positional form works with linCmt 1-cmt IV model", {
    m <- rxode2({
      cp <- linCmt(cl, v)
      if (t > 23 && t < 25) {
        bolus(50, central, 0, 0, 0)
      }
    })
    e <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    p <- c(cl = 1, v = 10)
    r <- suppressMessages(rxSolve(m, p, e))
    expect_true(nrow(r) > 0)
    expect_false(any(is.na(r$cp)))
    # IV model: cp > 0 immediately after t=0 dose, except possibly t=0 itself
    expect_true(all(r$cp[r$time > 0] > 0))
  })

  test_that("bolus() functional form (named args) works with linCmt", {
    m <- suppressMessages(rxode2(function() {
      ini({ ka <- 0.5; cl <- 1; v <- 10 })
      model({
        cp <- linCmt(ka, cl, v)
        if (t > 23 && t < 25) {
          bolus(50, cmt = depot)
        }
      })
    }))
    e <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    r <- suppressMessages(rxSolve(m, e))
    expect_true(nrow(r) > 0)
    expect_false(any(is.na(r$cp)))
    expect_true(all(r$cp >= 0))
    expect_true(all(r$cp[r$time > 0] > 0))
  })

  test_that("bolus() functional form (string cmt) works with linCmt", {
    # cmt = "depot" (string literal) must produce the bare compartment name
    # in the generated model, not a quoted token.
    m <- suppressMessages(rxode2(function() {
      ini({ ka <- 0.5; cl <- 1; v <- 10 })
      model({
        cp <- linCmt(ka, cl, v)
        if (t > 23 && t < 25) {
          bolus(50, cmt = "depot")
        }
      })
    }))
    # The generated dosing line should use the bare name `depot`, no quotes
    .txt <- paste(vapply(m$lstExpr, deparse1, character(1)), collapse = "\n")
    expect_true(grepl("bolus(50, depot", .txt, fixed = TRUE))
    expect_false(grepl("\"depot\"", .txt, fixed = TRUE))

    e <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    r <- suppressMessages(rxSolve(m, e))
    expect_true(nrow(r) > 0)
    expect_false(any(is.na(r$cp)))
    expect_true(all(r$cp >= 0))
    expect_true(all(r$cp[r$time > 0] > 0))
  })

  test_that("bolus() works in a mixed linCmt + ODE model (both compartments)", {
    # Mixed model: analytical linCmt PK feeding an ODE effect compartment.
    # bolus() to the linCmt depot AND (string cmt) to the ODE effect cmt.
    m <- suppressMessages(rxode2(function() {
      ini({ ka <- 0.5; cl <- 1; v <- 10; ke0 <- 0.3 })
      model({
        cp <- linCmt(ka, cl, v)
        d/dt(effect) <- ke0 * (cp - effect)
        if (t > 11 && t < 13) {
          bolus(20, cmt = "effect")
        }
        if (t > 23 && t < 25) {
          bolus(50, cmt = depot)
        }
      })
    }))
    # String cmt should generate a bare compartment name
    .txt <- paste(vapply(m$lstExpr, deparse1, character(1)), collapse = "\n")
    expect_true(grepl("bolus(20, effect", .txt, fixed = TRUE))
    expect_false(grepl("\"effect\"", .txt, fixed = TRUE))

    e <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    r <- suppressMessages(rxSolve(m, e))
    expect_true(nrow(r) > 0)
    expect_false(any(is.na(r$cp)))
    expect_false(any(is.na(r$effect)))
    expect_true(all(r$cp[r$time > 0] > 0))
    # effect compartment jumps after the bolus at t=12
    expect_true(r$effect[r$time == 13] > r$effect[r$time == 11])
  })

  test_that("bolus() in linCmt fires exactly once per qualifying observation", {
    # Bolus fires when t == 24 (t %% 24 == 0 && t > 0).
    # The extra dose should appear exactly once in the solved output.
    m <- rxode2({
      cp <- linCmt(ka, cl, v)
      if (t == 24) {
        bolus(50, depot, 0, 0, 0)
      }
    })
    e <- et(amt = 100, time = 0) |> et(seq(0, 48, by = 1))
    p <- c(ka = 0.5, cl = 1, v = 10)
    r <- suppressMessages(rxSolve(m, p, e, addDosing = TRUE))
    # Exactly one extra dose event at t=24 (from bolus push)
    doses <- r[r$evid == 1L & r$time == 24, ]
    expect_equal(nrow(doses), 1L,
                 label = "bolus() should fire exactly once at t=24")
  })
})

rxTest({
  test_that("rxPrune captures evid_ inside if clause", {
    m <- rxode2({
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - cl / vd * central
      cp <- central / vd
      if (t < 24) {
        evid_(t + 12, 1, 50, 1, 0, 0, 0, 0)
      }
    })
    pruned <- rxPrune(m)
    caps <- attr(pruned, "capturedEvid")

    expect_false(grepl("evid_", pruned))
    expect_true(grepl("rxCaptureId1", pruned))
    expect_length(caps, 1L)
    expect_equal(caps[[1]]$id, 1L)
    expect_match(caps[[1]]$original, "evid_")
    expect_match(caps[[1]]$condition, "t")     # condition contains t
    expect_equal(caps[[1]]$capVar, "rxCaptureId1")
  })

  test_that("rxPrune captures bolus with correct args inside if clause", {
    m <- rxode2({
      d/dt(A) <- -ka * A
      if (t < 24) {
        bolus(50, 1, 0, 0, 0)
      }
    })
    pruned <- rxPrune(m)
    caps <- attr(pruned, "capturedEvid")

    expect_false(grepl("bolus", pruned))
    expect_true(grepl("rxCaptureId1", pruned))
    expect_length(caps, 1L)
    expect_match(caps[[1]]$original, "bolus")
  })

  test_that("rxPrune captures all adaptive dosing function types", {
    # (evid_, bolus, infuse, infuseDur, reset, replace, multiply, phantom, obs)
    # Each is wrapped in a simple model; correct arg counts per grammar.
    cases <- list(
      list(fn = "evid_",     call = "evid_(t+12, 1, 50, 1, 0, 0, 0, 0)"),
      list(fn = "bolus",     call = "bolus(50, 1, 0, 0, 0)"),
      list(fn = "infuse",    call = "infuse(50, 10, 1, 0, 0, 0)"),
      list(fn = "infuseDur", call = "infuseDur(50, 2, 1, 0, 0, 0)"),
      list(fn = "reset",     call = "reset()"),
      list(fn = "replace",   call = "replace(50, 1)"),
      list(fn = "multiply",  call = "multiply(0.5, 1)"),
      list(fn = "phantom",   call = "phantom(50, 1, 0, 0, 0)"),
      list(fn = "obs",       call = "obs(cp)")
    )
    for (case in cases) {
      model_str <- paste0(
        "d/dt(A) <- -ka * A\ncp <- A/vd\nif (t > 0) {\n  ",
        case$call, "\n}"
      )
      m <- tryCatch(
        suppressMessages(rxode2(model_str)),
        error = function(e) NULL
      )
      if (is.null(m)) next   # skip if parser rejects (e.g. obs)
      pruned <- rxPrune(m)
      caps <- attr(pruned, "capturedEvid")
      expect_false(
        grepl(case$fn, pruned, fixed = TRUE),
        info = paste0(case$fn, " should be removed from pruned string")
      )
      expect_true(
        !is.null(caps) && length(caps) >= 1L,
        info = paste0(case$fn, " should produce at least one capture")
      )
      expect_match(caps[[1]]$original, case$fn,
                   info = paste0(case$fn, " original should be stored"))
    }
  })

  test_that("rxPrune captures unconditional adaptive dosing call with condition '1'", {
    m <- rxode2({
      d/dt(A) <- -ka * A
      bolus(50, 1, 0, 0, 0)
    })
    pruned <- rxPrune(m)
    caps <- attr(pruned, "capturedEvid")

    expect_false(grepl("bolus", pruned))
    expect_length(caps, 1L)
    expect_equal(caps[[1]]$condition, "1")
    expect_true(grepl("rxCaptureId1 <- 1", pruned, fixed = TRUE))
  })

  test_that("rxPrune captures nested if conditions by multiplying them", {
    m <- rxode2({
      d/dt(A) <- -ka * A
      if (t > 0) {
        if (t < 24) {
          bolus(50, 1, 0, 0, 0)
        }
      }
    })
    pruned <- rxPrune(m)
    caps <- attr(pruned, "capturedEvid")

    expect_length(caps, 1L)
    # Both conditions must appear in the combined condition string
    expect_match(caps[[1]]$condition, "t")
    # Conditions are multiplied together (two conditions -> two sets of parens with *)
    expect_match(caps[[1]]$condition, "\\*")
  })

  test_that("rxPrune assigns unique sequential IDs to multiple adaptive dosing calls", {
    m <- rxode2({
      d/dt(A) <- -ka * A
      if (t < 12) {
        bolus(50, 1, 0, 0, 0)
      }
      if (t >= 12) {
        replace(10, 1)
      }
    })
    pruned <- rxPrune(m)
    caps <- attr(pruned, "capturedEvid")

    expect_length(caps, 2L)
    expect_equal(caps[[1]]$id, 1L)
    expect_equal(caps[[2]]$id, 2L)
    expect_true(grepl("rxCaptureId1", pruned))
    expect_true(grepl("rxCaptureId2", pruned))
    expect_match(caps[[1]]$original, "bolus")
    expect_match(caps[[2]]$original, "replace")
  })

  test_that("rxPrune attaches no capturedEvid attribute when no adaptive dosing present", {
    m <- rxode2({
      if (t < 12) cp <- A / vd
      d/dt(A) <- -ka * A
    })
    pruned <- rxPrune(m)
    caps <- attr(pruned, "capturedEvid")
    expect_true(is.null(caps) || length(caps) == 0L)
  })

  test_that(".restoreAdaptiveDosing builds correct if-wrapped restore lines", {
    caps <- list(
      list(id = 1L, original = "bolus(50, 1, 0, 0, 0)", condition = "(t<24)",
           capVar = "rxCaptureId1"),
      list(id = 2L, original = "reset()", condition = "1",
           capVar = "rxCaptureId2")
    )
    lines <- .restoreAdaptiveDosing(caps)
    expect_equal(lines[[1]], "if (rxCaptureId1) { bolus(50, 1, 0, 0, 0) }")
    expect_equal(lines[[2]], "if (rxCaptureId2) { reset() }")
  })

  test_that(".rxLoadPrune stores captures and restore lines in rxS env", {
    rxReq("symengine")
    m <- rxode2({
      d/dt(A) <- -ka * A
      if (t < 24) {
        bolus(50, 1, 0, 0, 0)
      }
    })
    env <- suppressMessages(rxode2:::.rxLoadPrune(m, doConst = TRUE))
    expect_true(!is.null(env$..capturedEvid))
    expect_length(env$..capturedEvid, 1L)
    expect_true(!is.null(env$..restoreLines))
    expect_match(env$..restoreLines[[1]], "bolus")
    expect_match(env$..restoreLines[[1]], "rxCaptureId1")
  })

  test_that("rxPrune captures adaptive dosing in else branch (negated condition)", {
    m <- rxode2({
      if (t < 12) {
        x <- 1
      } else {
        bolus(50, 1, 0, 0, 0)
      }
      d/dt(A) <- -ka * A
    })
    pruned <- rxPrune(m)
    caps <- attr(pruned, "capturedEvid")

    expect_length(caps, 1L)
    # The else branch condition is negated: 1-(t<12) form
    expect_match(caps[[1]]$condition, "1-")
  })
})

#' @export
rxUdfUi.linMod <- function(fun) {
  eval(fun)
}
attr(rxUdfUi.linMod, "nargs") <- 2L

#' @export
rxUdfUi.linMod0 <- function(fun) {
  eval(fun)
}
attr(rxUdfUi.linMod0, "nargs") <- 2L

#' @export
rxUdfUi.linModB <- function(fun) {
  eval(fun)
}
attr(rxUdfUi.linModB, "nargs") <- 2L

#' @export
rxUdfUi.linModB0 <- function(fun) {
  eval(fun)
}
attr(rxUdfUi.linModB0, "nargs") <- 2L

#' @export
rxUdfUi.linModA <- function(fun) {
  eval(fun)
}
attr(rxUdfUi.linModA, "nargs") <- 2L

#' @export
rxUdfUi.linModA0 <- function(fun) {
  eval(fun)
}
attr(rxUdfUi.linModA0, "nargs") <- 2L

#' @export
rxUdfUi.linModD <- function(fun) {
  eval(fun)
}
attr(rxUdfUi.linModD, "nargs") <- 3L

#' @export
rxUdfUi.linModD0 <- function(fun) {
  eval(fun)
}
attr(rxUdfUi.linModD0, "nargs") <- 3L

#' @export
rxUdfUi.linModM <- function(fun) {
  eval(fun)
}
attr(rxUdfUi.linModM, "nargs") <- 2L

#' @export
rxUdfUi.linModM0 <- function(fun) {
  eval(fun)
}
attr(rxUdfUi.linModM0, "nargs") <- 2L

#' @export
rxUdfUi.default <- function(fun) {
  stop("rxode2 user defined function '", fun, "' not supported", call.=FALSE) # nocov
}

#' Linear model to replace in rxode2 ui model
#'
#' @param variable The variable that the rxode2 will be made on.
#'
#' @param power The power of the polynomial that will be generated.
#'
#' @param intercept Boolean that tells if the intercept be generated.
#'
#' @param type the type of linear model replacement to be used.
#'
#' @param num the number the particular model is being generated. If
#'   unspecified, query using `rxUdfUiNum()`.
#'
#' @param iniDf the initialization `data.frame`, if `NULL` query using
#'   `rxUdfUiIniDf()`
#'
#' @param dv the dependent variable to use to generate the initial
#'   estimates from the data. If `NULL` query using `rxUdfUiData()`.
#'
#' @param data logical that tells if the initial estimates of the
#'   linear model should be estimated from the data.
#'
#' @param mv logical that tell if the model variables need to be used
#'   to generate model variables.
#'
#' @param ... arguments that are passed to `linMod()` for the other
#'   abbreviations of `linMod()`
#'
#' @return a list for use in when generating the `rxode2` ui model see
#'   `rxUdfUi()` for details.
#'
#' @export
#' @family User functions
#' @author Matthew L. Fidler
#' @examples
#'
#' linMod(x, 3)
linMod <- function(variable, power, dv="dv",
                   intercept=TRUE,type=c("replace", "before", "after"),
                   num=NULL, iniDf=NULL, data=FALSE, mv=FALSE) {
  .dv <- as.character(substitute(dv))
  .tmp <- suppressWarnings(try(force(dv), silent=TRUE))
  if (!inherits(.tmp, "try-error")) {
    if (is.character(.tmp)) {
      .dv <- dv
    }
  }
  .var <- as.character(substitute(variable))
  .tmp <- try(force(variable), silent=TRUE)
  .doExp3 <- FALSE
  if (!inherits(.tmp, "try-error")) {
    if (is.character(.tmp)) {
      .var <- variable
    } else if (!inherits(.tmp, "formula")) {
      .dv <- as.character(substitute(dv))
      .tmp <- suppressWarnings(try(force(dv), silent=TRUE))
      if (!inherits(.tmp, "try-error")) {
        if (is.character(.tmp)) {
          .dv <- dv
        }
      }
    } else if (length(variable) == 2) {
      if (!identical(variable[[1]], quote(`~`))) {
        stop("unexpected formula, needs to be the form ~x^3",
             call.=FALSE)
      }
      .doExp3 <- TRUE
      .exp3 <- variable[[2]]
    } else {
      if (length(variable) != 3) {
        stop("unexpected formula, needs to be the form dv~x^3",
             call.=FALSE)
      }
      if (!identical(variable[[1]], quote(`~`))) {
        stop("unexpected formula, needs to be the form dv~x^3",
             call.=FALSE)
      }
      .dv <- as.character(variable[[2]])
      data <- TRUE
      .exp3 <- variable[[3]]
      .doExp3 <- TRUE
    }
    if (.doExp3) {
      if (length(.exp3) == 1) {
        .var <- variable <- as.character(.exp3)
        power <- 1
      } else if (length(.exp3) == 3) {
        if (!identical(.exp3[[1]], quote(`^`))) {
          stop("unexpected formula, needs to be the form dv~x^3",
               call.=FALSE)
        }
        if (!is.numeric(.exp3[[3]])) {
          stop("unexpected formula, needs to be the form dv~x^3",
               call.=FALSE)
        }
        .var <- variable <- as.character(.exp3[[2]])
        power <- .exp3[[3]]
      } else {
        stop("unexpected formula, needs to be the form dv~x^3",
             call.=FALSE)
      }
    }
  }
  checkmate::assertCharacter(.var, len=1L, any.missing=FALSE, pattern = "^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$", min.chars=1L,
                             .var.name="variable")
  checkmate::assertCharacter(.dv, len=1L, any.missing=FALSE, pattern = "^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$", min.chars=1L,
                             .var.name="dv")
  checkmate::assertLogical(intercept, len=1L, any.missing=FALSE)
  checkmate::assertIntegerish(power, lower=ifelse(intercept, 0L, 1L), len=1L)
  if (is.null(num)) {
    num <- rxUdfUiNum()
  }
  checkmate::assertIntegerish(num, lower=1, any.missing=FALSE, len=1)
  if (mv && is.null(rxUdfUiMv())) {
    if (intercept) {
      return(list(replace=paste0("linModM(", .var, ", ", power, ")"),
                  uiUseMv=TRUE))
    } else {
      return(list(replace=paste0("linModM0(", .var, ", ", power, ")"),
                  uiUseMv=TRUE))
    }
  }
  if (data && is.null(rxUdfUiData())) {
    if (intercept) {
      return(list(replace=paste0("linModD(", .var, ", ", power, ", ", .dv, ")"),
                  uiUseData=TRUE))
    } else {
      return(list(replace=paste0("linModD0(", .var, ", ", power, ",", .dv, ")"),
                  uiUseData=TRUE))
    }
  }
  if (is.null(iniDf)) {
    iniDf <- rxUdfUiIniDf()
  }
  assertIniDf(iniDf, null.ok=TRUE)
  type <- match.arg(type)
  .mv <- rxUdfUiMv()
  if (!is.null(.mv)) {
    .varsMv <- c(.mv$lhs, .mv$params, .mv$state)
    .pre <- paste0(.var, num, rxIntToLetter(seq_len(power+ifelse(intercept, 1L, 0L))-1L))
    .pre <- vapply(.pre, function(v) {
      if (v %in% .varsMv) {
        paste0("rx.linMod.", v)
      } else {
        v
      }
    }, character(1), USE.NAMES=FALSE)
  } else {
    .pre <- paste0("rx.linMod.", .var, num, rxIntToLetter(seq_len(power+ifelse(intercept, 1L, 0L))-1L))
  }

  if (!is.null(iniDf)) {
    .theta <- iniDf[!is.na(iniDf$ntheta),,drop=FALSE]
    if (length(.theta$ntheta) > 0L) {
      .maxTheta <- max(.theta$ntheta)
      .theta1 <- .theta[1,]
    } else {
      .maxTheta <- 0L
      .theta1 <- .rxBlankIni("theta")
    }
    .theta1$lower <- -Inf
    .theta1$upper <- Inf
    .theta1$fix <- FALSE
    .theta1$label <- NA_character_
    .theta1$backTransform <- NA_character_
    .theta1$condition <- NA_character_
    .theta1$err <- NA_character_
    .est <- rep(0, length(.pre))
    if (data) {
      .dat <- rxUdfUiData()
      .wdv <- which(tolower(names(.dat)) == tolower(.dv))
      if (length(.wdv) == 0L) {
        warning(.dv, "not found in data, so no initial estimates will be set to zero")
      } else {
        names(.dat)[.wdv] <- .dv
        .model <-
          stats::lm(
            stats::as.formula(
              paste0(.dv, " ~ stats::poly(", .var, ",", power, ")",
                     ifelse(intercept, "", "+0"))),
            data=rxUdfUiData()
          )
        .est <- coef(.model)
      }
    }
    .cur <- c(list(.theta),
              lapply(seq_along(.pre), function(i) {
                .cur <- .theta1
                .cur$name <- .pre[i]
                .cur$est <- .est[i]
                .cur$ntheta <- .maxTheta+i
                .cur
              }))
    .theta <- do.call(`rbind`, .cur)
    .eta <- iniDf[is.na(iniDf$neta),,drop=FALSE]
    .iniDf <- rbind(.theta, .eta)
  } else {
    .iniDf <- NULL
  }
  .linMod <- paste(vapply(seq_along(.pre),
                          function(i) {
                            if (intercept) {
                              if (i == 1) return(.pre[i])
                              if (i == 2) return(paste0(.pre[i], "*", .var))
                              paste0(.pre[i], "*", paste0(.var,"^", i-1L))
                            } else {
                              if (i == 1) return(paste0(.pre[i], "*", .var))
                              paste0(.pre[i], "*", paste0(.var,"^", i))
                            }
                          }, character(1)), collapse="+")
  if (type == "replace") {
    list(replace=.linMod,
         iniDf=.iniDf )
  } else if (type == "before") {
    .replace <- paste0("rx.linMod.", .var, ".f", num)
    list(before=paste0(.replace, " <- ", .linMod),
         replace=.replace,
         iniDf=.iniDf)
  } else if (type == "after") {
    .replace <- paste0("rx.linMod.", .var, ".f", num)
    list(after=paste0(.replace, " <- ", .linMod),
         replace="0",
         iniDf=.iniDf)
  }
}
#' @describeIn linMod linear model without intercept
#' @export
linMod0 <- function(...,intercept=FALSE) {
  linMod(..., intercept=intercept)
}

#' @describeIn linMod linear model before where it occurs
#' @export
linModB <- function(..., type="before") {
  linMod(..., type=type)
}

#' @describeIn linMod linear model before where the user function occurs
#' @export
linModB0 <- function(..., intercept=FALSE, type="before") {
  linMod(..., intercept=intercept, type=type)
}
#' @describeIn linMod linear model after where the user function occurs
linModA <- function(..., type="after") {
  linMod(..., type=type)
}

#' @describeIn linMod liner model without an intercept placed after where the user function occurs
#' @export
linModA0 <- function(..., intercept=FALSE, type="after") {
  linMod(..., intercept=intercept, type=type)
}

#' @describeIn linMod linear model where initial estimates are generated from the data
#' @export
linModD <- function(..., intercept=TRUE, data=TRUE) {
  linMod(..., intercept=intercept, data=data)
}

#' @describeIn linMod linear model where initial estimates are generated from the data (no intercept)
#' @export
linModD0 <- function(..., intercept=FALSE, data=TRUE) {
  linMod(..., intercept=intercept, data=data)
}

#' @describeIn linMod linear model where the model variables are used to generate the model variables
#' @export
linModM <- function(..., intercept=TRUE, mv=TRUE) {
  linMod(..., intercept=intercept, mv=mv)
}
#' @describeIn linMod linear model where the model variables are used to generate the model variables (no intercept)
#' @export
linModM0 <- function(..., intercept=FALSE, mv=TRUE) {
  linMod(..., intercept=intercept, mv=mv)
}

rxTest({

  udf <- function(x, y, ...) {
    x + y
  }

  expect_error(rxode2parse("b <- udf(x, y)"))

  udf <- function(x, y) {
    x + y
  }

  expect_error(rxode2parse("b <- udf(x, y)"), NA)

  expect_error(rxode2parse("b <- udf(x, y, z)"))

  rxode2parse("b <- udf(x, y)", code="udf.c")

  expect_true(file.exists("udf.c"))

  if (file.exists("udf.c")) {
    lines <- readLines("udf.c")
    unlink("udf.c")
    expect_false(file.exists("udf.c"))
  }

  .w <- which(grepl("b =_udf(\"udf\",", lines, fixed=TRUE))
  expect_true(length(.w) > 0)

  .w <- which(grepl("double __udf[2]", lines, fixed=TRUE))
  expect_true(length(.w) > 0)


  e <- et(1:10) |> as.data.frame()

  e$x <- 1:10
  e$y <- 21:30


  gg <- function(x, y) {
    x + y
  }

  f <- rxode2({
    z = gg(x, y)
  })


  test_that("udf1 works well", {
    expect_warning(rxSolve(f, e))

    d <- suppressWarnings(rxSolve(f, e))

    expect_true(all(d$z == d$x + d$y))
  })


  # now modify gg
  gg <- function(x, y, z) {
    x + y + z
  }

  test_that("udf with 3 arguments works", {
    expect_error(rxSolve(f, e))
  })

  # now modify gg back to 2 arguments
  gg <- function(x, y) {
    x * y
  }

  test_that("when changing gg the results will be different", {
    # different solve results but still runs

    d <- suppressWarnings(rxSolve(f, e))

    expect_true(all(d$z == d$x * d$y))

  })

  rm(gg)

  test_that("Without a udf, the solve errors", {
    expect_error(rxSolve(f, e))
  })


  gg <- function(x, ...) {
    x
  }

  test_that("cannot solve with udf functions that have ...", {
    expect_error(rxSolve(f, e))
  })

  gg <- function(x, y) {
    stop("running me")
  }

  test_that("functions that error will error the solve",{
    expect_error(rxSolve(f, e))
  })

  gg <- function(x, y) {
    "running "
  }


  test_that("runs with improper output will error", {
    expect_error(rxSolve(f, e))
  })

  gg <- function(x, y) {
    "3"
  }

  test_that("error for invalid input", {
    expect_error(rxSolve(f, e))
  })

  gg <- function(x, y) {
    3L
  }

  test_that("test symengine functions work with udf funs", {

    expect_equal(rxToSE("gg(x,y)"), "gg(x, y)")

    expect_error(rxToSE("gg()"), "user function")

    expect_error(rxFromSE("Derivative(gg(a,b),a)"), NA)

    expect_error(rxFromSE("Derivative(gg(a),a)"))

    expect_error(rxFromSE("Derivative(gg(),a)"))
  })

  gg <- function(x, ...) {
    x
  }

  test_that("test that functions with ... will error symengine translation", {
    expect_error(rxToSE("gg(x,y)"))

    expect_error(rxFromSE("Derivative(gg(a,b),a)"))
  })

  ## manual functions in C vs R functions

  gg <- function(x, y) {
    x + y
  }

  test_that("R vs C functions", {
    d <- suppressWarnings(rxSolve(f, e))
    expect_true(all(d$z == d$x + d$y))
  })

  # now add a C function with different values
  rxFun("gg", c("x", "y"),
        "double gg(double x, double y) { return x*y;}")


  test_that("C functions rule", {


    d <- suppressWarnings(rxSolve(f, e))

    expect_true(all(d$z == d$x * d$y))

  })

  rxRmFun("gg")

  test_that("c conversion", {

    udf <- function(x, y) {
      a <- x + y
      b <- a ^ 2
      a + b
    }

    expect_true(grepl("R_pow_di[(]", rxFun2c(udf)[[1]]$cCode))

    udf <- function(x, y) {
      a <- x + y
      b <- a ^ x
      a + b
    }

    expect_true(grepl("R_pow[(]", rxFun2c(udf)[[1]]$cCode))

    udf <- function(x, y) {
      a <- x + y
      b <- cos(a) + x
      a + b
    }

    expect_true(grepl("cos[(]", rxFun2c(udf)[[1]]$cCode))

    udf <- function(x, y) {
      if (a < b) {
        return(b ^ 2)
      }
      a + b
    }

    expect_true(grepl("if [(]", rxFun2c(udf)[[1]]$cCode))


    udf <- function(x, y) {
      a <- x
      b <- x ^ 2 + a
      if (a < b) {
        return(b ^ 2)
      } else {
        a + b
      }
    }

    expect_true(grepl("else [{]", rxFun2c(udf)[[1]]$cCode))

    udf <- function(x, y) {
      a <- x
      b <- x ^ 2 + a
      if (a < b) {
        return(b ^ 2)
      } else if (a > b + 3) {
        return(a + b)
      }
      a ^ 2 + b ^ 2
    }

    expect_true(grepl("else if [(]", rxFun2c(udf)[[1]]$cCode))

    udf <- function(x, y) {
      a <- x
      b <- x ^ 2 + a
      if (a < b) {
        return(b ^ 2)
      } else if (a > b + 3) {
        b <- 3
        if (a > 2) {
          a <- 2
        }
        return(a + b)
      }
      a ^ 2 + b ^ 2
    }

    expect_true(grepl("else if [(]", rxFun2c(udf)[[1]]$cCode))

    udf <- function(x, y) {
      a <- x + y
      x <- a ^ 2
      x
    }

    expect_error(rxFun2c(udf)[[1]]$cCode)


    udf <- function(x, y) {
      a <- x
      b <- x ^ 2 + a
      if (a < b) {
        b ^ 2
      } else {
        a + b
      }
    }

    rxFun(udf)
    rxRmFun("udf")

  })

  test_that("udf with model functions", {

    gg <- function(x, y) {
      x/y
    }

    # Step 1 - Create a model specification
    f <- function() {
      ini({
        KA <- .291
        CL <- 18.6
        V2 <- 40.2
        Q <- 10.5
        V3 <- 297.0
        Kin <- 1.0
        Kout <- 1.0
        EC50 <- 200.0
      })
      model({
        # A 4-compartment model, 3 PK and a PD (effect) compartment
        # (notice state variable names 'depot', 'centr', 'peri', 'eff')
        C2 <- gg(centr, V2)
        C3 <- peri/V3
        d/dt(depot) <- -KA*depot
        d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
        d/dt(peri)  <-                    Q*C2 - Q*C3
        d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
        eff(0) <- 1
      })
    }

    u <- f()

    # this pre-compiles and displays the simulation model
    u$simulationModel

    # Step 2 - Create the model input as an EventTable,
    # including dosing and observation (sampling) events

    # QD (once daily) dosing for 5 days.

    qd <- eventTable(amount.units = "ug", time.units = "hours")
    qd$add.dosing(dose = 10000, nbr.doses = 5, dosing.interval = 24)

    # Sample the system hourly during the first day, every 8 hours
    # then after

    qd$add.sampling(0:24)
    qd$add.sampling(seq(from = 24 + 8, to = 5 * 24, by = 8))

    # Step 3 - set starting parameter estimates and initial
    # values of the state

    # Step 4 - Fit the model to the data
    expect_error(suppressWarnings(solve(u, qd)), NA)

    u1 <- u$simulationModel

    expect_error(suppressWarnings(solve(u1, qd)), NA)

    u2 <- u$simulationIniModel
    expect_error(suppressWarnings(solve(u2, qd)), NA)

    expect_error(suppressWarnings(rxSolve(f, qd)), NA)
  })

  test_that("symengine load", {

    mod <- "tke=THETA[1];\nprop.sd=THETA[2];\neta.ke=ETA[1];\nke=gg(tke,exp(eta.ke));\nipre=gg(10,exp(-ke*t));\nlipre=log(ipre);\nrx_yj_~2;\nrx_lambda_~1;\nrx_low_~0;\nrx_hi_~1;\nrx_pred_f_~ipre;\nrx_pred_~rx_pred_f_;\nrx_r_~(rx_pred_f_*prop.sd)^2;\n"

    gg <- function(x, y) {
      x * y
    }

    expect_error(rxS(mod, TRUE, TRUE), NA)

    rxFun(gg)

    rm(gg)

    expect_error(rxS(mod, TRUE, TRUE), NA)

    rxRmFun("gg")


  })

})


rxTest({
  test_that("udf type 2 (that changes ui models upon parsing)", {

    expect_error(rxModelVars("a <- linMod(x, 3)"), NA)
    expect_error(rxModelVars("a <- linMod(x, 3, b)"))
    expect_error(rxModelVars("a <- linMod(x)"))
    expect_error(rxModelVars("a <- linMod()"))

    f <- rxode2({
      a <- linMod(x, 3)
    })

    e <- et(1:10)

    expect_error(rxSolve(f, e, c(x=2)), "ui user function")

    # Test a linear model construction

    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linMod(time, 3)
        b <-  d
      })
    }

    tmp <- f()

    expect_equal(tmp$iniDf$name,
                 c("d", "rx.linMod.time1a", "rx.linMod.time1b", "rx.linMod.time1c",
                   "rx.linMod.time1d"))

    expect_equal(modelExtract(tmp, a),
                 "a <- (rx.linMod.time1a + rx.linMod.time1b * time + rx.linMod.time1c * time^2 + rx.linMod.time1d * time^3)")

    # Test a linear model construction without an intercept
    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linMod0(time, 3) + d
      })
    }

    tmp <- f()

    expect_equal(tmp$iniDf$name,
                 c("d", "rx.linMod.time1a", "rx.linMod.time1b", "rx.linMod.time1c"))


    expect_equal(modelExtract(tmp, a),
                 "a <- (rx.linMod.time1a * time + rx.linMod.time1b * time^2 + rx.linMod.time1c * time^3) + d")

    # Now test the use of 2 linear models in the UI
    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linMod(time, 3)
        b <- linMod(time, 3)
        c <- d
      })
    }

    tmp <- f()

    expect_equal(tmp$iniDf$name,
                 c("d", "rx.linMod.time1a", "rx.linMod.time1b", "rx.linMod.time1c", "rx.linMod.time1d",
                   "rx.linMod.time2a", "rx.linMod.time2b", "rx.linMod.time2c", "rx.linMod.time2d"))

    expect_equal(modelExtract(tmp, a),
                 "a <- (rx.linMod.time1a + rx.linMod.time1b * time + rx.linMod.time1c * time^2 + rx.linMod.time1d * time^3)")

    expect_equal(modelExtract(tmp, b),
                 "b <- (rx.linMod.time2a + rx.linMod.time2b * time + rx.linMod.time2c * time^2 + rx.linMod.time2d * time^3)")


    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linModB(time, 3)
        b <-  d
      })
    }

    tmp <- f()

    expect_equal(modelExtract(tmp, rx.linMod.time.f1),
                 "rx.linMod.time.f1 <- rx.linMod.time1a + rx.linMod.time1b * time + rx.linMod.time1c * time^2 + rx.linMod.time1d * time^3")

    expect_equal(modelExtract(tmp, a),
                 "a <- rx.linMod.time.f1")

    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linModB0(time, 3) + d
      })
    }

    tmp <- f()

    expect_equal(modelExtract(tmp, rx.linMod.time.f1),
                 "rx.linMod.time.f1 <- rx.linMod.time1a * time + rx.linMod.time1b * time^2 + rx.linMod.time1c * time^3")

    expect_equal(modelExtract(tmp, a),
                 "a <- rx.linMod.time.f1 + d")

    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linModA(time, 1) + d
      })
    }

    tmp <- f()

    expect_equal(modelExtract(tmp, rx.linMod.time.f1),
                 "rx.linMod.time.f1 <- rx.linMod.time1a + rx.linMod.time1b * time")

    expect_equal(modelExtract(tmp, a),
                 "a <- 0 + d")

    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linModA0(time, 1) + d
      })
    }

    tmp <- f()

    expect_equal(modelExtract(tmp, rx.linMod.time.f1),
                 "rx.linMod.time.f1 <- rx.linMod.time1a * time")

    expect_equal(modelExtract(tmp, a),
                 "a <- 0 + d")

    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linMod(power=3, variable="x") + d
      })
    }

    tmp <- f()

    expect_equal(modelExtract(tmp, a),
                 "a <- (rx.linMod.x1a + rx.linMod.x1b * x + rx.linMod.x1c * x^2 + rx.linMod.x1d * x^3) + d")

    expect_false(tmp$uiUseData)


    ## Formula interface
    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linMod0(dv~x^3) + d
      })
    }

    tmp <- f()

    expect_equal(modelExtract(tmp, a),
                 "a <- linModD0(x, 3, dv) + d")
    expect_true(tmp$uiUseData)

    ## Formula interface
    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linMod0(~x^3) + d
      })
    }

    tmp <- f()

    expect_equal(modelExtract(tmp, a),
                 "a <- (rx.linMod.x1a * x + rx.linMod.x1b * x^2 + rx.linMod.x1c * x^3) + d")

    ## Formula interface
    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linMod0(~x^6) + d
      })
    }

    tmp <- f()

    expect_equal(modelExtract(tmp, a),
                 "a <- (rx.linMod.x1a * x + rx.linMod.x1b * x^2 + rx.linMod.x1c * x^3 + rx.linMod.x1d * x^4 + rx.linMod.x1e * x^5 + rx.linMod.x1f * x^6) + d")


    # This checks to make sure that the variables are not in the model
    # before adding them
    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linModM0(~x^6) + d
      })
    }

    tmp <- f()

    expect_equal(modelExtract(tmp, a),
                 "a <- (x1a * x + x1b * x^2 + x1c * x^3 + x1d * x^4 + x1e * x^5 + x1f * x^6) + d")


    f <- function() {
      ini({
        d <- 4
      })
      model({
        a <- linModM(~x^6) + d
      })
    }

    tmp <- f()

    expect_equal(modelExtract(tmp, a),
                 "a <- (x1a + x1b * x + x1c * x^2 + x1d * x^3 + x1e * x^4 + x1f * x^5 + x1g * x^6) + d")

    rxWithSeed(42, {

      q <- seq(from=0, to=20, by=0.1)

      y <- 500 + 42*q^2 + 0.4 * (q-10)^3

      df <- data.frame(q=q, y=y)

      f <- function() {
        model({
          a <- linMod(y~q^3)
        })
      }

      f <- f()

      expect_equal(modelExtract(f, a),
                   "a <- linModD(q, 3, y)")

      rxUdfUiData(df)

      try({
        if (f$uiUseData) {
          f <- rxode2(as.function(f))
          expect_false(any(f$theta == 0))
        }
      })
      rxUdfUiData(NULL)
    })

  })
})

#' Fork a model's meta environment for piping
#'
#' Shallow copy: only the meta environment itself is new, so the piped and
#' original model no longer share the `.simModelBase` cache slot (which is
#' dropped here, as in .copyEnv/.copyUi).  Bindings are carried over as-is --
#' nested environments stay shared and closures keep their environment -- so
#' reference-semantics metadata behaves exactly as it did when meta was shared.
#'
#' @param env meta environment to fork
#'
#' @return New meta environment
#' @author Matthew L. Fidler
#' @noRd
.copyMetaForPiping <- function(env) {
  .ret <- new.env(parent=parent.env(env))
  lapply(ls(envir=env, all.names=TRUE), function(item) {
    if (item == ".simModelBase") {
      return(NULL)
    }
    assign(item, get(item, envir=env), envir=.ret)
  })
  .ret
}

#' Copy an environment and all of its contents
#'
#' This is a recursive copy that creates a new environment for every
#' environment in the original environment.  This is used to copy the
#' rxUi object so that it can be modified without modifying the
#' original.
#'
#' @param env Environment to copy
#'
#' @return Copied environment
#' @author Matthew L. Fidler
#' @noRd
.copyEnv <- function(env, visited=new.env(hash=TRUE, parent=emptyenv())) {
  .addr <- environmentName(env)
  if (identical(.addr, "") || is.na(.addr)) .addr <- format(env)
  if (exists(.addr, envir=visited, inherits=FALSE)) {
    return(get(.addr, envir=visited, inherits=FALSE))
  }
  .ret <- new.env(parent=emptyenv())
  assign(.addr, .ret, envir=visited)
  lapply(ls(envir=env, all.names=TRUE), function(item) {
    if (item == ".simModelBase") {
      return(NULL)
    }
    if (is.environment(get(item, envir=env))) {
      assign(item, .copyEnv(get(item, envir=env), visited), envir=.ret)
    } else {
      assign(item, get(item, envir=env), envir=.ret)
    }
  })
  .ret
}

#' This copies the rxode2 UI object so it can be modified
#'
#' @param ui Original UI object
#' @return Copied UI object
#' @author Matthew L. Fidler
#' @export
.copyUi <- function(ui) {
  if (is.list(ui) ||
        inherits(ui, "raw")) {
    return(rxUiDecompress(ui))
  }
  .ret <- new.env(parent=emptyenv())
  .visited <- new.env(hash=TRUE, parent=emptyenv())
  lapply(ls(envir=ui, all.names=TRUE), function(item) {
    if (item == ".simModelBase") {
      return(NULL)
    }
    if (is.environment(get(item, envir=ui))) {
      assign(item, .copyEnv(get(item, envir=ui), .visited), envir=.ret)
    } else {
      assign(item, get(item, envir=ui), envir=.ret)
    }
  })
  class(.ret) <- class(ui)
  .ret
}

#' Expand the quoted lines to include relevant lines from UI
#'
#' @param cur This is the current piped in `rxUi` interface
#'
#' @param iniDf This is the ini data frame from the prior ui
#'
#' @param charExpression character vector of the current expression
#'
#' @return `NULL` if there is no lines to add OR a list of the lines
#'   from the `cur` UI removing parameters that are not in the
#'   destination `ini()`
#'
#' @author Matthew L Fidler
#' @noRd
.quoteExpandRxUi <- function(cur, iniDf, charExpression) {
  if (is.null(iniDf)) return(NULL)
  .cur <- try(as.rxUi(cur), silent=TRUE)
  if (!inherits(.cur, "rxUi")) return(NULL)
  .curLotri <- lotri::as.lotri(.cur$iniDf)
  .lotriEst <- lotri::lotriEst(.curLotri)
  attr(.curLotri, "lotriEst") <- NULL
  .ini1 <- NULL
  if (!is.null(.lotriEst)) {
    .w <- which(.lotriEst$name %in% iniDf$name)
    .drop <- NULL
    if (length(.w) == 0L) {
     .drop <- .lotriEst$name
    } else {
      .drop <- .lotriEst$name[-.w]
      .lotriEst <- .lotriEst[.w, ]
      .ret <- list()
      attr(.ret, "lotriEst") <- .lotriEst
      class(.ret) <- "lotriFix"
      .ini1 <- as.data.frame(.ret)
    }
  }
  .dn <- dimnames(.curLotri)
  .ini2 <- NULL
  if (!is.null(.dn)) {
    .dn <- .dn[[1]]
    .w <- which(.dn %in% iniDf$name)
    if (length(.w) == 0L) {
      .drop <- c(.drop, .dn)
    } else  {
      .drop <- c(.drop, .dn[-.w])
      .curLotri <- .curLotri[.w, .w]
      class(.curLotri) <- c("lotriFix", "matrix", "array")
      .ini2 <- as.data.frame(.curLotri)
    }
  }
  .iniDf <- rbind(.ini1, .ini2)
  if (is.null(.iniDf)) {
    cli::cli_alert_info(paste0("piping '", charExpression, "' has no parameters in common with model and does nothing"))
    return(list())
  }
  if (length(.drop) > 0) {
    cli::cli_alert_info(paste0("ignoring following estimates in '", charExpression, "': ",
                               paste(.drop, collapse=", ")))
  }
  .ini <- lotri::lotriDataFrameToLotriExpression(.iniDf, useIni = TRUE)
  .ini <- .ini[[2]]
  .ini <- as.list(.ini)[-1]
  .env <- new.env(parent=emptyenv())
  .env$labels <- NULL
  .env$lastlhs <- NULL
  .env$lhsvars <- NULL
  .ini <- lapply(seq_along(.ini), function(i) {
    .cur <- .ini[[i]]
    if (is.call(.cur) && identical(.cur[[1]], quote(`label`))) {
      if (is.null(.env$lastlhs)) {
        stop("do not know where to put label", call.=FALSE) # should not get here
      }
      .env$labels <- c(.env$labels, i)
      return(as.call(list(quote(`<-`), str2lang(.env$lastlhs), .cur)))
    }
    if (is.call(.cur) && (identical(.cur[[1]], quote(`<-`)) ||
                            identical(.cur[[1]], quote(`~`)))) {
      if (is.call(.cur[[2]])) {
        if (identical(.cur[[2]], quote(`+`))) {
          .env$lhsvars <- c(.env$lhsvars, vapply(as.list(.cur)[-1], function(x) {
            as.character(x)
          }, character(1), USE.NAMES=FALSE))
        }
      } else {
        .char <- as.character(.cur[[2]])
        .env$lastlhs <- .char
        .env$lhsvars <- c(.env$lhsvars, .char)
      }
    }
    .cur
  })
  if (!is.null(.env$labels)) {
    if (getOption("rxode2.ignoreLabels", TRUE)) {
      .ini <- .ini[-.env$labels]
      .minfo("the labels from the piped model do not overwrite old labels\nto change use 'options(rxode2.ignoreLabels=FALSE)'")
    } else {
      .minfo("the labels from the piped model overwrite old labels\nto change use 'options(rxode2.ignoreLabels=TRUE)'")
    }
  }
  return(.ini)
}

#' This expands a list of expressions
#'
#' @param lines These are the expressions as a list
#'
#' @param bracketsOrCs This is the indicator of the bracket lines ie
#'   `{}` or concatenations ie `c()`, that are expanded
#'
#' @param iniDf initial conditions from the previous/parent rxUi
#'
#' @return Single list of expressions; `a=b` becomes `a<-b` in this
#'   expression
#'
#' @author Matthew L. Fidler
#' @noRd
.quoteExpandBracketsOrCs <- function(lines, bracketsOrCs, envir=envir, iniDf=NULL) {
  if (length(bracketsOrCs) == 0) return(lines)
  .expandedForm <- NULL
  .currentLine <- 1
  for (.b in bracketsOrCs) {
    .bracketExpression <- lines[[.b]]
    .cur <- NULL
    .curEta <- FALSE
    if (length(.bracketExpression) == 1) {
      # evaulate expression
      .cur <- try(eval(.bracketExpression, envir=envir), silent=TRUE)
      if (inherits(.cur, "try-error")) {
      } else if (length(.cur) > 1) {
        if (inherits(.cur, "character")) {
          if (is.null(names(.cur))) {
            .cur <- lapply(.cur, function(x) {
              str2lang(x)
            })
          } else {
            .cur <- lapply(names(.cur), function(x) {
              str2lang(paste0(x, "<-", .cur[[x]]))
            })
          }
          .cur <- as.call(c(list(quote(`{`)),.cur))
          .bracketExpression <- .cur
        } else if (identical(.cur[[1]], quote(`{`))) {
          .bracketExpression <- .cur
        }
      }
    } else if (length(.bracketExpression) == 2 &&
                 identical(.bracketExpression[[1]], quote(`~`))) {
      # evaulate expression
      .cur <- try(eval(.bracketExpression[[2]], envir=envir), silent=TRUE)
      .curEta <- TRUE
      if (inherits(.cur, "try-error")) {
      } else if (length(.cur) > 1) {
        if (inherits(.cur, "character")) {
          if (is.null(names(.cur))) {
            .cur <- lapply(.cur, function(x) {
              str2lang(x)
            })
          } else {
            .cur <- lapply(names(.cur), function(x) {
              str2lang(paste0(x, "~", .cur[[x]]))
            })
          }
          .cur <- as.call(c(list(quote(`{`)),.cur))
          .bracketExpression <- .cur
        } else if (identical(.cur[[1]], quote(`{`))) {
          .bracketExpression <- .cur
        }
      }
    }
    .unlistedBrackets <- NULL
    if (length(.bracketExpression) > 1) {
      if (identical(.bracketExpression[[1]], quote(`{`))) {
        .unlistedBrackets <- lapply(seq_along(.bracketExpression)[-1],
                                    function(i) {
                                      .c <- .bracketExpression[[i]]
                                      if (identical(.c[[1]], quote(`=`))) {
                                        .c[[1]] <- quote(`<-`)
                                      }
                                      .c
                                    })
      }
    }
    if (is.null(.unlistedBrackets)) {
      if (is.null(.cur)) {
        ## evaluate to vector and then put it in place
        .cur <- eval(.bracketExpression, envir=envir)
      }
      if (inherits(.cur, "<-") || inherits(.cur, "call")) {
        .unlistedBrackets <- .cur
      } else if (inherits(.cur, "numeric")) {
        if (is.null(names(.cur))) {
          stop("cannot figure out what to do with the unnamed vector", call.=FALSE)
        }
        .unlistedBrackets <- lapply(names(.cur), function(.n) {
          if (.curEta) {
            bquote(.(str2lang(.n)) ~ .(setNames(.cur[.n], NULL)))
          } else {
            bquote(.(str2lang(.n)) <- .(setNames(.cur[.n], NULL)))
          }
        })
      } else if (!inherits(.cur, "rxUi") &&
                   inherits(.cur, "list")) {
        if (is.null(names(.cur))) {
          stop("cannot figure out what to do with the unnamed list", call.=FALSE)
        }
        .unlistedBrackets <- lapply(names(.cur), function(.n) {
          .v <- .cur[[.n]]
          if (inherits(.v, "numeric")) {
            if (.curEta) {
              bquote(.(str2lang(.n)) ~ .(setNames(.cur[[.n]], NULL)))
            } else {
              bquote(.(str2lang(.n)) <- .(setNames(.cur[[.n]], NULL)))
            }
          } else {
            stop("one of the list items supplied to piping is non-numeric", call.=FALSE)
          }
        })
      } else if (inherits(.cur, "matrix")) {
        .cur2 <- .cur
        .unlistedBrackets <- as.list(lotri::lotriAsExpression(.cur2, plusNames=TRUE)[[-1]])[-1]
      } else if (inherits(.cur, "character") && !is.null(names(.cur))) {
        .unlistedBrackets <- lapply(paste(names(.cur),"=", setNames(.cur, NULL)),
                                    str2lang)
      } else if (inherits(.cur, "character") && length(.cur) == 1) {
        .unlistedBrackets <- try(str2lang(.cur), silent=TRUE)
        if (inherits(.unlistedBrackets, "try-error")) {
          stop("vectors and list need to be named numeric expression", call.=FALSE)
        }
        if (identical(.unlistedBrackets[[1]], quote(`=`))) {
          .unlistedBrackets[[1]] <- quote(`<-`)
        }
        .unlistedBrackets <- list(.unlistedBrackets)
      } else {
        .ini <- .quoteExpandRxUi(.cur, iniDf=iniDf, charExpression=deparse1(.bracketExpression))
        if (is.null(.ini)) stop("vectors and list need to be named numeric expression", call.=FALSE)
        .expandedForm <- c(.expandedForm, .ini)
      }
    }
    if (.currentLine == .b) {
      .expandedForm <- c(.expandedForm, .unlistedBrackets)
    } else {
      .expandedForm <- c(.expandedForm, lines[seq(.currentLine, .b - 1)],
                         .unlistedBrackets)
    }
    .currentLine <- .b + 1
  }
  if (.currentLine <= length(lines)) {
    .expandedForm <- c(.expandedForm, lines[seq(.currentLine, length(lines))])
  }
  .expandedForm
}

#' This function collapses the lotri line form to the plus form
#'
#' @param expressionList Expression list that is input to change into
#'   matrix expression form the new line expressions to the classic
#'   plus expressions.
#' @return expression list where lotri line for covariance matrices
#'   are translated to classic plus form.
#' @author Matthew L. Fidler
#' @noRd
#' @examples
#'
#' tmp <- list(str2lang("d ~ 1"),
#'             str2lang("e ~ c(0.5, 3)"),
#'             str2lang("cp ~ add(add.sd)"),
#'             str2lang("cp ~ add(add.sd) + prop(prop.sd)"),
#'             str2lang("cp ~ + add(add.sd)"))
#'
#' .collapseLotriLineFormToPlusForm(tmp)
#'
.collapseLotriLineFormToPlusForm <- function(expressionList) {
  .env <- new.env(parent=emptyenv())
  .env$ret <- expressionList
  .env$lst <- list()
  .env$last <- NA_integer_

  .f <- function() {
    if (!is.na(.env$last)) {
      .val <- as.call(c(list(quote(`{`)), .env$lst))
      .val <- as.call(c(str2lang("lotri::lotri"), .val))
      .val <- suppressMessages(try(eval(.val), silent=TRUE))
      if (inherits(.val, "try-error")) {
        for (.j in seq_along(.env$lst)) {
          .env$ret[[.env$last + .j - 1L]] <- .env$lst[[.j]]
        }
      } else {
        .val <- lotri::lotriAsExpression(.val, plusNames=TRUE)
        .val <- lapply(seq_along(.val)[-1],
                       function(i){
                         .val[[i]]
                       })[[1]]
        .val <- lapply(seq_along(.val)[-1],
                       function(i){
                         .val[[i]]
                       })
        for (.j in seq_along(.val)) {
          .env$ret[[.env$last + .j - 1L]] <- .val[[.j]]
        }
      }
      .env$lst <- list()
      .env$last <- NA_integer_
    }
  }
  for (.i in seq_along(.env$ret)) {
    .cur <- .env$ret[[.i]]
    if (is.call(.cur) && identical(.cur[[1]], quote(`~`)) &&
          length(.cur) == 3L &&
          length(.cur[[2]]) == 1L # excludes ll(cp) ~ 1
        ) {
      .isLotri <- TRUE
      # Check to see if this is an error call
      if (is.call(.cur[[3]])) {
        .call <- deparse1(.cur[[3]][[1]])
        if (.call == "+" &&
              length(.cur[[3]]) >= 2 &&
              is.call(.cur[[3]][[2]])) {
          .call <- deparse1(.cur[[3]][[2]][[1]])
        }
        if (.call %in% names(.errDist)) {
          .isLotri <- FALSE
        }
      }
      if (.isLotri) {
        if (is.na(.env$last)) {
          .env$last <- .i
        }
        .env$ret[[.i]] <- NA
        .env$lst <- c(.env$lst, .cur)
      }
    } else {
      .f()
    }
  }
  .f()
  .w <- which(vapply(seq_along(.env$ret), function(i) {
    !(length(.env$ret[[i]]) == 1L && is.na(.env$ret[[i]]))
  }, logical(1), USE.NAMES=FALSE))
  lapply(.w, function(i) { .env$ret[[i]]})
}

.nsEnv <- new.env(parent=emptyenv())

.nsEnv$.quoteCallInfoLinesAppend <- NULL
#' Returns quoted call information
#'
#' @param callInfo Call information
#'
#' @param envir Environment for evaluation (if needed)
#'
#' @param iniDf The parent model `iniDf` when piping in a `ini` block
#'   (`NULL` otherwise)
#'
#' @return Quote call information.  for `name=expression`, change to
#'   `name<-expression` in quoted call list. For expressions that are
#'   within brackets ie `{}`, unlist the brackets as if they were
#'   called in one single sequence.
#'
#' @author Matthew L. Fidler
#'
#' @export
.quoteCallInfoLines <- function(callInfo, envir=parent.frame(), iniDf=NULL) {
  .bracket <- rep(FALSE, length.out=length(callInfo))
  .env <- environment()
  .nsEnv$.quoteCallInfoLinesAppend <- NULL
  .ret <- lapply(seq_along(callInfo), function(i) {
    .name <- names(callInfo)[i]
    if (!is.null(.name)) {
      if (.name == "append") {
        .append <- callInfo[[i]]
        if (identical(.append, quote(TRUE)) ||
              identical(.append, quote(FALSE)) ||
              identical(.append, quote(NA))) {
        } else {
          .nsEnv$.quoteCallInfoLinesAppend <- eval(call("quote", .append))
        }
        return(NULL)
      } else if (.name %in% c("envir",  "auto", "iniDf", "cov")) {
        return(NULL)
      } else if (.name != "") {
        # Changed named items to
        return(as.call(list(quote(`<-`), str2lang(.name),
                            eval(call("quote", callInfo[[i]])))))
      }
    }
    .quoted <- eval(call("quote", callInfo[[i]]))
    if (missing(.quoted)) {
      # Capture empty arguments (rxode2#688)
      warning("empty argument ignored")
      return(NULL)
    } else if (length(.quoted) == 1) {
      if (identical(.quoted, quote(`diag`)) ||
            (is.call(.quoted) && identical(.quoted[[1]], quote(`diag`)))) {
        .quoted <- str2lang("~diag()")
      } else {
        .bracket[i] <- TRUE
        assign(".bracket", .bracket, envir=.env)
      }
    } else if (length(.quoted) >= 1 &&
                 identical(.quoted[[1]], quote(`diag`))) {
      .quoted <- as.call(c(list(quote(`~`)), .quoted))
    } else if (identical(.quoted[[1]], quote(`diag`))) {
    } else if (identical(.quoted[[1]], quote(`{`)) ||
          identical(.quoted[[1]], quote(`c`)) ||
          identical(.quoted[[1]], quote(`list`))) {
      .bracket[i] <- TRUE
      assign(".bracket", .bracket, envir=.env)
    } else if (identical(.quoted[[1]], quote(`as.formula`))) {
      .quoted <- .quoted[[2]]
    } else if (identical(.quoted[[1]], quote(`~`))) {
      if (length(.quoted) == 2 &&
            exists(as.character(.quoted[[2]]), envir)) {
        # this is a "bracketed" assignment of etas
        .bracket[i] <- TRUE
        assign(".bracket", .bracket, envir=.env)
      } else {
        if (length(.quoted) == 3L && !is.null(.quoted[[3]])) {
          .quoted[[3]] <- .iniSimplifyFixUnfix(.quoted[[3]])
          if (identical(.quoted[[3]], quote(`fix`)) ||
                identical(.quoted[[3]], quote(`unfix`))) {
            .quoted <- as.call(list(quote(`<-`), .quoted[[2]], .quoted[[3]]))
          }
        }

      }
    } else if (identical(.quoted[[1]], quote(`$`))) {
      .tmp <- try(eval(.quoted), silent=TRUE)
      if (!inherits(.tmp, "try-error")) {
        .quoted <- .tmp
        if (inherits(.quoted, "character")) {
          .quoted <- str2lang(.quoted)
        }
      }
    }
    .quoted
  })
  .w <- which(.bracket)
  .ret <- .quoteExpandBracketsOrCs(.ret, .w, envir=envir, iniDf=iniDf)
  .ret <- lapply(seq_along(.ret), function(i) {
    if (identical(.ret[[i]][[1]], quote(`$`))) {
      .r <- eval(.ret[[i]], envir=envir)
      if (inherits(.r, "character")) {
        .r <- str2lang(.r)
      }
      return(.r)
    }
    .ret[[i]]
  })
  .collapseLotriLineFormToPlusForm(.ret[vapply(seq_along(.ret), function(i) {
    !is.null(.ret[[i]])
  }, logical(1), USE.NAMES=FALSE)])
}

rxode2.logo <- "  _____         ____  _____  ______
 |  __ \\       / __ \\|  __ \\|  ____| %s
 | |__) |__  _| |  | | |  | | |__
 |  _  / \\ \\/ / |  | | |  | |  __|
 | | \\ \\  >  <| |__| | |__| | |____
 |_|  \\_\\/_/\\_\\\\____/|_____/|______|
"

.rxVersion <- c(version = "calc", repo = "https://github.com/nlmixr2/rxode2", md5 = rxode2.md5)

#' Version and repository for this dparser package.
#'
#' @param echo Boolean to echo the text logo, by default FALSE
#' @param extra extra text to display in the logo
#' @return A character vector with the version and repository.
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
rxVersion <- function(extra = "", echo = FALSE) {
  if (echo) {
    message(sprintf(rxode2.logo, sprintf("%s%s", version, extra)), appendLF = FALSE)
  }
  return(invisible(.rxVersion))
}

#' Log likelihood for normal distribution
#'
#' @param x Observation
#' @param mean Mean for the likelihood
#' @param sd Standard deviation for the likelihood
#' @param full Add the data frame showing x, mean, sd as well as the
#'   fx and derivatives
#'
#' @details
#'
#' In an `rxode2()` model, you can use `llikNorm()` but you have to
#' use all arguments.  You can also get the derivatives with
#' `llikNormDmean()` and `llikNormDsd()`
#'
#' @return data frame with `fx` for the pdf value of with `dMean` and
#'   `dSd` that has the derivatives with respect to the parameters at
#'   the observation time-point
#' @author Matthew L. Fidler
#' @export
#' @importFrom Rcpp sourceCpp
#' @examples
#' 
#' \donttest{
#' 
#' llikNorm(0)
#'
#' llikNorm(seq(-2,2,length.out=10), full=TRUE)
#' 
#' # With rxode2 you can use:
#'  
#' et <- et(-3, 3, length.out=10)
#' et$mu <- 0
#' et$sigma <- 1
#' 
#' model <- function(){
#'   model({
#'     fx <- llikNorm(time, mu, sigma)
#'     dMean <- llikNormDmean(time, mu, sigma)
#'     dSd <- llikNormDsd(time, mu, sigma)
#'   })
#'  }
#'
#' ret <- rxSolve(model, et)
#' ret
#' }
llikNorm <- function(x, mean = 0, sd = 1, full=FALSE) {
  rxode2ll::llikNorm(x, mean, sd, full)
}
#' log-likelihood for the Poisson distribution
#'
#' @param x non negative integers
#' @param lambda non-negative means
#' @inheritParams llikNorm
#' @details
#'
#' In an `rxode2()` model, you can use `llikPois()` but you have to
#' use all arguments.  You can also get the derivatives with
#' `llikPoisDlambda()`
#'
#' @return data frame with `fx` for the pdf value of with
#'   `dLambda` that has the derivatives with respect to the parameters at
#'   the observation time-point
#' @author Matthew L. Fidler
#' @export
#' @examples
#' \donttest{
#' llikPois(0:7, lambda = 1)
#'
#' llikPois(0:7, lambda = 4, full=TRUE)
#'
#' # In rxode2 you can use:
#'
#' et <- et(0:10)
#' et$lambda <- 0.5
#' 
#' model <- function() {
#'   model({
#'     fx <- llikPois(time, lambda)
#'     dLambda <- llikPoisDlambda(time, lambda)
#'   })
#' }
#'
#' rxSolve(model, et)
#' }
llikPois <- function(x, lambda, full=FALSE) {
  rxode2ll::llikPois(x, lambda, full)
}

#' Calculate the log likelihood of the binomial function (and its derivatives)
#' 
#' @param x  Number of successes
#' @param size Size of trial
#' @param prob probability of success
#' 
#' @inheritParams llikNorm
#'
#' @details
#' 
#' In an `rxode2()` model, you can use `llikBinom()` but you have to
#' use all arguments.  You can also get the derivative of `prob` with
#' `llikBinomDprob()`
#' 
#' @return data frame with `fx` for the pdf value of with
#'   `dProb` that has the derivatives with respect to the parameters at
#'   the observation time-point
#' @author Matthew L. Fidler
#' @export 
#' @examples
#' \donttest{
#' llikBinom(46:54, 100, 0.5)
#'
#' llikBinom(46:54, 100, 0.5, TRUE)
#'
#' # In rxode2 you can use:
#' 
#' et <- et(46:54)
#' et$size <- 100
#' et$prob <-0.5
#'
#' model <- function() {
#'   model({
#'     fx <- llikBinom(time, size, prob)
#'     dProb <- llikBinomDprob(time, size, prob)
#'   })
#' }
#'
#' rxSolve(model, et)
#' }
llikBinom <- function(x, size, prob, full=FALSE) {
  rxode2ll::llikBinom(x, size, prob, full)
}

#' Calculate the log likelihood of the negative binomial function (and its derivatives)
#' 
#' @param x  Number of successes
#' @param size Size of trial
#' @param prob probability of success
#' 
#' @inheritParams llikNorm
#'
#' @details
#' In an `rxode2()` model, you can use `llikNbinom()` but you have to
#' use all arguments.  You can also get the derivative of `prob` with
#' `llikNbinomDprob()`
#' @return data frame with `fx` for the pdf value of with
#'   `dProb` that has the derivatives with respect to the parameters at
#'   the observation time-point
#' @author Matthew L. Fidler
#' @export 
#' @examples
#' \donttest{
#' llikNbinom(46:54, 100, 0.5)
#'
#' llikNbinom(46:54, 100, 0.5, TRUE)
#'
#' # In rxode2 you can use:
#' 
#' et <- et(46:54)
#' et$size <- 100
#' et$prob <-0.5
#'
#' model <- function() {
#'   model({
#'     fx <- llikNbinom(time, size, prob)
#'     dProb <- llikNbinomDprob(time, size, prob)
#'   })
#' }
#'
#' rxSolve(model, et)
#' }
llikNbinom <- function(x, size, prob, full=FALSE) {
  rxode2ll::llikNbinom(x, size, prob, full)
}

#' Calculate the log likelihood of the negative binomial function (and its derivatives)
#' 
#' @param x  Number of successes
#' 
#' @param size Size of trial
#' 
#' @param mu mu parameter for negative binomial
#' 
#' @inheritParams llikNorm
#'
#' @details
#' 
#' In an `rxode2()` model, you can use `llikNbinomMu()` but you have to
#' use all arguments.  You can also get the derivative of `mu` with
#' `llikNbinomMuDmu()`
#' 
#' @return data frame with `fx` for the pdf value of with
#'   `dProb` that has the derivatives with respect to the parameters at
#'   the observation time-point
#' 
#' @author Matthew L. Fidler
#' 
#' @export 
#' @examples
#' \donttest{
#' llikNbinomMu(46:54, 100, 40)
#'
#' llikNbinomMu(46:54, 100, 40, TRUE)
#' 
#' et <- et(46:54)
#' et$size <- 100
#' et$mu <- 40
#'
#' model <- function() {
#'   model({
#'     fx <- llikNbinomMu(time, size, mu)
#'     dProb <- llikNbinomMuDmu(time, size, mu)
#'   })
#' }
#'
#' rxSolve(model, et)
#' }
llikNbinomMu <- function(x, size, mu, full=FALSE) {
  rxode2ll::llikNbinomMu(x, size, mu, full)
}

#' Calculate the log likelihood of the binomial function (and its derivatives)
#'
#' @inheritParams stats::dbeta
#' 
#' @inheritParams llikNorm
#'
#' @details
#' 
#' In an `rxode2()` model, you can use `llikBeta()` but you have to
#' use all arguments.  You can also get the derivative of `shape1` and `shape2` with
#' `llikBetaDshape1()` and `llikBetaDshape2()`.
#' 
#' @return data frame with `fx` for the log pdf value of with
#'   `dShape1` and `dShape2` that has the derivatives with respect to the parameters at
#'   the observation time-point
#' 
#' @author Matthew L. Fidler
#' 
#' @export
#' 
#' @examples
#' 
#' \donttest{
#'
#' x <- seq(1e-4, 1 - 1e-4, length.out = 21)
#' 
#' llikBeta(x, 0.5, 0.5)
#'
#' llikBeta(x, 1, 3, TRUE)
#' 
#' et <- et(seq(1e-4, 1-1e-4, length.out=21))
#' et$shape1 <- 0.5
#' et$shape2 <- 1.5
#'
#' model <- function() {
#'   model({
#'     fx <- llikBeta(time, shape1, shape2)
#'     dShape1 <- llikBetaDshape1(time, shape1, shape2)
#'     dShape2 <- llikBetaDshape2(time, shape1, shape2)
#'   })
#' }
#'
#' rxSolve(model, et)
#' }
llikBeta <- function(x, shape1, shape2, full=FALSE) {
  rxode2ll::llikBeta(x, shape1, shape2, full)
}

#' Log likelihood of T and it's derivatives (from stan) 
#'
#' @param x  Observation
#' @inheritParams llikNorm
#' @inheritParams stats::dnorm
#' @inheritParams stats::dt
#' @return data frame with `fx` for the log pdf value of with `dDf`
#'   `dMean` and `dSd` that has the derivatives with respect to the parameters at
#'   the observation time-point
#' @author Matthew L. Fidler
#' @details
#' In an `rxode2()` model, you can use `llikT()` but you have to
#' use all arguments.  You can also get the derivative of `df`, `mean` and `sd` with
#' `llikTDdf()`, `llikTDmean()` and `llikTDsd()`.
#' @export 
#' @examples
#'
#' \donttest{
#' x <- seq(-3, 3, length.out = 21)
#'
#' llikT(x, 7, 0, 1)
#'
#' llikT(x, 15, 0, 1, full=TRUE)
#'
#' et <- et(-3, 3, length.out=10)
#' et$nu <- 7
#' et$mean <- 0
#' et$sd <- 1
#'
#' model <- function() {
#'   model({
#'     fx <- llikT(time, nu, mean, sd)
#'     dDf <- llikTDdf(time, nu, mean, sd)
#'     dMean <- llikTDmean(time, nu, mean, sd)
#'     dSd   <- llikTDsd(time, nu, mean, sd)
#'   })
#' }
#'
#' rxSolve(model, et)
#' }
#'
llikT <- function(x, df, mean=0, sd=1, full=FALSE) {
  rxode2ll::llikT(x, df, mean, sd)
}

#' log likelihood and derivatives for chi-squared distribution
#'
#' @param x variable that is distributed by chi-squared distribution
#' @inheritParams llikNorm
#' @inheritParams stats::dchisq
#' @return data frame with `fx` for the log pdf value of with `dDf`
#'   that has the derivatives with respect to the `df` parameter
#'   the observation time-point
#' @author Matthew L. Fidler
#' @export 
#' @details
#' In an `rxode2()` model, you can use `llikChisq()` but you have to
#' use the x and df arguments.  You can also get the derivative of `df` with
#' `llikChisqDdf()`.
#' @examples
#'
#' \donttest{
#' llikChisq(1, df = 1:3, full=TRUE)
#' 
#' llikChisq(1, df = 6:9)
#' 
#' et <- et(1:3)
#' et$x <- 1
#'
#' model <- function() {
#'   model({
#'    fx <- llikChisq(x, time)
#'    dDf <- llikChisqDdf(x, time)
#'   })
#' }
#'
#' rxSolve(model, et)
#' }
llikChisq <- function(x, df, full=FALSE) {
  rxode2ll::llikChisq(x, df, full)
}

#' log likelihood and derivatives for exponential distribution
#' 
#' @param x variable that is distributed by exponential distribution
#' @inheritParams llikNorm
#' @inheritParams stats::dexp
#' @return data frame with `fx` for the log pdf value of with `dRate`
#'   that has the derivatives with respect to the `rate` parameter
#'   the observation time-point
#' 
#' @author Matthew L. Fidler
#' 
#' @export
#' 
#' @details
#' In an `rxode2()` model, you can use `llikExp()` but you have to
#' use the x and rate arguments.  You can also get the derivative of `rate` with
#' `llikExpDrate()`.
#' 
#' @examples
#' \donttest{
#' llikExp(1, 1:3)
#'
#' llikExp(1, 1:3, full=TRUE)
#'
#' # You can use rxode2 for these too:
#'
#' et <- et(1:3)
#' et$x <- 1
#'
#' model <- function() {
#'   model({
#'     fx <- llikExp(x, time)
#'     dRate <- llikExpDrate(x, time)
#'   })
#' }
#'
#' rxSolve(model, et)
#' }
llikExp <- function(x, rate, full=FALSE) {
  rxode2ll::llikExp(x, rate, full)
}

#' log likelihood and derivatives for F distribution
#'
#' @param x variable that is distributed by f distribution
#' @inheritParams llikNorm
#' @inheritParams stats::df
#' @return data frame with `fx` for the log pdf value of with `dDf1` and `dDf2`
#'   that has the derivatives with respect to the `df1`/`df2` parameters at 
#'   the observation time-point
#' 
#' @author Matthew L. Fidler
#' 
#' @export
#' 
#' @details
#' In an `rxode2()` model, you can use `llikF()` but you have to
#' use the x and rate arguments.  You can also get the derivative of `df1` and `df2` with
#' `llikFDdf1()` and `llikFDdf2()`.
#' 
#' @examples
#'
#' \donttest{
#' x <- seq(0.001, 5, length.out = 100)
#'
#' llikF(x^2, 1, 5)
#'
#' model <- function(){
#'   model({
#'     fx <- llikF(time, df1, df2)
#'     dMean <- llikFDdf1(time, df1, df2)
#'     dSd <- llikFDdf2(time, df1, df2)
#'   })
#' }
#' 
#' et <- et(x)
#' et$df1 <- 1
#' et$df2 <- 5
#' 
#' rxSolve(model, et)
#' }
llikF <- function(x, df1, df2, full=FALSE) {
  rxode2ll::llikF(x, df1, df2, full)
}

#' log likelihood and derivatives for Geom distribution
#'
#' @param x variable distributed by a geom distribution
#' 
#' @inheritParams llikNorm
#' @inheritParams stats::dgeom
#' @return data frame with `fx` for the log pdf value of with `dProb`
#'   that has the derivatives with respect to the `prob` parameters at 
#'   the observation time-point
#' 
#' @author Matthew L. Fidler
#' 
#' @export
#' 
#' @details
#' In an `rxode2()` model, you can use `llikGeom()` but you have to
#' use the x and rate arguments.  You can also get the derivative of `prob` with
#' `llikGeomDprob()`.
#' 
#' @examples
#' 
#' \donttest{
#' 
#' llikGeom(1:10, 0.2)
#' 
#' et  <- et(1:10)
#' et$prob <- 0.2
#'  
#' model <- function() {
#'   model({
#'     fx <- llikGeom(time, prob)
#'     dProb <- llikGeomDprob(time, prob)
#'   })
#' }
#'
#' rxSolve(model, et)
#' }
llikGeom <- function(x, prob, full=FALSE) {
  rxode2ll::llikGeom(x, prob, full)
}

#' log likelihood and derivatives for Unif distribution
#'
#' @param x variable distributed by a uniform distribution
#' @param alpha is the lower limit of the uniform distribution
#' @param beta is the upper limit of the distribution
#' @inheritParams llikNorm
#' @inheritParams stats::dunif
#' @return data frame with `fx` for the log pdf value of with `dProb`
#'   that has the derivatives with respect to the `prob` parameters at 
#'   the observation time-point
#' 
#' @author Matthew L. Fidler
#' 
#' @export
#' 
#' @details
#' 
#' In an `rxode2()` model, you can use `llikUnif()` but you have to
#' use the x and rate arguments.  You can also get the derivative of `alpha` or `beta` with
#' `llikUnifDalpha()` and `llikUnifDbeta()`.
#' 
#' @examples
#'
#' \donttest{
#'
#' llikUnif(1, -2, 2)
#'
#' et  <- et(seq(1,1, length.out=4))
#' et$alpha <- -2
#' et$beta <- 2
#'  
#' model <- function() {
#'   model({
#'     fx <- llikUnif(time, alpha, beta)
#'     dAlpha<- llikUnifDalpha(time, alpha, beta)
#'     dBeta <- llikUnifDbeta(time, alpha, beta)
#'   })
#' }
#' 
#' rxSolve(model, et)
#' }
llikUnif <- function(x, alpha, beta, full=FALSE) {
  rxode2ll::llikUnif(x, alpha, beta, full)
}

#' log likelihood and derivatives for Weibull distribution
#' 
#' @param x variable distributed by a Weibull distribution
#' 
#' @inheritParams llikNorm
#' @inheritParams stats::dweibull
#' @return data frame with `fx` for the log pdf value of with `dProb`
#'   that has the derivatives with respect to the `prob` parameters at 
#'   the observation time-point
#' 
#' @author Matthew L. Fidler
#' 
#' @export
#' 
#' @details
#' 
#' In an `rxode2()` model, you can use `llikWeibull()` but you have to
#' use the x and rate arguments.  You can also get the derivative of `shape` or `scale` with
#' `llikWeibullDshape()` and `llikWeibullDscale()`.
#' 
#' @examples
#' \donttest{
#' llikWeibull(1, 1, 10)
#'
#' # rxode2 can use this too:
#'
#' et  <- et(seq(0.001, 1, length.out=10))
#' et$shape <- 1
#' et$scale <- 10
#'  
#' model <- function() {
#'   model({
#'     fx <- llikWeibull(time, shape, scale)
#'     dShape<- llikWeibullDshape(time, shape, scale)
#'     dScale <- llikWeibullDscale(time, shape, scale)
#'   })
#' }
#'
#' rxSolve(model, et)
#' }
llikWeibull <- function(x, shape, scale, full=FALSE) {
  rxode2ll::llikWeibull(x, shape, scale, full)
}


#' log likelihood and derivatives for Gamma distribution
#'
#' @param x variable that is distributed by gamma distribution
#' @param shape this is the distribution's shape parameter. Must be positive.
#' @param rate this is the distribution's rate parameters.  Must be positive.
#' 
#' @inheritParams llikNorm
#' 
#' @inheritParams stats::dgamma
#' 
#' @return data frame with `fx` for the log pdf value of with `dProb`
#'   that has the derivatives with respect to the `prob` parameters at 
#'   the observation time-point
#' 
#' @author Matthew L. Fidler
#' 
#' @export
#' 
#' @details
#' 
#' In an `rxode2()` model, you can use `llikGamma()` but you have to
#' use the x and rate arguments.  You can also get the derivative of `shape` or `rate` with
#' `llikGammaDshape()` and `llikGammaDrate()`.
#' 
#' @examples
#' \donttest{
#' 
#' llikGamma(1, 1, 10)
#' 
#' # You can use this in `rxode2` too:
#' 
#' et  <- et(seq(0.001, 1, length.out=10))
#' et$shape <- 1
#' et$rate <- 10
#'  
#' model <- function() {
#'   model({
#'     fx <- llikGamma(time, shape, rate)
#'     dShape<- llikGammaDshape(time, shape, rate)
#'     dRate <- llikGammaDrate(time, shape, rate)
#'   })
#' }
#' 
#' rxSolve(model, et)
#' }
llikGamma <- function(x, shape, rate, full=FALSE) {
  rxode2ll::llikGamma(x, shape, rate, full)
}


#' log likelihood of Cauchy distribution and it's derivatives (from stan) 
#'
#' @param x  Observation
#' @inheritParams llikNorm
#' @inheritParams stats::dnorm
#' @inheritParams stats::dcauchy
#' @return data frame with `fx` for the log pdf value of with 
#'   `dLocation` and `dScale` that has the derivatives with respect to the parameters at
#'   the observation time-point
#' @author Matthew L. Fidler
#' @details
#' In an `rxode2()` model, you can use `llikCauchy()` but you have to
#' use all arguments.  You can also get the derivative of `location` and `scale` with
#' `llikCauchyDlocation()` and `llikCauchyDscale()`.
#' @export 
#' @examples
#' \donttest{
#' x <- seq(-3, 3, length.out = 21)
#'
#' llikCauchy(x, 0, 1)
#'
#' llikCauchy(x, 3, 1, full=TRUE)
#'
#' et <- et(-3, 3, length.out=10)
#' et$location <- 0
#' et$scale <- 1
#'
#' model <- function() {
#'   model({
#'     fx <- llikCauchy(time, location, scale)
#'     dLocation <- llikCauchyDlocation(time, location, scale)
#'     dScale <- llikCauchyDscale(time, location, scale)
#'   })
#' }
#'
#' rxSolve(model, et)
#' }
llikCauchy <- function(x, location=0, scale=1, full=FALSE) {
  rxode2ll::llikCauchy(x, location, scale, full)
}

rxTest({
  test_that("number of llik calculations saved (for nlmixr2)", {
    # This is used in nlmixr2 generalized log likelihood in focei to
    # remove duplicate likelihood/likelihood gradient calls

    expect_equal(setNames(rxModelVars("a=3")$flags["nLlik"], NULL), 0L)
    expect_equal(setNames(rxModelVars("a=llikNorm(a,mu,sigma)")$flags["nLlik"], NULL), 1L)

    expect_equal(setNames(rxModelVars("a=llikXNorm(3, a,mu,sigma)")$flags["nLlik"], NULL), 4L)

  })

  test_that("log-liklihood tests for normal (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikNorm(x, mu, sigma)"), NA)
    expect_error(rxode2("tmp=llikNormDsd(x, mu, sigma)"), NA)
    expect_error(rxode2("tmp=llikNormDmean(x, mu, sigma)"), NA)

    expect_error(rxode2("tmp=llikXNorm(1, x, mu, sigma)"), NA)
    expect_error(rxode2("tmp=llikXNormDsd(1, x, mu, sigma)"), NA)
    expect_error(rxode2("tmp=llikXNormDmean(1, x, mu, sigma)"), NA)


    # Make sure they translate correctly:
    expect_equal(rxToSE("llikNorm(x, mu, sigma)"), "llikNorm(x,mu,sigma)")
    expect_equal(rxToSE("llikNormDsd(x, mu, sigma)"), "llikNormDsd(x,mu,sigma)")
    expect_equal(rxToSE("llikNormDmean(x, mu, sigma)"), "llikNormDmean(x,mu,sigma)")
    expect_equal(rxFromSE("llikNorm(x, mu, sigma)"), "llikNorm(x,mu,sigma)")
    expect_equal(rxFromSE("llikNormDsd(x, mu, sigma)"), "llikNormDsd(x,mu,sigma)")
    expect_equal(rxFromSE("llikNormDmean(x, mu, sigma)"), "llikNormDmean(x,mu,sigma)")

    expect_equal(rxToSE("llikXNorm(1, x, mu, sigma)"), "llikXNorm(1,x,mu,sigma)")
    expect_equal(rxToSE("llikXNormDsd(1, x, mu, sigma)"), "llikXNormDsd(1,x,mu,sigma)")
    expect_equal(rxToSE("llikXNormDmean(1, x, mu, sigma)"), "llikXNormDmean(1,x,mu,sigma)")
    expect_equal(rxFromSE("llikXNorm(1, x, mu, sigma)"), "llikXNorm(1,x,mu,sigma)")
    expect_equal(rxFromSE("llikXNormDsd(1, x, mu, sigma)"), "llikXNormDsd(1,x,mu,sigma)")
    expect_equal(rxFromSE("llikXNormDmean(1, x, mu, sigma)"), "llikXNormDmean(1,x,mu,sigma)")

    # Check the derivatives

    # this is forward difference with no correction
    #rxFromSE("Derivative(llikNorm(x,mu,sigma),x)")

    expect_equal(rxFromSE("Derivative(llikNorm(x,mu,sigma),mu)"), "llikNormDmean(x, mu, sigma)")
    expect_equal(rxFromSE("Derivative(llikNorm(x,mu,sigma),sigma)"), "llikNormDsd(x, mu, sigma)")
    expect_equal(rxFromSE("Derivative(llikXNorm(2,x,mu,sigma),mu)"), "llikXNormDmean(2, x, mu, sigma)")
    expect_equal(rxFromSE("Derivative(llikXNorm(2,x,mu,sigma),sigma)"), "llikXNormDsd(2, x, mu, sigma)")


    et <- et(-3, 3, length.out=10)
    et$mu <- 0
    et$sigma <- 1

    model <- rxode2({
      fx <- llikNorm(time, mu, sigma)
      dMean <- llikNormDmean(time, mu, sigma)
      dSd <- llikNormDsd(time, mu, sigma)
    })

    modelX <- rxode2({
      fx <- llikXNorm(1, time, mu, sigma)
      dMean <- llikXNormDmean(1, time, mu, sigma)
      dSd <- llikXNormDsd(1, time, mu, sigma)
    })


    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)
    fromR <- llikNorm(et$time, et$mu, et$sigma)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dMean, fromOde$dMean)
    expect_equal(fromR$dSd, fromOde$dSd)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dMean, fromOdeX$dMean)
    expect_equal(fromR$dSd, fromOdeX$dSd)

    expect_equal(fromR$fx, dnorm(fromOde$time, log=TRUE))

  })


  test_that("log-liklihood tests for pois (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikPois(x, lambda)"), NA)
    expect_error(rxode2("tmp=llikPoisDlambda(x, lambda)"), NA)

    expect_error(rxode2("tmp=llikXPois(1, x, lambda)"), NA)
    expect_error(rxode2("tmp=llikXPoisDlambda(1, x, lambda)"), NA)


    # Make sure they translate correctly:
    expect_equal(rxToSE("llikPois(x, lambda)"), "llikPois(x,lambda)")
    expect_equal(rxToSE("llikPoisDlambda(x, lambda)"), "llikPoisDlambda(x,lambda)")
    expect_equal(rxFromSE("llikPois(x, lambda)"), "llikPois(x,lambda)")
    expect_equal(rxFromSE("llikPoisDlambda(x, lambda)"), "llikPoisDlambda(x,lambda)")

    expect_equal(rxToSE("llikXPois(1, x, lambda)"), "llikXPois(1,x,lambda)")
    expect_equal(rxToSE("llikXPoisDlambda(1,x, lambda)"), "llikXPoisDlambda(1,x,lambda)")
    expect_equal(rxFromSE("llikXPois(2, x, lambda)"), "llikXPois(2,x,lambda)")
    expect_equal(rxFromSE("llikXPoisDlambda(2, x, lambda)"), "llikXPoisDlambda(2,x,lambda)")


    # Check the derivatives

    # this is forward difference with no correction
    #rxFromSE("Derivative(llikNorm(x,mu,sigma),x)")

    expect_equal(rxFromSE("Derivative(llikPois(x,lambda),lambda)"), "llikPoisDlambda(x, lambda)")

    expect_equal(rxFromSE("Derivative(llikXPois(1,x,lambda),lambda)"), "llikXPoisDlambda(1, x, lambda)")

    et <- et(0:10)
    et$lambda <- 0.5

    model <- rxode2({
      fx <- llikPois(time, lambda)
      dLambda <- llikPoisDlambda(time, lambda)
    })

    modelX <- rxode2({
      fx <- llikXPois(1, time, lambda)
      dLambda <- llikXPoisDlambda(1, time, lambda)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(model, et)

    fromR <- llikPois(et$time, et$lambda)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dLambda, fromOde$dLambda)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dLambda, fromOdeX$dLambda)

    expect_equal(fromR$fx, dpois(fromOde$time, lambda=0.5, log=TRUE))

  })

  test_that("log-liklihood tests for binom (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikBinom(x, size, prob)"), NA)
    expect_error(rxode2("tmp=llikBinomDprob(x, size, prob)"), NA)

    expect_error(rxode2("tmp=llikXBinom(1, x, size, prob)"), NA)
    expect_error(rxode2("tmp=llikXBinomDprob(1, x, size, prob)"), NA)


    # Make sure they translate correctly:
    expect_equal(rxToSE("llikBinom(x, size, prob)"), "llikBinom(x,size,prob)")
    expect_equal(rxToSE("llikBinomDprob(x, size, prob)"), "llikBinomDprob(x,size,prob)")
    expect_equal(rxFromSE("llikBinom(x, size, prob)"), "llikBinom(x,size,prob)")
    expect_equal(rxFromSE("llikPoisDlambda(x, lambda)"), "llikPoisDlambda(x,lambda)")

    expect_equal(rxToSE("llikXBinom(i,x, size, prob)"), "llikXBinom(i,x,size,prob)")
    expect_equal(rxToSE("llikXBinomDprob(i,x, size, prob)"), "llikXBinomDprob(i,x,size,prob)")
    expect_equal(rxFromSE("llikXBinom(i,x, size, prob)"), "llikXBinom(i,x,size,prob)")
    expect_equal(rxFromSE("llikXPoisDlambda(i,x, lambda)"), "llikXPoisDlambda(i,x,lambda)")

    # Check the derivatives

    # this is forward difference with no correction
    expect_equal(rxFromSE("Derivative(llikBinom(x,size,prob),size)"),"0")
    expect_equal(rxFromSE("Derivative(llikBinom(x,size, prob),prob)"),
                 "llikBinomDprob(x, size, prob)")

    expect_equal(rxFromSE("Derivative(llikXBinom(i,x,size,prob),size)"),"0")
    expect_equal(rxFromSE("Derivative(llikXBinom(i,x,size, prob),prob)"), "llikXBinomDprob(i, x, size, prob)")

    et <- et(0:10)
    et$size <- 100
    et$prob <- 0.5

    model <- rxode2({
      fx <- llikBinom(time, size, prob)
      dProb <- llikBinomDprob(time, size, prob)
    })

    modelX <- rxode2({
      fx <- llikXBinom(1, time, size, prob)
      dProb <- llikXBinomDprob(1, time, size, prob)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikBinom(et$time, et$size, et$prob, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dProb, fromOde$dProb)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dProb, fromOdeX$dProb)

    expect_equal(fromR$fx, dbinom(fromOde$time, size=100, prob=0.5, log=TRUE))
  })

  test_that("log-liklihood tests for nbinom (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikNbinom(x, size, prob)"), NA)
    expect_error(rxode2("tmp=llikNbinomDprob(x, size, prob)"), NA)

    expect_error(rxode2("tmp=llikXNbinom(1, x, size, prob)"), NA)
    expect_error(rxode2("tmp=llikXNbinomDprob(1, x, size, prob)"), NA)


    # Make sure they translate correctly:
    expect_equal(rxToSE("llikNbinom(x, size, prob)"), "llikNbinom(x,size,prob)")
    expect_equal(rxToSE("llikNbinomDprob(x, size, prob)"), "llikNbinomDprob(x,size,prob)")
    expect_equal(rxFromSE("llikNbinom(x, size, prob)"), "llikNbinom(x,size,prob)")

    expect_equal(rxToSE("llikXNbinom(i,x, size, prob)"), "llikXNbinom(i,x,size,prob)")
    expect_equal(rxToSE("llikXNbinomDprob(i,x, size, prob)"), "llikXNbinomDprob(i,x,size,prob)")
    expect_equal(rxFromSE("llikXNbinom(i,x, size, prob)"), "llikXNbinom(i,x,size,prob)")

    # Check the derivatives

    # this is forward difference with no correction
    expect_equal(rxFromSE("Derivative(llikNbinom(x,size,prob),size)"),"0")
    expect_equal(rxFromSE("Derivative(llikNbinom(x,size, prob),prob)"), "llikNbinomDprob(x, size, prob)")

    expect_equal(rxFromSE("Derivative(llikXNbinom(i,x,size,prob),size)"),"0")
    expect_equal(rxFromSE("Derivative(llikXNbinom(i,x,size, prob),prob)"), "llikXNbinomDprob(i, x, size, prob)")

    et <- et(0:10)
    et$size <- 100
    et$prob <- 0.5

    model <- rxode2({
      fx <- llikNbinom(time, size, prob)
      dProb <- llikNbinomDprob(time, size, prob)
    })

    modelX <- rxode2({
      fx <- llikXNbinom(1, time, size, prob)
      dProb <- llikXNbinomDprob(1, time, size, prob)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikNbinom(et$time, et$size, et$prob, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dProb, fromOde$dProb)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dProb, fromOde$dProb)

    expect_equal(fromR$fx, dnbinom(fromOde$time, size=100, prob=0.5, log=TRUE))

  })


  test_that("log-liklihood tests for NbinomMu (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikNbinomMu(x, size, mu)"), NA)
    expect_error(rxode2("tmp=llikNbinomMuDmu(x, size, mu)"), NA)

    expect_error(rxode2("tmp=llikXNbinomMu(1, x, size, mu)"), NA)
    expect_error(rxode2("tmp=llikXNbinomMuDmu(1, x, size, mu)"), NA)


    # Make sure they translate correctly:
    expect_equal(rxToSE("llikNbinomMu(x, size, mu)"), "llikNbinomMu(x,size,mu)")
    expect_equal(rxToSE("llikNbinomMuDmu(x, size, mu)"), "llikNbinomMuDmu(x,size,mu)")
    expect_equal(rxFromSE("llikNbinomMu(x, size, mu)"), "llikNbinomMu(x,size,mu)")

    expect_equal(rxToSE("llikXNbinomMu(i,x, size, mu)"), "llikXNbinomMu(i,x,size,mu)")
    expect_equal(rxToSE("llikXNbinomMuDmu(i,x, size, mu)"), "llikXNbinomMuDmu(i,x,size,mu)")
    expect_equal(rxFromSE("llikXNbinomMu(i,x, size, mu)"), "llikXNbinomMu(i,x,size,mu)")

    # Check the derivatives

    # this is forward difference with no correction
    expect_equal(rxFromSE("Derivative(llikNbinomMu(x,size,mu),size)"),"0")
    expect_equal(rxFromSE("Derivative(llikNbinomMu(x,size, mu),mu)"), "llikNbinomMuDmu(x, size, mu)")

    expect_equal(rxFromSE("Derivative(llikXNbinomMu(i,x,size,mu),size)"),"0")
    expect_equal(rxFromSE("Derivative(llikXNbinomMu(i,x,size, mu),mu)"), "llikXNbinomMuDmu(i, x, size, mu)")

    et <- et(0:10)
    et$size <- 100
    et$mu <- 40

    model <- rxode2({
      fx <- llikNbinomMu(time, size, mu)
      dMu <- llikNbinomMuDmu(time, size, mu)
    })

    modelX <- rxode2({
      fx <- llikXNbinomMu(1, time, size, mu)
      dMu <- llikXNbinomMuDmu(1, time, size, mu)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikNbinomMu(et$time, et$size, et$mu, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dMu, fromOde$dMu)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dMu, fromOde$dMu)

    expect_equal(fromR$fx, dnbinom(fromOde$time, size=100, mu=40, log=TRUE))
  })

  test_that("log-liklihood tests for beta (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikBeta(x, shape1, shape2)"), NA)
    expect_error(rxode2("tmp=llikBetaDshape1(x, shape1, shape2)"), NA)
    expect_error(rxode2("tmp=llikBetaDshape2(x, shape1, shape2)"), NA)

    # Make sure they compile:
    expect_error(rxode2("tmp=llikXBeta(1, x, shape1, shape2)"), NA)
    expect_error(rxode2("tmp=llikXBetaDshape1(1, x, shape1, shape2)"), NA)
    expect_error(rxode2("tmp=llikXBetaDshape2(1, x, shape1, shape2)"), NA)

    # Make sure they translate correctly:
    expect_equal(rxToSE("llikBeta(x, shape1, shape2)"), "llikBeta(x,shape1,shape2)")
    expect_equal(rxToSE("llikBetaDshape1(x, shape1, shape2)"), "llikBetaDshape1(x,shape1,shape2)")
    expect_equal(rxToSE("llikBetaDshape2(x, shape1, shape2)"), "llikBetaDshape2(x,shape1,shape2)")
    expect_equal(rxFromSE("llikBeta(x, shape1, shape2)"), "llikBeta(x,shape1,shape2)")
    expect_equal(rxFromSE("llikBetaDshape1(x, shape1, shape2)"), "llikBetaDshape1(x,shape1,shape2)")
    expect_equal(rxFromSE("llikBetaDshape2(x, shape1, shape2)"), "llikBetaDshape2(x,shape1,shape2)")

    expect_equal(rxToSE("llikXBeta(i,x, shape1, shape2)"), "llikXBeta(i,x,shape1,shape2)")
    expect_equal(rxToSE("llikXBetaDshape1(i,x, shape1, shape2)"), "llikXBetaDshape1(i,x,shape1,shape2)")
    expect_equal(rxToSE("llikXBetaDshape2(i,x, shape1, shape2)"), "llikXBetaDshape2(i,x,shape1,shape2)")
    expect_equal(rxFromSE("llikXBeta(i,x, shape1, shape2)"), "llikXBeta(i,x,shape1,shape2)")
    expect_equal(rxFromSE("llikXBetaDshape1(i,x, shape1, shape2)"), "llikXBetaDshape1(i,x,shape1,shape2)")
    expect_equal(rxFromSE("llikXBetaDshape2(i,x, shape1, shape2)"), "llikXBetaDshape2(i,x,shape1,shape2)")

    # Check the derivatives
    expect_equal(rxFromSE("Derivative(llikBeta(x,shape1,shape2),shape1)"),"llikBetaDshape1(x, shape1, shape2)")
    expect_equal(rxFromSE("Derivative(llikBeta(x,shape1, shape2),shape2)"), "llikBetaDshape2(x, shape1, shape2)")

    expect_equal(rxFromSE("Derivative(llikXBeta(i,x,shape1,shape2),shape1)"),
                 "llikXBetaDshape1(i, x, shape1, shape2)")
    expect_equal(rxFromSE("Derivative(llikXBeta(i,x,shape1, shape2),shape2)"),
                 "llikXBetaDshape2(i, x, shape1, shape2)")

    et <- et(seq(1e-4, 1-1e-4, length.out=21))
    et$shape1 <- 0.5
    et$shape2 <- 1.5

    model <- rxode2({
      fx <- llikBeta(time, shape1, shape2)
      dShape1 <- llikBetaDshape1(time, shape1, shape2)
      dShape2 <- llikBetaDshape2(time, shape1, shape2)
    })

    modelX <- rxode2({
      fx <- llikXBeta(1, time, shape1, shape2)
      dShape1 <- llikXBetaDshape1(1,time, shape1, shape2)
      dShape2 <- llikXBetaDshape2(1,time, shape1, shape2)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikBeta(et$time, et$shape1, et$shape2, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dShape1, fromOde$dShape1)
    expect_equal(fromR$dShape2, fromOde$dShape2)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dShape1, fromOdeX$dShape1)
    expect_equal(fromR$dShape2, fromOdeX$dShape2)

    expect_equal(fromR$fx, dbeta(fromOde$time, shape1=0.5, shape2=1.5, log=TRUE))

  })


  test_that("log-liklihood tests for T (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikT(x, nu, mean, sd)"), NA)
    expect_error(rxode2("tmp=llikTDdf(x, nu, mean, sd)"), NA)
    expect_error(rxode2("tmp=llikTDmean(x, nu, mean, sd)"), NA)
    expect_error(rxode2("tmp=llikTDsd(x, nu, mean, sd)"), NA)

    expect_error(rxode2("tmp=llikXT(i,x, nu, mean, sd)"), NA)
    expect_error(rxode2("tmp=llikXTDdf(i,x, nu, mean, sd)"), NA)
    expect_error(rxode2("tmp=llikXTDmean(i,x, nu, mean, sd)"), NA)
    expect_error(rxode2("tmp=llikXTDsd(i,x, nu, mean, sd)"), NA)

    # Make sure they translate correctly:
    expect_equal(rxToSE("llikT(x, nu, mean, sd)"), "llikT(x,nu,mean,sd)")
    expect_equal(rxFromSE("llikT(x, nu, mean, sd)"), "llikT(x,nu,mean,sd)")

    expect_equal(rxToSE("llikTDdf(x, nu, mean, sd)"), "llikTDdf(x,nu,mean,sd)")
    expect_equal(rxFromSE("llikTDdf(x, nu, mean, sd)"), "llikTDdf(x,nu,mean,sd)")

    expect_equal(rxToSE("llikTDmean(x, nu, mean, sd)"), "llikTDmean(x,nu,mean,sd)")
    expect_equal(rxFromSE("llikTDmean(x, nu, mean, sd)"), "llikTDmean(x,nu,mean,sd)")

    expect_equal(rxToSE("llikTDsd(x, nu, mean, sd)"), "llikTDsd(x,nu,mean,sd)")
    expect_equal(rxFromSE("llikTDsd(x, nu, mean, sd)"), "llikTDsd(x,nu,mean,sd)")

    expect_equal(rxToSE("llikXT(i,x, nu, mean, sd)"), "llikXT(i,x,nu,mean,sd)")
    expect_equal(rxFromSE("llikXT(i,x, nu, mean, sd)"), "llikXT(i,x,nu,mean,sd)")

    expect_equal(rxToSE("llikXTDdf(i,x, nu, mean, sd)"), "llikXTDdf(i,x,nu,mean,sd)")
    expect_equal(rxFromSE("llikXTDdf(i,x, nu, mean, sd)"), "llikXTDdf(i,x,nu,mean,sd)")

    expect_equal(rxToSE("llikXTDmean(i,x, nu, mean, sd)"), "llikXTDmean(i,x,nu,mean,sd)")
    expect_equal(rxFromSE("llikXTDmean(i,x, nu, mean, sd)"), "llikXTDmean(i,x,nu,mean,sd)")

    expect_equal(rxToSE("llikXTDsd(i,x, nu, mean, sd)"), "llikXTDsd(i,x,nu,mean,sd)")
    expect_equal(rxFromSE("llikXTDsd(i,x, nu, mean, sd)"), "llikXTDsd(i,x,nu,mean,sd)")

    # Check the derivatives
    expect_equal(rxFromSE("Derivative(llikT(x,nu, mean, sd),nu)"),"llikTDdf(x, nu, mean, sd)")

    expect_equal(rxFromSE("Derivative(llikT(x, nu, mean, sd), mean)"),
                 "llikTDmean(x, nu, mean, sd)")

    expect_equal(rxFromSE("Derivative(llikT(x, nu, mean, sd), sd)"),
                 "llikTDsd(x, nu, mean, sd)")


    #
    expect_equal(rxFromSE("Derivative(llikXT(x, i,nu, mean, sd),nu)"),"llikXTDdf(x, i, nu, mean, sd)")

    expect_equal(rxFromSE("Derivative(llikXT(x, i, nu, mean, sd), mean)"),
                 "llikXTDmean(x, i, nu, mean, sd)")

    expect_equal(rxFromSE("Derivative(llikXT(x, i, nu, mean, sd), sd)"),
                 "llikXTDsd(x, i, nu, mean, sd)")


    # Check rxode2 internals with R exported
    et <- et(-3, 3, length.out=10)
    et$nu <- 7
    et$mean <- 0
    et$sd <- 1

    model <- rxode2({
      fx <- llikT(time, nu, mean, sd)
      dDf <- llikTDdf(time, nu, mean, sd)
      dMean <- llikTDmean(time, nu, mean, sd)
      dSd   <- llikTDsd(time, nu, mean, sd)
    })

    modelX <- rxode2({
      fx <- llikXT(1, time, nu, mean, sd)
      dDf <- llikXTDdf(1, time, nu, mean, sd)
      dMean <- llikXTDmean(1, time, nu, mean, sd)
      dSd   <- llikXTDsd(1, time, nu, mean, sd)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikT(et$time, et$nu, et$mean, et$sd, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dDf, fromOde$dDf)
    expect_equal(fromR$dMean, fromOde$dMean)
    expect_equal(fromR$dSd, fromOde$dSd)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dDf, fromOdeX$dDf)
    expect_equal(fromR$dMean, fromOdeX$dMean)
    expect_equal(fromR$dSd, fromOdeX$dSd)

    expect_equal(fromR$fx, dt(fromOde$time, df=7, log=TRUE))
  })


  test_that("log-liklihood tests for chi-squared (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikChisq(x, nu)"), NA)
    expect_error(rxode2("tmp=llikChisqDdf(x, nu)"), NA)

    expect_error(rxode2("tmp=llikXChisq(1, x, nu)"), NA)
    expect_error(rxode2("tmp=llikXChisqDdf(1, x, nu)"), NA)

    # Make sure they translate correctly:
    expect_equal(rxToSE("llikChisq(x, nu)"), "llikChisq(x,nu)")
    expect_equal(rxFromSE("llikChisq(x, nu)"), "llikChisq(x,nu)")

    expect_equal(rxToSE("llikChisqDdf(x, nu)"), "llikChisqDdf(x,nu)")
    expect_equal(rxFromSE("llikChisqDdf(x, nu)"), "llikChisqDdf(x,nu)")

    #
    expect_equal(rxToSE("llikXChisq(i,x, nu)"), "llikXChisq(i,x,nu)")
    expect_equal(rxFromSE("llikXChisq(i,x, nu)"), "llikXChisq(i,x,nu)")

    expect_equal(rxToSE("llikXChisqDdf(i,x, nu)"), "llikXChisqDdf(i,x,nu)")
    expect_equal(rxFromSE("llikXChisqDdf(i,x, nu)"), "llikXChisqDdf(i,x,nu)")


    # Check the derivatives
    expect_equal(rxFromSE("Derivative(llikChisq(x,nu),nu)"),"llikChisqDdf(x, nu)")
    expect_equal(rxFromSE("Derivative(llikXChisq(1,x,nu),nu)"),"llikXChisqDdf(1, x, nu)")

    # Check rxode2 internals with R exported
    et <- et(1:3)
    et$x <- 1

    model <- rxode2({
      fx <- llikChisq(x, time)
      dDf <- llikChisqDdf(x, time)
    })

    modelX <- rxode2({
      fx <- llikXChisq(1, x, time)
      dDf <- llikXChisqDdf(1, x, time)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikChisq(et$x,et$time, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dDf, fromOde$dDf)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dDf, fromOdeX$dDf)


    expect_equal(fromR$fx, dchisq(1, fromOde$time, log=TRUE))

  })

  test_that("log-liklihood tests for exponential (including derivatives)", {
    # Make sure they compile:
    expect_error(rxode2("tmp=llikExp(x, nu)"), NA)
    expect_error(rxode2("tmp=llikExpDrate(x, nu)"), NA)

    expect_error(rxode2("tmp=llikXExp(1, x, nu)"), NA)
    expect_error(rxode2("tmp=llikXExpDrate(1, x, nu)"), NA)

    # Make sure they translate correctly:
    expect_equal(rxToSE("llikExp(x, nu)"), "llikExp(x,nu)")
    expect_equal(rxFromSE("llikExp(x, nu)"), "llikExp(x,nu)")

    expect_equal(rxToSE("llikExpDrate(x, nu)"), "llikExpDrate(x,nu)")
    expect_equal(rxFromSE("llikExpDrate(x, nu)"), "llikExpDrate(x,nu)")
    #
    expect_equal(rxToSE("llikXExp(i,x, nu)"), "llikXExp(i,x,nu)")
    expect_equal(rxFromSE("llikXExp(i,x, nu)"), "llikXExp(i,x,nu)")

    expect_equal(rxToSE("llikXExpDrate(i,x, nu)"), "llikXExpDrate(i,x,nu)")
    expect_equal(rxFromSE("llikXExpDrate(i,x, nu)"), "llikXExpDrate(i,x,nu)")

    # Check the derivatives
    expect_equal(rxFromSE("Derivative(llikExp(x,nu),nu)"),"llikExpDrate(x, nu)")
    expect_equal(rxFromSE("Derivative(llikXExp(i,x,nu),nu)"),"llikXExpDrate(i, x, nu)")

    # Check rxode2 internals with R exported
    et <- et(1:3)
    et$x <- 1

    model <- rxode2({
      fx <- llikExp(x, time)
      dRate <- llikExpDrate(x, time)
    })

    modelX <- rxode2({
      fx <- llikXExp(1, x, time)
      dRate <- llikXExpDrate(1, x, time)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikExp(et$x,et$time, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dRate, fromOde$dRate)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dRate, fromOdeX$dRate)

    expect_equal(fromR$fx, dexp(1, fromOde$time, log=TRUE))
  })


  test_that("log-liklihood tests for f (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikF(x, df1, df2)"), NA)
    expect_error(rxode2("tmp=llikFDdf1(x, df1, df2)"), NA)
    expect_error(rxode2("tmp=llikFDdf2(x, df1, df2)"), NA)

    expect_error(rxode2("tmp=llikXF(1, x, df1, df2)"), NA)
    expect_error(rxode2("tmp=llikXFDdf1(1, x, df1, df2)"), NA)
    expect_error(rxode2("tmp=llikXFDdf2(1, x, df1, df2)"), NA)

    expect_equal(rxToSE("llikF(x, df1, df2)"), "llikF(x,df1,df2)")
    expect_equal(rxFromSE("llikF(x, df1, df2)"), "llikF(x,df1,df2)")

    expect_equal(rxToSE("llikFDdf1(x, df1, df2)"), "llikFDdf1(x,df1,df2)")
    expect_equal(rxFromSE("llikFDdf1(x, df1, df2)"), "llikFDdf1(x,df1,df2)")

    expect_equal(rxToSE("llikFDdf2(x, df1, df2)"), "llikFDdf2(x,df1,df2)")
    expect_equal(rxFromSE("llikFDdf2(x, df1, df2)"), "llikFDdf2(x,df1,df2)")

    #
    expect_equal(rxToSE("llikXF(i,x, df1, df2)"), "llikXF(i,x,df1,df2)")
    expect_equal(rxFromSE("llikXF(i,x, df1, df2)"), "llikXF(i,x,df1,df2)")

    expect_equal(rxToSE("llikXFDdf1(i,x, df1, df2)"), "llikXFDdf1(i,x,df1,df2)")
    expect_equal(rxFromSE("llikXFDdf1(i,x, df1, df2)"), "llikXFDdf1(i,x,df1,df2)")

    expect_equal(rxToSE("llikXFDdf2(i,x, df1, df2)"), "llikXFDdf2(i,x,df1,df2)")
    expect_equal(rxFromSE("llikXFDdf2(i,x, df1, df2)"), "llikXFDdf2(i,x,df1,df2)")

    # Check the derivatives
    expect_equal(rxFromSE("Derivative(llikF(x,df1,df2),df1)"),"llikFDdf1(x, df1, df2)")
    expect_equal(rxFromSE("Derivative(llikF(x,df1,df2),df2)"),"llikFDdf2(x, df1, df2)")

    expect_equal(rxFromSE("Derivative(llikXF(1,x,df1,df2),df1)"),"llikXFDdf1(1, x, df1, df2)")
    expect_equal(rxFromSE("Derivative(llikXF(1,x,df1,df2),df2)"),"llikXFDdf2(1, x, df1, df2)")

    # Check rxode2 internals with R exported

    x <- seq(0.001, 5, length.out = 100)

    et <- et(x)
    et$df1 <- 1
    et$df2 <- 5

    model <- rxode2({
      fx <- llikF(time, df1, df2)
      dDf1 <- llikFDdf1(time, df1, df2)
      dDf2 <- llikFDdf2(time, df1, df2)
    })

    modelX <- rxode2({
      fx <- llikXF(1, time, df1, df2)
      dDf1 <- llikXFDdf1(1, time, df1, df2)
      dDf2 <- llikXFDdf2(1, time, df1, df2)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikF(et$time,et$df1, et$df2, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dDf1, fromOde$dDf1)
    expect_equal(fromR$dDf2, fromOde$dDf2)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dDf1, fromOdeX$dDf1)
    expect_equal(fromR$dDf2, fromOdeX$dDf2)

    expect_equal(fromR$fx, df(et$time, 1, 5, log=TRUE))

  })


  test_that("log-liklihood tests for geom (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikGeom(x, p)"), NA)
    expect_error(rxode2("tmp=llikGeomDprob(x, p)"), NA)

    expect_error(rxode2("tmp=llikXGeom(1, x, p)"), NA)
    expect_error(rxode2("tmp=llikXGeomDprob(1, x, p)"), NA)

    expect_equal(rxToSE("llikGeom(x, p)"), "llikGeom(x,p)")
    expect_equal(rxFromSE("llikGeom(x, p)"), "llikGeom(x,p)")

    expect_equal(rxToSE("llikGeomDprob(x, p)"), "llikGeomDprob(x,p)")
    expect_equal(rxFromSE("llikGeomDprob(x, p)"), "llikGeomDprob(x,p)")

    #
    expect_equal(rxToSE("llikGeom(x, p)"), "llikGeom(x,p)")
    expect_equal(rxFromSE("llikGeom(x, p)"), "llikGeom(x,p)")

    expect_equal(rxToSE("llikXGeomDprob(i, x, p)"), "llikXGeomDprob(i,x,p)")
    expect_equal(rxFromSE("llikXGeomDprob(i, x, p)"), "llikXGeomDprob(i,x,p)")


    # Check the derivatives
    expect_equal(rxFromSE("Derivative(llikGeom(x,p),p)"),"llikGeomDprob(x, p)")
    expect_equal(rxFromSE("Derivative(llikXGeom(i,x,p),p)"),"llikXGeomDprob(i, x, p)")

    # Check rxode2 internals with R exported
    et  <- et(1:10)
    et$prob <- 0.2

    model <- rxode2({
      fx <- llikGeom(time, prob)
      dProb <- llikGeomDprob(time, prob)
    })

    modelX <- rxode2({
      fx <- llikXGeom(1, time, prob)
      dProb <- llikXGeomDprob(1, time, prob)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikGeom(et$time, et$prob, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dProb, fromOde$dProb)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dProb, fromOdeX$dProb)

    expect_equal(fromR$fx, dgeom(et$time, 0.2, log=TRUE))
  })


  test_that("log-liklihood tests for unif (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikUnif(x, alpha, beta)"), NA)
    expect_error(rxode2("tmp=llikUnifDalpha(x, alpha, beta)"), NA)
    expect_error(rxode2("tmp=llikUnifDbeta(x, alpha, beta)"), NA)

    expect_error(rxode2("tmp=llikXUnif(i, x, alpha, beta)"), NA)
    expect_error(rxode2("tmp=llikXUnifDalpha(i, x, alpha, beta)"), NA)
    expect_error(rxode2("tmp=llikXUnifDbeta(i, x, alpha, beta)"), NA)

    expect_equal(rxToSE("llikUnif(x, alpha, beta)"), "llikUnif(x,alpha,beta)")
    expect_equal(rxFromSE("llikUnif(x, alpha, beta)"), "llikUnif(x,alpha,beta)")

    expect_equal(rxToSE("llikUnifDalpha(x, alpha, beta)"), "llikUnifDalpha(x,alpha,beta)")
    expect_equal(rxFromSE("llikUnifDalpha(x, alpha, beta)"), "llikUnifDalpha(x,alpha,beta)")

    expect_equal(rxToSE("llikUnifDbeta(x, alpha, beta)"), "llikUnifDbeta(x,alpha,beta)")
    expect_equal(rxFromSE("llikUnifDbeta(x, alpha, beta)"), "llikUnifDbeta(x,alpha,beta)")

    #
    expect_equal(rxToSE("llikXUnif(x,i, alpha, beta)"), "llikXUnif(x,i,alpha,beta)")
    expect_equal(rxFromSE("llikXUnif(x,i, alpha, beta)"), "llikXUnif(x,i,alpha,beta)")

    expect_equal(rxToSE("llikXUnifDalpha(x,i, alpha, beta)"), "llikXUnifDalpha(x,i,alpha,beta)")
    expect_equal(rxFromSE("llikXUnifDalpha(x,i, alpha, beta)"), "llikXUnifDalpha(x,i,alpha,beta)")

    expect_equal(rxToSE("llikXUnifDbeta(x,i, alpha, beta)"), "llikXUnifDbeta(x,i,alpha,beta)")
    expect_equal(rxFromSE("llikXUnifDbeta(x,i, alpha, beta)"), "llikXUnifDbeta(x,i,alpha,beta)")

    # Check the derivatives
    expect_equal(rxFromSE("Derivative(llikUnif(x,alpha, beta),alpha)"),"llikUnifDalpha(x, alpha, beta)")
    expect_equal(rxFromSE("Derivative(llikUnif(x,alpha, beta),beta)"),"llikUnifDbeta(x, alpha, beta)")

    expect_equal(rxFromSE("Derivative(llikXUnif(i,x,alpha, beta),alpha)"),
                 "llikXUnifDalpha(i, x, alpha, beta)")
    expect_equal(rxFromSE("Derivative(llikXUnif(i,x,alpha, beta),beta)"),
                 "llikXUnifDbeta(i, x, alpha, beta)")

    # Check rxode2 internals with R exported
    et  <- et(seq(-4,4, length.out=10))
    et$alpha <- -2
    et$beta <- 2

    model <- rxode2({
      fx <- llikUnif(time, alpha, beta)
      dAlpha<- llikUnifDalpha(time, alpha, beta)
      dBeta <- llikUnifDbeta(time, alpha, beta)
    })

    modelX <- rxode2({
      fx <- llikXUnif(1, time, alpha, beta)
      dAlpha<- llikXUnifDalpha(1, time, alpha, beta)
      dBeta <- llikXUnifDbeta(1, time, alpha, beta)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikUnif(et$time, -2, 2, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dAlpha, fromOde$dAlpha)
    expect_equal(fromR$dBeta, fromOde$dBeta)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dAlpha, fromOdeX$dAlpha)
    expect_equal(fromR$dBeta, fromOdeX$dBeta)

    expect_equal(fromR$fx, dunif(et$time,-2, 2, log=TRUE))

  })


  test_that("log-liklihood tests for weibull (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikWeibull(x, shape, scale)"), NA)
    expect_error(rxode2("tmp=llikWeibullDshape(x, shape, scale)"), NA)
    expect_error(rxode2("tmp=llikWeibullDscale(x, shape, scale)"), NA)

    expect_error(rxode2("tmp=llikXWeibull(1, x, shape, scale)"), NA)
    expect_error(rxode2("tmp=llikXWeibullDshape(1, x, shape, scale)"), NA)
    expect_error(rxode2("tmp=llikXWeibullDscale(1, x, shape, scale)"), NA)

    expect_equal(rxToSE("llikWeibull(x, shape, scale)"), "llikWeibull(x,shape,scale)")
    expect_equal(rxFromSE("llikWeibull(x, shape, scale)"), "llikWeibull(x,shape,scale)")

    expect_equal(rxToSE("llikWeibullDshape(x, shape, scale)"), "llikWeibullDshape(x,shape,scale)")
    expect_equal(rxFromSE("llikWeibullDshape(x, shape, scale)"), "llikWeibullDshape(x,shape,scale)")

    expect_equal(rxToSE("llikWeibullDscale(x, shape, scale)"), "llikWeibullDscale(x,shape,scale)")
    expect_equal(rxFromSE("llikWeibullDscale(x, shape, scale)"), "llikWeibullDscale(x,shape,scale)")

    #
    expect_equal(rxToSE("llikXWeibull(i,x, shape, scale)"), "llikXWeibull(i,x,shape,scale)")
    expect_equal(rxFromSE("llikXWeibull(i,x, shape, scale)"), "llikXWeibull(i,x,shape,scale)")

    expect_equal(rxToSE("llikXWeibullDshape(i,x, shape, scale)"), "llikXWeibullDshape(i,x,shape,scale)")
    expect_equal(rxFromSE("llikXWeibullDshape(i,x, shape, scale)"), "llikXWeibullDshape(i,x,shape,scale)")

    expect_equal(rxToSE("llikXWeibullDscale(i,x, shape, scale)"), "llikXWeibullDscale(i,x,shape,scale)")
    expect_equal(rxFromSE("llikXWeibullDscale(i,x, shape, scale)"), "llikXWeibullDscale(i,x,shape,scale)")

    # Check the derivatives
    expect_equal(rxFromSE("Derivative(llikWeibull(x,shape, scale),shape)"),"llikWeibullDshape(x, shape, scale)")
    expect_equal(rxFromSE("Derivative(llikWeibull(x,shape, scale),scale)"),"llikWeibullDscale(x, shape, scale)")

    expect_equal(rxFromSE("Derivative(llikXWeibull(i, x, shape, scale),shape)"),
                 "llikXWeibullDshape(i, x, shape, scale)")

    expect_equal(rxFromSE("Derivative(llikXWeibull(i, x, shape, scale),scale)"),
                 "llikXWeibullDscale(i, x, shape, scale)")


    # Check rxode2 internals with R exported
    et  <- et(seq(0.01,4, length.out=10))
    et$shape <- 1
    et$scale <- 10

    model <- rxode2({
      fx <- llikWeibull(time, shape, scale)
      dShape<- llikWeibullDshape(time, shape, scale)
      dScale <- llikWeibullDscale(time, shape, scale)
    })

    modelX <- rxode2({
      fx <- llikXWeibull(1, time, shape, scale)
      dShape<- llikXWeibullDshape(1, time, shape, scale)
      dScale <- llikXWeibullDscale(1, time, shape, scale)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikWeibull(et$time, 1, 10, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dShape, fromOde$dShape)
    expect_equal(fromR$dScale, fromOde$dScale)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dShape, fromOdeX$dShape)
    expect_equal(fromR$dScale, fromOdeX$dScale)

    expect_equal(fromR$fx, dweibull(et$time, 1, 10, log=TRUE))

  })


  test_that("log-liklihood tests for gamma (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikGamma(x, shape, rate)"), NA)
    expect_error(rxode2("tmp=llikGammaDshape(x, shape, rate)"), NA)
    expect_error(rxode2("tmp=llikGammaDrate(x, shape, rate)"), NA)

    expect_error(rxode2("tmp=llikXGamma(1, x, shape, rate)"), NA)
    expect_error(rxode2("tmp=llikXGammaDshape(1, x, shape, rate)"), NA)
    expect_error(rxode2("tmp=llikXGammaDrate(1, x, shape, rate)"), NA)


    expect_equal(rxToSE("llikGamma(x, shape, rate)"), "llikGamma(x,shape,rate)")
    expect_equal(rxFromSE("llikGamma(x, shape, rate)"), "llikGamma(x,shape,rate)")

    expect_equal(rxToSE("llikGammaDshape(x, shape, rate)"), "llikGammaDshape(x,shape,rate)")
    expect_equal(rxFromSE("llikGammaDshape(x, shape, rate)"), "llikGammaDshape(x,shape,rate)")

    expect_equal(rxToSE("llikGammaDrate(x, shape, rate)"), "llikGammaDrate(x,shape,rate)")
    expect_equal(rxFromSE("llikGammaDrate(x, shape, rate)"), "llikGammaDrate(x,shape,rate)")

    #
    expect_equal(rxToSE("llikXGamma(x,i, shape, rate)"), "llikXGamma(x,i,shape,rate)")
    expect_equal(rxFromSE("llikXGamma(x,i, shape, rate)"), "llikXGamma(x,i,shape,rate)")

    expect_equal(rxToSE("llikXGammaDshape(x,i, shape, rate)"), "llikXGammaDshape(x,i,shape,rate)")
    expect_equal(rxFromSE("llikXGammaDshape(x,i, shape, rate)"), "llikXGammaDshape(x,i,shape,rate)")

    expect_equal(rxToSE("llikXGammaDrate(x,i, shape, rate)"), "llikXGammaDrate(x,i,shape,rate)")
    expect_equal(rxFromSE("llikXGammaDrate(x,i, shape, rate)"), "llikXGammaDrate(x,i,shape,rate)")

    # Check the derivatives
    expect_equal(rxFromSE("Derivative(llikGamma(x,shape, rate),shape)"),"llikGammaDshape(x, shape, rate)")
    expect_equal(rxFromSE("Derivative(llikGamma(x,shape, rate),rate)"),"llikGammaDrate(x, shape, rate)")

    expect_equal(rxFromSE("Derivative(llikXGamma(i, x, shape, rate),shape)"),
                 "llikXGammaDshape(i, x, shape, rate)")
    expect_equal(rxFromSE("Derivative(llikXGamma(i, x, shape, rate),rate)"),
                 "llikXGammaDrate(i, x, shape, rate)")


    # Check rxode2 internals with R exported
    et  <- et(seq(0.01,4, length.out=10))
    et$shape <- 1
    et$rate <- 10

    model <- rxode2({
      fx <- llikGamma(time, shape, rate)
      dShape<- llikGammaDshape(time, shape, rate)
      dRate <- llikGammaDrate(time, shape, rate)
    })

    modelX <- rxode2({
      fx <- llikXGamma(1, time, shape, rate)
      dShape<- llikXGammaDshape(1, time, shape, rate)
      dRate <- llikXGammaDrate(1, time, shape, rate)
    })

    fromOde <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikGamma(et$time, 1, 10, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dShape, fromOde$dShape)
    expect_equal(fromR$dRate, fromOde$dRate)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dShape, fromOdeX$dShape)
    expect_equal(fromR$dRate, fromOdeX$dRate)

    expect_equal(fromR$fx, dgamma(et$time, 1, rate=10, log=TRUE))

    et  <- et(seq(0.01, 4, length.out = 10)) |>
      et(id = 1:200)
    et$shape <- 1
    et$rate <- 10

    model <- rxode2({
      fx <- llikGamma(time, shape, rate)
      dShape<- llikGammaDshape(time, shape, rate)
      dRate <- llikGammaDrate(time, shape, rate)
    })

    modelX <- rxode2({
      fx <- llikXGamma(1, time, shape, rate)
      dShape<- llikXGammaDshape(1, time, shape, rate)
      dRate <- llikXGammaDrate(1, time, shape, rate)
    })

    expect_error(rxSolve(modelX, et, cores=2), NA)

  })


  test_that("log-liklihood tests for cauchy (including derivatives)", {

    # Make sure they compile:
    expect_error(rxode2("tmp=llikCauchy(x, location, scale)"), NA)
    expect_error(rxode2("tmp=llikCauchyDlocation(x, location, scale)"), NA)
    expect_error(rxode2("tmp=llikCauchyDscale(x, location, scale)"), NA)

    expect_error(rxode2("tmp=llikXCauchy(1, x, location, scale)"), NA)
    expect_error(rxode2("tmp=llikXCauchyDlocation(1, x, location, scale)"), NA)
    expect_error(rxode2("tmp=llikXCauchyDscale(1, x, location, scale)"), NA)


    expect_equal(rxToSE("llikCauchy(x, location, scale)"), "llikCauchy(x,location,scale)")
    expect_equal(rxFromSE("llikCauchy(x, location, scale)"), "llikCauchy(x,location,scale)")

    expect_equal(rxToSE("llikCauchyDlocation(x, location, scale)"), "llikCauchyDlocation(x,location,scale)")
    expect_equal(rxFromSE("llikCauchyDlocation(x, location, scale)"), "llikCauchyDlocation(x,location,scale)")

    expect_equal(rxToSE("llikCauchyDscale(x, location, scale)"), "llikCauchyDscale(x,location,scale)")
    expect_equal(rxFromSE("llikCauchyDscale(x, location, scale)"), "llikCauchyDscale(x,location,scale)")

    #
    expect_equal(rxToSE("llikXCauchy(i,x, location, scale)"), "llikXCauchy(i,x,location,scale)")
    expect_equal(rxFromSE("llikXCauchy(i,x, location, scale)"), "llikXCauchy(i,x,location,scale)")

    expect_equal(rxToSE("llikXCauchyDlocation(i,x, location, scale)"), "llikXCauchyDlocation(i,x,location,scale)")
    expect_equal(rxFromSE("llikXCauchyDlocation(i,x, location, scale)"), "llikXCauchyDlocation(i,x,location,scale)")

    expect_equal(rxToSE("llikXCauchyDscale(i,x, location, scale)"), "llikXCauchyDscale(i,x,location,scale)")
    expect_equal(rxFromSE("llikXCauchyDscale(i,x, location, scale)"), "llikXCauchyDscale(i,x,location,scale)")


    # Check the derivatives
    expect_equal(rxFromSE("Derivative(llikCauchy(x,location, scale),location)"),"llikCauchyDlocation(x, location, scale)")
    expect_equal(rxFromSE("Derivative(llikCauchy(x,location, scale),scale)"),"llikCauchyDscale(x, location, scale)")

    #
    expect_equal(rxFromSE("Derivative(llikXCauchy(i, x, location, scale),location)"),
                 "llikXCauchyDlocation(i, x, location, scale)")
    expect_equal(rxFromSE("Derivative(llikXCauchy(i, x, location, scale),scale)"),
                 "llikXCauchyDscale(i, x, location, scale)")

    # Check rxode2 internals with R exported
    et  <- et(seq(0.01,4, length.out=10))
    et$location <- 1
    et$scale <- 10

    model <- rxode2({
      fx <- llikCauchy(time, location, scale)
      dLocation<- llikCauchyDlocation(time, location, scale)
      dScale <- llikCauchyDscale(time, location, scale)
    })

    modelX <- rxode2({
      fx <- llikXCauchy(1, time, location, scale)
      dLocation<- llikXCauchyDlocation(1, time, location, scale)
      dScale <- llikXCauchyDscale(1, time, location, scale)
    })

    fromOde  <- rxSolve(model, et)
    fromOdeX <- rxSolve(modelX, et)

    fromR <- llikCauchy(et$time, 1, 10, full=TRUE)

    expect_equal(fromR$fx, fromOde$fx)
    expect_equal(fromR$dLocation, fromOde$dLocation)
    expect_equal(fromR$dScale, fromOde$dScale)

    expect_equal(fromR$fx, fromOdeX$fx)
    expect_equal(fromR$dLocation, fromOdeX$dLocation)
    expect_equal(fromR$dScale, fromOdeX$dScale)

    expect_equal(fromR$fx, dcauchy(et$time, location=1, scale=10, log=TRUE))


    et  <- et(seq(0.01,4, length.out=10)) |>
      et(id=1:200)
    et$location <- 1
    et$scale <- 10

    model <- rxode2({
      fx <- llikCauchy(time, location, scale)
      dLocation<- llikCauchyDlocation(time, location, scale)
      dScale <- llikCauchyDscale(time, location, scale)
    })

    modelX <- rxode2({
      fx <- llikXCauchy(1, time, location, scale)
      dLocation<- llikXCauchyDlocation(1, time, location, scale)
      dScale <- llikXCauchyDscale(1, time, location, scale)
    })

    expect_error(rxSolve(model, et, cores=2), NA)
    expect_error(rxSolve(modelX, et, cores=2), NA)

  })
})

.rxIndLinStrategy <- "curState"
#' This sets the inductive linearization strategy for matrix building
#'
#' When there is more than one state in a ODE that cannot be
#' separated this specifies how it is incorporated into the matrix
#' exponential.
#'
#' @param strategy The strategy for inductive linearization matrix building
#'
#' * `curState` Prefer parameterizing in terms of the current
#'    state, followed by the first state observed in the term.
#' * `split` Split the parameterization between all states in the
#'   term by dividing each by the number of states in the term and then
#'   adding a matrix term for each state.
#'
#' @return Nothing
#' @author Matthew L. Fidler
#' @export
rxIndLinStrategy <- function(strategy = c("curState", "split")) {
  assignInMyNamespace(".rxIndLinStrategy", match.arg(strategy))
}

.rxIndLinState <- NULL
#' Set the preferred factoring by state
#'
#' @param preferred A list of each state's preferred factorization
#' @return Nothing
#' @author Matthew Fidler
#' @export
rxIndLinState <- function(preferred = NULL) {
  if (is.null(preferred)) {
    return(assignInMyNamespace(".rxIndLinState", preferred))
  }
  checkmate::assertList(preferred, names = "unique")
  lapply(seq_along(preferred), function(x) {
    if (!checkmate::checkCharacter(preferred[[x]],
      names = "unnamed"
    )) {
      stop(sprintf(gettext("'rxIndLinState' list element '%s' must be a unnamed character vector"), names(preferred)[x]),
        call. = FALSE
      )
    }
  })
  assignInMyNamespace(".rxIndLinState", preferred)
}

.rxIndLinLine <- function(line, states, state0) {
  .tmp <- tryCatch(
    symengine::expand(line), ## Expand line; simplifies X^1->X etc
    error = function(e) symengine::S(as.character(line)) ## fall back for complex exprs
  )
  .tmp <- rxFromSE(.tmp) ## Convert SE->rxode2; Changes things like X^1 -> X

  .ret <- eval(parse(text = paste0("rxSplitPlusQ(quote(", .tmp, "))")))
  .lst <- list()
  .fullIndLin <- FALSE
  ## This adds the x to the name environment in the .lst list above.
  .addIt <- function(x, name) {
    .cur <- .lst
    if (any(names(.lst) == name)) {
      .cur[[name]] <- c(.cur[[name]], x)
    } else {
      .cur[[name]] <- x
    }
    .lst <<- .cur
  }
  ## This fixes the multiplication so that -1*x becomes -x and *1* becomes *
  ## Also changes y^2*z*1/y to y^1 to y
  ## This collapses a character vector with "*" between it
  ## If we have *1/x this becomes simply /x
  .multCollapse <- function(x) {
    as.character(symengine::S(paste(x, collapse = "*")))
  }
  sapply(.ret, function(x) {
    .mult <- eval(parse(text = paste0("rxSplitPlusQ(quote(", x, "),mult=TRUE)")))
    .mult <- unlist(lapply(.mult, function(x) {
      if (x == "-1") {
        return(x)
      }
      .min <- substr(x, 0, 1)
      if (.min == "-") {
        c("-1", substr(x, 2, nchar(x)))
      } else {
        return(x)
      }
    }))

    .curStates <- unlist(lapply(.mult, function(x) {
      .pars <- rxModelVars(paste0("rx_expr=", x))$params
      return(.pars[.pars %in% states])
    }))
    .addState <- function(.state, .mult) {
      .num <- which(.mult == .state)
      if (length(.num) == 0) {
        .fullIndLin <<- TRUE
        .rest <- c(.mult, paste0("1/", .state))
        .expr <- .multCollapse(.rest)
        .addIt(.expr, .state)
      } else {
        if (length(.num) > 1) .fullIndLin <<- TRUE
        .num <- .num[1]
        .rest <- .mult[-.num]
        if (length(.rest) == 0) {
          .addIt("1", .state)
        } else if (length(.rest) == 1 && .rest[1] == "-1") {
          .addIt("-1", .state)
        } else {
          if (!.fullIndLin) {
            ## Check to see if this is inductive or
            ## depends on other states.
            .otherStates <- unlist(lapply(.rest, function(x) {
              .pars <- rxModelVars(paste0("rx_expr=", x))$params
              return(.pars[.pars %in% states])
            }))
            if (length(.otherStates) > 0) .fullIndLin <<- TRUE
          }
          .expr <- .multCollapse(.rest)
          .addIt(.expr, .state)
        }
      }
    }
    if (length(.curStates) == 1) {
      .addState(.curStates, .mult)
    } else if (length(.curStates) > 1) {
      if (.rxIndLinStrategy == "split") {
        ## Use strategy #3, split between all the compartments
        .extra <- paste0("1/", length(.curStates))
        for (.s in .curStates) {
          .addState(.s, c(.mult, .extra))
        }
      } else {
        .pref <- .rxIndLinState[[state0]]
        .addPref <- FALSE
        if (!is.null(.pref)) {
          .pref <- intersect(.pref, .curStates)
          if (length(.pref) > 0) {
            .addState(.pref[1], .mult)
            .addPref <- TRUE
          }
        }
        if (!.addPref) {
          if (any(.curStates == state0)) {
            ## If there is d/dt(state1) = ... (state1*state2*state3) ...
            ## or some other complex expression prefer expressing
            .addState(state0, .mult)
          } else {
            ## Otherwise just use the first state identified.
            ## Strategy #1 just add the first
            .addState(.curStates[1], .mult)
          }
        }
      }
    } else {
      ## This is some "constant" or time-base experssion that
      ## does not depend on state.  Hence it is the extra constant
      .expr <- .multCollapse(.mult)
      .addIt(.expr, "_rxF")
    }
  })
  .ret <- sapply(c(states, "_rxF"), function(x) {
    if (any(names(.lst) == x)) {
      return(gsub("[+][-]", "-", paste(.lst[[x]], collapse = "+")))
    }
    return("0")
  })
  rxTick()
  return(c(.ret, paste0(.fullIndLin)))
}

.indLinInfo <- list()

#' This creates the inductive linearization pieces to integrate into
#' a rxode2 model.
#'
#' @param model rxode2 model type of object
#'
#' @param doConst Replace constants with values; By default this is
#'     `FALSE`.
#'
#' @return List:
#'
#' *  Matrix Exponential initial matrix A for exp(t*A)
#'
#' * Inductive Linerization vector for F
#'
#' * Extra rxode2 code for model generation
#'
#' * Generated C code for model variables; With ME only this will
#' be a list of size 1, otherwise it is a list of size 2.
#'
#' @author Matthew Fidler
#' @noRd
.rxIndLin <- function(model, doConst = FALSE) {
  rxReq("symengine")
  .env <- .rxLoadPrune(model, doConst = doConst)
  .states <- rxState(.env)
  rxProgress(length(.states))
  .minfo("create inductive linearization matrices")
  on.exit({
    rxProgressAbort()
  })
  .ret <- eval(parse(text = rxIndLin_(.states)))
  .w <- setNames(which(.ret[, "indLin"] == "TRUE") - 1, NULL)
  .fullIndLin <- length(.w) > 0
  .ret0 <- .ret[.states, .states, drop = FALSE]
  .ret1 <- .ret[, "_rxF", drop = FALSE]
  .code <- c(paste0("_rxM=", as.vector(.ret0), ";"))
  if (all(.ret1 == "0")) {
    .ret <- list(
      A = .ret0,
      f = NULL,
      fullIndLin = .fullIndLin,
      wIndLin = .w
    )
  } else {
    .ret <- list(
      A = .ret0,
      f = .ret1,
      fullIndLin = .fullIndLin,
      wIndLin = .w
    )
    .code <- c(.code, paste("_rxF=", as.vector(.ret1)))
  }
  assignInMyNamespace(".indLinInfo", .ret)
  rxProgressStop()
  .malert("indLin is in development and results subject to change")
  return(paste(.code, collapse = "\n"))
}

#' Rectified Linear Unit (ReLU) Activation Function
#'
#' This function applies the Rectified Linear Unit (ReLU) activation function to the input numeric vector.
#' The ReLU function is defined as the positive part of its argument: \eqn{f(x) = max(0, x)}.
#'
#' @family Activation Functions
#' @param x A numeric vector. All elements must be finite and
#'   non-missing.
#' @return A numeric vector where the ReLU function has been applied
#'   to each element of `x`.
#' @author Matthew Fidler
#' @export
#' @examples
#'
#' ReLU(c(-1, 0, 1, 2))
#'
#' \donttest{
#' # Can also be used in rxode2:
#' x <- rxode2({
#'    r=ReLU(time)
#' })
#'
#' e <- et(c(-1, 0, 1, 2))
#'
#' rxSolve(x, e)
#' }
#'
ReLU <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 2L)
}
#' Derivative of the Rectified Linear Unit (ReLU) Activation Function
#'
#' This function applies the derivative of the Rectified Linear Unit
#' (ReLU) activation function to the input numeric vector.
#'
#' @param x A numeric vector. All elements must be finite and
#'  non-missing.
#'
#' @family Activation Functions
#'
#' @return A numeric vector where the derivative of the ReLU function
#' @export
#' @examples
#'
#' dReLU(c(-1, 0, 1, 2))
#'
#' \donttest{
#' # Can also be used in rxode2:
#' x <- rxode2({
#'    r=dReLU(time)
#' })
#'
#' e <- et(c(-1, 0, 1, 2))
#'
#' rxSolve(x, e)
#' }
dReLU <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 7L)
}

# backward-compatible list
.rxUiBackward <- c(
  "model.desc"="modelDesc",
  "fun.txt"="funTxt",
  "all.covs"="allCovs"
)

#' Convert rxode2 UI object to object for `rxUiGet`
#'
#' @param obj rxode2 ui object
#' @param arg argument that you are trying to get from rxui
#' @param exact exact argument
#' @return object for `rxUiGet`
#' @author Matthew L. Fidler
#' @noRd
.uiToRxUiGet <- function(obj, arg, exact=TRUE) {
  if (is.list(obj) || inherits(obj, "raw")) obj <- rxUiDecompress(obj)
  .lst <- list(obj, exact)
  .arg <- .rxUiBackward[arg]
  if (is.na(.arg)) .arg <- arg
  class(.lst) <- c(.arg, "rxUiGet")
  .lst
}

#' @export
`$.rxUi` <- function(obj, arg, exact = TRUE) {
  # need to assign environment correctly for UDF
  #
  # The model() and rxode2() assign the parent environments for UDF
  # parsing, if the object is in that environment lock it and then
  # unlock on exit
  .udfEnvSet(list(parent.frame(1), parent.frame(2)))
  .obj <- .uiToRxUiGet(obj=obj, arg=arg, exact=exact)
  if (.rstudioComplete()){
    # If Rstudio is running completion, then we need to simply
    # return a dummy object so it doesn't calculate the value.
    #
    # However, if the object actually exists, then use the rxUiGet.default method
    # to get the value.
    .v <- as.character(utils::methods("rxUiGet"))
    .cls <- class(.obj)[1]
    .method <- paste0("rxUiGet.", .cls)
    if (.method %in% .v) {
      # If there is a rstudio value in the method, assume that is what you
      # wish to return for the rstudio auto-completion method
      .rstudio <- attr(utils::getS3method("rxUiGet", .cls), "rstudio")
      if (is.null(.rstudio)) {
        return(list("calculated value"))
      } else if (is.na(.rstudio)) {
        # If it is NA, then return the value
        return(rxUiGet(.obj))
      } else {
        return(.rstudio)
      }
    }
  }
  rxUiGet(.obj)
}

#' S3 for getting information from UI model
#'
#' @param x list of (UIenvironment, exact).  UI environment is the
#'   parsed function for rxode2.  `exact` is a boolean that says if an
#'   exact match is required.
#' @param ... Other arguments
#' @return value that was requested from the UI object
#' @author Matthew Fidler
#' @export
rxUiGet <- function(x, ...) {
  if (!inherits(x, "rxUiGet")) {
    stop("object is wrong type for `rxUiGet`")
  }
  UseMethod("rxUiGet")
}

#' @rdname rxUiGet
#' @export
rxUiGet.levels <- function(x, ...) {
  .x <- x[[1]]
  .mv <- rxModelVars(.x)
  .str <- .mv$strAssign
  .names <- names(.str)
  lapply(vapply(seq_along(.str), function(i) {
    paste0("levels(", .names[i], ") <- ",
           deparse1(.str[[i]]))
  }, character(1), USE.NAMES=FALSE),
  str2lang)
}
attr(rxUiGet.levels, "rstudio") <- quote(levels(a))

#' @rdname rxUiGet
#' @export
rxUiGet.state <- function(x, ...) {
  .ui <- x[[1]]
  rxModelVars(.ui)$state
}
attr(rxUiGet.state, "desc") <- "states associated with the model (in order)"
attr(rxUiGet.state, "rstudio") <- "state" # character

#' @rdname rxUiGet
#' @export
rxUiGet.stateDf <- function(x, ...) {
  .ui <- x[[1]]
  .mv <- rxModelVars(.ui)
  .state <- .mv$state
  .cmt <- .getCmtNum(.mv)
  if (.mv$flags["linCmtFlg"] != 0) {
    data.frame("Compartment Number"=seq_along(.cmt),
               "Compartment Name"=names(.cmt),
               "Rate"=cmtSupportsInfusion_(.cmt, .mv),
               "Off"=cmtSupportsOff_(.cmt, .mv),
               "Internal #"=setNames(.cmt,NULL),
               check.names=FALSE)
  } else {
    data.frame("Compartment Number"=seq_along(.cmt),
               "Compartment Name"=names(.cmt),
               check.names=FALSE)
  }
}
attr(rxUiGet.stateDf, "desc") <- "states and cmt number data.frame"
attr(rxUiGet.stateDf, "rstudio") <- NA # passthrough

#' @export
#' @rdname rxUiGet
rxUiGet.statePropDf <- function(x,...) {
  .ui <- x[[1]]
  .mv <- rxModelVars(.ui)
  do.call(rbind, lapply(seq_along(.mv$stateProp),
                 function(i) {
                   .prop <- .mv$stateProp[i]
                   if (length(.prop) != 1) return(NULL)
                   if (.prop == 0) return(NULL)
                   .name <- names(.mv$stateProp)[i]
                   .props <- character(0)
                   if (bitwAnd(.prop, 1)) {
                     .props <- c(.props, "ini")
                   }
                   if (bitwAnd(.prop, 2)) {
                     .props <- c(.props, "f")
                   }
                   if (bitwAnd(.prop, 4)) {
                     .props <- c(.props, "alag")
                   }
                   if (bitwAnd(.prop, 8)) {
                     .props <- c(.props, "rate")
                   }
                   if (bitwAnd(.prop, 16)) {
                     .props <- c(.props, "dur")
                   }
                   if (length(.props) == 0) return(NULL)
                   data.frame("Compartment"=.name,
                              "Property"=.props)
                 }))
}
attr(rxUiGet.statePropDf, "rstudio") <- NA

#' @export
#' @rdname rxUiGet
rxUiGet.props <- function(x, ...) {
  .x <- x[[1]]
  .ini <- .x$iniDf
  .w <- !is.na(.ini$ntheta) & is.na(.ini$err)
  .pop <- .ini$name[.w]
  .w <- !is.na(.ini$ntheta) & !is.na(.ini$err)
  .resid <- .ini$name[.w]
  .w <- !is.na(.ini$neta1)
  .cnds <- unique(.ini$condition[.w])
  .var <- lapply(.cnds,
                 function(cnd) {
                   .w <- which(.ini$condition == cnd &
                                 .ini$neta1 == .ini$neta2)
                   .ini$name[.w]
                 })
  .mv <- rxGetModel(.x)
  .lin <- FALSE
  .doseExtra <- character(0)
  .mv <- rxModelVars(.x)
  if (!is.null(.x$.linCmtM)) {
    .lin <- TRUE
    if (.mv$extraCmt == 2L) {
      .doseExtra <- c("depot", "central")
    } else if (.mv$extraCmt == 1L) {
      .doseExtra <- "central"
    }
  }
  .predDf <- .x$predDf
  if (!.lin && any(.predDf$linCmt)) {
    .lin <- TRUE
    if (.mv$flags["ka"] == 1L) {
      .doseExtra <- c("depot", "central")
    } else {
      .doseExtra <- "central"
    }
  }
  .dose <- c(.doseExtra, .x$state)
  names(.var) <- .cnds
  .lhs <- .mv$lhs
  .state <- .mv$state
  .end <- .x$predDf$var
  .end <- .end[.end %in% c(.lhs, .state)]
  .lhs <- .lhs[!(.lhs %in% .end)]
  .varLhs <- .x$varLhs
  .primary <- .lhs[.lhs %in% .varLhs]
  .secondary <- .lhs[!(.lhs %in% .primary)]
  list(pop=.pop,
       resid=.resid,
       group=.var,
       linCmt=.lin,
       cmt=.dose,
       output=list(primary=.primary,
                   secondary=.secondary,
                   endpoint=.end,
                   state=.x$state),
       cmtProp=rxUiGet.statePropDf(x,...))
}
attr(rxUiGet.props, "desc") <- "rxode2 model properties"

#' @export
#' @rdname rxUiGet
rxUiGet.theta <- function(x, ...) {
  .x <- x[[1]]
  .ini <- .x$iniDf
  .w <- !is.na(.ini$ntheta)
  setNames(.ini$est[.w], .ini$name[.w])
}
attr(rxUiGet.theta, "desc") <- "Initial Population/Fixed Effects estimates, theta"
attr(rxUiGet.theta, "rstudio") <- c("theta"=1.0) # named vector

#' @export
#' @rdname rxUiGet
rxUiGet.lstChr <- function(x, ...) {
  vapply(get("lstExpr", envir=x[[1]]),
         function(x) {
           deparse1(x)
         }, character(1), USE.NAMES=FALSE)
}
#attr(rxUiGet.lstChr, "desc") <- "Get a character vector of the model expressions (by line)"

#' @export
#' @rdname rxUiGet
rxUiGet.omega <- function(x, ...) {
  .x <- x[[1]]
  .lotri <- lotri::as.lotri(.x$iniDf)
  if (inherits(.lotri, "matrix")) {
    attr(.lotri, "lotriEst") <- NULL
    class(.lotri) <- NULL
  } else {
    attr(.lotri, "lotriEst") <- NULL
    if (length(.lotri) == 0) {
      .lotri <- NULL
    } else if (!inherits(.lotri, "lotri")) {
      class(.lotri) <- c("lotri", class(.lotri))
    }
  }
  .lotri
}
attr(rxUiGet.omega, "desc") <- "Initial Random Effects variability matrix, omega"
attr(rxUiGet.omega, "rstudio") <- lotri::lotri(a+b ~ c(1, .1, 1))

#' @export
#' @rdname rxUiGet
rxUiGet.funTxt <- function(x, ...) {
  paste(rxUiGet.lstChr(x, ...), collapse="\n")
}
attr(rxUiGet.funTxt, "desc") <- "Get function text for the model({}) block"
attr(rxUiGet.funTxt, "rstudio") <- "model text" # character

#' @export
#' @rdname rxUiGet
rxUiGet.allCovs <- function(x, ...) {
  get("covariates", envir=x[[1]])
}
attr(rxUiGet.allCovs, "desc") <- "Get all covariates defined in the model"
attr(rxUiGet.allCovs, "rstudio") <- "covariates" # character

#' @export
#' @rdname rxUiGet
rxUiGet.muRefTable <- function(x, ...) {
  .x <- x[[1]]
  .exact <- x[[2]]
  .muRef <- get("muRefDataFrame", .x)
  if (length(.muRef$theta) == 0) return(NULL)
  .muRefCov <- rbind(get("muRefCovariateDataFrame", .x),
                     get("mu2RefCovariateReplaceDataFrame", .x)[,c("theta", "covariate", "covariateParameter")])
  if (length(.muRefCov$theta) > 0) {
    .env <- new.env(parent=emptyenv())
    lapply(seq_along(.muRefCov$theta), function(i) {
      .theta <- .muRefCov$theta[i]
      .cov <- paste0(.muRefCov$covariate[i], "*", .muRefCov$covariateParameter[i])
      if (exists(.theta, .env)) {
        assign(.theta, c(get(.theta, .env), .cov), .env)
      } else {
        assign(.theta, .cov, .env)
      }
    })
    .muRef$covariates <- vapply(.muRef$theta, function(theta) {
      if (exists(theta, .env)) {
        return(paste(get(theta, .env), collapse=" + "))
      }
      return("")
    }, character(1), USE.NAMES=FALSE)
  }
  .muRef
}
attr(rxUiGet.muRefTable, "desc") <- "table of mu-referenced items in a model"
attr(rxUiGet.muRefTable, "rstudio") <- NA # passthrough

#' @rdname rxUiGet
#' @export
rxUiGet.multipleEndpoint <- function(x, ...) {
  .x <- x[[1]]
  .exact <- x[[2]]
  .info <- get("predDf", .x)
  if (is.null(.info)) {
    return(invisible())
  }
  if (length(.info$cond) == 1) return(NULL)
  if (getOption("rxode2.combine.dvid", TRUE)) {
    .info <- .info[order(.info$dvid), ]
  }
  .info <- with(.info, data.frame(
    variable = paste(var, "~", ifelse(use.utf(), "\u2026", "...")),
    cmt = paste0("cmt='", cond, "' or cmt=", cmt),
    "dvid*" = ifelse(is.na(dvid), "",
                     paste0("dvid='", cond, "' or dvid=", dvid)),
    check.names = FALSE,
    stringsAsFactors=FALSE))
  if (!getOption("rxode2.combine.dvid", TRUE)) {
    .info <- .info[, names(.info) != "dvid*"]
  }
  .info
}
attr(rxUiGet.multipleEndpoint, "desc") <- "table of multiple endpoint translations"
attr(rxUiGet.multipleEndpoint, "rstudio") <- NA
#' This is a generic function for deparsing certain objects when
#' printing out a rxode2 object.  Currently it is used for any meta-information
#'
#' @param object object to be deparsed
#' @param var variable name to be assigned
#' @return parsed R expression that can be used for printing and
#'   `as.function()` calls.
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' mat <- matrix(c(1, 0.1, 0.1, 1), 2, 2, dimnames=list(c("a", "b"), c("a", "b")))
#'
#' rxUiDeparse(matrix(c(1, 0.1, 0.1, 1), 2, 2, dimnames=list(c("a", "b"), c("a", "b"))), "x")
rxUiDeparse <- function(object, var) {
 UseMethod("rxUiDeparse")
}

#' @rdname rxUiDeparse
#' @export
rxUiDeparse.lotriFix <- function(object, var) {
  .val <- lotri::lotriAsExpression(object)
  bquote(.(str2lang(var)) <- .(.val))
}

#' @rdname rxUiDeparse
#' @export
rxUiDeparse.default <- function(object, var) {
  # This is a default method for deparsing objects
  if (checkmate::testMatrix(object, any.missing=FALSE,
                            row.names="strict", col.names="strict")) {
    .dn <- dimnames(object)
    if (identical(.dn[[1]], .dn[[2]]) && isSymmetric(object)) {
      return(rxUiDeparse.lotriFix(object, var))
    }
  }
  .ret <- try(str2lang(paste0(var, "<-", deparse1(object))), silent=TRUE)
  if (inherits(.ret, "try-error")) {
    .ret <- str2lang("NULL")
  }
  .ret
}

#' @rdname rxUiGet
#' @export
rxUiGet.funPrint <- function(x, ...) {
  .x <- x[[1]]
  .ls <- ls(.x$meta, all.names=TRUE)
  .hasIni <- length(.x$iniDf$cond) > 0
  .ret <- vector("list", length(.ls) + ifelse(.hasIni, 3, 2))
  .ret[[1]] <- quote(`{`)
  for (.i in seq_along(.ls)) {
    if (.i == ".simModelBase") {
      next
    }
    .var <- .ls[.i]
    .val <- .x$meta[[.ls[.i]]]
    .ret[[.i + 1]] <- rxUiDeparse(.val, .var)
  }
  .theta <- x$theta
  .omega <- x$omega
  if (.hasIni) {
    .len <- length(.ls)
    .ret[[.len + 2]] <- .x$iniFun
    .ret[[.len + 3]] <- .x$modelFun
  } else {
    .len <- length(.ls)
    .ret[[.len + 2]] <- .x$modelFun
  }
  .ret
}
attr(rxUiGet.funPrint, "desc") <- "Normalized, quoted model function (for printing)"

#' @export
#' @rdname rxUiGet
rxUiGet.fun <- function(x, ...) {
  .ret <- rxUiGet.funPrint(x, ...)
  .ret2 <- function() {

  }
  body(.ret2) <- as.call(.ret)
  .ret2
}
attr(rxUiGet.fun, "desc") <- "Normalized model function"
attr(rxUiGet.fun, "rstudio") <- function() {}

#' @export
#' @rdname rxUiGet
rxUiGet.funPartsDigest <- function(x, ...) {
  .ui <- x[[1]]
  list(
    normModel = .ui$mv0$model["normModel"],
    iniDf = .ui$iniDf,
    errLinesI = .ui$predDf$line,
    errLines = vapply(.ui$predDf$line, function(l) {
      deparse1(.ui$lstExpr[[l]])
    }, character(1), USE.NAMES=FALSE),
    # Now get environment specific differences in the model
    # This changes how models can be expressed (and their output)
    allow.ini=getOption("rxode2.syntax.allow.ini", TRUE),
    # Defined lower level functions and udf functions
    definedFuns=  ls(.udfEnv$symengineFs, all.names=TRUE),
    # Defined rxUdfUi methods
    uiFuns=as.character(utils::methods("rxUdfUi")),
    # Add version of rxode2
    rxVersion=.rxVersion
  )
}

#' @export
#' @rdname rxUiGet
rxUiGet.md5 <- function(x, ...) {
  digest::digest(rxUiGet.funPartsDigest(x, ...), algo="md5")
}
attr(rxUiGet.md5, "desc") <- "MD5 hash of the UI model"
attr(rxUiGet.md5, "rstudio") <- "md5 hash of the model"

#' @export
#' @rdname rxUiGet
rxUiGet.sha1 <- function(x, ...) {
  digest::digest(rxUiGet.funPartsDigest(x, ...), algo="sha1")
}
attr(rxUiGet.sha1, "desc") <- "SHA1 hash of the UI model"
attr(rxUiGet.sha1, "rstudio") <- "sha1 hash of the model"

sha1.rxUi <- function(x, digits = 14L, zapsmall = 7L, ..., algo = "sha1")  {
  digest::sha1(rxUiGet.funPartsDigest(list(x)),
               digits=digits, zapsmall=zapsmall, ..., algo=algo)
}

#' @export
#' @rdname rxUiGet
rxUiGet.ini <- function(x, ...) {
  get("iniDf", x[[1]])
}
attr(rxUiGet.ini, "desc") <- "Model initilizations/bounds object"
attr(rxUiGet.ini,"rstudio") <- NA

#'@export
#' @rdname rxUiGet
rxUiGet.iniFun <- function(x, ...) {
  .x <- x[[1]]
  .arg <- class(x)[1]
  lotri::lotriDataFrameToLotriExpression(.x$iniDf, useIni=TRUE)
}
attr(rxUiGet.iniFun, "desc") <- "normalized, quoted `ini()` block"
attr(rxUiGet.iniFun, "rstudio") <- quote(ini({}))

#' @export
#' @rdname rxUiGet
rxUiGet.modelFun <- function(x, ...) {
  .x <- x[[1]]
  bquote(model(.(as.call(c(quote(`{`),.x$lstExpr)))))
}
attr(rxUiGet.modelFun, "desc") <- "normalized, quoted `model()` block"
attr(rxUiGet.modelFun, "rstudio") <- quote(model({}))

#' @export
#' @rdname rxUiGet
rxUiGet.model <- rxUiGet.modelFun


#' @export
#' @rdname rxUiGet
rxUiGet.modelDesc <- function(x, ...) {
  .mv <- get("mv0", x[[1]])
  .mvL <- get("mvL", x[[1]])
  if (!is.null(.mvL)) {
    # With the new linear models, need to remove anything from them
    .rxUiLinCompartmentNames <-   c("depot",
                                    "central",
                                    "peripheral1",
                                    "peripheral2")
    .state <- .mvL$state[!(
      (.mvL$state %in% .rxUiLinCompartmentNames) |
      startsWith(.mvL$state, "rx__sens_")
    )]
    return(sprintf(
      "rxode2-based solved PK %s-compartment model%s%s", .mvL$flags["ncmt"],
      ifelse(.mv$extraCmt == 2, " with first-order absorption", ""),
      ifelse(length(.state) == 0L, "",
             sprintf(" mixed with free from %d-cmt ODE model",
                     length(.state)))
    ))
  } else if (length(.mv$state) > 0) {
    return(sprintf("rxode2-based free-form %d-cmt ODE model", length(.mv$state)))
  } else {
    return("rxode2-based Pred model")
  }
}
attr(rxUiGet.modelDesc, "desc") <- "Model description (ie linear compartment, pred, ode etc)"
attr(rxUiGet.modelDesc, "rstudio") <- "model description"

#' @export
#' @rdname rxUiGet
rxUiGet.thetaLower <- function(x, ...) {
  .x <- x[[1]]
  .ini <- .x$iniDf
  .w <- !is.na(.ini$ntheta)
  setNames(.ini$lower[.w], .ini$name[.w])
}
attr(rxUiGet.thetaLower, "desc") <- "thetaLower"
attr(rxUiGet.thetaLower, "rstudio") <- c("thetaLower"=1.0) # named vector

#' @export
#' @rdname rxUiGet
rxUiGet.thetaUpper <- function(x, ...) {
  .x <- x[[1]]
  .ini <- .x$iniDf
  .w <- !is.na(.ini$ntheta)
  setNames(.ini$upper[.w], .ini$name[.w])
}
attr(rxUiGet.thetaUpper, "desc") -> "thetaUpper"
attr(rxUiGet.thetaUpper, "rstudio") <- c("thetaUpper"=1.0) # named vector

#' @export
#' @rdname rxUiGet
rxUiGet.lhsVar <- function(x, ...) {
  .x <- x[[1]]
  .eta <- get("etaLhsDf", .x)
  .theta <- get("thetaLhsDf", .x)
  .cov <- get("covLhsDf", .x)
  if (exists("levelLhsDf", .x)) {
    .level <- get("levelLhsDf", .x)
  } else {
    .level <- list(level=character(0),
                  lhs=character(0))
  }
  setNames(c(.eta$eta, .theta$theta, .cov$cov,
             .level$level),
           c(.eta$lhs, .theta$lhs, .cov$lhs,
             .level$lhs))
}

#' @export
#' @rdname rxUiGet
rxUiGet.varLhs <- function(x, ...) {
  .x <- x[[1]]
  .eta <- get("etaLhsDf", .x)
  .theta <- get("thetaLhsDf", .x)
  .cov <- get("covLhsDf", .x)
  if (exists("levelLhsDf", .x)) {
    .level <- get("levelLhsDf", .x)
  } else {
    .level <- list(level=character(0),
                   lhs=character(0))
  }
  setNames(c(.eta$lhs, .theta$lhs, .cov$lhs,
             .level$lhs),
           c(.eta$eta, .theta$theta, .cov$cov,
             .level$level))
}
attr(rxUiGet.varLhs, "desc") <- "var->lhs translation"
attr(rxUiGet.varLhs, "rstudio") <- c("varLhs"="lhs") # character

#' @export
#' @rdname rxUiGet
rxUiGet.lhsEta <- function(x, ...) {
  .x <- x[[1]]
  .eta <- get("etaLhsDf", .x)
  setNames(.eta$eta,.eta$lhs)
}
attr(rxUiGet.lhsEta, "desc") <- "lhs->eta translation"
attr(rxUiGet.lhsEta, "rstudio") <- c("lhsEta"="eta") # character

#' @export
#' @rdname rxUiGet
rxUiGet.lhsTheta <- function(x, ...) {
  .x <- x[[1]]
  .eta <- get("thetaLhsDf", .x)
  setNames(.eta$theta, .eta$lhs)
}
attr(rxUiGet.lhsTheta, "desc") <- "lhs->theta translation"
attr(rxUiGet.lhsTheta, "rstudio") <- c("lhsTheta"="theta") # character

#' @export
#' @rdname rxUiGet
rxUiGet.lhsCov <- function(x, ...) {
  .x <- x[[1]]
  .cov <- get("covLhsDf", .x)
  setNames(.cov$cov, .cov$lhs)
}
attr(rxUiGet.lhsCov, "desc") <- "lhs->cov translation"
attr(rxUiGet.lhsCov, "rstudio") <- c("lhsCov"="cov") # character

#' @export
#' @rdname rxUiGet
rxUiGet.etaLhs <- function(x, ...) {
  .x <- x[[1]]
  .eta <- get("etaLhsDf", .x)
  setNames(.eta$lhs, .eta$eta)
}
attr(rxUiGet.etaLhs, "desc") <- "eta->lhs translation"
attr(rxUiGet.etaLhs, "rstudio") <- c("etaLhs"="etaLhs") # character

#' @export
#' @rdname rxUiGet
rxUiGet.thetaLhs <- function(x, ...) {
  .x <- x[[1]]
  .theta <- get("thetaLhsDf", .x)
  setNames(.theta$lhs, .theta$theta)
}
attr(rxUiGet.thetaLhs, "desc") <- "theta->lhs translation"
attr(rxUiGet.thetaLhs, "rstudio") <- c("theta"="lhs")

#' @export
#' @rdname rxUiGet
rxUiGet.covLhs <- function(x, ...) {
  .x <- x[[1]]
  .cov <- get("covLhsDf", .x)
  setNames(.cov$lhs, .cov$cov)
}
attr(rxUiGet.covLhs, "desc") <- "cov->lhs translation"
attr(rxUiGet.covLhs, "rstudio") <- "covLhs" # character

#' @export
#' @rdname rxUiGet
rxUiGet.levelLhs <- function(x, ...) {
  .x <- x[[1]]
  if (exists("levelLhsDf", .x)) {
    .level <- get("levelLhsDf", .x)
  } else {
    .level <- list(level=character(0),
                  lhs=character(0))
  }
  setNames(.level$lhs, .level$level)
}
attr(rxUiGet.levelLhs, "desc") <- "level->lhs translation"
attr(rxUiGet.levelLhs, "rstudio") <- "levelLhs" # character

#' @export
#' @rdname rxUiGet
rxUiGet.default <- function(x, ...) {
  .arg <- class(x)[1]
  .ui <- x[[1]]
  if (!exists(.arg, envir=.ui)) {
    .meta <- get("meta", envir=.ui)
    if (exists(.arg, envir=.meta)) {
      return(get(.arg, envir=.meta))
    }
    return(NULL)
  }
  get(.arg, .ui)
}

.rxUiGetEnvInfo <- c("model"="Original Model (with comments if available)",
                     "meta"="Model meta information",
                     "iniDf"="Initialization data frame for UI")

.rxUiDevelop <- new.env(parent=emptyenv())
.rxUiDevelop$enable <- FALSE

#' rxUiDevelop - Enable/Disable rxUi development.  Here all $ completions are given
#'
#'
#' @param enable boolean to enable/disable rxUi development mode.
#' @return nothing, called for side effects
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' rxUiDevelop(TRUE)
#' rxUiDevelop(FALSE)
rxUiDevelop <- function(enable=TRUE) {
  .rxUiDevelop$enable <- enable
}

.rxUiGetSupportedDollars <- function() {
  .v <- as.character(utils::methods("rxUiGet"))
  .v <- .v[.v != "rxUiGet.default"]
  .cls <- vapply(.v, function(methodStr) {
    substr(methodStr, 9, nchar(methodStr))
  }, character(1), USE.NAMES=FALSE)
  .v <- vapply(.cls, function(cls) {
    .desc <- attr(utils::getS3method("rxUiGet", cls), "desc")
    if (is.null(.desc)) {
      if (.rxUiDevelop$enable) {
        .desc <- " "
      } else {
        .desc <- ""
      }
    }
    if (is.character(.desc)) {
      .desc
    } else {
      message("Bad description for rxUiGet method: ", cls,
              " - should be character or NULL, got: ", class(.desc))
      ""
    }
  }, character(1), USE.NAMES=TRUE)
  # Take out any "hidden methods"
  .w <- which(.v != "")
  .v <- c(.v[.w], .rxUiGetEnvInfo)
  .v
}

#' @export
str.rxUi <- function(object, ...) {
  cat("rxode2 model function\n")
  .s <- .rxUiGetSupportedDollars()
  cat(paste(strtrim(paste(vapply(names(.s), function(x) {
    .nchar <- nchar(x)
    if (.nchar >= 10) {
      return(paste0(" $ ", x, ": "))
    } else {
      return(paste0(" $ ", x, paste(rep(" ", 10 - .nchar), collapse=""), ": "))
    }
  }, character(1), USE.NAMES=FALSE), .s), 128), collapse="\n"))
  cat("\n")
  invisible()
}

#' @export
.DollarNames.rxUi <- function(x, pattern) {
  .cmp <- names(.rxUiGetSupportedDollars())
  grep(pattern, .cmp, value = TRUE)
}

test_that("rxUiDeparse()", {

  rxUiDeparse(rxControl(), "a")
  expect_equal(rxUiDeparse(rxControl(), "a"),
               str2lang("a <- rxControl()"))

  expect_equal(rxUiDeparse(rxControl(covsInterpolation="linear", method="dop853",
                                    naInterpolation="nocb", keepInterpolation="nocb", sigmaXform="variance",
                                    omegaXform="variance", returnType="data.frame", sumType="fsum", prodType="logify"), "a"),
               str2lang("a <- rxControl(method = \"dop853\", covsInterpolation = \"linear\", returnType = \"data.frame\", sigmaXform = \"variance\", sumType = \"fsum\", naInterpolation = \"nocb\", keepInterpolation = \"nocb\")"))

})

C dlsode.f -- DLSODE (Hindmarsh 1983) for rxode2
C
C  Extracted from deSolve package (ODEPACK double precision routines).
C  Contains: DLSODE (methods 106=Adams, 107=BDF), DSTODE, DPREPJ, DVNORM.
C
C  DLSODE_ERRWD calls renamed to DLSODE_ERRWD to avoid conflict with
C  rxode2's 2-arg xerrwd_ (rprintf.c).
C
C     The code in this file is was taken from 
C     https://www.netlib.org/odepack/
C     Original author: Hindmarsh, Alan C., (LLNL)
C     Adapted for use in R package deSolve by the deSolve authors.
C

*DECK DLSODE
      SUBROUTINE DLSODE (F, NEQ, Y, T, TOUT, ITOL, RTOL, ATOL, ITASK,
     1                  ISTATE, IOPT, RWORK, LRW, IWORK, LIW, JAC, MF,
     2                  rpar, ipar)
      EXTERNAL F, JAC
CKS: added rpar, ipar
      integer ipar(*)
      double precision rpar(*)

      INTEGER NEQ, ITOL, ITASK, ISTATE, IOPT, LRW, IWORK, LIW, MF
      DOUBLE PRECISION Y, T, TOUT, RTOL, ATOL, RWORK
      DIMENSION NEQ(*), Y(*), RTOL(*), ATOL(*), RWORK(LRW), IWORK(LIW)
C***BEGIN PROLOGUE  DLSODE
C***PURPOSE  Livermore Solver for Ordinary Differential Equations.
C            DLSODE solves the initial-value problem for stiff or
C            nonstiff systems of first-order ODE's,
C               dy/dt = f(t,y),   or, in component form,
C               dy(i)/dt = f(i) = f(i,t,y(1),y(2),...,y(N)),  i=1,...,N.
C***CATEGORY  I1A
C***TYPE      DOUBLE PRECISION (SLSODE-S, DLSODE-D)
C***KEYWORDS  ORDINARY DIFFERENTIAL EQUATIONS, INITIAL VALUE PROBLEM,
C             STIFF, NONSTIFF
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C             Center for Applied Scientific Computing, L-561
C             Lawrence Livermore National Laboratory
C             Livermore, CA 94551.
C***DESCRIPTION
C
C     NOTE: The "Usage" and "Arguments" sections treat only a subset of
C           available options, in condensed fashion.  The options
C           covered and the information supplied will support most
C           standard uses of DLSODE.
C
C           For more sophisticated uses, full details on all options are
C           given in the concluding section, headed "Long Description."
C           A synopsis of the DLSODE Long Description is provided at the
C           beginning of that section; general topics covered are:
C           - Elements of the call sequence; optional input and output
C           - Optional supplemental routines in the DLSODE package
C           - internal COMMON block
C
C     changes by Karline Soetaert.
C     NOTE for inclusion in R-package: the interface to F, Res and Jac has
C          been changed: now a double precision and an integer vector
C          rpar(*) and ipar(*) is also passed. This to allow output of
C          ordinary output variables.
C          These changes have been made consistently throughout the code
C          including subroutines in opkda1.f
C *Usage:
C     Communication between the user and the DLSODE package, for normal
C     situations, is summarized here.  This summary describes a subset
C     of the available options.  See "Long Description" for complete
C     details, including optional communication, nonstandard options,
C     and instructions for special situations.
C
C     A sample program is given in the "Examples" section.
C
C     Refer to the argument descriptions for the definitions of the
C     quantities that appear in the following sample declarations.
C
C     For MF = 10,
C        PARAMETER  (LRW = 20 + 16*NEQ,           LIW = 20)
C     For MF = 21 or 22,
C        PARAMETER  (LRW = 22 +  9*NEQ + NEQ**2,  LIW = 20 + NEQ)
C     For MF = 24 or 25,
C        PARAMETER  (LRW = 22 + 10*NEQ + (2*ML+MU)*NEQ,
C       *                                         LIW = 20 + NEQ)
C
C        EXTERNAL F, JAC
C        INTEGER  NEQ, ITOL, ITASK, ISTATE, IOPT, LRW, IWORK(LIW),
C       *         LIW, MF
C        DOUBLE PRECISION Y(NEQ), T, TOUT, RTOL, ATOL(ntol), RWORK(LRW)
C
C        CALL DLSODE (F, NEQ, Y, T, TOUT, ITOL, RTOL, ATOL, ITASK,
C       *            ISTATE, IOPT, RWORK, LRW, IWORK, LIW, JAC, MF)
C
C *Arguments:
C     F     :EXT    Name of subroutine for right-hand-side vector f.
C                   This name must be declared EXTERNAL in calling
C                   program.  The form of F must be:
C
C                   SUBROUTINE  F (NEQ, T, Y, YDOT)
C                   INTEGER  NEQ
C                   DOUBLE PRECISION  T, Y(*), YDOT(*)
C
C                   The inputs are NEQ, T, Y.  F is to set
C
C                   YDOT(i) = f(i,T,Y(1),Y(2),...,Y(NEQ)),
C                                                     i = 1, ..., NEQ .
C
C     NEQ   :IN     Number of first-order ODE's.
C
C     Y     :INOUT  Array of values of the y(t) vector, of length NEQ.
C                   Input:  For the first call, Y should contain the
C                           values of y(t) at t = T. (Y is an input
C                           variable only if ISTATE = 1.)
C                   Output: On return, Y will contain the values at the
C                           new t-value.
C
C     T     :INOUT  Value of the independent variable.  On return it
C                   will be the current value of t (normally TOUT).
C
C     TOUT  :IN     Next point where output is desired (.NE. T).
C
C     ITOL  :IN     1 or 2 according as ATOL (below) is a scalar or
C                   an array.
C
C     RTOL  :IN     Relative tolerance parameter (scalar).
C
C     ATOL  :IN     Absolute tolerance parameter (scalar or array).
C                   If ITOL = 1, ATOL need not be dimensioned.
C                   If ITOL = 2, ATOL must be dimensioned at least NEQ.
C
C                   The estimated local error in Y(i) will be controlled
C                   so as to be roughly less (in magnitude) than
C
C                   EWT(i) = RTOL*ABS(Y(i)) + ATOL     if ITOL = 1, or
C                   EWT(i) = RTOL*ABS(Y(i)) + ATOL(i)  if ITOL = 2.
C
C                   Thus the local error test passes if, in each
C                   component, either the absolute error is less than
C                   ATOL (or ATOL(i)), or the relative error is less
C                   than RTOL.
C
C                   Use RTOL = 0.0 for pure absolute error control, and
C                   use ATOL = 0.0 (or ATOL(i) = 0.0) for pure relative
C                   error control.  Caution:  Actual (global) errors may
C                   exceed these local tolerances, so choose them
C                   conservatively.
C
C     ITASK :IN     Flag indicating the task DLSODE is to perform.
C                   Use ITASK = 1 for normal computation of output
C                   values of y at t = TOUT.
C
C     ISTATE:INOUT  Index used for input and output to specify the state
C                   of the calculation.
C                   Input:
C                    1   This is the first call for a problem.
C                    2   This is a subsequent call.
C                   Output:
C                    1   Nothing was done, because TOUT was equal to T.
C                    2   DLSODE was successful (otherwise, negative).
C                        Note that ISTATE need not be modified after a
C                        successful return.
C                   -1   Excess work done on this call (perhaps wrong
C                        MF).
C                   -2   Excess accuracy requested (tolerances too
C                        small).
C                   -3   Illegal input detected (see printed message).
C                   -4   Repeated error test failures (check all
C                        inputs).
C                   -5   Repeated convergence failures (perhaps bad
C                        Jacobian supplied or wrong choice of MF or
C                        tolerances).
C                   -6   Error weight became zero during problem
C                        (solution component i vanished, and ATOL or
C                        ATOL(i) = 0.).
C
C     IOPT  :IN     Flag indicating whether optional inputs are used:
C                   0   No.
C                   1   Yes.  (See "Optional inputs" under "Long
C                       Description," Part 1.)
C
C     RWORK :WORK   Real work array of length at least:
C                   20 + 16*NEQ                    for MF = 10,
C                   22 +  9*NEQ + NEQ**2           for MF = 21 or 22,
C                   22 + 10*NEQ + (2*ML + MU)*NEQ  for MF = 24 or 25.
C
C     LRW   :IN     Declared length of RWORK (in user's DIMENSION
C                   statement).
C
C     IWORK :WORK   Integer work array of length at least:
C                   20        for MF = 10,
C                   20 + NEQ  for MF = 21, 22, 24, or 25.
C
C                   If MF = 24 or 25, input in IWORK(1),IWORK(2) the
C                   lower and upper Jacobian half-bandwidths ML,MU.
C
C                   On return, IWORK contains information that may be
C                   of interest to the user:
C
C            Name   Location   Meaning
C            -----  ---------  -----------------------------------------
C            NST    IWORK(11)  Number of steps taken for the problem so
C                              far.
C            NFE    IWORK(12)  Number of f evaluations for the problem
C                              so far.
C            NJE    IWORK(13)  Number of Jacobian evaluations (and of
C                              matrix LU decompositions) for the problem
C                              so far.
C            NQU    IWORK(14)  Method order last used (successfully).
C            LENRW  IWORK(17)  Length of RWORK actually required.  This
C                              is defined on normal returns and on an
C                              illegal input return for insufficient
C                              storage.
C            LENIW  IWORK(18)  Length of IWORK actually required.  This
C                              is defined on normal returns and on an
C                              illegal input return for insufficient
C                              storage.
C
C     LIW   :IN     Declared length of IWORK (in user's DIMENSION
C                   statement).
C
C     JAC   :EXT    Name of subroutine for Jacobian matrix (MF =
C                   21 or 24).  If used, this name must be declared
C                   EXTERNAL in calling program.  If not used, pass a
C                   dummy name.  The form of JAC must be:
C
C                   SUBROUTINE JAC (NEQ, T, Y, ML, MU, PD, NROWPD)
C                   INTEGER  NEQ, ML, MU, NROWPD
C                   DOUBLE PRECISION  T, Y(*), PD(NROWPD,*)
C
C                   See item c, under "Description" below for more
C                   information about JAC.
C
C     MF    :IN     Method flag.  Standard values are:
C                   10  Nonstiff (Adams) method, no Jacobian used.
C                   21  Stiff (BDF) method, user-supplied full Jacobian.
C                   22  Stiff method, internally generated full
C                       Jacobian.
C                   24  Stiff method, user-supplied banded Jacobian.
C                   25  Stiff method, internally generated banded
C                       Jacobian.
C
C *Description:
C     DLSODE solves the initial value problem for stiff or nonstiff
C     systems of first-order ODE's,
C
C        dy/dt = f(t,y) ,
C
C     or, in component form,
C
C        dy(i)/dt = f(i) = f(i,t,y(1),y(2),...,y(NEQ))
C                                                  (i = 1, ..., NEQ) .
C
C     DLSODE is a package based on the GEAR and GEARB packages, and on
C     the October 23, 1978, version of the tentative ODEPACK user
C     interface standard, with minor modifications.
C
C     The steps in solving such a problem are as follows.
C
C     a. First write a subroutine of the form
C
C           SUBROUTINE  F (NEQ, T, Y, YDOT,rpar,ipar)
C           INTEGER  NEQ,ipar(*)
C           DOUBLE PRECISION  T, Y(*), YDOT(*),rpar(*)
C
C        which supplies the vector function f by loading YDOT(i) with
C        f(i).
C
C     b. Next determine (or guess) whether or not the problem is stiff.
C        Stiffness occurs when the Jacobian matrix df/dy has an
C        eigenvalue whose real part is negative and large in magnitude
C        compared to the reciprocal of the t span of interest.  If the
C        problem is nonstiff, use method flag MF = 10.  If it is stiff,
C        there are four standard choices for MF, and DLSODE requires the
C        Jacobian matrix in some form.  This matrix is regarded either
C        as full (MF = 21 or 22), or banded (MF = 24 or 25).  In the
C        banded case, DLSODE requires two half-bandwidth parameters ML
C        and MU. These are, respectively, the widths of the lower and
C        upper parts of the band, excluding the main diagonal.  Thus the
C        band consists of the locations (i,j) with
C
C           i - ML <= j <= i + MU ,
C
C        and the full bandwidth is ML + MU + 1 .
C
C     c. If the problem is stiff, you are encouraged to supply the
C        Jacobian directly (MF = 21 or 24), but if this is not feasible,
C        DLSODE will compute it internally by difference quotients (MF =
C        22 or 25).  If you are supplying the Jacobian, write a
C        subroutine of the form
C
C           SUBROUTINE  JAC (NEQ, T, Y, ML, MU, PD, NROWPD,rpar,ipar)
C           INTEGER  NEQ, ML, MU, NRWOPD,ipar(*)
C           DOUBLE PRECISION  T, Y(*), PD(NROWPD,*),rpar(*)
C
C        which provides df/dy by loading PD as follows:
C        - For a full Jacobian (MF = 21), load PD(i,j) with df(i)/dy(j),
C          the partial derivative of f(i) with respect to y(j).  (Ignore
C          the ML and MU arguments in this case.)
C        - For a banded Jacobian (MF = 24), load PD(i-j+MU+1,j) with
C          df(i)/dy(j); i.e., load the diagonal lines of df/dy into the
C          rows of PD from the top down.
C        - In either case, only nonzero elements need be loaded.
C
C     d. Write a main program that calls subroutine DLSODE once for each
C        point at which answers are desired.  This should also provide
C        for possible use of logical unit 6 for output of error messages
C        by DLSODE.
C
C        Before the first call to DLSODE, set ISTATE = 1, set Y and T to
C        the initial values, and set TOUT to the first output point.  To
C        continue the integration after a successful return, simply
C        reset TOUT and call DLSODE again.  No other parameters need be
C        reset.
C
C *Examples:
C     The following is a simple example problem, with the coding needed
C     for its solution by DLSODE. The problem is from chemical kinetics,
C     and consists of the following three rate equations:
C
C        dy1/dt = -.04*y1 + 1.E4*y2*y3
C        dy2/dt = .04*y1 - 1.E4*y2*y3 - 3.E7*y2**2
C        dy3/dt = 3.E7*y2**2
C
C     on the interval from t = 0.0 to t = 4.E10, with initial conditions
C     y1 = 1.0, y2 = y3 = 0. The problem is stiff.
C
C     The following coding solves this problem with DLSODE, using
C     MF = 21 and printing results at t = .4, 4., ..., 4.E10.  It uses
C     ITOL = 2 and ATOL much smaller for y2 than for y1 or y3 because y2
C     has much smaller values.  At the end of the run, statistical
C     quantities of interest are printed.
C
C        EXTERNAL  FEX, JEX
C        INTEGER  IOPT, IOUT, ISTATE, ITASK, ITOL, IWORK(23), LIW, LRW,
C       *         MF, NEQ
C        DOUBLE PRECISION  ATOL(3), RTOL, RWORK(58), T, TOUT, Y(3)
C        NEQ = 3
C        Y(1) = 1.D0
C        Y(2) = 0.D0
C        Y(3) = 0.D0
C        T = 0.D0
C        TOUT = .4D0
C        ITOL = 2
C        RTOL = 1.D-4
C        ATOL(1) = 1.D-6
C        ATOL(2) = 1.D-10
C        ATOL(3) = 1.D-6
C        ITASK = 1
C        ISTATE = 1
C        IOPT = 0
C        LRW = 58
C        LIW = 23
C        MF = 21
C        DO 40 IOUT = 1,12
C          CALL DLSODE (FEX, NEQ, Y, T, TOUT, ITOL, RTOL, ATOL, ITASK,
C       *               ISTATE, IOPT, RWORK, LRW, IWORK, LIW, JEX, MF)
C          WRITE(6,20)  T, Y(1), Y(2), Y(3)
C    20    FORMAT(' At t =',D12.4,'   y =',3D14.6)
C          IF (ISTATE .LT. 0)  GO TO 80
C    40    TOUT = TOUT*10.D0
C        WRITE(6,60)  IWORK(11), IWORK(12), IWORK(13)
C    60  FORMAT(/' No. steps =',i4,',  No. f-s =',i4,',  No. J-s =',i4)
C        STOP
C    80  WRITE(6,90)  ISTATE
C    90  FORMAT(///' Error halt.. ISTATE =',I3)
C        STOP
C        END
C
C        SUBROUTINE  FEX (NEQ, T, Y, YDOT, rpar, ipar)
C        INTEGER  NEQ, ipar(*)
C        DOUBLE PRECISION  T, Y(3), YDOT(3), rpar(*)
C        YDOT(1) = -.04D0*Y(1) + 1.D4*Y(2)*Y(3)
C        YDOT(3) = 3.D7*Y(2)*Y(2)
C        YDOT(2) = -YDOT(1) - YDOT(3)
C        RETURN
C        END
C
C        SUBROUTINE  JEX (NEQ, T, Y, ML, MU, PD, NRPD, rpar, ipar)
C        INTEGER  NEQ, ML, MU, NRPD, ipar(*)
C        DOUBLE PRECISION  T, Y(3), PD(NRPD,3), rpar(*)
C        PD(1,1) = -.04D0
C        PD(1,2) = 1.D4*Y(3)
C        PD(1,3) = 1.D4*Y(2)
C        PD(2,1) = .04D0
C        PD(2,3) = -PD(1,3)
C        PD(3,2) = 6.D7*Y(2)
C        PD(2,2) = -PD(1,2) - PD(3,2)
C        RETURN
C        END
C
C     The output from this program (on a Cray-1 in single precision)
C     is as follows.
C
C     At t =  4.0000e-01   y =  9.851726e-01  3.386406e-05  1.479357e-02
C     At t =  4.0000e+00   y =  9.055142e-01  2.240418e-05  9.446344e-02
C     At t =  4.0000e+01   y =  7.158050e-01  9.184616e-06  2.841858e-01
C     At t =  4.0000e+02   y =  4.504846e-01  3.222434e-06  5.495122e-01
C     At t =  4.0000e+03   y =  1.831701e-01  8.940379e-07  8.168290e-01
C     At t =  4.0000e+04   y =  3.897016e-02  1.621193e-07  9.610297e-01
C     At t =  4.0000e+05   y =  4.935213e-03  1.983756e-08  9.950648e-01
C     At t =  4.0000e+06   y =  5.159269e-04  2.064759e-09  9.994841e-01
C     At t =  4.0000e+07   y =  5.306413e-05  2.122677e-10  9.999469e-01
C     At t =  4.0000e+08   y =  5.494530e-06  2.197825e-11  9.999945e-01
C     At t =  4.0000e+09   y =  5.129458e-07  2.051784e-12  9.999995e-01
C     At t =  4.0000e+10   y = -7.170603e-08 -2.868241e-13  1.000000e+00
C
C     No. steps = 330,  No. f-s = 405,  No. J-s = 69
C
C *Accuracy:
C     The accuracy of the solution depends on the choice of tolerances
C     RTOL and ATOL.  Actual (global) errors may exceed these local
C     tolerances, so choose them conservatively.
C
C *Cautions:
C     The work arrays should not be altered between calls to DLSODE for
C     the same problem, except possibly for the conditional and optional
C     inputs.
C
C *Portability:
C     Since NEQ is dimensioned inside DLSODE, some compilers may object
C     to a call to DLSODE with NEQ a scalar variable.  In this event,
C     use DIMENSION NEQ(1).  Similar remarks apply to RTOL and ATOL.
C
C     Note to Cray users:
C     For maximum efficiency, use the CFT77 compiler.  Appropriate
C     compiler optimization directives have been inserted for CFT77.
C
C *Reference:
C     Alan C. Hindmarsh, "ODEPACK, A Systematized Collection of ODE
C     Solvers," in Scientific Computing, R. S. Stepleman, et al., Eds.
C     (North-Holland, Amsterdam, 1983), pp. 55-64.
C
C *Long Description:
C     The following complete description of the user interface to
C     DLSODE consists of four parts:
C
C     1.  The call sequence to subroutine DLSODE, which is a driver
C         routine for the solver.  This includes descriptions of both
C         the call sequence arguments and user-supplied routines.
C         Following these descriptions is a description of optional
C         inputs available through the call sequence, and then a
C         description of optional outputs in the work arrays.
C
C     2.  Descriptions of other routines in the DLSODE package that may
C         be (optionally) called by the user.  These provide the ability
C         to alter error message handling, save and restore the internal
C         COMMON, and obtain specified derivatives of the solution y(t).
C
C     3.  Descriptions of COMMON block to be declared in overlay or
C         similar environments, or to be saved when doing an interrupt
C         of the problem and continued solution later.
C
C     4.  Description of two routines in the DLSODE package, either of
C         which the user may replace with his own version, if desired.
C         These relate to the measurement of errors.
C
C
C                         Part 1.  Call Sequence
C                         ----------------------
C
C     Arguments
C     ---------
C     The call sequence parameters used for input only are
C
C        F, NEQ, TOUT, ITOL, RTOL, ATOL, ITASK, IOPT, LRW, LIW, JAC, MF,
C
C     and those used for both input and output are
C
C        Y, T, ISTATE.
C
C     The work arrays RWORK and IWORK are also used for conditional and
C     optional inputs and optional outputs.  (The term output here
C     refers to the return from subroutine DLSODE to the user's calling
C     program.)
C
C     The legality of input parameters will be thoroughly checked on the
C     initial call for the problem, but not checked thereafter unless a
C     change in input parameters is flagged by ISTATE = 3 on input.
C
C     The descriptions of the call arguments are as follows.
C
C     F        The name of the user-supplied subroutine defining the ODE
C              system.  The system must be put in the first-order form
C              dy/dt = f(t,y), where f is a vector-valued function of
C              the scalar t and the vector y. Subroutine F is to compute
C              the function f. It is to have the form
C
C           SUBROUTINE  F (NEQ, T, Y, YDOT,rpar,ipar)
C           INTEGER  NEQ,ipar(*)
C           DOUBLE PRECISION  T, Y(*), YDOT(*),rpar(*)
C
C              where NEQ, T, and Y are input, and the array YDOT =
C              f(T,Y) is output.  Y and YDOT are arrays of length NEQ.
C              Subroutine F should not alter Y(1),...,Y(NEQ).  F must be
C              declared EXTERNAL in the calling program.
C
C              Subroutine F may access user-defined quantities in
C              NEQ(2),... and/or in Y(NEQ(1)+1),..., if NEQ is an array
C              (dimensioned in F) and/or Y has length exceeding NEQ(1).
C              See the descriptions of NEQ and Y below.
C
C              If quantities computed in the F routine are needed
C              externally to DLSODE, an extra call to F should be made
C              for this purpose, for consistent and accurate results.
C              If only the derivative dy/dt is needed, use DINTDY
C              instead.
C
C     NEQ      The size of the ODE system (number of first-order
C              ordinary differential equations).  Used only for input.
C              NEQ may be decreased, but not increased, during the
C              problem.  If NEQ is decreased (with ISTATE = 3 on input),
C              the remaining components of Y should be left undisturbed,
C              if these are to be accessed in F and/or JAC.
C
C              Normally, NEQ is a scalar, and it is generally referred
C              to as a scalar in this user interface description.
C              However, NEQ may be an array, with NEQ(1) set to the
C              system size.  (The DLSODE package accesses only NEQ(1).)
C              In either case, this parameter is passed as the NEQ
C              argument in all calls to F and JAC.  Hence, if it is an
C              array, locations NEQ(2),... may be used to store other
C              integer data and pass it to F and/or JAC.  Subroutines
C              F and/or JAC must include NEQ in a DIMENSION statement
C              in that case.
C
C     Y        A real array for the vector of dependent variables, of
C              length NEQ or more.  Used for both input and output on
C              the first call (ISTATE = 1), and only for output on
C              other calls.  On the first call, Y must contain the
C              vector of initial values.  On output, Y contains the
C              computed solution vector, evaluated at T. If desired,
C              the Y array may be used for other purposes between
C              calls to the solver.
C
C              This array is passed as the Y argument in all calls to F
C              and JAC.  Hence its length may exceed NEQ, and locations
C              Y(NEQ+1),... may be used to store other real data and
C              pass it to F and/or JAC.  (The DLSODE package accesses
C              only Y(1),...,Y(NEQ).)
C
C     T        The independent variable.  On input, T is used only on
C              the first call, as the initial point of the integration.
C              On output, after each call, T is the value at which a
C              computed solution Y is evaluated (usually the same as
C              TOUT).  On an error return, T is the farthest point
C              reached.
C
C     TOUT     The next value of T at which a computed solution is
C              desired.  Used only for input.
C
C              When starting the problem (ISTATE = 1), TOUT may be equal
C              to T for one call, then should not equal T for the next
C              call.  For the initial T, an input value of TOUT .NE. T
C              is used in order to determine the direction of the
C              integration (i.e., the algebraic sign of the step sizes)
C              and the rough scale of the problem.  Integration in
C              either direction (forward or backward in T) is permitted.
C
C              If ITASK = 2 or 5 (one-step modes), TOUT is ignored
C              after the first call (i.e., the first call with
C              TOUT .NE. T).  Otherwise, TOUT is required on every call.
C
C              If ITASK = 1, 3, or 4, the values of TOUT need not be
C              monotone, but a value of TOUT which backs up is limited
C              to the current internal T interval, whose endpoints are
C              TCUR - HU and TCUR.  (See "Optional Outputs" below for
C              TCUR and HU.)
C
C
C     ITOL     An indicator for the type of error control.  See
C              description below under ATOL.  Used only for input.
C
C     RTOL     A relative error tolerance parameter, either a scalar or
C              an array of length NEQ.  See description below under
C              ATOL.  Input only.
C
C     ATOL     An absolute error tolerance parameter, either a scalar or
C              an array of length NEQ.  Input only.
C
C              The input parameters ITOL, RTOL, and ATOL determine the
C              error control performed by the solver.  The solver will
C              control the vector e = (e(i)) of estimated local errors
C              in Y, according to an inequality of the form
C
C                 rms-norm of ( e(i)/EWT(i) ) <= 1,
C
C              where
C
C                 EWT(i) = RTOL(i)*ABS(Y(i)) + ATOL(i),
C
C              and the rms-norm (root-mean-square norm) here is
C
C                 rms-norm(v) = SQRT(sum v(i)**2 / NEQ).
C
C              Here EWT = (EWT(i)) is a vector of weights which must
C              always be positive, and the values of RTOL and ATOL
C              should all be nonnegative.  The following table gives the
C              types (scalar/array) of RTOL and ATOL, and the
C              corresponding form of EWT(i).
C
C              ITOL    RTOL      ATOL      EWT(i)
C              ----    ------    ------    -----------------------------
C              1       scalar    scalar    RTOL*ABS(Y(i)) + ATOL
C              2       scalar    array     RTOL*ABS(Y(i)) + ATOL(i)
C              3       array     scalar    RTOL(i)*ABS(Y(i)) + ATOL
C              4       array     array     RTOL(i)*ABS(Y(i)) + ATOL(i)
C
C              When either of these parameters is a scalar, it need not
C              be dimensioned in the user's calling program.
C
C              If none of the above choices (with ITOL, RTOL, and ATOL
C              fixed throughout the problem) is suitable, more general
C              error controls can be obtained by substituting
C              user-supplied routines for the setting of EWT and/or for
C              the norm calculation.  See Part 4 below.
C
C              If global errors are to be estimated by making a repeated
C              run on the same problem with smaller tolerances, then all
C              components of RTOL and ATOL (i.e., of EWT) should be
C              scaled down uniformly.
C
C     ITASK    An index specifying the task to be performed.  Input
C              only.  ITASK has the following values and meanings:
C              1   Normal computation of output values of y(t) at
C                  t = TOUT (by overshooting and interpolating).
C              2   Take one step only and return.
C              3   Stop at the first internal mesh point at or beyond
C                  t = TOUT and return.
C              4   Normal computation of output values of y(t) at
C                  t = TOUT but without overshooting t = TCRIT.  TCRIT
C                  must be input as RWORK(1).  TCRIT may be equal to or
C                  beyond TOUT, but not behind it in the direction of
C                  integration.  This option is useful if the problem
C                  has a singularity at or beyond t = TCRIT.
C              5   Take one step, without passing TCRIT, and return.
C                  TCRIT must be input as RWORK(1).
C
C              Note:  If ITASK = 4 or 5 and the solver reaches TCRIT
C              (within roundoff), it will return T = TCRIT (exactly) to
C              indicate this (unless ITASK = 4 and TOUT comes before
C              TCRIT, in which case answers at T = TOUT are returned
C              first).
C
C     ISTATE   An index used for input and output to specify the state
C              of the calculation.
C
C              On input, the values of ISTATE are as follows:
C              1   This is the first call for the problem
C                  (initializations will be done).  See "Note" below.
C              2   This is not the first call, and the calculation is to
C                  continue normally, with no change in any input
C                  parameters except possibly TOUT and ITASK.  (If ITOL,
C                  RTOL, and/or ATOL are changed between calls with
C                  ISTATE = 2, the new values will be used but not
C                  tested for legality.)
C              3   This is not the first call, and the calculation is to
C                  continue normally, but with a change in input
C                  parameters other than TOUT and ITASK.  Changes are
C                  allowed in NEQ, ITOL, RTOL, ATOL, IOPT, LRW, LIW, MF,
C                  ML, MU, and any of the optional inputs except H0.
C                  (See IWORK description for ML and MU.)
C
C              Note:  A preliminary call with TOUT = T is not counted as
C              a first call here, as no initialization or checking of
C              input is done.  (Such a call is sometimes useful for the
C              purpose of outputting the initial conditions.)  Thus the
C              first call for which TOUT .NE. T requires ISTATE = 1 on
C              input.
C
C              On output, ISTATE has the following values and meanings:
C               1  Nothing was done, as TOUT was equal to T with
C                  ISTATE = 1 on input.
C               2  The integration was performed successfully.
C              -1  An excessive amount of work (more than MXSTEP steps)
C                  was done on this call, before completing the
C                  requested task, but the integration was otherwise
C                  successful as far as T. (MXSTEP is an optional input
C                  and is normally 500.)  To continue, the user may
C                  simply reset ISTATE to a value >1 and call again (the
C                  excess work step counter will be reset to 0).  In
C                  addition, the user may increase MXSTEP to avoid this
C                  error return; see "Optional Inputs" below.
C              -2  Too much accuracy was requested for the precision of
C                  the machine being used.  This was detected before
C                  completing the requested task, but the integration
C                  was successful as far as T. To continue, the
C                  tolerance parameters must be reset, and ISTATE must
C                  be set to 3. The optional output TOLSF may be used
C                  for this purpose.  (Note:  If this condition is
C                  detected before taking any steps, then an illegal
C                  input return (ISTATE = -3) occurs instead.)
C              -3  Illegal input was detected, before taking any
C                  integration steps.  See written message for details.
C                  (Note:  If the solver detects an infinite loop of
C                  calls to the solver with illegal input, it will cause
C                  the run to stop.)
C              -4  There were repeated error-test failures on one
C                  attempted step, before completing the requested task,
C                  but the integration was successful as far as T.  The
C                  problem may have a singularity, or the input may be
C                  inappropriate.
C              -5  There were repeated convergence-test failures on one
C                  attempted step, before completing the requested task,
C                  but the integration was successful as far as T. This
C                  may be caused by an inaccurate Jacobian matrix, if
C                  one is being used.
C              -6  EWT(i) became zero for some i during the integration.
C                  Pure relative error control (ATOL(i)=0.0) was
C                  requested on a variable which has now vanished.  The
C                  integration was successful as far as T.
C
C              Note:  Since the normal output value of ISTATE is 2, it
C              does not need to be reset for normal continuation.  Also,
C              since a negative input value of ISTATE will be regarded
C              as illegal, a negative output value requires the user to
C              change it, and possibly other inputs, before calling the
C              solver again.
C
C     IOPT     An integer flag to specify whether any optional inputs
C              are being used on this call.  Input only.  The optional
C              inputs are listed under a separate heading below.
C              0   No optional inputs are being used.  Default values
C                  will be used in all cases.
C              1   One or more optional inputs are being used.
C
C     RWORK    A real working array (double precision).  The length of
C              RWORK must be at least
C
C                 20 + NYH*(MAXORD + 1) + 3*NEQ + LWM
C
C              where
C                 NYH = the initial value of NEQ,
C              MAXORD = 12 (if METH = 1) or 5 (if METH = 2) (unless a
C                       smaller value is given as an optional input),
C                 LWM = 0           if MITER = 0,
C                 LWM = NEQ**2 + 2  if MITER = 1 or 2,
C                 LWM = NEQ + 2     if MITER = 3, and
C                 LWM = (2*ML + MU + 1)*NEQ + 2
C                                   if MITER = 4 or 5.
C              (See the MF description below for METH and MITER.)
C
C              Thus if MAXORD has its default value and NEQ is constant,
C              this length is:
C              20 + 16*NEQ                    for MF = 10,
C              22 + 16*NEQ + NEQ**2           for MF = 11 or 12,
C              22 + 17*NEQ                    for MF = 13,
C              22 + 17*NEQ + (2*ML + MU)*NEQ  for MF = 14 or 15,
C              20 +  9*NEQ                    for MF = 20,
C              22 +  9*NEQ + NEQ**2           for MF = 21 or 22,
C              22 + 10*NEQ                    for MF = 23,
C              22 + 10*NEQ + (2*ML + MU)*NEQ  for MF = 24 or 25.
C
C              The first 20 words of RWORK are reserved for conditional
C              and optional inputs and optional outputs.
C
C              The following word in RWORK is a conditional input:
C              RWORK(1) = TCRIT, the critical value of t which the
C                         solver is not to overshoot.  Required if ITASK
C                         is 4 or 5, and ignored otherwise.  See ITASK.
C
C     LRW      The length of the array RWORK, as declared by the user.
C              (This will be checked by the solver.)
C
C     IWORK    An integer work array.  Its length must be at least
C              20       if MITER = 0 or 3 (MF = 10, 13, 20, 23), or
C              20 + NEQ otherwise (MF = 11, 12, 14, 15, 21, 22, 24, 25).
C              (See the MF description below for MITER.)  The first few
C              words of IWORK are used for conditional and optional
C              inputs and optional outputs.
C
C              The following two words in IWORK are conditional inputs:
C              IWORK(1) = ML   These are the lower and upper half-
C              IWORK(2) = MU   bandwidths, respectively, of the banded
C                              Jacobian, excluding the main diagonal.
C                         The band is defined by the matrix locations
C                         (i,j) with i - ML <= j <= i + MU. ML and MU
C                         must satisfy 0 <= ML,MU <= NEQ - 1. These are
C                         required if MITER is 4 or 5, and ignored
C                         otherwise.  ML and MU may in fact be the band
C                         parameters for a matrix to which df/dy is only
C                         approximately equal.
C
C     LIW      The length of the array IWORK, as declared by the user.
C              (This will be checked by the solver.)
C
C     Note:  The work arrays must not be altered between calls to DLSODE
C     for the same problem, except possibly for the conditional and
C     optional inputs, and except for the last 3*NEQ words of RWORK.
C     The latter space is used for internal scratch space, and so is
C     available for use by the user outside DLSODE between calls, if
C     desired (but not for use by F or JAC).
C
C     JAC      The name of the user-supplied routine (MITER = 1 or 4) to
C              compute the Jacobian matrix, df/dy, as a function of the
C              scalar t and the vector y.  (See the MF description below
C              for MITER.)  It is to have the form
C
C                 SUBROUTINE JAC (NEQ, T, Y, ML, MU, PD, NROWPD, rpar, ipar)
C                 integer ipar(*)
C                 DOUBLE PRECISION T, Y(*), PD(NROWPD,*), rpar(*)
C
C              where NEQ, T, Y, ML, MU, and NROWPD are input and the
C              array PD is to be loaded with partial derivatives
C              (elements of the Jacobian matrix) on output.  PD must be
C              given a first dimension of NROWPD.  T and Y have the same
C              meaning as in subroutine F.
C
C              In the full matrix case (MITER = 1), ML and MU are
C              ignored, and the Jacobian is to be loaded into PD in
C              columnwise manner, with df(i)/dy(j) loaded into PD(i,j).
C
C              In the band matrix case (MITER = 4), the elements within
C              the band are to be loaded into PD in columnwise manner,
C              with diagonal lines of df/dy loaded into the rows of PD.
C              Thus df(i)/dy(j) is to be loaded into PD(i-j+MU+1,j).  ML
C              and MU are the half-bandwidth parameters (see IWORK).
C              The locations in PD in the two triangular areas which
C              correspond to nonexistent matrix elements can be ignored
C              or loaded arbitrarily, as they are overwritten by DLSODE.
C
C              JAC need not provide df/dy exactly. A crude approximation
C              (possibly with a smaller bandwidth) will do.
C
C              In either case, PD is preset to zero by the solver, so
C              that only the nonzero elements need be loaded by JAC.
C              Each call to JAC is preceded by a call to F with the same
C              arguments NEQ, T, and Y. Thus to gain some efficiency,
C              intermediate quantities shared by both calculations may
C              be saved in a user COMMON block by F and not recomputed
C              by JAC, if desired.  Also, JAC may alter the Y array, if
C              desired.  JAC must be declared EXTERNAL in the calling
C              program.
C
C              Subroutine JAC may access user-defined quantities in
C              NEQ(2),... and/or in Y(NEQ(1)+1),... if NEQ is an array
C              (dimensioned in JAC) and/or Y has length exceeding
C              NEQ(1).  See the descriptions of NEQ and Y above.
C
C     MF       The method flag.  Used only for input.  The legal values
C              of MF are 10, 11, 12, 13, 14, 15, 20, 21, 22, 23, 24,
C              and 25.  MF has decimal digits METH and MITER:
C                 MF = 10*METH + MITER .
C
C              METH indicates the basic linear multistep method:
C              1   Implicit Adams method.
C              2   Method based on backward differentiation formulas
C                  (BDF's).
C
C              MITER indicates the corrector iteration method:
C              0   Functional iteration (no Jacobian matrix is
C                  involved).
C              1   Chord iteration with a user-supplied full (NEQ by
C                  NEQ) Jacobian.
C              2   Chord iteration with an internally generated
C                  (difference quotient) full Jacobian (using NEQ
C                  extra calls to F per df/dy value).
C              3   Chord iteration with an internally generated
C                  diagonal Jacobian approximation (using one extra call
C                  to F per df/dy evaluation).
C              4   Chord iteration with a user-supplied banded Jacobian.
C              5   Chord iteration with an internally generated banded
C                  Jacobian (using ML + MU + 1 extra calls to F per
C                  df/dy evaluation).
C
C              If MITER = 1 or 4, the user must supply a subroutine JAC
C              (the name is arbitrary) as described above under JAC.
C              For other values of MITER, a dummy argument can be used.
C
C     Optional Inputs
C     ---------------
C     The following is a list of the optional inputs provided for in the
C     call sequence.  (See also Part 2.)  For each such input variable,
C     this table lists its name as used in this documentation, its
C     location in the call sequence, its meaning, and the default value.
C     The use of any of these inputs requires IOPT = 1, and in that case
C     all of these inputs are examined.  A value of zero for any of
C     these optional inputs will cause the default value to be used.
C     Thus to use a subset of the optional inputs, simply preload
C     locations 5 to 10 in RWORK and IWORK to 0.0 and 0 respectively,
C     and then set those of interest to nonzero values.
C
C     Name    Location   Meaning and default value
C     ------  ---------  -----------------------------------------------
C     H0      RWORK(5)   Step size to be attempted on the first step.
C                        The default value is determined by the solver.
C     HMAX    RWORK(6)   Maximum absolute step size allowed.  The
C                        default value is infinite.
C     HMIN    RWORK(7)   Minimum absolute step size allowed.  The
C                        default value is 0.  (This lower bound is not
C                        enforced on the final step before reaching
C                        TCRIT when ITASK = 4 or 5.)
C     MAXORD  IWORK(5)   Maximum order to be allowed.  The default value
C                        is 12 if METH = 1, and 5 if METH = 2. (See the
C                        MF description above for METH.)  If MAXORD
C                        exceeds the default value, it will be reduced
C                        to the default value.  If MAXORD is changed
C                        during the problem, it may cause the current
C                        order to be reduced.
C     MXSTEP  IWORK(6)   Maximum number of (internally defined) steps
C                        allowed during one call to the solver.  The
C                        default value is 500.
C     MXHNIL  IWORK(7)   Maximum number of messages printed (per
C                        problem) warning that T + H = T on a step
C                        (H = step size).  This must be positive to
C                        result in a nondefault value.  The default
C                        value is 10.
C
C     Optional Outputs
C     ----------------
C     As optional additional output from DLSODE, the variables listed
C     below are quantities related to the performance of DLSODE which
C     are available to the user.  These are communicated by way of the
C     work arrays, but also have internal mnemonic names as shown.
C     Except where stated otherwise, all of these outputs are defined on
C     any successful return from DLSODE, and on any return with ISTATE =
C     -1, -2, -4, -5, or -6.  On an illegal input return (ISTATE = -3),
C     they will be unchanged from their existing values (if any), except
C     possibly for TOLSF, LENRW, and LENIW.  On any error return,
C     outputs relevant to the error will be defined, as noted below.
C
C     Name   Location   Meaning
C     -----  ---------  ------------------------------------------------
C     HU     RWORK(11)  Step size in t last used (successfully).
C     HCUR   RWORK(12)  Step size to be attempted on the next step.
C     TCUR   RWORK(13)  Current value of the independent variable which
C                       the solver has actually reached, i.e., the
C                       current internal mesh point in t. On output,
C                       TCUR will always be at least as far as the
C                       argument T, but may be farther (if interpolation
C                       was done).
C     TOLSF  RWORK(14)  Tolerance scale factor, greater than 1.0,
C                       computed when a request for too much accuracy
C                       was detected (ISTATE = -3 if detected at the
C                       start of the problem, ISTATE = -2 otherwise).
C                       If ITOL is left unaltered but RTOL and ATOL are
C                       uniformly scaled up by a factor of TOLSF for the
C                       next call, then the solver is deemed likely to
C                       succeed.  (The user may also ignore TOLSF and
C                       alter the tolerance parameters in any other way
C                       appropriate.)
C     NST    IWORK(11)  Number of steps taken for the problem so far.
C     NFE    IWORK(12)  Number of F evaluations for the problem so far.
C     NJE    IWORK(13)  Number of Jacobian evaluations (and of matrix LU
C                       decompositions) for the problem so far.
C     NQU    IWORK(14)  Method order last used (successfully).
C     NQCUR  IWORK(15)  Order to be attempted on the next step.
C     IMXER  IWORK(16)  Index of the component of largest magnitude in
C                       the weighted local error vector ( e(i)/EWT(i) ),
C                       on an error return with ISTATE = -4 or -5.
C     LENRW  IWORK(17)  Length of RWORK actually required.  This is
C                       defined on normal returns and on an illegal
C                       input return for insufficient storage.
C     LENIW  IWORK(18)  Length of IWORK actually required.  This is
C                       defined on normal returns and on an illegal
C                       input return for insufficient storage.
C
C     The following two arrays are segments of the RWORK array which may
C     also be of interest to the user as optional outputs.  For each
C     array, the table below gives its internal name, its base address
C     in RWORK, and its description.
C
C     Name  Base address  Description
C     ----  ------------  ----------------------------------------------
C     YH    21            The Nordsieck history array, of size NYH by
C                         (NQCUR + 1), where NYH is the initial value of
C                         NEQ.  For j = 0,1,...,NQCUR, column j + 1 of
C                         YH contains HCUR**j/factorial(j) times the jth
C                         derivative of the interpolating polynomial
C                         currently representing the solution, evaluated
C                         at t = TCUR.
C     ACOR  LENRW-NEQ+1   Array of size NEQ used for the accumulated
C                         corrections on each step, scaled on output to
C                         represent the estimated local error in Y on
C                         the last step.  This is the vector e in the
C                         description of the error control.  It is
C                         defined only on successful return from DLSODE.
C
C
C                    Part 2.  Other Callable Routines
C                    --------------------------------
C
C     The following are optional calls which the user may make to gain
C     additional capabilities in conjunction with DLSODE.
C
C     Form of call              Function
C     ------------------------  ----------------------------------------
C     CALL XSETUN(LUN)          Set the logical unit number, LUN, for
C                               output of messages from DLSODE, if the
C                               default is not desired.  The default
C                               value of LUN is 6. This call may be made
C                               at any time and will take effect
C                               immediately.
C     CALL XSETF(MFLAG)         Set a flag to control the printing of
C                               messages by DLSODE.  MFLAG = 0 means do
C                               not print.  (Danger:  this risks losing
C                               valuable information.)  MFLAG = 1 means
C                               print (the default).  This call may be
C                               made at any time and will take effect
C                               immediately.
C     CALL DSRCOM(RSAV,ISAV,JOB)  Saves and restores the contents of the
C                               internal COMMON blocks used by DLSODE
C                               (see Part 3 below).  RSAV must be a
C                               real array of length 218 or more, and
C                               ISAV must be an integer array of length
C                               37 or more.  JOB = 1 means save COMMON
C                               into RSAV/ISAV.  JOB = 2 means restore
C                               COMMON from same.  DSRCOM is useful if
C                               one is interrupting a run and restarting
C                               later, or alternating between two or
C                               more problems solved with DLSODE.
C     CALL DINTDY(,,,,,)        Provide derivatives of y, of various
C     (see below)               orders, at a specified point t, if
C                               desired.  It may be called only after a
C                               successful return from DLSODE.  Detailed
C                               instructions follow.
C
C     Detailed instructions for using DINTDY
C     --------------------------------------
C     The form of the CALL is:
C
C           CALL DINTDY (T, K, RWORK(21), NYH, DKY, IFLAG)
C
C     The input parameters are:
C
C     T          Value of independent variable where answers are
C                desired (normally the same as the T last returned by
C                DLSODE).  For valid results, T must lie between
C                TCUR - HU and TCUR.  (See "Optional Outputs" above
C                for TCUR and HU.)
C     K          Integer order of the derivative desired.  K must
C                satisfy 0 <= K <= NQCUR, where NQCUR is the current
C                order (see "Optional Outputs").  The capability
C                corresponding to K = 0, i.e., computing y(t), is
C                already provided by DLSODE directly.  Since
C                NQCUR >= 1, the first derivative dy/dt is always
C                available with DINTDY.
C     RWORK(21)  The base address of the history array YH.
C     NYH        Column length of YH, equal to the initial value of NEQ.
C
C     The output parameters are:
C
C     DKY        Real array of length NEQ containing the computed value
C                of the Kth derivative of y(t).
C     IFLAG      Integer flag, returned as 0 if K and T were legal,
C                -1 if K was illegal, and -2 if T was illegal.
C                On an error return, a message is also written.
C
C
C                          Part 3.  Common Blocks
C                          ----------------------
C
C     If DLSODE is to be used in an overlay situation, the user must
C     declare, in the primary overlay, the variables in:
C     (1) the call sequence to DLSODE,
C     (2) the internal COMMON block /DLS001/, of length 255
C         (218 double precision words followed by 37 integer words).
C
C     If DLSODE is used on a system in which the contents of internal
C     COMMON blocks are not preserved between calls, the user should
C     declare the above COMMON block in his main program to insure that
C     its contents are preserved.
C
C     If the solution of a given problem by DLSODE is to be interrupted
C     and then later continued, as when restarting an interrupted run or
C     alternating between two or more problems, the user should save,
C     following the return from the last DLSODE call prior to the
C     interruption, the contents of the call sequence variables and the
C     internal COMMON block, and later restore these values before the
C     next DLSODE call for that problem.   In addition, if XSETUN and/or
C     XSETF was called for non-default handling of error messages, then
C     these calls must be repeated.  To save and restore the COMMON
C     block, use subroutine DSRCOM (see Part 2 above).
C
C
C              Part 4.  Optionally Replaceable Solver Routines
C              -----------------------------------------------
C
C     Below are descriptions of two routines in the DLSODE package which
C     relate to the measurement of errors.  Either routine can be
C     replaced by a user-supplied version, if desired.  However, since
C     such a replacement may have a major impact on performance, it
C     should be done only when absolutely necessary, and only with great
C     caution.  (Note:  The means by which the package version of a
C     routine is superseded by the user's version may be system-
C     dependent.)
C
C     DEWSET
C     ------
C     The following subroutine is called just before each internal
C     integration step, and sets the array of error weights, EWT, as
C     described under ITOL/RTOL/ATOL above:
C
C           SUBROUTINE DEWSET (NEQ, ITOL, RTOL, ATOL, YCUR, EWT)
C
C     where NEQ, ITOL, RTOL, and ATOL are as in the DLSODE call
C     sequence, YCUR contains the current dependent variable vector,
C     and EWT is the array of weights set by DEWSET.
C
C     If the user supplies this subroutine, it must return in EWT(i)
C     (i = 1,...,NEQ) a positive quantity suitable for comparing errors
C     in Y(i) to.  The EWT array returned by DEWSET is passed to the
C     DVNORM routine (see below), and also used by DLSODE in the
C     computation of the optional output IMXER, the diagonal Jacobian
C     approximation, and the increments for difference quotient
C     Jacobians.
C
C     In the user-supplied version of DEWSET, it may be desirable to use
C     the current values of derivatives of y. Derivatives up to order NQ
C     are available from the history array YH, described above under
C     optional outputs.  In DEWSET, YH is identical to the YCUR array,
C     extended to NQ + 1 columns with a column length of NYH and scale
C     factors of H**j/factorial(j).  On the first call for the problem,
C     given by NST = 0, NQ is 1 and H is temporarily set to 1.0.
C     NYH is the initial value of NEQ.  The quantities NQ, H, and NST
C     can be obtained by including in SEWSET the statements:
C           DOUBLE PRECISION RLS
C           COMMON /DLS001/ RLS(218),ILS(37)
C           NQ = ILS(33)
C           NST = ILS(34)
C           H = RLS(212)
C     Thus, for example, the current value of dy/dt can be obtained as
C     YCUR(NYH+i)/H (i=1,...,NEQ) (and the division by H is unnecessary
C     when NST = 0).
C
C     DVNORM
C     ------
C     DVNORM is a real function routine which computes the weighted
C     root-mean-square norm of a vector v:
C
C        d = DVNORM (n, v, w)
C
C     where:
C     n = the length of the vector,
C     v = real array of length n containing the vector,
C     w = real array of length n containing weights,
C     d = SQRT( (1/n) * sum(v(i)*w(i))**2 ).
C
C     DVNORM is called with n = NEQ and with w(i) = 1.0/EWT(i), where
C     EWT is as set by subroutine DEWSET.
C
C     If the user supplies this function, it should return a nonnegative
C     value of DVNORM suitable for use in the error control in DLSODE.
C     None of the arguments should be altered by DVNORM.  For example, a
C     user-supplied DVNORM routine might:
C     - Substitute a max-norm of (v(i)*w(i)) for the rms-norm, or
C     - Ignore some components of v in the norm, with the effect of
C       suppressing the error control on those components of Y.
C  ---------------------------------------------------------------------
C***ROUTINES CALLED  DEWSET, DINTDY, DUMACH, DSTODE, DVNORM, DLSODE_ERRWD
C***COMMON BLOCKS    DLS001
C***REVISION HISTORY  (YYYYMMDD)
C 19791129  DATE WRITTEN
C 19791213  Minor changes to declarations; DELP init. in STODE.
C 19800118  Treat NEQ as array; integer declarations added throughout;
C           minor changes to prologue.
C 19800306  Corrected TESCO(1,NQP1) setting in CFODE.
C 19800519  Corrected access of YH on forced order reduction;
C           numerous corrections to prologues and other comments.
C 19800617  In main driver, added loading of SQRT(UROUND) in RWORK;
C           minor corrections to main prologue.
C 19800923  Added zero initialization of HU and NQU.
C 19801218  Revised DLSODE_ERRWD routine; minor corrections to main prologue.
C 19810401  Minor changes to comments and an error message.
C 19810814  Numerous revisions: replaced EWT by 1/EWT; used flags
C           JCUR, ICF, IERPJ, IERSL between STODE and subordinates;
C           added tuning parameters CCMAX, MAXCOR, MSBP, MXNCF;
C           reorganized returns from STODE; reorganized type decls.;
C           fixed message length in DLSODE_ERRWD; changed default LUNIT to 6;
C           changed Common lengths; changed comments throughout.
C 19870330  Major update by ACH: corrected comments throughout;
C           removed TRET from Common; rewrote EWSET with 4 loops;
C           fixed t test in INTDY; added Cray directives in STODE;
C           in STODE, fixed DELP init. and logic around PJAC call;
C           combined routines to save/restore Common;
C           passed LEVEL = 0 in error message calls (except run abort).
C 19890426  Modified prologue to SLATEC/LDOC format.  (FNF)
C 19890501  Many improvements to prologue.  (FNF)
C 19890503  A few final corrections to prologue.  (FNF)
C 19890504  Minor cosmetic changes.  (FNF)
C 19890510  Corrected description of Y in Arguments section.  (FNF)
C 19890517  Minor corrections to prologue.  (FNF)
C 19920514  Updated with prologue edited 891025 by G. Shaw for manual.
C 19920515  Converted source lines to upper case.  (FNF)
C 19920603  Revised DLSODE_ERRWD calls using mixed upper-lower case.  (ACH)
C 19920616  Revised prologue comment regarding CFT.  (ACH)
C 19921116  Revised prologue comments regarding Common.  (ACH).
C 19930326  Added comment about non-reentrancy.  (FNF)
C 19930723  Changed D1MACH to DUMACH. (FNF)
C 19930801  Removed ILLIN and NTREP from Common (affects driver logic);
C           minor changes to prologue and internal comments;
C           changed Hollerith strings to quoted strings;
C           changed internal comments to mixed case;
C           replaced DLSODE_ERRWD with new version using character type;
C           changed dummy dimensions from 1 to *. (ACH)
C 19930809  Changed to generic intrinsic names; changed names of
C           subprograms and Common blocks to DLSODE etc. (ACH)
C 19930929  Eliminated use of REAL intrinsic; other minor changes. (ACH)
C 20010412  Removed all 'own' variables from Common block /DLS001/
C           (affects declarations in 6 routines). (ACH)
C 20010509  Minor corrections to prologue. (ACH)
C 20031105  Restored 'own' variables to Common block /DLS001/, to
C           enable interrupt/restart feature. (ACH)
C 20031112  Added SAVE statements for data-loaded constants.
C
C***END PROLOGUE  DLSODE
C
C*Internal Notes:
C
C Other Routines in the DLSODE Package.
C
C In addition to Subroutine DLSODE, the DLSODE package includes the
C following subroutines and function routines:
C  DINTDY   computes an interpolated value of the y vector at t = TOUT.
C  DSTODE   is the core integrator, which does one step of the
C           integration and the associated error control.
C  DCFODE   sets all method coefficients and test constants.
C  DPREPJ   computes and preprocesses the Jacobian matrix J = df/dy
C           and the Newton iteration matrix P = I - h*l0*J.
C  DSOLSY   manages solution of linear system in chord iteration.
C  DEWSET   sets the error weight vector EWT before each step.
C  DVNORM   computes the weighted R.M.S. norm of a vector.
C  DSRCOM   is a user-callable routine to save and restore
C           the contents of the internal Common block.
C  DGEFA and DGESL   are routines from LINPACK for solving full
C           systems of linear algebraic equations.
C  DGBFA and DGBSL   are routines from LINPACK for solving banded
C           linear systems.
C  DUMACH   computes the unit roundoff in a machine-independent manner.
C  DLSODE_ERRWD, XSETUN, XSETF, IXSAV, IUMACH   handle the printing of all
C           error messages and warnings.  DLSODE_ERRWD is machine-dependent.
C Note: DVNORM, DUMACH, IXSAV, and IUMACH are function routines.
C All the others are subroutines.
C
C**End
C
C  Declare externals.
      EXTERNAL DPREPJ, DSOLSY
      DOUBLE PRECISION DUMACH, DVNORM
C
C  Declare all other variables.
      INTEGER INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH,
     1   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     2   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     3   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     4   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      INTEGER I, I1, I2, IFLAG, IMXER, KGO, LF0,
     1   LENIW, LENRW, LENWM, ML, MORD, MU, MXHNL0, MXSTP0
      DOUBLE PRECISION CONIT, CRATE, EL, ELCO, HOLD, RMAX, TESCO,
     1   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND
      DOUBLE PRECISION ATOLI, AYI, BIG, EWTI, H0, HMAX, HMX, RH, RTOLI,
     1   TCRIT, TDIST, TNEXT, TOL, TOLSF, TP, SIZE, SUM, W0
      DIMENSION MORD(2)
      LOGICAL IHIT
      CHARACTER(LEN=80) MSG
      SAVE MORD, MXSTP0, MXHNL0
C-----------------------------------------------------------------------
C The following internal Common block contains
C (a) variables which are local to any subroutine but whose values must
C     be preserved between calls to the routine ("own" variables), and
C (b) variables which are communicated between subroutines.
C The block DLS001 is declared in subroutines DLSODE, DINTDY, DSTODE,
C DPREPJ, and DSOLSY.
C Groups of variables are replaced by dummy arrays in the Common
C declarations in routines where those variables are not used.
C-----------------------------------------------------------------------
      COMMON /DLS001/ CONIT, CRATE, EL(13), ELCO(13,12),
     1   HOLD, RMAX, TESCO(3,12),
     2   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND,
     3   INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH,
     4   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     5   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     6   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     7   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
C
      DATA  MORD(1),MORD(2)/12,5/, MXSTP0/500/, MXHNL0/10/
C-----------------------------------------------------------------------
C Block A.
C This code block is executed on every call.
C It tests ISTATE and ITASK for legality and branches appropriately.
C If ISTATE .GT. 1 but the flag INIT shows that initialization has
C not yet been done, an error return occurs.
C If ISTATE = 1 and TOUT = T, return immediately.
C-----------------------------------------------------------------------
C KARLINE: INITIALISED IHIT TO AVOID COMPILER WARNINGS - SHOULD HAVE NO EFFEXT
      IHIT = .TRUE.
      LENWM = 0
C***FIRST EXECUTABLE STATEMENT  DLSODE
      IF (ISTATE .LT. 1 .OR. ISTATE .GT. 3) GO TO 601
      IF (ITASK .LT. 1 .OR. ITASK .GT. 5) GO TO 602
      IF (ISTATE .EQ. 1) GO TO 10
      IF (INIT .EQ. 0) GO TO 603
      IF (ISTATE .EQ. 2) GO TO 200
      GO TO 20
 10   INIT = 0
      IF (TOUT .EQ. T) RETURN
C-----------------------------------------------------------------------
C Block B.
C The next code block is executed for the initial call (ISTATE = 1),
C or for a continuation call with parameter changes (ISTATE = 3).
C It contains checking of all inputs and various initializations.
C
C First check legality of the non-optional inputs NEQ, ITOL, IOPT,
C MF, ML, and MU.
C-----------------------------------------------------------------------
 20   IF (NEQ(1) .LE. 0) GO TO 604
      IF (ISTATE .EQ. 1) GO TO 25
      IF (NEQ(1) .GT. N) GO TO 605
 25   N = NEQ(1)
      IF (ITOL .LT. 1 .OR. ITOL .GT. 4) GO TO 606
      IF (IOPT .LT. 0 .OR. IOPT .GT. 1) GO TO 607
      METH = MF/10
      MITER = MF - 10*METH
      IF (METH .LT. 1 .OR. METH .GT. 2) GO TO 608
      IF (MITER .LT. 0 .OR. MITER .GT. 5) GO TO 608
      IF (MITER .LE. 3) GO TO 30
      ML = IWORK(1)
      MU = IWORK(2)
      IF (ML .LT. 0 .OR. ML .GE. N) GO TO 609
      IF (MU .LT. 0 .OR. MU .GE. N) GO TO 610
 30   CONTINUE
C Next process and check the optional inputs. --------------------------
      IF (IOPT .EQ. 1) GO TO 40
      MAXORD = MORD(METH)
      MXSTEP = MXSTP0
      MXHNIL = MXHNL0
      IF (ISTATE .EQ. 1) H0 = 0.0D0
      HMXI = 0.0D0
      HMIN = 0.0D0
      GO TO 60
 40   MAXORD = IWORK(5)
      IF (MAXORD .LT. 0) GO TO 611
      IF (MAXORD .EQ. 0) MAXORD = 100
      MAXORD = MIN(MAXORD,MORD(METH))
      MXSTEP = IWORK(6)
      IF (MXSTEP .LT. 0) GO TO 612
      IF (MXSTEP .EQ. 0) MXSTEP = MXSTP0
      MXHNIL = IWORK(7)
      IF (MXHNIL .LT. 0) GO TO 613
      IF (MXHNIL .EQ. 0) MXHNIL = MXHNL0
      IF (ISTATE .NE. 1) GO TO 50
      H0 = RWORK(5)
      IF ((TOUT - T)*H0 .LT. 0.0D0) GO TO 614
 50   HMAX = RWORK(6)
      IF (HMAX .LT. 0.0D0) GO TO 615
      HMXI = 0.0D0
      IF (HMAX .GT. 0.0D0) HMXI = 1.0D0/HMAX
      HMIN = RWORK(7)
      IF (HMIN .LT. 0.0D0) GO TO 616
C-----------------------------------------------------------------------
C Set work array pointers and check lengths LRW and LIW.
C Pointers to segments of RWORK and IWORK are named by prefixing L to
C the name of the segment.  E.g., the segment YH starts at RWORK(LYH).
C Segments of RWORK (in order) are denoted  YH, WM, EWT, SAVF, ACOR.
C-----------------------------------------------------------------------
 60   LYH = 21
      IF (ISTATE .EQ. 1) NYH = N
      LWM = LYH + (MAXORD + 1)*NYH
      IF (MITER .EQ. 0) LENWM = 0
      IF (MITER .EQ. 1 .OR. MITER .EQ. 2) LENWM = N*N + 2
      IF (MITER .EQ. 3) LENWM = N + 2
      IF (MITER .GE. 4) LENWM = (2*ML + MU + 1)*N + 2
      LEWT = LWM + LENWM
      LSAVF = LEWT + N
      LACOR = LSAVF + N
      LENRW = LACOR + N - 1
      IWORK(17) = LENRW
      LIWM = 1
      LENIW = 20 + N
      IF (MITER .EQ. 0 .OR. MITER .EQ. 3) LENIW = 20
      IWORK(18) = LENIW
      IF (LENRW .GT. LRW) GO TO 617
      IF (LENIW .GT. LIW) GO TO 618
C Check RTOL and ATOL for legality. ------------------------------------
      RTOLI = RTOL(1)
      ATOLI = ATOL(1)
      DO 70 I = 1,N
        IF (ITOL .GE. 3) RTOLI = RTOL(I)
        IF (ITOL .EQ. 2 .OR. ITOL .EQ. 4) ATOLI = ATOL(I)
        IF (RTOLI .LT. 0.0D0) GO TO 619
        IF (ATOLI .LT. 0.0D0) GO TO 620
 70     CONTINUE
      IF (ISTATE .EQ. 1) GO TO 100
C If ISTATE = 3, set flag to signal parameter changes to DSTODE. -------
      JSTART = -1
      IF (NQ .LE. MAXORD) GO TO 90
C MAXORD was reduced below NQ.  Copy YH(*,MAXORD+2) into SAVF. ---------
      DO 80 I = 1,N
        RWORK(I+LSAVF-1) = RWORK(I+LWM-1)
 80   CONTINUE

C Reload WM(1) = RWORK(LWM), since LWM may have changed. ---------------
 90   IF (MITER .GT. 0) RWORK(LWM) = SQRT(UROUND)
      IF (N .EQ. NYH) GO TO 200
C NEQ was reduced.  Zero part of YH to avoid undefined references. -----
      I1 = LYH + L*NYH
      I2 = LYH + (MAXORD + 1)*NYH - 1
      IF (I1 .GT. I2) GO TO 200
      DO 95 I = I1,I2
        RWORK(I) = 0.0D0
 95   CONTINUE
      GO TO 200
C-----------------------------------------------------------------------
C Block C.
C The next block is for the initial call only (ISTATE = 1).
C It contains all remaining initializations, the initial call to F,
C and the calculation of the initial step size.
C The error weights in EWT are inverted after being loaded.
C-----------------------------------------------------------------------
 100  UROUND = DUMACH()
      TN = T
      IF (ITASK .NE. 4 .AND. ITASK .NE. 5) GO TO 110
      TCRIT = RWORK(1)
      IF ((TCRIT - TOUT)*(TOUT - T) .LT. 0.0D0) GO TO 625
      IF (H0 .NE. 0.0D0 .AND. (T + H0 - TCRIT)*H0 .GT. 0.0D0)
     1   H0 = TCRIT - T
 110  JSTART = 0
      IF (MITER .GT. 0) RWORK(LWM) = SQRT(UROUND)
      NHNIL = 0
      NST = 0
      NJE = 0
      NSLAST = 0
      HU = 0.0D0
      NQU = 0
      CCMAX = 0.3D0
      MAXCOR = 3
      MSBP = 20
      MXNCF = 10
C Initial call to F.  (LF0 points to YH(*,2).) -------------------------
      LF0 = LYH + NYH
      CALL F (NEQ, T, Y, RWORK(LF0), rpar, ipar)
      NFE = 1
C Load the initial value vector in YH. ---------------------------------
      DO 115 I = 1,N
        RWORK(I+LYH-1) = Y(I)
 115  CONTINUE
C Load and invert the EWT array.  (H is temporarily set to 1.0.) -------
      NQ = 1
      H = 1.0D0
      CALL DEWSET (N, ITOL, RTOL, ATOL, RWORK(LYH), RWORK(LEWT))
      DO 120 I = 1,N
        IF (RWORK(I+LEWT-1) .LE. 0.0D0) GO TO 621
        RWORK(I+LEWT-1) = 1.0D0/RWORK(I+LEWT-1)
 120  CONTINUE
C-----------------------------------------------------------------------
C The coding below computes the step size, H0, to be attempted on the
C first step, unless the user has supplied a value for this.
C First check that TOUT - T differs significantly from zero.
C A scalar tolerance quantity TOL is computed, as MAX(RTOL(I))
C if this is positive, or MAX(ATOL(I)/ABS(Y(I))) otherwise, adjusted
C so as to be between 100*UROUND and 1.0E-3.
C Then the computed value H0 is given by..
C                                      NEQ
C   H0**2 = TOL / ( w0**-2 + (1/NEQ) * SUM ( f(i)/ywt(i) )**2  )
C                                       1
C where   w0     = MAX ( ABS(T), ABS(TOUT) ),
C         f(i)   = i-th component of initial value of f,
C         ywt(i) = EWT(i)/TOL  (a weight for y(i)).
C The sign of H0 is inferred from the initial values of TOUT and T.
C-----------------------------------------------------------------------
      IF (H0 .NE. 0.0D0) GO TO 180
      TDIST = ABS(TOUT - T)
      W0 = MAX(ABS(T),ABS(TOUT))
      IF (TDIST .LT. 2.0D0*UROUND*W0) GO TO 622
      TOL = RTOL(1)
      IF (ITOL .LE. 2) GO TO 140
      DO 130 I = 1,N
        TOL = MAX(TOL,RTOL(I))
 130  CONTINUE
 140  IF (TOL .GT. 0.0D0) GO TO 160
      ATOLI = ATOL(1)
      DO 150 I = 1,N
        IF (ITOL .EQ. 2 .OR. ITOL .EQ. 4) ATOLI = ATOL(I)
        AYI = ABS(Y(I))
        IF (AYI .NE. 0.0D0) TOL = MAX(TOL,ATOLI/AYI)
 150    CONTINUE
 160  TOL = MAX(TOL,100.0D0*UROUND)
      TOL = MIN(TOL,0.001D0)
      SUM = DVNORM (N, RWORK(LF0), RWORK(LEWT))
      SUM = 1.0D0/(TOL*W0*W0) + TOL*SUM**2
      H0 = 1.0D0/SQRT(SUM)
      H0 = MIN(H0,TDIST)
      H0 = SIGN(H0,TOUT-T)
C Adjust H0 if necessary to meet HMAX bound. ---------------------------
 180  RH = ABS(H0)*HMXI
      IF (RH .GT. 1.0D0) H0 = H0/RH
C Load H with H0 and scale YH(*,2) by H0. ------------------------------
      H = H0
      DO 190 I = 1,N
        RWORK(I+LF0-1) = H0*RWORK(I+LF0-1)
 190  CONTINUE
      GO TO 270
C-----------------------------------------------------------------------
C Block D.
C The next code block is for continuation calls only (ISTATE = 2 or 3)
C and is to check stop conditions before taking a step.
C-----------------------------------------------------------------------
 200  NSLAST = NST
      IF (ITASK .EQ. 1) THEN
        GOTO 210
      ELSE IF (ITASK .EQ. 2) THEN
        GOTO 250
      ELSE IF (ITASK .EQ. 3) THEN
        GOTO 220
      ELSE IF (ITASK .EQ. 4) THEN
        GOTO 230
      ELSE IF (ITASK .EQ. 5) THEN
        GOTO 240
      ENDIF
C      GO TO (210, 250, 220, 230, 240), ITASK
 210  IF ((TN - TOUT)*H .LT. 0.0D0) GO TO 250
      CALL DINTDY (TOUT, 0, RWORK(LYH), NYH, Y, IFLAG)
      IF (IFLAG .NE. 0) GO TO 627
      T = TOUT
      GO TO 420
 220  TP = TN - HU*(1.0D0 + 100.0D0*UROUND)
      IF ((TP - TOUT)*H .GT. 0.0D0) GO TO 623
      IF ((TN - TOUT)*H .LT. 0.0D0) GO TO 250
      GO TO 400
 230  TCRIT = RWORK(1)
      IF ((TN - TCRIT)*H .GT. 0.0D0) GO TO 624
      IF ((TCRIT - TOUT)*H .LT. 0.0D0) GO TO 625
      IF ((TN - TOUT)*H .LT. 0.0D0) GO TO 245
      CALL DINTDY (TOUT, 0, RWORK(LYH), NYH, Y, IFLAG)
      IF (IFLAG .NE. 0) GO TO 627
      T = TOUT
      GO TO 420
 240  TCRIT = RWORK(1)
      IF ((TN - TCRIT)*H .GT. 0.0D0) GO TO 624
 245  HMX = ABS(TN) + ABS(H)
      IHIT = ABS(TN - TCRIT) .LE. 100.0D0*UROUND*HMX
      IF (IHIT) GO TO 400
      TNEXT = TN + H*(1.0D0 + 4.0D0*UROUND)
      IF ((TNEXT - TCRIT)*H .LE. 0.0D0) GO TO 250
      H = (TCRIT - TN)*(1.0D0 - 4.0D0*UROUND)
      IF (ISTATE .EQ. 2) JSTART = -2
C-----------------------------------------------------------------------
C Block E.
C The next block is normally executed for all calls and contains
C the call to the one-step core integrator DSTODE.
C
C This is a looping point for the integration steps.
C
C First check for too many steps being taken, update EWT (if not at
C start of problem), check for too much accuracy being requested, and
C check for H below the roundoff level in T.
C-----------------------------------------------------------------------
 250  CONTINUE
      IF ((NST-NSLAST) .GE. MXSTEP) GO TO 500
      CALL DEWSET (N, ITOL, RTOL, ATOL, RWORK(LYH), RWORK(LEWT))
      DO 260 I = 1,N
        IF (RWORK(I+LEWT-1) .LE. 0.0D0) GO TO 510
        RWORK(I+LEWT-1) = 1.0D0/RWORK(I+LEWT-1)
 260  CONTINUE
 270  TOLSF = UROUND*DVNORM (N, RWORK(LYH), RWORK(LEWT))
      IF (TOLSF .LE. 1.0D0) GO TO 280
      TOLSF = TOLSF*2.0D0
      IF (NST .EQ. 0) GO TO 626
      GO TO 520
 280  IF ((TN + H) .NE. TN) GO TO 290
      NHNIL = NHNIL + 1
      IF (NHNIL .GT. MXHNIL) GO TO 290
      MSG = 'DLSODE-  Warning..internal T (=R1) and H (=R2) are'
      CALL DLSODE_ERRWD (MSG, 50, 101, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      MSG='      such that in the machine, T + H = T on the next step  '
      CALL DLSODE_ERRWD (MSG, 60, 101, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      MSG = '      (H = step size). Solver will continue anyway'
      CALL DLSODE_ERRWD (MSG, 50, 101, 0, 0, 0, 0, 2, TN, H)
      IF (NHNIL .LT. MXHNIL) GO TO 290
      MSG = 'DLSODE-  Above warning has been issued I1 times.  '
      CALL DLSODE_ERRWD (MSG, 50, 102, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      MSG = '      It will not be issued again for this problem'
      CALL DLSODE_ERRWD (MSG, 50, 102, 0, 1, MXHNIL, 0, 0, 0.0D0, 0.0D0)
 290  CONTINUE
C-----------------------------------------------------------------------
C  CALL DSTODE(NEQ,Y,YH,NYH,YH,EWT,SAVF,ACOR,WM,IWM,F,JAC,DPREPJ,DSOLSY)
C-----------------------------------------------------------------------
      CALL DSTODE (NEQ, Y, RWORK(LYH), NYH, RWORK(LYH), RWORK(LEWT),
     1   RWORK(LSAVF), RWORK(LACOR), RWORK(LWM), IWORK(LIWM),
     2   F, JAC, DPREPJ, DSOLSY, rpar,ipar)
      KGO = 1 - KFLAG
      IF (KGO .EQ. 1) THEN
         GOTO 300
      ELSE IF (KGO .EQ. 2) THEN
         GOTO 530
      ELSE IF (KGO .EQ. 3) THEN
         GOTO 540
      ENDIF
C      GO TO (300, 530, 540), KGO
C-----------------------------------------------------------------------
C Block F.
C The following block handles the case of a successful return from the
C core integrator (KFLAG = 0).  Test for stop conditions.
C-----------------------------------------------------------------------
 300  INIT = 1

      IF (ITASK .EQ. 1) THEN
         GOTO 310
      ELSE IF (ITASK .EQ. 2) THEN
         GOTO 400
      ELSE IF (ITASK .EQ. 3) THEN
         GOTO 330
      ELSE IF (ITASK .EQ. 4) THEN
         GOTO 340
      ELSE IF (ITASK .EQ. 5) THEN
         GOTO 350
      ENDIF
C karline: change from
C      GO TO (310, 400, 330, 340, 350), ITASK
C ITASK = 1.  If TOUT has been reached, interpolate. -------------------
 310  IF ((TN - TOUT)*H .LT. 0.0D0) GO TO 250
      CALL DINTDY (TOUT, 0, RWORK(LYH), NYH, Y, IFLAG)
      T = TOUT
      GO TO 420
C ITASK = 3.  Jump to exit if TOUT was reached. ------------------------
 330  IF ((TN - TOUT)*H .GE. 0.0D0) GO TO 400
      GO TO 250
C ITASK = 4.  See if TOUT or TCRIT was reached.  Adjust H if necessary.
 340  IF ((TN - TOUT)*H .LT. 0.0D0) GO TO 345
      CALL DINTDY (TOUT, 0, RWORK(LYH), NYH, Y, IFLAG)
      T = TOUT
      GO TO 420
 345  HMX = ABS(TN) + ABS(H)
      IHIT = ABS(TN - TCRIT) .LE. 100.0D0*UROUND*HMX
      IF (IHIT) GO TO 400
      TNEXT = TN + H*(1.0D0 + 4.0D0*UROUND)
      IF ((TNEXT - TCRIT)*H .LE. 0.0D0) GO TO 250
      H = (TCRIT - TN)*(1.0D0 - 4.0D0*UROUND)
      JSTART = -2
      GO TO 250
C ITASK = 5.  See if TCRIT was reached and jump to exit. ---------------
 350  HMX = ABS(TN) + ABS(H)
      IHIT = ABS(TN - TCRIT) .LE. 100.0D0*UROUND*HMX
C-----------------------------------------------------------------------
C Block G.
C The following block handles all successful returns from DLSODE.
C If ITASK .NE. 1, Y is loaded from YH and T is set accordingly.
C ISTATE is set to 2, and the optional outputs are loaded into the
C work arrays before returning.
C-----------------------------------------------------------------------
 400  DO 410 I = 1,N
        Y(I) = RWORK(I+LYH-1)
 410  CONTINUE
      T = TN
      IF (ITASK .NE. 4 .AND. ITASK .NE. 5) GO TO 420
      IF (IHIT) T = TCRIT
 420  ISTATE = 2
      RWORK(11) = HU
      RWORK(12) = H
      RWORK(13) = TN
      IWORK(11) = NST
      IWORK(12) = NFE
      IWORK(13) = NJE
      IWORK(14) = NQU
      IWORK(15) = NQ
      RETURN
C-----------------------------------------------------------------------
C Block H.
C The following block handles all unsuccessful returns other than
C those for illegal input.  First the error message routine is called.
C If there was an error test or convergence test failure, IMXER is set.
C Then Y is loaded from YH and T is set to TN.  The optional outputs
C are loaded into the work arrays before returning.
C-----------------------------------------------------------------------
C The maximum number of steps was taken before reaching TOUT. ----------
 500  MSG = 'DLSODE-  At current T (=R1), MXSTEP (=I1) steps   '
      CALL DLSODE_ERRWD (MSG, 50, 201, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      MSG = '      taken on this call before reaching TOUT     '
      CALL DLSODE_ERRWD (MSG, 50, 201, 0, 1, MXSTEP, 0, 1, TN, 0.0D0)
      ISTATE = -1
      GO TO 580
C EWT(I) .LE. 0.0 for some I (not at start of problem). ----------------
 510  EWTI = RWORK(LEWT+I-1)
      MSG = 'DLSODE-  At T (=R1), EWT(I1) has become R2 .LE. 0.'
      CALL DLSODE_ERRWD (MSG, 50, 202, 0, 1, I, 0, 2, TN, EWTI)
      ISTATE = -6
      GO TO 580
C Too much accuracy requested for machine precision. -------------------
 520  MSG = 'DLSODE-  At T (=R1), too much accuracy requested  '
      CALL DLSODE_ERRWD (MSG, 50, 203, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      MSG = '      for precision of machine..  see TOLSF (=R2) '
      CALL DLSODE_ERRWD (MSG, 50, 203, 0, 0, 0, 0, 2, TN, TOLSF)
      RWORK(14) = TOLSF
      ISTATE = -2
      GO TO 580
C KFLAG = -1.  Error test failed repeatedly or with ABS(H) = HMIN. -----
 530  MSG = 'DLSODE-  At T(=R1) and step size H(=R2), the error'
      CALL DLSODE_ERRWD (MSG, 50, 204, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      MSG = '      test failed repeatedly or with ABS(H) = HMIN'
      CALL DLSODE_ERRWD (MSG, 50, 204, 0, 0, 0, 0, 2, TN, H)
      ISTATE = -4
      GO TO 560
C KFLAG = -2.  Convergence failed repeatedly or with ABS(H) = HMIN. ----
 540  MSG = 'DLSODE-  At T (=R1) and step size H (=R2), the    '
      CALL DLSODE_ERRWD (MSG, 50, 205, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      MSG = '      corrector convergence failed repeatedly     '
      CALL DLSODE_ERRWD (MSG, 50, 205, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      MSG = '      or with ABS(H) = HMIN   '
      CALL DLSODE_ERRWD (MSG, 30, 205, 0, 0, 0, 0, 2, TN, H)
      ISTATE = -5
C Compute IMXER if relevant. -------------------------------------------
 560  BIG = 0.0D0
      IMXER = 1
      DO 570 I = 1,N
        SIZE = ABS(RWORK(I+LACOR-1)*RWORK(I+LEWT-1))
        IF (BIG .GE. SIZE) GO TO 570
        BIG = SIZE
        IMXER = I
 570    CONTINUE
      IWORK(16) = IMXER
C Set Y vector, T, and optional outputs. -------------------------------
 580  DO 590 I = 1,N
        Y(I) = RWORK(I+LYH-1)
 590  CONTINUE
      T = TN
      RWORK(11) = HU
      RWORK(12) = H
      RWORK(13) = TN
      IWORK(11) = NST
      IWORK(12) = NFE
      IWORK(13) = NJE
      IWORK(14) = NQU
      IWORK(15) = NQ
      RETURN
C-----------------------------------------------------------------------
C Block I.
C The following block handles all error returns due to illegal input
C (ISTATE = -3), as detected before calling the core integrator.
C First the error message routine is called.  If the illegal input
C is a negative ISTATE, the run is aborted (apparent infinite loop).
C-----------------------------------------------------------------------
 601  MSG = 'DLSODE-  ISTATE (=I1) illegal '
      CALL DLSODE_ERRWD (MSG, 30, 1, 0, 1, ISTATE, 0, 0, 0.0D0, 0.0D0)
      IF (ISTATE .LT. 0) GO TO 800
      GO TO 700
 602  MSG = 'DLSODE-  ITASK (=I1) illegal  '
      CALL DLSODE_ERRWD (MSG, 30, 2, 0, 1, ITASK, 0, 0, 0.0D0, 0.0D0)
      GO TO 700
 603  MSG = 'DLSODE-  ISTATE .GT. 1 but DLSODE not initialized '
      CALL DLSODE_ERRWD (MSG, 50, 3, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      GO TO 700
 604  MSG = 'DLSODE-  NEQ (=I1) .LT. 1     '
      CALL DLSODE_ERRWD (MSG, 30, 4, 0, 1, NEQ(1), 0, 0, 0.0D0, 0.0D0)
      GO TO 700
 605  MSG = 'DLSODE-  ISTATE = 3 and NEQ increased (I1 to I2)  '
      CALL DLSODE_ERRWD (MSG, 50, 5, 0, 2, N, NEQ(1), 0, 0.0D0, 0.0D0)
      GO TO 700
 606  MSG = 'DLSODE-  ITOL (=I1) illegal   '
      CALL DLSODE_ERRWD (MSG, 30, 6, 0, 1, ITOL, 0, 0, 0.0D0, 0.0D0)
      GO TO 700
 607  MSG = 'DLSODE-  IOPT (=I1) illegal   '
      CALL DLSODE_ERRWD (MSG, 30, 7, 0, 1, IOPT, 0, 0, 0.0D0, 0.0D0)
      GO TO 700
 608  MSG = 'DLSODE-  MF (=I1) illegal     '
      CALL DLSODE_ERRWD (MSG, 30, 8, 0, 1, MF, 0, 0, 0.0D0, 0.0D0)
      GO TO 700
 609  MSG = 'DLSODE-  ML (=I1) illegal.. .LT.0 or .GE.NEQ (=I2)'
      CALL DLSODE_ERRWD (MSG, 50, 9, 0, 2, ML, NEQ(1), 0, 0.0D0, 0.0D0)
      GO TO 700
 610  MSG = 'DLSODE-  MU (=I1) illegal.. .LT.0 or .GE.NEQ (=I2)'
      CALL DLSODE_ERRWD (MSG, 50, 10, 0, 2, MU, NEQ(1), 0, 0.0D0, 0.0D0)
      GO TO 700
 611  MSG = 'DLSODE-  MAXORD (=I1) .LT. 0  '
      CALL DLSODE_ERRWD (MSG, 30, 11, 0, 1, MAXORD, 0, 0, 0.0D0, 0.0D0)
      GO TO 700
 612  MSG = 'DLSODE-  MXSTEP (=I1) .LT. 0  '
      CALL DLSODE_ERRWD (MSG, 30, 12, 0, 1, MXSTEP, 0, 0, 0.0D0, 0.0D0)
      GO TO 700
 613  MSG = 'DLSODE-  MXHNIL (=I1) .LT. 0  '
      CALL DLSODE_ERRWD (MSG, 30, 13, 0, 1, MXHNIL, 0, 0, 0.0D0, 0.0D0)
      GO TO 700
 614  MSG = 'DLSODE-  TOUT (=R1) behind T (=R2)      '
      CALL DLSODE_ERRWD (MSG, 40, 14, 0, 0, 0, 0, 2, TOUT, T)
      MSG = '      Integration direction is given by H0 (=R1)  '
      CALL DLSODE_ERRWD (MSG, 50, 14, 0, 0, 0, 0, 1, H0, 0.0D0)
      GO TO 700
 615  MSG = 'DLSODE-  HMAX (=R1) .LT. 0.0  '
      CALL DLSODE_ERRWD (MSG, 30, 15, 0, 0, 0, 0, 1, HMAX, 0.0D0)
      GO TO 700
 616  MSG = 'DLSODE-  HMIN (=R1) .LT. 0.0  '
      CALL DLSODE_ERRWD (MSG, 30, 16, 0, 0, 0, 0, 1, HMIN, 0.0D0)
      GO TO 700
 617  CONTINUE
      MSG='DLSODE-  RWORK length needed, LENRW (=I1), exceeds LRW (=I2)'
      CALL DLSODE_ERRWD (MSG, 60, 17, 0, 2, LENRW, LRW, 0, 0.0D0, 0.0D0)
      GO TO 700
 618  CONTINUE
      MSG='DLSODE-  IWORK length needed, LENIW (=I1), exceeds LIW (=I2)'
      CALL DLSODE_ERRWD (MSG, 60, 18, 0, 2, LENIW, LIW, 0, 0.0D0, 0.0D0)
      GO TO 700
 619  MSG = 'DLSODE-  RTOL(I1) is R1 .LT. 0.0        '
      CALL DLSODE_ERRWD (MSG, 40, 19, 0, 1, I, 0, 1, RTOLI, 0.0D0)
      GO TO 700
 620  MSG = 'DLSODE-  ATOL(I1) is R1 .LT. 0.0        '
      CALL DLSODE_ERRWD (MSG, 40, 20, 0, 1, I, 0, 1, ATOLI, 0.0D0)
      GO TO 700
 621  EWTI = RWORK(LEWT+I-1)
      MSG = 'DLSODE-  EWT(I1) is R1 .LE. 0.0         '
      CALL DLSODE_ERRWD (MSG, 40, 21, 0, 1, I, 0, 1, EWTI, 0.0D0)
      GO TO 700
 622  CONTINUE
      MSG='DLSODE-  TOUT (=R1) too close to T(=R2) to start integration'
      CALL DLSODE_ERRWD (MSG, 60, 22, 0, 0, 0, 0, 2, TOUT, T)
      GO TO 700
 623  CONTINUE
      MSG='DLSODE-  ITASK = I1 and TOUT (=R1) behind TCUR - HU (= R2)  '
      CALL DLSODE_ERRWD (MSG, 60, 23, 0, 1, ITASK, 0, 2, TOUT, TP)
      GO TO 700
 624  CONTINUE
      MSG='DLSODE-  ITASK = 4 OR 5 and TCRIT (=R1) behind TCUR (=R2)   '
      CALL DLSODE_ERRWD (MSG, 60, 24, 0, 0, 0, 0, 2, TCRIT, TN)
      GO TO 700
 625  CONTINUE
      MSG='DLSODE-  ITASK = 4 or 5 and TCRIT (=R1) behind TOUT (=R2)   '
      CALL DLSODE_ERRWD (MSG, 60, 25, 0, 0, 0, 0, 2, TCRIT, TOUT)
      GO TO 700
 626  MSG = 'DLSODE-  At start of problem, too much accuracy   '
      CALL DLSODE_ERRWD (MSG, 50, 26, 0, 0, 0, 0, 0, 0.0D0, 0.0D0)
      MSG='      requested for precision of machine..  See TOLSF (=R1) '
      CALL DLSODE_ERRWD (MSG, 60, 26, 0, 0, 0, 0, 1, TOLSF, 0.0D0)
      RWORK(14) = TOLSF
      GO TO 700
 627  MSG = 'DLSODE-  Trouble in DINTDY.  ITASK = I1, TOUT = R1'
      CALL DLSODE_ERRWD (MSG, 50, 27, 0, 1, ITASK, 0, 1, TOUT, 0.0D0)
C
 700  ISTATE = -3
      RETURN
C
 800  MSG = 'DLSODE-  Run aborted.. apparent infinite loop     '
      CALL DLSODE_ERRWD (MSG, 50, 303, 2, 0, 0, 0, 0, 0.0D0, 0.0D0)
      RETURN
C----------------------- END OF SUBROUTINE DLSODE ----------------------
      END
      SUBROUTINE DPREPJ (NEQ, Y, YH, NYH, EWT, FTEM, SAVF, WM, IWM,
     1   F, JAC,rpar,ipar)
C***BEGIN PROLOGUE  DPREPJ
C***SUBSIDIARY
C***PURPOSE  Compute and process Newton iteration matrix.
C***TYPE      DOUBLE PRECISION (SPREPJ-S, DPREPJ-D)
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***DESCRIPTION
C
C  DPREPJ is called by DSTODE to compute and process the matrix
C  P = I - h*el(1)*J , where J is an approximation to the Jacobian.
C  Here J is computed by the user-supplied routine JAC if
C  MITER = 1 or 4, or by finite differencing if MITER = 2, 3, or 5.
C  If MITER = 3, a diagonal approximation to J is used.
C  J is stored in WM and replaced by P.  If MITER .ne. 3, P is then
C  subjected to LU decomposition in preparation for later solution
C  of linear systems with P as coefficient matrix.  This is done
C  by DGEFA if MITER = 1 or 2, and by DGBFA if MITER = 4 or 5.
C
C  In addition to variables described in DSTODE and DLSODE prologues,
C  communication with DPREPJ uses the following:
C  Y     = array containing predicted values on entry.
C  FTEM  = work array of length N (ACOR in DSTODE).
C  SAVF  = array containing f evaluated at predicted y.
C  WM    = real work space for matrices.  On output it contains the
C          inverse diagonal matrix if MITER = 3 and the LU decomposition
C          of P if MITER is 1, 2 , 4, or 5.
C          Storage of matrix elements starts at WM(3).
C          WM also contains the following matrix-related data:
C          WM(1) = SQRT(UROUND), used in numerical Jacobian increments.
C          WM(2) = H*EL0, saved for later use if MITER = 3.
C  IWM   = integer work space containing pivot information, starting at
C          IWM(21), if MITER is 1, 2, 4, or 5.  IWM also contains band
C          parameters ML = IWM(1) and MU = IWM(2) if MITER is 4 or 5.
C  EL0   = EL(1) (input).
C  IERPJ = output error flag,  = 0 if no trouble, .gt. 0 if
C          P matrix found to be singular.
C  JCUR  = output flag = 1 to indicate that the Jacobian matrix
C          (or approximation) is now current.
C  This routine also uses the COMMON variables EL0, H, TN, UROUND,
C  MITER, N, NFE, and NJE.
C
C***SEE ALSO  DLSODE
C***ROUTINES CALLED  DGBFA, DGEFA, DVNORM
C***COMMON BLOCKS    DLS001
C***REVISION HISTORY  (YYMMDD)
C   791129  DATE WRITTEN
C   890501  Modified prologue to SLATEC/LDOC format.  (FNF)
C   890504  Minor cosmetic changes.  (FNF)
C   930809  Renamed to allow single/double precision versions. (ACH)
C   010418  Reduced size of Common block /DLS001/. (ACH)
C   031105  Restored 'own' variables to Common block /DLS001/, to
C           enable interrupt/restart feature. (ACH)
C***END PROLOGUE  DPREPJ
C**End
      EXTERNAL F, JAC
CKS: added rpar,ipar
      integer ipar(*)
      double precision rpar(*)
      INTEGER NEQ, NYH, IWM
      DOUBLE PRECISION Y, YH, EWT, FTEM, SAVF, WM
      DIMENSION NEQ(*), Y(*), YH(NYH,*), EWT(*), FTEM(*), SAVF(*),
     1   WM(*), IWM(*)
      INTEGER INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH01,
     1   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     2   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     3   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     4   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      DOUBLE PRECISION CONIT, CRATE, EL, ELCO, HOLD, RMAX, TESCO,
     1   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND
      COMMON /DLS001/ CONIT, CRATE, EL(13), ELCO(13,12),
     1   HOLD, RMAX, TESCO(3,12),
     2   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND,
     3   INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH01,
     4   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     5   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     6   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     7   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      INTEGER I, I1, I2, IER, II, J, J1, JJ, LENP,
     1   MBA, MBAND, MEB1, MEBAND, ML, ML3, MU, NP1
      DOUBLE PRECISION CON, DI, FAC, HL0, R, R0, SRUR, YI, YJ, YJJ,
     1   DVNORM
C
C***FIRST EXECUTABLE STATEMENT  DPREPJ
      NJE = NJE + 1
      IERPJ = 0
      JCUR = 1
      HL0 = H*EL0
      IF (MITER .EQ. 1) THEN
        GOTO 100
      ELSE IF (MITER .EQ. 2) THEN
        GOTO 200
      ELSE IF (MITER .EQ. 3) THEN
        GOTO 300
      ELSE IF (MITER .EQ. 4) THEN
        GOTO 400
      ELSE IF (MITER .EQ. 5) THEN
        GOTO 500
      ENDIF

C      GO TO (100, 200, 300, 400, 500), MITER
C If MITER = 1, call JAC and multiply by scalar. -----------------------
 100  LENP = N*N
      DO 110 I = 1,LENP
        WM(I+2) = 0.0D0
 110  CONTINUE
CKS
      CALL JAC (NEQ, TN, Y, 0, 0, WM(3), N,RPAR,IPAR)
      CON = -HL0
      DO 120 I = 1,LENP
        WM(I+2) = WM(I+2)*CON
 120  CONTINUE
      GO TO 240
C If MITER = 2, make N calls to F to approximate J. --------------------
 200  FAC = DVNORM (N, SAVF, EWT)
      R0 = 1000.0D0*ABS(H)*UROUND*N*FAC
      IF (R0 .EQ. 0.0D0) R0 = 1.0D0
      SRUR = WM(1)
      J1 = 2
      DO 230 J = 1,N
        YJ = Y(J)
        R = MAX(SRUR*ABS(YJ),R0/EWT(J))
        Y(J) = Y(J) + R
        FAC = -HL0/R
CKS
        CALL F (NEQ, TN, Y, FTEM, rpar, ipar)
        DO 220 I = 1,N
          WM(I+J1) = (FTEM(I) - SAVF(I))*FAC
 220    CONTINUE
        Y(J) = YJ
        J1 = J1 + N
 230    CONTINUE
      NFE = NFE + N
C Add identity matrix. -------------------------------------------------
 240  J = 3
      NP1 = N + 1
      DO 250 I = 1,N
        WM(J) = WM(J) + 1.0D0
        J = J + NP1
 250  CONTINUE
C Do LU decomposition on P. --------------------------------------------
      CALL DGEFA (WM(3), N, N, IWM(21), IER)
      IF (IER .NE. 0) IERPJ = 1
      RETURN
C If MITER = 3, construct a diagonal approximation to J and P. ---------
 300  WM(2) = HL0
      R = EL0*0.1D0
      DO 310 I = 1,N
        Y(I) = Y(I) + R*(H*SAVF(I) - YH(I,2))
 310  CONTINUE
CKS
      CALL F (NEQ, TN, Y, WM(3), rpar, ipar)
      NFE = NFE + 1
      DO 320 I = 1,N
        R0 = H*SAVF(I) - YH(I,2)
        DI = 0.1D0*R0 - H*(WM(I+2) - SAVF(I))
        WM(I+2) = 1.0D0
        IF (ABS(R0) .LT. UROUND/EWT(I)) GO TO 320
        IF (ABS(DI) .EQ. 0.0D0) GO TO 330
        WM(I+2) = 0.1D0*R0/DI
 320    CONTINUE
      RETURN
 330  IERPJ = 1
      RETURN
C If MITER = 4, call JAC and multiply by scalar. -----------------------
 400  ML = IWM(1)
      MU = IWM(2)
      ML3 = ML + 3
      MBAND = ML + MU + 1
      MEBAND = MBAND + ML
      LENP = MEBAND*N
      DO 410 I = 1,LENP
        WM(I+2) = 0.0D0
 410  CONTINUE
CKS
      CALL JAC (NEQ, TN, Y, ML, MU, WM(ML3), MEBAND,RPAR,IPAR)
      CON = -HL0
      DO 420 I = 1,LENP
        WM(I+2) = WM(I+2)*CON
 420  CONTINUE
      GO TO 570
C If MITER = 5, make MBAND calls to F to approximate J. ----------------
 500  ML = IWM(1)
      MU = IWM(2)
      MBAND = ML + MU + 1
      MBA = MIN(MBAND,N)
      MEBAND = MBAND + ML
      MEB1 = MEBAND - 1
      SRUR = WM(1)
      FAC = DVNORM (N, SAVF, EWT)
      R0 = 1000.0D0*ABS(H)*UROUND*N*FAC
      IF (R0 .EQ. 0.0D0) R0 = 1.0D0
      DO 560 J = 1,MBA
        DO 530 I = J,N,MBAND
          YI = Y(I)
          R = MAX(SRUR*ABS(YI),R0/EWT(I))
          Y(I) = Y(I) + R
 530    CONTINUE
CKS
        CALL F (NEQ, TN, Y, FTEM, rpar, ipar)
        DO 550 JJ = J,N,MBAND
          Y(JJ) = YH(JJ,1)
          YJJ = Y(JJ)
          R = MAX(SRUR*ABS(YJJ),R0/EWT(JJ))
          FAC = -HL0/R
          I1 = MAX(JJ-MU,1)
          I2 = MIN(JJ+ML,N)
          II = JJ*MEB1 - ML + 2
          DO 540 I = I1,I2
            WM(II+I) = (FTEM(I) - SAVF(I))*FAC
 540      CONTINUE
 550      CONTINUE
 560    CONTINUE
      NFE = NFE + MBA
C Add identity matrix. -------------------------------------------------
 570  II = MBAND + 2
      DO 580 I = 1,N
        WM(II) = WM(II) + 1.0D0
        II = II + MEBAND
 580  CONTINUE
C Do LU decomposition of P. --------------------------------------------
      CALL DGBFA (WM(3), MEBAND, N, ML, MU, IWM(21), IER)
      IF (IER .NE. 0) IERPJ = 1
      RETURN
C----------------------- END OF SUBROUTINE DPREPJ ----------------------
      END
      SUBROUTINE DSTODE (NEQ, Y, YH, NYH, YH1, EWT, SAVF, ACOR,
     1   WM, IWM, F, JAC, PJAC, SLVS,RPAR,IPAR)
C***BEGIN PROLOGUE  DSTODE
C***SUBSIDIARY
C***PURPOSE  Performs one step of an ODEPACK integration.
C***TYPE      DOUBLE PRECISION (SSTODE-S, DSTODE-D)
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***DESCRIPTION
C
C  DSTODE performs one step of the integration of an initial value
C  problem for a system of ordinary differential equations.
C  Note:  DSTODE is independent of the value of the iteration method
C  indicator MITER, when this is .ne. 0, and hence is independent
C  of the type of chord method used, or the Jacobian structure.
C  Communication with DSTODE is done with the following variables:
C
C  NEQ    = integer array containing problem size in NEQ(1), and
C           passed as the NEQ argument in all calls to F and JAC.
C  Y      = an array of length .ge. N used as the Y argument in
C           all calls to F and JAC.
C  YH     = an NYH by LMAX array containing the dependent variables
C           and their approximate scaled derivatives, where
C           LMAX = MAXORD + 1.  YH(i,j+1) contains the approximate
C           j-th derivative of y(i), scaled by h**j/factorial(j)
C           (j = 0,1,...,NQ).  on entry for the first step, the first
C           two columns of YH must be set from the initial values.
C  NYH    = a constant integer .ge. N, the first dimension of YH.
C  YH1    = a one-dimensional array occupying the same space as YH.
C  EWT    = an array of length N containing multiplicative weights
C           for local error measurements.  Local errors in Y(i) are
C           compared to 1.0/EWT(i) in various error tests.
C  SAVF   = an array of working storage, of length N.
C           Also used for input of YH(*,MAXORD+2) when JSTART = -1
C           and MAXORD .lt. the current order NQ.
C  ACOR   = a work array of length N, used for the accumulated
C           corrections.  On a successful return, ACOR(i) contains
C           the estimated one-step local error in Y(i).
C  WM,IWM = real and integer work arrays associated with matrix
C           operations in chord iteration (MITER .ne. 0).
C  PJAC   = name of routine to evaluate and preprocess Jacobian matrix
C           and P = I - h*el0*JAC, if a chord method is being used.
C  SLVS   = name of routine to solve linear system in chord iteration.
C  CCMAX  = maximum relative change in h*el0 before PJAC is called.
C  H      = the step size to be attempted on the next step.
C           H is altered by the error control algorithm during the
C           problem.  H can be either positive or negative, but its
C           sign must remain constant throughout the problem.
C  HMIN   = the minimum absolute value of the step size h to be used.
C  HMXI   = inverse of the maximum absolute value of h to be used.
C           HMXI = 0.0 is allowed and corresponds to an infinite hmax.
C           HMIN and HMXI may be changed at any time, but will not
C           take effect until the next change of h is considered.
C  TN     = the independent variable. TN is updated on each step taken.
C  JSTART = an integer used for input only, with the following
C           values and meanings:
C                0  perform the first step.
C            .gt.0  take a new step continuing from the last.
C               -1  take the next step with a new value of H, MAXORD,
C                     N, METH, MITER, and/or matrix parameters.
C               -2  take the next step with a new value of H,
C                     but with other inputs unchanged.
C           On return, JSTART is set to 1 to facilitate continuation.
C  KFLAG  = a completion code with the following meanings:
C                0  the step was succesful.
C               -1  the requested error could not be achieved.
C               -2  corrector convergence could not be achieved.
C               -3  fatal error in PJAC or SLVS.
C           A return with KFLAG = -1 or -2 means either
C           abs(H) = HMIN or 10 consecutive failures occurred.
C           On a return with KFLAG negative, the values of TN and
C           the YH array are as of the beginning of the last
C           step, and H is the last step size attempted.
C  MAXORD = the maximum order of integration method to be allowed.
C  MAXCOR = the maximum number of corrector iterations allowed.
C  MSBP   = maximum number of steps between PJAC calls (MITER .gt. 0).
C  MXNCF  = maximum number of convergence failures allowed.
C  METH/MITER = the method flags.  See description in driver.
C  N      = the number of first-order differential equations.
C  The values of CCMAX, H, HMIN, HMXI, TN, JSTART, KFLAG, MAXORD,
C  MAXCOR, MSBP, MXNCF, METH, MITER, and N are communicated via COMMON.
C
C***SEE ALSO  DLSODE
C***ROUTINES CALLED  DCFODE, DVNORM
C***COMMON BLOCKS    DLS001
C***REVISION HISTORY  (YYMMDD)
C   791129  DATE WRITTEN
C   890501  Modified prologue to SLATEC/LDOC format.  (FNF)
C   890503  Minor cosmetic changes.  (FNF)
C   930809  Renamed to allow single/double precision versions. (ACH)
C   010418  Reduced size of Common block /DLS001/. (ACH)
C   031105  Restored 'own' variables to Common block /DLS001/, to
C           enable interrupt/restart feature. (ACH)
C***END PROLOGUE  DSTODE
C**End
      EXTERNAL F, JAC, PJAC, SLVS
      INTEGER NEQ, NYH, IWM
CKS: added rpar,ipar
      integer ipar(*)
      double precision rpar(*)
      DOUBLE PRECISION Y, YH, YH1, EWT, SAVF, ACOR, WM
      DIMENSION NEQ(*), Y(*), YH(NYH,*), YH1(*), EWT(*), SAVF(*),
     1   ACOR(*), WM(*), IWM(*)
      INTEGER INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH01,
     1   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     2   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     3   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     4   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
      INTEGER I, I1, IREDO, IRET, J, JB, M, NCF, NEWQ
      DOUBLE PRECISION CONIT, CRATE, EL, ELCO, HOLD, RMAX, TESCO,
     2   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND
      DOUBLE PRECISION DCON, DDN, DEL, DELP, DSM, DUP, EXDN, EXSM, EXUP,
     1   R, RH, RHDN, RHSM, RHUP, TOLD, DVNORM
      COMMON /DLS001/ CONIT, CRATE, EL(13), ELCO(13,12),
     1   HOLD, RMAX, TESCO(3,12),
     2   CCMAX, EL0, H, HMIN, HMXI, HU, RC, TN, UROUND,
     3   INIT, MXSTEP, MXHNIL, NHNIL, NSLAST, NYH01,
     4   IALTH, IPUP, LMAX, MEO, NQNYH, NSLP,
     5   ICF, IERPJ, IERSL, JCUR, JSTART, KFLAG, L,
     6   LYH, LEWT, LACOR, LSAVF, LWM, LIWM, METH, MITER,
     7   MAXORD, MAXCOR, MSBP, MXNCF, N, NQ, NST, NFE, NJE, NQU
C
C***FIRST EXECUTABLE STATEMENT  DSTODE
      KFLAG = 0
      TOLD = TN
      NCF = 0
      IERPJ = 0
      IERSL = 0
      JCUR = 0
      ICF = 0
      IREDO = 0
      DELP = 0.0D0
      RH = 0.0D0
      IF (JSTART .GT. 0) GO TO 200
      IF (JSTART .EQ. -1) GO TO 100
      IF (JSTART .EQ. -2) GO TO 160
C-----------------------------------------------------------------------
C On the first call, the order is set to 1, and other variables are
C initialized.  RMAX is the maximum ratio by which H can be increased
C in a single step.  It is initially 1.E4 to compensate for the small
C initial H, but then is normally equal to 10.  If a failure
C occurs (in corrector convergence or error test), RMAX is set to 2
C for the next increase.
C-----------------------------------------------------------------------
      LMAX = MAXORD + 1
      NQ = 1
      L = 2
      IALTH = 2
      RMAX = 10000.0D0
      RC = 0.0D0
      EL0 = 1.0D0
      CRATE = 0.7D0
      HOLD = H
      MEO = METH
      NSLP = 0
      IPUP = MITER
      IRET = 3
      GO TO 140
C-----------------------------------------------------------------------
C The following block handles preliminaries needed when JSTART = -1.
C IPUP is set to MITER to force a matrix update.
C If an order increase is about to be considered (IALTH = 1),
C IALTH is reset to 2 to postpone consideration one more step.
C If the caller has changed METH, DCFODE is called to reset
C the coefficients of the method.
C If the caller has changed MAXORD to a value less than the current
C order NQ, NQ is reduced to MAXORD, and a new H chosen accordingly.
C If H is to be changed, YH must be rescaled.
C If H or METH is being changed, IALTH is reset to L = NQ + 1
C to prevent further changes in H for that many steps.
C-----------------------------------------------------------------------
 100  IPUP = MITER
      LMAX = MAXORD + 1
      IF (IALTH .EQ. 1) IALTH = 2
      IF (METH .EQ. MEO) GO TO 110
      CALL DCFODE (METH, ELCO, TESCO)
      MEO = METH
      IF (NQ .GT. MAXORD) GO TO 120
      IALTH = L
      IRET = 1
      GO TO 150
 110  IF (NQ .LE. MAXORD) GO TO 160
 120  NQ = MAXORD
      L = LMAX
      DO 125 I = 1,L
        EL(I) = ELCO(I,NQ)
 125  CONTINUE
      NQNYH = NQ*NYH
      RC = RC*EL(1)/EL0
      EL0 = EL(1)
      CONIT = 0.5D0/(NQ+2)
      DDN = DVNORM (N, SAVF, EWT)/TESCO(1,L)
      EXDN = 1.0D0/L
      RHDN = 1.0D0/(1.3D0*DDN**EXDN + 0.0000013D0)
      RH = MIN(RHDN,1.0D0)
      IREDO = 3
      IF (H .EQ. HOLD) GO TO 170
      RH = MIN(RH,ABS(H/HOLD))
      H = HOLD
      GO TO 175
C-----------------------------------------------------------------------
C DCFODE is called to get all the integration coefficients for the
C current METH.  Then the EL vector and related constants are reset
C whenever the order NQ is changed, or at the start of the problem.
C-----------------------------------------------------------------------
 140  CALL DCFODE (METH, ELCO, TESCO)
 150  DO 155 I = 1,L
        EL(I) = ELCO(I,NQ)
 155  CONTINUE
      NQNYH = NQ*NYH
      RC = RC*EL(1)/EL0
      EL0 = EL(1)
      CONIT = 0.5D0/(NQ+2)

      IF (IRET .EQ. 1) THEN
        GOTO 160
      ELSE IF (IRET .EQ. 2) THEN
        GOTO 170
      ELSE IF (IRET .EQ. 3) THEN
        GOTO 200
      ENDIF
C      GO TO (160, 170, 200), IRET
C-----------------------------------------------------------------------
C If H is being changed, the H ratio RH is checked against
C RMAX, HMIN, and HMXI, and the YH array rescaled.  IALTH is set to
C L = NQ + 1 to prevent a change of H for that many steps, unless
C forced by a convergence or error test failure.
C-----------------------------------------------------------------------
 160  IF (H .EQ. HOLD) GO TO 200
      RH = H/HOLD
      H = HOLD
      IREDO = 3
      GO TO 175
 170  RH = MAX(RH,HMIN/ABS(H))
 175  RH = MIN(RH,RMAX)
      RH = RH/MAX(1.0D0,ABS(H)*HMXI*RH)
      R = 1.0D0
      DO 190 J = 2,L
        R = R*RH
        DO 180 I = 1,N
          YH(I,J) = YH(I,J)*R
 180    CONTINUE
 190  CONTINUE
      H = H*RH
      RC = RC*RH
      IALTH = L
      IF (IREDO .EQ. 0) GO TO 690
C-----------------------------------------------------------------------
C This section computes the predicted values by effectively
C multiplying the YH array by the Pascal Triangle matrix.
C RC is the ratio of new to old values of the coefficient  H*EL(1).
C When RC differs from 1 by more than CCMAX, IPUP is set to MITER
C to force PJAC to be called, if a Jacobian is involved.
C In any case, PJAC is called at least every MSBP steps.
C-----------------------------------------------------------------------
 200  IF (ABS(RC-1.0D0) .GT. CCMAX) IPUP = MITER
      IF (NST .GE. NSLP+MSBP) IPUP = MITER
      TN = TN + H
      I1 = NQNYH + 1
      DO 215 JB = 1,NQ
        I1 = I1 - NYH
Cdir$ ivdep
        DO 210 I = I1,NQNYH
          YH1(I) = YH1(I) + YH1(I+NYH)
 210    CONTINUE
 215    CONTINUE
C-----------------------------------------------------------------------
C Up to MAXCOR corrector iterations are taken.  A convergence test is
C made on the R.M.S. norm of each correction, weighted by the error
C weight vector EWT.  The sum of the corrections is accumulated in the
C vector ACOR(i).  The YH array is not altered in the corrector loop.
C-----------------------------------------------------------------------
 220  M = 0
      DO 230 I = 1,N
        Y(I) = YH(I,1)
 230  CONTINUE
CKS
      CALL F (NEQ, TN, Y, SAVF, rpar, ipar)
      NFE = NFE + 1
      IF (IPUP .LE. 0) GO TO 250
C-----------------------------------------------------------------------
C If indicated, the matrix P = I - h*el(1)*J is reevaluated and
C preprocessed before starting the corrector iteration.  IPUP is set
C to 0 as an indicator that this has been done.
C-----------------------------------------------------------------------
CKS
      CALL PJAC (NEQ, Y, YH, NYH, EWT, ACOR, SAVF, WM, IWM, F, JAC,
     &rpar,ipar)
      IPUP = 0
      RC = 1.0D0
      NSLP = NST
      CRATE = 0.7D0
      IF (IERPJ .NE. 0) GO TO 430
 250  DO 260 I = 1,N
        ACOR(I) = 0.0D0
 260  CONTINUE
 270  IF (MITER .NE. 0) GO TO 350
C-----------------------------------------------------------------------
C In the case of functional iteration, update Y directly from
C the result of the last function evaluation.
C-----------------------------------------------------------------------
      DO 290 I = 1,N
        SAVF(I) = H*SAVF(I) - YH(I,2)
        Y(I) = SAVF(I) - ACOR(I)
 290  CONTINUE
      DEL = DVNORM (N, Y, EWT)
      DO 300 I = 1,N
        Y(I) = YH(I,1) + EL(1)*SAVF(I)
        ACOR(I) = SAVF(I)
 300  CONTINUE
      GO TO 400
C-----------------------------------------------------------------------
C In the case of the chord method, compute the corrector error,
C and solve the linear system with that as right-hand side and
C P as coefficient matrix.
C-----------------------------------------------------------------------
 350  DO 360 I = 1,N
        Y(I) = H*SAVF(I) - (YH(I,2) + ACOR(I))
 360  CONTINUE
      CALL SLVS (WM, IWM, Y, SAVF)
      IF (IERSL .LT. 0) GO TO 430
      IF (IERSL .GT. 0) GO TO 410
      DEL = DVNORM (N, Y, EWT)
      DO 380 I = 1,N
        ACOR(I) = ACOR(I) + Y(I)
        Y(I) = YH(I,1) + EL(1)*ACOR(I)
 380  CONTINUE
C-----------------------------------------------------------------------
C Test for convergence.  If M.gt.0, an estimate of the convergence
C rate constant is stored in CRATE, and this is used in the test.
C-----------------------------------------------------------------------
 400  IF (M .NE. 0) CRATE = MAX(0.2D0*CRATE,DEL/DELP)
      DCON = DEL*MIN(1.0D0,1.5D0*CRATE)/(TESCO(2,NQ)*CONIT)
      IF (DCON .LE. 1.0D0) GO TO 450
      M = M + 1
      IF (M .EQ. MAXCOR) GO TO 410
      IF (M .GE. 2 .AND. DEL .GT. 2.0D0*DELP) GO TO 410
      DELP = DEL
CKS
      CALL F (NEQ, TN, Y, SAVF, rpar, ipar)
      NFE = NFE + 1
      GO TO 270
C-----------------------------------------------------------------------
C The corrector iteration failed to converge.
C If MITER .ne. 0 and the Jacobian is out of date, PJAC is called for
C the next try.  Otherwise the YH array is retracted to its values
C before prediction, and H is reduced, if possible.  If H cannot be
C reduced or MXNCF failures have occurred, exit with KFLAG = -2.
C-----------------------------------------------------------------------
 410  IF (MITER .EQ. 0 .OR. JCUR .EQ. 1) GO TO 430
      ICF = 1
      IPUP = MITER
      GO TO 220
 430  ICF = 2
      NCF = NCF + 1
      RMAX = 2.0D0
      TN = TOLD
      I1 = NQNYH + 1
      DO 445 JB = 1,NQ
        I1 = I1 - NYH
Cdir$ ivdep
        DO 440 I = I1,NQNYH
          YH1(I) = YH1(I) - YH1(I+NYH)
 440    CONTINUE
 445    CONTINUE
      IF (IERPJ .LT. 0 .OR. IERSL .LT. 0) GO TO 680
      IF (ABS(H) .LE. HMIN*1.00001D0) GO TO 670
      IF (NCF .EQ. MXNCF) GO TO 670
      RH = 0.25D0
      IPUP = MITER
      IREDO = 1
      GO TO 170
C-----------------------------------------------------------------------
C The corrector has converged.  JCUR is set to 0
C to signal that the Jacobian involved may need updating later.
C The local error test is made and control passes to statement 500
C if it fails.
C-----------------------------------------------------------------------
 450  JCUR = 0
      IF (M .EQ. 0) DSM = DEL/TESCO(2,NQ)
      IF (M .GT. 0) DSM = DVNORM (N, ACOR, EWT)/TESCO(2,NQ)
      IF (DSM .GT. 1.0D0) GO TO 500
C-----------------------------------------------------------------------
C After a successful step, update the YH array.
C Consider changing H if IALTH = 1.  Otherwise decrease IALTH by 1.
C If IALTH is then 1 and NQ .lt. MAXORD, then ACOR is saved for
C use in a possible order increase on the next step.
C If a change in H is considered, an increase or decrease in order
C by one is considered also.  A change in H is made only if it is by a
C factor of at least 1.1.  If not, IALTH is set to 3 to prevent
C testing for that many steps.
C-----------------------------------------------------------------------
      KFLAG = 0
      IREDO = 0
      NST = NST + 1
      HU = H
      NQU = NQ
      DO 480 J = 1,L
        DO 470 I = 1,N
          YH(I,J) = YH(I,J) + EL(J)*ACOR(I)
 470    CONTINUE
 480  CONTINUE
      IALTH = IALTH - 1
      IF (IALTH .EQ. 0) GO TO 520
      IF (IALTH .GT. 1) GO TO 700
      IF (L .EQ. LMAX) GO TO 700
      DO 490 I = 1,N
        YH(I,LMAX) = ACOR(I)
 490  CONTINUE
      GO TO 700
C-----------------------------------------------------------------------
C The error test failed.  KFLAG keeps track of multiple failures.
C Restore TN and the YH array to their previous values, and prepare
C to try the step again.  Compute the optimum step size for this or
C one lower order.  After 2 or more failures, H is forced to decrease
C by a factor of 0.2 or less.
C-----------------------------------------------------------------------
 500  KFLAG = KFLAG - 1
      TN = TOLD
      I1 = NQNYH + 1
      DO 515 JB = 1,NQ
        I1 = I1 - NYH
Cdir$ ivdep
        DO 510 I = I1,NQNYH
          YH1(I) = YH1(I) - YH1(I+NYH)
 510    CONTINUE
 515    CONTINUE
      RMAX = 2.0D0
      IF (ABS(H) .LE. HMIN*1.00001D0) GO TO 660
      IF (KFLAG .LE. -3) GO TO 640
      IREDO = 2
      RHUP = 0.0D0
      GO TO 540
C-----------------------------------------------------------------------
C Regardless of the success or failure of the step, factors
C RHDN, RHSM, and RHUP are computed, by which H could be multiplied
C at order NQ - 1, order NQ, or order NQ + 1, respectively.
C In the case of failure, RHUP = 0.0 to avoid an order increase.
C The largest of these is determined and the new order chosen
C accordingly.  If the order is to be increased, we compute one
C additional scaled derivative.
C-----------------------------------------------------------------------
 520  RHUP = 0.0D0
      IF (L .EQ. LMAX) GO TO 540
      DO 530 I = 1,N
        SAVF(I) = ACOR(I) - YH(I,LMAX)
 530  CONTINUE
      DUP = DVNORM (N, SAVF, EWT)/TESCO(3,NQ)
      EXUP = 1.0D0/(L+1)
      RHUP = 1.0D0/(1.4D0*DUP**EXUP + 0.0000014D0)
 540  EXSM = 1.0D0/L
      RHSM = 1.0D0/(1.2D0*DSM**EXSM + 0.0000012D0)
      RHDN = 0.0D0
      IF (NQ .EQ. 1) GO TO 560
      DDN = DVNORM (N, YH(1,L), EWT)/TESCO(1,NQ)
      EXDN = 1.0D0/NQ
      RHDN = 1.0D0/(1.3D0*DDN**EXDN + 0.0000013D0)
 560  IF (RHSM .GE. RHUP) GO TO 570
      IF (RHUP .GT. RHDN) GO TO 590
      GO TO 580
 570  IF (RHSM .LT. RHDN) GO TO 580
      NEWQ = NQ
      RH = RHSM
      GO TO 620
 580  NEWQ = NQ - 1
      RH = RHDN
      IF (KFLAG .LT. 0 .AND. RH .GT. 1.0D0) RH = 1.0D0
      GO TO 620
 590  NEWQ = L
      RH = RHUP
      IF (RH .LT. 1.1D0) GO TO 610
      R = EL(L)/L
      DO 600 I = 1,N
        YH(I,NEWQ+1) = ACOR(I)*R
 600  CONTINUE
      GO TO 630
 610  IALTH = 3
      GO TO 700
 620  IF ((KFLAG .EQ. 0) .AND. (RH .LT. 1.1D0)) GO TO 610
      IF (KFLAG .LE. -2) RH = MIN(RH,0.2D0)
C-----------------------------------------------------------------------
C If there is a change of order, reset NQ, l, and the coefficients.
C In any case H is reset according to RH and the YH array is rescaled.
C Then exit from 690 if the step was OK, or redo the step otherwise.
C-----------------------------------------------------------------------
      IF (NEWQ .EQ. NQ) GO TO 170
 630  NQ = NEWQ
      L = NQ + 1
      IRET = 2
      GO TO 150
C-----------------------------------------------------------------------
C Control reaches this section if 3 or more failures have occured.
C If 10 failures have occurred, exit with KFLAG = -1.
C It is assumed that the derivatives that have accumulated in the
C YH array have errors of the wrong order.  Hence the first
C derivative is recomputed, and the order is set to 1.  Then
C H is reduced by a factor of 10, and the step is retried,
C until it succeeds or H reaches HMIN.
C-----------------------------------------------------------------------
 640  IF (KFLAG .EQ. -10) GO TO 660
      RH = 0.1D0
      RH = MAX(HMIN/ABS(H),RH)
      H = H*RH
      DO 645 I = 1,N
        Y(I) = YH(I,1)
 645  CONTINUE
CKS
      CALL F (NEQ, TN, Y, SAVF, rpar, ipar)
      NFE = NFE + 1
      DO 650 I = 1,N
        YH(I,2) = H*SAVF(I)
 650  CONTINUE
      IPUP = MITER
      IALTH = 5
      IF (NQ .EQ. 1) GO TO 200
      NQ = 1
      L = 2
      IRET = 3
      GO TO 150
C-----------------------------------------------------------------------
C All returns are made through this section.  H is saved in HOLD
C to allow the caller to change H on the next step.
C-----------------------------------------------------------------------
 660  KFLAG = -1
      GO TO 720
 670  KFLAG = -2
      GO TO 720
 680  KFLAG = -3
      GO TO 720
 690  RMAX = 10.0D0
 700  R = 1.0D0/TESCO(2,NQU)
      DO 710 I = 1,N
        ACOR(I) = ACOR(I)*R
 710  CONTINUE
 720  HOLD = H
      JSTART = 1
      RETURN
C----------------------- END OF SUBROUTINE DSTODE ----------------------
      END
      DOUBLE PRECISION FUNCTION DVNORM (N, V, W)
C***BEGIN PROLOGUE  DVNORM
C***SUBSIDIARY
C***PURPOSE  Weighted root-mean-square vector norm.
C***TYPE      DOUBLE PRECISION (SVNORM-S, DVNORM-D)
C***AUTHOR  Hindmarsh, Alan C., (LLNL)
C***DESCRIPTION
C
C  This function routine computes the weighted root-mean-square norm
C  of the vector of length N contained in the array V, with weights
C  contained in the array W of length N:
C    DVNORM = SQRT( (1/N) * SUM( V(i)*W(i) )**2 )
C
C***SEE ALSO  DLSODE
C***ROUTINES CALLED  (NONE)
C***REVISION HISTORY  (YYMMDD)
C   791129  DATE WRITTEN
C   890501  Modified prologue to SLATEC/LDOC format.  (FNF)
C   890503  Minor cosmetic changes.  (FNF)
C   930809  Renamed to allow single/double precision versions. (ACH)
C***END PROLOGUE  DVNORM
C**End
      INTEGER N,   I
      DOUBLE PRECISION V, W,   SUM
      DIMENSION V(N), W(N)
C
C***FIRST EXECUTABLE STATEMENT  DVNORM
      SUM = 0.0D0
      DO 10 I = 1,N
        SUM = SUM + (V(I)*W(I))**2
 10   CONTINUE
      DVNORM = SQRT(SUM/N)
      RETURN
C----------------------- END OF FUNCTION DVNORM ------------------------
      END

.parseEnv$.rxode2parseDf <-  structure(list(rxFun = c("linCmtA", "linCmtB", "rxnorm", "rxbinom", 
"rxnbinom", "rxnbinomMu", "rxcauchy", "rxchisq", "rxexp", "rxf", 
"rxgeom", "rxgamma", "rxbeta", "rxpois", "rxt_", "rxunif", "rxweibull", 
"rinorm", "ribinom", "rinbinom", "rinbinomMu", "ricauchy", "richisq", 
"riexp", "rif", "rigeom", "rigamma", "ribeta", "ripois", "rit_", 
"riunif", "riweibull", "ReLU", "dReLU", "GELU", "dGELU", "d2GELU", 
"d3GELU", "d4GELU", "softplus", "dsoftplus", "d2softplus", "d3softplus", 
"d4softplus", "SELU", "dSELU", "lReLU", "dlReLU", "Swish", "dSwish", 
"PReLU", "dPReLU", "dPReLUa", "dPReLUa1", "ELU", "dELU", "d2ELU", 
"d2aELU", "dELUa", "d2ELUa", "phi", "gammap", "gammaq", "gammapInv", 
"gammapInva", "gammaqInv", "gammaqInva", "uppergamma", "lowergamma", 
"gammapDer", "logit", "probit", "expit", "probitInv", "simeta", 
"simeps", "llikNorm", "llikNormDmean", "llikNormDsd", "llikPois", 
"llikPoisDlambda", "llikBinom", "llikBinomDprob", "llikNbinom", 
"llikNbinomDprob", "llikNbinomMu", "llikNbinomMuDmu", "llikBeta", 
"llikBetaDshape1", "llikBetaDshape2", "llikT", "llikTDdf", "llikTDmean", 
"llikTDsd", "llikChisq", "llikChisqDdf", "llikExp", "llikExpDrate", 
"llikF", "llikFDdf1", "llikFDdf2", "llikGeom", "llikGeomDp", 
"llikUnif", "llikUnifDalpha", "llikUnifDbeta", "llikWeibull", 
"llikWeibullDshape", "llikWeibullDscale", "llikGamma", "llikGammaDshape", 
"llikGammaDrate", "llikCauchy", "llikCauchyDlocation", "llikCauchyDscale"
), fun = c("linCmtA", "linCmtB", "rxnorm", "rxbinom", "rxnbinom", 
"rxnbinomMu", "rxcauchy", "rxchisq", "rxexp", "rxf", "rxgeom", 
"rxgamma", "rxbeta", "rxpois", "rxt_", "rxunif", "rxweibull", 
"rinorm", "ribinom", "rinbinom", "rinbinomMu", "ricauchy", "richisq", 
"riexp", "rif", "rigeom", "rigamma", "ribeta", "ripois", "rit_", 
"riunif", "riweibull", "ReLU", "dReLU", "GELU", "dGELU", "d2GELU", 
"d3GELU", "d4GELU", "softplus", "dsoftplus", "d2softplus", "d3softplus", 
"d4softplus", "SELU", "dSELU", "lReLU", "dlReLU", "Swish", "dSwish", 
"PReLU", "dPReLU", "dPReLUa", "dPReLUa1", "ELU", "dELU", "d2ELU", 
"d2aELU", "dELUa", "d2ELUa", "phi", "gammap", "gammaq", "gammapInv", 
"gammapInva", "gammaqInv", "gammaqInva", "uppergamma", "lowergamma", 
"gammapDer", "logit", "probit", "expit", "probitInv", "simeta", 
"simeps", "_llikNorm", "_llikNormDmean", "_llikNormDsd", "_llikPois", 
"_llikPoisDlambda", "_llikBinom", "_llikBinomDprob", "_llikNbinom", 
"_llikNbinomDprob", "_llikNbinomMu", "_llikNbinomMuDmu", "_llikBeta", 
"_llikBetaDshape1", "_llikBetaDshape2", "_llikT", "_llikTDdf", 
"_llikTDmean", "_llikTDsd", "_llikChisq", "_llikChisqDdf", "_llikExp", 
"_llikExpDrate", "_llikF", "_llikFDdf1", "_llikFDdf2", "_llikGeom", 
"_llikGeomDp", "_llikUnif", "_llikUnifDalpha", "_llikUnifDbeta", 
"_llikWeibull", "_llikWeibullDshape", "_llikWeibullDscale", "_llikGamma", 
"_llikGammaDshape", "_llikGammaDrate", "_llikCauchy", "_llikCauchyDlocation", 
"_llikCauchyDscale"), type = c("linCmtA_p", "linCmtB_p", "rxode2i_fn2", 
"rxode2i_rxbinom", "rxode2i_rxbinom", "rxode2i_rxbinom", "rxode2_fn2", 
"rxode2_fn", "rxode2_fn", "rxode2_fn2", "rxode2_ifn", "rxode2_fn2", 
"rxode2_fn2", "rxode2_ifn", "rxode2_fn", "rxode2_fn2", "rxode2_fn2", 
"rxode2i2_fn2", "rxode2i2_ribinom", "rxode2i2_ribinom", "rxode2i2_ribinom", 
"rxode2i2_fn2", "rxode2i2_fn", "rxode2i2_fn", "rxode2i2_fn2", 
"rxode2i2_ifn", "rxode2i2_fn2", "rxode2i2_fn2", "rxode2i2_ifn", 
"rxode2i2_fn", "rxode2i2_fn2", "rxode2i2_fn2", "rxode2_fn", "rxode2_fn", 
"rxode2_fn", "rxode2_fn", "rxode2_fn", "rxode2_fn", "rxode2_fn", 
"rxode2_fn", "rxode2_fn", "rxode2_fn", "rxode2_fn", "rxode2_fn", 
"rxode2_fn", "rxode2_fn", "rxode2_fn", "rxode2_fn", "rxode2_fn", 
"rxode2_fn", "rxode2_fn2", "rxode2_fn2", "rxode2_fn2", "rxode2_fn2", 
"rxode2_fn2", "rxode2_fn2", "rxode2_fn2", "rxode2_fn2", "rxode2_fn2", 
"rxode2_fn2", "rxode2_fn", "rxode2_fn2", "rxode2_fn2", "rxode2_fn2", 
"rxode2_fn2", "rxode2_fn2", "rxode2_fn2", "rxode2_fn2", "rxode2_fn2", 
"rxode2_fn2", "rxode2_fn3", "rxode2_fn3", "rxode2_fn3", "rxode2_fn3", 
"_simfun", "_simfun", "rxode2_llikNormFun", "rxode2_llikNormFun", 
"rxode2_llikNormFun", "rxode2_llikPoisFun", "rxode2_llikPoisFun", 
"rxode2_llikBinomFun", "rxode2_llikBinomFun", "rxode2_llikBinomFun", 
"rxode2_llikBinomFun", "rxode2_llikBinomFun", "rxode2_llikBinomFun", 
"rxode2_llikBetaFun", "rxode2_llikBetaFun", "rxode2_llikBetaFun", 
"rxode2_llikTFun", "rxode2_llikTFun", "rxode2_llikTFun", "rxode2_llikTFun", 
"rxode2_llikChisqFun", "rxode2_llikChisqFun", "rxode2_llikExpFun", 
"rxode2_llikExpFun", "rxode2_llikFFun", "rxode2_llikFFun", "rxode2_llikFFun", 
"rxode2_llikGeomFun", "rxode2_llikGeomFun", "rxode2_llikUnifFun", 
"rxode2_llikUnifFun", "rxode2_llikUnifFun", "rxode2_llikWeibullFun", 
"rxode2_llikWeibullFun", "rxode2_llikWeibullFun", "rxode2_llikGammaFun", 
"rxode2_llikGammaFun", "rxode2_llikGammaFun", "rxode2_llikCauchyFun", 
"rxode2_llikCauchyFun", "rxode2_llikCauchyFun"), package = c("rxode2", 
"rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", 
"rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", 
"rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", 
"rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", 
"rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", 
"rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", 
"rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", 
"rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", 
"rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", 
"rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2", 
"rxode2", "rxode2", "rxode2", "rxode2", "rxode2", "rxode2ll", 
"rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", 
"rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", 
"rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", 
"rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", 
"rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", 
"rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", "rxode2ll", 
"rxode2ll", "rxode2ll"), packageFun = c("linCmtA", "linCmtB", 
"rxnorm", "rxbinom", "rxnbinom", "rxnbinomMu", "rxcauchy", "rxchisq", 
"rxexp", "rxf", "rxgeom", "rxgamma", "rxbeta", "rxpois", "rxt_", 
"rxunif", "rxweibull", "rinorm", "ribinom", "rinbinom", "rinbinomMu", 
"ricauchy", "richisq", "riexp", "rif", "rigeom", "rigamma", "ribeta", 
"ripois", "rit_", "riunif", "riweibull", "ReLU", "dReLU", "GELU", 
"dGELU", "d2GELU", "d3GELU", "d4GELU", "softplus", "dsoftplus", 
"d2softplus", "d3softplus", "d4softplus", "SELU", "dSELU", "lReLU", 
"dlReLU", "Swish", "dSwish", "PReLU", "dPReLU", "dPReLUa", "dPReLUa1", 
"ELU", "dELU", "d2ELU", "d2aELU", "dELUa", "d2ELUa", "phi", "gammap", 
"gammaq", "gammapInv", "gammapInva", "gammaqInv", "gammaqInva", 
"uppergamma", "lowergamma", "gammapDer", "logit", "probit", "expit", 
"probitInv", "simeta", "simeps", "rxLlikNorm", "rxLlikNormDmean", 
"rxLlikNormDsd", "rxLlikPois", "rxLlikPoisDlambda", "rxLlikBinom", 
"rxLlikBinomDprob", "rxLlikNbinom", "rxLlikNbinomDprob", "rxLlikNbinomMu", 
"rxLlikNbinomMuDmu", "rxLlikBeta", "rxLlikBetaDshape1", "rxLlikBetaDshape2", 
"rxLlikT", "rxLlikTDdf", "rxLlikTDmean", "rxLlikTDsd", "rxLlikChisq", 
"rxLlikChisqDdf", "rxLlikExp", "rxLlikExpDrate", "rxLlikF", "rxLlikFDdf1", 
"rxLlikFDdf2", "rxLlikGeom", "rxLlikGeomDp", "rxLlikUnif", "rxLlikUnifDalpha", 
"rxLlikUnifDbeta", "rxLlikWeibull", "rxLlikWeibullDshape", "rxLlikWeibullDscale", 
"rxLlikGamma", "rxLlikGammaDshape", "rxLlikGammaDrate", "rxLlikCauchy", 
"rxLlikCauchyDlocation", "rxLlikCauchyDscale"), argMin = c(14L, 
15L, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 
1L, 1L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, NA, 2L, 2L, 2L, 
2L, 2L, 2L, 2L, 2L, 2L, NA, NA, NA, NA, NA, NA, 3L, 3L, 3L, 2L, 
2L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 4L, 4L, 4L, 4L, 2L, 2L, 
2L, 2L, 3L, 3L, 3L, 2L, NA, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 
3L, 3L, 3L), argMax = c(14L, 15L, NA, NA, NA, NA, NA, NA, NA, 
NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
NA, NA, NA, NA, NA, NA, NA, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 
1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 
2L, 2L, 2L, NA, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, NA, NA, NA, 
NA, NA, NA, 3L, 3L, 3L, 2L, 2L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 
3L, 4L, 4L, 4L, 4L, 2L, 2L, 2L, 2L, 3L, 3L, 3L, 2L, NA, 3L, 3L, 
3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L), threadSafe = c(1L, 1L, 
1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 
1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 
1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 
1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 
1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 
1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 
1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 
1L)), row.names = c(NA, -115L), class = "data.frame")

#' This function matches the call to 'plogis'
#'
#' @param fun the R language expression for plogis that is parsed
#' @return list of the formals for plogis with the arguments replaced
#'   by the values in the call
#' @noRd
#' @author Matthew L. Fidler
.plogisMatchCall <- function(fun) {
  .args <- as.list(formals(stats::plogis))
  .call <- as.list(match.call(stats::plogis, fun, expand.dots = FALSE))[-1L]
  .args[names(.call)] <- .call
  .args
}
#' This is the expression
#'
#' @param q the plogis `q` expression
#' @param location the `plogis` location expression
#' @param scale the `scale` expression
#' @param env the environment that evaluations are performed in.
#' @return the plogis()->expit() translation when lower.tail=TRUE
#' @noRd
#' @author Matthew L. Fidler
.plogisZLang <- function(q, location, scale, env = baseenv()) {
  .ret <- q
  if (!rxUdfUiIsValue(location, 0, env = env)) {
    .ret <- call("-", .ret, location) # q-location
  }
  if (!rxUdfUiIsValue(scale, 1, env = env)) {
    if (is.call(.ret) && identical(.ret[[1]], quote(`-`)) && length(.ret) == 3L) {
      .ret <- str2lang(paste0("(", deparse1(.ret), ")")) # (q-location)
    }
    .ret <- call("/", .ret, scale) # (q-location)/scale or q/scale
  }
  .ret
}
#' Plogis for lower.tail=FALSE
#'
#'
#' @param q the plogis `q` expression
#' @param location the `plogis` location expression
#' @param scale the `scale` expression
#' @param env the environment that evaluations are performed in.
#' @return the plogis()->expit() translation when lower.tail=FALSE
#' @noRd
#' @author Matthew L. Fidler
.plogisNegZLang <- function(q, location, scale, env = baseenv()) {
  if (rxUdfUiIsValue(location, 0, env = env)) {
    .ret <- call("-", q) # -q
  } else {
    .ret <- call("-", location, q) # location-q
  }
  if (!rxUdfUiIsValue(scale, 1, env = env)) {
    if (is.call(.ret) &&
          identical(.ret[[1]], quote(`-`)) &&
          length(.ret) == 3L) {
      .ret <- str2lang(paste0("(", deparse1(.ret), ")")) # -(location-q) or -(-q)
    }
    .ret <- call("/", .ret, scale) # -(location-q)/scale or -(-q)/scale
  }
  .ret
}
#' Change plogis to expit for RxUdfUi
#'
#' @param q the plogis `q` expression
#' @param location the `plogis` location expression
#' @param scale the `scale` expression
#' @param lower.tail the `lower.tail` expression
#' @param log.p the `log.p` expression
#' @param env the environment that evaluations are performed in.
#' @return the plogis()->expit() translation
#' @noRd
#' @author Matthew L. Fidler
.plogisRxLang <- function(q, location, scale, lower.tail, log.p, env = baseenv()) {
  if (isTRUE(lower.tail)) {
    .ret <- call("expit", .plogisZLang(q, location, scale, env = env))
  } else {
    .ret <- call("expit", .plogisNegZLang(q, location, scale, env = env))
  }
  if (isTRUE(log.p)) {
    .ret <- call("log", .ret)
  }
  .ret
}

# The plogis() ui translation to expit()
#' @export
rxUdfUi.plogis <- function(fun) {
  .args <- .plogisMatchCall(fun)
  .env <- parent.frame()
  .q <- rxUdfUiExpr(.args$q, env = .env)
  .location <- rxUdfUiExpr(.args$location, env = .env)
  .scale <- rxUdfUiExpr(.args$scale, env = .env)
  .lowerTail <- rxUdfUiFlag(.args$lower.tail, arg="lower.tail", funName="plogis", env = .env)
  .logP <- rxUdfUiFlag(.args$log.p, arg="log.p", funName="plogis", env = .env)
  list(replace = .plogisRxLang(.q, .location, .scale, .lowerTail, .logP, env = .env))
}

#' Update for rxUi
#'
#' @param object rxode2 UI object
#' @param ... Lines to update
#' @param envir Environment for evaluating ini({}) style calls
#' @return a new rxode2 updated UI object
#' @export
update.rxUi <- function(object, ..., envir=parent.frame()) {
  .modelLines <- .quoteCallInfoLines(match.call(expand.dots = TRUE)[-(1:2)], envir=envir)
  # x, ..., envir=parent.frame()
  x <- object
  .ret <- .copyUi(x) # copy so (as expected) old UI isn't affected by the call
  .modelHandleModelLines(.modelLines, .ret, modifyIni=TRUE, envir=envir)
}

#'@export
update.function <- function(object, ..., envir=parent.frame()) {
  .modelLines <- .quoteCallInfoLines(match.call(expand.dots = TRUE)[-(1:2)], envir=envir)
  # x, ..., envir=parent.frame()
  x <- object
  .ret <- rxode2(x) # copy so (as expected) old UI isn't affected by the call
  .modelHandleModelLines(.modelLines, .ret, modifyIni=TRUE, envir=envir)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Scaled Inverse Chi Squared distribution
#'
#' @param n Number of random samples
#'
#' @param nu degrees of freedom of inverse chi square
#'
#' @param scale  Scale of inverse chi squared distribution
#'         (default is 1).
#'
#' @return a vector of inverse chi squared deviates.
#'
#' @examples
#' rinvchisq(3, 4, 1) ## Scale = 1, degrees of freedom = 4
#' rinvchisq(2, 4, 2) ## Scale = 2, degrees of freedom = 4
#' @export
rinvchisq <- function(n = 1L, nu = 1.0, scale = 1) {
    .Call(`_rxode2_rinvchisq`, n, nu, scale)
}

#' One correlation sample from the LKJ distribution
#'
#' @param d The dimension of the correlation matrix
#'
#' @param eta The scaling parameter of the LKJ distribution.
#'   Must be > 1.  Also related to the degrees of freedom nu.
#'   eta = (nu-1)/2.
#'
#' @param cholesky boolean; If `TRUE` return the cholesky
#'   decomposition.
#'
#' @return A correlation sample from the LKJ distribution
#'
#' @author Matthew Fidler (translated to RcppArmadillo) and Emma Schwager
#' @export
#' @keywords internal
rLKJ1 <- function(d, eta = 1.0, cholesky = FALSE) {
    .Call(`_rxode2_rLKJ1`, d, eta, cholesky)
}

rLKJcv1 <- function(sd, eta = 1.0) {
    .Call(`_rxode2_rLKJcv1`, sd, eta)
}

rLKJcvLsd1 <- function(logSd, logSdSD, eta = 1.0) {
    .Call(`_rxode2_rLKJcvLsd1`, logSd, logSdSD, eta)
}

#' One correlation sample from the Inverse Wishart distribution
#'
#' This correlation is constructed by transformation of the Inverse Wishart
#' random covariate to a correlation.
#'
#' @inheritParams rLKJ1
#'
#' @param nu Degrees of freedom of the Wishart distribution
#'
#' @inheritParams cvPost
#'
#' @return One correlation sample from the inverse wishart
#'
#' @author Matthew Fidler
#' @keywords internal
#' @export
invWR1d <- function(d, nu, omegaIsChol = FALSE) {
    .Call(`_rxode2_invWR1d`, d, nu, omegaIsChol)
}

rcvC1 <- function(sdEst, nu = 3.0, diagXformType = 1L, rType = 1L, returnChol = FALSE) {
    .Call(`_rxode2_rcvC1`, sdEst, nu, diagXformType, rType, returnChol)
}

cvPost_ <- function(nuS, omegaS, nS, omegaIsCholS, returnCholS, typeS, diagXformTypeS) {
    .Call(`_rxode2_cvPost_`, nuS, omegaS, nS, omegaIsCholS, returnCholS, typeS, diagXformTypeS)
}

expandTheta_ <- function(thetaS, thetaMatS, thetaLowerS, thetaUpperS, nStudS, nCoresRVS) {
    .Call(`_rxode2_expandTheta_`, thetaS, thetaMatS, thetaLowerS, thetaUpperS, nStudS, nCoresRVS)
}

expandPars_ <- function(objectS, paramsS, eventsS, controlS) {
    .Call(`_rxode2_expandPars_`, objectS, paramsS, eventsS, controlS)
}

nestingInfo_ <- function(omega, data) {
    .Call(`_rxode2_nestingInfo_`, omega, data)
}

swapMatListWithCube_ <- function(inO) {
    .Call(`_rxode2_swapMatListWithCube_`, inO)
}

omegaListRse <- function(omegaIn) {
    .Call(`_rxode2_omegaListRse`, omegaIn)
}

#' Set the CVODE linear solver
#'
#' @param type Integer: 1=dense (default), 2=band (currently aliases to dense),
#'   3=gmres, 4=bicgstab, 5=tfqmr
#' @return NULL (invisibly)
#' @export
#' @keywords internal
setCvodeLinearSolver <- function(type) {
    invisible(.Call(`_rxode2_setCvodeLinearSolver`, type))
}

#' Get the Linear Compartment Information based on the model variables
#'
#' @param obj The model variables object
#'
#' @return A named integer vector with the linear sensitivity compartments.
#'
#' @noRd
getLinInfo_ <- function(mv) {
    .Call(`_rxode2_getLinInfo_`, mv)
}

#' Get the real compartment number based on NONMEM-style compartment
#' adjusting for linear solved systems.
#'
#' This export is mostly for testing purposes.
#'
#' @param cmt The compartment number provided with traditional NONMEM numbering
#'
#' @param mv The model variables list
#'
#' @return An integer vector with the real compartment numbers
#'
#' @noRd
getCmtNum_ <- function(cmt, mv, sens = TRUE) {
    .Call(`_rxode2_getCmtNum_`, cmt, mv, sens)
}

#' Set Initial conditions to time zero instead of the first observed/dosed time
#'
#' @param ini0 When `TRUE` (default), set initial conditions to time
#'   zero. Otherwise the initial conditions are the first observed
#'   time.
#'
#' @return the boolean ini0, though this is called for its side effects
#'
#' @export
rxSetIni0 <- function(ini0 = TRUE) {
    .Call(`_rxode2_rxSetIni0`, ini0)
}

etTransEvidIsObs <- function(isObsSexp) {
    .Call(`_rxode2_etTransEvidIsObs`, isObsSexp)
}

#' See if the NONMEM compartment number supports infusion
#'
#' This export is mostly for testing purposes.
#'
#' @param cmt The compartment number provided with traditional NONMEM numbering
#'
#' @param mv The model variables list
#'
#' @return An integer vector with the real compartment numbers
#'
#' @noRd
cmtSupportsInfusion_ <- function(cmt, mv) {
    .Call(`_rxode2_cmtSupportsInfusion_`, cmt, mv)
}

#' See if the NONMEM compartment number supports infusion
#'
#' This export is mostly for testing purposes.
#'
#' @param cmt The compartment number provided with traditional NONMEM numbering
#'
#' @param mv The model variables list
#'
#' @return An integer vector with the real compartment numbers
#'
#' @noRd
cmtSupportsOff_ <- function(cmt, mv) {
    .Call(`_rxode2_cmtSupportsOff_`, cmt, mv)
}

#' Event translation for rxode2
#'
#' @param inData Data frame to translate
#'
#' @param obj object where model variables can be extracted from
#'
#' @param addCmt Add compartment to data frame (default `FALSE`).
#'
#' @param dropUnits Boolean to drop the units (default `FALSE`).
#'
#' @param allTimeVar Treat all covariates as if they were time-varying
#'
#' @param keepDosingOnly keep the individuals who only have dosing records and any
#'   trailing dosing records after the last observation.
#'
#' @param combineDvid is a boolean indicating if rxode2 will use `DVID` on observation
#'     records to change the `cmt` value; Useful for multiple-endpoint nlmixr models.  By default
#'     this is determined by `option("rxode2.combine.dvid")` and if the option has not been set,
#'     this is `TRUE`. This typically does not affect rxode2 simulations.
#'
#' @param keep This is a named vector of items you want to keep in the final rxode2 dataset.
#'     For added rxode2 event records (if seen), last observation carried forward will be used.
#'
#' @param addlKeepsCov This determines if the additional dosing items
#'   repeats the dose only (`FALSE`) or keeps the covariates at the
#'   record of the dose (`TRUE`)
#'
#' @param addlDropSs When there are steady state doses with an `addl`
#'   specification the steady state flag is dropped with repeated
#'   doses (when `TRUE`) or retained (when `FALSE`)
#'
#' @param ssAtDoseTime Boolean that when `TRUE` back calculates the
#'   steady concentration at the actual time of dose, otherwise when
#'   `FALSE` the doses are shifted
#'
#' @inheritParams rxSolve
#'
#' @return Object for solving in rxode2
#'
#' @keywords internal
#'
#' @export
etTrans <- function(inData, obj, addCmt = FALSE, dropUnits = FALSE, allTimeVar = FALSE, keepDosingOnly = FALSE, combineDvid = NULL, keep = character(0), addlKeepsCov = FALSE, addlDropSs = TRUE, ssAtDoseTime = TRUE, iCov = NULL) {
    .Call(`_rxode2_etTrans`, inData, obj, addCmt, dropUnits, allTimeVar, keepDosingOnly, combineDvid, keep, addlKeepsCov, addlDropSs, ssAtDoseTime, iCov)
}

rxEtTransAsDataFrame_ <- function(inData1) {
    .Call(`_rxode2_rxEtTransAsDataFrame_`, inData1)
}

#' Expand grid internal function
#'
#' @param c1 character vector of items to be expanded
#' @param c2 second character vector of items to be expanded
#' @param type 0 for a typical data frame, 1 for symengine sensitivity expansion
#' @return data frame (when type = 0) or symengine string (when type=1)
#' @export
#' @keywords internal
rxExpandGrid_ <- function(c1, c2, type) {
    .Call(`_rxode2_rxExpandGrid_`, c1, c2, type)
}

#' Expand sensitivity
#'
#' @param state is the state to expand
#' @param calcSens is the variables to calculate sensitivity
#' @keywords internal
#' @return symengine string for expanded sensitivity
#' @export
rxExpandSens_ <- function(state, calcSens) {
    .Call(`_rxode2_rxExpandSens_`, state, calcSens)
}

#' Expand second order sensitivity
#'
#' @param state is the state to expand
#' @param s1 is the variables to calculate sensitivity
#' @param s2 is the variables to calculate second order sensitivity
#' @keywords internal
#' @return string for symengine second order sensitivity
#' @export
rxExpandSens2_ <- function(state, s1, s2) {
    .Call(`_rxode2_rxExpandSens2_`, state, s1, s2)
}

#' Expand third order sensitivity
#'
#' Builds the symengine code for the third order forward sensitivity
#' d/dt(d^3 state / d s1 d s2 d s3) by total-differentiating the stored
#' second order sensitivity RHS (for the pair s2,s3) with respect to the
#' newly added variable s1.  Mirrors \code{rxExpandSens2_}; assumes the
#' first and second order sensitivities (and the Jacobian) are already
#' loaded in the symengine \code{model} environment.
#'
#' @param state is the state to expand
#' @param s1 is the newly added (outer) sensitivity variable
#' @param s2 is the second sensitivity variable
#' @param s3 is the third sensitivity variable
#' @keywords internal
#' @return data frame of symengine third order sensitivity code
#' @export
rxExpandSens3_ <- function(state, s1, s2, s3) {
    .Call(`_rxode2_rxExpandSens3_`, state, s1, s2, s3)
}

#' Expand d(f)/d(eta)
#'
#' @param state is the state to expand
#' @param neta is the number of etas
#' @param pred type of prediction
#' @param isTheta logical, is the expansion actually for thetas instead of etas
#' @keywords internal
#' @return String of symengine expressions to evaluate to calculate df/deta
#' @export
rxExpandFEta_ <- function(state, neta, pred, isTheta = FALSE) {
    .Call(`_rxode2_rxExpandFEta_`, state, neta, pred, isTheta)
}

#' Rep R0 for foce
#'
#' @param neta ETA to substitute
#'
#' @return Returns a string of R code to substitute the rx_r expression in the symengine environment .s
#'
#' @keywords internal
#'
#' @export
rxRepR0_ <- function(neta) {
    .Call(`_rxode2_rxRepR0_`, neta)
}

rxExpandNesting <- function(obj, nestingInfo, compile = FALSE) {
    .Call(`_rxode2_rxExpandNesting`, obj, nestingInfo, compile)
}

#' Inductive linearization solver
#'
#' @param cSub = Current subject number
#' @param op - rxode2 solving options
#' @param tp - Prior time point/time zero
#' @param yp - Prior state;  vector size = neq; Final state is updated here
#' @param tf - Final Time
#' @param InfusionRate = Rates of each compartment;  vector size = neq
#' @param on Indicator for if the compartment is "on"
#' @param cache
#'    0 = no Cache
#'    When doIndLin == 0, cache > 0 = nInf-1
#' @param ME the rxode2 matrix exponential function
#' @param IndF The rxode2 Inductive Linearization function F
#'
#' @return Returns a status for solving
#'
#'   1 = Successful solve
#'
#'   -1 = Maximum number of iterations reached when doing
#'        inductive linearization
#' @name rxIndLin_
#' @noRd
NULL

rxIndLin_ <- function(states) {
    .Call(`_rxode2_rxIndLin_`, states)
}

convertId_ <- function(x) {
    .Call(`_rxode2_convertId_`, x)
}

linCmtModelDouble <- function(dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, type, tau, tinf, amt, bolusCmt, ndiff, sensType = 3L, sensH = 0.001) {
    .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, type, tau, tinf, amt, bolusCmt, ndiff, sensType, sensH)
}

#' Variance-covariance (non-Cholesky) Omega parameterization derivatives (C++)
#'
#' C++/RcppArmadillo implementation of the non-Cholesky Omega derivatives used
#' to build a FOCEI/FOCE observed-information covariance over the natural
#' variance-covariance scale.  Returns \eqn{\Omega^{-1}}, \eqn{\log|\Omega|},
#' and their first (and optionally second) derivatives with respect to each free
#' lower-triangular variance-covariance element \eqn{\omega_{ab}}, using
#'
#' \deqn{\partial \Omega^{-1}/\partial \omega_{ab} = -\Omega^{-1} E_{ab} \Omega^{-1}}
#' \deqn{\partial \log|\Omega|/\partial \omega_{ab} = \mathrm{tr}(\Omega^{-1} E_{ab})}
#'
#' where \eqn{E_{ab}} is the symmetric single-entry basis matrix.
#'
#' @param omega symmetric positive-definite random-effects covariance matrix.
#' @param order integer; `1` for first derivatives only, `2` (default) to also
#'   return the second derivatives needed for the covariance Hessian.
#' @return a list with `omegaInv`, `logDet`, the free-element index matrix
#'   `elements` (each row `c(a, b)`, `a >= b`), first derivatives
#'   `dOmegaInv` / `dLogDet`, and (when `order = 2`) second derivatives
#'   `d2OmegaInv` / `d2LogDet`.
#' @author Hidde van de Beek
#' @keywords internal
#' @export
rxOmegaVarCovDeriv_ <- function(omega, order = 2L) {
    .Call(`_rxode2_rxOmegaVarCovDeriv_`, omega, order)
}

rxQs <- function(x) {
    .Call(`_rxode2_rxQs`, x)
}

rxQr <- function(y) {
    .Call(`_rxode2_rxQr`, y)
}

rxode2parseSetRstudio <- function(isRstudio = FALSE) {
    .Call(`_rxode2_rxode2parseSetRstudio`, isRstudio)
}

#' Silence some of rxode2's C/C++ messages
#'
#' @param silent can be 0L "noisy"  or 1L "silent"
#'
#' @keywords internal
#' @return TRUE; called for side effects
#' @export
rxParseSetSilentErr <- function(silent) {
    .Call(`_rxode2_rxParseSetSilentErr`, silent)
}

#' Get the serialization type from raw vector
#'
#' @param raw A raw vector
#'
#' @keywords internal
#'
#' @export
#'
#' @return a string indicating the serialization type:
#'    "qs2", "qdata", "qs", "bzip2", "xz",  "base", or "unknown"
rxGetSerialType_ <- function(raw) {
    .Call(`_rxode2_rxGetSerialType_`, raw)
}

#' This handles the drop list
#'
#' @param drop a character vector representing the list of items to
#'   drop from the output
#'
#' @param input The input data frame (technically a data.frame is a
#'   special list)
#'
#' @param warnDrop is a boolean that will determine if nonsense drops
#'   should be warned about in R.
#'
#' @return a data-frame with the requested items dropped.
#' @noRd
NULL

#' This gets the rxModels_ environment from the rxode2 namespace
#'
#' @return `rxModels_` environment
#' @noRd
NULL

#' This allows assignment inside the `rxModels_` environment
#'
#' @param string that should be assigned
#'
#' @param a R/S expression (SEXP) to assign in the `rxModels_`
#'   environment
#'
#' @return nothing, called for side effects
#'
#' @noRd
NULL

#' Returns (unsurprisingly) if the file exists
#'
#' @param file name
#'
#' @return boolean of if the file exists
#' @noRd
NULL

#' If nothing else works, try dispatching the `rxModelVarsS3`
#' dispatch
#'
#' @param obj is the R object trying to get the model variables from
#'
#' @return model variables (or error from `rxModelVarsS3` default method)
#' @noRd
NULL

#' Get the rxModelVars from an rxode2 object
#'
#' @param obj is the rxode2 object to get the model variables from
#'
#' @return model variables
#' @noRd
NULL

#' This gives a blank model variables object
#'
#' @return blank model variables object (though has same structure)
#'
#' @noRd
NULL

#' Get the model variables object from a character object
#'
#' @param obj character vector to convert to model variables object
#'
#' @return model variables
#'
#' @noRd
NULL

#' List to convert Model variables object into
#'
#' @param obj R list to extract model variable from.
#'
#' @return model variable object
#' @noRd
NULL

#' Check the type of an object using Rcpp
#'
#' @param obj Object to check
#'
#' @param cls Type of class.  Only s3 classes for lists/environments
#'   and primitive classes are checked.  For matrix types they are
#'   distinguished as `numeric.matrix`, `integer.matrix`,
#'   `logical.matrix`, and `character.matrix` as well as the
#'   traditional `matrix` class. Additionally checks for
#'   `event.data.frame` which is an `data.frame` object with `time`,
#'   `evid` and `amt`. (UPPER, lower or Title cases accepted)
#'
#'
#' @return A boolean indicating if the object is a member of the class.
#'
#' @keywords internal
#'
#' @author Matthew L. Fidler
#'
#' @export
#'
rxIs <- function(obj, cls) {
    .Call(`_rxode2_rxIs`, obj, cls)
}

dynLoad <- function(dll) {
    .Call(`_rxode2_dynLoad`, dll)
}

#' Get the model variables for an object
#'
#' @param obj the R object to try to extract the model variable object from.
#'
#' @return model variable object or error
#'
#' @noRd
rxModelVars_ <- function(obj) {
    .Call(`_rxode2_rxModelVars_`, obj)
}

#' State variables
#'
#' This returns the model's compartments or states.
#'
#' @inheritParams rxModelVars
#'
#' @param state is a string indicating the state or compartment that
#'     you would like to lookup.
#'
#' @return If state is missing, return a character vector of all the states.
#'
#' If state is a string, return the compartment number of the named state.
#'
#' @seealso [rxode2()]
#' @family Query model information
#' @author Matthew L.Fidler
#' @export
rxState <- function(obj = NULL, state = NULL) {
    .Call(`_rxode2_rxState`, obj, state)
}

rxParams_ <- function(obj) {
    .Call(`_rxode2_rxParams_`, obj)
}

#' Jacobian and parameter derivatives
#'
#' Return Jacobain and parameter derivatives
#'
#' @inheritParams rxModelVars
#'
#' @return A list of the jacobian parameters defined in this rxode2
#'     object.
#'
#' @author Matthew L. Fidler
#'
#' @family Query model information
#' @export
rxDfdy <- function(obj) {
    .Call(`_rxode2_rxDfdy`, obj)
}

#' Left handed Variables
#'
#' This returns the model calculated variables
#'
#' @inheritParams rxModelVars
#'
#' @return a character vector listing the calculated parameters
#' @seealso \code{\link{rxode2}}
#' @family Query model information
#' @author Matthew L.Fidler
#' @export
rxLhs <- function(obj) {
    .Call(`_rxode2_rxLhs`, obj)
}

#' Initial Values and State values for a rxode2 object
#'
#' Returns the initial values of the rxDll object
#'
#' @param obj rxDll, rxode2, or named vector representing default
#'     initial arguments
#'
#' @param vec If supplied, named vector for the model.
#'
#' @param req Required names, and the required order for the ODE solver
#'
#' @param defaultValue a number or NA representing the default value for
#'     parameters missing in `vec`, but required in `req`.
#'
#' @param noerror is a boolean specifying if an error should be thrown
#'     for missing parameter values when `default` = `NA`
#'
#' @return Initial values of the rxDll object
#'
#' @keywords internal
#' @family Query model information
#' @author Matthew L.Fidler
#' @export
rxInits <- function(obj, vec = NULL, req = NULL, defaultValue = 0, noerror = FALSE, noini = FALSE, rxLines = FALSE) {
    .Call(`_rxode2_rxInits`, obj, vec, req, defaultValue, noerror, noini, rxLines)
}

#' Setup the initial conditions.
#'
#' @param obj rxode2 object
#' @param inits A numeric vector of initial conditions.
#' @return initial conditions that were setup
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
rxSetupIni <- function(obj, inits = NULL) {
    .Call(`_rxode2_rxSetupIni`, obj, inits)
}

#' Setup the initial conditions.
#'
#' @param obj rxode2 object
#'
#' @param scale A numeric vector scales
#'
#' @param extraArgs A list of extra args to parse for initial conditions.
#'
#' @author Matthew L. Fidler
#'
#' @keywords internal
#'
#' @return setup scale for changing compartment values
#'
#' @export
rxSetupScale <- function(obj, scale = NULL, extraArgs = NULL) {
    .Call(`_rxode2_rxSetupScale`, obj, scale, extraArgs)
}

atolRtolFactor_ <- function(factor) {
    invisible(.Call(`_rxode2_atolRtolFactor_`, factor))
}

#' Simulate Parameters from a Theta/Omega specification
#'
#' @param params Named Vector of rxode2 model parameters
#'
#' @param nObs Number of observations to simulate (with `sigma` matrix)
#'
#' @inheritParams rxSolve
#'
#' @param simSubjects boolean indicated rxode2 should simulate subjects in studies (`TRUE`,
#'         default) or studies (`FALSE`)
#'
#' @return a data frame with the simulated subjects
#'
#' @author Matthew L.Fidler
#'
#' @export
rxSimThetaOmega <- function(params = NULL, omega = NULL, omegaDf = NULL, omegaLower = as.numeric( c(R_NegInf)), omegaUpper = as.numeric( c(R_PosInf)), omegaIsChol = FALSE, omegaSeparation = "auto", omegaXform = 1L, nSub = 1L, thetaMat = NULL, thetaLower = as.numeric( c(R_NegInf)), thetaUpper = as.numeric( c(R_PosInf)), thetaDf = NULL, thetaIsChol = FALSE, nStud = 1L, sigma = NULL, sigmaLower = as.numeric( c(R_NegInf)), sigmaUpper = as.numeric( c(R_PosInf)), sigmaDf = NULL, sigmaIsChol = FALSE, sigmaSeparation = "auto", sigmaXform = 1L, nCoresRV = 1L, nObs = 1L, dfSub = 0, dfObs = 0, simSubjects = TRUE, simVariability = as.logical( c(NA_LOGICAL))) {
    .Call(`_rxode2_rxSimThetaOmega`, params, omega, omegaDf, omegaLower, omegaUpper, omegaIsChol, omegaSeparation, omegaXform, nSub, thetaMat, thetaLower, thetaUpper, thetaDf, thetaIsChol, nStud, sigma, sigmaLower, sigmaUpper, sigmaDf, sigmaIsChol, sigmaSeparation, sigmaXform, nCoresRV, nObs, dfSub, dfObs, simSubjects, simVariability)
}

rxSolveSetCurObj_ <- function(obj) {
    invisible(.Call(`_rxode2_rxSolveSetCurObj_`, obj))
}

#' Free the C solving/parsing information.
#'
#' Take the ODE C system and free it.
#'
#' @keywords internal
#' @return logical indicating if the memory was successfully freed
#' @export
rxSolveFree <- function() {
    .Call(`_rxode2_rxSolveFree`)
}

#' See if the memory is installed for a solve
#'
#' @return boolean saying if the memnory is currently free for rxode2
#' @keywords internal
#' @export
#' @author Matthew L. Fidler
rxSolveSetup <- function() {
    .Call(`_rxode2_rxSolveSetup`)
}

rxSolveFromRaw_ <- function(obj, rawObj, solveState, rxControl, specParams, extraArgs, params, events, inits) {
    .Call(`_rxode2_rxSolveFromRaw_`, obj, rawObj, solveState, rxControl, specParams, extraArgs, params, events, inits)
}

rxSolve_ <- function(obj, rxControl, specParams, extraArgs, params, events, inits, setupOnly) {
    .Call(`_rxode2_rxSolve_`, obj, rxControl, specParams, extraArgs, params, events, inits, setupOnly)
}

rxSolveDollarNames <- function(obj) {
    .Call(`_rxode2_rxSolveDollarNames`, obj)
}

rxSolveGet <- function(obj, arg, exact = TRUE) {
    .Call(`_rxode2_rxSolveGet`, obj, arg, exact)
}

rxSolveUpdate <- function(obj, arg = NULL, value = NULL) {
    .Call(`_rxode2_rxSolveUpdate`, obj, arg, value)
}

rxSolveSEXP <- function(objS, rxControlS, specParamsS, extraArgsS, paramsS, eventsS, initsS, setupOnlyS) {
    .Call(`_rxode2_rxSolveSEXP`, objS, rxControlS, specParamsS, extraArgsS, paramsS, eventsS, initsS, setupOnlyS)
}

rxRmModelLib_ <- function(str) {
    invisible(.Call(`_rxode2_rxRmModelLib_`, str))
}

#' Get rxode2 model from object
#' @param obj rxode2 family of objects
#' @return rxode2 model
#' @export
rxGetrxode2 <- function(obj) {
    .Call(`_rxode2_rxGetrxode2`, obj)
}

#' Checks if the rxode2 object was built with the current build
#'
#' @inheritParams rxModelVars
#'
#' @return boolean indicating if this was built with current rxode2
#'
#' @export
rxIsCurrent <- function(obj) {
    .Call(`_rxode2_rxIsCurrent`, obj)
}

#' Assign pointer based on model variables
#' @param object rxode2 family of objects
#' @return nothing, called for side effects
#' @export
rxAssignPtr <- function(object = NULL) {
    invisible(.Call(`_rxode2_rxAssignPtr`, object))
}

#' Return the DLL associated with the rxode2 object
#'
#' This will return the dynamic load library or shared object used to
#' run the C code for rxode2.
#'
#' @param obj A rxode2 family of objects or a character string of the
#'     model specification or location of a file with a model
#'     specification.
#'
#' @return a path of the library
#'
#' @keywords internal
#' @author Matthew L.Fidler
#' @export
rxDll <- function(obj) {
    .Call(`_rxode2_rxDll`, obj)
}

#' Return the C file associated with the rxode2 object
#'
#' This will return C code for generating the rxode2 DLL.
#'
#' @param obj A rxode2 family of objects or a character string of the
#'     model specification or location of a file with a model
#'     specification.
#'
#' @return a path of the library
#'
#' @keywords internal
#' @author Matthew L.Fidler
#' @export
rxC <- function(obj) {
    .Call(`_rxode2_rxC`, obj)
}

#' Determine if the DLL associated with the rxode2 object is loaded
#'
#' @param obj A rxode2 family of objects
#'
#' @return Boolean returning if the rxode2 library is loaded.
#'
#' @keywords internal
#' @author Matthew L.Fidler
#' @export
rxIsLoaded <- function(obj) {
    .Call(`_rxode2_rxIsLoaded`, obj)
}

#' Load rxode2 object
#'
#' @param obj A rxode2 family of objects
#'
#' @return Boolean returning if the rxode2 library is loaded.
#'
#' @keywords internal
#' @author Matthew L.Fidler
#' @export
rxDynLoad <- function(obj) {
    .Call(`_rxode2_rxDynLoad`, obj)
}

#' Lock/unlocking of rxode2 dll file
#'
#' @param obj A rxode2 family of objects
#'
#' @return nothing; called for side effects
#'
#' @export
rxLock <- function(obj) {
    .Call(`_rxode2_rxLock`, obj)
}

#' @rdname rxLock
#' @export
rxUnlock <- function(obj) {
    .Call(`_rxode2_rxUnlock`, obj)
}

#' Allow unloading of dlls
#'
#' @param allow boolean indicating if garbage collection will unload of rxode2 dlls.
#'
#' @return Boolean allow; called for side effects
#'
#' @examples
#'
#' # Garbage collection will not unload un-used rxode2 dlls
#' rxAllowUnload(FALSE);
#'
#' # Garbage collection will unload unused rxode2 dlls
#' rxAllowUnload(TRUE);
#' @export
#' @author Matthew Fidler
rxAllowUnload <- function(allow) {
    .Call(`_rxode2_rxAllowUnload`, allow)
}

rxUnloadAll_ <- function() {
    .Call(`_rxode2_rxUnloadAll_`)
}

#' Unload rxode2 object
#'
#' @param obj A rxode2 family of objects
#'
#' @return Boolean returning if the rxode2 library is loaded.
#'
#' @keywords internal
#' @author Matthew L.Fidler
#' @export
rxDynUnload <- function(obj) {
    .Call(`_rxode2_rxDynUnload`, obj)
}

#' Delete the DLL for the model
#'
#' This function deletes the DLL, but doesn't delete the model
#' information in the object.
#'
#' @param obj rxode2 family of objects
#'
#' @return A boolean stating if the operation was successful.
#'
#' @author Matthew L.Fidler
#' @export
rxDelete <- function(obj) {
    .Call(`_rxode2_rxDelete`, obj)
}

setRstudio <- function(isRstudio = FALSE) {
    .Call(`_rxode2_setRstudio`, isRstudio)
}

setProgSupported <- function(isSupported = 1L) {
    .Call(`_rxode2_setProgSupported`, isSupported)
}

getProgSupported <- function() {
    .Call(`_rxode2_getProgSupported`)
}

rxUpdateTrans_ <- function(ret, prefix, libName) {
    .Call(`_rxode2_rxUpdateTrans_`, ret, prefix, libName)
}

dropUnitsRxSolve <- function(x) {
    .Call(`_rxode2_dropUnitsRxSolve`, x)
}

#' Silence some of rxode2's C/C++ messages
#'
#' @param silent can be 0L "noisy"  or 1L "silent"
#'
#' @keywords internal
#' @return TRUE; called for side effects
#' @export
rxSetSilentErr <- function(silent) {
    .Call(`_rxode2_rxSetSilentErr`, silent)
}

#' Invert matrix using RcppArmadillo.
#'
#' @param matrix matrix to be inverted.
#'
#' @return inverse or pseudo inverse of matrix.
#'
#' @export
rxInv <- function(matrix) {
    .Call(`_rxode2_rxInv`, matrix)
}

#' Get Omega^-1 and derivatives
#'
#' @param invObjOrMatrix Object for inverse-type calculations.  If
#'   this is a matrix, setup the object for inversion
#'   [rxSymInvCholCreate()] with the default arguments and return a
#'   reactive s3 object.  Otherwise, use the inversion object to
#'   calculate the requested derivative/inverse.
#'
#' @param theta Thetas to be used for calculation.  If missing (`NULL`), a
#'     special s3 class is created and returned to access `Omega^1`
#'     objects as needed and cache them based on the theta that is
#'     used.
#'
#' @param type The type of object.  Currently the following types are
#'     supported:
#'
#' * `cholOmegaInv` gives the
#'     Cholesky decomposition of the Omega Inverse matrix.
#' * `omegaInv` gives the Omega Inverse matrix.
#' * `d(omegaInv)` gives the `d(Omega^-1)` withe respect to the
#'     theta parameter specified in `thetaNumber`.
#' * `d(D)` gives the `d(diagonal(Omega^-1))` with respect to
#'     the theta parameter specified in the `thetaNumber`
#'     parameter
#'
#' @param thetaNumber For types `d(omegaInv)` and `d(D)`,
#'     the theta number that the derivative is taken against.  This
#'     must be positive from 1 to the number of thetas defining the
#'     Omega matrix.
#'
#' @return Matrix based on parameters or environment with all the
#'     matrixes calculated in variables `omega`, `omegaInv`, `dOmega`,
#'     `dOmegaInv`.
#'
#' @author Matthew L. Fidler
#'
#' @export
rxSymInvChol <- function(invObjOrMatrix, theta = NULL, type = "cholOmegaInv", thetaNumber = 0L) {
    .Call(`_rxode2_rxSymInvChol`, invObjOrMatrix, theta, type, thetaNumber)
}

rxSymInvCholEnvCalculate <- function(obj, what, theta = NULL) {
    .Call(`_rxode2_rxSymInvCholEnvCalculate`, obj, what, theta)
}

#' Report the byte sizes of every rxode2 solver memory allocation
#'
#' Returns a named numeric vector of byte counts for every buffer rxode2
#' allocates during \code{rxSolve()}.  Values are computed with the same
#' formulas used by the actual allocator in \code{rxData.cpp} via the shared
#' \code{rxFillMemLayout()} function, so any change to the allocator
#' automatically changes the estimate.
#'
#' @param neq       Number of ODE states (\code{length(rxModelVars(model)$state)}).
#' @param stateSize Effective \code{state.size()} passed to the solver.
#'   For pure ODE models this equals \code{neq}; for linCmt-only models it
#'   may be 0.  Use \code{neq} when in doubt.
#' @param nlhs      Number of LHS (calculated) outputs.
#' @param npars     Number of model parameters (for \code{gpars} estimate).
#' @param neta      Number of random effects (etas).
#' @param neps      Number of residual-error levels (epsilons).
#' @param ncov      Number of time-varying covariates.
#' @param nsim      Number of simulations.
#' @param cores     Number of parallel OMP threads.
#' @param nMtime    Number of model measurement times.
#' @param extraCmt  Extra compartments (0, 1 = depot, 2 = depot+central).
#' @param linB      1 if using an analytical linear-compartment model, else 0.
#' @param nLlik     Number of log-likelihood terms (FOCEi).
#' @param nIndSim   Per-individual simulation count (typically \code{neta+neps}).
#' @param numLinSens Number of linear sensitivity parameters (FOCEi mixed models).
#' @param numLin    Number of linear compartment terms (FOCEi mixed models).
#' @param nsub      Number of subjects.
#' @param nallTotal Total events across all subjects (sum of obs + doses).
#' @param maxAllTimes Maximum events for any single subject.
#' @return Named numeric vector; each element is bytes for that allocation.
#'   Also includes \code{sizeofInd} (bytes per \code{rx_solving_options_ind}
#'   struct) and \code{rxLlikSaveSize} (the compile-time constant).
#' @noRd
rxMemoryComponents_ <- function(neq, stateSize, nlhs, npars, neta, neps, ncov, nsim, cores, nMtime, extraCmt, linB, nLlik, nIndSim, numLinSens, numLin, nsub, nallTotal, maxAllTimes) {
    .Call(`_rxode2_rxMemoryComponents_`, neq, stateSize, nlhs, npars, neta, neps, ncov, nsim, cores, nMtime, extraCmt, linB, nLlik, nIndSim, numLinSens, numLin, nsub, nallTotal, maxAllTimes)
}

rxOptRep_ <- function(input) {
    .Call(`_rxode2_rxOptRep_`, input)
}

#' Save the pre-integration rxode2 solver state to a binary file
#'
#' Called internally after event-table and parameter setup but before ODE
#' integration.  The file can later be restored with rxRestoreState_().
#'
#' @param path File path to write (created/overwritten).
#' @return Invisibly TRUE.
#' @noRd
rxSaveState_ <- function() {
    .Call(`_rxode2_rxSaveState_`)
}

#' Check whether a file was written by rxSaveState_
#' @noRd
rxIsSerializeFile_ <- function(rawSexp) {
    .Call(`_rxode2_rxIsSerializeFile_`, rawSexp)
}

#' Restore a pre-integration rxode2 solver state from binary file
#'
#' @param path File path written by rxSaveState_().
#' @return Invisibly TRUE.
#' @noRd
rxRestoreState_ <- function(rawSexp) {
    .Call(`_rxode2_rxRestoreState_`, rawSexp)
}

rxStack_ <- function(Data, vars = NULL) {
    .Call(`_rxode2_rxStack_`, Data, vars)
}

rxModelVarsStack <- function(x) {
    .Call(`_rxode2_rxModelVarsStack`, x)
}

rxRmvn_ <- function(A_, mu, sigma, ncores = 1L, isChol = FALSE) {
    .Call(`_rxode2_rxRmvn_`, A_, mu, sigma, ncores, isChol)
}

rxMvnrnd <- function(n, L, l, u, mu, a = 0.4, tol = 2.05) {
    .Call(`_rxode2_rxMvnrnd`, n, L, l, u, mu, a, tol)
}

rxCholperm <- function(Sig, l, u, eps = 1e-10) {
    .Call(`_rxode2_rxCholperm`, Sig, l, u, eps)
}

rxGradpsi <- function(y, L, l, u) {
    .Call(`_rxode2_rxGradpsi`, y, L, l, u)
}

rxNleq <- function(l, u, L) {
    .Call(`_rxode2_rxNleq`, l, u, L)
}

rxMvrandn_ <- function(A_, mu, sigma, lower, upper, ncores = 1L, a = 0.4, tol = 2.05, nlTol = 1e-10, nlMaxiter = 100L) {
    .Call(`_rxode2_rxMvrandn_`, A_, mu, sigma, lower, upper, ncores, a, tol, nlTol, nlMaxiter)
}

#' This seeds the engine based on the number of cores used in random number generation
#'
#' @param ncores is the number of cores to use.
#' @keywords internal
#' @return Nothing, called for side effects
#' @export
#' @examples
#' rxSeedEng()
rxSeedEng <- function(ncores = 1L) {
    .Call(`_rxode2_rxSeedEng`, ncores)
}

rxnbinomMu_ <- function(size, mu, n, ncores) {
    .Call(`_rxode2_rxnbinomMu_`, size, mu, n, ncores)
}

rxnbinom_ <- function(size, prob, n, ncores) {
    .Call(`_rxode2_rxnbinom_`, size, prob, n, ncores)
}

rxbinom_ <- function(n0, prob, n, ncores) {
    .Call(`_rxode2_rxbinom_`, n0, prob, n, ncores)
}

rxcauchy_ <- function(location, scale, n, ncores) {
    .Call(`_rxode2_rxcauchy_`, location, scale, n, ncores)
}

rxchisq_ <- function(df, n, ncores) {
    .Call(`_rxode2_rxchisq_`, df, n, ncores)
}

rxexp_ <- function(rate, n, ncores) {
    .Call(`_rxode2_rxexp_`, rate, n, ncores)
}

rxf_ <- function(df1, df2, n, ncores) {
    .Call(`_rxode2_rxf_`, df1, df2, n, ncores)
}

rxgamma_ <- function(shape, rate, n, ncores) {
    .Call(`_rxode2_rxgamma_`, shape, rate, n, ncores)
}

rxbeta_ <- function(shape1, shape2, n, ncores) {
    .Call(`_rxode2_rxbeta_`, shape1, shape2, n, ncores)
}

rxgeom_ <- function(prob, n, ncores) {
    .Call(`_rxode2_rxgeom_`, prob, n, ncores)
}

rxnorm_ <- function(mean, sd, n, ncores) {
    .Call(`_rxode2_rxnorm_`, mean, sd, n, ncores)
}

rxpois_ <- function(lambda, n, ncores) {
    .Call(`_rxode2_rxpois_`, lambda, n, ncores)
}

rxt__ <- function(df, n, ncores) {
    .Call(`_rxode2_rxt__`, df, n, ncores)
}

rxunif_ <- function(low, hi, n, ncores) {
    .Call(`_rxode2_rxunif_`, low, hi, n, ncores)
}

rxweibull_ <- function(shape, scale, n, ncores) {
    .Call(`_rxode2_rxweibull_`, shape, scale, n, ncores)
}

rxRmvn0 <- function(A_, mu, sigma, lower, upper, ncores = 1L, isChol = FALSE, a = 0.4, tol = 2.05, nlTol = 1e-10, nlMaxiter = 100L) {
    .Call(`_rxode2_rxRmvn0`, A_, mu, sigma, lower, upper, ncores, isChol, a, tol, nlTol, nlMaxiter)
}

rxRmvnSEXP <- function(nS, muS, sigmaS, lowerS, upperS, ncoresS, isCholS, keepNamesS, aS, tolS, nlTolS, nlMaxiterS) {
    .Call(`_rxode2_rxRmvnSEXP`, nS, muS, sigmaS, lowerS, upperS, ncoresS, isCholS, keepNamesS, aS, tolS, nlTolS, nlMaxiterS)
}

rpp_ <- function(nS, lambdaS, gammaS, probS, t0S, tmaxS, randomOrderS) {
    .Call(`_rxode2_rpp_`, nS, lambdaS, gammaS, probS, t0S, tmaxS, randomOrderS)
}

rxordSelect <- function(u, cs) {
    .Call(`_rxode2_rxordSelect`, u, cs)
}

isNullZero <- function(obj) {
    .Call(`_rxode2_isNullZero`, obj)
}

rxErf <- function(v) {
    .Call(`_rxode2_rxErf`, v)
}

binomProbsPredVec_ <- function(n, m, Y, M, doP = TRUE, tol = 1e-7) {
    .Call(`_rxode2_binomProbsPredVec_`, n, m, Y, M, doP, tol)
}

binomProbs_ <- function(x, probs, naRm, nIn, cont) {
    .Call(`_rxode2_binomProbs_`, x, probs, naRm, nIn, cont)
}

meanProbs_ <- function(x, probs, naRm, useT, pred, nIn) {
    .Call(`_rxode2_meanProbs_`, x, probs, naRm, useT, pred, nIn)
}

rxTest({
  test_that("test the matrices of the linear compartment solutions .solComp2", {

    v <- list(L = c(4.07546291005291, 0.0245370899470903),
              C1 = structure(c(0.759200006770939,-0.740571447916969,
                               0.240799993229061, 0.740571447916969),
                             dim = c(2L, 2L)),
              C2 = structure(c(-0.246857149305656, 0.240799993229061,
                               0.246857149305656, 0.759200006770939),
                             dim = c(2L, 2L)))

    expect_equal(.solComp2(k10=0.1, k12=3, k21=1), v)

    v <- list(L = c(5.8977832399092, 0.0122878774411843, 0.689928882649618),
              C1 = structure(c(0.86252789433926, -0.528317313419071, -0.319585969277925,
                               0.120784508040635, 0.366861472939276, 0.495310665672737,
                               0.016687597620105, 0.161455840479795, -0.175724696394812),
                             dim = c(3L, 3L)),
              C2 = structure(c(-0.17610577113969, 0.107868659665074, 0.0652511460029109,
                               0.122287157646425, 0.371425504011093, 0.501472700759773,
                               0.053818613493265, 0.520705836323833, -0.566723846762684),
                             dim = c(3L, 3L)),
              C3 = structure(c(-0.0798964923194813, 0.0489383595021832, 0.0296034459956658,
                               0.123827666418184, 0.37610452556983, 0.507789987948271,
                               -0.043931174098703, -0.425042885072013, 0.462606566056063),
                             dim = c(3L,  3L)))

    expect_equal(.solComp3(k10=0.1, k12=3, k21=1, k13=2, k31=0.5),
                 v)

  })

  if (requireNamespace("pmxTools", quietly = TRUE)) {

    f <- function(dt, CL=25, V=20, DOSE=100) {
      p1 <- CL
      v1 <- V
      p2 <- 0
      p3 <- 0
      p4 <- 0
      p5 <- 0
      ka <- 0
      alastNV <- DOSE
      rateNV <- 0
      oral0 <- 0
      trans <- 1
      ncmt <- 1
      deriv <- FALSE
      l <- .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)$val
      c(s=pmxTools::calc_sd_1cmt_linear_bolus(CL=CL, V=V, t=dt, dose=DOSE), l=l)
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             test_that(paste0("test the one compartment linear compartment solution at ", d), {
               v <- f(d)
               expect_equal(stats::setNames(v["s"], NULL),
                            stats::setNames(v["l"], NULL))
             })
           })


    f <- function(dt, CL=25, V=20, KA=2, DOSE=100) {
      p1 <- CL
      v1 <- V
      p2 <- 0
      p3 <- 0
      p4 <- 0
      p5 <- 0
      ka <- KA
      alastNV <- c(DOSE, 0)
      rateNV <- c(0, 0)
      oral0 <- 1
      trans <- 1
      ncmt <- 1
      deriv <- FALSE
      l <- .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)$val
      c(s=pmxTools::calc_sd_1cmt_linear_oral_1(CL=CL, V=V, ka=KA, t=dt, dose=DOSE),
        l=l)
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             test_that(paste0("test the oral 1-cmt linear compartment solution at ", d), {
               v <- f(d)
               expect_equal(stats::setNames(v["s"], NULL),
                            stats::setNames(v["l"], NULL))
             })
           })


    f <- function(t, CL=25, V=20, DOSE=100, tinf=1) {
      p1 <- CL
      v1 <- V
      p2 <- 0
      p3 <- 0
      p4 <- 0
      p5 <- 0
      ka <- 0
      deriv <- FALSE
      if (t < tinf) {
        alastNV <- 0
        rateNV <- DOSE/tinf
        dt <- t
      } else {
        alastNV <- pmxTools::calc_sd_1cmt_linear_infusion(CL=CL, V=V, t=tinf, dose=DOSE, tinf=tinf)*V # amt not conc
        rateNV <- 0
        dt <- t - tinf
      }
      oral0 <- 0
      trans <- 1
      ncmt <- 1
      l <- .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
      l <- l$val
      c(s=pmxTools::calc_sd_1cmt_linear_infusion(CL=CL, V=V, t=t, dose=DOSE, tinf=tinf),
        l=l)
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             test_that(paste0("test the iv infusion 1-cmt linear compartment solution at ", d), {
               v <- f(d)
               expect_equal(stats::setNames(v["s"], NULL),
                            stats::setNames(v["l"], NULL))
             })
           })

    f <- function(dt, V = 40, CL = 18, V2 = 297, Q = 10, DOSE=100) {
      p1 <- CL
      v1 <- V
      p2 <- Q
      p3 <- V2
      p4 <- 0
      p5 <- 0
      ka <- 0
      alastNV <- c(DOSE, 0)
      rateNV <- c(0, 0)
      oral0 <- 0
      trans <- 1
      ncmt <- 2
      deriv <- FALSE
      l <- .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)$val
      c(s=pmxTools::calc_sd_2cmt_linear_bolus(CL=CL, V=V, V2=V2, Q=Q,
                                              t=dt, dose=DOSE), l=l)
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             test_that(paste0("test the two compartment linear compartment solution at ", d), {
               v <- f(d)
               expect_equal(stats::setNames(v["s"], NULL),
                            stats::setNames(v["l"], NULL))
             })
           })

    f <- function(dt, V = 40, CL = 18, V2 = 297, Q = 10, DOSE=100, tinf=1) {
      p1 <- CL
      v1 <- V
      p2 <- Q
      p3 <- V2
      p4 <- 0
      p5 <- 0
      ka <- 0
      alastNV <- c(0, 0)
      rateNV <- DOSE/tinf
      oral0 <- 0
      trans <- 1
      ncmt <- 2
      extra <- 0
      deriv <- FALSE
      if (dt <= tinf) {
      } else {
        l <- .Call(`_rxode2_linCmtModelDouble`, tinf, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
        dt <- dt - tinf
        extra <- tinf
        rateNV <- 0
        alastNV <- l$Alast
      }
      l <- .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)$val
      c(s=pmxTools::calc_sd_2cmt_linear_infusion(CL=CL, V=V, V2=V2, Q=Q,
                                                 t=dt+extra, dose=DOSE, tinf=tinf),
        l=l)
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             test_that(paste0("test the two compartment iv infusion linear compartment solution at ", d), {
               v <- f(d)
               expect_equal(stats::setNames(v["s"], NULL),
                            stats::setNames(v["l"], NULL))
             })
           })

    # 3 compartment tests
    f <- function(dt,
                  V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400,
                  DOSE=100) {
      p1 <- CL
      v1 <- V
      p2 <- Q
      p3 <- V2
      p4 <- Q2
      p5 <- V3
      ka <- 0
      alastNV <- c(DOSE, 0, 0)
      rateNV <- c(0, 0)
      oral0 <- 0
      trans <- 1
      ncmt <- 3
      deriv <- FALSE
      l <- .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)$val
      c(s=pmxTools::calc_sd_3cmt_linear_bolus(CL=CL, V=V, V2=V2, Q=Q,
                                              V3=V3, Q3=Q2,
                                              t=dt, dose=DOSE), l=l)
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             test_that(paste0("test the two compartment linear compartment solution at ", d), {
               v <- f(d)
               expect_equal(stats::setNames(v["s"], NULL),
                            stats::setNames(v["l"], NULL))
             })
           })

    # 3 compartment tests, oral
    f <- function(dt,
                  V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400,
                  ka=2,
                  DOSE=100) {
      p1 <- CL
      v1 <- V
      p2 <- Q
      p3 <- V2
      p4 <- Q2
      p5 <- V3
      ka <- ka
      alastNV <- c(DOSE, 0, 0, 0)
      rateNV <- c(0, 0)
      oral0 <- 1
      trans <- 1
      ncmt <- 3
      deriv <- FALSE
      l <- .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)$val
      c(s=pmxTools::calc_sd_3cmt_linear_oral_1(CL=CL, V=V, V2=V2, Q=Q,
                                               V3=V3, Q3=Q2,
                                               ka=ka,
                                               t=dt, dose=DOSE), l=l)
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             test_that(paste0("test the three compartment linear oral compartment solution at ", d), {
               v <- f(d)
               expect_equal(stats::setNames(v["s"], NULL),
                            stats::setNames(v["l"], NULL))
             })
           })


    f <- function(dt, V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400,
                  DOSE=100, tinf=1) {
      p1 <- CL
      v1 <- V
      p2 <- Q
      p3 <- V2
      p4 <- Q2
      p5 <- V3
      ka <- 0
      alastNV <- c(0, 0, 0)
      rateNV <- DOSE/tinf
      oral0 <- 0
      trans <- 1
      ncmt <- 3
      extra <- 0
      deriv <- FALSE
      if (dt <= tinf) {
      } else {
        l <- .Call(`_rxode2_linCmtModelDouble`, tinf, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
        dt <- dt - tinf
        extra <- tinf
        rateNV <- 0
        alastNV <- l$Alast
      }
      l <- .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)$val
      c(s=pmxTools::calc_sd_3cmt_linear_infusion(CL=CL, V=V, V2=V2, Q=Q, V3=V3, Q3=Q2,
                                                 t=dt+extra, dose=DOSE, tinf=tinf),
        l=l)
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             test_that(paste0("test the three compartment linear infusion compartment solution at ", d), {
               v <- f(d)
               expect_equal(stats::setNames(v["s"], NULL),
                            stats::setNames(v["l"], NULL))
             })
           })


    ####################################################
    ## AD linCmt()
    ####################################################

    f0 <- function(dt, CL=25, V=20, DOSE=100,
                   alastNV=c(DOSE, 0, 0)) {
      p1 <- CL
      v1 <- V
      p2 <- 0
      p3 <- 0
      p4 <- 0
      p5 <- 0
      ka <- 0
      rateNV <- 0
      oral0 <- 0
      trans <- 1
      ncmt <- 1
      deriv <- TRUE
      l <- .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
      l
      #c(s=pmxTools::calc_sd_1cmt_linear_bolus(CL=CL, V=V, t=dt, dose=DOSE), l=l$val)
    }

    f <-  function(dt, CL=25, V=20, DOSE=100) {
      test_that(paste0("one compartment bolus t=", dt, ";CL=", CL,
                       "; V=", V, "; DOSE=", DOSE), {
                         t0 <- dt/2
                         f1 <- f0(t0, CL=CL, V=V, DOSE=DOSE)
                         expect_equal(pmxTools::calc_sd_1cmt_linear_bolus(CL=CL, V=V, t=t0, dose=DOSE),
                                      f1$val)
                         #
                         f2 <- f0(t0, CL=CL, V=V, DOSE=DOSE, alastNV=f1$Alast)
                         #
                         f3 <- f0(t0*2, CL=CL, V=V, DOSE=DOSE)
                         #
                         expect_equal(pmxTools::calc_sd_1cmt_linear_bolus(CL=CL, V=V, t=dt, dose=DOSE),
                                      f3$val)
                         # Value is in concentration
                         expect_equal(f2$val, f3$val)
                         # Unadjusted Jacobian
                         expect_equal(f2$J, f3$J)
                         # Jacobian adjusted to concentration
                         expect_equal(f2$Jg, f3$Jg)
                         # Alast and gradients are in amounts
                         expect_equal(f2$Alast, f3$Alast)
                       })
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             f(d, CL=25, V=20, DOSE=100)
           })

    f0 <- function(dt, CL=25, V=20, KA=2, DOSE=100,
                   alastNV=c(DOSE, 0, 0, 0, 0, 0)) {
      p1 <- CL
      v1 <- V
      p2 <- 0
      p3 <- 0
      p4 <- 0
      p5 <- 0
      ka <- KA
      alastNV <- alastNV
      rateNV <- c(0, 0)
      oral0 <- 1
      trans <- 1
      ncmt <- 1
      deriv <- TRUE
      l <- .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1,
                 p2, p3,
                 p4, p5,
                 ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
      l
    }

    f <-  function(dt, CL=25, V=20, KA=2, DOSE=100) {
      test_that(paste0("one compartment oral t=", dt, ";CL=", CL,
                       "; V=", V, "; KA=", KA, "; DOSE=", DOSE), {
                         t0 <- dt/2
                         f1 <- f0(t0, CL=CL, V=V, KA=KA, DOSE=DOSE)
                         expect_equal(pmxTools::calc_sd_1cmt_linear_oral_1(CL=CL, V=V, ka=KA, t=t0, dose=DOSE),
                                      f1$val)
                         #
                         f2 <- f0(t0, CL=CL, V=V, KA=KA, DOSE=DOSE, alastNV=f1$Alast)
                         #
                         f3 <- f0(t0*2, CL=CL, V=V, KA=KA, DOSE=DOSE)
                         #
                         expect_equal(pmxTools::calc_sd_1cmt_linear_oral_1(CL=CL, V=V, ka=KA, t=dt, dose=DOSE),
                                      f3$val)
                         # Value is in concentration
                         expect_equal(f2$val, f3$val)
                         # Unadjusted Jacobian
                         expect_equal(f2$J, f3$J)
                         # Jacobian adjusted to concentration
                         expect_equal(f2$Jg, f3$Jg)
                         # Alast and gradients are in amounts
                         expect_equal(f2$Alast, f3$Alast)
                       })
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             f(d, CL=25, V=20, KA=2, DOSE=100)
           })

    ## One compartment IV infusion

    f0i <- function(t, CL=25, V=20, DOSE=100, tinf=1,
                    alastNV=rep(0, 3), rateNV=DOSE/tinf) {
      p1 <- CL
      v1 <- V
      p2 <- 0
      p3 <- 0
      p4 <- 0
      p5 <- 0
      ka <- 0
      deriv <- TRUE
      oral0 <- 0
      trans <- 1
      ncmt <- 1
      if (all(rateNV == 0)) {
        # Now a simple eliminiation; Handles the case where the infusion has already past
        dt <- t
      } else if (t < tinf) {
        # infusion not complete
        dt <- t
      } else {
        #Infusion completes during the time
        v <- .Call(`_rxode2_linCmtModelDouble`, tinf, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
        if (t == tinf) return(v)
        # Infusion is now complete
        alastNV <- v$Alast
        rateNV <- 0
        dt <- t - tinf
      }
      .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
    }

    f <-  function(dt, CL=25, V=20, DOSE=100, tinf=1) {
      test_that(paste0("one compartment infusion t=", dt, ";CL=", CL,
                       "; V=", V, "; tinf=", tinf, "; DOSE=", DOSE), {
                         t0 <- dt/2
                         # See if more care is needed for IV comparison
                         if (dt < tinf) {
                           f1 <- f0i(t0, CL=CL, V=V, DOSE=DOSE, tinf=tinf)
                           expect_equal(pmxTools::calc_sd_1cmt_linear_infusion(CL=CL, V=V, t=t0, dose=DOSE, tinf=tinf),
                                        f1$val)
                           #
                           f2 <- f0i(t0, CL=CL, V=V, DOSE=DOSE, tinf=tinf, alastNV=f1$Alast)
                           #
                           f3 <- f0i(t0*2, CL=CL, V=V, DOSE=DOSE)
                           expect_equal(pmxTools::calc_sd_1cmt_linear_infusion(CL=CL, V=V, t=t0*2, dose=DOSE, tinf=tinf),
                                        f3$val)
                           # Unadjusted Jacobian
                           expect_equal(f2$J, f3$J)
                           # Jacobian adjusted to concentration
                           expect_equal(f2$Jg, f3$Jg)
                           # Alast and gradients are in amounts
                           expect_equal(f2$Alast, f3$Alast)
                           # Value is in concentration
                           expect_equal(f2$val, f3$val)
                         } else if (t0 > tinf) {
                           rateNV <- DOSE/tinf
                           f1 <- f0i(t0, CL=CL, V=V, DOSE=DOSE, tinf=tinf, rateNV=rateNV)
                           expect_equal(pmxTools::calc_sd_1cmt_linear_infusion(CL=CL, V=V, t=t0, dose=DOSE, tinf=tinf),
                                        f1$val)
                           #
                           f2 <- f0i(t0, CL=CL, V=V, DOSE=DOSE, tinf=tinf, rateNV=0,
                                     alastNV=f1$Alast)
                           expect_equal(pmxTools::calc_sd_1cmt_linear_infusion(CL=CL, V=V, t=t0*2, dose=DOSE, tinf=tinf),
                                        f2$val)
                           f3 <- f0i(t0*2, CL=CL, V=V, DOSE=DOSE, tinf=tinf)
                           expect_equal(pmxTools::calc_sd_1cmt_linear_infusion(CL=CL, V=V, t=t0*2, dose=DOSE, tinf=tinf),
                                        f3$val)
                           # Unadjusted Jacobian
                           expect_equal(f2$J, f3$J)
                           # Jacobian adjusted to concentration
                           expect_equal(f2$Jg, f3$Jg)
                           # Alast and gradients are in amounts
                           expect_equal(f2$Alast, f3$Alast)
                           # Value is in concentration
                           expect_equal(f2$val, f3$val)
                         } else {
                           rateNV <- DOSE/tinf
                           f1 <- f0i(t0, CL=CL, V=V, DOSE=DOSE, tinf=tinf, rateNV=rateNV)
                           expect_equal(pmxTools::calc_sd_1cmt_linear_infusion(CL=CL, V=V, t=t0, dose=DOSE, tinf=tinf),
                                        f1$val)
                           f3 <- f0i(t0*2, CL=CL, V=V, DOSE=DOSE, tinf=tinf)
                           expect_equal(pmxTools::calc_sd_1cmt_linear_infusion(CL=CL, V=V, t=t0*2, dose=DOSE, tinf=tinf),
                                        f3$val)
                           tinf2 <- tinf-t0
                           f2 <- f0i(t0, CL=CL, V=V, DOSE=DOSE, tinf=tinf2, rateNV=rateNV,
                                     alastNV=f1$Alast)
                           expect_equal(pmxTools::calc_sd_1cmt_linear_infusion(CL=CL, V=V, t=t0*2, dose=DOSE, tinf=tinf),
                                        f3$val)
                           # Unadjusted Jacobian
                           expect_equal(f2$J, f3$J)
                           # Jacobian adjusted to concentration
                           expect_equal(f2$Jg, f3$Jg)
                           # Alast and gradients are in amounts
                           expect_equal(f2$Alast, f3$Alast)
                           # Value is in concentration
                           expect_equal(f2$val, f3$val)
                         }
                       })
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             f(d, CL=25, V=20, tinf=1, DOSE=100)
           })

    ## Two compartment bolus
    f0 <- function(dt, V = 40, CL = 18, V2 = 297, Q = 10, DOSE=100,
                   alastNV=c(DOSE, rep(0, 9))) {
      p1 <- CL
      v1 <- V
      p2 <- Q
      p3 <- V2
      p4 <- 0
      p5 <- 0
      ka <- 0
      alastNV <- alastNV
      rateNV <- c(0, 0)
      oral0 <- 0
      trans <- 1
      ncmt <- 2
      deriv <- TRUE
      .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
    }

    f <-  function(dt, V = 40, CL = 18, V2 = 297, Q = 10, DOSE=100) {
      test_that(paste0("two compartment bolus t=", dt, ";CL=", CL,
                       "; V=", V, "; Q=", Q, "; V2=", V2, "; DOSE=", DOSE), {
                         t0 <- dt/2
                         f1 <- f0(t0, CL=CL, V=V, Q=Q, V2=V2, DOSE=DOSE)
                         expect_equal(pmxTools::calc_sd_2cmt_linear_bolus(CL=CL, V=V,
                                                                          V2=V2, Q=Q,
                                                                          t=t0, dose=DOSE),
                                      f1$val)
                         #
                         f2 <- f0(t0, CL=CL, V=V, Q=Q, V2=V2,
                                  DOSE=DOSE, alastNV=f1$Alast)
                         #
                         f3 <- f0(t0*2, CL=CL, V=V, Q=Q, V2=V2,
                                  DOSE=DOSE)
                         #
                         expect_equal(
                           pmxTools::calc_sd_2cmt_linear_bolus(CL=CL, V=V, V2=V2, Q=Q,
                                                               t=dt, dose=DOSE),
                           f3$val)
                         # Value is in concentration
                         expect_equal(f2$val, f3$val)
                         # Unadjusted Jacobian
                         expect_equal(f2$J, f3$J)
                         # Jacobian adjusted to concentration
                         expect_equal(f2$Jg, f3$Jg)
                         # Alast and gradients are in amounts
                         expect_equal(f2$Alast, f3$Alast)
                       })
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             f(d,V = 40, CL = 18, V2 = 297, Q = 10, DOSE=100)
           })

    ## Two compartment oral
    f0 <- function(dt, V = 40, CL = 18, V2 = 297, Q = 10, DOSE=100, ka=2,
                   alastNV=c(DOSE, rep(0, 13))) {
      p1 <- CL
      v1 <- V
      p2 <- Q
      p3 <- V2
      p4 <- 0
      p5 <- 0
      ka <- ka
      alastNV <- alastNV
      rateNV <- c(0, 0)
      oral0 <- 1
      trans <- 1
      ncmt <- 2
      deriv <- TRUE
      .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
    }

    f <-  function(dt, V = 40, CL = 18, V2 = 297, Q = 10, ka=2, DOSE=100) {
      test_that(paste0("two compartment oral t=", dt, ";CL=", CL,
                       "; V=", V, "; Q=", Q, "; V2=", V2, "; ka=", ka, "DOSE=", DOSE), {
                         t0 <- dt/2
                         f1 <- f0(t0, CL=CL, V=V, Q=Q, V2=V2, ka=ka, DOSE=DOSE)
                         expect_equal(pmxTools::calc_sd_2cmt_linear_oral_1(
                           CL=CL, V=V,
                           V2=V2, Q=Q, ka=ka,
                           t=t0, dose=DOSE),
                           f1$val)
                         #
                         f2 <- f0(t0, CL=CL, V=V, Q=Q, V2=V2, ka=ka,
                                  DOSE=DOSE, alastNV=f1$Alast)
                         #
                         f3 <- f0(t0*2, CL=CL, V=V, Q=Q, V2=V2, ka=ka,
                                  DOSE=DOSE)
                         #
                         expect_equal(
                           pmxTools::calc_sd_2cmt_linear_oral_1(
                             CL=CL, V=V, V2=V2, Q=Q, ka=ka,
                             t=dt, dose=DOSE),
                           f3$val)
                         # Value is in concentration
                         expect_equal(f2$val, f3$val)
                         # Unadjusted Jacobian
                         expect_equal(f2$J, f3$J)
                         # Jacobian adjusted to concentration
                         expect_equal(f2$Jg, f3$Jg)
                         # Alast and gradients are in amounts
                         expect_equal(f2$Alast, f3$Alast)
                       })
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             f(d,V = 40, CL = 18, V2 = 297, Q = 10, DOSE=100)
           })

    ## Two compartment IV infusion

    f0i <- function(t, CL=25, V=20, V2 = 297, Q = 10, DOSE=100, tinf=1,
                    alastNV=rep(0, 10), rateNV=DOSE/tinf) {
      p1 <- CL
      v1 <- V
      p2 <- Q
      p3 <- V2
      p4 <- 0
      p5 <- 0
      ka <- 0
      deriv <- TRUE
      oral0 <- 0
      trans <- 1
      ncmt <- 2
      if (all(rateNV == 0)) {
        # Now a simple eliminiation; Handles the case where the infusion has already past
        dt <- t
      } else if (t < tinf) {
        # infusion not complete
        dt <- t
      } else {
        #Infusion completes during the time
        v <- .Call(`_rxode2_linCmtModelDouble`, tinf, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
        if (t == tinf) return(v)
        # Infusion is now complete
        alastNV <- v$Alast
        rateNV <- 0
        dt <- t - tinf
      }
      .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
    }

    f <-  function(dt, CL=25, V=20, V2 = 297, Q = 10, DOSE=100, tinf=1) {
      test_that(paste0("two compartment infusion t=", dt, ";CL=", CL,
                       "; V2 = ", V2, "; Q = ", Q,
                       "; V=", V, "; tinf=", tinf, "; DOSE=", DOSE), {
                         t0 <- dt/2
                         # See if more care is needed for IV comparison
                         if (dt < tinf) {
                           f1 <- f0i(t0, CL=CL, V=V, V2=V2, Q=Q, DOSE=DOSE, tinf=tinf)
                           expect_equal(pmxTools::calc_sd_2cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2,
                             t=t0, dose=DOSE, tinf=tinf),
                             f1$val)
                           #
                           f2 <- f0i(t0, CL=CL, V=V, V2=V2, Q=Q,
                                     DOSE=DOSE, tinf=tinf, alastNV=f1$Alast)
                           #
                           f3 <- f0i(t0*2, CL=CL, V=V, Q=Q, V2=V2, DOSE=DOSE)
                           expect_equal(pmxTools::calc_sd_2cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2,
                             t=t0*2, dose=DOSE, tinf=tinf),
                             f3$val)
                           # Unadjusted Jacobian
                           expect_equal(f2$J, f3$J)
                           # Jacobian adjusted to concentration
                           expect_equal(f2$Jg, f3$Jg)
                           # Alast and gradients are in amounts
                           expect_equal(f2$Alast, f3$Alast)
                           # Value is in concentration
                           expect_equal(f2$val, f3$val)
                         } else if (t0 > tinf) {
                           rateNV <- DOSE/tinf
                           f1 <- f0i(t0, CL=CL, V=V, V2=V2, Q=Q,
                                     DOSE=DOSE, tinf=tinf, rateNV=rateNV)
                           expect_equal(pmxTools::calc_sd_2cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, t=t0, dose=DOSE, tinf=tinf),
                             f1$val)
                           #
                           f2 <- f0i(t0, CL=CL, V=V, V2=V2, Q=Q,
                                     DOSE=DOSE, tinf=tinf, rateNV=0,
                                     alastNV=f1$Alast)
                           expect_equal(pmxTools::calc_sd_2cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, t=t0*2, dose=DOSE, tinf=tinf),
                             f2$val)
                           f3 <- f0i(t0*2, CL=CL, V=V, Q=Q, V2=V2,
                                     DOSE=DOSE, tinf=tinf)
                           expect_equal(pmxTools::calc_sd_2cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, t=t0*2, dose=DOSE, tinf=tinf),
                             f3$val)
                           # Unadjusted Jacobian
                           expect_equal(f2$J, f3$J)
                           # Jacobian adjusted to concentration
                           expect_equal(f2$Jg, f3$Jg)
                           # Alast and gradients are in amounts
                           expect_equal(f2$Alast, f3$Alast)
                           # Value is in concentration
                           expect_equal(f2$val, f3$val)
                         } else {
                           rateNV <- DOSE/tinf
                           f1 <- f0i(t0, CL=CL, V=V, Q=Q, V2=V2,
                                     DOSE=DOSE, tinf=tinf, rateNV=rateNV)
                           expect_equal(pmxTools::calc_sd_2cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, t=t0, dose=DOSE, tinf=tinf),
                             f1$val)
                           f3 <- f0i(t0*2, CL=CL, V=V, Q=Q, V2=V2,
                                     DOSE=DOSE, tinf=tinf)
                           expect_equal(pmxTools::calc_sd_2cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, t=t0*2, dose=DOSE, tinf=tinf),
                             f3$val)
                           tinf2 <- tinf-t0
                           f2 <- f0i(t0, CL=CL, V=V, Q=Q, V2=V2,
                                     DOSE=DOSE, tinf=tinf2, rateNV=rateNV,
                                     alastNV=f1$Alast)
                           expect_equal(pmxTools::calc_sd_2cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, t=t0*2, dose=DOSE, tinf=tinf),
                             f3$val)
                           # Unadjusted Jacobian
                           expect_equal(f2$J, f3$J)
                           # Jacobian adjusted to concentration
                           expect_equal(f2$Jg, f3$Jg)
                           # Alast and gradients are in amounts
                           expect_equal(f2$Alast, f3$Alast)
                           # Value is in concentration
                           expect_equal(f2$val, f3$val)
                         }
                       })
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             f(d, CL=25, V=20, tinf=1, DOSE=100)
           })

    ## Three compartment model bolus
    f0 <- function(dt, V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400,
                   DOSE=100,
                   alastNV=c(DOSE, rep(0, 20))) {
      p1 <- CL
      v1 <- V
      p2 <- Q
      p3 <- V2
      p4 <- Q2
      p5 <- V3
      ka <- 0
      alastNV <- alastNV
      rateNV <- c(0, 0)
      oral0 <- 0
      trans <- 1
      ncmt <- 3
      deriv <- TRUE
      .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
    }

    f <-  function(dt, V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, DOSE=100) {
      test_that(paste0("two compartment bolus t=", dt, ";CL=", CL,
                       "; V=", V, "; Q=", Q, "; V2=", V2, "; Q2=", Q2,
                       "; V3=", V3,
                       "; DOSE=", DOSE), {
                         t0 <- dt/2
                         f1 <- f0(t0, CL=CL, V=V, Q=Q, V2=V2, Q2=Q2, V3=V3, DOSE=DOSE)
                         expect_equal(pmxTools::calc_sd_3cmt_linear_bolus(
                           CL=CL, V=V, V2=V2, Q=Q, V3=V3, Q3=Q2,
                           t=t0, dose=DOSE),
                           f1$val)
                         #
                         f2 <- f0(t0, CL=CL, V=V, Q=Q, V2=V2, Q2=Q2, V3=V3,
                                  DOSE=DOSE, alastNV=f1$Alast)
                         #
                         f3 <- f0(t0*2, CL=CL, V=V, Q=Q, V2=V2, Q2=Q2, V3=V3,
                                  DOSE=DOSE)
                         #
                         expect_equal(
                           pmxTools::calc_sd_3cmt_linear_bolus(
                             CL=CL, V=V, V2=V2, Q=Q, V3=V3, Q3=Q2,
                             t=dt, dose=DOSE),
                           f3$val)
                         # Value is in concentration
                         expect_equal(f2$val, f3$val)
                         # Unadjusted Jacobian
                         expect_equal(f2$J, f3$J)
                         # Jacobian adjusted to concentration
                         expect_equal(f2$Jg, f3$Jg)
                         # Alast and gradients are in amounts
                         expect_equal(f2$Alast, f3$Alast)
                       })
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             f(d,V = 40, CL = 18, V2 = 297, Q = 10, DOSE=100)
           })


    ## Three compartment model oral
    f0 <- function(dt, V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400,
                   ka=2,
                   DOSE=100,
                   alastNV=c(DOSE, rep(0, 25))) {
      p1 <- CL
      v1 <- V
      p2 <- Q
      p3 <- V2
      p4 <- Q2
      p5 <- V3
      ka <- ka
      alastNV <- alastNV
      rateNV <- c(0, 0)
      oral0 <- TRUE
      trans <- 1
      ncmt <- 3
      deriv <- TRUE
      .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
    }

    f <-  function(dt, V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400,
                   ka=2, DOSE=100) {
      test_that(paste0("three compartment oral t=", dt, ";CL=", CL,
                       "; V=", V, "; Q=", Q, "; V2=", V2, "; ka=", ka,
                       "; Q2=", Q2, "; V3=", V3, "; ka=", ka,
                       "; DOSE=", DOSE), {
                         t0 <- dt/2
                         f1 <- f0(t0, CL=CL, V=V, Q=Q, V2=V2, Q2=Q2, V3=V3, ka=ka, DOSE=DOSE)
                         expect_equal(pmxTools::calc_sd_3cmt_linear_oral_1(
                           CL=CL, V=V,
                           V2=V2, Q=Q, Q3=Q2, V3=V3, ka=ka,
                           t=t0, dose=DOSE),
                           f1$val)
                         #
                         f2 <- f0(t0, CL=CL, V=V, Q=Q, V2=V2, Q2=Q2, V3=V3, ka=ka,
                                  DOSE=DOSE, alastNV=f1$Alast)
                         #
                         f3 <- f0(t0*2, CL=CL, V=V, Q=Q, V2=V2, Q2=Q2, V3=V3, ka=ka,
                                  DOSE=DOSE)
                         #
                         expect_equal(
                           pmxTools::calc_sd_3cmt_linear_oral_1(
                             CL=CL, V=V, V2=V2, Q=Q, V3=V3, Q3=Q2, ka=ka,
                             t=dt, dose=DOSE),
                           f3$val)
                         # Value is in concentration
                         expect_equal(f2$val, f3$val)
                         # Unadjusted Jacobian
                         expect_equal(f2$J, f3$J)
                         # Jacobian adjusted to concentration
                         expect_equal(f2$Jg, f3$Jg)
                         # Alast and gradients are in amounts
                         expect_equal(f2$Alast, f3$Alast)
                       })
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             f(d,V = 40, CL = 18, V2 = 297, Q = 10, DOSE=100)
           })


    ## Three compartment IV infusion
    f0i <- function(t, V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400,
                    DOSE=100, tinf=1,
                    alastNV=rep(0, 21),
                    rateNV=DOSE/tinf) {
      p1 <- CL
      v1 <- V
      p2 <- Q
      p3 <- V2
      p4 <- Q2
      p5 <- V3
      ka <- 0
      deriv <- TRUE
      oral0 <- 0
      trans <- 1
      ncmt <- 3
      if (all(rateNV == 0)) {
        # Now a simple eliminiation; Handles the case where the infusion has already past
        dt <- t
      } else if (t < tinf) {
        # infusion not complete
        dt <- t
      } else {
        #Infusion completes during the time
        v <- .Call(`_rxode2_linCmtModelDouble`, tinf, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
        if (t == tinf) return(v)
        # Infusion is now complete
        alastNV <- v$Alast
        rateNV <- 0
        dt <- t - tinf
      }
      .Call(`_rxode2_linCmtModelDouble`, dt, p1, v1, p2, p3, p4, p5, ka, alastNV, rateNV, ncmt, oral0, trans, deriv, 0L, 0, 0, 0, 0, 127, 3L, 0.001)
    }

    f <-  function(dt, V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400,
                   DOSE=100, tinf=1) {
      test_that(paste0("three compartment infusion t=", dt, ";CL=", CL,
                       "; V2 = ", V2, "; Q = ", Q,
                       "; V3 = ", V3, "; Q2 = ", Q2,
                       "; V=", V, "; tinf=", tinf, "; DOSE=", DOSE), {
                         t0 <- dt/2
                         # See if more care is needed for IV comparison
                         if (dt < tinf) {
                           f1 <- f0i(t0, CL=CL, V=V, V2=V2, Q=Q, V3=V3, Q2=Q2,
                                     DOSE=DOSE, tinf=tinf)
                           expect_equal(pmxTools::calc_sd_3cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, Q3=Q2, V3=V3,
                             t=t0, dose=DOSE, tinf=tinf),
                             f1$val)
                           #
                           f2 <- f0i(t0, CL=CL, V=V, V2=V2, Q=Q, V3=V3, Q2=Q2,
                                     DOSE=DOSE, tinf=tinf, alastNV=f1$Alast)
                           #
                           f3 <- f0i(t0*2, CL=CL, V=V, Q=Q, V2=V2, V3=V3, Q2=Q2,
                                     DOSE=DOSE)
                           expect_equal(pmxTools::calc_sd_3cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, Q3=Q2, V3=V3,
                             t=t0*2, dose=DOSE, tinf=tinf),
                             f3$val)
                           # Unadjusted Jacobian
                           expect_equal(f2$J, f3$J)
                           # Jacobian adjusted to concentration
                           expect_equal(f2$Jg, f3$Jg)
                           # Alast and gradients are in amounts
                           expect_equal(f2$Alast, f3$Alast)
                           # Value is in concentration
                           expect_equal(f2$val, f3$val)
                         } else if (t0 > tinf) {
                           rateNV <- DOSE/tinf
                           f1 <- f0i(t0, CL=CL, V=V, V2=V2, Q=Q, V3=V3, Q2=Q2,
                                     DOSE=DOSE, tinf=tinf, rateNV=rateNV)
                           expect_equal(pmxTools::calc_sd_3cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, Q3=Q2, V3=V3,
                             t=t0, dose=DOSE, tinf=tinf),
                             f1$val)
                           f2 <- f0i(t0, CL=CL, V=V, V2=V2, Q=Q, V3=V3, Q2=Q2,
                                     DOSE=DOSE, tinf=tinf, rateNV=0,
                                     alastNV=f1$Alast)
                           expect_equal(pmxTools::calc_sd_3cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, Q3=Q2, V3=V3,
                             t=t0*2, dose=DOSE, tinf=tinf),
                             f2$val)
                           f3 <- f0i(t0*2, CL=CL, V=V, Q=Q, V2=V2, Q2=Q2, V3=V3,
                                     DOSE=DOSE, tinf=tinf)
                           expect_equal(pmxTools::calc_sd_3cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, Q3=Q2, V3=V3,
                             t=t0*2, dose=DOSE, tinf=tinf),
                             f3$val)
                           # Unadjusted Jacobian
                           expect_equal(f2$J, f3$J)
                           # Jacobian adjusted to concentration
                           expect_equal(f2$Jg, f3$Jg)
                           # Alast and gradients are in amounts
                           expect_equal(f2$Alast, f3$Alast)
                           # Value is in concentration
                           expect_equal(f2$val, f3$val)
                         } else {
                           rateNV <- DOSE/tinf
                           f1 <- f0i(t0, CL=CL, V=V, Q=Q, V2=V2, V3=V3, Q2=Q2,
                                     DOSE=DOSE, tinf=tinf, rateNV=rateNV)
                           expect_equal(pmxTools::calc_sd_3cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, Q3=Q2, V3=V3,
                             t=t0, dose=DOSE, tinf=tinf),
                             f1$val)
                           f3 <- f0i(t0*2, CL=CL, V=V, Q=Q, V2=V2, Q2=Q2, V3=V3,
                                     DOSE=DOSE, tinf=tinf)
                           expect_equal(pmxTools::calc_sd_3cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, Q3=Q2, V3=V3,
                             t=t0*2, dose=DOSE, tinf=tinf),
                             f3$val)
                           tinf2 <- tinf-t0
                           f2 <- f0i(t0, CL=CL, V=V, Q=Q, V2=V2, Q2=Q2, V3=V3,
                                     DOSE=DOSE, tinf=tinf2, rateNV=rateNV,
                                     alastNV=f1$Alast)
                           expect_equal(pmxTools::calc_sd_3cmt_linear_infusion(
                             CL=CL, V=V, Q=Q, V2=V2, Q3=Q2, V3=V3,
                             t=t0*2, dose=DOSE, tinf=tinf),
                             f3$val)
                           # Unadjusted Jacobian
                           expect_equal(f2$J, f3$J)
                           # Jacobian adjusted to concentration
                           expect_equal(f2$Jg, f3$Jg)
                           # Alast and gradients are in amounts
                           expect_equal(f2$Alast, f3$Alast)
                           # Value is in concentration
                           expect_equal(f2$val, f3$val)
                         }
                       })
    }

    lapply(seq(.1, 10, by=0.1),
           function(d) {
             f(d, CL=25, V=20, tinf=1, DOSE=100)
           })
  }

  testVal <- function(rx, val) {
    state <- rx$state
    state <- setNames(seq_along(state), state)

    expect_equal(.getCmtNum(rx),
                 state[val])
  }

  testInf <- function(rx, val) {
    state <- rx$state
    state <- setNames(seq_along(state), state)
    .num <- .getCmtNum(rx)
    expect_equal(setNames(cmtSupportsInfusion_(.num, rxModelVars(rx)),
                          names(.getCmtNum(rx))),
                 val)
  }

  testOff <- function(rx, val) {
    state <- rx$state
    state <- setNames(seq_along(state), state)
    .num <- .getCmtNum(rx)
    expect_equal(setNames(cmtSupportsOff_(.num, rxModelVars(rx)),
                          names(.num)),
                 val)
  }

  test_that("rxode2 parsing of linCmt() 1 compartment with ka", {

    rx <- function() {
      ini({
        popCl <- 1
        popV <- 20
        bsvCl ~ 0.1
        bsvV ~ 0.1
        popKeo <- 1.4
        bsvKeo ~ 0.1
        popKa <- 1
      })
      model({
        cl ~ popCl * exp(bsvCl)
        v ~ popV * exp(bsvV)
        ka ~ popKa
        keo ~ popKeo * exp(bsvKeo)
        popLagCentral <- 0
        popRateCentral <- 0
        popDurCentral <- 0
        bsvLagCentral <- 0
        bsvRateCentral <- 0
        bsvDurCentral <- 0
        alag(central) <- popLagCentral * exp(bsvLagCentral)
        rate(central) <- popRateCentral * exp(bsvRateCentral)
        dur(central) <- popDurCentral * exp(bsvDurCentral)
        cp <- linCmt()
        d/dt(ce) = keo*(cp-ce)
        effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
      })
    }


    rx <- rxode2({
      popCl <- 1
      popV <- 20
      bsvCl <-0
      bsvV <- 0
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      popKa <- 1
      ##
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      ka ~ popKa
      keo ~ popKeo * exp(bsvKeo)
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
    })

    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 0L, numLin = 2L, depotLin = 1L))

    testVal(rx, c("depot", "central", "ce"))

    testInf(rx, c(depot=TRUE, central=TRUE, ce=TRUE))

    testOff(rx, c(depot=FALSE, central=FALSE, ce=TRUE))

    expect_equal(rx$state, c("ce", "depot", "central"))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtA[(]", rxModelVars(rx)$model["normModel"]))

    rx <- rxode2({
      popCl <- 1
      popV <- 20
      popKa <- 1
      bsvCl <-0
      bsvV <- 0
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      ##
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      ka ~ popKa
      keo ~ popKeo * exp(bsvKeo)
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
      b=central
      c=peripheral1
    }, linCmtSens="linCmtB")

    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 4L, numLin = 2L, depotLin = 1L))

    testVal(rx, c("depot", "central", "ce",
                  "rx__sens_central_BY_p1", "rx__sens_central_BY_v1",
                  "rx__sens_central_BY_ka", "rx__sens_depot_BY_ka"))

    testInf(rx, c(depot=TRUE, central=TRUE, ce=TRUE,
                  rx__sens_central_BY_p1=FALSE, rx__sens_central_BY_v1=FALSE,
                  rx__sens_central_BY_ka=FALSE, rx__sens_depot_BY_ka=FALSE))

    testOff(rx, c(depot=FALSE, central=FALSE, ce=TRUE,
                  rx__sens_central_BY_p1=FALSE, rx__sens_central_BY_v1=FALSE,
                  rx__sens_central_BY_ka=FALSE, rx__sens_depot_BY_ka=FALSE))



    expect_equal(rx$state, c("ce",
                             "depot", "central",
                             "rx__sens_central_BY_p1", "rx__sens_central_BY_v1",
                             "rx__sens_central_BY_ka",
                             "rx__sens_depot_BY_ka"))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtB[(]", rxModelVars(rx)$model["normModel"]))
  })


  test_that("rxode2 parsing of linCmt() 1 compartment without ka", {

    rx <- rxode2({
      popCl <- 1
      popV <- 20
      bsvCl <-0
      bsvV <- 0
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      ##
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      keo ~ popKeo * exp(bsvKeo)
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
    })

    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 0L, numLin = 1L, depotLin = 0L))

    testVal(rx, c("central", "ce"))

    testInf(rx, c(central=TRUE, ce=TRUE))

    testOff(rx, c(central=FALSE, ce=TRUE))

    expect_equal(rx$state, c("ce", "central"))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtA[(]", rxModelVars(rx)$model["normModel"]))

    rx <- rxode2({
      popCl <- 1
      popV <- 20
      bsvCl <-0
      bsvV <- 0
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      ##
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      keo ~ popKeo * exp(bsvKeo)
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
      b=central
      c=peripheral1
    }, linCmtSens="linCmtB")

    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 2L, numLin = 1L, depotLin = 0L))

    testVal(rx, c("central","ce",
                  "rx__sens_central_BY_p1", "rx__sens_central_BY_v1"))

    testInf(rx, c(central=TRUE, ce=TRUE,
                  rx__sens_central_BY_p1=FALSE, rx__sens_central_BY_v1=FALSE))

    testOff(rx, c(central=FALSE, ce=TRUE,
                  rx__sens_central_BY_p1=FALSE, rx__sens_central_BY_v1=FALSE))


    expect_equal(rx$state, c("ce",
                             "central",
                             "rx__sens_central_BY_p1", "rx__sens_central_BY_v1"))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtB[(]", rxModelVars(rx)$model["normModel"]))

  })



  test_that("rxode2 parsing of linCmt() 2 compartment without ka", {

    rx <- rxode2({
      popCl <- 1
      popV <- 20
      popVp <- 10
      popQ <- 2
      bsvCl <-0
      bsvV <- 0
      bsvVp <- 0
      bsvQ <-0
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      ##
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      q ~ popQ * exp(bsvQ)
      vp ~ popVp * exp(bsvVp)
      keo ~ popKeo * exp(bsvKeo)
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
    })

    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 0L, numLin = 2L, depotLin = 0L))

    testVal(rx, c("central", "ce", "peripheral1"))

    testInf(rx, c("central"=TRUE, "ce"=TRUE, "peripheral1"=FALSE))

    testOff(rx, c("central"=FALSE, "ce"=TRUE, "peripheral1"=FALSE))

    expect_equal(rx$state, c("ce", "central", "peripheral1"))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtA[(]", rxModelVars(rx)$model["normModel"]))

    rx <- rxode2({
      popCl <- 1
      popV <- 20
      popVp <- 10
      popQ <- 2
      bsvCl <-0
      bsvV <- 0
      bsvVp <- 0
      bsvQ <-0
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      ##
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      q ~ popQ * exp(bsvQ)
      vp ~ popVp * exp(bsvVp)
      keo ~ popKeo * exp(bsvKeo)
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
      b=central
      c=peripheral1
    }, linCmtSens="linCmtB")

    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 8L, numLin = 2L, depotLin = 0L))

    testVal(rx, c("central", "ce", "peripheral1",
                  "rx__sens_central_BY_p1", "rx__sens_central_BY_v1",
                  "rx__sens_central_BY_p2", "rx__sens_central_BY_p3",
                  #
                  "rx__sens_peripheral1_BY_p1", "rx__sens_peripheral1_BY_v1",
                  "rx__sens_peripheral1_BY_p2", "rx__sens_peripheral1_BY_p3"))


    testInf(rx, c("central"=TRUE, "ce"=TRUE, "peripheral1"=FALSE,
                  "rx__sens_central_BY_p1"=FALSE, "rx__sens_central_BY_v1"=FALSE,
                  "rx__sens_central_BY_p2"=FALSE, "rx__sens_central_BY_p3"=FALSE,
                  #
                  "rx__sens_peripheral1_BY_p1"=FALSE, "rx__sens_peripheral1_BY_v1"=FALSE,
                  "rx__sens_peripheral1_BY_p2"=FALSE, "rx__sens_peripheral1_BY_p3"=FALSE))

    testOff(rx, c("central"=FALSE, "ce"=TRUE, "peripheral1"=FALSE,
                  "rx__sens_central_BY_p1"=FALSE, "rx__sens_central_BY_v1"=FALSE,
                  "rx__sens_central_BY_p2"=FALSE, "rx__sens_central_BY_p3"=FALSE,
                  #
                  "rx__sens_peripheral1_BY_p1"=FALSE, "rx__sens_peripheral1_BY_v1"=FALSE,
                  "rx__sens_peripheral1_BY_p2"=FALSE, "rx__sens_peripheral1_BY_p3"=FALSE))


    expect_equal(rx$state, c("ce", "central", "peripheral1",
                             "rx__sens_central_BY_p1", "rx__sens_central_BY_v1",
                             "rx__sens_central_BY_p2", "rx__sens_central_BY_p3",
                             #
                             "rx__sens_peripheral1_BY_p1", "rx__sens_peripheral1_BY_v1",
                             "rx__sens_peripheral1_BY_p2", "rx__sens_peripheral1_BY_p3"))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtB[(]", rxModelVars(rx)$model["normModel"]))
  })


  test_that("rxode2 parsing of linCmt() 2 compartment with ka", {

    rx <- rxode2({
      popCl <- 1
      popV <- 20
      popKa <- 1
      popVp <- 10
      popQ <- 2
      bsvCl <-0
      bsvV <- 0
      bsvKa <-0
      bsvVp <- 0
      bsvQ <-0
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      ##
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      ka ~ popKa * exp(bsvKa)
      q ~ popQ * exp(bsvQ)
      vp ~ popVp * exp(bsvVp)
      keo ~ popKeo * exp(bsvKeo)
      popLagDepot <- 0
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagDepot <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(depot) <- popLagDepot * exp(bsvLagDepot)
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
    })

    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 0L, numLin = 3L, depotLin = 1L))

    testVal(rx, c("depot", "central", "ce", "peripheral1"))

    testInf(rx, c("depot"=TRUE, "central"=TRUE, "ce"=TRUE, "peripheral1"=FALSE))

    testOff(rx, c("depot"=FALSE, "central"=FALSE, "ce"=TRUE, "peripheral1"=FALSE))

    expect_equal(rx$state, c("ce", "depot", "central", "peripheral1"))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtA[(]", rxModelVars(rx)$model["normModel"]))

    rx <- rxode2({
      popCl <- 1
      popV <- 20
      popKa <- 1
      popVp <- 10
      popQ <- 2
      bsvCl <-0
      bsvV <- 0
      bsvKa <-0
      bsvVp <- 0
      bsvQ <-0
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      ##
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      ka ~ popKa * exp(bsvKa)
      q ~ popQ * exp(bsvQ)
      vp ~ popVp * exp(bsvVp)
      keo ~ popKeo * exp(bsvKeo)
      popLagDepot <- 0
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagDepot <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(depot) <- popLagDepot * exp(bsvLagDepot)
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
      a=depot
      b=central
      c=peripheral1
    }, linCmtSens="linCmtB")

    testVal(rx, c("depot", "central", "ce", "peripheral1",
                  "rx__sens_central_BY_p1", "rx__sens_central_BY_v1",
                  "rx__sens_central_BY_p2", "rx__sens_central_BY_p3",
                  "rx__sens_central_BY_ka",
                  #
                  "rx__sens_peripheral1_BY_p1", "rx__sens_peripheral1_BY_v1",
                  "rx__sens_peripheral1_BY_p2", "rx__sens_peripheral1_BY_p3",
                  "rx__sens_peripheral1_BY_ka",
                  #
                  "rx__sens_depot_BY_ka"))

    testInf(rx, c("depot"=TRUE, "central"=TRUE, "ce"=TRUE, "peripheral1"=FALSE,
                  "rx__sens_central_BY_p1"=FALSE, "rx__sens_central_BY_v1"=FALSE,
                  "rx__sens_central_BY_p2"=FALSE, "rx__sens_central_BY_p3"=FALSE,
                  "rx__sens_central_BY_ka"=FALSE,
                  #
                  "rx__sens_peripheral1_BY_p1"=FALSE, "rx__sens_peripheral1_BY_v1"=FALSE,
                  "rx__sens_peripheral1_BY_p2"=FALSE, "rx__sens_peripheral1_BY_p3"=FALSE,
                  "rx__sens_peripheral1_BY_ka"=FALSE,
                  #
                  "rx__sens_depot_BY_ka"=FALSE))

    testOff(rx, c("depot"=FALSE, "central"=FALSE, "ce"=TRUE, "peripheral1"=FALSE,
                  "rx__sens_central_BY_p1"=FALSE, "rx__sens_central_BY_v1"=FALSE,
                  "rx__sens_central_BY_p2"=FALSE, "rx__sens_central_BY_p3"=FALSE,
                  "rx__sens_central_BY_ka"=FALSE,
                  #
                  "rx__sens_peripheral1_BY_p1"=FALSE, "rx__sens_peripheral1_BY_v1"=FALSE,
                  "rx__sens_peripheral1_BY_p2"=FALSE, "rx__sens_peripheral1_BY_p3"=FALSE,
                  "rx__sens_peripheral1_BY_ka"=FALSE,
                  #
                  "rx__sens_depot_BY_ka"=FALSE))


    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 11L, numLin = 3L, depotLin = 1L))

    expect_equal(rx$state, c("ce", "depot", "central", "peripheral1",
                             "rx__sens_central_BY_p1", "rx__sens_central_BY_v1",
                             "rx__sens_central_BY_p2", "rx__sens_central_BY_p3",
                             "rx__sens_central_BY_ka",
                             #
                             "rx__sens_peripheral1_BY_p1", "rx__sens_peripheral1_BY_v1",
                             "rx__sens_peripheral1_BY_p2", "rx__sens_peripheral1_BY_p3",
                             "rx__sens_peripheral1_BY_ka",
                             #
                             "rx__sens_depot_BY_ka"))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtB[(]", rxModelVars(rx)$model["normModel"]))
  })


  test_that("rxode2 parsing of linCmt() 3 compartment without ka", {

    rx <- rxode2({
      popCl <- 1
      popV <- 20
      popVp <- 10
      popQ <- 2
      popVp2 <- 1
      popQ2 <- 2
      bsvCl <-0
      bsvV <- 0
      bsvVp <- 0
      bsvQ <-0
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      ##
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      q ~ popQ * exp(bsvQ)
      vp ~ popVp * exp(bsvVp)
      vp2 ~ popVp2
      q2 ~ popQ2
      keo ~ popKeo * exp(bsvKeo)
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
    })

    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 0L, numLin = 3L, depotLin = 0L))

    testVal(rx, c("central", "ce", "peripheral1", "peripheral2"))

    testInf(rx, c("central"=TRUE, "ce"=TRUE, "peripheral1"=FALSE, "peripheral2"=FALSE))

    testOff(rx, c("central"=FALSE, "ce"=TRUE, "peripheral1"=FALSE, "peripheral2"=FALSE))

    expect_equal(rx$state, c("ce", "central", "peripheral1", "peripheral2"))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtA[(]", rxModelVars(rx)$model["normModel"]))

    rx <- rxode2({
      popCl <- 1
      popV <- 20
      popVp <- 10
      popQ <- 2
      bsvCl <-0
      bsvV <- 0
      bsvVp <- 0
      bsvQ <-0
      popVp2 <- 3
      popQ2 <- 4
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      ##
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      q ~ popQ * exp(bsvQ)
      vp ~ popVp * exp(bsvVp)
      vp2 ~ popVp2
      q2 ~ popQ2
      keo ~ popKeo * exp(bsvKeo)
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
      b=central
      c=peripheral1
    }, linCmtSens="linCmtB")

    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 18L, numLin = 3L, depotLin = 0L))

    testVal(rx,c("central", "ce", "peripheral1", "peripheral2",
                 "rx__sens_central_BY_p1", "rx__sens_central_BY_v1",
                 "rx__sens_central_BY_p2", "rx__sens_central_BY_p3",
                 "rx__sens_central_BY_p4", "rx__sens_central_BY_p5",
                 #
                 "rx__sens_peripheral1_BY_p1", "rx__sens_peripheral1_BY_v1",
                 "rx__sens_peripheral1_BY_p2", "rx__sens_peripheral1_BY_p3",
                 "rx__sens_peripheral1_BY_p4", "rx__sens_peripheral1_BY_p5",
                 #
                 "rx__sens_peripheral2_BY_p1", "rx__sens_peripheral2_BY_v1",
                 "rx__sens_peripheral2_BY_p2", "rx__sens_peripheral2_BY_p3",
                 "rx__sens_peripheral2_BY_p4", "rx__sens_peripheral2_BY_p5"))

    testInf(rx,c("central"=TRUE, "ce"=TRUE, "peripheral1"=FALSE, "peripheral2"=FALSE,
                 "rx__sens_central_BY_p1"=FALSE, "rx__sens_central_BY_v1"=FALSE,
                 "rx__sens_central_BY_p2"=FALSE, "rx__sens_central_BY_p3"=FALSE,
                 "rx__sens_central_BY_p4"=FALSE, "rx__sens_central_BY_p5"=FALSE,
                 #
                 "rx__sens_peripheral1_BY_p1"=FALSE, "rx__sens_peripheral1_BY_v1"=FALSE,
                 "rx__sens_peripheral1_BY_p2"=FALSE, "rx__sens_peripheral1_BY_p3"=FALSE,
                 "rx__sens_peripheral1_BY_p4"=FALSE, "rx__sens_peripheral1_BY_p5"=FALSE,
                 #
                 "rx__sens_peripheral2_BY_p1"=FALSE, "rx__sens_peripheral2_BY_v1"=FALSE,
                 "rx__sens_peripheral2_BY_p2"=FALSE, "rx__sens_peripheral2_BY_p3"=FALSE,
                 "rx__sens_peripheral2_BY_p4"=FALSE, "rx__sens_peripheral2_BY_p5"=FALSE))

    testOff(rx,c("central"=FALSE, "ce"=TRUE, "peripheral1"=FALSE, "peripheral2"=FALSE,
                 "rx__sens_central_BY_p1"=FALSE, "rx__sens_central_BY_v1"=FALSE,
                 "rx__sens_central_BY_p2"=FALSE, "rx__sens_central_BY_p3"=FALSE,
                 "rx__sens_central_BY_p4"=FALSE, "rx__sens_central_BY_p5"=FALSE,
                 #
                 "rx__sens_peripheral1_BY_p1"=FALSE, "rx__sens_peripheral1_BY_v1"=FALSE,
                 "rx__sens_peripheral1_BY_p2"=FALSE, "rx__sens_peripheral1_BY_p3"=FALSE,
                 "rx__sens_peripheral1_BY_p4"=FALSE, "rx__sens_peripheral1_BY_p5"=FALSE,
                 #
                 "rx__sens_peripheral2_BY_p1"=FALSE, "rx__sens_peripheral2_BY_v1"=FALSE,
                 "rx__sens_peripheral2_BY_p2"=FALSE, "rx__sens_peripheral2_BY_p3"=FALSE,
                 "rx__sens_peripheral2_BY_p4"=FALSE, "rx__sens_peripheral2_BY_p5"=FALSE))


    expect_equal(rx$state, c("ce", "central", "peripheral1", "peripheral2",
                             "rx__sens_central_BY_p1", "rx__sens_central_BY_v1",
                             "rx__sens_central_BY_p2", "rx__sens_central_BY_p3",
                             "rx__sens_central_BY_p4", "rx__sens_central_BY_p5",
                             #
                             "rx__sens_peripheral1_BY_p1", "rx__sens_peripheral1_BY_v1",
                             "rx__sens_peripheral1_BY_p2", "rx__sens_peripheral1_BY_p3",
                             "rx__sens_peripheral1_BY_p4", "rx__sens_peripheral1_BY_p5",
                             #
                             "rx__sens_peripheral2_BY_p1", "rx__sens_peripheral2_BY_v1",
                             "rx__sens_peripheral2_BY_p2", "rx__sens_peripheral2_BY_p3",
                             "rx__sens_peripheral2_BY_p4", "rx__sens_peripheral2_BY_p5"
                             ))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtB[(]", rxModelVars(rx)$model["normModel"]))

  })


  test_that("rxode2 parsing of linCmt() 3 compartment with ka", {

    rx <- rxode2({
      popCl <- 1
      popKa <- 4
      popV <- 20
      popVp <- 10
      popQ <- 2
      popVp2 <- 1
      popQ2 <- 2
      bsvCl <-0
      bsvV <- 0
      bsvVp <- 0
      bsvQ <-0
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      ##
      ka ~ popKa
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      q ~ popQ * exp(bsvQ)
      vp ~ popVp * exp(bsvVp)
      vp2 ~ popVp2
      q2 ~ popQ2
      keo ~ popKeo * exp(bsvKeo)
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
    })

    etTrans(et(amt=3,cmt=1) |> et(1:10), rx)

    etTrans(et(amt=3) |> et(1:10), rx)

    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 0L, numLin = 4L, depotLin = 1L))

    testVal(rx,
            c("depot", "central", "ce", "peripheral1", "peripheral2"))

    testInf(rx,
            c("depot"=TRUE, "central"=TRUE, "ce"=TRUE,
              "peripheral1"=FALSE, "peripheral2"=FALSE))

    testOff(rx,
            c("depot"=FALSE, "central"=FALSE, "ce"=TRUE,
              "peripheral1"=FALSE, "peripheral2"=FALSE))


    expect_equal(rx$state, c("ce", "depot", "central", "peripheral1", "peripheral2"))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtA[(]", rxModelVars(rx)$model["normModel"]))

    rx <- rxode2({
      popCl <- 1
      popKa <- 3
      popV <- 20
      popVp <- 10
      popQ <- 2
      bsvCl <-0
      bsvV <- 0
      bsvVp <- 0
      bsvQ <-0
      popVp2 <- 3
      popQ2 <- 4
      popKeo <- 1.4
      bsvKeo <- 0
      popE0 <- 0
      popEmax <- 1
      popEC50 <- 5
      popGamma <- 1
      bsvE0 <- 0
      bsvEmax <- 0
      bsvEC50 <- 0
      ##
      ka ~ popKa
      cl ~ popCl * exp(bsvCl)
      v ~ popV * exp(bsvV)
      q ~ popQ * exp(bsvQ)
      vp ~ popVp * exp(bsvVp)
      vp2 ~ popVp2
      q2 ~ popQ2
      keo ~ popKeo * exp(bsvKeo)
      popLagCentral <- 0
      popRateCentral <- 0
      popDurCentral <- 0
      bsvLagCentral <- 0
      bsvRateCentral <- 0
      bsvDurCentral <- 0
      alag(central) <- popLagCentral * exp(bsvLagCentral)
      rate(central) <- popRateCentral * exp(bsvRateCentral)
      dur(central) <- popDurCentral * exp(bsvDurCentral)
      cp <- linCmt()
      d/dt(ce) = keo*(cp-ce)
      effect = E0 - Emax*(Ce^gamma)/((Ce^gamma)+(Ec50^gamma));
      b=central
      c=peripheral1
    }, linCmtSens="linCmtB")

    expect_equal(getLinInfo_(rxModelVars(rx)),
                 c(numLinSens = 22L, numLin = 4L, depotLin = 1L))

    testVal(rx,
            c("depot", "central", "ce", "peripheral1", "peripheral2",
              "rx__sens_central_BY_p1", "rx__sens_central_BY_v1",
              "rx__sens_central_BY_p2", "rx__sens_central_BY_p3",
              "rx__sens_central_BY_p4", "rx__sens_central_BY_p5",
              "rx__sens_central_BY_ka",
              #
              "rx__sens_peripheral1_BY_p1", "rx__sens_peripheral1_BY_v1",
              "rx__sens_peripheral1_BY_p2", "rx__sens_peripheral1_BY_p3",
              "rx__sens_peripheral1_BY_p4", "rx__sens_peripheral1_BY_p5",
              "rx__sens_peripheral1_BY_ka",
              #
              "rx__sens_peripheral2_BY_p1", "rx__sens_peripheral2_BY_v1",
              "rx__sens_peripheral2_BY_p2", "rx__sens_peripheral2_BY_p3",
              "rx__sens_peripheral2_BY_p4", "rx__sens_peripheral2_BY_p5",
              "rx__sens_peripheral2_BY_ka",
              #
              "rx__sens_depot_BY_ka"))

    testInf(rx,
            c("depot"=TRUE, "central"=TRUE,
              "ce"=TRUE, "peripheral1"=FALSE, "peripheral2"=FALSE,
              "rx__sens_central_BY_p1"=FALSE, "rx__sens_central_BY_v1"=FALSE,
              "rx__sens_central_BY_p2"=FALSE, "rx__sens_central_BY_p3"=FALSE,
              "rx__sens_central_BY_p4"=FALSE, "rx__sens_central_BY_p5"=FALSE,
              "rx__sens_central_BY_ka"=FALSE,
              #
              "rx__sens_peripheral1_BY_p1"=FALSE, "rx__sens_peripheral1_BY_v1"=FALSE,
              "rx__sens_peripheral1_BY_p2"=FALSE, "rx__sens_peripheral1_BY_p3"=FALSE,
              "rx__sens_peripheral1_BY_p4"=FALSE, "rx__sens_peripheral1_BY_p5"=FALSE,
              "rx__sens_peripheral1_BY_ka"=FALSE,
              #
              "rx__sens_peripheral2_BY_p1"=FALSE, "rx__sens_peripheral2_BY_v1"=FALSE,
              "rx__sens_peripheral2_BY_p2"=FALSE, "rx__sens_peripheral2_BY_p3"=FALSE,
              "rx__sens_peripheral2_BY_p4"=FALSE, "rx__sens_peripheral2_BY_p5"=FALSE,
              "rx__sens_peripheral2_BY_ka"=FALSE,
              #
              "rx__sens_depot_BY_ka"=FALSE))

    testOff(rx,
            c("depot"=FALSE, "central"=FALSE,
              "ce"=TRUE, "peripheral1"=FALSE, "peripheral2"=FALSE,
              "rx__sens_central_BY_p1"=FALSE, "rx__sens_central_BY_v1"=FALSE,
              "rx__sens_central_BY_p2"=FALSE, "rx__sens_central_BY_p3"=FALSE,
              "rx__sens_central_BY_p4"=FALSE, "rx__sens_central_BY_p5"=FALSE,
              "rx__sens_central_BY_ka"=FALSE,
              #
              "rx__sens_peripheral1_BY_p1"=FALSE, "rx__sens_peripheral1_BY_v1"=FALSE,
              "rx__sens_peripheral1_BY_p2"=FALSE, "rx__sens_peripheral1_BY_p3"=FALSE,
              "rx__sens_peripheral1_BY_p4"=FALSE, "rx__sens_peripheral1_BY_p5"=FALSE,
              "rx__sens_peripheral1_BY_ka"=FALSE,
              #
              "rx__sens_peripheral2_BY_p1"=FALSE, "rx__sens_peripheral2_BY_v1"=FALSE,
              "rx__sens_peripheral2_BY_p2"=FALSE, "rx__sens_peripheral2_BY_p3"=FALSE,
              "rx__sens_peripheral2_BY_p4"=FALSE, "rx__sens_peripheral2_BY_p5"=FALSE,
              "rx__sens_peripheral2_BY_ka"=FALSE,
              #
              "rx__sens_depot_BY_ka"=FALSE))



    expect_equal(rx$state, c("ce", "depot", "central", "peripheral1", "peripheral2",
                             "rx__sens_central_BY_p1", "rx__sens_central_BY_v1",
                             "rx__sens_central_BY_p2", "rx__sens_central_BY_p3",
                             "rx__sens_central_BY_p4", "rx__sens_central_BY_p5",
                             "rx__sens_central_BY_ka",
                             #
                             "rx__sens_peripheral1_BY_p1", "rx__sens_peripheral1_BY_v1",
                             "rx__sens_peripheral1_BY_p2", "rx__sens_peripheral1_BY_p3",
                             "rx__sens_peripheral1_BY_p4", "rx__sens_peripheral1_BY_p5",
                             "rx__sens_peripheral1_BY_ka",
                             #
                             "rx__sens_peripheral2_BY_p1", "rx__sens_peripheral2_BY_v1",
                             "rx__sens_peripheral2_BY_p2", "rx__sens_peripheral2_BY_p3",
                             "rx__sens_peripheral2_BY_p4", "rx__sens_peripheral2_BY_p5",
                             "rx__sens_peripheral2_BY_ka",
                             #
                             "rx__sens_depot_BY_ka"
                             ))

    expect_equal(rx$stateExtra, character(0))

    expect_true(grepl("linCmtB[(]", rxModelVars(rx)$model["normModel"]))

  })


  if (requireNamespace("pmxTools", quietly = TRUE)) {

    test_that("rxode2 solving of 1 compartment steady state bolus", {

      f <- function(CL=6, V=25, DOSE=600, tau=24, deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- 0
        p3 <- 0
        p4 <- 0
        p5 <- 0
        ka <- 0
        if (deriv) {
          alastNV <- c(0, 0, 0)
        } else {
          alastNV <- 0
        }
        rateNV <- 0
        oral0 <- 0
        trans <- 1L
        ncmt <- 1
        type <- 3L # ss bolus
        amt <- DOSE
        tinf <- 1
        dt <- 0
        cmt <- 0L
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans, deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      expect_equal(pmxTools::calc_ss_1cmt_linear_bolus(t=0, CL=6, V=25, dose=600, tau=24),
                   f()$val)

      v <- f(deriv=TRUE)

      expect_equal(v$val,pmxTools::calc_ss_1cmt_linear_bolus(t=0, CL=6, V=25, dose=600, tau=24))

      expect_equal(dim(v$J), 1:2)
      expect_length(v$Jg, 2)
      expect_length(v$Alast, 3)
    })

    test_that("rxode2 solving of 1 compartment steady state infusion", {

      f <- function(CL=2, V=25, DOSE=600, tinf=1, tau=24, deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- 0
        p3 <- 0
        p4 <- 0
        p5 <- 0
        ka <- 0
        if (deriv) {
          alastNV <- c(0, 0, 0)
        } else {
          alastNV <- 0
        }
        rateNV <- DOSE/tinf
        oral0 <- 0
        trans <- 1L
        ncmt <- 1
        type <- 2L # ss infusion
        amt <- DOSE
        dt <- 0
        cmt <- 0L
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      expect_equal(pmxTools::calc_ss_1cmt_linear_infusion(tad=0, CL=2, V=25, dose=600, tinf=1, tau=24),
                   f()$val)

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   pmxTools::calc_ss_1cmt_linear_infusion(tad=0, CL=2, V=25, dose=600, tinf=1, tau=24))

      expect_equal(dim(v$J), 1:2)
      expect_length(v$Jg, 2)
      expect_length(v$Alast, 3)
    })

    test_that("rxode2 solving of 1 compartment steady state 1st order dosing", {

      f <- function(CL=2, V=25, DOSE=600, ka=0.25, tau=24, deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- 0
        p3 <- 0
        p4 <- 0
        p5 <- 0
        ka <- ka
        if (deriv) {
          alastNV <- c(0, 0, 0, 0, 0, 0)
        } else {
          alastNV <- c(0, 0)
        }
        rateNV <- c(0, 0)
        oral0 <- 1L
        trans <- 1L
        ncmt <- 1
        type <- 3L # bolus infusion
        amt <- DOSE
        dt <- 0
        cmt <- 0L
        tinf <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      expect_equal(f()$val,
                   pmxTools::calc_ss_1cmt_linear_oral_1(tad=0, CL=2, V=25, dose=600, ka=0.25, tau=24))


      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   pmxTools::calc_ss_1cmt_linear_oral_1(tad=0, CL=2, V=25, dose=600, ka=0.25, tau=24))

      expect_equal(dim(v$J), 2:3)
      expect_length(v$Jg, 3)
      expect_length(v$Alast, 6)
    })

    test_that("rxode2 solving of 1 compartment steady state for constant infusion", {
      ## Now test the solutions not covered by pmxTools by cross testing
      ## against rxode2 lsoda
      rx <- rxode2({
        CL <- 6
        V <- 25
        kel <- CL/V
        d/dt(central) = -kel*central
        cp <- central/V
      })

      e <- et(amt=0, rate=600, ss=1) |>
        et(time=seq(1:25))

      f <- function(CL=6, V=25, DOSE=600, deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- 0
        p3 <- 0
        p4 <- 0
        p5 <- 0
        ka <- 0
        if (deriv) {
          alastNV <- c(0, 0, 0)
        } else {
          alastNV <- 0
        }
        rateNV <- 600
        oral0 <- 0
        trans <- 1L
        ncmt <- 1
        type <- 1L # ss infinite infusion
        amt <- 0
        tinf <- 0
        dt <- 0
        cmt <- 0L
        tau <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      expect_equal(f()$val,
                   rxSolve(rx, e, addDosing =TRUE)$cp[1])

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   rxSolve(rx, e, addDosing =TRUE)$cp[1])

      expect_equal(dim(v$J), 1:2)
      expect_length(v$Jg, 2)
      expect_length(v$Alast, 3)
    })

    test_that("rxode2 solving of 1 compartment steady state for constant depot infusion", {

      rx <- rxode2({
        CL <- 6
        V <- 25
        ka <- 0.25
        kel <- CL/V
        d/dt(depot) = -ka*depot
        d/dt(central) = -kel*central + ka*depot
        cp <- central/V
      })

      e <- et(amt=0, rate=600, ss=1) |>
        et(time=seq(1:25))

      f <- function(CL=6, V=25, DOSE=600, ka=0.25, rate=c(600, 0), deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- 0
        p3 <- 0
        p4 <- 0
        p5 <- 0
        ka <- ka
        if (deriv) {
          alastNV <- c(0, 0, 0, 0, 0, 0)
        } else {
          alastNV <- c(0, 0)
        }
        rateNV <- rate
        oral0 <- 1L
        trans <- 1L
        ncmt <- 1
        type <- 1L # ss infinite infusion
        amt <- 0
        tinf <- 0
        dt <- 0
        cmt <- 0L
        tau <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      s <- rxSolve(rx, e, addDosing =TRUE)

      expect_equal(f()$val,
                   s$cp[1])

      expect_equal(f()$Alast,
                   c(s$depot[1], s$central[1]))

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1])

      expect_equal(v$Alast[1:2],
                   c(s$depot[1], s$central[1]))

      expect_equal(dim(v$J), 2:3)

      expect_length(v$Jg, 3)
      expect_length(v$Alast, 6)

      # Now infusion in the 2 compartment models to central

      e <- et(amt=0, rate=600, ss=1, cmt=2) |>
        et(time=seq(1:25))

      s <- rxSolve(rx, e, addDosing =TRUE)

      expect_equal(f(rate=c(0, 600))$val,
                   s$cp[1])

      expect_equal(f(rate=c(0, 600))$Alast,
                   c(s$depot[1], s$central[1]))

      v <- f(rate=c(0, 600), deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1])

      expect_equal(v$Alast[1:2],
                   c(s$depot[1], s$central[1]))

      expect_equal(dim(v$J), 2:3)

      expect_length(v$Jg, 3)
      expect_length(v$Alast, 6)

    })

    test_that("rxode2 solving of 1 compartment steady state for constant depot infusion", {

      f <- function(CL=2, V=25, DOSE=600, tinf=1, tau=6, deriv=FALSE, cmt=0L,
                    ka=0.25) {
        p1 <- CL
        v1 <- V
        p2 <- 0
        p3 <- 0
        p4 <- 0
        p5 <- 0
        ka <- ka
        if (deriv) {
          alastNV <- c(0, 0, 0, 0, 0, 0)
        } else {
          alastNV <- c(0, 0)
        }
        if (cmt == 0L) {
          rateNV <- c(DOSE/tinf, 0)
        } else {
          rateNV <- c(0, DOSE/tinf)
        }
        oral0 <- 1L
        trans <- 1L
        ncmt <- 1
        type <- 2L # ss infusion
        amt <- DOSE
        dt <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      rx <- rxode2({
        CL <- 2
        V <- 25
        ka <- 0.25
        kel <- CL/V
        d/dt(depot) = -ka*depot
        d/dt(central) = -kel*central + ka*depot
        cp <- central/V
      })

      e <- et(amt=600, rate=600, ss=1, ii=6, cmt=depot) |>
        et(time=seq(1:25))

      s <- rxSolve(rx, e, addDosing =TRUE)

      expect_equal(f()$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(f()$Alast,
                   c(s$depot[1], s$central[1]),
                   tolerance=1e-3)

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(v$Alast[1:2],
                   c(s$depot[1], s$central[1]),
                   tolerance = 1e-3)

      expect_equal(dim(v$J), 2:3)

      expect_length(v$Jg, 3)
      expect_length(v$Alast, 6)

      e <- et(amt=600, rate=600, ss=1, ii=6, cmt=central) |>
        et(time=seq(1:25))

      s <- rxSolve(rx, e, addDosing =TRUE)

      expect_equal(f(cmt=1L)$val,
                   s$cp[1], tolerance = 1e-3)

      expect_equal(f(cmt=1L)$Alast,
                   c(s$depot[1], s$central[1]),
                   tolerance = 1e-3)

      v <- f(cmt=1L, deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1], tolerance = 1e-3)

      expect_equal(v$Alast[1:2],
                   c(s$depot[1], s$central[1]), tolerance = 1e-3)

      expect_equal(dim(v$J), 2:3)

      expect_length(v$Jg, 3)
      expect_length(v$Alast, 6)

    })

    ####################################################################
    ## Test the 2 compartment steady state solutions
    ####################################################################

    test_that("rxode2 solving of 2 compartment steady state bolus", {

      f <- function(CL=7.5, V=20, V2 = 30, Q = 0.5, DOSE=10, tau=24, deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- Q
        p3 <- V2
        p4 <- 0
        p5 <- 0
        ka <- 0
        if (deriv) {
          alastNV <- rep(0, 10)
        } else {
          alastNV <- c(0, 0)
        }
        rateNV <- 0
        oral0 <- 0
        trans <- 1L
        ncmt <- 2
        type <- 3L # ss bolus
        amt <- DOSE
        tinf <- 1
        dt <- 0
        cmt <- 0L
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans, deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      expect_equal(pmxTools::calc_ss_2cmt_linear_bolus(0, CL = 7.5, V1 = 20, V2 = 30, Q = 0.5,
                                                       dose = 10, tau=24),
                   f()$val)

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   pmxTools::calc_ss_2cmt_linear_bolus(0, CL = 7.5, V1 = 20, V2 = 30, Q = 0.5,
                                                       dose = 10, tau=24))

      expect_equal(dim(v$J), c(2L, 4L))

      expect_length(v$Jg, 4)

      expect_length(v$Alast, 10)

    })


    test_that("rxode2 solving of 2 compartment steady state infusion", {

      f <- function(CL=7.5, V=20, V2=30, Q=0.5, DOSE=10, tinf=1, tau=12, deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- Q
        p3 <- V2
        p4 <- 0
        p5 <- 0
        ka <- 0
        if (deriv) {
          alastNV <- rep(0, 10)
        } else {
          alastNV <- c(0, 0)
        }
        rateNV <- DOSE/tinf
        oral0 <- 0
        trans <- 1L
        ncmt <- 2
        type <- 2L # ss infusion
        amt <- DOSE
        dt <- 0
        cmt <- 0L
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      expect_equal(pmxTools::calc_ss_2cmt_linear_infusion(tad = 0, CL = 7.5, V1 = 20, V2 = 30, Q = 0.5,
                                                          dose = 10, tinf = 1, tau = 12),
                   f()$val)

      v <- f(deriv=TRUE)

      expect_equal(pmxTools::calc_ss_2cmt_linear_infusion(tad = 0, CL = 7.5, V1 = 20, V2 = 30, Q = 0.5,
                                                          dose = 10, tinf = 1, tau = 12),
                   v$val)

      expect_equal(dim(v$J), c(2L, 4L))
      expect_length(v$Jg, 4)
      expect_length(v$Alast, 10)

    })

    test_that("rxode2 solving of 2 compartment steady state 1st order dosing", {

      f <- function(CL=7.5, V=20, V2=30, Q=0.5, DOSE=1000, ka=1, tau=24, deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- Q
        p3 <- V2
        p4 <- 0
        p5 <- 0
        ka <- ka
        if (deriv) {
          alastNV <- rep(0, 14)
        } else {
          alastNV <- c(0, 0, 0)
        }
        rateNV <- c(0, 0)
        oral0 <- 1L
        trans <- 1L
        ncmt <- 2
        type <- 3L # bolus infusion
        amt <- DOSE
        dt <- 0
        cmt <- 0L
        tinf <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      expect_equal(f()$val,
                   pmxTools::calc_ss_2cmt_linear_oral_1(tad=0, CL=7.5, V1=20, V2=30, Q=0.5, dose=1000, ka=1, tau=24))


      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   pmxTools::calc_ss_2cmt_linear_oral_1(tad=0, CL=7.5, V1=20, V2=30, Q=0.5, dose=1000, ka=1, tau=24))

      expect_equal(dim(v$J), c(3L, 5L))
      expect_length(v$Jg, 5)
      expect_length(v$Alast, 14)

    })

    test_that("rxode2 solving of 2 compartment steady state for constant infusion", {
      ## Now test the solutions not covered by pmxTools by cross testing
      ## against rxode2 lsoda

      rx <- rxode2({
        CL <- 6
        V <- 25
        Q <- 0.5
        V2 <- 30
        k12 <- Q/V
        k21 <- Q/V2
        kel <- CL/V
        d/dt(central) = -kel*central + k21*peripheral - k12*central
        d/dt(peripheral) = k12*central - k21*peripheral
        cp <- central/V
      })

      e <- et(amt=0, rate=600, ss=1) |>
        et(time=seq(1:25))

      f <- function(CL=6, V=25, Q=0.5, V2=30, DOSE=600, deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- Q
        p3 <- V2
        p4 <- 0
        p5 <- 0
        ka <- 0
        if (deriv) {
          alastNV <- rep(0, 10)
        } else {
          alastNV <- c(0, 0)
        }
        rateNV <- 600
        oral0 <- 0
        trans <- 1L
        ncmt <- 2
        type <- 1L # ss infinite infusion
        amt <- 0
        tinf <- 0
        dt <- 0
        cmt <- 0L
        tau <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      expect_equal(f()$val,
                   rxSolve(rx, e, addDosing =TRUE)$cp[1],
                   tolerance = 1e-3)

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   rxSolve(rx, e, addDosing =TRUE)$cp[1],
                   tolerance = 1e-3)

      expect_equal(dim(v$J), c(2L, 4L))
      expect_length(v$Jg, 4)
      expect_length(v$Alast, 10)

    })

    test_that("rxode2 solving of 2 compartment steady state for constant depot infusion", {

      rx <- rxode2({
        CL <- 6
        V <- 25
        Q <- 0.5
        V2 <- 30
        k12 <- Q/V
        k21 <- Q/V2
        kel <- CL/V
        ka <- 0.25
        d/dt(depot) = -ka*depot
        d/dt(central) = -kel*central + k21*peripheral - k12*central + ka*depot
        d/dt(peripheral) = k12*central - k21*peripheral
        cp <- central/V
      })


      e <- et(amt=0, rate=600, ss=1) |>
        et(time=seq(1:25))

      f <- function(CL=6, V=25, Q=0.5, V2=30, DOSE=600, ka=0.25, rate=c(600, 0), deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- Q
        p3 <- V2
        p4 <- 0
        p5 <- 0
        ka <- ka
        if (deriv) {
          alastNV <- rep(0, 14)
        } else {
          alastNV <- rep(0, 3)
        }
        rateNV <- rate
        oral0 <- 1L
        trans <- 1L
        ncmt <- 2
        type <- 1L # ss infinite infusion
        amt <- 0
        tinf <- 0
        dt <- 0
        cmt <- 0L
        tau <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      s <- rxSolve(rx, e, addDosing =TRUE)

      expect_equal(f()$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(f()$Alast[1:3],
                   c(s$depot[1], s$central[1], s$peripheral[1]),
                   tolerance=1e-3)

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(v$Alast[1:3],
                   c(s$depot[1], s$central[1], s$peripheral[1]),
                   tolerance=1e-3)

      expect_equal(dim(v$J), c(3L, 5L))

      expect_length(v$Jg, 5)
      expect_length(v$Alast, 14)

      # Now infusion in the 2 compartment models to central

      e <- et(amt=0, rate=600, ss=1, cmt=2) |>
        et(time=seq(1:25))

      s <- rxSolve(rx, e, addDosing =TRUE)

      expect_equal(f(rate=c(0, 600))$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(f(rate=c(0, 600))$Alast[1:3],
                   c(s$depot[1], s$central[1],s$peripheral[1]),
                   tolerance=1e-3)

      v <- f(rate=c(0, 600), deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(v$Alast[1:3],
                   c(s$depot[1], s$central[1], s$peripheral[1]),
                   tolerance=1e-3)

      expect_equal(dim(v$J), c(3L, 5L))

      expect_length(v$Jg, 5)
      expect_length(v$Alast, 14)

    })

    test_that("rxode2 solving of 2 compartment steady state for constant depot infusion", {

      f <- function(CL=2, V=25, Q=0.5, V2=30, DOSE=600, tinf=1, tau=6, deriv=FALSE, cmt=0L,
                    ka=0.25) {
        p1 <- CL
        v1 <- V
        p2 <- Q
        p3 <- V2
        p4 <- 0
        p5 <- 0
        ka <- ka
        if (deriv) {
          alastNV <- rep(0, 14)
        } else {
          alastNV <- rep(0, 3)
        }
        if (cmt == 0L) {
          rateNV <- c(DOSE/tinf, 0)
        } else {
          rateNV <- c(0, DOSE/tinf)
        }
        oral0 <- 1L
        trans <- 1L
        ncmt <- 2
        type <- 2L # ss infusion
        amt <- DOSE
        dt <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      rx <- rxode2({
        CL <- 2
        V <- 25
        Q <- 0.5
        V2 <- 30
        k12 <- Q/V
        k21 <- Q/V2
        kel <- CL/V
        ka <- 0.25
        d/dt(depot) = -ka*depot
        d/dt(central) = -kel*central + k21*peripheral - k12*central + ka*depot
        d/dt(peripheral) = k12*central - k21*peripheral
        cp <- central/V
      })


      e <- et(amt=600, rate=600, ss=1, ii=6, cmt=depot) |>
        et(time=seq(1:25))

      s <- rxSolve(rx, e, addDosing =TRUE)

      expect_equal(f()$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(f()$Alast,
                   c(s$depot[1], s$central[1], s$peripheral[1]),
                   tolerance=1e-3)

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(v$Alast[1:3],
                   c(s$depot[1], s$central[1], s$peripheral[1]),
                   tolerance = 1e-3)

      expect_equal(dim(v$J), c(3L, 5L))

      expect_length(v$Jg, 5)
      expect_length(v$Alast, 14)

      e <- et(amt=600, rate=600, ss=1, ii=6, cmt=central) |>
        et(time=seq(1:25))

      s <- rxSolve(rx, e, addDosing =TRUE)

      expect_equal(f(cmt=1L)$val,
                   s$cp[1], tolerance = 1e-3)

      expect_equal(f(cmt=1L)$Alast,
                   c(s$depot[1], s$central[1], s$peripheral[1]),
                   tolerance = 1e-3)

      v <- f(cmt=1L, deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1], tolerance = 1e-3)

      expect_equal(v$Alast[1:2],
                   c(s$depot[1], s$central[1]), tolerance = 1e-3)

      expect_equal(dim(v$J), c(3L, 5L))

      expect_length(v$Jg, 5L)
      expect_length(v$Alast, 14)

    })

    ####################################################################
    ## Test the 3 compartment steady state solutions
    ####################################################################

    test_that("rxode2 solving of 3 compartment steady state bolus", {

      f <- function(CL=3.5, V1=20, V2 = 500, V3=200, Q2=0.5, Q3 = 0.05, DOSE=100, tau=24, deriv=FALSE) {
        p1 <- CL
        v1 <- V1
        p2 <- Q2
        p3 <- V2
        p4 <- Q3
        p5 <- V3
        ka <- 0
        if (deriv) {
          alastNV <- rep(0, 21)
        } else {
          alastNV <- c(0, 0, 0)
        }
        rateNV <- 0
        oral0 <- 0
        trans <- 1L
        ncmt <- 3
        type <- 3L # ss bolus
        amt <- DOSE
        tinf <- 1
        dt <- 0
        cmt <- 0L
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans, deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      expect_equal(pmxTools::calc_ss_3cmt_linear_bolus(0, CL=3.5, V1=20, V2 = 500, V3=200, Q2=0.5, Q3 = 0.05, dose=100, tau=24),
                   f()$val)

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   pmxTools::calc_ss_3cmt_linear_bolus(0, CL=3.5, V1=20, V2 = 500, V3=200, Q2=0.5, Q3 = 0.05, dose=100, tau=24))

      expect_equal(dim(v$J), c(3L, 6L))
      expect_length(v$Jg, 6)
      expect_length(v$Alast, 21)

    })


    test_that("rxode2 solving of 3 compartment steady state infusion", {

      f <- function(CL=7.5, V=20, V2=30, Q=0.5, DOSE=10, tinf=1, tau=12, deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- Q
        p3 <- V2
        p4 <- 0
        p5 <- 0
        ka <- 0
        if (deriv) {
          alastNV <- rep(0, 10)
        } else {
          alastNV <- c(0, 0)
        }
        rateNV <- DOSE/tinf
        oral0 <- 0
        trans <- 1L
        ncmt <- 2
        type <- 2L # ss infusion
        amt <- DOSE
        dt <- 0
        cmt <- 0L
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      expect_equal(pmxTools::calc_ss_2cmt_linear_infusion(tad = 0, CL = 7.5, V1 = 20, V2 = 30, Q = 0.5,
                                                          dose = 10, tinf = 1, tau = 12),
                   f()$val)

      v <- f(deriv=TRUE)

      expect_equal(pmxTools::calc_ss_2cmt_linear_infusion(tad = 0, CL = 7.5, V1 = 20, V2 = 30, Q = 0.5,
                                                          dose = 10, tinf = 1, tau = 12),
                   v$val)

      expect_equal(dim(v$J), c(2L, 4L))
      expect_length(v$Jg, 4)
      expect_length(v$Alast, 10)

    })

    test_that("rxode2 solving of 3 compartment steady state 1st order dosing", {

      f <- function(CL=7.5, V1=20, V2=30, V3=200, Q2=0.5, Q3=0.05, ka=1, DOSE=100, tau=24, deriv=FALSE, ndiff=127) {
        p1 <- CL
        v1 <- V1
        p2 <- Q2
        p3 <- V2
        p4 <- Q3
        p5 <- V3
        ka <- ka
        if (deriv) {
          alastNV <- rep(0, 26)
        } else {
          alastNV <- c(0, 0, 0, 0)
        }
        rateNV <- c(0, 0)
        oral0 <- 1L
        trans <- 1L
        ncmt <- 3
        type <- 3L # bolus infusion
        amt <- DOSE
        dt <- 0
        cmt <- 0L
        tinf <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, ndiff, 3L, 0.001)
      }

      expect_equal(f()$val,
                   pmxTools::calc_ss_3cmt_linear_oral_1(tad = 0, CL = 7.5, V1 = 20,
                                                        V2 = 30, V3 = 200, Q2 = 0.5,
                                                        Q3 = 0.05, ka = 1, dose = 100,
                                                        tau = 24))


      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   pmxTools::calc_ss_3cmt_linear_oral_1(tad = 0, CL = 7.5, V1 = 20,
                                                        V2 = 30, V3 = 200, Q2 = 0.5,
                                                        Q3 = 0.05, ka = 1, dose = 100,
                                                        tau = 24))

      expect_equal(dim(v$J), c(4L, 7L))
      expect_length(v$Jg, 7)
      expect_length(v$Alast, 26)

    })

    test_that("rxode2 solving of 3 compartment steady state for constant infusion", {

      ## Now test the solutions not covered by pmxTools by cross testing
      ## against rxode2 lsoda

      rx <- rxode2({
        CL <- 6
        V <- 25
        V2 <- 30
        V3 <- 200
        Q2 <- 0.5
        Q3 <- 0.05
        k12 <- Q2/V
        k21 <- Q2/V2
        k13 <- Q3/V
        k31 <- Q3/V3
        kel <- CL/V
        d/dt(central) = -kel*central + k21*peripheral - k12*central +
          k31*peripheral2 - k13*central
        d/dt(peripheral) = k12*central - k21*peripheral
        d/dt(peripheral2) = k13*central - k31*peripheral2
        cp <- central/V
      })

      e <- et(amt=0, rate=600, ss=1) |>
        et(time=seq(1:25))

      f <- function(CL= 6, V = 25, V2 = 30, V3 = 200,
                    Q2= 0.5,
                    Q3= 0.05,
                    DOSE=600, deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- Q2
        p3 <- V2
        p4 <- Q3
        p5 <- V3
        ka <- 0
        if (deriv) {
          alastNV <- rep(0, 21)
        } else {
          alastNV <- c(0, 0, 0)
        }
        rateNV <- 600
        oral0 <- 0
        trans <- 1L
        ncmt <- 3
        type <- 1L # ss infinite infusion
        amt <- 0
        tinf <- 0
        dt <- 0
        cmt <- 0L
        tau <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      expect_equal(f()$val,
                   rxSolve(rx, e, addDosing =TRUE)$cp[1],
                   tolerance = 1e-3)

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   rxSolve(rx, e, addDosing =TRUE)$cp[1],
                   tolerance = 1e-3)

      expect_equal(dim(v$J), c(3L, 6L))
      expect_length(v$Jg, 6)
      expect_length(v$Alast, 21)

    })

    test_that("rxode2 solving of 3 compartment steady state for constant depot infusion", {

      rx <- rxode2({
        CL <- 6
        V  <- 25
        Q2 <- 0.5
        Q3 <- 0.05
        V2 <- 30
        V3 <- 300
        k12 <- Q2/V
        k21 <- Q2/V2
        k13 <- Q3/V
        k31 <- Q3/V3
        kel <- CL/V
        ka <- 0.25
        d/dt(depot) = -ka*depot
        d/dt(central) = -kel*central + ka*depot +
          k21*peripheral  - k12*central +
          k31*peripheral1 - k13*central
        d/dt(peripheral) = k12*central - k21*peripheral
        d/dt(peripheral1) = k13*central - k31*peripheral1
        cp <- central/V
      })

      e <- et(amt=0, rate=600, ss=1) |>
        et(time=seq(1:25))

      s <- rxSolve(rx, e, addDosing =TRUE)

      f <- function(CL=6, V=25, Q2=0.5, Q3=0.05, V2=30, V3=300, DOSE=600,
                    ka=0.25, rate=c(600, 0), deriv=FALSE) {
        p1 <- CL
        v1 <- V
        p2 <- Q2
        p3 <- V2
        p4 <- Q3
        p5 <- V3
        ka <- ka
        if (deriv) {
          alastNV <- rep(0, 26)
        } else {
          alastNV <- rep(0, 4)
        }
        rateNV <- rate
        oral0 <- 1L
        trans <- 1L
        ncmt <- 3
        type <- 1L # ss infinite infusion
        amt <- 0
        tinf <- 0
        dt <- 0
        cmt <- 0L
        tau <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      s <- rxSolve(rx, e, addDosing =TRUE)

      expect_equal(f()$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(f()$Alast[1:4],
                   c(s$depot[1], s$central[1], s$peripheral[1], s$peripheral1[1]),
                   tolerance=1e-3)

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(v$Alast[1:4],
                   c(s$depot[1], s$central[1], s$peripheral[1],
                     s$peripheral1[1]), tolerance=1e-3)

      expect_equal(dim(v$J), c(4L, 7L))

      expect_length(v$Jg, 7)
      expect_length(v$Alast, 26)

      # Now infusion in the 3 compartment models to central

      e <- et(amt=0, rate=600, ss=1, cmt=2) |>
        et(time=seq(1:25))

      s <- rxSolve(rx, e, addDosing =TRUE)

      expect_equal(f(rate=c(0, 600))$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(f(rate=c(0, 600))$Alast[1:3],
                   c(s$depot[1], s$central[1],s$peripheral[1]),
                   tolerance=1e-3)

      v <- f(rate=c(0, 600), deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(v$Alast[1:4],
                   c(s$depot[1], s$central[1], s$peripheral[1], s$peripheral1[1]),
                   tolerance=1e-3)

      expect_equal(dim(v$J), c(4L, 7L))

      expect_length(v$Jg, 7)
      expect_length(v$Alast, 26)

    })

    test_that("rxode2 solving of 3 compartment steady state for constant depot infusion", {

      f <- function(CL=6, V=25, Q2=0.5, Q3=0.05, V2=30, V3=300, DOSE=600, tinf=1, tau=6, deriv=FALSE, cmt=0L,
                    ka=0.25) {
        p1 <- CL
        v1 <- V
        p2 <- Q2
        p3 <- V2
        p4 <- Q3
        p5 <- V3
        ka <- ka
        if (deriv) {
          alastNV <- rep(0, 26)
        } else {
          alastNV <- rep(0, 4)
        }
        if (cmt == 0L) {
          rateNV <- c(DOSE/tinf, 0)
        } else {
          rateNV <- c(0, DOSE/tinf)
        }
        oral0 <- 1L
        trans <- 1L
        ncmt <- 3
        type <- 2L # ss infusion
        amt <- DOSE
        dt <- 0
        .Call(`_rxode2_linCmtModelDouble`, 0.1,
              p1, v1, p2, p3, p4, p5, ka,
              alastNV, rateNV, ncmt, oral0, trans,
              deriv, type, tau, tinf, amt, cmt, 127, 3L, 0.001)
      }

      rx <- rxode2({
        CL <- 6
        V  <- 25
        Q2 <- 0.5
        Q3 <- 0.05
        V2 <- 30
        V3 <- 300
        k12 <- Q2/V
        k21 <- Q2/V2
        k13 <- Q3/V
        k31 <- Q3/V3
        kel <- CL/V
        ka <- 0.25
        d/dt(depot) = -ka*depot
        d/dt(central) = -kel*central + ka*depot +
          k21*peripheral  - k12*central +
          k31*peripheral1 - k13*central
        d/dt(peripheral) = k12*central - k21*peripheral
        d/dt(peripheral1) = k13*central - k31*peripheral1
        cp <- central/V
      })



      e <- et(amt=600, rate=600, ss=1, ii=6, cmt=depot) |>
        et(time=0.5)

      s <- rxSolve(rx, e, addDosing =TRUE, maxsteps=700000)

      expect_equal(f()$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(f()$Alast,
                   c(s$depot[1], s$central[1], s$peripheral[1], s$peripheral1[1]),
                   tolerance=1e-3)

      v <- f(deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1], tolerance=1e-3)

      expect_equal(v$Alast[1:4],
                   c(s$depot[1], s$central[1], s$peripheral[1], s$peripheral1[1]),
                   tolerance = 1e-3)

      expect_equal(dim(v$J), c(4L, 7L))

      expect_length(v$Jg, 7)
      expect_length(v$Alast, 26)

      e <- et(amt=600, rate=600, ss=1, ii=6, cmt=central) |>
        et(time=0.5)

      s <- rxSolve(rx, e, addDosing =TRUE, , maxsteps=700000)

      expect_equal(f(cmt=1L)$val,
                   s$cp[1], tolerance = 1e-3)

      expect_equal(f(cmt=1L)$Alast,
                   c(s$depot[1], s$central[1], s$peripheral[1], s$peripheral1[1]),
                   tolerance = 1e-3)

      v <- f(cmt=1L, deriv=TRUE)

      expect_equal(v$val,
                   s$cp[1], tolerance = 1e-3)

      expect_equal(v$Alast[1:4],
                   c(s$depot[1], s$central[1], s$peripheral[1], s$peripheral1[1]),
                   tolerance = 1e-3)

      expect_equal(dim(v$J), c(4L, 7L))

      expect_length(v$Jg, 7L)
      expect_length(v$Alast, 26)

    })


  }
})

rxTest({

  one.cmt <- function() {
    ini({
      tka <- 0.45
      tcl <- log(c(0, 2.7, 100))
      tv <- 3.45
      eta.ka ~ 0.6
      eta.cl ~ 0.3
      eta.v ~ 0.1
      add.sd <- 0.7
    })
    model({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      v <- exp(tv + eta.v)
      linCmt() ~ add(add.sd)
    })
  }

  f <- one.cmt()

  cov <- function() {
    ini({
      tka <- 0.45
      tcl <- log(c(0, 2.7, 100))
      tv <- 3.45
      tvp <- 3.45
      cl.wt <- 0.1
      v.wt <- 0.1
      cl.sex <- 0.1
      v.sex <- 0.1
      cl.age <- 0.1
      v.age <- 0.1
      vp.wt <- 1
      vp.sex <- 1
      vp.age <- 1
      eta.ka ~ 0.6
      eta.cl ~ 0.3
      eta.v ~ 0.1
      add.sd <- 0.7
    })
    model({
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl + log(wt / 70) * cl.wt + sex * cl.sex + age * cl.age + 3)
      v  <- exp(tv + eta.v + wt * v.wt + sex * v.sex + age * v.age + 2)
      vp <- exp(tvp + wt * vp.wt + sex * vp.sex + age * vp.age)
      d/dt(depot) = -ka * depot
      d/dt(center) = ka * depot - cl/v * center
      cp = center/v
      cp ~ add(add.sd)
    })
  }

  f2 <- cov()

  test_that("rxode2 model variables work with ui objects", {
    expect_s3_class(rxModelVars(one.cmt), "rxModelVars")
    expect_s3_class(rxModelVars(f), "rxModelVars")
    expect_s3_class(rxModelVars(cov), "rxModelVars")
    expect_s3_class(rxModelVars(f2), "rxModelVars")
  })

  test_that("rxode2 rxState works with ui objects", {
    expect_equal(rxState(one.cmt), c("depot", "central"))
    expect_equal(rxState(f), c("depot", "central"))
    expect_equal(rxState(cov), c("depot", "center"))
    expect_equal(rxState(f2), c("depot", "center"))
  })

  test_that("rxode2 rxLhs works with ui objects", {
    expect_equal(rxLhs(one.cmt), c("ka", "cl", "v"))
    expect_equal(rxLhs(f), c("ka", "cl", "v"))
    expect_equal(rxLhs(cov), c("ka", "cl", "v", "vp", "cp"))
    expect_equal(rxLhs(f2), c("ka", "cl", "v", "vp", "cp"))
  })

  test_that("rxode2 rxParams works with ui objects", {
    expect_equal(rxParams(one.cmt), c("tka", "eta.ka", "tcl", "eta.cl", "tv", "eta.v"))
    expect_equal(rxParams(f), c("tka", "eta.ka", "tcl", "eta.cl", "tv", "eta.v"))
    expect_equal(rxParams(cov), c("tka", "eta.ka", "tcl", "eta.cl", "wt", "cl.wt", "sex", "cl.sex", "age", "cl.age", "tv", "eta.v", "v.wt", "v.sex", "v.age", "tvp", "vp.wt", "vp.sex", "vp.age"))
    expect_equal(rxParams(f2), c("tka", "eta.ka", "tcl", "eta.cl", "wt", "cl.wt", "sex", "cl.sex", "age", "cl.age", "tv", "eta.v", "v.wt", "v.sex", "v.age", "tvp", "vp.wt", "vp.sex", "vp.age"))
  })

  test_that("rxInit works with ui obects", {
    expect_equal(rxInits(one.cmt), structure(numeric(0), names = character(0)))
    expect_equal(rxInits(f), structure(numeric(0), names = character(0)))
    expect_equal(rxInits(cov), structure(numeric(0), names = character(0)))
    expect_equal(rxInits(f2), structure(numeric(0), names = character(0)))
  })

  test_that("data frame doesn't work by itself", {
    matt <- data.frame(a=3)
    expect_error(rxModelVars(matt))
  })
})

rxTest({
  test_that("Issue RxODE#50", {
    .file <- test_path("rxodedemo_ode.txt")
    
    sink(.file)
    cat("C2 = centr/V2;
C3 = peri/V3;
d/dt(depot) =-KAdepot;
d/dt(centr) = KAdepot - CLC2 - QC2 + QC3;
d/dt(peri) = QC2 - QC3;
d/dt(eff) = Kin - Kout(1-C2/(EC50+C2))*eff;")
    sink()
    
    expect_true(file.exists(.file))
    suppressMessages(
      m1 <-
        try(
          rxode2(filename = .file, modName = test_path("m1")),
          silent=TRUE
        )
    )
    expect_false(is(class(m1), "rxode2"))
    expect_true(file.exists(.file))
    
    sink(.file)
    cat("C2 = centr/V2;
C3 = peri/V3;
d/dt(depot) =-KA*depot;
d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
d/dt(peri)  =                    Q*C2 - Q*C3;
d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;")
    sink()
    
    expect_true(file.exists(.file))
    m1 <- rxode2(filename = .file, modName = test_path("m1"))
    expect_true(is(m1, "rxode2"))
    expect_true(file.exists(test_path("rxodedemo_ode.txt")))
    
    rxDelete(m1)
    unlink(test_path("rxodedemo_ode.txt"))
    if (dir.exists(test_path("m1.d"))) unlink(test_path("m1.d"), recursive = TRUE)
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("Index access the right compartment", {
      rigid.txt <- "
y1(0)    = 1
y2(0)    = 0
y3(0)    = 0.9
a1       = -2
a2       = 1.25
a3       = -0.5
d/dt(y1) = a1*y2*y3
d/dt(y2) = a2*y1*y3
d/dt(y3) = a3*y1*y2
"
      rigid <- rxode2(rigid.txt)

      expect_equal(rxState(rigid, "y1"), 1)
      expect_equal(rxState(rigid, "y2"), 2)
      expect_equal(rxState(rigid, "y3"), 3)
      expect_error(rxState(rigid, "matt"), "cannot locate compartment")
      expect_equal(rigid$get.index("y1"), 1)
      expect_equal(rigid$get.index("y2"), 2)
      expect_equal(rigid$get.index("y3"), 3)
      expect_error(rigid$get.index("matt"), "cannot locate compartment")
    })
  }
})

rxTest({
  test_that("Absolute Working Directory + Model Name; Issue RxODE#56", {
    skip_on_os("windows")
    skip_on_os("solaris")
    if (!dir.exists("/tmp/")) {
      skip()
    }
    ode <- " d/dt(test) = 0; "
    m1 <- rxode2(model = ode, modName = "m1", wd = "/tmp")
    expect_true(dir.exists("/tmp/m1.d"))
    rxDelete(m1)
    unlink("/tmp/m1.d", recursive = TRUE)
  })
}) 

rxTest({
  # Working Directory + Model Name (Issue RxODE#4)

  model_text <- "
  C2 = centr/V2;
  C3 = peri/V3;
  d/dt(depot) =-KA*depot;
  d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
  d/dt(peri)  =                    Q*C2 - Q*C3;
  d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
  "
  modName <- paste(sample(LETTERS, 5), collapse = "")
  model <- rxode2::rxode2(model = model_text, modName = modName, wd = "mytempdir")

  test_that("Model has the correct states", {
    expect_equal(
      rxState(model),
      c("depot", "centr", "peri", "eff")
    )
    expect_equal(
      rxParams(model),
      c("V2", "V3", "KA", "CL", "Q", "Kin", "Kout", "EC50")
    )
  })

  rxDelete(model)
  unlink("mytempdir", recursive = TRUE)
})

rxTest({
  test_that("single cvPost draw makes sense", {
    draw1 <- cvPost(3, matrix(c(1, .3, .3, 1), 2, 2))
    expect_true(is.matrix(draw1))
    expect_equal(dim(draw1), c(2L, 2L))
  })

  test_that("cvPost of 3 items make sense.", {
    rxWithSeed(42, {
      mat1 <- matrix(c(1, .3, .3, 1), 2, 2)
      draw3 <- cvPost(3, mat1, n = 3)
      drawNull <- cvPost(NULL, mat1, n = 1)
      expect_equal(drawNull, mat1)
    })
    rxWithSeed(
      42,
      draw3c <- cvPost(3, matrix(c(1, .3, .3, 1), 2, 2), n = 3, returnChol = TRUE)
    )
    rxWithSeed(
      42,
      draw3ct1 <- cvPost(3, matrix(c(1, .3, .3, 1), 2, 2), n = 3, returnChol = TRUE, type = 1)
    )
    expect_type(draw3, "list")
    expect_type(draw3c, "list")
    for (i in seq_along(draw3)) {
      expect_equal(dim(draw3[[i]]), c(2L, 2L))
      expect_equal(dim(draw3c[[i]]), c(2L, 2L))
      expect_equal(chol(draw3[[i]]), draw3c[[i]])
    }
    expect_equal(draw3c, draw3ct1)
    rxWithSeed(
      42,
      draw3c <- cvPost(3, chol(matrix(c(1, .3, .3, 1), 2, 2)), n = 3, omegaIsChol = TRUE)
    )
    rxWithSeed(
      42,
      draw3 <- cvPost(3, matrix(c(1, .3, .3, 1), 2, 2), n = 3)
    )
    for (i in seq_along(draw3)) {
      expect_equal(draw3[[1]], draw3c[[1]])
    }

    rxWithSeed(
      42,
      lkj <-
        cvPost(
          3,
          vapply(1:3, function(...) {
            rnorm(10)
          },
          numeric(10)),
          type = "lkj",
          returnChol = TRUE
        )
    )
    rxWithSeed(
      42,
      lkjTn <-
        cvPost(
          3,
          vapply(1:3, function(...) {
            rnorm(10)
          }, numeric(10)),
          type = 2,
          returnChol = TRUE, diagXformType = 5
        )
    )
    expect_equal(lkj, lkjTn)
  })

  test_that("rinvchisq produces proper output", {
    expect_equal(length(rinvchisq(3, 4, 1)), 3) ## Scale = 1, degrees of freedom = 4
    expect_equal(length(rinvchisq(3, 4, 1)), 3) ## Scale = 1, degrees of freedom = 4
  })

  test_that("rcvC1 single-variance (1x1) case does not abort", {
    ## Before the fix the 1x1 branch did `ret = ret(1,1)` -- element access on a
    ## default-constructed 0x0 matrix -> "Mat::operator(): index out of bounds".
    r <- rcvC1(sdEst = 2.0, diagXformType = 1L)
    expect_equal(dim(r), c(1L, 1L))
    ## diagXformType=1 (sqrt): sd = 1/sdEst^2 = 0.25; ret = sd^2 = 0.0625
    expect_equal(as.numeric(r), 0.0625)
  })

  test_that("rcvC1 separation strategy passes (d, nu) in the correct order", {
    ## Before the fix rinvWRcv1 called invWR1d(nu, d) with the arguments swapped
    ## (signature is invWR1d(int d, double nu)), so a valid d=3, nu=5 request
    ## failed with "'nu' must be greater than 'd'-1".
    expect_error(r <- rcvC1(sdEst = c(1, 1, 1), nu = 5, rType = 2L), NA)
    expect_equal(dim(r), c(3L, 3L))
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    ## Adapted from https://github.com/biobakery/banocc/blob/master/tests/testthat/test_utils_rlkj.R
    # context("Utilities - rLKJ1")
    d_vals <- c(2, 5) # seq(2, 200, 10)
    eta_vals <- c(1, 50) # seq(1, 200, 10)

    test_that("rLKJ1 returns a matrix", {
      for (d in d_vals) {
        for (eta in eta_vals) {
          eval(bquote(expect_true(is.matrix(rLKJ1(d = .(d), eta = .(eta))))))
          eval(bquote(
            expect_true(is.matrix(rLKJ1(d = .(d), eta = .(eta), cholesky = TRUE)))
          ))
        }
      }
    })

    test_that("rLKJ1 returns square matrix", {
      test_square <- function(d, e, ch) {
        r <- rLKJ1(d = d, eta = e, cholesky = ch)
        expect_equal(nrow(r), ncol(r))
      }
      for (d in d_vals) {
        for (eta in eta_vals) {
          test_square(d, eta, FALSE)
          test_square(d, eta, TRUE)
        }
      }
    })

    test_that("rLKJ1 returns matrix with d rows", {
      test_nrow <- function(d, e, ch) {
        eval(bquote(
          expect_equal(nrow(rLKJ1(d = .(d), eta = .(e), cholesky = .(ch))), .(d))
        ))
      }
      for (d in d_vals) {
        for (eta in eta_vals) {
          test_nrow(d, eta, TRUE)
          test_nrow(d, eta, FALSE)
        }
      }
    })

    test_that("rLKJ1 returns positive definite matrix if cholesky=FALSE", {
      test_pd <- function(d, e) {
        eval(bquote(expect_true(all(eigen(rLKJ1(d = .(d), eta = .(e)))$values > 0))))
      }
      for (d in d_vals) {
        for (eta in eta_vals) {
          test_pd(d, eta)
        }
      }
    })

    test_that("rLKJ1 returns symmetric matrix if cholesky=FALSE", {
      get_tri <- function(mat, type) {
        if (type == "lower") {
          t(mat)[upper.tri(mat)]
        } else {
          mat[upper.tri(mat)]
        }
      }
      test_symmetric <- function(d, e) {
        l <- rLKJ1(d = d, eta = e)
        expect_equal(
          get_tri(l, "lower"),
          get_tri(l, "upper")
        )
      }
      for (d in d_vals) {
        for (eta in eta_vals) {
          test_symmetric(d, eta)
        }
      }
    })

    test_that("rLKJ1 returns matrix with diagonal elts 1 if cholesky=FALSE", {
      test_diag <- function(d, e) {
        eval(bquote(
          expect_equal(diag(rLKJ1(d = .(d), eta = .(e))), rep(1, .(d)))
        ))
      }
      for (d in d_vals) {
        for (eta in eta_vals) {
          test_diag(d, eta)
        }
      }
    })

    test_that("rLKJ1 returns matrix with elts between -1 and 1 if cholesky=FALSE", {
      test_elts <- function(d, e) {
        r <- rLKJ1(d = d, eta = e)
        r[upper.tri(r)]
      }
      for (d in d_vals) {
        for (eta in eta_vals) {
          expect_true(all(test_elts(d, eta) >= -1))
          expect_true(all(test_elts(d, eta) <= 1))
        }
      }
    })

    test_that("rLKJ1 returns matrix with upper triangle of 0 if cholesky=TRUE", {
      get_uppertri <- function(mat) {
        mat[upper.tri(mat)]
      }
      test_uppertri <- function(d, e) {
        eval(bquote(
          expect_equal(
            get_uppertri(rLKJ1(d = .(d), eta = .(e), cholesky = TRUE)),
            rep(0, choose(.(d), 2))
          )
        ))
      }
      for (d in d_vals) {
        for (eta in eta_vals) {
          test_uppertri(d, eta)
        }
      }
    })

    test_that("rLKJ1 corr matrix is positive definite if cholesky=TRUE", {
      test_pd <- function(d, e) {
        l <- rLKJ1(d = d, eta = e, cholesky = TRUE)
        eigen(l %*% t(l))$values
      }
      for (d in d_vals) {
        for (eta in eta_vals) {
          expect_true(all(test_pd(d, eta) > 0))
        }
      }
    })

    test_that("rLKJ1 corr matrix is symmetric if cholesky=TRUE", {
      get_tri <- function(mat, type) {
        if (type == "lower") {
          t(mat)[upper.tri(mat)]
        } else {
          mat[upper.tri(mat)]
        }
      }
      test_symmetric <- function(d, e) {
        l <- rLKJ1(d = d, eta = e, cholesky = TRUE)
        r <- l %*% t(l)
        expect_equal(get_tri(r, "lower"), get_tri(r, "upper"))
      }
      for (d in d_vals) {
        for (eta in eta_vals) {
          test_symmetric(d, eta)
        }
      }
    })

    test_that("rLKJ1 corr matrix has diagonal elements of 1 if cholesky=TRUE", {
      test_diag <- function(d, e) {
        l <- rLKJ1(d = d, eta = e, cholesky = TRUE)
        diag(l %*% t(l))
      }
      for (d in d_vals) {
        for (eta in eta_vals) {
          expect_equal(test_diag(d, eta), rep(1, d))
        }
      }
    })

    test_that("rLKJ1 corr matrix has elts between -1 and 1 if cholesky=TRUE", {
      test_elts <- function(d, e) {
        l <- rLKJ1(d = d, eta = e, cholesky = TRUE)
        (l %*% t(l))[upper.tri(l)]
      }
      for (d in d_vals) {
        for (eta in eta_vals) {
          expect_true(all(test_elts(d, eta) >= -1))
          expect_true(all(test_elts(d, eta) <= 1))
        }
      }
    })

    test_that("rLKJ1 gives error if eta < 1", {
      err_string <- "must be >= 1"
      for (d in d_vals) {
        for (eta in c(-1, 0)) {
          expect_error(rLKJ1(d = d, eta = eta, cholesky = TRUE), err_string)
          expect_error(rLKJ1(d = d, eta = eta, cholesky = FALSE), err_string)
        }
      }
    })

    test_that("rLKJ1 gives error if d < 2", {
      err_string <- "must be > 1"
      for (d in c(-1, 0, 1)) {
        for (eta in eta_vals) {
          expect_error(rLKJ1(d = d, eta = eta, cholesky = TRUE), err_string)
          expect_error(rLKJ1(d = d, eta = eta, cholesky = FALSE), err_string)
        }
      }
    })

    # context("Utilities - invWR1d")
    nu_vals <- c(4.5, 50)
    d_vals <- c(2, 5)

    test_that("invWR1d returns a matrix", {
      for (d in d_vals) {
        for (nu in nu_vals) {
          expect_true(is.matrix(eval(bquote(invWR1d(d = .(d), nu = .(nu))))))
          eval(bquote(
            expect_true(is.matrix(invWR1d(d = .(d), nu = .(nu))))
          ))
        }
      }
    })

    test_that("invWR1d returns square matrix", {
      test_square <- function(d, e) {
        r <- invWR1d(d = d, nu = e)
        expect_equal(nrow(r), ncol(r))
      }
      for (d in d_vals) {
        for (nu in nu_vals) {
          test_square(d, nu)
        }
      }
    })

    test_that("invWR1d returns matrix with d rows", {
      test_nrow <- function(d, e) {
        eval(bquote(
          expect_equal(nrow(invWR1d(d = .(d), nu = .(e))), .(d))
        ))
      }
      for (d in d_vals) {
        for (nu in nu_vals) {
          test_nrow(d, nu)
        }
      }
    })

    test_that("invWR1d returns positive definite matrix if cholesky=FALSE", {
      test_pd <- function(d, e) {
        eval(bquote(expect_true(all(eigen(invWR1d(d = .(d), nu = .(e)))$values > 0))))
      }
      for (d in d_vals) {
        for (nu in nu_vals) {
          test_pd(d, nu)
        }
      }
    })

    test_that("invWR1d returns symmetric matrix", {
      get_tri <- function(mat, type) {
        if (type == "lower") {
          t(mat)[upper.tri(mat)]
        } else {
          mat[upper.tri(mat)]
        }
      }
      test_symmetric <- function(d, e) {
        l <- invWR1d(d = d, nu = e)
        expect_equal(
          get_tri(l, "lower"),
          get_tri(l, "upper")
        )
      }
      for (d in d_vals) {
        for (nu in nu_vals) {
          test_symmetric(d, nu)
        }
      }
    })

    test_that("invWR1d returns matrix with diagonal elts 1", {
      test_diag <- function(d, e) {
        eval(bquote(
          expect_equal(diag(invWR1d(d = .(d), nu = .(e))), rep(1, .(d)))
        ))
      }
      for (d in d_vals) {
        for (nu in nu_vals) {
          test_diag(d, nu)
        }
      }
    })

    test_that("invWR1d returns matrix with elts between -1 and 1", {
      test_elts <- function(d, e) {
        r <- invWR1d(d = d, nu = e)
        r[upper.tri(r)]
      }
      for (d in d_vals) {
        for (nu in nu_vals) {
          expect_true(all(test_elts(d, nu) >= -1))
          expect_true(all(test_elts(d, nu) <= 1))
        }
      }
    })

    test_that("invWR1d corr matrix is positive definite", {
      test_pd <- function(d, e) {
        l <- invWR1d(d = d, nu = e)
        eigen(l %*% t(l))$values
      }
      for (d in d_vals) {
        for (nu in nu_vals) {
          expect_true(all(test_pd(d, nu) > 0))
        }
      }
    })

    test_that("invWR1d corr matrix is symmetric", {
      get_tri <- function(mat, type) {
        if (type == "lower") {
          t(mat)[upper.tri(mat)]
        } else {
          mat[upper.tri(mat)]
        }
      }
      test_symmetric <- function(d, e) {
        l <- invWR1d(d = d, nu = e)
        r <- l %*% t(l)
        expect_equal(get_tri(r, "lower"), get_tri(r, "upper"))
      }
      for (d in d_vals) {
        for (nu in nu_vals) {
          test_symmetric(d, nu)
        }
      }
    })

    test_that("invWR1d gives error if nu < d-1", {
      err_string <- "'nu' must be greater than 'd'-1"
      for (d in d_vals) {
        for (nu in c(-1, 1)) {
          expect_error(invWR1d(d = d, nu = nu), err_string)
        }
      }
    })
  }
})

rxTest({
  # Test loading, and unloading of models

  rigid.txt <- "
y1(0)    = 1
y2(0)    = 0
y3(0)    = 0.9
a1       = -2
a2       = 1.25
a3       = -0.5
d/dt(y1) = a1*y2*y3
d/dt(y2) = a2*y1*y3
d/dt(y3) = a3*y1*y2
"

  rigid <- rxode2(rigid.txt)

  test_that("loading and unloading works.", {
    dll <- rxDll(rigid)
    expect_true(rxDllLoaded(rigid))
    expect_true(file.exists(dll))
    rxUnload(rigid)
    expect_false(rxDllLoaded(rigid))
    expect_true(file.exists(dll))
    rxLoad(rigid)
    expect_true(rxDllLoaded(rigid))
    expect_true(file.exists(dll))
    rxDelete(rigid)
    expect_false(rxDllLoaded(rigid))
    expect_false(file.exists(dll))
    rxLoad(rigid)
    expect_true(rxDllLoaded(rigid))
    expect_true(file.exists(dll))
    rxDelete(rigid)
    expect_false(rxDllLoaded(rigid))
    expect_false(file.exists(dll))
    ## Test $ syntax
    rigid$compile()
    expect_true(rxDllLoaded(rigid))
    expect_true(file.exists(dll))
    rigid$delete()
    expect_false(rxDllLoaded(rigid))
    expect_false(file.exists(dll))
    rigid$dynLoad()
    expect_true(rxDllLoaded(rigid))
    expect_true(file.exists(dll))
    rigid$dynUnload()
    expect_false(rxDllLoaded(rigid))
    expect_true(file.exists(dll))
  })
})

rxTest({
  m1 <- rxode2({
    C2 <- centr / V2
    C3 <- peri / V3
    d / dt(depot) <- -KA * depot
    d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
    d / dt(peri) <- Q * C2 - Q * C3
    d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
  })

  test_that("blank names works", {
    expect_equal(
      suppressWarnings(rxInits(m1, c(0, 0, 0, 1), rxState(m1), 0)),
      structure(c(0, 0, 0, 1), names = c("depot", "centr", "peri", "eff"))
    )
  })

  out <- rxode2("ini = 1; fun_ini = 2; fun = 4; addit = ini + fun_ini + pi + no_ini")

  test_that("Initial constants are correct", {
    expect_equal(rxInits(out)["pi"], c(pi = pi))
    expect_equal(rxInits(out)["ini"], c(ini = 1))
    expect_equal(rxInits(out)["fun_ini"], c(fun_ini = 2))
    expect_equal(rxInits(out)["fun"], c(fun = 4))
  })

  test_that("Constants are not included in get.modelVars()", {
    expect_equal(out$get.modelVars()$params, "no_ini")
  })

  .rxWithOptions(list(rxode2.syntax.allow.ini = FALSE), {
    out2 <- rxode2("ini = 1; fun_ini = 2; fun = 4; addit = ini + fun_ini + pi + no_ini")

    test_that("Initial constants only include pi.", {
      expect_equal(names(rxInits(out2)), "pi")
      expect_true(rxDll(out) != rxDll(out2))
      expect_equal(out2$get.modelVars()$params, "no_ini")
    })
  })

  .rxWithOptions(list(rxode2.syntax.allow.ini = TRUE), {

    ## out <- rxode2({
    ##     theta[1] = 3
    ##     eta[1] = 2
    ##     k = exp(theta[1] + eta[1])
    ##     d / dt(central) = -theta[1] * central
    ## })

    ## test_that("Allow THETA[#] and ETA[#]s.", {
    ## })

    fini <- rxode2({
      C2 <- centr / V2
      C3 <- peri / V3
      d / dt(depot) <- -KA * depot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- theta1 + eta1
    })

    expect_equal(c(depot = 0L, centr = 0L, peri = 0L, eff = 1L),
                 rxModelVars(fini)$stateProp)

    theta <-
      c(
        KA = 2.94E-01, CL = 1.86E+01, V2 = 4.02E+01, # central
        Q = 1.05E+01, V3 = 2.97E+02, # peripheral
        Kin = 1, Kout = 1, EC50 = 200, eta1 = 0, theta1 = 1
      ) # effects

      ev <- eventTable(amount.units = "mg", time.units = "hours")
      ev$add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12)
      ev$add.dosing(dose = 20000, nbr.doses = 5, start.time = 120, dosing.interval = 24)
      ev$add.sampling(0:240)

      s <- fini |> solve(theta, ev)

      expect_equal(as.data.frame(s)[1, "eff"], 1)

      rxDelete(fini)

      fini <- rxode2({
        C2 <- centr / V2
        C3 <- peri / V3
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
        eff(0) <- theta1 + eta1
      })

      # This should flag that the ini state has been used.
      expect_equal(rxModelVars(fini)$stateProp,
                   c(depot = 0L, centr = 0L, peri = 0L, eff = 1L))

      theta <-
        c(
          KA = 2.94E-01, CL = 1.86E+01, V2 = 4.02E+01, # central
          Q = 1.05E+01, V3 = 2.97E+02, # peripheral
          Kin = 1, Kout = 1, EC50 = 200, eta1 = 0, theta1 = 1
        ) # effects

        ev <- eventTable(amount.units = "mg", time.units = "hours")
        ev$add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12)
        ev$add.dosing(dose = 20000, nbr.doses = 5, start.time = 120, dosing.interval = 24)
        ev$add.sampling(0:240)

        s <- fini |> solve(theta, ev)

        expect_equal(as.data.frame(s)[1, "eff"], 1)

        theta["eta1"] <- 0.5

        s <- fini |> solve(theta, ev)
        expect_equal(as.data.frame(s)[1, "eff"], 1.5)

        theta["eta1"] <- -0.5

        s <- fini |> solve(theta, ev)
        expect_equal(as.data.frame(s)[1, "eff"], 0.5)



        ## Now with rxGetModel
        m1 <- rxGetModel({
          C2 <- centr / V2
          C3 <- peri / V3
          d / dt(depot) <- -KA * depot
          d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
          d / dt(peri) <- Q * C2 - Q * C3
          d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
        })

        test_that("blank names works", {
          expect_equal(
            suppressWarnings(rxInits(m1, c(0, 0, 0, 1), rxState(m1), 0)),
            structure(c(0, 0, 0, 1), names = c("depot", "centr", "peri", "eff"))
          )
        })
  })

  .rxWithOptions(list(rxode2.syntax.allow.ini = TRUE), {
    out <- rxGetModel("ini = 1; fun_ini = 2; fun = 4; addit = ini + fun_ini + pi + no_ini")

    test_that("Initial constants are correct", {
      expect_equal(rxInits(out)["pi"], c(pi = pi))
      expect_equal(rxInits(out)["ini"], c(ini = 1))
      expect_equal(rxInits(out)["fun_ini"], c(fun_ini = 2))
      expect_equal(rxInits(out)["fun"], c(fun = 4))
    })
  })

  .rxWithOptions(list(rxode2.syntax.allow.ini = FALSE), {
    out2 <- rxGetModel("ini = 1; fun_ini = 2; fun = 4; addit = ini + fun_ini + pi + no_ini")

    test_that("Initial constants only include pi.", {
      expect_equal(names(rxInits(out2)), "pi")
    })

    test_that("Initial conditions are zero length before and after compile", {
      expect_equal(rxModelVars("KA=exp(THETA[1]);\nCL=exp(THETA[2]+ETA[1]);\nV=exp(THETA[3]+ETA[2]);\nd/dt(depot)=-KA*depot;\nd/dt(centr)=KA*depot-CL/V*centr;\nrx_yj_=2;\nrx_lambda_=1;\nrx_pred_f_~centr;\nrx_pred_=centr;\nrx_r_=(THETA[4])^2;\n")$ini, structure(numeric(0), names = character(0)))
      tmp <- rxode2("KA=exp(THETA[1]);\nCL=exp(THETA[2]+ETA[1]);\nV=exp(THETA[3]+ETA[2]);\nd/dt(depot)=-KA*depot;\nd/dt(centr)=KA*depot-CL/V*centr;\nrx_yj_=2;\nrx_lambda_=1;\nrx_pred_f_~centr;\nrx_pred_=centr;\nrx_r_=(THETA[4])^2;\n")
      expect_equal(rxModelVars(tmp)$ini, structure(numeric(0), names = character(0)))
    })
  })
})

rxTest({
  ## To keep basename the same
  .rxWithOptions(list(
    "rxode2.basename.print" = "basename",
    "rxode2.dll.print" = "dll",
    "rxode2.c.print" = "cfile"
  ), {
    ## test_path("test-print.txt")
    .df <- expand.grid(color = c(TRUE, FALSE), unicode = c(TRUE, FALSE))
    lapply(seq_along(.df$color), function(.i) {
      .path <- test_path(sprintf(
        "test-print%s%s.txt", ifelse(.df$color[.i], "-color", ""),
        ifelse(.df$unicode[.i], "-unicode", "")
      ))
      test_that(paste0(.i, ":", .path), {
        verify_output(
          .path,
          {
            mod <- rxode2({
              a <- 6
              b <- 0.6 + a / 100
              kel <- b * 0.01
              V <- 10
              d / dt(intestine) <- -a * intestine
              d / dt(blood) <- a * intestine - b * blood
              l2 <- linCmt()
            })

            print(mod)

            summary(mod)
            str(mod) ## fragile

            print(mod$cmpMgr$rxDll())
            summary(mod$cmpMgr$rxDll())

            coef(mod)

            print(rxModelVars(mod))

            print(rxC(mod))
            summary(rxC(mod)) # too fragile

            print(mod$.rxDll)

            et <- eventTable(time.units = "days")
            et$add.sampling(seq(0, 10, by = 1 / 24))
            et$add.dosing(
              dose = 2 / 24, rate = 2, start.time = 0,
              nbr.doses = 10, dosing.interval = 1
            )

            print(et)
            summary(et)
            str(et)

            et2 <- et |> et(id = 1:10)
            print(et2)

            tmp <- class(et2)
            et2$matt <- 4
            class(et2) <- tmp
            print(et2)

            print(attr(class(et), ".rxode2.lst"))
            str(attr(class(et), ".rxode2.lst"))

            print(format(structure(0:7, class = "rxEvid")))
            print(structure(0:7, class = "rxEvid"))

            print(format(structure(c(-2, -1, 0, 1, 2), class = "rxRateDur")))
            print(structure(c(-2, -1, 0, 1, 2), class = "rxRateDur"))


            pk <- solve(mod, et)
            print(pk)

            print(pk, width = 40)

            print(pk, bound = "k")

            print(pk, n = 10)

            .rxWithOptions(
              list(rxode2.display.tbl = FALSE),
              print(pk)
            )

            summary(pk)
            str(pk)

            print(summary(pk, bound = "k"))


            modS <- rxode2(mod, calcSens = TRUE)

            print(coef(modS))

            ## Now "destroy" the rxode2 solved object and change the printout
            tmp <- class(pk)
            pk$matt <- 4
            class(tmp) <- tmp
            print(pk)
            class(pk) <- tmp
            summary(pk)

            noOde <- rxode2({
              matt <- a^2 + b^2
            })

            print(coef(noOde))


            rxUnload(mod)
            print(mod)
            print(mod$cmpMgr$rxDll())
            summary(mod)
            str(mod)

            rxDelete(mod)
            print(mod)
            print(mod$cmpMgr$rxDll())
            summary(mod)
            str(mod)

            mod <- rxode2(mod, indLin = TRUE)

            print(mod)
            summary(mod)
            str(mod)

            rxUnload(mod)
            print(mod)
            summary(mod)
            str(mod)

            rxDelete(mod)
            print(mod)
            summary(mod)
            str(mod)

            rxWithSeed(
              42,
              tmp <- matrix(rnorm(5^2), 5, 5)
            )
            mcov <- tcrossprod(tmp, tmp)
            v <- rxSymInvCholCreate(mcov, "sqrt")

            print(v)
            str(v)

            class(v) <- NULL
            v$env$theta <- NULL
            class(v) <- "rxSymInvCholEnv"
            print(v)

            mod <- rxode2("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
d/dt(blood)     = a*intestine - b*blood
")


            et <- eventTable()
            et$add.dosing(
              dose = 2 / 24, rate = 2, start.time = 0,
              nbr.doses = 10, dosing.interval = 1
            )
            et <- et |>
              et(0.05, evid = 2) |>
              et(amt = 3, time = 0.5, cmt = out) |>
              et(amt = 3, time = 0.1, cmt = intestine, ss = 1, ii = 3) |>
              et(amt = 3, time = 0.3, cmt = intestine, ss = 2, ii = 3) |>
              et(time = 0.2, cmt = "-intestine") |>
              as.data.frame()

            ett1 <- etTrans(et, mod, keepDosingOnly = TRUE)

            print(ett1)

            mod2 <- rxode2({
              ## the order of variables do not matter, the type of compartmental
              ## model is determined by the parameters specified.
              CL ~ TCL * exp(eta.Cl)
              C2 ~ linCmt(KA, CL, V2, Q, V3)
              eff(0) <- 1 + eff0^2 ## This specifies that the effect compartment starts at 1.
              d / dt(eff) ~ Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
              ##
              resp <- eff + err1
              pk <- C2 * exp(err2)
            })


            ev <- eventTable(amount.units = "mg", time.units = "hours") |>
              add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12, dosing.to = 2) |>
              add.dosing(dose = 20000, nbr.doses = 5, start.time = 120, dosing.interval = 24, dosing.to = 2) |>
              add.sampling(0:240)

            ## Add Residual differences
            sigma <- diag(2) * 0.05
            dimnames(sigma) <- list(c("err1", "err2"), c("err1", "err2"))

            pk3 <- rxSolve(mod2, c(
              KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
              Kin = 1, Kout = 1, EC50 = 200
            ),
            omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
            nSub = 4, ev, sigma = sigma, cores = 2
            )
            print(pk3)
            str(pk3)

            pk3a <- rxSolve(mod2, c(
              KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
              Kin = 1, Kout = 1, EC50 = 200
            ),
            omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
            nSub = 4, ev, sigma = sigma, cores = 2, addDosing = TRUE
            )
            print(pk3a)
            str(pk3a)

            ode <- rxode2({
              b <- -1
              d / dt(X) <- a * X + Y * Z
              d / dt(Y) <- b * (Y - Z)
              d / dt(Z) <- -X * Y + c * Y - Z
            })

            et <- eventTable(time.units = "hr") # default time units
            et$add.sampling(seq(from = 0, to = 100, by = 0.01))

            cov <- data.frame(c = et$get.EventTable()$time + units::set_units(1, h))

            et0 <- et

            et <- cbind(et, cov)

            et$extra <- 1

            out <- rxSolve(ode,
                           params = c(a = -8 / 3, b = -10),
                           events = et,
                           inits = c(X = 1, Y = 1, Z = 1),
                           covsInterpolation = "linear",
                           keep = "extra"
                           )
            print(out)
            str(out)

            mmModel <- rxode2(
            {
              ka <- 1
              Vc <- 1
              Vmax <- 0.00734
              Km <- 0.3672
              Cp <- center / Vc
              d / dt(center) <- -Vmax / (Km + Cp) * Cp + exp(-10 * t)
            },
            indLin = TRUE
            )

            print(mmModel)
            summary(mmModel)

            mmModel <- rxode2(
            {
              ka <- 1
              Vc <- 1
              Vmax <- 0.00734
              Km <- 0.3672
              Cp <- center / Vc
              d / dt(center) <- -Vmax / (Km + Cp) * Cp
            },
            indLin = TRUE
            )

            print(mmModel)
            summary(mmModel)

            mod2 <- rxode2({
              C2 <- centr / V2
              C3 ~ peri / V3
              CL ~ TCL * exp(eta.Cl)
              d / dt(depot) ~ -KA * depot
              d / dt(centr) ~ KA * depot - CL * C2 - Q * C2 + Q * C3
              d / dt(peri) ~ Q * C2 - Q * C3
              d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
              eff(0) <- 1000
              e1 <- err1
              e2 <- err2
              resp <- eff + e1
              pk <- C2 * exp(e2)
            })


            thetaMat <- diag(3) * 0.01
            dimnames(thetaMat) <- list(NULL, c("KA", "TCL", "V2"))
            sigma <- diag(2) * 0.05
            dimnames(sigma) <- list(c("err1", "err2"), c("err1", "err2"))

            ev <- eventTable(amount.units = "mg", time.units = "hours") |>
              add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12, dosing.to = 2) |>
              add.dosing(dose = 20000, nbr.doses = 5, start.time = 120, dosing.interval = 24, dosing.to = 2) |>
              add.sampling(0:240)


            pk4 <- rxSolve(mod2, c(
              KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
              Kin = 1, Kout = 1, EC50 = 200
            ),
            omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
            nSub = 4, nStud = 4, thetaMat = thetaMat, sigma = sigma, ev, cores = 1, method = "lsoda"
            )

            print(pk4)

            pk4 <- rxSolve(mod2, c(
              KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
              Kin = 1, Kout = 1, EC50 = 200
            ),
            omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
            nSub = 4, nStud = 4, dfObs = 4, dfSub = 4,
            thetaMat = thetaMat, sigma = sigma, ev,
            cores = 1, method = "lsoda"
            )

            print(pk4)

            pk4 <- rxSolve(mod2, c(
              KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
              Kin = 1, Kout = 1, EC50 = 200
            ),
            omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
            nSub = 4, nStud = 4,
            sigma = sigma, ev,
            cores = 1, method = "lsoda"
            )

            print(pk4)
          },
crayon = .df$color[.i],
unicode = .df$unicode[.i]
)
      })

      unlink(.path)
    })
  })
})

##     options(rxode2.unload.unused = TRUE)
##     test_that("Check garbage collection unloads DLLs", {
##       options(rxode2.unload.unused = TRUE)

##       ode <- rxode2({
##         b <- -1
##         d / dt(X) <- a * X + Y * Z
##         d / dt(Y) <- b * (Y - Z)
##         d / dt(Z) <- -X * Y + c * Y - Z
##       })

##       name <- basename(rxDll(ode))
##       name <- substr(name, 0, nchar(name) - nchar(.Platform$dynlib.ext))

##       expect_false(is.null(getLoadedDLLs()[[name]]))
##       Sys.sleep(0.5)
##       gc()
##       Sys.sleep(0.5)
##       expect_false(is.null(getLoadedDLLs()[[name]]))
##       rm(ode)
##       Sys.sleep(0.5)
##       gc()
##       Sys.sleep(0.5)
##       expect_true(is.null(getLoadedDLLs()[[name]]))

##       options(rxode2.unload.unused = FALSE)
##     })
##     options(rxode2.unload.unused = FALSE)

rxTest({
  test_that("occasions", {

    # Nesting tests

    mod <- rxode2({
      eff(0) <- 1
      C2 <- centr / V2 * (1 + prop.err)
      C3 <- peri / V3
      CL <- TCl * exp(eta.Cl + iov.Cl)
      KA <- TKA * exp(eta.Ka + iov.Ka)
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d/dt(peri) <- Q * C2 - Q * C3
      d/dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
    })

    mod.eta <- rxode2({
      eff(0) <- 1
      C2 <- centr / V2 * (1 + prop.err)
      C3 <- peri / V3
      CL <- TCl * exp(ETA[1] + iov.Cl)
      KA <- TKA * exp(ETA[2] + iov.Ka)
      d / dt(depot) <- -KA * depot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
    })

    et(amountUnits = "mg", timeUnits = "hours") |>
      et(amt = 10000, addl = 9, ii = 12, cmt = "depot") |>
      et(time = 120, amt = 2000, addl = 4, ii = 14, cmt = "depot") |>
      et(seq(0, 240, by = 4)) |>
      # Assumes sampling when there is no dosing information
      et(seq(0, 240, by = 4) + 0.1) |>
      ## adds 0.1 for separate eye
      et(id = 1:20) |>
      ## Add an occasion per dose
      dplyr::mutate(occ = cumsum(!is.na(amt))) |>
      dplyr::mutate(occ = ifelse(occ == 0, 1, occ)) |>
      dplyr::mutate(occ = 2 - occ %% 2) |>
      dplyr::mutate(eye = ifelse(round(time) == time, 1, 2)) |>
      dplyr::mutate(inv = ifelse(id < 10, 1, 2)) ->
      ev

      omega <- lotri(
        lotri(
          eta.Cl ~ 0.1,
          eta.Ka ~ 0.1
        ) | id(nu = 100),
        lotri(
          eye.Cl ~ 0.05,
          eye.Ka ~ 0.05
        ) | eye(nu = 50, same = 2),
        lotri(
          iov.Cl ~ 0.01,
          iov.Ka ~ 0.01
        ) | occ(nu = 200, same = 2),
        lotri(
          inv.Cl ~ 0.02,
          inv.Ka ~ 0.02
        ) | inv(nu = 10, same = 2)
      )
      attr(omega, "format") <- "THETA[%d]"
      attr(omega, "start") <- 2L

      ## cvPost(nu=1000, omega, 2)

      omega <- lotri(
        lotri(
          eta.Cl ~ 0.1,
          eta.Ka ~ 0.1
        ) | id(nu = 100),
        lotri(
          eye.Cl ~ 0.05,
          eye.Ka ~ 0.05
        ) | eye(nu = 50),
        lotri(
          iov.Cl ~ 0.01,
          iov.Ka ~ 0.01
        ) | occ(nu = 200),
        lotri(
          inv.Cl ~ 0.02,
          inv.Ka ~ 0.02
        ) | inv(nu = 10)
      )

      .ni <- nestingInfo_(omega, ev)

      expect_equal(.ni$below, c(eye = 2L, occ = 2L))
      expect_equal(.ni$above, c(inv = 2L))
      expect_s3_class(.ni$data$eye, "factor")
      expect_equal(attr(.ni$data$eye, "nu"), 40L)
      expect_s3_class(.ni$data$inv, "factor")
      expect_equal(attr(.ni$data$inv, "nu"), NULL)
      expect_s3_class(.ni$data$occ, "factor")
      expect_equal(attr(.ni$data$occ, "nu"), 40L)

      expect_equal(.ni$extraTheta, 4)
      expect_equal(.ni$extraEta, 8)

      .en <- rxExpandNesting(mod, .ni, compile = TRUE)

      .ett <- etTrans(.ni$data, .en$mod)

      theta <- c(
        KA = 2.94E-01, CL = 1.86E+01, V2 = 4.02E+01, # central
        Q = 1.05E+01, V3 = 2.97E+02, # peripheral
        Kin = 1, Kout = 1, EC50 = 200
      ) # effects

      thetaMat <- lotri(
        KA ~ 0.01,
        CL ~ 0.01,
        V2 ~ 0.01,
        Q ~ 0.01,
        V3 ~ 0.01,
        Kin ~ 0.01,
        Kout ~ 0.01,
        EC50 ~ 0.01
      )

      .ep <- .expandPars(mod, theta, ev,
                             control = rxControl(
                               thetaMat = thetaMat, omega = omega,
                               nSub = 40, nStud = 3))

      expect_equal(length(.ep$KA), 120L)
      expect_equal(length(unique(.ep$KA)), 3L)

      .ep <- .expandPars(mod, theta, ev,
                             control = rxControl(
                               thetaMat = thetaMat, omega = omega,
                               nStud = 3
                             )
                             )

      expect_equal(length(.rxModels[[".thetaL"]]), 3L)
      expect_equal(length(.rxModels[[".omegaL"]]), 3L)
      expect_equal(.rxModels[[".sigmaL"]], NULL)
      expect_equal(length(.ep$KA), 60L)
      expect_equal(length(unique(.ep$KA)), 3L)
      expect_true(any(names(.ep) == "eta.Cl"))

      .ep <- .expandPars(mod, theta, ev,
                             control = rxControl(
                               thetaMat = thetaMat, omega = omega,
                               nStud = 3, nSub = 20
                             )
                             )

      expect_equal(length(.rxModels[[".thetaL"]]), 3L)
      expect_equal(length(.rxModels[[".omegaL"]]), 3L)
      expect_equal(.rxModels[[".sigmaL"]], NULL)
      expect_equal(length(.ep$KA), 60L)
      expect_true(any(names(.ep) == "eta.Cl"))

      .ep <- .expandPars(mod, theta, ev,
                             control = rxControl(
                               thetaMat = thetaMat, omega = omega,
                               sigma = lotri(prop.err ~ 0.1), dfObs = 10,
                               nStud = 3, nSub = 20
                             )
                             )

      expect_equal(length(.rxModels[[".thetaL"]]), 3L)
      expect_equal(length(.rxModels[[".omegaL"]]), 3L)
      expect_equal(length(.rxModels[[".sigmaL"]]), 3L)
      expect_equal(length(.ep$KA), 60L)
      expect_true(any(names(.ep) == "eta.Cl"))

      .ep <- .expandPars(mod, theta, ev,
                             control = rxControl(
                               thetaMat = thetaMat,
                               sigma = lotri(prop.err ~ 0.1), dfObs = 10,
                               nStud = 3, nSub = 20
                             )
                             )

      expect_equal(.rxModels[[".thetaL"]], NULL)
      expect_equal(.rxModels[[".omegaL"]], NULL)
      expect_equal(length(.rxModels[[".sigmaL"]]), 3L)
      expect_equal(length(.ep$KA), 60L)
      expect_false(any(names(.ep) == "eta.Cl"))


      .ep <- .expandPars(mod, theta, ev,
                             control = rxControl(
                               sigma = lotri(prop.err ~ 0.1), dfObs = 10,
                               nStud = 3, nSub = 20
                             )
                             )

      expect_equal(.rxModels[[".thetaL"]], NULL)
      expect_equal(.rxModels[[".omegaL"]], NULL)
      expect_equal(length(.rxModels[[".sigmaL"]]), 3L)
      expect_equal(length(.ep$KA), 60L)
      expect_false(any(names(.ep) == "eta.Cl"))

      .ep <- .expandPars(mod, theta, ev,
                             control = rxControl(
                               sigma = lotri(prop.err ~ 0.1), dfObs = 10,
                               nStud = 3, nSub = 20
                             )
                             )

      expect_equal(.rxModels[[".thetaL"]], NULL)
      expect_equal(.rxModels[[".omegaL"]], NULL)
      expect_equal(length(.rxModels[[".sigmaL"]]), 3L)
      expect_equal(length(.ep$KA), 60L)
      expect_false(any(names(.ep) == "eta.Cl"))

      .ep <- .expandPars(mod, theta, ev,
                             control = rxControl(
                               omega = lotri(eta.Cl ~ 0.1), dfObs = 10,
                               nStud = 3, nSub = 20
                             )
                             )

      expect_equal(.rxModels[[".thetaL"]], NULL)
      expect_equal(.rxModels[[".omegaL"]], NULL)
      expect_equal(.rxModels[[".sigmaL"]], NULL)
      expect_equal(length(.ep$KA), 60L)
      expect_true(any(names(.ep) == "eta.Cl"))

      .ep <- .expandPars(mod, theta, ev,
                             control = rxControl(dfObs = 10, nStud = 3, nSub = 4)
                             )

      expect_equal(.rxModels[[".thetaL"]], NULL)
      expect_equal(.rxModels[[".omegaL"]], NULL)
      expect_equal(.rxModels[[".sigmaL"]], NULL)
      expect_equal(length(.ep$KA), 12L)
      expect_false(any(names(.ep) == "eta.Cl"))


      expect_error(.expandPars(mod, NULL, ev,
                                   control = rxControl(thetaMat = thetaMat, omega = omega, nStud = 3)
                                   ))

      .ep <- .expandPars(mod, NULL, ev,
                             control = rxControl(omega = omega, nStud = 3)
                             )

      expect_equal(length(.rxModels[[".thetaL"]]), 3L)
      expect_equal(length(.rxModels[[".omegaL"]]), 3L)
      expect_equal(.rxModels[[".sigmaL"]], NULL)
      expect_equal(length(.ep$eta.Ka), 60L)
      expect_true(any(names(.ep) == "eta.Cl"))

      .ep <- .expandPars(mod, NULL, ev,
                             control = rxControl(
                               omega = omega,
                               nStud = 3, dfObs = 100, nSub = 20, dfSub = 10
                             )
                             )

      expect_equal(length(.rxModels[[".thetaL"]]), 3L)
      expect_equal(length(.rxModels[[".omegaL"]]), 3L)
      expect_equal(.rxModels[[".sigmaL"]], NULL)
      expect_equal(length(.ep$eta.Ka), 60L)
      expect_true(any(names(.ep) == "eta.Cl"))

      .ep <- .expandPars(mod, theta, ev,
                             control = rxControl(
                               thetaMat = lotri(KA ~ 1, CL ~ 1),
                               omega = omega,
                               sigma = lotri(prop.err ~ 0.1), dfObs = 10,
                               nStud = 3, nSub = 20
                             )
                             )

      ## Test edge case -- no between or above occasion variability

      .ni <- nestingInfo_(
        lotri(lotri(eta.Cl ~ 0.1, eta.Ka ~ 0.1) | id(nu = 100)),
        ev
      )

      expect_equal(.ni$above, structure(integer(0), names = character(0)))
      expect_equal(.ni$below, structure(integer(0), names = character(0)))
      expect_equal(.ni$idName, "id")
      expect_s3_class(.ni$omega, "lotri")
      expect_equal(names(.ni$omega), "id")

      .en <- rxExpandNesting(mod, .ni)
  })

  test_that("nesting test from https://github.com/nlmixr2/rxode2random/issues/25", {

    mod <- rxode2({
      TABS = TV_TABS * exp(eta.TABS + iov.TABS)
      TR_Fbio = TV_TR_Fbio + eta.TR_Fbio + iov.TR_Fbio
      CL = TV_CL * exp(eta.CL)
      V1 = TV_V1 * exp(eta.V1)
      V2 = TV_V2 * exp(eta.V2)
      CLD = TV_CLD * exp(eta.CLD)
      KA = log(2) / (TABS/60)
      FBIO = 1 / (exp(-TR_Fbio) + 1)
      DC1 = AMT1/V1
      DC2 = AMT2/V2
      d/dt(AMTa) =       -KA * AMTa
      d/dt(AMT1) = FBIO * KA * AMTa - CLD * DC1 + CLD * DC2 - CL * DC1
      d/dt(AMT2) =                  + CLD * DC1 - CLD * DC2
      d/dt(AUC) = DC1
    })

    n <- 10

    theta <- c("TV_TABS" = 45,
              "TV_TR_Fbio" = logit(x = 0.85),
              "TV_CL" = 10,
              "TV_V1" = 10,
              "TV_V2" = 65,
              "TV_CLD" = 25)


    omega <- lotri::lotri(
      lotri::lotri(eta.TABS~0.25,
                   eta.TR_Fbio~0.20,
                   eta.CL~0.30,
                   eta.V1~0.30,
                   eta.V2~0.45,
                   eta.CLD~0.15) | id(nu=n),
      lotri::lotri(iov.TABS~0.15,
                   iov.TR_Fbio~0.15) | occ(nu=n*2))

    dosing <- et(amt=1000,
                addl=6,
                ii=24,
                evid=1,
                cmt="AMTa",
                time=0) |>
      et(amt=1000,
         addl=6,
         ii=24,
         evid=4,
         cmt="AMTa",
         time = 336) |>
      et(seq(0,168,0.5)) |>
      et(seq(336,672,0.5)) |>
      et(id=seq(1,n))

    dosing <- dplyr::mutate(dosing, occ = 1) |>
      dplyr::mutate(occ = ifelse(time>=336,2,occ))

    expect_error(rxSolve(object = mod,
                             theta,
                             omega=omega,
                             ev=dosing,
                             nDisplayProgress=100L), NA)
  })

  test_that("iov curEval", {

    one.cmt <- function() {
      ini({
        ## You may label each parameter with a comment
        tka <- 0.45 # Log Ka
        tcl <- log(c(0, 2.7, 100)) # Log Cl
        ## This works with interactive models
        ## You may also label the preceding line with label("label text")
        tv <- 3.45; label("log V")
        ## the label("Label name") works with all models
        eta.ka ~ 0.6
        eta.cl ~ 0.3
        eta.v ~ 0.1
        iov.cl ~ 0.1 | occ
        add.sd <- 0.7
      })
      model({
        ka <- exp(tka + eta.ka)
        cl <- exp(tcl + eta.cl + iov.cl)
        v <- exp(tv + eta.v)
        linCmt() ~ add(add.sd)
      })
    }

    f <- rxode2(one.cmt)

    curEval <- f$muRefCurEval

    expect_equal(curEval[curEval$parameter == "iov.cl", "curEval"],
                 "exp")

  })

  test_that("iov simulation with id()/occ() nesting (issue #323)", {

    # Classic-style event table with an occ column added by assignment;
    # this combination of id(nu=)/occ(nu=) nesting previously segfaulted
    # in the nesting expansion ("Not compatible with STRSXP: [type=NULL]").

    model <- rxode2({
      KA <- KA * exp(KA_eta)
      V  <- V * exp(V_eta)
      CL <- CL * exp(CL_eta + iov.cl)
      cp = A2 / V * (1 + eps)
      d/dt(A1) <- -KA * A1
      d/dt(A2) <-  KA * A1 - CL / V * A2
    })

    ev <- et()
    ev$add.dosing(dose = 100000, nbr.doses = 1, dosing.to = 1, start.time = 0)
    ev$add.dosing(dose = 100000, nbr.doses = 1, dosing.to = 1, start.time = 100)
    ev$add.sampling(0:200)
    ev$occ <- ifelse(as.numeric(ev$time) < 100, 1, 2)

    params <- c(KA = 1, CL = 10, V = 100)

    omega <- lotri(lotri(KA_eta + V_eta + CL_eta ~ c(0.1,
                                                     0.01, 0.1,
                                                     0.01, 0.01, 0.1)) | id(nu = 100),
                   lotri(iov.cl ~ 0.1) | occ(nu = 200))
    sigma <- lotri(eps ~ 0.2)

    set.seed(42)
    expect_error(
      sim <- rxSolve(model, events = ev, params = params,
                     sigma = sigma, omega = omega, nSub = 5),
      NA)

    # each subject should have exactly one iov.cl per occasion (two total)
    .d <- as.data.frame(sim)
    .nOcc <- tapply(.d$iov.cl, .d$sim.id, function(x) length(unique(x)))
    expect_true(all(.nOcc == 2))
  })
})

rxTest({
  # These functions are taken from TruncatedNormal for testing
  .rx <- loadNamespace("rxode2")

  lnNpr <-
    function(a, b) { ## computes ln(P(a<Z<b))
      ## where Z~N(0,1) very accurately for any 'a', 'b'
      p <- rep(0, length(a))
      ## case b>a>0
      I <- a > 0
      if (any(I)) {
        pa <- pnorm(a[I], lower.tail = FALSE, log.p = TRUE)
        pb <- pnorm(b[I], lower.tail = FALSE, log.p = TRUE)
        p[I] <- pa + log1p(-exp(pb - pa))
      }
      ## case a<b<0
      idx <- b < 0
      if (any(idx)) {
        pa <- pnorm(a[idx], log.p = TRUE)
        pb <- pnorm(b[idx], log.p = TRUE)
        p[idx] <- pb + log1p(-exp(pa - pb))
      }
      ## case a<0<b
      I <- !I & !idx
      if (any(I)) {
        pa <- pnorm(a[I])
        pb <- pnorm(b[I], lower.tail = FALSE)
        p[I] <- log1p(-pa - pb)
      }
      return(p)
    }

  cholperm <-
    function(Sig, l, u) {
      ##  Computes permuted lower Cholesky factor L for Sig
      ##  by permuting integration limit vectors l and u.
      ## Outputs perm, such that Sig(perm,perm)=L%*%t(L).
      ##
      ## Reference:
      ##  Gibson G. J., Glasbey C. A., Elston D. A. (1994),
      ##  "Monte Carlo evaluation of multivariate normal integrals and
      ##  sensitivity to variate ordering",
      ##  In: Advances in Numerical Methods and Applications, pages 120--126
      eps <- 10^-10 # round-off error tolerance
      d <- length(l)
      perm <- 1:d # keep track of permutation
      L <- matrix(0, d, d)
      z <- rep(0, d)
      for (j in 1:d) {
        pr <- rep(Inf, d) # compute marginal prob.
        I <- j:d # search remaining dimensions
        D <- diag(Sig)
        if (j > 2) {
          s <- D[I] - L[I, 1:(j - 1)]^2 %*% rep(1, j - 1)
        } else if (j == 2) {
          s <- D[I] - L[I, 1]^2
        } else {
          s <- D[I]
        }
        s[s < 0] <- eps
        s <- sqrt(s)
        if (j > 2) {
          cols <- L[I, 1:(j - 1)] %*% z[1:(j - 1)]
        } else if (j == 2) {
          cols <- L[I, 1] * z[1]
        } else {
          cols <- 0
        }
        tl <- (l[I] - cols) / s
        tu <- (u[I] - cols) / s
        pr[I] <- lnNpr(tl, tu)
        # find smallest marginal dimension
        k <- which.min(pr)
        # flip dimensions k-->j
        jk <- c(j, k)
        kj <- c(k, j)
        Sig[jk, ] <- Sig[kj, ]
        Sig[, jk] <- Sig[, kj] # update rows and cols of Sig
        L[jk, ] <- L[kj, ] # update only rows of L
        l[jk] <- l[kj]
        u[jk] <- u[kj] # update integration limits
        perm[jk] <- perm[kj] # keep track of permutation
        # construct L sequentially via Cholesky computation
        s <- Sig[j, j] - sum(L[j, 1:(j - 1)]^2)
        if (s < (-0.001)) {
          stop("Sigma is not positive semi-definite")
        }
        s[s < 0] <- eps
        L[j, j] <- sqrt(s)
        if (j < d) {
          if (j > 2) {
            L[(j + 1):d, j] <- (Sig[(j + 1):d, j] - L[(j + 1):d, 1:(j - 1)] %*% L[j, 1:(j - 1)]) / L[j, j]
          } else if (j == 2) {
            L[(j + 1):d, j] <- (Sig[(j + 1):d, j] - L[(j + 1):d, 1] * L[j, 1]) / L[j, j]
          } else if (j == 1) {
            L[(j + 1):d, j] <- Sig[(j + 1):d, j] / L[j, j]
          }
        }
        ## find mean value, z(j), of truncated normal:
        tl <- (l[j] - L[j, 1:j] %*% z[1:j]) / L[j, j]
        tu <- (u[j] - L[j, 1:j] %*% z[1:j]) / L[j, j]
        w <- lnNpr(tl, tu) # aids in computing expected value of trunc. normal
        z[j] <- (exp(-.5 * tl^2 - w) - exp(-.5 * tu^2 - w)) / sqrt(2 * pi)
      }
      return(list(L = L, l = l, u = u, perm = perm))
    }

  gradpsi <-
    function(y, L, l, u) { # implements grad_psi(x) to find optimal exponential twisting;
      # assume scaled 'L' with zero diagonal;
      d <- length(u)
      c <- rep(0, d)
      x <- c
      mu <- c
      x[1:(d - 1)] <- y[1:(d - 1)]
      mu[1:(d - 1)] <- y[d:(2 * d - 2)]
      # compute now ~l and ~u
      c[-1] <- L[-1, ] %*% x
      lt <- l - mu - c
      ut <- u - mu - c
      # compute gradients avoiding catastrophic cancellation
      w <- lnNpr(lt, ut)
      pl <- exp(-0.5 * lt^2 - w) / sqrt(2 * pi)
      pu <- exp(-0.5 * ut^2 - w) / sqrt(2 * pi)
      P <- pl - pu
      # output the gradient
      dfdx <- -mu[-d] + t(t(P) %*% L[, -d])
      dfdm <- mu - x + P
      grad <- c(dfdx, dfdm[-d])
      # here compute Jacobian matrix
      lt[is.infinite(lt)] <- 0
      ut[is.infinite(ut)] <- 0
      dP <- (-P^2) + lt * pl - ut * pu # dPdm
      DL <- rep(dP, 1, d) * L
      mx <- -diag(d) + DL
      xx <- t(L) %*% DL
      mx <- mx[1:(d - 1), 1:(d - 1)]
      xx <- xx[1:(d - 1), 1:(d - 1)]
      if (d > 2) {
        Jac <- rbind(cbind(xx, t(mx)), cbind(mx, diag(1 + dP[1:(d - 1)])))
      } else {
        Jac <- rbind(cbind(xx, t(mx)), cbind(mx, 1 + dP[1:(d - 1)]))
        dimnames(Jac) <- NULL
      }
      f <- list(grad = grad, Jac = Jac)
    }

  nleq <-
    function(l, u, L) {
      d <- length(l)
      x <- rep(0, 2 * d - 2) # initial point for Newton iteration
      err <- Inf
      iter <- 0
      while (err > 10^-10) {
        f <- gradpsi(x, L, l, u)
        Jac <- f$Jac
        grad <- f$grad
        del <- solve(Jac, -grad) # Newton correction
        x <- x + del
        err <- sum(grad^2)
        iter <- iter + 1
        if (iter > 100) {
          stop("Covariance matrix is ill-conditioned and method failed")
        }
      }
      return(x)
    }

  test_that("cholperm", {

    rxWithSeed(12, {
      d <- 5

      mu <- 1:d

      ## Creating covariance matrix
      tmp <- matrix(rnorm(d^2), d, d)
      mcov <- tcrossprod(tmp, tmp)

      r1 <- cholperm(mcov, -(1:5), 1:5)
      r2 <- .rx$rxCholperm(mcov, -(1:5), 1:5)

      expect_equal(r1$L, r2$L)
      expect_equal(r1$l, r2$l)
      expect_equal(r1$u, r2$u)
      expect_equal(r1$perm, r2$perm + 1)

      r1 <- cholperm(mcov, 1:5, 2 * (1:5))
      r2 <- .rx$rxCholperm(mcov, 1:5, 2 * 1:5)

      expect_equal(r1$L, r2$L)
      expect_equal(r1$l, r2$l)
      expect_equal(r1$u, r2$u)
      expect_equal(r1$perm, r2$perm + 1)

      r1 <- cholperm(mcov, -2 * (1:5), -(1:5))
      r2 <- .rx$rxCholperm(mcov, -2 * (1:5), -(1:5))

      expect_equal(r1$L, r2$L)
      expect_equal(r1$l, r2$l)
      expect_equal(r1$u, r2$u)
      expect_equal(r1$perm, r2$perm + 1)

      ## microbenchmark::microbenchmark(cholperm(mcov, -2 * (1:5), -(1:5)), rxCholperm(mcov, -2 * (1:5), -(1:5)))
      ## microbenchmark::microbenchmark(microbenchmark::cholperm(mcov, -2 * (1:5), -(1:5)), rxCholperm(mcov, -2 * (1:5), -(1:5)))
    })

  })

  test_that("gradpsi", {

    rxWithSeed(12, {
      d <- 5

      mu <- 1:d

      ## Creating covariance matrix
      tmp <- matrix(rnorm(d^2), d, d)
      mcov <- tcrossprod(tmp, tmp)
      r2 <- .rx$rxCholperm(mcov, -(1:5), 1:5)

      d <- length(r2$l)
      y <- seq(0, 2 * d - 2)
      l <- r2$l
      u <- r2$u
      L <- r2$L

      r1 <- gradpsi(y, L, l, u)
      r2 <- .rx$rxGradpsi(y, L, l, u)

      expect_equal(r1$Jac, r2$Jac)
      expect_equal(r1$grad, r2$grad)

      r1 <- gradpsi(rep(1, 2 * d - 2), L, l, u)
      r2 <- .rx$rxGradpsi(rep(1, 2 * d - 2), L, l, u)

      expect_equal(r1$Jac, r2$Jac)
      expect_equal(r1$grad, r2$grad)

      r1 <- gradpsi(rep(-1, 2 * d - 2), L, l, u)
      r2 <- .rx$rxGradpsi(rep(-1, 2 * d - 2), L, l, u)

      expect_equal(r1$Jac, r2$Jac)
      expect_equal(r1$grad, r2$grad)

      ## microbenchmark::microbenchmark(gradpsi(rep(-1,2*d-2), L, l, u), .rx$rxGradpsi(rep(-1,2*d-2), L, l, u));

      d <- 2

      mu <- 1:d

      ## Creating covariance matrix
      tmp <- matrix(rnorm(d^2), d, d)
      mcov <- tcrossprod(tmp, tmp)
      r2 <- .rx$rxCholperm(mcov, -(1:d), 1:d)

      d <- length(r2$l)
      y <- seq(0, 2 * d - 2)
      l <- r2$l
      u <- r2$u
      L <- r2$L


      r1 <- gradpsi(rep(-1, 2 * d - 2), L, l, u)
      r2 <- .rx$rxGradpsi(rep(-1, 2 * d - 2), L, l, u)

      expect_equal(r1$Jac, r2$Jac)
      expect_equal(r1$grad, r2$grad)
    })

  })


  test_that("nleq", {

    rxWithSeed(12, {
      d <- 5

      mu <- 1:d

      ## Creating covariance matrix
      tmp <- matrix(rnorm(d^2), d, d)
      mcov <- tcrossprod(tmp, tmp)
      r2 <- .rx$rxCholperm(mcov, -3 * (1:d), 2 * (1:d))

      expect_equal(.rx$rxNleq(r2$l, r2$u, r2$L), nleq(r2$l, r2$u, r2$L))
      ## microbenchmark::microbenchmark(rxNleq(r2$l, r2$u, r2$L), nleq(r2$l, r2$u, r2$L))

      d <- 2

      mu <- 1:d

      ## Creating covariance matrix
      tmp <- matrix(rnorm(d^2), d, d)
      mcov <- tcrossprod(tmp, tmp)
      r2 <- .rx$rxCholperm(mcov, -2 * (1:d), 3 * 1:d)

      expect_equal(.rx$rxNleq(r2$l, r2$u, r2$L), nleq(r2$l, r2$u, r2$L))
    })

  })

  test_that("rxMvnrnd", {

    rxWithSeed(12, {
      d <- 5

      mu <- 1:d

      ## Creating covariance matrix
      tmp <- matrix(rnorm(d^2), d, d)
      mcov <- tcrossprod(tmp, tmp)

      out <- .rx$rxCholperm(mcov, -2 * (1:5), 1:5)

      Lfull <- out$L
      l <- out$l
      u <- out$u
      D <- diag(Lfull)
      perm <- out$perm
      if (any(D < 10^-10)) {
        warning("Method may fail as covariance matrix is singular!")
      }
      L <- Lfull / D
      u <- u / D
      l <- l / D # rescale
      L <- L - diag(d) # remove diagonal
      # find optimal tilting parameter via non-linear equation solver
      xmu <- nleq(l, u, L) # nonlinear equation solver
      x <- xmu[1:(d - 1)]
      mu <- xmu[d:(2 * d - 2)] # assign saddlepoint x* and mu*

      fun <- function(n) {
        r1 <- .rx$rxMvnrnd(5, L, l, u, mu)
        expect_equal(length(r1$logpr), 5)
        expect_true(all(!duplicated(r1$logpr)))
        expect_equal(length(r1$Z[1, ]), 5)
        expect_true(all(!duplicated(r1$Z[1, ])))
      }
      fun(2)
      fun(5)
      fun(10)
    })

  })
})

rxTest({
  # Removing dll
  skip_on_os("windows")
  ode <- rxode2({
    b <- -1
    d / dt(X) <- a * X + Y * Z
    d / dt(Y) <- b * (Y - Z)
    d / dt(Z) <- -X * Y + c * Y - Z
  })

  dll <- rxDll(ode)

  test_that("dll exists", {
    expect_true(file.exists(dll))
    expect_true(file.exists(gsub("[.][^.]*$", ".c", dll)))
  })

  ode$delete()

  test_that("dll is removed", {
    expect_false(file.exists(dll))
    expect_false(file.exists(gsub("[.][^.]*$", ".c", dll)))
  })

  ode$load()

  test_that("dll exists #2", {
    expect_true(file.exists(dll))
    expect_true(file.exists(gsub("[.][^.]*$", ".c", dll)))
  })
})

#' @export
str.rxHidden <- function(object, ...) {
  cat("\r")
}


#' @importFrom utils str
#' @export
str.rxSolve <- function(object, ..., nchar.max = 128) {
  if (rxIs(object, "rxSolve")) {
    .dn <- .Call(`_rxode2_rxSolveDollarNames`, object)
    .max <- max(sapply(.dn, nchar))
    cat(sprintf(
      "Classes 'rxSolve' and 'data.frame':\t%s rows of  %s variables:\n",
      object$.check.nrow, object$.check.ncol
    ))
    for (.n in c(
      "sim.id", "id", "evid", "cmt", "ss", "amt", "rate", "dur",
      "ii"
    )) {
      if (any(names(object) == .n)) {
        cat(paste0(
          " $ ", .n,
          paste(rep(" ", .max - nchar(.n)), collapse = ""),
          ":"
        ))
        str(object[[.n]])
      }
    }
    cat(paste0(" $ time", paste(rep(" ", .max - 4), collapse = ""), ":"))
    str(object$time, nchar.max = nchar.max - .max - 8)
    .mv <- rxModelVars(object)
    if (length(.mv$lhs) > 0) {
      cat("Left Handed values ($lhs):\n")
      for (.l in .mv$lhs) {
        cat(paste0(
          sprintf(" $ %s", .l),
          paste(rep(" ", .max - nchar(.l)), collapse = ""),
          ":"
        ))
        str(object[[.l]], nchar.max = nchar.max - .max - nchar(.l) - 4)
      }
    }
    if (length(.mv$state) > 0) {
      cat("State values ($state):\n")
      for (.l in .mv$state) {
        cat(paste0(
          sprintf(" $ %s", .l),
          paste(rep(" ", .max - nchar(.l)), collapse = ""),
          ":"
        ))
        str(object[[.l]], nchar.max = nchar.max - .max - nchar(.l) - 4)
      }
      cat("State ini values:\n")
      for (.l in .mv$state) {
        cat(paste0(
          sprintf(" $ %s0", .l),
          paste(rep(" ", .max - nchar(.l) - 1), collapse = ""),
          ":"
        ))
        str(object[[paste0(.l, "0")]],
          nchar.max = nchar.max - .max - nchar(.l) - 5
        )
      }
    }
    if (length(.mv$params) > 0) {
      cat("Parameter values ($params):\n")
      for (.l in .mv$params) {
        cat(paste0(
          sprintf(" $ %s", .l),
          paste(rep(" ", .max - nchar(.l) - 1), collapse = ""),
          ":"
        ))
        str(object[[.l]], nchar.max = nchar.max - .max - nchar(.l) - 4)
      }
    }
    .vars <- c(
      "sim.id", "id", "evid", "cmt", "ss", "amt", "rate",
      "dur", "ii", "time", .mv$lhs, .mv$state
    )
    .n2 <- names(object)[!(names(object) %in% .vars)]
    if (length(.n2) > 0) {
      cat("Other Variables:\n")
      for (.l in .n2) {
        cat(paste0(
          sprintf(" $ %s", .l),
          paste(rep(" ", .max - nchar(.l)), collapse = ""),
          ":"
        ))
        str(object[[.l]], nchar.max = nchar.max - .max - nchar(.l) - 4)
      }
    }
    .dn <- .dn[!(.dn %in% c(
      names(object), paste0(.mv$state, "0"),
      "t", "params", "inits",
      .mv$params
    ))]
    .fns <- sapply(.dn, function(x) {
      inherits(`$.rxSolve`(object, x), "function")
    })
    .fns <- names(.fns[.fns])
    if (length(.fns) > 0) {
      cat("Functions:\n")
      for (.l in .fns) {
        cat(paste0(
          sprintf(" $ %s", .l),
          paste(rep(" ", .max - nchar(.l)), collapse = ""),
          ":"
        ))
        str(object[[.l]], nchar.max = nchar.max - .max - nchar(.l) - 4)
      }
    }
    .dn <- .dn[!(.dn %in% .fns)]
    if (length(.dn) > 0) {
      cat("Other:\n")
      for (.l in .dn) {
        cat(paste0(
          sprintf(" $ %s", .l),
          paste(rep(" ", .max - nchar(.l)), collapse = ""),
          ":"
        ))
        str(object[[.l]], nchar.max = nchar.max - .max - nchar(.l) - 4)
      }
    }
  } else {
    NextMethod()
  }
}

#' @export
str.rxEt <- function(object, ...) {
  cat("rxEt methods and properties:\n")
  cat(" $ get.EventTable   :function ()\n")
  cat(" $ get.obs.rec      :function ()  \n")
  cat(" $ get.nobs         :function ()  \n")
  cat(" $ add.dosing       :function ()  \n")
  cat(" $ clear.dosing     :function ()  \n")
  cat(" $ get.dosing       :function ()  \n")
  cat(" $ add.sampling     :function ()  \n")
  cat(" $ clear.sampling   :function ()  \n")
  cat(" $ get.sampling     :function ()  \n")
  cat(" $ get.units        :function ()  \n")
  cat(" $ import.EventTable:function ()  \n")
  cat(" $ copy             :function ()  \n")
  cat(" $ expand           :function ()  \n")
  return(invisible(NextMethod("str", ...)))
}

#' @export
str.rxSymInvCholEnv <- function(object, ...) {
  cat("Derivatives and Inverse of a matrix; Assigning theta will change these values.\n")
  cat(" $ theta             : Current parameters (on inverse Cholesky)\n")
  cat(" $ ntheta            : Number of parameters\n")
  cat(" $ chol.omegaInv     : chol(Omega^-1)\n")
  cat(" $ omegaInv          : Omega^-1\n")
  cat(" $ d.omegaInv        : d(Omega^-1)\n")
  cat(" $ d.D.omegaInv      : gives the d(diagonal(Omega^-1))\n")
  cat(" $ chol.omega        : chol(Omega)\n")
  cat(" $ omega             : Omega\n")
  cat(" $ log.det.OMGAinv.5 : log(det(Omega^-1))\n")
  cat(" $ tr.28             : -0.5*tr(Omega^-1 %*% d(Omega)) = 0.5*tr(d(Omega^-1) %*% Omega); (Almquist 2015 #28)\n")
  cat(" $ omega.47          : d(Omega^-1)*d(eta) (Almquist 2015 #47)\n")
  cat(" $ theta.diag        : indicator of diagonal theta values\n")
}

#' @export
str.rxode2 <- function(object, ...) {
  cat("rxode2 object methods and properties:\n")
  cat(" $ assignPtr()    : Assign C pointers\n")
  cat(" $ compile()      : compile rxode2 model\n")
  cat(" $ delete()       : delete rxode2 dll\n")
  cat(" $ dynLoad()      : load dll for rxode2 model\n")
  cat(" $ dynUnload()    : unload dll for rxode2 model\n")
  cat(" $ get.index(...) : Get compartment number\n")
  cat(" $ get.modelVars(): Get model variables\n")
  cat(" $ isLoaded()     : Is rxode2 model dll loaded\n")
  cat(" $ isValid()      : Is rxode2 model dll valid\n")
  cat(" $ load()         : Load rxode2 model\n")
  cat(" $ parse()        : Parse model (doesn't do anything anymore)\n")
  cat(" $ run(...)       : Run ODE model\n")
  cat(" $ solve(...)     : Solve ODE model\n")
  cat(" $ unload()       : Unload DLL for rxode2 model\n")
  .out <- utils::capture.output(utils::str(list(
    calcJac = object$calcJac,
    calcSens = object$calcSens,
    collapseModel = object$collapseModel,
    debug = object$debug,
    extraC = object$extraC,
    lhs = object$lhs,
    lib.name = object$lib.name,
    mdir = object$mdir,
    missing.modName = object$missing.modName,
    model = object$model,
    modName = object$modName,
    package = object$package,
    params = object$params,
    rxDll = object$rxDll,
    state = object$state,
    stateExtra = object$stateExtra,
    version = object$version,
    wd = object$wd
  )))
  .out <- .out[-1]
  sapply(.out, function(x) {
    cat(paste0(x, "\n"))
  })
  invisible()
}

!     MLF: Modification to return I when norm is 0
!-----------------------------------------------------------------------!
!                                                                       !
!     V A R I O U S    M A T H E M A T I C A L    U T I L I T I E S     !
!                                                                       !
!                       FORTRAN 77 PROCEDURES                           !
!-----------------------------------------------------------------------!

*     DGEMM -> DGEXX for LTO
* Idea comes from https://github.com/nmatzke/rexpokit/blob/2e7712fb5e6d3c9786e412a8b7f2ec7cced097ef/src/lapack/blas_mod.f#L1259-L1580
*----------------------------------------------------------------------|
      SUBROUTINE DGEXX ( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB,
     $                   BETA, C, LDC )
*     .. Scalar Arguments ..
      CHARACTER(LEN=1)   TRANSA, TRANSB
      INTEGER            M, N, K, LDA, LDB, LDC
      DOUBLE PRECISION   ALPHA, BETA
*     .. Array Arguments ..
*     2019-06-26_NJM: fix
*
* Error: Variable 'ldb' cannot appear in the expression at (1)
*
      DOUBLE PRECISION   A( LDA, * ), B( LDB, * ), C( LDC, * )

*
*     ..
*
*  Purpose
*  =======
*
*  DGEXX  performs one of the matrix-matrix operations
*
*     C := alpha*op( A )*op( B ) + beta*C,
*
*  where  op( X ) is one of
*
*     op( X ) = X   or   op( X ) = X',
*
*  alpha and beta are scalars, and A, B and C are matrices, with op( A )
*  an m by k matrix,  op( B )  a  k by n matrix and  C an m by n matrix.
*
*  Parameters
*  ==========
*
*  TRANSA - CHARACTER*1.
*           On entry, TRANSA specifies the form of op( A ) to be used in
*           the matrix multiplication as follows:
*
*              TRANSA = 'N' or 'n',  op( A ) = A.
*
*              TRANSA = 'T' or 't',  op( A ) = A'.
*
*              TRANSA = 'C' or 'c',  op( A ) = A'.
*
*           Unchanged on exit.
*
*  TRANSB - CHARACTER*1.
*           On entry, TRANSB specifies the form of op( B ) to be used in
*           the matrix multiplication as follows:
*
*              TRANSB = 'N' or 'n',  op( B ) = B.
*
*              TRANSB = 'T' or 't',  op( B ) = B'.
*
*              TRANSB = 'C' or 'c',  op( B ) = B'.
*
*           Unchanged on exit.
*
*  M      - INTEGER.
*           On entry,  M  specifies  the number  of rows  of the  matrix
*           op( A )  and of the  matrix  C.  M  must  be at least  zero.
*           Unchanged on exit.
*
*  N      - INTEGER.
*           On entry,  N  specifies the number  of columns of the matrix
*           op( B ) and the number of columns of the matrix C. N must be
*           at least zero.
*           Unchanged on exit.
*
*  K      - INTEGER.
*           On entry,  K  specifies  the number of columns of the matrix
*           op( A ) and the number of rows of the matrix op( B ). K must
*           be at least  zero.
*           Unchanged on exit.
*
*  ALPHA  - DOUBLE PRECISION.
*           On entry, ALPHA specifies the scalar alpha.
*           Unchanged on exit.
*
*  A      - DOUBLE PRECISION array of DIMENSION ( LDA, ka ), where ka is
*           k  when  TRANSA = 'N' or 'n',  and is  m  otherwise.
*           Before entry with  TRANSA = 'N' or 'n',  the leading  m by k
*           part of the array  A  must contain the matrix  A,  otherwise
*           the leading  k by m  part of the array  A  must contain  the
*           matrix A.
*           Unchanged on exit.
*
*  LDA    - INTEGER.
*           On entry, LDA specifies the first dimension of A as declared
*           in the calling (sub) program. When  TRANSA = 'N' or 'n' then
*           LDA must be at least  max( 1, m ), otherwise  LDA must be at
*           least  max( 1, k ).
*           Unchanged on exit.
*
*  B      - DOUBLE PRECISION array of DIMENSION ( LDB, kb ), where kb is
*           n  when  TRANSB = 'N' or 'n',  and is  k  otherwise.
*           Before entry with  TRANSB = 'N' or 'n',  the leading  k by n
*           part of the array  B  must contain the matrix  B,  otherwise
*           the leading  n by k  part of the array  B  must contain  the
*           matrix B.
*           Unchanged on exit.
*
*  LDB    - INTEGER.
*           On entry, LDB specifies the first dimension of B as declared
*           in the calling (sub) program. When  TRANSB = 'N' or 'n' then
*           LDB must be at least  max( 1, k ), otherwise  LDB must be at
*           least  max( 1, n ).
*           Unchanged on exit.
*
*  BETA   - DOUBLE PRECISION.
*           On entry,  BETA  specifies the scalar  beta.  When  BETA  is
*           supplied as zero then C need not be set on input.
*           Unchanged on exit.
*
*  C      - DOUBLE PRECISION array of DIMENSION ( LDC, n ).
*           Before entry, the leading  m by n  part of the array  C must
*           contain the matrix  C,  except when  beta  is zero, in which
*           case C need not be set on entry.
*           On exit, the array  C  is overwritten by the  m by n  matrix
*           ( alpha*op( A )*op( B ) + beta*C ).
*
*  LDC    - INTEGER.
*           On entry, LDC specifies the first dimension of C as declared
*           in  the  calling  (sub)  program.   LDC  must  be  at  least
*           max( 1, m ).
*           Unchanged on exit.
*
*
*  Level 3 Blas routine.
*
*  -- Written on 8-February-1989.
*     Jack Dongarra, Argonne National Laboratory.
*     Iain Duff, AERE Harwell.
*     Jeremy Du Croz, Numerical Algorithms Group Ltd.
*     Sven Hammarling, Numerical Algorithms Group Ltd.
*
*
*     .. External Functions ..
      LOGICAL            LSAMEX
      EXTERNAL           LSAMEX
*     .. External Subroutines ..
      EXTERNAL           XERBLA
*     .. Intrinsic Functions ..
      INTRINSIC          MAX
*     .. Local Scalars ..
      LOGICAL            NOTA, NOTB
      INTEGER            I, INFO, J, L, NCOLA, NROWA, NROWB
      DOUBLE PRECISION   TEMP
*     .. Parameters ..
      DOUBLE PRECISION   ONE         , ZERO
      PARAMETER        ( ONE = 1.0D+0, ZERO = 0.0D+0 )
*     ..
*     .. Executable Statements ..
*
*     Set  NOTA  and  NOTB  as  true if  A  and  B  respectively are not
*     transposed and set  NROWA, NCOLA and  NROWB  as the number of rows
*     and  columns of  A  and the  number of  rows  of  B  respectively.
*
      NOTA  = LSAMEX( TRANSA, 'N' )
      NOTB  = LSAMEX( TRANSB, 'N' )
      IF( NOTA )THEN
         NROWA = M
         NCOLA = K
      ELSE
         NROWA = K
         NCOLA = M
      END IF
      IF( NOTB )THEN
         NROWB = K
      ELSE
         NROWB = N
      END IF
*
*     Test the input parameters.
*
      INFO = 0
      IF(      ( .NOT.NOTA                 ).AND.
     $         ( .NOT.LSAMEX( TRANSA, 'C' ) ).AND.
     $         ( .NOT.LSAMEX( TRANSA, 'T' ) )      )THEN
         INFO = 1
      ELSE IF( ( .NOT.NOTB                 ).AND.
     $         ( .NOT.LSAMEX( TRANSB, 'C' ) ).AND.
     $         ( .NOT.LSAMEX( TRANSB, 'T' ) )      )THEN
         INFO = 2
      ELSE IF( M  .LT.0               )THEN
         INFO = 3
      ELSE IF( N  .LT.0               )THEN
         INFO = 4
      ELSE IF( K  .LT.0               )THEN
         INFO = 5
      ELSE IF( LDA.LT.MAX( 1, NROWA ) )THEN
         INFO = 8
      ELSE IF( LDB.LT.MAX( 1, NROWB ) )THEN
         INFO = 10
      ELSE IF( LDC.LT.MAX( 1, M     ) )THEN
         INFO = 13
      END IF
      IF( INFO.NE.0 )THEN
         CALL XERBLA( 'DGEXX ', INFO )
         RETURN
      END IF
*
*     Quick return if possible.
*
      IF( ( M.EQ.0 ).OR.( N.EQ.0 ).OR.
     $    ( ( ( ALPHA.EQ.ZERO ).OR.( K.EQ.0 ) ).AND.( BETA.EQ.ONE ) ) )
     $   RETURN
*
*     And if  alpha.eq.zero.
*
      IF( ALPHA.EQ.ZERO )THEN
         IF( BETA.EQ.ZERO )THEN
            DO 20, J = 1, N
               DO 10, I = 1, M
                  C( I, J ) = ZERO
   10          CONTINUE
   20       CONTINUE
         ELSE
            DO 40, J = 1, N
               DO 30, I = 1, M
                  C( I, J ) = BETA*C( I, J )
   30          CONTINUE
   40       CONTINUE
         END IF
         RETURN
      END IF
*
*     Start the operations.
*
      IF( NOTB )THEN
         IF( NOTA )THEN
*
*           Form  C := alpha*A*B + beta*C.
*
            DO 90, J = 1, N
               IF( BETA.EQ.ZERO )THEN
                  DO 50, I = 1, M
                     C( I, J ) = ZERO
   50             CONTINUE
               ELSE IF( BETA.NE.ONE )THEN
                  DO 60, I = 1, M
                     C( I, J ) = BETA*C( I, J )
   60             CONTINUE
               END IF
               DO 80, L = 1, K
                  IF( B( L, J ).NE.ZERO )THEN
                     TEMP = ALPHA*B( L, J )
                     DO 70, I = 1, M
                        C( I, J ) = C( I, J ) + TEMP*A( I, L )
   70                CONTINUE
                  END IF
   80          CONTINUE
   90       CONTINUE
         ELSE
*
*           Form  C := alpha*A'*B + beta*C
*
            DO 120, J = 1, N
               DO 110, I = 1, M
                  TEMP = ZERO
                  DO 100, L = 1, K
                     TEMP = TEMP + A( L, I )*B( L, J )
  100             CONTINUE
                  IF( BETA.EQ.ZERO )THEN
                     C( I, J ) = ALPHA*TEMP
                  ELSE
                     C( I, J ) = ALPHA*TEMP + BETA*C( I, J )
                  END IF
  110          CONTINUE
  120       CONTINUE
         END IF
      ELSE
         IF( NOTA )THEN
*
*           Form  C := alpha*A*B' + beta*C
*
            DO 170, J = 1, N
               IF( BETA.EQ.ZERO )THEN
                  DO 130, I = 1, M
                     C( I, J ) = ZERO
  130             CONTINUE
               ELSE IF( BETA.NE.ONE )THEN
                  DO 140, I = 1, M
                     C( I, J ) = BETA*C( I, J )
  140             CONTINUE
               END IF
               DO 160, L = 1, K
                  IF( B( J, L ).NE.ZERO )THEN
                     TEMP = ALPHA*B( J, L )
                     DO 150, I = 1, M
                        C( I, J ) = C( I, J ) + TEMP*A( I, L )
  150                CONTINUE
                  END IF
  160          CONTINUE
  170       CONTINUE
         ELSE
*
*           Form  C := alpha*A'*B' + beta*C
*
            DO 200, J = 1, N
               DO 190, I = 1, M
                  TEMP = ZERO
                  DO 180, L = 1, K
                     TEMP = TEMP + A( L, I )*B( J, L )
  180             CONTINUE
                  IF( BETA.EQ.ZERO )THEN
                     C( I, J ) = ALPHA*TEMP
                  ELSE
                     C( I, J ) = ALPHA*TEMP + BETA*C( I, J )
                  END IF
  190          CONTINUE
  200       CONTINUE
         END IF
      END IF
*
      RETURN
*
*     End of DGEXX .
*
      END
*----------------------------------------------------------------------|

*----------------------------------------------------------------------|
      LOGICAL          FUNCTION LSAMEX( CA, CB )
*
*  -- LAPACK auxiliary routine (version 1.1) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     February 29, 1992
*
*     .. Scalar Arguments ..
      CHARACTER          CA, CB
*     ..
*
*  Purpose
*  =======
*
*  LSAMEX returns .TRUE. if CA is the same letter as CB regardless of
*  case.
*
*  Arguments
*  =========
*
*  CA      (input) CHARACTER*1
*  CB      (input) CHARACTER*1
*          CA and CB specify the single characters to be compared.
*
*     .. Intrinsic Functions ..
      INTRINSIC          ICHAR
*     ..
*     .. Local Scalars ..
      INTEGER            INTA, INTB, ZCODE
*     ..
*     .. Executable Statements ..
*
*     Test if the characters are equal
*
      LSAMEX = CA.EQ.CB
      IF( LSAMEX )
     $   RETURN
*
*     Now test for equivalence if both characters are alphabetic.
*
      ZCODE = ICHAR( 'Z' )
*
*     Use 'Z' rather than 'A' so that ASCII can be detected on Prime
*     machines, on which ICHAR returns a value with bit 8 set.
*     ICHAR('A') on Prime machines returns 193 which is the same as
*     ICHAR('A') on an EBCDIC machine.
*
      INTA = ICHAR( CA )
      INTB = ICHAR( CB )
*
      IF( ZCODE.EQ.90 .OR. ZCODE.EQ.122 ) THEN
*
*        ASCII is assumed - ZCODE is the ASCII code of either lower or
*        upper case 'Z'.
*
         IF( INTA.GE.97 .AND. INTA.LE.122 ) INTA = INTA - 32
         IF( INTB.GE.97 .AND. INTB.LE.122 ) INTB = INTB - 32
*
      ELSE IF( ZCODE.EQ.233 .OR. ZCODE.EQ.169 ) THEN
*
*        EBCDIC is assumed - ZCODE is the EBCDIC code of either lower or
*        upper case 'Z'.
*
         IF( INTA.GE.129 .AND. INTA.LE.137 .OR.
     $       INTA.GE.145 .AND. INTA.LE.153 .OR.
     $       INTA.GE.162 .AND. INTA.LE.169 ) INTA = INTA + 64
         IF( INTB.GE.129 .AND. INTB.LE.137 .OR.
     $       INTB.GE.145 .AND. INTB.LE.153 .OR.
     $       INTB.GE.162 .AND. INTB.LE.169 ) INTB = INTB + 64
*
      ELSE IF( ZCODE.EQ.218 .OR. ZCODE.EQ.250 ) THEN
*
*        ASCII is assumed, on Prime machines - ZCODE is the ASCII code
*        plus 128 of either lower or upper case 'Z'.
*
         IF( INTA.GE.225 .AND. INTA.LE.250 ) INTA = INTA - 32
         IF( INTB.GE.225 .AND. INTB.LE.250 ) INTB = INTB - 32
      END IF
      LSAMEX = INTA.EQ.INTB
*
*     RETURN
*
*     End of LSAMEX
*
      END


!-----------------------------------------------------------------------!
      subroutine matexpRBS (ideg, m, t, H, iflag)

      IMPLICIT NONE
      INTEGER ideg, m, iflag
      DOUBLE PRECISION t, H(m,m)

!-----PURPOSE-----------------------------------------------------------!
!
!     COMPUTES EXP(T*H), THE MATRIX EXPONENTIAL OF A GENERAL MATRIX IN
!     FULL, USING THE IRREDUCIBLE RATIONAL PADE APPROXIMATION TO THE
!     EXPONENTIAL FUNCTION EXP(X) = R(X) = (+/-)( I + 2*(Q(X)/P(X)) ),
!     COMBINED WITH SCALING-AND-SQUARING.
!
!-----ARGUMENTS---------------------------------------------------------!
!
!     IDEG      : (INPUT) THE DEGREE OF THE DIAGONAL PADE TO BE USED.
!                 A VALUE OF 6 IS GENERALLY SATISFACTORY.
!
!     M         : (INPUT) ORDER OF H.
!
!     T         : (INPUT) TIME-SCALE (CAN BE < 0).
!
!     H(M,M)    : (INPUT) ARGUMENT MATRIX.
!
!     IFLAG     : (OUTPUT) EXIT FLAG.
!                      0 - NO PROBLEM
!                     <0 - PROBLEM
!
!-----------------------------------------------------------------------!
!     ROGER B. SIDJE (RBS@MATHS.UQ.EDU.AU) - 'RBS'
!     EXPOKIT: SOFTWARE PACKAGE FOR COMPUTING MATRIX EXPONENTIALS.
!     ACM - TRANSACTIONS ON MATHEMATICAL SOFTWARE, 24(1):130-156, 1998
!-----------------------------------------------------------------------!
!     MODIFIED TO RETURN A FLAG INSTEAD OF TERMINATING, WHEN TRYING
!     TO COMPUTE THE EXPONENTIAL OF A MATRIX WITH TOO LARGE ELEMENTS.
!
!     1) NIELS RODE KRISTENSEN, TECHNICAL UNIVERSITY OF DENMARK, 2000
!     2) ANDREAS S. CHRISTENSEN,TECHNICAL UNIVERSITY OF DENMARK, 2006
!     3) SOREN KLIM, IMM-DTU, 2007
!-----------------------------------------------------------------------!
!
      INTEGER  LWSP, NS, IPIV(M)
      DOUBLE PRECISION WSP(4*M*M+IDEG+1)
      INTEGER MM,I,J,K,IH2,IP,IQ,IUSED,IFREE,IODD,ICOEF,IPUT,IGET
      DOUBLE PRECISION HNORM,SCALE,SCALE2,CP,CQ

! "External" routines:
      INTRINSIC INT,ABS,DBLE,LOG,MAX
!---  CHECK RESTRICTIONS ON INPUT PARAMETERS ...
      MM = M*M
      IFLAG = 0
      LWSP = 4*M*M + IDEG +1
!
!---  INITIALISE POINTERS ...
!
      ICOEF = 1
      IH2 = ICOEF + (IDEG+1)
      IP  = IH2 + MM
      IQ  = IP + MM
      IFREE = IQ + MM

      NS=0

!
!---  INITIALISE ARRAYS ...
!
      DO I = 1, LWSP
         WSP(I) = 0.0D0
      ENDDO

      DO I = 1, M
         IPIV(I) = 0
      ENDDO

!
!---  SCALING: SEEK NS SUCH THAT ||T*H/2^NS|| < 1/2;
!     AND SET SCALE = T/2^NS ...
!

      DO J = 1,M
         DO I = 1,M
            WSP(I) = WSP(I) + ABS( H(I,J) )
         ENDDO
      ENDDO
      HNORM = 0.0D0
      DO I = 1,M
         HNORM = MAX( HNORM,WSP(I) )
      ENDDO
      HNORM = ABS( T*HNORM )
      IF (HNORM .EQ. 0.D0) THEN
! Deviation from original algorithm, use identity
         DO I = 1, M
            H(I,I)=1.D0
         ENDDO
         GOTO 200
      ENDIF
      NS = MAX(0, INT(LOG(HNORM)/LOG(2.))+2)
      SCALE = T / DBLE(2**NS)
      SCALE2 = SCALE*SCALE
!
!---  COMPUTE PADE COEFFICIENTS ...
!
      I = IDEG+1
      J = 2*IDEG+1
      WSP(ICOEF) = 1.0D0
      DO K = 1,IDEG
         WSP(ICOEF+K) = (WSP(ICOEF+K-1)*DBLE( I-K ))/DBLE( K*(J-K) )
      ENDDO
!
!---  H2 = SCALE2*H*H ...
!
      CALL DGEXX( 'N','N',M,M,M,SCALE2,H,M,H,M,0.0D0,WSP(IH2),M )
!
!---  INITIALIZE P (NUMERATOR) AND Q (DENOMINATOR) ...
!
      CP = WSP(ICOEF+IDEG-1)
      CQ = WSP(ICOEF+IDEG)
      DO J = 1,M
         DO I = 1,M
            WSP(IP + (J-1)*M + I-1) = 0.0D0
            WSP(IQ + (J-1)*M + I-1) = 0.0D0
         ENDDO
         WSP(IP + (J-1)*(M+1)) = CP
         WSP(IQ + (J-1)*(M+1)) = CQ
      ENDDO
!
!---  APPLY HORNER RULE ...
!
      IODD = 1
      K = IDEG - 1
 100  CONTINUE
      IUSED = IODD*IQ + (1-IODD)*IP
      CALL DGEXX( 'N','N',M,M,M, 1.0D0,WSP(IUSED),M,
     .             WSP(IH2),M, 0.0D0,WSP(IFREE),M )
      DO J = 1,M
         WSP(IFREE+(J-1)*(M+1)) = WSP(IFREE+(J-1)*(M+1))+WSP(ICOEF+K-1)
      ENDDO
      IP = (1-IODD)*IFREE + IODD*IP
      IQ = IODD*IFREE + (1-IODD)*IQ
      IFREE = IUSED
      IODD = 1-IODD
      K = K-1
      IF ( K.GT.0 )  GOTO 100
!
!---  OBTAIN (+/-)(I + 2*(P\Q)) ...
!
      IF ( IODD .EQ. 1 ) THEN
         CALL DGEXX( 'N','N',M,M,M, SCALE,WSP(IQ),M,
     .                H,M, 0.0D0,WSP(IFREE),M )
         IQ = IFREE
      ELSE
         CALL DGEXX( 'N','N',M,M,M, SCALE,WSP(IP),M,
     .                H,M, 0.0D0,WSP(IFREE),M )
         IP = IFREE
      ENDIF
      CALL DAXPY( MM, -1.0D0,WSP(IP),1, WSP(IQ),1 )
      CALL DGESV( M,M, WSP(IQ),M, IPIV, WSP(IP),M, IFLAG )
      IF ( IFLAG.NE.0 ) CALL RWarn ('PROBLEM IN DGESV (WITHIN DGPADM)')
      CALL DSCAL( MM, 2.0D0, WSP(IP), 1 )
      DO J = 1,M
         WSP(IP+(J-1)*(M+1)) = WSP(IP+(J-1)*(M+1)) + 1.0D0
      ENDDO
      IPUT = IP
      IF ( NS.EQ.0 .AND. IODD.EQ.1 ) THEN
         CALL DSCAL( MM, -1.0D0, WSP(IP), 1 )
      ELSE
!---  		SQUARING : EXP(T*H) = (EXP(T*H))^(2^NS) ...
!
         IODD = 1
         DO K = 1,NS
            IGET = IODD*IP + (1-IODD)*IQ
            IPUT = (1-IODD)*IP + IODD*IQ
            CALL DGEXX( 'N','N',M,M,M, 1.0D0,WSP(IGET),M, WSP(IGET),M,
     .           0.0D0,WSP(IPUT),M )
            IODD = 1-IODD
         ENDDO
      ENDIF

!---  COPY EXP(H*T) into H
      DO I= 1,M
         DO J=1,M
            H(I,J) = WSP(IPUT +(I-1) + (J-1)*M)
         ENDDO
      ENDDO
!-----------------------------------------------------------------------!
200   END

dfWishartCalcRse <- function(nu, omega, totN, rse, diag=TRUE) {
  .cv <- cvPost(nu, omega, totN)
  #.cv <- ivdoctr:::rinvwish(totN, nu, omega)
  #print(.cv)
  #max(omegaListRse(.cv)$rse)-rse
  if (diag) {
    mean(diag(omegaListRse(.cv)$rse))-rse
  } else {
    mean(as.vector(omegaListRse(.cv)$rse))-rse
  }
}

#' This uses simulations to match the rse
#'
#' @param omega represents the matrix for simulation
#' @param n This represents the number of subjects/samples this comes
#'   from (used to calculate rse).  When present it assumes the rse=
#'   sqrt(2)/sqrt(n)
#' @param rse This is the rse that we try to match, if not specified,
#'   it is derived from `n`
#' @param upper The upper boundary for root finding in terms of
#'   degrees of freedom.  If not specified, it is n*200
#' @param totN This represents the total number of simulated inverse
#'   wishart deviates
#' @param diag When `TRUE`, represents the rse to match is the
#'   diagonals, otherwise it is the total matrix.
#' @param seed to make the simulation reproducible, this represents
#'   the seed that is used for simulating the inverse Wishart
#'   distribution
#' @return output from `uniroot()` to find the right estimate
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' dfWishart(lotri::lotri(a+b~c(1, 0.5, 1)), 100)
#'
dfWishart <- function(omega, n, rse, upper, totN=1000, diag=TRUE, seed=1234) {
  checkmate::assertMatrix(omega, "numeric", min.rows=1, min.cols=1)
  if (!missing(rse) && !missing(n)) {
    stop("can only specify `n` or `rse` not both", call.=FALSE)
  }
  if (missing(rse) && !missing(n)) {
    checkmate::assertIntegerish(n, len=1, lower=1)
    rse <- sqrt(2)/sqrt(n)
    if (missing(upper)) {
      upper <- 200*n
    }
  } else if (missing(rse)) {
    stop("need to match rse with some metric", call.=FALSE)
  } else if (missing(upper)) {
    upper <- 200*(sqrt(2)/rse)^2
  }
  checkmate::assertNumeric(upper, len=1, lower=1)
  checkmate::assertIntegerish(totN, len=1, lower=1)
  .d <-  dim(omega)
  if (.d[1] != .d[2]) {
    stop("omega must be a square matrix",
         call.=FALSE)
  }
  # nu-p-3 > 0 so min for nu is
  .lower <- .d[1] + 3.1
  .upper <-  upper
  rxWithSeed(seed, {
    stats::uniroot(dfWishartCalcRse, lower=.lower, upper=.upper, omega=omega, totN=totN, rse=rse, diag=diag)
  })
}
#' Swaps the matrix list with a cube
#'
#' @param matrixListOrCube Either a list of 2-dimensional matrices or a cube of matrices
#' @return A list or a cube (opposite format as input)
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' # Create matrix list
#' matLst <- cvPost(10, lotri::lotri(a+b~c(1, 0.25, 1)), 3)
#' print(matLst)
#'
#' # Convert to cube
#' matCube <- swapMatListWithCube(matLst)
#' print(matCube)
#'
#' # Convert back to list
#' matLst2 <- swapMatListWithCube(matCube)
#' print(matLst2)
#'
swapMatListWithCube <- function(matrixListOrCube) {
  .dim <- dim(matrixListOrCube)
  if (length(.dim) == 3L) {
    return(.Call(`_rxode2_swapMatListWithCube_`, matrixListOrCube))
  } else if (length(.dim) > 0L) {
  } else if (inherits(matrixListOrCube, "list") && length(matrixListOrCube) > 0L) {
    .m0 <- matrixListOrCube[[1]]
    .dim <- dim(.m0)
    if (length(.dim) == 2L)     return(.Call(`_rxode2_swapMatListWithCube_`, matrixListOrCube))
  }
  stop("The input must be a cube or a list of matrices", call.=FALSE)
}

##' Save the pre-integration rxode2 solver state to a binary file
##'
##' This is called automatically when \code{rxControl(serializeFile="path")} is
##' supplied to \code{rxSolve()}.  The file can be reloaded later by passing it
##' as the \code{params} argument of \code{rxSolve()} together with the matching
##' model object.
##'
##' @param file Character(1).  Path of the file to write.
##' @return Invisibly \code{TRUE}.
##' @seealso \code{\link{rxLoadState}}
##' @keywords internal
##' @export
rxSaveState <- function(file) {
  checkmate::assertCharacter(file, len = 1L, any.missing = FALSE)
  .cState <- rxSaveState_()
  .rxSaveStateBundle(file, .cState)
  invisible(TRUE)
}

#' Internal function to bundle the C state with R metadata and write to disk
#' @param file The file path
#' @param cState The raw vector from C++
#' @param object Optional rxode2 model object used to persist model identity
#' @noRd
.rxSaveStateBundle <- function(file, cState, object = NULL, solveState = NULL,
                               params = NULL, events = NULL, inits = NULL) {
  .model <- rxModels_()
  .bundle <- list(
    cState = cState,
    keepFcov = .model$keepFcov,
    keepFcovType = .model$keepFcovType,
    idLevels = .model$idLevels,
    modelId = .rxSerializeModelId(object),
    solveState = solveState,
    params = params,
    events = events,
    inits = inits
  )
  .rawBundle <- serialize(.bundle, NULL)
  .compressedBundle <- memCompress(.rawBundle, type = "xz")
  writeBin(.compressedBundle, file)
}

#' Internal helper to return the parsed model md5 for serialization validation
#' @param object Optional rxode2 model object
#' @return Character scalar or NULL when model identity is unavailable
#' @noRd
.rxSerializeModelId <- function(object = NULL) {
  if (is.null(object)) {
    return(NULL)
  }
  .mv <- rxModelVars(object)
  .md5 <- .mv$md5["parsed_md5"]
  if (length(.md5) != 1L || is.na(.md5) || identical(unname(.md5), "")) {
    return(NULL)
  }
  unname(.md5)
}

#' Internal helper to read a serialized rxode2 state bundle
#' @param file Serialization file path
#' @return Bundled serialized state
#' @noRd
.rxReadStateBundle <- function(file) {
  checkmate::assertCharacter(file, len = 1L, any.missing = FALSE)
  if (!file.exists(file)) {
    stop("File does not exist", call. = FALSE)
  }
  .sz <- file.info(file)$size
  .rawDat <- readBin(file, "raw", .sz)
  .decompressed <- memDecompress(.rawDat, type = "xz")
  unserialize(.decompressed)
}

#' Internal helper to restore R-side metadata from a serialized bundle
#' @param bundle Deserialized bundle
#' @return Invisibly TRUE
#' @noRd
.rxRestoreStateBundle <- function(bundle) {
  .rxm <- rxModels_()
  if (!is.null(bundle$keepFcov)) {
    assign("keepFcov", bundle$keepFcov, envir = .rxm)
  } else {
    assign("keepFcov", NULL, envir = .rxm)
  }
  if (!is.null(bundle$keepFcovType)) {
    assign("keepFcovType", bundle$keepFcovType, envir = .rxm)
  } else {
    assign("keepFcovType", NULL, envir = .rxm)
  }
  if (!is.null(bundle$idLevels)) {
    assign("idLevels", bundle$idLevels, envir = .rxm)
  } else {
    assign("idLevels", NULL, envir = .rxm)
  }
  invisible(TRUE)
}

#' Internal helper to ensure a serialized bundle matches the supplied model
#' @param object rxode2 model object
#' @param bundle Deserialized bundle
#' @param file Serialization file path
#' @return Invisibly TRUE
#' @noRd
.rxValidateStateBundleModel <- function(object, bundle, file) {
  .modelId <- bundle$modelId
  if (is.null(.modelId) || length(.modelId) != 1L || is.na(.modelId) ||
      identical(.modelId, "")) {
    stop(sprintf("Serialization file '%s' is missing model identity", file),
         call. = FALSE)
  }
  .currentId <- .rxSerializeModelId(object)
  if (is.null(.currentId)) {
    stop("Unable to determine model identity for serialized solve", call. = FALSE)
  }
  if (!identical(.currentId, .modelId)) {
    stop(sprintf("Serialization file '%s' does not match the supplied model", file),
         call. = FALSE)
  }
  invisible(TRUE)
}

#' Check if a file is an rxode2 serialize file
#' @param file The file path
#' @return TRUE if it is a valid serialize file, FALSE otherwise
#' @noRd
.rxIsSerializeFile <- function(file) {
  if (!file.exists(file)) {
    return(FALSE)
  }
  .res <- tryCatch(
    {
      .bundle <- .rxReadStateBundle(file)
      if (is.list(.bundle) && !is.null(.bundle$cState)) {
        rxIsSerializeFile_(.bundle$cState)
      } else {
        FALSE
      }
    },
    error = function(e) {
      FALSE
    }
  )
  .res
}

##' Restore a pre-integration rxode2 solver state from a binary file
##'
##' Reads a file written by \code{\link{rxSaveState}} (or via
##' \code{rxControl(serializeFile=...)}) and reconstructs the live solver state
##' so that ODE integration can proceed from the saved pre-integration
##' checkpoint.
##'
##' This function is called automatically when \code{rxSolve(mod, "state.rxbin")}
##' detects a serialization file by its magic bytes and validates that the file
##' matches the supplied model.
##'
##' @param file Character(1).  Path of a file written by \code{\link{rxSaveState}}.
##' @return Invisibly \code{TRUE}.
##' @seealso \code{\link{rxSaveState}}
##' @keywords internal
##' @export
rxLoadState <- function(file) {
  checkmate::assertCharacter(file, len = 1L, any.missing = FALSE)
  .bundle <- .rxReadStateBundle(file)
  .rxRestoreStateBundle(.bundle)
  rxRestoreState_(.bundle$cState)
  invisible(TRUE)
}

#' Turn into a model expression
#'
#' @param x item to convert to a `model({})` expression
#' @return model expression
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' model <- quote(model({
#'   ka <- exp(tka + eta.ka)
#'   cl <- exp(tcl + eta.cl)
#'   v <- exp(tv + eta.v)
#'   d/dt(depot) = -ka * depot
#'   d/dt(center) = ka * depot - cl / v * center
#'   cp = center / v
#'   cp ~ add(add.sd)
#' }))
#'
#' as.model(model)
#'
#' one.compartment <- function() {
#'   ini({
#'     tka <- log(1.57)
#'     tcl <- log(2.72)
#'     tv <- log(31.5)
#'     eta.ka ~ 0.6
#'     eta.cl ~ 0.3
#'     eta.v ~ 0.1
#'     add.sd <- 0.7
#'   })
#'   model({
#'     ka <- exp(tka + eta.ka)
#'     cl <- exp(tcl + eta.cl)
#'     v <- exp(tv + eta.v)
#'     d/dt(depot) = -ka * depot
#'     d/dt(center) = ka * depot - cl / v * center
#'     cp = center / v
#'     cp ~ add(add.sd)
#'   })
#' }
#'
#' as.model(one.compartment)
#'
#' ui <- one.compartment()
#'
#' as.model(ui)
#'
#' model <- c("model({",
#'            "ka <- exp(tka + eta.ka)",
#'            "cl <- exp(tcl + eta.cl)",
#'            "v <- exp(tv + eta.v)",
#'            "d/dt(depot) = -ka * depot",
#'            "d/dt(center) = ka * depot - cl / v * center",
#'            "cp = center / v",
#'            "cp ~ add(add.sd)",
#'            "})")
#'
#' as.model(model)
#'
#' model <- paste(model, collapse="\n")
#'
#' as.model(model)
#'
as.model <- function(x) {
  UseMethod("as.model")
}
#' @rdname as.model
#' @export
as.model.character <- function(x) {
  .ret <- try(as.model(lapply(x, function(i) {
    str2lang(i)
  })), silent=TRUE)
  if (inherits(.ret, "try-error")) {
    .ret <- try(as.model(str2lang(paste(x, collapse="\n"))), silent=TRUE)
    if (inherits(.ret, "try-error")) {
      stop("error converting character vector to model({}) expression",
           call.=FALSE)
    }
  }
  .ret
}

#' @rdname as.model
#' @export
as.model.call <- function(x) {
  if (!identical(x[[1]], quote(`model`))) {
    stop("unsupported expression of model({}) block",
         call.=FALSE)
  }
  x
}

#' @rdname as.model
#' @export
as.model.list <- function(x) {
  if (inherits(x, "rxUi")) {
    return(x$model)
  }
  .lst <- lapply(seq_along(x), function(i) {
    if (is.language(x[[i]])) return(x[[i]])
    if (is.character(x[[i]])) return(str2lang(x[[i]]))
    stop("unsupported expression of model({}) block",
         call.=FALSE)
  })
  as.call(c(quote(`model`), as.call(c(quote(`{`), .lst))))
}

#' @rdname as.model
#' @export
as.model.default <- function(x) {
  .model <- try(as.rxUi(x), silent=TRUE)
  if (inherits(.model, "try-error")) {
    stop("do not know how to convert this to an `model` expression",
         call.=FALSE)
  }
  model(.model)
}

#' rxTheme is the ggplot2 theme for rxode2 plots
#'
#' @inheritParams ggplot2::theme_grey
#'
#' @param grid a Boolean indicating if the grid is on (`TRUE`) or off
#'   (`FALSE`). This could also be a character indicating `x` or `y`.
#'
#' @return ggplot2 theme used in rxode2
#' @family rxode2 plotting
#' @export
rxTheme <- function(base_size = 11, base_family = "",
                    base_line_size = base_size / 22,
                    base_rect_size = base_size / 22,
                    grid = TRUE) {
  half_line <- base_size / 2
  .greyTextAxisX <- ggplot2::element_text(
    color = "#808078",
    margin = ggplot2::margin(t = 0.8 * half_line / 2), vjust = 1
  )
  .greyTextAxisY <- ggplot2::element_text(
    color = "#808078",
    margin = ggplot2::margin(r = 0.8 * half_line / 2), hjust = 1
  )
  .greyLabTextX <- ggplot2::element_text(
    color = "#808078", face = "bold",
    margin = ggplot2::margin(t = half_line / 2),
    size = ggplot2::rel(1.1)
  )
  .greyLabTextY <- ggplot2::element_text(
    color = "#808078", face = "bold", angle = 90,
    margin = ggplot2::margin(r = half_line / 2),
    size = ggplot2::rel(1.1)
  )
  .title <- ggplot2::element_text(
    colour = "#808078", face = "bold", hjust = 0,
    size = ggplot2::rel(1.2),
    margin = ggplot2::margin(b = half_line)
  )
  .subTitle <- ggplot2::element_text(
    colour = "#808078", face = "bold", hjust = 0,
    margin = ggplot2::margin(b = half_line)
  )

  .greyTick <- ggplot2::element_line(color = "#808078")
  .panelGrid <-
    .greyX <- NULL
  .greyY <- NULL
  .blankGrid <- NULL
  if (inherits(grid, "character") || grid == TRUE) {
    .greyMajor <- ggplot2::element_line(color = "#BFBFB4")
    .panelGrid <- .greyMajor
    .greyMinor <- ggplot2::element_line(color = "#E6E6D8")
    .greyMajorX <- .greyMajor
    .greyMinorX <- .greyMinor
    .greyMajorY <- .greyMajor
    .greyMinorY <- .greyMinor
    if (inherits(grid, "character")) {
      if (regexpr("X", grid)[1] < 0) .greyMajorX <- ggplot2::element_blank()
      if (regexpr("Y", grid)[1] < 0) .greyMajorY <- ggplot2::element_blank()
      if (regexpr("x", grid)[1] < 0) .greyMinorX <- ggplot2::element_blank()
      if (regexpr("y", grid)[1] < 0) .greyMinorY <- ggplot2::element_blank()
    }
  } else {
    .panelGrid <- ggplot2::element_blank()
    .greyMajor <- .panelGrid
    .greyMinor <- .panelGrid
    .greyMajorX <- .greyMajor
    .greyMinorX <- .greyMinor
    .greyMajorY <- .greyMajor
    .greyMinorY <- .greyMinor
  }
  .theme <- ggplot2::theme_bw(
    base_size = base_size, base_family = base_family,
    base_line_size = base_line_size,
    base_rect_size = base_rect_size
  ) %+replace%
    ggplot2::theme(
      plot.title = .title,
      plot.subtitle = .title,
      panel.border = ggplot2::element_blank(),
      ## panel.background = ggplot2::element_rect(fill = "#FFFFF7", colour = NA),
      panel.grid = .panelGrid,
      panel.grid.minor = .greyMinor,
      panel.grid.minor.x = .greyMinorX,
      panel.grid.minor.y = .greyMinorY,
      panel.grid.major.x = .greyMajorX,
      panel.grid.major.y = .greyMajorY,
      panel.grid.major = .greyMajor,
      axis.text.x = .greyTextAxisX,
      axis.text.y = .greyTextAxisY,
      axis.title.x = .greyLabTextX,
      axis.title.y = .greyLabTextY,
      axis.ticks.x = .greyTick,
      axis.ticks.y = .greyTick,
      strip.text = ggplot2::element_text(
        color = "#FFFFF7", face = "bold",
        size = ggplot2::rel(1.2),
        margin = ggplot2::margin(0.5 * half_line, 0.5 * half_line, 0.5 * half_line, 0.5 * half_line)
      ),
      strip.background = ggplot2::element_rect(fill = "#808078", color = NA)
    )
  ## If above ggplot2 3.2.1 then add plot.title.position="plot"
  if (utils::packageVersion("ggplot2") > "3.2.1") {
    .theme <- .theme %+replace% ggplot2::theme(plot.title.position = "plot")
  }
}

.dropUnits <- function(x) {
  if (inherits(x, "units")) {
    return(units::drop_units(x))
  }
  return(x)
}

.plotTime <- function(.dat, xlab) {
  .xgxr <- getOption("rxode2.xgxr", TRUE) &&
    requireNamespace("xgxr", quietly = TRUE)
  .timex <- NULL
  .unit <- NULL
  .xgxrT <- function(.unit) {
    if (!.xgxr) {
      return(NULL)
    }
    if (is.null(.unit)) {
      return(NULL)
    }
    .timex <- xgxr::xgx_scale_x_time_units(.unit)
    if (inherits(.timex, "list")) {
      .w <- which(sapply(seq_along(.timex), function(x) {
        inherits(.timex[[x]], "labels")
      }))
      if (length(.w) > 0) {
        .timex <- .timex[-.w]
      }
    }
    return(.timex)
  }
  if (inherits(.dat$time, "units")) {
    .unit <- as.character(units(.dat$time))
    .unitNames <- c(s="second",
                    sec="second",
                    second="second",
                    min="minute",
                    minute="minute",
                    h="hour",
                    hr="hour",
                    hour="hour",
                    d="day",
                    day="day",
                    days="day",
                    week="week",
                    weeks="week",
                    month="month",
                    months="month",
                    year="year",
                    years="year",
                    yr="year")
    .unit <- setNames(.unitNames[.unit], NULL)
    if (is.na(.unit)) {
      .timex <- .xgxrT(NULL)
      .xlab <- xlab(xlab)
    }
    .dat$time <- .dropUnits(.dat$time)
    .timex <- .xgxrT(.unit)
    .xlab <- xlab(sprintf("%s [%s]", xlab, .unit))
  } else {
    .timex <- .xgxrT(NULL)
    .xlab <- xlab(xlab)
  }
  list(timex=.timex, xlab=.xlab, dat=.dat)
}

.plotLog <- function(.dat, .timex, log = "") {
  # match.arg() cannot be used because it doesn't work with an empty string
  # (https://stackoverflow.com/questions/41441170/failure-of-match-arg-for-the-empty-string)
  stopifnot(length(log) == 1)
  stopifnot(is.character(log))
  stopifnot(log %in% c("", "x", "y", "xy", "yx"))
  useLogX <- nchar(log) == 2L || log == "x"
  useLogY <- nchar(log) == 2L || log == "y"
  useXgxr <-
    getOption("rxode2.xgxr", TRUE) &&
    requireNamespace("xgxr", quietly = TRUE)
  .logx <- NULL
  .logy <- NULL
  if (useLogX) {
    stopifnot(".dat requires 'time' column"="time" %in% names(.dat))
    .dat <- .dat[.dat$time > 0, ]
    if (useXgxr) {
      .logx <- xgxr::xgx_scale_x_log10()
    } else {
      .logx <- ggplot2::scale_x_log10()
    }
    .timex <- NULL
  }
  if (useLogY) {
    if (useXgxr) {
      .logy <- xgxr::xgx_scale_y_log10()
    } else {
      .logy <- ggplot2::scale_y_log10()
    }
  }
  list(timex=.timex, logx=.logx, logy=.logy, dat=.dat)
}

#' Plot rxode2 objects
#'
#' @param x rxode2 object to plot
#' @param y Compartments or left-hand-side values to plot either as a bare name
#'   or as a character vector
#' @param ... Ignored
#' @param log Should "" (neither x nor y), "x", "y", or "xy" (or "yx") be
#'   log-scale?
#' @param xlab,ylab The x and y axis labels
#'
#' @return A ggplot2 object
#'
#' @family rxode2 plotting
#' @keywords Internal
#' @export
plot.rxSolve <- function(x, y, ..., log = "", xlab = "Time", ylab = "") {
  .data <- NULL
  .y <- as.character(substitute(y))
  .call0 <- match.call()[-(1:2)]
  .call <- as.list(.call0)
  .w <- which(names(.call) %in% c("x", "y", "log", "xlab", "ylab"))
  if (length(.w) > 0) {
    .call <- .call[-.w]
  }
  .cmts <- c(
    as.character(substitute(y)),
    names(vapply(as.character(.call), `c`, character(1), USE.NAMES=FALSE)),
    as.character(unlist(.call))
  )
  .cmts <- .cmts[.cmts != ""]
  .cmts <- unique(.cmts)
  if (length(.cmts) == 0L) {
    .cmts <- NULL
  } else {
    .doSim <- all(vapply(.cmts, function(x) {
      if (substr(x, 0, 4) == "sim.") return(TRUE)
      FALSE
    }, logical(1), USE.NAMES=FALSE))
    if (.doSim) {
      .vars <- vapply(.cmts, function(x) {
        substr(x, 5, nchar(x))
      }, character(1), USE.NAMES=FALSE)
      .mv <- rxModelVars(x)
      .good <- c(.mv$state, .mv$stateExtra)
      .cmts0 <- intersect(.vars, .good)
      if (length(.cmts0) == 0) {
        stop("the `sim` endpoints requested in the plot do not exist in the solved object: ",
             paste(.cmts, collapse=", "))
      } else if (length(.cmts0) != length(.cmts)) {
        .ignored <- .vars[!(.vars %in% .good)]
        warning("some `sim` requested items do not exist in the solved object and were ignored: ",
                paste(paste0("sim.", .ignored), collapse=", "))
      }
      .cmts <- paste0("sim.", .cmts0)
    } else {
      .cmts <- .cmts[!duplicated(.cmts)]
      .both <- c(rxState(x), rxLhs(x))
      .cmts0 <- intersect(.cmts, .both)
      if (length(.cmts0) == 0) {
        stop("the items requested in the plot do not exist in the solved object: ",
             paste(.cmts, collapse=", "))
      } else if (length(.cmts0) != length(.cmts)) {
        .ignored <- .cmts[!(.cmts %in% .both)]
        warning("some requested items do not exist in the solved object and were ignored: ",
                paste(.ignored, collapse=", "))
      }
      .cmts <- .cmts0
    }
  }
  .dat <- rxStack(x, .cmts)
  .cmts <- as.character(unique(.dat$trt))
  .nlvl <- 1L
  if (any(names(.dat) == "id")) {
    if (any(names(.dat) == "sim.id")) {
      if (any(names(.dat) == "resetno")) {
        .dat$id <- factor(paste0("id=", .dat$id, ", sim.id=", .dat$sim.id, ", resetno=", .dat$resetno))
      } else {
        .dat$id <- factor(paste0("id=", .dat$id, ", sim.id=", .dat$sim.id))
      }
    } else {
      if (any(names(.dat) == "resetno")) {
        .dat$id <- factor(paste0("id=", .dat$id, ", resetno=", .dat$resetno))
      } else {
        .dat$id <- factor(.dat$id)
      }
    }
    .nlvl <- length(levels(.dat$id))
    .dat2 <- .dat[rev(seq_along(.dat$id)), ]
    .dat2$label <- .dat$id
    .dat2$time <- .dropUnits(.dat2$time)
    row.names(.dat2) <- NULL
    .dat2 <- .dat2[!duplicated(paste0(.dat2$id, .dat2$trt)), ]
    .aes <- aes(.data$time, .data$value, color = .data$id)
    .aesG <- aes(.data$time, .data$value, group = .data$id)
    .aesLab <- aes(label = .data$label)
  } else if (any(names(.dat) == "sim.id")) {
    if (any(names(.dat) == "resetno")) {
      .dat$sim.id <- factor(paste0("sim.id=", .dat$sim.id, ", resetno=", .dat$resetno))
    } else {
      .dat$sim.id <- factor(.dat$sim.id)
    }
    .nlvl <- length(levels(.dat$sim.id))
    .dat2 <- .dat[rev(seq_along(.dat$sim.id)), ]
    .dat2$label <- .dat$sim.id
    .dat2$time <- .dropUnits(.dat2$time)
    row.names(.dat2) <- NULL
    .dat2 <- .dat2[!duplicated(paste0(.dat2$sim.id, .dat2$trt)), ]
    .aes <- aes(.data$time, .data$value, color = .data$sim.id)
    .aesG <- aes(.data$time, .data$value, group = .data$sim.id)
    .aesLab <- aes(label = .data$label)
  } else if (any(names(.dat) == "resetno")) {
    .dat$resetno <- factor(.dat$resetno)
    .nlvl <- length(levels(.dat$resetno))
    .dat2 <- .dat[rev(seq_along(.dat$resetno)), ]
    .dat2$label <- .dat$resetno
    .dat2$time <- .dropUnits(.dat2$time)
    row.names(.dat2) <- NULL
    .dat2 <- .dat2[!duplicated(paste0(.dat2$resetno, .dat2$trt)), ]
    .aes <- aes(.data$time, .data$value, color = .data$resetno)
    .aesG <- aes(.data$time, .data$value, group = .data$resetno)
    .aesLab <- aes(label = .data$label)
  } else {
    .aes <- aes(.data$time, .data$value)
  }
  .facet <- facet_wrap(~trt, scales = "free_y")
  if (length(.cmts) == 1) .facet <- NULL
  .ylab <- ylab(ylab)
  .theme <- rxTheme()
  if (!getOption("rxode2.theme", TRUE)) .theme <- NULL
  .repel <- NULL
  .legend <- NULL
  if (compareVersion(as.character(packageVersion("ggplot2")), "3.3.6.9000") < 0) {
    .line <- geom_line(size = 1.2)
  } else {
    .line <- geom_line(linewidth = 1.2)
  }
  .rxSpaghetti <- getOption("rxode2.spaghetti", 7L)
  .ggrepel <- getOption("rxode2.ggrepel", TRUE) &&
    requireNamespace("ggrepel", quietly = TRUE)
  if (.nlvl > 1 && .nlvl < .rxSpaghetti && .ggrepel && is.null(.facet)) {
    .repel <- ggrepel::geom_label_repel(.aesLab,
      data = .dat2, nudge_x = 1,
      fontface = "bold", size = 5
    )
    .legend <- ggplot2::guides(color = "none")
  } else {
    if (.nlvl < .rxSpaghetti) {
      .legend <- ggplot2::theme(legend.title = ggplot2::element_blank())
    } else {
      .legend <- ggplot2::guides(color = "none")
      if (compareVersion(as.character(packageVersion("ggplot2")), " 3.3.6.9000") < 0) {
        .line <- geom_line(size = 1.2, alpha = 0.2)
      } else {
        .line <- geom_line(linewidth = 1.2, alpha = 0.2)
      }
      .aes <- .aesG
    }
  }
  .lst <- .plotTime(.dat, xlab)
  .xlab <- .lst[["xlab"]]
  .lst <- .plotLog(.lst[["dat"]], .lst[["timex"]], log)
  ggplot(.lst[["dat"]], .aes) +
    .line +
    .facet +
    .theme +
    .repel +
    .lst[["timex"]] +
    .lst[["logx"]] +
    .lst[["logy"]] +
    .ylab +
    .xlab +
    .legend ->
  .gg
  .gg
}

#' @rdname plot.rxSolve
#' @keywords Internal
#' @export
plot.rxSolveConfint1 <- function(x, y, ..., xlab = "Time", ylab = "", log = "") {
  .data <- NULL
  .y <- as.character(substitute(y))
  .call0 <- match.call()[-(1:2)]
  .call <- as.list(.call0)
  .w <- which(names(.call) %in% c("x", "y", "log", "xlab", "ylab"))
  if (length(.w) > 0) {
    .call <- .call[-.w]
  }
  .cmts <- c(
    as.character(substitute(y)),
    names(vapply(as.character(.call), `c`, character(1), USE.NAMES=FALSE)),
    as.character(unlist(.call))
  )
  .cmts <- .cmts[.cmts != ""]
  .cmts <- unique(.cmts)
  .lvl <- attr(class(x), ".rx")$lvl
  .parm <- attr(class(x), ".rx")$parm
  if (length(.cmts) > 0) {
    .parm <- intersect(.parm, .cmts)
    x  <- x[x$trt %in% .parm,]
  }
  .by <- attr(class(x), ".rx")$by
  .aes <- aes(.data$time, .data$eff)
  .facet <- NULL
  .dat <- x
  .lst <- .plotTime(.dat, xlab)
  .xlab <- .lst[["xlab"]]
  .lst <- .plotLog(.lst[["dat"]], .lst[["timex"]], log)
  .timex <- .lst[["timex"]]
  .logx <- .lst[["logx"]]
  .logy <- .lst[["logy"]]
  .dat <- .lst[["dat"]]
  if (length(.parm) > 1) {
    if (length(.by) > 0) {
      .facet <- eval(str2lang(paste0("facet_wrap(~", paste(c("trt", .by), collapse="+"),
                                     ", scales = \"free_y\")")))
    } else {
      .facet <- facet_wrap(~trt, scales = "free_y")
    }
  } else if (length(.by) > 0) {
    .facet <- eval(str2lang(paste0("facet_wrap(~", paste(.by, collapse="+"),
                                   ", scales = \"free_y\")")))
  }
  if (compareVersion(as.character(packageVersion("ggplot2")), "3.4.0") < 0) {
    .line <- geom_line(size = 1.2, show.legend = !is.null(.facet))
  } else {
    .line <- geom_line(linewidth = 1.2, show.legend = !is.null(.facet))
  }
  .theme <- NULL
  if (getOption("rxode2.theme_bw", TRUE)) {
    .theme <- rxTheme()
  }
  .ylab <- ylab(ylab)
  .dat$p <- .dat$Percentile
  .d2 <- .dat[!duplicated(.dat$Percentile), ]
  .d2 <- .d2[.d2$Percentile != "50%", ]
  .dat0 <- .dat[.dat$Percentile == "50%", ]
  .dat0$low <- .dat$eff[.dat$Percentile == .d2$Percentile[1]]
  .dat0$up <- .dat$eff[.dat$Percentile == .d2$Percentile[2]]
  .aesR <- ggplot2::aes(ymin = .data$low, ymax = .data$up)
  .ribbon <- ggplot2::geom_ribbon(.aesR, alpha = 0.5, fill = "gray50")
  ## .title <- ggplot2::ggtitle(paste0("50% [", .d2$Percentile[1], ", ",
  ##                                 .d2$Percentile[2], "]"))
  ggplot2::ggplot(.dat0, .aes) +
    .ribbon +
    .line +
    .facet +
    .timex +
    .logx +
    .logy +
    .xlab +
    .ylab +
    .theme ->
  .ret
  return(.ret)
}

#' @rdname plot.rxSolve
#' @keywords Internal
#' @export
plot.rxSolveConfint2 <- function(x, y, ..., xlab = "Time", ylab = "", log = "") {
  .data <- NULL
  .y <- as.character(substitute(y))
  .call0 <- match.call()[-(1:2)]
  .call <- as.list(.call0)
  .w <- which(names(.call) %in% c("x", "y", "log", "xlab", "ylab"))
  if (length(.w) > 0) {
    .call <- .call[-.w]
  }
  .cmts <- c(
    as.character(substitute(y)),
    names(vapply(as.character(.call), `c`, character(1), USE.NAMES=FALSE)),
    as.character(unlist(.call))
  )
  .cmts <- .cmts[.cmts != ""]
  .cmts <- unique(.cmts)
  .parm <- attr(class(x), ".rx")$parm
  if (length(.cmts) > 0) {
    .parm <- intersect(.parm, .cmts)
    x  <- x[x$trt %in% .parm,]
  }
  .lvl <- attr(class(x), ".rx")$lvl
  .ci  <- attr(class(x), ".rx")$ci
  .by <- attr(class(x), ".rx")$by
  .aes <- aes(.data$time, .data$p50,
    color = .data$Percentile,
    fill = .data$Percentile,
    group = .data$p
  )
  .aesR <- ggplot2::aes_string(ymin = .ci[1], ymax = .ci[3])
  .facet <- NULL
  .dat <- x
  .lst <- .plotTime(.dat, xlab)
  .xlab <- .lst[["xlab"]]
  .lst <- .plotLog(.lst[["dat"]], .lst[["timex"]], log)
  .timex <- .lst[["timex"]]
  .logx <- .lst[["logx"]]
  .logy <- .lst[["logy"]]
  .dat <- .lst[["dat"]]
  if (length(.parm) > 1) {
    if (length(.by) > 0) {
      .facet <- eval(str2lang(paste0("facet_wrap(~", paste(c("trt", .by), collapse="+"),
                                     ", scales = \"free_y\")")))
    } else {
      .facet <- facet_wrap(~trt, scales = "free_y")
    }
  } else if (length(.by) > 0) {
    .facet <- eval(str2lang(paste0("facet_wrap(~", paste(.by, collapse="+"),
                                   ", scales = \"free_y\")")))
  }
  if (compareVersion(as.character(packageVersion("ggplot2")), "3.4.0") < 0) {
    .line <- geom_line(size = 1.1, show.legend = FALSE)
  } else {
    .line <- geom_line(linewidth = 1.1, show.legend = FALSE)
  }
  .theme <- NULL
  if (getOption("rxode2.theme_bw", TRUE)) {
    .theme <- rxTheme()
  }
  .ylab <- ylab(ylab)
  .dat$p <- .dat$Percentile
  .d2 <- .dat[!duplicated(.dat$Percentile), ]
  .d2 <- .d2[.d2$Percentile != "50%", ]
  .d2 <- sub("[%]", "", paste0(.d2$Percentile, collapse = " / "))
  .dat$Percentile <- paste(.dat$Percentile)
  .dat$Percentile[.dat$Percentile != "50%"] <- .d2
  .dat$Percentile <- factor(.dat$Percentile, c("50%", .d2))
  .ribbon <- ggplot2::geom_ribbon(.aesR, alpha = 0.5, col = NA, show.legend = FALSE)
  .leg1 <- ggplot2::scale_color_manual(values = c("black", "gray"))
  .leg2 <- ggplot2::scale_fill_manual(values = c("black", "gray"))
  ggplot2::ggplot(.dat, .aes) +
    .ribbon +
    .line +
    .facet +
    .timex +
    .logx +
    .logy +
    .xlab +
    .ylab +
    .leg1 +
    .leg2 +
    .theme ->
  .ret
  ## p1 <- time <- eff <-Percentile <-sim.id <-id <-p2 <-p50 <-p05 <- p95 <- . <- NULL
  ## .lvl <- attr(class(x), ".rx")$lvl
  ## .parm <- attr(class(x), ".rx")$parm
  ## .ret <- ggplot2::ggplot(x,ggplot2::aes(time,p50,col=Percentile,fill=Percentile)) +
  ##     ggplot2::geom_ribbon(ggplot2::aes_string(ymin=.lvl[1],ymax=.lvl[3]),alpha=0.5)+
  ##     ggplot2::geom_line(size=1.2);
  ## if (length(.parm) > 1){
  ##     .ret <- .ret + facet_wrap( ~ trt, scales="free_y")
  ## }
  ## if (getOption("rxode2.theme_bw", TRUE)){
  ##     .ret <- .ret + ggplot2::theme_bw()
  ## }
  return(.ret)
}

#' Transition ODEs written in d/dt() format to matrix exponential / inductive linearization format
#'
#' @param model rxode2 model, text, or function
#' @param doConst Replace constants with values; By default this is `FALSE`.
#' @param calcSens A character vector of parameter names for which sensitivities should be calculated.
#' @return A character string representing the matrix exponential model code
#' @author Matthew L. Fidler
#' @export
indLin <- function(model, doConst = FALSE, calcSens = NULL) {
  rxReq("symengine")
  if (!is.null(calcSens)) {
    return(rxSensMatExp(model = model, calcSens = calcSens, doConst = doConst))
  }
  
  # 1. Parse model to get model variables and load symengine environment
  .mv <- rxModelVars(model)
  .env <- .rxLoadPrune(model, doConst = doConst)
  .states <- rxState(.env)
  
  if (length(.states) == 0L) {
    stop("No state variables (compartments) found in the model.", call. = FALSE)
  }
  
  # 2. Call the C/C++ registered function to get inductive linearization matrices
  .ret <- eval(parse(text = rxIndLin_((.states))))
  
  # 3. Extract the coefficient matrix (rows/cols = states) and forcing function vector
  .ret0 <- .ret[.states, .states, drop = FALSE]
  .ret1 <- .ret[, "_rxF", drop = FALSE]
  
  # 4. Construct the new model code
  .code <- c("matExp()")
  
  # Add compartment declarations to ensure ordering
  for (.s in .states) {
    .code <- c(.code, paste0("cmt(", .s, ")"))
  }
  
  # Extract off-diagonal transfer rates and diagonal output/elimination rates
  for (j in seq_along(.states)) {
    .cmt1 <- .states[j]
    .offTerms <- list()
    
    # Off-diagonals: rate of transfer from cmt1 to cmt2
    for (i in seq_along(.states)) {
      .cmt2 <- .states[i]
      if (i != j) {
        .val <- .ret0[i, j]
        if (.val != "0") {
          .kname <- paste0("k_", .cmt1, "_", .cmt2)
          # Check if the expression is already the micro-constant name (either snake_case or dot notation)
          if (.val == .kname || .val == paste0("k.", .cmt1, ".", .cmt2)) {
            .code <- c(.code, paste0("param(", .val, ")"))
          } else {
            .code <- c(.code, paste0(.kname, " = ", .val))
          }
          .offTerms <- c(.offTerms, symengine::S(.val))
        }
      }
    }
    
    # Diagonal column sum: elimination/output rate from cmt1
    .diag <- .ret0[j, j]
    if (.diag != "0" || length(.offTerms) > 0) {
      .sumExpr <- symengine::S(.diag)
      for (.t in .offTerms) {
        .sumExpr <- .sumExpr + .t
      }
      .elimStr <- as.character(symengine::S(paste0("-(", rxFromSE(.sumExpr), ")")))
      if (.elimStr != "0") {
        .knameOut <- paste0("k_", .cmt1, "_output")
        if (.elimStr == .knameOut || .elimStr == paste0("k.", .cmt1, ".output")) {
          .code <- c(.code, paste0("param(", .elimStr, ")"))
        } else {
          .code <- c(.code, paste0(.knameOut, " = ", .elimStr))
        }
      }
    }
  }
  
  # Forcing functions: indLin property
  for (i in seq_along(.states)) {
    .cmt2 <- .states[i]
    .fVal <- .ret1[i, 1]
    if (.fVal != "0") {
      .code <- c(.code, paste0("indLin(", .cmt2, ") <- ", .fVal))
    }
  }
  
  # 5. Extract and preserve the non-ODE lines from the original normalized model
  .normModel <- .mv$model["normModel"]
  .lines <- unlist(strsplit(.normModel, "[\n;]"))
  .lines <- trimws(.lines)
  for (.l in .lines) {
    if (.l != "") {
      # Keep lines that do not define ODE derivatives or CMTs (since we output cmt declarations at the top)
      if (!grepl("^d/dt\\(", .l) && !grepl("^cmt\\(", .l)) {
        .code <- c(.code, .l)
      }
    }
  }
  
  # If there are no assignments in the code, append a dummy assignment to avoid "nothing in output queue to write" compiler error
  if (!any(grepl("=", .code) | grepl("<-", .code))) {
    .code <- c(.code, "dummy = 1")
  }
  
  return(paste(.code, collapse = "\n"))
}

#' Total derivative of an indLin/matExp Jacobian-entry expression
#'
#' `expr` is a scalar Jacobian-entry expression, differentiated wrt every
#' symbol it references: the explicit partial wrt `byVar`, plus a chain term
#' for every physical state and every pre-existing sensitivity compartment
#' symbol (the same symbol with `_BY_byVar` appended).  Generalizes
#' `.rxEventSensD2Expr()`'s state+sens coupling to arbitrary sens symbols.
#'
#' @param expr symengine expression (a Jacobian entry, or a total derivative
#'   of one built by a previous call to this function).
#' @param byVar Parameter name to differentiate wrt.
#' @param states Physical state names.
#' @return symengine expression for the total derivative.
#' @noRd
.rxIndLinTotalD <- function(expr, byVar, states) {
  .isZero <- function(.e) {
    .z <- rxFromSE(.e)
    .z == "0" || .z == "0.0" || .z == "-0"
  }
  .vars <- tryCatch(
    vapply(symengine::free_symbols(expr), as.character, character(1)),
    error = function(e) character(0)
  )
  .tot <- NULL
  .add <- function(.term) {
    if (is.null(.tot)) .tot <<- .term else .tot <<- .tot + .term
  }
  if (byVar %in% .vars) {
    .add(symengine::D(expr, symengine::S(byVar)))
  }
  for (.l in states) {
    if (!(.l %in% .vars)) next
    .dl <- symengine::D(expr, symengine::S(.l))
    if (!.isZero(.dl)) {
      .add(.dl * symengine::S(paste0("rx__sens_", .l, "_BY_", byVar, "__")))
    }
  }
  for (.s in .vars) {
    if (!startsWith(.s, "rx__sens_") || !endsWith(.s, "__")) next
    .ds <- symengine::D(expr, symengine::S(.s))
    if (.isZero(.ds)) next
    .target <- paste0(substring(.s, 1L, nchar(.s) - 2L), "_BY_", byVar, "__")
    .add(.ds * symengine::S(.target))
  }
  if (is.null(.tot)) symengine::S("0") else .tot
}

#' Apply `.rxIndLinTotalD()` repeatedly, one variable at a time
#'
#' @param base Starting symengine expression (a Jacobian entry).
#' @param byVars Character vector of variables to differentiate by, in order.
#' @param states Physical state names.
#' @return symengine expression for the repeated total derivative.
#' @noRd
.rxIndLinChainD <- function(base, byVars, states) {
  .e <- base
  for (.v in byVars) .e <- .rxIndLinTotalD(.e, .v, states)
  .e
}

#' Accumulator for non-depleting (`_nd`) cross-term contributions
#'
#' At second/third order, distinct terms can target the same `from -> to` pair
#' when differentiated parameters coincide; their coefficients must be summed,
#' not emitted as conflicting duplicate-LHS lines.  Accumulates by `(from, to)`
#' key in first-seen order, so emission has one line per pair.
#'
#' @return list with `add(from, to, val)` and `emit()` (character vector of
#'   `k_<from>_<to>_nd = <expr>` lines, skipping pairs that summed to zero).
#' @noRd
.rxIndLinNdAccumulator <- function() {
  .isZero <- function(.e) {
    .z <- rxFromSE(.e)
    .z == "0" || .z == "0.0" || .z == "-0"
  }
  .env <- new.env(parent = emptyenv())
  .order <- character(0)
  .add <- function(from, to, val) {
    if (.isZero(val)) return(invisible())
    .key <- paste0(from, "\r", to)
    if (!exists(.key, envir = .env, inherits = FALSE)) {
      .order <<- c(.order, .key)
    } else {
      val <- get(.key, envir = .env, inherits = FALSE) + val
    }
    assign(.key, val, envir = .env)
    invisible()
  }
  .emit <- function() {
    .lines <- character(0)
    for (.key in .order) {
      .val <- get(.key, envir = .env, inherits = FALSE)
      if (.isZero(.val)) next
      .parts <- strsplit(.key, "\r", fixed = TRUE)[[1L]]
      .lines <- c(.lines, paste0("k_", .parts[1L], "_", .parts[2L], "_nd = ", rxFromSE(.val)))
    }
    .lines
  }
  list(add = .add, emit = .emit)
}

#' Differentiate and expand a matrix exponential model with forward sensitivities
#'
#' @param model rxode2 model, text, or function
#' @param calcSens A character vector of parameter names for which sensitivities should be calculated.
#' @param calcSens2 character vector (or `NULL`) requesting second-order
#'   sensitivities `rx__sens_<x>_BY_<p>_BY_<q>__` (`p` over `calcSens`, `q` over
#'   `calcSens2`; every `calcSens2` element must also be in `calcSens`).
#'   Expressed as `k_from_to` micro-constant transfers like the first-order
#'   ones.  Ignored for `linCmt()` states (those use Stan forward-AD).
#' @param calcSens3 character vector (or `NULL`) requesting third-order
#'   sensitivities `rx__sens_<x>_BY_<p>_BY_<q>_BY_<r>__` (`r` over `calcSens3`).
#'   Requires `calcSens2`; every `calcSens3` element must also be in `calcSens2`.
#' @param doConst Replace constants with values; By default this is `FALSE`.
#' @param env A pre-loaded symengine environment (from `.rxLoadPrune()`) to
#'   reuse instead of reloading `model`; when `NULL` it is built internally.
#' @return A character string representing the matrix exponential sensitivity-expanded model code
#' @author Matthew L. Fidler
#' @export
rxSensMatExp <- function(model, calcSens, calcSens2 = NULL, calcSens3 = NULL, doConst = FALSE, env = NULL) {
  rxReq("symengine")
  if (!is.character(calcSens)) {
    stop("'calcSens' must be a character vector of parameter names.", call. = FALSE)
  }
  if (!is.null(calcSens2)) {
    if (!is.character(calcSens2)) {
      stop("'calcSens2' must be a character vector of parameter names.", call. = FALSE)
    }
    if (!all(calcSens2 %in% calcSens)) {
      stop("'calcSens2' must be a subset of 'calcSens' (every second-order parameter needs its own first-order sensitivity).", call. = FALSE)
    }
  }
  if (!is.null(calcSens3)) {
    if (is.null(calcSens2)) {
      stop("'calcSens3' requires 'calcSens2' to be supplied.", call. = FALSE)
    }
    if (!is.character(calcSens3)) {
      stop("'calcSens3' must be a character vector of parameter names.", call. = FALSE)
    }
    if (!all(calcSens3 %in% calcSens2)) {
      stop("'calcSens3' must be a subset of 'calcSens2' (every third-order parameter needs its own second-order sensitivity).", call. = FALSE)
    }
  }

  # 1. Load model into symengine environment (or reuse a pre-loaded one)
  .mv <- rxModelVars(model)
  if (is.null(env)) {
    .env <- .rxLoadPrune(model, doConst = doConst)
  } else {
    .env <- env
  }
  # Materialize d/dt(<state>) from matExp k_from_to constants (and any indLin()
  # forcing); a no-op for d/dt() input, which already carries rx__d_dt_<state>__.
  .rxInjectMatExpOdes(.env)
  # rxState returns all compartments; for sensitivity we only use non-output,
  # non-linCmt() states -- linCmt sensitivities come from Stan forward-AD, not
  # this Jacobian-based (matExp/indLin) expansion, at any order.
  .states <- setdiff(rxState(.env), c("output", .rxLinCmt(.mv)))

  if (length(.states) == 0L) {
    stop("No state variables (compartments) found in the model.", call. = FALSE)
  }

  # 2. Build the system Jacobian A[i, j] = d(d/dt X_i)/d X_j directly from the
  #    materialized derivatives.  For linear models A is constant; for nonlinear
  #    models (e.g. Michaelis-Menten) the entries are state-dependent expressions.
  .zero <- symengine::S("0")
  .isZero <- function(.e) {
    .z <- rxFromSE(.e)
    .z == "0" || .z == "0.0" || .z == "-0"
  }
  .rhs <- lapply(.states, function(.s) {
    .v <- paste0("rx__d_dt_", .s, "__")
    if (exists(.v, envir = .env, inherits = FALSE)) {
      base::get(.v, envir = .env, inherits = FALSE)
    } else {
      .zero
    }
  })
  names(.rhs) <- .states
  .stateSym <- lapply(.states, function(.s) symengine::S(.s))
  names(.stateSym) <- .states
  .A <- lapply(.states, function(.i) {
    .row <- lapply(.states, function(.j) symengine::D(.rhs[[.i]], .stateSym[[.j]]))
    names(.row) <- .states
    .row
  })
  names(.A) <- .states
  # elimination from compartment j: -(A[j,j] + sum_{i != j} A[i,j])
  .elimOf <- function(.j) {
    .e <- -.A[[.j]][[.j]]
    for (.i in .states) {
      if (.i != .j) .e <- .e - .A[[.i]][[.j]]
    }
    .e
  }

  # 3. Build model code
  .code <- c("matExp()")
  for (.s in .states) {
    .code <- c(.code, paste0("cmt(", .s, ")"))
  }
  for (.p in calcSens) {
    for (.s in .states) {
      .code <- c(.code, paste0("cmt(rx__sens_", .s, "_BY_", .p, "__)"))
    }
  }
  if (!is.null(calcSens2)) {
    for (.p in calcSens) {
      for (.q in calcSens2) {
        for (.s in .states) {
          .code <- c(.code, paste0("cmt(rx__sens_", .s, "_BY_", .p, "_BY_", .q, "__)"))
        }
      }
    }
  }
  if (!is.null(calcSens3)) {
    for (.p in calcSens) {
      for (.q in calcSens2) {
        for (.r in calcSens3) {
          for (.s in .states) {
            .code <- c(.code, paste0("cmt(rx__sens_", .s, "_BY_", .p, "_BY_", .q, "_BY_", .r, "__)"))
          }
        }
      }
    }
  }

  # 4. Original block: decompose A into k_from_to / k_from_output micro-constants.
  #    NB: rxFromSE() resolves its argument by name (substitute()), so the
  #    symengine entry must be bound to a plain local first.
  for (.j in .states) {
    for (.i in .states) {
      if (.i == .j) next
      .aij <- .A[[.i]][[.j]]
      if (!.isZero(.aij)) {
        .code <- c(.code, paste0("k_", .j, "_", .i, " = ", rxFromSE(.aij)))
      }
    }
    .elim <- .elimOf(.j)
    if (!.isZero(.elim)) {
      .code <- c(.code, paste0("k_", .j, "_output = ", rxFromSE(.elim)))
    }
  }

  # 4b. Explicit Jacobian (df/dy) lines from `.A`.  matExp()/indLin() models
  # have a no-op dydt(), so the event-sensitivity Jacobian column would be
  # zero; these df()/dy() lines populate calc_jac with the known Jacobian,
  # which handle_evid reads instead for these models.  Emitted unconditionally
  # (any eventSens="jump" solve needs it).
  for (.i in .states) {
    for (.j in .states) {
      .aij <- .A[[.i]][[.j]]
      if (!.isZero(.aij)) {
        .code <- c(.code, paste0("df(", .i, ")/dy(", .j, ") = ", rxFromSE(.aij)))
      }
    }
  }

  # 5. Sensitivity blocks for each parameter.
  for (.p in calcSens) {
    .pSym <- symengine::S(.p)
    .S <- function(.s) paste0("rx__sens_", .s, "_BY_", .p, "__")
    # 5a. Diagonal block: sensitivity states obey the same dynamics as the
    #     originals (reuse the original micro-constants).
    for (.j in .states) {
      for (.i in .states) {
        if (.i == .j) next
        if (!.isZero(.A[[.i]][[.j]])) {
          .code <- c(.code, paste0("k_", .S(.j), "_", .S(.i), " = k_", .j, "_", .i))
        }
      }
      if (!.isZero(.elimOf(.j))) {
        .code <- c(.code, paste0("k_", .S(.j), "_output = k_", .j, "_output"))
      }
    }
    # 5b. Cross terms: (dA/dp) * X enter the sensitivity states as non-depleting
    #     transfers X_j -> S^p_i with rate dA[i,j]/dp (matrix entry set directly,
    #     X_j is not depleted).
    for (.j in .states) {
      for (.i in .states) {
        .dAdp <- symengine::D(.A[[.i]][[.j]], .pSym)
        if (!.isZero(.dAdp)) {
          .code <- c(.code, paste0("k_", .j, "_", .S(.i), "_nd = ", rxFromSE(.dAdp)))
        }
      }
    }
  }

  # 5c. Second-order sensitivity blocks (Hessian path, if calcSens2 given).
  # For (p, q), rx__sens_<x>_BY_<p>_BY_<q>__ obeys the total-derivative-wrt-q
  # of the first-order ODE (A = Jacobian, dAdp_ij = d(A[i,j])/dp):
  #   d(S^{pq}_i)/dt = sum_k A_ik * S^{pq}_k                 [homogeneous: reuse]
  #                  + sum_k totalD_q(A_ik)  * S^p_k          [from S^p_k]
  #                  + sum_j dAdp_ij         * S^q_j          [from S^q_j]
  #                  + sum_j totalD_q(dAdp_ij) * X_j          [from X_j]
  # totalD_q() (.rxIndLinTotalD()) is the total derivative wrt q.  Coefficients
  # are computed index-by-index (not by scanning a pre-built sum's free
  # symbols, which fails when a coefficient depends on the state it multiplies).
  if (!is.null(calcSens2)) {
    for (.p in calcSens) {
      .pSym <- symengine::S(.p)
      .S1p <- function(.s) paste0("rx__sens_", .s, "_BY_", .p, "__")
      for (.q in calcSens2) {
        .S1q <- function(.s) paste0("rx__sens_", .s, "_BY_", .q, "__")
        .S2 <- function(.s) paste0("rx__sens_", .s, "_BY_", .p, "_BY_", .q, "__")
        # homogeneous block: S^{pq} obeys the same dynamics as X / S^p (reuse).
        for (.j in .states) {
          for (.i in .states) {
            if (.i == .j) next
            if (!.isZero(.A[[.i]][[.j]])) {
              .code <- c(.code, paste0("k_", .S2(.j), "_", .S2(.i), " = k_", .j, "_", .i))
            }
          }
          if (!.isZero(.elimOf(.j))) {
            .code <- c(.code, paste0("k_", .S2(.j), "_output = k_", .j, "_output"))
          }
        }
        # cross terms.  Accumulate by (from,to): when p == q (a diagonal
        # Hessian entry), the "from S^p_k" and "from S^q_j" families collapse
        # onto the same source compartment for k == j and MUST be summed, not
        # emitted as two conflicting k_..._nd lines for the same pair.
        .acc <- .rxIndLinNdAccumulator()
        for (.i in .states) {
          for (.k in .states) {
            .c2a <- .rxIndLinTotalD(.A[[.i]][[.k]], .q, .states) # from S^p_k
            .acc$add(.S1p(.k), .S2(.i), .c2a)
          }
          for (.j in .states) {
            .dAdp <- symengine::D(.A[[.i]][[.j]], .pSym)
            .acc$add(.S1q(.j), .S2(.i), .dAdp) # from S^q_j
            .c2c <- .rxIndLinTotalD(.dAdp, .q, .states) # from X_j
            .acc$add(.j, .S2(.i), .c2c)
          }
        }
        .code <- c(.code, .acc$emit())
      }
    }
  }

  # 5d. Third-order sensitivity blocks (if calcSens3 given).  For (p, q, r),
  # rx__sens_<x>_BY_<p>_BY_<q>_BY_<r>__ obeys the total-derivative-wrt-r of the
  # second-order ODE (pieces chained one variable further with
  # .rxIndLinChainD()), for every state pair (i,k)/(i,j):
  #   homogeneous:      A_ik                     -> S^{pqr}_k   (reuse)
  #   from S^{pq}_k:    totalD_r(A_ik)
  #   from S^{pr}_k:    totalD_q(A_ik)
  #   from S^p_k:       totalD_r(totalD_q(A_ik))
  #   from S^{qr}_j:    dAdp_ij
  #   from S^q_j:       totalD_r(dAdp_ij)
  #   from S^r_j:       totalD_q(dAdp_ij)
  #   from X_j:         totalD_r(totalD_q(dAdp_ij))
  if (!is.null(calcSens3)) {
    for (.p in calcSens) {
      .pSym <- symengine::S(.p)
      .S1p <- function(.s) paste0("rx__sens_", .s, "_BY_", .p, "__")
      for (.q in calcSens2) {
        .S1q <- function(.s) paste0("rx__sens_", .s, "_BY_", .q, "__")
        .S2pq <- function(.s) paste0("rx__sens_", .s, "_BY_", .p, "_BY_", .q, "__")
        for (.r in calcSens3) {
          .S1r <- function(.s) paste0("rx__sens_", .s, "_BY_", .r, "__")
          .S2pr <- function(.s) paste0("rx__sens_", .s, "_BY_", .p, "_BY_", .r, "__")
          .S2qr <- function(.s) paste0("rx__sens_", .s, "_BY_", .q, "_BY_", .r, "__")
          .S3 <- function(.s) paste0("rx__sens_", .s, "_BY_", .p, "_BY_", .q, "_BY_", .r, "__")
          # homogeneous block: S^{pqr} obeys the same dynamics (reuse).
          for (.j in .states) {
            for (.i in .states) {
              if (.i == .j) next
              if (!.isZero(.A[[.i]][[.j]])) {
                .code <- c(.code, paste0("k_", .S3(.j), "_", .S3(.i), " = k_", .j, "_", .i))
              }
            }
            if (!.isZero(.elimOf(.j))) {
              .code <- c(.code, paste0("k_", .S3(.j), "_output = k_", .j, "_output"))
            }
          }
          # cross terms.  Accumulate by (from,to): whenever two of p/q/r
          # coincide, several of the eight families below collapse onto the
          # same source compartment (e.g. p==q==r makes S2pq(k) and S2qr(k)
          # the same name) and their coefficients must be summed.
          .acc <- .rxIndLinNdAccumulator()
          for (.i in .states) {
            for (.k in .states) {
              .Aik <- .A[[.i]][[.k]]
              .acc$add(.S2pq(.k), .S3(.i), .rxIndLinTotalD(.Aik, .r, .states)) # from S^{pq}_k
              .acc$add(.S2pr(.k), .S3(.i), .rxIndLinTotalD(.Aik, .q, .states)) # from S^{pr}_k
              .acc$add(.S1p(.k), .S3(.i), .rxIndLinChainD(.Aik, c(.q, .r), .states)) # from S^p_k
            }
            for (.j in .states) {
              .dAdp <- symengine::D(.A[[.i]][[.j]], .pSym)
              .acc$add(.S2qr(.j), .S3(.i), .dAdp) # from S^{qr}_j
              .acc$add(.S1q(.j), .S3(.i), .rxIndLinTotalD(.dAdp, .r, .states)) # from S^q_j
              .acc$add(.S1r(.j), .S3(.i), .rxIndLinTotalD(.dAdp, .q, .states)) # from S^r_j
              .acc$add(.j, .S3(.i), .rxIndLinChainD(.dAdp, c(.q, .r), .states)) # from X_j
            }
          }
          .code <- c(.code, .acc$emit())
        }
      }
    }
  }

  # 6. Preserve non-ODE output / lhs lines from the original normalized model
  #    (e.g. cp = central/v); drop structural lines we re-emit above.
  .normModel <- .mv$model["normModel"]
  .lines <- trimws(unlist(strsplit(.normModel, "[\n;]")))
  for (.l in .lines) {
    if (nzchar(.l) && !grepl("^d/dt\\(", .l) && !grepl("^cmt\\(", .l) &&
        !grepl("^matExp\\(", .l) && !grepl("^indLin\\(", .l) &&
        !grepl("^k[_.][^=]*<?=", .l)) {
      .code <- c(.code, .l)
    }
  }

  return(paste(.code, collapse = "\n"))
}

#' @rdname indLin
#' @export
rxOdeToIndLin <- indLin

#' @rdname indLin
#' @export
rxToIndLin <- indLin

#' Scaled Exponential Linear Unit (SELU) Activation Function
#'
#' @family Activation Functions
#' @param x A numeric vector. All elements must be finite and
#'   non-missing.
#' @return A numeric vector where the ReLU function has been applied
#'   to each element of `x`.
#' @author Matthew Fidler
#' @export
#' @examples
#'
#' SELU(c(-1, 0, 1, 2))
#'
#' \donttest{
#' # Can also be used in rxode2:
#' x <- rxode2({
#'    r=SELU(time)
#' })
#'
#' e <- et(c(-1, 0, 1, 2))
#'
#' rxSolve(x, e)
#' }
#'
SELU <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 4L)
}

#' Derivative of the Scaled Exponential Linear Unit (SELU) Activation Function
#' @param x A numeric vector. All elements must be finite and
#'  non-missing.
#' @return A numeric vector where the derivative of the SELU function
#'  has been applied to each element of `x`.
#' @author Matthew Fidler
#' @family Activation Functions
#' @export
#' @examples
#' dSELU(c(-1, 0, 1, 2))
#' \donttest{
#' # Can also be used in rxode2:
#' x <- rxode2({
#'   r=dSELU(time)
#' })
#' e <- et(c(-1, 0, 1, 2))
#' rxSolve(x, e)
#' }
dSELU <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 17L)
}

.createCensBox <- function(data) {
  .dat <- data
  .df <- do.call(
    rbind,
    lapply(seq_along(.dat[, 1]), function(i) {
      .xlow <- .dat$x[i] - .dat$..width[i]
      .xhi <- .dat$x[i] + .dat$..width[i]
      .low <- with(.dat, ifelse(lower[i] == ..ni[i], -Inf, lower[i]))
      .hi <- with(.dat, ifelse(upper[i] == ..pi[i], Inf, upper[i]))
      .ret <- data.frame(
        x = c(.xlow, .xlow, .xhi, .xhi),
        y = c(.low, .hi, .hi, .low),
        group = i
      )
      .ret
    })
  )
  .addCols <- setdiff(colnames(data), c(colnames(.df), "..ni", "..pi", "..delta", "..width"))
  for (.col in .addCols) {
    .df[, .col] <- rep(data[, .col], each = 4)
  }
  .df
}

.setupGroupCensBox <- function(data, params) {
  .nm <- names(data)
  .hasCens <- "cens" %in% .nm
  .hasLimit <- "limit" %in% .nm
  .hasLower <- "lower" %in% .nm
  .hasUpper <- "upper" %in% .nm
  .censSet <- .hasCens || .hasLimit
  .lowerSet <- .hasLower || .hasUpper
  if (.censSet && .lowerSet) {
    stop("stat_cens cannot mix the (cens, limit) and (lower, upper) aesthetic pairs; supply only one pair",
      call. = FALSE
    )
  }
  if (!.censSet && !.lowerSet) {
    stop("stat_cens requires either the (lower, upper) or (cens, limit) aesthetic pair",
      call. = FALSE
    )
  }
  if (.hasLimit && !.hasCens) {
    stop("stat_cens requires `cens` when the `limit` aesthetic is supplied",
      call. = FALSE
    )
  }
  if (.lowerSet && !(.hasLower && .hasUpper)) {
    stop("stat_cens requires both `lower` and `upper` aesthetics when using the (lower, upper) pair",
      call. = FALSE
    )
  }
  .dat <- data
  if (.hasCens) {
    .dat$lower <- ifelse(.dat$cens == 0, NA_real_, -Inf)
    .dat$upper <- ifelse(.dat$cens == 0, NA_real_, Inf)
    if (.hasLimit) {
      .dat$upper <- ifelse(.dat$cens == 1, .dat$y, ifelse(is.na(.dat$limit), .dat$upper, .dat$limit))
      .dat$lower <- ifelse(.dat$cens == -1, .dat$y, ifelse(is.na(.dat$limit), .dat$lower, .dat$limit))
    } else {
      .dat$upper <- ifelse(.dat$cens == 1, .dat$y, .dat$upper)
      .dat$lower <- ifelse(.dat$cens == -1, .dat$y, .dat$lower)
    }
    data <- .dat
  }
  .dat <- data[!is.na(data$lower), , drop = FALSE]
  if (length(.dat[, 1]) > 0) {
    .r <- c(data$y, data$upper[is.finite(data$upper)], data$lower[is.finite(data$lower)])
    .r <- range(.r, na.rm = TRUE)
    .rw <- .r[2] - .r[1]
    .dat$..ni <- .r[1] - 0.01 * .rw
    .dat$..pi <- .r[2] + 0.01 * .rw
    .dat$..delta <- .rw
    .r <- range(data$x, na.rm = TRUE)
    .dat$..width <- (.r[2] - .r[1]) * params$width
    .dat$lower <- ifelse(is.finite(.dat$lower), .dat$lower, .dat$..ni)
    .dat$upper <- ifelse(is.finite(.dat$upper), .dat$upper, .dat$..pi)
  }
  .dat
}

.setupCensParams <- function(data, params) {
  if (is.null(params$width)) {
    params$width <- 0.01
  }
  params
}

GeomPolygonCens <- ggplot2::ggproto("GeomPolygonCens", ggplot2::GeomPolygon,
  default_aes = aes(
    colour = NA, fill = "red", size = 0.5, linetype = 1,
    alpha = 0.5
  )
)

StatCens <- ggplot2::ggproto("StatCens", ggplot2::Stat,
  setup_params = function(data, params) {
    .setupCensParams(data, params)
  },
  setup_data = function(data, params) {
    .setupGroupCensBox(data, params)
  },
  compute_panel = function(data, scales, width) {
    .createCensBox(data)
  },
  compute_group = function(data, scales, width) {
    .createCensBox(data)
  },
  required_aes = c("x", "y"),
  optional_aes = c("cens", "limit", "lower", "upper")
)

#' Censoring geom/stat
#'
#' This is a censoring geom that shows the left or right censoring specified in the nlmixr input data-set or fit
#'
#' Requires the following aesthetics:
#'
#'  - `x` represents the independent variable, often the time scale
#'  - `y` represents the dependent variable
#'
#' Plus exactly one of these aesthetic forms (the two forms cannot be mixed):
#'
#'  - `lower` and `upper` give the lower and upper bounds of the censoring
#'    interval directly. Both must be supplied. Use `-Inf`/`Inf` for an
#'    unbounded side; use `NA` on both for observed (non-censored) rows.
#'  - `cens` (with optional `limit`) gives the censoring indicator
#'    (`-1` right censored, `0` not censored, `1` left censored) and, when
#'    `limit` is also supplied, the associated limit value. Internally
#'    converted to `lower`/`upper`.
#'
#' Will add boxes representing the areas of the fit that were censored.
#'
#' @param width represents the width (in \% of range) of the
#'   censoring box
#' @inheritParams ggplot2::stat_identity
#' @return This returns a ggplot2 stat
#' @export
stat_cens <- function(mapping = NULL, data = NULL,
                      position = "identity", show.legend = NA,
                      inherit.aes = TRUE, width = 0.01, ...) {
  ggplot2::layer(
    stat = StatCens, data = data, mapping = mapping, geom = GeomPolygonCens,
    position = position, show.legend = show.legend, inherit.aes = inherit.aes,
    params = list(na.rm = TRUE, width = width, ...)
  )
}
#' @rdname stat_cens
#' @export
geom_cens <- stat_cens

rxTest({

  test_that(".setupGroupCensBox accepts lower/upper pair and produces polygon-ready data", {
    d <- data.frame(
      x = c(1, 2, 3),
      y = c(1, 2, 3),
      lower = c(NA_real_, -Inf, 2),
      upper = c(NA_real_, 0.5, Inf)
    )
    out <- .setupGroupCensBox(d, list(width = 0.01))
    expect_true(all(c("..ni", "..pi", "..width") %in% names(out)))
    # NA-lower rows are dropped (observed rows) -- two censored rows remain
    expect_equal(nrow(out), 2L)
  })

  test_that(".setupGroupCensBox cens/limit pair produces same lower/upper as direct form", {
    d_cl <- data.frame(
      x = c(1, 2, 3),
      y = c(1, 2, 3),
      cens = c(0, 1, -1),
      limit = c(NA_real_, 0.5, 5)
    )
    d_lu <- data.frame(
      x = c(1, 2, 3),
      y = c(1, 2, 3),
      lower = c(NA_real_, 0.5, 3),
      upper = c(NA_real_, 2,   5)
    )
    out_cl <- .setupGroupCensBox(d_cl, list(width = 0.01))
    out_lu <- .setupGroupCensBox(d_lu, list(width = 0.01))
    expect_equal(out_cl$lower, out_lu$lower)
    expect_equal(out_cl$upper, out_lu$upper)
  })

  test_that(".setupGroupCensBox rejects mixing the two aesthetic pairs", {
    d <- data.frame(x = 1, y = 1, cens = 1, limit = 0, lower = -Inf, upper = 0)
    expect_error(
      .setupGroupCensBox(d, list(width = 0.01)),
      regexp = "cannot mix"
    )
  })

  test_that(".setupGroupCensBox accepts cens alone (limit optional)", {
    d <- data.frame(x = 1:3, y = c(1, 2, 3), cens = c(0, 1, -1))
    out <- .setupGroupCensBox(d, list(width = 0.01))
    # cens == 0 row is dropped (lower == NA); the other two remain
    expect_equal(nrow(out), 2L)
  })

  test_that(".setupGroupCensBox rejects limit without cens", {
    d_limit_only <- data.frame(x = 1:2, y = 1:2, limit = c(NA_real_, 0.5))
    expect_error(
      .setupGroupCensBox(d_limit_only, list(width = 0.01)),
      regexp = "requires `cens` when the `limit`"
    )
  })

  test_that(".setupGroupCensBox rejects partial lower/upper pair", {
    d_lower_only <- data.frame(x = 1:2, y = 1:2, lower = c(NA_real_, -Inf))
    expect_error(
      .setupGroupCensBox(d_lower_only, list(width = 0.01)),
      regexp = "both `lower` and `upper`"
    )
    d_upper_only <- data.frame(x = 1:2, y = 1:2, upper = c(NA_real_, 0.5))
    expect_error(
      .setupGroupCensBox(d_upper_only, list(width = 0.01)),
      regexp = "both `lower` and `upper`"
    )
  })

  test_that(".setupGroupCensBox rejects when neither pair is given", {
    d <- data.frame(x = 1:2, y = 1:2)
    expect_error(
      .setupGroupCensBox(d, list(width = 0.01)),
      regexp = "either the \\(lower, upper\\) or \\(cens, limit\\)"
    )
  })

  test_that("geom_cens does not warn about unknown aesthetics", {
    d <- data.frame(
      x = 1:5,
      y = c(1, 2, 1, 2, 3),
      lower = c(NA_real_, -Inf, NA_real_, 3, NA_real_),
      upper = c(NA_real_, 0.5,  NA_real_, Inf, NA_real_)
    )
    p <- ggplot2::ggplot(d, ggplot2::aes(x = x, y = y)) +
      ggplot2::geom_point() +
      geom_cens(ggplot2::aes(lower = lower, upper = upper))
    expect_no_warning(suppressMessages(ggplot2::ggplot_build(p)))

    d2 <- data.frame(
      x = 1:5,
      y = c(1, 2, 1, 2, 3),
      cens = c(0, 1, 0, -1, 0),
      limit = c(NA_real_, 0.5, NA_real_, 3, NA_real_)
    )
    p2 <- ggplot2::ggplot(d2, ggplot2::aes(x = x, y = y)) +
      ggplot2::geom_point() +
      geom_cens(ggplot2::aes(cens = cens, limit = limit))
    expect_no_warning(suppressMessages(ggplot2::ggplot_build(p2)))
  })

})

#' Replace the mu referencing covariates  with 0
#'
#' @param x Expression to change
#' @param muRefCovariateDataFrame This is the mu referenced data frame from the ui
#' @return expression with mu referenced covariates replaced with zero
#' @author Matthew L. Fidler
#' @noRd
.dropMuCovs <- function(x, muRefCovariateDataFrame) {
  if (is.call(x)) {
    if (identical(x[[1]], quote(`*`)) &&
          length(x) == 3) {
      if (is.name(x[[2]]) && is.name(x[[3]])) {
        .v1 <- as.character(x[[2]])
        .v2 <- as.character(x[[3]])
        .w <- which(muRefCovariateDataFrame$covariate == .v1 & muRefCovariateDataFrame$covariateParameter == .v2)
        if (length(.w) == 1) return(0)
        .w <- which(muRefCovariateDataFrame$covariate == .v2 & muRefCovariateDataFrame$covariateParameter == .v1)
        if (length(.w) == 1) return(0)
      }
    }
    return(as.call(lapply(x, .dropMuCovs, muRefCovariateDataFrame=muRefCovariateDataFrame)))
  } else {
    return(x)
  }
}
#' This determines if the expression has a covariate that needs to be downgraded
#'
#' @param x Expression with mu-referenced covariate expressions removed
#' @param covariates character vector of known mu-referenced covariates
#' @param env Environment to store covariates that need to be downgraded.  Needs to start with `env$covs <-NULL`
#' @return Original expression, but called for side effects
#' @details
#' The `env$covs` will have a list of covariates that need to be degraded (possibly not unique)
#' @author Matthew L. Fidler
#' @noRd
.hasACovariateToDowngrade <- function(x, covariates, env) {
  if (is.call(x)) {
    return(as.call(lapply(x, .hasACovariateToDowngrade, covariates, env)))
  } else if (is.name(x)) {
    .n <- as.character(x)
    if (.n %in% covariates) {
      assign("covs", c(.n, env$covs), envir=env)
    }
    return(x)
  }
  x
}
#' Downgrade theta and eta in `$muRefCurEval` to additive expressions
#'
#' @param theta the population parameter estimate
#' @param ui rxode2 ui, that will be updated
#' @return Nothing called for side effects
#' @author Matthew L. Fidler
#' @noRd
.muRefDowngradeThetaAndEta <- function(theta, ui) {
  .mu <- ui$muRefDataFrame
  .mu <- .mu[.mu$theta == theta, ]
  .curEval <- ui$muRefCurEval
  .w <- which(.curEval$parameter == theta)
  if (length(.w) > 0) {
    .curEval[.w, "curEval"] <- ""
    .curEval[.w, "low"] <- NA_real_
    .curEval[.w, "hi"] <- NA_real_
  }
  if (length(.mu$eta) > 0) {
    .w <- which(.curEval$parameter %in% .mu$eta)
    if (length(.w) > 0) {
      .curEval[.w, "curEval"] <- ""
      .curEval[.w, "low"] <- NA_real_
      .curEval[.w, "hi"] <- NA_real_
    }
  }
  assign("muRefCurEval", .curEval, envir=ui)
  invisible()
}

#'  Downgrade covariate expression for covariate `cov`
#'
#'
#' @param cov Covariate to downgrade the expression
#' @param ui rxode2 ui
#' @return Nothing called for side effects
#' @author Matthew L. Fidler
#' @noRd
.muRefDowngradeForCovariate <- function(cov, ui) {
  .covDf <- ui$muRefCovariateDataFrame
  .covDf <- .covDf[.covDf$covariate == cov, ]
  lapply(.covDf$theta, .muRefDowngradeThetaAndEta, ui=ui)
  .covDf <- .covDf[.covDf$covariate != cov,, drop = FALSE]
  assign("muRefCovariateDataFrame", .covDf, envir=ui)
  invisible()
}
#' Cleanup after the mu reference downgrade
#'
#' @param ui rxode2 ui
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @noRd
.muRefDowngradeCleanup <- function(ui) {
  .names <- ui$iniDf$name[is.na(ui$iniDf$err)]
  .namesCurEval <- ui$muRefCurEval$parameter
  assign("muRefCurEval", do.call("rbind", c(list(ui$muRefCurEval), lapply(setdiff(.names, .namesCurEval), function(x) {
    data.frame(parameter=x, curEval="", low=NA_real_, hi=NA_real_)
  }))), envir=ui)
  .etaNames <- ui$iniDf$name[which(ui$iniDf$neta1 == ui$iniDf$neta2 & ui$iniDf$condition == "id")]
  .extra <- setdiff(.etaNames, ui$muRefDataFrame$eta)
  .extra <- setdiff(.extra, ui$nonMuEtas)
  if (length(.extra) > 0) {
    warning("some etas defaulted to non-mu referenced, possible parsing error: ",
            paste(.extra, collapse=", "),
            "\nas a work-around try putting the mu-referenced expression on a simple line",
            call.=FALSE)
    assign("nonMuEtas", c(ui$nonMuEtas, .extra), envir=ui)
  }
  invisible()
}

#' Downgrade mu referenced covariates
#'
#' When there is a mu referenced covariate and that covariate is also
#' used in a non-mu referenced expression, downgrade all the mu
#' referenced expression that relate to that covariate.
#'
#' This will also:
#'
#'  - add any missing parameters to the `$muRefDataFrame`
#'
#'  - add any etas that are not found in the mu-referenced
#'    expression to nonMuEtas (likely an error in parsing somewhere)
#'
#' @param ui rxode2 ui
#'
#' @return Nothing, called for side effects
#'
#' @author Matthew L. Fidler
#'
#' @noRd
.muRefDowngrade <- function(ui) {
  .covData <- ui$muRefCovariateDataFrame
  if (length(.covData$covariate) == 0) return(.muRefDowngradeCleanup(ui))
  # First drop any mu referenced covariates
  .lst2 <- lapply(ui$lstExpr, .dropMuCovs, muRefCovariateDataFrame=.covData)
  # now see if any of the covariates are still in the model
  .env <- new.env(parent=emptyenv())
  .env$covs <- NULL
  lapply(.lst2, .hasACovariateToDowngrade, unique(.covData$covariate), env=.env)
  .down <- unique(.env$covs)
  if (length(.down) == 0) return(.muRefDowngradeCleanup(ui))
  lapply(.down, .muRefDowngradeForCovariate, ui=ui)
  .muRefDowngradeCleanup(ui)
}

#' Bind the study parameters and individual parameters
#'
#' @param studyParameters These are the study parameters, often can
#'   be generated by sampling from a population.  This can be either
#'   a matrix or a data frame
#' @param individualParameters A data frame of individual parameters
#' @return Data frame that can be used in rxode2 simulations
#' @author Matthew Fidler
#' @examples
#'
#' # Function for coverting coefficient of covariance into a variance
#' lognCv <- function(x){log((x/100)^2+1)}
#'
#' set.seed(32)
#'
#' nSub  <-  100
#' nStud  <-  10
#'
#' #define theta
#' theta <- c(lka=log(0.5), # log ka
#'           lCl=log(5), # log Cl
#'           lV=log(300) # log V
#'           )
#'
#' #define theta Matrix
#' thetaMat <- lotri(lCl ~ lognCv(5),
#'                  lV  ~ lognCv(5),
#'                  lka ~ lognCv(5))
#'
#'
#' nev <- nSub*nStud
#'
#' ev1 <- data.frame(COV1=rnorm(nev,50,30),COV2=rnorm(nev,75,10),
#'                   COV3=sample(c(1.0,2.0),nev,replace=TRUE))
#'
#'
#' tmat <-rxRmvn(nStud, theta[dimnames(thetaMat)[[1]]], thetaMat)
#'
#' rxCbindStudyIndividual(tmat, ev1)
#'
#' @export
rxCbindStudyIndividual <- function(studyParameters, individualParameters) {
  if (!inherits(individualParameters, "data.frame")) {
    stop("'individualParameters' must  be a data frame", call. = FALSE)
  }
  if (inherits(studyParameters, "matrix")) {
    .dimnames <- dimnames(studyParameters)
    if (is.null(.dimnames)) {
      stop("'studyParameters' must be matrix with named columns", call. = FALSE)
    }
    .dimnames <- .dimnames[[2]]
    if (is.null(.dimnames)) {
      stop("'studyParameters' must be matrix with named columns", call. = FALSE)
    }
  } else if (!inherits(studyParameters, "data.frame")) {
    stop("'studyParameters' needs to be either a data frame or a matrix", call. = FALSE)
  }
  .Call(`_rxode2_rxCbindStudyIndividual`, studyParameters, individualParameters)
}

#' @export
confint.rxSolve <- function(object, parm = NULL, level = 0.95, ...) {
  sim.id <- id <- NULL # rcheck nonsense
  .SD <- . <- NULL # nolint
  `:=` <- NULL # nolint
  rxode2::rxReq("data.table")
  checkmate::assertNumeric(level, lower=0, upper=1, finite=TRUE, any.missing=FALSE)
  .args <- list(...)
  if (any(names(.args) == "doSim")) {
    checkmate::assertLogical(.args$doSim, len=1,
                             any.missing=FALSE, .var.name="doSim")
    .doSim <- .args$doSim
  } else {
    .doSim<-TRUE
  }
  .by <- NULL
  if (any(names(.args) == "by")) {
    .by <- .args$by
    checkmate::assertSubset(.by, names(object), .var.name="by")
  }
  .ci <- level
  if (any(names(.args) == "ci")) {
    .ci <- .args$ci
    if (inherits(.ci, "logical")) {
      checkmate::assertLogical(.ci, len=1, any.missing=FALSE, .var.name="ci")
      if (!.ci) {
        .ci <- 0.0
      }
    } else {
      checkmate::assertNumeric(.ci, lower=0, upper=1, finite=TRUE, any.missing=FALSE, .var.name="ci")
    }
  }
  .mean <- FALSE
  .binom <- FALSE
  .nC <- 0L
  .pred <- FALSE
  .useT <- TRUE
  .mM <- 500000
  .tol <- .Machine$double.eps^0.25
  if (any(names(.args) == "useT")) {
    .useT <- .args$useT
    checkmate::assertLogical(.useT, len=1, any.missing=FALSE, .var.name="useT")
  }
  if (any(names(.args) == "mean")) {
    .mean <- .args$mean
    if (inherits(.mean, "character") &&
          length(.mean) == 1L &&
          .mean == "binom") {
      .binom <- TRUE
      .mean <- FALSE
    } else {
      checkmate::assertLogical(.mean, len=1, any.missing=FALSE, .var.name="mean")
    }
  }
  if (any(names(.args) == "pred")) {
    .pred <- .args$pred
    checkmate::assertLogical(.pred, len=1, any.missing=FALSE, .var.name="pred")
  }
  if (any(names(.args) == "n")) {
    .nC <- unique(.args$n)
    checkmate::assertIntegerish(.nC, len=1, any.missing=FALSE, lower=0L, .var.name="n")
  }
  if (any(names(.args) == "m")) {
    .mM <- unique(.args$m)
    checkmate::assertIntegerish(.mM, len=1, any.missing=FALSE, lower=0L, .var.name="m")
  }
  if (any(names(.args) == "M")) {
    .mM <- unique(.args$M)
    checkmate::assertIntegerish(.mM, len=1, any.missing=FALSE, lower=1000L, .var.name="M")
  }
  if (any(names(.args) == "tol")) {
    .tol <- unique(.args$tol)
    checkmate::assertNumeric(.tol, len=1, any.missing=FALSE, lower=.Machine$double.eps, .var.name="tol")
  }
  .ciMethod <- "wald"
  if (any(names(.args) == "ciMethod")) {
    .ciMethod <- .args$method
  }
  .stk <- rxStack(object, parm, doSim=.doSim) # nolint
  if (!any(names(.stk) == "id") &&
        any(names(.stk) == "sim.id")) {
    names(.stk) <- gsub("sim.id", "id", names(.stk))
  }
  for (.v in .by) {
    .stk[[.v]] <- object[[.v]]
  }
  setDT(.stk) # nolint
  .a <- (1 - level) / 2
  .p <- c(.a, 0.5, 1 - .a)
  .c <- (1-.ci) / 2
  .p2 <- c(.c, 0.5, 1 - .c)
  .lst <- list(
    lvl = paste0("p", .p * 100),
    ci = paste0("p", .p2 * 100),
    parm = levels(.stk$trt),
    by = .by,
    mean = .mean,
    binom=.binom
  )
  class(.lst) <- "rxHidden"
  if (.ci ==0 || !any(names(.stk) == "sim.id") ||
        !isTRUE(object$env$.args$nStud > 1L)) {
    if (any(names(.stk) == "sim.id")) {
      .stk$id <- factor(paste(.stk$sim.id, .stk$id))
      .ntot <- length(levels(.stk$id))
      .stk$id <- as.integer(.stk$id)
    } else {
      .ntot <- object$env$.args$nSub
      if (.ntot == 1L && object$env$.args$nStud > 1L) {
        .ntot <- object$env$.args$nStud
      }
    }
    if (.ci == 0 || .ntot < 2500) {
      if (.ci != 0.0) {
        .mwarn("in order to put confidence bands around the intervals, you need at least 2500 simulations") # nolint
      }
      message("summarizing data...", appendLF = FALSE)
      if (.mean) {
        .stk <- .stk[, list(
          p1 = .p, eff = rxode2::meanProbs(.SD$value, probs = .p, na.rm = TRUE, useT=.useT,
                                           n=.nC, pred=.pred),
          Percentile = sprintf("%s%%", .p * 100)
        ),
        by = c("time", "trt", .by)
        ]
      } else if (.binom) {
        .stk <- .stk[, list(
          p1 = .p, eff = rxode2::binomProbs(.SD$value, probs = .p, na.rm = TRUE,
                                            n=.nC, m=.mM, M=.mM, tol=.tol,
                                            pred=.pred, ciMethod=.ciMethod),
          Percentile = sprintf("%s%%", .p * 100)
        ),
        by = c("time", "trt", .by)
        ]
      } else {
        .stk <- .stk[, list(
          p1 = .p, eff = stats::quantile(.SD$value, probs = .p, na.rm = TRUE),
          Percentile = sprintf("%s%%", .p * 100)
        ),
        by = c("time", "trt", .by)
        ]
      }
      if (requireNamespace("tibble", quietly = TRUE)) {
        .stk <- tibble::as_tibble(.stk)
      }
      .cls <- c("rxSolveConfint1", class(.stk))
      attr(.cls, ".rx") <- .lst
      class(.stk) <- .cls
      message("done")
      return(.stk)
    } else {
      .n <- round(sqrt(.ntot))
      if (!any(names(.stk) == "sim.id")) {
        .stk$sim.id <- .stk$id
      }
    }
  } else {
    .n <- object$env$.args$nStud
  }
  message("summarizing data...", appendLF = FALSE)
  .ret <- .stk[, id := sim.id %% .n]
  if (.mean) {
    .ret <- .ret[, list(p1 = .p,
                        eff = rxode2::meanProbs(.SD$value, probs = .p, na.rm = TRUE, n=.nC,
                                                useT=.useT,
                                                pred=.pred)),
                 by = c("id", "time", "trt", .by)]
  } else if (.binom) {
    .ret <- .ret[, list(p1 = .p,
                        eff = rxode2::binomProbs(.SD$value, probs = .p, na.rm = TRUE,
                                                 n=.nC, m=.mM, M=.mM, tol=.tol,
                                                 pred=.pred, ciMethod=.ciMethod)),
                 by = c("id", "time", "trt", .by)]
  } else {
    .ret <- .ret[, list(p1 = .p,
                        eff = stats::quantile(.SD$value, probs = .p, na.rm = TRUE)), by = c("id", "time", "trt", .by)]
  }
  .ret <- .ret[, setNames(as.list(stats::quantile(.SD$eff, probs = .p2, na.rm = TRUE)),
                          sprintf("p%s", .p2 * 100)),
               by = c("p1", "time", "trt", .by)]
  .ret$Percentile <- factor(sprintf("%s%%", .ret$p1 * 100))
  if (requireNamespace("tibble", quietly = TRUE)) {
    .ret <- tibble::as_tibble(.ret)
  }
  message("done")
  .cls <- c("rxSolveConfint2", class(.ret))
  attr(.cls, ".rx") <- .lst
  class(.ret) <- .cls
  .ret
}

.parseEnv$.parseFuns <- c("gammafn", "lgammafn", "lgamma", "loggamma", "digamma", "trigamma", 
"tetragamma", "pentagamma", "cospi", "sinpi", "tanpi", "log1p", 
"expm1", "factorial", "lfactorial", "lgamma1p", "log10", "log2", 
"log1pexp", "phi", "pnorm", "normcdf", "qnorm", "fabs", "pow", 
"R_pow", "R_pow_di", "Rx_pow_di", "Rx_pow", "lbeta", "%%", "abs", 
"acos", "acosh", "asin", "asinh", "atan", "atan2", "atanh", "beta", 
"cos", "cosh", "erf", "erfc", "exp", "gamma", "log", "polygamma", 
"rxTBS", "rxTBSi", "rxTBSd", "rxTBSd2", "rxTBSdL", "rxTBSdL2", 
"rxTBSdLx", "sin", "sinh", "sqrt", "tan", "tanh", "gammap", "floor", 
"round", "ceil", "trunc", "bessel_i", "bessel_j", "bessel_k", 
"bessel_y", "logspace_add", "logspace_sub", "fmax2", "fmin2", 
"sign", "fsign", "fprec", "fround", "ftrunc", "transit", "gammaq", 
"gammapDer", "gammapInv", "gammapInva", "gammaqInv", "gammaqInva", 
"lowergamma", "uppergamma", "linCmtA", "linCmtB", "max", "min", 
"logit", "expit", "probit", "probitInv", "tlast", "tlast0", "tfirst", 
"tfirst0", "lag", "lead", "lag0", "lead0", "dose", "podo", "dose0", 
"podo0", "dabs", "dabs2", "abs1", "dabs1", "erfinv", "abs0", 
"dosenum", "first", "last", "diff", "diff0", "is.nan", "is.na", 
"is.finite", "is.infinite", "llikPois", "llikPoisDlambda", "llikBinom", 
"llikBinomDprob", "llikNbinom", "llikNbinomDprob", "llikNbinomMu", 
"llikNbinomMuDmu", "llikBeta", "llikBetaDshape1", "llikBetaDshape2", 
"llikT", "llikTDdf", "llikTDmean", "llikTDsd", "llikChisq", "llikChisqDdf", 
"llikExp", "llikExpDrate", "llikF", "llikFDdf1", "llikFDdf2", 
"llikGeom", "llikGeomDprob", "llikUnif", "llikUnifDalpha", "llikUnifDbeta", 
"llikWeibull", "llikWeibullDshape", "llikWeibullDscale", "llikGamma", 
"llikGammaDshape", "llikGammaDrate", "llikCauchy", "llikCauchyDlocation", 
"llikCauchyDscale", "llikNorm", "llikNormDmean", "llikNormDsd", 
"llikXPois", "llikXPoisDlambda", "llikXBinom", "llikXBinomDprob", 
"llikXNbinomMu", "llikXNbinomMuDmu", "llikXNbinom", "llikXNbinomDprob", 
"llikXBeta", "llikXBetaDshape1", "llikXBetaDshape2", "llikXT", 
"llikXTDdf", "llikXTDmean", "llikXTDsd", "llikXChisq", "llikXChisqDdf", 
"llikXExp", "llikXExpDrate", "llikXF", "llikXFDdf1", "llikXFDdf2", 
"llikXGeom", "llikXGeomDprob", "llikXUnif", "llikXUnifDalpha", 
"llikXUnifDbeta", "llikXWeibull", "llikXWeibullDshape", "llikXWeibullDscale", 
"llikXGamma", "llikXGammaDshape", "llikXGammaDrate", "llikXCauchy", 
"llikXCauchyDlocation", "llikXCauchyDscale", "llikXNorm", "llikXNormDmean", 
"llikXNormDsd", "ReLU", "dReLU", "GELU", "dGELU", "d2GELU", "d3GELU", 
"d4GELU", "ELU", "dELU", "d2ELU", "d2aELU", "dELUa", "d2ELUa", 
"softplus", "dsoftplus", "d2softplus", "d3softplus", "d4softplus", 
"SELU", "dSELU", "lReLU", "dlReLU", "PReLU", "dPReLU", "d2PReLU", 
"dPReLUa", "dPReLUa1", "Swish", "dSwish", "linCmt", "rnorm", 
"rxnorm", "rxbinom", "rbinom", "rxcauchy", "rcauchy", "rchisq", 
"rxchisq", "rexp", "rxexp", "rbeta", "rxbeta", "rgeom", "rxgeom", 
"rxpois", "rpois", "rxt", "rt")
.parseEnv$.parseNum <- c(lgamma = 1, abs = 1, acos = 1, acosh = 1, asin = 1, asinh = 1, 
atan = 1, atan2 = 2, atanh = 1, beta = 2, cos = 1, cosh = 1, 
erf = 1, erfc = 1, exp = 1, gamma = 1, log = 1, polygamma = 2, 
rxTBS = 5, rxTBSi = 5, rxTBSd = 5, rxTBSd2 = 5, rxTBSdL = 5, 
rxTBSdL2 = 5, rxTBSdLx = 5, sin = 1, sinh = 1, sqrt = 1, tan = 1, 
tanh = 1, gammap = 2, floor = 1, round = 1, ceil = 1, trunc = 1, 
bessel_i = 3, bessel_j = 2, bessel_k = 3, bessel_y = 2, logspace_add = 2, 
logspace_sub = 2, fmax2 = 2, fmin2 = 2, sign = 1, fsign = 2, 
fprec = 2, fround = 2, ftrunc = 2, transit = NA, gammaq = 2, 
gammapDer = 2, gammapInv = 2, gammapInva = 2, gammaqInv = 2, 
gammaqInva = 2, lowergamma = 2, uppergamma = 2, linCmtA = 14, 
linCmtB = 15, max = NA, min = NA, logit = NA, expit = NA, probit = NA, 
probitInv = NA, tlast = NA, tlast0 = NA, tfirst = NA, tfirst0 = NA, 
lag = NA, lead = NA, lag0 = NA, lead0 = NA, dose = NA, podo = NA, 
dose0 = NA, podo0 = NA, dabs = 1, dabs2 = 1, abs1 = 1, dabs1 = 1, 
erfinv = 1, abs0 = 1, dosenum = 0, first = 1, last = 1, diff = 1, 
diff0 = 1, is.nan = 1, is.na = 1, is.finite = 1, is.infinite = 1, 
llikPois = 2, llikPoisDlambda = 2, llikBinom = 3, llikBinomDprob = 3, 
llikNbinom = 3, llikNbinomDprob = 3, llikNbinomMu = 3, llikNbinomMuDmu = 3, 
llikBeta = 3, llikBetaDshape1 = 3, llikBetaDshape2 = 3, llikT = 4, 
llikTDdf = 4, llikTDmean = 4, llikTDsd = 4, llikChisq = 2, llikChisqDdf = 2, 
llikExp = 2, llikExpDrate = 2, llikF = 3, llikFDdf1 = 3, llikFDdf2 = 3, 
llikGeom = 2, llikGeomDprob = 2, llikUnif = 3, llikUnifDalpha = 3, 
llikUnifDbeta = 3, llikWeibull = 3, llikWeibullDshape = 3, llikWeibullDscale = 3, 
llikGamma = 3, llikGammaDshape = 3, llikGammaDrate = 3, llikCauchy = 3, 
llikCauchyDlocation = 3, llikCauchyDscale = 3, llikNorm = 3, 
llikNormDmean = 3, llikNormDsd = 3, llikXPois = 3, llikXPoisDlambda = 3, 
llikXBinom = 4, llikXBinomDprob = 4, llikXNbinomMu = 4, llikXNbinomMuDmu = 4, 
llikXNbinom = 4, llikXNbinomDprob = 4, llikXBeta = 4, llikXBetaDshape1 = 4, 
llikXBetaDshape2 = 4, llikXT = 5, llikXTDdf = 5, llikXTDmean = 5, 
llikXTDsd = 5, llikXChisq = 3, llikXChisqDdf = 3, llikXExp = 3, 
llikXExpDrate = 3, llikXF = 4, llikXFDdf1 = 4, llikXFDdf2 = 4, 
llikXGeom = 3, llikXGeomDprob = 3, llikXUnif = 4, llikXUnifDalpha = 4, 
llikXUnifDbeta = 4, llikXWeibull = 4, llikXWeibullDshape = 4, 
llikXWeibullDscale = 4, llikXGamma = 4, llikXGammaDshape = 4, 
llikXGammaDrate = 4, llikXCauchy = 4, llikXCauchyDlocation = 4, 
llikXCauchyDscale = 4, llikXNorm = 4, llikXNormDmean = 4, llikXNormDsd = 4, 
ReLU = 1, dReLU = 1, GELU = 1, dGELU = 1, d2GELU = 1, d3GELU = 1, 
d4GELU = 1, ELU = 2, dELU = 2, d2ELU = 2, d2aELU = 2, dELUa = 2, 
d2ELUa = 2, softplus = 1, dsoftplus = 1, d2softplus = 1, d3softplus = 1, 
d4softplus = 1, SELU = 1, dSELU = 1, lReLU = 1, dlReLU = 1, PReLU = 2, 
dPReLU = 2, d2PReLU = 2, dPReLUa = 2, dPReLUa1 = 2, Swish = 1, 
dSwish = 1)

.udfUiEnv <- new.env(parent=emptyenv())

#' Reset the rxode2 ui environment variables
#'
#' @return NULL silently
#' @export
#'
#' @keywords internal
#' @author Matthew L. Fidler
#' @examples
#' rxUdfUiReset()
rxUdfUiReset <- function() {
  .udfUiEnv$num <- 1L
  .udfUiEnv$iniDf <- NULL
  .udfUiEnv$lhs <- NULL
  .udfUiEnv$data <- NULL
  .udfUiEnv$est <- NULL
  .udfUiEnv$control <- NULL
  .udfUiEnv$parsing <- FALSE
  .udfUiEnv$mv <- NULL
  .udfUiEnv$probs <- NULL
  invisible(NULL)
}

rxUdfUiReset()


#' This gives the current number in the ui of the particular function being called.
#'
#' If this is called outside of function parsing or the input is
#' unexpected this returns 1L. This is useful when writing replacement
#' UI functions
#'
#' @return integer greater than 1L
#' @family User functions
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' rxUdfUiNum()
rxUdfUiNum <- function() {
  if (checkmate::testIntegerish(.udfUiEnv$num, lower=1L, len=1L, any.missing=FALSE))  {
    as.integer(.udfUiEnv$num)
  } else {
    1L
  }
}

#' Is the expression actually a flag that can be used in the rxUdfUi
#' functions?
#'
#' This is useful when writing replacement
#' UI functions
#'
#' @param expr expression to evaluate
#'
#' @param arg argument name for error messages for the argument name
#'   in the `rxUdfUi` extension when the expression is not logical
#'
#' @param funName function name for the error message for the argument
#'   name in the `rxUdfUi` extension when the expression is not
#'   logical.
#'
#' @param env the environment in which to evaluate the expression (in case it is numeric)
#'
#' @return logical value of the expression if it can be evaluated to a
#'   scalar `TRUE`/`FALSE` value, otherwise an error is thrown.
#'
#' @export
#'
#' @author Matthew L. Fidler
#'
rxUdfUiFlag <- function(expr, arg="arg", funName="fun", env = baseenv()) {
  checkmate::assertCharacter(arg, len=1L, any.missing=FALSE)
  checkmate::assertCharacter(funName, len=1L, any.missing=FALSE)
  .val <- suppressWarnings(try(eval(expr, envir = env), silent = TRUE))
  if (inherits(.val, "try-error") ||
        !(checkmate::testLogical(.val, len=1L, any.missing=FALSE) ||
            checkmate::testIntegerish(.val, len=1L, any.missing=FALSE, min=0, max=1))) {
    stop(sprintf("'%s' requires '%s' to be a scalar TRUE/FALSE value in rxode2 model syntax",
                 funName,
                 arg),
         call. = FALSE)
  }
  as.logical(.val)
}

#' Get the rxode2 iniDf of the current UI being processed (or return NULL)
#'
#' @return Initial `data.frame` being processed or `NULL` for nothing.
#'
#' @export
#'
#' @author Matthew L. Fidler
#'
#' @examples
#'
#' rxUdfUiIniDf()
#'
rxUdfUiIniDf <- function() {
  if (testIniDf(.udfUiEnv$iniDf)) {
    .udfUiEnv$iniDf
  } else {
    NULL
  }
}
#' Return the lhs parsed language expression
#'
#' @return lhs language expression or NULL
#' @family User functions
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' rxUdfUiIniLhs()
#'
rxUdfUiIniLhs <- function() {
  if (is.language(.udfUiEnv$lhs)) {
    .udfUiEnv$lhs
  } else {
    NULL
  }
}
#' Is the expression actually equal to a value?
#'
#' This is used to determine if the location and scale parameters are
#' actually equal to their default values, which allows for more
#' efficient translation to expit expressions when possible.
#'
#' @param expr the R language expression to evaluate
#' @param value the value to compare against
#' @param env the environment in which to evaluate the expression
#' @return TRUE if the expression evaluates to a scalar value equal to
#'   the specified value, FALSE otherwise
#'
#' @export
#'
#' @author Matthew L. Fidler
rxUdfUiIsValue <- function(expr, value, env = baseenv()) {
  .val <- try(eval(expr, envir = env), silent = TRUE)
  if (inherits(.val, "try-error") ||
        length(.val) != 1L ||
        (!is.numeric(.val) && !is.logical(.val)) ||
        is.na(.val)) {
    return(FALSE)
  }
  identical(as.numeric(.val), as.numeric(value))
}
#' Return the model variables that is being processed or setup model
#' variables for processing
#'
#'
#' @param value when specified, this assigns the model variables to be
#'   processed, or resets it by assigning it to be `NULL`.
#'
#' @return value of the `modelVariables` being processed or `NULL`.
#'
#' @family User functions
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' rxUdfUiMv()
#'
rxUdfUiMv <- function(value) {
  if (missing(value)) {
    .udfUiEnv$mv
  } else if (inherits(value, "rxModelVars")) {
    .udfUiEnv$mv <- value
  } else if (is.null(value)) {
    .udfUiEnv$mv <- value
  } else {
    stop("rxUdfUiMt must be called with model variables, NULL, or without any arguments",
         call.=FALSE)
  }
}
#' Return the data.frame that is being processed or setup data.frame for processing
#'
#'
#' @param value when specified, this assigns the data.frame to be processed, or resets it by assigning it to be `NULL`.
#'
#' @return value of the `data.frame` being processed or `NULL`.
#'
#' @export
#' @family User functions
#' @author Matthew L. Fidler
#' @examples
#'
#' rxUdfUiData()
#'
rxUdfUiData <- function(value) {
  if (missing(value)) {
    .udfUiEnv$data
  } else if (is.data.frame(value)) {
    .udfUiEnv$data <- value
  } else if (is.null(value)) {
    .udfUiEnv$data <- value
  } else {
    stop("rxUdfUiData must be called with a data.frame, NULL, or without any arguments",
         call.=FALSE)
  }
}


#' Give the expression as a compressed model or expression
#'
#' Here it means that it evaluates to a variable or a number.
#'
#' @param expr the R language expression to evaluate
#'
#' @param env the environment in which to evaluate the expression
#'
#' @return expression
#'
#' @export
#'
#' @author Matthew L. Fidler
#'
rxUdfUiExpr <- function(expr, env = parent.frame()) {
  .val <- suppressWarnings(try(eval(expr, envir = env), silent = TRUE))
  if (!inherits(.val, "try-error") &&
        length(.val) == 1L &&
         (is.numeric(.val) || is.character(.val))) {
    return(str2lang(as.character(.val)))
  }
  expr
}

#' Return the control that is being processed or setup control for processing
#'
#' @param value when specified, this assigns the control to be
#'   processed, or resets it by assigning it to be `NULL`.
#'
#' @return value of the `data.frame` being processed or `NULL`.
#'
#' @export
#' @family User functions
#' @author Matthew L. Fidler
#' @examples
#'
#' rxUdfUiControl()
#'
rxUdfUiControl <- function(value) {
  if (missing(value)) {
    .udfUiEnv$control
  } else if (is.list(value)) {
    .udfUiEnv$control <- value
  } else if (is.null(value)) {
    .udfUiEnv$control <- value
  } else {
    stop("rxUdfUiControl must be called with a list, NULL, or without any arguments",
         call.=FALSE)
  }
  invisible(.udfUiEnv$control)

}
#' Return the current estimation method for the UI processing
#'
#' @param value when specified, this assigns the character value of
#'   the estimation method or NULL if there is nothing being estimated
#' @return value of the estimation method being processed or NULL
#' @family User functions
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' rxUdfUiEst()
#'
rxUdfUiEst <- function(value) {
  if (missing(value)) {
    .udfUiEnv$est
  } else if (checkmate::testCharacter(value, min.chars=1L, any.missing=FALSE, len=1L)) {
    .udfUiEnv$est <- value
  } else if (is.null(value)) {
    .udfUiEnv$est <- value
  } else {
    stop("rxUdfUiEst must be called with a character, NULL, or without any arguments",
         call.=FALSE)
  }
}
#' Returns if the current ui function is being parsed
#'
#' @return logical if the current ui function is being parsed
#' @family User functions
#' @export
#' @author Matthew L. Fidler
#' @examples
#' rxUdfUiParsing()
rxUdfUiParsing <- function() {
  .udfUiEnv$parsing
}


#' Handle User-Defined Functions in UI
#'
#' This function processes expressions to handle user-defined
#' functions in the UI.  It will see if there is any registered `s3`
#' generic in `rxUdfUi` and call that with the parsed function.  The
#' s3 generic is responsible for returning a list in the correct form
#' so that the parsed UI will be updated.
#'
#' @param expr The expression to be processed.
#' @param env The environment in which to evaluate the expression.
#' @return The processed expression.
#' @noRd
.handleUdfUi <- function(expr, env) {
  if (is.call(expr)) {
    if (length(expr) == 1L) {
      return(expr)
    }
    .c <- as.character(expr[[1]])
    .fun <- try(utils::getS3method("rxUdfUi", .c), silent=TRUE)
    if (inherits(.fun, "try-error")) {
      as.call(c(expr[[1]], lapply(expr[-1], .handleUdfUi, env=env)))
    } else {
      if (!exists(.c, envir=env$rxUdfUiCount)) {
        assign(.c, 0L, envir=env$rxUdfUiCount)
      }
      .num <- get(.c, envir=env$rxUdfUiCount) + 1L
      assign(.c, .num, envir=env$rxUdfUiCount)
      .udfUiEnv$num <- .num
      .udfUiEnv$iniDf <- env$df
      .udfUiEnv$lhs <- env$lhs
      .e <- .fun(expr)
      if (is.language(.e$replace)) {
        if (!identical(expr, .e$replace)) {
          env$redo <- TRUE
        }
        expr <- .e$replace
      } else if (length(.e$replace) == 1 &&
                   inherits(.e$replace, "character")) {
        .t <- try(str2lang(.e$replace), silent=TRUE)
        if (inherits(.t, "try-error")) {
          stop("rxode2 ui user function '", .c, "' failed to produce code that could be parsed '", .e$replace, "'",
               call.=FALSE)
        }
        if (!identical(expr, .t)) {
          env$redo <- TRUE
        }
        expr <- .t
      } else {

        stop("rxode2 ui user function '", .c, "' failed to produce code that could be parsed",
             call.=FALSE)
      }
      .handleUdifUiBeforeOrAfter("before", .e, env, .c)
      .handleUdifUiBeforeOrAfter("after", .e, env, .c)
      if (inherits(.e$iniDf, "data.frame")) {
        env$df <- .e$iniDf
      }
      if (is.null(.udfUiEnv$data) &&
            checkmate::testLogical(.e$uiUseData, len=1L, any.missing=FALSE)) {
        env$uiUseData <- .e$uiUseData
      }
      if (is.null(.udfUiEnv$mv) &&
            checkmate::testLogical(.e$uiUseMv, len=1L, any.missing=FALSE)) {
        env$uiUseMv <- .e$uiUseMv
      }
      if (!is.call(expr)) return(expr)
      expr <- as.call(c(expr[[1]], lapply(expr[-1], .handleUdfUi, env=env)))
      if (is.call(expr) && length(expr) >= 2L &&
            (identical(expr[[1]], quote(`+`)) ||
               identical(expr[[1]], quote(`-`)) ||
               identical(expr[[1]], quote(`^`)) ||
               identical(expr[[1]], quote(`/`)) ||
               identical(expr[[1]], quote(`*`)))) {
        expr <- str2lang(paste0("(", deparse1(expr), ")"))
      }
      expr
    }
  } else {
    expr
  }
}

#' This function is called when processing rxode2 user functions from
#' the models
#'
#'
#' @param fun this is the function that needs to be parsed and
#'   changed.  This is a R language expression
#'
#' @return This needs to return a list with the following elements:
#'
#' - `iniDf` -- the modified initial estimate data.frame
#'
#' - `before` -- any model code that needs to be added before the current line
#'
#' - `after` -- any model code that needs to be added after the current line
#'
#' - `replace` -- replacement code for this user function
#'
#' @export
#' @keywords internal
#' @author Matthew L. Fidler
rxUdfUi <- function(fun) {
  UseMethod("rxUdfUi")
}

#' Get the number of arguments for user defined functions for ui
#' replacement
#'
#' @param fun The rxode2 ui function to replace
#' @return The number of arguments needed for this function
#' @noRd
#' @author Matthew L. Fidler
.rxUdfUiNarg <- function(fun) {
  .cls <- try(utils::getS3method("rxUdfUi", fun), silent=TRUE)
  if (inherits(.cls, "try-error")) {
    return(NA_integer_)
  }
  .nargs <- attr(.cls, "nargs")
  if (is.null(.nargs)) {
    return(NA_integer_)
  }
  as.integer(.nargs)
}

#' Convert a positive integer to a letter series
#'
#' @param x integer to convert
#' @param base can be 2 to 26
#' @return a sequence of letters representing the number(s) input
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' rxIntToLetter(1:100)
#'
rxIntToLetter <- function(x, base=26L) {
  checkmate::testIntegerish(x, lower=0L, any.missing=FALSE)
  checkmate::testIntegerish(base, lower=2L, upper=26L, any.missing=FALSE, len=1L)
  .Call(`_rxode2_itoletter`, as.integer(x), as.integer(base), PACKAGE="rxode2")
}

#' Convert a positive  base
#'
#' @param x integer to convert
#' @param base can be 2 to 36
#' @return a sequence of letters and representing the number(s) input
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' rxIntToBase(1:100)
#'
rxIntToBase <- function(x, base=36L) {
  checkmate::testIntegerish(x, lower=0L, any.missing=FALSE)
  checkmate::testIntegerish(base, lower=2L, upper=36L, any.missing=FALSE, len=1L)
  .Call(`_rxode2_itostr`, as.integer(x), as.integer(base), PACKAGE="rxode2")
}

.handleUdifUiBeforeOrAfter <- function(type="before", e, env, fun) {
  .cur <- e[[type]]
  if (is.null(.cur)) return(invisible())
  if (is.list(.cur)) {
    .ret <- lapply(seq_along(.cur),
                   function(i) {
                     if (is.language(.cur[[i]])) {
                       .cur[[i]]
                     } else if (length(.cur[[i]]) == 1L &&
                                  inherits(.cur[[i]], "character")) {
                       .ret <- try(str2lang(.cur[[i]]), silent=TRUE)
                       if (inherits(.ret, "try-error")) {
                         stop("rxode2 ui user function '", fun, "' failed to produce code that could be parsed '", .cur[[i]],
                              "' in $",
                              type,
                              call.=FALSE)
                       }
                       .ret
                     } else {
                       stop("rxode2 ui user function '", fun, "' failed to produce code that could be parsed in $", type,
                            call.=FALSE)
                     }
                   })
    assign(type, c(get(type, env), .ret), envir=env)
  } else if (is.language(.cur)) {
    assign(type, c(get(type, env), list(.cur)), envir=env)
  } else if (inherits(.cur, "character")) {
    .ret <- lapply(seq_along(.cur),
                   function(i) {
                     .ret <- try(str2lang(.cur[[i]]), silent=TRUE)
                     if (inherits(.ret, "try-error")) {
                       stop("rxode2 ui user function '", fun, "' failed to produce code that could be parsed '", .cur[[i]],
                            "' in $",
                            type,
                            call.=FALSE)
                     }
                     .ret
                   })
    assign(type, c(get(type, env), .ret), envir=env)
  } else {
    stop("rxode2 ui user function '", fun, "' failed to produce code that could be parsed in $", type,
         call.=FALSE)
  }
}

## This is only for rxode2
## inst/include/rxode2_RcppExports.h is the header consumed by generated model
## code -- it should carry only <Rcpp.h>.  Strip RcppArmadillo and RcppEigen.
## src/RcppExports.cpp is the package's own implementation file -- it uses
## arma:: types and needs RcppArmadillo.h, but it must come BEFORE RcppEigen.h /
## Rcpp.h to satisfy newer RcppArmadillo's include-order requirement.

.strip_rcpp_guard <- function(l) {
  l <- l[regexpr("^[#]include <RcppArmadillo.h>", l) == -1]
  l <- l[regexpr("^[#]define R_STRICT_HEADERS", l) == -1]
  w <- which(regexpr("^#ifndef RCPP_H$", l) != -1)
  if (length(w) > 0) {
    .start <- w[1]
    .end <- which(regexpr("^#endif", l[seq(.start, length(l))]) != -1)[1] + .start - 1L
    if (!is.na(.end)) l <- l[-seq(.start, .end)]
  }
  l
}

## Header: strip RcppArmadillo, RcppEigen, and any stale guards.
## Generated model code needs only <Rcpp.h>.
.hdr_f <- "inst/include/rxode2_RcppExports.h"
.hdr_l <- .strip_rcpp_guard(readLines(.hdr_f))
.hdr_l <- .hdr_l[regexpr("^[#]include <RcppEigen.h>", .hdr_l) == -1]
.hdr_out <- file(.hdr_f, "wb")
writeLines(.hdr_l, .hdr_out)
close(.hdr_out)

## Implementation: clean up stale guards, then ensure RcppArmadillo.h appears
## BEFORE rxode2.h (which includes R.h / Rinternals.h).  RcppCommon.h defines
## R_NO_REMAP before R headers are pulled in, suppressing macros like `length`
## that would otherwise conflict with Armadillo / STL names.
.cpp_f <- "src/RcppExports.cpp"
.cpp_l <- .strip_rcpp_guard(readLines(.cpp_f))
if (!any(grepl("RcppArmadillo", .cpp_l, fixed = TRUE))) {
  .rx_line <- which(regexpr("#include.*rxode2\\.h", .cpp_l) != -1)
  if (length(.rx_line) > 0) {
    .cpp_l <- append(.cpp_l, "#include <RcppArmadillo.h>", after = .rx_line[1] - 1L)
  }
}
.cpp_out <- file(.cpp_f, "wb")
writeLines(.cpp_l, .cpp_out)
close(.cpp_out)

l <- readLines("R/RcppExports.R")
w <- which(regexpr("# Register entry points", l, fixed=TRUE) != -1)
if (length(w) >= 1) {
  w <- w[1]
  l <- l[seq(1, w-1)]
  RcppExports.R <- file("R/RcppExports.R", "wb")
  writeLines(l, RcppExports.R)
  close(RcppExports.R)
}

compilerPath <- tools::Rcmd("config CC", stdout=TRUE)

# To distinguish between them, check the version output
versionInfo <- try(system(paste(compilerPath, "--version"), intern = TRUE))
if (inherits(versionInfo, "try-error")) {
  .o2 <- "-O2 "
} else if (any(grepl("clang", versionInfo, ignore.case = TRUE))) {
  .o2 <- "-O3 -fno-math-errno -mtune=native "
} else if (any(grepl("gcc", versionInfo, ignore.case = TRUE))) {
  .o2 <- "-O3 -fno-math-errno -mtune=native "
} else {
  .o2 <- "-O2 "
}

.in <- suppressWarnings(readLines("src/Makevars.in"))
.in <- gsub("@ARMA@", file.path(find.package("RcppArmadillo"),"include"), .in)
.in <- gsub("@O2@", .o2, .in)
.in <- gsub("@BH@", file.path(find.package("BH"),"include"), .in)
.in <- gsub("@RCPP@", file.path(find.package("Rcpp"),"include"), .in)
.in <- gsub("@EG@", file.path(find.package("RcppEigen"),"include"), .in)


.sl <- paste(capture.output(StanHeaders:::LdFlags()),
             capture.output(RcppParallel:::RcppParallelLibs()))
if (.Platform$OS.type == "windows") {
  # rpath is not meaningful on Windows and can generate noisy linker flags.
  # The path is shQuote()d by StanHeaders, so match quoted forms first;
  # otherwise a path containing a space leaves an orphaned token behind.
  .sl <- gsub("\\s+-Wl,-rpath,('[^']*'|\"[^\"]*\"|[^[:space:]]+)", "", .sl)
  # RcppParallel >= 6.0.0 statically links TBB on Windows via Rtools and no
  # longer loads its tbb.dll stub, so StanHeaders' LdFlags() output
  # (-L<RcppParallel/lib dir> -ltbb -ltbbmalloc) leaves the built DLL with an
  # unresolvable runtime dependency on tbb.dll.  Strip those flags; TBB
  # symbols resolve through RcppParallel's own -lRcppParallel instead.  The
  # strip is keyed to the stale -L<RcppParallel/lib> signature so that a
  # future StanHeaders that emits corrected flags (a different -L, -ltbb12,
  # or -lRcppParallel) passes through untouched.  When TBB_LINK_LIB/TBB_LIB
  # point at a user-supplied TBB, StanHeaders emits flags for that copy on
  # purpose, so keep them too.
  .rp_ver <- tryCatch(utils::packageVersion("RcppParallel"), error = function(e) package_version("0.0.0"))
  .tbb_env <- Sys.getenv("TBB_LINK_LIB", Sys.getenv("TBB_LIB"))
  if (.rp_ver >= "6.0.0" && !dir.exists(.tbb_env)) {
    # Match ".../RcppParallel/lib" plus any arch subdir (x64, arm64, ...) in
    # shQuote()d (single-quoted), double-quoted, or unquoted form -- but not
    # ".../RcppParallel/libs" (-lRcppParallel's dir, still needed).
    .sl2 <- gsub("-L'[^']*RcppParallel[/\\\\]lib([/\\\\][^']*)?'", "", .sl)
    .sl2 <- gsub("-L\"[^\"]*RcppParallel[/\\\\]lib([/\\\\][^\"]*)?\"", "", .sl2)
    .sl2 <- gsub("-L[^-'\"[:space:]][^[:space:]]*RcppParallel[/\\\\]lib([/\\\\][^[:space:]]*)?(?=[[:space:]]|$)",
                 "", .sl2, perl = TRUE)
    if (!identical(.sl2, .sl)) {
      # The stale -L was present, so the -ltbb/-ltbbmalloc flags next to it
      # came from the same LdFlags() call; drop them with it.
      .sl <- gsub("-ltbbmalloc_proxy\\b", "", .sl2)
      .sl <- gsub("-ltbbmalloc\\b", "", .sl)
      .sl <- gsub("-ltbb\\b", "", .sl)
      .sl <- gsub("\\s+", " ", trimws(.sl))
    }
  }
}
.in <- gsub("@SL@", .sl, .in) #nolint

## SUNDIALS public headers are vendored in-tree (src/sundials_inc) so the
## vendored SUNDIALS .c sources always compile against the matching headers
## (see https://github.com/nlmixr2/rxode2/issues/1155).
.sundialsInc <- file.path("src", "sundials_inc")
if (!file.exists(file.path(.sundialsInc, "sundials", "sundials_config.h"))) {
  stop("Vendored SUNDIALS headers are missing from src/sundials_inc.\n",
       "These files are committed to the repository and must be present.\n",
       "Re-vendor them with 'Rscript build/vendor-sundials.R'.", call. = FALSE)
}

## CVODE C source and private impl headers are committed to src/.
## All of these files must be present; they are part of the package source.
.sundialsVendorFiles <- c(
  "cvode_diag_impl.h", "cvode_impl.h", "cvode_ls_impl.h", "cvode_proj_impl.h",
  "sundials_adiak_metadata.h", "sundials_cli.h", "sundials_cvode.c",
  "sundials_cvode_diag.c", "sundials_cvode_io.c", "sundials_cvode_ls.c",
  "sundials_cvode_nls.c", "sundials_cvode_proj.c",
  "sundials_datanode.h", "sundials_hashmap_impl.h", "sundials_iterative_impl.h",
  "sundials_logger_impl.h", "sundials_macros.h", "sundials_nvector_serial.c",
  "sundials_profiler_impl.h",
  "sundials_sundials_band.c", "sundials_sundials_cli.c",
  "sundials_sundials_context.c", "sundials_sundials_dense.c",
  "sundials_sundials_direct.c", "sundials_sundials_errors.c",
  "sundials_sundials_hashmap.c",
  "sundials_sundials_iterative.c", "sundials_sundials_linearsolver.c",
  "sundials_sundials_logger.c", "sundials_sundials_math.c",
  "sundials_sundials_matrix.c", "sundials_sundials_memory.c",
  "sundials_sundials_nonlinearsolver.c", "sundials_sundials_nvector.c",
  "sundials_sundials_profiler.c",
  "sundials_sundials_version.c",
  "sundials_sunlinsol_band.c", "sundials_sunlinsol_dense.c",
  "sundials_sunmatrix_band.c", "sundials_sunmatrix_dense.c",
  "sundials_sunmatrix_sparse.c", "sundials_sunnonlinsol_fixedpoint.c",
  "sundials_sunnonlinsol_newton.c", "sundials_system_memory.c",
  "sundials_utils.h", "sundials_debug.h", "sunlinsol_spgmr.c", "sunlinsol_spbcgs.c",
  "sunlinsol_sptfqmr.c", file.path("stl", "sunstl_vector.h")
)

.missing <- .sundialsVendorFiles[!file.exists(file.path("src", .sundialsVendorFiles))]
if (length(.missing) > 0) {
  stop(
    "Required SUNDIALS source files are missing from src/:\n  ",
    paste(.missing, collapse = "\n  "),
    "\nThese files are committed to the repository and must be present.\n",
    "Please ensure you have a complete checkout of the package source.",
    call. = FALSE
  )
}

## ---------------------------------------------------------------------------
## CRAN fixes for vendored SUNDIALS files
## These patches are applied unconditionally so they survive both fresh vendor
## and subsequent installs from the committed source tree.
## ---------------------------------------------------------------------------

## Fix 1a: sundials_sundials_errors.c
##  - abort() is forbidden in CRAN packages (use return instead)
##  - fprintf(stderr, ...) references a forbidden symbol
.ef <- "src/sundials_sundials_errors.c"
if (file.exists(.ef)) {
  .el <- readLines(.ef)
  .el <- gsub("abort();", "return;", .el, fixed = TRUE)
  .el <- gsub('fprintf(stderr, "%s", log_msg);', '', .el, fixed = TRUE)
  .ef_out <- file(.ef, "wb")
  writeLines(.el, .ef_out, sep = "\n")
  close(.ef_out)
}

## Fix 1b: sundials_sundials_logger.c
##  - Assignments and comparisons referencing stdout/stderr forbidden symbols
.lf <- "src/sundials_sundials_logger.c"
if (file.exists(.lf)) {
  ## Collapse to a single string for multi-line pattern matching
  .ll <- paste(readLines(.lf), collapse = "\n")
  ## Remove the multi-line fprintf(stderr, ...) call in sunLoggerPrintf
  .ll <- gsub("(?s)fprintf\\(stderr,[^;]*;", "", .ll, perl = TRUE)
  ## Replace stdout/stderr symbol references with NULL
  .ll <- gsub("fp = stdout;", "fp = NULL;", .ll, fixed = TRUE)
  .ll <- gsub("fp = stderr;", "fp = NULL;", .ll, fixed = TRUE)
  .ll <- gsub("fp && fp != stdout && fp != stderr", "fp", .ll, fixed = TRUE)
  .ll <- gsub("logger->error_fp   = stderr;", "logger->error_fp   = NULL;", .ll, fixed = TRUE)
  .ll <- gsub("logger->warning_fp = stdout;", "logger->warning_fp = NULL;", .ll, fixed = TRUE)
  .ll <- strsplit(.ll, "\n", fixed = TRUE)[[1]]
  .lf_out <- file(.lf, "wb")
  writeLines(.ll, .lf_out, sep = "\n")
  close(.lf_out)
}

## Fix 1c: sundials_nvector_serial.c
##  - N_VPrint_Serial passes stdout to N_VPrintFile_Serial (forbidden symbol)
.nf <- "src/sundials_nvector_serial.c"
if (file.exists(.nf)) {
  .nl <- readLines(.nf)
  .nl <- gsub("N_VPrintFile_Serial(x, stdout);",
               "/* N_VPrintFile_Serial stdout removed for CRAN */", .nl, fixed = TRUE)
  .nf_out <- file(.nf, "wb")
  writeLines(.nl, .nf_out, sep = "\n")
  close(.nf_out)
}

## Fix 1d: sundials_sundials_nvector.c
##  - printf() calls compile to puts() at -O3, which is a forbidden symbol
.nvf <- "src/sundials_sundials_nvector.c"
if (file.exists(.nvf)) {
  .nv <- readLines(.nvf)
  .nv <- gsub('printf("NULL Vector\\n");', '', .nv, fixed = TRUE)
  .nv <- gsub('printf("NULL Print Op\\n");', '', .nv, fixed = TRUE)
  .nvf_out <- file(.nvf, "wb")
  writeLines(.nv, .nvf_out, sep = "\n")
  close(.nvf_out)
}

## Fix 3: Remove deprecated SUNDIALS 7.x workspace-query function usage.
##  Null out the deprecated function pointers (ops->space, ops->nvspace) and
##  replace direct N_VSpace/SUNMatSpace/SUNLinSolSpace calls with zero-assignment
##  equivalents.  Also strips any pragma block injected by a previous run.
.strip_deprecated_pragma <- function(.lines) {
  if (length(.lines) >= 3 &&
      trimws(.lines[1]) == "#if defined(__GNUC__) || defined(__clang__)" &&
      grepl("Wdeprecated-declarations", .lines[2], fixed = TRUE) &&
      trimws(.lines[3]) == "#endif") {
    .lines <- .lines[-c(1L, 2L, 3L)]
  }
  .lines
}

## nvector_serial: null out deprecated N_VSpace_Serial function pointer
.nf3 <- "src/sundials_nvector_serial.c"
if (file.exists(.nf3)) {
  .nl3 <- .strip_deprecated_pragma(readLines(.nf3))
  .nl3 <- gsub("= N_VSpace_Serial;", "= NULL;", .nl3, fixed = TRUE)
  .nf3_out <- file(.nf3, "wb")
  writeLines(.nl3, .nf3_out, sep = "\n")
  close(.nf3_out)
}

## sunlinsol_band: null out deprecated SUNLinSolSpace_Band function pointer
.lbf <- "src/sundials_sunlinsol_band.c"
if (file.exists(.lbf)) {
  .lb <- .strip_deprecated_pragma(readLines(.lbf))
  .lb <- gsub("= SUNLinSolSpace_Band;", "= NULL;", .lb, fixed = TRUE)
  .lbf_out <- file(.lbf, "wb")
  writeLines(.lb, .lbf_out, sep = "\n")
  close(.lbf_out)
}

## sunlinsol_dense: null out deprecated SUNLinSolSpace_Dense function pointer
.ldf <- "src/sundials_sunlinsol_dense.c"
if (file.exists(.ldf)) {
  .ld <- .strip_deprecated_pragma(readLines(.ldf))
  .ld <- gsub("= SUNLinSolSpace_Dense;", "= NULL;", .ld, fixed = TRUE)
  .ldf_out <- file(.ldf, "wb")
  writeLines(.ld, .ldf_out, sep = "\n")
  close(.ldf_out)
}

## sunmatrix_band: null out deprecated SUNMatSpace_Band function pointer
.mbf <- "src/sundials_sunmatrix_band.c"
if (file.exists(.mbf)) {
  .mb <- .strip_deprecated_pragma(readLines(.mbf))
  .mb <- gsub("= SUNMatSpace_Band;", "= NULL;", .mb, fixed = TRUE)
  .mbf_out <- file(.mbf, "wb")
  writeLines(.mb, .mbf_out, sep = "\n")
  close(.mbf_out)
}

## sunmatrix_dense: null out deprecated SUNMatSpace_Dense function pointer
.mdf <- "src/sundials_sunmatrix_dense.c"
if (file.exists(.mdf)) {
  .md <- .strip_deprecated_pragma(readLines(.mdf))
  .md <- gsub("= SUNMatSpace_Dense;", "= NULL;", .md, fixed = TRUE)
  .mdf_out <- file(.mdf, "wb")
  writeLines(.md, .mdf_out, sep = "\n")
  close(.mdf_out)
}

## sunmatrix_sparse: null out deprecated SUNMatSpace_Sparse function pointer
.msf <- "src/sundials_sunmatrix_sparse.c"
if (file.exists(.msf)) {
  .ms <- .strip_deprecated_pragma(readLines(.msf))
  .ms <- gsub("= SUNMatSpace_Sparse;", "= NULL;", .ms, fixed = TRUE)
  .msf_out <- file(.msf, "wb")
  writeLines(.ms, .msf_out, sep = "\n")
  close(.msf_out)
}

## sundials_cvode: replace deprecated N_VSpace call with zero assignments
.cvf3 <- "src/sundials_cvode.c"
if (file.exists(.cvf3)) {
  .cv3 <- .strip_deprecated_pragma(readLines(.cvf3))
  .cv3 <- gsub("N_VSpace(y0, &lrw1, &liw1);", "lrw1 = 0; liw1 = 0;", .cv3, fixed = TRUE)
  .cvf3_out <- file(.cvf3, "wb")
  writeLines(.cv3, .cvf3_out, sep = "\n")
  close(.cvf3_out)
}

## sundials_cvode_ls: replace deprecated N_VSpace/SUNMatSpace/SUNLinSolSpace calls
.clf <- "src/sundials_cvode_ls.c"
if (file.exists(.clf)) {
  .cl <- .strip_deprecated_pragma(readLines(.clf))
  .cl <- gsub("N_VSpace(cv_mem->cv_tempv, &lrw1, &liw1);",
              "lrw1 = 0; liw1 = 0;", .cl, fixed = TRUE)
  .cl <- gsub("retval = SUNMatSpace(cvls_mem->savedJ, &lrw, &liw);",
              "lrw = 0; liw = 0; retval = 0;", .cl, fixed = TRUE)
  .cl <- gsub("retval = SUNLinSolSpace(cvls_mem->LS, &lrw, &liw);",
              "lrw = 0; liw = 0; retval = 0;", .cl, fixed = TRUE)
  .clf_out <- file(.clf, "wb")
  writeLines(.cl, .clf_out, sep = "\n")
  close(.clf_out)
}

## ---------------------------------------------------------------------------

## Compilation runs with src/ as the working directory, so a relative
## include path is sufficient (and avoids embedding build-tree paths).
.in <- gsub("@SUNDIALS_INC@", "-Isundials_inc", .in)

.sp_files <- c("sunlinsol_spgmr.c", "sunlinsol_spbcgs.c", "sunlinsol_sptfqmr.c")

.fix_monitoring_endif <- function(.lines) {
  .changed <- FALSE
  .i <- 1L
  while (.i <= length(.lines)) {
    if (trimws(.lines[.i]) == "#ifdef SUNDIALS_BUILD_WITH_MONITORING") {
      .j <- .i + 1L
      .inserted <- FALSE
      while (.j <= length(.lines)) {
        .tj <- trimws(.lines[.j])
        if (.tj == "#endif" || .tj == "#else") {
          break
        }
        if (grepl("Check if Atimes function has been set", .lines[.j], fixed = TRUE) &&
            any(grepl("SUNLINSOL_", .lines[seq.int(.i + 1L, .j - 1L)], fixed = TRUE))) {
          .lines <- append(.lines, "#endif", after = .j - 1L)
          .changed <- TRUE
          .inserted <- TRUE
          break
        }
        if (grepl("if\\s*\\(.*<=\\s*delta\\)", .lines[.j]) &&
            any(grepl("SUNLS_MSG_RESIDUAL", .lines[seq.int(.i + 1L, .j - 1L)], fixed = TRUE))) {
          .lines <- append(.lines, "#endif", after = .j - 1L)
          .changed <- TRUE
          .inserted <- TRUE
          break
        }
        .j <- .j + 1L
      }
      if (.inserted) {
        .i <- .i + 1L
      }
    }
    .i <- .i + 1L
  }
  attr(.lines, "changed") <- .changed
  .lines
}

# CRAN requires that compiled code not reference stdout.  The upstream
# SUNDIALS sources initialise info_file to stdout; patch it to NULL.
# print_level defaults to 0 so info_file is never dereferenced in practice.
for (.sp in file.path("src", .sp_files)) {
  if (file.exists(.sp)) {
    .sl <- .strip_deprecated_pragma(readLines(.sp))
    .sl <- gsub("content->info_file\\s*=\\s*stdout;",
                 "content->info_file = NULL;", .sl)
    .sl <- gsub("= SUNLinSolSpace_SPBCGS;", "= NULL;", .sl, fixed = TRUE)
    .sl <- gsub("= SUNLinSolSpace_SPGMR;", "= NULL;", .sl, fixed = TRUE)
    .sl <- gsub("= SUNLinSolSpace_SPTFQMR;", "= NULL;", .sl, fixed = TRUE)
    .sl <- gsub("N_VSpace(SPBCGS_CONTENT(S)->vtemp, &lrw1, &liw1);",
                "lrw1 = 0; liw1 = 0;", .sl, fixed = TRUE)
    .sl <- gsub("N_VSpace(SPGMR_CONTENT(S)->vtemp, &lrw1, &liw1);",
                "lrw1 = 0; liw1 = 0;", .sl, fixed = TRUE)
    .sl <- gsub("N_VSpace(SPTFQMR_CONTENT(S)->vtemp1, &lrw1, &liw1);",
                "lrw1 = 0; liw1 = 0;", .sl, fixed = TRUE)
    .sl <- .fix_monitoring_endif(.sl)
    .sp_out <- file(.sp, "wb")
    writeLines(.sl, .sp_out, sep = "\n")
    close(.sp_out)
  }
}

# Generate src/implicit_euler_rxode2.hpp from BH's copy of the same header,
# replacing the std::exit(0) call (CRAN-forbidden) with a C++ exception so
# the error can be caught and handled via the rxode2 OpenMP-safe badSolveExit
# pattern.  Generating from BH keeps us in sync with any future BH updates.
.bh_ie <- system.file("include", "boost", "numeric", "odeint", "stepper",
                       "implicit_euler.hpp", package = "BH")
if (!nzchar(.bh_ie)) {
  stop("BH package implicit_euler.hpp not found", call. = FALSE)
}
.ie_lines <- readLines(.bh_ie)
.ie_lines <- sub(
  "^(#include <boost/numeric/ublas/lu\\.hpp>)$",
  "\\1\n#include <stdexcept>",
  .ie_lines
)
.ie_lines <- gsub(
  "if\\( res != 0 \\) std::exit\\(0\\);",
  "if( res != 0 ) throw std::runtime_error(\"implicit Euler LU factorization singular\");",
  .ie_lines
)
.ie_out <- file("src/implicit_euler_rxode2.hpp", "wb")
writeLines(.ie_lines, .ie_out, sep = "\n")
close(.ie_out)


.badStan <- ""
.in <- gsub("@SH@", gsub("-I", "-@ISYSTEM@",
                         paste(capture.output(StanHeaders:::CxxFlags()), # nolint
                               capture.output(RcppParallel:::CxxFlags()), # nolint
                               paste0("-@ISYSTEM@'", system.file('include', package = 'StanHeaders', mustWork = TRUE), "'"),
                               paste0("-@ISYSTEM@'", system.file('include', 'src', package = 'StanHeaders', mustWork = TRUE), "'"),
                               .badStan)),
            .in)




if (.Platform$OS.type == "windows") {
  .makevars <- file("src/Makevars.win", "wb")
  .i <- "I"
} else {
  .makevars <- file("src/Makevars", "wb")
  if (any(grepl("Pop!_OS", utils::osVersion, fixed=TRUE)) ||
          any(grepl("Ubuntu", utils::osVersion, fixed=TRUE))) {
    .i <- "isystem"
  } else {
    .i <- "I"
  }
}

writeLines(gsub("@ISYSTEM@", .i, .in),
             .makevars)
close(.makevars)

## --- Compilation prerequisites: generate before anything that may fail ---
## sbuf.c and codegen2.h are required for compilation.  They must be written
## before the digest::digest() call below, which needs the 'digest' package
## (Suggests, not Imports) and may be absent on --no-suggests CI runners.

unlink("src/sbuf.c")
l <- readLines("inst/include/sbuf.c")
sbuf.c <- file("src/sbuf.c", "wb")
writeLines(l, sbuf.c)
close(sbuf.c)

unlink("src/codegen2.h")
l <- readLines("inst/include/rxode2_model_shared.c")

l <- l[l != ""]
l <- gsub(" *= *NULL;", "=NULL;", l)

def <- l
w <- which(regexpr("double _prod", def) != -1) - 1
def <- def[1:w]
def <- gsub("=NULL", "", def)
def <- gsub("[^ ]* *[*]?([^;]*);", "\\1", def)

def <- unique(c(def, c("_sum", "_udf", "_sign", "_prod", "_max", "_min", "_transit4P", "_transit3P", "_rxDelay", "_rxDelayD", "_rxDelayD2", "_rxDelayD3", "_rxPast", "_assignFuns0", "_assignFuns", "_getRxSolve_", "_solveData", "_rxord", "__assignFuns2")))

w0 <- which(grepl("double +_prod", l))[1]
r <- 1:(w0 - 1)
l0 <- l[r]
l <- l[-r]


w1 <- which(regexpr("dynamic start", l) != -1)
l1 <- l[1:w1]



w2 <- which(regexpr("dynamic stop", l) != -1)
l2 <- l[seq(w2, length(l))]

w3 <- which(regexpr("assign start", l2) != -1)

l3 <- l2[seq(w3, length(l2))]
l2 <- l2[1:w3]

w4 <- which(regexpr("assign stop", l3) != -1)
l3 <- l3[seq(w4, length(l3))]

dfP <- l[seq(w1+1, w2-1)]

dfP <- dfP[regexpr("^ *$", dfP)==-1]
df <- setNames(do.call("rbind",lapply(seq_along(dfP),
                                      function(i) {
                                        .r <- sub("^ *", "", dfP[[i]])
                                        .r <- sub("^([^ ]*) *= *[(]", "\\1,", .r)
                                        .r <- sub("^([^ ]*) *[)] *R_GetCCallable *[(] *\"", "\\1,", .r, perl=TRUE)
                                        .r <- sub("^([^ ]*) *\" *, *\"", "\\1,", .r, perl=TRUE)
                                        .r <- sub("^([^ ]*)\" *[)] *; *", "\\1",.r, perl=TRUE)
                                        data.frame(t(strsplit(.r, ",")[[1]]),stringsAsFactors = FALSE)
                                      })), c("fun", "type", "package", "packageFun"))

df$rxFun <- df$fun
df$argMax <- df$argMin <- NA_integer_
df$threadSafe <- 1L
df <- df[,c("rxFun", "fun", "type", "package", "packageFun", "argMin", "argMax", "threadSafe")]
df$rxFun <- gsub("_llik", "llik", df$rxFun)

def <- def[!(def %in% df$rxFun)]
def <- def[!(def %in% df$fun)]

.parseEnv <- new.env(parent=emptyenv())
source("R/parseFuns.R")

df$argMin <- vapply(df$rxFun, function(f) {
  .n <- .parseEnv$.parseNum[f]
  if (is.na(.n)) return(NA_integer_)
  .n <-setNames(.n, NULL)
  as.integer(.n)
}, integer(1), USE.NAMES=TRUE)

df$argMax <- df$argMin

dfStr <- deparse(df)
dfStr[1] <- paste(".parseEnv$.rxode2parseDf <- ", dfStr[1])

dfIni.R <- file("R/dfIni.R", "wb")
writeLines(dfStr,
           dfIni.R)
close(dfIni.R)

## deparse1 came from R 4.0, use deparse2
deparse2 <- function(expr, collapse = " ", width.cutoff = 500L, ...) {
  paste(deparse(expr, width.cutoff, ...), collapse = collapse)
}

final <- c("#include <time.h>",
           "#include <stdlib.h>",
           "unsigned long int __timeId=0;",
           "char *genRandomChar(void);",
           "void writeHeader(const char *md5, const char *extra) {",
           paste0("sAppend(&sbOut, \"#define ", def, " _rx%s%s%ld_", def, "_%s\\n\", extra, md5, __timeId++, genRandomChar());"),
           "}",
           "void writeBody0(void) {",
           paste0("sAppendN(&sbOut, ", vapply(paste0(l0, "\n"), deparse2, character(1)), ", ", nchar(l0) + 1, ");"),
           "}",
           "void writeBody1(void) {",
           paste0("sAppendN(&sbOut, ", vapply(paste0(l1, "\n"), deparse2, character(1)), ", ", nchar(l1) + 1, ");"),
           "}",
           "void writeBody2(void) {",
           paste0("sAppendN(&sbOut, ", vapply(paste0(l2, "\n"), deparse2, character(1)), ", ", nchar(l2) + 1, ");"),
           "}",
           "void writeBody3(void) {",
           paste0("sAppendN(&sbOut, ", vapply(paste0(l3, "\n"), deparse2, character(1)), ", ", nchar(l3) + 1, ");"),
           "}",
           "void writeFooter(void) {",
           paste0("sAppendN(&sbOut, \"#undef ", def, "\\n\", ", nchar(def) + 8, ");"),
           "}"
           )

codegen2.h <- file("src/codegen2.h", "wb")
writeLines(final,
           codegen2.h)
close(codegen2.h)

## --- Optional: model-cache MD5 (needs 'digest', which is in Suggests) ---
## Skipped gracefully when digest is not installed (e.g. --no-suggests CI).

if (requireNamespace("digest", quietly = TRUE)) {
  cpp <- list.files("src", pattern = "\\.(c|h|cpp|f)$")
  cpp <- cpp[!dir.exists(file.path("src", cpp))]
  include <- list.files("inst/include", recursive = TRUE)

  md5 <- digest::digest(c(lapply(c(paste0("src/", cpp),
                                   paste0("inst/include/", include)
                                   ), digest::digest, file = TRUE),
                          ""))
  unlink("R/rxode2_md5.R")
  md5file <- file("R/rxode2_md5.R", "wb")
  writeLines(sprintf("rxode2.md5 <- \"%s\"\n", md5), md5file)
  close(md5file)

  l <- readLines("DESCRIPTION")
  w <- which(regexpr("Version[:] *(.*)$", l) != -1)
  v <- gsub("Version[:] *(.*)$", "\\1", l[w])

  unlink("inst/include/rxode2parseVer.h")
  ode.h <- file("inst/include/rxode2parseVer.h", "wb")
  writeLines(c(sprintf("#define __VER_md5__ \"%s\"", md5),
               "#define __VER_repo__ \"https://github.com/nlmixr2/rxode2\"",
               sprintf("#define __VER_ver__ \"%s\"", v)),
             ode.h)
  close(ode.h)
}

#' Hard fix population variables
#'
#' @param item expression to consider
#' @param var variable list to fix
#' @param isLhs is the expression a lhs expression?
#' @return expression with variables replaced with constants
#' @noRd
#' @author Matthew L. Fidler
.rxFixPopVar <- function(item, var, isLhs=FALSE) {
  if (is.atomic(item)) {
    return(item)
  }
  if (is.name(item)) {
    .n <- as.character(item)
    .n <- var[.n]
    if (!is.na(.n)) {
      return(setNames(.n, NULL))
    }
    return(item)
  } else if (is.call(item)) {
    if (isLhs && identical(item[[1]], quote(`/`))) {
      # handle d/dt() differently so that d doesn't get renamed
      .num <- item[[2]]
      .denom <- item[[3]]
      if (is.call(.num)) .num <- as.call(lapply(.num, .rxFixPopVar, var=var, isLhs=TRUE))
      if (is.call(.denom)) .denom <- as.call(lapply(.denom, .rxFixPopVar, var=var, isLhs=TRUE))
      return(as.call(c(list(item[[1]]), .num, .denom)))
    } else if (isLhs && length(item) == 2L &&
                 is.numeric(item[[2]])) {
      .env <- new.env(parent=emptyenv())
      .env$new <- NULL
      lapply(seq_along(var),
             function(i) {
               if (!is.null(.env$new)) return(NULL)
               .curVal <- setNames(var[i], NULL)
               .old <- str2lang(names(var[i]))
               if (identical(item[[1]], .old)) {
                 .env$new <- .curVal
               }
               return(NULL)
             })
      if (!is.null(.env$new)) {
        # handle x(0) = items
        return(as.call(c(.env$new, lapply(item[-1], .rxFixPopVar, var=var, isLhs=isLhs))))
      }
    }
    if (identical(item[[1]], quote(`=`)) ||
          identical(item[[1]], quote(`<-`)) ||
          identical(item[[1]], quote(`~`))) {
      .elhs <- lapply(item[c(-1, -3)], .rxFixPopVar, var=var, isLhs=TRUE)
      .erhs <- lapply(item[c(-1, -2)], .rxFixPopVar, var=var, isLhs=FALSE)
      return(as.call(c(item[[1]], .elhs, .erhs)))
    } else {
      return(as.call(c(list(item[[1]]), lapply(item[-1], .rxFixPopVar, var=var, isLhs=isLhs))))
    }
  } else {
    stop("unknown expression", call.=FALSE)
  }
}

#' Apply the fixed population estimated parameters
#'
#' @param ui rxode2 ui function
#' @param returnNull boolean for if unchanged values should return the
#'   original ui (`FALSE`) or null (`TRUE`)
#' @return when `returnNull` is TRUE, NULL if nothing was changed, or
#'   the changed model ui.  When `returnNull` is FALSE, return a ui no
#'   matter if it is changed or not.
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' One.comp.transit.allo <- function() {
#'  ini({
#'    # Where initial conditions/variables are specified
#'    lktr <- log(1.15)  #log k transit (/h)
#'    lcl  <- log(0.15)  #log Cl (L/hr)
#'    lv   <- log(7)     #log V (L)
#'    ALLC <- fix(0.75)  #allometric exponent cl
#'    ALLV <- fix(1.00)  #allometric exponent v
#'    prop.err <- 0.15   #proportional error (SD/mean)
#'    add.err <- 0.6     #additive error (mg/L)
#'    eta.ktr ~ 0.5
#'    eta.cl ~ 0.1
#'    eta.v ~ 0.1
#'  })
#'  model({
#'    #Allometric scaling on weight
#'    cl <- exp(lcl + eta.cl + ALLC * logWT70)
#'    v  <- exp(lv + eta.v + ALLV * logWT70)
#'    ktr <- exp(lktr + eta.ktr)
#'    # RxODE-style differential equations are supported
#'    d/dt(depot)   = -ktr * depot
#'    d/dt(central) =  ktr * trans - (cl/v) * central
#'    d/dt(trans)   =  ktr * depot - ktr * trans
#'    ## Concentration is calculated
#'    cp = central/v
#'    # And is assumed to follow proportional and additive error
#'    cp ~ prop(prop.err) + add(add.err)
#'  })
#' }
#'
#' m <- rxFixPop(One.comp.transit.allo)
#' m
#'
#' # now everything is already fixed, so calling again will do nothing
#'
#' rxFixPop(m)
#'
#' # if you call it with returnNull=TRUE when no changes have been
#' # performed, the function will return NULL
#'
#' rxFixPop(m, returnNull=TRUE)
#'
rxFixPop <- function(ui, returnNull=FALSE) {
  checkmate::assertLogical(returnNull, any.missing = FALSE, len=1, null.ok=FALSE)
  .model <- rxUiDecompress(assertRxUi(ui))
  .model <- .copyUi(.model)
  .iniDf <- .model$iniDf
  # a mixture proportion (ui$mixProbs) is structural: mix() requires it as a named
  # model-block variable, so it cannot be literally substituted with its (fixed)
  # value -- doing so makes the re-parse below throw from mix().  Exclude it.
  .mixProbs <- tryCatch(.model$mixProbs, error=function(e) NULL)
  .w <- which(!is.na(.iniDf$ntheta) & is.na(.iniDf$err) & .iniDf$fix &
                !(.iniDf$name %in% .mixProbs))
  if (length(.w) == 0L) {
    if (returnNull) return(NULL)
    return(.model)
  }
  .v <- setNames(.iniDf$est[.w], .iniDf$name[.w])
  .lst <- lapply(.model$lstExpr,
                 function(e) {
                   .rxFixPopVar(e, .v)
                 })
  .iniDf <- .iniDf[-.w, ]
  .iniDf$ntheta <- ifelse(is.na(.iniDf$ntheta), NA_integer_, seq_along(.iniDf$ntheta))
  assign("iniDf", .iniDf, envir=.model)
  suppressMessages({
    model(.model) <- .lst
    .model
  })
}
.lineHasFixedResEnv <- new.env(parent=emptyenv())
.lineHasFixedResEnv$err <- NULL
#' Does this line have a fixed residual expression?
#'
#' @param line parsed line to check
#' @param errs errors to check against
#' @return FALSE
#' @noRd
#' @author Matthew L. Fidler
.lineHasFixedRes <- function(line, errs) {
  if (is.call(line)) {
    return(any(sapply(line, .lineHasFixedRes, errs=errs)))
  } else if (is.name(line)) {
    .cline <- as.character(line)
    if (.cline %in% errs) {
      .lineHasFixedResEnv$err <- .cline
      return(TRUE)
    } else {
      return(FALSE)
    }
  } else if (is.atomic(line)) {
    return(FALSE)
  } else {
    stop("unknown expression", call.=FALSE)
  }
}

#' Literally fix residual parameters
#'
#' @inheritParams rxFixPop
#' @return model with residual parameters literally fixed in the model
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' One.comp.transit.allo <- function() {
#'  ini({
#'    # Where initial conditions/variables are specified
#'    lktr <- log(1.15)  #log k transit (/h)
#'    lcl  <- log(0.15)  #log Cl (L/hr)
#'    lv   <- log(7)     #log V (L)
#'    ALLC <- 0.75  #allometric exponent cl
#'    ALLV <- 1.00  #allometric exponent v
#'    prop.err <- fix(0.15)   #proportional error (SD/mean)
#'    add.err <- fix(0.6)     #additive error (mg/L)
#'    eta.ktr ~ 0.5
#'    eta.cl ~ 0.1
#'    eta.v ~ 0.1
#'  })
#'  model({
#'    #Allometric scaling on weight
#'    cl <- exp(lcl + eta.cl + ALLC * logWT70)
#'    v  <- exp(lv + eta.v + ALLV * logWT70)
#'    ktr <- exp(lktr + eta.ktr)
#'    # RxODE-style differential equations are supported
#'    d/dt(depot)   = -ktr * depot
#'    d/dt(central) =  ktr * trans - (cl/v) * central
#'    d/dt(trans)   =  ktr * depot - ktr * trans
#'    ## Concentration is calculated
#'    cp = central/v
#'    # And is assumed to follow proportional and additive error
#'    cp ~ prop(prop.err) + add(add.err)
#'  })
#' }
#'
#' m <- rxFixRes(One.comp.transit.allo)
rxFixRes <- function(ui, returnNull=FALSE) {
  checkmate::assertLogical(returnNull, any.missing = FALSE, len=1, null.ok=FALSE)
  .model <- rxUiDecompress(assertRxUi(ui))
  .model <- .copyUi(.model)
  .iniDf <- .model$iniDf
  .w <- which(!is.na(.iniDf$ntheta) & !is.na(.iniDf$err) & .iniDf$fix)
  if (length(.w) == 0L) {
    if (returnNull) return(NULL)
    return(.model)
  }
  .v <- setNames(.iniDf$est[.w], .iniDf$name[.w])

  .lstExpr0 <- .model$lstExpr
  .env <- new.env(parent=emptyenv())
  .env$i <- 1
  .env$fix <- .iniDf$name[.w]
  .lst <- lapply(seq_len(length(.lstExpr0)+length(.w)),
                 function(i) {
                   .item <- .lstExpr0[[.env$i]]
                   if (is.call(.item) &&
                         identical(.item[[1]], quote(`~`)) &&
                         .lineHasFixedRes(.item, .env$fix)) {
                     .cerr <- .lineHasFixedResEnv$err
                     .env$fix <- .env$fix[.env$fix != .cerr]
                     str2lang(paste0(.cerr, " <- ", .v[.cerr]))
                   } else {
                     .env$i <- .env$i + 1L
                     .item
                   }
                 })

  .iniDf <- .iniDf[-.w, ]
  .iniDf$ntheta <- ifelse(is.na(.iniDf$ntheta), NA_integer_, seq_along(.iniDf$ntheta))
  assign("iniDf", .iniDf, envir=.model)
  suppressMessages({
    model(.model) <- .lst
    .model
  })
}

#' @return Inherited from parent routine
#' @name reexports
#' @keywords internal
NULL

#' @importFrom ggplot2 scale_type
#' @export
ggplot2::scale_type

#' @importFrom ggplot2 ggplot
#' @export
ggplot2::ggplot

#' @importFrom ggplot2 aes
#' @export
ggplot2::aes

#' @importFrom ggplot2 geom_line
#' @export
ggplot2::geom_line

#' @importFrom ggplot2 facet_wrap
#' @export
ggplot2::facet_wrap


#' @importFrom ggplot2 %+replace%
`%+replace%`

#' @importFrom ggplot2 ylab
#' @export
ggplot2::ylab

#' @importFrom ggplot2 xlab
#' @export
ggplot2::xlab

#' @importFrom ggplot2 waiver
#' @export
ggplot2::waiver

#' Empty Guide
#'
#' This empty guide draws nothing; It is included in rxode2 for
#' compatibility with ggplot 3.2
#'
#' @inheritParams ggplot2::guide_none
#' @return nothing, simply included to be compatible with ggplot 3.2
#' @export
#' @keywords internal
guide_none <- function(title = waiver(), position = waiver()) {
  stop("needs \"ggplot2\" 3.3.0", call. = FALSE)
}

#' @importFrom lotri lotri
#' @export
lotri::lotri

#' @importFrom ggplot2  label_value
#' @export
ggplot2::label_value

#' @importFrom ggplot2 label_both
#' @export
ggplot2::label_both

#' @importFrom ggplot2 label_context
#' @export
ggplot2::label_context

#' @importFrom ggplot2 label_wrap_gen
#' @export
ggplot2::label_wrap_gen

#' @importFrom ggplot2 label_context
#' @export
ggplot2::label_context

#' @importFrom ggplot2 scale_x_discrete
#' @export
ggplot2::scale_x_discrete

#' @importFrom ggplot2 scale_y_discrete
#' @export
ggplot2::scale_y_discrete

#' @importFrom ggplot2 scale_x_continuous
#' @export
ggplot2::scale_x_continuous

#' @importFrom ggplot2 scale_y_continuous
#' @export
ggplot2::scale_y_continuous

#' @importFrom ggplot2 scale_x_date
#' @export
ggplot2::scale_x_date

#' @importFrom ggplot2 scale_y_date
#' @export
ggplot2::scale_y_date

#' @importFrom ggplot2 expand_limits
#' @export
ggplot2::expand_limits

#' Get the base simulation model for simulation
#'
#' @param obj Fit Object
#' @return Simulation object
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
getBaseSimModel <- function(obj) {
  UseMethod("getBaseSimModel")
}

#' @rdname getBaseSimModel
#' @export
getBaseSimModel.default <- function(obj) {
  .ui <- assertRxUi(obj)
  .ret <- rxCombineErrorLines(.ui)
  if (identical(.ret[[2]][[2]], str2lang("params()"))) {
    .ret[[2]] <- .ret[[2]][-2]
  }
  .ret
}

#' Get the symengine for loading into symengine with `rxS()`
#'
#' @param obj Object
#' @return Simulation model ready to load into symeng
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
getBaseSymengineModel <- function(obj) {
  UseMethod("getBaseSymengineModel")
}

#'@export
getBaseSymengineModel.default <- function(obj) {
  .ui <- assertRxUi(obj)
  .x <- rxCombineErrorLines(.ui, cmtLines=FALSE, dvidLine=FALSE)
  if (identical(.x[[2]][[2]][[1]], quote(`params`))) {
    .x[[2]] <- .x[[2]][-2]
  }
  .x
}

#' Get the base simulation model for simulation with inis in the
#' underlying rxode2 model
#'
#' @param obj Fit Object
#' @return Simulation object
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
getBaseIniSimModel <- function(obj) {
  UseMethod("getBaseIniSimModel")
}

#' @rdname getBaseSimModel
#' @export
getBaseIniSimModel.default <- function(obj) {
  .ui <- assertRxUi(obj)
  .ret <- rxCombineErrorLines(.ui)
  .params <- .ret[[2]][[2]]
  .iniDf <- obj$iniDf
  .iniDf <- .iniDf[(is.na(.iniDf$neta1) | .iniDf$neta1 == .iniDf$neta2),]
  .ini <- lapply(seq_along(.iniDf$name), function(i) {
    if (!is.na(.iniDf$neta1[i])) {
      return(str2lang(paste0(.iniDf$name[i], "<- 0")))
    } else {
      return(str2lang(paste0(.iniDf$name[i], "<- ", .iniDf$est[i])))
    }
  })
  .predDf <- .ui$predDf
  if (is.null(.predDf)) {
    .sigma <- NULL
  } else {
    .sigma <- lapply(seq_along(.predDf$var), function(i) {
      if (.predDf$distribution[i] %in% c("dnorm",  "norm")) {
        str2lang(paste0("rxerr.", .predDf$var[i], "<- 1"))
      } else {
        NULL
      }
    })
    .w <- which(vapply(seq_along(.sigma),
                       function(i) {
                         !is.null(.sigma[[i]])
                       }, logical(1), USE.NAMES = FALSE))
    if (length(.w) == 0) {
      .sigma <- NULL
    } else {
      .sigma <- lapply(.w, function(i) {.sigma[[i]]})
    }
  }
  .mod <- lapply(seq_along(.ret[[2]])[-(1:2)], function(i){.ret[[2]][[i]]})
  if (identical(.params, str2lang("params()"))) {
    .params <- NULL
  }
  # now filter out directive lines; simulation model generation adds
  # back the normalized declarations explicitly.
  .mod <- .rxFilterOutPropsAndAdjustPredDf(.ui, predDf=NULL, lstExpr=.mod)
  .interp <- rxUiGet.interpLines(list(.ui))
  .splitDose <- rxUiGet.splitDoseLines(list(.ui))
  as.call(c(list(quote(`rxode2`)),
            as.call(c(list(quote(`{`)), .params, .interp, .splitDose,
                      .sigma, .ini, .mod))))
}

#' Return if the object can be stacked
#'
#' @param object object to test if it can be stacked
#' @return boolean to tell if an object can be stacked using rxode2
#' @export
#' @author Matthew L. Fidler
#' @examples
#'
#' is.rxStackData(NULL)
is.rxStackData <- function(object) {
  if (!inherits(object, "data.frame")) return(FALSE)
  .mv <- try(.Call(`_rxode2_rxModelVarsStack`, object), silent=TRUE)
  if (!inherits(.mv, "rxModelVars")) return(FALSE)
  .mv <- try(object$rxModelVars, silent=TRUE)
  if (!inherits(.mv, "rxModelVars")) return(FALSE)
  TRUE
}

#' Stack a solved object for things like default ggplot2 plot
#'
#' @param data is a rxode2 object to be stacked.
#'
#' @param vars Variables to include in stacked data; By default this
#'   is all the variables when vars is NULL.
#'
#' When vars is `sim` and comes from a `rxode2` ui simulation with
#' multiple endpoints (ie it has a `CMT` in the simulation), it will
#' rework the data as if it was stacked based the value based on the
#' compartments in the multiple endpoint model.
#'
#' When the vars is `sim.endpoint1` it will subset the stack to
#' endpoint1, you can also have `c("sim.endpoint1", "sim.endpoint2")
#' and the "stack" will subset to endpoint1 and endpoint2.
#'
#' When you specify the `sim` type variables they have to be all
#' prefixed with `sim` otherwise, the stack will not treat them
#' differently.
#'
#' @param doSim boolean that determines if the "sim" variable in a
#'   `rxSolve` dataset is actually "stacking" based on the endpoint
#'   (`TRUE`) or simply treating `sim` as a variable.
#'
#' @param doIpredSim boolean that determines if the "ipredSim" variable in a
#'  `rxSolve` dataset is actually "stacking" based on the endpoint
#'  (`TRUE`) or simply treating `ipredSim` as a variable.
#'
#' @return Stacked data with \code{value} and \code{trt}, where value is the values
#'   and \code{trt} is the state and \code{lhs} variables.
#'
#' @author Matthew Fidler
#' @export
rxStack <- function(data, vars = NULL, doSim=TRUE, doIpredSim=TRUE) {
  if (!is.rxStackData(data)) stop("this data cannot be used with `rxStack`", call.=FALSE)
  .nd <- names(data)
  checkmate::assertCharacter(vars, pattern="^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$", null.ok=TRUE)
  if (doSim) {
    .doSim <- all(vapply(vars, function(x) {
      if (x == "sim") return(TRUE)
      if (substr(x, 0, 4) == "sim.") return(TRUE)
      FALSE
    }, logical(1), USE.NAMES=FALSE))
  } else {
    .doSim <- FALSE
  }
  if (doIpredSim) {
    .doIpredSim <- all(vapply(vars, function(x) {
      if (x == "ipredSim") return(TRUE)
      if (substr(x, 0, 9) == "ipredSim.") return(TRUE)
      FALSE
    }, logical(1), USE.NAMES=FALSE))
  } else {
    .doIpredSim <- FALSE
  }
  if (length(vars) == 1L && (.doSim || .doIpredSim) &&
        any(.nd == "CMT")) {
    .vars <- vapply(vars, function(x) {
      if (.doSim) {
        if (x == "sim") return(NA_character_)
        if (substr(x, 0, 4) == "sim.") return(substr(x, 5, nchar(x)))
      }
      if (.doIpredSim) {
        if (x == "ipredSim") return(NA_character_)
        if (substr(x, 0, 9) == "ipredSim.") return(substr(x, 10, nchar(x)))
      }
      x
    }, character(1), USE.NAMES=FALSE)
    .vars <- .vars[!is.na(.vars)]
    .mv <- data$rxModelVars
    if (!is.null(.mv)) {
      .ret <- NULL
      if (any(.nd == "sim.id")) {
        .ret <- list(sim.id=data[["sim.id"]])
      }
      if (any(.nd == "id")) {
        .ret <- c(.ret, list(id=data[["id"]]))
      }
      if (any(.nd == "resetno")) {
        .ret <- c(.ret, list(resetno=data[["resetno"]]))
      }
      if (any(.nd == "evid")) {
        .ret <- c(.ret, list(evid=data[["evid"]]))
      }
      if (any(.nd == "amt")) {
        .ret <- c(.ret, list(amt=data[["amt"]]))
      }
      .cmt <- as.integer(data[["CMT"]])
      attr(.cmt, "levels") <- c(.mv$state, .mv$stateExtra)
      attr(.cmt, "class") <- "factor"
      if (.doSim) {
        .ret <- c(.ret, list(time=data[["time"]], value=data[["sim"]], trt=.cmt))
      }
      if (.doIpredSim) {
        .ret <- c(.ret,
                  list(time=data[["time"]], value=data[["ipredSim"]], trt=.cmt))
      }
      attr(.ret, "row.names") <- c(NA_integer_, length(.cmt))
      attr(.ret, "class") <- "data.frame"
      if (length(.vars) > 0L) {
        .ret <- .ret[.ret$trt %in% .vars,]
      }
      return(.ret)
    }
  }
  rxStack_(data, vars)
}

#' Collect warnings and just warn once.
#'
#' @param expr R expression
#' @param lst When `TRUE` return a list with
#'     list(object,warnings) instead of issuing the warnings.
#'     Otherwise, when `FALSE` issue the warnings and return the
#'     object.
#' @return The value of the expression or a list with the value of
#'     the expression and a list of warning messages
#' @author Matthew L. Fidler
#' @export
.collectWarnings <- function(expr, lst = FALSE) {
  .ws <- NULL
  .thisEnv <- environment()
  .ret <- suppressWarnings(
    withCallingHandlers(expr,
      warning = function(w) {
        assign(".ws", unique(c(w$message, .ws)), .thisEnv)
      }
    )
  )
  if (lst) {
    return(list(.ret, .ws))
  } else {
    for (.w in .ws) {
      warning(.w, call. = FALSE)
    }
    return(.ret)
  }
}

#' GELU activation function
#' @param x numeric vector
#' @return numeric vector
#' @family Activation Functions
#' @export
#' @examples
#'
#' GELU(c(-2, -1, 0, 1, 2))
#'
#' \donttest{
#' # you can use rxode2 as well
#' r <- rxode2({
#'   r = GELU(time)
#' })
#' et <- et(c(-2, -1, 0, 1, 2))
#' rxSolve(r, et)
#' }
#'
GELU <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 1L)
}


#' Derivatives of GELU
#'
#' @param x numeric vector
#' @return numeric vector
#' @family Activation Functions
#' @export
#' @examples
#' dGELU(c(-2, -1, 0, 1, 2))
#' d2GELU(c(-2, -1, 0, 1, 2))
#' d3GELU(c(-2, -1, 0, 1, 2))
#' d4GELU(c(-2, -1, 0, 1, 2))
#' \donttest{
#' # you can use rxode2 as well
#' r <- rxode2({
#'    r1 <- dGELU(time)
#'    r2 <- d2GELU(time)
#'    r3 <- d3GELU(time)
#'    r4 <- d4GELU(time)
#' })
#' et <- et(c(-2, -1, 0, 1, 2))
#' rxSolve(r, et)
#' }
dGELU <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 9L)
}

#' @rdname dGELU
#' @export
d2GELU <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 10L)
}

#' @rdname dGELU
#' @export
d3GELU <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 11L)
}

#' @rdname dGELU
#' @export
d4GELU <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 12L)
}

#' Leaky ReLU activation function
#'
#' @param x numeric vector
#' @return numeric vector
#' @family Activation Functions
#' @export
#' @examples
#'
#' lReLU(c(-1, 0, 1))
#'
#' \donttest{
#' # Can use in rxode2 as well
#'
#' r <- rxode2({r <- lReLU(time)})
#' e <- et(c(-1, 0, 1))
#' rxSolve(r, e)
#' }
lReLU <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 5L)
}

#' Derivative of Leaky ReLU activation function
#'
#' @param x numeric vector
#' @return numeric vector
#' @family Activation Functions
#' @export
#' @examples
#'
#' dlReLU(c(-1, 0, 1))
#'
#' \donttest{
#' # Can use in rxode2 as well
#'
#' r <- rxode2({r <- dlReLU(time)})
#' e <- et(c(-1, 0, 1))
#' rxSolve(r, e)
#' }
#'
dlReLU <- function(x) {
  checkmate::assertNumeric(x, finite=TRUE, any.missing=FALSE)
  .Call(`_rxode2_activationF`, x, 8L)
}

# rxEt public helpers -----------------------------------------------------

#' @importFrom utils .DollarNames
#' @export
.DollarNames.rxEt <- function(x, pattern) { # nolint
  if (is.rxEt(x)) { # nolint
    .envProps <- c("randomType", "canResize", "ids", "show", "ndose", "nobs")
    .methods <- c(
      "expand", "getSampling", "get.sampling", "getDosing", "get.dosing",
      "get.nobs", "get.obs.rec", "getEventTable", "get.EventTable",
      "copy", "importEventTable", "import.EventTable", "simulate",
      "clearDosing", "clear_dosing", "clear.dosing",
      "clearSampling", "clear_sampling", "clear.sampling",
      "addSampling", "add_sampling", "add.sampling",
      "addDosing", "add_dosing", "add.dosing",
      "get_units", "getUnits", "get.units", "units"
    )
    .dataCols <- rev(c("id", "low", "time", "high", "cmt", "amt", "rate", "ii", "addl", "evid", "ss", "dur"))
    return(grep(pattern, c(.envProps, .methods, .dataCols, "env"), value = TRUE))
  }
  character(0)
}

.isRxEt <- function(obj) {
  inherits(obj, "rxEt")
}


.etAddCls <- function(x) {
  if (.isRxEt(x)) {
    .x <- x
    .cls <- class(x)
    class(.x) <- "data.frame"
    if (!is.null(.x[["evid"]])) {
      class(.x[["evid"]]) <- "rxEvid"
      .tmp <- .x[["rate"]]
      .cls2 <- class(.tmp)
      if (!inherits(.cls2, "rxRateDur")) {
        class(.tmp) <- c("rxRateDur", .cls2)
      }
      .x[["rate"]] <- .tmp
      .tmp <- .x[["dur"]]
      .cls2 <- class(.tmp)
      if (!inherits(.cls2, "rxRateDur")) {
        class(.tmp) <- c("rxRateDur", .cls2)
      }
      .x[["dur"]] <- .tmp
      class(.x) <- .cls
      .x
    } else {
      x
    }
  } else {
    x
  }
}

# et generic and pipeline state ------------------------------------------

#' Event Table Function
#'
#' @param ... Times or event tables.  They can also be one of the named arguments below.
#'
#' @param time Time is the time of the dose or the sampling times.
#'     This can also be unspecified and is determined by the object
#'     type (list or numeric/integer).
#'
#' @param amt Amount of the dose. If specified, this assumes a dosing
#'     record, instead of a sampling record.
#'
#' @param evid Event ID; This can be:
#'
#' | Numeric Value | Description |
#' |---------------|-------------|
#' | 0             | An observation. This can also be specified as `evid=obs` |
#' | 1             | A dose observation.  This can also be specified as `evid=dose` |
#' | 2             | A non-dose event. This can also be specified as `evid=other` |
#' | 3             | A reset event.  This can also be specified as `evid=reset`. |
#' | 4             |Dose and reset event.  This can also be specified as `evid=doseReset` or `evid=resetDose` |
#'
#' Note a reset event resets all the compartment values to zero and turns off all infusions.
#'
#' @param cmt Compartment name or number.  If a number, this is an
#'   integer starting at 1.  Negative compartments turn off a
#'   compartment. If the compartment is a name, the compartment name
#'   is changed to the correct state/compartment number before
#'   running the simulation.  For a compartment named "-cmt" the
#'   compartment is turned off.
#'
#'     Can also specify `cmt` as `dosing.to`,
#'     `dose.to`, `doseTo`, `dosingTo`, and
#'     `state`.
#'
#' @param ii When specifying a dose, this is the inter-dose interval
#'     for `ss`, `addl` and `until` options (described below).
#'
#' @param addl The number of additional doses at a inter-dose
#'     interval after one dose.
#'
#' @param ss Steady state flag;  It can be one of:
#'
#' | Value | Description |
#' |------------|-------------|
#' | 0 | This dose is not a steady state dose
#' | 1 | This dose is a steady state dose with the between/inter-dose interval of `ii` |
#' | 2 | Superposition steady state |
#'
#' When `ss=2` the steady state dose that uses the super-position
#' principle to allow more complex steady states, like 10 mg in the
#' morning and 20 mg at night, or dosing at 8 am 12 pm and 8 pm
#' instead of every 12 hours.  Since it uses the super positioning
#' principle, it only makes sense when you know the kinetics are
#' linear.
#'
#' All other values of `SS` are currently invalid.
#'
#' @param rate When positive, this is the rate of infusion.  Otherwise:
#'
#' | Value | Description |
#' |-------|--------------------------------|
#' | 0     |  No infusion is on this record |
#' | -1    | Modeled rate (in rxode2:`rate(cmt) =`); Can be `et(rate=model)`. |
#' |-2     | Modeled duration (in rxode2: `dur(cmt) =`); Can be`et(dur=model)` or `et(rate=dur)`. |
#'
#' When a modeled bioavailability is applied to positive rates
#' (`rate` > 0), the duration of infusion is changed. This is
#' because the data specify the rate and amount, the only think that
#' modeled bioavailability can affect is duration.
#'
#' If instead you want the modeled bioavailability to increase the
#' rate of infusion instead of the duration of infusion, specify the
#' `dur` instead or model the duration with `rate=2`.
#'
#' @param dur Duration of infusion.  When `amt` and `dur`
#'     are specified the rate is calculated from the two data items.
#'     When `dur` is specified instead of `rate`, the
#'     bioavailability changes will increase rate instead of
#'     duration.
#'
#' @param until This is the time until the dosing should end.  It can
#'     be an easier way to figure out how many additional doses are
#'     needed over your sampling period.
#'
#' @param id A integer vector of ids to add or remove from the event
#'     table.  If the event table is identical for each ID, then you
#'     may expand it to include all the ids in this vector.  All the
#'     negative ids in this vector will be removed.
#'
#' @param amountUnits The units for the dosing records (`amt`)
#'
#' @param timeUnits The units for the time records (`time`)
#'
#' @param addSampling This is a boolean indicating if a sampling time
#'     should be added at the same time as a dosing time.  By default
#'     this is `FALSE`.
#'
#' @param x This is the first argument supplied to the event table.
#'     This is named to allow `et` to be used in a pipe-line
#'     with arbitrary objects.
#'
#' @inheritParams base::eval
#' @inheritParams base::seq
#' @return A new event table
#'
#' @template etExamples
#' @importFrom Rcpp evalCpp
#' @importFrom stats simulate end setNames start
#' @importFrom methods is
#' @export
et <- function(x, ..., envir = parent.frame()) {
  UseMethod("et")
}

.etInfo <- new.env(parent = emptyenv())
.etInfo$pipelineRx <- NULL
.etInfo$pipelineInits <- NULL
.etInfo$pipelineEvents <- NULL
.etInfo$pipelineParams <- NULL
.etInfo$pipelineICov <- NULL
.etInfo$pipelineKeep <- NULL
.etInfo$pipelineThetaMat <- NULL
.etInfo$pipelineOmega <- NULL
.etInfo$pipelineIov <- NULL
.etInfo$pipelineSigma <- NULL
.etInfo$pipelineDfObs <- NULL
.etInfo$pipelineDfSub <- NULL
.etInfo$pipelineNSub <- NULL
.etInfo$lastIdLvl <- NULL

.isNa1 <- function(x) {
  if (inherits(x, "logical") ||
        inherits(x, "numeric") ||
        inherits(x, "integer")) {
    if (length(x) == 1) {
      return(is.na(x))
    }
  }
  FALSE
}

.etInfo$pipelineNStud <- NULL

#' Assign in the rxode2 pipeline
#'
#'
#' @param obj  Object to assign.  If NA return the value.
#' @return The pipeline object (invisibly)
#' @author Matthew L. Fidler
#' @export
#' @keywords internal
.pipeRx <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineRx))
  .etInfo$pipelineRx <- obj
  invisible(obj)
}

#' @rdname dot-pipeRx
#' @export
.pipeInits <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineInits))
  .etInfo$pipelineInits <- obj
  invisible(obj)
}

#' @rdname dot-pipeRx
#' @export
.pipeEvents <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineEvents))
  .etInfo$pipelineEvents <- obj
  invisible(obj)
}

#' @rdname dot-pipeRx
#' @export
.pipeParams <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineParams))
  .etInfo$pipelineParams <- obj
  invisible(obj)
}

#' @rdname dot-pipeRx
#' @export
.pipeKeep <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineKeep))
  .etInfo$pipelineKeep <- obj
  invisible(obj)
}

#' @rdname dot-pipeRx
#' @export
.pipeThetaMat <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineThetaMat))
  .etInfo$pipelineThetaMat <- obj
  invisible(obj)
}

#' @rdname dot-pipeRx
#' @export
.pipeOmega <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineOmega))
  .etInfo$pipelineOmega <- obj
  invisible(obj)
}

#' @rdname dot-pipeRx
#' @export
.pipeSigma <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineSigma))
  .etInfo$pipelineSigma <- obj
  invisible(obj)
}

#' @rdname dot-pipeRx
#' @export
.pipeDfObs <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineDfObs))
  .etInfo$pipelineDfObs <- obj
  invisible(obj)
}

#' @rdname dot-pipeRx
#' @export
.pipeDfSub <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineDfSub))
  .etInfo$pipelineDfSub <- obj
  invisible(obj)
}

#' @rdname dot-pipeRx
#' @export
.pipeNSub <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineNSub))
  .etInfo$pipelineNSub <- obj
  invisible(obj)
}


#' @rdname dot-pipeRx
#' @export
.pipeNStud <- function(obj) {
  if (.isNa1(obj)) return(invisible(.etInfo$pipelineNStud))
  .etInfo$pipelineNStud <- obj
  invisible(obj)
}

#' Clear/Set pipeline
#'
#' @param rx rxode2 object
#' @keywords internal
#' @return None, clears rxode2 pipeline
#' @export
.clearPipe <- function(rx = NULL, inits = NULL,
                       events = NULL, params = NULL,
                       iCov = NULL, keep = NULL,
                       thetaMat = NULL, omega = NULL,
                       sigma = NULL, dfObs = NULL,
                       dfSub = NULL, nSub = NULL,
                       nStud = NULL) {
  .etInfo$pipelineRx <- rx
  .etInfo$pipelineInits <- inits
  .etInfo$pipelineEvents <- events
  .etInfo$pipelineParams <- params
  .etInfo$pipelineICov <- iCov
  .etInfo$pipelineKeep <- keep
  .etInfo$pipelineThetaMat <- thetaMat
  .etInfo$pipelineOmega <- omega
  .etInfo$pipelineSigma <- sigma
  .etInfo$pipelineDfObs <- dfObs
  .etInfo$pipelineDfSub <- dfSub
  .etInfo$pipelineNSub <- nSub
  .etInfo$pipelineNStud <- nStud
}

# et dispatch methods -----------------------------------------------------

#' @rdname et
#' @export
et.rxode2 <- function(x, ..., envir = parent.frame()) {
  .clearPipe()
  .etInfo$pipelineRx <- x
  do.call(et, c(list(...), list(envir = envir)), envir = envir)
}

#' @rdname et
#' @export
et.function <- et.rxode2

#' @rdname et
#' @export
et.rxUi <- et.rxode2

#' @rdname et
#' @export
et.rxSolve <- function(x, ..., envir = parent.frame()) {
  ## Need to extract:
  ## 1. rxode2 model
  .etInfo$pipelineRx <- x$.args.object
  ## 2. rxode2 parameters
  .etInfo$pipelineParams <- x$.args.par0
  .etInfo$pipelineICov <- x$.args$iCov
  .etInfo$pipelineKeep <- x$.args$keep
  ## 3. rxode2 inits
  .etInfo$pipelineInits <- x$.args.inits
  ## 4. rxode2 thetaMat
  .etInfo$pipelineThetaMat <- x$.args$thetaMat
  ## 5. rxode2 omega
  .etInfo$pipelineOmega <- x$.args$omega
  ## 6. rxode2 sigma
  .etInfo$pipelineSigma <- x$.args$sigma
  ## 7. rxode2 dfObs
  .etInfo$pipelineDfObs <- x$env$.args$dfObs
  ## 8. rxode2 dfSub
  .etInfo$pipelineDfSub <- x$env$.args$dfSub
  do.call(et, c(list(...), list(envir = envir)), envir = envir)
}

#' @rdname et
#' @export
et.rxParams <- function(x, ..., envir = parent.frame()) {
  ## Need to extract:
  ## 1. rxode2 model
  ## 2. rxode2 parameters
  if (!is.null(x$params)) .etInfo$pipelineParams <- x$params
  if (!is.null(x$iCov)) .etInfo$pipelineICov <- x$iCov
  if (!is.null(x$keep)) .etInfo$pipelineKeep <- x$keep
  ## 3. rxode2 inits
  if (!is.null(x$inits)) .etInfo$pipelineInits <- x$inits
  ## 4. rxode2 thetaMat
  if (!is.null(x$thetaMat)) .etInfo$pipelineThetaMat <- x$thetaMat
  ## 5. rxode2 omega
  if (!is.null(x$omega)) .etInfo$pipelineOmega <- x$omega
  ## 6. rxode2 sigma
  if (!is.null(x$sigma)) .etInfo$pipelineSigma <- x$sigma
  ## 7. rxode2 dfObs
  if (!is.null(x$dfObs)) .etInfo$pipelineDfObs <- x$dfObs
  ## 8. rxode2 dfSub
  if (!is.null(x$dfSub)) .etInfo$pipelineDfSub <- x$dfSub
  if (!is.null(x$nSub)) .etInfo$pipelineNSub <- x$nSub
  if (!is.null(x$nStud)) .etInfo$pipelineNStud <- x$nStud

  do.call(et, c(list(...), list(envir = envir)), envir = envir)
}

#' @rdname et
#' @export
et.default <- function(x, ..., time = NULL, amt = NULL, evid = NULL, cmt = NULL,
                       ii = NULL, addl = NULL, ss = NULL, rate = NULL, dur = NULL,
                       until = NULL, id = NULL,
                       amountUnits = NULL, timeUnits = NULL, addSampling = NULL,
                       envir = parent.frame(),
                       by = NULL, length.out = NULL) {

  # ---- Capture missing status ----
  .xMissing <- missing(x)
  .timeMissing <- missing(time)
  .amtMissing <- missing(amt)
  .evidMissing <- missing(evid)
  .cmtMissing <- missing(cmt)
  .iiMissing <- missing(ii)
  .addlMissing <- missing(addl)
  .ssMissing <- missing(ss)
  .rateMissing <- missing(rate)
  .durMissing <- missing(dur)
  .untilMissing <- missing(until)
  .idMissing <- missing(id)
  .amountUnitsMissing <- missing(amountUnits)
  .timeUnitsMissing <- missing(timeUnits)
  .addSamplingMissing <- missing(addSampling)

  # ---- Positional argument checks ----
  if (!.xMissing) {
    if (!is.data.frame(x) &&
          !is.rxEt(x) && # nolint
          !is.character(x)) {
      # et(1, 2, 3) -> x=1, ...=list(2, 3) in et.default
      .dots <- list(...)
      if (length(.dots) >= 2) {
        .dn <- names(.dots)
        if (is.null(.dn) || any(.dn[1:2] == "")) stop("unused positional arguments", call. = FALSE)
      }
    }
  }

  # ---- Determine base rxEt object ----
  .xIsRxEt <- !.xMissing && is.rxEt(x) # nolint
  if (.xIsRxEt) {
    .et <- x$copy()  # always copy: et() returns a new object, never mutates x
  } else {
    .et <- .newRxEt() # nolint
  }
  .envRef <- .rxEtEnv(.et) # nolint

  # ---- Conflicting alias checks ----
  .dotArgs <- list(...)
  id <- .etAssertArgsAndReturnId(.dotArgs, id, ii, amt, cmt, timeUnits, time, dur, # nolint
                                 .idMissing, .iiMissing, .amtMissing, .cmtMissing,
                                 .timeUnitsMissing, .timeMissing, .durMissing)

  # ---- seq helpers: by / length.out ----
  .ret <- .etHandleSeq(by, length.out, .xIsRxEt, .envRef, x, ..., envir = envir, # nolint
                       time = time, et = .et,  xMissing = .xMissing, timeMissing = .timeMissing)
  if (.ret$done) {
    return(.rxEtSyncData(.ret$et)) # nolint
  }

  # ---- Positional args / data.frame import ----
  .ret <- .etHandlePositional(x, ..., # nolint
                              time = time, xIsRxEt = .xIsRxEt, envir = envir, envRef = .envRef, et = .et,
                              xMissing = .xMissing, timeMissing = .timeMissing)
  if (.ret$done) {
    return(.rxEtSyncData(.ret$et)) # nolint
  }
  .posCmt  <- .ret$posCmt
  .listObs <- .ret$listObs
  time     <- .ret$time

  # ---- Units ----
  .etHandleUnits(.envRef, amountUnits, timeUnits, .dotArgs, # nolint
                 .amountUnitsMissing, .timeUnitsMissing)

  # ---- ID expansion ----
  .addingEvents <- (!.amtMissing || !.timeMissing || !.evidMissing || !.cmtMissing ||
                    !.iiMissing || !.addlMissing || !.ssMissing || !.rateMissing ||
                    !.durMissing || !.untilMissing || !.addSamplingMissing ||
                    !is.null(.dotArgs[["dose"]]) || !is.null(.listObs))
  .ret <- .etHandleId(id, .envRef, .xIsRxEt, envir, addingEvents = .addingEvents) # nolint
  .resolvedId  <- .ret$resolvedId
  .targetIds   <- .ret$targetIds
  .doResize    <- .ret$doResize
  .addedIds    <- .ret$addedIds
  .removedIds  <- .ret$removedIds
  .existingIds <- .ret$existingIds

  # ---- Resolve evid / cmt ----
  .evidExpr <- substitute(evid)
  .evidSym <- as.character(.evidExpr)
  .evidVal <- .etHandleEvid(.evidExpr, .evidSym, envir, .evidMissing) # nolint

  .cmtExpr <- substitute(cmt)
  .cmtSym <- as.character(.cmtExpr)
  .cmtVal  <- .etHandleCmt(.cmtExpr, .cmtSym, .posCmt, envir, .cmtMissing) # nolint

  # ---- Deferred list obs ----
  if (!is.null(.listObs)) {
    .dfObs <- .etObsChunk(.listObs, cmt = .cmtVal) # nolint
    .etAddChunk(.envRef, .dfObs, .targetIds) # nolint
    .envRef$nobs <- .envRef$nobs + length(.listObs) * max(1L, length(.targetIds))
    if (.amtMissing && is.null(amt)) {
      return(.rxEtSyncData(.et)) # nolint
    }
  }

  # ---- Dose record ----
  .amtExpr <- substitute(amt)
  .timeExpr <- substitute(time)
  .iiExpr <- substitute(ii)
  .addlExpr <- substitute(addl)
  .ssExpr <- substitute(ss)
  .rateExpr <- substitute(rate)
  .durExpr <- substitute(dur)
  .untilExpr <- substitute(until)
  .rateSym <- deparse(.rateExpr)

  .ret <- .etHandleDose(amt, .amtExpr, .dotArgs, .envRef, envir, time, .timeExpr, # nolint
                        .iiExpr, .addlExpr, .ssExpr, .rateExpr, .durExpr, .untilExpr,
                        .evidVal, .cmtVal, .resolvedId, .targetIds, addSampling, .et, .rateSym,
                        .amtMissing, .timeMissing, .iiMissing, .addlMissing,
                        .ssMissing, .rateMissing, .durMissing, .untilMissing,
                        .addSamplingMissing)
  if (.ret$done) {
    return(.rxEtSyncData(.ret$et)) # nolint
  }

  # ---- Infusion/SS dose without explicit amt ----
  .ret <- .etHandleInfusionNoAmt(amt, .amtExpr, .rateExpr, .ssExpr, .envRef, envir, time, .timeExpr, # nolint
                                 .iiExpr, .durExpr, .evidVal, .cmtVal, .targetIds, .et, .rateSym,
                                 .amtMissing, .rateMissing, .ssMissing, .timeMissing, .iiMissing, .durMissing)
  if (.ret$done) {
    return(.rxEtSyncData(.ret$et)) # nolint
  }

  # ---- Observation record ----
  .ret <- .etHandleObs(time, .timeExpr, envir, .envRef, .evidVal, # nolint
                       .cmtVal, .targetIds, .et, .timeMissing)
  if (.ret$done) {
    return(.rxEtSyncData(.ret$et)) # nolint
  }

  # ---- Piping pattern ----
  .ret <- .etHandlePiping(.xIsRxEt, x, time, .timeExpr, amt, .amtExpr, .dotArgs, .cmtVal, # nolint
                          .targetIds, .evidVal, .envRef, .et, .timeMissing, .amtMissing, envir)
  if (.ret$done) {
    return(.rxEtSyncData(.ret$et)) # nolint
  }

  # ---- ID-only resize ----
  if (.doResize) {
    .groups <- .etGetGroups(.envRef) # nolint
    if (length(.groups) > 0L) {
      .templateId <- if (length(.existingIds) > 0L) .existingIds[[1L]] else NA_integer_
      .templateGroup <- which(vapply(.groups, function(.g) .templateId %in% .g$ids, logical(1)))[1]
      if (is.na(.templateGroup)) {
        .templateGroup <- if (length(.groups) > 0L) 1L else NA_integer_
      }
      .templateData <- if (!is.na(.templateGroup)) .groups[[.templateGroup]]$data else NULL
      if (length(.groups) == 1L && .etGroupIdsEqual(.existingIds, .groups[[1]]$ids)) {
        .groups[[1]]$ids <- sort(unique(as.integer(.envRef$ids)))
      } else {
        for (.i in seq_along(.groups)) {
          .groups[[.i]]$ids <- intersect(.groups[[.i]]$ids, .envRef$ids)
        }
        .groups <- Filter(function(.g) length(.g$ids) > 0L, .groups)
        if (length(.addedIds) > 0L && length(.groups) > 0L) {
          .templateGroup <- which(vapply(.groups, function(.g) .templateId %in% .g$ids, logical(1)))[1]
          if (is.na(.templateGroup)) {
            .templateGroup <- 1L
          }
          .groups[[.templateGroup]]$ids <- sort(unique(c(.groups[[.templateGroup]]$ids, as.integer(.addedIds))))
        } else if (length(.addedIds) > 0L &&
                   length(.groups) == 0L &&
                   is.data.frame(.templateData) &&
                   nrow(.templateData) > 0L) {
          .groups <- list(list(
            ids = sort(unique(as.integer(.addedIds))),
            data = .templateData
          ))
        }
      }
      .etSetGroups(.envRef, .groups) # nolint
      .etResetCountsFromGroups(.envRef) # nolint
    } else {
      if (length(.addedIds) > 0L) {
        if (length(.existingIds) > 0) {
          .tid <- .existingIds[1]
        } else {
          .tid <- NA_integer_
        }
        if (!is.na(.tid) && .tid <= length(.envRef$chunks) && !is.null(.envRef$chunks[[.tid]])) {
          .template <- .envRef$chunks[[.tid]]
          for (.newId in .addedIds) {
            .newChunk <- .template
            .newChunk$id <- as.integer(.newId)
            .envRef$chunks[[.newId]] <- .newChunk
            .envRef$nobs  <- .envRef$nobs  + sum(.template$evid == 0L, na.rm = TRUE)
            .envRef$ndose <- .envRef$ndose + sum(.template$evid != 0L, na.rm = TRUE)
          }
        }
      }
      for (.rmId in .removedIds) {
        if (.rmId <= length(.envRef$chunks) && !is.null(.envRef$chunks[[.rmId]])) {
          .df <- .envRef$chunks[[.rmId]]
          .envRef$nobs  <- .envRef$nobs  - sum(.df$evid == 0L, na.rm = TRUE)
          .envRef$ndose <- .envRef$ndose - sum(.df$evid != 0L, na.rm = TRUE)
          .envRef$chunks[.rmId] <- list(NULL)
        }
      }
    }
  }

  # ---- Pure-evid ----
  if (!is.null(.evidVal)) {
    .df <- .etObsChunk(0.0, cmt = .cmtVal) # nolint
    if (.evidVal != 0L) .df$evid <- as.integer(.evidVal)
    .etAddChunk(.envRef, .df, .targetIds) # nolint
    if (.evidVal == 0L) .envRef$nobs <- .envRef$nobs + max(1L, length(.targetIds))
    else .envRef$ndose <- .envRef$ndose + max(1L, length(.targetIds))
    return(.rxEtSyncData(.et)) # nolint
  }

  .rxEtSyncData(.et) # nolint
}

# rxEt object access and mutation ----------------------------------------

#' @export
`$.rxEt` <- function(obj, arg) {
  # 0. "env" is an alias for the mutable .env environment
  if (arg %in% c("env", ".env")) {
    return(.rxEtEnv(obj)) # nolint
  }
  # 1. Check method functions in .env$methods (new-style) or direct list slots (internal mini-rxEt)
  .env <- .rxEtEnv(obj) # nolint
  if (is.environment(.env)) {
    if (!is.null(.env$methods)) {
      .direct <- .env$methods[[arg]]
    } else {
      .direct <- unclass(obj)[[arg]]
    }

  } else {
    .direct <- unclass(obj)[[arg]]
  }
  if (!is.null(.direct)) return(.direct)

  # 2. Check mutable env properties (nobs, ndose, units, show, ids, chunks)
  # "id" falls through to step 3 and returns the per-row id column
  # (data-frame semantics, #1154); unique ids are available via x$env$ids
  if (!is.null(.env) && exists(arg, envir = .env, inherits = FALSE)) {
    return(get(arg, envir = .env, inherits = FALSE))
  }

  # 3. Materialize and return data column (time, amt, evid, etc.)
  if (!is.null(.env)) {
    if (arg == "id") {
      # computed from row counts rather than .etMaterialize(), which
      # draws window/addl times from the RNG; the id column itself does
      # not depend on the drawn values (rows sort by id first)
      return(.etIdColumn(obj)) # nolint
    }
    .mat <- .etMaterialize(obj) # nolint
    if (arg %in% names(.mat)) {
      .val <- .mat[[arg]]
      if (requireNamespace("units", quietly = TRUE)) {
        .tu <- .env$units["time"]
        .du <- .env$units["dosing"]
        .hasTimeU <- !is.na(.tu) && nchar(.tu) > 0
        .hasDoseU <- !is.na(.du) && nchar(.du) > 0
        if (arg == "time" && .hasTimeU) {
          return(units::set_units(.val, .tu, mode = "standard"))
        } else if (arg == "amt" && .hasDoseU) {
          return(units::set_units(.val, .du, mode = "standard"))
        } else if (arg == "ii" && .hasTimeU) {
          return(units::set_units(.val, .tu, mode = "standard"))
        } else if (arg == "rate" && .hasDoseU && .hasTimeU) {
          .rateU <- paste0(.du, "/", .tu)
          return(units::set_units(.val, .rateU, mode = "standard"))
        }
      }
      return(.val)
    }
  }
  NULL
}

select.rxEt <- function(.data, ...) {
  .full <- tibble::as_tibble(as.data.frame(.data, all = TRUE))
  dplyr::select(.full, ...)
}

filter.rxEt <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  .full <- tibble::as_tibble(as.data.frame(.data, all = TRUE))
  dplyr::filter(.full, ..., .by = {{ .by }}, .preserve = .preserve)
}

rename.rxEt <- function(.data, ...) {
  .full <- tibble::as_tibble(as.data.frame(.data, all = TRUE))
  dplyr::rename(.full, ...)
}

mutate.rxEt <- function(.data, ...) {
  .full <- tibble::as_tibble(as.data.frame(.data, all = TRUE))
  dplyr::mutate(.full, ...)
}

#' @export
names.rxEt <- function(x) {
  .env <- .rxEtEnv(x) # nolint
  .n <- names(.env$show)
  for (.ch in .env$chunks) {
    if (!is.null(.ch)) {
      .extra <- setdiff(names(.ch), .n)
      if (length(.extra) > 0L) .n <- c(.n, .extra)
      break
    }
  }
  .n
}

# rxEt data-frame methods -------------------------------------------------

#' @export
`[.rxEt` <- function(x, i, j, drop = FALSE) {
  .env0 <- .rxEtEnv(x) # nolint
  if (missing(i) && missing(j)) return(x)
  if (missing(j)) {
    # Row-only subset: return rxEt
    .full <- .etMaterialize(x) # nolint
    if (!missing(i)) .etCheckLogicalRowIndex(i, nrow(.full)) # nolint
    .sub  <- if (missing(i)) .full else .full[i, , drop = FALSE]
    .newEnv <- new.env(parent = emptyenv())
    .newEnv$units      <- .env0$units
    .newEnv$show       <- .env0$show
    .newEnv$randomType <- .env0$randomType
    .newEnv$canResize  <- FALSE
    .newEnv$extraCols  <- .etExtraCols(.env0) # nolint
    .newEnv$chunks     <- list()
    if (nrow(.sub) > 0L) {
      .newEnv$ids  <- sort(unique(as.integer(.sub$id)))
      .newEnv$nobs  <- sum(.sub$evid == 0L, na.rm = TRUE)
      .newEnv$ndose <- sum(.sub$evid != 0L, na.rm = TRUE)
      for (.ii in .newEnv$ids)
        .newEnv$chunks[[.ii]] <- .sub[.sub$id == .ii, , drop = FALSE]
    } else {
      .newEnv$ids   <- 1L
      .newEnv$nobs  <- 0L
      .newEnv$ndose <- 0L
    }
    return(structure(c(list(.env = .newEnv), .etBuildMethods(.newEnv)), class = "rxEt")) # nolint
  }
  .mat <- as.data.frame(x, all = TRUE)
  if (missing(i)) return(.mat[, j, drop = drop])
  .etCheckLogicalRowIndex(i, nrow(.mat)) # nolint
  .mat[i, j, drop = drop]
}

#' Guard against silent logical recycling in rxEt row subsets (#1154)
#'
#' @param i row index supplied to `[.rxEt`
#' @param nrow number of rows in the materialized event table
#' @return nothing; stops when a logical index cannot recycle cleanly
#' @noRd
.etCheckLogicalRowIndex <- function(i, nrow) {
  if (is.logical(i) && length(i) > 1L && length(i) != nrow) {
    stop(sprintf("logical row index of length %d does not match the number of event table rows (%d)",
                 length(i), nrow), call. = FALSE)
  }
  invisible(NULL)
}

#' Non-canonical columns that were explicitly assigned on an event table
#'
#' Columns that merely rode along with an imported data frame are not
#' listed here, so they stay hidden in the default `as.data.frame()`
#' output the way they were before #1154.
#'
#' @param env event table environment (or `NULL`)
#' @return character vector of column names, possibly empty
#' @noRd
.etExtraCols <- function(env) {
  if (!is.environment(env)) return(character(0))
  .extra <- env$extraCols
  if (is.null(.extra)) return(character(0))
  as.character(.extra)
}

#' Columns displayed for an event table
#'
#' The canonical columns flagged in `show`, followed by any explicitly
#' assigned columns (e.g. `ev$wt <- 70`) so they are not silently
#' swallowed by the event table (#1154).
#'
#' @param nms names available in the data being displayed
#' @param show named logical vector of canonical columns to show
#' @param extraCols explicitly assigned non-canonical columns
#' @param pre columns to place first (e.g. `id`)
#' @return character vector of column names present in `nms`
#' @noRd
.etDisplayCols <- function(nms, show, extraCols = character(0), pre = character(0)) {
  .cols <- pre
  if (!is.null(show)) .cols <- c(.cols, names(show)[show])
  intersect(unique(c(.cols, extraCols)), nms)
}

#' Provenance for explicitly assigned columns carried on a data frame
#'
#' `as.data.frame()` tags the columns that were explicitly assigned so a
#' round trip back through `et()`/`$import.EventTable()` keeps showing
#' them instead of demoting them to hidden imported covariates (#1154).
#' A data frame built by hand carries no tag, so its covariates stay
#' hidden the way they always have.
#'
#' @param df object to read the tag from
#' @return character vector of column names, possibly empty
#' @noRd
.etExtraColsAttr <- function(df) {
  .extra <- attr(df, "rxEtExtraCols", exact = TRUE)
  if (is.null(.extra)) return(character(0))
  as.character(.extra)
}

#' Carry an imported data frame's assigned-column tag into an event table
#'
#' @param env event table environment to modify by reference
#' @param df object the tag was read from (before any renaming)
#' @param nms names of the data actually stored
#' @return nothing, called for the side effect on `env`
#' @noRd
.etImportExtraCols <- function(env, df, nms) {
  .etAddExtraCols(env, intersect(.etExtraColsAttr(df), nms)) # nolint
}

#' Mark an event table frame so it prints only its display columns
#'
#' Every column is kept for programmatic access; only the printed output
#' is limited to `.etDisplayCols()`, so an un-grouped `$get.dosing()` no
#' longer dumps hidden internal/covariate columns the way a compressed
#' preview never did.
#'
#' @param x data frame to mark, or `NULL`
#' @param env event table environment supplying `show`/`extraCols`
#' @return `x` with the `rxEtPreview` print class attached
#' @noRd
.etMarkDisplay <- function(x, env) {
  if (is.null(x) || inherits(x, "rxEtPreview")) return(x)
  attr(x, "rxEtShow") <- env$show
  attr(x, "rxEtExtraCols") <- .etExtraCols(env) # nolint
  attr(x, "rxEtMarkedCols") <- names(x)
  class(x) <- c("rxEtPreview", class(x))
  x
}

#' Columns hidden when an event table frame prints
#'
#' Only columns that were present -- and hidden -- when the frame was
#' marked are dropped, so a column added or renamed afterwards still
#' prints instead of silently disappearing.
#'
#' @param nms names of the data being printed
#' @param marked names present when the frame was marked, or `NULL`
#' @param show named logical vector of canonical columns to show
#' @param extraCols explicitly assigned non-canonical columns
#' @return character vector of columns to keep, in `nms` order
#' @noRd
.etKeepCols <- function(nms, marked, show, extraCols = character(0)) {
  if (is.null(marked)) marked <- nms
  .hide <- setdiff(marked, .etDisplayCols(marked, show, extraCols, pre = "id"))
  # not setdiff(): that would drop duplicated column names as well
  .keep <- nms[!(nms %in% .hide)]
  if (length(.keep) == 0L) return(nms)
  .keep
}

#' Drop the display marking once an event table frame is modified
#'
#' The marking describes the columns the accessor returned, so it is only
#' meaningful while the frame is unchanged.  Mutating one in place makes
#' it the caller's own data frame -- the same way `dplyr` verbs do via
#' `dplyr_reconstruct()` -- so the print class goes away rather than
#' hiding a column that was just assigned.
#'
#' @param x object being assigned into
#' @return `x` without the `rxEtPreview` class
#' @noRd
.etUnmarkDisplay <- function(x) {
  if (inherits(x, "rxEtPreview")) {
    class(x) <- setdiff(class(x), "rxEtPreview")
  }
  x
}

#' @export
`$<-.rxEtPreview` <- function(x, name, value) {
  .etUnmarkDisplay(NextMethod()) # nolint
}

#' @export
`[[<-.rxEtPreview` <- function(x, ..., value) {
  .etUnmarkDisplay(NextMethod()) # nolint
}

#' @export
`[<-.rxEtPreview` <- function(x, ..., value) {
  .etUnmarkDisplay(NextMethod()) # nolint
}

#' @export
`[.rxEtPreview` <- function(x, ...) {
  .ret <- NextMethod()
  # The marking says which columns were hidden when the accessor built the
  # frame, so it still describes a row subset -- `head(ev$get.dosing())` keeps
  # printing display columns only.  Picking columns makes it the caller's own
  # frame: `dplyr`'s column verbs (select(), relocate()) subset with `[`
  # rather than going through `dplyr_reconstruct()`, so this is what degrades
  # them to a plain data frame the way mutate()/filter() already were.  A bare
  # rename is deliberately still marked: `.etKeepCols()` hides only the names
  # marked at the time, so a renamed column keeps printing either way.
  # `[.data.frame` keeps the marker attributes on the row form but drops them
  # on the column form, so `x[TRUE]` would otherwise keep the class with
  # nothing left to drive it and print every column anyway.
  if (!is.data.frame(.ret) ||
        is.null(attr(.ret, "rxEtShow", exact = TRUE)) ||
        !identical(names(.ret), names(x))) {
    return(.etUnmarkDisplay(.ret)) # nolint
  }
  # A compressed preview is the exception: its groups count rows, and print()
  # both synthesizes the `id` column from them and heads the output with how
  # many individuals they cover, so dropping rows makes the marking a lie.
  if (!is.null(attr(.ret, "rxEtPreviewGroups", exact = TRUE)) &&
        !identical(nrow(.ret), nrow(x))) {
    return(.etUnmarkDisplay(.ret)) # nolint
  }
  .ret
}

#' Record a non-canonical column as explicitly assigned
#'
#' @param env event table environment to modify by reference
#' @param cols character vector of column names to add
#' @return nothing, called for the side effect on `env`
#' @noRd
.etAddExtraCols <- function(env, cols) {
  if (!is.environment(env) || length(cols) == 0L) return(invisible(NULL))
  env$extraCols <- unique(c(.etExtraCols(env), as.character(cols))) # nolint
  invisible(NULL)
}

#' Forget a non-canonical column that was removed from an event table
#'
#' @param env event table environment to modify by reference
#' @param cols character vector of column names to drop
#' @return nothing, called for the side effect on `env`
#' @noRd
.etDropExtraCols <- function(env, cols) {
  if (!is.environment(env) || length(cols) == 0L) return(invisible(NULL))
  env$extraCols <- setdiff(.etExtraCols(env), as.character(cols)) # nolint
  invisible(NULL)
}

#' @export
`$<-.rxEt` <- function(x, name, value) {
  .df <- .etMaterialize(x) # nolint
  .df[[name]] <- value
  .ret <- .rxEtRebuildShell(x, .df) # nolint
  # an explicitly assigned column should display, so the default
  # as.data.frame() output reflects what will be solved (#1154)
  .env <- .rxEtEnv(.ret) # nolint
  if (is.null(value)) {
    # the column was deleted, so it is neither shown nor tracked
    .etDropExtraCols(.env, name) # nolint
  } else if (name %in% names(.env$show)) {
    .env$show[name] <- TRUE
  } else {
    .etAddExtraCols(.env, name) # nolint
  }
  .ret
}

drop_units.rxEt <- function(x) {
  if (!requireNamespace("units", quietly = TRUE)) {
    stop("requires package 'units'", call. = FALSE)
  }
  if (is.rxEt(x)) { # nolint
    .env <- .rxEtEnv(x) # nolint
    .env$units["dosing"] <- NA_character_
    .env$units["time"]   <- NA_character_
    return(x)
  }
  stop("invalid event table", call. = FALSE)
}

set_units.rxEt <- function(x, value, ..., mode = .setUnitsMode()) { # nolint
  if (is.null(mode)) {
    stop("requires package 'units'", call. = FALSE)
  }
  if (missing(value)) {
    value <- .unitless() # nolint
  } else if (mode == "symbols") {
    value <- substitute(value)
    if (is.numeric(value) && !identical(value, 1) && !identical(value, 1L)) {
      stop("the only valid number defining a unit is '1', signifying a unitless unit", call. = FALSE)
    }
  }
  if (identical(value, .unitless())) { # nolint
    warning("clearing both amount and time units\n",
      "for more precise control use 'et(amountUnits=\"\")' or 'et(timeUnits=\"\")'",
      call. = FALSE
    )
    if (is.rxEt(x)) { # nolint
      .env <- .rxEtEnv(x) # nolint
      .env$units["dosing"] <- NA_character_
      .env$units["time"]   <- NA_character_
      return(x)
    }
    stop("invalid event table", call. = FALSE)
  } else {
    if (!inherits(value, "character")) value <- deparse(value)
    .isTime <- try(units::set_units(units::set_units(1, value, mode = "standard"), "sec"), silent = TRUE)
    if (inherits(.isTime, "try-error")) {
      ## Amount
      if (is.rxEt(x)) { # nolint
        .env <- .rxEtEnv(x) # nolint
        .env$units["dosing"] <- value
        return(x)
      }
      stop("invalid event table", call. = FALSE)
    } else {
      ##
      if (is.rxEt(x)) { # nolint
        .env <- .rxEtEnv(x) # nolint
        .env$units["time"] <- value
        return(x)
      }
      stop("invalid event table", call. = FALSE)
    }
  }
}

# Event-table solve dispatch and piping helpers ---------------------------

#' Dispatch solve to 'rxode2' solve
#'
#'
#' @param x rxode2 solve dispatch object
#' @param ...  other arguments
#' @return if 'rxode2'  is loaded, a solved object, otherwise an error
#' @author Matthew L. Fidler
#' @export
rxEtDispatchSolve <- function(x, ...) {
  UseMethod("rxEtDispatchSolve")
}

#' @rdname rxEtDispatchSolve
#' @export
rxEtDispatchSolve.default <- function(x, ...) {
  stop("need 'rxode2' loaded for piping to a simulation")
}

#' @export
simulate.rxEt <- function(object, nsim = 1, seed = NULL, ...) {
  .isPipe <- as.character(substitute(object))
  if (length(.isPipe) == 1 && .isPipe == ".") {
    .isPipe <- TRUE
  } else if (missing(object)) {
    .isPipe <- FALSE
  } else {
    .isPipe <- substitute(object)
    if (is.call(.isPipe) && length(.isPipe) >= 1L) {
      # This will assume the input is going to be an et compatible object
      .isPipe <- TRUE
    } else {
      if (is.symbol(.isPipe)) {
        .isPipe <- TRUE
      } else {
        .isPipe <- FALSE
      }
    }
  }
  if (is.null(.etInfo$pipelineRx) || !.isPipe) {
    if (!missing(nsim)) warning("'nsim' is ignored when simulating event tables", call. = FALSE)
    if (!is.null(seed)) set.seed(seed)
    if (is.rxEt(object)) { # nolint
      .env0 <- .rxEtEnv(object) # nolint
      .sim <- .etSimulateRepresentation(.env0) # nolint
      if (!isTRUE(.sim$hasWin)) {
        warning("simulating event table without windows returns identical event table", call. = FALSE)
      }
      .newEnv <- new.env(parent = emptyenv())
      .newEnv$units      <- .env0$units
      .newEnv$show       <- .env0$show
      .newEnv$ids        <- .env0$ids
      .newEnv$nobs       <- .env0$nobs
      .newEnv$ndose      <- .env0$ndose
      .newEnv$randomType <- NA_integer_
      .newEnv$canResize  <- FALSE
      .newEnv$extraCols  <- .etExtraCols(.env0) # nolint
      .newEnv$groups     <- .sim$groups
      .newEnv$chunks     <- .sim$chunks
      return(structure(c(list(.env = .newEnv),
                         .etBuildMethods(.newEnv)), # nolint
                       class = "rxEt"))
    }
    stop("invalid event table", call. = FALSE)
  } else {
    .ret <- list(object, ..., seed = seed, nsim = nsim)
    class(.ret) <- "rxode2et"
    rxEtDispatchSolve(.ret)
  }
}

# Public event-table constructors and combinators -------------------------

#' Add dosing to eventTable
#'
#' This adds a dosing event to the event table.  This is provided for
#' piping syntax through magrittr.  It can also be accessed by `eventTable$add.dosing(...)`
#'
#' @param eventTable eventTable object; When accessed from object it would be `eventTable$`
#' @param dose numeric scalar, dose amount in `amount.units`;
#' @param nbr.doses integer, number of doses;
#' @param dosing.interval required numeric scalar, time between doses
#'     in `time.units`, defaults to 24 of
#'     `time.units="hours"`;
#' @param dosing.to integer, compartment the dose goes into (first
#'     compartment by default);
#' @param rate for infusions, the rate of infusion (default is
#'     `NULL`, for bolus dosing;
#' @param amount.units optional string indicating the dosing units.
#'     Defaults to `NA` to indicate as per the original
#'     `EventTable` definition.
#' @param start.time required dosing start time;
#' @param do.sampling logical, should observation sampling records be
#'     added at the dosing times? Defaults to `FALSE`.
#' @param time.units optional string indicating the time units.
#'     Defaults to `"hours"` to indicate as per the original
#'     `EventTable` definition.
#' @param ... Other parameters passed to [et()].
#' @return eventTable with updated dosing (note the event table will
#'     be updated anyway)
#' @author Matthew L. Fidler
#' @template etExamples
#' @export
add.dosing <- function(eventTable, dose, nbr.doses = 1L,
                       dosing.interval = 24, dosing.to = 1L,
                       rate = NULL, amount.units = NA_character_,
                       start.time = 0.0, do.sampling = FALSE,
                       time.units = NA_character_, ...) {
  if (is.rxEt(eventTable)) { # nolint
    .lst <- list(
      x = eventTable,
      amt = dose,
      time = start.time,
      ii = if (nbr.doses > 1L) dosing.interval else 0.0,
      addl = as.integer(nbr.doses) - 1L,
      addSampling = do.sampling
    )
    if (!is.null(rate)) .lst$rate <- rate
    if (!is.na(amount.units)) .lst$amountUnits <- amount.units
    if (!is.na(time.units)) .lst$timeUnits <- time.units
    .extra <- list(...)
    if (is.null(.extra$cmt) && is.null(.extra$dosing.to) &&
          !identical(dosing.to, 1L)) {
      .lst$cmt <- dosing.to
    }
    return(invisible(do.call(et, c(.lst, .extra))))
  } else if (inherits(eventTable, "rxSolve")) {
    .et <- eventTable$get.EventTable()
    .newEt <- et(x = .et, amt = dose, time = start.time,
                 ii = if (nbr.doses > 1L) dosing.interval else 0.0,
                 addl = as.integer(nbr.doses) - 1L,
                 addSampling = do.sampling, ...)
    return(rxSolve(eventTable, events = .newEt, updateObject = FALSE)) # nolint
  }
  # Fallback for rxSolve objects and old-style EventTable via C++
  stop("invalid event table", call. = FALSE)
}

#' Add sampling to eventTable
#'
#' This adds a dosing event to the event table.  This is provided for
#' piping syntax through magrittr.  It can also be accessed by
#' `eventTable$add.sampling()`
#'
#' @param eventTable An eventTable object. When accessed from object it would be `eventTable$`
#' @param time a vector of time values (in `time.units`).
#' @param time.units an optional string specifying the time
#'     units. Defaults to the units specified when the
#'     `EventTable` was initialized.
#' @return eventTable with updated sampling.  (Note the event table
#'     will be updated even if you don't reassign the eventTable)
#' @template etExamples
#' @export
add.sampling <- function(eventTable, time, time.units = NA_character_) {
  if (is.rxEt(eventTable)) { # nolint
    eventTable$add.sampling(time, time.units = time.units)
    return(.rxEtSyncData(eventTable)) # nolint
  } else if (inherits(eventTable, "rxSolve")) {
    .et <- eventTable$get.EventTable()
    .newEt <- et(x = .et, time = time, timeUnits = time.units)
    return(rxSolve(eventTable, events = .newEt, updateObject = FALSE)) # nolint
  }
  # Fallback for rxSolve objects and old-style EventTable via C++
  stop("invalid event table", call. = FALSE)
}


#' Create an event table object
#'
#' Initializes an object of class \sQuote{EventTable} with methods for
#' adding and querying dosing and observation records
#'
#' @param amount.units string denoting the amount dosing units, e.g.,
#'      \dQuote{mg}, \dQuote{ug}. Default to `NA` to denote
#'      unspecified units.  It could also be a solved rxode2 object.  In
#'      that case, eventTable(obj) returns the eventTable that was used
#'      to solve the rxode2 object.
#'
#' @param time.units string denoting the time units, e.g.,
#'      \dQuote{hours}, \dQuote{days}. Default to `"hours"`.
#'
#'  An `eventTable` is an object that consists of a data.frame
#'  storing ordered time-stamped events of an (unspecified) PK/PD
#'  dynamic system, units (strings) for dosing and time records, plus a
#'  list of functions to add and extract event records.
#'
#'  Currently, events can be of two types: dosing events that represent
#'  inputs to the system and sampling time events that represent
#'  observations of the system with \sQuote{amount.units} and
#'  \sQuote{time.units}, respectively.
#'
#'
#' @return A modified data.frame with the following accessible functions:
#'
#' * `get.EventTable()` returns the current event table
#'
#' * [add.dosing()]  adds dosing records to the event table.
#'
#' * `get.dosing()` returns a data.frame of dosing records.
#'
#' * `clear.dosing()` clears or deletes all dosing from event table
#'
#' *  `[add.sampling()] adds sampling time observation records to the
#'        event table.
#'
#' * `get.sampling()`returns a data.frame of sampled observation records.
#'
#' * `clear.sampling()` removes all sampling from event table.
#'
#' * `get.obs.rec()` returns a logical vector indicating whether each
#'    event record represents an observation or not.
#'
#' * `get.nobs()` returns the number of observation (not dosing) records.
#'
#' * `get.units()` returns a two-element character vector with the
#'        dosing and time units, respectively
#'
#' * `copy()` makes a copy of the current event table. To create
#'        a copy of an event table object use `qd2 <- qd$copy()`
#'
#' * `expand()` Expands the event table for multi-subject solving.
#'    This is done by `qd$expand(400)` for a 400 subject data expansion
#'
#' @author Matthew Fidler, Melissa Hallow and Wenping Wang
#'
#' @seealso [et()]
#'
#' @examples
#' # create dosing and observation (sampling) events
#' # QD 50mg dosing, 5 days followed by 25mg 5 days
#' #
#' qd <- eventTable(amount.units = "mg", time.units = "days")
#' #
#' qd$add.dosing(dose = 50, nbr.doses = 5, dosing.interval = 1, do.sampling = FALSE)
#' #
#' # sample the system's drug amounts hourly the first day, then every 12 hours
#' # for the next 4 days
#' qd$add.sampling(seq(from = 0, to = 1, by = 1 / 24))
#' qd$add.sampling(seq(from = 1, to = 5, by = 12 / 24))
#' #
#' # print(qd$get.dosing())     # table of dosing records
#' print(qd$get.nobs()) # number of observation (not dosing) records
#' #
#' # BID dosing, 5 days
#' bid <- eventTable("mg", "days") # only dosing
#' bid$add.dosing(
#'   dose = 10000, nbr.doses = 2 * 5,
#'   dosing.interval = 12, do.sampling = FALSE
#' )
#' #
#' # Use the copy() method to create a copy (clone) of an existing
#' # event table (simple assignments just create a new reference to
#' # the same event table object (closure)).
#' #
#' bid.ext <- bid$copy() # three-day extension for a 2nd cohort
#' bid.ext$add.dosing(
#'   dose = 5000, nbr.doses = 2 * 3,
#'   start.time = 120, dosing.interval = 12, do.sampling = FALSE
#' )
#'
#' # You can also use the Piping operator to create a table
#'
#' qd2 <- eventTable(amount.units = "mg", time.units = "days") |>
#'   add.dosing(dose = 50, nbr.doses = 5, dosing.interval = 1, do.sampling = FALSE) |>
#'   add.sampling(seq(from = 0, to = 1, by = 1 / 24)) |>
#'   add.sampling(seq(from = 1, to = 5, by = 12 / 24))
#' # print(qd2$get.dosing())     # table of dosing records
#' print(qd2$get.nobs()) # number of observation (not dosing) records
#'
#' # Note that piping with |> will update the original table.
#'
#' qd3 <- qd2 |> add.sampling(seq(from = 5, to = 10, by = 6 / 24))
#' print(qd2$get.nobs())
#' print(qd3$get.nobs())
#' @keywords models data
#' @concept ordinary differential equations
#' @concept Nonlinear regression
#' @concept Pharmacokinetics (PK)
#' @concept Pharmacodynamics (PD)
#' @export
eventTable <- function(amount.units = NA, time.units = NA) {
  .amtU <- if (!missing(amount.units)) as.character(amount.units) else NA_character_
  .timU <- if (!missing(time.units))   as.character(time.units)   else NA_character_
  .newRxEt(amountUnits = .amtU, timeUnits = .timU) # nolint
}
# nolint end

#' Sequence of event tables
#'
#' This combines a sequence of event tables.
#'
#' @param ... The event tables and optionally time between event
#'     tables, called waiting times in this help document.
#'
#' @param samples How to handle samples when repeating an event
#'     table.  The options are:
#'
#' * `"clear"` Clear sampling records before combining the datasets
#' * `"use"` Use the sampling records when combining the datasets
#'
#' @param waitII This determines how waiting times between events are
#'     handled. The options are:
#'
#' * `"smart"` This "smart" handling of waiting times is the
#'   default option.  In this case, if the waiting time is above the
#'   last observed inter-dose interval in the first combined event
#'   table, then the actual time between doses is given by the wait
#'   time.  If it is smaller than the last observed inter-dose
#'   interval, the time between event tables is given by the inter-dose
#'   interval + the waiting time between event tables.
#'
#' * `"+ii"` In this case, the wait time is added to the
#'    inter-dose interval no matter the length of the wait time or
#'    inter-dose interval
#'
#' @param ii If there was no inter-dose intervals found in the event
#'     table, assume that the interdose interval is given by this
#'     `ii` value.  By default this is `24`.
#'
#' @return An event table
#'
#' @details
#'
#' This `seq`uences all the event tables in added in the
#' argument list `...`.  By default when combining the event
#' tables the offset is at least by the last inter-dose interval in
#' the prior event table (or `ii`).  If you separate any of the
#' event tables by a number, the event tables will be separated at
#' least the wait time defined by that number or the last inter-dose
#' interval.
#'
#' @template etExamples
#'
#' @export
etSeq <- function(..., samples = c("clear", "use"),
                  waitII = c("smart", "+ii"), ii = 24) {
  .samples    <- match.arg(samples)
  .waitType   <- match.arg(waitII)
  .args       <- list(...)
  .explicitIi <- !missing(ii)

  .chunks    <- list()
  .groups    <- list()
  .nobs      <- 0L
  .ndose     <- 0L
  .units     <- NULL
  .show      <- NULL
  .ids       <- integer(0)
  .timeDelta <- 0.0
  .lastIi    <- 0.0
  .lastDose  <- 0.0

  # Pre-scan: warn once if no event table has an inter-dose interval
  .etItems <- Filter(is.rxEt, .args) # nolint
  if (length(.etItems) > 1L) {
    .hasAnyIi <- any(vapply(.etItems, function(.a) {
      isTRUE(.rxEtEnv(.a)$show[["ii"]]) # nolint
    }, logical(1)))
    if (!.hasAnyIi) {
      warning("No inter-dose interval found in event tables; using ii=", ii, call. = FALSE)
    }
  }

  for (.item in .args) {
    if (is.rxEt(.item)) { # nolint
      .ret <- .etSeqHandleRxEt(.item, .units, .show, .ids, .timeDelta, .samples, # nolint
                               .chunks, .groups, .nobs, .ndose, .explicitIi, ii)
      .units     <- .ret$units
      .show      <- .ret$show
      .ids       <- .ret$ids
      .chunks    <- .ret$chunks
      .groups    <- .ret$groups
      .nobs      <- .ret$nobs
      .ndose     <- .ret$ndose
      .timeDelta <- .ret$timeDelta
      .lastIi    <- .ret$lastIi
      .lastDose  <- .ret$lastDose
    } else if (is.numeric(.item) || is.integer(.item)) {
      .timeDelta <- .etSeqHandleWait(.item, .waitType, .lastDose, .lastIi, ii, .timeDelta) # nolint
    }
  }

  .newEnv <- new.env(parent = emptyenv())
  .newEnv$chunks     <- .chunks
  .newEnv$groups     <- .groups
  .newEnv$units      <- if (!is.null(.units)) .units else c(dosing = NA_character_, time = NA_character_)
  if (!is.null(.show)) {
    .newShow <-  .show
  } else {
    .newShow <-  .etDefaultShow() # nolint
  }

  if (.explicitIi || .ndose > 0L) {
    .newShow["ii"]   <- TRUE
    .newShow["addl"] <- TRUE
  }
  .newEnv$show       <- .newShow
  if (length(.ids) > 1L) {
    .newEnv$ids        <- .ids
  } else {
    .newEnv$ids        <- 1L
  }

  .newEnv$nobs       <- .nobs
  .newEnv$ndose      <- .ndose
  .newEnv$randomType <- NA_integer_
  .newEnv$canResize  <- FALSE
  .newEnv$extraCols  <- unique(unlist(lapply(.etItems, function(.a) {
    .etExtraCols(.rxEtEnv(.a)) # nolint
  }), use.names = FALSE))
  if (length(.newEnv$ids) > 1L) {
    .newEnv$show["id"] <- TRUE
  }
  structure(c(list(.env = .newEnv),
              .etBuildMethods(.newEnv) # nolint
             ), class = "rxEt")
}
#' Combining event tables
#'
#' @inheritParams etSeq
#' @param id This is how rbind will handle ids.  There are two different types of options:
#'
#' * `merge` with `id="merge"`, the ids are merged together,
#' overlapping ids would be merged into a single event table.
#'
#' * `unique` with `id="unique"`, the ids will be renumbered
#' so that the ids in all the event tables are not overlapping.
#'
#' @param
#' deparse.level The `deparse.level` of a traditional
#'     `rbind` is ignored.
#'
#' @author Matthew L Fidler
#'
#' @return An event table
#'
#' @template etExamples
#'
#' @export
etRbind <- function(..., samples = c("use", "clear"),
                    waitII = c("smart", "+ii"),
                    id = c("merge", "unique")) {
  .samples  <- match.arg(samples)
  .uniqueId <- match.arg(id) == "unique"
  .ets <- list(...)

  .chunks  <- list()
  .nobs    <- 0L
  .ndose   <- 0L
  .units   <- NULL
  .show    <- NULL
  .ids     <- integer(0)
  .groups  <- list()
  .nextId  <- 0L

  for (.et in .ets) {
    if (!is.rxEt(.et)) next # nolint
    .env <- .rxEtEnv(.et) # nolint
    if (is.null(.units)) {
      .units <- .env$units
      .show  <- .env$show
    } else {
      .show <- .show | .env$show
    }
    # ID remapping for unique mode (always materializes)
    if (.uniqueId || .samples == "clear") {
      if (length(.groups) > 0L) {
        for (.g in .groups) {
          .chunks <- .addRowsToChunks(.chunks, .etMaterializeGroup(.g)) # nolint
        }
        .groups <- list()
      }
      .mat    <- .etMaterialize(.et) # nolint
      if (.uniqueId) {
        .oldIds <- sort(unique(.mat$id))
        .map    <- seq_along(.oldIds) + .nextId
        .nextId <- .nextId + length(.oldIds)
        .mat$id <- .map[match(.mat$id, .oldIds)]
        .ids    <- c(.ids, .map)
      } else {
        .ids <- sort(unique(c(.ids, .env$ids)))
      }
      if (.samples == "clear") {
        .mat <- .mat[.mat$evid != 0L, , drop = FALSE]
      }
      .chunks <- .addRowsToChunks(.chunks, .mat) # nolint
    } else {
      .etGroupsIn <- .etGetGroups(.env) # nolint
      if (length(.chunks) == 0L &&
          length(.etGroupsIn) > 0L &&
          length(intersect(.ids, .env$ids)) == 0L) {
        .groups <- c(.groups, lapply(.etGroupsIn, function(.g) {
          list(ids = as.integer(.g$ids), data = .g$data)
        }))
      } else if (length(.chunks) == 0L &&
                 length(.groups) == 1L &&
                 length(.etGroupsIn) == 1L &&
                 .etGroupIdsEqual(.groups[[1]]$ids, .etGroupsIn[[1]]$ids)) {
        .groups[[1]]$data <- as.data.frame(
          data.table::rbindlist(list(.groups[[1]]$data, .etGroupsIn[[1]]$data), fill = TRUE)
        )
      } else {
        if (length(.groups) > 0L) {
          for (.g in .groups) {
            .chunks <- .addRowsToChunks(.chunks, .etMaterializeGroup(.g)) # nolint
          }
          .groups <- list()
        }
        if (length(.etGroups(.env)) > 0L) { # nolint
          .chunks <- .addRowsToChunks(.chunks, .etMaterialize(.et)) # nolint
        } else {
          # Merge indexed chunks directly
          for (.ci in seq_along(.env$chunks)) {
            if (!is.null(.env$chunks[[.ci]])) {
              .existing <- if (.ci <= length(.chunks)) .chunks[[.ci]] else NULL
              .chunks[[.ci]] <- as.data.frame(
                data.table::rbindlist(list(.existing, .env$chunks[[.ci]]), fill = TRUE)
              )
            }
          }
        }
      }
      .ids    <- sort(unique(c(.ids, .env$ids)))
    }
    if (.samples == "use") {
      .nobs  <- .nobs + .env$nobs
    }
    .ndose <- .ndose + .env$ndose
  }

  .newEnv <- new.env(parent = emptyenv())
  .newEnv$chunks     <- .chunks
  .newEnv$groups     <- .groups
  if (!is.null(.units)) {
    .newEnv$units      <-  .units
  } else {
    .newEnv$units      <- c(dosing = NA_character_, time = NA_character_)
  }
  if (!is.null(.show))  {
    .newEnv$show       <- .show
  } else {
    .newEnv$show       <- .etDefaultShow() # nolint
  }
  if (length(.ids) > 1L) {
    .newEnv$ids        <- .ids
  } else {
    .newEnv$ids        <- 1L
  }
  .newEnv$nobs       <- .nobs
  .newEnv$ndose      <- .ndose
  .newEnv$randomType <- NA_integer_
  .newEnv$canResize  <- FALSE
  .newEnv$extraCols  <- unique(unlist(lapply(Filter(is.rxEt, .ets), function(.a) { # nolint
    .etExtraCols(.rxEtEnv(.a)) # nolint
  }), use.names = FALSE))
  if (length(.newEnv$ids) > 1L) {
    .newEnv$show["id"] <- TRUE
  }
  structure(c(list(.env = .newEnv),
              .etBuildMethods(.newEnv)), # nolint
            class = "rxEt")
}

#' @rdname etRbind
#' @export
rbind.rxEt <- function(..., deparse.level = 1) {
  if (!missing(deparse.level)) warning("'deparse.level' not used with rxode2 event tables", call. = FALSE)
  do.call(etRbind, list(...))
}

#' @rdname etSeq
#' @export
seq.rxEt <- function(...) {
  do.call(etSeq, list(...))
}

#' @export
c.rxEt <- function(...) {
  do.call(etSeq, list(...))
}

#' Repeat an rxode2 event table
#'
#' @param x An rxode2 event table
#' @param times Number of times to repeat the event table
#' @param length.out Invalid with rxode2 event tables, will throw an
#'     error if used.
#' @param each Invalid with rxode2 event tables, will throw an error
#'     if used.
#' @param n The number of times to repeat the event table.  Overrides
#'     `times`.
#' @param wait Waiting time between each repeated event table.  By
#'     default there is no waiting, or wait=0
#' @inheritParams et
#' @inheritParams etSeq
#' @template etExamples
#' @return An event table
#' @export
etRep <- function(x, times = 1, length.out = NA, each = NA, n = NULL, wait = 0, id = integer(0),
                  samples = c("clear", "use"),
                  waitII = c("smart", "+ii"), ii = 24) {
  if (!is.na(length.out)) stop("'length.out' makes no sense with event tables", call. = FALSE)
  if (!is.na(each))       stop("'each' makes no sense with event tables", call. = FALSE)
  if (!is.null(n)) times <- n
  if (is.rxEt(x)) { # nolint
    .xEnv <- .rxEtEnv(x) # nolint
    if (is.environment(.xEnv) && isFALSE(.xEnv$canResize)) {
      warning("event table has been expanded; rep may produce unexpected results", call. = FALSE)
    }
  }
  # Convert units wait to numeric in event table's time units
  if (inherits(wait, "units")) {
    if (!requireNamespace("units", quietly = TRUE)) {
      warning("'wait' has units but 'units' package not available; using numeric value", call. = FALSE)
      wait <- as.numeric(wait)
    } else {
      .timeU <- .rxEtEnv(x)$units["time"] # nolint
      if (!is.na(.timeU) && nchar(.timeU) > 0) {
        wait <- as.numeric(units::set_units(wait, .timeU, mode = "standard"))
      } else {
        warning("'wait' has units but event table has no time units; using numeric value", call. = FALSE)
        wait <- as.numeric(wait)
      }
    }
  }
  # Build alternating list: et, wait, et, wait, ...
  .lst <- vector("list", times * 2L)
  for (.i in seq_len(times)) {
    .lst[[.i * 2L - 1L]] <- x
    .lst[[.i * 2L]]      <- wait
  }
  do.call(etSeq, c(.lst, list(samples = samples, waitII = waitII, ii = ii)))
}

#' @rdname etRep
#' @export
rep.rxEt <- function(x, ...) {
  do.call(etRep, list(x = x, ...))
}

# rxEt coercion and data conversion --------------------------------------

#' Coerce object to data.frame
#'
#' @param x Object to coerce to et.
#' @param ... Other parameters
#' @return An event table
#' @export
as.et <- function(x, ...) {
  UseMethod("as.et")
}
#' @rdname as.et
#' @export
as.et.default <- function(x, ...) {
  .e <- et()
  .e$importEventTable(as.data.frame(x))
  .e
}

#' @export
as.data.frame.rxEt <- function(x, row.names = NULL, optional = FALSE, ...) {
  .env <- tryCatch(.rxEtEnv(x), error = function(e) NULL) # nolint
  .lst <- list(...)
  if (!is.environment(.env)) {
    # Old-style EventTable: strip rxEt class and delegate
    if (is.data.frame(x)) {
      .cls <- class(x)
      class(x) <- .cls[.cls != "rxEt"]
      return(as.data.frame(x, ...))
    }
    return(data.frame())
  }
  .show <- .env$show
  .full <- .etMaterialize(x) # nolint
  # keep explicitly assigned columns (e.g. covariates from ev$wt <- ...), #1154
  .extraCols <- intersect(.etExtraCols(.env), names(.full)) # nolint
  if (isTRUE(.lst$all)) {
    .ret <- .full
  } else {
    .showCols <- names(.show)[.show]
    .showCols <- intersect(.showCols, names(.full))
    .ret <- .full[, c(.showCols, setdiff(.extraCols, .showCols)), drop = FALSE]
  }
  # tag which columns were assigned so et(as.data.frame(ev)) still shows them
  if (length(.extraCols) > 0L) attr(.ret, "rxEtExtraCols") <- .extraCols
  .ret
}

.datatable.aware <- TRUE # nolint
#' Convert an event table to a data.table
#'
#' @inheritParams data.table::as.data.table
#'
#' @return data.table of event table
#'
#' @export
as.data.table.rxEt <- function(x, keep.rownames = FALSE, ...) { # nolint
  rxReq("data.table") # nolint
  data.table::as.data.table(as.data.frame.rxEt(x, ...), keep.rownames = keep.rownames)
}

#' Convert to tbl
#'
#' @param x rxode2 event table
#'
#' @param ... Other arguments to `as_tibble`
#'
#' @return tibble of event table
#'
#' @export
as_tibble.rxEt <- function(x, ...) {
  rxReq("tibble") # nolint
  tibble::as_tibble(as.data.frame.rxEt(x, ...), ...)
}


#' Expand additional doses
#'
#' @param et Event table to expand additional doses for.
#' @return New event table with `addl` doses expanded
#' @author Matthew Fidler
#' @examples
#' ev <- et(amt = 3, ii = 24, until = 240)
#' print(ev)
#' etExpand(ev) # expands event table, but doesn't modify it
#'
#' print(ev)
#'
#' ev$expand() ## Expands the current event table and saves it in ev
#' @export
etExpand <- function(et) {
  .env      <- .rxEtEnv(et) # nolint
  .newEnv <- new.env(parent = emptyenv())
  .newEnv$chunks <- list()
  .groups <- .etGetGroups(.env) # nolint
  if (length(.groups) > 0L) {
    .newEnv$groups <- lapply(.groups, function(.g) {
      list(ids = as.integer(.g$ids), data = .etExpandGroupData(.g$data, .env)) # nolint
    })
  } else {
    .mat      <- .etMaterialize(et) # nolint
    .expanded <- .etExpandAddl(.mat, .env) # nolint
    .newEnv$groups <- list()
    if (nrow(.expanded) > 0L) {
      .ids <- unique(as.integer(.expanded$id))
      for (.i in .ids) {
        .newEnv$chunks[[.i]] <- .expanded[.expanded$id == .i, , drop = FALSE]
      }
    }
  }
  .newEnv$units      <- .env$units
  .newEnv$show       <- .env$show
  .newEnv$ids        <- .env$ids
  .etResetCountsFromGroups(.newEnv) # nolint
  if (length(.newEnv$groups) == 0L) {
    .newEnv$nobs       <- sum(.expanded$evid == 0L)
    .newEnv$ndose      <- sum(.expanded$evid != 0L)
  }
  if (length(.newEnv$groups) > 0L) {
    .newEnv$show["addl"] <- any(vapply(.newEnv$groups, function(.g) {
      !is.null(.g$data$addl) && any(.g$data$addl != 0L, na.rm = TRUE)
    }, logical(1)))
  } else {
    .newEnv$show["addl"] <- !is.null(.expanded$addl) && any(.expanded$addl != 0L, na.rm = TRUE)
  }
  .newEnv$randomType <- NA_integer_
  .newEnv$canResize  <- FALSE
  .newEnv$extraCols  <- .etExtraCols(.env) # nolint
  structure(c(list(.env = .newEnv),
              .etBuildMethods(.newEnv)), # nolint
            class = "rxEt")
}

# Event-table display helper classes -------------------------------------

#' EVID formatting for tibble and other places.
#'
#' This is to make an EVID more readable by non
#' pharmacometricians. It displays what each means and allows it to
#' be displayed in a tibble.
#'
#' @param x Item to be converted to a rxode2 EVID specification.
#'
#' @param ... Other parameters
#'
#' @return rxEvid specification
#'
#' @examples
#'
#' rxEvid(1:7)
#' @export
rxEvid <- function(x) {
  structure(x, class = "rxEvid")
}

#' @rdname rxEvid
#' @export
as.rxEvid <- rxEvid

#' @rdname rxEvid
#' @export
c.rxEvid <- function(x, ...) {
  as.rxEvid(NextMethod())
}

#' @rdname rxEvid
#' @export
`[.rxEvid` <- function(x, ...) {
  as.rxEvid(NextMethod()) # nolint
}

.colorFmt.rxEvid <- function(x, ...) { # nolint
  .x <- unclass(x)
  if (is.numeric(.x)) {
    .x <-
      data.table::fcase(
        .x == 0, paste0(crayon::blue$bold("0"), ":", crayon::white("Observation")),
        .x == 1, paste0(crayon::blue$bold("1"), ":", crayon::yellow("Dose (Add)")),
        .x == 2, paste0(crayon::blue$bold("2"), ":", crayon::yellow("Other")),
        .x == 3, paste0(crayon::blue$bold("3"), ":", crayon::red("Reset")),
        .x == 4, paste0(crayon::blue$bold("4"), ":", crayon::red("Reset"), "&", crayon::yellow("Dose")),
        .x == 5, paste0(crayon::blue$bold("5"), ":", crayon::red("Replace")),
        .x == 6, paste0(crayon::blue$bold("6"), ":", crayon::yellow("Multiply")),
        .x == 7, paste0(crayon::blue$bold("7"), ":", crayon::yellow("Transit")),
        default=paste0(crayon::blue$red(.x), ":", crayon::red("Invalid"))
      )
  } else {
    .x <- paste0(crayon::blue$red(.x), ":", crayon::red("Invalid"))
  }
  format(.x, justify = "left")
}

#' @rdname rxEvid
#' @export
as.character.rxEvid <- function(x, ...) {
  .x <- unclass(x)
  if (is.numeric(.x)) {
    .x <-
      data.table::fcase(
        .x == 0, "0:Observation",
        .x == 1, "1:Dose (Add)",
        .x == 2, "2:Other",
        .x == 3, "3:Reset",
        .x == 4, "4:Reset&Dose",
        .x == 5, "5:Replace",
        .x == 6, "6:Multiply",
        .x == 7, "7:Transit",
        default = paste0(.x, ":Invalid")
      )
  } else {
    .x <- paste0(.x, ":Invalid")
  }
  .x
}



#' @rdname rxEvid
#' @export
`[[.rxEvid` <- function(x, ...) {
  as.rxEvid(NextMethod())
}

#' @rdname rxEvid
#' @param value It will be an error to set units for evid
#' @export
`units<-.rxEvid` <- function(x, value) {
  stop("'evid' is unitless", call. = FALSE)
}

#' @export
`[<-.rxEvid` <- function(x, i, value) {
  as.rxEvid(NextMethod())
}

# registered in .onLoad()
type_sum.rxEvid <- function(x) {
  "evid"
}

# registered in .onLoad()
pillar_shaft.rxEvid <- function(x, ...) {
  .x <- .colorFmt.rxEvid(x)
  pillar::new_pillar_shaft_simple(.x)
}

#' @export
as.data.frame.rxEvid <- base::as.data.frame.difftime

#' Creates a rxRateDur object
#'
#' This is primarily to display information about rate
#'
#' @param x rxRateDur data
#' @param ... Other parameters
#'
#' @return rxRateDur object
#'
#' @export
rxRateDur <- function(x) {
  structure(x, class = "rxRateDur")
}

#' @rdname rxRateDur
#' @export
`[.rxRateDur` <- function(x, ...) {
  as.rxRateDur(NextMethod())
}

#' @rdname rxRateDur
#' @export
as.rxRateDur <- rxRateDur
#' @rdname rxEvid
#' @export
c.rxRateDur <- function(x, ...) {
  as.rxRateDur(NextMethod())
}

#' @rdname rxRateDur
#' @export
as.character.rxRateDur <- function(x, ...) {
  .x <- unclass(x)
  .x <-
    ifelse(.x == -1, "-1:rate",
      ifelse(.x == -2, "-2:dur",
        ifelse(.x < 0, paste0(as.character(.x), ":Invalid"),
          sprintf(" %-8g", .x)
        )
      )
    )
  .x
}

.fmt <- function(x, width = 9) {
  .g <- sprintf(paste0(" %-", width - 1, "g"), unclass(x))
  .f <- sprintf(paste0(" %-", width - 1, "f"), unclass(x))
  .ncg <- nchar(.g)
  .ncf <- nchar(.f)
  .ret <- ifelse(.ncg == width, .g,
    ifelse(.ncf == width, .f, .g)
  )
  .ret
}


.colorFmt.rxRateDur <- function(x, ...) { # nolint
  .x <- unclass(x)
  .x <-
    ifelse(.x == -1, paste0(crayon::red("-1"), ":", crayon::yellow("rate")),
      ifelse(.x == -2, paste0(crayon::red("-2"), ":", crayon::yellow("dur")),
        ifelse(.x < 0, paste0(crayon::red(as.character(.x)), ":", crayon::red("Invalid")),
          .fmt(.x)
        )
      )
    )
  .x
}

#' @rdname rxRateDur
#' @export
`[[.rxRateDur` <- function(x, ...) {
  as.rxRateDur(NextMethod())
}

#' @export
`[<-.rxRateDur` <- function(x, i, value) {
  as.rxRateDur(NextMethod())
}

# registered in .onLoad()
type_sum.rxRateDur <- function(x) {
  .unit <- attr(x, "units")
  if (!is.null(.unit)) {
    .tmp <- x
    class(.tmp) <- "units"
    pillar::type_sum(.tmp)
  } else {
    "rate/dur"
  }
}

# registered in .onLoad()
pillar_shaft.rxRateDur <- function(x, ...) {
  .x <- .colorFmt.rxRateDur(x)
  pillar::new_pillar_shaft_simple(.x, align = "left", width = 10)
}

#' @export
as.data.frame.rxRateDur <- base::as.data.frame.difftime

set_units.rxRateDur <- function(x, value, ...,
                                mode = .setUnitsMode() # nolint
                                ) {
  if (is.null(mode)) {
    stop("requires package 'units'", call. = FALSE)
  }
  if (inherits(x, "units")) {
    .ret <- x
    .ret0 <- unclass(x)
    .w1 <- which(.ret0 == -1)
    .w2 <- which(.ret0 == -2)
    .lst <- as.list(match.call())[-1]
    class(.ret0) <- "units"
    .lst[[1]] <- .ret0
    .ret <- do.call(units::set_units, .lst)
    if (length(.w1) > 0) .ret[.w1] <- -1
    if (length(.w2) > 0) .ret[.w2] <- -2
    class(.ret) <- c("rxRateDur", "units")
    .ret
  } else {
    .lst <- as.list(match.call())[-1]
    .lst[[1]] <- unclass(x)
    .ret <- do.call(units::set_units, .lst)
    class(.ret) <- c("rxRateDur", "units")
    .ret
  }
}

rxTest({
  test_that("mixing omega and sigma with parameter data frame; RxODE#375", {

    lognCv <- function(x) {
      log((x / 100)^2 + 1)
    }

    mod2 <- rxode2({
      ## the order of variables do not matter, the type of compartmental
      ## model is determined by the parameters specified.
      CL ~ TCL * exp(eta.Cl) * (WT / 70)^0.75
      C2 ~ linCmt(KA, CL, V2, Q, V3)
      eff(0) <- 1 ## This specifies that the effect compartment starts at 1.
      d / dt(eff) ~ Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      ##
      resp <- eff + err1
      pk <- C2 * exp(err2)
    })


    ev <- eventTable(amount.units = "mg", time.units = "hours") |>
      add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12, dosing.to = 2) |>
      add.dosing(dose = 20000, nbr.doses = 5, start.time = 120, dosing.interval = 24, dosing.to = 2) |>
      add.sampling(0:240)

    ## Add Residual differences
    sigma <- diag(2) * 0.05
    dimnames(sigma) <- list(c("err1", "err2"), c("err1", "err2"))

    omega <- matrix(0.2, dimnames = list("eta.Cl", "eta.Cl"))

    theta <- c(
      KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
      Kin = 1, Kout = 1, EC50 = 200
    )

    thetaMat <- diag(length(theta)) * lognCv(5)
    dimnames(thetaMat) <- list(names(theta), names(theta))

    nStud <- 3

    nSub <- 12

    par <- rxRmvn(nStud, theta, thetaMat)

    par <- rxCbindStudyIndividual(par, data.frame(WT = rnorm(nStud * nSub, 70, 10)))

    expect_error(rxSolve(mod2, ev, par,
                         omega = omega, sigma = sigma, dfSub = 100, dfObs = 400,
                         nStud = nStud, nSub = nSub), NA)

    # Nesting:
    ## mod <- rxode2({
    ##   ## Clearance with individuals
    ##   eff(0) = 1
    ##   C2 = centr/V2*(1+prop.sd);
    ##   C3 = peri/V3;
    ##   CL =  TCl*exp(eta.Cl + eye.Cl + iov.Cl + inv.Cl) * (WT / 70)^0.75
    ##   KA = TKA * exp(eta.Ka + eye.Ka + iov.Cl + inv.Ka)
    ##   d/dt(depot) =-KA*depot;
    ##   d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    ##   d/dt(peri)  =                    Q*C2 - Q*C3;
    ##   d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    ##   ef0 = eff + add.sd
    ## })


    ## et(amountUnits="mg", timeUnits="hours") |>
    ##   et(amt=10000, addl=9,ii=12,cmt="depot") |>
    ##   et(time=120, amt=2000, addl=4, ii=14, cmt="depot") |>
    ##   et(seq(0, 240, by=4)) |> # Assumes sampling when there is no dosing information
    ##   et(seq(0, 240, by=4) + 0.1) |> ## adds 0.1 for separate eye
    ##   et(id=1:20) |>
    ##   ## Add an occasion per dose
    ##   dplyr::mutate(occ=cumsum(!is.na(amt))) |>
    ##   dplyr::mutate(occ=ifelse(occ == 0, 1, occ)) |>
    ##   dplyr::mutate(occ=2- occ %% 2) |>
    ##   dplyr::mutate(eye=ifelse(round(time) == time, 1, 2)) |>
    ##   dplyr::mutate(inv=ifelse(id < 10, 1, 2)) |>
    ##   dplyr::as_tibble() ->
    ##   ev


    ## theta <- c("TKA"=0.294, "TCl"=18.6, "V2"=40.2,
    ##            "Q"=10.5, "V3"=297, "Kin"=1, "Kout"=1, "EC50"=200)

    ## ## Creating covariance matrix
    ## tmp <- matrix(rnorm(8^2), 8, 8)
    ## tMat <- tcrossprod(tmp, tmp) / (8 ^ 2)
    ## dimnames(tMat) <- list(names(theta), names(theta))

    ## tMat

    ## nStud <- 4

    ## nSub <- 20

    ## par <- rxCbindStudyIndividual(rxRmvn(nStud, theta, tMat),
    ##                               data.frame(WT=rnorm(nStud * nSub, 70, 10)))

    ## omega <- lotri(lotri(eta.Cl ~ 0.1,
    ##                  eta.Ka ~ 0.1) | id(nu=100),
    ##            lotri(eye.Cl ~ 0.05,
    ##                  eye.Ka ~ 0.05) | eye(nu=200),
    ##            lotri(iov.Cl ~ 0.01,
    ##                  iov.Ka ~ 0.01) | occ(nu=200),
    ##            lotri(inv.Cl ~ 0.02,
    ##                  inv.Ka ~ 0.02) | inv(nu=10))


    ## sigma <- lotri(prop.sd ~ .25,
    ##            add.sd~ 0.125)

    ## s <- rxSolve(mod, par, ev, omega=omega,
    ##              sigma=sigma, sigmaDf=400,
    ##              nStud=nStud)
  })
})

rxTest({
  test_that("seq test for rxSolve", {

    m1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      ## Added modeled bioavaiblity, duration and rate
      fdepot <- 1
      durDepot <- 8
      rateDepot <- 1250
      C2 <- centr / V2
      C3 <- peri / V3
      d / dt(depot) <- -KA * depot
      f(depot) <- fdepot
      dur(depot) <- durDepot
      rate(depot) <- rateDepot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- 1
    })

    ev <- eventTable(amount.units = "mg", time.units = "hr")

    ev$add.dosing(dose = 10000, nbr.doses = 3) # loading doses

    ev <- ev |>
      add.dosing(dose = 5000, nbr.doses = 14, dosing.interval = 12) # maintenance


      t1 <- rxSolve(m1, ev, length.out = 10)
      expect_equal(length(t1$time), 10)

      t1 <- rxSolve(m1, ev, from = 1, to = 10, length.out = 10)
      expect_equal(as.integer(t1$time), 1:10)

      t1 <- rxSolve(m1, ev, from = 1, to = 10, by = 0.5)
      expect_equal(as.numeric(t1$time), seq(1, 10, by = 0.5))

      expect_error(rxSolve(m1, ev, from = 1:2, to = 10, by = 0.5), "'from'")
      expect_error(rxSolve(m1, ev, from = 1, to = 10:11, by = 0.5), "'to'")
      expect_error(rxSolve(m1, ev, from = 1, to = 10, by = c(0.5, 1)), "'by'")
      expect_error(rxSolve(m1, ev, from = 1, to = 10, length.out = 10:11), "'length.out'")
  })

  test_that("seq test for et", {
    expect_error(et(1, 2, 3))
    expect_equal(et(1, 10)$time, as.numeric(1:10))
    expect_equal(et(1)$time, seq(1))
    expect_equal(et(1, length.out = 4)$time, seq(1, length.out = 4))
    expect_equal(et(1, by = 0.1)$time, seq(1, length.out = 0.1))
    expect_equal(et(1, 10, by = 0.1)$time, seq(1, 10, by = 0.1))
    expect_equal(et(1, 10, length.out = 7)$time, seq(1, 10, length.out = 7))

    expect_error(et(0.5) |> et(1, 2, 3))
    expect_equal((et(0.5) |> et(1, 10))$time, c(0.5, seq(1, 10)))
    expect_equal((et(0.5) |> et(1))$time, c(0.5, seq(1)))
    expect_equal((et(0.5) |> et(1, length.out = 4))$time, c(0.5, seq(1, length.out = 4)))
    expect_equal((et(0.5) |> et(1, by = 0.1))$time, c(0.5, seq(1, by = 0.1)))
    expect_equal((et(0.5) |> et(1, 10, by = 0.1))$time, c(0.5, seq(1, 10, by = 0.1)))
    expect_equal((et(0.5) |> et(1, 10, length.out = 7))$time, c(0.5, seq(1, 10, length.out = 7)))
  })
})

rxTest({
  library(withr)

  test_that("ev$id has data-frame semantics (#1154)", {
    ev <- et(amt = 100, cmt = "depot") |> et(seq(0, 24, by = 1)) |> et(id = 1:3)
    .df <- as.data.frame(ev, all = TRUE)
    # per-row id column, matching as.data.frame(ev)$id
    expect_equal(ev$id, .df$id)
    expect_equal(length(ev$id), nrow(.df))
    # unique ids remain available via the env
    expect_equal(ev$env$ids, 1:3)
    # idiomatic subset returns only subject 3, doses included
    .sub <- as.data.frame(ev[ev$id == 3, ])
    expect_equal(unique(.sub$id), 3L)
    expect_equal(nrow(.sub), sum(.df$id == 3L))
    expect_equal(sum(.sub$evid != 0L), 1L)
    # a logical row index that cannot recycle cleanly errors
    expect_error(ev[c(TRUE, FALSE, TRUE), ], "logical row index")
    expect_error(ev[c(TRUE, FALSE, TRUE), "time"], "logical row index")
    # length-0 and length-1 logical indices keep data-frame semantics
    expect_equal(nrow(as.data.frame(ev[logical(0), ])), 0L)
    expect_equal(nrow(as.data.frame(ev[TRUE, ], all = TRUE)), nrow(.df))
    # reading ev$id does not consume the RNG stream (windowed addl doses
    # draw times on materialization)
    .evw <- et(data.frame(id = 1L, time = 0, low = 0, high = 1, cmt = "depot",
                          amt = 100, ii = 24, addl = 1L, evid = 1L))
    .a <- withr::with_seed(42, as.data.frame(.evw, all = TRUE))
    .b <- withr::with_seed(42, {
      invisible(.evw$id)
      as.data.frame(.evw, all = TRUE)
    })
    expect_equal(.a$time, .b$time)
    expect_equal(.evw$id, .a$id)
    # per-subject covariate assignment round-trips through as.data.frame
    ev$wt <- 50 + 20 * ev$id
    .df2 <- as.data.frame(ev)
    expect_true("wt" %in% names(.df2))
    expect_equal(as.numeric(tapply(.df2$wt, .df2$id, unique)), c(70, 90, 110))
    # assigning an unshown canonical column round-trips too
    .evc <- et(0:5)
    .evc$cmt <- "depot"
    expect_true("cmt" %in% names(as.data.frame(.evc)))
    expect_equal(unique(as.data.frame(.evc)$cmt), "depot")
  })

  test_that("covariates riding along on an imported data frame stay hidden", {
    # a covariate that came in with the imported data frame is not shown by
    # default; downstream packages join it back on and would otherwise get
    # duplicated `wt.x`/`wt.y` columns
    .ev <- et(data.frame(id = 1, time = 0, evid = 1, cmt = 1, amt = 10, wt = 70))
    expect_false("wt" %in% names(as.data.frame(.ev)))
    expect_true("wt" %in% names(as.data.frame(.ev, all = TRUE)))
    # the hidden covariate is still used when solving
    .ev <- add.sampling(.ev, 0:3)
    expect_false("wt" %in% names(as.data.frame(.ev)))
    .m <- rxode2("v = 10 * (wt / 70)\nd/dt(ac) = -0.1 * ac\ncc = ac / v")
    expect_equal(suppressWarnings(unique(rxSolve(.m, .ev)$v)), 10)
    # an explicitly assigned covariate is still shown, and survives copies,
    # row subsets and rbind
    .ev$wt2 <- 80
    expect_true("wt2" %in% names(as.data.frame(.ev)))
    expect_true("wt2" %in% names(as.data.frame(.ev$copy())))
    expect_true("wt2" %in% names(as.data.frame(.ev[1:2, ])))
    expect_true("wt2" %in% names(as.data.frame(etRbind(.ev, .ev))))
    expect_false("wt" %in% names(as.data.frame(etRbind(.ev, .ev))))
    # ... and through the vctrs rebuild path (vec_slice/vec_restore/vec_c)
    expect_true("wt2" %in% names(as.data.frame(vctrs::vec_slice(.ev, 1))))
    expect_false("wt" %in% names(as.data.frame(vctrs::vec_slice(.ev, 1))))
    expect_true("wt2" %in% names(as.data.frame(vctrs::vec_c(.ev, .ev))))
    # deleting the column forgets it rather than recording it
    .ev$wt2 <- NULL
    expect_false("wt2" %in% names(as.data.frame(.ev)))
    expect_false("wt2" %in% .rxEtEnv(.ev)$extraCols)
    .evn <- et(0:5)
    .evn$never <- NULL
    expect_equal(.rxEtEnv(.evn)$extraCols, character(0))
  })

  test_that("explicitly assigned columns survive etExpand/etSeq/simulate", {
    # these three rebuild the event table environment from scratch, so each
    # has to carry `extraCols` over or the assigned column silently becomes a
    # hidden imported covariate (#1154)
    .eve <- et(amt = 3, ii = 24, addl = 2) |> et(0:5) |> et(id = 1:3)
    .eve$wt <- 70
    expect_true("wt" %in% names(as.data.frame(etExpand(.eve))))
    expect_equal(.rxEtEnv(etExpand(.eve))$extraCols, "wt")

    .evq <- suppressWarnings(etSeq(.eve, .eve))
    expect_true("wt" %in% names(as.data.frame(.evq)))
    expect_equal(.rxEtEnv(.evq)$extraCols, "wt")

    .evs <- et(amt = 3) |> et(0:5)
    .evs$wt <- 70
    .evsim <- suppressWarnings(simulate(.evs))
    expect_true("wt" %in% names(as.data.frame(.evsim)))
    expect_equal(.rxEtEnv(.evsim)$extraCols, "wt")

    # a covariate that merely rode along is still not resurrected by any of them
    .evi <- et(data.frame(id = 1, time = 0, evid = 1, cmt = 1, amt = 10, wt = 70))
    .evi <- add.sampling(.evi, 0:3)
    expect_false("wt" %in% names(as.data.frame(etExpand(.evi))))
    expect_false("wt" %in% names(as.data.frame(suppressWarnings(simulate(.evi)))))
  })

  test_that("dplyr verbs degrade an event table preview to a plain data frame", {
    skip_if_not_installed("dplyr")
    # the preview marking describes the frame the accessor returned; once a
    # dplyr verb rewrites it, it is the caller's own data frame and must print
    # like one instead of hiding columns it no longer describes
    .ev <- et(amt = 3) |> et(0:5)
    .ev$wt <- 70
    .p <- .ev$get.dosing()
    expect_s3_class(.p, "rxEtPreview")
    for (.f in list(function(x) x[c("time", "amt")],
                    function(x) dplyr::select(x, "time", "amt"),
                    function(x) dplyr::relocate(x, "amt"),
                    function(x) dplyr::mutate(x, z = 1),
                    function(x) dplyr::filter(x, TRUE))) {
      .out <- .f(.p)
      expect_false(inherits(.out, "rxEtPreview"))
      expect_s3_class(.out, "data.frame")
    }
    # a bare rename keeps the marking on purpose -- `.etKeepCols()` hides only
    # the names that were marked, so the renamed column keeps printing
    expect_s3_class(dplyr::rename(.p, tm = "time"), "rxEtPreview")
    expect_true(any(grepl("\\btm\\b",
                          utils::capture.output(print(dplyr::rename(.p, tm = "time"))))))
    # a row subset keeps every column, so the marking still describes it and
    # the hidden columns stay hidden (this is what the `[` method must NOT undo)
    for (.r in list(.p[1, ], utils::head(.p), .p[0, ], .p[.p$amt > 0, ])) {
      expect_s3_class(.r, "rxEtPreview")
      expect_false(any(grepl("\\brate\\b", utils::capture.output(print(.r)))))
    }
    expect_false(inherits(dplyr::slice(.p, 1), "rxEtPreview"))
    # selecting every column by name is still a column subset: `[.data.frame`
    # drops the marker attributes there, so the class must not be left behind
    # pointing at nothing
    for (.a in list(.p[TRUE], .p[names(.p)], .p[, names(.p)])) {
      expect_false(inherits(.a, "rxEtPreview"))
      expect_null(attr(.a, "rxEtShow", exact = TRUE))
    }
    # `[` keeps base semantics: drop=, matrix indexing and row filters
    expect_equal(.p[, "amt"], as.data.frame(.p)[["amt"]])
    expect_false(inherits(.p[, "amt", drop = FALSE], "rxEtPreview"))
    expect_equal(ncol(.p[, "amt", drop = FALSE]), 1L)
    expect_equal(nrow(.p[.p$amt > 0, ]), sum(.p$amt > 0))
    expect_equal(.p[1, 2], as.data.frame(.p)[1, 2])
    # a compressed preview is the exception: its groups count rows and print()
    # builds the `id` column and the "N individuals" header from them, so a row
    # subset must give up the marking rather than print a header that lies
    .evg <- et(amt = 3) |> et(0:5)
    .evg$wt <- 70
    .g <- (.evg |> et(id = 1:4))$get.sampling()
    expect_s3_class(.g, "rxEtPreview")
    # the whole preview prints the synthesized `id` under a group header
    expect_false("id" %in% names(.g))
    expect_true(any(grepl("\\bid\\b", utils::capture.output(print(.g)))))
    expect_true(any(grepl("compressed preview", utils::capture.output(print(.g)))))
    expect_false(inherits(.g[1:2, ], "rxEtPreview"))
    expect_false(any(grepl("compressed preview",
                           utils::capture.output(print(.g[1:2, ])))))
    # every row still present keeps it, since the groups still describe it
    expect_s3_class(.g[seq_len(nrow(.g)), ], "rxEtPreview")
    # the untouched preview still hides the columns it was marked to hide,
    # but every column is there for programmatic access
    expect_false(any(grepl("\\brate\\b", utils::capture.output(print(.p)))))
    expect_true("rate" %in% names(.p))
    # once columns are picked it is a plain data frame that prints all of them,
    # including one that was hidden while it was marked
    .k <- .p[c("time", "rate")]
    expect_false(inherits(.k, "rxEtPreview"))
    expect_true(any(grepl("\\brate\\b", utils::capture.output(print(.k)))))
  })

  test_that("explicitly assigned columns show up in print(ev) (#1154)", {
    withr::local_options(list(width = 120))
    .hasWt <- function(x) any(grepl("\\bwt\\b", utils::capture.output(print(x))))
    .ev <- et(amt = 3, ii = 24, until = 96) |> et(seq(0, 96, by = 12))
    expect_false(.hasWt(.ev))
    .ev$wt <- 70
    # the assigned column is part of the printed tibble, not swallowed
    expect_true(.hasWt(.ev))
    expect_true(any(grepl("\\b70\\b", utils::capture.output(print(.ev)))))
    # and it is part of the accessor the printout points at
    expect_true("wt" %in% names(.ev$get.EventTable()))
    # deleting it takes it back out of the printout
    .ev$wt <- NULL
    expect_false(.hasWt(.ev))
    expect_false("wt" %in% names(.ev$get.EventTable()))
    # the compressed multi-subject preview shows it too
    .evi <- et(amt = 3) |> et(0:5)
    .evi$wt <- 70
    .evi <- .evi |> et(id = 1:4)
    .prev <- .etPreviewData(.rxEtEnv(.evi), "all")
    expect_s3_class(.prev, "rxEtPreview")
    expect_equal(attr(.prev, "rxEtExtraCols"), "wt")
    expect_true(.hasWt(.evi))
    expect_true(.hasWt(.prev))
    expect_true(.hasWt(.evi$get.dosing()))
    expect_true(.hasWt(.evi$get.sampling()))
    # a covariate that only rode along with an imported data frame stays
    # hidden even once another column is assigned explicitly
    .evd <- et(data.frame(time = c(0, 1), amt = c(100, NA), evid = c(1, 0),
                          covar = c(50, 50)))
    .evd$wt <- 70
    .hasCovar <- function(x) any(grepl("\\bcovar\\b", utils::capture.output(print(x))))
    expect_true(.hasWt(.evd))
    expect_false(.hasCovar(.evd))
    expect_true("wt" %in% names(.evd$get.EventTable()))
    expect_false("covar" %in% names(.evd$get.EventTable()))
    # an unshown canonical column stays hidden as well
    expect_false(any(grepl("\\bdur\\b", utils::capture.output(print(.evd)))))
    # a canonical column that is not shown by default is unaffected
    expect_equal(.etDisplayCols(c("time", "amt", "wt"),
                                c(time = TRUE, amt = TRUE, cmt = FALSE),
                                "wt"),
                 c("time", "amt", "wt"))
    # `pre` columns lead, and columns absent from the data are dropped
    expect_equal(.etDisplayCols(c("id", "time", "wt"),
                                c(time = TRUE, amt = TRUE),
                                c("wt", "gone"), pre = "id"),
                 c("id", "time", "wt"))
    # an extra column that is also canonical is not duplicated
    expect_equal(.etDisplayCols(c("time", "cmt"),
                                c(time = TRUE, cmt = TRUE),
                                "cmt"),
                 c("time", "cmt"))
  })

  test_that("get.dosing()/get.sampling() print display columns only (#1154)", {
    withr::local_options(list(width = 120))
    .out <- function(x) utils::capture.output(print(x))
    .has <- function(x, nm) any(grepl(paste0("\\b", nm, "\\b"), .out(x)))
    .evd <- et(data.frame(time = c(0, 1, 2), amt = c(100, NA, NA),
                          evid = c(1, 0, 0), covar = c(50, 50, 50)))
    .evd$wt <- 70
    for (.f in list(.evd$get.dosing(), .evd$get.sampling())) {
      # the un-grouped accessor now prints the same columns print(ev) does
      expect_s3_class(.f, "rxEtPreview")
      expect_true(.has(.f, "wt"))
      expect_true(.has(.f, "time"))
      expect_false(.has(.f, "covar"))
      expect_false(.has(.f, "dur"))
      expect_false(.has(.f, "high"))
      # but every column is still there for programmatic access
      expect_true(all(c("covar", "dur", "high", "wt") %in% names(.f)))
      expect_equal(.f$covar, rep(50, nrow(.f)))
    }
    # no compressed-preview header for an un-grouped table
    expect_false(any(grepl("compressed preview", .out(.evd$get.dosing()))))
    # a column renamed afterwards is not silently hidden
    .r <- .evd$get.dosing()
    names(.r)[names(.r) == "time"] <- "Time"
    expect_true(.has(.r, "Time"))
    expect_false(.has(.r, "covar"))
    # keeping only hidden columns prints them rather than nothing
    .k <- .evd$get.dosing()[, c("low", "high"), drop = FALSE]
    expect_true(.has(.k, "low"))
    # assigning into the frame makes it the caller's own data frame, so a
    # column reusing a hidden name is not silently suppressed
    .a <- .evd$get.dosing()
    .a$covar <- NULL
    .a$covar <- "new"
    expect_false(inherits(.a, "rxEtPreview"))
    expect_true(.has(.a, "covar"))
    .b <- .evd$get.dosing()
    .b[["covar"]] <- 1
    expect_false(inherits(.b, "rxEtPreview"))
    .c <- .evd$get.dosing()
    .c[1, "amt"] <- 5
    expect_false(inherits(.c, "rxEtPreview"))
    # duplicated column names are not collapsed away by the column filter
    expect_equal(.etKeepCols(c("time", "amt", "amt", "low"),
                             c("time", "amt", "amt", "low"),
                             c(time = TRUE, amt = TRUE, low = FALSE)),
                 c("time", "amt", "amt"))
    # nothing left to show falls back to showing everything
    expect_equal(.etKeepCols(c("low", "high"), c("low", "high"),
                             c(time = TRUE, low = FALSE, high = FALSE)),
                 c("low", "high"))
    # no marked columns recorded behaves like filtering on what is there
    expect_equal(.etKeepCols(c("time", "low"), NULL,
                             c(time = TRUE, low = FALSE)),
                 "time")
    # units and row names survive the display marking
    skip_if_not_installed("units")
    .evu <- et(timeUnits = "hr") |>
      et(amt = 100, ii = 12, until = 24) |>
      et(seq(0, 24, by = 6))
    withr::with_options(list(rxode2.homogenous = FALSE), {
      .d <- .evu$get.dosing()
      expect_true(inherits(.d$time, "units"))
      expect_equal(rownames(.d), "1")
      expect_equal(as.numeric(.d$ii), 12)
      expect_true(.has(.d, "addl"))
      expect_false(.has(.d, "low"))
    })
  })

  test_that("assigned columns survive a data frame round trip (#1154)", {
    withr::local_options(list(width = 120))
    .has <- function(x, nm) {
      any(grepl(paste0("\\b", nm, "\\b"), utils::capture.output(print(x))))
    }
    .ev <- et(data.frame(time = c(0, 1, 2), amt = c(100, NA, NA),
                         evid = c(1, 0, 0), covar = c(50, 50, 50)))
    .ev$wt <- 70
    .df <- as.data.frame(.ev)
    expect_equal(attr(.df, "rxEtExtraCols"), "wt")
    # et(), $import.EventTable(), $importEventTable() and as.et() all keep the
    # assigned column visible without promoting the imported covariate
    .rt <- et(.df)
    expect_true(.has(.rt, "wt"))
    expect_false(.has(.rt, "covar"))
    expect_equal(.etExtraCols(.rxEtEnv(.rt)), "wt")
    expect_true(.has(et(as.data.frame(.rt)), "wt")) # and again, repeatedly
    expect_true(.has(as.et(.df), "wt"))
    .i1 <- et()
    .i1$import.EventTable(.df)
    expect_true(.has(.i1, "wt"))
    expect_false(.has(.i1, "covar"))
    .i2 <- et()
    .i2$importEventTable(.df)
    expect_true(.has(.i2, "wt"))
    expect_false(.has(.i2, "covar"))
    # all = TRUE carries the hidden covariate as data but not as a shown column
    .all <- et(as.data.frame(.ev, all = TRUE))
    expect_true(.has(.all, "wt"))
    expect_false(.has(.all, "covar"))
    expect_true("covar" %in% names(as.data.frame(.all, all = TRUE)))
    # a NONMEM-style uppercase name is renamed on import, and the tag is
    # renamed with it so the column does not vanish (it shows before the trip)
    .up <- et(time = 1, amt = 100)
    .up$SS <- 3
    expect_true(.has(.up, "SS"))
    .upRt <- et(as.data.frame(.up))
    expect_equal(.etExtraCols(.rxEtEnv(.upRt)), "ss")
    expect_true(.has(.upRt, "ss"))
    # a tag naming a column that is not in the frame is ignored
    .bad <- as.data.frame(.ev)
    attr(.bad, "rxEtExtraCols") <- c("wt", "nosuchcolumn")
    expect_equal(.etExtraCols(.rxEtEnv(et(.bad))), "wt")
    # a data frame built by hand carries no tag, so its covariates stay hidden
    .plain <- et(data.frame(time = c(0, 1), amt = c(1, NA), evid = c(1, 0),
                            cv = c(2, 2)))
    expect_false(.has(.plain, "cv"))
    expect_null(attr(as.data.frame(.plain), "rxEtExtraCols"))
    # dropping the column drops the tag, so it does not come back
    .ev$wt <- NULL
    expect_null(attr(as.data.frame(.ev), "rxEtExtraCols"))
    expect_false(.has(et(as.data.frame(.ev)), "wt"))
  })

  test_that("dplyr verbs drop the get.dosing() display marking", {
    skip_if_not_installed("dplyr")
    withr::local_options(list(width = 120))
    .out <- function(x) utils::capture.output(print(x))
    .evd <- et(data.frame(time = c(0, 1, 2), amt = c(100, NA, NA),
                          evid = c(1, 0, 0), covar = c(50, 50, 50)))
    .evd$wt <- 70
    # dplyr keeps the class but drops the marker attributes, so degrade to a
    # plain data frame the way `rxEt` does instead of printing a stale subset
    .m <- dplyr::mutate(.evd$get.dosing(), wt2 = 1)
    expect_false(inherits(.m, "rxEtPreview"))
    expect_true(any(grepl("\\bwt2\\b", .out(.m))))
    .s <- dplyr::filter(.evd$get.sampling(), .data$time > 1)
    expect_false(inherits(.s, "rxEtPreview"))
    expect_equal(nrow(.s), 1L)
  })

  test_that("et import rate=-2", {

    d <- data.frame(id = c(1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1),
                    time = c(0, 2, 12.5, 24.5, 37, 48, 60.5, 72.5, 85.3, 96.5, 108.5, 112.5),
                    amt = c(25, 0, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 0),
                    rate = c(-2, 0, -2, -2, -2, -2, -2, -2, -2, -2, -2, 0))
    d2 <- as.data.frame(et(d))
    expect_equal(d2$rate, c(-2, NA_real_, -2, -2, -2, -2, -2, -2, -2, -2, -2, NA_real_))


  })


  for (radi in c(1, 2)) {
    forderForceBase(switch(radi,
                           TRUE,
                           FALSE
                           ))
    radix <- switch(radi,
                    "base::order",
                    "data.table::forder"
                    )

    # context(sprintf("Test event Table et(...) sort:%s", radix))
    et <- et()

    test_that("Empty event table check", {
      ## expect_s3_class(et$env,"rxHidden")
      expect_equal(et$nobs, 0L)
      expect_equal(et$ndose, 0L)
      expect_equal(et$get.EventTable(), NULL)
      expect_equal(et$get.dosing(), NULL)
      expect_equal(et$get.sampling(), NULL)
    })

    et <- et(1:10)

    test_that("10 sample items", {
      expect_equal(et$nobs, 10L)
      expect_equal(et$ndose, 0L)
      expect_s3_class(et$get.EventTable(), "data.frame")
      expect_true(is(et, "rxEt"))
      expect_false(et$show["id"])
      expect_false(et$show["cmt"])
      expect_equal(et$get.dosing(), NULL)
      expect_equal(length(et$get.sampling()$time), 10)
    })

    et <- et(1:10, "matt")

    test_that("compartment check", {
      expect_equal(et$nobs, 10L)
      expect_equal(et$ndose, 0L)
      expect_s3_class(et$get.EventTable(), "data.frame")
      expect_true(is(et, "rxEt"))
      expect_false(et$show["id"])
      expect_true(et$show["cmt"])
      expect_equal(et$get.dosing(), NULL)
      expect_equal(length(et$get.sampling()$time), 10)
    })

    et <- et(1:10, cmt = 1)
    test_that("compartment check", {
      expect_equal(et$nobs, 10L)
      expect_equal(et$ndose, 0L)
      expect_s3_class(et$get.EventTable(), "data.frame")
      expect_true(is(et, "rxEt"))
      expect_false(et$show["id"])
      expect_true(et$show["cmt"])
      expect_equal(et$get.dosing(), NULL)
      expect_equal(length(et$get.sampling()$time), 10)
      expect_true(all(et$cmt == 1L))
    })



    et1 <- et(1:10, id = 1:10)

    test_that("Observation only table check", {
      expect_equal(et1$nobs, 100L)
      expect_equal(et1$ndose, 0L)
      expect_s3_class(et1$get.EventTable(), "data.frame")
      expect_true(is(et1, "rxEt"))
      expect_true(et1$show["id"])
      expect_false(et1$show["cmt"])
      expect_equal(et1$get.dosing(), NULL)
      ## homogeneous path returns compressed (10 time-points); non-homogeneous returns 100
      expect_equal(length(et1$get.sampling()$time),
                   if (.rxGetHomogenous()) 10L else 100L)
    })

    test_that("get.sampling homogeneous vs non-homogeneous path", {
      ## default homogeneous: compressed per-group view
      withr::local_options(rxode2.homogenous = TRUE)
      expect_equal(length(et1$get.sampling()$time), 10L)
      ## non-homogeneous: fully materialized across all subjects
      withr::local_options(rxode2.homogenous = FALSE)
      expect_equal(length(et1$get.sampling()$time), 100L)
    })

    ## now resize down
    et2 <- et1 |>
      et(id = -(2:10))

    test_that("compartment check", {
      expect_equal(et2$nobs, 10L)
      expect_equal(et2$ndose, 0L)
      expect_s3_class(et2$get.EventTable(), "data.frame")
      expect_true(is(et, "rxEt"))
      expect_true(et2$show["id"])
      expect_false(et2$show["cmt"])
      expect_equal(et2$get.dosing(), NULL)
      expect_equal(length(et2$get.sampling()$time), 10)
    })


    ## now resize back up
    et3 <- et2 |>
      et(id = 1:10)

    test_that("Using simulate with et without windows will warn", {
      f1 <- as.data.frame(et3)
      f2 <- suppressWarnings({
        as.data.frame(simulate(et3))
      })
      expect_equal(f1$time, f2$time)
      expect_warning(simulate(et3))
    })

    eti <- et(amt = 10) |>
      et(id = 2:3)
    eti1 <- et(amt = 10, id = 3:4)
    eti2 <- et(amt = 10, id = 3)

    eti3 <- et(10, id = 3:4)
    eti4 <- et(10, id = 3)

    eti5 <- et(10) |>
      et(id = 2:3)

    eti6 <- et(amt = 10) |>
      et(id = 2)

    eti7 <- et(10) |>
      et(id = 2)


    eti8 <- et(10, id = 1)
    eti9 <- et(10) |>
      et(id = 1)

    eti10 <- et(amt = 10, id = 1)
    eti11 <- et(amt = 10) |>
      et(id = 1)

    ## Now look at windows

    eti12 <- et(list(c(10, 11)), id = 1)
    eti13 <- et(list(c(10, 11))) |>
      et(id = 1)

    eti14 <- et(list(c(10, 11)), amt = 10, id = 1)
    eti15 <- et(list(c(10, 11)), amt = 10) |>
      et(id = 1)


    eti16 <- et(list(c(10, 11)), id = 2)
    eti17 <- et(list(c(10, 11))) |>
      et(id = 2)

    eti18 <- et(list(c(10, 11)), amt = 10, id = 2)
    eti19 <- et(list(c(10, 11)), amt = 10) |>
      et(id = 2)

    eti20 <- et(list(c(10, 11)), id = 2:3)
    eti21 <- et(list(c(10, 11))) |>
      et(id = 2:3)

    eti22 <- et(list(c(10, 11)), amt = 10, id = 2:3)
    eti23 <- et(list(c(10, 11)), amt = 10) |>
      et(id = 2:3)

    ## Now do it with dosing/sampling windows

    test_that("Make sure that et only has IDs 2 and 3.", {
      expect_equal(unique(eti$id), 2:3)
      expect_equal(unique(eti1$id), 3:4)
      expect_equal(unique(eti2$id), 3)
      expect_equal(unique(eti3$id), 3:4)
      expect_equal(unique(eti4$id), 3)
      expect_equal(unique(eti5$id), 2:3)
      expect_equal(unique(eti6$id), 2)
      expect_equal(unique(eti7$id), 2)
      expect_equal(unique(eti8$id), 1)
      expect_true(eti8$env$show["id"])
      expect_equal(unique(eti9$id), 1)
      expect_true(eti9$env$show["id"])
      expect_equal(unique(eti10$id), 1)
      expect_true(eti10$env$show["id"])
      expect_equal(unique(eti11$id), 1)
      expect_true(eti11$env$show["id"])
      expect_equal(unique(eti12$id), 1)
      expect_true(eti12$env$show["id"])
      expect_equal(unique(eti13$id), 1)
      expect_true(eti13$env$show["id"])
      expect_equal(unique(eti14$id), 1)
      expect_true(eti14$env$show["id"])
      expect_equal(unique(eti15$id), 1)
      expect_true(eti15$env$show["id"])
      expect_equal(unique(eti16$id), 2)
      expect_true(eti16$env$show["id"])
      expect_equal(unique(eti17$id), 2)
      expect_true(eti17$env$show["id"])
      expect_equal(unique(eti18$id), 2)
      expect_true(eti18$env$show["id"])
      expect_equal(unique(eti19$id), 2)
      expect_true(eti19$env$show["id"])
      expect_equal(unique(eti20$id), 2:3)
      expect_true(eti20$env$show["id"])
      expect_equal(unique(eti21$id), 2:3)
      expect_true(eti21$env$show["id"])
      expect_equal(unique(eti22$id), 2:3)
      expect_true(eti22$env$show["id"])
      expect_equal(unique(eti23$id), 2:3)
      expect_true(eti23$env$show["id"])
    })


    test_that("Observation only table check", {
      expect_equal(et1$nobs, 100L)
      expect_equal(et1$ndose, 0L)
      expect_s3_class(et1$get.EventTable(), "data.frame")
      expect_true(is(et1, "rxEt"))
      expect_true(et1$show["id"])
      expect_false(et1$show["cmt"])
      expect_equal(et1$get.dosing(), NULL)
      ## homogeneous path returns compressed (10 time-points); non-homogeneous returns 100
      expect_equal(length(et1$get.sampling()$time),
                   if (.rxGetHomogenous()) 10L else 100L)
    })

    ## Check adding different units of time, rate, amt work
    test_that("units tests", {
      skip_if_not_installed("units")
      ## Make sure units are right.
      et3 <- et3 |> units::set_units(mg)
      e <- et(amount.units = "mg", time_units = "hr") |>
        add.sampling(seq(0, 24, by = 3)) |>
        add.dosing(1, 3) |>
        et(rate = 3, amt = 2, time = 120)

      e2 <- e |>
        et(amt = units::set_units(0.0003, "lb"), time = 0.5)

      expect_equal(e2$amt[e2$time == units::set_units(0.5, hr)],
                   units::set_units(units::set_units(0.0003, lb), mg))
      e2 <- e |> et(units::set_units(30, min))
      expect_true(any(e2$time == units::set_units(0.5, hr)))
      e2 <- e |> et(time = 0.25, rate = units::set_units(30, ug / min), amt = units::set_units(4, ug))
      tmp <- e2[e2$time == units::set_units(0.25, hr), ]
      expect_equal(units::set_units(1.8, mg / h), tmp$rate)
      expect_equal(units::set_units(0.004, mg), tmp$amt)
      e2 <- e |> et(time = 0.25, ii = units::set_units(30, min), amt = 4, addl = 4)
      expect_equal(e2$ii[e2$time == units::set_units(0.25, hr)], units::set_units(0.5, hr))

      ## Check importing wrong different ii and time units as well as different rate units work.
      e <- et(amount.units = "mg", time_units = "hr") |>
        add.sampling(seq(0, 24, by = 3)) |>
        add.dosing(1, 3) |>
        et(rate = 3, amt = 2, time = 120)

      etDf <- as.data.frame(e)
      etDf$rate <- units::set_units(etDf$rate, ug / s)
      etDf$ii <- units::set_units(etDf$ii, min)

      et <- et()
      et$import.EventTable(etDf)

      expect_equal(as.numeric(et$ii), as.numeric(e$ii))
      expect_equal(as.numeric(et$rate), as.numeric(e$rate))
    })

    test_that("seq works with wait", {

      e1 <- et(amt = 100, ii = 24, addl = 6)

      e2 <- et(amt = 200, ii = 24, addl = 6)

      e3 <- seq(e1, e2, e1)

      e4 <- seq(e1, wait = 72, e2, wait = 72, e1) |> as.data.frame()

      expect_equal(structure(list(
        time = c(0, 216, 432),
        amt = c(100, 200, 100),
        ii = c(24, 24, 24),
        addl = c(6L, 6L, 6L),
        evid = c(1L, 1L, 1L)
      ),
      class = "data.frame",
      row.names = c(NA, -3L)
      ), e4)

      e5 <- etSeq(e1, wait = 72, e2, wait = 72, e1, waitII = "+ii") |>
        as.data.frame()

      expect_equal(structure(list(
        time = c(0, 240, 480),
        amt = c(100, 200, 100),
        ii = c(24, 24, 24),
        addl = c(6L, 6L, 6L),
        evid = c(1L, 1L, 1L)
      ),
      class = "data.frame",
      row.names = c(NA, -3L)
      ), e5)

      e1 <- et(amt = 500)
      e2 <- et(amt = 250, ii = 24, addl = 4)

      expect_equal(
        structure(list(
          time = c(0, 24),
          amt = c(500, 250),
          ii = c(0, 24),
          addl = c(0L, 4L),
          evid = c(1L, 1L)
        ),
        class = "data.frame",
        row.names = c(NA, -2L)
        ),
        c(e1, e2) |> as.data.frame()
      )

      expect_equal(
        structure(list(
          time = c(0, 120, 144),
          amt = c(250, 500, 250),
          ii = c(24, 0, 24),
          addl = c(4L, 0L, 4L),
          evid = c(1L, 1L, 1L)
        ),
        class = "data.frame",
        row.names = c(NA, -3L)
        ),
        c(e2, e1, e2) |> as.data.frame()
      )

      e3 <- et(amt = 200)

      expect_warning(c(e1, e3))

      e4 <- suppressWarnings(c(e1, e3) |> as.data.frame())

      expect_equal(
        structure(list(
          time = c(0, 24),
          amt = c(500, 200),
          ii = c(0, 0),
          addl = c(0L, 0L),
          evid = c(1L, 1L)
        ),
        class = "data.frame",
        row.names = c(NA, -2L)
        ),
        e4
      )

      e4 <- suppressWarnings(c(e1, e3, ii = 12) |> as.data.frame())

      expect_equal(
        structure(list(
          time = c(0, 12),
          amt = c(500, 200),
          ii = c(0, 0),
          addl = c(0L, 0L),
          evid = c(1L, 1L)
        ),
        class = "data.frame",
        row.names = c(NA, -2L)
        ),
        e4
      )


      e1 <- et(amt = 100, ii = 24, addl = 6) |>
        et(seq(0, 2 * 168, by = 0.1))
      e2 <- c(e1, e1, samples = "use")
      expect_equal(range(e2$time), c(0, 672))

      ## combine without changing, use rbind
      e1 <- et(amt = 100, ii = 24, addl = 6, ID = 1:5)
      e2 <- et(amt = 50, ii = 12, addl = 13, ID = 1:3)
      e3 <- et(amt = 200, ii = 24, addl = 2, ID = 1:2)

      e4 <- rbind(e1, e2, e3)
      expect_equal(
        e4 |> dplyr::select(id, time, amt, ii, addl) |> as.data.frame(),
        structure(list(
          id = c(1L, 1L, 1L, 2L, 2L, 2L, 3L, 3L, 4L, 5L),
          time = c(0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
          amt = c(100, 50, 200, 100, 50, 200, 100, 50, 100, 100),
          ii = c(24, 12, 24, 24, 12, 24, 24, 12, 24, 24),
          addl = c(6L, 13L, 2L, 6L, 13L, 2L, 6L, 13L, 6L, 6L)
        ),
        class = "data.frame",
        row.names = c(NA, -10L)
        )
      )

      e4 <- rbind(e1, e2, e3, id = "unique")
      expect_equal(
        e4 |> dplyr::select(id, time, amt, ii, addl) |> as.data.frame(),
        structure(list(
          id = 1:10,
          time = c(0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
          amt = c(100, 100, 100, 100, 100, 50, 50, 50, 200, 200),
          ii = c(24, 24, 24, 24, 24, 12, 12, 12, 24, 24),
          addl = c(6L, 6L, 6L, 6L, 6L, 13L, 13L, 13L, 2L, 2L)
        ),
        class = "data.frame",
        row.names = c(NA, -10L)
        )
      )
    })

    ## FIXME test windows and dur
    test_that("no ii throws an error with addl", {
      expect_error(et(c(11, 13), amt = 10, addl = 3))
    })
    ## Test Windows
    et <- et(list(
      c(0, 1),
      c(1, 4),
      c(4, 8),
      c(8, 12),
      c(12, 24)
    )) |>
      et(amt = 10) |>
      et(c(11, 13), amt = 10, addl = 3, ii = 12)

    test_that("Using simulate with et works and gives a different data frame", {
      et2 <- as.data.frame(simulate(et))
      et1 <- as.data.frame(et)
      expect_false(all(et1$time == et2$time))
      et3 <- as.data.frame(et)
      expect_true(all(et1$time == et3$time))
      et$simulate()
      et3 <- as.data.frame(et)
      expect_false(all(et1$time == et3$time))
    })

    test_that("Low/High middle tests; i.e windows", {
      et2 <- et[which(!is.na(et$low)), ]
      expect_true(et$show["low"])
      expect_true(et$show["high"])
      expect_true(all(et2$time < et2$high))
      expect_true(all(et2$time > et2$low))
    })

    et <- et(list(
      c(0, 1),
      c(1, 4),
      c(4, 8),
      c(8, 12),
      c(12, 24)
    ), cmt = 1) |>
      et(amt = 10) |>
      et(c(11, 13), amt = 10, addl = 3, ii = 12)

    test_that("Low/High middle tests; i.e windows", {
      et2 <- et[which(!is.na(et$low)), ]
      expect_true(et$show["low"])
      expect_true(et$show["high"])
      expect_true(all(et2$time < et2$high))
      expect_true(all(et2$time > et2$low))
    })

    test_that("Window Errors", {
      expect_error(et(list(c(1, 0))))
      expect_error(et(list(c(0, 1, 2, 4))))
      expect_error(et(list(c(2, 3, 1))))
    })
    test_that("until is inclusive", {
      expect_equal(et(amt = 1, time = 50, until = 57.5, ii = 1.5)$addl, 5)
      expect_equal(et(amt = 1, time = 50, until = 57.49999, ii = 1.5)$addl, 4)
      expect_equal(et(amt = 1, time = 50, until = 57.50001, ii = 1.5)$addl, 5)
    })

    test_that("et expected errors", {
      expect_error(et(list(c(2, 1), c(3, 4)), amt = 3))
      expect_error(et(list(c(1, 2), 3, c(1, 2, 3)), amt = 3))
      expect_error(et(list(c(1, 2), 3, TRUE), amt = 3))
    })

    test_that("et steady state constant infusion", {
      expect_error(et(amt = 0, rate = 10, ii = 0, ss = 2))
      expect_error(et(amt = 0, rate = 10, ii = 2, ss = 1))
      expect_error(et(amt = 0, rate = -2, ii = 0, ss = 1))
      expect_error(et(rate = 10, ii = 0, ss = 2))
      expect_error(et(rate = 10, ii = 2, ss = 1))
      expect_error(et(rate = -2, ii = 0, ss = 1))

      t1 <- et(amt = 0, rate = 10, ii = 0, ss = 1) |> as.data.frame()
      t2 <- et(rate = 10, ii = 0, ss = 1) |> as.data.frame()
      expect_equal(t1, t2)

      t1 <- et(amt = 0, rate = -1, ii = 0, ss = 1) |> as.data.frame()
      t2 <- et(rate = -1, ii = 0, ss = 1) |> as.data.frame()
      expect_equal(t1, t2)
    })

    test_that("et addl expand", {
      ev <- et(amt = 3, ii = 24, until = 120)
      tmp <- etExpand(ev)
      expect_equal(ev$amt, 3)
      expect_equal(tmp$time, c(0, 24, 48, 72, 96, 120))
      ev$expand()
      expect_equal(ev$time, c(0, 24, 48, 72, 96, 120))
    })

    ev2 <- et(amt = 3, ii = 24, until = 120) |> et(amt = 3, rate = 7)

    test_that("data.frame conversion", {
      tmp <- data.frame(ev2)
      expect_equal(names(tmp), c("time", "amt", "rate", "ii", "addl", "evid"))
      expect_false(inherits(tmp$rate, "rxRateDur"))
      expect_false(inherits(tmp$evid, "rxEvid"))
    })

    test_that("tibble conversion", {
      tmp <- tibble::as_tibble(ev2)
      expect_equal(names(tmp), c("time", "amt", "rate", "ii", "addl", "evid"))
      expect_false(inherits(tmp$rate, "rxRateDur"))
      expect_false(inherits(tmp$evid, "rxEvid"))
    })
  }

  test_that("seq() args work; see #97", {
    et1 <- et() |> add.sampling(seq(0, 24, by = 3))

    et2 <- et(from = 0, to = 24, by = 3)

    expect_equal(et1$time, et2$time)
  })

  test_that("errors", {
    expect_error(et(ii = 12, dosing.interval = 14))
    expect_error(et(dose = 4, amt = 3))
    expect_error(et(cmt = 1, dosing.to = 3))
    expect_error(et(cmt = 1, dose.to = 3))
    expect_error(et(cmt = 1, state = 3))
    expect_error(et(amt.units = "mg", dose.units = "mg"))
    expect_error(et(time.units = "mg", timeUnits = "mg"))
    expect_error(et(time = 4, start.time = 5))
    expect_error(et(nbr.doses = 4, nbrDoses = 5))
    expect_error(et(dur = 4, duration = 5))
  })

  test_that("dose=0 is OK; see #192", {
    ev1 <- et(amt = 0, time = 10)
    ev2 <- eventTable()
    ev2$add.dosing(dose = 0, start.time = 10)

    expect_equal(as.data.frame(ev1), as.data.frame(ev2))
  })

  test_that("Issue #236 math in to/from", {
    expect_error(et(from = 0, to = 168 * 2 * 6, length.out = 168 * 2 * 6 + 1), NA)
  })

  test_that("Issue #257 numeric cmt vectorized", {
    ds4 <- c(1, 2, 3, 4)
    rate <- c(1.5, 2.5, 3.5, 4.5)
    expect_error(et() |> et(amt = ds4, rate = rate, cmt = 4), NA)
  })

  test_that("etRep #313", {
    skip_if_not_installed("units")
    sch1 <- et(timeUnits = "hr") |>
      et(amt = 100, ii = 24, until = units::set_units(2, "days"))

    toto <- rep(sch1, times = 10, wait = units::set_units(19, "days"))

    expect_equal(as.numeric(toto$time), seq(0, by = 504, length.out = 10))

    sch1 <- et(timeUnits = "hr") |>
      et(amt = 100, ii = 24, until = units::set_units(2, "days")) |>
      etExpand()

    toto1 <- etExpand(toto)

    expect_warning(toto <- rep(sch1, times = 10, wait = units::set_units(19, "days")))
  })

  test_that("'by' and 'length.out'", {
    expect_error(et(0, 3, by = 0.5, length.out = 3))
  })

  test_that("'amt' and 'time' have different values", {
    expect_error(et(amt = c(1, 2), time = c(1, 3, 4)))
  })

  test_that("'time' with invalid list", {
    expect_error(et(time = list(1, 2, 3)))
  })

  test_that("'is' for rxode2 event tables are 'rxEt'", {
    ev <- et()
    expect_true(is(ev, "rxEt"))
  })

  test_that("can use 'evid=0' with time entries", {
    expect_error(et(amt = 10, cmt = 1, time = 0, evid = 1, id = 1) |>
                   et(time = c(0, 10, 20), evid = 0), NA)

  })

  test_that("extra doses are not added (nlmixr2/rxode2et#2)", {
    foo <- et(amt=10, id=1:2) |> et(time=1, id=2:3)
    expect_equal(
      foo$id[!is.na(foo$amt)],
      1:2
    )
  })

  test_that("event table id, (rxode2et#4)", {
    expect_error(et(amt = 10, time = 0, evid = 1, id = 1:5) |>
                   et(amt = 100, time = 0, evid = 1, id = 6:10) |>
                   et(amt = 1000, time = 0, evid = 1, id = 11), NA)

  })

  test_that("event table non-zero time", {
    suppressWarnings(expect_warning(et(amt=1.153846, ii=24*7*6, until=24*7*6*2) |>
                                      et(amt=1.153846, time=24*7*6*(2+8),
                                         ii=24*7*8, until=24*7), "until"))
  })


  test_that("event table cmt needs to be evaluated #16", {

    dosing_df <- data.frame(
      DOSE = c(0.1, 0.5),
      CMT = c("A", "B"),
      TIME = c(0, 0)
    )

    samp_t <- c(0, 0.1, 0.5, 1, 2)

    # The below should work... but does not
    sub_df <- dosing_df[1, , drop = T]

    expect_error(et(
      amt = sub_df$DOSE,
      cmt = sub_df$CMT,
      time = sub_df$TIME,
      evid = 1,
      id = 1:5
    ) |>
      add.sampling(time = samp_t), NA)

  })

  test_that("toTrialDuration works", {
    trialEnd = 2
    ev <- et(data.frame(id = rep(1:2, 3),  time = c(13, 14, 13.5, 14.5, 15.3, 16.5)))
    res <- toTrialDuration(ev, trialEnd = trialEnd, interval = 0.5)
    expect_setequal(res$time, c(13, 13.5, 14, 14.5, 15, 14, 14.5, 15, 15.5, 16))
  })

  test_that("Ad issue #23", {

    cmti <- "Ad"
    dose_nmol <- 3
    dosing <- et(time = 0, amt = dose_nmol, cmt = cmti)

    expect_equal(dosing$cmt, "Ad")

    dosing <- et(time = 0, amt = dose_nmol, cmt = cmtj)

    expect_equal(dosing$cmt, "cmtj")

  })

  test_that("test import with NA time", {

    e <- et()

    expect_warning(e$importEventTable(data.frame(
      ID = "A",
      TIME = c(738.9333333, NA),
      CMT = c("depot", "central"),
      AMT = c(300, NA),
      EVID = c(1, 0),
      DOSE = 300,
      COVAR = 1
    )))

  })


  test_that("another import", {

    e <- et()

    expect_warning(e$importEventTable(data.frame(ID = c("A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A131", "A131", "A131", "A131",
                                                        "A131", "A131", "A131", "A131", "A132", "A132", "A132", "A132",
                                                        "A132", "A132", "A132", "A132", "A132", "A132", "A132", "A132",
                                                        "A132", "A132", "A132", "A132", "A132", "A132", "A132", "A132",
                                                        "A132", "A132", "A132", "A132", "A132", "A132", "A132", "A132",
                                                        "A132", "A132", "A132", "A132", "A132", "A132", "A132", "A132",
                                                        "A132", "A132", "A132", "A132", "A133", "A133", "A133", "A133",
                                                        "A133", "A133", "A133", "A133", "A133", "A133", "A133", "A133",
                                                        "A133", "A133", "A133", "A133", "A133", "A133", "A133", "A133",
                                                        "A133", "A133", "A133", "A133", "A133", "A133", "A133", "A133",
                                                        "A133", "A133", "A133", "A133", "A133", "A133", "A133", "A133",
                                                        "A133", "A133", "A133", "A134", "A134", "A134", "A134", "A134",
                                                        "A134", "A134", "A134", "A134", "A134", "A134", "A135", "A135",
                                                        "A135", "A135", "A135", "A135", "A135", "A135", "A135", "A135",
                                                        "A135", "A135", "A135", "A135", "A135", "A135", "A135", "A135",
                                                        "A135", "A135", "A135", "A135", "A135", "A135", "A135", "A135",
                                                        "A135", "A135", "A135", "A135", "A135"),
                                                 TIME = c(0, 0, 1.3,
                                                          1.883333333, 2.766666667, 4.85, 6.883333333, 7.816666667, 23.81666667,
                                                          47.81666667, 71.81666667, 95.81666667, 166.2, 166.2, 190.2, 214.2,
                                                          238.2, 262.2, 332.3333333, 332.3333333, 332.95, 333.5333333,
                                                          334.5333333, 336.5333333, 338.5333333, 339.45, 355.45, 379.45,
                                                          403.45, 427.45, 504.3166667, 504.3166667, 528.3166667, 552.3166667,
                                                          576.3166667, 672.3166667, 696.3166667, 720.3166667, 744.3166667,
                                                          840.3166667, 864.3166667, 888.3166667, 909.5333333, 909.5333333,
                                                          910.3666667, 910.8166667, 911.7, 913.45, 915.4, 917, 1009.783333,
                                                          1009.783333, 1012.3, 1034.3, 1058.3, 1082.3, 1178.3, 1202.3,
                                                          1226.3, 1250.3, 2016.1, 2016.1, 2040.1, 2064.1, 2088.1, 2184.1,
                                                          2208.1, 2232.1, 2256.1, 2352.1, 2376.1, 2400.1, 2424.1, 2522.1,
                                                          2522.1, 2522.45, 2544.45, 2568.45, 2592.45, 2688.45, 2712.45,
                                                          2736.45, 2760.45, 2856.45, 2880.45, 2904.45, 2928.45, 3025.8,
                                                          3025.8, 3028.266667, 3050.266667, 3074.266667, 3098.266667, 3194.266667,
                                                          3218.266667, 3242.266667, 3266.266667, 0, 0, 0.85, 24.35, 48.35,
                                                          72.35, 96.35, 336.35, 360.35, 384.35, 408.35, 432.35, 500.7,
                                                          500.7, 503.0166667, 525.0166667, 549.0166667, 573.0166667, 597.0166667,
                                                          666.1333333, 666.1333333, 666.9833333, 667.4833333, 668.4833333,
                                                          669.9833333, 672.0333333, 673.7, 689.7, 713.7, 737.7, 761.7,
                                                          833.7, 881.7, 905.7, 1001.7, 1025.7, 1049.7, 1082.816667, 1082.816667,
                                                          1085.7, 0, 0, 1.616666667, 2.033333333, 2.866666667, 4.866666667,
                                                          6.283333333, 24.28333333, 170.2833333, 170.2833333, 194.2833333,
                                                          218.2833333, 242.2833333, 261.6833333, 333.1666667, 357.1666667,
                                                          381.1666667, 405.1666667, 429.1666667, 505.35, 505.35, 508.3333333,
                                                          530.3333333, 554.3333333, 578.3333333, 595.4, 595.4, 596.9333333,
                                                          666.9333333, 690.9333333, 714.9333333, 738.9333333, NA, NA, NA,
                                                          NA, NA, NA, NA, 0, 0, 1.133333333, 1.716666667, 2.633333333,
                                                          4.416666667, 6.216666667, 24.21666667, 48.21666667, 72.21666667,
                                                          96.21666667, 0, 0, 0.9666666667, 1.466666667, 2.266666667, 4.35,
                                                          6.016666667, 24.01666667, 48.01666667, 168.0166667, 192.0166667,
                                                          216.0166667, 240.0166667, 264.0166667, 336.0166667, 360.0166667,
                                                          384.0166667, 408.0166667, 432.0166667, 503.3, 503.3, 505.9, 527.9,
                                                          551.9, 575.9, 599.9, 1175.566667, 1175.566667, 1178, 1200, 1224
                                                          ),
                                                 CMT = c("central", "depot", "central", "central", "central",
                                                         "central", "central", "central", "depot", "depot", "depot", "depot",
                                                         "central", "depot", "depot", "depot", "depot", "depot", "central",
                                                         "depot", "central", "central", "central", "central", "central",
                                                         "central", "depot", "depot", "depot", "depot", "central", "depot",
                                                         "depot", "depot", "depot", "depot", "depot", "depot", "depot",
                                                         "depot", "depot", "depot", "central", "depot", "central", "central",
                                                         "central", "central", "central", "central", "central", "depot",
                                                         "central", "depot", "depot", "depot", "depot", "depot", "depot",
                                                         "depot", "central", "depot", "depot", "depot", "depot", "depot",
                                                         "depot", "depot", "depot", "depot", "depot", "depot", "depot",
                                                         "central", "depot", "central", "depot", "depot", "depot", "depot",
                                                         "depot", "depot", "depot", "depot", "depot", "depot", "depot",
                                                         "central", "depot", "central", "depot", "depot", "depot", "depot",
                                                         "depot", "depot", "depot", "central", "depot", "central", "depot",
                                                         "depot", "depot", "depot", "depot", "depot", "depot", "depot",
                                                         "depot", "central", "depot", "central", "depot", "depot", "depot",
                                                         "depot", "central", "depot", "central", "central", "central",
                                                         "central", "central", "central", "depot", "depot", "depot", "depot",
                                                         "depot", "depot", "depot", "depot", "depot", "depot", "central",
                                                         "depot", "central", "central", "depot", "central", "central",
                                                         "central", "central", "central", "depot", "central", "depot",
                                                         "depot", "depot", "depot", "depot", "depot", "depot", "depot",
                                                         "depot", "depot", "central", "depot", "central", "depot", "depot",
                                                         "depot", "central", "depot", "central", "depot", "depot", "depot",
                                                         "depot", "central", "central", "central", "central", "central",
                                                         "central", "central", "central", "depot", "central", "central",
                                                         "central", "central", "central", "depot", "depot", "depot", "depot",
                                                         "central", "depot", "central", "central", "central", "central",
                                                         "central", "depot", "depot", "depot", "depot", "depot", "depot",
                                                         "depot", "depot", "depot", "depot", "depot", "depot", "central",
                                                         "depot", "central", "depot", "depot", "depot", "depot", "central",
                                                         "depot", "central", "depot", "depot"),
                                                 AMT = c(NA, 400, NA, NA,
                                                         NA, NA, NA, NA, 400, 400, 400, 400, NA, 400, 400, 400, 400, 400,
                                                         NA, 400, NA, NA, NA, NA, NA, NA, 400, 400, 400, 400, NA, 400,
                                                         400, 400, 400, 400, 400, 400, 400, 400, 400, 400, NA, 400, NA,
                                                         NA, NA, NA, NA, NA, NA, 400, NA, 400, 400, 400, 400, 400, 400,
                                                         400, NA, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400,
                                                         400, NA, 400, NA, 400, 400, 400, 400, 400, 400, 400, 400, 400,
                                                         400, 400, NA, 400, NA, 400, 400, 400, 400, 400, 400, 400, NA,
                                                         500, NA, 500, 500, 500, 500, 450, 450, 450, 450, 450, NA, 450,
                                                         NA, 450, 450, 450, 450, NA, 450, NA, NA, NA, NA, NA, NA, 450,
                                                         450, 450, 450, 450, 450, 450, 450, 450, 450, NA, 450, NA, NA,
                                                         350, NA, NA, NA, NA, NA, 350, NA, 300, 300, 300, 300, 300, 300,
                                                         300, 300, 300, 300, NA, 300, NA, 300, 300, 300, NA, 300, NA,
                                                         300, 300, 300, 300, NA, NA, NA, NA, NA, NA, NA, NA, 350, NA,
                                                         NA, NA, NA, NA, 350, 350, 350, 350, NA, 400, NA, NA, NA, NA,
                                                         NA, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400,
                                                         NA, 400, NA, 400, 400, 400, 400, NA, 400, NA, 400, 400),
                                                 DOSE = c(0,
                                                          400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400,
                                                          400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400,
                                                          400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400,
                                                          400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400,
                                                          400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400,
                                                          400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400,
                                                          400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400,
                                                          400, 400, 400, 400, 400, 0, 500, 500, 500, 500, 500, 500, 450,
                                                          450, 450, 450, 450, 450, 450, 450, 450, 450, 450, 450, 450, 450,
                                                          450, 450, 450, 450, 450, 450, 450, 450, 450, 450, 450, 450, 450,
                                                          450, 450, 450, 450, 450, 450, 0, 350, 350, 350, 350, 350, 350,
                                                          350, 350, 300, 300, 300, 300, 300, 300, 300, 300, 300, 300, 300,
                                                          300, 300, 300, 300, 300, 300, 300, 300, 300, 300, 300, 300, 300,
                                                          300, 300, 300, 300, 300, 300, 0, 350, 350, 350, 350, 350, 350,
                                                          350, 350, 350, 350, 0, 400, 400, 400, 400, 400, 400, 400, 400,
                                                          400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400,
                                                          400, 400, 400, 400, 400, 400, 400, 400, 400),
                                                 EVID = c(0, 1,
                                                          0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0,
                                                          0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1,
                                                          0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1,
                                                          1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                                                          1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1,
                                                          1, 1, 0, 1, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1,
                                                          1, 1, 1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1,
                                                          1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 0,
                                                          0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0,
                                                          0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1,
                                                          1, 0, 1, 0, 1, 1),
                                                 COVAR = c(1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                                                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
                                                           1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
                                                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                                           0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                                           1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
                                                 )))

  })

  test_that("as.character.rxEvid", {
    expect_equal(
      as.character.rxEvid(-1:9),
      c("-1:Invalid", "0:Observation", "1:Dose (Add)", "2:Other", "3:Reset",
        "4:Reset&Dose", "5:Replace", "6:Multiply", "7:Transit", "8:Invalid",
        "9:Invalid")
    )
    expect_equal(
      as.character.rxEvid("A"),
      "A:Invalid"
    )
    expect_equal(
      as.character.rxEvid(0.5),
      "0.5:Invalid"
    )
  })

  test_that("sampling windows versus PopED windows", {

    e1 <- et(list(c(0.1, 1, 50),
                  c(0.5, 2, 50),
                  c(0.5, 3, 50),
                  c(0.5, 25, 50),
                  c(0.5, 25, 50),
                  c(0.5, 30, 50),
                  c(0.5, 50, 80),
                  c(0.5, 60, 90))) |>
      et(amt=100)
    expect_equal(attr(class(e1), ".rxode2.lst")$randomType, 1L)

    expect_warning(simulate(e1))

    e2 <- et(list(c(0.1, 50),
                  c(0.5,  50),
                  c(0.5, 50),
                  c(0.5, 50),
                  c(0.5, 50),
                  c(0.5, 50),
                  c(0.5, 80),
                  c(0.5, 90))) |>
      et(amt=100)

    expect_equal(attr(class(e2), ".rxode2.lst")$randomType, 2L)

    expect_warning(simulate(e2), NA)

    e3 <- et(list(c(0.9, 1.1),
                  c(1.9, 2.1),
                  c(2.9, 3.1),
                  c(3.9, 4.1),
                  c(4.9, 5.1),
                  c(5.9, 6.1),
                  c(6.9, 7.1),
                  c(7.9, 8.1))) |>
      et(amt=100)

    expect_equal(attr(class(e3), ".rxode2.lst")$randomType, 2L)

    e4 <- et(list(c(4, 0.5, NA))) |>
      et(id=1:10000)

    expect_equal(attr(class(e4), ".rxode2.lst")$randomType, 3L)
    df4 <- as.data.frame(e4)
    expect_equal(df4$low[1], 4)
    expect_equal(df4$high[1], 0.5)

    e5 <- simulate(e4, seed=42)
    expect_equal(mean(as.numeric(e5$time)), 4, tolerance=0.1)
    expect_equal(sd(as.numeric(e5$time)), 0.5, tolerance=0.1)

    e6 <- et(list(c(4, 2, NA))) |>
      et(id=1:10000)
    expect_equal(attr(class(e6), ".rxode2.lst")$randomType, 3L)
    df6 <- as.data.frame(e6)
    expect_equal(df6$low[1], 4)
    expect_equal(df6$high[1], 2)

    e7 <- simulate(e6, seed=42)
    expect_equal(mean(as.numeric(e7$time)), 4, tolerance=0.1)
    expect_equal(sd(as.numeric(e7$time)), 2, tolerance=0.1)

  })

  test_that("dosing window simulation", {

    e1 <- et(amt=100, time=list(c(0, 1))) |>
      et(id=1:10000)
    e2 <- simulate(e1, seed=42)
    df <- as.data.frame(e2)
    expect_equal(mean(df$time), 0.5, tolerance=0.1)
    expect_equal(sd(df$time), sqrt(1/12), tolerance=0.1)

    e3 <- et(amt=100, time=list(c(4, 0.5, NA))) |>
      et(id=1:10000)
    e4 <- simulate(e3, seed=42)
    df <- as.data.frame(e4)
    expect_equal(mean(df$time), 4, tolerance=0.1)
    expect_equal(sd(df$time), 0.5, tolerance=0.1)

  })


  test_that("adding an ID in the improper order will not cause an issue", {

    e <- et(1:10) |>
      et(id=rev(1:10))

    expect_equal(unique(e$id), 1:10)

  })

  test_that("et improper piped arguments #722", {
    expect_error(et(amt=0, cmt=central, id=c(1:10)) |>
                   et(c(0,1), cmt=Cc), NA)
  })

  test_that("et can add doses per id (#725)", {
    expect_error(rxWithSeed(100, {

      sim.cov <- data.frame(id=1:10,
                            wt=rnorm(10, 70, 10))

      e_1 <-   et(id=sim.cov$id) %>%
        add.dosing(start.time = 0,dose = 600000*sim.cov$wt, nbr.doses = 14, dosing.interval = 8, dur=15/60) %>%
        add.dosing(start.time = (5+9)*24,dose = 600000*sim.cov$wt, nbr.doses = 14, dosing.interval = 8, dur=15/60) %>%
        add.sampling(seq(from = 0, to = 24*21, by = 0.25)) %>%
        as.data.frame %>%
        merge(sim.cov, by="id")

    }), NA)
  })

  test_that("when solving without observations, they are assumed (#858)", {

    mod3 <- function() {
      ini({
        TKA   <- 2.94E-01
        TCL   <- 1.86E+01
        TV2   <-4.02E+01
        TQ    <-1.05E+01
        TV3   <-2.97E+02
        TKin  <- 1
        TKout <- 1
        TEC50 <-200
      })
      model({
        KA            <- TKA
        CL            <- TCL * (WT / 70) ^ 0.75
        V2            <- TV2
        Q             <- TQ
        V3            <- TV3
        Kin           <- TKin
        Kout          <- TKout
        EC50          <- TEC50
        Tz            <- 8
        amp           <- 0.1
        C2            <- central/V2
        C3            <- peri/V3
        d/dt(depot)   <- -KA*depot
        d/dt(central) <- KA*depot - CL*C2 - Q*C2 + Q*C3
        d/dt(peri)    <-                    Q*C2 - Q*C3
        d/dt(eff)     <-  Kin - Kout*(1-C2/(EC50+C2))*eff
        eff(0)        <- 1  ## This specifies that the effect compartment starts at 1.
      })
    }

    ev <- et(amount.units="mg", time.units="hours") %>%
      et(amt=10000, cmt=1) %>%
      et(id=1:4)

    rxWithSeed(10, {
      r1 <- solve(mod3, ev,
                  iCov=data.frame(id=1:4, WT=rnorm(4, 70, 10)))
    }, rxseed = 10)

    expect_true(nrow(r1) > 0)
  })

  test_that("Fix for #732", {

    rxWithSeed(123, {
      # Setup the rxode2 event table
      eventTable <- et(amt=320, evid=1, cmt=1, time = 0) |> # nolint: object_name_linter.
        et(amt =  320, evid=1, cmt=2, time = 0) |>
        et(list(c(0.1, 0.4), # sampling windows for first profile
                c(0.3, 0.9),
                c(0.7, 1.5),
                c(1.5, 2.5),
                c(2.5, 4),
                c(4, 6),
                c(6, 8.5),
                c(8.75, 10),
                c( 10, 12),
                c(23, 27))) |>
        et(amt=320, evid=4, time=72, cmt=1) |> # reset & dose
        et(amt=320, evid=1, time=72, cmt=2) |> # dose afterward
        et(list(c(72.1, 72.4),
                c(72.3, 72.9),
                c(72.7, 73.5),
                c(73.5, 74.5),
                c(74.5, 78.5),
                c(76, 78),
                c(78, 80.5),
                c(80.75, 82),
                c(82, 84),
                c(95, 100))) |>
        et(id=1:14) # Number of subjects to sample


        # Now define the nlmixr2/rxode2 model used for both estimation and simulation
        mod <- function() {
          # Parameters
          ini({
            tka <- 0.45; label("Ka (first order absorption)")
            trate <- 0.4 ; label("Zero order rate")
            tcl <- 1; label("Cl")
            tv <- 3.45; label("V")
            fDepot <- logit(0.5) ; label("amount of dose in first order absorption")
            eta.ka ~ 0.6
            eta.cl ~ 0.3
            eta.v ~ 0.1
            add.sd <- 0.7
          })
          # and a model block with the error specification and model specification
          model({
            ka <- exp(tka + eta.ka)
            cl <- exp(tcl + eta.cl)
            v <- exp(tv + eta.v)
            d/dt(depot) = -ka * depot
            d/dt(center) = ka * depot - cl / v * center

            f(depot) <- expit(fDepot)
            f(center) <- 1-expit(fDepot)
            rate(center) <- exp(trate)

            cp = center / v
            cp ~ add(add.sd)
          })
        }

        mod <- mod()

        # Simulate the data
        d <- rxSolve(mod, eventTable, addDosing = TRUE)

        expect_equal(d |>
                       dplyr::filter(evid==1 & cmt==2) |> dplyr::pull(time) |> unique(),
                     c(0, 72))

        function() {
          ini({
            tka <- 0.45
            label("Ka (first order absorption)")
            trate <- 0.4
            label("Zero order rate")
            tcl <- 1
            label("Cl")
            tv <- 3.45
            label("V")
            fDepot <- 0
            label("amount of dose in first order absorption")
            add.sd <- c(0, 0.7)
            eta.ka ~ 0.6
            eta.cl ~ 0.3
            eta.v ~ 0.1
          })
          model({
            ka <- exp(tka + eta.ka)
            cl <- exp(tcl + eta.cl)
            v <- exp(tv + eta.v)

            f(depot) <- expit(fDepot, 0, 1)
            f(center) <- 1 - expit(fDepot, 0, 1)
            rate(center) <- exp(trate)
            cp = linCmt()
            cp ~ add(add.sd)
          })
        }
    }, rxseed = 123)
  })

  test_that("Only adds dose for subject 1, Fix for #723", {

    sub_ids = c(1,2,3,4,5)

    # creating observations
    obs_times = c(0:20)
    events = et(time=obs_times, id=sub_ids)

    for(sub_id in sub_ids){
      events = etRbind(
        events,
        # This does not work
        et(cmt = "Ac", amt=c(1,1,1), time = c(0, 1, 2), id=as.character(sub_id)))
    }

    events |> dplyr::filter(id ==2 & cmt == "Ac") |> nrow() |> expect_equal(3)

  })

  test_that("$expand() issue #721", {
    ev <- et()
    ev <- ev %>%
      et(id=1:10, amt=1:10, dosing.to=1, ii=14, addl=5, rate=(1:10)/(15/60/24))
    expect_error(ev$expand(), NA)
  })

})

rxTest({
  skip_if_not_installed("units")
  # Test reset event EVID=3

  ## 6.1
  mod <- rxode2({
    a <- 6
    b <- 0.6
    d / dt(intestine) <- -a * intestine
    d / dt(blood) <- a * intestine - b * blood
  })

  et <- eventTable(time.units = "day")

  et$add.sampling(seq(0, 10, by = 1 / 24))
  et$add.dosing(
    dose = 2 / 24, start.time = 0,
    nbr.doses = 10, dosing.interval = 1
  )

  et$add.dosing(start.time = 7.5, evid = 3, dose = 0)

  ms <- c(.methods0, "lsoda")

  m <- ms[1]

  for (m in ms) {

    x2 <- solve(mod, et, method = m)

    #x2 |> plot(blood)

    x27 <- x2 |>
      dplyr::filter(time >= units::set_units(7.5, "days")) |>
      dplyr::filter(time < units::set_units(8, "days"))

    zeros <- rep(0, length(x27$blood))

    test_that(sprintf("EVID=3 resets the system (%s)", m), {
      expect_true(any(x27$blood == zeros))
      expect_true(any(x27$intestine == zeros))
    })
  }

  et <- eventTable() |>
    add.dosing(dose = 3, nbr.doses = 6, dosing.interval = 8) |>
    add.sampling(seq(0, 48, length.out = 200))

  sol.1c <- rxode2({
    V <- 20
    Cl <- 2
    blood <- linCmt()
  })

  x2 <- solve(sol.1c, et)

  et1 <- et |> et(time = 9, evid = 3)

  et1 <- et1 |> units::set_units(h)

  x2 <- solve(sol.1c, et1)

  test_that("Solved Linear EVID=3", {
    expect_true(all((x2 |>
                       dplyr::filter(time > units::set_units(9, h)) |>
                       dplyr::filter(time < units::set_units(12, h)))$blood == 0))
  })
})

rxTest({
  test_that("'drop' test", {

    ode <- rxode2("
         d/dt(X) = a*X + Y*Z;
         d/dt(Y) = b*(Y - Z);
         d/dt(Z) = -X*Y + c*Y - Z;")

    et <- eventTable() # default time units
    et$add.sampling(seq(from = 0, to = 10, by = 1))

    out0 <- rxSolve(ode,
                    params = c(a = -8 / 3, b = -10, c = 28),
                    events = et, inits = c(X = 1, Y = 1, Z = 1)
                    )

    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10, c = 28),
                   events = et, inits = c(X = 1, Y = 1, Z = 1),
                   drop = "X"
                   )

    expect_equal(names(out), c("time", "Y", "Z"))
    expect_equal(out$time, out0$time)
    expect_equal(out$Y, out0$Y)
    expect_equal(out$Z, out0$Z)

    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10, c = 28),
                   events = et, inits = c(X = 1, Y = 1, Z = 1),
                   drop = c("X", "Y")
                   )

    expect_equal(names(out), c("time", "Z"))
    expect_equal(out$time, out0$time)
    expect_equal(out$Z, out0$Z)


    expect_warning(rxSolve(ode,
                           params = c(a = -8 / 3, b = -10, c = 28),
                           events = et, inits = c(X = 1, Y = 1, Z = 1),
                           drop = c("X", "Y", "ww")
                           ))

    expect_warning(rxSolve(ode,
                           params = c(a = -8 / 3, b = -10, c = 28),
                           events = et, inits = c(X = 1, Y = 1, Z = 1),
                           drop = c("X", "Y", "ww"), warnDrop = FALSE
                           ), NA)

    expect_warning(rxSolve(ode,
                           params = c(a = -8 / 3, b = -10, c = 28),
                           events = et, inits = c(X = 1, Y = 1, Z = 1),
                           drop = "time"
                           ))
  })
})

rxTest({
  # Test Transit compartment model

  dat1 <- structure(list(time = c(0, 0.0503, 0.1005, 0.1508, 0.201, 0.2513, 0.3015, 0.3518, 0.402, 0.4523, 0.5025, 0.5528, 0.603, 0.6533, 0.7035, 8.9447, 8.995, 9.0452, 9.0955, 9.1457, 9.196, 9.2462, 9.2965, 9.3467, 9.397, 9.4472, 9.4975, 9.5477, 9.598, 9.6482, 9.6985), k = c(0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814, 0.3814), ktr = c(57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027, 57.027), depot = c(0, 0, 0, 0.0044, 0.1348, 1.096, 3.9792, 8.5762, 13.1524, 16.2823, 17.8004, 18.2531, 18.1919, 17.9355, 17.6201, 0.7694, 0.7548, 0.7405, 0.7265, 0.7128, 0.6993, 0.6861, 0.6731, 0.6604, 0.6479, 0.6356, 0.6236, 0.6118, 0.6002, 0.5889, 0.5777), cen = c(0, 0, 0, 0, 8e-04, 0.0102, 0.0546, 0.1709, 0.3748, 0.6489, 0.9611, 1.285, 1.6058, 1.9171, 2.217, 2.4915, 2.4586, 2.426, 2.3939, 2.362, 2.3306, 2.2994, 2.2686, 2.2382, 2.2081, 2.1783, 2.1488, 2.1197, 2.091, 2.0625, 2.0344)), row.names = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L, 179L, 180L, 181L, 182L, 183L, 184L, 185L, 186L, 187L, 188L, 189L, 190L, 191L, 192L, 193L, 194L), class = "data.frame")

  mod <- rxode2("
## Table 3 from Savic 2007
cl = 17.2 # (L/hr)
vc = 45.1 # L
ka = 0.38 # 1/hr
mtt = 0.37 # hr
bio=1
n = 20.1
k = cl/vc
ktr = (n+1)/mtt
## note that lgammafn is the same as lgamma in R.
d/dt(depot) = exp(log(bio*podo(depot))+log(ktr)+n*log(ktr*tad(depot))-ktr*tad(depot)-lgammafn(n+1))-ka*depot
d/dt(cen) = ka*depot-k*cen
")

  et <- eventTable()
  et$add.sampling(seq(0, 10, length.out = 200))
  et$add.dosing(20, start.time = 0, evid=7)

  expect_error(rxSolve(mod, et, transitAbs = TRUE))

  transit <- rxSolve(mod, et)

  test_that("Transit absorption works", {
    expect_equal(
      round(as.data.frame(transit[c(1:15, seq(194 - 15, 194)), ]), 4),
      dat1
    )
  })

  test_that("Transit absorption is not specified, but still works.", {
    transit <- rxSolve(mod, et)
    expect_equal(
      round(as.data.frame(transit[c(1:15, seq(194 - 15, 194)), ]), 4),
      dat1
    )
  })

  mod <- rxode2("
## Table 3 from Savic 2007
cl = 17.2 # (L/hr)
vc = 45.1 # L
ka = 0.38 # 1/hr
mtt = 0.37 # hr
bio=1
n = 20.1
k = cl/vc
ktr = (n+1)/mtt
## note that lgamma1p is the same as lgamma(1+p) in R.
d/dt(depot) = exp(log(bio*podo(depot))+log(ktr)+n*log(ktr*tad(depot))-ktr*tad(depot)-lgamma1p(n))-ka*depot
d/dt(cen) = ka*depot-k*cen
")

  test_that("Transit absorption using lagmma1p works.", {
    transit <- rxSolve(mod, et)
    expect_equal(
      round(as.data.frame(transit[c(1:15, seq(194 - 15, 194)), ]), 4),
      dat1
    )
  })

  mod <- rxode2("
## Table 3 from Savic 2007
cl = 17.2 # (L/hr)
vc = 45.1 # L
ka = 0.38 # 1/hr
mtt = 0.37 # hr
bio=1
n = 20.1
k = cl/vc
ktr = (n+1)/mtt
d/dt(depot) = exp(log(bio*podo(depot))+log(ktr)+n*log(ktr*tad(depot))-ktr*tad(depot)-lgamma(n+1))-ka*depot
d/dt(cen) = ka*depot-k*cen
")

  test_that("Transit absorption using C's lgamma function.", {
    transit <- rxSolve(mod, et)
    expect_equal(round(as.data.frame(transit[c(1:15, seq(194 - 15, 194)), ]), 4), dat1)
  })

  mod <- rxode2("
## Table 3 from Savic 2007
cl = 17.2 # (L/hr)
vc = 45.1 # L
ka = 0.38 # 1/hr
mtt = 0.37 # hr
bio=1
n = 20.1
k = cl/vc
ktr = (n+1)/mtt
d/dt(depot) = exp(log(bio*podo())+log(ktr)+n*log(ktr*t)-ktr*t-lfactorial(n))-ka*depot
d/dt(cen) = ka*depot-k*cen
")

  test_that("Transit absorption using lfactorial", {
    transit <- rxSolve(mod, et)
    expect_equal(round(as.data.frame(transit[c(1:15, seq(194 - 15, 194)), ]), 4), dat1)
  })

  mod <- rxode2("
## Table 3 from Savic 2007
cl = 17.2 # (L/hr)
vc = 45.1 # L
ka = 0.38 # 1/hr
mtt = 0.37 # hr
bio=1
n = 20.1
k = cl/vc
ktr = (n+1)/mtt
d/dt(depot) = exp(log(bio*podo(depot))+log(ktr)+n*log(ktr*tad(depot))-ktr*tad(depot)-log(factorial(n)))-ka*depot
d/dt(cen) = ka*depot-k*cen
")

  test_that("Transit absorption using log(factorial)", {
    transit <- rxSolve(mod, et)
    expect_equal(round(as.data.frame(transit[c(1:15, seq(194 - 15, 194)), ]), 4), dat1)
  })

  mod <- rxode2("
## Table 3 from Savic 2007
cl = 17.2 # (L/hr)
vc = 45.1 # L
ka = 0.38 # 1/hr
mtt = 0.37 # hr
n = 20.1
k = cl/vc
## note that lgammafn is the same as lgamma in R.
d/dt(depot) = transit(n, mtt)-ka*depot
d/dt(cen) = ka*depot-k*cen
")

  # test dosing to central compartment
  etC <- eventTable()
  etC$add.sampling(seq(0, 10, length.out = 200))
  etC$add.dosing(20, start.time = 0, evid=1, cmt="cen")

  tmp <- rxSolve(mod, etC)
  expect_false(any(is.na(etC$cen)))

  test_that("Transit absorption is function that can take 2 arguments", {
    transit2 <- rxSolve(mod, et)
    expect_equal(
      round(as.data.frame(transit)[, names(transit) != "ktr"], 4),
      round(as.data.frame(transit2), 4)
    )
  })

  mod <- rxode2("
## Table 3 from Savic 2007
cl = 17.2 # (L/hr)
vc = 45.1 # L
ka = 0.38 # 1/hr
mtt = 0.37 # hr
bio=1
n = 20.1
k = cl/vc
d/dt(depot) = transit(n, mtt, bio)-ka*depot
d/dt(cen) = ka*depot-k*cen
")

  test_that("Transit absorption is function that can take 3 arguments", {
    transit2 <- rxSolve(mod, et)
    expect_equal(
      round(as.data.frame(transit)[, names(transit) != "ktr"], 4),
      round(as.data.frame(transit2), 4)
    )
  })

  mod <- rxode2("
## Table 3 from Savic 2007
cl = 17.2 # (L/hr)
vc = 45.1 # L
ka = 0.38 # 1/hr
mtt = 0.37 # hr
bio=1
n = 20.1
k = cl/vc
d/dt(depot) = transit(n, mtt, 1)-ka*depot
d/dt(cen) = ka*depot-k*cen
")

  test_that("Transit absorption can take numeric arguments", {
    transit2 <- rxSolve(mod, et)
    expect_equal(
      round(as.data.frame(transit)[, names(transit) != "ktr"], 4),
      round(as.data.frame(transit2), 4)
    )
  })

  test_that("transit can be a variable or ODE", {
    mod <- rxode2("
d/dt(transit) = -3
")

    expect_s3_class(mod, "rxode2")

    mod <- rxode2("
transit = matt + fun
")
    expect_s3_class(mod, "rxode2")
  })

  test_that("transit compartment works well with dual absorption (#804, #819)", {

    mod <- function() {
      ini({
        ## Table 3 from Savic 2007
        cl  <- 17.2 # (L/hr)
        vc  <- 45.1 # L
        ka  <- 0.38 # 1/hr
        mtt <- 1.37 # hr
        f2 <-0.5    # Fraction of 1st Order portion
        n   <- 20.101
      })
      model({
        k           <- cl/vc
        bio <- 1-f2
        d/dt(depot1) <- transit(n,mtt,bio)-ka*depot1
        d/dt(depot2) <- -ka*depot2
        f(depot2) <-f2
        d/dt(cen)   <- ka*depot1 + ka*depot2-k*cen
      })
    }

    ev1 <- et(0, 7, length.out=200) |>
      et(amt=20, cmt='depot1', evid=7) |>
      et(amt=20, cmt='depot2', evid=1)

    case1 <- rxSolve(mod, ev1)

    expect_true(max(case1$depot1) > 7.7)

  })
})

rxTest({
  test_that("state-dep mtime with state=0 fires at exactly mt1", {
    # mtime(t1) <- mt1 + state; state is constant at 0
    # => t1 fires at mt1 + 0 = mt1
    mod <- rxode2({
      mtime(t1) <- mt1 + offset
      d/dt(offset) <- 0
      d/dt(y) <- -kel * y
    })
    # Sparse sampling (by=3): 5 and 7 are NOT in the grid.
    # mtime events (evid 10-99) are observations that add their own time point.
    et <- eventTable() |>
      add.dosing(dose = 100, nbr.doses = 1, cmt = "y") |>
      add.sampling(seq(0, 21, by = 3))
    s <- solve(mod, et, inits = c(offset = 0, y = 0),
               params = c(mt1 = 5, kel = 0.1))
    expect_true(any(abs(s$time - 5) < 1e-6),
                label = "mtime fires at mt1 when state=0")
    expect_false(any(abs(s$time - 7) < 1e-6),
                 label = "mtime does not fire at mt1+2 when state=0")
  })

  test_that("state-dep mtime shifts by state value: mtime = mt1 + state_at_t1", {
    # mtime(t1) <- mt1 + offset; offset is constant at initial value S0
    # => t1 fires at mt1 + S0
    mod <- rxode2({
      mtime(t1) <- mt1 + offset
      d/dt(offset) <- 0
      d/dt(y) <- -kel * y
    })
    # Sparse sampling (by=3): 5, 7, and 3.5 are NOT in the grid.
    et <- eventTable() |>
      add.dosing(dose = 100, nbr.doses = 1, cmt = "y") |>
      add.sampling(seq(0, 21, by = 3))

    # S0 = 2: fires at 5 + 2 = 7, not at 5
    s2 <- solve(mod, et, inits = c(offset = 2, y = 0),
                params = c(mt1 = 5, kel = 0.1))
    expect_true(any(abs(s2$time - 7) < 1e-6),
                label = "mtime fires at mt1 + state (=7) when state=2")
    expect_false(any(abs(s2$time - 5) < 1e-6),
                 label = "mtime does not fire at bare mt1 when state=2")

    # S0 = -1.5: fires at 5 + (-1.5) = 3.5, not at 5
    s3 <- solve(mod, et, inits = c(offset = -1.5, y = 0),
                params = c(mt1 = 5, kel = 0.1))
    expect_true(any(abs(s3$time - 3.5) < 1e-6),
                label = "mtime fires at mt1 + state (=3.5) when state=-1.5")
    expect_false(any(abs(s3$time - 5) < 1e-6),
                 label = "mtime does not fire at bare mt1 when state=-1.5")
  })

  test_that("state-dep mtime: changing initial state shifts event time by same amount", {
    # Verify linearity: firing_time - mt1 == initial_state exactly
    mod <- rxode2({
      mtime(t1) <- mt1 + offset
      d/dt(offset) <- 0
      d/dt(y) <- -kel * y
    })
    mt1_val <- 3
    kel_val <- 0.2
    # Sampling by=0.7: none of the expected fire times (2.5, 3, 4, 5.5) fall on
    # the grid, so the mtime event adds its own time point for each S0.
    for (S0 in c(0, 1, 2.5, -0.5)) {
      et <- eventTable() |>
        add.dosing(dose = 100, nbr.doses = 1, cmt = "y") |>
        add.sampling(seq(0, 10, by = 0.7))
      s <- solve(mod, et, inits = c(offset = S0, y = 0),
                 params = c(mt1 = mt1_val, kel = kel_val))
      expected_time <- mt1_val + S0
      expect_true(any(abs(s$time - expected_time) < 1e-6),
                  label = paste0("mtime fires at mt1+S0=", expected_time,
                                 " when S0=", S0))
    }
  })

  test_that("state-dep mtime with decaying state: fires at mt1 + state(mt1)", {
    # mtime(t1) <- mt1 + state; state decays: state(t) = S0*exp(-ka*t)
    #
    # Correct semantics: the mtime expression is evaluated at the base time mt1
    # (the state-independent part of the expression).
    #   actual_event_time = mt1 + state(mt1) = mt1 + S0*exp(-ka*mt1)
    #
    # This is always AFTER mt1 (since state > 0 for positive S0).
    #
    mod <- rxode2({
      mtime(t1) <- mt1 + state
      d/dt(state) <- -ka * state
      d/dt(y) <- -kel * y
    })
    mt1_val <- 5
    ka_val  <- 0.3
    kel_val <- 0.1
    S0      <- 1.5

    # Correct: mt1 + state(mt1) = mt1 + S0*exp(-ka*mt1)
    T_actual <- mt1_val + S0 * exp(-ka_val * mt1_val)
    # Old wrong value: mt1 + state(0) = mt1 + S0
    T_old    <- mt1_val + S0  # = 6.5

    # Sparse sampling (by=2): T_actual and T_old are NOT in the grid.
    et <- eventTable() |>
      add.dosing(dose = 100, nbr.doses = 1, cmt = "y") |>
      add.sampling(seq(0, 15, by = 2))
    s <- solve(mod, et, inits = c(state = S0, y = 0),
               params = c(mt1 = mt1_val, ka = ka_val, kel = kel_val))

    expect_true(any(abs(s$time - T_actual) < 1e-4),
                label = paste0("mtime fires at mt1 + state(mt1) = ",
                               round(T_actual, 4)))
    expect_false(any(abs(s$time - T_old) < 1e-4),
                 label = paste0("mtime does NOT fire at old T0 = mt1 + state(0) = ",
                                round(T_old, 4)))
  })

  test_that("state-dependent mtime is parsed and solves without error", {
    mod <- rxode2({
      ka <- 0.04
      kel <- 0.6
      # mtime depends on state variable 'intestine'
      mtime(t1) <- intestine * 0.01
      d / dt(intestine) <- -ka * intestine
      d / dt(blood) <- ka * intestine - kel * blood
    })
    et <- eventTable() |>
      add.dosing(dose = 3, nbr.doses = 1) |>
      add.sampling(0:48)
    # Uses initial condition of intestine (100) to compute mtime at solve init
    expect_no_error(solve(mod, et, inits = c(intestine = 100, blood = 0)))
  })

  if (!.Call(`_rxode2_isIntel`)) {
    mod <- rxode2({
      ka <- 0.04
      kel <- 0.6
      mt1 <- 0.04
      mt2 <- 0.6
      mtime(t1) <- mt1
      mtime(t2) <- mt2
      if (t >= t1) {
        ka <- 0.4
      }
      if (t >= t2) {
        ka <- 4
      }
      d / dt(intestine) <- -ka * intestine
      d / dt(blood) <- ka * intestine - kel * blood
    })

    et <- eventTable() |>
      add.dosing(dose = 3, nbr.doses = 1) |>
      add.sampling(0:48)


    test_that("Solved model contains the model times", {
      s <- solve(mod, et)
      expect_true(any(s$time == 0.04))
      expect_true(any(s$time == 0.6))
      expect_false(any(s$time == 1.977))
      expect_false(any(s$time == 10.99))
      s <- solve(mod, et, c(mt1 = 1.977, mt2 = 10.09))
      expect_true(any(s$time == 1.977))
      expect_true(any(s$time == 10.09))
      expect_false(any(s$time == 0.04))
      expect_false(any(s$time == 0.6))
    })
  }
})

rxTest({

  expectSimWithout <- function(x) {
    expect_null(x$thetaMat)
    expect_null(x$omegaList)
    expect_null(x$sigmaList)
  }

  expectSimTheta <- function(x) {
    expect_type(x$thetaMat,"double")
    expect_true(inherits(x$thetaMat, "matrix"))
    expect_null(x$omegaList)
    expect_null(x$sigmaList)
  }

  expectSimAll <- function(x) {
    expect_type(x$thetaMat,"double")
    expect_true(inherits(x$thetaMat, "matrix"))
    expect_true(inherits(x$omegaList, "list"))
    expect_type(x$omegaList[[1]], "double")
    expect_true(inherits(x$omegaList[[1]], "matrix"))
    expect_true(inherits(x$sigmaList, "list"))
    expect_type(x$sigmaList[[1]], "double")
    expect_true(inherits(x$sigmaList[[1]], "matrix"))
  }

  for (meth in .methods0) {
    # context(sprintf("Test Parallel/Multi-subject Solve (%s)", meth))

    mod <- rxode2({
      d/dt(intestine) <- -a * intestine
      d/dt(blood) <- a * intestine - b * blood
    })

    et <- eventTable(time.units = "days")
    et$add.sampling(seq(0, 10, length.out = 50))
    et$add.dosing(
      dose = 2 / 24, rate = 2, strt.time = 0,
      nbr.doses = 10, dosing.interval = 1
    )

    p <- data.frame(a = 6, b = seq(0.4, 0.9, length.out = 4))

    pk1 <- suppressWarnings(rxSolve(mod, p, et, cores = 1, method = meth))

    expectSimWithout(pk1)

    pk2 <- suppressWarnings(rxSolve(mod, p, et, cores = 2, method = meth)) # CRAN requirement of at most 2 cores.

    expectSimWithout(pk2)

    test_that("Parallel Solve gives same results a single threaded solve", {
      expect_equal(as.data.frame(pk1), as.data.frame(pk2))
    })

    ## Test mixed solved and ODEs
    mod2 <- rxode2({
      ## the order of variables do not matter, the type of compartmental
      ## model is determined by the parameters specified.
      CL ~ TCL * exp(eta.Cl)
      C2 ~ linCmt(KA, CL, V2, Q, V3)
      eff(0) <- 1 ## This specifies that the effect compartment starts at 1.
      d / dt(eff) ~ Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      ##
      resp <- eff + err1
      pk <- C2 * exp(err2)
    })


    ev <- eventTable(amount.units = "mg", time.units = "hours") |>
      add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12, dosing.to = 2) |>
      add.dosing(dose = 20000, nbr.doses = 5, start.time = 120, dosing.interval = 24, dosing.to = 2) |>
      add.sampling(0:240)

    ## Add Residual differences
    sigma <- diag(2) * 0.05
    dimnames(sigma) <- list(c("err1", "err2"), c("err1", "err2"))

    pk3 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, ev, sigma = sigma, cores = 2, method = meth,
        ))

    expectSimWithout(pk3)

    pk3a <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, ev, sigma = sigma, cores = 2, method = meth, addDosing = TRUE
      ))

    expectSimWithout(pk3a)

    pk4 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, ev, sigma = sigma, cores = 1, method = meth
      ))

    expectSimWithout(pk4)

    test_that("Can solve the system.", {
      expect_true(rxIs(pk3, "data.frame"))
      expect_false(all(pk3$pk == 0))
      expect_true(rxIs(pk3a, "data.frame"))
      expect_false(all(pk3a$pk == 0))
      expect_true(rxIs(pk4, "data.frame"))
      expect_false(all(pk4$pk == 0))
    })

    mod3 <- rxode2({
      C2 <- prod(centr, 1 / V2)
      C3 ~ prod(peri, 1 / V3)
      CL ~ prod(TCL, exp(eta.Cl))
      d/dt(depot) ~ prod(-KA, depot)
      d/dt(centr) ~ sum(prod(KA, depot), -prod(CL, C2), -prod(Q, C2), prod(Q, C3))
      d/dt(peri) ~ sum(prod(Q, C2), -prod(Q, C3))
      d/dt(eff) <- sum(Kin, -prod(Kout, sum(1, -prod(C2, 1 / sum(EC50, C2))), eff))
      e1 <- err1
      e2 <- err2
      resp <- sum(eff, e1)
      pk <- prod(C2, exp(e2))
    })

    pk3 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, ev, sigma = sigma, cores = 2, method = meth
      ))

    expectSimWithout(pk3)

    pk4 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, ev, sigma = sigma, cores = 1, method = meth
      ))

    expectSimWithout(pk4)

    test_that("Can solve the system.", {
      expect_false(all(pk3$pk == 0))
      expect_true(rxIs(pk3, "data.frame"))
      expect_false(all(pk3$pk == 0))
      expect_true(rxIs(pk4, "data.frame"))
    })

    mod2 <- rxode2({
      C2 <- centr / V2
      C3 ~ peri / V3
      CL ~ TCL * exp(eta.Cl)
      d/dt(depot) ~ -KA * depot
      d/dt(centr) ~ KA * depot - CL * C2 - Q * C2 + Q * C3
      d/dt(peri) ~ Q * C2 - Q * C3
      d/dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- 1000
      e1 <- err1
      e2 <- err2
      resp <- eff + e1
      pk <- C2 * exp(e2)
    })

    pk3 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, ev, sigma = sigma, cores = 2, method = meth
      ))

    expectSimWithout(pk3)

    pk4 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, ev, sigma = sigma, cores = 1, method = meth
      ))

    expectSimWithout(pk4)

    test_that("Can solve the system.", {
      expect_false(all(pk3$pk == 0))
      expect_true(rxIs(pk3, "data.frame"))
      expect_false(all(pk4$pk == 0))
      expect_true(rxIs(pk4, "data.frame"))
    })

    pk2 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200, err1 = 0, err2 = 0),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, ev, cores = 1, method = meth
      ))

    expectSimWithout(pk2)

    pk3 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200, err1 = 0, err2 = 0),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, ev, cores = 2, method = meth
      ))

    expectSimWithout(pk3)


    ## "Study" Differences
    thetaMat <- diag(3) * 0.01
    dimnames(thetaMat) <- list(NULL, c("KA", "TCL", "V2"))

    pk4 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, nStud = 4, thetaMat = thetaMat, sigma = sigma, ev, cores = 1, method = meth
      ))

    expectSimTheta(pk4)

    pk5 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, nStud = 4, thetaMat = thetaMat, sigma = sigma, ev, cores = 2, method = meth
      ))

    expectSimTheta(pk5)

    pk6 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, nStud = 4, thetaMat = thetaMat, sigma = sigma, ev, cores = 1, dfSub = 4, dfObs = 4,
        method = meth
      ))

    expectSimAll(pk6)

    pk7 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, nStud = 4, thetaMat = thetaMat, sigma = sigma, ev, cores = 2, dfSub = 4, dfObs = 4,
        method = meth
      ))

    expectSimAll(pk7)

    test_that("Can solve the system.", {
      expect_true(rxIs(pk2, "data.frame"))
      expect_null(pk2$thetaMat)
      expect_null(pk2$omegaList)
      expect_null(pk2$sigmaList)
      expect_true(rxIs(pk3, "data.frame"))
      expect_null(pk3$thetaMat)
      expect_null(pk3$omegaList)
      expect_null(pk3$sigmaList)
      expect_true(rxIs(pk4, "data.frame"))
      expect_true(is(pk4$thetaMat, "matrix"))
      expect_null(pk4$omegaList)
      expect_null(pk4$sigmaList)
      expect_true(rxIs(pk5, "data.frame"))
      expect_true(is(pk5$thetaMat, "matrix"))
      expect_null(pk5$omegaList)
      expect_null(pk5$sigmaList)
      expect_true(rxIs(pk6, "data.frame"))
      expect_true(is(pk6$thetaMat, "matrix"))
      expect_type(pk6$omegaList, "list")
      expect_type(pk6$sigmaList, "list")
      expect_true(rxIs(pk7, "data.frame"))
      expect_true(is.matrix(pk7$thetaMat))
      expect_type(pk7$omegaList, "list")
      expect_type(pk7$sigmaList, "list")
    })
  }

  test_that("Can solve the system.", {

    ## Now Try multi-subject data.
    skip_if_not(file.exists(test_path("test-data-setup.rds")))
    dat <- readRDS(test_path("test-data-setup.rds"))

    pk7a <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        sigma = sigma, dat, cores = 1, method = meth
      ))

    expectSimWithout(pk7a)

    pk8 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        thetaMat = thetaMat, sigma = sigma, dat, nStud = 4, cores = 1, method = meth
      ))

    expectSimTheta(pk8)

    pk9 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        thetaMat = thetaMat, sigma = sigma, dat, nStud = 4, cores = 2, method = meth
      ))

    expectSimTheta(pk9)

    pk10 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        thetaMat = thetaMat, sigma = sigma, dat, nStud = 4, cores = 2, method = meth, simVariability = FALSE
      ))

    expectSimWithout(pk10)

    pk11 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        thetaMat = thetaMat, sigma = sigma, dat, nStud = 4, cores = 1, method = meth, simVariability = FALSE
      ))

    expectSimWithout(pk11)

    pk12 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        thetaMat = thetaMat, sigma = sigma, dat, nStud = 1, cores = 1, method = meth, dfSub = 4, dfObs = 4
      ))

    expectSimWithout(pk12)

    pk13 <-
      suppressWarnings(rxSolve(
        mod2,
        c(KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02, Kin = 1, Kout = 1, EC50 = 200),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        thetaMat = thetaMat, sigma = sigma, dat, nStud = 1, cores = 1, method = meth, dfSub = 4, dfObs = 4, simVariability=TRUE
      ))

    expectSimAll(pk13)

    expect_true(rxIs(pk8, "data.frame"))
    expect_true(rxIs(pk9, "data.frame"))
    expect_true(rxIs(pk10, "data.frame"))
    expect_true(rxIs(pk11, "data.frame"))

  })
})

rxTest({
  test_that("data.table", {
    mod <- rxode2({
      d / dt(intestine) <- -a * intestine
      d / dt(blood) <- a * intestine - b * blood
    })
    mod2 <- rxode2({
      C2 <- centr / V2
      C3 ~ peri / V3
      CL ~ TCL * exp(eta.Cl)
      d / dt(depot) ~ -KA * depot
      d / dt(centr) ~ KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) ~ Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- 1000
      e1 <- err1
      e2 <- err2
      resp <- eff + e1
      pk <- C2 * exp(e2)
    })

    et <- eventTable(time.units = "days")
    et$add.sampling(seq(0, 10, length.out = 50))
    et$add.dosing(
      dose = 2 / 24, rate = 2, strt.time = 0,
      nbr.doses = 10, dosing.interval = 1
    )

    p <- data.frame(a = 6, b = seq(0.4, 0.9, length.out = 4))
    dat <- NULL
    dataFile <- test_path("test-data-setup.rds")
    if (file.exists(dataFile)) {
      dat <- readRDS(dataFile)
    }

    for (rt in c("data.table", "tbl")) {
      p2 <- rxSolve(mod, p, et, cores = 1, returnType = rt)

      expect_s3_class(p2, rt)
      if (!is.null(dat)) {
        sigma <- diag(2) * 0.05
        dimnames(sigma) <- list(c("err1", "err2"), c("err1", "err2"))

        ev <- eventTable(amount.units = "mg", time.units = "hours") |>
          add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12, dosing.to = 2) |>
          add.dosing(dose = 20000, nbr.doses = 5, start.time = 120, dosing.interval = 24, dosing.to = 2) |>
          add.sampling(0:240)

        pk3 <- rxSolve(mod2, c(
          KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
          Kin = 1, Kout = 1, EC50 = 200
        ),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, ev, sigma = sigma, cores = 2, returnType = rt
        )

        expect_s3_class(pk3, rt)

        ## Should be no warning here...
        pk7a <- rxSolve(mod2, c(
          KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
          Kin = 1, Kout = 1, EC50 = 200
        ),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        sigma = sigma, dat, cores = 1, returnType = rt
        )

        expect_s3_class(pk7a, rt)

        thetaMat <- diag(3) * 0.01
        dimnames(thetaMat) <- list(NULL, c("KA", "TCL", "V2"))

        pk8 <- rxSolve(mod2, c(
          KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
          Kin = 1, Kout = 1, EC50 = 200
        ),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        thetaMat = thetaMat, sigma = sigma, dat, nStud = 4, cores = 1, returnType = rt
        )

        expect_s3_class(pk8, rt)
      }
    }
  })
})

rxTest({
  test_that("rxIs tests", {
    skip_if_not(file.exists(test_path("test-data-setup.rds")))
    dat <- readRDS(test_path("test-data-setup.rds"))
    dat <- tibble::as_tibble(dat)
    expect_true(rxIs(dat, "data.frame"))
    expect_true(rxIs(dat, "tbl"))
    expect_true(rxIs(dat, "tbl_df"))
    expect_true(rxIs(dat, "event.data.frame"))
    expect_true(rxIs(dat, "rx.event"))
    dat2 <- dat |>
      dplyr::mutate(amt = AMT, evid = EVID, time = TIME) |>
      dplyr::select(-AMT, -EVID, -TIME)
    expect_true(rxIs(dat2, "event.data.frame"))
    expect_true(rxIs(dat2, "rx.event"))
    dat3 <- dat |>
      dplyr::mutate(Amt = AMT, Evid = EVID, Time = TIME) |>
      dplyr::select(-AMT, -EVID, -TIME)
    expect_true(rxIs(dat3, "event.data.frame"))
    expect_true(rxIs(dat3, "rx.event"))
    dat4 <- dat |>
      dplyr::mutate(AMt = AMT, EVid = EVID, TIme = TIME) |>
      dplyr::select(-AMT, -EVID, -TIME)
    expect_false(rxIs(dat4, "event.data.frame"))
    expect_false(rxIs(dat4, "rx.event"))
    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev <- et(ev, id = 1:3)
    expect_true(rxIs(ev, "eventTable"))
    expect_true(rxIs(ev, "rxEt"))
    expect_true(rxIs(ev, "rx.event"))
    dat <- as.data.frame(dat)
    expect_true(rxIs(dat, "data.frame"))
    expect_false(rxIs(dat, "tbl"))
    expect_false(rxIs(dat, "tbl_df"))
    expect_true(rxIs(dat, "event.data.frame"))
    expect_true(rxIs(dat, "rx.event"))
    expect_false(rxIs(as.matrix(dat), "event.data.frame"))
    expect_true(rxIs(as.matrix(dat), "rx.event"))
    expect_true(rxIs(as.matrix(dat), "event.matrix"))
    dat2 <- dat |>
      dplyr::mutate(amt = AMT, evid = EVID, time = TIME) |>
      dplyr::select(-AMT, -EVID, -TIME)
    expect_true(rxIs(dat2, "event.data.frame"))
    expect_true(rxIs(dat2, "rx.event"))
    expect_true(rxIs(as.matrix(dat2), "event.matrix"))
    expect_true(rxIs(as.matrix(dat2), "rx.event"))
    dat3 <- dat |>
      dplyr::mutate(Amt = AMT, Evid = EVID, Time = TIME) |>
      dplyr::select(-AMT, -EVID, -TIME)
    expect_true(rxIs(dat3, "event.data.frame"))
    dat4 <- dat |>
      dplyr::mutate(AMt = AMT, EVid = EVID, TIme = TIME) |>
      dplyr::select(-AMT, -EVID, -TIME)
    expect_false(rxIs(as.matrix(dat4), "event.matrix"))
    expect_false(rxIs(as.matrix(dat4), "rx.event"))
    expect_false(rxIs(dat4, "event.data.frame"))
    expect_false(rxIs(dat4, "rx.event"))
    expect_false(rxIs(as.matrix(dat4), "event.matrix"))
    a <- numeric()
    expect_true(rxIs(a, "numeric"))
    a <- numeric(1)
    expect_true(rxIs(a, "numeric"))
    a <- matrix(a)
    expect_true(rxIs(a, "numeric.matrix"))
    expect_false(rxIs(a, "integer.matrix"))
    expect_false(rxIs(a, "logical.matrix"))
    expect_false(rxIs(a, "character.matrix"))
    expect_true(rxIs(a, "matrix"))
    a <- integer()
    expect_true(rxIs(a, "integer"))
    a <- integer(1)
    expect_true(rxIs(a, "integer"))
    a <- matrix(a)
    expect_false(rxIs(a, "numeric.matrix"))
    expect_true(rxIs(a, "integer.matrix"))
    expect_false(rxIs(a, "logical.matrix"))
    expect_false(rxIs(a, "character.matrix"))
    expect_true(rxIs(a, "matrix"))
    a <- logical()
    expect_true(rxIs(a, "logical"))
    a <- logical(1)
    expect_true(rxIs(a, "logical"))
    a <- matrix(a)
    expect_false(rxIs(a, "numeric.matrix"))
    expect_false(rxIs(a, "integer.matrix"))
    expect_true(rxIs(a, "logical.matrix"))
    expect_false(rxIs(a, "character.matrix"))
    expect_true(rxIs(a, "matrix"))
    a <- character()
    expect_true(rxIs(a, "character"))
    a <- character(1)
    expect_true(rxIs(a, "character"))
    a <- matrix(a)
    expect_false(rxIs(a, "numeric.matrix"))
    expect_false(rxIs(a, "integer.matrix"))
    expect_false(rxIs(a, "logical.matrix"))
    expect_true(rxIs(a, "character.matrix"))
    expect_true(rxIs(a, "matrix"))
    expect_true(rxIs(list(a = "b"), "list"))
    expect_true(rxIs(list(), "list"))
    expect_true(rxIs(list("b"), "list"))
    expect_false(rxIs(logical(), "list"))
    expect_false(rxIs(1.3, "integer"))
    expect_false(rxIs(1L, "numeric"))
    expect_false(rxIs(TRUE, "numeric"))
    expect_false(rxIs(list("b"), "numeric"))
    mod <- rxode2("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
d/dt(blood)     = a*intestine - b*blood
")
    expect_true(rxIs(mod, "rxode2"))
  })
})

rxTest({
  # -- Analytical background --------------------------------------------------
  # dur(depot) <- dur0 + state sets the infusion DURATION directly (rate=-2).
  # The infusion rate is derived as:  R = amt / (dur0 + state_at_start)
  #
  # For a depot with first-order drain rate ka and constant state S0:
  #
  #   D = dur0 + S0                        (infusion duration)
  #   R = amt / D                          (derived rate)
  #
  # During infusion (0 <= t <= D):
  #   depot(t) = (R/ka) * (1 - exp(-ka*t))
  #
  # After infusion (t > D):
  #   depot(t) = depot(D) * exp(-ka*(t-D))
  #            = (R/ka) * (1-exp(-ka*D)) * exp(-ka*(t-D))

  # -- Helper -----------------------------------------------------------------
  .depotAnalytic <- function(t, R, D, ka) {
    vapply(t, function(ti) {
      if (ti <= 0) return(0)
      if (ti <= D) {
        (R / ka) * (1 - exp(-ka * ti))
      } else {
        depot_D <- (R / ka) * (1 - exp(-ka * D))
        depot_D * exp(-ka * (ti - D))
      }
    }, numeric(1))
  }

  # -- Test 1 -----------------------------------------------------------------
  # state = 0 throughout  ->  dur(depot) = dur0 exactly.
  # Infusion duration = dur0; rate = amt/dur0; trajectory is the standard
  # fixed-duration analytical solution.
  test_that("state-dep dur with state=0 uses exactly dur0 (no state shift)", {
    mod <- rxode2({
      # Infusion duration is modelled as dur0 + state.
      # With state held at 0 this collapses to the fixed-duration case.
      dur(depot) <- dur0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- 0         # constant state; initial value is the only value
    })

    amt  <- 100
    dur0 <- 10
    ka   <- 0.5
    D    <- dur0          # state=0 -> effective duration = dur0
    R    <- amt / D       # derived rate
    times <- seq(0, 30, by = 0.5)

    et <- eventTable() |>
      add.dosing(dose = amt, rate = -2) |>  # rate=-2 -> use modelled duration
      add.sampling(times)

    s <- solve(mod, et,
               inits  = c(state = 0, depot = 0),
               params = c(dur0 = dur0, ka = ka))

    expected <- .depotAnalytic(times, R, D, ka)

    # Every sampled depot value must match the analytical trajectory computed
    # with duration = dur0.  A mismatch means the solver used a wrong duration.
    expect_equal(s$depot[s$time %in% times], expected,
                 tolerance = 1e-3,
                 label = "depot matches dur0-only analytical solution when state=0")
  })

  # -- Test 2 -----------------------------------------------------------------
  # state = S0 (constant)  ->  dur(depot) = dur0 + S0; rate = amt/(dur0+S0).
  # The full depot trajectory must match the closed-form solution for that rate.
  test_that("state-dep dur shifts infusion duration by state value: dur = dur0 + state_at_start", {
    mod <- rxode2({
      # Duration includes a state-variable contribution.
      # With d/dt(state)=0 the state stays at its initial value; the expected
      # duration is analytically known.
      dur(depot) <- dur0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- 0
    })

    amt  <- 100
    dur0 <- 10
    ka   <- 0.5
    times <- seq(0, 40, by = 0.5)

    for (S0 in c(2, 5, -3)) {
      D <- dur0 + S0          # effective duration at the moment the infusion starts
      R <- amt / D            # derived infusion rate

      et <- eventTable() |>
        add.dosing(dose = amt, rate = -2) |>
        add.sampling(times)

      s <- solve(mod, et,
                 inits  = c(state = S0, depot = 0),
                 params = c(dur0 = dur0, ka = ka))

      expected <- .depotAnalytic(times, R, D, ka)

      # The solver must use duration = dur0 + S0, not dur0 alone.
      # Verifying the full trajectory confirms both the duration magnitude and
      # the derived rate (R = amt/(dur0+S0)).
      expect_equal(s$depot[s$time %in% times], expected,
                   tolerance = 1e-3,
                   label = paste0("depot matches analytic solution for dur=",
                                  D, " (S0=", S0, ")"))

      # Confirm the trajectory with S0!=0 differs from the S0=0 baseline,
      # i.e., the state contribution to duration is not silently ignored.
      if (S0 != 0) {
        expected_s0 <- .depotAnalytic(times, amt / dur0, dur0, ka)
        expect_false(isTRUE(all.equal(expected, expected_s0)),
                     label = paste0("S0=", S0,
                                    " gives a different trajectory than state=0"))
      }
    }
  })

  # -- Test 3 -----------------------------------------------------------------
  # Infusion end time must be exactly dur0 + state_at_start.
  # We verify this by checking that depot(D) matches the analytical peak value
  # (R/ka)*(1-exp(-ka*D)).  A wrong duration would shift the peak.
  test_that("state-dep dur: infusion ends at exactly dur0 + state_at_start", {
    mod <- rxode2({
      dur(depot) <- dur0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- 0
    })

    amt  <- 100
    dur0 <- 5
    ka   <- 0.3

    for (S0 in c(0, 3, -1)) {
      D <- dur0 + S0
      R <- amt / D

      # Include the exact stop time D in the sampling grid
      times <- sort(unique(c(seq(0, D + 8, by = 0.1), D)))

      et <- eventTable() |>
        add.dosing(dose = amt, rate = -2) |>
        add.sampling(times)

      s <- solve(mod, et,
                 inits  = c(state = S0, depot = 0),
                 params = c(dur0 = dur0, ka = ka))

      # Analytical depot value at the exact infusion stop time
      depot_at_D_expected <- (R / ka) * (1 - exp(-ka * D))

      # Observed depot at the sample nearest D
      obs_at_D <- s$depot[which.min(abs(s$time - D))]

      # If the duration was wrong the infusion would stop at the wrong time,
      # yielding a depot value that is too large or too small.
      expect_equal(obs_at_D, depot_at_D_expected,
                   tolerance = 1e-3,
                   label = paste0("depot at infusion stop D=", round(D, 4),
                                  " matches analytic (S0=", S0, ")"))
    }
  })

  # -- Test 4 -----------------------------------------------------------------
  # Multiple doses: each dose's duration is evaluated at the state present when
  # THAT dose's infusion begins.  With decaying state, consecutive doses get
  # shorter (or longer) durations.
  # Verify no NAs and that the trajectory differs from the state=0 baseline.
  test_that("state-dep dur: consecutive doses evaluated at their own start state", {
    mod <- rxode2({
      # Duration depends on a decaying state variable, so each dose receives a
      # different infusion duration if spaced far enough apart.
      dur(depot) <- dur0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- -kstate * state
    })

    et <- et() |>
      et(amt = 50, rate = -2, time = 0) |>    # dose 1 - state is large
      et(amt = 50, rate = -2, time = 20) |>   # dose 2 - state has decayed
      et(seq(0, 35, by = 0.5))

    s <- rxSolve(mod, et,
                 params = c(dur0 = 5, ka = 0.4, kstate = 0.2),
                 inits  = c(depot = 0, state = 3))

    # No numerical failures
    expect_false(any(is.na(s$depot)),
                 label = "no NAs in depot with two state-dep dur doses")

    # The trajectory must differ from a state=0 run (different duration per dose)
    s0 <- rxSolve(mod, et,
                  params = c(dur0 = 5, ka = 0.4, kstate = 0.2),
                  inits  = c(depot = 0, state = 0))

    expect_false(isTRUE(all.equal(s$depot, s0$depot)),
                 label = "nonzero initial state gives different depot trajectory than state=0")
  })
})

rxTest({
  mod <- rxode2({
    a <- 6
    b <- 0.6
    d/dt(intestine) <- -a * intestine
    d/dt(blood) <- a * intestine - b * blood
  })

  mod2 <- rxode2({
    a <- 6
    b <- 0.6
    d/dt(intestine) <- -a * intestine
    f(intestine) <- 2
    d/dt(blood) <- a * intestine - b * blood
  })

  mod4 <- rxode2({
    a <- 6
    b <- 0.6
    d/dt(intestine) <- -a * intestine
    f(intestine) <- 0.5
    d/dt(blood) <- a * intestine - b * blood
  })

  mod5 <- rxode2({
    a <- 6
    b <- 0.6
    f <- 1
    d/dt(intestine) <- -a * intestine
    f(intestine) <- f
    d/dt(blood) <- a * intestine - b * blood
  })

  ms <- .methods0
  for (m in ms) {
    et <- eventTable(time.units = "days")
    et$add.sampling(seq(0, 10, by = 1 / 24))
    et$add.dosing(
      dose = 2 / 24, strt.time = 0,
      nbr.doses = 10, dosing.interval = 1
    )

    solve1 <- solve(mod, et, method = m)
    solve2 <- solve(mod2, et, method = m)
    test_that(sprintf("Bioavaibility changes dose (%s)", m), {
      expect_false(all(solve1$intestine == solve2$intestine))
      expect_false(all(solve1$blood == solve2$blood))
    })


    et <- eventTable(time.units = "days")
    et$add.sampling(seq(0, 10, by = 1 / 24))
    et$add.dosing(
      dose = 2 / 24 * 2, strt.time = 0,
      nbr.doses = 10, dosing.interval = 1
    )

    solve3 <- solve(mod, et, method = m)
    test_that(sprintf("F=2 is equivalent to doubling dosing (%s)", m), {
      expect_equal(solve2$intestine, solve3$intestine)
      expect_equal(solve2$blood, solve3$blood)
    })

    et <- eventTable(time.units = "days")
    et$add.sampling(seq(0, 10, by = 1 / 24))
    et$add.dosing(
      dose = 2 / 24 * 0.5, strt.time = 0,
      nbr.doses = 10, dosing.interval = 1
    )

    solve4a <- solve(mod, et, method = m)

    et <- eventTable(time.units = "days")
    et$add.sampling(seq(0, 10, by = 1 / 24))
    et$add.dosing(
      dose = 2 / 24, strt.time = 0,
      nbr.doses = 10, dosing.interval = 1
    )
    test_that(sprintf("F=0.5 is equivalent to halving the dose (%s)", m), {
      solve4b <- solve(mod4, et, method = m)

      expect_equal(solve4a$intestine, solve4b$intestine)
      expect_equal(solve4a$blood, solve4b$blood)
    })

    test_that(sprintf("F=0.5 works with parameter-based F (%s)", m), {
      solve5a <- solve(mod5, et, c(f = 0.5), method = m)

      expect_equal(solve4a$intestine, solve5a$intestine)
      expect_equal(solve4a$blood, solve5a$blood)
    })

    test_that(sprintf("F=2 works with parameter-based F (%s)", m), {
      solve5b <- solve(mod5, et, c(f = 2), method = m)

      expect_equal(solve3$intestine, solve5b$intestine)
      expect_equal(solve3$blood, solve5b$blood)
    })
  }
})

rxTest({
  # https://cran.r-project.org/web/packages/diffEq/vignettes/ODEinR.pdf p11
  mod <- rxode2("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
d/dt(blood)     = a*intestine - b*blood
")

  test_that("Warning for bad dose", {
    et <- eventTable(time.units = "days") |>
      add.sampling(seq(0, 10, by = 1 / 24)) |>
      add.dosing(dose = 2 / 24, rate = 2, start.time = 0, nbr.doses = 10, dosing.interval = 1, dosing.to = 3)

    expect_warning(solve(mod, et), rex::rex("dose to compartment 3 ignored (not in system; 'id=1')"))
  })

  test_that("No Warning for good dose", {
    et <- eventTable(time.units = "days") |>
      add.sampling(seq(0, 10, by = 1 / 24)) |>
      add.dosing(dose = 2 / 24, rate = 2, start.time = 0, nbr.doses = 10, dosing.interval = 1, dosing.to = 1)

    expect_warning(solve(mod, et), NA)
  })
})

rxTest({
  withr::with_tempdir({

    mod <- rxode2({
      d/dt(depot) = -ka * depot
      d/dt(centr) = ka * depot - cl / v * centr
      cp = centr / v
    })

    theta <- c(ka = 1.5, cl = 10, v = 50)

    ev <- et(amt = 100, ii = 24, addl = 2) |>
      et(seq(0, 72, by = 6))

    # Solve normally (reference)
    set.seed(42)
    ref <- rxSolve(mod, theta, ev, serializeFile = NULL)

    # Solve with serialization -- saves state before integration
    stateFile <- tempfile(fileext = ".rxbin")
    set.seed(42)
    ser <- rxSolve(mod, theta, ev, serializeFile = stateFile)

    test_that("rxSolve with serializeFile produces file", {
      expect_equal(ser, stateFile)
    })

    test_that("serialization file is created", {
      expect_true(file.exists(stateFile))
      expect_gt(file.size(stateFile), 0)
    })

    test_that("If file exists, rxSolve throws error with too many arguments", {
      expect_error(rxSolve(mod, theta, ev, serializeFile = stateFile))
    })


    test_that("rxIsSerializeFile detects magic bytes", {
      expect_true(.rxIsSerializeFile(stateFile))
      expect_false(.rxIsSerializeFile(tempfile()))  # non-existent -> FALSE
    })

    mod2 <- rxode2({
      d/dt(centr) = cl / v * centr
      cp = centr / v
    })

    test_that("solve from wrong model throws error", {
      expect_error(rxSolve(mod2, stateFile))
    })

    test_that("serialized solve rejects explicit serializeFile override", {
      expect_error(rxSolve(mod, stateFile, serializeFile = tempfile(fileext = ".rxbin")))
    })

    modFn <- function() {
      d/dt(depot) = -ka * depot
      d/dt(centr) = ka * depot - cl / v * centr
      cp = centr / v
    }

    test_that("serialized function-dispatch replay rejects iCov/keep overrides", {
      expect_error(
        rxSolve(modFn, stateFile, iCov = data.frame(id = 1:4, WT = c(70, 70, 80, 80))),
        "disallowed inputs"
      )
      expect_error(
        rxSolve(modFn, stateFile, keep = "grp"),
        "disallowed inputs"
      )
    })

    # Solve from file -- dispatch via rxSolve(mod, stateFile)
    set.seed(42)
    fromFile <- rxSolve(mod, stateFile)

    test_that("rxSolve(mod, stateFile) produces identical output", {
      expect_equal(ref$cp, fromFile$cp, tolerance = 1e-10)
    })

    test_that("serialized solve has same dimensions", {
      expect_equal(nrow(ref), nrow(fromFile))
      expect_equal(ncol(ref), ncol(fromFile))
    })

    groupedEv <- eventTable()
    groupedEv$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 24)
    groupedEv$add.sampling(seq(0, 48, by = 12))
    groupedEv <- et(groupedEv, id = 1:3)

    groupedRef <- rxSolve(mod, theta, groupedEv)
    groupedStateFile <- tempfile(fileext = ".rxbin")
    rxSolve(mod, theta, groupedEv, serializeFile = groupedStateFile)
    groupedBundle <- .rxReadStateBundle(groupedStateFile)
    groupedFromFile <- rxSolve(mod, groupedStateFile)

    test_that("grouped homogeneous serialize bundle keeps shared event attrs", {
      expect_s3_class(groupedBundle$events, "data.frame")
      expect_equal(attr(groupedBundle$events, "rxHomGroups"), list(1:3))
      expect_equal(attr(groupedBundle$events, "rxHomIdLevels"), c("1", "2", "3"))
      expect_equal(sort(unique(groupedBundle$events$id)), 1L)
    })

    test_that("grouped homogeneous replay from file matches direct solve", {
      expect_equal(
        as.data.frame(groupedRef)[, c("id", "time", "cp")],
        as.data.frame(groupedFromFile)[, c("id", "time", "cp")]
      )
    })

    doseOnlyEv <- eventTable()
    doseOnlyEv$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    doseOnlyEv <- et(doseOnlyEv, id = 1:4)

    doseOnlyRef <- rxSolve(mod, theta, doseOnlyEv, from = 0, to = 24, by = 12)
    doseOnlyStateFile <- tempfile(fileext = ".rxbin")
    rxSolve(mod, theta, doseOnlyEv, from = 0, to = 24, by = 12,
            serializeFile = doseOnlyStateFile)
    doseOnlyBundle <- .rxReadStateBundle(doseOnlyStateFile)
    doseOnlyFromFile <- rxSolve(mod, doseOnlyStateFile)
    doseOnlyTempReplay <- rxSolve(mod, theta, doseOnlyEv, from = 0, to = 24, by = 12,
                                  serializeFile = TRUE)

    test_that("grouped dose-only serialize bundle keeps shared event attrs", {
      expect_s3_class(doseOnlyBundle$events, "data.frame")
      expect_equal(attr(doseOnlyBundle$events, "rxHomGroups"), list(1:4))
      expect_equal(attr(doseOnlyBundle$events, "rxHomIdLevels"), c("1", "2", "3", "4"))
      expect_equal(sort(unique(doseOnlyBundle$events$id)), 1L)
      expect_equal(sum(doseOnlyBundle$events$evid == 0L), 3L)
    })

    test_that("grouped dose-only replay matches direct solve", {
      expect_equal(
        as.data.frame(doseOnlyRef)[, c("id", "time", "cp")],
        as.data.frame(doseOnlyFromFile)[, c("id", "time", "cp")]
      )
    })

    test_that("temporary grouped dose-only serialization replay matches direct solve", {
      expect_equal(
        as.data.frame(doseOnlyRef)[, c("id", "time", "cp")],
        as.data.frame(doseOnlyTempReplay)[, c("id", "time", "cp")]
      )
    })

    modICov <- rxode2({
      wtScale <- WT / 70
      d/dt(depot) = -ka * depot
      d/dt(centr) = ka * depot - cl * wtScale / v * centr
      cp = centr / v
    })
    thetaICov <- c(ka = 1.5, cl = 10, v = 50)
    iCov <- data.frame(id = 1:4, WT = c(70, 70, 80, 80))

    groupedICovEv <- eventTable()
    groupedICovEv$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 24)
    groupedICovEv$add.sampling(seq(0, 48, by = 12))
    groupedICovEv <- et(groupedICovEv, id = 1:4)
    groupedICovFile <- tempfile(fileext = ".rxbin")
    rxSolve(mod, theta, groupedICovEv, serializeFile = groupedICovFile)
    groupedICovBundle <- .rxReadStateBundle(groupedICovFile)
    groupedICovFileForModel <- tempfile(fileext = ".rxbin")
    rxSolve(modICov, thetaICov, groupedICovEv, iCov = iCov, serializeFile = groupedICovFileForModel)

    groupedICovFromBundle <- rxSolve(modICov, thetaICov, groupedICovBundle$events, iCov = iCov)
    groupedICovExpanded <- rxSolve(modICov, thetaICov, as.data.frame(groupedICovEv), iCov = iCov)

    test_that("grouped serialized event data supports direct iCov replay path", {
      expect_equal(
        as.data.frame(groupedICovFromBundle)[, c("id", "time", "cp")],
        as.data.frame(groupedICovExpanded)[, c("id", "time", "cp")]
      )
    })

    groupedKeepFromBundle <- suppressWarnings(
      rxSolve(modICov, thetaICov, groupedICovBundle$events, iCov = transform(iCov, grp = c("a", "a", "b", "b")), keep = "grp")
    )
    groupedKeepExpanded <- suppressWarnings(
      rxSolve(modICov, thetaICov, as.data.frame(groupedICovEv), iCov = transform(iCov, grp = c("a", "a", "b", "b")), keep = "grp")
    )

    test_that("grouped serialized event data keeps iCov-only keep columns on replay", {
      expect_equal(
        as.data.frame(groupedKeepFromBundle)[, c("id", "time", "cp", "grp")],
        as.data.frame(groupedKeepExpanded)[, c("id", "time", "cp", "grp")]
      )
    })

    groupedKeepFromBundleNoId <- suppressWarnings(
      rxSolve(modICov, thetaICov, groupedICovBundle$events,
              iCov = data.frame(WT = c(70, 70, 80, 80), grp = c("a", "a", "b", "b")),
              keep = "grp")
    )
    groupedKeepFromBundleWithId <- suppressWarnings(
      rxSolve(modICov, thetaICov, groupedICovBundle$events,
              iCov = data.frame(id = 1:4, WT = c(70, 70, 80, 80), grp = c("a", "a", "b", "b")),
              keep = "grp")
    )
    groupedKeepFromFileNoId <- suppressWarnings(
      rxSolve(modICov, groupedICovFileForModel,
              iCov = data.frame(WT = c(70, 70, 80, 80), grp = c("a", "a", "b", "b")),
              keep = "grp")
    )
    groupedKeepFromFileWithId <- suppressWarnings(
      rxSolve(modICov, groupedICovFileForModel,
              iCov = data.frame(id = 1:4, WT = c(70, 70, 80, 80), grp = c("a", "a", "b", "b")),
              keep = "grp")
    )

    test_that("grouped serialized bundle replay infers iCov ids when omitted", {
      expect_equal(
        as.data.frame(groupedKeepFromBundleNoId)[, c("id", "time", "cp", "grp")],
        as.data.frame(groupedKeepFromBundleWithId)[, c("id", "time", "cp", "grp")]
      )
    })

    test_that("grouped serialized file replay accepts iCov and infers omitted ids", {
      expect_equal(
        as.data.frame(groupedKeepFromFileNoId)[, c("id", "time", "cp", "grp")],
        as.data.frame(groupedKeepFromFileWithId)[, c("id", "time", "cp", "grp")]
      )
    })

    groupedDoseOnlyICovEv <- eventTable()
    groupedDoseOnlyICovEv$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    groupedDoseOnlyICovEv <- et(groupedDoseOnlyICovEv, id = 1:4)
    groupedDoseOnlyICovFile <- tempfile(fileext = ".rxbin")
    rxSolve(mod, theta, groupedDoseOnlyICovEv, from = 0, to = 24, by = 12,
            serializeFile = groupedDoseOnlyICovFile)
    groupedDoseOnlyICovBundle <- .rxReadStateBundle(groupedDoseOnlyICovFile)

    groupedDoseOnlyKeepFromBundle <- suppressWarnings(
      rxSolve(modICov, thetaICov, groupedDoseOnlyICovBundle$events,
              iCov = transform(iCov, grp = c("a", "a", "b", "b")),
              keep = "grp", from = 0, to = 24, by = 12)
    )
    groupedDoseOnlyKeepExpanded <- suppressWarnings(
      rxSolve(modICov, thetaICov, as.data.frame(groupedDoseOnlyICovEv),
              iCov = transform(iCov, grp = c("a", "a", "b", "b")),
              keep = "grp", from = 0, to = 24, by = 12)
    )
    groupedDoseOnlyKeepTempReplay <- suppressWarnings(
      rxSolve(modICov, thetaICov, groupedDoseOnlyICovEv,
              iCov = transform(iCov, grp = c("a", "a", "b", "b")),
              keep = "grp", from = 0, to = 24, by = 12, serializeFile = TRUE)
    )

    test_that("grouped serialized dose-only events keep iCov-only keep columns on replay", {
      expect_equal(
        as.data.frame(groupedDoseOnlyKeepFromBundle)[, c("id", "time", "cp", "grp")],
        as.data.frame(groupedDoseOnlyKeepExpanded)[, c("id", "time", "cp", "grp")]
      )
    })

    test_that("temporary grouped dose-only serialization replay keeps iCov-only keep columns", {
      expect_equal(
        as.data.frame(groupedDoseOnlyKeepTempReplay)[, c("id", "time", "cp", "grp")],
        as.data.frame(groupedDoseOnlyKeepExpanded)[, c("id", "time", "cp", "grp")]
      )
    })
  })
})

if (tolower(Sys.info()[["sysname"]]) == "linux") {
  rxTest({

    test_that("table step for linCmtB", {

      skip_if_not_installed("nlmixr2data")
      pars <- test_path("lincmt-solve-focei-sol.rds")
      skip_if_not(file.exists(pars))
      pars <- readRDS(pars)

      rx <- rxode2({
        param(THETA[1], THETA[2], THETA[3], THETA[4], ETA[1], ETA[2],
              ETA[3])
        rx_yj_ ~ 2
        rx_lambda_ ~ 1
        rx_hi_ ~ 1
        rx_low_ ~ 0
        rx_expr_0 ~ ETA[2] + THETA[2]
        rx_expr_1 ~ ETA[3] + THETA[3]
        rx_expr_2 ~ ETA[1] + THETA[1]
        rx_expr_3 ~ exp(rx_expr_0)
        rx_expr_4 ~ exp(rx_expr_1)
        rx_expr_5 ~ exp(rx_expr_2)
        rx_pred_ = linCmtB(rx__PTR__, t, 2, 1, 1, -1, -1, 1, rx_expr_3,
                           rx_expr_4, 0, 0, 0, 0, rx_expr_5)
        rx__sens_rx_pred__BY_ETA_1___ = rx_expr_5 *
          linCmtB(rx__PTR__,
                  t, 2, 1, 1, -2, 2, 1, rx_expr_3, rx_expr_4, 0, 0, 0,
                  0, rx_expr_5)
        rx__sens_rx_pred__BY_ETA_2___ = rx_expr_3 *
          linCmtB(rx__PTR__,
                  t, 2, 1, 1, -2, 0, 1, rx_expr_3, rx_expr_4, 0, 0, 0,
                  0, rx_expr_5)
        rx__sens_rx_pred__BY_ETA_3___ = rx_expr_4 *
          linCmtB(rx__PTR__,
                  t, 2, 1, 1, -2, 1, 1, rx_expr_3, rx_expr_4, 0, 0, 0,
                  0, rx_expr_5)
        rx_r_ = Rx_pow_di(THETA[4], 2)
        rx__sens_rx_r__BY_ETA_1___ = 0
        rx__sens_rx_r__BY_ETA_2___ = 0
        rx__sens_rx_r__BY_ETA_3___ = 0
        cmt(rxLinCmt)
        dvid(3)
      })


      expect_error(rxSolve(rx, pars,
                           nlmixr2data::theo_sd),
                   NA)

    })

    test_that("mixed ode/linCmt() zero observation issue(s)", {

      rxWithSeed(1, {

        pk <- c(cl=0.2,v2=3.5,q=.4,v3=3.5,ka=.2,f=.7)
        nn <- 100
        rands <- matrix(runif(4*nn),nn)
        pdpars <- dplyr::tibble(ec50=(300/28*pk["f"]/pk["cl"])*(0.1+0.9*rands[,1]),emax=-0.9+10.9*rands[,2],
                                gamma=1+3*rands[,3],ke0=log(2)/(5+45*rands[,4]))
        rxmod1 <- RxODE({
          Cp       <- linCmt(ka,cl,v2,v3,q)
          d/dt(Ce) <- (Cp-Ce)*ke0
          eff      <- 1*(1+emax*Ce**gamma/(ec50**gamma+Ce**gamma))
        })

        et1 <- et() |>
          et(c(seq(0,7*7,.2),seq(7*7,52*7,1))) |> ## sampling
          add.dosing(dose=300*pk["f"],dosing.to=1,nbr.doses=13,dosing.interval=28,start.time=0)  ## dosing

          res1 <- rxSolve(rxmod1,cbind(as.list(pk),pdpars),et1)

          ### Note 1: bug occurs when including several sets of parameters
          res1 <- rxSolve(rxmod1,cbind(as.list(pk),pdpars),et1)

          expect_length(res1 |> dplyr::filter(time>0 & Cp==0) |> dplyr::pull(time),0)

          ### Note 2: Bug also occurs when simulating one set of parameters at a time
          res2 <- do.call("rbind", lapply(1:nn, function(x) {
            rxSolve(rxmod1,unlist(c(pk,as.data.frame(pdpars[x,]))),et1)
          }))

          expect_length(res2 |> dplyr::filter(time>0 & Cp==0) |> dplyr::pull(time), 0)

      })
    })

  })

  rxTest({

    tol <- 5e-5 ## Current difference for all equations
    types <- 1:6

    for (type in types) {

      .txt <- switch(type,
                     "linear", #1
                     "sensitivity AD", #2
                     "sensitivity forward", #3
                     "sensitivity central", #4
                     "sensitivity forward3", #5
                     "sensitivity endpoint5") #6

      sens <- switch(type,
                     "linCmtA", # 1
                     "linCmtB", # 2
                     "linCmtB", # 3
                     "linCmtB", # 4
                     "linCmtB", # 5
                     "linCmtB") # 6

      linCmtSensType <- switch(type,
                               "AD", #1
                               "AD", #2
                               "forward", #3
                               "central", #4
                               "forward3", #5
                               "endpoint5") #6

      etSsB <- et() |>
        et(amt = 3) |>
        et(time = 4, amt = 3, ss = 1, ii = 24) |>
        et(amt = 3, ss = 2, ii = 24, time = 8) |>
        et(seq(0, 24, length.out = 200))

      etSsI <- et() |>
        et(amt = 3, rate = 1.5) |>
        et(time = 4, amt = 3, rate = 1.5, ss = 1, ii = 24) |>
        et(time = 8, amt = 3, rate = 1.5, ss = 2, ii = 24) |>
        et(seq(0, 24, length.out = 200))

      etSsR <- et(amt = 0, ss = 1, rate = 10000 / 8)

      ode.1c <- rxode2({
        C2 <- center / V
        d / dt(center) <- -CL * C2
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType)

      sol.1c <- rxode2({
        C2 <- linCmt(CL, V)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType)

      # context(sprintf("Test steady state solutions 1 cmt (%s)", .txt))

      o1 <- rxSolve(ode.1c, params = c(V = 20, CL = 25), events = etSsB)
      s1 <- rxSolve(sol.1c, params = c(V = 20, CL = 25), events = etSsB)

      test_that(sprintf("one compartment bolus steady state (%s)", .txt), {
        expect_equal(o1$C2, s1$C2, tolerance = tol)
      })

      o1 <- rxSolve(ode.1c, params = c(V = 20, CL = 25), events = etSsI, addDosing = TRUE, returnType="data.frame")

      expect_true("rate" %in% names(o1))

      o1 <- rxSolve(ode.1c, params = c(V = 20, CL = 25), events = etSsI)
      s1 <- rxSolve(sol.1c, params = c(V = 20, CL = 25), events = etSsI)
      test_that(sprintf("one compartment infusion tau steady state (%s)", .txt), {
        expect_equal(o1$C2, s1$C2, tolerance = tol)
      })

      o1 <- rxSolve(ode.1c, params = c(V = 20, CL = 25), events = etSsR,
                    addDosing = TRUE, returnType = "data.frame")

      expect_true("rate" %in% names(o1))
      expect_true("ss" %in% names(o1))
      expect_true("ii" %in% names(o1))

      o1 <- rxSolve(ode.1c, params = c(V = 20, CL = 25), events = etSsR)
      s1 <- rxSolve(sol.1c, params = c(V = 20, CL = 25), events = etSsR)
      test_that(sprintf("one compartment infusion steady state (%s)", .txt), {
        expect_equal(o1$C2, s1$C2, tolerance = tol)
      })

      # context(sprintf("Test steady state solutions 2 cmt (%s)", .txt))

      ode.2c <- rxode2(
      {
        C2 <- centr / V
        C3 <- peri / V2
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType)

      sol.2c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q1)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      o2 <- ode.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10), events = etSsB)
      s2 <- sol.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q1 = 10), events = etSsB)
      test_that("two compartment bolus steady state", {
        expect_equal(o2$C2, s2$C2, tolerance = tol)
      })

      o2 <- ode.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10), events = etSsI)
      s2 <- sol.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q1 = 10), events = etSsI)
      test_that("two compartment infusion steady state, tau", {
        expect_equal(o2$C2, s2$C2, tolerance = tol)
      })

      o2 <- ode.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10), events = etSsR)
      s2 <- sol.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q1 = 10), events = etSsR)
      test_that("two compartment infusion steady state", {
        expect_equal(o2$C2, s2$C2, tolerance = tol)
      })

      # context(sprintf("Test steady state solutions 3 cmt (%s)", .txt))

      ode.3c <- rxode2(
      {
        C2 <- centr / V
        C3 <- peri / V2
        C4 <- peri2 / V3
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(peri2) <- Q2 * C2 - Q2 * C4
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      sol.3c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, Q2, V3)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      o3 <- ode.3c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = etSsB)
      s3 <- sol.3c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = etSsB)
      test_that("three compartment bolus steady state", {
        expect_equal(o3$C2, s3$C2, tolerance = tol)
      })

      o3 <- ode.3c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = etSsI)
      s3 <- sol.3c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = etSsI)
      test_that("three compartment bolus steady state", {
        expect_equal(o3$C2, s3$C2, tolerance = tol)
      })

      o3 <- ode.3c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = etSsR)
      s3 <- sol.3c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = etSsR)
      test_that("three compartment bolus steady state", {
        expect_equal(o3$C2, s3$C2, tolerance = tol)
      })

      # context(sprintf("Test steady state solutions 1 cmt ka (%s)", .txt))

      ode.1c.ka <- rxode2(
      {
        C2 <- center / V
        d / dt(depot) <- -KA * depot
        d / dt(center) <- KA * depot - CL * C2
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      sol.1c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, KA)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      o1 <- ode.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsB)
      s1 <- sol.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsB)
      test_that(sprintf("one compartment bolus steady state to depot compartment (%s)",
                        .txt), {
                          expect_equal(o1$C2, s1$C2, tolerance = tol)
                        })

      o1 <- ode.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsI)
      s1 <- sol.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsI)
      test_that(sprintf("one compartment infusion steady state to depot compartment, tau (%s)", .txt), {
        expect_equal(o1$C2, s1$C2, tolerance = tol)
      })

      o1 <- ode.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsR)
      s1 <- sol.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsR)
      test_that(sprintf("one compartment infusion steady state to depot compartment (%s)", .txt), {
        expect_equal(o1$C2, s1$C2, tolerance = tol)
      })

      etSsB2 <- et() |>
        et(amt = 3, cmt = 2) |>
        et(time = 4, amt = 3, ss = 1, ii = 24, cmt = 2) |>
        et(amt = 3, ss = 2, ii = 24, time = 8, cmt = 2) |>
        et(seq(0, 24, length.out = 200))

      etSsI2 <- et() |>
        et(amt = 3, rate = 1.5, cmt = 2) |>
        et(time = 4, amt = 3, rate = 1.5, ss = 1, ii = 24, cmt = 2) |>
        et(time = 8, amt = 3, rate = 1.5, ss = 2, ii = 24, cmt = 2) |>
        et(seq(0, 24, length.out = 200))

      etSsR2 <- et(amt = 0, ss = 1, rate = 10000 / 8, cmt = 2)

      o1 <- ode.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsB2)
      s1 <- sol.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsB2)
      test_that("one compartment bolus steady state to central compartment", {
        expect_equal(o1$C2, s1$C2, tolerance = tol)
      })

      o1 <- ode.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsI2)
      s1 <- sol.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsI2)
      test_that("one compartment infusion steady state to central compartment, tau", {
        expect_equal(o1$C2, s1$C2, tolerance = tol)
      })

      o1 <- ode.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsR2)
      s1 <- sol.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = etSsR2)
      test_that("one compartment infusion steady state to central compartment", {
        expect_equal(o1$C2, s1$C2, tolerance = tol)
      })

      # context(sprintf("Test steady state solutions 2 cmt ka (%s)", .txt))

      ode.2c.ka <- rxode2(
      {
        C2 <- centr / V
        C3 <- peri / V2
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      sol.2c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, KA)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      o2 <- ode.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsB)
      s2 <- sol.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsB)
      test_that(sprintf("two compartment bolus steady state to depot compartment (%s)", .txt), {
        expect_equal(o2$C2, s2$C2, tolerance = tol)
      })

      o2 <- ode.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsI)
      s2 <- sol.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsI)
      test_that(sprintf("two compartment infusion steady state to depot compartment, tau (%s)", .txt), {
        expect_equal(o2$C2, s2$C2, tolerance = tol)
      })


      o2 <- ode.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsR)
      s2 <- sol.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsR)
      test_that(sprintf("two compartment infusion steady state to depot compartment, tau (%s)", .txt), {
        expect_equal(o2$C2, s2$C2, tolerance = tol)
      })

      o2 <- ode.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsB2)
      s2 <- sol.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsB2)
      test_that("two compartment bolus steady state to central compartment", {
        expect_equal(o2$C2, s2$C2, tolerance = tol)
      })

      o2 <- ode.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsI2)
      s2 <- sol.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsI2)
      test_that("two compartment infusion steady state to central compartment, tau", {
        expect_equal(o2$C2, s2$C2, tolerance = tol)
      })

      o2 <- ode.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsR2)
      s2 <- sol.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = etSsR2)
      test_that("two compartment infusion steady state to central compartment", {
        expect_equal(o2$C2, s2$C2, tolerance = tol)
      })

      # context(sprintf("Test steady state solutions 3 cmt ka (%s)", .txt))

      ode.3c.ka <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        C4 <- peri2 / V3
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(peri2) <- Q2 * C2 - Q2 * C4
      })

      sol.3c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, Q2, V3, KA)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      o3 <- ode.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsB)
      s3 <- sol.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsB)
      test_that(sprintf("three compartment bolus steady state to depot compartment (%s)",
                        .txt), {
                          expect_equal(o3$C2, s3$C2, tolerance = tol)
                        })

      o3 <- ode.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsI)
      s3 <- sol.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsI)
      test_that(sprintf("three compartment infusion steady state to depot compartment, tau (%s)", .txt), {
        expect_equal(o3$C2, s3$C2, tolerance = tol)
      })

      o3 <- ode.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsR)
      s3 <- sol.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsR)
      test_that("three compartment infusion steady state to depot compartment", {
        expect_equal(o3$C2, s3$C2, tolerance = tol)
      })

      ## B2
      o3 <- ode.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsB2)
      s3 <- sol.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsB2)
      test_that("three compartment bolus steady state to central compartment", {
        expect_equal(o3$C2, s3$C2, tolerance = tol)
      })

      o3 <- ode.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsI2)
      s3 <- sol.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsI2)
      test_that("three compartment infusion steady state to central compartment, tau", {
        expect_equal(o3$C2, s3$C2, tolerance = tol)
      })

      o3 <- ode.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsR2)
      s3 <- sol.3c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSsR2)
      test_that("three compartment infusion steady state to central compartment", {
        expect_equal(o3$C2, s3$C2, tolerance = tol)
      })

      # context(sprintf("Test the solved equations 1 cmt (%s)", .txt))

      et <- eventTable() |>
        add.dosing(dose = 3, nbr.doses = 6, dosing.interval = 8) |>
        add.sampling(seq(0, 48, length.out = 200))

      ode.1c <- rxode2({
        C2 <- center / V
        d / dt(center) <- -CL * C2
      })

      goodP <- function(model, cmt = 1L, ka = 0L) {
        test_that(sprintf("model '%s' parses to cmt=%d, ka=%d", as.character(substitute(model)), cmt, ka), {
          .flags <- rxModelVars(model)$flags
          expect_equal(setNames(.flags["ncmt"], NULL), cmt)
          expect_equal(setNames(.flags["ka"], NULL), ka)
        })
      }

      ## Solved systems can check the variables in the rxode2 statement
      ## to figure out what type of solved system is being requested
      ode.1cs <- rxode2(
      {
        V <- theta[1]
        CL <- theta[2]
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.1cs)

      # context(sprintf("Test the solved equations 2 cmt (%s)", .txt))

      ode.2cK <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        K <- CLx / V
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.2cK)

      ode.2cA1 <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        alpha <- CLx / V
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.2cA1)

      ode.2cA2 <- rxode2(
      {
        A <- 1 / theta[1]
        CLx <- theta[2]
        alpha <- CLx * A
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.2cA2)

      ## Instead of specifying parameters in the solved system, you can
      ## specify them in the linCmt variable.
      ode.1cs2 <- rxode2(
      {
        C2 <- linCmt(CL, V)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.1cs2)

      ## The solved systems can be mixed with ODE solving routines (to
      ## speed them up a bit...?)

      o.1c <- ode.1c |> solve(params = c(V = 20, CL = 25), events = et)

      s.1c <- ode.1cs2 |> solve(params = c(V = 20, CL = 25), events = et)

      s.2c <- ode.1cs |> solve(theta = c(20, 25), events = et)

      s.2cK <- ode.2cK |> solve(theta = c(20, 25), events = et)
      s.2cA1 <- ode.2cA1 |> solve(theta = c(20, 25), events = et)
      s.2cA2 <- ode.2cA2 |> solve(theta = c(20, 25), events = et)

      test_that("Gives the correct parameters for THETAs", {
        expect_equal(
          s.2c$params,
          structure(list("THETA[1]" = 20, "THETA[2]" = 25),
                    class = "data.frame",
                    row.names = c(NA, -1L)
                    )
        )
      })

      test_that("1 compartment solved models and ODEs same.", {
        expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
        expect_equal(o.1c$C2, s.2c$C2, tolerance = tol)
        expect_equal(o.1c$C2, s.2cK$C2, tolerance = tol)
        expect_equal(o.1c$C2, s.2cA1$C2, tolerance = tol)
        expect_equal(o.1c$C2, s.2cA2$C2, tolerance = tol)
      })

      ## Test steady state doses.
      etSs <- et() |>
        et(amt = 3) |>
        et(time = 4, amt = 3, ss = 1, ii = 24) |>
        et(amt = 3, ss = 2, ii = 24, time = 8) |>
        et(seq(0, 24, length.out = 200))

      o.1c <- ode.1c |> solve(params = c(V = 20, CL = 1), events = etSs)

      s.1c <- ode.1cs2 |> solve(params = c(V = 20, CL = 1), events = etSs)

      s.2c <- ode.1cs |> solve(theta = c(20, 1), events = etSs)

      test_that("1 compartment steady-state solved models and ODEs same.", {
        expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
        expect_equal(o.1c$C2, s.2c$C2, tolerance = tol)
      })

      ode.1c.ka <- rxode2({
        C2 <- center / V
        d / dt(depot) <- -KA * depot
        d / dt(center) <- KA * depot - CL * C2
      })

      sol.1c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, KA)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.1c.ka, ka = 1L)

      ode.2cK <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        Ka <- theta[3]
        K <- CLx / V
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.2cK, ka = 1L)

      ode.2cA1 <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        Ka <- theta[3]
        alpha <- CLx / V
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.2cA1, ka = 1L)

      ode.2cA2 <- rxode2(
      {
        A <- 1 / theta[1]
        CLx <- theta[2]
        Ka <- theta[3]
        alpha <- CLx * A
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.2cA2, ka = 1L)

      o.1c <- ode.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = et)

      s.1c <- sol.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = et)

      s.2cK <- ode.2cK |> solve(theta = unname(c(20, 25, KA = 2)), events = et)
      s.2cA1 <- ode.2cA1 |> solve(theta = unname(c(20, 25, KA = 2)), events = et)
      s.2cA2 <- ode.2cA2 |> solve(theta = unname(c(20, 25, KA = 2)), events = et)

      test_that("1 compartment oral solved models and ODEs same.", {
        expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
        expect_equal(o.1c$C2, s.2cK$C2, tolerance = tol)
        expect_equal(o.1c$C2, s.2cA1$C2, tolerance = tol)
        expect_equal(o.1c$C2, s.2cA2$C2, tolerance = tol)
      })

      ## Note the strange-looking dip at 4 hours.  This is because ss=1 resets the system first.
      o.1c <- ode.1c.ka |> solve(params = c(V = 20, CL = 2, KA = 2), events = etSs)

      s.1c <- sol.1c.ka |> solve(params = c(V = 20, CL = 2, KA = 2), events = etSs)

      test_that("1 compartment oral solved models steady state ODEs same.", {
        expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
      })

      ode.2c <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
      })

      sol.2c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q1)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2c, cmt = 2L)

      sol.2cK <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        K <- CLx / V
        K12 <- Qx / V
        K21 <- Qx / V2x
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2cK, cmt = 2L)

      ## A1 in terms of A, alpha, B, beta

      sol.2cA1 <- rxode2(
      {
        Vx <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        Kx <- CLx / Vx
        K12x <- Qx / Vx
        K21x <- Qx / V2x
        beta <- 0.5 * (K12x + K21x + Kx -
                         sqrt((K12x +
                                 K21x + Kx) * (K12x + K21x + Kx) - 4 * K21x *
                                Kx))
        alpha <- K21x * Kx / beta
        A <- (alpha - K21x) / (alpha - beta) / Vx
        B <- (beta - K21x) / (beta - alpha) / Vx
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2cA1, cmt = 2L)

      ## A2 V, alpha, beta, k21
      sol.2cA2 <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        Kx <- CLx / V
        K12x <- Qx / V
        K21 <- Qx / V2x
        beta <- 0.5 * (K12x + K21 + Kx -
                         sqrt((K12x +
                                 K21 + Kx) * (K12x + K21 + Kx) - 4 * K21 *
                                Kx))
        alpha <- K21 * Kx / beta
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2cA2, cmt = 2L)

      ## A3 alpha, beta, aob
      sol.2cA3 <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        Kx <- CLx / V
        K12x <- Qx / V
        K21x <- Qx / V2x
        beta <- 0.5 * (K12x + K21x + Kx -
                         sqrt((K12x +
                                 K21x + Kx) * (K12x + K21x + Kx) - 4 * K21x *
                                Kx))
        alpha <- K21x * Kx / beta
        Ax <- (alpha - K21x) / (alpha - beta) / V
        Bx <- (beta - K21x) / (beta - alpha) / V
        aob <- Ax / Bx
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2cA3, cmt = 2L)

      o.2c <- ode.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10), events = et)

      s.2c <- sol.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q1 = 10), events = et)

      s.2cK <- sol.2cK |> solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10)), events = et)

      s.2cA1 <- sol.2cA1 |> solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10)), events = et)

      s.2cA2 <- sol.2cA2 |> solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10)), events = et)
      s.2cA3 <- sol.2cA3 |> solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10)), events = et)

      test_that("2 compartment solved models and ODEs same.", {
        expect_equal(s.2cK$C2, s.2c$C2, tolerance = tol)
        expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)
        expect_equal(o.2c$C2, s.2cK$C2, tolerance = tol)
        expect_equal(o.2c$C2, s.2cA1$C2, tolerance = tol)
        expect_equal(o.2c$C2, s.2cA2$C2, tolerance = tol)
        expect_equal(o.2c$C2, s.2cA3$C2, tolerance = tol)
      })

      ode.2c.ka <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
      })

      sol.2c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, KA)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2c.ka, cmt = 2, ka = 1)

      sol.2cK <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        ka <- theta[5]
        K <- CLx / V
        K12 <- Qx / V
        K21 <- Qx / V2x
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2cK, cmt = 2, ka = 1)

      ## A1 in terms of A, alpha, B, beta

      sol.2cA1 <- rxode2(
      {
        Vx <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        ka <- theta[5]
        Kx <- CLx / Vx
        K12x <- Qx / Vx
        K21x <- Qx / V2x
        beta <- 0.5 * (K12x + K21x + Kx -
                         sqrt((K12x +
                                 K21x + Kx) * (K12x + K21x + Kx) - 4 * K21x *
                                Kx))
        alpha <- K21x * Kx / beta
        A <- (alpha - K21x) / (alpha - beta) / Vx
        B <- (beta - K21x) / (beta - alpha) / Vx
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2cA1, cmt = 2, ka = 1)

      ## A2 V, alpha, beta, k21
      sol.2cA2 <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        ka <- theta[5]
        Kx <- CLx / V
        K12x <- Qx / V
        K21 <- Qx / V2x
        beta <- 0.5 * (K12x + K21 + Kx -
                         sqrt((K12x +
                                 K21 + Kx) * (K12x + K21 + Kx) - 4 * K21 *
                                Kx))
        alpha <- K21 * Kx / beta
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2cA2, cmt = 2, ka = 1)

      ## A3 alpha, beta, aob
      sol.2cA3 <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        ka <- theta[5]
        Kx <- CLx / V
        K12x <- Qx / V
        K21x <- Qx / V2x
        beta <- 0.5 * (K12x + K21x + Kx -
                         sqrt((K12x +
                                 K21x + Kx) * (K12x + K21x + Kx) - 4 * K21x *
                                Kx))
        alpha <- K21x * Kx / beta
        Ax <- (alpha - K21x) / (alpha - beta) / V
        Bx <- (beta - K21x) / (beta - alpha) / V
        aob <- Ax / Bx
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2cA3, cmt = 2, ka = 1)

      sol.2cSS <- rxode2(
      {
        V <- theta[1]
        CL <- theta[2]
        V2x <- theta[3]
        Q <- theta[4]
        Ka <- theta[5]
        Vss <- V + V2x
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2cSS, cmt = 2, ka = 1)

      sol.2cT <- rxode2(
      {
        V <- theta[1]
        CL <- theta[2]
        VT <- theta[3]
        Q <- theta[4]
        Ka <- theta[5]
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2cT, cmt = 2, ka = 1)

      o.2c <- ode.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = et)
      s.2c <- sol.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = et)
      s.2cK <- sol.2cK |> solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, ka = 0.3)), events = et)
      s.2cA1 <- sol.2cA1 |> solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, ka = 0.3)), events = et)
      s.2cA2 <- sol.2cA2 |> solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, ka = 0.3)), events = et)
      s.2cA3 <- sol.2cA3 |> solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, ka = 0.3)), events = et)
      s.2cSS <- sol.2cSS |> solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, ka = 0.3)), events = et)
      s.2cT <- sol.2cT |> solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, ka = 0.3)), events = et)

      test_that("2 compartment oral solved models and ODEs same.", {
        expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)
        expect_equal(o.2c$C2, s.2cK$C2, tolerance = tol)
        expect_equal(o.2c$C2, s.2cA1$C2, tolerance = tol)
        expect_equal(o.2c$C2, s.2cA2$C2, tolerance = tol)
        expect_equal(o.2c$C2, s.2cA3$C2, tolerance = tol)
        expect_equal(o.2c$C2, s.2cSS$C2, tolerance = tol)
        expect_equal(o.2c$C2, s.2cT$C2, tolerance = tol)
      })

      o.2c <- ode.2c.ka |> solve(params = c(V = 40, CL = 1, V2 = 297, Q = 10, KA = 0.3), events = etSs)

      s.2c <- sol.2c.ka |> solve(params = c(V = 40, CL = 1, V2 = 297, Q = 10, KA = 0.3), events = etSs)

      test_that("2 compartment oral steady-state solved models and ODEs same.", {
        expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)
      })

      # context(sprintf("Test the solved equations 3 cmt (%s)", .txt))

      ode.3c <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        C4 <- peri2 / V3
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(peri2) <- Q2 * C2 - Q2 * C4
      })

      sol.3c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, Q2, V3)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3c, 3)

      sol.3cK <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        Q2x <- theta[5]
        V3x <- theta[6]
        K <- CLx / V
        K12 <- Qx / V
        K21 <- Qx / V2x
        k13 <- Q2x / V
        k31 <- Q2x / V3x
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3cK, 3)

      sol.3cA1 <- rxode2(
      {
        Vx <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        Q2x <- theta[5]
        V3x <- theta[6]
        Kx <- CLx / Vx
        K12x <- Qx / Vx
        K21x <- Qx / V2x
        K13x <- Q2x / Vx
        K31x <- Q2x / V3x
        a0 <- Kx * K21x * K31x
        a1 <- Kx * K31x + K21x * K31x + K21x * K13x +
          Kx * K21x + K31x * K12x
        a2 <- Kx + K12x + K13x + K21x + K31x
        p <- a1 - a2 * a2 / 3
        qq <- 2 * a2 * a2 * a2 / 27 - a1 * a2 / 3 + a0
        r1 <- sqrt(-p * p * p / 27)
        r2 <- 2 * r1^(1 / 3)
        theta <- acos(-qq / (2 * r1)) / 3
        alpha <- -(cos(theta) * r2 - a2 / 3)
        beta <- -(cos(theta + 2 / 3 * pi) * r2 - a2 / 3)
        gamma <- -(cos(theta + 4 / 3 * pi) * r2 - a2 / 3)
        A <- (K21x - alpha) * (K31x - alpha) / (alpha - beta) / (alpha - gamma) / Vx
        B <- (K21x - beta) * (K31x - beta) / (beta - alpha) / (beta - gamma) / Vx
        C <- (K21x - gamma) * (K31x - gamma) / (gamma - alpha) / (gamma - beta) / Vx
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3cA1, 3)

      sol.3cVp <- rxode2(
      {
        V <- theta[1]
        CL <- theta[2]
        Vp <- theta[3]
        Q <- theta[4]
        Q2 <- theta[5]
        Vp2 <- theta[6]
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3cVp, 3)

      sol.3cVt <- rxode2(
      {
        V <- theta[1]
        CL <- theta[2]
        Vt <- theta[3]
        Q <- theta[4]
        Q2 <- theta[5]
        Vt2 <- theta[6]
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3cVt, 3)

      o.3c <- ode.3c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = et)

      s.3c <- sol.3c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = et)

      s.3cK <- sol.3cK |>
        solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400)), events = et)
      s.3cA1 <- sol.3cA1 |>
        solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400)), events = et)
      s.3cVp <- sol.3cVp |>
        solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400)), events = et)
      s.3cVt <- sol.3cVt |>
        solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400)), events = et)

      test_that("3 compartment solved models and ODEs same.", {
        expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
        expect_equal(o.3c$C2, s.3cK$C2, tolerance = tol)
        expect_equal(o.3c$C2, s.3cA1$C2, tolerance = tol)
        expect_equal(o.3c$C2, s.3cVp$C2, tolerance = tol)
        expect_equal(o.3c$C2, s.3cVt$C2, tolerance = tol)
      })

      o.3c <- ode.3c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = etSs)

      s.3c <- sol.3c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = etSs)

      test_that("3 compartment solved models and ODEs same with steady state.", {
        expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
      })

      ode.3c.ka <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        C4 <- peri2 / V3
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(peri2) <- Q2 * C2 - Q2 * C4
      })

      sol.3c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, Q2, V3, KA)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3c.ka, 3, 1)

      sol.3cK <- rxode2(
      {
        V <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        Q2x <- theta[5]
        V3x <- theta[6]
        ka <- theta[7]
        K <- CLx / V
        K12 <- Qx / V
        K21 <- Qx / V2x
        k13 <- Q2x / V
        k31 <- Q2x / V3x
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3cK, 3, 1)

      sol.3cA1 <- rxode2(
      {
        Vx <- theta[1]
        CLx <- theta[2]
        V2x <- theta[3]
        Qx <- theta[4]
        Q2x <- theta[5]
        V3x <- theta[6]
        ka <- theta[7]
        Kx <- CLx / Vx
        K12x <- Qx / Vx
        K21x <- Qx / V2x
        K13x <- Q2x / Vx
        K31x <- Q2x / V3x
        a0 <- Kx * K21x * K31x
        a1 <- Kx * K31x + K21x * K31x + K21x * K13x +
          Kx * K21x + K31x * K12x
        a2 <- Kx + K12x + K13x + K21x + K31x
        p <- a1 - a2 * a2 / 3
        qq <- 2 * a2 * a2 * a2 / 27 - a1 * a2 / 3 + a0
        r1 <- sqrt(-p * p * p / 27)
        r2 <- 2 * r1^(1 / 3)
        theta <- acos(-qq / (2 * r1)) / 3
        alpha <- -(cos(theta) * r2 - a2 / 3)
        beta <- -(cos(theta + 2 / 3 * pi) * r2 - a2 / 3)
        gamma <- -(cos(theta + 4 / 3 * pi) * r2 - a2 / 3)
        A <- (K21x - alpha) * (K31x - alpha) / (alpha - beta) / (alpha - gamma) / Vx
        B <- (K21x - beta) * (K31x - beta) / (beta - alpha) / (beta - gamma) / Vx
        C <- (K21x - gamma) * (K31x - gamma) / (gamma - alpha) / (gamma - beta) / Vx
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3cA1, 3, 1)

      o.3c <- ode.3c.ka |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = et)
      s.3c <- sol.3c.ka |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = et)
      s.3cK <- sol.3cK |>
        solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3)), events = et)
      s.3cA1 <- sol.3cA1 |>
        solve(theta = unname(c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3)), events = et)

      test_that("3 compartment oral solved models and ODEs same.", {
        expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
        expect_equal(o.3c$C2, s.3cK$C2, tolerance = tol)
        expect_equal(o.3c$C2, s.3cA1$C2, tolerance = tol)
      })

      o.3c <- ode.3c.ka |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSs)
      s.3c <- sol.3c.ka |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = etSs)

      ## Again the 4 hour strange discontinuity because ss=1
      test_that("3 compartment oral solved models and ODEs same for steady state.", {
        expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
      })

      # context(sprintf("Infusion Models 1 cmt (%s)", .txt))

      et <- eventTable() |>
        add.dosing(dose = 3, rate = 1.5, nbr.doses = 6, dosing.interval = 8) |>
        add.sampling(seq(0, 48, length.out = 200))

      etSs <- et() |>
        et(amt = 3, rate = 1.5) |>
        et(time = 4, amt = 3, rate = 1.5, ss = 1, ii = 24) |>
        et(time = 8, amt = 3, rate = 1.5, ss = 2, ii = 24) |>
        et(seq(0, 24, length.out = 200))

      ode.1c <- rxode2({
        C2 <- center / V
        d / dt(center) <- -CL * C2
      })

      ## Solved systems can check the variables in the rxode2 statement
      ## to figure out what type of solved system is being requested
      ode.1cs <- rxode2(
      {
        V <- theta[1]
        CL <- theta[2]
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.1cs, 1)

      ## Instead of specifying parameters in the solved system, you can
      ## specify them in the linCmt variable.
      ode.1cs2 <- rxode2(
      {
        C2 <- linCmt(CL, V)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.1cs2, 1)

      ## The solved systems can be mixed with ODE solving routines (to
      ## speed them up a bit...?)

      o.1c <- ode.1c |>
        solve(params = c(V = 20, CL = 25), events = et)
      s.1c <- ode.1cs2 |>
        solve(params = c(V = 20, CL = 25), events = et)
      s.2c <- ode.1cs |>
        solve(theta = c(20, 25), events = et)

      test_that("1 compartment solved models and ODEs same.", {
        expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
        expect_equal(o.1c$C2, s.2c$C2, tolerance = tol)
      })

      o.1c <- ode.1c |>
        solve(params = c(V = 20, CL = 10), events = etSs)
      s.1c <- ode.1cs2 |>
        solve(params = c(V = 20, CL = 10), events = etSs)
      s.2c <- ode.1cs |>
        solve(theta = c(20, 10), events = etSs)

      test_that("1 compartment solved models and ODEs same; Steady State", {
        expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
        expect_equal(o.1c$C2, s.2c$C2, tolerance = tol)
      })

      # context(sprintf("Infusion Models 2 cmt (%s)", .txt))

      ode.2c <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
      })

      sol.2c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2c, 2)

      o.2c <- ode.2c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10), events = et)
      s.2c <- sol.2c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10), events = et)

      test_that("2 compartment solved models and ODEs same.", {
        expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)
      })

      o.2c <- ode.2c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10), events = etSs)
      s.2c <- sol.2c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10), events = etSs)

      test_that("2 compartment steady state solved models and ODEs same.", {
        expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)
      })

      # context(sprintf("Infusion Models 3 cmt (%s)", .txt))

      ode.3c <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        C4 <- peri2 / V3
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(peri2) <- Q2 * C2 - Q2 * C4
      })

      sol.3c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, Q2, V3)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3c, 3)

      o.3c <- ode.3c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = et)
      s.3c <- sol.3c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = et)

      test_that("3 compartment solved models and ODEs same.", {
        expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
      })

      o.3c <- ode.3c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = etSs)
      s.3c <- sol.3c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = etSs)

      test_that("3 compartment steady state solved models and ODEs same.", {
        expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
      })

      # context(sprintf("Infusion + Bolus 1 cmt (%s)", .txt))

      et <- eventTable() |>
        add.dosing(dose = 3, rate = 1.5, nbr.doses = 6, dosing.interval = 8) |>
        add.dosing(dose = 1.5, nbr.doses = 6, dosing.interval = 8) |>
        add.sampling(seq(0, 48, length.out = 200))

      ode.1c <- rxode2({
        C2 <- center / V
        d / dt(center) <- -CL * C2
      })

      ## Solved systems can check the variables in the rxode2 statement
      ## to figure out what type of solved system is being requested
      ode.1cs <- rxode2(
      {
        V <- theta[1]
        CL <- theta[2]
        C2 <- linCmt()
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.1cs, 1)

      ## Instead of specifying parameters in the solved system, you can
      ## specify them in the linCmt variable.
      ode.1cs2 <- rxode2(
      {
        C2 <- linCmt(CL, V)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.1cs2, 1)

      ## The solved systems can be mixed with ODE solving routines (to
      ## speed them up a bit...?)

      o.1c <- ode.1c |> solve(params = c(V = 20, CL = 25), events = et)

      s.1c <- ode.1cs2 |> solve(params = c(V = 20, CL = 25), events = et)

      s.2c <- ode.1cs |> solve(theta = c(20, 25), events = et)

      test_that("1 compartment solved models and ODEs same.", {
        expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
        expect_equal(o.1c$C2, s.2c$C2, tolerance = tol)
      })

      # context(sprintf("Infusion + Bolus 2 cmt (%s)", .txt))

      ode.2c <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
      })

      sol.2c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2c, 2)

      o.2c <- ode.2c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10), events = et)
      s.2c <- sol.2c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10), events = et)

      test_that("2 compartment solved models and ODEs same.", {
        expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)
      })

      # context(sprintf("Infusion + Bolus 3 cmt (%s)", .txt))

      ode.3c <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        C4 <- peri2 / V3
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(peri2) <- Q2 * C2 - Q2 * C4
      })

      sol.3c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, Q2, V3)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3c, 3)

      o.3c <- ode.3c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = et)
      s.3c <- sol.3c |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = et)

      test_that("3 compartment solved models and ODEs same.", {
        expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
      })

      # context(sprintf("Oral + Infusion + Bolus Models 1 cmt (%s)", .txt))

      et <- eventTable() |>
        add.dosing(dose = 3, rate = 1.5, nbr.doses = 3, dosing.interval = 16, cmt = 2) |>
        add.dosing(dose = 1.5, nbr.doses = 3, dosing.interval = 16, cmt = 2) |>
        add.dosing(dose = 1.5, nbr.doses = 3, dosing.interval = 16, cmt = 1, start.time = 8) |>
        add.sampling(seq(0, 48, length.out = 200))

      ode.1c.ka <- rxode2({
        C2 <- center / V
        d / dt(depot) <- -KA * depot
        d / dt(center) <- KA * depot - CL * C2
      })

      sol.1c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, KA)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.1c.ka, 1, 1)

      o.1c <- ode.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = et)
      s.1c <- sol.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2), events = et)

      test_that("1 compartment solved models and ODEs same for mixed oral, iv and infusion.", {
        expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
      })

      # context(sprintf("Oral + Infusion + Bolus Models 2 cmt (%s)", .txt))

      ode.2c.ka <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
      })

      sol.2c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, KA)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2c.ka, 2, 1)

      o.2c <- ode.2c.ka |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = et)
      s.2c <- sol.2c.ka |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3), events = et)

      test_that("2 compartment solved models and ODEs same for mixed oral, iv and infusion.", {
        expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)
      })

      # context(sprintf("Oral + Infusion + Bolus Models 3 cmt (%s)", .txt))

      ode.3c.ka <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        C4 <- peri2 / V3
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(peri2) <- Q2 * C2 - Q2 * C4
      })

      sol.3c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, Q2, V3, KA)
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3c.ka, 3, 1)

      o.3c <- ode.3c.ka |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = et)
      s.3c <- sol.3c.ka |>
        solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3), events = et)

      test_that("3 compartment solved models and ODEs same for mixed oral, iv and infusion.", {
        expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
      })

      # context(sprintf("Modeled bio-availability 1 cmt (%s)", .txt))

      et <- eventTable() |>
        add.dosing(dose = 3, rate = 1.5, nbr.doses = 3, dosing.interval = 16, cmt = 2) |>
        add.dosing(dose = 1.5, nbr.doses = 3, dosing.interval = 16, cmt = 2) |>
        add.dosing(dose = 1.5, nbr.doses = 3, dosing.interval = 16, cmt = 1, start.time = 8) |>
        add.sampling(seq(0, 48, length.out = 200))

      ode.1c.ka <- rxode2({
        C2 <- center / V
        d / dt(depot) <- -KA * depot
        d / dt(center) <- KA * depot - CL * C2
        f(depot) <- fDepot
        f(center) <- fCenter
      })

      sol.1c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, KA)
        f(depot) <- fDepot
        f(central) <- fCenter
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.1c.ka, 1, 1)

      for (fd in c(0.5, 1, 2)) {
        for (fc in c(0.5, 1, 2)) {
          o.1c <- ode.1c.ka |>
            solve(params = c(V = 20, CL = 25, KA = 2, fDepot = fd, fCenter = fc), events = et)
          s.1c <- sol.1c.ka |>
            solve(params = c(V = 20, CL = 25, KA = 2, fDepot = fd, fCenter = fc), events = et)
          test_that(sprintf("1 compartment solved models and ODEs same for mixed oral, iv and infusion + Fd=%f,Fc=%f", fd, fc), {
            expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
          })
        }
      }

      # context(sprintf("Modeled bio-availability 2 cmt (%s)", .txt))

      ode.2c.ka <- rxode2(
      {
        C2 <- centr / V
        C3 <- peri / V2
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
        f(depot) <- fDepot
        f(centr) <- fCenter
        ## FIXME:
        ## f(central) should throw an error
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      sol.2c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, KA)
        f(depot) <- fDepot
        f(central) <- fCenter
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2c.ka, 2, 1)

      for (fd in c(0.5, 1, 2)) {
        for (fc in c(0.5, 1, 2)) {
          o.2c <- ode.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3, fDepot = fd, fCenter = fc), events = et)
          s.2c <- sol.2c.ka |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3, fDepot = fd, fCenter = fc), events = et)
          test_that(sprintf("2 compartment solved models and ODEs same for mixed oral, iv and infusion + Fd=%f,Fc=%f", fd, fc), {
            expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)
          })
        }
      }

      # context(sprintf("Modeled bio-availability 3 cmt (%s)", .txt))

      ode.3c.ka <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        C4 <- peri2 / V3
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(peri2) <- Q2 * C2 - Q2 * C4
        f(depot) <- fDepot
        f(centr) <- fCenter
      })

      sol.3c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, Q2, V3, KA)
        f(depot) <- fDepot
        f(central) <- fCenter
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3c.ka, 3, 1)

      for (fd in c(0.5, 1, 2)) {
        for (fc in c(0.5, 1, 2)) {
          o.3c <- ode.3c.ka |>
            solve(params = c(
              V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7,
              V3 = 400, KA = 0.3, fDepot = fd, fCenter = fc
            ), events = et)
          s.3c <- sol.3c.ka |>
            solve(params = c(
              V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3,
              fDepot = fd, fCenter = fc
            ), events = et)
          test_that(sprintf("3 compartment solved models and ODEs same for mixed oral, iv and infusion + Fd=%f,Fc=%f", fd, fc), {
            expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
          })
        }
      }

      # context(sprintf("Modeled lag time 1 cmt (%s)", .txt))

      et <- eventTable() |>
        add.dosing(dose = 3, rate = 1.5, nbr.doses = 3, dosing.interval = 16, cmt = 2) |>
        add.dosing(dose = 1.5, nbr.doses = 3, dosing.interval = 16, cmt = 2) |>
        add.dosing(dose = 1.5, nbr.doses = 3, dosing.interval = 16, cmt = 1, start.time = 8) |>
        add.sampling(seq(0, 48, length.out = 200))

      ode.1c.ka <- rxode2(
      {
        C2 <- center / V
        d / dt(depot) <- -KA * depot
        d / dt(center) <- KA * depot - CL * C2
        alag(depot) <- lagDepot
        alag(center) <- lagCenter
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      sol.1c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, KA)
        alag(depot) <- lagDepot
        alag(central) <- lagCenter
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.1c.ka, 1, 1)

      for (fd in c(1, 2, 10)) {
        for (fc in c(1, 2, 10)) {
          o.1c <- ode.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2, lagDepot = fd, lagCenter = fc), events = et)
          s.1c <- sol.1c.ka |> solve(params = c(V = 20, CL = 25, KA = 2, lagDepot = fd, lagCenter = fc), events = et)
          test_that(sprintf("1 compartment solved models and ODEs same for mixed oral, iv and infusion + Fd=%f,Fc=%f", fd, fc), {
            expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
          })
        }
      }

      # context(sprintf("Modeled lag time 2 cmt (%s)", .txt))

      ode.2c.ka <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
        alag(depot) <- lagDepot
        alag(centr) <- lagCenter
      })

      sol.2c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, KA)
        alag(depot) <- lagDepot
        alag(central) <- lagCenter
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2c.ka, 2, 1)

      for (fd in c(1, 2, 10)) {
        for (fc in c(1, 2, 10)) {
          o.2c <- ode.2c.ka |>
            solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3, lagDepot = fd, lagCenter = fc), events = et)
          s.2c <- sol.2c.ka |>
            solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, KA = 0.3, lagDepot = fd, lagCenter = fc), events = et)
          test_that(sprintf("2 compartment solved models and ODEs same for mixed oral, iv and infusion + Fd=%f,Fc=%f", fd, fc), {
            expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)
          })
        }
      }

      # context(sprintf("Modeled lag time 3 cmt (%s)", .txt))

      ode.3c.ka <- rxode2(
      {
        C2 <- centr / V
        C3 <- peri / V2
        C4 <- peri2 / V3
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(peri2) <- Q2 * C2 - Q2 * C4
        alag(depot) <- lagDepot
        alag(centr) <- lagCenter
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      sol.3c.ka <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, Q2, V3, KA)
        alag(depot) <- lagDepot
        alag(central) <- lagCenter
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3c.ka, 3, 1)

      for (fd in c(1, 2, 10)) {
        for (fc in c(1, 2, 10)) {
          o.3c <- ode.3c.ka |> solve(params = c(
            V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7,
            V3 = 400, KA = 0.3, lagDepot = fd, lagCenter = fc
          ), events = et)
          s.3c <- sol.3c.ka |> solve(params = c(
            V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, KA = 0.3,
            lagDepot = fd, lagCenter = fc
          ), events = et)
          test_that(sprintf("3 compartment solved models and ODEs same for mixed oral, iv and infusion + Fd=%f,Fc=%f", fd, fc), {
            expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
          })
        }
      }

      # context(sprintf("Modeled rate 1 cmt (%s)", .txt))

      ode.1c <- rxode2({
        C2 <- center / V
        d / dt(center) <- -CL * C2
        rate(center) <- rt
      })

      sol.1c <- rxode2(
      {
        C2 <- linCmt(CL, V)
        rate(central) <- rt
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.1c, 1)

      et <- eventTable() |>
        add.dosing(dose = 3, rate = -1, nbr.doses = 3, cmt = 1, dosing.interval = 12) |>
        add.sampling(seq(0, 36, length.out = 200))

      for (rt in c(0.5, 1, 1.5)) {
        o.1c <- ode.1c |> solve(params = c(V = 20, CL = 25, rt = rt), events = et)
        s.1c <- sol.1c |> solve(params = c(V = 20, CL = 25, rt = rt), events = et)
        test_that(sprintf("1 compartment solved models and ODEs same for rate-modeled infusion: %s", rt), {
          expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
        })
      }

      # context(sprintf("Modeled rate 2 cmt (%s)", .txt))

      ode.2c <- rxode2(
      {
        C2 <- centr / V
        C3 <- peri / V2
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
        rate(centr) <- rt
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      sol.2c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q)
        rate(central) <- rt
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2c, 2)

      for (rt in c(0.5, 1, 1.5)) {
        o.2c <- ode.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, rt = rt), events = et)
        s.2c <- sol.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, rt = rt), events = et)
        test_that(sprintf("2 compartment solved models and ODEs same for rate-modeled infusion: %s", rt), {
          expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)
        })
      }

      # context(sprintf("Modeled rate 3 cmt (%s)", .txt))

      ode.3c <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        C4 <- peri2 / V3
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(peri2) <- Q2 * C2 - Q2 * C4
        rate(centr) <- rt
      })

      sol.3c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, Q2, V3)
        rate(central) <- rt
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3c, 3)

      for (rt in c(0.5, 1, 1.5)) {
        s.3c <- sol.3c |>
          solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, rt = rt), events = et)
        o.3c <- ode.3c |>
          solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, rt = rt), events = et)
        test_that(sprintf("3 compartment solved models and ODEs same for rate-modeled infusion: %s", rt), {
          expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
        })
      }

      # context(sprintf("Modeled duration 1 cmt (%s)", .txt))

      ode.1c <- rxode2({
        C2 <- center / V
        d / dt(center) <- -CL * C2
        dur(center) <- dr
      })

      sol.1c <- rxode2(
      {
        C2 <- linCmt(CL, V)
        dur(central) <- dr
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.1c, 1)

      et <- eventTable() |>
        add.dosing(dose = 3, rate = -2, nbr.doses = 3, cmt = 1, dosing.interval = 12) |>
        add.sampling(seq(0, 36, length.out = 200))

      for (dur in c(0.5, 1, 1.5)) {
        o.1c <- ode.1c |> solve(params = c(V = 20, CL = 25, dr = dur), events = et)
        s.1c <- sol.1c |> solve(params = c(V = 20, CL = 25, dr = dur), events = et)
        test_that(sprintf("1 compartment solved models and ODEs same for dur-modeled infusion: %s", dur), {
          expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)
        })
      }

      # context(sprintf("Modeled duration 2 cmt (%s)", .txt))

      ode.2c <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
        dur(centr) <- dr
      })

      sol.2c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q)
        dur(central) <- dr
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.2c, 2)

      for (dur in c(0.5, 1, 1.5)) {
        o.2c <- ode.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, dr = dur), events = et)
        s.2c <- sol.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, dr = dur), events = et)
        test_that(sprintf("2 compartment solved models and ODEs same for dur-modeled infusion: %s", dur), {
          expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)
        })
      }

      # context(sprintf("Modeled duration 3 cmt (%s)", .txt))

      ode.3c <- rxode2({
        C2 <- centr / V
        C3 <- peri / V2
        C4 <- peri2 / V3
        d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(peri2) <- Q2 * C2 - Q2 * C4
        dur(centr) <- dr
      })

      sol.3c <- rxode2(
      {
        C2 <- linCmt(V, CL, V2, Q, Q2, V3)
        dur(central) <- dr
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(sol.3c, 3)

      for (dur in c(0.5, 1, 1.5)) {
        o.3c <- ode.3c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, dr = dur), events = et)
        s.3c <- sol.3c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400, dr = dur), events = et)
        test_that(sprintf("3 compartment solved models and ODEs same for dur-modeled infusion: %s", dur), {
          expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
        })
      }

      test_that("central should throw error", {
        expect_error(expect_message(rxode2({
          C2 <- centr / V
          C3 <- peri / V2
          d / dt(depot) <- -KA * depot
          d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
          d / dt(peri) <- Q * C2 - Q * C3
          alag(depot) <- lagDepot
          alag(centr) <- lagCenter
          alag(central) <- matt
        })))
      })

      test_that("depot should throw error", {
        expect_error(expect_message(rxode2({
          C2 <- centr / V
          C3 <- peri / V2
          d / dt(dep) <- -KA * dep
          d / dt(centr) <- KA * dep - CL * C2 - Q * C2 + Q * C3
          d / dt(peri) <- Q * C2 - Q * C3
          alag(dep) <- lagDepot
          alag(centr) <- lagCenter
          alag(depot) <- matt
        })))
      })

      ode.1cs2 <- rxode2(
      {
        C2 <- linCmt(CL, V)
        mtime(t1) <- mt1
        mtime(t2) <- mt2
      },
      linCmtSens = sens, linCmtSensType = linCmtSensType
      )

      goodP(ode.1cs2, 1)

      et <- eventTable() |>
        add.dosing(dose = 3, nbr.doses = 6, dosing.interval = 8) |>
        add.sampling(0:48)

      s.1c <- ode.1cs2 |> solve(
        params = c(V = 20, CL = 25, mt1 = 0.5, mt2 = 1.75),
        events = et
      )

      test_that("mtime with solved systems work", {
        expect_equal(s.1c$time[1:4], c(0, 0.5, 1, 1.75))
      })

      # context(sprintf("evid=3 (%s)", .txt))

      test_that(sprintf("evid==3 (%s)", .txt), {
        ode.1c <- rxode2(
        {
          C2 <- center / V
          d / dt(center) <- -CL * C2
        },
        linCmtSens = sens, linCmtSensType = linCmtSensType
        )

        sol.1c <- rxode2(
        {
          C2 <- linCmt(CL, V)
        },
        linCmtSens = sens, linCmtSensType = linCmtSensType
        )

        et <- eventTable() |>
          add.dosing(dose = 3, nbr.doses = 6, dosing.interval = 8) |>
          et(time = 25, evid = 3) |>
          add.sampling(seq(0, 48, length.out = 200))

        o1 <- rxSolve(ode.1c, params = c(V = 20, CL = 25), events = et)
        s1 <- rxSolve(sol.1c, params = c(V = 20, CL = 25), events = et)

        expect_equal(o1$C2, s1$C2, tolerance = tol)
      })
    }
    if (length(types) > 1) {
      test_that("double linCmt has error", {
        expect_error(expect_message(rxode2({
          C2 <- linCmt(CL, V)
          C2 <- linCmt(CL, V)
        })))
      })

      # context(sprintf("Steady State Infusions (%s)", .txt))

      test_that("Steady state IV infusion", {
        ev <- et(amt = 0, ss = 1, rate = 10000 / 8)

        ode.1c <- rxode2({
          C2 <- center / V
          d / dt(center) <- -CL * C2
        })

        ode.1cs2 <- rxode2({
          C2 <- linCmt(CL, V)
        })

        goodP(ode.1cs2, 1)

        o.1c <- ode.1c |> solve(params = c(V = 20, CL = 25), events = ev)

        s.1c <- ode.1cs2 |> solve(params = c(V = 20, CL = 25), events = ev)

        expect_equal(o.1c$C2, s.1c$C2, tolerance = tol)

        ode.2c <- rxode2({
          C2 <- centr / V
          C3 <- peri / V2
          d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3
          d / dt(peri) <- Q * C2 - Q * C3
        })

        sol.2c <- rxode2({
          C2 <- linCmt(V, CL, V2, Q1)
        })

        goodP(sol.2c, 2)

        o.2c <- ode.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10), events = ev)

        s.2c <- sol.2c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q1 = 10), events = ev)

        expect_equal(o.2c$C2, s.2c$C2, tolerance = tol)

        ode.3c <- rxode2({
          C2 <- centr / V
          C3 <- peri / V2
          C4 <- peri2 / V3
          d / dt(centr) <- -CL * C2 - Q * C2 + Q * C3 - Q2 * C2 + Q2 * C4
          d / dt(peri) <- Q * C2 - Q * C3
          d / dt(peri2) <- Q2 * C2 - Q2 * C4
        })

        sol.3c <- rxode2({
          C2 <- linCmt(V, CL, V2, Q, Q2, V3)
        })

        goodP(sol.3c, 3)

        o.3c <- ode.3c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = ev)

        s.3c <- sol.3c |> solve(params = c(V = 40, CL = 18, V2 = 297, Q = 10, Q2 = 7, V3 = 400), events = ev)

        expect_equal(o.3c$C2, s.3c$C2, tolerance = tol)
      })

      test_that(paste("Issue RxODE#258, sens:", sens), {
        m258 <- rxode2(
        {
          ka <- 1
          cl <- 3.5
          vc <- 40
          Conc <- linCmt()
          alag(depot) <- 1
        },
        linCmtSens = sens, linCmtSensType = linCmtSensType
        )

        m258o <- rxode2({
          ka <- 1
          cl <- 3.5
          vc <- 40
          d / dt(depot) <- -ka * depot
          d / dt(central) <- ka * depot - cl / vc * central
          Conc <- central / vc
          alag(depot) <- 1
        })

        s1 <- m258 |>
          et(dose = 100, time = 0, addl = 6, ii = 24) |>
          et(0, 250, by = 0.1) |>
          rxSolve()

        s2 <- m258o |>
          et(dose = 100, time = 0, addl = 6, ii = 24) |>
          et(0, 250, by = 0.1) |>
          rxSolve()

        expect_equal(s1$Conc, s2$Conc, tolerance = tol)
      })

      tol <- 1e-5 ## Current difference for all equations
      type <- 1
    }
  })

  rxTest({
    test_that("zero clearance solve #261", {

      rx <- rxode2({
        param(tlag, tka, tv, tcl, cl_crcl, cl_hep, cl_cyp2d6, cl_cyp3a_inh,
              cl_cyp3a_ind, cl_sexf, cl_japanese, prop_err, add_err,
              iiv_v, iiv_cl, iiv_ka, crcl, hepatic_impair, cyp2d6,
              cyp3a_inh, cyp3a_ind, sex, japanese)
        rxerr.rxLinCmt = 1
        tlag = 0.447
        tka = 0.0935
        tv = 144
        tcl = 93.8
        cl_crcl = 0.303
        cl_hep = 0.422
        cl_cyp2d6 = 0.626
        cl_cyp3a_inh = 0.504
        cl_cyp3a_ind = 3.9
        cl_sexf = 0.903
        cl_japanese = 1.12
        prop_err = 0.0951
        add_err = 0.00123
        iiv_v = 0
        iiv_cl = 0
        iiv_ka = 0
        cl_crcl_i = (crcl/80)^cl_crcl
        cl_hep_i = cl_hep * hepatic_impair + (1 - hepatic_impair)
        cl_cyp2d6_i = cl_cyp2d6 * cyp2d6 + (1 - cyp2d6)
        cl_cyp3a_inh_i = cl_cyp3a_inh * cyp3a_inh + (1 - cyp3a_inh)
        cl_cyp3a_ind_i = cl_cyp3a_ind * cyp3a_ind + (1 - cyp3a_ind)
        cl_sexf_i = cl_sexf * (sex == "Female") + (sex == "Male")
        cl_japanese_i = cl_japanese * japanese + (1 - japanese)
        cl = tcl * exp(iiv_cl) * cl_crcl_i * cl_hep_i * cl_cyp2d6_i *
          cl_cyp3a_inh_i * cl_cyp3a_ind_i * cl_sexf_i * cl_japanese_i
        v = tv * exp(iiv_v)
        ka = tka * exp(iiv_ka)
        alag(depot) = tlag
        rx_yj_ ~ 2
        rx_lambda_ ~ 1
        rx_low_ ~ 0
        rx_hi_ ~ 1
        rx_pred_f_ ~ linCmt()
        rx_pred_ ~ rx_pred_f_
        rx_r_ ~ (add_err)^2 + (rx_pred_f_)^2 * (prop_err)^2
        ipredSim = rxTBSi(rx_pred_, rx_lambda_, rx_yj_, rx_low_,
                          rx_hi_)
        sim = rxTBSi(rx_pred_ + sqrt(rx_r_) * rxerr.rxLinCmt, rx_lambda_,
                     rx_yj_, rx_low_, rx_hi_)
        cmt(rxLinCmt)
        dvid(3)
      })

      d_sim <-
        et(amt=4) |>
        et(time=0:24) |>
        as.data.frame() |>
        dplyr::mutate(
          japanese=0,
          sex=0,
          cyp3a_ind=0,
          cyp3a_inh=0,
          cyp2d6=0,
          hepatic_impair=0,
          crcl=80
        )

      f <- rxSolve(rx, d_sim)

      expect_false(any(is.na(f$sim) | is.nan(f$sim)))
      expect_false(any(is.na(f$ipredSim) | is.nan(f$ipredSim)))

    })


    # test mixed ODE and solved models

    test_that("mixed ode and solved models", {

      ode <- rxode2({
        lka <- 0.45
        lcl <- 1
        lvc  <- 3.45
        ka <- exp(lka)
        cl <- exp(lcl)
        vc  <- exp(lvc)
        kel <- cl / vc
        d/dt(depot) <- -ka*depot
        d/dt(central) <- ka*depot-kel*central
        Cc <- central / vc
        d/dt(accum) <- Cc
      })

      lin <- rxode2({
        lka <- 0.45
        lcl <- 1
        lvc  <- 3.45
        ka <- exp(lka)
        cl <- exp(lcl)
        vc  <- exp(lvc)
        Cc <- linCmt(vc, cl, ka)
        d/dt(accum) <- Cc
      })

      d <-
        rbind(
          data.frame(AMT = 500, EVID = 1, CMT = "depot", TIME = 0),
          data.frame(AMT = 0, EVID = 0, CMT = "central", TIME = 0:5)
        )

      d$ID <- 1

      for (meth in c("liblsoda", "lsoda", "dop853", "f78")) {
        dOde <- rxSolve(ode, d, meth=meth)

        dLin <- rxSolve(lin, d, meth=meth)

        expect_equal(dOde$Cc, dLin$Cc, tolerance = 1e-4)

        expect_equal(dOde$accum, dLin$accum, tolerance = 1e-4)

      }
    })
  })

}

rxTest({
  .mod <- rxode2::rxode2({
    d/dt(y) <- -y
  })
  .et <- rxode2::eventTable()
  .et$add.sampling(seq(0, 5, length.out = 10))

  # method, hmin (NULL = omit), tolerance
  .solvers <- list(
    list(method = "ab",           hmin = 0.0001, tol = 0.05),
    list(method = "abm",          hmin = 0.0001, tol = 0.05),
    list(method = "bs",           hmin = NULL,   tol = 0.05),
    list(method = "ck54",         hmin = NULL,   tol = 0.05),
    list(method = "dop5",         hmin = NULL,   tol = 0.05),
    list(method = "dop54",        hmin = NULL,   tol = 0.05),
    list(method = "dop87",        hmin = NULL,   tol = 0.05),
    list(method = "gauss6",       hmin = 0.01,   tol = 0.05),
    list(method = "geng5",        hmin = 0.01,   tol = 0.05),
    list(method = "iem",          hmin = NULL,   tol = 0.05),
    list(method = "iiic6",        hmin = 0.01,   tol = 0.05),
    list(method = "radauiia5",    hmin = 0.01,   tol = 0.05),
    list(method = "rk43",         hmin = NULL,   tol = 0.05),
    list(method = "f32",          hmin = NULL,   tol = 0.05),
    list(method = "f78",          hmin = NULL,   tol = 0.05),
    list(method = "ros4",         hmin = NULL,   tol = 0.05),
    list(method = "ros43",        hmin = NULL,   tol = 0.05),
    list(method = "ros6",         hmin = 0.01,   tol = 0.05),
    list(method = "sdirk43",      hmin = NULL,   tol = 0.05),
    list(method = "vern65",       hmin = NULL,   tol = 0.05),
    list(method = "vern76",       hmin = NULL,   tol = 0.05),
    list(method = "vern98",       hmin = NULL,   tol = 0.05),
    list(method = "backwardEuler", hmin = 0.01,  tol = 0.2)
  )

  for (.cfg in .solvers) {
    local({
      .method <- .cfg$method
      .tol    <- .cfg$tol
      .hmin   <- .cfg$hmin
      test_that(paste(.method, "integrates 1-compartment model correctly"), {
        .args <- list(.mod, params = c(), events = .et,
                      inits = c(y = 1), method = .method)
        if (!is.null(.hmin)) .args$hmin <- .hmin
        out <- do.call(rxode2::rxSolve, .args)
        expect_s3_class(out, "rxSolve")
        expect_equal(out$y, exp(-out$time), tolerance = .tol)
      })
    })
  }
})

rxTest({
  test_that("implicit solvers use stage-specific evaluation times", {
    mod <- rxode2::rxode2({
      d/dt(y) <- time
    })
    et <- rxode2::eventTable()
    et$add.sampling(seq(0, 1, by = 0.1))
    .expected <- 0.5 * seq(0, 1, by = 0.1)^2

    for (.method in c("backwardEuler", "gauss6", "iiic6", "radauiia5",
                      "geng5", "sdirk43", "ros6")) {
      out <- rxode2::rxSolve(mod, params = c(), events = et, inits = c(y = 0),
                             method = .method, hmin = 1e-3, hmax = 0.05)
      .tol <- if (.method == "ros6") 1e-2 else 5e-3
      expect_equal(out$y, .expected, tolerance = .tol,
                   info = paste("method", .method))
    }
  })
})

rxTest({
  # -- Analytical helper ------------------------------------------------------
  # depot(t) for a single constant-rate infusion into a one-compartment depot
  # that drains with first-order rate constant ka.
  #
  #   R  = infusion rate (amount / time)
  #   D  = duration      = amt / R   (infusion starts at t = 0)
  #   ka = depot elimination rate constant
  #
  # During infusion (0 <= t <= D):
  #   depot(t) = (R/ka) * (1 - exp(-ka*t))
  #
  # After infusion (t > D):
  #   depot(t) = depot(D) * exp(-ka*(t-D))
  #            = (R/ka) * (1-exp(-ka*D)) * exp(-ka*(t-D))
  .depotAnalytic <- function(t, R, D, ka) {
    vapply(t, function(ti) {
      if (ti <= 0) return(0)
      if (ti <= D) {
        (R / ka) * (1 - exp(-ka * ti))
      } else {
        depot_D <- (R / ka) * (1 - exp(-ka * D))
        depot_D * exp(-ka * (ti - D))
      }
    }, numeric(1))
  }

  # -- Test 1 -----------------------------------------------------------------
  # state = 0 throughout  ->  rate(depot) = rate0 + 0 = rate0 exactly.
  # This is the baseline: no state dependence, behaviour identical to a fixed
  # rate infusion.
  test_that("state-dep rate with state=0 uses exactly rate0 (no state shift)", {
    mod <- rxode2({
      # Infusion rate is modelled as rate0 + state.
      # With state held at 0, this collapses to the plain fixed-rate case.
      rate(depot) <- rate0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- 0         # constant state (initial value is the only value)
    })

    amt   <- 100
    rate0 <- 10
    ka    <- 0.5
    R     <- rate0            # state=0 -> effective rate = rate0
    D     <- amt / R          # expected infusion duration
    times <- seq(0, 30, by = 0.5)

    et <- eventTable() |>
      add.dosing(dose = amt, rate = -1) |>  # rate=-1 -> use modeled rate
      add.sampling(times)

    s <- solve(mod, et,
               inits  = c(state = 0, depot = 0),
               params = c(rate0 = rate0, ka = ka))

    expected <- .depotAnalytic(times, R, D, ka)

    # Every sampled depot value must match the analytical trajectory computed
    # with rate = rate0.  A mismatch would mean the solver used a different rate.
    expect_equal(s$depot[s$time %in% times], expected,
                 tolerance = 1e-3,
                 label = "depot matches rate0-only analytical solution when state=0")
  })

  # -- Test 2 -----------------------------------------------------------------
  # state = S0 (constant, positive and negative) throughout.
  # The effective rate at infusion start is rate0 + S0; duration = amt/(rate0+S0).
  # The full depot trajectory must match the analytical formula for that rate.
  test_that("state-dep rate shifts infusion rate by state value: rate = rate0 + state_at_start", {
    mod <- rxode2({
      # Infusion rate includes a state-variable contribution.
      # With d/dt(state)=0 the state stays at its initial value for the whole
      # simulation, giving an analytically verifiable rate.
      rate(depot) <- rate0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- 0
    })

    amt   <- 100
    rate0 <- 10
    ka    <- 0.5
    times <- seq(0, 40, by = 0.5)

    for (S0 in c(2, 5, -3)) {
      R <- rate0 + S0         # effective rate at the moment the infusion starts
      D <- amt / R            # duration changes with state value

      et <- eventTable() |>
        add.dosing(dose = amt, rate = -1) |>
        add.sampling(times)

      s <- solve(mod, et,
                 inits  = c(state = S0, depot = 0),
                 params = c(rate0 = rate0, ka = ka))

      expected <- .depotAnalytic(times, R, D, ka)

      # The solver must use rate = rate0 + S0, not rate0 alone.
      # Verifying the full trajectory confirms both the rate magnitude and the
      # derived infusion duration (D = amt/R).
      expect_equal(s$depot[s$time %in% times], expected,
                   tolerance = 1e-3,
                   label = paste0("depot matches analytic solution for rate=",
                                  R, " (S0=", S0, ")"))

      # Confirm the trajectory with S0!=0 differs from the S0=0 baseline,
      # i.e., state truly shifts the rate and is not silently ignored.
      if (S0 != 0) {
        expected_s0 <- .depotAnalytic(times, rate0, amt / rate0, ka)
        expect_false(isTRUE(all.equal(expected, expected_s0)),
                     label = paste0("state S0=", S0,
                                    " produces a different trajectory than state=0"))
      }
    }
  })

  # -- Test 3 -----------------------------------------------------------------
  # The infusion end time must be exactly amt / (rate0 + state_at_start).
  # We probe this by checking depot at the expected stop time D and well after.
  # During the infusion depot is rising; after it decays.  Confirming depot(D)
  # matches the analytical peak locks in the exact stop time.
  test_that("state-dep rate: infusion ends at exactly amt/(rate0 + state_at_start)", {
    mod <- rxode2({
      rate(depot) <- rate0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- 0
    })

    amt   <- 100
    rate0 <- 5
    ka    <- 0.3

    for (S0 in c(0, 3, -1)) {
      R <- rate0 + S0
      D <- amt / R           # expected infusion end time

      # Sample finely around D so we capture exactly the stop-event time
      times <- sort(unique(c(seq(0, D + 8, by = 0.1), D)))

      et <- eventTable() |>
        add.dosing(dose = amt, rate = -1) |>
        add.sampling(times)

      s <- solve(mod, et,
                 inits  = c(state = S0, depot = 0),
                 params = c(rate0 = rate0, ka = ka))

      # depot(D) from the analytical formula
      depot_at_D_expected <- (R / ka) * (1 - exp(-ka * D))

      # observed depot at the exact stop-event sample time
      obs_at_D <- s$depot[which.min(abs(s$time - D))]

      # If rate was wrong the infusion would stop at a different time, yielding
      # a depot peak that is too large (rate too high, shorter duration) or too
      # small (rate too low, longer duration).
      expect_equal(obs_at_D, depot_at_D_expected,
                   tolerance = 1e-3,
                   label = paste0("depot at infusion stop D=", round(D, 4),
                                  " matches analytic (S0=", S0, ")"))
    }
  })

  # -- Test 4 -----------------------------------------------------------------
  # Multiple doses: each dose's rate is evaluated at the state present when
  # THAT dose's infusion begins.  Because state changes over time (exponential
  # decay here), consecutive doses get different effective rates and therefore
  # different durations.
  # We verify no NAs and that the two-dose trajectory is self-consistent (not
  # all identical) when the state is non-zero.
  test_that("state-dep rate: consecutive doses evaluated at their own start state", {
    mod <- rxode2({
      # rate depends on a decaying state variable so each dose sees a different
      # rate if the infusions are spaced far enough apart.
      rate(depot) <- rate0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- -kstate * state   # state decays between doses
    })

    et <- et() |>
      et(amt = 50, rate = -1, time = 0) |>    # dose 1 - state is large
      et(amt = 50, rate = -1, time = 15) |>   # dose 2 - state has decayed
      et(seq(0, 30, by = 0.5))

    s <- rxSolve(mod, et,
                 params = c(rate0 = 5, ka = 0.4, kstate = 0.2),
                 inits  = c(depot = 0, state = 4))

    # No numerical failures
    expect_false(any(is.na(s$depot)),
                 label = "no NAs in depot with two state-dep rate doses")

    # The trajectory around dose 2 must differ from what a state=0 run would
    # produce (state has not decayed to exactly 0 by t=15).
    s0 <- rxSolve(mod, et,
                  params = c(rate0 = 5, ka = 0.4, kstate = 0.2),
                  inits  = c(depot = 0, state = 0))

    expect_false(isTRUE(all.equal(s$depot, s0$depot)),
                 label = "nonzero initial state gives different depot trajectory than state=0")
  })
})

rxTest({

  test_that("cvode integrates simple exponential decay correctly", {
    mod <- rxode2({
      d/dt(y) <- -y
    })
    et <- eventTable()
    et$add.sampling(seq(0, 5, by = 0.5))
    out <- rxSolve(mod, params = c(), events = et,
                   inits = c(y = 1), method = "cvode")
    expect_s3_class(out, "rxSolve")
    expect_equal(out$y, exp(-out$time), tolerance = 1e-5)
  })

  test_that("cvode matches liblsoda on a stiff TMDD model", {
    # Two-compartment TMDD model (moderately stiff)
    mod <- rxode2({
      d/dt(C)  <- -kel * C - kon * C * R + koff * RC
      d/dt(R)  <- ksyn - kdeg * R - kon * C * R + koff * RC + kint * RC
      d/dt(RC) <- kon * C * R - koff * RC - kint * RC
    })
    parms <- c(kel = 0.1, kon = 0.5, koff = 0.05, ksyn = 10,
               kdeg = 0.1, kint = 0.2)
    et <- eventTable()
    et$add.dosing(dose = 100, start.time = 0, nbr.doses = 1, dosing.to = 1)
    et$add.sampling(seq(0, 48, by = 1))

    out_lsoda <- rxSolve(mod, params = parms, events = et,
                         inits = c(C = 0, R = 100, RC = 0),
                         method = "liblsoda")
    out_cvode <- rxSolve(mod, params = parms, events = et,
                         inits = c(C = 0, R = 100, RC = 0),
                         method = "cvode")

    expect_equal(out_cvode$C,  out_lsoda$C,  tolerance = 1e-4)
    expect_equal(out_cvode$R,  out_lsoda$R,  tolerance = 1e-4)
    expect_equal(out_cvode$RC, out_lsoda$RC, tolerance = 1e-4)
  })

  test_that("cvode linear solver options run without error", {
    mod <- rxode2({
      d/dt(y) <- -y
    })
    et <- eventTable()
    et$add.sampling(seq(0, 2, by = 0.5))

    for (.lsname in c("dense", "band", "gmres", "bicgstab", "tfqmr")) {
      out <- rxSolve(mod, params = c(), events = et,
                     inits = c(y = 1), method = "cvode",
                     cvodeLinSolver = .lsname)
      expect_s3_class(out, "rxSolve")
      expect_equal(out$y, exp(-out$time), tolerance = 1e-4)
    }
  })
})

rxTest({
  withr::local_options(list(rxode2.eventSens = "fd"))

  test_that("dose-history names as an lhs do not break analytic Jacobian (calcJac)", {
    ## A bare dose-history name (tad, dosenum, tlast, ...) used as a value expands
    ## to its functional form, but as an assignment *target* it must stay a plain
    ## symbol.  Otherwise the lhs is rewritten into the function body, producing an
    ## invalid assignment like `(t-tlast()) = t-tlast()` that fails calcJac.  FOCEi's
    ## finite-difference model appends exactly `tad = tad()` / `dosenum = dosenum()`,
    ## and delay()/DDE models force the Jacobian-requiring dense solver, so this hit
    ## every DDE fit (rxode2/nlmixr2#dde).
    expect_s3_class(
      suppressMessages(rxode2("d/dt(cen) <- -k*cen\ncen(0) <- 10\nk <- 0.2\ntad <- tad()",
                              calcJac = TRUE)), "rxode2")
    expect_s3_class(
      suppressMessages(rxode2("d/dt(cen) <- -k*cen\ncen(0) <- 10\nk <- 0.2\ndosenum <- dosenum()",
                              calcJac = TRUE)), "rxode2")
    ## the practical trigger: a delay() model with the FOCEi FD-model bookkeeping lines
    expect_s3_class(
      suppressMessages(rxode2(paste0(
        "d/dt(cen) <- -k*cen + kin*delay(cen, tau)\ncen(0) <- 10\n",
        "k <- 0.2\nkin <- 0.5\ntau <- 1\ntad <- tad()\ndosenum <- dosenum()"),
        calcJac = TRUE)), "rxode2")
    ## a bare dose-history name used as a value still expands (no regression)
    .m <- suppressMessages(rxode2("d/dt(cen) <- -k*cen\ncen(0) <- 10\nk <- 0.2\ny <- tad + 1"))
    expect_true(grepl("y=tad+1", rxNorm(.m), fixed = TRUE))
  })

  test_that("Specified jacobian is captured", {

    Vtpol2 <- rxode2("
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
")

    expect_true(any(rxDfdy(Vtpol2) == "df(y)/dy(dy)"))
    expect_true(any(rxDfdy(Vtpol2) == "df(dy)/dy(y)"))
    expect_true(any(rxDfdy(Vtpol2) == "df(dy)/dy(dy)"))
  })

  ## fixme multiple jacobian definitions raise an error.

  tmp <- rxode2("
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
df(dy)/dy(dy) = mu*(1-y^2)
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
")

  test_that("Doubled jacobain will compile correctly", {
    expect_s3_class(tmp, "rxode2")
  })

  test_that("Jacobian and sensitivity not specified.", {
    norm <- rxode2("
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
")

    expect_false(norm$calcJac)
    expect_false(norm$calcSens)
    rxDelete(norm)
  })

  test_that("Jacobian specified but sensitivity not specified.", {
    jac <- suppressMessages(rxode2("
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
", calcJac = TRUE))

    expect_true(jac$calcJac)
    expect_false(jac$calcSens)
    rxDelete(jac)
  })

  test_that("Sensitivity specified.", {
    sens <- suppressMessages(rxode2("
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
", calcSens = TRUE))

    expect_false(sens$calcJac)
    expect_true(sens$calcSens)
    rxDelete(sens)
  })

  test_that("Jac/Sens can be calculated from 'normal' model", {
    norm <- rxode2("
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
")
    jac <- norm
    jac <- suppressMessages(rxode2(jac, calcJac = TRUE))
    expect_true(jac$calcJac)
    expect_false(jac$calcSens)
    sens <- suppressMessages(rxode2(jac, calcSens = TRUE))
    expect_false(sens$calcJac)
    expect_true(sens$calcSens)
    full <- suppressMessages(rxode2(jac, calcSens = TRUE, calcJac = TRUE))
    expect_false(sens$calcJac)
    expect_true(sens$calcSens)
    rxDelete(jac)
    rxDelete(sens)
    rxDelete(norm)
    rxDelete(full)
  })

  test_that("Jacobian and sensitivity specified.", {
    sens <- suppressMessages(rxode2("
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
", calcSens = TRUE))

    norm <- suppressMessages(rxode2(sens, calcSens = FALSE))
    expect_false(norm$calcJac)
    expect_false(norm$calcSens)

    jac <- suppressMessages(rxode2(sens, calcJac = TRUE))
    expect_true(jac$calcJac)
    expect_false(jac$calcSens)
    expect_false(sens$calcJac)
    expect_true(sens$calcSens)
  })

  test_that("Conditional Sensitivities", {

    transit.if <- rxode2({
      ## Table 3 from Savic 2007
      cl <- 17.2 # (L/hr)
      vc <- 45.1 # L
      ka1 <- 0.38 # 1/hr
      ka2 <- 0.2 # 1/hr
      mtt <- 0.37 # hr
      bio <- 1
      n <- 20.1
      k <- cl / vc
      if (pop1 == 1) {
        ka <- ka1
      } else {
        ka <- ka2
      }
      d/dt(depot) <- transit(n, mtt, bio) - ka * depot
      d/dt(cen) <- ka * depot - k * cen
    })

    expect_false(transit.if$calcJac)
    expect_false(transit.if$calcSens)

    jac <- suppressMessages(rxode2(transit.if, calcJac = TRUE))
    expect_true(jac$calcJac)
    expect_false(jac$calcSens)

    sens <- suppressMessages(rxode2(transit.if, calcSens = TRUE))
    expect_false(sens$calcJac)
    expect_true(sens$calcSens)

    full <- suppressMessages(rxode2(transit.if, calcSens = TRUE, calcJac = TRUE))
    expect_true(full$calcJac)
    expect_true(full$calcSens)

  })

  test_that("Transit Sensitivities", {

    mod <- rxode2("
## Table 3 from Savic 2007
cl = 17.2 # (L/hr)
vc = 45.1 # L
ka = 0.38 # 1/hr
mtt = 0.37 # hr
bio=1
n = 20.1
k = cl/vc
ktr = (n+1)/mtt
## note that lgammafn is the same as lgamma in R.
d/dt(depot) = exp(log(bio*podo(depot))+log(ktr)+n*log(ktr*tad(depot))-ktr*tad(depot)-lgammafn(n+1))-ka*depot
d/dt(cen) = ka*depot-k*cen
")

    mod <- suppressMessages(rxode2(mod, calcSens = TRUE))

    et <- eventTable()
    et$add.sampling(seq(0, 10, length.out = 200))
    et$add.dosing(20, start.time = 0, evid=7)

    transit <- suppressWarnings({
      rxSolve(mod, et)
    })

    ## Used the log(0) protection since the depot_mtt sensitivity
    ## includes log(t*...) and t=0
    ##
    ## These results are now system dependent since it uses
    ## log(sqrt(DOUBLE_EPS)) and DOUBLE_EPS varies by platform, so
    ## just make sure the results are not NA.
    expect_true(all(!is.na(transit[["_sens_depot_mtt"]])))
    expect_equal(transit$depot.mtt, transit[["_sens_depot_mtt"]])
    expect_equal(transit$depot_mtt, transit[["_sens_depot_mtt"]])
    expect_equal(transit[["depot_mtt"]], transit[["_sens_depot_mtt"]])
    expect_equal(transit[["depot.mtt"]], transit[["_sens_depot_mtt"]])

    mod <- rxode2({
      ## Table 3 from Savic 2007
      cl <- 17.2 # (L/hr)
      vc <- 45.1 # L
      tvka <- 0.38 # 1/hr
      tvmtt <- 0.37 # hr
      ka <- tvka * exp(eta_ka)
      mtt <- tvmtt * exp(eta_mtt)
      bio <- 1
      n <- 20.1
      k <- cl / vc
      ktr <- (n + 1) / mtt
      ## note that lgammafn is the same as lgamma in R.
      d/dt(depot) <- exp(log(bio * podo(depot)) + log(ktr) + n * log(ktr * tad(depot)) - ktr * tad(depot) -
                           lgammafn(n + 1)) - ka * depot
      d/dt(cen) <- ka * depot - k * cen
    })

    transit <- suppressMessages(rxode2(mod, calcSens = c("eta_ka", "eta_mtt")))
    expect_true(all(!is.na(transit[["_sens_depot_mtt"]])))

    ## tmp <- rxode2(mod, calcSens=list(eta=c("eta_ka", "eta_mtt"), theta=c("cl", "vc")));
  })

  test_that("df(state)/dy(THETA[n]) and dy(ETA[n]) do not segfault", {
    # The grammar accepts a bracketed THETA[n]/ETA[n] in dy(), but handleDy()
    # never registered the synthetic _THETA_n_ / _ETA_n_ symbol, leaving tb.ix
    # = -2; that -2 was stored in tb.dy[] and assertCorrectDfDy() then read
    # tb.ss.line[-2] -- an out-of-bounds access that crashed R.  nlm/FOCEi hit
    # this by re-parsing their generated calcJac model (parameters expressed as
    # THETA[n]) in the residual/table step.  It must error or parse, never crash.
    # The operative property is simply "did not crash": rxModelVars() returns
    # normally, whether it parses the model or raises a clean R error.
    .safe <- function(m) {
      .r <- tryCatch(suppressMessages(rxModelVars(m)),
                     error = function(e) "clean-error")
      !is.null(.r)
    }
    expect_true(.safe("d/dt(x1)=-x1\nx1(0)=1\ny=x1\ndf(x1)/dy(THETA[1])=0\n"))
    expect_true(.safe("d/dt(x1)=-x1\nx1(0)=1\ny=x1\ndf(x1)/dy(ETA[1])=0\n"))
    # a full re-parsed calcJac DDE model (delay in a product of delayed states,
    # parameters as THETA[n], the FOCEi tad=/dosenum= bookkeeping) must not crash
    expect_true(.safe(paste0(
      "cmt(x1);\ncmt(x2);\nka=2495\nke=2.52\nV=2.79\n",
      "d/dt(x1)=-ka*x1\nd/dt(x2)=ka*x1-ke*x2\n",
      "d/dt(cen)=exp(THETA[1])*delay(x1, exp(THETA[2]))*delay(x2, exp(THETA[2]))\n",
      "cen(0)=0\ndf(cen)/dy(x1)=0\ndf(cen)/dy(THETA[1])=0\ndf(cen)/dy(THETA[2])=0\n",
      "df(cen)/dy(THETA[3])=0\ny=cen\ntad=t-tlast()\n")))
  })
})

rxTest({
  ## https://cran.r-project.org/web/packages/diffEq/vignettes/ODEinR.pdf p15

  Vtpol <- rxode2("
d/dt(y) = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
")


  Vtpol2 <- rxode2("
d/dt(y)  = dy
d/dt(dy) = mu*(1-y^2)*dy - y
## Jacobian
df(y)/dy(dy)  = 1
df(dy)/dy(y)  = -2*dy*mu*y - 1
df(dy)/dy(dy) = mu*(1-y^2)
## Initial conditions
y(0) = 2
dy(0) = 0
## mu
mu = 1 ## nonstiff; 10 moderately stiff; 1000 stiff
")

  et <- eventTable()
  et$add.sampling(seq(0, 30, by = 0.01))

  stiff <- solve(Vtpol, et)

  ## plot(stiff$time,stiff$y,type="l")
  counts <- data.frame(stiff$counts, mu = 1, digest = digest::digest(round(as.data.frame(stiff), 3)))

  for (i in 10^(1:7)) {
    stiff$mu <- i
    ## plot(stiff$time,stiff$y,type="l");
    ## title(sprintf("mu=%s",i));
    counts <- rbind(counts, data.frame(stiff$counts, mu = i, digest = digest::digest(round(as.data.frame(stiff), 3))))
  }

  test_that("No User jacobian are called.", {
    expect_true(all(counts$jac == 0))
  })

  ## Data sets match

  ## Changes based on platform.

  ## test_that("Solving of stiff systems by automatic jacobians is expected",{
  ##     expect_equal(digest(counts),"982ce8816e9e1f5db6ee315c71b6a7dc")
  ## })

  counts0 <- counts

  stiff <- solve(Vtpol2, et)
  ## plot(stiff$time,stiff$y,type="l")
  counts <- data.frame(stiff$counts, mu = 1, digest = digest::digest(round(as.data.frame(stiff), 3)))

  for (i in 10^(1:7)) {
    stiff$mu <- i
    ## plot(stiff$time,stiff$y,type="l");
    ## title(sprintf("mu=%s",i));
    counts <- rbind(counts, data.frame(stiff$counts, mu = i, digest = digest::digest(round(as.data.frame(stiff), 3))))
  }

  ## print(counts)

  ## changes based on platform.

  ## test_that("Some User jacobian are called.",{
  ##     expect_true(!all(counts$user_jac == 0))
  ## })

  test_that("Same solutions", {
    expect_true(all(counts0$digest == counts$digest))
  })
})

rxTest({
  # Tests for absorption lag time.

  mod <- rxode2({
    a <- 6
    b <- 0.6
    d/dt(intestine) <- -a * intestine
    d/dt(blood) <- a * intestine - b * blood
  })

  mod2 <- rxode2({
    a <- 6
    b <- 0.6
    d / dt(intestine) <- -a * intestine
    lag(intestine) <- 2
    d / dt(blood) <- a * intestine - b * blood
  })

  et <- #et(time=0, amt= 5, ss=1, ii=1, cmt=1) |>
    et(time=0, cmt=1, amt=5, addl=10, ii=4) |>
    et(seq(0, 24, by=0.25/24))

  tmp <- rxSolve(mod2, et)

  expect_error(plot(tmp, intestine), NA)
  ## vdiffr::expect_doppelganger("intestine-mod2a", plot(tmp, intestine))

  et <- #et(time=0, amt= 5, ss=1, ii=1, cmt=1) |>
    et(time=0, cmt=1, amt=5, ii=4, ss=1) |>
    et(time=4, cmt=1, amt=5, addl=10, ii=4) |>
    et(seq(0, 24, by=0.25/24))


  tmp <- rxSolve(mod2, et)

  ## vdiffr::expect_doppelganger("intestine-mod2b", plot(tmp, intestine))
  expect_error(plot(tmp, intestine), NA)

  ms <- .methods0
  for (m in ms) {
    skip_if_not_installed("units")

    obs <- units::set_units(seq(0, 10, by = 1 / 24), "days")

    et <- eventTable(time.units = "days")
    et$add.sampling(obs)
    et$add.dosing(
      dose = 2 / 24, start.time = 0,
      nbr.doses = 10, dosing.interval = 1
    )

    solve1 <- solve(mod, et, method = m)
    solve2 <- solve(mod2, et, method = m)

    test_that(sprintf("Test absorption lag-time with IV dosing (%s): Solves with lag times are different", m), {
      expect_equal(obs, solve1$time)
      expect_equal(obs, solve2$time)
      expect_false(all(solve1$intestine == solve2$intestine))
      expect_false(all(solve1$blood == solve2$blood))
    })

    et <- eventTable(time.units = "days")
    et$add.sampling(seq(0, 10, by = 1 / 24))
    et$add.dosing(
      dose = 2 / 24, start.time = 2,
      nbr.doses = 10, dosing.interval = 1
    )

    solve3 <- solve(mod, et, method = m)

    test_that(sprintf("Test absorption lag-time with IV dosing (%s): Absorption lag shifts event by 2", m), {
      expect_equal(obs, solve3$time)
      expect_equal(solve3$intestine, solve2$intestine, tolerance = 1e-5)
      expect_equal(solve3$blood, solve2$blood, tolerance = 1e-5)
    })

  }
})

rxTest({
  .mod <- rxode2({
    d/dt(intestine) <- -a * intestine
    d/dt(blood) <- a * intestine - b * blood
  })

  .et <- eventTable()
  .et$add.sampling(seq(0, 10, length.out = 10))
  .et$add.dosing(dose = 1, strt.time = 0)

  # Build a 4-subject event table (id column required for nsub > 1)
  .ev <- do.call(rbind, lapply(1:4, function(i) {
    d <- as.data.frame(.et)
    d$id <- i
    d
  }))

  .p <- data.frame(id = 1:4, a = 6, b = seq(0.4, 0.9, length.out = 4))

  test_that("tolFactor=NULL (default) solves without error", {
    expect_error(rxSolve(.mod, .p, .ev, tolFactor = NULL), NA)
  })

  test_that("tolFactor scalar >= 1 solves without error", {
    expect_error(rxSolve(.mod, .p, .ev, tolFactor = 10), NA)
  })

  test_that("tolFactor vector (one per subject) solves without error", {
    expect_error(rxSolve(.mod, .p, .ev, tolFactor = c(1, 2, 5, 10)), NA)
  })

  test_that("tolFactor named vector matched by subject ID solves without error", {
    expect_error(rxSolve(.mod, .p, .ev, tolFactor = c("2" = 5, "4" = 10)), NA)
  })

  test_that("tolFactor < 1 is rejected", {
    expect_error(rxSolve(.mod, .p, .ev, tolFactor = 0.5))
  })

  test_that("$tolFactor returns per-subject factors used in the solve", {
    .out <- rxSolve(.mod, .p, .ev, tolFactor = c(1, 2, 5, 10))
    .tf <- .out$tolFactor
    expect_length(.tf, 4)
    expect_equal(as.numeric(.tf), c(1, 2, 5, 10))
  })

  test_that("named tolFactor: matched subjects get factor, unmatched stay at 1", {
    .out <- rxSolve(.mod, .p, .ev, tolFactor = c("2" = 5, "4" = 10))
    .got <- as.numeric(.out$tolFactor)
    expect_equal(.got[1], 1)
    expect_equal(.got[2], 5)
    expect_equal(.got[3], 1)
    expect_equal(.got[4], 10)
  })

  test_that("tolFactor=1 for all subjects does not change results vs NULL", {
    .base <- rxSolve(.mod, .p, .ev, tolFactor = NULL)
    .tf1  <- rxSolve(.mod, .p, .ev, tolFactor = rep(1, 4))
    expect_equal(
      as.data.frame(.base)[, c("id", "time", "intestine", "blood")],
      as.data.frame(.tf1)[, c("id", "time", "intestine", "blood")]
    )
  })
})

rxTest({
  skip_if_not_installed("nlmixr2data")
  warfarin <- nlmixr2data::warfarin

  test_that("cmt() syntax makes sense", {

    mod <- rxode2({
      a <- 6
      b <- 0.6
      cmt(blood) # cmt = 1 now
      d / dt(intestine) <- -a * intestine
      d / dt(blood) <- a * intestine - b * blood
    })

    expect_equal(c("blood", "intestine"), rxState(mod))

    mod <- rxode2({
      a <- 6
      b <- 0.6
      d / dt(intestine) <- -a * intestine
      d / dt(blood) <- a * intestine - b * blood
    })

    expect_equal(c("intestine", "blood"), rxState(mod))


    expect_error(suppressMessages(rxode2({
      a <- 6
      b <- 0.6
      cmt(matt) # cmt = 1 now
      d/dt(intestine) <- -a * intestine
      d/dt(blood) <- a * intestine - b * blood
    })))


    tmp <- rxode2({
      a <- 6
      b <- 0.6
      cmt(blood) # cmt = 1 now
      cmt(intestine) # cmt = 2 now
      cmt(matt)
      d / dt(intestine) <- -a * intestine
      d / dt(blood) <- a * intestine - b * blood
    })
    expect_equal(tmp$stateExtra, "matt")

  })

  test_that("Compartment melding with dvid", {

    w <- rxode2({
      ktr <- exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      ##
      logit <- exp(poplogit + eta.emax)
      emax <- logit / (1 + logit)
      ec50 <- exp(tec50 + eta.ec50)
      kout <- exp(tkout + eta.kout)
      e0 <- exp(te0 + eta.e0)
      ##
      DCP <- center / v
      PD <- 1 - emax * DCP / (ec50 + DCP)
      ##
      pca(0) <- e0
      kin <- e0 * kout
      ##
      d / dt(depot) <- -ktr * depot
      d / dt(gut) <- ktr * depot - ka * gut
      d / dt(center) <- ka * gut - cl / v * center
      d / dt(pca) <- kin * PD - kout * pca
      ##
      cp <- center / v
      cmt(cp)
      ## These dvid w/ negative values are counted backward from the last
      dvid(-1, 4)
    })

    w2 <- nlmixr2data::warfarin

    tmp <- etTrans(w2, w, addCmt = TRUE)

    expect_equal(
      tmp$CMT[tmp$ID == 1],
      c(
        1L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 4L, 5L, 4L, 5L, 4L, 5L,
        4L, 4L, 4L, 4L
      )
    )

    w <- rxode2({
      ktr <- exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      ##
      logit <- exp(poplogit + eta.emax)
      emax <- logit / (1 + logit)
      ec50 <- exp(tec50 + eta.ec50)
      kout <- exp(tkout + eta.kout)
      e0 <- exp(te0 + eta.e0)
      ##
      DCP <- center / v
      PD <- 1 - emax * DCP / (ec50 + DCP)
      ##
      pca(0) <- e0
      kin <- e0 * kout
      ##
      d / dt(depot) <- -ktr * depot
      d / dt(gut) <- ktr * depot - ka * gut
      d / dt(center) <- ka * gut - cl / v * center
      d / dt(pca) <- kin * PD - kout * pca
      ##
      cp <- center / v
      cmt(cp)
      dvid(5, 4)
    })

    w2 <- nlmixr2data::warfarin

    tmp <- etTrans(w2, w, addCmt = TRUE)

    expect_equal(
      tmp$CMT[tmp$ID == 1],
      c(
        1L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 4L, 5L, 4L, 5L, 4L, 5L,
        4L, 4L, 4L, 4L
      )
    )

    w2$dvid <- paste(w2$dvid)
    tmp <- etTrans(w2, w, addCmt = TRUE)
    expect_equal(
      tmp$CMT[tmp$ID == 1],
      c(
        1L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 4L, 5L, 4L, 5L, 4L, 5L,
        4L, 4L, 4L, 4L
      )
    )

    w2 <- nlmixr2data::warfarin
    w2$dvid <- as.integer(w2$dvid)
    tmp <- etTrans(w2, w, addCmt = TRUE)
    expect_equal(
      tmp$CMT[tmp$ID == 1],
      c(
        1L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 4L, 5L, 4L, 5L, 4L, 5L,
        4L, 4L, 4L, 4L
      )
    )

    w2 <- nlmixr2data::warfarin
    w2$dvid <- as.integer(w2$dvid) * 10
    expect_warning(tmp <- etTrans(w2, w, addCmt = TRUE))
    expect_equal(
      tmp$CMT[tmp$ID == 1],
      c(
        1L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 4L, 5L, 4L, 5L, 4L, 5L,
        4L, 4L, 4L, 4L
      )
    )

    w <- rxode2({
      ktr <- exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      ##
      logit <- exp(poplogit + eta.emax)
      emax <- logit / (1 + logit)
      ec50 <- exp(tec50 + eta.ec50)
      kout <- exp(tkout + eta.kout)
      e0 <- exp(te0 + eta.e0)
      ##
      DCP <- center / v
      PD <- 1 - emax * DCP / (ec50 + DCP)
      ##
      pca(0) <- e0
      kin <- e0 * kout
      ##
      d / dt(depot) <- -ktr * depot
      d / dt(gut) <- ktr * depot - ka * gut
      d / dt(center) <- ka * gut - cl / v * center
      d / dt(pca) <- kin * PD - kout * pca
      ##
      cp <- center / v
      cmt(cp)
      dvid(5)
    })

    w2 <- nlmixr2data::warfarin
    w2$dvid <- as.integer(w2$dvid) * 10
    expect_warning(tmp <- etTrans(w2, w, addCmt = TRUE))
    expect_equal(
      tmp$CMT[tmp$ID == 1],
      c(
        1L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 1L, 5L, 1L, 5L, 1L, 5L,
        1L, 1L, 1L, 1L
      )
    )

    w <- rxode2({
      ktr <- exp(tktr + eta.ktr)
      ka <- exp(tka + eta.ka)
      cl <- exp(tcl + eta.cl)
      ##
      logit <- exp(poplogit + eta.emax)
      emax <- logit / (1 + logit)
      ec50 <- exp(tec50 + eta.ec50)
      kout <- exp(tkout + eta.kout)
      e0 <- exp(te0 + eta.e0)
      ##
      DCP <- center / v
      PD <- 1 - emax * DCP / (ec50 + DCP)
      ##
      pca(0) <- e0
      kin <- e0 * kout
      ##
      d / dt(depot) <- -ktr * depot
      d / dt(gut) <- ktr * depot - ka * gut
      d / dt(center) <- ka * gut - cl / v * center
      d / dt(pca) <- kin * PD - kout * pca
      ##
      cp <- center / v
      cmt(cp)
      dvid(4, 5)
    })

    w2 <- nlmixr2data::warfarin
    w2$dvid <- as.integer(w2$dvid)
    tmp <- etTrans(w2, w, addCmt = TRUE)
    expect_equal(
      tmp$CMT[tmp$ID == 1],
      c(
        1L, 4L, 4L, 4L, 4L, 4L, 4L, 4L, 4L, 5L, 4L, 5L, 4L, 5L, 4L,
        5L, 5L, 5L, 5L
      )
    )

    ## Warfarin model
    inner <- rxode2({
      cmt(depot)
      cmt(gut)
      cmt(center)
      cmt(effect)
      rx_expr_001 ~ -depot
      rx_expr_002 ~ -center
      rx_expr_003 ~ -ETA[4]
      rx_expr_004 ~ -effect
      rx_expr_005 ~ -2 * ETA[4]
      rx_expr_006 ~ 2 * THETA[4]
      rx_expr_007 ~ ETA[1] + THETA[1]
      rx_expr_008 ~ ETA[2] + THETA[2]
      rx_expr_009 ~ ETA[3] + THETA[3]
      rx_expr_010 ~ ETA[7] + THETA[9]
      rx_expr_011 ~ ETA[5] + THETA[7]
      rx_expr_012 ~ ETA[6] + THETA[8]
      rx_expr_013 ~ rx_expr_003 - THETA[4]
      rx_expr_014 ~ ETA[8] + THETA[10]
      rx_expr_015 ~ exp(rx_expr_007)
      rx_expr_016 ~ exp(rx_expr_008)
      rx_expr_017 ~ exp(rx_expr_009)
      rx_expr_018 ~ exp(rx_expr_010)
      rx_expr_019 ~ exp(rx_expr_011)
      rx_expr_020 ~ exp(rx_expr_012)
      rx_expr_021 ~ rx_expr_005 - rx_expr_006
      rx_expr_022 ~ Rx_pow_di(center, 2)
      rx_expr_023 ~ exp(rx_expr_013)
      rx_expr_024 ~ exp(rx_expr_014)
      rx_expr_025 ~ rx_expr_019 + 1
      rx_expr_026 ~ Rx_pow_di(THETA[5], 2)
      rx_expr_027 ~ gut * rx_expr_016
      rx_expr_028 ~ 2 * rx_expr_026
      rx_expr_029 ~ exp(rx_expr_021)
      rx_expr_030 ~ depot * rx_expr_015
      rx_expr_031 ~ rx_expr_001 * rx_expr_015
      rx_expr_032 ~ rx_expr_002 * rx_expr_017
      rx_expr_033 ~ rx_expr_004 * rx_expr_018
      rx_expr_034 ~ center * rx_expr_023
      rx_expr_035 ~ rx_expr_002 * rx_expr_023
      rx_expr_036 ~ rx_expr_028 * center
      rx_expr_037 ~ center * rx_expr_029
      rx_expr_038 ~ rx_expr_023 * rx_expr_019
      rx_expr_039 ~ rx__sens_gut_BY_ETA_1___ * rx_expr_016
      rx_expr_040 ~ rx__sens_gut_BY_ETA_2___ * rx_expr_016
      rx_expr_041 ~ rx__sens_gut_BY_ETA_3___ * rx_expr_016
      rx_expr_042 ~ rx__sens_gut_BY_ETA_4___ * rx_expr_016
      rx_expr_043 ~ rx__sens_gut_BY_ETA_5___ * rx_expr_016
      rx_expr_044 ~ rx__sens_gut_BY_ETA_6___ * rx_expr_016
      rx_expr_045 ~ rx__sens_gut_BY_ETA_7___ * rx_expr_016
      rx_expr_046 ~ rx__sens_gut_BY_ETA_8___ * rx_expr_016
      rx_expr_047 ~ rx__sens_depot_BY_ETA_1___ * rx_expr_015
      rx_expr_048 ~ rx_expr_034 + rx_expr_020
      rx_expr_049 ~ rx_expr_034 * rx_expr_019
      rx_expr_050 ~ rx_expr_032 * rx_expr_023
      rx_expr_051 ~ rx_expr_035 * rx_expr_019
      rx_expr_052 ~ rx_expr_037 * rx_expr_019
      rx_expr_053 ~ Rx_pow_di(rx_expr_048, 2)
      rx_expr_054 ~ (rx_expr_048) * (rx_expr_025)
      rx_expr_055 ~ rx_expr_053 * (rx_expr_025)
      rx_expr_056 ~ rx_expr_038 / (rx_expr_054)
      rx_expr_057 ~ rx_expr_051 / (rx_expr_054)
      rx_expr_058 ~ rx_expr_057 + 1
      rx_expr_059 ~ rx_expr_052 / (rx_expr_055)
      rx_expr_060 ~ (rx_expr_058) * rx_expr_018
      rx_expr_061 ~ rx_expr_060 * rx_expr_024
      rx_expr_062 ~ rx_expr_059 - rx_expr_056
      d / dt(depot) <- rx_expr_031
      d / dt(gut) <- rx_expr_030 - rx_expr_027
      d / dt(center) <- rx_expr_050 + rx_expr_027
      effect(0) <- rx_expr_024
      d / dt(effect) <- rx_expr_033 + rx_expr_061
      d / dt(rx__sens_depot_BY_ETA_1___) ~ rx_expr_031 - rx_expr_047
      d / dt(rx__sens_depot_BY_ETA_2___) ~ -rx__sens_depot_BY_ETA_2___ * rx_expr_015
      d / dt(rx__sens_depot_BY_ETA_3___) ~ -rx__sens_depot_BY_ETA_3___ * rx_expr_015
      d / dt(rx__sens_depot_BY_ETA_4___) ~ -rx__sens_depot_BY_ETA_4___ * rx_expr_015
      d / dt(rx__sens_depot_BY_ETA_5___) ~ -rx__sens_depot_BY_ETA_5___ * rx_expr_015
      d / dt(rx__sens_depot_BY_ETA_6___) ~ -rx__sens_depot_BY_ETA_6___ * rx_expr_015
      d / dt(rx__sens_depot_BY_ETA_7___) ~ -rx__sens_depot_BY_ETA_7___ * rx_expr_015
      d / dt(rx__sens_depot_BY_ETA_8___) ~ -rx__sens_depot_BY_ETA_8___ * rx_expr_015
      d / dt(rx__sens_gut_BY_ETA_1___) ~ rx_expr_030 + rx_expr_047 - rx_expr_039
      d / dt(rx__sens_gut_BY_ETA_2___) ~ -gut * rx_expr_016 + rx__sens_depot_BY_ETA_2___ * rx_expr_015 - rx_expr_040
      d / dt(rx__sens_gut_BY_ETA_3___) ~ rx__sens_depot_BY_ETA_3___ * rx_expr_015 - rx_expr_041
      d / dt(rx__sens_gut_BY_ETA_4___) ~ rx__sens_depot_BY_ETA_4___ * rx_expr_015 - rx_expr_042
      d / dt(rx__sens_gut_BY_ETA_5___) ~ rx__sens_depot_BY_ETA_5___ * rx_expr_015 - rx_expr_043
      d / dt(rx__sens_gut_BY_ETA_6___) ~ rx__sens_depot_BY_ETA_6___ * rx_expr_015 - rx_expr_044
      d / dt(rx__sens_gut_BY_ETA_7___) ~ rx__sens_depot_BY_ETA_7___ * rx_expr_015 - rx_expr_045
      d / dt(rx__sens_gut_BY_ETA_8___) ~ rx__sens_depot_BY_ETA_8___ * rx_expr_015 - rx_expr_046
      d / dt(rx__sens_center_BY_ETA_1___) ~ -rx__sens_center_BY_ETA_1___ * rx_expr_017 * rx_expr_023 + rx_expr_039
      d / dt(rx__sens_center_BY_ETA_2___) ~ rx_expr_027 - rx__sens_center_BY_ETA_2___ * rx_expr_017 * rx_expr_023 + rx_expr_040
      d / dt(rx__sens_center_BY_ETA_3___) ~ rx_expr_050 - rx__sens_center_BY_ETA_3___ * rx_expr_017 * rx_expr_023 + rx_expr_041
      d / dt(rx__sens_center_BY_ETA_4___) ~ center * rx_expr_017 * rx_expr_023 - rx__sens_center_BY_ETA_4___ * rx_expr_017 * rx_expr_023 + rx_expr_042
      d / dt(rx__sens_center_BY_ETA_5___) ~ -rx__sens_center_BY_ETA_5___ * rx_expr_017 * rx_expr_023 + rx_expr_043
      d / dt(rx__sens_center_BY_ETA_6___) ~ -rx__sens_center_BY_ETA_6___ * rx_expr_017 * rx_expr_023 + rx_expr_044
      d / dt(rx__sens_center_BY_ETA_7___) ~ -rx__sens_center_BY_ETA_7___ * rx_expr_017 * rx_expr_023 + rx_expr_045
      d / dt(rx__sens_center_BY_ETA_8___) ~ -rx__sens_center_BY_ETA_8___ * rx_expr_017 * rx_expr_023 + rx_expr_046
      d / dt(rx__sens_effect_BY_ETA_1___) ~ rx__sens_center_BY_ETA_1___ * (rx_expr_062) * rx_expr_018 * rx_expr_024 - rx__sens_effect_BY_ETA_1___ * rx_expr_018
      d / dt(rx__sens_effect_BY_ETA_2___) ~ rx__sens_center_BY_ETA_2___ * (rx_expr_062) * rx_expr_018 * rx_expr_024 - rx__sens_effect_BY_ETA_2___ * rx_expr_018
      d / dt(rx__sens_effect_BY_ETA_3___) ~ rx__sens_center_BY_ETA_3___ * (rx_expr_062) * rx_expr_018 * rx_expr_024 - rx__sens_effect_BY_ETA_3___ * rx_expr_018
      d / dt(rx__sens_effect_BY_ETA_4___) ~ rx__sens_center_BY_ETA_4___ * (rx_expr_062) * rx_expr_018 * rx_expr_024 - rx__sens_effect_BY_ETA_4___ * rx_expr_018 + (-rx_expr_022 * rx_expr_029 * rx_expr_019 / (rx_expr_055) + rx_expr_049 / (rx_expr_054)) * rx_expr_018 * rx_expr_024
      d / dt(rx__sens_effect_BY_ETA_5___) ~ rx__sens_center_BY_ETA_5___ * (rx_expr_062) * rx_expr_018 * rx_expr_024 - rx__sens_effect_BY_ETA_5___ * rx_expr_018 + (rx_expr_057 + rx_expr_034 * exp(2 * ETA[5] + 2 * THETA[7]) / ((rx_expr_048) * Rx_pow_di(rx_expr_025, 2))) * rx_expr_018 * rx_expr_024
      d / dt(rx__sens_effect_BY_ETA_6___) ~ rx_expr_049 * rx_expr_020 * rx_expr_018 * rx_expr_024 / (rx_expr_055) + rx__sens_center_BY_ETA_6___ * (rx_expr_062) * rx_expr_018 * rx_expr_024 - rx__sens_effect_BY_ETA_6___ * rx_expr_018
      d / dt(rx__sens_effect_BY_ETA_7___) ~ rx_expr_033 + rx__sens_center_BY_ETA_7___ * (rx_expr_062) * rx_expr_018 * rx_expr_024 - rx__sens_effect_BY_ETA_7___ * rx_expr_018 + rx_expr_061
      rx__sens_effect_BY_ETA_8___(0) <- rx_expr_024
      d / dt(rx__sens_effect_BY_ETA_8___) ~ rx__sens_center_BY_ETA_8___ * (rx_expr_062) * rx_expr_018 * rx_expr_024 - rx__sens_effect_BY_ETA_8___ * rx_expr_018 + rx_expr_061
      rx_yj_ ~ 2
      rx_lambda_ ~ 1
      if (CMT == 3) {
        rx_pred_ <- rx_expr_034
        rx__sens_rx_pred__BY_ETA_1___ <- rx__sens_center_BY_ETA_1___ * rx_expr_023
        rx__sens_rx_pred__BY_ETA_2___ <- rx__sens_center_BY_ETA_2___ * rx_expr_023
        rx__sens_rx_pred__BY_ETA_3___ <- rx__sens_center_BY_ETA_3___ * rx_expr_023
        rx__sens_rx_pred__BY_ETA_4___ <- rx_expr_035 + rx__sens_center_BY_ETA_4___ * rx_expr_023
        rx__sens_rx_pred__BY_ETA_5___ <- rx__sens_center_BY_ETA_5___ * rx_expr_023
        rx__sens_rx_pred__BY_ETA_6___ <- rx__sens_center_BY_ETA_6___ * rx_expr_023
        rx__sens_rx_pred__BY_ETA_7___ <- rx__sens_center_BY_ETA_7___ * rx_expr_023
        rx__sens_rx_pred__BY_ETA_8___ <- rx__sens_center_BY_ETA_8___ * rx_expr_023
        rx_r_ <- rx_expr_026 * rx_expr_022 * rx_expr_029 + Rx_pow_di(THETA[6], 2)
        rx__sens_rx_r__BY_ETA_1___ <- rx_expr_036 * rx__sens_center_BY_ETA_1___ * rx_expr_029
        rx__sens_rx_r__BY_ETA_2___ <- rx_expr_036 * rx__sens_center_BY_ETA_2___ * rx_expr_029
        rx__sens_rx_r__BY_ETA_3___ <- rx_expr_036 * rx__sens_center_BY_ETA_3___ * rx_expr_029
        rx__sens_rx_r__BY_ETA_4___ <- -2 * rx_expr_026 * rx_expr_022 * rx_expr_029 + rx_expr_036 * rx__sens_center_BY_ETA_4___ * rx_expr_029
        rx__sens_rx_r__BY_ETA_5___ <- rx_expr_036 * rx__sens_center_BY_ETA_5___ * rx_expr_029
        rx__sens_rx_r__BY_ETA_6___ <- rx_expr_036 * rx__sens_center_BY_ETA_6___ * rx_expr_029
        rx__sens_rx_r__BY_ETA_7___ <- rx_expr_036 * rx__sens_center_BY_ETA_7___ * rx_expr_029
        rx__sens_rx_r__BY_ETA_8___ <- rx_expr_036 * rx__sens_center_BY_ETA_8___ * rx_expr_029
      }
      if (CMT == 4) {
        rx_pred_ <- effect
        rx__sens_rx_pred__BY_ETA_1___ <- rx__sens_effect_BY_ETA_1___
        rx__sens_rx_pred__BY_ETA_2___ <- rx__sens_effect_BY_ETA_2___
        rx__sens_rx_pred__BY_ETA_3___ <- rx__sens_effect_BY_ETA_3___
        rx__sens_rx_pred__BY_ETA_4___ <- rx__sens_effect_BY_ETA_4___
        rx__sens_rx_pred__BY_ETA_5___ <- rx__sens_effect_BY_ETA_5___
        rx__sens_rx_pred__BY_ETA_6___ <- rx__sens_effect_BY_ETA_6___
        rx__sens_rx_pred__BY_ETA_7___ <- rx__sens_effect_BY_ETA_7___
        rx__sens_rx_pred__BY_ETA_8___ <- rx__sens_effect_BY_ETA_8___
        rx_r_ <- Rx_pow_di(THETA[11], 2)
        rx__sens_rx_r__BY_ETA_1___ <- 0
        rx__sens_rx_r__BY_ETA_2___ <- 0
        rx__sens_rx_r__BY_ETA_3___ <- 0
        rx__sens_rx_r__BY_ETA_4___ <- 0
        rx__sens_rx_r__BY_ETA_5___ <- 0
        rx__sens_rx_r__BY_ETA_6___ <- 0
        rx__sens_rx_r__BY_ETA_7___ <- 0
        rx__sens_rx_r__BY_ETA_8___ <- 0
      }
      cmt(cp)
      cmt(pca)
      dvid(3, 4, 5, 6)
    })

    # context("warfarin inner test")

    data.pkpd <- nlmixr2data::warfarin
    data.pkpd$dvid <- as.integer(data.pkpd$dvid)
    data.pkpd$cmt <- paste(factor(data.pkpd$dvid, c(1, 2), c("center", "effect")))
    data.pkpd$cmt[data.pkpd$amt > 0] <- "depot"

    t2 <- etTrans(data.pkpd[, names(data.pkpd) != "dvid"], inner)

    expect_equal(
      t2$CMT[t2$ID == 1],
      c(
        1L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 4L, 3L, 4L, 3L, 4L, 3L,
        4L, 4L, 4L, 4L
      )
    )

    data.pkpd$cmt <- paste(factor(data.pkpd$dvid, c(1, 2), c("cp", "pca")))
    data.pkpd$cmt[data.pkpd$amt > 0] <- "depot"

    t2 <- etTrans(data.pkpd[, names(data.pkpd) != "dvid"], inner)

    expect_equal(
      t2$CMT[t2$ID == 1],
      c(
        1L, 37L, 37L, 37L, 37L, 37L, 37L, 37L, 37L, 38L, 37L, 38L, 37L, 38L, 37L,
        38L, 38L, 38L, 38L
      )
    )

    data.pkpd$cmt <- factor(data.pkpd$cmt)

    t2 <- etTrans(data.pkpd[, names(data.pkpd) != "dvid"], inner)

    expect_equal(
      t2$CMT[t2$ID == 1],
      c(
        1L, 37L, 37L, 37L, 37L, 37L, 37L, 37L, 37L, 38L, 37L, 38L, 37L, 38L, 37L,
        38L, 38L, 38L, 38L
      )
    )

    data.pkpd$cmt <- paste(data.pkpd$cmt)
    .cmt <- data.pkpd$cmt
    data.pkpd$cmt <- 1
    data.pkpd$cmt[.cmt == "cp"] <- 5
    data.pkpd$cmt[.cmt == "pca"] <- 6

    ## This skips the sensitivity equations
    t2 <- etTrans(data.pkpd[, names(data.pkpd) != "dvid"], inner)

    expect_equal(
      t2$CMT[t2$ID == 1],
      c(
        1L, 37L, 37L, 37L, 37L, 37L, 37L, 37L, 37L, 38L, 37L, 38L, 37L, 38L, 37L,
        38L, 38L, 38L, 38L
      )
    )

    data.pkpd$cmt <- 1
    data.pkpd$cmt[.cmt == "cp"] <- 3
    data.pkpd$cmt[.cmt == "pca"] <- 4

    t2 <- etTrans(data.pkpd[, names(data.pkpd) != "dvid"], inner)

    expect_equal(
      t2$CMT[t2$ID == 1],
      c(
        1L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 4L, 3L, 4L, 3L, 4L, 3L,
        4L, 4L, 4L, 4L
      )
    )

    t2 <- etTrans(data.pkpd, inner)

    expect_equal(
      t2$CMT[t2$ID == 1],
      c(
        1L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 3L, 4L, 3L, 4L, 3L, 4L, 3L,
        4L, 4L, 4L, 4L
      )
    )

    data.pkpd$dvid[data.pkpd$dvid == 1] <- 3
    data.pkpd$dvid[data.pkpd$dvid == 2] <- 4

    expect_error(etTrans(data.pkpd, inner))

    t2 <- etTrans(data.pkpd[, names(data.pkpd) != "cmt"], inner, addCmt = TRUE)

    expect_equal(
      t2$CMT[t2$ID == 1],
      c(
        1L, 37L, 37L, 37L, 37L, 37L, 37L, 37L, 37L, 38L, 37L, 38L, 37L, 38L, 37L,
        38L, 38L, 38L, 38L
      )
    )

    ## TEST NA cmt/dvid w/factors
    ## TEST NA strings for cmt/dvid

    tmp <- rxode2({
      d / dt(depot) <- -ka * depot
      d / dt(center) <- ka * depot - cl / v * center
      cp <- center / v
      cp2 <- cp + nlmixr_extra_par
      cmt(central)
      cmt(c20)
    })

    expect_s3_class(tmp, "rxode2")

    # context("Check lhs allowed stateExtra while preserving lhs properties.")

    tmp <- rxode2({
      d / dt(depot) <- -ka * depot
      d / dt(center) <- ka * depot - cl / v * center
      cp <- center / v
      cmt(cp)
    })

    expect_s3_class(tmp, "rxode2")
    expect_equal(tmp$lhs, "cp")
    expect_equal(tmp$stateExtra, "cp")

    tmp <- rxode2({
      d / dt(depot) <- -ka * depot
      d / dt(center) <- ka * depot - cl / v * center
      cmt(cp)
      cp <- center / v
    })

    expect_s3_class(tmp, "rxode2")
    expect_equal(tmp$lhs, "cp")
    expect_equal(tmp$stateExtra, "cp")

    tmp <- rxode2({
      d / dt(depot) <- -ka * depot
      d / dt(center) <- ka * depot - cl / v * center
      cmt(cp)
      cp <- 3
    })

    expect_s3_class(tmp, "rxode2")
    expect_equal(tmp$lhs, "cp")
    expect_equal(tmp$stateExtra, "cp")

    tmp <- rxode2({
      d / dt(depot) <- -ka * depot
      d / dt(center) <- ka * depot - cl / v * center
      cp <- 3
      cmt(cp)
    })

    expect_s3_class(tmp, "rxode2")
    expect_equal(tmp$lhs, "cp")
    expect_equal(tmp$stateExtra, "cp")
  })

  test_that("cmt() and hidden lhs variables", {
    tmp <- rxode2({
      d / dt(depot) ~ -ka * depot
      d / dt(center) ~ ka * depot - cl / v * center
      cp ~ 3
      cmt(cp)
    })

    expect_s3_class(tmp, "rxode2")
    expect_equal(tmp$lhs, character(0))
    expect_equal(tmp$stateExtra, "cp")
  })

  test_that("compartment ordering does not affect c-code generation #785",
  {

    f1 <- function() {
      ini({
        TVCL <- c(0, 0.0308628127403366)
        TVV1 <- c(0, 0.132584062223269)
        TVQ <- c(0, 0.0107718034646091)
        TVV2 <- c(0, 0.0613517707801907)
        TVR0 <- c(0, 3.94111989610973)
        TVKON <- c(0, 1.31095109796496)
        TVKOFF <- c(0, 0.0758072298659602)
        TVKDEG <- c(0, 1.07290072658165)
        TVKINT <- c(0, 3.49422759360926)
        ADD.ERR.PK <- c(0, 0.446231462752061)
        ETA.CL ~ 0.0499998244205031
      })
      model({
        CL <- TVCL * exp(ETA.CL)
        V1 <- TVV1
        Q <- TVQ
        V2 <- TVV2
        R0 <- TVR0
        KON <- TVKON
        KOFF <- TVKOFF
        KDEG <- TVKDEG
        KINT <- TVKINT
        KEL <- CL/V1
        K12 <- Q/V1
        K21 <- Q/V2
        KD <- KOFF/KON
        KSYN <- R0 * KDEG
        A3(0) <- R0
        KEL <- 0
        d/dt(A1) = -(K12 + KEL) * A1 + K21 * A2 - KON * A1 *
          A3 + KOFF * A4 * V1
        d/dt(A2) = K12 * A1 - K21 * A2
        d/dt(A3) = KSYN - KDEG * A3 - KON * (A1/V1) * A3 + KOFF *
          A4
        d/dt(A4) = KON * (A1/V1) * A3 - (KINT + KOFF) * A4
        CFREE = A1/V1
        CTOT = CFREE + A4
        CP = log(CTOT)
        CP ~ add(ADD.ERR.PK)
        SPEC.OUT <- SPEC
        DGRP.OUT <- DGRP
        BW.OUT <- BW
      })
    }

    f1 <- f1 |> zeroRe("omega")

    f2 <- function() {
      ini({
        TVCL <- c(0, 0.0308628127403366)
        TVV1 <- c(0, 0.132584062223269)
        TVQ <- c(0, 0.0107718034646091)
        TVV2 <- c(0, 0.0613517707801907)
        TVR0 <- c(0, 3.94111989610973)
        TVKON <- c(0, 1.31095109796496)
        TVKOFF <- c(0, 0.0758072298659602)
        TVKDEG <- c(0, 1.07290072658165)
        TVKINT <- c(0, 3.49422759360926)
        ADD.ERR.PK <- c(0, 0.446231462752061)
        ETA.CL ~ 0.0499998244205031
      })
      model({
        cmt(A1)
        cmt(A2)
        cmt(A3)
        cmt(A4)
        CL <- TVCL * exp(ETA.CL)
        V1 <- TVV1
        Q <- TVQ
        V2 <- TVV2
        R0 <- TVR0
        KON <- TVKON
        KOFF <- TVKOFF
        KDEG <- TVKDEG
        KINT <- TVKINT
        KEL <- CL/V1
        K12 <- Q/V1
        K21 <- Q/V2
        KD <- KOFF/KON
        KSYN <- R0 * KDEG
        A3(0) <- R0
        KEL <- 0
        d/dt(A1) = -(K12 + KEL) * A1 + K21 * A2 - KON * A1 *
          A3 + KOFF * A4 * V1
        d/dt(A2) = K12 * A1 - K21 * A2
        d/dt(A3) = KSYN - KDEG * A3 - KON * (A1/V1) * A3 + KOFF *
          A4
        d/dt(A4) = KON * (A1/V1) * A3 - (KINT + KOFF) * A4
        CFREE = A1/V1
        CTOT = CFREE + A4
        CP = log(CTOT)
        CP ~ add(ADD.ERR.PK)
        SPEC.OUT <- SPEC
        DGRP.OUT <- DGRP
        BW.OUT <- BW
      })
    }

    f2 <- f2 |> zeroRe("omega")

    ev     <- eventTable()
    mybw   <- 70
    mymw   <- 150000
    mydgrp <- 50
    ev$add.dosing(dose=mydgrp*mybw*1000000/mymw ,nbr.doses= 13, dosing.interval= 7)
    ev$add.sampling(seq(0,13*7,length.out=7))

    ev2 <- ev |> as.data.frame() |>
      dplyr::mutate(BW=mybw, DGRP=mydgrp, SPEC=4)

    s2 <- rxSolve(f2, ev2, returnType="data.frame")

    s1 <- rxSolve(f1, ev2, returnType="data.frame")

    expect_equal(s1[,c("A1", "A2", "A3", "A4")],
                 s2[,c("A1", "A2", "A3", "A4")])

  })

})

rxTest({

  test_that("Parallel solve vs single vs for; Issue RxODE#450", {
    TV_CLr  <- 6.54  # L/h, (CLr/F)
    TV_CLnr <- 2.39  # L/h, (CLnr/F)
    TV_Vc   <- 95.1  # L, (V/F)
    TV_alag <- 0.145 # h,
    TV_D    <- 0.512 # h,
    TV_Q    <- 2.1   #L/h, (Q/F)
    TV_Vp   <- 23.3  #L, (Vp/F)

    OM_D_normal <- log((128/100)^2+1)
    D_trans     <- 0.0819
    OM_CLr      <- log((36.2/100)^2+1)
    OM_CLnr     <- log((43.6/100)^2+1)
    OM_Vc       <- log((14.4/100)^2+1)
    OM_Q        <- log((15.1/100)^2+1)
    OM_Vp       <- log((37.6/100)^2+1)
    OM_CLr_CLnr <- 0.101
    OM_CLr_Vc   <- 0.0066

    rwishart <- function(df, p = nrow(SqrtSigma), Sigma, SqrtSigma = diag(p)) {
      if (!missing(Sigma)) {
        tmp <- svd(Sigma)
        SqrtSigma <- sqrt(tmp$d) * t(tmp$u)
      }
      if ((Ident <- missing(SqrtSigma)) && missing(p))
        stop("either p, Sigma or SqrtSigma must be specified")
      Z <- matrix(0, p, p)
      diag(Z) <- sqrt(rchisq(p, df:(df - p + 1)))
      if (p > 1) {
        pseq <- 1:(p - 1)
        Z[rep(p * pseq, pseq) + unlist(lapply(pseq, seq))] <- rnorm(p *
                                                                      (p - 1)/2)
      }
      if (Ident)
        crossprod(Z)
      else crossprod(Z %*% SqrtSigma)
    }

    sample.etas <- function(df, R, n) {
      R.inv = solve(df*R)
      R0 = chol(.5*(R.inv+t(R.inv)))  ## R0 is an upper-diag matrix
      R1 = solve(rwishart(df, SqrtSigma=R0))
      eta = rxRmvn(n = n, mu= rep(0, nrow(R)), sigma= .5*(R1+t(R1)))
      eta
    }

    rxWithSeed(100, {

      nsubj <- 200

      n.pts.pk <- 300  # number of subjects in PK model fitting

      eta_CLr_CLnr_Vc <- as.data.frame(sample.etas(df = n.pts.pk,
                                                   R = lotri({eta_CLr + eta_CLnr + eta_Vc ~
                                                                c(OM_CLr,
                                                                  OM_CLr_CLnr, OM_CLnr,
                                                                  OM_CLr_Vc, 0, OM_Vc)
                                                   }),
                                                   n = nsubj))

      eta_D_trans <- data.frame(eta_D_normal = rnorm(mean=0,sd=sqrt(OM_D_normal),n=nsubj)) |>
        dplyr::mutate(eta_D = ((exp(eta_D_normal))^D_trans-1)/D_trans)

      par.pk <- data.frame(sim.id = seq(nsubj),
                           D = TV_D * exp(eta_D_trans$eta_D),
                           CLr = TV_CLr * exp(eta_CLr_CLnr_Vc$eta_CLr),
                           CLnr = TV_CLnr * exp(eta_CLr_CLnr_Vc$eta_CLnr),
                           Vc = TV_Vc * exp(eta_CLr_CLnr_Vc$eta_Vc),
                           Vp = TV_Vp * exp(rnorm(nsubj, 0, sqrt(OM_Vp))),
                           Q = TV_Q * exp(rnorm(nsubj, 0, sqrt(OM_Q))),
                           alag = TV_alag)

      # zero-order absorption with lag time 2-compartment
      mod <- rxode2({
        CL <- CLr+CLnr
        C2 <- central/Vc*1000
        all<- central+periph+output
        d/dt(central) <- - CL/Vc*central - Q/Vc*central + Q/Vp*periph
        d/dt(periph)  <- Q/Vc*central - Q/Vp*periph
        d/dt(output)  <- CL/Vc*central
        alag(central) <- alag
        dur(central) <- D
      })

      ev <-  et(amt=2,cmt="central",rate=-2,ii=24,addl=4) |>
        et(seq(0,120,0.1))

      bar2x <- rxSolve(mod, ev, params=par.pk, cores=2L, returnType="data.frame")

      bar1x <- rxSolve(mod, ev, params=par.pk, cores=1L, returnType="data.frame")

      expect_equal(bar1x, bar2x)

      bar3x <- rxSolve(mod, ev, params=par.pk)

      # add for loop for testing against
      res.all = NULL

      for (id in seq(nsubj)) {
        # now combine
        ev.new <- eventTable() |>
          add.dosing(dose = 2, nbr.doses = 5, dosing.interval = 24, dur = par.pk$D[par.pk$sim.id == id]) |>
          add.sampling(seq(0,120,0.1))
        theta <- c(CLr = par.pk$CLr[par.pk$sim.id == id],
                   CLnr = par.pk$CLnr[par.pk$sim.id == id],
                   Vc = par.pk$Vc[par.pk$sim.id == id],
                   Q = par.pk$Q[par.pk$sim.id == id],
                   Vp = par.pk$Vp[par.pk$sim.id == id],
                   alag = par.pk$alag[par.pk$sim.id == id],
                   D = par.pk$D[par.pk$sim.id == id])
        res.id = data.frame(sim.id=id, rxSolve(mod, theta, ev.new, returnType="data.frame"))

        expect_equal(theta, unlist(bar3x$params[bar3x$params$sim.id == id, -1]))

        ## tmp  <- rbind(data.frame(res.id, type="for"), data.frame(bar3x[bar3x$sim.id == id,], type="single"))
        ## ggplot(tmp, aes(time, C2, col=type)) + geom_line()

        row.names(res.id) <- NULL
        res2 <- as.data.frame(bar3x[bar3x$sim.id == id,])
        row.names(res2) <- NULL
        expect_equal(res.id, res2, tolerance=1e-4)

        res.all = rbind(res.all, res.id)
      }
      expect_equal(bar2x, res.all)
    })

  })
})

rxTest({
  skip_if_not_installed("units")
  # context("tad family of functions with odes")

  test_that("tad family works with ode", {

    mod1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      fDepot <- 1
      fPeri <- 1
      durDepot <- 1
      rateDepot <- 1
      durPeri <- 1
      ratePeri <- 1
      C2 <- centr / V2
      C3 <- peri / V3
      d/dt(depot) <- -KA * depot
      d/dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d/dt(peri) <- Q * C2 - Q * C3
      d/dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      f(depot) <- fDepot
      f(peri) <- fPeri
      dur(depot) = durDepot
      rate(depot) = rateDepot
      dur(peri) = durPeri
      rate(peri) = ratePeri
      ## TAD tests
      tad <- tad()
      tad0 <- tad0()
      dosen <- dosenum()
      tl <- tlast()
      tl0 <- tlast0()
      tafd <- tafd()
      tafd0 <- tafd0()
      tfirst <- tfirst()
      tfirst0 <- tfirst0()
      tadd <- tad(depot)
      tadd0 <- tad0(depot)
      tfirstd <- tfirst(depot)
      tfirstd0 <- tfirst0(depot)
      tlastd <- tlast(depot)
      tlastd0 <- tlast0(depot)
      tafdd <- tafd(depot)
      tafdd0 <- tafd0(depot)
      tadp <- tad(peri)
      tadp0 <- tad0(peri)
      tafdp <- tafd(peri)
      tafdp0 <- tafd0(peri)
      tfirstp <- tfirst(peri)
      tfirstp0 <- tfirst0(peri)
      tlastp <- tlast(peri)
      tlastp0 <-  tlast0(peri)
      dose <- dose()
      dosed <- dose(depot)
      dosep <- dose(peri)
    })

    ev <- et(amountUnits = "mg", timeUnits = "hours") |>
      et(time = 1, amt = 10000, addl = 9, ii = 12, cmt = "depot") |>
      et(time = 120, amt = 2000, addl = 4, ii = 14, cmt = "depot") |>
      et(time = 122, amt = 2200, addl = 4, ii = 14, cmt = "peri") |>
      et(0, 240, by = 3)

    r1 <- rxSolve(mod1, ev, addDosing = TRUE)

    r <- c(NA, 10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000,
            10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000,
            10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000,
            10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000,
            10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000,
            10000, 10000, 10000, 10000, 10000, 2000, 2000, 2200, 2200, 2200,
            2200, 2200, 2000, 2000, 2200, 2200, 2200, 2200, 2200, 2000, 2200,
            2200, 2200, 2200, 2200, 2000, 2000, 2200, 2200, 2200, 2200, 2200,
            2000, 2000, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200,
            2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200,
            2200, 2200)

    expect_equal(r1$dosed, c(NA, 10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000,
                             10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000,
                             10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000,
                             10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000,
                             10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000, 10000,
                             10000, 10000, 10000, 10000, 10000, 2000, 2000, 2000, 2000, 2000,
                             2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000,
                             2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000,
                             2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000,
                             2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000,
                             2000, 2000))

    expect_equal(r1$dosep, c(NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                             NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                             NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
                             NA, NA, NA, NA, NA, 2200, 2200, 2200, 2200, 2200, 2200, 2200,
                             2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200,
                             2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200,
                             2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200,
                             2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200, 2200))

    expect_equal(r1$dosen, c(
      0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4,
      4, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8,
      9, 9, 9, 9, 9, 10, 10, 10, 10, 11, 11, 12, 12, 12, 12, 12, 13,
      13, 14, 14, 14, 14, 14, 15, 16, 16, 16, 16, 16, 17, 17, 18, 18,
      18, 18, 18, 19, 19, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20
    ))
    r <- c(
      NA, 0, 2, 5, 8, 11, 0, 2, 5, 8, 11, 0, 2, 5, 8, 11, 0, 2, 5,
      8, 11, 0, 2, 5, 8, 11, 0, 2, 5, 8, 11, 0, 2, 5, 8, 11, 0, 2,
      5, 8, 11, 0, 2, 5, 8, 11, 0, 2, 5, 8, 0, 0, 0, 1, 4, 7, 10, 0,
      1, 0, 2, 5, 8, 11, 0, 0, 0, 3, 6, 9, 0, 0, 0, 1, 4, 7, 10, 0,
      1, 0, 2, 5, 8, 11, 14, 17, 20, 23, 26, 29, 32, 35, 38, 41, 44,
      47, 50, 53, 56, 59, 62)

    expect_equal(r1$tad, r)

    r <- ifelse(is.na(r), 0, r)

    expect_equal(r1$tad0, r)

    r <- c(
      NA, 0, 2, 5, 8, 11, 12, 14, 17, 20, 23, 24, 26, 29, 32, 35,
      36, 38, 41, 44, 47, 48, 50, 53, 56, 59, 60, 62, 65, 68, 71, 72,
      74, 77, 80, 83, 84, 86, 89, 92, 95, 96, 98, 101, 104, 107, 108,
      110, 113, 116, 119, 119, 121, 122, 125, 128, 131, 133, 134, 135,
      137, 140, 143, 146, 147, 149, 149, 152, 155, 158, 161, 161, 163,
      164, 167, 170, 173, 175, 176, 177, 179, 182, 185, 188, 191, 194,
      197, 200, 203, 206, 209, 212, 215, 218, 221, 224, 227, 230, 233,
      236, 239)

    expect_equal(r1$tafd, r)

    r <- ifelse(is.na(r), 0, r)
    expect_equal(r1$tafd0, r)

    r <- c(
      NA, 0, 2, 5, 8, 11, 0, 2, 5, 8, 11, 0, 2, 5, 8, 11, 0, 2, 5,
      8, 11, 0, 2, 5, 8, 11, 0, 2, 5, 8, 11, 0, 2, 5, 8, 11, 0, 2,
      5, 8, 11, 0, 2, 5, 8, 11, 0, 2, 5, 8, 0, 0, 2, 3, 6, 9, 12, 0,
      1, 2, 4, 7, 10, 13, 0, 2, 2, 5, 8, 11, 0, 0, 2, 3, 6, 9, 12,
      0, 1, 2, 4, 7, 10, 13, 16, 19, 22, 25, 28, 31, 34, 37, 40, 43,
      46, 49, 52, 55, 58, 61, 64)

    expect_equal(r1$tadd, r)

    r <- ifelse(is.na(r), 0, r)
    expect_equal(r1$tadd0, r)


    r <- c(
      NA, 0, 2, 5, 8, 11, 12, 14, 17, 20, 23, 24, 26, 29, 32, 35,
      36, 38, 41, 44, 47, 48, 50, 53, 56, 59, 60, 62, 65, 68, 71, 72,
      74, 77, 80, 83, 84, 86, 89, 92, 95, 96, 98, 101, 104, 107, 108,
      110, 113, 116, 119, 119, 121, 122, 125, 128, 131, 133, 134, 135,
      137, 140, 143, 146, 147, 149, 149, 152, 155, 158, 161, 161, 163,
      164, 167, 170, 173, 175, 176, 177, 179, 182, 185, 188, 191, 194,
      197, 200, 203, 206, 209, 212, 215, 218, 221, 224, 227, 230, 233,
      236, 239)

    expect_equal(r1$tafdd, r)

    r <- ifelse(is.na(r), 0, r)
    expect_equal(r1$tafdd0, r)

    r <- c(
      NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
      NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
      NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
      NA, NA, NA, NA, NA, 0, 1, 4, 7, 10, 12, 13, 0, 2, 5, 8, 11, 12,
      0, 0, 3, 6, 9, 12, 12, 0, 1, 4, 7, 10, 12, 13, 0, 2, 5, 8, 11,
      14, 17, 20, 23, 26, 29, 32, 35, 38, 41, 44, 47, 50, 53, 56, 59,
      62)

    expect_equal(r1$tadp, r)

    r <- ifelse(is.na(r), 0, r)
    expect_equal(r1$tadp0, r)

    expect_equal(r1$tfirst[1], NA_real_)
    expect_true(all(r1$tfirst[-1] == 1))

    r <- c(
      NA, 1, 1, 1, 1, 1, 13, 13, 13, 13, 13, 25, 25, 25, 25, 25,
      37, 37, 37, 37, 37, 49, 49, 49, 49, 49, 61, 61, 61, 61, 61, 73,
      73, 73, 73, 73, 85, 85, 85, 85, 85, 97, 97, 97, 97, 97, 109,
      109, 109, 109, 120, 120, 122, 122, 122, 122, 122, 134, 134, 136,
      136, 136, 136, 136, 148, 150, 150, 150, 150, 150, 162, 162, 164,
      164, 164, 164, 164, 176, 176, 178, 178, 178, 178, 178, 178, 178,
      178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178,
      178, 178)

    expect_equal(
      r1$tl,r)

    r <- ifelse(is.na(r), 0, r)
    expect_equal(r1$tl0,r)

    r <- c(
      NA, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1)

    expect_equal(r1$tfirstd, r)

    r <- ifelse(is.na(r), 0, r)
    expect_equal(r1$tfirstd0, r)

    r <- c(
      NA, 1, 1, 1, 1, 1, 13, 13, 13, 13, 13, 25, 25, 25, 25, 25,
      37, 37, 37, 37, 37, 49, 49, 49, 49, 49, 61, 61, 61, 61, 61, 73,
      73, 73, 73, 73, 85, 85, 85, 85, 85, 97, 97, 97, 97, 97, 109,
      109, 109, 109, 120, 120, 120, 120, 120, 120, 120, 134, 134, 134,
      134, 134, 134, 134, 148, 148, 148, 148, 148, 148, 162, 162, 162,
      162, 162, 162, 162, 176, 176, 176, 176, 176, 176, 176, 176, 176,
      176, 176, 176, 176, 176, 176, 176, 176, 176, 176, 176, 176, 176,
      176, 176)

    expect_equal(r1$tlastd, r)
    r <- ifelse(is.na(r), 0, r)

    expect_equal(r1$tlastd0, r)

    r <- c(
      NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
      NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
      NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
      NA, NA, NA, NA, NA, 122, 122, 122, 122, 122, 122, 122, 136, 136,
      136, 136, 136, 136, 150, 150, 150, 150, 150, 150, 150, 164, 164,
      164, 164, 164, 164, 164, 178, 178, 178, 178, 178, 178, 178, 178,
      178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178,
      178)
    expect_equal(
      r1$tlastp, r)

    r <- ifelse(is.na(r), 0, r)
    expect_equal(r1$tlastp0, r)
    r <- c(NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
      NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
      NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
      NA, NA, NA, NA, NA, 122, 122, 122, 122, 122, 122, 122, 122, 122,
      122, 122, 122, 122, 122, 122, 122, 122, 122, 122, 122, 122, 122,
      122, 122, 122, 122, 122, 122, 122, 122, 122, 122, 122, 122, 122,
      122, 122, 122, 122, 122, 122, 122, 122, 122, 122, 122, 122, 122,
      122)

    expect_equal(r1$tfirstp, r)

    r <-  ifelse(is.na(r), 0, r)
    expect_equal(r1$tfirstp0, r)

    f <- function(r1, r2) {
      expect_equal(r1$dose, r2$dose)
      expect_equal(r1$dosed, r2$dosed)
      expect_equal(r1$dosep, r2$dosep)
      expect_equal(r1$dosen, r2$dosen)
      expect_equal(r1$tad, r2$tad)
      expect_equal(r1$tafd, r2$tafd)
      expect_equal(r1$tadd, r2$tadd)
      expect_equal(r1$tafdd, r2$tafdd)
      expect_equal(r1$tadp, r2$tadp)
      expect_equal(r1$tl, r2$tl)
      expect_equal(r1$tfirst, r2$tfirst)
      expect_equal(r1$tfirstd, r2$tfirstd)
      expect_equal(r1$tlastd, r2$tlastd)
      expect_equal(r1$tlastp, r2$tlastp)
      expect_equal(r1$tfirstp, r2$tfirstp)
      expect_equal(r1$tad0, r2$tad0)
      expect_equal(r1$tafd0, r2$tafd0)
      expect_equal(r1$tadd0, r2$tadd0)
      expect_equal(r1$tafdd0, r2$tafdd0)
      expect_equal(r1$tadp0, r2$tadp0)
      expect_equal(r1$tl0, r2$tl0)
      expect_equal(r1$tfirst0, r2$tfirst0)
      expect_equal(r1$tfirstd0, r2$tfirstd0)
      expect_equal(r1$tlastd0, r2$tlastd0)
      expect_equal(r1$tlastp0, r2$tlastp0)
      expect_equal(r1$tfirstp0, r2$tfirstp0)
    }


    r2 <- rxSolve(mod1, c(fDepot=0.5, fPeri=0.25), ev, addDosing = TRUE)

    f(r1, r2)

    ev <- et(amountUnits = "mg", timeUnits = "hours") |>
      et(time = 1, amt = 10000, rate=10000 / 6, addl = 9, ii = 12, cmt = "depot") |>
      et(time = 120, amt = 2000, rate=2000 / 6, addl = 4, ii = 14, cmt = "depot") |>
      et(time = 122, amt = 2200, rate=2200 / 6, addl = 4, ii = 14, cmt = "peri") |>
      et(0, 240, by = 3)

    r2 <- rxSolve(mod1, ev, addDosing = TRUE)

    f(r1, r2)

    r2 <- rxSolve(mod1, c(fDepot=0.5, fPeri=0.25), ev, addDosing = TRUE)

    f(r1, r2)

    ev <- et(amountUnits = "mg", timeUnits = "hours") |>
      et(time = 1, amt = 10000, dur=6, addl = 9, ii = 12, cmt = "depot") |>
      et(time = 120, amt = 2000, dur=6, addl = 4, ii = 14, cmt = "depot") |>
      et(time = 122, amt = 2200, dur=6, addl = 4, ii = 14, cmt = "peri") |>
      et(0, 240, by = 3)

    r3 <- rxSolve(mod1, ev, addDosing=TRUE)

    f(r1, r3)

    r2 <- rxSolve(mod1, c(fDepot=0.5, fPeri=0.25), ev, addDosing = TRUE)

    f(r1, r2)

    ev <- et(amountUnits = "mg", timeUnits = "hours") |>
      et(time = 1, amt = 10000, rate = -1, addl = 9, ii = 12, cmt = "depot") |>
      et(time = 120, amt = 2000, rate = -1, addl = 4, ii = 14, cmt = "depot") |>
      et(time = 122, amt = 2200, rate= -1, addl = 4, ii = 14, cmt = "peri") |>
      et(0, 240, by = 3)

    r1 <- rxSolve(mod1, ev, addDosing = TRUE)

    r2 <- rxSolve(mod1, c(fDepot=0.5, fPeri=0.25), ev, addDosing = TRUE)

    f(r1, r2)

    ev <- et(amountUnits = "mg", timeUnits = "hours") |>
      et(time = 1, amt = 10000, rate = -2, addl = 9, ii = 12, cmt = "depot") |>
      et(time = 120, amt = 2000, rate = -2, addl = 4, ii = 14, cmt = "depot") |>
      et(time = 122, amt = 2200, rate= -2, addl = 4, ii = 14, cmt = "peri") |>
      et(0, 240, by = 3)

    r1 <- rxSolve(mod1, ev, addDosing = TRUE)

    r2 <- rxSolve(mod1, c(fDepot=0.5, fPeri=0.25), ev, addDosing = TRUE)

    f(r1, r2)

  })

  test_that("test parsing of ode", {
    ## now change the option
      expect_error(rxode2({
        KA <- 2.94E-01
        CL <- 1.86E+01
        V2 <- 4.02E+01
        Q <- 1.05E+01
        V3 <- 2.97E+02
        Kin <- 1
        Kout <- 1
        EC50 <- 200
        C2 <- centr / V2
        C3 <- peri / V3
        ## error in capturing early
        f <- tad(eff)
        d / dt(depot) <- -KA * depot
        d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
        d / dt(peri) <- Q * C2 - Q * C3
        d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      }), NA)
  })

  # context("tad family of functions with linCmt()")

  test_that("lincmt solution tad family", {

    sol.1c.ka <- rxode2({
      KA <- 2
      V <- 20
      CL <- 25
      C2 <- linCmt(V, CL, KA)
      tad <- tad()
      tafd <- tafd()
      tadd <- tad(depot)
      tafdd <- tafd(depot)
      tadc <- tad(central)
      tafdc <- tafd(central)
      dose <- dose()
      dosed <- dose(depot)
      dosec <- dose(central)
    })

    et <- eventTable() |>
      add.dosing(dose = 3, nbr.doses = 6, dosing.interval = 8) |>
      add.dosing(
        dose = 6, nbr.doses = 6, dosing.interval = 8,
        start.time = 2, dosing.to = "central"
      ) |>
      add.sampling(seq(0, 48, length.out = 200))

    s1 <- rxSolve(sol.1c.ka, et)

    sol.1c.ka <- rxode2({
      KA <- 2
      V <- 20
      CL <- 25
      C2 <- linCmt(V, CL, KA)
      tad <- tad()
      tafd <- tafd()
      tadc <- tad(central)
      tafdc <- tafd(central)
      dose <- dose()
      dosed <- dose(depot)
      dosec <- dose(central)
    })

    s2 <- rxSolve(sol.1c.ka, et)

    expect_equal(s2$tad, s1$tad)
    expect_equal(s2$tafd, s1$tafd)
    expect_equal(s2$tadc, s1$tadc)
    expect_equal(s2$tafdc, s1$tafdc)
    expect_equal(s2$dose, s1$dose)
    expect_equal(s2$dosed, s1$dosed)
    expect_equal(s2$dosec, s1$dosec)

    expect_error(rxode2({
      V <- 20
      CL <- 25
      C2 <- linCmt(V, CL)
      tad <- tad()
      tafd <- tafd()
      tadd <- tad(depot)
      tafdd <- tafd(depot)
      tadc <- tad(central)
      tafdc <- tafd(central)
    }))

    one.cmt <- rxode2({
      V <- 20
      CL <- 25
      C2 <- linCmt(V, CL)
      tad <- tad()
      tafd <- tafd()
      tadc <- tad(central)
      tafdc <- tafd(central)
      dose <- dose()
      dosec <- dose(central)
    })

    et <- eventTable() |>
      add.dosing(dose = 3, nbr.doses = 6, dosing.interval = 8) |>
      add.sampling(seq(0, 48, length.out = 200))

    s <- rxSolve(one.cmt, et)

    expect_equal(s$tad, s$tadc)
    expect_equal(s$tafd, s$tafdc)
    expect_equal(s$dose, s$dosec)
  })

  test_that("tad(<state>)/tlast(<state>) survive an injected cmt() observable (nlmixr2est#685)", {
    # An algebraic observable in a residual-error tilde makes rxUi inject a
    # cmt() compartment for the observable *after* the linCmt() compartments.
    # That extra compartment must not shift the slot indices that
    # tad(<state>)/tlast(<state>) resolve to.  In a linCmt() model the
    # depot/central compartments are added last (calcLinCmt()), so an earlier
    # cmt() observable used to sort before them and shifted every per-state
    # lookup -- tad(depot) returned NA and tad(central) returned the current
    # time instead of NA.
    et <-
      et(amt = 100, time = c(0, 24, 48), cmt = "depot") |>
      et(time = c(72, 96, 120), cmt = "central")

    pars <- c(lka = log(1), lkel = log(0.1), lvc = log(10))

    # mirrors the structure of an rxUi simulation model: tad()/tlast() before
    # linCmt(), with cmt(Cc) injected for the algebraic observable
    withCmt <- rxode2({
      ka  <- exp(lka)
      kel <- exp(lkel)
      vc  <- exp(lvc)
      tdDepot   <- tad(depot)
      tdCentral <- tad(central)
      tlDepot   <- tlast(depot)
      tdAny     <- tad()
      Cc <- linCmt()
      cmt(Cc)
    })

    withoutCmt <- rxode2({
      ka  <- exp(lka)
      kel <- exp(lkel)
      vc  <- exp(lvc)
      tdDepot   <- tad(depot)
      tdCentral <- tad(central)
      tlDepot   <- tlast(depot)
      tdAny     <- tad()
      Cc <- linCmt()
    })

    # the injected cmt() observable does not change the ODE-state ordering
    expect_equal(rxState(withCmt), c("depot", "central"))
    expect_equal(rxState(withCmt), rxState(withoutCmt))
    expect_equal(withCmt$stateExtra, "Cc")

    s1 <- rxSolve(withCmt, pars, et)
    s2 <- rxSolve(withoutCmt, pars, et)

    # the cmt() observable must not perturb the per-state lookups
    expect_equal(s1$tdDepot, s2$tdDepot)
    expect_equal(s1$tdCentral, s2$tdCentral)
    expect_equal(s1$tlDepot, s2$tlDepot)

    # only depot was dosed (at 0, 24, 48); observations are at 72, 96, 120
    obs <- s1[s1$time %in% c(72, 96, 120), ]
    expect_equal(obs$tdDepot, c(24, 48, 72))     # time since last depot dose
    expect_equal(obs$tdAny, c(24, 48, 72))       # tad(depot) == tad() here
    expect_equal(obs$tlDepot, c(48, 48, 48))     # last depot dose time
    expect_true(all(is.na(obs$tdCentral)))       # central was never dosed
  })

  test_that("tad(<state>) survives a cmt() observable declared before d/dt (nlmixr2est#685)", {
    # The same slot-shift bug can be triggered in a pure-ODE model when the
    # algebraic-observable cmt() compartment is declared *before* the d/dt
    # block.  The d/dt states only get their compartment index when the d/dt
    # line is parsed, so an earlier cmt() observable used to sort before them.
    # (When the cmt() observable comes after the d/dt block, as the nlmixr2 ui
    # appends it, the ODE states are already numbered and there is no shift.)
    et <-
      et(amt = 100, time = c(0, 24, 48), cmt = "depot") |>
      et(time = c(72, 96, 120), cmt = "central")

    pars <- c(lka = log(1), lkel = log(0.1), lvc = log(10))

    cmtBefore <- rxode2({
      tdDepot   <- tad(depot)
      tdCentral <- tad(central)
      ka  <- exp(lka)
      kel <- exp(lkel)
      vc  <- exp(lvc)
      Cc <- central / vc
      cmt(Cc)                  # declared before the d/dt block
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - kel * central
    })

    cmtAfter <- rxode2({
      tdDepot   <- tad(depot)
      tdCentral <- tad(central)
      ka  <- exp(lka)
      kel <- exp(lkel)
      vc  <- exp(lvc)
      d/dt(depot)   <- -ka * depot
      d/dt(central) <-  ka * depot - kel * central
      Cc <- central / vc
      cmt(Cc)
    })

    # the observable does not change the ODE-state ordering, regardless of
    # where cmt() is declared relative to the d/dt block
    expect_equal(rxState(cmtBefore), c("depot", "central"))
    expect_equal(rxState(cmtBefore), rxState(cmtAfter))
    expect_equal(cmtBefore$stateExtra, "Cc")

    s1 <- rxSolve(cmtBefore, pars, et)
    s2 <- rxSolve(cmtAfter, pars, et)

    expect_equal(s1$tdDepot, s2$tdDepot)
    expect_equal(s1$tdCentral, s2$tdCentral)

    obs <- s1[s1$time %in% c(72, 96, 120), ]
    expect_equal(obs$tdDepot, c(24, 48, 72))   # depot dose history is tracked
    expect_true(all(is.na(obs$tdCentral)))     # central was never dosed
  })

  # context("tad family of functions with linCmt()/ode mix")

  test_that("ode mixed", {

    mod3 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin0 <- 1
      Kout <- 1
      EC50 <- 200
      ## The linCmt() picks up the variables from above
      C2 <- linCmt()
      Tz <- 8
      amp <- 0.1
      eff(0) <- 1 ## This specifies that the effect compartment starts at 1.
      ## Kin changes based on time of day (like cortosol)
      Kin <- Kin0 + amp * cos(2 * pi * (ctime - Tz) / 24)
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      tadd <- tad(depot)
      tad <- tad()
      tade <- tad(eff)
      dose <- dose()
      dosed <- dose(depot)
      dosee <- dose(eff)
    })

    ev <- eventTable(amount.units = "mg", time.units = "hours") |>
      add.dosing(dose = 10000, nbr.doses = 1, dosing.to = 1) |>
      add.sampling(seq(0, 48, length.out = 100))


    ## Create data frame of 8 am dosing for the first dose This is done
    ## with base R but it can be done with dplyr or data.table
    ev$ctime <- (ev$time + units::set_units(8, hr)) %% 24

    x <- rxSolve(mod3, ev)

    expect_equal(x$tad, x$tadd)
    expect_true(all(is.na(x$tade)))

    expect_equal(x$dose, x$dosed)
    expect_true(all(is.na(x$dosee)))

    ev <- eventTable(amount.units = "mg", time.units = "hours") |>
      add.dosing(dose = 10000, nbr.doses = 1, dosing.to = 1) |>
      add.dosing(dose = -1, start.time = 6, nbr.doses = 1, dosing.to = 3) |>
      add.sampling(seq(0, 48, length.out = 20))


    ## Create data frame of 8 am dosing for the first dose This is done
    ## with base R but it can be done with dplyr or data.table
    ev$ctime <- (ev$time + units::set_units(8, hr)) %% 24

    x <- rxSolve(mod3, ev, addDosing = TRUE)

    expect_false(isTRUE(all.equal(x$tad, x$tadd)))

    expect_false(isTRUE(all.equal(x$tad, x$tade)))

  })

  test_that("tad parsing", {

    mod2 <- function() {
      ini({
        ## Table 3 from Savic 2007
        cl  <- 17.2 # (L/hr)
        vc  <- 45.1 # L
        ka  <- 0.38 # 1/hr
        mtt <- 1.37 # hr
        f2 <-0.5    # Fraction of 1st Order portion
        n   <- 20.1
      })
      model({
        k           <- cl/vc
        bio <- 1-f2
        ktr = (n+1)/mtt
        ## note that lgammafn is the same as lgamma in R.
        d/dt(depot1) = exp(log(bio*podo(depot))+
                             log(ktr)+n*log(ktr*tad(depot))-
                             ktr*tad(depot)-lgammafn(n+1))-ka*depot1
        d/dt(depot2) <- -ka*depot2
        f(depot2) <-f2
        d/dt(cen)   <- ka*depot1 + ka*depot2-k*cen
      })
    }

    expect_error(mod2())

    mod2 <- function() {
      ini({
        ## Table 3 from Savic 2007
        cl  <- 17.2 # (L/hr)
        vc  <- 45.1 # L
        ka  <- 0.38 # 1/hr
        mtt <- 1.37 # hr
        f2 <-0.5    # Fraction of 1st Order portion
        n   <- 20.1
      })
      model({
        k           <- cl/vc
        bio <- 1-f2
        ktr = (n+1)/mtt
        ## note that lgammafn is the same as lgamma in R.
        d/dt(depot1) = exp(log(bio*podo(depot1))+
                             log(ktr)+n*log(ktr*tad(depot1))-
                             ktr*tad(depot1)-lgammafn(n+1))-ka*depot1
        d/dt(depot2) <- -ka*depot2
        f(depot2) <-f2
        d/dt(cen)   <- ka*depot1 + ka*depot2-k*cen
      })
    }

    mod2 <- mod2()

  })
})

rxTest({
  # -- Analytical background --------------------------------------------------
  # f(depot) <- f0 + state scales the bioavailable fraction of each dose.
  # For a bolus dose of amount `amt` at time t_dose into a depot that drains
  # with first-order rate constant ka:
  #
  #   depot(t) = amt * (f0 + S0) * exp(-ka * (t - t_dose))   for t >= t_dose
  #
  # Multiple doses contribute additively.  Constant state (d/dt(state)=0)
  # keeps S0 fixed, making the expected trajectory fully analytical.

  # -- Test 1 -----------------------------------------------------------------
  # state = 0 throughout  ->  f(depot) = f0 + 0 = f0 exactly.
  # Effective bioavailable amount = amt * f0; trajectory is mono-exponential.
  test_that("state-dep f with state=0 uses exactly f0 (no state shift)", {
    mod <- rxode2({
      # Bioavailability depends on a state variable.
      # With state held at 0, this reduces to the fixed-f0 case.
      f(depot) <- f0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- 0         # constant state; initial value is the only value
    })

    amt  <- 100
    f0   <- 0.8
    ka   <- 0.5
    times <- seq(0, 20, by = 0.5)

    et <- eventTable() |>
      add.dosing(dose = amt, nbr.doses = 1) |>
      add.sampling(times)

    s <- solve(mod, et,
               inits  = c(state = 0, depot = 0),
               params = c(f0 = f0, ka = ka))

    # Expected: depot(t) = amt * f0 * exp(-ka * t)
    expected <- amt * f0 * exp(-ka * times)

    # A mismatch would mean the bioavailable fraction was not f0.
    expect_equal(s$depot[s$time %in% times], expected,
                 tolerance = 1e-4,
                 label = "depot matches amt*f0*exp(-ka*t) when state=0")
  })

  # -- Test 2 -----------------------------------------------------------------
  # state = S0 (constant, various values)  ->  f(depot) = f0 + S0 exactly.
  # Effective bioavailable amount = amt * (f0 + S0); different for each S0.
  test_that("state-dep f shifts bioavailable amount by state value: f = f0 + state_at_dose", {
    mod <- rxode2({
      # f(depot) is evaluated at the moment the dose fires.
      # With constant state the evaluation time doesn't matter.
      f(depot) <- f0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- 0
    })

    amt  <- 100
    f0   <- 0.8
    ka   <- 0.5
    times <- seq(0, 20, by = 0.5)

    for (S0 in c(0.1, 0.5, -0.2)) {
      et <- eventTable() |>
        add.dosing(dose = amt, nbr.doses = 1) |>
        add.sampling(times)

      s <- solve(mod, et,
                 inits  = c(state = S0, depot = 0),
                 params = c(f0 = f0, ka = ka))

      F_eff    <- f0 + S0                          # effective bioavailability
      expected <- amt * F_eff * exp(-ka * times)   # analytical trajectory

      # Full trajectory must match the analytical formula for f = f0 + S0.
      expect_equal(s$depot[s$time %in% times], expected,
                   tolerance = 1e-4,
                   label = paste0("depot matches amt*(f0+S0)*exp(-ka*t) for S0=", S0))

      # Also confirm that this differs from the S0=0 baseline,
      # i.e., the state contribution is not silently ignored.
      expected_s0 <- amt * f0 * exp(-ka * times)
      expect_false(isTRUE(all.equal(expected, expected_s0)),
                   label = paste0("S0=", S0, " gives a different trajectory than state=0"))
    }
  })

  # -- Test 3 -----------------------------------------------------------------
  # Multiple doses: each dose's bioavailable amount = dose * (f0 + state_at_dose_time).
  # With a decaying state, consecutive doses get decreasing f values and
  # therefore decreasing effective amounts.
  # depot(t) = sum over fired doses of:  amt*(f0+state(t_d)) * exp(-ka*(t-t_d))
  test_that("state-dep f: each dose evaluated at its own firing-time state", {
    mod <- rxode2({
      f(depot)    <- f0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- -kstate * state   # state decays so each dose sees a different f
    })

    f0     <- 0.8
    ka     <- 0.5
    kstate <- 0.3
    amt    <- 100

    dose_times <- c(0, 5)
    times      <- seq(0, 15, by = 1)

    et <- et() |>
      et(amt = amt, time = dose_times[1], cmt = "depot") |>
      et(amt = amt, time = dose_times[2], cmt = "depot") |>
      et(times)

    for (S0 in c(0, 0.2)) {
      s <- rxSolve(mod, et,
                   params = c(f0 = f0, ka = ka, kstate = kstate),
                   inits  = c(depot = 0, state = S0))

      # Analytical: state(t) = S0 * exp(-kstate * t)
      # f at each dose time:  f_d = f0 + S0*exp(-kstate*t_d)
      # contribution of dose d at observation time t:
      #   amt * f_d * exp(-ka*(t - t_d))  [if t >= t_d]
      expected_depot <- vapply(times, function(t) {
        contrib <- 0
        for (td in dose_times) {
          if (t >= td) {
            f_at_td  <- f0 + S0 * exp(-kstate * td)
            contrib  <- contrib + amt * f_at_td * exp(-ka * (t - td))
          }
        }
        contrib
      }, numeric(1))

      obs <- s$depot[s$time %in% times]
      expect_equal(obs, expected_depot,
                   tolerance = 1e-3,
                   label = paste0("two-dose depot matches analytic f=f0+state(t_dose), S0=", S0))
    }
  })

  # -- Test 4 -----------------------------------------------------------------
  # Robustness: no NAs produced when state is non-zero and f varies between
  # subjects (two-row data frame with different initial state values).
  test_that("state-dep f: no NAs across subjects with different initial state", {
    mod <- rxode2({
      f(depot) <- f0 + state
      d/dt(depot) <- -ka * depot
      d/dt(state) <- 0
    })

    et <- eventTable() |>
      add.dosing(dose = 100, nbr.doses = 1) |>
      add.sampling(seq(0, 10, by = 1))

    s0 <- solve(mod, et, inits = c(state = 0,   depot = 0),
                params = c(f0 = 0.8, ka = 0.5))
    s1 <- solve(mod, et, inits = c(state = 0.3, depot = 0),
                params = c(f0 = 0.8, ka = 0.5))

    expect_false(any(is.na(s0$depot)), label = "no NAs when state=0")
    expect_false(any(is.na(s1$depot)), label = "no NAs when state=0.3")
    # Different initial f -> different trajectories
    expect_false(isTRUE(all.equal(s0$depot, s1$depot)),
                 label = "different state gives different f and depot")
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    # locking tests

    ode <- rxode2({
      a <- -8 / 3
      b <- -10
      c <- -1
      d / dt(X) <- a * X + Y * Z
      d / dt(Y) <- b * (Y - Z)
      d / dt(Z) <- -X * Y + c * Y - Z
    })

    test_that("Unlocked when simply loading", {
      expect_true(rxModels_()[[rxDll(ode)]] == 0L)
    })

    et <- eventTable(time.units = "hr") # default time units
    et$add.sampling(seq(from = 0, to = 100, by = 0.01))

    out <- rxSolve(ode,
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1),
                   .setupOnly = TRUE
                   )

    test_that("Locked after .setupOnly", {
      expect_true(rxModels_()[[rxDll(ode)]] == 1L)
    })

    out <- rxSolve(ode,
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1)
                   )

    test_that("Unlocked after other solve", {
      expect_true(rxModels_()[[rxDll(ode)]] == 0L)
    })
  }
})

rxTest({
  skip_if_not_installed("units")
  skip_if_not_installed("devtools")
  skip_on_cran()

  test_that("build package", {

    mod <- rxode2({
      a <- 6
      b <- 0.6
      d / dt(intestine) <- -a * intestine
      d / dt(blood) <- a * intestine - b * blood
    })

    obs <- units::set_units(seq(0, 10, by = 1 / 24), "days")

    et <- eventTable(time.units = "days")
    et$add.sampling(obs)
    et$add.dosing(
      dose = 2 / 24, start.time = 0,
      nbr.doses = 10, dosing.interval = 1
    )

    solve1 <- rxSolve(mod, et, returnType = "data.frame")

    mod2 <- mod
    dir <- tempdir()

    t <- try(rxPkg(mod, mod2, package = "rxm", wd = dir))
    if (inherits(t, "try-error")) {
      skip("not important")
    } else {
      expect_true(TRUE)
      try(remove.packages("rxm"))
    }
  ## unlink(dir, recursive=TRUE)
    # when load_all is used, you get
    ## Error: package 'rxode2' required by 'rxm' could not be found
    ## rm(list=c("mod", "mod2"))

    ## expect_error(library(rxm), NA)

    ## expect_error(mod, NA)
    ## expect_error(mod2, NA)

    ## expect_s3_class(mod, "rxode2")

    ## expect_s3_class(mod2, "rxode2")

    ## solve2 <- rxSolve(mod, et, returnType = "data.frame")

    ## expect_equal(solve1, solve2)


    ## detach("package:rxm", unload=TRUE)
    ## expect_error(mod)
    ## expect_error(mod2)
  })
})

rxTest({
  ms <- c(.methods0, "lsoda")

  for (m in ms) {
    # context(sprintf("Test turning compartment off (%s)", m))
    mod1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      C2 <- centr / V2
      C3 <- peri / V3
      d / dt(depot) <- -KA * depot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
    })

    et <- eventTable() |>
      add.dosing(dose = 10000, nbr.doses = 2, dosing.interval = 8) |>
      et(cmt = "-depot", time = 4) |>
      add.sampling(seq(0, 24, by = 0.5))

    x <- rxSolve(mod1, et, method = m)

    test_that("off works correctly", {
      expect_equal((x |> dplyr::filter(time == 4.5))$depot, 0)
      expect_false((x |> dplyr::filter(time == 4.5))$centr == 0)
    })
  }
})

rxTest({
  test_that("rbind", {

    mod2 <- function() {
      ini({
        KA <- 2.94E-01
        TCL <- 1.86E+01
        V2 <- 4.02E+01
        Q <- 1.05E+01
        V3 <- 2.97E+02
        Kin <- 1
        Kout <- 1
        EC50 <- 200
        eta.Cl ~ 0.2
        eff.sd <- sqrt(0.05)
        c2.sd <- sqrt(0.05)
      })
      model({
        CL <- TCL * exp(eta.Cl)
        d/dt(depot) <- -KA * depot
        d/dt(centr) <- KA * depot - CL * (centr / V2) - Q * (centr / V2) + Q * (peri / V3)
        d/dt(peri) <- Q * (centr / V2) - Q * (peri / V3)
        C2 <- centr/V2
        d/dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
        eff(0) <- 1000
        eff ~ add(eff.sd)
        C2 ~ prop(c2.sd)
      })
    }

    ev <- eventTable() |>
      add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12, dosing.to = 2) |>
      add.dosing(dose = 20000, nbr.doses = 5, start.time = 120, dosing.interval = 24, dosing.to = 2) |>
      add.sampling(0:240) |>
      et(id=1:2) |>
      as.data.frame()

    ev$dvid <- ifelse(ev$evid == 0, 1, NA_real_)

    ev2 <- ev[which(ev$dvid == 1), ]
    ev2$dvid <- 2

    ev <- rbind(ev, ev2) |>
      dplyr::arrange(id, time)

    rownames(ev) <- NULL

    pk1 <-
      suppressWarnings(rxSolve(mod2, nStud = 4, ev, cores = 2))

    pk2 <-
      suppressWarnings(rxSolve(mod2, nStud = 4, ev, cores = 2))

    b1 <- rbind.rxSolve(pk1, pk2)

    b2 <- rbind.rxSolve(pk1, pk2, pk1)

    expect_true(inherits(b1, "rxSolve"))
    expect_true(inherits(b2, "rxSolve"))
    expect_error(rbind.rxSolve())
  })
})

rxTest({
  test_that(".DollarNames", {

    mod2 <- rxode2({
      ## the order of variables do not matter, the type of compartmental
      ## model is determined by the parameters specified.
      CL ~ TCL * exp(eta.Cl)
      C2 ~ linCmt(KA, CL, V2, Q, V3)
      eff(0) <- 1 ## This specifies that the effect compartment starts at 1.
      d / dt(eff) ~ Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      ##
      resp <- eff + err1
      pk <- C2 * exp(err2)
    })

    ev <- eventTable(amount.units = "mg", time.units = "hours") |>
      add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12, dosing.to = 2) |>
      add.dosing(dose = 20000, nbr.doses = 5, start.time = 120, dosing.interval = 24, dosing.to = 2) |>
      add.sampling(0:240)

    expect_equal(
      .DollarNames(ev, ""),
      c(
        "randomType", "canResize", "ids", "show", "ndose", "nobs", "expand", "getSampling",
        "get.sampling", "getDosing", "get.dosing", "get.nobs", "get.obs.rec",
        "getEventTable", "get.EventTable", "copy", "importEventTable",
        "import.EventTable", "simulate", "clearDosing", "clear_dosing",
        "clear.dosing", "clearSampling", "clear_sampling", "clear.sampling",
        "addSampling", "add_sampling", "add.sampling", "addDosing", "add_dosing",
        "add.dosing", "get_units", "getUnits", "get.units", "units",
        "dur", "ss", "evid", "addl", "ii", "rate", "amt", "cmt", "high",
        "time", "low", "id", "env"
      )
    )

    p <- data.frame(a = 6, b = seq(0.4, 0.9, length.out = 4))

    ## "Study" Differences
    thetaMat <- diag(3) * 0.01
    dimnames(thetaMat) <- list(NULL, c("KA", "TCL", "V2"))

    sigma <- diag(2) * 0.05
    dimnames(sigma) <- list(c("err1", "err2"), c("err1", "err2"))


    pk4 <-
      suppressWarnings(rxSolve(
        mod2,
        c(
          KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
          Kin = 1, Kout = 1, EC50 = 200
        ),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")), dfSub = 100, dfObs = 100,
        nSub = 4, nStud = 8, thetaMat = thetaMat, sigma = sigma, ev, cores = 1
      ))

    expect_equal(sort(.DollarNames(pk4, "")),
                 sort(c("add.dosing", "add.sampling", "central", "central0", "clear.dosing",
                        "clear.sampling", "counts", "depot", "depot0", "dll", "EC50",
                        "eff0", "env", "eta.Cl", "get.dosing", "get.EventTable", "get.nobs",
                        "get.obs.rec", "get.sampling", "get.units", "import.EventTable",
                        "inits", "KA", "Kin", "Kout", "model", "nobs", "omegaList", "params",
                        "peripheral1", "peripheral10", "pk", "Q", "resp", "rxode2", "sigmaList",
                        "sim.id", "sim.id", "t", "TCL", "thetaMat", "time", "tolFactor", "units",
                        "V2", "V3")))


    pk4 <-
      suppressWarnings(rxSolve(
        mod2,
        c(
          KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
          Kin = 1, Kout = 1, EC50 = 200
        ),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")), dfSub = 100,
        nSub = 4, nStud = 8, thetaMat = thetaMat, sigma = sigma, ev, cores = 1
      ))

    expect_equal(sort(.DollarNames(pk4, "")),
                 sort(c("add.dosing", "add.sampling", "central", "central0", "clear.dosing",
                        "clear.sampling", "counts", "depot", "depot0", "dll", "EC50",
                        "eff0", "env", "eta.Cl", "get.dosing", "get.EventTable", "get.nobs",
                        "get.obs.rec", "get.sampling", "get.units", "import.EventTable",
                        "inits", "KA", "Kin", "Kout", "model", "nobs", "omegaList", "params",
                        "peripheral1", "peripheral10", "pk", "Q", "resp", "rxode2", "sim.id",
                        "sim.id", "t", "TCL", "thetaMat", "time", "tolFactor", "units", "V2", "V3")))

    pk4 <-
      suppressWarnings(rxSolve(
        mod2,
        c(
          KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
          Kin = 1, Kout = 1, EC50 = 200
        ),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, nStud = 8, thetaMat = thetaMat, sigma = sigma, ev, cores = 1
      ))

    expect_equal(sort(.DollarNames(pk4, "")),
                 sort(c("add.dosing", "add.sampling", "central", "central0", "clear.dosing",
                        "clear.sampling", "counts", "depot", "depot0", "dll", "EC50",
                        "eff0", "env", "eta.Cl", "get.dosing", "get.EventTable", "get.nobs",
                        "get.obs.rec", "get.sampling", "get.units", "import.EventTable",
                        "inits", "KA", "Kin", "Kout", "model", "nobs", "params", "peripheral1",
                        "peripheral10", "pk", "Q", "resp", "rxode2", "sim.id", "sim.id",
                        "t", "TCL", "thetaMat", "time", "tolFactor", "units", "V2", "V3")))

    pk4 <-
      suppressWarnings(rxSolve(
        mod2,
        c(
          KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
          Kin = 1, Kout = 1, EC50 = 200
        ),
        omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
        nSub = 4, nStud = 8, sigma = sigma, ev, cores = 1))

    expect_equal(sort(.DollarNames(pk4, "")),
                 sort(c("add.dosing", "add.sampling", "central", "central0", "clear.dosing",
                   "clear.sampling", "counts", "depot", "depot0", "dll", "EC50",
                   "eff0", "env", "eta.Cl", "get.dosing", "get.EventTable", "get.nobs",
                   "get.obs.rec", "get.sampling", "get.units", "import.EventTable",
                   "inits", "KA", "Kin", "Kout", "model", "nobs", "params", "peripheral1",
                   "peripheral10", "pk", "Q", "resp", "rxode2", "sim.id", "sim.id",
                   "t", "TCL", "time", "tolFactor", "units", "V2", "V3")))
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    # test newind flag
    ode.1c <- rxode2({
      V <- 20
      Cl <- 1
      fc <- 1
      C2 <- center / V
      ni <- newind
      ni2 <- NEWIND
      d / dt(center) ~ -Cl * C2
      f(center) <- fc
    })

    et <- eventTable() |>
      add.dosing(dose = 3, nbr.doses = 6, dosing.interval = 8) |>
      add.sampling(0:4)

    et2 <- eventTable() |>
      add.dosing(start.time = 0.5, dose = 3, nbr.doses = 6, dosing.interval = 8) |>
      add.sampling(0:4)

    ms <- .methods0

    for (m in ms) {
      s <- solve(ode.1c, et, addDosing = TRUE, method = m)
      test_that("ode newind works with dose first", {
        expect_equal(s$ni[1], 1L)
        expect_equal(s$ni2[1], 1L)
        expect_true(all(s$ni[-1] == 2))
        expect_true(all(s$ni2[-1] == 2))
      })

      s2 <- solve(ode.1c, et2, addDosing = TRUE, method = m)

      test_that("ode newind works with dose first", {
        expect_equal(s2$ni[1], 1L)
        expect_equal(s2$ni2[1], 1L)
        expect_true(all(s2$ni[-1] == 2))
        expect_true(all(s2$ni2[-1] == 2))
      })
    }
  }
})

rxTest({
  model <- rxode2({
    C2 <- centr / V2
    C3 <- peri / V3
    #
    d / dt(depot) <- -KA * depot
    d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
    d / dt(peri) <- Q * C2 - Q * C3
    d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
    ##
    eff(0) <- 1
    ##
    ## central
    KA <- 2.94E-01
    CL <- 1.86E+01
    V2 <- 4.02E+01
    ## peripheral
    Q <- 1.05E+01
    V3 <- 2.97E+02
    ##
    ## effects
    Kin <- 1
    Kout <- 1
    EC50 <- 200
  })

  event_table <- rxode2::eventTable() |>
    add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12) |>
    add.sampling(0:240)

  test_that("Bad solve raises error, RxODE#5", {
    tf <- tempfile()
    sink(tf)
    expect_error(
      model |> solve(c(KA = -10), event_table, maxsteps = 5000),
      rex::rex("could not solve the system")
    )
    sink()
    unlink(tf)
  })
})

rxTest({
  .mod_osc <- rxode2::rxode2({
    d/dt(q) <- p
    d/dt(p) <- -q
  })
  .mod_odd <- rxode2::rxode2({
    d/dt(X) <- Y * Z
    d/dt(Y) <- -X * Z
    d/dt(Z) <- -X * Y
  })
  .et_osc <- rxode2::eventTable()
  .et_osc$add.sampling(seq(0, 2 * pi, length.out = 10))
  .et_odd <- rxode2::eventTable()
  .et_odd$add.sampling(0:10)

  # method, hmin (NULL = omit)
  # em uses hmin=0.001 for both tests (first-order; needs small step for accuracy)
  .symplectic <- list(
    list(method = "mm",     hmin = NULL),
    list(method = "em",     hmin = 0.001),
    list(method = "vv",     hmin = NULL),
    list(method = "sem",    hmin = NULL),
    list(method = "sb3a",   hmin = NULL),
    list(method = "sb3am4", hmin = NULL)
  )

  for (.cfg in .symplectic) {
    local({
      .method <- .cfg$method
      .hmin   <- .cfg$hmin
      test_that(paste(.method, "integrates harmonic oscillator correctly"), {
        .args <- list(.mod_osc, params = c(), events = .et_osc,
                      inits = c(q = 1, p = 0), method = .method)
        if (!is.null(.hmin)) .args$hmin <- .hmin
        out   <- do.call(rxode2::rxSolve, .args)
        times <- out$time
        expect_equal(out$q, cos(times),  tolerance = 0.05)
        expect_equal(out$p, -sin(times), tolerance = 0.05)
      })
      test_that(paste(.method, "solves odd number of states successfully"), {
        .args <- list(.mod_odd, params = c(), events = .et_odd,
                      inits = c(X = 1, Y = 1, Z = 1), method = .method)
        if (!is.null(.hmin)) .args$hmin <- .hmin
        out <- do.call(rxode2::rxSolve, .args)
        expect_s3_class(out, "rxSolve")
      })
    })
  }

  # mm-specific: verify that the order= parameter is respected
  test_that("mm uses the order option from rxSolve", {
    .et <- rxode2::eventTable()
    .et$add.sampling(seq(0, 2 * pi, length.out = 10))
    out2  <- rxode2::rxSolve(.mod_osc, params = c(), events = .et,
                              inits = c(q = 1, p = 0), method = "mm", order = 2L)
    out10 <- rxode2::rxSolve(.mod_osc, params = c(), events = .et,
                              inits = c(q = 1, p = 0), method = "mm", order = 10L)
    expect_s3_class(out2,  "rxSolve")
    expect_s3_class(out10, "rxSolve")
    expect_equal(out2$q,  cos(out2$time),  tolerance = 0.05)
    expect_equal(out10$q, cos(out10$time), tolerance = 0.05)
    expect_true(any(out2$q != out10$q))
  })
})

rxTest({
  test_that("Matrix exponential alone works", {
    # Test inductive linearization

    ## Case 1 ME alone from wikipedia
    mod <- suppressMessages(rxode2(
    {
      d / dt(x) <- 2 * x - y + z
      d / dt(y) <- 3 * y - 1 * z
      d / dt(z) <- 2 * x + y + 3 * z
      x(0) <- 0.1
      y(0) <- 0.1
      z(0) <- 0.1
    },
    indLin = TRUE
    ))

    m <- rxSolve(mod, et(seq(0, 24, length.out = 50)), method = "indLin")
    m2 <- rxSolve(mod, et(seq(0, 24, length.out = 50)), method = "lsoda")

    expect_equal(as.data.frame(m), as.data.frame(m2), tolerance = 1e-5)

    ## Now do without indLin in the rxode2

    mod <- rxode2({
      d / dt(x) <- 2 * x - y + z
      d / dt(y) <- 3 * y - 1 * z
      d / dt(z) <- 2 * x + y + 3 * z
      x(0) <- 0.1
      y(0) <- 0.1
      z(0) <- 0.1
    })

    m <- suppressMessages(rxSolve(mod, et(seq(0, 24, length.out = 50)), method = "indLin"))
    m2 <- rxSolve(mod, et(seq(0, 24, length.out = 50)), method = "lsoda")

    ## FIXME
    ## expect_equal(as.data.frame(m), as.data.frame(m2), tolerance = 1e-5)

    ## Case 2 ME alone with inhomogenous systems

    mod <- suppressMessages(rxode2(
    {
      d / dt(x) <- 2 * x - y + z + exp(-2 * t)
      d / dt(y) <- 3 * y - 1 * z
      d / dt(z) <- 2 * x + y + 3 * z + exp(-2 * t)
      x(0) <- 0.1
      y(0) <- 0.1
      z(0) <- 0.1
    },
    indLin = TRUE
    ))

    m <- rxSolve(mod, et(seq(0, 24, length.out = 50)), method = "indLin")
    m2 <- rxSolve(mod, et(seq(0, 24, length.out = 50)), method = "lsoda")

    ## gridExtra::grid.arrange(plot(m), plot(m2))

    ## FIXME?
    ## expect_equal(as.data.frame(m), as.data.frame(m2), tolerance =1e-5)

    mod <- suppressMessages(rxode2("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
d/dt(blood)     = a*intestine - b*blood
", indLin = TRUE))


    et <- eventTable(time.units = "days")
    et$add.sampling(seq(0, 10, by = 1 / 24))
    et$add.dosing(
      dose = 2 / 24, rate = 2, start.time = 0,
      nbr.doses = 10, dosing.interval = 1
    )

    pk <- rxSolve(mod, et, method = "indLin")
    pk2 <- rxSolve(mod, et, method = "liblsoda")
    expect_equal(as.data.frame(pk), as.data.frame(pk2), tolerance = 1e-5)

    ## plot(microbenchmark::microbenchmark(rxSolve(mod,et, method="indLin",indLinMatExpType=1L),rxSolve(mod,et, method="indLin",indLinMatExpType=2L), rxSolve(mod,et, method="indLin",indLinMatExpType=3L), rxSolve(mod,et, method="lsoda")), log="y")

    et2 <- eventTable(time.units = "days")
    et2$add.sampling(seq(0, 10, by = 1 / 24))
    et2$add.dosing(
      dose = 2, start.time = 0,
      nbr.doses = 10, dosing.interval = 1
    )

    pk <- rxSolve(mod, et2, method = "indLin")

    pk2 <- rxSolve(mod, et2, method = "liblsoda")

    expect_equal(as.data.frame(pk), as.data.frame(pk2), tolerance = 1e-5)

    ## Inductive linearization
    mmModel <- suppressMessages(rxode2(
    {
      ka <- 1
      Vc <- 1
      Vmax <- 0.00734
      Km <- 0.3672
      Cp <- center / Vc
      d / dt(center) <- -Vmax / (Km + Cp) * Cp
    },
    indLin = TRUE
    ))

    mmModel <- suppressMessages(rxode2(
    {
      ka <- 1
      Vc <- 1
      Vmax <- 0.00734
      Km <- 0.3672
      Cp <- center / Vc
      d / dt(center) <- -Vmax / (Km + Cp) * Cp + exp(-10 * t)
    },
    indLin = TRUE
    ))

    ## Inductive + 1x1 matrix
    ## FIXME this should be inductive too...
    mmModel <- suppressMessages(rxode2(
    {
      ka <- 1
      Vc <- 1
      Vmax <- 0.00734
      Km <- 0.3672
      d / dt(depot) <- -ka * depot
      d / dt(center) <- ka * depot - Vmax / (Km + Cp) * Cp
      Cp <- center / Vc
    },
    indLin = TRUE
    ))

    ## This is inductive
    mmModel <- suppressMessages(rxode2(
    {
      ka <- 1
      Vc <- 1
      Vmax <- 0.00734
      Km <- 0.3672
      d / dt(depot) <- -ka * depot
      Cp <- center / Vc
      d / dt(center) <- ka * depot - Vmax / (Km + Cp) * Cp
    },
    indLin = TRUE
    ))

    mmModel <- suppressMessages(rxode2(
    {
      ka <- 1
      Vc <- 1
      Vmax <- 0.00734
      Km <- 0.3672
      V4 <- 4.3
      Q <- 1.5
      K12 <- Q / Vc
      K21 <- Q / Vp
      Cp <- center / Vc
      d / dt(depot) <- -ka * depot
      d / dt(center) <- ka * depot - Vmax / (Km + Cp) * Cp + K21 * periph - K12 * center
      d / dt(periph) <- -K21 * periph + K12 * center
    },
    indLin = TRUE
    ))

    ## Inductive linearization
    mmModel <- suppressMessages(rxode2(
    {
      ka <- 1
      Vc <- 1
      Vmax <- 0.00734
      Km <- 0.3672
      d / dt(depot) <- -ka * depot
      Cp <- center / Vc
      d / dt(center) <- ka * depot - Vmax / (Km + Cp) * Cp
    },
    indLin = TRUE
    ))

    et <- eventTable(time.units = "days")
    et$add.sampling(seq(0, 10, by = 1 / 24))
    et$add.dosing(
      dose = 2, start.time = 0,
      nbr.doses = 10, dosing.interval = 6
    )

    pk <- rxSolve(mmModel, et, method = "indLin")
    pk2 <- rxSolve(mmModel, et, method = "liblsoda")

    ## gridExtra::grid.arrange(plot(pk), plot(pk2))

    expect_equal(as.data.frame(pk), as.data.frame(pk2), tolerance = 7e-5)

    mmModel <- suppressMessages(rxode2(
    {
      ka <- 1
      Vc <- 1
      Vmax <- 0.00734
      Km <- 0.3672
      d / dt(depot) <- -ka * depot
      Cp <- center / Vc
      d / dt(center) <- ka * depot - Vmax / (Km + Cp) * Cp + 5 * exp(-0.5 * t)
    },
    indLin = TRUE
    ))

    pk <- rxSolve(mmModel, et, method = "indLin")
    pk2 <- rxSolve(mmModel, et, method = "lsoda")

    ## gridExtra::grid.arrange(plot(pk), plot(pk2))
    ## These are not equal...
    ## expect_equal(as.data.frame(pk), as.data.frame(pk2), tolerance =7e-5)

    ## plot(microbenchmark::microbenchmark(rxSolve(mmModel,et, method="indLin",indLinMatExpType=1L),rxSolve(mmModel,et, method="indLin",indLinMatExpType=2L), rxSolve(mmModel,et, method="indLin",indLinMatExpType=3L), rxSolve(mmModel,et, method="lsoda")), log="y")

    ## Van der Pol Equation
    ## mu = 1000 stiff
    ## me = 1 non-stiff
    ## rxIndLinState(list(y="dy", dy="y"))
    rxIndLinState(NULL)
    rxIndLinStrategy()
    van1 <- suppressMessages(rxode2(
    {
      y(0) <- 2
      d / dt(y) <- dy
      d / dt(dy) <- mu * (1 - y^2) * dy - y
    },
    indLin = TRUE
    ))

    van <- van1

    rxIndLinState(list(y = "dy", dy = "y"))
    ## rxIndLinState(NULL)
    rxIndLinStrategy()
    van2 <- suppressMessages(rxode2(
    {
      y(0) <- 2
      d / dt(y) <- dy
      d / dt(dy) <- mu * (1 - y^2) * dy - y
    },
    indLin = TRUE
    ))

    ## rxIndLinState(list(y="dy", dy="y"))
    rxIndLinState(NULL)
    rxIndLinStrategy("split")
    van3 <- suppressMessages(rxode2(
    {
      y(0) <- 2
      d / dt(y) <- dy
      d / dt(dy) <- mu * (1 - y^2) * dy - y
    },
    indLin = TRUE
    ))

    et <- eventTable()
    ## 3000 causes weird behavior of indLin / lsoda
    et$add.sampling(seq(0, 20, length.out = 200))

    s1 <- rxSolve(van1, et, c(mu = 1000), method = "lsoda")
    s2 <- rxSolve(van1, et, c(mu = 1000), method = "indLin")
    ## s3 <- rxSolve(van, et, c(mu=1000), method="dop853")

    ## f <- function(mu = 1, ...) {
    ##   s1 <- rxSolve(van1, et, c(mu = mu), method = "lsoda") |> plot() +
    ##     ggtitle(sprintf("Lsoda mu=%s", mu))
    ##   s2 <- rxSolve(van1, et, c(mu = mu), method = "indLin", ...) |> plot() +
    ##     ggtitle(sprintf("indLin1 mu=%s", mu))
    ##   s3 <- rxSolve(van3, et, c(mu = mu), method = "indLin", ...) |> plot() +
    ##     ggtitle(sprintf("indLin3 mu=%s", mu))
    ##   ## s4 <- rxSolve(van3, et, c(mu=mu), method="indLin", ...) |> plot() +
    ##   ##     ggtitle(sprintf("indLin3 mu=%s", mu))
    ##   s4 <- rxSolve(van3, et, c(mu = mu), method = "dop853", ...) |> plot() +
    ##     ggtitle(sprintf("dop853 mu=%s", mu))
    ##   gridExtra::grid.arrange(s1, s2, s3, s4)
    ## }

    ## uses library animation
    ## saveGIF({
    ##     for (i in seq(0.1, 15, by=0.1)){
    ##         print(f(mu=i))
    ##     }
    ## }, movie.name="indLin-dop.gif", interval=0.1, nmax=30, ani.width=600, ani.hegith=300)

    expect_equal(as.data.frame(s1), as.data.frame(s2), tolerance = 1e-5)

    s1 <- rxSolve(van, et, c(mu = 1), method = "lsoda")
    s2 <- rxSolve(van, et, c(mu = 1), method = "indLin")
    ## expect_equal(as.data.frame(s1), as.data.frame(s2), tolerance =1e-4)
    ## s3 <- rxSolve(van, et, c(mu=1), method="dop853")

    ## s1 |> rename(y.lsoda=y, dy.lsoda=dy) |>
    ##     merge(s2) |> mutate(y.diff=y.lsoda - y) |>
    ##     ggplot(aes(time, y.diff)) + geom_line()

    ## gridExtra::grid.arrange(plot(s1), plot(s2))


    ## f <- function(mu=5){
    ##     s1 <- rxSolve(van, et, c(mu=mu), method="lsoda")
    ##     s2 <- rxSolve(van, et, c(mu=mu), method="indLin")
    ##     s1 |> rename(y.lsoda=y, dy.lsoda=dy) |>
    ##         merge(s2) |> mutate(y.diff=y.lsoda - y) |>
    ##         ggplot(aes(time, y.diff)) + geom_line() + ylim(-5, 5) +
    ##         ggtitle(paste0("mu=", mu)) ->
    ##         ret
    ##     return(ret)
    ## }

    ## expect_equal(as.data.frame(s1), as.data.frame(s2), tolerance =1e-4)

    ## gridExtra::grid.arrange(plot(s1), plot(s2), plot(s3))

    ## expect_equal(as.data.frame(s1), as.data.frame(s2))

    ## microbenchmark::microbenchmark(rxSolve(mmModel,et, method="indLin"),
    ##                                rxSolve(mmModel,et, method="liblsoda"))

    iSec <- suppressMessages(rxode2(
    {
      d / dt(Ga) <- -ka * Ga
      d / dt(Gt) <- ka * Ga - ka * Gt
      Gprod <- Gss * (Clg + Clgi * Iss)
      d / dt(Gc) <- ka * Gt - Gprod + Q / Vp * Gp - (Clg + Clgi * Ie + Q) / Vg * Gc
      Gc(0) <- Gss * Vg
      d / dt(Gp) <- -Q / Vp * Gp + Q / Vg * Gc
      d / dt(Ge) <- Gc * Kge - Ge * Kge
      d / dt(I) <- (Iss * Cli) * (1 + Sincr * Gt) * (Ge / Gss)^IPRG - Cli / Vi * I
      I(0) <- Iss * Vi
      d / dt(Ie) <- kie * I - kie * Ie
    },
    indLin = TRUE
    ))
  })

  test_that("hmax caps method='indLin' relinearization interval (task #8)", {
    # `indLin()`/`meOnly()` (src/expm.cpp) evaluate the ME/Jacobian ONCE per
    # call and treat it as constant over the WHOLE requested interval --
    # exact for a true (state-independent) matExp() model, but only a
    # first-order approximation for a state-dependent (indLin-forcing, e.g.
    # Michaelis-Menten) one. `hmax` previously had NO effect on this at
    # all: the relinearization interval was always exactly the gap between
    # requested output times, so a coarse sampling grid silently gave a
    # coarse (and potentially very wrong) answer for nonlinear models, with
    # no way for a user to ask for more accuracy. This checks (1) `hmax`
    # now measurably improves accuracy for a nonlinear (Michaelis-Menten
    # elimination) model, and (2) it makes no numerical difference for a
    # genuinely linear matExp model (same Jacobian regardless of how finely
    # the interval is subdivided).
    ode_code <- "
      vmax <- 10; km <- 5; v <- 20
      d/dt(central) = -vmax*central/(km+central)
    "
    pars <- c(vmax = 10, km = 5, v = 20)
    et_f <- et(amt = 100, cmt = "central") |> et(seq(0, 20, by = 0.5))
    mod_ode <- rxode2(ode_code)
    res_ode <- rxSolve(mod_ode, et_f, pars, atol = 1e-10, rtol = 1e-10)

    mod_mexp <- suppressMessages(rxode2(rxToIndLin(ode_code)))
    diff_coarse <- max(abs(
      rxSolve(mod_mexp, et_f, pars, method = "indLin", hmax = 0.5)$central - res_ode$central
    ))
    diff_fine <- max(abs(
      rxSolve(mod_mexp, et_f, pars, method = "indLin", hmax = 0.01)$central - res_ode$central
    ))
    expect_true(diff_fine < diff_coarse / 10)

    ode_code_lin <- "
      ka <- 0.5; cl <- 0.2; v <- 10
      d/dt(depot) = -ka*depot
      d/dt(central) = ka*depot - cl/v*central
    "
    pars_lin <- c(ka = 0.5, cl = 0.2, v = 10)
    et_lin <- et(amt = 100, cmt = "depot") |> et(seq(0, 20, by = 0.5))
    mod_mexp_lin <- suppressMessages(rxode2(rxToIndLin(ode_code_lin)))
    r_coarse <- rxSolve(mod_mexp_lin, et_lin, pars_lin, method = "indLin", hmax = 0.5)$central
    r_fine <- rxSolve(mod_mexp_lin, et_lin, pars_lin, method = "indLin", hmax = 0.01)$central
    expect_equal(r_coarse, r_fine, tolerance = 1e-6)
  })
})

if (!.Call(`_rxode2_isIntel`)) {
  modelCode <- "
TVBASE=14.3;
BASE=TVBASE*exp(EBASE);
SLD(0)=BASE;
d/dt(SLD)=-0.5*SLD;
"
  m1 <- rxode2(modelCode)
  ev <- eventTable() |>
    add.sampling(c(0, 1))
  e <- rxSolve(m1, ev, params = c(EBASE = 0.3))
  test_that("Make sure initial conditions work, Issue RxODE#67: baseline works", {
    expect_equal(c(19.3029809483368, 19.3029809483368), e$BASE, tolerance = 1e-6)
  })
}

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    test_that("test meanProb()", {

      x <- rnorm(10)
      m <- mean(x)
      s <- sd(x)
      v <- var(x)
      mn <- min(x)
      mx <- max(x)
      t1 <- c("0%"=mn, "25%"=m+(s/sqrt(10))*qt(0.25, 9), "50%"=m, "75%"=m+(s/sqrt(10))*qt(0.75, 9), "100%"=mx)
      expect_equal(meanProbs(x), t1)

      t1.100 <- c("0%"=mn, "25%"=m+(s/sqrt(100))*qt(0.25, 99), "50%"=m, "75%"=m+(s/sqrt(100))*qt(0.75, 99), "100%"=mx)
      expect_equal(meanProbs(x, n=100), t1.100)

      t1.p100 <- c("0%"=mn, "25%"=m+s*sqrt(1.0+(1.0/100))*qt(0.25, 99), "50%"=m, "75%"=m+s*sqrt(1.0+(1.0/100))*qt(0.75, 99), "100%"=mx)
      expect_equal(meanProbs(x, n=100, pred=TRUE), t1.p100)

      t2 <- c(c("mean"=m, "var"=v, "sd"=s, "min"=mn, "max"=mx, "n"=10), t1)
      expect_equal(meanProbs(x, onlyProbs=FALSE), t2)
      x2 <- c(x, NA_real_)
      setNames(rep(NA_real_, length(t1)),names(t1))

      expect_equal(meanProbs(x2), setNames(rep(NA_real_, length(t1)),names(t1)))

      expect_equal(meanProbs(x2, onlyProbs=FALSE),
                   setNames(rep(NA_real_, length(t2)),names(t2)))

      expect_equal(meanProbs(x2, na.rm=TRUE), t1)

      expect_equal(meanProbs(x2, onlyProbs=FALSE, na.rm=TRUE), t2)

      mod <- rxode2({
        d/dt(intestine) <- -a * intestine
        d/dt(blood) <- a * intestine - b * blood
      })

      et <- eventTable()
      et$add.sampling(seq(0, 10, length.out = 50))
      et$add.dosing(
        dose = 2/24, rate = 2, strt.time = 0,
        nbr.doses = 10, dosing.interval = 1
      )

      p <- data.frame(a = 6, b = seq(0.4, 0.9, length.out = 4))

      pk1 <- suppressWarnings(rxSolve(mod, p, et, cores = 1))

      ci1 <- confint(pk1, "blood", mean=TRUE)

      # use dplyr
      ci2 <- pk1 |>
        dplyr::group_by(time) |>
        dplyr::reframe(eff=meanProbs(blood, c(0.025, 0.5, 0.975), na.rm=TRUE, names=FALSE),
                       p1=c(0.025, 0.5, 0.975), Percentile=c("2.5%", "50%", "97.5%")) |>
        dplyr::arrange(time, p1)

      expect_equal(ci1$eff, ci2$eff)

      ci1 <- confint(pk1, "blood", mean=TRUE, pred=TRUE)

      ci2 <- pk1 |>
        dplyr::group_by(time) |>
        dplyr::reframe(eff=meanProbs(blood, c(0.025, 0.5, 0.975), na.rm=TRUE, names=FALSE, pred=TRUE),
                       p1=c(0.025, 0.5, 0.975), Percentile=c("2.5%", "50%", "97.5%")) |>
        dplyr::arrange(time, p1)

      expect_equal(ci1$eff, ci2$eff)

      ci1 <- confint(pk1, "blood", mean=TRUE, n=100)

      ci2 <- pk1 |>
        dplyr::group_by(time) |>
        dplyr::reframe(eff=meanProbs(blood, c(0.025, 0.5, 0.975), na.rm=TRUE, names=FALSE,
                                     n=100),
                       p1=c(0.025, 0.5, 0.975), Percentile=c("2.5%", "50%", "97.5%")) |>
        dplyr::arrange(time, p1)

      expect_equal(ci1$eff, ci2$eff)


      ci1 <- confint(pk1, "blood", mean=TRUE, pred=TRUE, n=100)

      ci2 <- pk1 |>
        dplyr::group_by(time) |>
        dplyr::reframe(eff=meanProbs(blood, c(0.025, 0.5, 0.975), na.rm=TRUE, names=FALSE,
                                     pred=TRUE, n=100),
                       p1=c(0.025, 0.5, 0.975), Percentile=c("2.5%", "50%", "97.5%")) |>
        dplyr::arrange(time, p1)

      expect_equal(ci1$eff, ci2$eff)


      ci1 <- confint(pk1, "blood", mean=TRUE, useT=FALSE)

      # use dplyr
      ci2 <- pk1 |>
        dplyr::group_by(time) |>
        dplyr::reframe(eff=meanProbs(blood, c(0.025, 0.5, 0.975), na.rm=TRUE, names=FALSE, useT=FALSE),
                       p1=c(0.025, 0.5, 0.975), Percentile=c("2.5%", "50%", "97.5%")) |>
        dplyr::arrange(time, p1)

      expect_equal(ci1$eff, ci2$eff)

      ci1 <- confint(pk1, "blood", mean=TRUE, useT=FALSE, pred=TRUE)

      # use dplyr
      ci2 <- pk1 |>
        dplyr::group_by(time) |>
        dplyr::reframe(eff=meanProbs(blood, c(0.025, 0.5, 0.975), na.rm=TRUE, names=FALSE, useT=FALSE,
                                     pred=TRUE),
                       p1=c(0.025, 0.5, 0.975), Percentile=c("2.5%", "50%", "97.5%")) |>
        dplyr::arrange(time, p1)

      expect_equal(ci1$eff, ci2$eff)

    })
  }
})

rxTest({
  mod <- rxode2(paste(sprintf("d/dt(amt%s) = -k1*amt%s", 1:105, 1:105), collapse = ";"))

  et <- eventTable() |> add.sampling(0:180)

  test_that("initial conditions work", {
    inits <- structure(rep(10, 105), names = sprintf("amt%s", 1:105))

    dat <- mod |> rxSolve(et, c(k1 = 0.1), inits = inits)

    for (j in seq(3, 106)) {
      expect_equal(dat[, 2], dat[, j])
    }
  })

  test_that("Events work", {
    for (i in 1:105) {
      et$add.dosing(10, dosing.to = i, start.time = 0)
    }

    dat <- mod |> rxSolve(et, c(k1 = 0.1))

    for (j in seq(3, 106)) {
      expect_equal(dat[, 2], dat[, j])
    }
  })
})

rxTest({
  test_that("Test ui solve", {

    # Step 1 - Create a model specification
    f <- function() {
      ini({
        KA <- .291
        CL <- 18.6
        V2 <- 40.2
        Q <- 10.5
        V3 <- 297.0
        Kin <- 1.0
        Kout <- 1.0
        EC50 <- 200.0
      })
      model({
        # A 4-compartment model, 3 PK and a PD (effect) compartment
        # (notice state variable names 'depot', 'centr', 'peri', 'eff')
        C2 <- centr/V2
        C3 <- peri/V3
        d/dt(depot) <- -KA*depot
        d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
        d/dt(peri)  <-                    Q*C2 - Q*C3
        d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
        eff(0) <- 1
      })
    }

    u <- f()

    # this pre-compiles and displays the simulation model
    u$simulationModel

    # Step 2 - Create the model input as an EventTable,
    # including dosing and observation (sampling) events

    # QD (once daily) dosing for 5 days.

    qd <- eventTable(amount.units = "ug", time.units = "hours")
    qd$add.dosing(dose = 10000, nbr.doses = 5, dosing.interval = 24)

    # Sample the system hourly during the first day, every 8 hours
    # then after

    qd$add.sampling(0:24)
    qd$add.sampling(seq(from = 24 + 8, to = 5 * 24, by = 8))

    # Step 3 - set starting parameter estimates and initial
    # values of the state


    # Step 4 - Fit the model to the data
    expect_error(solve(u, qd), NA)

    u1 <- u$simulationModel

    expect_error(solve(u1, qd), NA)

    u2 <- u$simulationIniModel

    expect_error(solve(u2, qd), NA)

  })

  test_that("grouped rxEt ui data source stays compressed", {
    ev <- eventTable()
    ev$add.dosing(dose = 100, nbr.doses = 2, dosing.interval = 12)
    ev$add.sampling(c(0, 1, 2, 12, 13, 24))
    ev <- et(ev, id = 1:5000)

    uiData <- .rxSolveUiEventData(ev)
    full <- as.data.frame(ev)
    grp <- .etGetGroups(.rxEtEnv(ev))

    expect_true(is.data.frame(uiData))
    expect_equal(nrow(uiData), nrow(grp[[1]]$data))
    expect_lt(nrow(uiData), nrow(full))
    expect_equal(unique(uiData$id), 1L)
  })
})

rxTest({
  mod <- rxode2("
a = 6
b = 0.6
d/dt(intestine) = -a*intestine
d/dt(blood)     = a*intestine - b*blood
")

  et <- eventTable()
  et$add.sampling(seq(0, 10, by = 1 / 24))
  et$add.dosing(
    dose = 2 / 24, rate = 2, start.time = 0,
    nbr.doses = 10, dosing.interval = 1
  )
  et <- et |> et(0.05, evid = 2)

  s1 <- solve(mod, et, addDosing = FALSE)

  s2 <- solve(mod, et, addDosing = NULL)

  s3 <- solve(mod, et, addDosing = NA)

  s4 <- solve(mod, et, addDosing = TRUE)


  s1tbs <- solve(mod, et, addDosing = FALSE, returnType = "data.frame.TBS")

  s2tbs <- solve(mod, et, addDosing = NULL, returnType = "data.frame.TBS")

  s3tbs <- solve(mod, et, addDosing = NA, returnType = "data.frame.TBS")

  s4tbs <- solve(mod, et, addDosing = TRUE, returnType = "data.frame.TBS")

  test_that("evid is included with evid=2", {
    expect_true("evid" %in% names(s1))
    expect_false("evid" %in% names(s2))
    expect_true("evid" %in% names(s3))
    expect_true("evid" %in% names(s4))
    expect_true("evid" %in% names(s1tbs))
    expect_false("evid" %in% names(s2tbs))
    expect_true("evid" %in% names(s3tbs))
    expect_true("evid" %in% names(s4tbs))
  })

  test_that("Includes and ignores EVID=2", {
    expect_true(any(s1$time == 0.05))
    expect_true(any(s1tbs$time == 0.05))
    expect_false(any(s2$time == 0.05))
    expect_false(any(s2tbs$time == 0.05))
  })

  test_that("addDosing=TRUE does not add an extra row", {
    expect_equal(sort(unique(s4$evid)), c(0L, 1L, 2L))
    expect_equal(sort(unique(s4tbs$evid)), c(0L, 1L, 2L))
  })

  ## Test mixed solved and ODEs
  mod2 <- rxode2({
    ## the order of variables do not matter, the type of compartmental
    ## model is determined by the parameters specified.
    CL ~ TCL * exp(eta.Cl)
    C2 ~ linCmt(KA, CL, V2, Q, V3)
    eff(0) <- 1 ## This specifies that the effect compartment starts at 1.
    d / dt(eff) ~ Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
    ##
    resp <- eff + err1
    pk <- C2 * exp(err2)
  })

  sigma <- diag(2) * 0.05
  dimnames(sigma) <- list(c("err1", "err2"), c("err1", "err2"))


  ev <- eventTable() |>
    add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12, dosing.to = 2) |>
    add.dosing(dose = 20000, nbr.doses = 5, start.time = 120, dosing.interval = 24, dosing.to = 2) |>
    add.sampling(0:240)

  ev <- ev |> et(0.5, evid = 2)

  pk4 <- rxSolve(mod2, c(
    KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
    Kin = 1, Kout = 1, EC50 = 200
  ),
  omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
  nSub = 4, events = ev, sigma = sigma, cores = 1, addDosing = TRUE
  )

  pk5 <- rxSolve(mod2, c(
    KA = 2.94E-01, TCL = 1.86E+01, V2 = 4.02E+01, Q = 1.05E+01, V3 = 2.97E+02,
    Kin = 1, Kout = 1, EC50 = 200
  ),
  omega = matrix(0.2, dimnames = list("eta.Cl", "eta.Cl")),
  nSub = 4, events = ev, sigma = sigma, cores = 1, addDosing = NULL
  )

  test_that("Multi-subject solves keep evid=2", {
    expect_true(any(pk4$time == 0.5))
    expect_false(any(pk5$time == 0.5))
  })
})

rxTest({
  if (!.Call(`_rxode2_isIntel`)) {
    ## Example 3.1 from
    ## "Solving Differential Equations in R" by Soetaert et al (2012)
    ## https://cran.r-project.org/web/packages/diffEq/vignettes/ODEinR.pdf Example #1
    skip_if_not(file.exists(test_path("test-example-3-1.rds")))
    df.test <- readRDS(test_path("test-example-3-1.rds"))

    ms <- .methods0
    ode <- rxode2(model = "d/dt(y) = r * y * (1.0 - y/K);")
    .rxWithOptions(list(digits = 6), {
      for (meth in ms) {
        # context(sprintf("Example 3-1 (%s)", meth))

        ## create event table with times at which we observe the system
        et <- eventTable(time.units = NA)
        et$add.sampling(seq(from = 0, to = 20, by = 0.2))

        ## same model, different initial values (2 and 12)
        out1 <- ode$solve(params = c(r = 1, K = 10), events = et, inits = c(y = 2))
        out2 <- ode$solve(params = c(r = 1, K = 10), events = et, inits = c(y = 12))

        ## matplot(x = out1[,1], y = cbind(out1[,2], out2[,2]), type = "l",
        ##    main = "logistic growth", xlab="time", ylab="")

        ## Now use a non-stiff solver
        out2.ns <-
          ode$solve(params = c(r = 1, K = 10), events = et, inits = c(y = 12), method = "dop853")

        df <- round(cbind(out1, out2, out2.ns), 3)

        test_that("Runs example 3.1 correctly", {
          expect_equal(df, round(df.test, 3))
        })

        head(cbind(out1, out2, out2.ns), n = 15)
      }
    })
  }
})

rxTest({
  test_that("get information from rxUi", {

    ## Test mixed solved and ODEs
    mod2 <- function() {
      sigma <- lotri({
        err1 ~ 0.05
        err2 ~ 0.05
      })
      ini({
        KA = 2.94E-01
        TCL = 1.86E+01
        V2 = 4.02E+01
        Q = 1.05E+01
        V3 = 2.97E+02
        Kin = 1
        Kout = 1
        EC50 = 200
        eta.Cl ~ 0.2
      })
      model({
        ## the order of variables do not matter, the type of compartmental
        ## model is determined by the parameters specified.
        CL <- TCL * exp(eta.Cl)
        C2 <- linCmt(KA, CL, V2, Q, V3)
        eff(0) <- 1 ## This specifies that the effect compartment starts at 1.
        d/dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
        ##
        resp <- eff + err1
        pk <- C2 * exp(err2)
      })
    }

    f <- mod2()

    ev <- eventTable() |>
      add.dosing(dose = 10000, nbr.doses = 10, dosing.interval = 12, dosing.to = 2) |>
      add.dosing(dose = 20000, nbr.doses = 5, start.time = 120, dosing.interval = 24, dosing.to = 2) |>
      add.sampling(0:240)

    ev <- ev |> et(0.5, evid = 2)

    pk4 <- rxSolve(f, events=ev, nSub=4, cores = 1, addDosing = TRUE)
    expect_true(inherits(pk4, "rxSolve"))

  })

})

rxTest({
  test_that("rxFlag works", {
    
    mod2 <- rxode2({
      a <- 6
      b <- 0.6
      d / dt(intestine) <- -a * intestine
      lag(intestine) <- 2
      d / dt(blood) <- a * intestine - b * blood
      printf("%f\n", rxFlag)
      flag <- rxFlag
    }, fullPrint=FALSE)
    
    obs <- units::set_units(seq(0, 10, by = 1 / 24), "days")

    et <- eventTable(time.units = "days")
    et$add.sampling(obs)
    et$add.dosing(
      dose = 2 / 24, start.time = 0,
      nbr.doses = 10, dosing.interval = 1
    )

    sink("flag.csv")
    cat("flag\n")
    solve2 <- solve(mod2, et)
    sink()
    f <- read.csv("flag.csv")
    unlink("flag.csv")

    expect_true(all(solve2$flag == 11))
    expect_true(all(f$flag == 1))

    mod3 <- rxode2({
      a <- 6
      b <- 0.6
      d / dt(intestine) <- -a * intestine
      lag(intestine) <- 2
      d / dt(blood) <- a * intestine - b * blood
      printf("%f\n", rxFlag)
      flag <- rxFlag
    }, fullPrint=TRUE)

    sink("flag.csv")
    cat("flag\n")
    solve3 <- solve(mod3, et)
    sink()
    f3 <- read.csv("flag.csv")
    unlink("flag.csv")

    expect_true(all(solve2$flag == 11))
    expect_equal(sort(unique(f3$flag)), c(1, 5, 11))

  })
})

rxTest({
  # test ODE parsing for syntax errors
  # context("Test errors with non-symmetric matrices")

  ## Jauslin's IGI (ogtt) model
  ode <- "
# volumes in L
# Cl in L/min
  Vg = tVg*exp(eta.Vg);
  Q =tQ*exp(eta.Q);
  Vi = tVi*exp(eta.Vi);
  Clg = tClg*exp(eta.Clg)
  Clgi = tClgi*exp(eta.Clgi)
  Cli = tCli*exp(eta.Cli)
  kcp = Q/Vg;
  kpc = Q/Vp;
  kg = Clg/Vg;
  kgi = Clgi/Vg;
  ki = Cli/Vi;
  ktr = n/mtt;
  lnfac = log(2.5066)+(n+0.5)*log(n)-n+log(1+1/(12*n));
  absg = exp(log(75000*(0.811/10))+log(ktr) + n*log(ktr*t+0.00001)-ktr*t-lnfac);
  #Gcm1 = ((Ge1/Gss)+0.001)^GPRG;
  Gcm2 = ((Ge2/Gss)+0.001)^IPRG;
  Iabsg = 1 + (Emax*absg)/(absg+CA50);
  Isec = Iss*Cli*Gcm2*Iabsg;
  Gpro = Gss*(kg + (kgi*Iss))*Vg;

  depot(0) = 0;
  trn(0) = 0;
  Gc(0) = Gss*Vg;
  #Ge1(0) = Gss;
  Ge2(0) = Gss;
  Gp(0) = kcp*Gss*Vg/kpc;
  Ie(0) = Iss;
  I(0) = Iss*Vi;
  Isec_0 = Iss*ki*Vi;

  d/dt(depot) = -absg;
  d/dt(trn) = absg - ka*trn;
  d/dt(Gc) = ka*trn + Gss*(kg + (kgi*Iss))*Vg - (kg + (kgi*Ie))*Gc - (Gc*kcp - Gp*kpc);
  d/dt(Gp) = Gc*kcp - Gp*kpc;
  d/dt(Ge2) = keog*(Gc/Vg) - keog*Ge2;
  d/dt(I) = (Iss*ki*Vi*(((Ge2/Gss)+0.001)^IPRG)*Iabsg) - ki*I;
  d/dt(Ie) = keoi*(I/Vi) - keoi*Ie;

  Cg = Gc/Vg
  Cp = Gp/Vp
  Ci = I/Vi

"

  mod <- rxode2(model = ode)

  theta <- c(
    tVg = 9.33, Vp = 8.56, tQ = 0.442,
    tClg = 0.0287, tClgi = 0.0059,
    Iss = 9.3, Gss = 150,
    tCli = 1.22, tVi = 6.09,
    IPRG = 1.42,
    keog = 0.0289, keoi = 0.0213,
    mtt = 34.9, n = 1.27,
    Emax = 1.47, ka = 0.02865, CA50 = 14.8
  )

  omega1 <- matrix(c(0.0887, -0.192, 0.0855, 0.0, 0.73, -0.12, 0.0, 0.0, 0.165), 3, 3,
                   dimnames = list(NULL, c("eta.Vg", "eta.Q", "eta.Vi"))
                   )
  omega2 <- matrix(0.352, dimnames = list(NULL, "eta.Clg"))
  omega3 <- matrix(0.207, dimnames = list(NULL, "eta.Clgi"))
  omega4 <- matrix(0.0852, dimnames = list(NULL, "eta.Cli"))

  et <- eventTable(amount.units = "mg", time.units = "min")
  et$add.dosing(dose = 75000, nbr.doses = 1, start.time = 0, dosing.to = 1)
  et$add.sampling(0:360)

  et <- et |> et(id = 1:7)

  rxWithSeed(
    42,
    {
      test_that("non-symmetric omegas throw errors", {
        expect_warning(
          rxSolve(mod, theta, et, omega = list(omega1, omega2, omega3, omega4)),
          "omega.*symmetric"
        )
      })

      test_that("non-symmetric sigmas throw errors", {
        expect_warning(
          rxSolve(mod, theta, et,
                  omega = lotri(eta.Cli ~ 0.0854),
                  sigma = list(omega1, omega2, omega3)
                  ),
          "sigma.*symmetric"
        )
        expect_warning(
          expect_warning(rxSolve(mod, theta, et,
                                 sigma = list(omega1, omega2, omega3, omega4)
                                 ),
          "sigma.*symmetric"))
      })

      tMat <- mod$params
      tMat <- tMat[regexpr("eta", tMat) == -1]
      tM <- diag(length(tMat))
      dimnames(tM) <- list(tMat, tMat)
      tM[1, 2] <- 2

      omega <- lotri({
        eta.Vg + eta.Q + eta.Vi ~
          c(
            0.0887,
            -0.1920, 0.73,
            0.0855, -0.12, 0.165
          )
        eta.Clg ~ 0.352
        eta.Clgi ~ 0.207
        eta.Cli ~ 0.0852
      })

      test_that("non-symmetric sigmas throw errors", {
        expect_warning(
          rxSolve(mod, theta, et, thetaMat = tM, omega = omega),
          "thetaMat.*symmetric"
        )
      })
    }
  )
})

rxTest({
  ms <- .methods0
  for (m in ms) {

    et <- eventTable() |>
      add.dosing(dose = 3, nbr.doses = 6, dosing.interval = 8) |>
      add.sampling(seq(0, 48, length.out = 200))

    ode.1c <- rxode2({
      V <- 20
      Cl <- 1
      fc <- 1
      C2 <- center / V
      d / dt(center) ~ -Cl * C2
      f(center) <- fc
    })

    ode.1cR <- rxode2({
      V <- 20
      Cl <- 1
      C2 <- center / V
      fc <- 1
      d / dt(center) ~ -Cl * C2
      rate(center) <- rateIn
      f(center) <- fc
    })

    ode.1cD <- rxode2({
      V <- 20
      Cl <- 1
      C2 <- center / V
      fc <- 1
      d / dt(center) ~ -Cl * C2
      dur(center) <- durIn
      f(center) <- fc
    })

    d <- 3
    et2 <- eventTable() |>
      add.dosing(dose = d, nbr.doses = 2, dosing.interval = 8) |>
      add.sampling(seq(0, 48, length.out = 200))

    x2 <- solve(ode.1c, et2, method = m)

    c0 <- setNames(d / rxInits(ode.1c)["V"], NULL)

    ke <- with(as.list(rxInits(ode.1c)), {
      Cl / V
    })

    # context(sprintf("Steady state IV Bolus (%s)", m))

    test_that(paste0("Non steady state dose makes sense; meth=",m), {
      expect_equal(x2$C2[1], c0)
    })

    tol <- 1e-3

    for (ii in seq(2, 24, by = 2)) {
      ## now setup bolus steady state evid
      et3 <- et() |>
        et(amt = d, ss = 1, ii = ii) |>
        et(amt = d, time = ii) |>
        et(seq(0, 48, length.out = 200))
      x2 <- solve(ode.1c, et3, method = m)
      test_that(paste("Steady State dose makes sense for ii=", ii,"; meth=", m), {
        expect_equal(x2$C2[1], c0 / (1 - exp(-ke * ii)), tolerance = tol)
      })
    }

    # context(sprintf("Steady state Infusions (%s)", m))

    ## Need to test:
    ## - Fixed Infusion w/rate (check)
    ## - Fixed Infusion w/rate + bioavailability change (check)
    ## - Fixed Infusion w/dur (check)
    ## - Fixed Infusion w/dur + bioavailability change
    ## - Modeled rate (check)
    ## - Modeled rate + bioavailability change (check)
    ## - Modeled duration (check)
    ## - Modeled duration + bioavailability change
    for (dur in c(0.5, 1)) {
      for (ii in seq(2, 24, by = 2)) {
        ## Fixed rate
        et3 <- et() |>
          et(amt = d, ss = 1, ii = ii, rate = d / dur) |>
          et(time = ii, amt = d, ii = ii, addl = floor(24 / ii), rate = d / dur) |>
          et(c(dur, seq(0, 24, length.out = 200)))
        x2 <- solve(ode.1c, et3, method = m, maxsteps = 10000)
        infMax <- with(
          as.list(rxInit(ode.1c)),
          d / (Cl * dur) * (1 - exp(-ke * dur)) / (1 - exp(-ke * ii))
        )
        inf0 <- with(
          as.list(rxInit(ode.1c)),
          infMax * exp(-ke * (ii - dur))
        )

        test_that(paste("Infusion Steady State dose makes sense for ii=", ii, " dur=", dur, "(rate); meth=", m), {
          expect_equal(x2$C2[x2$time == dur], infMax, tolerance = tol)
          expect_equal(x2$C2[1], inf0, tolerance = tol)
        })

        ## Fixed duration
        et3 <- et() |>
          et(amt = d, ss = 1, ii = ii, dur = dur) |>
          et(c(dur, seq(0, 24, length.out = 19)))
        x2 <- solve(ode.1c, et3, method = m, maxsteps = 10000)
        test_that(paste("Infusion Steady State dose makes sense for ii=", ii, " dur=", dur, "(dur); met=", m), {
          expect_equal(x2$C2[x2 == dur], infMax, tolerance = tol)
          expect_equal(x2$C2[1], inf0, tolerance = tol)
        })
        ## rate modeled
        et3 <- et() |>
          et(amt = d, ss = 1, ii = ii, rate = -1) |>
          et(c(dur, seq(0, 24, length.out = 19)))
        x2 <- rxSolve(ode.1cR, et3, c(rateIn = d / dur), method = m, maxsteps = 10000)
        test_that(paste("Infusion Steady State dose makes sense for ii=", ii, " dur=", dur, "(rate modeled); meth=", m), {
          expect_equal(x2$C2[x2 == dur], infMax, tolerance = tol)
          expect_equal(x2$C2[1], inf0, tolerance = tol)
        })
        ## duration modeled
        et3 <- et() |>
          et(amt = d, ss = 1, ii = ii, rate = -2) |>
          et(c(dur, seq(0, 24, length.out = 19)))
        x2 <- rxSolve(ode.1cD, et3, c(durIn = dur), method = m, maxsteps = 10000)
        test_that(paste("Infusion Steady State dose makes sense for ii=", ii, " dur=", dur, "(dur modeled); meth=", m), {
          expect_equal(x2$C2[x2 == dur], infMax, tolerance = tol)
          expect_equal(x2$C2[1], inf0, tolerance = tol)
        })
        for (f in c(0.5, 1)) {
          if (dur * f < ii) {
            ## Now add modeled bioavailability change
            ## That changes duration
            infMax <- with(
              as.list(rxInit(ode.1c)),
              f * d / (Cl * dur * f) * (1 - exp(-ke * dur * f)) / (1 - exp(-ke * ii))
            )
            inf0 <- with(
              as.list(rxInit(ode.1c)),
              infMax * exp(-ke * (ii - dur * f))
            )
            et3 <- et() |>
              et(amt = d, ss = 1, ii = ii, rate = d / dur) |>
              et(time = ii, amt = d, ii = ii, addl = floor(24 / ii), rate = d / dur) |>
              et(c(dur * f, seq(0, 24, length.out = 400)))
            x2 <- solve(ode.1c, et3, c(fc = f), method = m, maxsteps = 10000)
            test_that(paste("Infusion Steady State dose makes sense for f= ", f, "ii=", ii, " dur=", dur, "(rate); meth=", m), {
              expect_equal(x2$C2[x2 == dur * f], infMax, tolerance = tol)
              expect_equal(x2$C2[1], inf0, tolerance = tol)
            })
            ## rate modeled
            et3 <- et() |>
              et(amt = d, ss = 1, ii = ii, rate = -1) |>
              et(c(dur * f, seq(0, 24, length.out = 19)))
            x2 <- rxSolve(ode.1cR, et3, c(fc = f, rateIn = d / dur), method = m, maxsteps = 10000)
            test_that(paste("Infusion Steady State dose makes sense for f= ", f, " ii=", ii, " dur=", dur, "(rate modeled); meth=", m), {
              expect_equal(x2$C2[x2 == dur * f], infMax, tolerance = tol)
              expect_equal(x2$C2[1], inf0, tolerance = tol)
            })
          }
          ## Add modeled bioavailability change
          ## That changes rate
          infMax <- with(
            as.list(rxInit(ode.1c)),
            f * d / (Cl * dur) * (1 - exp(-ke * dur)) / (1 - exp(-ke * ii))
          )
          inf0 <- with(
            as.list(rxInit(ode.1c)),
            infMax * exp(-ke * (ii - dur))
          )
          et3 <- et() |>
            et(amt = d, ss = 1, ii = ii, dur = dur) |>
            et(time = ii, amt = d, ii = ii, addl = floor(24 / ii), dur = dur) |>
            et(unique(c(dur, seq(0, 24, length.out = 200))))
          x2 <- solve(ode.1c, et3, c(fc = f), method = m, maxsteps = 10000)
          test_that(paste("Infusion Steady State dose makes sense for f= ", f, "ii=", ii, " dur=", dur, "(dur); meth=", m), {
            expect_equal(x2$C2[x2 == dur], infMax, tolerance = tol)
            expect_equal(x2$C2[1], inf0, tolerance = tol)
          })
          et3 <- et() |>
            et(amt = d, ss = 1, ii = ii, rate = -2) |>
            et(c(dur, seq(0, 24, length.out = 19)))
          x2 <- rxSolve(ode.1cD, et3, c(fc = f, durIn = dur), method = m, maxsteps = 10000)
          test_that(paste("Infusion Steady State dose makes sense for f=", f, "ii=", ii, " dur=", dur, "(dur modeled); meth=", m), {
            expect_equal(x2$C2[x2 == dur], infMax, tolerance = tol)
            expect_equal(x2$C2[1], inf0, tolerance = tol)
          })
        }
      }
    }

    ## steady-state = 2 for bolus
    # context(sprintf("Bolus SS=2 (%s)", m))
    test_that(paste0("bolus SS=2; meth=", m), {
      e2 <- et(amt = 20, ii = 24, ss = 2, time = 12) |>
        et(seq(0, 24, length.out = 100))
      s2 <- rxSolve(ode.1c, e2)
      e3 <- et(amt = 20, ii = 24, ss = 1, time = 12) |>
        et(seq(0, 24, length.out = 100))
      s3 <- rxSolve(ode.1c, e3)
      expect_equal(as.data.frame(s2), as.data.frame(s3))
      e1 <- et(amt = 10, ii = 24, ss = 1, time = 0) |>
        et(seq(0, 24, length.out = 100))
      s1 <- rxSolve(ode.1c, e1)
      e2 <- et(amt = 20, ii = 24, ss = 2, time = 12) |>
        et(seq(0, 24, length.out = 100))
      s2 <- rxSolve(ode.1c, e2)
      e3 <- c(e1, e2, ii = 0) |>
        et(seq(0, 24, length.out = 100))
      s3 <- rxSolve(ode.1c, e3)
      expect_equal(s1$C2 + s2$C2, s3$C2, tolerance = tol)
      for (f in c(0.5, 2)) {
        s1 <- rxSolve(ode.1c, e1, c(fc = f))
        s2 <- rxSolve(ode.1c, e2, c(fc = f))
        s3 <- rxSolve(ode.1c, e3, c(fc = f))
        expect_equal(s1$C2 + s2$C2, s3$C2, tolerance = tol)
      }
    })
    # context(sprintf("IV Infusion SS=2 (%s)", m))
    d <- 10
    ## Changing Bioavailability causes changes in these results...
    for (f in c(0.5, 1, 2)) {
      for (dur in c(1, 2)) {
        ## Fixed rate
        e1 <- et() |>
          et(amt = d, ss = 1, ii = 24, rate = d / dur) |>
          et(seq(0, 24, length.out = 200))
        s1 <- solve(ode.1c, e1, c(fc = f), method = m, maxsteps = 1000000)
        e2 <- et() |>
          et(time = 12, amt = 2 * d, ss = 2, ii = 24, rate = d * 2 / dur) |>
          et(seq(0, 24, length.out = 200))
        if (f == 1) {
          s2 <- solve(ode.1c, e2, c(fc = f), method = m, maxsteps = 1000000)
          e3 <- et() |>
            et(amt = d, ss = 1, ii = 24, rate = d / dur) |>
            et(time = 12, amt = 2 * d, ss = 2, ii = 24, rate = d * 2 / dur) |>
            et(seq(0, 24, length.out = 200))
          s3 <- solve(ode.1c, e3, c(fc = f), method = m, maxsteps = 1000000)
          test_that(paste("Infusion SS=2 dose makes sense for f=", f, " dur=", dur, "(rate); meth=", m), {
            expect_equal(s1$C2 + s2$C2, s3$C2, tolerance = 1e-4)
          })
        } else {
          test_that(paste("Infusion SS=2 dose makes sense for f=", f, " dur=", dur, "(dur); meth=", m), {
            expect_error(solve(ode.1c, e2, c(fc = f), method = m, maxsteps = 1000000))
          })
        }
        ## Fixed duration
        e1 <- et() |>
          et(amt = d, ss = 1, ii = 24, dur = dur) |>
          et(seq(0, 24, length.out = 200))
        s1 <- solve(ode.1c, e1, c(fc = f), method = m, maxsteps = 1000000)
        e2 <- et() |>
          et(time = 12, amt = 2 * d, ss = 2, ii = 24, dur = dur) |>
          et(seq(0, 24, length.out = 200))
        if (f == 1) {
          s2 <- solve(ode.1c, e2, c(fc = f), method = m, maxsteps = 1000000)
          e3 <- et() |>
            et(amt = d, ss = 1, ii = 24, rate = d / dur) |>
            et(time = 12, amt = 2 * d, ss = 2, ii = 24, rate = d * 2 / dur) |>
            et(seq(0, 24, length.out = 200))
          s3 <- solve(ode.1c, e3, c(fc = f), method = m, maxsteps = 1000000)
          test_that(paste("Infusion SS=2 dose makes sense for f=", f, " dur=", dur, "(dur); meth=", m), {
            expect_equal(s1$C2 + s2$C2, s3$C2, tolerance = 1e-4)
          })
        } else {
          test_that(paste("Infusion SS=2 dose makes sense for f=", f, " dur=", dur, "(dur); meth=", m), {
            expect_error(solve(ode.1c, e2, c(fc = f), method = m, maxsteps = 1000000))
          })
        }
        ## Modeled rate when used with SS=2
        e1 <- et() |>
          et(amt = d, ss = 1, ii = 24, rate = d / dur) |>
          et(seq(0, 24, length.out = 200))
        s1 <- solve(ode.1cR, e1, c(fc = f, rateIn = 2 * d / dur), method = m, maxsteps = 1000000)
        e2 <- et() |>
          et(time = 12, amt = 2 * d, ss = 2, ii = 24, rate = -1) |>
          et(seq(0, 24, length.out = 200))
        if (f == 1) {
          s2 <- solve(ode.1cR, e2, c(fc = f, rateIn = 2 * d / dur), method = m, maxsteps = 1000000)
          e3 <- et() |>
            et(amt = d, ss = 1, ii = 24, rate = d / dur) |>
            et(time = 12, amt = 2 * d, ss = 2, ii = 24, rate = -1) |>
            et(seq(0, 24, length.out = 200))
          s3 <- solve(ode.1cR, e3, c(fc = f, rateIn = 2 * d / dur), method = m, maxsteps = 1000000)
          test_that(paste("Infusion SS=2 dose makes sense for f=", f, " dur=", dur, "(modeled rate); meth=", m), {
            expect_equal(s1$C2 + s2$C2, s3$C2, tolerance = 1e-4)
          })
        } else {
          test_that(paste("Infusion SS=2 dose makes sense for f=", f, " dur=", dur, "(modeled rate); meth=", m), {
            expect_error(solve(ode.1cR, e2, c(fc = f, rateIn = 2 * d / dur), method = m, maxsteps = 1000000))
          })
        }
        ## Modeled duration when used with SS=2
        e1 <- et() |>
          et(amt = d, ss = 1, ii = 24, dur = dur) |>
          et(seq(0, 24, length.out = 200))
        s1 <- solve(ode.1cD, e1, c(fc = f, durIn = dur), method = m, maxsteps = 1000000)
        e2 <- et() |>
          et(time = 12, amt = 2 * d, ss = 2, ii = 24, rate = -2) |>
          et(seq(0, 24, length.out = 200))
        if (f == 1) {
          s2 <- solve(ode.1cD, e2, c(fc = f, durIn = dur), method = m, maxsteps = 1000000)
          e3 <- et() |>
            et(amt = d, ss = 1, ii = 24, dur = dur) |>
            et(time = 12, amt = 2 * d, ss = 2, ii = 24, rate = -2) |>
            et(seq(0, 24, length.out = 200))
          s3 <- solve(ode.1cD, e3, c(fc = f, durIn = dur), method = m, maxsteps = 1000000)
          test_that(paste("Infusion SS=2 dose makes sense for f=", f, " dur=", dur, "(modeled duration); meth=", m), {
            expect_equal(s1$C2 + s2$C2, s3$C2, tolerance = 1e-4)
          })
        } else {
          test_that(paste("Infusion SS=2 dose makes sense for f=", f, " dur=", dur, "(modeled duration); meth=", m), {
            expect_error(solve(ode.1cD, e2, c(fc = f, durIn = dur), method = m, maxsteps = 1000000))
          })
        }
      }
    }

    # context("Constant Infusion steady state")

    test_that(paste0("constant infusion steady state; meth=", m), {
      ode.1cR <- rxode2({
        V <- 20
        Cl <- 1
        C2 <- center / V
        fc <- 1
        d / dt(center) ~ -Cl * C2
        rate(center) <- rateIn
        f(center) <- fc
      })

      s <- rxSolve(ode.1cR, c(rateIn = 3), et(amt = 0, rate = 10, ss = 1))

      expect_equal(s$C2[1], 10, tolerance = 1e-5)

      s <- rxSolve(ode.1cR, c(rateIn = 3), et(amt = 0, rate = -1, ss = 1))

      expect_equal(s$C2[1], 3, tolerance = 1e-5)
    })
  }
})

rxTest({
  # Test Scaling Factors
  ode <-
    rxode2({
      b <- -1
      d / dt(X) <- a * X + Y * Z
      d / dt(Y) <- b * (Y - Z)
      d / dt(Z) <- -X * Y + c * Y - Z
    })

  et <- eventTable() # default time units
  et$add.sampling(seq(from = 0, to = 100, by = 0.01))

  et$c <- et$time + 1

  out <- rxSolve(ode,
                 params = c(a = -8 / 3, b = -10),
                 events = et,
                 inits = c(X = 1, Y = 1, Z = 1)
                 )

  x1 <- out$X
  y1 <- out$Y
  z1 <- out$Z

  ## Test scaling factor
  test_that("Scaling factors specified by S#= work", {
    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10),
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1),
                   S1 = 2
                   )
    expect_equal(out$X, x1 / 2)
    expect_equal(out$Y, y1)
    expect_equal(out$Z, z1)

    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10),
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1),
                   S2 = 2
                   )
    expect_equal(out$X, x1)
    expect_equal(out$Y, y1 / 2)
    expect_equal(out$Z, z1)
    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10),
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1),
                   S3 = 2
                   )
    expect_equal(out$X, x1)
    expect_equal(out$Y, y1)
    expect_equal(out$Z, z1 / 2)
  })

  test_that("Multiple Scaling Factors", {
    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10),
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1),
                   S1 = 2, S2 = 2
                   )
    expect_equal(out$X, x1 / 2)
    expect_equal(out$Y, y1 / 2)
    expect_equal(out$Z, z1)
  })

  test_that("Lower Case scaling factor", {
    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10),
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1),
                   s1 = 2
                   )
    expect_equal(out$X, x1 / 2)
    expect_equal(out$Y, y1)
    expect_equal(out$Z, z1)
  })

  test_that("Duplicate Scaling factors raise error", {
    expect_error(out <- rxSolve(ode,
                                params = c(a = -8 / 3, b = -10),
                                events = et,
                                inits = c(X = 1, Y = 1, Z = 1),
                                s1 = 2, s1 = 4
                                ))
  })

  test_that("Scaling with two methods creates an error", {
    expect_error(out <- rxSolve(ode,
                                params = c(a = -8 / 3, b = -10),
                                events = et,
                                inits = c(X = 1, Y = 1, Z = 1),
                                scale = c(X = 2),
                                s1 = 2
                                ))
  })

  test_that("Scaling for a unnumbered compartment rases a warning", {
    expect_warning(out <- rxSolve(ode,
                                  params = c(a = -8 / 3, b = -10),
                                  events = et,
                                  inits = c(X = 1, Y = 1, Z = 1),
                                  s1 = 2, s7 = 3
                                  ))
    expect_equal(out$X, x1 / 2)
    expect_equal(out$Y, y1)
    expect_equal(out$Z, z1)
  })

  test_that("Scaling factors specified by scale(X=...) work", {
    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10),
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1),
                   scale = c(X = 2)
                   )
    expect_equal(out$X, x1 / 2)
    expect_equal(out$Y, y1)
    expect_equal(out$Z, z1)

    ## Issue #18
    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10),
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1),
                   scale = list(X = 2)
                   )
    expect_equal(out$X, x1 / 2)
    expect_equal(out$Y, y1)
    expect_equal(out$Z, z1)

    expect_error(rxSolve(ode,
                         params = c(a = -8 / 3, b = -10),
                         events = et,
                         inits = c(X = 1, Y = 1, Z = 1),
                         scale = list(X = 1:2)
                         ))
    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10),
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1),
                   scale = c(Y = 2)
                   )
    expect_equal(out$X, x1)
    expect_equal(out$Y, y1 / 2)
    expect_equal(out$Z, z1)

    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10),
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1),
                   scale = c(Z = 2)
                   )
    expect_equal(out$X, x1)
    expect_equal(out$Y, y1)
    expect_equal(out$Z, z1 / 2)

    out <- rxSolve(ode,
                   params = c(a = -8 / 3, b = -10),
                   events = et,
                   inits = c(X = 1, Y = 1, Z = 1),
                   scale = c(X = 2, Z = 2)
                   )
    expect_equal(out$X, x1 / 2)
    expect_equal(out$Y, y1)
    expect_equal(out$Z, z1 / 2)
  })
})

#' Creates a logical matrix for block matrixes.
#'
#' @param mat Matrix
#' @param i Row/column where block matrix should be setup.
#'
#' @return A logical matrix returning where the elements should be
#'     zero.
#'
#' @keywords internal
#' @export
rxBlockZeros <- function(mat, i) {
  return(!((row(mat) > i & col(mat) > i) | (row(mat) <= i & col(mat) <= i)))
}
rxIsBlock <- function(mat, i) {
  if (missing(i)) {
    for (j in 1:(dim(mat)[1] - 1)) {
      if (rxIsBlock(mat, j)) {
        return(TRUE)
      } else {
        return(FALSE)
      }
    }
  } else {
    return(all(mat[rxBlockZeros(mat, i)] == 0))
  }
}

## Version #2
rxSymInvC2 <- function(mat1, diag.xform = c("sqrt", "log", "identity"),
                       allow.cache = TRUE) {
  rxReq("symengine")
  if (!all(as.vector(mat1) == 1)) {
    stop("this has to be a matrix of all 1s or 0s", call. = FALSE)
  }
  if (any(diag(mat1) == 0)) {
    stop("diagonal elements must be non-zero", call. = FALSE)
  }
  cache.file <- file.path(
    rxTempDir(),
    sprintf(
      "rx_%s2.inv",
      digest::digest(deparse(list(mat1, diag.xform)))
    )
  )
  cache.file2 <- file.path(system.file("inv", package = "rxode2"), cache.file)
  if (allow.cache && file.exists(cache.file)) {
    load(cache.file)
    return(ret)
  } else if (allow.cache && file.exists(cache.file2)) {
    load(cache.file2)
    return(ret)
  } else {
    diag.xform <- match.arg(diag.xform)
    message("diagonal form: ", diag.xform)
    num <- as.vector(mat1[upper.tri(mat1, TRUE)])
    i <- 0
    num <- sapply(num, function(x) {
      if (x == 1) {
        i <<- i + 1
        return(i)
      } else {
        return(0)
      }
    })
    mat1[upper.tri(mat1, TRUE)] <- num - 1
    mat1[lower.tri(mat1)] <- t(mat1)[lower.tri(mat1)]
    d <- dim(mat1)[1]
    mat1 <- paste0("t", mat1)
    mat1[mat1 == "t-1"] <- "0"
    mat1 <- matrix(mat1, d)

    diags <- -2 - as.numeric(sapply(diag(mat1), function(x) {
      substring(x, 2)
    }))
    mat2 <- mat1
    if (diag.xform == "sqrt") {
      ## The diagonal elements are assumed to be estimated as sqrt
      diag(mat2) <- sprintf("%s=2", diag(mat2))
      diag(mat1) <- sprintf("%s^2", diag(mat1))
    } else if (diag.xform == "log") {
      ## The diagonal elements are assumed to be estimated as log
      diag(mat2) <- sprintf("%s=3", diag(mat2))
      diag(mat1) <- sprintf("exp(%s)", diag(mat1))
    } else {
      ## The diagonal elements are assumed to be estimated as identity
      diag(mat2) <- sprintf("%s=4", diag(mat2))
      diag(mat1) <- sprintf("(%s)", diag(mat1))
    }
    ## Cholesky is upper tri
    mat1[lower.tri(mat1)] <- "0"
    mat2[lower.tri(mat2)] <- "0"
    mat2[upper.tri(mat2)] <- sprintf("%s=5", mat2[upper.tri(mat2)])
    mat2 <- as.vector(mat2)
    mat2 <- mat2[mat2 != "0"]
    omat <- fmat <- mat1
    vars <- paste0("t", seq(0, i - 1))
    sdiag <- sprintf("(%s)^2", diag(omat))
    se.mat <- symengine::Matrix(omat)
    message("calculate symbolic inverse: t(chol.mat) %*% chol.mat ...", appendLF = FALSE)
    se.inv <- symengine::t(se.mat) %*% se.mat
    message("done")
    ## Then take the derivatives
    ## These are used in equations #28 and #47
    ##
    ##
    ## - Omega^-1 %*% dOmega %*% Omega^-1 = d(Omega^-1)
    ##
    ## In Equation #28
    ##
    ## Therefore:
    ##  1/2*t(eta) %*% (Omega^-1 %*% dOmega %*% Omega^-1)  %*% eta =
    ## -1/2*t(eta)*d(Omega^-1)*eta
    ##
    ## The second part is:
    ##
    ##  -1/2*tr(Omega^-1*dOmega) = +1/2*tr(Omega^-1 %*% Omega %*% d(Omega^-1)*Omega) or
    ## +1/2*tr(d(Omega^-1)*Omega);
    ##
    ## Omega needs to be inverted, but not symbolically.
    ##
    ## In fact the whole of dOmega does not need to be calculated,
    ## rather the diff(D_Omega^-1) where the D is the LDL^T
    ## factorization

    ## Equation #29 uses d(Omega^-1)
    ##
    ## Equation #47 uses
    ## -t(eta)*Omega^-1*(dOmega)*Omega^-1*d(eta)
    ## t(eta)*d(Omega^-1)*d(eta)
    ## Therefore NO symbolic derivatives of anything but d(Omega^-1) are required; These are below:
    cnt.i <- 0
    cnt <- function() {
      message(".", appendLF = FALSE)
      if (cnt.i %% 5 == 0) {
        message(cnt.i, apendLF = FALSE)
      }
      if (cnt.i %% 50 == 0) {
        message("", appendLF = TRUE)
      }
      cnt.i <<- cnt.i + 1
    }
    i <- 0
    j <- 0

    .m1 <- symengine::Matrix(sdiag)
    diag <- paste(lapply(diags, function(x) {
      .m <- symengine::D(.m1, symengine::S(paste0("t", -(x + 2))))
      .n <- dim(.m)[1]
      .str <- paste(sapply(seq(1, .n), function(d) {
        sprintf("      REAL(ret)[%s] = %s;", d - 1, seC(.m[d, 1]))
      }), collapse = "\n")
      sprintf(
        "    %sif (theta_n == %s){\n%s\n    }", ifelse(-(x + 2) == 0, "", "else "), x - 1,
        .str
      )
    }), collapse = "\n")
    ## FIXME: this derivative expression is always the same. There
    ## should be a simpler way to express this...
    i <- 0
    omega0 <- sprintf(
      "    if (theta_n == 0){\n%s\n    }",
      paste(sapply(as.vector(omat), function(x) {
        ret <- sprintf("      REAL(ret)[%s] = %s;", i, seC(x))
        i <<- i + 1
        return(ret)
      }), collapse = "\n")
    )
    i <- 0
    omega1 <- sprintf(
      "    else if (theta_n == -1){\n%s\n    }",
      paste(sapply(as.vector(se.inv), function(x) {
        cnt()
        ret <- sprintf("      REAL(ret)[%s] = %s;", i, seC(x))
        i <<- i + 1
        return(ret)
      }), collapse = "\n")
    )
    omega1p <- paste(unlist(lapply(vars, function(x) {
      i <<- 0
      j <<- j + 1
      sprintf(
        "    else if (theta_n == %s){\n%s\n    }", j,
        paste(sapply(
          as.vector(symengine::D(se.inv, symengine::S(x))),
          function(x) {
            ret <- sprintf("      REAL(ret)[%s] = %s;", i, seC(x))
            i <<- i + 1
            return(ret)
          }
        ), collapse = "\n")
      )
    })), collapse = "\n")
    ##

    mat2 <- sprintf(

      "if (theta_n== NA_INTEGER){\n    SEXP ret=  PROTECT(Rf_allocVector(INTSXP,%s));\n%s\n    UNPROTECT(1);\n    return(ret);  \n}\n",
      length(mat2), paste(paste(gsub(rex::rex("t", capture(any_numbers), "="), "    INTEGER(ret)[\\1]=", mat2), ";", sep = ""),
        collapse = "\n"
      )
    )
    matExpr <- sprintf("  if (theta_n >= -1){\n    SEXP ret = PROTECT(Rf_allocMatrix(REALSXP, %s, %s));for (int i = 0; i < %s; i++){REAL(ret)[i]=0;}\n", d, d, d * d)
    vecExpr <- sprintf("    UNPROTECT(1);\n    return(ret);\n  } else {\n    SEXP ret = PROTECT(Rf_allocVector(REALSXP, %s));for(int i = 0; i < %s; i++){REAL(ret)[i]=0;}\n%s\n    UNPROTECT(1);\n    return(ret);\n  }", d, d, diag)
    src <- sprintf(
      "  int theta_n = INTEGER(tn)[0];\n  %s\nif (theta_n == -2){\n    SEXP ret = PROTECT(Rf_allocVector(INTSXP, 1));\n    INTEGER(ret)[0] = %s;\n    UNPROTECT(1);\n    return ret;\n  }\n  else if (theta_n < %s || theta_n > %s){\n    Rf_error(\"d(Omega^-1) derivative outside bounds\");\n  }\n  else if (Rf_length(theta) != %s){\n    Rf_error(\"requires vector with %s arguments\");\n  }\n%s\n%s\n%s",
      mat2, length(vars), min(diags) - 1, length(vars), length(vars), length(vars),
      paste0(matExpr, omega0), omega1, paste0(omega1p, "\n", vecExpr)
    )
    src <- strsplit(src, "\n")[[1]]
    reg <- rex::rex(any_spaces, "REAL(ret)[", any_numbers, "]", any_spaces, "=", any_spaces, "0", any_spaces, ";")
    ## Take out the =0; expressions
    w <- which(regexpr(reg, src) != -1)
    if (length(w) > 0) {
      src <- paste(src[-w], collapse = "\n")
    } else {
      src <- paste(src, collapse = "\n")
    }
    message("done")
    fmat <- matrix(sapply(as.vector(fmat), function(x) {
      force(x)
      return(rxFromSE(x))
    }), d)
    ret <- paste0("#define warning Rf_warning\n#define Rx_pow_di R_pow_di\n#define Rx_pow R_pow\n", src)
    ret <- list(ret, fmat)
    if (allow.cache) {
      save(ret, file = cache.file)
    }
    return(ret)
  }
}

#' Return the dimension of the built-in derivatives/inverses
#'
#' @keywords internal
#'
#' @return dimension of built-in derivatives/inverses
#'
#' @export
rxSymInvCholN <- function() {
  .Call(`_rxCholInv`, 0L, NULL, NULL)
}

rxSymInvCreate2C <- function(src) {
  return(inline::cfunction(signature(theta = "numeric", tn = "integer"), src, includes = "\n#include <Rmath.h>\n"))
}


## rxSymInvCreateC_.slow <- NULL
rxSymInvCreateC_ <- function(mat, diag.xform = c("sqrt", "log", "identity")) {
  diag.xform <- match.arg(diag.xform)
  mat2 <- mat
  mat2 <- rxInv(mat2)
  mat2 <- try({
    chol(mat2)
  }, silent = TRUE)
  if (inherits(mat2, "try-error")) {
    stop("initial 'omega' matrix inverse is non-positive definite", call. = FALSE)
  }
  mat3 <- mat2
  if (diag.xform == "sqrt") {
    diag(mat3) <- sqrt(diag(mat3))
  } else if (diag.xform == "log") {
    diag(mat3) <- log(diag(mat3))
  }
  w <- which(as.vector(lower.tri(mat3, TRUE)) * 1 == 1)
  elts <- as.vector(mat3)[w]
  ini <- as.vector(mat3)[as.vector(upper.tri(mat3, TRUE))]
  th.unscaled <- NULL
  for (i in seq_along(elts)) {
    if (elts[i] != 0) {
      th.unscaled[length(th.unscaled) + 1] <- elts[i]
    }
  }
  mat1 <- (mat != 0) * 1
  if (length(mat1) == 1) {
    mat1 <- matrix(mat1, 1)
  }
  dmat <- dim(mat1)[1] - 1
  block <- list()
  last <- 1
  if (dmat != 0) {
    for (i in 1:dmat) {
      if (all(mat1[rxBlockZeros(mat1, i)] == 0)) {
        s <- seq(last, i)
        cur <- matrix(as.double(mat[s, s]), length(s))
        last <- i + 1
        block[[length(block) + 1]] <- cur
      }
    }
  }
  if (length(block) != 0) {
    s <- seq(last, dmat + 1)
    cur <- matrix(as.double(mat[s, s]), length(s))
    block[[length(block) + 1]] <- cur
  }
  if (length(block) == 0) {
    if (diag.xform == "sqrt" && dim(mat1)[1] <= .Call(`_rxCholInv`, 0L, NULL, NULL)) {
      fmat <- mat1
      num <- as.vector(mat1[upper.tri(mat1, TRUE)])
      i <- 0
      num <- sapply(num, function(x) {
        if (x == 1) {
          i <<- i + 1
          return(i)
        } else {
          return(0)
        }
      })
      fmat[upper.tri(fmat, TRUE)] <- num - 1
      fmat[lower.tri(fmat)] <- t(fmat)[lower.tri(fmat)]
      d <- dim(fmat)[1]
      fmat <- paste0("t", fmat)
      fmat[fmat == "t-1"] <- "0"
      fmat <- matrix(fmat, d)
      if (diag.xform == "sqrt") {
        diag(fmat) <- sprintf("%s^2", diag(fmat))
      } else if (diag.xform == "log") {
        diag(fmat) <- sprintf("exp(%s)", diag(fmat))
      } else {
        diag(fmat) <- sprintf("(%s)", diag(fmat))
      }
      w <- which(fmat[upper.tri(fmat, TRUE)] != "0")
      if (length(w) == 0) {
        stop("zero matrix", call. = FALSE)
      }
      ## signature(theta="numeric", tn="integer")
      ## FIXME move these functions to Cpp?
      fn <- eval(bquote(function(theta, tn) {
        if (is.null(tn)) {
          .ret <- matrix(rep(TRUE, .(d * d)), nrow = .(d))
          .ret[lower.tri(.ret)] <- FALSE
          return(.ret[lower.tri(.ret, TRUE)])
        } else if (is.na(tn)) {
          new.theta <- rep(0.0, .((d + 1) * d / 2))
          new.theta[.(w)] <- theta
          return(.Call(`_rxCholInv`, .(as.integer(d)), as.double(new.theta), NA_integer_))
        }
        if (tn == -2L) {
          return(.(length(w)))
        }
        new.theta <- rep(0.0, .((d + 1) * d / 2))
        new.theta[.(w)] <- theta
        return(.Call(`_rxCholInv`, .(as.integer(d)), as.double(new.theta), as.integer(tn)))
      }))
      ret <- list(
        fmat = fmat,
        fn = fn,
        ini = ini,
        cache = TRUE
      )
      class(ret) <- "rxSymInvChol"
      return(ret)
    } else {
      ret <- rxSymInvC2(
        mat1 = mat1,
        diag.xform = diag.xform
      )
      th <- th.unscaled
      ret <- list(
        fmat = ret[[2]],
        ini = ini,
        fn = rxSymInvCreate2C(ret[[1]])
      )
      class(ret) <- "rxSymInvChol"
      return(ret)
    }
  } else {
    mat <- Matrix::.bdiag(block)
    matI <- lapply(block, rxSymInvCreateC_, diag.xform = diag.xform)
    ini <- unlist(lapply(matI, function(x) {
      x$ini
    }))
    ntheta <- sum(sapply(matI, function(x) {
      return(x$fn(NULL, -2L))
    }))
    i <- 1
    theta.part <- lapply(matI, function(x) {
      len <- x$fn(NULL, -2L)
      ret <- as.integer(seq(i, by = 1, length.out = len))
      i <<- max(ret) + 1
      return(ret)
    })
    ## FIXME move these to C/C++
    ## Drop the dependency on Matrix (since this is partially run in R)
    fn <- eval(bquote(function(theta, tn) {
      force(matI)
      theta.part <- .(theta.part)
      if (is.null(tn)) {
        if (is.null(theta)) {
          return(unlist(lapply(
            theta.part,
            function(x) {
              .j <- .k <- 1
              .ret <- logical(length(x))
              for (.i in seq_along(x)) {
                if (.j == .k) {
                  .ret[.i] <- TRUE
                  .k <- .k + 1
                  .j <- 1
                } else {
                  .ret[.i] <- FALSE
                  .j <- .j + 1
                }
              }
              return(.ret)
            }
          )))
        }
      } else if (!is.na(tn)) {
        if (tn == -2L) {
          return(.(ntheta))
        }
      }

      lst <- lapply(
        seq_along(theta.part),
        function(x) {
          mt <- matI[[x]]
          w <- theta.part[[x]]
          new.theta <- theta[w]
          ctn <- as.integer(tn)
          if (is.na(ctn)) {
            return(mt$fn(as.double(new.theta), ctn))
          } else if (ctn == -1L || ctn == 0L) {
            return(mt$fn(as.double(new.theta), ctn))
          } else {
            ## the ctn should refer to the theta relative to
            ## the current submatrix; However this function
            ## has the theta number relative to the whole
            ## matrix.
            if (ctn > 0L) {
              if (ctn > max(w) | ctn < min(w)) {
                mat <- mt$fn(as.double(new.theta), 0L)
                d <- dim(mat)[1]
                return(matrix(rep(0, d * d), d))
              } else {
                ctn <- as.integer(ctn - min(w) + 1)
                return(mt$fn(as.double(new.theta), ctn))
              }
            } else {
              ctn <- as.integer(-ctn - 2)
              if (ctn > max(w) | ctn < min(w)) {
                vec <- mt$fn(as.double(new.theta), -3L)
                d <- length(vec)
                return(rep(0, d))
              } else {
                ctn <- as.integer(-(ctn - min(w) + 1) - 2L)
                return(mt$fn(as.double(new.theta), ctn))
              }
            }
          }
        }
      )
      if (is.na(tn)) {
        return(unlist(lst))
      } else if (tn >= -1) {
        ## FIXME: lotriMat should take unnamed
        ## return(lotri::lotriMat(lst))
        return(as.matrix(Matrix::bdiag(lst)))
      } else {
        return(unlist(lst))
      }
    }))

    ret <- list(
      fmat = mat,
      ini = ini,
      fn = fn
    )
    class(ret) <- "rxSymInvChol"
    return(ret)
  }
}

#' Creates an object for calculating Omega/Omega^-1 and derivatives
#'
#' @param mat Initial Omega matrix
#' @param diag.xform transformation to diagonal elements of OMEGA. or `chol(Omega^-1)`
#' @param create.env -- Create an environment to calculate the inverses. (By default TRUE)
#' @param envir -- Environment to evaluate function, bu default it is the parent frame.
#' @return A rxSymInv object OR a rxSymInv environment
#' @author Matthew L. Fidler
#' @keywords internal
#' @export
rxSymInvCholCreate <- function(mat,
                               diag.xform = c("sqrt", "log", "identity"),
                               create.env = TRUE, envir = parent.frame()) {
  args <- as.list(match.call(expand.dots = TRUE))[-1]
  args <- args[names(args) != "create.env"]
  if (create.env) {
    rxi <- do.call(rxSymInvCreateC_, args, envir = envir)
    ret <- rxSymInvChol(rxi)
    ret$theta <- rxi$ini
    return(ret)
  } else {
    return(do.call(rxSymInvCreateC_, args, envir = envir))
  }
}

#' @export
`$.rxSymInvCholEnv` <- function(obj, arg, exact = TRUE) {
  return(.Call(`_rxode2_rxSymInvCholEnvCalculate`, obj, arg, NULL))
}

#' @export
"$<-.rxSymInvCholEnv" <- function(obj, arg, value) {
  return(.Call(`_rxode2_rxSymInvCholEnvCalculate`, obj, arg, value))
}


## For the inner problem, only Omega^-1 is needed for the
## optimization.  To finalize the likelihood for the individual, you
## need -1/2*log(det(2*pi*Omega)) Which was log.det.OMGAinv.5.  In the
## case of the cholesky decomposition parametrization this becomes
## sum(log(diag)); Therefore for inner problem only calculate these
## two quantities.  For outer problem, or gradient evaluation more is needed.

rxTest({
  rxWithSeed(
    42,
    {
      dgs <- c("sqrt", "log", "identity")
      for (dg in dgs) {
        for (d in seq(1, rxSymInvCholN())) {
          test_that(sprintf("Omega Cholesky %sx%s, %s", d, d, dg), {
            ## Creating covariance matrix
            tmp <- matrix(rnorm(d^2), d, d)
            mcov <- tcrossprod(tmp, tmp)
            v <- suppressMessages(rxSymInvCholCreate(mcov, dg))
            expect_equal(v$ntheta, sum((lower.tri(mcov, TRUE)) * 1))
            expect_equal(length(v$xType), sum((lower.tri(mcov, TRUE)) * 1))
            expect_equal(v$omega, mcov, tolerance = 1e-4)
            expect_equal(v$omegaInv, solve(mcov), tolerance = 1e-4)
            expect_equal(v$chol.omegaInv, chol(solve(mcov)), tolerance = 1e-4)
            expect_equal(v$chol.omega, chol(mcov), tolerance = 1e-4)
            expect_equal(v$log.det.OMGAinv.5, 0.5 * log(det(solve(mcov))), tolerance = 1e-4)
            expect_equal(length(v$d.omegaInv), v$ntheta, tolerance = 1e-4)
            expect_equal(length(v$d.D.omegaInv), v$ntheta, tolerance = 1e-4)
            ## This is to make sure there is no run-time error in calculation
            expect_type(v$tr.28, "double")
            expect_type(v$omega.47, "list")
            if (d != 1) {
              expect_error(v$theta <- 3)
            } else {
              v$theta <- 3 # Should work
            }
          })
        }

        test_that("diagonal indicator give correct values", {
          tmp <- rxSymInvCholCreate(matrix(c(1, 0.9, 0, 0.9, 1, 0, 0, 0, 1), ncol = 3))
          expect_equal(tmp$theta.diag, c(TRUE, FALSE, TRUE, TRUE))
          tmp <- rxSymInvCholCreate(matrix(c(1, 0.9, 0.9, 1), ncol = 2))
          expect_equal(tmp$theta.diag, c(TRUE, FALSE, TRUE))
        })
      }
    }
  )
})

#' @export
#' @rdname model
model.function <- function(x, ..., append=NULL, auto=getOption("rxode2.autoVarPiping", TRUE),
                           cov=NULL, envir=parent.frame()) {
  .modelLines <- .quoteCallInfoLines(match.call(expand.dots = TRUE)[-(1:2)], envir=envir)
  .ret <- rxUiDecompress(rxode2(x))
  if (length(.modelLines) == 0) return(.ret$modelFun)
  .modelHandleModelLines(.modelLines, .ret, modifyIni=FALSE, append=append, auto=auto,
                         cov=cov, envir=envir)
}

#' @export
#' @rdname model
model.rxUi <- function(x, ..., append=NULL, auto=getOption("rxode2.autoVarPiping", TRUE),
                       cov=NULL, envir=parent.frame()) {
  .modelLines <- .quoteCallInfoLines(match.call(expand.dots = TRUE)[-(1:2)], envir=envir)
  .ret <- rxUiDecompress(.copyUi(x)) # copy so (as expected) old UI isn't affected by the call
  if (length(.modelLines) == 0) return(.ret$modelFun)
  .ret <- .modelHandleModelLines(.modelLines, .ret, modifyIni=FALSE, append=append, auto=auto,
                                 cov=cov, envir=envir)
  # need to adjust since the model function was from a rxui object
  .x <- rxUiDecompress(x)
  .ret <- rxUiDecompress(.ret)
  .ret <- .newModelAdjust(.ret, .x)
  .ret <- rxUiCompress(.ret)
  .cls <- setdiff(class(x), class(.ret))
  if (length(.cls) > 0) {
    class(.ret) <- c(.cls, class(.ret))
  }
  .ret
}

#' @export
#' @rdname model
model.rxode2 <- function(x, ..., append=NULL, auto=getOption("rxode2.autoVarPiping", TRUE),
                         cov=NULL, envir=parent.frame()) {
  .modelLines <- .quoteCallInfoLines(match.call(expand.dots = TRUE)[-(1:2)], envir=envir)
  x <- as.function(x)
  .ret <- suppressMessages(rxUiDecompress(rxode2(x)))
  if (length(.modelLines) == 0) return(.ret$modelFun)
  .modelHandleModelLines(.modelLines, .ret, modifyIni=FALSE, append=append, auto=auto,
                         cov=cov, envir=envir)
}

#' @export
#' @rdname model
model.rxModelVars <- model.rxode2

#'  Handle model lines
#'
#' @param modelLines The model lines that are being considered
#' @param rxui The rxode2 UI object
#' @param modifyIni Should the ini({}) be considered
#' @param envir Environment for evaluation
#' @inheritParams model
#' @return New UI
#' @author Matthew L. Fidler
#' @export
.modelHandleModelLines <- function(modelLines, rxui, modifyIni=FALSE, append=NULL,
                                   auto=getOption("rxode2.autoVarPiping", TRUE),
                                   cov=NULL, envir) {
  checkmate::assertLogical(modifyIni, any.missing=FALSE, len=1)
  ## checkmate::assertLogical(append, any.missing=TRUE, len=1)
  checkmate::assertLogical(auto, any.missing=TRUE, len=1)
  checkmate::assertCharacter(cov, pattern="^[.]*[a-zA-Z]+[a-zA-Z0-9._]*$", null.ok=TRUE)
  .varSelect$cov <- cov
  .doAppend <- FALSE
  rxui <- rxUiDecompress(rxui)
  .ll <- length(rxui$lstExpr)
  if (!is.null(.nsEnv$.quoteCallInfoLinesAppend)) {
    if (identical(.nsEnv$.quoteCallInfoLinesAppend, quote(Inf))) {
      .nsEnv$.quoteCallInfoLinesAppend <- NULL
      append <- TRUE
    } else if (identical(.nsEnv$.quoteCallInfoLinesAppend, quote(-Inf))) {
      .nsEnv$.quoteCallInfoLinesAppend <- NULL
      append <- NA
    } else if (identical(.nsEnv$.quoteCallInfoLinesAppend, quote(0))) {
      .nsEnv$.quoteCallInfoLinesAppend <- NULL
      append <- NA
    } else if (checkmate::testIntegerish(.nsEnv$.quoteCallInfoLinesAppend, lower=.ll)) {
      .nsEnv$.quoteCallInfoLinesAppend <- NULL
      append <- TRUE
    }
  }
  if (!is.null(.nsEnv$.quoteCallInfoLinesAppend)) {
    if (checkmate::testIntegerish(.nsEnv$.quoteCallInfoLinesAppend, lower=0, upper=.ll)) {
      .nsEnv$.quoteCallInfoLinesAppend <- .getLhs(rxui$lstExpr[[.nsEnv$.quoteCallInfoLinesAppend]])
    } else if (checkmate::testCharacter(.nsEnv$.quoteCallInfoLinesAppend, len=1, any.missing=FALSE,
                                        min.chars = 1)) {
      .tmp <- try(str2lang(.nsEnv$.quoteCallInfoLinesAppend), silent=TRUE)
      if (inherits(.tmp, "try-error")) {
        stop("'append' must refer to a LHS model line when a character",
             call. = FALSE)
      }
      .nsEnv$.quoteCallInfoLinesAppend <- .tmp
    }
    .w <- which(vapply(seq_len(.ll),
                       function(i) {
                         .lhs <- .getLhs(rxui$lstExpr[[i]])
                         identical(.lhs, .nsEnv$.quoteCallInfoLinesAppend)
                       }, logical(1), USE.NAMES=FALSE))
    if (length(.w) == 0) {
      .var <- deparse1(.nsEnv$.quoteCallInfoLinesAppend)
      .stop <- TRUE
      if (!is.null(.asFunctionEnv$rx)) {
        if (any(.asFunctionEnv$rx$lhs == .var)) {
          .iniDf <- rxui$iniDf
          .tmp <- .asFunctionEnv$rx
          .cls <- class(.asFunctionEnv$rx)
          class(.asFunctionEnv$rx) <- "rxode2"
          .fun <- suppressMessages(as.function(.asFunctionEnv$rx))
          class(.asFunctionEnv$rx) <- .cls
          rxui <- suppressMessages(as.rxUi(.fun))
          ini(rxui) <- .iniDf
          rxui <- rxUiDecompress(rxui)
          .ll <- length(rxui$lstExpr)
          .w <- which(vapply(seq_len(.ll),
                             function(i) {
                               .lhs <- .getLhs(rxui$lstExpr[[i]])
                               identical(.lhs, .nsEnv$.quoteCallInfoLinesAppend)
                             }, logical(1), USE.NAMES=FALSE))
          if (length(.w) == 0) {
            .stop <- TRUE
          } else {
            .stop <- FALSE
          }
        }
      }
      if (.stop) {
        stop("cannot find '", .var, "' in lhs model, cannot append", call.=FALSE)
      }
    }
    .w <- max(.w)
    if (.w == .ll) {
      assign("lstExpr", c(rxui$lstExpr, modelLines), envir=rxui)
    } else {
      assign("lstExpr", c(rxui$lstExpr[seq(1, .w)], modelLines, rxui$lstExpr[seq(.w+1, .ll)]),
             envir=rxui)
    }
    .doAppend <- TRUE
  } else if (is.logical(append) && length(append) == 1L && (is.na(append) || !append)) {
    assign("lstExpr", c(modelLines, rxui$lstExpr), envir=rxui)
    .doAppend <- TRUE
  } else if (isTRUE(append)) {
    assign("lstExpr", c(rxui$lstExpr, modelLines), envir=rxui)
    .doAppend <- TRUE
  }
  if (.doAppend) {
    # in pre-pending or appending, lines are only added
    .lhs <- character(0)
    .rhs <- character(0)
    .lhs0 <- c(rxui$mv0$lhs, rxui$mv0$state, rxui$allCovs, rxui$iniDf$name)
    for (x in modelLines) {
      .isTilde <- .isEndpoint(x)
      if (.isTilde || .isAssignment(x)) {
        .rhs <- unique(c(.getVariablesFromExpression(.getRhs(x), ignorePipe=.isTilde), .rhs))
        .lhs <- unique(c(.getVariablesFromExpression(.getLhs(x)), .lhs))
      }
      .rhs <- setdiff(.rhs, c(.lhs, .lhs0))
      if (isTRUE(auto)) {
        for (v in .rhs) {
          .isCov <- grepl(.varSelect$covariateExceptions, tolower(v))
          .isTheta <- !.isCov && grepl(.varSelect$thetaModelReg, v)
          .isEta <- !.isCov && grepl(.varSelect$etaModelReg, v)
          if (.isTilde || .isTheta || .isEta) {
            .addVariableToIniDf(v, rxui,
                                promote=ifelse(.isTilde,NA,
                                               TRUE))
            .lhs <- c(.lhs, v)
          }
        }
      }
    }
    return(rxUiCompress(rxui$fun()))
  }
  .modifyModelLines(lines = modelLines, rxui = rxui, modifyIni = modifyIni, envir = envir)
  .v <- .getAddedOrRemovedVariablesFromNonErrorLines(rxui)
  if (length(.v$rm) > 0) {
    lapply(.v$rm, function(x) {
      .removeVariableFromIniDf(x, rxui, promote=ifelse(x %in% .v$err, NA, FALSE))
    })
  }
  if (length(.v$new) > 0) {
    lapply(.v$new, function(x) {
      .isErr <- x %in% .v$err
      if (auto || .isErr) {
        .addVariableToIniDf(x, rxui, promote=ifelse(.isErr, NA, FALSE))
      } else if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
        .minfo(paste0("add covariate {.code ", x, "}"))
      }
    })
  }
  rxUiCompress(rxui$fun())
}

# Determine if the input is an endpoint by being 3 long and the call part being
# a tilde
.isEndpoint <- function(expr) {
  .matchesLangTemplate(expr, str2lang(". ~ ."))
}

# Determine if the input is an assignment by being 3 long and the call part
# being either the left arrow, right arrow, or equal sign
.isAssignment <- function(expr) {
  .matchesLangTemplate(expr, str2lang(". <- .")) ||
    .matchesLangTemplate(expr, str2lang(". = ."))
}

.isTildeExpr <- function(expr) {
  .matchesLangTemplate(expr, str2lang("~ ."))
}

.isIniDropExpression <- function(expr) {
  .matchesLangTemplate(expr, str2lang("- ."))
}

# get the left hand side of an assignment or endpoint; returns NULL if the input
# is not an assignment or endpoint
.getLhs <- function(expr) {
  ret <- NULL
  if (.isAssignment(expr) || .isEndpoint(expr)) {
    ret <- expr[[2]]
  }
  ret
}

.getRhs <- function(expr, ignorePipe=FALSE) {
  ret <- NULL
  if (.isAssignment(expr) || .isEndpoint(expr)) {
    ret <- expr[[3]]
  }
  ret
}

.getModelLineEquivalentLhsExpressionDropEndpoint <- function(expr) {
  ret <- NULL
  if (.isEndpoint(expr)) {
    lhs <- .getLhs(expr)
    if (.matchesLangTemplate(lhs, str2lang("-.")) ||
          .matchesLangTemplate(lhs, str2lang(". <- NULL"))) {
      # If it is a drop expression with a minus sign, grab the non-minus part
      ret <- lhs[[2]]
    }
  }
  ret
}

.getModelLineEquivalentLhsExpressionDropDdt <- function(expr) {
  .expr3 <- NULL
  if (.matchesLangTemplate(x = expr, template = str2lang("-d/dt(.name)")) ||
        .matchesLangTemplate(x = expr, template = str2lang("d/dt(.name) <- NULL"))) {
    .expr3 <- expr
    # remove the minus sign from the numerator
    .expr3[[2]] <- .expr3[[2]][[2]]
  }
  .expr3
}

#' This gives a equivalent left handed expression
#'
#' @param expr This exchanges expressions ie f() to F()
#' @return NULL if there isn't an equivalent expression, or the
#'   equivalent expression.
#' @author Matthew L. Fidler
#' @noRd
.getModelLineEquivalentLhsExpression <- function(expr) {
  .expr3 <- .getModelLineEquivalentLhsExpressionDropDdt(expr)
  if (is.null(.expr3)) .expr3 <- .getModelLineEquivalentLhsExpressionDropEndpoint(expr)
  if (length(expr) == 2L) {
    .expr1 <- expr[[1]]
    .expr2 <- expr[[2]]
    if (identical(.expr1, quote(`f`))) {
      .expr3 <- eval(parse(text=paste0("quote(F(",as.character(.expr2),"))")))
    }
    if (identical(.expr1, quote(`F`))) {
      .expr3 <- eval(parse(text=paste0("quote(f(",as.character(.expr2),"))")))
    }
    if (identical(.expr1, quote(`lag`))) {
      .expr3 <- eval(parse(text=paste0("quote(alag(",as.character(.expr2),"))")))
    }
    if (identical(.expr1, quote(`alag`))) {
      .expr3 <- eval(parse(text=paste0("quote(lag(",as.character(.expr2),"))")))
    }
    if (identical(.expr1, quote(`-`))) {
      .expr3 <- .expr2
    }
  }
 .expr3
}

#' Get the model line number from the expression
#'
#' @param expr The Left handed expression to find
#' @param altExpr The alternate expression to find (or `NULL`), this
#'   is for expressions that are equivalent like `f(center)` and
#'   `F(center)`.
#' @param useErrorLine This is a boolean indicating if error lines
#'   should be considered (`TRUE`).
#' @param errLines This is an integer vector of the lines where an
#'   error is defined in the model.
#' @param origLines This is a list of lines in the `model({})` block
#'   of the equation.
#' @param rxui the UI model
#' @param returnAllLines Return all line numbers for the lhs, even
#'   when there are duplicates. (default `FALSE`)
#' @return For duplicated lines: `NULL` for duplicated lines (when
#'   `returnAllLines` is FALSE) or all the line numbers (when
#'   `returnAllLines` is TRUE).
#'
#' For non-duplicated lines return  the line number (if
#'   it is found) and `NA` if it is not found.
#'
#' @author Matthew L. Fidler
#' @noRd
.getModelineFromExpressionsAndOriginalLines <- function(expr, altExpr, useErrorLine,
                                                        errLines, origLines, rxui,
                                                        returnAllLines=FALSE) {
  .ret <- NA_integer_
  .multipleEndpointModel <- length(errLines) != 1L
  for (.i in seq_along(origLines)) {
    .isErrorLine <- .i %in% errLines
    if (returnAllLines ||
          (useErrorLine && .isErrorLine) ||
          (!useErrorLine && !.isErrorLine)) {
      .expr <- origLines[[.i]]
      if (identical(.expr[[2]], expr)) {
        if (is.na(.ret)) {
          .ret <- .i
        } else if (returnAllLines) {
          .ret <- c(.ret, .i)
        } else {
          return(NULL)
        }
      } else if (!is.null(altExpr)) {
        if (identical(.expr[[2]], altExpr)) {
          if (is.na(.ret)) {
            .ret <- .i
          } else if (returnAllLines) {
            .ret <- c(.ret, .i)
          } else {
            return(NULL)
          }
        }
      } else if (useErrorLine  && !.multipleEndpointModel) {
        if (is.na(.ret)) {
          if (.isNormOrTErrorExpression(.expr)) {
            # Make sure the lhs is included in the model prediction
            .var <- deparse1(expr)
            .modelVars <- c(rxui$mv0$lhs, rxui$mv0$state)
            if (!(.var %in% .modelVars)) {
              stop("the variable '", .var, "' must be in the defined the model for piping this: '",deparse(.expr), "'",
                   call.=FALSE)
            }
          }
          if (!identical(.expr[[2]], expr)) {
            warning("with single endpoint model prediction '", deparse1(.expr[[2]]), "' is changed to '", expr, "'",
                    call.=FALSE)
          }
          .ret <- .i
        } else {
          return(NULL)
        }
      }
    }
  }
  .ret
}
#' Get the negative model line for ode based on ode property
#'
#' @param expr Left handed expression
#' @param origLines List of original lines (expressions)
#' @param errorLine Consider error lines
#' @return `NA` if the ODE isn't found. Negative number for the lowest
#'   line of the ODE compartment.
#' @author Matthew L. Fidler
#' @noRd
.getNegativeModelLineForDiffFromProperty <- function(expr, origLines, errorLine) {
  .ret <- NA_integer_
  if (!errorLine && length(expr) == 2L) {
    .expr1 <- expr[[1]]
    .expr2 <- expr[[2]]
    if (is.numeric(.expr2)) {
      .state <- as.character(.expr1)
    } else {
      .state <- as.character(.expr2)
    }
    if (is.numeric(.expr2) ||
          identical(.expr1, quote(`f`)) ||
          identical(.expr1, quote(`F`)) ||
          identical(.expr1, quote(`alag`)) ||
          identical(.expr1, quote(`lag`)) ||
          identical(.expr1, quote(`rate`)) ||
          identical(.expr1, quote(`dur`))) {
      .expr3 <- eval(parse(text=paste0("quote(d/dt(",.state,"))")))
      for (.i in seq_along(origLines)) {
        .expr <- origLines[[.i]]
        if (identical(.expr[[2]], .expr3)) {
          if (is.na(.ret)) {
            .ret <- -.i
          } else {
            .ret <- min(-.i, .ret)
          }
        }
      }
    }
  }
  .ret
}

#' Find the line of the original expression
#'
#' @param expr Expression
#'
#' @param rxui rxode2 UI
#'
#' @param errorLine When TRUE, Should the error line be considered, or
#'   consider the non-error line
#'
#' @param returnAllLines A boolean to determine if all line numbers
#'   should be returned, by default `FALSE`
#'
#' @return The return can be:
#'
#'  - Line number of from the original model, if the model matches one
#'  line uniquely (or an integer vector of all the lines matching when `returnAllLines=TRUE`)
#'
#'  - Negative number; This indicates the model property (ie
#'  `f(depot)`) is not found in the model, but the differential
#'  equation `d/dt(depot)` is defined uniquely i the model and is
#'  found at the -(line number) returned.
#'
#'  - `NA` which means the line (or related line) was not found in the
#'  model
#'
#'  - `NULL` which means a line is duplicated in the model
#'
#' @author Matthew L. Fidler
#'
#' @noRd
.getModelLineFromExpression <- function(lhsExpr, rxui, errorLine=FALSE, returnAllLines=FALSE) {
  .origLines <- rxui$lstExpr
  .errLines <- rxui$predDf$line
  .expr3 <- .getModelLineEquivalentLhsExpression(lhsExpr)
  .ret <- .getModelineFromExpressionsAndOriginalLines(lhsExpr, .expr3, errorLine, .errLines, .origLines, rxui, returnAllLines)
  if (is.null(.ret)) {
    return(NULL)
  } else if (length(.ret) > 1) {
    return(.ret)
  } else if (!is.na(.ret)) {
    return(.ret)
  }
  .getNegativeModelLineForDiffFromProperty(lhsExpr, .origLines, errorLine)
}


#' @export
rxUiGet.mvFromExpression <- function(x, ...) {
  .x <- x[[1]]
  .exact <- x[[2]]
  if (is.null(.x$predDf)) {
    eval(call("rxModelVars",as.call(c(list(quote(`{`)), .x$lstExpr))))
  } else {
    eval(call("rxModelVars",as.call(c(list(quote(`{`)), .x$lstExpr[-.x$predDf$line]))))
  }
}
attr(rxUiGet.mvFromExpression, "desc") <- "Calculate model variables from stored (possibly changed) expression"

#' Describe if the piping expression is a drop expression
#'
#' @param line expression for line/expression
#' @return `TRUE` if this is a drop expression like `|> model(-v)`, otherwise `FALSE`
#' @author Matthew L. Fidler
#' @noRd
.isDropExpression <- function(line) {
  if (!is.null(.getModelLineEquivalentLhsExpressionDropDdt(line))) {
    return(TRUE)
  }
  if (!is.null(.getModelLineEquivalentLhsExpressionDropEndpoint(line))) {
    return(TRUE)
  }
  if (length(line) == 2L) {
    if (identical(line[[1]], quote(`-`))) {
      if (is.name(line[[2]])) {
        return(TRUE)
      } else if (is.call(line[[2]]) && length(line[[2]]) == 2L) {
        if (is.name(line[[2]][[2]]) &&
              as.character(line[[2]][[1]]) %in% c("F", "f", "alag", "lag", "dur", "rate")) {
          return(TRUE)
        } else if (identical(line[[2]][[2]], 0)) {
          return(TRUE)
        }
      }
    }
  }
  FALSE
}

.getAdditionalDropLines <- function(line, rxui, isErr, isDrop) {
  .dropCmt <- .getModelLineEquivalentLhsExpressionDropDdt(line)
  if (!is.null(.dropCmt)) {
    .state <- .dropCmt[[3]][[2]]
    .types <- list(quote(`f`), quote(`F`), quote(`alag`),
                   quote(`lag`), quote(`dur`), quote(`rate`), NULL)
    .types <- lapply(seq_along(.types),
                     function(i) {
                       .cur <- .types[[i]]
                       if (is.null(.cur)) {
                         as.call(list(.state, 0))
                       } else {
                         as.call(list(.cur, .state))
                       }
                     })
    return(unique(do.call("c", lapply(seq_along(.types), function(i) {
      .v <- .getModelLineFromExpression(.types[[i]], rxui, isErr, isDrop)
      if (is.na(.v[1])) {
        return(NULL)
      } else if (.v[1] < 0) {
        return(NULL)
      } else {
        return(.v)
      }
    }))))
  }
  NULL
}
#' This checks the different types of drop assignments
#'
#'
#' @param prefix The prefix of the drop assignment
#' @param line The line expression to check
#' @return logical to say if this matches the prefix
#' @author Matthew L. Fidler
#' @noRd
.isDropNullType <- function(prefix, line) {
  .e1 <- str2lang(paste0(prefix, " <- NULL"))
  .e2 <- str2lang(paste0(prefix, " = NULL"))
  .e3 <- str2lang(paste0(prefix, " ~ NULL"))
  if (.matchesLangTemplate(line, .e1)) return(TRUE)
  if (.matchesLangTemplate(line, .e3)) return(TRUE)
  if (.matchesLangTemplate(line, .e2)) return(TRUE)
  FALSE
}
#' This changes NULL assignment line to a -drop line
#'
#' @param line Line to change if necessary
#' @return Drop line normalized to be `-line` instead of  `line <- NULL`
#' @author Matthew L. Fidler
#' @noRd
.changeDropNullLine <- function(line) {
  if (.isDropNullType("d/dt(.name)", line)) {
    .ret <- line[[2]]
    .ret[[2]] <- as.call(list(quote(`-`), .ret[[2]]))
    return(.ret)
  }
  if (.isDropNullType(".name", line) ||
        .isDropNullType("lag(.name)", line) ||
        .isDropNullType("alag(.name)", line) ||
        .isDropNullType("f(.name)", line) ||
        .isDropNullType("F(.name)", line) ||
        .isDropNullType("rate(.name)", line) ||
        .isDropNullType("dur(.name)", line) ||
        .isDropNullType(".name(0)", line)
        ) {
    return(as.call(list(quote(`-`), line[[2]])))
  }
  line
}

#'  Modify the error lines/expression
#'
#' @param lines quoted lines to modify
#' @param rxui UI to save information
#' @param modifyIni Modify ini({}) types of calls
#' @param envir Environment
#' @return Nothing, called for the side effects
#' @author Matthew L. Fidler
#' @noRd
.modifyModelLines <- function(lines, rxui, modifyIni=FALSE, envir) {
  .err <- NULL
  .env <- environment()
  lapply(lines, function(line) {
    line <- .changeDropNullLine(line)
    if (modifyIni && .isQuotedLineRhsModifiesEstimates(line, rxui)) {
      .iniHandleFixOrUnfix(line, rxui, envir=envir)
    } else {
      .isErr  <- .isErrorExpression(line)
      .isDrop <- .isDropExpression(line)
      if (.isDrop && .isErr) {
        .ret <- .getModelLineFromExpression(.getModelLineEquivalentLhsExpression(line), rxui, .isErr, FALSE)
      } else if (.isDrop) {
        .ret <- .getModelLineFromExpression(.getModelLineEquivalentLhsExpression(line), rxui, .isErr, .isDrop)
        .ret <- c(.ret, .getAdditionalDropLines(line, rxui, .isErr, .isDrop))
      } else {
        .ret <- .getModelLineFromExpression(.getLhs(line), rxui, .isErr, .isDrop)
      }
      if (length(.ret)  == 1) {
        if (.isErr && is.na(.ret)) {
          stop("the error '", deparse1(line[[2]]), "' is not in the multiple-endpoint model and cannot be modified",
               call.=FALSE)
        }
      }
      if (is.null(.ret)) {
        assign(".err",
               c(.err, paste0("the lhs expression '", deparse1(line[[2]]), "' is duplicated in the model and cannot be modified by piping")),
               envir=.env)
      } else if (is.na(.ret[1])) {
          assign(".err",
                 c(.err, paste0("the lhs expression '", deparse1(line[[2]]), "' is not in the model and cannot be modified by piping")),
                 envir=.env)
      } else if (all(.ret > 0)) {
        if (.isDrop) {
          .lstExpr <- get("lstExpr", rxui)
          .predDf <- get("predDf", rxui)
          .ret0 <- sort(.ret, decreasing=TRUE)
          if (length(.predDf$cond) == 1L && any(.ret0 %in% .predDf$line)) {
            .predDf <- NULL
            .lstExpr <- .lstExpr[-.ret]
          } else {
            for (.i in .ret0) {
              ## Drop lines that match
              .w <- which(.predDf$line == .i)
              if (length(.w) > 0) {
                .predDf <- .predDf[-.w,, drop = FALSE]
              }
              # renumber lines greater
              .w <- which(.predDf$line > .i)
              if (length(.w) > 0) {
                .predDf$line[.w] <- .predDf$line[.w] - 1L
              }
              .lstExpr <- .lstExpr[-.i]
            }
          }
          assign("predDf", .predDf, rxui)
          assign("lstExpr", .lstExpr, rxui)
          assign(".recalculate", TRUE, rxui)
        } else {
          if (.isErr) {
            .throwIfInvalidTilde(line)
          }
          .lstExpr <- get("lstExpr", rxui)
          .lstExpr[[.ret]] <- line
          assign("lstExpr", .lstExpr, rxui)
          assign(".recalculate", TRUE, rxui)
        }
      } else {
        if (.isErr) {
          .throwIfInvalidTilde(line)
        }
        .lstExpr <- get("lstExpr", rxui)
        .lstExpr[[length(.lstExpr) + 1]] <- line
        assign("lstExpr", .lstExpr, rxui)
        assign(".recalculate", TRUE, rxui)
      }
    }
    NULL
  })
  if (!is.null(.err)) {
    stop(paste(.err, collapse="\n"), call.=FALSE)
  }
}

#' Get the Variables from the expression
#'
#' @param x Expression
#' @return Character vector of variables
#' @author Matthew L. Fidler
#' @noRd
.getVariablesFromExpression <- function(x, ignorePipe=FALSE) {
  if (is.atomic(x)) {
    character()
  } else if (is.name(x)) {
    return(as.character(x))
  } else if (.matchesLangTemplate(x, str2lang("d/dt(.name)"))) {
    # ODE expressions only pull out the state name and not "d" or "dt"
    return(as.character(x[[3]][[2]]))
  } else {
    if (is.call(x)) {
      if (ignorePipe && identical(x[[1]], quote(`|`))) {
        return(.getVariablesFromExpression(x[[2]]))
      } else {
        x1 <- x[-1]
      }
    } else {
      x1 <- x
    }
    unique(unlist(lapply(x1, .getVariablesFromExpression, ignorePipe=ignorePipe)))
  }
}

#' @export
rxUiGet.errParams <- function(x, ...) {
  .x <- x[[1]]
  .exact <- x[[2]]
  unlist(lapply(.x$lstExpr[.x$predDf$line], function(x) {
    .getVariablesFromExpression(x[[3]])
  }))
}
attr(rxUiGet.errParams, "desc") <- "Get the error-associated variables"
attr(rxUiGet.errParams, "rstudio") <- "character"

#' Get the added or removed variables
#'
#' @param rxui This is the rxode2 UI object
#'
#' @return A list with the removed and added error objects
#'
#' @author Matthew L. Fidler
#'
#' @noRd
.getAddedOrRemovedVariablesFromNonErrorLines <- function(rxui) {
  .old <- rxui$mv0$params
  .new <- rxui$mvFromExpression$params
  .both <- intersect(.old, .new)
  .rm1 <- setdiff(.old, .both)
  .new1 <- setdiff(.new, .both)

  .old <- rxui$errParams0
  .err <- rxui$errParams
  .new <- rxui$errParams
  .both <- intersect(.old, .new)
  .rm2 <- setdiff(.old, .both)
  .new2 <- setdiff(.new, .both)

  .new <- c(.new1, .new2)
  .new <- setdiff(.new, rxui$mv0$lhs)

  .rm <- c(.rm1, .rm2)

  .rm <- setdiff(.rm, .err)

  list(rm=.rm, new=.new, err=.err)
}

#' Remove a single variable from the initialization data frame
#'
#' @param var Variable that is removed
#' @param rxui UI function where the initial estimate data frame is modified
#' @inheritParams .addVariableToIniDf
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @noRd
.removeVariableFromIniDf <- function(var, rxui, promote=FALSE) {
  .iniDf <- rxui$iniDf
  .w <- which(.iniDf$name == var)
  if (length(.w) == 1L) {
    .neta <- .iniDf$neta1[.w]
    .iniDf <- .iniDf[-.w, ]
    if (!is.na(.neta)) {
      # Here we remove any assocaited covariance terms that remain
      .w1 <- which(.iniDf$neta1 == .neta)
      if (length(.w1) > 0) .iniDf <- .iniDf[-.w1, ]
      .w1 <- which(.iniDf$neta2 == .neta)
      if (length(.w1) > 0) .iniDf <- .iniDf[-.w1, ]
      if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
        .mwarn(paste0("remove between subject variability {.code ", var, "}"))
      }
    } else if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
      if (is.na(promote)) {
        .mwarn(paste0("remove residual parameter {.code ", var, "}"))
      } else {
        .mwarn(paste0("remove population parameter {.code ", var, "}"))
      }
    }
    assign("iniDf", .iniDf, rxui)
  }
  invisible()
}

.varSelect <- new.env(parent=emptyenv())
#' Set the variables for the model piping automatic covarite selection
#'
#' @param thetamodelVars This is the prefixes for the theta model
#'   variables in a regular expression
#' @param covariateExceptions This is a regular expression of
#'   covariates that should always be covariates
#' @param etaParts This is the list of eta prefixes/post-fixes that
#'   identify a variable as a between subject variability
#' @return Nothing, called for side effects
#' @export
#' @author Matthew L. Fidler
#' @details
#'
#' This is called once at startup to set the defaults, though you can
#' change this if you wish so that piping can work differently for
#' your individual setup
#'
rxSetPipingAuto <- function(thetamodelVars=rex::rex(or("tv", "t", "pop", "POP", "Pop",
                                                     "TV", "T", "cov", "err", "eff")),
                          covariateExceptions = rex::rex(start, or("wt", "sex", "crcl", "kout"), end),
                          etaParts=c("eta", "ETA", "Eta", "ppv", "PPV", "Ppv", "iiv", "Iiv",
                                     "bsv", "Bsv", "BSV","bpv", "Bpv", "BPV", "psv", "PSV",
                                     "Psv")
                          ) {
  .varSelect$thetamodelVars <- thetamodelVars
  .varSelect$thetaModelReg <- rex::rex(or(
    group(start, thetamodelVars),
    group(thetamodelVars, end)))
  .varSelect$covariateExceptions <- covariateExceptions
  .varSelect$etaParts <- etaParts
  .varSelect$etaModelReg <- rex::rex(or(group(start, or(etaParts)),
                                        group(or(etaParts), end)))
  .varSelect$covariateNames <- NULL
  .varSelect$cov <- NULL
}

rxSetPipingAuto()

.rxIniDfTemplate <-
  data.frame(
    ntheta = NA_integer_,
    neta1 = NA_real_,
    neta2 = NA_real_,
    name = NA_character_,
    lower = -Inf,
    est = NA_real_,
    upper = Inf,
    fix = FALSE,
    label = NA_character_,
    backTransform = NA_character_,
    condition = NA_character_,
    err = NA_character_
  )

#' Assign covariates for piping
#'
#' @param covariates NULL (for no covariates), or the list of
#'   covariates. nlmixr uses this function to set covariates if you
#'   pipe from a nlmixr fit.
#'
#' @return Nothing, called for side effects
#'
#' @author Matthew L. Fidler
#'
#' @export
#'
#' @examples
#'
#' # First set the name of known covariates
#' # Note this is case sensitive
#'
#' rxSetCovariateNamesForPiping(c("WT","HT", "TC"))
#'
#' one.compartment <- function() {
#'  ini({
#'    tka <- 0.45 ; label("Log Ka")
#'    tcl <- 1 ; label("Log Cl")
#'    tv <- 3.45 ; label("Log V")
#'    eta.ka ~ 0.6
#'    eta.cl ~ 0.3
#'    eta.v ~ 0.1
#'    add.err <- 0.7
#'  })
#'  model({
#'    ka <- exp(tka + eta.ka)
#'    cl <- exp(tcl + eta.cl)
#'    v <- exp(tv + eta.v)
#'    d / dt(depot) <- -ka * depot
#'    d/dt(depot) <- -ka * depot
#'    d / dt(center) <- ka * depot - cl / v * center
#'    cp <- center / v
#'    cp ~ add(add.err)
#'  })
#' }
#'
#' # now TC is detected as a covariate instead of a population parameter
#'
#' one.compartment |>
#'   model({ka <- exp(tka + eta.ka + TC * cov_C)})
#'
#' # You can turn it off by simply adding it back
#'
#' rxSetCovariateNamesForPiping()
#'
#' one.compartment |>
#'   model({ka <- exp(tka + eta.ka + TC * cov_C)})
#'
#' # The covariates you set with `rxSetCovariateNamesForPiping()`
#' # are turned off every time you solve (or fit in nlmixr)
#'
#' @export
rxSetCovariateNamesForPiping <- function(covariates=NULL) {
  if (!is.null(covariates)) {
    checkmate::assertCharacter(covariates, any.missing=FALSE, unique=TRUE)
  }
  .varSelect$covariateNames <-  covariates
}
#' Add a single variable from the initialization data frame
#'
#' @param var Variable that is added
#' @param rxui UI function where the initial estimate data frame is
#'   modified
#' @param toEta This boolean determines if it should be added to the
#'   etas (`TRUE`), thetas (`FALSE`) or determined by `.etaModelReg`
#'   (`NA`)
#' @param value This is the value to assign to the ini block
#' @param promote This boolean determines if the parameter was
#'   promoted from an initial covariate parameter (`TRUE`).  If so it
#'   changes the verbose message to the user.  If not, (`FALSE`) it
#'   will only add if it is a covariate.  If `NA` it will assume that
#'   the parameter is an error term. When `promote` is `TRUE`, it will
#'   also remove the parameter from the `$allCovs`.
#' @return Nothing, called for side effects
#' @author Matthew L. Fidler
#' @noRd
.addVariableToIniDf <- function(var, rxui, toEta=NA, value=1, promote=FALSE) {
  if (var %in% c("pi", "M_E", "M_E", "E", "M_PI", "M_PI_2",
                 "M_PI_4", "M_1_PI", "M_2_PI", "M_2PI", "M_SQRT_PI",
                 "M_2_SQRTPI", "M_1_SQRT_2PI", "M_SQRT2", "M_SQRT_3",
                 "M_SQRT_32", "M_SQRT_2dPI", "M_LN_SQRT_PI",
                 "M_LN_SQRT_2PI", "M_LN_SQRT_PId2", "M_LOG10_2",
                 "M_LOG2E", "M_LOG10E", "M_LN2", "M_LN10")) {
    return(invisible())
  }
  if (!is.null(.varSelect$cov)) {
    if (var %in% .varSelect$cov) {
      if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
        .minfo(paste0("add covariate {.code ", var, "} (as requested by cov option)"))
      }
      return(invisible())
    }
  }
  .mv <- rxModelVars(rxui)
  .ini <- .mv$ini
  .ini <- .ini[which(!is.na(.ini))]
  if (var %in% names(.ini)) {
    return(invisible())
  }
  .iniDf <- rxui$iniDf
  .isEta <- TRUE
  checkmate::assertLogical(toEta, len=1)
  if (is.na(promote)) {
    checkmate::assertNumeric(value, len=1, any.missing=FALSE)
  } else if (promote) {
    if (is.na(toEta)) {
      checkmate::assertNumeric(value, len=1, any.missing=FALSE)
    } else if (toEta) {
      checkmate::assertNumeric(value, len=1, any.missing=TRUE)
    } else {
      if (length(value) %in% c(2, 3)) {
        # This is for promotion to parameter, only.  Additional bound setting
        # happens elsewhere.
        value <- value[2]
      }
      checkmate::assertNumeric(value, len=1, any.missing=FALSE)
    }
  } else {
    checkmate::assertNumeric(value, len=1, any.missing=FALSE)
  }
  checkmate::assertLogical(promote, len=1)
  if (is.na(toEta)) {
    .isEta <- (regexpr(.varSelect$etaModelReg, var)  != -1)
  } else  {
    .isEta <- toEta
  }
  if (.isEta) {
    if (all(is.na(.iniDf$neta1))) {
      .eta <- 1
    } else {
      .eta <- max(.iniDf$neta1, na.rm=TRUE) + 1
    }
    .extra <- .rxIniDfTemplate
    .extra$est <- value
    .extra$neta1 <- .eta
    .extra$neta2 <- .eta
    .extra$name <- var
    .extra$condition <- "id"
    if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
      if (is.na(promote)) {
      } else if (promote) {
        if (is.na(value))  {
          value <- 1
          .minfo(paste0("promote {.code ", var, "} to between subject variability"))
        } else {
          .minfo(paste0("promote {.code ", var, "} to between subject variability with initial estimate {.number ", value, "}"))
        }
        .cov <- get("covariates", envir=rxui)
        .cov <- .cov[.cov != var]
        assign("covariates", .cov, envir=rxui)
      } else {
        .minfo(paste0("add between subject variability {.code ", var, "} and set estimate to {.number ", value, "}"))
      }
    }
    assign("iniDf", rbind(.iniDf, .extra), envir=rxui)
  } else {
    if (is.na(promote)) {
    } else if (!promote) {
      if (regexpr(.varSelect$covariateExceptions, tolower(var)) != -1 ||
            regexpr(.varSelect$thetaModelReg, var, perl=TRUE) == -1) {
        if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
          .minfo(paste0("add covariate {.code ", var, "}"))
        }
        return(invisible())
      }
      if (!is.null(.varSelect$covariateNames)) {
        if (var %in% .varSelect$covariateNames) {
          if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
            .minfo(paste0("add covariate {.code ", var, "} (known covariate)"))
          }
          return(invisible())
        }
      }
    }
    if (all(is.na(.iniDf$ntheta))) {
      .theta <- 1
    } else {
      .theta <- max(.iniDf$ntheta, na.rm=TRUE) + 1
    }
    .extra <- .rxIniDfTemplate
    .extra$est <- value
    .extra$ntheta <- .theta
    .extra$name <- var
    if (isTRUE(getOption("rxode2.verbose.pipe", TRUE))) {
      if (is.na(promote)) {
        .minfo(paste0("add residual parameter {.code ", var, "} and set estimate to {.number ", value, "}"))
      } else if (promote) {
        .minfo(paste0("promote {.code ", var, "} to population parameter with initial estimate {.number ", value, "}"))
        # need to reassess model for mu2 enhancement
        assign("iniDf", rbind(.iniDf, .extra), envir=rxui)
        rxui2 <- rxui
        model(rxui2) <- rxui$lstExpr
        rxui2 <- rxUiDecompress(rxui2)
        for (i in ls(envir=rxui2, all.names=TRUE)) {
          assign(i, get(i, envir=rxui2), envir=rxui)
        }
        return(invisible())
      } else {
        .minfo(paste0("add population parameter {.code ", var, "} and set estimate to {.number ", value, "}"))
      }
    }
    assign("iniDf", rbind(.iniDf, .extra), envir=rxui)
  }
  invisible()
}

rxTest({
  # Test pipeline style of interacting with rxode2

  mod <- rxode2({
    eff(0) <- 1
    C2 <- centr / V2
    C3 <- peri / V3
    CL <- TCl * exp(eta.Cl) ## This is coded as a variable in the model
    d/dt(depot) <- -KA * depot
    d/dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
    d/dt(peri) <- Q * C2 - Q * C3
    d/dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
  })

  fun <- function(type) {
    rxWithSeed(
      42,
      {
        p1 <- mod |>
          rxParams(
            params = c(
              KA = 2.94E-01, TCl = 1.86E+01, V2 = 4.02E+01, # central
              Q = 1.05E+01, V3 = 2.97E+02, # peripheral
              Kin = 1, Kout = 1, EC50 = 200
            ),
            inits = c(eff = 1),
            omega = lotri(eta.Cl ~ 0.4^2)
          ) |>
          et(amountUnits = "mg", timeUnits = "hours") |>
          et(amt = 10000, cmt = 2, ii = 12, until = 48) |>
          et(seq(0, 48, length.out = 100))
        if (type == "rxSolve") {
          p1 <- p1 |>
            rxSolve(nSub = 30)
        } else if (type == "solve") {
          p1 <- p1 |>
            solve(nSub = 30)
        } else if (type == "simulate") {
          p1 <- p1 |>
            simulate(nSub = 30)
        } else if (type == "predict") {
          p1 <- p1 |>
            predict(nSub = 30)
        }
      }
    )
    ##
    rxWithSeed(
      42,
      {
        p2 <- mod |>
          et(amountUnits = "mg", timeUnits = "hours") |>
          et(amt = 10000, cmt = 2, ii = 12, until = 48) |>
          et(seq(0, 48, length.out = 100)) |>
          rxParams(
            params = c(
              KA = 2.94E-01, TCl = 1.86E+01, V2 = 4.02E+01, # central
              Q = 1.05E+01, V3 = 2.97E+02, # peripheral
              Kin = 1, Kout = 1, EC50 = 200
            ),
            inits = c(eff = 1),
            omega = lotri(eta.Cl ~ 0.4^2)
          )
        if (type == "rxSolve") {
          p2 <- p2 |>
            rxSolve(nSub = 30)
        } else if (type == "solve") {
          p2 <- p2 |>
            solve(nSub = 30)
        } else if (type == "simulate") {
          p2 <- p2 |>
            simulate(nSub = 30)
        } else if (type == "predict") {
          p2 <- p2 |>
            predict(nSub = 30)
        }
      }
    )
    test_that(sprintf(
      "mod > et > rxParams > %s == mod > rxParams > et > %s",
      type, type
    ), {
      expect_equal(as.data.frame(p1), as.data.frame(p2))
    })
  }

  fun("rxSolve")
  fun("solve")
  fun("simulate")
  fun("predict")

  p1 <- mod |>
    rxParams(
      params = c(
        KA = 2.94E-01, TCl = 1.86E+01, V2 = 4.02E+01, # central
        Q = 1.05E+01, V3 = 2.97E+02, # peripheral
        Kin = 1, Kout = 1, EC50 = 200
      ),
      inits = c(eff = 1),
      omega = lotri(eta.Cl ~ 0.4^2)
    ) |>
    et(amountUnits = "mg", timeUnits = "hours") |>
    et(amt = 10000, cmt = 2, ii = 12, until = 48) |>
    et(seq(0, 48, length.out = 100)) |>
    rxSolve(nSub = 4)

  ps1 <- p1 |>
    rxParams(inits = c(eff = 2), dfSub = 4) |>
    rxSolve(nSub = 6, nStud = 3)

  test_that("can update parameters from solve", {
    expect_true(is(ps1, "rxSolve"))
    expect_false(is.null(ps1$omegaList))
  })

  ps2 <- p1 |>
    et(amt = 10000, cmt = 2, ii = 24, until = 48) |>
    et(seq(0, 48, length.out = 100)) |>
    rxSolve(nSub = 4)

  test_that("Can update event table in pipline solve", {
    expect_true(is(ps1, "rxSolve"))
  })
})

test_that("drop linCmt() endpoint (#355)", {
  ui <- function() {
    ini({
      tcl <- 1
      tvc <- 1
      addSd <- 1
    })
    model({
      cl <- tcl
      vc <- tvc
      linCmt() ~ add(addSd)
    })
  }
  suppressMessages(
    expect_error(newmod <- model(ui, -linCmt()~.), NA)
  )
  expect_equal(
    newmod$lstExpr,
    list(
      str2lang("cl <- tcl"),
      str2lang("vc <- tvc")
    )
  )
})

test_that("Compartment should not be added to ini (rxode2#336)", {
  uifun <- function() {
    ini({
      a <- 2
      propSd <- c(0, 0.3)
    })
    model({
      d/dt(tumor) <- - a*tumor
      tumor ~ prop(propSd)
    })
  }

  rx_orig <- rxode2(uifun)
  rx_mod <-
    model(
      rx_orig,
      d/dt(transit2) <- (tumor - transit2)/a,
      append = TRUE
    )
  expect_equal(rx_mod$state, c("tumor", "transit2"))
  expect_equal(rx_mod$ini$est, c(2, 0.3))
  expect_equal(rx_mod$ini$name, c("a", "propSd"))
})

# Tests of individual functions ####

test_that(".getModelLineEquivalentLhsExpressionDropDdt", {
  expect_null(.getModelLineEquivalentLhsExpressionDropDdt(str2lang("d/dt(a)")))
  expect_equal(
    .getModelLineEquivalentLhsExpressionDropDdt(str2lang("-d/dt(a)")),
    str2lang("d/dt(a)")
  )
})

test_that(".getModelLineEquivalentLhsExpressionDropEndpoint", {
  # drop a normal endpoint
  expect_equal(
    .getModelLineEquivalentLhsExpressionDropEndpoint(str2lang("-a~.")),
    str2lang("a")
  )
  # don't drop when not requested (only negation matches)
  expect_null(
    .getModelLineEquivalentLhsExpressionDropEndpoint(str2lang("a~."))
  )
  # don't drop assignment (only endpoints are matched)
  expect_null(
    .getModelLineEquivalentLhsExpressionDropEndpoint(str2lang("-a <- ."))
  )
  # don't drop a name
  expect_null(
    .getModelLineEquivalentLhsExpressionDropEndpoint(str2lang("a"))
  )
  # don't drop a negated name
  expect_null(
    .getModelLineEquivalentLhsExpressionDropEndpoint(str2lang("-a"))
  )
  # drop linCmt() (issue #355)
  expect_equal(
    .getModelLineEquivalentLhsExpressionDropEndpoint(str2lang("-linCmt()~.")),
    str2lang("linCmt()")
  )
})

test_that(".getVariablesFromExpression", {
  expect_equal(.getVariablesFromExpression(""), character())
  expect_equal(.getVariablesFromExpression(5), character())
  expect_equal(.getVariablesFromExpression(as.name("a")), "a")
  expect_equal(.getVariablesFromExpression(str2lang("a~b")), c("a", "b"))
  # only pull the state from an ODE expression
  expect_equal(.getVariablesFromExpression(str2lang("d/dt(foo)")), "foo")
  expect_equal(.getVariablesFromExpression(str2lang("d(foo)")), "foo")
  expect_equal(.getVariablesFromExpression(str2lang("d(foo)|bar"), ignorePipe = TRUE), "foo")
})

test_that(".getLhs, .getRhs", {
  expect_equal(.getLhs(str2lang("a~b")), as.name("a"))
  expect_equal(.getLhs(str2lang("a~b|c")), str2lang("a"))
  expect_equal(.getLhs(str2lang("a~b+d|c")), str2lang("a"))
  expect_equal(.getLhs(str2lang("linCmt()~b+d|c")), str2lang("linCmt()"))

  expect_equal(.getRhs(str2lang("a~b")), as.name("b"))
  expect_equal(.getRhs(str2lang("a~b|c")), str2lang("b|c"))
  expect_equal(.getRhs(str2lang("a~b+d|c")), str2lang("b+d|c"))
})

rxTest({
  test_that("issue RxODE#272 and RxODE#273", {
    m1 <- rxode2({
      KA <- 2.94E-01
      CL <- 1.86E+01
      V2 <- 4.02E+01
      Q <- 1.05E+01
      V3 <- 2.97E+02
      Kin <- 1
      Kout <- 1
      EC50 <- 200
      ## Added modeled bioavaiblity, duration and rate
      fdepot <- 1
      durDepot <- 8
      rateDepot <- 1250
      C2 <- centr / V2
      C3 <- peri / V3
      d / dt(depot) <- -KA * depot
      f(depot) <- fdepot
      dur(depot) <- durDepot
      rate(depot) <- rateDepot
      d / dt(centr) <- KA * depot - CL * C2 - Q * C2 + Q * C3
      d / dt(peri) <- Q * C2 - Q * C3
      d / dt(eff) <- Kin - Kout * (1 - C2 / (EC50 + C2)) * eff
      eff(0) <- 1
    })

    ev <- et(timeUnits = "hr") |>
      et(amt = 10000, ii = 12, addl = 3) |>
      et(time = 6, cmt = "-depot", evid = 2, ii = 12, addl = 3) |>
      et(seq(0, 24, length.out = 100))

    expect_error(s <- rxSolve(m1, ev), NA)
    expect_false(s$C2[3] == 0) # 273
  })
})

rxTest({
  # Regression test for a heap-buffer-overflow in syncIdx()
  # (inst/include/rxode2parseHandleEvid.h).  handle_evid() advances ind->ixds
  # past the last dose and syncIdx() dereferenced ind->idose[ind->ixds] before
  # bounds-checking ixds against ind->ndoses, reading one past the idose array.
  #
  # The corruption is silent under a normal build (the solve still returns the
  # right values -- the out-of-bounds value is discarded by the re-sync), so
  # this test cannot fail on the value alone; it exists to exercise the exact
  # solve path under the package's ASAN / valgrind CI, where the overflow is
  # caught.  An oral two-compartment model with a modeled-duration infusion
  # (rate = -2 -> dur()), a lag time (alag()) and a bioavailability split
  # (f(A1)/f(A2)) dosed into two compartments drives ixds past ndoses.  Model
  # taken from the RxODE#435 regression.
  test_that("syncIdx does not read idose out of bounds (modeled dur + alag, multi-cmt dosing)", {
    rx <- rxode2({
      K20 <- CL / VC
      S2 <- VC
      d / dt(A1) <- -KA * A1
      d / dt(A2) <- KA * A1 - K20 * A2
      dur(A2) <- D2
      alag(A2) <- ALAG2
      f(A1) <- F1
      f(A2) <- 1 - F1
      CP <- A2 / S2
    })

    theta <- c(KA = 3, CL = 5, VC = 100, D2 = 4, F1 = 0, ALAG2 = 4)

    ev <- et(amt = 100000, rate = -2, cmt = 2) |>
      et(seq(0, 24, length.out = 100)) |>
      et(amt = 100000, cmt = 1)

    d <- suppressWarnings(rxSolve(rx, ev, theta, returnType = "data.frame"))

    expect_equal(nrow(d), 100L)
    expect_false(anyNA(d$CP))
    expect_true(all(is.finite(d$CP)))
    # lock in the known-good steady value so a regression that corrupts the
    # dose index (and therefore the solve) is also caught
    expect_equal(max(d$CP), 906.346235, tolerance = 1e-4)
  })
})

rxTest({
  ms <- c("liblsoda", "lsoda", "dop853")

  for (m in ms) {
    # context(sprintf("Modeled rate (%s)", m))

    mod.rate <- rxode2({
      a <- 6
      b <- 0.6
      f <- 1
      ri <- 10
      li <- 0
      d/dt(intestine) <- -a * intestine
      f(intestine) <- f
      rate(intestine) <- ri
      alag(intestine) <- li
      d/dt(blood) <- a * intestine - b * blood
    })

    et <- et() |>
      et(amt = 2 / 24, rate = -1, 0, addl = 9, ii = 1) |>
      et(seq(0, 10, by = 1 / 24))


    et2 <- et() |>
      et(amt = 2 / 24, rate = 2, 0, addl = 9, ii = 1) |>
      et(seq(0, 10, by = 1 / 24))

    f1.a <- rxSolve(mod.rate, et, c(ri = 2, f = 1, li = 0), method = m)
    f1.b <- rxSolve(mod.rate, et2, c(ri = 0.1, f = 1, li = 0), method = m)

    test_that("rate=2, f=1, lag=0 makes sense for modeled rate", {
      expect_equal(seq(0, 10, by = 1 / 24), f1.a$time)
      expect_equal(seq(0, 10, by = 1 / 24), f1.b$time)
      expect_equal(as.data.frame(f1.a), as.data.frame(f1.b))
    })

    f1.a <- rxSolve(mod.rate, et, c(ri = 2, f = 1, li = 0.5), method = m)
    f1.b <- rxSolve(mod.rate, et2, c(ri = 0.1, f = 1, li = 0.5), method = m)

    test_that("rate=2, f=1, lag=0.5 makes sense for modeled rate", {
      expect_equal(seq(0, 10, by = 1 / 24), f1.a$time)
      expect_equal(seq(0, 10, by = 1 / 24), f1.b$time)
      expect_equal(as.data.frame(f1.a), as.data.frame(f1.b))
    })

    f1.a <- rxSolve(mod.rate, et, c(ri = 2, f = 2, li = 0.5), method = m)
    f1.b <- rxSolve(mod.rate, et2, c(ri = 0.1, f = 2, li = 0.5), method = m)

    test_that("rate=2, f=2, lag=0.5 makes sense for modeled rate & fixed rate", {
      expect_equal(seq(0, 10, by = 1 / 24), f1.a$time)
      expect_equal(seq(0, 10, by = 1 / 24), f1.b$time)
      expect_equal(as.data.frame(f1.a), as.data.frame(f1.b))
    })

    et <- et() |>
      et(amt = 2 / 24, rate = -1) |>
      et(amt = 0.2, time = 1) |>
      et(seq(0, 10, by = 1 / 24))

    et2 <- et() |>
      et(amt = 2 / 24, rate = 2) |>
      et(amt = 0.2, time = 1) |>
      et(seq(0, 10, by = 1 / 24))

    f.a <- rxSolve(mod.rate, et, c(ri = 2, f = 2, li = 0.3), method = m)
    f.b <- rxSolve(mod.rate, et2, c(ri = 1e-6, f = 2, li = 0.3), method = m)

    test_that("rate=2 makes sense for modeled rate", {
      expect_equal(seq(0, 10, by = 1 / 24), f.a$time)
      expect_equal(seq(0, 10, by = 1 / 24), f.b$time)
      expect_equal(as.data.frame(f.a), as.data.frame(f.b))
    })

    f.a <- rxSolve(mod.rate, et, c(ri = 0.125, f = 2, li = 0.3), method = m)

    et2 <- et() |>
      et(amt = 2 / 24, rate = 0.125) |>
      et(amt = 0.2, time = 1) |>
      et(seq(0, 10, by = 1 / 24))

    f.b <- rxSolve(mod.rate, et2, c(ri = 1e-6, f = 2, li = 0.3), method = m)

    test_that("rate=0.125, f=2, lag=0.3 with 2 doses makes sense for modeled rate", {
      expect_equal(seq(0, 10, by = 1 / 24), f.a$time)
      expect_equal(seq(0, 10, by = 1 / 24), f.b$time)
      expect_equal(round(as.data.frame(f.a), 4), round(as.data.frame(f.b), 4))
    })

    f.a <- rxSolve(mod.rate, et, c(ri = 0.05, f = 2, li = 0.3), method = m)

    et2 <- et() |>
      et(amt = 2 / 24, rate = 0.05) |>
      et(amt = 0.2, time = 1) |>
      et(seq(0, 10, by = 1 / 24))

    f.b <- rxSolve(mod.rate, et2, c(ri = 1, f = 2, li = 0.3), method = m)

    et3 <- et() |>
      et(amt = 2 / 24 * 2, rate = 0.05) |>
      et(amt = 0.2 * 2, time = 1) |>
      et(seq(0, 10, by = 1 / 24))

    f.c <- rxSolve(mod.rate, et3, c(ri = 1, f = 1, li = 0.3), method = m)

    test_that("rate=0.05, f=2, lag=0.3 makes sense for modeled rate", {
      expect_equal(seq(0, 10, by = 1 / 24), f.a$time)
      expect_equal(seq(0, 10, by = 1 / 24), f.b$time)
      expect_equal(as.data.frame(f.a), as.data.frame(f.b))
    })

    test_that("bad rates (zero/negative) throw errors", {
      expect_error(rxSolve(mod.rate, et, c(ri = -1, f = 1, li = 0.3), method = m))
      expect_error(rxSolve(mod.rate, et, c(ri = 0, f = 1, li = 0.3), method = m))
    })

    mod.dur <- rxode2({
      a <- 6
      b <- 0.6
      f <- 1
      di <- 3
      li <- 0
      d / dt(intestine) <- -a * intestine
      f(intestine) <- f
      dur(intestine) <- di
      alag(intestine) <- li
      d/dt(blood) <- a * intestine - b * blood
    })

    test_that("Error when rate is requested but not in table", {
      expect_error(rxSolve(mod.dur, et, c(ri = 1, f = 1, li = 0.3), method = m))
    })

    # context(sprintf("Modeled duration (%s)", m))

    ## Now model duration
    et <- et(amt = 2 / 24 * 2, rate = -2) |>
      et(amt = 0.2 * 2, time = 2) |>
      et(seq(0, 10, by = 1 / 24))

    test_that("Error is thrown without modeled duration", {
      expect_error({
        rxSolve(mod.rate, et, c(ri = 1, f = 1, li = 0.3), method = m)
      })
    })

    f.a <- rxSolve(mod.dur, et, c(di = 0.5, f = 1, li = 0.3), method = m)

    et2 <- et() |>
      et(amt = 2 / 24 * 2, dur = 0.5) |>
      et(amt = 0.2 * 2, time = 2) |>
      et(seq(0, 10, by = 1 / 24))

    f.b <- rxSolve(mod.dur, et2, c(di = 10, f = 1, li = 0.3), method = m)

    test_that("Duration 0.5, F=1, li=0.3", {
      expect_equal(as.data.frame(f.a), as.data.frame(f.b))
    })

    et2 <- et() |>
      et(amt = 2 / 24 * 2, dur = 0.5) |>
      et(amt = 0.2 * 2, time = 2) |>
      et(seq(0, 10, by = 1 / 24))

    f.a <- rxSolve(mod.dur, et, c(di = 0.5, f = 1, li = 0.3), method = m)

    f.b <- rxSolve(mod.dur, et2, c(di = 10, f = 1, li = 0.3), method = m)

    test_that("Duration 0.5, F=1, li=0.3", {
      expect_equal(as.data.frame(f.a), as.data.frame(f.b))
    })

    f.a <- rxSolve(mod.dur, et, c(di = 0.5, f = 2, li = 0.3), method = m)

    f.b <- rxSolve(mod.dur, et2, c(di = 10, f = 2, li = 0.3), method = m)

    test_that("Duration 0.5, F=1, li=0.3", {
      expect_equal(as.data.frame(f.a), as.data.frame(f.b))
    })
  }

  test_that("state-dep f now allowed with modeled rate/dur; fixed-rate f(state) and rate(state)/dur(state) at zero still error; Issues RxODE#216 and RxODE#222", {
    mod.rate <- rxode2({
      a <- 6
      b <- 0.6
      f <- 1
      ri <- 10
      li <- 0
      d / dt(intestine) <- -a * intestine
      f(intestine) <- f
      rate(intestine) <- ri * intestine
      alag(intestine) <- li
      d / dt(blood) <- a * intestine - b * blood
    })

    et <- et() |>
      et(amt = 2 / 24, rate = -1, 0, addl = 9, ii = 1) |>
      et(seq(0, 10, by = 1 / 24))

    expect_error(rxSolve(mod.rate, et))

    mod.rate <- rxode2({
      a <- 6
      b <- 0.6
      f <- 1
      ri <- 10
      li <- 0
      d / dt(intestine) <- -a * intestine
      f(intestine) <- f
      dur(intestine) <- ri * intestine
      alag(intestine) <- li
      d / dt(blood) <- a * intestine - b * blood
    })

    et <- et() |>
      et(amt = 2 / 24, rate = -1, 0, addl = 9, ii = 1) |>
      et(seq(0, 10, by = 1 / 24))

    expect_error(rxSolve(mod.rate, et))

    mod.rate <- rxode2({
      a <- 6
      b <- 0.6
      f <- 1
      ri <- 10
      li <- 0
      d / dt(intestine) <- -a * intestine
      f(intestine) <- f * intestine
      rate(intestine) <- ri
      alag(intestine) <- li
      d / dt(blood) <- a * intestine - b * blood
    })

    et <- et() |>
      et(amt = 2 / 24, rate = 1, 0, addl = 9, ii = 1) |>
      et(seq(0, 10, by = 1 / 24))

    expect_error(rxSolve(mod.rate, et))

    et <- et() |>
      et(amt = 2 / 24, rate = -1, 0, addl = 9, ii = 1) |>
      et(seq(0, 10, by = 1 / 24))

    # f(intestine) <- f * intestine is now allowed with modeled rate (rate=-1)
    expect_no_error(rxSolve(mod.rate, et))

    mod.rate <- rxode2({
      a <- 6
      b <- 0.6
      f <- 1
      ri <- 10
      li <- 0
      d / dt(intestine) <- -a * intestine
      f(intestine) <- f * intestine
      dur(intestine) <- ri
      alag(intestine) <- li
      d / dt(blood) <- a * intestine - b * blood
    })

    et <- et() |>
      et(amt = 2 / 24, rate = -2, 0, addl = 9, ii = 1) |>
      et(seq(0, 10, by = 1 / 24))

    # f(intestine) <- f * intestine is now allowed with modeled duration (rate=-2)
    expect_no_error(rxSolve(mod.rate, et))
  })

  test_that("Test that state-based bio-availability's does work", {
    CLp <- 25
    CLh <- 1
    fuinc <- 1
    Nh <- 0.4
    Vh <- 3.9
    V1 <- 1800
    ke <- 0.02
    sample.int <- 24
    Sample.V <- 20
    Days <- 5
    ini <- c(A1 = 1000, A2 = 0, V.ms = 0)
    mod <- "
  d/dt(V.ms) = 0

  Vme = ke * time
  V1 = V1 - Vme - V.ms
  C1 = A1 / V1
  C2 = A2 / V2
  d/dt(A1) =  CLp * fuinc * (C2 - C1)
  f(A1)    = (V1 - Sample.V) / V1
  d/dt(A2) =  CLp * fuinc * (C1 - C2) - CLh * C2 * Nh
  d/dt(M) = CLh * C2 * Nh
"
    mod <- rxode2(model = mod)

    par <- c(
      CLp = CLp, CLh = CLh, fuinc = fuinc, Nh = Nh, V2 = Vh,
      V1 = V1 - Sample.V, Sample.V = Sample.V, ke = ke, Cp1 = 0
    )

    sim.ev <- et(seq(0, Days * 60 * 24, by = 1)) |>
      et(cmt = "V.ms", evid = 4, amt = Sample.V, nbr = 5, ii = sample.int * 60) |>
      et(cmt = "A1", time = 1440, evid = 6, amt = 1)

    expect_error(rxSolve(mod, par, sim.ev, inits = ini, addDosing = TRUE), NA)
  })
})

rxTest({
  test_that("rate and duration data mismatches; Issue RxODE#445", {

    TV_CLr <- 6.54 # L/h, (CLr/F)
    TV_CLnr <- 2.39 # L/h, (CLnr/F)
    TV_Vc <- 95.1 # L, (V/F)
    TV_alag <- 0.145 # h,
    TV_D <- 0.512 # h,
    TV_Q <- 2.1 # L/h, (Q/F)
    TV_Vp <- 23.3 # L, (Vp/F)

    # zero-order absorption with lag time 2-compartment

    mod <- rxode2({
      CL <- CLr + CLnr
      C2 <- central / Vc * 1000
      d / dt(central) <- -CL / Vc * central - Q / Vc * central + Q / Vp * periph
      d / dt(periph) <- Q / Vc * central - Q / Vp * periph
      alag(central) <- alag
      dur(central) <- D
    })

    ev <- et(amt = 2, cmt = "central", ii = 24, addl = 4) |>
      et(seq(0, 120, 0.1))

    ### check whether infusion duration was considered in simulation

    theta2 <- c(D = TV_D * 10, CLr = TV_CLr, CLnr = TV_CLnr, Vc = TV_Vc, Vp = TV_Vp, Q = TV_Q, alag = TV_alag)
    #expect_warning(sim2 <- rxSolve(mod, theta2, ev), "dur()")


    ev2 <- et(amt = 2, cmt = "central", ii = 24, addl = 4, rate = -2) |>
      et(seq(0, 120, 0.1))

    theta2 <- c(D = TV_D * 10, CLr = TV_CLr, CLnr = TV_CLnr, Vc = TV_Vc, Vp = TV_Vp, Q = TV_Q, alag = TV_alag)
    expect_warning(sim2 <- rxSolve(mod, theta2, ev2), NA)

    mod <- rxode2({
      CL <- CLr + CLnr
      C2 <- central / Vc * 1000
      d / dt(central) <- -CL / Vc * central - Q / Vc * central + Q / Vp * periph
      d / dt(periph) <- Q / Vc * central - Q / Vp * periph
      alag(central) <- alag
      rate(central) <- D
    })

    theta2 <- c(D = TV_D * 10, CLr = TV_CLr, CLnr = TV_CLnr, Vc = TV_Vc, Vp = TV_Vp, Q = TV_Q, alag = TV_alag)
    #expect_warning(sim2 <- rxSolve(mod, theta2, ev), "rate()")

    ev3 <- et(amt = 2, cmt = "central", ii = 24, addl = 4, rate = -1) |>
      et(seq(0, 120, 0.1))

    theta2 <- c(D = TV_D * 10, CLr = TV_CLr, CLnr = TV_CLnr, Vc = TV_Vc, Vp = TV_Vp, Q = TV_Q, alag = TV_alag)
    expect_warning(sim2 <- rxSolve(mod, theta2, ev3), NA)
  })
})

## Order test: integrate dx/dt = p*t^(p-1) over [0,1] with x(0)=0.
## A method of order p integrates this exactly (within machine epsilon).
## Mirrors rklib's rk_test_order.F90 test.
##
## NOTE: Fixed-step methods use hmin=1 (single step over [0,1]).
##       Adaptive methods use atol=rtol=1 (single coarse step).
##
## Flagged methods needing extra validation (transcription-error risk):
##   * f108, f1210, f1412: very many stages; also tested to tighter
##     tolerance in the decay test.
##   (!) FSAL methods: incorrect cache logic caught by decay test (multi-step).
##   (!) Low-storage (ls44, ls54): custom 2-register scheme, order tested here.
rxTest({
  .eps <- .Machine$double.eps
  .tol <- 1e4 * .eps   # 10000*eps, same threshold as rklib
  .ev  <- rxode2::et(1)

  ## -- Fixed-step methods ---------------------------------------------------
  ## Each entry: list(method, order). Uses hmin=1 for a single step.
  .fixedCases <- list(
    list(method = "euler", order = 1L),
    list(method = "midpoint",order = 2L),
    list(method = "heun",  order = 2L),
    list(method = "ssp22", order = 2L),
    list(method = "rk3",   order = 3L),
    list(method = "ssp53", order = 3L),
    list(method = "s4",    order = 4L),
    list(method = "r4",    order = 4L),
    list(method = "ls44",  order = 4L),
    list(method = "ls54",  order = 4L),
    list(method = "ssp54", order = 4L),
    list(method = "s5",    order = 5L),
    list(method = "rk5",   order = 5L),
    list(method = "c5",    order = 5L),
    list(method = "l5",    order = 5L),
    list(method = "lk5a",  order = 5L),
    list(method = "lk5b",  order = 5L),
    list(method = "b6",    order = 6L),
    list(method = "s7",    order = 7L),
    list(method = "s8_10",order = 8L),
    list(method = "cv8",   order = 8L),
    list(method = "s8_12", order = 8L),
    list(method = "s10",   order = 10L),
    list(method = "z10",   order = 10L),
    list(method = "o10",   order = 10L),
    list(method = "h10",   order = 10L)
  )

  ## -- Variable-step methods ------------------------------------------------
  ## Use atol=rtol=1 so one coarse step is taken (mimics fixed-step mode).
  .var_cases <- list(
    ## aliases
    list(method = "ck54",   order = 5L),
    list(method = "dp54",   order = 5L),
    list(method = "ssp33",  order = 3L),
    list(method = "dp87",   order = 8L),
    list(method = "v65e",   order = 6L),
    list(method = "v76e",   order = 7L),
    list(method = "v98e",   order = 9L),
    ## new variable-step
    list(method = "bs32",   order = 3L),
    list(method = "f45",    order = 4L),
    list(method = "t54",    order = 5L),
    list(method = "s54",    order = 5L),
    list(method = "pp54",   order = 5L),
    list(method = "pp54b",  order = 5L),
    list(method = "bs54",   order = 5L),
    list(method = "ss54",   order = 5L),
    list(method = "dp65",   order = 6L),
    list(method = "c65",    order = 6L),
    list(method = "tp64",   order = 6L),
    list(method = "v65r",   order = 6L),
    list(method = "v65",    order = 6L),
    list(method = "dverk65",  order = 6L),
    list(method = "tf65",   order = 6L),
    list(method = "tp75",   order = 7L),
    list(method = "tmy7",   order = 7L),
    list(method = "tmy7s",  order = 7L),
    list(method = "v76r",   order = 7L),
    list(method = "ss76",   order = 7L),
    list(method = "dp85",   order = 8L),
    list(method = "tp86",   order = 8L),
    list(method = "v87e",   order = 8L),
    list(method = "v87r",   order = 8L),
    list(method = "ev87",   order = 8L),
    list(method = "k87",    order = 8L),
    list(method = "t98a",   order = 9L),
    list(method = "v98r",   order = 9L),
    list(method = "s98",    order = 9L),
    list(method = "f108",   order = 10L),   # * extra transcription risk
    list(method = "c108",   order = 10L),
    list(method = "b109",   order = 10L),
    list(method = "s1110a", order = 11L),
    list(method = "f1210",  order = 12L),   # * extra transcription risk
    list(method = "o129",   order = 12L),
    list(method = "f1412",  order = 14L)    # * extra transcription risk
  )

  ## Build one model per unique polynomial order before running any tests.
  ## Methods sharing the same order reuse the same compiled model object.
  .all_orders <- unique(c(
    sapply(.fixedCases, `[[`, "order"),
    sapply(.var_cases,  `[[`, "order")
  ))
  .poly_mods <- lapply(.all_orders, function(.p) {
    rxode2::rxode2(paste0("d/dt(x) <- ", .p, "*t^", .p - 1L))
  })
  names(.poly_mods) <- as.character(.all_orders)

  for (.cfg in .fixedCases) {
    local({
      .meth  <- .cfg$method
      .order <- .cfg$order
      .mod   <- .poly_mods[[as.character(.order)]]
      test_that(paste("order test (fixed):", .meth, "order", .order), {
        .s <- rxode2::rxSolve(.mod, .ev, inits = c(x = 0),
                               method = .meth, hmin = 1.0, cores = 1)
        expect_lt(abs(.s$x[1] - 1.0), .tol,
                  label = paste(.meth, "order test error"))
      })
    })
  }

  for (.cfg in .var_cases) {
    local({
      .meth  <- .cfg$method
      .order <- .cfg$order
      .mod   <- .poly_mods[[as.character(.order)]]
      test_that(paste("order test (adaptive):", .meth, "order", .order), {
        .s <- rxode2::rxSolve(.mod, .ev, inits = c(x = 0),
                               method = .meth, atol = 1e10, rtol = 1e10, cores = 1)
        expect_lt(abs(.s$x[1] - 1.0), .tol,
                  label = paste(.meth, "order test error"))
      })
    })
  }
})

if (!.Call(`_rxode2_isIntel`)) {
  test_that("Demo runs successfully", {
    skip_if_not(dir.exists(file.path(system.file(package = "rxode2"), "demo")), "demo not installed")
    expect_error(suppressWarnings({
      demo("demo1", "rxode2", ask = FALSE, echo = FALSE)
    }), NA)
  })
}

rxTest({
  # sigdig sets the ODE solver tolerances with one simple, solver-independent
  # formula: rtol = 10^-sigdig, atol = 10^(-sigdig-3).  Sensitivity solves match
  # the main solve; steady-state solves run one order looser.  Same for every solver
  # (stiff, non-stiff, auto-switching).

  test_that("sigdig sets the same tolerances for every solver", {
    for (m in c("liblsoda", "lsoda", "dop853", "dop5", "ros4", "cvode", "bdf",
                "dop853+ros4")) {
      .c <- rxControl(sigdig = 4, method = m)
      expect_equal(.c$rtol, 1e-4, info = m)
      expect_equal(.c$atol, 1e-7, info = m)
      expect_equal(.c$rtolSens, 1e-4, info = m)
      expect_equal(.c$atolSens, 1e-7, info = m)
      expect_equal(.c$ssRtol[1], 1e-3, info = m)
      expect_equal(.c$ssAtol[1], 1e-6, info = m)
      expect_equal(.c$ssRtolSens[1], 1e-3, info = m)
      expect_equal(.c$ssAtolSens[1], 1e-6, info = m)
    }
  })

  test_that("sigdig scales the tolerances by one order per significant digit", {
    .c <- rxControl(sigdig = 6)
    expect_equal(.c$rtol, 1e-6)
    expect_equal(.c$atol, 1e-9)
    .d <- rxControl(sigdig = 3)
    expect_equal(.d$rtol, 1e-3)
    expect_equal(.d$atol, 1e-6)
  })

  test_that("an explicit atol/rtol overrides the sigdig-derived value", {
    .c <- rxControl(sigdig = 4, atol = 1e-12, rtol = 1e-10)
    expect_equal(.c$atol, 1e-12)
    expect_equal(.c$rtol, 1e-10)
  })

  test_that("each tolerance is resolved independently of the others", {
    # an explicit atol/rtol wins for the main solve but does not propagate to the
    # sensitivity or steady-state tolerances; those stay sigdig-derived
    .c <- rxControl(sigdig = 4, atol = 1e-12, rtol = 1e-10)
    expect_equal(.c$atolSens, 1e-7)
    expect_equal(.c$rtolSens, 1e-4)
    expect_equal(.c$ssAtol[1], 1e-6)
    expect_equal(.c$ssRtol[1], 1e-3)
    # and an explicit sensitivity tolerance wins over the sigdig-derived one
    .d <- rxControl(sigdig = 4, atolSens = 1e-12, rtolSens = 1e-10)
    expect_equal(.d$atolSens, 1e-12)
    expect_equal(.d$rtolSens, 1e-10)
    expect_equal(.d$atol, 1e-7)
    expect_equal(.d$rtol, 1e-4)
  })

  test_that("sigdig is unused unless it is explicitly requested", {
    # sigdig defaults to NULL and must not touch any tolerance
    .d <- rxControl()
    expect_null(.d$sigdig)
    expect_equal(.d$atol, 1e-8)
    expect_equal(.d$rtol, 1e-6)
    expect_equal(.d$atolSens, 1e-8)
    expect_equal(.d$rtolSens, 1e-6)
    expect_equal(.d$ssAtol[1], 1e-8)
    expect_equal(.d$ssRtol[1], 1e-6)
    expect_equal(.d$ssAtolSens[1], 1e-8)
    expect_equal(.d$ssRtolSens[1], 1e-6)
    # an explicit NULL is the same as not passing it
    expect_equal(rxControl(sigdig = NULL)$rtol, 1e-6)
  })
})

rxTest({
  test_that("state-dep rate uses per-subject initial state at sort phase", {
    # Two subjects with different initial central concentrations
    # rate(depot) depends on central, so infusion duration differs between subjects
    mod <- rxode2({
      rate(depot) <- central * 0.1 + 1.0
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - kel * central
    })
    et <- et(amt = 100, rate = -1) %>% et(seq(0, 24, by = 1))
    s0 <- rxSolve(mod, et, c(ka = 0.5, kel = 0.2),
                  inits = c(depot = 0, central = 0))
    s5 <- rxSolve(mod, et, c(ka = 0.5, kel = 0.2),
                  inits = c(depot = 0, central = 5))
    # With central=5, rate is higher so infusion ends sooner ->
    # the subsequent trajectories should differ
    expect_false(isTRUE(all.equal(s0$central, s5$central)))
    expect_false(any(is.na(s0$central)))
    expect_false(any(is.na(s5$central)))
  })

  test_that("runtime re-sort places stop event correctly after rate recompute", {
    mod <- rxode2({
      rate(depot) <- central + 2
      d/dt(depot) <- -ka * depot
      d/dt(central) <- ka * depot - kel * central
    })
    et <- et() |>
      et(amt = 100, rate = -1, time = 0) |>
      et(amt = 100, rate = -1, time = 2) |>
      et(seq(0, 10))
    s <- rxSolve(mod, et, c(ka = 0.5, kel = 0.2),
                 inits = c(depot = 0, central = 3))
    expect_false(any(is.na(s$central)))
  })
})

rxTest({
  test_that("mixed oral iv etc model; Issue RxODE#435", {
    rx1 <- rxode2({
      K20 <- CL / VC
      S2 <- VC
      d / dt(A1) <- -KA * A1
      d / dt(A2) <- KA * A1 - K20 * A2
      dur(A2) <- D2
      alag(A2) <- ALAG2
      CP <- A2 / S2
    })

    rx2 <- rxode2({
      K20 <- CL / VC
      S2 <- VC
      d / dt(A1) <- -KA * A1
      d / dt(A2) <- KA * A1 - K20 * A2
      dur(A2) <- D2
      alag(A2) <- ALAG2
      f(A1) <- F1
      f(A2) <- 1 - F1
      CP <- A2 / S2
    })

    t <- c(
      KA = 3,
      CL = 5,
      VC = 100,
      D2 = 4,
      F1 = 0,
      ALAG2 = 4
    )

    et1 <- et(amt = 100000, rate = -2, cmt = 2) |>
      et(seq(0, 24, length.out = 100))

    et2 <- et1 |>
      et(amt = 100000, cmt = 1)

    a <- rxSolve(rx1, et1, t, returnType = "data.frame")
    b <- rxSolve(rx2, et2, t, returnType = "data.frame")

    expect_equal(a, b)
  })
})

if (!.Call(`_rxode2_isIntel`)) {
  test_that("Bioavailability doesn't impact other cmts (issue RxODE#160)", {
    rx <- rxode2({
      ka <- .2
      BA <- 0.9
      cl <- 0.3 # fixed at value for model1/base_2comp_iv
      v1 <- 0.1
      q <- 0.01
      v2 <- 0.1
      d / dt(depot) <- -ka * depot
      f(depot) <- BA
      d / dt(centr) <- ka * depot + (q / v2) * periph - (q / v1) * centr -
        (cl / v1) * centr
      d / dt(periph) <- (q / v1) * centr - (q / v2) * periph
    })

    s <- rxSolve(rx, et(amt = 100, cmt = "centr"))

    expect_false(all(s$centr == 0))
  })
}

rxTest({

  test_that("dur=0 is treated as a bolus in non steady-state doses (ODE)", {

    f <- rxode2({
      cl <- 1.1
      v <- 20
      d/dt(central) <- - (cl/v)*central
      dur(central) <- dur2
      cp <- central/(v/1000)
    })

    e <- et(rate=-2, amt=100)
    s <- rxSolve(f, e, c(dur2=0))

    e <- et(amt=100)
    s2 <- rxSolve(f, e, c(dur2=100))

    expect_equal(s$central, s2$central)
    expect_equal(s$cp, s2$cp)

  })

  test_that("dur=0 is treated as a bolus in non steady-state doses (linCmt())", {

    f <- rxode2({
      cl <- 1.1
      v <- 20
      cp <- linCmt()
      dur(central) <- dur2
    })

    e <- et(rate=-2, amt=100)
    s <- rxSolve(f, e, c(dur2=0))

    e <- et(amt=100)
    s2 <- rxSolve(f, e, c(dur2=100))

    expect_equal(s$central, s2$central)
    expect_equal(s$cp, s2$cp)

  })

  test_that("dur=0 with f is treated the same as bolus with f (ODE)", {

    f <- rxode2({
      cl <- 1.1
      v <- 20
      d/dt(central) <- - (cl/v)*central
      dur(central) <- dur2
      f(central) <- fcentral
      cp <- central/(v/1000)
    })

    e <- et(rate=-2, amt=100)
    s <- rxSolve(f, e, c(dur2=0, fcentral=2))
    sb <- rxSolve(f, e, c(dur2=0, fcentral=0.5))

    e <- et(amt=100)
    s2 <- rxSolve(f, e, c(dur2=100, fcentral=2))
    s2b <- rxSolve(f, e, c(dur2=100, fcentral=0.5))

    expect_equal(s$central, s2$central)
    expect_equal(s$cp, s2$cp)

  })

  test_that("dur=0 with f is treated the same as bolus with f (linCmt)", {

    f <- rxode2({
      cl <- 1.1
      v <- 20
      cp <- linCmt()
      dur(central) <- dur2
      f(central) <- fcentral
    })

    e <- et(rate=-2, amt=100)
    s <- rxSolve(f, e, c(dur2=0, fcentral=2))
    sb <- rxSolve(f, e, c(dur2=0, fcentral=0.5))

    e <- et(amt=100)
    s2 <- rxSolve(f, e, c(dur2=100, fcentral=2))
    s2b <- rxSolve(f, e, c(dur2=100, fcentral=0.5))

    expect_equal(s$central, s2$central)
    expect_equal(s$cp, s2$cp)

  })

  test_that("dur=0 is treated as a bolus in non steady-state doses (ODE, SS)", {

    f <- rxode2({
      cl <- 1.1
      v <- 20
      d/dt(central) <- - (cl/v)*central
      dur(central) <- dur2
      cp <- central/(v/1000)
    })

    e <- et(rate=-2, amt=100, ss=1, ii=12)
    s <- rxSolve(f, e, c(dur2=0))

    e <- et(amt=100, ss=1, ii=12)
    s2 <- rxSolve(f, e, c(dur2=100))

    expect_equal(s$central, s2$central)
    expect_equal(s$cp, s2$cp)

    f <- rxode2({
      cl <- 1.1
      v <- 20
      d/dt(central) <- - (cl/v)*central
      dur(central) <- dur2
      lag(central) <- lag2
      cp <- central/(v/1000)
    })

    e <- et(rate=-2, amt=100, ss=1, ii=12)
    s <- rxSolve(f, e, c(dur2=0, lag2=2))

    e <- et(amt=100, ss=1, ii=12)
    s2 <- rxSolve(f, e, c(dur2=100, lag2=2))

    expect_equal(s$central, s2$central, tolerance=1e-6)
    expect_equal(s$cp, s2$cp, tolerance=1e-6)

    expect_equal(s$time, s2$time)

  })


  test_that("dur=0 is treated as a bolus in non steady-state doses (linCmt, SS)", {

    f <- rxode2({
      cl <- 1.1
      v <- 20
      cp <- linCmt()
      dur(central) <- dur2
    })

    e <- et(rate=-2, amt=100, ss=1, ii=12)
    s <- rxSolve(f, e, c(dur2=0))

    e <- et(amt=100, ss=1, ii=12)
    s2 <- rxSolve(f, e, c(dur2=100))

    expect_equal(s$central, s2$central)
    expect_equal(s$cp, s2$cp)

    f <- rxode2({
      cl <- 1.1
      v <- 20
      cp <- linCmt()
      dur(central) <- dur2
      lag(central) <- lag2
    })

    e <- et(rate=-2, amt=100, ss=1, ii=12)
    s <- rxSolve(f, e, c(dur2=0, lag2=2))

    e <- et(amt=100, ss=1, ii=12)
    s2 <- rxSolve(f, e, c(dur2=100, lag2=2))

    expect_equal(s$central, s2$central, tolerance=1e-6)
    expect_equal(s$cp, s2$cp, tolerance=1e-6)

    expect_equal(s$time, s2$time)

  })


})

test_that("safeZero solving", {


  m <- rxode2({
    nan0 = 0/0
    nanLog = log(-1)
    nanPow = 0^-1
    rPow = 0^-1.5
  })

  et <- et(1)

  unsafe <- rxSolve(m, et, safeZero=FALSE, safePow=FALSE, safeLog=FALSE, useStdPow = TRUE)
  expect_true(is.nan(unsafe$nan0))
  expect_true(is.nan(unsafe$nanLog))
  expect_false(is.finite(unsafe$nanPow))
  expect_false(is.finite(unsafe$rPow))

  safe <- rxSolve(m, et, safeZero=TRUE, safePow=TRUE, safeLog=TRUE, useStdPow=FALSE)

  expect_false(is.nan(safe$nan0))
  expect_false(is.nan(safe$nanLog))
  expect_true(is.finite(safe$nanPow))
  expect_true(is.finite(safe$rPow))

  safe <- rxSolve(m, et, safeZero=TRUE, safePow=FALSE, safeLog=TRUE, useStdPow=FALSE)

  expect_false(is.na(safe$nan0))
  expect_false(is.na(safe$nanLog))
  expect_false(is.finite(safe$nanPow))
  expect_false(is.finite(safe$rPow))

})

test_that("when reserved variables are assigned, parser should error",{

  expect_error(rxode2({
    C2 <- centr/V2
    C3 <- peri/V3
    d/dt(depot) <- -KA*depot
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
    d/dt(peri)  <- Q*C2 - Q*C3
    eff(0) <- 1
    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
    t <- 1
  }))

})

rxTest({
  test_that("RxODE#393", {
    filePath <- test_path("test-issue-393.rds")

    skip_if_not(file.exists(filePath))

    d <- readRDS(filePath)

    mod1 <- rxode2({
      #  CLH = THETA[1];
      #  CLD = THETA[2];
      #  V_centr = THETA[3];
      #  V_peri = THETA[4];
      CLR <- 0
      ka <- 0
      FQ_LV <- 2.604
      BW <- 6.2 # kg
      Qh <- FQ_LV * BW
      RBP <- 0.679
      Fh <- 1 - CLH / RBP / Qh
      FaFg <- 1
      C_centr <- A_centr / V_centr  #           venous (VE) total  venous blood
      C_peri <- A_peri / V_peri  #  			arterial (AR) total  arterial blood
      d / dt(A_centr) <- -C_centr * (CLR + CLH) + C_peri * CLD - C_centr * CLD + ka * A_absorption  # 	central
      d / dt(A_peri) <- -C_peri * CLD + C_centr * CLD  # peripheral
      d / dt(A_absorption) <- -ka * A_absorption
      f(A_absorption) <- Fh * FaFg
    })

    mod2 <- rxode2({
      CLH <- THETA[1]
      CLD <- THETA[2]
      V_centr <- THETA[3]
      V_peri <- THETA[4]
      CLR <- 0
      ka <- 0
      FQ_LV <- 2.604
      BW <- 6.2 # kg
      Qh <- FQ_LV * BW
      RBP <- 0.679
      Fh <- 1 - CLH / RBP / Qh
      FaFg <- 1
      C_centr <- A_centr / V_centr  #           venous (VE) total  venous blood
      C_peri <- A_peri / V_peri  #  			arterial (AR) total  arterial blood
      d / dt(A_centr) <- -C_centr * (CLR + CLH) + C_peri * CLD - C_centr * CLD + ka * A_absorption  # 	central
      d / dt(A_peri) <- -C_peri * CLD + C_centr * CLD  # peripheral
      d / dt(A_absorption) <- -ka * A_absorption
      f(A_absorption) <- Fh * FaFg
    })


    s1 <- rxSolve(mod1, d, params = c(CLH = 2, CLD = 10, V_centr = 77, V_peri = 10), returnType = "data.frame")
    s2 <- rxSolve(mod2, d, params = setNames(c(2, 10, 77, 10), paste0("THETA[", 1:4, "]")), returnType = "data.frame")
    s2 <- s2[, names(s1)]
    expect_equal(s1, s2)

    s3 <- rxSolve(mod2, d,
                  params = c(c(CLR = 0, ka = 0, FQ_LV = 2.604, BW = 6.2, RBP = 0.679, FaFg = 1), setNames(c(2, 10, 77, 10), paste0("THETA[", 1:4, "]"))),
                  returnType = "data.frame"
                  )

    s3 <- s3[, names(s1)]
    expect_equal(s1, s3)

    s4 <- rxSolve(mod2, d, theta = c(2, 10, 77, 10), returnType = "data.frame")

    s4 <- s4[, names(s1)]
    expect_equal(s1, s3)

    s5 <- rxSolve(mod2, d, theta = c(2, 10, 77, 10))

    s6 <- rxSolve(mod2, d, params = setNames(c(2, 10, 77, 10), paste0("THETA[", 1:4, "]")))
  })
})

test_that("LHS string assignment without explicit levels does not segfault", {
  # This reproduces a condition where an LHS string variable is assigned
  # but no levels are present in the dataset or model factors, which
  # previously caused rxode2_df.cpp to access a nullptr colI during OpenMP output.
  
  mod <- rxode2({
    limit <- "none"
    d/dt(a) <- -a
  })
  
  dat <- data.frame(time=seq(0, 10, by=1), evid=0)
  
  # Before the fix, rxSolve would throw a segmentation fault (core dumped) here.
  # Now it should gracefully complete and limit should be a factor column.
  expect_error(res <- rxSolve(mod, dat), NA)
  
  expect_true(inherits(res, "rxSolve"))
  expect_true("limit" %in% names(res))
})

rxTest({
  ode.1c <- rxode2({
    V <- 20
    Cl <- 1
    fc <- 1
    C2 <- center / V
    d/dt(center) ~ -Cl * C2
    f(center) <- fc
  })

  dur <- 0.5
  ii <- 2
  m <- "liblsoda"
  d <- 3

  for (m in .methods0) {
    test_that(sprintf("missing evid tests for %s", m), {
      for (dur in c(0.5, 1)) {
        for (ii in seq(2, 24, by = 2)) {
          et <- et() |>
            et(amt = d, rate = d / dur) |>
            et(time = ii, amt = d, ii = ii, addl = floor(24 / ii), rate = d / dur) |>
            et(c(dur, seq(0, 24, length.out = 200)))

          x2 <- solve(ode.1c, et, method = m, maxsteps = 10000)

          et <- data.frame(et)
          et <- et[, names(et) != "evid"]

          expect_warning(x3 <- solve(ode.1c, et, method = m, maxsteps = 10000), NA)

          expect_equal(data.frame(x2), data.frame(x3))

          et2 <- et(et)

          expect_equal(data.frame(et), data.frame(et2)[, c("time", "amt", "rate", "ii", "addl")])


          d <- 3
          et <- et() |>
            et(amt = d, dur = dur) |>
            et(time = ii, amt = d, ii = ii, addl = floor(24 / ii), dur = dur) |>
            et(c(dur, seq(0, 24, length.out = 200)))

          x2 <- solve(ode.1c, et, method = m, maxsteps = 10000)

          et <- data.frame(et)
          et <- et[, names(et) != "evid"]

          expect_warning(x3 <- solve(ode.1c, et, method = m, maxsteps = 10000), NA)

          expect_equal(data.frame(x2), data.frame(x3))

          et <- et() |>
            et(amt = d, ii = ii, ss = 1, dur = dur) |>
            et(c(dur, seq(0, 24, length.out = 200)))

          x2 <- solve(ode.1c, et, method = m, maxsteps = 10000)

          et <- data.frame(et)
          et <- et[, names(et) != "evid"]

          expect_warning(x3 <- solve(ode.1c, et, method = m, maxsteps = 10000), NA)

          expect_equal(data.frame(x2), data.frame(x3))
        }
      }
    })
  }
})

rxTest({

  mod <- rxode2({
    d/dt(depot) <- -KA * depot
    d/dt(centr) <- KA * depot - CL / V * centr
    d/dt(peri)  <- Q / V * centr - Q / V2 * peri
  })

  # 16 subjects -- large enough to span multiple OMP threads
  p <- data.frame(
    KA = 0.3,
    CL = seq(1, 8, length.out = 16),
    V  = 40,
    Q  = 10,
    V2 = 300
  )

  et <- et(seq(0, 48, by = 1)) |> et(amt = 100)

  s1 <- suppressWarnings(rxSolve(mod, p, et, method = "dop853", cores = 1))
  s2 <- suppressWarnings(rxSolve(mod, p, et, method = "dop853", cores = 2))

  test_that("dop853 cores=1 and cores=2 give identical results", {
    expect_equal(as.data.frame(s1), as.data.frame(s2))
  })

  # Per-compartment tolerance equivalence: explicitly equal vector tolerances
  # must produce the same result as the scalar default path since rtol2/atol2
  # are populated from RTOL/ATOL when a scalar is supplied.
  s_default <- suppressWarnings(
    rxSolve(mod, p, et, method = "dop853", cores = 1, atol = 1e-8, rtol = 1e-6)
  )
  s_vec <- suppressWarnings(
    rxSolve(mod, p, et, method = "dop853", cores = 1,
            atol = rep(1e-8, 3), rtol = rep(1e-6, 3))
  )

  test_that("scalar and equal per-compartment tolerances give identical results for dop853", {
    expect_equal(as.data.frame(s_default), as.data.frame(s_vec))
  })

  # Race condition guard: repeated parallel calls must be deterministic
  s2a <- suppressWarnings(rxSolve(mod, p, et, method = "dop853", cores = 2))
  s2b <- suppressWarnings(rxSolve(mod, p, et, method = "dop853", cores = 2))

  test_that("parallel dop853 is deterministic across repeated calls", {
    expect_equal(as.data.frame(s2a), as.data.frame(s2b))
  })

  # Dense-path parity: cores=1 vs cores=2
  s_d1 <- suppressWarnings(rxSolve(mod, p, et, method = "dop853",
                                    dense = TRUE, cores = 1))
  s_d2 <- suppressWarnings(rxSolve(mod, p, et, method = "dop853",
                                    dense = TRUE, cores = 2))

  test_that("dop853 dense: cores=1 and cores=2 give identical results", {
    expect_equal(as.data.frame(s_d1), as.data.frame(s_d2))
  })

  s_nd <- suppressWarnings(rxSolve(mod, p, et, method = "dop853",
                                    dense = FALSE, cores = 1))

  test_that("dop853 dense=TRUE matches dense=FALSE (cores=1)", {
    expect_equal(as.data.frame(s_d1), as.data.frame(s_nd), tolerance = 1e-5)
  })

  s_d2a <- suppressWarnings(rxSolve(mod, p, et, method = "dop853",
                                     dense = TRUE, cores = 2))
  s_d2b <- suppressWarnings(rxSolve(mod, p, et, method = "dop853",
                                     dense = TRUE, cores = 2))

  test_that("parallel dense dop853 is deterministic across repeated calls", {
    expect_equal(as.data.frame(s_d2a), as.data.frame(s_d2b))
  })

  # nmtest dataset: dense vs non-dense must give identical cp for ODE model
  skip_if_not_installed("nlmixr2data")
  d <- nlmixr2data::nmtest

  f <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    d/dt(depot)   <- -ka * depot
    d/dt(central) <- ka * depot - (cl / v) * central
    f(central)    <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central)  <- dur2
    cp <- central / (v / 1000)
  })

  fl <- rxode2({
    cl <- 1.1
    v <- 20
    ka <- 1.5
    d/dt(depot)   <- -ka * depot
    d/dt(central) <- ka * depot - (cl / v) * central
    lag(central)  <- lagt
    f(central)    <- bioav
    if (mode == 1) rate(central) <- rat2
    if (mode == 2) dur(central)  <- dur2
    cp <- central / (v / 1000)
  })

  lapply(unique(d$id), function(id) {
    di <- d[d$id == id, ]
    for (addlDropSs in c(TRUE, FALSE)) {
      test_that(paste0("dop853 dense==nodense nmtest id:", id,
                       " addlDropSs:", addlDropSs), {
        s_nodense <- rxSolve(f, di, method = "dop853", dense = FALSE,
                             addlDropSs = addlDropSs)
        s_dense   <- rxSolve(f, di, method = "dop853",
                             addlDropSs = addlDropSs)
        expect_equal(s_nodense$cp, s_dense$cp, tolerance = 1e-5)
      })
      test_that(paste0("dop853 dense==nodense nmtest id:", id,
                       " alag addlDropSs:", addlDropSs), {
        s_nodense <- rxSolve(fl, di, method = "dop853", dense = FALSE,
                             addlDropSs = addlDropSs)
        s_dense   <- rxSolve(fl, di, method = "dop853",
                             addlDropSs = addlDropSs)
        expect_equal(s_nodense$cp, s_dense$cp, tolerance = 1e-5)
      })
    }
  })
})

rxTest({
  .m <- rxode2({
    d/dt(depot)   <- -ka * depot
    d/dt(central) <- ka * depot - cl / vd * central
    cp <- central / vd
  })
  .p <- c(ka = 0.5, cl = 1, vd = 10)
  .e <- et(amt = 100, time = 0) |> et(seq(0, 24, by = 1))

  test_that("indOwnAlloc=TRUE gives same results as indOwnAlloc=FALSE (single subject)", {
    .r1 <- rxSolve(.m, .p, .e, indOwnAlloc = TRUE)
    .r2 <- rxSolve(.m, .p, .e, indOwnAlloc = FALSE)
    expect_equal(.r1$cp,    .r2$cp,    tolerance = 1e-8)
    expect_equal(.r1$depot, .r2$depot, tolerance = 1e-8)
  })

  test_that("indOwnAlloc=TRUE works with multiple subjects", {
    # params as a per-individual data frame, shared event table
    .params <- data.frame(
      id = 1:5,
      ka = 0.3 + (1:5) * 0.1,
      cl = 1.0,
      vd = 10.0
    )
    .r1 <- rxSolve(.m, .params, .e, indOwnAlloc = TRUE)
    .r2 <- rxSolve(.m, .params, .e, indOwnAlloc = FALSE)
    expect_equal(.r1$cp, .r2$cp, tolerance = 1e-8)
  })

  test_that("repeated solves with indOwnAlloc=TRUE do not crash", {
    for (.i in seq_len(5)) {
      .r <- rxSolve(.m, .p, .e, indOwnAlloc = TRUE)
      gc()
    }
    .r    <- rxSolve(.m, .p, .e, indOwnAlloc = TRUE)
    .rref <- rxSolve(.m, .p, .e, indOwnAlloc = FALSE)
    expect_equal(.r$cp, .rref$cp, tolerance = 1e-8)
  })
})

if (!.Call(`_rxode2_isIntel`)) {
  test_that("issue RxODE#213 doesn't drop patients, 1e+5 doesn't drop", {
    rx <- rxode2({
      ka <- .2
      BA <- 0.9
      cl <- 0.3 # fixed at value for model1/base_2comp_iv
      v1 <- 0.1
      q <- 0.01
      v2 <- 0.1
      d / dt(depot) <- -ka * depot
      f(depot) <- BA
      d / dt(centr) <- ka * depot + (q / v2) * periph - (q / v1) * centr -
        (cl / v1) * centr
      d / dt(periph) <- (q / v1) * centr - (q / v2) * periph
    })

    dat <- data.frame(
      id = c(1e5, 1e5 + 1),
      ka = .2,
      BA = 0.9,
      cl = 0.3, # fixed at value for model1/base_2comp_iv
      v1 = 0.1,
      q = 0.01,
      v2 = 0.1
    )

    expect_warning(
      rxSolve(rx, params = dat, events = et(amt = 100, cmt = "centr", id = c(1e5, 1e5 + 1))),
      NA
    )

    dat <- data.frame(
      id = as.integer(c(1e5, 1e5 + 1)),
      ka = .2,
      BA = 0.9,
      cl = 0.3, # fixed at value for model1/base_2comp_iv
      v1 = 0.1,
      q = 0.01,
      v2 = 0.1
    )
    expect_warning(
      rxSolve(rx, params = dat, events = et(amt = 100, cmt = "centr", id = c(1e5, 1e5 + 1))),
      NA
    )
  })
}

if (!.Call(`_rxode2_isIntel`)) {
  test_that("Assign statement", {
    expect_error(capture.output(rxode2({
      x <- max(x, 0)
      d / dt(x) <- x - a * y
      d / dt(y) <- b * x - 2 * y
    })))
  })
}

rxTest({
  test_that("RxODE#398", {
    filename <- test_path("test-issue-398.rds")
    skip_if_not(file.exists(filename))

    NM_data <- readRDS(filename)

    # Define and Compile model
    mod1 <- rxode2({
      CLH_int <- THETA[1]
      CLD <- THETA[2]
      V_centr <- THETA[3]
      V_peri <- THETA[4]
      fup <- 0.28 #
      RBP <- 1
      CLR <- 0
      ka <- 10
      FQ_LV <- 1.26
      BW <- 70 # kg
      Qh <- FQ_LV * BW
      CLH <- Qh * CLH_int * fup / (Qh + CLH_int * fup / RBP)
      Fh <- 1 - CLH / RBP / Qh
      FaFg <- 0.82
      C_centr <- A_centr / V_centr #           venous (VE) total  venous blood
      C_peri <- A_peri / V_peri #  			arterial (AR) total  arterial blood
      d / dt(A_centr) <- -C_centr * (CLR + CLH) + C_peri * CLD - C_centr * CLD + ka * A_absorption # 	central
      d / dt(A_peri) <- -C_peri * CLD + C_centr * CLD # peripheral
      d / dt(A_absorption) <- -ka * A_absorption
      f(A_absorption) <- Fh * FaFg
      dur(A_absorption) <- THETA[5]
      DV <- DV # copying data columns from data set for fitting
      MDV <- MDV
      PRED <- C_centr
    })

    # simulation
    THETA <- c(10, 2.619, 10, 22.14, 5)
    suppressWarnings(
      expect_error(solve(mod1, NM_data, params = setNames(THETA, paste0("THETA[", 1:5, "]"))), NA)
    )
  })
})

rxTest({
  withr::with_tempdir({
    test.dir <- tempfile("Rx_base-")
    dir.create(test.dir, showWarnings = FALSE)
    ode <- "d/dt(y) = r * y * (1.0 - y/K);"
    fn <- file.path(test.dir, "exam3.1.txt")
    writeLines(ode, fn)

    m1 <- rxode2(model = ode, modName = "m1")

    test_that("ODE inside a string", {
      expect_s3_class(m1, "rxode2")
    })

    m2 <- rxode2(filename = fn, modName = "f1")
    test_that("ODE inside a text file", {
      expect_s3_class(m2, "rxode2")
    })

    test_that("arguments model= and filename= are mutually exclusive.", {
      expect_error(rxode2(model = ode, filename = fn), "must specify exactly one of 'model' or 'filename'")
    })

    unlink(test.dir, recursive = TRUE)
  })
})

#!/bin/bash
export R_HOME=/usr/lib/R/
export R_HOME_DIR=/usr/lib/R
export R_SHARE_DIR=/usr/share/R/share
export R_INCLUDE_DIR=/usr/share/R/include
export R_DOC_DIR=/usr/share/R/doc
. /usr/lib/R/etc/ldpaths

LD_PRELOAD=/usr/lib/x86_64-linux-gnu/libprofiler.so CPUPROFILE=sample.profile /usr/lib/R/bin/exec/R -f linCmtTest.R

#google-pprof --text /usr/bin/R ./sample.profile
google-pprof --web /usr/bin/R ./sample.profile

---
title: "rxode2 syntax"
output:
  learnr::tutorial:
    progressive: true
runtime: shiny_prerendered
description: >
  This covers the basic syntax supported for `rxode2` models and how to compile the models using `rxode2`
---

```{r setup, include=FALSE}
library(learnr)
library(rxode2)
library(ggplot2)
knitr::opts_chunk$set(echo = FALSE)
```

From Rstudio 1.3+ tutorial pane you can use the expand button
(![](images/expandTutorial.png)) to navigate tutorial sections.

## Creating Compiled models

`rxode2` currently supports differential equations using Leibniz
notation, that is `d/dt(x)`. These differential equations can be in
one of the following locations:

1. In a separate model definition file
2. Inline in a string
3. In an R expression passed to `rxode2` in brackets `{}`

Each of these can be directly passed into the `rxode2` function.  This
translates the input into compiled C code that can be used to quickly
run ODE solutions.

### Creating an rxode2 model in 3 ways

*This is a simple exercise to create a rxode2 model from more than one model definition*

Change the below R code to:

- Move the model definition to an inline string `ode`
- Once the inline string method is complete, use the `rxode2({})`
  method to specify the ODE inline.

```{r rxode2_model, exercise=TRUE}
## This example creates an ODE with a file tmp.rxode2
## The learnr won't create the file, so change this to be an inline
## string
sink("tmp.rxode2")
cat("d/dt(center)=-kel*center\n")
sink()
model <- rxode2("tmp.rxode2")
print(model)
```


<div id="rxode2_model-hint">
You may wish to take out the `sink` statements and change the `cat` to `ode=`
</div>

### Create an rxode2 model inline

Now change the string-based model to an inline `rxode2` model:

```{r rxode2_inline, exercise=TRUE}
## This is an ode model defined by a string;
## Change it to be defined inline, that is rxode2({})
out = "d/dt(center)=-kel*center"
model <- rxode2(out)
print(model)
```


```{r rxode2_inline-solution}
model <- rxode2({
  d/dt(center) <- -kel*center
})
print(model)
```

## Pre-Quiz - ODE syntax

We briefly touched on the ODE syntax in `rxode2`.  Based on what you
know so far, What is the way you specify ODE equations in `rxode2`?

```{r odeQuiz}
quiz(
  question("What is the syntax for ODEs in rxode2",
    answer("`DXDT_state = -k*state`"),
    answer("`ddt_state = -k*state`"),
    answer("`DADT(1)=-K*A(1)`"),
    answer("`d/dt(state)=-k*state`", correct = TRUE),
    answer("`d/dy(state)=-m*state`")
  ),
  question("Do you have to use `=` for rxode2 ODE assignments?",
    answer("Yes, rxode2 requires `=` for ODE assignment"),
    answer("No, You may use `=` or `<-` to assign ODE equations", correct=TRUE)
  )
)
```

## Simple and time derivative statements

The most fundamental part of an `rxode2` model are simple assignments
and time-derivative assignments:

- **simple** assignments, where the left hand is an identifier (i.e.,
  variable)
- special **time-derivative** assignments, where the left hand
  specifies the change of the amount in the corresponding state
  variable (compartment) with respect to time e.g., `d/dt(depot)`.

Once these variables are assigned in the rxode2 model, rxode2 parses the
model and separates the following types of variables:

- `state` - This gives a list of names the ODE state variables. These
  are in order of how they appear in the model; The first compartment
  is the first compartment listed.
- `params` - This gives a list of names of parameters that need to be
  specified (as a defined parameter, supplied parameter, or
  covariate).  If you use `summary(rxode2Model)` it will show you what
  parameters still need to be defined.
- `lhs` - This gives a parameter that is calculated instead of
  supplied or used as a ODE `state`.

Each of these properties can be accessed with the `$` R operator. For
example if you have an `rxode2` compiled model `mod`, you can access
the state names by `mod$state`

### Exercise - Fix rxode2 model and solve

With the knowledge of how simple statements are defined, you can
figure out what is going wrong with a simple `rxode2` model.

In this exercise, fix the model by adding a parameter value to the
following code assign the missing parameter value to 1

```{r rxParams, exercise=TRUE}
##
mod1 <-rxode2({
    KA <- .294
    CL=18.6
    V2=40.2 # central
    Q =10.5
    V3=297.# peripheral
    Kin=1
    EC50=200
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
})
# You can see what parameters are needed by the summary function:
# summary(mod1)

# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |> et(0,24,length.out=24)) |>
  plot(centr, eff)
```

```{r rxParams-solution}
##
mod1 <-rxode2({
    KA <- .294
    CL=18.6
    V2=40.2 # central
    Q =10.5
    V3=297.# peripheral
    Kin=1
    Kout = 1 # Missing parameter
    EC50=200
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
})
# You can see what parameters are needed by the summary function:
# summary(mod1)

# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |> et(0,24,length.out=24)) |>
  plot(centr, eff)
```


## Special state properties

In addition to the simple statements, and the time-deriviate
statements defining states, there are a few special state properties
that can be added to each state:

- **initial-condition** assignments where the left hand
  specifies the compartment of the initial condition being specified,
  e.g. `depot(0) = 0`

- **bioavailability** is a modeled parameter or constant that changes
  the magnitude of an event/dose by multiplying the dose/event by this
  amount. `f(depot)=`

- **absorption lag time** is a modeled parameter that shifts the time
  of the dose/event by lag time specified; It is specified by `alag`
  on the compartment, like `alag(depot)=`

- **modeled rate** when the event data specify the rate should be
  modeled (`rate=-1`) `rate` for a compartment can specify/model the
  rate; This is specified by `rate(depot) = `

- **modeled duration** when the event data specify the duration should
  be modeled (`rate=-2`), `dur` for a compartment can specify the
  duration of the infusion: `dur(depot) = `

More information about the event specification for the duration/rate
  items can be found in the [rxode2 events
  vignette](https://nlmixr2.github.io/rxode2/articles/rxode2-event-types.html).

**Important:** By default, these special properties can only be defined after a state has been defined by `d/dt(state)=`.

### Exercise - change bioavailibilty and effect initial condition

With the old model:

- Change the bioavailability of the `depot` compartment to `0.5`
- Set the initial `eff` compartment value to `1`.

```{r rxF, exercise=TRUE}
mod1 <-rxode2({
    KA <- .294
    CL=18.6
    V2=40.2 # central
    Q =10.5
    V3=297.# peripheral
    Kin=1
    Kout = 1
    EC50=200
    C2 = centr/V2;
    C3 = peri/V3;
    f(depot)=0.5
    d/dt(depot) =-KA*depot;
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
})
# You can see what parameters are needed by the summary function:
# summary(mod1)

# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |> et(0,24,length.out=24)) |>
  plot(centr, eff)
```

```{r rxF-solution}
##
mod1 <-rxode2({
    KA <- .294
    CL=18.6
    V2=40.2 # central
    Q =10.5
    V3=297.# peripheral
    Kin=1
    Kout = 1 # Missing parameter
    EC50=200
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    f(depot) = 0.5
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    eff(0) = 1
})
# You can see what parameters are needed by the summary function:
# summary(mod1)

# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |> et(0,24,length.out=24)) |>
  plot(centr, eff)
```

## Model Times and Compartment order

We discussed simple assigments, time-derivative statements
`d/dt(state)=...` and special properties you can add to the state
`alag(state)=...`.  There are a few additional convience statements
you can add to your model to change the compartment behavior

- **Compartment declaration statements** `cmt(compartmentName)`. This
  can change the compartment order.  The order of states in the
  `rxode2` model is determined by the order you see `cmt()` statements
  and `d/dt()=` statements. Therefore, to adjust the default dose (ie
  `cmt=1`) you simply need to specify right `cmt()`at the top of the
  `rxode2` model. This is useful for matching event table compartments
  to what you want them to be.

- **modeled times**.  These model times are specified by
  `mtime(var)=time`.  This adds a modeled time to the output dataset.
  This can be set to a numeric value or a model variable.  It is
  useful to add an observation point at the end of an infusion or the
  end of a entero-hepatic recycling model.

### Compartment declaration exercise

The following code explicitly states that `depot` is the first
compartment.

Consider a situation where you want to change the `cmt=1` event to the
`central` compartment to deliver a bolus dose.  Adjust the code below
to change to bolus dose to an iv infusion by switching the default or
first compartment to `centr`


```{r cmt,exercise=TRUE}
mod1 <-rxode2({
  cmt(depot) # the first compartment/default dose is depot
  KA <- .294
  CL=18.6
  V2=40.2 # central
  Q =10.5
  V3=297.# peripheral
  Kin=1
  Kout = 1
  EC50=200
  C2 = centr/V2;
  C3 = peri/V3;
  d/dt(depot) =-KA*depot;
  d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
  d/dt(peri)  =                    Q*C2 - Q*C3;
  d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
  eff(0) = 1
})
# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |> et(0,24,length.out=24)) |>
  plot(centr, eff)
```


```{r cmt-solution}
mod1 <-rxode2({
  cmt(centr) # the first compartment/default dose is depot
  KA <- .294
  CL=18.6
  V2=40.2 # central
  Q =10.5
  V3=297.# peripheral
  Kin=1
  Kout = 1
  EC50=200
  C2 = centr/V2;
  C3 = peri/V3;
  d/dt(depot) =-KA*depot;
  d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
  d/dt(peri)  =                    Q*C2 - Q*C3;
  d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
  eff(0) = 1
})
# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |> et(0,24,length.out=24)) |>
  plot(centr, eff)
```


### Add modeled time

If you notice the output plot above, it seems that the central compartment received 2 different bolus doses.  This is a consequence of the sampling time used.

In this exercise add a modeling time to observe the `12.001`
time-point to see if the plot changes.  Recall to assign a variable to
an observation you wrap it in the `mtime()` function, ie
`mtime(var)=X`

```{r mtime, exercise=TRUE}
mod1 <-rxode2({
  cmt(centr) # the first compartment/default dose is depot
  KA <- .294
  CL=18.6
  V2=40.2 # central
  Q =10.5
  V3=297.# peripheral
  Kin=1
  Kout = 1
  EC50=200
  C2 = centr/V2;
  C3 = peri/V3;
  d/dt(depot) =-KA*depot;
  d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
  d/dt(peri)  =                    Q*C2 - Q*C3;
  d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
  eff(0) = 1
})
# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |>
          et(0,24,length.out=24)) |>
  plot(centr, eff)
```



```{r mtime-solution}
mod1 <-rxode2({
  cmt(centr) # the first compartment/default dose is depot
  KA <- .294
  CL=18.6
  V2=40.2 # central
  Q =10.5
  V3=297.0 # peripheral
  Kin=1
  Kout = 1
  EC50=200
  C2 = centr/V2;
  C3 = peri/V3;
  mtime(mt1)=12.001
  d/dt(depot) =-KA*depot;
  d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
  d/dt(peri)  =                    Q*C2 - Q*C3;
  d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
  eff(0) = 1
})

# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |>
          et(0,24,length.out=24)) |>
  plot(centr, eff)
```

### Mtime fixes the plots

You can now see that the doses seem the same at time zero and the 12
hour time-points.

## If/else clauses

`rxode2` also supports the standard `if`/`else` clauses; A more common
use of `mtime()=` variables is to make sure that a time-point is
captured when a key variable changes. The `rxode2` `if` and `else`
allow time-changes to occur in a model like you would use in a
standard `rxode2` block.  Using the internal variable `t` for time you
can adjust parameters and make sure they show up in the model.

### Exercise 2 -- Make sure that the change point is observed

Change the model below to make sure the 2.5 hour point is observed as
a modeling time-point.

```{r if-else, exercise=TRUE}
library(ggplot2)
mod1 <-rxode2({
  cmt(centr) # the first compartment/default dose is depot
  KA <- .294
  V2=40.2 # central
  Q =10.5
  V3=297.0 # peripheral
  if (t < 2.5){
    CL = 18.6
  } else {
    CL=40
  }
  Kin=1
  Kout = 1
  EC50=200
  C2 = centr/V2;
  C3 = peri/V3;
  mtime(mt1)=12.001
  d/dt(depot) =-KA*depot;
  d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
  d/dt(peri)  =                    Q*C2 - Q*C3;
  d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
  eff(0) = 1
})

# This is a simple code that plots the ODE system.
# D2on't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |>
          et(0,24,length.out=24)) |>
  plot(centr) + scale_y_log10()
```

```{r if-else-solution}
mod1 <-rxode2({
  cmt(centr) # the first compartment/default dose is depot
  KA <- .294
  V2=40.2 # central
  Q =10.5
  V3=297.0 # peripheral
  mtime(ttrans) = 2.5
  if (t < ttrans){
    CL = 18.6
  } else {
    CL=40
  }
  Kin=1
  Kout = 1
  EC50=200
  C2 = centr/V2;
  C3 = peri/V3;
  mtime(mt1)=12.001
  d/dt(depot) =-KA*depot;
  d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
  d/dt(peri)  =                    Q*C2 - Q*C3;
  d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
  eff(0) = 1
})

# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |>
          et(0,24,length.out=24)) |>
  plot(centr) + scale_y_log10()
```

## Suppressing output

One last note about `rxode2`, you can suppress columns in the solved
output (by default `rxode2` outputs all `$state` values and `$lhs`
values).  This is done by modifing either the simple assigment or
time-derivative assignment `d/dt(state)=...` to use `~` instead of the
standard assignment operators `<-` or `=`

### Exercise: Suppressing all output exept C2

Instead of creating a large matrix of output, change the output to
include only `C2`

```{r suppress, exercise=TRUE}
mod1 <-rxode2({
  cmt(centr) # the first compartment/default dose is depot
  KA <- .294
  V2=40.2 # central
  Q =10.5
  V3=297.0 # peripheral
  mtime(ttrans) = 2.5
  if (t < ttrans){
    CL = 18.6
  } else {
    CL=40
  }
  Kin=1
  Kout = 1
  EC50=200
  C2 = centr/V2;
  C3 = peri/V3;
  mtime(mt1)=12.001
  d/dt(depot) =-KA*depot;
  d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
  d/dt(peri)  =                    Q*C2 - Q*C3;
  d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
  eff(0) = 1
})

# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |>
          et(0,24,length.out=24)) |>
  plot()
```



```{r suppress-solution}
mod1 <-rxode2({
  cmt(centr) # the first compartment/default dose is depot
  KA ~ .294
  V2~40.2 # central
  Q ~10.5
  V3~297.0 # peripheral
  mtime(ttrans) ~ 2.5
  if (t < ttrans){
    CL ~ 18.6
  } else {
    CL~40
  }
  Kin~1
  Kout ~ 1
  EC50~200
  C2 <- centr/V2;
  C3 ~ peri/V3;
  mtime(mt1)~12.001
  d/dt(depot) ~ -KA*depot;
  d/dt(centr) ~ KA*depot - CL*C2 - Q*C2 + Q*C3;
  d/dt(peri)  ~                    Q*C2 - Q*C3;
  d/dt(eff)  ~ Kin - Kout*(1-C2/(EC50+C2))*eff;
  eff(0) = 1
})

# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |>
          et(0,24,length.out=24)) |>
  plot()
```

## PK solved systems

PK Solved models are also simple to create. You simply place the
`linCmt()` psuedo-function into your code.  The `linCmt()` function
figures out the type of model to use based on the parameter names
specified.

Most often, pharmacometric models are parameterized in terms of volume
and clearances. Clearances are specified by NONMEM-style names of
`CL`, `Q`, `Q1`, `Q2`, etc. or distributional clearances `CLD`,
`CLD2`. Volumes are specified by Central (`VC` or `V`),
Peripheral/Tissue (`VP`, `VT`).

Another popular parameterization is in terms of micro-constants. rxode2
assumes compartment `1` is the central compartment.  The elimination
constant would be specified by `K`, `Ke` or `Kel`.

The last parameterization possible is using `alpha` and `V` and/or
`A`/`B`/`C`

To see some tables of tested parameters for `linCmt()` you can go to
the [`linCmt()`
vignette](https://nlmixr2.github.io/rxode2/articles/rxode2-model-types.html#solved-compartment-models)

Once the `linCmt()` sleuthing is complete, the `1`, `2` or `3`
compartment model solution is used as the value of `linCmt()`.

The compartments where you can dose in a linear solved system are
`depot` (for oral absorption models) and `central`.  Without any
additional ODEs, these compartments are numbered `depot=1` and
`central=2`.  You can add some properties to these compartments like
bioavailibilty changes `f(depot)=0.5` changes the bioavailability to
0.5.

When the absorption constant `ka` is missing, you may only dose to the
`central` compartment.  Without any additional ODEs the compartment
number is `central=1`.


### Exercise 1: Change a one-compartment bolus model

The model below is a 2 compartment IV model; Change the model to be a 2-compartment
oral absorption model by adding a `Ka` parameter.

```{r lincmt, exercise=TRUE}
mod1 <-rxode2({
  V2~40.2 # central
  Q ~10.5
  V3~297.0 # peripheral
  CL ~ 18.6
  C2 = linCmt()
})

# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24) |>
          et(0,24,length.out=24)) |>
  plot()
```


<div id="lincmt-hint">
Put in the `KA = .294` parameter to change the model
</div>


### Exercise 2: Mixing ODE and PK solved systems

One of the nice features of `rxode2` is it can also mix `linCmt()`
solutions and ODE solutions.

As an exercise change the below model to drop the `depot` and
`central` ODE definitions:

```{r linCmt-mix, exercise=TRUE}
mod1 <-rxode2({
  KA ~ .294
  V2~40.2 # central
  Q ~10.5
  V3~297.0 # peripheral
  CL ~ 18.6
  Kin~1
  Kout ~ 1
  EC50~200
  C2 <- central/V2;
  C3 ~ peri/V3;
  mtime(mt1)~12.001
  d/dt(depot) ~ -KA*depot;
  d/dt(central) ~ KA*depot - CL*C2 - Q*C2 + Q*C3;
  d/dt(peri)  ~                    Q*C2 - Q*C3;
  d/dt(eff) = Kin - Kout*(1-C2/(EC50+C2))*eff;
  eff(0) = 1
})

# This is a simple code that plots the ODE system.
# Don't worry about `et` if you don't understand it now
# It will be covered in another tutorial
rxSolve(mod1,et(amt=10000, ii=12, until=24, cmt='depot') |>
          et(0,24,length.out=24)) |>
  plot()
```

<div id="linCmt-mix-hint">
Simply drop the `d/dt(central)`, `d/dt(depot)` and `C3` lines; Change `C2=linCmt()`
</div>

### Exercise 3: Compartment number with mixed `linCmt`/`ODE` models

You may have noticed that the `et` dosed to `cmt='depot'`.  This is so
that the solutions are the same between the ODE model and the solved system.

The compartment numbers have changed when mixing `ODE`s and `linCmt()` models.

In this exercise use `print(mod1)` to get an idea of what the
compartment numbers are for the model.

```{r linCmt-cmt, exercise=TRUE}
mod1 <-rxode2({
  KA ~ .294
  V2~40.2 # central
  Q ~10.5
  V3~297.0 # peripheral
  CL ~ 18.6
  Kin~1
  Kout ~ 1
  EC50~200
  C2 <- linCmt()
  d/dt(eff) = Kin - Kout*(1-C2/(EC50+C2))*eff;
  eff(0) = 1
})
```

### Mixing ODE models and `linCmt()` models add solved systems at the end of the model

You can see the `linCmt()` model has a variable `$stateExtra` which
lists the extra states.  In this case the `depot` compartment is the
`2`nd compartment since it is after the state `eff` compartment.  The
extra states always are numbered after the ODE states have been
specified.

**Note:** `rxode2` cannot currently use `cmt()` on the `depot` and
`central` compartments for solved systems to change their compartment
number.


## Jacobian-derivatives

There is one more type of syntax that you may see in `rxode2` models, the **Jacobain** syntax; ie:

- special **Jacobian-derivative** assignments, where the left hand
  specifies the change in the compartment ode with respect to a
  variable.  For example, if `d/dt(y) = dy`, then a Jacobian for this
  compartment can be specified as `df(y)/dy(dy) = 1`.

### Seeing Jacobian derivatives

The Following code is an example of Jacobain syntax:

```{r jac, exercise=TRUE}
van <- rxode2({
  d/dt(dy)=y
  d/dt(y)=mu*(1-dy^2)*y-dy
})

van <- rxode2({
    d/dt(dy) = y
    d/dt(y) = -dy + y * mu * (1 - dy^2)
    df(dy)/dy(dy) = 0
    df(y)/dy(dy) = -1 - 2 * y * mu * dy
    df(dy)/dy(y) = 1
    df(y)/dy(y) = mu * (1 - dy^2)
    df(dy)/dy(mu) = 0
    df(y)/dy(mu) = y * (1 - dy^2)
})

## The lsoda (not liblsoda) method is the only solving method that
## uses the Jacobian specification currently
```

Note if `rxode2` is setup completely, you can automatically calculate
the Jacoabian by `rxode2({}, calcJac=TRUE)`

## Summary

`rxode2` model syntax comprises of:

- **simple** assignments, where the left hand is an identifier (i.e.,
  variable)
- special **time-derivative** assignments, where the left hand
  specifies the change of the amount in the corresponding state
  variable (compartment) with respect to time e.g., `d/dt(depot)`:
- special **initial-condition** assignments where the left hand
  specifies the compartment of the initial condition being specified,
  e.g. `depot(0) = 0`
- special model event changes including **bioavailability**
  (`f(depot)=1`), **lag time** (`lag(depot)=0`), **modeled rate**
  (`rate(depot)=2`) and **modeled duration** (`dur(depot)=2`). An
  example of these model features and the event specification for the
  modeled infusions the rxode2 data specification is found in [rxode2
  events vignette](https://nlmixr2.github.io/rxode2/articles/rxode2-event-types.html).
- **Compartment declaration statements**, which can change the default
  dosing compartment and the assumed compartment number(s) as well as
  add extra compartment names at the end (useful for multiple-endpoint
  nlmixr models); These are specified by `cmt(compartmentName)`
- special **change point syntax, or model times**.  These model times are
  specified by `mtime(var)=time`
- special **Jacobian-derivative** assignments, where the left hand
  specifies the change in the compartment ode with respect to a
  variable.  For example, if `d/dt(y) = dy`, then a Jacobian for this
  compartment can be specified as `df(y)/dy(dy) = 1`.  There may be
  some advantage to obtaining the solution or specifying the Jacobian
  for very stiff ODE systems.  However, for the few stiff systems we
  tried with LSODA, this actually slightly slowed down the solving.

Note that assignment can be done by `=` or `<-`.

Additionally, assignment can be done with the `~` operator, which
causes rxode2 to use the variable/expression while solving but suppress
output to either the matrix or data-frame returned in R.

Also note `linCmt()` can be used to specify linear compartmental models.

---
title: "rxode2 ODE solving syntax"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{rxode2 ODE solving syntax}
  %\VignetteEngine{knitr::rmarkdown_notangle}
  %\VignetteEncoding{UTF-8}
---


```{r, echo=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>")
library(rxode2)
```

## Introduction

This briefly describes the syntax used to define models
that `rxode2` will translate into R-callable compiled code. It also
describes the communication of variables between `R` and the
`rxode2` modeling specification.


## Creating rxode2 models

```{r child='../man/rmdhunks/rxode2-create-models.Rmd'}
```

## Syntax

```{r child='../man/rmdhunks/rxode2-syntax-hunk.Rmd'}
```


## Note

The ODE specification mini-language is parsed with the help of the
open source tool \code{dparser}, Plevyak (2015).

## Example model

Below is a commented example to quickly show the capabilities of
`rxode2` syntax.

## Example

``` r
f <- function() {
  ini({

  })
  model({
    # An rxode2 model specification (this line is a comment).

    if(comed==0) {  # concomitant medication (con-med)?
      F <- 1.0     # full bioavailability w.o. con-med
    } else {
      F <- 0.80    # 20% reduced bioavailability
    }

    C2 <- centr/V2  # concentration in the central compartment
    C3 <- peri/V3   # concentration in the peripheral compartment

    # ODE describing the PK and PD

    d/dt(depot) <- -KA*depot
    d/dt(centr) <- F*KA*depot - CL*C2 - Q*C2 + Q*C3
    d/dt(peri)  <-                      Q*C2 - Q*C3
    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
    eff(0)      <- 1
  })
}
```



## Interface and data handling between R and the generated C code

Users specify which variables are the dynamic system's state variables
via the `d/dt(identifier)` operator as part of the model specification,
and which are model parameters via the `params=` argument in `rxode2`
`solve()` method:

``` r
m1 <- rxode2(model = ode, modName = "m1")

# model parameters -- a named vector is required
theta <-
   c(KA=0.29, CL=18.6, V2=40.2, Q=10.5, V3=297, Kin=1, Kout=1, EC50=200)

# state variables and their amounts at time 0 (the use of names is
# encouraged, but not required)
inits <- c(depot=0, centr=0, peri=0, eff=1)

# qd1 is an eventTable specification with a set of dosing and sampling
# records (code not shown here)

solve(theta, event = qd1, inits = inits)
```


The values of these variables at pre-specified time points are saved
during model fitting/integration and returned as part of the fitted
values (see the function `et()`, to define a set of time points when
to capture the values of these variables) and returned as part of the
modeling output.

The ODE specification mini-language is parsed with the help of the
open source tool *DParser*, Plevyak (2015).

---
title: "rxode2 events"
output:
  learnr::tutorial:
    progressive: true
runtime: shiny_prerendered
description: >
  This covers basic events supported for `rxode2` models with an example rxode2 model
---

```{r setup, include=FALSE}
library(learnr)
library(rxode2)
library(ggplot2)
## Adapted from gradethis::grade_result for now
## Wait for gradethis to make it to CRAN
## Model from rxode2 tutorial
m1 <-rxode2({
    KA=2.94E-01;
    CL=1.86E+01;
    V2=4.02E+01;
    Q=1.05E+01;
    V3=2.97E+02;
    Kin=1;
    Kout=1;
    EC50=200;
    ## Added modeled bioavaiblity, duration and rate
    fdepot = 1;
    durDepot = 8;
    rateDepot = 1250;
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    f(depot) = fdepot
    dur(depot) = durDepot
    rate(depot) = rateDepot
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    eff(0) = 1
});

```
From Rstudio 1.3+ tutorial pane you can use the expand button
(![](images/expandTutorial.png)) to navigate tutorial sections.

## Creating events

### Model for Event discussion

For these models, we can illustrate by using the model shared in the
rxode2 tutorial; Compile it so we can continue:

```{r m1, exercise=TRUE}
## Model from rxode2 tutorial
m1 <-rxode2({
    KA=2.94E-01;
    CL=1.86E+01;
    V2=4.02E+01;
    Q=1.05E+01;
    V3=2.97E+02;
    Kin=1;
    Kout=1;
    EC50=200;
    ## Added modeled bioavaiblity, duration and rate
    fdepot = 1;
    durDepot = 8;
    rateDepot = 1250;
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    f(depot) = fdepot
    dur(depot) = durDepot
    rate(depot) = rateDepot
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
    eff(0) = 1
});

print(summary(m1))
```
### What is an event table

An event table in rxode2 is a specialized data frame that acts as a
container for all of rxode2's events and observation times.  While not
required to solve, the event table allows easy generation of a data
table in the format that `rxode2` expects.

To create an rxode2 event table you may use the code `eventTable()`,
`et()`, or even create your own data frame with the right event
information contained in it.  This is closely related to the [types of
events that rxode2 supports](rxode2-event-types.html).

Once the event table has been created you can add
sampling/observations or doses by piping or direct access.

This is a short table of arguments the two main functions (`et()` and
`add.dosing()`:

| add.dosing()    | et() | Description                         |
|-----------------|------|-------------------------------------|
| dose            | amt  | Dose/Rate/Duration amount           |
| nbr.doses       | addl | Additional doses or number of doses |
| dosing.interval | ii   | Dosing Interval                     |
| dosing.to       | cmt  | Dosing Compartment                  |
| rate            | rate | Infusion rate                       |
| start.time      | time | Dosing start time                   |
|                 | dur  | Infusion Duration                   |


Sampling times can be added with `add.sampling(` sampling times `)` or
`et(` sampling times `)`.  [Dosing intervals and sampling
windows](#add-doses-and-samples-within-a-sampling-window) are also
supported.

### Exercise: Create an empty table  and add a 100 mg dose

```{r eventTable, exercise=TRUE}
e <- et()## you can add the units by `amount.units` and `time.units`

print(e)
```

<div id="eventTable-hint">
You may add a dose by `$add.dosing(dose=...)` or `e |> et(amt=...)`
</div>

### Solving the system

With the model built from `rxode2`, and the event table built above, we
can solve and plot the scenario; Run the code and see if you notice
anything:

```{r, preSample, exercise=TRUE}
ev <- et() |>
    et(dose=5000, dosing.interval=12, until=48)

rxSolve(m1, ev) |> plot(C2)
```

### Quiz 1

Lets see if you can recall everything so far:

```{r unitsQuiz, echo=FALSE}
quiz(
  question("What function(s) can create event table for solving in rxode2?",
    answer("`data.frame()`", correct=TRUE),
    answer("`eventTable()`", correct=TRUE),
    answer("`et()`", correct=TRUE),
    answer("ev()", correct = FALSE),
    answer("add.sampling()", correct=FALSE),
    answer("add.dosing()", correct=FALSE)
  ),
  question("What did you notice about the event table plot?",
           answer("There was a missing dose at the end of the plot"),
           answer("The samples times were chose by rxode2", correct=TRUE))
)
```

## Adding sampling

In the last example you may have noticed that the sample times were
chosen by `rxode2`; This is because the event table that was supplied
to the solving routine only included dosing information.  If you add
observations to the event table you can control the observations of
each event.

Recall `rxode2` you can use `add.sampling(` samples `)` or `et(` samples `)` to add
observations to an rxode2 event table.

### Exercise, add observations for the first 24 hours

As an exercise, add an observation for the first 24 hours every 0.5 hours:

```{r addObs, exercise=TRUE}
ev <- et() |>
    et(dose=5000, ii=12, until=48)

rxSolve(m1, ev) |> plot(C2)
```

```{r addObs-solution}
ev <- et() |>
    et(dose=5000, ii=12, until=48) |>
    et(seq(0, 24, by=0.5))

rxSolve(m1, ev) |> plot(C2)
```

### `et` convenience notation

If you are like me, you may have used `seq()` as your solution to
adding observations to the event table.  For the function `et()` you
can simply drop the `seq()` and enclose everything in the `et()`
function.  For `et(seq(0,24,by=0.5))` this can be written as
`et(0,24,by=0.5)`.

As an exercise change the event table below to reflect this convenience:


```{r addObsEt, exercise=TRUE}
ev <- et() |>
    et(dose=5000, ii=12, until=48) |>
    et(seq(0, 24, by=0.5))

rxSolve(m1, ev) |> plot(C2)
```

```{r addObsEt-solution}
ev <- et() |>
    et(dose=5000, ii=12, until=48) |>
    et(seq(0, 24, by=0.5))

rxSolve(m1, ev) |> plot(C2)
```


## Event classification

Here are the legal entries to a rxode2 event table (which can be a
`tibble`, `data.frame`, `data.table` or `rxode2` event table):

| Column | Meaning               | Notes                                                                        |
|-----------|-----------------------|------------------------------------------------------------------------------|
| id        | Individual identifier | Will convert to a factor/integer (sorted)
| time      | Individual time       | For each ID must be ascending and non-negative                               |
| amt       | dose amount           | Positive for doses zero/NA for observations                                  |
| rate      | infusion rate         | When specified the infusion duration will be dur=amt/rate                    |
|           |                       | rate = -1, rate modeled; rate = -2, duration modeled                         |
| dur       | infusion duration     | When specified the infusion rate will be rate = amt/dur                      |
| evid      | event ID              | 0=Observation; 1=Dose; 2=Other; 3=Reset; 4=Reset+Dose; 5=Replace; 6=Multiply |
| cmt       | Compartment           | Represents compartment #/name for dose/observation                           |
| ss        | Steady State Flag     | 0 = non-steady-state; 1=steady state; 2=steady state +prior states           |
| ii        | Inter-dose Interval   | Time between doses.                                                          |
| addl      | # of additional doses | Number of doses like the current dose.                                       |

### Quiz -- Legal event names

What are some of the legal event names:

```{r eventColQuiz,echo=FALSE}
quiz(
  question("What are legal names for event columns:",
           answer("steadyState", correct=FALSE),
           answer("ii", correct=TRUE),
           answer("dosingInterval", correct=FALSE),
           answer("rate", correct=TRUE),
           answer("dur", correct = TRUE),
           answer("ss", correct=TRUE)))
```

### Event differences

If you are familiar with `NONMEM` or `Monolix`, you may notice that
`rxode2` event tables are very similar to the `NONMEM`/`Monolix`
dataset convention with these exceptions:

 - The compartment data item (`cmt`) can be a string/factor with
   compartment names
   - You may turn off a compartment with a negative compartment number
     or "-cmt" where cmt is the compartment name.
   - The compartment data item (`cmt`) can still be a number, the
     number of the compartment is defined by the appearance of the
     compartment name in the model. This can be tedious to count, so
     you can specify compartment numbers easier by using the
     `cmt(cmtName)` at the beginning of the model.
 - An additional column, `dur` can specify the duration of infusions;
    - Bioavailability changes will change the rate of infusion since
      `dur`/`amt` are fixed in the input data.
    - Similarly, when specifying `rate`/`amt` for an infusion, the
      bioavailability will change the infusion duration since
      `rate`/`amt` are fixed in the input data.
 - Some infrequent NONMEM columns are not supported: `pcmt`, `call`.
 - Additional events are supported:
   - `evid=5` or replace event; This replaces the value of a
       compartment with the value specified in the `amt` column. This
       is equivalent to `deSolve`=`replace`.
   - `evid=6` or multiply event; This multiplies the value in the
     compartment with the value specified by the `amt` column.  This
     is equivalent to `deSolve`=`multiply`.

### Event ID (evid quiz)

Now that we covered all the event types, quick quiz on what they all
mean:

```{r quizDiff, echo=FALSE}
quiz(
  question("What does evid=1 mean?",
           answer("An observation event", correct=FALSE),
           answer("A reset event", correct=FALSE),
           answer("A reset then dose", correct=FALSE),
           answer("A multiplication event", correct=FALSE),
           answer("An addivite event (adding say one dose)", correct=TRUE),
           answer("Other type of event", correct=FALSE),
           answer("Replacement event", correct=FALSE),
           answer("Nothing; illegal evid", correct=FALSE)),
  question("What does evid=0 mean?",
           answer("An observation event", correct=TRUE),
           answer("A reset event", correct=FALSE),
           answer("A reset then dose", correct=FALSE),
           answer("A multiplication event", correct=FALSE),
           answer("An addivite event (adding say one dose)", correct=FALSE),
           answer("Other type of event", correct=FALSE),
           answer("Replacement event", correct=FALSE),
           answer("Nothing; illegal evid", correct=FALSE)),
  question("What does evid=2 mean?",
           answer("An observation event", correct=FALSE),
           answer("A reset event", correct=FALSE),
           answer("A reset then dose", correct=FALSE),
           answer("A multiplication event", correct=FALSE),
           answer("An addivite event (adding say one dose)", correct=FALSE),
           answer("Other type of event", correct=TRUE),
           answer("Replacement event", correct=FALSE),
           answer("Nothing; illegal evid", correct=FALSE)),
  question("What does evid=3 mean?",
           answer("An observation event", correct=FALSE),
           answer("A reset event", correct=TRUE),
           answer("A reset then dose", correct=FALSE),
           answer("A multiplication event", correct=FALSE),
           answer("An addivite event (adding say one dose)", correct=FALSE),
           answer("Other type of event", correct=FALSE),
           answer("Replacement event", correct=FALSE),
           answer("Nothing; illegal evid", correct=FALSE)),
  question("What does evid=4 mean?",
           answer("An observation event", correct=FALSE),
           answer("A reset event", correct=FALSE),
           answer("A reset then dose", correct=TRUE),
           answer("A multiplication event", correct=FALSE),
           answer("An addivite event (adding say one dose)", correct=FALSE),
           answer("Other type of event", correct=FALSE),
           answer("Replacement event", correct=FALSE),
           answer("Nothing; illegal evid", correct=FALSE)),
  question("What does evid=5 mean?",
           answer("An observation event", correct=FALSE),
           answer("A reset event", correct=FALSE),
           answer("A reset then dose", correct=FALSE),
           answer("A multiplication event", correct=FALSE),
           answer("An addivite event (adding say one dose)", correct=FALSE),
           answer("Other type of event", correct=FALSE),
           answer("Replacement event", correct=TRUE),
           answer("Nothing; illegal evid", correct=FALSE)),
  question("What does evid=6 mean?",
           answer("An observation event", correct=FALSE),
           answer("A reset event", correct=FALSE),
           answer("A reset then dose", correct=FALSE),
           answer("A multiplication event", correct=TRUE),
           answer("An addivite event (adding say one dose)", correct=FALSE),
           answer("Other type of event", correct=FALSE),
           answer("Replacement event", correct=FALSE),
           answer("Nothing; illegal evid", correct=FALSE)),
  question("What does evid=7 mean?",
           answer("An observation event", correct=FALSE),
           answer("A reset event", correct=FALSE),
           answer("A reset then dose", correct=FALSE),
           answer("A multiplication event", correct=FALSE),
           answer("An addivite event (adding say one dose)", correct=FALSE),
           answer("Other type of event", correct=FALSE),
           answer("Replacement event", correct=FALSE),
           answer("Nothing; illegal evid", correct=TRUE))
  )
```

### Remembering event types

One of the challenges of flexible event types, is remembering the
event identifiers; If you ever want to figure out what an event type
is, it is print out in an `rxode2` event table.

In this exercise, print out event types `1`  to `6` by creating an event table:

```{r etPrintout, exercise=TRUE}
e <- et() |>
    et(time=1, amt=1, evid=1) |>
    et(time=2, amt=2, evid=2) |>
    et(time=3, amt=3, evid=3) |>
    et(time=4, amt=4, evid=4)

## add evid=5 through 6 to the above event table

print(e)
```

```{r etPrintout-solution}
e <- et() |>
    et(time=1, amt=1, evid=1) |>
    et(time=2, amt=2, evid=2) |>
    et(time=3, amt=3, evid=3) |>
    et(time=4, amt=4, evid=4) |>
    et(time=5, amt=5, evid=5) |>
    et(time=6, amt=6, evid=6)
## add evid=5 through 6 to the above event table

print(e)
```

### Convert to data.frame

Notice that the `evid` shows `#:Description`, this is only for
display; When you convert the event table to a data.frame, it becomes
a number.

Use `as.data.frame` to convert the event table `e` to an data.frame and print-out the results:

```{r etPrintoutDf, exercise=TRUE}
e <- et() |>
    et(time=1, amt=1, evid=1) |>
    et(time=2, amt=2, evid=2) |>
    et(time=3, amt=3, evid=3) |>
    et(time=4, amt=4, evid=4) |>
    et(time=6, amt=6, evid=6)

print(e)
```


```{r etPrintoutDf-solution}
e <- et() |>
    et(time=1, amt=1, evid=1) |>
    et(time=2, amt=2, evid=2) |>
    et(time=3, amt=3, evid=3) |>
    et(time=4, amt=4, evid=4) |>
    et(time=6, amt=6, evid=6) |>
    as.data.frame

print(e)

```

## Bolus/Additive Doses

A bolus dose is the default type of dose in rxode2 and only requires
the `amt`/`dose`.  Note that this uses the convenience function `et()`
described in the [rxode2 event tables](rxode2-event-table.html)

```{r, bolus, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12,until=24) |>
    et(seq(0, 24, length.out=100))

print(ev)

rxSolve(m1, ev) |> plot(C2) +
    xlab("Time")
```
## Infusions

There are a few different type of infusions that rxode2 supports:

- Constant Rate Infusion (`rate`)
- Constant Duration Infusion (`dur`)
- Estimated Rate of Infusion
- Estimated Duration of Infusion


### Constant Infusion (in terms of duration and rate)

The next type of event is an infusion; There are two ways to specify
an infusion; The first is the `dur` keyword.

An example of this is:

```{r, durInf, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12,until=24, dur=8) |>
    et(seq(0, 24, length.out=100))

print(ev)

rxSolve(m1, ev) |> plot(depot, C2) +
    xlab("Time")
```


It can be also specified by the `rate` component:

```{r, rateInf, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12,until=24, rate=10000/8) |>
    et(seq(0, 24, length.out=100))

print(ev)

rxSolve(m1, ev) |> plot(depot, C2) +
    xlab("Time")
```

### Did you spot a difference in the event tables?

If you have a keen eye you may have noticed the two event tables.  The
first had a `rate` column and the second has a `dur` column.  Keep
this in mind as we continue our infusion discussion.

### What does bioavailability do to the system?

Now that we have infusion, a natural question is how does
bioavailability affect the `rxode2` system?

Let check; Change the bioavailability to `fdepot=0.25` when the event
table is parameterized in terms of `rate` to see what happens (you may
have to run the code then change it to see the difference):

```{r, rateFinf, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12,until=24, rate=10000/8) |>
    et(seq(0, 24, length.out=100))

## Change the fdepot to 0.25
rxSolve(m1, ev, c(fdepot=1)) |> plot(depot, C2) +
    xlab("Time")
```

### Bioavailability change for durations

In the case of modeling `rate`, a bioavailability decrease,
decreases the infusion duration, as in `NONMEM`.

Conversely, a bioavailability increase should therefore increase the infusion
duration; A simple check could be to change the `fdepot` to `2`:

```{r, rateFinf2, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12,until=24, rate=10000/8) |>
    et(seq(0, 24, length.out=100))

## Change the fdepot to 2
## If you cannot clearly see the effect, continue to the next section
rxSolve(m1, ev, c(fdepot=1)) |> plot(depot, C2) +
    xlab("Time")
```

### Rationale for infusion changes when `rate` is specified

The rationale for this behavior is that the `rate` and `amt` are
specified by the event table, so the only thing that can change with a
bioavailability increase is the duration of the infusion.

However a bioavailability change may not affect the infusion duration of an IV
infusion, so this thinking about IV infusion bioavilability has led to
confusion in the past.

This is the reason why parameterizing using the duration can be useful.

### Exercise -- parameterize using duration

If you specify the `amt` and `dur` components in the event table,
bioavailability changes affect the `rate` of infusion.

Note that if there is no bioavialability change the solution is the
same; Change the bioavilability to see the effect:

```{r durInf2, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12,until=24, dur=8) |>
    et(0, 24, length.out=100)

## If you cannot see a clear difference, go to the next section
rxSolve(m1, ev, c(fdepot=1)) |> plot(depot, C2) +
    xlab("Time")
```

### Duration effect

Did you figure out the effect?  It can be hard to notice; the `y`-axis
changed; To more clearly see the effect you can look at the
side-by-side comparison; Run the following code to see the difference:

```{r durSideBySide, exercise=TRUE}
library(ggplot2)
library(gridExtra)

ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12,until=24, dur=8) |>
    et(0, 24, length.out=100)

p1 <- rxSolve(m1, ev, c(fdepot=1.25)) |> plot(depot) +
    xlab("Time") + ylim(0,5000)

p2 <- rxSolve(m1, ev, c(fdepot=0.25)) |> plot(depot) +
    xlab("Time")+ ylim(0,5000)

grid.arrange(p1,p2, nrow=1)
```

Now you can see the clear changes in infusion `rate` when `dur` and
`amt` is constant.  This is more likely what would happen in clinical
practice; The duration of infusion is the same, but the rate is
changed.

### Modeled rate

You may also model rate.  This is equivalent to NONMEM's `rate=-1` and
is how rxode2's event table specifies the data item as well. You can
also use `rate=model` as a mnemonic.  Note that you chan change the
`rateDepot` to change the rate of the infusion.  Also `fdepot` in this
case changes the duration of infusion.

```{r modelRate, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12,until=24, rate=model) |>
    et(seq(0, 24, length.out=100))

print(ev)

rxSolve(m1, ev, c(rateDepot=10000/3, fdepot=1)) |> plot(depot, C2) +
    xlab("Time")
```

Note that the event table gives you clues about the type of event
`rate=-1` is by the text.  As before, this is simply a displayed
output and the actual value is `-1`


### Modeled duration

You can model the duration, which is equivalent to NONMEM's
`rate=-2`. As a mnemonic you can use the `dur=model` instead of
`rate=-2`


As an exercise you can change the duration of infusion (`durDepot`)
and the `fdepot`.  The `durDepot` will change the duration of the
infusion; The `fdepot` will change the infusion rate:

```{r modelDur, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12,until=24, dur=model) |>
    et(seq(0, 24, length.out=100))

print(ev)

rxSolve(m1, ev, c(durDepot=7, fdepot =1)) |> plot(depot, C2) +
    xlab("Time")
```

## Steady State dosing

These doses are solved until a steady state is reached with a constant
inter-dose interval.  There are three types of steady state doses that
`rxode2` supports:

- Simple steady state solutions
- Complex steady state solutions
- Infusion steady state solutions

The `rxode2` event table needs to be setup specially for each type of
steady state record.

### Simple Steady State dosing

Simple steady-state dosing can be added by using the `ss=1` flag
coupled with the inter-dose interval flag `ii`.  This is interpreted
as the same bolus dose or infusion being applied separated by a dosing
interval `ii` until the steady state is reached.

This particular example has a steady-state bolus dose separated by 12
hours; You may run it and see how it behaves.

As an exercise you can change it to a 30 minute duration infusion with
`dur=0.5`; You can also explore the effect of the inter-dose interval `ii`
on the steady state concentrations

```{r ssBolus, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12, ss=1) |>
    et(seq(0, 24, length.out=100))

print(ev)

rxSolve(m1, ev) |> plot(depot, C2)
```


```{r ssBolus-solution}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12, ss=1, dur=0.5) |>
    et(seq(0, 24, length.out=100))

rxSolve(m1, ev) |> plot(depot, C2)
```

### Simple steady state dosing observations

You probably observed the steady state solution did not start at zero,
but was at a higher concentration, as expected.  This simple way of
specifying steady state works for both bolus doses and infusion doses;
It also assumes a constant interval between doses;

But, what if the interval or dose was not constant; For example if you
were dosing `100` mg in the morning and `150` mg in the evening.  How
would you get the steady state solution?

The answer is adding a bit of complexity to the steady state solutions

### Complex steady state

By using the `ss=2` flag, you can use the super-positioning principle
in linear kinetics to get steady state nonstandard dosing
(i.e. morning 100 mg vs evening 150 mg). This is done by:

 - Saving all the state values
 - Resetting all the states and solving the system to steady state
 - Adding back all the prior state values

This takes a full dose cycle to reach a true steady state solution;

As an exercise, run the example below and then change the evening dose to `200` mg:

```{r complexSS, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=24, ss=1) |>
    et(time=12, amt=15000, ii=24, ss=2) |>
    et(time=24, amt=10000, ii=24, addl=3) |>
    et(time=36, amt=15000, ii=24, addl=3) |>
    et(seq(0, 64, length.out=500))

library(ggplot2)

rxSolve(m1, ev,maxsteps=10000) |> plot(C2) +
    annotate("rect", xmin=0, xmax=24, ymin=-Inf, ymax=Inf, alpha=0.2) +
    annotate("text", x=12.5, y=7, label="Initial Steady State Period") +
    annotate("text", x=44,   y=7, label="Steady State AM/PM dosing")

```

### Infusion steady states

The last type of steady state that rxode2 supports is steady-state
constant infusion rate.  This can be specified the same way as NONMEM,
that is:

- No inter-dose interval `ii`=`0`
- A steady state dose, ie `ss`=`1`
- Either a positive rate (`rate`>0) or a estimated rate `rate`=`-1`.
- A zero dose, ie `amt`=`0`
- Once the steady-state constant infusion is achieved, the infusion is
  turned off when using this record, just like NONMEM.

Note that `rate`=`-2` where we model the duration of infusion doesn't
make much sense since we are solving the infusion until steady state.
The duration is specified by the steady state solution.

Also note that bioavailability changes on this steady state infusion
also do not make sense because they neither change the `rate` or the
duration of the steady state infusion.  Hence modeled bioavailability
on this type of dosing event is ignored.

As an exercise you can run this example and then change the rate to
see how it affects the steady state solution; The steady state
solution (on the top) should be the same as the long term infusion (on
the bottom)

```{r infSSzero, exercise=TRUE}
library(gridExtra)
ev <- et(timeUnits="hr") |>
    ## The infusion amount is unknown, rather we are using as much as
    ## needed to get to steady state so `ss=0`
    et(amt=0, ss=1,rate=10000/8)

p1 <- rxSolve(m1, ev) |> plot(C2, eff)


ev <- et(timeUnits="hr") |>
    ## Use a very large dose so that the infusion reaches steady state
    et(amt=200000, rate=10000/8) |>
    et(0, 250, length.out=1000)

p2 <- rxSolve(m1, ev) |> plot(C2, eff)


grid.arrange(p1,p2, ncol=1)

```

## Resetting or turning off Compartments

rxode2 supports two different types of events that may be useful:

- Reset events
- Turning off compartments

### Resetting events

Reset events are implemented by `evid=3` or `evid=reset`, for reset
and `evid=4` for reset and dose.  The exercise below resets the dose
`6 hours` post dose.

Run the following example to see what happens 6 hours post-dose:

```{r resetExample, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12, addl=3) |>
    et(time=6, evid=reset) |> ## this is a mnemonic for evid=3
    et(seq(0, 24, length.out=100))

rxSolve(m1, ev) |> plot(depot,C2, eff)
```

### Instead of reset, what about reset and dose?

You can see all the compartments are reset to their initial values at
`time=6`.  The next dose start the dosing cycle over.

What if you used the reset and dose event `evid=4` of `amt=1000`?

Adjust the last example to see what happens:

```{r resetDoseExample, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12, addl=3) |>
    et(time=6, evid=reset) |> ## Change this to evid=4
    et(seq(0, 24, length.out=100))

rxSolve(m1, ev) |> plot(depot,C2, eff)
```

```{r resetDoseExample-solution}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12, addl=3) |>
    et(time=6, evid=4) |> ## Change this to evid=4
    et(seq(0, 24, length.out=100))

rxSolve(m1, ev) |> plot(depot,C2, eff)
```


### Turning off compartments

In this case, the whole system is reset and the dose is given to the
depot compartment.  This related but not the same as turning off a
compartment.

What is the difference between "resetting" the system and turning off
compartments?

To find out try turning off the `depot` compartment; This is done by
changing the reset event to `evid=2` and turning off the compartment.
This can be done by dosing either to the negative compartment, ie `-#` or  dosing to the `-cmtName` compartment.  In this case it can be `-1` or `-depot`

As an exercise change the reset dose to turn off the depot compartment:

```{r depotOff, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12, addl=3) |>
    et(time=6, evid=4) |> ## turn off the compartment with evid=2 and cmt=?
    et(seq(0, 24, length.out=100))

rxSolve(m1, ev) |> plot(depot,C2, eff)
```

```{r depotOff-solution}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12, addl=3) |>
    et(time=6, cmt="-depot", evid=2) |>
    et(seq(0, 24, length.out=100))

rxSolve(m1, ev) |> plot(depot,C2, eff)
```

### Turning the compartment back on?

In this case, the depot is turned off; The depot compartment
concentrations  are set to the initial values but the other compartment
concentrations/levels are not reset.

So what do you need to do to turn the compartment back on? Add another
dose to the `depot` is administered the depot compartment is turned
back on. This dose can be even a bolus of `0` just to turn the
compartment back on.

**Note:** a dose to a compartment only turns back on the compartment
that was dosed.  Hence if you turn off the effect compartment, it
continues to be off after another dose to the depot.

As an exercise, run the model without turning on the effect
compartment; Then as an exercise turn the effect compartment back on
at `time=12`:

```{r cmtOn, exercise=TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12, addl=3) |>
    et(time=6, cmt="-eff", evid=2) |>
    et(seq(0, 24, length.out=100))

rxSolve(m1, ev) |> plot(depot,C2, eff)
```

```{r cmtOn-solution}
ev <- et(timeUnits="hr") |>
    et(amt=10000, ii=12, addl=3) |>
    et(time=6, cmt="-eff", evid=2) |>
    et(time=12,cmt="eff",evid=2) |>
    et(seq(0, 24, length.out=100))

rxSolve(m1, ev) |> plot(depot,C2, eff)
```

### Result of turning on the effect compartment again

Once the effect compartment was turned back on, you can see a sharp
increase in effect since the amount in the `C2` starts to equilabrate
with the effect compartment

## Combining/Expanding event tables

In this section of the events tutorial we will cover:

- Multi-subject event table creation/expansion

- Combining event tables

- Sequencing event tables

- Repeating event tables

### Expanding the event table into a multi-subject event table

You may want to expand the event table into a multiple subject event
table; `et()` will expand the event table by duplicating an event
table for each individual specified in `id=...`

As an exercise, expand the solve to be 4 IDs

```{r idExpand, exercise = TRUE}
ev <- et(timeUnits="hr") |>
    et(amt=10000, until = set_units(3, days), ii=12) |> # loading doses
    et(seq(0,48,length.out=200)) |>
    et(id=1:4)

print(ev)

set.seed(42)
rxSolve(m1, ev,
      params=data.frame(KA=0.294*exp(rnorm(4)), 18.6*exp(rnorm(4)))) |>
    plot(C2)
```

<div id="idExpand-hint">
You can add `et(id=1:4)` to the end of the event table pipline  to solve for 4 individuals
</div>

### Combining event tables

You may have noticed that the `rxSolve` will work if you add
`et(id=1:4)` or if you do not; Where the expanding of individual IDs
really shines is by combining arms in a study with different sampling
times or sampling intervals.  There the IDs will not have the exact
same structure.

To combine event tables, you can use `c`, `seq`, `rep`, and `rbind`.

### Sequencing event tables for complex dosing

One way to combine event table is to sequence them by `c()`, `seq()` or
`etSeq()`. This takes the two dosing groups and adds at least one
inter-dose interval between them:


`etSeq()` can combine two event tables or time intervals; As an
exercise, change the event table to have qd dosing followed by bid
dosing:

```{r bidQd, exercise=TRUE}
## bid for 5 days
bid <- et(timeUnits="hr") |>
       et(amt=10000,ii=12,until=set_units(5, "days"))

## qd for 5 days
qd <- et(timeUnits="hr") |>
      et(amt=20000,ii=24,until=set_units(5, "days"))

## bid for 5 days followed by qd for 5 days
## Change to qd for 5 days and bid for 5 days
et <- seq(bid,qd) |> et(seq(0,11*24,length.out=100));

rxSolve(m1, et) |> plot(C2)
```

### Using `seq` to add a drug holiday

You can also add drug holidays using the `etSeq` or `seq` function.
The way to do this is to simply add a time-period between the two dosing tables;  For example

```{r, echo=TRUE, eval=FALSE}
seq(eventTable1,separationTime,eventTable2)
```

Would combine `eventTable1` add a `separationTime` and finally add `eventTable2`

Below is an example of a 3 day drug holiday; Change it to a week of drug holiday

```{r etSeq2, exercise=TRUE}
## bid for 5 days
bid <- et(timeUnits="hr") |>
       et(amt=10000,ii=12,until=set_units(5, "days"))

## qd for 5 days
qd <- et(timeUnits="hr") |>
      et(amt=20000,ii=24,until=set_units(5, "days"))

## bid for 5 days followed by qd for 5 days
et <- seq(bid,set_units(3, day), qd) |>
    et(seq(0,18*24,length.out=100));

rxSolve(m1, et) |> plot(C2)
```

<div id="etSeq2-hint">
You can change `set_units(3, day)` to `set_units(1, week)`
</div>

### Notes on adding the one-week drug holiday

Note that in this worked-out example, the time between the `bid` and
the `qd` event tables is exactly one week, not 1 week plus 24 hours
because of the inter-dose interval.  If you want that behavior, you
can sequence it using the `wait="+ii"`.

Also note, that `rxode2` assumes that the dosing is what you want to
space the event tables by, and clears out any sampling records when
you combine the event tables.  If that is not true, you can also use
the option `samples="use"`

### Repeating event tables

You can have an event table that you can repeat with `etRep` or `rep`.

For example 4 rounds of 2 weeks on QD therapy and 1 week off of
therapy can be simply specified by the default exercise below.

As an exercise, change the drug holiday to 2 weeks


```{r etRep,exercise=TRUE}
qd <-et(timeUnits = "hr") |> et(amt=10000, ii=24, until=set_units(2, "weeks"), cmt="depot")

et <- rep(qd, times=4, wait=set_units(1,"weeks")) |>
      add.sampling(set_units(seq(0, 12.5,by=0.005),weeks))

rxSolve(m1, et)  |> plot(C2)
```

### Combining event tables with rbind

You may combine event tables with `rbind`.  This does not consider the
event times when combining the event tables, but keeps them the same
times. If you space the event tables by a waiting period, it also does
not consider the inter-dose interval.

Using the previous `seq` you can clearly see the difference.

Please run the bid/qd example to have it fresh in you memory.  After
you have run it, change `seq` to `rbind` to see what changesr

```{r rbind, exercise=TRUE}
## bid for 5 days
bid <- et(timeUnits="hr") |>
       et(amt=10000,ii=12,until=set_units(5, "days"))

## qd for 5 days
qd <- et(timeUnits="hr") |>
      et(amt=20000,ii=24,until=set_units(5, "days"))

et <- seq(bid,qd) |>
    et(seq(0,18*24,length.out=500));

rxSolve(m1, et) |> plot(C2)
```

### rbind vs seq

As you noticed in the example `rbind` combined the event tables as
they are, creating a double dose at the `qd` time and a single dose at
the `bid` time.  What about if you use a waiting period?

### Using rbind with a waiting period

Still the waiting period applies (but does not consider the inter-dose interval).

As an exercise, change the waiting period to 3 days.

```{r rbindWait, exercise=TRUE}
## bid for 5 days
bid <- et(timeUnits="hr") |>
       et(amt=10000,ii=12,until=set_units(5, "days"))

## qd for 5 days
qd <- et(timeUnits="hr") |>
      et(amt=20000,ii=24,until=set_units(5, "days"))


et <- rbind(bid,wait=set_units(10,days),qd) |>
    et(seq(0,18*24,length.out=500));

rxSolve(m1, et) |> plot(C2)
```

### rbind with a waiting interval of 3 days

Notice that when waiting for 3 days, the days 3-5 duplicate the QD
doses

### Combining multi-subject event tables with `rbind`

You can also bind the tables together and make each ID in the event
table unique; This can be good to combine cohorts with different
expected dosing and sampling times.  This requires the `id="unique"`
option;  Using the first example shows how this is different in this case:

```{r rbindId, exercise=TRUE}
## bid for 5 days
bid <- et(timeUnits="hr") |>
       et(amt=10000,ii=12,until=set_units(5, "days"))

## qd for 5 days
qd <- et(timeUnits="hr") |>
      et(amt=20000,ii=24,until=set_units(5, "days"))


## bid for 5 days
et <- etRbind(bid,qd, id="unique") |>
    et(seq(0,150,length.out=500));

library(ggplot2)
rxSolve(m1, et) |> plot(C2) + facet_wrap( ~ id)
```

## Add samples/doses in a window
### Dosing windows

In addition to adding fixed doses and fixed sampling times, you can
have windows where you sample and draw doses from.  For dosing windows
you specify the time as an ordered numerical vector with the lowest
dosing time and the highest dosing time inside a list.

In the initial example, the dosing is within a 6 hour dosing window;
This can be specified by `list(c(0,6))`;

For an exercise change the dosing windows to `c(0,2)`:

```{r doseWindow, exercise=TRUE}
set.seed(42)
ev <- et(timeUnits="hr") |>
    et(time=list(c(0,6)), amt=10000, until = set_units(2, days), ii=12) |> # loading doses
    et(id=1:4)

print(ev)
ev <- ev |> et(seq(0,48,length.out=200))

solve(m1, ev, params=data.frame(KA=0.294*exp(rnorm(4)), 18.6*exp(rnorm(4)))) |>
    plot(C2)
```

```{r doseWindow-solution}
set.seed(42)
ev <- et(timeUnits="hr") |>
    et(time=list(c(0,2)), amt=10000, until = set_units(2, days), ii=12) |> # loading doses
    et(id=1:4) |>
    et(seq(0,48,length.out=200))

solve(m1, ev, params=data.frame(KA=0.294*exp(rnorm(4)), 18.6*exp(rnorm(4)))) |> plot(C2)
```

### Sampling windows

The same sort of thing can be specified with sampling times.  To
specify the sampling times in terms of a sampling window, you can
create a list of the sampling times.  Each sampling time will be a two
element ordered numeric vector.

```{r sampleTime, exercise=TRUE}
set.seed(42)
ev <- et(timeUnits="hr") |>
    et(time=list(c(0,2)), amt=10000, until = set_units(2, days), ii=12) |> # loading doses
    et(id=1:4)

## Create 20 samples in the first 24 hours and 20 samples in the second 24 hours
samples <- c(lapply(1:20, function(...){c(0,24)}),
             lapply(1:20, function(...){c(20,48)}))

## Add the random collection to the event table
ev <- ev |> et(samples)

library(ggplot2)
solve(m1, ev, params=data.frame(KA=0.294*exp(rnorm(4)), 18.6*exp(rnorm(4)))) |> plot(C2) + geom_point()
```

This shows the flexibility in dosing and sampling that the rxode2 event tables allow.

## Units in event tables

As commented in the last exercise, you may specify units in the table
event table. When specified, the units use the `units` package to keep
track of the units and convert them if needed.  Additionally,
`ggforce` uses them to label the `ggplot` axes.  The `set_units` and
`drop_units` are useful to set and drop the rxode2 event table units.

This example creates a data-set with doses in `mg` and time in `hr`.

### Adding units in data-set and converting to different overall units

In this exercise use `set_units` to change the dose to `g` and the
time to `days`:

```{r units1, exercise=TRUE}
ev <- et(amount.units="mg", time.units="hr") |>
    et(dose=5000, dosing.interval=12, until=48) |>
    etExpand() ## Expand dosing so there is no ii/addl record

## Use set_units to change the  dose to g and time to days
print(ev)
```

<div id="units1-hint">
Use `ev |> set_units(g) |> set_units(day)`
</div>

### Dropping units

You may have used the `units` package in the past; In the case of
`rxode2` event tables, you do not have to apply to the column(s) in the
dataset, but `rxode2` applies the unit changes where it needs to apply them.

This is also true when dropping units in an event table.  In this
exercise, drop the units using `drop_units` from the event table dataset:

```{r units2, exercise=TRUE}
ev <- et(amount.units="mg", time.units="hr") |>
    et(dose=5000, dosing.interval=12, until=48) |>
    etExpand() |> ## Expand dosing so there is no ii/addl record
    set_units(day) |> # change the units to day/grams
    set_units(g)

## Use drop_units to remove unit information
print(ev)
```

<div id="units2-hint">
Use `drop_units()` on the `ev` dataset
</div>


### Dropping units keeps the original transformation

If you simply dropped the units at the end of the prior example, you
may have noticed that the event table `ev` did not revert to the units
originally specified in the event table.  It will keep the conversions
completed before unit information is dropped.

You could also add units to the event table by simply using `set_units`.

In this example, try setting the units to `mg` and `hr` by the `set_units` function:

```{r units3, exercise=TRUE}
ev <- et() |>
    et(dose=5000, dosing.interval=12, until=48) |>
    etExpand() ## Expand dosing so there is no ii/addl record

print(ev)
```

<div id="units3-hint">
Use the `set_units(mg) |> set_units(hr)` to change the units
</div>

## Using deSolve data frames

Sometimes you may have a `deSolve` model you may wish to convert to
`rxode2` to take advantage of its threading capabilities or
pre-compiled models.

Using the `deSolve`'s example for the event data frame and adapting it
for `rxode2` you simply wrap the `deSolve` data frame in the `et()` function;

For example, using the model and event table the March 2020
`deSolve`'s vignette, we can reproduce the results

```{r deSolve, exercise=TRUE}
rx <- rxode2({
    d/dt(v1) = -0.1 * v1
    d/dt(v2) = -0.1 * v2
    v1(0) = 1
    v2(0) = 2
})

eventdat <- data.frame(var = c("v1", "v2", "v2", "v1"), time = c(1, 1, 5, 9),
                       value = c(1, 2, 3, 4),
                       method = c("add", "mult", "rep", "add"))

print(eventdat)


## When we have a
et <- et(eventdat) |> et(0, 10, by=0.1)

print(et)

## In this case we have to add the dosing records
rxSolve(rx, et, times=seq(0, 10, by=0.1)) |> plot()
```

The model equations can be specified through a text string, a model
file or an R expression. Both differential and algebraic equations are
permitted. Differential equations are specified by `d/dt(var_name) = `. Each
equation can be separated by a semicolon.

To load `rxode2` package and compile the model:

```{r}
library(rxode2)

mod1 <- function() {
  ini({
    # central
    KA=2.94E-01
    CL=1.86E+01
    V2=4.02E+01
    # peripheral
    Q=1.05E+01
    V3=2.97E+02
    # effects
    Kin=1
    Kout=1
    EC50=200
  })
  model({
    C2 <- centr/V2
    C3 <- peri/V3
    d/dt(depot) <- -KA*depot
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
    d/dt(peri)  <- Q*C2 - Q*C3
    eff(0) <- 1
    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
  })
}
```

Model parameters may be specified in the `ini({})` model block,
initial conditions can be specified within the model with the `cmt(0)=
X`, like in this model `eff(0) <- 1`.

You may also specify between subject variability initial conditions
and residual error components just like nlmixr2.  This allows a single
interface for `nlmixr2`/`rxode2` models.  Also note, the classic
`rxode2` interface still works just like it did in the past (so don't
worry about breaking code at this time).

In fact, you can get the classic `rxode2` model `$simulationModel` in
the ui object:

```r
mod1 <- mod1() # create the ui object (can also use `rxode2(mod1)`)
mod1

summary(mod1$simulationModel)
```



## Specify Dosing and sampling in rxode2

`rxode2` provides a simple and very flexible way to specify dosing and
sampling through functions that generate an event table. First, an
empty event table is generated through the "et()" function. This
has an interface that is similar to NONMEM event tables:

```{r}
ev  <- et(amountUnits="mg", timeUnits="hours") |>
  et(amt=10000, addl=9,ii=12,cmt="depot") |>
  et(time=120, amt=2000, addl=4, ii=14, cmt="depot") |>
  et(0:240) # Add sampling
```

You can see from the above code, you can dose to the compartment named
in the rxode2 model.  This slight deviation from NONMEM can reduce the
need for compartment renumbering.

These events can also be combined and expanded (to multi-subject
events and complex regimens) with `rbind`, `c`, `seq`, and `rep`. For
more information about creating complex dosing regimens using rxode2
see the [rxode2 events
vignette](https://nlmixr2.github.io/rxode2/articles/rxode2-event-types.html).


## Solving ODEs

The ODE can now be solved using `rxSolve`:


```{r}
x <- mod1 |> rxSolve(ev)
x
```

This returns a modified data frame.  You can see the compartment
values in the plot below:

```{r intro-central}
library(ggplot2)
plot(x,C2) + ylab("Central Concentration")
```

Or,

```{r intro-effect}
plot(x,eff)  + ylab("Effect")
```

Note that the labels are automatically labeled with the units from the
initial event table. rxode2 extracts `units` to label the plot (if they
are present).

rxode2 automatically assigns compartment numbers when parsing. For
example, with the Mavoglurant PBPK model the following model may be
used:

```{r}
library(rxode2)

pbpk <- function() {
  model({
    KbBR = exp(lKbBR)
    KbMU = exp(lKbMU)
    KbAD = exp(lKbAD)
    CLint= exp(lCLint + eta.LClint)
    KbBO = exp(lKbBO)
    KbRB = exp(lKbRB)

    ## Regional blood flows
    # Cardiac output (L/h) from White et al (1968)
    CO  = (187.00*WT^0.81)*60/1000 
    QHT = 4.0 *CO/100
    QBR = 12.0*CO/100
    QMU = 17.0*CO/100
    QAD = 5.0 *CO/100
    QSK = 5.0 *CO/100
    QSP = 3.0 *CO/100
    QPA = 1.0 *CO/100
    QLI = 25.5*CO/100
    QST = 1.0 *CO/100
    QGU = 14.0*CO/100
    # Hepatic artery blood flow
    QHA = QLI - (QSP + QPA + QST + QGU) 
    QBO = 5.0 *CO/100
    QKI = 19.0*CO/100
    QRB = CO - (QHT + QBR + QMU + QAD + QSK + QLI + QBO + QKI)
    QLU = QHT + QBR + QMU + QAD + QSK + QLI + QBO + QKI + QRB

    ## Organs' volumes = organs' weights / organs' density
    VLU = (0.76 *WT/100)/1.051
    VHT = (0.47 *WT/100)/1.030
    VBR = (2.00 *WT/100)/1.036
    VMU = (40.00*WT/100)/1.041
    VAD = (21.42*WT/100)/0.916
    VSK = (3.71 *WT/100)/1.116
    VSP = (0.26 *WT/100)/1.054
    VPA = (0.14 *WT/100)/1.045
    VLI = (2.57 *WT/100)/1.040
    VST = (0.21 *WT/100)/1.050
    VGU = (1.44 *WT/100)/1.043
    VBO = (14.29*WT/100)/1.990
    VKI = (0.44 *WT/100)/1.050
    VAB = (2.81 *WT/100)/1.040
    VVB = (5.62 *WT/100)/1.040
    VRB = (3.86 *WT/100)/1.040

    ## Fixed parameters
    BP = 0.61      # Blood:plasma partition coefficient
    fup = 0.028    # Fraction unbound in plasma
    fub = fup/BP   # Fraction unbound in blood

    KbLU = exp(0.8334)
    KbHT = exp(1.1205)
    KbSK = exp(-.5238)
    KbSP = exp(0.3224)
    KbPA = exp(0.3224)
    KbLI = exp(1.7604)
    KbST = exp(0.3224)
    KbGU = exp(1.2026)
    KbKI = exp(1.3171)


    ##-----------------------------------------
    S15 = VVB*BP/1000
    C15 = Venous_Blood/S15

    ##-----------------------------------------
    d/dt(Lungs) = QLU*(Venous_Blood/VVB - Lungs/KbLU/VLU)
    d/dt(Heart) = QHT*(Arterial_Blood/VAB - Heart/KbHT/VHT)
    d/dt(Brain) = QBR*(Arterial_Blood/VAB - Brain/KbBR/VBR)
    d/dt(Muscles) = QMU*(Arterial_Blood/VAB - Muscles/KbMU/VMU)
    d/dt(Adipose) = QAD*(Arterial_Blood/VAB - Adipose/KbAD/VAD)
    d/dt(Skin) = QSK*(Arterial_Blood/VAB - Skin/KbSK/VSK)
    d/dt(Spleen) = QSP*(Arterial_Blood/VAB - Spleen/KbSP/VSP)
    d/dt(Pancreas) = QPA*(Arterial_Blood/VAB - Pancreas/KbPA/VPA)
    d/dt(Liver) = QHA*Arterial_Blood/VAB + QSP*Spleen/KbSP/VSP +
      QPA*Pancreas/KbPA/VPA + QST*Stomach/KbST/VST +
      QGU*Gut/KbGU/VGU - CLint*fub*Liver/KbLI/VLI - QLI*Liver/KbLI/VLI
    d/dt(Stomach) = QST*(Arterial_Blood/VAB - Stomach/KbST/VST)
    d/dt(Gut) = QGU*(Arterial_Blood/VAB - Gut/KbGU/VGU)
    d/dt(Bones) = QBO*(Arterial_Blood/VAB - Bones/KbBO/VBO)
    d/dt(Kidneys) = QKI*(Arterial_Blood/VAB - Kidneys/KbKI/VKI)
    d/dt(Arterial_Blood) = QLU*(Lungs/KbLU/VLU - Arterial_Blood/VAB)
    d/dt(Venous_Blood) = QHT*Heart/KbHT/VHT + QBR*Brain/KbBR/VBR +
      QMU*Muscles/KbMU/VMU + QAD*Adipose/KbAD/VAD + QSK*Skin/KbSK/VSK +
      QLI*Liver/KbLI/VLI + QBO*Bones/KbBO/VBO + QKI*Kidneys/KbKI/VKI +
      QRB*Rest_of_Body/KbRB/VRB - QLU*Venous_Blood/VVB
    d/dt(Rest_of_Body) = QRB*(Arterial_Blood/VAB - Rest_of_Body/KbRB/VRB)
  })
}
```

If you look at the printout, you can see where rxode2 assigned the compartment number(s)

```{r}
pbpk <- pbpk()
print(pbpk)
```

You can also see this with the classic rxode2 model. In that case you
use the `summary()` function:

```{r}
pbpk <- pbpk$simulationModel
summary(pbpk)
```

In this case, `Venous_Blood` is assigned to compartment `15`.
Figuring this out can be inconvenient and also lead to re-numbering
compartment in simulation or estimation datasets. While it is easy and
probably clearer to specify the [compartment by
name](rxode2-events.html), other tools only support compartment
numbers.  Therefore, having a way to number compartment easily can
lead to less data modification between multiple tools.

You can also append "compartments" to the model.  Because of the ODE
solving internals, you cannot add fake compartments to the model until
after all the differential equations are defined.

For example this is legal:

```{r}
ode.1c.ka <- function(){
  model({
    C2 = center/V
    d / dt(depot) = -KA * depot
    d/dt(center) = KA * depot - CL*C2
    cmt(eff)
  })
}

ode.1c.ka <- ode.1c.ka()
print(ode.1c.ka)
```

You can see this more clearly with the underlying classic `rxode2` model:

```{r}
ode.1c.ka$simulationModel
```

But compartments defined before all the differential equations is not supported;  So the model below:

```
ode.1c.ka <- rxode2({
    cmt(eff)
    C2 = center/V;
    d / dt(depot) = -KA * depot
    d/dt(center) = KA * depot - CL*C2
})
```

will give an error:

```
Error in rxModelVars_(obj) : 
  Evaluation error: Compartment 'eff' needs differential equations defined.
```

## Creating rxode2 models


The ODE-based model specification may be coded inside four places:

- Inside a `rxode2({})` block statements:

``` r
library(rxode2)
mod <- rxode2({
  # simple assignment
  C2 <- centr/V2

  # time-derivative assignment
  d/dt(centr) <- F*KA*depot - CL*C2 - Q*C2 + Q*C3;
})
```

- Inside a `rxode2("")` string statement:

``` r
mod <- rxode2("
  # simple assignment
  C2 <- centr/V2

  # time-derivative assignment
  d/dt(centr) <- F*KA*depot - CL*C2 - Q*C2 + Q*C3;
")
```

- In a file name to be loaded by rxode2:

``` r
writeLines("
  # simple assignment
  C2 <- centr/V2

  # time-derivative assignment
  d/dt(centr) <- F*KA*depot - CL*C2 - Q*C2 + Q*C3;
", "modelFile.rxode2")
mod <- rxode2(filename='modelFile.rxode2')
unlink("modelFile.rxode2")
```

- In a model function which can be parsed by `rxode2`:

``` r
mod <- function() {
  model({
    # simple assignment
    C2 <- centr/V2

    # time-derivative assignment
    d/dt(centr) <- F*KA*depot - CL*C2 - Q*C2 + Q*C3;
  })
}

mod <- rxode2(mod) # or simply mod() if the model is at the end of the function

# These model functions often have residual components and initial
# (`ini({})`) conditions attached as well. For example the
# theophylline model can be written as:

one.compartment <- function() {
  ini({
    tka <- 0.45 # Log Ka
    tcl <- 1 # Log Cl
    tv <- 3.45    # Log V
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v <- exp(tv + eta.v)
    d/dt(depot) = -ka * depot
    d/dt(center) = ka * depot - cl / v * center
    cp = center / v
    cp ~ add(add.sd)
  })
}

# after parsing the model
mod <- one.compartment()
```

For the block statement, character string or text file an internal
`rxode2` compilation manager translates the ODE system into C,
compiles it and loads it into the R session.  The
call to `rxode2` produces an object of class `rxode2` which
consists of a list-like structure (environment) with various member
functions.

For the last type of model (a model function), a call to `rxode2`
creates a parsed `rxode2` ui that can be translated to the `rxode2`
compilation model.

``` r
mod$simulationModel

# or
mod$simulationIniModel
```

This is the same type of function required for `nlmixr2` estimation
and can be extended and modified by model piping.  For this reason
will be focused on in the documentation.

---
editor_options:
  markdown:
    wrap: 72
---

This basic model specification consists of one or more statements
optionally terminated by semi-colons `;` and optional comments (comments
are delimited by `#` and an end-of-line).

A block of statements is a set of statements delimited by curly braces,
`{ ... }`.

Statements can be either assignments, conditional `if`/`else if`/`else`,
`while` loops (can be exited by `break`), special statements, or
printing statements (for debugging/testing).

Assignment statements can be:

-   **simple** assignments, where the left hand is an identifier (i.e.,
    variable).  This includes string assignments

-   special **time-derivative** assignments, where the left hand
    specifies the change of the amount in the corresponding state
    variable (compartment) with respect to time e.g., `d/dt(depot)`:

-   special **initial-condition** assignments where the left hand
    specifies the compartment of the initial condition being specified,
    e.g. `depot(0) = 0`

-   special model event changes including **bioavailability**
    (`f(depot)=1`), **lag time** (`alag(depot)=0`), **modeled rate**
    (`rate(depot)=2`) and **modeled duration** (`dur(depot)=2`). An
    example of these model features and the event specification for the
    modeled infusions the rxode2 data specification is found in [rxode2
    events vignette](https://nlmixr2.github.io/rxode2/articles/rxode2-event-types.html).

-   special **change point syntax, or model times**. These model times
    are specified by `mtime(var)=time`

-   special **Jacobian-derivative** assignments, where the left hand
    specifies the change in the compartment ode with respect to a
    variable. For example, if `d/dt(y) = dy`, then a Jacobian for this
    compartment can be specified as `df(y)/dy(dy) = 1`. There may be
    some advantage to obtaining the solution or specifying the Jacobian
    for very stiff ODE systems. However, for the few stiff systems we
    tried with LSODA, this actually slightly slowed down the solving.

- Special **string value declarations** which tell what values a
  string variable will take within a `rxode2` solving structure.
  These values will then cause a factor to be created for this
  variable on solving the `rxode2` model.  As such, they are declared
  in much the same way as `R`, that is: `labels(a) <- c("a1", "a2")`.

Note that assignment can be done by `=`, `<-` or `~`.

When assigning with the `~` operator, the **simple assignments** and
**time-derivative** assignments will not be output. Note that with the
`rxode2` model functions assignment with `~` can also be overloaded with
a residual distribution specification.

Special statements can be:

-   **Compartment declaration statements**, which can change the default
    dosing compartment and the assumed compartment number(s) as well as
    add extra compartment names at the end (useful for multiple-endpoint
    nlmixr models); These are specified by `cmt(compartmentName)`

-   **Parameter declaration statements**, which can make sure the input
    parameters are in a certain order instead of ordering the parameters
    by the order they are parsed. This is useful for keeping the
    parameter order the same when using 2 different ODE models. These
    are specified by `param(par1, par2,...)`

-   **Variable interpolation statements**, which tells the
    interpolation for specific covariates.  These include `locf(cov1,
    cov2, ...)` for last observation carried forward, `nocb(cov1,
    cov2, ...)`  for next observation carried backward, `linear(cov1,
    cov2, ...)` for linear interpolation and `midpoint(cov1, cov2,
    ...)` for midpoint interpolation.

An example model is shown below:

```
   # simple assignment
   C2 <- centr/V2

   # time-derivative assignment
   d/dt(centr) <- F*KA*depot - CL*C2 - Q*C2 + Q*C3;
```

Expressions in assignment and `if` statements can be numeric or logical.

Numeric expressions can include the following numeric operators
`+, -, *, /, ^, %%` and those mathematical functions defined in the C or the
R math libraries (e.g., `fabs`, `exp`, `log`, `sin`, `abs`).

You may also access the R's functions in the [R math
libraries](https://cran.r-project.org/doc/manuals/r-release/R-exts.html#Numerical-analysis-subroutines),
like `lgammafn` for the log gamma function.

The `rxode2` syntax is case-sensitive, i.e., `ABC` is different than
`abc`, `Abc`, `ABc`, etc.

### Identifiers

Like R, Identifiers (variable names) may consist of one or more
alphanumeric, underscore `_` or period `.` characters, but the first
character cannot be a digit or underscore `_`.

Identifiers in a model specification can refer to:

-   State variables in the dynamic system (e.g., compartments in a
    pharmacokinetics model).
-   Implied input variable, `t` (time), `tlast` (last time point), and
    `podo` (oral dose, in the undocumented case of absorption transit
    models).
-   Special constants like `pi` or [R's predefined
    constants](https://cran.r-project.org/doc/manuals/r-release/R-exts.html#Mathematical-constants).
-   Model parameters (e.g., `ka` rate of absorption, `CL` clearance,
    etc.)
-   Others, as created by assignments as part of the model
    specification; these are referred as *LHS* (left-hand side)
    variable.

Currently, the `rxode2` modeling language only recognizes system state
variables and "parameters", thus, any values that need to be passed from
R to the ODE model (e.g., `age`) should be either passed in the `params`
argument of the integrator function `rxSolve()` or be in the supplied
event data-set.

There are certain variable names that are in the `rxode2` event tables.
To avoid confusion, the following event table-related items cannot be
assigned, or used as a state but can be accessed in the rxode2 code:

-   `cmt`
-   `dvid`
-   `addl`
-   `ss`
-   `amt`
-   `dur`
-   `rate`
-   `Rprintf`
-   `print`
-   `printf`
-   `id`

However the following variables are cannot be used in a model
specification:

-   `evid`
-   `ii`

Sometimes rxode2 generates variables that are fed back to rxode2.
Similarly, nlmixr2 generates some variables that are used in nlmixr
estimation and simulation. These variables start with the either the
`rx` or `nlmixr` prefixes. To avoid any problems, it is suggested to not
use these variables starting with either the `rx` or `nlmixr` prefixes.

## Logical Operators

Logical operators support the standard R operators `==`, `!=` `>=` `<=`
`>` and `<`. Like R these can be in `if()` or `while()` statements,
`ifelse()` expressions. Additionally they can be in a standard
assignment. For instance, the following is valid:

```
cov1 = covm*(sexf == "female") + covm*(sexf != "female")
```

Notice that you can also use character expressions in comparisons. This
convenience comes at a cost since character comparisons are slower than
numeric expressions. Unlike R, `as.numeric` or `as.integer` for these
logical statements is not only not needed, but will cause an syntax
error if you try to use the function.

## Supported functions

All the supported functions in rxode2 can be seen with the
`rxSupportedFuns()`.

A brief description of the built-in functions are in the following
table:

```{r rxSyntaxFunctions, echo=FALSE, screenshot.force=FALSE}
.rxDocTable(rxSyntaxFunctions, caption="rxode2 supported functions")
```

Note that `lag(cmt) =` is equivalent to `alag(cmt) =` and not the same
as `= lag(wt)`

## Reserved keywords

There are a few reserved keywords in a rxode2 model. They are in the
following table:

```{r rxReservedKeywords, echo=FALSE, screenshot.force=FALSE}
.rxDocTable(rxReservedKeywords, caption="rxode2 supported keywords")
```

Note that `rxFlag` will always output `11` or `calc_lhs` since that is
where the final variables are calculated, though you can tweak or test
certain parts of `rxode2` by using this flag.

## Residual functions when using rxode2 functions

In addition to `~` hiding output for certain types of output, it also is
used to specify a residual output or endpoint when the input is an
`rxode2` model function (that includes the residual in the `model({})`
block).

These specifications are of the form:

``` r
var ~ add(add.sd)
```

Indicating the variable `var` is the variable that represents the
individual central tendencies of the model and it also represents the
compartment specification in the data-set.

You can also change the compartment name using the `|` syntax, that is:

``` r
var ~ add(add.sd) | cmt
```

In the above case `var` represents the central tendency and `cmt`
represents the compartment or `dvid` specification.

### Transformations

For normal and related distributions, you can apply the transformation
on both sides by using some keywords/functions to apply these
transformations.

```{r transformationCode, echo=FALSE, screenshot.force=FALSE}
.rxDocTable(dplyr::tribble(~"Transformation", ~"rxode2/nlmixr2 code",
                           "Box-Cox", "+boxCox(lambda)",
                           "Yeo-Johnson", "+yeoJohnson(lambda)",
                           "logit-normal", "+logitNorm(logit.sd, low, hi)",
                           "probit-normal", "+probitNorm(probid.sd, low, hi)",
                           "log-normal","+lnorm(lnorm.sd)"),
            caption="rxode2/nlmixr2 normal/t/cauchy transformations")
```

By default for the likelihood for all of these transformations is
calculated on the **untransformed** scale.

For bounded variables like logit-normal or probit-normal the low and
high values are defaulted to 0 and 1 if missing.

For models where you wish to have a proportional model on one of these
transformation you can replace the standard deviation with `NA`

To allow for more transformations, `lnorm()`, `probitNorm()` and
`logitNorm()` can be combined the variance stabilizing `yeoJohnson()`
transformation.

### Normal and t-related distributions

For the normal and t-related distributions, we wanted to keep the
ability to use skewed distributions additive and proportional in the
t/cauchy-space, so these distributions are specified differently in
comparison to the other supported distributions within `nlmixr2`:

```{r llikTable, echo=FALSE, screenshot.force=FALSE}
.rxDocTable(dplyr::tribble(
  ~"Distribution", ~"How to Add", ~"Example",
  "Normal (log-likelihood)", "+dnorm()", "cc ~ add(add.sd) + dnorm()",
  "T-distribution", "+dt(df)", "cc ~a dd(add.sd) + dt(df)",
  "Cauchy (t with df=1)", "+dcauchy()", "cc ~ add(add.sd) + dcauchy()"),
            caption="rxode2/nlmixr2 normal/t/cauchy log-liklihood specification")
```

Note that with the normal and t-related distributions `nlmixr2` will
calculate `cwres` and `npde` under the normal assumption to help assess
the goodness of the fit of the model.

Also note that the `+dnorm()` is mostly for testing purposes and will
slow down the estimation procedure in `nlmixr2`. We suggest not adding
it (except for explicit testing). When there are multiple endpoint
models that mix non-normal and normal distributions, the whole problem
is shifted to a log-likelihood method for estimation in `nlmixr2`.

### Notes on additive + proportional models

There are two different ways to specify additive and proportional
models, which we will call **combined1** and **combined2**, the same way
that Monolix calls the two distributions (to avoid between software
differences in naming).

The first, **combined1**, assumes that the additive and proportional
differences are on the standard deviation scale, or:

y=f+(a+b* f^c)*err

The second, **combined2**, assumes that the additive and proportional
differences are combined on a variance scale:

y=f+\[sqrt(a^2+b^2 *f^(2c))\]*err

The default in `nlmixr2`/`rxode2` if not otherwise specified is
**combined2** since it mirrors how adding 2 normal distributions in
statistics will add their variances (not the standard deviations).
However, the **combined1** can describe the data possibly even better
than **combined2** so both are possible options in `rxode2`/`nlmixr2`.

### Autocorrelated residuals (`ar()`)

Adding `ar(cor)` to a normal residual makes successive residuals of the
same endpoint follow a continuous-time AR(1) process:

``` r
cp ~ add(add.sd) + ar(ar1.cor)
```

Here `ar1.cor` is the lag-one correlation and must be in `[0, 1)`. The
correlation between two residuals separated by a time gap `dt` is
`cor^dt`, so irregular sampling is handled directly and the marginal
(stationary) residual variance is unchanged; the first observation of
each subject/endpoint uses the marginal distribution. `ar()` combines
with `add()`/`prop()` normal residuals and uses the same `| cmt`
endpoint syntax as the other residual functions.

The `ar()` specification is identical in `rxode2` and `nlmixr2`: the same
model line simulates the autocorrelated residual in `rxode2` and
estimates the correlation in `nlmixr2` (the `nlm`, `focei`/`foce` and
`saem` families support it; `fo`, `foi`, `nlme` and `nls` do not). See
Karlsson et al. (1995) for the continuous-time AR(1) residual model.

### Distributions of known likelihoods

For residuals that are not related to normal, t-distribution or cauchy,
often the residual specification is of the form:

``` r
cmt ~ dbeta(alpha, beta)
```

Where the compartment specification is on the left handed side of the
specification.

For generalized likelihood you can specify:

``` r
ll(cmt) ~ llik specification
```

### Ordinal likelihoods

Finally, ordinal likelihoods/simulations can be specified in 2 ways. The
first is:

``` r
err ~ c(p0, p1, p2)
```

Here `err` represents the compartment and `p0` is the probability of
being in a specific category:

| Category | Probability |
|----------|-------------|
| 1        | p0          |
| 2        | p1          |
| 3        | p2          |
| 4        | 1-p0-p1-p2  |

It is up to the model to ensure that the sum of the `p` values are less
than `1`. Additionally you can write an arbitrary number of categories
in the ordinal model described above.

It seems a little off that `p0` is the probability for category `1` and
sometimes scores are in non-whole numbers. This can be modeled as
follows:

``` r
err ~ c(p0=0, p1=1, p2=2, 3)
```

Here the numeric categories are specified explicitly, and the
probabilities remain the same:

| Category | Probability |
|----------|-------------|
| 0        | p0          |
| 1        | p1          |
| 2        | p2          |
| 3        | 1-p0-p1-p2  |

### General table of supported residual distributions

In general all the that are supported are in the following table
(available in `rxode2::rxResidualError`)

```{r rxResidualError, echo=FALSE, screenshot.force=FALSE}
.rxDocTable(rxResidualError, "rxode2/nlmixr2 esidual error specification")
```

# Note on strings in rxode2

Strings are converted to double values inside of `rxode2`, hence you
can refer to them as an integer corresponding to the string value or
the string value itself.  For covariates these are calculated on the
fly based on your data and you should likely not try this, though you
should be aware.  For strings defined in the model, this is fixed and
both could be used.

For example:
```
if (APGAR == 10 || APGAR == 8 || APGAR == 9) {
    tAPGAR <- "High"
  } else if (APGAR == 1 || APGAR == 2 || APGAR == 3) {
    tAPGAR <- "Low"
  } else if (APGAR == 4 || APGAR == 5 || APGAR == 6 || APGAR == 7) {
    tAPGAR <- "Med"
  } else {
    tAPGAR<- "Med"
  }
```

Could also be replaced by:

```
if (APGAR == 10 || APGAR == 8 || APGAR == 9) {
    tAPGAR <- "High"
  } else if (APGAR == 1 || APGAR == 2 || APGAR == 3) {
    tAPGAR <- "Low"
  } else if (APGAR == 4 || APGAR == 5 || APGAR == 6 || APGAR == 7) {
    tAPGAR <- "Med"
  } else {
    tAPGAR<- 3
  }
```

Since `"Med"` is already defined

If you wanted you can pre-declare what levels it has (and the order)
to give you better control of this:

```
levels(tAPGAR) <- c("Med", "Low", "High")
if (APGAR == 10 || APGAR == 8 || APGAR == 9) {
    tAPGAR <- 3
  } else if (APGAR == 1 || APGAR == 2 || APGAR == 3) {
    tAPGAR <- 2
  } else if (APGAR == 4 || APGAR == 5 || APGAR == 6 || APGAR == 7) {
    tAPGAR <- 1
  } else {
    tAPGAR<- 1
  }
```

You can see that the number changed since the declaration change the
numbers in each variable for `tAPGAR`.  These `levels()` statements
need to be declared before the variable occurs to ensure the numbering
is consistent with what is declared.

To add the compartments to the rxode2 model in the order you desire you
simply need to pre-declare the compartments with `cmt`.  For example
specifying is `Venous_Blood` and `Skin` to be the 1st and 2nd
compartments, respectively, is simple:

```{r}
pbpk2 <- function() {
  model({
    ## Now this is the first compartment, ie cmt=1
    cmt(Venous_Blood)
    ## Skin may be a compartment you wish to dose to as well,
    ##  so it is now cmt=2
    cmt(Skin)
    KbBR = exp(lKbBR)
    KbMU = exp(lKbMU)
    KbAD = exp(lKbAD)
    CLint= exp(lCLint + eta.LClint)
    KbBO = exp(lKbBO)
    KbRB = exp(lKbRB)

    ## Regional blood flows
    # Cardiac output (L/h) from White et al (1968)m
    CO  = (187.00*WT^0.81)*60/1000;
    QHT = 4.0 *CO/100;
    QBR = 12.0*CO/100;
    QMU = 17.0*CO/100;
    QAD = 5.0 *CO/100;
    QSK = 5.0 *CO/100;
    QSP = 3.0 *CO/100;
    QPA = 1.0 *CO/100;
    QLI = 25.5*CO/100;
    QST = 1.0 *CO/100;
    QGU = 14.0*CO/100;
    QHA = QLI - (QSP + QPA + QST + QGU); # Hepatic artery blood flow
    QBO = 5.0 *CO/100;
    QKI = 19.0*CO/100;
    QRB = CO - (QHT + QBR + QMU + QAD + QSK + QLI + QBO + QKI);
    QLU = QHT + QBR + QMU + QAD + QSK + QLI + QBO + QKI + QRB;

    ## Organs' volumes = organs' weights / organs' density
    VLU = (0.76 *WT/100)/1.051;
    VHT = (0.47 *WT/100)/1.030;
    VBR = (2.00 *WT/100)/1.036;
    VMU = (40.00*WT/100)/1.041;
    VAD = (21.42*WT/100)/0.916;
    VSK = (3.71 *WT/100)/1.116;
    VSP = (0.26 *WT/100)/1.054;
    VPA = (0.14 *WT/100)/1.045;
    VLI = (2.57 *WT/100)/1.040;
    VST = (0.21 *WT/100)/1.050;
    VGU = (1.44 *WT/100)/1.043;
    VBO = (14.29*WT/100)/1.990;
    VKI = (0.44 *WT/100)/1.050;
    VAB = (2.81 *WT/100)/1.040;
    VVB = (5.62 *WT/100)/1.040;
    VRB = (3.86 *WT/100)/1.040;

    ## Fixed parameters
    BP = 0.61;      # Blood:plasma partition coefficient
    fup = 0.028;    # Fraction unbound in plasma
    fub = fup/BP;   # Fraction unbound in blood

    KbLU = exp(0.8334);
    KbHT = exp(1.1205);
    KbSK = exp(-.5238);
    KbSP = exp(0.3224);
    KbPA = exp(0.3224);
    KbLI = exp(1.7604);
    KbST = exp(0.3224);
    KbGU = exp(1.2026);
    KbKI = exp(1.3171);


    ##-----------------------------------------
    S15 = VVB*BP/1000;
    C15 = Venous_Blood/S15

    ##-----------------------------------------
    d/dt(Lungs) = QLU*(Venous_Blood/VVB - Lungs/KbLU/VLU);
    d/dt(Heart) = QHT*(Arterial_Blood/VAB - Heart/KbHT/VHT);
    d/dt(Brain) = QBR*(Arterial_Blood/VAB - Brain/KbBR/VBR);
    d/dt(Muscles) = QMU*(Arterial_Blood/VAB - Muscles/KbMU/VMU);
    d/dt(Adipose) = QAD*(Arterial_Blood/VAB - Adipose/KbAD/VAD);
    d/dt(Skin) = QSK*(Arterial_Blood/VAB - Skin/KbSK/VSK);
    d/dt(Spleen) = QSP*(Arterial_Blood/VAB - Spleen/KbSP/VSP);
    d/dt(Pancreas) = QPA*(Arterial_Blood/VAB - Pancreas/KbPA/VPA);
    d/dt(Liver) = QHA*Arterial_Blood/VAB + QSP*Spleen/KbSP/VSP +
      QPA*Pancreas/KbPA/VPA + QST*Stomach/KbST/VST + QGU*Gut/KbGU/VGU -
      CLint*fub*Liver/KbLI/VLI - QLI*Liver/KbLI/VLI;
      d/dt(Stomach) = QST*(Arterial_Blood/VAB - Stomach/KbST/VST);
      d/dt(Gut) = QGU*(Arterial_Blood/VAB - Gut/KbGU/VGU);
      d/dt(Bones) = QBO*(Arterial_Blood/VAB - Bones/KbBO/VBO);
      d/dt(Kidneys) = QKI*(Arterial_Blood/VAB - Kidneys/KbKI/VKI);
      d/dt(Arterial_Blood) = QLU*(Lungs/KbLU/VLU - Arterial_Blood/VAB);
      d/dt(Venous_Blood) = QHT*Heart/KbHT/VHT + QBR*Brain/KbBR/VBR +
        QMU*Muscles/KbMU/VMU + QAD*Adipose/KbAD/VAD + QSK*Skin/KbSK/VSK +
        QLI*Liver/KbLI/VLI + QBO*Bones/KbBO/VBO + QKI*Kidneys/KbKI/VKI +
        QRB*Rest_of_Body/KbRB/VRB - QLU*Venous_Blood/VVB;
        d/dt(Rest_of_Body) = QRB*(Arterial_Blood/VAB - Rest_of_Body/KbRB/VRB);
  })
}
```

You can see this change in the simple printout

```{r}
pbpk2 <- pbpk2()
pbpk2
```

The first two compartments are `Venous_Blood` followed by `Skin`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxLinCmt.R
\name{rxGetLin}
\alias{rxGetLin}
\title{Get the linear compartment model true function}
\usage{
rxGetLin(model, linCmtSens = c("linCmtA", "linCmtB"), verbose = FALSE)
}
\arguments{
\item{model}{This is the ODE model specification.  It can be:
\itemize{
\item a string containing the set of ordinary differential
equations (ODE) and other expressions defining the changes in
the dynamic system.
\item a file name where the ODE system equation is contained
}

An ODE expression enclosed in \verb{\\\{\\\}}

(see also the \code{filename} argument). For
details, see the sections \dQuote{Details} and
\verb{rxode2 Syntax} below.}

\item{linCmtSens}{The method to calculate the linCmt() solutions}

\item{verbose}{When \code{TRUE} be verbose with the linear
compartmental model}
}
\value{
model with linCmt() replaced with linCmtA()
}
\description{
Get the linear compartment model true function
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{.symengineFs}
\alias{.symengineFs}
\title{Return symengineFs from user functions}
\usage{
.symengineFs()
}
\value{
symengineFs from user functions
}
\description{
Return symengineFs from user functions
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err-foceiBase.R
\name{.rxGetPredictionDVTransform}
\alias{.rxGetPredictionDVTransform}
\title{Get the Observation transformation}
\usage{
.rxGetPredictionDVTransform(env, pred1, yj)
}
\arguments{
\item{env}{Environment for the parsed model}

\item{pred1}{The \code{data.frame} of the current error}

\item{yj}{The transformation number for the current error}
}
\value{
The transformation expression
}
\description{
Get the Observation transformation
}
\author{
Matthew Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indLin.R
\name{rxSensMatExp}
\alias{rxSensMatExp}
\title{Differentiate and expand a matrix exponential model with forward sensitivities}
\usage{
rxSensMatExp(
  model,
  calcSens,
  calcSens2 = NULL,
  calcSens3 = NULL,
  doConst = FALSE,
  env = NULL
)
}
\arguments{
\item{model}{rxode2 model, text, or function}

\item{calcSens}{A character vector of parameter names for which sensitivities should be calculated.}

\item{calcSens2}{character vector (or \code{NULL}) requesting second-order
sensitivities \verb{rx__sens_<x>_BY_<p>_BY_<q>__} (\code{p} over \code{calcSens}, \code{q} over
\code{calcSens2}; every \code{calcSens2} element must also be in \code{calcSens}).
Expressed as \code{k_from_to} micro-constant transfers like the first-order
ones.  Ignored for \code{linCmt()} states (those use Stan forward-AD).}

\item{calcSens3}{character vector (or \code{NULL}) requesting third-order
sensitivities \verb{rx__sens_<x>_BY_<p>_BY_<q>_BY_<r>__} (\code{r} over \code{calcSens3}).
Requires \code{calcSens2}; every \code{calcSens3} element must also be in \code{calcSens2}.}

\item{doConst}{Replace constants with values; By default this is \code{FALSE}.}

\item{env}{A pre-loaded symengine environment (from \code{.rxLoadPrune()}) to
reuse instead of reloading \code{model}; when \code{NULL} it is built internally.}
}
\value{
A character string representing the matrix exponential sensitivity-expanded model code
}
\description{
Differentiate and expand a matrix exponential model with forward sensitivities
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxIsCurrent}
\alias{rxIsCurrent}
\title{Checks if the rxode2 object was built with the current build}
\usage{
rxIsCurrent(obj)
}
\arguments{
\item{obj}{rxode2 family of objects}
}
\value{
boolean indicating if this was built with current rxode2
}
\description{
Checks if the rxode2 object was built with the current build
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{eventTable}
\alias{eventTable}
\title{Create an event table object}
\usage{
eventTable(amount.units = NA, time.units = NA)
}
\arguments{
\item{amount.units}{string denoting the amount dosing units, e.g.,
\dQuote{mg}, \dQuote{ug}. Default to \code{NA} to denote
unspecified units.  It could also be a solved rxode2 object.  In
that case, eventTable(obj) returns the eventTable that was used
to solve the rxode2 object.}

\item{time.units}{string denoting the time units, e.g.,
\dQuote{hours}, \dQuote{days}. Default to \code{"hours"}.

An \code{eventTable} is an object that consists of a data.frame
storing ordered time-stamped events of an (unspecified) PK/PD
dynamic system, units (strings) for dosing and time records, plus a
list of functions to add and extract event records.

Currently, events can be of two types: dosing events that represent
inputs to the system and sampling time events that represent
observations of the system with \sQuote{amount.units} and
\sQuote{time.units}, respectively.}
}
\value{
A modified data.frame with the following accessible functions:
\itemize{
\item \code{get.EventTable()} returns the current event table
\item \code{\link[=add.dosing]{add.dosing()}}  adds dosing records to the event table.
\item \code{get.dosing()} returns a data.frame of dosing records.
\item \code{clear.dosing()} clears or deletes all dosing from event table
\item `\code{\link[=add.sampling]{add.sampling()}} adds sampling time observation records to the
event table.
\item \code{get.sampling()}returns a data.frame of sampled observation records.
\item \code{clear.sampling()} removes all sampling from event table.
\item \code{get.obs.rec()} returns a logical vector indicating whether each
event record represents an observation or not.
\item \code{get.nobs()} returns the number of observation (not dosing) records.
\item \code{get.units()} returns a two-element character vector with the
dosing and time units, respectively
\item \code{copy()} makes a copy of the current event table. To create
a copy of an event table object use \code{qd2 <- qd$copy()}
\item \code{expand()} Expands the event table for multi-subject solving.
This is done by \code{qd$expand(400)} for a 400 subject data expansion
}
}
\description{
Initializes an object of class \sQuote{EventTable} with methods for
adding and querying dosing and observation records
}
\examples{
# create dosing and observation (sampling) events
# QD 50mg dosing, 5 days followed by 25mg 5 days
#
qd <- eventTable(amount.units = "mg", time.units = "days")
#
qd$add.dosing(dose = 50, nbr.doses = 5, dosing.interval = 1, do.sampling = FALSE)
#
# sample the system's drug amounts hourly the first day, then every 12 hours
# for the next 4 days
qd$add.sampling(seq(from = 0, to = 1, by = 1 / 24))
qd$add.sampling(seq(from = 1, to = 5, by = 12 / 24))
#
# print(qd$get.dosing())     # table of dosing records
print(qd$get.nobs()) # number of observation (not dosing) records
#
# BID dosing, 5 days
bid <- eventTable("mg", "days") # only dosing
bid$add.dosing(
  dose = 10000, nbr.doses = 2 * 5,
  dosing.interval = 12, do.sampling = FALSE
)
#
# Use the copy() method to create a copy (clone) of an existing
# event table (simple assignments just create a new reference to
# the same event table object (closure)).
#
bid.ext <- bid$copy() # three-day extension for a 2nd cohort
bid.ext$add.dosing(
  dose = 5000, nbr.doses = 2 * 3,
  start.time = 120, dosing.interval = 12, do.sampling = FALSE
)

# You can also use the Piping operator to create a table

qd2 <- eventTable(amount.units = "mg", time.units = "days") |>
  add.dosing(dose = 50, nbr.doses = 5, dosing.interval = 1, do.sampling = FALSE) |>
  add.sampling(seq(from = 0, to = 1, by = 1 / 24)) |>
  add.sampling(seq(from = 1, to = 5, by = 12 / 24))
# print(qd2$get.dosing())     # table of dosing records
print(qd2$get.nobs()) # number of observation (not dosing) records

# Note that piping with |> will update the original table.

qd3 <- qd2 |> add.sampling(seq(from = 5, to = 10, by = 6 / 24))
print(qd2$get.nobs())
print(qd3$get.nobs())
}
\seealso{
\code{\link[=et]{et()}}
}
\author{
Matthew Fidler, Melissa Hallow and Wenping Wang
}
\concept{Nonlinear regression}
\concept{Pharmacodynamics (PD)}
\concept{Pharmacokinetics (PK)}
\concept{ordinary differential equations}
\keyword{data}
\keyword{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxReservedKeywords.R
\docType{data}
\name{rxReservedKeywords}
\alias{rxReservedKeywords}
\title{A list and description of rxode2 supported reserved keywords}
\format{
A data frame with 3 columns and 31 rows
\describe{
\item{Reserved Name}{Reserved Keyword Name}
\item{Meaning}{Reserved Keyword Meaning}
\item{Alias}{Keyword Alias}
}
}
\usage{
rxReservedKeywords
}
\description{
A list and description of rxode2 supported reserved keywords
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui.R
\name{rxUiDecompress}
\alias{rxUiDecompress}
\alias{rxUiCompress}
\title{Compress/Decompress \code{rxode2} ui}
\usage{
rxUiDecompress(ui)

rxUiCompress(ui)
}
\arguments{
\item{ui}{rxode2 ui object}
}
\value{
A compressed or decompressed rxui object
}
\description{
Compress/Decompress \code{rxode2} ui
}
\examples{
one.cmt <- function() {
  ini({
    ## You may label each parameter with a comment
    tka <- 0.45 # Log Ka
    tcl <- log(c(0, 2.7, 100)) # Log Cl
    ## This works with interactive models
    ## You may also label the preceding line with label("label text")
    tv <- 3.45; label("log V")
    ## the label("Label name") works with all models
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v <- exp(tv + eta.v)
    linCmt() ~ add(add.sd) | tmp
  })
}

f <- rxode2(one.cmt)
print(class(f))
print(is.environment(f))

f  <- rxUiDecompress(f)
print(class(f))
print(is.environment(f))

f  <- rxUiCompress(f)
print(class(f))
print(is.environment(f))

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assert.R
\name{assertVariableNew}
\alias{assertVariableNew}
\title{Assert a variable would be new to the model}
\usage{
assertVariableNew(ui, x)
}
\arguments{
\item{ui}{rxode2 ui model}

\item{x}{would the variable \code{x} variable be new in the model}
}
\value{
nothing, but will error if \code{x} would not be new
}
\description{
Assert a variable would be new to the model
}
\seealso{
Other Assertions:
\code{\link[=assertCompartmentExists]{assertCompartmentExists()}},
\code{\link[=assertCompartmentName]{assertCompartmentName()}},
\code{\link[=assertCompartmentNew]{assertCompartmentNew()}},
\code{\link[=assertRxUi]{assertRxUi()}},
\code{\link[=assertVariableExists]{assertVariableExists()}},
\code{\link[=testIniDf]{testIniDf()}},
\code{\link[=testRxUnbounded]{testRxUnbounded()}}
}
\author{
Matthew L. Fidler
}
\concept{Assertions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxOom.R
\name{rxSolveChunked}
\alias{rxSolveChunked}
\title{Solve an ODE model in memory-safe chunks}
\usage{
rxSolveChunked(
  object,
  params = NULL,
  events = NULL,
  inits = NULL,
  ...,
  chunkSize,
  seed = NULL,
  parallel = 0L
)
}
\arguments{
\item{object}{rxode2 model}

\item{params}{model parameters}

\item{events}{event table or \code{rxEtFile}}

\item{inits}{initial conditions}

\item{...}{additional arguments passed to \code{rxControl()}}

\item{chunkSize}{number of subjects per chunk (auto-computed from free RAM if omitted)}

\item{seed}{random seed (sets before solving if not NULL)}

\item{parallel}{number of mirai daemons for parallel chunk solving (0 = serial)}
}
\value{
An \code{rxSolveOom} object
}
\description{
Splits subjects into chunks sized to fit in available RAM, solves each chunk,
and writes output to parquet (or rds) files. Returns an \code{rxSolveOom}
object that lazily reads chunks on demand.
}
\details{
The storage/query engine used by the resulting \code{rxSolveOom} object is
controlled by the \code{rxode2.oom.backend} option:
\describe{
\item{\code{"auto"}}{(default) DuckDB if installed, else arrow, else rds.}
\item{\code{"duckdb"}}{lazy DuckDB SQL queries over arrow-written parquet.}
\item{\code{"arrow"}}{arrow parquet reads/writes, no DuckDB.}
\item{\code{"rds"}}{plain rds files, no arrow or DuckDB.}
}
A requested engine that is not installed silently degrades
(duckdb \eqn{\to} arrow \eqn{\to} rds).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/asRxui.R
\name{as.rxUi}
\alias{as.rxUi}
\alias{as.rxUi.rxode2}
\alias{as.rxUi.rxode2tos}
\alias{as.rxUi.rxModelVars}
\alias{as.rxUi.function}
\alias{as.rxUi.rxUi}
\alias{as.rxUi.default}
\title{As rxode2 ui}
\usage{
as.rxUi(x)

\method{as.rxUi}{rxode2}(x)

\method{as.rxUi}{rxode2tos}(x)

\method{as.rxUi}{rxModelVars}(x)

\method{as.rxUi}{`function`}(x)

\method{as.rxUi}{rxUi}(x)

\method{as.rxUi}{default}(x)
}
\arguments{
\item{x}{Object to convert to \code{rxUi} object}
}
\value{
rxUi object (or error if it cannot be converted)
}
\description{
As rxode2 ui
}
\examples{

mod1 <- function() {
 ini({
   # central 
   KA=2.94E-01
   CL=1.86E+01
   V2=4.02E+01
   # peripheral
   Q=1.05E+01
   V3=2.97E+02
   # effects
   Kin=1
   Kout=1
   EC50=200 
 })
 model({
   C2 <- centr/V2
   C3 <- peri/V3
   d/dt(depot) <- -KA*depot
   d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
   d/dt(peri)  <- Q*C2 - Q*C3
   eff(0) <- 1
   d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
 })
}

as.rxUi(mod1)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxSerialize.R
\name{rxLoadState}
\alias{rxLoadState}
\title{Restore a pre-integration rxode2 solver state from a binary file}
\usage{
rxLoadState(file)
}
\arguments{
\item{file}{Character(1).  Path of a file written by \code{\link{rxSaveState}}.}
}
\value{
Invisibly \code{TRUE}.
}
\description{
Reads a file written by \code{\link{rxSaveState}} (or via
\code{rxControl(serializeFile=...)}) and reconstructs the live solver state
so that ODE integration can proceed from the saved pre-integration
checkpoint.
}
\details{
This function is called automatically when \code{rxSolve(mod, "state.rxbin")}
detects a serialization file by its magic bytes and validates that the file
matches the supplied model.
}
\seealso{
\code{\link{rxSaveState}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.rxCoef}
\alias{print.rxCoef}
\title{Print the rxCoef object}
\usage{
\method{print}{rxCoef}(x, ...)
}
\arguments{
\item{x}{rxCoef object}
}
\value{
original object
}
\description{
This prints out the user supplied arguments for rxCoef object
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxIndLin.R
\name{rxIndLinStrategy}
\alias{rxIndLinStrategy}
\title{This sets the inductive linearization strategy for matrix building}
\usage{
rxIndLinStrategy(strategy = c("curState", "split"))
}
\arguments{
\item{strategy}{The strategy for inductive linearization matrix building
\itemize{
\item \code{curState} Prefer parameterizing in terms of the current
state, followed by the first state observed in the term.
\item \code{split} Split the parameterization between all states in the
term by dividing each by the number of states in the term and then
adding a matrix term for each state.
}}
}
\value{
Nothing
}
\description{
When there is more than one state in a ODE that cannot be
separated this specifies how it is incorporated into the matrix
exponential.
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evidPush.R, R/rudfui.R, R/rxrandom.R
\name{rxUdfUi.evid_}
\alias{rxUdfUi.evid_}
\alias{rxUdfUi.bolus}
\alias{rxUdfUi.infuse}
\alias{rxUdfUi.infuseDur}
\alias{rxUdfUi.replace}
\alias{rxUdfUi.multiply}
\alias{rxUdfUi.phantom}
\alias{rxUdfUi.obs}
\alias{rxUdfUi}
\alias{rxUdfUi.rinbinom}
\title{This function is called when processing rxode2 user functions from
the models}
\usage{
\method{rxUdfUi}{evid_}(fun)

\method{rxUdfUi}{bolus}(fun)

\method{rxUdfUi}{infuse}(fun)

\method{rxUdfUi}{infuseDur}(fun)

\method{rxUdfUi}{replace}(fun)

\method{rxUdfUi}{multiply}(fun)

\method{rxUdfUi}{phantom}(fun)

\method{rxUdfUi}{obs}(fun)

rxUdfUi(fun)

\method{rxUdfUi}{rinbinom}(fun)
}
\arguments{
\item{fun}{this is the function that needs to be parsed and
changed.  This is a R language expression}
}
\value{
This needs to return a list with the following elements:
\itemize{
\item \code{iniDf} -- the modified initial estimate data.frame
\item \code{before} -- any model code that needs to be added before the current line
\item \code{after} -- any model code that needs to be added after the current line
\item \code{replace} -- replacement code for this user function
}
}
\description{
This function is called when processing rxode2 user functions from
the models
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-bind.R
\name{rxAppendModel}
\alias{rxAppendModel}
\title{Append two rxui models together}
\usage{
rxAppendModel(..., common = TRUE)
}
\arguments{
\item{...}{models to append together}

\item{common}{boolean; when \code{TRUE} warn if the models have no variables in common}
}
\value{
New model with both models appended together
}
\description{
Append two rxui models together
}
\examples{

\donttest{

ocmt <- function() {
  ini({
    tka <- exp(0.45) # Ka
    tcl <- exp(1) # Cl
    tv <- exp(3.45); # log V
    ## the label("Label name") works with all models
    add.sd <- 0.7
  })
  model({
    ka <- tka
    cl <- tcl
    v <- tv
    d/dt(depot) <- -ka * depot
    d/dt(center) <- ka * depot - cl / v * center
    cp <- center / v
    cp ~ add(add.sd)
  })
}

idr <- function() {
  ini({
    tkin <- log(1)
    tkout <- log(1)
    tic50 <- log(10)
    gamma <- fix(1)
    idr.sd <- 1
  })
  model({
    kin <- exp(tkin)
    kout <- exp(tkout)
    ic50 <- exp(tic50)
    d/dt(eff) <- kin - kout*(1-ceff^gamma/(ic50^gamma+ceff^gamma))
    eff ~ add(idr.sd)
  })
}

rxAppendModel(ocmt |>
  model(ceff=cp,append=TRUE), idr)

}

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxOom.R
\name{as.arrow}
\alias{as.arrow}
\alias{as.arrow.rxSolveOom}
\title{Convert an rxSolveOom result to a lazy Arrow Dataset}
\usage{
as.arrow(x, ...)

\method{as.arrow}{rxSolveOom}(x, ...)
}
\arguments{
\item{x}{An \code{rxSolveOom} object.}

\item{...}{Ignored.}
}
\value{
An \code{arrow::Dataset}.
}
\description{
Opens all parquet chunk files as a single lazy \code{arrow::Dataset} using
\code{arrow::open_dataset()}.  The dataset can be filtered and selected with
dplyr verbs before calling \code{dplyr::collect()} to materialise.  Requires
the \code{arrow} package.
}
\details{
This is an rxode2 generic: \pkg{arrow} provides \code{open_dataset()} but no
lazy-dataset coercion generic to dispatch on, so rxode2 defines its own.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/rxode2.R
\name{rxInits}
\alias{rxInits}
\alias{rxInit}
\title{Initial Values and State values for a rxode2 object}
\usage{
rxInits(
  obj,
  vec = NULL,
  req = NULL,
  defaultValue = 0,
  noerror = FALSE,
  noini = FALSE,
  rxLines = FALSE
)

rxInit(
  obj,
  vec = NULL,
  req = NULL,
  defaultValue = 0,
  noerror = FALSE,
  noini = FALSE,
  rxLines = FALSE
)
}
\arguments{
\item{obj}{rxDll, rxode2, or named vector representing default
initial arguments}

\item{vec}{If supplied, named vector for the model.}

\item{req}{Required names, and the required order for the ODE solver}

\item{defaultValue}{a number or NA representing the default value for
parameters missing in \code{vec}, but required in \code{req}.}

\item{noerror}{is a boolean specifying if an error should be thrown
for missing parameter values when \code{default} = \code{NA}}
}
\value{
Initial values of the rxDll object
}
\description{
Returns the initial values of the rxDll object
}
\seealso{
Other Query model information:
\code{\link[=rxDfdy]{rxDfdy()}},
\code{\link[=rxLhs]{rxLhs()}},
\code{\link[=rxModelVars]{rxModelVars()}},
\code{\link[=rxParams]{rxParams()}},
\code{\link[=rxState]{rxState()}}
}
\author{
Matthew L.Fidler
}
\concept{Query model information}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikF}
\alias{llikF}
\title{log likelihood and derivatives for F distribution}
\usage{
llikF(x, df1, df2, full = FALSE)
}
\arguments{
\item{x}{variable that is distributed by f distribution}

\item{df1, df2}{degrees of freedom.  \code{Inf} is allowed.}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the log pdf value of with \code{dDf1} and \code{dDf2}
that has the derivatives with respect to the \code{df1}/\code{df2} parameters at
the observation time-point
}
\description{
log likelihood and derivatives for F distribution
}
\details{
In an \code{rxode2()} model, you can use \code{llikF()} but you have to
use the x and rate arguments.  You can also get the derivative of \code{df1} and \code{df2} with
\code{llikFDdf1()} and \code{llikFDdf2()}.
}
\examples{

\donttest{
x <- seq(0.001, 5, length.out = 100)

llikF(x^2, 1, 5)

model <- function(){
  model({
    fx <- llikF(time, df1, df2)
    dMean <- llikFDdf1(time, df1, df2)
    dSd <- llikFDdf2(time, df1, df2)
  })
}

et <- et(x)
et$df1 <- 1
et$df2 <- 5

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err.R
\name{rxPreferredDistributionName}
\alias{rxPreferredDistributionName}
\title{Change distribution name to the preferred distribution name term}
\usage{
rxPreferredDistributionName(dist)
}
\arguments{
\item{dist}{This is the input distribution}
}
\value{
Preferred distribution term
}
\description{
This is determined by the internal preferred condition name list
\code{.errIdenticalDists}
}
\examples{

rxPreferredDistributionName("dt")

rxPreferredDistributionName("add")

# can be vectorized

rxPreferredDistributionName(c("add","dt"))

}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tran.R
\name{rxode2parseAssignTranslation}
\alias{rxode2parseAssignTranslation}
\title{This assigns the c level linkages for a roxde2 model}
\usage{
rxode2parseAssignTranslation(df)
}
\arguments{
\item{df}{data frame containing the character column names rxFun,
fun, type, package, packageFun and the integer column names
argMin and argMax}
}
\value{
Nothing called for side effects
}
\description{
This assigns the c level linkages for a roxde2 model
}
\examples{

rxode2parseAssignTranslation(rxode2parseGetTranslation())

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxLinCmt.R
\name{findLhs}
\alias{findLhs}
\title{Find the assignments in R expression}
\usage{
findLhs(x)
}
\arguments{
\item{x}{R expression}
}
\value{
list of assigned parameters
}
\description{
Find the assignments in R expression
}
\author{
Hadley Wickham and Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-assign-parts.R
\name{rxode2<-}
\alias{rxode2<-}
\alias{rxode2<-.function}
\alias{rxode2<-.default}
\alias{rxode<-}
\alias{RxODE<-}
\title{Set the function body of an rxUi object while retaining other object
information (like data)}
\usage{
rxode2(x, envir = environment(x)) <- value

\method{rxode2}{`function`}(x, envir = environment(x)) <- value

\method{rxode2}{default}(x, envir = environment(x)) <- value

rxode(x, envir = environment(x)) <- value

RxODE(x, envir = environment(x)) <- value
}
\arguments{
\item{x}{The rxUi object}

\item{envir}{environment where the assignment ocurs}

\item{value}{the value that will be assigned}
}
\value{
The rxode2 ui/function
}
\description{
Set the function body of an rxUi object while retaining other object
information (like data)
}
\examples{

one.compartment <- function() {
  ini({
    tka <- log(1.57); label("Ka")
    tcl <- log(2.72); label("Cl")
    tv <- log(31.5); label("V")
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v <- exp(tv + eta.v)
    d/dt(depot) = -ka * depot
    d/dt(center) = ka * depot - cl / v * center
    cp = center / v
    cp ~ add(add.sd)
  })
}

two.compartment <- function() {
  ini({
    lka <- 0.45 ; label("Absorption rate (Ka)")
    lcl <- 1 ; label("Clearance (CL)")
    lvc  <- 3 ; label("Central volume of distribution (V)")
    lvp  <- 5 ; label("Peripheral volume of distribution (Vp)")
    lq  <- 0.1 ; label("Intercompartmental clearance (Q)")
    propSd <- 0.5 ; label("Proportional residual error (fraction)")
  })
  model({
    ka <- exp(lka)
    cl <- exp(lcl)
    vc <- exp(lvc)
    vp <- exp(lvp)
    q  <- exp(lq)
    kel <- cl/vc
    k12 <- q/vc
    k21 <- q/vp
    d/dt(depot) <- -ka*depot
    d/dt(central) <-  ka*depot - kel*central - k12*central + k21*peripheral1
    d/dt(peripheral1) <- k12*central - k21*peripheral1
    cp <- central / vc
    cp ~ prop(propSd)
  })
}

ui <- rxode2(one.compartment)

rxode2(ui) <- two.compartment

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{binomProbs}
\alias{binomProbs}
\alias{binomProbs.default}
\title{Calculate expected confidence bands with binomial sampling distribution}
\usage{
binomProbs(x, ...)

\method{binomProbs}{default}(
  x,
  probs = c(0.025, 0.05, 0.5, 0.95, 0.975),
  na.rm = FALSE,
  names = TRUE,
  onlyProbs = TRUE,
  n = 0L,
  m = 0L,
  pred = FALSE,
  piMethod = "lim",
  M = 5e+05,
  tol = .Machine$double.eps^0.25,
  ciMethod = c("wilson", "wilsonCorrect", "agrestiCoull", "wald", "wc", "ac"),
  ...
)
}
\arguments{
\item{x}{numeric vector whose mean and probability based confidence
values are wanted, NA and NaN values are not allowed in numeric
vectors unless \code{na.rm} is \code{TRUE}.}

\item{...}{Arguments passed to default method, allows many
different methods to be applied.}

\item{probs}{numeric vector of probabilities with values in the
interval 0 to 1, inclusive. When 0, it represents the maximum
observed, when 1, it represents the maximum observed. When 0.5 it
represents the expected probability (mean).}

\item{na.rm}{logical; if true, any NA and NaN's are removed from
\code{x} before the quantiles are computed.}

\item{names}{logical; if true, the result has a names attribute.}

\item{onlyProbs}{logical; if true, only return the probability
based confidence interval/prediction interval estimates,
otherwise return extra statistics.}

\item{n}{integer/integerish; this is the n used to calculate the
prediction or confidence interval.  When \code{n=0} (default) use the
number of non-\code{NA} observations.  When calculating the prediction
interval, this represents the number of observations used in the
input ("true") distribution.}

\item{m}{integer.  When using the prediction interval this
represents the number of samples that will be observed in the
future for the prediction interval.}

\item{pred}{Use a prediction interval instead of a confidence
interval.  By default this is \code{FALSE}.}

\item{piMethod}{gives the prediction interval method (currently only lim) from Lu 2020}

\item{M}{number of simulations to run for the LIM PI.}

\item{tol}{tolerance of root finding in the LIM prediction interval}

\item{ciMethod}{gives the method for calculating the confidence
interval.

Can be:
\itemize{
\item "argestiCoull" or "ac" -- Agresti-Coull method. For a 95\\% confidence
interval, this method does not use the concept   of "adding 2
successes and 2 failures," but rather uses the formulas explicitly
described in the following link:
}

https://en.wikipedia.org/wiki/Binomial_proportion_confidence_interval#Agresti-Coull_Interval.
\itemize{
\item "wilson" -- Wilson Method
\item "wilsonCorrect" or "wc" -- Wilson method with continuity correction
\item "wald" -- Wald confidence interval or standard z approximation.
}}
}
\value{
By default the return has the probabilities as names (if
named) with the points where the expected distribution are
located given the sampling mean and standard deviation. If
\code{onlyProbs=FALSE} then it would prepend mean, variance, standard
deviation, minimum, maximum and number of non-NA observations.
}
\description{
This is meant to perform in the same way as \code{quantile()} so it can
be a drop in replacement for code using \code{quantile()} but using
distributional assumptions.
}
\details{
It is used for confidence intervals with rxode2 solved objects using
\code{confint(mean="binom")}
}
\examples{

x<- rbinom(7001, p=0.375, size=1)
binomProbs(x)

# you can also use the prediction interval
\donttest{
binomProbs(x, pred=TRUE)
}

# Can get some extra statistics if you request onlyProbs=FALSE
binomProbs(x, onlyProbs=FALSE)

x[2] <- NA_real_

binomProbs(x, onlyProbs=FALSE)

binomProbs(x, na.rm=TRUE)

}
\references{
\itemize{
\item Newcombe, R. G. (1998). "Two-sided confidence intervals for the single
proportion: comparison of seven methods". Statistics
in Medicine. 17 (8):
857-872. doi:10.1002/(SICI)1097-0258(19980430)17:8<857::AID-SIM777>3.0.CO;2-E. PMID
9595616.
\item Hezhi Lu, Hua Jin,
A new prediction interval for binomial random variable based on inferential models,
Journal of Statistical Planning and Inference,
Volume 205,
2020,
Pages 156-174,
ISSN 0378-3758,
https://doi.org/10.1016/j.jspi.2019.07.001.
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{rxGetSeed}
\alias{rxGetSeed}
\title{Get the rxode2 seed}
\usage{
rxGetSeed()
}
\value{
rxode2 seed state or -1 when the seed isn't set
}
\description{
Get the rxode2 seed
}
\examples{

# without setting seed

rxGetSeed()
# Now set the seed
rxSetSeed(42)

rxGetSeed()

rxnorm()

rxGetSeed()

# don't use the rxode2 seed again

rxSetSeed(-1)

rxGetSeed()

rxnorm()

rxGetSeed()

}
\seealso{
rxSetSeed, rxWithSeed, rxWithPreserveSeed
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getBaseSim.R
\name{getBaseIniSimModel}
\alias{getBaseIniSimModel}
\title{Get the base simulation model for simulation with inis in the
underlying rxode2 model}
\usage{
getBaseIniSimModel(obj)
}
\arguments{
\item{obj}{Fit Object}
}
\value{
Simulation object
}
\description{
Get the base simulation model for simulation with inis in the
underlying rxode2 model
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxsolve.R
\name{rxSolve}
\alias{rxSolve}
\alias{rxSolve.function}
\alias{rxSolve.rxUi}
\alias{rxSolve.rxode2tos}
\alias{rxSolve.nlmixr2FitData}
\alias{rxSolve.nlmixr2FitCore}
\alias{rxSolve.default}
\alias{update.rxSolve}
\alias{rxSolve.rxSolve}
\alias{predict.rxode2}
\alias{predict.function}
\alias{predict.rxUi}
\alias{predict.rxSolve}
\alias{predict.rxEt}
\alias{predict.rxParams}
\alias{simulate.rxode2}
\alias{simulate.rxSolve}
\alias{simulate.rxParams}
\alias{solve.rxSolve}
\alias{solve.rxUi}
\alias{solve.function}
\alias{solve.rxode2}
\alias{solve.rxParams}
\alias{solve.rxEt}
\alias{rxControl}
\title{Options, Solving & Simulation of an ODE/solved system}
\usage{
rxSolve(
  object,
  params = NULL,
  events = NULL,
  inits = NULL,
  scale = NULL,
  method = c("liblsoda", "lsoda", "dop853", "indLin", "f78", "rk4", "ck54", "ab", "abm",
    "dop5", "bs", "ros4", "iem", "sem", "sb3a", "sb3am4", "vv", "mm", "em", "cvode",
    "trapz", "ssp3", "f32", "rk43", "dop54", "vern65", "vern76", "dop87", "vern98",
    "ros43", "ros6", "backwardEuler", "gauss6", "iiic6", "radauiia5", "geng5", "sdirk43",
    "euler", "midpoint", "heun", "ssp22", "rk3", "ssp53", "s4", "r4", "ls44", "ls54",
    "ssp54", "s5", "rk5", "c5", "l5", "lk5a", "lk5b", "b6", "s7", "s8_10", "cv8",
    "s8_12", "s10", 
     "z10", "o10", "h10", "dp54", "v65e", "v76e", "dp87", "v98e",
    "ssp33", "bs32", "ssp43", "f45", "t54", "s54", "pp54", "pp54b", "bs54", "ss54",
    "dp65", "c65", "tp64", "v65r", "v65", "dverk65", "tf65", "tp75", "tmy7", "tmy7s",
    "v76r", "ss76", "v78", "dverk78", "dp85", "tp86", "v87e", "v87r", "ev87", "k87",
    "f89", "v89", "t98a", "v98r", "s98", "f108", "c108", "b109", "s1110a", "f1210",
    "o129", "f1412", "lsode", "bdf", "rk4s", "eulers", "midpoints", "heuns", "dop5s",
    "dop853s", "ck54s", "bs32s", "vern65s", 
     "vern76s", "dop87s", "f78s", "ros4s",
    "radauiia5s", "backwardEulers", "gauss6s", "sdirk43s", "iiic6s", "ros43s", "ros6s",
    "geng5s", "rk3s", "rk43s", "cvodesadj", "liblsodaadj", "abs", "dop54s", "dp54s",
    "vern98s", "f45s", "t54s", "pp54s", "pp54bs", "bs54s", "ss54s", "dp65s", "c65s",
    "tp64s", "v65rs", "dverk65s", "tf65s", "tp75s", "tmy7sadj", "tmy7adj", "v76rs",
    "ss76s", "v78s", "dverk78s", "dp85s", "tp86s", "v87es", "v87rs", "ev87s", "k87s",
    "v89s", "t98as", "v98rs", "s98s", "c108s", "b109s", 
     "s1110as", "o129s"),
  sigdig = NULL,
  atol = 1e-08,
  rtol = 1e-06,
  maxsteps = 70000L,
  hmin = 0,
  hmax = NA_real_,
  hmaxSd = 0,
  hini = 0,
  maxordn = 12L,
  maxords = 5L,
  order = 5L,
  ...,
  cores,
  covsInterpolation = c("locf", "linear", "nocb", "midpoint"),
  naInterpolation = c("locf", "nocb"),
  keepInterpolation = c("na", "locf", "nocb"),
  addCov = TRUE,
  sigma = NULL,
  sigmaDf = NULL,
  sigmaLower = -Inf,
  sigmaUpper = Inf,
  nCoresRV = 1L,
  sigmaIsChol = FALSE,
  sigmaSeparation = c("auto", "lkj", "separation"),
  sigmaXform = c("identity", "variance", "log", "nlmixrSqrt", "nlmixrLog",
    "nlmixrIdentity"),
  nDisplayProgress = 10000L,
  amountUnits = NA_character_,
  timeUnits = "hours",
  theta = NULL,
  thetaLower = -Inf,
  thetaUpper = Inf,
  eta = NULL,
  addDosing = FALSE,
  stateTrim = Inf,
  updateObject = FALSE,
  omega = NULL,
  omegaDf = NULL,
  omegaIsChol = FALSE,
  omegaSeparation = c("auto", "lkj", "separation"),
  omegaXform = c("variance", "identity", "log", "nlmixrSqrt", "nlmixrLog",
    "nlmixrIdentity"),
  omegaLower = -Inf,
  omegaUpper = Inf,
  nSub = 1L,
  thetaMat = NULL,
  thetaDf = NULL,
  thetaIsChol = FALSE,
  nStud = 1L,
  dfSub = 0,
  dfObs = 0,
  returnType = c("rxSolve", "matrix", "data.frame", "data.frame.TBS", "data.table",
    "tbl", "tibble"),
  seed = NULL,
  nsim = NULL,
  minSS = 10L,
  maxSS = 10000L,
  infSSstep = 12,
  strictSS = TRUE,
  istateReset = TRUE,
  subsetNonmem = TRUE,
  maxAtolRtolFactor = 0.1,
  from = NULL,
  to = NULL,
  by = NULL,
  length.out = NULL,
  iCov = NULL,
  keep = NULL,
  indLinPhiTol = 1e-07,
  indLinPhiM = 0L,
  indLinMatExpType = c("expokit", "Al-Mohy", "arma"),
  indLinMatExpOrder = 6L,
  drop = NULL,
  idFactor = TRUE,
  mxhnil = 0,
  hmxi = 0,
  warnIdSort = TRUE,
  warnDrop = TRUE,
  ssAtol = 1e-08,
  ssRtol = 1e-06,
  safeZero = TRUE,
  safeLog = TRUE,
  safePow = TRUE,
  sumType = c("pairwise", "fsum", "kahan", "neumaier", "c"),
  prodType = c("long double", "double", "logify"),
  resample = NULL,
  resampleID = TRUE,
  maxwhile = 1e+05,
  atolSens = 1e-08,
  rtolSens = 1e-06,
  ssAtolSens = 1e-08,
  ssRtolSens = 1e-06,
  simVariability = NA,
  nLlikAlloc = NULL,
  useStdPow = FALSE,
  naTimeHandle = c("ignore", "warn", "error"),
  addlKeepsCov = FALSE,
  addlDropSs = TRUE,
  ssAtDoseTime = TRUE,
  ss2cancelAllPending = FALSE,
  ssSolved = TRUE,
  linCmtSensType = c("auto", "endpoint5", "endpoint5G", "forward3", "forward3G", "AD",
    "ADr", "central", "forward", "forwardG", "forwardH", "centralH", "forward3H",
    "endpointH5", "forwardG"),
  linCmtSensH = 1e-04,
  linCmtGillFtol = 0,
  linCmtGillK = 20L,
  linCmtGillStep = 4,
  linCmtGillRtol = sqrt(.Machine$double.eps),
  linCmtShiErr = sqrt(.Machine$double.eps),
  linCmtShiMax = 20L,
  linCmtScale = FALSE,
  linCmtHcmt = NULL,
  linCmtHmeanI = c("geometric", "arithmetic", "harmonic"),
  linCmtHmeanO = c("geometric", "arithmetic", "harmonic"),
  linCmtSuspect = 1e-06,
  linCmtForwardMax = 2L,
  indOwnAlloc = NA,
  maxExtra = 1000L,
  tolFactor = NULL,
  serializeFile = NULL,
  dense = FALSE,
  cvodeLinSolver = c("dense", "band", "gmres", "bicgstab", "tfqmr"),
  autoSwitchNonstifftol = 9/10,
  autoSwitchStifftol = 9/10,
  autoSwitchDtfac = 2,
  autoSwitchMaxStiff = 10L,
  autoSwitchMaxNonstiff = 3L,
  autoSwitchStiffFirst = FALSE,
  autoSwitchSwitchMax = 5L,
  stiff2 = 0L,
  useLinCmt = getOption("rxode2.useLinCmt", TRUE),
  file = NULL,
  chunkSize = NULL,
  parallel = 0L,
  envir = parent.frame()
)

\method{rxSolve}{`function`}(
  object,
  params = NULL,
  events = NULL,
  inits = NULL,
  ...,
  theta = NULL,
  eta = NULL,
  envir = parent.frame()
)

\method{rxSolve}{rxUi}(
  object,
  params = NULL,
  events = NULL,
  inits = NULL,
  ...,
  useLinCmt = TRUE,
  theta = NULL,
  eta = NULL,
  envir = parent.frame()
)

\method{rxSolve}{rxode2tos}(
  object,
  params = NULL,
  events = NULL,
  inits = NULL,
  ...,
  useLinCmt = TRUE,
  theta = NULL,
  eta = NULL,
  envir = parent.frame()
)

\method{rxSolve}{nlmixr2FitData}(
  object,
  params = NULL,
  events = NULL,
  inits = NULL,
  ...,
  theta = NULL,
  eta = NULL,
  envir = parent.frame()
)

\method{rxSolve}{nlmixr2FitCore}(
  object,
  params = NULL,
  events = NULL,
  inits = NULL,
  ...,
  theta = NULL,
  eta = NULL,
  envir = parent.frame()
)

\method{rxSolve}{default}(
  object,
  params = NULL,
  events = NULL,
  inits = NULL,
  ...,
  indOwnAlloc = TRUE,
  theta = NULL,
  eta = NULL,
  envir = parent.frame()
)

\method{update}{rxSolve}(object, ...)

\method{rxSolve}{rxSolve}(
  object,
  params = NULL,
  events = NULL,
  inits = NULL,
  ...,
  theta = NULL,
  eta = NULL,
  envir = parent.frame()
)

\method{predict}{rxode2}(object, ...)

\method{predict}{`function`}(object, ...)

\method{predict}{rxUi}(object, ...)

\method{predict}{rxSolve}(object, ...)

\method{predict}{rxEt}(object, ...)

\method{predict}{rxParams}(object, ...)

\method{simulate}{rxode2}(object, nsim = 1L, seed = NULL, ...)

\method{simulate}{rxSolve}(object, nsim = 1L, seed = NULL, ...)

\method{simulate}{rxParams}(object, nsim = 1L, seed = NULL, ...)

\method{solve}{rxSolve}(a, b, ...)

\method{solve}{rxUi}(a, b, ...)

\method{solve}{`function`}(a, b, ...)

\method{solve}{rxode2}(a, b, ...)

\method{solve}{rxParams}(a, b, ...)

\method{solve}{rxEt}(a, b, ...)

rxControl(
  ...,
  params = NULL,
  events = NULL,
  inits = NULL,
  envir = parent.frame()
)
}
\arguments{
\item{object}{is a either a rxode2 family of objects, or a file-name
with a rxode2 model specification, or a string with a rxode2
model specification.}

\item{params}{a numeric named vector with values for every
parameter in the ODE system; the names must correspond to the
parameter identifiers used in the ODE specification;}

\item{events}{an \code{eventTable} object describing the input
(e.g., doses) to the dynamic system and observation sampling
time points (see \code{\link[=eventTable]{eventTable()}});}

\item{inits}{a vector of initial values of the state variables
(e.g., amounts in each compartment), and the order in this
vector must be the same as the state variables (e.g., PK/PD
compartments);}

\item{scale}{a numeric named vector with scaling for ode
parameters of the system.  The names must correspond to the
parameter identifiers in the ODE specification. Each of the
ODE variables will be divided by the scaling factor.  For
example \code{scale=c(center=2)} will divide the center ODE
variable by 2.}

\item{method}{The method for solving ODEs.  Currently this supports:
\itemize{
\item \code{"liblsoda"} thread safe lsoda.  This supports parallel
thread-based solving, and ignores user Jacobian specification.
\item \code{"lsoda"} -- LSODA solver.  Does not support parallel thread-based
solving, but allows user Jacobian specification.
\item \code{"dop853"} -- DOP853 solver.  Does not support parallel thread-based
solving nor user Jacobian specification
\item \code{"indLin"} -- Solving through inductive linearization.  The rxode2 dll
must be setup specially to use this solving routine.
\item \code{"f78"} -- Runge-Kutta Fehlberg 78 solver using Boost's odeint library.
\item \code{"rk4"} -- Runge-Kutta 4 solver using Boost's odeint library.
\item \code{"ck54"} -- Cash-Karp 5(4) solver using Boost's odeint library.
\item \code{"ab"} -- Adams-Bashforth multi-step solver with a direct implementation
(order 1-8, default 5). Uses RK4 to initialize the derivative history and
then applies the Adams-Bashforth extrapolation formula. Is a fixed-step
method (step size controlled by \code{hmin}).
\item \code{"abm"} -- Adams-Bashforth-Moulton solver using Boost's odeint library.
\item \code{"dop5"} -- DOPRI5 solver using Boost's odeint library (supports dense output).
\item \code{"bs"} -- Bulirsch-Stoer solver using Boost's odeint library (supports dense output).
\item \code{"ros4"} \strong{(implicit)} -- Rosenbrock 4 solver using Boost's odeint library (supports dense output).
Requires an analytical Jacobian (auto-computed from the model when not provided);
falls back to \code{liblsoda} if Jacobian generation fails.
\item \code{"ros43"} \strong{(implicit)} -- Kaps-Rentrop 4th-order A-stable Rosenbrock method (GRK4A) from libode.
Adaptive step size via embedded 3rd-order error estimate.
Requires an analytical Jacobian (auto-computed from the model when not provided);
falls back to \code{liblsoda} if Jacobian generation fails.
\item \code{"ros6"} \strong{(implicit)} -- Kaps-Wanner 6th-order A-stable Rosenbrock method (ROW6A) from libode.
Fixed step size (set with \code{hmin}). Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"backwardEuler"} \strong{(implicit)} -- Backward Euler (BDF-1), 1st-order L-stable fully implicit method from libode.
Fixed step size (set with \code{hmin}). Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"gauss6"} \strong{(implicit)} -- Gauss-Legendre 6th-order A-stable fully-implicit method (3 stages) from libode.
Fixed step size. Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"iiic6"} \strong{(implicit)} -- Lobatto IIIC 6th-order L-stable fully-implicit method (4 stages) from libode.
Fixed step size. Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"radauiia5"} \strong{(implicit)} -- Radau IIA 5th-order L-stable fully-implicit method (3 stages) from libode.
Fixed step size (set with \code{hmin}). Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"geng5"} \strong{(implicit)} -- Geng 5th-order symplectic fully-implicit method (3 stages) from libode.
Fixed step size. Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"sdirk43"} \strong{(implicit)} -- L-stable SDIRK 4(3) pair from libode.
Adaptive step size via embedded 3rd-order error estimate.
Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"iem"} \strong{(implicit)} -- Implicit Euler solver using Boost's odeint library.
Requires an explicit Jacobian (auto-generated via \code{calcJac=TRUE} when not provided).
Uses Boost.uBLAS vectors and is a fixed-step method (step size controlled by \code{hmin}).
\item \code{"sem"} -- Symplectic Euler solver using Boost's odeint library.
Requires splitting the state into coordinate-momentum pairs (and
automatically pads odd-dimensional systems).  Is a fixed-step
method (step size controlled by \code{hmin}).
\item \code{"sb3a"} -- Symplectic Runge-Kutta-Nystrom SB3A McLachlan stepper
using Boost's odeint library.  Requires splitting the state into
coordinate-momentum pairs (and automatically pads odd-dimensional
systems).  Is a fixed-step method (step size controlled by
\code{hmin}).
\item \code{"sb3am4"} -- Symplectic Runge-Kutta-Nystrom SB3A M4 McLachlan
stepper using Boost's odeint library.  Requires splitting the
state into coordinate-momentum pairs (and automatically pads
odd-dimensional systems).  Is a fixed-step method (step size
controlled by \code{hmin}).
\item \code{"vv"} -- Velocity Verlet stepper using Boost's odeint library.
Requires splitting the state into coordinate-momentum pairs (and
automatically pads odd-dimensional systems).  Is a fixed-step
method (step size controlled by \code{hmin}).
\item \code{"mm"} -- Modified Midpoint stepper using Boost's odeint library.
Is a fixed-step method (step size controlled by \code{hmin}) and uses
the \code{order} parameter to configure the number of intermediate
steps.
\item \code{"em"} -- Explicit Euler stepper using Boost's odeint library.
Is a fixed-step method (step size controlled by \code{hmin}).
\item \code{"cvode"} -- CVODE BDF stiff solver from the SUNDIALS library (sources
and headers vendored into rxode2).  Supports thread-parallel solving and
per-compartment absolute tolerances.
\item \code{"trapz"} -- Explicit trapezoidal rule (Heun's method), a 2nd-order
fixed-step Runge-Kutta method implemented via the libode library.
Each inter-event interval is integrated with fixed steps of size
\code{hmin} (default \code{0.01} when \code{hmin=0}).  If \code{hmin} would require
more than \code{maxsteps} steps to span the interval, the step size is
automatically enlarged to stay within that limit.  When an
inter-event interval is shorter than the nominal step size (e.g.,
two doses very close together), the step is silently clamped to the
interval length so that short intervals are always handled correctly.
Supports parallel thread-based solving and steady-state (\code{ss=1})
dosing.  The method does not use \code{atol} or \code{rtol} (fixed-step; no
error control).  Steady-state convergence is governed by \code{ssAtol},
\code{ssRtol}, \code{minSS}, \code{maxSS}, and \code{strictSS}.
\item \code{"ssp3"} -- Strong Stability-Preserving Runge-Kutta of order 3 (SSP-RK3,
Shu-Osher method), implemented via the libode library.  A 3-stage,
3rd-order explicit fixed-step method with superior non-oscillatory
properties for problems with discontinuities or sharp fronts.  Uses
the Butcher tableau c2=1, a21=1; c3=1/2, a31=1/4, a32=1/4;
b1=1/6, b2=1/6, b3=2/3.  The step-size behavior, clamping, and
steady-state options (\code{hmin}, \code{maxsteps}, \code{ssAtol}, \code{ssRtol},
\code{minSS}, \code{maxSS}, \code{strictSS}) are identical to \code{"trapz"}.  Does
not use \code{atol} or \code{rtol} (fixed-step; no error control).  Supports
parallel thread-based solving and steady-state (\code{ss=1}) dosing.
\item \code{"f32"} -- Fehlberg 3(2) embedded pair from libode (class \code{OdeRKF32}).
A 3-stage, 3rd-order adaptive method with a built-in 2nd-order error
estimate for automatic step-size control.  Tableau: c2=1, a21=1;
c3=1/2, a31=1/4, a32=1/4; primary (3rd-order) weights d1=1/6, d2=1/6,
d3=4/6; embedded (2nd-order) weights b1=1/2, b2=1/2.  Uses \code{atol} and \code{rtol}
for error control.  The \code{hmin}
parameter sets the initial step size (default \code{0.01}); subsequent steps
are chosen adaptively.  The total number of accepted and rejected steps
is bounded by \code{maxsteps}.  Supports parallel thread-based solving and
steady-state (\code{ss=1}) dosing with convergence governed by \code{ssAtol},
\code{ssRtol}, \code{minSS}, \code{maxSS}, and \code{strictSS}.
\item \code{"rk43"} -- Runge-Kutta 4(3) pair from libode.  A 5-stage, 4th-order
adaptive method with a built-in 3rd-order embedded error estimate for
automatic step-size control.  Tableau: c2=1/3, a21=1/3; c3=2/3,
a31=-1/3, a32=1; c4=1, a41=1, a42=-1, a43=1; a5j=bj (FSAL).
Primary (4th-order) weights b1=b4=1/8, b2=b3=3/8; embedded (3rd-order)
weights d1=1/12, d2=1/2, d3=1/4, d5=1/6.  Since a5j=bj the 5th stage
evaluation is reused as the 1st stage of the next step (FSAL).
Uses \code{atol} and \code{rtol} for error control.  The \code{hmin} parameter sets
the initial step size (default \code{0.01}); subsequent steps are chosen
adaptively.  The
total number of steps is bounded by \code{maxsteps}.  Supports parallel
thread-based solving and steady-state (\code{ss=1}) dosing with convergence
governed by \code{ssAtol}, \code{ssRtol}, \code{minSS}, \code{maxSS}, and \code{strictSS}.
\item \code{"dop54"} -- Dormand-Prince 5(4) FSAL method (Dormand & Prince 1980),
implemented via the libode library.  The same algorithm as MATLAB's
\code{ode45}.  A 7-stage, 5th-order adaptive method with an embedded
4th-order error estimate for automatic step-size control (FSAL: the
7th stage evaluation is reused as the 1st stage of the next step,
giving effectively 6 function evaluations per accepted step).  Uses
\code{atol} and \code{rtol} for error control.  The \code{hmin} parameter sets the
initial step size (default \code{0.01}); subsequent steps are chosen
adaptively.  The total number of steps is bounded by \code{maxsteps}.
Supports parallel thread-based solving and steady-state (\code{ss=1})
dosing with convergence governed by \code{ssAtol}, \code{ssRtol}, \code{minSS},
\code{maxSS}, and \code{strictSS}.  NaN/Inf in derivatives (e.g. from NA
parameters) is detected immediately and the solve exits with NA
output for the affected subject.
\item \code{"vern65"} -- Jim Verner's "most efficient" 6(5) FSAL pair (9 stages),
implemented via the libode library using coefficients from Verner's own
website (RKV65.IIIXb.Efficient).  A 6th-order adaptive method with an
embedded 5th-order error estimate.  The sparse tableau (many zero
a-coefficients) reduces function evaluations; the FSAL property means the
9th stage evaluation is reused as the 1st stage of the next step, giving
effectively 8 function evaluations per accepted step.  Uses \code{atol} and
\code{rtol} for error control.  The \code{hmin} parameter sets the initial step
size (default \code{0.01}); subsequent steps are chosen adaptively.  The total
number of steps is bounded by \code{maxsteps}.  Supports parallel thread-based
solving and steady-state (\code{ss=1}) dosing with convergence governed by
\code{ssAtol}, \code{ssRtol}, \code{minSS}, \code{maxSS}, and \code{strictSS}.  NaN/Inf in
derivatives is detected immediately and the solve exits with NA output.
\item \code{"vern76"} -- Jim Verner's "most efficient" 7(6) pair (10 stages),
implemented via the libode library using coefficients from Verner's own
website (RKV76.IIa.Efficient).  A 7th-order adaptive method with an
embedded 6th-order error estimate.  The sparse tableau reduces function
evaluations per step.  Uses \code{atol} and \code{rtol} for error control.  The
\code{hmin} parameter sets the initial step size (default \code{0.01}); subsequent
steps are chosen adaptively.  The total number of steps is bounded by
\code{maxsteps}.  Supports parallel thread-based solving and steady-state
(\code{ss=1}) dosing with convergence governed by \code{ssAtol}, \code{ssRtol},
\code{minSS}, \code{maxSS}, and \code{strictSS}.  NaN/Inf in derivatives is detected
immediately and the solve exits with NA output.
\item \code{"dop87"} -- Dormand-Prince 8(7) pair (13 stages), implemented via the
libode library using coefficients from Hairer, Norsett and Wanner (1993)
"Solving ODEs I" (2nd ed.).  An 8th-order adaptive method with an
embedded 7th-order error estimate.  Both solutions are computed from the
original state in the final step loop.  Uses \code{atol} and \code{rtol} for error
control.  The \code{hmin} parameter sets the initial step size (default
\code{0.01}); subsequent steps are chosen adaptively.  The total number of
steps is bounded by \code{maxsteps}.  Supports parallel thread-based solving
and steady-state (\code{ss=1}) dosing with convergence governed by \code{ssAtol},
\code{ssRtol}, \code{minSS}, \code{maxSS}, and \code{strictSS}.  NaN/Inf in derivatives is
detected immediately and the solve exits with NA output.
\item \code{"vern98"} -- Jim Verner's "most efficient" 9(8) pair (16 stages),
implemented via the libode library using coefficients from Verner's own
website (RKV98.IIa.Efficient).  A 9th-order adaptive method with an
embedded 8th-order error estimate.  The highly sparse tableau (stages
8-16 use only \eqn{k_{0}}{x_0} and \eqn{k_{5..}}{k_5..}) minimizes function evaluations per
step.  Both solutions are computed from the original state in the final
step loop.  Uses \code{atol} and \code{rtol} for error control.  The \code{hmin}
parameter sets the initial step size (default \code{0.01}); subsequent steps
are chosen adaptively.  The total number of steps is bounded by
\code{maxsteps}.  Supports parallel thread-based solving and steady-state
(\code{ss=1}) dosing with convergence governed by \code{ssAtol}, \code{ssRtol},
\code{minSS}, \code{maxSS}, and \code{strictSS}.  NaN/Inf in derivatives is detected
immediately and the solve exits with NA output.
}

\strong{Aliases for existing methods} (no additional C++ code; \code{hmin}, \code{atol},
\code{rtol}, and \code{maxsteps} follow the aliased method):
\itemize{
\item \code{"dp54"} -- alias for \code{"dop54"} (Dormand-Prince 5(4) FSAL, libode).
\item \code{"v65e"} -- alias for \code{"vern65"} (Verner 6(5) efficient, libode).
\item \code{"v76e"} -- alias for \code{"vern76"} (Verner 7(6) efficient, libode).
\item \code{"dp87"} -- alias for \code{"dop87"} (Dormand-Prince 8(7), libode).
\item \code{"v98e"} -- alias for \code{"vern98"} (Verner 9(8) efficient, libode).
\item \code{"ssp33"} -- alias for \code{"ssp3"} (Strong Stability-Preserving RK3, libode).
}

\strong{libode fixed-step explicit methods}.  All use \code{hmin} as the fixed step
size (default \code{0.01} when \code{hmin=0}); \code{atol} and \code{rtol} are ignored (no
error control); \code{maxsteps} bounds the total number of steps; NaN/Inf in
derivatives sets \code{ind$err} and the subject exits with NA output.  All
support parallel thread-based solving.
\itemize{
\item \code{"euler"} -- Forward (explicit) Euler, 1st-order, 1 stage.  Requires a
very small step size for accuracy (e.g., \code{hmin=1e-4}); intended for
pedagogical use or method-comparison baselines.
\item \code{"midpoint"} -- Explicit midpoint rule, 2nd-order, 2 stages.
\item \code{"heun"} -- Heun's method (explicit trapezoid), 2nd-order, 2 stages.
Identical in structure to \code{"trapz"} but uses the libode fixed-step driver.
\item \code{"ssp22"} -- Strong Stability-Preserving 2-stage 2nd-order method
(SSP-RK22), 2 stages.  Superior non-oscillatory properties near
discontinuities.
\item \code{"rk3"} -- Classical 3rd-order Runge-Kutta (Kutta 1901), 3 stages.
\item \code{"ssp53"} -- Strong Stability-Preserving 5-stage 3rd-order method
(SSP-RK53), 5 stages.  High SSP coefficient for hyperbolic PDEs or
event-heavy ODE systems.
\item \code{"s4"} -- Shanks 4th-order method, 4 stages.
\item \code{"r4"} -- Ralston's 4th-order method, 4 stages.  Minimizes local
truncation error among classical 4-stage 4th-order methods.
\item \code{"ls44"} -- Low-storage 4th-order method, 4 stages.  Uses a 2-register
update scheme that minimizes memory bandwidth at the cost of a less
general tableau structure.
\item \code{"ls54"} -- Low-storage 4th-order method, 5 stages.  Five-stage
variant of the 2-register low-storage scheme.
\item \code{"ssp54"} -- Strong Stability-Preserving 5-stage 4th-order method
(SSP-RK54), 5 stages.
\item \code{"s5"} -- Shanks 5th-order method, 5 stages.
\item \code{"rk5"} -- Classical 5th-order Runge-Kutta, 6 stages.
\item \code{"c5"} -- Cassity 5th-order method, 6 stages.
\item \code{"l5"} -- Lawson 5th-order method, 6 stages.
\item \code{"lk5a"} -- Luther-Konen 5th-order method, variant A, 6 stages.
\item \code{"lk5b"} -- Luther-Konen 5th-order method, variant B, 6 stages.
\item \code{"b6"} -- Butcher 6th-order method, 7 stages.
\item \code{"s7"} -- Shanks 7th-order method, 9 stages.
\item \code{"s8_10"} -- Shanks 8th-order method, 10 stages.
\item \code{"cv8"} -- Cooper-Verner 8th-order method, 11 stages.
\item \code{"s8_12"} -- Shanks 8th-order method, 12 stages.
\item \code{"s10"} -- Stepanov 10th-order method, 15 stages.  Requires a moderate
step size (e.g., \code{hmin=1.0}) because a single step is accurate to very
high order.
\item \code{"z10"} -- Zhang 10th-order method, 16 stages.
\item \code{"o10"} -- Ono 10th-order method, 17 stages.
\item \code{"h10"} -- Hairer 10th-order method, 17 stages.
}

\strong{libode adaptive (variable-step) explicit methods}.  All use \code{atol} and
\code{rtol} for error control;
\code{hmin} sets the initial step size (default \code{0.01} when \code{hmin=0}); \code{hmax}
sets the maximum step size; \code{maxsteps} bounds total steps; NaN/Inf in
derivatives sets \code{ind$err} and the subject exits with NA output.  All
support parallel thread-based solving and steady-state (\code{ss=1}) dosing
(convergence governed by \code{ssAtol}, \code{ssRtol}, \code{minSS}, \code{maxSS}, \code{strictSS}).
\itemize{
\item \code{"bs32"} -- Bogacki-Shampine 3(2) FSAL pair, 4 stages (Bogacki &
Shampine 1989).  3rd-order primary with 2nd-order embedded error estimate.
FSAL: the 4th-stage evaluation is reused as the 1st stage of the next
step.  The same algorithm as Julia \code{BS3()} and MATLAB \code{ode23}.
\item \code{"ssp43"} -- Strong Stability-Preserving 4(3) pair, 4 stages.
Adaptive SSP method with a 3rd-order embedded error estimate.
\item \code{"f45"} -- Fehlberg 4(5) pair, 6 stages (Fehlberg 1970).  4th-order
primary solution with a 5th-order embedded estimate used for error
control.
\item \code{"t54"} -- Tsitouras 5(4) FSAL pair, 7 stages (Tsitouras 2011).
5th-order primary with 4th-order embedded error estimate.  FSAL: the
7th stage is reused as the 1st stage of the next step.  The same
Butcher tableau as Julia's \code{Tsit5()}.
\item \code{"s54"} -- Stepanov 5(4) FSAL pair, 7 stages.  5th-order primary
with 4th-order embedded error estimate.
\item \code{"pp54"} -- Papakostas-Papageorgiou 5(4) FSAL pair, 7 stages.
\item \code{"pp54b"} -- Papakostas-Papageorgiou 5(4) variant B FSAL pair,
7 stages.
\item \code{"bs54"} -- Bogacki-Shampine 5(4) pair, 8 stages.  5th-order primary
with 4th-order embedded error estimate (non-FSAL).
\item \code{"ss54"} -- Sharp-Smart 5(4) pair, 7 stages.
\item \code{"dp65"} -- Dormand-Prince 6(5) pair, 8 stages.  6th-order primary
with 5th-order embedded error estimate.
\item \code{"c65"} -- Calvo 6(5) pair, 9 stages.
\item \code{"tp64"} -- Tsitouras-Papakostas 6(4) pair, 7 stages.  6th-order
primary with 4th-order embedded error estimate.
\item \code{"v65r"} -- Verner "robust" 6(5) FSAL pair, 9 stages.  Robust
variant of Verner's 6(5) family with wider stability region.
\item \code{"v65"} -- Verner 6(5) pair, 8 stages (non-FSAL).
\item \code{"dverk65"} -- Verner DVERK 6(5) pair, 8 stages.  Coefficients from
the classic DVERK Fortran code distributed by Hull and Enright.
\item \code{"tf65"} -- Tsitouras-Famelis 6(5) FSAL pair, 9 stages.
\item \code{"tp75"} -- Tsitouras-Papakostas 7(5) pair, 9 stages.  7th-order
primary with 5th-order embedded error estimate.
\item \code{"tmy7"} -- Tanaka-Muramatsu-Yamashita 7th-order pair, 10 stages.
The same family as Julia's \code{TanYam7()}.
\item \code{"tmy7s"} -- Tanaka-Muramatsu-Yamashita 7th-order stable variant,
10 stages.  Alternative coefficient set with wider stability region.
\item \code{"v76r"} -- Verner "robust" 7(6) pair, 10 stages.
\item \code{"ss76"} -- Sharp-Smart 7(6) pair, 11 stages.
\item \code{"v78"} -- Verner 7(8) pair, 13 stages.  7th-order primary with
8th-order embedded error estimate.  Closest rxode2 analog to Julia
\code{Vern8()}.
\item \code{"dverk78"} -- Verner DVERK 7(8) pair, 13 stages.  Coefficients from
the classic DVERK Fortran code; companion to \code{"dverk65"}.  Also close
to Julia \code{Vern8()}.
\item \code{"dp85"} -- Dormand-Prince 8(5) pair, 12 stages.  8th-order primary
with 5th-order embedded error estimate.
\item \code{"tp86"} -- Tsitouras-Papakostas 8(6) pair, 12 stages.  The same
family as Julia's \code{TsitPap8()}.
\item \code{"v87e"} -- Verner "efficient" 8(7) pair, 13 stages.
\item \code{"v87r"} -- Verner "robust" 8(7) pair, 13 stages.
\item \code{"ev87"} -- Enright-Verner 8(7) pair, 13 stages.
\item \code{"k87"} -- Kovalnogov-Fedorov-Karpukhina-Simos 8(7) pair, 13 stages.
\item \code{"f89"} -- Fehlberg 8(9) pair, 17 stages.  8th-order primary with
9th-order embedded estimate.  Same family as MATLAB \code{ode89}.
\item \code{"v89"} -- Verner 8(9) pair, 16 stages.  Alternative to \code{"f89"}
in the same order bracket.  Also close to MATLAB \code{ode89}.
\item \code{"t98a"} -- Tsitouras 9(8) variant A pair, 16 stages.
\item \code{"v98r"} -- Verner "robust" 9(8) pair, 16 stages.
\item \code{"s98"} -- Sharp 9(8) pair, 16 stages.
\item \code{"f108"} -- Feagin 10(8) pair, 17 stages (Feagin 2007).  10th-order
primary with 8th-order embedded error estimate.  The same method as
Julia's \code{Feagin10()}.  The large number of stages carries elevated
transcription-error risk; verified against Williams' libode canonical order test.
\item \code{"c108"} -- Curtis 10(8) pair, 21 stages.
\item \code{"b109"} -- Baker 10(9) pair, 21 stages.
\item \code{"s1110a"} -- Stone 11(10) variant A pair, 26 stages.  11th-order
primary; very few published references.
\item \code{"f1210"} -- Feagin 12(10) pair, 25 stages (Feagin 2007).  12th-order
primary with 10th-order embedded error estimate.  The same method as
Julia's \code{Feagin12()}.  Elevated transcription-error risk due to many
stages; verified against Williams' libode canonical order test.
\item \code{"o129"} -- Ono 12(9) pair, 29 stages.
\item \code{"f1412"} -- Feagin 14(12) pair, 35 stages (Feagin 2007).  14th-order
primary with 12th-order embedded error estimate.  The same method as
Julia's \code{Feagin14()}.  The highest-order method in rxode2; elevated
transcription-error risk due to 35 stages; verified against
Williams' libode canonical order test.
}

\strong{deSolve-derived Fortran solvers} (from the deSolve R package).  Both
use non-reentrant Fortran COMMON blocks and therefore run single-threaded
only, regardless of the \code{cores} setting.  They are BDF (Backward
Differentiation Formula) stiff solvers and are most useful when you want
behavior comparable to deSolve's \code{lsode()} or \code{bdf()}/\code{vode()} functions.
\itemize{
\item \code{"lsode"} -- DLSODE (Hindmarsh 1983, ODEPACK) in Adams (non-stiff) mode,
MF=10.  Variable-order Adams method, order 1-12; no Jacobian evaluation.
Corresponds to deSolve's \code{lsode(method="adams")}.  Adaptive step-size;
error controlled by \code{atol} and \code{rtol}.
\strong{Not thread-safe} -- always runs on a single core.
\item \code{"bdf"} -- DLSODE (Hindmarsh 1983, ODEPACK) in BDF (stiff) mode, MF=22.
Variable-order BDF method, order 1-5; internally generated dense Jacobian.
Corresponds to deSolve's \code{bdf()} / \code{vode(method="bdf")}.  Adaptive
step-size; error controlled by \code{atol} and \code{rtol}.
\strong{Not thread-safe} -- always runs on a single core.
}}

\item{sigdig}{Specifies the "significant digits" that the ODE
solving requests.  This is \code{NULL} by default, and while it is
\code{NULL} it has no effect at all: \code{rxSolve()} uses the standard
\code{atol}/\code{rtol} (and the standard sensitivity and steady-state
tolerances).  \code{sigdig} only changes a tolerance when you ask for
it explicitly.

When it is supplied, the tolerances are derived with one
solver-independent formula -- the same for stiff, non-stiff and
auto-switching solvers.  The \code{rtol} exponent IS \code{sigdig} and
\code{atol} sits three orders below it:
\itemize{
\item \code{rtol = 10^(-sigdig)}, \code{atol = 10^(-sigdig-3)}
\item the sensitivity tolerances match the main solve, so
\code{rtolSens = rtol} and \code{atolSens = atol} (gradients and
covariances are built from them)
\item the steady-state tolerances run one order looser than the
corresponding main tolerance, so \code{ssRtol = ssRtolSens = 10*rtol}
and \code{ssAtol = ssAtolSens = 10*atol}
}

Each of these is set only when you did not pass that tolerance
yourself; a tolerance you supply always wins.  Because they are
resolved independently, an explicit \code{atol}/\code{rtol} overrides the
main solve but does \emph{not} propagate to the sensitivity or
steady-state tolerances -- set those directly if you need them
changed too.

This mapping matches how \code{nlmixr2est} derives solver tolerances
from its optimization \code{sigdig}, so a \code{sigdig} used for estimation
and the same \code{sigdig} used for a plain \code{rxSolve()} mean the same
thing.  Note it is keyed to \code{sigdig} as a request for that many
significant digits, and is looser than the \code{atol}/\code{rtol} defaults
for small \code{sigdig} -- at \code{sigdig = 4} it gives \code{rtol = 1e-4}
against a default \code{rtol = 1e-6}.  Raise \code{sigdig}, or set
\code{atol}/\code{rtol} directly, when you want a tighter solve.}

\item{atol}{a numeric absolute tolerance (1e-8 by default) used
by the ODE solver to determine if a good solution has been
achieved;  This is also used in the solved linear model to check
if prior doses do not add anything to the solution.}

\item{rtol}{a numeric relative tolerance (\code{1e-6} by default) used
by the ODE solver to determine if a good solution has been
achieved. This is also used in the solved linear model to check
if prior doses do not add anything to the solution.}

\item{maxsteps}{maximum number of (internally defined) steps allowed
during one call to the solver. (5000 by default)}

\item{hmin}{The minimum absolute step size allowed. The default
value is 0.

\if{html}{\out{<div class="sourceCode">}}\preformatted{For the fixed-step Boost methods `"rk4"`, `"trapz"`, `"ssp3"`, `"ab"`,
`"abm"`, `"sem"`, `"sb3a"`, `"sb3am4"`, `"vv"`, `"mm"`, `"em"`, `"ros6"`,
`"backwardEuler"`, `"gauss6"`, `"iiic6"`, `"radauiia5"`, `"geng5"`, and the
libode fixed-step family (`"euler"`, `"midpoint"`, `"heun"`, `"ssp22"`,
`"rk3"`, `"ssp53"`, `"s4"`, `"r4"`, `"ls44"`, `"ls54"`,
`"ssp54"`, `"s5"`, `"rk5"`, `"c5"`, `"l5"`, `"lk5a"`, `"lk5b"`,
`"b6"`, `"s7"`, `"s8_10"`, `"cv8"`, `"s8_12"`, `"s10"`, `"z10"`,
`"o10"`, `"h10"`), this specifies the fixed step size.
If `hmin=0` (the default), it uses a default of `0.01` for `"rk4"`,
`"trapz"`, `"ssp3"`, `"ros6"`, `"backwardEuler"`, `"gauss6"`, `"iiic6"`,
`"radauiia5"`, `"geng5"`, and all libode fixed-step methods; `0.0001` for
`"ab"`, `"abm"`, `"sem"`, `"sb3a"`, `"sb3am4"`, `"vv"`, `"mm"`, and `"em"`.
If the requested step size would cause the number of steps to exceed
`maxsteps`, the step size is automatically increased to ensure the
integration completes within the `maxsteps` limit.  For `"trapz"`, the step
is also silently clamped to the interval length when the inter-event
interval is shorter than the nominal step size, so short intervals (e.g.,
between closely spaced doses) are always handled correctly.

For the adaptive methods `"f78"`, `"ck54"`, `"dop5"`, `"bs"`, `"f32"`,
`"rk43"`, `"dop54"`, `"vern65"`, `"vern76"`, `"dop87"`, `"vern98"`,
`"ros43"`, `"sdirk43"`, and all libode adaptive methods (`"bs32"`,
`"ssp43"`, `"f45"`, `"t54"`, `"s54"`, `"pp54"`, `"pp54b"`,
`"bs54"`, `"ss54"`, `"dp65"`, `"c65"`, `"tp64"`, `"v65r"`,
`"v65"`, `"dverk65"`, `"tf65"`, `"tp75"`, `"tmy7"`, `"tmy7s"`,
`"v76r"`, `"ss76"`, `"v78"`, `"dverk78"`, `"dp85"`, `"tp86"`,
`"v87e"`, `"v87r"`, `"ev87"`, `"k87"`, `"f89"`, `"v89"`,
`"t98a"`, `"v98r"`, `"s98"`, `"f108"`, `"c108"`, `"b109"`,
`"s1110a"`, `"f1210"`, `"o129"`, `"f1412"`, the libode aliases
`"dp54"`, `"v65e"`, `"v76e"`, `"dp87"`, `"v98e"`,
`"ssp33"`, and the deSolve-derived methods `"lsode"`, `"bdf"`),
this specifies the initial step size (default `0.01` when
`hmin=0`); subsequent steps are chosen adaptively using `atol`, `rtol`,
and `maxsteps`.
}\if{html}{\out{</div>}}}

\item{hmax}{The maximum absolute step size allowed.  When
\code{hmax=NA} (default), uses the average difference +
hmaxSd*sd in times and sampling events. The \code{hmaxSd} is a user
specified parameter and which defaults to zero.  When
\code{hmax=NULL} rxode2 uses the maximum difference in times in
your sampling and events.  The value 0 is equivalent to infinite
maximum absolute step size.

Note that for dense output methods (\code{"dop853"}, \code{"dop5"}, \code{"bs"},
\code{"ros4"}), \code{hmax} defaults to \code{NULL} to allow the solvers to
determine the step size when \code{dense=TRUE}

For \code{method="indLin"}, \code{hmax} caps how long an interval is treated
as having a CONSTANT Jacobian/matrix-exponential term before
re-linearizing (previously silently ignored for this method). For a
true (state-independent) \code{matExp()} model this makes no numerical
difference; for an \code{indLin()}-forcing model with a state-dependent
term (e.g. Michaelis-Menten elimination), the default \code{hmax} (tied to
the spacing of your own sampling/dosing times) may be too coarse for
the desired accuracy -- set an explicit, smaller \code{hmax} to force more
frequent relinearization.}

\item{hmaxSd}{The number of standard deviations of the time
difference to add to hmax. The default is 0}

\item{hini}{The step size to be attempted on the first step. The
default value is determined by the solver (when \code{hini = 0})}

\item{maxordn}{The maximum order to be allowed for the nonstiff
(Adams) method.  The default is 12.  It can be between 1 and
12.}

\item{maxords}{The maximum order to be allowed for the stiff (BDF)
method.  The default value is 5.  This can be between 1 and 5.}

\item{order}{The order for the \code{"ab"}, \code{"abm"}, and \code{"mm"} methods.
For \code{"ab"} and \code{"abm"}, the default is 5, and it can be between 1 and 8.
For \code{"mm"} (Modified Midpoint), it represents the number of intermediate steps,
the default is 5, and it must be a positive integer (>= 1).}

\item{...}{Other arguments including scaling factors for each
compartment.  This includes \code{S# =} numeric will scale a compartment
\code{#} by a dividing the compartment amount by the scale factor,
like NONMEM.}

\item{cores}{Number of cores used in parallel ODE solving.  This
is equivalent to calling \code{\link[=setRxThreads]{setRxThreads()}}}

\item{covsInterpolation}{specifies the interpolation method for
time-varying covariates. When solving ODEs it often samples
times outside the sampling time specified in \code{events}.
When this happens, the time varying covariates are
interpolated.  Currently this can be:
\itemize{
\item \code{"linear"} interpolation, which interpolates the covariate
by solving the line between the observed covariates and extrapolating the new
covariate value.
\item \code{"locf"} -- Last observation carried forward (the default).
\item \code{"nocb"} -- Next Observation Carried Backward.  This is the same method
that NONMEM uses.
\item \code{"midpoint"} Last observation carried forward to midpoint; Next observation
carried backward to midpoint.

For time-varying covariates where a missing value is present, the
interpolation method will use either "locf" or "nocb" throughout
if they are the type of covariate interpolation that is selected.

When using the linear or midpoint interpolation, the lower point
in the interpolation will use locf to interpolate missing
covariates and the upper point will use the nocb to interpolate
missing covariates.
}}

\item{naInterpolation}{specifies the interpolation method for
time-varying covariates when the instantaneous value is \code{NA} (not
during an explicit interpolation) and the \code{covsInterpolation} is
either \code{"midpoint"} or \code{"linear"}. This can be:
\itemize{
\item \code{"locf"} -- last observation carried forward (default)
\item \code{"nocb"} -- next observation carried backward.
}

This will look for the prior value (backwards/locf) when
instantaneously missing, or the next value when instantaneously
missing.  If all the covariates are missing and you find the
end/beginning of the individual record, switch direction.  If all
are really missing, then return missing.}

\item{keepInterpolation}{specifies the interpolation method for
variables in the \code{keep} column.  When \code{nlmixr2} creates \code{mtime},
\code{addl} doses etc, these items were not originally in the dataset.
The interpolation methods you can choose are:
\itemize{
\item \code{"locf"} -- last observation carried forward (default)
\item \code{"nocb"} -- next observation carried backward.
\item \code{"na"} -- no interpolation, simply put \code{NA} for the
interpolated \code{keep} covariates.
}}

\item{addCov}{A boolean indicating if covariates should be added
to the output matrix or data frame. By default this is
disabled.}

\item{sigma}{Named sigma covariance or Cholesky decomposition of a
covariance matrix.  The names of the columns indicate
parameters that are simulated.  These are simulated for every
observation in the solved system. When \code{sigma} is \code{NA} and you are using it with a
\code{rxode2} ui model, the unexplained variability described by
the \code{sigma} matrix are set to zero.}

\item{sigmaDf}{Degrees of freedom of the sigma t-distribution.  By
default it is equivalent to \code{Inf}, or a normal distribution.}

\item{sigmaLower}{Lower bounds for simulated unexplained variability (by default -Inf)}

\item{sigmaUpper}{Upper bounds for simulated unexplained variability (by default Inf)}

\item{nCoresRV}{Number of cores used for the simulation of the
sigma variables.  By default this is 1. To reproduce the results
you need to run on the same platform with the same number of
cores. This is the reason this is set to be one, regardless of
what the number of cores are used in threaded ODE solving.}

\item{sigmaIsChol}{Boolean indicating if the sigma is in the
Cholesky decomposition instead of a symmetric covariance}

\item{sigmaSeparation}{separation strategy for sigma;

Tells the type of separation strategy when
simulating covariance with parameter uncertainty with standard
deviations modeled in the \code{thetaMat} matrix.
\itemize{
\item \code{"lkj"} simulates the correlation matrix from the
\code{rLKJ1} matrix with the distribution parameter \code{eta}
equal to the degrees of freedom \code{nu} by \code{(nu-1)/2}
\item \code{"separation"} simulates from the identity inverse Wishart
covariance matrix with \code{nu} degrees of freedom.  This is then
converted to a covariance matrix and augmented with the modeled
standard deviations.  While computationally more complex than the
\code{"lkj"} prior, it performs better when the covariance matrix
size is greater or equal to 10
\item \code{"auto"} chooses \code{"lkj"} when the dimension of the
matrix is less than 10 and \code{"separation"} when greater
than equal to 10.
}}

\item{sigmaXform}{When taking \code{sigma} values from the \code{thetaMat}
simulations (using the separation strategy for covariance
simulation), how should the \code{thetaMat} values be turned int
standard deviation values:
\itemize{
\item \code{identity} This is when standard deviation values are
directly modeled by the \code{params} and \code{thetaMat} matrix
\item \code{variance} This is when the \code{params} and \code{thetaMat}
simulates the variance that are directly modeled by the
\code{thetaMat} matrix
\item \code{log} This is when the \code{params} and \code{thetaMat}
simulates \code{log(sd)}
\item \code{nlmixrSqrt} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix
with the \verb{x\\^2} modeled along the diagonal.  This only works
with a diagonal matrix.
\item \code{nlmixrLog} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix
with the \verb{exp(x\\^2)} along the diagonal.  This only works
with a diagonal matrix.
\item \code{nlmixrIdentity} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix.
This only works with a diagonal matrix.
}}

\item{nDisplayProgress}{An integer indicating the minimum number
of c-based solves before a progress bar is shown.  By default
this is 10,000.}

\item{amountUnits}{This supplies the dose units of a data frame
supplied instead of an event table.  This is for importing the
data as an rxode2 event table.}

\item{timeUnits}{This supplies the time units of a data frame
supplied instead of an event table.  This is for importing the
data as an rxode2 event table.}

\item{theta}{A vector of parameters that will be named \verb{THETA\[#\]} and
added to parameters}

\item{thetaLower}{Lower bounds for simulated population parameter
variability (by default \code{-Inf})}

\item{thetaUpper}{Upper bounds for simulated population unexplained
variability (by default \code{Inf})}

\item{eta}{A vector of parameters that will be named \verb{ETA\[#\]} and
added to parameters}

\item{addDosing}{Boolean indicating if the solve should add rxode2
EVID and related columns.  This will also include dosing
information and estimates at the doses.  Be default, rxode2
only includes estimates at the observations. (default
\code{FALSE}). When \code{addDosing} is \code{NULL}, only
include \code{EVID=0} on solve and exclude any model-times or
\code{EVID=2}. If \code{addDosing} is \code{NA} the classic
\code{rxode2} EVID events are returned. When \code{addDosing} is \code{TRUE}
add the event information in NONMEM-style format; If
\code{subsetNonmem=FALSE} rxode2 will also include extra event types
(\code{EVID}) for ending infusion and modeled times:
\itemize{
\item \code{EVID=-1} when the modeled rate infusions are turned
off (matches \code{rate=-1})
\item \code{EVID=-2} When the modeled duration infusions are
turned off (matches \code{rate=-2})
\item \code{EVID=-10} When the specified \code{rate} infusions are
turned off (matches \code{rate>0})
\item \code{EVID=-20} When the specified \code{dur} infusions are
turned off (matches \code{dur>0})
\item \verb{EVID=101,102,103,...} Modeled time where 101 is the
first model time, 102 is the second etc.
}}

\item{stateTrim}{When amounts/concentrations in one of the states
are above this value, trim them to be this value. By default
Inf.  Also trims to -stateTrim for large negative
amounts/concentrations.  If you want to trim between a range
say \code{c(0, 2000000)} you may specify 2 values with a lower and
upper range to make sure all state values are in the
reasonable range.}

\item{updateObject}{This is an internally used flag to update the
rxode2 solved object (when supplying an rxode2 solved object) as
well as returning a new object.  You probably should not
modify it's \code{FALSE} default unless you are willing to
have unexpected results.}

\item{omega}{Estimate of Covariance matrix. When omega is a list,
assume it is a block matrix and convert it to a full matrix for
simulations.  When \code{omega} is \code{NA} and you are using it with a
\code{rxode2} ui model, the between subject variability described by
the \code{omega} matrix are set to zero.}

\item{omegaDf}{The degrees of freedom of a t-distribution for
simulation.  By default this is \code{NULL} which is
equivalent to \code{Inf} degrees, or to simulate from a normal
distribution instead of a t-distribution.}

\item{omegaIsChol}{Indicates if the \code{omega} supplied is a
Cholesky decomposed matrix instead of the traditional
symmetric matrix.}

\item{omegaSeparation}{Omega separation strategy

Tells the type of separation strategy when
simulating covariance with parameter uncertainty with standard
deviations modeled in the \code{thetaMat} matrix.
\itemize{
\item \code{"lkj"} simulates the correlation matrix from the
\code{rLKJ1} matrix with the distribution parameter \code{eta}
equal to the degrees of freedom \code{nu} by \code{(nu-1)/2}
\item \code{"separation"} simulates from the identity inverse Wishart
covariance matrix with \code{nu} degrees of freedom.  This is then
converted to a covariance matrix and augmented with the modeled
standard deviations.  While computationally more complex than the
\code{"lkj"} prior, it performs better when the covariance matrix
size is greater or equal to 10
\item \code{"auto"} chooses \code{"lkj"} when the dimension of the
matrix is less than 10 and \code{"separation"} when greater
than equal to 10.
}}

\item{omegaXform}{When taking \code{omega} values from the \code{thetaMat}
simulations (using the separation strategy for covariance
simulation), how should the \code{thetaMat} values be turned int
standard deviation values:
\itemize{
\item \code{identity} This is when standard deviation values are
directly modeled by the \code{params} and \code{thetaMat} matrix
\item \code{variance} This is when the \code{params} and \code{thetaMat}
simulates the variance that are directly modeled by the
\code{thetaMat} matrix
\item \code{log} This is when the \code{params} and \code{thetaMat}
simulates \code{log(sd)}
\item \code{nlmixrSqrt} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix
with the \verb{x\\^2} modeled along the diagonal.  This only works
with a diagonal matrix.
\item \code{nlmixrLog} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix
with the \verb{exp(x\\^2)} along the diagonal.  This only works
with a diagonal matrix.
\item \code{nlmixrIdentity} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix.
This only works with a diagonal matrix.
}}

\item{omegaLower}{Lower bounds for simulated ETAs (by default -Inf)}

\item{omegaUpper}{Upper bounds for simulated ETAs (by default Inf)}

\item{nSub}{Number between subject variabilities (\code{ETAs}) simulated for every
realization of the parameters.}

\item{thetaMat}{Named theta matrix.}

\item{thetaDf}{The degrees of freedom of a t-distribution for
simulation.  By default this is \code{NULL} which is
equivalent to \code{Inf} degrees, or to simulate from a normal
distribution instead of a \code{t}-distribution.}

\item{thetaIsChol}{Indicates if the \code{theta} supplied is a
Cholesky decomposed matrix instead of the traditional
symmetric matrix.}

\item{nStud}{Number virtual studies to characterize uncertainty in estimated
parameters.}

\item{dfSub}{Degrees of freedom to sample the between subject variability matrix from the
inverse Wishart distribution (scaled) or scaled inverse chi squared distribution.}

\item{dfObs}{Degrees of freedom to sample the unexplained variability matrix from the
inverse Wishart distribution (scaled) or scaled inverse chi squared distribution.}

\item{returnType}{This tells what type of object is returned.  The
currently supported types are:
\itemize{
\item \code{"rxSolve"} (default) will return a reactive data frame
that can change easily change different pieces of the solve and
update the data frame.  This is the currently standard solving
method in rxode2,  is used for \code{rxSolve(object, ...)}, \code{solve(object,...)},
\item \code{"data.frame"} -- returns a plain, non-reactive data
frame; Currently very slightly faster than \code{returnType="matrix"}
\item \code{"matrix"} -- returns a plain matrix with column names attached
to the solved object.  This is what is used \code{object$run} as well as \code{object$solve}
\item \code{"data.table"} -- returns a \code{data.table}; The \code{data.table} is
created by reference (ie \code{setDt()}), which should be fast.
\item \code{"tbl"} or \code{"tibble"} returns a tibble format.
}}

\item{seed}{an object specifying if and how the random number
generator should be initialized}

\item{nsim}{represents the number of simulations.  For rxode2, if
you supply single subject event tables (created with
\verb{[eventTable()]})}

\item{minSS}{Minimum number of iterations for a steady-state dose}

\item{maxSS}{Maximum number of iterations for a steady-state dose}

\item{infSSstep}{Step size for determining if a constant infusion
has reached steady state.  By default this is large value,
12.}

\item{strictSS}{Boolean indicating if a strict steady-state is
required. If a strict steady-state is (\code{TRUE}) required
then at least \code{minSS} doses are administered and the
total number of steady states doses will continue until
\code{maxSS} is reached, or \code{atol} and \code{rtol} for
every compartment have been reached.  However, if ODE solving
problems occur after the \code{minSS} has been reached the
whole subject is considered an invalid solve. If
\code{strictSS} is \code{FALSE} then as long as \code{minSS}
has been reached the last good solve before ODE solving
problems occur is considered the steady state, even though
either \code{atol}, \code{rtol} or \code{maxSS} have not
been achieved.}

\item{istateReset}{When \code{TRUE}, reset the \code{ISTATE} variable to 1 for
lsoda and liblsoda with doses, like \code{deSolve}; When \code{FALSE}, do
not reset the \code{ISTATE} variable with doses.}

\item{subsetNonmem}{subset to NONMEM compatible EVIDs only.  By
default \code{TRUE}.}

\item{maxAtolRtolFactor}{The maximum \code{atol}/\code{rtol} that
FOCEi and other routines may adjust to.  By default 0.1}

\item{from}{When there is no observations in the event table,
start observations at this value. By default this is zero.}

\item{to}{When there is no observations in the event table, end
observations at this value. By default this is 24 + maximum
dose time.}

\item{by}{When there are no observations in the event table, this
is the amount to increment for the observations between \code{from}
and \code{to}.}

\item{length.out}{The number of observations to create if there
isn't any observations in the event table. By default this is 200.}

\item{iCov}{A data frame of individual non-time varying covariates
to combine with the \code{events} dataset.  The \code{iCov} dataset has one
covariate per ID and should match the event table}

\item{keep}{Columns to keep from either the input dataset or the
\code{iCov} dataset.  With the \code{iCov} dataset, the column
is kept once per line.  For the input dataset, if any records
are added to the data LOCF (Last Observation Carried forward)
imputation is performed.}

\item{indLinPhiTol}{the requested accuracy tolerance on
exponential matrix.}

\item{indLinPhiM}{the maximum size for the Krylov basis}

\item{indLinMatExpType}{This is them matrix exponential type that
is use for rxode2.  Currently the following are supported:
\itemize{
\item \code{Al-Mohy} Uses the exponential matrix method of Al-Mohy Higham (2009)
\item \code{arma} Use the exponential matrix from RcppArmadillo
\item \code{expokit} Use the exponential matrix from Roger B. Sidje (1998)
}}

\item{indLinMatExpOrder}{an integer, the order of approximation to
be used, for the \code{Al-Mohy} and \code{expokit} values.
The best value for this depends on machine precision (and
slightly on the matrix). We use \code{6} as a default.}

\item{drop}{Columns to drop from the output}

\item{idFactor}{This boolean indicates if original ID values
should be maintained. This changes the default sequentially
ordered ID to a factor with the original ID values in the
original dataset.  By default this is enabled.}

\item{mxhnil}{maximum number of messages printed (per problem)
warning that \code{T + H = T} on a step (\code{H} = step size).  This must
be positive to result in a non-default value.  The default
value is 0 (or infinite).}

\item{hmxi}{inverse of the maximum absolute value of \code{H} to are used.
hmxi = 0.0 is allowed and corresponds to an infinite \verb{hmax1 (default).  }hmin\code{and}hmxi\verb{may be changed at any time, but will not take effect until the next change of}H\verb{is considered. This option is only considered with}method="liblsoda"`.}

\item{warnIdSort}{Warn if the ID is not present and rxode2 assumes
the order of the parameters/iCov are the same as the order of
the parameters in the input dataset.}

\item{warnDrop}{Warn if column(s) were supposed to be dropped, but
were not present.}

\item{ssAtol}{Steady state atol convergence factor.  Can be
a vector based on each state.}

\item{ssRtol}{Steady state rtol convergence factor.  Can be a
vector based on each state.}

\item{safeZero}{Use safe zero divide. By default
this is turned on but you may turn it off if you wish.}

\item{safeLog}{Use safe log.  When enabled if your value that you are taking log() of is negative or zero, this will return \verb{log(machine epsilon)}.  By default this is turned on.}

\item{safePow}{Use safe powers.  When enabled if your power is
negative and your base is zero, this will return the \verb{machine epsilon^(negative power)}.  By default this is turned on.}

\item{sumType}{Sum type to use for \code{sum()} in
rxode2 code blocks.

\code{pairwise} uses the pairwise sum (fast, default)

\code{fsum} uses the PreciseSum package's fsum function (most accurate)

\code{kahan} uses Kahan correction

\code{neumaier} uses Neumaier correction

\code{c} uses no correction: default/native summing}

\item{prodType}{Product to use for \code{prod()} in rxode2 blocks

\verb{long double} converts to long double, performs the
multiplication and then converts back.

\code{double} uses the standard double scale for multiplication.}

\item{resample}{A character vector of model variables to resample
from the input dataset; This sampling is done with replacement.
When \code{NULL} or \code{FALSE} no resampling is done.  When
\code{TRUE} resampling is done on all covariates in the input
dataset}

\item{resampleID}{boolean representing if the resampling should be
done on an individual basis \code{TRUE} (ie. a whole patient is
selected) or each covariate is resampled independent of the
subject identifier \code{FALSE}.  When \code{resampleID=TRUE}
correlations of parameters are retained, where as when
\code{resampleID=FALSE} ignores patient covariate correaltions.
Hence the default is \code{resampleID=TRUE}.}

\item{maxwhile}{represents the maximum times a while loop is
evaluated before exiting.  By default this is 100000}

\item{atolSens}{Sensitivity atol, can be different than atol with
liblsoda.  This allows a less accurate solve for gradients (if desired)}

\item{rtolSens}{Sensitivity rtol, can be different than rtol with
liblsoda.  This allows a less accurate solve for gradients (if desired)}

\item{ssAtolSens}{Sensitivity absolute tolerance (atol) for
calculating if steady state has been achieved for sensitivity compartments.}

\item{ssRtolSens}{Sensitivity relative tolerance (rtol) for
calculating if steady state has been achieved for sensitivity compartments.}

\item{simVariability}{determines if the variability is simulated.
When \code{NA} (default) this is determined by the solver.}

\item{nLlikAlloc}{The number of log likelihood endpoints that are
used in the model.  This allows independent log likelihood per
endpoint in focei for nlmixr2.  It likely shouldn't be set,
though it won't hurt anything if you do (just may take up more
memory for larger allocations).}

\item{useStdPow}{This uses C's \code{pow} for exponentiation instead of
R's \code{R_pow} or \code{R_pow_di}.  By default this is \code{FALSE}}

\item{naTimeHandle}{Determines what time of handling happens when
the time becomes \code{NA}: current options are:
\itemize{
\item \code{ignore} this ignores the \code{NA} time input and passes it through.
\item \code{warn} (default) this will produce a warning at the end of the
solve, but continues solving passing through the \code{NA} time
\item \code{error} this will stop this solve if this is not a parallel
solved ODE (otherwise stopping can crash R)
}}

\item{addlKeepsCov}{This determines if the additional dosing items
repeats the dose only (\code{FALSE}) or keeps the covariates at the
record of the dose (\code{TRUE})}

\item{addlDropSs}{When there are steady state doses with an \code{addl}
specification the steady state flag is dropped with repeated
doses (when \code{TRUE}) or retained (when \code{FALSE})}

\item{ssAtDoseTime}{Boolean that when \code{TRUE} back calculates the
steady concentration at the actual time of dose, otherwise when
\code{FALSE} the doses are shifted}

\item{ss2cancelAllPending}{When \code{TRUE} the \code{SS=2} event type
cancels all pending doses like \code{SS=1}.  When \code{FALSE} the pending
doses not canceled with \code{SS=2} (the infusions started before
\code{SS=2} occurred are canceled, though).}

\item{ssSolved}{When \code{TRUE} this will return the solved steady
state solutions for the linear compartment model.  When \code{FALSE}
this will solve to steady state using the linear solutions
instead.  This is only used when the method only has \code{linCmt()}
and does not mix ODEs with the solution.  The default is \code{TRUE}.}

\item{linCmtSensType}{The type of linear compartment
sensitivity/gradients to use.  The current options are:
\itemize{
\item \code{auto} -- for one compartment models this will use the \code{AD}
method, for 2 and 3 compartment model this will use \code{forwardG}.
\item \code{AD} -- automatic differentiation (using stan math).  This now
uses forward-mode AD (\verb{stan::math::fvar}), which differentiates the
same analytic solution as the reverse-mode path but on the C++
stack with no reverse autodiff tape, matching the reverse result to
round-off while avoiding the per-call tape setup/teardown.
\item \code{ADr} -- reverse-mode automatic differentiation (the prior \code{AD}
behavior, \verb{stan::math::jacobian}).  Kept as an escape hatch for
validation; \code{AD} (forward) is preferred.
\item \code{forward} -- forward sensitivity where the step size is
determined by shi 2021 optimization (only once per problem)
\item \code{forwardG} -- forward sensitivity where the step size is determined by the
Gill 1983 optimization for forward differences (only once per problem).
\item \code{central} -- central sensitivity where the step size is
determined by shi 2021 optimization (only once per problem)
\item \code{forward3} -- three point central difference where step size is
determined by shi 2021 optimization for central differences (only
once per problem)
\item \code{endpoint5} -- five point endpoint difference where step size is
determined by the shi 2021 optimization for central differences
(only once per problem)
\item \code{fowardH} -- forward sensitivity where the step size is fixed
\item \code{centralH} -- central sensitivity where the step size is fixed
\item \code{forward3H} -- three point central difference where step size is fixed
\item \code{endpoint5H} -- five point endpoint difference where step size is fixed
}}

\item{linCmtSensH}{The step size for the forward and central
differences when using the option \code{centralH}, \code{forwardH},
\code{foward3H} or \code{endpoint5H} options.
#'}

\item{linCmtGillFtol}{The gillFtol is the gradient error tolerance that
is acceptable before issuing a warning/error about the gradient estimates.}

\item{linCmtGillK}{The total number of possible steps to determine the
optimal forward/central difference step size per parameter (by
the Gill 1983 method).  If 0, no optimal step size is
determined.  Otherwise this is the optimal step size
determined.}

\item{linCmtGillStep}{When looking for the optimal forward difference
step size, this is This is the step size to increase the
initial estimate by.  So each iteration the new step size =
(prior step size)*gillStep}

\item{linCmtGillRtol}{The relative tolerance used for Gill 1983
optimal step size determination.}

\item{linCmtShiErr}{Shi difference error}

\item{linCmtShiMax}{The maximum number of steps for the
optimization of the forward-difference step size in linear
compartment numeric difference.}

\item{linCmtScale}{The scale of the linear compartment model.  This
is applied to sensitivity approximation using numeric
differences.  When \code{TRUE} or \code{NULL} use default scaling, when
\code{FALSE} use no scaling.  If it is one elment numeric, the value
is duplicated 7 times and applies to all the parameters.
Otherwise this is a seven element numeric vector implying the
scaling for each of the linear compartmental model parameters.}

\item{linCmtHcmt}{This represents the compartments considered when
optimizing the forward difference step size.  When a character
vector it can be any of the following (multiple allowed):
\itemize{
\item \code{"depot"} -- depot compartment
\item \code{"central"} -- central compartment
\item \code{"peripheral1"} -- peripheral compartment
\item \code{"peripheral2"} -- second peripheral compartment
\item \code{"concentration"} -- concentration value (i.e. central
compartment/volume)
}}

\item{linCmtHmeanI}{This represents the type of sum done for each
time-point of the linear solved systems (as defined by \code{"linCmtHcmt"}).
\itemize{
\item \code{"arithmetic"} -- gives the arithmetic mean
\item \code{"geometric"} -- gives the geometric mean
\item \code{"harmonic"} -- gives the harmonic mean
}}

\item{linCmtHmeanO}{This represents the type of sum done for the
overall problem of the linear solved systems (first each time
point mean is calculated with \code{linCmtHmeanI}).
\itemize{
\item \code{"arithmetic"} -- gives the arithmetic mean
\item \code{"geometric"} -- gives the geometric mean
\item \code{"harmonic"} -- gives the harmonic mean
}}

\item{linCmtSuspect}{The tolerance for gradients in linear
compartment solutions to re-compute when gradients seem to be
zero.}

\item{linCmtForwardMax}{The maximum number of points in a forward
difference to take while calculating the gradients.  This is an
integer from 1 to 3.  There is at least 1 extra point taken for
gradient calculation, if the gradient is suspect another is taken
(if this value is 2), and finally a third is calculated if the
gradient is still suspect.}

\item{indOwnAlloc}{Logical; when \code{TRUE} each individual's \code{dose},
\code{ii}, \code{all_times}, and \code{solve} arrays are allocated independently
via \code{malloc}/\code{calloc} rather than as pointers into a single
global buffer.  This enables per-individual reallocation for dose
and observation-pushing with \code{evid_()} and related
functions. When \code{FALSE} these arrays are allocated as a single
global buffer, which is slightly faster and slightly more memory
efficient but does not allow for dynamic dosing/observations.  By
default this is \code{NA} which automatically decides based on if
there is any dosing or observation pushing in the model.}

\item{maxExtra}{Integer; maximum number of events (doses and
observations) that \code{evid_()} may push per individual per solve.
When an individual exceeds this limit the solve is aborted for
that individual (output filled with \code{NA}) and an error is raised
after the full parallel solve completes.  Set to \code{0L} to allow
unlimited pushes (use with care -- cascading \code{evid_()} calls can
grow without bound).  Default is \code{100L}.}

\item{tolFactor}{A per-individual tolerance multiplier (>= 1.0, or
\code{NULL} for no effect).  When supplied, each individual's
\code{atol}, \code{rtol}, \code{ssAtol}, and \code{ssRtol} are multiplied by
this factor before the ODE solver is called, loosening the
tolerances for that individual.  This is useful when a small
number of subjects are numerically stiff and would otherwise
cause solve failures: pass a larger factor for the
problematic subjects and leave the rest at 1.0 (or omit them).
The effective tolerance is always capped at \code{maxAtolRtolFactor}.

\if{html}{\out{<div class="sourceCode">}}\preformatted{`tolFactor` may be:
* `NULL` (default) -- no adjustment applied.
* A single numeric value -- the same factor is applied to
  the first `length(tolFactor)` subjects in order.
* A numeric vector -- applied element-wise to subjects in
  the order they appear.
* A **named** numeric vector -- names are matched to subject
  IDs; unmatched subjects retain `tolFactor = 1.0`.

The per-subject factors used (after matching) are stored back
in the solved object and accessible via `$tolFactor`.
}\if{html}{\out{</div>}}}

\item{serializeFile}{Controls serialization-driven solves. When
this is a file path, \code{rxSolve()} writes the pre-integration
serialized state to that file and returns the file path invisibly
without running the solve. When this is \code{TRUE}, \code{rxSolve()}
writes the serialized state to a temporary \code{.rxbin} file, solves
by reloading from that file, returns the solve result, and then
removes the temporary file (mostly for testing). When solving
from an existing serialization file via \code{rxSolve(model, "state.rxbin")}, only the model and serialization file are
allowed because the file already stores the solve inputs and
controls.}

\item{dense}{Logical; when \code{TRUE} and the method supports dense output,
enables continuous interpolation so the solver can take large internal
steps and reconstruct the solution cheaply at each observation time.
Dense-capable single methods are \code{"dop853"}, \code{"dop5"}, \code{"bs"}, and
\code{"ros4"}.  For composite AutoSwitch methods (e.g. \code{"dop5+ros4"}), dense
output is enabled only when \strong{both} the primary and stiff secondary
support dense output; \code{"ros4"} is the only stiff method that does, so
valid dense composites are \code{"dop853+ros4"}, \code{"dop5+ros4"}, and
\code{"bs+ros4"}.  A warning is issued and \code{dense} is set to \code{FALSE} when a
composite stiff secondary does not support dense output.  Silently
ignored for non-dense single methods.  Not yet supported for \code{linCmt()}
models (a warning is emitted and the standard path is used instead).}

\item{cvodeLinSolver}{Character; selects the linear solver used by the CVODE
integrator when \code{method = "cvode"}.  Ignored for all other methods.
Available choices and when to use them:
\itemize{
\item \code{"dense"} (default) -- Direct LU factorization of the full Jacobian.
Best for small to medium systems (typically fewer than ~50 compartments).
Requires \code{O(n^2)} storage and \code{O(n^3)} work per Newton iteration.
\item \code{"band"} -- Currently aliases to \code{"dense"}.  Kept for compatibility
until rxode2 can determine model bandwidth safely during solve setup.
\item \code{"gmres"} -- Generalized Minimal Residual iterative Krylov solver.
Avoids explicit Jacobian formation, making it practical for large stiff
systems (tens to hundreds of compartments) where LU factorization would
dominate runtime.  Generally the first iterative solver to try.
\item \code{"bicgstab"} -- Bi-Conjugate Gradient Stabilized iterative Krylov
solver.  Similar use case to \code{"gmres"} but can converge faster on some
non-symmetric problems.  Try if \code{"gmres"} is slow or fails to converge.
\item \code{"tfqmr"} -- Transpose-Free Quasi-Minimal Residual iterative Krylov
solver.  Another alternative for large systems; avoids the matrix
transpose required by classical QMR.
}

For most pharmacometric models with fewer than ~50 compartments the default
\code{"dense"} solver is fastest.  Switch to an iterative solver (\code{"gmres"} is
a good first choice) only for large QSP or PBPK models where Jacobian
factorization becomes the bottleneck.}

\item{autoSwitchNonstifftol}{Numeric in \verb{(0, 1]}; stiffness ratio threshold
used when the solver is in non-stiff mode.  If
\verb{rho * |dt| / S(primary) > autoSwitchNonstifftol}, the interval is
considered stiff and the secondary solver is tried.  Default \code{9/10}.}

\item{autoSwitchStifftol}{Numeric in \verb{(0, 1]}; non-stiffness ratio threshold
used when the solver is in stiff mode.  If
\verb{rho * |dt| / S(primary) < autoSwitchStifftol}, the interval is considered
non-stiff and the switch-back counter is incremented.  Default \code{9/10}.}

\item{autoSwitchDtfac}{Numeric \verb{>= 1}; factor by which the suggested step
size is multiplied when switching to the stiff solver, and divided when
switching back.  Default \code{2.0}.}

\item{autoSwitchMaxStiff}{Integer; number of consecutive stiff-detected
intervals before permanently switching to the stiff solver.  Default \code{10L}.}

\item{autoSwitchMaxNonstiff}{Integer; number of consecutive non-stiff
intervals (while in stiff mode) before switching back to the fast
non-stiff solver.  Default \code{3L}.}

\item{autoSwitchStiffFirst}{Logical; when \code{TRUE}, start each subject solve
with the stiff solver instead of the non-stiff primary.  Default \code{FALSE}.}

\item{autoSwitchSwitchMax}{Non-negative integer; minimum number of
integration intervals that must elapse after a switch before the solver
is allowed to switch back in the opposite direction.  Acts as an
oscillation guard: if the stiffness ratio fluctuates near a threshold,
this prevents rapid back-and-forth between solvers.  Set to \code{0L} to
disable the guard entirely.  Default \code{5L}.}

\item{stiff2}{Integer method code for the stiff secondary solver used in
AutoSwitch composite methods.  Normally set automatically when \code{method} is
a composite string of the form \code{"primary+stiff"} (e.g.
\code{"dop853+ros4"}); the \code{"+"} notation is the preferred way to configure
composite solving and \code{stiff2} need not be supplied directly.  When
supplied as a raw integer it must be a stiff method code as returned by
\code{\link[=odeMethodToInt]{odeMethodToInt()}}; \code{0L} (the default) disables the stiff secondary and
causes the solver to run as a plain non-composite method.  Dense output
(\code{dense = TRUE}) is silently disabled when \code{stiff2} names a stiff method
that does not support dense output; \code{"ros4"} (code \code{13L}) is the only
stiff secondary that does support it.}

\item{useLinCmt}{Logical; when \code{TRUE} and the model contains
linear-compartment ODEs that can be solved analytically,
automatically convert them to a \code{linCmt()} call before solving.
The detection and conversion use \code{\link[=odeToLin]{odeToLin()}}; the converted
model is cached so the compilation cost is paid only once.  Set
to \code{FALSE} to keep the original ODE solver.  This flag is also
stored in the returned \code{\link[=rxControl]{rxControl()}} object so that downstream
hooks (e.g. in nlmixr2) can read and apply it.  The default is to
use the value of \code{rxode2.useLinCmt} option (which when specified
is \code{TRUE} by default).}

\item{file}{Character string giving a file path prefix for out-of-memory
chunk solving. When set, \code{rxSolve()} splits subjects into chunks, writes
each chunk's result to \verb{<file>_chunk_NNNNN.parquet} (or \code{.rds} if
\code{arrow} is unavailable), and returns an \code{rxSolveOom} object. The manifest
is saved to \verb{<file>_manifest.rds}. Set to \code{NULL} (default) to use the
normal in-memory path.}

\item{chunkSize}{Integer; number of subjects per chunk when \code{file} is
set. If \code{NULL} (default), the chunk size is auto-computed from available
free RAM using \code{rxMemoryEstimate()}.}

\item{parallel}{Integer; number of \code{mirai} daemons to use for parallel
chunk solving when \code{file} is set. \code{0L} (default) uses serial solving.
Requires the \code{mirai} package.}

\item{envir}{is the environment to look for R user functions
(defaults to parent environment)}

\item{a}{when using \code{solve()}, this is equivalent to the
\code{object} argument.  If you specify \code{object} later in
the argument list it overwrites this parameter.}

\item{b}{when using \code{solve()}, this is equivalent to the
\code{params} argument.  If you specify \code{params} as a
named argument, this overwrites the output}
}
\value{
An \dQuote{rxSolve} solve object that stores the solved
value in a special data.frame or other type as determined by
\code{returnType}. By default this has as many rows as there are
sampled time points and as many columns as system variables (as
defined by the ODEs and additional assignments in the rxode2 model
code).  It also stores information about the call to allow
dynamic updating of the solved object.

The operations for the object are similar to a data-frame, but
expand the \code{$} and \verb{[[""]]} access operators and assignment
operators to resolve based on different parameter values, initial
conditions, solver parameters, or events (by updating the \code{time}
variable).

You can call the \code{\link[=eventTable]{eventTable()}} methods on the solved object to
update the event table and resolve the system of equations.
}
\description{
This uses rxode2 family of objects, file, or model specification to
solve a ODE system.  There are many options for a solved rxode2
model, the first are the required \code{object}, and \code{events} with the
some-times optional \code{params} and \code{inits}.
}
\details{
The rest of the document focus on the different ODE solving
methods, followed by the core solving method's options, rxode2 event
handling options, rxode2's numerical stability options, rxode2's
output options, and finally internal rxode2 options or compatibility
options.
}
\references{
"New Scaling and Squaring Algorithm for the Matrix Exponential", by
Awad H. Al-Mohy and Nicholas J. Higham, August 2009

Roger B. Sidje (1998).  EXPOKIT: Software package for computing
matrix exponentials.  ACM - Transactions on Mathematical Software
\emph{24}(1), 130-156.

Hindmarsh, A. C.
\emph{ODEPACK, A Systematized Collection of ODE Solvers}.
Scientific Computing, R. S. Stepleman et al. (Eds.),
North-Holland, Amsterdam, 1983, pp. 55-64.

Petzold, L. R.
\emph{Automatic Selection of Methods for Solving Stiff and Nonstiff
Systems of Ordinary Differential Equations}.
Siam J. Sci. Stat. Comput. 4 (1983), pp. 136-148.

Hairer, E., Norsett, S. P., and Wanner, G.
\emph{Solving ordinary differential equations I, nonstiff problems}.
2nd edition, Springer Series in Computational Mathematics,
Springer-Verlag (1993).
}
\seealso{
\code{\link[=rxode2]{rxode2()}}
}
\author{
Matthew Fidler, Melissa Hallow and  Wenping Wang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxRaw.R
\name{rxSerialize}
\alias{rxSerialize}
\title{Serialize an R Object to a Raw Vector}
\usage{
rxSerialize(x, type = c("xz", "bzip2", "base"))
}
\arguments{
\item{x}{object to serialize}

\item{type}{serialization type; one of "base", "xz" or "bzip2".}
}
\value{
raw vector
}
\description{
Serialize an R Object to a Raw Vector
}
\examples{

rxSerialize(mtcars)

rxRawToC(mtcars)

}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piping.R
\name{.copyUi}
\alias{.copyUi}
\title{This copies the rxode2 UI object so it can be modified}
\usage{
.copyUi(ui)
}
\arguments{
\item{ui}{Original UI object}
}
\value{
Copied UI object
}
\description{
This copies the rxode2 UI object so it can be modified
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{gammaq}
\alias{gammaq}
\title{Gammaq: normalized upper incomplete gamma function}
\usage{
gammaq(a, z)
}
\arguments{
\item{a}{The numeric 'a' parameter in the normalized upper
incomplete gamma}

\item{z}{The numeric 'z' parameter in the normalized upper
incomplete gamma}
}
\value{
gammaq results
}
\description{
This is the gamma_q from the boost library
}
\details{
The gamma q function is given by:

gammaq = uppergamma(a, z)/gamma(a)
}
\examples{

gammaq(1, 3)
gammaq(1:3, 3)
gammaq(1, 1:3)
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{setCvodeLinearSolver}
\alias{setCvodeLinearSolver}
\title{Set the CVODE linear solver}
\usage{
setCvodeLinearSolver(type)
}
\arguments{
\item{type}{Integer: 1=dense (default), 2=band (currently aliases to dense),
3=gmres, 4=bicgstab, 5=tfqmr}
}
\value{
NULL (invisibly)
}
\description{
Set the CVODE linear solver
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assert.R
\name{assertCompartmentNew}
\alias{assertCompartmentNew}
\title{Verify that a compartment would be new to the model}
\usage{
assertCompartmentNew(ui, x)
}
\arguments{
\item{ui}{is the model to test that a model paramet exists}

\item{x}{The value to test}
}
\value{
The value or an error
}
\description{
Verify that a compartment would be new to the model
}
\seealso{
Other Assertions:
\code{\link[=assertCompartmentExists]{assertCompartmentExists()}},
\code{\link[=assertCompartmentName]{assertCompartmentName()}},
\code{\link[=assertRxUi]{assertRxUi()}},
\code{\link[=assertVariableExists]{assertVariableExists()}},
\code{\link[=assertVariableNew]{assertVariableNew()}},
\code{\link[=testIniDf]{testIniDf()}},
\code{\link[=testRxUnbounded]{testRxUnbounded()}}
}
\author{
Matthew Fidler & Bill Denney
}
\concept{Assertions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{gammapDer}
\alias{gammapDer}
\title{gammapDer:  derivative of gammap}
\usage{
gammapDer(a, z)
}
\arguments{
\item{a}{The numeric 'a' parameter in the upper
incomplete gamma}

\item{z}{The numeric 'z' parameter in the upper
incomplete gamma}
}
\value{
lowergamma results
}
\description{
This is the gamma_p_derivative from the boost library
}
\examples{

gammapDer(1:3, 3)

gammapDer(1, 1:3)
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/rxode2.R
\name{rxDynUnload}
\alias{rxDynUnload}
\alias{rxUnload}
\title{Unload rxode2 object}
\usage{
rxDynUnload(obj)

rxUnload(obj)
}
\arguments{
\item{obj}{A rxode2 family of objects}
}
\value{
Boolean returning if the rxode2 library is loaded.
}
\description{
Unload rxode2 object
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{probit}
\alias{probit}
\alias{probitInv}
\title{probit and inverse probit functions}
\usage{
probit(x, low = 0, high = 1)

probitInv(x, low = 0, high = 1)
}
\arguments{
\item{x}{Input value(s) in range [low,high] to translate -Inf to
Inf}

\item{low}{Lowest value in the range}

\item{high}{Highest value in the range}
}
\value{
values from probit, probitInv and probitNormInfo
}
\description{
probit and inverse probit functions
}
\examples{

probit(0.25)

probitInv(-0.674)

probitNormInfo(probit(0.25), sd = 0.1)

probitNormInfo(probit(1, 0, 10), sd = 1, low = 0, high = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikNbinom}
\alias{llikNbinom}
\title{Calculate the log likelihood of the negative binomial function (and its derivatives)}
\usage{
llikNbinom(x, size, prob, full = FALSE)
}
\arguments{
\item{x}{Number of successes}

\item{size}{Size of trial}

\item{prob}{probability of success}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the pdf value of with
\code{dProb} that has the derivatives with respect to the parameters at
the observation time-point
}
\description{
Calculate the log likelihood of the negative binomial function (and its derivatives)
}
\details{
In an \code{rxode2()} model, you can use \code{llikNbinom()} but you have to
use all arguments.  You can also get the derivative of \code{prob} with
\code{llikNbinomDprob()}
}
\examples{
\donttest{
llikNbinom(46:54, 100, 0.5)

llikNbinom(46:54, 100, 0.5, TRUE)

# In rxode2 you can use:

et <- et(46:54)
et$size <- 100
et$prob <-0.5

model <- function() {
  model({
    fx <- llikNbinom(time, size, prob)
    dProb <- llikNbinomDprob(time, size, prob)
  })
}

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/etNew.R
\name{is.rxEt}
\alias{is.rxEt}
\title{Check if object is an rxEt event table}
\usage{
is.rxEt(x)
}
\arguments{
\item{x}{object to test}
}
\value{
logical
}
\description{
Check if object is an rxEt event table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{rxWithSeed}
\alias{rxWithSeed}
\alias{rxWithPreserveSeed}
\title{Preserved seed and possibly set the seed}
\usage{
rxWithSeed(
  seed,
  code,
  rxseed = rxGetSeed(),
  kind = "default",
  normal.kind = "default",
  sample.kind = "default"
)

rxWithPreserveSeed(code)
}
\arguments{
\item{seed}{R seed to use for the session}

\item{code}{Is the code to evaluate}

\item{rxseed}{is the rxode2 seed that is being preserved}

\item{kind}{character or \code{NULL}.  If \code{kind} is a character
    string, set \R's RNG to the kind desired.  Use \code{"default"} to
    return to the \R default.  See \sQuote{Details} for the
    interpretation of \code{NULL}.}

\item{normal.kind}{character string or \code{NULL}.  If it is a character
    string, set the method of Normal generation.  Use \code{"default"}
    to return to the \R default.  \code{NULL} makes no change.}

\item{sample.kind}{character string or \code{NULL}.  If it is a character
    string, set the method of discrete uniform generation (used in 
    \code{\link[base]{sample}}, for instance).  Use \code{"default"} to return to 
    the \R default.  \code{NULL} makes no change.}
}
\value{
returns whatever the code is returning
}
\description{
Preserved seed and possibly set the seed
}
\examples{

rxGetSeed()
rxWithSeed(1, {
   print(rxGetSeed())
   rxnorm()
   print(rxGetSeed())
   rxnorm()
}, rxseed=3)

}
\seealso{
rxGetSeed, rxSetSeed
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxValidate.R
\name{rxValidate}
\alias{rxValidate}
\alias{rxTest}
\title{Validate rxode2
This allows easy validation/qualification of nlmixr by running the
testing suite on your system.}
\usage{
rxValidate(type = NULL, skipOnCran = TRUE)

rxTest(type = NULL, skipOnCran = TRUE)
}
\arguments{
\item{type}{Type of test or filter of test type, When this is an
expression, evaluate the contents, respecting \code{skipOnCran}.  In
the expression form, stray progress messages are muffled unless
\code{options(rxode2.test.verbose = TRUE)} is set.}

\item{skipOnCran}{when \code{TRUE} skip the test on CRAN.}
}
\value{
nothing
}
\description{
Validate rxode2
This allows easy validation/qualification of nlmixr by running the
testing suite on your system.
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxSeedEng}
\alias{rxSeedEng}
\title{This seeds the engine based on the number of cores used in random number generation}
\usage{
rxSeedEng(ncores = 1L)
}
\arguments{
\item{ncores}{is the number of cores to use.}
}
\value{
Nothing, called for side effects
}
\description{
This seeds the engine based on the number of cores used in random number generation
}
\examples{
rxSeedEng()
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/rxode2.R
\name{rxDynLoad}
\alias{rxDynLoad}
\alias{rxLoad}
\title{Load rxode2 object}
\usage{
rxDynLoad(obj)

rxLoad(obj)
}
\arguments{
\item{obj}{A rxode2 family of objects}
}
\value{
Boolean returning if the rxode2 library is loaded.
}
\description{
Load rxode2 object
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{rxEtDispatchSolve}
\alias{rxEtDispatchSolve}
\alias{rxEtDispatchSolve.default}
\title{Dispatch solve to 'rxode2' solve}
\usage{
rxEtDispatchSolve(x, ...)

\method{rxEtDispatchSolve}{default}(x, ...)
}
\arguments{
\item{x}{rxode2 solve dispatch object}

\item{...}{other arguments}
}
\value{
if 'rxode2'  is loaded, a solved object, otherwise an error
}
\description{
Dispatch solve to 'rxode2' solve
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assert.R
\name{assertCompartmentName}
\alias{assertCompartmentName}
\alias{assertVariableName}
\alias{assertParameterValue}
\alias{assertExists}
\alias{testExists}
\title{Verify that a value is a valid nlmixr2 compartment name}
\usage{
assertCompartmentName(x)

assertVariableName(x)

assertParameterValue(x)

assertExists(ui, x)

testExists(ui, x)
}
\arguments{
\item{x}{The value to test}

\item{ui}{when needed, this is the rxode2/nlmixr2 model}
}
\value{
The value or an error
}
\description{
Verify that a value is a valid nlmixr2 compartment name
}
\section{Functions}{
\itemize{
\item \code{assertVariableName()}: Verify that a value is a valid
nlmixr2 variable name

\item \code{assertParameterValue()}: Verify that a value is a valid
nlmixr2 parameter value

\item \code{assertExists()}: Assert compartment/variable exists

\item \code{testExists()}: Test compartment/variable exists

}}
\seealso{
Other Assertions:
\code{\link[=assertCompartmentExists]{assertCompartmentExists()}},
\code{\link[=assertCompartmentNew]{assertCompartmentNew()}},
\code{\link[=assertRxUi]{assertRxUi()}},
\code{\link[=assertVariableExists]{assertVariableExists()}},
\code{\link[=assertVariableNew]{assertVariableNew()}},
\code{\link[=testIniDf]{testIniDf()}},
\code{\link[=testRxUnbounded]{testRxUnbounded()}}
}
\author{
Bill Denney
}
\concept{Assertions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.rxWithWd}
\alias{.rxWithWd}
\title{Temporarily set options then restore them while running code}
\usage{
.rxWithWd(wd, code)
}
\arguments{
\item{wd}{working directory to temporarily set the system to while
evaluating the code}

\item{code}{The code to run during the sink}
}
\value{
value of code
}
\description{
Temporarily set options then restore them while running code
}
\examples{

.rxWithWd(tempdir(), {
  getwd()
})

getwd()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dfWishart.R
\name{dfWishart}
\alias{dfWishart}
\title{This uses simulations to match the rse}
\usage{
dfWishart(omega, n, rse, upper, totN = 1000, diag = TRUE, seed = 1234)
}
\arguments{
\item{omega}{represents the matrix for simulation}

\item{n}{This represents the number of subjects/samples this comes
from (used to calculate rse).  When present it assumes the rse=
sqrt(2)/sqrt(n)}

\item{rse}{This is the rse that we try to match, if not specified,
it is derived from \code{n}}

\item{upper}{The upper boundary for root finding in terms of
degrees of freedom.  If not specified, it is n*200}

\item{totN}{This represents the total number of simulated inverse
wishart deviates}

\item{diag}{When \code{TRUE}, represents the rse to match is the
diagonals, otherwise it is the total matrix.}

\item{seed}{to make the simulation reproducible, this represents
the seed that is used for simulating the inverse Wishart
distribution}
}
\value{
output from \code{uniroot()} to find the right estimate
}
\description{
This uses simulations to match the rse
}
\examples{

dfWishart(lotri::lotri(a+b~c(1, 0.5, 1)), 100)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikChisq}
\alias{llikChisq}
\title{log likelihood and derivatives for chi-squared distribution}
\usage{
llikChisq(x, df, full = FALSE)
}
\arguments{
\item{x}{variable that is distributed by chi-squared distribution}

\item{df}{degrees of freedom (non-negative, but can be non-integer).}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the log pdf value of with \code{dDf}
that has the derivatives with respect to the \code{df} parameter
the observation time-point
}
\description{
log likelihood and derivatives for chi-squared distribution
}
\details{
In an \code{rxode2()} model, you can use \code{llikChisq()} but you have to
use the x and df arguments.  You can also get the derivative of \code{df} with
\code{llikChisqDdf()}.
}
\examples{

\donttest{
llikChisq(1, df = 1:3, full=TRUE)

llikChisq(1, df = 6:9)

et <- et(1:3)
et$x <- 1

model <- function() {
  model({
   fx <- llikChisq(x, time)
   dDf <- llikChisqDdf(x, time)
  })
}

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evidPush.R
\name{replace}
\alias{replace}
\title{Replace dose the rxode2 system in the model}
\usage{
replace(amt, cmt = 1)
}
\arguments{
\item{amt}{Numeric dose amount (for dose events) or \code{0} for
observations.  When \code{rate > 0} this is interpreted as the total
infusion amount and the infusion duration is \code{amt / rate}.}

\item{cmt}{Integer compartment number (1-based) to which the dose is
applied.  Default is 1}
}
\value{
This function is only meaningful inside an rxode2 model; it
returns \code{NULL} invisibly if called from R directly (after signaling
an error).
}
\description{
Replace dose the rxode2 system in the model
}
\details{
\subsection{Behavior inside a model}{

\code{replace()} is evaluated at every output time point (when the solver is
exactly at a scheduled event time).  The pushed event is inserted into
the individual's event timeline and the solver visits it now or at the
specified future time.

The number of events that may be pushed per individual is limited by
the \code{maxExtra} argument of \code{\link[=rxSolve]{rxSolve()}}.  When \code{maxExtra = 0}
(the default) there is no limit.  Exceeding the limit causes an
error.

Past-time pushes (where \code{time < t}) are silently ignored and counted;
a warning is issued after solving.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxSolveFree}
\alias{rxSolveFree}
\title{Free the C solving/parsing information.}
\usage{
rxSolveFree()
}
\value{
logical indicating if the memory was successfully freed
}
\description{
Take the ODE C system and free it.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{rxStateOde}
\alias{rxStateOde}
\title{Get the ODE states only}
\usage{
rxStateOde(obj)
}
\arguments{
\item{obj}{rxode2 object}
}
\value{
ODE states only
}
\description{
Get the ODE states only
}
\examples{
\donttest{
mod <- rxode2({
  Cp <- linCmt(Cl, V, Q2, V2, Q3, V3)
  ke0 <- log(2)/(50)
  d/dt(Ce) <- (Cp-Ce)*ke0
})

rxStateOde(mod)

rxState(mod)

mod <- rxode2({
  Cp <- linCmt(Cl, V, Q2, V2, Q3, V3, ka)
  ke0 <- log(2)/(50)
  d/dt(Ce) <- (Cp-Ce)*ke0
}, linCmtSens="linCmtB")

rxStateOde(mod)

rxState(mod)
}

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evidPush.R
\name{obs}
\alias{obs}
\title{Add observations to a model}
\usage{
obs(...)
}
\arguments{
\item{...}{Numeric values to be added as observations after the
current time \code{t}.  They would be added as events with \code{evid = 0}
and \code{amt = 0} at the specified future time points.  The time
points are relative to the current model time \code{t} (i.e.,
\code{obs(12)} adds an observation at \code{t + 12}).  If this is a ui
model, you can also specify observations with obs(seq(0, 24,
by=0.5)) to add observations every 0.5 time units from t to t+24
since it will resolve to the obs() inside of the final rxode2
model.}
}
\value{
This function is only meaningful inside an rxode2 model; it
returns \code{NULL} invisibly if called from R directly (after signaling
an error).
}
\description{
Add observations to a model
}
\details{
\subsection{Behavior inside a model}{

\code{obs()} is evaluated at every output time point (when the solver is
exactly at a scheduled event time).  The pushed event is inserted into
the individual's event timeline and the solver visits it now or at the
specified future time.

The number of events that may be pushed per individual is limited by
the \code{maxExtra} argument of \code{\link[=rxSolve]{rxSolve()}}.  When \code{maxExtra = 0}
(the default) there is no limit.  Exceeding the limit causes an
error.

Past-time pushes (where \code{time < t}) are silently ignored and counted;
a warning is issued after solving.
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tran.R
\name{rxode2parseGetPointerAssignment}
\alias{rxode2parseGetPointerAssignment}
\title{This function gets the currently assigned function pointer assignments}
\usage{
rxode2parseGetPointerAssignment()
}
\value{
The currently assigned pointer assignments
}
\description{
This function gets the currently assigned function pointer assignments
}
\examples{
rxode2parseGetTranslation()
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/d.R
\name{rxode2parseD}
\alias{rxode2parseD}
\title{This gives the derivative table for rxode2}
\usage{
rxode2parseD()
}
\value{
Derivative table environment for rxode2
}
\description{
This will help allow registration of functions in \code{rxode2}
}
\details{
This environment is a derivative table;

For example:

Derivative(f(a,b,c), a) = fa()
Derivative(f(a,b,c), b) = fb()
Derivative(f(a,b,c), c) = fc()

Then the derivative table for \code{f} would be:

assign("f", list(fa(a,b,c), fb(a,b,c), fc(a,b,c)), rxode2parseD())

fa translates the arguments to the derivative with respect to a
fb translates the arguments to the derivative with respect to b

If any of the list is NULL then rxode2 won't know how to take a
derivative with respect to the argument.

If the list is shorter than the length of the arguments then the
argument then the derivative of arguments that are not specified
cannot be taken.
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{add.dosing}
\alias{add.dosing}
\title{Add dosing to eventTable}
\usage{
add.dosing(
  eventTable,
  dose,
  nbr.doses = 1L,
  dosing.interval = 24,
  dosing.to = 1L,
  rate = NULL,
  amount.units = NA_character_,
  start.time = 0,
  do.sampling = FALSE,
  time.units = NA_character_,
  ...
)
}
\arguments{
\item{eventTable}{eventTable object; When accessed from object it would be \verb{eventTable$}}

\item{dose}{numeric scalar, dose amount in \code{amount.units};}

\item{nbr.doses}{integer, number of doses;}

\item{dosing.interval}{required numeric scalar, time between doses
in \code{time.units}, defaults to 24 of
\code{time.units="hours"};}

\item{dosing.to}{integer, compartment the dose goes into (first
compartment by default);}

\item{rate}{for infusions, the rate of infusion (default is
\code{NULL}, for bolus dosing;}

\item{amount.units}{optional string indicating the dosing units.
Defaults to \code{NA} to indicate as per the original
\code{EventTable} definition.}

\item{start.time}{required dosing start time;}

\item{do.sampling}{logical, should observation sampling records be
added at the dosing times? Defaults to \code{FALSE}.}

\item{time.units}{optional string indicating the time units.
Defaults to \code{"hours"} to indicate as per the original
\code{EventTable} definition.}

\item{...}{Other parameters passed to \code{\link[=et]{et()}}.}
}
\value{
eventTable with updated dosing (note the event table will
be updated anyway)
}
\description{
This adds a dosing event to the event table.  This is provided for
piping syntax through magrittr.  It can also be accessed by \code{eventTable$add.dosing(...)}
}
\examples{

\dontrun{

library(rxode2)
library(units)

# Model from rxode2 tutorial
# Using a nlmixr2 style function

mod1 <-function(){
  ini({
    KA <- 2.94E-01
    CL <- 1.86E+01
    V2 <- 4.02E+01
    Q <- 1.05E+01
    V3 <- 2.97E+02
    Kin <- 1
    Kout <- 1
    EC50 <- 200
  })
 model({
    C2 <- centr/V2
    C3 <- peri/V3
    d/dt(depot) <- -KA*depot
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
    d/dt(peri)  <-                    Q*C2 - Q*C3
    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
 })
}

## These are making the more complex regimens of the rxode2 tutorial

## bid for 5 days
bid <- et(timeUnits="hr") |>
       et(amt=10000,ii=12,until=set_units(5, "days"))

## qd for 5 days
qd <- et(timeUnits="hr") |>
      et(amt=20000,ii=24,until=set_units(5, "days"))

## bid for 5 days followed by qd for 5 days

et <- seq(bid,qd) |>
      et(seq(0,11*24,length.out=100))

bidQd <- rxSolve(mod1, et)

plot(bidQd, C2)


## Now Infusion for 5 days followed by oral for 5 days

##  note you can dose to a named compartment instead of using the compartment number
infusion <- et(timeUnits = "hr") |>
      et(amt=10000, rate=5000, ii=24, until=set_units(5, "days"), cmt="centr")


qd <- et(timeUnits = "hr") |>
  et(amt=10000, ii=24, until=set_units(5, "days"), cmt="depot")

et <- seq(infusion,qd)

infusionQd <- rxSolve(mod1, et)

plot(infusionQd, C2)

## 2wk-on, 1wk-off

qd <- et(timeUnits = "hr") |>
      et(amt=10000, ii=24, until=set_units(2, "weeks"), cmt="depot")

et <- seq(qd, set_units(1,"weeks"), qd) |>
     add.sampling(set_units(seq(0, 5.5,by=0.005),weeks))

wkOnOff <- rxSolve(mod1, et)

plot(wkOnOff, C2)

## You can also repeat the cycle easily with the rep function

qd <-et(timeUnits = "hr") |>
     et(amt=10000, ii=24, until=set_units(2, "weeks"), cmt="depot")

et <- etRep(qd, times=4, wait=set_units(1,"weeks")) |>
      add.sampling(set_units(seq(0, 12.5,by=0.005),weeks))

repCycle4 <- rxSolve(mod1, et)

plot(repCycle4, C2)

}
}
\references{
Wang W, Hallow K, James D (2015). "A Tutorial on rxode2: Simulating
Differential Equation Pharmacometric Models in R." CPT:
Pharmacometrics and Systems Pharmacology, 5(1), 3-10. ISSN 2163-8306
}
\seealso{
\code{\link{eventTable}}, \code{\link{add.sampling}},
\code{\link{add.dosing}}, \code{\link{et}},
\code{\link{etRep}}, \code{\link{etRbind}},
\code{\link{rxode2}}
}
\author{
Matthew L. Fidler

Matthew L Fidler, Wenping Wang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err-foceiBase.R
\name{.handleSingleErrTypeNormOrTFoceiBase}
\alias{.handleSingleErrTypeNormOrTFoceiBase}
\title{Handle the single error for normal or t distributions}
\usage{
.handleSingleErrTypeNormOrTFoceiBase(
  env,
  pred1,
  errNum = 1L,
  rxPredLlik = TRUE,
  arNorm = FALSE
)
}
\arguments{
\item{env}{Environment for the parsed model}

\item{pred1}{The \code{data.frame} of the current error}

\item{errNum}{The number of the error specification in the nlmixr2 model}

\item{rxPredLlik}{A boolean indicating if the log likelihood should
be calculated for non-normal distributions.  By default \code{TRUE}.}

\item{arNorm}{A boolean; when \code{TRUE}, emit the AR(1) norm
(conditional mean/variance) form used by the focei inner model.
By default \code{FALSE}.}
}
\value{
A list of the lines added.  The lines will contain
\itemize{
\item \code{rx_yj_} which is an integer that corresponds to the
transformation type.
\item \code{rx_lambda_} is the transformation lambda
\item \code{rx_low_} The lower boundary of the transformation
\item \code{rx_hi_} The upper boundary of the transformation
\item \code{rx_pred_f_} The prediction function
\item \code{rx_pred_} The transformed prediction function
\item \code{rx_r_} The transformed variance
}
}
\description{
Handle the single error for normal or t distributions
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxPrune.R
\name{rxPrune}
\alias{rxPrune}
\title{Prune branches (ie if/else) from rxode2}
\usage{
rxPrune(x)
}
\arguments{
\item{x}{rxode2 model that can be accessed by rxNorm}
}
\value{
Pruned rxode2 model text.
}
\description{
This prunes branches (ie if/else) from rxode2.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxStack.R
\name{is.rxStackData}
\alias{is.rxStackData}
\title{Return if the object can be stacked}
\usage{
is.rxStackData(object)
}
\arguments{
\item{object}{object to test if it can be stacked}
}
\value{
boolean to tell if an object can be stacked using rxode2
}
\description{
Return if the object can be stacked
}
\examples{

is.rxStackData(NULL)
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dde.R
\name{delay}
\alias{delay}
\title{Delayed state for delay differential equations}
\usage{
delay(state, T)
}
\arguments{
\item{state}{An ODE state (compartment) defined in the model whose
past value is required.}

\item{T}{The delay duration.  May be a constant, a parameter, a
covariate, or any model expression.  The value returned is the state
at time \code{t - T}.}
}
\value{
Inside an rxode2 model, the value of \code{state} at the past time
\code{t - T}.  Before the start of integration the constant
initial-condition history is used.
}
\description{
\code{delay(state, T)} evaluates an ODE state at the past time \code{t - T}, turning
an ordinary differential equation model into a delay differential equation
(DDE). The semantics match the \code{delay()} function of Monolix.
}
\details{
Delayed states are interpolated from the solver's dense output, so delay
models are solved on a dense path: the default method becomes the dense
AutoSwitch composite \code{"dop853+ros4"}, and dense methods such as \code{"dop853"}
or \code{"ros4"} also work.  The step size is capped to the smallest delay, and
methods that cannot record dense history raise an error.  The dense-output
and delay-history machinery is adapted from the \code{dde} package by Rich
FitzJohn and Wes Hinsley (Imperial College of Science, Technology and
Medicine).
}
\examples{
\donttest{
# Classic linear delay differential equation y'(t) = -y(t - 1)
dde <- rxode2({
  y(0) <- 1
  d/dt(y) <- -delay(y, 1)
})

s <- rxSolve(dde, et(seq(0, 5, by = 0.1)))

# Delayed (Hutchinson) logistic growth
hutch <- rxode2({
  r <- 0.5
  K <- 10
  tau <- 1
  N(0) <- 2
  d/dt(N) <- r * N * (1 - delay(N, tau) / K)
})

s2 <- rxSolve(hutch, et(seq(0, 40, by = 0.5)))
}

}
\seealso{
\code{\link[=rxSolve]{rxSolve()}}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{rxord}
\alias{rxord}
\title{Simulate ordinal value}
\usage{
rxord(...)
}
\arguments{
\item{...}{the probabilities to be simulated.  These should sum up to a number below one.}
}
\value{
A number from 1 to the (number of probabilities specified + 1)
}
\description{
Simulate ordinal value
}
\details{
The values entered into the 'rxord' simulation will simulate the
probability of falling each group. If it falls outside of the
specified probabilities, it will simulate the group (number of
probabilities specified + 1)
}
\examples{

# This will give values 1, and 2
rxord(0.5)
rxord(0.5)
rxord(0.5)
rxord(0.5)

# This will give values 1, 2 and 3
rxord(0.3, 0.3)
rxord(0.3, 0.3)
rxord(0.3, 0.3)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikExp}
\alias{llikExp}
\title{log likelihood and derivatives for exponential distribution}
\usage{
llikExp(x, rate, full = FALSE)
}
\arguments{
\item{x}{variable that is distributed by exponential distribution}

\item{rate}{vector of rates.}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the log pdf value of with \code{dRate}
that has the derivatives with respect to the \code{rate} parameter
the observation time-point
}
\description{
log likelihood and derivatives for exponential distribution
}
\details{
In an \code{rxode2()} model, you can use \code{llikExp()} but you have to
use the x and rate arguments.  You can also get the derivative of \code{rate} with
\code{llikExpDrate()}.
}
\examples{
\donttest{
llikExp(1, 1:3)

llikExp(1, 1:3, full=TRUE)

# You can use rxode2 for these too:

et <- et(1:3)
et$x <- 1

model <- function() {
  model({
    fx <- llikExp(x, time)
    dRate <- llikExpDrate(x, time)
  })
}

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symengine.R
\name{rxFun}
\alias{rxFun}
\alias{rxRmFun}
\title{Add/Create C functions for use in rxode2}
\usage{
rxFun(name, args, cCode)

rxRmFun(name)
}
\arguments{
\item{name}{This can either give the name of the user function or
be a simple R function that you wish to convert to C.  If you
have rxode2 convert the R function to C, the name of the function
will match the function name provided and the number of arguments
will match the R function provided.  Hence, if you are providing
an R function for conversion to C, the rest of the arguments are
implied.}

\item{args}{This gives the arguments of the user function}

\item{cCode}{This is the C-code for the new function}
}
\description{
Add/Create C functions for use in rxode2
}
\examples{
\donttest{
# Right now rxode2 is not aware of the function fun
# Therefore it cannot translate it to symengine or
# Compile a model with it.

try(rxode2("a=fun(a,b,c)"))

# Note for this approach to work, it cannot interfere with C
# function names or reserved rxode2 special terms.  Therefore
# f(x) would not work since f is an alias for bioavailability.

fun <- "
double fun(double a, double b, double c) {
  return a*a+b*a+c;
}
" # C-code for function

rxFun("fun", c("a", "b", "c"), fun) ## Added function

# Now rxode2 knows how to translate this function to symengine

rxToSE("fun(a,b,c)")

# And will take a central difference when calculating derivatives

rxFromSE("Derivative(fun(a,b,c),a)")

## Of course, you could specify the derivative table manually
rxD("fun", list(
  function(a, b, c) {
    paste0("2*", a, "+", b)
  },
  function(a, b, c) {
    return(a)
  },
  function(a, b, c) {
    return("0.0")
  }
))

rxFromSE("Derivative(fun(a,b,c),a)")

# You can also remove the functions by `rxRmFun`

rxRmFun("fun")

# you can also use R functions directly in rxode2


gg <- function(x, y) {
  x + y
}

f <- rxode2({
 z = gg(x, y)
})


e <- et(1:10) |> as.data.frame()

e$x <- 1:10
e$y <- 21:30

rxSolve(f, e)

# Note that since it touches R, it can only run single-threaded.
# There are also requirements for the function:
#
# 1. It accepts one value per argument (numeric)
#
# 2. It returns one numeric value

# If it is a simple function (like gg) you can also convert it to C
# using rxFun and load it into rxode2

rxFun(gg)

rxSolve(f, e)

# to stop the recompile simply reassign the function
f <- rxode2(f)

rxSolve(f, e)

rxRmFun("gg")
rm(gg)
rm(f)


# You can also automatically convert a R function to R code (and
# calculate first derivatives)

fun <- function(a, b, c) {
  a^2+b*a+c
}

rxFun(fun)

# You can see the R code if you want with rxC

message(rxC("fun"))

# you can also remove both the function and the
# derivatives with rxRmFun("fun")

rxRmFun("fun")

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-getControl.R
\name{rxGetControl}
\alias{rxGetControl}
\title{rxGetControl option from ui}
\usage{
rxGetControl(ui, option, default)
}
\arguments{
\item{ui}{rxode2 ui object}

\item{option}{Option to get}

\item{default}{Default value}
}
\value{
Option (if present) or default value
}
\description{
rxGetControl option from ui
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxSymInv.R
\name{rxSymInvCholN}
\alias{rxSymInvCholN}
\title{Return the dimension of the built-in derivatives/inverses}
\usage{
rxSymInvCholN()
}
\value{
dimension of built-in derivatives/inverses
}
\description{
Return the dimension of the built-in derivatives/inverses
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxRaw.R
\name{rxDeserialize}
\alias{rxDeserialize}
\title{Deserialize a Raw Vector or String to an R Object}
\usage{
rxDeserialize(x)
}
\arguments{
\item{x}{raw vector or string to deserialize}
}
\value{
R object
}
\description{
Deserialize a Raw Vector or String to an R Object
}
\examples{

rxDeserialize(rxSerialize(mtcars))

}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.rxLinCmt}
\alias{.rxLinCmt}
\title{Get the linear compartment model states}
\usage{
.rxLinCmt(obj)
}
\arguments{
\item{obj}{rxode2 type of object}
}
\value{
compartment names
}
\description{
Get the linear compartment model states
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode-options.R
\name{rxForget}
\alias{rxForget}
\title{Clear memoise caches for rxode2}
\usage{
rxForget()
}
\value{
nothing; called for side effects
}
\description{
Clear memoise caches for rxode2
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxAssignPtr}
\alias{rxAssignPtr}
\title{Assign pointer based on model variables}
\usage{
rxAssignPtr(object = NULL)
}
\arguments{
\item{object}{rxode2 family of objects}
}
\value{
nothing, called for side effects
}
\description{
Assign pointer based on model variables
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{rxCat}
\alias{rxCat}
\title{Use cat when rxode2.verbose is TRUE}
\usage{
rxCat(a, ...)
}
\arguments{
\item{...}{Parameters sent to cat}
}
\value{
nothing
}
\description{
Use cat when rxode2.verbose is TRUE
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjoint.R
\name{.rxAdjointSolveEvalC}
\alias{.rxAdjointSolveEvalC}
\title{Evaluate a prebuilt full-trajectory adjoint solve in C++}
\usage{
.rxAdjointSolveEvalC(
  build,
  params,
  outTimes,
  denseBy = 0.01,
  atol = 1e-10,
  rtol = 1e-10
)
}
\arguments{
\item{build}{object returned by \code{\link[=.rxAdjointSolveBuild]{.rxAdjointSolveBuild()}}.}

\item{params}{named numeric vector of parameter values.}

\item{outTimes}{numeric vector of output times.}

\item{denseBy, atol, rtol}{as in \code{\link[=.rxAdjointSolve]{.rxAdjointSolve()}}.}
}
\value{
data.frame with a \code{time} column and one
\verb{rx__sens_<state>_BY_<param>__} column per (state, param) pair.
}
\description{
Full-trajectory counterpart of \code{\link[=.rxAdjointGradEvalC]{.rxAdjointGradEvalC()}}: reconstructs
\code{dy_k(t_i)/dp} for every output state, time and parameter in one backward
sweep (\code{rxode2AdjointTrajSweep}), covering all dosing duals.  Numerically
equivalent to \code{\link[=.rxAdjointSolve]{.rxAdjointSolve()}}.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err-sim.R
\name{rxGetDistributionSimulationLines}
\alias{rxGetDistributionSimulationLines}
\alias{rxGetDistributionSimulationLines.norm}
\alias{rxGetDistributionSimulationLines.dnorm}
\alias{rxGetDistributionSimulationLines.t}
\alias{rxGetDistributionSimulationLines.cauchy}
\alias{rxGetDistributionSimulationLines.ordinal}
\alias{rxGetDistributionSimulationLines.default}
\alias{rxGetDistributionSimulationLines.rxUi}
\title{This is a S3 method for getting the distribution lines for a rxode2 simulation}
\usage{
rxGetDistributionSimulationLines(line)

\method{rxGetDistributionSimulationLines}{norm}(line)

\method{rxGetDistributionSimulationLines}{dnorm}(line)

\method{rxGetDistributionSimulationLines}{t}(line)

\method{rxGetDistributionSimulationLines}{cauchy}(line)

\method{rxGetDistributionSimulationLines}{ordinal}(line)

\method{rxGetDistributionSimulationLines}{default}(line)

\method{rxGetDistributionSimulationLines}{rxUi}(line)
}
\arguments{
\item{line}{Parsed rxode2 model environment}
}
\value{
Lines for the simulation of \code{ipred} and \code{dv}. This is based
on the idea that the focei parameters are defined
}
\description{
This is a S3 method for getting the distribution lines for a rxode2 simulation
}
\author{
Matthew Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode-options.R
\name{.rxode2ptrs}
\alias{.rxode2ptrs}
\title{Get the rxode2 function pointers}
\usage{
.rxode2ptrs()
}
\value{
a list of function pointers
}
\description{
This function is used to get the function pointers for rxode2.  This is
used to allow rxode2 to have binary linkage to nlmixr2est.
}
\examples{

.rxode2ptrs()
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-assign-parts.R
\name{model<-}
\alias{model<-}
\title{Assign the model block in the rxode2 related object}
\usage{
model(x, envir = environment(x)) <- value
}
\arguments{
\item{x}{rxode2 related object}

\item{envir}{Environment where assignment occurs}

\item{value}{Value of the object}
}
\value{
rxode2 related object
}
\description{
Assign the model block in the rxode2 related object
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modlib.R
\name{rxPkg}
\alias{rxPkg}
\title{Creates a package from compiled rxode2 models}
\usage{
rxPkg(
  ...,
  package,
  wd = getwd(),
  action = c("install", "build", "binary", "create"),
  license = c("gpl3", "lgpl", "mit", "agpl3"),
  name = "Firstname Lastname",
  fields = list()
)
}
\arguments{
\item{...}{Models to build a package from}

\item{package}{String of the package name to create}

\item{wd}{character string with a working directory where to
create a subdirectory according to \code{modName}. When
specified, a subdirectory named after the
\dQuote{\code{modName.d}} will be created and populated with a
C file, a dynamic loading library, plus various other working
files. If missing, the files are created (and removed) in the
temporary directory, and the rxode2 DLL for the model is
created in the current directory named \verb{rx_????_platform}, for
example \code{rx_129f8f97fb94a87ca49ca8dafe691e1e_i386.dll}}

\item{action}{Type of action to take after package is created}

\item{license}{is the type of license for the package.}

\item{name}{Full name of author}

\item{fields}{A named list of fields to add to \code{DESCRIPTION}, potentially
overriding default values. See \code{\link[usethis:use_description]{use_description()}} for how you can set
personalized defaults using package options.}
}
\value{
this function returns nothing and is used for its side effects
}
\description{
Creates a package from compiled rxode2 models
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tran.R
\name{rxode2parseGetPackagesToLoad}
\alias{rxode2parseGetPackagesToLoad}
\alias{rxode2parseAssignPackagesToLoad}
\title{Control the packages that are loaded when a \code{rxode2} model dll is loaded}
\usage{
rxode2parseGetPackagesToLoad()

rxode2parseAssignPackagesToLoad(pkgs = rxode2parseGetPackagesToLoad())
}
\arguments{
\item{pkgs}{The packages to make sure are loaded every time you load an rxode2 model.}
}
\value{
List of packages to load
}
\description{
Control the packages that are loaded when a \code{rxode2} model dll is loaded
}
\examples{

rxode2parseGetPackagesToLoad()

rxode2parseAssignPackagesToLoad(rxode2parseGetPackagesToLoad())
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxJacobian.R
\name{.rxSens}
\alias{.rxSens}
\title{Sensitivity for model}
\usage{
.rxSens(model, vars, vars2, vars3)
}
\arguments{
\item{model}{symengine model environment}

\item{vars}{Variables for single sensitivity}

\item{vars2}{if present, 2 parameter sensitivity}
}
\value{
Sensitivity
}
\description{
Sensitivity for model
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxRaw.R
\name{rxRawToC}
\alias{rxRawToC}
\title{Convert a Raw Vector or R object to C Code}
\usage{
rxRawToC(raw, type = c("xz", "base", "bzip2"))
}
\arguments{
\item{raw}{vector or R object to convert}

\item{type}{serialization type; one of "base", "xz" or "bzip2".}
}
\value{
raw vector for conversion to C in the codegen procedure
}
\description{
This function converts a raw vector or R object to C code that
is used in the rxode2 model
}
\examples{

rxRawToC(mtcars)

}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode-options.R
\name{rxTempDir}
\alias{rxTempDir}
\title{Get the rxode2 temporary directory}
\usage{
rxTempDir()
}
\value{
rxode2 temporary directory.
}
\description{
Get the rxode2 temporary directory
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piping-ini.R, R/ui.R
\name{ini.rxUi}
\alias{ini.rxUi}
\alias{ini.default}
\alias{ini}
\title{Ini block for rxode2/nlmixr models}
\usage{
\method{ini}{rxUi}(x, ..., envir = parent.frame(), append = NULL)

\method{ini}{default}(x, ..., envir = parent.frame(), append = NULL)

ini(x, ..., envir = parent.frame(), append = NULL)
}
\arguments{
\item{x}{expression}

\item{...}{Other expressions for \code{ini()} function}

\item{envir}{the \code{environment} in which unevaluated model
expressions is to be evaluated.  May also be \code{NULL}, a list, a
data frame, a pairlist or an integer as specified to \code{sys.call}.}

\item{append}{Reorder theta parameters.  \code{NULL} means no change to
parameter order.  A parameter name (as a character string) means to put the
new parameter after the named parameter.  A number less than or equal to
zero means to put the parameter at the beginning of the list.  A number
greater than the last parameter number means to put the parameter at the
end of the list.}
}
\value{
ini block
}
\description{
The ini block controls initial conditions for 'theta' (fixed effects),
'omega' (random effects), and 'sigma' (residual error) elements of the model.
}
\details{
The \code{ini()} function is used in two different ways.  The main way that
it is used is to set the initial conditions and associated attributes
(described below) in a model.  The other way that it is used is for updating
the initial conditions in a model, often using the pipe operator.

'theta' and 'sigma' can be set using either \code{<-} or \code{=} such as
\code{tvCL <- 1} or equivalently \code{tvCL = 1}.  'omega' can be set with a
\code{~} such as \code{etaCL ~ 0.1}.

Parameters can be named or unnamed (though named parameters are preferred).
A named parameter is set using the name on the left of the assignment while
unnamed parameters are set without an assignment operator.  \code{tvCL <- 1}
would set a named parameter of \code{tvCL} to \code{1}.  Unnamed parameters
are set using just the value, such as \code{1}.

For some estimation methods, lower and upper bounds can be set for 'theta'
and 'sigma' values.  To set a lower and/or upper bound, use a vector of
values.  The vector is \code{c(lower, estimate, upper)}.  The vector may be
given with just the estimate (\code{estimate}), the lower bound and
estimate (\code{c(lower, estimate)}), or all three (\code{c(lower, estimate,
upper)}).  To set an estimate and upper bound without a lower bound, set the
lower bound to \code{-Inf}, \code{c(-Inf, estimate, upper)}.  When an
estimation method does not support bounds, the bounds will be ignored with a
warning.

'omega' values can be set as a single value or as the values of a
lower-triangular matrix.  The values may be set as either a
variance-covariance matrix (the default) or as a correlation matrix
for the off-diagonals with the standard deviations on the
diagonals.  Names may be set on the left side of the \code{~}.  To
set a variance-covariance matrix with variance values of 2 and 3
and a covariance of -2.5 use \code{~c(2, 2.5, 3)}.  To set the same
matrix with names of \code{iivKa} and \code{iivCL}, use \code{iivKa
+ iivCL~c(2, 2.5, 3)}.  To set a correlation matrix with standard
deviations on the diagonal, use \code{cor()} like \code{iivKa +
iivCL~cor(2, -0.5, 3)}.  As of rxode2 3.0 you can also use
\code{iivKa ~ 2, iivCL ~ c(2.5, 3)} for covariance matrices as
well.

Values may be fixed (and therefore not estimated) using either the name
\code{fixed} at the end of the assignment or by calling \code{fixed()} as a
function for the value to fix.  For 'theta' and 'sigma', either the estimate
or the full definition (including lower and upper bounds) may be included in
the fixed setting.  For example, the following are all effectively equivalent
to set a 'theta' or 'sigma' to a fixed value (because the lower and upper
bounds are ignored for a fixed value): \code{tvCL <- fixed(1)}, \code{tvCL <-
fixed(0, 1)}, \code{tvCL <- fixed(0, 1, 2)}, \code{tvCL <- c(0, fixed(1),
2)}, or \code{tvCL <- c(0, 1, fixed)}.  For 'omega' assignment, the full
block or none of the block must be set as \code{fixed}.  Examples of setting
an 'omega' value as fixed are: \code{iivKa~fixed(1)}, \code{iivKa +
iivCL~fixed(1, 2, 3)}, or \code{iivKa + iivCL~c(1, 2, 3, fixed)}.  Anywhere
that \code{fixed} is used, \code{FIX}, \code{FIXED}, or \code{fix} may be
used equivalently.

For any value, standard mathematical operators or functions may be used to
define the value.  For example, \code{log(2)} and \code{24*30} may be used to
define a value anywhere that a number can be used (e.g. lower bound,
estimate, upper bound, variance, etc.).

Values may be labeled using the \code{label()} function after the assignment.
Labels are are used to make reporting easier by giving a human-readable
description of the parameter, but the labels do not have any effect on
estimation.  The typical way to set a label so that the parameter \code{tvCL}
has a label of "Typical Value of Clearance (L/hr)" is \code{tvCL <- 1;
label("Typical Value of Clearance (L/hr)")}.

Off diagonal values of 'omega' can be set to zero using the
\code{diag()} to remove all off-diagonals can be removed with
\code{ini(diag())}.  To remove covariances of 'omega' item with \code{iivKa},
you can use \verb{|> ini(diag(iivKa))}.  Or to remove covariances that
contain either \code{iivKa} or \code{iivCl} you can use \verb{|> ini(diag(iivKa, iivCl))}.  For finer control you can remove the covariance between
two items (like \code{iivKa} and \code{iivCl}) by `|> ini(-cov(iivKa, iivCl))

\code{rxode2}/\code{nlmixr2} will attempt to determine some
back-transformations for the user.  For example, \code{CL <- exp(tvCL)} will
detect that \code{tvCL} must be back-transformed by \code{exp()} for easier
interpretation.  When you want to control the back-transformation, you can
specify the back-transformation using \code{backTransform()} after the
assignment.  For example, to set the
back-transformation to \code{exp()}, you can use \code{tvCL <- 1;
backTransform(exp())}.
}
\examples{
# Set the ini() block in a model
one.compartment <- function() {
  ini({
    tka <- log(1.57); label("Ka")
    tcl <- log(2.72); label("Cl")
    tv <- log(31.5); label("V")
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v <- exp(tv + eta.v)
    d/dt(depot) = -ka * depot
    d/dt(center) = ka * depot - cl / v * center
    cp = center / v
    cp ~ add(add.sd)
  })
}

# Use piping to update initial conditions
one.compartment |> ini(tka <- log(2))
one.compartment |> ini(tka <- label("Absorption rate, Ka (1/hr)"))
# Move the tka parameter to be just below the tv parameter (affects parameter
# summary table, only)
one.compartment |> ini(tka <- label("Absorption rate, Ka (1/hr)"), append = "tv")
# When programming with rxode2/nlmixr2, it may be easier to pass strings in
# to modify the ini
one.compartment |> ini("tka <- log(2)")
}
\seealso{
Other Initial conditions:
\code{\link[=zeroRe]{zeroRe()}}
}
\author{
Matthew Fidler
}
\concept{Initial conditions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assert.R
\name{testRxLinCmt}
\alias{testRxLinCmt}
\alias{assertRxLinCmt}
\title{Test if rxode2 uses linear solved systems}
\usage{
testRxLinCmt(ui, extra = "", .var.name = .vname(ui))

assertRxLinCmt(ui, extra = "", .var.name = .vname(ui))
}
\arguments{
\item{ui}{rxode2 model}

\item{extra}{Extra text to append to the error message (like
"for focei")}

\item{.var.name}{[\code{character(1)}]\cr
Name of the checked object to print in assertions. Defaults to
the heuristic implemented in \code{\link[checkmate]{vname}}.}
}
\value{
TRUE if the model uses linear solved systems, FALSE otherwise
}
\description{
Test if rxode2 uses linear solved systems
}
\section{Functions}{
\itemize{
\item \code{assertRxLinCmt()}: Assert that the rxode2 uses linear solved systems

}}
\examples{

one.cmt <- function() {
  ini({
   ## You may label each parameter with a comment
   tka <- 0.45 # Log Ka
   tcl <- log(c(0, 2.7, 100)) # Log Cl
   ## This works with interactive models
   ## You may also label the preceding line with label("label text")
   tv <- 3.45; label("log V")
   ## the label("Label name") works with all models
   eta.ka ~ 0.6
   eta.cl ~ 0.3
   eta.v ~ 0.1
   add.sd <- 0.7
 })
 model({
   ka <- exp(tka + eta.ka)
   cl <- exp(tcl + eta.cl)
   v <- exp(tv + eta.v)
   linCmt() ~ add(add.sd)
 })
}

testRxLinCmt(one.cmt)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lrelu.R
\name{dlReLU}
\alias{dlReLU}
\title{Derivative of Leaky ReLU activation function}
\usage{
dlReLU(x)
}
\arguments{
\item{x}{numeric vector}
}
\value{
numeric vector
}
\description{
Derivative of Leaky ReLU activation function
}
\examples{

dlReLU(c(-1, 0, 1))

\donttest{
# Can use in rxode2 as well

r <- rxode2({r <- dlReLU(time)})
e <- et(c(-1, 0, 1))
rxSolve(r, e)
}

}
\seealso{
Other Activation Functions:
\code{\link[=ELU]{ELU()}},
\code{\link[=GELU]{GELU()}},
\code{\link[=PReLU]{PReLU()}},
\code{\link[=ReLU]{ReLU()}},
\code{\link[=SELU]{SELU()}},
\code{\link[=Swish]{Swish()}},
\code{\link[=dELU]{dELU()}},
\code{\link[=dGELU]{dGELU()}},
\code{\link[=dPReLU]{dPReLU()}},
\code{\link[=dReLU]{dReLU()}},
\code{\link[=dSELU]{dSELU()}},
\code{\link[=dSwish]{dSwish()}},
\code{\link[=dsoftplus]{dsoftplus()}},
\code{\link[=lReLU]{lReLU()}},
\code{\link[=softplus]{softplus()}}
}
\concept{Activation Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dsl.R
\name{rxExpandIfElse}
\alias{rxExpandIfElse}
\title{Expand if/else clauses into multiple different types of lines.}
\usage{
rxExpandIfElse(model, removeInis = TRUE, removePrint = TRUE)
}
\arguments{
\item{model}{Model can be a character, or a rxode2 model.  It needs
to have normalized syntax, that is \code{if (...){}} has to be
on the same line.  The \verb{else} statement must be on its
own line with the closing bracket of the \code{if} statement
on the previous line.  This \verb{else} statment must also
contain the opening bracket, like the code \verb{else \{\}}}

\item{removeInis}{A boolean indicating if parameter
initializations should be removed from the model.}

\item{removePrint}{A boolean indicating if printing statements
should be removed from the model.}
}
\value{
A named character vector. The names of the vector are the
logical conditions, the values are the lines that satisfy the
logical conditions.
}
\description{
Expand if/else clauses into multiple different types of lines.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjointDiscrete.R
\name{.rxDiscreteForwardSens}
\alias{.rxDiscreteForwardSens}
\title{Discrete forward sensitivity of a fixed-step RK4 solve}
\usage{
.rxDiscreteForwardSens(build, X0, params, h, nStep, doses = NULL, S0 = NULL)
}
\arguments{
\item{build}{object from \code{\link[=.rxDiscreteAdjointBuild]{.rxDiscreteAdjointBuild()}}.}

\item{X0}{named numeric initial state (order = \code{build$st}).}

\item{params}{named numeric parameter vector.}

\item{h}{fixed step size.}

\item{nStep}{number of RK4 steps (final time \verb{= h * nStep}).}

\item{doses}{optional list of additive-bolus dose specs, each
\verb{list(step = <0-based grid step>, cmt = <state name>, amt = <numeric>)};
applied (with bioavailability \code{f(cmt)} if defined) at the start of the step.}

\item{S0}{optional initial sensitivity (\verb{ns x np}); default zero.}
}
\value{
list with \code{XN}, \code{SN} (\verb{ns x np}), \code{Sall} (per-step \code{S}), \code{stages}.
}
\description{
Integrates the state and its exact derivative \code{S = dX/dtheta} through the RK4
map, storing the stages for the discrete adjoint.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxSetIni0}
\alias{rxSetIni0}
\title{Set Initial conditions to time zero instead of the first observed/dosed time}
\usage{
rxSetIni0(ini0 = TRUE)
}
\arguments{
\item{ini0}{When \code{TRUE} (default), set initial conditions to time
zero. Otherwise the initial conditions are the first observed
time.}
}
\value{
the boolean ini0, though this is called for its side effects
}
\description{
Set Initial conditions to time zero instead of the first observed/dosed time
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxUdfUiNum}
\alias{rxUdfUiNum}
\title{This gives the current number in the ui of the particular function being called.}
\usage{
rxUdfUiNum()
}
\value{
integer greater than 1L
}
\description{
If this is called outside of function parsing or the input is
unexpected this returns 1L. This is useful when writing replacement
UI functions
}
\examples{

rxUdfUiNum()
}
\seealso{
Other User functions:
\code{\link[=linMod]{linMod()}},
\code{\link[=rxUdfUiControl]{rxUdfUiControl()}},
\code{\link[=rxUdfUiData]{rxUdfUiData()}},
\code{\link[=rxUdfUiEst]{rxUdfUiEst()}},
\code{\link[=rxUdfUiIniLhs]{rxUdfUiIniLhs()}},
\code{\link[=rxUdfUiMv]{rxUdfUiMv()}},
\code{\link[=rxUdfUiParsing]{rxUdfUiParsing()}}
}
\author{
Matthew L. Fidler
}
\concept{User functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress.R
\name{rxProgress}
\alias{rxProgress}
\alias{rxTick}
\alias{rxProgressStop}
\alias{rxProgressAbort}
\title{rxode2 progress bar functions}
\usage{
rxProgress(num, core = 0L)

rxTick()

rxProgressStop(clear = TRUE)

rxProgressAbort(error = "Aborted calculation")
}
\arguments{
\item{num}{Tot number of operations to track}

\item{core}{Number of cores to show.  If below 1, don't show
number of cores}

\item{clear}{Boolean telling if you should clear the progress bar
after completion (as if it wasn't displayed).  By default this is TRUE}

\item{error}{With rxProgressAbort this is the error that is displayed}
}
\value{
All return NULL invisibly.
}
\description{
\code{rxProgress} sets up the progress bar
}
\details{
\code{rxTick} is a progress bar tick

\code{rxProgressStop} stop progress bar

\code{rxProgressAbort} shows an abort if \code{rxProgressStop}
wasn't called.
}
\examples{

f <- function() {
  on.exit({
    rxProgressAbort()
  })
  rxProgress(100)
  for (i in 1:100) {
    rxTick()
    Sys.sleep(1 / 100)
  }
  rxProgressStop()
}
\donttest{
f()
}

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symengine.R
\name{rxSplitPlusQ}
\alias{rxSplitPlusQ}
\title{This function splits a function based on + or - terms}
\usage{
rxSplitPlusQ(x, level = 0, mult = FALSE)
}
\arguments{
\item{x}{Quoted R expression for splitting}

\item{level}{Internal level of parsing}

\item{mult}{boolean to split based on * and / expressions instead.
By default this is turned off.}
}
\value{
character vector of the split expressions
}
\description{
It uses the parser and does not disturb terms within other
functions.  For example:
}
\details{
a\emph{exp(b+c)+d}log(e-f)-g*f

would return

c("a * exp(b + c)", "d * log(e - f)", "- g * f")
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxJacobian.R
\name{.rxJacobian}
\alias{.rxJacobian}
\title{Internal function for calculating the Jacobian}
\usage{
.rxJacobian(model, vars = TRUE)
}
\arguments{
\item{model}{symengine environment}

\item{vars}{Variables}
}
\value{
Jacobian information
}
\description{
Internal function for calculating the Jacobian
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assert.R
\name{assertRxUi}
\alias{assertRxUi}
\alias{assertRxUiPrediction}
\alias{assertRxUiIovNoCor}
\alias{assertRxUiNoMix}
\alias{assertRxUiNoAutoregressive}
\alias{assertRxUiSingleEndpoint}
\alias{assertRxUiTransformNormal}
\alias{assertRxUiNormal}
\alias{assertRxUiMuRefOnly}
\alias{assertRxUiEstimatedResiduals}
\alias{assertRxUiPopulationOnly}
\alias{assertRxUiMixedOnly}
\alias{assertRxUiRandomOnIdOnly}
\title{Assert properties of the rxUi models}
\usage{
assertRxUi(ui, extra = "", .var.name = .vname(ui))

assertRxUiPrediction(ui, extra = "", .var.name = .vname(ui))

assertRxUiIovNoCor(ui, extra = "", .var.name = .vname(ui))

assertRxUiNoMix(ui, extra = "", .var.name = .vname(ui))

assertRxUiNoAutoregressive(ui, extra = "", .var.name = .vname(ui))

assertRxUiSingleEndpoint(ui, extra = "", .var.name = .vname(ui))

assertRxUiTransformNormal(ui, extra = "", .var.name = .vname(ui))

assertRxUiNormal(ui, extra = "", .var.name = .vname(ui))

assertRxUiMuRefOnly(ui, extra = "", .var.name = .vname(ui))

assertRxUiEstimatedResiduals(ui, extra = "", .var.name = .vname(ui))

assertRxUiPopulationOnly(ui, extra = "", .var.name = .vname(ui))

assertRxUiMixedOnly(ui, extra = "", .var.name = .vname(ui))

assertRxUiRandomOnIdOnly(ui, extra = "", .var.name = .vname(ui))
}
\arguments{
\item{ui}{Model to check}

\item{extra}{Extra text to append to the error message (like
"for focei")}

\item{.var.name}{[\code{character(1)}]\cr
Name of the checked object to print in assertions. Defaults to
the heuristic implemented in \code{\link[checkmate]{vname}}.}
}
\value{
the rxUi model
}
\description{
Assert properties of the rxUi models
}
\details{
These functions have different types of assertions
\itemize{
\item \code{assertRxUi} -- Make sure this is a proper rxode2 model (if not throw error)
\item \code{assertRxUiSingleEndpoint} -- Make sure the rxode2 model is only
a single endpoint model (if not throw error)
\item \code{assertRxUiTransformNormal} -- Make sure that the model residual
distribution is normal or transformably normal
\item \code{assertRxUiNormal} -- Make sure that the model residual distribution is normal
\item \code{assertRxUiEstimatedResiduals} -- Make sure that the residual error
parameters are estimated (not modeled).
\item \code{assertRxUiPopulationOnly} -- Make sure the model is the population only
model (no mixed effects)
\item \code{assertRxUiMixedOnly} -- Make sure the model is a mixed effect model (not a
population effect, only)
\item \code{assertRxUiPrediction} -- Make sure the model has predictions
\item \code{assertRxUiMuRefOnly} -- Make sure that all the parameters are mu-referenced
\item \code{assertRxUiRandomOnIdOnly} -- Make sure there are only random effects at the ID level
\item \code{assertRxUiIovNoCor} -- Make sure that the IOV model does not have any correlations
\item \code{assertRxUiNoMix} -- Make sure that the model does not have a mixture model inside it
\item \code{assertRxUiNoAutoregressive} -- Make sure the model does not have an
autoregressive residual (ie \code{ar()}); used by estimation methods that do
not support it
}
}
\examples{

\donttest{
one.cmt <- function() {
 ini({
   tka <- 0.45; label("Ka")
   tcl <- log(c(0, 2.7, 100)); label("Cl")
   tv <- 3.45; label("V")
   eta.ka ~ 0.6
   eta.cl ~ 0.3
   eta.v ~ 0.1
   add.sd <- 0.7
 })
 model({
   ka <- exp(tka + eta.ka)
   cl <- exp(tcl + eta.cl)
   v <- exp(tv + eta.v)
   linCmt() ~ add(add.sd)
 })
}

assertRxUi(one.cmt)
# assertRxUi(rnorm) # will fail

assertRxUiSingleEndpoint(one.cmt)
}
}
\seealso{
Other Assertions:
\code{\link[=assertCompartmentExists]{assertCompartmentExists()}},
\code{\link[=assertCompartmentName]{assertCompartmentName()}},
\code{\link[=assertCompartmentNew]{assertCompartmentNew()}},
\code{\link[=assertVariableExists]{assertVariableExists()}},
\code{\link[=assertVariableNew]{assertVariableNew()}},
\code{\link[=testIniDf]{testIniDf()}},
\code{\link[=testRxUnbounded]{testRxUnbounded()}}
}
\author{
Matthew L. Fidler
}
\concept{Assertions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{as.data.table.rxEt}
\alias{as.data.table.rxEt}
\title{Convert an event table to a data.table}
\usage{
\method{as.data.table}{rxEt}(x, keep.rownames = FALSE, ...)
}
\arguments{
\item{x}{An R object.}

\item{keep.rownames}{Default is \code{FALSE}. If \code{TRUE}, adds the input object's names as a separate column named \code{"rn"}. \code{keep.rownames = "id"} names the column \code{"id"} instead. For lists and when calling \code{data.table()}, names from the first named vector are extracted and used as row names, similar to \code{data.frame()} behavior.}

\item{...}{Additional arguments to be passed to or from other methods.}
}
\value{
data.table of event table
}
\description{
Convert an event table to a data.table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxRaw.R
\name{rxOldQsDes}
\alias{rxOldQsDes}
\title{Deserialize a Raw Vector using qs}
\usage{
rxOldQsDes(x)
}
\arguments{
\item{x}{raw vector to deserialize using qs (not on CRAN any more)}
}
\value{
qs::deserialize (if qs is available) of x
}
\description{
Deserialize a Raw Vector using qs
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxSymInv.R
\name{rxSymInvCholCreate}
\alias{rxSymInvCholCreate}
\title{Creates an object for calculating Omega/Omega^-1 and derivatives}
\usage{
rxSymInvCholCreate(
  mat,
  diag.xform = c("sqrt", "log", "identity"),
  create.env = TRUE,
  envir = parent.frame()
)
}
\arguments{
\item{mat}{Initial Omega matrix}

\item{diag.xform}{transformation to diagonal elements of OMEGA. or \code{chol(Omega^-1)}}

\item{create.env}{-- Create an environment to calculate the inverses. (By default TRUE)}

\item{envir}{-- Environment to evaluate function, bu default it is the parent frame.}
}
\value{
A rxSymInv object OR a rxSymInv environment
}
\description{
Creates an object for calculating Omega/Omega^-1 and derivatives
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjointDiscrete.R
\name{.rxAdjointMethodStiff}
\alias{.rxAdjointMethodStiff}
\title{Does an adjoint (\code{s}) solve method require the stiff analytic Jacobian?}
\usage{
.rxAdjointMethodStiff(method)
}
\arguments{
\item{method}{adjoint method name(s) -- a single name, a composite vector
\code{c("dop853s","ros4s")}, or a \code{"dop853s+ros4s"} string.}
}
\value{
\code{TRUE} if any component is a stiff adjoint method.
}
\description{
Does an adjoint (\code{s}) solve method require the stiff analytic Jacobian?
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxModelVars}
\alias{rxModelVars}
\alias{rxModelVarsS3}
\alias{rxModelVarsS3.rxUi}
\alias{rxModelVarsS3.default}
\title{All model variables for a rxode2 object}
\usage{
rxModelVars(obj)

rxModelVarsS3(obj)

\method{rxModelVarsS3}{rxUi}(obj)

\method{rxModelVarsS3}{default}(obj)
}
\arguments{
\item{obj}{rxode2 family of objects}
}
\value{
A list of rxode2 model properties including:
\itemize{
\item \code{params}  a character vector of names of the model parameters
\item \code{lhs} a character vector of the names of the model calculated parameters
\item \code{state} a character vector of the compartments in rxode2 object
\item \code{trans} a named vector of translated model properties
including what type of jacobian is specified, the \code{C} function prefixes,
as well as the \code{C} functions names to be called through the compiled model.
\item \code{md5} a named vector that gives the digest of the model (\code{file_md5}) and the parsed model
(\code{parsed_md5})
\item \code{model}  a named vector giving the input model (\code{model}),
normalized model (no comments and standard syntax for parsing, \code{normModel}),
and interim code that is used to generate the final C file \code{parseModel}
}
}
\description{
Return all the known model variables for a specified rxode2 object
}
\details{
These items are only calculated after compilation; they are
built-into the rxode2 compiled DLL.

To allow extension, an s3 hook is added in the function \code{rxModelVarsS3}.
}
\seealso{
Other Query model information:
\code{\link[=rxDfdy]{rxDfdy()}},
\code{\link[=rxInits]{rxInits()}},
\code{\link[=rxLhs]{rxLhs()}},
\code{\link[=rxParams]{rxParams()}},
\code{\link[=rxState]{rxState()}}
}
\author{
Matthew L. Fidler
}
\concept{Query model information}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjoint.R
\name{.rxAdjointSolve}
\alias{.rxAdjointSolve}
\title{Solve the full-trajectory adjoint sensitivities for a model}
\usage{
.rxAdjointSolve(
  object,
  params,
  events,
  calcSens,
  outTimes,
  adjStates = NULL,
  denseBy = 0.01,
  atol = 1e-10,
  rtol = 1e-10
)
}
\arguments{
\item{object}{model definition accepted by \code{\link[=rxode2]{rxode2()}} (text,
\code{rxode2} object, or function/ui).}

\item{params}{named numeric vector of parameter values.}

\item{events}{event table / data used to define dosing.}

\item{calcSens}{character vector of parameter names to differentiate with
respect to.}

\item{outTimes}{numeric vector of output (observation) times at which the
sensitivities are requested.}

\item{adjStates}{character vector of output states of interest.  Defaults to
all ODE states (full forward-sensitivity parity).}

\item{denseBy}{grid spacing for the forward checkpoint trajectory; smaller
values reduce the covariate-interpolation error.}

\item{atol, rtol}{solver tolerances used for both the forward checkpoint and
the backward sweeps.}
}
\value{
data.frame with a \code{time} column and one
\verb{rx__sens_<state>_BY_<param>__} column per (state, param) pair.
}
\description{
Reconstructs \code{dy_k(t_i)/dp} for every output state \code{k}, parameter \code{p} and
output time \code{t_i} via backward (adjoint) integration, returning the same
\verb{rx__sens_<state>_BY_<param>__} columns that forward sensitivities produce.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxExpandSens_}
\alias{rxExpandSens_}
\title{Expand sensitivity}
\usage{
rxExpandSens_(state, calcSens)
}
\arguments{
\item{state}{is the state to expand}

\item{calcSens}{is the variables to calculate sensitivity}
}
\value{
symengine string for expanded sensitivity
}
\description{
Expand sensitivity
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxOom.R
\name{rxEventTableFile}
\alias{rxEventTableFile}
\title{Create a file-backed event table reference}
\usage{
rxEventTableFile(
  path,
  format = c("auto", "parquet", "csv", "fst", "rds"),
  id = "id"
)
}
\arguments{
\item{path}{path to the file containing the event table}

\item{format}{file format: "auto" (detect from extension), "parquet", "csv", "fst", or "rds"}

\item{id}{name of the subject ID column}
}
\value{
An \code{rxEtFile} object
}
\description{
Create a file-backed event table reference
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{gammap}
\alias{gammap}
\title{Gammap: normalized lower incomplete gamma function}
\usage{
gammap(a, z)
}
\arguments{
\item{a}{The numeric 'a' parameter in the normalized lower
incomplete gamma}

\item{z}{The numeric 'z' parameter in the normalized lower
incomplete gamma}
}
\value{
gammap results
}
\description{
This is the gamma_p from the boost library
}
\details{
The gamma p function is given by:

gammap = lowergamma(a, z)/gamma(a)
}
\examples{

gammap(1, 3)
gammap(1:3, 3)
gammap(1, 1:3)
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxLastCompile}
\alias{rxLastCompile}
\title{Get the last compiled model information as alist}
\usage{
rxLastCompile()
}
\value{
A list contains the following elements:
\itemize{
\item \code{msg} the message for a bad compilation, or NULL if successful.
\item \code{stdout} the standard output from the compilation
\item \code{stderr} the standard error from the compilation
\item \code{c} the code code that was used
}

This list will be returned invisibly, but the function will
also message the contents to the console.
}
\description{
Get the last compiled model information as alist
}
\examples{
\donttest{
rxode2({
  a <- b
})
rxLastCompile()
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/etTran.R
\name{etTrans}
\alias{etTrans}
\title{Event translation for rxode2}
\usage{
etTrans(
  inData,
  obj,
  addCmt = FALSE,
  dropUnits = FALSE,
  allTimeVar = FALSE,
  keepDosingOnly = FALSE,
  combineDvid = NULL,
  keep = character(0),
  addlKeepsCov = FALSE,
  addlDropSs = TRUE,
  ssAtDoseTime = TRUE,
  iCov = NULL
)

etTrans(
  inData,
  obj,
  addCmt = FALSE,
  dropUnits = FALSE,
  allTimeVar = FALSE,
  keepDosingOnly = FALSE,
  combineDvid = NULL,
  keep = character(0),
  addlKeepsCov = FALSE,
  addlDropSs = TRUE,
  ssAtDoseTime = TRUE,
  iCov = NULL
)
}
\arguments{
\item{inData}{Data frame to translate}

\item{obj}{object where model variables can be extracted from}

\item{addCmt}{Add compartment to data frame (default \code{FALSE}).}

\item{dropUnits}{Boolean to drop the units (default \code{FALSE}).}

\item{allTimeVar}{Treat all covariates as if they were time-varying}

\item{keepDosingOnly}{keep the individuals who only have dosing records and any
trailing dosing records after the last observation.}

\item{combineDvid}{is a boolean indicating if rxode2 will use \code{DVID} on observation
records to change the \code{cmt} value; Useful for multiple-endpoint nlmixr models.  By default
this is determined by \code{option("rxode2.combine.dvid")} and if the option has not been set,
this is \code{TRUE}. This typically does not affect rxode2 simulations.}

\item{keep}{This is a named vector of items you want to keep in the final rxode2 dataset.
For added rxode2 event records (if seen), last observation carried forward will be used.}

\item{addlKeepsCov}{This determines if the additional dosing items
repeats the dose only (\code{FALSE}) or keeps the covariates at the
record of the dose (\code{TRUE})}

\item{addlDropSs}{When there are steady state doses with an \code{addl}
specification the steady state flag is dropped with repeated
doses (when \code{TRUE}) or retained (when \code{FALSE})}

\item{ssAtDoseTime}{Boolean that when \code{TRUE} back calculates the
steady concentration at the actual time of dose, otherwise when
\code{FALSE} the doses are shifted}

\item{iCov}{A data frame of individual non-time varying covariates
to combine with the \code{events} dataset.  The \code{iCov} dataset has one
covariate per ID and should match the event table}
}
\value{
Object for solving in rxode2
}
\description{
Event translation for rxode2
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{rxPp}
\alias{rxPp}
\title{Simulate a from a Poisson process}
\usage{
rxPp(
  n,
  lambda,
  gamma = 1,
  prob = NULL,
  t0 = 0,
  tmax = Inf,
  randomOrder = FALSE
)
}
\arguments{
\item{n}{Number of time points to simulate in the Poisson process}

\item{lambda}{Rate of Poisson process}

\item{gamma}{Asymmetry rate of Poisson process.  When gamma=1.0,
this simulates a homogenous Poisson process.  When gamma<1.0,
the Poisson process has more events early, when gamma > 1.0,
the Poisson process has more events late in the process.

When gamma is non-zero, the tmax should not be infinite but indicate
the end of the Poisson process to be simulated.  In most
pharamcometric cases, this will be the end of the study.
Internally this uses a rate of:

l(t) = lambda\emph{gamma}(t/tmax)^(gamma-1)}

\item{prob}{When specified, this is a probability function with
one argument, time, that gives the probability that a Poisson
time t is accepted as a rejection time.}

\item{t0}{the starting time of the Poisson process}

\item{tmax}{the maximum time of the Poisson process}

\item{randomOrder}{when \code{TRUE} randomize the order of the Poisson
events.  By default (\code{FALSE}) it returns the Poisson process is
in order of how the events occurred.}
}
\value{
This returns a vector of the Poisson process times; If the dropout is >=
tmax, then all the rest of the times are = tmax to indicate the
dropout is equal to or after tmax.
}
\description{
Simulate a from a Poisson process
}
\examples{

## Sample homogenous Poisson process of rate 1/10
rxPp(10, 1 / 10)

## Sample inhomogenous Poisson rate of 1/10

rxPp(10, 1 / 10, gamma = 2, tmax = 100)

## Typically the Poisson process times are in a sequential order,
## using randomOrder gives the Poisson process in random order

rxPp(10, 1 / 10, gamma = 2, tmax = 10, randomOrder = TRUE)

## This uses an arbitrary function to sample a non-homogenous Poisson process

rxPp(10, 1 / 10, prob = function(x) {
  1/(1+abs(x))
})

}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piping-model.R
\name{rxSetPipingAuto}
\alias{rxSetPipingAuto}
\title{Set the variables for the model piping automatic covarite selection}
\usage{
rxSetPipingAuto(
  thetamodelVars = rex::rex(or("tv", "t", "pop", "POP", "Pop", "TV", "T", "cov", "err",
    "eff")),
  covariateExceptions = rex::rex(start, or("wt", "sex", "crcl", "kout"), end),
  etaParts = c("eta", "ETA", "Eta", "ppv", "PPV", "Ppv", "iiv", "Iiv", "bsv", "Bsv",
    "BSV", "bpv", "Bpv", "BPV", "psv", "PSV", "Psv")
)
}
\arguments{
\item{thetamodelVars}{This is the prefixes for the theta model
variables in a regular expression}

\item{covariateExceptions}{This is a regular expression of
covariates that should always be covariates}

\item{etaParts}{This is the list of eta prefixes/post-fixes that
identify a variable as a between subject variability}
}
\value{
Nothing, called for side effects
}
\description{
Set the variables for the model piping automatic covarite selection
}
\details{
This is called once at startup to set the defaults, though you can
change this if you wish so that piping can work differently for
your individual setup
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{.rxSEeqUsr}
\alias{.rxSEeqUsr}
\title{Return the equivalents symengine user functions from C}
\usage{
.rxSEeqUsr()
}
\value{
equivalent symengine user functions
}
\description{
Return the equivalents symengine user functions from C
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{etRbind}
\alias{etRbind}
\alias{rbind.rxEt}
\title{Combining event tables}
\usage{
etRbind(
  ...,
  samples = c("use", "clear"),
  waitII = c("smart", "+ii"),
  id = c("merge", "unique")
)

\method{rbind}{rxEt}(..., deparse.level = 1)
}
\arguments{
\item{...}{The event tables and optionally time between event
tables, called waiting times in this help document.}

\item{samples}{How to handle samples when repeating an event
table.  The options are:
\itemize{
\item \code{"clear"} Clear sampling records before combining the datasets
\item \code{"use"} Use the sampling records when combining the datasets
}}

\item{waitII}{This determines how waiting times between events are
handled. The options are:
\itemize{
\item \code{"smart"} This "smart" handling of waiting times is the
default option.  In this case, if the waiting time is above the
last observed inter-dose interval in the first combined event
table, then the actual time between doses is given by the wait
time.  If it is smaller than the last observed inter-dose
interval, the time between event tables is given by the inter-dose
interval + the waiting time between event tables.
\item \code{"+ii"} In this case, the wait time is added to the
inter-dose interval no matter the length of the wait time or
inter-dose interval
}}

\item{id}{This is how rbind will handle ids.  There are two different types of options:
\itemize{
\item \code{merge} with \code{id="merge"}, the ids are merged together,
overlapping ids would be merged into a single event table.
\item \code{unique} with \code{id="unique"}, the ids will be renumbered
so that the ids in all the event tables are not overlapping.
}}

\item{deparse.level}{The \code{deparse.level} of a traditional
\code{rbind} is ignored.}
}
\value{
An event table
}
\description{
Combining event tables
}
\examples{

\dontrun{

library(rxode2)
library(units)

# Model from rxode2 tutorial
# Using a nlmixr2 style function

mod1 <-function(){
  ini({
    KA <- 2.94E-01
    CL <- 1.86E+01
    V2 <- 4.02E+01
    Q <- 1.05E+01
    V3 <- 2.97E+02
    Kin <- 1
    Kout <- 1
    EC50 <- 200
  })
 model({
    C2 <- centr/V2
    C3 <- peri/V3
    d/dt(depot) <- -KA*depot
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
    d/dt(peri)  <-                    Q*C2 - Q*C3
    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
 })
}

## These are making the more complex regimens of the rxode2 tutorial

## bid for 5 days
bid <- et(timeUnits="hr") |>
       et(amt=10000,ii=12,until=set_units(5, "days"))

## qd for 5 days
qd <- et(timeUnits="hr") |>
      et(amt=20000,ii=24,until=set_units(5, "days"))

## bid for 5 days followed by qd for 5 days

et <- seq(bid,qd) |>
      et(seq(0,11*24,length.out=100))

bidQd <- rxSolve(mod1, et)

plot(bidQd, C2)


## Now Infusion for 5 days followed by oral for 5 days

##  note you can dose to a named compartment instead of using the compartment number
infusion <- et(timeUnits = "hr") |>
      et(amt=10000, rate=5000, ii=24, until=set_units(5, "days"), cmt="centr")


qd <- et(timeUnits = "hr") |>
  et(amt=10000, ii=24, until=set_units(5, "days"), cmt="depot")

et <- seq(infusion,qd)

infusionQd <- rxSolve(mod1, et)

plot(infusionQd, C2)

## 2wk-on, 1wk-off

qd <- et(timeUnits = "hr") |>
      et(amt=10000, ii=24, until=set_units(2, "weeks"), cmt="depot")

et <- seq(qd, set_units(1,"weeks"), qd) |>
     add.sampling(set_units(seq(0, 5.5,by=0.005),weeks))

wkOnOff <- rxSolve(mod1, et)

plot(wkOnOff, C2)

## You can also repeat the cycle easily with the rep function

qd <-et(timeUnits = "hr") |>
     et(amt=10000, ii=24, until=set_units(2, "weeks"), cmt="depot")

et <- etRep(qd, times=4, wait=set_units(1,"weeks")) |>
      add.sampling(set_units(seq(0, 12.5,by=0.005),weeks))

repCycle4 <- rxSolve(mod1, et)

plot(repCycle4, C2)

}
}
\references{
Wang W, Hallow K, James D (2015). "A Tutorial on rxode2: Simulating
Differential Equation Pharmacometric Models in R." CPT:
Pharmacometrics and Systems Pharmacology, 5(1), 3-10. ISSN 2163-8306
}
\seealso{
\code{\link{eventTable}}, \code{\link{add.sampling}},
\code{\link{add.dosing}}, \code{\link{et}},
\code{\link{etRep}}, \code{\link{etRbind}},
\code{\link{rxode2}}
}
\author{
Matthew L Fidler

Matthew L Fidler, Wenping Wang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-rename.R
\name{rxRename}
\alias{rxRename}
\alias{.rxRename}
\alias{rename.rxUi}
\alias{rename.function}
\alias{rxRename.rxUi}
\alias{rxRename.function}
\alias{rxRename.default}
\title{Rename items inside of a \code{rxode2} ui model}
\usage{
rxRename(.data, ..., envir = parent.frame())

.rxRename(.data, ..., envir = parent.frame())

rename.rxUi(.data, ...)

rename.function(.data, ...)

\method{rxRename}{rxUi}(.data, ...)

\method{rxRename}{`function`}(.data, ...)

\method{rxRename}{default}(.data, ...)
}
\arguments{
\item{.data}{rxode2 ui function, named data to be consistent with \code{dplyr::rename()}}

\item{...}{rename items}

\item{envir}{Environment for evaluation}
}
\value{
New model with items renamed
}
\description{
\code{rxRename()} changes the names of individual variables, lhs, and ode states using
\code{new_name = old_name} syntax
}
\details{
This is similar to \code{dplyr}'s \code{rename()} function.  When \code{dplyr} is
loaded, the \code{s3} methods work for the ui objects.

Note that the \code{.rxRename()} is the internal function that is called
when renaming and is likely not what you need to call unless you
are writing your own extension of the function
}
\examples{

ocmt <- function() {
  ini({
    tka <- exp(0.45) # Ka
    tcl <- exp(1) # Cl
    ## This works with interactive models
    ## You may also label the preceding line with label("label text")
    tv <- exp(3.45) # log V
    ## the label("Label name") works with all models
    add.sd <- 0.7
  })
  model({
    ka <- tka
    cl <- tcl
    v <- tv
    d/dt(depot) = -ka * depot
    d/dt(center) = ka * depot - cl / v * center
    cp = center / v
    cp ~ add(add.sd)
  })
}

ocmt |> rxRename(cpParent=cp)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxChain2}
\alias{rxChain2}
\alias{rxChain2.default}
\alias{rxChain2.EventTable}
\title{Second command in chaining commands}
\usage{
rxChain2(obj, solvedObject)

\method{rxChain2}{default}(obj, solvedObject)

\method{rxChain2}{EventTable}(obj, solvedObject)
}
\arguments{
\item{obj}{the object being added/chained/piped to the solved object}

\item{solvedObject}{the solved object}
}
\value{
chained operation
}
\description{
This is s3 method is called internally with \code{+} and \verb{\\\%>\\\%} operators.
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{.rxSetSeed}
\alias{.rxSetSeed}
\title{Set the random seed for the session}
\usage{
.rxSetSeed(seed)
}
\arguments{
\item{seed}{Seed from \code{.rxGetSeed()}}
}
\value{
Nothing, called for its side effects
}
\description{
This sets the random seed for the session.  This is used internally
to set the random seed for the session.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{phi}
\alias{phi}
\title{Cumulative distribution of standard normal}
\usage{
phi(q)
}
\arguments{
\item{q}{vector of quantiles}
}
\value{
cumulative distribution of standard normal distribution
}
\description{
Cumulative distribution of standard normal
}
\examples{

# phi is equivalent to pnorm(x)
phi(3)

# See
pnorm(3)

# This is provided for NONMEM-like compatibility in rxode2 models
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.rxode2}
\alias{summary.rxode2}
\title{Print expanded information about the rxode2 object.}
\usage{
\method{summary}{rxode2}(object, ...)
}
\arguments{
\item{object}{rxode2 object}

\item{...}{Ignored parameters}
}
\value{
object is returned
}
\description{
This prints the expanded information about the rxode2 object.
}
\author{
Matthew L.Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dsl.R
\name{rxSumProdModel}
\alias{rxSumProdModel}
\title{Recast model in terms of sum/prod}
\usage{
rxSumProdModel(model, expand = FALSE, sum = TRUE, prod = TRUE)
}
\arguments{
\item{model}{rxode2 model}

\item{expand}{Boolean indicating if the expression is expanded.}

\item{sum}{Use sum(...)}

\item{prod}{Use prod(...)}
}
\value{
model string with prod(.) and sum(.) for all these
operations.
}
\description{
Recast model in terms of sum/prod
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxMemoryEstimate.R
\name{rxMemoryEstimate}
\alias{rxMemoryEstimate}
\title{Estimate memory required by rxSolve() for a given dataset and model}
\usage{
rxMemoryEstimate(
  dat,
  model = NULL,
  control = NULL,
  neq = 1L,
  stateSize = neq,
  nlhs = 0L,
  npars = neq,
  neta = 0L,
  neps = 0L,
  ncov = 0L,
  nsim = 1L,
  cores = 1L,
  nMtime = 0L,
  extraCmt = 0L,
  linB = FALSE,
  nLlik = 0L,
  nIndSim = NULL,
  numLinSens = 0L,
  numLin = 0L
)
}
\arguments{
\item{dat}{A \code{\link{rxMemSummary}}, a data.frame with
\code{nobs}/\code{ndoses} columns, or a full event-table
data.frame with an \code{evid} column. This can also be a
serialized solve state file path, a solve state bundle list
containing an \code{events} element, or an \code{rxSolve}
object (using its stored solve events).}

\item{model}{Optional rxode2 model object.  When supplied,
\code{neq}, \code{nlhs}, \code{npars}, \code{extraCmt},
\code{linB}, \code{nMtime}, \code{nLlik}, and \code{nIndSim} are
extracted automatically.}

\item{control}{Optional \code{\link{rxControl}} object.  When
supplied, \code{cores}, \code{nsim}, \code{neta} (from
\code{omega}), \code{neps} (from \code{sigma}), and \code{nLlik}
(adjusted by \code{nLlikAlloc}) are overridden automatically.}

\item{neq}{Number of ODE states.}

\item{stateSize}{Effective \code{state.size()} seen by the solver.
Equals \code{neq} for pure ODE models; may differ for linCmt-only
models.  Defaults to \code{neq}.}

\item{nlhs}{Number of LHS (calculated) output variables.}

\item{npars}{Number of model parameters (drives \code{gpars} size).}

\item{neta}{Number of random effects (etas).}

\item{neps}{Number of residual-error levels (epsilons).}

\item{ncov}{Number of time-varying covariates.}

\item{nsim}{Number of simulations.}

\item{cores}{Number of parallel OMP threads.}

\item{nMtime}{Number of model measurement times.}

\item{extraCmt}{Extra compartments (0, 1 = depot, 2 =
depot+central).}

\item{linB}{\code{TRUE}/\code{1} if using a linear-compartment
model.}

\item{nLlik}{Number of log-likelihood terms (FOCEi use).}

\item{nIndSim}{Per-individual simulation count.  Defaults to
\code{neta + neps} when not supplied explicitly.}

\item{numLinSens}{Number of linear sensitivity parameters (FOCEi +
linCmt).}

\item{numLin}{Number of linear compartment terms (FOCEi + linCmt).}
}
\value{
A named list of class \code{"rxMemoryEstimate"} whose
elements are raw byte counts plus \code{outputData},
\code{ramBytes}, \code{freeRamBytes}, \code{total},
\code{sizeofInd}, and \code{rxLlikSaveSize}.
}
\description{
Accepts either a pre-summarized per-ID table (an \code{\link{rxMemSummary}}
or any data.frame with \code{nobs} and \code{ndoses} columns) or a full
event-table data.frame with an \code{evid} column.  Model dimensions can
be supplied via a compiled rxode2 model object or overridden individually.
}
\details{
The byte counts are computed by \code{rxMemoryComponents_()} which calls
the same \code{rxFillMemLayout()} used by the real allocator, so any
change to the allocation formulas propagates here automatically.
}
\examples{

\donttest{

mod <- rxode2::rxode2({
  d/dt(depot)  <- -ka * depot
  d/dt(center) <- ka * depot - cl / v * center
  cp           <- center / v
})

ev <- rxode2::et(amt = 100, ii = 24, until = 168) |>
  rxode2::et(seq(0, 168, by = 1))

# Basic estimate from event table and model
rxMemoryEstimate(as.data.frame(ev), model = mod)

# With rxControl: population simulation with omega and 4 cores
ctrl <- rxode2::rxControl(
  cores = 4L,
  omega = lotri::lotri(eta.ka ~ 0.09, eta.cl ~ 0.04)
)
rxMemoryEstimate(as.data.frame(ev), model = mod, control = ctrl)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{etRep}
\alias{etRep}
\alias{rep.rxEt}
\title{Repeat an rxode2 event table}
\usage{
etRep(
  x,
  times = 1,
  length.out = NA,
  each = NA,
  n = NULL,
  wait = 0,
  id = integer(0),
  samples = c("clear", "use"),
  waitII = c("smart", "+ii"),
  ii = 24
)

\method{rep}{rxEt}(x, ...)
}
\arguments{
\item{x}{An rxode2 event table}

\item{times}{Number of times to repeat the event table}

\item{length.out}{Invalid with rxode2 event tables, will throw an
error if used.}

\item{each}{Invalid with rxode2 event tables, will throw an error
if used.}

\item{n}{The number of times to repeat the event table.  Overrides
\code{times}.}

\item{wait}{Waiting time between each repeated event table.  By
default there is no waiting, or wait=0}

\item{id}{A integer vector of ids to add or remove from the event
table.  If the event table is identical for each ID, then you
may expand it to include all the ids in this vector.  All the
negative ids in this vector will be removed.}

\item{samples}{How to handle samples when repeating an event
table.  The options are:
\itemize{
\item \code{"clear"} Clear sampling records before combining the datasets
\item \code{"use"} Use the sampling records when combining the datasets
}}

\item{waitII}{This determines how waiting times between events are
handled. The options are:
\itemize{
\item \code{"smart"} This "smart" handling of waiting times is the
default option.  In this case, if the waiting time is above the
last observed inter-dose interval in the first combined event
table, then the actual time between doses is given by the wait
time.  If it is smaller than the last observed inter-dose
interval, the time between event tables is given by the inter-dose
interval + the waiting time between event tables.
\item \code{"+ii"} In this case, the wait time is added to the
inter-dose interval no matter the length of the wait time or
inter-dose interval
}}

\item{ii}{When specifying a dose, this is the inter-dose interval
for \code{ss}, \code{addl} and \code{until} options (described below).}

\item{...}{Times or event tables.  They can also be one of the named arguments below.}
}
\value{
An event table
}
\description{
Repeat an rxode2 event table
}
\examples{

\dontrun{

library(rxode2)
library(units)

# Model from rxode2 tutorial
# Using a nlmixr2 style function

mod1 <-function(){
  ini({
    KA <- 2.94E-01
    CL <- 1.86E+01
    V2 <- 4.02E+01
    Q <- 1.05E+01
    V3 <- 2.97E+02
    Kin <- 1
    Kout <- 1
    EC50 <- 200
  })
 model({
    C2 <- centr/V2
    C3 <- peri/V3
    d/dt(depot) <- -KA*depot
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
    d/dt(peri)  <-                    Q*C2 - Q*C3
    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
 })
}

## These are making the more complex regimens of the rxode2 tutorial

## bid for 5 days
bid <- et(timeUnits="hr") |>
       et(amt=10000,ii=12,until=set_units(5, "days"))

## qd for 5 days
qd <- et(timeUnits="hr") |>
      et(amt=20000,ii=24,until=set_units(5, "days"))

## bid for 5 days followed by qd for 5 days

et <- seq(bid,qd) |>
      et(seq(0,11*24,length.out=100))

bidQd <- rxSolve(mod1, et)

plot(bidQd, C2)


## Now Infusion for 5 days followed by oral for 5 days

##  note you can dose to a named compartment instead of using the compartment number
infusion <- et(timeUnits = "hr") |>
      et(amt=10000, rate=5000, ii=24, until=set_units(5, "days"), cmt="centr")


qd <- et(timeUnits = "hr") |>
  et(amt=10000, ii=24, until=set_units(5, "days"), cmt="depot")

et <- seq(infusion,qd)

infusionQd <- rxSolve(mod1, et)

plot(infusionQd, C2)

## 2wk-on, 1wk-off

qd <- et(timeUnits = "hr") |>
      et(amt=10000, ii=24, until=set_units(2, "weeks"), cmt="depot")

et <- seq(qd, set_units(1,"weeks"), qd) |>
     add.sampling(set_units(seq(0, 5.5,by=0.005),weeks))

wkOnOff <- rxSolve(mod1, et)

plot(wkOnOff, C2)

## You can also repeat the cycle easily with the rep function

qd <-et(timeUnits = "hr") |>
     et(amt=10000, ii=24, until=set_units(2, "weeks"), cmt="depot")

et <- etRep(qd, times=4, wait=set_units(1,"weeks")) |>
      add.sampling(set_units(seq(0, 12.5,by=0.005),weeks))

repCycle4 <- rxSolve(mod1, et)

plot(repCycle4, C2)

}
}
\references{
Wang W, Hallow K, James D (2015). "A Tutorial on rxode2: Simulating
Differential Equation Pharmacometric Models in R." CPT:
Pharmacometrics and Systems Pharmacology, 5(1), 3-10. ISSN 2163-8306
}
\seealso{
\code{\link{eventTable}}, \code{\link{add.sampling}},
\code{\link{add.dosing}}, \code{\link{et}},
\code{\link{etRep}}, \code{\link{etRbind}},
\code{\link{rxode2}}
}
\author{
Matthew L Fidler, Wenping Wang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assert.R
\name{rxHasAr}
\alias{rxHasAr}
\title{Test if a model uses an autoregressive (\code{ar()}) residual}
\usage{
rxHasAr(ui)
}
\arguments{
\item{ui}{rxode2 user interface model}
}
\value{
logical, \code{TRUE} if any endpoint carries an \code{ar()} correlation
(either an estimated correlation or a modeled/literal one)
}
\description{
Test if a model uses an autoregressive (\code{ar()}) residual
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{gammapInv}
\alias{gammapInv}
\alias{gammapInva}
\title{gammapInv and gammapInva:  Inverses of normalized gammap function}
\usage{
gammapInv(a, p)

gammapInva(x, p)
}
\arguments{
\item{a}{The numeric 'a' parameter in the upper
incomplete gamma}

\item{p}{The numeric 'p' parameter in the upper incomplete gamma}

\item{x}{The numeric 'x' parameter in the upper incomplete gamma}
}
\value{
inverse gammap results
}
\description{
gammapInv and gammapInva:  Inverses of normalized gammap function
}
\details{
With the equation:

p = gammap(a, x)

The 'gammapInv' function returns a value 'x' that satisfies the
equation above

The 'gammapInva' function returns a value 'q' that satisfies the
equation above

NOTE: gammapInva is slow
}
\examples{

gammapInv(1:3, 0.5)

gammapInv(1, 1:3 / 3.1)

gammapInv(1:3, 1:3 / 3.1)

gammapInva(1:3, 1:3 / 3.1)
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{.cbindOme}
\alias{.cbindOme}
\title{cbind Ome}
\usage{
.cbindOme(et, mat, n)
}
\arguments{
\item{et}{The theta data frame}

\item{mat}{The full matrix simulation from omegas}

\item{n}{number of subject simulated}
}
\value{
data frame with et combined with simulated omega matrix values
}
\description{
cbind Ome
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{.pipeRx}
\alias{.pipeRx}
\alias{.pipeInits}
\alias{.pipeEvents}
\alias{.pipeParams}
\alias{.pipeKeep}
\alias{.pipeThetaMat}
\alias{.pipeOmega}
\alias{.pipeSigma}
\alias{.pipeDfObs}
\alias{.pipeDfSub}
\alias{.pipeNSub}
\alias{.pipeNStud}
\title{Assign in the rxode2 pipeline}
\usage{
.pipeRx(obj)

.pipeInits(obj)

.pipeEvents(obj)

.pipeParams(obj)

.pipeKeep(obj)

.pipeThetaMat(obj)

.pipeOmega(obj)

.pipeSigma(obj)

.pipeDfObs(obj)

.pipeDfSub(obj)

.pipeNSub(obj)

.pipeNStud(obj)
}
\arguments{
\item{obj}{Object to assign.  If NA return the value.}
}
\value{
The pipeline object (invisibly)
}
\description{
Assign in the rxode2 pipeline
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxSetupScale}
\alias{rxSetupScale}
\title{Setup the initial conditions.}
\usage{
rxSetupScale(obj, scale = NULL, extraArgs = NULL)
}
\arguments{
\item{obj}{rxode2 object}

\item{scale}{A numeric vector scales}

\item{extraArgs}{A list of extra args to parse for initial conditions.}
}
\value{
setup scale for changing compartment values
}
\description{
Setup the initial conditions.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxUdfUiExpr}
\alias{rxUdfUiExpr}
\title{Give the expression as a compressed model or expression}
\usage{
rxUdfUiExpr(expr, env = parent.frame())
}
\arguments{
\item{expr}{the R language expression to evaluate}

\item{env}{the environment in which to evaluate the expression}
}
\value{
expression
}
\description{
Here it means that it evaluates to a variable or a number.
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{uppergamma}
\alias{uppergamma}
\title{uppergamma:  upper incomplete gamma function}
\usage{
uppergamma(a, z)
}
\arguments{
\item{a}{The numeric 'a' parameter in the upper
incomplete gamma}

\item{z}{The numeric 'z' parameter in the upper
incomplete gamma}
}
\value{
uppergamma results
}
\description{
This is the tgamma from the boost library
}
\details{
The uppergamma function is given by:

\eqn{uppergamma(a, z) = \int_{z}^{\infty}t^{a-1}\cdot e^{-t} dt}
}
\examples{

uppergamma(1, 3)

uppergamma(1:3, 3)

uppergamma(1, 1:3)
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxOom.R
\name{head.rxSolveOom}
\alias{head.rxSolveOom}
\title{First rows of an out-of-memory solved object}
\usage{
\method{head}{rxSolveOom}(x, n = 6L, ...)
}
\arguments{
\item{x}{An \code{rxSolveOom} object.}

\item{n}{Number of rows to return.}

\item{...}{Ignored.}
}
\value{
A \code{data.frame} with the first \code{n} rows.
}
\description{
Reads only the first \code{n} rows from the parquet (or rds) chunks,
preferring DuckDB (which pushes the row limit into the parquet reader) and
falling back to arrow / rds.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genShinyApp.template.R
\name{genShinyApp.template}
\alias{genShinyApp.template}
\alias{write.template.server}
\alias{write.template.ui}
\title{Generate an example (template) of a dosing regimen shiny app}
\usage{
genShinyApp.template(
  appDir = "shinyExample",
  verbose = TRUE,
  ODE.config = list(ode = "model", params = c(KA = 0.294), inits = c(eff = 1), method =
    "lsoda", atol = 1e-08, rtol = 1e-06)
)

write.template.server(appDir)

write.template.ui(appDir, statevars)
}
\arguments{
\item{appDir}{a string with a directory where to store the shiny
app, by default is \code{"shinyExample"}. The directory
\code{appDir} will be created if it does not exist.}

\item{verbose}{logical specifying whether to write messages as the
shiny app is generated. Defaults to \code{TRUE}.}

\item{ODE.config}{model name compiled and list of parameters sent to \code{\link[=rxSolve]{rxSolve()}}.}

\item{statevars}{List of statevars passed to to the \code{\link[=write.template.ui]{write.template.ui()}} function.  This usually isn't called directly.

A PK/PD model is defined using \code{\link[=rxode2]{rxode2()}}, and
a set of parameters and initial values are defined.  Then
the appropriate R scripts for the shiny's user interface \code{ui.R}
and the server logic \code{server.R} are created in the
directory \code{appDir}.

The function evaluates the following PK/PD model by default:
\preformatted{
    C2 = centr/V2;
    C3 = peri/V3;
    d/dt(depot) =-KA*depot;
    d/dt(centr) = KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  =                    Q*C2 - Q*C3;
    d/dt(eff)  = Kin - Kout*(1-C2/(EC50+C2))*eff;
}

This can be changed by the \code{ODE.config} parameter.

To launch the shiny app, simply issue the \code{runApp(appDir)}
R command.}
}
\value{
None, these functions are used for their side effects.
}
\description{
Create a complete shiny application for exploring dosing regimens
given a (hardcoded) PK/PD model.
}
\note{
These functions create a simple, but working example of a
dosing regimen simulation web application. Users may want to
modify the code to experiment creating shiny applications for
their specific \code{rxode2} models.
}
\examples{
\donttest{
# create in a temp dir so nothing is left in the working directory
.appDir <- tempfile("myapp")
on.exit(unlink(.appDir, recursive = TRUE, force = TRUE))
# create the shiny app example (template)
genShinyApp.template(appDir = .appDir)
# run the shiny app
if (requireNamespace("shiny", quietly=TRUE)) {
  library(shiny)
  # runApp(.appDir) # Won't launch in environments without browsers
}
}
}
\seealso{
\code{\link[=rxode2]{rxode2()}},\code{\link[=eventTable]{eventTable()}}, and the package \pkg{shiny} (\url{https://shiny.posit.co}).
}
\concept{PK/PD}
\concept{pharmacometrics}
\keyword{nonlinear}
\keyword{simulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxGetSerialType_}
\alias{rxGetSerialType_}
\title{Get the serialization type from raw vector}
\usage{
rxGetSerialType_(raw)
}
\arguments{
\item{raw}{A raw vector}
}
\value{
a string indicating the serialization type:
"qs2", "qdata", "qs", "bzip2", "xz",  "base", or "unknown"
}
\description{
Get the serialization type from raw vector
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{erf}
\alias{erf}
\title{Error function}
\usage{
erf(x)
}
\arguments{
\item{x}{vector or real values}
}
\value{
erf of x
}
\description{
Error function
}
\examples{
erf(1.0)
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.rxWithOptions}
\alias{.rxWithOptions}
\title{Temporarily set options then restore them while running code}
\usage{
.rxWithOptions(ops, code)
}
\arguments{
\item{ops}{list of options that will be temporarily set for the
\code{code}}

\item{code}{The code to run during the sink}
}
\value{
value of code
}
\description{
Temporarily set options then restore them while running code
}
\examples{

.rxWithOptions(list(digits = 21), {
  print(pi)
})

print(pi)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symengine.R
\name{rxParseErr}
\alias{rxParseErr}
\title{Prepare Error function for inclusion in rxode2}
\usage{
rxParseErr(
  x,
  baseTheta,
  ret = "rx_r_",
  init = NULL,
  addProp = c("combined2", "combined1")
)
}
\arguments{
\item{x}{error function}

\item{baseTheta}{Base theta to start numbering add(.) and prop(.) from.}

\item{ret}{Internal return type.  Should not be changed by the user...}

\item{init}{Initialization vector}
}
\value{
rxode2 transformed text
}
\description{
Prepare Error function for inclusion in rxode2
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/register-s3.R
\name{.s3register}
\alias{.s3register}
\title{Register a method for a suggested dependency}
\usage{
.s3register(generic, class, method = NULL)
}
\arguments{
\item{generic}{Name of the generic in the form \code{pkg::generic}.}

\item{class}{Name of the class}

\item{method}{Optionally, the implementation of the method. By default,
this will be found by looking for a function called \code{generic.class}
in the package environment.

Note that providing \code{method} can be dangerous if you use
devtools. When the namespace of the method is reloaded by
\code{devtools::load_all()}, the function will keep inheriting from
the old namespace. This might cause crashes because of dangling
\code{.Call()} pointers.}
}
\value{
nothing; called for side effects
}
\description{
Generally, the recommend way to register an S3 method is to use the
\code{S3Method()} namespace directive (often generated automatically by the
\verb{@export} roxygen2 tag). However, this technique requires that the generic
be in an imported package, and sometimes you want to suggest a package,
and only provide a method when that package is loaded. \code{s3_register()}
can be called from your package's \code{.onLoad()} to dynamically register
a method only if the generic's package is loaded. (To avoid taking a
dependency on vctrs for this one function, please feel free to copy
and paste the function source into your own package.)
}
\details{
For R 3.5.0 and later, \code{s3_register()} is also useful when demonstrating
class creation in a vignette, since method lookup no longer always involves
the lexical scope. For R 3.6.0 and later, you can achieve a similar effect
by using "delayed method registration", i.e. placing the following in your
\code{NAMESPACE} file:

\if{html}{\out{<div class="sourceCode">}}\preformatted{if (getRversion() >= "3.6.0") \{
  S3method(package::generic, class)
\}
}\if{html}{\out{</div>}}
}
\examples{
# A typical use case is to dynamically register tibble/pillar methods
# for your class. That way you avoid creating a hard dependency on packages
# that are not essential, while still providing finer control over
# printing when they are used.

.onLoad <- function(...) {
  .s3Register("pillar::pillar_shaft", "vctrs_vctr")
  .s3Register("tibble::type_sum", "vctrs_vctr")
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fixef.R
\name{coef.rxUi}
\alias{coef.rxUi}
\alias{coef.function}
\title{Extract the initial coefficients from an rxode2 model (rxUi)}
\usage{
\method{coef}{rxUi}(
  object,
  level = c("theta", "fixed", "omega", "random", "all", "both"),
  ...
)

\method{coef}{`function`}(
  object,
  level = c("theta", "fixed", "omega", "random", "all", "both"),
  ...
)
}
\arguments{
\item{object}{rxode2 ui model or a function that evaluates to one}

\item{level}{which coefficients to return: \code{"theta"}/\code{"fixed"} (the default)
returns the named fixed-effect estimates; \code{"omega"}/\code{"random"} returns the
random-effect variability matrix (or \code{NULL} when the model has no random
effects); \code{"all"}/\code{"both"} returns a list with \code{theta} and \code{omega}.}

\item{...}{other parameters (currently ignored)}
}
\value{
depends on \code{level}: a named numeric vector of fixed effects, the
\code{omega} matrix (or \code{NULL}), or a list with both \code{theta} and \code{omega}.
}
\description{
For an rxode2 model specification (\code{rxUi}) there are no realized random
effects (etas), so \code{coef()} returns the fixed-effect (\code{theta}) estimates by
default.  The random-effect variability matrix (\code{omega}) can be requested
with \code{level="omega"}, or both can be returned together with \code{level="all"}.
}
\examples{
one.compartment <- function() {
  ini({
    tka <- 0.45
    tcl <- 1
    tv <- 3.45
    eta.ka ~ 0.6
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl)
    v <- exp(tv)
    d/dt(depot) <- -ka * depot
    d/dt(center) <- ka * depot - cl / v * center
    cp <- center / v
  })
}

ui <- one.compartment()
coef(ui)
coef(ui, level="omega")
coef(ui, level="all")
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjointDiscrete.R
\name{.rxAdjointExpand}
\alias{.rxAdjointExpand}
\title{Build the adjoint-expansion model text for an in-engine discrete-adjoint solve}
\usage{
.rxAdjointExpand(object, calcSens, stiff = FALSE)
}
\arguments{
\item{object}{model definition accepted by \code{\link[=rxode2]{rxode2()}}.}

\item{calcSens}{character vector of parameters to differentiate w.r.t.}
}
\value{
list with \code{text} (expanded model text), \code{st}, \code{ns}, \code{np},
\code{calcSens}, and lhs layout \code{fxOff} (=0), \code{fpOff} (=ns*ns), \code{nlhsAdj}.
}
\description{
Exposes the state Jacobian \code{F_X = dF/dy} and parameter forcing
\code{F_p = dF/dtheta} as rxode2 lhs assignments spliced onto the base ODE, so the
C++ backward sweep can evaluate them at any stage state via \code{calc_lhs}.  The
lhs are emitted in a fixed order: all \code{F_X} entries row-major
(\code{rx__adjFX_i_j__} at index \code{i*ns + j}), then \code{F_p} (\code{rx__adjFP_i_p__} at
\code{ns*ns + i*np + p}).
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{rxFunParse}
\alias{rxFunParse}
\alias{rxRmFunParse}
\title{Add user function to rxode2}
\usage{
rxFunParse(name, args, cCode)

rxRmFunParse(name)
}
\arguments{
\item{name}{This gives the name of the user function}

\item{args}{This gives the arguments of the user function}

\item{cCode}{This is the C-code for the new function}
}
\value{
nothing
}
\description{
This adds a user function to rxode2 that can be called.  If needed,
these functions can be differentiated by numerical differences or
by adding the derivatives to rxode2's internal derivative table
with rxode2's \code{rxD} function
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-getControl.R
\name{rxRemoveControl}
\alias{rxRemoveControl}
\title{rxRemoveControl options for UI object}
\usage{
rxRemoveControl(ui)
}
\arguments{
\item{ui}{rxode2 ui object}
}
\value{
Nothing, called for side effects
}
\description{
rxRemoveControl options for UI object
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxReload}
\alias{rxReload}
\title{Reload rxode2 DLL}
\usage{
rxReload()
}
\value{
boolean of if the object is reloaded
}
\description{
Can be useful for debugging
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxSetSilentErr}
\alias{rxSetSilentErr}
\title{Silence some of rxode2's C/C++ messages}
\usage{
rxSetSilentErr(silent)
}
\arguments{
\item{silent}{can be 0L "noisy"  or 1L "silent"}
}
\value{
TRUE; called for side effects
}
\description{
Silence some of rxode2's C/C++ messages
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{rxgamma}
\alias{rxgamma}
\title{Simulate gamma variable from threefry generator}
\usage{
rxgamma(shape, rate = 1, n = 1L, ncores = 1L)
}
\arguments{
\item{shape}{The shape of the gamma random variable}

\item{rate}{an alternative way to specify the scale.}

\item{n}{number of observations. If \code{length(n) > 1}, the length
    is taken to be the number required.}

\item{ncores}{Number of cores for the simulation

\code{rxnorm} simulates using the threefry sitmo generator.

\code{rxnormV} used to simulate with the vandercorput simulator, but
since it didn't satisfy the normal properties it was changed to simple be
an alias of \code{rxnorm}. It is no longer supported in \code{rxode2({})} blocks}
}
\value{
gamma random deviates
}
\description{
Care should be taken with this method not to encounter the
birthday problem, described
\url{https://www.johndcook.com/blog/2016/01/29/random-number-generator-seed-mistakes/}.
Since the \code{sitmo} \code{threefry}, this currently generates
one random deviate from the uniform distribution to seed the
engine \code{threefry} and then run the code.
}
\details{
Therefore, a simple call to the random number generated followed by a second
call to random number generated may have identical seeds.  As the number of
random number generator calls are increased the probability that the
birthday problem will increase.

The key to avoid this problem is to either run all simulations in the
\code{rxode2} environment once (therefore one seed or series of seeds
for the whole simulation), pre-generate all random variables
used for the simulation, or seed the rxode2 engine with \code{rxSetSeed()}

Internally each ID is seeded with a unique number so that the
results do not depend on the number of cores used.
}
\examples{

\donttest{

## Use threefry engine

rxgamma(0.5, n = 10) # with rxgamma you have to explicitly state n
rxgamma(5, n = 10, ncores = 2) # You can parallelize the simulation using openMP

rxgamma(1)


## This example uses `rxbeta` directly in the model

rx <- function() {
  model({
    a <- rxgamma(2)
  })
}

et <- et(1, id = 1:2)


s <- rxSolve(rx, et)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{logit}
\alias{logit}
\alias{expit}
\alias{logitNormInfo}
\alias{probitNormInfo}
\title{logit and inverse logit (expit) functions}
\usage{
logit(x, low = 0, high = 1)

expit(alpha, low = 0, high = 1)

logitNormInfo(mean = 0, sd = 1, low = 0, high = 1, abs.tol = 1e-06, ...)

probitNormInfo(mean = 0, sd = 1, low = 0, high = 1, abs.tol = 1e-06, ...)
}
\arguments{
\item{x}{Input value(s) in range [low,high] to translate -Inf to
Inf}

\item{low}{Lowest value in the range}

\item{high}{Highest value in the range}

\item{alpha}{Infinite value(s) to translate to range of [low,
high]}

\item{mean}{logit-scale mean}

\item{sd}{logit-scale standard deviation}

\item{abs.tol}{absolute accuracy requested.}

\item{...}{other parameters passed to \code{integrate()}}
}
\value{
values from logit and expit
}
\description{
logit and inverse logit (expit) functions
}
\details{
logit is given by:

logit(p) = -log(1/p-1)

where:

p = x-low/high-low

expit is given by:

expit(p, low, high) = (high-low)/(1+exp(-alpha)) + low

The \code{logitNormInfo()} gives the mean, variance and coefficient of
variability on the untransformed scale.
}
\examples{

logit(0.25)

expit(-1.09)

logitNormInfo(logit(0.25), sd = 0.1)

logitNormInfo(logit(1, 0, 10), sd = 1, low = 0, high = 10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxCbindStudyIndividual.R
\name{rxCbindStudyIndividual}
\alias{rxCbindStudyIndividual}
\title{Bind the study parameters and individual parameters}
\usage{
rxCbindStudyIndividual(studyParameters, individualParameters)
}
\arguments{
\item{studyParameters}{These are the study parameters, often can
be generated by sampling from a population.  This can be either
a matrix or a data frame}

\item{individualParameters}{A data frame of individual parameters}
}
\value{
Data frame that can be used in rxode2 simulations
}
\description{
Bind the study parameters and individual parameters
}
\examples{

# Function for coverting coefficient of covariance into a variance
lognCv <- function(x){log((x/100)^2+1)}

set.seed(32)

nSub  <-  100
nStud  <-  10

#define theta
theta <- c(lka=log(0.5), # log ka
          lCl=log(5), # log Cl
          lV=log(300) # log V
          )

#define theta Matrix
thetaMat <- lotri(lCl ~ lognCv(5),
                 lV  ~ lognCv(5),
                 lka ~ lognCv(5))


nev <- nSub*nStud

ev1 <- data.frame(COV1=rnorm(nev,50,30),COV2=rnorm(nev,75,10),
                  COV3=sample(c(1.0,2.0),nev,replace=TRUE))


tmat <-rxRmvn(nStud, theta[dimnames(thetaMat)[[1]]], thetaMat)

rxCbindStudyIndividual(tmat, ev1)

}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err.R
\name{rxDemoteAddErr}
\alias{rxDemoteAddErr}
\title{Demote the error type}
\usage{
rxDemoteAddErr(errType)
}
\arguments{
\item{errType}{Error type factor}
}
\value{
Demoted Error Type
}
\description{
Demote the error type
}
\examples{
rxErrTypeCombine("add") |>
  rxErrTypeCombine("prop")

# This removes the internal additive error
rxErrTypeCombine("add") |>
  rxErrTypeCombine("prop") |>
  rxDemoteAddErr()

# This is used for logitNorm(NA), the additive portion is stripped

}
\author{
Matthew Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gelu.R
\name{dGELU}
\alias{dGELU}
\alias{d2GELU}
\alias{d3GELU}
\alias{d4GELU}
\title{Derivatives of GELU}
\usage{
dGELU(x)

d2GELU(x)

d3GELU(x)

d4GELU(x)
}
\arguments{
\item{x}{numeric vector}
}
\value{
numeric vector
}
\description{
Derivatives of GELU
}
\examples{
dGELU(c(-2, -1, 0, 1, 2))
d2GELU(c(-2, -1, 0, 1, 2))
d3GELU(c(-2, -1, 0, 1, 2))
d4GELU(c(-2, -1, 0, 1, 2))
\donttest{
# you can use rxode2 as well
r <- rxode2({
   r1 <- dGELU(time)
   r2 <- d2GELU(time)
   r3 <- d3GELU(time)
   r4 <- d4GELU(time)
})
et <- et(c(-2, -1, 0, 1, 2))
rxSolve(r, et)
}
}
\seealso{
Other Activation Functions:
\code{\link[=ELU]{ELU()}},
\code{\link[=GELU]{GELU()}},
\code{\link[=PReLU]{PReLU()}},
\code{\link[=ReLU]{ReLU()}},
\code{\link[=SELU]{SELU()}},
\code{\link[=Swish]{Swish()}},
\code{\link[=dELU]{dELU()}},
\code{\link[=dPReLU]{dPReLU()}},
\code{\link[=dReLU]{dReLU()}},
\code{\link[=dSELU]{dSELU()}},
\code{\link[=dSwish]{dSwish()}},
\code{\link[=dlReLU]{dlReLU()}},
\code{\link[=dsoftplus]{dsoftplus()}},
\code{\link[=lReLU]{lReLU()}},
\code{\link[=softplus]{softplus()}}
}
\concept{Activation Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxTrans}
\alias{rxTrans}
\alias{rxTrans.default}
\alias{rxTrans.character}
\title{Translate the model to C code if needed}
\usage{
rxTrans(
  model,
  modelPrefix = "",
  md5 = "",
  modName = NULL,
  modVars = FALSE,
  ...
)

\method{rxTrans}{default}(
  model,
  modelPrefix = "",
  md5 = "",
  modName = NULL,
  modVars = FALSE,
  ...
)

\method{rxTrans}{character}(
  model,
  modelPrefix = "",
  md5 = "",
  modName = NULL,
  modVars = FALSE,
  eventSensKey = .rxEventSensCacheKey,
  ...
)
}
\arguments{
\item{model}{This is the ODE model specification.  It can be:
\itemize{
\item a string containing the set of ordinary differential
equations (ODE) and other expressions defining the changes in
the dynamic system.
\item a file name where the ODE system equation is contained
}

An ODE expression enclosed in \verb{\\\{\\\}}

(see also the \code{filename} argument). For
details, see the sections \dQuote{Details} and
\verb{rxode2 Syntax} below.}

\item{modelPrefix}{Prefix of the model functions that will be
compiled to make sure that multiple rxode2 objects can coexist
in the same R session.}

\item{md5}{Is the md5 of the model before parsing, and is used to
embed the md5 into DLL, and then provide for functions like
\code{\link[=rxModelVars]{rxModelVars()}}.}

\item{modName}{a string to be used as the model name. This string
is used for naming various aspects of the computations,
including generating C symbol names, dynamic libraries,
etc. Therefore, it is necessary that \code{modName} consists of
simple ASCII alphanumeric characters starting with a letter.}

\item{modVars}{returns the model variables instead of the named
vector of translated properties.}

\item{...}{Ignored parameters.}

\item{eventSensKey}{event ("jump") sensitivity mode key folded into
the parsed md5 cache key so the same model text translated in
different modes (e.g. \code{"fd"} vs \code{"jump"}) compiles to distinct
DLLs instead of colliding on a cached translation.  Defaults to
the session global \code{.rxEventSensCacheKey} (set by \code{\link[=rxode2]{rxode2()}}
before parsing); an empty string leaves the md5 unchanged so the
legacy \code{"fd"} path keeps existing caches valid.  It is a formal
argument (rather than read from the global in the body) because
\code{rxTrans.character} is memoised and \code{memoise} keys on
default-valued formals, so the cache distinguishes the modes.}
}
\value{
a named vector of translated model properties
including what type of jacobian is specified, the \code{C} function prefixes,
as well as the \code{C} functions names to be called through the compiled model.
}
\description{
This function translates the model to C code, if needed
}
\seealso{
\code{\link[=rxode2]{rxode2()}}, \code{\link[=rxCompile]{rxCompile()}}.
}
\author{
Matthew L.Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxUdfUiParsing}
\alias{rxUdfUiParsing}
\title{Returns if the current ui function is being parsed}
\usage{
rxUdfUiParsing()
}
\value{
logical if the current ui function is being parsed
}
\description{
Returns if the current ui function is being parsed
}
\examples{
rxUdfUiParsing()
}
\seealso{
Other User functions:
\code{\link[=linMod]{linMod()}},
\code{\link[=rxUdfUiControl]{rxUdfUiControl()}},
\code{\link[=rxUdfUiData]{rxUdfUiData()}},
\code{\link[=rxUdfUiEst]{rxUdfUiEst()}},
\code{\link[=rxUdfUiIniLhs]{rxUdfUiIniLhs()}},
\code{\link[=rxUdfUiMv]{rxUdfUiMv()}},
\code{\link[=rxUdfUiNum]{rxUdfUiNum()}}
}
\author{
Matthew L. Fidler
}
\concept{User functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/swish.R
\name{dSwish}
\alias{dSwish}
\title{Derivative of the Swish Activation Function}
\usage{
dSwish(x)
}
\arguments{
\item{x}{A numeric vector. All elements must be finite and
non-missing.}
}
\value{
A numeric vector where the derivative of the SELU function
has been applied to each element of \code{x}.
}
\description{
Derivative of the Swish Activation Function
}
\examples{
dSwish(c(-1, 0, 1, 2))

\donttest{
# Can also be used in rxode2:
x <- rxode2({
  r <- dSwish(time)
})
e <- et(c(-1, 0, 1, 2))
rxSolve(x, e)
}
}
\seealso{
Other Activation Functions:
\code{\link[=ELU]{ELU()}},
\code{\link[=GELU]{GELU()}},
\code{\link[=PReLU]{PReLU()}},
\code{\link[=ReLU]{ReLU()}},
\code{\link[=SELU]{SELU()}},
\code{\link[=Swish]{Swish()}},
\code{\link[=dELU]{dELU()}},
\code{\link[=dGELU]{dGELU()}},
\code{\link[=dPReLU]{dPReLU()}},
\code{\link[=dReLU]{dReLU()}},
\code{\link[=dSELU]{dSELU()}},
\code{\link[=dlReLU]{dlReLU()}},
\code{\link[=dsoftplus]{dsoftplus()}},
\code{\link[=lReLU]{lReLU()}},
\code{\link[=softplus]{softplus()}}
}
\author{
Matthew Fidler
}
\concept{Activation Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxAllowUnload}
\alias{rxAllowUnload}
\title{Allow unloading of dlls}
\usage{
rxAllowUnload(allow)
}
\arguments{
\item{allow}{boolean indicating if garbage collection will unload of rxode2 dlls.}
}
\value{
Boolean allow; called for side effects
}
\description{
Allow unloading of dlls
}
\examples{

# Garbage collection will not unload un-used rxode2 dlls
rxAllowUnload(FALSE);

# Garbage collection will unload unused rxode2 dlls
rxAllowUnload(TRUE);
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxDelete}
\alias{rxDelete}
\title{Delete the DLL for the model}
\usage{
rxDelete(obj)
}
\arguments{
\item{obj}{rxode2 family of objects}
}
\value{
A boolean stating if the operation was successful.
}
\description{
This function deletes the DLL, but doesn't delete the model
information in the object.
}
\author{
Matthew L.Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjoint.R
\name{.rxAdjointGradEval}
\alias{.rxAdjointGradEval}
\title{Evaluate a prebuilt adjoint objective gradient at given parameters}
\usage{
.rxAdjointGradEval(
  build,
  params,
  obsTimes,
  obs,
  weight = 1,
  denseBy = 0.01,
  atol = 1e-10,
  rtol = 1e-10
)
}
\arguments{
\item{build}{object returned by \code{\link[=.rxAdjointGradBuild]{.rxAdjointGradBuild()}}.}

\item{params}{named numeric vector of parameter values.}

\item{obsTimes}{numeric observation times.}

\item{obs}{numeric observed values aligned with \code{obsTimes}.}

\item{weight}{numeric scalar or vector of least-squares observation weights
(ignored when \code{errModel} is supplied).}

\item{denseBy}{grid spacing for the forward checkpoint trajectory; smaller
values reduce the covariate-interpolation error.}

\item{atol, rtol}{solver tolerances used for both the forward checkpoint and
the backward sweeps.}
}
\value{
named numeric vector \code{dG/dtheta} over the build's \code{calcSens}.
}
\description{
Evaluate a prebuilt adjoint objective gradient at given parameters
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxJacobian.R
\name{.rxAdjoint}
\alias{.rxAdjoint}
\title{Adjoint (backward) sensitivity equations for a model}
\usage{
.rxAdjoint(model, vars, states)
}
\arguments{
\item{model}{symengine model environment (as returned by \code{rxS}), with
\code{.rxJacobian(model, c(rxStateOde(model), vars))} already called.}

\item{vars}{character vector of parameter names to differentiate with
respect to (the \code{p} in \code{dy/dp}).  Defaults to \code{model$..vars}.}

\item{states}{character vector of output states of interest (the \code{k}).
Defaults to all ODE states, matching the full coverage of forward
sensitivities.}
}
\value{
character vector of \code{d/dt(...) = ...} lines (costate block followed
by quadrature block).  The costate (\verb{rx__adjLambda_*}) lines are internal
scaffolding; only the quadrature (\verb{rx__sens_*}) lines carry user-visible
output.  Also stashes the result in \code{model$..adjoint}.
}
\description{
Symbolically generates the continuous-adjoint ODE system mirroring the
forward sensitivity output (\code{.rxSens}), reusing the same
\verb{rx__sens_<state>_BY_<param>__} output names.  For each output state \code{k} it
emits two backward-in-time blocks: a costate
\verb{d/dt(rx__adjLambda_<k>_<i>__) = -(J^T lambda)_i} and a quadrature
\verb{d/dt(rx__sens_<k>_BY_<p>__) = -(lambda^T df/dp)_p}.  The derivatives are
taken from the \verb{rx__df_*} symbols \code{.rxJacobian()} already materialised, so no
new differentiation is done; the costate symbols are registered as bare
symengine \code{Symbol}s.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjoint.R
\name{.rxAdjointGrad}
\alias{.rxAdjointGrad}
\title{Adjoint gradient of an objective (single backward sweep)}
\usage{
.rxAdjointGrad(
  object,
  params,
  events,
  calcSens,
  pred,
  obsTimes,
  obs,
  weight = 1,
  errModel = NULL,
  denseBy = 0.01,
  atol = 1e-10,
  rtol = 1e-10
)
}
\arguments{
\item{object}{model definition accepted by \code{\link[=rxode2]{rxode2()}} (text,
\code{rxode2} object, or function/ui).}

\item{params}{named numeric vector of parameter values.}

\item{events}{event table / data used to define dosing.}

\item{calcSens}{character vector of parameter names to differentiate with
respect to.}

\item{pred}{character prediction expression \code{f} (function of states and
parameters), e.g. \code{"center/v"}.}

\item{obsTimes}{numeric observation times.}

\item{obs}{numeric observed values aligned with \code{obsTimes}.}

\item{weight}{numeric scalar or vector of least-squares observation weights
(ignored when \code{errModel} is supplied).}

\item{errModel}{\code{NULL} for least squares, or a list with character entries
\code{add} and/or \code{prop} naming the additive / proportional residual-error
parameters, selecting the FOCEi -2LL objective \code{v = add^2 + (prop*f)^2}.}

\item{denseBy}{grid spacing for the forward checkpoint trajectory; smaller
values reduce the covariate-interpolation error.}

\item{atol, rtol}{solver tolerances used for both the forward checkpoint and
the backward sweeps.}
}
\value{
named numeric vector \code{dG/dtheta} over \code{calcSens}.
}
\description{
Computes \code{dG/dtheta} for every parameter with a single backward sweep.
Supports least squares (default,
\verb{G = sum_i 1/2 weight_i (f_i - obs_i)^2}) or, when \code{errModel} is given, the
FOCEi -2 log-likelihood \verb{G = sum_i r_i^2/v_i + log(v_i)} with
\code{v_i = add^2 + (prop*f_i)^2}.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxJacobian.R
\name{rxExpandGrid}
\alias{rxExpandGrid}
\title{Faster expand.grid}
\usage{
rxExpandGrid(x, y, type = 0L)
}
\arguments{
\item{x}{first element (must be character)}

\item{y}{second element (must be character)}

\item{type}{Internal type=0L is traditional expand grid and
type=1L is Jacobian expand grid (adds symbols)}
}
\value{
Expand grid
}
\description{
Only support x and y as characters right now
}
\examples{

##
rxExpandGrid(letters, letters)

## Another fast method; See
## https://stackoverflow.com/questions/10405637/use-outer-instead-of-expand-grid

expand.grid.jc <- function(seq1, seq2) {
  cbind(
    Var1 = rep.int(seq1, length(seq2)),
    Var2 = rep.int(seq2, rep.int(length(seq1), length(seq2)))
  )
}
\donttest{
microbenchmark::microbenchmark(rxExpandGrid(letters, letters), expand.grid.jc(letters, letters))
}
}
\author{
Matthew Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxSerialize.R
\name{rxSaveState}
\alias{rxSaveState}
\title{Save the pre-integration rxode2 solver state to a binary file}
\usage{
rxSaveState(file)
}
\arguments{
\item{file}{Character(1).  Path of the file to write.}
}
\value{
Invisibly \code{TRUE}.
}
\description{
This is called automatically when \code{rxControl(serializeFile="path")} is
supplied to \code{rxSolve()}.  The file can be reloaded later by passing it
as the \code{params} argument of \code{rxSolve()} together with the matching
model object.
}
\seealso{
\code{\link{rxLoadState}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rinvchisq}
\alias{rinvchisq}
\title{Scaled Inverse Chi Squared distribution}
\usage{
rinvchisq(n = 1L, nu = 1, scale = 1)
}
\arguments{
\item{n}{Number of random samples}

\item{nu}{degrees of freedom of inverse chi square}

\item{scale}{Scale of inverse chi squared distribution
(default is 1).}
}
\value{
a vector of inverse chi squared deviates.
}
\description{
Scaled Inverse Chi Squared distribution
}
\examples{
rinvchisq(3, 4, 1) ## Scale = 1, degrees of freedom = 4
rinvchisq(2, 4, 2) ## Scale = 2, degrees of freedom = 4
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxParseSuppressMsg.R
\name{rxParseSuppressMsg}
\alias{rxParseSuppressMsg}
\title{Respect suppress messages}
\usage{
rxParseSuppressMsg()
}
\value{
Nothing
}
\description{
This turns on the silent REprintf in C when \code{suppressMessages()} is
turned on. This makes the \code{REprintf} act like \code{messages} in R,
they can be suppressed with \code{suppressMessages()}
}
\examples{

# rxParseSuppressMsg() is called with rxode2()

# Note the errors are output to the console

try(rxode2parse("d/dt(matt)=/3"), silent = TRUE)

# When using suppressMessages, the output is suppressed

suppressMessages(try(rxode2parse("d/dt(matt)=/3"), silent = TRUE))

# In rxode2, we use REprintf so that interrupted threads do not crash R
# if there is a user interrupt. This isn't captured by R's messages, but
# This interface allows the `suppressMessages()` to suppress the C printing
# as well

# If you  want to suppress messages from rxode2 in other packages, you can use
# this function
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getBaseSim.R
\name{getBaseSimModel}
\alias{getBaseSimModel}
\alias{getBaseSimModel.default}
\alias{getBaseIniSimModel.default}
\title{Get the base simulation model for simulation}
\usage{
getBaseSimModel(obj)

\method{getBaseSimModel}{default}(obj)

\method{getBaseIniSimModel}{default}(obj)
}
\arguments{
\item{obj}{Fit Object}
}
\value{
Simulation object
}
\description{
Get the base simulation model for simulation
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjoint.R
\name{.rxAdjointGradBuild}
\alias{.rxAdjointGradBuild}
\title{Build the symbolic part of an adjoint objective gradient (compile once)}
\usage{
.rxAdjointGradBuild(object, calcSens, pred, events, errModel = NULL)
}
\arguments{
\item{object}{model definition accepted by \code{\link[=rxode2]{rxode2()}} (text,
\code{rxode2} object, or function/ui).}

\item{calcSens}{character vector of parameter names to differentiate with
respect to.}

\item{pred}{character prediction expression \code{f} (function of states and
parameters), e.g. \code{"center/v"}.}

\item{events}{event table / data used to define dosing.}

\item{errModel}{\code{NULL} for least squares, or a list with character entries
\code{add} and/or \code{prop} naming the additive / proportional residual-error
parameters, selecting the FOCEi -2LL objective \code{v = add^2 + (prop*f)^2}.}
}
\value{
an opaque list consumed by \code{\link[=.rxAdjointGradEval]{.rxAdjointGradEval()}}.
}
\description{
Performs all symbolic work (symengine differentiation, \code{rxFromSE}, model
compilation) up front and returns a reusable object for
\code{\link[=.rxAdjointGradEval]{.rxAdjointGradEval()}}, so an optimiser evaluates the gradient at many
parameter values without re-running symbolic code.  Build before any
\code{rxSolve} to avoid a \code{load_all}-only symengine dispatch fragility.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxsolve.R
\name{+.rxSolve}
\alias{+.rxSolve}
\title{Update Solved object with '+'}
\usage{
\method{+}{rxSolve}(solved, new)
}
\arguments{
\item{solved}{Solved object}

\item{new}{New information added to the table.}
}
\value{
new solved object
}
\description{
Update Solved object with '+'
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{rxSetSum}
\alias{rxSetSum}
\title{Defunct setting of sum}
\usage{
rxSetSum(type = c("pairwise", "fsum", "kahan", "neumaier", "c"))
}
\arguments{
\item{type}{used to be type of product}
}
\value{
nothing
}
\description{
Defunct setting of sum
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode-options.R
\name{.rxMemoisedFns}
\alias{.rxMemoisedFns}
\title{Clear memoise cache for rxode2}
\usage{
.rxMemoisedFns
}
\value{
nothing; called for side effects
}
\description{
Clear memoise cache for rxode2
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{.rxLastCompileSuccess}
\alias{.rxLastCompileSuccess}
\title{Was the last compilation successful?}
\usage{
.rxLastCompileSuccess(set)
}
\arguments{
\item{set}{Logical flag. If missing, returns the current compilation
status. If provided, sets the compilation status.}
}
\value{
this determines if the last compilation was successful.  This is useful for
debugging and testing purposes.
}
\description{
Was the last compilation successful?
}
\examples{

.rxLastCompileSuccess()

}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.toClassicEvid}
\alias{.toClassicEvid}
\title{This converts NONMEM-style EVIDs to classic RxODE events}
\usage{
.toClassicEvid(cmt = 1L, amt = 0, rate = 0, dur = 0, ii = 0, evid = 0L, ss = 0)
}
\arguments{
\item{cmt}{compartment flag}

\item{amt}{dose amount}

\item{rate}{dose rate}

\item{dur}{dose duration}

\item{ii}{inter-dose interval}

\item{evid}{event id}

\item{ss}{steady state}
}
\value{
classic evids, excluding evids that are added (you need to
add them manually) or simply use etTran.  This is mostly for
testing and really shouldn't be used directly.
}
\description{
This converts NONMEM-style EVIDs to classic RxODE events
}
\examples{
.toClassicEvid(cmt=10, amt=3, evid=1)
.toClassicEvid(cmt=10, amt=3, rate=2, evid=1)
.toClassicEvid(cmt=10, amt=3, rate=-1, evid=1)
.toClassicEvid(cmt=10, amt=3, rate=-2, evid=1)
.toClassicEvid(cmt=10, amt=3, dur=2, evid=1)
.toClassicEvid(cmt=304, amt=3, dur=2, evid=1)
.toClassicEvid(cmt=7, amt=0, rate=2, evid=1, ss=1)
.toClassicEvid(cmt=-10, amt=3, evid=1)
.toClassicEvid(cmt=10, amt=3, evid=5)
.toClassicEvid(cmt=6, amt=3, evid=6)
.toClassicEvid(cmt=6, amt=3, evid=7)
.toClassicEvid(evid=2)
.toClassicEvid(evid=4)
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{rxSetProd}
\alias{rxSetProd}
\title{Defunct setting of product}
\usage{
rxSetProd(type = c("long double", "double", "logify"))
}
\arguments{
\item{type}{used to be type of product}
}
\value{
nothing
}
\description{
Defunct setting of product
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxIndLin.R
\name{rxIndLinState}
\alias{rxIndLinState}
\title{Set the preferred factoring by state}
\usage{
rxIndLinState(preferred = NULL)
}
\arguments{
\item{preferred}{A list of each state's preferred factorization}
}
\value{
Nothing
}
\description{
Set the preferred factoring by state
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/etTran.R
\name{.getLastIdLvl}
\alias{.getLastIdLvl}
\title{Get the last \code{idLvl}}
\usage{
.getLastIdLvl()
}
\value{
Last \code{idLvl}
}
\description{
Get the last \code{idLvl}
}
\examples{
.getLastIdLvl()
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{getRxThreads}
\alias{getRxThreads}
\alias{setRxThreads}
\alias{rxCores}
\title{Get/Set the number of threads that rxode2 uses}
\usage{
getRxThreads(verbose = FALSE)

setRxThreads(threads = NULL, percent = NULL, throttle = NULL)

rxCores(verbose = FALSE)
}
\arguments{
\item{verbose}{Display the value of relevant OpenMP settings}

\item{threads}{NULL (default) rereads environment variables
specified by OpenMP (\code{OMP_THREAD_LIMIT}).  If they are not
specified, this determines based the logical cores available. 0
means to use all logical CPU threads available. Otherwise a number >= 1}

\item{percent}{If provided it should be a number between 2 and
100; the percentage of logical CPUs to use. By default on
startup, 50 percent.}

\item{throttle}{2 (default) means that, roughly speaking, a
single thread will be used when number subjects solved for is <=2, 2 threads when
the number of all points is <=4, etc. The throttle is to speed up small data
tasks (especially when repeated many times) by not incurring the
overhead of managing multiple threads.

The throttle will also suppress sorting which ID will be solved first
when there are (nsubject solved)*throttle <= nthreads.  In
\code{rxode2} this sorting occurs to minimize the time for waiting for
another thread to finish. If the last item solved is has a long
solving time, all the other solving have to wait for that last
costly solving to occur. If the items which are likely to take
more time are solved first, this wait is less likely to have an
impact on the overall solving time.

In rxode2 the ids are sorted by the individual number of solving
points (largest first). It also has a C interface that allows
these ids to be resorted by total time spent solving the
equation.  This allows packages like nlmixr to sort by solving
time if needed.

Overall the the number of threads is throttled (restricted) for
small tasks and sorting for ids are suppressed.}
}
\value{
number of threads that rxode2 uses
}
\description{
Get/Set the number of threads that rxode2 uses
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{rxTheme}
\alias{rxTheme}
\title{rxTheme is the ggplot2 theme for rxode2 plots}
\usage{
rxTheme(
  base_size = 11,
  base_family = "",
  base_line_size = base_size/22,
  base_rect_size = base_size/22,
  grid = TRUE
)
}
\arguments{
\item{base_size}{base font size, given in pts.}

\item{base_family}{base font family}

\item{base_line_size}{base size for line elements}

\item{base_rect_size}{base size for rect elements}

\item{grid}{a Boolean indicating if the grid is on (\code{TRUE}) or off
(\code{FALSE}). This could also be a character indicating \code{x} or \code{y}.}
}
\value{
ggplot2 theme used in rxode2
}
\description{
rxTheme is the ggplot2 theme for rxode2 plots
}
\seealso{
Other rxode2 plotting:
\code{\link[=plot.rxSolve]{plot.rxSolve()}}
}
\concept{rxode2 plotting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{meanProbs}
\alias{meanProbs}
\alias{meanProbs.default}
\title{Calculate expected confidence bands or prediction intreval with normal or t sampling distribution}
\usage{
meanProbs(x, ...)

\method{meanProbs}{default}(
  x,
  probs = seq(0, 1, 0.25),
  na.rm = FALSE,
  names = TRUE,
  useT = TRUE,
  onlyProbs = TRUE,
  pred = FALSE,
  n = 0L,
  ...
)
}
\arguments{
\item{x}{numeric vector whose mean and probability based confidence
values are wanted, NA and NaN values are not allowed in numeric
vectors unless 'na.rm' is 'TRUE'.}

\item{...}{Arguments passed to default method, allows many
different methods to be applied.}

\item{probs}{numeric vector of probabilities with values in the
interval from 0 to 1 .}

\item{na.rm}{logical; if true, any NA and NaN's are removed from
\code{x} before the quantiles are computed.}

\item{names}{logical; if true, the result has a names attribute.}

\item{useT}{logical; if true, use the t-distribution to calculate
the confidence-based estimates. If false use the normal
distribution to calculate the confidence based estimates.}

\item{onlyProbs}{logical; if true, only return the probability
based confidence interval estimates, otherwise return}

\item{pred}{logical; if true use the prediction interval instead of
the confidence interval}

\item{n}{integer/integerish; this is the n used to calculate the
prediction or confidence interval.  When \code{n=0} (default) use the
number of non-\code{NA} observations.}
}
\value{
By default the return has the probabilities as names (if
named) with the points where the expected distribution are
located given the sampling mean and standard deviation. If
\code{onlyProbs=FALSE} then it would prepend mean, variance, standard
deviation, minimum, maximum and number of non-NA observations.
}
\description{
The generic function \code{meanProbs} produces expected confidence bands
under either the t distribution or the normal sampling
distribution. This uses \code{qnorm()} or \code{qt()} with the mean and
standard deviation.
}
\details{
For a single probability, p, it uses either:

mean + qt(p, df=n)*sd/sqrt(n)

or

mean + qnorm(p)*sd/sqrt(n)

The smallest observation corresponds to a probability of 0 and the
largest to a probability of 1 and the mean corresponds to 0.5.

The mean and standard deviation of the sample is calculated based
on Welford's method for a single pass.

This is meant to perform in the same way as \code{quantile()} so it can
be a drop in replacement for code using \code{quantile()} but using
distributional assumptions.
}
\examples{

quantile(x<- rnorm(1001))
meanProbs(x)

# Can get some extra statistics if you request onlyProbs=FALSE
meanProbs(x, onlyProbs=FALSE)

x[2] <- NA_real_

meanProbs(x, onlyProbs=FALSE)

quantile(x<- rnorm(42))

meanProbs(x)

meanProbs(x, useT=FALSE)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{rxSetProgressBar}
\alias{rxSetProgressBar}
\title{Set timing for progress bar}
\usage{
rxSetProgressBar(seconds = 1)
}
\arguments{
\item{seconds}{This sets the number of seconds that need to elapse
before drawing the next segment of the progress bar.  When
this is zero or below this turns off the progress bar.}
}
\value{
nothing, used for side effects
}
\description{
Set timing for progress bar
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxCondition}
\alias{rxCondition}
\title{Current Condition for rxode2 object}
\usage{
rxCondition(obj, condition = NULL)
}
\arguments{
\item{obj}{rxode2 object}

\item{condition}{If specified and is one of the conditions in the
rxode2 object (as determined by \code{\link[=rxExpandIfElse]{rxExpandIfElse()}}),
assign the rxode2 current condition to this parameter.  If the
condition is not one of the known condition, the condition is
set to \code{NULL}, implying no conditioning currently used.}
}
\value{
Current condition for rxode2 object
}
\description{
Current Condition for rxode2 object
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxOptExpr.R
\name{rxOptExpr}
\alias{rxOptExpr}
\title{Optimize rxode2 for computer evaluation}
\usage{
rxOptExpr(x, msg = "model", chunkLines = 40L, parallel = 0L)
}
\arguments{
\item{x}{rxode2 model that can be accessed by rxNorm}

\item{msg}{This is the name of type of object that rxode2 is
optimizing that will in the message when optimizing.  For
example "model" will produce the following message while
optimizing the model:

finding duplicate expressions in model...}

\item{chunkLines}{Integer; when positive (the default is 40),
a model longer than this many lines is optimized in contiguous
cost-balanced chunks of roughly this many lines instead of in a
single pass; a model at or under it is optimized whole, exactly
as before.  \code{0} always optimizes the whole model at once.

\if{html}{\out{<div class="sourceCode">}}\preformatted{Chunking pays off for a large machine-generated model -- a
sensitivity- or Jacobian-augmented model, say.  Normalizing a
model (`rxNorm()`, i.e. parsing it) is strongly superlinear in
its size, and for such a model it, not the common subexpression
search, is what dominates: optimizing a 275-line augmented model
takes ~113s, of which the subexpression search is only ~15s.
Chunking amortizes that parse -- `rxOptExpr()` normalizes each
chunk on its own -- taking the same model to ~11s:

\\tabular\{rrrr\}\{
  lines \\tab whole \\tab chunked \\tab \\cr
  34 \\tab 0.5s \\tab 0.5s \\tab (a typical model: not chunked) \\cr
  119 \\tab 2.0s \\tab 0.8s \\tab 2.5x \\cr
  149 \\tab 22.2s \\tab 3.9s \\tab 5.7x \\cr
  275 \\tab 112.7s \\tab 10.6s \\tab 10.7x \\cr
\}

Common subexpressions are then only shared within a chunk, so the
model is equivalent but carries more temporaries.  That costs no
measurable solve time, but it does make the C compilation of the
model somewhat slower, which partly offsets the gain.

A chunk is a fragment, so it can fail to optimize where the whole
model would not.  If any chunk fails, the whole model is optimized
instead, so a malformed model still raises the error the unchunked
call raises; falling back costs the unchunked time only on that
rare path.

Chunking therefore does not give the same optimized text as the
whole-model call -- it shares fewer subexpressions and so carries
more temporaries -- but it gives an equivalent model: the same
states and parameters, the same solution, and the same errors.
}\if{html}{\out{</div>}}}

\item{parallel}{Integer; number of \code{mirai} daemons used to optimize
the chunks in parallel.  Only used when the model is chunked.  It
carries the same semantics as \code{rxControl(cores=)}: \code{0} (the
default) means the rxode2 thread setting \code{rxCores()}, so CRAN and
users tune it with the same knob as the solver (\code{setRxThreads()},
\code{OMP_THREAD_LIMIT}) or by passing \verb{parallel=} directly; \code{1} runs
the chunks serially.
It is capped by the number of chunks and by \code{rxCores()}, so it
will not oversubscribe past the threads the user asked for.

\if{html}{\out{<div class="sourceCode">}}\preformatted{An existing `mirai` daemon pool is used as-is and left running.
Otherwise a pool is started for the call and shut down when it
returns; that startup (loading rxode2 into each daemon) costs a
few seconds, so a pool is only started when the model splits into
at least 4 chunks, where the parallel win covers it.
}\if{html}{\out{</div>}}}
}
\value{
Optimized rxode2 model text.  The order and type lhs and
state variables is maintained while the evaluation is sped up.
While parameters names are maintained, their order may be
modified.
}
\description{
This optimizes rxode2 code for computer evaluation by only
calculating redundant expressions once.
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{.clearPipe}
\alias{.clearPipe}
\title{Clear/Set pipeline}
\usage{
.clearPipe(
  rx = NULL,
  inits = NULL,
  events = NULL,
  params = NULL,
  iCov = NULL,
  keep = NULL,
  thetaMat = NULL,
  omega = NULL,
  sigma = NULL,
  dfObs = NULL,
  dfSub = NULL,
  nSub = NULL,
  nStud = NULL
)
}
\arguments{
\item{rx}{rxode2 object}
}
\value{
None, clears rxode2 pipeline
}
\description{
Clear/Set pipeline
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxode2}
\alias{rxode2}
\alias{RxODE}
\alias{rxode}
\title{Create an ODE-based model specification}
\usage{
rxode2(
  model,
  modName = basename(wd),
  wd = getwd(),
  filename = NULL,
  extraC = NULL,
  debug = FALSE,
  calcJac = NULL,
  calcSens = NULL,
  calcSens2 = NULL,
  calcSens3 = NULL,
  collapseModel = FALSE,
  package = NULL,
  ...,
  linCmtSens = c("linCmtA", "linCmtB"),
  indLin = FALSE,
  eventSens = NULL,
  verbose = FALSE,
  fullPrint = getOption("rxode2.fullPrint", FALSE),
  envir = parent.frame()
)

RxODE(
  model,
  modName = basename(wd),
  wd = getwd(),
  filename = NULL,
  extraC = NULL,
  debug = FALSE,
  calcJac = NULL,
  calcSens = NULL,
  calcSens2 = NULL,
  calcSens3 = NULL,
  collapseModel = FALSE,
  package = NULL,
  ...,
  linCmtSens = c("linCmtA", "linCmtB"),
  indLin = FALSE,
  eventSens = NULL,
  verbose = FALSE,
  fullPrint = getOption("rxode2.fullPrint", FALSE),
  envir = parent.frame()
)

rxode(
  model,
  modName = basename(wd),
  wd = getwd(),
  filename = NULL,
  extraC = NULL,
  debug = FALSE,
  calcJac = NULL,
  calcSens = NULL,
  calcSens2 = NULL,
  calcSens3 = NULL,
  collapseModel = FALSE,
  package = NULL,
  ...,
  linCmtSens = c("linCmtA", "linCmtB"),
  indLin = FALSE,
  eventSens = NULL,
  verbose = FALSE,
  fullPrint = getOption("rxode2.fullPrint", FALSE),
  envir = parent.frame()
)
}
\arguments{
\item{model}{This is the ODE model specification.  It can be:
\itemize{
\item a string containing the set of ordinary differential
equations (ODE) and other expressions defining the changes in
the dynamic system.
\item a file name where the ODE system equation is contained
}

An ODE expression enclosed in \verb{\\\{\\\}}

(see also the \code{filename} argument). For
details, see the sections \dQuote{Details} and
\verb{rxode2 Syntax} below.}

\item{modName}{a string to be used as the model name. This string
is used for naming various aspects of the computations,
including generating C symbol names, dynamic libraries,
etc. Therefore, it is necessary that \code{modName} consists of
simple ASCII alphanumeric characters starting with a letter.}

\item{wd}{character string with a working directory where to
create a subdirectory according to \code{modName}. When
specified, a subdirectory named after the
\dQuote{\code{modName.d}} will be created and populated with a
C file, a dynamic loading library, plus various other working
files. If missing, the files are created (and removed) in the
temporary directory, and the rxode2 DLL for the model is
created in the current directory named \verb{rx_????_platform}, for
example \code{rx_129f8f97fb94a87ca49ca8dafe691e1e_i386.dll}}

\item{filename}{A file name or connection object where the
ODE-based model specification resides. Only one of \code{model}
or \code{filename} may be specified.}

\item{extraC}{Extra c code to include in the model.  This can be
useful to specify functions in the model.  These C functions
should usually take \code{double} precision arguments, and
return \code{double} precision values.}

\item{debug}{is a boolean indicating if the executable should be
compiled with verbose debugging information turned on.}

\item{calcJac}{boolean indicating if rxode2 will calculate the
Jacobain according to the specified ODEs.}

\item{calcSens}{boolean indicating if rxode2 will calculate the
sensitivities according to the specified ODEs.  May also be a
character vector of the states/parameters whose first-order
sensitivities (\verb{rx__sens_<state>_BY_<param>__}) should be
generated.}

\item{calcSens2}{character vector (or \code{NULL}) requesting
second-order sensitivities in addition to the first-order ones
from \code{calcSens}.  When supplied, rxode2 also generates the
\verb{rx__sens_<state>_BY_<p>_BY_<q>__} compartments (the Hessian
path), where \code{p} ranges over \code{calcSens} and \code{q} over
\code{calcSens2}.  Used, for example, for population (\code{THETA})
second-order event sensitivities.  \code{NULL} (the default) skips the
second-order generation.}

\item{calcSens3}{character vector (or \code{NULL}) requesting third-order
sensitivities in addition to the first- and second-order ones.
Requires \code{calcSens2} to also be supplied (every \code{calcSens3}
parameter needs its own already-built second-order sensitivity
compartment, from the pairing of \code{calcSens2} with \code{calcSens3}, to
reference -- so \code{calcSens3} should be a subset of \code{calcSens2},
which itself should be a subset of \code{calcSens}, mirroring how
\code{calcSens2} is used everywhere else in rxode2 today).  When
supplied, generates the \verb{rx__sens_<state>_BY_<p>_BY_<q>_BY_<r>__}
compartments, where \code{p} ranges over \code{calcSens}, \code{q} over
\code{calcSens2}, and \code{r} over \code{calcSens3}, via \code{rxExpandSens3_()}.
\code{NULL} (the default) skips the third-order generation.}

\item{collapseModel}{boolean indicating if rxode2 will remove all
LHS variables when calculating sensitivities.}

\item{package}{Package name for pre-compiled binaries.}

\item{...}{ignored arguments.}

\item{linCmtSens}{The method to calculate the linCmt() solutions}

\item{indLin}{Calculate inductive linearization matrices and
compile with inductive linearization support.}

\item{eventSens}{controls how dosing/event-parameter (alag, F, rate,
dur, amt) sensitivities are computed when sensitivities are
generated: \code{"jump"} injects the analytic event ("jump")
sensitivities into the sensitivity states at each dosing event,
\code{"fd"} keeps the legacy finite-difference behavior (the
backward-compatible opt-out), and \code{"both"} computes both for
cross-checking.  \code{NULL} (the default) uses
\code{getOption("rxode2.eventSens", "fd")}.  When not \code{"fd"} and
\code{calcSens} is supplied, \code{calcJac} is forced to \code{TRUE} so the
Jacobian is available for the jump injection.}

\item{verbose}{When \code{TRUE} be verbose with the linear
compartmental model}

\item{fullPrint}{When using \code{printf} within the model, if \code{TRUE}
print on every step (except ME/indLin), otherwise when \code{FALSE}
print only when calculating the \code{d/dt}}

\item{envir}{is the environment to look for R user functions
(defaults to parent environment)}
}
\value{
An object (environment) of class \code{rxode2} (see Chambers and Temple Lang (2001))
consisting of the following list of strings and functions:

\if{html}{\out{<div class="sourceCode">}}\preformatted{* `model` a character string holding the source model specification.
* `get.modelVars`a function that returns a list with 3 character
    vectors, `params`, `state`, and `lhs` of variable names used in the model
    specification. These will be output when the model is computed (i.e., the ODE solved by integration).

  * `solve`\{this function solves (integrates) the ODE. This
      is done by passing the code to [rxSolve()].
      This is as if you called `rxSolve(rxode2object, ...)`,
      but returns a matrix instead of a rxSolve object.

      `params`: a numeric named vector with values for every parameter
      in the ODE system; the names must correspond to the parameter
      identifiers used in the ODE specification;

      `events`: an `eventTable` object describing the
      input (e.g., doses) to the dynamic system and observation
      sampling time points (see  [eventTable()]);

      `inits`: a vector of initial values of the state variables
      (e.g., amounts in each compartment), and the order in this vector
      must be the same as the state variables (e.g., PK/PD compartments);


      `stiff`: a logical (`TRUE` by default) indicating whether
      the ODE system is stiff or not.

      For stiff ODE systems (`stiff = TRUE`), `rxode2` uses
      the LSODA (Livermore Solver for Ordinary Differential Equations)
      Fortran package, which implements an automatic method switching
      for stiff and non-stiff problems along the integration interval,
      authored by Hindmarsh and Petzold (2003).

      For non-stiff systems (`stiff = FALSE`), `rxode2` uses `DOP853`,
      an explicit Runge-Kutta method of order 8(5, 3) of Dormand and Prince
      as implemented in C by Hairer and Wanner (1993).

      `trans_abs`: a logical (`FALSE` by default) indicating
      whether to fit a transit absorption term
      (TODO: need further documentation and example);

      `atol`: a numeric absolute tolerance (1e-08 by default);

      `rtol`: a numeric relative tolerance (1e-06 by default).

      The output of \\dQuote\{solve\} is a matrix with as many rows as there
      are sampled time points and as many columns as system variables
      (as defined by the ODEs and additional assignments in the rxode2 model
          code).\}

  * `isValid` a function that (naively) checks for model validity,
      namely that the C object code reflects the latest model
      specification.
  * `version` a string with the version of the `rxode2`
      object (not the package).
  * `dynLoad` a function with one `force = FALSE` argument
      that dynamically loads the object code if needed.
  * `dynUnload` a function with no argument that unloads
      the model object code.
  * `delete` removes all created model files, including C and DLL files.
      The model object is no longer valid and should be removed, e.g.,
      `rm(m1)`.
  * `run` deprecated, use `solve`.
  * `get.index` deprecated.
  * `getObj` internal (not user callable) function.
}\if{html}{\out{</div>}}
}
\description{
Create a dynamic ODE-based model object suitably for translation
into fast C code
}
\details{
The \code{Rx} in the name \code{rxode2} is meant to suggest the
abbreviation \emph{Rx} for a medical prescription, and thus to
suggest the package emphasis on pharmacometrics modeling, including
pharmacokinetics (PK), pharmacodynamics (PD), disease progression,
drug-disease modeling, etc.

\subsection{Creating rxode2 models}{

The ODE-based model specification may be coded inside four places:
\itemize{
\item Inside a \code{rxode2({})} block statements:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(rxode2)
mod <- rxode2(\{
  # simple assignment
  C2 <- centr/V2

  # time-derivative assignment
  d/dt(centr) <- F*KA*depot - CL*C2 - Q*C2 + Q*C3;
\})
}\if{html}{\out{</div>}}
\itemize{
\item Inside a \code{rxode2("")} string statement:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{mod <- rxode2("
  # simple assignment
  C2 <- centr/V2

  # time-derivative assignment
  d/dt(centr) <- F*KA*depot - CL*C2 - Q*C2 + Q*C3;
")
}\if{html}{\out{</div>}}
\itemize{
\item In a file name to be loaded by rxode2:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{writeLines("
  # simple assignment
  C2 <- centr/V2

  # time-derivative assignment
  d/dt(centr) <- F*KA*depot - CL*C2 - Q*C2 + Q*C3;
", "modelFile.rxode2")
mod <- rxode2(filename='modelFile.rxode2')
unlink("modelFile.rxode2")
}\if{html}{\out{</div>}}
\itemize{
\item In a model function which can be parsed by \code{rxode2}:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{mod <- function() \{
  model(\{
    # simple assignment
    C2 <- centr/V2

    # time-derivative assignment
    d/dt(centr) <- F*KA*depot - CL*C2 - Q*C2 + Q*C3;
  \})
\}

mod <- rxode2(mod) # or simply mod() if the model is at the end of the function

# These model functions often have residual components and initial
# (`ini(\{\})`) conditions attached as well. For example the
# theophylline model can be written as:

one.compartment <- function() \{
  ini(\{
    tka <- 0.45 # Log Ka
    tcl <- 1 # Log Cl
    tv <- 3.45    # Log V
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  \})
  model(\{
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v <- exp(tv + eta.v)
    d/dt(depot) = -ka * depot
    d/dt(center) = ka * depot - cl / v * center
    cp = center / v
    cp ~ add(add.sd)
  \})
\}

# after parsing the model
mod <- one.compartment()
}\if{html}{\out{</div>}}

For the block statement, character string or text file an internal
\code{rxode2} compilation manager translates the ODE system into C, compiles
it and loads it into the R session. The call to \code{rxode2} produces an
object of class \code{rxode2} which consists of a list-like structure
(environment) with various member functions.

For the last type of model (a model function), a call to \code{rxode2}
creates a parsed \code{rxode2} ui that can be translated to the \code{rxode2}
compilation model.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{mod$simulationModel

# or
mod$simulationIniModel
}\if{html}{\out{</div>}}

This is the same type of function required for \code{nlmixr2} estimation and
can be extended and modified by model piping. For this reason will be
focused on in the documentation.
}

This basic model specification consists of one or more statements
optionally terminated by semi-colons \verb{;} and optional comments (comments
are delimited by \verb{#} and an end-of-line).

A block of statements is a set of statements delimited by curly braces,
\code{{ ... }}.

Statements can be either assignments, conditional \code{if}/\verb{else if}/\verb{else},
\code{while} loops (can be exited by \code{break}), special statements, or
printing statements (for debugging/testing).

Assignment statements can be:
\itemize{
\item \strong{simple} assignments, where the left hand is an identifier (i.e.,
variable). This includes string assignments
\item special \strong{time-derivative} assignments, where the left hand specifies
the change of the amount in the corresponding state variable
(compartment) with respect to time e.g., \code{d/dt(depot)}:
\item special \strong{initial-condition} assignments where the left hand
specifies the compartment of the initial condition being specified,
e.g. \code{depot(0) = 0}
\item special model event changes including \strong{bioavailability}
(\code{f(depot)=1}), \strong{lag time} (\code{alag(depot)=0}), \strong{modeled rate}
(\code{rate(depot)=2}) and \strong{modeled duration} (\code{dur(depot)=2}). An
example of these model features and the event specification for the
modeled infusions the rxode2 data specification is found in \href{https://nlmixr2.github.io/rxode2/articles/rxode2-event-types.html}{rxode2 events vignette}.
\item special \strong{change point syntax, or model times}. These model times are
specified by \code{mtime(var)=time}
\item special \strong{Jacobian-derivative} assignments, where the left hand
specifies the change in the compartment ode with respect to a
variable. For example, if \code{d/dt(y) = dy}, then a Jacobian for this
compartment can be specified as \code{df(y)/dy(dy) = 1}. There may be some
advantage to obtaining the solution or specifying the Jacobian for
very stiff ODE systems. However, for the few stiff systems we tried
with LSODA, this actually slightly slowed down the solving.
\item Special \strong{string value declarations} which tell what values a string
variable will take within a \code{rxode2} solving structure. These values
will then cause a factor to be created for this variable on solving
the \code{rxode2} model. As such, they are declared in much the same way as
\code{R}, that is: \code{labels(a) <- c("a1", "a2")}.
}

Note that assignment can be done by \code{=}, \verb{<-} or \code{~}.

When assigning with the \code{~} operator, the \strong{simple assignments} and
\strong{time-derivative} assignments will not be output. Note that with the
\code{rxode2} model functions assignment with \code{~} can also be overloaded with
a residual distribution specification.

Special statements can be:
\itemize{
\item \strong{Compartment declaration statements}, which can change the default
dosing compartment and the assumed compartment number(s) as well as
add extra compartment names at the end (useful for multiple-endpoint
nlmixr models); These are specified by \code{cmt(compartmentName)}
\item \strong{Parameter declaration statements}, which can make sure the input
parameters are in a certain order instead of ordering the parameters
by the order they are parsed. This is useful for keeping the parameter
order the same when using 2 different ODE models. These are specified
by \code{param(par1, par2,...)}
\item \strong{Variable interpolation statements}, which tells the interpolation
for specific covariates. These include \code{locf(cov1,     cov2, ...)} for
last observation carried forward, \code{nocb(cov1,     cov2, ...)} for next
observation carried backward, \code{linear(cov1,     cov2, ...)} for linear
interpolation and \code{midpoint(cov1, cov2,     ...)} for midpoint
interpolation.
}

An example model is shown below:

\if{html}{\out{<div class="sourceCode">}}\preformatted{   # simple assignment
   C2 <- centr/V2

   # time-derivative assignment
   d/dt(centr) <- F*KA*depot - CL*C2 - Q*C2 + Q*C3;
}\if{html}{\out{</div>}}

Expressions in assignment and \code{if} statements can be numeric or logical.

Numeric expressions can include the following numeric operators
\verb{+, -, *, /, ^, \%\%} and those mathematical functions defined in the C or
the R math libraries (e.g., \code{fabs}, \code{exp}, \code{log}, \code{sin}, \code{abs}).

You may also access the R’s functions in the \href{https://cran.r-project.org/doc/manuals/r-release/R-exts.html#Numerical-analysis-subroutines}{R math libraries},
like \code{lgammafn} for the log gamma function.

The \code{rxode2} syntax is case-sensitive, i.e., \code{ABC} is different than
\code{abc}, \code{Abc}, \code{ABc}, etc.
\subsection{Identifiers}{

Like R, Identifiers (variable names) may consist of one or more
alphanumeric, underscore \verb{_} or period \code{.} characters, but the first
character cannot be a digit or underscore \verb{_}.

Identifiers in a model specification can refer to:
\itemize{
\item State variables in the dynamic system (e.g., compartments in a
pharmacokinetics model).
\item Implied input variable, \code{t} (time), \code{tlast} (last time point), and
\code{podo} (oral dose, in the undocumented case of absorption transit
models).
\item Special constants like \code{pi} or \href{https://cran.r-project.org/doc/manuals/r-release/R-exts.html#Mathematical-constants}{R’s predefined constants}.
\item Model parameters (e.g., \code{ka} rate of absorption, \code{CL} clearance, etc.)
\item Others, as created by assignments as part of the model specification;
these are referred as \emph{LHS} (left-hand side) variable.
}

Currently, the \code{rxode2} modeling language only recognizes system state
variables and “parameters”, thus, any values that need to be passed from
R to the ODE model (e.g., \code{age}) should be either passed in the \code{params}
argument of the integrator function \code{rxSolve()} or be in the supplied
event data-set.

There are certain variable names that are in the \code{rxode2} event tables.
To avoid confusion, the following event table-related items cannot be
assigned, or used as a state but can be accessed in the rxode2 code:
\itemize{
\item \code{cmt}
\item \code{dvid}
\item \code{addl}
\item \code{ss}
\item \code{amt}
\item \code{dur}
\item \code{rate}
\item \code{Rprintf}
\item \code{print}
\item \code{printf}
\item \code{id}
}

However the following variables are cannot be used in a model
specification:
\itemize{
\item \code{evid}
\item \code{ii}
}

Sometimes rxode2 generates variables that are fed back to rxode2.
Similarly, nlmixr2 generates some variables that are used in nlmixr
estimation and simulation. These variables start with the either the
\code{rx} or \code{nlmixr} prefixes. To avoid any problems, it is suggested to not
use these variables starting with either the \code{rx} or \code{nlmixr} prefixes.
}

\subsection{Logical Operators}{

Logical operators support the standard R operators \code{==}, \code{!=} \code{>=} \code{<=}
\code{>} and \code{<}. Like R these can be in \verb{if()} or \verb{while()} statements,
\code{ifelse()} expressions. Additionally they can be in a standard
assignment. For instance, the following is valid:

\if{html}{\out{<div class="sourceCode">}}\preformatted{cov1 = covm*(sexf == "female") + covm*(sexf != "female")
}\if{html}{\out{</div>}}

Notice that you can also use character expressions in comparisons. This
convenience comes at a cost since character comparisons are slower than
numeric expressions. Unlike R, \code{as.numeric} or \code{as.integer} for these
logical statements is not only not needed, but will cause an syntax
error if you try to use the function.
}

\subsection{Supported functions}{

All the supported functions in rxode2 can be seen with the
\code{rxSupportedFuns()}.

A brief description of the built-in functions are in the following
table:\tabular{lll}{
   Function \tab Description \tab Aliases \cr
   gamma(x) \tab The Gamma function \tab gammafn \cr
   lgamma(x) \tab Natural logarithm of absolute value of gamma function \tab digamma \cr
   digamma(x) \tab First derivative of lgamma \tab  \cr
   trigamma(x) \tab Second derivative of lgamma \tab  \cr
   tetragamma(x) \tab Third derivative of lgamma \tab  \cr
   pentagamma(x) \tab Fourth derivative of lgamma \tab  \cr
   psigamma(x, deriv) \tab n-th derivative of Psi, the digamma function, which is the derivative of lgammafn. In other words, digamma(x) is the same as psigamma(x,0), trigamma(x) == psigamma(x,1), etc. \tab  \cr
   cospi(x) \tab cos(pi*x) \tab  \cr
   sinpi(x) \tab sin(pi*x) \tab  \cr
   tanpi(x) \tab tan(pi*x) \tab  \cr
   beta(a, b) \tab Beta function \tab  \cr
   lbeta(a, b) \tab log Beta function \tab  \cr
   bessel_i(x, nu, expo) \tab Bessel function type I with index nu \tab expo==1 is unscaled expo==2 is scaled by exp(-x) \cr
   bessel_j(x, nu) \tab Bessel function type J with index nu \tab  \cr
   bessel_k(x, ku, expo) \tab Bessel function type K with index nu \tab expo==1 is unscaled expo==2 is scaled by exp(x) \cr
   bessel_y(x, nu) \tab Bessel function type Y with index nu \tab  \cr
   R_pow(x, y) \tab x^y \tab  \cr
   R_pow_di(x, I) \tab x^y \tab y is an integer \cr
   log1pmx \tab log(1+x) - x \tab  \cr
   log1pexp \tab log(1+exp(x)) \tab  \cr
   expm1(x) \tab exp(x)-1 \tab  \cr
   lgamma1p(x) \tab log(gamma(x+1)) \tab  \cr
   sign(x) \tab Compute the signum function where sign(x) is 1, 0 -1 \tab  \cr
   fsign(x, y) \tab abs(x)*sign(y) \tab  \cr
   fprec(x, digits) \tab x rounded to digits (after the decimal point, used by signif() \tab  \cr
   fround(x, digits) \tab Round, used by R’s round() \tab  \cr
   ftrunc(x) \tab Truncated towards zero \tab  \cr
   abs(x) \tab absolute value of x \tab fabs \cr
   sin(x) \tab sine of x \tab  \cr
   cos(x) \tab cos of x \tab  \cr
   tan(x) \tab tan of x \tab  \cr
   factorial(x) \tab factorial of x \tab  \cr
   lfactorial(x) \tab log(factorial(x)) \tab  \cr
   log10(x) \tab log base 10 \tab  \cr
   log2(x) \tab log base 2 \tab  \cr
   pnorm(x) \tab Normal CDF of x \tab normcdf, phi \cr
   qnorm(x) \tab Normal pdf of x \tab norminv \cr
   probit(x, low=0, hi=1) \tab Probit (normal pdf) of x transforming into a range \tab  \cr
   probitInv(q, low=0, hi=1) \tab Inverse probit of x transforming into a range \tab  \cr
   acos(x) \tab Inverse cosine \tab  \cr
   asin(x) \tab Inverse sine \tab  \cr
   atan(x) \tab Inverse tangent \tab  \cr
   atan2(a, b) \tab Four quadrant inverse tangent \tab  \cr
   sinh(x) \tab Hyperbolic sine \tab  \cr
   cosh(x) \tab Hyperbolic cosine \tab  \cr
   tanh(x) \tab Hyperbolic tangent \tab  \cr
   floor(x) \tab Downward rounding \tab  \cr
   ceil(x) \tab Upward rounding \tab  \cr
   logit(x, low=0, hi=1) \tab Logit transformation of x transforming into a range \tab  \cr
   expit(x, low=0, hi=1) \tab expit transofmration in range \tab invLogit, logitInv \cr
   gammaq(a, z) \tab Normalized incomplete gamma from boost \tab  \cr
   gammaqInv(a, q) \tab Normalized incomplete gamma inverse from boost \tab  \cr
   ifelse(cond, trueValue, falseValue) \tab if else function \tab  \cr
   gammap(a, z) \tab Normalized lower incomplete gamma from boost \tab  \cr
   gammapInv(a, p) \tab Inverse of Normalized lower incomplete gamma from boost \tab  \cr
   gammapInva(x, p) \tab Inverse of Normalized lower incomplete gamma from boost \tab  \cr
   rxnorm(x) \tab Generate one deviate of from a normal distribution for each observation scale \tab  \cr
   rxnormV(x) \tab Generate one deviate from low discrepancy normal for each observation \tab  \cr
   rxcauchy \tab Generate one deviate from the cauchy distribution for each observation \tab  \cr
   rxchisq \tab Generate one deviate from the chisq distribution for each observation \tab  \cr
   rxexp \tab Generate one deviate from the exponential distribution for each observation \tab  \cr
   rxf \tab Generate one deviate from low discrepancy normal for each observation \tab  \cr
   rxgamma \tab Generate one deviate from the gamma distribution for each observation \tab  \cr
   rxbeta \tab Generate one deviate from the beta distribution for each observation \tab  \cr
   rxgeom \tab Generate one deviate from the geometric distribution for each observation \tab  \cr
   rxpois \tab Generate one deviate from the poission distribution for each observation \tab  \cr
   rxt \tab Generate one deviate from the t distribtuion for each observation \tab  \cr
   tad() or tad(x) \tab Time after dose (tad()) or time after dose for a compartment tad(cmt); no dose=NA \tab  \cr
   tad0() or tad0(x) \tab Time after dose (tad0()) or time after dose for a compartment tad0(cmt); no dose=0 \tab  \cr
   tafd() or tafd(x) \tab Time after first dose (tafd()) or time after first dose for a compartment tafd(cmt); no dose=NA \tab  \cr
   tafd0() or tafd0(x) \tab Time after first dose (tafd()) or time after first dose for a compartment tafd(cmt); no dose=NA \tab  \cr
   dosenum() \tab Dose Number \tab  \cr
   tlast() or tlast(cmt) \tab Time of Last dose; This takes into consideration any lag time, so if there is a dose at time 3 and a lag of 1, the time of last dose would be 4. tlast(cmt) calculates the time since last dose of a compartment; no dose=NA \tab  \cr
   tlast0() or tlast0(cmt) \tab Time of Last dose; This takes into consideration any lag time, so if there is a dose at time 3 and a lag of 1, the time of last dose would be 4. tlast(cmt) calculates the time since last dose of a compartment; no dose=0 \tab  \cr
   tfirst() or tfirst(cmt) \tab Time since first dose or time since first dose of a compartment; no dose=NA \tab  \cr
   tfirst0() or tfirst0(cmt) \tab Time since first dose or time since first dose of a compartment; no dose=0 \tab  \cr
   prod(…) \tab product of terms; This uses PreciseSums so the product will not have as much floating point errors (though it will take longer) \tab  \cr
   sum(…) \tab sum of terms; This uses PreciseSums so the product will not have as much floating point errors (though it will take longer) \tab  \cr
   max(…) \tab maximum of a group of numbers \tab  \cr
   min(…) \tab Min of a group of numbers \tab  \cr
   lag(parameter, number=1) \tab Get the lag of an input parameter; You can specify a number of lagged observations \tab  \cr
   lead(parameter, number=2) \tab Get the lead of an input parameter; You can specify a number of lead observation \tab  \cr
   diff(par, number=1) \tab Get the difference between the current parameter and the last parameter; Can change the parameter number \tab  \cr
   lag0(parameter, number=1) \tab Like lag() but returns 0 instead of NA when there is no prior value (the first record) \tab  \cr
   lead0(parameter, number=1) \tab Like lead() but returns 0 instead of NA when there is no following value (the last record) \tab  \cr
   diff0(par, number=1) \tab Like diff() but returns the value itself (previous value treated as 0) when there is no prior value \tab  \cr
   first(par) \tab Get the first value of an input parameter \tab  \cr
   last(par) \tab Get the last value of an input parameter \tab  \cr
   transit() \tab The transit compartment psuedo function \tab  \cr
   delay(state, T) \tab Value of an ODE state delayed by T time units (delay differential equations); requires a dense solver \tab  \cr
   is.na() \tab Determine if a value is NA \tab  \cr
   is.nan() \tab Determine if a value is NaN \tab  \cr
   is.infinite() \tab Check to see if the value is infinite \tab  \cr
   rinorm(x) \tab Generate one deviate of from a normal distribution for each individual \tab  \cr
   rinormV(x) \tab Generate one deviate from low discrepancy normal for each individual \tab  \cr
   ricauchy \tab Generate one deviate from the cauchy distribution for each individual \tab  \cr
   richisq \tab Generate one deviate from the chisq distribution for each individual \tab  \cr
   riexp \tab Generate one deviate from the exponential distribution for each individual \tab  \cr
   rif \tab Generate one deviate from low discrepancy normal for each individual \tab  \cr
   rigamma \tab Generate one deviate from the gamma distribution for each individual \tab  \cr
   ribeta \tab Generate one deviate from the beta distribution for each individual \tab  \cr
   rigeom \tab Generate one deviate from the geometric distribution for each individual \tab  \cr
   ropois \tab Generate one deviate from the poission distribution for each individual \tab  \cr
   rit \tab Generate one deviate from the t distribtuion for each individual \tab  \cr
   simeps \tab Simulate EPS from possibly truncated sigma matrix. Will take sigma matrix from the current study. Simulated at the very last moment. \tab  \cr
   simeta \tab Simulate ETA from possibly truncated omega matrix. Will take the omega matrix from the current study. Simulated at the initilization of the ODE system or the intialization of lhs \tab  \cr
}


Note that \verb{lag(cmt) =} is equivalent to \verb{alag(cmt) =} and not the same
as \verb{= lag(wt)}
}

\subsection{Reserved keywords}{

There are a few reserved keywords in a rxode2 model. They are in the
following table:\tabular{lll}{
   Reserved Name \tab Meaning \tab Alias \cr
   time \tab solver time \tab t \cr
   podo \tab In Transit compartment models, last dose amount \tab  \cr
   tlast \tab Time of Last dose \tab  \cr
   M_E \tab Exp(1) \tab  \cr
   M_LOG2E \tab log2(e) \tab  \cr
   M_LOG10E \tab log10(e) \tab  \cr
   M_LN2 \tab log(2) \tab  \cr
   M_LN10 \tab log(10) \tab  \cr
   M_PI \tab pi \tab  \cr
   M_PI_2 \tab pi/2 \tab  \cr
   M_PI_4 \tab pi/4 \tab  \cr
   M_1_PI \tab 1/pi \tab  \cr
   M_2_PI \tab 2/pi \tab  \cr
   M_2_SQRTPI \tab 2/sqrt(pi) \tab  \cr
   M_SQRT2 \tab sqrt(2) \tab  \cr
   M_SQRT1_2 \tab 1/sqrt(2) \tab  \cr
   M_SQRT_3 \tab sqrt(3) \tab  \cr
   M_SQRT_32 \tab sqrt(32) \tab  \cr
   M_LOG10_2 \tab Log10(2) \tab  \cr
   M_2PI \tab 2*pi \tab  \cr
   M_SQRT_PI \tab sqrt(pi) \tab  \cr
   M_1_SQRT_2PI \tab 1/(sqrt(2*pi)) \tab  \cr
   M_LN_SQRT_PI \tab log(sqrt(pi)) \tab  \cr
   M_LN_SQRT_2PI \tab log(sqrt(2*pi)) \tab  \cr
   M_LN_SQRT_PId2 \tab log(sqrt(pi/2)) \tab  \cr
   pi \tab pi \tab  \cr
   NA \tab R’s NA value \tab  \cr
   NaN \tab Not a Number Value \tab  \cr
   Inf \tab Infinite Value \tab  \cr
   newind \tab 1: First record of individual; 2: Subsequent record of individual \tab NEWIND \cr
   rxFlag \tab Flag for what part of the rxode2 model is being run; 1: ddt; 2: jac; 3: ini; 4: F; 5: lag; 6: rate; 7: dur; 8: mtime; 9: matrix exponential; 10: inductive linearization; 11: lhs \tab  \cr
}


Note that \code{rxFlag} will always output \code{11} or \code{calc_lhs} since that is
where the final variables are calculated, though you can tweak or test
certain parts of \code{rxode2} by using this flag.
}

\subsection{Residual functions when using rxode2 functions}{

In addition to \code{~} hiding output for certain types of output, it also is
used to specify a residual output or endpoint when the input is an
\code{rxode2} model function (that includes the residual in the \code{model({})}
block).

These specifications are of the form:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{var ~ add(add.sd)
}\if{html}{\out{</div>}}

Indicating the variable \code{var} is the variable that represents the
individual central tendencies of the model and it also represents the
compartment specification in the data-set.

You can also change the compartment name using the \code{|} syntax, that is:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{var ~ add(add.sd) | cmt
}\if{html}{\out{</div>}}

In the above case \code{var} represents the central tendency and \code{cmt}
represents the compartment or \code{dvid} specification.
\subsection{Transformations}{

For normal and related distributions, you can apply the transformation
on both sides by using some keywords/functions to apply these
transformations.\tabular{ll}{
   Transformation \tab rxode2/nlmixr2 code \cr
   Box-Cox \tab +boxCox(lambda) \cr
   Yeo-Johnson \tab +yeoJohnson(lambda) \cr
   logit-normal \tab +logitNorm(logit.sd, low, hi) \cr
   probit-normal \tab +probitNorm(probid.sd, low, hi) \cr
   log-normal \tab +lnorm(lnorm.sd) \cr
}


By default for the likelihood for all of these transformations is
calculated on the \strong{untransformed} scale.

For bounded variables like logit-normal or probit-normal the low and
high values are defaulted to 0 and 1 if missing.

For models where you wish to have a proportional model on one of these
transformation you can replace the standard deviation with \code{NA}

To allow for more transformations, \code{lnorm()}, \code{probitNorm()} and
\code{logitNorm()} can be combined the variance stabilizing \code{yeoJohnson()}
transformation.
}

\subsection{Normal and t-related distributions}{

For the normal and t-related distributions, we wanted to keep the
ability to use skewed distributions additive and proportional in the
t/cauchy-space, so these distributions are specified differently in
comparison to the other supported distributions within \code{nlmixr2}:\tabular{lll}{
   Distribution \tab How to Add \tab Example \cr
   Normal (log-likelihood) \tab +dnorm() \tab cc ~ add(add.sd) + dnorm() \cr
   T-distribution \tab +dt(df) \tab cc ~a dd(add.sd) + dt(df) \cr
   Cauchy (t with df=1) \tab +dcauchy() \tab cc ~ add(add.sd) + dcauchy() \cr
}


Note that with the normal and t-related distributions \code{nlmixr2} will
calculate \code{cwres} and \code{npde} under the normal assumption to help assess
the goodness of the fit of the model.

Also note that the \code{+dnorm()} is mostly for testing purposes and will
slow down the estimation procedure in \code{nlmixr2}. We suggest not adding
it (except for explicit testing). When there are multiple endpoint
models that mix non-normal and normal distributions, the whole problem
is shifted to a log-likelihood method for estimation in \code{nlmixr2}.
}

\subsection{Notes on additive + proportional models}{

There are two different ways to specify additive and proportional
models, which we will call \strong{combined1} and \strong{combined2}, the same way
that Monolix calls the two distributions (to avoid between software
differences in naming).

The first, \strong{combined1}, assumes that the additive and proportional
differences are on the standard deviation scale, or:

y=f+(a+b* f^c)*err

The second, \strong{combined2}, assumes that the additive and proportional
differences are combined on a variance scale:

y=f+$$sqrt(a^2+b^2 *f^(2c))$$*err

The default in \code{nlmixr2}/\code{rxode2} if not otherwise specified is
\strong{combined2} since it mirrors how adding 2 normal distributions in
statistics will add their variances (not the standard deviations).
However, the \strong{combined1} can describe the data possibly even better
than \strong{combined2} so both are possible options in \code{rxode2}/\code{nlmixr2}.
}

\subsection{Autocorrelated residuals (\code{ar()})}{

Adding \code{ar(cor)} to a normal residual makes successive residuals of the
same endpoint follow a continuous-time AR(1) process:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cp ~ add(add.sd) + ar(ar1.cor)
}\if{html}{\out{</div>}}

Here \code{ar1.cor} is the lag-one correlation and must be in \verb{[0, 1)}. The
correlation between two residuals separated by a time gap \code{dt} is
\code{cor^dt}, so irregular sampling is handled directly and the marginal
(stationary) residual variance is unchanged; the first observation of
each subject/endpoint uses the marginal distribution. \code{ar()} combines
with \code{add()}/\code{prop()} normal residuals and uses the same \verb{| cmt}
endpoint syntax as the other residual functions.

The \code{ar()} specification is identical in \code{rxode2} and \code{nlmixr2}: the
same model line simulates the autocorrelated residual in \code{rxode2} and
estimates the correlation in \code{nlmixr2} (the \code{nlm}, \code{focei}/\code{foce} and
\code{saem} families support it; \code{fo}, \code{foi}, \code{nlme} and \code{nls} do not). See
Karlsson et al. (1995) for the continuous-time AR(1) residual model.
}

\subsection{Distributions of known likelihoods}{

For residuals that are not related to normal, t-distribution or cauchy,
often the residual specification is of the form:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cmt ~ dbeta(alpha, beta)
}\if{html}{\out{</div>}}

Where the compartment specification is on the left handed side of the
specification.

For generalized likelihood you can specify:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ll(cmt) ~ llik specification
}\if{html}{\out{</div>}}
}

\subsection{Ordinal likelihoods}{

Finally, ordinal likelihoods/simulations can be specified in 2 ways. The
first is:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{err ~ c(p0, p1, p2)
}\if{html}{\out{</div>}}

Here \code{err} represents the compartment and \code{p0} is the probability of
being in a specific category:\tabular{ll}{
   Category \tab Probability \cr
   1 \tab p0 \cr
   2 \tab p1 \cr
   3 \tab p2 \cr
   4 \tab 1-p0-p1-p2 \cr
}


It is up to the model to ensure that the sum of the \code{p} values are less
than \code{1}. Additionally you can write an arbitrary number of categories
in the ordinal model described above.

It seems a little off that \code{p0} is the probability for category \code{1} and
sometimes scores are in non-whole numbers. This can be modeled as
follows:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{err ~ c(p0=0, p1=1, p2=2, 3)
}\if{html}{\out{</div>}}

Here the numeric categories are specified explicitly, and the
probabilities remain the same:\tabular{ll}{
   Category \tab Probability \cr
   0 \tab p0 \cr
   1 \tab p1 \cr
   2 \tab p2 \cr
   3 \tab 1-p0-p1-p2 \cr
}

}

\subsection{General table of supported residual distributions}{

In general all the that are supported are in the following table
(available in \code{rxode2::rxResidualError})\tabular{llllll}{
   Error model \tab Functional Form \tab Transformation \tab code \tab addProp \tab lhs \cr
   constant \tab  \tab None \tab var ~ add(add.sd) \tab  \tab response variable \cr
   proportional \tab  \tab None \tab var ~ prop(prop.sd) \tab  \tab response variable \cr
   power \tab  \tab None \tab var ~ pow(pow.sd, exponent) \tab  \tab response variable \cr
   additive+proportional \tab combined1 \tab None \tab var ~ add(add.sd) + prop(prop.sd) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional \tab combined2 \tab None \tab var ~ add(add.sd) + prop(prop.sd) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power \tab combined1 \tab None \tab var ~ add(add.sd) + pow(pow.sd, exponent) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power \tab combined2 \tab None \tab var ~ add(add.sd) + pow(pow.sd, exponent) + combined2() \tab addProp=2 \tab response variable \cr
   constant \tab  \tab log \tab var ~ lnorm(add.sd) \tab  \tab response variable \cr
   proportional \tab  \tab log \tab var ~ lnorm(NA) + prop(prop.sd) \tab  \tab response variable \cr
   power \tab  \tab log \tab var ~ lnorm(NA) + pow(pow.sd, exponent) \tab  \tab response variable \cr
   additive+proportional \tab combined1 \tab log \tab var ~ lnorm(add.sd) + prop(prop.sd) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional \tab combined2 \tab log \tab var ~ lnorm(add.sd) + prop(prop.sd) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power \tab combined1 \tab log \tab var ~ lnorm(add.sd) + pow(pow.sd, exponent) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power \tab combined2 \tab log \tab var ~ lnorm(add.sd) + pow(pow.sd, exponent) + combined2() \tab addProp=2 \tab response variable \cr
   constant \tab  \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) \tab  \tab response variable \cr
   proportional \tab  \tab boxCox \tab var ~ boxCox(lambda) + prop(prop.sd) \tab  \tab response variable \cr
   power \tab  \tab boxCox \tab var ~ boxCox(lambda) + pow(pow.sd, exponent) \tab  \tab response variable \cr
   additive+proportional \tab combined1 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + prop(prop.sd) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional \tab combined2 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + prop(prop.sd) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power \tab combined1 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + pow(pop.sd, exponent) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power \tab combined2 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + pow(pop.sd, exponent) + combined2() \tab addProp=2 \tab response variable \cr
   constant \tab  \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) \tab  \tab response variable \cr
   proportional \tab  \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + prop(prop.sd) \tab  \tab response variable \cr
   power \tab  \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + pow(pow.sd, exponent) \tab  \tab response variable \cr
   additive+proportional \tab combined1 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + prop(prop.sd) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional \tab combined2 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + prop(prop.sd) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power \tab combined1 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + pow(pop.sd, exponent) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power \tab combined2 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + pow(pop.sd, exponent) + combined2() \tab addProp=2 \tab response variable \cr
   constant \tab  \tab logit \tab var ~ logitNorm(logit.sd) \tab  \tab response variable \cr
   proportional \tab  \tab logit \tab var ~ logitNorm(NA) + prop(prop.sd) \tab  \tab response variable \cr
   power \tab  \tab logit \tab var ~ logitNorm(NA) + pow(pow.sd, exponent) \tab  \tab response variable \cr
   additive+proportional \tab combined1 \tab logit \tab var ~ logitNorm(logit.sd) + prop(prop.sd) \tab addProp=1 \tab response variable \cr
   additive+proportional \tab combined2 \tab logit \tab var ~ logitNorm(logit.sd) + prop(prop.sd) \tab addProp=2 \tab response variable \cr
   additive+power \tab combined1 \tab logit \tab var ~ logitNorm(logit.sd) + pow(pow.sd, exponent) \tab addProp=1 \tab response variable \cr
   additive+power \tab combined2 \tab logit \tab var ~ logitNorm(logit.sd) + pow(pow.sd, exponent) \tab addProp=2 \tab response variable \cr
   additive \tab  \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) \tab  \tab response variable \cr
   proportional \tab  \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(NA) + prop(prop.sd) \tab  \tab response variable \cr
   power \tab  \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(NA) + pow(pow.sd, exponent) \tab  \tab response variable \cr
   additive+proportional \tab combined1 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + prop(prop.sd) \tab addProp=1 \tab response variable \cr
   additive+proportional \tab combined2 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + prop(prop.sd) \tab addProp=2 \tab response variable \cr
   additive+power \tab combined1 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + pow(pow.sd, exponent) \tab addProp=1 \tab response variable \cr
   additive+power \tab combined2 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + pow(pow.sd, exponent) \tab addProp=2 \tab response variable \cr
   constant \tab  \tab logit \tab var ~ probitNorm(probit.sd) \tab  \tab response variable \cr
   proportional \tab  \tab probit \tab var ~ probitNorm(NA) + prop(prop.sd) \tab  \tab response variable \cr
   power \tab  \tab probit \tab var ~ probitNorm(NA) + pow(pow.sd, exponent) \tab  \tab response variable \cr
   additive+proportional \tab combined1 \tab probit \tab var ~ probitNorm(probit.sd) + prop(prop.sd) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional \tab combined2 \tab probit \tab var ~ probitNorm(probit.sd) + prop(prop.sd) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power \tab combined1 \tab probit \tab var ~ probitNorm(probit.sd) + pow(pow.sd, exponent) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power \tab combined2 \tab probit \tab var ~ probitNorm(probit.sd) + pow(pow.sd, exponent) + combined2() \tab addProp=2 \tab response variable \cr
   additive \tab  \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) \tab  \tab response variable \cr
   proportional \tab  \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(NA) + prop(prop.sd) \tab  \tab response variable \cr
   power \tab  \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(NA) + pow(pow.sd, exponent) \tab  \tab response variable \cr
   additive+proportional \tab combined1 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + prop(prop.sd) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional \tab combined2 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + prop(prop.sd) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power \tab combined1 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + pow(pow.sd, exponent) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power \tab combined2 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + pow(pow.sd, exponent) + combined2() \tab addProp=2 \tab response variable \cr
   constant+t \tab  \tab None \tab var ~ add(add.sd) + dt(df) \tab  \tab response variable \cr
   proportional+t \tab  \tab None \tab var ~ prop(prop.sd) + dt(df) \tab  \tab response variable \cr
   power+t \tab  \tab None \tab var ~ pow(pow.sd, exponent) + dt(df) \tab  \tab response variable \cr
   additive+proportional+t \tab combined1 \tab None \tab var ~ add(add.sd) + prop(prop.sd) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+t \tab combined2 \tab None \tab var ~ add(add.sd) + prop(prop.sd) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+t \tab combined1 \tab None \tab var ~ add(add.sd) + pow(pow.sd, exponent) + dt(df) +combined1() \tab addProp=1 \tab response variable \cr
   additive+power+t \tab combined2 \tab None \tab var ~ add(add.sd) + pow(pow.sd, exponent) + dt(df) +combined2() \tab addProp=2 \tab response variable \cr
   constant+t \tab  \tab log \tab var ~ lnorm(add.sd) + dt(df) \tab  \tab response variable \cr
   proportional+t \tab  \tab log \tab var ~ lnorm(NA) + prop(prop.sd) + dt(df) \tab  \tab response variable \cr
   power+t \tab  \tab log \tab var ~ lnorm(NA) + pow(pow.sd, exponent) + dt(df) \tab  \tab response variable \cr
   additive+proportional+t \tab combined1 \tab log \tab var ~ lnorm(add.sd) + prop(prop.sd) + dt(df) +combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+t \tab combined2 \tab log \tab var ~ lnorm(add.sd) + prop(prop.sd) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+t \tab combined1 \tab log \tab var ~ lnorm(add.sd) + pow(pow.sd, exponent) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+t \tab combined2 \tab log \tab var ~ lnorm(add.sd) + pow(pow.sd, exponent) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   constant+t \tab  \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd)+dt(df) \tab  \tab response variable \cr
   proportional+t \tab  \tab boxCox \tab var ~ boxCox(lambda) + prop(prop.sd)+dt(df) \tab  \tab response variable \cr
   power+t \tab  \tab boxCox \tab var ~ boxCox(lambda) + pow(pow.sd, exponent)+dt(df) \tab  \tab response variable \cr
   additive+proportional+t \tab combined1 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + prop(prop.sd) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+t \tab combined2 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + prop(prop.sd) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+t \tab combined1 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + pow(pop.sd, exponent) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+t \tab combined2 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + pow(pop.sd, exponent) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   constant+t \tab  \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + dt(df) \tab  \tab response variable \cr
   proportional+t \tab  \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + prop(prop.sd) + dt(df) \tab  \tab response variable \cr
   power+t \tab  \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + pow(pow.sd, exponent) + dt(df) \tab  \tab response variable \cr
   additive+proportional+t \tab combined1 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + prop(prop.sd) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+t \tab combined2 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + prop(prop.sd) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+t \tab combined1 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + pow(pop.sd, exponent) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+t \tab combined2 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + pow(pop.sd, exponent) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   constant+t \tab  \tab logit \tab var ~ logitNorm(logit.sd)+dt(df) \tab  \tab response variable \cr
   proportional+t \tab  \tab logit \tab var ~ logitNorm(NA) + prop(prop.sd)+dt(df) \tab  \tab response variable \cr
   power+t \tab  \tab logit \tab var ~ logitNorm(NA) + pow(pow.sd, exponent) + dt(df) \tab  \tab response variable \cr
   additive+proportional+t \tab combined1 \tab logit \tab var ~ logitNorm(logit.sd) + prop(prop.sd) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+t \tab combined2 \tab logit \tab var ~ logitNorm(logit.sd) + prop(prop.sd) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+t \tab combined1 \tab logit \tab var ~ logitNorm(logit.sd) + pow(pow.sd, exponent) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+t \tab combined2 \tab logit \tab var ~ logitNorm(logit.sd) + pow(pow.sd, exponent) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   additive+t \tab  \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + dt(df) \tab  \tab response variable \cr
   proportional+t \tab  \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(NA) + prop(prop.sd) + dt(df) \tab  \tab response variable \cr
   power+t \tab  \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(NA) + pow(pow.sd, exponent) + dt(df) \tab  \tab response variable \cr
   additive+proportional+t \tab combined1 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + prop(prop.sd) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+t \tab combined2 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + prop(prop.sd) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+t \tab combined1 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + pow(pow.sd, exponent) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+t \tab combined2 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + pow(pow.sd, exponent) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   constant+t \tab  \tab logit \tab var ~ probitNorm(probit.sd) + dt(df) \tab  \tab response variable \cr
   proportional+t \tab  \tab probit \tab var ~ probitNorm(NA) + prop(prop.sd) + dt(df) \tab  \tab response variable \cr
   power+t \tab  \tab probit \tab var ~ probitNorm(NA) + pow(pow.sd, exponent) + dt(df) \tab  \tab response variable \cr
   additive+proportional+t \tab combined1 \tab probit \tab var ~ probitNorm(probit.sd) + prop(prop.sd) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+t \tab combined2 \tab probit \tab var ~ probitNorm(probit.sd) + prop(prop.sd) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+t \tab combined1 \tab probit \tab var ~ probitNorm(probit.sd) + pow(pow.sd, exponent) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+t \tab combined2 \tab probit \tab var ~ probitNorm(probit.sd) + pow(pow.sd, exponent) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   additive+t \tab  \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + dt(df) \tab  \tab response variable \cr
   proportional+t \tab  \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(NA) + prop(prop.sd) + dt(df) \tab  \tab response variable \cr
   power+t \tab  \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(NA) + pow(pow.sd, exponent) + dt(df) \tab  \tab response variable \cr
   additive+proportional+t \tab combined1 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + prop(prop.sd) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+t \tab combined2 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + prop(prop.sd) + dt(df) + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+t \tab combined1 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + pow(pow.sd, exponent) + dt(df) + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+t \tab combined2 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + pow(pow.sd, exponent) + dt(df) +combined2() \tab addProp=2 \tab response variable \cr
   constant+cauchy \tab  \tab None \tab var ~ add(add.sd) + dcauchy() \tab  \tab response variable \cr
   proportional+cauchy \tab  \tab None \tab var ~ prop(prop.sd) + dcauchy() \tab  \tab response variable \cr
   power+cauchy \tab  \tab None \tab var ~ pow(pow.sd, exponent) + dcauchy() \tab  \tab response variable \cr
   additive+proportional+cauchy \tab combined1 \tab None \tab var ~ add(add.sd) + prop(prop.sd) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+cauchy \tab combined2 \tab None \tab var ~ add(add.sd) + prop(prop.sd) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+cauchy \tab combined1 \tab None \tab var ~ add(add.sd) + pow(pow.sd, exponent) + dcauchy() +combined1() \tab addProp=1 \tab response variable \cr
   additive+power+cauchy \tab combined2 \tab None \tab var ~ add(add.sd) + pow(pow.sd, exponent) + dcauchy() +combined2() \tab addProp=2 \tab response variable \cr
   constant+cauchy \tab  \tab log \tab var ~ lnorm(add.sd) + dcauchy() \tab  \tab response variable \cr
   proportional+cauchy \tab  \tab log \tab var ~ lnorm(NA) + prop(prop.sd) + dcauchy() \tab  \tab response variable \cr
   power+cauchy \tab  \tab log \tab var ~ lnorm(NA) + pow(pow.sd, exponent) + dcauchy() \tab  \tab response variable \cr
   additive+proportional+cauchy \tab combined1 \tab log \tab var ~ lnorm(add.sd) + prop(prop.sd) + dcauchy() +combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+cauchy \tab combined2 \tab log \tab var ~ lnorm(add.sd) + prop(prop.sd) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+cauchy \tab combined1 \tab log \tab var ~ lnorm(add.sd) + pow(pow.sd, exponent) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+cauchy \tab combined2 \tab log \tab var ~ lnorm(add.sd) + pow(pow.sd, exponent) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   constant+cauchy \tab  \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd)+dcauchy() \tab  \tab response variable \cr
   proportional+cauchy \tab  \tab boxCox \tab var ~ boxCox(lambda) + prop(prop.sd)+dcauchy() \tab  \tab response variable \cr
   power+cauchy \tab  \tab boxCox \tab var ~ boxCox(lambda) + pow(pow.sd, exponent)+dcauchy() \tab  \tab response variable \cr
   additive+proportional+cauchy \tab combined1 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + prop(prop.sd) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+cauchy \tab combined2 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + prop(prop.sd) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+cauchy \tab combined1 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + pow(pop.sd, exponent) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+cauchy \tab combined2 \tab boxCox \tab var ~ boxCox(lambda) + add(add.sd) + pow(pop.sd, exponent) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   constant+cauchy \tab  \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + dcauchy() \tab  \tab response variable \cr
   proportional+cauchy \tab  \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + prop(prop.sd) + dcauchy() \tab  \tab response variable \cr
   power+cauchy \tab  \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + pow(pow.sd, exponent) + dcauchy() \tab  \tab response variable \cr
   additive+proportional+cauchy \tab combined1 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + prop(prop.sd) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+cauchy \tab combined2 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + prop(prop.sd) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+cauchy \tab combined1 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + pow(pop.sd, exponent) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+cauchy \tab combined2 \tab yeoJohnson \tab var ~ yeoJohnson(lambda) + add(add.sd) + pow(pop.sd, exponent) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   constant+cauchy \tab  \tab logit \tab var ~ logitNorm(logit.sd)+dcauchy() \tab  \tab response variable \cr
   proportional+cauchy \tab  \tab logit \tab var ~ logitNorm(NA) + prop(prop.sd)+dcauchy() \tab  \tab response variable \cr
   power+cauchy \tab  \tab logit \tab var ~ logitNorm(NA) + pow(pow.sd, exponent) + dcauchy() \tab  \tab response variable \cr
   additive+proportional+cauchy \tab combined1 \tab logit \tab var ~ logitNorm(logit.sd) + prop(prop.sd) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+cauchy \tab combined2 \tab logit \tab var ~ logitNorm(logit.sd) + prop(prop.sd) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+cauchy \tab combined1 \tab logit \tab var ~ logitNorm(logit.sd) + pow(pow.sd, exponent) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+cauchy \tab combined2 \tab logit \tab var ~ logitNorm(logit.sd) + pow(pow.sd, exponent) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   additive+cauchy \tab  \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + dcauchy() \tab  \tab response variable \cr
   proportional+cauchy \tab  \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(NA) + prop(prop.sd) + dcauchy() \tab  \tab response variable \cr
   power+cauchy \tab  \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(NA) + pow(pow.sd, exponent) + dcauchy() \tab  \tab response variable \cr
   additive+proportional+cauchy \tab combined1 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + prop(prop.sd) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+cauchy \tab combined2 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + prop(prop.sd) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+cauchy \tab combined1 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + pow(pow.sd, exponent) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+cauchy \tab combined2 \tab yeoJohnson(logit()) \tab var ~ yeoJohnson(lambda) + logitNorm(logit.sd) + pow(pow.sd, exponent) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   constant+cauchy \tab  \tab logit \tab var ~ probitNorm(probit.sd) + dcauchy() \tab  \tab response variable \cr
   proportional+cauchy \tab  \tab probit \tab var ~ probitNorm(NA) + prop(prop.sd) + dcauchy() \tab  \tab response variable \cr
   power+cauchy \tab  \tab probit \tab var ~ probitNorm(NA) + pow(pow.sd, exponent) + dcauchy() \tab  \tab response variable \cr
   additive+proportional+cauchy \tab combined1 \tab probit \tab var ~ probitNorm(probit.sd) + prop(prop.sd) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+cauchy \tab combined2 \tab probit \tab var ~ probitNorm(probit.sd) + prop(prop.sd) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+cauchy \tab combined1 \tab probit \tab var ~ probitNorm(probit.sd) + pow(pow.sd, exponent) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+cauchy \tab combined2 \tab probit \tab var ~ probitNorm(probit.sd) + pow(pow.sd, exponent) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   additive+cauchy \tab  \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + dcauchy() \tab  \tab response variable \cr
   proportional+cauchy \tab  \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(NA) + prop(prop.sd) + dcauchy() \tab  \tab response variable \cr
   power+cauchy \tab  \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(NA) + pow(pow.sd, exponent) + dcauchy() \tab  \tab response variable \cr
   additive+proportional+cauchy \tab combined1 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + prop(prop.sd) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+proportional+cauchy \tab combined2 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + prop(prop.sd) + dcauchy() + combined2() \tab addProp=2 \tab response variable \cr
   additive+power+cauchy \tab combined1 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + pow(pow.sd, exponent) + dcauchy() + combined1() \tab addProp=1 \tab response variable \cr
   additive+power+cauchy \tab combined2 \tab yeoJohnson(probit()) \tab var ~ yeoJohnson(lambda) + probitNorm(probit.sd) + pow(pow.sd, exponent) + dcauchy() +combined2() \tab addProp=2 \tab response variable \cr
   autocorrelated (constant) \tab  \tab None \tab var ~ add(add.sd) + ar(cor) \tab  \tab response variable \cr
   autocorrelated (additive+proportional) \tab  \tab None \tab var ~ add(add.sd) + prop(prop.sd) + ar(cor) \tab  \tab response variable \cr
   poission \tab  \tab none \tab cmt ~ dpois(lamba) \tab  \tab compartment specification \cr
   binomial \tab  \tab none \tab cmt ~ dbinom(n, p) \tab  \tab compartment specification \cr
   beta \tab  \tab none \tab cmt ~ dbeta(alpha, beta) \tab  \tab compartment specification \cr
   chisq \tab  \tab none \tab cmt ~ dchisq(nu) \tab  \tab compartment specification \cr
   exponential \tab  \tab none \tab cmt ~ dexp(r) \tab  \tab compartment specification \cr
   uniform \tab  \tab none \tab cmt ~ dunif(a, b) \tab  \tab compartment specification \cr
   weibull \tab  \tab none \tab cmt ~ dweibull(a, b) \tab  \tab compartment specification \cr
   gamma \tab  \tab none \tab cmt ~ dgamma(a, b) \tab  \tab compartment specification \cr
   geometric \tab  \tab none \tab cmt ~ dgeom(a) \tab  \tab compartment specification \cr
   negative binomial form #1 \tab  \tab none \tab cmt ~ dnbinom(n, p) \tab  \tab compartment specification \cr
   negative binomial form #2 \tab  \tab none \tab cmt ~ dnbinomMu(size, mu) \tab  \tab compartment specification \cr
   ordinal probability \tab  \tab none \tab cmt ~ c(p0=0, p1=1, p2=2, 3) \tab  \tab compartment specification \cr
   log-likelihood \tab  \tab none \tab ll(cmt) ~ log likelihood expression \tab  \tab likelihood + compartment expression \cr
}

}

}
}
\section{Note on strings in rxode2}{
Strings are converted to double values inside of \code{rxode2}, hence you can
refer to them as an integer corresponding to the string value or the
string value itself. For covariates these are calculated on the fly
based on your data and you should likely not try this, though you should
be aware. For strings defined in the model, this is fixed and both could
be used.

For example:

\if{html}{\out{<div class="sourceCode">}}\preformatted{if (APGAR == 10 || APGAR == 8 || APGAR == 9) \{
    tAPGAR <- "High"
  \} else if (APGAR == 1 || APGAR == 2 || APGAR == 3) \{
    tAPGAR <- "Low"
  \} else if (APGAR == 4 || APGAR == 5 || APGAR == 6 || APGAR == 7) \{
    tAPGAR <- "Med"
  \} else \{
    tAPGAR<- "Med"
  \}
}\if{html}{\out{</div>}}

Could also be replaced by:

\if{html}{\out{<div class="sourceCode">}}\preformatted{if (APGAR == 10 || APGAR == 8 || APGAR == 9) \{
    tAPGAR <- "High"
  \} else if (APGAR == 1 || APGAR == 2 || APGAR == 3) \{
    tAPGAR <- "Low"
  \} else if (APGAR == 4 || APGAR == 5 || APGAR == 6 || APGAR == 7) \{
    tAPGAR <- "Med"
  \} else \{
    tAPGAR<- 3
  \}
}\if{html}{\out{</div>}}

Since \code{"Med"} is already defined

If you wanted you can pre-declare what levels it has (and the order) to
give you better control of this:

\if{html}{\out{<div class="sourceCode">}}\preformatted{levels(tAPGAR) <- c("Med", "Low", "High")
if (APGAR == 10 || APGAR == 8 || APGAR == 9) \{
    tAPGAR <- 3
  \} else if (APGAR == 1 || APGAR == 2 || APGAR == 3) \{
    tAPGAR <- 2
  \} else if (APGAR == 4 || APGAR == 5 || APGAR == 6 || APGAR == 7) \{
    tAPGAR <- 1
  \} else \{
    tAPGAR<- 1
  \}
}\if{html}{\out{</div>}}

You can see that the number changed since the declaration change the
numbers in each variable for \code{tAPGAR}. These \code{levels()} statements need
to be declared before the variable occurs to ensure the numbering is
consistent with what is declared.
}

\examples{
\donttest{

mod <- function() {
  ini({
    KA   <- .291
    CL   <- 18.6
    V2   <- 40.2
    Q    <- 10.5
    V3   <- 297.0
    Kin  <- 1.0
    Kout <- 1.0
    EC50 <- 200.0
  })
  model({
    # A 4-compartment model, 3 PK and a PD (effect) compartment
    # (notice state variable names 'depot', 'centr', 'peri', 'eff')
    C2 <- centr/V2
    C3 <- peri/V3
    d/dt(depot) <- -KA*depot;
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3;
    d/dt(peri)  <-                    Q*C2 - Q*C3;
    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff;
    eff(0)      <- 1
  })
}

m1 <- rxode2(mod)
print(m1)

# Step 2 - Create the model input as an EventTable,
# including dosing and observation (sampling) events

# QD (once daily) dosing for 5 days.

qd <- et(amountUnits = "ug", timeUnits = "hours") |>
  et(amt = 10000, addl = 4, ii = 24)

# Sample the system hourly during the first day, every 8 hours
# then after
qd <- qd |> et(0:24) |>
  et(from = 24 + 8, to = 5 * 24, by = 8)

# Step 3 - solve the system

qd.cp <- rxSolve(m1, qd)

head(qd.cp)

}

}
\references{
Chamber, J. M. and Temple Lang, D. (2001)
\emph{Object Oriented Programming in R}.
R News, Vol. 1, No. 3, September 2001.
\url{https://cran.r-project.org/doc/Rnews/Rnews_2001-3.pdf}.

Hindmarsh, A. C.
\emph{ODEPACK, A Systematized Collection of ODE Solvers}.
Scientific Computing, R. S. Stepleman et al. (Eds.),
North-Holland, Amsterdam, 1983, pp. 55-64.

Petzold, L. R.
\emph{Automatic Selection of Methods for Solving Stiff and Nonstiff
Systems of Ordinary Differential Equations}.
Siam J. Sci. Stat. Comput. 4 (1983), pp. 136-148.

Hairer, E., Norsett, S. P., and Wanner, G.
\emph{Solving ordinary differential equations I, nonstiff problems}.
2nd edition, Springer Series in Computational Mathematics,
Springer-Verlag (1993).

Plevyak, J.
\emph{\code{dparser}}, \url{https://dparser.sourceforge.net/}. Web. 12 Oct. 2015.
}
\seealso{
\code{\link[=eventTable]{eventTable()}}, \code{\link[=et]{et()}}, \code{\link[=add.sampling]{add.sampling()}}, \code{\link[=add.dosing]{add.dosing()}}
}
\author{
Melissa Hallow, Wenping Wang and Matthew Fidler
}
\concept{Nonlinear regression}
\concept{ODE models}
\concept{Ordinary differential equations}
\concept{Pharmacodynamics (PD)}
\concept{Pharmacokinetics (PK)}
\keyword{models}
\keyword{nonlinear}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikCauchy}
\alias{llikCauchy}
\title{log likelihood of Cauchy distribution and it's derivatives (from stan)}
\usage{
llikCauchy(x, location = 0, scale = 1, full = FALSE)
}
\arguments{
\item{x}{Observation}

\item{location, scale}{location and scale parameters.}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the log pdf value of with
\code{dLocation} and \code{dScale} that has the derivatives with respect to the parameters at
the observation time-point
}
\description{
log likelihood of Cauchy distribution and it's derivatives (from stan)
}
\details{
In an \code{rxode2()} model, you can use \code{llikCauchy()} but you have to
use all arguments.  You can also get the derivative of \code{location} and \code{scale} with
\code{llikCauchyDlocation()} and \code{llikCauchyDscale()}.
}
\examples{
\donttest{
x <- seq(-3, 3, length.out = 21)

llikCauchy(x, 0, 1)

llikCauchy(x, 3, 1, full=TRUE)

et <- et(-3, 3, length.out=10)
et$location <- 0
et$scale <- 1

model <- function() {
  model({
    fx <- llikCauchy(time, location, scale)
    dLocation <- llikCauchyDlocation(time, location, scale)
    dScale <- llikCauchyDscale(time, location, scale)
  })
}

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxSetupIni}
\alias{rxSetupIni}
\title{Setup the initial conditions.}
\usage{
rxSetupIni(obj, inits = NULL)
}
\arguments{
\item{obj}{rxode2 object}

\item{inits}{A numeric vector of initial conditions.}
}
\value{
initial conditions that were setup
}
\description{
Setup the initial conditions.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.rxModelVars}
\alias{print.rxModelVars}
\title{Print Values}
\usage{
\method{print}{rxModelVars}(x, ...)
}
\arguments{
\item{x}{an object used to select a method.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
This returns invisibly the model variables object
}
\description{
\code{print} prints its argument and returns it \emph{invisibly} (via
  \code{\link[base]{invisible}(x)}).  It is a generic function which means that
  new printing methods can be easily added for new \code{\link[base]{class}}es.
}
\details{
The default method, \code{\link[base]{print.default}} has its own help page.
  Use \code{\link{methods}("print")} to get all the methods for the
  \code{print} generic.

  \code{print.factor} allows some customization and is used for printing
  \code{\link[base]{ordered}} factors as well.

  \code{print.table} for printing \code{\link[base]{table}}s allows other
  customization.  As of R 3.0.0, it only prints a description in case of a table
  with 0-extents (this can happen if a classifier has no valid data).

  See \code{\link[base]{noquote}} as an example of a class whose main
  purpose is a specific \code{print} method.
}
\examples{
require(stats)

ts(1:20)  #-- print is the "Default function" --> print.ts(.) is called
for(i in 1:3) print(1:i)

## Printing of factors
attenu$station ## 117 levels -> 'max.levels' depending on width

## ordered factors: levels  "l1 < l2 < .."
esoph$agegp[1:12]
esoph$alcgp[1:12]

## Printing of sparse (contingency) tables
set.seed(521)
t1 <- round(abs(rt(200, df = 1.8)))
t2 <- round(abs(rt(200, df = 1.4)))
table(t1, t2) # simple
print(table(t1, t2), zero.print = ".") # nicer to read

## same for non-integer "table":
T <- table(t2,t1)
T <- T * (1+round(rlnorm(length(T)))/4)
print(T, zero.print = ".") # quite nicer,
print.table(T[,2:8] * 1e9, digits=3, zero.print = ".")
## still slightly inferior to  Matrix::Matrix(T)  for larger T

## Corner cases with empty extents:
table(1, NA) # < table of extent 1 x 0 >
}
\references{
{\if{html}{\out{<span id="reference+print.Rd+R+3AChambers+2BHastie+3A1992"></span>}}Chambers JM, Hastie TJ (1992).
\emph{Statistical Models in S}.
Chapman & Hall, London.
ISBN 9780412830402.}
}
\seealso{
The default method \code{\link[base]{print.default}}, and help for the
  methods above; further \code{\link[base]{options}}, \code{\link[base]{noquote}}.

  For more customizable (but cumbersome) printing, see
  \code{\link[base]{cat}}, \code{\link[base]{format}} or also \code{\link[base]{write}}.
  For a simple prototypical print method, see
  \code{\link[tools]{.print.via.format}} in package \pkg{tools}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxUdfUiIniLhs}
\alias{rxUdfUiIniLhs}
\title{Return the lhs parsed language expression}
\usage{
rxUdfUiIniLhs()
}
\value{
lhs language expression or NULL
}
\description{
Return the lhs parsed language expression
}
\examples{

rxUdfUiIniLhs()

}
\seealso{
Other User functions:
\code{\link[=linMod]{linMod()}},
\code{\link[=rxUdfUiControl]{rxUdfUiControl()}},
\code{\link[=rxUdfUiData]{rxUdfUiData()}},
\code{\link[=rxUdfUiEst]{rxUdfUiEst()}},
\code{\link[=rxUdfUiMv]{rxUdfUiMv()}},
\code{\link[=rxUdfUiNum]{rxUdfUiNum()}},
\code{\link[=rxUdfUiParsing]{rxUdfUiParsing()}}
}
\author{
Matthew L. Fidler
}
\concept{User functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxExpandGrid_}
\alias{rxExpandGrid_}
\title{Expand grid internal function}
\usage{
rxExpandGrid_(c1, c2, type)
}
\arguments{
\item{c1}{character vector of items to be expanded}

\item{c2}{second character vector of items to be expanded}

\item{type}{0 for a typical data frame, 1 for symengine sensitivity expansion}
}
\value{
data frame (when type = 0) or symengine string (when type=1)
}
\description{
Expand grid internal function
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err.R
\name{.rxBlankIni}
\alias{.rxBlankIni}
\title{Get a blank, theta1, or eta1 initialization block for iniDf}
\usage{
.rxBlankIni(type = c("empty", "theta", "eta"))
}
\arguments{
\item{type}{type of initialization block to return}
}
\value{
A data.frame with the appropriate number/type of columns.

For type="empty", the data.frame will have 0 rows but all the correct types.

For type="theta", the data.frame will have 1 row with the correct
types and default values.  The "name" and "est" will likely need to
be updated.

For type="eta", the data.frame will have 1 row with the correct
types and default values for the a single eta being added.  The
"name" and "est" will likely need to be updated.
}
\description{
Get a blank, theta1, or eta1 initialization block for iniDf
}
\examples{

.rxBlankIni("empty")

.rxBlankIni("theta")

.rxBlankIni("eta")

}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxRepR0_}
\alias{rxRepR0_}
\title{Rep R0 for foce}
\usage{
rxRepR0_(neta)
}
\arguments{
\item{neta}{ETA to substitute}
}
\value{
Returns a string of R code to substitute the rx_r expression in the symengine environment .s
}
\description{
Rep R0 for foce
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.rxode2}
\alias{print.rxode2}
\title{Print information about the rxode2 object.}
\usage{
\method{print}{rxode2}(x, ...)
}
\arguments{
\item{x}{An rxode2 object}

\item{...}{Ignored parameters}
}
\value{
original object
}
\description{
This prints the model name and its status for being able to be solved
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{etSeq}
\alias{etSeq}
\alias{seq.rxEt}
\title{Sequence of event tables}
\usage{
etSeq(..., samples = c("clear", "use"), waitII = c("smart", "+ii"), ii = 24)

\method{seq}{rxEt}(...)
}
\arguments{
\item{...}{The event tables and optionally time between event
tables, called waiting times in this help document.}

\item{samples}{How to handle samples when repeating an event
table.  The options are:
\itemize{
\item \code{"clear"} Clear sampling records before combining the datasets
\item \code{"use"} Use the sampling records when combining the datasets
}}

\item{waitII}{This determines how waiting times between events are
handled. The options are:
\itemize{
\item \code{"smart"} This "smart" handling of waiting times is the
default option.  In this case, if the waiting time is above the
last observed inter-dose interval in the first combined event
table, then the actual time between doses is given by the wait
time.  If it is smaller than the last observed inter-dose
interval, the time between event tables is given by the inter-dose
interval + the waiting time between event tables.
\item \code{"+ii"} In this case, the wait time is added to the
inter-dose interval no matter the length of the wait time or
inter-dose interval
}}

\item{ii}{If there was no inter-dose intervals found in the event
table, assume that the interdose interval is given by this
\code{ii} value.  By default this is \code{24}.}
}
\value{
An event table
}
\description{
This combines a sequence of event tables.
}
\details{
This \code{seq}uences all the event tables in added in the
argument list \code{...}.  By default when combining the event
tables the offset is at least by the last inter-dose interval in
the prior event table (or \code{ii}).  If you separate any of the
event tables by a number, the event tables will be separated at
least the wait time defined by that number or the last inter-dose
interval.
}
\examples{

\dontrun{

library(rxode2)
library(units)

# Model from rxode2 tutorial
# Using a nlmixr2 style function

mod1 <-function(){
  ini({
    KA <- 2.94E-01
    CL <- 1.86E+01
    V2 <- 4.02E+01
    Q <- 1.05E+01
    V3 <- 2.97E+02
    Kin <- 1
    Kout <- 1
    EC50 <- 200
  })
 model({
    C2 <- centr/V2
    C3 <- peri/V3
    d/dt(depot) <- -KA*depot
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
    d/dt(peri)  <-                    Q*C2 - Q*C3
    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
 })
}

## These are making the more complex regimens of the rxode2 tutorial

## bid for 5 days
bid <- et(timeUnits="hr") |>
       et(amt=10000,ii=12,until=set_units(5, "days"))

## qd for 5 days
qd <- et(timeUnits="hr") |>
      et(amt=20000,ii=24,until=set_units(5, "days"))

## bid for 5 days followed by qd for 5 days

et <- seq(bid,qd) |>
      et(seq(0,11*24,length.out=100))

bidQd <- rxSolve(mod1, et)

plot(bidQd, C2)


## Now Infusion for 5 days followed by oral for 5 days

##  note you can dose to a named compartment instead of using the compartment number
infusion <- et(timeUnits = "hr") |>
      et(amt=10000, rate=5000, ii=24, until=set_units(5, "days"), cmt="centr")


qd <- et(timeUnits = "hr") |>
  et(amt=10000, ii=24, until=set_units(5, "days"), cmt="depot")

et <- seq(infusion,qd)

infusionQd <- rxSolve(mod1, et)

plot(infusionQd, C2)

## 2wk-on, 1wk-off

qd <- et(timeUnits = "hr") |>
      et(amt=10000, ii=24, until=set_units(2, "weeks"), cmt="depot")

et <- seq(qd, set_units(1,"weeks"), qd) |>
     add.sampling(set_units(seq(0, 5.5,by=0.005),weeks))

wkOnOff <- rxSolve(mod1, et)

plot(wkOnOff, C2)

## You can also repeat the cycle easily with the rep function

qd <-et(timeUnits = "hr") |>
     et(amt=10000, ii=24, until=set_units(2, "weeks"), cmt="depot")

et <- etRep(qd, times=4, wait=set_units(1,"weeks")) |>
      add.sampling(set_units(seq(0, 12.5,by=0.005),weeks))

repCycle4 <- rxSolve(mod1, et)

plot(repCycle4, C2)

}
}
\references{
Wang W, Hallow K, James D (2015). "A Tutorial on rxode2: Simulating
Differential Equation Pharmacometric Models in R." CPT:
Pharmacometrics and Systems Pharmacology, 5(1), 3-10. ISSN 2163-8306
}
\seealso{
\code{\link{eventTable}}, \code{\link{add.sampling}},
\code{\link{add.dosing}}, \code{\link{et}},
\code{\link{etRep}}, \code{\link{etRbind}},
\code{\link{rxode2}}
}
\author{
Matthew L Fidler, Wenping Wang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eventSens.R
\name{rxEventSensDeactivate}
\alias{rxEventSensDeactivate}
\title{Turn off the rxode2 event-sensitivity jump injection}
\usage{
rxEventSensDeactivate()
}
\value{
invisibly \code{NULL}.
}
\description{
Resets the runtime gate set by \code{\link[=rxEventSensLoadModel]{rxEventSensLoadModel()}} (active = 0) after a
direct C++ solve loop completes, so later unrelated solves are unaffected.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikBeta}
\alias{llikBeta}
\title{Calculate the log likelihood of the binomial function (and its derivatives)}
\usage{
llikBeta(x, shape1, shape2, full = FALSE)
}
\arguments{
\item{x}{Observation}

\item{shape1, shape2}{non-negative parameters of the Beta distribution.}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the log pdf value of with
\code{dShape1} and \code{dShape2} that has the derivatives with respect to the parameters at
the observation time-point
}
\description{
Calculate the log likelihood of the binomial function (and its derivatives)
}
\details{
In an \code{rxode2()} model, you can use \code{llikBeta()} but you have to
use all arguments.  You can also get the derivative of \code{shape1} and \code{shape2} with
\code{llikBetaDshape1()} and \code{llikBetaDshape2()}.
}
\examples{

\donttest{

x <- seq(1e-4, 1 - 1e-4, length.out = 21)

llikBeta(x, 0.5, 0.5)

llikBeta(x, 1, 3, TRUE)

et <- et(seq(1e-4, 1-1e-4, length.out=21))
et$shape1 <- 0.5
et$shape2 <- 1.5

model <- function() {
  model({
    fx <- llikBeta(time, shape1, shape2)
    dShape1 <- llikBetaDshape1(time, shape1, shape2)
    dShape2 <- llikBetaDshape2(time, shape1, shape2)
  })
}

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err.R
\name{rxErrTypeCombine}
\alias{rxErrTypeCombine}
\title{Combine transformations and error structures}
\usage{
rxErrTypeCombine(oldErrType, newErrType)
}
\arguments{
\item{oldErrType}{This is the old transformation, by default is
zero representing no prior transformation. This parameter is
first to allow piping. When the parameter \code{addTransform} is
missing and \code{oldErrType} is a character value, this functions
swaps \code{oldErrType} and \code{addTransform} and assigns
\code{oldErrType} to zero assuming that there is no prior
distribution.}

\item{newErrType}{This is the new distribution that is being
"added" to the current transformation.  These assumes the inputs
are in the preferred distribution name, as determined by
\code{rxPreferredDistributionName()}}
}
\value{
The new transformation as a factor
}
\description{
Combine error information to figure out what transformation is
being applied for the current endpoint
}
\examples{

rxErrTypeCombine("probitNorm")

rxErrTypeCombine("probitNorm") |>
  rxErrTypeCombine("boxCox")


}
\author{
Matthew Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.uiArg}
\alias{.uiArg}
\title{Handle arguments for ui functions}
\usage{
.uiArg(char, f, dp)
}
\arguments{
\item{char}{This is the character equivalent of the argument}

\item{f}{This is the forced equivalent of the argument}

\item{dp}{This is deparsed expression}
}
\value{
character representing the underlying rxode2 code for the argument
}
\description{
Note this is an internal function but it is exported in case it is
useful.
}
\examples{

.uiArg("1.0", 1.0, "1.0")
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxLock}
\alias{rxLock}
\alias{rxUnlock}
\title{Lock/unlocking of rxode2 dll file}
\usage{
rxLock(obj)

rxUnlock(obj)
}
\arguments{
\item{obj}{A rxode2 family of objects}
}
\value{
nothing; called for side effects
}
\description{
Lock/unlocking of rxode2 dll file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err-foceiBase.R
\name{.rxGetHiBoundaryPred1AndIni}
\alias{.rxGetHiBoundaryPred1AndIni}
\title{Get the upper boundary condition when the transformation it}
\usage{
.rxGetHiBoundaryPred1AndIni(env, pred1)
}
\arguments{
\item{env}{Environment for the parsed model}

\item{pred1}{The \code{data.frame} of the current error}
}
\value{
Upper Boundary
}
\description{
Get the upper boundary condition when the transformation it
}
\author{
Matthew Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikPois}
\alias{llikPois}
\title{log-likelihood for the Poisson distribution}
\usage{
llikPois(x, lambda, full = FALSE)
}
\arguments{
\item{x}{non negative integers}

\item{lambda}{non-negative means}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the pdf value of with
\code{dLambda} that has the derivatives with respect to the parameters at
the observation time-point
}
\description{
log-likelihood for the Poisson distribution
}
\details{
In an \code{rxode2()} model, you can use \code{llikPois()} but you have to
use all arguments.  You can also get the derivatives with
\code{llikPoisDlambda()}
}
\examples{
\donttest{
llikPois(0:7, lambda = 1)

llikPois(0:7, lambda = 4, full=TRUE)

# In rxode2 you can use:

et <- et(0:10)
et$lambda <- 0.5

model <- function() {
  model({
    fx <- llikPois(time, lambda)
    dLambda <- llikPoisDlambda(time, lambda)
  })
}

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikWeibull}
\alias{llikWeibull}
\title{log likelihood and derivatives for Weibull distribution}
\usage{
llikWeibull(x, shape, scale, full = FALSE)
}
\arguments{
\item{x}{variable distributed by a Weibull distribution}

\item{shape, scale}{shape and scale parameters, the latter defaulting to 1.}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the log pdf value of with \code{dProb}
that has the derivatives with respect to the \code{prob} parameters at
the observation time-point
}
\description{
log likelihood and derivatives for Weibull distribution
}
\details{
In an \code{rxode2()} model, you can use \code{llikWeibull()} but you have to
use the x and rate arguments.  You can also get the derivative of \code{shape} or \code{scale} with
\code{llikWeibullDshape()} and \code{llikWeibullDscale()}.
}
\examples{
\donttest{
llikWeibull(1, 1, 10)

# rxode2 can use this too:

et  <- et(seq(0.001, 1, length.out=10))
et$shape <- 1
et$scale <- 10
 
model <- function() {
  model({
    fx <- llikWeibull(time, shape, scale)
    dShape<- llikWeibullDshape(time, shape, scale)
    dScale <- llikWeibullDscale(time, shape, scale)
  })
}

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjoint.R
\name{rxSolveAdjoint}
\alias{rxSolveAdjoint}
\title{Solve a model with adjoint (backward) sensitivities, drop-in for forward sens}
\usage{
rxSolveAdjoint(
  object,
  params,
  events,
  calcSens,
  adjStates = NULL,
  denseBy = 0.01,
  atol = 1e-08,
  rtol = 1e-06,
  ...
)
}
\arguments{
\item{object}{model definition accepted by \code{\link[=rxode2]{rxode2()}}.}

\item{params}{named numeric vector of parameter values.}

\item{events}{event table / data used to define dosing and sampling times;
sampling (\code{evid==0}) times become the sensitivity output times.}

\item{calcSens}{character vector of parameter names to differentiate with
respect to.}

\item{adjStates}{character vector of output states of interest; defaults to
all ODE states (full forward-sensitivity parity).}

\item{denseBy, atol, rtol}{passed to the adjoint checkpoint solve; see
\code{\link[=.rxAdjointSolveEvalC]{.rxAdjointSolveEvalC()}}.}

\item{...}{additional arguments passed to the primal \code{\link[=rxSolve]{rxSolve()}} call.}
}
\value{
the standard \code{rxSolve()} output (as \code{returnType="data.frame"}) with
\verb{rx__sens_<state>_BY_<param>__} columns appended.
}
\description{
Solves \code{object} as \code{\link[=rxSolve]{rxSolve()}} would, then appends
\verb{rx__sens_<state>_BY_<param>__} columns computed via adjoint (backward)
sensitivity analysis.  Same column names and output structure as
\code{rxSolve(object, ..., calcSens=)}.  Prefer \code{\link[=.rxAdjointGrad]{.rxAdjointGrad()}} when only a
scalar objective gradient is needed.
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.rxDocTable}
\alias{.rxDocTable}
\title{Print out a table in the documentation}
\usage{
.rxDocTable(table, caption = "none")
}
\arguments{
\item{table}{data frame}

\item{caption}{a character vector representing the caption for the latex table}
}
\value{
based on the \code{knitr} context:
\itemize{
\item output a \code{kableExtra::kbl} for \code{latex} output
\item output a \code{DT::datatable} for html output
\item otherwise output a \code{knitr::kable}
}
}
\description{
Print out a table in the documentation
}
\examples{
.rxDocTable(rxReservedKeywords)
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxFormat.R
\name{rxHtml}
\alias{rxHtml}
\alias{rxHtml.rxSolve}
\title{Format rxSolve and related objects as html.}
\usage{
rxHtml(x, ...)

\method{rxHtml}{rxSolve}(x, ...)
}
\arguments{
\item{x}{rxode2 object}

\item{...}{Extra arguments sent to kable}
}
\value{
html code for rxSolve object
}
\description{
Format rxSolve and related objects as html.
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{.vecDf}
\alias{.vecDf}
\title{Convert numeric vector to repeated data.frame}
\usage{
.vecDf(vec, n)
}
\arguments{
\item{vec}{Named input vector}

\item{n}{Number of columns}
}
\value{
Data frame with repeated vec
}
\description{
Convert numeric vector to repeated data.frame
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxsolve.R
\name{rxIsDense}
\alias{rxIsDense}
\title{Check whether an ODE solving method supports dense output}
\usage{
rxIsDense(method)
}
\arguments{
\item{method}{A character vector of method names or an integerish vector of
method codes (as returned by \code{\link[=odeMethodToInt]{odeMethodToInt()}}).  Vectorised.}
}
\value{
A logical vector the same length as \code{method}.  \code{TRUE} if the
corresponding method supports dense output.
}
\description{
Dense output (continuous interpolation) allows the solver to reconstruct
the solution at any point within a completed step without extra function
evaluations.  This lets the solver take large internal steps that span many
requested output times, then interpolate cheaply -- improving both speed and
accuracy on densely sampled grids.
}
\details{
Dense-capable single methods are:
\code{"dop853"} (0), \code{"dop5"} (10), \code{"bs"} (11), \code{"ros4"} (13).

For composite AutoSwitch methods (e.g. \code{"dop5+ros4"}), \code{TRUE} is returned
only when \strong{both} the primary and stiff secondary are dense-capable.
\code{"ros4"} is the only stiff method with dense support, so the valid dense
composites are \code{"dop853+ros4"}, \code{"dop5+ros4"}, and \code{"bs+ros4"}.
}
\examples{
rxIsDense("dop853")                       # TRUE
rxIsDense("ros4")                         # TRUE
rxIsDense("liblsoda")                     # FALSE
rxIsDense(c("dop5", "bs", "cvode"))       # TRUE TRUE FALSE
rxIsDense("dop5+ros4")                    # TRUE  (both dense)
rxIsDense("dop5+ros43")                   # FALSE (ros43 not dense)

}
\seealso{
\code{\link[=rxIsStiff]{rxIsStiff()}}, \code{\link[=rxIsNonStiff]{rxIsNonStiff()}}, \code{\link[=odeMethodToInt]{odeMethodToInt()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-fix.R
\name{rxFixPop}
\alias{rxFixPop}
\title{Apply the fixed population estimated parameters}
\usage{
rxFixPop(ui, returnNull = FALSE)
}
\arguments{
\item{ui}{rxode2 ui function}

\item{returnNull}{boolean for if unchanged values should return the
original ui (\code{FALSE}) or null (\code{TRUE})}
}
\value{
when \code{returnNull} is TRUE, NULL if nothing was changed, or
the changed model ui.  When \code{returnNull} is FALSE, return a ui no
matter if it is changed or not.
}
\description{
Apply the fixed population estimated parameters
}
\examples{

One.comp.transit.allo <- function() {
 ini({
   # Where initial conditions/variables are specified
   lktr <- log(1.15)  #log k transit (/h)
   lcl  <- log(0.15)  #log Cl (L/hr)
   lv   <- log(7)     #log V (L)
   ALLC <- fix(0.75)  #allometric exponent cl
   ALLV <- fix(1.00)  #allometric exponent v
   prop.err <- 0.15   #proportional error (SD/mean)
   add.err <- 0.6     #additive error (mg/L)
   eta.ktr ~ 0.5
   eta.cl ~ 0.1
   eta.v ~ 0.1
 })
 model({
   #Allometric scaling on weight
   cl <- exp(lcl + eta.cl + ALLC * logWT70)
   v  <- exp(lv + eta.v + ALLV * logWT70)
   ktr <- exp(lktr + eta.ktr)
   # RxODE-style differential equations are supported
   d/dt(depot)   = -ktr * depot
   d/dt(central) =  ktr * trans - (cl/v) * central
   d/dt(trans)   =  ktr * depot - ktr * trans
   ## Concentration is calculated
   cp = central/v
   # And is assumed to follow proportional and additive error
   cp ~ prop(prop.err) + add(add.err)
 })
}

m <- rxFixPop(One.comp.transit.allo)
m

# now everything is already fixed, so calling again will do nothing

rxFixPop(m)

# if you call it with returnNull=TRUE when no changes have been
# performed, the function will return NULL

rxFixPop(m, returnNull=TRUE)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elu.R
\name{dELU}
\alias{dELU}
\alias{d2ELU}
\alias{d2aELU}
\alias{dELUa}
\alias{d2ELUa}
\title{Derivatives of the Exponential Linear Unit (ELU) Activation Function}
\usage{
dELU(x, alpha = 1)

d2ELU(x, alpha = 1)

d2aELU(x, alpha = 1)

dELUa(x, alpha = 1)

d2ELUa(x, alpha = 1)
}
\arguments{
\item{x}{A numeric vector. All elements must be finite and
non-missing.}

\item{alpha}{A numeric scalar. All elements must be finite and
non-missing.}
}
\value{
A numeric vector where the derivative(s) of the ELU function has been applied
to each element of \code{x}.
}
\description{
Derivatives of the Exponential Linear Unit (ELU) Activation Function
}
\examples{
dELU(c(-1, 0, 1, 2), 2)
d2ELU(c(-1, 0, 1, 2), 2)
d2aELU(c(-1, 0, 1, 2), 2)
dELUa(c(-1, 0, 1, 2), 2)
d2ELUa(c(-1, 0, 1, 2), 2)

\donttest{
# Can also be used in rxode2:
r <- rxode2({
  r1=dELU(time, 2)
  r2=d2ELU(time, 2)
  r2a=d2aELU(time, 2)
  ra=dELUa(time, 2)
  r2a=d2ELUa(time, 2)
})

e <- et(c(-1, 0, 1, 2))
rxSolve(r, e)
}
}
\seealso{
Other Activation Functions:
\code{\link[=ELU]{ELU()}},
\code{\link[=GELU]{GELU()}},
\code{\link[=PReLU]{PReLU()}},
\code{\link[=ReLU]{ReLU()}},
\code{\link[=SELU]{SELU()}},
\code{\link[=Swish]{Swish()}},
\code{\link[=dGELU]{dGELU()}},
\code{\link[=dPReLU]{dPReLU()}},
\code{\link[=dReLU]{dReLU()}},
\code{\link[=dSELU]{dSELU()}},
\code{\link[=dSwish]{dSwish()}},
\code{\link[=dlReLU]{dlReLU()}},
\code{\link[=dsoftplus]{dsoftplus()}},
\code{\link[=lReLU]{lReLU()}},
\code{\link[=softplus]{softplus()}}
}
\author{
Matthew L. Fidler
}
\concept{Activation Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxC}
\alias{rxC}
\title{Return the C file associated with the rxode2 object}
\usage{
rxC(obj)
}
\arguments{
\item{obj}{A rxode2 family of objects or a character string of the
model specification or location of a file with a model
specification.}
}
\value{
a path of the library
}
\description{
This will return C code for generating the rxode2 DLL.
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eventSens.R
\name{rxEventSensLoadModel}
\alias{rxEventSensLoadModel}
\title{Point the rxode2 event-sensitivity globals at a jump-sensitivity model}
\usage{
rxEventSensLoadModel(model)
}
\arguments{
\item{model}{A built jump-sensitivity model (carrying \code{eventSensInfo}).}
}
\value{
invisibly \code{TRUE} when the jumps were activated, \code{FALSE} otherwise
(fd model / no sensitivities).
}
\description{
For downstream packages (e.g. nlmixr2est's FOCEi) that solve a sensitivity
model through a direct C++ \code{ind_solve()} loop, bypassing \code{rxSolve()}: sets
rxode2's event ("jump") sensitivity function pointers and runtime dims to
\code{model} and turns the jumps on.  The jump blocks are bounds-guarded, so
smaller models solved afterwards skip the injection safely.  Pair with
\code{rxEventSensDeactivate()} after the run.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piping-model.R, R/ui.R
\name{model.function}
\alias{model.function}
\alias{model.rxUi}
\alias{model.rxode2}
\alias{model.rxModelVars}
\alias{model}
\alias{model.default}
\title{Model block for rxode2/nlmixr models}
\usage{
\method{model}{`function`}(
  x,
  ...,
  append = NULL,
  auto = getOption("rxode2.autoVarPiping", TRUE),
  cov = NULL,
  envir = parent.frame()
)

\method{model}{rxUi}(
  x,
  ...,
  append = NULL,
  auto = getOption("rxode2.autoVarPiping", TRUE),
  cov = NULL,
  envir = parent.frame()
)

\method{model}{rxode2}(
  x,
  ...,
  append = NULL,
  auto = getOption("rxode2.autoVarPiping", TRUE),
  cov = NULL,
  envir = parent.frame()
)

\method{model}{rxModelVars}(
  x,
  ...,
  append = NULL,
  auto = getOption("rxode2.autoVarPiping", TRUE),
  cov = NULL,
  envir = parent.frame()
)

model(
  x,
  ...,
  append = FALSE,
  auto = getOption("rxode2.autoVarPiping", TRUE),
  cov = NULL,
  envir = parent.frame()
)

\method{model}{default}(x, ..., append = FALSE, cov = NULL, envir = parent.frame())
}
\arguments{
\item{x}{model expression}

\item{...}{Other arguments}

\item{append}{This is a boolean to determine if the lines are
appended in piping.  The possible values for this is:
\itemize{
\item \code{TRUE} which is when the lines are appended to the model instead of replaced
\item \code{FALSE} when the lines are replaced in the model (default)
\item \code{NA} is when the lines are pre-pended to the model instead of replaced
\item \verb{lhs expression}, which will append the lines after the last observed line of the expression \code{lhs}
}}

\item{auto}{This boolean tells if piping automatically selects the
parameters should be characterized as a population parameter,
between subject variability, or a covariate.  When \code{TRUE} this
automatic selection occurs.  When \code{FALSE} this automatic
selection is turned off and everything is added as a covariate
(which can be promoted to a parameter with the \code{ini} statement).
By default this is \code{TRUE}, but it can be changed by
\code{options(rxode2.autoVarPiping=FALSE)}.}

\item{cov}{is a character vector of variables that should be
assumed to be covariates.  This will override automatic promotion
to a population parameter estimate (or an eta)}

\item{envir}{the \code{environment} in which unevaluated model
expressions is to be evaluated.  May also be \code{NULL}, a list, a
data frame, a pairlist or an integer as specified to \code{sys.call}.}
}
\value{
Model block with ini information included.  \code{ini} must be
called before \code{model} block
}
\description{
Model block for rxode2/nlmixr models
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxDfdy}
\alias{rxDfdy}
\title{Jacobian and parameter derivatives}
\usage{
rxDfdy(obj)
}
\arguments{
\item{obj}{rxode2 family of objects}
}
\value{
A list of the jacobian parameters defined in this rxode2
object.
}
\description{
Return Jacobain and parameter derivatives
}
\seealso{
Other Query model information:
\code{\link[=rxInits]{rxInits()}},
\code{\link[=rxLhs]{rxLhs()}},
\code{\link[=rxModelVars]{rxModelVars()}},
\code{\link[=rxParams]{rxParams()}},
\code{\link[=rxState]{rxState()}}
}
\author{
Matthew L. Fidler
}
\concept{Query model information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxResidualError.R
\docType{data}
\name{rxResidualError}
\alias{rxResidualError}
\title{A description of Rode2 supported residual errors}
\format{
A data frame with 6 columns and 183 rows
\describe{
\item{Error model}{A description of the type of residual error}
\item{Functional Form}{For additive and proportional what functional form is used}
\item{Transformation}{The type of transformation that is done on the DV and the prediction}
\item{code}{Example code for the residual error type}
\item{addProp}{The type of add+prop residual error default that would be equivalent}
\item{lhs}{what the left handed side of the specification represents, either a response variable, or a compartment specification}
}
}
\usage{
rxResidualError
}
\description{
A description of Rode2 supported residual errors
}
\references{
The transformation-based and autocorrelated (\code{ar()}, AR(1)) residual
error models follow Karlsson MO, Beal SL, Sheiner LB. Three new residual
error models for population PK/PD analyses. J Pharmacokinet Biopharm.
1995;23(6):651-672. doi:10.1007/BF02353466.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{coef.rxode2}
\alias{coef.rxode2}
\title{Return the rxode2 coefficients}
\usage{
\method{coef}{rxode2}(object, ...)
}
\arguments{
\item{object}{is an rxode2 object}

\item{...}{ignored arguments}
}
\value{
a rxCoef object with the following
\itemize{
\item \code{params}  is a list of strings for parameters for the rxode2 object
\item \code{state} is a list of strings for the names of each state in
the rxode2 object.
\item \code{ini} is the model specified default values for the
parameters.
\item \code{rxode2} is the referring rxode2 object
}
}
\description{
This returns the parameters , state variables
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{rxSetSeed}
\alias{rxSetSeed}
\title{Set the parallel seed for rxode2 random number generation}
\usage{
rxSetSeed(seed)
}
\arguments{
\item{seed}{An integer that represents the rxode2 parallel and
internal random number generator seed.  When positive, use this
seed for random number generation and increment and reseed any
parallel or new engines that are being called. When negative,
turn off the rxode2 seed and generate a seed from the R's uniform
random number generator.  Best practice is to set this seed.}
}
\value{
Nothing, called for its side effects
}
\description{
This sets the seed for the rxode2 parallel random number generation.
If set, then whenever a seed is set for the threefry or
vandercorput simulation engine, it will use this seed, increment
for the number of seeds and continue with the sequence the next
time the random number generator is called.
}
\details{
In contrast, when this is not called, the time that the
vandercorput or threefry simulation engines are seeded it comes
from a uniform random number generated from the standard R random
seed.  This may cause a duplicate seed based on the R seed state.
This means that there could be correlations between simulations
that do not exist This will avoid the birthday problem picking
exactly the same seed using the seed state of the R random number
generator.  The more times the seed is called, the more likely this
becomes.
}
\examples{

rxSetSeed(42)

# seed with generator 42
rxnorm()

# Use R's random number generator
rnorm(1)

rxSetSeed(42)

# reproduces the same number
rxnorm()

# But R's random number is not the same

rnorm(1)

# If we reset this to use the R's seed
# (internally rxode2 uses a uniform random number to span seeds)
# This can lead to duplicate sequences and seeds

rxSetSeed(-1)

# Now set seed works for both.

# This is not recommended, but illustrates the different types of
# seeds that can be generated.

set.seed(42)

rxnorm()

rnorm(1)

set.seed(42)

rxnorm()

rnorm(1)

}
\references{
JD Cook. (2016). Random number generator seed mistakes.
\url{https://www.johndcook.com/blog/2016/01/29/random-number-generator-seed-mistakes/}
}
\seealso{
rxGetSeed, rxWithSeed, rxWithPreserveSeed
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxsolve.R
\name{rxIsAutoSwitch}
\alias{rxIsAutoSwitch}
\title{Check whether an ODE method string is an AutoSwitch composite pair}
\usage{
rxIsAutoSwitch(method)
}
\arguments{
\item{method}{Character vector of method names.}
}
\value{
Logical vector the same length as \code{method}.
}
\description{
Returns \code{TRUE} for strings containing \code{"+"}, such as \code{"dop853+ros43"}.
}
\examples{
rxIsAutoSwitch("dop853+ros43")  # TRUE
rxIsAutoSwitch("dop853")        # FALSE
rxIsAutoSwitch(c("dop853+ros43", "liblsoda"))  # TRUE FALSE

}
\seealso{
\code{\link[=odeMethodToInt]{odeMethodToInt()}}, \code{\link[=rxIsNonStiff]{rxIsNonStiff()}}, \code{\link[=rxIsStiff]{rxIsStiff()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.rxIsLinCmt}
\alias{.rxIsLinCmt}
\title{Internal function to tell if the linCmt() is the model variables}
\usage{
.rxIsLinCmt()
}
\value{
0 or 1
}
\description{
Internal function to tell if the linCmt() is the model variables
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evidPush.R
\name{evid_}
\alias{evid_}
\title{Push a future dose or observation event from within an rxode2 model}
\usage{
evid_(time, evid, amt, cmt = 1, rate = 0, ii = 0, addl = 0, ss = 0)
}
\arguments{
\item{time}{Numeric. The time at which the event should occur.  Must
be greater than the current model time \code{t}; events in the past are
silently counted and a warning is issued after solving.}

\item{evid}{Integer event ID.  Follows NONMEM/rxode2 conventions:
\describe{
\item{\code{0}}{Observation (adds an output row, no dose).}
\item{\code{1}}{Dose (bolus or infusion depending on \code{rate}).}
\item{\code{2}}{Other type observation (passed through).}
\item{\code{3}}{Reset all compartments.}
\item{\code{4}}{Reset then dose.}
\item{\code{5}}{Replace compartment amount.}
\item{\code{6}}{Multiply compartment amount.}
\item{\code{7}}{Phantom/transit dose.}
\item{\verb{>= 100}}{Classic rxode2 internal evid; passed through verbatim.}
}}

\item{amt}{Numeric dose amount (for dose events) or \code{0} for
observations.  When \code{rate > 0} this is interpreted as the total
infusion amount and the infusion duration is \code{amt / rate}.}

\item{cmt}{Integer compartment number (1-based) to which the dose is
applied.  Default is 1}

\item{rate}{Numeric infusion rate.
\describe{
\item{\code{0}}{Bolus dose (default).}
\item{\verb{> 0}}{Fixed infusion rate; duration = \code{amt / rate}.}
\item{\code{-1}}{Rate defined by the model (\verb{rate_<cmt>} variable).}
\item{\code{-2}}{Duration defined by the model (\verb{dur_<cmt>} variable).}
}}

\item{ii}{Numeric inter-dose interval.  Used together with \code{addl} to
schedule repeat doses at \code{time}, \code{time + ii}, \code{time + 2*ii}, ...,
\code{time + addl*ii}.  Also required when \code{ss > 0}. Default 0}

\item{addl}{Integer number of \emph{additional} doses beyond the first.
The total number of doses pushed is \code{addl + 1}, spaced \code{ii} apart.
Each dose is pushed as a standalone event (not as a periodic
schedule in the event table).}

\item{ss}{Integer steady-state flag applied to the \emph{first} dose only
(\code{addl} repetitions always use \code{ss = 0}).
\describe{
\item{\code{0}}{No steady-state (default).}
\item{\code{1}}{Steady-state additive: add SS solution to current state.}
\item{\code{2}}{Steady-state replace: replace current state with SS solution.}
}}
}
\value{
This function is only meaningful inside an rxode2 model; it
returns \code{NULL} invisibly if called from R directly (after signaling
an error).
}
\description{
\code{evid_()} is a model-only function that schedules a future dosing or
observation event during ODE solving.  It is evaluated at each output
time point and inserts the requested event into the individual's event
timeline for future processing.

This function has no meaning outside an rxode2 model block and will
throw an error if called directly from R.
}
\details{
\subsection{Behavior inside a model}{

\code{evid_()} is evaluated at every output time point (when the solver is
exactly at a scheduled event time).  The pushed event is inserted into
the individual's event timeline and the solver visits it at the
specified future time.

The number of events that may be pushed per individual is limited by
the \code{maxExtra} argument of \code{\link[=rxSolve]{rxSolve()}}.  When \code{maxExtra = 0}
(the default) there is no limit.  Exceeding the limit causes an
error.

Past-time pushes (where \code{time <= t}) are silently ignored and counted;
a warning is issued after solving.
}

\subsection{Relationship to NONMEM event columns}{

The argument order and names mirror the standard NONMEM dataset
columns: \code{TIME}, \code{EVID}, \code{AMT}, \code{CMT}, \code{RATE}, \code{II}, \code{ADDL}, \code{SS}.
}
}
\examples{
\donttest{
# Push a single bolus of 50 mg to compartment 1 at t + 12
m <- rxode2({
  d/dt(depot)   <- -ka * depot
  d/dt(central) <- ka * depot - cl / vd * central
  cp <- central / vd
  if (t < 24) {
    evid_(t + 12, 1, 50, 1, 0, 0, 0, 0)
  }
})

# Push three boluses (addl = 2) at t+6, t+18, t+30 (ii = 12)
m2 <- rxode2({
  d/dt(depot)   <- -ka * depot
  d/dt(central) <- ka * depot - cl / vd * central
  cp <- central / vd
  if (t < 1) {
    evid_(t + 6, 1, 50, 1, 0, 12, 2, 0)
  }
})
}

}
\seealso{
\code{\link[=rxSolve]{rxSolve()}} for the \code{maxExtra} control argument.  Other
more convenient ways to interact with adaptive observations and
events include \code{\link[=bolus]{bolus()}} \code{\link[=infuse]{infuse()}}, \code{\link[=infuseDur]{infuseDur()}},
\code{\link[=reset]{reset()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxNumLoaded}
\alias{rxNumLoaded}
\title{Get the number of loaded rxode2 DLLs}
\usage{
rxNumLoaded()
}
\value{
Number of loaded rxode2 DLLs
}
\description{
Get the number of loaded rxode2 DLLs
}
\examples{
rxNumLoaded()
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.getWh}
\alias{.getWh}
\title{Get the internal breakdown of an evid}
\usage{
.getWh(i)
}
\arguments{
\item{i}{evid to breakdown}
}
\value{
named evid integer vector
}
\description{
Get the internal breakdown of an evid
}
\examples{

.getWh(1001)
.getWh(10401)

}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symengine.R
\name{rxS}
\alias{rxS}
\title{Load a model into a symengine environment}
\usage{
rxS(x, doConst = TRUE, promoteLinSens = FALSE, envir = parent.frame())
}
\arguments{
\item{x}{rxode2 object}

\item{doConst}{Load constants into the environment as well.}

\item{promoteLinSens}{Promote solved linear compartment systems to
sensitivity-based solutions.}

\item{envir}{default is \code{NULL}; Environment to put symengine
variables in.}
}
\value{
rxode2/symengine environment
}
\description{
Load a model into a symengine environment
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{.udfEnvSetUdf}
\alias{.udfEnvSetUdf}
\title{Use the udf model variable information to get the environment where
the functions exists}
\usage{
.udfEnvSetUdf(udf)
}
\arguments{
\item{udf}{modelVars$udf, integer vector with NA_integer_ for the
address of the environment where the functions exist}
}
\value{
nothing called for side effects
}
\description{
Use the udf model variable information to get the environment where
the functions exists
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{rxnormV}
\alias{rxnormV}
\alias{rxnorm}
\title{Simulate random normal variable from threefry generator}
\usage{
rxnormV(mean = 0, sd = 1, n = 1L, ncores = 1L)

rxnorm(mean = 0, sd = 1, n = 1L, ncores = 1L)
}
\arguments{
\item{mean}{vector of means.}

\item{sd}{vector of standard deviations.}

\item{n}{number of observations}

\item{ncores}{Number of cores for the simulation

\code{rxnorm} simulates using the threefry sitmo generator.

\code{rxnormV} used to simulate with the vandercorput simulator, but
since it didn't satisfy the normal properties it was changed to simple be
an alias of \code{rxnorm}. It is no longer supported in \code{rxode2({})} blocks}
}
\value{
normal random number deviates
}
\description{
Simulate random normal variable from threefry generator
}
\examples{
\donttest{
## Use threefry engine

rxnorm(n = 10) # with rxnorm you have to explicitly state n
rxnorm(n = 10, ncores = 2) # You can parallelize the simulation using openMP

rxnorm(2, 3) ## The first 2 arguments are the mean and standard deviation


## This example uses `rxnorm` directly in the model

rx <- function() {
  model({
    a <- rxnorm()
  })
}

et <- et(1, id = 1:2)

s <- rxSolve(rx, et)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.rxWithSink}
\alias{.rxWithSink}
\alias{.rxWithSinkBoth}
\title{With one sink, then release}
\usage{
.rxWithSink(file, code)

.rxWithSinkBoth(file, code)
}
\arguments{
\item{file}{the path to the file sink while running the \code{code}}

\item{code}{The code to run during the sink}
}
\value{
Will return the results of the \code{code} section
}
\description{
With one sink, then release
}
\details{
\code{.rxWithSink} captures output from \code{cat}

\code{.rxWithSinkBoth} captures output from \code{cat} and \code{message}
}
\examples{

t <- tempfile()
.rxWithSink(t, cat("message\n"))
cat("cat2\n") # now you can see the cat2
lines <- readLines(t)
unlink(t)
}
\author{
Matthew Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxReq.R
\name{rxReq}
\alias{rxReq}
\title{Require namespace, otherwise throw error.}
\usage{
rxReq(pkg)
}
\arguments{
\item{pkg}{Package required for function to work.}
}
\value{
Nothing
}
\description{
Require namespace, otherwise throw error.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxJacobian.R
\name{.rxFwdSensJacBlock}
\alias{.rxFwdSensJacBlock}
\title{Full-system Jacobian for a first-order forward-sensitivity model, reusing F_X}
\usage{
.rxFwdSensJacBlock(model, state, params)
}
\arguments{
\item{model}{symengine model environment (post \code{.rxJacobian})}

\item{state}{base ODE state names}

\item{params}{sensitivity parameter names (\code{calcSens})}
}
\value{
character vector of \code{df(state)/dy(state)} lines for the full system
}
\description{
Builds the full df()/dy() block for the sensitivity-expanded system (base
states + \verb{rx__sens_*} variational compartments) from the \code{F_X}/\code{F_p} pieces
\code{.rxJacobian(model)} already produced.  The base x base and sens x sens blocks
are \code{F_X} (reused, no differentiation); only the sens x base block needs new
\code{symengine::D} and it vanishes for state-independent (linear) systems.
Requires \code{.rxJacobian(model)} (vars=TRUE) to have run.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splitBolus.R
\name{splitBolus}
\alias{splitBolus}
\title{Split bolus doses across compartments inside an rxode2 model}
\usage{
splitBolus(cmt, ...)
}
\arguments{
\item{cmt}{Source compartment name.}

\item{...}{Target compartment names. Provide at least one target
compartment.}
}
\value{
This function is only meaningful inside an rxode2 model; it
errors when called directly from R.
}
\description{
\code{splitBolus()} is a model-only directive that rewrites bolus doses
aimed at \code{cmt} into parallel bolus doses to the target compartments.
Each target compartment receives the full original amount. The
source-compartment bolus is replaced, not retained.

This rewrite applies both to event tables translated by \code{\link[=etTrans]{etTrans()}}
and to future doses scheduled during solving with \code{\link[=evid_]{evid_()}}.

The source compartment may also appear in the target list, but the
target compartments in \code{...} must be unique.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{rxDerived}
\alias{rxDerived}
\title{Calculate derived parameters for the 1-, 2-, and 3- compartment
linear models.}
\usage{
rxDerived(..., verbose = FALSE, digits = 0)
}
\arguments{
\item{...}{The input can be:
\itemize{
\item A data frame with PK parameters in it; This should ideally
be a data frame with one pk parameter per row since it will
output a data frame with one PK parameter per row.
\item PK parameters as either a vector or a scalar
}}

\item{verbose}{boolean that when TRUE provides a message about the detected pk parameters
and the detected compartmental model.  By default this is \code{FALSE}.}

\item{digits}{represents the number of significant digits for the
output; If the number is zero or below (default), do not round.}
}
\value{
Return a data.frame of derived PK parameters for a 1-, 2-,
or 3-compartment linear model given provided clearances and
volumes based on the inferred model type.

The model parameters that will be provided in the data frame are:
\itemize{
\item \code{vc}: Central Volume (for 1-, 2- and 3-
compartment models)
\item \code{kel}: First-order elimination rate (for 1-, 2-, and
3-compartment models)
\item \code{k12}: First-order rate of transfer from central to
first peripheral compartment; (for 2- and 3-compartment models)
\item \code{k21}: First-order rate of transfer from first
peripheral to central compartment, (for 2- and 3-compartment
models)
\item \code{k13}: First-order rate of transfer from central to
second peripheral compartment; (3-compartment model)
\item \code{k31}: First-order rate of transfer from second
peripheral to central compartment (3-compartment model)
\item \code{vp}: Peripheral Volume (for 2- and 3- compartment models)
\item \code{vp2}: Peripheral Volume for 3rd compartment (3- compartment model)
\item \code{vss}: Volume of distribution at steady state; (1-, 2-, and 3-compartment models)
\item \code{t12alpha}: \eqn{t_{1/2,\alpha}}; (1-, 2-, and 3-compartment models)
\item \code{t12beta}: \eqn{t_{1/2,\beta}}; (2- and 3-compartment models)
\item \code{t12gamma}: \eqn{t_{1/2,\gamma}}; (3-compartment model)
\item \code{alpha}: \eqn{\alpha}; (1-, 2-, and 3-compartment models)
\item \code{beta}: \eqn{\beta}; (2- and 3-compartment models)
\item \code{gamma}: \eqn{\beta}; (3-compartment model)
\item \code{A}: true \code{A}; (1-, 2-, and 3-compartment models)
\item \code{B}: true \code{B}; (2- and 3-compartment models)
\item \code{C}: true \code{C}; (3-compartment model)
\item \code{fracA}: fractional A; (1-, 2-, and 3-compartment models)
\item \code{fracB}: fractional B; (2- and 3-compartment models)
\item \code{fracC}: fractional C; (3-compartment model)
}
}
\description{
This calculates the derived parameters based on what is provided
in a data frame or arguments
}
\examples{

## Note that rxode2 parses the names to figure out the best PK parameter

params <- rxDerived(cl = 29.4, v = 23.4, Vp = 114, vp2 = 4614, q = 270, q2 = 73)

## That is why this gives the same results as the value before

params <- rxDerived(CL = 29.4, V1 = 23.4, V2 = 114, V3 = 4614, Q2 = 270, Q3 = 73)

## You may also use micro-constants alpha/beta etc.

params <- rxDerived(k12 = 0.1, k21 = 0.2, k13 = 0.3, k31 = 0.4, kel = 10, v = 10)

## or you can mix vectors and scalars

params <- rxDerived(CL = 29.4, V = 1:3)

## If you want, you can round to a number of significant digits
## with the `digits` argument:

params <- rxDerived(CL = 29.4, V = 1:3, digits = 2)
}
\references{
Shafer S. L. \code{CONVERT.XLS}

Rowland M, Tozer TN. Clinical Pharmacokinetics and Pharmacodynamics: Concepts and Applications (4th). Clipping Williams & Wilkins, Philadelphia, 2010.
}
\author{
Matthew Fidler and documentation from Justin Wilkins, \email{justin.wilkins@occams.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.minfo}
\alias{.minfo}
\alias{.malert}
\alias{.mwarn}
\alias{.msuccess}
\title{Internal messaging statements}
\usage{
.minfo(text, ..., .envir = parent.frame())

.malert(text, ..., .envir = parent.frame())

.mwarn(text, ..., .envir = parent.frame())

.msuccess(text, ..., .envir = parent.frame())
}
\arguments{
\item{text}{Text}

\item{...}{Other arguments}

\item{.envir}{Environment to evaluate in}
}
\value{
Nothing, called for side effects
}
\description{
Internal messaging statements
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-amt.R
\name{stat_amt}
\alias{stat_amt}
\alias{geom_amt}
\title{Dosing/Amt geom/stat}
\usage{
stat_amt(
  mapping = NULL,
  data = NULL,
  position = "identity",
  show.legend = NA,
  inherit.aes = TRUE,
  ...
)

geom_amt(
  mapping = NULL,
  data = NULL,
  position = "identity",
  show.legend = NA,
  inherit.aes = TRUE,
  ...
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[ggplot2:aes]{aes()}}. If specified and
\code{inherit.aes = TRUE} (the default), it is combined with the default mapping
at the top level of the plot. You must supply \code{mapping} if there is no plot
mapping.}

\item{data}{The data to be displayed in this layer. There are three
options:

If \code{NULL}, the default, the data is inherited from the plot
data as specified in the call to \code{\link[ggplot2:ggplot]{ggplot()}}.

A \code{data.frame}, or other object, will override the plot
data. All objects will be fortified to produce a data frame. See
\code{\link[ggplot2:fortify]{fortify()}} for which variables will be created.

A \code{function} will be called with a single argument,
the plot data. The return value must be a \code{data.frame}, and
will be used as the layer data. A \code{function} can be created
from a \code{formula} (e.g. \code{~ head(.x, 10)}).}

\item{position}{A position adjustment to use on the data for this layer. This
can be used in various ways, including to prevent overplotting and
improving the display. The \code{position} argument accepts the following:
\itemize{
\item The result of calling a position function, such as \code{position_jitter()}.
This method allows for passing extra arguments to the position.
\item A string naming the position adjustment. To give the position as a
string, strip the function name of the \code{position_} prefix. For example,
to use \code{position_jitter()}, give the position as \code{"jitter"}.
\item For more information and other ways to specify the position, see the
\link[ggplot2:layer_positions]{layer position} documentation.
}}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped.
\code{FALSE} never includes, and \code{TRUE} always includes.
It can also be a named logical vector to finely select the aesthetics to
display. To include legend keys for all levels, even
when no data exists, use \code{TRUE}.  If \code{NA}, all levels are shown in legend,
but unobserved levels are omitted.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics,
rather than combining with them. This is most useful for helper functions
that define both data and aesthetics and shouldn't inherit behaviour from
the default plot specification, e.g. \code{\link[ggplot2:annotation_borders]{annotation_borders()}}.}

\item{...}{Other arguments passed on to \code{\link[ggplot2:layer]{layer()}}'s \code{params} argument. These
arguments broadly fall into one of 4 categories below. Notably, further
arguments to the \code{position} argument, or aesthetics that are required
can \emph{not} be passed through \code{...}. Unknown arguments that are not part
of the 4 categories below are ignored.
\itemize{
\item Static aesthetics that are not mapped to a scale, but are at a fixed
value and apply to the layer as a whole. For example, \code{colour = "red"}
or \code{linewidth = 3}. The geom's documentation has an \strong{Aesthetics}
section that lists the available options. The 'required' aesthetics
cannot be passed on to the \code{params}. Please note that while passing
unmapped aesthetics as vectors is technically possible, the order and
required length is not guaranteed to be parallel to the input data.
\item When constructing a layer using
a \verb{stat_*()} function, the \code{...} argument can be used to pass on
parameters to the \code{geom} part of the layer. An example of this is
\code{stat_density(geom = "area", outline.type = "both")}. The geom's
documentation lists which parameters it can accept.
\item Inversely, when constructing a layer using a
\verb{geom_*()} function, the \code{...} argument can be used to pass on parameters
to the \code{stat} part of the layer. An example of this is
\code{geom_area(stat = "density", adjust = 0.5)}. The stat's documentation
lists which parameters it can accept.
\item The \code{key_glyph} argument of \code{\link[ggplot2:layer]{layer()}} may also be passed on through
\code{...}. This can be one of the functions described as
\link[ggplot2:draw_key]{key glyphs}, to change the display of the layer in the legend.
}}
}
\value{
This returns a stat_amt in context of a ggplot2 plot
}
\description{
This is a dosing geom that shows the vertical lines where a dose occurs
}
\details{
Requires the following aesthetics:

\itemize{
\item x representing the x values, usually time
\item amt representing the dosing values;  They are missing or zero when no dose is given
}
}
\examples{

\donttest{
library(rxode2)
library(units)


## Model from RxODE tutorial
mod1 <- function() {
  ini({
    KA <- 2.94E-01
    CL <- 1.86E+01
    V2 <- 4.02E+01
    Q  <- 1.05E+01
    V3 <- 2.97E+02
    Kin <- 1
    Kout <- 1
    EC50 <- 200
  })
  model({
    C2 <- centr/V2
    C3 <- peri/V3
    d/dt(depot) <- -KA*depot
    d/dt(centr) <-  KA*depot - CL*C2 - Q*C2 + Q*C3
    d/dt(peri)  <-                    Q*C2 - Q*C3
    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
  })
}

## These are making the more complex regimens of the rxode2 tutorial

## bid for 5 days
bid <- et(timeUnits="hr") |>
  et(amt=10000,ii=12,until=set_units(5, "days"))

## qd for 5 days
qd <- et(timeUnits="hr") |>
      et(amt=20000,ii=24,until=set_units(5, "days"))

## bid for 5 days followed by qd for 5 days

et <- seq(bid,qd) |>
      et(seq(0,11*24,length.out=100))

bidQd <- rxSolve(mod1, et, addDosing=TRUE)

# by default dotted and under-stated
plot(bidQd, C2) + geom_amt(aes(amt=amt))

# of course you can make it a bit more visible

plot(bidQd, C2) + geom_amt(aes(amt=amt), col="red", lty=1, linewidth=1.2)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxGetrxode2}
\alias{rxGetrxode2}
\title{Get rxode2 model from object}
\usage{
rxGetrxode2(obj)
}
\arguments{
\item{obj}{rxode2 family of objects}
}
\value{
rxode2 model
}
\description{
Get rxode2 model from object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjointDiscrete.R
\name{.rxDiscreteAdjointBuild}
\alias{.rxDiscreteAdjointBuild}
\title{Build symbolic evaluators for the discrete adjoint (RHS, Jacobian, forcing)}
\usage{
.rxDiscreteAdjointBuild(object, calcSens)
}
\arguments{
\item{object}{model definition accepted by \code{\link[=rxode2]{rxode2()}}.}

\item{calcSens}{character vector of parameters to differentiate w.r.t.}
}
\value{
a list with \code{st} (state names), \code{ns}, \code{np}, \code{calcSens}, and evaluator
closures \code{Fe(x, params)}, \code{FXe(x, params)} (\verb{ns x ns}), \code{FPe(x, params)}
(\verb{ns x np}).
}
\description{
Parses the model once and returns closures that evaluate the ODE RHS \code{F}, the
state Jacobian \code{dF/dy} and the parameter forcing \code{dF/dtheta} at an arbitrary
state (needed at interior RK4 stage points).  Pair with
\code{\link[=.rxDiscreteForwardSens]{.rxDiscreteForwardSens()}} and \code{\link[=.rxDiscreteAdjointGrad]{.rxDiscreteAdjointGrad()}}.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxMemoryEstimate.R
\name{rxMemSummary}
\alias{rxMemSummary}
\title{Create a per-ID event summary for memory estimation}
\usage{
rxMemSummary(nobs, ndoses, id = seq_along(nobs))
}
\arguments{
\item{nobs}{Integer vector of observation counts per ID.}

\item{ndoses}{Integer vector of dose event counts per ID.}

\item{id}{Integer or character vector of subject IDs (optional).}
}
\value{
A data.frame with class \code{"rxMemSummary"}.
}
\description{
Create a per-ID event summary for memory estimation
}
\examples{

# Three subjects with known observation and dose counts
rxMemSummary(nobs = c(48L, 96L, 48L), ndoses = c(7L, 14L, 7L))

# Explicit subject IDs
rxMemSummary(nobs = c(10L, 20L), ndoses = c(5L, 5L), id = c(101L, 102L))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikNorm}
\alias{llikNorm}
\title{Log likelihood for normal distribution}
\usage{
llikNorm(x, mean = 0, sd = 1, full = FALSE)
}
\arguments{
\item{x}{Observation}

\item{mean}{Mean for the likelihood}

\item{sd}{Standard deviation for the likelihood}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the pdf value of with \code{dMean} and
\code{dSd} that has the derivatives with respect to the parameters at
the observation time-point
}
\description{
Log likelihood for normal distribution
}
\details{
In an \code{rxode2()} model, you can use \code{llikNorm()} but you have to
use all arguments.  You can also get the derivatives with
\code{llikNormDmean()} and \code{llikNormDsd()}
}
\examples{

\donttest{

llikNorm(0)

llikNorm(seq(-2,2,length.out=10), full=TRUE)

# With rxode2 you can use:
 
et <- et(-3, 3, length.out=10)
et$mu <- 0
et$sigma <- 1

model <- function(){
  model({
    fx <- llikNorm(time, mu, sigma)
    dMean <- llikNormDmean(time, mu, sigma)
    dSd <- llikNormDsd(time, mu, sigma)
  })
 }

ret <- rxSolve(model, et)
ret
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indLin.R
\name{indLin}
\alias{indLin}
\alias{rxOdeToIndLin}
\alias{rxToIndLin}
\title{Transition ODEs written in d/dt() format to matrix exponential / inductive linearization format}
\usage{
indLin(model, doConst = FALSE, calcSens = NULL)

rxOdeToIndLin(model, doConst = FALSE, calcSens = NULL)

rxToIndLin(model, doConst = FALSE, calcSens = NULL)
}
\arguments{
\item{model}{rxode2 model, text, or function}

\item{doConst}{Replace constants with values; By default this is \code{FALSE}.}

\item{calcSens}{A character vector of parameter names for which sensitivities should be calculated.}
}
\value{
A character string representing the matrix exponential model code
}
\description{
Transition ODEs written in d/dt() format to matrix exponential / inductive linearization format
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{gammaqInv}
\alias{gammaqInv}
\alias{gammaqInva}
\title{gammaqInv and gammaqInva:  Inverses of normalized gammaq function}
\usage{
gammaqInv(a, q)

gammaqInva(x, q)
}
\arguments{
\item{a}{The numeric 'a' parameter in the upper
incomplete gamma}

\item{q}{The numeric 'q' parameter in the upper
incomplete gamma}

\item{x}{The numeric 'x' parameter in the upper incomplete gamma}
}
\value{
inverse gammaq results
}
\description{
gammaqInv and gammaqInva:  Inverses of normalized gammaq function
}
\details{
With the equation:

q = gammaq(a, x)

The 'gammaqInv' function returns a value 'x' that satisfies the
equation above

The 'gammaqInva' function returns a value 'a' that satisfies the
equation above

NOTE: gammaqInva is slow
}
\examples{

gammaqInv(1:3, 0.5)

gammaqInv(1, 1:3 / 3)

gammaqInv(1:3, 1:3 / 3.1)

gammaqInva(1:3, 1:3 / 3.1)
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dde.R
\name{.rxDelaySensJumpMap}
\alias{.rxDelaySensJumpMap}
\alias{.rxDelaySensJumpEvents}
\alias{.rxDelaySensJump}
\alias{.rxDelaySensJumpMapCached}
\title{Dose-induced breaking-point jump for forward delay sensitivities}
\usage{
.rxDelaySensJumpMap(model, calcSens)

.rxDelaySensJumpEvents(jumpMap, st, events)

.rxDelaySensJump(model, calcSens, events)

.rxDelaySensJumpMapCached(model, calcSens, keyTxt = NULL)
}
\arguments{
\item{model}{base ODE model (anything \code{rxNorm()} accepts).}

\item{calcSens}{character vector of sensitivity parameters.}

\item{jumpMap}{the jump map from \code{.rxDelaySensJumpMap()} (\verb{$jumpMap}).}

\item{st}{ODE state names (\verb{$st} from \code{.rxDelaySensJumpMap()}), for numeric-cmt
resolution.}

\item{events}{an rxode2 event table (to mirror the state-j doses onto the
sensitivity compartments).}

\item{keyTxt}{normalized model text used as the cache key (pass \code{rxNorm(model)}
if already computed, else it is derived).}
}
\value{
\code{.rxDelaySensJumpMap()}: list with \code{alagf}, \code{jumpMap}, and \code{st}, or
\code{NULL} when the model has no parameter-dependent delay.  \code{.rxDelaySensJump()}:
list with \code{alagf} and \code{events}, or \code{NULL}.
}
\description{
A dose propagated through \code{delay(y_j, T(p))} makes the 1st-order sensitivity
jump at \code{t = t_dose + T} by \verb{[S_i] = -(d f_i/d delay(y_j,T)) * [y_j] * dT/dp},
reproduced here as a modeled bolus of \verb{[y_j]} on the sensitivity compartment
with modeled lag \code{T} and bioavailability \verb{-(d f_i/d delay)*dT/dp}.
\code{.rxDelaySensJumpMap()} does the model-only symengine analysis (cached via
\code{.rxDelaySensJumpMapCached()}); \code{.rxDelaySensJumpEvents()} rbinds the
mirroring doses per solve; \code{.rxDelaySensJump()} does both.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tran.R
\name{rxode2parse}
\alias{rxode2parse}
\title{Internal translation to get model variables list}
\usage{
rxode2parse(
  model,
  linear = FALSE,
  linCmtSens = c("linCmtA", "linCmtB"),
  verbose = FALSE,
  code = NULL,
  envir = parent.frame()
)
}
\arguments{
\item{model}{Model (either file name or string)}

\item{linear}{boolean indicating if linear compartment model should
be generated from \code{linCmt()} (default FALSE)}

\item{linCmtSens}{Linear compartment model sensitivity type}

\item{verbose}{is a boolean indicating the type of model detected
with \code{linCmt()} parsing}

\item{code}{is a file name where the c code is written to (for
testing purposes mostly, it needs \code{rxode2} to do anything fancy)}

\item{envir}{is the environment to look for R user functions
(defaults to parent environment)}
}
\value{
A rxModelVars object that has the model variables of a
rxode2 syntax expression
}
\description{
Internal translation to get model variables list
}
\examples{
rxode2parse("a=3")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R, R/format.R, R/print.R
\name{rxEvid}
\alias{rxEvid}
\alias{as.rxEvid}
\alias{c.rxEvid}
\alias{[.rxEvid}
\alias{as.character.rxEvid}
\alias{[[.rxEvid}
\alias{units<-.rxEvid}
\alias{c.rxRateDur}
\alias{format.rxEvid}
\alias{format.rxRateDur}
\alias{print.rxEvid}
\title{EVID formatting for tibble and other places.}
\usage{
rxEvid(x)

as.rxEvid(x)

\method{c}{rxEvid}(x, ...)

\method{[}{rxEvid}(x, ...)

\method{as.character}{rxEvid}(x, ...)

\method{[[}{rxEvid}(x, ...)

\method{units}{rxEvid}(x) <- value

\method{c}{rxRateDur}(x, ...)

\method{format}{rxEvid}(x, ...)

\method{format}{rxRateDur}(x, ...)

\method{print}{rxEvid}(x, ...)
}
\arguments{
\item{x}{Item to be converted to a rxode2 EVID specification.}

\item{...}{Other parameters}

\item{value}{It will be an error to set units for evid}
}
\value{
rxEvid specification
}
\description{
This is to make an EVID more readable by non
pharmacometricians. It displays what each means and allows it to
be displayed in a tibble.
}
\examples{

rxEvid(1:7)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjointDiscrete.R
\name{.rxAdjointModel}
\alias{.rxAdjointModel}
\title{Build (and cache) the compiled adjoint-expansion model for rk4s}
\usage{
.rxAdjointModel(object, calcSens, stiff = FALSE)
}
\arguments{
\item{object}{model definition accepted by \code{\link[=rxode2]{rxode2()}}.}

\item{calcSens}{character vector of parameters to differentiate w.r.t.}
}
\value{
list with \code{model} (compiled rxode2) and \code{info} (the
\code{\link[=.rxAdjointExpand]{.rxAdjointExpand()}} layout).
}
\description{
Runs \code{\link[=.rxAdjointExpand]{.rxAdjointExpand()}} and compiles the result, caching by the normalized
model text + \code{calcSens} so an optimizer that resolves the gradient many times
compiles once.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxParams.R
\name{rxParams}
\alias{rxParams}
\alias{rxParams.rxode2}
\alias{rxParams.rxSolve}
\alias{rxParams.rxEt}
\alias{rxParam}
\title{Parameters specified by the model}
\usage{
rxParams(obj, ...)

\method{rxParams}{rxode2}(
  obj,
  constants = TRUE,
  ...,
  params = NULL,
  inits = NULL,
  iCov = NULL,
  keep = NULL,
  thetaMat = NULL,
  omega = NULL,
  dfSub = NULL,
  sigma = NULL,
  dfObs = NULL,
  nSub = NULL,
  nStud = NULL
)

\method{rxParams}{rxSolve}(
  obj,
  constants = TRUE,
  ...,
  params = NULL,
  inits = NULL,
  iCov = NULL,
  keep = NULL,
  thetaMat = NULL,
  omega = NULL,
  dfSub = NULL,
  sigma = NULL,
  dfObs = NULL,
  nSub = NULL,
  nStud = NULL
)

\method{rxParams}{rxEt}(
  obj,
  ...,
  params = NULL,
  inits = NULL,
  iCov = NULL,
  keep = NULL,
  thetaMat = NULL,
  omega = NULL,
  dfSub = NULL,
  sigma = NULL,
  dfObs = NULL,
  nSub = NULL,
  nStud = NULL
)

rxParam(obj, ...)
}
\arguments{
\item{obj}{rxode2 family of objects}

\item{...}{Other arguments including scaling factors for each
compartment.  This includes \code{S# =} numeric will scale a compartment
\code{#} by a dividing the compartment amount by the scale factor,
like NONMEM.}

\item{constants}{is a boolean indicting if constants should be
included in the list of parameters. Currently rxode2 parses
constants into variables in case you wish to change them
without recompiling the rxode2 model.}

\item{params}{a numeric named vector with values for every
parameter in the ODE system; the names must correspond to the
parameter identifiers used in the ODE specification;}

\item{inits}{a vector of initial values of the state variables
(e.g., amounts in each compartment), and the order in this
vector must be the same as the state variables (e.g., PK/PD
compartments);}

\item{iCov}{A data frame of individual non-time varying covariates
to combine with the \code{events} dataset.  The \code{iCov} dataset has one
covariate per ID and should match the event table}

\item{keep}{Columns to keep from either the input dataset or the
\code{iCov} dataset.  With the \code{iCov} dataset, the column
is kept once per line.  For the input dataset, if any records
are added to the data LOCF (Last Observation Carried forward)
imputation is performed.}

\item{thetaMat}{Named theta matrix.}

\item{omega}{Estimate of Covariance matrix. When omega is a list,
assume it is a block matrix and convert it to a full matrix for
simulations.  When \code{omega} is \code{NA} and you are using it with a
\code{rxode2} ui model, the between subject variability described by
the \code{omega} matrix are set to zero.}

\item{dfSub}{Degrees of freedom to sample the between subject variability matrix from the
inverse Wishart distribution (scaled) or scaled inverse chi squared distribution.}

\item{sigma}{Named sigma covariance or Cholesky decomposition of a
covariance matrix.  The names of the columns indicate
parameters that are simulated.  These are simulated for every
observation in the solved system. When \code{sigma} is \code{NA} and you are using it with a
\code{rxode2} ui model, the unexplained variability described by
the \code{sigma} matrix are set to zero.}

\item{dfObs}{Degrees of freedom to sample the unexplained variability matrix from the
inverse Wishart distribution (scaled) or scaled inverse chi squared distribution.}

\item{nSub}{Number between subject variabilities (\code{ETAs}) simulated for every
realization of the parameters.}

\item{nStud}{Number virtual studies to characterize uncertainty in estimated
parameters.}
}
\value{
When extracting the parameters from an rxode2 model, a
character vector listing the parameters in the model.
}
\description{
This returns the model's parameters that are required to solve the
ODE system, and can be used to pipe parameters into an rxode2 solve
}
\seealso{
Other Query model information:
\code{\link[=rxDfdy]{rxDfdy()}},
\code{\link[=rxInits]{rxInits()}},
\code{\link[=rxLhs]{rxLhs()}},
\code{\link[=rxModelVars]{rxModelVars()}},
\code{\link[=rxState]{rxState()}}
}
\author{
Matthew L.Fidler
}
\concept{Query model information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.rxDll}
\alias{summary.rxDll}
\title{Summary of rxDll object}
\usage{
\method{summary}{rxDll}(object, ...)
}
\arguments{
\item{object}{RxDll object}

\item{...}{Other arguments.  Includes \code{noprint}, which is a
logical telling if the object should print the rxDll object
first. By default this is FALSE}
}
\value{
object is returned
}
\description{
This gives expanded information about the rxDll object
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjointDiscrete.R
\name{rxSolveAdjointRk4}
\alias{rxSolveAdjointRk4}
\title{Solve a model with an in-engine discrete-adjoint (\code{s}) method}
\usage{
rxSolveAdjointRk4(
  object,
  events,
  params = NULL,
  calcSens,
  method = "rk4s",
  ...
)
}
\arguments{
\item{object}{model definition accepted by \code{\link[=rxode2]{rxode2()}}.}

\item{events}{an rxode2 event table / data set.}

\item{params}{optional named parameter vector or data frame (per subject).}

\item{calcSens}{character vector of parameters to differentiate w.r.t.}

\item{method}{adjoint solve method (default \code{"rk4s"}).  May be a composite
(\code{"dop853s+ros4s"}).}

\item{...}{passed to \code{\link[=rxSolve]{rxSolve()}}.}
}
\value{
the solved data frame with \verb{rx__sens_<state>_BY_<param>__} columns.
}
\description{
Applies the (cached) adjoint expansion, solves with the requested adjoint
method, and drops the internal \verb{rx__adj*} lhs, returning the full-trajectory
\verb{rx__sens_<state>_BY_<param>__} columns.  The stiff analytic Jacobian is
emitted into the expansion automatically, inferred from \code{method}.
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collectWarnings.R
\name{.collectWarnings}
\alias{.collectWarnings}
\title{Collect warnings and just warn once.}
\usage{
.collectWarnings(expr, lst = FALSE)
}
\arguments{
\item{expr}{R expression}

\item{lst}{When \code{TRUE} return a list with
list(object,warnings) instead of issuing the warnings.
Otherwise, when \code{FALSE} issue the warnings and return the
object.}
}
\value{
The value of the expression or a list with the value of
the expression and a list of warning messages
}
\description{
Collect warnings and just warn once.
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rLKJ1}
\alias{rLKJ1}
\title{One correlation sample from the LKJ distribution}
\usage{
rLKJ1(d, eta = 1, cholesky = FALSE)
}
\arguments{
\item{d}{The dimension of the correlation matrix}

\item{eta}{The scaling parameter of the LKJ distribution.
Must be > 1.  Also related to the degrees of freedom nu.
eta = (nu-1)/2.}

\item{cholesky}{boolean; If \code{TRUE} return the cholesky
decomposition.}
}
\value{
A correlation sample from the LKJ distribution
}
\description{
One correlation sample from the LKJ distribution
}
\author{
Matthew Fidler (translated to RcppArmadillo) and Emma Schwager
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{rxRmvn}
\alias{rxRmvn}
\title{Simulate from a (truncated) multivariate normal}
\usage{
rxRmvn(
  n,
  mu = NULL,
  sigma,
  lower = -Inf,
  upper = Inf,
  ncores = 1,
  isChol = FALSE,
  keepNames = TRUE,
  a = 0.4,
  tol = 2.05,
  nlTol = 1e-10,
  nlMaxiter = 100L
)
}
\arguments{
\item{n}{Number of random row vectors to be simulated OR the
matrix to use for simulation (faster).}

\item{mu}{mean vector}

\item{sigma}{Covariance matrix for multivariate normal or a list
of covariance matrices. If a list of covariance matrix, each
matrix will simulate \code{n} matrices and combine them to a full
matrix}

\item{lower}{is a vector of the lower bound for the truncated
multivariate norm}

\item{upper}{is a vector of the upper bound for the truncated
multivariate norm}

\item{ncores}{Number of cores used in the simulation}

\item{isChol}{A boolean indicating if \code{sigma} is a cholesky
decomposition of the covariance matrix.}

\item{keepNames}{Keep the names from either the mean or covariance
matrix.}

\item{a}{threshold for switching between methods; They can be
tuned for maximum speed;  There are three cases that are considered:

case 1: a < l < u

case 2: l < u < -a

case 3: otherwise

where l=lower and u = upper}

\item{tol}{When case 3 is used from the above possibilities, the
tol value controls the acceptance rejection and
inverse-transformation;

When abs(u-l)>tol, uses accept-reject from randn}

\item{nlTol}{Tolerance for newton line-search}

\item{nlMaxiter}{Maximum iterations for newton line-search}
}
\value{
If \code{n==integer} (default) the output is an (n x d) matrix
where the i-th row is the i-th simulated vector.

If \code{is.matrix(n)} then the random vector are store in \code{n},
which is provided by the user, and the function returns
\code{NULL} invisibly.
}
\description{
This is simulated with the fast, thread-safe threefry simulator
and can use multiple cores to generate the random deviates.
}
\examples{

## From mvnfast
## Unlike mvnfast, uses threefry simulation

d <- 5
mu <- 1:d

# Creating covariance matrix
tmp <- matrix(rnorm(d^2), d, d)
mcov <- tcrossprod(tmp, tmp)


set.seed(414)
rxRmvn(4, 1:d, mcov)

set.seed(414)
rxRmvn(4, 1:d, mcov)

set.seed(414)
rxRmvn(4, 1:d, mcov, ncores = 2) # r.v. generated on the second core are different

###### Here we create the matrix that will hold the simulated
#  random variables upfront.
A <- matrix(NA, 4, d)
class(A) <- "numeric" # This is important. We need the elements of A to be of class "numeric".

set.seed(414)
rxRmvn(A, 1:d, mcov, ncores = 2) # This returns NULL ...
A # ... but the result is here

## You can also simulate from a truncated normal:

rxRmvn(10, 1:d, mcov, lower = 1:d - 1, upper = 1:d + 1)


# You can also simulate from different matrices (if they match
# dimensions) by using a list of matrices.

matL <- lapply(1:4, function(...) {
  tmp <- matrix(rnorm(d^2), d, d)
  tcrossprod(tmp, tmp)
})


rxRmvn(4, setNames(1:d, paste0("a", 1:d)), matL)
}
\references{
John K. Salmon, Mark A. Moraes, Ron O. Dror, and David
E. Shaw (2011). Parallel Random Numbers: As Easy as 1, 2, 3.
D. E. Shaw Research, New York, NY 10036, USA.

The thread safe multivariate normal was inspired from the \code{mvnfast} package by Matteo Fasiolo \url{https://CRAN.R-project.org/package=mvnfast}

The concept of the truncated multivariate normal was
taken from Zdravko Botev Botev (2017) \doi{10.1111/rssb.12162}
and Botev and L'Ecuyer (2015) \doi{10.1109/WSC.2015.7408180} and
converted to thread safe simulation;
}
\author{
Matthew Fidler, Zdravko Botev and some from Matteo Fasiolo
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{.rxSensStrippable}
\alias{.rxSensStrippable}
\title{Get model properties without compiling it.}
\usage{
.rxSensStrippable(mv)
}
\arguments{
\item{mv}{rxode2 model variables

Sensitivity-named states (\verb{rx__sens_<state>_BY_<var>__}) are
stripped of their \code{d/dt()} when calcJac/calcSens regenerate the
sensitivity block.  A state that is read by \code{delay()} is a
load-bearing ODE (\code{delay(X, T)} requires \code{d/dt(X)}), so it must
never be stripped -- e.g. nlmixr2est's analytic-cov augmented model
supplies delayed sensitivity ODEs directly.  Returns the sens
states safe to strip (those not referenced by any delay() term).}
}
\value{
rxode2 trans list
}
\description{
Get model properties without compiling it.
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.matchesLangTemplate}
\alias{.matchesLangTemplate}
\title{Check if a language object matches a template language object}
\usage{
.matchesLangTemplate(x, template)
}
\arguments{
\item{x}{The object to check}

\item{template}{The template object it should match}
}
\value{
TRUE if it matches, FALSE, otherwise
}
\description{
\itemize{
\item{If \code{template == str2lang(".")}, it will match anything.}
\item{If \code{template == str2lang(".name")}, it will match any name.}
\item{If \code{template == str2lang(".call()")}, it will match any call.}
}
}
\examples{
.matchesLangTemplate(str2lang("d/dt(foo)"), str2lang("d/dt(.name)"))
.matchesLangTemplate(str2lang("d/dt(foo)"), str2lang("d/foo(.name)"))
.matchesLangTemplate(str2lang("d/dt(foo)"), str2lang("d/."))
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.rxLinCmtGen}
\alias{.rxLinCmtGen}
\title{Internal function to generate the model variables for a linCmt() model}
\usage{
.rxLinCmtGen(lenState, vars)
}
\arguments{
\item{lenState}{Length of the state}

\item{vars}{Variables in the model}
}
\value{
Model variables of expanded linCmt model
}
\description{
Internal function to generate the model variables for a linCmt() model
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linMod.R
\name{linMod}
\alias{linMod}
\alias{linMod0}
\alias{linModB}
\alias{linModB0}
\alias{linModA}
\alias{linModA0}
\alias{linModD}
\alias{linModD0}
\alias{linModM}
\alias{linModM0}
\title{Linear model to replace in rxode2 ui model}
\usage{
linMod(
  variable,
  power,
  dv = "dv",
  intercept = TRUE,
  type = c("replace", "before", "after"),
  num = NULL,
  iniDf = NULL,
  data = FALSE,
  mv = FALSE
)

linMod0(..., intercept = FALSE)

linModB(..., type = "before")

linModB0(..., intercept = FALSE, type = "before")

linModA(..., type = "after")

linModA0(..., intercept = FALSE, type = "after")

linModD(..., intercept = TRUE, data = TRUE)

linModD0(..., intercept = FALSE, data = TRUE)

linModM(..., intercept = TRUE, mv = TRUE)

linModM0(..., intercept = FALSE, mv = TRUE)
}
\arguments{
\item{variable}{The variable that the rxode2 will be made on.}

\item{power}{The power of the polynomial that will be generated.}

\item{dv}{the dependent variable to use to generate the initial
estimates from the data. If \code{NULL} query using \code{rxUdfUiData()}.}

\item{intercept}{Boolean that tells if the intercept be generated.}

\item{type}{the type of linear model replacement to be used.}

\item{num}{the number the particular model is being generated. If
unspecified, query using \code{rxUdfUiNum()}.}

\item{iniDf}{the initialization \code{data.frame}, if \code{NULL} query using
\code{rxUdfUiIniDf()}}

\item{data}{logical that tells if the initial estimates of the
linear model should be estimated from the data.}

\item{mv}{logical that tell if the model variables need to be used
to generate model variables.}

\item{...}{arguments that are passed to \code{linMod()} for the other
abbreviations of \code{linMod()}}
}
\value{
a list for use in when generating the \code{rxode2} ui model see
\code{rxUdfUi()} for details.
}
\description{
Linear model to replace in rxode2 ui model
}
\section{Functions}{
\itemize{
\item \code{linMod0()}: linear model without intercept

\item \code{linModB()}: linear model before where it occurs

\item \code{linModB0()}: linear model before where the user function occurs

\item \code{linModA()}: linear model after where the user function occurs

\item \code{linModA0()}: liner model without an intercept placed after where the user function occurs

\item \code{linModD()}: linear model where initial estimates are generated from the data

\item \code{linModD0()}: linear model where initial estimates are generated from the data (no intercept)

\item \code{linModM()}: linear model where the model variables are used to generate the model variables

\item \code{linModM0()}: linear model where the model variables are used to generate the model variables (no intercept)

}}
\examples{

linMod(x, 3)
}
\seealso{
Other User functions:
\code{\link[=rxUdfUiControl]{rxUdfUiControl()}},
\code{\link[=rxUdfUiData]{rxUdfUiData()}},
\code{\link[=rxUdfUiEst]{rxUdfUiEst()}},
\code{\link[=rxUdfUiIniLhs]{rxUdfUiIniLhs()}},
\code{\link[=rxUdfUiMv]{rxUdfUiMv()}},
\code{\link[=rxUdfUiNum]{rxUdfUiNum()}},
\code{\link[=rxUdfUiParsing]{rxUdfUiParsing()}}
}
\author{
Matthew L. Fidler
}
\concept{User functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-getControl.R
\name{rxSetControl}
\alias{rxSetControl}
\title{rxSetControl options for UI object}
\usage{
rxSetControl(ui, control)
}
\arguments{
\item{ui}{rxode2 ui object}

\item{control}{Default value}
}
\value{
Nothing, called for side effects
}
\description{
rxSetControl options for UI object
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikUnif}
\alias{llikUnif}
\title{log likelihood and derivatives for Unif distribution}
\usage{
llikUnif(x, alpha, beta, full = FALSE)
}
\arguments{
\item{x}{variable distributed by a uniform distribution}

\item{alpha}{is the lower limit of the uniform distribution}

\item{beta}{is the upper limit of the distribution}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the log pdf value of with \code{dProb}
that has the derivatives with respect to the \code{prob} parameters at
the observation time-point
}
\description{
log likelihood and derivatives for Unif distribution
}
\details{
In an \code{rxode2()} model, you can use \code{llikUnif()} but you have to
use the x and rate arguments.  You can also get the derivative of \code{alpha} or \code{beta} with
\code{llikUnifDalpha()} and \code{llikUnifDbeta()}.
}
\examples{

\donttest{

llikUnif(1, -2, 2)

et  <- et(seq(1,1, length.out=4))
et$alpha <- -2
et$beta <- 2
 
model <- function() {
  model({
    fx <- llikUnif(time, alpha, beta)
    dAlpha<- llikUnifDalpha(time, alpha, beta)
    dBeta <- llikUnifDbeta(time, alpha, beta)
  })
}

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symengine.R
\name{rxD}
\alias{rxD}
\title{Add to rxode2's derivative tables}
\usage{
rxD(name, derivatives)
}
\arguments{
\item{name}{Function Name}

\item{derivatives}{A list of functions. Each function takes the
same number of arguments as the original function.  The first
function will construct the derivative with respect to the
first argument; The second function will construct the
derivitive with respect to the second argument, and so on.}
}
\value{
nothing
}
\description{
Add to rxode2's derivative tables
}
\examples{
## Add an arbitrary list of derivative functions
## In this case the fun(x,y) is assumed to be 0.5*x^2+0.5*y^2

rxD("fun", list(
  function(x, y) {
    return(x)
  },
  function(x, y) {
    return(y)
  }
))
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{as.et}
\alias{as.et}
\alias{as.et.default}
\title{Coerce object to data.frame}
\usage{
as.et(x, ...)

\method{as.et}{default}(x, ...)
}
\arguments{
\item{x}{Object to coerce to et.}

\item{...}{Other parameters}
}
\value{
An event table
}
\description{
Coerce object to data.frame
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symengine.R
\name{rxSupportedFuns}
\alias{rxSupportedFuns}
\title{Get list of supported functions}
\usage{
rxSupportedFuns()
}
\value{
list of supported functions in rxode2
}
\description{
Get list of supported functions
}
\examples{
rxSupportedFuns()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxInv}
\alias{rxInv}
\title{Invert matrix using RcppArmadillo.}
\usage{
rxInv(matrix)
}
\arguments{
\item{matrix}{matrix to be inverted.}
}
\value{
inverse or pseudo inverse of matrix.
}
\description{
Invert matrix using RcppArmadillo.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlogit.R
\name{mlogit}
\alias{mlogit}
\title{mlogit -- Convert multiple probabilities to log-scale numbers}
\usage{
mlogit(
  ...,
  maxiter = 10000,
  rtol = 1e-10,
  atol = 1e-12,
  ctol = 1e-12,
  returnRoot = FALSE
)
}
\arguments{
\item{...}{numeric probabilities to convert to log-scale numbers.
These probabilities must add to a number less than 1 and are used
in the mix() estimation algorithm.}

\item{maxiter}{maximal number of iterations allowed.
  }

\item{rtol}{relative error tolerance, either a scalar or a vector, one
    value for each element in the unknown x.
  }

\item{atol}{absolute error tolerance, either a scalar or a vector, one
    value for each element in x.
  }

\item{ctol}{a scalar. If between two iterations, the maximal change in
    the variable values is less than this amount, then it is assumed that
    the root is found.
  }

\item{returnRoot}{logical; If TRUE, return the root object,
otherwise return the root itself.}
}
\value{
A numeric vector of the log-scale numbers for use in
regressions where the sum of a set of probabilities must add to
be one.
}
\description{
These multiple probabilities need to add up to be less than 1.
}
\details{
Once converted to log-scale numbers, they can be used in the
in the \code{mexpit()} to get the probabilities back with the following equation

\deqn{p_i = \frac{e^{x_i}}{1+\sum_{j=1}^{N-1} e^{x_j}}}

This ensures one remaining probability will add to one, that is

\deqn{p_N = \frac{1}{1+\sum_{j=1}^{N-1} e^{x_j}}}

Unfortunately, the log-scale inverse cannot be solved analytically,
so it is solved with the \code{rootSolve::multiroot()} function.

You may adjust some of the root finding options when using this
function.

When running \code{nlmixr2} with mixture models (ie. \code{mix()} models),
the \code{mlogit()} function is called in the probabilities and the
log-based values are used in the optimization problem.  The
probabilities are determined by the \code{mexpit()} function.
}
\examples{

mlogit(0.1, 0.2, 0.3)

mlogit(0.1, 0.2, 0.3, returnRoot = TRUE)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symengine.R
\name{rxToSE}
\alias{rxToSE}
\alias{.rxToSE}
\alias{rxFromSE}
\alias{.rxFromSE}
\title{rxode2 to symengine environment}
\usage{
rxToSE(
  x,
  envir = NULL,
  progress = FALSE,
  promoteLinSens = TRUE,
  parent = parent.frame()
)

.rxToSE(x, envir = NULL, progress = FALSE)

rxFromSE(
  x,
  unknownDerivatives = c("forward", "central", "error"),
  parent = parent.frame()
)

.rxFromSE(x)
}
\arguments{
\item{x}{expression}

\item{envir}{default is \code{NULL}; Environment to put symengine
variables in.}

\item{progress}{shows progress bar if true.}

\item{promoteLinSens}{Promote solved linear compartment systems to
sensitivity-based solutions.}

\item{parent}{is the parent environment to look for R-based user functions}

\item{unknownDerivatives}{When handling derivatives from unknown
functions, the translator will translate into different types
of numeric derivatives.  The currently supported methods are:

\if{html}{\out{<div class="sourceCode">}}\preformatted{- `forward` for forward differences
- `central` for central differences
- `error` for throwing an error for unknown derivatives
}\if{html}{\out{</div>}}}
}
\value{
An rxode2 symengine environment
}
\description{
rxode2 to symengine environment
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odeToLin.R
\name{odeToLin}
\alias{odeToLin}
\title{Convert ODE-based linear compartment models to analytical linCmt() form}
\usage{
odeToLin(ui)
}
\arguments{
\item{ui}{rxUi-like model object (function, rxUi, or anything accepted by
\code{as.rxUi}).}
}
\value{
An rxUi model with ODE equations replaced by \code{linCmt()}, or
(with a message) the original model unchanged if conversion is not possible.
}
\description{
Detects whether a model's ODE equations form a standard 1-3 compartment PK
system and, if so, replaces the \code{d/dt()} equations and the central
output assignment (e.g. \code{cp <- central/v}) with \code{cp <- linCmt()},
preserving all other model lines.  Detection requires linear ODE
right-hand sides, 1-4 compartments in a standard depot/central/peripheral
topology, and one output line of the form \code{var <- central_cmt / v_expr}.
Named parameter assignments are retained so \code{linCmt()} can infer the
parameterization.
}
\examples{
oneCmtOde <- function() {
  ini({
    tka <- 0.45
    tcl <- log(2.7)
    tv <- 3.45
    add.sd <- 0.7
  })
  model({
    ka  <- exp(tka)
    cl  <- exp(tcl)
    v   <- exp(tv)
    d/dt(depot)   <- -ka * depot
    d/dt(central) <- ka * depot - cl/v * central
    cp <- central / v
    cp ~ add(add.sd)
  })
}
linCmtModel <- odeToLin(oneCmtOde)

}
\seealso{
\code{\link{linToOde}} for the inverse transformation.
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{invWR1d}
\alias{invWR1d}
\title{One correlation sample from the Inverse Wishart distribution}
\usage{
invWR1d(d, nu, omegaIsChol = FALSE)
}
\arguments{
\item{d}{The dimension of the correlation matrix}

\item{nu}{Degrees of freedom of the Wishart distribution}

\item{omegaIsChol}{is an indicator of if the omega matrix is in
the Cholesky decomposition. This is only used when \code{type="invWishart"}}
}
\value{
One correlation sample from the inverse wishart
}
\description{
This correlation is constructed by transformation of the Inverse Wishart
random covariate to a correlation.
}
\author{
Matthew Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err-foceiBase.R
\name{.rxGetLambdaFromPred1AndIni}
\alias{.rxGetLambdaFromPred1AndIni}
\title{Get the lambda value based on the pred information}
\usage{
.rxGetLambdaFromPred1AndIni(env, pred1)
}
\arguments{
\item{env}{Environment that has the environment}

\item{pred1}{Single error data frame}
}
\value{
Lambda expression
}
\description{
Get the lambda value based on the pred information
}
\author{
Matthew Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/omegaVarCov.R
\name{rxOmegaVarCovDeriv}
\alias{rxOmegaVarCovDeriv}
\title{Variance-covariance (non-Cholesky) Omega parameterization derivatives}
\usage{
rxOmegaVarCovDeriv(omega, order = 2L)
}
\arguments{
\item{omega}{symmetric positive-definite random-effects covariance matrix.}

\item{order}{integer; \code{1} for first derivatives only, \code{2} (default) to also
return the second derivatives needed for the covariance Hessian.}
}
\value{
a list with \code{omegaInv}, \code{logDet}, the free-element index matrix
\code{elements} (each row \code{c(a, b)}, \code{a >= b}), first derivatives
\code{dOmegaInv} / \code{dLogDet}, and (when \code{order = 2}) second derivatives
\code{d2OmegaInv} / \code{d2LogDet}.
}
\description{
A non-Cholesky \code{Omega} path that differentiates with respect to the
variance-covariance entries directly (for reporting SEs on the natural scale
or building an analytic covariance over the \code{Omega} elements).  Returns
\code{Omega^{-1}}, \verb{log|Omega|}, and their first (and optionally second)
derivatives with respect to each free lower-triangular element
\verb{omega_\{ab\}}, via
}
\details{
\deqn{\partial \Omega^{-1}/\partial \omega_{ab} = -\Omega^{-1} E_{ab} \Omega^{-1}}
\deqn{\partial \log|\Omega|/\partial \omega_{ab} = \mathrm{tr}(\Omega^{-1} E_{ab})}

where \eqn{E_{ab}} is the symmetric single-entry basis matrix.
}
\author{
Hidde van de Beek
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxUdfUiIniDf}
\alias{rxUdfUiIniDf}
\title{Get the rxode2 iniDf of the current UI being processed (or return NULL)}
\usage{
rxUdfUiIniDf()
}
\value{
Initial \code{data.frame} being processed or \code{NULL} for nothing.
}
\description{
Get the rxode2 iniDf of the current UI being processed (or return NULL)
}
\examples{

rxUdfUiIniDf()

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxSimThetaOmega}
\alias{rxSimThetaOmega}
\title{Simulate Parameters from a Theta/Omega specification}
\usage{
rxSimThetaOmega(
  params = NULL,
  omega = NULL,
  omegaDf = NULL,
  omegaLower = as.numeric(c(R_NegInf)),
  omegaUpper = as.numeric(c(R_PosInf)),
  omegaIsChol = FALSE,
  omegaSeparation = "auto",
  omegaXform = 1L,
  nSub = 1L,
  thetaMat = NULL,
  thetaLower = as.numeric(c(R_NegInf)),
  thetaUpper = as.numeric(c(R_PosInf)),
  thetaDf = NULL,
  thetaIsChol = FALSE,
  nStud = 1L,
  sigma = NULL,
  sigmaLower = as.numeric(c(R_NegInf)),
  sigmaUpper = as.numeric(c(R_PosInf)),
  sigmaDf = NULL,
  sigmaIsChol = FALSE,
  sigmaSeparation = "auto",
  sigmaXform = 1L,
  nCoresRV = 1L,
  nObs = 1L,
  dfSub = 0,
  dfObs = 0,
  simSubjects = TRUE,
  simVariability = as.logical(c(NA_LOGICAL))
)
}
\arguments{
\item{params}{Named Vector of rxode2 model parameters}

\item{omega}{Estimate of Covariance matrix. When omega is a list,
assume it is a block matrix and convert it to a full matrix for
simulations.  When \code{omega} is \code{NA} and you are using it with a
\code{rxode2} ui model, the between subject variability described by
the \code{omega} matrix are set to zero.}

\item{omegaDf}{The degrees of freedom of a t-distribution for
simulation.  By default this is \code{NULL} which is
equivalent to \code{Inf} degrees, or to simulate from a normal
distribution instead of a t-distribution.}

\item{omegaLower}{Lower bounds for simulated ETAs (by default -Inf)}

\item{omegaUpper}{Upper bounds for simulated ETAs (by default Inf)}

\item{omegaIsChol}{Indicates if the \code{omega} supplied is a
Cholesky decomposed matrix instead of the traditional
symmetric matrix.}

\item{omegaSeparation}{Omega separation strategy

Tells the type of separation strategy when
simulating covariance with parameter uncertainty with standard
deviations modeled in the \code{thetaMat} matrix.
\itemize{
\item \code{"lkj"} simulates the correlation matrix from the
\code{rLKJ1} matrix with the distribution parameter \code{eta}
equal to the degrees of freedom \code{nu} by \code{(nu-1)/2}
\item \code{"separation"} simulates from the identity inverse Wishart
covariance matrix with \code{nu} degrees of freedom.  This is then
converted to a covariance matrix and augmented with the modeled
standard deviations.  While computationally more complex than the
\code{"lkj"} prior, it performs better when the covariance matrix
size is greater or equal to 10
\item \code{"auto"} chooses \code{"lkj"} when the dimension of the
matrix is less than 10 and \code{"separation"} when greater
than equal to 10.
}}

\item{omegaXform}{When taking \code{omega} values from the \code{thetaMat}
simulations (using the separation strategy for covariance
simulation), how should the \code{thetaMat} values be turned int
standard deviation values:
\itemize{
\item \code{identity} This is when standard deviation values are
directly modeled by the \code{params} and \code{thetaMat} matrix
\item \code{variance} This is when the \code{params} and \code{thetaMat}
simulates the variance that are directly modeled by the
\code{thetaMat} matrix
\item \code{log} This is when the \code{params} and \code{thetaMat}
simulates \code{log(sd)}
\item \code{nlmixrSqrt} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix
with the \verb{x\\^2} modeled along the diagonal.  This only works
with a diagonal matrix.
\item \code{nlmixrLog} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix
with the \verb{exp(x\\^2)} along the diagonal.  This only works
with a diagonal matrix.
\item \code{nlmixrIdentity} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix.
This only works with a diagonal matrix.
}}

\item{nSub}{Number between subject variabilities (\code{ETAs}) simulated for every
realization of the parameters.}

\item{thetaMat}{Named theta matrix.}

\item{thetaLower}{Lower bounds for simulated population parameter
variability (by default \code{-Inf})}

\item{thetaUpper}{Upper bounds for simulated population unexplained
variability (by default \code{Inf})}

\item{thetaDf}{The degrees of freedom of a t-distribution for
simulation.  By default this is \code{NULL} which is
equivalent to \code{Inf} degrees, or to simulate from a normal
distribution instead of a \code{t}-distribution.}

\item{thetaIsChol}{Indicates if the \code{theta} supplied is a
Cholesky decomposed matrix instead of the traditional
symmetric matrix.}

\item{nStud}{Number virtual studies to characterize uncertainty in estimated
parameters.}

\item{sigma}{Named sigma covariance or Cholesky decomposition of a
covariance matrix.  The names of the columns indicate
parameters that are simulated.  These are simulated for every
observation in the solved system. When \code{sigma} is \code{NA} and you are using it with a
\code{rxode2} ui model, the unexplained variability described by
the \code{sigma} matrix are set to zero.}

\item{sigmaLower}{Lower bounds for simulated unexplained variability (by default -Inf)}

\item{sigmaUpper}{Upper bounds for simulated unexplained variability (by default Inf)}

\item{sigmaDf}{Degrees of freedom of the sigma t-distribution.  By
default it is equivalent to \code{Inf}, or a normal distribution.}

\item{sigmaIsChol}{Boolean indicating if the sigma is in the
Cholesky decomposition instead of a symmetric covariance}

\item{sigmaSeparation}{separation strategy for sigma;

Tells the type of separation strategy when
simulating covariance with parameter uncertainty with standard
deviations modeled in the \code{thetaMat} matrix.
\itemize{
\item \code{"lkj"} simulates the correlation matrix from the
\code{rLKJ1} matrix with the distribution parameter \code{eta}
equal to the degrees of freedom \code{nu} by \code{(nu-1)/2}
\item \code{"separation"} simulates from the identity inverse Wishart
covariance matrix with \code{nu} degrees of freedom.  This is then
converted to a covariance matrix and augmented with the modeled
standard deviations.  While computationally more complex than the
\code{"lkj"} prior, it performs better when the covariance matrix
size is greater or equal to 10
\item \code{"auto"} chooses \code{"lkj"} when the dimension of the
matrix is less than 10 and \code{"separation"} when greater
than equal to 10.
}}

\item{sigmaXform}{When taking \code{sigma} values from the \code{thetaMat}
simulations (using the separation strategy for covariance
simulation), how should the \code{thetaMat} values be turned int
standard deviation values:
\itemize{
\item \code{identity} This is when standard deviation values are
directly modeled by the \code{params} and \code{thetaMat} matrix
\item \code{variance} This is when the \code{params} and \code{thetaMat}
simulates the variance that are directly modeled by the
\code{thetaMat} matrix
\item \code{log} This is when the \code{params} and \code{thetaMat}
simulates \code{log(sd)}
\item \code{nlmixrSqrt} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix
with the \verb{x\\^2} modeled along the diagonal.  This only works
with a diagonal matrix.
\item \code{nlmixrLog} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix
with the \verb{exp(x\\^2)} along the diagonal.  This only works
with a diagonal matrix.
\item \code{nlmixrIdentity} This is when the \code{params} and
\code{thetaMat} simulates the inverse cholesky decomposed matrix.
This only works with a diagonal matrix.
}}

\item{nCoresRV}{Number of cores used for the simulation of the
sigma variables.  By default this is 1. To reproduce the results
you need to run on the same platform with the same number of
cores. This is the reason this is set to be one, regardless of
what the number of cores are used in threaded ODE solving.}

\item{nObs}{Number of observations to simulate (with \code{sigma} matrix)}

\item{dfSub}{Degrees of freedom to sample the between subject variability matrix from the
inverse Wishart distribution (scaled) or scaled inverse chi squared distribution.}

\item{dfObs}{Degrees of freedom to sample the unexplained variability matrix from the
inverse Wishart distribution (scaled) or scaled inverse chi squared distribution.}

\item{simSubjects}{boolean indicated rxode2 should simulate subjects in studies (\code{TRUE},
default) or studies (\code{FALSE})}

\item{simVariability}{determines if the variability is simulated.
When \code{NA} (default) this is determined by the solver.}
}
\value{
a data frame with the simulated subjects
}
\description{
Simulate Parameters from a Theta/Omega specification
}
\author{
Matthew L.Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tran.R
\name{rxode2parseAssignPointerTranslation}
\alias{rxode2parseAssignPointerTranslation}
\title{This sets function gets the currently assigned function pointer assignments}
\usage{
rxode2parseAssignPointerTranslation(var)
}
\arguments{
\item{var}{List of packages where pointer assignment will be called.}
}
\value{
Nothing, called for side effects
}
\description{
This sets function gets the currently assigned function pointer assignments
}
\examples{
rxode2parseAssignPointerTranslation("rxode2")
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.rxSolve}
\alias{plot.rxSolve}
\alias{plot.rxSolveConfint1}
\alias{plot.rxSolveConfint2}
\title{Plot rxode2 objects}
\usage{
\method{plot}{rxSolve}(x, y, ..., log = "", xlab = "Time", ylab = "")

\method{plot}{rxSolveConfint1}(x, y, ..., xlab = "Time", ylab = "", log = "")

\method{plot}{rxSolveConfint2}(x, y, ..., xlab = "Time", ylab = "", log = "")
}
\arguments{
\item{x}{rxode2 object to plot}

\item{y}{Compartments or left-hand-side values to plot either as a bare name
or as a character vector}

\item{...}{Ignored}

\item{log}{Should "" (neither x nor y), "x", "y", or "xy" (or "yx") be
log-scale?}

\item{xlab, ylab}{The x and y axis labels}
}
\value{
A ggplot2 object
}
\description{
Plot rxode2 objects
}
\seealso{
Other rxode2 plotting:
\code{\link[=rxTheme]{rxTheme()}}
}
\concept{rxode2 plotting}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxShiny.R
\name{rxShiny}
\alias{rxShiny}
\alias{rxShiny.rxSolve}
\alias{rxShiny.default}
\title{Use Shiny to help develop an rxode2 model}
\usage{
rxShiny(
  object,
  params = NULL,
  events = NULL,
  inits = NULL,
  ...,
  data = data.frame()
)

\method{rxShiny}{rxSolve}(
  object,
  params = NULL,
  events = NULL,
  inits = NULL,
  ...,
  data = data.frame()
)

\method{rxShiny}{default}(
  object = NULL,
  params = NULL,
  events = NULL,
  inits = NULL,
  ...,
  data = data.frame()
)
}
\arguments{
\item{object}{A rxode2 family of objects. If not supplied a
2-compartment indirect effect model is used.  If it is
supplied, use the model associated with the rxode2 object for
the model exploration.}

\item{params}{Initial parameters for model}

\item{events}{Event information (currently ignored)}

\item{inits}{Initial estimates for model}

\item{...}{Other arguments passed to rxShiny.  Currently doesn't
do anything.}

\item{data}{Any data that you would like to plot.  If the data has
a \code{time} variable as well as a compartment or calculated
variable that matches the rxode2 model, the data will be added
to the plot of a specific compartment or calculated variable.}
}
\value{
Nothing; Starts a shiny server
}
\description{
Use Shiny to help develop an rxode2 model
}
\author{
Zufar Mulyukov and Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{as_tibble.rxEt}
\alias{as_tibble.rxEt}
\title{Convert to tbl}
\usage{
as_tibble.rxEt(x, ...)
}
\arguments{
\item{x}{rxode2 event table}

\item{...}{Other arguments to \code{as_tibble}}
}
\value{
tibble of event table
}
\description{
Convert to tbl
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxPrune.R
\name{.rxPrune}
\alias{.rxPrune}
\title{Internal Pruning function}
\usage{
.rxPrune(x, envir = parent.frame(), strAssign = list())
}
\arguments{
\item{x}{List of quoted lines}

\item{envir}{Environment where information is stored}

\item{strAssign}{string assignment list from  model variables}
}
\value{
Pruned model code
}
\description{
Internal Pruning function
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{.udfEnvReset}
\alias{.udfEnvReset}
\title{Lock/Unlock environment for getting R user functions}
\usage{
.udfEnvReset(lock = TRUE)
}
\arguments{
\item{lock}{logical to see if environment to look for user defined
functions is locked.  If it is locked then environments are not
assigned.  When NULL returns lock status}
}
\value{
lock status
}
\description{
Lock/Unlock environment for getting R user functions
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxsolve.R
\name{rxIsStiff}
\alias{rxIsStiff}
\title{Check whether an ODE solving method is a purely stiff solver}
\usage{
rxIsStiff(method)
}
\arguments{
\item{method}{A character vector of method names or an integerish vector of
method codes (as returned by \code{\link[=odeMethodToInt]{odeMethodToInt()}}).  Vectorised.}
}
\value{
A logical vector the same length as \code{method}.  \code{TRUE} if the
corresponding method is a purely stiff solver.
}
\description{
Returns \code{TRUE} for methods that are designed exclusively for stiff systems
and will never switch to a non-stiff algorithm.  Solvers like \code{"lsoda"} and
\code{"liblsoda"} that automatically switch between stiff and non-stiff algorithms
return \code{FALSE}.
}
\details{
Stiff-only methods are:
\code{"ros4"} (13), \code{"iem"} (14), \code{"cvode"} (21), \code{"ros43"} (31), \code{"ros6"} (32),
\code{"backwardEuler"} (33), \code{"gauss6"} (34), \code{"iiic6"} (35), \code{"radauiia5"} (36),
\code{"geng5"} (37), \code{"sdirk43"} (38), \code{"bdf"} (107).
}
\examples{
rxIsStiff("bdf")                           # TRUE
rxIsStiff("lsoda")                         # FALSE (switches)
rxIsStiff("liblsoda")                      # FALSE (switches)
rxIsStiff(c("cvode", "dop853", "ros43"))   # TRUE FALSE TRUE

}
\seealso{
\code{\link[=rxIsNonStiff]{rxIsNonStiff()}}, \code{\link[=rxIsImplicit]{rxIsImplicit()}}, \code{\link[=odeMethodToInt]{odeMethodToInt()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxChain}
\alias{rxChain}
\alias{+.solveRxDll}
\title{rxChain  Chain or add item to solved system of equations}
\usage{
rxChain(obj1, obj2)

\method{+}{solveRxDll}(obj1, obj2)
}
\arguments{
\item{obj1}{Solved object.}

\item{obj2}{New object to be added/piped/chained to solved object.}
}
\value{
When \code{newObject} is an event table, return a new
solved object with the new event table.
}
\description{
Add item to solved system of equations
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxLhs}
\alias{rxLhs}
\title{Left handed Variables}
\usage{
rxLhs(obj)
}
\arguments{
\item{obj}{rxode2 family of objects}
}
\value{
a character vector listing the calculated parameters
}
\description{
This returns the model calculated variables
}
\seealso{
\code{\link{rxode2}}

Other Query model information:
\code{\link[=rxDfdy]{rxDfdy()}},
\code{\link[=rxInits]{rxInits()}},
\code{\link[=rxModelVars]{rxModelVars()}},
\code{\link[=rxParams]{rxParams()}},
\code{\link[=rxState]{rxState()}}
}
\author{
Matthew L.Fidler
}
\concept{Query model information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assert.R
\name{assertCompartmentExists}
\alias{assertCompartmentExists}
\alias{testCompartmentExists}
\title{Verify that the compartment exists in a model}
\usage{
assertCompartmentExists(ui, x)

testCompartmentExists(ui, x)
}
\arguments{
\item{ui}{is the model to test}

\item{x}{The value to test (can be a vector of strings)}
}
\value{
the value of the compartment that exists; if it is a vector
returns the first item that matches
}
\description{
Verify that the compartment exists in a model
}
\section{Functions}{
\itemize{
\item \code{testCompartmentExists()}: Test if compartment exists

}}
\seealso{
Other Assertions:
\code{\link[=assertCompartmentName]{assertCompartmentName()}},
\code{\link[=assertCompartmentNew]{assertCompartmentNew()}},
\code{\link[=assertRxUi]{assertRxUi()}},
\code{\link[=assertVariableExists]{assertVariableExists()}},
\code{\link[=assertVariableNew]{assertVariableNew()}},
\code{\link[=testIniDf]{testIniDf()}},
\code{\link[=testRxUnbounded]{testRxUnbounded()}}
}
\author{
Matthew Fidler & Bill Denney
}
\concept{Assertions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxCompile}
\alias{rxCompile}
\alias{rxCompile.rxModelVars}
\alias{rxCompile.character}
\alias{rxCompile.rxDll}
\alias{rxCompile.rxode2}
\title{Compile a model if needed}
\usage{
rxCompile(
  model,
  dir,
  prefix,
  force = FALSE,
  modName = NULL,
  package = NULL,
  ...
)

\method{rxCompile}{rxModelVars}(
  model,
  dir = NULL,
  prefix = NULL,
  force = FALSE,
  modName = NULL,
  package = NULL,
  eventSensCode = rep("", 13L),
  ...
)

\method{rxCompile}{character}(
  model,
  dir = NULL,
  prefix = NULL,
  force = FALSE,
  modName = NULL,
  package = NULL,
  eventSensCode = rep("", 13L),
  ...
)

\method{rxCompile}{rxDll}(model, ...)

\method{rxCompile}{rxode2}(model, ...)
}
\arguments{
\item{model}{This is the ODE model specification.  It can be:
\itemize{
\item a string containing the set of ordinary differential
equations (ODE) and other expressions defining the changes in
the dynamic system.
\item a file name where the ODE system equation is contained
}

An ODE expression enclosed in \verb{\\\{\\\}}

(see also the \code{filename} argument). For
details, see the sections \dQuote{Details} and
\verb{rxode2 Syntax} below.}

\item{dir}{This is the model directory where the C file will be
stored for compiling.

\if{html}{\out{<div class="sourceCode">}}\preformatted{If unspecified, the C code is stored in a temporary directory,
then the model is compiled and moved to the current directory.
Afterwards the C code is removed.

If specified, the C code is stored in the specified directory
and then compiled in that directory.  The C code is not removed
after the DLL is created in the same directory.  This can be
useful to debug the c-code outputs.
}\if{html}{\out{</div>}}}

\item{prefix}{is a string indicating the prefix to use in the C
based functions.  If missing, it is calculated based on file
name, or md5 of parsed model.}

\item{force}{is a boolean stating if the (re)compile should be
forced if rxode2 detects that the models are the same as already
generated.}

\item{modName}{a string to be used as the model name. This string
is used for naming various aspects of the computations,
including generating C symbol names, dynamic libraries,
etc. Therefore, it is necessary that \code{modName} consists of
simple ASCII alphanumeric characters starting with a letter.}

\item{package}{Package name for pre-compiled binaries.}

\item{...}{Other arguments sent to the \code{\link[=rxTrans]{rxTrans()}}
function.}

\item{eventSensCode}{character vector of length 13 giving the C body
lines for the event ("jump") sensitivity helper functions
\code{dLag}/\code{dF}/\code{dRate}/\code{dDur}/\code{d2F}/\code{d2Lag}/\code{d2Rate}/\code{d2Dur}/\code{d3F}/\code{dFQ}/
\code{dLagJac}/\code{dLagQ}/\code{dDurQ} (in that order).  These are the per-model
dosing-parameter total-derivative assignments produced by the
event-sensitivity code generator and are emitted verbatim into the
corresponding generated functions.  The default of thirteen empty
strings produces trivial (no-op) helpers, which is what
non-sensitivity models and \code{eventSens = "fd"} models use.}
}
\value{
An rxDll object that has the following components
\itemize{
\item \code{dll} DLL path
\item \code{model} model specification
\item \code{.c} A function to call C code in the correct context from the DLL
using the \code{\link[=.C]{.C()}} function.
\item \code{.call} A function to call C code in the correct context from the DLL
using the \code{\link[=.Call]{.Call()}} function.
\item \code{args} A list of the arguments used to create the rxDll object.
}
}
\description{
This is the compilation workhorse creating the rxode2 model DLL
files.
}
\seealso{
\code{\link[=rxode2]{rxode2()}}
}
\author{
Matthew L.Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjoint.R
\name{.rxAdjointSolveBuild}
\alias{.rxAdjointSolveBuild}
\title{Build the symbolic part of a full-trajectory adjoint solve (compile once)}
\usage{
.rxAdjointSolveBuild(object, calcSens, events, adjStates = NULL)
}
\arguments{
\item{object}{model definition accepted by \code{\link[=rxode2]{rxode2()}} (text,
\code{rxode2} object, or function/ui).}

\item{calcSens}{character vector of parameter names to differentiate with
respect to.}

\item{events}{event table / data used to define dosing.}

\item{adjStates}{character vector of output states of interest.  Defaults to
all ODE states (full forward-sensitivity parity).}
}
\value{
an opaque list consumed by \code{\link[=.rxAdjointSolveEvalC]{.rxAdjointSolveEvalC()}}.
}
\description{
Performs all symbolic work (Jacobian, dose-dual expressions, model
compilation) once and returns a reusable object for \code{\link[=.rxAdjointSolveEvalC]{.rxAdjointSolveEvalC()}},
the C++-backed counterpart of \code{\link[=.rxAdjointSolve]{.rxAdjointSolve()}}.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxUdfUiFlag}
\alias{rxUdfUiFlag}
\title{Is the expression actually a flag that can be used in the rxUdfUi
functions?}
\usage{
rxUdfUiFlag(expr, arg = "arg", funName = "fun", env = baseenv())
}
\arguments{
\item{expr}{expression to evaluate}

\item{arg}{argument name for error messages for the argument name
in the \code{rxUdfUi} extension when the expression is not logical}

\item{funName}{function name for the error message for the argument
name in the \code{rxUdfUi} extension when the expression is not
logical.}

\item{env}{the environment in which to evaluate the expression (in case it is numeric)}
}
\value{
logical value of the expression if it can be evaluated to a
scalar \code{TRUE}/\code{FALSE} value, otherwise an error is thrown.
}
\description{
This is useful when writing replacement
UI functions
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piping.R
\name{.quoteCallInfoLines}
\alias{.quoteCallInfoLines}
\title{Returns quoted call information}
\usage{
.quoteCallInfoLines(callInfo, envir = parent.frame(), iniDf = NULL)
}
\arguments{
\item{callInfo}{Call information}

\item{envir}{Environment for evaluation (if needed)}

\item{iniDf}{The parent model \code{iniDf} when piping in a \code{ini} block
(\code{NULL} otherwise)}
}
\value{
Quote call information.  for \code{name=expression}, change to
\code{name<-expression} in quoted call list. For expressions that are
within brackets ie \code{{}}, unlist the brackets as if they were
called in one single sequence.
}
\description{
Returns quoted call information
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err-foceiBase.R
\name{.rxGetVarianceForErrorType}
\alias{.rxGetVarianceForErrorType}
\title{Get Variance for error type}
\usage{
.rxGetVarianceForErrorType(env, pred1)
}
\arguments{
\item{env}{Environment}

\item{pred1}{Pred for one end-point}
}
\value{
Variance error type
}
\description{
Get Variance for error type
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{.rxGetSeed}
\alias{.rxGetSeed}
\title{Get the random seed for the session}
\usage{
.rxGetSeed()
}
\value{
A list with the following components:
\itemize{
\item \code{seed} The random seed for the session.  This is the value of
\code{.Random.seed} in the global environment.
\item \code{kind} The random number generator kind for the session.  This is
the value of \code{RNGkind()} for the session.
\item \code{rxseed} The rxode2 random seed for the session.  This
is the value of \code{rxGetSeed()} for the session.
}
}
\description{
Get the random seed for the session
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{rxRateDur}
\alias{rxRateDur}
\alias{[.rxRateDur}
\alias{as.rxRateDur}
\alias{as.character.rxRateDur}
\alias{[[.rxRateDur}
\title{Creates a rxRateDur object}
\usage{
rxRateDur(x)

\method{[}{rxRateDur}(x, ...)

as.rxRateDur(x)

\method{as.character}{rxRateDur}(x, ...)

\method{[[}{rxRateDur}(x, ...)
}
\arguments{
\item{x}{rxRateDur data}

\item{...}{Other parameters}
}
\value{
rxRateDur object
}
\description{
This is primarily to display information about rate
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxSolveSetup}
\alias{rxSolveSetup}
\title{See if the memory is installed for a solve}
\usage{
rxSolveSetup()
}
\value{
boolean saying if the memnory is currently free for rxode2
}
\description{
See if the memory is installed for a solve
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxUiGet.R, R/rxsolve.R
\name{rxUiDeparse}
\alias{rxUiDeparse}
\alias{rxUiDeparse.lotriFix}
\alias{rxUiDeparse.default}
\alias{rxUiDeparse.rxControl}
\title{This is a generic function for deparsing certain objects when
printing out a rxode2 object.  Currently it is used for any meta-information}
\usage{
rxUiDeparse(object, var)

\method{rxUiDeparse}{lotriFix}(object, var)

\method{rxUiDeparse}{default}(object, var)

\method{rxUiDeparse}{rxControl}(object, var)
}
\arguments{
\item{object}{object to be deparsed}

\item{var}{variable name to be assigned}
}
\value{
parsed R expression that can be used for printing and
\code{as.function()} calls.
}
\description{
This is a generic function for deparsing certain objects when
printing out a rxode2 object.  Currently it is used for any meta-information

rxUiDeparse.rxControl(rxControl(covsInterpolation="linear", method="dop853",
naInterpolation="nocb", keepInterpolation="nocb", sigmaXform="variance",
omegaXform="variance", returnType="data.frame", sumType="fsum", prodType="logify"),
"ctl")
}
\examples{

mat <- matrix(c(1, 0.1, 0.1, 1), 2, 2, dimnames=list(c("a", "b"), c("a", "b")))

rxUiDeparse(matrix(c(1, 0.1, 0.1, 1), 2, 2, dimnames=list(c("a", "b"), c("a", "b"))), "x")
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piping-ini.R
\name{.iniHandleAppend}
\alias{.iniHandleAppend}
\title{Reorder rows in iniDf}
\usage{
.iniHandleAppend(expr, rxui, envir, append)
}
\arguments{
\item{expr}{Expression for parsing}

\item{rxui}{User interface function}

\item{envir}{Environment for parsing}

\item{append}{Reorder theta parameters.  \code{NULL} means no change to
parameter order.  A parameter name (as a character string) means to put the
new parameter after the named parameter.  A number less than or equal to
zero means to put the parameter at the beginning of the list.  A number
greater than the last parameter number means to put the parameter at the
end of the list.}
}
\value{
Nothing, called for side effects
}
\description{
Reorder rows in iniDf
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{et}
\alias{et}
\alias{et.rxode2}
\alias{et.function}
\alias{et.rxUi}
\alias{et.rxSolve}
\alias{et.rxParams}
\alias{et.default}
\title{Event Table Function}
\usage{
et(x, ..., envir = parent.frame())

\method{et}{rxode2}(x, ..., envir = parent.frame())

\method{et}{`function`}(x, ..., envir = parent.frame())

\method{et}{rxUi}(x, ..., envir = parent.frame())

\method{et}{rxSolve}(x, ..., envir = parent.frame())

\method{et}{rxParams}(x, ..., envir = parent.frame())

\method{et}{default}(
  x,
  ...,
  time = NULL,
  amt = NULL,
  evid = NULL,
  cmt = NULL,
  ii = NULL,
  addl = NULL,
  ss = NULL,
  rate = NULL,
  dur = NULL,
  until = NULL,
  id = NULL,
  amountUnits = NULL,
  timeUnits = NULL,
  addSampling = NULL,
  envir = parent.frame(),
  by = NULL,
  length.out = NULL
)
}
\arguments{
\item{x}{This is the first argument supplied to the event table.
This is named to allow \code{et} to be used in a pipe-line
with arbitrary objects.}

\item{...}{Times or event tables.  They can also be one of the named arguments below.}

\item{envir}{the \code{\link[base]{environment}} in which \code{expr} is to
    be evaluated.  May also be \code{NULL}, a list, a data frame,
    a pairlist or an integer as specified to \code{\link[base]{sys.call}}.}

\item{time}{Time is the time of the dose or the sampling times.
This can also be unspecified and is determined by the object
type (list or numeric/integer).}

\item{amt}{Amount of the dose. If specified, this assumes a dosing
record, instead of a sampling record.}

\item{evid}{Event ID; This can be:\tabular{ll}{
   Numeric Value \tab Description \cr
   0 \tab An observation. This can also be specified as \code{evid=obs} \cr
   1 \tab A dose observation.  This can also be specified as \code{evid=dose} \cr
   2 \tab A non-dose event. This can also be specified as \code{evid=other} \cr
   3 \tab A reset event.  This can also be specified as \code{evid=reset}. \cr
   4 \tab Dose and reset event.  This can also be specified as \code{evid=doseReset} or \code{evid=resetDose} \cr
}


Note a reset event resets all the compartment values to zero and turns off all infusions.}

\item{cmt}{Compartment name or number.  If a number, this is an
integer starting at 1.  Negative compartments turn off a
compartment. If the compartment is a name, the compartment name
is changed to the correct state/compartment number before
running the simulation.  For a compartment named "-cmt" the
compartment is turned off.

\if{html}{\out{<div class="sourceCode">}}\preformatted{Can also specify `cmt` as `dosing.to`,
`dose.to`, `doseTo`, `dosingTo`, and
`state`.
}\if{html}{\out{</div>}}}

\item{ii}{When specifying a dose, this is the inter-dose interval
for \code{ss}, \code{addl} and \code{until} options (described below).}

\item{addl}{The number of additional doses at a inter-dose
interval after one dose.}

\item{ss}{Steady state flag;  It can be one of:\tabular{ll}{
   Value \tab Description \cr
   0 \tab This dose is not a steady state dose \cr
   1 \tab This dose is a steady state dose with the between/inter-dose interval of \code{ii} \cr
   2 \tab Superposition steady state \cr
}


When \code{ss=2} the steady state dose that uses the super-position
principle to allow more complex steady states, like 10 mg in the
morning and 20 mg at night, or dosing at 8 am 12 pm and 8 pm
instead of every 12 hours.  Since it uses the super positioning
principle, it only makes sense when you know the kinetics are
linear.

All other values of \code{SS} are currently invalid.}

\item{rate}{When positive, this is the rate of infusion.  Otherwise:\tabular{ll}{
   Value \tab Description \cr
   0 \tab No infusion is on this record \cr
   -1 \tab Modeled rate (in rxode2:\verb{rate(cmt) =}); Can be \code{et(rate=model)}. \cr
   -2 \tab Modeled duration (in rxode2: \verb{dur(cmt) =}); Can be\code{et(dur=model)} or \code{et(rate=dur)}. \cr
}


When a modeled bioavailability is applied to positive rates
(\code{rate} > 0), the duration of infusion is changed. This is
because the data specify the rate and amount, the only think that
modeled bioavailability can affect is duration.

If instead you want the modeled bioavailability to increase the
rate of infusion instead of the duration of infusion, specify the
\code{dur} instead or model the duration with \code{rate=2}.}

\item{dur}{Duration of infusion.  When \code{amt} and \code{dur}
are specified the rate is calculated from the two data items.
When \code{dur} is specified instead of \code{rate}, the
bioavailability changes will increase rate instead of
duration.}

\item{until}{This is the time until the dosing should end.  It can
be an easier way to figure out how many additional doses are
needed over your sampling period.}

\item{id}{A integer vector of ids to add or remove from the event
table.  If the event table is identical for each ID, then you
may expand it to include all the ids in this vector.  All the
negative ids in this vector will be removed.}

\item{amountUnits}{The units for the dosing records (\code{amt})}

\item{timeUnits}{The units for the time records (\code{time})}

\item{addSampling}{This is a boolean indicating if a sampling time
should be added at the same time as a dosing time.  By default
this is \code{FALSE}.}

\item{by}{number: increment of the sequence.}

\item{length.out}{desired length of the sequence.  A
    non-negative number, which for \code{seq} and \code{seq.int} will be
    rounded up if fractional.}
}
\value{
A new event table
}
\description{
Event Table Function
}
\examples{

\dontrun{

library(rxode2)
library(units)

# Model from rxode2 tutorial
# Using a nlmixr2 style function

mod1 <-function(){
  ini({
    KA <- 2.94E-01
    CL <- 1.86E+01
    V2 <- 4.02E+01
    Q <- 1.05E+01
    V3 <- 2.97E+02
    Kin <- 1
    Kout <- 1
    EC50 <- 200
  })
 model({
    C2 <- centr/V2
    C3 <- peri/V3
    d/dt(depot) <- -KA*depot
    d/dt(centr) <- KA*depot - CL*C2 - Q*C2 + Q*C3
    d/dt(peri)  <-                    Q*C2 - Q*C3
    d/dt(eff)   <- Kin - Kout*(1-C2/(EC50+C2))*eff
 })
}

## These are making the more complex regimens of the rxode2 tutorial

## bid for 5 days
bid <- et(timeUnits="hr") |>
       et(amt=10000,ii=12,until=set_units(5, "days"))

## qd for 5 days
qd <- et(timeUnits="hr") |>
      et(amt=20000,ii=24,until=set_units(5, "days"))

## bid for 5 days followed by qd for 5 days

et <- seq(bid,qd) |>
      et(seq(0,11*24,length.out=100))

bidQd <- rxSolve(mod1, et)

plot(bidQd, C2)


## Now Infusion for 5 days followed by oral for 5 days

##  note you can dose to a named compartment instead of using the compartment number
infusion <- et(timeUnits = "hr") |>
      et(amt=10000, rate=5000, ii=24, until=set_units(5, "days"), cmt="centr")


qd <- et(timeUnits = "hr") |>
  et(amt=10000, ii=24, until=set_units(5, "days"), cmt="depot")

et <- seq(infusion,qd)

infusionQd <- rxSolve(mod1, et)

plot(infusionQd, C2)

## 2wk-on, 1wk-off

qd <- et(timeUnits = "hr") |>
      et(amt=10000, ii=24, until=set_units(2, "weeks"), cmt="depot")

et <- seq(qd, set_units(1,"weeks"), qd) |>
     add.sampling(set_units(seq(0, 5.5,by=0.005),weeks))

wkOnOff <- rxSolve(mod1, et)

plot(wkOnOff, C2)

## You can also repeat the cycle easily with the rep function

qd <-et(timeUnits = "hr") |>
     et(amt=10000, ii=24, until=set_units(2, "weeks"), cmt="depot")

et <- etRep(qd, times=4, wait=set_units(1,"weeks")) |>
      add.sampling(set_units(seq(0, 12.5,by=0.005),weeks))

repCycle4 <- rxSolve(mod1, et)

plot(repCycle4, C2)

}
}
\references{
Wang W, Hallow K, James D (2015). "A Tutorial on rxode2: Simulating
Differential Equation Pharmacometric Models in R." CPT:
Pharmacometrics and Systems Pharmacology, 5(1), 3-10. ISSN 2163-8306
}
\seealso{
\code{\link{eventTable}}, \code{\link{add.sampling}},
\code{\link{add.dosing}}, \code{\link{et}},
\code{\link{etRep}}, \code{\link{etRbind}},
\code{\link{rxode2}}
}
\author{
Matthew L Fidler, Wenping Wang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{.udfEnvSet}
\alias{.udfEnvSet}
\title{Setup the UDF environment (for querying user defined funtions)}
\usage{
.udfEnvSet(env)
}
\arguments{
\item{env}{environment where user defined functions are queried. If NULL return current environment}
}
\value{
environment
}
\description{
Setup the UDF environment (for querying user defined funtions)
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linCmt.R
\name{.solComp3}
\alias{.solComp3}
\title{Calculate the lambdas and coefficients of the three compartment model}
\usage{
.solComp3(k10, k12, k21, k13, k31)
}
\arguments{
\item{k10}{elimination rate}

\item{k12}{rate from central to peripheral compartment}

\item{k21}{rate from peripheral to central compartment}

\item{k13}{rate from central to peripheral compartment #2}

\item{k31}{rate from peripheral compartment #2 to central}
}
\value{
List with \code{L} vector and matrices \code{C1}, \code{C2} and \code{C3}
}
\description{
Calculate the lambdas and coefficients of the three compartment model
}
\examples{
.solComp3(k10=0.1, k12=3, k21=1, k13=2, k31=0.5)
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{.rxC}
\alias{.rxC}
\title{Return the C code of an internal function}
\usage{
.rxC(fun)
}
\arguments{
\item{fun}{is the string of a function that you wish to get the C
code for}
}
\value{
C code if found (as a string) or NULL if not found
}
\description{
Return the C code of an internal function
}
\author{
Matthew Fider
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikT}
\alias{llikT}
\title{Log likelihood of T and it's derivatives (from stan)}
\usage{
llikT(x, df, mean = 0, sd = 1, full = FALSE)
}
\arguments{
\item{x}{Observation}

\item{df}{degrees of freedom (\eqn{> 0}, maybe non-integer).  \code{df
      = Inf} is allowed.}

\item{mean}{Mean for the likelihood}

\item{sd}{Standard deviation for the likelihood}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the log pdf value of with \code{dDf}
\code{dMean} and \code{dSd} that has the derivatives with respect to the parameters at
the observation time-point
}
\description{
Log likelihood of T and it's derivatives (from stan)
}
\details{
In an \code{rxode2()} model, you can use \code{llikT()} but you have to
use all arguments.  You can also get the derivative of \code{df}, \code{mean} and \code{sd} with
\code{llikTDdf()}, \code{llikTDmean()} and \code{llikTDsd()}.
}
\examples{

\donttest{
x <- seq(-3, 3, length.out = 21)

llikT(x, 7, 0, 1)

llikT(x, 15, 0, 1, full=TRUE)

et <- et(-3, 3, length.out=10)
et$nu <- 7
et$mean <- 0
et$sd <- 1

model <- function() {
  model({
    fx <- llikT(time, nu, mean, sd)
    dDf <- llikTDdf(time, nu, mean, sd)
    dMean <- llikTDmean(time, nu, mean, sd)
    dSd   <- llikTDsd(time, nu, mean, sd)
  })
}

rxSolve(model, et)
}

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piping-model.R
\name{.modelHandleModelLines}
\alias{.modelHandleModelLines}
\title{Handle model lines}
\usage{
.modelHandleModelLines(
  modelLines,
  rxui,
  modifyIni = FALSE,
  append = NULL,
  auto = getOption("rxode2.autoVarPiping", TRUE),
  cov = NULL,
  envir
)
}
\arguments{
\item{modelLines}{The model lines that are being considered}

\item{rxui}{The rxode2 UI object}

\item{modifyIni}{Should the ini({}) be considered}

\item{append}{This is a boolean to determine if the lines are
appended in piping.  The possible values for this is:
\itemize{
\item \code{TRUE} which is when the lines are appended to the model instead of replaced
\item \code{FALSE} when the lines are replaced in the model (default)
\item \code{NA} is when the lines are pre-pended to the model instead of replaced
\item \verb{lhs expression}, which will append the lines after the last observed line of the expression \code{lhs}
}}

\item{auto}{This boolean tells if piping automatically selects the
parameters should be characterized as a population parameter,
between subject variability, or a covariate.  When \code{TRUE} this
automatic selection occurs.  When \code{FALSE} this automatic
selection is turned off and everything is added as a covariate
(which can be promoted to a parameter with the \code{ini} statement).
By default this is \code{TRUE}, but it can be changed by
\code{options(rxode2.autoVarPiping=FALSE)}.}

\item{cov}{is a character vector of variables that should be
assumed to be covariates.  This will override automatic promotion
to a population parameter estimate (or an eta)}

\item{envir}{Environment for evaluation}
}
\value{
New UI
}
\description{
Handle model lines
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{boxCox}
\alias{boxCox}
\alias{boxCoxInv}
\alias{yeoJohnson}
\alias{yeoJohnsonInv}
\title{boxCox/yeoJohnson and inverse boxCox/yeoJohnson functions}
\usage{
boxCox(x, lambda = 1)

boxCoxInv(x, lambda = 1)

yeoJohnson(x, lambda = 1)

yeoJohnsonInv(x, lambda = 1)
}
\arguments{
\item{x}{input value(s) to transform}

\item{lambda}{lambda value for the transformation}
}
\value{
values from boxCox and boxCoxInv
}
\description{
boxCox/yeoJohnson and inverse boxCox/yeoJohnson functions
}
\examples{

boxCox(10, 0.5)

boxCoxInv(4.32, 0.5)

yeoJohnson(10, 0.5)

yeoJohnsonInv(4.32, 0.5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxExpandSens3_}
\alias{rxExpandSens3_}
\title{Expand third order sensitivity}
\usage{
rxExpandSens3_(state, s1, s2, s3)
}
\arguments{
\item{state}{is the state to expand}

\item{s1}{is the newly added (outer) sensitivity variable}

\item{s2}{is the second sensitivity variable}

\item{s3}{is the third sensitivity variable}
}
\value{
data frame of symengine third order sensitivity code
}
\description{
Builds the symengine code for the third order forward sensitivity
d/dt(d^3 state / d s1 d s2 d s3) by total-differentiating the stored
second order sensitivity RHS (for the pair s2,s3) with respect to the
newly added variable s1.  Mirrors \code{rxExpandSens2_}; assumes the
first and second order sensitivities (and the Jacobian) are already
loaded in the symengine \code{model} environment.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{rxUnloadAll}
\alias{rxUnloadAll}
\title{Unloads all rxode2 compiled DLLs}
\usage{
rxUnloadAll(set = TRUE)
}
\arguments{
\item{set}{If specified and \code{TRUE}, unloads all rxode2 dlls.  If
specified and \code{FALSE}, block a simple \code{rxUnloadAll()} will not
actually unload all dlls (helps with CRAN ASAN check)}
}
\value{
boolean telling if \code{rxUnloadAll()} completed the unloading procedure
}
\description{
Unloads all rxode2 compiled DLLs
}
\examples{

# print(rxUnloadAll())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{.udfExists}
\alias{.udfExists}
\title{See if the UI function exists in given environment.}
\usage{
.udfExists(fun, nargs, envir, doList = TRUE)
}
\arguments{
\item{fun}{Function to check}

\item{nargs}{Number of args to check}

\item{envir}{Environment to check}

\item{doList}{A boolean to see if the functions in .udfEnv$fun
should be checked too.  By default TRUE, but this is called
recursively for each function (and set to FALSE)}
}
\value{
logical declaring if the udf function exists in this environment
}
\description{
If other functions have been declared, make sure they exist too.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxModels_}
\alias{rxModels_}
\title{Get the rxModels  information}
\usage{
rxModels_(env = TRUE)
}
\arguments{
\item{env}{boolean that returns the environment where models are stored (TRUE), or the currently assigned rxode2 model variables (FALSE).}
}
\value{
internal rxModels information environment
}
\description{
Get the rxModels  information
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxStack.R
\name{rxStack}
\alias{rxStack}
\title{Stack a solved object for things like default ggplot2 plot}
\usage{
rxStack(data, vars = NULL, doSim = TRUE, doIpredSim = TRUE)
}
\arguments{
\item{data}{is a rxode2 object to be stacked.}

\item{vars}{Variables to include in stacked data; By default this
is all the variables when vars is NULL.

When vars is \code{sim} and comes from a \code{rxode2} ui simulation with
multiple endpoints (ie it has a \code{CMT} in the simulation), it will
rework the data as if it was stacked based the value based on the
compartments in the multiple endpoint model.

When the vars is \code{sim.endpoint1} it will subset the stack to
endpoint1, you can also have `c("sim.endpoint1", "sim.endpoint2")
and the "stack" will subset to endpoint1 and endpoint2.

When you specify the \code{sim} type variables they have to be all
prefixed with \code{sim} otherwise, the stack will not treat them
differently.}

\item{doSim}{boolean that determines if the "sim" variable in a
\code{rxSolve} dataset is actually "stacking" based on the endpoint
(\code{TRUE}) or simply treating \code{sim} as a variable.}

\item{doIpredSim}{boolean that determines if the "ipredSim" variable in a
\code{rxSolve} dataset is actually "stacking" based on the endpoint
(\code{TRUE}) or simply treating \code{ipredSim} as a variable.}
}
\value{
Stacked data with \code{value} and \code{trt}, where value is the values
and \code{trt} is the state and \code{lhs} variables.
}
\description{
Stack a solved object for things like default ggplot2 plot
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err-foceiBase.R
\name{.rxGetPredictionFTransform}
\alias{.rxGetPredictionFTransform}
\title{Get the prediction transformation}
\usage{
.rxGetPredictionFTransform(env, pred1, yj)
}
\arguments{
\item{env}{Environment for the parsed model}

\item{pred1}{The \code{data.frame} of the current error}

\item{yj}{The transformation number for the current error}
}
\value{
The transformation expression
}
\description{
Get the prediction transformation
}
\author{
Matthew Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/asIni.R
\name{as.ini}
\alias{as.ini}
\alias{as.ini.character}
\alias{as.ini.data.frame}
\alias{as.ini.call}
\alias{as.ini.lotriFix}
\alias{as.ini.matrix}
\alias{as.ini.default}
\title{Turn into an ini block for initialization}
\usage{
as.ini(x)

\method{as.ini}{character}(x)

\method{as.ini}{data.frame}(x)

\method{as.ini}{call}(x)

\method{as.ini}{lotriFix}(x)

\method{as.ini}{matrix}(x)

\method{as.ini}{default}(x)
}
\arguments{
\item{x}{Item to convert to a rxode2/nlmixr2 ui ini expression}
}
\value{
rxode2 ini expression
}
\description{
Turn into an ini block for initialization
}
\examples{

ini <- quote(ini({
  tka <- log(1.57)
  tcl <- log(2.72)
  tv <- log(31.5)
  eta.ka ~ 0.6
  eta.cl ~ 0.3
  eta.v ~ 0.1
  add.sd <- 0.7
}))

as.ini(ini)

l <- quote(lotri({
  tka <- log(1.57)
  tcl <- log(2.72)
  tv <- log(31.5)
  eta.ka ~ 0.6
  eta.cl ~ 0.3
  eta.v ~ 0.1
  add.sd <- 0.7
 }))

as.ini(l)

m <- lotri({
   eta.ka ~ 0.6
   eta.cl ~ 0.3
   eta.v ~ 0.1
})

as.ini(m)

one.compartment <- function() {
  ini({
    tka <- log(1.57)
    tcl <- log(2.72)
    tv <- log(31.5)
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v <- exp(tv + eta.v)
    d/dt(depot) = -ka * depot
    d/dt(center) = ka * depot - cl / v * center
    cp = center / v
    cp ~ add(add.sd)
  })
}

as.ini(one.compartment)

ui <- one.compartment()

as.ini(ui)

ui$iniDf

as.ini(ui$iniDf)

ini <- c("ini({",
          "tka <- log(1.57)",
          "tcl <- log(2.72)",
          "tv <- log(31.5)",
          "eta.ka ~ 0.6",
          "eta.cl ~ 0.3",
          "eta.v ~ 0.1",
          "add.sd <- 0.7",
          "})")

as.ini(ini)

ini <- paste(ini, collapse="\n")

as.ini(ini)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxState}
\alias{rxState}
\title{State variables}
\usage{
rxState(obj = NULL, state = NULL)
}
\arguments{
\item{obj}{rxode2 family of objects}

\item{state}{is a string indicating the state or compartment that
you would like to lookup.}
}
\value{
If state is missing, return a character vector of all the states.

If state is a string, return the compartment number of the named state.
}
\description{
This returns the model's compartments or states.
}
\seealso{
\code{\link[=rxode2]{rxode2()}}

Other Query model information:
\code{\link[=rxDfdy]{rxDfdy()}},
\code{\link[=rxInits]{rxInits()}},
\code{\link[=rxLhs]{rxLhs()}},
\code{\link[=rxModelVars]{rxModelVars()}},
\code{\link[=rxParams]{rxParams()}}
}
\author{
Matthew L.Fidler
}
\concept{Query model information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode-options.R
\name{rxSyncOptions}
\alias{rxSyncOptions}
\title{Sync options with rxode2 variables}
\usage{
rxSyncOptions(setDefaults = c("none", "permissive", "strict"))
}
\arguments{
\item{setDefaults}{This will setup rxode2's default solving options with the following options:
\itemize{
\item \code{"none"} leave the options alone
\item \code{"permissive"} This is a permissive option set similar to R language specifications.
\item \code{"strict"} This is a strict option set similar to the original
rxode2(). It requires semicolons at the end of lines and equals for
assignment
}}
}
\value{
nothing; called for side effects
}
\description{
Sync options with rxode2 variables
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxsolve.R
\name{rxControlUpdateSens}
\alias{rxControlUpdateSens}
\title{This updates the tolerances based on the sensitivity equations}
\usage{
rxControlUpdateSens(rxControl, sensCmt = NULL, ncmt = NULL)
}
\arguments{
\item{rxControl}{Input list or rxControl type of list}

\item{sensCmt}{Number of sensitivity compartments}

\item{ncmt}{Number of compartments}
}
\value{
Updated rxControl where \verb{$atol}, \verb{$rtol}, \verb{$ssAtol}
\verb{$ssRtol} are updated with different sensitivities for the normal
ODEs (first) and a different sensitivity for the larger
compartments (sensitivities).
}
\description{
This assumes the normal ODE equations are the first equations and
the ODE is expanded by the forward sensitivities or other type of
sensitivity (like adjoint)
}
\examples{

tmp <- rxControl()

tmp2 <- rxControlUpdateSens(tmp, 3, 6)

tmp2$atol
tmp2$rtol
tmp2$ssAtol
tmp2$ssRtol
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{.useUtf}
\alias{.useUtf}
\title{Internal function to figure out if this session supports Unicode}
\usage{
.useUtf()
}
\value{
boolean indicating if this session supports Unicode
}
\description{
Internal function to figure out if this session supports Unicode
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fixef.R
\name{fixef.rxUi}
\alias{fixef.rxUi}
\alias{fixef.function}
\title{Extract fixed effects from an rxode2 model (rxUi)}
\usage{
fixef.rxUi(object, ...)

fixef.function(object, ...)
}
\arguments{
\item{object}{rxode2 ui model or a function that evaluates to one}

\item{...}{other parameters (currently ignored)}
}
\value{
named numeric vector of the fixed-effect (theta) estimates
}
\description{
Extract fixed effects from an rxode2 model (rxUi)
}
\examples{
one.compartment <- function() {
  ini({
    tka <- 0.45
    tcl <- 1
    tv <- 3.45
    eta.ka ~ 0.6
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl)
    v <- exp(tv)
    d/dt(depot) <- -ka * depot
    d/dt(center) <- ka * depot - cl / v * center
    cp <- center / v
  })
}

ui <- one.compartment()
nlme::fixef(ui)
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxsolve.R
\name{odeMethodToInt}
\alias{odeMethodToInt}
\title{Conversion between character and integer ODE integration methods for rxode2}
\usage{
odeMethodToInt(
  method = c("liblsoda", "lsoda", "dop853", "indLin", "f78", "rk4", "ck54", "ab", "abm",
    "dop5", "bs", "ros4", "iem", "sem", "sb3a", "sb3am4", "vv", "mm", "em", "cvode",
    "trapz", "ssp3", "f32", "rk43", "dop54", "vern65", "vern76", "dop87", "vern98",
    "ros43", "ros6", "backwardEuler", "gauss6", "iiic6", "radauiia5", "geng5", "sdirk43",
    "euler", "midpoint", "heun", "ssp22", "rk3", "ssp53", "s4", "r4", "ls44", "ls54",
    "ssp54", "s5", "rk5", "c5", "l5", "lk5a", "lk5b", "b6", "s7", "s8_10", "cv8",
    "s8_12", "s10", 
     "z10", "o10", "h10", "dp54", "v65e", "v76e", "dp87", "v98e",
    "ssp33", "bs32", "ssp43", "f45", "t54", "s54", "pp54", "pp54b", "bs54", "ss54",
    "dp65", "c65", "tp64", "v65r", "v65", "dverk65", "tf65", "tp75", "tmy7", "tmy7s",
    "v76r", "ss76", "v78", "dverk78", "dp85", "tp86", "v87e", "v87r", "ev87", "k87",
    "f89", "v89", "t98a", "v98r", "s98", "f108", "c108", "b109", "s1110a", "f1210",
    "o129", "f1412", "lsode", "bdf", "rk4s", "eulers", "midpoints", "heuns", "dop5s",
    "dop853s", "ck54s", "bs32s", "vern65s", 
     "vern76s", "dop87s", "f78s", "ros4s",
    "radauiia5s", "backwardEulers", "gauss6s", "sdirk43s", "iiic6s", "ros43s", "ros6s",
    "geng5s", "rk3s", "rk43s", "cvodesadj", "liblsodaadj", "abs", "dop54s", "dp54s",
    "vern98s", "f45s", "t54s", "pp54s", "pp54bs", "bs54s", "ss54s", "dp65s", "c65s",
    "tp64s", "v65rs", "dverk65s", "tf65s", "tp75s", "tmy7sadj", "tmy7adj", "v76rs",
    "ss76s", "v78s", "dverk78s", "dp85s", "tp86s", "v87es", "v87rs", "ev87s", "k87s",
    "v89s", "t98as", "v98rs", "s98s", "c108s", "b109s", 
     "s1110as", "o129s")
)
}
\arguments{
\item{method}{The method for solving ODEs.  Currently this supports:
\itemize{
\item \code{"liblsoda"} thread safe lsoda.  This supports parallel
thread-based solving, and ignores user Jacobian specification.
\item \code{"lsoda"} -- LSODA solver.  Does not support parallel thread-based
solving, but allows user Jacobian specification.
\item \code{"dop853"} -- DOP853 solver.  Does not support parallel thread-based
solving nor user Jacobian specification
\item \code{"indLin"} -- Solving through inductive linearization.  The rxode2 dll
must be setup specially to use this solving routine.
\item \code{"f78"} -- Runge-Kutta Fehlberg 78 solver using Boost's odeint library.
\item \code{"rk4"} -- Runge-Kutta 4 solver using Boost's odeint library.
\item \code{"ck54"} -- Cash-Karp 5(4) solver using Boost's odeint library.
\item \code{"ab"} -- Adams-Bashforth multi-step solver with a direct implementation
(order 1-8, default 5). Uses RK4 to initialize the derivative history and
then applies the Adams-Bashforth extrapolation formula. Is a fixed-step
method (step size controlled by \code{hmin}).
\item \code{"abm"} -- Adams-Bashforth-Moulton solver using Boost's odeint library.
\item \code{"dop5"} -- DOPRI5 solver using Boost's odeint library (supports dense output).
\item \code{"bs"} -- Bulirsch-Stoer solver using Boost's odeint library (supports dense output).
\item \code{"ros4"} \strong{(implicit)} -- Rosenbrock 4 solver using Boost's odeint library (supports dense output).
Requires an analytical Jacobian (auto-computed from the model when not provided);
falls back to \code{liblsoda} if Jacobian generation fails.
\item \code{"ros43"} \strong{(implicit)} -- Kaps-Rentrop 4th-order A-stable Rosenbrock method (GRK4A) from libode.
Adaptive step size via embedded 3rd-order error estimate.
Requires an analytical Jacobian (auto-computed from the model when not provided);
falls back to \code{liblsoda} if Jacobian generation fails.
\item \code{"ros6"} \strong{(implicit)} -- Kaps-Wanner 6th-order A-stable Rosenbrock method (ROW6A) from libode.
Fixed step size (set with \code{hmin}). Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"backwardEuler"} \strong{(implicit)} -- Backward Euler (BDF-1), 1st-order L-stable fully implicit method from libode.
Fixed step size (set with \code{hmin}). Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"gauss6"} \strong{(implicit)} -- Gauss-Legendre 6th-order A-stable fully-implicit method (3 stages) from libode.
Fixed step size. Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"iiic6"} \strong{(implicit)} -- Lobatto IIIC 6th-order L-stable fully-implicit method (4 stages) from libode.
Fixed step size. Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"radauiia5"} \strong{(implicit)} -- Radau IIA 5th-order L-stable fully-implicit method (3 stages) from libode.
Fixed step size (set with \code{hmin}). Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"geng5"} \strong{(implicit)} -- Geng 5th-order symplectic fully-implicit method (3 stages) from libode.
Fixed step size. Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"sdirk43"} \strong{(implicit)} -- L-stable SDIRK 4(3) pair from libode.
Adaptive step size via embedded 3rd-order error estimate.
Requires Jacobian; falls back to \code{liblsoda} if unavailable.
\item \code{"iem"} \strong{(implicit)} -- Implicit Euler solver using Boost's odeint library.
Requires an explicit Jacobian (auto-generated via \code{calcJac=TRUE} when not provided).
Uses Boost.uBLAS vectors and is a fixed-step method (step size controlled by \code{hmin}).
\item \code{"sem"} -- Symplectic Euler solver using Boost's odeint library.
Requires splitting the state into coordinate-momentum pairs (and
automatically pads odd-dimensional systems).  Is a fixed-step
method (step size controlled by \code{hmin}).
\item \code{"sb3a"} -- Symplectic Runge-Kutta-Nystrom SB3A McLachlan stepper
using Boost's odeint library.  Requires splitting the state into
coordinate-momentum pairs (and automatically pads odd-dimensional
systems).  Is a fixed-step method (step size controlled by
\code{hmin}).
\item \code{"sb3am4"} -- Symplectic Runge-Kutta-Nystrom SB3A M4 McLachlan
stepper using Boost's odeint library.  Requires splitting the
state into coordinate-momentum pairs (and automatically pads
odd-dimensional systems).  Is a fixed-step method (step size
controlled by \code{hmin}).
\item \code{"vv"} -- Velocity Verlet stepper using Boost's odeint library.
Requires splitting the state into coordinate-momentum pairs (and
automatically pads odd-dimensional systems).  Is a fixed-step
method (step size controlled by \code{hmin}).
\item \code{"mm"} -- Modified Midpoint stepper using Boost's odeint library.
Is a fixed-step method (step size controlled by \code{hmin}) and uses
the \code{order} parameter to configure the number of intermediate
steps.
\item \code{"em"} -- Explicit Euler stepper using Boost's odeint library.
Is a fixed-step method (step size controlled by \code{hmin}).
\item \code{"cvode"} -- CVODE BDF stiff solver from the SUNDIALS library (sources
and headers vendored into rxode2).  Supports thread-parallel solving and
per-compartment absolute tolerances.
\item \code{"trapz"} -- Explicit trapezoidal rule (Heun's method), a 2nd-order
fixed-step Runge-Kutta method implemented via the libode library.
Each inter-event interval is integrated with fixed steps of size
\code{hmin} (default \code{0.01} when \code{hmin=0}).  If \code{hmin} would require
more than \code{maxsteps} steps to span the interval, the step size is
automatically enlarged to stay within that limit.  When an
inter-event interval is shorter than the nominal step size (e.g.,
two doses very close together), the step is silently clamped to the
interval length so that short intervals are always handled correctly.
Supports parallel thread-based solving and steady-state (\code{ss=1})
dosing.  The method does not use \code{atol} or \code{rtol} (fixed-step; no
error control).  Steady-state convergence is governed by \code{ssAtol},
\code{ssRtol}, \code{minSS}, \code{maxSS}, and \code{strictSS}.
\item \code{"ssp3"} -- Strong Stability-Preserving Runge-Kutta of order 3 (SSP-RK3,
Shu-Osher method), implemented via the libode library.  A 3-stage,
3rd-order explicit fixed-step method with superior non-oscillatory
properties for problems with discontinuities or sharp fronts.  Uses
the Butcher tableau c2=1, a21=1; c3=1/2, a31=1/4, a32=1/4;
b1=1/6, b2=1/6, b3=2/3.  The step-size behavior, clamping, and
steady-state options (\code{hmin}, \code{maxsteps}, \code{ssAtol}, \code{ssRtol},
\code{minSS}, \code{maxSS}, \code{strictSS}) are identical to \code{"trapz"}.  Does
not use \code{atol} or \code{rtol} (fixed-step; no error control).  Supports
parallel thread-based solving and steady-state (\code{ss=1}) dosing.
\item \code{"f32"} -- Fehlberg 3(2) embedded pair from libode (class \code{OdeRKF32}).
A 3-stage, 3rd-order adaptive method with a built-in 2nd-order error
estimate for automatic step-size control.  Tableau: c2=1, a21=1;
c3=1/2, a31=1/4, a32=1/4; primary (3rd-order) weights d1=1/6, d2=1/6,
d3=4/6; embedded (2nd-order) weights b1=1/2, b2=1/2.  Uses \code{atol} and \code{rtol}
for error control.  The \code{hmin}
parameter sets the initial step size (default \code{0.01}); subsequent steps
are chosen adaptively.  The total number of accepted and rejected steps
is bounded by \code{maxsteps}.  Supports parallel thread-based solving and
steady-state (\code{ss=1}) dosing with convergence governed by \code{ssAtol},
\code{ssRtol}, \code{minSS}, \code{maxSS}, and \code{strictSS}.
\item \code{"rk43"} -- Runge-Kutta 4(3) pair from libode.  A 5-stage, 4th-order
adaptive method with a built-in 3rd-order embedded error estimate for
automatic step-size control.  Tableau: c2=1/3, a21=1/3; c3=2/3,
a31=-1/3, a32=1; c4=1, a41=1, a42=-1, a43=1; a5j=bj (FSAL).
Primary (4th-order) weights b1=b4=1/8, b2=b3=3/8; embedded (3rd-order)
weights d1=1/12, d2=1/2, d3=1/4, d5=1/6.  Since a5j=bj the 5th stage
evaluation is reused as the 1st stage of the next step (FSAL).
Uses \code{atol} and \code{rtol} for error control.  The \code{hmin} parameter sets
the initial step size (default \code{0.01}); subsequent steps are chosen
adaptively.  The
total number of steps is bounded by \code{maxsteps}.  Supports parallel
thread-based solving and steady-state (\code{ss=1}) dosing with convergence
governed by \code{ssAtol}, \code{ssRtol}, \code{minSS}, \code{maxSS}, and \code{strictSS}.
\item \code{"dop54"} -- Dormand-Prince 5(4) FSAL method (Dormand & Prince 1980),
implemented via the libode library.  The same algorithm as MATLAB's
\code{ode45}.  A 7-stage, 5th-order adaptive method with an embedded
4th-order error estimate for automatic step-size control (FSAL: the
7th stage evaluation is reused as the 1st stage of the next step,
giving effectively 6 function evaluations per accepted step).  Uses
\code{atol} and \code{rtol} for error control.  The \code{hmin} parameter sets the
initial step size (default \code{0.01}); subsequent steps are chosen
adaptively.  The total number of steps is bounded by \code{maxsteps}.
Supports parallel thread-based solving and steady-state (\code{ss=1})
dosing with convergence governed by \code{ssAtol}, \code{ssRtol}, \code{minSS},
\code{maxSS}, and \code{strictSS}.  NaN/Inf in derivatives (e.g. from NA
parameters) is detected immediately and the solve exits with NA
output for the affected subject.
\item \code{"vern65"} -- Jim Verner's "most efficient" 6(5) FSAL pair (9 stages),
implemented via the libode library using coefficients from Verner's own
website (RKV65.IIIXb.Efficient).  A 6th-order adaptive method with an
embedded 5th-order error estimate.  The sparse tableau (many zero
a-coefficients) reduces function evaluations; the FSAL property means the
9th stage evaluation is reused as the 1st stage of the next step, giving
effectively 8 function evaluations per accepted step.  Uses \code{atol} and
\code{rtol} for error control.  The \code{hmin} parameter sets the initial step
size (default \code{0.01}); subsequent steps are chosen adaptively.  The total
number of steps is bounded by \code{maxsteps}.  Supports parallel thread-based
solving and steady-state (\code{ss=1}) dosing with convergence governed by
\code{ssAtol}, \code{ssRtol}, \code{minSS}, \code{maxSS}, and \code{strictSS}.  NaN/Inf in
derivatives is detected immediately and the solve exits with NA output.
\item \code{"vern76"} -- Jim Verner's "most efficient" 7(6) pair (10 stages),
implemented via the libode library using coefficients from Verner's own
website (RKV76.IIa.Efficient).  A 7th-order adaptive method with an
embedded 6th-order error estimate.  The sparse tableau reduces function
evaluations per step.  Uses \code{atol} and \code{rtol} for error control.  The
\code{hmin} parameter sets the initial step size (default \code{0.01}); subsequent
steps are chosen adaptively.  The total number of steps is bounded by
\code{maxsteps}.  Supports parallel thread-based solving and steady-state
(\code{ss=1}) dosing with convergence governed by \code{ssAtol}, \code{ssRtol},
\code{minSS}, \code{maxSS}, and \code{strictSS}.  NaN/Inf in derivatives is detected
immediately and the solve exits with NA output.
\item \code{"dop87"} -- Dormand-Prince 8(7) pair (13 stages), implemented via the
libode library using coefficients from Hairer, Norsett and Wanner (1993)
"Solving ODEs I" (2nd ed.).  An 8th-order adaptive method with an
embedded 7th-order error estimate.  Both solutions are computed from the
original state in the final step loop.  Uses \code{atol} and \code{rtol} for error
control.  The \code{hmin} parameter sets the initial step size (default
\code{0.01}); subsequent steps are chosen adaptively.  The total number of
steps is bounded by \code{maxsteps}.  Supports parallel thread-based solving
and steady-state (\code{ss=1}) dosing with convergence governed by \code{ssAtol},
\code{ssRtol}, \code{minSS}, \code{maxSS}, and \code{strictSS}.  NaN/Inf in derivatives is
detected immediately and the solve exits with NA output.
\item \code{"vern98"} -- Jim Verner's "most efficient" 9(8) pair (16 stages),
implemented via the libode library using coefficients from Verner's own
website (RKV98.IIa.Efficient).  A 9th-order adaptive method with an
embedded 8th-order error estimate.  The highly sparse tableau (stages
8-16 use only \eqn{k_{0}}{x_0} and \eqn{k_{5..}}{k_5..}) minimizes function evaluations per
step.  Both solutions are computed from the original state in the final
step loop.  Uses \code{atol} and \code{rtol} for error control.  The \code{hmin}
parameter sets the initial step size (default \code{0.01}); subsequent steps
are chosen adaptively.  The total number of steps is bounded by
\code{maxsteps}.  Supports parallel thread-based solving and steady-state
(\code{ss=1}) dosing with convergence governed by \code{ssAtol}, \code{ssRtol},
\code{minSS}, \code{maxSS}, and \code{strictSS}.  NaN/Inf in derivatives is detected
immediately and the solve exits with NA output.
}

\strong{Aliases for existing methods} (no additional C++ code; \code{hmin}, \code{atol},
\code{rtol}, and \code{maxsteps} follow the aliased method):
\itemize{
\item \code{"dp54"} -- alias for \code{"dop54"} (Dormand-Prince 5(4) FSAL, libode).
\item \code{"v65e"} -- alias for \code{"vern65"} (Verner 6(5) efficient, libode).
\item \code{"v76e"} -- alias for \code{"vern76"} (Verner 7(6) efficient, libode).
\item \code{"dp87"} -- alias for \code{"dop87"} (Dormand-Prince 8(7), libode).
\item \code{"v98e"} -- alias for \code{"vern98"} (Verner 9(8) efficient, libode).
\item \code{"ssp33"} -- alias for \code{"ssp3"} (Strong Stability-Preserving RK3, libode).
}

\strong{libode fixed-step explicit methods}.  All use \code{hmin} as the fixed step
size (default \code{0.01} when \code{hmin=0}); \code{atol} and \code{rtol} are ignored (no
error control); \code{maxsteps} bounds the total number of steps; NaN/Inf in
derivatives sets \code{ind$err} and the subject exits with NA output.  All
support parallel thread-based solving.
\itemize{
\item \code{"euler"} -- Forward (explicit) Euler, 1st-order, 1 stage.  Requires a
very small step size for accuracy (e.g., \code{hmin=1e-4}); intended for
pedagogical use or method-comparison baselines.
\item \code{"midpoint"} -- Explicit midpoint rule, 2nd-order, 2 stages.
\item \code{"heun"} -- Heun's method (explicit trapezoid), 2nd-order, 2 stages.
Identical in structure to \code{"trapz"} but uses the libode fixed-step driver.
\item \code{"ssp22"} -- Strong Stability-Preserving 2-stage 2nd-order method
(SSP-RK22), 2 stages.  Superior non-oscillatory properties near
discontinuities.
\item \code{"rk3"} -- Classical 3rd-order Runge-Kutta (Kutta 1901), 3 stages.
\item \code{"ssp53"} -- Strong Stability-Preserving 5-stage 3rd-order method
(SSP-RK53), 5 stages.  High SSP coefficient for hyperbolic PDEs or
event-heavy ODE systems.
\item \code{"s4"} -- Shanks 4th-order method, 4 stages.
\item \code{"r4"} -- Ralston's 4th-order method, 4 stages.  Minimizes local
truncation error among classical 4-stage 4th-order methods.
\item \code{"ls44"} -- Low-storage 4th-order method, 4 stages.  Uses a 2-register
update scheme that minimizes memory bandwidth at the cost of a less
general tableau structure.
\item \code{"ls54"} -- Low-storage 4th-order method, 5 stages.  Five-stage
variant of the 2-register low-storage scheme.
\item \code{"ssp54"} -- Strong Stability-Preserving 5-stage 4th-order method
(SSP-RK54), 5 stages.
\item \code{"s5"} -- Shanks 5th-order method, 5 stages.
\item \code{"rk5"} -- Classical 5th-order Runge-Kutta, 6 stages.
\item \code{"c5"} -- Cassity 5th-order method, 6 stages.
\item \code{"l5"} -- Lawson 5th-order method, 6 stages.
\item \code{"lk5a"} -- Luther-Konen 5th-order method, variant A, 6 stages.
\item \code{"lk5b"} -- Luther-Konen 5th-order method, variant B, 6 stages.
\item \code{"b6"} -- Butcher 6th-order method, 7 stages.
\item \code{"s7"} -- Shanks 7th-order method, 9 stages.
\item \code{"s8_10"} -- Shanks 8th-order method, 10 stages.
\item \code{"cv8"} -- Cooper-Verner 8th-order method, 11 stages.
\item \code{"s8_12"} -- Shanks 8th-order method, 12 stages.
\item \code{"s10"} -- Stepanov 10th-order method, 15 stages.  Requires a moderate
step size (e.g., \code{hmin=1.0}) because a single step is accurate to very
high order.
\item \code{"z10"} -- Zhang 10th-order method, 16 stages.
\item \code{"o10"} -- Ono 10th-order method, 17 stages.
\item \code{"h10"} -- Hairer 10th-order method, 17 stages.
}

\strong{libode adaptive (variable-step) explicit methods}.  All use \code{atol} and
\code{rtol} for error control;
\code{hmin} sets the initial step size (default \code{0.01} when \code{hmin=0}); \code{hmax}
sets the maximum step size; \code{maxsteps} bounds total steps; NaN/Inf in
derivatives sets \code{ind$err} and the subject exits with NA output.  All
support parallel thread-based solving and steady-state (\code{ss=1}) dosing
(convergence governed by \code{ssAtol}, \code{ssRtol}, \code{minSS}, \code{maxSS}, \code{strictSS}).
\itemize{
\item \code{"bs32"} -- Bogacki-Shampine 3(2) FSAL pair, 4 stages (Bogacki &
Shampine 1989).  3rd-order primary with 2nd-order embedded error estimate.
FSAL: the 4th-stage evaluation is reused as the 1st stage of the next
step.  The same algorithm as Julia \code{BS3()} and MATLAB \code{ode23}.
\item \code{"ssp43"} -- Strong Stability-Preserving 4(3) pair, 4 stages.
Adaptive SSP method with a 3rd-order embedded error estimate.
\item \code{"f45"} -- Fehlberg 4(5) pair, 6 stages (Fehlberg 1970).  4th-order
primary solution with a 5th-order embedded estimate used for error
control.
\item \code{"t54"} -- Tsitouras 5(4) FSAL pair, 7 stages (Tsitouras 2011).
5th-order primary with 4th-order embedded error estimate.  FSAL: the
7th stage is reused as the 1st stage of the next step.  The same
Butcher tableau as Julia's \code{Tsit5()}.
\item \code{"s54"} -- Stepanov 5(4) FSAL pair, 7 stages.  5th-order primary
with 4th-order embedded error estimate.
\item \code{"pp54"} -- Papakostas-Papageorgiou 5(4) FSAL pair, 7 stages.
\item \code{"pp54b"} -- Papakostas-Papageorgiou 5(4) variant B FSAL pair,
7 stages.
\item \code{"bs54"} -- Bogacki-Shampine 5(4) pair, 8 stages.  5th-order primary
with 4th-order embedded error estimate (non-FSAL).
\item \code{"ss54"} -- Sharp-Smart 5(4) pair, 7 stages.
\item \code{"dp65"} -- Dormand-Prince 6(5) pair, 8 stages.  6th-order primary
with 5th-order embedded error estimate.
\item \code{"c65"} -- Calvo 6(5) pair, 9 stages.
\item \code{"tp64"} -- Tsitouras-Papakostas 6(4) pair, 7 stages.  6th-order
primary with 4th-order embedded error estimate.
\item \code{"v65r"} -- Verner "robust" 6(5) FSAL pair, 9 stages.  Robust
variant of Verner's 6(5) family with wider stability region.
\item \code{"v65"} -- Verner 6(5) pair, 8 stages (non-FSAL).
\item \code{"dverk65"} -- Verner DVERK 6(5) pair, 8 stages.  Coefficients from
the classic DVERK Fortran code distributed by Hull and Enright.
\item \code{"tf65"} -- Tsitouras-Famelis 6(5) FSAL pair, 9 stages.
\item \code{"tp75"} -- Tsitouras-Papakostas 7(5) pair, 9 stages.  7th-order
primary with 5th-order embedded error estimate.
\item \code{"tmy7"} -- Tanaka-Muramatsu-Yamashita 7th-order pair, 10 stages.
The same family as Julia's \code{TanYam7()}.
\item \code{"tmy7s"} -- Tanaka-Muramatsu-Yamashita 7th-order stable variant,
10 stages.  Alternative coefficient set with wider stability region.
\item \code{"v76r"} -- Verner "robust" 7(6) pair, 10 stages.
\item \code{"ss76"} -- Sharp-Smart 7(6) pair, 11 stages.
\item \code{"v78"} -- Verner 7(8) pair, 13 stages.  7th-order primary with
8th-order embedded error estimate.  Closest rxode2 analog to Julia
\code{Vern8()}.
\item \code{"dverk78"} -- Verner DVERK 7(8) pair, 13 stages.  Coefficients from
the classic DVERK Fortran code; companion to \code{"dverk65"}.  Also close
to Julia \code{Vern8()}.
\item \code{"dp85"} -- Dormand-Prince 8(5) pair, 12 stages.  8th-order primary
with 5th-order embedded error estimate.
\item \code{"tp86"} -- Tsitouras-Papakostas 8(6) pair, 12 stages.  The same
family as Julia's \code{TsitPap8()}.
\item \code{"v87e"} -- Verner "efficient" 8(7) pair, 13 stages.
\item \code{"v87r"} -- Verner "robust" 8(7) pair, 13 stages.
\item \code{"ev87"} -- Enright-Verner 8(7) pair, 13 stages.
\item \code{"k87"} -- Kovalnogov-Fedorov-Karpukhina-Simos 8(7) pair, 13 stages.
\item \code{"f89"} -- Fehlberg 8(9) pair, 17 stages.  8th-order primary with
9th-order embedded estimate.  Same family as MATLAB \code{ode89}.
\item \code{"v89"} -- Verner 8(9) pair, 16 stages.  Alternative to \code{"f89"}
in the same order bracket.  Also close to MATLAB \code{ode89}.
\item \code{"t98a"} -- Tsitouras 9(8) variant A pair, 16 stages.
\item \code{"v98r"} -- Verner "robust" 9(8) pair, 16 stages.
\item \code{"s98"} -- Sharp 9(8) pair, 16 stages.
\item \code{"f108"} -- Feagin 10(8) pair, 17 stages (Feagin 2007).  10th-order
primary with 8th-order embedded error estimate.  The same method as
Julia's \code{Feagin10()}.  The large number of stages carries elevated
transcription-error risk; verified against Williams' libode canonical order test.
\item \code{"c108"} -- Curtis 10(8) pair, 21 stages.
\item \code{"b109"} -- Baker 10(9) pair, 21 stages.
\item \code{"s1110a"} -- Stone 11(10) variant A pair, 26 stages.  11th-order
primary; very few published references.
\item \code{"f1210"} -- Feagin 12(10) pair, 25 stages (Feagin 2007).  12th-order
primary with 10th-order embedded error estimate.  The same method as
Julia's \code{Feagin12()}.  Elevated transcription-error risk due to many
stages; verified against Williams' libode canonical order test.
\item \code{"o129"} -- Ono 12(9) pair, 29 stages.
\item \code{"f1412"} -- Feagin 14(12) pair, 35 stages (Feagin 2007).  14th-order
primary with 12th-order embedded error estimate.  The same method as
Julia's \code{Feagin14()}.  The highest-order method in rxode2; elevated
transcription-error risk due to 35 stages; verified against
Williams' libode canonical order test.
}

\strong{deSolve-derived Fortran solvers} (from the deSolve R package).  Both
use non-reentrant Fortran COMMON blocks and therefore run single-threaded
only, regardless of the \code{cores} setting.  They are BDF (Backward
Differentiation Formula) stiff solvers and are most useful when you want
behavior comparable to deSolve's \code{lsode()} or \code{bdf()}/\code{vode()} functions.
\itemize{
\item \code{"lsode"} -- DLSODE (Hindmarsh 1983, ODEPACK) in Adams (non-stiff) mode,
MF=10.  Variable-order Adams method, order 1-12; no Jacobian evaluation.
Corresponds to deSolve's \code{lsode(method="adams")}.  Adaptive step-size;
error controlled by \code{atol} and \code{rtol}.
\strong{Not thread-safe} -- always runs on a single core.
\item \code{"bdf"} -- DLSODE (Hindmarsh 1983, ODEPACK) in BDF (stiff) mode, MF=22.
Variable-order BDF method, order 1-5; internally generated dense Jacobian.
Corresponds to deSolve's \code{bdf()} / \code{vode(method="bdf")}.  Adaptive
step-size; error controlled by \code{atol} and \code{rtol}.
\strong{Not thread-safe} -- always runs on a single core.
}}
}
\value{
An integer for the method (unless the input is NULL, in which case,
see the details)
}
\description{
If \code{NULL} is given as the method, all choices are returned as a named
vector.
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dfWishart.R
\name{swapMatListWithCube}
\alias{swapMatListWithCube}
\title{Swaps the matrix list with a cube}
\usage{
swapMatListWithCube(matrixListOrCube)
}
\arguments{
\item{matrixListOrCube}{Either a list of 2-dimensional matrices or a cube of matrices}
}
\value{
A list or a cube (opposite format as input)
}
\description{
Swaps the matrix list with a cube
}
\examples{

# Create matrix list
matLst <- cvPost(10, lotri::lotri(a+b~c(1, 0.25, 1)), 3)
print(matLst)

# Convert to cube
matCube <- swapMatListWithCube(matLst)
print(matCube)

# Convert back to list
matLst2 <- swapMatListWithCube(matCube)
print(matLst2)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxSyntaxFunctions.R
\docType{data}
\name{rxSyntaxFunctions}
\alias{rxSyntaxFunctions}
\title{A list and description of Rode supported syntax functions}
\format{
A data frame with 3 columns and 106 rows
\describe{
\item{Function}{Reserved function Name}
\item{Description}{Description of function}
\item{Aliases}{Function Aliases}
}
}
\usage{
rxSyntaxFunctions
}
\description{
A list and description of Rode supported syntax functions
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxNorm}
\alias{rxNorm}
\title{Get the normalized model}
\usage{
rxNorm(obj, condition = NULL, removeInis, removeJac, removeSens)
}
\arguments{
\item{obj}{rxode2 family of objects}

\item{condition}{Character string of a logical condition to use
for subsetting the normalized model.  When missing, and a
condition is not set via \code{rxCondition}, return the whole
code with all the conditional settings intact.  When a condition
is set with \code{rxCondition}, use that condition.}

\item{removeInis}{A boolean indicating if parameter initialization
will be removed from the model}

\item{removeJac}{A boolean indicating if the Jacobians will be
removed.}

\item{removeSens}{A boolean indicating if the sensitivities will
be removed.}
}
\value{
Normalized Normal syntax (no comments)
}
\description{
This get the syntax preferred model for processing
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mix.R
\name{mix}
\alias{mix}
\title{Specify a mixture model of variables}
\usage{
mix(...)
}
\arguments{
\item{...}{Arguments to the mixture model.

The first call to the mixture model function takes an odd number of
arguments (at least 3).

For example the model we could have:

cl = mix(cl1, p1, cl2, p2, cl3)

Here there is a mixture of three clearance variables, \code{cl1}, \code{cl2},
and \code{cl3}, at a probability of \code{p1}, \code{p2}, and the last one is
assumed to be \code{1 - p1 - p2}.

For simulations this is selected randomly.  For estimations this is
selected by the data for each individual.

After the first call when the number of populations has been
established, you can also call the mixture model with the number of
populations, for example:

v = mix(v1, v2, v3)

The ui function will translate this to the following model:

v = mix(v1, p1, v2, p2, v3)

This is because the first call to \code{mix()} sets the probabilities.
In rxode2/nlmixr2 these probabilities should be conserved between
the models.  These probabilities also have to be defined in the ini
block directly.}
}
\value{
The mixture model replacement for the underlying rxode2 model.
}
\description{
Specify a mixture model of variables
}
\examples{

# This is an example of a mixture model
# Where there are 2 different clearance populations

one.cmt <- function() {
  ini({
    tka <- 0.45 # Log Ka
    tcl1 <- log(c(0, 2.7, 100)) # Log Cl
    tcl2 <- log(c(0, 0.1, 120)) # Log Cl
    tv <- 3.45; label("log V")
    p1 <- 0.3
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka + eta.ka)
    # This is the example mixture model
    cl <- mix(exp(tcl1 + eta.cl), p1, exp(tcl2 + eta.cl))
    v <- exp(tv + eta.v)
    me <- mixest # This is the assigned mixture estimate
    mn <- mixnum # This is the number of mixture estimate in the model
    # This is the uniform mixture estimate used in simualtion to
    # determine the population
    mu <- mixunif
    linCmt() ~ add(add.sd)
  })
}

\donttest{

s <- rxSolve(one.cmt, et(amt=320, ii=12, addl=2, cmt=1) |>
                      et(seq(0, 72)) |>
                      et(id=1:20))

plot(s, ipredSim)

}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{.expandPars}
\alias{.expandPars}
\title{Expand parameters}
\usage{
.expandPars(object, params, events, control)
}
\arguments{
\item{object}{rxode2 model variables object}

\item{params}{parameters to expand}

\item{events}{event table to help with the expansion}

\item{control}{control structure to help with the parameter generation}
}
\value{
Expanded parameters for simulation
}
\description{
Expand parameters
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxsolve.R
\name{rxIsImplicit}
\alias{rxIsImplicit}
\title{Check whether an ODE solving method requires a Jacobian (is implicit)}
\usage{
rxIsImplicit(method)
}
\arguments{
\item{method}{A character vector of method names or an integerish vector of
method codes (as returned by \code{\link[=odeMethodToInt]{odeMethodToInt()}}).  Vectorised.}
}
\value{
A logical vector the same length as \code{method}.  \code{TRUE} if the
corresponding method is implicit and requires a Jacobian.
}
\description{
Implicit ODE solvers linearise the system at each step and therefore
require the Jacobian df/dy.  rxode2 auto-computes a symbolic Jacobian
when one of these methods is requested and falls back to \code{"liblsoda"} if
symbolic differentiation fails.  Use this function to test whether a
method string or integer code corresponds to an implicit solver.
}
\details{
Implicit methods are:
\code{"ros4"} (13), \code{"iem"} (14), \code{"ros43"} (31), \code{"ros6"} (32),
\code{"backwardEuler"} (33), \code{"gauss6"} (34), \code{"iiic6"} (35), \code{"radauiia5"} (36),
\code{"geng5"} (37), \code{"sdirk43"} (38).
}
\examples{
rxIsImplicit("ros4")              # TRUE
rxIsImplicit("liblsoda")          # FALSE
rxIsImplicit(c("ros43", "dop853", "iem"))  # TRUE FALSE TRUE
rxIsImplicit(13L)                 # TRUE  (ros4)
rxIsImplicit(0L)                  # FALSE (dop853)

}
\seealso{
\code{\link[=odeMethodToInt]{odeMethodToInt()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err.R
\name{.getErrDist}
\alias{.getErrDist}
\alias{.getNamedArgumentsToPredDf}
\title{Get defined residual information for parsing elsewhere}
\usage{
.getErrDist()

.getNamedArgumentsToPredDf()
}
\value{
A list of error distribution information
}
\description{
Get defined residual information for parsing elsewhere
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.rxLinNcmt}
\alias{.rxLinNcmt}
\title{Get the number of linear compartments}
\usage{
.rxLinNcmt(obj)
}
\arguments{
\item{obj}{rxode2 object}
}
\value{
A named integer vector with the number of linear compartments (numLin, numLinSens and depotLin)
}
\description{
Get the number of linear compartments
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/asModel.R
\name{as.model}
\alias{as.model}
\alias{as.model.character}
\alias{as.model.call}
\alias{as.model.list}
\alias{as.model.default}
\title{Turn into a model expression}
\usage{
as.model(x)

\method{as.model}{character}(x)

\method{as.model}{call}(x)

\method{as.model}{list}(x)

\method{as.model}{default}(x)
}
\arguments{
\item{x}{item to convert to a \code{model({})} expression}
}
\value{
model expression
}
\description{
Turn into a model expression
}
\examples{

model <- quote(model({
  ka <- exp(tka + eta.ka)
  cl <- exp(tcl + eta.cl)
  v <- exp(tv + eta.v)
  d/dt(depot) = -ka * depot
  d/dt(center) = ka * depot - cl / v * center
  cp = center / v
  cp ~ add(add.sd)
}))

as.model(model)

one.compartment <- function() {
  ini({
    tka <- log(1.57)
    tcl <- log(2.72)
    tv <- log(31.5)
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v <- exp(tv + eta.v)
    d/dt(depot) = -ka * depot
    d/dt(center) = ka * depot - cl / v * center
    cp = center / v
    cp ~ add(add.sd)
  })
}

as.model(one.compartment)

ui <- one.compartment()

as.model(ui)

model <- c("model({",
           "ka <- exp(tka + eta.ka)",
           "cl <- exp(tcl + eta.cl)",
           "v <- exp(tv + eta.v)",
           "d/dt(depot) = -ka * depot",
           "d/dt(center) = ka * depot - cl / v * center",
           "cp = center / v",
           "cp ~ add(add.sd)",
           "})")

as.model(model)

model <- paste(model, collapse="\n")

as.model(model)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err-sim.R, R/rxUiGet.R
\name{rxUiGet.cmtLines}
\alias{rxUiGet.cmtLines}
\alias{rxUiGet.dvidLine}
\alias{rxUiGet.paramsLine}
\alias{rxUiGet.interpLines}
\alias{rxUiGet.splitDose}
\alias{rxUiGet.splitDoseLines}
\alias{rxUiGet.simulationSigma}
\alias{rxUiGet.simulationModel}
\alias{rxUiGet.symengineModelNoPrune}
\alias{rxUiGet.symengineModelPrune}
\alias{rxUiGet.simulationIniModel}
\alias{rxUiGet}
\alias{rxUiGet.levels}
\alias{rxUiGet.state}
\alias{rxUiGet.stateDf}
\alias{rxUiGet.statePropDf}
\alias{rxUiGet.props}
\alias{rxUiGet.theta}
\alias{rxUiGet.lstChr}
\alias{rxUiGet.omega}
\alias{rxUiGet.funTxt}
\alias{rxUiGet.allCovs}
\alias{rxUiGet.muRefTable}
\alias{rxUiGet.multipleEndpoint}
\alias{rxUiGet.funPrint}
\alias{rxUiGet.fun}
\alias{rxUiGet.funPartsDigest}
\alias{rxUiGet.md5}
\alias{rxUiGet.sha1}
\alias{rxUiGet.ini}
\alias{rxUiGet.iniFun}
\alias{rxUiGet.modelFun}
\alias{rxUiGet.model}
\alias{rxUiGet.modelDesc}
\alias{rxUiGet.thetaLower}
\alias{rxUiGet.thetaUpper}
\alias{rxUiGet.lhsVar}
\alias{rxUiGet.varLhs}
\alias{rxUiGet.lhsEta}
\alias{rxUiGet.lhsTheta}
\alias{rxUiGet.lhsCov}
\alias{rxUiGet.etaLhs}
\alias{rxUiGet.thetaLhs}
\alias{rxUiGet.covLhs}
\alias{rxUiGet.levelLhs}
\alias{rxUiGet.default}
\title{S3 for getting information from UI model}
\usage{
\method{rxUiGet}{cmtLines}(x, ...)

\method{rxUiGet}{dvidLine}(x, ...)

\method{rxUiGet}{paramsLine}(x, ...)

\method{rxUiGet}{interpLines}(x, ...)

\method{rxUiGet}{splitDose}(x, ...)

\method{rxUiGet}{splitDoseLines}(x, ...)

\method{rxUiGet}{simulationSigma}(x, ...)

\method{rxUiGet}{simulationModel}(x, ...)

\method{rxUiGet}{symengineModelNoPrune}(x, ...)

\method{rxUiGet}{symengineModelPrune}(x, ...)

\method{rxUiGet}{simulationIniModel}(x, ...)

rxUiGet(x, ...)

\method{rxUiGet}{levels}(x, ...)

\method{rxUiGet}{state}(x, ...)

\method{rxUiGet}{stateDf}(x, ...)

\method{rxUiGet}{statePropDf}(x, ...)

\method{rxUiGet}{props}(x, ...)

\method{rxUiGet}{theta}(x, ...)

\method{rxUiGet}{lstChr}(x, ...)

\method{rxUiGet}{omega}(x, ...)

\method{rxUiGet}{funTxt}(x, ...)

\method{rxUiGet}{allCovs}(x, ...)

\method{rxUiGet}{muRefTable}(x, ...)

\method{rxUiGet}{multipleEndpoint}(x, ...)

\method{rxUiGet}{funPrint}(x, ...)

\method{rxUiGet}{fun}(x, ...)

\method{rxUiGet}{funPartsDigest}(x, ...)

\method{rxUiGet}{md5}(x, ...)

\method{rxUiGet}{sha1}(x, ...)

\method{rxUiGet}{ini}(x, ...)

\method{rxUiGet}{iniFun}(x, ...)

\method{rxUiGet}{modelFun}(x, ...)

\method{rxUiGet}{model}(x, ...)

\method{rxUiGet}{modelDesc}(x, ...)

\method{rxUiGet}{thetaLower}(x, ...)

\method{rxUiGet}{thetaUpper}(x, ...)

\method{rxUiGet}{lhsVar}(x, ...)

\method{rxUiGet}{varLhs}(x, ...)

\method{rxUiGet}{lhsEta}(x, ...)

\method{rxUiGet}{lhsTheta}(x, ...)

\method{rxUiGet}{lhsCov}(x, ...)

\method{rxUiGet}{etaLhs}(x, ...)

\method{rxUiGet}{thetaLhs}(x, ...)

\method{rxUiGet}{covLhs}(x, ...)

\method{rxUiGet}{levelLhs}(x, ...)

\method{rxUiGet}{default}(x, ...)
}
\arguments{
\item{x}{list of (UIenvironment, exact).  UI environment is the
parsed function for rxode2.  \code{exact} is a boolean that says if an
exact match is required.}

\item{...}{Other arguments}
}
\value{
value that was requested from the UI object
}
\description{
S3 for getting information from UI model
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxUdfUiIsValue}
\alias{rxUdfUiIsValue}
\title{Is the expression actually equal to a value?}
\usage{
rxUdfUiIsValue(expr, value, env = baseenv())
}
\arguments{
\item{expr}{the R language expression to evaluate}

\item{value}{the value to compare against}

\item{env}{the environment in which to evaluate the expression}
}
\value{
TRUE if the expression evaluates to a scalar value equal to
the specified value, FALSE otherwise
}
\description{
This is used to determine if the location and scale parameters are
actually equal to their default values, which allows for more
efficient translation to expit expressions when possible.
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err-sim.R
\name{rxCombineErrorLines}
\alias{rxCombineErrorLines}
\title{Combine Error Lines and create rxode2 expression}
\usage{
rxCombineErrorLines(
  uiModel,
  errLines = NULL,
  prefixLines = NULL,
  paramsLine = NULL,
  modelVars = FALSE,
  cmtLines = TRUE,
  dvidLine = TRUE,
  lstExpr = NULL,
  useIf = TRUE,
  interpLines = NULL,
  splitDoseLines = NULL,
  levelLines = NULL
)
}
\arguments{
\item{uiModel}{UI model}

\item{errLines}{Error lines; If missing, get the error lines from
\code{rxGetDistributionSimulationLines()}}

\item{prefixLines}{Prefix lines, after param statement}

\item{paramsLine}{Params line, if not present.}

\item{modelVars}{Return model vars instead of rxode2 statement}

\item{cmtLines}{Include trailing \code{cmt} lines}

\item{dvidLine}{Include trailing \code{dvid()} specification}

\item{lstExpr}{A list of expressions for model, or NULL.  When NULL
defaults to the model expressions accessible by
\code{uiModel$lstExpr}.}

\item{useIf}{Use an \verb{if (CMT == X)} for endpoints}

\item{interpLines}{Interpolation lines, if not present}

\item{splitDoseLines}{Split dose lines, if not present}

\item{levelLines}{Levels lines for assigned strings.  If not
present, use the interpolation lines from the current model.}
}
\value{
quoted expression that can be evaluated to compiled rxode2
model
}
\description{
Combine Error Lines and create rxode2 expression
}
\details{
This is exported to allow other functions to mangle the error lines
to make other types of estimation methods (if needed)
}
\examples{

\donttest{

one.cmt <- function() {
  ini({
    ## You may label each parameter with a comment
    tka <- 0.45 # Log Ka
    tcl <- log(c(0, 2.7, 100)) # Log Cl
    ## This works with interactive models
    ## You may also label the preceding line with label("label text")
    tv <- 3.45; label("log V")
    ## the label("Label name") works with all models
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v <- exp(tv + eta.v)
    linCmt() ~ add(add.sd)
  })
}

f <- rxode2(one.cmt)

# You can get the simulation model easily by
rxCombineErrorLines(f)

# You can then get the compiled model by simply evaluting the model:
r <- eval(rxCombineErrorLines(f))

# This also works with multile endpoint models:
pk.turnover.emax <- function() {
  ini({
    tktr <- log(1)
    tka <- log(1)
    tcl <- log(0.1)
    tv <- log(10)
    ##
    eta.ktr ~ 1
    eta.ka ~ 1
    eta.cl ~ 2
    eta.v ~ 1
    prop.err <- 0.1
    pkadd.err <- 0.1
    ##
    temax <- logit(0.8)
    tec50 <- log(0.5)
    tkout <- log(0.05)
    te0 <- log(100)
    ##
    eta.emax ~ .5
    eta.ec50  ~ .5
    eta.kout ~ .5
    eta.e0 ~ .5
    ##
    pdadd.err <- 10
  })
  model({
    ktr <- exp(tktr + eta.ktr)
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v <- exp(tv + eta.v)
    ##
    emax=expit(temax+eta.emax)
    ec50 =  exp(tec50 + eta.ec50)
    kout = exp(tkout + eta.kout)
    e0 = exp(te0 + eta.e0)
    ##
    DCP = center/v
    PD=1-emax*DCP/(ec50+DCP)
    ##
    effect(0) = e0
    kin = e0*kout
    ##
    d/dt(depot) = -ktr * depot
    d/dt(gut) =  ktr * depot -ka * gut
    d/dt(center) =  ka * gut - cl / v * center
    d/dt(effect) = kin*PD -kout*effect
    ##
    cp = center / v
    cp ~ prop(prop.err) + add(pkadd.err)
    effect ~ add(pdadd.err)
  })
}

f <- rxode2(pk.turnover.emax)
rxCombineErrorLines(f)

# Note that in the parsed form, you can also get the compiled rxode2
# model with $simulationModel

f$simulationModel

}
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/softplus.R
\name{softplus}
\alias{softplus}
\title{Softplus Activation Function}
\usage{
softplus(x)
}
\arguments{
\item{x}{numeric vector}
}
\value{
numeric vector
}
\description{
Softplus Activation Function
}
\examples{

softplus(c(-1, 0, 1, 2))

\donttest{
# You can use rxode2 too:

r <- rxode2({
 s <- softplus(time)
})

e <- et(c(-1, 0, 1, 2))

rxSolve(r, e)
}

}
\seealso{
Other Activation Functions:
\code{\link[=ELU]{ELU()}},
\code{\link[=GELU]{GELU()}},
\code{\link[=PReLU]{PReLU()}},
\code{\link[=ReLU]{ReLU()}},
\code{\link[=SELU]{SELU()}},
\code{\link[=Swish]{Swish()}},
\code{\link[=dELU]{dELU()}},
\code{\link[=dGELU]{dGELU()}},
\code{\link[=dPReLU]{dPReLU()}},
\code{\link[=dReLU]{dReLU()}},
\code{\link[=dSELU]{dSELU()}},
\code{\link[=dSwish]{dSwish()}},
\code{\link[=dlReLU]{dlReLU()}},
\code{\link[=dsoftplus]{dsoftplus()}},
\code{\link[=lReLU]{lReLU()}}
}
\author{
Matthew L. Fidler
}
\concept{Activation Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piping-model.R
\name{rxSetCovariateNamesForPiping}
\alias{rxSetCovariateNamesForPiping}
\title{Assign covariates for piping}
\usage{
rxSetCovariateNamesForPiping(covariates = NULL)
}
\arguments{
\item{covariates}{NULL (for no covariates), or the list of
covariates. nlmixr uses this function to set covariates if you
pipe from a nlmixr fit.}
}
\value{
Nothing, called for side effects
}
\description{
Assign covariates for piping
}
\examples{

# First set the name of known covariates
# Note this is case sensitive

rxSetCovariateNamesForPiping(c("WT","HT", "TC"))

one.compartment <- function() {
 ini({
   tka <- 0.45 ; label("Log Ka")
   tcl <- 1 ; label("Log Cl")
   tv <- 3.45 ; label("Log V")
   eta.ka ~ 0.6
   eta.cl ~ 0.3
   eta.v ~ 0.1
   add.err <- 0.7
 })
 model({
   ka <- exp(tka + eta.ka)
   cl <- exp(tcl + eta.cl)
   v <- exp(tv + eta.v)
   d / dt(depot) <- -ka * depot
   d/dt(depot) <- -ka * depot
   d / dt(center) <- ka * depot - cl / v * center
   cp <- center / v
   cp ~ add(add.err)
 })
}

# now TC is detected as a covariate instead of a population parameter

one.compartment |>
  model({ka <- exp(tka + eta.ka + TC * cov_C)})

# You can turn it off by simply adding it back

rxSetCovariateNamesForPiping()

one.compartment |>
  model({ka <- exp(tka + eta.ka + TC * cov_C)})

# The covariates you set with `rxSetCovariateNamesForPiping()`
# are turned off every time you solve (or fit in nlmixr)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piping-update.R
\name{update.rxUi}
\alias{update.rxUi}
\title{Update for rxUi}
\usage{
\method{update}{rxUi}(object, ..., envir = parent.frame())
}
\arguments{
\item{object}{rxode2 UI object}

\item{...}{Lines to update}

\item{envir}{Environment for evaluating ini({}) style calls}
}
\value{
a new rxode2 updated UI object
}
\description{
Update for rxUi
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjoint.R
\name{.rxAdjointGradPop}
\alias{.rxAdjointGradPop}
\title{Adjoint gradient of a POPULATION objective (sum over subjects)}
\usage{
.rxAdjointGradPop(
  object,
  params,
  events,
  calcSens,
  pred,
  data,
  errModel = NULL,
  denseBy = 0.01,
  useC = TRUE,
  atol = 1e-10,
  rtol = 1e-10
)
}
\arguments{
\item{object}{model definition accepted by \code{\link[=rxode2]{rxode2()}} (text,
\code{rxode2} object, or function/ui).}

\item{params}{named numeric vector of parameter values.}

\item{events}{event table / data used to define dosing.}

\item{calcSens}{character vector of parameter names to differentiate with
respect to.}

\item{pred}{character prediction expression \code{f} (function of states and
parameters), e.g. \code{"center/v"}.}

\item{data}{data.frame of observations with columns \code{id}, \code{time}, and \code{dv};
one block of rows per subject.  Any additional columns are treated as
per-subject covariates (first value per subject) merged into \code{params}.}

\item{errModel}{\code{NULL} for least squares, or a list with character entries
\code{add} and/or \code{prop} naming the additive / proportional residual-error
parameters, selecting the FOCEi -2LL objective \code{v = add^2 + (prop*f)^2}.}

\item{denseBy}{grid spacing for the forward checkpoint trajectory; smaller
values reduce the covariate-interpolation error.}

\item{useC}{use the C++ sweep (\code{\link[=.rxAdjointGradEvalC]{.rxAdjointGradEvalC()}}); otherwise the R
reference (\code{\link[=.rxAdjointGradEval]{.rxAdjointGradEval()}}).}

\item{atol, rtol}{solver tolerances used for both the forward checkpoint and
the backward sweeps.}
}
\value{
named numeric vector \code{dOFV/dtheta} over \code{calcSens}.
}
\description{
Computes \verb{dOFV/dtheta = sum_subjects dG_i/dtheta} with one backward sweep per
subject.  The symbolic model is built once (\code{\link[=.rxAdjointGradBuild]{.rxAdjointGradBuild()}}) and
reused per subject.
Assumes a shared dosing regimen across subjects.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxSymInv.R
\name{rxBlockZeros}
\alias{rxBlockZeros}
\title{Creates a logical matrix for block matrixes.}
\usage{
rxBlockZeros(mat, i)
}
\arguments{
\item{mat}{Matrix}

\item{i}{Row/column where block matrix should be setup.}
}
\value{
A logical matrix returning where the elements should be
zero.
}
\description{
Creates a logical matrix for block matrixes.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/err-foceiBase.R
\name{.rxGetLowBoundaryPred1AndIni}
\alias{.rxGetLowBoundaryPred1AndIni}
\title{Get the lower boundary condition when the transformation requires it}
\usage{
.rxGetLowBoundaryPred1AndIni(env, pred1)
}
\arguments{
\item{env}{Environment for the parsed model}

\item{pred1}{The \code{data.frame} of the current error}
}
\value{
Lower Boundary
}
\description{
Get the lower boundary condition when the transformation requires it
}
\author{
Matthew Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{.extraCnow}
\alias{.extraCnow}
\title{Get the extraCnow for compiling}
\usage{
.extraCnow()
}
\value{
string of extraC information
}
\description{
Get the extraCnow for compiling
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxIs}
\alias{rxIs}
\title{Check the type of an object using Rcpp}
\usage{
rxIs(obj, cls)
}
\arguments{
\item{obj}{Object to check}

\item{cls}{Type of class.  Only s3 classes for lists/environments
and primitive classes are checked.  For matrix types they are
distinguished as \code{numeric.matrix}, \code{integer.matrix},
\code{logical.matrix}, and \code{character.matrix} as well as the
traditional \code{matrix} class. Additionally checks for
\code{event.data.frame} which is an \code{data.frame} object with \code{time},
\code{evid} and \code{amt}. (UPPER, lower or Title cases accepted)}
}
\value{
A boolean indicating if the object is a member of the class.
}
\description{
Check the type of an object using Rcpp
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piping-ini.R
\name{zeroRe}
\alias{zeroRe}
\title{Set random effects and residual error to zero}
\usage{
zeroRe(object, which = c("omega", "sigma"), fix = TRUE)
}
\arguments{
\item{object}{The model to modify}

\item{which}{The types of parameters to set to zero}

\item{fix}{Should the parameters be fixed to the zero value?}
}
\value{
The \code{object} with some parameters set to zero
}
\description{
Set random effects and residual error to zero
}
\examples{
one.compartment <- function() {
  ini({
    tka <- log(1.57); label("Ka")
    tcl <- log(2.72); label("Cl")
    tv <- log(31.5); label("V")
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v <- exp(tv + eta.v)
    d/dt(depot) = -ka * depot
    d/dt(center) = ka * depot - cl / v * center
    cp = center / v
    cp ~ add(add.sd)
  })
}
zeroRe(one.compartment)
}
\seealso{
Other Initial conditions:
\code{\link[=ini.rxUi]{ini.rxUi()}}
}
\author{
Bill Denney
}
\concept{Initial conditions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{rxClean}
\alias{rxClean}
\title{Cleanup anonymous DLLs by unloading them}
\usage{
rxClean(wd)
}
\arguments{
\item{wd}{What directory should be cleaned; (DEPRECIATED), this no
longer does anything.

This unloads all rxode2 anonymous dlls.}
}
\value{
TRUE if successful
}
\description{
This cleans up any rxode2 loaded DLLs
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assert.R
\name{assertVariableExists}
\alias{assertVariableExists}
\alias{testVariableExists}
\title{Assert a variable exists in the model}
\usage{
assertVariableExists(ui, x)

testVariableExists(ui, x)
}
\arguments{
\item{ui}{rxode2 ui model}

\item{x}{does the \code{x} variable exist in the model.  If it is a
vector of variable check to see if any exists, but all must be
valid nlmixr2 variable names}
}
\value{
variable that matches, in the case of multiple variables,
the first that matches.  If nothing matches return error
}
\description{
Assert a variable exists in the model
}
\section{Functions}{
\itemize{
\item \code{testVariableExists()}: Test if variable exists

}}
\seealso{
Other Assertions:
\code{\link[=assertCompartmentExists]{assertCompartmentExists()}},
\code{\link[=assertCompartmentName]{assertCompartmentName()}},
\code{\link[=assertCompartmentNew]{assertCompartmentNew()}},
\code{\link[=assertRxUi]{assertRxUi()}},
\code{\link[=assertVariableNew]{assertVariableNew()}},
\code{\link[=testIniDf]{testIniDf()}},
\code{\link[=testRxUnbounded]{testRxUnbounded()}}
}
\author{
Matthew L. Fidler
}
\concept{Assertions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.rxDll}
\alias{print.rxDll}
\title{Print rxDll object}
\usage{
\method{print}{rxDll}(x, ...)
}
\value{
original object
}
\description{
This tells if the rxDll is loaded, ready and/or deleted.
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxUdfUiMv}
\alias{rxUdfUiMv}
\title{Return the model variables that is being processed or setup model
variables for processing}
\usage{
rxUdfUiMv(value)
}
\arguments{
\item{value}{when specified, this assigns the model variables to be
processed, or resets it by assigning it to be \code{NULL}.}
}
\value{
value of the \code{modelVariables} being processed or \code{NULL}.
}
\description{
Return the model variables that is being processed or setup model
variables for processing
}
\examples{

rxUdfUiMv()

}
\seealso{
Other User functions:
\code{\link[=linMod]{linMod()}},
\code{\link[=rxUdfUiControl]{rxUdfUiControl()}},
\code{\link[=rxUdfUiData]{rxUdfUiData()}},
\code{\link[=rxUdfUiEst]{rxUdfUiEst()}},
\code{\link[=rxUdfUiIniLhs]{rxUdfUiIniLhs()}},
\code{\link[=rxUdfUiNum]{rxUdfUiNum()}},
\code{\link[=rxUdfUiParsing]{rxUdfUiParsing()}}
}
\author{
Matthew L. Fidler
}
\concept{User functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-getControl.R
\name{rxAssignControlValue}
\alias{rxAssignControlValue}
\title{Assign Control Variable}
\usage{
rxAssignControlValue(ui, option, value)
}
\arguments{
\item{ui}{rxode2 ui function}

\item{option}{Option name in the control to modify}

\item{value}{Value of control to modify}
}
\value{
Nothing; called for the side effects
}
\description{
Assign Control Variable
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rstudio.R
\name{.rstudioComplete}
\alias{.rstudioComplete}
\title{Determine if Rstudio completion 2025+ is running}
\usage{
.rstudioComplete()
}
\value{
boolean, is Rstudio completion 2025+ running
}
\description{
Determine if Rstudio completion 2025+ is running
}
\examples{
# Determines if rstudio is running completion
.rstudioComplete()
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxUiGet.R
\name{rxUiDevelop}
\alias{rxUiDevelop}
\title{rxUiDevelop - Enable/Disable rxUi development.  Here all $ completions are given}
\usage{
rxUiDevelop(enable = TRUE)
}
\arguments{
\item{enable}{boolean to enable/disable rxUi development mode.}
}
\value{
nothing, called for side effects
}
\description{
rxUiDevelop - Enable/Disable rxUi development.  Here all $ completions are given
}
\examples{

rxUiDevelop(TRUE)
rxUiDevelop(FALSE)
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxLinCmt.R
\name{linToOde}
\alias{linToOde}
\title{Convert linCmt rxUi models to ODE rxUi models}
\usage{
linToOde(ui)
}
\arguments{
\item{ui}{rxUi-like model object}
}
\value{
rxUi model with \code{linCmt()} translated to explicit ODEs
}
\description{
Convert linCmt rxUi models to ODE rxUi models
}
\examples{
oneCmt <- function() {
  ini({
    tka <- 0.45
    tcl <- log(2.7)
    tv <- 3.45
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka)
    cl <- exp(tcl)
    v <- exp(tv)
    cp <- linCmt()
    cp ~ add(add.sd)
  })
}

oneCmtOde <- linToOde(oneCmt)



pkpd <- function() {
  ini({
    ka <- 1
    cl <- 2
    v <- 20
    kin <- 1
    kout <- 1
    ec50 <- 2
  })
  model({
    cp <- linCmt(ka, cl, v)
    eff(0) <- 1
    d/dt(eff) <- kin - kout * (1 - cp/(ec50 + cp)) * eff
  })
}

pkpdOde <- linToOde(pkpd)

}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{.udfMd5Info}
\alias{.udfMd5Info}
\title{Get the udf strings for creating model md5}
\usage{
.udfMd5Info()
}
\value{
string vector
}
\description{
Get the udf strings for creating model md5
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikGeom}
\alias{llikGeom}
\title{log likelihood and derivatives for Geom distribution}
\usage{
llikGeom(x, prob, full = FALSE)
}
\arguments{
\item{x}{variable distributed by a geom distribution}

\item{prob}{probability of success in each trial. \code{0 < prob <= 1}.}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the log pdf value of with \code{dProb}
that has the derivatives with respect to the \code{prob} parameters at
the observation time-point
}
\description{
log likelihood and derivatives for Geom distribution
}
\details{
In an \code{rxode2()} model, you can use \code{llikGeom()} but you have to
use the x and rate arguments.  You can also get the derivative of \code{prob} with
\code{llikGeomDprob()}.
}
\examples{

\donttest{

llikGeom(1:10, 0.2)

et  <- et(1:10)
et$prob <- 0.2
 
model <- function() {
  model({
    fx <- llikGeom(time, prob)
    dProb <- llikGeomDprob(time, prob)
  })
}

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dde.R
\name{.rxPastBaseLinesFromEnv}
\alias{.rxPastBaseLinesFromEnv}
\title{Base past(state, tau) <- expr history lines from a symengine env}
\usage{
.rxPastBaseLinesFromEnv(model)
}
\arguments{
\item{model}{a symengine environment (as from \code{.loadSymengine}/\code{rxS}).}
}
\value{
character vector of past() lines, or NULL if none.
}
\description{
Rebuilds the base \code{past(state,tau)=expr} line(s) from the stored
\code{rx__pastRhs_STATE__} / \code{rx__pastTau_STATE__} text; used by gradient-free
estimators (SAEM) whose symengine env is built without sensitivities.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piping-ini.R
\name{.iniGetAppendArg}
\alias{.iniGetAppendArg}
\title{This gets the append arg for the ini({}) piping}
\usage{
.iniGetAppendArg(f, s)
}
\arguments{
\item{f}{this is the \code{try(force(append))} argument,}

\item{s}{this is the \code{as.character(substitute(append))} argument}
}
\value{
corrected ini piping argument

This is exported for creating new ini methods that have the same
requirements for piping
}
\description{
This gets the append arg for the ini({}) piping
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode-options.R
\name{rxCreateCache}
\alias{rxCreateCache}
\title{This will create the cache directory for rxode2 to save between sessions}
\usage{
rxCreateCache()
}
\value{
nothing
}
\description{
When run, if the \code{R_user_dir} for rxode2's cache isn't present,
create the cache
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxrandom.R
\name{cvPost}
\alias{cvPost}
\title{Sample a covariance Matrix from the Posterior Inverse Wishart
distribution.}
\usage{
cvPost(
  nu,
  omega,
  n = 1L,
  omegaIsChol = FALSE,
  returnChol = FALSE,
  type = c("invWishart", "lkj", "separation"),
  diagXformType = c("log", "identity", "variance", "nlmixrSqrt", "nlmixrLog",
    "nlmixrIdentity")
)
}
\arguments{
\item{nu}{Degrees of Freedom (Number of Observations) for
covariance matrix simulation.}

\item{omega}{Either the estimate of covariance matrix or the
estimated standard deviations in matrix form each row forming
the standard deviation simulated values}

\item{n}{Number of Matrices to sample.  By default this is 1.
This is only useful when \code{omega} is a matrix.  Otherwise
it is determined by the number of rows in the input
\code{omega} matrix of standard deviations}

\item{omegaIsChol}{is an indicator of if the omega matrix is in
the Cholesky decomposition. This is only used when \code{type="invWishart"}}

\item{returnChol}{Return the Cholesky decomposition of the
covariance matrix sample. This is only used when \code{type="invWishart"}}

\item{type}{The type of covariance posterior that is being
simulated.  This can be:
\itemize{
\item \code{invWishart} The posterior is an inverse wishart; This allows
for correlations between parameters to be modeled.  All the
uncertainty in the parameter is captured in the degrees of freedom
parameter.
\item \code{lkj} The posterior separates the standard deviation
estimates (modeled outside and provided in the \code{omega}
argument) and the correlation estimates. The correlation estimate
is simulated with the \code{\link[=rLKJ1]{rLKJ1()}}.  This simulation uses
the relationship \code{eta=(nu-1)/2}.  This is relationship based
on the proof of the relationship between the restricted
LKJ-distribution and inverse wishart distribution (XXXXXX).  Once
the correlation posterior is calculated, the estimated standard
deviations are then combined with the simulated correlation matrix
to create the covariance matrix.
\item \code{separation} Like the \code{lkj} option, this separates out
the estimation of the correlation and standard deviation.  Instead
of using the \code{LKJ} distribution to simulate the correlation,
it simulates the inverse wishart of the identity matrix and
converts the result to a correlation matrix.  This correlation
matrix is then used with the standard deviation to calculate the
simulated covariance matrix.
}}

\item{diagXformType}{Diagonal transformation type.  These could be:
\itemize{
\item \code{log} The standard deviations are log transformed, so the
actual standard deviations are exp(omega)
\item \code{identity} The standard deviations are not transformed. The
standard deviations are not transformed;  They should be positive.
\item \code{variance} The variances are specified in the \code{omega}
matrix; They are transformed into standard deviations.
\item \code{nlmixrSqrt} These standard deviations come from an nlmixr
omega matrix where diag(chol(inv(omega))) = x^2
\item \code{nlmixrLog} These standard deviations come from a nlmixr
omega matrix omega matrix where diag(chol(solve(omega))) = exp(x)
\item \code{nlmixrIdentity} These standard deviations come from a nlmixr
omega matrix omega matrix where diag(chol(solve(omega))) = x
}

The nlmixr transformations only make sense when there is no
off-diagonal correlations modeled.}
}
\value{
a matrix (n=1) or a list of matrices  (n > 1)
}
\description{
Note this Inverse wishart rescaled to match the original scale of
the covariance matrix.
}
\details{
If your covariance matrix is a 1x1 matrix, this uses an scaled
inverse chi-squared which is equivalent to the Inverse Wishart
distribution in the uni-directional case.

In general, the separation strategy is preferred for diagonal
matrices.  If the dimension of the matrix is below 10, \code{lkj}
is numerically faster than \code{separation} method.  However, the
\code{lkj} method has densities too close to zero (XXXX) when the
dimension is above 10.  In that case, though computationally more
expensive \code{separation} method performs better.

For matrices with modeled covariances, the easiest method to use
is the inverse Wishart which allows the simulation of correlation
matrices (XXXX).  This method is more well suited for well behaved
matrices, that is the variance components are not too low or too
high.  When modeling nonlinear mixed effects modeling matrices
with too high or low variances are considered sub-optimal in
describing a system.  With these rules in mind, it is reasonable
to use the inverse Wishart.
}
\examples{

## Sample a single covariance.
draw1 <- cvPost(3, matrix(c(1, .3, .3, 1), 2, 2))

## Sample 3 covariances
set.seed(42)
draw3 <- cvPost(3, matrix(c(1, .3, .3, 1), 2, 2), n = 3)

## Sample 3 covariances, but return the cholesky decomposition
set.seed(42)
draw3c <- cvPost(3, matrix(c(1, .3, .3, 1), 2, 2), n = 3, returnChol = TRUE)

## Sample 3 covariances with lognormal standard deviations via LKJ
## correlation sample
cvPost(3, sapply(1:3, function(...) {
  rnorm(10)
}), type = "lkj")

## or return cholesky decomposition
cvPost(3, sapply(1:3, function(...) {
  rnorm(10)
}),
type = "lkj",
returnChol = TRUE
)

## Sample 3 covariances with lognormal standard deviations via separation
## strategy using inverse Wishart correlation sample
cvPost(3, sapply(1:3, function(...) {
  rnorm(10)
}), type = "separation")

## or returning the cholesky decomposition
cvPost(3, sapply(1:3, function(...) {
  rnorm(10)
}),
type = "separation",
returnChol = TRUE
)
}
\references{
Alvarez I, Niemi J and Simpson M. (2014) \emph{Bayesian Inference for a
Covariance Matrix}. Conference on Applied Statistics in Agriculture.

Wang1 Z, Wu Y, and Chu H. (2018) \emph{On Equivalence of the LKJ
distribution and the restricted Wishart
distribution}. <doi:10.48550/arXiv.1809.047463
}
\author{
Matthew L.Fidler & Wenping Wang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxGetModel}
\alias{rxGetModel}
\title{Get the rxode2 model variables without compiling the model}
\usage{
rxGetModel(
  model,
  calcSens = NULL,
  calcJac = NULL,
  collapseModel = NULL,
  indLin = FALSE,
  calcSens2 = NULL,
  calcSens3 = NULL
)
}
\arguments{
\item{model}{The model to get the model variables from.  It can
be: rxode2 model text (a character string, \code{rxModelText}, or a
\code{{}} expression), a function or call whose body is the model, an
\code{rxode2} object, an \code{rxModelVars} object, an \code{rxDll} object, or
anything else \code{rxModelVars()} accepts.}

\item{calcSens}{boolean indicating if rxode2 will calculate the
sensitivities according to the specified ODEs.  May also be a
character vector of the states/parameters whose first-order
sensitivities (\verb{rx__sens_<state>_BY_<param>__}) should be
generated.}

\item{calcJac}{boolean indicating if rxode2 will calculate the
Jacobain according to the specified ODEs.}

\item{collapseModel}{boolean indicating if rxode2 will remove all
LHS variables when calculating sensitivities.}

\item{indLin}{Calculate inductive linearization matrices and
compile with inductive linearization support.}

\item{calcSens2}{character vector (or \code{NULL}) requesting
second-order sensitivities in addition to the first-order ones
from \code{calcSens}.  When supplied, rxode2 also generates the
\verb{rx__sens_<state>_BY_<p>_BY_<q>__} compartments (the Hessian
path), where \code{p} ranges over \code{calcSens} and \code{q} over
\code{calcSens2}.  Used, for example, for population (\code{THETA})
second-order event sensitivities.  \code{NULL} (the default) skips the
second-order generation.}

\item{calcSens3}{character vector (or \code{NULL}) requesting third-order
sensitivities in addition to the first- and second-order ones.
Requires \code{calcSens2} to also be supplied (every \code{calcSens3}
parameter needs its own already-built second-order sensitivity
compartment, from the pairing of \code{calcSens2} with \code{calcSens3}, to
reference -- so \code{calcSens3} should be a subset of \code{calcSens2},
which itself should be a subset of \code{calcSens}, mirroring how
\code{calcSens2} is used everywhere else in rxode2 today).  When
supplied, generates the \verb{rx__sens_<state>_BY_<p>_BY_<q>_BY_<r>__}
compartments, where \code{p} ranges over \code{calcSens}, \code{q} over
\code{calcSens2}, and \code{r} over \code{calcSens3}, via \code{rxExpandSens3_()}.
\code{NULL} (the default) skips the third-order generation.}
}
\value{
An \code{rxModelVars} object of the (possibly rewritten) model
}
\description{
Normalizes the input to an \code{rxModelVars} object and optionally
rewrites the model text first: adding/stripping forward
sensitivity equations (\code{calcSens}/\code{calcSens2}/\code{calcSens3}),
adding/removing the \code{df()/dy()} Jacobian block (\code{calcJac}),
dropping LHS definitions (\code{collapseModel}), or converting to an
inductive linearization model (\code{indLin}).
}
\examples{

rxGetModel("d/dt(depot) = -ka*depot; d/dt(central) = ka*depot - kel*central")

}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-fix.R
\name{rxFixRes}
\alias{rxFixRes}
\title{Literally fix residual parameters}
\usage{
rxFixRes(ui, returnNull = FALSE)
}
\arguments{
\item{ui}{rxode2 ui function}

\item{returnNull}{boolean for if unchanged values should return the
original ui (\code{FALSE}) or null (\code{TRUE})}
}
\value{
model with residual parameters literally fixed in the model
}
\description{
Literally fix residual parameters
}
\examples{

One.comp.transit.allo <- function() {
 ini({
   # Where initial conditions/variables are specified
   lktr <- log(1.15)  #log k transit (/h)
   lcl  <- log(0.15)  #log Cl (L/hr)
   lv   <- log(7)     #log V (L)
   ALLC <- 0.75  #allometric exponent cl
   ALLV <- 1.00  #allometric exponent v
   prop.err <- fix(0.15)   #proportional error (SD/mean)
   add.err <- fix(0.6)     #additive error (mg/L)
   eta.ktr ~ 0.5
   eta.cl ~ 0.1
   eta.v ~ 0.1
 })
 model({
   #Allometric scaling on weight
   cl <- exp(lcl + eta.cl + ALLC * logWT70)
   v  <- exp(lv + eta.v + ALLV * logWT70)
   ktr <- exp(lktr + eta.ktr)
   # RxODE-style differential equations are supported
   d/dt(depot)   = -ktr * depot
   d/dt(central) =  ktr * trans - (cl/v) * central
   d/dt(trans)   =  ktr * depot - ktr * trans
   ## Concentration is calculated
   cp = central/v
   # And is assumed to follow proportional and additive error
   cp ~ prop(prop.err) + add(add.err)
 })
}

m <- rxFixRes(One.comp.transit.allo)
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikGamma}
\alias{llikGamma}
\title{log likelihood and derivatives for Gamma distribution}
\usage{
llikGamma(x, shape, rate, full = FALSE)
}
\arguments{
\item{x}{variable that is distributed by gamma distribution}

\item{shape}{this is the distribution's shape parameter. Must be positive.}

\item{rate}{this is the distribution's rate parameters.  Must be positive.}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the log pdf value of with \code{dProb}
that has the derivatives with respect to the \code{prob} parameters at
the observation time-point
}
\description{
log likelihood and derivatives for Gamma distribution
}
\details{
In an \code{rxode2()} model, you can use \code{llikGamma()} but you have to
use the x and rate arguments.  You can also get the derivative of \code{shape} or \code{rate} with
\code{llikGammaDshape()} and \code{llikGammaDrate()}.
}
\examples{
\donttest{

llikGamma(1, 1, 10)

# You can use this in `rxode2` too:

et  <- et(seq(0.001, 1, length.out=10))
et$shape <- 1
et$rate <- 10
 
model <- function() {
  model({
    fx <- llikGamma(time, shape, rate)
    dShape<- llikGammaDshape(time, shape, rate)
    dRate <- llikGammaDrate(time, shape, rate)
  })
}

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxExpandFEta_}
\alias{rxExpandFEta_}
\title{Expand d(f)/d(eta)}
\usage{
rxExpandFEta_(state, neta, pred, isTheta = FALSE)
}
\arguments{
\item{state}{is the state to expand}

\item{neta}{is the number of etas}

\item{pred}{type of prediction}

\item{isTheta}{logical, is the expansion actually for thetas instead of etas}
}
\value{
String of symengine expressions to evaluate to calculate df/deta
}
\description{
Expand d(f)/d(eta)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linCmt.R
\name{.solComp2}
\alias{.solComp2}
\title{Calculate the lambdas and coefficients of the two compartment model}
\usage{
.solComp2(k10, k12, k21)
}
\arguments{
\item{k10}{elimination rate}

\item{k12}{rate from central to peripheral compartment}

\item{k21}{rate from peripheral to central compartment}
}
\value{
List with \code{L} vector and matrices \code{C1} and \code{C2}
}
\description{
Calculate the lambdas and coefficients of the two compartment model
}
\examples{
.solComp2(k10=0.1, k12=3, k21=1)
}
\author{
Matthew L. Fidler based on \code{wnl} package/paper, implemented
in C/C++
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modlib.R
\name{rxUse}
\alias{rxUse}
\title{Use model object in your package}
\usage{
rxUse(obj, overwrite = TRUE, compress = "bzip2", internal = FALSE)
}
\arguments{
\item{obj}{model to save.}

\item{overwrite}{By default, \code{use_data()} will not overwrite existing
files. If you really want to do so, set this to \code{TRUE}.}

\item{compress}{Choose the type of compression used by \code{\link[=save]{save()}}.
Should be one of "gzip", "bzip2", or "xz".}

\item{internal}{If this is run internally.  By default this is FALSE}
}
\value{
Nothing; This is used for its side effects and shouldn't be called by a user
}
\description{
Use model object in your package
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assert.R
\name{testRxUnbounded}
\alias{testRxUnbounded}
\alias{assertRxUnbounded}
\alias{warnRxBounded}
\title{Test if the rxode2 model has any parameters with user defined boundaries}
\usage{
testRxUnbounded(ui)

assertRxUnbounded(ui, extra = "", .var.name = .vname(ui))

warnRxBounded(ui, extra = "", .var.name = .vname(ui))
}
\arguments{
\item{ui}{rxode2 ui}

\item{extra}{extra information to append to the error message}

\item{.var.name}{variable name}
}
\value{
boolean indicating if any parameters have user defined boundaries
}
\description{
Test if the rxode2 model has any parameters with user defined boundaries
}
\section{Functions}{
\itemize{
\item \code{assertRxUnbounded()}: Assert that the rxode2 model has any parameters with user defined boundaries

\item \code{warnRxBounded()}: Warn that the rxode2 model has any parameters with user defined boundaries

}}
\examples{

one.cmt <- function() {
  ini({
    tka <- 0.45; label("Ka")
    tcl <- log(c(0, 2.7, 100)); label("Cl")
    tv <- 3.45; label("V")
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v <- exp(tv + eta.v)
    linCmt() ~ add(add.sd)
  })
}

testRxUnbounded(one.cmt)

try(assertRxUnbounded(one.cmt))

warnRxBounded(one.cmt)
}
\seealso{
Other Assertions:
\code{\link[=assertCompartmentExists]{assertCompartmentExists()}},
\code{\link[=assertCompartmentName]{assertCompartmentName()}},
\code{\link[=assertCompartmentNew]{assertCompartmentNew()}},
\code{\link[=assertRxUi]{assertRxUi()}},
\code{\link[=assertVariableExists]{assertVariableExists()}},
\code{\link[=assertVariableNew]{assertVariableNew()}},
\code{\link[=testIniDf]{testIniDf()}}
}
\author{
Matthew L. Fidler
}
\concept{Assertions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelExtract.R
\name{modelExtract}
\alias{modelExtract}
\alias{modelExtract.function}
\alias{modelExtract.rxUi}
\alias{modelExtract.rxode2}
\alias{modelExtract.rxModelVars}
\alias{modelExtract.default}
\title{Extract model lines from a rxui model}
\usage{
modelExtract(
  x,
  ...,
  expression = FALSE,
  endpoint = FALSE,
  lines = FALSE,
  envir = parent.frame()
)

\method{modelExtract}{`function`}(
  x,
  ...,
  expression = FALSE,
  endpoint = FALSE,
  lines = FALSE,
  envir = parent.frame()
)

\method{modelExtract}{rxUi}(
  x,
  ...,
  expression = FALSE,
  endpoint = FALSE,
  lines = FALSE,
  envir = parent.frame()
)

\method{modelExtract}{rxode2}(
  x,
  ...,
  expression = FALSE,
  endpoint = FALSE,
  lines = FALSE,
  envir = parent.frame()
)

\method{modelExtract}{rxModelVars}(
  x,
  ...,
  expression = FALSE,
  endpoint = FALSE,
  lines = FALSE,
  envir = parent.frame()
)

\method{modelExtract}{default}(
  x,
  ...,
  expression = FALSE,
  endpoint = FALSE,
  lines = FALSE,
  envir = parent.frame()
)
}
\arguments{
\item{x}{model to extract lines from}

\item{...}{variables to extract. When it is missing, it will
extract the entire model (conditioned on the endpoint option
below)}

\item{expression}{return expressions (if \code{TRUE}) or strings (if
\code{FALSE})}

\item{endpoint}{include endpoint.  This can be:
\itemize{
\item \code{NA}    -- Missing means include both the endpoint and non-endpoint lines
\item \code{TRUE}  -- Only include endpoint lines
\item \code{FALSE} -- Only include non-endpoint lines
}}

\item{lines}{is a boolean.  When \code{TRUE} this will add the lines as
an attribute to the output value ie \code{attr(, "lines")}}

\item{envir}{Environment for evaluating variables}
}
\value{
expressions or strings of extracted lines. Note if there is
a duplicated lhs expression in the line, it will return both
lines
}
\description{
Extract model lines from a rxui model
}
\examples{

one.compartment <- function() {
  ini({
    tka <- 0.45 # Log Ka
    tcl <- 1 # Log Cl
    tv <- 3.45    # Log V
    eta.ka ~ 0.6
    eta.cl ~ 0.3
    eta.v ~ 0.1
    add.sd <- 0.7
  })
  model({
    ka <- exp(tka + eta.ka)
    cl <- exp(tcl + eta.cl)
    v  <- exp(tv + eta.v)
    d/dt(depot)  <- -ka * depot
    d/dt(center) <-  ka * depot - cl / v * center
    cp <- center / v
    cp ~ add(add.sd)
  })
 }

 f <- one.compartment()

 modelExtract(f, cp)

 modelExtract(one.compartment, d/dt(depot))

 # from variable
 var <- "d/dt(depot)"

 modelExtract(one.compartment, var)

 modelExtract(f, endpoint=NA, lines=TRUE, expression=TRUE)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CTS.R
\name{toTrialDuration}
\alias{toTrialDuration}
\title{Convert event data to trial duration data
A helper function to create a custom event table.
The observation time will start from the first event time (baseline) and end at trial duration.
The interval is the spacing between each observation.}
\usage{
toTrialDuration(ev, trialEnd, interval, writeDir = NULL)
}
\arguments{
\item{ev}{event data}

\item{trialEnd}{extend trial duration. Must be same time unit as event data}

\item{interval}{observation interval. Must be same time unit as event data}

\item{writeDir}{if not NULL, write the output to a csv file}
}
\description{
Convert event data to trial duration data
A helper function to create a custom event table.
The observation time will start from the first event time (baseline) and end at trial duration.
The interval is the spacing between each observation.
}
\examples{
 # Create event table with unique time for each ID
 ev = et(data.frame(id = rep(1:10, 3),  time = runif(min = 10, max = 20, n = 30)))

 # select the duration and spacing interval (assuming time is in years)
 toTrialDuration(ev, trialEnd = 1.5, interval = 0.2)

}
\author{
Omar Elashkar
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reexport.R
\name{guide_none}
\alias{guide_none}
\title{Empty Guide}
\usage{
guide_none(title = waiver(), position = waiver())
}
\arguments{
\item{title}{A character string or expression indicating a title of guide.
If \code{NULL}, the title is not shown. By default
(\code{\link[ggplot2:waiver]{waiver()}}), the name of the scale object or the name
specified in \code{\link[ggplot2:labs]{labs()}} is used for the title.}

\item{position}{Where this guide should be drawn: one of top, bottom,
left, or right.}
}
\value{
nothing, simply included to be compatible with ggplot 3.2
}
\description{
This empty guide draws nothing; It is included in rxode2 for
compatibility with ggplot 3.2
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxSymInvChol}
\alias{rxSymInvChol}
\title{Get Omega^-1 and derivatives}
\usage{
rxSymInvChol(
  invObjOrMatrix,
  theta = NULL,
  type = "cholOmegaInv",
  thetaNumber = 0L
)
}
\arguments{
\item{invObjOrMatrix}{Object for inverse-type calculations.  If
this is a matrix, setup the object for inversion
\code{\link[=rxSymInvCholCreate]{rxSymInvCholCreate()}} with the default arguments and return a
reactive s3 object.  Otherwise, use the inversion object to
calculate the requested derivative/inverse.}

\item{theta}{Thetas to be used for calculation.  If missing (\code{NULL}), a
special s3 class is created and returned to access \code{Omega^1}
objects as needed and cache them based on the theta that is
used.}

\item{type}{The type of object.  Currently the following types are
supported:
\itemize{
\item \code{cholOmegaInv} gives the
Cholesky decomposition of the Omega Inverse matrix.
\item \code{omegaInv} gives the Omega Inverse matrix.
\item \code{d(omegaInv)} gives the \code{d(Omega^-1)} withe respect to the
theta parameter specified in \code{thetaNumber}.
\item \code{d(D)} gives the \code{d(diagonal(Omega^-1))} with respect to
the theta parameter specified in the \code{thetaNumber}
parameter
}}

\item{thetaNumber}{For types \code{d(omegaInv)} and \code{d(D)},
the theta number that the derivative is taken against.  This
must be positive from 1 to the number of thetas defining the
Omega matrix.}
}
\value{
Matrix based on parameters or environment with all the
matrixes calculated in variables \code{omega}, \code{omegaInv}, \code{dOmega},
\code{dOmegaInv}.
}
\description{
Get Omega^-1 and derivatives
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxsolve.R
\name{is.rxSolve}
\alias{is.rxSolve}
\title{Check to see if this is an rxSolve object.}
\usage{
is.rxSolve(x)
}
\arguments{
\item{x}{object to check to see if it is rxSolve

If this is an rxSolve object that has expired strip all rxSolve
information.}
}
\value{
boolean indicating if this is a \code{rxSolve} object
}
\description{
Check to see if this is an rxSolve object.
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-rename.R
\name{.assertRenameErrorModelLine}
\alias{.assertRenameErrorModelLine}
\title{This function asserts the requested rename makes sense}
\usage{
.assertRenameErrorModelLine(line, vars)
}
\arguments{
\item{line}{quoted call information line}

\item{vars}{Variables contained within the mdel}
}
\value{
list(new, old, newChar, oldChar)
}
\description{
It returns the new expression, old expression, new variable and old
variable
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxUdfUiEst}
\alias{rxUdfUiEst}
\title{Return the current estimation method for the UI processing}
\usage{
rxUdfUiEst(value)
}
\arguments{
\item{value}{when specified, this assigns the character value of
the estimation method or NULL if there is nothing being estimated}
}
\value{
value of the estimation method being processed or NULL
}
\description{
Return the current estimation method for the UI processing
}
\examples{

rxUdfUiEst()

}
\seealso{
Other User functions:
\code{\link[=linMod]{linMod()}},
\code{\link[=rxUdfUiControl]{rxUdfUiControl()}},
\code{\link[=rxUdfUiData]{rxUdfUiData()}},
\code{\link[=rxUdfUiIniLhs]{rxUdfUiIniLhs()}},
\code{\link[=rxUdfUiMv]{rxUdfUiMv()}},
\code{\link[=rxUdfUiNum]{rxUdfUiNum()}},
\code{\link[=rxUdfUiParsing]{rxUdfUiParsing()}}
}
\author{
Matthew L. Fidler
}
\concept{User functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getBaseSim.R
\name{getBaseSymengineModel}
\alias{getBaseSymengineModel}
\title{Get the symengine for loading into symengine with \code{rxS()}}
\usage{
getBaseSymengineModel(obj)
}
\arguments{
\item{obj}{Object}
}
\value{
Simulation model ready to load into symeng
}
\description{
Get the symengine for loading into symengine with \code{rxS()}
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assert.R
\name{testIniDf}
\alias{testIniDf}
\alias{assertIniDf}
\title{This function tests if this object is a iniDf as needed by the UI}
\usage{
testIniDf(iniDf)

assertIniDf(iniDf, extra = "", .var.name = .vname(iniDf), null.ok = FALSE)
}
\arguments{
\item{iniDf}{the object to test if it is a rxode2 ui \code{iniDf} data.frame}

\item{extra}{information to append to the error message}

\item{.var.name}{[\code{character(1)}]\cr
Name of the checked object to print in assertions. Defaults to
the heuristic implemented in \code{\link[checkmate]{vname}}.}

\item{null.ok}{[\code{logical(1)}]\cr
If set to \code{TRUE}, \code{x} may also be \code{NULL}.
In this case only a type check of \code{x} is performed, all additional checks are disabled.}
}
\value{
boolean, indicating if the object is a valid initialization data frame
}
\description{
This function tests if this object is a iniDf as needed by the UI
}
\section{Functions}{
\itemize{
\item \code{assertIniDf()}: Assert that the object is a valid rxode2 ui initialization data frame

}}
\examples{
testIniDf(TRUE)
}
\seealso{
Other Assertions:
\code{\link[=assertCompartmentExists]{assertCompartmentExists()}},
\code{\link[=assertCompartmentName]{assertCompartmentName()}},
\code{\link[=assertCompartmentNew]{assertCompartmentNew()}},
\code{\link[=assertRxUi]{assertRxUi()}},
\code{\link[=assertVariableExists]{assertVariableExists()}},
\code{\link[=assertVariableNew]{assertVariableNew()}},
\code{\link[=testRxUnbounded]{testRxUnbounded()}}
}
\author{
Matthew L. Fidler
}
\concept{Assertions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reexport.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{scale_type}
\alias{ggplot}
\alias{aes}
\alias{geom_line}
\alias{facet_wrap}
\alias{ylab}
\alias{xlab}
\alias{waiver}
\alias{lotri}
\alias{label_value}
\alias{label_both}
\alias{label_context}
\alias{label_wrap_gen}
\alias{scale_x_discrete}
\alias{scale_y_discrete}
\alias{scale_x_continuous}
\alias{scale_y_continuous}
\alias{scale_x_date}
\alias{scale_y_date}
\alias{expand_limits}
\title{Objects exported from other packages}
\value{
Inherited from parent routine
}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{ggplot2}{\code{\link[ggplot2:aes]{aes()}}, \code{\link[ggplot2:expand_limits]{expand_limits()}}, \code{\link[ggplot2:facet_wrap]{facet_wrap()}}, \code{\link[ggplot2:geom_line]{geom_line()}}, \code{\link[ggplot2:ggplot]{ggplot()}}, \code{\link[ggplot2:label_both]{label_both()}}, \code{\link[ggplot2:label_context]{label_context()}}, \code{\link[ggplot2:label_context]{label_context()}}, \code{\link[ggplot2:label_value]{label_value()}}, \code{\link[ggplot2:label_wrap_gen]{label_wrap_gen()}}, \code{\link[ggplot2:scale_type]{scale_type()}}, \code{\link[ggplot2:scale_x_continuous]{scale_x_continuous()}}, \code{\link[ggplot2:scale_x_date]{scale_x_date()}}, \code{\link[ggplot2:scale_x_discrete]{scale_x_discrete()}}, \code{\link[ggplot2:scale_y_continuous]{scale_y_continuous()}}, \code{\link[ggplot2:scale_y_date]{scale_y_date()}}, \code{\link[ggplot2:scale_y_discrete]{scale_y_discrete()}}, \code{\link[ggplot2:waiver]{waiver()}}, \code{\link[ggplot2:xlab]{xlab()}}, \code{\link[ggplot2:ylab]{ylab()}}}

  \item{lotri}{\code{\link[lotri:lotri]{lotri()}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symengine.R
\name{rxParsePk}
\alias{rxParsePk}
\title{Parse PK function for inclusion in rxode2}
\usage{
rxParsePk(x, init = NULL)
}
\arguments{
\item{x}{PK function}

\item{init}{Initialization vector}
}
\value{
rxode2 transformed text.
}
\description{
Parse PK function for inclusion in rxode2
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/et.R
\name{etExpand}
\alias{etExpand}
\title{Expand additional doses}
\usage{
etExpand(et)
}
\arguments{
\item{et}{Event table to expand additional doses for.}
}
\value{
New event table with \code{addl} doses expanded
}
\description{
Expand additional doses
}
\examples{
ev <- et(amt = 3, ii = 24, until = 240)
print(ev)
etExpand(ev) # expands event table, but doesn't modify it

print(ev)

ev$expand() ## Expands the current event table and saves it in ev
}
\author{
Matthew Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjointDiscrete.R
\name{.rxDiscreteAdjointGrad}
\alias{.rxDiscreteAdjointGrad}
\title{Discrete adjoint gradient of a trajectory objective}
\usage{
.rxDiscreteAdjointGrad(
  build,
  stages,
  params,
  h,
  obsSteps,
  cov,
  doses = NULL,
  lam0Cov = NULL
)
}
\arguments{
\item{build}{object from \code{\link[=.rxDiscreteAdjointBuild]{.rxDiscreteAdjointBuild()}}.}

\item{stages}{RK4 stage list from \code{\link[=.rxDiscreteForwardSens]{.rxDiscreteForwardSens()}}.}

\item{params}{named numeric parameter vector.}

\item{h}{fixed step size (same as the forward solve).}

\item{obsSteps}{integer step indices at which observation covectors apply.}

\item{cov}{list of covector vectors (length \code{ns}), aligned with \code{obsSteps}.}

\item{doses}{optional list of additive-bolus dose specs (same format as
\code{\link[=.rxDiscreteForwardSens]{.rxDiscreteForwardSens()}}); the exact transpose of the forward dose jump
is applied, contributing \code{amt * dF/dtheta * lambda_c} to \code{dG/dtheta}.}

\item{lam0Cov}{optional initial-condition sensitivity \code{dX0/dtheta} (\verb{ns x np})
to add the terminal transversality \verb{lambda(t0)^T dX0/dtheta}; default none.}
}
\value{
named numeric vector \code{dG/dtheta} over \code{build$calcSens}.
}
\description{
Given the RK4 stages from \code{\link[=.rxDiscreteForwardSens]{.rxDiscreteForwardSens()}} and, per observation
step, the objective covector \verb{c_n = dG/dX_\{step_n\}}, computes \code{dG/dtheta} by
reverse-mode differentiation through the RK4 steps -- equal to
\verb{sum_n c_n^T S_\{step_n\}} to machine precision, in one backward pass.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxIntToLetter}
\alias{rxIntToLetter}
\title{Convert a positive integer to a letter series}
\usage{
rxIntToLetter(x, base = 26L)
}
\arguments{
\item{x}{integer to convert}

\item{base}{can be 2 to 26}
}
\value{
a sequence of letters representing the number(s) input
}
\description{
Convert a positive integer to a letter series
}
\examples{

rxIntToLetter(1:100)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llik.R
\name{llikNbinomMu}
\alias{llikNbinomMu}
\title{Calculate the log likelihood of the negative binomial function (and its derivatives)}
\usage{
llikNbinomMu(x, size, mu, full = FALSE)
}
\arguments{
\item{x}{Number of successes}

\item{size}{Size of trial}

\item{mu}{mu parameter for negative binomial}

\item{full}{Add the data frame showing x, mean, sd as well as the
fx and derivatives}
}
\value{
data frame with \code{fx} for the pdf value of with
\code{dProb} that has the derivatives with respect to the parameters at
the observation time-point
}
\description{
Calculate the log likelihood of the negative binomial function (and its derivatives)
}
\details{
In an \code{rxode2()} model, you can use \code{llikNbinomMu()} but you have to
use all arguments.  You can also get the derivative of \code{mu} with
\code{llikNbinomMuDmu()}
}
\examples{
\donttest{
llikNbinomMu(46:54, 100, 40)

llikNbinomMu(46:54, 100, 40, TRUE)

et <- et(46:54)
et$size <- 100
et$mu <- 40

model <- function() {
  model({
    fx <- llikNbinomMu(time, size, mu)
    dProb <- llikNbinomMuDmu(time, size, mu)
  })
}

rxSolve(model, et)
}
}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxDll}
\alias{rxDll}
\title{Return the DLL associated with the rxode2 object}
\usage{
rxDll(obj)
}
\arguments{
\item{obj}{A rxode2 family of objects or a character string of the
model specification or location of a file with a model
specification.}
}
\value{
a path of the library
}
\description{
This will return the dynamic load library or shared object used to
run the C code for rxode2.
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxUdfUiReset}
\alias{rxUdfUiReset}
\title{Reset the rxode2 ui environment variables}
\usage{
rxUdfUiReset()
}
\value{
NULL silently
}
\description{
Reset the rxode2 ui environment variables
}
\examples{
rxUdfUiReset()
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.convertId}
\alias{.convertId}
\title{Convert a factor/char to an id}
\usage{
.convertId(a)
}
\arguments{
\item{a}{value to convert to an id}
}
\value{
id factor
}
\description{
Convert a factor/char to an id
}
\examples{
.convertId("a")
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.rxTransform}
\alias{.rxTransform}
\title{rxode2 general transformation function}
\usage{
.rxTransform(
  x,
  lambda = 1,
  low = 0,
  high = 1,
  transform = c("boxCox", "yeoJohnson", "untransformed", "lnorm", "logit",
    "logit + yeoJohnson", "probit", "probit + yeoJohnson", "logit + boxCox",
    "probit + boxCox"),
  inverse = FALSE
)
}
\arguments{
\item{x}{value that will be transformed}

\item{lambda}{lambda value for the transformation}

\item{low}{lower bound for the transformation}

\item{high}{upper bound for the transformation}

\item{transform}{transformation to use (can be integer or string
matching supported transformations)}

\item{inverse}{boolean if the inverse transformation should be performed}
}
\value{
transformed value
}
\description{
rxode2 general transformation function
}
\examples{

logit(0.25)

.rxTransform(0.25, transform="logit")

expit(-1.09)

.rxTransform(-1.09, transform="logit", inverse=TRUE)

}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxsolve.R
\name{rxIsNonStiff}
\alias{rxIsNonStiff}
\title{Check whether an ODE solving method is a purely non-stiff solver}
\usage{
rxIsNonStiff(method)
}
\arguments{
\item{method}{A character vector of method names or an integerish vector of
method codes (as returned by \code{\link[=odeMethodToInt]{odeMethodToInt()}}).  Vectorised.}
}
\value{
A logical vector the same length as \code{method}.  \code{TRUE} if the
corresponding method is a purely non-stiff solver.
}
\description{
Returns \code{TRUE} for methods that are designed exclusively for non-stiff
systems.  Solvers like \code{"lsoda"} and \code{"liblsoda"} that automatically switch
between stiff and non-stiff algorithms return \code{FALSE}, as do all stiff-only
methods.
}
\details{
Switchers (\code{"lsoda"} = 1, \code{"liblsoda"} = 2) and the inductive linearisation
solver (\code{"indLin"} = 3) are excluded from both \code{\link[=rxIsStiff]{rxIsStiff()}} and
\code{rxIsNonStiff()}.
}
\examples{
rxIsNonStiff("dop853")                        # TRUE
rxIsNonStiff("lsoda")                         # FALSE (switches)
rxIsNonStiff("bdf")                           # FALSE (stiff-only)
rxIsNonStiff(c("lsode", "cvode", "bs32"))     # TRUE FALSE TRUE

}
\seealso{
\code{\link[=rxIsStiff]{rxIsStiff()}}, \code{\link[=rxIsImplicit]{rxIsImplicit()}}, \code{\link[=odeMethodToInt]{odeMethodToInt()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{.extraC}
\alias{.extraC}
\title{Generate extraC information for rxode2 models}
\usage{
.extraC(extraC = NULL)
}
\arguments{
\item{extraC}{Additional extraC from rxode2 compile optioioins}
}
\value{
Nothing, called for side effects
}
\description{
Generate extraC information for rxode2 models
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudfui.R
\name{rxIntToBase}
\alias{rxIntToBase}
\title{Convert a positive  base}
\usage{
rxIntToBase(x, base = 36L)
}
\arguments{
\item{x}{integer to convert}

\item{base}{can be 2 to 36}
}
\value{
a sequence of letters and representing the number(s) input
}
\description{
Convert a positive  base
}
\examples{

rxIntToBase(1:100)

}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rudf.R
\name{.udfCallFunArg}
\alias{.udfCallFunArg}
\title{Get the function name with the current arguments as a string}
\usage{
.udfCallFunArg(fun, args)
}
\arguments{
\item{fun}{function name}

\item{args}{arguments}
}
\value{
string of the form 'fun(arg1, arg2)':
}
\description{
Get the function name with the current arguments as a string
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode2.R
\name{rxMd5}
\alias{rxMd5}
\title{Return the md5 of an rxode2 object or file}
\usage{
rxMd5(model, ...)
}
\arguments{
\item{model}{This is the ODE model specification.  It can be:
\itemize{
\item a string containing the set of ordinary differential
equations (ODE) and other expressions defining the changes in
the dynamic system.
\item a file name where the ODE system equation is contained
}

An ODE expression enclosed in \verb{\\\{\\\}}

(see also the \code{filename} argument). For
details, see the sections \dQuote{Details} and
\verb{rxode2 Syntax} below.}

\item{...}{ignored arguments}
}
\value{
If this is a rxode2 object, return a named list:
\itemize{
\item \code{file_md5} is the model's file's md5
\item \code{parsed_md5}  is the parsed model's file's md5.
}

Otherwise return the md5 based on the arguments provided
}
\description{
This md5 is based on the model and possibly the extra c code
supplied for the model.  In addition the md5 is based on syntax
options, compiled rxode2 library md5, and the rxode2
version/repository.
}
\author{
Matthew L.Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forder.R
\name{forderForceBase}
\alias{forderForceBase}
\title{Force using base order for rxode2 radix sorting}
\usage{
forderForceBase(forceBase = FALSE)
}
\arguments{
\item{forceBase}{boolean indicating if rxode2 should use R's
\code{\link[=order]{order()}} for radix sorting instead of
\code{data.table}'s parallel radix sorting.}
}
\value{
value of \code{forceBase} (can change if \code{data.table} is not
available)
}
\description{
Force using base order for rxode2 radix sorting
}
\examples{
\donttest{
forderForceBase(TRUE) # Use base `order` for rxode2 sorts
forderForceBase(FALSE) # Use `data.table` for rxode2 sorts
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/version.R
\name{rxVersion}
\alias{rxVersion}
\title{Version and repository for this dparser package.}
\usage{
rxVersion(extra = "", echo = FALSE)
}
\arguments{
\item{extra}{extra text to display in the logo}

\item{echo}{Boolean to echo the text logo, by default FALSE}
}
\value{
A character vector with the version and repository.
}
\description{
Version and repository for this dparser package.
}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.rxTransInfo}
\alias{.rxTransInfo}
\title{Get the information about the rxode2 derived parameter transformation}
\usage{
.rxTransInfo(...)
}
\arguments{
\item{...}{Parameters translated, should be unquoted and not assigned to anything.}
}
\value{
Translation information; This list contains:
\itemize{
\item \verb{$str} A named string of the parameters as seen in the underlying C/C++
code. The parameters that are NA are not used in the linear
compartment model calculations.
\item \verb{$ncmt} the number of compartments in the model
\item \verb{$trans} the rxode2 translation number of the parameterization
}

This contains the linCmt()
translation number, the number of compartments and the parameters
}
\description{
Get the information about the rxode2 derived parameter transformation
}
\examples{

.rxTransInfo(cl, v , Vp, vp2, q, q2)

.rxTransInfo(k12, k21, k13, k31, kel, v)

.rxTransInfo(k12, k21, k13, k31, kel, v, ka)

.rxTransInfo(CL, V)

}
\author{
Matthew L. Fidler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlogit.R
\name{mexpit}
\alias{mexpit}
\alias{dmexpit}
\title{mexpit -- Convert log-scale numbers to probabilities}
\usage{
mexpit(...)

dmexpit(...)
}
\arguments{
\item{...}{numeric log-scale numbers to convert to probabilities.}
}
\value{
Probabilities that add up to a number less than 1.
}
\description{
This function converts log-scale numbers to probabilities.
}
\details{
The probabilities are calculated using the following equation:

\deqn{p_i = \frac{e^{x_i}}{1+\sum_{j=1}^{N-1} e^{x_j}}}

This ensures one remaining probability will add to one, that is

\deqn{p_N = \frac{1}{1+\sum_{j=1}^{N-1} e^{x_j}}}

For the function \code{dmexpit()}, the element-wise derivatives are calculated; that is, it returns the diagonal of the Jacobian matrix, \eqn{d p_i / d x_i}, not the full Jacobian with off-diagonal terms.
}
\examples{

m <- mlogit(0.1, 0.2, 0.3)
mexpit(m)

# derivatives
dmexpit(m)

p <- mexpit(-3, 0.5, 3)
mlogit(p)


}
\author{
Matthew L. Fidler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{rxExpandSens2_}
\alias{rxExpandSens2_}
\title{Expand second order sensitivity}
\usage{
rxExpandSens2_(state, s1, s2)
}
\arguments{
\item{state}{is the state to expand}

\item{s1}{is the variables to calculate sensitivity}

\item{s2}{is the variables to calculate second order sensitivity}
}
\value{
string for symengine second order sensitivity
}
\description{
Expand second order sensitivity
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxRaw.R
\name{rxGetDefaultSerialize}
\alias{rxGetDefaultSerialize}
\title{Get the Default Serialization Type}
\usage{
rxGetDefaultSerialize()
}
\value{
string indicating the default serialization type
}
\description{
Get the Default Serialization Type
}
\examples{

rxGetDefaultSerialize()

}
\author{
Matthew L. Fidler
}
